{!Wrapper uses FHIRBase, FHIRBase_Wrapper, FHIRTypes, FHIRTypes_Wrapper, DateAndTime, DateAndTime_Wrapper}

unit FHIRResources;

{
  Copyright (c) 2011+, HL7, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without modification, 
  are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice, this 
     list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the following disclaimer in the documentation 
     and/or other materials provided with the distribution.
   * Neither the name of HL7 nor the names of its contributors may be used to 
     endorse or promote products derived from this software without specific 
     prior written permission.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
  POSSIBILITY OF SUCH DAMAGE.
  
}


{$IFDEF FHIR-DSTU}
This is the dev branch of the FHIR code
{$ENDIF}

interface

// FHIR v1.0.2 generated Sat, Oct 24, 2015 07:41+1100

uses
  SysUtils, Classes, StringSupport, DecimalSupport, AdvBuffers, DateAndTime, FHIRBase, FHIRTypes;

Type
  {@Enum TFhirResourceType
    Enumeration of known resource types
  }
  TFhirResourceType = (
    frtNull, {@enum.value Resource type not known / not Specified }
    frtAccount, {@enum.value A financial tool for tracking value accrued for a particular purpose.  In the healthcare field, used to track charges for a patient, cost centres, etc. }
    frtAllergyIntolerance, {@enum.value Risk of harmful or undesirable, physiological response which is unique to an individual and associated with exposure to a substance. }
    frtAppointment, {@enum.value A booking of a healthcare event among patient(s), practitioner(s), related person(s) and/or device(s) for a specific date/time. This may result in one or more Encounter(s). }
    frtAppointmentResponse, {@enum.value A reply to an appointment request for a patient and/or practitioner(s), such as a confirmation or rejection. }
    frtAuditEvent, {@enum.value A record of an event made for purposes of maintaining a security log. Typical uses include detection of intrusion attempts and monitoring for inappropriate usage. }
    frtBasic, {@enum.value Basic is used for handling concepts not yet defined in FHIR, narrative-only resources that don't map to an existing resource, and custom resources not appropriate for inclusion in the FHIR specification. }
    frtBinary, {@enum.value A binary resource can contain any content, whether text, image, pdf, zip archive, etc. }
    frtBodySite, {@enum.value Record details about the anatomical location of a specimen or body part.  This resource may be used when a coded concept does not provide the necessary detail needed for the use case. }
    frtBundle, {@enum.value A container for a collection of resources. }
    frtCarePlan, {@enum.value Describes the intention of how one or more practitioners intend to deliver care for a particular patient, group or community for a period of time, possibly limited to care for a specific condition or set of conditions. }
    frtClaim, {@enum.value A provider issued list of services and products provided, or to be provided, to a patient which is provided to an insurer for payment recovery. }
    frtClaimResponse, {@enum.value This resource provides the adjudication details from the processing of a Claim resource. }
    frtClinicalImpression, {@enum.value A record of a clinical assessment performed to determine what problem(s) may affect the patient and before planning the treatments or management strategies that are best to manage a patient's condition. Assessments are often 1:1 with a clinical consultation / encounter,  but this varies greatly depending on the clinical workflow. This resource is called "ClinicalImpression" rather than "ClinicalAssessment" to avoid confusion with the recording of assessment tools such as Apgar score. }
    frtCommunication, {@enum.value An occurrence of information being transmitted; e.g. an alert that was sent to a responsible provider, a public health agency was notified about a reportable condition. }
    frtCommunicationRequest, {@enum.value A request to convey information; e.g. the CDS system proposes that an alert be sent to a responsible provider, the CDS system proposes that the public health agency be notified about a reportable condition. }
    frtComposition, {@enum.value A set of healthcare-related information that is assembled together into a single logical document that provides a single coherent statement of meaning, establishes its own context and that has clinical attestation with regard to who is making the statement. While a Composition defines the structure, it does not actually contain the content: rather the full content of a document is contained in a Bundle, of which the Composition is the first resource contained. }
    frtConceptMap, {@enum.value A statement of relationships from one set of concepts to one or more other concepts - either code systems or data elements, or classes in class models. }
    frtCondition, {@enum.value Use to record detailed information about conditions, problems or diagnoses recognized by a clinician. There are many uses including: recording a diagnosis during an encounter; populating a problem list or a summary statement, such as a discharge summary. }
    frtConformance, {@enum.value A conformance statement is a set of capabilities of a FHIR Server that may be used as a statement of actual server functionality or a statement of required or desired server implementation. }
    frtContract, {@enum.value A formal agreement between parties regarding the conduct of business, exchange of information or other matters. }
    frtCoverage, {@enum.value Financial instrument which may be used to pay for or reimburse health care products and services. }
    frtDataElement, {@enum.value The formal description of a single piece of information that can be gathered and reported. }
    frtDetectedIssue, {@enum.value Indicates an actual or potential clinical issue with or between one or more active or proposed clinical actions for a patient; e.g. Drug-drug interaction, Ineffective treatment frequency, Procedure-condition conflict, etc. }
    frtDevice, {@enum.value This resource identifies an instance of a manufactured item that is used in the provision of healthcare without being substantially changed through that activity. The device may be a medical or non-medical device.  Medical devices includes durable (reusable) medical equipment, implantable devices, as well as disposable equipment used for diagnostic, treatment, and research for healthcare and public health.  Non-medical devices may include items such as a machine, cellphone, computer, application, etc. }
    frtDeviceComponent, {@enum.value Describes the characteristics, operational status and capabilities of a medical-related component of a medical device. }
    frtDeviceMetric, {@enum.value Describes a measurement, calculation or setting capability of a medical device. }
    frtDeviceUseRequest, {@enum.value Represents a request for a patient to employ a medical device. The device may be an implantable device, or an external assistive device, such as a walker. }
    frtDeviceUseStatement, {@enum.value A record of a device being used by a patient where the record is the result of a report from the patient or another clinician. }
    frtDiagnosticOrder, {@enum.value A record of a request for a diagnostic investigation service to be performed. }
    frtDiagnosticReport, {@enum.value The findings and interpretation of diagnostic  tests performed on patients, groups of patients, devices, and locations, and/or specimens derived from these. The report includes clinical context such as requesting and provider information, and some mix of atomic results, images, textual and coded interpretations, and formatted representation of diagnostic reports. }
    frtDocumentManifest, {@enum.value A manifest that defines a set of documents. }
    frtDocumentReference, {@enum.value A reference to a document . }
    frtEligibilityRequest, {@enum.value This resource provides the insurance eligibility details from the insurer regarding a specified coverage and optionally some class of service. }
    frtEligibilityResponse, {@enum.value This resource provides eligibility and plan details from the processing of an Eligibility resource. }
    frtEncounter, {@enum.value An interaction between a patient and healthcare provider(s) for the purpose of providing healthcare service(s) or assessing the health status of a patient. }
    frtEnrollmentRequest, {@enum.value This resource provides the insurance enrollment details to the insurer regarding a specified coverage. }
    frtEnrollmentResponse, {@enum.value This resource provides enrollment and plan details from the processing of an Enrollment resource. }
    frtEpisodeOfCare, {@enum.value An association between a patient and an organization / healthcare provider(s) during which time encounters may occur. The managing organization assumes a level of responsibility for the patient during this time. }
    frtExplanationOfBenefit, {@enum.value This resource provides: the claim details; adjudication details from the processing of a Claim; and optionally account balance information, for informing the subscriber of the benefits provided. }
    frtFamilyMemberHistory, {@enum.value Significant health events and conditions for a person related to the patient relevant in the context of care for the patient. }
    frtFlag, {@enum.value Prospective warnings of potential issues when providing care to the patient. }
    frtGoal, {@enum.value Describes the intended objective(s) for a patient, group or organization care, for example, weight loss, restoring an activity of daily living, obtaining herd immunity via immunization, meeting a process improvement objective, etc. }
    frtGroup, {@enum.value Represents a defined collection of entities that may be discussed or acted upon collectively but which are not expected to act collectively and are not formally or legally recognized; i.e. a collection of entities that isn't an Organization. }
    frtHealthcareService, {@enum.value The details of a healthcare service available at a location. }
    frtImagingObjectSelection, {@enum.value A manifest of a set of DICOM Service-Object Pair Instances (SOP Instances).  The referenced SOP Instances (images or other content) are for a single patient, and may be from one or more studies. The referenced SOP Instances have been selected for a purpose, such as quality assurance, conference, or consult. Reflecting that range of purposes, typical ImagingObjectSelection resources may include all SOP Instances in a study (perhaps for sharing through a Health Information Exchange); key images from multiple studies (for reference by a referring or treating physician); a multi-frame ultrasound instance ("cine" video clip) and a set of measurements taken from that instance (for inclusion in a teaching file); and so on. }
    frtImagingStudy, {@enum.value Representation of the content produced in a DICOM imaging study. A study comprises a set of series, each of which includes a set of Service-Object Pair Instances (SOP Instances - images or other data) acquired or produced in a common context.  A series is of only one modality (e.g. X-ray, CT, MR, ultrasound), but a study may have multiple series of different modalities. }
    frtImmunization, {@enum.value Describes the event of a patient being administered a vaccination or a record of a vaccination as reported by a patient, a clinician or another party and may include vaccine reaction information and what vaccination protocol was followed. }
    frtImmunizationRecommendation, {@enum.value A patient's point-in-time immunization and recommendation (i.e. forecasting a patient's immunization eligibility according to a published schedule) with optional supporting justification. }
    frtImplementationGuide, {@enum.value A set of rules or how FHIR is used to solve a particular problem. This resource is used to gather all the parts of an implementation guide into a logical whole, and to publish a computable definition of all the parts. }
    frtList, {@enum.value A set of information summarized from a list of other resources. }
    frtLocation, {@enum.value Details and position information for a physical place where services are provided  and resources and participants may be stored, found, contained or accommodated. }
    frtMedia, {@enum.value A photo, video, or audio recording acquired or used in healthcare. The actual content may be inline or provided by direct reference. }
    frtMedication, {@enum.value This resource is primarily used for the identification and definition of a medication. It covers the ingredients and the packaging for a medication. }
    frtMedicationAdministration, {@enum.value Describes the event of a patient consuming or otherwise being administered a medication.  This may be as simple as swallowing a tablet or it may be a long running infusion.  Related resources tie this event to the authorizing prescription, and the specific encounter between patient and health care practitioner. }
    frtMedicationDispense, {@enum.value Indicates that a medication product is to be or has been dispensed for a named person/patient.  This includes a description of the medication product (supply) provided and the instructions for administering the medication.  The medication dispense is the result of a pharmacy system responding to a medication order. }
    frtMedicationOrder, {@enum.value An order for both supply of the medication and the instructions for administration of the medication to a patient. The resource is called "MedicationOrder" rather than "MedicationPrescription" to generalize the use across inpatient and outpatient settings as well as for care plans, etc. }
    frtMedicationStatement, {@enum.value A record of a medication that is being consumed by a patient.   A MedicationStatement may indicate that the patient may be taking the medication now, or has taken the medication in the past or will be taking the medication in the future.  The source of this information can be the patient, significant other (such as a family member or spouse), or a clinician.  A common scenario where this information is captured is during the history taking process during a patient visit or stay.   The medication information may come from e.g. the patient's memory, from a prescription bottle,  or from a list of medications the patient, clinician or other party maintains 

The primary difference between a medication statement and a medication administration is that the medication administration has complete administration information and is based on actual administration information from the person who administered the medication.  A medication statement is often, if not always, less specific.  There is no required date/time when the medication was administered, in fact we only know that a source has reported the patient is taking this medication, where details such as time, quantity, or rate or even medication product may be incomplete or missing or less precise.  As stated earlier, the medication statement information may come from the patient's memory, from a prescription bottle or from a list of medications the patient, clinician or other party maintains.  Medication administration is more formal and is not missing detailed information. }
    frtMessageHeader, {@enum.value The header for a message exchange that is either requesting or responding to an action.  The reference(s) that are the subject of the action as well as other information related to the action are typically transmitted in a bundle in which the MessageHeader resource instance is the first resource in the bundle. }
    frtNamingSystem, {@enum.value A curated namespace that issues unique symbols within that namespace for the identification of concepts, people, devices, etc.  Represents a "System" used within the Identifier and Coding data types. }
    frtNutritionOrder, {@enum.value A request to supply a diet, formula feeding (enteral) or oral nutritional supplement to a patient/resident. }
    frtObservation, {@enum.value Measurements and simple assertions made about a patient, device or other subject. }
    frtOperationDefinition, {@enum.value A formal computable definition of an operation (on the RESTful interface) or a named query (using the search interaction). }
    frtOperationOutcome, {@enum.value A collection of error, warning or information messages that result from a system action. }
    frtOrder, {@enum.value A request to perform an action. }
    frtOrderResponse, {@enum.value A response to an order. }
    frtOrganization, {@enum.value A formally or informally recognized grouping of people or organizations formed for the purpose of achieving some form of collective action.  Includes companies, institutions, corporations, departments, community groups, healthcare practice groups, etc. }
    frtParameters, {@enum.value This special resource type is used to represent an operation request and response (operations.html). It has no other use, and there is no RESTful endpoint associated with it. }
    frtPatient, {@enum.value Demographics and other administrative information about an individual or animal receiving care or other health-related services. }
    frtPaymentNotice, {@enum.value This resource provides the status of the payment for goods and services rendered, and the request and response resource references. }
    frtPaymentReconciliation, {@enum.value This resource provides payment details and claim references supporting a bulk payment. }
    frtPerson, {@enum.value Demographics and administrative information about a person independent of a specific health-related context. }
    frtPractitioner, {@enum.value A person who is directly or indirectly involved in the provisioning of healthcare. }
    frtProcedure, {@enum.value An action that is or was performed on a patient. This can be a physical intervention like an operation, or less invasive like counseling or hypnotherapy. }
    frtProcedureRequest, {@enum.value A request for a procedure to be performed. May be a proposal or an order. }
    frtProcessRequest, {@enum.value This resource provides the target, request and response, and action details for an action to be performed by the target on or about existing resources. }
    frtProcessResponse, {@enum.value This resource provides processing status, errors and notes from the processing of a resource. }
    frtProvenance, {@enum.value Provenance of a resource is a record that describes entities and processes involved in producing and delivering or otherwise influencing that resource. Provenance provides a critical foundation for assessing authenticity, enabling trust, and allowing reproducibility. Provenance assertions are a form of contextual metadata and can themselves become important records with their own provenance. Provenance statement indicates clinical significance in terms of confidence in authenticity, reliability, and trustworthiness, integrity, and stage in lifecycle (e.g. Document Completion - has the artifact been legally authenticated), all of which may impact security, privacy, and trust policies. }
    frtQuestionnaire, {@enum.value A structured set of questions intended to guide the collection of answers. The questions are ordered and grouped into coherent subsets, corresponding to the structure of the grouping of the underlying questions. }
    frtQuestionnaireResponse, {@enum.value A structured set of questions and their answers. The questions are ordered and grouped into coherent subsets, corresponding to the structure of the grouping of the underlying questions. }
    frtReferralRequest, {@enum.value Used to record and send details about a request for referral service or transfer of a patient to the care of another provider or provider organization. }
    frtRelatedPerson, {@enum.value Information about a person that is involved in the care for a patient, but who is not the target of healthcare, nor has a formal responsibility in the care process. }
    frtRiskAssessment, {@enum.value An assessment of the likely outcome(s) for a patient or other subject as well as the likelihood of each outcome. }
    frtSchedule, {@enum.value A container for slot(s) of time that may be available for booking appointments. }
    frtSearchParameter, {@enum.value A search parameter that defines a named search item that can be used to search/filter on a resource. }
    frtSlot, {@enum.value A slot of time on a schedule that may be available for booking appointments. }
    frtSpecimen, {@enum.value A sample to be used for analysis. }
    frtStructureDefinition, {@enum.value A definition of a FHIR structure. This resource is used to describe the underlying resources, data types defined in FHIR, and also for describing extensions, and constraints on resources and data types. }
    frtSubscription, {@enum.value The subscription resource is used to define a push based subscription from a server to another system. Once a subscription is registered with the server, the server checks every resource that is created or updated, and if the resource matches the given criteria, it sends a message on the defined "channel" so that another system is able to take an appropriate action. }
    frtSubstance, {@enum.value A homogeneous material with a definite composition. }
    frtSupplyDelivery, {@enum.value Record of delivery of what is supplied. }
    frtSupplyRequest, {@enum.value A record of a request for a medication, substance or device used in the healthcare setting. }
    frtTestScript, {@enum.value TestScript is a resource that specifies a suite of tests against a FHIR server implementation to determine compliance against the FHIR specification. }
    frtValueSet, {@enum.value A value set specifies a set of codes drawn from one or more code systems. }
    frtVisionPrescription); {@enum.value An authorization for the supply of glasses and/or contact lenses to a patient. }

  TFhirResourceTypeSet = set of TFhirResourceType;
  {@Enum TSearchParamsAccount
    Search Parameters for Account
  }
  TSearchParamsAccount = (
    spAccount__content, {@enum.value "_content" spAccount__content Search on the entire content of the resource }
    spAccount__id, {@enum.value "_id" spAccount__id Logical id of this artifact }
    spAccount__lastUpdated, {@enum.value "_lastUpdated" spAccount__lastUpdated When the resource version last changed }
    spAccount__profile, {@enum.value "_profile" spAccount__profile Profiles this resource claims to conform to }
    spAccount__query, {@enum.value "_query" spAccount__query A custom search profile that describes a specific defined query operation }
    spAccount__security, {@enum.value "_security" spAccount__security Security Labels applied to this resource }
    spAccount__tag, {@enum.value "_tag" spAccount__tag Tags applied to this resource }
    spAccount__text, {@enum.value "_text" spAccount__text Search on the narrative of the resource }
    spAccount_Balance, {@enum.value "balance" spAccount_Balance How much is in account? }
    spAccount_Identifier, {@enum.value "identifier" spAccount_Identifier Account number }
    spAccount_Name, {@enum.value "name" spAccount_Name Human-readable label }
    spAccount_Owner, {@enum.value "owner" spAccount_Owner Who is responsible? }
    spAccount_Patient, {@enum.value "patient" spAccount_Patient What is account tied to? }
    spAccount_Period, {@enum.value "period" spAccount_Period Transaction window }
    spAccount_Status, {@enum.value "status" spAccount_Status active | inactive }
    spAccount_Subject, {@enum.value "subject" spAccount_Subject What is account tied to? }
    spAccount_Type); {@enum.value "type" spAccount_Type E.g. patient, expense, depreciation }

  {@Enum TSearchParamsAllergyIntolerance
    Search Parameters for AllergyIntolerance
  }
  TSearchParamsAllergyIntolerance = (
    spAllergyIntolerance__content, {@enum.value "_content" spAllergyIntolerance__content Search on the entire content of the resource }
    spAllergyIntolerance__id, {@enum.value "_id" spAllergyIntolerance__id Logical id of this artifact }
    spAllergyIntolerance__lastUpdated, {@enum.value "_lastUpdated" spAllergyIntolerance__lastUpdated When the resource version last changed }
    spAllergyIntolerance__profile, {@enum.value "_profile" spAllergyIntolerance__profile Profiles this resource claims to conform to }
    spAllergyIntolerance__query, {@enum.value "_query" spAllergyIntolerance__query A custom search profile that describes a specific defined query operation }
    spAllergyIntolerance__security, {@enum.value "_security" spAllergyIntolerance__security Security Labels applied to this resource }
    spAllergyIntolerance__tag, {@enum.value "_tag" spAllergyIntolerance__tag Tags applied to this resource }
    spAllergyIntolerance__text, {@enum.value "_text" spAllergyIntolerance__text Search on the narrative of the resource }
    spAllergyIntolerance_Category, {@enum.value "category" spAllergyIntolerance_Category food | medication | environment | other - Category of Substance }
    spAllergyIntolerance_Criticality, {@enum.value "criticality" spAllergyIntolerance_Criticality CRITL | CRITH | CRITU }
    spAllergyIntolerance_Date, {@enum.value "date" spAllergyIntolerance_Date When recorded }
    spAllergyIntolerance_Identifier, {@enum.value "identifier" spAllergyIntolerance_Identifier External ids for this item }
    spAllergyIntolerance_Last_date, {@enum.value "last-date" spAllergyIntolerance_Last_date Date(/time) of last known occurrence of a reaction }
    spAllergyIntolerance_Manifestation, {@enum.value "manifestation" spAllergyIntolerance_Manifestation Clinical symptoms/signs associated with the Event }
    spAllergyIntolerance_Onset, {@enum.value "onset" spAllergyIntolerance_Onset Date(/time) when manifestations showed }
    spAllergyIntolerance_Patient, {@enum.value "patient" spAllergyIntolerance_Patient Who the sensitivity is for }
    spAllergyIntolerance_Recorder, {@enum.value "recorder" spAllergyIntolerance_Recorder Who recorded the sensitivity }
    spAllergyIntolerance_Reporter, {@enum.value "reporter" spAllergyIntolerance_Reporter Source of the information about the allergy }
    spAllergyIntolerance_Route, {@enum.value "route" spAllergyIntolerance_Route How the subject was exposed to the substance }
    spAllergyIntolerance_Severity, {@enum.value "severity" spAllergyIntolerance_Severity mild | moderate | severe (of event as a whole) }
    spAllergyIntolerance_Status, {@enum.value "status" spAllergyIntolerance_Status active | unconfirmed | confirmed | inactive | resolved | refuted | entered-in-error }
    spAllergyIntolerance_Substance, {@enum.value "substance" spAllergyIntolerance_Substance Substance, (or class) considered to be responsible for risk }
    spAllergyIntolerance_Type); {@enum.value "type" spAllergyIntolerance_Type allergy | intolerance - Underlying mechanism (if known) }

  {@Enum TSearchParamsAppointment
    Search Parameters for Appointment
  }
  TSearchParamsAppointment = (
    spAppointment__content, {@enum.value "_content" spAppointment__content Search on the entire content of the resource }
    spAppointment__id, {@enum.value "_id" spAppointment__id Logical id of this artifact }
    spAppointment__lastUpdated, {@enum.value "_lastUpdated" spAppointment__lastUpdated When the resource version last changed }
    spAppointment__profile, {@enum.value "_profile" spAppointment__profile Profiles this resource claims to conform to }
    spAppointment__query, {@enum.value "_query" spAppointment__query A custom search profile that describes a specific defined query operation }
    spAppointment__security, {@enum.value "_security" spAppointment__security Security Labels applied to this resource }
    spAppointment__tag, {@enum.value "_tag" spAppointment__tag Tags applied to this resource }
    spAppointment__text, {@enum.value "_text" spAppointment__text Search on the narrative of the resource }
    spAppointment_Actor, {@enum.value "actor" spAppointment_Actor Any one of the individuals participating in the appointment }
    spAppointment_Date, {@enum.value "date" spAppointment_Date Appointment date/time. }
    spAppointment_Identifier, {@enum.value "identifier" spAppointment_Identifier An Identifier of the Appointment }
    spAppointment_Location, {@enum.value "location" spAppointment_Location This location is listed in the participants of the appointment }
    spAppointment_Part_status, {@enum.value "part-status" spAppointment_Part_status The Participation status of the subject, or other participant on the appointment. Can be used to locate participants that have not responded to meeting requests. }
    spAppointment_Patient, {@enum.value "patient" spAppointment_Patient One of the individuals of the appointment is this patient }
    spAppointment_Practitioner, {@enum.value "practitioner" spAppointment_Practitioner One of the individuals of the appointment is this practitioner }
    spAppointment_Status); {@enum.value "status" spAppointment_Status The overall status of the appointment }

  {@Enum TSearchParamsAppointmentResponse
    Search Parameters for AppointmentResponse
  }
  TSearchParamsAppointmentResponse = (
    spAppointmentResponse__content, {@enum.value "_content" spAppointmentResponse__content Search on the entire content of the resource }
    spAppointmentResponse__id, {@enum.value "_id" spAppointmentResponse__id Logical id of this artifact }
    spAppointmentResponse__lastUpdated, {@enum.value "_lastUpdated" spAppointmentResponse__lastUpdated When the resource version last changed }
    spAppointmentResponse__profile, {@enum.value "_profile" spAppointmentResponse__profile Profiles this resource claims to conform to }
    spAppointmentResponse__query, {@enum.value "_query" spAppointmentResponse__query A custom search profile that describes a specific defined query operation }
    spAppointmentResponse__security, {@enum.value "_security" spAppointmentResponse__security Security Labels applied to this resource }
    spAppointmentResponse__tag, {@enum.value "_tag" spAppointmentResponse__tag Tags applied to this resource }
    spAppointmentResponse__text, {@enum.value "_text" spAppointmentResponse__text Search on the narrative of the resource }
    spAppointmentResponse_Actor, {@enum.value "actor" spAppointmentResponse_Actor The Person, Location/HealthcareService or Device that this appointment response replies for }
    spAppointmentResponse_Appointment, {@enum.value "appointment" spAppointmentResponse_Appointment The appointment that the response is attached to }
    spAppointmentResponse_Identifier, {@enum.value "identifier" spAppointmentResponse_Identifier An Identifier in this appointment response }
    spAppointmentResponse_Location, {@enum.value "location" spAppointmentResponse_Location This Response is for this Location }
    spAppointmentResponse_Part_status, {@enum.value "part-status" spAppointmentResponse_Part_status The participants acceptance status for this appointment }
    spAppointmentResponse_Patient, {@enum.value "patient" spAppointmentResponse_Patient This Response is for this Patient }
    spAppointmentResponse_Practitioner); {@enum.value "practitioner" spAppointmentResponse_Practitioner This Response is for this Practitioner }

  {@Enum TSearchParamsAuditEvent
    Search Parameters for AuditEvent
  }
  TSearchParamsAuditEvent = (
    spAuditEvent__content, {@enum.value "_content" spAuditEvent__content Search on the entire content of the resource }
    spAuditEvent__id, {@enum.value "_id" spAuditEvent__id Logical id of this artifact }
    spAuditEvent__lastUpdated, {@enum.value "_lastUpdated" spAuditEvent__lastUpdated When the resource version last changed }
    spAuditEvent__profile, {@enum.value "_profile" spAuditEvent__profile Profiles this resource claims to conform to }
    spAuditEvent__query, {@enum.value "_query" spAuditEvent__query A custom search profile that describes a specific defined query operation }
    spAuditEvent__security, {@enum.value "_security" spAuditEvent__security Security Labels applied to this resource }
    spAuditEvent__tag, {@enum.value "_tag" spAuditEvent__tag Tags applied to this resource }
    spAuditEvent__text, {@enum.value "_text" spAuditEvent__text Search on the narrative of the resource }
    spAuditEvent_Action, {@enum.value "action" spAuditEvent_Action Type of action performed during the event }
    spAuditEvent_Address, {@enum.value "address" spAuditEvent_Address Identifier for the network access point of the user device }
    spAuditEvent_Altid, {@enum.value "altid" spAuditEvent_Altid Alternative User id e.g. authentication }
    spAuditEvent_Date, {@enum.value "date" spAuditEvent_Date Time when the event occurred on source }
    spAuditEvent_Desc, {@enum.value "desc" spAuditEvent_Desc Instance-specific descriptor for Object }
    spAuditEvent_Identity, {@enum.value "identity" spAuditEvent_Identity Specific instance of object (e.g. versioned) }
    spAuditEvent_Name, {@enum.value "name" spAuditEvent_Name Human-meaningful name for the user }
    spAuditEvent_Object_type, {@enum.value "object-type" spAuditEvent_Object_type Type of object involved }
    spAuditEvent_Participant, {@enum.value "participant" spAuditEvent_Participant Direct reference to resource }
    spAuditEvent_Patient, {@enum.value "patient" spAuditEvent_Patient Direct reference to resource }
    spAuditEvent_Policy, {@enum.value "policy" spAuditEvent_Policy Policy that authorized event }
    spAuditEvent_Reference, {@enum.value "reference" spAuditEvent_Reference Specific instance of resource (e.g. versioned) }
    spAuditEvent_Site, {@enum.value "site" spAuditEvent_Site Logical source location within the enterprise }
    spAuditEvent_Source, {@enum.value "source" spAuditEvent_Source The identity of source detecting the event }
    spAuditEvent_Subtype, {@enum.value "subtype" spAuditEvent_Subtype More specific type/id for the event }
    spAuditEvent_Type, {@enum.value "type" spAuditEvent_Type Type/identifier of event }
    spAuditEvent_User); {@enum.value "user" spAuditEvent_User Unique identifier for the user }

  {@Enum TSearchParamsBasic
    Search Parameters for Basic
  }
  TSearchParamsBasic = (
    spBasic__content, {@enum.value "_content" spBasic__content Search on the entire content of the resource }
    spBasic__id, {@enum.value "_id" spBasic__id Logical id of this artifact }
    spBasic__lastUpdated, {@enum.value "_lastUpdated" spBasic__lastUpdated When the resource version last changed }
    spBasic__profile, {@enum.value "_profile" spBasic__profile Profiles this resource claims to conform to }
    spBasic__query, {@enum.value "_query" spBasic__query A custom search profile that describes a specific defined query operation }
    spBasic__security, {@enum.value "_security" spBasic__security Security Labels applied to this resource }
    spBasic__tag, {@enum.value "_tag" spBasic__tag Tags applied to this resource }
    spBasic__text, {@enum.value "_text" spBasic__text Search on the narrative of the resource }
    spBasic_Author, {@enum.value "author" spBasic_Author Who created }
    spBasic_Code, {@enum.value "code" spBasic_Code Kind of Resource }
    spBasic_Created, {@enum.value "created" spBasic_Created When created }
    spBasic_Identifier, {@enum.value "identifier" spBasic_Identifier Business identifier }
    spBasic_Patient, {@enum.value "patient" spBasic_Patient Identifies the focus of this resource }
    spBasic_Subject); {@enum.value "subject" spBasic_Subject Identifies the focus of this resource }

  {@Enum TSearchParamsBinary
    Search Parameters for Binary
  }
  TSearchParamsBinary = (
    spBinary__content, {@enum.value "_content" spBinary__content Search on the entire content of the resource }
    spBinary__id, {@enum.value "_id" spBinary__id Logical id of this artifact }
    spBinary__lastUpdated, {@enum.value "_lastUpdated" spBinary__lastUpdated When the resource version last changed }
    spBinary__profile, {@enum.value "_profile" spBinary__profile Profiles this resource claims to conform to }
    spBinary__query, {@enum.value "_query" spBinary__query A custom search profile that describes a specific defined query operation }
    spBinary__security, {@enum.value "_security" spBinary__security Security Labels applied to this resource }
    spBinary__tag, {@enum.value "_tag" spBinary__tag Tags applied to this resource }
    spBinary__text, {@enum.value "_text" spBinary__text Search on the narrative of the resource }
    spBinary_Contenttype); {@enum.value "contenttype" spBinary_Contenttype MimeType of the binary content }

  {@Enum TSearchParamsBodySite
    Search Parameters for BodySite
  }
  TSearchParamsBodySite = (
    spBodySite__content, {@enum.value "_content" spBodySite__content Search on the entire content of the resource }
    spBodySite__id, {@enum.value "_id" spBodySite__id Logical id of this artifact }
    spBodySite__lastUpdated, {@enum.value "_lastUpdated" spBodySite__lastUpdated When the resource version last changed }
    spBodySite__profile, {@enum.value "_profile" spBodySite__profile Profiles this resource claims to conform to }
    spBodySite__query, {@enum.value "_query" spBodySite__query A custom search profile that describes a specific defined query operation }
    spBodySite__security, {@enum.value "_security" spBodySite__security Security Labels applied to this resource }
    spBodySite__tag, {@enum.value "_tag" spBodySite__tag Tags applied to this resource }
    spBodySite__text, {@enum.value "_text" spBodySite__text Search on the narrative of the resource }
    spBodySite_Code, {@enum.value "code" spBodySite_Code Named anatomical location }
    spBodySite_Identifier, {@enum.value "identifier" spBodySite_Identifier Identifier for this instance of the anatomical location }
    spBodySite_Patient); {@enum.value "patient" spBodySite_Patient Patient to whom bodysite belongs }

  {@Enum TSearchParamsBundle
    Search Parameters for Bundle
  }
  TSearchParamsBundle = (
    spBundle__content, {@enum.value "_content" spBundle__content Search on the entire content of the resource }
    spBundle__id, {@enum.value "_id" spBundle__id Logical id of this artifact }
    spBundle__lastUpdated, {@enum.value "_lastUpdated" spBundle__lastUpdated When the resource version last changed }
    spBundle__profile, {@enum.value "_profile" spBundle__profile Profiles this resource claims to conform to }
    spBundle__query, {@enum.value "_query" spBundle__query A custom search profile that describes a specific defined query operation }
    spBundle__security, {@enum.value "_security" spBundle__security Security Labels applied to this resource }
    spBundle__tag, {@enum.value "_tag" spBundle__tag Tags applied to this resource }
    spBundle__text, {@enum.value "_text" spBundle__text Search on the narrative of the resource }
    spBundle_Composition, {@enum.value "composition" spBundle_Composition The first resource in the bundle, if the bundle type is "document" - this is a composition, and this parameter provides access to searches its contents }
    spBundle_Message, {@enum.value "message" spBundle_Message The first resource in the bundle, if the bundle type is "message" - this is a message header, and this parameter provides access to search its contents }
    spBundle_Type); {@enum.value "type" spBundle_Type document | message | transaction | transaction-response | batch | batch-response | history | searchset | collection }

  {@Enum TSearchParamsCarePlan
    Search Parameters for CarePlan
  }
  TSearchParamsCarePlan = (
    spCarePlan__content, {@enum.value "_content" spCarePlan__content Search on the entire content of the resource }
    spCarePlan__id, {@enum.value "_id" spCarePlan__id Logical id of this artifact }
    spCarePlan__lastUpdated, {@enum.value "_lastUpdated" spCarePlan__lastUpdated When the resource version last changed }
    spCarePlan__profile, {@enum.value "_profile" spCarePlan__profile Profiles this resource claims to conform to }
    spCarePlan__query, {@enum.value "_query" spCarePlan__query A custom search profile that describes a specific defined query operation }
    spCarePlan__security, {@enum.value "_security" spCarePlan__security Security Labels applied to this resource }
    spCarePlan__tag, {@enum.value "_tag" spCarePlan__tag Tags applied to this resource }
    spCarePlan__text, {@enum.value "_text" spCarePlan__text Search on the narrative of the resource }
    spCarePlan_Activitycode, {@enum.value "activitycode" spCarePlan_Activitycode Detail type of activity }
    spCarePlan_Activitydate, {@enum.value "activitydate" spCarePlan_Activitydate Specified date occurs within period specified by CarePlan.activity.timingSchedule }
    spCarePlan_Activityreference, {@enum.value "activityreference" spCarePlan_Activityreference Activity details defined in specific resource }
    spCarePlan_Condition, {@enum.value "condition" spCarePlan_Condition Health issues this plan addresses }
    spCarePlan_Date, {@enum.value "date" spCarePlan_Date Time period plan covers }
    spCarePlan_Goal, {@enum.value "goal" spCarePlan_Goal Desired outcome of plan }
    spCarePlan_Participant, {@enum.value "participant" spCarePlan_Participant Who is involved }
    spCarePlan_Patient, {@enum.value "patient" spCarePlan_Patient Who care plan is for }
    spCarePlan_Performer, {@enum.value "performer" spCarePlan_Performer Matches if the practitioner is listed as a performer in any of the "simple" activities.  (For performers of the detailed activities, chain through the activitydetail search parameter.) }
    spCarePlan_Related, {@enum.value "related" spCarePlan_Related A combination of the type of relationship and the related plan }
    spCarePlan_Relatedcode, {@enum.value "relatedcode" spCarePlan_Relatedcode includes | replaces | fulfills }
    spCarePlan_Relatedplan, {@enum.value "relatedplan" spCarePlan_Relatedplan Plan relationship exists with }
    spCarePlan_Subject); {@enum.value "subject" spCarePlan_Subject Who care plan is for }

  {@Enum TSearchParamsClaim
    Search Parameters for Claim
  }
  TSearchParamsClaim = (
    spClaim__content, {@enum.value "_content" spClaim__content Search on the entire content of the resource }
    spClaim__id, {@enum.value "_id" spClaim__id Logical id of this artifact }
    spClaim__lastUpdated, {@enum.value "_lastUpdated" spClaim__lastUpdated When the resource version last changed }
    spClaim__profile, {@enum.value "_profile" spClaim__profile Profiles this resource claims to conform to }
    spClaim__query, {@enum.value "_query" spClaim__query A custom search profile that describes a specific defined query operation }
    spClaim__security, {@enum.value "_security" spClaim__security Security Labels applied to this resource }
    spClaim__tag, {@enum.value "_tag" spClaim__tag Tags applied to this resource }
    spClaim__text, {@enum.value "_text" spClaim__text Search on the narrative of the resource }
    spClaim_Identifier, {@enum.value "identifier" spClaim_Identifier The primary identifier of the financial resource }
    spClaim_Patient, {@enum.value "patient" spClaim_Patient Patient }
    spClaim_Priority, {@enum.value "priority" spClaim_Priority Processing priority requested }
    spClaim_Provider, {@enum.value "provider" spClaim_Provider Provider responsible for the claim }
    spClaim_Use); {@enum.value "use" spClaim_Use The kind of financial resource }

  {@Enum TSearchParamsClaimResponse
    Search Parameters for ClaimResponse
  }
  TSearchParamsClaimResponse = (
    spClaimResponse__content, {@enum.value "_content" spClaimResponse__content Search on the entire content of the resource }
    spClaimResponse__id, {@enum.value "_id" spClaimResponse__id Logical id of this artifact }
    spClaimResponse__lastUpdated, {@enum.value "_lastUpdated" spClaimResponse__lastUpdated When the resource version last changed }
    spClaimResponse__profile, {@enum.value "_profile" spClaimResponse__profile Profiles this resource claims to conform to }
    spClaimResponse__query, {@enum.value "_query" spClaimResponse__query A custom search profile that describes a specific defined query operation }
    spClaimResponse__security, {@enum.value "_security" spClaimResponse__security Security Labels applied to this resource }
    spClaimResponse__tag, {@enum.value "_tag" spClaimResponse__tag Tags applied to this resource }
    spClaimResponse__text, {@enum.value "_text" spClaimResponse__text Search on the narrative of the resource }
    spClaimResponse_Identifier); {@enum.value "identifier" spClaimResponse_Identifier The identity of the insurer }

  {@Enum TSearchParamsClinicalImpression
    Search Parameters for ClinicalImpression
  }
  TSearchParamsClinicalImpression = (
    spClinicalImpression__content, {@enum.value "_content" spClinicalImpression__content Search on the entire content of the resource }
    spClinicalImpression__id, {@enum.value "_id" spClinicalImpression__id Logical id of this artifact }
    spClinicalImpression__lastUpdated, {@enum.value "_lastUpdated" spClinicalImpression__lastUpdated When the resource version last changed }
    spClinicalImpression__profile, {@enum.value "_profile" spClinicalImpression__profile Profiles this resource claims to conform to }
    spClinicalImpression__query, {@enum.value "_query" spClinicalImpression__query A custom search profile that describes a specific defined query operation }
    spClinicalImpression__security, {@enum.value "_security" spClinicalImpression__security Security Labels applied to this resource }
    spClinicalImpression__tag, {@enum.value "_tag" spClinicalImpression__tag Tags applied to this resource }
    spClinicalImpression__text, {@enum.value "_text" spClinicalImpression__text Search on the narrative of the resource }
    spClinicalImpression_Action, {@enum.value "action" spClinicalImpression_Action Actions taken during assessment }
    spClinicalImpression_Assessor, {@enum.value "assessor" spClinicalImpression_Assessor The clinician performing the assessment }
    spClinicalImpression_Date, {@enum.value "date" spClinicalImpression_Date When the assessment occurred }
    spClinicalImpression_Finding, {@enum.value "finding" spClinicalImpression_Finding Specific text or code for finding }
    spClinicalImpression_Investigation, {@enum.value "investigation" spClinicalImpression_Investigation Record of a specific investigation }
    spClinicalImpression_Patient, {@enum.value "patient" spClinicalImpression_Patient The patient being assessed }
    spClinicalImpression_Plan, {@enum.value "plan" spClinicalImpression_Plan Plan of action after assessment }
    spClinicalImpression_Previous, {@enum.value "previous" spClinicalImpression_Previous Reference to last assessment }
    spClinicalImpression_Problem, {@enum.value "problem" spClinicalImpression_Problem General assessment of patient state }
    spClinicalImpression_Resolved, {@enum.value "resolved" spClinicalImpression_Resolved Diagnoses/conditions resolved since previous assessment }
    spClinicalImpression_Ruledout, {@enum.value "ruledout" spClinicalImpression_Ruledout Specific text of code for diagnosis }
    spClinicalImpression_Status, {@enum.value "status" spClinicalImpression_Status in-progress | completed | entered-in-error }
    spClinicalImpression_Trigger, {@enum.value "trigger" spClinicalImpression_Trigger Request or event that necessitated this assessment }
    spClinicalImpression_Trigger_code); {@enum.value "trigger-code" spClinicalImpression_Trigger_code Request or event that necessitated this assessment }

  {@Enum TSearchParamsCommunication
    Search Parameters for Communication
  }
  TSearchParamsCommunication = (
    spCommunication__content, {@enum.value "_content" spCommunication__content Search on the entire content of the resource }
    spCommunication__id, {@enum.value "_id" spCommunication__id Logical id of this artifact }
    spCommunication__lastUpdated, {@enum.value "_lastUpdated" spCommunication__lastUpdated When the resource version last changed }
    spCommunication__profile, {@enum.value "_profile" spCommunication__profile Profiles this resource claims to conform to }
    spCommunication__query, {@enum.value "_query" spCommunication__query A custom search profile that describes a specific defined query operation }
    spCommunication__security, {@enum.value "_security" spCommunication__security Security Labels applied to this resource }
    spCommunication__tag, {@enum.value "_tag" spCommunication__tag Tags applied to this resource }
    spCommunication__text, {@enum.value "_text" spCommunication__text Search on the narrative of the resource }
    spCommunication_Category, {@enum.value "category" spCommunication_Category Message category }
    spCommunication_Encounter, {@enum.value "encounter" spCommunication_Encounter Encounter leading to message }
    spCommunication_Identifier, {@enum.value "identifier" spCommunication_Identifier Unique identifier }
    spCommunication_Medium, {@enum.value "medium" spCommunication_Medium A channel of communication }
    spCommunication_Patient, {@enum.value "patient" spCommunication_Patient Focus of message }
    spCommunication_Received, {@enum.value "received" spCommunication_Received When received }
    spCommunication_Recipient, {@enum.value "recipient" spCommunication_Recipient Message recipient }
    spCommunication_Request, {@enum.value "request" spCommunication_Request CommunicationRequest producing this message }
    spCommunication_Sender, {@enum.value "sender" spCommunication_Sender Message sender }
    spCommunication_Sent, {@enum.value "sent" spCommunication_Sent When sent }
    spCommunication_Status, {@enum.value "status" spCommunication_Status in-progress | completed | suspended | rejected | failed }
    spCommunication_Subject); {@enum.value "subject" spCommunication_Subject Focus of message }

  {@Enum TSearchParamsCommunicationRequest
    Search Parameters for CommunicationRequest
  }
  TSearchParamsCommunicationRequest = (
    spCommunicationRequest__content, {@enum.value "_content" spCommunicationRequest__content Search on the entire content of the resource }
    spCommunicationRequest__id, {@enum.value "_id" spCommunicationRequest__id Logical id of this artifact }
    spCommunicationRequest__lastUpdated, {@enum.value "_lastUpdated" spCommunicationRequest__lastUpdated When the resource version last changed }
    spCommunicationRequest__profile, {@enum.value "_profile" spCommunicationRequest__profile Profiles this resource claims to conform to }
    spCommunicationRequest__query, {@enum.value "_query" spCommunicationRequest__query A custom search profile that describes a specific defined query operation }
    spCommunicationRequest__security, {@enum.value "_security" spCommunicationRequest__security Security Labels applied to this resource }
    spCommunicationRequest__tag, {@enum.value "_tag" spCommunicationRequest__tag Tags applied to this resource }
    spCommunicationRequest__text, {@enum.value "_text" spCommunicationRequest__text Search on the narrative of the resource }
    spCommunicationRequest_Category, {@enum.value "category" spCommunicationRequest_Category Message category }
    spCommunicationRequest_Encounter, {@enum.value "encounter" spCommunicationRequest_Encounter Encounter leading to message }
    spCommunicationRequest_Identifier, {@enum.value "identifier" spCommunicationRequest_Identifier Unique identifier }
    spCommunicationRequest_Medium, {@enum.value "medium" spCommunicationRequest_Medium A channel of communication }
    spCommunicationRequest_Patient, {@enum.value "patient" spCommunicationRequest_Patient Focus of message }
    spCommunicationRequest_Priority, {@enum.value "priority" spCommunicationRequest_Priority Message urgency }
    spCommunicationRequest_Recipient, {@enum.value "recipient" spCommunicationRequest_Recipient Message recipient }
    spCommunicationRequest_Requested, {@enum.value "requested" spCommunicationRequest_Requested When ordered or proposed }
    spCommunicationRequest_Requester, {@enum.value "requester" spCommunicationRequest_Requester An individual who requested a communication }
    spCommunicationRequest_Sender, {@enum.value "sender" spCommunicationRequest_Sender Message sender }
    spCommunicationRequest_Status, {@enum.value "status" spCommunicationRequest_Status proposed | planned | requested | received | accepted | in-progress | completed | suspended | rejected | failed }
    spCommunicationRequest_Subject, {@enum.value "subject" spCommunicationRequest_Subject Focus of message }
    spCommunicationRequest_Time); {@enum.value "time" spCommunicationRequest_Time When scheduled }

  {@Enum TSearchParamsComposition
    Search Parameters for Composition
  }
  TSearchParamsComposition = (
    spComposition__content, {@enum.value "_content" spComposition__content Search on the entire content of the resource }
    spComposition__id, {@enum.value "_id" spComposition__id Logical id of this artifact }
    spComposition__lastUpdated, {@enum.value "_lastUpdated" spComposition__lastUpdated When the resource version last changed }
    spComposition__profile, {@enum.value "_profile" spComposition__profile Profiles this resource claims to conform to }
    spComposition__query, {@enum.value "_query" spComposition__query A custom search profile that describes a specific defined query operation }
    spComposition__security, {@enum.value "_security" spComposition__security Security Labels applied to this resource }
    spComposition__tag, {@enum.value "_tag" spComposition__tag Tags applied to this resource }
    spComposition__text, {@enum.value "_text" spComposition__text Search on the narrative of the resource }
    spComposition_Attester, {@enum.value "attester" spComposition_Attester Who attested the composition }
    spComposition_Author, {@enum.value "author" spComposition_Author Who and/or what authored the composition }
    spComposition_Class, {@enum.value "class" spComposition_Class Categorization of Composition }
    spComposition_Confidentiality, {@enum.value "confidentiality" spComposition_Confidentiality As defined by affinity domain }
    spComposition_Context, {@enum.value "context" spComposition_Context Code(s) that apply to the event being documented }
    spComposition_Date, {@enum.value "date" spComposition_Date Composition editing time }
    spComposition_Encounter, {@enum.value "encounter" spComposition_Encounter Context of the Composition }
    spComposition_Entry, {@enum.value "entry" spComposition_Entry A reference to data that supports this section }
    spComposition_Identifier, {@enum.value "identifier" spComposition_Identifier Logical identifier of composition (version-independent) }
    spComposition_Patient, {@enum.value "patient" spComposition_Patient Who and/or what the composition is about }
    spComposition_Period, {@enum.value "period" spComposition_Period The period covered by the documentation }
    spComposition_Section, {@enum.value "section" spComposition_Section Classification of section (recommended) }
    spComposition_Status, {@enum.value "status" spComposition_Status preliminary | final | amended | entered-in-error }
    spComposition_Subject, {@enum.value "subject" spComposition_Subject Who and/or what the composition is about }
    spComposition_Title, {@enum.value "title" spComposition_Title Human Readable name/title }
    spComposition_Type); {@enum.value "type" spComposition_Type Kind of composition (LOINC if possible) }

  {@Enum TSearchParamsConceptMap
    Search Parameters for ConceptMap
  }
  TSearchParamsConceptMap = (
    spConceptMap__content, {@enum.value "_content" spConceptMap__content Search on the entire content of the resource }
    spConceptMap__id, {@enum.value "_id" spConceptMap__id Logical id of this artifact }
    spConceptMap__lastUpdated, {@enum.value "_lastUpdated" spConceptMap__lastUpdated When the resource version last changed }
    spConceptMap__profile, {@enum.value "_profile" spConceptMap__profile Profiles this resource claims to conform to }
    spConceptMap__query, {@enum.value "_query" spConceptMap__query A custom search profile that describes a specific defined query operation }
    spConceptMap__security, {@enum.value "_security" spConceptMap__security Security Labels applied to this resource }
    spConceptMap__tag, {@enum.value "_tag" spConceptMap__tag Tags applied to this resource }
    spConceptMap__text, {@enum.value "_text" spConceptMap__text Search on the narrative of the resource }
    spConceptMap_Context, {@enum.value "context" spConceptMap_Context A use context assigned to the concept map }
    spConceptMap_Date, {@enum.value "date" spConceptMap_Date The concept map publication date }
    spConceptMap_Dependson, {@enum.value "dependson" spConceptMap_Dependson Reference to element/field/ValueSet mapping depends on }
    spConceptMap_Description, {@enum.value "description" spConceptMap_Description Text search in the description of the concept map }
    spConceptMap_Identifier, {@enum.value "identifier" spConceptMap_Identifier Additional identifier for the concept map }
    spConceptMap_Name, {@enum.value "name" spConceptMap_Name Name of the concept map }
    spConceptMap_Product, {@enum.value "product" spConceptMap_Product Reference to element/field/ValueSet mapping depends on }
    spConceptMap_Publisher, {@enum.value "publisher" spConceptMap_Publisher Name of the publisher of the concept map }
    spConceptMap_Source, {@enum.value "source" spConceptMap_Source Identifies the source of the concepts which are being mapped }
    spConceptMap_Sourcecode, {@enum.value "sourcecode" spConceptMap_Sourcecode Identifies element being mapped }
    spConceptMap_Sourcesystem, {@enum.value "sourcesystem" spConceptMap_Sourcesystem Code System (if value set crosses code systems) }
    spConceptMap_Sourceuri, {@enum.value "sourceuri" spConceptMap_Sourceuri Identifies the source of the concepts which are being mapped }
    spConceptMap_Status, {@enum.value "status" spConceptMap_Status Status of the concept map }
    spConceptMap_Target, {@enum.value "target" spConceptMap_Target Provides context to the mappings }
    spConceptMap_Targetcode, {@enum.value "targetcode" spConceptMap_Targetcode Code that identifies the target element }
    spConceptMap_Targetsystem, {@enum.value "targetsystem" spConceptMap_Targetsystem System of the target (if necessary) }
    spConceptMap_Url, {@enum.value "url" spConceptMap_Url The URL of the concept map }
    spConceptMap_Version); {@enum.value "version" spConceptMap_Version The version identifier of the concept map }

  {@Enum TSearchParamsCondition
    Search Parameters for Condition
  }
  TSearchParamsCondition = (
    spCondition__content, {@enum.value "_content" spCondition__content Search on the entire content of the resource }
    spCondition__id, {@enum.value "_id" spCondition__id Logical id of this artifact }
    spCondition__lastUpdated, {@enum.value "_lastUpdated" spCondition__lastUpdated When the resource version last changed }
    spCondition__profile, {@enum.value "_profile" spCondition__profile Profiles this resource claims to conform to }
    spCondition__query, {@enum.value "_query" spCondition__query A custom search profile that describes a specific defined query operation }
    spCondition__security, {@enum.value "_security" spCondition__security Security Labels applied to this resource }
    spCondition__tag, {@enum.value "_tag" spCondition__tag Tags applied to this resource }
    spCondition__text, {@enum.value "_text" spCondition__text Search on the narrative of the resource }
    spCondition_Asserter, {@enum.value "asserter" spCondition_Asserter Person who asserts this condition }
    spCondition_Body_site, {@enum.value "body-site" spCondition_Body_site Anatomical location, if relevant }
    spCondition_Category, {@enum.value "category" spCondition_Category The category of the condition }
    spCondition_Clinicalstatus, {@enum.value "clinicalstatus" spCondition_Clinicalstatus The clinical status of the condition }
    spCondition_Code, {@enum.value "code" spCondition_Code Code for the condition }
    spCondition_Date_recorded, {@enum.value "date-recorded" spCondition_Date_recorded A date, when the Condition statement was documented }
    spCondition_Encounter, {@enum.value "encounter" spCondition_Encounter Encounter when condition first asserted }
    spCondition_Evidence, {@enum.value "evidence" spCondition_Evidence Manifestation/symptom }
    spCondition_Identifier, {@enum.value "identifier" spCondition_Identifier A unique identifier of the condition record }
    spCondition_Onset, {@enum.value "onset" spCondition_Onset Date related onsets (dateTime and Period) }
    spCondition_Onset_info, {@enum.value "onset-info" spCondition_Onset_info Other onsets (boolean, age, range, string) }
    spCondition_Patient, {@enum.value "patient" spCondition_Patient Who has the condition? }
    spCondition_Severity, {@enum.value "severity" spCondition_Severity The severity of the condition }
    spCondition_Stage); {@enum.value "stage" spCondition_Stage Simple summary (disease specific) }

  {@Enum TSearchParamsConformance
    Search Parameters for Conformance
  }
  TSearchParamsConformance = (
    spConformance__content, {@enum.value "_content" spConformance__content Search on the entire content of the resource }
    spConformance__id, {@enum.value "_id" spConformance__id Logical id of this artifact }
    spConformance__lastUpdated, {@enum.value "_lastUpdated" spConformance__lastUpdated When the resource version last changed }
    spConformance__profile, {@enum.value "_profile" spConformance__profile Profiles this resource claims to conform to }
    spConformance__query, {@enum.value "_query" spConformance__query A custom search profile that describes a specific defined query operation }
    spConformance__security, {@enum.value "_security" spConformance__security Security Labels applied to this resource }
    spConformance__tag, {@enum.value "_tag" spConformance__tag Tags applied to this resource }
    spConformance__text, {@enum.value "_text" spConformance__text Search on the narrative of the resource }
    spConformance_Date, {@enum.value "date" spConformance_Date The conformance statement publication date }
    spConformance_Description, {@enum.value "description" spConformance_Description Text search in the description of the conformance statement }
    spConformance_Event, {@enum.value "event" spConformance_Event Event code in a conformance statement }
    spConformance_Fhirversion, {@enum.value "fhirversion" spConformance_Fhirversion The version of FHIR }
    spConformance_Format, {@enum.value "format" spConformance_Format formats supported (xml | json | mime type) }
    spConformance_Mode, {@enum.value "mode" spConformance_Mode Mode - restful (server/client) or messaging (sender/receiver) }
    spConformance_Name, {@enum.value "name" spConformance_Name Name of the conformance statement }
    spConformance_Profile, {@enum.value "profile" spConformance_Profile A profile id invoked in a conformance statement }
    spConformance_Publisher, {@enum.value "publisher" spConformance_Publisher Name of the publisher of the conformance statement }
    spConformance_Resource, {@enum.value "resource" spConformance_Resource Name of a resource mentioned in a conformance statement }
    spConformance_Security, {@enum.value "security" spConformance_Security OAuth | SMART-on-FHIR | NTLM | Basic | Kerberos | Certificates }
    spConformance_Software, {@enum.value "software" spConformance_Software Part of a the name of a software application }
    spConformance_Status, {@enum.value "status" spConformance_Status The current status of the conformance statement }
    spConformance_Supported_profile, {@enum.value "supported-profile" spConformance_Supported_profile Profiles for use cases supported }
    spConformance_Url, {@enum.value "url" spConformance_Url The uri that identifies the conformance statement }
    spConformance_Version); {@enum.value "version" spConformance_Version The version identifier of the conformance statement }

  {@Enum TSearchParamsContract
    Search Parameters for Contract
  }
  TSearchParamsContract = (
    spContract__content, {@enum.value "_content" spContract__content Search on the entire content of the resource }
    spContract__id, {@enum.value "_id" spContract__id Logical id of this artifact }
    spContract__lastUpdated, {@enum.value "_lastUpdated" spContract__lastUpdated When the resource version last changed }
    spContract__profile, {@enum.value "_profile" spContract__profile Profiles this resource claims to conform to }
    spContract__query, {@enum.value "_query" spContract__query A custom search profile that describes a specific defined query operation }
    spContract__security, {@enum.value "_security" spContract__security Security Labels applied to this resource }
    spContract__tag, {@enum.value "_tag" spContract__tag Tags applied to this resource }
    spContract__text, {@enum.value "_text" spContract__text Search on the narrative of the resource }
    spContract_Actor, {@enum.value "actor" spContract_Actor Contract Actor Type }
    spContract_Identifier, {@enum.value "identifier" spContract_Identifier The identity of the contract }
    spContract_Patient, {@enum.value "patient" spContract_Patient The identity of the target of the contract (if a patient) }
    spContract_Signer, {@enum.value "signer" spContract_Signer Contract Signatory Party }
    spContract_Subject); {@enum.value "subject" spContract_Subject The identity of the target of the contract }

  {@Enum TSearchParamsCoverage
    Search Parameters for Coverage
  }
  TSearchParamsCoverage = (
    spCoverage__content, {@enum.value "_content" spCoverage__content Search on the entire content of the resource }
    spCoverage__id, {@enum.value "_id" spCoverage__id Logical id of this artifact }
    spCoverage__lastUpdated, {@enum.value "_lastUpdated" spCoverage__lastUpdated When the resource version last changed }
    spCoverage__profile, {@enum.value "_profile" spCoverage__profile Profiles this resource claims to conform to }
    spCoverage__query, {@enum.value "_query" spCoverage__query A custom search profile that describes a specific defined query operation }
    spCoverage__security, {@enum.value "_security" spCoverage__security Security Labels applied to this resource }
    spCoverage__tag, {@enum.value "_tag" spCoverage__tag Tags applied to this resource }
    spCoverage__text, {@enum.value "_text" spCoverage__text Search on the narrative of the resource }
    spCoverage_Dependent, {@enum.value "dependent" spCoverage_Dependent Dependent number }
    spCoverage_Group, {@enum.value "group" spCoverage_Group Group identifier }
    spCoverage_Identifier, {@enum.value "identifier" spCoverage_Identifier The primary identifier of the insured }
    spCoverage_Issuer, {@enum.value "issuer" spCoverage_Issuer The identity of the insurer }
    spCoverage_Plan, {@enum.value "plan" spCoverage_Plan A plan or policy identifier }
    spCoverage_Sequence, {@enum.value "sequence" spCoverage_Sequence Sequence number }
    spCoverage_Subplan, {@enum.value "subplan" spCoverage_Subplan Sub-plan identifier }
    spCoverage_Type); {@enum.value "type" spCoverage_Type The kind of coverage }

  {@Enum TSearchParamsDataElement
    Search Parameters for DataElement
  }
  TSearchParamsDataElement = (
    spDataElement__content, {@enum.value "_content" spDataElement__content Search on the entire content of the resource }
    spDataElement__id, {@enum.value "_id" spDataElement__id Logical id of this artifact }
    spDataElement__lastUpdated, {@enum.value "_lastUpdated" spDataElement__lastUpdated When the resource version last changed }
    spDataElement__profile, {@enum.value "_profile" spDataElement__profile Profiles this resource claims to conform to }
    spDataElement__query, {@enum.value "_query" spDataElement__query A custom search profile that describes a specific defined query operation }
    spDataElement__security, {@enum.value "_security" spDataElement__security Security Labels applied to this resource }
    spDataElement__tag, {@enum.value "_tag" spDataElement__tag Tags applied to this resource }
    spDataElement__text, {@enum.value "_text" spDataElement__text Search on the narrative of the resource }
    spDataElement_Code, {@enum.value "code" spDataElement_Code A code for the data element (server may choose to do subsumption) }
    spDataElement_Context, {@enum.value "context" spDataElement_Context A use context assigned to the data element }
    spDataElement_Date, {@enum.value "date" spDataElement_Date The data element publication date }
    spDataElement_Description, {@enum.value "description" spDataElement_Description Text search in the description of the data element.  This corresponds to the definition of the first DataElement.element. }
    spDataElement_Identifier, {@enum.value "identifier" spDataElement_Identifier The identifier of the data element }
    spDataElement_Name, {@enum.value "name" spDataElement_Name Name of the data element }
    spDataElement_Publisher, {@enum.value "publisher" spDataElement_Publisher Name of the publisher of the data element }
    spDataElement_Status, {@enum.value "status" spDataElement_Status The current status of the data element }
    spDataElement_Stringency, {@enum.value "stringency" spDataElement_Stringency The stringency of the data element definition }
    spDataElement_Url, {@enum.value "url" spDataElement_Url The official URL for the data element }
    spDataElement_Version); {@enum.value "version" spDataElement_Version The version identifier of the data element }

  {@Enum TSearchParamsDetectedIssue
    Search Parameters for DetectedIssue
  }
  TSearchParamsDetectedIssue = (
    spDetectedIssue__content, {@enum.value "_content" spDetectedIssue__content Search on the entire content of the resource }
    spDetectedIssue__id, {@enum.value "_id" spDetectedIssue__id Logical id of this artifact }
    spDetectedIssue__lastUpdated, {@enum.value "_lastUpdated" spDetectedIssue__lastUpdated When the resource version last changed }
    spDetectedIssue__profile, {@enum.value "_profile" spDetectedIssue__profile Profiles this resource claims to conform to }
    spDetectedIssue__query, {@enum.value "_query" spDetectedIssue__query A custom search profile that describes a specific defined query operation }
    spDetectedIssue__security, {@enum.value "_security" spDetectedIssue__security Security Labels applied to this resource }
    spDetectedIssue__tag, {@enum.value "_tag" spDetectedIssue__tag Tags applied to this resource }
    spDetectedIssue__text, {@enum.value "_text" spDetectedIssue__text Search on the narrative of the resource }
    spDetectedIssue_Author, {@enum.value "author" spDetectedIssue_Author The provider or device that identified the issue }
    spDetectedIssue_Category, {@enum.value "category" spDetectedIssue_Category Issue Category, e.g. drug-drug, duplicate therapy, etc. }
    spDetectedIssue_Date, {@enum.value "date" spDetectedIssue_Date When identified }
    spDetectedIssue_Identifier, {@enum.value "identifier" spDetectedIssue_Identifier Unique id for the detected issue }
    spDetectedIssue_Implicated, {@enum.value "implicated" spDetectedIssue_Implicated Problem resource }
    spDetectedIssue_Patient); {@enum.value "patient" spDetectedIssue_Patient Associated patient }

  {@Enum TSearchParamsDevice
    Search Parameters for Device
  }
  TSearchParamsDevice = (
    spDevice__content, {@enum.value "_content" spDevice__content Search on the entire content of the resource }
    spDevice__id, {@enum.value "_id" spDevice__id Logical id of this artifact }
    spDevice__lastUpdated, {@enum.value "_lastUpdated" spDevice__lastUpdated When the resource version last changed }
    spDevice__profile, {@enum.value "_profile" spDevice__profile Profiles this resource claims to conform to }
    spDevice__query, {@enum.value "_query" spDevice__query A custom search profile that describes a specific defined query operation }
    spDevice__security, {@enum.value "_security" spDevice__security Security Labels applied to this resource }
    spDevice__tag, {@enum.value "_tag" spDevice__tag Tags applied to this resource }
    spDevice__text, {@enum.value "_text" spDevice__text Search on the narrative of the resource }
    spDevice_Identifier, {@enum.value "identifier" spDevice_Identifier Instance id from manufacturer, owner, and others }
    spDevice_Location, {@enum.value "location" spDevice_Location A location, where the resource is found }
    spDevice_Manufacturer, {@enum.value "manufacturer" spDevice_Manufacturer The manufacturer of the device }
    spDevice_Model, {@enum.value "model" spDevice_Model The model of the device }
    spDevice_Organization, {@enum.value "organization" spDevice_Organization The organization responsible for the device }
    spDevice_Patient, {@enum.value "patient" spDevice_Patient Patient information, if the resource is affixed to a person }
    spDevice_Type, {@enum.value "type" spDevice_Type The type of the device }
    spDevice_Udi, {@enum.value "udi" spDevice_Udi FDA mandated Unique Device Identifier }
    spDevice_Url); {@enum.value "url" spDevice_Url Network address to contact device }

  {@Enum TSearchParamsDeviceComponent
    Search Parameters for DeviceComponent
  }
  TSearchParamsDeviceComponent = (
    spDeviceComponent__content, {@enum.value "_content" spDeviceComponent__content Search on the entire content of the resource }
    spDeviceComponent__id, {@enum.value "_id" spDeviceComponent__id Logical id of this artifact }
    spDeviceComponent__lastUpdated, {@enum.value "_lastUpdated" spDeviceComponent__lastUpdated When the resource version last changed }
    spDeviceComponent__profile, {@enum.value "_profile" spDeviceComponent__profile Profiles this resource claims to conform to }
    spDeviceComponent__query, {@enum.value "_query" spDeviceComponent__query A custom search profile that describes a specific defined query operation }
    spDeviceComponent__security, {@enum.value "_security" spDeviceComponent__security Security Labels applied to this resource }
    spDeviceComponent__tag, {@enum.value "_tag" spDeviceComponent__tag Tags applied to this resource }
    spDeviceComponent__text, {@enum.value "_text" spDeviceComponent__text Search on the narrative of the resource }
    spDeviceComponent_Parent, {@enum.value "parent" spDeviceComponent_Parent The parent DeviceComponent resource }
    spDeviceComponent_Source, {@enum.value "source" spDeviceComponent_Source The device source }
    spDeviceComponent_Type); {@enum.value "type" spDeviceComponent_Type The device component type }

  {@Enum TSearchParamsDeviceMetric
    Search Parameters for DeviceMetric
  }
  TSearchParamsDeviceMetric = (
    spDeviceMetric__content, {@enum.value "_content" spDeviceMetric__content Search on the entire content of the resource }
    spDeviceMetric__id, {@enum.value "_id" spDeviceMetric__id Logical id of this artifact }
    spDeviceMetric__lastUpdated, {@enum.value "_lastUpdated" spDeviceMetric__lastUpdated When the resource version last changed }
    spDeviceMetric__profile, {@enum.value "_profile" spDeviceMetric__profile Profiles this resource claims to conform to }
    spDeviceMetric__query, {@enum.value "_query" spDeviceMetric__query A custom search profile that describes a specific defined query operation }
    spDeviceMetric__security, {@enum.value "_security" spDeviceMetric__security Security Labels applied to this resource }
    spDeviceMetric__tag, {@enum.value "_tag" spDeviceMetric__tag Tags applied to this resource }
    spDeviceMetric__text, {@enum.value "_text" spDeviceMetric__text Search on the narrative of the resource }
    spDeviceMetric_Category, {@enum.value "category" spDeviceMetric_Category The category of the metric }
    spDeviceMetric_Identifier, {@enum.value "identifier" spDeviceMetric_Identifier The identifier of the metric }
    spDeviceMetric_Parent, {@enum.value "parent" spDeviceMetric_Parent The parent DeviceMetric resource }
    spDeviceMetric_Source, {@enum.value "source" spDeviceMetric_Source The device resource }
    spDeviceMetric_Type); {@enum.value "type" spDeviceMetric_Type The component type }

  {@Enum TSearchParamsDeviceUseRequest
    Search Parameters for DeviceUseRequest
  }
  TSearchParamsDeviceUseRequest = (
    spDeviceUseRequest__content, {@enum.value "_content" spDeviceUseRequest__content Search on the entire content of the resource }
    spDeviceUseRequest__id, {@enum.value "_id" spDeviceUseRequest__id Logical id of this artifact }
    spDeviceUseRequest__lastUpdated, {@enum.value "_lastUpdated" spDeviceUseRequest__lastUpdated When the resource version last changed }
    spDeviceUseRequest__profile, {@enum.value "_profile" spDeviceUseRequest__profile Profiles this resource claims to conform to }
    spDeviceUseRequest__query, {@enum.value "_query" spDeviceUseRequest__query A custom search profile that describes a specific defined query operation }
    spDeviceUseRequest__security, {@enum.value "_security" spDeviceUseRequest__security Security Labels applied to this resource }
    spDeviceUseRequest__tag, {@enum.value "_tag" spDeviceUseRequest__tag Tags applied to this resource }
    spDeviceUseRequest__text, {@enum.value "_text" spDeviceUseRequest__text Search on the narrative of the resource }
    spDeviceUseRequest_Device, {@enum.value "device" spDeviceUseRequest_Device Device requested }
    spDeviceUseRequest_Patient, {@enum.value "patient" spDeviceUseRequest_Patient Search by subject - a patient }
    spDeviceUseRequest_Subject); {@enum.value "subject" spDeviceUseRequest_Subject Search by subject }

  {@Enum TSearchParamsDeviceUseStatement
    Search Parameters for DeviceUseStatement
  }
  TSearchParamsDeviceUseStatement = (
    spDeviceUseStatement__content, {@enum.value "_content" spDeviceUseStatement__content Search on the entire content of the resource }
    spDeviceUseStatement__id, {@enum.value "_id" spDeviceUseStatement__id Logical id of this artifact }
    spDeviceUseStatement__lastUpdated, {@enum.value "_lastUpdated" spDeviceUseStatement__lastUpdated When the resource version last changed }
    spDeviceUseStatement__profile, {@enum.value "_profile" spDeviceUseStatement__profile Profiles this resource claims to conform to }
    spDeviceUseStatement__query, {@enum.value "_query" spDeviceUseStatement__query A custom search profile that describes a specific defined query operation }
    spDeviceUseStatement__security, {@enum.value "_security" spDeviceUseStatement__security Security Labels applied to this resource }
    spDeviceUseStatement__tag, {@enum.value "_tag" spDeviceUseStatement__tag Tags applied to this resource }
    spDeviceUseStatement__text, {@enum.value "_text" spDeviceUseStatement__text Search on the narrative of the resource }
    spDeviceUseStatement_Device, {@enum.value "device" spDeviceUseStatement_Device Search by device }
    spDeviceUseStatement_Patient, {@enum.value "patient" spDeviceUseStatement_Patient Search by subject - a patient }
    spDeviceUseStatement_Subject); {@enum.value "subject" spDeviceUseStatement_Subject Search by subject }

  {@Enum TSearchParamsDiagnosticOrder
    Search Parameters for DiagnosticOrder
  }
  TSearchParamsDiagnosticOrder = (
    spDiagnosticOrder__content, {@enum.value "_content" spDiagnosticOrder__content Search on the entire content of the resource }
    spDiagnosticOrder__id, {@enum.value "_id" spDiagnosticOrder__id Logical id of this artifact }
    spDiagnosticOrder__lastUpdated, {@enum.value "_lastUpdated" spDiagnosticOrder__lastUpdated When the resource version last changed }
    spDiagnosticOrder__profile, {@enum.value "_profile" spDiagnosticOrder__profile Profiles this resource claims to conform to }
    spDiagnosticOrder__query, {@enum.value "_query" spDiagnosticOrder__query A custom search profile that describes a specific defined query operation }
    spDiagnosticOrder__security, {@enum.value "_security" spDiagnosticOrder__security Security Labels applied to this resource }
    spDiagnosticOrder__tag, {@enum.value "_tag" spDiagnosticOrder__tag Tags applied to this resource }
    spDiagnosticOrder__text, {@enum.value "_text" spDiagnosticOrder__text Search on the narrative of the resource }
    spDiagnosticOrder_Actor, {@enum.value "actor" spDiagnosticOrder_Actor Who recorded or did this }
    spDiagnosticOrder_Bodysite, {@enum.value "bodysite" spDiagnosticOrder_Bodysite Location of requested test (if applicable) }
    spDiagnosticOrder_Code, {@enum.value "code" spDiagnosticOrder_Code Code to indicate the item (test or panel) being ordered }
    spDiagnosticOrder_Encounter, {@enum.value "encounter" spDiagnosticOrder_Encounter The encounter that this diagnostic order is associated with }
    spDiagnosticOrder_Event_date, {@enum.value "event-date" spDiagnosticOrder_Event_date The date at which the event happened }
    spDiagnosticOrder_Event_status, {@enum.value "event-status" spDiagnosticOrder_Event_status proposed | draft | planned | requested | received | accepted | in-progress | review | completed | cancelled | suspended | rejected | failed }
    spDiagnosticOrder_Event_status_date, {@enum.value "event-status-date" spDiagnosticOrder_Event_status_date A combination of past-status and date }
    spDiagnosticOrder_Identifier, {@enum.value "identifier" spDiagnosticOrder_Identifier Identifiers assigned to this order }
    spDiagnosticOrder_Item_date, {@enum.value "item-date" spDiagnosticOrder_Item_date The date at which the event happened }
    spDiagnosticOrder_Item_past_status, {@enum.value "item-past-status" spDiagnosticOrder_Item_past_status proposed | draft | planned | requested | received | accepted | in-progress | review | completed | cancelled | suspended | rejected | failed }
    spDiagnosticOrder_Item_status, {@enum.value "item-status" spDiagnosticOrder_Item_status proposed | draft | planned | requested | received | accepted | in-progress | review | completed | cancelled | suspended | rejected | failed }
    spDiagnosticOrder_Item_status_date, {@enum.value "item-status-date" spDiagnosticOrder_Item_status_date A combination of item-past-status and item-date }
    spDiagnosticOrder_Orderer, {@enum.value "orderer" spDiagnosticOrder_Orderer Who ordered the test }
    spDiagnosticOrder_Patient, {@enum.value "patient" spDiagnosticOrder_Patient Who and/or what test is about }
    spDiagnosticOrder_Specimen, {@enum.value "specimen" spDiagnosticOrder_Specimen If the whole order relates to specific specimens }
    spDiagnosticOrder_Status, {@enum.value "status" spDiagnosticOrder_Status proposed | draft | planned | requested | received | accepted | in-progress | review | completed | cancelled | suspended | rejected | failed }
    spDiagnosticOrder_Subject); {@enum.value "subject" spDiagnosticOrder_Subject Who and/or what test is about }

  {@Enum TSearchParamsDiagnosticReport
    Search Parameters for DiagnosticReport
  }
  TSearchParamsDiagnosticReport = (
    spDiagnosticReport__content, {@enum.value "_content" spDiagnosticReport__content Search on the entire content of the resource }
    spDiagnosticReport__id, {@enum.value "_id" spDiagnosticReport__id Logical id of this artifact }
    spDiagnosticReport__lastUpdated, {@enum.value "_lastUpdated" spDiagnosticReport__lastUpdated When the resource version last changed }
    spDiagnosticReport__profile, {@enum.value "_profile" spDiagnosticReport__profile Profiles this resource claims to conform to }
    spDiagnosticReport__query, {@enum.value "_query" spDiagnosticReport__query A custom search profile that describes a specific defined query operation }
    spDiagnosticReport__security, {@enum.value "_security" spDiagnosticReport__security Security Labels applied to this resource }
    spDiagnosticReport__tag, {@enum.value "_tag" spDiagnosticReport__tag Tags applied to this resource }
    spDiagnosticReport__text, {@enum.value "_text" spDiagnosticReport__text Search on the narrative of the resource }
    spDiagnosticReport_Category, {@enum.value "category" spDiagnosticReport_Category Which diagnostic discipline/department created the report }
    spDiagnosticReport_Code, {@enum.value "code" spDiagnosticReport_Code The code for the report as a whole, as opposed to codes for the atomic results, which are the names on the observation resource referred to from the result }
    spDiagnosticReport_Date, {@enum.value "date" spDiagnosticReport_Date The clinically relevant time of the report }
    spDiagnosticReport_Diagnosis, {@enum.value "diagnosis" spDiagnosticReport_Diagnosis A coded diagnosis on the report }
    spDiagnosticReport_Encounter, {@enum.value "encounter" spDiagnosticReport_Encounter The Encounter when the order was made }
    spDiagnosticReport_Identifier, {@enum.value "identifier" spDiagnosticReport_Identifier An identifier for the report }
    spDiagnosticReport_Image, {@enum.value "image" spDiagnosticReport_Image A reference to the image source. }
    spDiagnosticReport_Issued, {@enum.value "issued" spDiagnosticReport_Issued When the report was issued }
    spDiagnosticReport_Patient, {@enum.value "patient" spDiagnosticReport_Patient The subject of the report if a patient }
    spDiagnosticReport_Performer, {@enum.value "performer" spDiagnosticReport_Performer Who was the source of the report (organization) }
    spDiagnosticReport_Request, {@enum.value "request" spDiagnosticReport_Request Reference to the test or procedure request. }
    spDiagnosticReport_Result, {@enum.value "result" spDiagnosticReport_Result Link to an atomic result (observation resource) }
    spDiagnosticReport_Specimen, {@enum.value "specimen" spDiagnosticReport_Specimen The specimen details }
    spDiagnosticReport_Status, {@enum.value "status" spDiagnosticReport_Status The status of the report }
    spDiagnosticReport_Subject); {@enum.value "subject" spDiagnosticReport_Subject The subject of the report }

  {@Enum TSearchParamsDocumentManifest
    Search Parameters for DocumentManifest
  }
  TSearchParamsDocumentManifest = (
    spDocumentManifest__content, {@enum.value "_content" spDocumentManifest__content Search on the entire content of the resource }
    spDocumentManifest__id, {@enum.value "_id" spDocumentManifest__id Logical id of this artifact }
    spDocumentManifest__lastUpdated, {@enum.value "_lastUpdated" spDocumentManifest__lastUpdated When the resource version last changed }
    spDocumentManifest__profile, {@enum.value "_profile" spDocumentManifest__profile Profiles this resource claims to conform to }
    spDocumentManifest__query, {@enum.value "_query" spDocumentManifest__query A custom search profile that describes a specific defined query operation }
    spDocumentManifest__security, {@enum.value "_security" spDocumentManifest__security Security Labels applied to this resource }
    spDocumentManifest__tag, {@enum.value "_tag" spDocumentManifest__tag Tags applied to this resource }
    spDocumentManifest__text, {@enum.value "_text" spDocumentManifest__text Search on the narrative of the resource }
    spDocumentManifest_Author, {@enum.value "author" spDocumentManifest_Author Who and/or what authored the manifest }
    spDocumentManifest_Content_ref, {@enum.value "content-ref" spDocumentManifest_Content_ref Contents of this set of documents }
    spDocumentManifest_Created, {@enum.value "created" spDocumentManifest_Created When this document manifest created }
    spDocumentManifest_Description, {@enum.value "description" spDocumentManifest_Description Human-readable description (title) }
    spDocumentManifest_Identifier, {@enum.value "identifier" spDocumentManifest_Identifier Unique Identifier for the set of documents }
    spDocumentManifest_Patient, {@enum.value "patient" spDocumentManifest_Patient The subject of the set of documents }
    spDocumentManifest_Recipient, {@enum.value "recipient" spDocumentManifest_Recipient Intended to get notified about this set of documents }
    spDocumentManifest_Related_id, {@enum.value "related-id" spDocumentManifest_Related_id Identifiers of things that are related }
    spDocumentManifest_Related_ref, {@enum.value "related-ref" spDocumentManifest_Related_ref Related Resource }
    spDocumentManifest_Source, {@enum.value "source" spDocumentManifest_Source The source system/application/software }
    spDocumentManifest_Status, {@enum.value "status" spDocumentManifest_Status current | superseded | entered-in-error }
    spDocumentManifest_Subject, {@enum.value "subject" spDocumentManifest_Subject The subject of the set of documents }
    spDocumentManifest_Type); {@enum.value "type" spDocumentManifest_Type Kind of document set }

  {@Enum TSearchParamsDocumentReference
    Search Parameters for DocumentReference
  }
  TSearchParamsDocumentReference = (
    spDocumentReference__content, {@enum.value "_content" spDocumentReference__content Search on the entire content of the resource }
    spDocumentReference__id, {@enum.value "_id" spDocumentReference__id Logical id of this artifact }
    spDocumentReference__lastUpdated, {@enum.value "_lastUpdated" spDocumentReference__lastUpdated When the resource version last changed }
    spDocumentReference__profile, {@enum.value "_profile" spDocumentReference__profile Profiles this resource claims to conform to }
    spDocumentReference__query, {@enum.value "_query" spDocumentReference__query A custom search profile that describes a specific defined query operation }
    spDocumentReference__security, {@enum.value "_security" spDocumentReference__security Security Labels applied to this resource }
    spDocumentReference__tag, {@enum.value "_tag" spDocumentReference__tag Tags applied to this resource }
    spDocumentReference__text, {@enum.value "_text" spDocumentReference__text Search on the narrative of the resource }
    spDocumentReference_Authenticator, {@enum.value "authenticator" spDocumentReference_Authenticator Who/what authenticated the document }
    spDocumentReference_Author, {@enum.value "author" spDocumentReference_Author Who and/or what authored the document }
    spDocumentReference_Class, {@enum.value "class" spDocumentReference_Class Categorization of document }
    spDocumentReference_Created, {@enum.value "created" spDocumentReference_Created Document creation time }
    spDocumentReference_Custodian, {@enum.value "custodian" spDocumentReference_Custodian Organization which maintains the document }
    spDocumentReference_Description, {@enum.value "description" spDocumentReference_Description Human-readable description (title) }
    spDocumentReference_Encounter, {@enum.value "encounter" spDocumentReference_Encounter Context of the document  content }
    spDocumentReference_Event, {@enum.value "event" spDocumentReference_Event Main Clinical Acts Documented }
    spDocumentReference_Facility, {@enum.value "facility" spDocumentReference_Facility Kind of facility where patient was seen }
    spDocumentReference_Format, {@enum.value "format" spDocumentReference_Format Format/content rules for the document }
    spDocumentReference_Identifier, {@enum.value "identifier" spDocumentReference_Identifier Master Version Specific Identifier }
    spDocumentReference_Indexed, {@enum.value "indexed" spDocumentReference_Indexed When this document reference created }
    spDocumentReference_Language, {@enum.value "language" spDocumentReference_Language Human language of the content (BCP-47) }
    spDocumentReference_Location, {@enum.value "location" spDocumentReference_Location Uri where the data can be found }
    spDocumentReference_Patient, {@enum.value "patient" spDocumentReference_Patient Who/what is the subject of the document }
    spDocumentReference_Period, {@enum.value "period" spDocumentReference_Period Time of service that is being documented }
    spDocumentReference_Related_id, {@enum.value "related-id" spDocumentReference_Related_id Identifier of related objects or events }
    spDocumentReference_Related_ref, {@enum.value "related-ref" spDocumentReference_Related_ref Related Resource }
    spDocumentReference_Relatesto, {@enum.value "relatesto" spDocumentReference_Relatesto Target of the relationship }
    spDocumentReference_Relation, {@enum.value "relation" spDocumentReference_Relation replaces | transforms | signs | appends }
    spDocumentReference_Relationship, {@enum.value "relationship" spDocumentReference_Relationship Combination of relation and relatesTo }
    spDocumentReference_Securitylabel, {@enum.value "securitylabel" spDocumentReference_Securitylabel Document security-tags }
    spDocumentReference_Setting, {@enum.value "setting" spDocumentReference_Setting Additional details about where the content was created (e.g. clinical specialty) }
    spDocumentReference_Status, {@enum.value "status" spDocumentReference_Status current | superseded | entered-in-error }
    spDocumentReference_Subject, {@enum.value "subject" spDocumentReference_Subject Who/what is the subject of the document }
    spDocumentReference_Type); {@enum.value "type" spDocumentReference_Type Kind of document (LOINC if possible) }

  {@Enum TSearchParamsEligibilityRequest
    Search Parameters for EligibilityRequest
  }
  TSearchParamsEligibilityRequest = (
    spEligibilityRequest__content, {@enum.value "_content" spEligibilityRequest__content Search on the entire content of the resource }
    spEligibilityRequest__id, {@enum.value "_id" spEligibilityRequest__id Logical id of this artifact }
    spEligibilityRequest__lastUpdated, {@enum.value "_lastUpdated" spEligibilityRequest__lastUpdated When the resource version last changed }
    spEligibilityRequest__profile, {@enum.value "_profile" spEligibilityRequest__profile Profiles this resource claims to conform to }
    spEligibilityRequest__query, {@enum.value "_query" spEligibilityRequest__query A custom search profile that describes a specific defined query operation }
    spEligibilityRequest__security, {@enum.value "_security" spEligibilityRequest__security Security Labels applied to this resource }
    spEligibilityRequest__tag, {@enum.value "_tag" spEligibilityRequest__tag Tags applied to this resource }
    spEligibilityRequest__text, {@enum.value "_text" spEligibilityRequest__text Search on the narrative of the resource }
    spEligibilityRequest_Identifier); {@enum.value "identifier" spEligibilityRequest_Identifier The business identifier of the Eligibility }

  {@Enum TSearchParamsEligibilityResponse
    Search Parameters for EligibilityResponse
  }
  TSearchParamsEligibilityResponse = (
    spEligibilityResponse__content, {@enum.value "_content" spEligibilityResponse__content Search on the entire content of the resource }
    spEligibilityResponse__id, {@enum.value "_id" spEligibilityResponse__id Logical id of this artifact }
    spEligibilityResponse__lastUpdated, {@enum.value "_lastUpdated" spEligibilityResponse__lastUpdated When the resource version last changed }
    spEligibilityResponse__profile, {@enum.value "_profile" spEligibilityResponse__profile Profiles this resource claims to conform to }
    spEligibilityResponse__query, {@enum.value "_query" spEligibilityResponse__query A custom search profile that describes a specific defined query operation }
    spEligibilityResponse__security, {@enum.value "_security" spEligibilityResponse__security Security Labels applied to this resource }
    spEligibilityResponse__tag, {@enum.value "_tag" spEligibilityResponse__tag Tags applied to this resource }
    spEligibilityResponse__text, {@enum.value "_text" spEligibilityResponse__text Search on the narrative of the resource }
    spEligibilityResponse_Identifier); {@enum.value "identifier" spEligibilityResponse_Identifier The business identifier of the Explanation of Benefit }

  {@Enum TSearchParamsEncounter
    Search Parameters for Encounter
  }
  TSearchParamsEncounter = (
    spEncounter__content, {@enum.value "_content" spEncounter__content Search on the entire content of the resource }
    spEncounter__id, {@enum.value "_id" spEncounter__id Logical id of this artifact }
    spEncounter__lastUpdated, {@enum.value "_lastUpdated" spEncounter__lastUpdated When the resource version last changed }
    spEncounter__profile, {@enum.value "_profile" spEncounter__profile Profiles this resource claims to conform to }
    spEncounter__query, {@enum.value "_query" spEncounter__query A custom search profile that describes a specific defined query operation }
    spEncounter__security, {@enum.value "_security" spEncounter__security Security Labels applied to this resource }
    spEncounter__tag, {@enum.value "_tag" spEncounter__tag Tags applied to this resource }
    spEncounter__text, {@enum.value "_text" spEncounter__text Search on the narrative of the resource }
    spEncounter_Appointment, {@enum.value "appointment" spEncounter_Appointment The appointment that scheduled this encounter }
    spEncounter_Condition, {@enum.value "condition" spEncounter_Condition Reason the encounter takes place (resource) }
    spEncounter_Date, {@enum.value "date" spEncounter_Date A date within the period the Encounter lasted }
    spEncounter_Episodeofcare, {@enum.value "episodeofcare" spEncounter_Episodeofcare Episode(s) of care that this encounter should be recorded against }
    spEncounter_Identifier, {@enum.value "identifier" spEncounter_Identifier Identifier(s) by which this encounter is known }
    spEncounter_Incomingreferral, {@enum.value "incomingreferral" spEncounter_Incomingreferral The ReferralRequest that initiated this encounter }
    spEncounter_Indication, {@enum.value "indication" spEncounter_Indication Reason the encounter takes place (resource) }
    spEncounter_Length, {@enum.value "length" spEncounter_Length Length of encounter in days }
    spEncounter_Location, {@enum.value "location" spEncounter_Location Location the encounter takes place }
    spEncounter_Location_period, {@enum.value "location-period" spEncounter_Location_period Time period during which the patient was present at the location }
    spEncounter_Part_of, {@enum.value "part-of" spEncounter_Part_of Another Encounter this encounter is part of }
    spEncounter_Participant, {@enum.value "participant" spEncounter_Participant Persons involved in the encounter other than the patient }
    spEncounter_Participant_type, {@enum.value "participant-type" spEncounter_Participant_type Role of participant in encounter }
    spEncounter_Patient, {@enum.value "patient" spEncounter_Patient The patient present at the encounter }
    spEncounter_Practitioner, {@enum.value "practitioner" spEncounter_Practitioner Persons involved in the encounter other than the patient }
    spEncounter_Procedure, {@enum.value "procedure" spEncounter_Procedure Reason the encounter takes place (resource) }
    spEncounter_Reason, {@enum.value "reason" spEncounter_Reason Reason the encounter takes place (code) }
    spEncounter_Special_arrangement, {@enum.value "special-arrangement" spEncounter_Special_arrangement Wheelchair, translator, stretcher, etc. }
    spEncounter_Status, {@enum.value "status" spEncounter_Status planned | arrived | in-progress | onleave | finished | cancelled }
    spEncounter_Type); {@enum.value "type" spEncounter_Type Specific type of encounter }

  {@Enum TSearchParamsEnrollmentRequest
    Search Parameters for EnrollmentRequest
  }
  TSearchParamsEnrollmentRequest = (
    spEnrollmentRequest__content, {@enum.value "_content" spEnrollmentRequest__content Search on the entire content of the resource }
    spEnrollmentRequest__id, {@enum.value "_id" spEnrollmentRequest__id Logical id of this artifact }
    spEnrollmentRequest__lastUpdated, {@enum.value "_lastUpdated" spEnrollmentRequest__lastUpdated When the resource version last changed }
    spEnrollmentRequest__profile, {@enum.value "_profile" spEnrollmentRequest__profile Profiles this resource claims to conform to }
    spEnrollmentRequest__query, {@enum.value "_query" spEnrollmentRequest__query A custom search profile that describes a specific defined query operation }
    spEnrollmentRequest__security, {@enum.value "_security" spEnrollmentRequest__security Security Labels applied to this resource }
    spEnrollmentRequest__tag, {@enum.value "_tag" spEnrollmentRequest__tag Tags applied to this resource }
    spEnrollmentRequest__text, {@enum.value "_text" spEnrollmentRequest__text Search on the narrative of the resource }
    spEnrollmentRequest_Identifier, {@enum.value "identifier" spEnrollmentRequest_Identifier The business identifier of the Enrollment }
    spEnrollmentRequest_Patient, {@enum.value "patient" spEnrollmentRequest_Patient The party to be enrolled }
    spEnrollmentRequest_Subject); {@enum.value "subject" spEnrollmentRequest_Subject The party to be enrolled }

  {@Enum TSearchParamsEnrollmentResponse
    Search Parameters for EnrollmentResponse
  }
  TSearchParamsEnrollmentResponse = (
    spEnrollmentResponse__content, {@enum.value "_content" spEnrollmentResponse__content Search on the entire content of the resource }
    spEnrollmentResponse__id, {@enum.value "_id" spEnrollmentResponse__id Logical id of this artifact }
    spEnrollmentResponse__lastUpdated, {@enum.value "_lastUpdated" spEnrollmentResponse__lastUpdated When the resource version last changed }
    spEnrollmentResponse__profile, {@enum.value "_profile" spEnrollmentResponse__profile Profiles this resource claims to conform to }
    spEnrollmentResponse__query, {@enum.value "_query" spEnrollmentResponse__query A custom search profile that describes a specific defined query operation }
    spEnrollmentResponse__security, {@enum.value "_security" spEnrollmentResponse__security Security Labels applied to this resource }
    spEnrollmentResponse__tag, {@enum.value "_tag" spEnrollmentResponse__tag Tags applied to this resource }
    spEnrollmentResponse__text, {@enum.value "_text" spEnrollmentResponse__text Search on the narrative of the resource }
    spEnrollmentResponse_Identifier); {@enum.value "identifier" spEnrollmentResponse_Identifier The business identifier of the Explanation of Benefit }

  {@Enum TSearchParamsEpisodeOfCare
    Search Parameters for EpisodeOfCare
  }
  TSearchParamsEpisodeOfCare = (
    spEpisodeOfCare__content, {@enum.value "_content" spEpisodeOfCare__content Search on the entire content of the resource }
    spEpisodeOfCare__id, {@enum.value "_id" spEpisodeOfCare__id Logical id of this artifact }
    spEpisodeOfCare__lastUpdated, {@enum.value "_lastUpdated" spEpisodeOfCare__lastUpdated When the resource version last changed }
    spEpisodeOfCare__profile, {@enum.value "_profile" spEpisodeOfCare__profile Profiles this resource claims to conform to }
    spEpisodeOfCare__query, {@enum.value "_query" spEpisodeOfCare__query A custom search profile that describes a specific defined query operation }
    spEpisodeOfCare__security, {@enum.value "_security" spEpisodeOfCare__security Security Labels applied to this resource }
    spEpisodeOfCare__tag, {@enum.value "_tag" spEpisodeOfCare__tag Tags applied to this resource }
    spEpisodeOfCare__text, {@enum.value "_text" spEpisodeOfCare__text Search on the narrative of the resource }
    spEpisodeOfCare_Care_manager, {@enum.value "care-manager" spEpisodeOfCare_Care_manager Care manager/care co-ordinator for the patient }
    spEpisodeOfCare_Condition, {@enum.value "condition" spEpisodeOfCare_Condition Conditions/problems/diagnoses this episode of care is for }
    spEpisodeOfCare_Date, {@enum.value "date" spEpisodeOfCare_Date The provided date search value falls within the episode of care's period }
    spEpisodeOfCare_Identifier, {@enum.value "identifier" spEpisodeOfCare_Identifier Identifier(s) for the EpisodeOfCare }
    spEpisodeOfCare_Incomingreferral, {@enum.value "incomingreferral" spEpisodeOfCare_Incomingreferral Incoming Referral Request }
    spEpisodeOfCare_Organization, {@enum.value "organization" spEpisodeOfCare_Organization The organization that has assumed the specific responsibilities of this EpisodeOfCare }
    spEpisodeOfCare_Patient, {@enum.value "patient" spEpisodeOfCare_Patient Patient for this episode of care }
    spEpisodeOfCare_Status, {@enum.value "status" spEpisodeOfCare_Status The current status of the Episode of Care as provided (does not check the status history collection) }
    spEpisodeOfCare_Team_member, {@enum.value "team-member" spEpisodeOfCare_Team_member A Practitioner or Organization allocated to the care team for this EpisodeOfCare }
    spEpisodeOfCare_Type); {@enum.value "type" spEpisodeOfCare_Type Type/class  - e.g. specialist referral, disease management }

  {@Enum TSearchParamsExplanationOfBenefit
    Search Parameters for ExplanationOfBenefit
  }
  TSearchParamsExplanationOfBenefit = (
    spExplanationOfBenefit__content, {@enum.value "_content" spExplanationOfBenefit__content Search on the entire content of the resource }
    spExplanationOfBenefit__id, {@enum.value "_id" spExplanationOfBenefit__id Logical id of this artifact }
    spExplanationOfBenefit__lastUpdated, {@enum.value "_lastUpdated" spExplanationOfBenefit__lastUpdated When the resource version last changed }
    spExplanationOfBenefit__profile, {@enum.value "_profile" spExplanationOfBenefit__profile Profiles this resource claims to conform to }
    spExplanationOfBenefit__query, {@enum.value "_query" spExplanationOfBenefit__query A custom search profile that describes a specific defined query operation }
    spExplanationOfBenefit__security, {@enum.value "_security" spExplanationOfBenefit__security Security Labels applied to this resource }
    spExplanationOfBenefit__tag, {@enum.value "_tag" spExplanationOfBenefit__tag Tags applied to this resource }
    spExplanationOfBenefit__text, {@enum.value "_text" spExplanationOfBenefit__text Search on the narrative of the resource }
    spExplanationOfBenefit_Identifier); {@enum.value "identifier" spExplanationOfBenefit_Identifier The business identifier of the Explanation of Benefit }

  {@Enum TSearchParamsFamilyMemberHistory
    Search Parameters for FamilyMemberHistory
  }
  TSearchParamsFamilyMemberHistory = (
    spFamilyMemberHistory__content, {@enum.value "_content" spFamilyMemberHistory__content Search on the entire content of the resource }
    spFamilyMemberHistory__id, {@enum.value "_id" spFamilyMemberHistory__id Logical id of this artifact }
    spFamilyMemberHistory__lastUpdated, {@enum.value "_lastUpdated" spFamilyMemberHistory__lastUpdated When the resource version last changed }
    spFamilyMemberHistory__profile, {@enum.value "_profile" spFamilyMemberHistory__profile Profiles this resource claims to conform to }
    spFamilyMemberHistory__query, {@enum.value "_query" spFamilyMemberHistory__query A custom search profile that describes a specific defined query operation }
    spFamilyMemberHistory__security, {@enum.value "_security" spFamilyMemberHistory__security Security Labels applied to this resource }
    spFamilyMemberHistory__tag, {@enum.value "_tag" spFamilyMemberHistory__tag Tags applied to this resource }
    spFamilyMemberHistory__text, {@enum.value "_text" spFamilyMemberHistory__text Search on the narrative of the resource }
    spFamilyMemberHistory_Code, {@enum.value "code" spFamilyMemberHistory_Code A search by a condition code }
    spFamilyMemberHistory_Date, {@enum.value "date" spFamilyMemberHistory_Date When history was captured/updated }
    spFamilyMemberHistory_Gender, {@enum.value "gender" spFamilyMemberHistory_Gender A search by a gender code of a family member }
    spFamilyMemberHistory_Identifier, {@enum.value "identifier" spFamilyMemberHistory_Identifier A search by a record identifier }
    spFamilyMemberHistory_Patient, {@enum.value "patient" spFamilyMemberHistory_Patient The identity of a subject to list family member history items for }
    spFamilyMemberHistory_Relationship); {@enum.value "relationship" spFamilyMemberHistory_Relationship A search by a relationship type }

  {@Enum TSearchParamsFlag
    Search Parameters for Flag
  }
  TSearchParamsFlag = (
    spFlag__content, {@enum.value "_content" spFlag__content Search on the entire content of the resource }
    spFlag__id, {@enum.value "_id" spFlag__id Logical id of this artifact }
    spFlag__lastUpdated, {@enum.value "_lastUpdated" spFlag__lastUpdated When the resource version last changed }
    spFlag__profile, {@enum.value "_profile" spFlag__profile Profiles this resource claims to conform to }
    spFlag__query, {@enum.value "_query" spFlag__query A custom search profile that describes a specific defined query operation }
    spFlag__security, {@enum.value "_security" spFlag__security Security Labels applied to this resource }
    spFlag__tag, {@enum.value "_tag" spFlag__tag Tags applied to this resource }
    spFlag__text, {@enum.value "_text" spFlag__text Search on the narrative of the resource }
    spFlag_Author, {@enum.value "author" spFlag_Author Flag creator }
    spFlag_Date, {@enum.value "date" spFlag_Date Time period when flag is active }
    spFlag_Encounter, {@enum.value "encounter" spFlag_Encounter Alert relevant during encounter }
    spFlag_Patient, {@enum.value "patient" spFlag_Patient The identity of a subject to list flags for }
    spFlag_Subject); {@enum.value "subject" spFlag_Subject The identity of a subject to list flags for }

  {@Enum TSearchParamsGoal
    Search Parameters for Goal
  }
  TSearchParamsGoal = (
    spGoal__content, {@enum.value "_content" spGoal__content Search on the entire content of the resource }
    spGoal__id, {@enum.value "_id" spGoal__id Logical id of this artifact }
    spGoal__lastUpdated, {@enum.value "_lastUpdated" spGoal__lastUpdated When the resource version last changed }
    spGoal__profile, {@enum.value "_profile" spGoal__profile Profiles this resource claims to conform to }
    spGoal__query, {@enum.value "_query" spGoal__query A custom search profile that describes a specific defined query operation }
    spGoal__security, {@enum.value "_security" spGoal__security Security Labels applied to this resource }
    spGoal__tag, {@enum.value "_tag" spGoal__tag Tags applied to this resource }
    spGoal__text, {@enum.value "_text" spGoal__text Search on the narrative of the resource }
    spGoal_Category, {@enum.value "category" spGoal_Category E.g. Treatment, dietary, behavioral, etc. }
    spGoal_Identifier, {@enum.value "identifier" spGoal_Identifier External Ids for this goal }
    spGoal_Patient, {@enum.value "patient" spGoal_Patient Who this goal is intended for }
    spGoal_Status, {@enum.value "status" spGoal_Status proposed | planned | accepted | rejected | in-progress | achieved | sustaining | on-hold | cancelled }
    spGoal_Subject, {@enum.value "subject" spGoal_Subject Who this goal is intended for }
    spGoal_Targetdate); {@enum.value "targetdate" spGoal_Targetdate Reach goal on or before }

  {@Enum TSearchParamsGroup
    Search Parameters for Group
  }
  TSearchParamsGroup = (
    spGroup__content, {@enum.value "_content" spGroup__content Search on the entire content of the resource }
    spGroup__id, {@enum.value "_id" spGroup__id Logical id of this artifact }
    spGroup__lastUpdated, {@enum.value "_lastUpdated" spGroup__lastUpdated When the resource version last changed }
    spGroup__profile, {@enum.value "_profile" spGroup__profile Profiles this resource claims to conform to }
    spGroup__query, {@enum.value "_query" spGroup__query A custom search profile that describes a specific defined query operation }
    spGroup__security, {@enum.value "_security" spGroup__security Security Labels applied to this resource }
    spGroup__tag, {@enum.value "_tag" spGroup__tag Tags applied to this resource }
    spGroup__text, {@enum.value "_text" spGroup__text Search on the narrative of the resource }
    spGroup_Actual, {@enum.value "actual" spGroup_Actual Descriptive or actual }
    spGroup_Characteristic, {@enum.value "characteristic" spGroup_Characteristic Kind of characteristic }
    spGroup_Characteristic_value, {@enum.value "characteristic-value" spGroup_Characteristic_value A composite of both characteristic and value }
    spGroup_Code, {@enum.value "code" spGroup_Code The kind of resources contained }
    spGroup_Exclude, {@enum.value "exclude" spGroup_Exclude Group includes or excludes }
    spGroup_Identifier, {@enum.value "identifier" spGroup_Identifier Unique id }
    spGroup_Member, {@enum.value "member" spGroup_Member Reference to the group member }
    spGroup_Type, {@enum.value "type" spGroup_Type The type of resources the group contains }
    spGroup_Value); {@enum.value "value" spGroup_Value Value held by characteristic }

  {@Enum TSearchParamsHealthcareService
    Search Parameters for HealthcareService
  }
  TSearchParamsHealthcareService = (
    spHealthcareService__content, {@enum.value "_content" spHealthcareService__content Search on the entire content of the resource }
    spHealthcareService__id, {@enum.value "_id" spHealthcareService__id Logical id of this artifact }
    spHealthcareService__lastUpdated, {@enum.value "_lastUpdated" spHealthcareService__lastUpdated When the resource version last changed }
    spHealthcareService__profile, {@enum.value "_profile" spHealthcareService__profile Profiles this resource claims to conform to }
    spHealthcareService__query, {@enum.value "_query" spHealthcareService__query A custom search profile that describes a specific defined query operation }
    spHealthcareService__security, {@enum.value "_security" spHealthcareService__security Security Labels applied to this resource }
    spHealthcareService__tag, {@enum.value "_tag" spHealthcareService__tag Tags applied to this resource }
    spHealthcareService__text, {@enum.value "_text" spHealthcareService__text Search on the narrative of the resource }
    spHealthcareService_Characteristic, {@enum.value "characteristic" spHealthcareService_Characteristic One of the HealthcareService's characteristics }
    spHealthcareService_Identifier, {@enum.value "identifier" spHealthcareService_Identifier External identifiers for this item }
    spHealthcareService_Location, {@enum.value "location" spHealthcareService_Location The location of the Healthcare Service }
    spHealthcareService_Name, {@enum.value "name" spHealthcareService_Name A portion of the Healthcare service name }
    spHealthcareService_Organization, {@enum.value "organization" spHealthcareService_Organization The organization that provides this Healthcare Service }
    spHealthcareService_Programname, {@enum.value "programname" spHealthcareService_Programname One of the Program Names serviced by this HealthcareService }
    spHealthcareService_Servicecategory, {@enum.value "servicecategory" spHealthcareService_Servicecategory Service Category of the Healthcare Service }
    spHealthcareService_Servicetype); {@enum.value "servicetype" spHealthcareService_Servicetype The type of service provided by this healthcare service }

  {@Enum TSearchParamsImagingObjectSelection
    Search Parameters for ImagingObjectSelection
  }
  TSearchParamsImagingObjectSelection = (
    spImagingObjectSelection__content, {@enum.value "_content" spImagingObjectSelection__content Search on the entire content of the resource }
    spImagingObjectSelection__id, {@enum.value "_id" spImagingObjectSelection__id Logical id of this artifact }
    spImagingObjectSelection__lastUpdated, {@enum.value "_lastUpdated" spImagingObjectSelection__lastUpdated When the resource version last changed }
    spImagingObjectSelection__profile, {@enum.value "_profile" spImagingObjectSelection__profile Profiles this resource claims to conform to }
    spImagingObjectSelection__query, {@enum.value "_query" spImagingObjectSelection__query A custom search profile that describes a specific defined query operation }
    spImagingObjectSelection__security, {@enum.value "_security" spImagingObjectSelection__security Security Labels applied to this resource }
    spImagingObjectSelection__tag, {@enum.value "_tag" spImagingObjectSelection__tag Tags applied to this resource }
    spImagingObjectSelection__text, {@enum.value "_text" spImagingObjectSelection__text Search on the narrative of the resource }
    spImagingObjectSelection_Author, {@enum.value "author" spImagingObjectSelection_Author Author of key DICOM object selection }
    spImagingObjectSelection_Authoring_time, {@enum.value "authoring-time" spImagingObjectSelection_Authoring_time Time of key DICOM object selection authoring }
    spImagingObjectSelection_Identifier, {@enum.value "identifier" spImagingObjectSelection_Identifier UID of key DICOM object selection }
    spImagingObjectSelection_Patient, {@enum.value "patient" spImagingObjectSelection_Patient Subject of key DICOM object selection }
    spImagingObjectSelection_Selected_study, {@enum.value "selected-study" spImagingObjectSelection_Selected_study Study selected in key DICOM object selection }
    spImagingObjectSelection_Title); {@enum.value "title" spImagingObjectSelection_Title Title of key DICOM object selection }

  {@Enum TSearchParamsImagingStudy
    Search Parameters for ImagingStudy
  }
  TSearchParamsImagingStudy = (
    spImagingStudy__content, {@enum.value "_content" spImagingStudy__content Search on the entire content of the resource }
    spImagingStudy__id, {@enum.value "_id" spImagingStudy__id Logical id of this artifact }
    spImagingStudy__lastUpdated, {@enum.value "_lastUpdated" spImagingStudy__lastUpdated When the resource version last changed }
    spImagingStudy__profile, {@enum.value "_profile" spImagingStudy__profile Profiles this resource claims to conform to }
    spImagingStudy__query, {@enum.value "_query" spImagingStudy__query A custom search profile that describes a specific defined query operation }
    spImagingStudy__security, {@enum.value "_security" spImagingStudy__security Security Labels applied to this resource }
    spImagingStudy__tag, {@enum.value "_tag" spImagingStudy__tag Tags applied to this resource }
    spImagingStudy__text, {@enum.value "_text" spImagingStudy__text Search on the narrative of the resource }
    spImagingStudy_Accession, {@enum.value "accession" spImagingStudy_Accession The accession identifier for the study }
    spImagingStudy_Bodysite, {@enum.value "bodysite" spImagingStudy_Bodysite The body site studied }
    spImagingStudy_Dicom_class, {@enum.value "dicom-class" spImagingStudy_Dicom_class The type of the instance }
    spImagingStudy_Modality, {@enum.value "modality" spImagingStudy_Modality The modality of the series }
    spImagingStudy_Order, {@enum.value "order" spImagingStudy_Order The order for the image }
    spImagingStudy_Patient, {@enum.value "patient" spImagingStudy_Patient Who the study is about }
    spImagingStudy_Series, {@enum.value "series" spImagingStudy_Series The identifier of the series of images }
    spImagingStudy_Started, {@enum.value "started" spImagingStudy_Started When the study was started }
    spImagingStudy_Study, {@enum.value "study" spImagingStudy_Study The study identifier for the image }
    spImagingStudy_Uid); {@enum.value "uid" spImagingStudy_Uid The instance unique identifier }

  {@Enum TSearchParamsImmunization
    Search Parameters for Immunization
  }
  TSearchParamsImmunization = (
    spImmunization__content, {@enum.value "_content" spImmunization__content Search on the entire content of the resource }
    spImmunization__id, {@enum.value "_id" spImmunization__id Logical id of this artifact }
    spImmunization__lastUpdated, {@enum.value "_lastUpdated" spImmunization__lastUpdated When the resource version last changed }
    spImmunization__profile, {@enum.value "_profile" spImmunization__profile Profiles this resource claims to conform to }
    spImmunization__query, {@enum.value "_query" spImmunization__query A custom search profile that describes a specific defined query operation }
    spImmunization__security, {@enum.value "_security" spImmunization__security Security Labels applied to this resource }
    spImmunization__tag, {@enum.value "_tag" spImmunization__tag Tags applied to this resource }
    spImmunization__text, {@enum.value "_text" spImmunization__text Search on the narrative of the resource }
    spImmunization_Date, {@enum.value "date" spImmunization_Date Vaccination  (non)-Administration Date }
    spImmunization_Dose_sequence, {@enum.value "dose-sequence" spImmunization_Dose_sequence Dose number within series }
    spImmunization_Identifier, {@enum.value "identifier" spImmunization_Identifier Business identifier }
    spImmunization_Location, {@enum.value "location" spImmunization_Location The service delivery location or facility in which the vaccine was / was to be administered }
    spImmunization_Lot_number, {@enum.value "lot-number" spImmunization_Lot_number Vaccine Lot Number }
    spImmunization_Manufacturer, {@enum.value "manufacturer" spImmunization_Manufacturer Vaccine Manufacturer }
    spImmunization_Notgiven, {@enum.value "notgiven" spImmunization_Notgiven Administrations which were not given }
    spImmunization_Patient, {@enum.value "patient" spImmunization_Patient The patient for the vaccination record }
    spImmunization_Performer, {@enum.value "performer" spImmunization_Performer The practitioner who administered the vaccination }
    spImmunization_Reaction, {@enum.value "reaction" spImmunization_Reaction Additional information on reaction }
    spImmunization_Reaction_date, {@enum.value "reaction-date" spImmunization_Reaction_date When reaction started }
    spImmunization_Reason, {@enum.value "reason" spImmunization_Reason Why immunization occurred }
    spImmunization_Reason_not_given, {@enum.value "reason-not-given" spImmunization_Reason_not_given Explanation of reason vaccination was not administered }
    spImmunization_Requester, {@enum.value "requester" spImmunization_Requester The practitioner who ordered the vaccination }
    spImmunization_Status, {@enum.value "status" spImmunization_Status Immunization event status }
    spImmunization_Vaccine_code); {@enum.value "vaccine-code" spImmunization_Vaccine_code Vaccine Product Administered }

  {@Enum TSearchParamsImmunizationRecommendation
    Search Parameters for ImmunizationRecommendation
  }
  TSearchParamsImmunizationRecommendation = (
    spImmunizationRecommendation__content, {@enum.value "_content" spImmunizationRecommendation__content Search on the entire content of the resource }
    spImmunizationRecommendation__id, {@enum.value "_id" spImmunizationRecommendation__id Logical id of this artifact }
    spImmunizationRecommendation__lastUpdated, {@enum.value "_lastUpdated" spImmunizationRecommendation__lastUpdated When the resource version last changed }
    spImmunizationRecommendation__profile, {@enum.value "_profile" spImmunizationRecommendation__profile Profiles this resource claims to conform to }
    spImmunizationRecommendation__query, {@enum.value "_query" spImmunizationRecommendation__query A custom search profile that describes a specific defined query operation }
    spImmunizationRecommendation__security, {@enum.value "_security" spImmunizationRecommendation__security Security Labels applied to this resource }
    spImmunizationRecommendation__tag, {@enum.value "_tag" spImmunizationRecommendation__tag Tags applied to this resource }
    spImmunizationRecommendation__text, {@enum.value "_text" spImmunizationRecommendation__text Search on the narrative of the resource }
    spImmunizationRecommendation_Date, {@enum.value "date" spImmunizationRecommendation_Date Date recommendation created }
    spImmunizationRecommendation_Dose_number, {@enum.value "dose-number" spImmunizationRecommendation_Dose_number Recommended dose number }
    spImmunizationRecommendation_Dose_sequence, {@enum.value "dose-sequence" spImmunizationRecommendation_Dose_sequence Dose number within sequence }
    spImmunizationRecommendation_Identifier, {@enum.value "identifier" spImmunizationRecommendation_Identifier Business identifier }
    spImmunizationRecommendation_Information, {@enum.value "information" spImmunizationRecommendation_Information Patient observations supporting recommendation }
    spImmunizationRecommendation_Patient, {@enum.value "patient" spImmunizationRecommendation_Patient Who this profile is for }
    spImmunizationRecommendation_Status, {@enum.value "status" spImmunizationRecommendation_Status Vaccine administration status }
    spImmunizationRecommendation_Support, {@enum.value "support" spImmunizationRecommendation_Support Past immunizations supporting recommendation }
    spImmunizationRecommendation_Vaccine_type); {@enum.value "vaccine-type" spImmunizationRecommendation_Vaccine_type Vaccine recommendation applies to }

  {@Enum TSearchParamsImplementationGuide
    Search Parameters for ImplementationGuide
  }
  TSearchParamsImplementationGuide = (
    spImplementationGuide__content, {@enum.value "_content" spImplementationGuide__content Search on the entire content of the resource }
    spImplementationGuide__id, {@enum.value "_id" spImplementationGuide__id Logical id of this artifact }
    spImplementationGuide__lastUpdated, {@enum.value "_lastUpdated" spImplementationGuide__lastUpdated When the resource version last changed }
    spImplementationGuide__profile, {@enum.value "_profile" spImplementationGuide__profile Profiles this resource claims to conform to }
    spImplementationGuide__query, {@enum.value "_query" spImplementationGuide__query A custom search profile that describes a specific defined query operation }
    spImplementationGuide__security, {@enum.value "_security" spImplementationGuide__security Security Labels applied to this resource }
    spImplementationGuide__tag, {@enum.value "_tag" spImplementationGuide__tag Tags applied to this resource }
    spImplementationGuide__text, {@enum.value "_text" spImplementationGuide__text Search on the narrative of the resource }
    spImplementationGuide_Context, {@enum.value "context" spImplementationGuide_Context A use context assigned to the structure }
    spImplementationGuide_Date, {@enum.value "date" spImplementationGuide_Date The implementation guide publication date }
    spImplementationGuide_Dependency, {@enum.value "dependency" spImplementationGuide_Dependency Where to find dependency }
    spImplementationGuide_Description, {@enum.value "description" spImplementationGuide_Description Text search in the description of the implementation guide }
    spImplementationGuide_Experimental, {@enum.value "experimental" spImplementationGuide_Experimental If for testing purposes, not real usage }
    spImplementationGuide_Name, {@enum.value "name" spImplementationGuide_Name Name of the implementation guide }
    spImplementationGuide_Publisher, {@enum.value "publisher" spImplementationGuide_Publisher Name of the publisher of the implementation guide }
    spImplementationGuide_Status, {@enum.value "status" spImplementationGuide_Status The current status of the implementation guide }
    spImplementationGuide_Url, {@enum.value "url" spImplementationGuide_Url Absolute URL used to reference this Implementation Guide }
    spImplementationGuide_Version); {@enum.value "version" spImplementationGuide_Version The version identifier of the implementation guide }

  {@Enum TSearchParamsList
    Search Parameters for List
  }
  TSearchParamsList = (
    spList__content, {@enum.value "_content" spList__content Search on the entire content of the resource }
    spList__id, {@enum.value "_id" spList__id Logical id of this artifact }
    spList__lastUpdated, {@enum.value "_lastUpdated" spList__lastUpdated When the resource version last changed }
    spList__profile, {@enum.value "_profile" spList__profile Profiles this resource claims to conform to }
    spList__query, {@enum.value "_query" spList__query A custom search profile that describes a specific defined query operation }
    spList__security, {@enum.value "_security" spList__security Security Labels applied to this resource }
    spList__tag, {@enum.value "_tag" spList__tag Tags applied to this resource }
    spList__text, {@enum.value "_text" spList__text Search on the narrative of the resource }
    spList_Code, {@enum.value "code" spList_Code What the purpose of this list is }
    spList_Date, {@enum.value "date" spList_Date When the list was prepared }
    spList_Empty_reason, {@enum.value "empty-reason" spList_Empty_reason Why list is empty }
    spList_Encounter, {@enum.value "encounter" spList_Encounter Context in which list created }
    spList_Item, {@enum.value "item" spList_Item Actual entry }
    spList_Notes, {@enum.value "notes" spList_Notes Comments about the list }
    spList_Patient, {@enum.value "patient" spList_Patient If all resources have the same subject }
    spList_Source, {@enum.value "source" spList_Source Who and/or what defined the list contents (aka Author) }
    spList_Status, {@enum.value "status" spList_Status current | retired | entered-in-error }
    spList_Subject, {@enum.value "subject" spList_Subject If all resources have the same subject }
    spList_Title); {@enum.value "title" spList_Title Descriptive name for the list }

  {@Enum TSearchParamsLocation
    Search Parameters for Location
  }
  TSearchParamsLocation = (
    spLocation__content, {@enum.value "_content" spLocation__content Search on the entire content of the resource }
    spLocation__id, {@enum.value "_id" spLocation__id Logical id of this artifact }
    spLocation__lastUpdated, {@enum.value "_lastUpdated" spLocation__lastUpdated When the resource version last changed }
    spLocation__profile, {@enum.value "_profile" spLocation__profile Profiles this resource claims to conform to }
    spLocation__query, {@enum.value "_query" spLocation__query A custom search profile that describes a specific defined query operation }
    spLocation__security, {@enum.value "_security" spLocation__security Security Labels applied to this resource }
    spLocation__tag, {@enum.value "_tag" spLocation__tag Tags applied to this resource }
    spLocation__text, {@enum.value "_text" spLocation__text Search on the narrative of the resource }
    spLocation_Address, {@enum.value "address" spLocation_Address A (part of the) address of the location }
    spLocation_Address_city, {@enum.value "address-city" spLocation_Address_city A city specified in an address }
    spLocation_Address_country, {@enum.value "address-country" spLocation_Address_country A country specified in an address }
    spLocation_Address_postalcode, {@enum.value "address-postalcode" spLocation_Address_postalcode A postal code specified in an address }
    spLocation_Address_state, {@enum.value "address-state" spLocation_Address_state A state specified in an address }
    spLocation_Address_use, {@enum.value "address-use" spLocation_Address_use A use code specified in an address }
    spLocation_Identifier, {@enum.value "identifier" spLocation_Identifier Unique code or number identifying the location to its users }
    spLocation_Name, {@enum.value "name" spLocation_Name A (portion of the) name of the location }
    spLocation_Near, {@enum.value "near" spLocation_Near The coordinates expressed as [lat],[long] (using the WGS84 datum, see notes) to find locations near to (servers may search using a square rather than a circle for efficiency) }
    spLocation_Near_distance, {@enum.value "near-distance" spLocation_Near_distance A distance quantity to limit the near search to locations within a specific distance }
    spLocation_Organization, {@enum.value "organization" spLocation_Organization Searches for locations that are managed by the provided organization }
    spLocation_Partof, {@enum.value "partof" spLocation_Partof The location of which this location is a part }
    spLocation_Status, {@enum.value "status" spLocation_Status Searches for locations with a specific kind of status }
    spLocation_Type); {@enum.value "type" spLocation_Type A code for the type of location }

  {@Enum TSearchParamsMedia
    Search Parameters for Media
  }
  TSearchParamsMedia = (
    spMedia__content, {@enum.value "_content" spMedia__content Search on the entire content of the resource }
    spMedia__id, {@enum.value "_id" spMedia__id Logical id of this artifact }
    spMedia__lastUpdated, {@enum.value "_lastUpdated" spMedia__lastUpdated When the resource version last changed }
    spMedia__profile, {@enum.value "_profile" spMedia__profile Profiles this resource claims to conform to }
    spMedia__query, {@enum.value "_query" spMedia__query A custom search profile that describes a specific defined query operation }
    spMedia__security, {@enum.value "_security" spMedia__security Security Labels applied to this resource }
    spMedia__tag, {@enum.value "_tag" spMedia__tag Tags applied to this resource }
    spMedia__text, {@enum.value "_text" spMedia__text Search on the narrative of the resource }
    spMedia_Created, {@enum.value "created" spMedia_Created Date attachment was first created }
    spMedia_Identifier, {@enum.value "identifier" spMedia_Identifier Identifier(s) for the image }
    spMedia_Operator, {@enum.value "operator" spMedia_Operator The person who generated the image }
    spMedia_Patient, {@enum.value "patient" spMedia_Patient Who/What this Media is a record of }
    spMedia_Subject, {@enum.value "subject" spMedia_Subject Who/What this Media is a record of }
    spMedia_Subtype, {@enum.value "subtype" spMedia_Subtype The type of acquisition equipment/process }
    spMedia_Type, {@enum.value "type" spMedia_Type photo | video | audio }
    spMedia_View); {@enum.value "view" spMedia_View Imaging view, e.g. Lateral or Antero-posterior }

  {@Enum TSearchParamsMedication
    Search Parameters for Medication
  }
  TSearchParamsMedication = (
    spMedication__content, {@enum.value "_content" spMedication__content Search on the entire content of the resource }
    spMedication__id, {@enum.value "_id" spMedication__id Logical id of this artifact }
    spMedication__lastUpdated, {@enum.value "_lastUpdated" spMedication__lastUpdated When the resource version last changed }
    spMedication__profile, {@enum.value "_profile" spMedication__profile Profiles this resource claims to conform to }
    spMedication__query, {@enum.value "_query" spMedication__query A custom search profile that describes a specific defined query operation }
    spMedication__security, {@enum.value "_security" spMedication__security Security Labels applied to this resource }
    spMedication__tag, {@enum.value "_tag" spMedication__tag Tags applied to this resource }
    spMedication__text, {@enum.value "_text" spMedication__text Search on the narrative of the resource }
    spMedication_Code, {@enum.value "code" spMedication_Code Codes that identify this medication }
    spMedication_Container, {@enum.value "container" spMedication_Container E.g. box, vial, blister-pack }
    spMedication_Content, {@enum.value "content" spMedication_Content A product in the package }
    spMedication_Form, {@enum.value "form" spMedication_Form powder | tablets | carton + }
    spMedication_Ingredient, {@enum.value "ingredient" spMedication_Ingredient The product contained }
    spMedication_Manufacturer); {@enum.value "manufacturer" spMedication_Manufacturer Manufacturer of the item }

  {@Enum TSearchParamsMedicationAdministration
    Search Parameters for MedicationAdministration
  }
  TSearchParamsMedicationAdministration = (
    spMedicationAdministration__content, {@enum.value "_content" spMedicationAdministration__content Search on the entire content of the resource }
    spMedicationAdministration__id, {@enum.value "_id" spMedicationAdministration__id Logical id of this artifact }
    spMedicationAdministration__lastUpdated, {@enum.value "_lastUpdated" spMedicationAdministration__lastUpdated When the resource version last changed }
    spMedicationAdministration__profile, {@enum.value "_profile" spMedicationAdministration__profile Profiles this resource claims to conform to }
    spMedicationAdministration__query, {@enum.value "_query" spMedicationAdministration__query A custom search profile that describes a specific defined query operation }
    spMedicationAdministration__security, {@enum.value "_security" spMedicationAdministration__security Security Labels applied to this resource }
    spMedicationAdministration__tag, {@enum.value "_tag" spMedicationAdministration__tag Tags applied to this resource }
    spMedicationAdministration__text, {@enum.value "_text" spMedicationAdministration__text Search on the narrative of the resource }
    spMedicationAdministration_Code, {@enum.value "code" spMedicationAdministration_Code Return administrations of this medication code }
    spMedicationAdministration_Device, {@enum.value "device" spMedicationAdministration_Device Return administrations with this administration device identity }
    spMedicationAdministration_Effectivetime, {@enum.value "effectivetime" spMedicationAdministration_Effectivetime Date administration happened (or did not happen) }
    spMedicationAdministration_Encounter, {@enum.value "encounter" spMedicationAdministration_Encounter Return administrations that share this encounter }
    spMedicationAdministration_Identifier, {@enum.value "identifier" spMedicationAdministration_Identifier Return administrations with this external identifier }
    spMedicationAdministration_Medication, {@enum.value "medication" spMedicationAdministration_Medication Return administrations of this medication resource }
    spMedicationAdministration_Notgiven, {@enum.value "notgiven" spMedicationAdministration_Notgiven Administrations that were not made }
    spMedicationAdministration_Patient, {@enum.value "patient" spMedicationAdministration_Patient The identity of a patient to list administrations  for }
    spMedicationAdministration_Practitioner, {@enum.value "practitioner" spMedicationAdministration_Practitioner Who administered substance }
    spMedicationAdministration_Prescription, {@enum.value "prescription" spMedicationAdministration_Prescription The identity of a prescription to list administrations from }
    spMedicationAdministration_Status); {@enum.value "status" spMedicationAdministration_Status MedicationAdministration event status (for example one of active/paused/completed/nullified) }

  {@Enum TSearchParamsMedicationDispense
    Search Parameters for MedicationDispense
  }
  TSearchParamsMedicationDispense = (
    spMedicationDispense__content, {@enum.value "_content" spMedicationDispense__content Search on the entire content of the resource }
    spMedicationDispense__id, {@enum.value "_id" spMedicationDispense__id Logical id of this artifact }
    spMedicationDispense__lastUpdated, {@enum.value "_lastUpdated" spMedicationDispense__lastUpdated When the resource version last changed }
    spMedicationDispense__profile, {@enum.value "_profile" spMedicationDispense__profile Profiles this resource claims to conform to }
    spMedicationDispense__query, {@enum.value "_query" spMedicationDispense__query A custom search profile that describes a specific defined query operation }
    spMedicationDispense__security, {@enum.value "_security" spMedicationDispense__security Security Labels applied to this resource }
    spMedicationDispense__tag, {@enum.value "_tag" spMedicationDispense__tag Tags applied to this resource }
    spMedicationDispense__text, {@enum.value "_text" spMedicationDispense__text Search on the narrative of the resource }
    spMedicationDispense_Code, {@enum.value "code" spMedicationDispense_Code Return dispenses of this medicine code }
    spMedicationDispense_Destination, {@enum.value "destination" spMedicationDispense_Destination Return dispenses that should be sent to a specific destination }
    spMedicationDispense_Dispenser, {@enum.value "dispenser" spMedicationDispense_Dispenser Return all dispenses performed by a specific individual }
    spMedicationDispense_Identifier, {@enum.value "identifier" spMedicationDispense_Identifier Return dispenses with this external identifier }
    spMedicationDispense_Medication, {@enum.value "medication" spMedicationDispense_Medication Return dispenses of this medicine resource }
    spMedicationDispense_Patient, {@enum.value "patient" spMedicationDispense_Patient The identity of a patient to list dispenses  for }
    spMedicationDispense_Prescription, {@enum.value "prescription" spMedicationDispense_Prescription The identity of a prescription to list dispenses from }
    spMedicationDispense_Receiver, {@enum.value "receiver" spMedicationDispense_Receiver Who collected the medication }
    spMedicationDispense_Responsibleparty, {@enum.value "responsibleparty" spMedicationDispense_Responsibleparty Return all dispenses with the specified responsible party }
    spMedicationDispense_Status, {@enum.value "status" spMedicationDispense_Status Status of the dispense }
    spMedicationDispense_Type, {@enum.value "type" spMedicationDispense_Type Return all dispenses of a specific type }
    spMedicationDispense_Whenhandedover, {@enum.value "whenhandedover" spMedicationDispense_Whenhandedover Date when medication handed over to patient (outpatient setting), or supplied to ward or clinic (inpatient setting) }
    spMedicationDispense_Whenprepared); {@enum.value "whenprepared" spMedicationDispense_Whenprepared Date when medication prepared }

  {@Enum TSearchParamsMedicationOrder
    Search Parameters for MedicationOrder
  }
  TSearchParamsMedicationOrder = (
    spMedicationOrder__content, {@enum.value "_content" spMedicationOrder__content Search on the entire content of the resource }
    spMedicationOrder__id, {@enum.value "_id" spMedicationOrder__id Logical id of this artifact }
    spMedicationOrder__lastUpdated, {@enum.value "_lastUpdated" spMedicationOrder__lastUpdated When the resource version last changed }
    spMedicationOrder__profile, {@enum.value "_profile" spMedicationOrder__profile Profiles this resource claims to conform to }
    spMedicationOrder__query, {@enum.value "_query" spMedicationOrder__query A custom search profile that describes a specific defined query operation }
    spMedicationOrder__security, {@enum.value "_security" spMedicationOrder__security Security Labels applied to this resource }
    spMedicationOrder__tag, {@enum.value "_tag" spMedicationOrder__tag Tags applied to this resource }
    spMedicationOrder__text, {@enum.value "_text" spMedicationOrder__text Search on the narrative of the resource }
    spMedicationOrder_Code, {@enum.value "code" spMedicationOrder_Code Return administrations of this medication code }
    spMedicationOrder_Datewritten, {@enum.value "datewritten" spMedicationOrder_Datewritten Return prescriptions written on this date }
    spMedicationOrder_Encounter, {@enum.value "encounter" spMedicationOrder_Encounter Return prescriptions with this encounter identifier }
    spMedicationOrder_Identifier, {@enum.value "identifier" spMedicationOrder_Identifier Return prescriptions with this external identifier }
    spMedicationOrder_Medication, {@enum.value "medication" spMedicationOrder_Medication Return administrations of this medication reference }
    spMedicationOrder_Patient, {@enum.value "patient" spMedicationOrder_Patient The identity of a patient to list orders  for }
    spMedicationOrder_Prescriber, {@enum.value "prescriber" spMedicationOrder_Prescriber Who ordered the medication(s) }
    spMedicationOrder_Status); {@enum.value "status" spMedicationOrder_Status Status of the prescription }

  {@Enum TSearchParamsMedicationStatement
    Search Parameters for MedicationStatement
  }
  TSearchParamsMedicationStatement = (
    spMedicationStatement__content, {@enum.value "_content" spMedicationStatement__content Search on the entire content of the resource }
    spMedicationStatement__id, {@enum.value "_id" spMedicationStatement__id Logical id of this artifact }
    spMedicationStatement__lastUpdated, {@enum.value "_lastUpdated" spMedicationStatement__lastUpdated When the resource version last changed }
    spMedicationStatement__profile, {@enum.value "_profile" spMedicationStatement__profile Profiles this resource claims to conform to }
    spMedicationStatement__query, {@enum.value "_query" spMedicationStatement__query A custom search profile that describes a specific defined query operation }
    spMedicationStatement__security, {@enum.value "_security" spMedicationStatement__security Security Labels applied to this resource }
    spMedicationStatement__tag, {@enum.value "_tag" spMedicationStatement__tag Tags applied to this resource }
    spMedicationStatement__text, {@enum.value "_text" spMedicationStatement__text Search on the narrative of the resource }
    spMedicationStatement_Code, {@enum.value "code" spMedicationStatement_Code Return administrations of this medication code }
    spMedicationStatement_Effectivedate, {@enum.value "effectivedate" spMedicationStatement_Effectivedate Date when patient was taking (or not taking) the medication }
    spMedicationStatement_Identifier, {@enum.value "identifier" spMedicationStatement_Identifier Return statements with this external identifier }
    spMedicationStatement_Medication, {@enum.value "medication" spMedicationStatement_Medication Return administrations of this medication reference }
    spMedicationStatement_Patient, {@enum.value "patient" spMedicationStatement_Patient The identity of a patient to list statements  for }
    spMedicationStatement_Source, {@enum.value "source" spMedicationStatement_Source Who the information in the statement came from }
    spMedicationStatement_Status); {@enum.value "status" spMedicationStatement_Status Return statements that match the given status }

  {@Enum TSearchParamsMessageHeader
    Search Parameters for MessageHeader
  }
  TSearchParamsMessageHeader = (
    spMessageHeader__content, {@enum.value "_content" spMessageHeader__content Search on the entire content of the resource }
    spMessageHeader__id, {@enum.value "_id" spMessageHeader__id Logical id of this artifact }
    spMessageHeader__lastUpdated, {@enum.value "_lastUpdated" spMessageHeader__lastUpdated When the resource version last changed }
    spMessageHeader__profile, {@enum.value "_profile" spMessageHeader__profile Profiles this resource claims to conform to }
    spMessageHeader__query, {@enum.value "_query" spMessageHeader__query A custom search profile that describes a specific defined query operation }
    spMessageHeader__security, {@enum.value "_security" spMessageHeader__security Security Labels applied to this resource }
    spMessageHeader__tag, {@enum.value "_tag" spMessageHeader__tag Tags applied to this resource }
    spMessageHeader__text, {@enum.value "_text" spMessageHeader__text Search on the narrative of the resource }
    spMessageHeader_Author, {@enum.value "author" spMessageHeader_Author The source of the decision }
    spMessageHeader_Code, {@enum.value "code" spMessageHeader_Code ok | transient-error | fatal-error }
    spMessageHeader_Data, {@enum.value "data" spMessageHeader_Data The actual content of the message }
    spMessageHeader_Destination, {@enum.value "destination" spMessageHeader_Destination Name of system }
    spMessageHeader_Destination_uri, {@enum.value "destination-uri" spMessageHeader_Destination_uri Actual destination address or id }
    spMessageHeader_Enterer, {@enum.value "enterer" spMessageHeader_Enterer The source of the data entry }
    spMessageHeader_Event, {@enum.value "event" spMessageHeader_Event Code for the event this message represents }
    spMessageHeader_Receiver, {@enum.value "receiver" spMessageHeader_Receiver Intended "real-world" recipient for the data }
    spMessageHeader_Response_id, {@enum.value "response-id" spMessageHeader_Response_id Id of original message }
    spMessageHeader_Responsible, {@enum.value "responsible" spMessageHeader_Responsible Final responsibility for event }
    spMessageHeader_Source, {@enum.value "source" spMessageHeader_Source Name of system }
    spMessageHeader_Source_uri, {@enum.value "source-uri" spMessageHeader_Source_uri Actual message source address or id }
    spMessageHeader_Target, {@enum.value "target" spMessageHeader_Target Particular delivery destination within the destination }
    spMessageHeader_Timestamp); {@enum.value "timestamp" spMessageHeader_Timestamp Time that the message was sent }

  {@Enum TSearchParamsNamingSystem
    Search Parameters for NamingSystem
  }
  TSearchParamsNamingSystem = (
    spNamingSystem__content, {@enum.value "_content" spNamingSystem__content Search on the entire content of the resource }
    spNamingSystem__id, {@enum.value "_id" spNamingSystem__id Logical id of this artifact }
    spNamingSystem__lastUpdated, {@enum.value "_lastUpdated" spNamingSystem__lastUpdated When the resource version last changed }
    spNamingSystem__profile, {@enum.value "_profile" spNamingSystem__profile Profiles this resource claims to conform to }
    spNamingSystem__query, {@enum.value "_query" spNamingSystem__query A custom search profile that describes a specific defined query operation }
    spNamingSystem__security, {@enum.value "_security" spNamingSystem__security Security Labels applied to this resource }
    spNamingSystem__tag, {@enum.value "_tag" spNamingSystem__tag Tags applied to this resource }
    spNamingSystem__text, {@enum.value "_text" spNamingSystem__text Search on the narrative of the resource }
    spNamingSystem_Contact, {@enum.value "contact" spNamingSystem_Contact Name of a individual to contact }
    spNamingSystem_Context, {@enum.value "context" spNamingSystem_Context Content intends to support these contexts }
    spNamingSystem_Date, {@enum.value "date" spNamingSystem_Date Publication Date(/time) }
    spNamingSystem_Id_type, {@enum.value "id-type" spNamingSystem_Id_type oid | uuid | uri | other }
    spNamingSystem_Kind, {@enum.value "kind" spNamingSystem_Kind codesystem | identifier | root }
    spNamingSystem_Name, {@enum.value "name" spNamingSystem_Name Human-readable label }
    spNamingSystem_Period, {@enum.value "period" spNamingSystem_Period When is identifier valid? }
    spNamingSystem_Publisher, {@enum.value "publisher" spNamingSystem_Publisher Name of the publisher (Organization or individual) }
    spNamingSystem_Replaced_by, {@enum.value "replaced-by" spNamingSystem_Replaced_by Use this instead }
    spNamingSystem_Responsible, {@enum.value "responsible" spNamingSystem_Responsible Who maintains system namespace? }
    spNamingSystem_Status, {@enum.value "status" spNamingSystem_Status draft | active | retired }
    spNamingSystem_Telecom, {@enum.value "telecom" spNamingSystem_Telecom Contact details for individual or publisher }
    spNamingSystem_Type, {@enum.value "type" spNamingSystem_Type e.g. driver,  provider,  patient, bank etc. }
    spNamingSystem_Value); {@enum.value "value" spNamingSystem_Value The unique identifier }

  {@Enum TSearchParamsNutritionOrder
    Search Parameters for NutritionOrder
  }
  TSearchParamsNutritionOrder = (
    spNutritionOrder__content, {@enum.value "_content" spNutritionOrder__content Search on the entire content of the resource }
    spNutritionOrder__id, {@enum.value "_id" spNutritionOrder__id Logical id of this artifact }
    spNutritionOrder__lastUpdated, {@enum.value "_lastUpdated" spNutritionOrder__lastUpdated When the resource version last changed }
    spNutritionOrder__profile, {@enum.value "_profile" spNutritionOrder__profile Profiles this resource claims to conform to }
    spNutritionOrder__query, {@enum.value "_query" spNutritionOrder__query A custom search profile that describes a specific defined query operation }
    spNutritionOrder__security, {@enum.value "_security" spNutritionOrder__security Security Labels applied to this resource }
    spNutritionOrder__tag, {@enum.value "_tag" spNutritionOrder__tag Tags applied to this resource }
    spNutritionOrder__text, {@enum.value "_text" spNutritionOrder__text Search on the narrative of the resource }
    spNutritionOrder_Additive, {@enum.value "additive" spNutritionOrder_Additive Type of module component to add to the feeding }
    spNutritionOrder_Datetime, {@enum.value "datetime" spNutritionOrder_Datetime Return nutrition orders requested on this date }
    spNutritionOrder_Encounter, {@enum.value "encounter" spNutritionOrder_Encounter Return nutrition orders with this encounter identifier }
    spNutritionOrder_Formula, {@enum.value "formula" spNutritionOrder_Formula Type of enteral or infant formula }
    spNutritionOrder_Identifier, {@enum.value "identifier" spNutritionOrder_Identifier Return nutrition orders with this external identifier }
    spNutritionOrder_Oraldiet, {@enum.value "oraldiet" spNutritionOrder_Oraldiet Type of diet that can be consumed orally (i.e., take via the mouth). }
    spNutritionOrder_Patient, {@enum.value "patient" spNutritionOrder_Patient The identity of the person who requires the diet, formula or nutritional supplement }
    spNutritionOrder_Provider, {@enum.value "provider" spNutritionOrder_Provider The identify of the provider who placed the nutrition order }
    spNutritionOrder_Status, {@enum.value "status" spNutritionOrder_Status Status of the nutrition order. }
    spNutritionOrder_Supplement); {@enum.value "supplement" spNutritionOrder_Supplement Type of supplement product requested }

  {@Enum TSearchParamsObservation
    Search Parameters for Observation
  }
  TSearchParamsObservation = (
    spObservation__content, {@enum.value "_content" spObservation__content Search on the entire content of the resource }
    spObservation__id, {@enum.value "_id" spObservation__id Logical id of this artifact }
    spObservation__lastUpdated, {@enum.value "_lastUpdated" spObservation__lastUpdated When the resource version last changed }
    spObservation__profile, {@enum.value "_profile" spObservation__profile Profiles this resource claims to conform to }
    spObservation__query, {@enum.value "_query" spObservation__query A custom search profile that describes a specific defined query operation }
    spObservation__security, {@enum.value "_security" spObservation__security Security Labels applied to this resource }
    spObservation__tag, {@enum.value "_tag" spObservation__tag Tags applied to this resource }
    spObservation__text, {@enum.value "_text" spObservation__text Search on the narrative of the resource }
    spObservation_Category, {@enum.value "category" spObservation_Category The classification of the type of observation }
    spObservation_Code, {@enum.value "code" spObservation_Code The code of the observation type }
    spObservation_Code_value_x, {@enum.value "code-value-[x]" spObservation_Code_value_x Both code and one of the value parameters }
    spObservation_Component_code, {@enum.value "component-code" spObservation_Component_code The component code of the observation type }
    spObservation_Component_code_value_x, {@enum.value "component-code-value-[x]" spObservation_Component_code_value_x Both component code and one of the component value parameters }
    spObservation_Component_data_absent_reason, {@enum.value "component-data-absent-reason" spObservation_Component_data_absent_reason The reason why the expected value in the element Observation.component.value[x] is missing. }
    spObservation_Component_value_concept, {@enum.value "component-value-concept" spObservation_Component_value_concept The value of the component observation, if the value is a CodeableConcept }
    spObservation_Component_value_quantity, {@enum.value "component-value-quantity" spObservation_Component_value_quantity The value of the component observation, if the value is a Quantity, or a SampledData (just search on the bounds of the values in sampled data) }
    spObservation_Component_value_string, {@enum.value "component-value-string" spObservation_Component_value_string The value of the component observation, if the value is a string, and also searches in CodeableConcept.text }
    spObservation_Data_absent_reason, {@enum.value "data-absent-reason" spObservation_Data_absent_reason The reason why the expected value in the element Observation.value[x] is missing. }
    spObservation_Date, {@enum.value "date" spObservation_Date Obtained date/time. If the obtained element is a period, a date that falls in the period }
    spObservation_Device, {@enum.value "device" spObservation_Device The Device that generated the observation data. }
    spObservation_Encounter, {@enum.value "encounter" spObservation_Encounter Healthcare event related to the observation }
    spObservation_Identifier, {@enum.value "identifier" spObservation_Identifier The unique id for a particular observation }
    spObservation_Patient, {@enum.value "patient" spObservation_Patient The subject that the observation is about (if patient) }
    spObservation_Performer, {@enum.value "performer" spObservation_Performer Who performed the observation }
    spObservation_Related, {@enum.value "related" spObservation_Related Related Observations - search on related-type and related-target together }
    spObservation_Related_target, {@enum.value "related-target" spObservation_Related_target Resource that is related to this one }
    spObservation_Related_type, {@enum.value "related-type" spObservation_Related_type has-member | derived-from | sequel-to | replaces | qualified-by | interfered-by }
    spObservation_Specimen, {@enum.value "specimen" spObservation_Specimen Specimen used for this observation }
    spObservation_Status, {@enum.value "status" spObservation_Status The status of the observation }
    spObservation_Subject, {@enum.value "subject" spObservation_Subject The subject that the observation is about }
    spObservation_Value_concept, {@enum.value "value-concept" spObservation_Value_concept The value of the observation, if the value is a CodeableConcept }
    spObservation_Value_date, {@enum.value "value-date" spObservation_Value_date The value of the observation, if the value is a date or period of time }
    spObservation_Value_quantity, {@enum.value "value-quantity" spObservation_Value_quantity The value of the observation, if the value is a Quantity, or a SampledData (just search on the bounds of the values in sampled data) }
    spObservation_Value_string); {@enum.value "value-string" spObservation_Value_string The value of the observation, if the value is a string, and also searches in CodeableConcept.text }

  {@Enum TSearchParamsOperationDefinition
    Search Parameters for OperationDefinition
  }
  TSearchParamsOperationDefinition = (
    spOperationDefinition__content, {@enum.value "_content" spOperationDefinition__content Search on the entire content of the resource }
    spOperationDefinition__id, {@enum.value "_id" spOperationDefinition__id Logical id of this artifact }
    spOperationDefinition__lastUpdated, {@enum.value "_lastUpdated" spOperationDefinition__lastUpdated When the resource version last changed }
    spOperationDefinition__profile, {@enum.value "_profile" spOperationDefinition__profile Profiles this resource claims to conform to }
    spOperationDefinition__query, {@enum.value "_query" spOperationDefinition__query A custom search profile that describes a specific defined query operation }
    spOperationDefinition__security, {@enum.value "_security" spOperationDefinition__security Security Labels applied to this resource }
    spOperationDefinition__tag, {@enum.value "_tag" spOperationDefinition__tag Tags applied to this resource }
    spOperationDefinition__text, {@enum.value "_text" spOperationDefinition__text Search on the narrative of the resource }
    spOperationDefinition_Base, {@enum.value "base" spOperationDefinition_Base Marks this as a profile of the base }
    spOperationDefinition_Code, {@enum.value "code" spOperationDefinition_Code Name used to invoke the operation }
    spOperationDefinition_Date, {@enum.value "date" spOperationDefinition_Date Date for this version of the operation definition }
    spOperationDefinition_Instance, {@enum.value "instance" spOperationDefinition_Instance Invoke on an instance? }
    spOperationDefinition_Kind, {@enum.value "kind" spOperationDefinition_Kind operation | query }
    spOperationDefinition_Name, {@enum.value "name" spOperationDefinition_Name Informal name for this operation }
    spOperationDefinition_Profile, {@enum.value "profile" spOperationDefinition_Profile Profile on the type }
    spOperationDefinition_Publisher, {@enum.value "publisher" spOperationDefinition_Publisher Name of the publisher (Organization or individual) }
    spOperationDefinition_Status, {@enum.value "status" spOperationDefinition_Status draft | active | retired }
    spOperationDefinition_System, {@enum.value "system" spOperationDefinition_System Invoke at the system level? }
    spOperationDefinition_Type, {@enum.value "type" spOperationDefinition_Type Invoke at resource level for these type }
    spOperationDefinition_Url, {@enum.value "url" spOperationDefinition_Url Logical URL to reference this operation definition }
    spOperationDefinition_Version); {@enum.value "version" spOperationDefinition_Version Logical id for this version of the operation definition }

  {@Enum TSearchParamsOperationOutcome
    Search Parameters for OperationOutcome
  }
  TSearchParamsOperationOutcome = (
    spOperationOutcome__content, {@enum.value "_content" spOperationOutcome__content Search on the entire content of the resource }
    spOperationOutcome__id, {@enum.value "_id" spOperationOutcome__id Logical id of this artifact }
    spOperationOutcome__lastUpdated, {@enum.value "_lastUpdated" spOperationOutcome__lastUpdated When the resource version last changed }
    spOperationOutcome__profile, {@enum.value "_profile" spOperationOutcome__profile Profiles this resource claims to conform to }
    spOperationOutcome__query, {@enum.value "_query" spOperationOutcome__query A custom search profile that describes a specific defined query operation }
    spOperationOutcome__security, {@enum.value "_security" spOperationOutcome__security Security Labels applied to this resource }
    spOperationOutcome__tag, {@enum.value "_tag" spOperationOutcome__tag Tags applied to this resource }
    spOperationOutcome__text); {@enum.value "_text" spOperationOutcome__text Search on the narrative of the resource }

  {@Enum TSearchParamsOrder
    Search Parameters for Order
  }
  TSearchParamsOrder = (
    spOrder__content, {@enum.value "_content" spOrder__content Search on the entire content of the resource }
    spOrder__id, {@enum.value "_id" spOrder__id Logical id of this artifact }
    spOrder__lastUpdated, {@enum.value "_lastUpdated" spOrder__lastUpdated When the resource version last changed }
    spOrder__profile, {@enum.value "_profile" spOrder__profile Profiles this resource claims to conform to }
    spOrder__query, {@enum.value "_query" spOrder__query A custom search profile that describes a specific defined query operation }
    spOrder__security, {@enum.value "_security" spOrder__security Security Labels applied to this resource }
    spOrder__tag, {@enum.value "_tag" spOrder__tag Tags applied to this resource }
    spOrder__text, {@enum.value "_text" spOrder__text Search on the narrative of the resource }
    spOrder_Date, {@enum.value "date" spOrder_Date When the order was made }
    spOrder_Detail, {@enum.value "detail" spOrder_Detail What action is being ordered }
    spOrder_Identifier, {@enum.value "identifier" spOrder_Identifier Instance id from source, target, and/or  others }
    spOrder_Patient, {@enum.value "patient" spOrder_Patient Patient this order is about }
    spOrder_Source, {@enum.value "source" spOrder_Source Who initiated the order }
    spOrder_Subject, {@enum.value "subject" spOrder_Subject Patient this order is about }
    spOrder_Target, {@enum.value "target" spOrder_Target Who is intended to fulfill the order }
    spOrder_When, {@enum.value "when" spOrder_When A formal schedule }
    spOrder_When_code); {@enum.value "when_code" spOrder_When_code Code specifies when request should be done. The code may simply be a priority code }

  {@Enum TSearchParamsOrderResponse
    Search Parameters for OrderResponse
  }
  TSearchParamsOrderResponse = (
    spOrderResponse__content, {@enum.value "_content" spOrderResponse__content Search on the entire content of the resource }
    spOrderResponse__id, {@enum.value "_id" spOrderResponse__id Logical id of this artifact }
    spOrderResponse__lastUpdated, {@enum.value "_lastUpdated" spOrderResponse__lastUpdated When the resource version last changed }
    spOrderResponse__profile, {@enum.value "_profile" spOrderResponse__profile Profiles this resource claims to conform to }
    spOrderResponse__query, {@enum.value "_query" spOrderResponse__query A custom search profile that describes a specific defined query operation }
    spOrderResponse__security, {@enum.value "_security" spOrderResponse__security Security Labels applied to this resource }
    spOrderResponse__tag, {@enum.value "_tag" spOrderResponse__tag Tags applied to this resource }
    spOrderResponse__text, {@enum.value "_text" spOrderResponse__text Search on the narrative of the resource }
    spOrderResponse_Code, {@enum.value "code" spOrderResponse_Code pending | review | rejected | error | accepted | cancelled | replaced | aborted | completed }
    spOrderResponse_Date, {@enum.value "date" spOrderResponse_Date When the response was made }
    spOrderResponse_Fulfillment, {@enum.value "fulfillment" spOrderResponse_Fulfillment Details of the outcome of performing the order }
    spOrderResponse_Identifier, {@enum.value "identifier" spOrderResponse_Identifier Identifiers assigned to this order by the orderer or by the receiver }
    spOrderResponse_Request, {@enum.value "request" spOrderResponse_Request The order that this is a response to }
    spOrderResponse_Who); {@enum.value "who" spOrderResponse_Who Who made the response }

  {@Enum TSearchParamsOrganization
    Search Parameters for Organization
  }
  TSearchParamsOrganization = (
    spOrganization__content, {@enum.value "_content" spOrganization__content Search on the entire content of the resource }
    spOrganization__id, {@enum.value "_id" spOrganization__id Logical id of this artifact }
    spOrganization__lastUpdated, {@enum.value "_lastUpdated" spOrganization__lastUpdated When the resource version last changed }
    spOrganization__profile, {@enum.value "_profile" spOrganization__profile Profiles this resource claims to conform to }
    spOrganization__query, {@enum.value "_query" spOrganization__query A custom search profile that describes a specific defined query operation }
    spOrganization__security, {@enum.value "_security" spOrganization__security Security Labels applied to this resource }
    spOrganization__tag, {@enum.value "_tag" spOrganization__tag Tags applied to this resource }
    spOrganization__text, {@enum.value "_text" spOrganization__text Search on the narrative of the resource }
    spOrganization_Active, {@enum.value "active" spOrganization_Active Whether the organization's record is active }
    spOrganization_Address, {@enum.value "address" spOrganization_Address A (part of the) address of the Organization }
    spOrganization_Address_city, {@enum.value "address-city" spOrganization_Address_city A city specified in an address }
    spOrganization_Address_country, {@enum.value "address-country" spOrganization_Address_country A country specified in an address }
    spOrganization_Address_postalcode, {@enum.value "address-postalcode" spOrganization_Address_postalcode A postal code specified in an address }
    spOrganization_Address_state, {@enum.value "address-state" spOrganization_Address_state A state specified in an address }
    spOrganization_Address_use, {@enum.value "address-use" spOrganization_Address_use A use code specified in an address }
    spOrganization_Identifier, {@enum.value "identifier" spOrganization_Identifier Any identifier for the organization (not the accreditation issuer's identifier) }
    spOrganization_Name, {@enum.value "name" spOrganization_Name A portion of the organization's name }
    spOrganization_Partof, {@enum.value "partof" spOrganization_Partof Search all organizations that are part of the given organization }
    spOrganization_Phonetic, {@enum.value "phonetic" spOrganization_Phonetic A portion of the organization's name using some kind of phonetic matching algorithm }
    spOrganization_Type); {@enum.value "type" spOrganization_Type A code for the type of organization }

  {@Enum TSearchParamsPatient
    Search Parameters for Patient
  }
  TSearchParamsPatient = (
    spPatient__content, {@enum.value "_content" spPatient__content Search on the entire content of the resource }
    spPatient__id, {@enum.value "_id" spPatient__id Logical id of this artifact }
    spPatient__lastUpdated, {@enum.value "_lastUpdated" spPatient__lastUpdated When the resource version last changed }
    spPatient__profile, {@enum.value "_profile" spPatient__profile Profiles this resource claims to conform to }
    spPatient__query, {@enum.value "_query" spPatient__query A custom search profile that describes a specific defined query operation }
    spPatient__security, {@enum.value "_security" spPatient__security Security Labels applied to this resource }
    spPatient__tag, {@enum.value "_tag" spPatient__tag Tags applied to this resource }
    spPatient__text, {@enum.value "_text" spPatient__text Search on the narrative of the resource }
    spPatient_Active, {@enum.value "active" spPatient_Active Whether the patient record is active }
    spPatient_Address, {@enum.value "address" spPatient_Address An address in any kind of address/part of the patient }
    spPatient_Address_city, {@enum.value "address-city" spPatient_Address_city A city specified in an address }
    spPatient_Address_country, {@enum.value "address-country" spPatient_Address_country A country specified in an address }
    spPatient_Address_postalcode, {@enum.value "address-postalcode" spPatient_Address_postalcode A postalCode specified in an address }
    spPatient_Address_state, {@enum.value "address-state" spPatient_Address_state A state specified in an address }
    spPatient_Address_use, {@enum.value "address-use" spPatient_Address_use A use code specified in an address }
    spPatient_Animal_breed, {@enum.value "animal-breed" spPatient_Animal_breed The breed for animal patients }
    spPatient_Animal_species, {@enum.value "animal-species" spPatient_Animal_species The species for animal patients }
    spPatient_Birthdate, {@enum.value "birthdate" spPatient_Birthdate The patient's date of birth }
    spPatient_Careprovider, {@enum.value "careprovider" spPatient_Careprovider Patient's nominated care provider, could be a care manager, not the organization that manages the record }
    spPatient_Deathdate, {@enum.value "deathdate" spPatient_Deathdate The date of death has been provided and satisfies this search value }
    spPatient_Deceased, {@enum.value "deceased" spPatient_Deceased This patient has been marked as deceased, or as a death date entered }
    spPatient_Email, {@enum.value "email" spPatient_Email A value in an email contact }
    spPatient_Family, {@enum.value "family" spPatient_Family A portion of the family name of the patient }
    spPatient_Gender, {@enum.value "gender" spPatient_Gender Gender of the patient }
    spPatient_Given, {@enum.value "given" spPatient_Given A portion of the given name of the patient }
    spPatient_Identifier, {@enum.value "identifier" spPatient_Identifier A patient identifier }
    spPatient_Language, {@enum.value "language" spPatient_Language Language code (irrespective of use value) }
    spPatient_Link, {@enum.value "link" spPatient_Link All patients linked to the given patient }
    spPatient_Name, {@enum.value "name" spPatient_Name A portion of either family or given name of the patient }
    spPatient_Organization, {@enum.value "organization" spPatient_Organization The organization at which this person is a patient }
    spPatient_Phone, {@enum.value "phone" spPatient_Phone A value in a phone contact }
    spPatient_Phonetic, {@enum.value "phonetic" spPatient_Phonetic A portion of either family or given name using some kind of phonetic matching algorithm }
    spPatient_Telecom); {@enum.value "telecom" spPatient_Telecom The value in any kind of telecom details of the patient }

  {@Enum TSearchParamsPaymentNotice
    Search Parameters for PaymentNotice
  }
  TSearchParamsPaymentNotice = (
    spPaymentNotice__content, {@enum.value "_content" spPaymentNotice__content Search on the entire content of the resource }
    spPaymentNotice__id, {@enum.value "_id" spPaymentNotice__id Logical id of this artifact }
    spPaymentNotice__lastUpdated, {@enum.value "_lastUpdated" spPaymentNotice__lastUpdated When the resource version last changed }
    spPaymentNotice__profile, {@enum.value "_profile" spPaymentNotice__profile Profiles this resource claims to conform to }
    spPaymentNotice__query, {@enum.value "_query" spPaymentNotice__query A custom search profile that describes a specific defined query operation }
    spPaymentNotice__security, {@enum.value "_security" spPaymentNotice__security Security Labels applied to this resource }
    spPaymentNotice__tag, {@enum.value "_tag" spPaymentNotice__tag Tags applied to this resource }
    spPaymentNotice__text, {@enum.value "_text" spPaymentNotice__text Search on the narrative of the resource }
    spPaymentNotice_Identifier); {@enum.value "identifier" spPaymentNotice_Identifier The business identifier of the Eligibility }

  {@Enum TSearchParamsPaymentReconciliation
    Search Parameters for PaymentReconciliation
  }
  TSearchParamsPaymentReconciliation = (
    spPaymentReconciliation__content, {@enum.value "_content" spPaymentReconciliation__content Search on the entire content of the resource }
    spPaymentReconciliation__id, {@enum.value "_id" spPaymentReconciliation__id Logical id of this artifact }
    spPaymentReconciliation__lastUpdated, {@enum.value "_lastUpdated" spPaymentReconciliation__lastUpdated When the resource version last changed }
    spPaymentReconciliation__profile, {@enum.value "_profile" spPaymentReconciliation__profile Profiles this resource claims to conform to }
    spPaymentReconciliation__query, {@enum.value "_query" spPaymentReconciliation__query A custom search profile that describes a specific defined query operation }
    spPaymentReconciliation__security, {@enum.value "_security" spPaymentReconciliation__security Security Labels applied to this resource }
    spPaymentReconciliation__tag, {@enum.value "_tag" spPaymentReconciliation__tag Tags applied to this resource }
    spPaymentReconciliation__text, {@enum.value "_text" spPaymentReconciliation__text Search on the narrative of the resource }
    spPaymentReconciliation_Identifier); {@enum.value "identifier" spPaymentReconciliation_Identifier The business identifier of the Explanation of Benefit }

  {@Enum TSearchParamsPerson
    Search Parameters for Person
  }
  TSearchParamsPerson = (
    spPerson__content, {@enum.value "_content" spPerson__content Search on the entire content of the resource }
    spPerson__id, {@enum.value "_id" spPerson__id Logical id of this artifact }
    spPerson__lastUpdated, {@enum.value "_lastUpdated" spPerson__lastUpdated When the resource version last changed }
    spPerson__profile, {@enum.value "_profile" spPerson__profile Profiles this resource claims to conform to }
    spPerson__query, {@enum.value "_query" spPerson__query A custom search profile that describes a specific defined query operation }
    spPerson__security, {@enum.value "_security" spPerson__security Security Labels applied to this resource }
    spPerson__tag, {@enum.value "_tag" spPerson__tag Tags applied to this resource }
    spPerson__text, {@enum.value "_text" spPerson__text Search on the narrative of the resource }
    spPerson_Address, {@enum.value "address" spPerson_Address An address in any kind of address/part }
    spPerson_Address_city, {@enum.value "address-city" spPerson_Address_city A city specified in an address }
    spPerson_Address_country, {@enum.value "address-country" spPerson_Address_country A country specified in an address }
    spPerson_Address_postalcode, {@enum.value "address-postalcode" spPerson_Address_postalcode A postal code specified in an address }
    spPerson_Address_state, {@enum.value "address-state" spPerson_Address_state A state specified in an address }
    spPerson_Address_use, {@enum.value "address-use" spPerson_Address_use A use code specified in an address }
    spPerson_Birthdate, {@enum.value "birthdate" spPerson_Birthdate The person's date of birth }
    spPerson_Email, {@enum.value "email" spPerson_Email A value in an email contact }
    spPerson_Gender, {@enum.value "gender" spPerson_Gender The gender of the person }
    spPerson_Identifier, {@enum.value "identifier" spPerson_Identifier A person Identifier }
    spPerson_Link, {@enum.value "link" spPerson_Link Any link has this Patient, Person, RelatedPerson or Practitioner reference }
    spPerson_Name, {@enum.value "name" spPerson_Name A portion of name in any name part }
    spPerson_Organization, {@enum.value "organization" spPerson_Organization The organization at which this person record is being managed }
    spPerson_Patient, {@enum.value "patient" spPerson_Patient The Person links to this Patient }
    spPerson_Phone, {@enum.value "phone" spPerson_Phone A value in a phone contact }
    spPerson_Phonetic, {@enum.value "phonetic" spPerson_Phonetic A portion of name using some kind of phonetic matching algorithm }
    spPerson_Practitioner, {@enum.value "practitioner" spPerson_Practitioner The Person links to this Practitioner }
    spPerson_Relatedperson, {@enum.value "relatedperson" spPerson_Relatedperson The Person links to this RelatedPerson }
    spPerson_Telecom); {@enum.value "telecom" spPerson_Telecom The value in any kind of contact }

  {@Enum TSearchParamsPractitioner
    Search Parameters for Practitioner
  }
  TSearchParamsPractitioner = (
    spPractitioner__content, {@enum.value "_content" spPractitioner__content Search on the entire content of the resource }
    spPractitioner__id, {@enum.value "_id" spPractitioner__id Logical id of this artifact }
    spPractitioner__lastUpdated, {@enum.value "_lastUpdated" spPractitioner__lastUpdated When the resource version last changed }
    spPractitioner__profile, {@enum.value "_profile" spPractitioner__profile Profiles this resource claims to conform to }
    spPractitioner__query, {@enum.value "_query" spPractitioner__query A custom search profile that describes a specific defined query operation }
    spPractitioner__security, {@enum.value "_security" spPractitioner__security Security Labels applied to this resource }
    spPractitioner__tag, {@enum.value "_tag" spPractitioner__tag Tags applied to this resource }
    spPractitioner__text, {@enum.value "_text" spPractitioner__text Search on the narrative of the resource }
    spPractitioner_Address, {@enum.value "address" spPractitioner_Address An address in any kind of address/part }
    spPractitioner_Address_city, {@enum.value "address-city" spPractitioner_Address_city A city specified in an address }
    spPractitioner_Address_country, {@enum.value "address-country" spPractitioner_Address_country A country specified in an address }
    spPractitioner_Address_postalcode, {@enum.value "address-postalcode" spPractitioner_Address_postalcode A postalCode specified in an address }
    spPractitioner_Address_state, {@enum.value "address-state" spPractitioner_Address_state A state specified in an address }
    spPractitioner_Address_use, {@enum.value "address-use" spPractitioner_Address_use A use code specified in an address }
    spPractitioner_Communication, {@enum.value "communication" spPractitioner_Communication One of the languages that the practitioner can communicate with }
    spPractitioner_Email, {@enum.value "email" spPractitioner_Email A value in an email contact }
    spPractitioner_Family, {@enum.value "family" spPractitioner_Family A portion of the family name }
    spPractitioner_Gender, {@enum.value "gender" spPractitioner_Gender Gender of the practitioner }
    spPractitioner_Given, {@enum.value "given" spPractitioner_Given A portion of the given name }
    spPractitioner_Identifier, {@enum.value "identifier" spPractitioner_Identifier A practitioner's Identifier }
    spPractitioner_Location, {@enum.value "location" spPractitioner_Location One of the locations at which this practitioner provides care }
    spPractitioner_Name, {@enum.value "name" spPractitioner_Name A portion of either family or given name }
    spPractitioner_Organization, {@enum.value "organization" spPractitioner_Organization The identity of the organization the practitioner represents / acts on behalf of }
    spPractitioner_Phone, {@enum.value "phone" spPractitioner_Phone A value in a phone contact }
    spPractitioner_Phonetic, {@enum.value "phonetic" spPractitioner_Phonetic A portion of either family or given name using some kind of phonetic matching algorithm }
    spPractitioner_Role, {@enum.value "role" spPractitioner_Role The practitioner can perform this role at for the organization }
    spPractitioner_Specialty, {@enum.value "specialty" spPractitioner_Specialty The practitioner has this specialty at an organization }
    spPractitioner_Telecom); {@enum.value "telecom" spPractitioner_Telecom The value in any kind of contact }

  {@Enum TSearchParamsProcedure
    Search Parameters for Procedure
  }
  TSearchParamsProcedure = (
    spProcedure__content, {@enum.value "_content" spProcedure__content Search on the entire content of the resource }
    spProcedure__id, {@enum.value "_id" spProcedure__id Logical id of this artifact }
    spProcedure__lastUpdated, {@enum.value "_lastUpdated" spProcedure__lastUpdated When the resource version last changed }
    spProcedure__profile, {@enum.value "_profile" spProcedure__profile Profiles this resource claims to conform to }
    spProcedure__query, {@enum.value "_query" spProcedure__query A custom search profile that describes a specific defined query operation }
    spProcedure__security, {@enum.value "_security" spProcedure__security Security Labels applied to this resource }
    spProcedure__tag, {@enum.value "_tag" spProcedure__tag Tags applied to this resource }
    spProcedure__text, {@enum.value "_text" spProcedure__text Search on the narrative of the resource }
    spProcedure_Code, {@enum.value "code" spProcedure_Code A code to identify a  procedure }
    spProcedure_Date, {@enum.value "date" spProcedure_Date Date/Period the procedure was performed }
    spProcedure_Encounter, {@enum.value "encounter" spProcedure_Encounter The encounter associated with the procedure }
    spProcedure_Identifier, {@enum.value "identifier" spProcedure_Identifier A unique identifier for a procedure }
    spProcedure_Location, {@enum.value "location" spProcedure_Location Where the procedure happened }
    spProcedure_Patient, {@enum.value "patient" spProcedure_Patient Search by subject - a patient }
    spProcedure_Performer, {@enum.value "performer" spProcedure_Performer The reference to the practitioner }
    spProcedure_Subject); {@enum.value "subject" spProcedure_Subject Search by subject }

  {@Enum TSearchParamsProcedureRequest
    Search Parameters for ProcedureRequest
  }
  TSearchParamsProcedureRequest = (
    spProcedureRequest__content, {@enum.value "_content" spProcedureRequest__content Search on the entire content of the resource }
    spProcedureRequest__id, {@enum.value "_id" spProcedureRequest__id Logical id of this artifact }
    spProcedureRequest__lastUpdated, {@enum.value "_lastUpdated" spProcedureRequest__lastUpdated When the resource version last changed }
    spProcedureRequest__profile, {@enum.value "_profile" spProcedureRequest__profile Profiles this resource claims to conform to }
    spProcedureRequest__query, {@enum.value "_query" spProcedureRequest__query A custom search profile that describes a specific defined query operation }
    spProcedureRequest__security, {@enum.value "_security" spProcedureRequest__security Security Labels applied to this resource }
    spProcedureRequest__tag, {@enum.value "_tag" spProcedureRequest__tag Tags applied to this resource }
    spProcedureRequest__text, {@enum.value "_text" spProcedureRequest__text Search on the narrative of the resource }
    spProcedureRequest_Encounter, {@enum.value "encounter" spProcedureRequest_Encounter Encounter request created during }
    spProcedureRequest_Identifier, {@enum.value "identifier" spProcedureRequest_Identifier A unique identifier of the Procedure Request }
    spProcedureRequest_Orderer, {@enum.value "orderer" spProcedureRequest_Orderer Who made request }
    spProcedureRequest_Patient, {@enum.value "patient" spProcedureRequest_Patient Search by subject - a patient }
    spProcedureRequest_Performer, {@enum.value "performer" spProcedureRequest_Performer Who should perform the procedure }
    spProcedureRequest_Subject); {@enum.value "subject" spProcedureRequest_Subject Search by subject }

  {@Enum TSearchParamsProcessRequest
    Search Parameters for ProcessRequest
  }
  TSearchParamsProcessRequest = (
    spProcessRequest__content, {@enum.value "_content" spProcessRequest__content Search on the entire content of the resource }
    spProcessRequest__id, {@enum.value "_id" spProcessRequest__id Logical id of this artifact }
    spProcessRequest__lastUpdated, {@enum.value "_lastUpdated" spProcessRequest__lastUpdated When the resource version last changed }
    spProcessRequest__profile, {@enum.value "_profile" spProcessRequest__profile Profiles this resource claims to conform to }
    spProcessRequest__query, {@enum.value "_query" spProcessRequest__query A custom search profile that describes a specific defined query operation }
    spProcessRequest__security, {@enum.value "_security" spProcessRequest__security Security Labels applied to this resource }
    spProcessRequest__tag, {@enum.value "_tag" spProcessRequest__tag Tags applied to this resource }
    spProcessRequest__text, {@enum.value "_text" spProcessRequest__text Search on the narrative of the resource }
    spProcessRequest_Action, {@enum.value "action" spProcessRequest_Action The action requested by this resource }
    spProcessRequest_Identifier, {@enum.value "identifier" spProcessRequest_Identifier The business identifier of the ProcessRequest }
    spProcessRequest_Organization, {@enum.value "organization" spProcessRequest_Organization The organization who generated this request }
    spProcessRequest_Provider); {@enum.value "provider" spProcessRequest_Provider The provider who regenerated this request }

  {@Enum TSearchParamsProcessResponse
    Search Parameters for ProcessResponse
  }
  TSearchParamsProcessResponse = (
    spProcessResponse__content, {@enum.value "_content" spProcessResponse__content Search on the entire content of the resource }
    spProcessResponse__id, {@enum.value "_id" spProcessResponse__id Logical id of this artifact }
    spProcessResponse__lastUpdated, {@enum.value "_lastUpdated" spProcessResponse__lastUpdated When the resource version last changed }
    spProcessResponse__profile, {@enum.value "_profile" spProcessResponse__profile Profiles this resource claims to conform to }
    spProcessResponse__query, {@enum.value "_query" spProcessResponse__query A custom search profile that describes a specific defined query operation }
    spProcessResponse__security, {@enum.value "_security" spProcessResponse__security Security Labels applied to this resource }
    spProcessResponse__tag, {@enum.value "_tag" spProcessResponse__tag Tags applied to this resource }
    spProcessResponse__text, {@enum.value "_text" spProcessResponse__text Search on the narrative of the resource }
    spProcessResponse_Identifier, {@enum.value "identifier" spProcessResponse_Identifier The business identifier of the Explanation of Benefit }
    spProcessResponse_Organization, {@enum.value "organization" spProcessResponse_Organization The organization who generated this resource }
    spProcessResponse_Request, {@enum.value "request" spProcessResponse_Request The reference to the claim }
    spProcessResponse_Requestorganization, {@enum.value "requestorganization" spProcessResponse_Requestorganization The Organization who is responsible the request transaction }
    spProcessResponse_Requestprovider); {@enum.value "requestprovider" spProcessResponse_Requestprovider The Provider who is responsible the request transaction }

  {@Enum TSearchParamsProvenance
    Search Parameters for Provenance
  }
  TSearchParamsProvenance = (
    spProvenance__content, {@enum.value "_content" spProvenance__content Search on the entire content of the resource }
    spProvenance__id, {@enum.value "_id" spProvenance__id Logical id of this artifact }
    spProvenance__lastUpdated, {@enum.value "_lastUpdated" spProvenance__lastUpdated When the resource version last changed }
    spProvenance__profile, {@enum.value "_profile" spProvenance__profile Profiles this resource claims to conform to }
    spProvenance__query, {@enum.value "_query" spProvenance__query A custom search profile that describes a specific defined query operation }
    spProvenance__security, {@enum.value "_security" spProvenance__security Security Labels applied to this resource }
    spProvenance__tag, {@enum.value "_tag" spProvenance__tag Tags applied to this resource }
    spProvenance__text, {@enum.value "_text" spProvenance__text Search on the narrative of the resource }
    spProvenance_Agent, {@enum.value "agent" spProvenance_Agent Individual, device or organization playing role }
    spProvenance_End, {@enum.value "end" spProvenance_End End time with inclusive boundary, if not ongoing }
    spProvenance_Entity, {@enum.value "entity" spProvenance_Entity Identity of entity }
    spProvenance_Entitytype, {@enum.value "entitytype" spProvenance_Entitytype The type of resource in this entity }
    spProvenance_Location, {@enum.value "location" spProvenance_Location Where the activity occurred, if relevant }
    spProvenance_Patient, {@enum.value "patient" spProvenance_Patient Target Reference(s) (usually version specific) }
    spProvenance_Sigtype, {@enum.value "sigtype" spProvenance_Sigtype Indication of the reason the entity signed the object(s) }
    spProvenance_Start, {@enum.value "start" spProvenance_Start Starting time with inclusive boundary }
    spProvenance_Target, {@enum.value "target" spProvenance_Target Target Reference(s) (usually version specific) }
    spProvenance_Userid); {@enum.value "userid" spProvenance_Userid Authorization-system identifier for the agent }

  {@Enum TSearchParamsQuestionnaire
    Search Parameters for Questionnaire
  }
  TSearchParamsQuestionnaire = (
    spQuestionnaire__content, {@enum.value "_content" spQuestionnaire__content Search on the entire content of the resource }
    spQuestionnaire__id, {@enum.value "_id" spQuestionnaire__id Logical id of this artifact }
    spQuestionnaire__lastUpdated, {@enum.value "_lastUpdated" spQuestionnaire__lastUpdated When the resource version last changed }
    spQuestionnaire__profile, {@enum.value "_profile" spQuestionnaire__profile Profiles this resource claims to conform to }
    spQuestionnaire__query, {@enum.value "_query" spQuestionnaire__query A custom search profile that describes a specific defined query operation }
    spQuestionnaire__security, {@enum.value "_security" spQuestionnaire__security Security Labels applied to this resource }
    spQuestionnaire__tag, {@enum.value "_tag" spQuestionnaire__tag Tags applied to this resource }
    spQuestionnaire__text, {@enum.value "_text" spQuestionnaire__text Search on the narrative of the resource }
    spQuestionnaire_Code, {@enum.value "code" spQuestionnaire_Code A code that corresponds to the questionnaire or one of its groups }
    spQuestionnaire_Date, {@enum.value "date" spQuestionnaire_Date When the questionnaire was last changed }
    spQuestionnaire_Identifier, {@enum.value "identifier" spQuestionnaire_Identifier An identifier for the questionnaire }
    spQuestionnaire_Publisher, {@enum.value "publisher" spQuestionnaire_Publisher The author of the questionnaire }
    spQuestionnaire_Status, {@enum.value "status" spQuestionnaire_Status The status of the questionnaire }
    spQuestionnaire_Title, {@enum.value "title" spQuestionnaire_Title All or part of the name of the questionnaire (title for the root group of the questionnaire) }
    spQuestionnaire_Version); {@enum.value "version" spQuestionnaire_Version The business version of the questionnaire }

  {@Enum TSearchParamsQuestionnaireResponse
    Search Parameters for QuestionnaireResponse
  }
  TSearchParamsQuestionnaireResponse = (
    spQuestionnaireResponse__content, {@enum.value "_content" spQuestionnaireResponse__content Search on the entire content of the resource }
    spQuestionnaireResponse__id, {@enum.value "_id" spQuestionnaireResponse__id Logical id of this artifact }
    spQuestionnaireResponse__lastUpdated, {@enum.value "_lastUpdated" spQuestionnaireResponse__lastUpdated When the resource version last changed }
    spQuestionnaireResponse__profile, {@enum.value "_profile" spQuestionnaireResponse__profile Profiles this resource claims to conform to }
    spQuestionnaireResponse__query, {@enum.value "_query" spQuestionnaireResponse__query A custom search profile that describes a specific defined query operation }
    spQuestionnaireResponse__security, {@enum.value "_security" spQuestionnaireResponse__security Security Labels applied to this resource }
    spQuestionnaireResponse__tag, {@enum.value "_tag" spQuestionnaireResponse__tag Tags applied to this resource }
    spQuestionnaireResponse__text, {@enum.value "_text" spQuestionnaireResponse__text Search on the narrative of the resource }
    spQuestionnaireResponse_Author, {@enum.value "author" spQuestionnaireResponse_Author The author of the questionnaire }
    spQuestionnaireResponse_Authored, {@enum.value "authored" spQuestionnaireResponse_Authored When the questionnaire was authored }
    spQuestionnaireResponse_Encounter, {@enum.value "encounter" spQuestionnaireResponse_Encounter Encounter during which questionnaire was authored }
    spQuestionnaireResponse_Patient, {@enum.value "patient" spQuestionnaireResponse_Patient The patient that is the subject of the questionnaire }
    spQuestionnaireResponse_Questionnaire, {@enum.value "questionnaire" spQuestionnaireResponse_Questionnaire The questionnaire the answers are provided for }
    spQuestionnaireResponse_Source, {@enum.value "source" spQuestionnaireResponse_Source The person who answered the questions }
    spQuestionnaireResponse_Status, {@enum.value "status" spQuestionnaireResponse_Status The status of the questionnaire response }
    spQuestionnaireResponse_Subject); {@enum.value "subject" spQuestionnaireResponse_Subject The subject of the questionnaire }

  {@Enum TSearchParamsReferralRequest
    Search Parameters for ReferralRequest
  }
  TSearchParamsReferralRequest = (
    spReferralRequest__content, {@enum.value "_content" spReferralRequest__content Search on the entire content of the resource }
    spReferralRequest__id, {@enum.value "_id" spReferralRequest__id Logical id of this artifact }
    spReferralRequest__lastUpdated, {@enum.value "_lastUpdated" spReferralRequest__lastUpdated When the resource version last changed }
    spReferralRequest__profile, {@enum.value "_profile" spReferralRequest__profile Profiles this resource claims to conform to }
    spReferralRequest__query, {@enum.value "_query" spReferralRequest__query A custom search profile that describes a specific defined query operation }
    spReferralRequest__security, {@enum.value "_security" spReferralRequest__security Security Labels applied to this resource }
    spReferralRequest__tag, {@enum.value "_tag" spReferralRequest__tag Tags applied to this resource }
    spReferralRequest__text, {@enum.value "_text" spReferralRequest__text Search on the narrative of the resource }
    spReferralRequest_Date, {@enum.value "date" spReferralRequest_Date Creation or activation date }
    spReferralRequest_Patient, {@enum.value "patient" spReferralRequest_Patient Who the referral is about }
    spReferralRequest_Priority, {@enum.value "priority" spReferralRequest_Priority The priority assigned to the referral }
    spReferralRequest_Recipient, {@enum.value "recipient" spReferralRequest_Recipient The person that the referral was sent to }
    spReferralRequest_Requester, {@enum.value "requester" spReferralRequest_Requester Requester of referral / transfer of care }
    spReferralRequest_Specialty, {@enum.value "specialty" spReferralRequest_Specialty The specialty that the referral is for }
    spReferralRequest_Status, {@enum.value "status" spReferralRequest_Status The status of the referral }
    spReferralRequest_Type); {@enum.value "type" spReferralRequest_Type The type of the referral }

  {@Enum TSearchParamsRelatedPerson
    Search Parameters for RelatedPerson
  }
  TSearchParamsRelatedPerson = (
    spRelatedPerson__content, {@enum.value "_content" spRelatedPerson__content Search on the entire content of the resource }
    spRelatedPerson__id, {@enum.value "_id" spRelatedPerson__id Logical id of this artifact }
    spRelatedPerson__lastUpdated, {@enum.value "_lastUpdated" spRelatedPerson__lastUpdated When the resource version last changed }
    spRelatedPerson__profile, {@enum.value "_profile" spRelatedPerson__profile Profiles this resource claims to conform to }
    spRelatedPerson__query, {@enum.value "_query" spRelatedPerson__query A custom search profile that describes a specific defined query operation }
    spRelatedPerson__security, {@enum.value "_security" spRelatedPerson__security Security Labels applied to this resource }
    spRelatedPerson__tag, {@enum.value "_tag" spRelatedPerson__tag Tags applied to this resource }
    spRelatedPerson__text, {@enum.value "_text" spRelatedPerson__text Search on the narrative of the resource }
    spRelatedPerson_Address, {@enum.value "address" spRelatedPerson_Address An address in any kind of address/part }
    spRelatedPerson_Address_city, {@enum.value "address-city" spRelatedPerson_Address_city A city specified in an address }
    spRelatedPerson_Address_country, {@enum.value "address-country" spRelatedPerson_Address_country A country specified in an address }
    spRelatedPerson_Address_postalcode, {@enum.value "address-postalcode" spRelatedPerson_Address_postalcode A postal code specified in an address }
    spRelatedPerson_Address_state, {@enum.value "address-state" spRelatedPerson_Address_state A state specified in an address }
    spRelatedPerson_Address_use, {@enum.value "address-use" spRelatedPerson_Address_use A use code specified in an address }
    spRelatedPerson_Birthdate, {@enum.value "birthdate" spRelatedPerson_Birthdate The Related Person's date of birth }
    spRelatedPerson_Email, {@enum.value "email" spRelatedPerson_Email A value in an email contact }
    spRelatedPerson_Gender, {@enum.value "gender" spRelatedPerson_Gender Gender of the person }
    spRelatedPerson_Identifier, {@enum.value "identifier" spRelatedPerson_Identifier A patient Identifier }
    spRelatedPerson_Name, {@enum.value "name" spRelatedPerson_Name A portion of name in any name part }
    spRelatedPerson_Patient, {@enum.value "patient" spRelatedPerson_Patient The patient this person is related to }
    spRelatedPerson_Phone, {@enum.value "phone" spRelatedPerson_Phone A value in a phone contact }
    spRelatedPerson_Phonetic, {@enum.value "phonetic" spRelatedPerson_Phonetic A portion of name using some kind of phonetic matching algorithm }
    spRelatedPerson_Telecom); {@enum.value "telecom" spRelatedPerson_Telecom The value in any kind of contact }

  {@Enum TSearchParamsRiskAssessment
    Search Parameters for RiskAssessment
  }
  TSearchParamsRiskAssessment = (
    spRiskAssessment__content, {@enum.value "_content" spRiskAssessment__content Search on the entire content of the resource }
    spRiskAssessment__id, {@enum.value "_id" spRiskAssessment__id Logical id of this artifact }
    spRiskAssessment__lastUpdated, {@enum.value "_lastUpdated" spRiskAssessment__lastUpdated When the resource version last changed }
    spRiskAssessment__profile, {@enum.value "_profile" spRiskAssessment__profile Profiles this resource claims to conform to }
    spRiskAssessment__query, {@enum.value "_query" spRiskAssessment__query A custom search profile that describes a specific defined query operation }
    spRiskAssessment__security, {@enum.value "_security" spRiskAssessment__security Security Labels applied to this resource }
    spRiskAssessment__tag, {@enum.value "_tag" spRiskAssessment__tag Tags applied to this resource }
    spRiskAssessment__text, {@enum.value "_text" spRiskAssessment__text Search on the narrative of the resource }
    spRiskAssessment_Condition, {@enum.value "condition" spRiskAssessment_Condition Condition assessed }
    spRiskAssessment_Date, {@enum.value "date" spRiskAssessment_Date When was assessment made? }
    spRiskAssessment_Encounter, {@enum.value "encounter" spRiskAssessment_Encounter Where was assessment performed? }
    spRiskAssessment_Identifier, {@enum.value "identifier" spRiskAssessment_Identifier Unique identifier for the assessment }
    spRiskAssessment_Method, {@enum.value "method" spRiskAssessment_Method Evaluation mechanism }
    spRiskAssessment_Patient, {@enum.value "patient" spRiskAssessment_Patient Who/what does assessment apply to? }
    spRiskAssessment_Performer, {@enum.value "performer" spRiskAssessment_Performer Who did assessment? }
    spRiskAssessment_Subject); {@enum.value "subject" spRiskAssessment_Subject Who/what does assessment apply to? }

  {@Enum TSearchParamsSchedule
    Search Parameters for Schedule
  }
  TSearchParamsSchedule = (
    spSchedule__content, {@enum.value "_content" spSchedule__content Search on the entire content of the resource }
    spSchedule__id, {@enum.value "_id" spSchedule__id Logical id of this artifact }
    spSchedule__lastUpdated, {@enum.value "_lastUpdated" spSchedule__lastUpdated When the resource version last changed }
    spSchedule__profile, {@enum.value "_profile" spSchedule__profile Profiles this resource claims to conform to }
    spSchedule__query, {@enum.value "_query" spSchedule__query A custom search profile that describes a specific defined query operation }
    spSchedule__security, {@enum.value "_security" spSchedule__security Security Labels applied to this resource }
    spSchedule__tag, {@enum.value "_tag" spSchedule__tag Tags applied to this resource }
    spSchedule__text, {@enum.value "_text" spSchedule__text Search on the narrative of the resource }
    spSchedule_Actor, {@enum.value "actor" spSchedule_Actor The individual(HealthcareService, Practitioner, Location, ...) to find a Schedule for }
    spSchedule_Date, {@enum.value "date" spSchedule_Date Search for Schedule resources that have a period that contains this date specified }
    spSchedule_Identifier, {@enum.value "identifier" spSchedule_Identifier A Schedule Identifier }
    spSchedule_Type); {@enum.value "type" spSchedule_Type The type of appointments that can be booked into associated slot(s) }

  {@Enum TSearchParamsSearchParameter
    Search Parameters for SearchParameter
  }
  TSearchParamsSearchParameter = (
    spSearchParameter__content, {@enum.value "_content" spSearchParameter__content Search on the entire content of the resource }
    spSearchParameter__id, {@enum.value "_id" spSearchParameter__id Logical id of this artifact }
    spSearchParameter__lastUpdated, {@enum.value "_lastUpdated" spSearchParameter__lastUpdated When the resource version last changed }
    spSearchParameter__profile, {@enum.value "_profile" spSearchParameter__profile Profiles this resource claims to conform to }
    spSearchParameter__query, {@enum.value "_query" spSearchParameter__query A custom search profile that describes a specific defined query operation }
    spSearchParameter__security, {@enum.value "_security" spSearchParameter__security Security Labels applied to this resource }
    spSearchParameter__tag, {@enum.value "_tag" spSearchParameter__tag Tags applied to this resource }
    spSearchParameter__text, {@enum.value "_text" spSearchParameter__text Search on the narrative of the resource }
    spSearchParameter_Base, {@enum.value "base" spSearchParameter_Base The resource type this search parameter applies to }
    spSearchParameter_Code, {@enum.value "code" spSearchParameter_Code Code used in URL }
    spSearchParameter_Description, {@enum.value "description" spSearchParameter_Description Documentation for  search parameter }
    spSearchParameter_Name, {@enum.value "name" spSearchParameter_Name Informal name for this search parameter }
    spSearchParameter_Target, {@enum.value "target" spSearchParameter_Target Types of resource (if a resource reference) }
    spSearchParameter_Type, {@enum.value "type" spSearchParameter_Type number | date | string | token | reference | composite | quantity | uri }
    spSearchParameter_Url); {@enum.value "url" spSearchParameter_Url Absolute URL used to reference this search parameter }

  {@Enum TSearchParamsSlot
    Search Parameters for Slot
  }
  TSearchParamsSlot = (
    spSlot__content, {@enum.value "_content" spSlot__content Search on the entire content of the resource }
    spSlot__id, {@enum.value "_id" spSlot__id Logical id of this artifact }
    spSlot__lastUpdated, {@enum.value "_lastUpdated" spSlot__lastUpdated When the resource version last changed }
    spSlot__profile, {@enum.value "_profile" spSlot__profile Profiles this resource claims to conform to }
    spSlot__query, {@enum.value "_query" spSlot__query A custom search profile that describes a specific defined query operation }
    spSlot__security, {@enum.value "_security" spSlot__security Security Labels applied to this resource }
    spSlot__tag, {@enum.value "_tag" spSlot__tag Tags applied to this resource }
    spSlot__text, {@enum.value "_text" spSlot__text Search on the narrative of the resource }
    spSlot_Fb_type, {@enum.value "fb-type" spSlot_Fb_type The free/busy status of the appointment }
    spSlot_Identifier, {@enum.value "identifier" spSlot_Identifier A Slot Identifier }
    spSlot_Schedule, {@enum.value "schedule" spSlot_Schedule The Schedule Resource that we are seeking a slot within }
    spSlot_Slot_type, {@enum.value "slot-type" spSlot_Slot_type The type of appointments that can be booked into the slot }
    spSlot_Start); {@enum.value "start" spSlot_Start Appointment date/time. }

  {@Enum TSearchParamsSpecimen
    Search Parameters for Specimen
  }
  TSearchParamsSpecimen = (
    spSpecimen__content, {@enum.value "_content" spSpecimen__content Search on the entire content of the resource }
    spSpecimen__id, {@enum.value "_id" spSpecimen__id Logical id of this artifact }
    spSpecimen__lastUpdated, {@enum.value "_lastUpdated" spSpecimen__lastUpdated When the resource version last changed }
    spSpecimen__profile, {@enum.value "_profile" spSpecimen__profile Profiles this resource claims to conform to }
    spSpecimen__query, {@enum.value "_query" spSpecimen__query A custom search profile that describes a specific defined query operation }
    spSpecimen__security, {@enum.value "_security" spSpecimen__security Security Labels applied to this resource }
    spSpecimen__tag, {@enum.value "_tag" spSpecimen__tag Tags applied to this resource }
    spSpecimen__text, {@enum.value "_text" spSpecimen__text Search on the narrative of the resource }
    spSpecimen_Accession, {@enum.value "accession" spSpecimen_Accession The accession number associated with the specimen }
    spSpecimen_Bodysite, {@enum.value "bodysite" spSpecimen_Bodysite The code for the body site from where the specimen originated }
    spSpecimen_Collected, {@enum.value "collected" spSpecimen_Collected The date the specimen was collected }
    spSpecimen_Collector, {@enum.value "collector" spSpecimen_Collector Who collected the specimen }
    spSpecimen_Container, {@enum.value "container" spSpecimen_Container The kind of specimen container }
    spSpecimen_Container_id, {@enum.value "container-id" spSpecimen_Container_id The unique identifier associated with the specimen container }
    spSpecimen_Identifier, {@enum.value "identifier" spSpecimen_Identifier The unique identifier associated with the specimen }
    spSpecimen_Parent, {@enum.value "parent" spSpecimen_Parent The parent of the specimen }
    spSpecimen_Patient, {@enum.value "patient" spSpecimen_Patient The patient the specimen comes from }
    spSpecimen_Subject, {@enum.value "subject" spSpecimen_Subject The subject of the specimen }
    spSpecimen_Type); {@enum.value "type" spSpecimen_Type The specimen type }

  {@Enum TSearchParamsStructureDefinition
    Search Parameters for StructureDefinition
  }
  TSearchParamsStructureDefinition = (
    spStructureDefinition__content, {@enum.value "_content" spStructureDefinition__content Search on the entire content of the resource }
    spStructureDefinition__id, {@enum.value "_id" spStructureDefinition__id Logical id of this artifact }
    spStructureDefinition__lastUpdated, {@enum.value "_lastUpdated" spStructureDefinition__lastUpdated When the resource version last changed }
    spStructureDefinition__profile, {@enum.value "_profile" spStructureDefinition__profile Profiles this resource claims to conform to }
    spStructureDefinition__query, {@enum.value "_query" spStructureDefinition__query A custom search profile that describes a specific defined query operation }
    spStructureDefinition__security, {@enum.value "_security" spStructureDefinition__security Security Labels applied to this resource }
    spStructureDefinition__tag, {@enum.value "_tag" spStructureDefinition__tag Tags applied to this resource }
    spStructureDefinition__text, {@enum.value "_text" spStructureDefinition__text Search on the narrative of the resource }
    spStructureDefinition_Abstract, {@enum.value "abstract" spStructureDefinition_Abstract Whether the structure is abstract }
    spStructureDefinition_Base, {@enum.value "base" spStructureDefinition_Base Structure that this set of constraints applies to }
    spStructureDefinition_Base_path, {@enum.value "base-path" spStructureDefinition_Base_path Path that identifies the base element }
    spStructureDefinition_Code, {@enum.value "code" spStructureDefinition_Code A code for the profile }
    spStructureDefinition_Context, {@enum.value "context" spStructureDefinition_Context A use context assigned to the structure }
    spStructureDefinition_Context_type, {@enum.value "context-type" spStructureDefinition_Context_type resource | datatype | mapping | extension }
    spStructureDefinition_Date, {@enum.value "date" spStructureDefinition_Date The profile publication date }
    spStructureDefinition_Description, {@enum.value "description" spStructureDefinition_Description Text search in the description of the profile }
    spStructureDefinition_Display, {@enum.value "display" spStructureDefinition_Display Use this name when displaying the value }
    spStructureDefinition_Experimental, {@enum.value "experimental" spStructureDefinition_Experimental If for testing purposes, not real usage }
    spStructureDefinition_Ext_context, {@enum.value "ext-context" spStructureDefinition_Ext_context Where the extension can be used in instances }
    spStructureDefinition_Identifier, {@enum.value "identifier" spStructureDefinition_Identifier The identifier of the profile }
    spStructureDefinition_Kind, {@enum.value "kind" spStructureDefinition_Kind datatype | resource | logical }
    spStructureDefinition_Name, {@enum.value "name" spStructureDefinition_Name Name of the profile }
    spStructureDefinition_Path, {@enum.value "path" spStructureDefinition_Path A path that is constrained in the profile }
    spStructureDefinition_Publisher, {@enum.value "publisher" spStructureDefinition_Publisher Name of the publisher of the profile }
    spStructureDefinition_Status, {@enum.value "status" spStructureDefinition_Status The current status of the profile }
    spStructureDefinition_Type, {@enum.value "type" spStructureDefinition_Type Any datatype or resource, including abstract ones }
    spStructureDefinition_Url, {@enum.value "url" spStructureDefinition_Url Absolute URL used to reference this StructureDefinition }
    spStructureDefinition_Valueset, {@enum.value "valueset" spStructureDefinition_Valueset A vocabulary binding reference }
    spStructureDefinition_Version); {@enum.value "version" spStructureDefinition_Version The version identifier of the profile }

  {@Enum TSearchParamsSubscription
    Search Parameters for Subscription
  }
  TSearchParamsSubscription = (
    spSubscription__content, {@enum.value "_content" spSubscription__content Search on the entire content of the resource }
    spSubscription__id, {@enum.value "_id" spSubscription__id Logical id of this artifact }
    spSubscription__lastUpdated, {@enum.value "_lastUpdated" spSubscription__lastUpdated When the resource version last changed }
    spSubscription__profile, {@enum.value "_profile" spSubscription__profile Profiles this resource claims to conform to }
    spSubscription__query, {@enum.value "_query" spSubscription__query A custom search profile that describes a specific defined query operation }
    spSubscription__security, {@enum.value "_security" spSubscription__security Security Labels applied to this resource }
    spSubscription__tag, {@enum.value "_tag" spSubscription__tag Tags applied to this resource }
    spSubscription__text, {@enum.value "_text" spSubscription__text Search on the narrative of the resource }
    spSubscription_Contact, {@enum.value "contact" spSubscription_Contact Contact details for source (e.g. troubleshooting) }
    spSubscription_Criteria, {@enum.value "criteria" spSubscription_Criteria Rule for server push criteria }
    spSubscription_Payload, {@enum.value "payload" spSubscription_Payload Mimetype to send, or blank for no payload }
    spSubscription_Status, {@enum.value "status" spSubscription_Status requested | active | error | off }
    spSubscription_Tag, {@enum.value "tag" spSubscription_Tag A tag to add to matching resources }
    spSubscription_Type, {@enum.value "type" spSubscription_Type rest-hook | websocket | email | sms | message }
    spSubscription_Url); {@enum.value "url" spSubscription_Url Where the channel points to }

  {@Enum TSearchParamsSubstance
    Search Parameters for Substance
  }
  TSearchParamsSubstance = (
    spSubstance__content, {@enum.value "_content" spSubstance__content Search on the entire content of the resource }
    spSubstance__id, {@enum.value "_id" spSubstance__id Logical id of this artifact }
    spSubstance__lastUpdated, {@enum.value "_lastUpdated" spSubstance__lastUpdated When the resource version last changed }
    spSubstance__profile, {@enum.value "_profile" spSubstance__profile Profiles this resource claims to conform to }
    spSubstance__query, {@enum.value "_query" spSubstance__query A custom search profile that describes a specific defined query operation }
    spSubstance__security, {@enum.value "_security" spSubstance__security Security Labels applied to this resource }
    spSubstance__tag, {@enum.value "_tag" spSubstance__tag Tags applied to this resource }
    spSubstance__text, {@enum.value "_text" spSubstance__text Search on the narrative of the resource }
    spSubstance_Category, {@enum.value "category" spSubstance_Category The category of the substance }
    spSubstance_Code, {@enum.value "code" spSubstance_Code The code of the substance }
    spSubstance_Container_identifier, {@enum.value "container-identifier" spSubstance_Container_identifier Identifier of the package/container }
    spSubstance_Expiry, {@enum.value "expiry" spSubstance_Expiry Expiry date of package or container of substance }
    spSubstance_Identifier, {@enum.value "identifier" spSubstance_Identifier Unique identifier for the substance }
    spSubstance_Quantity, {@enum.value "quantity" spSubstance_Quantity Amount of substance in the package }
    spSubstance_Substance); {@enum.value "substance" spSubstance_Substance A component of the substance }

  {@Enum TSearchParamsSupplyDelivery
    Search Parameters for SupplyDelivery
  }
  TSearchParamsSupplyDelivery = (
    spSupplyDelivery__content, {@enum.value "_content" spSupplyDelivery__content Search on the entire content of the resource }
    spSupplyDelivery__id, {@enum.value "_id" spSupplyDelivery__id Logical id of this artifact }
    spSupplyDelivery__lastUpdated, {@enum.value "_lastUpdated" spSupplyDelivery__lastUpdated When the resource version last changed }
    spSupplyDelivery__profile, {@enum.value "_profile" spSupplyDelivery__profile Profiles this resource claims to conform to }
    spSupplyDelivery__query, {@enum.value "_query" spSupplyDelivery__query A custom search profile that describes a specific defined query operation }
    spSupplyDelivery__security, {@enum.value "_security" spSupplyDelivery__security Security Labels applied to this resource }
    spSupplyDelivery__tag, {@enum.value "_tag" spSupplyDelivery__tag Tags applied to this resource }
    spSupplyDelivery__text, {@enum.value "_text" spSupplyDelivery__text Search on the narrative of the resource }
    spSupplyDelivery_Identifier, {@enum.value "identifier" spSupplyDelivery_Identifier External identifier }
    spSupplyDelivery_Patient, {@enum.value "patient" spSupplyDelivery_Patient Patient for whom the item is supplied }
    spSupplyDelivery_Receiver, {@enum.value "receiver" spSupplyDelivery_Receiver Who collected the Supply }
    spSupplyDelivery_Status, {@enum.value "status" spSupplyDelivery_Status in-progress | completed | abandoned }
    spSupplyDelivery_Supplier); {@enum.value "supplier" spSupplyDelivery_Supplier Dispenser }

  {@Enum TSearchParamsSupplyRequest
    Search Parameters for SupplyRequest
  }
  TSearchParamsSupplyRequest = (
    spSupplyRequest__content, {@enum.value "_content" spSupplyRequest__content Search on the entire content of the resource }
    spSupplyRequest__id, {@enum.value "_id" spSupplyRequest__id Logical id of this artifact }
    spSupplyRequest__lastUpdated, {@enum.value "_lastUpdated" spSupplyRequest__lastUpdated When the resource version last changed }
    spSupplyRequest__profile, {@enum.value "_profile" spSupplyRequest__profile Profiles this resource claims to conform to }
    spSupplyRequest__query, {@enum.value "_query" spSupplyRequest__query A custom search profile that describes a specific defined query operation }
    spSupplyRequest__security, {@enum.value "_security" spSupplyRequest__security Security Labels applied to this resource }
    spSupplyRequest__tag, {@enum.value "_tag" spSupplyRequest__tag Tags applied to this resource }
    spSupplyRequest__text, {@enum.value "_text" spSupplyRequest__text Search on the narrative of the resource }
    spSupplyRequest_Date, {@enum.value "date" spSupplyRequest_Date When the request was made }
    spSupplyRequest_Identifier, {@enum.value "identifier" spSupplyRequest_Identifier Unique identifier }
    spSupplyRequest_Kind, {@enum.value "kind" spSupplyRequest_Kind The kind of supply (central, non-stock, etc.) }
    spSupplyRequest_Patient, {@enum.value "patient" spSupplyRequest_Patient Patient for whom the item is supplied }
    spSupplyRequest_Source, {@enum.value "source" spSupplyRequest_Source Who initiated this order }
    spSupplyRequest_Status, {@enum.value "status" spSupplyRequest_Status requested | completed | failed | cancelled }
    spSupplyRequest_Supplier); {@enum.value "supplier" spSupplyRequest_Supplier Who is intended to fulfill the request }

  {@Enum TSearchParamsTestScript
    Search Parameters for TestScript
  }
  TSearchParamsTestScript = (
    spTestScript__content, {@enum.value "_content" spTestScript__content Search on the entire content of the resource }
    spTestScript__id, {@enum.value "_id" spTestScript__id Logical id of this artifact }
    spTestScript__lastUpdated, {@enum.value "_lastUpdated" spTestScript__lastUpdated When the resource version last changed }
    spTestScript__profile, {@enum.value "_profile" spTestScript__profile Profiles this resource claims to conform to }
    spTestScript__query, {@enum.value "_query" spTestScript__query A custom search profile that describes a specific defined query operation }
    spTestScript__security, {@enum.value "_security" spTestScript__security Security Labels applied to this resource }
    spTestScript__tag, {@enum.value "_tag" spTestScript__tag Tags applied to this resource }
    spTestScript__text, {@enum.value "_text" spTestScript__text Search on the narrative of the resource }
    spTestScript_Description, {@enum.value "description" spTestScript_Description Natural language description of the TestScript }
    spTestScript_Identifier, {@enum.value "identifier" spTestScript_Identifier External identifier }
    spTestScript_Name, {@enum.value "name" spTestScript_Name Informal name for this TestScript }
    spTestScript_Testscript_capability, {@enum.value "testscript-capability" spTestScript_Testscript_capability TestScript required and validated capability }
    spTestScript_Testscript_setup_capability, {@enum.value "testscript-setup-capability" spTestScript_Testscript_setup_capability TestScript setup required and validated capability }
    spTestScript_Testscript_test_capability, {@enum.value "testscript-test-capability" spTestScript_Testscript_test_capability TestScript test required and validated capability }
    spTestScript_Url); {@enum.value "url" spTestScript_Url Absolute URL used to reference this TestScript }

  {@Enum TSearchParamsValueSet
    Search Parameters for ValueSet
  }
  TSearchParamsValueSet = (
    spValueSet__content, {@enum.value "_content" spValueSet__content Search on the entire content of the resource }
    spValueSet__id, {@enum.value "_id" spValueSet__id Logical id of this artifact }
    spValueSet__lastUpdated, {@enum.value "_lastUpdated" spValueSet__lastUpdated When the resource version last changed }
    spValueSet__profile, {@enum.value "_profile" spValueSet__profile Profiles this resource claims to conform to }
    spValueSet__query, {@enum.value "_query" spValueSet__query A custom search profile that describes a specific defined query operation }
    spValueSet__security, {@enum.value "_security" spValueSet__security Security Labels applied to this resource }
    spValueSet__tag, {@enum.value "_tag" spValueSet__tag Tags applied to this resource }
    spValueSet__text, {@enum.value "_text" spValueSet__text Search on the narrative of the resource }
    spValueSet_Code, {@enum.value "code" spValueSet_Code A code defined in the value set }
    spValueSet_Context, {@enum.value "context" spValueSet_Context A use context assigned to the value set }
    spValueSet_Date, {@enum.value "date" spValueSet_Date The value set publication date }
    spValueSet_Description, {@enum.value "description" spValueSet_Description Text search in the description of the value set }
    spValueSet_Expansion, {@enum.value "expansion" spValueSet_Expansion Uniquely identifies this expansion }
    spValueSet_Identifier, {@enum.value "identifier" spValueSet_Identifier The identifier for the value set }
    spValueSet_Name, {@enum.value "name" spValueSet_Name The name of the value set }
    spValueSet_Publisher, {@enum.value "publisher" spValueSet_Publisher Name of the publisher of the value set }
    spValueSet_Reference, {@enum.value "reference" spValueSet_Reference A code system included or excluded in the value set or an imported value set }
    spValueSet_Status, {@enum.value "status" spValueSet_Status The status of the value set }
    spValueSet_System, {@enum.value "system" spValueSet_System The system for any codes defined by this value set }
    spValueSet_Url, {@enum.value "url" spValueSet_Url The logical URL for the value set }
    spValueSet_Version); {@enum.value "version" spValueSet_Version The version identifier of the value set }

  {@Enum TSearchParamsVisionPrescription
    Search Parameters for VisionPrescription
  }
  TSearchParamsVisionPrescription = (
    spVisionPrescription__content, {@enum.value "_content" spVisionPrescription__content Search on the entire content of the resource }
    spVisionPrescription__id, {@enum.value "_id" spVisionPrescription__id Logical id of this artifact }
    spVisionPrescription__lastUpdated, {@enum.value "_lastUpdated" spVisionPrescription__lastUpdated When the resource version last changed }
    spVisionPrescription__profile, {@enum.value "_profile" spVisionPrescription__profile Profiles this resource claims to conform to }
    spVisionPrescription__query, {@enum.value "_query" spVisionPrescription__query A custom search profile that describes a specific defined query operation }
    spVisionPrescription__security, {@enum.value "_security" spVisionPrescription__security Security Labels applied to this resource }
    spVisionPrescription__tag, {@enum.value "_tag" spVisionPrescription__tag Tags applied to this resource }
    spVisionPrescription__text, {@enum.value "_text" spVisionPrescription__text Search on the narrative of the resource }
    spVisionPrescription_Datewritten, {@enum.value "datewritten" spVisionPrescription_Datewritten Return prescriptions written on this date }
    spVisionPrescription_Encounter, {@enum.value "encounter" spVisionPrescription_Encounter Return prescriptions with this encounter identifier }
    spVisionPrescription_Identifier, {@enum.value "identifier" spVisionPrescription_Identifier Return prescriptions with this external identifier }
    spVisionPrescription_Patient, {@enum.value "patient" spVisionPrescription_Patient The identity of a patient to list dispenses for }
    spVisionPrescription_Prescriber); {@enum.value "prescriber" spVisionPrescription_Prescriber Who authorizes the vision product }

Type
  TFhirResource = class;
  TFhirResourceList = class;
  TFhirDomainResource = class;
  TFhirDomainResourceList = class;
  TFhirParametersParameter = class;
  TFhirParametersParameterList = class;
  TFhirParameters = class;
  TFhirParametersList = class;
  TFhirAccount = class;
  TFhirAccountList = class;
  TFhirAllergyIntoleranceReaction = class;
  TFhirAllergyIntoleranceReactionList = class;
  TFhirAllergyIntolerance = class;
  TFhirAllergyIntoleranceList = class;
  TFhirAppointmentParticipant = class;
  TFhirAppointmentParticipantList = class;
  TFhirAppointment = class;
  TFhirAppointmentList = class;
  TFhirAppointmentResponse = class;
  TFhirAppointmentResponseList = class;
  TFhirAuditEventEvent = class;
  TFhirAuditEventEventList = class;
  TFhirAuditEventParticipant = class;
  TFhirAuditEventParticipantList = class;
  TFhirAuditEventParticipantNetwork = class;
  TFhirAuditEventParticipantNetworkList = class;
  TFhirAuditEventSource = class;
  TFhirAuditEventSourceList = class;
  TFhirAuditEventObject = class;
  TFhirAuditEventObjectList = class;
  TFhirAuditEventObjectDetail = class;
  TFhirAuditEventObjectDetailList = class;
  TFhirAuditEvent = class;
  TFhirAuditEventList = class;
  TFhirBasic = class;
  TFhirBasicList = class;
  TFhirBinary = class;
  TFhirBinaryList = class;
  TFhirBodySite = class;
  TFhirBodySiteList = class;
  TFhirBundleLink = class;
  TFhirBundleLinkList = class;
  TFhirBundleEntry = class;
  TFhirBundleEntryList = class;
  TFhirBundleEntrySearch = class;
  TFhirBundleEntrySearchList = class;
  TFhirBundleEntryRequest = class;
  TFhirBundleEntryRequestList = class;
  TFhirBundleEntryResponse = class;
  TFhirBundleEntryResponseList = class;
  TFhirBundle = class;
  TFhirBundleList = class;
  TFhirCarePlanRelatedPlan = class;
  TFhirCarePlanRelatedPlanList = class;
  TFhirCarePlanParticipant = class;
  TFhirCarePlanParticipantList = class;
  TFhirCarePlanActivity = class;
  TFhirCarePlanActivityList = class;
  TFhirCarePlanActivityDetail = class;
  TFhirCarePlanActivityDetailList = class;
  TFhirCarePlan = class;
  TFhirCarePlanList = class;
  TFhirClaimPayee = class;
  TFhirClaimPayeeList = class;
  TFhirClaimDiagnosis = class;
  TFhirClaimDiagnosisList = class;
  TFhirClaimCoverage = class;
  TFhirClaimCoverageList = class;
  TFhirClaimItem = class;
  TFhirClaimItemList = class;
  TFhirClaimItemDetail = class;
  TFhirClaimItemDetailList = class;
  TFhirClaimItemDetailSubDetail = class;
  TFhirClaimItemDetailSubDetailList = class;
  TFhirClaimItemProsthesis = class;
  TFhirClaimItemProsthesisList = class;
  TFhirClaimMissingTeeth = class;
  TFhirClaimMissingTeethList = class;
  TFhirClaim = class;
  TFhirClaimList = class;
  TFhirClaimResponseItem = class;
  TFhirClaimResponseItemList = class;
  TFhirClaimResponseItemAdjudication = class;
  TFhirClaimResponseItemAdjudicationList = class;
  TFhirClaimResponseItemDetail = class;
  TFhirClaimResponseItemDetailList = class;
  TFhirClaimResponseItemDetailAdjudication = class;
  TFhirClaimResponseItemDetailAdjudicationList = class;
  TFhirClaimResponseItemDetailSubDetail = class;
  TFhirClaimResponseItemDetailSubDetailList = class;
  TFhirClaimResponseItemDetailSubDetailAdjudication = class;
  TFhirClaimResponseItemDetailSubDetailAdjudicationList = class;
  TFhirClaimResponseAddItem = class;
  TFhirClaimResponseAddItemList = class;
  TFhirClaimResponseAddItemAdjudication = class;
  TFhirClaimResponseAddItemAdjudicationList = class;
  TFhirClaimResponseAddItemDetail = class;
  TFhirClaimResponseAddItemDetailList = class;
  TFhirClaimResponseAddItemDetailAdjudication = class;
  TFhirClaimResponseAddItemDetailAdjudicationList = class;
  TFhirClaimResponseError = class;
  TFhirClaimResponseErrorList = class;
  TFhirClaimResponseNote = class;
  TFhirClaimResponseNoteList = class;
  TFhirClaimResponseCoverage = class;
  TFhirClaimResponseCoverageList = class;
  TFhirClaimResponse = class;
  TFhirClaimResponseList = class;
  TFhirClinicalImpressionInvestigations = class;
  TFhirClinicalImpressionInvestigationsList = class;
  TFhirClinicalImpressionFinding = class;
  TFhirClinicalImpressionFindingList = class;
  TFhirClinicalImpressionRuledOut = class;
  TFhirClinicalImpressionRuledOutList = class;
  TFhirClinicalImpression = class;
  TFhirClinicalImpressionList = class;
  TFhirCommunicationPayload = class;
  TFhirCommunicationPayloadList = class;
  TFhirCommunication = class;
  TFhirCommunicationList = class;
  TFhirCommunicationRequestPayload = class;
  TFhirCommunicationRequestPayloadList = class;
  TFhirCommunicationRequest = class;
  TFhirCommunicationRequestList = class;
  TFhirCompositionAttester = class;
  TFhirCompositionAttesterList = class;
  TFhirCompositionEvent = class;
  TFhirCompositionEventList = class;
  TFhirCompositionSection = class;
  TFhirCompositionSectionList = class;
  TFhirComposition = class;
  TFhirCompositionList = class;
  TFhirConceptMapContact = class;
  TFhirConceptMapContactList = class;
  TFhirConceptMapElement = class;
  TFhirConceptMapElementList = class;
  TFhirConceptMapElementTarget = class;
  TFhirConceptMapElementTargetList = class;
  TFhirConceptMapElementTargetDependsOn = class;
  TFhirConceptMapElementTargetDependsOnList = class;
  TFhirConceptMap = class;
  TFhirConceptMapList = class;
  TFhirConditionStage = class;
  TFhirConditionStageList = class;
  TFhirConditionEvidence = class;
  TFhirConditionEvidenceList = class;
  TFhirCondition = class;
  TFhirConditionList = class;
  TFhirConformanceContact = class;
  TFhirConformanceContactList = class;
  TFhirConformanceSoftware = class;
  TFhirConformanceSoftwareList = class;
  TFhirConformanceImplementation = class;
  TFhirConformanceImplementationList = class;
  TFhirConformanceRest = class;
  TFhirConformanceRestList = class;
  TFhirConformanceRestSecurity = class;
  TFhirConformanceRestSecurityList = class;
  TFhirConformanceRestSecurityCertificate = class;
  TFhirConformanceRestSecurityCertificateList = class;
  TFhirConformanceRestResource = class;
  TFhirConformanceRestResourceList = class;
  TFhirConformanceRestResourceInteraction = class;
  TFhirConformanceRestResourceInteractionList = class;
  TFhirConformanceRestResourceSearchParam = class;
  TFhirConformanceRestResourceSearchParamList = class;
  TFhirConformanceRestInteraction = class;
  TFhirConformanceRestInteractionList = class;
  TFhirConformanceRestOperation = class;
  TFhirConformanceRestOperationList = class;
  TFhirConformanceMessaging = class;
  TFhirConformanceMessagingList = class;
  TFhirConformanceMessagingEndpoint = class;
  TFhirConformanceMessagingEndpointList = class;
  TFhirConformanceMessagingEvent = class;
  TFhirConformanceMessagingEventList = class;
  TFhirConformanceDocument = class;
  TFhirConformanceDocumentList = class;
  TFhirConformance = class;
  TFhirConformanceList = class;
  TFhirContractActor = class;
  TFhirContractActorList = class;
  TFhirContractValuedItem = class;
  TFhirContractValuedItemList = class;
  TFhirContractSigner = class;
  TFhirContractSignerList = class;
  TFhirContractTerm = class;
  TFhirContractTermList = class;
  TFhirContractTermActor = class;
  TFhirContractTermActorList = class;
  TFhirContractTermValuedItem = class;
  TFhirContractTermValuedItemList = class;
  TFhirContractFriendly = class;
  TFhirContractFriendlyList = class;
  TFhirContractLegal = class;
  TFhirContractLegalList = class;
  TFhirContractRule = class;
  TFhirContractRuleList = class;
  TFhirContract = class;
  TFhirContractList = class;
  TFhirCoverage = class;
  TFhirCoverageList = class;
  TFhirDataElementContact = class;
  TFhirDataElementContactList = class;
  TFhirDataElementMapping = class;
  TFhirDataElementMappingList = class;
  TFhirDataElement = class;
  TFhirDataElementList = class;
  TFhirDetectedIssueMitigation = class;
  TFhirDetectedIssueMitigationList = class;
  TFhirDetectedIssue = class;
  TFhirDetectedIssueList = class;
  TFhirDevice = class;
  TFhirDeviceList = class;
  TFhirDeviceComponentProductionSpecification = class;
  TFhirDeviceComponentProductionSpecificationList = class;
  TFhirDeviceComponent = class;
  TFhirDeviceComponentList = class;
  TFhirDeviceMetricCalibration = class;
  TFhirDeviceMetricCalibrationList = class;
  TFhirDeviceMetric = class;
  TFhirDeviceMetricList = class;
  TFhirDeviceUseRequest = class;
  TFhirDeviceUseRequestList = class;
  TFhirDeviceUseStatement = class;
  TFhirDeviceUseStatementList = class;
  TFhirDiagnosticOrderEvent = class;
  TFhirDiagnosticOrderEventList = class;
  TFhirDiagnosticOrderItem = class;
  TFhirDiagnosticOrderItemList = class;
  TFhirDiagnosticOrder = class;
  TFhirDiagnosticOrderList = class;
  TFhirDiagnosticReportImage = class;
  TFhirDiagnosticReportImageList = class;
  TFhirDiagnosticReport = class;
  TFhirDiagnosticReportList = class;
  TFhirDocumentManifestContent = class;
  TFhirDocumentManifestContentList = class;
  TFhirDocumentManifestRelated = class;
  TFhirDocumentManifestRelatedList = class;
  TFhirDocumentManifest = class;
  TFhirDocumentManifestList = class;
  TFhirDocumentReferenceRelatesTo = class;
  TFhirDocumentReferenceRelatesToList = class;
  TFhirDocumentReferenceContent = class;
  TFhirDocumentReferenceContentList = class;
  TFhirDocumentReferenceContext = class;
  TFhirDocumentReferenceContextList = class;
  TFhirDocumentReferenceContextRelated = class;
  TFhirDocumentReferenceContextRelatedList = class;
  TFhirDocumentReference = class;
  TFhirDocumentReferenceList = class;
  TFhirEligibilityRequest = class;
  TFhirEligibilityRequestList = class;
  TFhirEligibilityResponse = class;
  TFhirEligibilityResponseList = class;
  TFhirEncounterStatusHistory = class;
  TFhirEncounterStatusHistoryList = class;
  TFhirEncounterParticipant = class;
  TFhirEncounterParticipantList = class;
  TFhirEncounterHospitalization = class;
  TFhirEncounterHospitalizationList = class;
  TFhirEncounterLocation = class;
  TFhirEncounterLocationList = class;
  TFhirEncounter = class;
  TFhirEncounterList = class;
  TFhirEnrollmentRequest = class;
  TFhirEnrollmentRequestList = class;
  TFhirEnrollmentResponse = class;
  TFhirEnrollmentResponseList = class;
  TFhirEpisodeOfCareStatusHistory = class;
  TFhirEpisodeOfCareStatusHistoryList = class;
  TFhirEpisodeOfCareCareTeam = class;
  TFhirEpisodeOfCareCareTeamList = class;
  TFhirEpisodeOfCare = class;
  TFhirEpisodeOfCareList = class;
  TFhirExplanationOfBenefit = class;
  TFhirExplanationOfBenefitList = class;
  TFhirFamilyMemberHistoryCondition = class;
  TFhirFamilyMemberHistoryConditionList = class;
  TFhirFamilyMemberHistory = class;
  TFhirFamilyMemberHistoryList = class;
  TFhirFlag = class;
  TFhirFlagList = class;
  TFhirGoalOutcome = class;
  TFhirGoalOutcomeList = class;
  TFhirGoal = class;
  TFhirGoalList = class;
  TFhirGroupCharacteristic = class;
  TFhirGroupCharacteristicList = class;
  TFhirGroupMember = class;
  TFhirGroupMemberList = class;
  TFhirGroup = class;
  TFhirGroupList = class;
  TFhirHealthcareServiceServiceType = class;
  TFhirHealthcareServiceServiceTypeList = class;
  TFhirHealthcareServiceAvailableTime = class;
  TFhirHealthcareServiceAvailableTimeList = class;
  TFhirHealthcareServiceNotAvailable = class;
  TFhirHealthcareServiceNotAvailableList = class;
  TFhirHealthcareService = class;
  TFhirHealthcareServiceList = class;
  TFhirImagingObjectSelectionStudy = class;
  TFhirImagingObjectSelectionStudyList = class;
  TFhirImagingObjectSelectionStudySeries = class;
  TFhirImagingObjectSelectionStudySeriesList = class;
  TFhirImagingObjectSelectionStudySeriesInstance = class;
  TFhirImagingObjectSelectionStudySeriesInstanceList = class;
  TFhirImagingObjectSelectionStudySeriesInstanceFrames = class;
  TFhirImagingObjectSelectionStudySeriesInstanceFramesList = class;
  TFhirImagingObjectSelection = class;
  TFhirImagingObjectSelectionList = class;
  TFhirImagingStudySeries = class;
  TFhirImagingStudySeriesList = class;
  TFhirImagingStudySeriesInstance = class;
  TFhirImagingStudySeriesInstanceList = class;
  TFhirImagingStudy = class;
  TFhirImagingStudyList = class;
  TFhirImmunizationExplanation = class;
  TFhirImmunizationExplanationList = class;
  TFhirImmunizationReaction = class;
  TFhirImmunizationReactionList = class;
  TFhirImmunizationVaccinationProtocol = class;
  TFhirImmunizationVaccinationProtocolList = class;
  TFhirImmunization = class;
  TFhirImmunizationList = class;
  TFhirImmunizationRecommendationRecommendation = class;
  TFhirImmunizationRecommendationRecommendationList = class;
  TFhirImmunizationRecommendationRecommendationDateCriterion = class;
  TFhirImmunizationRecommendationRecommendationDateCriterionList = class;
  TFhirImmunizationRecommendationRecommendationProtocol = class;
  TFhirImmunizationRecommendationRecommendationProtocolList = class;
  TFhirImmunizationRecommendation = class;
  TFhirImmunizationRecommendationList = class;
  TFhirImplementationGuideContact = class;
  TFhirImplementationGuideContactList = class;
  TFhirImplementationGuideDependency = class;
  TFhirImplementationGuideDependencyList = class;
  TFhirImplementationGuidePackage = class;
  TFhirImplementationGuidePackageList = class;
  TFhirImplementationGuidePackageResource = class;
  TFhirImplementationGuidePackageResourceList = class;
  TFhirImplementationGuideGlobal = class;
  TFhirImplementationGuideGlobalList = class;
  TFhirImplementationGuidePage = class;
  TFhirImplementationGuidePageList = class;
  TFhirImplementationGuide = class;
  TFhirImplementationGuideList = class;
  TFhirListEntry = class;
  TFhirListEntryList = class;
  TFhirList = class;
  TFhirListList = class;
  TFhirLocationPosition = class;
  TFhirLocationPositionList = class;
  TFhirLocation = class;
  TFhirLocationList = class;
  TFhirMedia = class;
  TFhirMediaList = class;
  TFhirMedicationProduct = class;
  TFhirMedicationProductList = class;
  TFhirMedicationProductIngredient = class;
  TFhirMedicationProductIngredientList = class;
  TFhirMedicationProductBatch = class;
  TFhirMedicationProductBatchList = class;
  TFhirMedicationPackage = class;
  TFhirMedicationPackageList = class;
  TFhirMedicationPackageContent = class;
  TFhirMedicationPackageContentList = class;
  TFhirMedication = class;
  TFhirMedicationList = class;
  TFhirMedicationAdministrationDosage = class;
  TFhirMedicationAdministrationDosageList = class;
  TFhirMedicationAdministration = class;
  TFhirMedicationAdministrationList = class;
  TFhirMedicationDispenseDosageInstruction = class;
  TFhirMedicationDispenseDosageInstructionList = class;
  TFhirMedicationDispenseSubstitution = class;
  TFhirMedicationDispenseSubstitutionList = class;
  TFhirMedicationDispense = class;
  TFhirMedicationDispenseList = class;
  TFhirMedicationOrderDosageInstruction = class;
  TFhirMedicationOrderDosageInstructionList = class;
  TFhirMedicationOrderDispenseRequest = class;
  TFhirMedicationOrderDispenseRequestList = class;
  TFhirMedicationOrderSubstitution = class;
  TFhirMedicationOrderSubstitutionList = class;
  TFhirMedicationOrder = class;
  TFhirMedicationOrderList = class;
  TFhirMedicationStatementDosage = class;
  TFhirMedicationStatementDosageList = class;
  TFhirMedicationStatement = class;
  TFhirMedicationStatementList = class;
  TFhirMessageHeaderResponse = class;
  TFhirMessageHeaderResponseList = class;
  TFhirMessageHeaderSource = class;
  TFhirMessageHeaderSourceList = class;
  TFhirMessageHeaderDestination = class;
  TFhirMessageHeaderDestinationList = class;
  TFhirMessageHeader = class;
  TFhirMessageHeaderList = class;
  TFhirNamingSystemContact = class;
  TFhirNamingSystemContactList = class;
  TFhirNamingSystemUniqueId = class;
  TFhirNamingSystemUniqueIdList = class;
  TFhirNamingSystem = class;
  TFhirNamingSystemList = class;
  TFhirNutritionOrderOralDiet = class;
  TFhirNutritionOrderOralDietList = class;
  TFhirNutritionOrderOralDietNutrient = class;
  TFhirNutritionOrderOralDietNutrientList = class;
  TFhirNutritionOrderOralDietTexture = class;
  TFhirNutritionOrderOralDietTextureList = class;
  TFhirNutritionOrderSupplement = class;
  TFhirNutritionOrderSupplementList = class;
  TFhirNutritionOrderEnteralFormula = class;
  TFhirNutritionOrderEnteralFormulaList = class;
  TFhirNutritionOrderEnteralFormulaAdministration = class;
  TFhirNutritionOrderEnteralFormulaAdministrationList = class;
  TFhirNutritionOrder = class;
  TFhirNutritionOrderList = class;
  TFhirObservationReferenceRange = class;
  TFhirObservationReferenceRangeList = class;
  TFhirObservationRelated = class;
  TFhirObservationRelatedList = class;
  TFhirObservationComponent = class;
  TFhirObservationComponentList = class;
  TFhirObservation = class;
  TFhirObservationList = class;
  TFhirOperationDefinitionContact = class;
  TFhirOperationDefinitionContactList = class;
  TFhirOperationDefinitionParameter = class;
  TFhirOperationDefinitionParameterList = class;
  TFhirOperationDefinitionParameterBinding = class;
  TFhirOperationDefinitionParameterBindingList = class;
  TFhirOperationDefinition = class;
  TFhirOperationDefinitionList = class;
  TFhirOperationOutcomeIssue = class;
  TFhirOperationOutcomeIssueList = class;
  TFhirOperationOutcome = class;
  TFhirOperationOutcomeList = class;
  TFhirOrderWhen = class;
  TFhirOrderWhenList = class;
  TFhirOrder = class;
  TFhirOrderList = class;
  TFhirOrderResponse = class;
  TFhirOrderResponseList = class;
  TFhirOrganizationContact = class;
  TFhirOrganizationContactList = class;
  TFhirOrganization = class;
  TFhirOrganizationList = class;
  TFhirPatientContact = class;
  TFhirPatientContactList = class;
  TFhirPatientAnimal = class;
  TFhirPatientAnimalList = class;
  TFhirPatientCommunication = class;
  TFhirPatientCommunicationList = class;
  TFhirPatientLink = class;
  TFhirPatientLinkList = class;
  TFhirPatient = class;
  TFhirPatientList = class;
  TFhirPaymentNotice = class;
  TFhirPaymentNoticeList = class;
  TFhirPaymentReconciliationDetail = class;
  TFhirPaymentReconciliationDetailList = class;
  TFhirPaymentReconciliationNote = class;
  TFhirPaymentReconciliationNoteList = class;
  TFhirPaymentReconciliation = class;
  TFhirPaymentReconciliationList = class;
  TFhirPersonLink = class;
  TFhirPersonLinkList = class;
  TFhirPerson = class;
  TFhirPersonList = class;
  TFhirPractitionerPractitionerRole = class;
  TFhirPractitionerPractitionerRoleList = class;
  TFhirPractitionerQualification = class;
  TFhirPractitionerQualificationList = class;
  TFhirPractitioner = class;
  TFhirPractitionerList = class;
  TFhirProcedurePerformer = class;
  TFhirProcedurePerformerList = class;
  TFhirProcedureFocalDevice = class;
  TFhirProcedureFocalDeviceList = class;
  TFhirProcedure = class;
  TFhirProcedureList = class;
  TFhirProcedureRequest = class;
  TFhirProcedureRequestList = class;
  TFhirProcessRequestItem = class;
  TFhirProcessRequestItemList = class;
  TFhirProcessRequest = class;
  TFhirProcessRequestList = class;
  TFhirProcessResponseNotes = class;
  TFhirProcessResponseNotesList = class;
  TFhirProcessResponse = class;
  TFhirProcessResponseList = class;
  TFhirProvenanceAgent = class;
  TFhirProvenanceAgentList = class;
  TFhirProvenanceAgentRelatedAgent = class;
  TFhirProvenanceAgentRelatedAgentList = class;
  TFhirProvenanceEntity = class;
  TFhirProvenanceEntityList = class;
  TFhirProvenance = class;
  TFhirProvenanceList = class;
  TFhirQuestionnaireGroup = class;
  TFhirQuestionnaireGroupList = class;
  TFhirQuestionnaireGroupQuestion = class;
  TFhirQuestionnaireGroupQuestionList = class;
  TFhirQuestionnaire = class;
  TFhirQuestionnaireList = class;
  TFhirQuestionnaireResponseGroup = class;
  TFhirQuestionnaireResponseGroupList = class;
  TFhirQuestionnaireResponseGroupQuestion = class;
  TFhirQuestionnaireResponseGroupQuestionList = class;
  TFhirQuestionnaireResponseGroupQuestionAnswer = class;
  TFhirQuestionnaireResponseGroupQuestionAnswerList = class;
  TFhirQuestionnaireResponse = class;
  TFhirQuestionnaireResponseList = class;
  TFhirReferralRequest = class;
  TFhirReferralRequestList = class;
  TFhirRelatedPerson = class;
  TFhirRelatedPersonList = class;
  TFhirRiskAssessmentPrediction = class;
  TFhirRiskAssessmentPredictionList = class;
  TFhirRiskAssessment = class;
  TFhirRiskAssessmentList = class;
  TFhirSchedule = class;
  TFhirScheduleList = class;
  TFhirSearchParameterContact = class;
  TFhirSearchParameterContactList = class;
  TFhirSearchParameter = class;
  TFhirSearchParameterList = class;
  TFhirSlot = class;
  TFhirSlotList = class;
  TFhirSpecimenCollection = class;
  TFhirSpecimenCollectionList = class;
  TFhirSpecimenTreatment = class;
  TFhirSpecimenTreatmentList = class;
  TFhirSpecimenContainer = class;
  TFhirSpecimenContainerList = class;
  TFhirSpecimen = class;
  TFhirSpecimenList = class;
  TFhirStructureDefinitionContact = class;
  TFhirStructureDefinitionContactList = class;
  TFhirStructureDefinitionMapping = class;
  TFhirStructureDefinitionMappingList = class;
  TFhirStructureDefinitionSnapshot = class;
  TFhirStructureDefinitionSnapshotList = class;
  TFhirStructureDefinitionDifferential = class;
  TFhirStructureDefinitionDifferentialList = class;
  TFhirStructureDefinition = class;
  TFhirStructureDefinitionList = class;
  TFhirSubscriptionChannel = class;
  TFhirSubscriptionChannelList = class;
  TFhirSubscription = class;
  TFhirSubscriptionList = class;
  TFhirSubstanceInstance = class;
  TFhirSubstanceInstanceList = class;
  TFhirSubstanceIngredient = class;
  TFhirSubstanceIngredientList = class;
  TFhirSubstance = class;
  TFhirSubstanceList = class;
  TFhirSupplyDelivery = class;
  TFhirSupplyDeliveryList = class;
  TFhirSupplyRequestWhen = class;
  TFhirSupplyRequestWhenList = class;
  TFhirSupplyRequest = class;
  TFhirSupplyRequestList = class;
  TFhirTestScriptContact = class;
  TFhirTestScriptContactList = class;
  TFhirTestScriptMetadata = class;
  TFhirTestScriptMetadataList = class;
  TFhirTestScriptMetadataLink = class;
  TFhirTestScriptMetadataLinkList = class;
  TFhirTestScriptMetadataCapability = class;
  TFhirTestScriptMetadataCapabilityList = class;
  TFhirTestScriptFixture = class;
  TFhirTestScriptFixtureList = class;
  TFhirTestScriptVariable = class;
  TFhirTestScriptVariableList = class;
  TFhirTestScriptSetup = class;
  TFhirTestScriptSetupList = class;
  TFhirTestScriptSetupAction = class;
  TFhirTestScriptSetupActionList = class;
  TFhirTestScriptSetupActionOperation = class;
  TFhirTestScriptSetupActionOperationList = class;
  TFhirTestScriptSetupActionOperationRequestHeader = class;
  TFhirTestScriptSetupActionOperationRequestHeaderList = class;
  TFhirTestScriptSetupActionAssert = class;
  TFhirTestScriptSetupActionAssertList = class;
  TFhirTestScriptTest = class;
  TFhirTestScriptTestList = class;
  TFhirTestScriptTestAction = class;
  TFhirTestScriptTestActionList = class;
  TFhirTestScriptTeardown = class;
  TFhirTestScriptTeardownList = class;
  TFhirTestScriptTeardownAction = class;
  TFhirTestScriptTeardownActionList = class;
  TFhirTestScript = class;
  TFhirTestScriptList = class;
  TFhirValueSetContact = class;
  TFhirValueSetContactList = class;
  TFhirValueSetCodeSystem = class;
  TFhirValueSetCodeSystemList = class;
  TFhirValueSetCodeSystemConcept = class;
  TFhirValueSetCodeSystemConceptList = class;
  TFhirValueSetCodeSystemConceptDesignation = class;
  TFhirValueSetCodeSystemConceptDesignationList = class;
  TFhirValueSetCompose = class;
  TFhirValueSetComposeList = class;
  TFhirValueSetComposeInclude = class;
  TFhirValueSetComposeIncludeList = class;
  TFhirValueSetComposeIncludeConcept = class;
  TFhirValueSetComposeIncludeConceptList = class;
  TFhirValueSetComposeIncludeFilter = class;
  TFhirValueSetComposeIncludeFilterList = class;
  TFhirValueSetExpansion = class;
  TFhirValueSetExpansionList = class;
  TFhirValueSetExpansionParameter = class;
  TFhirValueSetExpansionParameterList = class;
  TFhirValueSetExpansionContains = class;
  TFhirValueSetExpansionContainsList = class;
  TFhirValueSet = class;
  TFhirValueSetList = class;
  TFhirVisionPrescriptionDispense = class;
  TFhirVisionPrescriptionDispenseList = class;
  TFhirVisionPrescription = class;
  TFhirVisionPrescriptionList = class;

  {@Class TFhirResource : TFHIRBase
    This is the base resource type for everything.
  }
  {!.Net HL7Connect.Fhir.Resource}
  TFhirResource = {abstract} class (TFHIRBase)
  private
    FId : TFhirId;
    FMeta : TFhirMeta;
    FImplicitRules : TFhirUri;
    FLanguage : TFhirCode;
    Procedure SetId(value : TFhirId);
    Function GetIdST : String;
    Procedure SetIdST(value : String);
    Procedure SetMeta(value : TFhirMeta);
    Procedure SetImplicitRules(value : TFhirUri);
    Function GetImplicitRulesST : String;
    Procedure SetImplicitRulesST(value : String);
    Procedure SetLanguage(value : TFhirCode);
    Function GetLanguageST : String;
    Procedure SetLanguageST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
    function GetResourceType : TFhirResourceType; virtual; abstract;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirResource; overload;
    function Clone : TFhirResource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    Property ResourceType : TFhirResourceType read GetResourceType;

    {@member id
      Typed access to The logical id of the resource, as used in the URL for the resource. Once assigned, this value never changes.
    }
    property id : String read GetIdST write SetIdST;
    {@member idElement
      The logical id of the resource, as used in the URL for the resource. Once assigned, this value never changes.
    }
    property idElement : TFhirId read FId write SetId;

    {@member meta
      Typed access to The metadata about the resource. This is content that is maintained by the infrastructure. Changes to the content may not always be associated with version changes to the resource. (defined for API consistency)
    }
    property meta : TFhirMeta read FMeta write SetMeta;
    {@member metaElement
      The metadata about the resource. This is content that is maintained by the infrastructure. Changes to the content may not always be associated with version changes to the resource.
    }
    property metaElement : TFhirMeta read FMeta write SetMeta;

    {@member implicitRules
      Typed access to A reference to a set of rules that were followed when the resource was constructed, and which must be understood when processing the content.
    }
    property implicitRules : String read GetImplicitRulesST write SetImplicitRulesST;
    {@member implicitRulesElement
      A reference to a set of rules that were followed when the resource was constructed, and which must be understood when processing the content.
    }
    property implicitRulesElement : TFhirUri read FImplicitRules write SetImplicitRules;

    {@member language
      Typed access to The base language in which the resource is written.
    }
    property language : String read GetLanguageST write SetLanguageST;
    {@member languageElement
      The base language in which the resource is written.
    }
    property languageElement : TFhirCode read FLanguage write SetLanguage;

  end;

  TFhirResourceClass = class of TFhirResource;

  TFhirResourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirResourceList;
    function GetCurrent : TFhirResource;
  public
    Constructor Create(list : TFhirResourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirResource read GetCurrent;
  end;


  {@Class TFhirResourceList
    A list of FhirResource
  }
  {!.Net HL7Connect.Fhir.ResourceList}
  TFhirResourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirResource;
    procedure SetItemN(index : Integer; value : TFhirResource);
  public
    {!script hide}
    function Link : TFhirResourceList; Overload;
    function Clone : TFhirResourceList; Overload;
    function GetEnumerator : TFhirResourceListEnumerator;
    {!script show}
    

    
    {@member AddItem
      Add an already existing FhirResource to the end of the list.
    }
    procedure AddItem(value : TFhirResource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirResource) : Integer;
    

    {@member InsertItem
       Insert an existing FhirResource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirResource);
    
    {@member Item
       Get the iIndexth FhirResource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirResource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirResource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirResource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirResources[index : Integer] : TFhirResource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDomainResource : TFhirResource
    A resource that includes narrative, extensions, and contained resources.
  }
  {!.Net HL7Connect.Fhir.DomainResource}
  TFhirDomainResource = {abstract} class (TFhirResource)
  private
    FText : TFhirNarrative;
    FcontainedList : TFhirResourceList;
    FextensionList : TFhirExtensionList;
    FmodifierExtensionList : TFhirExtensionList;
    Procedure SetText(value : TFhirNarrative);
    function GetContainedList : TFhirResourceList;
    function GetHasContainedList : Boolean;
    function GetExtensionList : TFhirExtensionList;
    function GetHasExtensionList : Boolean;
    function GetModifierExtensionList : TFhirExtensionList;
    function GetHasModifierExtensionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDomainResource; overload;
    function Clone : TFhirDomainResource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member text
      Typed access to A human-readable narrative that contains a summary of the resource, and may be used to represent the content of the resource to a human. The narrative need not encode all the structured data, but is required to contain sufficient detail to make it "clinically safe" for a human to just read the narrative. Resource definitions may define what content should be represented in the narrative to ensure clinical safety. (defined for API consistency)
    }
    property text : TFhirNarrative read FText write SetText;
    {@member textElement
      A human-readable narrative that contains a summary of the resource, and may be used to represent the content of the resource to a human. The narrative need not encode all the structured data, but is required to contain sufficient detail to make it "clinically safe" for a human to just read the narrative. Resource definitions may define what content should be represented in the narrative to ensure clinical safety.
    }
    property textElement : TFhirNarrative read FText write SetText;

    {@member containedList
      These resources do not have an independent existence apart from the resource that contains them - they cannot be identified independently, and nor can they have their own independent transaction scope.
    }
    property containedList : TFhirResourceList read GetContainedList;
    property hasContainedList : boolean read GetHasContainedList;

    {@member extensionList
      May be used to represent additional information that is not part of the basic definition of the resource. In order to make the use of extensions safe and manageable, there is a strict set of governance  applied to the definition and use of extensions. Though any implementer is allowed to define an extension, there is a set of requirements that SHALL be met as part of the definition of the extension.
    }
    property extensionList : TFhirExtensionList read GetExtensionList;
    property hasExtensionList : boolean read GetHasExtensionList;

    {@member modifierExtensionList
      May be used to represent additional information that is not part of the basic definition of the resource, and that modifies the understanding of the element that contains it. Usually modifier elements provide negation or qualification. In order to make the use of extensions safe and manageable, there is a strict set of governance applied to the definition and use of extensions. Though any implementer is allowed to define an extension, there is a set of requirements that SHALL be met as part of the definition of the extension. Applications processing a resource are required to check for modifier extensions.
    }
    property modifierExtensionList : TFhirExtensionList read GetModifierExtensionList;
    property hasModifierExtensionList : boolean read GetHasModifierExtensionList;

  end;


  TFhirDomainResourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDomainResourceList;
    function GetCurrent : TFhirDomainResource;
  public
    Constructor Create(list : TFhirDomainResourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDomainResource read GetCurrent;
  end;


  {@Class TFhirDomainResourceList
    A list of FhirDomainResource
  }
  {!.Net HL7Connect.Fhir.DomainResourceList}
  TFhirDomainResourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDomainResource;
    procedure SetItemN(index : Integer; value : TFhirDomainResource);
  public
    {!script hide}
    function Link : TFhirDomainResourceList; Overload;
    function Clone : TFhirDomainResourceList; Overload;
    function GetEnumerator : TFhirDomainResourceListEnumerator;
    {!script show}
    

    
    {@member AddItem
      Add an already existing FhirDomainResource to the end of the list.
    }
    procedure AddItem(value : TFhirDomainResource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDomainResource) : Integer;
    

    {@member InsertItem
       Insert an existing FhirDomainResource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDomainResource);
    
    {@member Item
       Get the iIndexth FhirDomainResource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDomainResource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDomainResource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDomainResource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDomainResources[index : Integer] : TFhirDomainResource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirParametersParameter : TFhirElement
    A parameter passed to or received from the operation.
  }
  {!.Net HL7Connect.Fhir.ParametersParameter}
  TFhirParametersParameter = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FValue : TFhirType;
    FResource : TFhirResource;
    FpartList : TFhirParametersParameterList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetValue(value : TFhirType);
    Procedure SetResource(value : TFhirResource);
    function GetPartList : TFhirParametersParameterList;
    function GetHasPartList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirParametersParameter; overload;
    function Clone : TFhirParametersParameter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of the parameter (reference to the operation definition).
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of the parameter (reference to the operation definition).
    }
    property nameElement : TFhirString read FName write SetName;

    {@member value
      Typed access to If the parameter is a data type. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      If the parameter is a data type.
    }
    property valueElement : TFhirType read FValue write SetValue;

    {@member resource
      Typed access to If the parameter is a whole resource. (defined for API consistency)
    }
    property resource : TFhirResource read FResource write SetResource;
    {@member resourceElement
      If the parameter is a whole resource.
    }
    property resourceElement : TFhirResource read FResource write SetResource;

    {@member partList
      A named part of a parameter. In many implementation context, a set of named parts is known as a "Tuple".
    }
    property partList : TFhirParametersParameterList read GetPartList;
    property hasPartList : boolean read GetHasPartList;

  end;


  TFhirParametersParameterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirParametersParameterList;
    function GetCurrent : TFhirParametersParameter;
  public
    Constructor Create(list : TFhirParametersParameterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirParametersParameter read GetCurrent;
  end;


  {@Class TFhirParametersParameterList
    A list of FhirParametersParameter
  }
  {!.Net HL7Connect.Fhir.ParametersParameterList}
  TFhirParametersParameterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirParametersParameter;
    procedure SetItemN(index : Integer; value : TFhirParametersParameter);
  public
    {!script hide}
    function Link : TFhirParametersParameterList; Overload;
    function Clone : TFhirParametersParameterList; Overload;
    function GetEnumerator : TFhirParametersParameterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirParametersParameter to the end of the list.
    }
    function Append : TFhirParametersParameter;

    
    {@member AddItem
      Add an already existing FhirParametersParameter to the end of the list.
    }
    procedure AddItem(value : TFhirParametersParameter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirParametersParameter) : Integer;
    

    {@member Insert
      Insert FhirParametersParameter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirParametersParameter;
    

    {@member InsertItem
       Insert an existing FhirParametersParameter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirParametersParameter);
    
    {@member Item
       Get the iIndexth FhirParametersParameter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirParametersParameter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirParametersParameter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirParametersParameter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirParametersParameters[index : Integer] : TFhirParametersParameter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirParameters : TFhirResource
    This special resource type is used to represent an operation request and response (operations.html). It has no other use, and there is no RESTful endpoint associated with it.
  }
  {!.Net HL7Connect.Fhir.Parameters}
  TFhirParameters = class (TFhirResource)
  private
    FparameterList : TFhirParametersParameterList;
    function GetParameterList : TFhirParametersParameterList;
    function GetHasParameterList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirParameters; overload;
    function Clone : TFhirParameters; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member parameterList
      A parameter passed to or received from the operation.
    }
    property parameterList : TFhirParametersParameterList read GetParameterList;
    property hasParameterList : boolean read GetHasParameterList;

  end;


  TFhirParametersListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirParametersList;
    function GetCurrent : TFhirParameters;
  public
    Constructor Create(list : TFhirParametersList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirParameters read GetCurrent;
  end;


  {@Class TFhirParametersList
    A list of FhirParameters
  }
  {!.Net HL7Connect.Fhir.ParametersList}
  TFhirParametersList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirParameters;
    procedure SetItemN(index : Integer; value : TFhirParameters);
  public
    {!script hide}
    function Link : TFhirParametersList; Overload;
    function Clone : TFhirParametersList; Overload;
    function GetEnumerator : TFhirParametersListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirParameters to the end of the list.
    }
    function Append : TFhirParameters;

    
    {@member AddItem
      Add an already existing FhirParameters to the end of the list.
    }
    procedure AddItem(value : TFhirParameters); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirParameters) : Integer;
    

    {@member Insert
      Insert FhirParameters before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirParameters;
    

    {@member InsertItem
       Insert an existing FhirParameters before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirParameters);
    
    {@member Item
       Get the iIndexth FhirParameters. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirParameters. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirParameters);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirParameters;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirParameters[index : Integer] : TFhirParameters read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAccount : TFhirDomainResource
    A financial tool for tracking value accrued for a particular purpose.  In the healthcare field, used to track charges for a patient, cost centres, etc.
  }
  {!.Net HL7Connect.Fhir.Account}
  TFhirAccount = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FName : TFhirString;
    FType_ : TFhirCodeableConcept;
    FStatus : TFhirEnum;
    FActivePeriod : TFhirPeriod;
    FCurrency : TFhirCoding;
    FBalance : TFhirQuantity;
    FCoveragePeriod : TFhirPeriod;
    FSubject : TFhirReference{Resource};
    FOwner : TFhirReference{TFhirOrganization};
    FDescription : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirAccountStatus;
    Procedure SetStatusST(value : TFhirAccountStatus);
    Procedure SetActivePeriod(value : TFhirPeriod);
    Procedure SetCurrency(value : TFhirCoding);
    Procedure SetBalance(value : TFhirQuantity);
    Procedure SetCoveragePeriod(value : TFhirPeriod);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetOwner(value : TFhirReference{TFhirOrganization});
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAccount; overload;
    function Clone : TFhirAccount; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Unique identifier used to reference the account.  May or may not be intended for human use (e.g. credit card number).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member name
      Typed access to Name used for the account when displaying it to humans in reports, etc.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Name used for the account when displaying it to humans in reports, etc.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member type_
      Typed access to Categorizes the account for reporting and searching purposes. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Categorizes the account for reporting and searching purposes.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member status
      Indicates whether the account is presently used/useable or not.
    }
    property status : TFhirAccountStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member activePeriod
      Typed access to Indicates the period of time over which the account is allowed. (defined for API consistency)
    }
    property activePeriod : TFhirPeriod read FActivePeriod write SetActivePeriod;
    {@member activePeriodElement
      Indicates the period of time over which the account is allowed.
    }
    property activePeriodElement : TFhirPeriod read FActivePeriod write SetActivePeriod;

    {@member currency
      Typed access to Identifies the currency to which transactions must be converted when crediting or debiting the account. (defined for API consistency)
    }
    property currency : TFhirCoding read FCurrency write SetCurrency;
    {@member currencyElement
      Identifies the currency to which transactions must be converted when crediting or debiting the account.
    }
    property currencyElement : TFhirCoding read FCurrency write SetCurrency;

    {@member balance
      Typed access to Represents the sum of all credits less all debits associated with the account.  Might be positive, zero or negative. (defined for API consistency)
    }
    property balance : TFhirQuantity read FBalance write SetBalance;
    {@member balanceElement
      Represents the sum of all credits less all debits associated with the account.  Might be positive, zero or negative.
    }
    property balanceElement : TFhirQuantity read FBalance write SetBalance;

    {@member coveragePeriod
      Typed access to Identifies the period of time the account applies to; e.g. accounts created per fiscal year, quarter, etc. (defined for API consistency)
    }
    property coveragePeriod : TFhirPeriod read FCoveragePeriod write SetCoveragePeriod;
    {@member coveragePeriodElement
      Identifies the period of time the account applies to; e.g. accounts created per fiscal year, quarter, etc.
    }
    property coveragePeriodElement : TFhirPeriod read FCoveragePeriod write SetCoveragePeriod;

    {@member subject
      Typed access to Identifies the patient, device, practitioner, location or other object the account is associated with. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Identifies the patient, device, practitioner, location or other object the account is associated with.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member owner
      Typed access to Indicates the organization, department, etc. with responsibility for the account. (defined for API consistency)
    }
    property owner : TFhirReference{TFhirOrganization} read FOwner write SetOwner;
    {@member ownerElement
      Indicates the organization, department, etc. with responsibility for the account.
    }
    property ownerElement : TFhirReference{TFhirOrganization} read FOwner write SetOwner;

    {@member description
      Typed access to Provides additional information about what the account tracks and how it is used.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Provides additional information about what the account tracks and how it is used.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

  end;


  TFhirAccountListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAccountList;
    function GetCurrent : TFhirAccount;
  public
    Constructor Create(list : TFhirAccountList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAccount read GetCurrent;
  end;


  {@Class TFhirAccountList
    A list of FhirAccount
  }
  {!.Net HL7Connect.Fhir.AccountList}
  TFhirAccountList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAccount;
    procedure SetItemN(index : Integer; value : TFhirAccount);
  public
    {!script hide}
    function Link : TFhirAccountList; Overload;
    function Clone : TFhirAccountList; Overload;
    function GetEnumerator : TFhirAccountListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAccount to the end of the list.
    }
    function Append : TFhirAccount;

    
    {@member AddItem
      Add an already existing FhirAccount to the end of the list.
    }
    procedure AddItem(value : TFhirAccount); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAccount) : Integer;
    

    {@member Insert
      Insert FhirAccount before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAccount;
    

    {@member InsertItem
       Insert an existing FhirAccount before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAccount);
    
    {@member Item
       Get the iIndexth FhirAccount. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAccount. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAccount);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAccount;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAccounts[index : Integer] : TFhirAccount read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAllergyIntoleranceReaction : TFhirElement
    Details about each adverse reaction event linked to exposure to the identified Substance.
  }
  {!.Net HL7Connect.Fhir.AllergyIntoleranceReaction}
  TFhirAllergyIntoleranceReaction = class (TFhirBackboneElement)
  private
    FSubstance : TFhirCodeableConcept;
    FCertainty : TFhirEnum;
    FmanifestationList : TFhirCodeableConceptList;
    FDescription : TFhirString;
    FOnset : TFhirDateTime;
    FSeverity : TFhirEnum;
    FExposureRoute : TFhirCodeableConcept;
    FNote : TFhirAnnotation;
    Procedure SetSubstance(value : TFhirCodeableConcept);
    Procedure SetCertainty(value : TFhirEnum);
    Function GetCertaintyST : TFhirReactionEventCertainty;
    Procedure SetCertaintyST(value : TFhirReactionEventCertainty);
    function GetManifestationList : TFhirCodeableConceptList;
    function GetHasManifestationList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetOnset(value : TFhirDateTime);
    Function GetOnsetST : TDateAndTime;
    Procedure SetOnsetST(value : TDateAndTime);
    Procedure SetSeverity(value : TFhirEnum);
    Function GetSeverityST : TFhirReactionEventSeverity;
    Procedure SetSeverityST(value : TFhirReactionEventSeverity);
    Procedure SetExposureRoute(value : TFhirCodeableConcept);
    Procedure SetNote(value : TFhirAnnotation);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAllergyIntoleranceReaction; overload;
    function Clone : TFhirAllergyIntoleranceReaction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member substance
      Typed access to Identification of the specific substance considered to be responsible for the Adverse Reaction event. Note: the substance for a specific reaction may be different to the substance identified as the cause of the risk, but must be consistent with it. For instance, it may be a more specific substance (e.g. a brand medication) or a composite substance that includes the identified substance. It must be clinically safe to only process the AllergyIntolerance.substance and ignore the AllergyIntolerance.event.substance. (defined for API consistency)
    }
    property substance : TFhirCodeableConcept read FSubstance write SetSubstance;
    {@member substanceElement
      Identification of the specific substance considered to be responsible for the Adverse Reaction event. Note: the substance for a specific reaction may be different to the substance identified as the cause of the risk, but must be consistent with it. For instance, it may be a more specific substance (e.g. a brand medication) or a composite substance that includes the identified substance. It must be clinically safe to only process the AllergyIntolerance.substance and ignore the AllergyIntolerance.event.substance.
    }
    property substanceElement : TFhirCodeableConcept read FSubstance write SetSubstance;

    {@member certainty
      Statement about the degree of clinical certainty that the specific substance was the cause of the manifestation in this reaction event.
    }
    property certainty : TFhirReactionEventCertainty read GetCertaintyST write SetCertaintyST;
    property certaintyElement : TFhirEnum read FCertainty write SetCertainty;

    {@member manifestationList
      Clinical symptoms and/or signs that are observed or associated with the adverse reaction event.
    }
    property manifestationList : TFhirCodeableConceptList read GetManifestationList;
    property hasManifestationList : boolean read GetHasManifestationList;

    {@member description
      Typed access to Text description about the reaction as a whole, including details of the manifestation if required.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Text description about the reaction as a whole, including details of the manifestation if required.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member onset
      Typed access to Record of the date and/or time of the onset of the Reaction.
    }
    property onset : TDateAndTime read GetOnsetST write SetOnsetST;
    {@member onsetElement
      Record of the date and/or time of the onset of the Reaction.
    }
    property onsetElement : TFhirDateTime read FOnset write SetOnset;

    {@member severity
      Clinical assessment of the severity of the reaction event as a whole, potentially considering multiple different manifestations.
    }
    property severity : TFhirReactionEventSeverity read GetSeverityST write SetSeverityST;
    property severityElement : TFhirEnum read FSeverity write SetSeverity;

    {@member exposureRoute
      Typed access to Identification of the route by which the subject was exposed to the substance. (defined for API consistency)
    }
    property exposureRoute : TFhirCodeableConcept read FExposureRoute write SetExposureRoute;
    {@member exposureRouteElement
      Identification of the route by which the subject was exposed to the substance.
    }
    property exposureRouteElement : TFhirCodeableConcept read FExposureRoute write SetExposureRoute;

    {@member note
      Typed access to Additional text about the adverse reaction event not captured in other fields. (defined for API consistency)
    }
    property note : TFhirAnnotation read FNote write SetNote;
    {@member noteElement
      Additional text about the adverse reaction event not captured in other fields.
    }
    property noteElement : TFhirAnnotation read FNote write SetNote;

  end;


  TFhirAllergyIntoleranceReactionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAllergyIntoleranceReactionList;
    function GetCurrent : TFhirAllergyIntoleranceReaction;
  public
    Constructor Create(list : TFhirAllergyIntoleranceReactionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAllergyIntoleranceReaction read GetCurrent;
  end;


  {@Class TFhirAllergyIntoleranceReactionList
    A list of FhirAllergyIntoleranceReaction
  }
  {!.Net HL7Connect.Fhir.AllergyIntoleranceReactionList}
  TFhirAllergyIntoleranceReactionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAllergyIntoleranceReaction;
    procedure SetItemN(index : Integer; value : TFhirAllergyIntoleranceReaction);
  public
    {!script hide}
    function Link : TFhirAllergyIntoleranceReactionList; Overload;
    function Clone : TFhirAllergyIntoleranceReactionList; Overload;
    function GetEnumerator : TFhirAllergyIntoleranceReactionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAllergyIntoleranceReaction to the end of the list.
    }
    function Append : TFhirAllergyIntoleranceReaction;

    
    {@member AddItem
      Add an already existing FhirAllergyIntoleranceReaction to the end of the list.
    }
    procedure AddItem(value : TFhirAllergyIntoleranceReaction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAllergyIntoleranceReaction) : Integer;
    

    {@member Insert
      Insert FhirAllergyIntoleranceReaction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAllergyIntoleranceReaction;
    

    {@member InsertItem
       Insert an existing FhirAllergyIntoleranceReaction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAllergyIntoleranceReaction);
    
    {@member Item
       Get the iIndexth FhirAllergyIntoleranceReaction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAllergyIntoleranceReaction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAllergyIntoleranceReaction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAllergyIntoleranceReaction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAllergyIntoleranceReactions[index : Integer] : TFhirAllergyIntoleranceReaction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAllergyIntolerance : TFhirDomainResource
    Risk of harmful or undesirable, physiological response which is unique to an individual and associated with exposure to a substance.
  }
  {!.Net HL7Connect.Fhir.AllergyIntolerance}
  TFhirAllergyIntolerance = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FOnset : TFhirDateTime;
    FRecordedDate : TFhirDateTime;
    FRecorder : TFhirReference{Resource};
    FPatient : TFhirReference{TFhirPatient};
    FReporter : TFhirReference{Resource};
    FSubstance : TFhirCodeableConcept;
    FStatus : TFhirEnum;
    FCriticality : TFhirEnum;
    FType_ : TFhirEnum;
    FCategory : TFhirEnum;
    FLastOccurence : TFhirDateTime;
    FNote : TFhirAnnotation;
    FreactionList : TFhirAllergyIntoleranceReactionList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetOnset(value : TFhirDateTime);
    Function GetOnsetST : TDateAndTime;
    Procedure SetOnsetST(value : TDateAndTime);
    Procedure SetRecordedDate(value : TFhirDateTime);
    Function GetRecordedDateST : TDateAndTime;
    Procedure SetRecordedDateST(value : TDateAndTime);
    Procedure SetRecorder(value : TFhirReference{Resource});
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetReporter(value : TFhirReference{Resource});
    Procedure SetSubstance(value : TFhirCodeableConcept);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirAllergyIntoleranceStatus;
    Procedure SetStatusST(value : TFhirAllergyIntoleranceStatus);
    Procedure SetCriticality(value : TFhirEnum);
    Function GetCriticalityST : TFhirAllergyIntoleranceCriticality;
    Procedure SetCriticalityST(value : TFhirAllergyIntoleranceCriticality);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirAllergyIntoleranceType;
    Procedure SetType_ST(value : TFhirAllergyIntoleranceType);
    Procedure SetCategory(value : TFhirEnum);
    Function GetCategoryST : TFhirAllergyIntoleranceCategory;
    Procedure SetCategoryST(value : TFhirAllergyIntoleranceCategory);
    Procedure SetLastOccurence(value : TFhirDateTime);
    Function GetLastOccurenceST : TDateAndTime;
    Procedure SetLastOccurenceST(value : TDateAndTime);
    Procedure SetNote(value : TFhirAnnotation);
    function GetReactionList : TFhirAllergyIntoleranceReactionList;
    function GetHasReactionList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAllergyIntolerance; overload;
    function Clone : TFhirAllergyIntolerance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this allergy/intolerance concern that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member onset
      Typed access to Record of the date and/or time of the onset of the Allergy or Intolerance.
    }
    property onset : TDateAndTime read GetOnsetST write SetOnsetST;
    {@member onsetElement
      Record of the date and/or time of the onset of the Allergy or Intolerance.
    }
    property onsetElement : TFhirDateTime read FOnset write SetOnset;

    {@member recordedDate
      Typed access to Date when the sensitivity was recorded.
    }
    property recordedDate : TDateAndTime read GetRecordedDateST write SetRecordedDateST;
    {@member recordedDateElement
      Date when the sensitivity was recorded.
    }
    property recordedDateElement : TFhirDateTime read FRecordedDate write SetRecordedDate;

    {@member recorder
      Typed access to Individual who recorded the record and takes responsibility for its conten. (defined for API consistency)
    }
    property recorder : TFhirReference{Resource} read FRecorder write SetRecorder;
    {@member recorderElement
      Individual who recorded the record and takes responsibility for its conten.
    }
    property recorderElement : TFhirReference{Resource} read FRecorder write SetRecorder;

    {@member patient
      Typed access to The patient who has the allergy or intolerance. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient who has the allergy or intolerance.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member reporter
      Typed access to The source of the information about the allergy that is recorded. (defined for API consistency)
    }
    property reporter : TFhirReference{Resource} read FReporter write SetReporter;
    {@member reporterElement
      The source of the information about the allergy that is recorded.
    }
    property reporterElement : TFhirReference{Resource} read FReporter write SetReporter;

    {@member substance
      Typed access to Identification of a substance, or a class of substances, that is considered to be responsible for the adverse reaction risk. (defined for API consistency)
    }
    property substance : TFhirCodeableConcept read FSubstance write SetSubstance;
    {@member substanceElement
      Identification of a substance, or a class of substances, that is considered to be responsible for the adverse reaction risk.
    }
    property substanceElement : TFhirCodeableConcept read FSubstance write SetSubstance;

    {@member status
      Assertion about certainty associated with the propensity, or potential risk, of a reaction to the identified Substance.
    }
    property status : TFhirAllergyIntoleranceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member criticality
      Estimate of the potential clinical harm, or seriousness, of the reaction to the identified Substance.
    }
    property criticality : TFhirAllergyIntoleranceCriticality read GetCriticalityST write SetCriticalityST;
    property criticalityElement : TFhirEnum read FCriticality write SetCriticality;

    {@member type_
      Identification of the underlying physiological mechanism for the reaction risk.
    }
    property type_ : TFhirAllergyIntoleranceType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member category
      Category of the identified Substance.
    }
    property category : TFhirAllergyIntoleranceCategory read GetCategoryST write SetCategoryST;
    property categoryElement : TFhirEnum read FCategory write SetCategory;

    {@member lastOccurence
      Typed access to Represents the date and/or time of the last known occurrence of a reaction event.
    }
    property lastOccurence : TDateAndTime read GetLastOccurenceST write SetLastOccurenceST;
    {@member lastOccurenceElement
      Represents the date and/or time of the last known occurrence of a reaction event.
    }
    property lastOccurenceElement : TFhirDateTime read FLastOccurence write SetLastOccurence;

    {@member note
      Typed access to Additional narrative about the propensity for the Adverse Reaction, not captured in other fields. (defined for API consistency)
    }
    property note : TFhirAnnotation read FNote write SetNote;
    {@member noteElement
      Additional narrative about the propensity for the Adverse Reaction, not captured in other fields.
    }
    property noteElement : TFhirAnnotation read FNote write SetNote;

    {@member reactionList
      Details about each adverse reaction event linked to exposure to the identified Substance.
    }
    property reactionList : TFhirAllergyIntoleranceReactionList read GetReactionList;
    property hasReactionList : boolean read GetHasReactionList;

  end;


  TFhirAllergyIntoleranceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAllergyIntoleranceList;
    function GetCurrent : TFhirAllergyIntolerance;
  public
    Constructor Create(list : TFhirAllergyIntoleranceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAllergyIntolerance read GetCurrent;
  end;


  {@Class TFhirAllergyIntoleranceList
    A list of FhirAllergyIntolerance
  }
  {!.Net HL7Connect.Fhir.AllergyIntoleranceList}
  TFhirAllergyIntoleranceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAllergyIntolerance;
    procedure SetItemN(index : Integer; value : TFhirAllergyIntolerance);
  public
    {!script hide}
    function Link : TFhirAllergyIntoleranceList; Overload;
    function Clone : TFhirAllergyIntoleranceList; Overload;
    function GetEnumerator : TFhirAllergyIntoleranceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAllergyIntolerance to the end of the list.
    }
    function Append : TFhirAllergyIntolerance;

    
    {@member AddItem
      Add an already existing FhirAllergyIntolerance to the end of the list.
    }
    procedure AddItem(value : TFhirAllergyIntolerance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAllergyIntolerance) : Integer;
    

    {@member Insert
      Insert FhirAllergyIntolerance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAllergyIntolerance;
    

    {@member InsertItem
       Insert an existing FhirAllergyIntolerance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAllergyIntolerance);
    
    {@member Item
       Get the iIndexth FhirAllergyIntolerance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAllergyIntolerance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAllergyIntolerance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAllergyIntolerance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAllergyIntolerances[index : Integer] : TFhirAllergyIntolerance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAppointmentParticipant : TFhirElement
    List of participants involved in the appointment.
  }
  {!.Net HL7Connect.Fhir.AppointmentParticipant}
  TFhirAppointmentParticipant = class (TFhirBackboneElement)
  private
    Ftype_List : TFhirCodeableConceptList;
    FActor : TFhirReference{Resource};
    FRequired : TFhirEnum;
    FStatus : TFhirEnum;
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    Procedure SetActor(value : TFhirReference{Resource});
    Procedure SetRequired(value : TFhirEnum);
    Function GetRequiredST : TFhirParticipantrequired;
    Procedure SetRequiredST(value : TFhirParticipantrequired);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirParticipationstatus;
    Procedure SetStatusST(value : TFhirParticipationstatus);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAppointmentParticipant; overload;
    function Clone : TFhirAppointmentParticipant; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_List
      Role of participant in the appointment.
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member actor
      Typed access to A Person, Location/HealthcareService or Device that is participating in the appointment. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      A Person, Location/HealthcareService or Device that is participating in the appointment.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

    {@member required
      Is this participant required to be present at the meeting. This covers a use-case where 2 doctors need to meet to discuss the results for a specific patient, and the patient is not required to be present.
    }
    property required : TFhirParticipantrequired read GetRequiredST write SetRequiredST;
    property requiredElement : TFhirEnum read FRequired write SetRequired;

    {@member status
      Participation status of the Patient.
    }
    property status : TFhirParticipationstatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

  end;


  TFhirAppointmentParticipantListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAppointmentParticipantList;
    function GetCurrent : TFhirAppointmentParticipant;
  public
    Constructor Create(list : TFhirAppointmentParticipantList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAppointmentParticipant read GetCurrent;
  end;


  {@Class TFhirAppointmentParticipantList
    A list of FhirAppointmentParticipant
  }
  {!.Net HL7Connect.Fhir.AppointmentParticipantList}
  TFhirAppointmentParticipantList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAppointmentParticipant;
    procedure SetItemN(index : Integer; value : TFhirAppointmentParticipant);
  public
    {!script hide}
    function Link : TFhirAppointmentParticipantList; Overload;
    function Clone : TFhirAppointmentParticipantList; Overload;
    function GetEnumerator : TFhirAppointmentParticipantListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAppointmentParticipant to the end of the list.
    }
    function Append : TFhirAppointmentParticipant;

    
    {@member AddItem
      Add an already existing FhirAppointmentParticipant to the end of the list.
    }
    procedure AddItem(value : TFhirAppointmentParticipant); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAppointmentParticipant) : Integer;
    

    {@member Insert
      Insert FhirAppointmentParticipant before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAppointmentParticipant;
    

    {@member InsertItem
       Insert an existing FhirAppointmentParticipant before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAppointmentParticipant);
    
    {@member Item
       Get the iIndexth FhirAppointmentParticipant. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAppointmentParticipant. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAppointmentParticipant);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAppointmentParticipant;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAppointmentParticipants[index : Integer] : TFhirAppointmentParticipant read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAppointment : TFhirDomainResource
    A booking of a healthcare event among patient(s), practitioner(s), related person(s) and/or device(s) for a specific date/time. This may result in one or more Encounter(s).
  }
  {!.Net HL7Connect.Fhir.Appointment}
  TFhirAppointment = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FType_ : TFhirCodeableConcept;
    FReason : TFhirCodeableConcept;
    FPriority : TFhirUnsignedInt;
    FDescription : TFhirString;
    FStart : TFhirInstant;
    FEnd_ : TFhirInstant;
    FMinutesDuration : TFhirPositiveInt;
    FslotList : TFhirReferenceList{TFhirSlot};
    FComment : TFhirString;
    FparticipantList : TFhirAppointmentParticipantList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirAppointmentstatus;
    Procedure SetStatusST(value : TFhirAppointmentstatus);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetReason(value : TFhirCodeableConcept);
    Procedure SetPriority(value : TFhirUnsignedInt);
    Function GetPriorityST : String;
    Procedure SetPriorityST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetStart(value : TFhirInstant);
    Function GetStartST : TDateAndTime;
    Procedure SetStartST(value : TDateAndTime);
    Procedure SetEnd_(value : TFhirInstant);
    Function GetEnd_ST : TDateAndTime;
    Procedure SetEnd_ST(value : TDateAndTime);
    Procedure SetMinutesDuration(value : TFhirPositiveInt);
    Function GetMinutesDurationST : String;
    Procedure SetMinutesDurationST(value : String);
    function GetSlotList : TFhirReferenceList{TFhirSlot};
    function GetHasSlotList : Boolean;
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
    function GetParticipantList : TFhirAppointmentParticipantList;
    function GetHasParticipantList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAppointment; overload;
    function Clone : TFhirAppointment; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this appointment concern that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      The overall status of the Appointment. Each of the participants has their own participation status which indicates their involvement in the process, however this status indicates the shared status.
    }
    property status : TFhirAppointmentstatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member type_
      Typed access to The type of appointment that is being booked (This may also be associated with participants for location, and/or a HealthcareService). (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The type of appointment that is being booked (This may also be associated with participants for location, and/or a HealthcareService).
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member reason
      Typed access to The reason that this appointment is being scheduled. This is more clinical than administrative. (defined for API consistency)
    }
    property reason : TFhirCodeableConcept read FReason write SetReason;
    {@member reasonElement
      The reason that this appointment is being scheduled. This is more clinical than administrative.
    }
    property reasonElement : TFhirCodeableConcept read FReason write SetReason;

    {@member priority
      Typed access to The priority of the appointment. Can be used to make informed decisions if needing to re-prioritize appointments. (The iCal Standard specifies 0 as undefined, 1 as highest, 9 as lowest priority).
    }
    property priority : String read GetPriorityST write SetPriorityST;
    {@member priorityElement
      The priority of the appointment. Can be used to make informed decisions if needing to re-prioritize appointments. (The iCal Standard specifies 0 as undefined, 1 as highest, 9 as lowest priority).
    }
    property priorityElement : TFhirUnsignedInt read FPriority write SetPriority;

    {@member description
      Typed access to The brief description of the appointment as would be shown on a subject line in a meeting request, or appointment list. Detailed or expanded information should be put in the comment field.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      The brief description of the appointment as would be shown on a subject line in a meeting request, or appointment list. Detailed or expanded information should be put in the comment field.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member start
      Typed access to Date/Time that the appointment is to take place.
    }
    property start : TDateAndTime read GetStartST write SetStartST;
    {@member startElement
      Date/Time that the appointment is to take place.
    }
    property startElement : TFhirInstant read FStart write SetStart;

    {@member end_
      Typed access to Date/Time that the appointment is to conclude.
    }
    property end_ : TDateAndTime read GetEnd_ST write SetEnd_ST;
    {@member end_Element
      Date/Time that the appointment is to conclude.
    }
    property end_Element : TFhirInstant read FEnd_ write SetEnd_;

    {@member minutesDuration
      Typed access to Number of minutes that the appointment is to take. This can be less than the duration between the start and end times (where actual time of appointment is only an estimate or is a planned appointment request).
    }
    property minutesDuration : String read GetMinutesDurationST write SetMinutesDurationST;
    {@member minutesDurationElement
      Number of minutes that the appointment is to take. This can be less than the duration between the start and end times (where actual time of appointment is only an estimate or is a planned appointment request).
    }
    property minutesDurationElement : TFhirPositiveInt read FMinutesDuration write SetMinutesDuration;

    {@member slotList
      The slot that this appointment is filling. If provided then the schedule will not be provided as slots are not recursive, and the start/end values MUST be the same as from the slot.
    }
    property slotList : TFhirReferenceList{TFhirSlot} read GetSlotList;
    property hasSlotList : boolean read GetHasSlotList;

    {@member comment
      Typed access to Additional comments about the appointment.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      Additional comments about the appointment.
    }
    property commentElement : TFhirString read FComment write SetComment;

    {@member participantList
      List of participants involved in the appointment.
    }
    property participantList : TFhirAppointmentParticipantList read GetParticipantList;
    property hasParticipantList : boolean read GetHasParticipantList;

  end;


  TFhirAppointmentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAppointmentList;
    function GetCurrent : TFhirAppointment;
  public
    Constructor Create(list : TFhirAppointmentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAppointment read GetCurrent;
  end;


  {@Class TFhirAppointmentList
    A list of FhirAppointment
  }
  {!.Net HL7Connect.Fhir.AppointmentList}
  TFhirAppointmentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAppointment;
    procedure SetItemN(index : Integer; value : TFhirAppointment);
  public
    {!script hide}
    function Link : TFhirAppointmentList; Overload;
    function Clone : TFhirAppointmentList; Overload;
    function GetEnumerator : TFhirAppointmentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAppointment to the end of the list.
    }
    function Append : TFhirAppointment;

    
    {@member AddItem
      Add an already existing FhirAppointment to the end of the list.
    }
    procedure AddItem(value : TFhirAppointment); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAppointment) : Integer;
    

    {@member Insert
      Insert FhirAppointment before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAppointment;
    

    {@member InsertItem
       Insert an existing FhirAppointment before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAppointment);
    
    {@member Item
       Get the iIndexth FhirAppointment. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAppointment. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAppointment);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAppointment;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAppointments[index : Integer] : TFhirAppointment read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAppointmentResponse : TFhirDomainResource
    A reply to an appointment request for a patient and/or practitioner(s), such as a confirmation or rejection.
  }
  {!.Net HL7Connect.Fhir.AppointmentResponse}
  TFhirAppointmentResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FAppointment : TFhirReference{TFhirAppointment};
    FStart : TFhirInstant;
    FEnd_ : TFhirInstant;
    FparticipantTypeList : TFhirCodeableConceptList;
    FActor : TFhirReference{Resource};
    FParticipantStatus : TFhirEnum;
    FComment : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetAppointment(value : TFhirReference{TFhirAppointment});
    Procedure SetStart(value : TFhirInstant);
    Function GetStartST : TDateAndTime;
    Procedure SetStartST(value : TDateAndTime);
    Procedure SetEnd_(value : TFhirInstant);
    Function GetEnd_ST : TDateAndTime;
    Procedure SetEnd_ST(value : TDateAndTime);
    function GetParticipantTypeList : TFhirCodeableConceptList;
    function GetHasParticipantTypeList : Boolean;
    Procedure SetActor(value : TFhirReference{Resource});
    Procedure SetParticipantStatus(value : TFhirEnum);
    Function GetParticipantStatusST : TFhirParticipantstatus;
    Procedure SetParticipantStatusST(value : TFhirParticipantstatus);
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAppointmentResponse; overload;
    function Clone : TFhirAppointmentResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this appointment response concern that are defined by business processes and/ or used to refer to it when a direct URL reference to the resource itself is not appropriate.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member appointment
      Typed access to Appointment that this response is replying to. (defined for API consistency)
    }
    property appointment : TFhirReference{TFhirAppointment} read FAppointment write SetAppointment;
    {@member appointmentElement
      Appointment that this response is replying to.
    }
    property appointmentElement : TFhirReference{TFhirAppointment} read FAppointment write SetAppointment;

    {@member start
      Typed access to Date/Time that the appointment is to take place, or requested new start time.
    }
    property start : TDateAndTime read GetStartST write SetStartST;
    {@member startElement
      Date/Time that the appointment is to take place, or requested new start time.
    }
    property startElement : TFhirInstant read FStart write SetStart;

    {@member end_
      Typed access to This may be either the same as the appointment request to confirm the details of the appointment, or alternately a new time to request a re-negotiation of the end time.
    }
    property end_ : TDateAndTime read GetEnd_ST write SetEnd_ST;
    {@member end_Element
      This may be either the same as the appointment request to confirm the details of the appointment, or alternately a new time to request a re-negotiation of the end time.
    }
    property end_Element : TFhirInstant read FEnd_ write SetEnd_;

    {@member participantTypeList
      Role of participant in the appointment.
    }
    property participantTypeList : TFhirCodeableConceptList read GetParticipantTypeList;
    property hasParticipantTypeList : boolean read GetHasParticipantTypeList;

    {@member actor
      Typed access to A Person, Location/HealthcareService or Device that is participating in the appointment. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      A Person, Location/HealthcareService or Device that is participating in the appointment.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

    {@member participantStatus
      Participation status of the participant. When the status is declined or tentative if the start/end times are different to the appointment, then these times should be interpreted as a requested time change. When the status is accepted, the times can either be the time of the appointment (as a confirmation of the time) or can be empty.
    }
    property participantStatus : TFhirParticipantstatus read GetParticipantStatusST write SetParticipantStatusST;
    property participantStatusElement : TFhirEnum read FParticipantStatus write SetParticipantStatus;

    {@member comment
      Typed access to Additional comments about the appointment.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      Additional comments about the appointment.
    }
    property commentElement : TFhirString read FComment write SetComment;

  end;


  TFhirAppointmentResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAppointmentResponseList;
    function GetCurrent : TFhirAppointmentResponse;
  public
    Constructor Create(list : TFhirAppointmentResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAppointmentResponse read GetCurrent;
  end;


  {@Class TFhirAppointmentResponseList
    A list of FhirAppointmentResponse
  }
  {!.Net HL7Connect.Fhir.AppointmentResponseList}
  TFhirAppointmentResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAppointmentResponse;
    procedure SetItemN(index : Integer; value : TFhirAppointmentResponse);
  public
    {!script hide}
    function Link : TFhirAppointmentResponseList; Overload;
    function Clone : TFhirAppointmentResponseList; Overload;
    function GetEnumerator : TFhirAppointmentResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAppointmentResponse to the end of the list.
    }
    function Append : TFhirAppointmentResponse;

    
    {@member AddItem
      Add an already existing FhirAppointmentResponse to the end of the list.
    }
    procedure AddItem(value : TFhirAppointmentResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAppointmentResponse) : Integer;
    

    {@member Insert
      Insert FhirAppointmentResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAppointmentResponse;
    

    {@member InsertItem
       Insert an existing FhirAppointmentResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAppointmentResponse);
    
    {@member Item
       Get the iIndexth FhirAppointmentResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAppointmentResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAppointmentResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAppointmentResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAppointmentResponses[index : Integer] : TFhirAppointmentResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventEvent : TFhirElement
    Identifies the name, action type, time, and disposition of the audited event.
  }
  {!.Net HL7Connect.Fhir.AuditEventEvent}
  TFhirAuditEventEvent = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FsubtypeList : TFhirCodingList;
    FAction : TFhirEnum;
    FDateTime : TFhirInstant;
    FOutcome : TFhirEnum;
    FOutcomeDesc : TFhirString;
    FpurposeOfEventList : TFhirCodingList;
    Procedure SetType_(value : TFhirCoding);
    function GetSubtypeList : TFhirCodingList;
    function GetHasSubtypeList : Boolean;
    Procedure SetAction(value : TFhirEnum);
    Function GetActionST : TFhirAuditEventAction;
    Procedure SetActionST(value : TFhirAuditEventAction);
    Procedure SetDateTime(value : TFhirInstant);
    Function GetDateTimeST : TDateAndTime;
    Procedure SetDateTimeST(value : TDateAndTime);
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirAuditEventOutcome;
    Procedure SetOutcomeST(value : TFhirAuditEventOutcome);
    Procedure SetOutcomeDesc(value : TFhirString);
    Function GetOutcomeDescST : String;
    Procedure SetOutcomeDescST(value : String);
    function GetPurposeOfEventList : TFhirCodingList;
    function GetHasPurposeOfEventList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventEvent; overload;
    function Clone : TFhirAuditEventEvent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Identifier for a family of the event.  For example, a menu item, program, rule, policy, function code, application name or URL. It identifies the performed function. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      Identifier for a family of the event.  For example, a menu item, program, rule, policy, function code, application name or URL. It identifies the performed function.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member subtypeList
      Identifier for the category of event.
    }
    property subtypeList : TFhirCodingList read GetSubtypeList;
    property hasSubtypeList : boolean read GetHasSubtypeList;

    {@member action
      Indicator for type of action performed during the event that generated the audit.
    }
    property action : TFhirAuditEventAction read GetActionST write SetActionST;
    property actionElement : TFhirEnum read FAction write SetAction;

    {@member dateTime
      Typed access to The time when the event occurred on the source.
    }
    property dateTime : TDateAndTime read GetDateTimeST write SetDateTimeST;
    {@member dateTimeElement
      The time when the event occurred on the source.
    }
    property dateTimeElement : TFhirInstant read FDateTime write SetDateTime;

    {@member outcome
      Indicates whether the event succeeded or failed.
    }
    property outcome : TFhirAuditEventOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member outcomeDesc
      Typed access to A free text description of the outcome of the event.
    }
    property outcomeDesc : String read GetOutcomeDescST write SetOutcomeDescST;
    {@member outcomeDescElement
      A free text description of the outcome of the event.
    }
    property outcomeDescElement : TFhirString read FOutcomeDesc write SetOutcomeDesc;

    {@member purposeOfEventList
      The purposeOfUse (reason) that was used during the event being recorded.
    }
    property purposeOfEventList : TFhirCodingList read GetPurposeOfEventList;
    property hasPurposeOfEventList : boolean read GetHasPurposeOfEventList;

  end;


  TFhirAuditEventEventListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventEventList;
    function GetCurrent : TFhirAuditEventEvent;
  public
    Constructor Create(list : TFhirAuditEventEventList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventEvent read GetCurrent;
  end;


  {@Class TFhirAuditEventEventList
    A list of FhirAuditEventEvent
  }
  {!.Net HL7Connect.Fhir.AuditEventEventList}
  TFhirAuditEventEventList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventEvent;
    procedure SetItemN(index : Integer; value : TFhirAuditEventEvent);
  public
    {!script hide}
    function Link : TFhirAuditEventEventList; Overload;
    function Clone : TFhirAuditEventEventList; Overload;
    function GetEnumerator : TFhirAuditEventEventListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventEvent to the end of the list.
    }
    function Append : TFhirAuditEventEvent;

    
    {@member AddItem
      Add an already existing FhirAuditEventEvent to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventEvent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventEvent) : Integer;
    

    {@member Insert
      Insert FhirAuditEventEvent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventEvent;
    

    {@member InsertItem
       Insert an existing FhirAuditEventEvent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventEvent);
    
    {@member Item
       Get the iIndexth FhirAuditEventEvent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventEvent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventEvent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventEvent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventEvents[index : Integer] : TFhirAuditEventEvent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventParticipant : TFhirElement
    A person, a hardware device or software process.
  }
  {!.Net HL7Connect.Fhir.AuditEventParticipant}
  TFhirAuditEventParticipant = class (TFhirBackboneElement)
  private
    FroleList : TFhirCodeableConceptList;
    FReference : TFhirReference{Resource};
    FUserId : TFhirIdentifier;
    FAltId : TFhirString;
    FName : TFhirString;
    FRequestor : TFhirBoolean;
    FLocation : TFhirReference{TFhirLocation};
    FpolicyList : TFhirUriList;
    FMedia : TFhirCoding;
    FNetwork : TFhirAuditEventParticipantNetwork;
    FpurposeOfUseList : TFhirCodingList;
    function GetRoleList : TFhirCodeableConceptList;
    function GetHasRoleList : Boolean;
    Procedure SetReference(value : TFhirReference{Resource});
    Procedure SetUserId(value : TFhirIdentifier);
    Procedure SetAltId(value : TFhirString);
    Function GetAltIdST : String;
    Procedure SetAltIdST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetRequestor(value : TFhirBoolean);
    Function GetRequestorST : Boolean;
    Procedure SetRequestorST(value : Boolean);
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    function GetPolicyList : TFhirUriList;
    function GetHasPolicyList : Boolean;
    Procedure SetMedia(value : TFhirCoding);
    Procedure SetNetwork(value : TFhirAuditEventParticipantNetwork);
    function GetPurposeOfUseList : TFhirCodingList;
    function GetHasPurposeOfUseList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventParticipant; overload;
    function Clone : TFhirAuditEventParticipant; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member roleList
      Specification of the role(s) the user plays when performing the event. Usually the codes used in this element are local codes defined by the role-based access control security system used in the local context.
    }
    property roleList : TFhirCodeableConceptList read GetRoleList;
    property hasRoleList : boolean read GetHasRoleList;

    {@member reference
      Typed access to Direct reference to a resource that identifies the participant. (defined for API consistency)
    }
    property reference : TFhirReference{Resource} read FReference write SetReference;
    {@member referenceElement
      Direct reference to a resource that identifies the participant.
    }
    property referenceElement : TFhirReference{Resource} read FReference write SetReference;

    {@member userId
      Typed access to Unique identifier for the user actively participating in the event. (defined for API consistency)
    }
    property userId : TFhirIdentifier read FUserId write SetUserId;
    {@member userIdElement
      Unique identifier for the user actively participating in the event.
    }
    property userIdElement : TFhirIdentifier read FUserId write SetUserId;

    {@member altId
      Typed access to Alternative Participant Identifier. For a human, this should be a user identifier text string from authentication system. This identifier would be one known to a common authentication system (e.g. single sign-on), if available.
    }
    property altId : String read GetAltIdST write SetAltIdST;
    {@member altIdElement
      Alternative Participant Identifier. For a human, this should be a user identifier text string from authentication system. This identifier would be one known to a common authentication system (e.g. single sign-on), if available.
    }
    property altIdElement : TFhirString read FAltId write SetAltId;

    {@member name
      Typed access to Human-meaningful name for the user.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Human-meaningful name for the user.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member requestor
      Typed access to Indicator that the user is or is not the requestor, or initiator, for the event being audited.
    }
    property requestor : Boolean read GetRequestorST write SetRequestorST;
    {@member requestorElement
      Indicator that the user is or is not the requestor, or initiator, for the event being audited.
    }
    property requestorElement : TFhirBoolean read FRequestor write SetRequestor;

    {@member location
      Typed access to Where the event occurred. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      Where the event occurred.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member policyList
      The policy or plan that authorized the activity being recorded. Typically, a single activity may have multiple applicable policies, such as patient consent, guarantor funding, etc. The policy would also indicate the security token used.
    }
    property policyList : TFhirUriList read GetPolicyList;
    property hasPolicyList : boolean read GetHasPolicyList;

    {@member media
      Typed access to Type of media involved. Used when the event is about exporting/importing onto media. (defined for API consistency)
    }
    property media : TFhirCoding read FMedia write SetMedia;
    {@member mediaElement
      Type of media involved. Used when the event is about exporting/importing onto media.
    }
    property mediaElement : TFhirCoding read FMedia write SetMedia;

    {@member network
      Typed access to Logical network location for application activity, if the activity has a network location. (defined for API consistency)
    }
    property network : TFhirAuditEventParticipantNetwork read FNetwork write SetNetwork;
    {@member networkElement
      Logical network location for application activity, if the activity has a network location.
    }
    property networkElement : TFhirAuditEventParticipantNetwork read FNetwork write SetNetwork;

    {@member purposeOfUseList
      The reason (purpose of use), specific to this participant, that was used during the event being recorded.
    }
    property purposeOfUseList : TFhirCodingList read GetPurposeOfUseList;
    property hasPurposeOfUseList : boolean read GetHasPurposeOfUseList;

  end;


  TFhirAuditEventParticipantListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventParticipantList;
    function GetCurrent : TFhirAuditEventParticipant;
  public
    Constructor Create(list : TFhirAuditEventParticipantList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventParticipant read GetCurrent;
  end;


  {@Class TFhirAuditEventParticipantList
    A list of FhirAuditEventParticipant
  }
  {!.Net HL7Connect.Fhir.AuditEventParticipantList}
  TFhirAuditEventParticipantList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventParticipant;
    procedure SetItemN(index : Integer; value : TFhirAuditEventParticipant);
  public
    {!script hide}
    function Link : TFhirAuditEventParticipantList; Overload;
    function Clone : TFhirAuditEventParticipantList; Overload;
    function GetEnumerator : TFhirAuditEventParticipantListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventParticipant to the end of the list.
    }
    function Append : TFhirAuditEventParticipant;

    
    {@member AddItem
      Add an already existing FhirAuditEventParticipant to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventParticipant); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventParticipant) : Integer;
    

    {@member Insert
      Insert FhirAuditEventParticipant before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventParticipant;
    

    {@member InsertItem
       Insert an existing FhirAuditEventParticipant before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventParticipant);
    
    {@member Item
       Get the iIndexth FhirAuditEventParticipant. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventParticipant. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventParticipant);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventParticipant;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventParticipants[index : Integer] : TFhirAuditEventParticipant read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventParticipantNetwork : TFhirElement
    Logical network location for application activity, if the activity has a network location.
  }
  {!.Net HL7Connect.Fhir.AuditEventParticipantNetwork}
  TFhirAuditEventParticipantNetwork = class (TFhirBackboneElement)
  private
    FAddress : TFhirString;
    FType_ : TFhirEnum;
    Procedure SetAddress(value : TFhirString);
    Function GetAddressST : String;
    Procedure SetAddressST(value : String);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirNetworkType;
    Procedure SetType_ST(value : TFhirNetworkType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventParticipantNetwork; overload;
    function Clone : TFhirAuditEventParticipantNetwork; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member address
      Typed access to An identifier for the network access point of the user device for the audit event.
    }
    property address : String read GetAddressST write SetAddressST;
    {@member addressElement
      An identifier for the network access point of the user device for the audit event.
    }
    property addressElement : TFhirString read FAddress write SetAddress;

    {@member type_
      An identifier for the type of network access point that originated the audit event.
    }
    property type_ : TFhirNetworkType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

  end;


  TFhirAuditEventParticipantNetworkListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventParticipantNetworkList;
    function GetCurrent : TFhirAuditEventParticipantNetwork;
  public
    Constructor Create(list : TFhirAuditEventParticipantNetworkList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventParticipantNetwork read GetCurrent;
  end;


  {@Class TFhirAuditEventParticipantNetworkList
    A list of FhirAuditEventParticipantNetwork
  }
  {!.Net HL7Connect.Fhir.AuditEventParticipantNetworkList}
  TFhirAuditEventParticipantNetworkList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventParticipantNetwork;
    procedure SetItemN(index : Integer; value : TFhirAuditEventParticipantNetwork);
  public
    {!script hide}
    function Link : TFhirAuditEventParticipantNetworkList; Overload;
    function Clone : TFhirAuditEventParticipantNetworkList; Overload;
    function GetEnumerator : TFhirAuditEventParticipantNetworkListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventParticipantNetwork to the end of the list.
    }
    function Append : TFhirAuditEventParticipantNetwork;

    
    {@member AddItem
      Add an already existing FhirAuditEventParticipantNetwork to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventParticipantNetwork); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventParticipantNetwork) : Integer;
    

    {@member Insert
      Insert FhirAuditEventParticipantNetwork before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventParticipantNetwork;
    

    {@member InsertItem
       Insert an existing FhirAuditEventParticipantNetwork before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventParticipantNetwork);
    
    {@member Item
       Get the iIndexth FhirAuditEventParticipantNetwork. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventParticipantNetwork. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventParticipantNetwork);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventParticipantNetwork;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventParticipantNetworks[index : Integer] : TFhirAuditEventParticipantNetwork read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventSource : TFhirElement
    Application systems and processes.
  }
  {!.Net HL7Connect.Fhir.AuditEventSource}
  TFhirAuditEventSource = class (TFhirBackboneElement)
  private
    FSite : TFhirString;
    FIdentifier : TFhirIdentifier;
    Ftype_List : TFhirCodingList;
    Procedure SetSite(value : TFhirString);
    Function GetSiteST : String;
    Procedure SetSiteST(value : String);
    Procedure SetIdentifier(value : TFhirIdentifier);
    function GetType_List : TFhirCodingList;
    function GetHasType_List : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventSource; overload;
    function Clone : TFhirAuditEventSource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member site
      Typed access to Logical source location within the healthcare enterprise network.  For example, a hospital or other provider location within a multi-entity provider group.
    }
    property site : String read GetSiteST write SetSiteST;
    {@member siteElement
      Logical source location within the healthcare enterprise network.  For example, a hospital or other provider location within a multi-entity provider group.
    }
    property siteElement : TFhirString read FSite write SetSite;

    {@member identifier
      Typed access to Identifier of the source where the event was detected. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifier of the source where the event was detected.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member type_List
      Code specifying the type of source where event originated.
    }
    property type_List : TFhirCodingList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

  end;


  TFhirAuditEventSourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventSourceList;
    function GetCurrent : TFhirAuditEventSource;
  public
    Constructor Create(list : TFhirAuditEventSourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventSource read GetCurrent;
  end;


  {@Class TFhirAuditEventSourceList
    A list of FhirAuditEventSource
  }
  {!.Net HL7Connect.Fhir.AuditEventSourceList}
  TFhirAuditEventSourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventSource;
    procedure SetItemN(index : Integer; value : TFhirAuditEventSource);
  public
    {!script hide}
    function Link : TFhirAuditEventSourceList; Overload;
    function Clone : TFhirAuditEventSourceList; Overload;
    function GetEnumerator : TFhirAuditEventSourceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventSource to the end of the list.
    }
    function Append : TFhirAuditEventSource;

    
    {@member AddItem
      Add an already existing FhirAuditEventSource to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventSource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventSource) : Integer;
    

    {@member Insert
      Insert FhirAuditEventSource before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventSource;
    

    {@member InsertItem
       Insert an existing FhirAuditEventSource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventSource);
    
    {@member Item
       Get the iIndexth FhirAuditEventSource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventSource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventSource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventSource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventSources[index : Integer] : TFhirAuditEventSource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventObject : TFhirElement
    Specific instances of data or objects that have been accessed.
  }
  {!.Net HL7Connect.Fhir.AuditEventObject}
  TFhirAuditEventObject = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirIdentifier;
    FReference : TFhirReference{TFhirReference};
    FType_ : TFhirCoding;
    FRole : TFhirCoding;
    FLifecycle : TFhirCoding;
    FsecurityLabelList : TFhirCodingList;
    FName : TFhirString;
    FDescription : TFhirString;
    FQuery : TFhirBase64Binary;
    FdetailList : TFhirAuditEventObjectDetailList;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetReference(value : TFhirReference{TFhirReference});
    Procedure SetType_(value : TFhirCoding);
    Procedure SetRole(value : TFhirCoding);
    Procedure SetLifecycle(value : TFhirCoding);
    function GetSecurityLabelList : TFhirCodingList;
    function GetHasSecurityLabelList : Boolean;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetQuery(value : TFhirBase64Binary);
    Function GetQueryST : TBytes;
    Procedure SetQueryST(value : TBytes);
    function GetDetailList : TFhirAuditEventObjectDetailList;
    function GetHasDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventObject; overload;
    function Clone : TFhirAuditEventObject; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to Identifies a specific instance of the participant object. The reference should always be version specific. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifies a specific instance of the participant object. The reference should always be version specific.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member reference
      Typed access to Identifies a specific instance of the participant object. The reference should always be version specific. (defined for API consistency)
    }
    property reference : TFhirReference{TFhirReference} read FReference write SetReference;
    {@member referenceElement
      Identifies a specific instance of the participant object. The reference should always be version specific.
    }
    property referenceElement : TFhirReference{TFhirReference} read FReference write SetReference;

    {@member type_
      Typed access to The type of the object that was involved in this audit event. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of the object that was involved in this audit event.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member role
      Typed access to Code representing the functional application role of Participant Object being audited. (defined for API consistency)
    }
    property role : TFhirCoding read FRole write SetRole;
    {@member roleElement
      Code representing the functional application role of Participant Object being audited.
    }
    property roleElement : TFhirCoding read FRole write SetRole;

    {@member lifecycle
      Typed access to Identifier for the data life-cycle stage for the participant object. (defined for API consistency)
    }
    property lifecycle : TFhirCoding read FLifecycle write SetLifecycle;
    {@member lifecycleElement
      Identifier for the data life-cycle stage for the participant object.
    }
    property lifecycleElement : TFhirCoding read FLifecycle write SetLifecycle;

    {@member securityLabelList
      Denotes security labels for the identified object.
    }
    property securityLabelList : TFhirCodingList read GetSecurityLabelList;
    property hasSecurityLabelList : boolean read GetHasSecurityLabelList;

    {@member name
      Typed access to An instance-specific descriptor of the Participant Object ID audited, such as a person's name.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      An instance-specific descriptor of the Participant Object ID audited, such as a person's name.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member description
      Typed access to Text that describes the object in more detail.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Text that describes the object in more detail.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member query
      Typed access to The actual query for a query-type participant object.
    }
    property query : TBytes read GetQueryST write SetQueryST;
    {@member queryElement
      The actual query for a query-type participant object.
    }
    property queryElement : TFhirBase64Binary read FQuery write SetQuery;

    {@member detailList
      Additional Information about the Object.
    }
    property detailList : TFhirAuditEventObjectDetailList read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirAuditEventObjectListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventObjectList;
    function GetCurrent : TFhirAuditEventObject;
  public
    Constructor Create(list : TFhirAuditEventObjectList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventObject read GetCurrent;
  end;


  {@Class TFhirAuditEventObjectList
    A list of FhirAuditEventObject
  }
  {!.Net HL7Connect.Fhir.AuditEventObjectList}
  TFhirAuditEventObjectList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventObject;
    procedure SetItemN(index : Integer; value : TFhirAuditEventObject);
  public
    {!script hide}
    function Link : TFhirAuditEventObjectList; Overload;
    function Clone : TFhirAuditEventObjectList; Overload;
    function GetEnumerator : TFhirAuditEventObjectListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventObject to the end of the list.
    }
    function Append : TFhirAuditEventObject;

    
    {@member AddItem
      Add an already existing FhirAuditEventObject to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventObject); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventObject) : Integer;
    

    {@member Insert
      Insert FhirAuditEventObject before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventObject;
    

    {@member InsertItem
       Insert an existing FhirAuditEventObject before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventObject);
    
    {@member Item
       Get the iIndexth FhirAuditEventObject. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventObject. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventObject);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventObject;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventObjects[index : Integer] : TFhirAuditEventObject read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEventObjectDetail : TFhirElement
    Additional Information about the Object.
  }
  {!.Net HL7Connect.Fhir.AuditEventObjectDetail}
  TFhirAuditEventObjectDetail = class (TFhirBackboneElement)
  private
    FType_ : TFhirString;
    FValue : TFhirBase64Binary;
    Procedure SetType_(value : TFhirString);
    Function GetType_ST : String;
    Procedure SetType_ST(value : String);
    Procedure SetValue(value : TFhirBase64Binary);
    Function GetValueST : TBytes;
    Procedure SetValueST(value : TBytes);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEventObjectDetail; overload;
    function Clone : TFhirAuditEventObjectDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Name of the property.
    }
    property type_ : String read GetType_ST write SetType_ST;
    {@member type_Element
      Name of the property.
    }
    property type_Element : TFhirString read FType_ write SetType_;

    {@member value
      Typed access to Property value.
    }
    property value : TBytes read GetValueST write SetValueST;
    {@member valueElement
      Property value.
    }
    property valueElement : TFhirBase64Binary read FValue write SetValue;

  end;


  TFhirAuditEventObjectDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventObjectDetailList;
    function GetCurrent : TFhirAuditEventObjectDetail;
  public
    Constructor Create(list : TFhirAuditEventObjectDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEventObjectDetail read GetCurrent;
  end;


  {@Class TFhirAuditEventObjectDetailList
    A list of FhirAuditEventObjectDetail
  }
  {!.Net HL7Connect.Fhir.AuditEventObjectDetailList}
  TFhirAuditEventObjectDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEventObjectDetail;
    procedure SetItemN(index : Integer; value : TFhirAuditEventObjectDetail);
  public
    {!script hide}
    function Link : TFhirAuditEventObjectDetailList; Overload;
    function Clone : TFhirAuditEventObjectDetailList; Overload;
    function GetEnumerator : TFhirAuditEventObjectDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEventObjectDetail to the end of the list.
    }
    function Append : TFhirAuditEventObjectDetail;

    
    {@member AddItem
      Add an already existing FhirAuditEventObjectDetail to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEventObjectDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEventObjectDetail) : Integer;
    

    {@member Insert
      Insert FhirAuditEventObjectDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEventObjectDetail;
    

    {@member InsertItem
       Insert an existing FhirAuditEventObjectDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEventObjectDetail);
    
    {@member Item
       Get the iIndexth FhirAuditEventObjectDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEventObjectDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEventObjectDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEventObjectDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEventObjectDetails[index : Integer] : TFhirAuditEventObjectDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirAuditEvent : TFhirDomainResource
    A record of an event made for purposes of maintaining a security log. Typical uses include detection of intrusion attempts and monitoring for inappropriate usage.
  }
  {!.Net HL7Connect.Fhir.AuditEvent}
  TFhirAuditEvent = class (TFhirDomainResource)
  private
    FEvent : TFhirAuditEventEvent;
    FparticipantList : TFhirAuditEventParticipantList;
    FSource : TFhirAuditEventSource;
    Fobject_List : TFhirAuditEventObjectList;
    Procedure SetEvent(value : TFhirAuditEventEvent);
    function GetParticipantList : TFhirAuditEventParticipantList;
    function GetHasParticipantList : Boolean;
    Procedure SetSource(value : TFhirAuditEventSource);
    function GetObject_List : TFhirAuditEventObjectList;
    function GetHasObject_List : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirAuditEvent; overload;
    function Clone : TFhirAuditEvent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member event
      Typed access to Identifies the name, action type, time, and disposition of the audited event. (defined for API consistency)
    }
    property event : TFhirAuditEventEvent read FEvent write SetEvent;
    {@member eventElement
      Identifies the name, action type, time, and disposition of the audited event.
    }
    property eventElement : TFhirAuditEventEvent read FEvent write SetEvent;

    {@member participantList
      A person, a hardware device or software process.
    }
    property participantList : TFhirAuditEventParticipantList read GetParticipantList;
    property hasParticipantList : boolean read GetHasParticipantList;

    {@member source
      Typed access to Application systems and processes. (defined for API consistency)
    }
    property source : TFhirAuditEventSource read FSource write SetSource;
    {@member sourceElement
      Application systems and processes.
    }
    property sourceElement : TFhirAuditEventSource read FSource write SetSource;

    {@member object_List
      Specific instances of data or objects that have been accessed.
    }
    property object_List : TFhirAuditEventObjectList read GetObject_List;
    property hasObject_List : boolean read GetHasObject_List;

  end;


  TFhirAuditEventListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirAuditEventList;
    function GetCurrent : TFhirAuditEvent;
  public
    Constructor Create(list : TFhirAuditEventList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirAuditEvent read GetCurrent;
  end;


  {@Class TFhirAuditEventList
    A list of FhirAuditEvent
  }
  {!.Net HL7Connect.Fhir.AuditEventList}
  TFhirAuditEventList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirAuditEvent;
    procedure SetItemN(index : Integer; value : TFhirAuditEvent);
  public
    {!script hide}
    function Link : TFhirAuditEventList; Overload;
    function Clone : TFhirAuditEventList; Overload;
    function GetEnumerator : TFhirAuditEventListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirAuditEvent to the end of the list.
    }
    function Append : TFhirAuditEvent;

    
    {@member AddItem
      Add an already existing FhirAuditEvent to the end of the list.
    }
    procedure AddItem(value : TFhirAuditEvent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirAuditEvent) : Integer;
    

    {@member Insert
      Insert FhirAuditEvent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirAuditEvent;
    

    {@member InsertItem
       Insert an existing FhirAuditEvent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirAuditEvent);
    
    {@member Item
       Get the iIndexth FhirAuditEvent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirAuditEvent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirAuditEvent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirAuditEvent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirAuditEvents[index : Integer] : TFhirAuditEvent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBasic : TFhirDomainResource
    Basic is used for handling concepts not yet defined in FHIR, narrative-only resources that don't map to an existing resource, and custom resources not appropriate for inclusion in the FHIR specification.
  }
  {!.Net HL7Connect.Fhir.Basic}
  TFhirBasic = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FCode : TFhirCodeableConcept;
    FSubject : TFhirReference{TFhirReference};
    FAuthor : TFhirReference{Resource};
    FCreated : TFhirDate;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSubject(value : TFhirReference{TFhirReference});
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetCreated(value : TFhirDate);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBasic; overload;
    function Clone : TFhirBasic; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier assigned to the resource for business purposes, outside the context of FHIR.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member code
      Typed access to Identifies the 'type' of resource - equivalent to the resource name for other resources. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Identifies the 'type' of resource - equivalent to the resource name for other resources.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member subject
      Typed access to Identifies the patient, practitioner, device or any other resource that is the "focus" of this resource. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirReference} read FSubject write SetSubject;
    {@member subjectElement
      Identifies the patient, practitioner, device or any other resource that is the "focus" of this resource.
    }
    property subjectElement : TFhirReference{TFhirReference} read FSubject write SetSubject;

    {@member author
      Typed access to Indicates who was responsible for creating the resource instance. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      Indicates who was responsible for creating the resource instance.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member created
      Typed access to Identifies when the resource was first created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      Identifies when the resource was first created.
    }
    property createdElement : TFhirDate read FCreated write SetCreated;

  end;


  TFhirBasicListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBasicList;
    function GetCurrent : TFhirBasic;
  public
    Constructor Create(list : TFhirBasicList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBasic read GetCurrent;
  end;


  {@Class TFhirBasicList
    A list of FhirBasic
  }
  {!.Net HL7Connect.Fhir.BasicList}
  TFhirBasicList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBasic;
    procedure SetItemN(index : Integer; value : TFhirBasic);
  public
    {!script hide}
    function Link : TFhirBasicList; Overload;
    function Clone : TFhirBasicList; Overload;
    function GetEnumerator : TFhirBasicListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBasic to the end of the list.
    }
    function Append : TFhirBasic;

    
    {@member AddItem
      Add an already existing FhirBasic to the end of the list.
    }
    procedure AddItem(value : TFhirBasic); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBasic) : Integer;
    

    {@member Insert
      Insert FhirBasic before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBasic;
    

    {@member InsertItem
       Insert an existing FhirBasic before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBasic);
    
    {@member Item
       Get the iIndexth FhirBasic. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBasic. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBasic);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBasic;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBasics[index : Integer] : TFhirBasic read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBinary : TFhirResource
    A binary resource can contain any content, whether text, image, pdf, zip archive, etc.
  }
  {!.Net HL7Connect.Fhir.Binary}
  TFhirBinary = class (TFhirResource)
  private
    FContentType : TFhirCode;
    FContent : TFhirBase64Binary;
    Procedure SetContentType(value : TFhirCode);
    Function GetContentTypeST : String;
    Procedure SetContentTypeST(value : String);
    Procedure SetContent(value : TFhirBase64Binary);
    Function GetContentST : TBytes;
    Procedure SetContentST(value : TBytes);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBinary; overload;
    function Clone : TFhirBinary; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member contentType
      Typed access to MimeType of the binary content represented as a standard MimeType (BCP 13).
    }
    property contentType : String read GetContentTypeST write SetContentTypeST;
    {@member contentTypeElement
      MimeType of the binary content represented as a standard MimeType (BCP 13).
    }
    property contentTypeElement : TFhirCode read FContentType write SetContentType;

    {@member content
      Typed access to The actual content, base64 encoded.
    }
    property content : TBytes read GetContentST write SetContentST;
    {@member contentElement
      The actual content, base64 encoded.
    }
    property contentElement : TFhirBase64Binary read FContent write SetContent;

  end;


  TFhirBinaryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBinaryList;
    function GetCurrent : TFhirBinary;
  public
    Constructor Create(list : TFhirBinaryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBinary read GetCurrent;
  end;


  {@Class TFhirBinaryList
    A list of FhirBinary
  }
  {!.Net HL7Connect.Fhir.BinaryList}
  TFhirBinaryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBinary;
    procedure SetItemN(index : Integer; value : TFhirBinary);
  public
    {!script hide}
    function Link : TFhirBinaryList; Overload;
    function Clone : TFhirBinaryList; Overload;
    function GetEnumerator : TFhirBinaryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBinary to the end of the list.
    }
    function Append : TFhirBinary;

    
    {@member AddItem
      Add an already existing FhirBinary to the end of the list.
    }
    procedure AddItem(value : TFhirBinary); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBinary) : Integer;
    

    {@member Insert
      Insert FhirBinary before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBinary;
    

    {@member InsertItem
       Insert an existing FhirBinary before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBinary);
    
    {@member Item
       Get the iIndexth FhirBinary. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBinary. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBinary);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBinary;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBinaries[index : Integer] : TFhirBinary read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBodySite : TFhirDomainResource
    Record details about the anatomical location of a specimen or body part.  This resource may be used when a coded concept does not provide the necessary detail needed for the use case.
  }
  {!.Net HL7Connect.Fhir.BodySite}
  TFhirBodySite = class (TFhirDomainResource)
  private
    FPatient : TFhirReference{TFhirPatient};
    FidentifierList : TFhirIdentifierList;
    FCode : TFhirCodeableConcept;
    FmodifierList : TFhirCodeableConceptList;
    FDescription : TFhirString;
    FimageList : TFhirAttachmentList;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetModifierList : TFhirCodeableConceptList;
    function GetHasModifierList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetImageList : TFhirAttachmentList;
    function GetHasImageList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBodySite; overload;
    function Clone : TFhirBodySite; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member patient
      Typed access to The person to which the body site belongs. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The person to which the body site belongs.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member identifierList
      Identifier for this instance of the anatomical location.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member code
      Typed access to Named anatomical location - ideally coded where possible. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Named anatomical location - ideally coded where possible.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member modifierList
      Modifier to refine the anatomical location.  These include modifiers for laterality, relative location, directionality, number, and plane.
    }
    property modifierList : TFhirCodeableConceptList read GetModifierList;
    property hasModifierList : boolean read GetHasModifierList;

    {@member description
      Typed access to Description of anatomical location.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Description of anatomical location.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member imageList
      Image or images used to identify a location.
    }
    property imageList : TFhirAttachmentList read GetImageList;
    property hasImageList : boolean read GetHasImageList;

  end;


  TFhirBodySiteListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBodySiteList;
    function GetCurrent : TFhirBodySite;
  public
    Constructor Create(list : TFhirBodySiteList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBodySite read GetCurrent;
  end;


  {@Class TFhirBodySiteList
    A list of FhirBodySite
  }
  {!.Net HL7Connect.Fhir.BodySiteList}
  TFhirBodySiteList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBodySite;
    procedure SetItemN(index : Integer; value : TFhirBodySite);
  public
    {!script hide}
    function Link : TFhirBodySiteList; Overload;
    function Clone : TFhirBodySiteList; Overload;
    function GetEnumerator : TFhirBodySiteListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBodySite to the end of the list.
    }
    function Append : TFhirBodySite;

    
    {@member AddItem
      Add an already existing FhirBodySite to the end of the list.
    }
    procedure AddItem(value : TFhirBodySite); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBodySite) : Integer;
    

    {@member Insert
      Insert FhirBodySite before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBodySite;
    

    {@member InsertItem
       Insert an existing FhirBodySite before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBodySite);
    
    {@member Item
       Get the iIndexth FhirBodySite. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBodySite. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBodySite);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBodySite;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBodySites[index : Integer] : TFhirBodySite read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundleLink : TFhirElement
    A series of links that provide context to this bundle.
  }
  {!.Net HL7Connect.Fhir.BundleLink}
  TFhirBundleLink = class (TFhirBackboneElement)
  private
    FRelation : TFhirString;
    FUrl : TFhirUri;
    Procedure SetRelation(value : TFhirString);
    Function GetRelationST : String;
    Procedure SetRelationST(value : String);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundleLink; overload;
    function Clone : TFhirBundleLink; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member relation
      Typed access to A name which details the functional use for this link - see [[http://www.iana.org/assignments/link-relations/link-relations.xhtml]].
    }
    property relation : String read GetRelationST write SetRelationST;
    {@member relationElement
      A name which details the functional use for this link - see [[http://www.iana.org/assignments/link-relations/link-relations.xhtml]].
    }
    property relationElement : TFhirString read FRelation write SetRelation;

    {@member url
      Typed access to The reference details for the link.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      The reference details for the link.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

  end;


  TFhirBundleLinkListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleLinkList;
    function GetCurrent : TFhirBundleLink;
  public
    Constructor Create(list : TFhirBundleLinkList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundleLink read GetCurrent;
  end;


  {@Class TFhirBundleLinkList
    A list of FhirBundleLink
  }
  {!.Net HL7Connect.Fhir.BundleLinkList}
  TFhirBundleLinkList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundleLink;
    procedure SetItemN(index : Integer; value : TFhirBundleLink);
  public
    {!script hide}
    function Link : TFhirBundleLinkList; Overload;
    function Clone : TFhirBundleLinkList; Overload;
    function GetEnumerator : TFhirBundleLinkListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundleLink to the end of the list.
    }
    function Append : TFhirBundleLink;

    
    {@member AddItem
      Add an already existing FhirBundleLink to the end of the list.
    }
    procedure AddItem(value : TFhirBundleLink); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundleLink) : Integer;
    

    {@member Insert
      Insert FhirBundleLink before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundleLink;
    

    {@member InsertItem
       Insert an existing FhirBundleLink before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundleLink);
    
    {@member Item
       Get the iIndexth FhirBundleLink. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundleLink. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundleLink);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundleLink;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundleLinks[index : Integer] : TFhirBundleLink read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundleEntry : TFhirElement
    An entry in a bundle resource - will either contain a resource, or information about a resource (transactions and history only).
  }
  {!.Net HL7Connect.Fhir.BundleEntry}
  TFhirBundleEntry = class (TFhirBackboneElement)
  private
    Flink_List : TFhirBundleLinkList;
    FFullUrl : TFhirUri;
    FResource : TFhirResource;
    FSearch : TFhirBundleEntrySearch;
    FRequest : TFhirBundleEntryRequest;
    FResponse : TFhirBundleEntryResponse;
    function GetLink_List : TFhirBundleLinkList;
    function GetHasLink_List : Boolean;
    Procedure SetFullUrl(value : TFhirUri);
    Function GetFullUrlST : String;
    Procedure SetFullUrlST(value : String);
    Procedure SetResource(value : TFhirResource);
    Procedure SetSearch(value : TFhirBundleEntrySearch);
    Procedure SetRequest(value : TFhirBundleEntryRequest);
    Procedure SetResponse(value : TFhirBundleEntryResponse);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundleEntry; overload;
    function Clone : TFhirBundleEntry; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member link_List
      A series of links that provide context to this entry.
    }
    property link_List : TFhirBundleLinkList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

    {@member fullUrl
      Typed access to The Absolute URL for the resource. This must be provided for all resources. The fullUrl SHALL not disagree with the id in the resource. The fullUrl is a version independent reference to the resource.
    }
    property fullUrl : String read GetFullUrlST write SetFullUrlST;
    {@member fullUrlElement
      The Absolute URL for the resource. This must be provided for all resources. The fullUrl SHALL not disagree with the id in the resource. The fullUrl is a version independent reference to the resource.
    }
    property fullUrlElement : TFhirUri read FFullUrl write SetFullUrl;

    {@member resource
      Typed access to The Resources for the entry. (defined for API consistency)
    }
    property resource : TFhirResource read FResource write SetResource;
    {@member resourceElement
      The Resources for the entry.
    }
    property resourceElement : TFhirResource read FResource write SetResource;

    {@member search
      Typed access to Information about the search process that lead to the creation of this entry. (defined for API consistency)
    }
    property search : TFhirBundleEntrySearch read FSearch write SetSearch;
    {@member searchElement
      Information about the search process that lead to the creation of this entry.
    }
    property searchElement : TFhirBundleEntrySearch read FSearch write SetSearch;

    {@member request
      Typed access to Additional information about how this entry should be processed as part of a transaction. (defined for API consistency)
    }
    property request : TFhirBundleEntryRequest read FRequest write SetRequest;
    {@member requestElement
      Additional information about how this entry should be processed as part of a transaction.
    }
    property requestElement : TFhirBundleEntryRequest read FRequest write SetRequest;

    {@member response
      Typed access to Additional information about how this entry should be processed as part of a transaction. (defined for API consistency)
    }
    property response : TFhirBundleEntryResponse read FResponse write SetResponse;
    {@member responseElement
      Additional information about how this entry should be processed as part of a transaction.
    }
    property responseElement : TFhirBundleEntryResponse read FResponse write SetResponse;

  end;


  TFhirBundleEntryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleEntryList;
    function GetCurrent : TFhirBundleEntry;
  public
    Constructor Create(list : TFhirBundleEntryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundleEntry read GetCurrent;
  end;


  {@Class TFhirBundleEntryList
    A list of FhirBundleEntry
  }
  {!.Net HL7Connect.Fhir.BundleEntryList}
  TFhirBundleEntryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundleEntry;
    procedure SetItemN(index : Integer; value : TFhirBundleEntry);
  public
    {!script hide}
    function Link : TFhirBundleEntryList; Overload;
    function Clone : TFhirBundleEntryList; Overload;
    function GetEnumerator : TFhirBundleEntryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundleEntry to the end of the list.
    }
    function Append : TFhirBundleEntry;

    
    {@member AddItem
      Add an already existing FhirBundleEntry to the end of the list.
    }
    procedure AddItem(value : TFhirBundleEntry); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundleEntry) : Integer;
    

    {@member Insert
      Insert FhirBundleEntry before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundleEntry;
    

    {@member InsertItem
       Insert an existing FhirBundleEntry before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundleEntry);
    
    {@member Item
       Get the iIndexth FhirBundleEntry. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundleEntry. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundleEntry);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundleEntry;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundleEntries[index : Integer] : TFhirBundleEntry read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundleEntrySearch : TFhirElement
    Information about the search process that lead to the creation of this entry.
  }
  {!.Net HL7Connect.Fhir.BundleEntrySearch}
  TFhirBundleEntrySearch = class (TFhirBackboneElement)
  private
    FMode : TFhirEnum;
    FScore : TFhirDecimal;
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirSearchEntryMode;
    Procedure SetModeST(value : TFhirSearchEntryMode);
    Procedure SetScore(value : TFhirDecimal);
    Function GetScoreST : String;
    Procedure SetScoreST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundleEntrySearch; overload;
    function Clone : TFhirBundleEntrySearch; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member mode
      Why this entry is in the result set - whether it's included as a match or because of an _include requirement.
    }
    property mode : TFhirSearchEntryMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member score
      Typed access to When searching, the server's search ranking score for the entry.
    }
    property score : String read GetScoreST write SetScoreST;
    {@member scoreElement
      When searching, the server's search ranking score for the entry.
    }
    property scoreElement : TFhirDecimal read FScore write SetScore;

  end;


  TFhirBundleEntrySearchListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleEntrySearchList;
    function GetCurrent : TFhirBundleEntrySearch;
  public
    Constructor Create(list : TFhirBundleEntrySearchList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundleEntrySearch read GetCurrent;
  end;


  {@Class TFhirBundleEntrySearchList
    A list of FhirBundleEntrySearch
  }
  {!.Net HL7Connect.Fhir.BundleEntrySearchList}
  TFhirBundleEntrySearchList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundleEntrySearch;
    procedure SetItemN(index : Integer; value : TFhirBundleEntrySearch);
  public
    {!script hide}
    function Link : TFhirBundleEntrySearchList; Overload;
    function Clone : TFhirBundleEntrySearchList; Overload;
    function GetEnumerator : TFhirBundleEntrySearchListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundleEntrySearch to the end of the list.
    }
    function Append : TFhirBundleEntrySearch;

    
    {@member AddItem
      Add an already existing FhirBundleEntrySearch to the end of the list.
    }
    procedure AddItem(value : TFhirBundleEntrySearch); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundleEntrySearch) : Integer;
    

    {@member Insert
      Insert FhirBundleEntrySearch before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundleEntrySearch;
    

    {@member InsertItem
       Insert an existing FhirBundleEntrySearch before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundleEntrySearch);
    
    {@member Item
       Get the iIndexth FhirBundleEntrySearch. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundleEntrySearch. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundleEntrySearch);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundleEntrySearch;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundleEntrySearches[index : Integer] : TFhirBundleEntrySearch read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundleEntryRequest : TFhirElement
    Additional information about how this entry should be processed as part of a transaction.
  }
  {!.Net HL7Connect.Fhir.BundleEntryRequest}
  TFhirBundleEntryRequest = class (TFhirBackboneElement)
  private
    FMethod : TFhirEnum;
    FUrl : TFhirUri;
    FIfNoneMatch : TFhirString;
    FIfModifiedSince : TFhirInstant;
    FIfMatch : TFhirString;
    FIfNoneExist : TFhirString;
    Procedure SetMethod(value : TFhirEnum);
    Function GetMethodST : TFhirHttpVerb;
    Procedure SetMethodST(value : TFhirHttpVerb);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetIfNoneMatch(value : TFhirString);
    Function GetIfNoneMatchST : String;
    Procedure SetIfNoneMatchST(value : String);
    Procedure SetIfModifiedSince(value : TFhirInstant);
    Function GetIfModifiedSinceST : TDateAndTime;
    Procedure SetIfModifiedSinceST(value : TDateAndTime);
    Procedure SetIfMatch(value : TFhirString);
    Function GetIfMatchST : String;
    Procedure SetIfMatchST(value : String);
    Procedure SetIfNoneExist(value : TFhirString);
    Function GetIfNoneExistST : String;
    Procedure SetIfNoneExistST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundleEntryRequest; overload;
    function Clone : TFhirBundleEntryRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member method
      The HTTP verb for this entry in either a update history, or a transaction/ transaction response.
    }
    property method : TFhirHttpVerb read GetMethodST write SetMethodST;
    property methodElement : TFhirEnum read FMethod write SetMethod;

    {@member url
      Typed access to The URL for this entry, relative to the root (the address to which the request is posted).
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      The URL for this entry, relative to the root (the address to which the request is posted).
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member ifNoneMatch
      Typed access to If the ETag values match, return a 304 Not modified status. See the API documentation for ["Conditional Read"](http.html#cread).
    }
    property ifNoneMatch : String read GetIfNoneMatchST write SetIfNoneMatchST;
    {@member ifNoneMatchElement
      If the ETag values match, return a 304 Not modified status. See the API documentation for ["Conditional Read"](http.html#cread).
    }
    property ifNoneMatchElement : TFhirString read FIfNoneMatch write SetIfNoneMatch;

    {@member ifModifiedSince
      Typed access to Only perform the operation if the last updated date matches. See the API documentation for ["Conditional Read"](http.html#cread).
    }
    property ifModifiedSince : TDateAndTime read GetIfModifiedSinceST write SetIfModifiedSinceST;
    {@member ifModifiedSinceElement
      Only perform the operation if the last updated date matches. See the API documentation for ["Conditional Read"](http.html#cread).
    }
    property ifModifiedSinceElement : TFhirInstant read FIfModifiedSince write SetIfModifiedSince;

    {@member ifMatch
      Typed access to Only perform the operation if the Etag value matches. For more information, see the API section ["Managing Resource Contention"](http.html#concurrency).
    }
    property ifMatch : String read GetIfMatchST write SetIfMatchST;
    {@member ifMatchElement
      Only perform the operation if the Etag value matches. For more information, see the API section ["Managing Resource Contention"](http.html#concurrency).
    }
    property ifMatchElement : TFhirString read FIfMatch write SetIfMatch;

    {@member ifNoneExist
      Typed access to Instruct the server not to perform the create if a specified resource already exists. For further information, see the API documentation for ["Conditional Create"](http.html#ccreate). This is just the query portion of the URL - what follows the "?" (not including the "?").
    }
    property ifNoneExist : String read GetIfNoneExistST write SetIfNoneExistST;
    {@member ifNoneExistElement
      Instruct the server not to perform the create if a specified resource already exists. For further information, see the API documentation for ["Conditional Create"](http.html#ccreate). This is just the query portion of the URL - what follows the "?" (not including the "?").
    }
    property ifNoneExistElement : TFhirString read FIfNoneExist write SetIfNoneExist;

  end;


  TFhirBundleEntryRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleEntryRequestList;
    function GetCurrent : TFhirBundleEntryRequest;
  public
    Constructor Create(list : TFhirBundleEntryRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundleEntryRequest read GetCurrent;
  end;


  {@Class TFhirBundleEntryRequestList
    A list of FhirBundleEntryRequest
  }
  {!.Net HL7Connect.Fhir.BundleEntryRequestList}
  TFhirBundleEntryRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundleEntryRequest;
    procedure SetItemN(index : Integer; value : TFhirBundleEntryRequest);
  public
    {!script hide}
    function Link : TFhirBundleEntryRequestList; Overload;
    function Clone : TFhirBundleEntryRequestList; Overload;
    function GetEnumerator : TFhirBundleEntryRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundleEntryRequest to the end of the list.
    }
    function Append : TFhirBundleEntryRequest;

    
    {@member AddItem
      Add an already existing FhirBundleEntryRequest to the end of the list.
    }
    procedure AddItem(value : TFhirBundleEntryRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundleEntryRequest) : Integer;
    

    {@member Insert
      Insert FhirBundleEntryRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundleEntryRequest;
    

    {@member InsertItem
       Insert an existing FhirBundleEntryRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundleEntryRequest);
    
    {@member Item
       Get the iIndexth FhirBundleEntryRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundleEntryRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundleEntryRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundleEntryRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundleEntryRequests[index : Integer] : TFhirBundleEntryRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundleEntryResponse : TFhirElement
    Additional information about how this entry should be processed as part of a transaction.
  }
  {!.Net HL7Connect.Fhir.BundleEntryResponse}
  TFhirBundleEntryResponse = class (TFhirBackboneElement)
  private
    FStatus : TFhirString;
    FLocation : TFhirUri;
    FEtag : TFhirString;
    FLastModified : TFhirInstant;
    Procedure SetStatus(value : TFhirString);
    Function GetStatusST : String;
    Procedure SetStatusST(value : String);
    Procedure SetLocation(value : TFhirUri);
    Function GetLocationST : String;
    Procedure SetLocationST(value : String);
    Procedure SetEtag(value : TFhirString);
    Function GetEtagST : String;
    Procedure SetEtagST(value : String);
    Procedure SetLastModified(value : TFhirInstant);
    Function GetLastModifiedST : TDateAndTime;
    Procedure SetLastModifiedST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundleEntryResponse; overload;
    function Clone : TFhirBundleEntryResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member status
      Typed access to The status code returned by processing this entry.
    }
    property status : String read GetStatusST write SetStatusST;
    {@member statusElement
      The status code returned by processing this entry.
    }
    property statusElement : TFhirString read FStatus write SetStatus;

    {@member location
      Typed access to The location header created by processing this operation.
    }
    property location : String read GetLocationST write SetLocationST;
    {@member locationElement
      The location header created by processing this operation.
    }
    property locationElement : TFhirUri read FLocation write SetLocation;

    {@member etag
      Typed access to The etag for the resource, it the operation for the entry produced a versioned resource.
    }
    property etag : String read GetEtagST write SetEtagST;
    {@member etagElement
      The etag for the resource, it the operation for the entry produced a versioned resource.
    }
    property etagElement : TFhirString read FEtag write SetEtag;

    {@member lastModified
      Typed access to The date/time that the resource was modified on the server.
    }
    property lastModified : TDateAndTime read GetLastModifiedST write SetLastModifiedST;
    {@member lastModifiedElement
      The date/time that the resource was modified on the server.
    }
    property lastModifiedElement : TFhirInstant read FLastModified write SetLastModified;

  end;


  TFhirBundleEntryResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleEntryResponseList;
    function GetCurrent : TFhirBundleEntryResponse;
  public
    Constructor Create(list : TFhirBundleEntryResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundleEntryResponse read GetCurrent;
  end;


  {@Class TFhirBundleEntryResponseList
    A list of FhirBundleEntryResponse
  }
  {!.Net HL7Connect.Fhir.BundleEntryResponseList}
  TFhirBundleEntryResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundleEntryResponse;
    procedure SetItemN(index : Integer; value : TFhirBundleEntryResponse);
  public
    {!script hide}
    function Link : TFhirBundleEntryResponseList; Overload;
    function Clone : TFhirBundleEntryResponseList; Overload;
    function GetEnumerator : TFhirBundleEntryResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundleEntryResponse to the end of the list.
    }
    function Append : TFhirBundleEntryResponse;

    
    {@member AddItem
      Add an already existing FhirBundleEntryResponse to the end of the list.
    }
    procedure AddItem(value : TFhirBundleEntryResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundleEntryResponse) : Integer;
    

    {@member Insert
      Insert FhirBundleEntryResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundleEntryResponse;
    

    {@member InsertItem
       Insert an existing FhirBundleEntryResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundleEntryResponse);
    
    {@member Item
       Get the iIndexth FhirBundleEntryResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundleEntryResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundleEntryResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundleEntryResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundleEntryResponses[index : Integer] : TFhirBundleEntryResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirBundle : TFhirResource
    A container for a collection of resources.
  }
  {!.Net HL7Connect.Fhir.Bundle}
  TFhirBundle = class (TFhirResource)
  private
    FType_ : TFhirEnum;
    FTotal : TFhirUnsignedInt;
    Flink_List : TFhirBundleLinkList;
    FentryList : TFhirBundleEntryList;
    FSignature : TFhirSignature;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirBundleType;
    Procedure SetType_ST(value : TFhirBundleType);
    Procedure SetTotal(value : TFhirUnsignedInt);
    Function GetTotalST : String;
    Procedure SetTotalST(value : String);
    function GetLink_List : TFhirBundleLinkList;
    function GetHasLink_List : Boolean;
    function GetEntryList : TFhirBundleEntryList;
    function GetHasEntryList : Boolean;
    Procedure SetSignature(value : TFhirSignature);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirBundle; overload;
    function Clone : TFhirBundle; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member type_
      Indicates the purpose of this bundle- how it was intended to be used.
    }
    property type_ : TFhirBundleType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member total
      Typed access to If a set of search matches, this is the total number of matches for the search (as opposed to the number of results in this bundle).
    }
    property total : String read GetTotalST write SetTotalST;
    {@member totalElement
      If a set of search matches, this is the total number of matches for the search (as opposed to the number of results in this bundle).
    }
    property totalElement : TFhirUnsignedInt read FTotal write SetTotal;

    {@member link_List
      A series of links that provide context to this bundle.
    }
    property link_List : TFhirBundleLinkList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

    {@member entryList
      An entry in a bundle resource - will either contain a resource, or information about a resource (transactions and history only).
    }
    property entryList : TFhirBundleEntryList read GetEntryList;
    property hasEntryList : boolean read GetHasEntryList;

    {@member signature
      Typed access to Digital Signature - base64 encoded. XML DigSIg or a JWT. (defined for API consistency)
    }
    property signature : TFhirSignature read FSignature write SetSignature;
    {@member signatureElement
      Digital Signature - base64 encoded. XML DigSIg or a JWT.
    }
    property signatureElement : TFhirSignature read FSignature write SetSignature;

  end;


  TFhirBundleListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirBundleList;
    function GetCurrent : TFhirBundle;
  public
    Constructor Create(list : TFhirBundleList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirBundle read GetCurrent;
  end;


  {@Class TFhirBundleList
    A list of FhirBundle
  }
  {!.Net HL7Connect.Fhir.BundleList}
  TFhirBundleList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirBundle;
    procedure SetItemN(index : Integer; value : TFhirBundle);
  public
    {!script hide}
    function Link : TFhirBundleList; Overload;
    function Clone : TFhirBundleList; Overload;
    function GetEnumerator : TFhirBundleListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirBundle to the end of the list.
    }
    function Append : TFhirBundle;

    
    {@member AddItem
      Add an already existing FhirBundle to the end of the list.
    }
    procedure AddItem(value : TFhirBundle); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirBundle) : Integer;
    

    {@member Insert
      Insert FhirBundle before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirBundle;
    

    {@member InsertItem
       Insert an existing FhirBundle before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirBundle);
    
    {@member Item
       Get the iIndexth FhirBundle. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirBundle. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirBundle);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirBundle;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirBundles[index : Integer] : TFhirBundle read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCarePlanRelatedPlan : TFhirElement
    Identifies CarePlans with some sort of formal relationship to the current plan.
  }
  {!.Net HL7Connect.Fhir.CarePlanRelatedPlan}
  TFhirCarePlanRelatedPlan = class (TFhirBackboneElement)
  private
    FCode : TFhirEnum;
    FPlan : TFhirReference{TFhirCarePlan};
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirCarePlanRelationship;
    Procedure SetCodeST(value : TFhirCarePlanRelationship);
    Procedure SetPlan(value : TFhirReference{TFhirCarePlan});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCarePlanRelatedPlan; overload;
    function Clone : TFhirCarePlanRelatedPlan; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Identifies the type of relationship this plan has to the target plan.
    }
    property code : TFhirCarePlanRelationship read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member plan
      Typed access to A reference to the plan to which a relationship is asserted. (defined for API consistency)
    }
    property plan : TFhirReference{TFhirCarePlan} read FPlan write SetPlan;
    {@member planElement
      A reference to the plan to which a relationship is asserted.
    }
    property planElement : TFhirReference{TFhirCarePlan} read FPlan write SetPlan;

  end;


  TFhirCarePlanRelatedPlanListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCarePlanRelatedPlanList;
    function GetCurrent : TFhirCarePlanRelatedPlan;
  public
    Constructor Create(list : TFhirCarePlanRelatedPlanList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCarePlanRelatedPlan read GetCurrent;
  end;


  {@Class TFhirCarePlanRelatedPlanList
    A list of FhirCarePlanRelatedPlan
  }
  {!.Net HL7Connect.Fhir.CarePlanRelatedPlanList}
  TFhirCarePlanRelatedPlanList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCarePlanRelatedPlan;
    procedure SetItemN(index : Integer; value : TFhirCarePlanRelatedPlan);
  public
    {!script hide}
    function Link : TFhirCarePlanRelatedPlanList; Overload;
    function Clone : TFhirCarePlanRelatedPlanList; Overload;
    function GetEnumerator : TFhirCarePlanRelatedPlanListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCarePlanRelatedPlan to the end of the list.
    }
    function Append : TFhirCarePlanRelatedPlan;

    
    {@member AddItem
      Add an already existing FhirCarePlanRelatedPlan to the end of the list.
    }
    procedure AddItem(value : TFhirCarePlanRelatedPlan); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCarePlanRelatedPlan) : Integer;
    

    {@member Insert
      Insert FhirCarePlanRelatedPlan before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCarePlanRelatedPlan;
    

    {@member InsertItem
       Insert an existing FhirCarePlanRelatedPlan before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCarePlanRelatedPlan);
    
    {@member Item
       Get the iIndexth FhirCarePlanRelatedPlan. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCarePlanRelatedPlan. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCarePlanRelatedPlan);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCarePlanRelatedPlan;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCarePlanRelatedPlans[index : Integer] : TFhirCarePlanRelatedPlan read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCarePlanParticipant : TFhirElement
    Identifies all people and organizations who are expected to be involved in the care envisioned by this plan.
  }
  {!.Net HL7Connect.Fhir.CarePlanParticipant}
  TFhirCarePlanParticipant = class (TFhirBackboneElement)
  private
    FRole : TFhirCodeableConcept;
    FMember : TFhirReference{Resource};
    Procedure SetRole(value : TFhirCodeableConcept);
    Procedure SetMember(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCarePlanParticipant; overload;
    function Clone : TFhirCarePlanParticipant; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member role
      Typed access to Indicates specific responsibility of an individual within the care plan; e.g. "Primary physician", "Team coordinator", "Caregiver", etc. (defined for API consistency)
    }
    property role : TFhirCodeableConcept read FRole write SetRole;
    {@member roleElement
      Indicates specific responsibility of an individual within the care plan; e.g. "Primary physician", "Team coordinator", "Caregiver", etc.
    }
    property roleElement : TFhirCodeableConcept read FRole write SetRole;

    {@member member
      Typed access to The specific person or organization who is participating/expected to participate in the care plan. (defined for API consistency)
    }
    property member : TFhirReference{Resource} read FMember write SetMember;
    {@member memberElement
      The specific person or organization who is participating/expected to participate in the care plan.
    }
    property memberElement : TFhirReference{Resource} read FMember write SetMember;

  end;


  TFhirCarePlanParticipantListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCarePlanParticipantList;
    function GetCurrent : TFhirCarePlanParticipant;
  public
    Constructor Create(list : TFhirCarePlanParticipantList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCarePlanParticipant read GetCurrent;
  end;


  {@Class TFhirCarePlanParticipantList
    A list of FhirCarePlanParticipant
  }
  {!.Net HL7Connect.Fhir.CarePlanParticipantList}
  TFhirCarePlanParticipantList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCarePlanParticipant;
    procedure SetItemN(index : Integer; value : TFhirCarePlanParticipant);
  public
    {!script hide}
    function Link : TFhirCarePlanParticipantList; Overload;
    function Clone : TFhirCarePlanParticipantList; Overload;
    function GetEnumerator : TFhirCarePlanParticipantListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCarePlanParticipant to the end of the list.
    }
    function Append : TFhirCarePlanParticipant;

    
    {@member AddItem
      Add an already existing FhirCarePlanParticipant to the end of the list.
    }
    procedure AddItem(value : TFhirCarePlanParticipant); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCarePlanParticipant) : Integer;
    

    {@member Insert
      Insert FhirCarePlanParticipant before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCarePlanParticipant;
    

    {@member InsertItem
       Insert an existing FhirCarePlanParticipant before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCarePlanParticipant);
    
    {@member Item
       Get the iIndexth FhirCarePlanParticipant. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCarePlanParticipant. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCarePlanParticipant);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCarePlanParticipant;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCarePlanParticipants[index : Integer] : TFhirCarePlanParticipant read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCarePlanActivity : TFhirElement
    Identifies a planned action to occur as part of the plan.  For example, a medication to be used, lab tests to perform, self-monitoring, education, etc.
  }
  {!.Net HL7Connect.Fhir.CarePlanActivity}
  TFhirCarePlanActivity = class (TFhirBackboneElement)
  private
    FactionResultingList : TFhirReferenceList{TFhirReference};
    FprogressList : TFhirAnnotationList;
    FReference : TFhirReference{Resource};
    FDetail : TFhirCarePlanActivityDetail;
    function GetActionResultingList : TFhirReferenceList{TFhirReference};
    function GetHasActionResultingList : Boolean;
    function GetProgressList : TFhirAnnotationList;
    function GetHasProgressList : Boolean;
    Procedure SetReference(value : TFhirReference{Resource});
    Procedure SetDetail(value : TFhirCarePlanActivityDetail);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCarePlanActivity; overload;
    function Clone : TFhirCarePlanActivity; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member actionResultingList
      Resources that describe follow-on actions resulting from the plan, such as drug prescriptions, encounter records, appointments, etc.
    }
    property actionResultingList : TFhirReferenceList{TFhirReference} read GetActionResultingList;
    property hasActionResultingList : boolean read GetHasActionResultingList;

    {@member progressList
      Notes about the adherence/status/progress of the activity.
    }
    property progressList : TFhirAnnotationList read GetProgressList;
    property hasProgressList : boolean read GetHasProgressList;

    {@member reference
      Typed access to The details of the proposed activity represented in a specific resource. (defined for API consistency)
    }
    property reference : TFhirReference{Resource} read FReference write SetReference;
    {@member referenceElement
      The details of the proposed activity represented in a specific resource.
    }
    property referenceElement : TFhirReference{Resource} read FReference write SetReference;

    {@member detail
      Typed access to A simple summary of a planned activity suitable for a general care plan system (e.g. form driven) that doesn't know about specific resources such as procedure etc. (defined for API consistency)
    }
    property detail : TFhirCarePlanActivityDetail read FDetail write SetDetail;
    {@member detailElement
      A simple summary of a planned activity suitable for a general care plan system (e.g. form driven) that doesn't know about specific resources such as procedure etc.
    }
    property detailElement : TFhirCarePlanActivityDetail read FDetail write SetDetail;

  end;


  TFhirCarePlanActivityListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCarePlanActivityList;
    function GetCurrent : TFhirCarePlanActivity;
  public
    Constructor Create(list : TFhirCarePlanActivityList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCarePlanActivity read GetCurrent;
  end;


  {@Class TFhirCarePlanActivityList
    A list of FhirCarePlanActivity
  }
  {!.Net HL7Connect.Fhir.CarePlanActivityList}
  TFhirCarePlanActivityList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCarePlanActivity;
    procedure SetItemN(index : Integer; value : TFhirCarePlanActivity);
  public
    {!script hide}
    function Link : TFhirCarePlanActivityList; Overload;
    function Clone : TFhirCarePlanActivityList; Overload;
    function GetEnumerator : TFhirCarePlanActivityListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCarePlanActivity to the end of the list.
    }
    function Append : TFhirCarePlanActivity;

    
    {@member AddItem
      Add an already existing FhirCarePlanActivity to the end of the list.
    }
    procedure AddItem(value : TFhirCarePlanActivity); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCarePlanActivity) : Integer;
    

    {@member Insert
      Insert FhirCarePlanActivity before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCarePlanActivity;
    

    {@member InsertItem
       Insert an existing FhirCarePlanActivity before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCarePlanActivity);
    
    {@member Item
       Get the iIndexth FhirCarePlanActivity. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCarePlanActivity. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCarePlanActivity);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCarePlanActivity;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCarePlanActivities[index : Integer] : TFhirCarePlanActivity read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCarePlanActivityDetail : TFhirElement
    A simple summary of a planned activity suitable for a general care plan system (e.g. form driven) that doesn't know about specific resources such as procedure etc.
  }
  {!.Net HL7Connect.Fhir.CarePlanActivityDetail}
  TFhirCarePlanActivityDetail = class (TFhirBackboneElement)
  private
    FCategory : TFhirCodeableConcept;
    FCode : TFhirCodeableConcept;
    FreasonCodeList : TFhirCodeableConceptList;
    FreasonReferenceList : TFhirReferenceList{TFhirCondition};
    FgoalList : TFhirReferenceList{TFhirGoal};
    FStatus : TFhirEnum;
    FStatusReason : TFhirCodeableConcept;
    FProhibited : TFhirBoolean;
    FScheduled : TFhirType;
    FLocation : TFhirReference{TFhirLocation};
    FperformerList : TFhirReferenceList{Resource};
    FProduct : TFhirType;
    FDailyAmount : TFhirQuantity;
    FQuantity : TFhirQuantity;
    FDescription : TFhirString;
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetReasonCodeList : TFhirCodeableConceptList;
    function GetHasReasonCodeList : Boolean;
    function GetReasonReferenceList : TFhirReferenceList{TFhirCondition};
    function GetHasReasonReferenceList : Boolean;
    function GetGoalList : TFhirReferenceList{TFhirGoal};
    function GetHasGoalList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirCarePlanActivityStatus;
    Procedure SetStatusST(value : TFhirCarePlanActivityStatus);
    Procedure SetStatusReason(value : TFhirCodeableConcept);
    Procedure SetProhibited(value : TFhirBoolean);
    Function GetProhibitedST : Boolean;
    Procedure SetProhibitedST(value : Boolean);
    Procedure SetScheduled(value : TFhirType);
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    function GetPerformerList : TFhirReferenceList{Resource};
    function GetHasPerformerList : Boolean;
    Procedure SetProduct(value : TFhirType);
    Procedure SetDailyAmount(value : TFhirQuantity);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCarePlanActivityDetail; overload;
    function Clone : TFhirCarePlanActivityDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member category
      Typed access to High-level categorization of the type of activity in a care plan. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      High-level categorization of the type of activity in a care plan.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member code
      Typed access to Detailed description of the type of planned activity; e.g. What lab test, what procedure, what kind of encounter. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Detailed description of the type of planned activity; e.g. What lab test, what procedure, what kind of encounter.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member reasonCodeList
      Provides the rationale that drove the inclusion of this particular activity as part of the plan.
    }
    property reasonCodeList : TFhirCodeableConceptList read GetReasonCodeList;
    property hasReasonCodeList : boolean read GetHasReasonCodeList;

    {@member reasonReferenceList
      Provides the health condition(s) that drove the inclusion of this particular activity as part of the plan.
    }
    property reasonReferenceList : TFhirReferenceList{TFhirCondition} read GetReasonReferenceList;
    property hasReasonReferenceList : boolean read GetHasReasonReferenceList;

    {@member goalList
      Internal reference that identifies the goals that this activity is intended to contribute towards meeting.
    }
    property goalList : TFhirReferenceList{TFhirGoal} read GetGoalList;
    property hasGoalList : boolean read GetHasGoalList;

    {@member status
      Identifies what progress is being made for the specific activity.
    }
    property status : TFhirCarePlanActivityStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member statusReason
      Typed access to Provides reason why the activity isn't yet started, is on hold, was cancelled, etc. (defined for API consistency)
    }
    property statusReason : TFhirCodeableConcept read FStatusReason write SetStatusReason;
    {@member statusReasonElement
      Provides reason why the activity isn't yet started, is on hold, was cancelled, etc.
    }
    property statusReasonElement : TFhirCodeableConcept read FStatusReason write SetStatusReason;

    {@member prohibited
      Typed access to If true, indicates that the described activity is one that must NOT be engaged in when following the plan.
    }
    property prohibited : Boolean read GetProhibitedST write SetProhibitedST;
    {@member prohibitedElement
      If true, indicates that the described activity is one that must NOT be engaged in when following the plan.
    }
    property prohibitedElement : TFhirBoolean read FProhibited write SetProhibited;

    {@member scheduled
      Typed access to The period, timing or frequency upon which the described activity is to occur. (defined for API consistency)
    }
    property scheduled : TFhirType read FScheduled write SetScheduled;
    {@member scheduledElement
      The period, timing or frequency upon which the described activity is to occur.
    }
    property scheduledElement : TFhirType read FScheduled write SetScheduled;

    {@member location
      Typed access to Identifies the facility where the activity will occur; e.g. home, hospital, specific clinic, etc. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      Identifies the facility where the activity will occur; e.g. home, hospital, specific clinic, etc.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member performerList
      Identifies who's expected to be involved in the activity.
    }
    property performerList : TFhirReferenceList{Resource} read GetPerformerList;
    property hasPerformerList : boolean read GetHasPerformerList;

    {@member product
      Typed access to Identifies the food, drug or other product to be consumed or supplied in the activity. (defined for API consistency)
    }
    property product : TFhirType read FProduct write SetProduct;
    {@member productElement
      Identifies the food, drug or other product to be consumed or supplied in the activity.
    }
    property productElement : TFhirType read FProduct write SetProduct;

    {@member dailyAmount
      Typed access to Identifies the quantity expected to be consumed in a given day. (defined for API consistency)
    }
    property dailyAmount : TFhirQuantity read FDailyAmount write SetDailyAmount;
    {@member dailyAmountElement
      Identifies the quantity expected to be consumed in a given day.
    }
    property dailyAmountElement : TFhirQuantity read FDailyAmount write SetDailyAmount;

    {@member quantity
      Typed access to Identifies the quantity expected to be supplied, administered or consumed by the subject. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      Identifies the quantity expected to be supplied, administered or consumed by the subject.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member description
      Typed access to This provides a textual description of constraints on the intended activity occurrence, including relation to other activities.  It may also include objectives, pre-conditions and end-conditions.  Finally, it may convey specifics about the activity such as body site, method, route, etc.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      This provides a textual description of constraints on the intended activity occurrence, including relation to other activities.  It may also include objectives, pre-conditions and end-conditions.  Finally, it may convey specifics about the activity such as body site, method, route, etc.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

  end;


  TFhirCarePlanActivityDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCarePlanActivityDetailList;
    function GetCurrent : TFhirCarePlanActivityDetail;
  public
    Constructor Create(list : TFhirCarePlanActivityDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCarePlanActivityDetail read GetCurrent;
  end;


  {@Class TFhirCarePlanActivityDetailList
    A list of FhirCarePlanActivityDetail
  }
  {!.Net HL7Connect.Fhir.CarePlanActivityDetailList}
  TFhirCarePlanActivityDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCarePlanActivityDetail;
    procedure SetItemN(index : Integer; value : TFhirCarePlanActivityDetail);
  public
    {!script hide}
    function Link : TFhirCarePlanActivityDetailList; Overload;
    function Clone : TFhirCarePlanActivityDetailList; Overload;
    function GetEnumerator : TFhirCarePlanActivityDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCarePlanActivityDetail to the end of the list.
    }
    function Append : TFhirCarePlanActivityDetail;

    
    {@member AddItem
      Add an already existing FhirCarePlanActivityDetail to the end of the list.
    }
    procedure AddItem(value : TFhirCarePlanActivityDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCarePlanActivityDetail) : Integer;
    

    {@member Insert
      Insert FhirCarePlanActivityDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCarePlanActivityDetail;
    

    {@member InsertItem
       Insert an existing FhirCarePlanActivityDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCarePlanActivityDetail);
    
    {@member Item
       Get the iIndexth FhirCarePlanActivityDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCarePlanActivityDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCarePlanActivityDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCarePlanActivityDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCarePlanActivityDetails[index : Integer] : TFhirCarePlanActivityDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCarePlan : TFhirDomainResource
    Describes the intention of how one or more practitioners intend to deliver care for a particular patient, group or community for a period of time, possibly limited to care for a specific condition or set of conditions.
  }
  {!.Net HL7Connect.Fhir.CarePlan}
  TFhirCarePlan = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FStatus : TFhirEnum;
    FContext : TFhirReference{Resource};
    FPeriod : TFhirPeriod;
    FauthorList : TFhirReferenceList{Resource};
    FModified : TFhirDateTime;
    FcategoryList : TFhirCodeableConceptList;
    FDescription : TFhirString;
    FaddressesList : TFhirReferenceList{TFhirCondition};
    FsupportList : TFhirReferenceList{TFhirReference};
    FrelatedPlanList : TFhirCarePlanRelatedPlanList;
    FparticipantList : TFhirCarePlanParticipantList;
    FgoalList : TFhirReferenceList{TFhirGoal};
    FactivityList : TFhirCarePlanActivityList;
    FNote : TFhirAnnotation;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirCarePlanStatus;
    Procedure SetStatusST(value : TFhirCarePlanStatus);
    Procedure SetContext(value : TFhirReference{Resource});
    Procedure SetPeriod(value : TFhirPeriod);
    function GetAuthorList : TFhirReferenceList{Resource};
    function GetHasAuthorList : Boolean;
    Procedure SetModified(value : TFhirDateTime);
    Function GetModifiedST : TDateAndTime;
    Procedure SetModifiedST(value : TDateAndTime);
    function GetCategoryList : TFhirCodeableConceptList;
    function GetHasCategoryList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetAddressesList : TFhirReferenceList{TFhirCondition};
    function GetHasAddressesList : Boolean;
    function GetSupportList : TFhirReferenceList{TFhirReference};
    function GetHasSupportList : Boolean;
    function GetRelatedPlanList : TFhirCarePlanRelatedPlanList;
    function GetHasRelatedPlanList : Boolean;
    function GetParticipantList : TFhirCarePlanParticipantList;
    function GetHasParticipantList : Boolean;
    function GetGoalList : TFhirReferenceList{TFhirGoal};
    function GetHasGoalList : Boolean;
    function GetActivityList : TFhirCarePlanActivityList;
    function GetHasActivityList : Boolean;
    Procedure SetNote(value : TFhirAnnotation);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCarePlan; overload;
    function Clone : TFhirCarePlan; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this care plan that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to Identifies the patient or group whose intended care is described by the plan. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Identifies the patient or group whose intended care is described by the plan.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member status
      Indicates whether the plan is currently being acted upon, represents future intentions or is now a historical record.
    }
    property status : TFhirCarePlanStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member context
      Typed access to Identifies the context in which this particular CarePlan is defined. (defined for API consistency)
    }
    property context : TFhirReference{Resource} read FContext write SetContext;
    {@member contextElement
      Identifies the context in which this particular CarePlan is defined.
    }
    property contextElement : TFhirReference{Resource} read FContext write SetContext;

    {@member period
      Typed access to Indicates when the plan did (or is intended to) come into effect and end. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      Indicates when the plan did (or is intended to) come into effect and end.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member authorList
      Identifies the individual(s) or ogranization who is responsible for the content of the care plan.
    }
    property authorList : TFhirReferenceList{Resource} read GetAuthorList;
    property hasAuthorList : boolean read GetHasAuthorList;

    {@member modified
      Typed access to Identifies the most recent date on which the plan has been revised.
    }
    property modified : TDateAndTime read GetModifiedST write SetModifiedST;
    {@member modifiedElement
      Identifies the most recent date on which the plan has been revised.
    }
    property modifiedElement : TFhirDateTime read FModified write SetModified;

    {@member categoryList
      Identifies what "kind" of plan this is to support differentiation between multiple co-existing plans; e.g. "Home health", "psychiatric", "asthma", "disease management", "wellness plan", etc.
    }
    property categoryList : TFhirCodeableConceptList read GetCategoryList;
    property hasCategoryList : boolean read GetHasCategoryList;

    {@member description
      Typed access to A description of the scope and nature of the plan.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A description of the scope and nature of the plan.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member addressesList
      Identifies the conditions/problems/concerns/diagnoses/etc. whose management and/or mitigation are handled by this plan.
    }
    property addressesList : TFhirReferenceList{TFhirCondition} read GetAddressesList;
    property hasAddressesList : boolean read GetHasAddressesList;

    {@member supportList
      Identifies portions of the patient's record that specifically influenced the formation of the plan.  These might include co-morbidities, recent procedures, limitations, recent assessments, etc.
    }
    property supportList : TFhirReferenceList{TFhirReference} read GetSupportList;
    property hasSupportList : boolean read GetHasSupportList;

    {@member relatedPlanList
      Identifies CarePlans with some sort of formal relationship to the current plan.
    }
    property relatedPlanList : TFhirCarePlanRelatedPlanList read GetRelatedPlanList;
    property hasRelatedPlanList : boolean read GetHasRelatedPlanList;

    {@member participantList
      Identifies all people and organizations who are expected to be involved in the care envisioned by this plan.
    }
    property participantList : TFhirCarePlanParticipantList read GetParticipantList;
    property hasParticipantList : boolean read GetHasParticipantList;

    {@member goalList
      Describes the intended objective(s) of carrying out the care plan.
    }
    property goalList : TFhirReferenceList{TFhirGoal} read GetGoalList;
    property hasGoalList : boolean read GetHasGoalList;

    {@member activityList
      Identifies a planned action to occur as part of the plan.  For example, a medication to be used, lab tests to perform, self-monitoring, education, etc.
    }
    property activityList : TFhirCarePlanActivityList read GetActivityList;
    property hasActivityList : boolean read GetHasActivityList;

    {@member note
      Typed access to General notes about the care plan not covered elsewhere. (defined for API consistency)
    }
    property note : TFhirAnnotation read FNote write SetNote;
    {@member noteElement
      General notes about the care plan not covered elsewhere.
    }
    property noteElement : TFhirAnnotation read FNote write SetNote;

  end;


  TFhirCarePlanListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCarePlanList;
    function GetCurrent : TFhirCarePlan;
  public
    Constructor Create(list : TFhirCarePlanList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCarePlan read GetCurrent;
  end;


  {@Class TFhirCarePlanList
    A list of FhirCarePlan
  }
  {!.Net HL7Connect.Fhir.CarePlanList}
  TFhirCarePlanList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCarePlan;
    procedure SetItemN(index : Integer; value : TFhirCarePlan);
  public
    {!script hide}
    function Link : TFhirCarePlanList; Overload;
    function Clone : TFhirCarePlanList; Overload;
    function GetEnumerator : TFhirCarePlanListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCarePlan to the end of the list.
    }
    function Append : TFhirCarePlan;

    
    {@member AddItem
      Add an already existing FhirCarePlan to the end of the list.
    }
    procedure AddItem(value : TFhirCarePlan); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCarePlan) : Integer;
    

    {@member Insert
      Insert FhirCarePlan before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCarePlan;
    

    {@member InsertItem
       Insert an existing FhirCarePlan before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCarePlan);
    
    {@member Item
       Get the iIndexth FhirCarePlan. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCarePlan. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCarePlan);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCarePlan;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCarePlans[index : Integer] : TFhirCarePlan read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimPayee : TFhirElement
    The party to be reimbursed for the services.
  }
  {!.Net HL7Connect.Fhir.ClaimPayee}
  TFhirClaimPayee = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    FPerson : TFhirReference{TFhirPatient};
    Procedure SetType_(value : TFhirCoding);
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetPerson(value : TFhirReference{TFhirPatient});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimPayee; overload;
    function Clone : TFhirClaimPayee; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Party to be reimbursed: Subscriber, provider, other. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      Party to be reimbursed: Subscriber, provider, other.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member provider
      Typed access to The provider who is to be reimbursed for the claim (the party to whom any benefit is assigned). (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The provider who is to be reimbursed for the claim (the party to whom any benefit is assigned).
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization who is to be reimbursed for the claim (the party to whom any benefit is assigned). (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization who is to be reimbursed for the claim (the party to whom any benefit is assigned).
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member person
      Typed access to The person other than the subscriber who is to be reimbursed for the claim (the party to whom any benefit is assigned). (defined for API consistency)
    }
    property person : TFhirReference{TFhirPatient} read FPerson write SetPerson;
    {@member personElement
      The person other than the subscriber who is to be reimbursed for the claim (the party to whom any benefit is assigned).
    }
    property personElement : TFhirReference{TFhirPatient} read FPerson write SetPerson;

  end;


  TFhirClaimPayeeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimPayeeList;
    function GetCurrent : TFhirClaimPayee;
  public
    Constructor Create(list : TFhirClaimPayeeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimPayee read GetCurrent;
  end;


  {@Class TFhirClaimPayeeList
    A list of FhirClaimPayee
  }
  {!.Net HL7Connect.Fhir.ClaimPayeeList}
  TFhirClaimPayeeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimPayee;
    procedure SetItemN(index : Integer; value : TFhirClaimPayee);
  public
    {!script hide}
    function Link : TFhirClaimPayeeList; Overload;
    function Clone : TFhirClaimPayeeList; Overload;
    function GetEnumerator : TFhirClaimPayeeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimPayee to the end of the list.
    }
    function Append : TFhirClaimPayee;

    
    {@member AddItem
      Add an already existing FhirClaimPayee to the end of the list.
    }
    procedure AddItem(value : TFhirClaimPayee); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimPayee) : Integer;
    

    {@member Insert
      Insert FhirClaimPayee before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimPayee;
    

    {@member InsertItem
       Insert an existing FhirClaimPayee before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimPayee);
    
    {@member Item
       Get the iIndexth FhirClaimPayee. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimPayee. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimPayee);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimPayee;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimPayees[index : Integer] : TFhirClaimPayee read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimDiagnosis : TFhirElement
    Ordered list of patient diagnosis for which care is sought.
  }
  {!.Net HL7Connect.Fhir.ClaimDiagnosis}
  TFhirClaimDiagnosis = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FDiagnosis : TFhirCoding;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetDiagnosis(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimDiagnosis; overload;
    function Clone : TFhirClaimDiagnosis; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to Sequence of diagnosis which serves to order and provide a link.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      Sequence of diagnosis which serves to order and provide a link.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member diagnosis
      Typed access to The diagnosis. (defined for API consistency)
    }
    property diagnosis : TFhirCoding read FDiagnosis write SetDiagnosis;
    {@member diagnosisElement
      The diagnosis.
    }
    property diagnosisElement : TFhirCoding read FDiagnosis write SetDiagnosis;

  end;


  TFhirClaimDiagnosisListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimDiagnosisList;
    function GetCurrent : TFhirClaimDiagnosis;
  public
    Constructor Create(list : TFhirClaimDiagnosisList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimDiagnosis read GetCurrent;
  end;


  {@Class TFhirClaimDiagnosisList
    A list of FhirClaimDiagnosis
  }
  {!.Net HL7Connect.Fhir.ClaimDiagnosisList}
  TFhirClaimDiagnosisList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimDiagnosis;
    procedure SetItemN(index : Integer; value : TFhirClaimDiagnosis);
  public
    {!script hide}
    function Link : TFhirClaimDiagnosisList; Overload;
    function Clone : TFhirClaimDiagnosisList; Overload;
    function GetEnumerator : TFhirClaimDiagnosisListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimDiagnosis to the end of the list.
    }
    function Append : TFhirClaimDiagnosis;

    
    {@member AddItem
      Add an already existing FhirClaimDiagnosis to the end of the list.
    }
    procedure AddItem(value : TFhirClaimDiagnosis); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimDiagnosis) : Integer;
    

    {@member Insert
      Insert FhirClaimDiagnosis before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimDiagnosis;
    

    {@member InsertItem
       Insert an existing FhirClaimDiagnosis before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimDiagnosis);
    
    {@member Item
       Get the iIndexth FhirClaimDiagnosis. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimDiagnosis. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimDiagnosis);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimDiagnosis;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimDiagnoses[index : Integer] : TFhirClaimDiagnosis read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimCoverage : TFhirElement
    Financial instrument by which payment information for health care.
  }
  {!.Net HL7Connect.Fhir.ClaimCoverage}
  TFhirClaimCoverage = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FFocal : TFhirBoolean;
    FCoverage : TFhirReference{TFhirCoverage};
    FBusinessArrangement : TFhirString;
    FRelationship : TFhirCoding;
    FpreAuthRefList : TFhirStringList;
    FClaimResponse : TFhirReference{TFhirClaimResponse};
    FOriginalRuleset : TFhirCoding;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetFocal(value : TFhirBoolean);
    Function GetFocalST : Boolean;
    Procedure SetFocalST(value : Boolean);
    Procedure SetCoverage(value : TFhirReference{TFhirCoverage});
    Procedure SetBusinessArrangement(value : TFhirString);
    Function GetBusinessArrangementST : String;
    Procedure SetBusinessArrangementST(value : String);
    Procedure SetRelationship(value : TFhirCoding);
    function GetPreAuthRefList : TFhirStringList;
    function GetHasPreAuthRefList : Boolean;
    Procedure SetClaimResponse(value : TFhirReference{TFhirClaimResponse});
    Procedure SetOriginalRuleset(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimCoverage; overload;
    function Clone : TFhirClaimCoverage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to A service line item.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      A service line item.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member focal
      Typed access to The instance number of the Coverage which is the focus for adjudication. The Coverage against which the claim is to be adjudicated.
    }
    property focal : Boolean read GetFocalST write SetFocalST;
    {@member focalElement
      The instance number of the Coverage which is the focus for adjudication. The Coverage against which the claim is to be adjudicated.
    }
    property focalElement : TFhirBoolean read FFocal write SetFocal;

    {@member coverage
      Typed access to Reference to the program or plan identification, underwriter or payor. (defined for API consistency)
    }
    property coverage : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;
    {@member coverageElement
      Reference to the program or plan identification, underwriter or payor.
    }
    property coverageElement : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;

    {@member businessArrangement
      Typed access to The contract number of a business agreement which describes the terms and conditions.
    }
    property businessArrangement : String read GetBusinessArrangementST write SetBusinessArrangementST;
    {@member businessArrangementElement
      The contract number of a business agreement which describes the terms and conditions.
    }
    property businessArrangementElement : TFhirString read FBusinessArrangement write SetBusinessArrangement;

    {@member relationship
      Typed access to The relationship of the patient to the subscriber. (defined for API consistency)
    }
    property relationship : TFhirCoding read FRelationship write SetRelationship;
    {@member relationshipElement
      The relationship of the patient to the subscriber.
    }
    property relationshipElement : TFhirCoding read FRelationship write SetRelationship;

    {@member preAuthRefList
      A list of references from the Insurer to which these services pertain.
    }
    property preAuthRefList : TFhirStringList read GetPreAuthRefList;
    property hasPreAuthRefList : boolean read GetHasPreAuthRefList;

    {@member claimResponse
      Typed access to The Coverages adjudication details. (defined for API consistency)
    }
    property claimResponse : TFhirReference{TFhirClaimResponse} read FClaimResponse write SetClaimResponse;
    {@member claimResponseElement
      The Coverages adjudication details.
    }
    property claimResponseElement : TFhirReference{TFhirClaimResponse} read FClaimResponse write SetClaimResponse;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

  end;


  TFhirClaimCoverageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimCoverageList;
    function GetCurrent : TFhirClaimCoverage;
  public
    Constructor Create(list : TFhirClaimCoverageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimCoverage read GetCurrent;
  end;


  {@Class TFhirClaimCoverageList
    A list of FhirClaimCoverage
  }
  {!.Net HL7Connect.Fhir.ClaimCoverageList}
  TFhirClaimCoverageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimCoverage;
    procedure SetItemN(index : Integer; value : TFhirClaimCoverage);
  public
    {!script hide}
    function Link : TFhirClaimCoverageList; Overload;
    function Clone : TFhirClaimCoverageList; Overload;
    function GetEnumerator : TFhirClaimCoverageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimCoverage to the end of the list.
    }
    function Append : TFhirClaimCoverage;

    
    {@member AddItem
      Add an already existing FhirClaimCoverage to the end of the list.
    }
    procedure AddItem(value : TFhirClaimCoverage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimCoverage) : Integer;
    

    {@member Insert
      Insert FhirClaimCoverage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimCoverage;
    

    {@member InsertItem
       Insert an existing FhirClaimCoverage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimCoverage);
    
    {@member Item
       Get the iIndexth FhirClaimCoverage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimCoverage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimCoverage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimCoverage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimCoverages[index : Integer] : TFhirClaimCoverage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimItem : TFhirElement
    First tier of goods and services.
  }
  {!.Net HL7Connect.Fhir.ClaimItem}
  TFhirClaimItem = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FType_ : TFhirCoding;
    FProvider : TFhirReference{TFhirPractitioner};
    FdiagnosisLinkIdList : TFhirPositiveIntList;
    FService : TFhirCoding;
    FServiceDate : TFhirDate;
    FQuantity : TFhirQuantity;
    FUnitPrice : TFhirQuantity;
    FFactor : TFhirDecimal;
    FPoints : TFhirDecimal;
    FNet : TFhirQuantity;
    FUdi : TFhirCoding;
    FBodySite : TFhirCoding;
    FsubSiteList : TFhirCodingList;
    FmodifierList : TFhirCodingList;
    FdetailList : TFhirClaimItemDetailList;
    FProsthesis : TFhirClaimItemProsthesis;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    function GetDiagnosisLinkIdList : TFhirPositiveIntList;
    function GetHasDiagnosisLinkIdList : Boolean;
    Procedure SetService(value : TFhirCoding);
    Procedure SetServiceDate(value : TFhirDate);
    Function GetServiceDateST : TDateAndTime;
    Procedure SetServiceDateST(value : TDateAndTime);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetUnitPrice(value : TFhirQuantity);
    Procedure SetFactor(value : TFhirDecimal);
    Function GetFactorST : String;
    Procedure SetFactorST(value : String);
    Procedure SetPoints(value : TFhirDecimal);
    Function GetPointsST : String;
    Procedure SetPointsST(value : String);
    Procedure SetNet(value : TFhirQuantity);
    Procedure SetUdi(value : TFhirCoding);
    Procedure SetBodySite(value : TFhirCoding);
    function GetSubSiteList : TFhirCodingList;
    function GetHasSubSiteList : Boolean;
    function GetModifierList : TFhirCodingList;
    function GetHasModifierList : Boolean;
    function GetDetailList : TFhirClaimItemDetailList;
    function GetHasDetailList : Boolean;
    Procedure SetProsthesis(value : TFhirClaimItemProsthesis);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimItem; overload;
    function Clone : TFhirClaimItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to A service line number.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      A service line number.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member type_
      Typed access to The type of product or service. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of product or service.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member provider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member diagnosisLinkIdList
      Diagnosis applicable for this service or product line.
    }
    property diagnosisLinkIdList : TFhirPositiveIntList read GetDiagnosisLinkIdList;
    property hasDiagnosisLinkIdList : boolean read GetHasDiagnosisLinkIdList;

    {@member service
      Typed access to If a grouping item then 'GROUP' otherwise it is a node therefore a code to indicate the Professional Service or Product supplied. (defined for API consistency)
    }
    property service : TFhirCoding read FService write SetService;
    {@member serviceElement
      If a grouping item then 'GROUP' otherwise it is a node therefore a code to indicate the Professional Service or Product supplied.
    }
    property serviceElement : TFhirCoding read FService write SetService;

    {@member serviceDate
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property serviceDate : TDateAndTime read GetServiceDateST write SetServiceDateST;
    {@member serviceDateElement
      The date when the enclosed suite of services were performed or completed.
    }
    property serviceDateElement : TFhirDate read FServiceDate write SetServiceDate;

    {@member quantity
      Typed access to The number of repetitions of a service or product. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The number of repetitions of a service or product.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member unitPrice
      Typed access to If the item is a node then this is the fee for the product or service, otherwise this is the total of the fees for the children of the group. (defined for API consistency)
    }
    property unitPrice : TFhirQuantity read FUnitPrice write SetUnitPrice;
    {@member unitPriceElement
      If the item is a node then this is the fee for the product or service, otherwise this is the total of the fees for the children of the group.
    }
    property unitPriceElement : TFhirQuantity read FUnitPrice write SetUnitPrice;

    {@member factor
      Typed access to A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factor : String read GetFactorST write SetFactorST;
    {@member factorElement
      A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factorElement : TFhirDecimal read FFactor write SetFactor;

    {@member points
      Typed access to An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property points : String read GetPointsST write SetPointsST;
    {@member pointsElement
      An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property pointsElement : TFhirDecimal read FPoints write SetPoints;

    {@member net
      Typed access to The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied. (defined for API consistency)
    }
    property net : TFhirQuantity read FNet write SetNet;
    {@member netElement
      The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied.
    }
    property netElement : TFhirQuantity read FNet write SetNet;

    {@member udi
      Typed access to List of Unique Device Identifiers associated with this line item. (defined for API consistency)
    }
    property udi : TFhirCoding read FUdi write SetUdi;
    {@member udiElement
      List of Unique Device Identifiers associated with this line item.
    }
    property udiElement : TFhirCoding read FUdi write SetUdi;

    {@member bodySite
      Typed access to Physical service site on the patient (limb, tooth, etc.). (defined for API consistency)
    }
    property bodySite : TFhirCoding read FBodySite write SetBodySite;
    {@member bodySiteElement
      Physical service site on the patient (limb, tooth, etc.).
    }
    property bodySiteElement : TFhirCoding read FBodySite write SetBodySite;

    {@member subSiteList
      A region or surface of the site, e.g. limb region or tooth surface(s).
    }
    property subSiteList : TFhirCodingList read GetSubSiteList;
    property hasSubSiteList : boolean read GetHasSubSiteList;

    {@member modifierList
      Item typification or modifiers codes, e.g. for Oral whether the treatment is cosmetic or associated with TMJ, or an appliance was lost or stolen.
    }
    property modifierList : TFhirCodingList read GetModifierList;
    property hasModifierList : boolean read GetHasModifierList;

    {@member detailList
      Second tier of goods and services.
    }
    property detailList : TFhirClaimItemDetailList read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

    {@member prosthesis
      Typed access to The materials and placement date of prior fixed prosthesis. (defined for API consistency)
    }
    property prosthesis : TFhirClaimItemProsthesis read FProsthesis write SetProsthesis;
    {@member prosthesisElement
      The materials and placement date of prior fixed prosthesis.
    }
    property prosthesisElement : TFhirClaimItemProsthesis read FProsthesis write SetProsthesis;

  end;


  TFhirClaimItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimItemList;
    function GetCurrent : TFhirClaimItem;
  public
    Constructor Create(list : TFhirClaimItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimItem read GetCurrent;
  end;


  {@Class TFhirClaimItemList
    A list of FhirClaimItem
  }
  {!.Net HL7Connect.Fhir.ClaimItemList}
  TFhirClaimItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimItem;
    procedure SetItemN(index : Integer; value : TFhirClaimItem);
  public
    {!script hide}
    function Link : TFhirClaimItemList; Overload;
    function Clone : TFhirClaimItemList; Overload;
    function GetEnumerator : TFhirClaimItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimItem to the end of the list.
    }
    function Append : TFhirClaimItem;

    
    {@member AddItem
      Add an already existing FhirClaimItem to the end of the list.
    }
    procedure AddItem(value : TFhirClaimItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimItem) : Integer;
    

    {@member Insert
      Insert FhirClaimItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimItem;
    

    {@member InsertItem
       Insert an existing FhirClaimItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimItem);
    
    {@member Item
       Get the iIndexth FhirClaimItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimItems[index : Integer] : TFhirClaimItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimItemDetail : TFhirElement
    Second tier of goods and services.
  }
  {!.Net HL7Connect.Fhir.ClaimItemDetail}
  TFhirClaimItemDetail = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FType_ : TFhirCoding;
    FService : TFhirCoding;
    FQuantity : TFhirQuantity;
    FUnitPrice : TFhirQuantity;
    FFactor : TFhirDecimal;
    FPoints : TFhirDecimal;
    FNet : TFhirQuantity;
    FUdi : TFhirCoding;
    FsubDetailList : TFhirClaimItemDetailSubDetailList;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetService(value : TFhirCoding);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetUnitPrice(value : TFhirQuantity);
    Procedure SetFactor(value : TFhirDecimal);
    Function GetFactorST : String;
    Procedure SetFactorST(value : String);
    Procedure SetPoints(value : TFhirDecimal);
    Function GetPointsST : String;
    Procedure SetPointsST(value : String);
    Procedure SetNet(value : TFhirQuantity);
    Procedure SetUdi(value : TFhirCoding);
    function GetSubDetailList : TFhirClaimItemDetailSubDetailList;
    function GetHasSubDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimItemDetail; overload;
    function Clone : TFhirClaimItemDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to A service line number.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      A service line number.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member type_
      Typed access to The type of product or service. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of product or service.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member service
      Typed access to If a grouping item then 'GROUP' otherwise it is a node therefore a code to indicate the Professional Service or Product supplied. (defined for API consistency)
    }
    property service : TFhirCoding read FService write SetService;
    {@member serviceElement
      If a grouping item then 'GROUP' otherwise it is a node therefore a code to indicate the Professional Service or Product supplied.
    }
    property serviceElement : TFhirCoding read FService write SetService;

    {@member quantity
      Typed access to The number of repetitions of a service or product. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The number of repetitions of a service or product.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member unitPrice
      Typed access to If the item is a node then this is the fee for the product or service, otherwise this is the total of the fees for the children of the group. (defined for API consistency)
    }
    property unitPrice : TFhirQuantity read FUnitPrice write SetUnitPrice;
    {@member unitPriceElement
      If the item is a node then this is the fee for the product or service, otherwise this is the total of the fees for the children of the group.
    }
    property unitPriceElement : TFhirQuantity read FUnitPrice write SetUnitPrice;

    {@member factor
      Typed access to A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factor : String read GetFactorST write SetFactorST;
    {@member factorElement
      A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factorElement : TFhirDecimal read FFactor write SetFactor;

    {@member points
      Typed access to An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property points : String read GetPointsST write SetPointsST;
    {@member pointsElement
      An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property pointsElement : TFhirDecimal read FPoints write SetPoints;

    {@member net
      Typed access to The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied. (defined for API consistency)
    }
    property net : TFhirQuantity read FNet write SetNet;
    {@member netElement
      The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied.
    }
    property netElement : TFhirQuantity read FNet write SetNet;

    {@member udi
      Typed access to List of Unique Device Identifiers associated with this line item. (defined for API consistency)
    }
    property udi : TFhirCoding read FUdi write SetUdi;
    {@member udiElement
      List of Unique Device Identifiers associated with this line item.
    }
    property udiElement : TFhirCoding read FUdi write SetUdi;

    {@member subDetailList
      Third tier of goods and services.
    }
    property subDetailList : TFhirClaimItemDetailSubDetailList read GetSubDetailList;
    property hasSubDetailList : boolean read GetHasSubDetailList;

  end;


  TFhirClaimItemDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimItemDetailList;
    function GetCurrent : TFhirClaimItemDetail;
  public
    Constructor Create(list : TFhirClaimItemDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimItemDetail read GetCurrent;
  end;


  {@Class TFhirClaimItemDetailList
    A list of FhirClaimItemDetail
  }
  {!.Net HL7Connect.Fhir.ClaimItemDetailList}
  TFhirClaimItemDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimItemDetail;
    procedure SetItemN(index : Integer; value : TFhirClaimItemDetail);
  public
    {!script hide}
    function Link : TFhirClaimItemDetailList; Overload;
    function Clone : TFhirClaimItemDetailList; Overload;
    function GetEnumerator : TFhirClaimItemDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimItemDetail to the end of the list.
    }
    function Append : TFhirClaimItemDetail;

    
    {@member AddItem
      Add an already existing FhirClaimItemDetail to the end of the list.
    }
    procedure AddItem(value : TFhirClaimItemDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimItemDetail) : Integer;
    

    {@member Insert
      Insert FhirClaimItemDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimItemDetail;
    

    {@member InsertItem
       Insert an existing FhirClaimItemDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimItemDetail);
    
    {@member Item
       Get the iIndexth FhirClaimItemDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimItemDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimItemDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimItemDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimItemDetails[index : Integer] : TFhirClaimItemDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimItemDetailSubDetail : TFhirElement
    Third tier of goods and services.
  }
  {!.Net HL7Connect.Fhir.ClaimItemDetailSubDetail}
  TFhirClaimItemDetailSubDetail = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FType_ : TFhirCoding;
    FService : TFhirCoding;
    FQuantity : TFhirQuantity;
    FUnitPrice : TFhirQuantity;
    FFactor : TFhirDecimal;
    FPoints : TFhirDecimal;
    FNet : TFhirQuantity;
    FUdi : TFhirCoding;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetService(value : TFhirCoding);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetUnitPrice(value : TFhirQuantity);
    Procedure SetFactor(value : TFhirDecimal);
    Function GetFactorST : String;
    Procedure SetFactorST(value : String);
    Procedure SetPoints(value : TFhirDecimal);
    Function GetPointsST : String;
    Procedure SetPointsST(value : String);
    Procedure SetNet(value : TFhirQuantity);
    Procedure SetUdi(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimItemDetailSubDetail; overload;
    function Clone : TFhirClaimItemDetailSubDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to A service line number.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      A service line number.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member type_
      Typed access to The type of product or service. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of product or service.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member service
      Typed access to The fee for an additional  service or product or charge. (defined for API consistency)
    }
    property service : TFhirCoding read FService write SetService;
    {@member serviceElement
      The fee for an additional  service or product or charge.
    }
    property serviceElement : TFhirCoding read FService write SetService;

    {@member quantity
      Typed access to The number of repetitions of a service or product. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The number of repetitions of a service or product.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member unitPrice
      Typed access to The fee for an additional  service or product or charge. (defined for API consistency)
    }
    property unitPrice : TFhirQuantity read FUnitPrice write SetUnitPrice;
    {@member unitPriceElement
      The fee for an additional  service or product or charge.
    }
    property unitPriceElement : TFhirQuantity read FUnitPrice write SetUnitPrice;

    {@member factor
      Typed access to A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factor : String read GetFactorST write SetFactorST;
    {@member factorElement
      A real number that represents a multiplier used in determining the overall value of services delivered and/or goods received. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factorElement : TFhirDecimal read FFactor write SetFactor;

    {@member points
      Typed access to An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property points : String read GetPointsST write SetPointsST;
    {@member pointsElement
      An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the good or service delivered. The concept of Points allows for assignment of point values for services and/or goods, such that a monetary amount can be assigned to each point.
    }
    property pointsElement : TFhirDecimal read FPoints write SetPoints;

    {@member net
      Typed access to The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied. (defined for API consistency)
    }
    property net : TFhirQuantity read FNet write SetNet;
    {@member netElement
      The quantity times the unit price for an additional  service or product or charge. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied.
    }
    property netElement : TFhirQuantity read FNet write SetNet;

    {@member udi
      Typed access to List of Unique Device Identifiers associated with this line item. (defined for API consistency)
    }
    property udi : TFhirCoding read FUdi write SetUdi;
    {@member udiElement
      List of Unique Device Identifiers associated with this line item.
    }
    property udiElement : TFhirCoding read FUdi write SetUdi;

  end;


  TFhirClaimItemDetailSubDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimItemDetailSubDetailList;
    function GetCurrent : TFhirClaimItemDetailSubDetail;
  public
    Constructor Create(list : TFhirClaimItemDetailSubDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimItemDetailSubDetail read GetCurrent;
  end;


  {@Class TFhirClaimItemDetailSubDetailList
    A list of FhirClaimItemDetailSubDetail
  }
  {!.Net HL7Connect.Fhir.ClaimItemDetailSubDetailList}
  TFhirClaimItemDetailSubDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimItemDetailSubDetail;
    procedure SetItemN(index : Integer; value : TFhirClaimItemDetailSubDetail);
  public
    {!script hide}
    function Link : TFhirClaimItemDetailSubDetailList; Overload;
    function Clone : TFhirClaimItemDetailSubDetailList; Overload;
    function GetEnumerator : TFhirClaimItemDetailSubDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimItemDetailSubDetail to the end of the list.
    }
    function Append : TFhirClaimItemDetailSubDetail;

    
    {@member AddItem
      Add an already existing FhirClaimItemDetailSubDetail to the end of the list.
    }
    procedure AddItem(value : TFhirClaimItemDetailSubDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimItemDetailSubDetail) : Integer;
    

    {@member Insert
      Insert FhirClaimItemDetailSubDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimItemDetailSubDetail;
    

    {@member InsertItem
       Insert an existing FhirClaimItemDetailSubDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimItemDetailSubDetail);
    
    {@member Item
       Get the iIndexth FhirClaimItemDetailSubDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimItemDetailSubDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimItemDetailSubDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimItemDetailSubDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimItemDetailSubDetails[index : Integer] : TFhirClaimItemDetailSubDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimItemProsthesis : TFhirElement
    The materials and placement date of prior fixed prosthesis.
  }
  {!.Net HL7Connect.Fhir.ClaimItemProsthesis}
  TFhirClaimItemProsthesis = class (TFhirBackboneElement)
  private
    FInitial : TFhirBoolean;
    FPriorDate : TFhirDate;
    FPriorMaterial : TFhirCoding;
    Procedure SetInitial(value : TFhirBoolean);
    Function GetInitialST : Boolean;
    Procedure SetInitialST(value : Boolean);
    Procedure SetPriorDate(value : TFhirDate);
    Function GetPriorDateST : TDateAndTime;
    Procedure SetPriorDateST(value : TDateAndTime);
    Procedure SetPriorMaterial(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimItemProsthesis; overload;
    function Clone : TFhirClaimItemProsthesis; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member initial
      Typed access to Indicates whether this is the initial placement of a fixed prosthesis.
    }
    property initial : Boolean read GetInitialST write SetInitialST;
    {@member initialElement
      Indicates whether this is the initial placement of a fixed prosthesis.
    }
    property initialElement : TFhirBoolean read FInitial write SetInitial;

    {@member priorDate
      Typed access to Date of the initial placement.
    }
    property priorDate : TDateAndTime read GetPriorDateST write SetPriorDateST;
    {@member priorDateElement
      Date of the initial placement.
    }
    property priorDateElement : TFhirDate read FPriorDate write SetPriorDate;

    {@member priorMaterial
      Typed access to Material of the prior denture or bridge prosthesis. (Oral). (defined for API consistency)
    }
    property priorMaterial : TFhirCoding read FPriorMaterial write SetPriorMaterial;
    {@member priorMaterialElement
      Material of the prior denture or bridge prosthesis. (Oral).
    }
    property priorMaterialElement : TFhirCoding read FPriorMaterial write SetPriorMaterial;

  end;


  TFhirClaimItemProsthesisListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimItemProsthesisList;
    function GetCurrent : TFhirClaimItemProsthesis;
  public
    Constructor Create(list : TFhirClaimItemProsthesisList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimItemProsthesis read GetCurrent;
  end;


  {@Class TFhirClaimItemProsthesisList
    A list of FhirClaimItemProsthesis
  }
  {!.Net HL7Connect.Fhir.ClaimItemProsthesisList}
  TFhirClaimItemProsthesisList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimItemProsthesis;
    procedure SetItemN(index : Integer; value : TFhirClaimItemProsthesis);
  public
    {!script hide}
    function Link : TFhirClaimItemProsthesisList; Overload;
    function Clone : TFhirClaimItemProsthesisList; Overload;
    function GetEnumerator : TFhirClaimItemProsthesisListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimItemProsthesis to the end of the list.
    }
    function Append : TFhirClaimItemProsthesis;

    
    {@member AddItem
      Add an already existing FhirClaimItemProsthesis to the end of the list.
    }
    procedure AddItem(value : TFhirClaimItemProsthesis); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimItemProsthesis) : Integer;
    

    {@member Insert
      Insert FhirClaimItemProsthesis before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimItemProsthesis;
    

    {@member InsertItem
       Insert an existing FhirClaimItemProsthesis before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimItemProsthesis);
    
    {@member Item
       Get the iIndexth FhirClaimItemProsthesis. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimItemProsthesis. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimItemProsthesis);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimItemProsthesis;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimItemProstheses[index : Integer] : TFhirClaimItemProsthesis read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimMissingTeeth : TFhirElement
    A list of teeth which would be expected but are not found due to having been previously  extracted or for other reasons.
  }
  {!.Net HL7Connect.Fhir.ClaimMissingTeeth}
  TFhirClaimMissingTeeth = class (TFhirBackboneElement)
  private
    FTooth : TFhirCoding;
    FReason : TFhirCoding;
    FExtractionDate : TFhirDate;
    Procedure SetTooth(value : TFhirCoding);
    Procedure SetReason(value : TFhirCoding);
    Procedure SetExtractionDate(value : TFhirDate);
    Function GetExtractionDateST : TDateAndTime;
    Procedure SetExtractionDateST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimMissingTeeth; overload;
    function Clone : TFhirClaimMissingTeeth; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member tooth
      Typed access to The code identifying which tooth is missing. (defined for API consistency)
    }
    property tooth : TFhirCoding read FTooth write SetTooth;
    {@member toothElement
      The code identifying which tooth is missing.
    }
    property toothElement : TFhirCoding read FTooth write SetTooth;

    {@member reason
      Typed access to Missing reason may be: E-extraction, O-other. (defined for API consistency)
    }
    property reason : TFhirCoding read FReason write SetReason;
    {@member reasonElement
      Missing reason may be: E-extraction, O-other.
    }
    property reasonElement : TFhirCoding read FReason write SetReason;

    {@member extractionDate
      Typed access to The date of the extraction either known from records or patient reported estimate.
    }
    property extractionDate : TDateAndTime read GetExtractionDateST write SetExtractionDateST;
    {@member extractionDateElement
      The date of the extraction either known from records or patient reported estimate.
    }
    property extractionDateElement : TFhirDate read FExtractionDate write SetExtractionDate;

  end;


  TFhirClaimMissingTeethListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimMissingTeethList;
    function GetCurrent : TFhirClaimMissingTeeth;
  public
    Constructor Create(list : TFhirClaimMissingTeethList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimMissingTeeth read GetCurrent;
  end;


  {@Class TFhirClaimMissingTeethList
    A list of FhirClaimMissingTeeth
  }
  {!.Net HL7Connect.Fhir.ClaimMissingTeethList}
  TFhirClaimMissingTeethList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimMissingTeeth;
    procedure SetItemN(index : Integer; value : TFhirClaimMissingTeeth);
  public
    {!script hide}
    function Link : TFhirClaimMissingTeethList; Overload;
    function Clone : TFhirClaimMissingTeethList; Overload;
    function GetEnumerator : TFhirClaimMissingTeethListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimMissingTeeth to the end of the list.
    }
    function Append : TFhirClaimMissingTeeth;

    
    {@member AddItem
      Add an already existing FhirClaimMissingTeeth to the end of the list.
    }
    procedure AddItem(value : TFhirClaimMissingTeeth); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimMissingTeeth) : Integer;
    

    {@member Insert
      Insert FhirClaimMissingTeeth before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimMissingTeeth;
    

    {@member InsertItem
       Insert an existing FhirClaimMissingTeeth before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimMissingTeeth);
    
    {@member Item
       Get the iIndexth FhirClaimMissingTeeth. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimMissingTeeth. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimMissingTeeth);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimMissingTeeth;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimMissingTeeths[index : Integer] : TFhirClaimMissingTeeth read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaim : TFhirDomainResource
    A provider issued list of services and products provided, or to be provided, to a patient which is provided to an insurer for payment recovery.
  }
  {!.Net HL7Connect.Fhir.Claim}
  TFhirClaim = class (TFhirDomainResource)
  private
    FType_ : TFhirEnum;
    FidentifierList : TFhirIdentifierList;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FTarget : TFhirReference{TFhirOrganization};
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    FUse : TFhirEnum;
    FPriority : TFhirCoding;
    FFundsReserve : TFhirCoding;
    FEnterer : TFhirReference{TFhirPractitioner};
    FFacility : TFhirReference{TFhirLocation};
    FPrescription : TFhirReference{Resource};
    FOriginalPrescription : TFhirReference{TFhirMedicationOrder};
    FPayee : TFhirClaimPayee;
    FReferral : TFhirReference{TFhirReferralRequest};
    FdiagnosisList : TFhirClaimDiagnosisList;
    FconditionList : TFhirCodingList;
    FPatient : TFhirReference{TFhirPatient};
    FcoverageList : TFhirClaimCoverageList;
    FexceptionList : TFhirCodingList;
    FSchool : TFhirString;
    FAccident : TFhirDate;
    FAccidentType : TFhirCoding;
    FinterventionExceptionList : TFhirCodingList;
    FitemList : TFhirClaimItemList;
    FadditionalMaterialsList : TFhirCodingList;
    FmissingTeethList : TFhirClaimMissingTeethList;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirClaimTypeLink;
    Procedure SetType_ST(value : TFhirClaimTypeLink);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetTarget(value : TFhirReference{TFhirOrganization});
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetUse(value : TFhirEnum);
    Function GetUseST : TFhirClaimUseLink;
    Procedure SetUseST(value : TFhirClaimUseLink);
    Procedure SetPriority(value : TFhirCoding);
    Procedure SetFundsReserve(value : TFhirCoding);
    Procedure SetEnterer(value : TFhirReference{TFhirPractitioner});
    Procedure SetFacility(value : TFhirReference{TFhirLocation});
    Procedure SetPrescription(value : TFhirReference{Resource});
    Procedure SetOriginalPrescription(value : TFhirReference{TFhirMedicationOrder});
    Procedure SetPayee(value : TFhirClaimPayee);
    Procedure SetReferral(value : TFhirReference{TFhirReferralRequest});
    function GetDiagnosisList : TFhirClaimDiagnosisList;
    function GetHasDiagnosisList : Boolean;
    function GetConditionList : TFhirCodingList;
    function GetHasConditionList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    function GetCoverageList : TFhirClaimCoverageList;
    function GetHasCoverageList : Boolean;
    function GetExceptionList : TFhirCodingList;
    function GetHasExceptionList : Boolean;
    Procedure SetSchool(value : TFhirString);
    Function GetSchoolST : String;
    Procedure SetSchoolST(value : String);
    Procedure SetAccident(value : TFhirDate);
    Function GetAccidentST : TDateAndTime;
    Procedure SetAccidentST(value : TDateAndTime);
    Procedure SetAccidentType(value : TFhirCoding);
    function GetInterventionExceptionList : TFhirCodingList;
    function GetHasInterventionExceptionList : Boolean;
    function GetItemList : TFhirClaimItemList;
    function GetHasItemList : Boolean;
    function GetAdditionalMaterialsList : TFhirCodingList;
    function GetHasAdditionalMaterialsList : Boolean;
    function GetMissingTeethList : TFhirClaimMissingTeethList;
    function GetHasMissingTeethList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaim; overload;
    function Clone : TFhirClaim; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member type_
      The category of claim this is.
    }
    property type_ : TFhirClaimTypeLink read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member identifierList
      The business identifier for the instance: invoice number, claim number, pre-determination or pre-authorization number.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member ruleset
      Typed access to The version of the specification on which this instance relies. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the specification on which this instance relies.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The version of the specification from which the original instance was created. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The version of the specification from which the original instance was created.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member target
      Typed access to Insurer Identifier, typical BIN number (6 digit). (defined for API consistency)
    }
    property target : TFhirReference{TFhirOrganization} read FTarget write SetTarget;
    {@member targetElement
      Insurer Identifier, typical BIN number (6 digit).
    }
    property targetElement : TFhirReference{TFhirOrganization} read FTarget write SetTarget;

    {@member provider
      Typed access to The provider which is responsible for the bill, claim pre-determination, pre-authorization. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The provider which is responsible for the bill, claim pre-determination, pre-authorization.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization which is responsible for the bill, claim pre-determination, pre-authorization. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization which is responsible for the bill, claim pre-determination, pre-authorization.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member use
      Complete (Bill or Claim), Proposed (Pre-Authorization), Exploratory (Pre-determination).
    }
    property use : TFhirClaimUseLink read GetUseST write SetUseST;
    property useElement : TFhirEnum read FUse write SetUse;

    {@member priority
      Typed access to Immediate (stat), best effort (normal), deferred (deferred). (defined for API consistency)
    }
    property priority : TFhirCoding read FPriority write SetPriority;
    {@member priorityElement
      Immediate (stat), best effort (normal), deferred (deferred).
    }
    property priorityElement : TFhirCoding read FPriority write SetPriority;

    {@member fundsReserve
      Typed access to In the case of a Pre-Determination/Pre-Authorization the provider may request that funds in the amount of the expected Benefit be reserved ('Patient' or 'Provider') to pay for the Benefits determined on the subsequent claim(s). 'None' explicitly indicates no funds reserving is requested. (defined for API consistency)
    }
    property fundsReserve : TFhirCoding read FFundsReserve write SetFundsReserve;
    {@member fundsReserveElement
      In the case of a Pre-Determination/Pre-Authorization the provider may request that funds in the amount of the expected Benefit be reserved ('Patient' or 'Provider') to pay for the Benefits determined on the subsequent claim(s). 'None' explicitly indicates no funds reserving is requested.
    }
    property fundsReserveElement : TFhirCoding read FFundsReserve write SetFundsReserve;

    {@member enterer
      Typed access to Person who created the invoice/claim/pre-determination or pre-authorization. (defined for API consistency)
    }
    property enterer : TFhirReference{TFhirPractitioner} read FEnterer write SetEnterer;
    {@member entererElement
      Person who created the invoice/claim/pre-determination or pre-authorization.
    }
    property entererElement : TFhirReference{TFhirPractitioner} read FEnterer write SetEnterer;

    {@member facility
      Typed access to Facility where the services were provided. (defined for API consistency)
    }
    property facility : TFhirReference{TFhirLocation} read FFacility write SetFacility;
    {@member facilityElement
      Facility where the services were provided.
    }
    property facilityElement : TFhirReference{TFhirLocation} read FFacility write SetFacility;

    {@member prescription
      Typed access to Prescription to support the dispensing of Pharmacy or Vision products. (defined for API consistency)
    }
    property prescription : TFhirReference{Resource} read FPrescription write SetPrescription;
    {@member prescriptionElement
      Prescription to support the dispensing of Pharmacy or Vision products.
    }
    property prescriptionElement : TFhirReference{Resource} read FPrescription write SetPrescription;

    {@member originalPrescription
      Typed access to Original prescription to support the dispensing of pharmacy services, medications or products. (defined for API consistency)
    }
    property originalPrescription : TFhirReference{TFhirMedicationOrder} read FOriginalPrescription write SetOriginalPrescription;
    {@member originalPrescriptionElement
      Original prescription to support the dispensing of pharmacy services, medications or products.
    }
    property originalPrescriptionElement : TFhirReference{TFhirMedicationOrder} read FOriginalPrescription write SetOriginalPrescription;

    {@member payee
      Typed access to The party to be reimbursed for the services. (defined for API consistency)
    }
    property payee : TFhirClaimPayee read FPayee write SetPayee;
    {@member payeeElement
      The party to be reimbursed for the services.
    }
    property payeeElement : TFhirClaimPayee read FPayee write SetPayee;

    {@member referral
      Typed access to The referral resource which lists the date, practitioner, reason and other supporting information. (defined for API consistency)
    }
    property referral : TFhirReference{TFhirReferralRequest} read FReferral write SetReferral;
    {@member referralElement
      The referral resource which lists the date, practitioner, reason and other supporting information.
    }
    property referralElement : TFhirReference{TFhirReferralRequest} read FReferral write SetReferral;

    {@member diagnosisList
      Ordered list of patient diagnosis for which care is sought.
    }
    property diagnosisList : TFhirClaimDiagnosisList read GetDiagnosisList;
    property hasDiagnosisList : boolean read GetHasDiagnosisList;

    {@member conditionList
      List of patient conditions for which care is sought.
    }
    property conditionList : TFhirCodingList read GetConditionList;
    property hasConditionList : boolean read GetHasConditionList;

    {@member patient
      Typed access to Patient Resource. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      Patient Resource.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member coverageList
      Financial instrument by which payment information for health care.
    }
    property coverageList : TFhirClaimCoverageList read GetCoverageList;
    property hasCoverageList : boolean read GetHasCoverageList;

    {@member exceptionList
      Factors which may influence the applicability of coverage.
    }
    property exceptionList : TFhirCodingList read GetExceptionList;
    property hasExceptionList : boolean read GetHasExceptionList;

    {@member school
      Typed access to Name of school for over-aged dependents.
    }
    property school : String read GetSchoolST write SetSchoolST;
    {@member schoolElement
      Name of school for over-aged dependents.
    }
    property schoolElement : TFhirString read FSchool write SetSchool;

    {@member accident
      Typed access to Date of an accident which these services are addressing.
    }
    property accident : TDateAndTime read GetAccidentST write SetAccidentST;
    {@member accidentElement
      Date of an accident which these services are addressing.
    }
    property accidentElement : TFhirDate read FAccident write SetAccident;

    {@member accidentType
      Typed access to Type of accident: work, auto, etc. (defined for API consistency)
    }
    property accidentType : TFhirCoding read FAccidentType write SetAccidentType;
    {@member accidentTypeElement
      Type of accident: work, auto, etc.
    }
    property accidentTypeElement : TFhirCoding read FAccidentType write SetAccidentType;

    {@member interventionExceptionList
      A list of intervention and exception codes which may influence the adjudication of the claim.
    }
    property interventionExceptionList : TFhirCodingList read GetInterventionExceptionList;
    property hasInterventionExceptionList : boolean read GetHasInterventionExceptionList;

    {@member itemList
      First tier of goods and services.
    }
    property itemList : TFhirClaimItemList read GetItemList;
    property hasItemList : boolean read GetHasItemList;

    {@member additionalMaterialsList
      Code to indicate that Xrays, images, emails, documents, models or attachments are being sent in support of this submission.
    }
    property additionalMaterialsList : TFhirCodingList read GetAdditionalMaterialsList;
    property hasAdditionalMaterialsList : boolean read GetHasAdditionalMaterialsList;

    {@member missingTeethList
      A list of teeth which would be expected but are not found due to having been previously  extracted or for other reasons.
    }
    property missingTeethList : TFhirClaimMissingTeethList read GetMissingTeethList;
    property hasMissingTeethList : boolean read GetHasMissingTeethList;

  end;


  TFhirClaimListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimList;
    function GetCurrent : TFhirClaim;
  public
    Constructor Create(list : TFhirClaimList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaim read GetCurrent;
  end;


  {@Class TFhirClaimList
    A list of FhirClaim
  }
  {!.Net HL7Connect.Fhir.ClaimList}
  TFhirClaimList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaim;
    procedure SetItemN(index : Integer; value : TFhirClaim);
  public
    {!script hide}
    function Link : TFhirClaimList; Overload;
    function Clone : TFhirClaimList; Overload;
    function GetEnumerator : TFhirClaimListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaim to the end of the list.
    }
    function Append : TFhirClaim;

    
    {@member AddItem
      Add an already existing FhirClaim to the end of the list.
    }
    procedure AddItem(value : TFhirClaim); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaim) : Integer;
    

    {@member Insert
      Insert FhirClaim before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaim;
    

    {@member InsertItem
       Insert an existing FhirClaim before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaim);
    
    {@member Item
       Get the iIndexth FhirClaim. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaim. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaim);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaim;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaims[index : Integer] : TFhirClaim read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItem : TFhirElement
    The first tier service adjudications for submitted services.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItem}
  TFhirClaimResponseItem = class (TFhirBackboneElement)
  private
    FSequenceLinkId : TFhirPositiveInt;
    FnoteNumberList : TFhirPositiveIntList;
    FadjudicationList : TFhirClaimResponseItemAdjudicationList;
    FdetailList : TFhirClaimResponseItemDetailList;
    Procedure SetSequenceLinkId(value : TFhirPositiveInt);
    Function GetSequenceLinkIdST : String;
    Procedure SetSequenceLinkIdST(value : String);
    function GetNoteNumberList : TFhirPositiveIntList;
    function GetHasNoteNumberList : Boolean;
    function GetAdjudicationList : TFhirClaimResponseItemAdjudicationList;
    function GetHasAdjudicationList : Boolean;
    function GetDetailList : TFhirClaimResponseItemDetailList;
    function GetHasDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItem; overload;
    function Clone : TFhirClaimResponseItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkId
      Typed access to A service line number.
    }
    property sequenceLinkId : String read GetSequenceLinkIdST write SetSequenceLinkIdST;
    {@member sequenceLinkIdElement
      A service line number.
    }
    property sequenceLinkIdElement : TFhirPositiveInt read FSequenceLinkId write SetSequenceLinkId;

    {@member noteNumberList
      A list of note references to the notes provided below.
    }
    property noteNumberList : TFhirPositiveIntList read GetNoteNumberList;
    property hasNoteNumberList : boolean read GetHasNoteNumberList;

    {@member adjudicationList
      The adjudications results.
    }
    property adjudicationList : TFhirClaimResponseItemAdjudicationList read GetAdjudicationList;
    property hasAdjudicationList : boolean read GetHasAdjudicationList;

    {@member detailList
      The second tier service adjudications for submitted services.
    }
    property detailList : TFhirClaimResponseItemDetailList read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirClaimResponseItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemList;
    function GetCurrent : TFhirClaimResponseItem;
  public
    Constructor Create(list : TFhirClaimResponseItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItem read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemList
    A list of FhirClaimResponseItem
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemList}
  TFhirClaimResponseItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItem;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItem);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemList; Overload;
    function Clone : TFhirClaimResponseItemList; Overload;
    function GetEnumerator : TFhirClaimResponseItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItem to the end of the list.
    }
    function Append : TFhirClaimResponseItem;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItem to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItem) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItem;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItem);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItems[index : Integer] : TFhirClaimResponseItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItemAdjudication : TFhirElement
    The adjudications results.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemAdjudication}
  TFhirClaimResponseItemAdjudication = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FAmount : TFhirQuantity;
    FValue : TFhirDecimal;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetAmount(value : TFhirQuantity);
    Procedure SetValue(value : TFhirDecimal);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItemAdjudication; overload;
    function Clone : TFhirClaimResponseItemAdjudication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member amount
      Typed access to Monetary amount associated with the code. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Monetary amount associated with the code.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

    {@member value
      Typed access to A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property valueElement : TFhirDecimal read FValue write SetValue;

  end;


  TFhirClaimResponseItemAdjudicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemAdjudicationList;
    function GetCurrent : TFhirClaimResponseItemAdjudication;
  public
    Constructor Create(list : TFhirClaimResponseItemAdjudicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItemAdjudication read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemAdjudicationList
    A list of FhirClaimResponseItemAdjudication
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemAdjudicationList}
  TFhirClaimResponseItemAdjudicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItemAdjudication;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItemAdjudication);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemAdjudicationList; Overload;
    function Clone : TFhirClaimResponseItemAdjudicationList; Overload;
    function GetEnumerator : TFhirClaimResponseItemAdjudicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItemAdjudication to the end of the list.
    }
    function Append : TFhirClaimResponseItemAdjudication;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItemAdjudication to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItemAdjudication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItemAdjudication) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItemAdjudication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItemAdjudication;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItemAdjudication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItemAdjudication);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemAdjudication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemAdjudication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItemAdjudication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItemAdjudication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItemAdjudications[index : Integer] : TFhirClaimResponseItemAdjudication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItemDetail : TFhirElement
    The second tier service adjudications for submitted services.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetail}
  TFhirClaimResponseItemDetail = class (TFhirBackboneElement)
  private
    FSequenceLinkId : TFhirPositiveInt;
    FadjudicationList : TFhirClaimResponseItemDetailAdjudicationList;
    FsubDetailList : TFhirClaimResponseItemDetailSubDetailList;
    Procedure SetSequenceLinkId(value : TFhirPositiveInt);
    Function GetSequenceLinkIdST : String;
    Procedure SetSequenceLinkIdST(value : String);
    function GetAdjudicationList : TFhirClaimResponseItemDetailAdjudicationList;
    function GetHasAdjudicationList : Boolean;
    function GetSubDetailList : TFhirClaimResponseItemDetailSubDetailList;
    function GetHasSubDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItemDetail; overload;
    function Clone : TFhirClaimResponseItemDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkId
      Typed access to A service line number.
    }
    property sequenceLinkId : String read GetSequenceLinkIdST write SetSequenceLinkIdST;
    {@member sequenceLinkIdElement
      A service line number.
    }
    property sequenceLinkIdElement : TFhirPositiveInt read FSequenceLinkId write SetSequenceLinkId;

    {@member adjudicationList
      The adjudications results.
    }
    property adjudicationList : TFhirClaimResponseItemDetailAdjudicationList read GetAdjudicationList;
    property hasAdjudicationList : boolean read GetHasAdjudicationList;

    {@member subDetailList
      The third tier service adjudications for submitted services.
    }
    property subDetailList : TFhirClaimResponseItemDetailSubDetailList read GetSubDetailList;
    property hasSubDetailList : boolean read GetHasSubDetailList;

  end;


  TFhirClaimResponseItemDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemDetailList;
    function GetCurrent : TFhirClaimResponseItemDetail;
  public
    Constructor Create(list : TFhirClaimResponseItemDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItemDetail read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemDetailList
    A list of FhirClaimResponseItemDetail
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailList}
  TFhirClaimResponseItemDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItemDetail;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItemDetail);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemDetailList; Overload;
    function Clone : TFhirClaimResponseItemDetailList; Overload;
    function GetEnumerator : TFhirClaimResponseItemDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItemDetail to the end of the list.
    }
    function Append : TFhirClaimResponseItemDetail;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItemDetail to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItemDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItemDetail) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItemDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItemDetail;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItemDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItemDetail);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItemDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItemDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItemDetails[index : Integer] : TFhirClaimResponseItemDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItemDetailAdjudication : TFhirElement
    The adjudications results.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailAdjudication}
  TFhirClaimResponseItemDetailAdjudication = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FAmount : TFhirQuantity;
    FValue : TFhirDecimal;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetAmount(value : TFhirQuantity);
    Procedure SetValue(value : TFhirDecimal);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItemDetailAdjudication; overload;
    function Clone : TFhirClaimResponseItemDetailAdjudication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member amount
      Typed access to Monetary amount associated with the code. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Monetary amount associated with the code.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

    {@member value
      Typed access to A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property valueElement : TFhirDecimal read FValue write SetValue;

  end;


  TFhirClaimResponseItemDetailAdjudicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemDetailAdjudicationList;
    function GetCurrent : TFhirClaimResponseItemDetailAdjudication;
  public
    Constructor Create(list : TFhirClaimResponseItemDetailAdjudicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItemDetailAdjudication read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemDetailAdjudicationList
    A list of FhirClaimResponseItemDetailAdjudication
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailAdjudicationList}
  TFhirClaimResponseItemDetailAdjudicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItemDetailAdjudication;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItemDetailAdjudication);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemDetailAdjudicationList; Overload;
    function Clone : TFhirClaimResponseItemDetailAdjudicationList; Overload;
    function GetEnumerator : TFhirClaimResponseItemDetailAdjudicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItemDetailAdjudication to the end of the list.
    }
    function Append : TFhirClaimResponseItemDetailAdjudication;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItemDetailAdjudication to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItemDetailAdjudication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItemDetailAdjudication) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItemDetailAdjudication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItemDetailAdjudication;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItemDetailAdjudication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItemDetailAdjudication);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailAdjudication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailAdjudication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItemDetailAdjudication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItemDetailAdjudication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItemDetailAdjudications[index : Integer] : TFhirClaimResponseItemDetailAdjudication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItemDetailSubDetail : TFhirElement
    The third tier service adjudications for submitted services.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailSubDetail}
  TFhirClaimResponseItemDetailSubDetail = class (TFhirBackboneElement)
  private
    FSequenceLinkId : TFhirPositiveInt;
    FadjudicationList : TFhirClaimResponseItemDetailSubDetailAdjudicationList;
    Procedure SetSequenceLinkId(value : TFhirPositiveInt);
    Function GetSequenceLinkIdST : String;
    Procedure SetSequenceLinkIdST(value : String);
    function GetAdjudicationList : TFhirClaimResponseItemDetailSubDetailAdjudicationList;
    function GetHasAdjudicationList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItemDetailSubDetail; overload;
    function Clone : TFhirClaimResponseItemDetailSubDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkId
      Typed access to A service line number.
    }
    property sequenceLinkId : String read GetSequenceLinkIdST write SetSequenceLinkIdST;
    {@member sequenceLinkIdElement
      A service line number.
    }
    property sequenceLinkIdElement : TFhirPositiveInt read FSequenceLinkId write SetSequenceLinkId;

    {@member adjudicationList
      The adjudications results.
    }
    property adjudicationList : TFhirClaimResponseItemDetailSubDetailAdjudicationList read GetAdjudicationList;
    property hasAdjudicationList : boolean read GetHasAdjudicationList;

  end;


  TFhirClaimResponseItemDetailSubDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemDetailSubDetailList;
    function GetCurrent : TFhirClaimResponseItemDetailSubDetail;
  public
    Constructor Create(list : TFhirClaimResponseItemDetailSubDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItemDetailSubDetail read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemDetailSubDetailList
    A list of FhirClaimResponseItemDetailSubDetail
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailSubDetailList}
  TFhirClaimResponseItemDetailSubDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItemDetailSubDetail;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItemDetailSubDetail);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemDetailSubDetailList; Overload;
    function Clone : TFhirClaimResponseItemDetailSubDetailList; Overload;
    function GetEnumerator : TFhirClaimResponseItemDetailSubDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItemDetailSubDetail to the end of the list.
    }
    function Append : TFhirClaimResponseItemDetailSubDetail;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItemDetailSubDetail to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItemDetailSubDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItemDetailSubDetail) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItemDetailSubDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItemDetailSubDetail;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItemDetailSubDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItemDetailSubDetail);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailSubDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailSubDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItemDetailSubDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItemDetailSubDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItemDetailSubDetails[index : Integer] : TFhirClaimResponseItemDetailSubDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseItemDetailSubDetailAdjudication : TFhirElement
    The adjudications results.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailSubDetailAdjudication}
  TFhirClaimResponseItemDetailSubDetailAdjudication = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FAmount : TFhirQuantity;
    FValue : TFhirDecimal;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetAmount(value : TFhirQuantity);
    Procedure SetValue(value : TFhirDecimal);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseItemDetailSubDetailAdjudication; overload;
    function Clone : TFhirClaimResponseItemDetailSubDetailAdjudication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member amount
      Typed access to Monetary amount associated with the code. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Monetary amount associated with the code.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

    {@member value
      Typed access to A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property valueElement : TFhirDecimal read FValue write SetValue;

  end;


  TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseItemDetailSubDetailAdjudicationList;
    function GetCurrent : TFhirClaimResponseItemDetailSubDetailAdjudication;
  public
    Constructor Create(list : TFhirClaimResponseItemDetailSubDetailAdjudicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseItemDetailSubDetailAdjudication read GetCurrent;
  end;


  {@Class TFhirClaimResponseItemDetailSubDetailAdjudicationList
    A list of FhirClaimResponseItemDetailSubDetailAdjudication
  }
  {!.Net HL7Connect.Fhir.ClaimResponseItemDetailSubDetailAdjudicationList}
  TFhirClaimResponseItemDetailSubDetailAdjudicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseItemDetailSubDetailAdjudication;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseItemDetailSubDetailAdjudication);
  public
    {!script hide}
    function Link : TFhirClaimResponseItemDetailSubDetailAdjudicationList; Overload;
    function Clone : TFhirClaimResponseItemDetailSubDetailAdjudicationList; Overload;
    function GetEnumerator : TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseItemDetailSubDetailAdjudication to the end of the list.
    }
    function Append : TFhirClaimResponseItemDetailSubDetailAdjudication;

    
    {@member AddItem
      Add an already existing FhirClaimResponseItemDetailSubDetailAdjudication to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseItemDetailSubDetailAdjudication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseItemDetailSubDetailAdjudication) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseItemDetailSubDetailAdjudication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseItemDetailSubDetailAdjudication;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseItemDetailSubDetailAdjudication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseItemDetailSubDetailAdjudication);
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailSubDetailAdjudication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseItemDetailSubDetailAdjudication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseItemDetailSubDetailAdjudication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseItemDetailSubDetailAdjudication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseItemDetailSubDetailAdjudications[index : Integer] : TFhirClaimResponseItemDetailSubDetailAdjudication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseAddItem : TFhirElement
    The first tier service adjudications for payor added services.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItem}
  TFhirClaimResponseAddItem = class (TFhirBackboneElement)
  private
    FsequenceLinkIdList : TFhirPositiveIntList;
    FService : TFhirCoding;
    FFee : TFhirQuantity;
    FnoteNumberLinkIdList : TFhirPositiveIntList;
    FadjudicationList : TFhirClaimResponseAddItemAdjudicationList;
    FdetailList : TFhirClaimResponseAddItemDetailList;
    function GetSequenceLinkIdList : TFhirPositiveIntList;
    function GetHasSequenceLinkIdList : Boolean;
    Procedure SetService(value : TFhirCoding);
    Procedure SetFee(value : TFhirQuantity);
    function GetNoteNumberLinkIdList : TFhirPositiveIntList;
    function GetHasNoteNumberLinkIdList : Boolean;
    function GetAdjudicationList : TFhirClaimResponseAddItemAdjudicationList;
    function GetHasAdjudicationList : Boolean;
    function GetDetailList : TFhirClaimResponseAddItemDetailList;
    function GetHasDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseAddItem; overload;
    function Clone : TFhirClaimResponseAddItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkIdList
      List of input service items which this service line is intended to replace.
    }
    property sequenceLinkIdList : TFhirPositiveIntList read GetSequenceLinkIdList;
    property hasSequenceLinkIdList : boolean read GetHasSequenceLinkIdList;

    {@member service
      Typed access to A code to indicate the Professional Service or Product supplied. (defined for API consistency)
    }
    property service : TFhirCoding read FService write SetService;
    {@member serviceElement
      A code to indicate the Professional Service or Product supplied.
    }
    property serviceElement : TFhirCoding read FService write SetService;

    {@member fee
      Typed access to The fee charged for the professional service or product.. (defined for API consistency)
    }
    property fee : TFhirQuantity read FFee write SetFee;
    {@member feeElement
      The fee charged for the professional service or product..
    }
    property feeElement : TFhirQuantity read FFee write SetFee;

    {@member noteNumberLinkIdList
      A list of note references to the notes provided below.
    }
    property noteNumberLinkIdList : TFhirPositiveIntList read GetNoteNumberLinkIdList;
    property hasNoteNumberLinkIdList : boolean read GetHasNoteNumberLinkIdList;

    {@member adjudicationList
      The adjudications results.
    }
    property adjudicationList : TFhirClaimResponseAddItemAdjudicationList read GetAdjudicationList;
    property hasAdjudicationList : boolean read GetHasAdjudicationList;

    {@member detailList
      The second tier service adjudications for payor added services.
    }
    property detailList : TFhirClaimResponseAddItemDetailList read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirClaimResponseAddItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseAddItemList;
    function GetCurrent : TFhirClaimResponseAddItem;
  public
    Constructor Create(list : TFhirClaimResponseAddItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseAddItem read GetCurrent;
  end;


  {@Class TFhirClaimResponseAddItemList
    A list of FhirClaimResponseAddItem
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemList}
  TFhirClaimResponseAddItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseAddItem;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseAddItem);
  public
    {!script hide}
    function Link : TFhirClaimResponseAddItemList; Overload;
    function Clone : TFhirClaimResponseAddItemList; Overload;
    function GetEnumerator : TFhirClaimResponseAddItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseAddItem to the end of the list.
    }
    function Append : TFhirClaimResponseAddItem;

    
    {@member AddItem
      Add an already existing FhirClaimResponseAddItem to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseAddItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseAddItem) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseAddItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseAddItem;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseAddItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseAddItem);
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseAddItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseAddItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseAddItems[index : Integer] : TFhirClaimResponseAddItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseAddItemAdjudication : TFhirElement
    The adjudications results.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemAdjudication}
  TFhirClaimResponseAddItemAdjudication = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FAmount : TFhirQuantity;
    FValue : TFhirDecimal;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetAmount(value : TFhirQuantity);
    Procedure SetValue(value : TFhirDecimal);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseAddItemAdjudication; overload;
    function Clone : TFhirClaimResponseAddItemAdjudication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member amount
      Typed access to Monetary amount associated with the code. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Monetary amount associated with the code.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

    {@member value
      Typed access to A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property valueElement : TFhirDecimal read FValue write SetValue;

  end;


  TFhirClaimResponseAddItemAdjudicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseAddItemAdjudicationList;
    function GetCurrent : TFhirClaimResponseAddItemAdjudication;
  public
    Constructor Create(list : TFhirClaimResponseAddItemAdjudicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseAddItemAdjudication read GetCurrent;
  end;


  {@Class TFhirClaimResponseAddItemAdjudicationList
    A list of FhirClaimResponseAddItemAdjudication
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemAdjudicationList}
  TFhirClaimResponseAddItemAdjudicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseAddItemAdjudication;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseAddItemAdjudication);
  public
    {!script hide}
    function Link : TFhirClaimResponseAddItemAdjudicationList; Overload;
    function Clone : TFhirClaimResponseAddItemAdjudicationList; Overload;
    function GetEnumerator : TFhirClaimResponseAddItemAdjudicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseAddItemAdjudication to the end of the list.
    }
    function Append : TFhirClaimResponseAddItemAdjudication;

    
    {@member AddItem
      Add an already existing FhirClaimResponseAddItemAdjudication to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseAddItemAdjudication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseAddItemAdjudication) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseAddItemAdjudication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseAddItemAdjudication;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseAddItemAdjudication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseAddItemAdjudication);
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemAdjudication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemAdjudication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseAddItemAdjudication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseAddItemAdjudication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseAddItemAdjudications[index : Integer] : TFhirClaimResponseAddItemAdjudication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseAddItemDetail : TFhirElement
    The second tier service adjudications for payor added services.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemDetail}
  TFhirClaimResponseAddItemDetail = class (TFhirBackboneElement)
  private
    FService : TFhirCoding;
    FFee : TFhirQuantity;
    FadjudicationList : TFhirClaimResponseAddItemDetailAdjudicationList;
    Procedure SetService(value : TFhirCoding);
    Procedure SetFee(value : TFhirQuantity);
    function GetAdjudicationList : TFhirClaimResponseAddItemDetailAdjudicationList;
    function GetHasAdjudicationList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseAddItemDetail; overload;
    function Clone : TFhirClaimResponseAddItemDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member service
      Typed access to A code to indicate the Professional Service or Product supplied. (defined for API consistency)
    }
    property service : TFhirCoding read FService write SetService;
    {@member serviceElement
      A code to indicate the Professional Service or Product supplied.
    }
    property serviceElement : TFhirCoding read FService write SetService;

    {@member fee
      Typed access to The fee charged for the professional service or product.. (defined for API consistency)
    }
    property fee : TFhirQuantity read FFee write SetFee;
    {@member feeElement
      The fee charged for the professional service or product..
    }
    property feeElement : TFhirQuantity read FFee write SetFee;

    {@member adjudicationList
      The adjudications results.
    }
    property adjudicationList : TFhirClaimResponseAddItemDetailAdjudicationList read GetAdjudicationList;
    property hasAdjudicationList : boolean read GetHasAdjudicationList;

  end;


  TFhirClaimResponseAddItemDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseAddItemDetailList;
    function GetCurrent : TFhirClaimResponseAddItemDetail;
  public
    Constructor Create(list : TFhirClaimResponseAddItemDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseAddItemDetail read GetCurrent;
  end;


  {@Class TFhirClaimResponseAddItemDetailList
    A list of FhirClaimResponseAddItemDetail
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemDetailList}
  TFhirClaimResponseAddItemDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseAddItemDetail;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseAddItemDetail);
  public
    {!script hide}
    function Link : TFhirClaimResponseAddItemDetailList; Overload;
    function Clone : TFhirClaimResponseAddItemDetailList; Overload;
    function GetEnumerator : TFhirClaimResponseAddItemDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseAddItemDetail to the end of the list.
    }
    function Append : TFhirClaimResponseAddItemDetail;

    
    {@member AddItem
      Add an already existing FhirClaimResponseAddItemDetail to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseAddItemDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseAddItemDetail) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseAddItemDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseAddItemDetail;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseAddItemDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseAddItemDetail);
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseAddItemDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseAddItemDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseAddItemDetails[index : Integer] : TFhirClaimResponseAddItemDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseAddItemDetailAdjudication : TFhirElement
    The adjudications results.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemDetailAdjudication}
  TFhirClaimResponseAddItemDetailAdjudication = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FAmount : TFhirQuantity;
    FValue : TFhirDecimal;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetAmount(value : TFhirQuantity);
    Procedure SetValue(value : TFhirDecimal);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseAddItemDetailAdjudication; overload;
    function Clone : TFhirClaimResponseAddItemDetailAdjudication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      Code indicating: Co-Pay, deductible, eligible, benefit, tax, etc.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member amount
      Typed access to Monetary amount associated with the code. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Monetary amount associated with the code.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

    {@member value
      Typed access to A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      A non-monetary value for example a percentage. Mutually exclusive to the amount element above.
    }
    property valueElement : TFhirDecimal read FValue write SetValue;

  end;


  TFhirClaimResponseAddItemDetailAdjudicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseAddItemDetailAdjudicationList;
    function GetCurrent : TFhirClaimResponseAddItemDetailAdjudication;
  public
    Constructor Create(list : TFhirClaimResponseAddItemDetailAdjudicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseAddItemDetailAdjudication read GetCurrent;
  end;


  {@Class TFhirClaimResponseAddItemDetailAdjudicationList
    A list of FhirClaimResponseAddItemDetailAdjudication
  }
  {!.Net HL7Connect.Fhir.ClaimResponseAddItemDetailAdjudicationList}
  TFhirClaimResponseAddItemDetailAdjudicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseAddItemDetailAdjudication;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseAddItemDetailAdjudication);
  public
    {!script hide}
    function Link : TFhirClaimResponseAddItemDetailAdjudicationList; Overload;
    function Clone : TFhirClaimResponseAddItemDetailAdjudicationList; Overload;
    function GetEnumerator : TFhirClaimResponseAddItemDetailAdjudicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseAddItemDetailAdjudication to the end of the list.
    }
    function Append : TFhirClaimResponseAddItemDetailAdjudication;

    
    {@member AddItem
      Add an already existing FhirClaimResponseAddItemDetailAdjudication to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseAddItemDetailAdjudication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseAddItemDetailAdjudication) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseAddItemDetailAdjudication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseAddItemDetailAdjudication;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseAddItemDetailAdjudication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseAddItemDetailAdjudication);
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemDetailAdjudication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseAddItemDetailAdjudication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseAddItemDetailAdjudication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseAddItemDetailAdjudication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseAddItemDetailAdjudications[index : Integer] : TFhirClaimResponseAddItemDetailAdjudication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseError : TFhirElement
    Mutually exclusive with Services Provided (Item).
  }
  {!.Net HL7Connect.Fhir.ClaimResponseError}
  TFhirClaimResponseError = class (TFhirBackboneElement)
  private
    FSequenceLinkId : TFhirPositiveInt;
    FDetailSequenceLinkId : TFhirPositiveInt;
    FSubdetailSequenceLinkId : TFhirPositiveInt;
    FCode : TFhirCoding;
    Procedure SetSequenceLinkId(value : TFhirPositiveInt);
    Function GetSequenceLinkIdST : String;
    Procedure SetSequenceLinkIdST(value : String);
    Procedure SetDetailSequenceLinkId(value : TFhirPositiveInt);
    Function GetDetailSequenceLinkIdST : String;
    Procedure SetDetailSequenceLinkIdST(value : String);
    Procedure SetSubdetailSequenceLinkId(value : TFhirPositiveInt);
    Function GetSubdetailSequenceLinkIdST : String;
    Procedure SetSubdetailSequenceLinkIdST(value : String);
    Procedure SetCode(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseError; overload;
    function Clone : TFhirClaimResponseError; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkId
      Typed access to The sequence number of the line item submitted which contains the error. This value is omitted when the error is elsewhere.
    }
    property sequenceLinkId : String read GetSequenceLinkIdST write SetSequenceLinkIdST;
    {@member sequenceLinkIdElement
      The sequence number of the line item submitted which contains the error. This value is omitted when the error is elsewhere.
    }
    property sequenceLinkIdElement : TFhirPositiveInt read FSequenceLinkId write SetSequenceLinkId;

    {@member detailSequenceLinkId
      Typed access to The sequence number of the addition within the line item submitted which contains the error. This value is omitted when the error is not related to an Addition.
    }
    property detailSequenceLinkId : String read GetDetailSequenceLinkIdST write SetDetailSequenceLinkIdST;
    {@member detailSequenceLinkIdElement
      The sequence number of the addition within the line item submitted which contains the error. This value is omitted when the error is not related to an Addition.
    }
    property detailSequenceLinkIdElement : TFhirPositiveInt read FDetailSequenceLinkId write SetDetailSequenceLinkId;

    {@member subdetailSequenceLinkId
      Typed access to The sequence number of the addition within the line item submitted which contains the error. This value is omitted when the error is not related to an Addition.
    }
    property subdetailSequenceLinkId : String read GetSubdetailSequenceLinkIdST write SetSubdetailSequenceLinkIdST;
    {@member subdetailSequenceLinkIdElement
      The sequence number of the addition within the line item submitted which contains the error. This value is omitted when the error is not related to an Addition.
    }
    property subdetailSequenceLinkIdElement : TFhirPositiveInt read FSubdetailSequenceLinkId write SetSubdetailSequenceLinkId;

    {@member code
      Typed access to An error code,froma specified code system, which details why the claim could not be adjudicated. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      An error code,froma specified code system, which details why the claim could not be adjudicated.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

  end;


  TFhirClaimResponseErrorListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseErrorList;
    function GetCurrent : TFhirClaimResponseError;
  public
    Constructor Create(list : TFhirClaimResponseErrorList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseError read GetCurrent;
  end;


  {@Class TFhirClaimResponseErrorList
    A list of FhirClaimResponseError
  }
  {!.Net HL7Connect.Fhir.ClaimResponseErrorList}
  TFhirClaimResponseErrorList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseError;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseError);
  public
    {!script hide}
    function Link : TFhirClaimResponseErrorList; Overload;
    function Clone : TFhirClaimResponseErrorList; Overload;
    function GetEnumerator : TFhirClaimResponseErrorListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseError to the end of the list.
    }
    function Append : TFhirClaimResponseError;

    
    {@member AddItem
      Add an already existing FhirClaimResponseError to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseError); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseError) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseError before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseError;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseError before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseError);
    
    {@member Item
       Get the iIndexth FhirClaimResponseError. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseError. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseError);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseError;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseErrors[index : Integer] : TFhirClaimResponseError read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseNote : TFhirElement
    Note text.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseNote}
  TFhirClaimResponseNote = class (TFhirBackboneElement)
  private
    FNumber : TFhirPositiveInt;
    FType_ : TFhirCoding;
    FText : TFhirString;
    Procedure SetNumber(value : TFhirPositiveInt);
    Function GetNumberST : String;
    Procedure SetNumberST(value : String);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseNote; overload;
    function Clone : TFhirClaimResponseNote; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member number
      Typed access to An integer associated with each note which may be referred to from each service line item.
    }
    property number : String read GetNumberST write SetNumberST;
    {@member numberElement
      An integer associated with each note which may be referred to from each service line item.
    }
    property numberElement : TFhirPositiveInt read FNumber write SetNumber;

    {@member type_
      Typed access to The note purpose: Print/Display. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The note purpose: Print/Display.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member text
      Typed access to The note text.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      The note text.
    }
    property textElement : TFhirString read FText write SetText;

  end;


  TFhirClaimResponseNoteListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseNoteList;
    function GetCurrent : TFhirClaimResponseNote;
  public
    Constructor Create(list : TFhirClaimResponseNoteList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseNote read GetCurrent;
  end;


  {@Class TFhirClaimResponseNoteList
    A list of FhirClaimResponseNote
  }
  {!.Net HL7Connect.Fhir.ClaimResponseNoteList}
  TFhirClaimResponseNoteList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseNote;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseNote);
  public
    {!script hide}
    function Link : TFhirClaimResponseNoteList; Overload;
    function Clone : TFhirClaimResponseNoteList; Overload;
    function GetEnumerator : TFhirClaimResponseNoteListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseNote to the end of the list.
    }
    function Append : TFhirClaimResponseNote;

    
    {@member AddItem
      Add an already existing FhirClaimResponseNote to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseNote); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseNote) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseNote before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseNote;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseNote before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseNote);
    
    {@member Item
       Get the iIndexth FhirClaimResponseNote. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseNote. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseNote);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseNote;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseNotes[index : Integer] : TFhirClaimResponseNote read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponseCoverage : TFhirElement
    Financial instrument by which payment information for health care.
  }
  {!.Net HL7Connect.Fhir.ClaimResponseCoverage}
  TFhirClaimResponseCoverage = class (TFhirBackboneElement)
  private
    FSequence : TFhirPositiveInt;
    FFocal : TFhirBoolean;
    FCoverage : TFhirReference{TFhirCoverage};
    FBusinessArrangement : TFhirString;
    FRelationship : TFhirCoding;
    FpreAuthRefList : TFhirStringList;
    FClaimResponse : TFhirReference{TFhirClaimResponse};
    FOriginalRuleset : TFhirCoding;
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetFocal(value : TFhirBoolean);
    Function GetFocalST : Boolean;
    Procedure SetFocalST(value : Boolean);
    Procedure SetCoverage(value : TFhirReference{TFhirCoverage});
    Procedure SetBusinessArrangement(value : TFhirString);
    Function GetBusinessArrangementST : String;
    Procedure SetBusinessArrangementST(value : String);
    Procedure SetRelationship(value : TFhirCoding);
    function GetPreAuthRefList : TFhirStringList;
    function GetHasPreAuthRefList : Boolean;
    Procedure SetClaimResponse(value : TFhirReference{TFhirClaimResponse});
    Procedure SetOriginalRuleset(value : TFhirCoding);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponseCoverage; overload;
    function Clone : TFhirClaimResponseCoverage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequence
      Typed access to A service line item.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      A service line item.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member focal
      Typed access to The instance number of the Coverage which is the focus for adjudication. The Coverage against which the claim is to be adjudicated.
    }
    property focal : Boolean read GetFocalST write SetFocalST;
    {@member focalElement
      The instance number of the Coverage which is the focus for adjudication. The Coverage against which the claim is to be adjudicated.
    }
    property focalElement : TFhirBoolean read FFocal write SetFocal;

    {@member coverage
      Typed access to Reference to the program or plan identification, underwriter or payor. (defined for API consistency)
    }
    property coverage : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;
    {@member coverageElement
      Reference to the program or plan identification, underwriter or payor.
    }
    property coverageElement : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;

    {@member businessArrangement
      Typed access to The contract number of a business agreement which describes the terms and conditions.
    }
    property businessArrangement : String read GetBusinessArrangementST write SetBusinessArrangementST;
    {@member businessArrangementElement
      The contract number of a business agreement which describes the terms and conditions.
    }
    property businessArrangementElement : TFhirString read FBusinessArrangement write SetBusinessArrangement;

    {@member relationship
      Typed access to The relationship of the patient to the subscriber. (defined for API consistency)
    }
    property relationship : TFhirCoding read FRelationship write SetRelationship;
    {@member relationshipElement
      The relationship of the patient to the subscriber.
    }
    property relationshipElement : TFhirCoding read FRelationship write SetRelationship;

    {@member preAuthRefList
      A list of references from the Insurer to which these services pertain.
    }
    property preAuthRefList : TFhirStringList read GetPreAuthRefList;
    property hasPreAuthRefList : boolean read GetHasPreAuthRefList;

    {@member claimResponse
      Typed access to The Coverages adjudication details. (defined for API consistency)
    }
    property claimResponse : TFhirReference{TFhirClaimResponse} read FClaimResponse write SetClaimResponse;
    {@member claimResponseElement
      The Coverages adjudication details.
    }
    property claimResponseElement : TFhirReference{TFhirClaimResponse} read FClaimResponse write SetClaimResponse;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

  end;


  TFhirClaimResponseCoverageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseCoverageList;
    function GetCurrent : TFhirClaimResponseCoverage;
  public
    Constructor Create(list : TFhirClaimResponseCoverageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponseCoverage read GetCurrent;
  end;


  {@Class TFhirClaimResponseCoverageList
    A list of FhirClaimResponseCoverage
  }
  {!.Net HL7Connect.Fhir.ClaimResponseCoverageList}
  TFhirClaimResponseCoverageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponseCoverage;
    procedure SetItemN(index : Integer; value : TFhirClaimResponseCoverage);
  public
    {!script hide}
    function Link : TFhirClaimResponseCoverageList; Overload;
    function Clone : TFhirClaimResponseCoverageList; Overload;
    function GetEnumerator : TFhirClaimResponseCoverageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponseCoverage to the end of the list.
    }
    function Append : TFhirClaimResponseCoverage;

    
    {@member AddItem
      Add an already existing FhirClaimResponseCoverage to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponseCoverage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponseCoverage) : Integer;
    

    {@member Insert
      Insert FhirClaimResponseCoverage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponseCoverage;
    

    {@member InsertItem
       Insert an existing FhirClaimResponseCoverage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponseCoverage);
    
    {@member Item
       Get the iIndexth FhirClaimResponseCoverage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponseCoverage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponseCoverage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponseCoverage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponseCoverages[index : Integer] : TFhirClaimResponseCoverage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClaimResponse : TFhirDomainResource
    This resource provides the adjudication details from the processing of a Claim resource.
  }
  {!.Net HL7Connect.Fhir.ClaimResponse}
  TFhirClaimResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirClaim};
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    FOutcome : TFhirEnum;
    FDisposition : TFhirString;
    FPayeeType : TFhirCoding;
    FitemList : TFhirClaimResponseItemList;
    FaddItemList : TFhirClaimResponseAddItemList;
    FerrorList : TFhirClaimResponseErrorList;
    FTotalCost : TFhirQuantity;
    FUnallocDeductable : TFhirQuantity;
    FTotalBenefit : TFhirQuantity;
    FPaymentAdjustment : TFhirQuantity;
    FPaymentAdjustmentReason : TFhirCoding;
    FPaymentDate : TFhirDate;
    FPaymentAmount : TFhirQuantity;
    FPaymentRef : TFhirIdentifier;
    FReserved : TFhirCoding;
    FForm : TFhirCoding;
    FnoteList : TFhirClaimResponseNoteList;
    FcoverageList : TFhirClaimResponseCoverageList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirClaim});
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirRemittanceOutcome;
    Procedure SetOutcomeST(value : TFhirRemittanceOutcome);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetPayeeType(value : TFhirCoding);
    function GetItemList : TFhirClaimResponseItemList;
    function GetHasItemList : Boolean;
    function GetAddItemList : TFhirClaimResponseAddItemList;
    function GetHasAddItemList : Boolean;
    function GetErrorList : TFhirClaimResponseErrorList;
    function GetHasErrorList : Boolean;
    Procedure SetTotalCost(value : TFhirQuantity);
    Procedure SetUnallocDeductable(value : TFhirQuantity);
    Procedure SetTotalBenefit(value : TFhirQuantity);
    Procedure SetPaymentAdjustment(value : TFhirQuantity);
    Procedure SetPaymentAdjustmentReason(value : TFhirCoding);
    Procedure SetPaymentDate(value : TFhirDate);
    Function GetPaymentDateST : TDateAndTime;
    Procedure SetPaymentDateST(value : TDateAndTime);
    Procedure SetPaymentAmount(value : TFhirQuantity);
    Procedure SetPaymentRef(value : TFhirIdentifier);
    Procedure SetReserved(value : TFhirCoding);
    Procedure SetForm(value : TFhirCoding);
    function GetNoteList : TFhirClaimResponseNoteList;
    function GetHasNoteList : Boolean;
    function GetCoverageList : TFhirClaimResponseCoverageList;
    function GetHasCoverageList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClaimResponse; overload;
    function Clone : TFhirClaimResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource referrence. (defined for API consistency)
    }
    property request : TFhirReference{TFhirClaim} read FRequest write SetRequest;
    {@member requestElement
      Original request resource referrence.
    }
    property requestElement : TFhirReference{TFhirClaim} read FRequest write SetRequest;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member organization
      Typed access to The Insurer who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The Insurer who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

    {@member outcome
      Transaction status: error, complete.
    }
    property outcome : TFhirRemittanceOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member payeeType
      Typed access to Party to be reimbursed: Subscriber, provider, other. (defined for API consistency)
    }
    property payeeType : TFhirCoding read FPayeeType write SetPayeeType;
    {@member payeeTypeElement
      Party to be reimbursed: Subscriber, provider, other.
    }
    property payeeTypeElement : TFhirCoding read FPayeeType write SetPayeeType;

    {@member itemList
      The first tier service adjudications for submitted services.
    }
    property itemList : TFhirClaimResponseItemList read GetItemList;
    property hasItemList : boolean read GetHasItemList;

    {@member addItemList
      The first tier service adjudications for payor added services.
    }
    property addItemList : TFhirClaimResponseAddItemList read GetAddItemList;
    property hasAddItemList : boolean read GetHasAddItemList;

    {@member errorList
      Mutually exclusive with Services Provided (Item).
    }
    property errorList : TFhirClaimResponseErrorList read GetErrorList;
    property hasErrorList : boolean read GetHasErrorList;

    {@member totalCost
      Typed access to The total cost of the services reported. (defined for API consistency)
    }
    property totalCost : TFhirQuantity read FTotalCost write SetTotalCost;
    {@member totalCostElement
      The total cost of the services reported.
    }
    property totalCostElement : TFhirQuantity read FTotalCost write SetTotalCost;

    {@member unallocDeductable
      Typed access to The amount of deductible applied which was not allocated to any particular service line. (defined for API consistency)
    }
    property unallocDeductable : TFhirQuantity read FUnallocDeductable write SetUnallocDeductable;
    {@member unallocDeductableElement
      The amount of deductible applied which was not allocated to any particular service line.
    }
    property unallocDeductableElement : TFhirQuantity read FUnallocDeductable write SetUnallocDeductable;

    {@member totalBenefit
      Typed access to Total amount of benefit payable (Equal to sum of the Benefit amounts from all detail lines and additions less the Unallocated Deductible). (defined for API consistency)
    }
    property totalBenefit : TFhirQuantity read FTotalBenefit write SetTotalBenefit;
    {@member totalBenefitElement
      Total amount of benefit payable (Equal to sum of the Benefit amounts from all detail lines and additions less the Unallocated Deductible).
    }
    property totalBenefitElement : TFhirQuantity read FTotalBenefit write SetTotalBenefit;

    {@member paymentAdjustment
      Typed access to Adjustment to the payment of this transaction which is not related to adjudication of this transaction. (defined for API consistency)
    }
    property paymentAdjustment : TFhirQuantity read FPaymentAdjustment write SetPaymentAdjustment;
    {@member paymentAdjustmentElement
      Adjustment to the payment of this transaction which is not related to adjudication of this transaction.
    }
    property paymentAdjustmentElement : TFhirQuantity read FPaymentAdjustment write SetPaymentAdjustment;

    {@member paymentAdjustmentReason
      Typed access to Reason for the payment adjustment. (defined for API consistency)
    }
    property paymentAdjustmentReason : TFhirCoding read FPaymentAdjustmentReason write SetPaymentAdjustmentReason;
    {@member paymentAdjustmentReasonElement
      Reason for the payment adjustment.
    }
    property paymentAdjustmentReasonElement : TFhirCoding read FPaymentAdjustmentReason write SetPaymentAdjustmentReason;

    {@member paymentDate
      Typed access to Estimated payment data.
    }
    property paymentDate : TDateAndTime read GetPaymentDateST write SetPaymentDateST;
    {@member paymentDateElement
      Estimated payment data.
    }
    property paymentDateElement : TFhirDate read FPaymentDate write SetPaymentDate;

    {@member paymentAmount
      Typed access to Payable less any payment adjustment. (defined for API consistency)
    }
    property paymentAmount : TFhirQuantity read FPaymentAmount write SetPaymentAmount;
    {@member paymentAmountElement
      Payable less any payment adjustment.
    }
    property paymentAmountElement : TFhirQuantity read FPaymentAmount write SetPaymentAmount;

    {@member paymentRef
      Typed access to Payment identifier. (defined for API consistency)
    }
    property paymentRef : TFhirIdentifier read FPaymentRef write SetPaymentRef;
    {@member paymentRefElement
      Payment identifier.
    }
    property paymentRefElement : TFhirIdentifier read FPaymentRef write SetPaymentRef;

    {@member reserved
      Typed access to Status of funds reservation (For provider, for Patient, None). (defined for API consistency)
    }
    property reserved : TFhirCoding read FReserved write SetReserved;
    {@member reservedElement
      Status of funds reservation (For provider, for Patient, None).
    }
    property reservedElement : TFhirCoding read FReserved write SetReserved;

    {@member form
      Typed access to The form to be used for printing the content. (defined for API consistency)
    }
    property form : TFhirCoding read FForm write SetForm;
    {@member formElement
      The form to be used for printing the content.
    }
    property formElement : TFhirCoding read FForm write SetForm;

    {@member noteList
      Note text.
    }
    property noteList : TFhirClaimResponseNoteList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

    {@member coverageList
      Financial instrument by which payment information for health care.
    }
    property coverageList : TFhirClaimResponseCoverageList read GetCoverageList;
    property hasCoverageList : boolean read GetHasCoverageList;

  end;


  TFhirClaimResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClaimResponseList;
    function GetCurrent : TFhirClaimResponse;
  public
    Constructor Create(list : TFhirClaimResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClaimResponse read GetCurrent;
  end;


  {@Class TFhirClaimResponseList
    A list of FhirClaimResponse
  }
  {!.Net HL7Connect.Fhir.ClaimResponseList}
  TFhirClaimResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClaimResponse;
    procedure SetItemN(index : Integer; value : TFhirClaimResponse);
  public
    {!script hide}
    function Link : TFhirClaimResponseList; Overload;
    function Clone : TFhirClaimResponseList; Overload;
    function GetEnumerator : TFhirClaimResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClaimResponse to the end of the list.
    }
    function Append : TFhirClaimResponse;

    
    {@member AddItem
      Add an already existing FhirClaimResponse to the end of the list.
    }
    procedure AddItem(value : TFhirClaimResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClaimResponse) : Integer;
    

    {@member Insert
      Insert FhirClaimResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClaimResponse;
    

    {@member InsertItem
       Insert an existing FhirClaimResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClaimResponse);
    
    {@member Item
       Get the iIndexth FhirClaimResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClaimResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClaimResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClaimResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClaimResponses[index : Integer] : TFhirClaimResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClinicalImpressionInvestigations : TFhirElement
    One or more sets of investigations (signs, symptions, etc.). The actual grouping of investigations vary greatly depending on the type and context of the assessment. These investigations may include data generated during the assessment process, or data previously generated and recorded that is pertinent to the outcomes.
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionInvestigations}
  TFhirClinicalImpressionInvestigations = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FitemList : TFhirReferenceList{Resource};
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetItemList : TFhirReferenceList{Resource};
    function GetHasItemList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClinicalImpressionInvestigations; overload;
    function Clone : TFhirClinicalImpressionInvestigations; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A name/code for the group ("set") of investigations. Typically, this will be something like "signs", "symptoms", "clinical", "diagnostic", but the list is not constrained, and others such groups such as (exposure|family|travel|nutitirional) history may be used. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A name/code for the group ("set") of investigations. Typically, this will be something like "signs", "symptoms", "clinical", "diagnostic", but the list is not constrained, and others such groups such as (exposure|family|travel|nutitirional) history may be used.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member itemList
      A record of a specific investigation that was undertaken.
    }
    property itemList : TFhirReferenceList{Resource} read GetItemList;
    property hasItemList : boolean read GetHasItemList;

  end;


  TFhirClinicalImpressionInvestigationsListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClinicalImpressionInvestigationsList;
    function GetCurrent : TFhirClinicalImpressionInvestigations;
  public
    Constructor Create(list : TFhirClinicalImpressionInvestigationsList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClinicalImpressionInvestigations read GetCurrent;
  end;


  {@Class TFhirClinicalImpressionInvestigationsList
    A list of FhirClinicalImpressionInvestigations
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionInvestigationsList}
  TFhirClinicalImpressionInvestigationsList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClinicalImpressionInvestigations;
    procedure SetItemN(index : Integer; value : TFhirClinicalImpressionInvestigations);
  public
    {!script hide}
    function Link : TFhirClinicalImpressionInvestigationsList; Overload;
    function Clone : TFhirClinicalImpressionInvestigationsList; Overload;
    function GetEnumerator : TFhirClinicalImpressionInvestigationsListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClinicalImpressionInvestigations to the end of the list.
    }
    function Append : TFhirClinicalImpressionInvestigations;

    
    {@member AddItem
      Add an already existing FhirClinicalImpressionInvestigations to the end of the list.
    }
    procedure AddItem(value : TFhirClinicalImpressionInvestigations); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClinicalImpressionInvestigations) : Integer;
    

    {@member Insert
      Insert FhirClinicalImpressionInvestigations before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClinicalImpressionInvestigations;
    

    {@member InsertItem
       Insert an existing FhirClinicalImpressionInvestigations before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClinicalImpressionInvestigations);
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionInvestigations. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionInvestigations. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClinicalImpressionInvestigations);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClinicalImpressionInvestigations;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClinicalImpressionInvestigations[index : Integer] : TFhirClinicalImpressionInvestigations read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClinicalImpressionFinding : TFhirElement
    Specific findings or diagnoses that was considered likely or relevant to ongoing treatment.
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionFinding}
  TFhirClinicalImpressionFinding = class (TFhirBackboneElement)
  private
    FItem : TFhirCodeableConcept;
    FCause : TFhirString;
    Procedure SetItem(value : TFhirCodeableConcept);
    Procedure SetCause(value : TFhirString);
    Function GetCauseST : String;
    Procedure SetCauseST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClinicalImpressionFinding; overload;
    function Clone : TFhirClinicalImpressionFinding; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member item
      Typed access to Specific text of code for finding or diagnosis. (defined for API consistency)
    }
    property item : TFhirCodeableConcept read FItem write SetItem;
    {@member itemElement
      Specific text of code for finding or diagnosis.
    }
    property itemElement : TFhirCodeableConcept read FItem write SetItem;

    {@member cause
      Typed access to Which investigations support finding or diagnosis.
    }
    property cause : String read GetCauseST write SetCauseST;
    {@member causeElement
      Which investigations support finding or diagnosis.
    }
    property causeElement : TFhirString read FCause write SetCause;

  end;


  TFhirClinicalImpressionFindingListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClinicalImpressionFindingList;
    function GetCurrent : TFhirClinicalImpressionFinding;
  public
    Constructor Create(list : TFhirClinicalImpressionFindingList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClinicalImpressionFinding read GetCurrent;
  end;


  {@Class TFhirClinicalImpressionFindingList
    A list of FhirClinicalImpressionFinding
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionFindingList}
  TFhirClinicalImpressionFindingList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClinicalImpressionFinding;
    procedure SetItemN(index : Integer; value : TFhirClinicalImpressionFinding);
  public
    {!script hide}
    function Link : TFhirClinicalImpressionFindingList; Overload;
    function Clone : TFhirClinicalImpressionFindingList; Overload;
    function GetEnumerator : TFhirClinicalImpressionFindingListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClinicalImpressionFinding to the end of the list.
    }
    function Append : TFhirClinicalImpressionFinding;

    
    {@member AddItem
      Add an already existing FhirClinicalImpressionFinding to the end of the list.
    }
    procedure AddItem(value : TFhirClinicalImpressionFinding); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClinicalImpressionFinding) : Integer;
    

    {@member Insert
      Insert FhirClinicalImpressionFinding before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClinicalImpressionFinding;
    

    {@member InsertItem
       Insert an existing FhirClinicalImpressionFinding before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClinicalImpressionFinding);
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionFinding. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionFinding. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClinicalImpressionFinding);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClinicalImpressionFinding;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClinicalImpressionFindings[index : Integer] : TFhirClinicalImpressionFinding read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClinicalImpressionRuledOut : TFhirElement
    Diagnosis considered not possible.
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionRuledOut}
  TFhirClinicalImpressionRuledOut = class (TFhirBackboneElement)
  private
    FItem : TFhirCodeableConcept;
    FReason : TFhirString;
    Procedure SetItem(value : TFhirCodeableConcept);
    Procedure SetReason(value : TFhirString);
    Function GetReasonST : String;
    Procedure SetReasonST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClinicalImpressionRuledOut; overload;
    function Clone : TFhirClinicalImpressionRuledOut; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member item
      Typed access to Specific text of code for diagnosis. (defined for API consistency)
    }
    property item : TFhirCodeableConcept read FItem write SetItem;
    {@member itemElement
      Specific text of code for diagnosis.
    }
    property itemElement : TFhirCodeableConcept read FItem write SetItem;

    {@member reason
      Typed access to Grounds for elimination.
    }
    property reason : String read GetReasonST write SetReasonST;
    {@member reasonElement
      Grounds for elimination.
    }
    property reasonElement : TFhirString read FReason write SetReason;

  end;


  TFhirClinicalImpressionRuledOutListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClinicalImpressionRuledOutList;
    function GetCurrent : TFhirClinicalImpressionRuledOut;
  public
    Constructor Create(list : TFhirClinicalImpressionRuledOutList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClinicalImpressionRuledOut read GetCurrent;
  end;


  {@Class TFhirClinicalImpressionRuledOutList
    A list of FhirClinicalImpressionRuledOut
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionRuledOutList}
  TFhirClinicalImpressionRuledOutList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClinicalImpressionRuledOut;
    procedure SetItemN(index : Integer; value : TFhirClinicalImpressionRuledOut);
  public
    {!script hide}
    function Link : TFhirClinicalImpressionRuledOutList; Overload;
    function Clone : TFhirClinicalImpressionRuledOutList; Overload;
    function GetEnumerator : TFhirClinicalImpressionRuledOutListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClinicalImpressionRuledOut to the end of the list.
    }
    function Append : TFhirClinicalImpressionRuledOut;

    
    {@member AddItem
      Add an already existing FhirClinicalImpressionRuledOut to the end of the list.
    }
    procedure AddItem(value : TFhirClinicalImpressionRuledOut); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClinicalImpressionRuledOut) : Integer;
    

    {@member Insert
      Insert FhirClinicalImpressionRuledOut before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClinicalImpressionRuledOut;
    

    {@member InsertItem
       Insert an existing FhirClinicalImpressionRuledOut before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClinicalImpressionRuledOut);
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionRuledOut. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClinicalImpressionRuledOut. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClinicalImpressionRuledOut);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClinicalImpressionRuledOut;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClinicalImpressionRuledOuts[index : Integer] : TFhirClinicalImpressionRuledOut read GetItemN write SetItemN; default;
  End;


  {@Class TFhirClinicalImpression : TFhirDomainResource
    A record of a clinical assessment performed to determine what problem(s) may affect the patient and before planning the treatments or management strategies that are best to manage a patient's condition. Assessments are often 1:1 with a clinical consultation / encounter,  but this varies greatly depending on the clinical workflow. This resource is called "ClinicalImpression" rather than "ClinicalAssessment" to avoid confusion with the recording of assessment tools such as Apgar score.
  }
  {!.Net HL7Connect.Fhir.ClinicalImpression}
  TFhirClinicalImpression = class (TFhirDomainResource)
  private
    FPatient : TFhirReference{TFhirPatient};
    FAssessor : TFhirReference{TFhirPractitioner};
    FStatus : TFhirEnum;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FPrevious : TFhirReference{TFhirClinicalImpression};
    FproblemList : TFhirReferenceList{Resource};
    FTrigger : TFhirType;
    FinvestigationsList : TFhirClinicalImpressionInvestigationsList;
    FProtocol : TFhirUri;
    FSummary : TFhirString;
    FfindingList : TFhirClinicalImpressionFindingList;
    FresolvedList : TFhirCodeableConceptList;
    FruledOutList : TFhirClinicalImpressionRuledOutList;
    FPrognosis : TFhirString;
    FplanList : TFhirReferenceList{Resource};
    FactionList : TFhirReferenceList{Resource};
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetAssessor(value : TFhirReference{TFhirPractitioner});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirClinicalImpressionStatus;
    Procedure SetStatusST(value : TFhirClinicalImpressionStatus);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetPrevious(value : TFhirReference{TFhirClinicalImpression});
    function GetProblemList : TFhirReferenceList{Resource};
    function GetHasProblemList : Boolean;
    Procedure SetTrigger(value : TFhirType);
    function GetInvestigationsList : TFhirClinicalImpressionInvestigationsList;
    function GetHasInvestigationsList : Boolean;
    Procedure SetProtocol(value : TFhirUri);
    Function GetProtocolST : String;
    Procedure SetProtocolST(value : String);
    Procedure SetSummary(value : TFhirString);
    Function GetSummaryST : String;
    Procedure SetSummaryST(value : String);
    function GetFindingList : TFhirClinicalImpressionFindingList;
    function GetHasFindingList : Boolean;
    function GetResolvedList : TFhirCodeableConceptList;
    function GetHasResolvedList : Boolean;
    function GetRuledOutList : TFhirClinicalImpressionRuledOutList;
    function GetHasRuledOutList : Boolean;
    Procedure SetPrognosis(value : TFhirString);
    Function GetPrognosisST : String;
    Procedure SetPrognosisST(value : String);
    function GetPlanList : TFhirReferenceList{Resource};
    function GetHasPlanList : Boolean;
    function GetActionList : TFhirReferenceList{Resource};
    function GetHasActionList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirClinicalImpression; overload;
    function Clone : TFhirClinicalImpression; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member patient
      Typed access to The patient being assessed. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient being assessed.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member assessor
      Typed access to The clinician performing the assessment. (defined for API consistency)
    }
    property assessor : TFhirReference{TFhirPractitioner} read FAssessor write SetAssessor;
    {@member assessorElement
      The clinician performing the assessment.
    }
    property assessorElement : TFhirReference{TFhirPractitioner} read FAssessor write SetAssessor;

    {@member status
      Identifies the workflow status of the assessment.
    }
    property status : TFhirClinicalImpressionStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member date
      Typed access to The point in time at which the assessment was concluded (not when it was recorded).
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The point in time at which the assessment was concluded (not when it was recorded).
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A summary of the context and/or cause of the assessment - why / where was it peformed, and what patient events/sstatus prompted it.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A summary of the context and/or cause of the assessment - why / where was it peformed, and what patient events/sstatus prompted it.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member previous
      Typed access to A reference to the last assesment that was conducted bon this patient. Assessments are often/usually ongoing in nature; a care provider (practitioner or team) will make new assessments on an ongoing basis as new data arises or the patient's conditions changes. (defined for API consistency)
    }
    property previous : TFhirReference{TFhirClinicalImpression} read FPrevious write SetPrevious;
    {@member previousElement
      A reference to the last assesment that was conducted bon this patient. Assessments are often/usually ongoing in nature; a care provider (practitioner or team) will make new assessments on an ongoing basis as new data arises or the patient's conditions changes.
    }
    property previousElement : TFhirReference{TFhirClinicalImpression} read FPrevious write SetPrevious;

    {@member problemList
      This a list of the general problems/conditions for a patient.
    }
    property problemList : TFhirReferenceList{Resource} read GetProblemList;
    property hasProblemList : boolean read GetHasProblemList;

    {@member trigger
      Typed access to The request or event that necessitated this assessment. This may be a diagnosis, a Care Plan, a Request Referral, or some other resource. (defined for API consistency)
    }
    property trigger : TFhirType read FTrigger write SetTrigger;
    {@member triggerElement
      The request or event that necessitated this assessment. This may be a diagnosis, a Care Plan, a Request Referral, or some other resource.
    }
    property triggerElement : TFhirType read FTrigger write SetTrigger;

    {@member investigationsList
      One or more sets of investigations (signs, symptions, etc.). The actual grouping of investigations vary greatly depending on the type and context of the assessment. These investigations may include data generated during the assessment process, or data previously generated and recorded that is pertinent to the outcomes.
    }
    property investigationsList : TFhirClinicalImpressionInvestigationsList read GetInvestigationsList;
    property hasInvestigationsList : boolean read GetHasInvestigationsList;

    {@member protocol
      Typed access to Reference to a specific published clinical protocol that was followed during this assessment, and/or that provides evidence in support of the diagnosis.
    }
    property protocol : String read GetProtocolST write SetProtocolST;
    {@member protocolElement
      Reference to a specific published clinical protocol that was followed during this assessment, and/or that provides evidence in support of the diagnosis.
    }
    property protocolElement : TFhirUri read FProtocol write SetProtocol;

    {@member summary
      Typed access to A text summary of the investigations and the diagnosis.
    }
    property summary : String read GetSummaryST write SetSummaryST;
    {@member summaryElement
      A text summary of the investigations and the diagnosis.
    }
    property summaryElement : TFhirString read FSummary write SetSummary;

    {@member findingList
      Specific findings or diagnoses that was considered likely or relevant to ongoing treatment.
    }
    property findingList : TFhirClinicalImpressionFindingList read GetFindingList;
    property hasFindingList : boolean read GetHasFindingList;

    {@member resolvedList
      Diagnoses/conditions resolved since the last assessment.
    }
    property resolvedList : TFhirCodeableConceptList read GetResolvedList;
    property hasResolvedList : boolean read GetHasResolvedList;

    {@member ruledOutList
      Diagnosis considered not possible.
    }
    property ruledOutList : TFhirClinicalImpressionRuledOutList read GetRuledOutList;
    property hasRuledOutList : boolean read GetHasRuledOutList;

    {@member prognosis
      Typed access to Estimate of likely outcome.
    }
    property prognosis : String read GetPrognosisST write SetPrognosisST;
    {@member prognosisElement
      Estimate of likely outcome.
    }
    property prognosisElement : TFhirString read FPrognosis write SetPrognosis;

    {@member planList
      Plan of action after assessment.
    }
    property planList : TFhirReferenceList{Resource} read GetPlanList;
    property hasPlanList : boolean read GetHasPlanList;

    {@member actionList
      Actions taken during assessment.
    }
    property actionList : TFhirReferenceList{Resource} read GetActionList;
    property hasActionList : boolean read GetHasActionList;

  end;


  TFhirClinicalImpressionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirClinicalImpressionList;
    function GetCurrent : TFhirClinicalImpression;
  public
    Constructor Create(list : TFhirClinicalImpressionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirClinicalImpression read GetCurrent;
  end;


  {@Class TFhirClinicalImpressionList
    A list of FhirClinicalImpression
  }
  {!.Net HL7Connect.Fhir.ClinicalImpressionList}
  TFhirClinicalImpressionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirClinicalImpression;
    procedure SetItemN(index : Integer; value : TFhirClinicalImpression);
  public
    {!script hide}
    function Link : TFhirClinicalImpressionList; Overload;
    function Clone : TFhirClinicalImpressionList; Overload;
    function GetEnumerator : TFhirClinicalImpressionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirClinicalImpression to the end of the list.
    }
    function Append : TFhirClinicalImpression;

    
    {@member AddItem
      Add an already existing FhirClinicalImpression to the end of the list.
    }
    procedure AddItem(value : TFhirClinicalImpression); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirClinicalImpression) : Integer;
    

    {@member Insert
      Insert FhirClinicalImpression before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirClinicalImpression;
    

    {@member InsertItem
       Insert an existing FhirClinicalImpression before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirClinicalImpression);
    
    {@member Item
       Get the iIndexth FhirClinicalImpression. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirClinicalImpression. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirClinicalImpression);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirClinicalImpression;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirClinicalImpressions[index : Integer] : TFhirClinicalImpression read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCommunicationPayload : TFhirElement
    Text, attachment(s), or resource(s) that was communicated to the recipient.
  }
  {!.Net HL7Connect.Fhir.CommunicationPayload}
  TFhirCommunicationPayload = class (TFhirBackboneElement)
  private
    FContent : TFhirType;
    Procedure SetContent(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCommunicationPayload; overload;
    function Clone : TFhirCommunicationPayload; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member content
      Typed access to A communicated content (or for multi-part communications, one portion of the communication). (defined for API consistency)
    }
    property content : TFhirType read FContent write SetContent;
    {@member contentElement
      A communicated content (or for multi-part communications, one portion of the communication).
    }
    property contentElement : TFhirType read FContent write SetContent;

  end;


  TFhirCommunicationPayloadListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCommunicationPayloadList;
    function GetCurrent : TFhirCommunicationPayload;
  public
    Constructor Create(list : TFhirCommunicationPayloadList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCommunicationPayload read GetCurrent;
  end;


  {@Class TFhirCommunicationPayloadList
    A list of FhirCommunicationPayload
  }
  {!.Net HL7Connect.Fhir.CommunicationPayloadList}
  TFhirCommunicationPayloadList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCommunicationPayload;
    procedure SetItemN(index : Integer; value : TFhirCommunicationPayload);
  public
    {!script hide}
    function Link : TFhirCommunicationPayloadList; Overload;
    function Clone : TFhirCommunicationPayloadList; Overload;
    function GetEnumerator : TFhirCommunicationPayloadListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCommunicationPayload to the end of the list.
    }
    function Append : TFhirCommunicationPayload;

    
    {@member AddItem
      Add an already existing FhirCommunicationPayload to the end of the list.
    }
    procedure AddItem(value : TFhirCommunicationPayload); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCommunicationPayload) : Integer;
    

    {@member Insert
      Insert FhirCommunicationPayload before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCommunicationPayload;
    

    {@member InsertItem
       Insert an existing FhirCommunicationPayload before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCommunicationPayload);
    
    {@member Item
       Get the iIndexth FhirCommunicationPayload. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCommunicationPayload. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCommunicationPayload);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCommunicationPayload;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCommunicationPayloads[index : Integer] : TFhirCommunicationPayload read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCommunication : TFhirDomainResource
    An occurrence of information being transmitted; e.g. an alert that was sent to a responsible provider, a public health agency was notified about a reportable condition.
  }
  {!.Net HL7Connect.Fhir.Communication}
  TFhirCommunication = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FCategory : TFhirCodeableConcept;
    FSender : TFhirReference{Resource};
    FrecipientList : TFhirReferenceList{Resource};
    FpayloadList : TFhirCommunicationPayloadList;
    FmediumList : TFhirCodeableConceptList;
    FStatus : TFhirEnum;
    FEncounter : TFhirReference{TFhirEncounter};
    FSent : TFhirDateTime;
    FReceived : TFhirDateTime;
    FreasonList : TFhirCodeableConceptList;
    FSubject : TFhirReference{TFhirPatient};
    FRequestDetail : TFhirReference{TFhirCommunicationRequest};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetSender(value : TFhirReference{Resource});
    function GetRecipientList : TFhirReferenceList{Resource};
    function GetHasRecipientList : Boolean;
    function GetPayloadList : TFhirCommunicationPayloadList;
    function GetHasPayloadList : Boolean;
    function GetMediumList : TFhirCodeableConceptList;
    function GetHasMediumList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirCommunicationStatus;
    Procedure SetStatusST(value : TFhirCommunicationStatus);
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetSent(value : TFhirDateTime);
    Function GetSentST : TDateAndTime;
    Procedure SetSentST(value : TDateAndTime);
    Procedure SetReceived(value : TFhirDateTime);
    Function GetReceivedST : TDateAndTime;
    Procedure SetReceivedST(value : TDateAndTime);
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    Procedure SetSubject(value : TFhirReference{TFhirPatient});
    Procedure SetRequestDetail(value : TFhirReference{TFhirCommunicationRequest});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCommunication; overload;
    function Clone : TFhirCommunication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifiers associated with this Communication that are defined by business processes and/ or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member category
      Typed access to The type of message conveyed such as alert, notification, reminder, instruction, etc. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      The type of message conveyed such as alert, notification, reminder, instruction, etc.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member sender
      Typed access to The entity (e.g. person, organization, clinical information system, or device) which was the source of the communication. (defined for API consistency)
    }
    property sender : TFhirReference{Resource} read FSender write SetSender;
    {@member senderElement
      The entity (e.g. person, organization, clinical information system, or device) which was the source of the communication.
    }
    property senderElement : TFhirReference{Resource} read FSender write SetSender;

    {@member recipientList
      The entity (e.g. person, organization, clinical information system, or device) which was the target of the communication. If receipts need to be tracked by individual, a separate resource instance will need to be created for each recipient. ?Multiple recipient communications are intended where either a receipt(s) is not tracked (e.g. a mass mail-out) or is captured in aggregate (all emails confirmed received by a particular time).
    }
    property recipientList : TFhirReferenceList{Resource} read GetRecipientList;
    property hasRecipientList : boolean read GetHasRecipientList;

    {@member payloadList
      Text, attachment(s), or resource(s) that was communicated to the recipient.
    }
    property payloadList : TFhirCommunicationPayloadList read GetPayloadList;
    property hasPayloadList : boolean read GetHasPayloadList;

    {@member mediumList
      A channel that was used for this communication (e.g. email, fax).
    }
    property mediumList : TFhirCodeableConceptList read GetMediumList;
    property hasMediumList : boolean read GetHasMediumList;

    {@member status
      The status of the transmission.
    }
    property status : TFhirCommunicationStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member encounter
      Typed access to The encounter within which the communication was sent. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter within which the communication was sent.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member sent
      Typed access to The time when this communication was sent.
    }
    property sent : TDateAndTime read GetSentST write SetSentST;
    {@member sentElement
      The time when this communication was sent.
    }
    property sentElement : TFhirDateTime read FSent write SetSent;

    {@member received
      Typed access to The time when this communication arrived at the destination.
    }
    property received : TDateAndTime read GetReceivedST write SetReceivedST;
    {@member receivedElement
      The time when this communication arrived at the destination.
    }
    property receivedElement : TFhirDateTime read FReceived write SetReceived;

    {@member reasonList
      The reason or justification for the communication.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member subject
      Typed access to The patient who was the focus of this communication. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirPatient} read FSubject write SetSubject;
    {@member subjectElement
      The patient who was the focus of this communication.
    }
    property subjectElement : TFhirReference{TFhirPatient} read FSubject write SetSubject;

    {@member requestDetail
      Typed access to The communication request that was responsible for producing this communication. (defined for API consistency)
    }
    property requestDetail : TFhirReference{TFhirCommunicationRequest} read FRequestDetail write SetRequestDetail;
    {@member requestDetailElement
      The communication request that was responsible for producing this communication.
    }
    property requestDetailElement : TFhirReference{TFhirCommunicationRequest} read FRequestDetail write SetRequestDetail;

  end;


  TFhirCommunicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCommunicationList;
    function GetCurrent : TFhirCommunication;
  public
    Constructor Create(list : TFhirCommunicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCommunication read GetCurrent;
  end;


  {@Class TFhirCommunicationList
    A list of FhirCommunication
  }
  {!.Net HL7Connect.Fhir.CommunicationList}
  TFhirCommunicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCommunication;
    procedure SetItemN(index : Integer; value : TFhirCommunication);
  public
    {!script hide}
    function Link : TFhirCommunicationList; Overload;
    function Clone : TFhirCommunicationList; Overload;
    function GetEnumerator : TFhirCommunicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCommunication to the end of the list.
    }
    function Append : TFhirCommunication;

    
    {@member AddItem
      Add an already existing FhirCommunication to the end of the list.
    }
    procedure AddItem(value : TFhirCommunication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCommunication) : Integer;
    

    {@member Insert
      Insert FhirCommunication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCommunication;
    

    {@member InsertItem
       Insert an existing FhirCommunication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCommunication);
    
    {@member Item
       Get the iIndexth FhirCommunication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCommunication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCommunication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCommunication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCommunications[index : Integer] : TFhirCommunication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCommunicationRequestPayload : TFhirElement
    Text, attachment(s), or resource(s) to be communicated to the recipient.
  }
  {!.Net HL7Connect.Fhir.CommunicationRequestPayload}
  TFhirCommunicationRequestPayload = class (TFhirBackboneElement)
  private
    FContent : TFhirType;
    Procedure SetContent(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCommunicationRequestPayload; overload;
    function Clone : TFhirCommunicationRequestPayload; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member content
      Typed access to The communicated content (or for multi-part communications, one portion of the communication). (defined for API consistency)
    }
    property content : TFhirType read FContent write SetContent;
    {@member contentElement
      The communicated content (or for multi-part communications, one portion of the communication).
    }
    property contentElement : TFhirType read FContent write SetContent;

  end;


  TFhirCommunicationRequestPayloadListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCommunicationRequestPayloadList;
    function GetCurrent : TFhirCommunicationRequestPayload;
  public
    Constructor Create(list : TFhirCommunicationRequestPayloadList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCommunicationRequestPayload read GetCurrent;
  end;


  {@Class TFhirCommunicationRequestPayloadList
    A list of FhirCommunicationRequestPayload
  }
  {!.Net HL7Connect.Fhir.CommunicationRequestPayloadList}
  TFhirCommunicationRequestPayloadList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCommunicationRequestPayload;
    procedure SetItemN(index : Integer; value : TFhirCommunicationRequestPayload);
  public
    {!script hide}
    function Link : TFhirCommunicationRequestPayloadList; Overload;
    function Clone : TFhirCommunicationRequestPayloadList; Overload;
    function GetEnumerator : TFhirCommunicationRequestPayloadListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCommunicationRequestPayload to the end of the list.
    }
    function Append : TFhirCommunicationRequestPayload;

    
    {@member AddItem
      Add an already existing FhirCommunicationRequestPayload to the end of the list.
    }
    procedure AddItem(value : TFhirCommunicationRequestPayload); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCommunicationRequestPayload) : Integer;
    

    {@member Insert
      Insert FhirCommunicationRequestPayload before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCommunicationRequestPayload;
    

    {@member InsertItem
       Insert an existing FhirCommunicationRequestPayload before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCommunicationRequestPayload);
    
    {@member Item
       Get the iIndexth FhirCommunicationRequestPayload. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCommunicationRequestPayload. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCommunicationRequestPayload);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCommunicationRequestPayload;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCommunicationRequestPayloads[index : Integer] : TFhirCommunicationRequestPayload read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCommunicationRequest : TFhirDomainResource
    A request to convey information; e.g. the CDS system proposes that an alert be sent to a responsible provider, the CDS system proposes that the public health agency be notified about a reportable condition.
  }
  {!.Net HL7Connect.Fhir.CommunicationRequest}
  TFhirCommunicationRequest = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FCategory : TFhirCodeableConcept;
    FSender : TFhirReference{Resource};
    FrecipientList : TFhirReferenceList{Resource};
    FpayloadList : TFhirCommunicationRequestPayloadList;
    FmediumList : TFhirCodeableConceptList;
    FRequester : TFhirReference{Resource};
    FStatus : TFhirEnum;
    FEncounter : TFhirReference{TFhirEncounter};
    FScheduled : TFhirType;
    FreasonList : TFhirCodeableConceptList;
    FRequestedOn : TFhirDateTime;
    FSubject : TFhirReference{TFhirPatient};
    FPriority : TFhirCodeableConcept;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetSender(value : TFhirReference{Resource});
    function GetRecipientList : TFhirReferenceList{Resource};
    function GetHasRecipientList : Boolean;
    function GetPayloadList : TFhirCommunicationRequestPayloadList;
    function GetHasPayloadList : Boolean;
    function GetMediumList : TFhirCodeableConceptList;
    function GetHasMediumList : Boolean;
    Procedure SetRequester(value : TFhirReference{Resource});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirCommunicationRequestStatus;
    Procedure SetStatusST(value : TFhirCommunicationRequestStatus);
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetScheduled(value : TFhirType);
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    Procedure SetRequestedOn(value : TFhirDateTime);
    Function GetRequestedOnST : TDateAndTime;
    Procedure SetRequestedOnST(value : TDateAndTime);
    Procedure SetSubject(value : TFhirReference{TFhirPatient});
    Procedure SetPriority(value : TFhirCodeableConcept);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCommunicationRequest; overload;
    function Clone : TFhirCommunicationRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      A unique ID of this request for reference purposes. It must be provided if user wants it returned as part of any output, otherwise it will be autogenerated, if needed, by CDS system. Does not need to be the actual ID of the source system.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member category
      Typed access to The type of message to be sent such as alert, notification, reminder, instruction, etc. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      The type of message to be sent such as alert, notification, reminder, instruction, etc.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member sender
      Typed access to The entity (e.g. person, organization, clinical information system, or device) which is to be the source of the communication. (defined for API consistency)
    }
    property sender : TFhirReference{Resource} read FSender write SetSender;
    {@member senderElement
      The entity (e.g. person, organization, clinical information system, or device) which is to be the source of the communication.
    }
    property senderElement : TFhirReference{Resource} read FSender write SetSender;

    {@member recipientList
      The entity (e.g. person, organization, clinical information system, or device) which is the intended target of the communication.
    }
    property recipientList : TFhirReferenceList{Resource} read GetRecipientList;
    property hasRecipientList : boolean read GetHasRecipientList;

    {@member payloadList
      Text, attachment(s), or resource(s) to be communicated to the recipient.
    }
    property payloadList : TFhirCommunicationRequestPayloadList read GetPayloadList;
    property hasPayloadList : boolean read GetHasPayloadList;

    {@member mediumList
      A channel that was used for this communication (e.g. email, fax).
    }
    property mediumList : TFhirCodeableConceptList read GetMediumList;
    property hasMediumList : boolean read GetHasMediumList;

    {@member requester
      Typed access to The responsible person who authorizes this order, e.g. physician. This may be different than the author of the order statement, e.g. clerk, who may have entered the statement into the order entry application. (defined for API consistency)
    }
    property requester : TFhirReference{Resource} read FRequester write SetRequester;
    {@member requesterElement
      The responsible person who authorizes this order, e.g. physician. This may be different than the author of the order statement, e.g. clerk, who may have entered the statement into the order entry application.
    }
    property requesterElement : TFhirReference{Resource} read FRequester write SetRequester;

    {@member status
      The status of the proposal or order.
    }
    property status : TFhirCommunicationRequestStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member encounter
      Typed access to The encounter within which the communication request was created. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter within which the communication request was created.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member scheduled
      Typed access to The time when this communication is to occur. (defined for API consistency)
    }
    property scheduled : TFhirType read FScheduled write SetScheduled;
    {@member scheduledElement
      The time when this communication is to occur.
    }
    property scheduledElement : TFhirType read FScheduled write SetScheduled;

    {@member reasonList
      The reason or justification for the communication request.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member requestedOn
      Typed access to The time when the request was made.
    }
    property requestedOn : TDateAndTime read GetRequestedOnST write SetRequestedOnST;
    {@member requestedOnElement
      The time when the request was made.
    }
    property requestedOnElement : TFhirDateTime read FRequestedOn write SetRequestedOn;

    {@member subject
      Typed access to The patient who is the focus of this communication request. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirPatient} read FSubject write SetSubject;
    {@member subjectElement
      The patient who is the focus of this communication request.
    }
    property subjectElement : TFhirReference{TFhirPatient} read FSubject write SetSubject;

    {@member priority
      Typed access to Characterizes how quickly the proposed act must be initiated. Includes concepts such as stat, urgent, routine. (defined for API consistency)
    }
    property priority : TFhirCodeableConcept read FPriority write SetPriority;
    {@member priorityElement
      Characterizes how quickly the proposed act must be initiated. Includes concepts such as stat, urgent, routine.
    }
    property priorityElement : TFhirCodeableConcept read FPriority write SetPriority;

  end;


  TFhirCommunicationRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCommunicationRequestList;
    function GetCurrent : TFhirCommunicationRequest;
  public
    Constructor Create(list : TFhirCommunicationRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCommunicationRequest read GetCurrent;
  end;


  {@Class TFhirCommunicationRequestList
    A list of FhirCommunicationRequest
  }
  {!.Net HL7Connect.Fhir.CommunicationRequestList}
  TFhirCommunicationRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCommunicationRequest;
    procedure SetItemN(index : Integer; value : TFhirCommunicationRequest);
  public
    {!script hide}
    function Link : TFhirCommunicationRequestList; Overload;
    function Clone : TFhirCommunicationRequestList; Overload;
    function GetEnumerator : TFhirCommunicationRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCommunicationRequest to the end of the list.
    }
    function Append : TFhirCommunicationRequest;

    
    {@member AddItem
      Add an already existing FhirCommunicationRequest to the end of the list.
    }
    procedure AddItem(value : TFhirCommunicationRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCommunicationRequest) : Integer;
    

    {@member Insert
      Insert FhirCommunicationRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCommunicationRequest;
    

    {@member InsertItem
       Insert an existing FhirCommunicationRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCommunicationRequest);
    
    {@member Item
       Get the iIndexth FhirCommunicationRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCommunicationRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCommunicationRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCommunicationRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCommunicationRequests[index : Integer] : TFhirCommunicationRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCompositionAttester : TFhirElement
    A participant who has attested to the accuracy of the composition/document.
  }
  {!.Net HL7Connect.Fhir.CompositionAttester}
  TFhirCompositionAttester = class (TFhirBackboneElement)
  private
    FMode : TFhirEnumList;
    FTime : TFhirDateTime;
    FParty : TFhirReference{Resource};
    function GetMode : TFhirEnumList;
    function GetHasMode : Boolean;
    Function GetModeST : TFhirCompositionAttestationModeList;
    Procedure SetModeST(value : TFhirCompositionAttestationModeList);
    Procedure SetTime(value : TFhirDateTime);
    Function GetTimeST : TDateAndTime;
    Procedure SetTimeST(value : TDateAndTime);
    Procedure SetParty(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCompositionAttester; overload;
    function Clone : TFhirCompositionAttester; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member mode
      The type of attestation the authenticator offers.
    }
    property mode : TFhirCompositionAttestationModeList read GetModeST write SetModeST;
    property modeList : TFhirEnumList read GetMode;
    property hasMode : boolean read GetHasMode;
    {@member time
      Typed access to When composition was attested by the party.
    }
    property time : TDateAndTime read GetTimeST write SetTimeST;
    {@member timeElement
      When composition was attested by the party.
    }
    property timeElement : TFhirDateTime read FTime write SetTime;

    {@member party
      Typed access to Who attested the composition in the specified way. (defined for API consistency)
    }
    property party : TFhirReference{Resource} read FParty write SetParty;
    {@member partyElement
      Who attested the composition in the specified way.
    }
    property partyElement : TFhirReference{Resource} read FParty write SetParty;

  end;


  TFhirCompositionAttesterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCompositionAttesterList;
    function GetCurrent : TFhirCompositionAttester;
  public
    Constructor Create(list : TFhirCompositionAttesterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCompositionAttester read GetCurrent;
  end;


  {@Class TFhirCompositionAttesterList
    A list of FhirCompositionAttester
  }
  {!.Net HL7Connect.Fhir.CompositionAttesterList}
  TFhirCompositionAttesterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCompositionAttester;
    procedure SetItemN(index : Integer; value : TFhirCompositionAttester);
  public
    {!script hide}
    function Link : TFhirCompositionAttesterList; Overload;
    function Clone : TFhirCompositionAttesterList; Overload;
    function GetEnumerator : TFhirCompositionAttesterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCompositionAttester to the end of the list.
    }
    function Append : TFhirCompositionAttester;

    
    {@member AddItem
      Add an already existing FhirCompositionAttester to the end of the list.
    }
    procedure AddItem(value : TFhirCompositionAttester); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCompositionAttester) : Integer;
    

    {@member Insert
      Insert FhirCompositionAttester before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCompositionAttester;
    

    {@member InsertItem
       Insert an existing FhirCompositionAttester before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCompositionAttester);
    
    {@member Item
       Get the iIndexth FhirCompositionAttester. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCompositionAttester. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCompositionAttester);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCompositionAttester;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCompositionAttesters[index : Integer] : TFhirCompositionAttester read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCompositionEvent : TFhirElement
    The clinical service, such as a colonoscopy or an appendectomy, being documented.
  }
  {!.Net HL7Connect.Fhir.CompositionEvent}
  TFhirCompositionEvent = class (TFhirBackboneElement)
  private
    FcodeList : TFhirCodeableConceptList;
    FPeriod : TFhirPeriod;
    FdetailList : TFhirReferenceList{TFhirReference};
    function GetCodeList : TFhirCodeableConceptList;
    function GetHasCodeList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    function GetDetailList : TFhirReferenceList{TFhirReference};
    function GetHasDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCompositionEvent; overload;
    function Clone : TFhirCompositionEvent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member codeList
      This list of codes represents the main clinical acts, such as a colonoscopy or an appendectomy, being documented. In some cases, the event is inherent in the typeCode, such as a "History and Physical Report" in which the procedure being documented is necessarily a "History and Physical" act.
    }
    property codeList : TFhirCodeableConceptList read GetCodeList;
    property hasCodeList : boolean read GetHasCodeList;

    {@member period
      Typed access to The period of time covered by the documentation. There is no assertion that the documentation is a complete representation for this period, only that it documents events during this time. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time covered by the documentation. There is no assertion that the documentation is a complete representation for this period, only that it documents events during this time.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member detailList
      The description and/or reference of the event(s) being documented. For example, this could be used to document such a colonoscopy or an appendectomy.
    }
    property detailList : TFhirReferenceList{TFhirReference} read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirCompositionEventListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCompositionEventList;
    function GetCurrent : TFhirCompositionEvent;
  public
    Constructor Create(list : TFhirCompositionEventList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCompositionEvent read GetCurrent;
  end;


  {@Class TFhirCompositionEventList
    A list of FhirCompositionEvent
  }
  {!.Net HL7Connect.Fhir.CompositionEventList}
  TFhirCompositionEventList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCompositionEvent;
    procedure SetItemN(index : Integer; value : TFhirCompositionEvent);
  public
    {!script hide}
    function Link : TFhirCompositionEventList; Overload;
    function Clone : TFhirCompositionEventList; Overload;
    function GetEnumerator : TFhirCompositionEventListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCompositionEvent to the end of the list.
    }
    function Append : TFhirCompositionEvent;

    
    {@member AddItem
      Add an already existing FhirCompositionEvent to the end of the list.
    }
    procedure AddItem(value : TFhirCompositionEvent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCompositionEvent) : Integer;
    

    {@member Insert
      Insert FhirCompositionEvent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCompositionEvent;
    

    {@member InsertItem
       Insert an existing FhirCompositionEvent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCompositionEvent);
    
    {@member Item
       Get the iIndexth FhirCompositionEvent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCompositionEvent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCompositionEvent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCompositionEvent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCompositionEvents[index : Integer] : TFhirCompositionEvent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCompositionSection : TFhirElement
    The root of the sections that make up the composition.
  }
  {!.Net HL7Connect.Fhir.CompositionSection}
  TFhirCompositionSection = class (TFhirBackboneElement)
  private
    FTitle : TFhirString;
    FCode : TFhirCodeableConcept;
    FText : TFhirNarrative;
    FMode : TFhirEnum;
    FOrderedBy : TFhirCodeableConcept;
    FentryList : TFhirReferenceList{TFhirReference};
    FEmptyReason : TFhirCodeableConcept;
    FsectionList : TFhirCompositionSectionList;
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetText(value : TFhirNarrative);
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirListMode;
    Procedure SetModeST(value : TFhirListMode);
    Procedure SetOrderedBy(value : TFhirCodeableConcept);
    function GetEntryList : TFhirReferenceList{TFhirReference};
    function GetHasEntryList : Boolean;
    Procedure SetEmptyReason(value : TFhirCodeableConcept);
    function GetSectionList : TFhirCompositionSectionList;
    function GetHasSectionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCompositionSection; overload;
    function Clone : TFhirCompositionSection; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member title
      Typed access to The label for this particular section.  This will be part of the rendered content for the document, and is often used to build a table of contents.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      The label for this particular section.  This will be part of the rendered content for the document, and is often used to build a table of contents.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member code
      Typed access to A code identifying the kind of content contained within the section. This must be consistent with the section title. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code identifying the kind of content contained within the section. This must be consistent with the section title.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member text
      Typed access to A human-readable narrative that contains the attested content of the section, used to represent the content of the resource to a human. The narrative need not encode all the structured data, but is required to contain sufficient detail to make it "clinically safe" for a human to just read the narrative. (defined for API consistency)
    }
    property text : TFhirNarrative read FText write SetText;
    {@member textElement
      A human-readable narrative that contains the attested content of the section, used to represent the content of the resource to a human. The narrative need not encode all the structured data, but is required to contain sufficient detail to make it "clinically safe" for a human to just read the narrative.
    }
    property textElement : TFhirNarrative read FText write SetText;

    {@member mode
      How the entry list was prepared - whether it is a working list that is suitable for being maintained on an ongoing basis, or if it represents a snapshot of a list of items from another source, or whether it is a prepared list where items may be marked as added, modified or deleted.
    }
    property mode : TFhirListMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member orderedBy
      Typed access to Specifies the order applied to the items in the section entries. (defined for API consistency)
    }
    property orderedBy : TFhirCodeableConcept read FOrderedBy write SetOrderedBy;
    {@member orderedByElement
      Specifies the order applied to the items in the section entries.
    }
    property orderedByElement : TFhirCodeableConcept read FOrderedBy write SetOrderedBy;

    {@member entryList
      A reference to the actual resource from which the narrative in the section is derived.
    }
    property entryList : TFhirReferenceList{TFhirReference} read GetEntryList;
    property hasEntryList : boolean read GetHasEntryList;

    {@member emptyReason
      Typed access to If the section is empty, why the list is empty. An empty section typically has some text explaining the empty reason. (defined for API consistency)
    }
    property emptyReason : TFhirCodeableConcept read FEmptyReason write SetEmptyReason;
    {@member emptyReasonElement
      If the section is empty, why the list is empty. An empty section typically has some text explaining the empty reason.
    }
    property emptyReasonElement : TFhirCodeableConcept read FEmptyReason write SetEmptyReason;

    {@member sectionList
      A nested sub-section within this section.
    }
    property sectionList : TFhirCompositionSectionList read GetSectionList;
    property hasSectionList : boolean read GetHasSectionList;

  end;


  TFhirCompositionSectionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCompositionSectionList;
    function GetCurrent : TFhirCompositionSection;
  public
    Constructor Create(list : TFhirCompositionSectionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCompositionSection read GetCurrent;
  end;


  {@Class TFhirCompositionSectionList
    A list of FhirCompositionSection
  }
  {!.Net HL7Connect.Fhir.CompositionSectionList}
  TFhirCompositionSectionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCompositionSection;
    procedure SetItemN(index : Integer; value : TFhirCompositionSection);
  public
    {!script hide}
    function Link : TFhirCompositionSectionList; Overload;
    function Clone : TFhirCompositionSectionList; Overload;
    function GetEnumerator : TFhirCompositionSectionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCompositionSection to the end of the list.
    }
    function Append : TFhirCompositionSection;

    
    {@member AddItem
      Add an already existing FhirCompositionSection to the end of the list.
    }
    procedure AddItem(value : TFhirCompositionSection); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCompositionSection) : Integer;
    

    {@member Insert
      Insert FhirCompositionSection before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCompositionSection;
    

    {@member InsertItem
       Insert an existing FhirCompositionSection before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCompositionSection);
    
    {@member Item
       Get the iIndexth FhirCompositionSection. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCompositionSection. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCompositionSection);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCompositionSection;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCompositionSections[index : Integer] : TFhirCompositionSection read GetItemN write SetItemN; default;
  End;


  {@Class TFhirComposition : TFhirDomainResource
    A set of healthcare-related information that is assembled together into a single logical document that provides a single coherent statement of meaning, establishes its own context and that has clinical attestation with regard to who is making the statement. While a Composition defines the structure, it does not actually contain the content: rather the full content of a document is contained in a Bundle, of which the Composition is the first resource contained.
  }
  {!.Net HL7Connect.Fhir.Composition}
  TFhirComposition = class (TFhirDomainResource)
  private
    FIdentifier : TFhirIdentifier;
    FDate : TFhirDateTime;
    FType_ : TFhirCodeableConcept;
    FClass_ : TFhirCodeableConcept;
    FTitle : TFhirString;
    FStatus : TFhirEnum;
    FConfidentiality : TFhirEnum;
    FSubject : TFhirReference{TFhirReference};
    FauthorList : TFhirReferenceList{Resource};
    FattesterList : TFhirCompositionAttesterList;
    FCustodian : TFhirReference{TFhirOrganization};
    FeventList : TFhirCompositionEventList;
    FEncounter : TFhirReference{TFhirEncounter};
    FsectionList : TFhirCompositionSectionList;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetClass_(value : TFhirCodeableConcept);
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirCompositionStatus;
    Procedure SetStatusST(value : TFhirCompositionStatus);
    Procedure SetConfidentiality(value : TFhirEnum);
    Function GetConfidentialityST : TFhirV3Confidentiality;
    Procedure SetConfidentialityST(value : TFhirV3Confidentiality);
    Procedure SetSubject(value : TFhirReference{TFhirReference});
    function GetAuthorList : TFhirReferenceList{Resource};
    function GetHasAuthorList : Boolean;
    function GetAttesterList : TFhirCompositionAttesterList;
    function GetHasAttesterList : Boolean;
    Procedure SetCustodian(value : TFhirReference{TFhirOrganization});
    function GetEventList : TFhirCompositionEventList;
    function GetHasEventList : Boolean;
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    function GetSectionList : TFhirCompositionSectionList;
    function GetHasSectionList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirComposition; overload;
    function Clone : TFhirComposition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifier
      Typed access to Logical identifier for the composition, assigned when created. This identifier stays constant as the composition is changed over time. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Logical identifier for the composition, assigned when created. This identifier stays constant as the composition is changed over time.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member date
      Typed access to The composition editing time, when the composition was last logically changed by the author.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The composition editing time, when the composition was last logically changed by the author.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member type_
      Typed access to Specifies the particular kind of composition (e.g. History and Physical, Discharge Summary, Progress Note). This usually equates to the purpose of making the composition. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Specifies the particular kind of composition (e.g. History and Physical, Discharge Summary, Progress Note). This usually equates to the purpose of making the composition.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member class_
      Typed access to A categorization for the type of the composition - helps for indexing and searching. This may be implied by or derived from the code specified in the Composition Type. (defined for API consistency)
    }
    property class_ : TFhirCodeableConcept read FClass_ write SetClass_;
    {@member class_Element
      A categorization for the type of the composition - helps for indexing and searching. This may be implied by or derived from the code specified in the Composition Type.
    }
    property class_Element : TFhirCodeableConcept read FClass_ write SetClass_;

    {@member title
      Typed access to Official human-readable label for the composition.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      Official human-readable label for the composition.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member status
      The workflow/clinical status of this composition. The status is a marker for the clinical standing of the document.
    }
    property status : TFhirCompositionStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member confidentiality
      The code specifying the level of confidentiality of the Composition.
    }
    property confidentiality : TFhirV3Confidentiality read GetConfidentialityST write SetConfidentialityST;
    property confidentialityElement : TFhirEnum read FConfidentiality write SetConfidentiality;

    {@member subject
      Typed access to Who or what the composition is about. The composition can be about a person, (patient or healthcare practitioner), a device (e.g. a machine) or even a group of subjects (such as a document about a herd of livestock, or a set of patients that share a common exposure). (defined for API consistency)
    }
    property subject : TFhirReference{TFhirReference} read FSubject write SetSubject;
    {@member subjectElement
      Who or what the composition is about. The composition can be about a person, (patient or healthcare practitioner), a device (e.g. a machine) or even a group of subjects (such as a document about a herd of livestock, or a set of patients that share a common exposure).
    }
    property subjectElement : TFhirReference{TFhirReference} read FSubject write SetSubject;

    {@member authorList
      Identifies who is responsible for the information in the composition, not necessarily who typed it in.
    }
    property authorList : TFhirReferenceList{Resource} read GetAuthorList;
    property hasAuthorList : boolean read GetHasAuthorList;

    {@member attesterList
      A participant who has attested to the accuracy of the composition/document.
    }
    property attesterList : TFhirCompositionAttesterList read GetAttesterList;
    property hasAttesterList : boolean read GetHasAttesterList;

    {@member custodian
      Typed access to Identifies the organization or group who is responsible for ongoing maintenance of and access to the composition/document information. (defined for API consistency)
    }
    property custodian : TFhirReference{TFhirOrganization} read FCustodian write SetCustodian;
    {@member custodianElement
      Identifies the organization or group who is responsible for ongoing maintenance of and access to the composition/document information.
    }
    property custodianElement : TFhirReference{TFhirOrganization} read FCustodian write SetCustodian;

    {@member eventList
      The clinical service, such as a colonoscopy or an appendectomy, being documented.
    }
    property eventList : TFhirCompositionEventList read GetEventList;
    property hasEventList : boolean read GetHasEventList;

    {@member encounter
      Typed access to Describes the clinical encounter or type of care this documentation is associated with. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      Describes the clinical encounter or type of care this documentation is associated with.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member sectionList
      The root of the sections that make up the composition.
    }
    property sectionList : TFhirCompositionSectionList read GetSectionList;
    property hasSectionList : boolean read GetHasSectionList;

  end;


  TFhirCompositionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCompositionList;
    function GetCurrent : TFhirComposition;
  public
    Constructor Create(list : TFhirCompositionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirComposition read GetCurrent;
  end;


  {@Class TFhirCompositionList
    A list of FhirComposition
  }
  {!.Net HL7Connect.Fhir.CompositionList}
  TFhirCompositionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirComposition;
    procedure SetItemN(index : Integer; value : TFhirComposition);
  public
    {!script hide}
    function Link : TFhirCompositionList; Overload;
    function Clone : TFhirCompositionList; Overload;
    function GetEnumerator : TFhirCompositionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirComposition to the end of the list.
    }
    function Append : TFhirComposition;

    
    {@member AddItem
      Add an already existing FhirComposition to the end of the list.
    }
    procedure AddItem(value : TFhirComposition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirComposition) : Integer;
    

    {@member Insert
      Insert FhirComposition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirComposition;
    

    {@member InsertItem
       Insert an existing FhirComposition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirComposition);
    
    {@member Item
       Get the iIndexth FhirComposition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirComposition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirComposition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirComposition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCompositions[index : Integer] : TFhirComposition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConceptMapContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.ConceptMapContact}
  TFhirConceptMapContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConceptMapContact; overload;
    function Clone : TFhirConceptMapContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the concept map.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the concept map.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirConceptMapContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConceptMapContactList;
    function GetCurrent : TFhirConceptMapContact;
  public
    Constructor Create(list : TFhirConceptMapContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConceptMapContact read GetCurrent;
  end;


  {@Class TFhirConceptMapContactList
    A list of FhirConceptMapContact
  }
  {!.Net HL7Connect.Fhir.ConceptMapContactList}
  TFhirConceptMapContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConceptMapContact;
    procedure SetItemN(index : Integer; value : TFhirConceptMapContact);
  public
    {!script hide}
    function Link : TFhirConceptMapContactList; Overload;
    function Clone : TFhirConceptMapContactList; Overload;
    function GetEnumerator : TFhirConceptMapContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConceptMapContact to the end of the list.
    }
    function Append : TFhirConceptMapContact;

    
    {@member AddItem
      Add an already existing FhirConceptMapContact to the end of the list.
    }
    procedure AddItem(value : TFhirConceptMapContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConceptMapContact) : Integer;
    

    {@member Insert
      Insert FhirConceptMapContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConceptMapContact;
    

    {@member InsertItem
       Insert an existing FhirConceptMapContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConceptMapContact);
    
    {@member Item
       Get the iIndexth FhirConceptMapContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConceptMapContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConceptMapContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConceptMapContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConceptMapContacts[index : Integer] : TFhirConceptMapContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConceptMapElement : TFhirElement
    Mappings for an individual concept in the source to one or more concepts in the target.
  }
  {!.Net HL7Connect.Fhir.ConceptMapElement}
  TFhirConceptMapElement = class (TFhirBackboneElement)
  private
    FCodeSystem : TFhirUri;
    FCode : TFhirCode;
    FtargetList : TFhirConceptMapElementTargetList;
    Procedure SetCodeSystem(value : TFhirUri);
    Function GetCodeSystemST : String;
    Procedure SetCodeSystemST(value : String);
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    function GetTargetList : TFhirConceptMapElementTargetList;
    function GetHasTargetList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConceptMapElement; overload;
    function Clone : TFhirConceptMapElement; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member codeSystem
      Typed access to An absolute URI that identifies the Code System (if the source is a value set that crosses more than one code system).
    }
    property codeSystem : String read GetCodeSystemST write SetCodeSystemST;
    {@member codeSystemElement
      An absolute URI that identifies the Code System (if the source is a value set that crosses more than one code system).
    }
    property codeSystemElement : TFhirUri read FCodeSystem write SetCodeSystem;

    {@member code
      Typed access to Identity (code or path) or the element/item being mapped.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      Identity (code or path) or the element/item being mapped.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member targetList
      A concept from the target value set that this concept maps to.
    }
    property targetList : TFhirConceptMapElementTargetList read GetTargetList;
    property hasTargetList : boolean read GetHasTargetList;

  end;


  TFhirConceptMapElementListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConceptMapElementList;
    function GetCurrent : TFhirConceptMapElement;
  public
    Constructor Create(list : TFhirConceptMapElementList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConceptMapElement read GetCurrent;
  end;


  {@Class TFhirConceptMapElementList
    A list of FhirConceptMapElement
  }
  {!.Net HL7Connect.Fhir.ConceptMapElementList}
  TFhirConceptMapElementList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConceptMapElement;
    procedure SetItemN(index : Integer; value : TFhirConceptMapElement);
  public
    {!script hide}
    function Link : TFhirConceptMapElementList; Overload;
    function Clone : TFhirConceptMapElementList; Overload;
    function GetEnumerator : TFhirConceptMapElementListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConceptMapElement to the end of the list.
    }
    function Append : TFhirConceptMapElement;

    
    {@member AddItem
      Add an already existing FhirConceptMapElement to the end of the list.
    }
    procedure AddItem(value : TFhirConceptMapElement); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConceptMapElement) : Integer;
    

    {@member Insert
      Insert FhirConceptMapElement before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConceptMapElement;
    

    {@member InsertItem
       Insert an existing FhirConceptMapElement before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConceptMapElement);
    
    {@member Item
       Get the iIndexth FhirConceptMapElement. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConceptMapElement. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConceptMapElement);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConceptMapElement;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConceptMapElements[index : Integer] : TFhirConceptMapElement read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConceptMapElementTarget : TFhirElement
    A concept from the target value set that this concept maps to.
  }
  {!.Net HL7Connect.Fhir.ConceptMapElementTarget}
  TFhirConceptMapElementTarget = class (TFhirBackboneElement)
  private
    FCodeSystem : TFhirUri;
    FCode : TFhirCode;
    FEquivalence : TFhirEnum;
    FComments : TFhirString;
    FdependsOnList : TFhirConceptMapElementTargetDependsOnList;
    FproductList : TFhirConceptMapElementTargetDependsOnList;
    Procedure SetCodeSystem(value : TFhirUri);
    Function GetCodeSystemST : String;
    Procedure SetCodeSystemST(value : String);
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetEquivalence(value : TFhirEnum);
    Function GetEquivalenceST : TFhirConceptMapEquivalence;
    Procedure SetEquivalenceST(value : TFhirConceptMapEquivalence);
    Procedure SetComments(value : TFhirString);
    Function GetCommentsST : String;
    Procedure SetCommentsST(value : String);
    function GetDependsOnList : TFhirConceptMapElementTargetDependsOnList;
    function GetHasDependsOnList : Boolean;
    function GetProductList : TFhirConceptMapElementTargetDependsOnList;
    function GetHasProductList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConceptMapElementTarget; overload;
    function Clone : TFhirConceptMapElementTarget; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member codeSystem
      Typed access to An absolute URI that identifies the code system of the target code (if the target is a value set that cross code systems).
    }
    property codeSystem : String read GetCodeSystemST write SetCodeSystemST;
    {@member codeSystemElement
      An absolute URI that identifies the code system of the target code (if the target is a value set that cross code systems).
    }
    property codeSystemElement : TFhirUri read FCodeSystem write SetCodeSystem;

    {@member code
      Typed access to Identity (code or path) or the element/item that the map refers to.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      Identity (code or path) or the element/item that the map refers to.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member equivalence
      The equivalence between the source and target concepts (counting for the dependencies and products). The equivalence is read from target to source (e.g. the target is 'wider' than the source).
    }
    property equivalence : TFhirConceptMapEquivalence read GetEquivalenceST write SetEquivalenceST;
    property equivalenceElement : TFhirEnum read FEquivalence write SetEquivalence;

    {@member comments
      Typed access to A description of status/issues in mapping that conveys additional information not represented in  the structured data.
    }
    property comments : String read GetCommentsST write SetCommentsST;
    {@member commentsElement
      A description of status/issues in mapping that conveys additional information not represented in  the structured data.
    }
    property commentsElement : TFhirString read FComments write SetComments;

    {@member dependsOnList
      A set of additional dependencies for this mapping to hold. This mapping is only applicable if the specified element can be resolved, and it has the specified value.
    }
    property dependsOnList : TFhirConceptMapElementTargetDependsOnList read GetDependsOnList;
    property hasDependsOnList : boolean read GetHasDependsOnList;

    {@member productList
      A set of additional outcomes from this mapping to other elements. To properly execute this mapping, the specified element must be mapped to some data element or source that is in context. The mapping may still be useful without a place for the additional data elements, but the equivalence cannot be relied on.
    }
    property productList : TFhirConceptMapElementTargetDependsOnList read GetProductList;
    property hasProductList : boolean read GetHasProductList;

  end;


  TFhirConceptMapElementTargetListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConceptMapElementTargetList;
    function GetCurrent : TFhirConceptMapElementTarget;
  public
    Constructor Create(list : TFhirConceptMapElementTargetList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConceptMapElementTarget read GetCurrent;
  end;


  {@Class TFhirConceptMapElementTargetList
    A list of FhirConceptMapElementTarget
  }
  {!.Net HL7Connect.Fhir.ConceptMapElementTargetList}
  TFhirConceptMapElementTargetList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConceptMapElementTarget;
    procedure SetItemN(index : Integer; value : TFhirConceptMapElementTarget);
  public
    {!script hide}
    function Link : TFhirConceptMapElementTargetList; Overload;
    function Clone : TFhirConceptMapElementTargetList; Overload;
    function GetEnumerator : TFhirConceptMapElementTargetListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConceptMapElementTarget to the end of the list.
    }
    function Append : TFhirConceptMapElementTarget;

    
    {@member AddItem
      Add an already existing FhirConceptMapElementTarget to the end of the list.
    }
    procedure AddItem(value : TFhirConceptMapElementTarget); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConceptMapElementTarget) : Integer;
    

    {@member Insert
      Insert FhirConceptMapElementTarget before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConceptMapElementTarget;
    

    {@member InsertItem
       Insert an existing FhirConceptMapElementTarget before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConceptMapElementTarget);
    
    {@member Item
       Get the iIndexth FhirConceptMapElementTarget. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConceptMapElementTarget. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConceptMapElementTarget);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConceptMapElementTarget;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConceptMapElementTargets[index : Integer] : TFhirConceptMapElementTarget read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConceptMapElementTargetDependsOn : TFhirElement
    A set of additional dependencies for this mapping to hold. This mapping is only applicable if the specified element can be resolved, and it has the specified value.
  }
  {!.Net HL7Connect.Fhir.ConceptMapElementTargetDependsOn}
  TFhirConceptMapElementTargetDependsOn = class (TFhirBackboneElement)
  private
    FElement : TFhirUri;
    FCodeSystem : TFhirUri;
    FCode : TFhirString;
    Procedure SetElement(value : TFhirUri);
    Function GetElementST : String;
    Procedure SetElementST(value : String);
    Procedure SetCodeSystem(value : TFhirUri);
    Function GetCodeSystemST : String;
    Procedure SetCodeSystemST(value : String);
    Procedure SetCode(value : TFhirString);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConceptMapElementTargetDependsOn; overload;
    function Clone : TFhirConceptMapElementTargetDependsOn; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member element
      Typed access to A reference to a specific concept that holds a coded value. This can be an element in a FHIR resource, or a specific reference to a data element in a different specification (e.g. HL7 v2) or a general reference to a kind of data field, or a reference to a value set with an appropriately narrow definition.
    }
    property element : String read GetElementST write SetElementST;
    {@member elementElement
      A reference to a specific concept that holds a coded value. This can be an element in a FHIR resource, or a specific reference to a data element in a different specification (e.g. HL7 v2) or a general reference to a kind of data field, or a reference to a value set with an appropriately narrow definition.
    }
    property elementElement : TFhirUri read FElement write SetElement;

    {@member codeSystem
      Typed access to An absolute URI that identifies the code system of the dependency code (if the source/dependency is a value set that crosses code systems).
    }
    property codeSystem : String read GetCodeSystemST write SetCodeSystemST;
    {@member codeSystemElement
      An absolute URI that identifies the code system of the dependency code (if the source/dependency is a value set that crosses code systems).
    }
    property codeSystemElement : TFhirUri read FCodeSystem write SetCodeSystem;

    {@member code
      Typed access to Identity (code or path) or the element/item/ValueSet that the map depends on / refers to.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      Identity (code or path) or the element/item/ValueSet that the map depends on / refers to.
    }
    property codeElement : TFhirString read FCode write SetCode;

  end;


  TFhirConceptMapElementTargetDependsOnListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConceptMapElementTargetDependsOnList;
    function GetCurrent : TFhirConceptMapElementTargetDependsOn;
  public
    Constructor Create(list : TFhirConceptMapElementTargetDependsOnList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConceptMapElementTargetDependsOn read GetCurrent;
  end;


  {@Class TFhirConceptMapElementTargetDependsOnList
    A list of FhirConceptMapElementTargetDependsOn
  }
  {!.Net HL7Connect.Fhir.ConceptMapElementTargetDependsOnList}
  TFhirConceptMapElementTargetDependsOnList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConceptMapElementTargetDependsOn;
    procedure SetItemN(index : Integer; value : TFhirConceptMapElementTargetDependsOn);
  public
    {!script hide}
    function Link : TFhirConceptMapElementTargetDependsOnList; Overload;
    function Clone : TFhirConceptMapElementTargetDependsOnList; Overload;
    function GetEnumerator : TFhirConceptMapElementTargetDependsOnListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConceptMapElementTargetDependsOn to the end of the list.
    }
    function Append : TFhirConceptMapElementTargetDependsOn;

    
    {@member AddItem
      Add an already existing FhirConceptMapElementTargetDependsOn to the end of the list.
    }
    procedure AddItem(value : TFhirConceptMapElementTargetDependsOn); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConceptMapElementTargetDependsOn) : Integer;
    

    {@member Insert
      Insert FhirConceptMapElementTargetDependsOn before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConceptMapElementTargetDependsOn;
    

    {@member InsertItem
       Insert an existing FhirConceptMapElementTargetDependsOn before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConceptMapElementTargetDependsOn);
    
    {@member Item
       Get the iIndexth FhirConceptMapElementTargetDependsOn. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConceptMapElementTargetDependsOn. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConceptMapElementTargetDependsOn);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConceptMapElementTargetDependsOn;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConceptMapElementTargetDependsOns[index : Integer] : TFhirConceptMapElementTargetDependsOn read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConceptMap : TFhirDomainResource
    A statement of relationships from one set of concepts to one or more other concepts - either code systems or data elements, or classes in class models.
  }
  {!.Net HL7Connect.Fhir.ConceptMap}
  TFhirConceptMap = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FIdentifier : TFhirIdentifier;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirConceptMapContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FRequirements : TFhirString;
    FCopyright : TFhirString;
    FSource : TFhirType;
    FTarget : TFhirType;
    FelementList : TFhirConceptMapElementList;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirConceptMapContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetSource(value : TFhirType);
    Procedure SetTarget(value : TFhirType);
    function GetElementList : TFhirConceptMapElementList;
    function GetHasElementList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConceptMap; overload;
    function Clone : TFhirConceptMap; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this concept map when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this concept map is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this concept map when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this concept map is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member identifier
      Typed access to Formal identifier that is used to identify this concept map when it is represented in other formats, or referenced in a specification, model, design or an instance. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Formal identifier that is used to identify this concept map when it is represented in other formats, or referenced in a specification, model, design or an instance.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member version
      Typed access to The identifier that is used to identify this version of the concept map when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the concept map when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name describing the concept map.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name describing the concept map.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the concept map.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to This ConceptMap was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This ConceptMap was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the concept map.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the concept map.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirConceptMapContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the concept map was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the concept map changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the concept map was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the concept map changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the use of the concept map - reason for definition, conditions of use, etc.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the use of the concept map - reason for definition, conditions of use, etc.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of concept map instances.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member requirements
      Typed access to Explains why this concept map is needed and why it has been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this concept map is needed and why it has been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member copyright
      Typed access to A copyright statement relating to the concept map and/or its contents.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the concept map and/or its contents.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member source
      Typed access to The source value set that specifies the concepts that are being mapped. (defined for API consistency)
    }
    property source : TFhirType read FSource write SetSource;
    {@member sourceElement
      The source value set that specifies the concepts that are being mapped.
    }
    property sourceElement : TFhirType read FSource write SetSource;

    {@member target
      Typed access to The target value set provides context to the mappings. Note that the mapping is made between concepts, not between value sets, but the value set provides important context about how the concept mapping choices are made. (defined for API consistency)
    }
    property target : TFhirType read FTarget write SetTarget;
    {@member targetElement
      The target value set provides context to the mappings. Note that the mapping is made between concepts, not between value sets, but the value set provides important context about how the concept mapping choices are made.
    }
    property targetElement : TFhirType read FTarget write SetTarget;

    {@member elementList
      Mappings for an individual concept in the source to one or more concepts in the target.
    }
    property elementList : TFhirConceptMapElementList read GetElementList;
    property hasElementList : boolean read GetHasElementList;

  end;


  TFhirConceptMapListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConceptMapList;
    function GetCurrent : TFhirConceptMap;
  public
    Constructor Create(list : TFhirConceptMapList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConceptMap read GetCurrent;
  end;


  {@Class TFhirConceptMapList
    A list of FhirConceptMap
  }
  {!.Net HL7Connect.Fhir.ConceptMapList}
  TFhirConceptMapList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConceptMap;
    procedure SetItemN(index : Integer; value : TFhirConceptMap);
  public
    {!script hide}
    function Link : TFhirConceptMapList; Overload;
    function Clone : TFhirConceptMapList; Overload;
    function GetEnumerator : TFhirConceptMapListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConceptMap to the end of the list.
    }
    function Append : TFhirConceptMap;

    
    {@member AddItem
      Add an already existing FhirConceptMap to the end of the list.
    }
    procedure AddItem(value : TFhirConceptMap); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConceptMap) : Integer;
    

    {@member Insert
      Insert FhirConceptMap before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConceptMap;
    

    {@member InsertItem
       Insert an existing FhirConceptMap before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConceptMap);
    
    {@member Item
       Get the iIndexth FhirConceptMap. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConceptMap. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConceptMap);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConceptMap;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConceptMaps[index : Integer] : TFhirConceptMap read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConditionStage : TFhirElement
    Clinical stage or grade of a condition. May include formal severity assessments.
  }
  {!.Net HL7Connect.Fhir.ConditionStage}
  TFhirConditionStage = class (TFhirBackboneElement)
  private
    FSummary : TFhirCodeableConcept;
    FassessmentList : TFhirReferenceList{Resource};
    Procedure SetSummary(value : TFhirCodeableConcept);
    function GetAssessmentList : TFhirReferenceList{Resource};
    function GetHasAssessmentList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConditionStage; overload;
    function Clone : TFhirConditionStage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member summary
      Typed access to A simple summary of the stage such as "Stage 3". The determination of the stage is disease-specific. (defined for API consistency)
    }
    property summary : TFhirCodeableConcept read FSummary write SetSummary;
    {@member summaryElement
      A simple summary of the stage such as "Stage 3". The determination of the stage is disease-specific.
    }
    property summaryElement : TFhirCodeableConcept read FSummary write SetSummary;

    {@member assessmentList
      Reference to a formal record of the evidence on which the staging assessment is based.
    }
    property assessmentList : TFhirReferenceList{Resource} read GetAssessmentList;
    property hasAssessmentList : boolean read GetHasAssessmentList;

  end;


  TFhirConditionStageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConditionStageList;
    function GetCurrent : TFhirConditionStage;
  public
    Constructor Create(list : TFhirConditionStageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConditionStage read GetCurrent;
  end;


  {@Class TFhirConditionStageList
    A list of FhirConditionStage
  }
  {!.Net HL7Connect.Fhir.ConditionStageList}
  TFhirConditionStageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConditionStage;
    procedure SetItemN(index : Integer; value : TFhirConditionStage);
  public
    {!script hide}
    function Link : TFhirConditionStageList; Overload;
    function Clone : TFhirConditionStageList; Overload;
    function GetEnumerator : TFhirConditionStageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConditionStage to the end of the list.
    }
    function Append : TFhirConditionStage;

    
    {@member AddItem
      Add an already existing FhirConditionStage to the end of the list.
    }
    procedure AddItem(value : TFhirConditionStage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConditionStage) : Integer;
    

    {@member Insert
      Insert FhirConditionStage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConditionStage;
    

    {@member InsertItem
       Insert an existing FhirConditionStage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConditionStage);
    
    {@member Item
       Get the iIndexth FhirConditionStage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConditionStage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConditionStage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConditionStage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConditionStages[index : Integer] : TFhirConditionStage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConditionEvidence : TFhirElement
    Supporting Evidence / manifestations that are the basis on which this condition is suspected or confirmed.
  }
  {!.Net HL7Connect.Fhir.ConditionEvidence}
  TFhirConditionEvidence = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FdetailList : TFhirReferenceList{TFhirReference};
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetDetailList : TFhirReferenceList{TFhirReference};
    function GetHasDetailList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConditionEvidence; overload;
    function Clone : TFhirConditionEvidence; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A manifestation or symptom that led to the recording of this condition. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A manifestation or symptom that led to the recording of this condition.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member detailList
      Links to other relevant information, including pathology reports.
    }
    property detailList : TFhirReferenceList{TFhirReference} read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirConditionEvidenceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConditionEvidenceList;
    function GetCurrent : TFhirConditionEvidence;
  public
    Constructor Create(list : TFhirConditionEvidenceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConditionEvidence read GetCurrent;
  end;


  {@Class TFhirConditionEvidenceList
    A list of FhirConditionEvidence
  }
  {!.Net HL7Connect.Fhir.ConditionEvidenceList}
  TFhirConditionEvidenceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConditionEvidence;
    procedure SetItemN(index : Integer; value : TFhirConditionEvidence);
  public
    {!script hide}
    function Link : TFhirConditionEvidenceList; Overload;
    function Clone : TFhirConditionEvidenceList; Overload;
    function GetEnumerator : TFhirConditionEvidenceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConditionEvidence to the end of the list.
    }
    function Append : TFhirConditionEvidence;

    
    {@member AddItem
      Add an already existing FhirConditionEvidence to the end of the list.
    }
    procedure AddItem(value : TFhirConditionEvidence); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConditionEvidence) : Integer;
    

    {@member Insert
      Insert FhirConditionEvidence before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConditionEvidence;
    

    {@member InsertItem
       Insert an existing FhirConditionEvidence before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConditionEvidence);
    
    {@member Item
       Get the iIndexth FhirConditionEvidence. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConditionEvidence. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConditionEvidence);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConditionEvidence;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConditionEvidences[index : Integer] : TFhirConditionEvidence read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCondition : TFhirDomainResource
    Use to record detailed information about conditions, problems or diagnoses recognized by a clinician. There are many uses including: recording a diagnosis during an encounter; populating a problem list or a summary statement, such as a discharge summary.
  }
  {!.Net HL7Connect.Fhir.Condition}
  TFhirCondition = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FPatient : TFhirReference{TFhirPatient};
    FEncounter : TFhirReference{TFhirEncounter};
    FAsserter : TFhirReference{Resource};
    FDateRecorded : TFhirDate;
    FCode : TFhirCodeableConcept;
    FCategory : TFhirCodeableConcept;
    FClinicalStatus : TFhirEnum;
    FVerificationStatus : TFhirEnum;
    FSeverity : TFhirCodeableConcept;
    FOnset : TFhirType;
    FAbatement : TFhirType;
    FStage : TFhirConditionStage;
    FevidenceList : TFhirConditionEvidenceList;
    FbodySiteList : TFhirCodeableConceptList;
    FNotes : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetAsserter(value : TFhirReference{Resource});
    Procedure SetDateRecorded(value : TFhirDate);
    Function GetDateRecordedST : TDateAndTime;
    Procedure SetDateRecordedST(value : TDateAndTime);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetClinicalStatus(value : TFhirEnum);
    Function GetClinicalStatusST : TFhirValuesetConditionClinical;
    Procedure SetClinicalStatusST(value : TFhirValuesetConditionClinical);
    Procedure SetVerificationStatus(value : TFhirEnum);
    Function GetVerificationStatusST : TFhirConditionVerStatus;
    Procedure SetVerificationStatusST(value : TFhirConditionVerStatus);
    Procedure SetSeverity(value : TFhirCodeableConcept);
    Procedure SetOnset(value : TFhirType);
    Procedure SetAbatement(value : TFhirType);
    Procedure SetStage(value : TFhirConditionStage);
    function GetEvidenceList : TFhirConditionEvidenceList;
    function GetHasEvidenceList : Boolean;
    function GetBodySiteList : TFhirCodeableConceptList;
    function GetHasBodySiteList : Boolean;
    Procedure SetNotes(value : TFhirString);
    Function GetNotesST : String;
    Procedure SetNotesST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCondition; overload;
    function Clone : TFhirCondition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this condition that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member patient
      Typed access to Indicates the patient who the condition record is associated with. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      Indicates the patient who the condition record is associated with.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member encounter
      Typed access to Encounter during which the condition was first asserted. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      Encounter during which the condition was first asserted.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member asserter
      Typed access to Individual who is making the condition statement. (defined for API consistency)
    }
    property asserter : TFhirReference{Resource} read FAsserter write SetAsserter;
    {@member asserterElement
      Individual who is making the condition statement.
    }
    property asserterElement : TFhirReference{Resource} read FAsserter write SetAsserter;

    {@member dateRecorded
      Typed access to A date, when  the Condition statement was documented.
    }
    property dateRecorded : TDateAndTime read GetDateRecordedST write SetDateRecordedST;
    {@member dateRecordedElement
      A date, when  the Condition statement was documented.
    }
    property dateRecordedElement : TFhirDate read FDateRecorded write SetDateRecorded;

    {@member code
      Typed access to Identification of the condition, problem or diagnosis. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Identification of the condition, problem or diagnosis.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member category
      Typed access to A category assigned to the condition. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      A category assigned to the condition.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member clinicalStatus
      The clinical status of the condition.
    }
    property clinicalStatus : TFhirValuesetConditionClinical read GetClinicalStatusST write SetClinicalStatusST;
    property clinicalStatusElement : TFhirEnum read FClinicalStatus write SetClinicalStatus;

    {@member verificationStatus
      The verification status to support the clinical status of the condition.
    }
    property verificationStatus : TFhirConditionVerStatus read GetVerificationStatusST write SetVerificationStatusST;
    property verificationStatusElement : TFhirEnum read FVerificationStatus write SetVerificationStatus;

    {@member severity
      Typed access to A subjective assessment of the severity of the condition as evaluated by the clinician. (defined for API consistency)
    }
    property severity : TFhirCodeableConcept read FSeverity write SetSeverity;
    {@member severityElement
      A subjective assessment of the severity of the condition as evaluated by the clinician.
    }
    property severityElement : TFhirCodeableConcept read FSeverity write SetSeverity;

    {@member onset
      Typed access to Estimated or actual date or date-time  the condition began, in the opinion of the clinician. (defined for API consistency)
    }
    property onset : TFhirType read FOnset write SetOnset;
    {@member onsetElement
      Estimated or actual date or date-time  the condition began, in the opinion of the clinician.
    }
    property onsetElement : TFhirType read FOnset write SetOnset;

    {@member abatement
      Typed access to The date or estimated date that the condition resolved or went into remission. This is called "abatement" because of the many overloaded connotations associated with "remission" or "resolution" - Conditions are never really resolved, but they can abate. (defined for API consistency)
    }
    property abatement : TFhirType read FAbatement write SetAbatement;
    {@member abatementElement
      The date or estimated date that the condition resolved or went into remission. This is called "abatement" because of the many overloaded connotations associated with "remission" or "resolution" - Conditions are never really resolved, but they can abate.
    }
    property abatementElement : TFhirType read FAbatement write SetAbatement;

    {@member stage
      Typed access to Clinical stage or grade of a condition. May include formal severity assessments. (defined for API consistency)
    }
    property stage : TFhirConditionStage read FStage write SetStage;
    {@member stageElement
      Clinical stage or grade of a condition. May include formal severity assessments.
    }
    property stageElement : TFhirConditionStage read FStage write SetStage;

    {@member evidenceList
      Supporting Evidence / manifestations that are the basis on which this condition is suspected or confirmed.
    }
    property evidenceList : TFhirConditionEvidenceList read GetEvidenceList;
    property hasEvidenceList : boolean read GetHasEvidenceList;

    {@member bodySiteList
      The anatomical location where this condition manifests itself.
    }
    property bodySiteList : TFhirCodeableConceptList read GetBodySiteList;
    property hasBodySiteList : boolean read GetHasBodySiteList;

    {@member notes
      Typed access to Additional information about the Condition. This is a general notes/comments entry  for description of the Condition, its diagnosis and prognosis.
    }
    property notes : String read GetNotesST write SetNotesST;
    {@member notesElement
      Additional information about the Condition. This is a general notes/comments entry  for description of the Condition, its diagnosis and prognosis.
    }
    property notesElement : TFhirString read FNotes write SetNotes;

  end;


  TFhirConditionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConditionList;
    function GetCurrent : TFhirCondition;
  public
    Constructor Create(list : TFhirConditionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCondition read GetCurrent;
  end;


  {@Class TFhirConditionList
    A list of FhirCondition
  }
  {!.Net HL7Connect.Fhir.ConditionList}
  TFhirConditionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCondition;
    procedure SetItemN(index : Integer; value : TFhirCondition);
  public
    {!script hide}
    function Link : TFhirConditionList; Overload;
    function Clone : TFhirConditionList; Overload;
    function GetEnumerator : TFhirConditionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCondition to the end of the list.
    }
    function Append : TFhirCondition;

    
    {@member AddItem
      Add an already existing FhirCondition to the end of the list.
    }
    procedure AddItem(value : TFhirCondition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCondition) : Integer;
    

    {@member Insert
      Insert FhirCondition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCondition;
    

    {@member InsertItem
       Insert an existing FhirCondition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCondition);
    
    {@member Item
       Get the iIndexth FhirCondition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCondition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCondition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCondition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConditions[index : Integer] : TFhirCondition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.ConformanceContact}
  TFhirConformanceContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceContact; overload;
    function Clone : TFhirConformanceContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the conformance.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the conformance.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirConformanceContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceContactList;
    function GetCurrent : TFhirConformanceContact;
  public
    Constructor Create(list : TFhirConformanceContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceContact read GetCurrent;
  end;


  {@Class TFhirConformanceContactList
    A list of FhirConformanceContact
  }
  {!.Net HL7Connect.Fhir.ConformanceContactList}
  TFhirConformanceContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceContact;
    procedure SetItemN(index : Integer; value : TFhirConformanceContact);
  public
    {!script hide}
    function Link : TFhirConformanceContactList; Overload;
    function Clone : TFhirConformanceContactList; Overload;
    function GetEnumerator : TFhirConformanceContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceContact to the end of the list.
    }
    function Append : TFhirConformanceContact;

    
    {@member AddItem
      Add an already existing FhirConformanceContact to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceContact) : Integer;
    

    {@member Insert
      Insert FhirConformanceContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceContact;
    

    {@member InsertItem
       Insert an existing FhirConformanceContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceContact);
    
    {@member Item
       Get the iIndexth FhirConformanceContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceContacts[index : Integer] : TFhirConformanceContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceSoftware : TFhirElement
    Software that is covered by this conformance statement.  It is used when the conformance statement describes the capabilities of a particular software version, independent of an installation.
  }
  {!.Net HL7Connect.Fhir.ConformanceSoftware}
  TFhirConformanceSoftware = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FVersion : TFhirString;
    FReleaseDate : TFhirDateTime;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetReleaseDate(value : TFhirDateTime);
    Function GetReleaseDateST : TDateAndTime;
    Procedure SetReleaseDateST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceSoftware; overload;
    function Clone : TFhirConformanceSoftware; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to Name software is known by.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Name software is known by.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member version
      Typed access to The version identifier for the software covered by this statement.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version identifier for the software covered by this statement.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member releaseDate
      Typed access to Date this version of the software released.
    }
    property releaseDate : TDateAndTime read GetReleaseDateST write SetReleaseDateST;
    {@member releaseDateElement
      Date this version of the software released.
    }
    property releaseDateElement : TFhirDateTime read FReleaseDate write SetReleaseDate;

  end;


  TFhirConformanceSoftwareListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceSoftwareList;
    function GetCurrent : TFhirConformanceSoftware;
  public
    Constructor Create(list : TFhirConformanceSoftwareList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceSoftware read GetCurrent;
  end;


  {@Class TFhirConformanceSoftwareList
    A list of FhirConformanceSoftware
  }
  {!.Net HL7Connect.Fhir.ConformanceSoftwareList}
  TFhirConformanceSoftwareList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceSoftware;
    procedure SetItemN(index : Integer; value : TFhirConformanceSoftware);
  public
    {!script hide}
    function Link : TFhirConformanceSoftwareList; Overload;
    function Clone : TFhirConformanceSoftwareList; Overload;
    function GetEnumerator : TFhirConformanceSoftwareListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceSoftware to the end of the list.
    }
    function Append : TFhirConformanceSoftware;

    
    {@member AddItem
      Add an already existing FhirConformanceSoftware to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceSoftware); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceSoftware) : Integer;
    

    {@member Insert
      Insert FhirConformanceSoftware before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceSoftware;
    

    {@member InsertItem
       Insert an existing FhirConformanceSoftware before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceSoftware);
    
    {@member Item
       Get the iIndexth FhirConformanceSoftware. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceSoftware. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceSoftware);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceSoftware;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceSoftwares[index : Integer] : TFhirConformanceSoftware read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceImplementation : TFhirElement
    Identifies a specific implementation instance that is described by the conformance statement - i.e. a particular installation, rather than the capabilities of a software program.
  }
  {!.Net HL7Connect.Fhir.ConformanceImplementation}
  TFhirConformanceImplementation = class (TFhirBackboneElement)
  private
    FDescription : TFhirString;
    FUrl : TFhirUri;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceImplementation; overload;
    function Clone : TFhirConformanceImplementation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member description
      Typed access to Information about the specific installation that this conformance statement relates to.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Information about the specific installation that this conformance statement relates to.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member url
      Typed access to An absolute base URL for the implementation.  This forms the base for REST interfaces as well as the mailbox and document interfaces.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute base URL for the implementation.  This forms the base for REST interfaces as well as the mailbox and document interfaces.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

  end;


  TFhirConformanceImplementationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceImplementationList;
    function GetCurrent : TFhirConformanceImplementation;
  public
    Constructor Create(list : TFhirConformanceImplementationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceImplementation read GetCurrent;
  end;


  {@Class TFhirConformanceImplementationList
    A list of FhirConformanceImplementation
  }
  {!.Net HL7Connect.Fhir.ConformanceImplementationList}
  TFhirConformanceImplementationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceImplementation;
    procedure SetItemN(index : Integer; value : TFhirConformanceImplementation);
  public
    {!script hide}
    function Link : TFhirConformanceImplementationList; Overload;
    function Clone : TFhirConformanceImplementationList; Overload;
    function GetEnumerator : TFhirConformanceImplementationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceImplementation to the end of the list.
    }
    function Append : TFhirConformanceImplementation;

    
    {@member AddItem
      Add an already existing FhirConformanceImplementation to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceImplementation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceImplementation) : Integer;
    

    {@member Insert
      Insert FhirConformanceImplementation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceImplementation;
    

    {@member InsertItem
       Insert an existing FhirConformanceImplementation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceImplementation);
    
    {@member Item
       Get the iIndexth FhirConformanceImplementation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceImplementation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceImplementation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceImplementation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceImplementations[index : Integer] : TFhirConformanceImplementation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRest : TFhirElement
    A definition of the restful capabilities of the solution, if any.
  }
  {!.Net HL7Connect.Fhir.ConformanceRest}
  TFhirConformanceRest = class (TFhirBackboneElement)
  private
    FMode : TFhirEnum;
    FDocumentation : TFhirString;
    FSecurity : TFhirConformanceRestSecurity;
    FresourceList : TFhirConformanceRestResourceList;
    FinteractionList : TFhirConformanceRestInteractionList;
    FTransactionMode : TFhirEnum;
    FsearchParamList : TFhirConformanceRestResourceSearchParamList;
    FoperationList : TFhirConformanceRestOperationList;
    FcompartmentList : TFhirUriList;
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirRestfulConformanceMode;
    Procedure SetModeST(value : TFhirRestfulConformanceMode);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
    Procedure SetSecurity(value : TFhirConformanceRestSecurity);
    function GetResourceList : TFhirConformanceRestResourceList;
    function GetHasResourceList : Boolean;
    function GetInteractionList : TFhirConformanceRestInteractionList;
    function GetHasInteractionList : Boolean;
    Procedure SetTransactionMode(value : TFhirEnum);
    Function GetTransactionModeST : TFhirTransactionMode;
    Procedure SetTransactionModeST(value : TFhirTransactionMode);
    function GetSearchParamList : TFhirConformanceRestResourceSearchParamList;
    function GetHasSearchParamList : Boolean;
    function GetOperationList : TFhirConformanceRestOperationList;
    function GetHasOperationList : Boolean;
    function GetCompartmentList : TFhirUriList;
    function GetHasCompartmentList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRest; overload;
    function Clone : TFhirConformanceRest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member mode
      Identifies whether this portion of the statement is describing ability to initiate or receive restful operations.
    }
    property mode : TFhirRestfulConformanceMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member documentation
      Typed access to Information about the system's restful capabilities that apply across all applications, such as security.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Information about the system's restful capabilities that apply across all applications, such as security.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

    {@member security
      Typed access to Information about security implementation from an interface perspective - what a client needs to know. (defined for API consistency)
    }
    property security : TFhirConformanceRestSecurity read FSecurity write SetSecurity;
    {@member securityElement
      Information about security implementation from an interface perspective - what a client needs to know.
    }
    property securityElement : TFhirConformanceRestSecurity read FSecurity write SetSecurity;

    {@member resourceList
      A specification of the restful capabilities of the solution for a specific resource type.
    }
    property resourceList : TFhirConformanceRestResourceList read GetResourceList;
    property hasResourceList : boolean read GetHasResourceList;

    {@member interactionList
      A specification of restful operations supported by the system.
    }
    property interactionList : TFhirConformanceRestInteractionList read GetInteractionList;
    property hasInteractionList : boolean read GetHasInteractionList;

    {@member transactionMode
      A code that indicates how transactions are supported.
    }
    property transactionMode : TFhirTransactionMode read GetTransactionModeST write SetTransactionModeST;
    property transactionModeElement : TFhirEnum read FTransactionMode write SetTransactionMode;

    {@member searchParamList
      Search parameters that are supported for searching all resources for implementations to support and/or make use of - either references to ones defined in the specification, or additional ones defined for/by the implementation.
    }
    property searchParamList : TFhirConformanceRestResourceSearchParamList read GetSearchParamList;
    property hasSearchParamList : boolean read GetHasSearchParamList;

    {@member operationList
      Definition of an operation or a named query and with its parameters and their meaning and type.
    }
    property operationList : TFhirConformanceRestOperationList read GetOperationList;
    property hasOperationList : boolean read GetHasOperationList;

    {@member compartmentList
      An absolute URI which is a reference to the definition of a compartment hosted by the system.
    }
    property compartmentList : TFhirUriList read GetCompartmentList;
    property hasCompartmentList : boolean read GetHasCompartmentList;

  end;


  TFhirConformanceRestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestList;
    function GetCurrent : TFhirConformanceRest;
  public
    Constructor Create(list : TFhirConformanceRestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRest read GetCurrent;
  end;


  {@Class TFhirConformanceRestList
    A list of FhirConformanceRest
  }
  {!.Net HL7Connect.Fhir.ConformanceRestList}
  TFhirConformanceRestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRest;
    procedure SetItemN(index : Integer; value : TFhirConformanceRest);
  public
    {!script hide}
    function Link : TFhirConformanceRestList; Overload;
    function Clone : TFhirConformanceRestList; Overload;
    function GetEnumerator : TFhirConformanceRestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRest to the end of the list.
    }
    function Append : TFhirConformanceRest;

    
    {@member AddItem
      Add an already existing FhirConformanceRest to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRest) : Integer;
    

    {@member Insert
      Insert FhirConformanceRest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRest;
    

    {@member InsertItem
       Insert an existing FhirConformanceRest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRest);
    
    {@member Item
       Get the iIndexth FhirConformanceRest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRests[index : Integer] : TFhirConformanceRest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestSecurity : TFhirElement
    Information about security implementation from an interface perspective - what a client needs to know.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestSecurity}
  TFhirConformanceRestSecurity = class (TFhirBackboneElement)
  private
    FCors : TFhirBoolean;
    FserviceList : TFhirCodeableConceptList;
    FDescription : TFhirString;
    FcertificateList : TFhirConformanceRestSecurityCertificateList;
    Procedure SetCors(value : TFhirBoolean);
    Function GetCorsST : Boolean;
    Procedure SetCorsST(value : Boolean);
    function GetServiceList : TFhirCodeableConceptList;
    function GetHasServiceList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetCertificateList : TFhirConformanceRestSecurityCertificateList;
    function GetHasCertificateList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestSecurity; overload;
    function Clone : TFhirConformanceRestSecurity; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member cors
      Typed access to Server adds CORS headers when responding to requests - this enables javascript applications to use the server.
    }
    property cors : Boolean read GetCorsST write SetCorsST;
    {@member corsElement
      Server adds CORS headers when responding to requests - this enables javascript applications to use the server.
    }
    property corsElement : TFhirBoolean read FCors write SetCors;

    {@member serviceList
      Types of security services are supported/required by the system.
    }
    property serviceList : TFhirCodeableConceptList read GetServiceList;
    property hasServiceList : boolean read GetHasServiceList;

    {@member description
      Typed access to General description of how security works.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      General description of how security works.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member certificateList
      Certificates associated with security profiles.
    }
    property certificateList : TFhirConformanceRestSecurityCertificateList read GetCertificateList;
    property hasCertificateList : boolean read GetHasCertificateList;

  end;


  TFhirConformanceRestSecurityListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestSecurityList;
    function GetCurrent : TFhirConformanceRestSecurity;
  public
    Constructor Create(list : TFhirConformanceRestSecurityList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestSecurity read GetCurrent;
  end;


  {@Class TFhirConformanceRestSecurityList
    A list of FhirConformanceRestSecurity
  }
  {!.Net HL7Connect.Fhir.ConformanceRestSecurityList}
  TFhirConformanceRestSecurityList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestSecurity;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestSecurity);
  public
    {!script hide}
    function Link : TFhirConformanceRestSecurityList; Overload;
    function Clone : TFhirConformanceRestSecurityList; Overload;
    function GetEnumerator : TFhirConformanceRestSecurityListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestSecurity to the end of the list.
    }
    function Append : TFhirConformanceRestSecurity;

    
    {@member AddItem
      Add an already existing FhirConformanceRestSecurity to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestSecurity); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestSecurity) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestSecurity before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestSecurity;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestSecurity before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestSecurity);
    
    {@member Item
       Get the iIndexth FhirConformanceRestSecurity. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestSecurity. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestSecurity);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestSecurity;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestSecurities[index : Integer] : TFhirConformanceRestSecurity read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestSecurityCertificate : TFhirElement
    Certificates associated with security profiles.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestSecurityCertificate}
  TFhirConformanceRestSecurityCertificate = class (TFhirBackboneElement)
  private
    FType_ : TFhirCode;
    FBlob : TFhirBase64Binary;
    Procedure SetType_(value : TFhirCode);
    Function GetType_ST : String;
    Procedure SetType_ST(value : String);
    Procedure SetBlob(value : TFhirBase64Binary);
    Function GetBlobST : TBytes;
    Procedure SetBlobST(value : TBytes);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestSecurityCertificate; overload;
    function Clone : TFhirConformanceRestSecurityCertificate; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Mime type for certificate.
    }
    property type_ : String read GetType_ST write SetType_ST;
    {@member type_Element
      Mime type for certificate.
    }
    property type_Element : TFhirCode read FType_ write SetType_;

    {@member blob
      Typed access to Actual certificate.
    }
    property blob : TBytes read GetBlobST write SetBlobST;
    {@member blobElement
      Actual certificate.
    }
    property blobElement : TFhirBase64Binary read FBlob write SetBlob;

  end;


  TFhirConformanceRestSecurityCertificateListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestSecurityCertificateList;
    function GetCurrent : TFhirConformanceRestSecurityCertificate;
  public
    Constructor Create(list : TFhirConformanceRestSecurityCertificateList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestSecurityCertificate read GetCurrent;
  end;


  {@Class TFhirConformanceRestSecurityCertificateList
    A list of FhirConformanceRestSecurityCertificate
  }
  {!.Net HL7Connect.Fhir.ConformanceRestSecurityCertificateList}
  TFhirConformanceRestSecurityCertificateList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestSecurityCertificate;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestSecurityCertificate);
  public
    {!script hide}
    function Link : TFhirConformanceRestSecurityCertificateList; Overload;
    function Clone : TFhirConformanceRestSecurityCertificateList; Overload;
    function GetEnumerator : TFhirConformanceRestSecurityCertificateListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestSecurityCertificate to the end of the list.
    }
    function Append : TFhirConformanceRestSecurityCertificate;

    
    {@member AddItem
      Add an already existing FhirConformanceRestSecurityCertificate to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestSecurityCertificate); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestSecurityCertificate) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestSecurityCertificate before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestSecurityCertificate;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestSecurityCertificate before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestSecurityCertificate);
    
    {@member Item
       Get the iIndexth FhirConformanceRestSecurityCertificate. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestSecurityCertificate. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestSecurityCertificate);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestSecurityCertificate;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestSecurityCertificates[index : Integer] : TFhirConformanceRestSecurityCertificate read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestResource : TFhirElement
    A specification of the restful capabilities of the solution for a specific resource type.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResource}
  TFhirConformanceRestResource = class (TFhirBackboneElement)
  private
    FType_ : TFhirCode;
    FProfile : TFhirReference{TFhirStructureDefinition};
    FinteractionList : TFhirConformanceRestResourceInteractionList;
    FVersioning : TFhirEnum;
    FReadHistory : TFhirBoolean;
    FUpdateCreate : TFhirBoolean;
    FConditionalCreate : TFhirBoolean;
    FConditionalUpdate : TFhirBoolean;
    FConditionalDelete : TFhirEnum;
    FsearchIncludeList : TFhirStringList;
    FsearchRevIncludeList : TFhirStringList;
    FsearchParamList : TFhirConformanceRestResourceSearchParamList;
    Procedure SetType_(value : TFhirCode);
    Function GetType_ST : String;
    Procedure SetType_ST(value : String);
    Procedure SetProfile(value : TFhirReference{TFhirStructureDefinition});
    function GetInteractionList : TFhirConformanceRestResourceInteractionList;
    function GetHasInteractionList : Boolean;
    Procedure SetVersioning(value : TFhirEnum);
    Function GetVersioningST : TFhirVersioningPolicy;
    Procedure SetVersioningST(value : TFhirVersioningPolicy);
    Procedure SetReadHistory(value : TFhirBoolean);
    Function GetReadHistoryST : Boolean;
    Procedure SetReadHistoryST(value : Boolean);
    Procedure SetUpdateCreate(value : TFhirBoolean);
    Function GetUpdateCreateST : Boolean;
    Procedure SetUpdateCreateST(value : Boolean);
    Procedure SetConditionalCreate(value : TFhirBoolean);
    Function GetConditionalCreateST : Boolean;
    Procedure SetConditionalCreateST(value : Boolean);
    Procedure SetConditionalUpdate(value : TFhirBoolean);
    Function GetConditionalUpdateST : Boolean;
    Procedure SetConditionalUpdateST(value : Boolean);
    Procedure SetConditionalDelete(value : TFhirEnum);
    Function GetConditionalDeleteST : TFhirConditionalDeleteStatus;
    Procedure SetConditionalDeleteST(value : TFhirConditionalDeleteStatus);
    function GetSearchIncludeList : TFhirStringList;
    function GetHasSearchIncludeList : Boolean;
    function GetSearchRevIncludeList : TFhirStringList;
    function GetHasSearchRevIncludeList : Boolean;
    function GetSearchParamList : TFhirConformanceRestResourceSearchParamList;
    function GetHasSearchParamList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestResource; overload;
    function Clone : TFhirConformanceRestResource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to A type of resource exposed via the restful interface.
    }
    property type_ : String read GetType_ST write SetType_ST;
    {@member type_Element
      A type of resource exposed via the restful interface.
    }
    property type_Element : TFhirCode read FType_ write SetType_;

    {@member profile
      Typed access to A specification of the profile that describes the solution's overall support for the resource, including any constraints on cardinality, bindings, lengths or other limitations. See further discussion in [Using Profiles]. (defined for API consistency)
    }
    property profile : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;
    {@member profileElement
      A specification of the profile that describes the solution's overall support for the resource, including any constraints on cardinality, bindings, lengths or other limitations. See further discussion in [Using Profiles].
    }
    property profileElement : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;

    {@member interactionList
      Identifies a restful operation supported by the solution.
    }
    property interactionList : TFhirConformanceRestResourceInteractionList read GetInteractionList;
    property hasInteractionList : boolean read GetHasInteractionList;

    {@member versioning
      This field is set to no-version to specify that the system does not support (server) or use (client) versioning for this resource type. If this has some other value, the server must at least correctly track and populate the versionId meta-property on resources. If the value is 'versioned-update', then the server supports all the versioning features, including using e-tags for version integrity in the API.
    }
    property versioning : TFhirVersioningPolicy read GetVersioningST write SetVersioningST;
    property versioningElement : TFhirEnum read FVersioning write SetVersioning;

    {@member readHistory
      Typed access to A flag for whether the server is able to return past versions as part of the vRead operation.
    }
    property readHistory : Boolean read GetReadHistoryST write SetReadHistoryST;
    {@member readHistoryElement
      A flag for whether the server is able to return past versions as part of the vRead operation.
    }
    property readHistoryElement : TFhirBoolean read FReadHistory write SetReadHistory;

    {@member updateCreate
      Typed access to A flag to indicate that the server allows or needs to allow the client to create new identities on the server (e.g. that is, the client PUTs to a location where there is no existing resource). Allowing this operation means that the server allows the client to create new identities on the server.
    }
    property updateCreate : Boolean read GetUpdateCreateST write SetUpdateCreateST;
    {@member updateCreateElement
      A flag to indicate that the server allows or needs to allow the client to create new identities on the server (e.g. that is, the client PUTs to a location where there is no existing resource). Allowing this operation means that the server allows the client to create new identities on the server.
    }
    property updateCreateElement : TFhirBoolean read FUpdateCreate write SetUpdateCreate;

    {@member conditionalCreate
      Typed access to A flag that indicates that the server supports conditional create.
    }
    property conditionalCreate : Boolean read GetConditionalCreateST write SetConditionalCreateST;
    {@member conditionalCreateElement
      A flag that indicates that the server supports conditional create.
    }
    property conditionalCreateElement : TFhirBoolean read FConditionalCreate write SetConditionalCreate;

    {@member conditionalUpdate
      Typed access to A flag that indicates that the server supports conditional update.
    }
    property conditionalUpdate : Boolean read GetConditionalUpdateST write SetConditionalUpdateST;
    {@member conditionalUpdateElement
      A flag that indicates that the server supports conditional update.
    }
    property conditionalUpdateElement : TFhirBoolean read FConditionalUpdate write SetConditionalUpdate;

    {@member conditionalDelete
      A code that indicates how the server supports conditional delete.
    }
    property conditionalDelete : TFhirConditionalDeleteStatus read GetConditionalDeleteST write SetConditionalDeleteST;
    property conditionalDeleteElement : TFhirEnum read FConditionalDelete write SetConditionalDelete;

    {@member searchIncludeList
      A list of _include values supported by the server.
    }
    property searchIncludeList : TFhirStringList read GetSearchIncludeList;
    property hasSearchIncludeList : boolean read GetHasSearchIncludeList;

    {@member searchRevIncludeList
      A list of _revinclude (reverse include) values supported by the server.
    }
    property searchRevIncludeList : TFhirStringList read GetSearchRevIncludeList;
    property hasSearchRevIncludeList : boolean read GetHasSearchRevIncludeList;

    {@member searchParamList
      Search parameters for implementations to support and/or make use of - either references to ones defined in the specification, or additional ones defined for/by the implementation.
    }
    property searchParamList : TFhirConformanceRestResourceSearchParamList read GetSearchParamList;
    property hasSearchParamList : boolean read GetHasSearchParamList;

  end;


  TFhirConformanceRestResourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestResourceList;
    function GetCurrent : TFhirConformanceRestResource;
  public
    Constructor Create(list : TFhirConformanceRestResourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestResource read GetCurrent;
  end;


  {@Class TFhirConformanceRestResourceList
    A list of FhirConformanceRestResource
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResourceList}
  TFhirConformanceRestResourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestResource;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestResource);
  public
    {!script hide}
    function Link : TFhirConformanceRestResourceList; Overload;
    function Clone : TFhirConformanceRestResourceList; Overload;
    function GetEnumerator : TFhirConformanceRestResourceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestResource to the end of the list.
    }
    function Append : TFhirConformanceRestResource;

    
    {@member AddItem
      Add an already existing FhirConformanceRestResource to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestResource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestResource) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestResource before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestResource;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestResource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestResource);
    
    {@member Item
       Get the iIndexth FhirConformanceRestResource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestResource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestResource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestResource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestResources[index : Integer] : TFhirConformanceRestResource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestResourceInteraction : TFhirElement
    Identifies a restful operation supported by the solution.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResourceInteraction}
  TFhirConformanceRestResourceInteraction = class (TFhirBackboneElement)
  private
    FCode : TFhirEnum;
    FDocumentation : TFhirString;
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirTypeRestfulInteraction;
    Procedure SetCodeST(value : TFhirTypeRestfulInteraction);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestResourceInteraction; overload;
    function Clone : TFhirConformanceRestResourceInteraction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Coded identifier of the operation, supported by the system resource.
    }
    property code : TFhirTypeRestfulInteraction read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member documentation
      Typed access to Guidance specific to the implementation of this operation, such as 'delete is a logical delete' or 'updates are only allowed with version id' or 'creates permitted from pre-authorized certificates only'.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Guidance specific to the implementation of this operation, such as 'delete is a logical delete' or 'updates are only allowed with version id' or 'creates permitted from pre-authorized certificates only'.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

  end;


  TFhirConformanceRestResourceInteractionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestResourceInteractionList;
    function GetCurrent : TFhirConformanceRestResourceInteraction;
  public
    Constructor Create(list : TFhirConformanceRestResourceInteractionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestResourceInteraction read GetCurrent;
  end;


  {@Class TFhirConformanceRestResourceInteractionList
    A list of FhirConformanceRestResourceInteraction
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResourceInteractionList}
  TFhirConformanceRestResourceInteractionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestResourceInteraction;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestResourceInteraction);
  public
    {!script hide}
    function Link : TFhirConformanceRestResourceInteractionList; Overload;
    function Clone : TFhirConformanceRestResourceInteractionList; Overload;
    function GetEnumerator : TFhirConformanceRestResourceInteractionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestResourceInteraction to the end of the list.
    }
    function Append : TFhirConformanceRestResourceInteraction;

    
    {@member AddItem
      Add an already existing FhirConformanceRestResourceInteraction to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestResourceInteraction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestResourceInteraction) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestResourceInteraction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestResourceInteraction;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestResourceInteraction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestResourceInteraction);
    
    {@member Item
       Get the iIndexth FhirConformanceRestResourceInteraction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestResourceInteraction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestResourceInteraction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestResourceInteraction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestResourceInteractions[index : Integer] : TFhirConformanceRestResourceInteraction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestResourceSearchParam : TFhirElement
    Search parameters for implementations to support and/or make use of - either references to ones defined in the specification, or additional ones defined for/by the implementation.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResourceSearchParam}
  TFhirConformanceRestResourceSearchParam = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FDefinition : TFhirUri;
    FType_ : TFhirEnum;
    FDocumentation : TFhirString;
    FtargetList : TFhirCodeList;
    FModifier : TFhirEnumList;
    FchainList : TFhirStringList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDefinition(value : TFhirUri);
    Function GetDefinitionST : String;
    Procedure SetDefinitionST(value : String);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirSearchParamType;
    Procedure SetType_ST(value : TFhirSearchParamType);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
    function GetTargetList : TFhirCodeList;
    function GetHasTargetList : Boolean;
    function GetModifier : TFhirEnumList;
    function GetHasModifier : Boolean;
    Function GetModifierST : TFhirSearchModifierCodeList;
    Procedure SetModifierST(value : TFhirSearchModifierCodeList);
    function GetChainList : TFhirStringList;
    function GetHasChainList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestResourceSearchParam; overload;
    function Clone : TFhirConformanceRestResourceSearchParam; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of the search parameter used in the interface.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of the search parameter used in the interface.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member definition
      Typed access to An absolute URI that is a formal reference to where this parameter was first defined, so that a client can be confident of the meaning of the search parameter (a reference to [[[SearchParameter.url]]]).
    }
    property definition : String read GetDefinitionST write SetDefinitionST;
    {@member definitionElement
      An absolute URI that is a formal reference to where this parameter was first defined, so that a client can be confident of the meaning of the search parameter (a reference to [[[SearchParameter.url]]]).
    }
    property definitionElement : TFhirUri read FDefinition write SetDefinition;

    {@member type_
      The type of value a search parameter refers to, and how the content is interpreted.
    }
    property type_ : TFhirSearchParamType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member documentation
      Typed access to This allows documentation of any distinct behaviors about how the search parameter is used.  For example, text matching algorithms.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      This allows documentation of any distinct behaviors about how the search parameter is used.  For example, text matching algorithms.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

    {@member targetList
      Types of resource (if a resource is referenced).
    }
    property targetList : TFhirCodeList read GetTargetList;
    property hasTargetList : boolean read GetHasTargetList;

    {@member modifier
      A modifier supported for the search parameter.
    }
    property modifier : TFhirSearchModifierCodeList read GetModifierST write SetModifierST;
    property modifierList : TFhirEnumList read GetModifier;
    property hasModifier : boolean read GetHasModifier;
    {@member chainList
      Contains the names of any search parameters which may be chained to the containing search parameter. Chained parameters may be added to search parameters of type reference, and specify that resources will only be returned if they contain a reference to a resource which matches the chained parameter value. Values for this field should be drawn from Conformance.rest.resource.searchParam.name on the target resource type.
    }
    property chainList : TFhirStringList read GetChainList;
    property hasChainList : boolean read GetHasChainList;

  end;


  TFhirConformanceRestResourceSearchParamListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestResourceSearchParamList;
    function GetCurrent : TFhirConformanceRestResourceSearchParam;
  public
    Constructor Create(list : TFhirConformanceRestResourceSearchParamList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestResourceSearchParam read GetCurrent;
  end;


  {@Class TFhirConformanceRestResourceSearchParamList
    A list of FhirConformanceRestResourceSearchParam
  }
  {!.Net HL7Connect.Fhir.ConformanceRestResourceSearchParamList}
  TFhirConformanceRestResourceSearchParamList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestResourceSearchParam;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestResourceSearchParam);
  public
    {!script hide}
    function Link : TFhirConformanceRestResourceSearchParamList; Overload;
    function Clone : TFhirConformanceRestResourceSearchParamList; Overload;
    function GetEnumerator : TFhirConformanceRestResourceSearchParamListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestResourceSearchParam to the end of the list.
    }
    function Append : TFhirConformanceRestResourceSearchParam;

    
    {@member AddItem
      Add an already existing FhirConformanceRestResourceSearchParam to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestResourceSearchParam); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestResourceSearchParam) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestResourceSearchParam before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestResourceSearchParam;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestResourceSearchParam before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestResourceSearchParam);
    
    {@member Item
       Get the iIndexth FhirConformanceRestResourceSearchParam. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestResourceSearchParam. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestResourceSearchParam);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestResourceSearchParam;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestResourceSearchParams[index : Integer] : TFhirConformanceRestResourceSearchParam read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestInteraction : TFhirElement
    A specification of restful operations supported by the system.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestInteraction}
  TFhirConformanceRestInteraction = class (TFhirBackboneElement)
  private
    FCode : TFhirEnum;
    FDocumentation : TFhirString;
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirSystemRestfulInteraction;
    Procedure SetCodeST(value : TFhirSystemRestfulInteraction);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestInteraction; overload;
    function Clone : TFhirConformanceRestInteraction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      A coded identifier of the operation, supported by the system.
    }
    property code : TFhirSystemRestfulInteraction read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member documentation
      Typed access to Guidance specific to the implementation of this operation, such as limitations on the kind of transactions allowed, or information about system wide search is implemented.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Guidance specific to the implementation of this operation, such as limitations on the kind of transactions allowed, or information about system wide search is implemented.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

  end;


  TFhirConformanceRestInteractionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestInteractionList;
    function GetCurrent : TFhirConformanceRestInteraction;
  public
    Constructor Create(list : TFhirConformanceRestInteractionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestInteraction read GetCurrent;
  end;


  {@Class TFhirConformanceRestInteractionList
    A list of FhirConformanceRestInteraction
  }
  {!.Net HL7Connect.Fhir.ConformanceRestInteractionList}
  TFhirConformanceRestInteractionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestInteraction;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestInteraction);
  public
    {!script hide}
    function Link : TFhirConformanceRestInteractionList; Overload;
    function Clone : TFhirConformanceRestInteractionList; Overload;
    function GetEnumerator : TFhirConformanceRestInteractionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestInteraction to the end of the list.
    }
    function Append : TFhirConformanceRestInteraction;

    
    {@member AddItem
      Add an already existing FhirConformanceRestInteraction to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestInteraction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestInteraction) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestInteraction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestInteraction;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestInteraction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestInteraction);
    
    {@member Item
       Get the iIndexth FhirConformanceRestInteraction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestInteraction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestInteraction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestInteraction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestInteractions[index : Integer] : TFhirConformanceRestInteraction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceRestOperation : TFhirElement
    Definition of an operation or a named query and with its parameters and their meaning and type.
  }
  {!.Net HL7Connect.Fhir.ConformanceRestOperation}
  TFhirConformanceRestOperation = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FDefinition : TFhirReference{TFhirOperationDefinition};
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDefinition(value : TFhirReference{TFhirOperationDefinition});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceRestOperation; overload;
    function Clone : TFhirConformanceRestOperation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of a query, which is used in the _query parameter when the query is called.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of a query, which is used in the _query parameter when the query is called.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member definition
      Typed access to Where the formal definition can be found. (defined for API consistency)
    }
    property definition : TFhirReference{TFhirOperationDefinition} read FDefinition write SetDefinition;
    {@member definitionElement
      Where the formal definition can be found.
    }
    property definitionElement : TFhirReference{TFhirOperationDefinition} read FDefinition write SetDefinition;

  end;


  TFhirConformanceRestOperationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceRestOperationList;
    function GetCurrent : TFhirConformanceRestOperation;
  public
    Constructor Create(list : TFhirConformanceRestOperationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceRestOperation read GetCurrent;
  end;


  {@Class TFhirConformanceRestOperationList
    A list of FhirConformanceRestOperation
  }
  {!.Net HL7Connect.Fhir.ConformanceRestOperationList}
  TFhirConformanceRestOperationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceRestOperation;
    procedure SetItemN(index : Integer; value : TFhirConformanceRestOperation);
  public
    {!script hide}
    function Link : TFhirConformanceRestOperationList; Overload;
    function Clone : TFhirConformanceRestOperationList; Overload;
    function GetEnumerator : TFhirConformanceRestOperationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceRestOperation to the end of the list.
    }
    function Append : TFhirConformanceRestOperation;

    
    {@member AddItem
      Add an already existing FhirConformanceRestOperation to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceRestOperation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceRestOperation) : Integer;
    

    {@member Insert
      Insert FhirConformanceRestOperation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceRestOperation;
    

    {@member InsertItem
       Insert an existing FhirConformanceRestOperation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceRestOperation);
    
    {@member Item
       Get the iIndexth FhirConformanceRestOperation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceRestOperation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceRestOperation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceRestOperation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceRestOperations[index : Integer] : TFhirConformanceRestOperation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceMessaging : TFhirElement
    A description of the messaging capabilities of the solution.
  }
  {!.Net HL7Connect.Fhir.ConformanceMessaging}
  TFhirConformanceMessaging = class (TFhirBackboneElement)
  private
    FendpointList : TFhirConformanceMessagingEndpointList;
    FReliableCache : TFhirUnsignedInt;
    FDocumentation : TFhirString;
    FeventList : TFhirConformanceMessagingEventList;
    function GetEndpointList : TFhirConformanceMessagingEndpointList;
    function GetHasEndpointList : Boolean;
    Procedure SetReliableCache(value : TFhirUnsignedInt);
    Function GetReliableCacheST : String;
    Procedure SetReliableCacheST(value : String);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
    function GetEventList : TFhirConformanceMessagingEventList;
    function GetHasEventList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceMessaging; overload;
    function Clone : TFhirConformanceMessaging; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member endpointList
      An endpoint (network accessible address) to which messages and/or replies are to be sent.
    }
    property endpointList : TFhirConformanceMessagingEndpointList read GetEndpointList;
    property hasEndpointList : boolean read GetHasEndpointList;

    {@member reliableCache
      Typed access to Length if the receiver's reliable messaging cache in minutes (if a receiver) or how long the cache length on the receiver should be (if a sender).
    }
    property reliableCache : String read GetReliableCacheST write SetReliableCacheST;
    {@member reliableCacheElement
      Length if the receiver's reliable messaging cache in minutes (if a receiver) or how long the cache length on the receiver should be (if a sender).
    }
    property reliableCacheElement : TFhirUnsignedInt read FReliableCache write SetReliableCache;

    {@member documentation
      Typed access to Documentation about the system's messaging capabilities for this endpoint not otherwise documented by the conformance statement.  For example, process for becoming an authorized messaging exchange partner.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Documentation about the system's messaging capabilities for this endpoint not otherwise documented by the conformance statement.  For example, process for becoming an authorized messaging exchange partner.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

    {@member eventList
      A description of the solution's support for an event at this end-point.
    }
    property eventList : TFhirConformanceMessagingEventList read GetEventList;
    property hasEventList : boolean read GetHasEventList;

  end;


  TFhirConformanceMessagingListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceMessagingList;
    function GetCurrent : TFhirConformanceMessaging;
  public
    Constructor Create(list : TFhirConformanceMessagingList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceMessaging read GetCurrent;
  end;


  {@Class TFhirConformanceMessagingList
    A list of FhirConformanceMessaging
  }
  {!.Net HL7Connect.Fhir.ConformanceMessagingList}
  TFhirConformanceMessagingList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceMessaging;
    procedure SetItemN(index : Integer; value : TFhirConformanceMessaging);
  public
    {!script hide}
    function Link : TFhirConformanceMessagingList; Overload;
    function Clone : TFhirConformanceMessagingList; Overload;
    function GetEnumerator : TFhirConformanceMessagingListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceMessaging to the end of the list.
    }
    function Append : TFhirConformanceMessaging;

    
    {@member AddItem
      Add an already existing FhirConformanceMessaging to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceMessaging); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceMessaging) : Integer;
    

    {@member Insert
      Insert FhirConformanceMessaging before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceMessaging;
    

    {@member InsertItem
       Insert an existing FhirConformanceMessaging before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceMessaging);
    
    {@member Item
       Get the iIndexth FhirConformanceMessaging. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceMessaging. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceMessaging);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceMessaging;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceMessagings[index : Integer] : TFhirConformanceMessaging read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceMessagingEndpoint : TFhirElement
    An endpoint (network accessible address) to which messages and/or replies are to be sent.
  }
  {!.Net HL7Connect.Fhir.ConformanceMessagingEndpoint}
  TFhirConformanceMessagingEndpoint = class (TFhirBackboneElement)
  private
    FProtocol : TFhirCoding;
    FAddress : TFhirUri;
    Procedure SetProtocol(value : TFhirCoding);
    Procedure SetAddress(value : TFhirUri);
    Function GetAddressST : String;
    Procedure SetAddressST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceMessagingEndpoint; overload;
    function Clone : TFhirConformanceMessagingEndpoint; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member protocol
      Typed access to A list of the messaging transport protocol(s) identifiers, supported by this endpoint. (defined for API consistency)
    }
    property protocol : TFhirCoding read FProtocol write SetProtocol;
    {@member protocolElement
      A list of the messaging transport protocol(s) identifiers, supported by this endpoint.
    }
    property protocolElement : TFhirCoding read FProtocol write SetProtocol;

    {@member address
      Typed access to The network address of the end-point. For solutions that do not use network addresses for routing, it can be just an identifier.
    }
    property address : String read GetAddressST write SetAddressST;
    {@member addressElement
      The network address of the end-point. For solutions that do not use network addresses for routing, it can be just an identifier.
    }
    property addressElement : TFhirUri read FAddress write SetAddress;

  end;


  TFhirConformanceMessagingEndpointListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceMessagingEndpointList;
    function GetCurrent : TFhirConformanceMessagingEndpoint;
  public
    Constructor Create(list : TFhirConformanceMessagingEndpointList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceMessagingEndpoint read GetCurrent;
  end;


  {@Class TFhirConformanceMessagingEndpointList
    A list of FhirConformanceMessagingEndpoint
  }
  {!.Net HL7Connect.Fhir.ConformanceMessagingEndpointList}
  TFhirConformanceMessagingEndpointList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceMessagingEndpoint;
    procedure SetItemN(index : Integer; value : TFhirConformanceMessagingEndpoint);
  public
    {!script hide}
    function Link : TFhirConformanceMessagingEndpointList; Overload;
    function Clone : TFhirConformanceMessagingEndpointList; Overload;
    function GetEnumerator : TFhirConformanceMessagingEndpointListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceMessagingEndpoint to the end of the list.
    }
    function Append : TFhirConformanceMessagingEndpoint;

    
    {@member AddItem
      Add an already existing FhirConformanceMessagingEndpoint to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceMessagingEndpoint); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceMessagingEndpoint) : Integer;
    

    {@member Insert
      Insert FhirConformanceMessagingEndpoint before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceMessagingEndpoint;
    

    {@member InsertItem
       Insert an existing FhirConformanceMessagingEndpoint before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceMessagingEndpoint);
    
    {@member Item
       Get the iIndexth FhirConformanceMessagingEndpoint. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceMessagingEndpoint. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceMessagingEndpoint);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceMessagingEndpoint;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceMessagingEndpoints[index : Integer] : TFhirConformanceMessagingEndpoint read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceMessagingEvent : TFhirElement
    A description of the solution's support for an event at this end-point.
  }
  {!.Net HL7Connect.Fhir.ConformanceMessagingEvent}
  TFhirConformanceMessagingEvent = class (TFhirBackboneElement)
  private
    FCode : TFhirCoding;
    FCategory : TFhirEnum;
    FMode : TFhirEnum;
    FFocus : TFhirCode;
    FRequest : TFhirReference{TFhirStructureDefinition};
    FResponse : TFhirReference{TFhirStructureDefinition};
    FDocumentation : TFhirString;
    Procedure SetCode(value : TFhirCoding);
    Procedure SetCategory(value : TFhirEnum);
    Function GetCategoryST : TFhirMessageSignificanceCategory;
    Procedure SetCategoryST(value : TFhirMessageSignificanceCategory);
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirMessageConformanceEventMode;
    Procedure SetModeST(value : TFhirMessageConformanceEventMode);
    Procedure SetFocus(value : TFhirCode);
    Function GetFocusST : String;
    Procedure SetFocusST(value : String);
    Procedure SetRequest(value : TFhirReference{TFhirStructureDefinition});
    Procedure SetResponse(value : TFhirReference{TFhirStructureDefinition});
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceMessagingEvent; overload;
    function Clone : TFhirConformanceMessagingEvent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A coded identifier of a supported messaging event. (defined for API consistency)
    }
    property code : TFhirCoding read FCode write SetCode;
    {@member codeElement
      A coded identifier of a supported messaging event.
    }
    property codeElement : TFhirCoding read FCode write SetCode;

    {@member category
      The impact of the content of the message.
    }
    property category : TFhirMessageSignificanceCategory read GetCategoryST write SetCategoryST;
    property categoryElement : TFhirEnum read FCategory write SetCategory;

    {@member mode
      The mode of this event declaration - whether application is sender or receiver.
    }
    property mode : TFhirMessageConformanceEventMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member focus
      Typed access to A resource associated with the event.  This is the resource that defines the event.
    }
    property focus : String read GetFocusST write SetFocusST;
    {@member focusElement
      A resource associated with the event.  This is the resource that defines the event.
    }
    property focusElement : TFhirCode read FFocus write SetFocus;

    {@member request
      Typed access to Information about the request for this event. (defined for API consistency)
    }
    property request : TFhirReference{TFhirStructureDefinition} read FRequest write SetRequest;
    {@member requestElement
      Information about the request for this event.
    }
    property requestElement : TFhirReference{TFhirStructureDefinition} read FRequest write SetRequest;

    {@member response
      Typed access to Information about the response for this event. (defined for API consistency)
    }
    property response : TFhirReference{TFhirStructureDefinition} read FResponse write SetResponse;
    {@member responseElement
      Information about the response for this event.
    }
    property responseElement : TFhirReference{TFhirStructureDefinition} read FResponse write SetResponse;

    {@member documentation
      Typed access to Guidance on how this event is handled, such as internal system trigger points, business rules, etc.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Guidance on how this event is handled, such as internal system trigger points, business rules, etc.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

  end;


  TFhirConformanceMessagingEventListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceMessagingEventList;
    function GetCurrent : TFhirConformanceMessagingEvent;
  public
    Constructor Create(list : TFhirConformanceMessagingEventList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceMessagingEvent read GetCurrent;
  end;


  {@Class TFhirConformanceMessagingEventList
    A list of FhirConformanceMessagingEvent
  }
  {!.Net HL7Connect.Fhir.ConformanceMessagingEventList}
  TFhirConformanceMessagingEventList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceMessagingEvent;
    procedure SetItemN(index : Integer; value : TFhirConformanceMessagingEvent);
  public
    {!script hide}
    function Link : TFhirConformanceMessagingEventList; Overload;
    function Clone : TFhirConformanceMessagingEventList; Overload;
    function GetEnumerator : TFhirConformanceMessagingEventListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceMessagingEvent to the end of the list.
    }
    function Append : TFhirConformanceMessagingEvent;

    
    {@member AddItem
      Add an already existing FhirConformanceMessagingEvent to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceMessagingEvent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceMessagingEvent) : Integer;
    

    {@member Insert
      Insert FhirConformanceMessagingEvent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceMessagingEvent;
    

    {@member InsertItem
       Insert an existing FhirConformanceMessagingEvent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceMessagingEvent);
    
    {@member Item
       Get the iIndexth FhirConformanceMessagingEvent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceMessagingEvent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceMessagingEvent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceMessagingEvent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceMessagingEvents[index : Integer] : TFhirConformanceMessagingEvent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformanceDocument : TFhirElement
    A document definition.
  }
  {!.Net HL7Connect.Fhir.ConformanceDocument}
  TFhirConformanceDocument = class (TFhirBackboneElement)
  private
    FMode : TFhirEnum;
    FDocumentation : TFhirString;
    FProfile : TFhirReference{TFhirStructureDefinition};
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirDocumentMode;
    Procedure SetModeST(value : TFhirDocumentMode);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
    Procedure SetProfile(value : TFhirReference{TFhirStructureDefinition});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformanceDocument; overload;
    function Clone : TFhirConformanceDocument; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member mode
      Mode of this document declaration - whether application is producer or consumer.
    }
    property mode : TFhirDocumentMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member documentation
      Typed access to A description of how the application supports or uses the specified document profile.  For example, when are documents created, what action is taken with consumed documents, etc.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      A description of how the application supports or uses the specified document profile.  For example, when are documents created, what action is taken with consumed documents, etc.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

    {@member profile
      Typed access to A constraint on a resource used in the document. (defined for API consistency)
    }
    property profile : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;
    {@member profileElement
      A constraint on a resource used in the document.
    }
    property profileElement : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;

  end;


  TFhirConformanceDocumentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceDocumentList;
    function GetCurrent : TFhirConformanceDocument;
  public
    Constructor Create(list : TFhirConformanceDocumentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformanceDocument read GetCurrent;
  end;


  {@Class TFhirConformanceDocumentList
    A list of FhirConformanceDocument
  }
  {!.Net HL7Connect.Fhir.ConformanceDocumentList}
  TFhirConformanceDocumentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformanceDocument;
    procedure SetItemN(index : Integer; value : TFhirConformanceDocument);
  public
    {!script hide}
    function Link : TFhirConformanceDocumentList; Overload;
    function Clone : TFhirConformanceDocumentList; Overload;
    function GetEnumerator : TFhirConformanceDocumentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformanceDocument to the end of the list.
    }
    function Append : TFhirConformanceDocument;

    
    {@member AddItem
      Add an already existing FhirConformanceDocument to the end of the list.
    }
    procedure AddItem(value : TFhirConformanceDocument); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformanceDocument) : Integer;
    

    {@member Insert
      Insert FhirConformanceDocument before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformanceDocument;
    

    {@member InsertItem
       Insert an existing FhirConformanceDocument before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformanceDocument);
    
    {@member Item
       Get the iIndexth FhirConformanceDocument. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformanceDocument. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformanceDocument);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformanceDocument;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformanceDocuments[index : Integer] : TFhirConformanceDocument read GetItemN write SetItemN; default;
  End;


  {@Class TFhirConformance : TFhirDomainResource
    A conformance statement is a set of capabilities of a FHIR Server that may be used as a statement of actual server functionality or a statement of required or desired server implementation.
  }
  {!.Net HL7Connect.Fhir.Conformance}
  TFhirConformance = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirConformanceContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FRequirements : TFhirString;
    FCopyright : TFhirString;
    FKind : TFhirEnum;
    FSoftware : TFhirConformanceSoftware;
    FImplementation_ : TFhirConformanceImplementation;
    FFhirVersion : TFhirId;
    FAcceptUnknown : TFhirEnum;
    FformatList : TFhirCodeList;
    FprofileList : TFhirReferenceList{TFhirStructureDefinition};
    FrestList : TFhirConformanceRestList;
    FmessagingList : TFhirConformanceMessagingList;
    FdocumentList : TFhirConformanceDocumentList;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirConformanceContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetKind(value : TFhirEnum);
    Function GetKindST : TFhirConformanceStatementKind;
    Procedure SetKindST(value : TFhirConformanceStatementKind);
    Procedure SetSoftware(value : TFhirConformanceSoftware);
    Procedure SetImplementation_(value : TFhirConformanceImplementation);
    Procedure SetFhirVersion(value : TFhirId);
    Function GetFhirVersionST : String;
    Procedure SetFhirVersionST(value : String);
    Procedure SetAcceptUnknown(value : TFhirEnum);
    Function GetAcceptUnknownST : TFhirUnknownContentCode;
    Procedure SetAcceptUnknownST(value : TFhirUnknownContentCode);
    function GetFormatList : TFhirCodeList;
    function GetHasFormatList : Boolean;
    function GetProfileList : TFhirReferenceList{TFhirStructureDefinition};
    function GetHasProfileList : Boolean;
    function GetRestList : TFhirConformanceRestList;
    function GetHasRestList : Boolean;
    function GetMessagingList : TFhirConformanceMessagingList;
    function GetHasMessagingList : Boolean;
    function GetDocumentList : TFhirConformanceDocumentList;
    function GetHasDocumentList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirConformance; overload;
    function Clone : TFhirConformance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this conformance statement when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this conformance statement is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this conformance statement when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this conformance statement is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member version
      Typed access to The identifier that is used to identify this version of the conformance statement when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the conformance statement when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name identifying the conformance statement.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the conformance statement.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of this conformance statement.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to A flag to indicate that this conformance statement is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      A flag to indicate that this conformance statement is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the conformance.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the conformance.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirConformanceContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date  (and optionally time) when the conformance statement was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the conformance statement changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date  (and optionally time) when the conformance statement was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the conformance statement changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the conformance statement and its use. Typically, this is used when the conformance statement describes a desired rather than an actual solution, for example as a formal expression of requirements as part of an RFP.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the conformance statement and its use. Typically, this is used when the conformance statement describes a desired rather than an actual solution, for example as a formal expression of requirements as part of an RFP.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member requirements
      Typed access to Explains why this conformance statement is needed and why it's been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this conformance statement is needed and why it's been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member copyright
      Typed access to A copyright statement relating to the conformance statement and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the system described by the conformance statement.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the conformance statement and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the system described by the conformance statement.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member kind
      The way that this statement is intended to be used, to describe an actual running instance of software, a particular product (kind not instance of software) or a class of implementation (e.g. a desired purchase).
    }
    property kind : TFhirConformanceStatementKind read GetKindST write SetKindST;
    property kindElement : TFhirEnum read FKind write SetKind;

    {@member software
      Typed access to Software that is covered by this conformance statement.  It is used when the conformance statement describes the capabilities of a particular software version, independent of an installation. (defined for API consistency)
    }
    property software : TFhirConformanceSoftware read FSoftware write SetSoftware;
    {@member softwareElement
      Software that is covered by this conformance statement.  It is used when the conformance statement describes the capabilities of a particular software version, independent of an installation.
    }
    property softwareElement : TFhirConformanceSoftware read FSoftware write SetSoftware;

    {@member implementation_
      Typed access to Identifies a specific implementation instance that is described by the conformance statement - i.e. a particular installation, rather than the capabilities of a software program. (defined for API consistency)
    }
    property implementation_ : TFhirConformanceImplementation read FImplementation_ write SetImplementation_;
    {@member implementation_Element
      Identifies a specific implementation instance that is described by the conformance statement - i.e. a particular installation, rather than the capabilities of a software program.
    }
    property implementation_Element : TFhirConformanceImplementation read FImplementation_ write SetImplementation_;

    {@member fhirVersion
      Typed access to The version of the FHIR specification on which this conformance statement is based.
    }
    property fhirVersion : String read GetFhirVersionST write SetFhirVersionST;
    {@member fhirVersionElement
      The version of the FHIR specification on which this conformance statement is based.
    }
    property fhirVersionElement : TFhirId read FFhirVersion write SetFhirVersion;

    {@member acceptUnknown
      A code that indicates whether the application accepts unknown elements or extensions when reading resources.
    }
    property acceptUnknown : TFhirUnknownContentCode read GetAcceptUnknownST write SetAcceptUnknownST;
    property acceptUnknownElement : TFhirEnum read FAcceptUnknown write SetAcceptUnknown;

    {@member formatList
      A list of the formats supported by this implementation using their content types.
    }
    property formatList : TFhirCodeList read GetFormatList;
    property hasFormatList : boolean read GetHasFormatList;

    {@member profileList
      A list of profiles that represent different use cases supported by the system. For a server, "supported by the system" means the system hosts/produces a set of resources that are conformant to a particular profile, and allows clients that use its services to search using this profile and to find appropriate data. For a client, it means the system will search by this profile and process data according to the guidance implicit in the profile. See further discussion in [Using Profiles].
    }
    property profileList : TFhirReferenceList{TFhirStructureDefinition} read GetProfileList;
    property hasProfileList : boolean read GetHasProfileList;

    {@member restList
      A definition of the restful capabilities of the solution, if any.
    }
    property restList : TFhirConformanceRestList read GetRestList;
    property hasRestList : boolean read GetHasRestList;

    {@member messagingList
      A description of the messaging capabilities of the solution.
    }
    property messagingList : TFhirConformanceMessagingList read GetMessagingList;
    property hasMessagingList : boolean read GetHasMessagingList;

    {@member documentList
      A document definition.
    }
    property documentList : TFhirConformanceDocumentList read GetDocumentList;
    property hasDocumentList : boolean read GetHasDocumentList;

  end;


  TFhirConformanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirConformanceList;
    function GetCurrent : TFhirConformance;
  public
    Constructor Create(list : TFhirConformanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirConformance read GetCurrent;
  end;


  {@Class TFhirConformanceList
    A list of FhirConformance
  }
  {!.Net HL7Connect.Fhir.ConformanceList}
  TFhirConformanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirConformance;
    procedure SetItemN(index : Integer; value : TFhirConformance);
  public
    {!script hide}
    function Link : TFhirConformanceList; Overload;
    function Clone : TFhirConformanceList; Overload;
    function GetEnumerator : TFhirConformanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirConformance to the end of the list.
    }
    function Append : TFhirConformance;

    
    {@member AddItem
      Add an already existing FhirConformance to the end of the list.
    }
    procedure AddItem(value : TFhirConformance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirConformance) : Integer;
    

    {@member Insert
      Insert FhirConformance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirConformance;
    

    {@member InsertItem
       Insert an existing FhirConformance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirConformance);
    
    {@member Item
       Get the iIndexth FhirConformance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirConformance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirConformance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirConformance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirConformances[index : Integer] : TFhirConformance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractActor : TFhirElement
    List of Contract actors.
  }
  {!.Net HL7Connect.Fhir.ContractActor}
  TFhirContractActor = class (TFhirBackboneElement)
  private
    FEntity : TFhirReference{Resource};
    FroleList : TFhirCodeableConceptList;
    Procedure SetEntity(value : TFhirReference{Resource});
    function GetRoleList : TFhirCodeableConceptList;
    function GetHasRoleList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractActor; overload;
    function Clone : TFhirContractActor; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member entity
      Typed access to Who or what actors are assigned roles in this Contract. (defined for API consistency)
    }
    property entity : TFhirReference{Resource} read FEntity write SetEntity;
    {@member entityElement
      Who or what actors are assigned roles in this Contract.
    }
    property entityElement : TFhirReference{Resource} read FEntity write SetEntity;

    {@member roleList
      Role type of actors assigned roles in this Contract.
    }
    property roleList : TFhirCodeableConceptList read GetRoleList;
    property hasRoleList : boolean read GetHasRoleList;

  end;


  TFhirContractActorListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractActorList;
    function GetCurrent : TFhirContractActor;
  public
    Constructor Create(list : TFhirContractActorList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractActor read GetCurrent;
  end;


  {@Class TFhirContractActorList
    A list of FhirContractActor
  }
  {!.Net HL7Connect.Fhir.ContractActorList}
  TFhirContractActorList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractActor;
    procedure SetItemN(index : Integer; value : TFhirContractActor);
  public
    {!script hide}
    function Link : TFhirContractActorList; Overload;
    function Clone : TFhirContractActorList; Overload;
    function GetEnumerator : TFhirContractActorListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractActor to the end of the list.
    }
    function Append : TFhirContractActor;

    
    {@member AddItem
      Add an already existing FhirContractActor to the end of the list.
    }
    procedure AddItem(value : TFhirContractActor); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractActor) : Integer;
    

    {@member Insert
      Insert FhirContractActor before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractActor;
    

    {@member InsertItem
       Insert an existing FhirContractActor before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractActor);
    
    {@member Item
       Get the iIndexth FhirContractActor. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractActor. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractActor);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractActor;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractActors[index : Integer] : TFhirContractActor read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractValuedItem : TFhirElement
    Contract Valued Item List.
  }
  {!.Net HL7Connect.Fhir.ContractValuedItem}
  TFhirContractValuedItem = class (TFhirBackboneElement)
  private
    FEntity : TFhirType;
    FIdentifier : TFhirIdentifier;
    FEffectiveTime : TFhirDateTime;
    FQuantity : TFhirQuantity;
    FUnitPrice : TFhirQuantity;
    FFactor : TFhirDecimal;
    FPoints : TFhirDecimal;
    FNet : TFhirQuantity;
    Procedure SetEntity(value : TFhirType);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetEffectiveTime(value : TFhirDateTime);
    Function GetEffectiveTimeST : TDateAndTime;
    Procedure SetEffectiveTimeST(value : TDateAndTime);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetUnitPrice(value : TFhirQuantity);
    Procedure SetFactor(value : TFhirDecimal);
    Function GetFactorST : String;
    Procedure SetFactorST(value : String);
    Procedure SetPoints(value : TFhirDecimal);
    Function GetPointsST : String;
    Procedure SetPointsST(value : String);
    Procedure SetNet(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractValuedItem; overload;
    function Clone : TFhirContractValuedItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member entity
      Typed access to Specific type of Contract Valued Item that may be priced. (defined for API consistency)
    }
    property entity : TFhirType read FEntity write SetEntity;
    {@member entityElement
      Specific type of Contract Valued Item that may be priced.
    }
    property entityElement : TFhirType read FEntity write SetEntity;

    {@member identifier
      Typed access to Identifies a Contract Valued Item instance. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifies a Contract Valued Item instance.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member effectiveTime
      Typed access to Indicates the time during which this Contract ValuedItem information is effective.
    }
    property effectiveTime : TDateAndTime read GetEffectiveTimeST write SetEffectiveTimeST;
    {@member effectiveTimeElement
      Indicates the time during which this Contract ValuedItem information is effective.
    }
    property effectiveTimeElement : TFhirDateTime read FEffectiveTime write SetEffectiveTime;

    {@member quantity
      Typed access to Specifies the units by which the Contract Valued Item is measured or counted, and quantifies the countable or measurable Contract Valued Item instances. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      Specifies the units by which the Contract Valued Item is measured or counted, and quantifies the countable or measurable Contract Valued Item instances.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member unitPrice
      Typed access to A Contract Valued Item unit valuation measure. (defined for API consistency)
    }
    property unitPrice : TFhirQuantity read FUnitPrice write SetUnitPrice;
    {@member unitPriceElement
      A Contract Valued Item unit valuation measure.
    }
    property unitPriceElement : TFhirQuantity read FUnitPrice write SetUnitPrice;

    {@member factor
      Typed access to A real number that represents a multiplier used in determining the overall value of the Contract Valued Item delivered. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factor : String read GetFactorST write SetFactorST;
    {@member factorElement
      A real number that represents a multiplier used in determining the overall value of the Contract Valued Item delivered. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factorElement : TFhirDecimal read FFactor write SetFactor;

    {@member points
      Typed access to An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the Contract Valued Item delivered. The concept of Points allows for assignment of point values for a Contract Valued Item, such that a monetary amount can be assigned to each point.
    }
    property points : String read GetPointsST write SetPointsST;
    {@member pointsElement
      An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the Contract Valued Item delivered. The concept of Points allows for assignment of point values for a Contract Valued Item, such that a monetary amount can be assigned to each point.
    }
    property pointsElement : TFhirDecimal read FPoints write SetPoints;

    {@member net
      Typed access to Expresses the product of the Contract Valued Item unitQuantity and the unitPriceAmt. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied. (defined for API consistency)
    }
    property net : TFhirQuantity read FNet write SetNet;
    {@member netElement
      Expresses the product of the Contract Valued Item unitQuantity and the unitPriceAmt. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied.
    }
    property netElement : TFhirQuantity read FNet write SetNet;

  end;


  TFhirContractValuedItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractValuedItemList;
    function GetCurrent : TFhirContractValuedItem;
  public
    Constructor Create(list : TFhirContractValuedItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractValuedItem read GetCurrent;
  end;


  {@Class TFhirContractValuedItemList
    A list of FhirContractValuedItem
  }
  {!.Net HL7Connect.Fhir.ContractValuedItemList}
  TFhirContractValuedItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractValuedItem;
    procedure SetItemN(index : Integer; value : TFhirContractValuedItem);
  public
    {!script hide}
    function Link : TFhirContractValuedItemList; Overload;
    function Clone : TFhirContractValuedItemList; Overload;
    function GetEnumerator : TFhirContractValuedItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractValuedItem to the end of the list.
    }
    function Append : TFhirContractValuedItem;

    
    {@member AddItem
      Add an already existing FhirContractValuedItem to the end of the list.
    }
    procedure AddItem(value : TFhirContractValuedItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractValuedItem) : Integer;
    

    {@member Insert
      Insert FhirContractValuedItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractValuedItem;
    

    {@member InsertItem
       Insert an existing FhirContractValuedItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractValuedItem);
    
    {@member Item
       Get the iIndexth FhirContractValuedItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractValuedItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractValuedItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractValuedItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractValuedItems[index : Integer] : TFhirContractValuedItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractSigner : TFhirElement
    Party signing this Contract.
  }
  {!.Net HL7Connect.Fhir.ContractSigner}
  TFhirContractSigner = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FParty : TFhirReference{Resource};
    FSignature : TFhirString;
    Procedure SetType_(value : TFhirCoding);
    Procedure SetParty(value : TFhirReference{Resource});
    Procedure SetSignature(value : TFhirString);
    Function GetSignatureST : String;
    Procedure SetSignatureST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractSigner; overload;
    function Clone : TFhirContractSigner; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Role of this Contract signer, e.g. notary, grantee. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      Role of this Contract signer, e.g. notary, grantee.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member party
      Typed access to Party which is a signator to this Contract. (defined for API consistency)
    }
    property party : TFhirReference{Resource} read FParty write SetParty;
    {@member partyElement
      Party which is a signator to this Contract.
    }
    property partyElement : TFhirReference{Resource} read FParty write SetParty;

    {@member signature
      Typed access to Legally binding Contract DSIG signature contents in Base64.
    }
    property signature : String read GetSignatureST write SetSignatureST;
    {@member signatureElement
      Legally binding Contract DSIG signature contents in Base64.
    }
    property signatureElement : TFhirString read FSignature write SetSignature;

  end;


  TFhirContractSignerListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractSignerList;
    function GetCurrent : TFhirContractSigner;
  public
    Constructor Create(list : TFhirContractSignerList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractSigner read GetCurrent;
  end;


  {@Class TFhirContractSignerList
    A list of FhirContractSigner
  }
  {!.Net HL7Connect.Fhir.ContractSignerList}
  TFhirContractSignerList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractSigner;
    procedure SetItemN(index : Integer; value : TFhirContractSigner);
  public
    {!script hide}
    function Link : TFhirContractSignerList; Overload;
    function Clone : TFhirContractSignerList; Overload;
    function GetEnumerator : TFhirContractSignerListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractSigner to the end of the list.
    }
    function Append : TFhirContractSigner;

    
    {@member AddItem
      Add an already existing FhirContractSigner to the end of the list.
    }
    procedure AddItem(value : TFhirContractSigner); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractSigner) : Integer;
    

    {@member Insert
      Insert FhirContractSigner before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractSigner;
    

    {@member InsertItem
       Insert an existing FhirContractSigner before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractSigner);
    
    {@member Item
       Get the iIndexth FhirContractSigner. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractSigner. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractSigner);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractSigner;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractSigners[index : Integer] : TFhirContractSigner read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractTerm : TFhirElement
    One or more Contract Provisions, which may be related and conveyed as a group, and may contain nested groups.
  }
  {!.Net HL7Connect.Fhir.ContractTerm}
  TFhirContractTerm = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirIdentifier;
    FIssued : TFhirDateTime;
    FApplies : TFhirPeriod;
    FType_ : TFhirCodeableConcept;
    FSubType : TFhirCodeableConcept;
    FSubject : TFhirReference{TFhirReference};
    FactionList : TFhirCodeableConceptList;
    FactionReasonList : TFhirCodeableConceptList;
    FactorList : TFhirContractTermActorList;
    FText : TFhirString;
    FvaluedItemList : TFhirContractTermValuedItemList;
    FgroupList : TFhirContractTermList;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetIssued(value : TFhirDateTime);
    Function GetIssuedST : TDateAndTime;
    Procedure SetIssuedST(value : TDateAndTime);
    Procedure SetApplies(value : TFhirPeriod);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetSubType(value : TFhirCodeableConcept);
    Procedure SetSubject(value : TFhirReference{TFhirReference});
    function GetActionList : TFhirCodeableConceptList;
    function GetHasActionList : Boolean;
    function GetActionReasonList : TFhirCodeableConceptList;
    function GetHasActionReasonList : Boolean;
    function GetActorList : TFhirContractTermActorList;
    function GetHasActorList : Boolean;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    function GetValuedItemList : TFhirContractTermValuedItemList;
    function GetHasValuedItemList : Boolean;
    function GetGroupList : TFhirContractTermList;
    function GetHasGroupList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractTerm; overload;
    function Clone : TFhirContractTerm; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to Unique identifier for this particular Contract Provision. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Unique identifier for this particular Contract Provision.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member issued
      Typed access to When this Contract Provision was issued.
    }
    property issued : TDateAndTime read GetIssuedST write SetIssuedST;
    {@member issuedElement
      When this Contract Provision was issued.
    }
    property issuedElement : TFhirDateTime read FIssued write SetIssued;

    {@member applies
      Typed access to Relevant time or time-period when this Contract Provision is applicable. (defined for API consistency)
    }
    property applies : TFhirPeriod read FApplies write SetApplies;
    {@member appliesElement
      Relevant time or time-period when this Contract Provision is applicable.
    }
    property appliesElement : TFhirPeriod read FApplies write SetApplies;

    {@member type_
      Typed access to Type of Contract Provision such as specific requirements, purposes for actions, obligations, prohibitions, e.g. life time maximum benefit. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Type of Contract Provision such as specific requirements, purposes for actions, obligations, prohibitions, e.g. life time maximum benefit.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member subType
      Typed access to Subtype of this Contract Provision, e.g. life time maximum payment for a contract term for specific valued item, e.g. disability payment. (defined for API consistency)
    }
    property subType : TFhirCodeableConcept read FSubType write SetSubType;
    {@member subTypeElement
      Subtype of this Contract Provision, e.g. life time maximum payment for a contract term for specific valued item, e.g. disability payment.
    }
    property subTypeElement : TFhirCodeableConcept read FSubType write SetSubType;

    {@member subject
      Typed access to Who or what this Contract Provision is about. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirReference} read FSubject write SetSubject;
    {@member subjectElement
      Who or what this Contract Provision is about.
    }
    property subjectElement : TFhirReference{TFhirReference} read FSubject write SetSubject;

    {@member actionList
      Action stipulated by this Contract Provision.
    }
    property actionList : TFhirCodeableConceptList read GetActionList;
    property hasActionList : boolean read GetHasActionList;

    {@member actionReasonList
      Reason or purpose for the action stipulated by this Contract Provision.
    }
    property actionReasonList : TFhirCodeableConceptList read GetActionReasonList;
    property hasActionReasonList : boolean read GetHasActionReasonList;

    {@member actorList
      List of actors participating in this Contract Provision.
    }
    property actorList : TFhirContractTermActorList read GetActorList;
    property hasActorList : boolean read GetHasActorList;

    {@member text
      Typed access to Human readable form of this Contract Provision.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Human readable form of this Contract Provision.
    }
    property textElement : TFhirString read FText write SetText;

    {@member valuedItemList
      Contract Provision Valued Item List.
    }
    property valuedItemList : TFhirContractTermValuedItemList read GetValuedItemList;
    property hasValuedItemList : boolean read GetHasValuedItemList;

    {@member groupList
      Nested group of Contract Provisions.
    }
    property groupList : TFhirContractTermList read GetGroupList;
    property hasGroupList : boolean read GetHasGroupList;

  end;


  TFhirContractTermListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractTermList;
    function GetCurrent : TFhirContractTerm;
  public
    Constructor Create(list : TFhirContractTermList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractTerm read GetCurrent;
  end;


  {@Class TFhirContractTermList
    A list of FhirContractTerm
  }
  {!.Net HL7Connect.Fhir.ContractTermList}
  TFhirContractTermList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractTerm;
    procedure SetItemN(index : Integer; value : TFhirContractTerm);
  public
    {!script hide}
    function Link : TFhirContractTermList; Overload;
    function Clone : TFhirContractTermList; Overload;
    function GetEnumerator : TFhirContractTermListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractTerm to the end of the list.
    }
    function Append : TFhirContractTerm;

    
    {@member AddItem
      Add an already existing FhirContractTerm to the end of the list.
    }
    procedure AddItem(value : TFhirContractTerm); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractTerm) : Integer;
    

    {@member Insert
      Insert FhirContractTerm before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractTerm;
    

    {@member InsertItem
       Insert an existing FhirContractTerm before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractTerm);
    
    {@member Item
       Get the iIndexth FhirContractTerm. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractTerm. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractTerm);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractTerm;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractTerms[index : Integer] : TFhirContractTerm read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractTermActor : TFhirElement
    List of actors participating in this Contract Provision.
  }
  {!.Net HL7Connect.Fhir.ContractTermActor}
  TFhirContractTermActor = class (TFhirBackboneElement)
  private
    FEntity : TFhirReference{Resource};
    FroleList : TFhirCodeableConceptList;
    Procedure SetEntity(value : TFhirReference{Resource});
    function GetRoleList : TFhirCodeableConceptList;
    function GetHasRoleList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractTermActor; overload;
    function Clone : TFhirContractTermActor; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member entity
      Typed access to The actor assigned a role in this Contract Provision. (defined for API consistency)
    }
    property entity : TFhirReference{Resource} read FEntity write SetEntity;
    {@member entityElement
      The actor assigned a role in this Contract Provision.
    }
    property entityElement : TFhirReference{Resource} read FEntity write SetEntity;

    {@member roleList
      Role played by the actor assigned this role in this Contract Provision.
    }
    property roleList : TFhirCodeableConceptList read GetRoleList;
    property hasRoleList : boolean read GetHasRoleList;

  end;


  TFhirContractTermActorListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractTermActorList;
    function GetCurrent : TFhirContractTermActor;
  public
    Constructor Create(list : TFhirContractTermActorList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractTermActor read GetCurrent;
  end;


  {@Class TFhirContractTermActorList
    A list of FhirContractTermActor
  }
  {!.Net HL7Connect.Fhir.ContractTermActorList}
  TFhirContractTermActorList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractTermActor;
    procedure SetItemN(index : Integer; value : TFhirContractTermActor);
  public
    {!script hide}
    function Link : TFhirContractTermActorList; Overload;
    function Clone : TFhirContractTermActorList; Overload;
    function GetEnumerator : TFhirContractTermActorListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractTermActor to the end of the list.
    }
    function Append : TFhirContractTermActor;

    
    {@member AddItem
      Add an already existing FhirContractTermActor to the end of the list.
    }
    procedure AddItem(value : TFhirContractTermActor); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractTermActor) : Integer;
    

    {@member Insert
      Insert FhirContractTermActor before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractTermActor;
    

    {@member InsertItem
       Insert an existing FhirContractTermActor before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractTermActor);
    
    {@member Item
       Get the iIndexth FhirContractTermActor. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractTermActor. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractTermActor);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractTermActor;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractTermActors[index : Integer] : TFhirContractTermActor read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractTermValuedItem : TFhirElement
    Contract Provision Valued Item List.
  }
  {!.Net HL7Connect.Fhir.ContractTermValuedItem}
  TFhirContractTermValuedItem = class (TFhirBackboneElement)
  private
    FEntity : TFhirType;
    FIdentifier : TFhirIdentifier;
    FEffectiveTime : TFhirDateTime;
    FQuantity : TFhirQuantity;
    FUnitPrice : TFhirQuantity;
    FFactor : TFhirDecimal;
    FPoints : TFhirDecimal;
    FNet : TFhirQuantity;
    Procedure SetEntity(value : TFhirType);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetEffectiveTime(value : TFhirDateTime);
    Function GetEffectiveTimeST : TDateAndTime;
    Procedure SetEffectiveTimeST(value : TDateAndTime);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetUnitPrice(value : TFhirQuantity);
    Procedure SetFactor(value : TFhirDecimal);
    Function GetFactorST : String;
    Procedure SetFactorST(value : String);
    Procedure SetPoints(value : TFhirDecimal);
    Function GetPointsST : String;
    Procedure SetPointsST(value : String);
    Procedure SetNet(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractTermValuedItem; overload;
    function Clone : TFhirContractTermValuedItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member entity
      Typed access to Specific type of Contract Provision Valued Item that may be priced. (defined for API consistency)
    }
    property entity : TFhirType read FEntity write SetEntity;
    {@member entityElement
      Specific type of Contract Provision Valued Item that may be priced.
    }
    property entityElement : TFhirType read FEntity write SetEntity;

    {@member identifier
      Typed access to Identifies a Contract Provision Valued Item instance. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifies a Contract Provision Valued Item instance.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member effectiveTime
      Typed access to Indicates the time during which this Contract Term ValuedItem information is effective.
    }
    property effectiveTime : TDateAndTime read GetEffectiveTimeST write SetEffectiveTimeST;
    {@member effectiveTimeElement
      Indicates the time during which this Contract Term ValuedItem information is effective.
    }
    property effectiveTimeElement : TFhirDateTime read FEffectiveTime write SetEffectiveTime;

    {@member quantity
      Typed access to Specifies the units by which the Contract Provision Valued Item is measured or counted, and quantifies the countable or measurable Contract Term Valued Item instances. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      Specifies the units by which the Contract Provision Valued Item is measured or counted, and quantifies the countable or measurable Contract Term Valued Item instances.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member unitPrice
      Typed access to A Contract Provision Valued Item unit valuation measure. (defined for API consistency)
    }
    property unitPrice : TFhirQuantity read FUnitPrice write SetUnitPrice;
    {@member unitPriceElement
      A Contract Provision Valued Item unit valuation measure.
    }
    property unitPriceElement : TFhirQuantity read FUnitPrice write SetUnitPrice;

    {@member factor
      Typed access to A real number that represents a multiplier used in determining the overall value of the Contract Provision Valued Item delivered. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factor : String read GetFactorST write SetFactorST;
    {@member factorElement
      A real number that represents a multiplier used in determining the overall value of the Contract Provision Valued Item delivered. The concept of a Factor allows for a discount or surcharge multiplier to be applied to a monetary amount.
    }
    property factorElement : TFhirDecimal read FFactor write SetFactor;

    {@member points
      Typed access to An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the Contract Provision Valued Item delivered. The concept of Points allows for assignment of point values for a Contract ProvisionValued Item, such that a monetary amount can be assigned to each point.
    }
    property points : String read GetPointsST write SetPointsST;
    {@member pointsElement
      An amount that expresses the weighting (based on difficulty, cost and/or resource intensiveness) associated with the Contract Provision Valued Item delivered. The concept of Points allows for assignment of point values for a Contract ProvisionValued Item, such that a monetary amount can be assigned to each point.
    }
    property pointsElement : TFhirDecimal read FPoints write SetPoints;

    {@member net
      Typed access to Expresses the product of the Contract Provision Valued Item unitQuantity and the unitPriceAmt. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied. (defined for API consistency)
    }
    property net : TFhirQuantity read FNet write SetNet;
    {@member netElement
      Expresses the product of the Contract Provision Valued Item unitQuantity and the unitPriceAmt. For example, the formula: unit Quantity * unit Price (Cost per Point) * factor Number  * points = net Amount. Quantity, factor and points are assumed to be 1 if not supplied.
    }
    property netElement : TFhirQuantity read FNet write SetNet;

  end;


  TFhirContractTermValuedItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractTermValuedItemList;
    function GetCurrent : TFhirContractTermValuedItem;
  public
    Constructor Create(list : TFhirContractTermValuedItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractTermValuedItem read GetCurrent;
  end;


  {@Class TFhirContractTermValuedItemList
    A list of FhirContractTermValuedItem
  }
  {!.Net HL7Connect.Fhir.ContractTermValuedItemList}
  TFhirContractTermValuedItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractTermValuedItem;
    procedure SetItemN(index : Integer; value : TFhirContractTermValuedItem);
  public
    {!script hide}
    function Link : TFhirContractTermValuedItemList; Overload;
    function Clone : TFhirContractTermValuedItemList; Overload;
    function GetEnumerator : TFhirContractTermValuedItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractTermValuedItem to the end of the list.
    }
    function Append : TFhirContractTermValuedItem;

    
    {@member AddItem
      Add an already existing FhirContractTermValuedItem to the end of the list.
    }
    procedure AddItem(value : TFhirContractTermValuedItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractTermValuedItem) : Integer;
    

    {@member Insert
      Insert FhirContractTermValuedItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractTermValuedItem;
    

    {@member InsertItem
       Insert an existing FhirContractTermValuedItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractTermValuedItem);
    
    {@member Item
       Get the iIndexth FhirContractTermValuedItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractTermValuedItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractTermValuedItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractTermValuedItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractTermValuedItems[index : Integer] : TFhirContractTermValuedItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractFriendly : TFhirElement
    The "patient friendly language" versionof the Contract in whole or in parts. "Patient friendly language" means the representation of the Contract and Contract Provisions in a manner that is readily accessible and understandable by a layperson in accordance with best practices for communication styles that ensure that those agreeing to or signing the Contract understand the roles, actions, obligations, responsibilities, and implication of the agreement.
  }
  {!.Net HL7Connect.Fhir.ContractFriendly}
  TFhirContractFriendly = class (TFhirBackboneElement)
  private
    FContent : TFhirType;
    Procedure SetContent(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractFriendly; overload;
    function Clone : TFhirContractFriendly; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member content
      Typed access to Human readable rendering of this Contract in a format and representation intended to enhance comprehension and ensure understandability. (defined for API consistency)
    }
    property content : TFhirType read FContent write SetContent;
    {@member contentElement
      Human readable rendering of this Contract in a format and representation intended to enhance comprehension and ensure understandability.
    }
    property contentElement : TFhirType read FContent write SetContent;

  end;


  TFhirContractFriendlyListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractFriendlyList;
    function GetCurrent : TFhirContractFriendly;
  public
    Constructor Create(list : TFhirContractFriendlyList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractFriendly read GetCurrent;
  end;


  {@Class TFhirContractFriendlyList
    A list of FhirContractFriendly
  }
  {!.Net HL7Connect.Fhir.ContractFriendlyList}
  TFhirContractFriendlyList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractFriendly;
    procedure SetItemN(index : Integer; value : TFhirContractFriendly);
  public
    {!script hide}
    function Link : TFhirContractFriendlyList; Overload;
    function Clone : TFhirContractFriendlyList; Overload;
    function GetEnumerator : TFhirContractFriendlyListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractFriendly to the end of the list.
    }
    function Append : TFhirContractFriendly;

    
    {@member AddItem
      Add an already existing FhirContractFriendly to the end of the list.
    }
    procedure AddItem(value : TFhirContractFriendly); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractFriendly) : Integer;
    

    {@member Insert
      Insert FhirContractFriendly before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractFriendly;
    

    {@member InsertItem
       Insert an existing FhirContractFriendly before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractFriendly);
    
    {@member Item
       Get the iIndexth FhirContractFriendly. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractFriendly. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractFriendly);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractFriendly;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractFriendlies[index : Integer] : TFhirContractFriendly read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractLegal : TFhirElement
    List of Legal expressions or representations of this Contract.
  }
  {!.Net HL7Connect.Fhir.ContractLegal}
  TFhirContractLegal = class (TFhirBackboneElement)
  private
    FContent : TFhirType;
    Procedure SetContent(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractLegal; overload;
    function Clone : TFhirContractLegal; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member content
      Typed access to Contract legal text in human renderable form. (defined for API consistency)
    }
    property content : TFhirType read FContent write SetContent;
    {@member contentElement
      Contract legal text in human renderable form.
    }
    property contentElement : TFhirType read FContent write SetContent;

  end;


  TFhirContractLegalListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractLegalList;
    function GetCurrent : TFhirContractLegal;
  public
    Constructor Create(list : TFhirContractLegalList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractLegal read GetCurrent;
  end;


  {@Class TFhirContractLegalList
    A list of FhirContractLegal
  }
  {!.Net HL7Connect.Fhir.ContractLegalList}
  TFhirContractLegalList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractLegal;
    procedure SetItemN(index : Integer; value : TFhirContractLegal);
  public
    {!script hide}
    function Link : TFhirContractLegalList; Overload;
    function Clone : TFhirContractLegalList; Overload;
    function GetEnumerator : TFhirContractLegalListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractLegal to the end of the list.
    }
    function Append : TFhirContractLegal;

    
    {@member AddItem
      Add an already existing FhirContractLegal to the end of the list.
    }
    procedure AddItem(value : TFhirContractLegal); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractLegal) : Integer;
    

    {@member Insert
      Insert FhirContractLegal before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractLegal;
    

    {@member InsertItem
       Insert an existing FhirContractLegal before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractLegal);
    
    {@member Item
       Get the iIndexth FhirContractLegal. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractLegal. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractLegal);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractLegal;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractLegals[index : Integer] : TFhirContractLegal read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContractRule : TFhirElement
    List of Computable Policy Rule Language Representations of this Contract.
  }
  {!.Net HL7Connect.Fhir.ContractRule}
  TFhirContractRule = class (TFhirBackboneElement)
  private
    FContent : TFhirType;
    Procedure SetContent(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContractRule; overload;
    function Clone : TFhirContractRule; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member content
      Typed access to Computable Contract conveyed using a policy rule language (e.g. XACML, DKAL, SecPal). (defined for API consistency)
    }
    property content : TFhirType read FContent write SetContent;
    {@member contentElement
      Computable Contract conveyed using a policy rule language (e.g. XACML, DKAL, SecPal).
    }
    property contentElement : TFhirType read FContent write SetContent;

  end;


  TFhirContractRuleListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractRuleList;
    function GetCurrent : TFhirContractRule;
  public
    Constructor Create(list : TFhirContractRuleList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContractRule read GetCurrent;
  end;


  {@Class TFhirContractRuleList
    A list of FhirContractRule
  }
  {!.Net HL7Connect.Fhir.ContractRuleList}
  TFhirContractRuleList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContractRule;
    procedure SetItemN(index : Integer; value : TFhirContractRule);
  public
    {!script hide}
    function Link : TFhirContractRuleList; Overload;
    function Clone : TFhirContractRuleList; Overload;
    function GetEnumerator : TFhirContractRuleListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContractRule to the end of the list.
    }
    function Append : TFhirContractRule;

    
    {@member AddItem
      Add an already existing FhirContractRule to the end of the list.
    }
    procedure AddItem(value : TFhirContractRule); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContractRule) : Integer;
    

    {@member Insert
      Insert FhirContractRule before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContractRule;
    

    {@member InsertItem
       Insert an existing FhirContractRule before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContractRule);
    
    {@member Item
       Get the iIndexth FhirContractRule. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContractRule. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContractRule);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContractRule;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContractRules[index : Integer] : TFhirContractRule read GetItemN write SetItemN; default;
  End;


  {@Class TFhirContract : TFhirDomainResource
    A formal agreement between parties regarding the conduct of business, exchange of information or other matters.
  }
  {!.Net HL7Connect.Fhir.Contract}
  TFhirContract = class (TFhirDomainResource)
  private
    FIdentifier : TFhirIdentifier;
    FIssued : TFhirDateTime;
    FApplies : TFhirPeriod;
    FsubjectList : TFhirReferenceList{TFhirReference};
    FauthorityList : TFhirReferenceList{TFhirOrganization};
    FdomainList : TFhirReferenceList{TFhirLocation};
    FType_ : TFhirCodeableConcept;
    FsubTypeList : TFhirCodeableConceptList;
    FactionList : TFhirCodeableConceptList;
    FactionReasonList : TFhirCodeableConceptList;
    FactorList : TFhirContractActorList;
    FvaluedItemList : TFhirContractValuedItemList;
    FsignerList : TFhirContractSignerList;
    FtermList : TFhirContractTermList;
    FBinding : TFhirType;
    FfriendlyList : TFhirContractFriendlyList;
    FlegalList : TFhirContractLegalList;
    FruleList : TFhirContractRuleList;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetIssued(value : TFhirDateTime);
    Function GetIssuedST : TDateAndTime;
    Procedure SetIssuedST(value : TDateAndTime);
    Procedure SetApplies(value : TFhirPeriod);
    function GetSubjectList : TFhirReferenceList{TFhirReference};
    function GetHasSubjectList : Boolean;
    function GetAuthorityList : TFhirReferenceList{TFhirOrganization};
    function GetHasAuthorityList : Boolean;
    function GetDomainList : TFhirReferenceList{TFhirLocation};
    function GetHasDomainList : Boolean;
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetSubTypeList : TFhirCodeableConceptList;
    function GetHasSubTypeList : Boolean;
    function GetActionList : TFhirCodeableConceptList;
    function GetHasActionList : Boolean;
    function GetActionReasonList : TFhirCodeableConceptList;
    function GetHasActionReasonList : Boolean;
    function GetActorList : TFhirContractActorList;
    function GetHasActorList : Boolean;
    function GetValuedItemList : TFhirContractValuedItemList;
    function GetHasValuedItemList : Boolean;
    function GetSignerList : TFhirContractSignerList;
    function GetHasSignerList : Boolean;
    function GetTermList : TFhirContractTermList;
    function GetHasTermList : Boolean;
    Procedure SetBinding(value : TFhirType);
    function GetFriendlyList : TFhirContractFriendlyList;
    function GetHasFriendlyList : Boolean;
    function GetLegalList : TFhirContractLegalList;
    function GetHasLegalList : Boolean;
    function GetRuleList : TFhirContractRuleList;
    function GetHasRuleList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirContract; overload;
    function Clone : TFhirContract; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifier
      Typed access to Unique identifier for this Contract. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Unique identifier for this Contract.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member issued
      Typed access to When this  Contract was issued.
    }
    property issued : TDateAndTime read GetIssuedST write SetIssuedST;
    {@member issuedElement
      When this  Contract was issued.
    }
    property issuedElement : TFhirDateTime read FIssued write SetIssued;

    {@member applies
      Typed access to Relevant time or time-period when this Contract is applicable. (defined for API consistency)
    }
    property applies : TFhirPeriod read FApplies write SetApplies;
    {@member appliesElement
      Relevant time or time-period when this Contract is applicable.
    }
    property appliesElement : TFhirPeriod read FApplies write SetApplies;

    {@member subjectList
      Who and/or what this Contract is about: typically a Patient, Organization, or valued items such as goods and services.
    }
    property subjectList : TFhirReferenceList{TFhirReference} read GetSubjectList;
    property hasSubjectList : boolean read GetHasSubjectList;

    {@member authorityList
      A formally or informally recognized grouping of people, principals, organizations, or jurisdictions formed for the purpose of achieving some form of collective action such as the promulgation, administration and enforcement of contracts and policies.
    }
    property authorityList : TFhirReferenceList{TFhirOrganization} read GetAuthorityList;
    property hasAuthorityList : boolean read GetHasAuthorityList;

    {@member domainList
      Recognized governance framework or system operating with a circumscribed scope in accordance with specified principles, policies, processes or procedures for managing rights, actions, or behaviors of parties or principals relative to resources.
    }
    property domainList : TFhirReferenceList{TFhirLocation} read GetDomainList;
    property hasDomainList : boolean read GetHasDomainList;

    {@member type_
      Typed access to Type of Contract such as an insurance policy, real estate contract, a will, power of attorny, Privacy or Security policy , trust framework agreement, etc. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Type of Contract such as an insurance policy, real estate contract, a will, power of attorny, Privacy or Security policy , trust framework agreement, etc.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member subTypeList
      More specific type or specialization of an overarching or more general contract such as auto insurance, home owner  insurance, prenupial agreement, Advanced-Directive, or privacy consent.
    }
    property subTypeList : TFhirCodeableConceptList read GetSubTypeList;
    property hasSubTypeList : boolean read GetHasSubTypeList;

    {@member actionList
      Action stipulated by this Contract.
    }
    property actionList : TFhirCodeableConceptList read GetActionList;
    property hasActionList : boolean read GetHasActionList;

    {@member actionReasonList
      Reason for action stipulated by this Contract.
    }
    property actionReasonList : TFhirCodeableConceptList read GetActionReasonList;
    property hasActionReasonList : boolean read GetHasActionReasonList;

    {@member actorList
      List of Contract actors.
    }
    property actorList : TFhirContractActorList read GetActorList;
    property hasActorList : boolean read GetHasActorList;

    {@member valuedItemList
      Contract Valued Item List.
    }
    property valuedItemList : TFhirContractValuedItemList read GetValuedItemList;
    property hasValuedItemList : boolean read GetHasValuedItemList;

    {@member signerList
      Party signing this Contract.
    }
    property signerList : TFhirContractSignerList read GetSignerList;
    property hasSignerList : boolean read GetHasSignerList;

    {@member termList
      One or more Contract Provisions, which may be related and conveyed as a group, and may contain nested groups.
    }
    property termList : TFhirContractTermList read GetTermList;
    property hasTermList : boolean read GetHasTermList;

    {@member binding
      Typed access to Legally binding Contract: This is the signed and legally recognized representation of the Contract, which is considered the "source of truth" and which would be the basis for legal action related to enforcement of this Contract. (defined for API consistency)
    }
    property binding : TFhirType read FBinding write SetBinding;
    {@member bindingElement
      Legally binding Contract: This is the signed and legally recognized representation of the Contract, which is considered the "source of truth" and which would be the basis for legal action related to enforcement of this Contract.
    }
    property bindingElement : TFhirType read FBinding write SetBinding;

    {@member friendlyList
      The "patient friendly language" versionof the Contract in whole or in parts. "Patient friendly language" means the representation of the Contract and Contract Provisions in a manner that is readily accessible and understandable by a layperson in accordance with best practices for communication styles that ensure that those agreeing to or signing the Contract understand the roles, actions, obligations, responsibilities, and implication of the agreement.
    }
    property friendlyList : TFhirContractFriendlyList read GetFriendlyList;
    property hasFriendlyList : boolean read GetHasFriendlyList;

    {@member legalList
      List of Legal expressions or representations of this Contract.
    }
    property legalList : TFhirContractLegalList read GetLegalList;
    property hasLegalList : boolean read GetHasLegalList;

    {@member ruleList
      List of Computable Policy Rule Language Representations of this Contract.
    }
    property ruleList : TFhirContractRuleList read GetRuleList;
    property hasRuleList : boolean read GetHasRuleList;

  end;


  TFhirContractListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirContractList;
    function GetCurrent : TFhirContract;
  public
    Constructor Create(list : TFhirContractList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirContract read GetCurrent;
  end;


  {@Class TFhirContractList
    A list of FhirContract
  }
  {!.Net HL7Connect.Fhir.ContractList}
  TFhirContractList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirContract;
    procedure SetItemN(index : Integer; value : TFhirContract);
  public
    {!script hide}
    function Link : TFhirContractList; Overload;
    function Clone : TFhirContractList; Overload;
    function GetEnumerator : TFhirContractListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirContract to the end of the list.
    }
    function Append : TFhirContract;

    
    {@member AddItem
      Add an already existing FhirContract to the end of the list.
    }
    procedure AddItem(value : TFhirContract); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirContract) : Integer;
    

    {@member Insert
      Insert FhirContract before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirContract;
    

    {@member InsertItem
       Insert an existing FhirContract before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirContract);
    
    {@member Item
       Get the iIndexth FhirContract. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirContract. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirContract);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirContract;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirContracts[index : Integer] : TFhirContract read GetItemN write SetItemN; default;
  End;


  {@Class TFhirCoverage : TFhirDomainResource
    Financial instrument which may be used to pay for or reimburse health care products and services.
  }
  {!.Net HL7Connect.Fhir.Coverage}
  TFhirCoverage = class (TFhirDomainResource)
  private
    FIssuer : TFhirReference{TFhirOrganization};
    FBin : TFhirIdentifier;
    FPeriod : TFhirPeriod;
    FType_ : TFhirCoding;
    FSubscriberId : TFhirIdentifier;
    FidentifierList : TFhirIdentifierList;
    FGroup : TFhirString;
    FPlan : TFhirString;
    FSubPlan : TFhirString;
    FDependent : TFhirPositiveInt;
    FSequence : TFhirPositiveInt;
    FSubscriber : TFhirReference{TFhirPatient};
    FNetwork : TFhirIdentifier;
    FcontractList : TFhirReferenceList{TFhirContract};
    Procedure SetIssuer(value : TFhirReference{TFhirOrganization});
    Procedure SetBin(value : TFhirIdentifier);
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetSubscriberId(value : TFhirIdentifier);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetGroup(value : TFhirString);
    Function GetGroupST : String;
    Procedure SetGroupST(value : String);
    Procedure SetPlan(value : TFhirString);
    Function GetPlanST : String;
    Procedure SetPlanST(value : String);
    Procedure SetSubPlan(value : TFhirString);
    Function GetSubPlanST : String;
    Procedure SetSubPlanST(value : String);
    Procedure SetDependent(value : TFhirPositiveInt);
    Function GetDependentST : String;
    Procedure SetDependentST(value : String);
    Procedure SetSequence(value : TFhirPositiveInt);
    Function GetSequenceST : String;
    Procedure SetSequenceST(value : String);
    Procedure SetSubscriber(value : TFhirReference{TFhirPatient});
    Procedure SetNetwork(value : TFhirIdentifier);
    function GetContractList : TFhirReferenceList{TFhirContract};
    function GetHasContractList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirCoverage; overload;
    function Clone : TFhirCoverage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member issuer
      Typed access to The program or plan underwriter or payor. (defined for API consistency)
    }
    property issuer : TFhirReference{TFhirOrganization} read FIssuer write SetIssuer;
    {@member issuerElement
      The program or plan underwriter or payor.
    }
    property issuerElement : TFhirReference{TFhirOrganization} read FIssuer write SetIssuer;

    {@member bin
      Typed access to Business Identification Number (BIN number) used to identify the routing  of eclaims if the insurer themselves don't have a BIN number for all of their business. (defined for API consistency)
    }
    property bin : TFhirIdentifier read FBin write SetBin;
    {@member binElement
      Business Identification Number (BIN number) used to identify the routing  of eclaims if the insurer themselves don't have a BIN number for all of their business.
    }
    property binElement : TFhirIdentifier read FBin write SetBin;

    {@member period
      Typed access to Time period during which the coverage is in force. A missing start date indicates the start date isn't known, a missing end date means the coverage is continuing to be in force. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      Time period during which the coverage is in force. A missing start date indicates the start date isn't known, a missing end date means the coverage is continuing to be in force.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member type_
      Typed access to The type of coverage: social program, medical plan, accident coverage (workers compensation, auto), group health. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of coverage: social program, medical plan, accident coverage (workers compensation, auto), group health.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member subscriberId
      Typed access to The id issued to the subscriber. (defined for API consistency)
    }
    property subscriberId : TFhirIdentifier read FSubscriberId write SetSubscriberId;
    {@member subscriberIdElement
      The id issued to the subscriber.
    }
    property subscriberIdElement : TFhirIdentifier read FSubscriberId write SetSubscriberId;

    {@member identifierList
      The main (and possibly only) identifier for the coverage - often referred to as a Member Id, Subscriber Id, Certificate number or Personal Health Number or Case ID.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member group
      Typed access to Identifies a style or collective of coverage issues by the underwriter, for example may be used to identify a class of coverage or employer group. May also be referred to as a Policy or Group ID.
    }
    property group : String read GetGroupST write SetGroupST;
    {@member groupElement
      Identifies a style or collective of coverage issues by the underwriter, for example may be used to identify a class of coverage or employer group. May also be referred to as a Policy or Group ID.
    }
    property groupElement : TFhirString read FGroup write SetGroup;

    {@member plan
      Typed access to Identifies a style or collective of coverage issues by the underwriter, for example may be used to identify a class of coverage or employer group. May also be referred to as a Policy or Group ID.
    }
    property plan : String read GetPlanST write SetPlanST;
    {@member planElement
      Identifies a style or collective of coverage issues by the underwriter, for example may be used to identify a class of coverage or employer group. May also be referred to as a Policy or Group ID.
    }
    property planElement : TFhirString read FPlan write SetPlan;

    {@member subPlan
      Typed access to Identifies a sub-style or sub-collective of coverage issues by the underwriter, for example may be used to identify a specific employer group within a class of employers. May be referred to as a Section or Division ID.
    }
    property subPlan : String read GetSubPlanST write SetSubPlanST;
    {@member subPlanElement
      Identifies a sub-style or sub-collective of coverage issues by the underwriter, for example may be used to identify a specific employer group within a class of employers. May be referred to as a Section or Division ID.
    }
    property subPlanElement : TFhirString read FSubPlan write SetSubPlan;

    {@member dependent
      Typed access to A unique identifier for a dependent under the coverage.
    }
    property dependent : String read GetDependentST write SetDependentST;
    {@member dependentElement
      A unique identifier for a dependent under the coverage.
    }
    property dependentElement : TFhirPositiveInt read FDependent write SetDependent;

    {@member sequence
      Typed access to An optional counter for a particular instance of the identified coverage which increments upon each renewal.
    }
    property sequence : String read GetSequenceST write SetSequenceST;
    {@member sequenceElement
      An optional counter for a particular instance of the identified coverage which increments upon each renewal.
    }
    property sequenceElement : TFhirPositiveInt read FSequence write SetSequence;

    {@member subscriber
      Typed access to The party who 'owns' the insurance contractual relationship to the policy or to whom the benefit of the policy is due. (defined for API consistency)
    }
    property subscriber : TFhirReference{TFhirPatient} read FSubscriber write SetSubscriber;
    {@member subscriberElement
      The party who 'owns' the insurance contractual relationship to the policy or to whom the benefit of the policy is due.
    }
    property subscriberElement : TFhirReference{TFhirPatient} read FSubscriber write SetSubscriber;

    {@member network
      Typed access to The identifier for a community of providers. (defined for API consistency)
    }
    property network : TFhirIdentifier read FNetwork write SetNetwork;
    {@member networkElement
      The identifier for a community of providers.
    }
    property networkElement : TFhirIdentifier read FNetwork write SetNetwork;

    {@member contractList
      The policy(s) which constitute this insurance coverage.
    }
    property contractList : TFhirReferenceList{TFhirContract} read GetContractList;
    property hasContractList : boolean read GetHasContractList;

  end;


  TFhirCoverageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirCoverageList;
    function GetCurrent : TFhirCoverage;
  public
    Constructor Create(list : TFhirCoverageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirCoverage read GetCurrent;
  end;


  {@Class TFhirCoverageList
    A list of FhirCoverage
  }
  {!.Net HL7Connect.Fhir.CoverageList}
  TFhirCoverageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirCoverage;
    procedure SetItemN(index : Integer; value : TFhirCoverage);
  public
    {!script hide}
    function Link : TFhirCoverageList; Overload;
    function Clone : TFhirCoverageList; Overload;
    function GetEnumerator : TFhirCoverageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirCoverage to the end of the list.
    }
    function Append : TFhirCoverage;

    
    {@member AddItem
      Add an already existing FhirCoverage to the end of the list.
    }
    procedure AddItem(value : TFhirCoverage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirCoverage) : Integer;
    

    {@member Insert
      Insert FhirCoverage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirCoverage;
    

    {@member InsertItem
       Insert an existing FhirCoverage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirCoverage);
    
    {@member Item
       Get the iIndexth FhirCoverage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirCoverage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirCoverage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirCoverage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirCoverages[index : Integer] : TFhirCoverage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDataElementContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.DataElementContact}
  TFhirDataElementContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDataElementContact; overload;
    function Clone : TFhirDataElementContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the data element.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the data element.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirDataElementContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDataElementContactList;
    function GetCurrent : TFhirDataElementContact;
  public
    Constructor Create(list : TFhirDataElementContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDataElementContact read GetCurrent;
  end;


  {@Class TFhirDataElementContactList
    A list of FhirDataElementContact
  }
  {!.Net HL7Connect.Fhir.DataElementContactList}
  TFhirDataElementContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDataElementContact;
    procedure SetItemN(index : Integer; value : TFhirDataElementContact);
  public
    {!script hide}
    function Link : TFhirDataElementContactList; Overload;
    function Clone : TFhirDataElementContactList; Overload;
    function GetEnumerator : TFhirDataElementContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDataElementContact to the end of the list.
    }
    function Append : TFhirDataElementContact;

    
    {@member AddItem
      Add an already existing FhirDataElementContact to the end of the list.
    }
    procedure AddItem(value : TFhirDataElementContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDataElementContact) : Integer;
    

    {@member Insert
      Insert FhirDataElementContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDataElementContact;
    

    {@member InsertItem
       Insert an existing FhirDataElementContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDataElementContact);
    
    {@member Item
       Get the iIndexth FhirDataElementContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDataElementContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDataElementContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDataElementContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDataElementContacts[index : Integer] : TFhirDataElementContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDataElementMapping : TFhirElement
    Identifies a specification (other than a terminology) that the elements which make up the DataElement have some correspondence with.
  }
  {!.Net HL7Connect.Fhir.DataElementMapping}
  TFhirDataElementMapping = class (TFhirBackboneElement)
  private
    FIdentity : TFhirId;
    FUri : TFhirUri;
    FName : TFhirString;
    FComments : TFhirString;
    Procedure SetIdentity(value : TFhirId);
    Function GetIdentityST : String;
    Procedure SetIdentityST(value : String);
    Procedure SetUri(value : TFhirUri);
    Function GetUriST : String;
    Procedure SetUriST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetComments(value : TFhirString);
    Function GetCommentsST : String;
    Procedure SetCommentsST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDataElementMapping; overload;
    function Clone : TFhirDataElementMapping; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identity
      Typed access to An internal id that is used to identify this mapping set when specific mappings are made on a per-element basis.
    }
    property identity : String read GetIdentityST write SetIdentityST;
    {@member identityElement
      An internal id that is used to identify this mapping set when specific mappings are made on a per-element basis.
    }
    property identityElement : TFhirId read FIdentity write SetIdentity;

    {@member uri
      Typed access to An absolute URI that identifies the specification that this mapping is expressed to.
    }
    property uri : String read GetUriST write SetUriST;
    {@member uriElement
      An absolute URI that identifies the specification that this mapping is expressed to.
    }
    property uriElement : TFhirUri read FUri write SetUri;

    {@member name
      Typed access to A name for the specification that is being mapped to.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A name for the specification that is being mapped to.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member comments
      Typed access to Comments about this mapping, including version notes, issues, scope limitations, and other important notes for usage.
    }
    property comments : String read GetCommentsST write SetCommentsST;
    {@member commentsElement
      Comments about this mapping, including version notes, issues, scope limitations, and other important notes for usage.
    }
    property commentsElement : TFhirString read FComments write SetComments;

  end;


  TFhirDataElementMappingListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDataElementMappingList;
    function GetCurrent : TFhirDataElementMapping;
  public
    Constructor Create(list : TFhirDataElementMappingList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDataElementMapping read GetCurrent;
  end;


  {@Class TFhirDataElementMappingList
    A list of FhirDataElementMapping
  }
  {!.Net HL7Connect.Fhir.DataElementMappingList}
  TFhirDataElementMappingList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDataElementMapping;
    procedure SetItemN(index : Integer; value : TFhirDataElementMapping);
  public
    {!script hide}
    function Link : TFhirDataElementMappingList; Overload;
    function Clone : TFhirDataElementMappingList; Overload;
    function GetEnumerator : TFhirDataElementMappingListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDataElementMapping to the end of the list.
    }
    function Append : TFhirDataElementMapping;

    
    {@member AddItem
      Add an already existing FhirDataElementMapping to the end of the list.
    }
    procedure AddItem(value : TFhirDataElementMapping); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDataElementMapping) : Integer;
    

    {@member Insert
      Insert FhirDataElementMapping before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDataElementMapping;
    

    {@member InsertItem
       Insert an existing FhirDataElementMapping before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDataElementMapping);
    
    {@member Item
       Get the iIndexth FhirDataElementMapping. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDataElementMapping. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDataElementMapping);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDataElementMapping;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDataElementMappings[index : Integer] : TFhirDataElementMapping read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDataElement : TFhirDomainResource
    The formal description of a single piece of information that can be gathered and reported.
  }
  {!.Net HL7Connect.Fhir.DataElement}
  TFhirDataElement = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FidentifierList : TFhirIdentifierList;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirDataElementContactList;
    FDate : TFhirDateTime;
    FuseContextList : TFhirCodeableConceptList;
    FCopyright : TFhirString;
    FStringency : TFhirEnum;
    FmappingList : TFhirDataElementMappingList;
    FelementList : TFhirElementDefinitionList;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirDataElementContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetStringency(value : TFhirEnum);
    Function GetStringencyST : TFhirDataelementStringency;
    Procedure SetStringencyST(value : TFhirDataelementStringency);
    function GetMappingList : TFhirDataElementMappingList;
    function GetHasMappingList : Boolean;
    function GetElementList : TFhirElementDefinitionList;
    function GetHasElementList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDataElement; overload;
    function Clone : TFhirDataElement; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this data element when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this data element is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this data element when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this data element is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member identifierList
      Formal identifier that is used to identify this data element when it is represented in other formats, or referenced in a specification, model, design or an instance.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member version
      Typed access to The identifier that is used to identify this version of the data element when it is referenced in a StructureDefinition, Questionnaire or instance. This is an arbitrary value managed by the definition author manually.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the data element when it is referenced in a StructureDefinition, Questionnaire or instance. This is an arbitrary value managed by the definition author manually.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to The term used by humans to refer to the data element.  Should ideally be unique within the context in which the data element is expected to be used.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The term used by humans to refer to the data element.  Should ideally be unique within the context in which the data element is expected to be used.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the data element.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to A flag to indicate that this search data element definition is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      A flag to indicate that this search data element definition is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the data element.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the data element.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirDataElementContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the data element was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the data element  changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the data element was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the data element  changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of data element definitions.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member copyright
      Typed access to A copyright statement relating to the definition of the data element. Copyright statements are generally legal restrictions on the use and publishing of the details of the definition of the data element.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the definition of the data element. Copyright statements are generally legal restrictions on the use and publishing of the details of the definition of the data element.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member stringency
      Identifies how precise the data element is in its definition.
    }
    property stringency : TFhirDataelementStringency read GetStringencyST write SetStringencyST;
    property stringencyElement : TFhirEnum read FStringency write SetStringency;

    {@member mappingList
      Identifies a specification (other than a terminology) that the elements which make up the DataElement have some correspondence with.
    }
    property mappingList : TFhirDataElementMappingList read GetMappingList;
    property hasMappingList : boolean read GetHasMappingList;

    {@member elementList
      Defines the structure, type, allowed values and other constraining characteristics of the data element.
    }
    property elementList : TFhirElementDefinitionList read GetElementList;
    property hasElementList : boolean read GetHasElementList;

  end;


  TFhirDataElementListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDataElementList;
    function GetCurrent : TFhirDataElement;
  public
    Constructor Create(list : TFhirDataElementList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDataElement read GetCurrent;
  end;


  {@Class TFhirDataElementList
    A list of FhirDataElement
  }
  {!.Net HL7Connect.Fhir.DataElementList}
  TFhirDataElementList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDataElement;
    procedure SetItemN(index : Integer; value : TFhirDataElement);
  public
    {!script hide}
    function Link : TFhirDataElementList; Overload;
    function Clone : TFhirDataElementList; Overload;
    function GetEnumerator : TFhirDataElementListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDataElement to the end of the list.
    }
    function Append : TFhirDataElement;

    
    {@member AddItem
      Add an already existing FhirDataElement to the end of the list.
    }
    procedure AddItem(value : TFhirDataElement); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDataElement) : Integer;
    

    {@member Insert
      Insert FhirDataElement before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDataElement;
    

    {@member InsertItem
       Insert an existing FhirDataElement before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDataElement);
    
    {@member Item
       Get the iIndexth FhirDataElement. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDataElement. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDataElement);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDataElement;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDataElements[index : Integer] : TFhirDataElement read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDetectedIssueMitigation : TFhirElement
    Indicates an action that has been taken or is committed to to reduce or eliminate the likelihood of the risk identified by the detected issue from manifesting.  Can also reflect an observation of known mitigating factors that may reduce/eliminate the need for any action.
  }
  {!.Net HL7Connect.Fhir.DetectedIssueMitigation}
  TFhirDetectedIssueMitigation = class (TFhirBackboneElement)
  private
    FAction : TFhirCodeableConcept;
    FDate : TFhirDateTime;
    FAuthor : TFhirReference{TFhirPractitioner};
    Procedure SetAction(value : TFhirCodeableConcept);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetAuthor(value : TFhirReference{TFhirPractitioner});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDetectedIssueMitigation; overload;
    function Clone : TFhirDetectedIssueMitigation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member action
      Typed access to Describes the action that was taken or the observation that was made that reduces/eliminates the risk associated with the identified issue. (defined for API consistency)
    }
    property action : TFhirCodeableConcept read FAction write SetAction;
    {@member actionElement
      Describes the action that was taken or the observation that was made that reduces/eliminates the risk associated with the identified issue.
    }
    property actionElement : TFhirCodeableConcept read FAction write SetAction;

    {@member date
      Typed access to Indicates when the mitigating action was documented.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      Indicates when the mitigating action was documented.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member author
      Typed access to Identifies the practitioner who determined the mitigation and takes responsibility for the mitigation step occurring. (defined for API consistency)
    }
    property author : TFhirReference{TFhirPractitioner} read FAuthor write SetAuthor;
    {@member authorElement
      Identifies the practitioner who determined the mitigation and takes responsibility for the mitigation step occurring.
    }
    property authorElement : TFhirReference{TFhirPractitioner} read FAuthor write SetAuthor;

  end;


  TFhirDetectedIssueMitigationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDetectedIssueMitigationList;
    function GetCurrent : TFhirDetectedIssueMitigation;
  public
    Constructor Create(list : TFhirDetectedIssueMitigationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDetectedIssueMitigation read GetCurrent;
  end;


  {@Class TFhirDetectedIssueMitigationList
    A list of FhirDetectedIssueMitigation
  }
  {!.Net HL7Connect.Fhir.DetectedIssueMitigationList}
  TFhirDetectedIssueMitigationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDetectedIssueMitigation;
    procedure SetItemN(index : Integer; value : TFhirDetectedIssueMitigation);
  public
    {!script hide}
    function Link : TFhirDetectedIssueMitigationList; Overload;
    function Clone : TFhirDetectedIssueMitigationList; Overload;
    function GetEnumerator : TFhirDetectedIssueMitigationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDetectedIssueMitigation to the end of the list.
    }
    function Append : TFhirDetectedIssueMitigation;

    
    {@member AddItem
      Add an already existing FhirDetectedIssueMitigation to the end of the list.
    }
    procedure AddItem(value : TFhirDetectedIssueMitigation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDetectedIssueMitigation) : Integer;
    

    {@member Insert
      Insert FhirDetectedIssueMitigation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDetectedIssueMitigation;
    

    {@member InsertItem
       Insert an existing FhirDetectedIssueMitigation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDetectedIssueMitigation);
    
    {@member Item
       Get the iIndexth FhirDetectedIssueMitigation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDetectedIssueMitigation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDetectedIssueMitigation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDetectedIssueMitigation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDetectedIssueMitigations[index : Integer] : TFhirDetectedIssueMitigation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDetectedIssue : TFhirDomainResource
    Indicates an actual or potential clinical issue with or between one or more active or proposed clinical actions for a patient; e.g. Drug-drug interaction, Ineffective treatment frequency, Procedure-condition conflict, etc.
  }
  {!.Net HL7Connect.Fhir.DetectedIssue}
  TFhirDetectedIssue = class (TFhirDomainResource)
  private
    FPatient : TFhirReference{TFhirPatient};
    FCategory : TFhirCodeableConcept;
    FSeverity : TFhirEnum;
    FimplicatedList : TFhirReferenceList{TFhirReference};
    FDetail : TFhirString;
    FDate : TFhirDateTime;
    FAuthor : TFhirReference{Resource};
    FIdentifier : TFhirIdentifier;
    FReference : TFhirUri;
    FmitigationList : TFhirDetectedIssueMitigationList;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetSeverity(value : TFhirEnum);
    Function GetSeverityST : TFhirDetectedissueSeverity;
    Procedure SetSeverityST(value : TFhirDetectedissueSeverity);
    function GetImplicatedList : TFhirReferenceList{TFhirReference};
    function GetHasImplicatedList : Boolean;
    Procedure SetDetail(value : TFhirString);
    Function GetDetailST : String;
    Procedure SetDetailST(value : String);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetReference(value : TFhirUri);
    Function GetReferenceST : String;
    Procedure SetReferenceST(value : String);
    function GetMitigationList : TFhirDetectedIssueMitigationList;
    function GetHasMitigationList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDetectedIssue; overload;
    function Clone : TFhirDetectedIssue; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member patient
      Typed access to Indicates the patient whose record the detected issue is associated with. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      Indicates the patient whose record the detected issue is associated with.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member category
      Typed access to Identifies the general type of issue identified. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      Identifies the general type of issue identified.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member severity
      Indicates the degree of importance associated with the identified issue based on the potential impact on the patient.
    }
    property severity : TFhirDetectedissueSeverity read GetSeverityST write SetSeverityST;
    property severityElement : TFhirEnum read FSeverity write SetSeverity;

    {@member implicatedList
      Indicates the resource representing the current activity or proposed activity that is potentially problematic.
    }
    property implicatedList : TFhirReferenceList{TFhirReference} read GetImplicatedList;
    property hasImplicatedList : boolean read GetHasImplicatedList;

    {@member detail
      Typed access to A textual explanation of the detected issue.
    }
    property detail : String read GetDetailST write SetDetailST;
    {@member detailElement
      A textual explanation of the detected issue.
    }
    property detailElement : TFhirString read FDetail write SetDetail;

    {@member date
      Typed access to The date or date-time when the detected issue was initially identified.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date or date-time when the detected issue was initially identified.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member author
      Typed access to Individual or device responsible for the issue being raised.  For example, a decision support application or a pharmacist conducting a medication review. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      Individual or device responsible for the issue being raised.  For example, a decision support application or a pharmacist conducting a medication review.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member identifier
      Typed access to Business identifier associated with the detected issue record. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Business identifier associated with the detected issue record.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member reference
      Typed access to The literature, knowledge-base or similar reference that describes the propensity for the detected issue identified.
    }
    property reference : String read GetReferenceST write SetReferenceST;
    {@member referenceElement
      The literature, knowledge-base or similar reference that describes the propensity for the detected issue identified.
    }
    property referenceElement : TFhirUri read FReference write SetReference;

    {@member mitigationList
      Indicates an action that has been taken or is committed to to reduce or eliminate the likelihood of the risk identified by the detected issue from manifesting.  Can also reflect an observation of known mitigating factors that may reduce/eliminate the need for any action.
    }
    property mitigationList : TFhirDetectedIssueMitigationList read GetMitigationList;
    property hasMitigationList : boolean read GetHasMitigationList;

  end;


  TFhirDetectedIssueListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDetectedIssueList;
    function GetCurrent : TFhirDetectedIssue;
  public
    Constructor Create(list : TFhirDetectedIssueList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDetectedIssue read GetCurrent;
  end;


  {@Class TFhirDetectedIssueList
    A list of FhirDetectedIssue
  }
  {!.Net HL7Connect.Fhir.DetectedIssueList}
  TFhirDetectedIssueList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDetectedIssue;
    procedure SetItemN(index : Integer; value : TFhirDetectedIssue);
  public
    {!script hide}
    function Link : TFhirDetectedIssueList; Overload;
    function Clone : TFhirDetectedIssueList; Overload;
    function GetEnumerator : TFhirDetectedIssueListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDetectedIssue to the end of the list.
    }
    function Append : TFhirDetectedIssue;

    
    {@member AddItem
      Add an already existing FhirDetectedIssue to the end of the list.
    }
    procedure AddItem(value : TFhirDetectedIssue); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDetectedIssue) : Integer;
    

    {@member Insert
      Insert FhirDetectedIssue before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDetectedIssue;
    

    {@member InsertItem
       Insert an existing FhirDetectedIssue before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDetectedIssue);
    
    {@member Item
       Get the iIndexth FhirDetectedIssue. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDetectedIssue. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDetectedIssue);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDetectedIssue;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDetectedIssues[index : Integer] : TFhirDetectedIssue read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDevice : TFhirDomainResource
    This resource identifies an instance of a manufactured item that is used in the provision of healthcare without being substantially changed through that activity. The device may be a medical or non-medical device.  Medical devices includes durable (reusable) medical equipment, implantable devices, as well as disposable equipment used for diagnostic, treatment, and research for healthcare and public health.  Non-medical devices may include items such as a machine, cellphone, computer, application, etc.
  }
  {!.Net HL7Connect.Fhir.Device}
  TFhirDevice = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FType_ : TFhirCodeableConcept;
    FnoteList : TFhirAnnotationList;
    FStatus : TFhirEnum;
    FManufacturer : TFhirString;
    FModel : TFhirString;
    FVersion : TFhirString;
    FManufactureDate : TFhirDateTime;
    FExpiry : TFhirDateTime;
    FUdi : TFhirString;
    FLotNumber : TFhirString;
    FOwner : TFhirReference{TFhirOrganization};
    FLocation : TFhirReference{TFhirLocation};
    FPatient : TFhirReference{TFhirPatient};
    FcontactList : TFhirContactPointList;
    FUrl : TFhirUri;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetNoteList : TFhirAnnotationList;
    function GetHasNoteList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDevicestatus;
    Procedure SetStatusST(value : TFhirDevicestatus);
    Procedure SetManufacturer(value : TFhirString);
    Function GetManufacturerST : String;
    Procedure SetManufacturerST(value : String);
    Procedure SetModel(value : TFhirString);
    Function GetModelST : String;
    Procedure SetModelST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetManufactureDate(value : TFhirDateTime);
    Function GetManufactureDateST : TDateAndTime;
    Procedure SetManufactureDateST(value : TDateAndTime);
    Procedure SetExpiry(value : TFhirDateTime);
    Function GetExpiryST : TDateAndTime;
    Procedure SetExpiryST(value : TDateAndTime);
    Procedure SetUdi(value : TFhirString);
    Function GetUdiST : String;
    Procedure SetUdiST(value : String);
    Procedure SetLotNumber(value : TFhirString);
    Function GetLotNumberST : String;
    Procedure SetLotNumberST(value : String);
    Procedure SetOwner(value : TFhirReference{TFhirOrganization});
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    function GetContactList : TFhirContactPointList;
    function GetHasContactList : Boolean;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDevice; overload;
    function Clone : TFhirDevice; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Unique instance identifiers assigned to a device by organizations like manufacturers or owners. If the identifier identifies the type of device, Device.type should be used.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member type_
      Typed access to Code or identifier to identify a kind of device. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Code or identifier to identify a kind of device.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member noteList
      Descriptive information, usage information or implantation information that is not captured in an existing element.
    }
    property noteList : TFhirAnnotationList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

    {@member status
      Status of the Device availability.
    }
    property status : TFhirDevicestatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member manufacturer
      Typed access to A name of the manufacturer.
    }
    property manufacturer : String read GetManufacturerST write SetManufacturerST;
    {@member manufacturerElement
      A name of the manufacturer.
    }
    property manufacturerElement : TFhirString read FManufacturer write SetManufacturer;

    {@member model
      Typed access to The "model" is an identifier assigned by the manufacturer to identify the product by its type. This number is shared by the all devices sold as the same type.
    }
    property model : String read GetModelST write SetModelST;
    {@member modelElement
      The "model" is an identifier assigned by the manufacturer to identify the product by its type. This number is shared by the all devices sold as the same type.
    }
    property modelElement : TFhirString read FModel write SetModel;

    {@member version
      Typed access to The version of the device, if the device has multiple releases under the same model, or if the device is software or carries firmware.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version of the device, if the device has multiple releases under the same model, or if the device is software or carries firmware.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member manufactureDate
      Typed access to The date and time when the device was manufactured.
    }
    property manufactureDate : TDateAndTime read GetManufactureDateST write SetManufactureDateST;
    {@member manufactureDateElement
      The date and time when the device was manufactured.
    }
    property manufactureDateElement : TFhirDateTime read FManufactureDate write SetManufactureDate;

    {@member expiry
      Typed access to The date and time beyond which this device is no longer valid or should not be used (if applicable).
    }
    property expiry : TDateAndTime read GetExpiryST write SetExpiryST;
    {@member expiryElement
      The date and time beyond which this device is no longer valid or should not be used (if applicable).
    }
    property expiryElement : TFhirDateTime read FExpiry write SetExpiry;

    {@member udi
      Typed access to United States Food and Drug Administration mandated Unique Device Identifier (UDI). Use the human readable information (the content that the user sees, which is sometimes different to the exact syntax represented in the barcode)  - see http://www.fda.gov/MedicalDevices/DeviceRegulationandGuidance/UniqueDeviceIdentification/default.htm.
    }
    property udi : String read GetUdiST write SetUdiST;
    {@member udiElement
      United States Food and Drug Administration mandated Unique Device Identifier (UDI). Use the human readable information (the content that the user sees, which is sometimes different to the exact syntax represented in the barcode)  - see http://www.fda.gov/MedicalDevices/DeviceRegulationandGuidance/UniqueDeviceIdentification/default.htm.
    }
    property udiElement : TFhirString read FUdi write SetUdi;

    {@member lotNumber
      Typed access to Lot number assigned by the manufacturer.
    }
    property lotNumber : String read GetLotNumberST write SetLotNumberST;
    {@member lotNumberElement
      Lot number assigned by the manufacturer.
    }
    property lotNumberElement : TFhirString read FLotNumber write SetLotNumber;

    {@member owner
      Typed access to An organization that is responsible for the provision and ongoing maintenance of the device. (defined for API consistency)
    }
    property owner : TFhirReference{TFhirOrganization} read FOwner write SetOwner;
    {@member ownerElement
      An organization that is responsible for the provision and ongoing maintenance of the device.
    }
    property ownerElement : TFhirReference{TFhirOrganization} read FOwner write SetOwner;

    {@member location
      Typed access to The place where the device can be found. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      The place where the device can be found.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member patient
      Typed access to Patient information, if the resource is affixed to a person. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      Patient information, if the resource is affixed to a person.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member contactList
      Contact details for an organization or a particular human that is responsible for the device.
    }
    property contactList : TFhirContactPointList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member url
      Typed access to A network address on which the device may be contacted directly.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      A network address on which the device may be contacted directly.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

  end;


  TFhirDeviceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceList;
    function GetCurrent : TFhirDevice;
  public
    Constructor Create(list : TFhirDeviceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDevice read GetCurrent;
  end;


  {@Class TFhirDeviceList
    A list of FhirDevice
  }
  {!.Net HL7Connect.Fhir.DeviceList}
  TFhirDeviceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDevice;
    procedure SetItemN(index : Integer; value : TFhirDevice);
  public
    {!script hide}
    function Link : TFhirDeviceList; Overload;
    function Clone : TFhirDeviceList; Overload;
    function GetEnumerator : TFhirDeviceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDevice to the end of the list.
    }
    function Append : TFhirDevice;

    
    {@member AddItem
      Add an already existing FhirDevice to the end of the list.
    }
    procedure AddItem(value : TFhirDevice); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDevice) : Integer;
    

    {@member Insert
      Insert FhirDevice before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDevice;
    

    {@member InsertItem
       Insert an existing FhirDevice before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDevice);
    
    {@member Item
       Get the iIndexth FhirDevice. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDevice. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDevice);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDevice;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDevices[index : Integer] : TFhirDevice read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceComponentProductionSpecification : TFhirElement
    Describes the production specification such as component revision, serial number, etc.
  }
  {!.Net HL7Connect.Fhir.DeviceComponentProductionSpecification}
  TFhirDeviceComponentProductionSpecification = class (TFhirBackboneElement)
  private
    FSpecType : TFhirCodeableConcept;
    FComponentId : TFhirIdentifier;
    FProductionSpec : TFhirString;
    Procedure SetSpecType(value : TFhirCodeableConcept);
    Procedure SetComponentId(value : TFhirIdentifier);
    Procedure SetProductionSpec(value : TFhirString);
    Function GetProductionSpecST : String;
    Procedure SetProductionSpecST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceComponentProductionSpecification; overload;
    function Clone : TFhirDeviceComponentProductionSpecification; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member specType
      Typed access to Describes the specification type, such as, serial number, part number, hardware revision, software revision, etc. (defined for API consistency)
    }
    property specType : TFhirCodeableConcept read FSpecType write SetSpecType;
    {@member specTypeElement
      Describes the specification type, such as, serial number, part number, hardware revision, software revision, etc.
    }
    property specTypeElement : TFhirCodeableConcept read FSpecType write SetSpecType;

    {@member componentId
      Typed access to Describes the internal component unique identification. This is a provision for manufacture specific standard components using a private OID. 11073-10101 has a partition for private OID semantic that the manufacture can make use of. (defined for API consistency)
    }
    property componentId : TFhirIdentifier read FComponentId write SetComponentId;
    {@member componentIdElement
      Describes the internal component unique identification. This is a provision for manufacture specific standard components using a private OID. 11073-10101 has a partition for private OID semantic that the manufacture can make use of.
    }
    property componentIdElement : TFhirIdentifier read FComponentId write SetComponentId;

    {@member productionSpec
      Typed access to Describes the printable string defining the component.
    }
    property productionSpec : String read GetProductionSpecST write SetProductionSpecST;
    {@member productionSpecElement
      Describes the printable string defining the component.
    }
    property productionSpecElement : TFhirString read FProductionSpec write SetProductionSpec;

  end;


  TFhirDeviceComponentProductionSpecificationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceComponentProductionSpecificationList;
    function GetCurrent : TFhirDeviceComponentProductionSpecification;
  public
    Constructor Create(list : TFhirDeviceComponentProductionSpecificationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceComponentProductionSpecification read GetCurrent;
  end;


  {@Class TFhirDeviceComponentProductionSpecificationList
    A list of FhirDeviceComponentProductionSpecification
  }
  {!.Net HL7Connect.Fhir.DeviceComponentProductionSpecificationList}
  TFhirDeviceComponentProductionSpecificationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceComponentProductionSpecification;
    procedure SetItemN(index : Integer; value : TFhirDeviceComponentProductionSpecification);
  public
    {!script hide}
    function Link : TFhirDeviceComponentProductionSpecificationList; Overload;
    function Clone : TFhirDeviceComponentProductionSpecificationList; Overload;
    function GetEnumerator : TFhirDeviceComponentProductionSpecificationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceComponentProductionSpecification to the end of the list.
    }
    function Append : TFhirDeviceComponentProductionSpecification;

    
    {@member AddItem
      Add an already existing FhirDeviceComponentProductionSpecification to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceComponentProductionSpecification); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceComponentProductionSpecification) : Integer;
    

    {@member Insert
      Insert FhirDeviceComponentProductionSpecification before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceComponentProductionSpecification;
    

    {@member InsertItem
       Insert an existing FhirDeviceComponentProductionSpecification before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceComponentProductionSpecification);
    
    {@member Item
       Get the iIndexth FhirDeviceComponentProductionSpecification. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceComponentProductionSpecification. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceComponentProductionSpecification);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceComponentProductionSpecification;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceComponentProductionSpecifications[index : Integer] : TFhirDeviceComponentProductionSpecification read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceComponent : TFhirDomainResource
    Describes the characteristics, operational status and capabilities of a medical-related component of a medical device.
  }
  {!.Net HL7Connect.Fhir.DeviceComponent}
  TFhirDeviceComponent = class (TFhirDomainResource)
  private
    FType_ : TFhirCodeableConcept;
    FIdentifier : TFhirIdentifier;
    FLastSystemChange : TFhirInstant;
    FSource : TFhirReference{TFhirDevice};
    FParent : TFhirReference{TFhirDeviceComponent};
    FoperationalStatusList : TFhirCodeableConceptList;
    FParameterGroup : TFhirCodeableConcept;
    FMeasurementPrinciple : TFhirEnum;
    FproductionSpecificationList : TFhirDeviceComponentProductionSpecificationList;
    FLanguageCode : TFhirCodeableConcept;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetLastSystemChange(value : TFhirInstant);
    Function GetLastSystemChangeST : TDateAndTime;
    Procedure SetLastSystemChangeST(value : TDateAndTime);
    Procedure SetSource(value : TFhirReference{TFhirDevice});
    Procedure SetParent(value : TFhirReference{TFhirDeviceComponent});
    function GetOperationalStatusList : TFhirCodeableConceptList;
    function GetHasOperationalStatusList : Boolean;
    Procedure SetParameterGroup(value : TFhirCodeableConcept);
    Procedure SetMeasurementPrinciple(value : TFhirEnum);
    Function GetMeasurementPrincipleST : TFhirMeasurementPrinciple;
    Procedure SetMeasurementPrincipleST(value : TFhirMeasurementPrinciple);
    function GetProductionSpecificationList : TFhirDeviceComponentProductionSpecificationList;
    function GetHasProductionSpecificationList : Boolean;
    Procedure SetLanguageCode(value : TFhirCodeableConcept);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceComponent; overload;
    function Clone : TFhirDeviceComponent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member type_
      Typed access to Describes the specific component type as defined in the object-oriented or metric nomenclature partition. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Describes the specific component type as defined in the object-oriented or metric nomenclature partition.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member identifier
      Typed access to Describes the local assigned unique identification by the software. For example: handle ID. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Describes the local assigned unique identification by the software. For example: handle ID.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member lastSystemChange
      Typed access to Describes the timestamp for the most recent system change which includes device configuration or setting change.
    }
    property lastSystemChange : TDateAndTime read GetLastSystemChangeST write SetLastSystemChangeST;
    {@member lastSystemChangeElement
      Describes the timestamp for the most recent system change which includes device configuration or setting change.
    }
    property lastSystemChangeElement : TFhirInstant read FLastSystemChange write SetLastSystemChange;

    {@member source
      Typed access to Describes the link to the source Device that contains administrative device information such as manufacture, serial number, etc. (defined for API consistency)
    }
    property source : TFhirReference{TFhirDevice} read FSource write SetSource;
    {@member sourceElement
      Describes the link to the source Device that contains administrative device information such as manufacture, serial number, etc.
    }
    property sourceElement : TFhirReference{TFhirDevice} read FSource write SetSource;

    {@member parent
      Typed access to Describes the link to the parent resource. For example: Channel is linked to its VMD parent. (defined for API consistency)
    }
    property parent : TFhirReference{TFhirDeviceComponent} read FParent write SetParent;
    {@member parentElement
      Describes the link to the parent resource. For example: Channel is linked to its VMD parent.
    }
    property parentElement : TFhirReference{TFhirDeviceComponent} read FParent write SetParent;

    {@member operationalStatusList
      Indicates current operational status of the device. For example: On, Off, Standby, etc.
    }
    property operationalStatusList : TFhirCodeableConceptList read GetOperationalStatusList;
    property hasOperationalStatusList : boolean read GetHasOperationalStatusList;

    {@member parameterGroup
      Typed access to Describes the parameter group supported by the current device component that is based on some nomenclature, e.g. cardiovascular. (defined for API consistency)
    }
    property parameterGroup : TFhirCodeableConcept read FParameterGroup write SetParameterGroup;
    {@member parameterGroupElement
      Describes the parameter group supported by the current device component that is based on some nomenclature, e.g. cardiovascular.
    }
    property parameterGroupElement : TFhirCodeableConcept read FParameterGroup write SetParameterGroup;

    {@member measurementPrinciple
      Describes the physical principle of the measurement. For example: thermal, chemical, acoustical, etc.
    }
    property measurementPrinciple : TFhirMeasurementPrinciple read GetMeasurementPrincipleST write SetMeasurementPrincipleST;
    property measurementPrincipleElement : TFhirEnum read FMeasurementPrinciple write SetMeasurementPrinciple;

    {@member productionSpecificationList
      Describes the production specification such as component revision, serial number, etc.
    }
    property productionSpecificationList : TFhirDeviceComponentProductionSpecificationList read GetProductionSpecificationList;
    property hasProductionSpecificationList : boolean read GetHasProductionSpecificationList;

    {@member languageCode
      Typed access to Describes the language code for the human-readable text string produced by the device. This language code will follow the IETF language tag. Example: en-US. (defined for API consistency)
    }
    property languageCode : TFhirCodeableConcept read FLanguageCode write SetLanguageCode;
    {@member languageCodeElement
      Describes the language code for the human-readable text string produced by the device. This language code will follow the IETF language tag. Example: en-US.
    }
    property languageCodeElement : TFhirCodeableConcept read FLanguageCode write SetLanguageCode;

  end;


  TFhirDeviceComponentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceComponentList;
    function GetCurrent : TFhirDeviceComponent;
  public
    Constructor Create(list : TFhirDeviceComponentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceComponent read GetCurrent;
  end;


  {@Class TFhirDeviceComponentList
    A list of FhirDeviceComponent
  }
  {!.Net HL7Connect.Fhir.DeviceComponentList}
  TFhirDeviceComponentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceComponent;
    procedure SetItemN(index : Integer; value : TFhirDeviceComponent);
  public
    {!script hide}
    function Link : TFhirDeviceComponentList; Overload;
    function Clone : TFhirDeviceComponentList; Overload;
    function GetEnumerator : TFhirDeviceComponentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceComponent to the end of the list.
    }
    function Append : TFhirDeviceComponent;

    
    {@member AddItem
      Add an already existing FhirDeviceComponent to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceComponent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceComponent) : Integer;
    

    {@member Insert
      Insert FhirDeviceComponent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceComponent;
    

    {@member InsertItem
       Insert an existing FhirDeviceComponent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceComponent);
    
    {@member Item
       Get the iIndexth FhirDeviceComponent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceComponent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceComponent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceComponent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceComponents[index : Integer] : TFhirDeviceComponent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceMetricCalibration : TFhirElement
    Describes the calibrations that have been performed or that are required to be performed.
  }
  {!.Net HL7Connect.Fhir.DeviceMetricCalibration}
  TFhirDeviceMetricCalibration = class (TFhirBackboneElement)
  private
    FType_ : TFhirEnum;
    FState : TFhirEnum;
    FTime : TFhirInstant;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirMetricCalibrationType;
    Procedure SetType_ST(value : TFhirMetricCalibrationType);
    Procedure SetState(value : TFhirEnum);
    Function GetStateST : TFhirMetricCalibrationState;
    Procedure SetStateST(value : TFhirMetricCalibrationState);
    Procedure SetTime(value : TFhirInstant);
    Function GetTimeST : TDateAndTime;
    Procedure SetTimeST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceMetricCalibration; overload;
    function Clone : TFhirDeviceMetricCalibration; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Describes the type of the calibration method.
    }
    property type_ : TFhirMetricCalibrationType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member state
      Describes the state of the calibration.
    }
    property state : TFhirMetricCalibrationState read GetStateST write SetStateST;
    property stateElement : TFhirEnum read FState write SetState;

    {@member time
      Typed access to Describes the time last calibration has been performed.
    }
    property time : TDateAndTime read GetTimeST write SetTimeST;
    {@member timeElement
      Describes the time last calibration has been performed.
    }
    property timeElement : TFhirInstant read FTime write SetTime;

  end;


  TFhirDeviceMetricCalibrationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceMetricCalibrationList;
    function GetCurrent : TFhirDeviceMetricCalibration;
  public
    Constructor Create(list : TFhirDeviceMetricCalibrationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceMetricCalibration read GetCurrent;
  end;


  {@Class TFhirDeviceMetricCalibrationList
    A list of FhirDeviceMetricCalibration
  }
  {!.Net HL7Connect.Fhir.DeviceMetricCalibrationList}
  TFhirDeviceMetricCalibrationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceMetricCalibration;
    procedure SetItemN(index : Integer; value : TFhirDeviceMetricCalibration);
  public
    {!script hide}
    function Link : TFhirDeviceMetricCalibrationList; Overload;
    function Clone : TFhirDeviceMetricCalibrationList; Overload;
    function GetEnumerator : TFhirDeviceMetricCalibrationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceMetricCalibration to the end of the list.
    }
    function Append : TFhirDeviceMetricCalibration;

    
    {@member AddItem
      Add an already existing FhirDeviceMetricCalibration to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceMetricCalibration); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceMetricCalibration) : Integer;
    

    {@member Insert
      Insert FhirDeviceMetricCalibration before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceMetricCalibration;
    

    {@member InsertItem
       Insert an existing FhirDeviceMetricCalibration before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceMetricCalibration);
    
    {@member Item
       Get the iIndexth FhirDeviceMetricCalibration. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceMetricCalibration. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceMetricCalibration);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceMetricCalibration;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceMetricCalibrations[index : Integer] : TFhirDeviceMetricCalibration read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceMetric : TFhirDomainResource
    Describes a measurement, calculation or setting capability of a medical device.
  }
  {!.Net HL7Connect.Fhir.DeviceMetric}
  TFhirDeviceMetric = class (TFhirDomainResource)
  private
    FType_ : TFhirCodeableConcept;
    FIdentifier : TFhirIdentifier;
    FUnit_ : TFhirCodeableConcept;
    FSource : TFhirReference{TFhirDevice};
    FParent : TFhirReference{TFhirDeviceComponent};
    FOperationalStatus : TFhirEnum;
    FColor : TFhirEnum;
    FCategory : TFhirEnum;
    FMeasurementPeriod : TFhirTiming;
    FcalibrationList : TFhirDeviceMetricCalibrationList;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetUnit_(value : TFhirCodeableConcept);
    Procedure SetSource(value : TFhirReference{TFhirDevice});
    Procedure SetParent(value : TFhirReference{TFhirDeviceComponent});
    Procedure SetOperationalStatus(value : TFhirEnum);
    Function GetOperationalStatusST : TFhirMetricOperationalStatus;
    Procedure SetOperationalStatusST(value : TFhirMetricOperationalStatus);
    Procedure SetColor(value : TFhirEnum);
    Function GetColorST : TFhirMetricColor;
    Procedure SetColorST(value : TFhirMetricColor);
    Procedure SetCategory(value : TFhirEnum);
    Function GetCategoryST : TFhirMetricCategory;
    Procedure SetCategoryST(value : TFhirMetricCategory);
    Procedure SetMeasurementPeriod(value : TFhirTiming);
    function GetCalibrationList : TFhirDeviceMetricCalibrationList;
    function GetHasCalibrationList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceMetric; overload;
    function Clone : TFhirDeviceMetric; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member type_
      Typed access to Describes the type of the metric. For example: Heart Rate, PEEP Setting, etc. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Describes the type of the metric. For example: Heart Rate, PEEP Setting, etc.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member identifier
      Typed access to Describes the unique identification of this metric that has been assigned by the device or gateway software. For example: handle ID.  It should be noted that in order to make the identifier unique, the system element of the identifier should be set to the unique identifier of the device. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Describes the unique identification of this metric that has been assigned by the device or gateway software. For example: handle ID.  It should be noted that in order to make the identifier unique, the system element of the identifier should be set to the unique identifier of the device.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member unit_
      Typed access to Describes the unit that an observed value determined for this metric will have. For example: Percent, Seconds, etc. (defined for API consistency)
    }
    property unit_ : TFhirCodeableConcept read FUnit_ write SetUnit_;
    {@member unit_Element
      Describes the unit that an observed value determined for this metric will have. For example: Percent, Seconds, etc.
    }
    property unit_Element : TFhirCodeableConcept read FUnit_ write SetUnit_;

    {@member source
      Typed access to Describes the link to the  Device that this DeviceMetric belongs to and that contains administrative device information such as manufacture, serial number, etc. (defined for API consistency)
    }
    property source : TFhirReference{TFhirDevice} read FSource write SetSource;
    {@member sourceElement
      Describes the link to the  Device that this DeviceMetric belongs to and that contains administrative device information such as manufacture, serial number, etc.
    }
    property sourceElement : TFhirReference{TFhirDevice} read FSource write SetSource;

    {@member parent
      Typed access to Describes the link to the  DeviceComponent that this DeviceMetric belongs to and that provide information about the location of this DeviceMetric in the containment structure of the parent Device. An example would be a DeviceComponent that represents a Channel. This reference can be used by a client application to distinguish DeviceMetrics that have the same type, but should be interpreted based on their containment location. (defined for API consistency)
    }
    property parent : TFhirReference{TFhirDeviceComponent} read FParent write SetParent;
    {@member parentElement
      Describes the link to the  DeviceComponent that this DeviceMetric belongs to and that provide information about the location of this DeviceMetric in the containment structure of the parent Device. An example would be a DeviceComponent that represents a Channel. This reference can be used by a client application to distinguish DeviceMetrics that have the same type, but should be interpreted based on their containment location.
    }
    property parentElement : TFhirReference{TFhirDeviceComponent} read FParent write SetParent;

    {@member operationalStatus
      Indicates current operational state of the device. For example: On, Off, Standby, etc.
    }
    property operationalStatus : TFhirMetricOperationalStatus read GetOperationalStatusST write SetOperationalStatusST;
    property operationalStatusElement : TFhirEnum read FOperationalStatus write SetOperationalStatus;

    {@member color
      Describes the color representation for the metric. This is often used to aid clinicians to track and identify parameter types by color. In practice, consider a Patient Monitor that has ECG/HR and Pleth for example; the parameters are displayed in different characteristic colors, such as HR-blue, BP-green, and PR and SpO2- magenta.
    }
    property color : TFhirMetricColor read GetColorST write SetColorST;
    property colorElement : TFhirEnum read FColor write SetColor;

    {@member category
      Indicates the category of the observation generation process. A DeviceMetric can be for example a setting, measurement, or calculation.
    }
    property category : TFhirMetricCategory read GetCategoryST write SetCategoryST;
    property categoryElement : TFhirEnum read FCategory write SetCategory;

    {@member measurementPeriod
      Typed access to Describes the measurement repetition time. This is not necessarily the same as the update period. The measurement repetition time can range from milliseconds up to hours. An example for a measurement repetition time in the range of milliseconds is the sampling rate of an ECG. An example for a measurement repetition time in the range of hours is a NIBP that is triggered automatically every hour. The update period may be different than the measurement repetition time, if the device does not update the published observed value with the same frequency as it was measured. (defined for API consistency)
    }
    property measurementPeriod : TFhirTiming read FMeasurementPeriod write SetMeasurementPeriod;
    {@member measurementPeriodElement
      Describes the measurement repetition time. This is not necessarily the same as the update period. The measurement repetition time can range from milliseconds up to hours. An example for a measurement repetition time in the range of milliseconds is the sampling rate of an ECG. An example for a measurement repetition time in the range of hours is a NIBP that is triggered automatically every hour. The update period may be different than the measurement repetition time, if the device does not update the published observed value with the same frequency as it was measured.
    }
    property measurementPeriodElement : TFhirTiming read FMeasurementPeriod write SetMeasurementPeriod;

    {@member calibrationList
      Describes the calibrations that have been performed or that are required to be performed.
    }
    property calibrationList : TFhirDeviceMetricCalibrationList read GetCalibrationList;
    property hasCalibrationList : boolean read GetHasCalibrationList;

  end;


  TFhirDeviceMetricListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceMetricList;
    function GetCurrent : TFhirDeviceMetric;
  public
    Constructor Create(list : TFhirDeviceMetricList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceMetric read GetCurrent;
  end;


  {@Class TFhirDeviceMetricList
    A list of FhirDeviceMetric
  }
  {!.Net HL7Connect.Fhir.DeviceMetricList}
  TFhirDeviceMetricList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceMetric;
    procedure SetItemN(index : Integer; value : TFhirDeviceMetric);
  public
    {!script hide}
    function Link : TFhirDeviceMetricList; Overload;
    function Clone : TFhirDeviceMetricList; Overload;
    function GetEnumerator : TFhirDeviceMetricListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceMetric to the end of the list.
    }
    function Append : TFhirDeviceMetric;

    
    {@member AddItem
      Add an already existing FhirDeviceMetric to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceMetric); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceMetric) : Integer;
    

    {@member Insert
      Insert FhirDeviceMetric before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceMetric;
    

    {@member InsertItem
       Insert an existing FhirDeviceMetric before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceMetric);
    
    {@member Item
       Get the iIndexth FhirDeviceMetric. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceMetric. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceMetric);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceMetric;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceMetrics[index : Integer] : TFhirDeviceMetric read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceUseRequest : TFhirDomainResource
    Represents a request for a patient to employ a medical device. The device may be an implantable device, or an external assistive device, such as a walker.
  }
  {!.Net HL7Connect.Fhir.DeviceUseRequest}
  TFhirDeviceUseRequest = class (TFhirDomainResource)
  private
    FBodySite : TFhirType;
    FStatus : TFhirEnum;
    FDevice : TFhirReference{TFhirDevice};
    FEncounter : TFhirReference{TFhirEncounter};
    FidentifierList : TFhirIdentifierList;
    FindicationList : TFhirCodeableConceptList;
    FnotesList : TFhirStringList;
    FprnReasonList : TFhirCodeableConceptList;
    FOrderedOn : TFhirDateTime;
    FRecordedOn : TFhirDateTime;
    FSubject : TFhirReference{TFhirPatient};
    FTiming : TFhirType;
    FPriority : TFhirEnum;
    Procedure SetBodySite(value : TFhirType);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDeviceUseRequestStatus;
    Procedure SetStatusST(value : TFhirDeviceUseRequestStatus);
    Procedure SetDevice(value : TFhirReference{TFhirDevice});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetIndicationList : TFhirCodeableConceptList;
    function GetHasIndicationList : Boolean;
    function GetNotesList : TFhirStringList;
    function GetHasNotesList : Boolean;
    function GetPrnReasonList : TFhirCodeableConceptList;
    function GetHasPrnReasonList : Boolean;
    Procedure SetOrderedOn(value : TFhirDateTime);
    Function GetOrderedOnST : TDateAndTime;
    Procedure SetOrderedOnST(value : TDateAndTime);
    Procedure SetRecordedOn(value : TFhirDateTime);
    Function GetRecordedOnST : TDateAndTime;
    Procedure SetRecordedOnST(value : TDateAndTime);
    Procedure SetSubject(value : TFhirReference{TFhirPatient});
    Procedure SetTiming(value : TFhirType);
    Procedure SetPriority(value : TFhirEnum);
    Function GetPriorityST : TFhirDeviceUseRequestPriority;
    Procedure SetPriorityST(value : TFhirDeviceUseRequestPriority);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceUseRequest; overload;
    function Clone : TFhirDeviceUseRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member bodySite
      Typed access to Indicates the site on the subject's body where the device should be used ( i.e. the target site). (defined for API consistency)
    }
    property bodySite : TFhirType read FBodySite write SetBodySite;
    {@member bodySiteElement
      Indicates the site on the subject's body where the device should be used ( i.e. the target site).
    }
    property bodySiteElement : TFhirType read FBodySite write SetBodySite;

    {@member status
      The status of the request.
    }
    property status : TFhirDeviceUseRequestStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member device
      Typed access to The details of the device  to be used. (defined for API consistency)
    }
    property device : TFhirReference{TFhirDevice} read FDevice write SetDevice;
    {@member deviceElement
      The details of the device  to be used.
    }
    property deviceElement : TFhirReference{TFhirDevice} read FDevice write SetDevice;

    {@member encounter
      Typed access to An encounter that provides additional context in which this request is made. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      An encounter that provides additional context in which this request is made.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member identifierList
      Identifiers assigned to this order by the orderer or by the receiver.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member indicationList
      Reason or justification for the use of this device.
    }
    property indicationList : TFhirCodeableConceptList read GetIndicationList;
    property hasIndicationList : boolean read GetHasIndicationList;

    {@member notesList
      Details about this request that were not represented at all or sufficiently in one of the attributes provided in a class. These may include for example a comment, an instruction, or a note associated with the statement.
    }
    property notesList : TFhirStringList read GetNotesList;
    property hasNotesList : boolean read GetHasNotesList;

    {@member prnReasonList
      The proposed act must be performed if the indicated conditions occur, e.g.., shortness of breath, SpO2 less than x%.
    }
    property prnReasonList : TFhirCodeableConceptList read GetPrnReasonList;
    property hasPrnReasonList : boolean read GetHasPrnReasonList;

    {@member orderedOn
      Typed access to The time when the request was made.
    }
    property orderedOn : TDateAndTime read GetOrderedOnST write SetOrderedOnST;
    {@member orderedOnElement
      The time when the request was made.
    }
    property orderedOnElement : TFhirDateTime read FOrderedOn write SetOrderedOn;

    {@member recordedOn
      Typed access to The time at which the request was made/recorded.
    }
    property recordedOn : TDateAndTime read GetRecordedOnST write SetRecordedOnST;
    {@member recordedOnElement
      The time at which the request was made/recorded.
    }
    property recordedOnElement : TFhirDateTime read FRecordedOn write SetRecordedOn;

    {@member subject
      Typed access to The patient who will use the device. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirPatient} read FSubject write SetSubject;
    {@member subjectElement
      The patient who will use the device.
    }
    property subjectElement : TFhirReference{TFhirPatient} read FSubject write SetSubject;

    {@member timing
      Typed access to The timing schedule for the use of the device The Schedule data type allows many different expressions, for example. "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013". (defined for API consistency)
    }
    property timing : TFhirType read FTiming write SetTiming;
    {@member timingElement
      The timing schedule for the use of the device The Schedule data type allows many different expressions, for example. "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013".
    }
    property timingElement : TFhirType read FTiming write SetTiming;

    {@member priority
      Characterizes how quickly the  use of device must be initiated. Includes concepts such as stat, urgent, routine.
    }
    property priority : TFhirDeviceUseRequestPriority read GetPriorityST write SetPriorityST;
    property priorityElement : TFhirEnum read FPriority write SetPriority;

  end;


  TFhirDeviceUseRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceUseRequestList;
    function GetCurrent : TFhirDeviceUseRequest;
  public
    Constructor Create(list : TFhirDeviceUseRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceUseRequest read GetCurrent;
  end;


  {@Class TFhirDeviceUseRequestList
    A list of FhirDeviceUseRequest
  }
  {!.Net HL7Connect.Fhir.DeviceUseRequestList}
  TFhirDeviceUseRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceUseRequest;
    procedure SetItemN(index : Integer; value : TFhirDeviceUseRequest);
  public
    {!script hide}
    function Link : TFhirDeviceUseRequestList; Overload;
    function Clone : TFhirDeviceUseRequestList; Overload;
    function GetEnumerator : TFhirDeviceUseRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceUseRequest to the end of the list.
    }
    function Append : TFhirDeviceUseRequest;

    
    {@member AddItem
      Add an already existing FhirDeviceUseRequest to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceUseRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceUseRequest) : Integer;
    

    {@member Insert
      Insert FhirDeviceUseRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceUseRequest;
    

    {@member InsertItem
       Insert an existing FhirDeviceUseRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceUseRequest);
    
    {@member Item
       Get the iIndexth FhirDeviceUseRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceUseRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceUseRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceUseRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceUseRequests[index : Integer] : TFhirDeviceUseRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDeviceUseStatement : TFhirDomainResource
    A record of a device being used by a patient where the record is the result of a report from the patient or another clinician.
  }
  {!.Net HL7Connect.Fhir.DeviceUseStatement}
  TFhirDeviceUseStatement = class (TFhirDomainResource)
  private
    FBodySite : TFhirType;
    FWhenUsed : TFhirPeriod;
    FDevice : TFhirReference{TFhirDevice};
    FidentifierList : TFhirIdentifierList;
    FindicationList : TFhirCodeableConceptList;
    FnotesList : TFhirStringList;
    FRecordedOn : TFhirDateTime;
    FSubject : TFhirReference{TFhirPatient};
    FTiming : TFhirType;
    Procedure SetBodySite(value : TFhirType);
    Procedure SetWhenUsed(value : TFhirPeriod);
    Procedure SetDevice(value : TFhirReference{TFhirDevice});
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetIndicationList : TFhirCodeableConceptList;
    function GetHasIndicationList : Boolean;
    function GetNotesList : TFhirStringList;
    function GetHasNotesList : Boolean;
    Procedure SetRecordedOn(value : TFhirDateTime);
    Function GetRecordedOnST : TDateAndTime;
    Procedure SetRecordedOnST(value : TDateAndTime);
    Procedure SetSubject(value : TFhirReference{TFhirPatient});
    Procedure SetTiming(value : TFhirType);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDeviceUseStatement; overload;
    function Clone : TFhirDeviceUseStatement; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member bodySite
      Typed access to Indicates the site on the subject's body where the device was used ( i.e. the target site). (defined for API consistency)
    }
    property bodySite : TFhirType read FBodySite write SetBodySite;
    {@member bodySiteElement
      Indicates the site on the subject's body where the device was used ( i.e. the target site).
    }
    property bodySiteElement : TFhirType read FBodySite write SetBodySite;

    {@member whenUsed
      Typed access to The time period over which the device was used. (defined for API consistency)
    }
    property whenUsed : TFhirPeriod read FWhenUsed write SetWhenUsed;
    {@member whenUsedElement
      The time period over which the device was used.
    }
    property whenUsedElement : TFhirPeriod read FWhenUsed write SetWhenUsed;

    {@member device
      Typed access to The details of the device used. (defined for API consistency)
    }
    property device : TFhirReference{TFhirDevice} read FDevice write SetDevice;
    {@member deviceElement
      The details of the device used.
    }
    property deviceElement : TFhirReference{TFhirDevice} read FDevice write SetDevice;

    {@member identifierList
      An external identifier for this statement such as an IRI.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member indicationList
      Reason or justification for the use of the device.
    }
    property indicationList : TFhirCodeableConceptList read GetIndicationList;
    property hasIndicationList : boolean read GetHasIndicationList;

    {@member notesList
      Details about the device statement that were not represented at all or sufficiently in one of the attributes provided in a class. These may include for example a comment, an instruction, or a note associated with the statement.
    }
    property notesList : TFhirStringList read GetNotesList;
    property hasNotesList : boolean read GetHasNotesList;

    {@member recordedOn
      Typed access to The time at which the statement was made/recorded.
    }
    property recordedOn : TDateAndTime read GetRecordedOnST write SetRecordedOnST;
    {@member recordedOnElement
      The time at which the statement was made/recorded.
    }
    property recordedOnElement : TFhirDateTime read FRecordedOn write SetRecordedOn;

    {@member subject
      Typed access to The patient who used the device. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirPatient} read FSubject write SetSubject;
    {@member subjectElement
      The patient who used the device.
    }
    property subjectElement : TFhirReference{TFhirPatient} read FSubject write SetSubject;

    {@member timing
      Typed access to How often the device was used. (defined for API consistency)
    }
    property timing : TFhirType read FTiming write SetTiming;
    {@member timingElement
      How often the device was used.
    }
    property timingElement : TFhirType read FTiming write SetTiming;

  end;


  TFhirDeviceUseStatementListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDeviceUseStatementList;
    function GetCurrent : TFhirDeviceUseStatement;
  public
    Constructor Create(list : TFhirDeviceUseStatementList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDeviceUseStatement read GetCurrent;
  end;


  {@Class TFhirDeviceUseStatementList
    A list of FhirDeviceUseStatement
  }
  {!.Net HL7Connect.Fhir.DeviceUseStatementList}
  TFhirDeviceUseStatementList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDeviceUseStatement;
    procedure SetItemN(index : Integer; value : TFhirDeviceUseStatement);
  public
    {!script hide}
    function Link : TFhirDeviceUseStatementList; Overload;
    function Clone : TFhirDeviceUseStatementList; Overload;
    function GetEnumerator : TFhirDeviceUseStatementListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDeviceUseStatement to the end of the list.
    }
    function Append : TFhirDeviceUseStatement;

    
    {@member AddItem
      Add an already existing FhirDeviceUseStatement to the end of the list.
    }
    procedure AddItem(value : TFhirDeviceUseStatement); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDeviceUseStatement) : Integer;
    

    {@member Insert
      Insert FhirDeviceUseStatement before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDeviceUseStatement;
    

    {@member InsertItem
       Insert an existing FhirDeviceUseStatement before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDeviceUseStatement);
    
    {@member Item
       Get the iIndexth FhirDeviceUseStatement. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDeviceUseStatement. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDeviceUseStatement);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDeviceUseStatement;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDeviceUseStatements[index : Integer] : TFhirDeviceUseStatement read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDiagnosticOrderEvent : TFhirElement
    A summary of the events of interest that have occurred as the request is processed; e.g. when the order was made, various processing steps (specimens received), when it was completed.
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrderEvent}
  TFhirDiagnosticOrderEvent = class (TFhirBackboneElement)
  private
    FStatus : TFhirEnum;
    FDescription : TFhirCodeableConcept;
    FDateTime : TFhirDateTime;
    FActor : TFhirReference{Resource};
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDiagnosticOrderStatus;
    Procedure SetStatusST(value : TFhirDiagnosticOrderStatus);
    Procedure SetDescription(value : TFhirCodeableConcept);
    Procedure SetDateTime(value : TFhirDateTime);
    Function GetDateTimeST : TDateAndTime;
    Procedure SetDateTimeST(value : TDateAndTime);
    Procedure SetActor(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDiagnosticOrderEvent; overload;
    function Clone : TFhirDiagnosticOrderEvent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member status
      The status for the event.
    }
    property status : TFhirDiagnosticOrderStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member description
      Typed access to Additional information about the event that occurred - e.g. if the status remained unchanged. (defined for API consistency)
    }
    property description : TFhirCodeableConcept read FDescription write SetDescription;
    {@member descriptionElement
      Additional information about the event that occurred - e.g. if the status remained unchanged.
    }
    property descriptionElement : TFhirCodeableConcept read FDescription write SetDescription;

    {@member dateTime
      Typed access to The date/time at which the event occurred.
    }
    property dateTime : TDateAndTime read GetDateTimeST write SetDateTimeST;
    {@member dateTimeElement
      The date/time at which the event occurred.
    }
    property dateTimeElement : TFhirDateTime read FDateTime write SetDateTime;

    {@member actor
      Typed access to The person responsible for performing or recording the action. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      The person responsible for performing or recording the action.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

  end;


  TFhirDiagnosticOrderEventListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDiagnosticOrderEventList;
    function GetCurrent : TFhirDiagnosticOrderEvent;
  public
    Constructor Create(list : TFhirDiagnosticOrderEventList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDiagnosticOrderEvent read GetCurrent;
  end;


  {@Class TFhirDiagnosticOrderEventList
    A list of FhirDiagnosticOrderEvent
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrderEventList}
  TFhirDiagnosticOrderEventList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDiagnosticOrderEvent;
    procedure SetItemN(index : Integer; value : TFhirDiagnosticOrderEvent);
  public
    {!script hide}
    function Link : TFhirDiagnosticOrderEventList; Overload;
    function Clone : TFhirDiagnosticOrderEventList; Overload;
    function GetEnumerator : TFhirDiagnosticOrderEventListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDiagnosticOrderEvent to the end of the list.
    }
    function Append : TFhirDiagnosticOrderEvent;

    
    {@member AddItem
      Add an already existing FhirDiagnosticOrderEvent to the end of the list.
    }
    procedure AddItem(value : TFhirDiagnosticOrderEvent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDiagnosticOrderEvent) : Integer;
    

    {@member Insert
      Insert FhirDiagnosticOrderEvent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDiagnosticOrderEvent;
    

    {@member InsertItem
       Insert an existing FhirDiagnosticOrderEvent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDiagnosticOrderEvent);
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrderEvent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrderEvent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDiagnosticOrderEvent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDiagnosticOrderEvent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDiagnosticOrderEvents[index : Integer] : TFhirDiagnosticOrderEvent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDiagnosticOrderItem : TFhirElement
    The specific diagnostic investigations that are requested as part of this request. Sometimes, there can only be one item per request, but in most contexts, more than one investigation can be requested.
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrderItem}
  TFhirDiagnosticOrderItem = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FspecimenList : TFhirReferenceList{TFhirSpecimen};
    FBodySite : TFhirCodeableConcept;
    FStatus : TFhirEnum;
    FeventList : TFhirDiagnosticOrderEventList;
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
    function GetHasSpecimenList : Boolean;
    Procedure SetBodySite(value : TFhirCodeableConcept);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDiagnosticOrderStatus;
    Procedure SetStatusST(value : TFhirDiagnosticOrderStatus);
    function GetEventList : TFhirDiagnosticOrderEventList;
    function GetHasEventList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDiagnosticOrderItem; overload;
    function Clone : TFhirDiagnosticOrderItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A code that identifies a particular diagnostic investigation, or panel of investigations, that have been requested. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code that identifies a particular diagnostic investigation, or panel of investigations, that have been requested.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member specimenList
      If the item is related to a specific specimen.
    }
    property specimenList : TFhirReferenceList{TFhirSpecimen} read GetSpecimenList;
    property hasSpecimenList : boolean read GetHasSpecimenList;

    {@member bodySite
      Typed access to Anatomical location where the request test should be performed.  This is the target site. (defined for API consistency)
    }
    property bodySite : TFhirCodeableConcept read FBodySite write SetBodySite;
    {@member bodySiteElement
      Anatomical location where the request test should be performed.  This is the target site.
    }
    property bodySiteElement : TFhirCodeableConcept read FBodySite write SetBodySite;

    {@member status
      The status of this individual item within the order.
    }
    property status : TFhirDiagnosticOrderStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member eventList
      A summary of the events of interest that have occurred as this item of the request is processed.
    }
    property eventList : TFhirDiagnosticOrderEventList read GetEventList;
    property hasEventList : boolean read GetHasEventList;

  end;


  TFhirDiagnosticOrderItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDiagnosticOrderItemList;
    function GetCurrent : TFhirDiagnosticOrderItem;
  public
    Constructor Create(list : TFhirDiagnosticOrderItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDiagnosticOrderItem read GetCurrent;
  end;


  {@Class TFhirDiagnosticOrderItemList
    A list of FhirDiagnosticOrderItem
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrderItemList}
  TFhirDiagnosticOrderItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDiagnosticOrderItem;
    procedure SetItemN(index : Integer; value : TFhirDiagnosticOrderItem);
  public
    {!script hide}
    function Link : TFhirDiagnosticOrderItemList; Overload;
    function Clone : TFhirDiagnosticOrderItemList; Overload;
    function GetEnumerator : TFhirDiagnosticOrderItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDiagnosticOrderItem to the end of the list.
    }
    function Append : TFhirDiagnosticOrderItem;

    
    {@member AddItem
      Add an already existing FhirDiagnosticOrderItem to the end of the list.
    }
    procedure AddItem(value : TFhirDiagnosticOrderItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDiagnosticOrderItem) : Integer;
    

    {@member Insert
      Insert FhirDiagnosticOrderItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDiagnosticOrderItem;
    

    {@member InsertItem
       Insert an existing FhirDiagnosticOrderItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDiagnosticOrderItem);
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrderItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrderItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDiagnosticOrderItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDiagnosticOrderItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDiagnosticOrderItems[index : Integer] : TFhirDiagnosticOrderItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDiagnosticOrder : TFhirDomainResource
    A record of a request for a diagnostic investigation service to be performed.
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrder}
  TFhirDiagnosticOrder = class (TFhirDomainResource)
  private
    FSubject : TFhirReference{Resource};
    FOrderer : TFhirReference{TFhirPractitioner};
    FidentifierList : TFhirIdentifierList;
    FEncounter : TFhirReference{TFhirEncounter};
    FreasonList : TFhirCodeableConceptList;
    FsupportingInformationList : TFhirReferenceList{Resource};
    FspecimenList : TFhirReferenceList{TFhirSpecimen};
    FStatus : TFhirEnum;
    FPriority : TFhirEnum;
    FeventList : TFhirDiagnosticOrderEventList;
    FitemList : TFhirDiagnosticOrderItemList;
    FnoteList : TFhirAnnotationList;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetOrderer(value : TFhirReference{TFhirPractitioner});
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    function GetSupportingInformationList : TFhirReferenceList{Resource};
    function GetHasSupportingInformationList : Boolean;
    function GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
    function GetHasSpecimenList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDiagnosticOrderStatus;
    Procedure SetStatusST(value : TFhirDiagnosticOrderStatus);
    Procedure SetPriority(value : TFhirEnum);
    Function GetPriorityST : TFhirDiagnosticOrderPriority;
    Procedure SetPriorityST(value : TFhirDiagnosticOrderPriority);
    function GetEventList : TFhirDiagnosticOrderEventList;
    function GetHasEventList : Boolean;
    function GetItemList : TFhirDiagnosticOrderItemList;
    function GetHasItemList : Boolean;
    function GetNoteList : TFhirAnnotationList;
    function GetHasNoteList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDiagnosticOrder; overload;
    function Clone : TFhirDiagnosticOrder; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member subject
      Typed access to Who or what the investigation is to be performed on. This is usually a human patient, but diagnostic tests can also be requested on animals, groups of humans or animals, devices such as dialysis machines, or even locations (typically for environmental scans). (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Who or what the investigation is to be performed on. This is usually a human patient, but diagnostic tests can also be requested on animals, groups of humans or animals, devices such as dialysis machines, or even locations (typically for environmental scans).
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member orderer
      Typed access to The practitioner that holds legal responsibility for ordering the investigation. (defined for API consistency)
    }
    property orderer : TFhirReference{TFhirPractitioner} read FOrderer write SetOrderer;
    {@member ordererElement
      The practitioner that holds legal responsibility for ordering the investigation.
    }
    property ordererElement : TFhirReference{TFhirPractitioner} read FOrderer write SetOrderer;

    {@member identifierList
      Identifiers assigned to this order instance by the orderer and/or  the receiver and/or order fulfiller.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member encounter
      Typed access to An encounter that provides additional information about the healthcare context in which this request is made. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      An encounter that provides additional information about the healthcare context in which this request is made.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member reasonList
      An explanation or justification for why this diagnostic investigation is being requested.   This is often for billing purposes.  May relate to the resources referred to in supportingInformation.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member supportingInformationList
      Additional clinical information about the patient or specimen that may influence test interpretations.  This includes observations explicitly requested by the producer(filler) to provide context or supporting information needed to complete the order.
    }
    property supportingInformationList : TFhirReferenceList{Resource} read GetSupportingInformationList;
    property hasSupportingInformationList : boolean read GetHasSupportingInformationList;

    {@member specimenList
      One or more specimens that the diagnostic investigation is about.
    }
    property specimenList : TFhirReferenceList{TFhirSpecimen} read GetSpecimenList;
    property hasSpecimenList : boolean read GetHasSpecimenList;

    {@member status
      The status of the order.
    }
    property status : TFhirDiagnosticOrderStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member priority
      The clinical priority associated with this order.
    }
    property priority : TFhirDiagnosticOrderPriority read GetPriorityST write SetPriorityST;
    property priorityElement : TFhirEnum read FPriority write SetPriority;

    {@member eventList
      A summary of the events of interest that have occurred as the request is processed; e.g. when the order was made, various processing steps (specimens received), when it was completed.
    }
    property eventList : TFhirDiagnosticOrderEventList read GetEventList;
    property hasEventList : boolean read GetHasEventList;

    {@member itemList
      The specific diagnostic investigations that are requested as part of this request. Sometimes, there can only be one item per request, but in most contexts, more than one investigation can be requested.
    }
    property itemList : TFhirDiagnosticOrderItemList read GetItemList;
    property hasItemList : boolean read GetHasItemList;

    {@member noteList
      Any other notes associated with this patient, specimen or order (e.g. "patient hates needles").
    }
    property noteList : TFhirAnnotationList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

  end;


  TFhirDiagnosticOrderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDiagnosticOrderList;
    function GetCurrent : TFhirDiagnosticOrder;
  public
    Constructor Create(list : TFhirDiagnosticOrderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDiagnosticOrder read GetCurrent;
  end;


  {@Class TFhirDiagnosticOrderList
    A list of FhirDiagnosticOrder
  }
  {!.Net HL7Connect.Fhir.DiagnosticOrderList}
  TFhirDiagnosticOrderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDiagnosticOrder;
    procedure SetItemN(index : Integer; value : TFhirDiagnosticOrder);
  public
    {!script hide}
    function Link : TFhirDiagnosticOrderList; Overload;
    function Clone : TFhirDiagnosticOrderList; Overload;
    function GetEnumerator : TFhirDiagnosticOrderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDiagnosticOrder to the end of the list.
    }
    function Append : TFhirDiagnosticOrder;

    
    {@member AddItem
      Add an already existing FhirDiagnosticOrder to the end of the list.
    }
    procedure AddItem(value : TFhirDiagnosticOrder); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDiagnosticOrder) : Integer;
    

    {@member Insert
      Insert FhirDiagnosticOrder before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDiagnosticOrder;
    

    {@member InsertItem
       Insert an existing FhirDiagnosticOrder before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDiagnosticOrder);
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrder. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDiagnosticOrder. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDiagnosticOrder);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDiagnosticOrder;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDiagnosticOrders[index : Integer] : TFhirDiagnosticOrder read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDiagnosticReportImage : TFhirElement
    A list of key images associated with this report. The images are generally created during the diagnostic process, and may be directly of the patient, or of treated specimens (i.e. slides of interest).
  }
  {!.Net HL7Connect.Fhir.DiagnosticReportImage}
  TFhirDiagnosticReportImage = class (TFhirBackboneElement)
  private
    FComment : TFhirString;
    FLink_ : TFhirReference{TFhirMedia};
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
    Procedure SetLink_(value : TFhirReference{TFhirMedia});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDiagnosticReportImage; overload;
    function Clone : TFhirDiagnosticReportImage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member comment
      Typed access to A comment about the image. Typically, this is used to provide an explanation for why the image is included, or to draw the viewer's attention to important features.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      A comment about the image. Typically, this is used to provide an explanation for why the image is included, or to draw the viewer's attention to important features.
    }
    property commentElement : TFhirString read FComment write SetComment;

    {@member link_
      Typed access to Reference to the image source. (defined for API consistency)
    }
    property link_ : TFhirReference{TFhirMedia} read FLink_ write SetLink_;
    {@member link_Element
      Reference to the image source.
    }
    property link_Element : TFhirReference{TFhirMedia} read FLink_ write SetLink_;

  end;


  TFhirDiagnosticReportImageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDiagnosticReportImageList;
    function GetCurrent : TFhirDiagnosticReportImage;
  public
    Constructor Create(list : TFhirDiagnosticReportImageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDiagnosticReportImage read GetCurrent;
  end;


  {@Class TFhirDiagnosticReportImageList
    A list of FhirDiagnosticReportImage
  }
  {!.Net HL7Connect.Fhir.DiagnosticReportImageList}
  TFhirDiagnosticReportImageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDiagnosticReportImage;
    procedure SetItemN(index : Integer; value : TFhirDiagnosticReportImage);
  public
    {!script hide}
    function Link : TFhirDiagnosticReportImageList; Overload;
    function Clone : TFhirDiagnosticReportImageList; Overload;
    function GetEnumerator : TFhirDiagnosticReportImageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDiagnosticReportImage to the end of the list.
    }
    function Append : TFhirDiagnosticReportImage;

    
    {@member AddItem
      Add an already existing FhirDiagnosticReportImage to the end of the list.
    }
    procedure AddItem(value : TFhirDiagnosticReportImage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDiagnosticReportImage) : Integer;
    

    {@member Insert
      Insert FhirDiagnosticReportImage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDiagnosticReportImage;
    

    {@member InsertItem
       Insert an existing FhirDiagnosticReportImage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDiagnosticReportImage);
    
    {@member Item
       Get the iIndexth FhirDiagnosticReportImage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDiagnosticReportImage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDiagnosticReportImage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDiagnosticReportImage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDiagnosticReportImages[index : Integer] : TFhirDiagnosticReportImage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDiagnosticReport : TFhirDomainResource
    The findings and interpretation of diagnostic  tests performed on patients, groups of patients, devices, and locations, and/or specimens derived from these. The report includes clinical context such as requesting and provider information, and some mix of atomic results, images, textual and coded interpretations, and formatted representation of diagnostic reports.
  }
  {!.Net HL7Connect.Fhir.DiagnosticReport}
  TFhirDiagnosticReport = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FCategory : TFhirCodeableConcept;
    FCode : TFhirCodeableConcept;
    FSubject : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FEffective : TFhirType;
    FIssued : TFhirInstant;
    FPerformer : TFhirReference{Resource};
    FrequestList : TFhirReferenceList{Resource};
    FspecimenList : TFhirReferenceList{TFhirSpecimen};
    FresultList : TFhirReferenceList{TFhirObservation};
    FimagingStudyList : TFhirReferenceList{Resource};
    FimageList : TFhirDiagnosticReportImageList;
    FConclusion : TFhirString;
    FcodedDiagnosisList : TFhirCodeableConceptList;
    FpresentedFormList : TFhirAttachmentList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDiagnosticReportStatus;
    Procedure SetStatusST(value : TFhirDiagnosticReportStatus);
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetEffective(value : TFhirType);
    Procedure SetIssued(value : TFhirInstant);
    Function GetIssuedST : TDateAndTime;
    Procedure SetIssuedST(value : TDateAndTime);
    Procedure SetPerformer(value : TFhirReference{Resource});
    function GetRequestList : TFhirReferenceList{Resource};
    function GetHasRequestList : Boolean;
    function GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
    function GetHasSpecimenList : Boolean;
    function GetResultList : TFhirReferenceList{TFhirObservation};
    function GetHasResultList : Boolean;
    function GetImagingStudyList : TFhirReferenceList{Resource};
    function GetHasImagingStudyList : Boolean;
    function GetImageList : TFhirDiagnosticReportImageList;
    function GetHasImageList : Boolean;
    Procedure SetConclusion(value : TFhirString);
    Function GetConclusionST : String;
    Procedure SetConclusionST(value : String);
    function GetCodedDiagnosisList : TFhirCodeableConceptList;
    function GetHasCodedDiagnosisList : Boolean;
    function GetPresentedFormList : TFhirAttachmentList;
    function GetHasPresentedFormList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDiagnosticReport; overload;
    function Clone : TFhirDiagnosticReport; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The local ID assigned to the report by the order filler, usually by the Information System of the diagnostic service provider.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      The status of the diagnostic report as a whole.
    }
    property status : TFhirDiagnosticReportStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member category
      Typed access to A code that classifies the clinical discipline, department or diagnostic service that created the report (e.g. cardiology, biochemistry, hematology, MRI). This is used for searching, sorting and display purposes. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      A code that classifies the clinical discipline, department or diagnostic service that created the report (e.g. cardiology, biochemistry, hematology, MRI). This is used for searching, sorting and display purposes.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member code
      Typed access to A code or name that describes this diagnostic report. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code or name that describes this diagnostic report.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member subject
      Typed access to The subject of the report. Usually, but not always, this is a patient. However diagnostic services also perform analyses on specimens collected from a variety of other sources. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The subject of the report. Usually, but not always, this is a patient. However diagnostic services also perform analyses on specimens collected from a variety of other sources.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member encounter
      Typed access to The link to the health care event (encounter) when the order was made. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The link to the health care event (encounter) when the order was made.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member effective
      Typed access to The time or time-period the observed values are related to. When the subject of the report is a patient, this is usually either the time of the procedure or of specimen collection(s), but very often the source of the date/time is not known, only the date/time itself. (defined for API consistency)
    }
    property effective : TFhirType read FEffective write SetEffective;
    {@member effectiveElement
      The time or time-period the observed values are related to. When the subject of the report is a patient, this is usually either the time of the procedure or of specimen collection(s), but very often the source of the date/time is not known, only the date/time itself.
    }
    property effectiveElement : TFhirType read FEffective write SetEffective;

    {@member issued
      Typed access to The date and time that this version of the report was released from the source diagnostic service.
    }
    property issued : TDateAndTime read GetIssuedST write SetIssuedST;
    {@member issuedElement
      The date and time that this version of the report was released from the source diagnostic service.
    }
    property issuedElement : TFhirInstant read FIssued write SetIssued;

    {@member performer
      Typed access to The diagnostic service that is responsible for issuing the report. (defined for API consistency)
    }
    property performer : TFhirReference{Resource} read FPerformer write SetPerformer;
    {@member performerElement
      The diagnostic service that is responsible for issuing the report.
    }
    property performerElement : TFhirReference{Resource} read FPerformer write SetPerformer;

    {@member requestList
      Details concerning a test or procedure requested.
    }
    property requestList : TFhirReferenceList{Resource} read GetRequestList;
    property hasRequestList : boolean read GetHasRequestList;

    {@member specimenList
      Details about the specimens on which this diagnostic report is based.
    }
    property specimenList : TFhirReferenceList{TFhirSpecimen} read GetSpecimenList;
    property hasSpecimenList : boolean read GetHasSpecimenList;

    {@member resultList
      Observations that are part of this diagnostic report. Observations can be simple name/value pairs (e.g. "atomic" results), or they can be grouping observations that include references to other members of the group (e.g. "panels").
    }
    property resultList : TFhirReferenceList{TFhirObservation} read GetResultList;
    property hasResultList : boolean read GetHasResultList;

    {@member imagingStudyList
      One or more links to full details of any imaging performed during the diagnostic investigation. Typically, this is imaging performed by DICOM enabled modalities, but this is not required. A fully enabled PACS viewer can use this information to provide views of the source images.
    }
    property imagingStudyList : TFhirReferenceList{Resource} read GetImagingStudyList;
    property hasImagingStudyList : boolean read GetHasImagingStudyList;

    {@member imageList
      A list of key images associated with this report. The images are generally created during the diagnostic process, and may be directly of the patient, or of treated specimens (i.e. slides of interest).
    }
    property imageList : TFhirDiagnosticReportImageList read GetImageList;
    property hasImageList : boolean read GetHasImageList;

    {@member conclusion
      Typed access to Concise and clinically contextualized narrative interpretation of the diagnostic report.
    }
    property conclusion : String read GetConclusionST write SetConclusionST;
    {@member conclusionElement
      Concise and clinically contextualized narrative interpretation of the diagnostic report.
    }
    property conclusionElement : TFhirString read FConclusion write SetConclusion;

    {@member codedDiagnosisList
      Codes for the conclusion.
    }
    property codedDiagnosisList : TFhirCodeableConceptList read GetCodedDiagnosisList;
    property hasCodedDiagnosisList : boolean read GetHasCodedDiagnosisList;

    {@member presentedFormList
      Rich text representation of the entire result as issued by the diagnostic service. Multiple formats are allowed but they SHALL be semantically equivalent.
    }
    property presentedFormList : TFhirAttachmentList read GetPresentedFormList;
    property hasPresentedFormList : boolean read GetHasPresentedFormList;

  end;


  TFhirDiagnosticReportListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDiagnosticReportList;
    function GetCurrent : TFhirDiagnosticReport;
  public
    Constructor Create(list : TFhirDiagnosticReportList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDiagnosticReport read GetCurrent;
  end;


  {@Class TFhirDiagnosticReportList
    A list of FhirDiagnosticReport
  }
  {!.Net HL7Connect.Fhir.DiagnosticReportList}
  TFhirDiagnosticReportList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDiagnosticReport;
    procedure SetItemN(index : Integer; value : TFhirDiagnosticReport);
  public
    {!script hide}
    function Link : TFhirDiagnosticReportList; Overload;
    function Clone : TFhirDiagnosticReportList; Overload;
    function GetEnumerator : TFhirDiagnosticReportListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDiagnosticReport to the end of the list.
    }
    function Append : TFhirDiagnosticReport;

    
    {@member AddItem
      Add an already existing FhirDiagnosticReport to the end of the list.
    }
    procedure AddItem(value : TFhirDiagnosticReport); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDiagnosticReport) : Integer;
    

    {@member Insert
      Insert FhirDiagnosticReport before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDiagnosticReport;
    

    {@member InsertItem
       Insert an existing FhirDiagnosticReport before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDiagnosticReport);
    
    {@member Item
       Get the iIndexth FhirDiagnosticReport. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDiagnosticReport. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDiagnosticReport);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDiagnosticReport;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDiagnosticReports[index : Integer] : TFhirDiagnosticReport read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentManifestContent : TFhirElement
    The list of Documents included in the manifest.
  }
  {!.Net HL7Connect.Fhir.DocumentManifestContent}
  TFhirDocumentManifestContent = class (TFhirBackboneElement)
  private
    FP : TFhirType;
    Procedure SetP(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentManifestContent; overload;
    function Clone : TFhirDocumentManifestContent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member p
      Typed access to The list of references to document content, or Attachment that consist of the parts of this document manifest. Usually, these would be document references, but direct references to Media or Attachments are also allowed. (defined for API consistency)
    }
    property p : TFhirType read FP write SetP;
    {@member pElement
      The list of references to document content, or Attachment that consist of the parts of this document manifest. Usually, these would be document references, but direct references to Media or Attachments are also allowed.
    }
    property pElement : TFhirType read FP write SetP;

  end;


  TFhirDocumentManifestContentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentManifestContentList;
    function GetCurrent : TFhirDocumentManifestContent;
  public
    Constructor Create(list : TFhirDocumentManifestContentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentManifestContent read GetCurrent;
  end;


  {@Class TFhirDocumentManifestContentList
    A list of FhirDocumentManifestContent
  }
  {!.Net HL7Connect.Fhir.DocumentManifestContentList}
  TFhirDocumentManifestContentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentManifestContent;
    procedure SetItemN(index : Integer; value : TFhirDocumentManifestContent);
  public
    {!script hide}
    function Link : TFhirDocumentManifestContentList; Overload;
    function Clone : TFhirDocumentManifestContentList; Overload;
    function GetEnumerator : TFhirDocumentManifestContentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentManifestContent to the end of the list.
    }
    function Append : TFhirDocumentManifestContent;

    
    {@member AddItem
      Add an already existing FhirDocumentManifestContent to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentManifestContent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentManifestContent) : Integer;
    

    {@member Insert
      Insert FhirDocumentManifestContent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentManifestContent;
    

    {@member InsertItem
       Insert an existing FhirDocumentManifestContent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentManifestContent);
    
    {@member Item
       Get the iIndexth FhirDocumentManifestContent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentManifestContent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentManifestContent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentManifestContent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentManifestContents[index : Integer] : TFhirDocumentManifestContent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentManifestRelated : TFhirElement
    Related identifiers or resources associated with the DocumentManifest.
  }
  {!.Net HL7Connect.Fhir.DocumentManifestRelated}
  TFhirDocumentManifestRelated = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirIdentifier;
    FRef : TFhirReference{TFhirReference};
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetRef(value : TFhirReference{TFhirReference});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentManifestRelated; overload;
    function Clone : TFhirDocumentManifestRelated; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to Related identifier to this DocumentManifest.  For example, Order numbers, accession numbers, XDW workflow numbers. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Related identifier to this DocumentManifest.  For example, Order numbers, accession numbers, XDW workflow numbers.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member ref
      Typed access to Related Resource to this DocumentManifest. For example, Order, DiagnosticOrder,  Procedure, EligibilityRequest, etc. (defined for API consistency)
    }
    property ref : TFhirReference{TFhirReference} read FRef write SetRef;
    {@member refElement
      Related Resource to this DocumentManifest. For example, Order, DiagnosticOrder,  Procedure, EligibilityRequest, etc.
    }
    property refElement : TFhirReference{TFhirReference} read FRef write SetRef;

  end;


  TFhirDocumentManifestRelatedListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentManifestRelatedList;
    function GetCurrent : TFhirDocumentManifestRelated;
  public
    Constructor Create(list : TFhirDocumentManifestRelatedList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentManifestRelated read GetCurrent;
  end;


  {@Class TFhirDocumentManifestRelatedList
    A list of FhirDocumentManifestRelated
  }
  {!.Net HL7Connect.Fhir.DocumentManifestRelatedList}
  TFhirDocumentManifestRelatedList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentManifestRelated;
    procedure SetItemN(index : Integer; value : TFhirDocumentManifestRelated);
  public
    {!script hide}
    function Link : TFhirDocumentManifestRelatedList; Overload;
    function Clone : TFhirDocumentManifestRelatedList; Overload;
    function GetEnumerator : TFhirDocumentManifestRelatedListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentManifestRelated to the end of the list.
    }
    function Append : TFhirDocumentManifestRelated;

    
    {@member AddItem
      Add an already existing FhirDocumentManifestRelated to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentManifestRelated); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentManifestRelated) : Integer;
    

    {@member Insert
      Insert FhirDocumentManifestRelated before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentManifestRelated;
    

    {@member InsertItem
       Insert an existing FhirDocumentManifestRelated before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentManifestRelated);
    
    {@member Item
       Get the iIndexth FhirDocumentManifestRelated. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentManifestRelated. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentManifestRelated);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentManifestRelated;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentManifestRelateds[index : Integer] : TFhirDocumentManifestRelated read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentManifest : TFhirDomainResource
    A manifest that defines a set of documents.
  }
  {!.Net HL7Connect.Fhir.DocumentManifest}
  TFhirDocumentManifest = class (TFhirDomainResource)
  private
    FMasterIdentifier : TFhirIdentifier;
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FrecipientList : TFhirReferenceList{Resource};
    FType_ : TFhirCodeableConcept;
    FauthorList : TFhirReferenceList{Resource};
    FCreated : TFhirDateTime;
    FSource : TFhirUri;
    FStatus : TFhirEnum;
    FDescription : TFhirString;
    FcontentList : TFhirDocumentManifestContentList;
    FrelatedList : TFhirDocumentManifestRelatedList;
    Procedure SetMasterIdentifier(value : TFhirIdentifier);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    function GetRecipientList : TFhirReferenceList{Resource};
    function GetHasRecipientList : Boolean;
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetAuthorList : TFhirReferenceList{Resource};
    function GetHasAuthorList : Boolean;
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetSource(value : TFhirUri);
    Function GetSourceST : String;
    Procedure SetSourceST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDocumentReferenceStatus;
    Procedure SetStatusST(value : TFhirDocumentReferenceStatus);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetContentList : TFhirDocumentManifestContentList;
    function GetHasContentList : Boolean;
    function GetRelatedList : TFhirDocumentManifestRelatedList;
    function GetHasRelatedList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentManifest; overload;
    function Clone : TFhirDocumentManifest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member masterIdentifier
      Typed access to A single identifier that uniquely identifies this manifest. Principally used to refer to the manifest in non-FHIR contexts. (defined for API consistency)
    }
    property masterIdentifier : TFhirIdentifier read FMasterIdentifier write SetMasterIdentifier;
    {@member masterIdentifierElement
      A single identifier that uniquely identifies this manifest. Principally used to refer to the manifest in non-FHIR contexts.
    }
    property masterIdentifierElement : TFhirIdentifier read FMasterIdentifier write SetMasterIdentifier;

    {@member identifierList
      Other identifiers associated with the document manifest, including version independent  identifiers.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to Who or what the set of documents is about. The documents can be about a person, (patient or healthcare practitioner), a device (i.e. machine) or even a group of subjects (such as a document about a herd of farm animals, or a set of patients that share a common exposure). If the documents cross more than one subject, then more than one subject is allowed here (unusual use case). (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Who or what the set of documents is about. The documents can be about a person, (patient or healthcare practitioner), a device (i.e. machine) or even a group of subjects (such as a document about a herd of farm animals, or a set of patients that share a common exposure). If the documents cross more than one subject, then more than one subject is allowed here (unusual use case).
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member recipientList
      A patient, practitioner, or organization for which this set of documents is intended.
    }
    property recipientList : TFhirReferenceList{Resource} read GetRecipientList;
    property hasRecipientList : boolean read GetHasRecipientList;

    {@member type_
      Typed access to Specifies the kind of this set of documents (e.g. Patient Summary, Discharge Summary, Prescription, etc.). The type of a set of documents may be the same as one of the documents in it - especially if there is only one - but it may be wider. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Specifies the kind of this set of documents (e.g. Patient Summary, Discharge Summary, Prescription, etc.). The type of a set of documents may be the same as one of the documents in it - especially if there is only one - but it may be wider.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member authorList
      Identifies who is responsible for creating the manifest, and adding  documents to it.
    }
    property authorList : TFhirReferenceList{Resource} read GetAuthorList;
    property hasAuthorList : boolean read GetHasAuthorList;

    {@member created
      Typed access to When the document manifest was created for submission to the server (not necessarily the same thing as the actual resource last modified time, since it may be modified, replicated, etc.).
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      When the document manifest was created for submission to the server (not necessarily the same thing as the actual resource last modified time, since it may be modified, replicated, etc.).
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member source
      Typed access to Identifies the source system, application, or software that produced the document manifest.
    }
    property source : String read GetSourceST write SetSourceST;
    {@member sourceElement
      Identifies the source system, application, or software that produced the document manifest.
    }
    property sourceElement : TFhirUri read FSource write SetSource;

    {@member status
      The status of this document manifest.
    }
    property status : TFhirDocumentReferenceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member description
      Typed access to Human-readable description of the source document. This is sometimes known as the "title".
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Human-readable description of the source document. This is sometimes known as the "title".
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member contentList
      The list of Documents included in the manifest.
    }
    property contentList : TFhirDocumentManifestContentList read GetContentList;
    property hasContentList : boolean read GetHasContentList;

    {@member relatedList
      Related identifiers or resources associated with the DocumentManifest.
    }
    property relatedList : TFhirDocumentManifestRelatedList read GetRelatedList;
    property hasRelatedList : boolean read GetHasRelatedList;

  end;


  TFhirDocumentManifestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentManifestList;
    function GetCurrent : TFhirDocumentManifest;
  public
    Constructor Create(list : TFhirDocumentManifestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentManifest read GetCurrent;
  end;


  {@Class TFhirDocumentManifestList
    A list of FhirDocumentManifest
  }
  {!.Net HL7Connect.Fhir.DocumentManifestList}
  TFhirDocumentManifestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentManifest;
    procedure SetItemN(index : Integer; value : TFhirDocumentManifest);
  public
    {!script hide}
    function Link : TFhirDocumentManifestList; Overload;
    function Clone : TFhirDocumentManifestList; Overload;
    function GetEnumerator : TFhirDocumentManifestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentManifest to the end of the list.
    }
    function Append : TFhirDocumentManifest;

    
    {@member AddItem
      Add an already existing FhirDocumentManifest to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentManifest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentManifest) : Integer;
    

    {@member Insert
      Insert FhirDocumentManifest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentManifest;
    

    {@member InsertItem
       Insert an existing FhirDocumentManifest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentManifest);
    
    {@member Item
       Get the iIndexth FhirDocumentManifest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentManifest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentManifest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentManifest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentManifests[index : Integer] : TFhirDocumentManifest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentReferenceRelatesTo : TFhirElement
    Relationships that this document has with other document references that already exist.
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceRelatesTo}
  TFhirDocumentReferenceRelatesTo = class (TFhirBackboneElement)
  private
    FCode : TFhirEnum;
    FTarget : TFhirReference{TFhirDocumentReference};
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirDocumentRelationshipType;
    Procedure SetCodeST(value : TFhirDocumentRelationshipType);
    Procedure SetTarget(value : TFhirReference{TFhirDocumentReference});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentReferenceRelatesTo; overload;
    function Clone : TFhirDocumentReferenceRelatesTo; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      The type of relationship that this document has with anther document.
    }
    property code : TFhirDocumentRelationshipType read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member target
      Typed access to The target document of this relationship. (defined for API consistency)
    }
    property target : TFhirReference{TFhirDocumentReference} read FTarget write SetTarget;
    {@member targetElement
      The target document of this relationship.
    }
    property targetElement : TFhirReference{TFhirDocumentReference} read FTarget write SetTarget;

  end;


  TFhirDocumentReferenceRelatesToListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentReferenceRelatesToList;
    function GetCurrent : TFhirDocumentReferenceRelatesTo;
  public
    Constructor Create(list : TFhirDocumentReferenceRelatesToList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentReferenceRelatesTo read GetCurrent;
  end;


  {@Class TFhirDocumentReferenceRelatesToList
    A list of FhirDocumentReferenceRelatesTo
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceRelatesToList}
  TFhirDocumentReferenceRelatesToList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentReferenceRelatesTo;
    procedure SetItemN(index : Integer; value : TFhirDocumentReferenceRelatesTo);
  public
    {!script hide}
    function Link : TFhirDocumentReferenceRelatesToList; Overload;
    function Clone : TFhirDocumentReferenceRelatesToList; Overload;
    function GetEnumerator : TFhirDocumentReferenceRelatesToListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentReferenceRelatesTo to the end of the list.
    }
    function Append : TFhirDocumentReferenceRelatesTo;

    
    {@member AddItem
      Add an already existing FhirDocumentReferenceRelatesTo to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentReferenceRelatesTo); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentReferenceRelatesTo) : Integer;
    

    {@member Insert
      Insert FhirDocumentReferenceRelatesTo before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentReferenceRelatesTo;
    

    {@member InsertItem
       Insert an existing FhirDocumentReferenceRelatesTo before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentReferenceRelatesTo);
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceRelatesTo. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceRelatesTo. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentReferenceRelatesTo);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentReferenceRelatesTo;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentReferenceRelatesTos[index : Integer] : TFhirDocumentReferenceRelatesTo read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentReferenceContent : TFhirElement
    The document and format referenced. There may be multiple content element repetitions, each with a different format.
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContent}
  TFhirDocumentReferenceContent = class (TFhirBackboneElement)
  private
    FAttachment : TFhirAttachment;
    FformatList : TFhirCodingList;
    Procedure SetAttachment(value : TFhirAttachment);
    function GetFormatList : TFhirCodingList;
    function GetHasFormatList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentReferenceContent; overload;
    function Clone : TFhirDocumentReferenceContent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member attachment
      Typed access to The document or url of the document along with critical metadata to prove content has integrity. (defined for API consistency)
    }
    property attachment : TFhirAttachment read FAttachment write SetAttachment;
    {@member attachmentElement
      The document or url of the document along with critical metadata to prove content has integrity.
    }
    property attachmentElement : TFhirAttachment read FAttachment write SetAttachment;

    {@member formatList
      An identifier of the document encoding, structure, and template that the document conforms to beyond the base format indicated in the mimeType.
    }
    property formatList : TFhirCodingList read GetFormatList;
    property hasFormatList : boolean read GetHasFormatList;

  end;


  TFhirDocumentReferenceContentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentReferenceContentList;
    function GetCurrent : TFhirDocumentReferenceContent;
  public
    Constructor Create(list : TFhirDocumentReferenceContentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentReferenceContent read GetCurrent;
  end;


  {@Class TFhirDocumentReferenceContentList
    A list of FhirDocumentReferenceContent
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContentList}
  TFhirDocumentReferenceContentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentReferenceContent;
    procedure SetItemN(index : Integer; value : TFhirDocumentReferenceContent);
  public
    {!script hide}
    function Link : TFhirDocumentReferenceContentList; Overload;
    function Clone : TFhirDocumentReferenceContentList; Overload;
    function GetEnumerator : TFhirDocumentReferenceContentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentReferenceContent to the end of the list.
    }
    function Append : TFhirDocumentReferenceContent;

    
    {@member AddItem
      Add an already existing FhirDocumentReferenceContent to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentReferenceContent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentReferenceContent) : Integer;
    

    {@member Insert
      Insert FhirDocumentReferenceContent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentReferenceContent;
    

    {@member InsertItem
       Insert an existing FhirDocumentReferenceContent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentReferenceContent);
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentReferenceContent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentReferenceContent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentReferenceContents[index : Integer] : TFhirDocumentReferenceContent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentReferenceContext : TFhirElement
    The clinical context in which the document was prepared.
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContext}
  TFhirDocumentReferenceContext = class (TFhirBackboneElement)
  private
    FEncounter : TFhirReference{TFhirEncounter};
    FeventList : TFhirCodeableConceptList;
    FPeriod : TFhirPeriod;
    FFacilityType : TFhirCodeableConcept;
    FPracticeSetting : TFhirCodeableConcept;
    FSourcePatientInfo : TFhirReference{TFhirPatient};
    FrelatedList : TFhirDocumentReferenceContextRelatedList;
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    function GetEventList : TFhirCodeableConceptList;
    function GetHasEventList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetFacilityType(value : TFhirCodeableConcept);
    Procedure SetPracticeSetting(value : TFhirCodeableConcept);
    Procedure SetSourcePatientInfo(value : TFhirReference{TFhirPatient});
    function GetRelatedList : TFhirDocumentReferenceContextRelatedList;
    function GetHasRelatedList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentReferenceContext; overload;
    function Clone : TFhirDocumentReferenceContext; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member encounter
      Typed access to Describes the clinical encounter or type of care that the document content is associated with. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      Describes the clinical encounter or type of care that the document content is associated with.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member eventList
      This list of codes represents the main clinical acts, such as a colonoscopy or an appendectomy, being documented. In some cases, the event is inherent in the typeCode, such as a "History and Physical Report" in which the procedure being documented is necessarily a "History and Physical" act.
    }
    property eventList : TFhirCodeableConceptList read GetEventList;
    property hasEventList : boolean read GetHasEventList;

    {@member period
      Typed access to The time period over which the service that is described by the document was provided. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The time period over which the service that is described by the document was provided.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member facilityType
      Typed access to The kind of facility where the patient was seen. (defined for API consistency)
    }
    property facilityType : TFhirCodeableConcept read FFacilityType write SetFacilityType;
    {@member facilityTypeElement
      The kind of facility where the patient was seen.
    }
    property facilityTypeElement : TFhirCodeableConcept read FFacilityType write SetFacilityType;

    {@member practiceSetting
      Typed access to This property may convey specifics about the practice setting where the content was created, often reflecting the clinical specialty. (defined for API consistency)
    }
    property practiceSetting : TFhirCodeableConcept read FPracticeSetting write SetPracticeSetting;
    {@member practiceSettingElement
      This property may convey specifics about the practice setting where the content was created, often reflecting the clinical specialty.
    }
    property practiceSettingElement : TFhirCodeableConcept read FPracticeSetting write SetPracticeSetting;

    {@member sourcePatientInfo
      Typed access to The Patient Information as known when the document was published. May be a reference to a version specific, or contained. (defined for API consistency)
    }
    property sourcePatientInfo : TFhirReference{TFhirPatient} read FSourcePatientInfo write SetSourcePatientInfo;
    {@member sourcePatientInfoElement
      The Patient Information as known when the document was published. May be a reference to a version specific, or contained.
    }
    property sourcePatientInfoElement : TFhirReference{TFhirPatient} read FSourcePatientInfo write SetSourcePatientInfo;

    {@member relatedList
      Related identifiers or resources associated with the DocumentReference.
    }
    property relatedList : TFhirDocumentReferenceContextRelatedList read GetRelatedList;
    property hasRelatedList : boolean read GetHasRelatedList;

  end;


  TFhirDocumentReferenceContextListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentReferenceContextList;
    function GetCurrent : TFhirDocumentReferenceContext;
  public
    Constructor Create(list : TFhirDocumentReferenceContextList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentReferenceContext read GetCurrent;
  end;


  {@Class TFhirDocumentReferenceContextList
    A list of FhirDocumentReferenceContext
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContextList}
  TFhirDocumentReferenceContextList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentReferenceContext;
    procedure SetItemN(index : Integer; value : TFhirDocumentReferenceContext);
  public
    {!script hide}
    function Link : TFhirDocumentReferenceContextList; Overload;
    function Clone : TFhirDocumentReferenceContextList; Overload;
    function GetEnumerator : TFhirDocumentReferenceContextListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentReferenceContext to the end of the list.
    }
    function Append : TFhirDocumentReferenceContext;

    
    {@member AddItem
      Add an already existing FhirDocumentReferenceContext to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentReferenceContext); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentReferenceContext) : Integer;
    

    {@member Insert
      Insert FhirDocumentReferenceContext before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentReferenceContext;
    

    {@member InsertItem
       Insert an existing FhirDocumentReferenceContext before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentReferenceContext);
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContext. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContext. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentReferenceContext);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentReferenceContext;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentReferenceContexts[index : Integer] : TFhirDocumentReferenceContext read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentReferenceContextRelated : TFhirElement
    Related identifiers or resources associated with the DocumentReference.
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContextRelated}
  TFhirDocumentReferenceContextRelated = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirIdentifier;
    FRef : TFhirReference{TFhirReference};
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetRef(value : TFhirReference{TFhirReference});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentReferenceContextRelated; overload;
    function Clone : TFhirDocumentReferenceContextRelated; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to Related identifier to this DocumentReference. If both id and ref are present they shall refer to the same thing. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Related identifier to this DocumentReference. If both id and ref are present they shall refer to the same thing.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member ref
      Typed access to Related Resource to this DocumentReference. If both id and ref are present they shall refer to the same thing. (defined for API consistency)
    }
    property ref : TFhirReference{TFhirReference} read FRef write SetRef;
    {@member refElement
      Related Resource to this DocumentReference. If both id and ref are present they shall refer to the same thing.
    }
    property refElement : TFhirReference{TFhirReference} read FRef write SetRef;

  end;


  TFhirDocumentReferenceContextRelatedListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentReferenceContextRelatedList;
    function GetCurrent : TFhirDocumentReferenceContextRelated;
  public
    Constructor Create(list : TFhirDocumentReferenceContextRelatedList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentReferenceContextRelated read GetCurrent;
  end;


  {@Class TFhirDocumentReferenceContextRelatedList
    A list of FhirDocumentReferenceContextRelated
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceContextRelatedList}
  TFhirDocumentReferenceContextRelatedList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentReferenceContextRelated;
    procedure SetItemN(index : Integer; value : TFhirDocumentReferenceContextRelated);
  public
    {!script hide}
    function Link : TFhirDocumentReferenceContextRelatedList; Overload;
    function Clone : TFhirDocumentReferenceContextRelatedList; Overload;
    function GetEnumerator : TFhirDocumentReferenceContextRelatedListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentReferenceContextRelated to the end of the list.
    }
    function Append : TFhirDocumentReferenceContextRelated;

    
    {@member AddItem
      Add an already existing FhirDocumentReferenceContextRelated to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentReferenceContextRelated); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentReferenceContextRelated) : Integer;
    

    {@member Insert
      Insert FhirDocumentReferenceContextRelated before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentReferenceContextRelated;
    

    {@member InsertItem
       Insert an existing FhirDocumentReferenceContextRelated before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentReferenceContextRelated);
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContextRelated. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentReferenceContextRelated. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentReferenceContextRelated);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentReferenceContextRelated;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentReferenceContextRelateds[index : Integer] : TFhirDocumentReferenceContextRelated read GetItemN write SetItemN; default;
  End;


  {@Class TFhirDocumentReference : TFhirDomainResource
    A reference to a document .
  }
  {!.Net HL7Connect.Fhir.DocumentReference}
  TFhirDocumentReference = class (TFhirDomainResource)
  private
    FMasterIdentifier : TFhirIdentifier;
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FType_ : TFhirCodeableConcept;
    FClass_ : TFhirCodeableConcept;
    FauthorList : TFhirReferenceList{Resource};
    FCustodian : TFhirReference{TFhirOrganization};
    FAuthenticator : TFhirReference{Resource};
    FCreated : TFhirDateTime;
    FIndexed : TFhirInstant;
    FStatus : TFhirEnum;
    FDocStatus : TFhirCodeableConcept;
    FrelatesToList : TFhirDocumentReferenceRelatesToList;
    FDescription : TFhirString;
    FsecurityLabelList : TFhirCodeableConceptList;
    FcontentList : TFhirDocumentReferenceContentList;
    FContext : TFhirDocumentReferenceContext;
    Procedure SetMasterIdentifier(value : TFhirIdentifier);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetClass_(value : TFhirCodeableConcept);
    function GetAuthorList : TFhirReferenceList{Resource};
    function GetHasAuthorList : Boolean;
    Procedure SetCustodian(value : TFhirReference{TFhirOrganization});
    Procedure SetAuthenticator(value : TFhirReference{Resource});
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetIndexed(value : TFhirInstant);
    Function GetIndexedST : TDateAndTime;
    Procedure SetIndexedST(value : TDateAndTime);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirDocumentReferenceStatus;
    Procedure SetStatusST(value : TFhirDocumentReferenceStatus);
    Procedure SetDocStatus(value : TFhirCodeableConcept);
    function GetRelatesToList : TFhirDocumentReferenceRelatesToList;
    function GetHasRelatesToList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetSecurityLabelList : TFhirCodeableConceptList;
    function GetHasSecurityLabelList : Boolean;
    function GetContentList : TFhirDocumentReferenceContentList;
    function GetHasContentList : Boolean;
    Procedure SetContext(value : TFhirDocumentReferenceContext);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirDocumentReference; overload;
    function Clone : TFhirDocumentReference; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member masterIdentifier
      Typed access to Document identifier as assigned by the source of the document. This identifier is specific to this version of the document. This unique identifier may be used elsewhere to identify this version of the document. (defined for API consistency)
    }
    property masterIdentifier : TFhirIdentifier read FMasterIdentifier write SetMasterIdentifier;
    {@member masterIdentifierElement
      Document identifier as assigned by the source of the document. This identifier is specific to this version of the document. This unique identifier may be used elsewhere to identify this version of the document.
    }
    property masterIdentifierElement : TFhirIdentifier read FMasterIdentifier write SetMasterIdentifier;

    {@member identifierList
      Other identifiers associated with the document, including version independent identifiers.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to Who or what the document is about. The document can be about a person, (patient or healthcare practitioner), a device (e.g. a machine) or even a group of subjects (such as a document about a herd of farm animals, or a set of patients that share a common exposure). (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Who or what the document is about. The document can be about a person, (patient or healthcare practitioner), a device (e.g. a machine) or even a group of subjects (such as a document about a herd of farm animals, or a set of patients that share a common exposure).
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member type_
      Typed access to Specifies the particular kind of document referenced  (e.g. History and Physical, Discharge Summary, Progress Note). This usually equates to the purpose of making the document referenced. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Specifies the particular kind of document referenced  (e.g. History and Physical, Discharge Summary, Progress Note). This usually equates to the purpose of making the document referenced.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member class_
      Typed access to A categorization for the type of document referenced - helps for indexing and searching. This may be implied by or derived from the code specified in the DocumentReference.type. (defined for API consistency)
    }
    property class_ : TFhirCodeableConcept read FClass_ write SetClass_;
    {@member class_Element
      A categorization for the type of document referenced - helps for indexing and searching. This may be implied by or derived from the code specified in the DocumentReference.type.
    }
    property class_Element : TFhirCodeableConcept read FClass_ write SetClass_;

    {@member authorList
      Identifies who is responsible for adding the information to the document.
    }
    property authorList : TFhirReferenceList{Resource} read GetAuthorList;
    property hasAuthorList : boolean read GetHasAuthorList;

    {@member custodian
      Typed access to Identifies the organization or group who is responsible for ongoing maintenance of and access to the document. (defined for API consistency)
    }
    property custodian : TFhirReference{TFhirOrganization} read FCustodian write SetCustodian;
    {@member custodianElement
      Identifies the organization or group who is responsible for ongoing maintenance of and access to the document.
    }
    property custodianElement : TFhirReference{TFhirOrganization} read FCustodian write SetCustodian;

    {@member authenticator
      Typed access to Which person or organization authenticates that this document is valid. (defined for API consistency)
    }
    property authenticator : TFhirReference{Resource} read FAuthenticator write SetAuthenticator;
    {@member authenticatorElement
      Which person or organization authenticates that this document is valid.
    }
    property authenticatorElement : TFhirReference{Resource} read FAuthenticator write SetAuthenticator;

    {@member created
      Typed access to When the document was created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      When the document was created.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member indexed
      Typed access to When the document reference was created.
    }
    property indexed : TDateAndTime read GetIndexedST write SetIndexedST;
    {@member indexedElement
      When the document reference was created.
    }
    property indexedElement : TFhirInstant read FIndexed write SetIndexed;

    {@member status
      The status of this document reference.
    }
    property status : TFhirDocumentReferenceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member docStatus
      Typed access to The status of the underlying document. (defined for API consistency)
    }
    property docStatus : TFhirCodeableConcept read FDocStatus write SetDocStatus;
    {@member docStatusElement
      The status of the underlying document.
    }
    property docStatusElement : TFhirCodeableConcept read FDocStatus write SetDocStatus;

    {@member relatesToList
      Relationships that this document has with other document references that already exist.
    }
    property relatesToList : TFhirDocumentReferenceRelatesToList read GetRelatesToList;
    property hasRelatesToList : boolean read GetHasRelatesToList;

    {@member description
      Typed access to Human-readable description of the source document. This is sometimes known as the "title".
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Human-readable description of the source document. This is sometimes known as the "title".
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member securityLabelList
      A set of Security-Tag codes specifying the level of privacy/security of the Document. Note that DocumentReference.meta.security contains the security labels of the "reference" to the document, while DocumentReference.securityLabel contains a snapshot of the security labels on the document the reference refers to.
    }
    property securityLabelList : TFhirCodeableConceptList read GetSecurityLabelList;
    property hasSecurityLabelList : boolean read GetHasSecurityLabelList;

    {@member contentList
      The document and format referenced. There may be multiple content element repetitions, each with a different format.
    }
    property contentList : TFhirDocumentReferenceContentList read GetContentList;
    property hasContentList : boolean read GetHasContentList;

    {@member context
      Typed access to The clinical context in which the document was prepared. (defined for API consistency)
    }
    property context : TFhirDocumentReferenceContext read FContext write SetContext;
    {@member contextElement
      The clinical context in which the document was prepared.
    }
    property contextElement : TFhirDocumentReferenceContext read FContext write SetContext;

  end;


  TFhirDocumentReferenceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirDocumentReferenceList;
    function GetCurrent : TFhirDocumentReference;
  public
    Constructor Create(list : TFhirDocumentReferenceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirDocumentReference read GetCurrent;
  end;


  {@Class TFhirDocumentReferenceList
    A list of FhirDocumentReference
  }
  {!.Net HL7Connect.Fhir.DocumentReferenceList}
  TFhirDocumentReferenceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirDocumentReference;
    procedure SetItemN(index : Integer; value : TFhirDocumentReference);
  public
    {!script hide}
    function Link : TFhirDocumentReferenceList; Overload;
    function Clone : TFhirDocumentReferenceList; Overload;
    function GetEnumerator : TFhirDocumentReferenceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirDocumentReference to the end of the list.
    }
    function Append : TFhirDocumentReference;

    
    {@member AddItem
      Add an already existing FhirDocumentReference to the end of the list.
    }
    procedure AddItem(value : TFhirDocumentReference); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirDocumentReference) : Integer;
    

    {@member Insert
      Insert FhirDocumentReference before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirDocumentReference;
    

    {@member InsertItem
       Insert an existing FhirDocumentReference before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirDocumentReference);
    
    {@member Item
       Get the iIndexth FhirDocumentReference. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirDocumentReference. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirDocumentReference);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirDocumentReference;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirDocumentReferences[index : Integer] : TFhirDocumentReference read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEligibilityRequest : TFhirDomainResource
    This resource provides the insurance eligibility details from the insurer regarding a specified coverage and optionally some class of service.
  }
  {!.Net HL7Connect.Fhir.EligibilityRequest}
  TFhirEligibilityRequest = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FTarget : TFhirReference{TFhirOrganization};
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetTarget(value : TFhirReference{TFhirOrganization});
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEligibilityRequest; overload;
    function Clone : TFhirEligibilityRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when this resource was created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when this resource was created.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member target
      Typed access to The Insurer who is target  of the request. (defined for API consistency)
    }
    property target : TFhirReference{TFhirOrganization} read FTarget write SetTarget;
    {@member targetElement
      The Insurer who is target  of the request.
    }
    property targetElement : TFhirReference{TFhirOrganization} read FTarget write SetTarget;

    {@member provider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

  end;


  TFhirEligibilityRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEligibilityRequestList;
    function GetCurrent : TFhirEligibilityRequest;
  public
    Constructor Create(list : TFhirEligibilityRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEligibilityRequest read GetCurrent;
  end;


  {@Class TFhirEligibilityRequestList
    A list of FhirEligibilityRequest
  }
  {!.Net HL7Connect.Fhir.EligibilityRequestList}
  TFhirEligibilityRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEligibilityRequest;
    procedure SetItemN(index : Integer; value : TFhirEligibilityRequest);
  public
    {!script hide}
    function Link : TFhirEligibilityRequestList; Overload;
    function Clone : TFhirEligibilityRequestList; Overload;
    function GetEnumerator : TFhirEligibilityRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEligibilityRequest to the end of the list.
    }
    function Append : TFhirEligibilityRequest;

    
    {@member AddItem
      Add an already existing FhirEligibilityRequest to the end of the list.
    }
    procedure AddItem(value : TFhirEligibilityRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEligibilityRequest) : Integer;
    

    {@member Insert
      Insert FhirEligibilityRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEligibilityRequest;
    

    {@member InsertItem
       Insert an existing FhirEligibilityRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEligibilityRequest);
    
    {@member Item
       Get the iIndexth FhirEligibilityRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEligibilityRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEligibilityRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEligibilityRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEligibilityRequests[index : Integer] : TFhirEligibilityRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEligibilityResponse : TFhirDomainResource
    This resource provides eligibility and plan details from the processing of an Eligibility resource.
  }
  {!.Net HL7Connect.Fhir.EligibilityResponse}
  TFhirEligibilityResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirEligibilityRequest};
    FOutcome : TFhirEnum;
    FDisposition : TFhirString;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirEligibilityRequest});
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirRemittanceOutcome;
    Procedure SetOutcomeST(value : TFhirRemittanceOutcome);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEligibilityResponse; overload;
    function Clone : TFhirEligibilityResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource reference. (defined for API consistency)
    }
    property request : TFhirReference{TFhirEligibilityRequest} read FRequest write SetRequest;
    {@member requestElement
      Original request resource reference.
    }
    property requestElement : TFhirReference{TFhirEligibilityRequest} read FRequest write SetRequest;

    {@member outcome
      Transaction status: error, complete.
    }
    property outcome : TFhirRemittanceOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member organization
      Typed access to The Insurer who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The Insurer who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

  end;


  TFhirEligibilityResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEligibilityResponseList;
    function GetCurrent : TFhirEligibilityResponse;
  public
    Constructor Create(list : TFhirEligibilityResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEligibilityResponse read GetCurrent;
  end;


  {@Class TFhirEligibilityResponseList
    A list of FhirEligibilityResponse
  }
  {!.Net HL7Connect.Fhir.EligibilityResponseList}
  TFhirEligibilityResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEligibilityResponse;
    procedure SetItemN(index : Integer; value : TFhirEligibilityResponse);
  public
    {!script hide}
    function Link : TFhirEligibilityResponseList; Overload;
    function Clone : TFhirEligibilityResponseList; Overload;
    function GetEnumerator : TFhirEligibilityResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEligibilityResponse to the end of the list.
    }
    function Append : TFhirEligibilityResponse;

    
    {@member AddItem
      Add an already existing FhirEligibilityResponse to the end of the list.
    }
    procedure AddItem(value : TFhirEligibilityResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEligibilityResponse) : Integer;
    

    {@member Insert
      Insert FhirEligibilityResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEligibilityResponse;
    

    {@member InsertItem
       Insert an existing FhirEligibilityResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEligibilityResponse);
    
    {@member Item
       Get the iIndexth FhirEligibilityResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEligibilityResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEligibilityResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEligibilityResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEligibilityResponses[index : Integer] : TFhirEligibilityResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEncounterStatusHistory : TFhirElement
    The status history permits the encounter resource to contain the status history without needing to read through the historical versions of the resource, or even have the server store them.
  }
  {!.Net HL7Connect.Fhir.EncounterStatusHistory}
  TFhirEncounterStatusHistory = class (TFhirBackboneElement)
  private
    FStatus : TFhirEnum;
    FPeriod : TFhirPeriod;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirEncounterState;
    Procedure SetStatusST(value : TFhirEncounterState);
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEncounterStatusHistory; overload;
    function Clone : TFhirEncounterStatusHistory; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member status
      planned | arrived | in-progress | onleave | finished | cancelled.
    }
    property status : TFhirEncounterState read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member period
      Typed access to The time that the episode was in the specified status. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The time that the episode was in the specified status.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirEncounterStatusHistoryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEncounterStatusHistoryList;
    function GetCurrent : TFhirEncounterStatusHistory;
  public
    Constructor Create(list : TFhirEncounterStatusHistoryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEncounterStatusHistory read GetCurrent;
  end;


  {@Class TFhirEncounterStatusHistoryList
    A list of FhirEncounterStatusHistory
  }
  {!.Net HL7Connect.Fhir.EncounterStatusHistoryList}
  TFhirEncounterStatusHistoryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEncounterStatusHistory;
    procedure SetItemN(index : Integer; value : TFhirEncounterStatusHistory);
  public
    {!script hide}
    function Link : TFhirEncounterStatusHistoryList; Overload;
    function Clone : TFhirEncounterStatusHistoryList; Overload;
    function GetEnumerator : TFhirEncounterStatusHistoryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEncounterStatusHistory to the end of the list.
    }
    function Append : TFhirEncounterStatusHistory;

    
    {@member AddItem
      Add an already existing FhirEncounterStatusHistory to the end of the list.
    }
    procedure AddItem(value : TFhirEncounterStatusHistory); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEncounterStatusHistory) : Integer;
    

    {@member Insert
      Insert FhirEncounterStatusHistory before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEncounterStatusHistory;
    

    {@member InsertItem
       Insert an existing FhirEncounterStatusHistory before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEncounterStatusHistory);
    
    {@member Item
       Get the iIndexth FhirEncounterStatusHistory. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEncounterStatusHistory. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEncounterStatusHistory);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEncounterStatusHistory;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEncounterStatusHistories[index : Integer] : TFhirEncounterStatusHistory read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEncounterParticipant : TFhirElement
    The?list of?people?responsible for providing the service.
  }
  {!.Net HL7Connect.Fhir.EncounterParticipant}
  TFhirEncounterParticipant = class (TFhirBackboneElement)
  private
    Ftype_List : TFhirCodeableConceptList;
    FPeriod : TFhirPeriod;
    FIndividual : TFhirReference{Resource};
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetIndividual(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEncounterParticipant; overload;
    function Clone : TFhirEncounterParticipant; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_List
      Role of participant in encounter.
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member period
      Typed access to The period of time that the specified participant was present during the encounter. These can overlap or be sub-sets of the overall encounters period. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time that the specified participant was present during the encounter. These can overlap or be sub-sets of the overall encounters period.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member individual
      Typed access to Persons involved in the encounter other than the patient. (defined for API consistency)
    }
    property individual : TFhirReference{Resource} read FIndividual write SetIndividual;
    {@member individualElement
      Persons involved in the encounter other than the patient.
    }
    property individualElement : TFhirReference{Resource} read FIndividual write SetIndividual;

  end;


  TFhirEncounterParticipantListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEncounterParticipantList;
    function GetCurrent : TFhirEncounterParticipant;
  public
    Constructor Create(list : TFhirEncounterParticipantList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEncounterParticipant read GetCurrent;
  end;


  {@Class TFhirEncounterParticipantList
    A list of FhirEncounterParticipant
  }
  {!.Net HL7Connect.Fhir.EncounterParticipantList}
  TFhirEncounterParticipantList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEncounterParticipant;
    procedure SetItemN(index : Integer; value : TFhirEncounterParticipant);
  public
    {!script hide}
    function Link : TFhirEncounterParticipantList; Overload;
    function Clone : TFhirEncounterParticipantList; Overload;
    function GetEnumerator : TFhirEncounterParticipantListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEncounterParticipant to the end of the list.
    }
    function Append : TFhirEncounterParticipant;

    
    {@member AddItem
      Add an already existing FhirEncounterParticipant to the end of the list.
    }
    procedure AddItem(value : TFhirEncounterParticipant); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEncounterParticipant) : Integer;
    

    {@member Insert
      Insert FhirEncounterParticipant before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEncounterParticipant;
    

    {@member InsertItem
       Insert an existing FhirEncounterParticipant before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEncounterParticipant);
    
    {@member Item
       Get the iIndexth FhirEncounterParticipant. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEncounterParticipant. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEncounterParticipant);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEncounterParticipant;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEncounterParticipants[index : Integer] : TFhirEncounterParticipant read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEncounterHospitalization : TFhirElement
    Details about the admission to a healthcare service.
  }
  {!.Net HL7Connect.Fhir.EncounterHospitalization}
  TFhirEncounterHospitalization = class (TFhirBackboneElement)
  private
    FPreAdmissionIdentifier : TFhirIdentifier;
    FOrigin : TFhirReference{TFhirLocation};
    FAdmitSource : TFhirCodeableConcept;
    FadmittingDiagnosisList : TFhirReferenceList{TFhirCondition};
    FReAdmission : TFhirCodeableConcept;
    FdietPreferenceList : TFhirCodeableConceptList;
    FspecialCourtesyList : TFhirCodeableConceptList;
    FspecialArrangementList : TFhirCodeableConceptList;
    FDestination : TFhirReference{TFhirLocation};
    FDischargeDisposition : TFhirCodeableConcept;
    FdischargeDiagnosisList : TFhirReferenceList{TFhirCondition};
    Procedure SetPreAdmissionIdentifier(value : TFhirIdentifier);
    Procedure SetOrigin(value : TFhirReference{TFhirLocation});
    Procedure SetAdmitSource(value : TFhirCodeableConcept);
    function GetAdmittingDiagnosisList : TFhirReferenceList{TFhirCondition};
    function GetHasAdmittingDiagnosisList : Boolean;
    Procedure SetReAdmission(value : TFhirCodeableConcept);
    function GetDietPreferenceList : TFhirCodeableConceptList;
    function GetHasDietPreferenceList : Boolean;
    function GetSpecialCourtesyList : TFhirCodeableConceptList;
    function GetHasSpecialCourtesyList : Boolean;
    function GetSpecialArrangementList : TFhirCodeableConceptList;
    function GetHasSpecialArrangementList : Boolean;
    Procedure SetDestination(value : TFhirReference{TFhirLocation});
    Procedure SetDischargeDisposition(value : TFhirCodeableConcept);
    function GetDischargeDiagnosisList : TFhirReferenceList{TFhirCondition};
    function GetHasDischargeDiagnosisList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEncounterHospitalization; overload;
    function Clone : TFhirEncounterHospitalization; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member preAdmissionIdentifier
      Typed access to Pre-admission identifier. (defined for API consistency)
    }
    property preAdmissionIdentifier : TFhirIdentifier read FPreAdmissionIdentifier write SetPreAdmissionIdentifier;
    {@member preAdmissionIdentifierElement
      Pre-admission identifier.
    }
    property preAdmissionIdentifierElement : TFhirIdentifier read FPreAdmissionIdentifier write SetPreAdmissionIdentifier;

    {@member origin
      Typed access to The location from which the patient came before admission. (defined for API consistency)
    }
    property origin : TFhirReference{TFhirLocation} read FOrigin write SetOrigin;
    {@member originElement
      The location from which the patient came before admission.
    }
    property originElement : TFhirReference{TFhirLocation} read FOrigin write SetOrigin;

    {@member admitSource
      Typed access to From where patient was admitted (physician referral, transfer). (defined for API consistency)
    }
    property admitSource : TFhirCodeableConcept read FAdmitSource write SetAdmitSource;
    {@member admitSourceElement
      From where patient was admitted (physician referral, transfer).
    }
    property admitSourceElement : TFhirCodeableConcept read FAdmitSource write SetAdmitSource;

    {@member admittingDiagnosisList
      The admitting diagnosis field is used to record the diagnosis codes as reported by admitting practitioner. This could be different or in addition to the conditions reported as reason-condition(s) for the encounter.
    }
    property admittingDiagnosisList : TFhirReferenceList{TFhirCondition} read GetAdmittingDiagnosisList;
    property hasAdmittingDiagnosisList : boolean read GetHasAdmittingDiagnosisList;

    {@member reAdmission
      Typed access to Whether this hospitalization is a readmission and why if known. (defined for API consistency)
    }
    property reAdmission : TFhirCodeableConcept read FReAdmission write SetReAdmission;
    {@member reAdmissionElement
      Whether this hospitalization is a readmission and why if known.
    }
    property reAdmissionElement : TFhirCodeableConcept read FReAdmission write SetReAdmission;

    {@member dietPreferenceList
      Diet preferences reported by the patient.
    }
    property dietPreferenceList : TFhirCodeableConceptList read GetDietPreferenceList;
    property hasDietPreferenceList : boolean read GetHasDietPreferenceList;

    {@member specialCourtesyList
      Special courtesies (VIP, board member).
    }
    property specialCourtesyList : TFhirCodeableConceptList read GetSpecialCourtesyList;
    property hasSpecialCourtesyList : boolean read GetHasSpecialCourtesyList;

    {@member specialArrangementList
      Wheelchair, translator, stretcher, etc.
    }
    property specialArrangementList : TFhirCodeableConceptList read GetSpecialArrangementList;
    property hasSpecialArrangementList : boolean read GetHasSpecialArrangementList;

    {@member destination
      Typed access to Location to which the patient is discharged. (defined for API consistency)
    }
    property destination : TFhirReference{TFhirLocation} read FDestination write SetDestination;
    {@member destinationElement
      Location to which the patient is discharged.
    }
    property destinationElement : TFhirReference{TFhirLocation} read FDestination write SetDestination;

    {@member dischargeDisposition
      Typed access to Category or kind of location after discharge. (defined for API consistency)
    }
    property dischargeDisposition : TFhirCodeableConcept read FDischargeDisposition write SetDischargeDisposition;
    {@member dischargeDispositionElement
      Category or kind of location after discharge.
    }
    property dischargeDispositionElement : TFhirCodeableConcept read FDischargeDisposition write SetDischargeDisposition;

    {@member dischargeDiagnosisList
      The final diagnosis given a patient before release from the hospital after all testing, surgery, and workup are complete.
    }
    property dischargeDiagnosisList : TFhirReferenceList{TFhirCondition} read GetDischargeDiagnosisList;
    property hasDischargeDiagnosisList : boolean read GetHasDischargeDiagnosisList;

  end;


  TFhirEncounterHospitalizationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEncounterHospitalizationList;
    function GetCurrent : TFhirEncounterHospitalization;
  public
    Constructor Create(list : TFhirEncounterHospitalizationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEncounterHospitalization read GetCurrent;
  end;


  {@Class TFhirEncounterHospitalizationList
    A list of FhirEncounterHospitalization
  }
  {!.Net HL7Connect.Fhir.EncounterHospitalizationList}
  TFhirEncounterHospitalizationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEncounterHospitalization;
    procedure SetItemN(index : Integer; value : TFhirEncounterHospitalization);
  public
    {!script hide}
    function Link : TFhirEncounterHospitalizationList; Overload;
    function Clone : TFhirEncounterHospitalizationList; Overload;
    function GetEnumerator : TFhirEncounterHospitalizationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEncounterHospitalization to the end of the list.
    }
    function Append : TFhirEncounterHospitalization;

    
    {@member AddItem
      Add an already existing FhirEncounterHospitalization to the end of the list.
    }
    procedure AddItem(value : TFhirEncounterHospitalization); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEncounterHospitalization) : Integer;
    

    {@member Insert
      Insert FhirEncounterHospitalization before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEncounterHospitalization;
    

    {@member InsertItem
       Insert an existing FhirEncounterHospitalization before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEncounterHospitalization);
    
    {@member Item
       Get the iIndexth FhirEncounterHospitalization. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEncounterHospitalization. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEncounterHospitalization);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEncounterHospitalization;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEncounterHospitalizations[index : Integer] : TFhirEncounterHospitalization read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEncounterLocation : TFhirElement
    List of locations where  the patient has been during this encounter.
  }
  {!.Net HL7Connect.Fhir.EncounterLocation}
  TFhirEncounterLocation = class (TFhirBackboneElement)
  private
    FLocation : TFhirReference{TFhirLocation};
    FStatus : TFhirEnum;
    FPeriod : TFhirPeriod;
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirEncounterLocationStatus;
    Procedure SetStatusST(value : TFhirEncounterLocationStatus);
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEncounterLocation; overload;
    function Clone : TFhirEncounterLocation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member location
      Typed access to The location where the encounter takes place. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      The location where the encounter takes place.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member status
      The status of the participants' presence at the specified location during the period specified. If the participant is is no longer at the location, then the period will have an end date/time.
    }
    property status : TFhirEncounterLocationStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member period
      Typed access to Time period during which the patient was present at the location. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      Time period during which the patient was present at the location.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirEncounterLocationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEncounterLocationList;
    function GetCurrent : TFhirEncounterLocation;
  public
    Constructor Create(list : TFhirEncounterLocationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEncounterLocation read GetCurrent;
  end;


  {@Class TFhirEncounterLocationList
    A list of FhirEncounterLocation
  }
  {!.Net HL7Connect.Fhir.EncounterLocationList}
  TFhirEncounterLocationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEncounterLocation;
    procedure SetItemN(index : Integer; value : TFhirEncounterLocation);
  public
    {!script hide}
    function Link : TFhirEncounterLocationList; Overload;
    function Clone : TFhirEncounterLocationList; Overload;
    function GetEnumerator : TFhirEncounterLocationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEncounterLocation to the end of the list.
    }
    function Append : TFhirEncounterLocation;

    
    {@member AddItem
      Add an already existing FhirEncounterLocation to the end of the list.
    }
    procedure AddItem(value : TFhirEncounterLocation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEncounterLocation) : Integer;
    

    {@member Insert
      Insert FhirEncounterLocation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEncounterLocation;
    

    {@member InsertItem
       Insert an existing FhirEncounterLocation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEncounterLocation);
    
    {@member Item
       Get the iIndexth FhirEncounterLocation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEncounterLocation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEncounterLocation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEncounterLocation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEncounterLocations[index : Integer] : TFhirEncounterLocation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEncounter : TFhirDomainResource
    An interaction between a patient and healthcare provider(s) for the purpose of providing healthcare service(s) or assessing the health status of a patient.
  }
  {!.Net HL7Connect.Fhir.Encounter}
  TFhirEncounter = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FstatusHistoryList : TFhirEncounterStatusHistoryList;
    FClass_ : TFhirEnum;
    Ftype_List : TFhirCodeableConceptList;
    FPriority : TFhirCodeableConcept;
    FPatient : TFhirReference{TFhirPatient};
    FepisodeOfCareList : TFhirReferenceList{TFhirEpisodeOfCare};
    FincomingReferralList : TFhirReferenceList{TFhirReferralRequest};
    FparticipantList : TFhirEncounterParticipantList;
    FAppointment : TFhirReference{TFhirAppointment};
    FPeriod : TFhirPeriod;
    FLength : TFhirQuantity;
    FreasonList : TFhirCodeableConceptList;
    FindicationList : TFhirReferenceList{Resource};
    FHospitalization : TFhirEncounterHospitalization;
    FlocationList : TFhirEncounterLocationList;
    FServiceProvider : TFhirReference{TFhirOrganization};
    FPartOf : TFhirReference{TFhirEncounter};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirEncounterState;
    Procedure SetStatusST(value : TFhirEncounterState);
    function GetStatusHistoryList : TFhirEncounterStatusHistoryList;
    function GetHasStatusHistoryList : Boolean;
    Procedure SetClass_(value : TFhirEnum);
    Function GetClass_ST : TFhirEncounterClass;
    Procedure SetClass_ST(value : TFhirEncounterClass);
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    Procedure SetPriority(value : TFhirCodeableConcept);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    function GetEpisodeOfCareList : TFhirReferenceList{TFhirEpisodeOfCare};
    function GetHasEpisodeOfCareList : Boolean;
    function GetIncomingReferralList : TFhirReferenceList{TFhirReferralRequest};
    function GetHasIncomingReferralList : Boolean;
    function GetParticipantList : TFhirEncounterParticipantList;
    function GetHasParticipantList : Boolean;
    Procedure SetAppointment(value : TFhirReference{TFhirAppointment});
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetLength(value : TFhirQuantity);
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    function GetIndicationList : TFhirReferenceList{Resource};
    function GetHasIndicationList : Boolean;
    Procedure SetHospitalization(value : TFhirEncounterHospitalization);
    function GetLocationList : TFhirEncounterLocationList;
    function GetHasLocationList : Boolean;
    Procedure SetServiceProvider(value : TFhirReference{TFhirOrganization});
    Procedure SetPartOf(value : TFhirReference{TFhirEncounter});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEncounter; overload;
    function Clone : TFhirEncounter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier(s) by which this encounter is known.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      planned | arrived | in-progress | onleave | finished | cancelled.
    }
    property status : TFhirEncounterState read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member statusHistoryList
      The status history permits the encounter resource to contain the status history without needing to read through the historical versions of the resource, or even have the server store them.
    }
    property statusHistoryList : TFhirEncounterStatusHistoryList read GetStatusHistoryList;
    property hasStatusHistoryList : boolean read GetHasStatusHistoryList;

    {@member class_
      inpatient | outpatient | ambulatory | emergency +.
    }
    property class_ : TFhirEncounterClass read GetClass_ST write SetClass_ST;
    property class_Element : TFhirEnum read FClass_ write SetClass_;

    {@member type_List
      Specific type of encounter (e.g. e-mail consultation, surgical day-care, skilled nursing, rehabilitation).
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member priority
      Typed access to Indicates the urgency of the encounter. (defined for API consistency)
    }
    property priority : TFhirCodeableConcept read FPriority write SetPriority;
    {@member priorityElement
      Indicates the urgency of the encounter.
    }
    property priorityElement : TFhirCodeableConcept read FPriority write SetPriority;

    {@member patient
      Typed access to The patient present at the encounter. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient present at the encounter.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member episodeOfCareList
      Where a specific encounter should be classified as a part of a specific episode(s) of care this field should be used. This association can facilitate grouping of related encounters together for a specific purpose, such as government reporting, issue tracking, association via a common problem.  The association is recorded on the encounter as these are typically created after the episode of care, and grouped on entry rather than editing the episode of care to append another encounter to it (the episode of care could span years).
    }
    property episodeOfCareList : TFhirReferenceList{TFhirEpisodeOfCare} read GetEpisodeOfCareList;
    property hasEpisodeOfCareList : boolean read GetHasEpisodeOfCareList;

    {@member incomingReferralList
      The referral request this encounter satisfies (incoming referral).
    }
    property incomingReferralList : TFhirReferenceList{TFhirReferralRequest} read GetIncomingReferralList;
    property hasIncomingReferralList : boolean read GetHasIncomingReferralList;

    {@member participantList
      The?list of?people?responsible for providing the service.
    }
    property participantList : TFhirEncounterParticipantList read GetParticipantList;
    property hasParticipantList : boolean read GetHasParticipantList;

    {@member appointment
      Typed access to The appointment that scheduled this encounter. (defined for API consistency)
    }
    property appointment : TFhirReference{TFhirAppointment} read FAppointment write SetAppointment;
    {@member appointmentElement
      The appointment that scheduled this encounter.
    }
    property appointmentElement : TFhirReference{TFhirAppointment} read FAppointment write SetAppointment;

    {@member period
      Typed access to The start and end time of the encounter. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The start and end time of the encounter.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member length
      Typed access to Quantity of time the encounter lasted. This excludes the time during leaves of absence. (defined for API consistency)
    }
    property length : TFhirQuantity read FLength write SetLength;
    {@member lengthElement
      Quantity of time the encounter lasted. This excludes the time during leaves of absence.
    }
    property lengthElement : TFhirQuantity read FLength write SetLength;

    {@member reasonList
      Reason the encounter takes place, expressed as a code. For admissions, this can be used for a coded admission diagnosis.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member indicationList
      Reason the encounter takes place, as specified using information from another resource. For admissions, this is the admission diagnosis. The indication will typically be a Condition (with other resources referenced in the evidence.detail), or a Procedure.
    }
    property indicationList : TFhirReferenceList{Resource} read GetIndicationList;
    property hasIndicationList : boolean read GetHasIndicationList;

    {@member hospitalization
      Typed access to Details about the admission to a healthcare service. (defined for API consistency)
    }
    property hospitalization : TFhirEncounterHospitalization read FHospitalization write SetHospitalization;
    {@member hospitalizationElement
      Details about the admission to a healthcare service.
    }
    property hospitalizationElement : TFhirEncounterHospitalization read FHospitalization write SetHospitalization;

    {@member locationList
      List of locations where  the patient has been during this encounter.
    }
    property locationList : TFhirEncounterLocationList read GetLocationList;
    property hasLocationList : boolean read GetHasLocationList;

    {@member serviceProvider
      Typed access to An organization that is in charge of maintaining the information of this Encounter (e.g. who maintains the report or the master service catalog item, etc.). This MAY be the same as the organization on the Patient record, however it could be different. This MAY not be not the Service Delivery Location's Organization. (defined for API consistency)
    }
    property serviceProvider : TFhirReference{TFhirOrganization} read FServiceProvider write SetServiceProvider;
    {@member serviceProviderElement
      An organization that is in charge of maintaining the information of this Encounter (e.g. who maintains the report or the master service catalog item, etc.). This MAY be the same as the organization on the Patient record, however it could be different. This MAY not be not the Service Delivery Location's Organization.
    }
    property serviceProviderElement : TFhirReference{TFhirOrganization} read FServiceProvider write SetServiceProvider;

    {@member partOf
      Typed access to Another Encounter of which this encounter is a part of (administratively or in time). (defined for API consistency)
    }
    property partOf : TFhirReference{TFhirEncounter} read FPartOf write SetPartOf;
    {@member partOfElement
      Another Encounter of which this encounter is a part of (administratively or in time).
    }
    property partOfElement : TFhirReference{TFhirEncounter} read FPartOf write SetPartOf;

  end;


  TFhirEncounterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEncounterList;
    function GetCurrent : TFhirEncounter;
  public
    Constructor Create(list : TFhirEncounterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEncounter read GetCurrent;
  end;


  {@Class TFhirEncounterList
    A list of FhirEncounter
  }
  {!.Net HL7Connect.Fhir.EncounterList}
  TFhirEncounterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEncounter;
    procedure SetItemN(index : Integer; value : TFhirEncounter);
  public
    {!script hide}
    function Link : TFhirEncounterList; Overload;
    function Clone : TFhirEncounterList; Overload;
    function GetEnumerator : TFhirEncounterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEncounter to the end of the list.
    }
    function Append : TFhirEncounter;

    
    {@member AddItem
      Add an already existing FhirEncounter to the end of the list.
    }
    procedure AddItem(value : TFhirEncounter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEncounter) : Integer;
    

    {@member Insert
      Insert FhirEncounter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEncounter;
    

    {@member InsertItem
       Insert an existing FhirEncounter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEncounter);
    
    {@member Item
       Get the iIndexth FhirEncounter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEncounter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEncounter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEncounter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEncounters[index : Integer] : TFhirEncounter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEnrollmentRequest : TFhirDomainResource
    This resource provides the insurance enrollment details to the insurer regarding a specified coverage.
  }
  {!.Net HL7Connect.Fhir.EnrollmentRequest}
  TFhirEnrollmentRequest = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FTarget : TFhirReference{TFhirOrganization};
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    FSubject : TFhirReference{TFhirPatient};
    FCoverage : TFhirReference{TFhirCoverage};
    FRelationship : TFhirCoding;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetTarget(value : TFhirReference{TFhirOrganization});
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetSubject(value : TFhirReference{TFhirPatient});
    Procedure SetCoverage(value : TFhirReference{TFhirCoverage});
    Procedure SetRelationship(value : TFhirCoding);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEnrollmentRequest; overload;
    function Clone : TFhirEnrollmentRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when this resource was created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when this resource was created.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member target
      Typed access to The Insurer who is target  of the request. (defined for API consistency)
    }
    property target : TFhirReference{TFhirOrganization} read FTarget write SetTarget;
    {@member targetElement
      The Insurer who is target  of the request.
    }
    property targetElement : TFhirReference{TFhirOrganization} read FTarget write SetTarget;

    {@member provider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member subject
      Typed access to Patient Resource. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirPatient} read FSubject write SetSubject;
    {@member subjectElement
      Patient Resource.
    }
    property subjectElement : TFhirReference{TFhirPatient} read FSubject write SetSubject;

    {@member coverage
      Typed access to Reference to the program or plan identification, underwriter or payor. (defined for API consistency)
    }
    property coverage : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;
    {@member coverageElement
      Reference to the program or plan identification, underwriter or payor.
    }
    property coverageElement : TFhirReference{TFhirCoverage} read FCoverage write SetCoverage;

    {@member relationship
      Typed access to The relationship of the patient to the subscriber. (defined for API consistency)
    }
    property relationship : TFhirCoding read FRelationship write SetRelationship;
    {@member relationshipElement
      The relationship of the patient to the subscriber.
    }
    property relationshipElement : TFhirCoding read FRelationship write SetRelationship;

  end;


  TFhirEnrollmentRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEnrollmentRequestList;
    function GetCurrent : TFhirEnrollmentRequest;
  public
    Constructor Create(list : TFhirEnrollmentRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEnrollmentRequest read GetCurrent;
  end;


  {@Class TFhirEnrollmentRequestList
    A list of FhirEnrollmentRequest
  }
  {!.Net HL7Connect.Fhir.EnrollmentRequestList}
  TFhirEnrollmentRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEnrollmentRequest;
    procedure SetItemN(index : Integer; value : TFhirEnrollmentRequest);
  public
    {!script hide}
    function Link : TFhirEnrollmentRequestList; Overload;
    function Clone : TFhirEnrollmentRequestList; Overload;
    function GetEnumerator : TFhirEnrollmentRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEnrollmentRequest to the end of the list.
    }
    function Append : TFhirEnrollmentRequest;

    
    {@member AddItem
      Add an already existing FhirEnrollmentRequest to the end of the list.
    }
    procedure AddItem(value : TFhirEnrollmentRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEnrollmentRequest) : Integer;
    

    {@member Insert
      Insert FhirEnrollmentRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEnrollmentRequest;
    

    {@member InsertItem
       Insert an existing FhirEnrollmentRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEnrollmentRequest);
    
    {@member Item
       Get the iIndexth FhirEnrollmentRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEnrollmentRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEnrollmentRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEnrollmentRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEnrollmentRequests[index : Integer] : TFhirEnrollmentRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEnrollmentResponse : TFhirDomainResource
    This resource provides enrollment and plan details from the processing of an Enrollment resource.
  }
  {!.Net HL7Connect.Fhir.EnrollmentResponse}
  TFhirEnrollmentResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirEnrollmentRequest};
    FOutcome : TFhirEnum;
    FDisposition : TFhirString;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirEnrollmentRequest});
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirRemittanceOutcome;
    Procedure SetOutcomeST(value : TFhirRemittanceOutcome);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEnrollmentResponse; overload;
    function Clone : TFhirEnrollmentResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource reference. (defined for API consistency)
    }
    property request : TFhirReference{TFhirEnrollmentRequest} read FRequest write SetRequest;
    {@member requestElement
      Original request resource reference.
    }
    property requestElement : TFhirReference{TFhirEnrollmentRequest} read FRequest write SetRequest;

    {@member outcome
      Transaction status: error, complete.
    }
    property outcome : TFhirRemittanceOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member organization
      Typed access to The Insurer who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The Insurer who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

  end;


  TFhirEnrollmentResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEnrollmentResponseList;
    function GetCurrent : TFhirEnrollmentResponse;
  public
    Constructor Create(list : TFhirEnrollmentResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEnrollmentResponse read GetCurrent;
  end;


  {@Class TFhirEnrollmentResponseList
    A list of FhirEnrollmentResponse
  }
  {!.Net HL7Connect.Fhir.EnrollmentResponseList}
  TFhirEnrollmentResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEnrollmentResponse;
    procedure SetItemN(index : Integer; value : TFhirEnrollmentResponse);
  public
    {!script hide}
    function Link : TFhirEnrollmentResponseList; Overload;
    function Clone : TFhirEnrollmentResponseList; Overload;
    function GetEnumerator : TFhirEnrollmentResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEnrollmentResponse to the end of the list.
    }
    function Append : TFhirEnrollmentResponse;

    
    {@member AddItem
      Add an already existing FhirEnrollmentResponse to the end of the list.
    }
    procedure AddItem(value : TFhirEnrollmentResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEnrollmentResponse) : Integer;
    

    {@member Insert
      Insert FhirEnrollmentResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEnrollmentResponse;
    

    {@member InsertItem
       Insert an existing FhirEnrollmentResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEnrollmentResponse);
    
    {@member Item
       Get the iIndexth FhirEnrollmentResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEnrollmentResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEnrollmentResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEnrollmentResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEnrollmentResponses[index : Integer] : TFhirEnrollmentResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEpisodeOfCareStatusHistory : TFhirElement
    The history of statuses that the EpisodeOfCare has been through (without requiring processing the history of the resource).
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCareStatusHistory}
  TFhirEpisodeOfCareStatusHistory = class (TFhirBackboneElement)
  private
    FStatus : TFhirEnum;
    FPeriod : TFhirPeriod;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirEpisodeOfCareStatus;
    Procedure SetStatusST(value : TFhirEpisodeOfCareStatus);
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEpisodeOfCareStatusHistory; overload;
    function Clone : TFhirEpisodeOfCareStatusHistory; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member status
      planned | waitlist | active | onhold | finished | cancelled.
    }
    property status : TFhirEpisodeOfCareStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member period
      Typed access to The period during this EpisodeOfCare that the specific status applied. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period during this EpisodeOfCare that the specific status applied.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirEpisodeOfCareStatusHistoryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEpisodeOfCareStatusHistoryList;
    function GetCurrent : TFhirEpisodeOfCareStatusHistory;
  public
    Constructor Create(list : TFhirEpisodeOfCareStatusHistoryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEpisodeOfCareStatusHistory read GetCurrent;
  end;


  {@Class TFhirEpisodeOfCareStatusHistoryList
    A list of FhirEpisodeOfCareStatusHistory
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCareStatusHistoryList}
  TFhirEpisodeOfCareStatusHistoryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEpisodeOfCareStatusHistory;
    procedure SetItemN(index : Integer; value : TFhirEpisodeOfCareStatusHistory);
  public
    {!script hide}
    function Link : TFhirEpisodeOfCareStatusHistoryList; Overload;
    function Clone : TFhirEpisodeOfCareStatusHistoryList; Overload;
    function GetEnumerator : TFhirEpisodeOfCareStatusHistoryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEpisodeOfCareStatusHistory to the end of the list.
    }
    function Append : TFhirEpisodeOfCareStatusHistory;

    
    {@member AddItem
      Add an already existing FhirEpisodeOfCareStatusHistory to the end of the list.
    }
    procedure AddItem(value : TFhirEpisodeOfCareStatusHistory); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEpisodeOfCareStatusHistory) : Integer;
    

    {@member Insert
      Insert FhirEpisodeOfCareStatusHistory before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEpisodeOfCareStatusHistory;
    

    {@member InsertItem
       Insert an existing FhirEpisodeOfCareStatusHistory before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEpisodeOfCareStatusHistory);
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCareStatusHistory. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCareStatusHistory. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEpisodeOfCareStatusHistory);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEpisodeOfCareStatusHistory;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEpisodeOfCareStatusHistories[index : Integer] : TFhirEpisodeOfCareStatusHistory read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEpisodeOfCareCareTeam : TFhirElement
    The list of practitioners that may be facilitating this episode of care for specific purposes.
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCareCareTeam}
  TFhirEpisodeOfCareCareTeam = class (TFhirBackboneElement)
  private
    FroleList : TFhirCodeableConceptList;
    FPeriod : TFhirPeriod;
    FMember : TFhirReference{Resource};
    function GetRoleList : TFhirCodeableConceptList;
    function GetHasRoleList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetMember(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEpisodeOfCareCareTeam; overload;
    function Clone : TFhirEpisodeOfCareCareTeam; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member roleList
      The role this team member is taking within this episode of care.
    }
    property roleList : TFhirCodeableConceptList read GetRoleList;
    property hasRoleList : boolean read GetHasRoleList;

    {@member period
      Typed access to The period of time this practitioner is performing some role within the episode of care. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time this practitioner is performing some role within the episode of care.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member member
      Typed access to The practitioner (or Organization) within the team. (defined for API consistency)
    }
    property member : TFhirReference{Resource} read FMember write SetMember;
    {@member memberElement
      The practitioner (or Organization) within the team.
    }
    property memberElement : TFhirReference{Resource} read FMember write SetMember;

  end;


  TFhirEpisodeOfCareCareTeamListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEpisodeOfCareCareTeamList;
    function GetCurrent : TFhirEpisodeOfCareCareTeam;
  public
    Constructor Create(list : TFhirEpisodeOfCareCareTeamList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEpisodeOfCareCareTeam read GetCurrent;
  end;


  {@Class TFhirEpisodeOfCareCareTeamList
    A list of FhirEpisodeOfCareCareTeam
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCareCareTeamList}
  TFhirEpisodeOfCareCareTeamList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEpisodeOfCareCareTeam;
    procedure SetItemN(index : Integer; value : TFhirEpisodeOfCareCareTeam);
  public
    {!script hide}
    function Link : TFhirEpisodeOfCareCareTeamList; Overload;
    function Clone : TFhirEpisodeOfCareCareTeamList; Overload;
    function GetEnumerator : TFhirEpisodeOfCareCareTeamListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEpisodeOfCareCareTeam to the end of the list.
    }
    function Append : TFhirEpisodeOfCareCareTeam;

    
    {@member AddItem
      Add an already existing FhirEpisodeOfCareCareTeam to the end of the list.
    }
    procedure AddItem(value : TFhirEpisodeOfCareCareTeam); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEpisodeOfCareCareTeam) : Integer;
    

    {@member Insert
      Insert FhirEpisodeOfCareCareTeam before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEpisodeOfCareCareTeam;
    

    {@member InsertItem
       Insert an existing FhirEpisodeOfCareCareTeam before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEpisodeOfCareCareTeam);
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCareCareTeam. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCareCareTeam. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEpisodeOfCareCareTeam);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEpisodeOfCareCareTeam;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEpisodeOfCareCareTeams[index : Integer] : TFhirEpisodeOfCareCareTeam read GetItemN write SetItemN; default;
  End;


  {@Class TFhirEpisodeOfCare : TFhirDomainResource
    An association between a patient and an organization / healthcare provider(s) during which time encounters may occur. The managing organization assumes a level of responsibility for the patient during this time.
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCare}
  TFhirEpisodeOfCare = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FstatusHistoryList : TFhirEpisodeOfCareStatusHistoryList;
    Ftype_List : TFhirCodeableConceptList;
    FconditionList : TFhirReferenceList{TFhirCondition};
    FPatient : TFhirReference{TFhirPatient};
    FManagingOrganization : TFhirReference{TFhirOrganization};
    FPeriod : TFhirPeriod;
    FreferralRequestList : TFhirReferenceList{TFhirReferralRequest};
    FCareManager : TFhirReference{TFhirPractitioner};
    FcareTeamList : TFhirEpisodeOfCareCareTeamList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirEpisodeOfCareStatus;
    Procedure SetStatusST(value : TFhirEpisodeOfCareStatus);
    function GetStatusHistoryList : TFhirEpisodeOfCareStatusHistoryList;
    function GetHasStatusHistoryList : Boolean;
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    function GetConditionList : TFhirReferenceList{TFhirCondition};
    function GetHasConditionList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetManagingOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetPeriod(value : TFhirPeriod);
    function GetReferralRequestList : TFhirReferenceList{TFhirReferralRequest};
    function GetHasReferralRequestList : Boolean;
    Procedure SetCareManager(value : TFhirReference{TFhirPractitioner});
    function GetCareTeamList : TFhirEpisodeOfCareCareTeamList;
    function GetHasCareTeamList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirEpisodeOfCare; overload;
    function Clone : TFhirEpisodeOfCare; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier(s) by which this EpisodeOfCare is known.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      planned | waitlist | active | onhold | finished | cancelled.
    }
    property status : TFhirEpisodeOfCareStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member statusHistoryList
      The history of statuses that the EpisodeOfCare has been through (without requiring processing the history of the resource).
    }
    property statusHistoryList : TFhirEpisodeOfCareStatusHistoryList read GetStatusHistoryList;
    property hasStatusHistoryList : boolean read GetHasStatusHistoryList;

    {@member type_List
      A classification of the type of encounter; e.g. specialist referral, disease management, type of funded care.
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member conditionList
      A list of conditions/problems/diagnoses that this episode of care is intended to be providing care for.
    }
    property conditionList : TFhirReferenceList{TFhirCondition} read GetConditionList;
    property hasConditionList : boolean read GetHasConditionList;

    {@member patient
      Typed access to The patient that this EpisodeOfCare applies to. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient that this EpisodeOfCare applies to.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member managingOrganization
      Typed access to The organization that has assumed the specific responsibilities for the specified duration. (defined for API consistency)
    }
    property managingOrganization : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;
    {@member managingOrganizationElement
      The organization that has assumed the specific responsibilities for the specified duration.
    }
    property managingOrganizationElement : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;

    {@member period
      Typed access to The interval during which the managing organization assumes the defined responsibility. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The interval during which the managing organization assumes the defined responsibility.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member referralRequestList
      Referral Request(s) that are fulfilled by this EpisodeOfCare, incoming referrals.
    }
    property referralRequestList : TFhirReferenceList{TFhirReferralRequest} read GetReferralRequestList;
    property hasReferralRequestList : boolean read GetHasReferralRequestList;

    {@member careManager
      Typed access to The practitioner that is the care manager/care co-ordinator for this patient. (defined for API consistency)
    }
    property careManager : TFhirReference{TFhirPractitioner} read FCareManager write SetCareManager;
    {@member careManagerElement
      The practitioner that is the care manager/care co-ordinator for this patient.
    }
    property careManagerElement : TFhirReference{TFhirPractitioner} read FCareManager write SetCareManager;

    {@member careTeamList
      The list of practitioners that may be facilitating this episode of care for specific purposes.
    }
    property careTeamList : TFhirEpisodeOfCareCareTeamList read GetCareTeamList;
    property hasCareTeamList : boolean read GetHasCareTeamList;

  end;


  TFhirEpisodeOfCareListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirEpisodeOfCareList;
    function GetCurrent : TFhirEpisodeOfCare;
  public
    Constructor Create(list : TFhirEpisodeOfCareList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirEpisodeOfCare read GetCurrent;
  end;


  {@Class TFhirEpisodeOfCareList
    A list of FhirEpisodeOfCare
  }
  {!.Net HL7Connect.Fhir.EpisodeOfCareList}
  TFhirEpisodeOfCareList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirEpisodeOfCare;
    procedure SetItemN(index : Integer; value : TFhirEpisodeOfCare);
  public
    {!script hide}
    function Link : TFhirEpisodeOfCareList; Overload;
    function Clone : TFhirEpisodeOfCareList; Overload;
    function GetEnumerator : TFhirEpisodeOfCareListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirEpisodeOfCare to the end of the list.
    }
    function Append : TFhirEpisodeOfCare;

    
    {@member AddItem
      Add an already existing FhirEpisodeOfCare to the end of the list.
    }
    procedure AddItem(value : TFhirEpisodeOfCare); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirEpisodeOfCare) : Integer;
    

    {@member Insert
      Insert FhirEpisodeOfCare before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirEpisodeOfCare;
    

    {@member InsertItem
       Insert an existing FhirEpisodeOfCare before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirEpisodeOfCare);
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCare. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirEpisodeOfCare. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirEpisodeOfCare);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirEpisodeOfCare;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirEpisodeOfCares[index : Integer] : TFhirEpisodeOfCare read GetItemN write SetItemN; default;
  End;


  {@Class TFhirExplanationOfBenefit : TFhirDomainResource
    This resource provides: the claim details; adjudication details from the processing of a Claim; and optionally account balance information, for informing the subscriber of the benefits provided.
  }
  {!.Net HL7Connect.Fhir.ExplanationOfBenefit}
  TFhirExplanationOfBenefit = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirClaim};
    FOutcome : TFhirEnum;
    FDisposition : TFhirString;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirClaim});
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirRemittanceOutcome;
    Procedure SetOutcomeST(value : TFhirRemittanceOutcome);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirExplanationOfBenefit; overload;
    function Clone : TFhirExplanationOfBenefit; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource reference. (defined for API consistency)
    }
    property request : TFhirReference{TFhirClaim} read FRequest write SetRequest;
    {@member requestElement
      Original request resource reference.
    }
    property requestElement : TFhirReference{TFhirClaim} read FRequest write SetRequest;

    {@member outcome
      Transaction status: error, complete.
    }
    property outcome : TFhirRemittanceOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member organization
      Typed access to The Insurer who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The Insurer who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

  end;


  TFhirExplanationOfBenefitListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirExplanationOfBenefitList;
    function GetCurrent : TFhirExplanationOfBenefit;
  public
    Constructor Create(list : TFhirExplanationOfBenefitList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirExplanationOfBenefit read GetCurrent;
  end;


  {@Class TFhirExplanationOfBenefitList
    A list of FhirExplanationOfBenefit
  }
  {!.Net HL7Connect.Fhir.ExplanationOfBenefitList}
  TFhirExplanationOfBenefitList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirExplanationOfBenefit;
    procedure SetItemN(index : Integer; value : TFhirExplanationOfBenefit);
  public
    {!script hide}
    function Link : TFhirExplanationOfBenefitList; Overload;
    function Clone : TFhirExplanationOfBenefitList; Overload;
    function GetEnumerator : TFhirExplanationOfBenefitListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirExplanationOfBenefit to the end of the list.
    }
    function Append : TFhirExplanationOfBenefit;

    
    {@member AddItem
      Add an already existing FhirExplanationOfBenefit to the end of the list.
    }
    procedure AddItem(value : TFhirExplanationOfBenefit); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirExplanationOfBenefit) : Integer;
    

    {@member Insert
      Insert FhirExplanationOfBenefit before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirExplanationOfBenefit;
    

    {@member InsertItem
       Insert an existing FhirExplanationOfBenefit before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirExplanationOfBenefit);
    
    {@member Item
       Get the iIndexth FhirExplanationOfBenefit. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirExplanationOfBenefit. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirExplanationOfBenefit);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirExplanationOfBenefit;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirExplanationOfBenefits[index : Integer] : TFhirExplanationOfBenefit read GetItemN write SetItemN; default;
  End;


  {@Class TFhirFamilyMemberHistoryCondition : TFhirElement
    The significant Conditions (or condition) that the family member had. This is a repeating section to allow a system to represent more than one condition per resource, though there is nothing stopping multiple resources - one per condition.
  }
  {!.Net HL7Connect.Fhir.FamilyMemberHistoryCondition}
  TFhirFamilyMemberHistoryCondition = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FOutcome : TFhirCodeableConcept;
    FOnset : TFhirType;
    FNote : TFhirAnnotation;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetOutcome(value : TFhirCodeableConcept);
    Procedure SetOnset(value : TFhirType);
    Procedure SetNote(value : TFhirAnnotation);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirFamilyMemberHistoryCondition; overload;
    function Clone : TFhirFamilyMemberHistoryCondition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to The actual condition specified. Could be a coded condition (like MI or Diabetes) or a less specific string like 'cancer' depending on how much is known about the condition and the capabilities of the creating system. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      The actual condition specified. Could be a coded condition (like MI or Diabetes) or a less specific string like 'cancer' depending on how much is known about the condition and the capabilities of the creating system.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member outcome
      Typed access to Indicates what happened as a result of this condition.  If the condition resulted in death, deceased date is captured on the relation. (defined for API consistency)
    }
    property outcome : TFhirCodeableConcept read FOutcome write SetOutcome;
    {@member outcomeElement
      Indicates what happened as a result of this condition.  If the condition resulted in death, deceased date is captured on the relation.
    }
    property outcomeElement : TFhirCodeableConcept read FOutcome write SetOutcome;

    {@member onset
      Typed access to Either the age of onset, range of approximate age or descriptive string can be recorded.  For conditions with multiple occurrences, this describes the first known occurrence. (defined for API consistency)
    }
    property onset : TFhirType read FOnset write SetOnset;
    {@member onsetElement
      Either the age of onset, range of approximate age or descriptive string can be recorded.  For conditions with multiple occurrences, this describes the first known occurrence.
    }
    property onsetElement : TFhirType read FOnset write SetOnset;

    {@member note
      Typed access to An area where general notes can be placed about this specific condition. (defined for API consistency)
    }
    property note : TFhirAnnotation read FNote write SetNote;
    {@member noteElement
      An area where general notes can be placed about this specific condition.
    }
    property noteElement : TFhirAnnotation read FNote write SetNote;

  end;


  TFhirFamilyMemberHistoryConditionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirFamilyMemberHistoryConditionList;
    function GetCurrent : TFhirFamilyMemberHistoryCondition;
  public
    Constructor Create(list : TFhirFamilyMemberHistoryConditionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirFamilyMemberHistoryCondition read GetCurrent;
  end;


  {@Class TFhirFamilyMemberHistoryConditionList
    A list of FhirFamilyMemberHistoryCondition
  }
  {!.Net HL7Connect.Fhir.FamilyMemberHistoryConditionList}
  TFhirFamilyMemberHistoryConditionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirFamilyMemberHistoryCondition;
    procedure SetItemN(index : Integer; value : TFhirFamilyMemberHistoryCondition);
  public
    {!script hide}
    function Link : TFhirFamilyMemberHistoryConditionList; Overload;
    function Clone : TFhirFamilyMemberHistoryConditionList; Overload;
    function GetEnumerator : TFhirFamilyMemberHistoryConditionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirFamilyMemberHistoryCondition to the end of the list.
    }
    function Append : TFhirFamilyMemberHistoryCondition;

    
    {@member AddItem
      Add an already existing FhirFamilyMemberHistoryCondition to the end of the list.
    }
    procedure AddItem(value : TFhirFamilyMemberHistoryCondition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirFamilyMemberHistoryCondition) : Integer;
    

    {@member Insert
      Insert FhirFamilyMemberHistoryCondition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirFamilyMemberHistoryCondition;
    

    {@member InsertItem
       Insert an existing FhirFamilyMemberHistoryCondition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirFamilyMemberHistoryCondition);
    
    {@member Item
       Get the iIndexth FhirFamilyMemberHistoryCondition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirFamilyMemberHistoryCondition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirFamilyMemberHistoryCondition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirFamilyMemberHistoryCondition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirFamilyMemberHistoryConditions[index : Integer] : TFhirFamilyMemberHistoryCondition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirFamilyMemberHistory : TFhirDomainResource
    Significant health events and conditions for a person related to the patient relevant in the context of care for the patient.
  }
  {!.Net HL7Connect.Fhir.FamilyMemberHistory}
  TFhirFamilyMemberHistory = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FPatient : TFhirReference{TFhirPatient};
    FDate : TFhirDateTime;
    FStatus : TFhirEnum;
    FName : TFhirString;
    FRelationship : TFhirCodeableConcept;
    FGender : TFhirEnum;
    FBorn : TFhirType;
    FAge : TFhirType;
    FDeceased : TFhirType;
    FNote : TFhirAnnotation;
    FconditionList : TFhirFamilyMemberHistoryConditionList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirHistoryStatus;
    Procedure SetStatusST(value : TFhirHistoryStatus);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetRelationship(value : TFhirCodeableConcept);
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetBorn(value : TFhirType);
    Procedure SetAge(value : TFhirType);
    Procedure SetDeceased(value : TFhirType);
    Procedure SetNote(value : TFhirAnnotation);
    function GetConditionList : TFhirFamilyMemberHistoryConditionList;
    function GetHasConditionList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirFamilyMemberHistory; overload;
    function Clone : TFhirFamilyMemberHistory; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this family member history record that are defined by business processes and/ or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member patient
      Typed access to The person who this history concerns. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The person who this history concerns.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member date
      Typed access to The date (and possibly time) when the family member history was taken.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date (and possibly time) when the family member history was taken.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member status
      A code specifying a state of a Family Member History record.
    }
    property status : TFhirHistoryStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member name
      Typed access to This will either be a name or a description; e.g. "Aunt Susan", "my cousin with the red hair".
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      This will either be a name or a description; e.g. "Aunt Susan", "my cousin with the red hair".
    }
    property nameElement : TFhirString read FName write SetName;

    {@member relationship
      Typed access to The type of relationship this person has to the patient (father, mother, brother etc.). (defined for API consistency)
    }
    property relationship : TFhirCodeableConcept read FRelationship write SetRelationship;
    {@member relationshipElement
      The type of relationship this person has to the patient (father, mother, brother etc.).
    }
    property relationshipElement : TFhirCodeableConcept read FRelationship write SetRelationship;

    {@member gender
      Administrative Gender - the gender that the relative is considered to have for administration and record keeping purposes.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member born
      Typed access to The actual or approximate date of birth of the relative. (defined for API consistency)
    }
    property born : TFhirType read FBorn write SetBorn;
    {@member bornElement
      The actual or approximate date of birth of the relative.
    }
    property bornElement : TFhirType read FBorn write SetBorn;

    {@member age
      Typed access to The actual or approximate age of the relative at the time the family member history is recorded. (defined for API consistency)
    }
    property age : TFhirType read FAge write SetAge;
    {@member ageElement
      The actual or approximate age of the relative at the time the family member history is recorded.
    }
    property ageElement : TFhirType read FAge write SetAge;

    {@member deceased
      Typed access to Deceased flag or the actual or approximate age of the relative at the time of death for the family member history record. (defined for API consistency)
    }
    property deceased : TFhirType read FDeceased write SetDeceased;
    {@member deceasedElement
      Deceased flag or the actual or approximate age of the relative at the time of death for the family member history record.
    }
    property deceasedElement : TFhirType read FDeceased write SetDeceased;

    {@member note
      Typed access to This property allows a non condition-specific note to the made about the related person. Ideally, the note would be in the condition property, but this is not always possible. (defined for API consistency)
    }
    property note : TFhirAnnotation read FNote write SetNote;
    {@member noteElement
      This property allows a non condition-specific note to the made about the related person. Ideally, the note would be in the condition property, but this is not always possible.
    }
    property noteElement : TFhirAnnotation read FNote write SetNote;

    {@member conditionList
      The significant Conditions (or condition) that the family member had. This is a repeating section to allow a system to represent more than one condition per resource, though there is nothing stopping multiple resources - one per condition.
    }
    property conditionList : TFhirFamilyMemberHistoryConditionList read GetConditionList;
    property hasConditionList : boolean read GetHasConditionList;

  end;


  TFhirFamilyMemberHistoryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirFamilyMemberHistoryList;
    function GetCurrent : TFhirFamilyMemberHistory;
  public
    Constructor Create(list : TFhirFamilyMemberHistoryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirFamilyMemberHistory read GetCurrent;
  end;


  {@Class TFhirFamilyMemberHistoryList
    A list of FhirFamilyMemberHistory
  }
  {!.Net HL7Connect.Fhir.FamilyMemberHistoryList}
  TFhirFamilyMemberHistoryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirFamilyMemberHistory;
    procedure SetItemN(index : Integer; value : TFhirFamilyMemberHistory);
  public
    {!script hide}
    function Link : TFhirFamilyMemberHistoryList; Overload;
    function Clone : TFhirFamilyMemberHistoryList; Overload;
    function GetEnumerator : TFhirFamilyMemberHistoryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirFamilyMemberHistory to the end of the list.
    }
    function Append : TFhirFamilyMemberHistory;

    
    {@member AddItem
      Add an already existing FhirFamilyMemberHistory to the end of the list.
    }
    procedure AddItem(value : TFhirFamilyMemberHistory); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirFamilyMemberHistory) : Integer;
    

    {@member Insert
      Insert FhirFamilyMemberHistory before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirFamilyMemberHistory;
    

    {@member InsertItem
       Insert an existing FhirFamilyMemberHistory before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirFamilyMemberHistory);
    
    {@member Item
       Get the iIndexth FhirFamilyMemberHistory. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirFamilyMemberHistory. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirFamilyMemberHistory);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirFamilyMemberHistory;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirFamilyMemberHistories[index : Integer] : TFhirFamilyMemberHistory read GetItemN write SetItemN; default;
  End;


  {@Class TFhirFlag : TFhirDomainResource
    Prospective warnings of potential issues when providing care to the patient.
  }
  {!.Net HL7Connect.Fhir.Flag}
  TFhirFlag = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FCategory : TFhirCodeableConcept;
    FStatus : TFhirEnum;
    FPeriod : TFhirPeriod;
    FSubject : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FAuthor : TFhirReference{Resource};
    FCode : TFhirCodeableConcept;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirFlagStatus;
    Procedure SetStatusST(value : TFhirFlagStatus);
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetCode(value : TFhirCodeableConcept);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirFlag; overload;
    function Clone : TFhirFlag; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier assigned to the flag for external use (outside the FHIR environment).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member category
      Typed access to Allows an flag to be divided into different categories like clinical, administrative etc. Intended to be used as a means of filtering which flags are displayed to particular user or in a given context. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      Allows an flag to be divided into different categories like clinical, administrative etc. Intended to be used as a means of filtering which flags are displayed to particular user or in a given context.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member status
      Supports basic workflow.
    }
    property status : TFhirFlagStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member period
      Typed access to The period of time from the activation of the flag to inactivation of the flag. If the flag is active, the end of the period should be unspecified. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time from the activation of the flag to inactivation of the flag. If the flag is active, the end of the period should be unspecified.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member subject
      Typed access to The patient, location, group , organization , or practitioner this is about record this flag is associated with. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The patient, location, group , organization , or practitioner this is about record this flag is associated with.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member encounter
      Typed access to This alert is only relevant during the encounter. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      This alert is only relevant during the encounter.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member author
      Typed access to The person, organization or device that created the flag. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      The person, organization or device that created the flag.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member code
      Typed access to The coded value or textual component of the flag to display to the user. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      The coded value or textual component of the flag to display to the user.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

  end;


  TFhirFlagListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirFlagList;
    function GetCurrent : TFhirFlag;
  public
    Constructor Create(list : TFhirFlagList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirFlag read GetCurrent;
  end;


  {@Class TFhirFlagList
    A list of FhirFlag
  }
  {!.Net HL7Connect.Fhir.FlagList}
  TFhirFlagList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirFlag;
    procedure SetItemN(index : Integer; value : TFhirFlag);
  public
    {!script hide}
    function Link : TFhirFlagList; Overload;
    function Clone : TFhirFlagList; Overload;
    function GetEnumerator : TFhirFlagListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirFlag to the end of the list.
    }
    function Append : TFhirFlag;

    
    {@member AddItem
      Add an already existing FhirFlag to the end of the list.
    }
    procedure AddItem(value : TFhirFlag); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirFlag) : Integer;
    

    {@member Insert
      Insert FhirFlag before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirFlag;
    

    {@member InsertItem
       Insert an existing FhirFlag before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirFlag);
    
    {@member Item
       Get the iIndexth FhirFlag. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirFlag. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirFlag);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirFlag;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirFlags[index : Integer] : TFhirFlag read GetItemN write SetItemN; default;
  End;


  {@Class TFhirGoalOutcome : TFhirElement
    Identifies the change (or lack of change) at the point where the goal was deepmed to be cancelled or achieved.
  }
  {!.Net HL7Connect.Fhir.GoalOutcome}
  TFhirGoalOutcome = class (TFhirBackboneElement)
  private
    FResult : TFhirType;
    Procedure SetResult(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirGoalOutcome; overload;
    function Clone : TFhirGoalOutcome; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member result
      Typed access to Details of what's changed (or not changed). (defined for API consistency)
    }
    property result : TFhirType read FResult write SetResult;
    {@member resultElement
      Details of what's changed (or not changed).
    }
    property resultElement : TFhirType read FResult write SetResult;

  end;


  TFhirGoalOutcomeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirGoalOutcomeList;
    function GetCurrent : TFhirGoalOutcome;
  public
    Constructor Create(list : TFhirGoalOutcomeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirGoalOutcome read GetCurrent;
  end;


  {@Class TFhirGoalOutcomeList
    A list of FhirGoalOutcome
  }
  {!.Net HL7Connect.Fhir.GoalOutcomeList}
  TFhirGoalOutcomeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirGoalOutcome;
    procedure SetItemN(index : Integer; value : TFhirGoalOutcome);
  public
    {!script hide}
    function Link : TFhirGoalOutcomeList; Overload;
    function Clone : TFhirGoalOutcomeList; Overload;
    function GetEnumerator : TFhirGoalOutcomeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirGoalOutcome to the end of the list.
    }
    function Append : TFhirGoalOutcome;

    
    {@member AddItem
      Add an already existing FhirGoalOutcome to the end of the list.
    }
    procedure AddItem(value : TFhirGoalOutcome); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirGoalOutcome) : Integer;
    

    {@member Insert
      Insert FhirGoalOutcome before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirGoalOutcome;
    

    {@member InsertItem
       Insert an existing FhirGoalOutcome before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirGoalOutcome);
    
    {@member Item
       Get the iIndexth FhirGoalOutcome. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirGoalOutcome. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirGoalOutcome);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirGoalOutcome;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirGoalOutcomes[index : Integer] : TFhirGoalOutcome read GetItemN write SetItemN; default;
  End;


  {@Class TFhirGoal : TFhirDomainResource
    Describes the intended objective(s) for a patient, group or organization care, for example, weight loss, restoring an activity of daily living, obtaining herd immunity via immunization, meeting a process improvement objective, etc.
  }
  {!.Net HL7Connect.Fhir.Goal}
  TFhirGoal = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FStart : TFhirType;
    FTarget : TFhirType;
    FcategoryList : TFhirCodeableConceptList;
    FDescription : TFhirString;
    FStatus : TFhirEnum;
    FStatusDate : TFhirDate;
    FStatusReason : TFhirCodeableConcept;
    FAuthor : TFhirReference{Resource};
    FPriority : TFhirCodeableConcept;
    FaddressesList : TFhirReferenceList{Resource};
    FnoteList : TFhirAnnotationList;
    FoutcomeList : TFhirGoalOutcomeList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetStart(value : TFhirType);
    Procedure SetTarget(value : TFhirType);
    function GetCategoryList : TFhirCodeableConceptList;
    function GetHasCategoryList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirGoalStatus;
    Procedure SetStatusST(value : TFhirGoalStatus);
    Procedure SetStatusDate(value : TFhirDate);
    Function GetStatusDateST : TDateAndTime;
    Procedure SetStatusDateST(value : TDateAndTime);
    Procedure SetStatusReason(value : TFhirCodeableConcept);
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetPriority(value : TFhirCodeableConcept);
    function GetAddressesList : TFhirReferenceList{Resource};
    function GetHasAddressesList : Boolean;
    function GetNoteList : TFhirAnnotationList;
    function GetHasNoteList : Boolean;
    function GetOutcomeList : TFhirGoalOutcomeList;
    function GetHasOutcomeList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirGoal; overload;
    function Clone : TFhirGoal; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this care plan that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to Identifies the patient, group or organization for whom the goal is being established. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Identifies the patient, group or organization for whom the goal is being established.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member start
      Typed access to The date or event after which the goal should begin being pursued. (defined for API consistency)
    }
    property start : TFhirType read FStart write SetStart;
    {@member startElement
      The date or event after which the goal should begin being pursued.
    }
    property startElement : TFhirType read FStart write SetStart;

    {@member target
      Typed access to Indicates either the date or the duration after start by which the goal should be met. (defined for API consistency)
    }
    property target : TFhirType read FTarget write SetTarget;
    {@member targetElement
      Indicates either the date or the duration after start by which the goal should be met.
    }
    property targetElement : TFhirType read FTarget write SetTarget;

    {@member categoryList
      Indicates a category the goal falls within.
    }
    property categoryList : TFhirCodeableConceptList read GetCategoryList;
    property hasCategoryList : boolean read GetHasCategoryList;

    {@member description
      Typed access to Human-readable description of a specific desired objective of care.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Human-readable description of a specific desired objective of care.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member status
      Indicates whether the goal has been reached and is still considered relevant.
    }
    property status : TFhirGoalStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member statusDate
      Typed access to Identifies when the current status.  I.e. When initially created, when achieved, when cancelled, etc.
    }
    property statusDate : TDateAndTime read GetStatusDateST write SetStatusDateST;
    {@member statusDateElement
      Identifies when the current status.  I.e. When initially created, when achieved, when cancelled, etc.
    }
    property statusDateElement : TFhirDate read FStatusDate write SetStatusDate;

    {@member statusReason
      Typed access to Captures the reason for the current status. (defined for API consistency)
    }
    property statusReason : TFhirCodeableConcept read FStatusReason write SetStatusReason;
    {@member statusReasonElement
      Captures the reason for the current status.
    }
    property statusReasonElement : TFhirCodeableConcept read FStatusReason write SetStatusReason;

    {@member author
      Typed access to Indicates whose goal this is - patient goal, practitioner goal, etc. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      Indicates whose goal this is - patient goal, practitioner goal, etc.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member priority
      Typed access to Identifies the mutually agreed level of importance associated with reaching/sustaining the goal. (defined for API consistency)
    }
    property priority : TFhirCodeableConcept read FPriority write SetPriority;
    {@member priorityElement
      Identifies the mutually agreed level of importance associated with reaching/sustaining the goal.
    }
    property priorityElement : TFhirCodeableConcept read FPriority write SetPriority;

    {@member addressesList
      The identified conditions and other health record elements that are intended to be addressed by the goal.
    }
    property addressesList : TFhirReferenceList{Resource} read GetAddressesList;
    property hasAddressesList : boolean read GetHasAddressesList;

    {@member noteList
      Any comments related to the goal.
    }
    property noteList : TFhirAnnotationList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

    {@member outcomeList
      Identifies the change (or lack of change) at the point where the goal was deepmed to be cancelled or achieved.
    }
    property outcomeList : TFhirGoalOutcomeList read GetOutcomeList;
    property hasOutcomeList : boolean read GetHasOutcomeList;

  end;


  TFhirGoalListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirGoalList;
    function GetCurrent : TFhirGoal;
  public
    Constructor Create(list : TFhirGoalList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirGoal read GetCurrent;
  end;


  {@Class TFhirGoalList
    A list of FhirGoal
  }
  {!.Net HL7Connect.Fhir.GoalList}
  TFhirGoalList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirGoal;
    procedure SetItemN(index : Integer; value : TFhirGoal);
  public
    {!script hide}
    function Link : TFhirGoalList; Overload;
    function Clone : TFhirGoalList; Overload;
    function GetEnumerator : TFhirGoalListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirGoal to the end of the list.
    }
    function Append : TFhirGoal;

    
    {@member AddItem
      Add an already existing FhirGoal to the end of the list.
    }
    procedure AddItem(value : TFhirGoal); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirGoal) : Integer;
    

    {@member Insert
      Insert FhirGoal before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirGoal;
    

    {@member InsertItem
       Insert an existing FhirGoal before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirGoal);
    
    {@member Item
       Get the iIndexth FhirGoal. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirGoal. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirGoal);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirGoal;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirGoals[index : Integer] : TFhirGoal read GetItemN write SetItemN; default;
  End;


  {@Class TFhirGroupCharacteristic : TFhirElement
    Identifies the traits shared by members of the group.
  }
  {!.Net HL7Connect.Fhir.GroupCharacteristic}
  TFhirGroupCharacteristic = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FValue : TFhirType;
    FExclude : TFhirBoolean;
    FPeriod : TFhirPeriod;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetValue(value : TFhirType);
    Procedure SetExclude(value : TFhirBoolean);
    Function GetExcludeST : Boolean;
    Procedure SetExcludeST(value : Boolean);
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirGroupCharacteristic; overload;
    function Clone : TFhirGroupCharacteristic; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A code that identifies the kind of trait being asserted. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code that identifies the kind of trait being asserted.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member value
      Typed access to The value of the trait that holds (or does not hold - see 'exclude') for members of the group. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      The value of the trait that holds (or does not hold - see 'exclude') for members of the group.
    }
    property valueElement : TFhirType read FValue write SetValue;

    {@member exclude
      Typed access to If true, indicates the characteristic is one that is NOT held by members of the group.
    }
    property exclude : Boolean read GetExcludeST write SetExcludeST;
    {@member excludeElement
      If true, indicates the characteristic is one that is NOT held by members of the group.
    }
    property excludeElement : TFhirBoolean read FExclude write SetExclude;

    {@member period
      Typed access to The period over which the characteristic is tested; e.g. the patient had an operation during the month of June. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period over which the characteristic is tested; e.g. the patient had an operation during the month of June.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirGroupCharacteristicListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirGroupCharacteristicList;
    function GetCurrent : TFhirGroupCharacteristic;
  public
    Constructor Create(list : TFhirGroupCharacteristicList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirGroupCharacteristic read GetCurrent;
  end;


  {@Class TFhirGroupCharacteristicList
    A list of FhirGroupCharacteristic
  }
  {!.Net HL7Connect.Fhir.GroupCharacteristicList}
  TFhirGroupCharacteristicList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirGroupCharacteristic;
    procedure SetItemN(index : Integer; value : TFhirGroupCharacteristic);
  public
    {!script hide}
    function Link : TFhirGroupCharacteristicList; Overload;
    function Clone : TFhirGroupCharacteristicList; Overload;
    function GetEnumerator : TFhirGroupCharacteristicListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirGroupCharacteristic to the end of the list.
    }
    function Append : TFhirGroupCharacteristic;

    
    {@member AddItem
      Add an already existing FhirGroupCharacteristic to the end of the list.
    }
    procedure AddItem(value : TFhirGroupCharacteristic); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirGroupCharacteristic) : Integer;
    

    {@member Insert
      Insert FhirGroupCharacteristic before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirGroupCharacteristic;
    

    {@member InsertItem
       Insert an existing FhirGroupCharacteristic before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirGroupCharacteristic);
    
    {@member Item
       Get the iIndexth FhirGroupCharacteristic. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirGroupCharacteristic. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirGroupCharacteristic);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirGroupCharacteristic;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirGroupCharacteristics[index : Integer] : TFhirGroupCharacteristic read GetItemN write SetItemN; default;
  End;


  {@Class TFhirGroupMember : TFhirElement
    Identifies the resource instances that are members of the group.
  }
  {!.Net HL7Connect.Fhir.GroupMember}
  TFhirGroupMember = class (TFhirBackboneElement)
  private
    FEntity : TFhirReference{Resource};
    FPeriod : TFhirPeriod;
    FInactive : TFhirBoolean;
    Procedure SetEntity(value : TFhirReference{Resource});
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetInactive(value : TFhirBoolean);
    Function GetInactiveST : Boolean;
    Procedure SetInactiveST(value : Boolean);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirGroupMember; overload;
    function Clone : TFhirGroupMember; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member entity
      Typed access to A reference to the entity that is a member of the group. Must be consistent with Group.type. (defined for API consistency)
    }
    property entity : TFhirReference{Resource} read FEntity write SetEntity;
    {@member entityElement
      A reference to the entity that is a member of the group. Must be consistent with Group.type.
    }
    property entityElement : TFhirReference{Resource} read FEntity write SetEntity;

    {@member period
      Typed access to The period that the member was in the group, if known. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period that the member was in the group, if known.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member inactive
      Typed access to A flag to indicate that the member is no longer in the group, but previously may have been a member.
    }
    property inactive : Boolean read GetInactiveST write SetInactiveST;
    {@member inactiveElement
      A flag to indicate that the member is no longer in the group, but previously may have been a member.
    }
    property inactiveElement : TFhirBoolean read FInactive write SetInactive;

  end;


  TFhirGroupMemberListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirGroupMemberList;
    function GetCurrent : TFhirGroupMember;
  public
    Constructor Create(list : TFhirGroupMemberList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirGroupMember read GetCurrent;
  end;


  {@Class TFhirGroupMemberList
    A list of FhirGroupMember
  }
  {!.Net HL7Connect.Fhir.GroupMemberList}
  TFhirGroupMemberList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirGroupMember;
    procedure SetItemN(index : Integer; value : TFhirGroupMember);
  public
    {!script hide}
    function Link : TFhirGroupMemberList; Overload;
    function Clone : TFhirGroupMemberList; Overload;
    function GetEnumerator : TFhirGroupMemberListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirGroupMember to the end of the list.
    }
    function Append : TFhirGroupMember;

    
    {@member AddItem
      Add an already existing FhirGroupMember to the end of the list.
    }
    procedure AddItem(value : TFhirGroupMember); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirGroupMember) : Integer;
    

    {@member Insert
      Insert FhirGroupMember before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirGroupMember;
    

    {@member InsertItem
       Insert an existing FhirGroupMember before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirGroupMember);
    
    {@member Item
       Get the iIndexth FhirGroupMember. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirGroupMember. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirGroupMember);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirGroupMember;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirGroupMembers[index : Integer] : TFhirGroupMember read GetItemN write SetItemN; default;
  End;


  {@Class TFhirGroup : TFhirDomainResource
    Represents a defined collection of entities that may be discussed or acted upon collectively but which are not expected to act collectively and are not formally or legally recognized; i.e. a collection of entities that isn't an Organization.
  }
  {!.Net HL7Connect.Fhir.Group}
  TFhirGroup = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FType_ : TFhirEnum;
    FActual : TFhirBoolean;
    FCode : TFhirCodeableConcept;
    FName : TFhirString;
    FQuantity : TFhirUnsignedInt;
    FcharacteristicList : TFhirGroupCharacteristicList;
    FmemberList : TFhirGroupMemberList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirGroupType;
    Procedure SetType_ST(value : TFhirGroupType);
    Procedure SetActual(value : TFhirBoolean);
    Function GetActualST : Boolean;
    Procedure SetActualST(value : Boolean);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetQuantity(value : TFhirUnsignedInt);
    Function GetQuantityST : String;
    Procedure SetQuantityST(value : String);
    function GetCharacteristicList : TFhirGroupCharacteristicList;
    function GetHasCharacteristicList : Boolean;
    function GetMemberList : TFhirGroupMemberList;
    function GetHasMemberList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirGroup; overload;
    function Clone : TFhirGroup; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      A unique business identifier for this group.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member type_
      Identifies the broad classification of the kind of resources the group includes.
    }
    property type_ : TFhirGroupType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member actual
      Typed access to If true, indicates that the resource refers to a specific group of real individuals.  If false, the group defines a set of intended individuals.
    }
    property actual : Boolean read GetActualST write SetActualST;
    {@member actualElement
      If true, indicates that the resource refers to a specific group of real individuals.  If false, the group defines a set of intended individuals.
    }
    property actualElement : TFhirBoolean read FActual write SetActual;

    {@member code
      Typed access to Provides a specific type of resource the group includes; e.g. "cow", "syringe", etc. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Provides a specific type of resource the group includes; e.g. "cow", "syringe", etc.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member name
      Typed access to A label assigned to the group for human identification and communication.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A label assigned to the group for human identification and communication.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member quantity
      Typed access to A count of the number of resource instances that are part of the group.
    }
    property quantity : String read GetQuantityST write SetQuantityST;
    {@member quantityElement
      A count of the number of resource instances that are part of the group.
    }
    property quantityElement : TFhirUnsignedInt read FQuantity write SetQuantity;

    {@member characteristicList
      Identifies the traits shared by members of the group.
    }
    property characteristicList : TFhirGroupCharacteristicList read GetCharacteristicList;
    property hasCharacteristicList : boolean read GetHasCharacteristicList;

    {@member memberList
      Identifies the resource instances that are members of the group.
    }
    property memberList : TFhirGroupMemberList read GetMemberList;
    property hasMemberList : boolean read GetHasMemberList;

  end;


  TFhirGroupListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirGroupList;
    function GetCurrent : TFhirGroup;
  public
    Constructor Create(list : TFhirGroupList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirGroup read GetCurrent;
  end;


  {@Class TFhirGroupList
    A list of FhirGroup
  }
  {!.Net HL7Connect.Fhir.GroupList}
  TFhirGroupList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirGroup;
    procedure SetItemN(index : Integer; value : TFhirGroup);
  public
    {!script hide}
    function Link : TFhirGroupList; Overload;
    function Clone : TFhirGroupList; Overload;
    function GetEnumerator : TFhirGroupListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirGroup to the end of the list.
    }
    function Append : TFhirGroup;

    
    {@member AddItem
      Add an already existing FhirGroup to the end of the list.
    }
    procedure AddItem(value : TFhirGroup); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirGroup) : Integer;
    

    {@member Insert
      Insert FhirGroup before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirGroup;
    

    {@member InsertItem
       Insert an existing FhirGroup before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirGroup);
    
    {@member Item
       Get the iIndexth FhirGroup. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirGroup. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirGroup);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirGroup;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirGroups[index : Integer] : TFhirGroup read GetItemN write SetItemN; default;
  End;


  {@Class TFhirHealthcareServiceServiceType : TFhirElement
    A specific type of service that may be delivered or performed.
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceServiceType}
  TFhirHealthcareServiceServiceType = class (TFhirBackboneElement)
  private
    FType_ : TFhirCodeableConcept;
    FspecialtyList : TFhirCodeableConceptList;
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetSpecialtyList : TFhirCodeableConceptList;
    function GetHasSpecialtyList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirHealthcareServiceServiceType; overload;
    function Clone : TFhirHealthcareServiceServiceType; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The specific type of service being delivered or performed. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The specific type of service being delivered or performed.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member specialtyList
      Collection of specialties handled by the service site. This is more of a medical term.
    }
    property specialtyList : TFhirCodeableConceptList read GetSpecialtyList;
    property hasSpecialtyList : boolean read GetHasSpecialtyList;

  end;


  TFhirHealthcareServiceServiceTypeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirHealthcareServiceServiceTypeList;
    function GetCurrent : TFhirHealthcareServiceServiceType;
  public
    Constructor Create(list : TFhirHealthcareServiceServiceTypeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirHealthcareServiceServiceType read GetCurrent;
  end;


  {@Class TFhirHealthcareServiceServiceTypeList
    A list of FhirHealthcareServiceServiceType
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceServiceTypeList}
  TFhirHealthcareServiceServiceTypeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirHealthcareServiceServiceType;
    procedure SetItemN(index : Integer; value : TFhirHealthcareServiceServiceType);
  public
    {!script hide}
    function Link : TFhirHealthcareServiceServiceTypeList; Overload;
    function Clone : TFhirHealthcareServiceServiceTypeList; Overload;
    function GetEnumerator : TFhirHealthcareServiceServiceTypeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirHealthcareServiceServiceType to the end of the list.
    }
    function Append : TFhirHealthcareServiceServiceType;

    
    {@member AddItem
      Add an already existing FhirHealthcareServiceServiceType to the end of the list.
    }
    procedure AddItem(value : TFhirHealthcareServiceServiceType); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirHealthcareServiceServiceType) : Integer;
    

    {@member Insert
      Insert FhirHealthcareServiceServiceType before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirHealthcareServiceServiceType;
    

    {@member InsertItem
       Insert an existing FhirHealthcareServiceServiceType before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirHealthcareServiceServiceType);
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceServiceType. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceServiceType. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirHealthcareServiceServiceType);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirHealthcareServiceServiceType;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirHealthcareServiceServiceTypes[index : Integer] : TFhirHealthcareServiceServiceType read GetItemN write SetItemN; default;
  End;


  {@Class TFhirHealthcareServiceAvailableTime : TFhirElement
    A collection of times that the Service Site is available.
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceAvailableTime}
  TFhirHealthcareServiceAvailableTime = class (TFhirBackboneElement)
  private
    FDaysOfWeek : TFhirEnumList;
    FAllDay : TFhirBoolean;
    FAvailableStartTime : TFhirTime;
    FAvailableEndTime : TFhirTime;
    function GetDaysOfWeek : TFhirEnumList;
    function GetHasDaysOfWeek : Boolean;
    Function GetDaysOfWeekST : TFhirDaysOfWeekList;
    Procedure SetDaysOfWeekST(value : TFhirDaysOfWeekList);
    Procedure SetAllDay(value : TFhirBoolean);
    Function GetAllDayST : Boolean;
    Procedure SetAllDayST(value : Boolean);
    Procedure SetAvailableStartTime(value : TFhirTime);
    Function GetAvailableStartTimeST : String;
    Procedure SetAvailableStartTimeST(value : String);
    Procedure SetAvailableEndTime(value : TFhirTime);
    Function GetAvailableEndTimeST : String;
    Procedure SetAvailableEndTimeST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirHealthcareServiceAvailableTime; overload;
    function Clone : TFhirHealthcareServiceAvailableTime; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member daysOfWeek
      Indicates which days of the week are available between the start and end Times.
    }
    property daysOfWeek : TFhirDaysOfWeekList read GetDaysOfWeekST write SetDaysOfWeekST;
    property daysOfWeekList : TFhirEnumList read GetDaysOfWeek;
    property hasDaysOfWeek : boolean read GetHasDaysOfWeek;
    {@member allDay
      Typed access to Is this always available? (hence times are irrelevant) e.g. 24 hour service.
    }
    property allDay : Boolean read GetAllDayST write SetAllDayST;
    {@member allDayElement
      Is this always available? (hence times are irrelevant) e.g. 24 hour service.
    }
    property allDayElement : TFhirBoolean read FAllDay write SetAllDay;

    {@member availableStartTime
      Typed access to The opening time of day. Note: If the AllDay flag is set, then this time is ignored.
    }
    property availableStartTime : String read GetAvailableStartTimeST write SetAvailableStartTimeST;
    {@member availableStartTimeElement
      The opening time of day. Note: If the AllDay flag is set, then this time is ignored.
    }
    property availableStartTimeElement : TFhirTime read FAvailableStartTime write SetAvailableStartTime;

    {@member availableEndTime
      Typed access to The closing time of day. Note: If the AllDay flag is set, then this time is ignored.
    }
    property availableEndTime : String read GetAvailableEndTimeST write SetAvailableEndTimeST;
    {@member availableEndTimeElement
      The closing time of day. Note: If the AllDay flag is set, then this time is ignored.
    }
    property availableEndTimeElement : TFhirTime read FAvailableEndTime write SetAvailableEndTime;

  end;


  TFhirHealthcareServiceAvailableTimeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirHealthcareServiceAvailableTimeList;
    function GetCurrent : TFhirHealthcareServiceAvailableTime;
  public
    Constructor Create(list : TFhirHealthcareServiceAvailableTimeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirHealthcareServiceAvailableTime read GetCurrent;
  end;


  {@Class TFhirHealthcareServiceAvailableTimeList
    A list of FhirHealthcareServiceAvailableTime
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceAvailableTimeList}
  TFhirHealthcareServiceAvailableTimeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirHealthcareServiceAvailableTime;
    procedure SetItemN(index : Integer; value : TFhirHealthcareServiceAvailableTime);
  public
    {!script hide}
    function Link : TFhirHealthcareServiceAvailableTimeList; Overload;
    function Clone : TFhirHealthcareServiceAvailableTimeList; Overload;
    function GetEnumerator : TFhirHealthcareServiceAvailableTimeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirHealthcareServiceAvailableTime to the end of the list.
    }
    function Append : TFhirHealthcareServiceAvailableTime;

    
    {@member AddItem
      Add an already existing FhirHealthcareServiceAvailableTime to the end of the list.
    }
    procedure AddItem(value : TFhirHealthcareServiceAvailableTime); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirHealthcareServiceAvailableTime) : Integer;
    

    {@member Insert
      Insert FhirHealthcareServiceAvailableTime before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirHealthcareServiceAvailableTime;
    

    {@member InsertItem
       Insert an existing FhirHealthcareServiceAvailableTime before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirHealthcareServiceAvailableTime);
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceAvailableTime. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceAvailableTime. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirHealthcareServiceAvailableTime);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirHealthcareServiceAvailableTime;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirHealthcareServiceAvailableTimes[index : Integer] : TFhirHealthcareServiceAvailableTime read GetItemN write SetItemN; default;
  End;


  {@Class TFhirHealthcareServiceNotAvailable : TFhirElement
    The HealthcareService is not available during this period of time due to the provided reason.
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceNotAvailable}
  TFhirHealthcareServiceNotAvailable = class (TFhirBackboneElement)
  private
    FDescription : TFhirString;
    FDuring : TFhirPeriod;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetDuring(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirHealthcareServiceNotAvailable; overload;
    function Clone : TFhirHealthcareServiceNotAvailable; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member description
      Typed access to The reason that can be presented to the user as to why this time is not available.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      The reason that can be presented to the user as to why this time is not available.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member during
      Typed access to Service is not available (seasonally or for a public holiday) from this date. (defined for API consistency)
    }
    property during : TFhirPeriod read FDuring write SetDuring;
    {@member duringElement
      Service is not available (seasonally or for a public holiday) from this date.
    }
    property duringElement : TFhirPeriod read FDuring write SetDuring;

  end;


  TFhirHealthcareServiceNotAvailableListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirHealthcareServiceNotAvailableList;
    function GetCurrent : TFhirHealthcareServiceNotAvailable;
  public
    Constructor Create(list : TFhirHealthcareServiceNotAvailableList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirHealthcareServiceNotAvailable read GetCurrent;
  end;


  {@Class TFhirHealthcareServiceNotAvailableList
    A list of FhirHealthcareServiceNotAvailable
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceNotAvailableList}
  TFhirHealthcareServiceNotAvailableList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirHealthcareServiceNotAvailable;
    procedure SetItemN(index : Integer; value : TFhirHealthcareServiceNotAvailable);
  public
    {!script hide}
    function Link : TFhirHealthcareServiceNotAvailableList; Overload;
    function Clone : TFhirHealthcareServiceNotAvailableList; Overload;
    function GetEnumerator : TFhirHealthcareServiceNotAvailableListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirHealthcareServiceNotAvailable to the end of the list.
    }
    function Append : TFhirHealthcareServiceNotAvailable;

    
    {@member AddItem
      Add an already existing FhirHealthcareServiceNotAvailable to the end of the list.
    }
    procedure AddItem(value : TFhirHealthcareServiceNotAvailable); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirHealthcareServiceNotAvailable) : Integer;
    

    {@member Insert
      Insert FhirHealthcareServiceNotAvailable before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirHealthcareServiceNotAvailable;
    

    {@member InsertItem
       Insert an existing FhirHealthcareServiceNotAvailable before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirHealthcareServiceNotAvailable);
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceNotAvailable. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirHealthcareServiceNotAvailable. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirHealthcareServiceNotAvailable);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirHealthcareServiceNotAvailable;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirHealthcareServiceNotAvailables[index : Integer] : TFhirHealthcareServiceNotAvailable read GetItemN write SetItemN; default;
  End;


  {@Class TFhirHealthcareService : TFhirDomainResource
    The details of a healthcare service available at a location.
  }
  {!.Net HL7Connect.Fhir.HealthcareService}
  TFhirHealthcareService = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FProvidedBy : TFhirReference{TFhirOrganization};
    FServiceCategory : TFhirCodeableConcept;
    FserviceTypeList : TFhirHealthcareServiceServiceTypeList;
    FLocation : TFhirReference{TFhirLocation};
    FServiceName : TFhirString;
    FComment : TFhirString;
    FExtraDetails : TFhirString;
    FPhoto : TFhirAttachment;
    FtelecomList : TFhirContactPointList;
    FcoverageAreaList : TFhirReferenceList{TFhirLocation};
    FserviceProvisionCodeList : TFhirCodeableConceptList;
    FEligibility : TFhirCodeableConcept;
    FEligibilityNote : TFhirString;
    FprogramNameList : TFhirStringList;
    FcharacteristicList : TFhirCodeableConceptList;
    FreferralMethodList : TFhirCodeableConceptList;
    FPublicKey : TFhirString;
    FAppointmentRequired : TFhirBoolean;
    FavailableTimeList : TFhirHealthcareServiceAvailableTimeList;
    FnotAvailableList : TFhirHealthcareServiceNotAvailableList;
    FAvailabilityExceptions : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetProvidedBy(value : TFhirReference{TFhirOrganization});
    Procedure SetServiceCategory(value : TFhirCodeableConcept);
    function GetServiceTypeList : TFhirHealthcareServiceServiceTypeList;
    function GetHasServiceTypeList : Boolean;
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    Procedure SetServiceName(value : TFhirString);
    Function GetServiceNameST : String;
    Procedure SetServiceNameST(value : String);
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
    Procedure SetExtraDetails(value : TFhirString);
    Function GetExtraDetailsST : String;
    Procedure SetExtraDetailsST(value : String);
    Procedure SetPhoto(value : TFhirAttachment);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    function GetCoverageAreaList : TFhirReferenceList{TFhirLocation};
    function GetHasCoverageAreaList : Boolean;
    function GetServiceProvisionCodeList : TFhirCodeableConceptList;
    function GetHasServiceProvisionCodeList : Boolean;
    Procedure SetEligibility(value : TFhirCodeableConcept);
    Procedure SetEligibilityNote(value : TFhirString);
    Function GetEligibilityNoteST : String;
    Procedure SetEligibilityNoteST(value : String);
    function GetProgramNameList : TFhirStringList;
    function GetHasProgramNameList : Boolean;
    function GetCharacteristicList : TFhirCodeableConceptList;
    function GetHasCharacteristicList : Boolean;
    function GetReferralMethodList : TFhirCodeableConceptList;
    function GetHasReferralMethodList : Boolean;
    Procedure SetPublicKey(value : TFhirString);
    Function GetPublicKeyST : String;
    Procedure SetPublicKeyST(value : String);
    Procedure SetAppointmentRequired(value : TFhirBoolean);
    Function GetAppointmentRequiredST : Boolean;
    Procedure SetAppointmentRequiredST(value : Boolean);
    function GetAvailableTimeList : TFhirHealthcareServiceAvailableTimeList;
    function GetHasAvailableTimeList : Boolean;
    function GetNotAvailableList : TFhirHealthcareServiceNotAvailableList;
    function GetHasNotAvailableList : Boolean;
    Procedure SetAvailabilityExceptions(value : TFhirString);
    Function GetAvailabilityExceptionsST : String;
    Procedure SetAvailabilityExceptionsST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirHealthcareService; overload;
    function Clone : TFhirHealthcareService; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External identifiers for this item.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member providedBy
      Typed access to The organization that provides this healthcare service. (defined for API consistency)
    }
    property providedBy : TFhirReference{TFhirOrganization} read FProvidedBy write SetProvidedBy;
    {@member providedByElement
      The organization that provides this healthcare service.
    }
    property providedByElement : TFhirReference{TFhirOrganization} read FProvidedBy write SetProvidedBy;

    {@member serviceCategory
      Typed access to Identifies the broad category of service being performed or delivered. (defined for API consistency)
    }
    property serviceCategory : TFhirCodeableConcept read FServiceCategory write SetServiceCategory;
    {@member serviceCategoryElement
      Identifies the broad category of service being performed or delivered.
    }
    property serviceCategoryElement : TFhirCodeableConcept read FServiceCategory write SetServiceCategory;

    {@member serviceTypeList
      A specific type of service that may be delivered or performed.
    }
    property serviceTypeList : TFhirHealthcareServiceServiceTypeList read GetServiceTypeList;
    property hasServiceTypeList : boolean read GetHasServiceTypeList;

    {@member location
      Typed access to The location where this healthcare service may be provided. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      The location where this healthcare service may be provided.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member serviceName
      Typed access to Further description of the service as it would be presented to a consumer while searching.
    }
    property serviceName : String read GetServiceNameST write SetServiceNameST;
    {@member serviceNameElement
      Further description of the service as it would be presented to a consumer while searching.
    }
    property serviceNameElement : TFhirString read FServiceName write SetServiceName;

    {@member comment
      Typed access to Any additional description of the service and/or any specific issues not covered by the other attributes, which can be displayed as further detail under the serviceName.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      Any additional description of the service and/or any specific issues not covered by the other attributes, which can be displayed as further detail under the serviceName.
    }
    property commentElement : TFhirString read FComment write SetComment;

    {@member extraDetails
      Typed access to Extra details about the service that can't be placed in the other fields.
    }
    property extraDetails : String read GetExtraDetailsST write SetExtraDetailsST;
    {@member extraDetailsElement
      Extra details about the service that can't be placed in the other fields.
    }
    property extraDetailsElement : TFhirString read FExtraDetails write SetExtraDetails;

    {@member photo
      Typed access to If there is a photo/symbol associated with this HealthcareService, it may be included here to facilitate quick identification of the service in a list. (defined for API consistency)
    }
    property photo : TFhirAttachment read FPhoto write SetPhoto;
    {@member photoElement
      If there is a photo/symbol associated with this HealthcareService, it may be included here to facilitate quick identification of the service in a list.
    }
    property photoElement : TFhirAttachment read FPhoto write SetPhoto;

    {@member telecomList
      List of contacts related to this specific healthcare service.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member coverageAreaList
      The location(s) that this service is available to (not where the service is provided).
    }
    property coverageAreaList : TFhirReferenceList{TFhirLocation} read GetCoverageAreaList;
    property hasCoverageAreaList : boolean read GetHasCoverageAreaList;

    {@member serviceProvisionCodeList
      The code(s) that detail the conditions under which the healthcare service is available/offered.
    }
    property serviceProvisionCodeList : TFhirCodeableConceptList read GetServiceProvisionCodeList;
    property hasServiceProvisionCodeList : boolean read GetHasServiceProvisionCodeList;

    {@member eligibility
      Typed access to Does this service have specific eligibility requirements that need to be met in order to use the service? (defined for API consistency)
    }
    property eligibility : TFhirCodeableConcept read FEligibility write SetEligibility;
    {@member eligibilityElement
      Does this service have specific eligibility requirements that need to be met in order to use the service?
    }
    property eligibilityElement : TFhirCodeableConcept read FEligibility write SetEligibility;

    {@member eligibilityNote
      Typed access to Describes the eligibility conditions for the service.
    }
    property eligibilityNote : String read GetEligibilityNoteST write SetEligibilityNoteST;
    {@member eligibilityNoteElement
      Describes the eligibility conditions for the service.
    }
    property eligibilityNoteElement : TFhirString read FEligibilityNote write SetEligibilityNote;

    {@member programNameList
      Program Names that can be used to categorize the service.
    }
    property programNameList : TFhirStringList read GetProgramNameList;
    property hasProgramNameList : boolean read GetHasProgramNameList;

    {@member characteristicList
      Collection of characteristics (attributes).
    }
    property characteristicList : TFhirCodeableConceptList read GetCharacteristicList;
    property hasCharacteristicList : boolean read GetHasCharacteristicList;

    {@member referralMethodList
      Ways that the service accepts referrals, if this is not provided then it is implied that no referral is required.
    }
    property referralMethodList : TFhirCodeableConceptList read GetReferralMethodList;
    property hasReferralMethodList : boolean read GetHasReferralMethodList;

    {@member publicKey
      Typed access to The public part of the 'keys' allocated to an Organization by an accredited body to support secure exchange of data over the internet. To be provided by the Organization, where available.
    }
    property publicKey : String read GetPublicKeyST write SetPublicKeyST;
    {@member publicKeyElement
      The public part of the 'keys' allocated to an Organization by an accredited body to support secure exchange of data over the internet. To be provided by the Organization, where available.
    }
    property publicKeyElement : TFhirString read FPublicKey write SetPublicKey;

    {@member appointmentRequired
      Typed access to Indicates whether or not a prospective consumer will require an appointment for a particular service at a site to be provided by the Organization. Indicates if an appointment is required for access to this service.
    }
    property appointmentRequired : Boolean read GetAppointmentRequiredST write SetAppointmentRequiredST;
    {@member appointmentRequiredElement
      Indicates whether or not a prospective consumer will require an appointment for a particular service at a site to be provided by the Organization. Indicates if an appointment is required for access to this service.
    }
    property appointmentRequiredElement : TFhirBoolean read FAppointmentRequired write SetAppointmentRequired;

    {@member availableTimeList
      A collection of times that the Service Site is available.
    }
    property availableTimeList : TFhirHealthcareServiceAvailableTimeList read GetAvailableTimeList;
    property hasAvailableTimeList : boolean read GetHasAvailableTimeList;

    {@member notAvailableList
      The HealthcareService is not available during this period of time due to the provided reason.
    }
    property notAvailableList : TFhirHealthcareServiceNotAvailableList read GetNotAvailableList;
    property hasNotAvailableList : boolean read GetHasNotAvailableList;

    {@member availabilityExceptions
      Typed access to A description of site availability exceptions, e.g. public holiday availability. Succinctly describing all possible exceptions to normal site availability as details in the available Times and not available Times.
    }
    property availabilityExceptions : String read GetAvailabilityExceptionsST write SetAvailabilityExceptionsST;
    {@member availabilityExceptionsElement
      A description of site availability exceptions, e.g. public holiday availability. Succinctly describing all possible exceptions to normal site availability as details in the available Times and not available Times.
    }
    property availabilityExceptionsElement : TFhirString read FAvailabilityExceptions write SetAvailabilityExceptions;

  end;


  TFhirHealthcareServiceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirHealthcareServiceList;
    function GetCurrent : TFhirHealthcareService;
  public
    Constructor Create(list : TFhirHealthcareServiceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirHealthcareService read GetCurrent;
  end;


  {@Class TFhirHealthcareServiceList
    A list of FhirHealthcareService
  }
  {!.Net HL7Connect.Fhir.HealthcareServiceList}
  TFhirHealthcareServiceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirHealthcareService;
    procedure SetItemN(index : Integer; value : TFhirHealthcareService);
  public
    {!script hide}
    function Link : TFhirHealthcareServiceList; Overload;
    function Clone : TFhirHealthcareServiceList; Overload;
    function GetEnumerator : TFhirHealthcareServiceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirHealthcareService to the end of the list.
    }
    function Append : TFhirHealthcareService;

    
    {@member AddItem
      Add an already existing FhirHealthcareService to the end of the list.
    }
    procedure AddItem(value : TFhirHealthcareService); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirHealthcareService) : Integer;
    

    {@member Insert
      Insert FhirHealthcareService before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirHealthcareService;
    

    {@member InsertItem
       Insert an existing FhirHealthcareService before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirHealthcareService);
    
    {@member Item
       Get the iIndexth FhirHealthcareService. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirHealthcareService. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirHealthcareService);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirHealthcareService;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirHealthcareServices[index : Integer] : TFhirHealthcareService read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingObjectSelectionStudy : TFhirElement
    Study identity and locating information of the DICOM SOP instances in the selection.
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudy}
  TFhirImagingObjectSelectionStudy = class (TFhirBackboneElement)
  private
    FUid : TFhirOid;
    FUrl : TFhirUri;
    FImagingStudy : TFhirReference{TFhirImagingStudy};
    FseriesList : TFhirImagingObjectSelectionStudySeriesList;
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetImagingStudy(value : TFhirReference{TFhirImagingStudy});
    function GetSeriesList : TFhirImagingObjectSelectionStudySeriesList;
    function GetHasSeriesList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingObjectSelectionStudy; overload;
    function Clone : TFhirImagingObjectSelectionStudy; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member uid
      Typed access to Study instance UID of the SOP instances in the selection.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Study instance UID of the SOP instances in the selection.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member url
      Typed access to WADO-RS URL to retrieve the study. Note that this URL retrieves all SOP instances of the study, not only those in the selection.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      WADO-RS URL to retrieve the study. Note that this URL retrieves all SOP instances of the study, not only those in the selection.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member imagingStudy
      Typed access to Reference to the Imaging Study in FHIR form. (defined for API consistency)
    }
    property imagingStudy : TFhirReference{TFhirImagingStudy} read FImagingStudy write SetImagingStudy;
    {@member imagingStudyElement
      Reference to the Imaging Study in FHIR form.
    }
    property imagingStudyElement : TFhirReference{TFhirImagingStudy} read FImagingStudy write SetImagingStudy;

    {@member seriesList
      Series identity and locating information of the DICOM SOP instances in the selection.
    }
    property seriesList : TFhirImagingObjectSelectionStudySeriesList read GetSeriesList;
    property hasSeriesList : boolean read GetHasSeriesList;

  end;


  TFhirImagingObjectSelectionStudyListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingObjectSelectionStudyList;
    function GetCurrent : TFhirImagingObjectSelectionStudy;
  public
    Constructor Create(list : TFhirImagingObjectSelectionStudyList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingObjectSelectionStudy read GetCurrent;
  end;


  {@Class TFhirImagingObjectSelectionStudyList
    A list of FhirImagingObjectSelectionStudy
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudyList}
  TFhirImagingObjectSelectionStudyList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingObjectSelectionStudy;
    procedure SetItemN(index : Integer; value : TFhirImagingObjectSelectionStudy);
  public
    {!script hide}
    function Link : TFhirImagingObjectSelectionStudyList; Overload;
    function Clone : TFhirImagingObjectSelectionStudyList; Overload;
    function GetEnumerator : TFhirImagingObjectSelectionStudyListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingObjectSelectionStudy to the end of the list.
    }
    function Append : TFhirImagingObjectSelectionStudy;

    
    {@member AddItem
      Add an already existing FhirImagingObjectSelectionStudy to the end of the list.
    }
    procedure AddItem(value : TFhirImagingObjectSelectionStudy); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingObjectSelectionStudy) : Integer;
    

    {@member Insert
      Insert FhirImagingObjectSelectionStudy before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingObjectSelectionStudy;
    

    {@member InsertItem
       Insert an existing FhirImagingObjectSelectionStudy before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingObjectSelectionStudy);
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudy. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudy. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingObjectSelectionStudy);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingObjectSelectionStudy;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingObjectSelectionStudies[index : Integer] : TFhirImagingObjectSelectionStudy read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingObjectSelectionStudySeries : TFhirElement
    Series identity and locating information of the DICOM SOP instances in the selection.
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeries}
  TFhirImagingObjectSelectionStudySeries = class (TFhirBackboneElement)
  private
    FUid : TFhirOid;
    FUrl : TFhirUri;
    FinstanceList : TFhirImagingObjectSelectionStudySeriesInstanceList;
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    function GetInstanceList : TFhirImagingObjectSelectionStudySeriesInstanceList;
    function GetHasInstanceList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingObjectSelectionStudySeries; overload;
    function Clone : TFhirImagingObjectSelectionStudySeries; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member uid
      Typed access to Series instance UID of the SOP instances in the selection.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Series instance UID of the SOP instances in the selection.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member url
      Typed access to WADO-RS URL to retrieve the series. Note that this URL retrieves all SOP instances of the series not only those in the selection.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      WADO-RS URL to retrieve the series. Note that this URL retrieves all SOP instances of the series not only those in the selection.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member instanceList
      Identity and locating information of the selected DICOM SOP instances.
    }
    property instanceList : TFhirImagingObjectSelectionStudySeriesInstanceList read GetInstanceList;
    property hasInstanceList : boolean read GetHasInstanceList;

  end;


  TFhirImagingObjectSelectionStudySeriesListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingObjectSelectionStudySeriesList;
    function GetCurrent : TFhirImagingObjectSelectionStudySeries;
  public
    Constructor Create(list : TFhirImagingObjectSelectionStudySeriesList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingObjectSelectionStudySeries read GetCurrent;
  end;


  {@Class TFhirImagingObjectSelectionStudySeriesList
    A list of FhirImagingObjectSelectionStudySeries
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeriesList}
  TFhirImagingObjectSelectionStudySeriesList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingObjectSelectionStudySeries;
    procedure SetItemN(index : Integer; value : TFhirImagingObjectSelectionStudySeries);
  public
    {!script hide}
    function Link : TFhirImagingObjectSelectionStudySeriesList; Overload;
    function Clone : TFhirImagingObjectSelectionStudySeriesList; Overload;
    function GetEnumerator : TFhirImagingObjectSelectionStudySeriesListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingObjectSelectionStudySeries to the end of the list.
    }
    function Append : TFhirImagingObjectSelectionStudySeries;

    
    {@member AddItem
      Add an already existing FhirImagingObjectSelectionStudySeries to the end of the list.
    }
    procedure AddItem(value : TFhirImagingObjectSelectionStudySeries); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingObjectSelectionStudySeries) : Integer;
    

    {@member Insert
      Insert FhirImagingObjectSelectionStudySeries before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingObjectSelectionStudySeries;
    

    {@member InsertItem
       Insert an existing FhirImagingObjectSelectionStudySeries before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingObjectSelectionStudySeries);
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeries. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeries. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingObjectSelectionStudySeries);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingObjectSelectionStudySeries;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingObjectSelectionStudySeries[index : Integer] : TFhirImagingObjectSelectionStudySeries read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingObjectSelectionStudySeriesInstance : TFhirElement
    Identity and locating information of the selected DICOM SOP instances.
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeriesInstance}
  TFhirImagingObjectSelectionStudySeriesInstance = class (TFhirBackboneElement)
  private
    FSopClass : TFhirOid;
    FUid : TFhirOid;
    FUrl : TFhirUri;
    FframesList : TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
    Procedure SetSopClass(value : TFhirOid);
    Function GetSopClassST : String;
    Procedure SetSopClassST(value : String);
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    function GetFramesList : TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
    function GetHasFramesList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingObjectSelectionStudySeriesInstance; overload;
    function Clone : TFhirImagingObjectSelectionStudySeriesInstance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sopClass
      Typed access to SOP class UID of the selected instance.
    }
    property sopClass : String read GetSopClassST write SetSopClassST;
    {@member sopClassElement
      SOP class UID of the selected instance.
    }
    property sopClassElement : TFhirOid read FSopClass write SetSopClass;

    {@member uid
      Typed access to SOP Instance UID of the selected instance.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      SOP Instance UID of the selected instance.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member url
      Typed access to WADO-RS URL to retrieve the DICOM SOP Instance.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      WADO-RS URL to retrieve the DICOM SOP Instance.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member framesList
      Identity and location information of the frames in the selected instance.
    }
    property framesList : TFhirImagingObjectSelectionStudySeriesInstanceFramesList read GetFramesList;
    property hasFramesList : boolean read GetHasFramesList;

  end;


  TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingObjectSelectionStudySeriesInstanceList;
    function GetCurrent : TFhirImagingObjectSelectionStudySeriesInstance;
  public
    Constructor Create(list : TFhirImagingObjectSelectionStudySeriesInstanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingObjectSelectionStudySeriesInstance read GetCurrent;
  end;


  {@Class TFhirImagingObjectSelectionStudySeriesInstanceList
    A list of FhirImagingObjectSelectionStudySeriesInstance
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeriesInstanceList}
  TFhirImagingObjectSelectionStudySeriesInstanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstance;
    procedure SetItemN(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstance);
  public
    {!script hide}
    function Link : TFhirImagingObjectSelectionStudySeriesInstanceList; Overload;
    function Clone : TFhirImagingObjectSelectionStudySeriesInstanceList; Overload;
    function GetEnumerator : TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingObjectSelectionStudySeriesInstance to the end of the list.
    }
    function Append : TFhirImagingObjectSelectionStudySeriesInstance;

    
    {@member AddItem
      Add an already existing FhirImagingObjectSelectionStudySeriesInstance to the end of the list.
    }
    procedure AddItem(value : TFhirImagingObjectSelectionStudySeriesInstance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingObjectSelectionStudySeriesInstance) : Integer;
    

    {@member Insert
      Insert FhirImagingObjectSelectionStudySeriesInstance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstance;
    

    {@member InsertItem
       Insert an existing FhirImagingObjectSelectionStudySeriesInstance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstance);
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeriesInstance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeriesInstance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingObjectSelectionStudySeriesInstances[index : Integer] : TFhirImagingObjectSelectionStudySeriesInstance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingObjectSelectionStudySeriesInstanceFrames : TFhirElement
    Identity and location information of the frames in the selected instance.
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeriesInstanceFrames}
  TFhirImagingObjectSelectionStudySeriesInstanceFrames = class (TFhirBackboneElement)
  private
    FframeNumbersList : TFhirUnsignedIntList;
    FUrl : TFhirUri;
    function GetFrameNumbersList : TFhirUnsignedIntList;
    function GetHasFrameNumbersList : Boolean;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingObjectSelectionStudySeriesInstanceFrames; overload;
    function Clone : TFhirImagingObjectSelectionStudySeriesInstanceFrames; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member frameNumbersList
      The frame numbers in the frame set.
    }
    property frameNumbersList : TFhirUnsignedIntList read GetFrameNumbersList;
    property hasFrameNumbersList : boolean read GetHasFrameNumbersList;

    {@member url
      Typed access to WADO-RS URL to retrieve the DICOM frames.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      WADO-RS URL to retrieve the DICOM frames.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

  end;


  TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
    function GetCurrent : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
  public
    Constructor Create(list : TFhirImagingObjectSelectionStudySeriesInstanceFramesList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingObjectSelectionStudySeriesInstanceFrames read GetCurrent;
  end;


  {@Class TFhirImagingObjectSelectionStudySeriesInstanceFramesList
    A list of FhirImagingObjectSelectionStudySeriesInstanceFrames
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionStudySeriesInstanceFramesList}
  TFhirImagingObjectSelectionStudySeriesInstanceFramesList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
    procedure SetItemN(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstanceFrames);
  public
    {!script hide}
    function Link : TFhirImagingObjectSelectionStudySeriesInstanceFramesList; Overload;
    function Clone : TFhirImagingObjectSelectionStudySeriesInstanceFramesList; Overload;
    function GetEnumerator : TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingObjectSelectionStudySeriesInstanceFrames to the end of the list.
    }
    function Append : TFhirImagingObjectSelectionStudySeriesInstanceFrames;

    
    {@member AddItem
      Add an already existing FhirImagingObjectSelectionStudySeriesInstanceFrames to the end of the list.
    }
    procedure AddItem(value : TFhirImagingObjectSelectionStudySeriesInstanceFrames); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingObjectSelectionStudySeriesInstanceFrames) : Integer;
    

    {@member Insert
      Insert FhirImagingObjectSelectionStudySeriesInstanceFrames before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
    

    {@member InsertItem
       Insert an existing FhirImagingObjectSelectionStudySeriesInstanceFrames before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstanceFrames);
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeriesInstanceFrames. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelectionStudySeriesInstanceFrames. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingObjectSelectionStudySeriesInstanceFrames);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingObjectSelectionStudySeriesInstanceFrames[index : Integer] : TFhirImagingObjectSelectionStudySeriesInstanceFrames read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingObjectSelection : TFhirDomainResource
    A manifest of a set of DICOM Service-Object Pair Instances (SOP Instances).  The referenced SOP Instances (images or other content) are for a single patient, and may be from one or more studies. The referenced SOP Instances have been selected for a purpose, such as quality assurance, conference, or consult. Reflecting that range of purposes, typical ImagingObjectSelection resources may include all SOP Instances in a study (perhaps for sharing through a Health Information Exchange); key images from multiple studies (for reference by a referring or treating physician); a multi-frame ultrasound instance ("cine" video clip) and a set of measurements taken from that instance (for inclusion in a teaching file); and so on.
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelection}
  TFhirImagingObjectSelection = class (TFhirDomainResource)
  private
    FUid : TFhirOid;
    FPatient : TFhirReference{TFhirPatient};
    FTitle : TFhirCodeableConcept;
    FDescription : TFhirString;
    FAuthor : TFhirReference{Resource};
    FAuthoringTime : TFhirDateTime;
    FstudyList : TFhirImagingObjectSelectionStudyList;
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetTitle(value : TFhirCodeableConcept);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetAuthoringTime(value : TFhirDateTime);
    Function GetAuthoringTimeST : TDateAndTime;
    Procedure SetAuthoringTimeST(value : TDateAndTime);
    function GetStudyList : TFhirImagingObjectSelectionStudyList;
    function GetHasStudyList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingObjectSelection; overload;
    function Clone : TFhirImagingObjectSelection; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member uid
      Typed access to Instance UID of the DICOM KOS SOP Instances represented in this resource.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Instance UID of the DICOM KOS SOP Instances represented in this resource.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member patient
      Typed access to A patient resource reference which is the patient subject of all DICOM SOP Instances in this ImagingObjectSelection. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A patient resource reference which is the patient subject of all DICOM SOP Instances in this ImagingObjectSelection.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member title
      Typed access to The reason for, or significance of, the selection of objects referenced in the resource. (defined for API consistency)
    }
    property title : TFhirCodeableConcept read FTitle write SetTitle;
    {@member titleElement
      The reason for, or significance of, the selection of objects referenced in the resource.
    }
    property titleElement : TFhirCodeableConcept read FTitle write SetTitle;

    {@member description
      Typed access to Text description of the DICOM SOP instances selected in the ImagingObjectSelection. This should be aligned with the content of the title element, and can provide further explanation of the SOP instances in the selection.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Text description of the DICOM SOP instances selected in the ImagingObjectSelection. This should be aligned with the content of the title element, and can provide further explanation of the SOP instances in the selection.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member author
      Typed access to Author of ImagingObjectSelection. It can be a human author or a device which made the decision of the SOP instances selected. For example, a radiologist selected a set of imaging SOP instances to attach in a diagnostic report, and a CAD application may author a selection to describe SOP instances it used to generate a detection conclusion. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      Author of ImagingObjectSelection. It can be a human author or a device which made the decision of the SOP instances selected. For example, a radiologist selected a set of imaging SOP instances to attach in a diagnostic report, and a CAD application may author a selection to describe SOP instances it used to generate a detection conclusion.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member authoringTime
      Typed access to Date and time when the selection of the referenced instances were made. It is (typically) different from the creation date of the selection resource, and from dates associated with the referenced instances (e.g. capture time of the referenced image).
    }
    property authoringTime : TDateAndTime read GetAuthoringTimeST write SetAuthoringTimeST;
    {@member authoringTimeElement
      Date and time when the selection of the referenced instances were made. It is (typically) different from the creation date of the selection resource, and from dates associated with the referenced instances (e.g. capture time of the referenced image).
    }
    property authoringTimeElement : TFhirDateTime read FAuthoringTime write SetAuthoringTime;

    {@member studyList
      Study identity and locating information of the DICOM SOP instances in the selection.
    }
    property studyList : TFhirImagingObjectSelectionStudyList read GetStudyList;
    property hasStudyList : boolean read GetHasStudyList;

  end;


  TFhirImagingObjectSelectionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingObjectSelectionList;
    function GetCurrent : TFhirImagingObjectSelection;
  public
    Constructor Create(list : TFhirImagingObjectSelectionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingObjectSelection read GetCurrent;
  end;


  {@Class TFhirImagingObjectSelectionList
    A list of FhirImagingObjectSelection
  }
  {!.Net HL7Connect.Fhir.ImagingObjectSelectionList}
  TFhirImagingObjectSelectionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingObjectSelection;
    procedure SetItemN(index : Integer; value : TFhirImagingObjectSelection);
  public
    {!script hide}
    function Link : TFhirImagingObjectSelectionList; Overload;
    function Clone : TFhirImagingObjectSelectionList; Overload;
    function GetEnumerator : TFhirImagingObjectSelectionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingObjectSelection to the end of the list.
    }
    function Append : TFhirImagingObjectSelection;

    
    {@member AddItem
      Add an already existing FhirImagingObjectSelection to the end of the list.
    }
    procedure AddItem(value : TFhirImagingObjectSelection); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingObjectSelection) : Integer;
    

    {@member Insert
      Insert FhirImagingObjectSelection before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingObjectSelection;
    

    {@member InsertItem
       Insert an existing FhirImagingObjectSelection before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingObjectSelection);
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelection. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingObjectSelection. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingObjectSelection);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingObjectSelection;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingObjectSelections[index : Integer] : TFhirImagingObjectSelection read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingStudySeries : TFhirElement
    Each study has one or more series of images or other content.
  }
  {!.Net HL7Connect.Fhir.ImagingStudySeries}
  TFhirImagingStudySeries = class (TFhirBackboneElement)
  private
    FNumber : TFhirUnsignedInt;
    FModality : TFhirCoding;
    FUid : TFhirOid;
    FDescription : TFhirString;
    FNumberOfInstances : TFhirUnsignedInt;
    FAvailability : TFhirEnum;
    FUrl : TFhirUri;
    FBodySite : TFhirCoding;
    FLaterality : TFhirCoding;
    FStarted : TFhirDateTime;
    FinstanceList : TFhirImagingStudySeriesInstanceList;
    Procedure SetNumber(value : TFhirUnsignedInt);
    Function GetNumberST : String;
    Procedure SetNumberST(value : String);
    Procedure SetModality(value : TFhirCoding);
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetNumberOfInstances(value : TFhirUnsignedInt);
    Function GetNumberOfInstancesST : String;
    Procedure SetNumberOfInstancesST(value : String);
    Procedure SetAvailability(value : TFhirEnum);
    Function GetAvailabilityST : TFhirInstanceAvailability;
    Procedure SetAvailabilityST(value : TFhirInstanceAvailability);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetBodySite(value : TFhirCoding);
    Procedure SetLaterality(value : TFhirCoding);
    Procedure SetStarted(value : TFhirDateTime);
    Function GetStartedST : TDateAndTime;
    Procedure SetStartedST(value : TDateAndTime);
    function GetInstanceList : TFhirImagingStudySeriesInstanceList;
    function GetHasInstanceList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingStudySeries; overload;
    function Clone : TFhirImagingStudySeries; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member number
      Typed access to The Numeric identifier of this series in the study.
    }
    property number : String read GetNumberST write SetNumberST;
    {@member numberElement
      The Numeric identifier of this series in the study.
    }
    property numberElement : TFhirUnsignedInt read FNumber write SetNumber;

    {@member modality
      Typed access to The modality of this series sequence. (defined for API consistency)
    }
    property modality : TFhirCoding read FModality write SetModality;
    {@member modalityElement
      The modality of this series sequence.
    }
    property modalityElement : TFhirCoding read FModality write SetModality;

    {@member uid
      Typed access to Formal identifier for this series.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Formal identifier for this series.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member description
      Typed access to A description of the series.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A description of the series.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member numberOfInstances
      Typed access to Number of SOP Instances in Series.
    }
    property numberOfInstances : String read GetNumberOfInstancesST write SetNumberOfInstancesST;
    {@member numberOfInstancesElement
      Number of SOP Instances in Series.
    }
    property numberOfInstancesElement : TFhirUnsignedInt read FNumberOfInstances write SetNumberOfInstances;

    {@member availability
      Availability of series (online, offline or nearline).
    }
    property availability : TFhirInstanceAvailability read GetAvailabilityST write SetAvailabilityST;
    property availabilityElement : TFhirEnum read FAvailability write SetAvailability;

    {@member url
      Typed access to URI/URL specifying the location of the referenced series using WADO-RS.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      URI/URL specifying the location of the referenced series using WADO-RS.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member bodySite
      Typed access to Body part examined. See  DICOM Part 16 Annex L for the mapping from DICOM to Snomed CT. (defined for API consistency)
    }
    property bodySite : TFhirCoding read FBodySite write SetBodySite;
    {@member bodySiteElement
      Body part examined. See  DICOM Part 16 Annex L for the mapping from DICOM to Snomed CT.
    }
    property bodySiteElement : TFhirCoding read FBodySite write SetBodySite;

    {@member laterality
      Typed access to Laterality if body site is paired anatomic structure and laterality is not pre-coordinated in body site code. (defined for API consistency)
    }
    property laterality : TFhirCoding read FLaterality write SetLaterality;
    {@member lateralityElement
      Laterality if body site is paired anatomic structure and laterality is not pre-coordinated in body site code.
    }
    property lateralityElement : TFhirCoding read FLaterality write SetLaterality;

    {@member started
      Typed access to The date and time the series was started.
    }
    property started : TDateAndTime read GetStartedST write SetStartedST;
    {@member startedElement
      The date and time the series was started.
    }
    property startedElement : TFhirDateTime read FStarted write SetStarted;

    {@member instanceList
      A single SOP Instance within the series, e.g. an image, or presentation state.
    }
    property instanceList : TFhirImagingStudySeriesInstanceList read GetInstanceList;
    property hasInstanceList : boolean read GetHasInstanceList;

  end;


  TFhirImagingStudySeriesListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingStudySeriesList;
    function GetCurrent : TFhirImagingStudySeries;
  public
    Constructor Create(list : TFhirImagingStudySeriesList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingStudySeries read GetCurrent;
  end;


  {@Class TFhirImagingStudySeriesList
    A list of FhirImagingStudySeries
  }
  {!.Net HL7Connect.Fhir.ImagingStudySeriesList}
  TFhirImagingStudySeriesList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingStudySeries;
    procedure SetItemN(index : Integer; value : TFhirImagingStudySeries);
  public
    {!script hide}
    function Link : TFhirImagingStudySeriesList; Overload;
    function Clone : TFhirImagingStudySeriesList; Overload;
    function GetEnumerator : TFhirImagingStudySeriesListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingStudySeries to the end of the list.
    }
    function Append : TFhirImagingStudySeries;

    
    {@member AddItem
      Add an already existing FhirImagingStudySeries to the end of the list.
    }
    procedure AddItem(value : TFhirImagingStudySeries); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingStudySeries) : Integer;
    

    {@member Insert
      Insert FhirImagingStudySeries before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingStudySeries;
    

    {@member InsertItem
       Insert an existing FhirImagingStudySeries before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingStudySeries);
    
    {@member Item
       Get the iIndexth FhirImagingStudySeries. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingStudySeries. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingStudySeries);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingStudySeries;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingStudySeries[index : Integer] : TFhirImagingStudySeries read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingStudySeriesInstance : TFhirElement
    A single SOP Instance within the series, e.g. an image, or presentation state.
  }
  {!.Net HL7Connect.Fhir.ImagingStudySeriesInstance}
  TFhirImagingStudySeriesInstance = class (TFhirBackboneElement)
  private
    FNumber : TFhirUnsignedInt;
    FUid : TFhirOid;
    FSopClass : TFhirOid;
    FType_ : TFhirString;
    FTitle : TFhirString;
    FcontentList : TFhirAttachmentList;
    Procedure SetNumber(value : TFhirUnsignedInt);
    Function GetNumberST : String;
    Procedure SetNumberST(value : String);
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetSopClass(value : TFhirOid);
    Function GetSopClassST : String;
    Procedure SetSopClassST(value : String);
    Procedure SetType_(value : TFhirString);
    Function GetType_ST : String;
    Procedure SetType_ST(value : String);
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    function GetContentList : TFhirAttachmentList;
    function GetHasContentList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingStudySeriesInstance; overload;
    function Clone : TFhirImagingStudySeriesInstance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member number
      Typed access to The number of instance in the series.
    }
    property number : String read GetNumberST write SetNumberST;
    {@member numberElement
      The number of instance in the series.
    }
    property numberElement : TFhirUnsignedInt read FNumber write SetNumber;

    {@member uid
      Typed access to Formal identifier for this image or other content.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Formal identifier for this image or other content.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member sopClass
      Typed access to DICOM instance  type.
    }
    property sopClass : String read GetSopClassST write SetSopClassST;
    {@member sopClassElement
      DICOM instance  type.
    }
    property sopClassElement : TFhirOid read FSopClass write SetSopClass;

    {@member type_
      Typed access to A human-friendly SOP Class name.
    }
    property type_ : String read GetType_ST write SetType_ST;
    {@member type_Element
      A human-friendly SOP Class name.
    }
    property type_Element : TFhirString read FType_ write SetType_;

    {@member title
      Typed access to The description of the instance.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      The description of the instance.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member contentList
      Content of the instance or a rendering thereof (e.g. a JPEG of an image, or an XML of a structured report). May be represented for example by inline encoding; by a URL reference to a WADO-RS service that makes the instance available; or to a FHIR Resource (e.g. Media, Document, etc.). Multiple content attachments may be used for alternate representations of the instance.
    }
    property contentList : TFhirAttachmentList read GetContentList;
    property hasContentList : boolean read GetHasContentList;

  end;


  TFhirImagingStudySeriesInstanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingStudySeriesInstanceList;
    function GetCurrent : TFhirImagingStudySeriesInstance;
  public
    Constructor Create(list : TFhirImagingStudySeriesInstanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingStudySeriesInstance read GetCurrent;
  end;


  {@Class TFhirImagingStudySeriesInstanceList
    A list of FhirImagingStudySeriesInstance
  }
  {!.Net HL7Connect.Fhir.ImagingStudySeriesInstanceList}
  TFhirImagingStudySeriesInstanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingStudySeriesInstance;
    procedure SetItemN(index : Integer; value : TFhirImagingStudySeriesInstance);
  public
    {!script hide}
    function Link : TFhirImagingStudySeriesInstanceList; Overload;
    function Clone : TFhirImagingStudySeriesInstanceList; Overload;
    function GetEnumerator : TFhirImagingStudySeriesInstanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingStudySeriesInstance to the end of the list.
    }
    function Append : TFhirImagingStudySeriesInstance;

    
    {@member AddItem
      Add an already existing FhirImagingStudySeriesInstance to the end of the list.
    }
    procedure AddItem(value : TFhirImagingStudySeriesInstance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingStudySeriesInstance) : Integer;
    

    {@member Insert
      Insert FhirImagingStudySeriesInstance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingStudySeriesInstance;
    

    {@member InsertItem
       Insert an existing FhirImagingStudySeriesInstance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingStudySeriesInstance);
    
    {@member Item
       Get the iIndexth FhirImagingStudySeriesInstance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingStudySeriesInstance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingStudySeriesInstance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingStudySeriesInstance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingStudySeriesInstances[index : Integer] : TFhirImagingStudySeriesInstance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImagingStudy : TFhirDomainResource
    Representation of the content produced in a DICOM imaging study. A study comprises a set of series, each of which includes a set of Service-Object Pair Instances (SOP Instances - images or other data) acquired or produced in a common context.  A series is of only one modality (e.g. X-ray, CT, MR, ultrasound), but a study may have multiple series of different modalities.
  }
  {!.Net HL7Connect.Fhir.ImagingStudy}
  TFhirImagingStudy = class (TFhirDomainResource)
  private
    FStarted : TFhirDateTime;
    FPatient : TFhirReference{TFhirPatient};
    FUid : TFhirOid;
    FAccession : TFhirIdentifier;
    FidentifierList : TFhirIdentifierList;
    ForderList : TFhirReferenceList{TFhirDiagnosticOrder};
    FmodalityListList : TFhirCodingList;
    FReferrer : TFhirReference{TFhirPractitioner};
    FAvailability : TFhirEnum;
    FUrl : TFhirUri;
    FNumberOfSeries : TFhirUnsignedInt;
    FNumberOfInstances : TFhirUnsignedInt;
    Fprocedure_List : TFhirReferenceList{TFhirProcedure};
    FInterpreter : TFhirReference{TFhirPractitioner};
    FDescription : TFhirString;
    FseriesList : TFhirImagingStudySeriesList;
    Procedure SetStarted(value : TFhirDateTime);
    Function GetStartedST : TDateAndTime;
    Procedure SetStartedST(value : TDateAndTime);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetUid(value : TFhirOid);
    Function GetUidST : String;
    Procedure SetUidST(value : String);
    Procedure SetAccession(value : TFhirIdentifier);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetOrderList : TFhirReferenceList{TFhirDiagnosticOrder};
    function GetHasOrderList : Boolean;
    function GetModalityListList : TFhirCodingList;
    function GetHasModalityListList : Boolean;
    Procedure SetReferrer(value : TFhirReference{TFhirPractitioner});
    Procedure SetAvailability(value : TFhirEnum);
    Function GetAvailabilityST : TFhirInstanceAvailability;
    Procedure SetAvailabilityST(value : TFhirInstanceAvailability);
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetNumberOfSeries(value : TFhirUnsignedInt);
    Function GetNumberOfSeriesST : String;
    Procedure SetNumberOfSeriesST(value : String);
    Procedure SetNumberOfInstances(value : TFhirUnsignedInt);
    Function GetNumberOfInstancesST : String;
    Procedure SetNumberOfInstancesST(value : String);
    function GetProcedure_List : TFhirReferenceList{TFhirProcedure};
    function GetHasProcedure_List : Boolean;
    Procedure SetInterpreter(value : TFhirReference{TFhirPractitioner});
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetSeriesList : TFhirImagingStudySeriesList;
    function GetHasSeriesList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImagingStudy; overload;
    function Clone : TFhirImagingStudy; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member started
      Typed access to Date and Time the study started.
    }
    property started : TDateAndTime read GetStartedST write SetStartedST;
    {@member startedElement
      Date and Time the study started.
    }
    property startedElement : TFhirDateTime read FStarted write SetStarted;

    {@member patient
      Typed access to The patient imaged in the study. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient imaged in the study.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member uid
      Typed access to Formal identifier for the study.
    }
    property uid : String read GetUidST write SetUidST;
    {@member uidElement
      Formal identifier for the study.
    }
    property uidElement : TFhirOid read FUid write SetUid;

    {@member accession
      Typed access to Accession Number is an identifier related to some aspect of imaging workflow and data management. Usage may vary across different institutions.  See for instance [IHE Radiology Technical Framework Volume 1 Appendix A](http://www.ihe.net/uploadedFiles/Documents/Radiology/IHE_RAD_TF_Rev13.0_Vol1_FT_2014-07-30.pdf). (defined for API consistency)
    }
    property accession : TFhirIdentifier read FAccession write SetAccession;
    {@member accessionElement
      Accession Number is an identifier related to some aspect of imaging workflow and data management. Usage may vary across different institutions.  See for instance [IHE Radiology Technical Framework Volume 1 Appendix A](http://www.ihe.net/uploadedFiles/Documents/Radiology/IHE_RAD_TF_Rev13.0_Vol1_FT_2014-07-30.pdf).
    }
    property accessionElement : TFhirIdentifier read FAccession write SetAccession;

    {@member identifierList
      Other identifiers for the study.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member orderList
      A list of the diagnostic orders that resulted in this imaging study being performed.
    }
    property orderList : TFhirReferenceList{TFhirDiagnosticOrder} read GetOrderList;
    property hasOrderList : boolean read GetHasOrderList;

    {@member modalityListList
      A list of all the Series.ImageModality values that are actual acquisition modalities, i.e. those in the DICOM Context Group 29 (value set OID 1.2.840.10008.6.1.19).
    }
    property modalityListList : TFhirCodingList read GetModalityListList;
    property hasModalityListList : boolean read GetHasModalityListList;

    {@member referrer
      Typed access to The requesting/referring physician. (defined for API consistency)
    }
    property referrer : TFhirReference{TFhirPractitioner} read FReferrer write SetReferrer;
    {@member referrerElement
      The requesting/referring physician.
    }
    property referrerElement : TFhirReference{TFhirPractitioner} read FReferrer write SetReferrer;

    {@member availability
      Availability of study (online, offline or nearline).
    }
    property availability : TFhirInstanceAvailability read GetAvailabilityST write SetAvailabilityST;
    property availabilityElement : TFhirEnum read FAvailability write SetAvailability;

    {@member url
      Typed access to WADO-RS resource where Study is available.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      WADO-RS resource where Study is available.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member numberOfSeries
      Typed access to Number of Series in Study.
    }
    property numberOfSeries : String read GetNumberOfSeriesST write SetNumberOfSeriesST;
    {@member numberOfSeriesElement
      Number of Series in Study.
    }
    property numberOfSeriesElement : TFhirUnsignedInt read FNumberOfSeries write SetNumberOfSeries;

    {@member numberOfInstances
      Typed access to Number of SOP Instances in Study.
    }
    property numberOfInstances : String read GetNumberOfInstancesST write SetNumberOfInstancesST;
    {@member numberOfInstancesElement
      Number of SOP Instances in Study.
    }
    property numberOfInstancesElement : TFhirUnsignedInt read FNumberOfInstances write SetNumberOfInstances;

    {@member procedure_List
      Type of procedure performed.
    }
    property procedure_List : TFhirReferenceList{TFhirProcedure} read GetProcedure_List;
    property hasProcedure_List : boolean read GetHasProcedure_List;

    {@member interpreter
      Typed access to Who read the study and interpreted the images or other content. (defined for API consistency)
    }
    property interpreter : TFhirReference{TFhirPractitioner} read FInterpreter write SetInterpreter;
    {@member interpreterElement
      Who read the study and interpreted the images or other content.
    }
    property interpreterElement : TFhirReference{TFhirPractitioner} read FInterpreter write SetInterpreter;

    {@member description
      Typed access to Institution-generated description or classification of the Study performed.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Institution-generated description or classification of the Study performed.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member seriesList
      Each study has one or more series of images or other content.
    }
    property seriesList : TFhirImagingStudySeriesList read GetSeriesList;
    property hasSeriesList : boolean read GetHasSeriesList;

  end;


  TFhirImagingStudyListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImagingStudyList;
    function GetCurrent : TFhirImagingStudy;
  public
    Constructor Create(list : TFhirImagingStudyList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImagingStudy read GetCurrent;
  end;


  {@Class TFhirImagingStudyList
    A list of FhirImagingStudy
  }
  {!.Net HL7Connect.Fhir.ImagingStudyList}
  TFhirImagingStudyList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImagingStudy;
    procedure SetItemN(index : Integer; value : TFhirImagingStudy);
  public
    {!script hide}
    function Link : TFhirImagingStudyList; Overload;
    function Clone : TFhirImagingStudyList; Overload;
    function GetEnumerator : TFhirImagingStudyListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImagingStudy to the end of the list.
    }
    function Append : TFhirImagingStudy;

    
    {@member AddItem
      Add an already existing FhirImagingStudy to the end of the list.
    }
    procedure AddItem(value : TFhirImagingStudy); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImagingStudy) : Integer;
    

    {@member Insert
      Insert FhirImagingStudy before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImagingStudy;
    

    {@member InsertItem
       Insert an existing FhirImagingStudy before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImagingStudy);
    
    {@member Item
       Get the iIndexth FhirImagingStudy. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImagingStudy. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImagingStudy);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImagingStudy;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImagingStudies[index : Integer] : TFhirImagingStudy read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationExplanation : TFhirElement
    Reasons why a vaccine was or was not administered.
  }
  {!.Net HL7Connect.Fhir.ImmunizationExplanation}
  TFhirImmunizationExplanation = class (TFhirBackboneElement)
  private
    FreasonList : TFhirCodeableConceptList;
    FreasonNotGivenList : TFhirCodeableConceptList;
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    function GetReasonNotGivenList : TFhirCodeableConceptList;
    function GetHasReasonNotGivenList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationExplanation; overload;
    function Clone : TFhirImmunizationExplanation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member reasonList
      Reasons why a vaccine was administered.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member reasonNotGivenList
      Reason why a vaccine was not administered.
    }
    property reasonNotGivenList : TFhirCodeableConceptList read GetReasonNotGivenList;
    property hasReasonNotGivenList : boolean read GetHasReasonNotGivenList;

  end;


  TFhirImmunizationExplanationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationExplanationList;
    function GetCurrent : TFhirImmunizationExplanation;
  public
    Constructor Create(list : TFhirImmunizationExplanationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationExplanation read GetCurrent;
  end;


  {@Class TFhirImmunizationExplanationList
    A list of FhirImmunizationExplanation
  }
  {!.Net HL7Connect.Fhir.ImmunizationExplanationList}
  TFhirImmunizationExplanationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationExplanation;
    procedure SetItemN(index : Integer; value : TFhirImmunizationExplanation);
  public
    {!script hide}
    function Link : TFhirImmunizationExplanationList; Overload;
    function Clone : TFhirImmunizationExplanationList; Overload;
    function GetEnumerator : TFhirImmunizationExplanationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationExplanation to the end of the list.
    }
    function Append : TFhirImmunizationExplanation;

    
    {@member AddItem
      Add an already existing FhirImmunizationExplanation to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationExplanation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationExplanation) : Integer;
    

    {@member Insert
      Insert FhirImmunizationExplanation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationExplanation;
    

    {@member InsertItem
       Insert an existing FhirImmunizationExplanation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationExplanation);
    
    {@member Item
       Get the iIndexth FhirImmunizationExplanation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationExplanation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationExplanation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationExplanation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationExplanations[index : Integer] : TFhirImmunizationExplanation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationReaction : TFhirElement
    Categorical data indicating that an adverse event is associated in time to an immunization.
  }
  {!.Net HL7Connect.Fhir.ImmunizationReaction}
  TFhirImmunizationReaction = class (TFhirBackboneElement)
  private
    FDate : TFhirDateTime;
    FDetail : TFhirReference{TFhirObservation};
    FReported : TFhirBoolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDetail(value : TFhirReference{TFhirObservation});
    Procedure SetReported(value : TFhirBoolean);
    Function GetReportedST : Boolean;
    Procedure SetReportedST(value : Boolean);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationReaction; overload;
    function Clone : TFhirImmunizationReaction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member date
      Typed access to Date of reaction to the immunization.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      Date of reaction to the immunization.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member detail
      Typed access to Details of the reaction. (defined for API consistency)
    }
    property detail : TFhirReference{TFhirObservation} read FDetail write SetDetail;
    {@member detailElement
      Details of the reaction.
    }
    property detailElement : TFhirReference{TFhirObservation} read FDetail write SetDetail;

    {@member reported
      Typed access to Self-reported indicator.
    }
    property reported : Boolean read GetReportedST write SetReportedST;
    {@member reportedElement
      Self-reported indicator.
    }
    property reportedElement : TFhirBoolean read FReported write SetReported;

  end;


  TFhirImmunizationReactionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationReactionList;
    function GetCurrent : TFhirImmunizationReaction;
  public
    Constructor Create(list : TFhirImmunizationReactionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationReaction read GetCurrent;
  end;


  {@Class TFhirImmunizationReactionList
    A list of FhirImmunizationReaction
  }
  {!.Net HL7Connect.Fhir.ImmunizationReactionList}
  TFhirImmunizationReactionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationReaction;
    procedure SetItemN(index : Integer; value : TFhirImmunizationReaction);
  public
    {!script hide}
    function Link : TFhirImmunizationReactionList; Overload;
    function Clone : TFhirImmunizationReactionList; Overload;
    function GetEnumerator : TFhirImmunizationReactionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationReaction to the end of the list.
    }
    function Append : TFhirImmunizationReaction;

    
    {@member AddItem
      Add an already existing FhirImmunizationReaction to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationReaction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationReaction) : Integer;
    

    {@member Insert
      Insert FhirImmunizationReaction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationReaction;
    

    {@member InsertItem
       Insert an existing FhirImmunizationReaction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationReaction);
    
    {@member Item
       Get the iIndexth FhirImmunizationReaction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationReaction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationReaction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationReaction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationReactions[index : Integer] : TFhirImmunizationReaction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationVaccinationProtocol : TFhirElement
    Contains information about the protocol(s) under which the vaccine was administered.
  }
  {!.Net HL7Connect.Fhir.ImmunizationVaccinationProtocol}
  TFhirImmunizationVaccinationProtocol = class (TFhirBackboneElement)
  private
    FDoseSequence : TFhirPositiveInt;
    FDescription : TFhirString;
    FAuthority : TFhirReference{TFhirOrganization};
    FSeries : TFhirString;
    FSeriesDoses : TFhirPositiveInt;
    FtargetDiseaseList : TFhirCodeableConceptList;
    FDoseStatus : TFhirCodeableConcept;
    FDoseStatusReason : TFhirCodeableConcept;
    Procedure SetDoseSequence(value : TFhirPositiveInt);
    Function GetDoseSequenceST : String;
    Procedure SetDoseSequenceST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetAuthority(value : TFhirReference{TFhirOrganization});
    Procedure SetSeries(value : TFhirString);
    Function GetSeriesST : String;
    Procedure SetSeriesST(value : String);
    Procedure SetSeriesDoses(value : TFhirPositiveInt);
    Function GetSeriesDosesST : String;
    Procedure SetSeriesDosesST(value : String);
    function GetTargetDiseaseList : TFhirCodeableConceptList;
    function GetHasTargetDiseaseList : Boolean;
    Procedure SetDoseStatus(value : TFhirCodeableConcept);
    Procedure SetDoseStatusReason(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationVaccinationProtocol; overload;
    function Clone : TFhirImmunizationVaccinationProtocol; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member doseSequence
      Typed access to Nominal position in a series.
    }
    property doseSequence : String read GetDoseSequenceST write SetDoseSequenceST;
    {@member doseSequenceElement
      Nominal position in a series.
    }
    property doseSequenceElement : TFhirPositiveInt read FDoseSequence write SetDoseSequence;

    {@member description
      Typed access to Contains the description about the protocol under which the vaccine was administered.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Contains the description about the protocol under which the vaccine was administered.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member authority
      Typed access to Indicates the authority who published the protocol.  E.g. ACIP. (defined for API consistency)
    }
    property authority : TFhirReference{TFhirOrganization} read FAuthority write SetAuthority;
    {@member authorityElement
      Indicates the authority who published the protocol.  E.g. ACIP.
    }
    property authorityElement : TFhirReference{TFhirOrganization} read FAuthority write SetAuthority;

    {@member series
      Typed access to One possible path to achieve presumed immunity against a disease - within the context of an authority.
    }
    property series : String read GetSeriesST write SetSeriesST;
    {@member seriesElement
      One possible path to achieve presumed immunity against a disease - within the context of an authority.
    }
    property seriesElement : TFhirString read FSeries write SetSeries;

    {@member seriesDoses
      Typed access to The recommended number of doses to achieve immunity.
    }
    property seriesDoses : String read GetSeriesDosesST write SetSeriesDosesST;
    {@member seriesDosesElement
      The recommended number of doses to achieve immunity.
    }
    property seriesDosesElement : TFhirPositiveInt read FSeriesDoses write SetSeriesDoses;

    {@member targetDiseaseList
      The targeted disease.
    }
    property targetDiseaseList : TFhirCodeableConceptList read GetTargetDiseaseList;
    property hasTargetDiseaseList : boolean read GetHasTargetDiseaseList;

    {@member doseStatus
      Typed access to Indicates if the immunization event should "count" against  the protocol. (defined for API consistency)
    }
    property doseStatus : TFhirCodeableConcept read FDoseStatus write SetDoseStatus;
    {@member doseStatusElement
      Indicates if the immunization event should "count" against  the protocol.
    }
    property doseStatusElement : TFhirCodeableConcept read FDoseStatus write SetDoseStatus;

    {@member doseStatusReason
      Typed access to Provides an explanation as to why an immunization event should or should not count against the protocol. (defined for API consistency)
    }
    property doseStatusReason : TFhirCodeableConcept read FDoseStatusReason write SetDoseStatusReason;
    {@member doseStatusReasonElement
      Provides an explanation as to why an immunization event should or should not count against the protocol.
    }
    property doseStatusReasonElement : TFhirCodeableConcept read FDoseStatusReason write SetDoseStatusReason;

  end;


  TFhirImmunizationVaccinationProtocolListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationVaccinationProtocolList;
    function GetCurrent : TFhirImmunizationVaccinationProtocol;
  public
    Constructor Create(list : TFhirImmunizationVaccinationProtocolList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationVaccinationProtocol read GetCurrent;
  end;


  {@Class TFhirImmunizationVaccinationProtocolList
    A list of FhirImmunizationVaccinationProtocol
  }
  {!.Net HL7Connect.Fhir.ImmunizationVaccinationProtocolList}
  TFhirImmunizationVaccinationProtocolList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationVaccinationProtocol;
    procedure SetItemN(index : Integer; value : TFhirImmunizationVaccinationProtocol);
  public
    {!script hide}
    function Link : TFhirImmunizationVaccinationProtocolList; Overload;
    function Clone : TFhirImmunizationVaccinationProtocolList; Overload;
    function GetEnumerator : TFhirImmunizationVaccinationProtocolListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationVaccinationProtocol to the end of the list.
    }
    function Append : TFhirImmunizationVaccinationProtocol;

    
    {@member AddItem
      Add an already existing FhirImmunizationVaccinationProtocol to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationVaccinationProtocol); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationVaccinationProtocol) : Integer;
    

    {@member Insert
      Insert FhirImmunizationVaccinationProtocol before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationVaccinationProtocol;
    

    {@member InsertItem
       Insert an existing FhirImmunizationVaccinationProtocol before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationVaccinationProtocol);
    
    {@member Item
       Get the iIndexth FhirImmunizationVaccinationProtocol. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationVaccinationProtocol. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationVaccinationProtocol);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationVaccinationProtocol;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationVaccinationProtocols[index : Integer] : TFhirImmunizationVaccinationProtocol read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunization : TFhirDomainResource
    Describes the event of a patient being administered a vaccination or a record of a vaccination as reported by a patient, a clinician or another party and may include vaccine reaction information and what vaccination protocol was followed.
  }
  {!.Net HL7Connect.Fhir.Immunization}
  TFhirImmunization = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FDate : TFhirDateTime;
    FVaccineCode : TFhirCodeableConcept;
    FPatient : TFhirReference{TFhirPatient};
    FWasNotGiven : TFhirBoolean;
    FReported : TFhirBoolean;
    FPerformer : TFhirReference{TFhirPractitioner};
    FRequester : TFhirReference{TFhirPractitioner};
    FEncounter : TFhirReference{TFhirEncounter};
    FManufacturer : TFhirReference{TFhirOrganization};
    FLocation : TFhirReference{TFhirLocation};
    FLotNumber : TFhirString;
    FExpirationDate : TFhirDate;
    FSite : TFhirCodeableConcept;
    FRoute : TFhirCodeableConcept;
    FDoseQuantity : TFhirQuantity;
    FnoteList : TFhirAnnotationList;
    FExplanation : TFhirImmunizationExplanation;
    FreactionList : TFhirImmunizationReactionList;
    FvaccinationProtocolList : TFhirImmunizationVaccinationProtocolList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirMedicationAdminStatus;
    Procedure SetStatusST(value : TFhirMedicationAdminStatus);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetVaccineCode(value : TFhirCodeableConcept);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetWasNotGiven(value : TFhirBoolean);
    Function GetWasNotGivenST : Boolean;
    Procedure SetWasNotGivenST(value : Boolean);
    Procedure SetReported(value : TFhirBoolean);
    Function GetReportedST : Boolean;
    Procedure SetReportedST(value : Boolean);
    Procedure SetPerformer(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequester(value : TFhirReference{TFhirPractitioner});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetManufacturer(value : TFhirReference{TFhirOrganization});
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    Procedure SetLotNumber(value : TFhirString);
    Function GetLotNumberST : String;
    Procedure SetLotNumberST(value : String);
    Procedure SetExpirationDate(value : TFhirDate);
    Function GetExpirationDateST : TDateAndTime;
    Procedure SetExpirationDateST(value : TDateAndTime);
    Procedure SetSite(value : TFhirCodeableConcept);
    Procedure SetRoute(value : TFhirCodeableConcept);
    Procedure SetDoseQuantity(value : TFhirQuantity);
    function GetNoteList : TFhirAnnotationList;
    function GetHasNoteList : Boolean;
    Procedure SetExplanation(value : TFhirImmunizationExplanation);
    function GetReactionList : TFhirImmunizationReactionList;
    function GetHasReactionList : Boolean;
    function GetVaccinationProtocolList : TFhirImmunizationVaccinationProtocolList;
    function GetHasVaccinationProtocolList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunization; overload;
    function Clone : TFhirImmunization; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      A unique identifier assigned to this immunization record.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      Indicates the current status of the vaccination event.
    }
    property status : TFhirMedicationAdminStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member date
      Typed access to Date vaccine administered or was to be administered.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      Date vaccine administered or was to be administered.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member vaccineCode
      Typed access to Vaccine that was administered or was to be administered. (defined for API consistency)
    }
    property vaccineCode : TFhirCodeableConcept read FVaccineCode write SetVaccineCode;
    {@member vaccineCodeElement
      Vaccine that was administered or was to be administered.
    }
    property vaccineCodeElement : TFhirCodeableConcept read FVaccineCode write SetVaccineCode;

    {@member patient
      Typed access to The patient who either received or did not receive the immunization. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient who either received or did not receive the immunization.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member wasNotGiven
      Typed access to Indicates if the vaccination was or was not given.
    }
    property wasNotGiven : Boolean read GetWasNotGivenST write SetWasNotGivenST;
    {@member wasNotGivenElement
      Indicates if the vaccination was or was not given.
    }
    property wasNotGivenElement : TFhirBoolean read FWasNotGiven write SetWasNotGiven;

    {@member reported
      Typed access to True if this administration was reported rather than directly administered.
    }
    property reported : Boolean read GetReportedST write SetReportedST;
    {@member reportedElement
      True if this administration was reported rather than directly administered.
    }
    property reportedElement : TFhirBoolean read FReported write SetReported;

    {@member performer
      Typed access to Clinician who administered the vaccine. (defined for API consistency)
    }
    property performer : TFhirReference{TFhirPractitioner} read FPerformer write SetPerformer;
    {@member performerElement
      Clinician who administered the vaccine.
    }
    property performerElement : TFhirReference{TFhirPractitioner} read FPerformer write SetPerformer;

    {@member requester
      Typed access to Clinician who ordered the vaccination. (defined for API consistency)
    }
    property requester : TFhirReference{TFhirPractitioner} read FRequester write SetRequester;
    {@member requesterElement
      Clinician who ordered the vaccination.
    }
    property requesterElement : TFhirReference{TFhirPractitioner} read FRequester write SetRequester;

    {@member encounter
      Typed access to The visit or admission or other contact between patient and health care provider the immunization was performed as part of. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The visit or admission or other contact between patient and health care provider the immunization was performed as part of.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member manufacturer
      Typed access to Name of vaccine manufacturer. (defined for API consistency)
    }
    property manufacturer : TFhirReference{TFhirOrganization} read FManufacturer write SetManufacturer;
    {@member manufacturerElement
      Name of vaccine manufacturer.
    }
    property manufacturerElement : TFhirReference{TFhirOrganization} read FManufacturer write SetManufacturer;

    {@member location
      Typed access to The service delivery location where the vaccine administration occurred. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      The service delivery location where the vaccine administration occurred.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member lotNumber
      Typed access to Lot number of the  vaccine product.
    }
    property lotNumber : String read GetLotNumberST write SetLotNumberST;
    {@member lotNumberElement
      Lot number of the  vaccine product.
    }
    property lotNumberElement : TFhirString read FLotNumber write SetLotNumber;

    {@member expirationDate
      Typed access to Date vaccine batch expires.
    }
    property expirationDate : TDateAndTime read GetExpirationDateST write SetExpirationDateST;
    {@member expirationDateElement
      Date vaccine batch expires.
    }
    property expirationDateElement : TFhirDate read FExpirationDate write SetExpirationDate;

    {@member site
      Typed access to Body site where vaccine was administered. (defined for API consistency)
    }
    property site : TFhirCodeableConcept read FSite write SetSite;
    {@member siteElement
      Body site where vaccine was administered.
    }
    property siteElement : TFhirCodeableConcept read FSite write SetSite;

    {@member route
      Typed access to The path by which the vaccine product is taken into the body. (defined for API consistency)
    }
    property route : TFhirCodeableConcept read FRoute write SetRoute;
    {@member routeElement
      The path by which the vaccine product is taken into the body.
    }
    property routeElement : TFhirCodeableConcept read FRoute write SetRoute;

    {@member doseQuantity
      Typed access to The quantity of vaccine product that was administered. (defined for API consistency)
    }
    property doseQuantity : TFhirQuantity read FDoseQuantity write SetDoseQuantity;
    {@member doseQuantityElement
      The quantity of vaccine product that was administered.
    }
    property doseQuantityElement : TFhirQuantity read FDoseQuantity write SetDoseQuantity;

    {@member noteList
      Extra information about the immunization that is not conveyed by the other attributes.
    }
    property noteList : TFhirAnnotationList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

    {@member explanation
      Typed access to Reasons why a vaccine was or was not administered. (defined for API consistency)
    }
    property explanation : TFhirImmunizationExplanation read FExplanation write SetExplanation;
    {@member explanationElement
      Reasons why a vaccine was or was not administered.
    }
    property explanationElement : TFhirImmunizationExplanation read FExplanation write SetExplanation;

    {@member reactionList
      Categorical data indicating that an adverse event is associated in time to an immunization.
    }
    property reactionList : TFhirImmunizationReactionList read GetReactionList;
    property hasReactionList : boolean read GetHasReactionList;

    {@member vaccinationProtocolList
      Contains information about the protocol(s) under which the vaccine was administered.
    }
    property vaccinationProtocolList : TFhirImmunizationVaccinationProtocolList read GetVaccinationProtocolList;
    property hasVaccinationProtocolList : boolean read GetHasVaccinationProtocolList;

  end;


  TFhirImmunizationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationList;
    function GetCurrent : TFhirImmunization;
  public
    Constructor Create(list : TFhirImmunizationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunization read GetCurrent;
  end;


  {@Class TFhirImmunizationList
    A list of FhirImmunization
  }
  {!.Net HL7Connect.Fhir.ImmunizationList}
  TFhirImmunizationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunization;
    procedure SetItemN(index : Integer; value : TFhirImmunization);
  public
    {!script hide}
    function Link : TFhirImmunizationList; Overload;
    function Clone : TFhirImmunizationList; Overload;
    function GetEnumerator : TFhirImmunizationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunization to the end of the list.
    }
    function Append : TFhirImmunization;

    
    {@member AddItem
      Add an already existing FhirImmunization to the end of the list.
    }
    procedure AddItem(value : TFhirImmunization); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunization) : Integer;
    

    {@member Insert
      Insert FhirImmunization before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunization;
    

    {@member InsertItem
       Insert an existing FhirImmunization before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunization);
    
    {@member Item
       Get the iIndexth FhirImmunization. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunization. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunization);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunization;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizations[index : Integer] : TFhirImmunization read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationRecommendationRecommendation : TFhirElement
    Vaccine administration recommendations.
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendation}
  TFhirImmunizationRecommendationRecommendation = class (TFhirBackboneElement)
  private
    FDate : TFhirDateTime;
    FVaccineCode : TFhirCodeableConcept;
    FDoseNumber : TFhirPositiveInt;
    FForecastStatus : TFhirCodeableConcept;
    FdateCriterionList : TFhirImmunizationRecommendationRecommendationDateCriterionList;
    FProtocol : TFhirImmunizationRecommendationRecommendationProtocol;
    FsupportingImmunizationList : TFhirReferenceList{TFhirImmunization};
    FsupportingPatientInformationList : TFhirReferenceList{Resource};
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetVaccineCode(value : TFhirCodeableConcept);
    Procedure SetDoseNumber(value : TFhirPositiveInt);
    Function GetDoseNumberST : String;
    Procedure SetDoseNumberST(value : String);
    Procedure SetForecastStatus(value : TFhirCodeableConcept);
    function GetDateCriterionList : TFhirImmunizationRecommendationRecommendationDateCriterionList;
    function GetHasDateCriterionList : Boolean;
    Procedure SetProtocol(value : TFhirImmunizationRecommendationRecommendationProtocol);
    function GetSupportingImmunizationList : TFhirReferenceList{TFhirImmunization};
    function GetHasSupportingImmunizationList : Boolean;
    function GetSupportingPatientInformationList : TFhirReferenceList{Resource};
    function GetHasSupportingPatientInformationList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationRecommendationRecommendation; overload;
    function Clone : TFhirImmunizationRecommendationRecommendation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member date
      Typed access to The date the immunization recommendation was created.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date the immunization recommendation was created.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member vaccineCode
      Typed access to Vaccine that pertains to the recommendation. (defined for API consistency)
    }
    property vaccineCode : TFhirCodeableConcept read FVaccineCode write SetVaccineCode;
    {@member vaccineCodeElement
      Vaccine that pertains to the recommendation.
    }
    property vaccineCodeElement : TFhirCodeableConcept read FVaccineCode write SetVaccineCode;

    {@member doseNumber
      Typed access to This indicates the next recommended dose number (e.g. dose 2 is the next recommended dose).
    }
    property doseNumber : String read GetDoseNumberST write SetDoseNumberST;
    {@member doseNumberElement
      This indicates the next recommended dose number (e.g. dose 2 is the next recommended dose).
    }
    property doseNumberElement : TFhirPositiveInt read FDoseNumber write SetDoseNumber;

    {@member forecastStatus
      Typed access to Vaccine administration status. (defined for API consistency)
    }
    property forecastStatus : TFhirCodeableConcept read FForecastStatus write SetForecastStatus;
    {@member forecastStatusElement
      Vaccine administration status.
    }
    property forecastStatusElement : TFhirCodeableConcept read FForecastStatus write SetForecastStatus;

    {@member dateCriterionList
      Vaccine date recommendations.  For example, earliest date to administer, latest date to administer, etc.
    }
    property dateCriterionList : TFhirImmunizationRecommendationRecommendationDateCriterionList read GetDateCriterionList;
    property hasDateCriterionList : boolean read GetHasDateCriterionList;

    {@member protocol
      Typed access to Contains information about the protocol under which the vaccine was administered. (defined for API consistency)
    }
    property protocol : TFhirImmunizationRecommendationRecommendationProtocol read FProtocol write SetProtocol;
    {@member protocolElement
      Contains information about the protocol under which the vaccine was administered.
    }
    property protocolElement : TFhirImmunizationRecommendationRecommendationProtocol read FProtocol write SetProtocol;

    {@member supportingImmunizationList
      Immunization event history that supports the status and recommendation.
    }
    property supportingImmunizationList : TFhirReferenceList{TFhirImmunization} read GetSupportingImmunizationList;
    property hasSupportingImmunizationList : boolean read GetHasSupportingImmunizationList;

    {@member supportingPatientInformationList
      Patient Information that supports the status and recommendation.  This includes patient observations, adverse reactions and allergy/intolerance information.
    }
    property supportingPatientInformationList : TFhirReferenceList{Resource} read GetSupportingPatientInformationList;
    property hasSupportingPatientInformationList : boolean read GetHasSupportingPatientInformationList;

  end;


  TFhirImmunizationRecommendationRecommendationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationRecommendationRecommendationList;
    function GetCurrent : TFhirImmunizationRecommendationRecommendation;
  public
    Constructor Create(list : TFhirImmunizationRecommendationRecommendationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationRecommendationRecommendation read GetCurrent;
  end;


  {@Class TFhirImmunizationRecommendationRecommendationList
    A list of FhirImmunizationRecommendationRecommendation
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendationList}
  TFhirImmunizationRecommendationRecommendationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationRecommendationRecommendation;
    procedure SetItemN(index : Integer; value : TFhirImmunizationRecommendationRecommendation);
  public
    {!script hide}
    function Link : TFhirImmunizationRecommendationRecommendationList; Overload;
    function Clone : TFhirImmunizationRecommendationRecommendationList; Overload;
    function GetEnumerator : TFhirImmunizationRecommendationRecommendationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationRecommendationRecommendation to the end of the list.
    }
    function Append : TFhirImmunizationRecommendationRecommendation;

    
    {@member AddItem
      Add an already existing FhirImmunizationRecommendationRecommendation to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationRecommendationRecommendation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationRecommendationRecommendation) : Integer;
    

    {@member Insert
      Insert FhirImmunizationRecommendationRecommendation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationRecommendationRecommendation;
    

    {@member InsertItem
       Insert an existing FhirImmunizationRecommendationRecommendation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationRecommendationRecommendation);
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationRecommendationRecommendation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationRecommendationRecommendation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationRecommendationRecommendations[index : Integer] : TFhirImmunizationRecommendationRecommendation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationRecommendationRecommendationDateCriterion : TFhirElement
    Vaccine date recommendations.  For example, earliest date to administer, latest date to administer, etc.
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendationDateCriterion}
  TFhirImmunizationRecommendationRecommendationDateCriterion = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FValue : TFhirDateTime;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetValue(value : TFhirDateTime);
    Function GetValueST : TDateAndTime;
    Procedure SetValueST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationRecommendationRecommendationDateCriterion; overload;
    function Clone : TFhirImmunizationRecommendationRecommendationDateCriterion; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Date classification of recommendation.  For example, earliest date to give, latest date to give, etc. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Date classification of recommendation.  For example, earliest date to give, latest date to give, etc.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member value
      Typed access to The date whose meaning is specified by dateCriterion.code.
    }
    property value : TDateAndTime read GetValueST write SetValueST;
    {@member valueElement
      The date whose meaning is specified by dateCriterion.code.
    }
    property valueElement : TFhirDateTime read FValue write SetValue;

  end;


  TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationRecommendationRecommendationDateCriterionList;
    function GetCurrent : TFhirImmunizationRecommendationRecommendationDateCriterion;
  public
    Constructor Create(list : TFhirImmunizationRecommendationRecommendationDateCriterionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationRecommendationRecommendationDateCriterion read GetCurrent;
  end;


  {@Class TFhirImmunizationRecommendationRecommendationDateCriterionList
    A list of FhirImmunizationRecommendationRecommendationDateCriterion
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendationDateCriterionList}
  TFhirImmunizationRecommendationRecommendationDateCriterionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationRecommendationRecommendationDateCriterion;
    procedure SetItemN(index : Integer; value : TFhirImmunizationRecommendationRecommendationDateCriterion);
  public
    {!script hide}
    function Link : TFhirImmunizationRecommendationRecommendationDateCriterionList; Overload;
    function Clone : TFhirImmunizationRecommendationRecommendationDateCriterionList; Overload;
    function GetEnumerator : TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationRecommendationRecommendationDateCriterion to the end of the list.
    }
    function Append : TFhirImmunizationRecommendationRecommendationDateCriterion;

    
    {@member AddItem
      Add an already existing FhirImmunizationRecommendationRecommendationDateCriterion to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationRecommendationRecommendationDateCriterion); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationRecommendationRecommendationDateCriterion) : Integer;
    

    {@member Insert
      Insert FhirImmunizationRecommendationRecommendationDateCriterion before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationRecommendationRecommendationDateCriterion;
    

    {@member InsertItem
       Insert an existing FhirImmunizationRecommendationRecommendationDateCriterion before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationRecommendationRecommendationDateCriterion);
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendationDateCriterion. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendationDateCriterion. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationRecommendationRecommendationDateCriterion);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationRecommendationRecommendationDateCriterion;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationRecommendationRecommendationDateCriterions[index : Integer] : TFhirImmunizationRecommendationRecommendationDateCriterion read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationRecommendationRecommendationProtocol : TFhirElement
    Contains information about the protocol under which the vaccine was administered.
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendationProtocol}
  TFhirImmunizationRecommendationRecommendationProtocol = class (TFhirBackboneElement)
  private
    FDoseSequence : TFhirInteger;
    FDescription : TFhirString;
    FAuthority : TFhirReference{TFhirOrganization};
    FSeries : TFhirString;
    Procedure SetDoseSequence(value : TFhirInteger);
    Function GetDoseSequenceST : String;
    Procedure SetDoseSequenceST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetAuthority(value : TFhirReference{TFhirOrganization});
    Procedure SetSeries(value : TFhirString);
    Function GetSeriesST : String;
    Procedure SetSeriesST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationRecommendationRecommendationProtocol; overload;
    function Clone : TFhirImmunizationRecommendationRecommendationProtocol; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member doseSequence
      Typed access to Indicates the nominal position in a series of the next dose.  This is the recommended dose number as per a specified protocol.
    }
    property doseSequence : String read GetDoseSequenceST write SetDoseSequenceST;
    {@member doseSequenceElement
      Indicates the nominal position in a series of the next dose.  This is the recommended dose number as per a specified protocol.
    }
    property doseSequenceElement : TFhirInteger read FDoseSequence write SetDoseSequence;

    {@member description
      Typed access to Contains the description about the protocol under which the vaccine was administered.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Contains the description about the protocol under which the vaccine was administered.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member authority
      Typed access to Indicates the authority who published the protocol.  For example, ACIP. (defined for API consistency)
    }
    property authority : TFhirReference{TFhirOrganization} read FAuthority write SetAuthority;
    {@member authorityElement
      Indicates the authority who published the protocol.  For example, ACIP.
    }
    property authorityElement : TFhirReference{TFhirOrganization} read FAuthority write SetAuthority;

    {@member series
      Typed access to One possible path to achieve presumed immunity against a disease - within the context of an authority.
    }
    property series : String read GetSeriesST write SetSeriesST;
    {@member seriesElement
      One possible path to achieve presumed immunity against a disease - within the context of an authority.
    }
    property seriesElement : TFhirString read FSeries write SetSeries;

  end;


  TFhirImmunizationRecommendationRecommendationProtocolListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationRecommendationRecommendationProtocolList;
    function GetCurrent : TFhirImmunizationRecommendationRecommendationProtocol;
  public
    Constructor Create(list : TFhirImmunizationRecommendationRecommendationProtocolList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationRecommendationRecommendationProtocol read GetCurrent;
  end;


  {@Class TFhirImmunizationRecommendationRecommendationProtocolList
    A list of FhirImmunizationRecommendationRecommendationProtocol
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationRecommendationProtocolList}
  TFhirImmunizationRecommendationRecommendationProtocolList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationRecommendationRecommendationProtocol;
    procedure SetItemN(index : Integer; value : TFhirImmunizationRecommendationRecommendationProtocol);
  public
    {!script hide}
    function Link : TFhirImmunizationRecommendationRecommendationProtocolList; Overload;
    function Clone : TFhirImmunizationRecommendationRecommendationProtocolList; Overload;
    function GetEnumerator : TFhirImmunizationRecommendationRecommendationProtocolListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationRecommendationRecommendationProtocol to the end of the list.
    }
    function Append : TFhirImmunizationRecommendationRecommendationProtocol;

    
    {@member AddItem
      Add an already existing FhirImmunizationRecommendationRecommendationProtocol to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationRecommendationRecommendationProtocol); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationRecommendationRecommendationProtocol) : Integer;
    

    {@member Insert
      Insert FhirImmunizationRecommendationRecommendationProtocol before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationRecommendationRecommendationProtocol;
    

    {@member InsertItem
       Insert an existing FhirImmunizationRecommendationRecommendationProtocol before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationRecommendationRecommendationProtocol);
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendationProtocol. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendationRecommendationProtocol. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationRecommendationRecommendationProtocol);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationRecommendationRecommendationProtocol;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationRecommendationRecommendationProtocols[index : Integer] : TFhirImmunizationRecommendationRecommendationProtocol read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImmunizationRecommendation : TFhirDomainResource
    A patient's point-in-time immunization and recommendation (i.e. forecasting a patient's immunization eligibility according to a published schedule) with optional supporting justification.
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendation}
  TFhirImmunizationRecommendation = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FPatient : TFhirReference{TFhirPatient};
    FrecommendationList : TFhirImmunizationRecommendationRecommendationList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    function GetRecommendationList : TFhirImmunizationRecommendationRecommendationList;
    function GetHasRecommendationList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImmunizationRecommendation; overload;
    function Clone : TFhirImmunizationRecommendation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      A unique identifier assigned to this particular recommendation record.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member patient
      Typed access to The patient for whom the recommendations are for. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient for whom the recommendations are for.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member recommendationList
      Vaccine administration recommendations.
    }
    property recommendationList : TFhirImmunizationRecommendationRecommendationList read GetRecommendationList;
    property hasRecommendationList : boolean read GetHasRecommendationList;

  end;


  TFhirImmunizationRecommendationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImmunizationRecommendationList;
    function GetCurrent : TFhirImmunizationRecommendation;
  public
    Constructor Create(list : TFhirImmunizationRecommendationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImmunizationRecommendation read GetCurrent;
  end;


  {@Class TFhirImmunizationRecommendationList
    A list of FhirImmunizationRecommendation
  }
  {!.Net HL7Connect.Fhir.ImmunizationRecommendationList}
  TFhirImmunizationRecommendationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImmunizationRecommendation;
    procedure SetItemN(index : Integer; value : TFhirImmunizationRecommendation);
  public
    {!script hide}
    function Link : TFhirImmunizationRecommendationList; Overload;
    function Clone : TFhirImmunizationRecommendationList; Overload;
    function GetEnumerator : TFhirImmunizationRecommendationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImmunizationRecommendation to the end of the list.
    }
    function Append : TFhirImmunizationRecommendation;

    
    {@member AddItem
      Add an already existing FhirImmunizationRecommendation to the end of the list.
    }
    procedure AddItem(value : TFhirImmunizationRecommendation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImmunizationRecommendation) : Integer;
    

    {@member Insert
      Insert FhirImmunizationRecommendation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImmunizationRecommendation;
    

    {@member InsertItem
       Insert an existing FhirImmunizationRecommendation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImmunizationRecommendation);
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImmunizationRecommendation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImmunizationRecommendation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImmunizationRecommendation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImmunizationRecommendations[index : Integer] : TFhirImmunizationRecommendation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuideContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideContact}
  TFhirImplementationGuideContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuideContact; overload;
    function Clone : TFhirImplementationGuideContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the implementation guide.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the implementation guide.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirImplementationGuideContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuideContactList;
    function GetCurrent : TFhirImplementationGuideContact;
  public
    Constructor Create(list : TFhirImplementationGuideContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuideContact read GetCurrent;
  end;


  {@Class TFhirImplementationGuideContactList
    A list of FhirImplementationGuideContact
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideContactList}
  TFhirImplementationGuideContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuideContact;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuideContact);
  public
    {!script hide}
    function Link : TFhirImplementationGuideContactList; Overload;
    function Clone : TFhirImplementationGuideContactList; Overload;
    function GetEnumerator : TFhirImplementationGuideContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuideContact to the end of the list.
    }
    function Append : TFhirImplementationGuideContact;

    
    {@member AddItem
      Add an already existing FhirImplementationGuideContact to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuideContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuideContact) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuideContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuideContact;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuideContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuideContact);
    
    {@member Item
       Get the iIndexth FhirImplementationGuideContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuideContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuideContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuideContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuideContacts[index : Integer] : TFhirImplementationGuideContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuideDependency : TFhirElement
    Another implementation guide that this implementation depends on. Typically, an implementation guide uses value sets, profiles etc.defined in other implementation guides.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideDependency}
  TFhirImplementationGuideDependency = class (TFhirBackboneElement)
  private
    FType_ : TFhirEnum;
    FUri : TFhirUri;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirGuideDependencyType;
    Procedure SetType_ST(value : TFhirGuideDependencyType);
    Procedure SetUri(value : TFhirUri);
    Function GetUriST : String;
    Procedure SetUriST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuideDependency; overload;
    function Clone : TFhirImplementationGuideDependency; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      How the dependency is represented when the guide is published.
    }
    property type_ : TFhirGuideDependencyType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member uri
      Typed access to Where the dependency is located.
    }
    property uri : String read GetUriST write SetUriST;
    {@member uriElement
      Where the dependency is located.
    }
    property uriElement : TFhirUri read FUri write SetUri;

  end;


  TFhirImplementationGuideDependencyListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuideDependencyList;
    function GetCurrent : TFhirImplementationGuideDependency;
  public
    Constructor Create(list : TFhirImplementationGuideDependencyList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuideDependency read GetCurrent;
  end;


  {@Class TFhirImplementationGuideDependencyList
    A list of FhirImplementationGuideDependency
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideDependencyList}
  TFhirImplementationGuideDependencyList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuideDependency;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuideDependency);
  public
    {!script hide}
    function Link : TFhirImplementationGuideDependencyList; Overload;
    function Clone : TFhirImplementationGuideDependencyList; Overload;
    function GetEnumerator : TFhirImplementationGuideDependencyListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuideDependency to the end of the list.
    }
    function Append : TFhirImplementationGuideDependency;

    
    {@member AddItem
      Add an already existing FhirImplementationGuideDependency to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuideDependency); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuideDependency) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuideDependency before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuideDependency;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuideDependency before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuideDependency);
    
    {@member Item
       Get the iIndexth FhirImplementationGuideDependency. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuideDependency. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuideDependency);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuideDependency;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuideDependencies[index : Integer] : TFhirImplementationGuideDependency read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuidePackage : TFhirElement
    A logical group of resources. Logical groups can be used when building pages.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePackage}
  TFhirImplementationGuidePackage = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FDescription : TFhirString;
    FresourceList : TFhirImplementationGuidePackageResourceList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetResourceList : TFhirImplementationGuidePackageResourceList;
    function GetHasResourceList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuidePackage; overload;
    function Clone : TFhirImplementationGuidePackage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name for the group, as used in page.package.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name for the group, as used in page.package.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member description
      Typed access to Human readable text describing the package.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Human readable text describing the package.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member resourceList
      A resource that is part of the implementation guide. Conformance resources (value set, structure definition, conformance statements etc.) are obvious candidates for inclusion, but any kind of resource can be included as an example resource.
    }
    property resourceList : TFhirImplementationGuidePackageResourceList read GetResourceList;
    property hasResourceList : boolean read GetHasResourceList;

  end;


  TFhirImplementationGuidePackageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuidePackageList;
    function GetCurrent : TFhirImplementationGuidePackage;
  public
    Constructor Create(list : TFhirImplementationGuidePackageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuidePackage read GetCurrent;
  end;


  {@Class TFhirImplementationGuidePackageList
    A list of FhirImplementationGuidePackage
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePackageList}
  TFhirImplementationGuidePackageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuidePackage;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuidePackage);
  public
    {!script hide}
    function Link : TFhirImplementationGuidePackageList; Overload;
    function Clone : TFhirImplementationGuidePackageList; Overload;
    function GetEnumerator : TFhirImplementationGuidePackageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuidePackage to the end of the list.
    }
    function Append : TFhirImplementationGuidePackage;

    
    {@member AddItem
      Add an already existing FhirImplementationGuidePackage to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuidePackage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuidePackage) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuidePackage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuidePackage;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuidePackage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuidePackage);
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePackage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePackage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuidePackage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuidePackage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuidePackages[index : Integer] : TFhirImplementationGuidePackage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuidePackageResource : TFhirElement
    A resource that is part of the implementation guide. Conformance resources (value set, structure definition, conformance statements etc.) are obvious candidates for inclusion, but any kind of resource can be included as an example resource.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePackageResource}
  TFhirImplementationGuidePackageResource = class (TFhirBackboneElement)
  private
    FPurpose : TFhirEnum;
    FName : TFhirString;
    FDescription : TFhirString;
    FAcronym : TFhirString;
    FSource : TFhirType;
    FExampleFor : TFhirReference{TFhirStructureDefinition};
    Procedure SetPurpose(value : TFhirEnum);
    Function GetPurposeST : TFhirGuideResourcePurpose;
    Procedure SetPurposeST(value : TFhirGuideResourcePurpose);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetAcronym(value : TFhirString);
    Function GetAcronymST : String;
    Procedure SetAcronymST(value : String);
    Procedure SetSource(value : TFhirType);
    Procedure SetExampleFor(value : TFhirReference{TFhirStructureDefinition});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuidePackageResource; overload;
    function Clone : TFhirImplementationGuidePackageResource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member purpose
      Why the resource is included in the guide.
    }
    property purpose : TFhirGuideResourcePurpose read GetPurposeST write SetPurposeST;
    property purposeElement : TFhirEnum read FPurpose write SetPurpose;

    {@member name
      Typed access to A human assigned name for the resource. All resources SHOULD have a name, but the name may be extracted from the resource (e.g. ValueSet.name).
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A human assigned name for the resource. All resources SHOULD have a name, but the name may be extracted from the resource (e.g. ValueSet.name).
    }
    property nameElement : TFhirString read FName write SetName;

    {@member description
      Typed access to A description of the reason that a resource has been included in the implementation guide.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A description of the reason that a resource has been included in the implementation guide.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member acronym
      Typed access to A short code that may be used to identify the resource throughout the implementation guide.
    }
    property acronym : String read GetAcronymST write SetAcronymST;
    {@member acronymElement
      A short code that may be used to identify the resource throughout the implementation guide.
    }
    property acronymElement : TFhirString read FAcronym write SetAcronym;

    {@member source
      Typed access to Where this resource is found. (defined for API consistency)
    }
    property source : TFhirType read FSource write SetSource;
    {@member sourceElement
      Where this resource is found.
    }
    property sourceElement : TFhirType read FSource write SetSource;

    {@member exampleFor
      Typed access to Another resource that this resource is an example for. This is mostly used for resources that are included as examples of StructureDefinitions. (defined for API consistency)
    }
    property exampleFor : TFhirReference{TFhirStructureDefinition} read FExampleFor write SetExampleFor;
    {@member exampleForElement
      Another resource that this resource is an example for. This is mostly used for resources that are included as examples of StructureDefinitions.
    }
    property exampleForElement : TFhirReference{TFhirStructureDefinition} read FExampleFor write SetExampleFor;

  end;


  TFhirImplementationGuidePackageResourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuidePackageResourceList;
    function GetCurrent : TFhirImplementationGuidePackageResource;
  public
    Constructor Create(list : TFhirImplementationGuidePackageResourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuidePackageResource read GetCurrent;
  end;


  {@Class TFhirImplementationGuidePackageResourceList
    A list of FhirImplementationGuidePackageResource
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePackageResourceList}
  TFhirImplementationGuidePackageResourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuidePackageResource;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuidePackageResource);
  public
    {!script hide}
    function Link : TFhirImplementationGuidePackageResourceList; Overload;
    function Clone : TFhirImplementationGuidePackageResourceList; Overload;
    function GetEnumerator : TFhirImplementationGuidePackageResourceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuidePackageResource to the end of the list.
    }
    function Append : TFhirImplementationGuidePackageResource;

    
    {@member AddItem
      Add an already existing FhirImplementationGuidePackageResource to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuidePackageResource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuidePackageResource) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuidePackageResource before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuidePackageResource;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuidePackageResource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuidePackageResource);
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePackageResource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePackageResource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuidePackageResource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuidePackageResource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuidePackageResources[index : Integer] : TFhirImplementationGuidePackageResource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuideGlobal : TFhirElement
    A set of profiles that all resources covered by this implementation guide must conform to.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideGlobal}
  TFhirImplementationGuideGlobal = class (TFhirBackboneElement)
  private
    FType_ : TFhirCode;
    FProfile : TFhirReference{TFhirStructureDefinition};
    Procedure SetType_(value : TFhirCode);
    Function GetType_ST : String;
    Procedure SetType_ST(value : String);
    Procedure SetProfile(value : TFhirReference{TFhirStructureDefinition});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuideGlobal; overload;
    function Clone : TFhirImplementationGuideGlobal; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The type of resource that all instances must conform to.
    }
    property type_ : String read GetType_ST write SetType_ST;
    {@member type_Element
      The type of resource that all instances must conform to.
    }
    property type_Element : TFhirCode read FType_ write SetType_;

    {@member profile
      Typed access to A reference to the profile that all instances must conform to. (defined for API consistency)
    }
    property profile : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;
    {@member profileElement
      A reference to the profile that all instances must conform to.
    }
    property profileElement : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;

  end;


  TFhirImplementationGuideGlobalListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuideGlobalList;
    function GetCurrent : TFhirImplementationGuideGlobal;
  public
    Constructor Create(list : TFhirImplementationGuideGlobalList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuideGlobal read GetCurrent;
  end;


  {@Class TFhirImplementationGuideGlobalList
    A list of FhirImplementationGuideGlobal
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideGlobalList}
  TFhirImplementationGuideGlobalList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuideGlobal;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuideGlobal);
  public
    {!script hide}
    function Link : TFhirImplementationGuideGlobalList; Overload;
    function Clone : TFhirImplementationGuideGlobalList; Overload;
    function GetEnumerator : TFhirImplementationGuideGlobalListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuideGlobal to the end of the list.
    }
    function Append : TFhirImplementationGuideGlobal;

    
    {@member AddItem
      Add an already existing FhirImplementationGuideGlobal to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuideGlobal); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuideGlobal) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuideGlobal before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuideGlobal;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuideGlobal before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuideGlobal);
    
    {@member Item
       Get the iIndexth FhirImplementationGuideGlobal. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuideGlobal. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuideGlobal);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuideGlobal;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuideGlobals[index : Integer] : TFhirImplementationGuideGlobal read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuidePage : TFhirElement
    A page / section in the implementation guide. The root page is the implementation guide home page.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePage}
  TFhirImplementationGuidePage = class (TFhirBackboneElement)
  private
    FSource : TFhirUri;
    FName : TFhirString;
    FKind : TFhirEnum;
    Ftype_List : TFhirCodeList;
    FpackageList : TFhirStringList;
    FFormat : TFhirCode;
    FpageList : TFhirImplementationGuidePageList;
    Procedure SetSource(value : TFhirUri);
    Function GetSourceST : String;
    Procedure SetSourceST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetKind(value : TFhirEnum);
    Function GetKindST : TFhirGuidePageKind;
    Procedure SetKindST(value : TFhirGuidePageKind);
    function GetType_List : TFhirCodeList;
    function GetHasType_List : Boolean;
    function GetPackageList : TFhirStringList;
    function GetHasPackageList : Boolean;
    Procedure SetFormat(value : TFhirCode);
    Function GetFormatST : String;
    Procedure SetFormatST(value : String);
    function GetPageList : TFhirImplementationGuidePageList;
    function GetHasPageList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuidePage; overload;
    function Clone : TFhirImplementationGuidePage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member source
      Typed access to The source address for the page.
    }
    property source : String read GetSourceST write SetSourceST;
    {@member sourceElement
      The source address for the page.
    }
    property sourceElement : TFhirUri read FSource write SetSource;

    {@member name
      Typed access to A short name used to represent this page in navigational structures such as table of contents, bread crumbs, etc.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A short name used to represent this page in navigational structures such as table of contents, bread crumbs, etc.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member kind
      The kind of page that this is. Some pages are autogenerated (list, example), and other kinds are of interest so that tools can navigate the user to the page of interest.
    }
    property kind : TFhirGuidePageKind read GetKindST write SetKindST;
    property kindElement : TFhirEnum read FKind write SetKind;

    {@member type_List
      For constructed pages, what kind of resources to include in the list.
    }
    property type_List : TFhirCodeList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member packageList
      For constructed pages, a list of packages to include in the page (or else empty for everything).
    }
    property packageList : TFhirStringList read GetPackageList;
    property hasPackageList : boolean read GetHasPackageList;

    {@member format
      Typed access to The format of the page.
    }
    property format : String read GetFormatST write SetFormatST;
    {@member formatElement
      The format of the page.
    }
    property formatElement : TFhirCode read FFormat write SetFormat;

    {@member pageList
      Nested Pages/Sections under this page.
    }
    property pageList : TFhirImplementationGuidePageList read GetPageList;
    property hasPageList : boolean read GetHasPageList;

  end;


  TFhirImplementationGuidePageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuidePageList;
    function GetCurrent : TFhirImplementationGuidePage;
  public
    Constructor Create(list : TFhirImplementationGuidePageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuidePage read GetCurrent;
  end;


  {@Class TFhirImplementationGuidePageList
    A list of FhirImplementationGuidePage
  }
  {!.Net HL7Connect.Fhir.ImplementationGuidePageList}
  TFhirImplementationGuidePageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuidePage;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuidePage);
  public
    {!script hide}
    function Link : TFhirImplementationGuidePageList; Overload;
    function Clone : TFhirImplementationGuidePageList; Overload;
    function GetEnumerator : TFhirImplementationGuidePageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuidePage to the end of the list.
    }
    function Append : TFhirImplementationGuidePage;

    
    {@member AddItem
      Add an already existing FhirImplementationGuidePage to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuidePage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuidePage) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuidePage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuidePage;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuidePage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuidePage);
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuidePage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuidePage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuidePage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuidePages[index : Integer] : TFhirImplementationGuidePage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirImplementationGuide : TFhirDomainResource
    A set of rules or how FHIR is used to solve a particular problem. This resource is used to gather all the parts of an implementation guide into a logical whole, and to publish a computable definition of all the parts.
  }
  {!.Net HL7Connect.Fhir.ImplementationGuide}
  TFhirImplementationGuide = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirImplementationGuideContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FCopyright : TFhirString;
    FFhirVersion : TFhirId;
    FdependencyList : TFhirImplementationGuideDependencyList;
    FpackageList : TFhirImplementationGuidePackageList;
    FglobalList : TFhirImplementationGuideGlobalList;
    FbinaryList : TFhirUriList;
    FPage : TFhirImplementationGuidePage;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirImplementationGuideContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetFhirVersion(value : TFhirId);
    Function GetFhirVersionST : String;
    Procedure SetFhirVersionST(value : String);
    function GetDependencyList : TFhirImplementationGuideDependencyList;
    function GetHasDependencyList : Boolean;
    function GetPackageList : TFhirImplementationGuidePackageList;
    function GetHasPackageList : Boolean;
    function GetGlobalList : TFhirImplementationGuideGlobalList;
    function GetHasGlobalList : Boolean;
    function GetBinaryList : TFhirUriList;
    function GetHasBinaryList : Boolean;
    Procedure SetPage(value : TFhirImplementationGuidePage);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirImplementationGuide; overload;
    function Clone : TFhirImplementationGuide; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this implementation guide when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this implementation guide is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this implementation guide when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this implementation guide is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member version
      Typed access to The identifier that is used to identify this version of the Implementation Guide when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the Implementation Guide author manually.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the Implementation Guide when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the Implementation Guide author manually.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name identifying the Implementation Guide.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the Implementation Guide.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the Implementation Guide.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to This Implementation Guide was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This Implementation Guide was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the implementation guide.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the implementation guide.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirImplementationGuideContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the implementation guide was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the implementation guide changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the implementation guide was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the implementation guide changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the Implementation Guide and its use.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the Implementation Guide and its use.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of implementation guides. The most common use of this element is to represent the country / jurisdiction for which this implementation guide was defined.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member copyright
      Typed access to A copyright statement relating to the implementation guide and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the implementation guide and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member fhirVersion
      Typed access to The version of the FHIR specification on which this ImplementationGuide is based - this is the formal version of the specification, without the revision number, e.g. [publication].[major].[minor], which is 1.0.2 for this version.
    }
    property fhirVersion : String read GetFhirVersionST write SetFhirVersionST;
    {@member fhirVersionElement
      The version of the FHIR specification on which this ImplementationGuide is based - this is the formal version of the specification, without the revision number, e.g. [publication].[major].[minor], which is 1.0.2 for this version.
    }
    property fhirVersionElement : TFhirId read FFhirVersion write SetFhirVersion;

    {@member dependencyList
      Another implementation guide that this implementation depends on. Typically, an implementation guide uses value sets, profiles etc.defined in other implementation guides.
    }
    property dependencyList : TFhirImplementationGuideDependencyList read GetDependencyList;
    property hasDependencyList : boolean read GetHasDependencyList;

    {@member packageList
      A logical group of resources. Logical groups can be used when building pages.
    }
    property packageList : TFhirImplementationGuidePackageList read GetPackageList;
    property hasPackageList : boolean read GetHasPackageList;

    {@member globalList
      A set of profiles that all resources covered by this implementation guide must conform to.
    }
    property globalList : TFhirImplementationGuideGlobalList read GetGlobalList;
    property hasGlobalList : boolean read GetHasGlobalList;

    {@member binaryList
      A binary file that is included in the  implementation guide when it is published.
    }
    property binaryList : TFhirUriList read GetBinaryList;
    property hasBinaryList : boolean read GetHasBinaryList;

    {@member page
      Typed access to A page / section in the implementation guide. The root page is the implementation guide home page. (defined for API consistency)
    }
    property page : TFhirImplementationGuidePage read FPage write SetPage;
    {@member pageElement
      A page / section in the implementation guide. The root page is the implementation guide home page.
    }
    property pageElement : TFhirImplementationGuidePage read FPage write SetPage;

  end;


  TFhirImplementationGuideListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirImplementationGuideList;
    function GetCurrent : TFhirImplementationGuide;
  public
    Constructor Create(list : TFhirImplementationGuideList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirImplementationGuide read GetCurrent;
  end;


  {@Class TFhirImplementationGuideList
    A list of FhirImplementationGuide
  }
  {!.Net HL7Connect.Fhir.ImplementationGuideList}
  TFhirImplementationGuideList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirImplementationGuide;
    procedure SetItemN(index : Integer; value : TFhirImplementationGuide);
  public
    {!script hide}
    function Link : TFhirImplementationGuideList; Overload;
    function Clone : TFhirImplementationGuideList; Overload;
    function GetEnumerator : TFhirImplementationGuideListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirImplementationGuide to the end of the list.
    }
    function Append : TFhirImplementationGuide;

    
    {@member AddItem
      Add an already existing FhirImplementationGuide to the end of the list.
    }
    procedure AddItem(value : TFhirImplementationGuide); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirImplementationGuide) : Integer;
    

    {@member Insert
      Insert FhirImplementationGuide before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirImplementationGuide;
    

    {@member InsertItem
       Insert an existing FhirImplementationGuide before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirImplementationGuide);
    
    {@member Item
       Get the iIndexth FhirImplementationGuide. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirImplementationGuide. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirImplementationGuide);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirImplementationGuide;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirImplementationGuides[index : Integer] : TFhirImplementationGuide read GetItemN write SetItemN; default;
  End;


  {@Class TFhirListEntry : TFhirElement
    Entries in this list.
  }
  {!.Net HL7Connect.Fhir.ListEntry}
  TFhirListEntry = class (TFhirBackboneElement)
  private
    FFlag : TFhirCodeableConcept;
    FDeleted : TFhirBoolean;
    FDate : TFhirDateTime;
    FItem : TFhirReference{TFhirReference};
    Procedure SetFlag(value : TFhirCodeableConcept);
    Procedure SetDeleted(value : TFhirBoolean);
    Function GetDeletedST : Boolean;
    Procedure SetDeletedST(value : Boolean);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetItem(value : TFhirReference{TFhirReference});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirListEntry; overload;
    function Clone : TFhirListEntry; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member flag
      Typed access to The flag allows the system constructing the list to indicate the role and significance of the item in the list. (defined for API consistency)
    }
    property flag : TFhirCodeableConcept read FFlag write SetFlag;
    {@member flagElement
      The flag allows the system constructing the list to indicate the role and significance of the item in the list.
    }
    property flagElement : TFhirCodeableConcept read FFlag write SetFlag;

    {@member deleted
      Typed access to True if this item is marked as deleted in the list.
    }
    property deleted : Boolean read GetDeletedST write SetDeletedST;
    {@member deletedElement
      True if this item is marked as deleted in the list.
    }
    property deletedElement : TFhirBoolean read FDeleted write SetDeleted;

    {@member date
      Typed access to When this item was added to the list.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      When this item was added to the list.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member item
      Typed access to A reference to the actual resource from which data was derived. (defined for API consistency)
    }
    property item : TFhirReference{TFhirReference} read FItem write SetItem;
    {@member itemElement
      A reference to the actual resource from which data was derived.
    }
    property itemElement : TFhirReference{TFhirReference} read FItem write SetItem;

  end;


  TFhirListEntryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirListEntryList;
    function GetCurrent : TFhirListEntry;
  public
    Constructor Create(list : TFhirListEntryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirListEntry read GetCurrent;
  end;


  {@Class TFhirListEntryList
    A list of FhirListEntry
  }
  {!.Net HL7Connect.Fhir.ListEntryList}
  TFhirListEntryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirListEntry;
    procedure SetItemN(index : Integer; value : TFhirListEntry);
  public
    {!script hide}
    function Link : TFhirListEntryList; Overload;
    function Clone : TFhirListEntryList; Overload;
    function GetEnumerator : TFhirListEntryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirListEntry to the end of the list.
    }
    function Append : TFhirListEntry;

    
    {@member AddItem
      Add an already existing FhirListEntry to the end of the list.
    }
    procedure AddItem(value : TFhirListEntry); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirListEntry) : Integer;
    

    {@member Insert
      Insert FhirListEntry before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirListEntry;
    

    {@member InsertItem
       Insert an existing FhirListEntry before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirListEntry);
    
    {@member Item
       Get the iIndexth FhirListEntry. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirListEntry. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirListEntry);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirListEntry;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirListEntries[index : Integer] : TFhirListEntry read GetItemN write SetItemN; default;
  End;


  {@Class TFhirList : TFhirDomainResource
    A set of information summarized from a list of other resources.
  }
  {!.Net HL7Connect.Fhir.List}
  TFhirList = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FTitle : TFhirString;
    FCode : TFhirCodeableConcept;
    FSubject : TFhirReference{Resource};
    FSource : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FStatus : TFhirEnum;
    FDate : TFhirDateTime;
    FOrderedBy : TFhirCodeableConcept;
    FMode : TFhirEnum;
    FNote : TFhirString;
    FentryList : TFhirListEntryList;
    FEmptyReason : TFhirCodeableConcept;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetSource(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirListStatus;
    Procedure SetStatusST(value : TFhirListStatus);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetOrderedBy(value : TFhirCodeableConcept);
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirListMode;
    Procedure SetModeST(value : TFhirListMode);
    Procedure SetNote(value : TFhirString);
    Function GetNoteST : String;
    Procedure SetNoteST(value : String);
    function GetEntryList : TFhirListEntryList;
    function GetHasEntryList : Boolean;
    Procedure SetEmptyReason(value : TFhirCodeableConcept);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirList; overload;
    function Clone : TFhirList; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier for the List assigned for business purposes outside the context of FHIR.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member title
      Typed access to A label for the list assigned by the author.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      A label for the list assigned by the author.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member code
      Typed access to This code defines the purpose of the list - why it was created. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      This code defines the purpose of the list - why it was created.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member subject
      Typed access to The common subject (or patient) of the resources that are in the list, if there is one. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The common subject (or patient) of the resources that are in the list, if there is one.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member source
      Typed access to The entity responsible for deciding what the contents of the list were. Where the list was created by a human, this is the same as the author of the list. (defined for API consistency)
    }
    property source : TFhirReference{Resource} read FSource write SetSource;
    {@member sourceElement
      The entity responsible for deciding what the contents of the list were. Where the list was created by a human, this is the same as the author of the list.
    }
    property sourceElement : TFhirReference{Resource} read FSource write SetSource;

    {@member encounter
      Typed access to The encounter that is the context in which this list was created. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter that is the context in which this list was created.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member status
      Indicates the current state of this list.
    }
    property status : TFhirListStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member date
      Typed access to The date that the list was prepared.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date that the list was prepared.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member orderedBy
      Typed access to What order applies to the items in the list. (defined for API consistency)
    }
    property orderedBy : TFhirCodeableConcept read FOrderedBy write SetOrderedBy;
    {@member orderedByElement
      What order applies to the items in the list.
    }
    property orderedByElement : TFhirCodeableConcept read FOrderedBy write SetOrderedBy;

    {@member mode
      How this list was prepared - whether it is a working list that is suitable for being maintained on an ongoing basis, or if it represents a snapshot of a list of items from another source, or whether it is a prepared list where items may be marked as added, modified or deleted.
    }
    property mode : TFhirListMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member note
      Typed access to Comments that apply to the overall list.
    }
    property note : String read GetNoteST write SetNoteST;
    {@member noteElement
      Comments that apply to the overall list.
    }
    property noteElement : TFhirString read FNote write SetNote;

    {@member entryList
      Entries in this list.
    }
    property entryList : TFhirListEntryList read GetEntryList;
    property hasEntryList : boolean read GetHasEntryList;

    {@member emptyReason
      Typed access to If the list is empty, why the list is empty. (defined for API consistency)
    }
    property emptyReason : TFhirCodeableConcept read FEmptyReason write SetEmptyReason;
    {@member emptyReasonElement
      If the list is empty, why the list is empty.
    }
    property emptyReasonElement : TFhirCodeableConcept read FEmptyReason write SetEmptyReason;

  end;


  TFhirListListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirListList;
    function GetCurrent : TFhirList;
  public
    Constructor Create(list : TFhirListList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirList read GetCurrent;
  end;


  {@Class TFhirListList
    A list of FhirList
  }
  {!.Net HL7Connect.Fhir.ListList}
  TFhirListList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirList;
    procedure SetItemN(index : Integer; value : TFhirList);
  public
    {!script hide}
    function Link : TFhirListList; Overload;
    function Clone : TFhirListList; Overload;
    function GetEnumerator : TFhirListListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirList to the end of the list.
    }
    function Append : TFhirList;

    
    {@member AddItem
      Add an already existing FhirList to the end of the list.
    }
    procedure AddItem(value : TFhirList); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirList) : Integer;
    

    {@member Insert
      Insert FhirList before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirList;
    

    {@member InsertItem
       Insert an existing FhirList before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirList);
    
    {@member Item
       Get the iIndexth FhirList. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirList. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirList);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirList;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirLists[index : Integer] : TFhirList read GetItemN write SetItemN; default;
  End;


  {@Class TFhirLocationPosition : TFhirElement
    The absolute geographic location of the Location, expressed using the WGS84 datum (This is the same co-ordinate system used in KML).
  }
  {!.Net HL7Connect.Fhir.LocationPosition}
  TFhirLocationPosition = class (TFhirBackboneElement)
  private
    FLongitude : TFhirDecimal;
    FLatitude : TFhirDecimal;
    FAltitude : TFhirDecimal;
    Procedure SetLongitude(value : TFhirDecimal);
    Function GetLongitudeST : String;
    Procedure SetLongitudeST(value : String);
    Procedure SetLatitude(value : TFhirDecimal);
    Function GetLatitudeST : String;
    Procedure SetLatitudeST(value : String);
    Procedure SetAltitude(value : TFhirDecimal);
    Function GetAltitudeST : String;
    Procedure SetAltitudeST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirLocationPosition; overload;
    function Clone : TFhirLocationPosition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member longitude
      Typed access to Longitude. The value domain and the interpretation are the same as for the text of the longitude element in KML (see notes below).
    }
    property longitude : String read GetLongitudeST write SetLongitudeST;
    {@member longitudeElement
      Longitude. The value domain and the interpretation are the same as for the text of the longitude element in KML (see notes below).
    }
    property longitudeElement : TFhirDecimal read FLongitude write SetLongitude;

    {@member latitude
      Typed access to Latitude. The value domain and the interpretation are the same as for the text of the latitude element in KML (see notes below).
    }
    property latitude : String read GetLatitudeST write SetLatitudeST;
    {@member latitudeElement
      Latitude. The value domain and the interpretation are the same as for the text of the latitude element in KML (see notes below).
    }
    property latitudeElement : TFhirDecimal read FLatitude write SetLatitude;

    {@member altitude
      Typed access to Altitude. The value domain and the interpretation are the same as for the text of the altitude element in KML (see notes below).
    }
    property altitude : String read GetAltitudeST write SetAltitudeST;
    {@member altitudeElement
      Altitude. The value domain and the interpretation are the same as for the text of the altitude element in KML (see notes below).
    }
    property altitudeElement : TFhirDecimal read FAltitude write SetAltitude;

  end;


  TFhirLocationPositionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirLocationPositionList;
    function GetCurrent : TFhirLocationPosition;
  public
    Constructor Create(list : TFhirLocationPositionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirLocationPosition read GetCurrent;
  end;


  {@Class TFhirLocationPositionList
    A list of FhirLocationPosition
  }
  {!.Net HL7Connect.Fhir.LocationPositionList}
  TFhirLocationPositionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirLocationPosition;
    procedure SetItemN(index : Integer; value : TFhirLocationPosition);
  public
    {!script hide}
    function Link : TFhirLocationPositionList; Overload;
    function Clone : TFhirLocationPositionList; Overload;
    function GetEnumerator : TFhirLocationPositionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirLocationPosition to the end of the list.
    }
    function Append : TFhirLocationPosition;

    
    {@member AddItem
      Add an already existing FhirLocationPosition to the end of the list.
    }
    procedure AddItem(value : TFhirLocationPosition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirLocationPosition) : Integer;
    

    {@member Insert
      Insert FhirLocationPosition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirLocationPosition;
    

    {@member InsertItem
       Insert an existing FhirLocationPosition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirLocationPosition);
    
    {@member Item
       Get the iIndexth FhirLocationPosition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirLocationPosition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirLocationPosition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirLocationPosition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirLocationPositions[index : Integer] : TFhirLocationPosition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirLocation : TFhirDomainResource
    Details and position information for a physical place where services are provided  and resources and participants may be stored, found, contained or accommodated.
  }
  {!.Net HL7Connect.Fhir.Location}
  TFhirLocation = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FName : TFhirString;
    FDescription : TFhirString;
    FMode : TFhirEnum;
    FType_ : TFhirCodeableConcept;
    FtelecomList : TFhirContactPointList;
    FAddress : TFhirAddress;
    FPhysicalType : TFhirCodeableConcept;
    FPosition : TFhirLocationPosition;
    FManagingOrganization : TFhirReference{TFhirOrganization};
    FPartOf : TFhirReference{TFhirLocation};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirLocationStatus;
    Procedure SetStatusST(value : TFhirLocationStatus);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetMode(value : TFhirEnum);
    Function GetModeST : TFhirLocationMode;
    Procedure SetModeST(value : TFhirLocationMode);
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetAddress(value : TFhirAddress);
    Procedure SetPhysicalType(value : TFhirCodeableConcept);
    Procedure SetPosition(value : TFhirLocationPosition);
    Procedure SetManagingOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetPartOf(value : TFhirReference{TFhirLocation});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirLocation; overload;
    function Clone : TFhirLocation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Unique code or number identifying the location to its users.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      active | suspended | inactive.
    }
    property status : TFhirLocationStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member name
      Typed access to Name of the location as used by humans. Does not need to be unique.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Name of the location as used by humans. Does not need to be unique.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member description
      Typed access to Description of the Location, which helps in finding or referencing the place.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Description of the Location, which helps in finding or referencing the place.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member mode
      Indicates whether a resource instance represents a specific location or a class of locations.
    }
    property mode : TFhirLocationMode read GetModeST write SetModeST;
    property modeElement : TFhirEnum read FMode write SetMode;

    {@member type_
      Typed access to Indicates the type of function performed at the location. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Indicates the type of function performed at the location.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member telecomList
      The contact details of communication devices available at the location. This can include phone numbers, fax numbers, mobile numbers, email addresses and web sites.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member address
      Typed access to Physical location. (defined for API consistency)
    }
    property address : TFhirAddress read FAddress write SetAddress;
    {@member addressElement
      Physical location.
    }
    property addressElement : TFhirAddress read FAddress write SetAddress;

    {@member physicalType
      Typed access to Physical form of the location, e.g. building, room, vehicle, road. (defined for API consistency)
    }
    property physicalType : TFhirCodeableConcept read FPhysicalType write SetPhysicalType;
    {@member physicalTypeElement
      Physical form of the location, e.g. building, room, vehicle, road.
    }
    property physicalTypeElement : TFhirCodeableConcept read FPhysicalType write SetPhysicalType;

    {@member position
      Typed access to The absolute geographic location of the Location, expressed using the WGS84 datum (This is the same co-ordinate system used in KML). (defined for API consistency)
    }
    property position : TFhirLocationPosition read FPosition write SetPosition;
    {@member positionElement
      The absolute geographic location of the Location, expressed using the WGS84 datum (This is the same co-ordinate system used in KML).
    }
    property positionElement : TFhirLocationPosition read FPosition write SetPosition;

    {@member managingOrganization
      Typed access to The organization responsible for the provisioning and upkeep of the location. (defined for API consistency)
    }
    property managingOrganization : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;
    {@member managingOrganizationElement
      The organization responsible for the provisioning and upkeep of the location.
    }
    property managingOrganizationElement : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;

    {@member partOf
      Typed access to Another Location which this Location is physically part of. (defined for API consistency)
    }
    property partOf : TFhirReference{TFhirLocation} read FPartOf write SetPartOf;
    {@member partOfElement
      Another Location which this Location is physically part of.
    }
    property partOfElement : TFhirReference{TFhirLocation} read FPartOf write SetPartOf;

  end;


  TFhirLocationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirLocationList;
    function GetCurrent : TFhirLocation;
  public
    Constructor Create(list : TFhirLocationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirLocation read GetCurrent;
  end;


  {@Class TFhirLocationList
    A list of FhirLocation
  }
  {!.Net HL7Connect.Fhir.LocationList}
  TFhirLocationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirLocation;
    procedure SetItemN(index : Integer; value : TFhirLocation);
  public
    {!script hide}
    function Link : TFhirLocationList; Overload;
    function Clone : TFhirLocationList; Overload;
    function GetEnumerator : TFhirLocationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirLocation to the end of the list.
    }
    function Append : TFhirLocation;

    
    {@member AddItem
      Add an already existing FhirLocation to the end of the list.
    }
    procedure AddItem(value : TFhirLocation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirLocation) : Integer;
    

    {@member Insert
      Insert FhirLocation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirLocation;
    

    {@member InsertItem
       Insert an existing FhirLocation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirLocation);
    
    {@member Item
       Get the iIndexth FhirLocation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirLocation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirLocation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirLocation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirLocations[index : Integer] : TFhirLocation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedia : TFhirDomainResource
    A photo, video, or audio recording acquired or used in healthcare. The actual content may be inline or provided by direct reference.
  }
  {!.Net HL7Connect.Fhir.Media}
  TFhirMedia = class (TFhirDomainResource)
  private
    FType_ : TFhirEnum;
    FSubtype : TFhirCodeableConcept;
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FOperator : TFhirReference{TFhirPractitioner};
    FView : TFhirCodeableConcept;
    FDeviceName : TFhirString;
    FHeight : TFhirPositiveInt;
    FWidth : TFhirPositiveInt;
    FFrames : TFhirPositiveInt;
    FDuration : TFhirUnsignedInt;
    FContent : TFhirAttachment;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirDigitalMediaType;
    Procedure SetType_ST(value : TFhirDigitalMediaType);
    Procedure SetSubtype(value : TFhirCodeableConcept);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetOperator(value : TFhirReference{TFhirPractitioner});
    Procedure SetView(value : TFhirCodeableConcept);
    Procedure SetDeviceName(value : TFhirString);
    Function GetDeviceNameST : String;
    Procedure SetDeviceNameST(value : String);
    Procedure SetHeight(value : TFhirPositiveInt);
    Function GetHeightST : String;
    Procedure SetHeightST(value : String);
    Procedure SetWidth(value : TFhirPositiveInt);
    Function GetWidthST : String;
    Procedure SetWidthST(value : String);
    Procedure SetFrames(value : TFhirPositiveInt);
    Function GetFramesST : String;
    Procedure SetFramesST(value : String);
    Procedure SetDuration(value : TFhirUnsignedInt);
    Function GetDurationST : String;
    Procedure SetDurationST(value : String);
    Procedure SetContent(value : TFhirAttachment);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedia; overload;
    function Clone : TFhirMedia; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member type_
      Whether the media is a photo (still image), an audio recording, or a video recording.
    }
    property type_ : TFhirDigitalMediaType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member subtype
      Typed access to Details of the type of the media - usually, how it was acquired (what type of device). If images sourced from a DICOM system, are wrapped in a Media resource, then this is the modality. (defined for API consistency)
    }
    property subtype : TFhirCodeableConcept read FSubtype write SetSubtype;
    {@member subtypeElement
      Details of the type of the media - usually, how it was acquired (what type of device). If images sourced from a DICOM system, are wrapped in a Media resource, then this is the modality.
    }
    property subtypeElement : TFhirCodeableConcept read FSubtype write SetSubtype;

    {@member identifierList
      Identifiers associated with the image - these may include identifiers for the image itself, identifiers for the context of its collection (e.g. series ids) and context ids such as accession numbers or other workflow identifiers.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to Who/What this Media is a record of. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Who/What this Media is a record of.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member operator
      Typed access to The person who administered the collection of the image. (defined for API consistency)
    }
    property operator : TFhirReference{TFhirPractitioner} read FOperator write SetOperator;
    {@member operatorElement
      The person who administered the collection of the image.
    }
    property operatorElement : TFhirReference{TFhirPractitioner} read FOperator write SetOperator;

    {@member view
      Typed access to The name of the imaging view e.g. Lateral or Antero-posterior (AP). (defined for API consistency)
    }
    property view : TFhirCodeableConcept read FView write SetView;
    {@member viewElement
      The name of the imaging view e.g. Lateral or Antero-posterior (AP).
    }
    property viewElement : TFhirCodeableConcept read FView write SetView;

    {@member deviceName
      Typed access to The name of the device / manufacturer of the device  that was used to make the recording.
    }
    property deviceName : String read GetDeviceNameST write SetDeviceNameST;
    {@member deviceNameElement
      The name of the device / manufacturer of the device  that was used to make the recording.
    }
    property deviceNameElement : TFhirString read FDeviceName write SetDeviceName;

    {@member height
      Typed access to Height of the image in pixels (photo/video).
    }
    property height : String read GetHeightST write SetHeightST;
    {@member heightElement
      Height of the image in pixels (photo/video).
    }
    property heightElement : TFhirPositiveInt read FHeight write SetHeight;

    {@member width
      Typed access to Width of the image in pixels (photo/video).
    }
    property width : String read GetWidthST write SetWidthST;
    {@member widthElement
      Width of the image in pixels (photo/video).
    }
    property widthElement : TFhirPositiveInt read FWidth write SetWidth;

    {@member frames
      Typed access to The number of frames in a photo. This is used with a multi-page fax, or an imaging acquisition context that takes multiple slices in a single image, or an animated gif. If there is more than one frame, this SHALL have a value in order to alert interface software that a multi-frame capable rendering widget is required.
    }
    property frames : String read GetFramesST write SetFramesST;
    {@member framesElement
      The number of frames in a photo. This is used with a multi-page fax, or an imaging acquisition context that takes multiple slices in a single image, or an animated gif. If there is more than one frame, this SHALL have a value in order to alert interface software that a multi-frame capable rendering widget is required.
    }
    property framesElement : TFhirPositiveInt read FFrames write SetFrames;

    {@member duration
      Typed access to The duration of the recording in seconds - for audio and video.
    }
    property duration : String read GetDurationST write SetDurationST;
    {@member durationElement
      The duration of the recording in seconds - for audio and video.
    }
    property durationElement : TFhirUnsignedInt read FDuration write SetDuration;

    {@member content
      Typed access to The actual content of the media - inline or by direct reference to the media source file. (defined for API consistency)
    }
    property content : TFhirAttachment read FContent write SetContent;
    {@member contentElement
      The actual content of the media - inline or by direct reference to the media source file.
    }
    property contentElement : TFhirAttachment read FContent write SetContent;

  end;


  TFhirMediaListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMediaList;
    function GetCurrent : TFhirMedia;
  public
    Constructor Create(list : TFhirMediaList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedia read GetCurrent;
  end;


  {@Class TFhirMediaList
    A list of FhirMedia
  }
  {!.Net HL7Connect.Fhir.MediaList}
  TFhirMediaList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedia;
    procedure SetItemN(index : Integer; value : TFhirMedia);
  public
    {!script hide}
    function Link : TFhirMediaList; Overload;
    function Clone : TFhirMediaList; Overload;
    function GetEnumerator : TFhirMediaListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedia to the end of the list.
    }
    function Append : TFhirMedia;

    
    {@member AddItem
      Add an already existing FhirMedia to the end of the list.
    }
    procedure AddItem(value : TFhirMedia); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedia) : Integer;
    

    {@member Insert
      Insert FhirMedia before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedia;
    

    {@member InsertItem
       Insert an existing FhirMedia before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedia);
    
    {@member Item
       Get the iIndexth FhirMedia. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedia. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedia);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedia;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedia[index : Integer] : TFhirMedia read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationProduct : TFhirElement
    Information that only applies to products (not packages).
  }
  {!.Net HL7Connect.Fhir.MedicationProduct}
  TFhirMedicationProduct = class (TFhirBackboneElement)
  private
    FForm : TFhirCodeableConcept;
    FingredientList : TFhirMedicationProductIngredientList;
    FbatchList : TFhirMedicationProductBatchList;
    Procedure SetForm(value : TFhirCodeableConcept);
    function GetIngredientList : TFhirMedicationProductIngredientList;
    function GetHasIngredientList : Boolean;
    function GetBatchList : TFhirMedicationProductBatchList;
    function GetHasBatchList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationProduct; overload;
    function Clone : TFhirMedicationProduct; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member form
      Typed access to Describes the form of the item.  Powder; tablets; carton. (defined for API consistency)
    }
    property form : TFhirCodeableConcept read FForm write SetForm;
    {@member formElement
      Describes the form of the item.  Powder; tablets; carton.
    }
    property formElement : TFhirCodeableConcept read FForm write SetForm;

    {@member ingredientList
      Identifies a particular constituent of interest in the product.
    }
    property ingredientList : TFhirMedicationProductIngredientList read GetIngredientList;
    property hasIngredientList : boolean read GetHasIngredientList;

    {@member batchList
      Information about a group of medication produced or packaged from one production run.
    }
    property batchList : TFhirMedicationProductBatchList read GetBatchList;
    property hasBatchList : boolean read GetHasBatchList;

  end;


  TFhirMedicationProductListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationProductList;
    function GetCurrent : TFhirMedicationProduct;
  public
    Constructor Create(list : TFhirMedicationProductList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationProduct read GetCurrent;
  end;


  {@Class TFhirMedicationProductList
    A list of FhirMedicationProduct
  }
  {!.Net HL7Connect.Fhir.MedicationProductList}
  TFhirMedicationProductList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationProduct;
    procedure SetItemN(index : Integer; value : TFhirMedicationProduct);
  public
    {!script hide}
    function Link : TFhirMedicationProductList; Overload;
    function Clone : TFhirMedicationProductList; Overload;
    function GetEnumerator : TFhirMedicationProductListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationProduct to the end of the list.
    }
    function Append : TFhirMedicationProduct;

    
    {@member AddItem
      Add an already existing FhirMedicationProduct to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationProduct); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationProduct) : Integer;
    

    {@member Insert
      Insert FhirMedicationProduct before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationProduct;
    

    {@member InsertItem
       Insert an existing FhirMedicationProduct before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationProduct);
    
    {@member Item
       Get the iIndexth FhirMedicationProduct. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationProduct. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationProduct);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationProduct;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationProducts[index : Integer] : TFhirMedicationProduct read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationProductIngredient : TFhirElement
    Identifies a particular constituent of interest in the product.
  }
  {!.Net HL7Connect.Fhir.MedicationProductIngredient}
  TFhirMedicationProductIngredient = class (TFhirBackboneElement)
  private
    FItem : TFhirReference{Resource};
    FAmount : TFhirRatio;
    Procedure SetItem(value : TFhirReference{Resource});
    Procedure SetAmount(value : TFhirRatio);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationProductIngredient; overload;
    function Clone : TFhirMedicationProductIngredient; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member item
      Typed access to The actual ingredient - either a substance (simple ingredient) or another medication. (defined for API consistency)
    }
    property item : TFhirReference{Resource} read FItem write SetItem;
    {@member itemElement
      The actual ingredient - either a substance (simple ingredient) or another medication.
    }
    property itemElement : TFhirReference{Resource} read FItem write SetItem;

    {@member amount
      Typed access to Specifies how many (or how much) of the items there are in this Medication.  For example, 250 mg per tablet. (defined for API consistency)
    }
    property amount : TFhirRatio read FAmount write SetAmount;
    {@member amountElement
      Specifies how many (or how much) of the items there are in this Medication.  For example, 250 mg per tablet.
    }
    property amountElement : TFhirRatio read FAmount write SetAmount;

  end;


  TFhirMedicationProductIngredientListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationProductIngredientList;
    function GetCurrent : TFhirMedicationProductIngredient;
  public
    Constructor Create(list : TFhirMedicationProductIngredientList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationProductIngredient read GetCurrent;
  end;


  {@Class TFhirMedicationProductIngredientList
    A list of FhirMedicationProductIngredient
  }
  {!.Net HL7Connect.Fhir.MedicationProductIngredientList}
  TFhirMedicationProductIngredientList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationProductIngredient;
    procedure SetItemN(index : Integer; value : TFhirMedicationProductIngredient);
  public
    {!script hide}
    function Link : TFhirMedicationProductIngredientList; Overload;
    function Clone : TFhirMedicationProductIngredientList; Overload;
    function GetEnumerator : TFhirMedicationProductIngredientListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationProductIngredient to the end of the list.
    }
    function Append : TFhirMedicationProductIngredient;

    
    {@member AddItem
      Add an already existing FhirMedicationProductIngredient to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationProductIngredient); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationProductIngredient) : Integer;
    

    {@member Insert
      Insert FhirMedicationProductIngredient before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationProductIngredient;
    

    {@member InsertItem
       Insert an existing FhirMedicationProductIngredient before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationProductIngredient);
    
    {@member Item
       Get the iIndexth FhirMedicationProductIngredient. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationProductIngredient. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationProductIngredient);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationProductIngredient;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationProductIngredients[index : Integer] : TFhirMedicationProductIngredient read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationProductBatch : TFhirElement
    Information about a group of medication produced or packaged from one production run.
  }
  {!.Net HL7Connect.Fhir.MedicationProductBatch}
  TFhirMedicationProductBatch = class (TFhirBackboneElement)
  private
    FLotNumber : TFhirString;
    FExpirationDate : TFhirDateTime;
    Procedure SetLotNumber(value : TFhirString);
    Function GetLotNumberST : String;
    Procedure SetLotNumberST(value : String);
    Procedure SetExpirationDate(value : TFhirDateTime);
    Function GetExpirationDateST : TDateAndTime;
    Procedure SetExpirationDateST(value : TDateAndTime);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationProductBatch; overload;
    function Clone : TFhirMedicationProductBatch; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member lotNumber
      Typed access to The assigned lot number of a batch of the specified product.
    }
    property lotNumber : String read GetLotNumberST write SetLotNumberST;
    {@member lotNumberElement
      The assigned lot number of a batch of the specified product.
    }
    property lotNumberElement : TFhirString read FLotNumber write SetLotNumber;

    {@member expirationDate
      Typed access to When this specific batch of product will expire.
    }
    property expirationDate : TDateAndTime read GetExpirationDateST write SetExpirationDateST;
    {@member expirationDateElement
      When this specific batch of product will expire.
    }
    property expirationDateElement : TFhirDateTime read FExpirationDate write SetExpirationDate;

  end;


  TFhirMedicationProductBatchListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationProductBatchList;
    function GetCurrent : TFhirMedicationProductBatch;
  public
    Constructor Create(list : TFhirMedicationProductBatchList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationProductBatch read GetCurrent;
  end;


  {@Class TFhirMedicationProductBatchList
    A list of FhirMedicationProductBatch
  }
  {!.Net HL7Connect.Fhir.MedicationProductBatchList}
  TFhirMedicationProductBatchList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationProductBatch;
    procedure SetItemN(index : Integer; value : TFhirMedicationProductBatch);
  public
    {!script hide}
    function Link : TFhirMedicationProductBatchList; Overload;
    function Clone : TFhirMedicationProductBatchList; Overload;
    function GetEnumerator : TFhirMedicationProductBatchListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationProductBatch to the end of the list.
    }
    function Append : TFhirMedicationProductBatch;

    
    {@member AddItem
      Add an already existing FhirMedicationProductBatch to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationProductBatch); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationProductBatch) : Integer;
    

    {@member Insert
      Insert FhirMedicationProductBatch before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationProductBatch;
    

    {@member InsertItem
       Insert an existing FhirMedicationProductBatch before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationProductBatch);
    
    {@member Item
       Get the iIndexth FhirMedicationProductBatch. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationProductBatch. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationProductBatch);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationProductBatch;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationProductBatches[index : Integer] : TFhirMedicationProductBatch read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationPackage : TFhirElement
    Information that only applies to packages (not products).
  }
  {!.Net HL7Connect.Fhir.MedicationPackage}
  TFhirMedicationPackage = class (TFhirBackboneElement)
  private
    FContainer : TFhirCodeableConcept;
    FcontentList : TFhirMedicationPackageContentList;
    Procedure SetContainer(value : TFhirCodeableConcept);
    function GetContentList : TFhirMedicationPackageContentList;
    function GetHasContentList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationPackage; overload;
    function Clone : TFhirMedicationPackage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member container
      Typed access to The kind of container that this package comes as. (defined for API consistency)
    }
    property container : TFhirCodeableConcept read FContainer write SetContainer;
    {@member containerElement
      The kind of container that this package comes as.
    }
    property containerElement : TFhirCodeableConcept read FContainer write SetContainer;

    {@member contentList
      A set of components that go to make up the described item.
    }
    property contentList : TFhirMedicationPackageContentList read GetContentList;
    property hasContentList : boolean read GetHasContentList;

  end;


  TFhirMedicationPackageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationPackageList;
    function GetCurrent : TFhirMedicationPackage;
  public
    Constructor Create(list : TFhirMedicationPackageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationPackage read GetCurrent;
  end;


  {@Class TFhirMedicationPackageList
    A list of FhirMedicationPackage
  }
  {!.Net HL7Connect.Fhir.MedicationPackageList}
  TFhirMedicationPackageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationPackage;
    procedure SetItemN(index : Integer; value : TFhirMedicationPackage);
  public
    {!script hide}
    function Link : TFhirMedicationPackageList; Overload;
    function Clone : TFhirMedicationPackageList; Overload;
    function GetEnumerator : TFhirMedicationPackageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationPackage to the end of the list.
    }
    function Append : TFhirMedicationPackage;

    
    {@member AddItem
      Add an already existing FhirMedicationPackage to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationPackage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationPackage) : Integer;
    

    {@member Insert
      Insert FhirMedicationPackage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationPackage;
    

    {@member InsertItem
       Insert an existing FhirMedicationPackage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationPackage);
    
    {@member Item
       Get the iIndexth FhirMedicationPackage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationPackage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationPackage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationPackage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationPackages[index : Integer] : TFhirMedicationPackage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationPackageContent : TFhirElement
    A set of components that go to make up the described item.
  }
  {!.Net HL7Connect.Fhir.MedicationPackageContent}
  TFhirMedicationPackageContent = class (TFhirBackboneElement)
  private
    FItem : TFhirReference{TFhirMedication};
    FAmount : TFhirQuantity;
    Procedure SetItem(value : TFhirReference{TFhirMedication});
    Procedure SetAmount(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationPackageContent; overload;
    function Clone : TFhirMedicationPackageContent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member item
      Typed access to Identifies one of the items in the package. (defined for API consistency)
    }
    property item : TFhirReference{TFhirMedication} read FItem write SetItem;
    {@member itemElement
      Identifies one of the items in the package.
    }
    property itemElement : TFhirReference{TFhirMedication} read FItem write SetItem;

    {@member amount
      Typed access to The amount of the product that is in the package. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      The amount of the product that is in the package.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

  end;


  TFhirMedicationPackageContentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationPackageContentList;
    function GetCurrent : TFhirMedicationPackageContent;
  public
    Constructor Create(list : TFhirMedicationPackageContentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationPackageContent read GetCurrent;
  end;


  {@Class TFhirMedicationPackageContentList
    A list of FhirMedicationPackageContent
  }
  {!.Net HL7Connect.Fhir.MedicationPackageContentList}
  TFhirMedicationPackageContentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationPackageContent;
    procedure SetItemN(index : Integer; value : TFhirMedicationPackageContent);
  public
    {!script hide}
    function Link : TFhirMedicationPackageContentList; Overload;
    function Clone : TFhirMedicationPackageContentList; Overload;
    function GetEnumerator : TFhirMedicationPackageContentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationPackageContent to the end of the list.
    }
    function Append : TFhirMedicationPackageContent;

    
    {@member AddItem
      Add an already existing FhirMedicationPackageContent to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationPackageContent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationPackageContent) : Integer;
    

    {@member Insert
      Insert FhirMedicationPackageContent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationPackageContent;
    

    {@member InsertItem
       Insert an existing FhirMedicationPackageContent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationPackageContent);
    
    {@member Item
       Get the iIndexth FhirMedicationPackageContent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationPackageContent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationPackageContent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationPackageContent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationPackageContents[index : Integer] : TFhirMedicationPackageContent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedication : TFhirDomainResource
    This resource is primarily used for the identification and definition of a medication. It covers the ingredients and the packaging for a medication.
  }
  {!.Net HL7Connect.Fhir.Medication}
  TFhirMedication = class (TFhirDomainResource)
  private
    FCode : TFhirCodeableConcept;
    FIsBrand : TFhirBoolean;
    FManufacturer : TFhirReference{TFhirOrganization};
    FProduct : TFhirMedicationProduct;
    FPackage : TFhirMedicationPackage;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetIsBrand(value : TFhirBoolean);
    Function GetIsBrandST : Boolean;
    Procedure SetIsBrandST(value : Boolean);
    Procedure SetManufacturer(value : TFhirReference{TFhirOrganization});
    Procedure SetProduct(value : TFhirMedicationProduct);
    Procedure SetPackage(value : TFhirMedicationPackage);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedication; overload;
    function Clone : TFhirMedication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member code
      Typed access to A code (or set of codes) that specify this medication, or a textual description if no code is available. Usage note: This could be a standard medication code such as a code from RxNorm, SNOMED CT, IDMP etc. It could also be a national or local formulary code, optionally with translations to other code systems. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code (or set of codes) that specify this medication, or a textual description if no code is available. Usage note: This could be a standard medication code such as a code from RxNorm, SNOMED CT, IDMP etc. It could also be a national or local formulary code, optionally with translations to other code systems.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member isBrand
      Typed access to Set to true if the item is attributable to a specific manufacturer.
    }
    property isBrand : Boolean read GetIsBrandST write SetIsBrandST;
    {@member isBrandElement
      Set to true if the item is attributable to a specific manufacturer.
    }
    property isBrandElement : TFhirBoolean read FIsBrand write SetIsBrand;

    {@member manufacturer
      Typed access to Describes the details of the manufacturer. (defined for API consistency)
    }
    property manufacturer : TFhirReference{TFhirOrganization} read FManufacturer write SetManufacturer;
    {@member manufacturerElement
      Describes the details of the manufacturer.
    }
    property manufacturerElement : TFhirReference{TFhirOrganization} read FManufacturer write SetManufacturer;

    {@member product
      Typed access to Information that only applies to products (not packages). (defined for API consistency)
    }
    property product : TFhirMedicationProduct read FProduct write SetProduct;
    {@member productElement
      Information that only applies to products (not packages).
    }
    property productElement : TFhirMedicationProduct read FProduct write SetProduct;

    {@member package
      Typed access to Information that only applies to packages (not products). (defined for API consistency)
    }
    property package : TFhirMedicationPackage read FPackage write SetPackage;
    {@member packageElement
      Information that only applies to packages (not products).
    }
    property packageElement : TFhirMedicationPackage read FPackage write SetPackage;

  end;


  TFhirMedicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationList;
    function GetCurrent : TFhirMedication;
  public
    Constructor Create(list : TFhirMedicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedication read GetCurrent;
  end;


  {@Class TFhirMedicationList
    A list of FhirMedication
  }
  {!.Net HL7Connect.Fhir.MedicationList}
  TFhirMedicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedication;
    procedure SetItemN(index : Integer; value : TFhirMedication);
  public
    {!script hide}
    function Link : TFhirMedicationList; Overload;
    function Clone : TFhirMedicationList; Overload;
    function GetEnumerator : TFhirMedicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedication to the end of the list.
    }
    function Append : TFhirMedication;

    
    {@member AddItem
      Add an already existing FhirMedication to the end of the list.
    }
    procedure AddItem(value : TFhirMedication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedication) : Integer;
    

    {@member Insert
      Insert FhirMedication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedication;
    

    {@member InsertItem
       Insert an existing FhirMedication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedication);
    
    {@member Item
       Get the iIndexth FhirMedication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedications[index : Integer] : TFhirMedication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationAdministrationDosage : TFhirElement
    Describes the medication dosage information details e.g. dose, rate, site, route, etc.
  }
  {!.Net HL7Connect.Fhir.MedicationAdministrationDosage}
  TFhirMedicationAdministrationDosage = class (TFhirBackboneElement)
  private
    FText : TFhirString;
    FSite : TFhirType;
    FRoute : TFhirCodeableConcept;
    FMethod : TFhirCodeableConcept;
    FQuantity : TFhirQuantity;
    FRate : TFhirType;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetSite(value : TFhirType);
    Procedure SetRoute(value : TFhirCodeableConcept);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetRate(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationAdministrationDosage; overload;
    function Clone : TFhirMedicationAdministrationDosage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member text
      Typed access to Free text dosage instructions can be used for cases where the instructions are too complex to code. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Free text dosage instructions can be used for cases where the instructions are too complex to code. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication.
    }
    property textElement : TFhirString read FText write SetText;

    {@member site
      Typed access to A coded specification of the anatomic site where the medication first entered the body.  For example, "left arm". (defined for API consistency)
    }
    property site : TFhirType read FSite write SetSite;
    {@member siteElement
      A coded specification of the anatomic site where the medication first entered the body.  For example, "left arm".
    }
    property siteElement : TFhirType read FSite write SetSite;

    {@member route
      Typed access to A code specifying the route or physiological path of administration of a therapeutic agent into or onto the patient.  For example, topical, intravenous, etc. (defined for API consistency)
    }
    property route : TFhirCodeableConcept read FRoute write SetRoute;
    {@member routeElement
      A code specifying the route or physiological path of administration of a therapeutic agent into or onto the patient.  For example, topical, intravenous, etc.
    }
    property routeElement : TFhirCodeableConcept read FRoute write SetRoute;

    {@member method
      Typed access to A coded value indicating the method by which the medication is intended to be or was introduced into or on the body.  This attribute will most often NOT be populated.  It is most commonly used for injections.  For example, Slow Push, Deep IV. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      A coded value indicating the method by which the medication is intended to be or was introduced into or on the body.  This attribute will most often NOT be populated.  It is most commonly used for injections.  For example, Slow Push, Deep IV.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member quantity
      Typed access to The amount of the medication given at one administration event.   Use this value when the administration is essentially an instantaneous event such as a swallowing a tablet or giving an injection. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of the medication given at one administration event.   Use this value when the administration is essentially an instantaneous event such as a swallowing a tablet or giving an injection.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member rate
      Typed access to Identifies the speed with which the medication was or will be introduced into the patient.  Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.  Currently we do not specify a default of '1' in the denominator, but this is being discussed.  Other examples:  200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours. (defined for API consistency)
    }
    property rate : TFhirType read FRate write SetRate;
    {@member rateElement
      Identifies the speed with which the medication was or will be introduced into the patient.  Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.  Currently we do not specify a default of '1' in the denominator, but this is being discussed.  Other examples:  200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours.
    }
    property rateElement : TFhirType read FRate write SetRate;

  end;


  TFhirMedicationAdministrationDosageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationAdministrationDosageList;
    function GetCurrent : TFhirMedicationAdministrationDosage;
  public
    Constructor Create(list : TFhirMedicationAdministrationDosageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationAdministrationDosage read GetCurrent;
  end;


  {@Class TFhirMedicationAdministrationDosageList
    A list of FhirMedicationAdministrationDosage
  }
  {!.Net HL7Connect.Fhir.MedicationAdministrationDosageList}
  TFhirMedicationAdministrationDosageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationAdministrationDosage;
    procedure SetItemN(index : Integer; value : TFhirMedicationAdministrationDosage);
  public
    {!script hide}
    function Link : TFhirMedicationAdministrationDosageList; Overload;
    function Clone : TFhirMedicationAdministrationDosageList; Overload;
    function GetEnumerator : TFhirMedicationAdministrationDosageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationAdministrationDosage to the end of the list.
    }
    function Append : TFhirMedicationAdministrationDosage;

    
    {@member AddItem
      Add an already existing FhirMedicationAdministrationDosage to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationAdministrationDosage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationAdministrationDosage) : Integer;
    

    {@member Insert
      Insert FhirMedicationAdministrationDosage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationAdministrationDosage;
    

    {@member InsertItem
       Insert an existing FhirMedicationAdministrationDosage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationAdministrationDosage);
    
    {@member Item
       Get the iIndexth FhirMedicationAdministrationDosage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationAdministrationDosage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationAdministrationDosage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationAdministrationDosage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationAdministrationDosages[index : Integer] : TFhirMedicationAdministrationDosage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationAdministration : TFhirDomainResource
    Describes the event of a patient consuming or otherwise being administered a medication.  This may be as simple as swallowing a tablet or it may be a long running infusion.  Related resources tie this event to the authorizing prescription, and the specific encounter between patient and health care practitioner.
  }
  {!.Net HL7Connect.Fhir.MedicationAdministration}
  TFhirMedicationAdministration = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FPatient : TFhirReference{TFhirPatient};
    FPractitioner : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FPrescription : TFhirReference{TFhirMedicationOrder};
    FWasNotGiven : TFhirBoolean;
    FreasonNotGivenList : TFhirCodeableConceptList;
    FreasonGivenList : TFhirCodeableConceptList;
    FEffectiveTime : TFhirType;
    FMedication : TFhirType;
    FdeviceList : TFhirReferenceList{TFhirDevice};
    FNote : TFhirString;
    FDosage : TFhirMedicationAdministrationDosage;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirMedicationAdminStatus;
    Procedure SetStatusST(value : TFhirMedicationAdminStatus);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetPractitioner(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetPrescription(value : TFhirReference{TFhirMedicationOrder});
    Procedure SetWasNotGiven(value : TFhirBoolean);
    Function GetWasNotGivenST : Boolean;
    Procedure SetWasNotGivenST(value : Boolean);
    function GetReasonNotGivenList : TFhirCodeableConceptList;
    function GetHasReasonNotGivenList : Boolean;
    function GetReasonGivenList : TFhirCodeableConceptList;
    function GetHasReasonGivenList : Boolean;
    Procedure SetEffectiveTime(value : TFhirType);
    Procedure SetMedication(value : TFhirType);
    function GetDeviceList : TFhirReferenceList{TFhirDevice};
    function GetHasDeviceList : Boolean;
    Procedure SetNote(value : TFhirString);
    Function GetNoteST : String;
    Procedure SetNoteST(value : String);
    Procedure SetDosage(value : TFhirMedicationAdministrationDosage);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationAdministration; overload;
    function Clone : TFhirMedicationAdministration; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External identifier - FHIR will generate its own internal identifiers (probably URLs) which do not need to be explicitly managed by the resource.  The identifier here is one that would be used by another non-FHIR system - for example an automated medication pump would provide a record each time it operated; an administration while the patient was off the ward might be made with a different system and entered after the event.  Particularly important if these records have to be updated.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      Will generally be set to show that the administration has been completed.  For some long running administrations such as infusions it is possible for an administration to be started but not completed or it may be paused while some other process is under way.
    }
    property status : TFhirMedicationAdminStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member patient
      Typed access to The person or animal receiving the medication. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The person or animal receiving the medication.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member practitioner
      Typed access to The individual who was responsible for giving the medication to the patient. (defined for API consistency)
    }
    property practitioner : TFhirReference{Resource} read FPractitioner write SetPractitioner;
    {@member practitionerElement
      The individual who was responsible for giving the medication to the patient.
    }
    property practitionerElement : TFhirReference{Resource} read FPractitioner write SetPractitioner;

    {@member encounter
      Typed access to The visit, admission or other contact between patient and health care provider the medication administration was performed as part of. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The visit, admission or other contact between patient and health care provider the medication administration was performed as part of.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member prescription
      Typed access to The original request, instruction or authority to perform the administration. (defined for API consistency)
    }
    property prescription : TFhirReference{TFhirMedicationOrder} read FPrescription write SetPrescription;
    {@member prescriptionElement
      The original request, instruction or authority to perform the administration.
    }
    property prescriptionElement : TFhirReference{TFhirMedicationOrder} read FPrescription write SetPrescription;

    {@member wasNotGiven
      Typed access to Set this to true if the record is saying that the medication was NOT administered.
    }
    property wasNotGiven : Boolean read GetWasNotGivenST write SetWasNotGivenST;
    {@member wasNotGivenElement
      Set this to true if the record is saying that the medication was NOT administered.
    }
    property wasNotGivenElement : TFhirBoolean read FWasNotGiven write SetWasNotGiven;

    {@member reasonNotGivenList
      A code indicating why the administration was not performed.
    }
    property reasonNotGivenList : TFhirCodeableConceptList read GetReasonNotGivenList;
    property hasReasonNotGivenList : boolean read GetHasReasonNotGivenList;

    {@member reasonGivenList
      A code indicating why the medication was given.
    }
    property reasonGivenList : TFhirCodeableConceptList read GetReasonGivenList;
    property hasReasonGivenList : boolean read GetHasReasonGivenList;

    {@member effectiveTime
      Typed access to A specific date/time or interval of time during which the administration took place (or did not take place, when the 'notGiven' attribute is true). For many administrations, such as swallowing a tablet the use of dateTime is more appropriate. (defined for API consistency)
    }
    property effectiveTime : TFhirType read FEffectiveTime write SetEffectiveTime;
    {@member effectiveTimeElement
      A specific date/time or interval of time during which the administration took place (or did not take place, when the 'notGiven' attribute is true). For many administrations, such as swallowing a tablet the use of dateTime is more appropriate.
    }
    property effectiveTimeElement : TFhirType read FEffectiveTime write SetEffectiveTime;

    {@member medication
      Typed access to Identifies the medication that was administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications. (defined for API consistency)
    }
    property medication : TFhirType read FMedication write SetMedication;
    {@member medicationElement
      Identifies the medication that was administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications.
    }
    property medicationElement : TFhirType read FMedication write SetMedication;

    {@member deviceList
      The device used in administering the medication to the patient.  For example, a particular infusion pump.
    }
    property deviceList : TFhirReferenceList{TFhirDevice} read GetDeviceList;
    property hasDeviceList : boolean read GetHasDeviceList;

    {@member note
      Typed access to Extra information about the medication administration that is not conveyed by the other attributes.
    }
    property note : String read GetNoteST write SetNoteST;
    {@member noteElement
      Extra information about the medication administration that is not conveyed by the other attributes.
    }
    property noteElement : TFhirString read FNote write SetNote;

    {@member dosage
      Typed access to Describes the medication dosage information details e.g. dose, rate, site, route, etc. (defined for API consistency)
    }
    property dosage : TFhirMedicationAdministrationDosage read FDosage write SetDosage;
    {@member dosageElement
      Describes the medication dosage information details e.g. dose, rate, site, route, etc.
    }
    property dosageElement : TFhirMedicationAdministrationDosage read FDosage write SetDosage;

  end;


  TFhirMedicationAdministrationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationAdministrationList;
    function GetCurrent : TFhirMedicationAdministration;
  public
    Constructor Create(list : TFhirMedicationAdministrationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationAdministration read GetCurrent;
  end;


  {@Class TFhirMedicationAdministrationList
    A list of FhirMedicationAdministration
  }
  {!.Net HL7Connect.Fhir.MedicationAdministrationList}
  TFhirMedicationAdministrationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationAdministration;
    procedure SetItemN(index : Integer; value : TFhirMedicationAdministration);
  public
    {!script hide}
    function Link : TFhirMedicationAdministrationList; Overload;
    function Clone : TFhirMedicationAdministrationList; Overload;
    function GetEnumerator : TFhirMedicationAdministrationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationAdministration to the end of the list.
    }
    function Append : TFhirMedicationAdministration;

    
    {@member AddItem
      Add an already existing FhirMedicationAdministration to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationAdministration); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationAdministration) : Integer;
    

    {@member Insert
      Insert FhirMedicationAdministration before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationAdministration;
    

    {@member InsertItem
       Insert an existing FhirMedicationAdministration before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationAdministration);
    
    {@member Item
       Get the iIndexth FhirMedicationAdministration. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationAdministration. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationAdministration);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationAdministration;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationAdministrations[index : Integer] : TFhirMedicationAdministration read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationDispenseDosageInstruction : TFhirElement
    Indicates how the medication is to be used by the patient.
  }
  {!.Net HL7Connect.Fhir.MedicationDispenseDosageInstruction}
  TFhirMedicationDispenseDosageInstruction = class (TFhirBackboneElement)
  private
    FText : TFhirString;
    FAdditionalInstructions : TFhirCodeableConcept;
    FTiming : TFhirTiming;
    FAsNeeded : TFhirType;
    FSite : TFhirType;
    FRoute : TFhirCodeableConcept;
    FMethod : TFhirCodeableConcept;
    FDose : TFhirType;
    FRate : TFhirType;
    FMaxDosePerPeriod : TFhirRatio;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetAdditionalInstructions(value : TFhirCodeableConcept);
    Procedure SetTiming(value : TFhirTiming);
    Procedure SetAsNeeded(value : TFhirType);
    Procedure SetSite(value : TFhirType);
    Procedure SetRoute(value : TFhirCodeableConcept);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetDose(value : TFhirType);
    Procedure SetRate(value : TFhirType);
    Procedure SetMaxDosePerPeriod(value : TFhirRatio);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationDispenseDosageInstruction; overload;
    function Clone : TFhirMedicationDispenseDosageInstruction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member text
      Typed access to Free text dosage instructions can be used for cases where the instructions are too complex to code. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Free text dosage instructions can be used for cases where the instructions are too complex to code. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication.
    }
    property textElement : TFhirString read FText write SetText;

    {@member additionalInstructions
      Typed access to Additional instructions such as "Swallow with plenty of water" which may or may not be coded. (defined for API consistency)
    }
    property additionalInstructions : TFhirCodeableConcept read FAdditionalInstructions write SetAdditionalInstructions;
    {@member additionalInstructionsElement
      Additional instructions such as "Swallow with plenty of water" which may or may not be coded.
    }
    property additionalInstructionsElement : TFhirCodeableConcept read FAdditionalInstructions write SetAdditionalInstructions;

    {@member timing
      Typed access to The timing schedule for giving the medication to the patient.  The Schedule data type allows many different expressions.  For example, "Every  8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:";  "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013". (defined for API consistency)
    }
    property timing : TFhirTiming read FTiming write SetTiming;
    {@member timingElement
      The timing schedule for giving the medication to the patient.  The Schedule data type allows many different expressions.  For example, "Every  8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:";  "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013".
    }
    property timingElement : TFhirTiming read FTiming write SetTiming;

    {@member asNeeded
      Typed access to Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept).  

Specifically if 'boolean' datatype is selected, then the following logic applies:  If set to True, this indicates that the medication is only taken when needed, within the specified schedule. (defined for API consistency)
    }
    property asNeeded : TFhirType read FAsNeeded write SetAsNeeded;
    {@member asNeededElement
      Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept).  

Specifically if 'boolean' datatype is selected, then the following logic applies:  If set to True, this indicates that the medication is only taken when needed, within the specified schedule.
    }
    property asNeededElement : TFhirType read FAsNeeded write SetAsNeeded;

    {@member site
      Typed access to A coded specification of the anatomic site where the medication first enters the body. (defined for API consistency)
    }
    property site : TFhirType read FSite write SetSite;
    {@member siteElement
      A coded specification of the anatomic site where the medication first enters the body.
    }
    property siteElement : TFhirType read FSite write SetSite;

    {@member route
      Typed access to A code specifying the route or physiological path of administration of a therapeutic agent into or onto a subject. (defined for API consistency)
    }
    property route : TFhirCodeableConcept read FRoute write SetRoute;
    {@member routeElement
      A code specifying the route or physiological path of administration of a therapeutic agent into or onto a subject.
    }
    property routeElement : TFhirCodeableConcept read FRoute write SetRoute;

    {@member method
      Typed access to A coded value indicating the method by which the medication is intended to be or was introduced into or on the body. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      A coded value indicating the method by which the medication is intended to be or was introduced into or on the body.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member dose
      Typed access to The amount of therapeutic or other substance given at one administration event. (defined for API consistency)
    }
    property dose : TFhirType read FDose write SetDose;
    {@member doseElement
      The amount of therapeutic or other substance given at one administration event.
    }
    property doseElement : TFhirType read FDose write SetDose;

    {@member rate
      Typed access to Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours. (defined for API consistency)
    }
    property rate : TFhirType read FRate write SetRate;
    {@member rateElement
      Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours.
    }
    property rateElement : TFhirType read FRate write SetRate;

    {@member maxDosePerPeriod
      Typed access to The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time,  e.g. 1000mg in 24 hours. (defined for API consistency)
    }
    property maxDosePerPeriod : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;
    {@member maxDosePerPeriodElement
      The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time,  e.g. 1000mg in 24 hours.
    }
    property maxDosePerPeriodElement : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;

  end;


  TFhirMedicationDispenseDosageInstructionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationDispenseDosageInstructionList;
    function GetCurrent : TFhirMedicationDispenseDosageInstruction;
  public
    Constructor Create(list : TFhirMedicationDispenseDosageInstructionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationDispenseDosageInstruction read GetCurrent;
  end;


  {@Class TFhirMedicationDispenseDosageInstructionList
    A list of FhirMedicationDispenseDosageInstruction
  }
  {!.Net HL7Connect.Fhir.MedicationDispenseDosageInstructionList}
  TFhirMedicationDispenseDosageInstructionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationDispenseDosageInstruction;
    procedure SetItemN(index : Integer; value : TFhirMedicationDispenseDosageInstruction);
  public
    {!script hide}
    function Link : TFhirMedicationDispenseDosageInstructionList; Overload;
    function Clone : TFhirMedicationDispenseDosageInstructionList; Overload;
    function GetEnumerator : TFhirMedicationDispenseDosageInstructionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationDispenseDosageInstruction to the end of the list.
    }
    function Append : TFhirMedicationDispenseDosageInstruction;

    
    {@member AddItem
      Add an already existing FhirMedicationDispenseDosageInstruction to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationDispenseDosageInstruction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationDispenseDosageInstruction) : Integer;
    

    {@member Insert
      Insert FhirMedicationDispenseDosageInstruction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationDispenseDosageInstruction;
    

    {@member InsertItem
       Insert an existing FhirMedicationDispenseDosageInstruction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationDispenseDosageInstruction);
    
    {@member Item
       Get the iIndexth FhirMedicationDispenseDosageInstruction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationDispenseDosageInstruction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationDispenseDosageInstruction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationDispenseDosageInstruction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationDispenseDosageInstructions[index : Integer] : TFhirMedicationDispenseDosageInstruction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationDispenseSubstitution : TFhirElement
    Indicates whether or not substitution was made as part of the dispense.  In some cases substitution will be expected but does not happen, in other cases substitution is not expected but does happen.  This block explains what substitution did or did not happen and why.
  }
  {!.Net HL7Connect.Fhir.MedicationDispenseSubstitution}
  TFhirMedicationDispenseSubstitution = class (TFhirBackboneElement)
  private
    FType_ : TFhirCodeableConcept;
    FreasonList : TFhirCodeableConceptList;
    FresponsiblePartyList : TFhirReferenceList{TFhirPractitioner};
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    function GetResponsiblePartyList : TFhirReferenceList{TFhirPractitioner};
    function GetHasResponsiblePartyList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationDispenseSubstitution; overload;
    function Clone : TFhirMedicationDispenseSubstitution; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to A code signifying whether a different drug was dispensed from what was prescribed. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      A code signifying whether a different drug was dispensed from what was prescribed.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member reasonList
      Indicates the reason for the substitution of (or lack of substitution) from what was prescribed.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member responsiblePartyList
      The person or organization that has primary responsibility for the substitution.
    }
    property responsiblePartyList : TFhirReferenceList{TFhirPractitioner} read GetResponsiblePartyList;
    property hasResponsiblePartyList : boolean read GetHasResponsiblePartyList;

  end;


  TFhirMedicationDispenseSubstitutionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationDispenseSubstitutionList;
    function GetCurrent : TFhirMedicationDispenseSubstitution;
  public
    Constructor Create(list : TFhirMedicationDispenseSubstitutionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationDispenseSubstitution read GetCurrent;
  end;


  {@Class TFhirMedicationDispenseSubstitutionList
    A list of FhirMedicationDispenseSubstitution
  }
  {!.Net HL7Connect.Fhir.MedicationDispenseSubstitutionList}
  TFhirMedicationDispenseSubstitutionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationDispenseSubstitution;
    procedure SetItemN(index : Integer; value : TFhirMedicationDispenseSubstitution);
  public
    {!script hide}
    function Link : TFhirMedicationDispenseSubstitutionList; Overload;
    function Clone : TFhirMedicationDispenseSubstitutionList; Overload;
    function GetEnumerator : TFhirMedicationDispenseSubstitutionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationDispenseSubstitution to the end of the list.
    }
    function Append : TFhirMedicationDispenseSubstitution;

    
    {@member AddItem
      Add an already existing FhirMedicationDispenseSubstitution to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationDispenseSubstitution); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationDispenseSubstitution) : Integer;
    

    {@member Insert
      Insert FhirMedicationDispenseSubstitution before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationDispenseSubstitution;
    

    {@member InsertItem
       Insert an existing FhirMedicationDispenseSubstitution before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationDispenseSubstitution);
    
    {@member Item
       Get the iIndexth FhirMedicationDispenseSubstitution. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationDispenseSubstitution. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationDispenseSubstitution);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationDispenseSubstitution;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationDispenseSubstitutions[index : Integer] : TFhirMedicationDispenseSubstitution read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationDispense : TFhirDomainResource
    Indicates that a medication product is to be or has been dispensed for a named person/patient.  This includes a description of the medication product (supply) provided and the instructions for administering the medication.  The medication dispense is the result of a pharmacy system responding to a medication order.
  }
  {!.Net HL7Connect.Fhir.MedicationDispense}
  TFhirMedicationDispense = class (TFhirDomainResource)
  private
    FIdentifier : TFhirIdentifier;
    FStatus : TFhirEnum;
    FPatient : TFhirReference{TFhirPatient};
    FDispenser : TFhirReference{TFhirPractitioner};
    FauthorizingPrescriptionList : TFhirReferenceList{TFhirMedicationOrder};
    FType_ : TFhirCodeableConcept;
    FQuantity : TFhirQuantity;
    FDaysSupply : TFhirQuantity;
    FMedication : TFhirType;
    FWhenPrepared : TFhirDateTime;
    FWhenHandedOver : TFhirDateTime;
    FDestination : TFhirReference{TFhirLocation};
    FreceiverList : TFhirReferenceList{Resource};
    FNote : TFhirString;
    FdosageInstructionList : TFhirMedicationDispenseDosageInstructionList;
    FSubstitution : TFhirMedicationDispenseSubstitution;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirMedicationDispenseStatus;
    Procedure SetStatusST(value : TFhirMedicationDispenseStatus);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetDispenser(value : TFhirReference{TFhirPractitioner});
    function GetAuthorizingPrescriptionList : TFhirReferenceList{TFhirMedicationOrder};
    function GetHasAuthorizingPrescriptionList : Boolean;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetDaysSupply(value : TFhirQuantity);
    Procedure SetMedication(value : TFhirType);
    Procedure SetWhenPrepared(value : TFhirDateTime);
    Function GetWhenPreparedST : TDateAndTime;
    Procedure SetWhenPreparedST(value : TDateAndTime);
    Procedure SetWhenHandedOver(value : TFhirDateTime);
    Function GetWhenHandedOverST : TDateAndTime;
    Procedure SetWhenHandedOverST(value : TDateAndTime);
    Procedure SetDestination(value : TFhirReference{TFhirLocation});
    function GetReceiverList : TFhirReferenceList{Resource};
    function GetHasReceiverList : Boolean;
    Procedure SetNote(value : TFhirString);
    Function GetNoteST : String;
    Procedure SetNoteST(value : String);
    function GetDosageInstructionList : TFhirMedicationDispenseDosageInstructionList;
    function GetHasDosageInstructionList : Boolean;
    Procedure SetSubstitution(value : TFhirMedicationDispenseSubstitution);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationDispense; overload;
    function Clone : TFhirMedicationDispense; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifier
      Typed access to Identifier assigned by the dispensing facility - this is an identifier assigned outside FHIR. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifier assigned by the dispensing facility - this is an identifier assigned outside FHIR.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member status
      A code specifying the state of the set of dispense events.
    }
    property status : TFhirMedicationDispenseStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member patient
      Typed access to A link to a resource representing the person to whom the medication will be given. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A link to a resource representing the person to whom the medication will be given.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member dispenser
      Typed access to The individual responsible for dispensing the medication. (defined for API consistency)
    }
    property dispenser : TFhirReference{TFhirPractitioner} read FDispenser write SetDispenser;
    {@member dispenserElement
      The individual responsible for dispensing the medication.
    }
    property dispenserElement : TFhirReference{TFhirPractitioner} read FDispenser write SetDispenser;

    {@member authorizingPrescriptionList
      Indicates the medication order that is being dispensed against.
    }
    property authorizingPrescriptionList : TFhirReferenceList{TFhirMedicationOrder} read GetAuthorizingPrescriptionList;
    property hasAuthorizingPrescriptionList : boolean read GetHasAuthorizingPrescriptionList;

    {@member type_
      Typed access to Indicates the type of dispensing event that is performed. For example, Trial Fill, Completion of Trial, Partial Fill, Emergency Fill, Samples, etc. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Indicates the type of dispensing event that is performed. For example, Trial Fill, Completion of Trial, Partial Fill, Emergency Fill, Samples, etc.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member quantity
      Typed access to The amount of medication that has been dispensed. Includes unit of measure. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of medication that has been dispensed. Includes unit of measure.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member daysSupply
      Typed access to The amount of medication expressed as a timing amount. (defined for API consistency)
    }
    property daysSupply : TFhirQuantity read FDaysSupply write SetDaysSupply;
    {@member daysSupplyElement
      The amount of medication expressed as a timing amount.
    }
    property daysSupplyElement : TFhirQuantity read FDaysSupply write SetDaysSupply;

    {@member medication
      Typed access to Identifies the medication being administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications. (defined for API consistency)
    }
    property medication : TFhirType read FMedication write SetMedication;
    {@member medicationElement
      Identifies the medication being administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications.
    }
    property medicationElement : TFhirType read FMedication write SetMedication;

    {@member whenPrepared
      Typed access to The time when the dispensed product was packaged and reviewed.
    }
    property whenPrepared : TDateAndTime read GetWhenPreparedST write SetWhenPreparedST;
    {@member whenPreparedElement
      The time when the dispensed product was packaged and reviewed.
    }
    property whenPreparedElement : TFhirDateTime read FWhenPrepared write SetWhenPrepared;

    {@member whenHandedOver
      Typed access to The time the dispensed product was provided to the patient or their representative.
    }
    property whenHandedOver : TDateAndTime read GetWhenHandedOverST write SetWhenHandedOverST;
    {@member whenHandedOverElement
      The time the dispensed product was provided to the patient or their representative.
    }
    property whenHandedOverElement : TFhirDateTime read FWhenHandedOver write SetWhenHandedOver;

    {@member destination
      Typed access to Identification of the facility/location where the medication was shipped to, as part of the dispense event. (defined for API consistency)
    }
    property destination : TFhirReference{TFhirLocation} read FDestination write SetDestination;
    {@member destinationElement
      Identification of the facility/location where the medication was shipped to, as part of the dispense event.
    }
    property destinationElement : TFhirReference{TFhirLocation} read FDestination write SetDestination;

    {@member receiverList
      Identifies the person who picked up the medication.  This will usually be a patient or their caregiver, but some cases exist where it can be a healthcare professional.
    }
    property receiverList : TFhirReferenceList{Resource} read GetReceiverList;
    property hasReceiverList : boolean read GetHasReceiverList;

    {@member note
      Typed access to Extra information about the dispense that could not be conveyed in the other attributes.
    }
    property note : String read GetNoteST write SetNoteST;
    {@member noteElement
      Extra information about the dispense that could not be conveyed in the other attributes.
    }
    property noteElement : TFhirString read FNote write SetNote;

    {@member dosageInstructionList
      Indicates how the medication is to be used by the patient.
    }
    property dosageInstructionList : TFhirMedicationDispenseDosageInstructionList read GetDosageInstructionList;
    property hasDosageInstructionList : boolean read GetHasDosageInstructionList;

    {@member substitution
      Typed access to Indicates whether or not substitution was made as part of the dispense.  In some cases substitution will be expected but does not happen, in other cases substitution is not expected but does happen.  This block explains what substitution did or did not happen and why. (defined for API consistency)
    }
    property substitution : TFhirMedicationDispenseSubstitution read FSubstitution write SetSubstitution;
    {@member substitutionElement
      Indicates whether or not substitution was made as part of the dispense.  In some cases substitution will be expected but does not happen, in other cases substitution is not expected but does happen.  This block explains what substitution did or did not happen and why.
    }
    property substitutionElement : TFhirMedicationDispenseSubstitution read FSubstitution write SetSubstitution;

  end;


  TFhirMedicationDispenseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationDispenseList;
    function GetCurrent : TFhirMedicationDispense;
  public
    Constructor Create(list : TFhirMedicationDispenseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationDispense read GetCurrent;
  end;


  {@Class TFhirMedicationDispenseList
    A list of FhirMedicationDispense
  }
  {!.Net HL7Connect.Fhir.MedicationDispenseList}
  TFhirMedicationDispenseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationDispense;
    procedure SetItemN(index : Integer; value : TFhirMedicationDispense);
  public
    {!script hide}
    function Link : TFhirMedicationDispenseList; Overload;
    function Clone : TFhirMedicationDispenseList; Overload;
    function GetEnumerator : TFhirMedicationDispenseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationDispense to the end of the list.
    }
    function Append : TFhirMedicationDispense;

    
    {@member AddItem
      Add an already existing FhirMedicationDispense to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationDispense); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationDispense) : Integer;
    

    {@member Insert
      Insert FhirMedicationDispense before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationDispense;
    

    {@member InsertItem
       Insert an existing FhirMedicationDispense before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationDispense);
    
    {@member Item
       Get the iIndexth FhirMedicationDispense. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationDispense. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationDispense);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationDispense;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationDispenses[index : Integer] : TFhirMedicationDispense read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationOrderDosageInstruction : TFhirElement
    Indicates how the medication is to be used by the patient.
  }
  {!.Net HL7Connect.Fhir.MedicationOrderDosageInstruction}
  TFhirMedicationOrderDosageInstruction = class (TFhirBackboneElement)
  private
    FText : TFhirString;
    FAdditionalInstructions : TFhirCodeableConcept;
    FTiming : TFhirTiming;
    FAsNeeded : TFhirType;
    FSite : TFhirType;
    FRoute : TFhirCodeableConcept;
    FMethod : TFhirCodeableConcept;
    FDose : TFhirType;
    FRate : TFhirType;
    FMaxDosePerPeriod : TFhirRatio;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetAdditionalInstructions(value : TFhirCodeableConcept);
    Procedure SetTiming(value : TFhirTiming);
    Procedure SetAsNeeded(value : TFhirType);
    Procedure SetSite(value : TFhirType);
    Procedure SetRoute(value : TFhirCodeableConcept);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetDose(value : TFhirType);
    Procedure SetRate(value : TFhirType);
    Procedure SetMaxDosePerPeriod(value : TFhirRatio);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationOrderDosageInstruction; overload;
    function Clone : TFhirMedicationOrderDosageInstruction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member text
      Typed access to Free text dosage instructions can be used for cases where the instructions are too complex to code.  The content of this attribute does not include the name or description of the medication. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication. It is expected that the text instructions will always be populated.  If the dosage.timing attribute is also populated, then the dosage.text should reflect the same information as the timing.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Free text dosage instructions can be used for cases where the instructions are too complex to code.  The content of this attribute does not include the name or description of the medication. When coded instructions are present, the free text instructions may still be present for display to humans taking or administering the medication. It is expected that the text instructions will always be populated.  If the dosage.timing attribute is also populated, then the dosage.text should reflect the same information as the timing.
    }
    property textElement : TFhirString read FText write SetText;

    {@member additionalInstructions
      Typed access to Additional instructions such as "Swallow with plenty of water" which may or may not be coded. (defined for API consistency)
    }
    property additionalInstructions : TFhirCodeableConcept read FAdditionalInstructions write SetAdditionalInstructions;
    {@member additionalInstructionsElement
      Additional instructions such as "Swallow with plenty of water" which may or may not be coded.
    }
    property additionalInstructionsElement : TFhirCodeableConcept read FAdditionalInstructions write SetAdditionalInstructions;

    {@member timing
      Typed access to The timing schedule for giving the medication to the patient. The Schedule data type allows many different expressions. For example: "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013". (defined for API consistency)
    }
    property timing : TFhirTiming read FTiming write SetTiming;
    {@member timingElement
      The timing schedule for giving the medication to the patient. The Schedule data type allows many different expressions. For example: "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013".
    }
    property timingElement : TFhirTiming read FTiming write SetTiming;

    {@member asNeeded
      Typed access to Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept). (defined for API consistency)
    }
    property asNeeded : TFhirType read FAsNeeded write SetAsNeeded;
    {@member asNeededElement
      Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept).
    }
    property asNeededElement : TFhirType read FAsNeeded write SetAsNeeded;

    {@member site
      Typed access to A coded specification of the anatomic site where the medication first enters the body. (defined for API consistency)
    }
    property site : TFhirType read FSite write SetSite;
    {@member siteElement
      A coded specification of the anatomic site where the medication first enters the body.
    }
    property siteElement : TFhirType read FSite write SetSite;

    {@member route
      Typed access to A code specifying the route or physiological path of administration of a therapeutic agent into or onto a patient's body. (defined for API consistency)
    }
    property route : TFhirCodeableConcept read FRoute write SetRoute;
    {@member routeElement
      A code specifying the route or physiological path of administration of a therapeutic agent into or onto a patient's body.
    }
    property routeElement : TFhirCodeableConcept read FRoute write SetRoute;

    {@member method
      Typed access to A coded value indicating the method by which the medication is introduced into or onto the body. Most commonly used for injections.  For examples, Slow Push; Deep IV. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      A coded value indicating the method by which the medication is introduced into or onto the body. Most commonly used for injections.  For examples, Slow Push; Deep IV.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member dose
      Typed access to The amount of therapeutic or other substance given at one administration event. (defined for API consistency)
    }
    property dose : TFhirType read FDose write SetDose;
    {@member doseElement
      The amount of therapeutic or other substance given at one administration event.
    }
    property doseElement : TFhirType read FDose write SetDose;

    {@member rate
      Typed access to Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours. (defined for API consistency)
    }
    property rate : TFhirType read FRate write SetRate;
    {@member rateElement
      Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours.
    }
    property rateElement : TFhirType read FRate write SetRate;

    {@member maxDosePerPeriod
      Typed access to The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time.  For example, 1000mg in 24 hours. (defined for API consistency)
    }
    property maxDosePerPeriod : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;
    {@member maxDosePerPeriodElement
      The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time.  For example, 1000mg in 24 hours.
    }
    property maxDosePerPeriodElement : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;

  end;


  TFhirMedicationOrderDosageInstructionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationOrderDosageInstructionList;
    function GetCurrent : TFhirMedicationOrderDosageInstruction;
  public
    Constructor Create(list : TFhirMedicationOrderDosageInstructionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationOrderDosageInstruction read GetCurrent;
  end;


  {@Class TFhirMedicationOrderDosageInstructionList
    A list of FhirMedicationOrderDosageInstruction
  }
  {!.Net HL7Connect.Fhir.MedicationOrderDosageInstructionList}
  TFhirMedicationOrderDosageInstructionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationOrderDosageInstruction;
    procedure SetItemN(index : Integer; value : TFhirMedicationOrderDosageInstruction);
  public
    {!script hide}
    function Link : TFhirMedicationOrderDosageInstructionList; Overload;
    function Clone : TFhirMedicationOrderDosageInstructionList; Overload;
    function GetEnumerator : TFhirMedicationOrderDosageInstructionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationOrderDosageInstruction to the end of the list.
    }
    function Append : TFhirMedicationOrderDosageInstruction;

    
    {@member AddItem
      Add an already existing FhirMedicationOrderDosageInstruction to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationOrderDosageInstruction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationOrderDosageInstruction) : Integer;
    

    {@member Insert
      Insert FhirMedicationOrderDosageInstruction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationOrderDosageInstruction;
    

    {@member InsertItem
       Insert an existing FhirMedicationOrderDosageInstruction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationOrderDosageInstruction);
    
    {@member Item
       Get the iIndexth FhirMedicationOrderDosageInstruction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationOrderDosageInstruction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationOrderDosageInstruction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationOrderDosageInstruction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationOrderDosageInstructions[index : Integer] : TFhirMedicationOrderDosageInstruction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationOrderDispenseRequest : TFhirElement
    Indicates the specific details for the dispense or medication supply part of a medication order (also known as a Medication Prescription).  Note that this information is NOT always sent with the order.  There may be in some settings (e.g. hospitals) institutional or system support for completing the dispense details in the pharmacy department.
  }
  {!.Net HL7Connect.Fhir.MedicationOrderDispenseRequest}
  TFhirMedicationOrderDispenseRequest = class (TFhirBackboneElement)
  private
    FMedication : TFhirType;
    FValidityPeriod : TFhirPeriod;
    FNumberOfRepeatsAllowed : TFhirPositiveInt;
    FQuantity : TFhirQuantity;
    FExpectedSupplyDuration : TFhirQuantity;
    Procedure SetMedication(value : TFhirType);
    Procedure SetValidityPeriod(value : TFhirPeriod);
    Procedure SetNumberOfRepeatsAllowed(value : TFhirPositiveInt);
    Function GetNumberOfRepeatsAllowedST : String;
    Procedure SetNumberOfRepeatsAllowedST(value : String);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetExpectedSupplyDuration(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationOrderDispenseRequest; overload;
    function Clone : TFhirMedicationOrderDispenseRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member medication
      Typed access to Identifies the medication being administered. This is a link to a resource that represents the medication which may be the details of the medication or simply an attribute carrying a code that identifies the medication from a known list of medications. (defined for API consistency)
    }
    property medication : TFhirType read FMedication write SetMedication;
    {@member medicationElement
      Identifies the medication being administered. This is a link to a resource that represents the medication which may be the details of the medication or simply an attribute carrying a code that identifies the medication from a known list of medications.
    }
    property medicationElement : TFhirType read FMedication write SetMedication;

    {@member validityPeriod
      Typed access to This indicates the validity period of a prescription (stale dating the Prescription). (defined for API consistency)
    }
    property validityPeriod : TFhirPeriod read FValidityPeriod write SetValidityPeriod;
    {@member validityPeriodElement
      This indicates the validity period of a prescription (stale dating the Prescription).
    }
    property validityPeriodElement : TFhirPeriod read FValidityPeriod write SetValidityPeriod;

    {@member numberOfRepeatsAllowed
      Typed access to An integer indicating the number of additional times (aka refills or repeats) the patient can receive the prescribed medication.   Usage Notes: This integer does NOT include the original order dispense.   This means that if an order indicates dispense 30 tablets plus  "3 repeats", then the order can be dispensed a total of 4 times and the patient can receive a total of 120 tablets.
    }
    property numberOfRepeatsAllowed : String read GetNumberOfRepeatsAllowedST write SetNumberOfRepeatsAllowedST;
    {@member numberOfRepeatsAllowedElement
      An integer indicating the number of additional times (aka refills or repeats) the patient can receive the prescribed medication.   Usage Notes: This integer does NOT include the original order dispense.   This means that if an order indicates dispense 30 tablets plus  "3 repeats", then the order can be dispensed a total of 4 times and the patient can receive a total of 120 tablets.
    }
    property numberOfRepeatsAllowedElement : TFhirPositiveInt read FNumberOfRepeatsAllowed write SetNumberOfRepeatsAllowed;

    {@member quantity
      Typed access to The amount that is to be dispensed for one fill. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount that is to be dispensed for one fill.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member expectedSupplyDuration
      Typed access to Identifies the period time over which the supplied product is expected to be used, or the length of time the dispense is expected to last. (defined for API consistency)
    }
    property expectedSupplyDuration : TFhirQuantity read FExpectedSupplyDuration write SetExpectedSupplyDuration;
    {@member expectedSupplyDurationElement
      Identifies the period time over which the supplied product is expected to be used, or the length of time the dispense is expected to last.
    }
    property expectedSupplyDurationElement : TFhirQuantity read FExpectedSupplyDuration write SetExpectedSupplyDuration;

  end;


  TFhirMedicationOrderDispenseRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationOrderDispenseRequestList;
    function GetCurrent : TFhirMedicationOrderDispenseRequest;
  public
    Constructor Create(list : TFhirMedicationOrderDispenseRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationOrderDispenseRequest read GetCurrent;
  end;


  {@Class TFhirMedicationOrderDispenseRequestList
    A list of FhirMedicationOrderDispenseRequest
  }
  {!.Net HL7Connect.Fhir.MedicationOrderDispenseRequestList}
  TFhirMedicationOrderDispenseRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationOrderDispenseRequest;
    procedure SetItemN(index : Integer; value : TFhirMedicationOrderDispenseRequest);
  public
    {!script hide}
    function Link : TFhirMedicationOrderDispenseRequestList; Overload;
    function Clone : TFhirMedicationOrderDispenseRequestList; Overload;
    function GetEnumerator : TFhirMedicationOrderDispenseRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationOrderDispenseRequest to the end of the list.
    }
    function Append : TFhirMedicationOrderDispenseRequest;

    
    {@member AddItem
      Add an already existing FhirMedicationOrderDispenseRequest to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationOrderDispenseRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationOrderDispenseRequest) : Integer;
    

    {@member Insert
      Insert FhirMedicationOrderDispenseRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationOrderDispenseRequest;
    

    {@member InsertItem
       Insert an existing FhirMedicationOrderDispenseRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationOrderDispenseRequest);
    
    {@member Item
       Get the iIndexth FhirMedicationOrderDispenseRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationOrderDispenseRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationOrderDispenseRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationOrderDispenseRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationOrderDispenseRequests[index : Integer] : TFhirMedicationOrderDispenseRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationOrderSubstitution : TFhirElement
    Indicates whether or not substitution can or should be part of the dispense. In some cases substitution must happen, in other cases substitution must not happen, and in others it does not matter. This block explains the prescriber's intent. If nothing is specified substitution may be done.
  }
  {!.Net HL7Connect.Fhir.MedicationOrderSubstitution}
  TFhirMedicationOrderSubstitution = class (TFhirBackboneElement)
  private
    FType_ : TFhirCodeableConcept;
    FReason : TFhirCodeableConcept;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetReason(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationOrderSubstitution; overload;
    function Clone : TFhirMedicationOrderSubstitution; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to A code signifying whether a different drug should be dispensed from what was prescribed. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      A code signifying whether a different drug should be dispensed from what was prescribed.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member reason
      Typed access to Indicates the reason for the substitution, or why substitution must or must not be performed. (defined for API consistency)
    }
    property reason : TFhirCodeableConcept read FReason write SetReason;
    {@member reasonElement
      Indicates the reason for the substitution, or why substitution must or must not be performed.
    }
    property reasonElement : TFhirCodeableConcept read FReason write SetReason;

  end;


  TFhirMedicationOrderSubstitutionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationOrderSubstitutionList;
    function GetCurrent : TFhirMedicationOrderSubstitution;
  public
    Constructor Create(list : TFhirMedicationOrderSubstitutionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationOrderSubstitution read GetCurrent;
  end;


  {@Class TFhirMedicationOrderSubstitutionList
    A list of FhirMedicationOrderSubstitution
  }
  {!.Net HL7Connect.Fhir.MedicationOrderSubstitutionList}
  TFhirMedicationOrderSubstitutionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationOrderSubstitution;
    procedure SetItemN(index : Integer; value : TFhirMedicationOrderSubstitution);
  public
    {!script hide}
    function Link : TFhirMedicationOrderSubstitutionList; Overload;
    function Clone : TFhirMedicationOrderSubstitutionList; Overload;
    function GetEnumerator : TFhirMedicationOrderSubstitutionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationOrderSubstitution to the end of the list.
    }
    function Append : TFhirMedicationOrderSubstitution;

    
    {@member AddItem
      Add an already existing FhirMedicationOrderSubstitution to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationOrderSubstitution); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationOrderSubstitution) : Integer;
    

    {@member Insert
      Insert FhirMedicationOrderSubstitution before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationOrderSubstitution;
    

    {@member InsertItem
       Insert an existing FhirMedicationOrderSubstitution before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationOrderSubstitution);
    
    {@member Item
       Get the iIndexth FhirMedicationOrderSubstitution. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationOrderSubstitution. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationOrderSubstitution);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationOrderSubstitution;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationOrderSubstitutions[index : Integer] : TFhirMedicationOrderSubstitution read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationOrder : TFhirDomainResource
    An order for both supply of the medication and the instructions for administration of the medication to a patient. The resource is called "MedicationOrder" rather than "MedicationPrescription" to generalize the use across inpatient and outpatient settings as well as for care plans, etc.
  }
  {!.Net HL7Connect.Fhir.MedicationOrder}
  TFhirMedicationOrder = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FDateWritten : TFhirDateTime;
    FStatus : TFhirEnum;
    FDateEnded : TFhirDateTime;
    FReasonEnded : TFhirCodeableConcept;
    FPatient : TFhirReference{TFhirPatient};
    FPrescriber : TFhirReference{TFhirPractitioner};
    FEncounter : TFhirReference{TFhirEncounter};
    FReason : TFhirType;
    FNote : TFhirString;
    FMedication : TFhirType;
    FdosageInstructionList : TFhirMedicationOrderDosageInstructionList;
    FDispenseRequest : TFhirMedicationOrderDispenseRequest;
    FSubstitution : TFhirMedicationOrderSubstitution;
    FPriorPrescription : TFhirReference{TFhirMedicationOrder};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetDateWritten(value : TFhirDateTime);
    Function GetDateWrittenST : TDateAndTime;
    Procedure SetDateWrittenST(value : TDateAndTime);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirMedicationOrderStatus;
    Procedure SetStatusST(value : TFhirMedicationOrderStatus);
    Procedure SetDateEnded(value : TFhirDateTime);
    Function GetDateEndedST : TDateAndTime;
    Procedure SetDateEndedST(value : TDateAndTime);
    Procedure SetReasonEnded(value : TFhirCodeableConcept);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetPrescriber(value : TFhirReference{TFhirPractitioner});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetReason(value : TFhirType);
    Procedure SetNote(value : TFhirString);
    Function GetNoteST : String;
    Procedure SetNoteST(value : String);
    Procedure SetMedication(value : TFhirType);
    function GetDosageInstructionList : TFhirMedicationOrderDosageInstructionList;
    function GetHasDosageInstructionList : Boolean;
    Procedure SetDispenseRequest(value : TFhirMedicationOrderDispenseRequest);
    Procedure SetSubstitution(value : TFhirMedicationOrderSubstitution);
    Procedure SetPriorPrescription(value : TFhirReference{TFhirMedicationOrder});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationOrder; overload;
    function Clone : TFhirMedicationOrder; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External identifier - one that would be used by another non-FHIR system - for example a re-imbursement system might issue its own id for each prescription that is created.  This is particularly important where FHIR only provides part of an entire workflow process where records have to be tracked through an entire system.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member dateWritten
      Typed access to The date (and perhaps time) when the prescription was written.
    }
    property dateWritten : TDateAndTime read GetDateWrittenST write SetDateWrittenST;
    {@member dateWrittenElement
      The date (and perhaps time) when the prescription was written.
    }
    property dateWrittenElement : TFhirDateTime read FDateWritten write SetDateWritten;

    {@member status
      A code specifying the state of the order.  Generally this will be active or completed state.
    }
    property status : TFhirMedicationOrderStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member dateEnded
      Typed access to The date (and perhaps time) when the prescription was stopped.
    }
    property dateEnded : TDateAndTime read GetDateEndedST write SetDateEndedST;
    {@member dateEndedElement
      The date (and perhaps time) when the prescription was stopped.
    }
    property dateEndedElement : TFhirDateTime read FDateEnded write SetDateEnded;

    {@member reasonEnded
      Typed access to The reason why the prescription was stopped, if it was. (defined for API consistency)
    }
    property reasonEnded : TFhirCodeableConcept read FReasonEnded write SetReasonEnded;
    {@member reasonEndedElement
      The reason why the prescription was stopped, if it was.
    }
    property reasonEndedElement : TFhirCodeableConcept read FReasonEnded write SetReasonEnded;

    {@member patient
      Typed access to A link to a resource representing the person to whom the medication will be given. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A link to a resource representing the person to whom the medication will be given.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member prescriber
      Typed access to The healthcare professional responsible for authorizing the prescription. (defined for API consistency)
    }
    property prescriber : TFhirReference{TFhirPractitioner} read FPrescriber write SetPrescriber;
    {@member prescriberElement
      The healthcare professional responsible for authorizing the prescription.
    }
    property prescriberElement : TFhirReference{TFhirPractitioner} read FPrescriber write SetPrescriber;

    {@member encounter
      Typed access to A link to a resource that identifies the particular occurrence of contact between patient and health care provider. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      A link to a resource that identifies the particular occurrence of contact between patient and health care provider.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member reason
      Typed access to Can be the reason or the indication for writing the prescription. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      Can be the reason or the indication for writing the prescription.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member note
      Typed access to Extra information about the prescription that could not be conveyed by the other attributes.
    }
    property note : String read GetNoteST write SetNoteST;
    {@member noteElement
      Extra information about the prescription that could not be conveyed by the other attributes.
    }
    property noteElement : TFhirString read FNote write SetNote;

    {@member medication
      Typed access to Identifies the medication being administered. This is a link to a resource that represents the medication which may be the details of the medication or simply an attribute carrying a code that identifies the medication from a known list of medications. (defined for API consistency)
    }
    property medication : TFhirType read FMedication write SetMedication;
    {@member medicationElement
      Identifies the medication being administered. This is a link to a resource that represents the medication which may be the details of the medication or simply an attribute carrying a code that identifies the medication from a known list of medications.
    }
    property medicationElement : TFhirType read FMedication write SetMedication;

    {@member dosageInstructionList
      Indicates how the medication is to be used by the patient.
    }
    property dosageInstructionList : TFhirMedicationOrderDosageInstructionList read GetDosageInstructionList;
    property hasDosageInstructionList : boolean read GetHasDosageInstructionList;

    {@member dispenseRequest
      Typed access to Indicates the specific details for the dispense or medication supply part of a medication order (also known as a Medication Prescription).  Note that this information is NOT always sent with the order.  There may be in some settings (e.g. hospitals) institutional or system support for completing the dispense details in the pharmacy department. (defined for API consistency)
    }
    property dispenseRequest : TFhirMedicationOrderDispenseRequest read FDispenseRequest write SetDispenseRequest;
    {@member dispenseRequestElement
      Indicates the specific details for the dispense or medication supply part of a medication order (also known as a Medication Prescription).  Note that this information is NOT always sent with the order.  There may be in some settings (e.g. hospitals) institutional or system support for completing the dispense details in the pharmacy department.
    }
    property dispenseRequestElement : TFhirMedicationOrderDispenseRequest read FDispenseRequest write SetDispenseRequest;

    {@member substitution
      Typed access to Indicates whether or not substitution can or should be part of the dispense. In some cases substitution must happen, in other cases substitution must not happen, and in others it does not matter. This block explains the prescriber's intent. If nothing is specified substitution may be done. (defined for API consistency)
    }
    property substitution : TFhirMedicationOrderSubstitution read FSubstitution write SetSubstitution;
    {@member substitutionElement
      Indicates whether or not substitution can or should be part of the dispense. In some cases substitution must happen, in other cases substitution must not happen, and in others it does not matter. This block explains the prescriber's intent. If nothing is specified substitution may be done.
    }
    property substitutionElement : TFhirMedicationOrderSubstitution read FSubstitution write SetSubstitution;

    {@member priorPrescription
      Typed access to A link to a resource representing an earlier order or prescription that this order supersedes. (defined for API consistency)
    }
    property priorPrescription : TFhirReference{TFhirMedicationOrder} read FPriorPrescription write SetPriorPrescription;
    {@member priorPrescriptionElement
      A link to a resource representing an earlier order or prescription that this order supersedes.
    }
    property priorPrescriptionElement : TFhirReference{TFhirMedicationOrder} read FPriorPrescription write SetPriorPrescription;

  end;


  TFhirMedicationOrderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationOrderList;
    function GetCurrent : TFhirMedicationOrder;
  public
    Constructor Create(list : TFhirMedicationOrderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationOrder read GetCurrent;
  end;


  {@Class TFhirMedicationOrderList
    A list of FhirMedicationOrder
  }
  {!.Net HL7Connect.Fhir.MedicationOrderList}
  TFhirMedicationOrderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationOrder;
    procedure SetItemN(index : Integer; value : TFhirMedicationOrder);
  public
    {!script hide}
    function Link : TFhirMedicationOrderList; Overload;
    function Clone : TFhirMedicationOrderList; Overload;
    function GetEnumerator : TFhirMedicationOrderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationOrder to the end of the list.
    }
    function Append : TFhirMedicationOrder;

    
    {@member AddItem
      Add an already existing FhirMedicationOrder to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationOrder); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationOrder) : Integer;
    

    {@member Insert
      Insert FhirMedicationOrder before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationOrder;
    

    {@member InsertItem
       Insert an existing FhirMedicationOrder before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationOrder);
    
    {@member Item
       Get the iIndexth FhirMedicationOrder. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationOrder. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationOrder);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationOrder;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationOrders[index : Integer] : TFhirMedicationOrder read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationStatementDosage : TFhirElement
    Indicates how the medication is/was used by the patient.
  }
  {!.Net HL7Connect.Fhir.MedicationStatementDosage}
  TFhirMedicationStatementDosage = class (TFhirBackboneElement)
  private
    FText : TFhirString;
    FTiming : TFhirTiming;
    FAsNeeded : TFhirType;
    FSite : TFhirType;
    FRoute : TFhirCodeableConcept;
    FMethod : TFhirCodeableConcept;
    FQuantity : TFhirType;
    FRate : TFhirType;
    FMaxDosePerPeriod : TFhirRatio;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetTiming(value : TFhirTiming);
    Procedure SetAsNeeded(value : TFhirType);
    Procedure SetSite(value : TFhirType);
    Procedure SetRoute(value : TFhirCodeableConcept);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetQuantity(value : TFhirType);
    Procedure SetRate(value : TFhirType);
    Procedure SetMaxDosePerPeriod(value : TFhirRatio);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationStatementDosage; overload;
    function Clone : TFhirMedicationStatementDosage; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member text
      Typed access to Free text dosage information as reported about a patient's medication use. When coded dosage information is present, the free text may still be present for display to humans.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Free text dosage information as reported about a patient's medication use. When coded dosage information is present, the free text may still be present for display to humans.
    }
    property textElement : TFhirString read FText write SetText;

    {@member timing
      Typed access to The timing schedule for giving the medication to the patient.  The Schedule data type allows many different expressions, for example.  "Every  8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:";  "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013". (defined for API consistency)
    }
    property timing : TFhirTiming read FTiming write SetTiming;
    {@member timingElement
      The timing schedule for giving the medication to the patient.  The Schedule data type allows many different expressions, for example.  "Every  8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:";  "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013".
    }
    property timingElement : TFhirTiming read FTiming write SetTiming;

    {@member asNeeded
      Typed access to Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept).  

Specifically if 'boolean' datatype is selected, then the following logic applies:  If set to True, this indicates that the medication is only taken when needed, within the specified schedule. (defined for API consistency)
    }
    property asNeeded : TFhirType read FAsNeeded write SetAsNeeded;
    {@member asNeededElement
      Indicates whether the Medication is only taken when needed within a specific dosing schedule (Boolean option), or it indicates the precondition for taking the Medication (CodeableConcept).  

Specifically if 'boolean' datatype is selected, then the following logic applies:  If set to True, this indicates that the medication is only taken when needed, within the specified schedule.
    }
    property asNeededElement : TFhirType read FAsNeeded write SetAsNeeded;

    {@member site
      Typed access to A coded specification of or a reference to the anatomic site where the medication first enters the body. (defined for API consistency)
    }
    property site : TFhirType read FSite write SetSite;
    {@member siteElement
      A coded specification of or a reference to the anatomic site where the medication first enters the body.
    }
    property siteElement : TFhirType read FSite write SetSite;

    {@member route
      Typed access to A code specifying the route or physiological path of administration of a therapeutic agent into or onto a subject. (defined for API consistency)
    }
    property route : TFhirCodeableConcept read FRoute write SetRoute;
    {@member routeElement
      A code specifying the route or physiological path of administration of a therapeutic agent into or onto a subject.
    }
    property routeElement : TFhirCodeableConcept read FRoute write SetRoute;

    {@member method
      Typed access to A coded value indicating the method by which the medication is intended to be or was introduced into or on the body.  This attribute will most often NOT be populated.  It is most commonly used for injections.  For example, Slow Push, Deep IV. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      A coded value indicating the method by which the medication is intended to be or was introduced into or on the body.  This attribute will most often NOT be populated.  It is most commonly used for injections.  For example, Slow Push, Deep IV.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member quantity
      Typed access to The amount of therapeutic or other substance given at one administration event. (defined for API consistency)
    }
    property quantity : TFhirType read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of therapeutic or other substance given at one administration event.
    }
    property quantityElement : TFhirType read FQuantity write SetQuantity;

    {@member rate
      Typed access to Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours. (defined for API consistency)
    }
    property rate : TFhirType read FRate write SetRate;
    {@member rateElement
      Identifies the speed with which the medication was or will be introduced into the patient. Typically the rate for an infusion e.g. 100 ml per 1 hour or 100 ml/hr.  May also be expressed as a rate per unit of time e.g. 500 ml per 2 hours.   Currently we do not specify a default of '1' in the denominator, but this is being discussed. Other examples: 200 mcg/min or 200 mcg/1 minute; 1 liter/8 hours.
    }
    property rateElement : TFhirType read FRate write SetRate;

    {@member maxDosePerPeriod
      Typed access to The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time.  For example, 1000mg in 24 hours. (defined for API consistency)
    }
    property maxDosePerPeriod : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;
    {@member maxDosePerPeriodElement
      The maximum total quantity of a therapeutic substance that may be administered to a subject over the period of time.  For example, 1000mg in 24 hours.
    }
    property maxDosePerPeriodElement : TFhirRatio read FMaxDosePerPeriod write SetMaxDosePerPeriod;

  end;


  TFhirMedicationStatementDosageListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationStatementDosageList;
    function GetCurrent : TFhirMedicationStatementDosage;
  public
    Constructor Create(list : TFhirMedicationStatementDosageList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationStatementDosage read GetCurrent;
  end;


  {@Class TFhirMedicationStatementDosageList
    A list of FhirMedicationStatementDosage
  }
  {!.Net HL7Connect.Fhir.MedicationStatementDosageList}
  TFhirMedicationStatementDosageList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationStatementDosage;
    procedure SetItemN(index : Integer; value : TFhirMedicationStatementDosage);
  public
    {!script hide}
    function Link : TFhirMedicationStatementDosageList; Overload;
    function Clone : TFhirMedicationStatementDosageList; Overload;
    function GetEnumerator : TFhirMedicationStatementDosageListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationStatementDosage to the end of the list.
    }
    function Append : TFhirMedicationStatementDosage;

    
    {@member AddItem
      Add an already existing FhirMedicationStatementDosage to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationStatementDosage); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationStatementDosage) : Integer;
    

    {@member Insert
      Insert FhirMedicationStatementDosage before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationStatementDosage;
    

    {@member InsertItem
       Insert an existing FhirMedicationStatementDosage before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationStatementDosage);
    
    {@member Item
       Get the iIndexth FhirMedicationStatementDosage. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationStatementDosage. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationStatementDosage);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationStatementDosage;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationStatementDosages[index : Integer] : TFhirMedicationStatementDosage read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMedicationStatement : TFhirDomainResource
    A record of a medication that is being consumed by a patient.   A MedicationStatement may indicate that the patient may be taking the medication now, or has taken the medication in the past or will be taking the medication in the future.  The source of this information can be the patient, significant other (such as a family member or spouse), or a clinician.  A common scenario where this information is captured is during the history taking process during a patient visit or stay.   The medication information may come from e.g. the patient's memory, from a prescription bottle,  or from a list of medications the patient, clinician or other party maintains 

The primary difference between a medication statement and a medication administration is that the medication administration has complete administration information and is based on actual administration information from the person who administered the medication.  A medication statement is often, if not always, less specific.  There is no required date/time when the medication was administered, in fact we only know that a source has reported the patient is taking this medication, where details such as time, quantity, or rate or even medication product may be incomplete or missing or less precise.  As stated earlier, the medication statement information may come from the patient's memory, from a prescription bottle or from a list of medications the patient, clinician or other party maintains.  Medication administration is more formal and is not missing detailed information.
  }
  {!.Net HL7Connect.Fhir.MedicationStatement}
  TFhirMedicationStatement = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FPatient : TFhirReference{TFhirPatient};
    FInformationSource : TFhirReference{Resource};
    FDateAsserted : TFhirDateTime;
    FStatus : TFhirEnum;
    FWasNotTaken : TFhirBoolean;
    FreasonNotTakenList : TFhirCodeableConceptList;
    FReasonForUse : TFhirType;
    FEffective : TFhirType;
    FNote : TFhirString;
    FsupportingInformationList : TFhirReferenceList{TFhirReference};
    FMedication : TFhirType;
    FdosageList : TFhirMedicationStatementDosageList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetInformationSource(value : TFhirReference{Resource});
    Procedure SetDateAsserted(value : TFhirDateTime);
    Function GetDateAssertedST : TDateAndTime;
    Procedure SetDateAssertedST(value : TDateAndTime);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirMedicationStatementStatus;
    Procedure SetStatusST(value : TFhirMedicationStatementStatus);
    Procedure SetWasNotTaken(value : TFhirBoolean);
    Function GetWasNotTakenST : Boolean;
    Procedure SetWasNotTakenST(value : Boolean);
    function GetReasonNotTakenList : TFhirCodeableConceptList;
    function GetHasReasonNotTakenList : Boolean;
    Procedure SetReasonForUse(value : TFhirType);
    Procedure SetEffective(value : TFhirType);
    Procedure SetNote(value : TFhirString);
    Function GetNoteST : String;
    Procedure SetNoteST(value : String);
    function GetSupportingInformationList : TFhirReferenceList{TFhirReference};
    function GetHasSupportingInformationList : Boolean;
    Procedure SetMedication(value : TFhirType);
    function GetDosageList : TFhirMedicationStatementDosageList;
    function GetHasDosageList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMedicationStatement; overload;
    function Clone : TFhirMedicationStatement; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External identifier - FHIR will generate its own internal identifiers (probably URLs) which do not need to be explicitly managed by the resource.  The identifier here is one that would be used by another non-FHIR system - for example an automated medication pump would provide a record each time it operated; an administration while the patient was off the ward might be made with a different system and entered after the event.  Particularly important if these records have to be updated.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member patient
      Typed access to The person or animal who is/was taking the medication. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The person or animal who is/was taking the medication.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member informationSource
      Typed access to The person who provided the information about the taking of this medication. (defined for API consistency)
    }
    property informationSource : TFhirReference{Resource} read FInformationSource write SetInformationSource;
    {@member informationSourceElement
      The person who provided the information about the taking of this medication.
    }
    property informationSourceElement : TFhirReference{Resource} read FInformationSource write SetInformationSource;

    {@member dateAsserted
      Typed access to The date when the medication statement was asserted by the information source.
    }
    property dateAsserted : TDateAndTime read GetDateAssertedST write SetDateAssertedST;
    {@member dateAssertedElement
      The date when the medication statement was asserted by the information source.
    }
    property dateAssertedElement : TFhirDateTime read FDateAsserted write SetDateAsserted;

    {@member status
      A code representing the patient or other source's judgment about the state of the medication used that this statement is about.  Generally this will be active or completed.
    }
    property status : TFhirMedicationStatementStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member wasNotTaken
      Typed access to Set this to true if the record is saying that the medication was NOT taken.
    }
    property wasNotTaken : Boolean read GetWasNotTakenST write SetWasNotTakenST;
    {@member wasNotTakenElement
      Set this to true if the record is saying that the medication was NOT taken.
    }
    property wasNotTakenElement : TFhirBoolean read FWasNotTaken write SetWasNotTaken;

    {@member reasonNotTakenList
      A code indicating why the medication was not taken.
    }
    property reasonNotTakenList : TFhirCodeableConceptList read GetReasonNotTakenList;
    property hasReasonNotTakenList : boolean read GetHasReasonNotTakenList;

    {@member reasonForUse
      Typed access to A reason for why the medication is being/was taken. (defined for API consistency)
    }
    property reasonForUse : TFhirType read FReasonForUse write SetReasonForUse;
    {@member reasonForUseElement
      A reason for why the medication is being/was taken.
    }
    property reasonForUseElement : TFhirType read FReasonForUse write SetReasonForUse;

    {@member effective
      Typed access to The interval of time during which it is being asserted that the patient was taking the medication (or was not taking, when the wasNotGiven element is true). (defined for API consistency)
    }
    property effective : TFhirType read FEffective write SetEffective;
    {@member effectiveElement
      The interval of time during which it is being asserted that the patient was taking the medication (or was not taking, when the wasNotGiven element is true).
    }
    property effectiveElement : TFhirType read FEffective write SetEffective;

    {@member note
      Typed access to Provides extra information about the medication statement that is not conveyed by the other attributes.
    }
    property note : String read GetNoteST write SetNoteST;
    {@member noteElement
      Provides extra information about the medication statement that is not conveyed by the other attributes.
    }
    property noteElement : TFhirString read FNote write SetNote;

    {@member supportingInformationList
      Allows linking the MedicationStatement to the underlying MedicationOrder, or to other information that supports the MedicationStatement.
    }
    property supportingInformationList : TFhirReferenceList{TFhirReference} read GetSupportingInformationList;
    property hasSupportingInformationList : boolean read GetHasSupportingInformationList;

    {@member medication
      Typed access to Identifies the medication being administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications. (defined for API consistency)
    }
    property medication : TFhirType read FMedication write SetMedication;
    {@member medicationElement
      Identifies the medication being administered. This is either a link to a resource representing the details of the medication or a simple attribute carrying a code that identifies the medication from a known list of medications.
    }
    property medicationElement : TFhirType read FMedication write SetMedication;

    {@member dosageList
      Indicates how the medication is/was used by the patient.
    }
    property dosageList : TFhirMedicationStatementDosageList read GetDosageList;
    property hasDosageList : boolean read GetHasDosageList;

  end;


  TFhirMedicationStatementListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMedicationStatementList;
    function GetCurrent : TFhirMedicationStatement;
  public
    Constructor Create(list : TFhirMedicationStatementList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMedicationStatement read GetCurrent;
  end;


  {@Class TFhirMedicationStatementList
    A list of FhirMedicationStatement
  }
  {!.Net HL7Connect.Fhir.MedicationStatementList}
  TFhirMedicationStatementList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMedicationStatement;
    procedure SetItemN(index : Integer; value : TFhirMedicationStatement);
  public
    {!script hide}
    function Link : TFhirMedicationStatementList; Overload;
    function Clone : TFhirMedicationStatementList; Overload;
    function GetEnumerator : TFhirMedicationStatementListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMedicationStatement to the end of the list.
    }
    function Append : TFhirMedicationStatement;

    
    {@member AddItem
      Add an already existing FhirMedicationStatement to the end of the list.
    }
    procedure AddItem(value : TFhirMedicationStatement); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMedicationStatement) : Integer;
    

    {@member Insert
      Insert FhirMedicationStatement before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMedicationStatement;
    

    {@member InsertItem
       Insert an existing FhirMedicationStatement before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMedicationStatement);
    
    {@member Item
       Get the iIndexth FhirMedicationStatement. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMedicationStatement. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMedicationStatement);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMedicationStatement;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMedicationStatements[index : Integer] : TFhirMedicationStatement read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMessageHeaderResponse : TFhirElement
    Information about the message that this message is a response to.  Only present if this message is a response.
  }
  {!.Net HL7Connect.Fhir.MessageHeaderResponse}
  TFhirMessageHeaderResponse = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirId;
    FCode : TFhirEnum;
    FDetails : TFhirReference{TFhirOperationOutcome};
    Procedure SetIdentifier(value : TFhirId);
    Function GetIdentifierST : String;
    Procedure SetIdentifierST(value : String);
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirResponseCode;
    Procedure SetCodeST(value : TFhirResponseCode);
    Procedure SetDetails(value : TFhirReference{TFhirOperationOutcome});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMessageHeaderResponse; overload;
    function Clone : TFhirMessageHeaderResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to The id of the message that this message is a response to.
    }
    property identifier : String read GetIdentifierST write SetIdentifierST;
    {@member identifierElement
      The id of the message that this message is a response to.
    }
    property identifierElement : TFhirId read FIdentifier write SetIdentifier;

    {@member code
      Code that identifies the type of response to the message - whether it was successful or not, and whether it should be resent or not.
    }
    property code : TFhirResponseCode read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member details
      Typed access to Full details of any issues found in the message. (defined for API consistency)
    }
    property details : TFhirReference{TFhirOperationOutcome} read FDetails write SetDetails;
    {@member detailsElement
      Full details of any issues found in the message.
    }
    property detailsElement : TFhirReference{TFhirOperationOutcome} read FDetails write SetDetails;

  end;


  TFhirMessageHeaderResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMessageHeaderResponseList;
    function GetCurrent : TFhirMessageHeaderResponse;
  public
    Constructor Create(list : TFhirMessageHeaderResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMessageHeaderResponse read GetCurrent;
  end;


  {@Class TFhirMessageHeaderResponseList
    A list of FhirMessageHeaderResponse
  }
  {!.Net HL7Connect.Fhir.MessageHeaderResponseList}
  TFhirMessageHeaderResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMessageHeaderResponse;
    procedure SetItemN(index : Integer; value : TFhirMessageHeaderResponse);
  public
    {!script hide}
    function Link : TFhirMessageHeaderResponseList; Overload;
    function Clone : TFhirMessageHeaderResponseList; Overload;
    function GetEnumerator : TFhirMessageHeaderResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMessageHeaderResponse to the end of the list.
    }
    function Append : TFhirMessageHeaderResponse;

    
    {@member AddItem
      Add an already existing FhirMessageHeaderResponse to the end of the list.
    }
    procedure AddItem(value : TFhirMessageHeaderResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMessageHeaderResponse) : Integer;
    

    {@member Insert
      Insert FhirMessageHeaderResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMessageHeaderResponse;
    

    {@member InsertItem
       Insert an existing FhirMessageHeaderResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMessageHeaderResponse);
    
    {@member Item
       Get the iIndexth FhirMessageHeaderResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMessageHeaderResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMessageHeaderResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMessageHeaderResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMessageHeaderResponses[index : Integer] : TFhirMessageHeaderResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMessageHeaderSource : TFhirElement
    The source application from which this message originated.
  }
  {!.Net HL7Connect.Fhir.MessageHeaderSource}
  TFhirMessageHeaderSource = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FSoftware : TFhirString;
    FVersion : TFhirString;
    FContact : TFhirContactPoint;
    FEndpoint : TFhirUri;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetSoftware(value : TFhirString);
    Function GetSoftwareST : String;
    Procedure SetSoftwareST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetContact(value : TFhirContactPoint);
    Procedure SetEndpoint(value : TFhirUri);
    Function GetEndpointST : String;
    Procedure SetEndpointST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMessageHeaderSource; overload;
    function Clone : TFhirMessageHeaderSource; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to Human-readable name for the source system.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Human-readable name for the source system.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member software
      Typed access to May include configuration or other information useful in debugging.
    }
    property software : String read GetSoftwareST write SetSoftwareST;
    {@member softwareElement
      May include configuration or other information useful in debugging.
    }
    property softwareElement : TFhirString read FSoftware write SetSoftware;

    {@member version
      Typed access to Can convey versions of multiple systems in situations where a message passes through multiple hands.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      Can convey versions of multiple systems in situations where a message passes through multiple hands.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member contact
      Typed access to An e-mail, phone, website or other contact point to use to resolve issues with message communications. (defined for API consistency)
    }
    property contact : TFhirContactPoint read FContact write SetContact;
    {@member contactElement
      An e-mail, phone, website or other contact point to use to resolve issues with message communications.
    }
    property contactElement : TFhirContactPoint read FContact write SetContact;

    {@member endpoint
      Typed access to Identifies the routing target to send acknowledgements to.
    }
    property endpoint : String read GetEndpointST write SetEndpointST;
    {@member endpointElement
      Identifies the routing target to send acknowledgements to.
    }
    property endpointElement : TFhirUri read FEndpoint write SetEndpoint;

  end;


  TFhirMessageHeaderSourceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMessageHeaderSourceList;
    function GetCurrent : TFhirMessageHeaderSource;
  public
    Constructor Create(list : TFhirMessageHeaderSourceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMessageHeaderSource read GetCurrent;
  end;


  {@Class TFhirMessageHeaderSourceList
    A list of FhirMessageHeaderSource
  }
  {!.Net HL7Connect.Fhir.MessageHeaderSourceList}
  TFhirMessageHeaderSourceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMessageHeaderSource;
    procedure SetItemN(index : Integer; value : TFhirMessageHeaderSource);
  public
    {!script hide}
    function Link : TFhirMessageHeaderSourceList; Overload;
    function Clone : TFhirMessageHeaderSourceList; Overload;
    function GetEnumerator : TFhirMessageHeaderSourceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMessageHeaderSource to the end of the list.
    }
    function Append : TFhirMessageHeaderSource;

    
    {@member AddItem
      Add an already existing FhirMessageHeaderSource to the end of the list.
    }
    procedure AddItem(value : TFhirMessageHeaderSource); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMessageHeaderSource) : Integer;
    

    {@member Insert
      Insert FhirMessageHeaderSource before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMessageHeaderSource;
    

    {@member InsertItem
       Insert an existing FhirMessageHeaderSource before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMessageHeaderSource);
    
    {@member Item
       Get the iIndexth FhirMessageHeaderSource. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMessageHeaderSource. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMessageHeaderSource);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMessageHeaderSource;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMessageHeaderSources[index : Integer] : TFhirMessageHeaderSource read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMessageHeaderDestination : TFhirElement
    The destination application which the message is intended for.
  }
  {!.Net HL7Connect.Fhir.MessageHeaderDestination}
  TFhirMessageHeaderDestination = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FTarget : TFhirReference{TFhirDevice};
    FEndpoint : TFhirUri;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetTarget(value : TFhirReference{TFhirDevice});
    Procedure SetEndpoint(value : TFhirUri);
    Function GetEndpointST : String;
    Procedure SetEndpointST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMessageHeaderDestination; overload;
    function Clone : TFhirMessageHeaderDestination; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to Human-readable name for the target system.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Human-readable name for the target system.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member target
      Typed access to Identifies the target end system in situations where the initial message transmission is to an intermediary system. (defined for API consistency)
    }
    property target : TFhirReference{TFhirDevice} read FTarget write SetTarget;
    {@member targetElement
      Identifies the target end system in situations where the initial message transmission is to an intermediary system.
    }
    property targetElement : TFhirReference{TFhirDevice} read FTarget write SetTarget;

    {@member endpoint
      Typed access to Indicates where the message should be routed to.
    }
    property endpoint : String read GetEndpointST write SetEndpointST;
    {@member endpointElement
      Indicates where the message should be routed to.
    }
    property endpointElement : TFhirUri read FEndpoint write SetEndpoint;

  end;


  TFhirMessageHeaderDestinationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMessageHeaderDestinationList;
    function GetCurrent : TFhirMessageHeaderDestination;
  public
    Constructor Create(list : TFhirMessageHeaderDestinationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMessageHeaderDestination read GetCurrent;
  end;


  {@Class TFhirMessageHeaderDestinationList
    A list of FhirMessageHeaderDestination
  }
  {!.Net HL7Connect.Fhir.MessageHeaderDestinationList}
  TFhirMessageHeaderDestinationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMessageHeaderDestination;
    procedure SetItemN(index : Integer; value : TFhirMessageHeaderDestination);
  public
    {!script hide}
    function Link : TFhirMessageHeaderDestinationList; Overload;
    function Clone : TFhirMessageHeaderDestinationList; Overload;
    function GetEnumerator : TFhirMessageHeaderDestinationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMessageHeaderDestination to the end of the list.
    }
    function Append : TFhirMessageHeaderDestination;

    
    {@member AddItem
      Add an already existing FhirMessageHeaderDestination to the end of the list.
    }
    procedure AddItem(value : TFhirMessageHeaderDestination); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMessageHeaderDestination) : Integer;
    

    {@member Insert
      Insert FhirMessageHeaderDestination before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMessageHeaderDestination;
    

    {@member InsertItem
       Insert an existing FhirMessageHeaderDestination before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMessageHeaderDestination);
    
    {@member Item
       Get the iIndexth FhirMessageHeaderDestination. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMessageHeaderDestination. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMessageHeaderDestination);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMessageHeaderDestination;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMessageHeaderDestinations[index : Integer] : TFhirMessageHeaderDestination read GetItemN write SetItemN; default;
  End;


  {@Class TFhirMessageHeader : TFhirDomainResource
    The header for a message exchange that is either requesting or responding to an action.  The reference(s) that are the subject of the action as well as other information related to the action are typically transmitted in a bundle in which the MessageHeader resource instance is the first resource in the bundle.
  }
  {!.Net HL7Connect.Fhir.MessageHeader}
  TFhirMessageHeader = class (TFhirDomainResource)
  private
    FTimestamp : TFhirInstant;
    FEvent : TFhirCoding;
    FResponse : TFhirMessageHeaderResponse;
    FSource : TFhirMessageHeaderSource;
    FdestinationList : TFhirMessageHeaderDestinationList;
    FEnterer : TFhirReference{TFhirPractitioner};
    FAuthor : TFhirReference{TFhirPractitioner};
    FReceiver : TFhirReference{Resource};
    FResponsible : TFhirReference{Resource};
    FReason : TFhirCodeableConcept;
    FdataList : TFhirReferenceList{TFhirReference};
    Procedure SetTimestamp(value : TFhirInstant);
    Function GetTimestampST : TDateAndTime;
    Procedure SetTimestampST(value : TDateAndTime);
    Procedure SetEvent(value : TFhirCoding);
    Procedure SetResponse(value : TFhirMessageHeaderResponse);
    Procedure SetSource(value : TFhirMessageHeaderSource);
    function GetDestinationList : TFhirMessageHeaderDestinationList;
    function GetHasDestinationList : Boolean;
    Procedure SetEnterer(value : TFhirReference{TFhirPractitioner});
    Procedure SetAuthor(value : TFhirReference{TFhirPractitioner});
    Procedure SetReceiver(value : TFhirReference{Resource});
    Procedure SetResponsible(value : TFhirReference{Resource});
    Procedure SetReason(value : TFhirCodeableConcept);
    function GetDataList : TFhirReferenceList{TFhirReference};
    function GetHasDataList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirMessageHeader; overload;
    function Clone : TFhirMessageHeader; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member timestamp
      Typed access to The time that the message was sent.
    }
    property timestamp : TDateAndTime read GetTimestampST write SetTimestampST;
    {@member timestampElement
      The time that the message was sent.
    }
    property timestampElement : TFhirInstant read FTimestamp write SetTimestamp;

    {@member event
      Typed access to Code that identifies the event this message represents and connects it with its definition. Events defined as part of the FHIR specification have the system value "http://hl7.org/fhir/message-events". (defined for API consistency)
    }
    property event : TFhirCoding read FEvent write SetEvent;
    {@member eventElement
      Code that identifies the event this message represents and connects it with its definition. Events defined as part of the FHIR specification have the system value "http://hl7.org/fhir/message-events".
    }
    property eventElement : TFhirCoding read FEvent write SetEvent;

    {@member response
      Typed access to Information about the message that this message is a response to.  Only present if this message is a response. (defined for API consistency)
    }
    property response : TFhirMessageHeaderResponse read FResponse write SetResponse;
    {@member responseElement
      Information about the message that this message is a response to.  Only present if this message is a response.
    }
    property responseElement : TFhirMessageHeaderResponse read FResponse write SetResponse;

    {@member source
      Typed access to The source application from which this message originated. (defined for API consistency)
    }
    property source : TFhirMessageHeaderSource read FSource write SetSource;
    {@member sourceElement
      The source application from which this message originated.
    }
    property sourceElement : TFhirMessageHeaderSource read FSource write SetSource;

    {@member destinationList
      The destination application which the message is intended for.
    }
    property destinationList : TFhirMessageHeaderDestinationList read GetDestinationList;
    property hasDestinationList : boolean read GetHasDestinationList;

    {@member enterer
      Typed access to The person or device that performed the data entry leading to this message. Where there is more than one candidate, pick the most proximal to the message. Can provide other enterers in extensions. (defined for API consistency)
    }
    property enterer : TFhirReference{TFhirPractitioner} read FEnterer write SetEnterer;
    {@member entererElement
      The person or device that performed the data entry leading to this message. Where there is more than one candidate, pick the most proximal to the message. Can provide other enterers in extensions.
    }
    property entererElement : TFhirReference{TFhirPractitioner} read FEnterer write SetEnterer;

    {@member author
      Typed access to The logical author of the message - the person or device that decided the described event should happen. Where there is more than one candidate, pick the most proximal to the MessageHeader. Can provide other authors in extensions. (defined for API consistency)
    }
    property author : TFhirReference{TFhirPractitioner} read FAuthor write SetAuthor;
    {@member authorElement
      The logical author of the message - the person or device that decided the described event should happen. Where there is more than one candidate, pick the most proximal to the MessageHeader. Can provide other authors in extensions.
    }
    property authorElement : TFhirReference{TFhirPractitioner} read FAuthor write SetAuthor;

    {@member receiver
      Typed access to Allows data conveyed by a message to be addressed to a particular person or department when routing to a specific application isn't sufficient. (defined for API consistency)
    }
    property receiver : TFhirReference{Resource} read FReceiver write SetReceiver;
    {@member receiverElement
      Allows data conveyed by a message to be addressed to a particular person or department when routing to a specific application isn't sufficient.
    }
    property receiverElement : TFhirReference{Resource} read FReceiver write SetReceiver;

    {@member responsible
      Typed access to The person or organization that accepts overall responsibility for the contents of the message. The implication is that the message event happened under the policies of the responsible party. (defined for API consistency)
    }
    property responsible : TFhirReference{Resource} read FResponsible write SetResponsible;
    {@member responsibleElement
      The person or organization that accepts overall responsibility for the contents of the message. The implication is that the message event happened under the policies of the responsible party.
    }
    property responsibleElement : TFhirReference{Resource} read FResponsible write SetResponsible;

    {@member reason
      Typed access to Coded indication of the cause for the event - indicates  a reason for the occurrence of the event that is a focus of this message. (defined for API consistency)
    }
    property reason : TFhirCodeableConcept read FReason write SetReason;
    {@member reasonElement
      Coded indication of the cause for the event - indicates  a reason for the occurrence of the event that is a focus of this message.
    }
    property reasonElement : TFhirCodeableConcept read FReason write SetReason;

    {@member dataList
      The actual data of the message - a reference to the root/focus class of the event.
    }
    property dataList : TFhirReferenceList{TFhirReference} read GetDataList;
    property hasDataList : boolean read GetHasDataList;

  end;


  TFhirMessageHeaderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirMessageHeaderList;
    function GetCurrent : TFhirMessageHeader;
  public
    Constructor Create(list : TFhirMessageHeaderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirMessageHeader read GetCurrent;
  end;


  {@Class TFhirMessageHeaderList
    A list of FhirMessageHeader
  }
  {!.Net HL7Connect.Fhir.MessageHeaderList}
  TFhirMessageHeaderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirMessageHeader;
    procedure SetItemN(index : Integer; value : TFhirMessageHeader);
  public
    {!script hide}
    function Link : TFhirMessageHeaderList; Overload;
    function Clone : TFhirMessageHeaderList; Overload;
    function GetEnumerator : TFhirMessageHeaderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirMessageHeader to the end of the list.
    }
    function Append : TFhirMessageHeader;

    
    {@member AddItem
      Add an already existing FhirMessageHeader to the end of the list.
    }
    procedure AddItem(value : TFhirMessageHeader); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirMessageHeader) : Integer;
    

    {@member Insert
      Insert FhirMessageHeader before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirMessageHeader;
    

    {@member InsertItem
       Insert an existing FhirMessageHeader before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirMessageHeader);
    
    {@member Item
       Get the iIndexth FhirMessageHeader. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirMessageHeader. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirMessageHeader);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirMessageHeader;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirMessageHeaders[index : Integer] : TFhirMessageHeader read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNamingSystemContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.NamingSystemContact}
  TFhirNamingSystemContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNamingSystemContact; overload;
    function Clone : TFhirNamingSystemContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the naming system.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the naming system.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirNamingSystemContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNamingSystemContactList;
    function GetCurrent : TFhirNamingSystemContact;
  public
    Constructor Create(list : TFhirNamingSystemContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNamingSystemContact read GetCurrent;
  end;


  {@Class TFhirNamingSystemContactList
    A list of FhirNamingSystemContact
  }
  {!.Net HL7Connect.Fhir.NamingSystemContactList}
  TFhirNamingSystemContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNamingSystemContact;
    procedure SetItemN(index : Integer; value : TFhirNamingSystemContact);
  public
    {!script hide}
    function Link : TFhirNamingSystemContactList; Overload;
    function Clone : TFhirNamingSystemContactList; Overload;
    function GetEnumerator : TFhirNamingSystemContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNamingSystemContact to the end of the list.
    }
    function Append : TFhirNamingSystemContact;

    
    {@member AddItem
      Add an already existing FhirNamingSystemContact to the end of the list.
    }
    procedure AddItem(value : TFhirNamingSystemContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNamingSystemContact) : Integer;
    

    {@member Insert
      Insert FhirNamingSystemContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNamingSystemContact;
    

    {@member InsertItem
       Insert an existing FhirNamingSystemContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNamingSystemContact);
    
    {@member Item
       Get the iIndexth FhirNamingSystemContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNamingSystemContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNamingSystemContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNamingSystemContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNamingSystemContacts[index : Integer] : TFhirNamingSystemContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNamingSystemUniqueId : TFhirElement
    Indicates how the system may be identified when referenced in electronic exchange.
  }
  {!.Net HL7Connect.Fhir.NamingSystemUniqueId}
  TFhirNamingSystemUniqueId = class (TFhirBackboneElement)
  private
    FType_ : TFhirEnum;
    FValue : TFhirString;
    FPreferred : TFhirBoolean;
    FPeriod : TFhirPeriod;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirNamingsystemIdentifierType;
    Procedure SetType_ST(value : TFhirNamingsystemIdentifierType);
    Procedure SetValue(value : TFhirString);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
    Procedure SetPreferred(value : TFhirBoolean);
    Function GetPreferredST : Boolean;
    Procedure SetPreferredST(value : Boolean);
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNamingSystemUniqueId; overload;
    function Clone : TFhirNamingSystemUniqueId; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Identifies the unique identifier scheme used for this particular identifier.
    }
    property type_ : TFhirNamingsystemIdentifierType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member value
      Typed access to The string that should be sent over the wire to identify the code system or identifier system.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      The string that should be sent over the wire to identify the code system or identifier system.
    }
    property valueElement : TFhirString read FValue write SetValue;

    {@member preferred
      Typed access to Indicates whether this identifier is the "preferred" identifier of this type.
    }
    property preferred : Boolean read GetPreferredST write SetPreferredST;
    {@member preferredElement
      Indicates whether this identifier is the "preferred" identifier of this type.
    }
    property preferredElement : TFhirBoolean read FPreferred write SetPreferred;

    {@member period
      Typed access to Identifies the period of time over which this identifier is considered appropriate to refer to the naming system.  Outside of this window, the identifier might be non-deterministic. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      Identifies the period of time over which this identifier is considered appropriate to refer to the naming system.  Outside of this window, the identifier might be non-deterministic.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirNamingSystemUniqueIdListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNamingSystemUniqueIdList;
    function GetCurrent : TFhirNamingSystemUniqueId;
  public
    Constructor Create(list : TFhirNamingSystemUniqueIdList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNamingSystemUniqueId read GetCurrent;
  end;


  {@Class TFhirNamingSystemUniqueIdList
    A list of FhirNamingSystemUniqueId
  }
  {!.Net HL7Connect.Fhir.NamingSystemUniqueIdList}
  TFhirNamingSystemUniqueIdList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNamingSystemUniqueId;
    procedure SetItemN(index : Integer; value : TFhirNamingSystemUniqueId);
  public
    {!script hide}
    function Link : TFhirNamingSystemUniqueIdList; Overload;
    function Clone : TFhirNamingSystemUniqueIdList; Overload;
    function GetEnumerator : TFhirNamingSystemUniqueIdListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNamingSystemUniqueId to the end of the list.
    }
    function Append : TFhirNamingSystemUniqueId;

    
    {@member AddItem
      Add an already existing FhirNamingSystemUniqueId to the end of the list.
    }
    procedure AddItem(value : TFhirNamingSystemUniqueId); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNamingSystemUniqueId) : Integer;
    

    {@member Insert
      Insert FhirNamingSystemUniqueId before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNamingSystemUniqueId;
    

    {@member InsertItem
       Insert an existing FhirNamingSystemUniqueId before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNamingSystemUniqueId);
    
    {@member Item
       Get the iIndexth FhirNamingSystemUniqueId. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNamingSystemUniqueId. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNamingSystemUniqueId);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNamingSystemUniqueId;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNamingSystemUniqueIds[index : Integer] : TFhirNamingSystemUniqueId read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNamingSystem : TFhirDomainResource
    A curated namespace that issues unique symbols within that namespace for the identification of concepts, people, devices, etc.  Represents a "System" used within the Identifier and Coding data types.
  }
  {!.Net HL7Connect.Fhir.NamingSystem}
  TFhirNamingSystem = class (TFhirDomainResource)
  private
    FName : TFhirString;
    FStatus : TFhirEnum;
    FKind : TFhirEnum;
    FPublisher : TFhirString;
    FcontactList : TFhirNamingSystemContactList;
    FResponsible : TFhirString;
    FDate : TFhirDateTime;
    FType_ : TFhirCodeableConcept;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FUsage : TFhirString;
    FuniqueIdList : TFhirNamingSystemUniqueIdList;
    FReplacedBy : TFhirReference{TFhirNamingSystem};
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetKind(value : TFhirEnum);
    Function GetKindST : TFhirNamingsystemType;
    Procedure SetKindST(value : TFhirNamingsystemType);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirNamingSystemContactList;
    function GetHasContactList : Boolean;
    Procedure SetResponsible(value : TFhirString);
    Function GetResponsibleST : String;
    Procedure SetResponsibleST(value : String);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetUsage(value : TFhirString);
    Function GetUsageST : String;
    Procedure SetUsageST(value : String);
    function GetUniqueIdList : TFhirNamingSystemUniqueIdList;
    function GetHasUniqueIdList : Boolean;
    Procedure SetReplacedBy(value : TFhirReference{TFhirNamingSystem});
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNamingSystem; overload;
    function Clone : TFhirNamingSystem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member name
      Typed access to The descriptive name of this particular identifier type or code system.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The descriptive name of this particular identifier type or code system.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      Indicates whether the naming system is "ready for use" or not.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member kind
      Indicates the purpose for the naming system - what kinds of things does it make unique?
    }
    property kind : TFhirNamingsystemType read GetKindST write SetKindST;
    property kindElement : TFhirEnum read FKind write SetKind;

    {@member publisher
      Typed access to The name of the individual or organization that published the naming system.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the naming system.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirNamingSystemContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member responsible
      Typed access to The name of the organization that is responsible for issuing identifiers or codes for this namespace and ensuring their non-collision.
    }
    property responsible : String read GetResponsibleST write SetResponsibleST;
    {@member responsibleElement
      The name of the organization that is responsible for issuing identifiers or codes for this namespace and ensuring their non-collision.
    }
    property responsibleElement : TFhirString read FResponsible write SetResponsible;

    {@member date
      Typed access to The date  (and optionally time) when the system was registered or published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the registration changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date  (and optionally time) when the system was registered or published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the registration changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member type_
      Typed access to Categorizes a naming system for easier search by grouping related naming systems. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Categorizes a naming system for easier search by grouping related naming systems.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member description
      Typed access to Details about what the namespace identifies including scope, granularity, version labeling, etc.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Details about what the namespace identifies including scope, granularity, version labeling, etc.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of naming systems.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member usage
      Typed access to Provides guidance on the use of the namespace, including the handling of formatting characters, use of upper vs. lower case, etc.
    }
    property usage : String read GetUsageST write SetUsageST;
    {@member usageElement
      Provides guidance on the use of the namespace, including the handling of formatting characters, use of upper vs. lower case, etc.
    }
    property usageElement : TFhirString read FUsage write SetUsage;

    {@member uniqueIdList
      Indicates how the system may be identified when referenced in electronic exchange.
    }
    property uniqueIdList : TFhirNamingSystemUniqueIdList read GetUniqueIdList;
    property hasUniqueIdList : boolean read GetHasUniqueIdList;

    {@member replacedBy
      Typed access to For naming systems that are retired, indicates the naming system that should be used in their place (if any). (defined for API consistency)
    }
    property replacedBy : TFhirReference{TFhirNamingSystem} read FReplacedBy write SetReplacedBy;
    {@member replacedByElement
      For naming systems that are retired, indicates the naming system that should be used in their place (if any).
    }
    property replacedByElement : TFhirReference{TFhirNamingSystem} read FReplacedBy write SetReplacedBy;

  end;


  TFhirNamingSystemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNamingSystemList;
    function GetCurrent : TFhirNamingSystem;
  public
    Constructor Create(list : TFhirNamingSystemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNamingSystem read GetCurrent;
  end;


  {@Class TFhirNamingSystemList
    A list of FhirNamingSystem
  }
  {!.Net HL7Connect.Fhir.NamingSystemList}
  TFhirNamingSystemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNamingSystem;
    procedure SetItemN(index : Integer; value : TFhirNamingSystem);
  public
    {!script hide}
    function Link : TFhirNamingSystemList; Overload;
    function Clone : TFhirNamingSystemList; Overload;
    function GetEnumerator : TFhirNamingSystemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNamingSystem to the end of the list.
    }
    function Append : TFhirNamingSystem;

    
    {@member AddItem
      Add an already existing FhirNamingSystem to the end of the list.
    }
    procedure AddItem(value : TFhirNamingSystem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNamingSystem) : Integer;
    

    {@member Insert
      Insert FhirNamingSystem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNamingSystem;
    

    {@member InsertItem
       Insert an existing FhirNamingSystem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNamingSystem);
    
    {@member Item
       Get the iIndexth FhirNamingSystem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNamingSystem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNamingSystem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNamingSystem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNamingSystems[index : Integer] : TFhirNamingSystem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderOralDiet : TFhirElement
    Diet given orally in contrast to enteral (tube) feeding.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDiet}
  TFhirNutritionOrderOralDiet = class (TFhirBackboneElement)
  private
    Ftype_List : TFhirCodeableConceptList;
    FscheduleList : TFhirTimingList;
    FnutrientList : TFhirNutritionOrderOralDietNutrientList;
    FtextureList : TFhirNutritionOrderOralDietTextureList;
    FfluidConsistencyTypeList : TFhirCodeableConceptList;
    FInstruction : TFhirString;
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    function GetScheduleList : TFhirTimingList;
    function GetHasScheduleList : Boolean;
    function GetNutrientList : TFhirNutritionOrderOralDietNutrientList;
    function GetHasNutrientList : Boolean;
    function GetTextureList : TFhirNutritionOrderOralDietTextureList;
    function GetHasTextureList : Boolean;
    function GetFluidConsistencyTypeList : TFhirCodeableConceptList;
    function GetHasFluidConsistencyTypeList : Boolean;
    Procedure SetInstruction(value : TFhirString);
    Function GetInstructionST : String;
    Procedure SetInstructionST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderOralDiet; overload;
    function Clone : TFhirNutritionOrderOralDiet; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_List
      The kind of diet or dietary restriction such as fiber restricted diet or diabetic diet.
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member scheduleList
      The time period and frequency at which the diet should be given.
    }
    property scheduleList : TFhirTimingList read GetScheduleList;
    property hasScheduleList : boolean read GetHasScheduleList;

    {@member nutrientList
      Class that defines the quantity and type of nutrient modifications required for the oral diet.
    }
    property nutrientList : TFhirNutritionOrderOralDietNutrientList read GetNutrientList;
    property hasNutrientList : boolean read GetHasNutrientList;

    {@member textureList
      Class that describes any texture modifications required for the patient to safely consume various types of solid foods.
    }
    property textureList : TFhirNutritionOrderOralDietTextureList read GetTextureList;
    property hasTextureList : boolean read GetHasTextureList;

    {@member fluidConsistencyTypeList
      The required consistency (e.g. honey-thick, nectar-thick, thin, thickened.) of liquids or fluids served to the patient.
    }
    property fluidConsistencyTypeList : TFhirCodeableConceptList read GetFluidConsistencyTypeList;
    property hasFluidConsistencyTypeList : boolean read GetHasFluidConsistencyTypeList;

    {@member instruction
      Typed access to Free text or additional instructions or information pertaining to the oral diet.
    }
    property instruction : String read GetInstructionST write SetInstructionST;
    {@member instructionElement
      Free text or additional instructions or information pertaining to the oral diet.
    }
    property instructionElement : TFhirString read FInstruction write SetInstruction;

  end;


  TFhirNutritionOrderOralDietListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderOralDietList;
    function GetCurrent : TFhirNutritionOrderOralDiet;
  public
    Constructor Create(list : TFhirNutritionOrderOralDietList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderOralDiet read GetCurrent;
  end;


  {@Class TFhirNutritionOrderOralDietList
    A list of FhirNutritionOrderOralDiet
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDietList}
  TFhirNutritionOrderOralDietList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderOralDiet;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderOralDiet);
  public
    {!script hide}
    function Link : TFhirNutritionOrderOralDietList; Overload;
    function Clone : TFhirNutritionOrderOralDietList; Overload;
    function GetEnumerator : TFhirNutritionOrderOralDietListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderOralDiet to the end of the list.
    }
    function Append : TFhirNutritionOrderOralDiet;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderOralDiet to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderOralDiet); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderOralDiet) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderOralDiet before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderOralDiet;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderOralDiet before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderOralDiet);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDiet. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDiet. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderOralDiet);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderOralDiet;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderOralDiets[index : Integer] : TFhirNutritionOrderOralDiet read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderOralDietNutrient : TFhirElement
    Class that defines the quantity and type of nutrient modifications required for the oral diet.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDietNutrient}
  TFhirNutritionOrderOralDietNutrient = class (TFhirBackboneElement)
  private
    FModifier : TFhirCodeableConcept;
    FAmount : TFhirQuantity;
    Procedure SetModifier(value : TFhirCodeableConcept);
    Procedure SetAmount(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderOralDietNutrient; overload;
    function Clone : TFhirNutritionOrderOralDietNutrient; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member modifier
      Typed access to The nutrient that is being modified such as carbohydrate or sodium. (defined for API consistency)
    }
    property modifier : TFhirCodeableConcept read FModifier write SetModifier;
    {@member modifierElement
      The nutrient that is being modified such as carbohydrate or sodium.
    }
    property modifierElement : TFhirCodeableConcept read FModifier write SetModifier;

    {@member amount
      Typed access to The quantity of the specified nutrient to include in diet. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      The quantity of the specified nutrient to include in diet.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

  end;


  TFhirNutritionOrderOralDietNutrientListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderOralDietNutrientList;
    function GetCurrent : TFhirNutritionOrderOralDietNutrient;
  public
    Constructor Create(list : TFhirNutritionOrderOralDietNutrientList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderOralDietNutrient read GetCurrent;
  end;


  {@Class TFhirNutritionOrderOralDietNutrientList
    A list of FhirNutritionOrderOralDietNutrient
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDietNutrientList}
  TFhirNutritionOrderOralDietNutrientList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderOralDietNutrient;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderOralDietNutrient);
  public
    {!script hide}
    function Link : TFhirNutritionOrderOralDietNutrientList; Overload;
    function Clone : TFhirNutritionOrderOralDietNutrientList; Overload;
    function GetEnumerator : TFhirNutritionOrderOralDietNutrientListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderOralDietNutrient to the end of the list.
    }
    function Append : TFhirNutritionOrderOralDietNutrient;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderOralDietNutrient to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderOralDietNutrient); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderOralDietNutrient) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderOralDietNutrient before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderOralDietNutrient;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderOralDietNutrient before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderOralDietNutrient);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDietNutrient. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDietNutrient. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderOralDietNutrient);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderOralDietNutrient;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderOralDietNutrients[index : Integer] : TFhirNutritionOrderOralDietNutrient read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderOralDietTexture : TFhirElement
    Class that describes any texture modifications required for the patient to safely consume various types of solid foods.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDietTexture}
  TFhirNutritionOrderOralDietTexture = class (TFhirBackboneElement)
  private
    FModifier : TFhirCodeableConcept;
    FFoodType : TFhirCodeableConcept;
    Procedure SetModifier(value : TFhirCodeableConcept);
    Procedure SetFoodType(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderOralDietTexture; overload;
    function Clone : TFhirNutritionOrderOralDietTexture; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member modifier
      Typed access to Any texture modifications (for solid foods) that should be made, e.g. easy to chew, chopped, ground, and pureed. (defined for API consistency)
    }
    property modifier : TFhirCodeableConcept read FModifier write SetModifier;
    {@member modifierElement
      Any texture modifications (for solid foods) that should be made, e.g. easy to chew, chopped, ground, and pureed.
    }
    property modifierElement : TFhirCodeableConcept read FModifier write SetModifier;

    {@member foodType
      Typed access to The food type(s) (e.g. meats, all foods)  that the texture modification applies to.  This could be all foods types. (defined for API consistency)
    }
    property foodType : TFhirCodeableConcept read FFoodType write SetFoodType;
    {@member foodTypeElement
      The food type(s) (e.g. meats, all foods)  that the texture modification applies to.  This could be all foods types.
    }
    property foodTypeElement : TFhirCodeableConcept read FFoodType write SetFoodType;

  end;


  TFhirNutritionOrderOralDietTextureListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderOralDietTextureList;
    function GetCurrent : TFhirNutritionOrderOralDietTexture;
  public
    Constructor Create(list : TFhirNutritionOrderOralDietTextureList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderOralDietTexture read GetCurrent;
  end;


  {@Class TFhirNutritionOrderOralDietTextureList
    A list of FhirNutritionOrderOralDietTexture
  }
  {!.Net HL7Connect.Fhir.NutritionOrderOralDietTextureList}
  TFhirNutritionOrderOralDietTextureList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderOralDietTexture;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderOralDietTexture);
  public
    {!script hide}
    function Link : TFhirNutritionOrderOralDietTextureList; Overload;
    function Clone : TFhirNutritionOrderOralDietTextureList; Overload;
    function GetEnumerator : TFhirNutritionOrderOralDietTextureListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderOralDietTexture to the end of the list.
    }
    function Append : TFhirNutritionOrderOralDietTexture;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderOralDietTexture to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderOralDietTexture); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderOralDietTexture) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderOralDietTexture before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderOralDietTexture;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderOralDietTexture before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderOralDietTexture);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDietTexture. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderOralDietTexture. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderOralDietTexture);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderOralDietTexture;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderOralDietTextures[index : Integer] : TFhirNutritionOrderOralDietTexture read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderSupplement : TFhirElement
    Oral nutritional products given in order to add further nutritional value to the patient's diet.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderSupplement}
  TFhirNutritionOrderSupplement = class (TFhirBackboneElement)
  private
    FType_ : TFhirCodeableConcept;
    FProductName : TFhirString;
    FscheduleList : TFhirTimingList;
    FQuantity : TFhirQuantity;
    FInstruction : TFhirString;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetProductName(value : TFhirString);
    Function GetProductNameST : String;
    Procedure SetProductNameST(value : String);
    function GetScheduleList : TFhirTimingList;
    function GetHasScheduleList : Boolean;
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetInstruction(value : TFhirString);
    Function GetInstructionST : String;
    Procedure SetInstructionST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderSupplement; overload;
    function Clone : TFhirNutritionOrderSupplement; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The kind of nutritional supplement product required such as a high protein or pediatric clear liquid supplement. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The kind of nutritional supplement product required such as a high protein or pediatric clear liquid supplement.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member productName
      Typed access to The product or brand name of the nutritional supplement such as "Acme Protein Shake".
    }
    property productName : String read GetProductNameST write SetProductNameST;
    {@member productNameElement
      The product or brand name of the nutritional supplement such as "Acme Protein Shake".
    }
    property productNameElement : TFhirString read FProductName write SetProductName;

    {@member scheduleList
      The time period and frequency at which the supplement(s) should be given.
    }
    property scheduleList : TFhirTimingList read GetScheduleList;
    property hasScheduleList : boolean read GetHasScheduleList;

    {@member quantity
      Typed access to The amount of the nutritional supplement to be given. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of the nutritional supplement to be given.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member instruction
      Typed access to Free text or additional instructions or information pertaining to the oral supplement.
    }
    property instruction : String read GetInstructionST write SetInstructionST;
    {@member instructionElement
      Free text or additional instructions or information pertaining to the oral supplement.
    }
    property instructionElement : TFhirString read FInstruction write SetInstruction;

  end;


  TFhirNutritionOrderSupplementListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderSupplementList;
    function GetCurrent : TFhirNutritionOrderSupplement;
  public
    Constructor Create(list : TFhirNutritionOrderSupplementList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderSupplement read GetCurrent;
  end;


  {@Class TFhirNutritionOrderSupplementList
    A list of FhirNutritionOrderSupplement
  }
  {!.Net HL7Connect.Fhir.NutritionOrderSupplementList}
  TFhirNutritionOrderSupplementList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderSupplement;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderSupplement);
  public
    {!script hide}
    function Link : TFhirNutritionOrderSupplementList; Overload;
    function Clone : TFhirNutritionOrderSupplementList; Overload;
    function GetEnumerator : TFhirNutritionOrderSupplementListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderSupplement to the end of the list.
    }
    function Append : TFhirNutritionOrderSupplement;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderSupplement to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderSupplement); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderSupplement) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderSupplement before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderSupplement;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderSupplement before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderSupplement);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderSupplement. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderSupplement. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderSupplement);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderSupplement;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderSupplements[index : Integer] : TFhirNutritionOrderSupplement read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderEnteralFormula : TFhirElement
    Feeding provided through the gastrointestinal tract via a tube, catheter, or stoma that delivers nutrition distal to the oral cavity.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderEnteralFormula}
  TFhirNutritionOrderEnteralFormula = class (TFhirBackboneElement)
  private
    FBaseFormulaType : TFhirCodeableConcept;
    FBaseFormulaProductName : TFhirString;
    FAdditiveType : TFhirCodeableConcept;
    FAdditiveProductName : TFhirString;
    FCaloricDensity : TFhirQuantity;
    FRouteofAdministration : TFhirCodeableConcept;
    FadministrationList : TFhirNutritionOrderEnteralFormulaAdministrationList;
    FMaxVolumeToDeliver : TFhirQuantity;
    FAdministrationInstruction : TFhirString;
    Procedure SetBaseFormulaType(value : TFhirCodeableConcept);
    Procedure SetBaseFormulaProductName(value : TFhirString);
    Function GetBaseFormulaProductNameST : String;
    Procedure SetBaseFormulaProductNameST(value : String);
    Procedure SetAdditiveType(value : TFhirCodeableConcept);
    Procedure SetAdditiveProductName(value : TFhirString);
    Function GetAdditiveProductNameST : String;
    Procedure SetAdditiveProductNameST(value : String);
    Procedure SetCaloricDensity(value : TFhirQuantity);
    Procedure SetRouteofAdministration(value : TFhirCodeableConcept);
    function GetAdministrationList : TFhirNutritionOrderEnteralFormulaAdministrationList;
    function GetHasAdministrationList : Boolean;
    Procedure SetMaxVolumeToDeliver(value : TFhirQuantity);
    Procedure SetAdministrationInstruction(value : TFhirString);
    Function GetAdministrationInstructionST : String;
    Procedure SetAdministrationInstructionST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderEnteralFormula; overload;
    function Clone : TFhirNutritionOrderEnteralFormula; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member baseFormulaType
      Typed access to The type of enteral or infant formula such as an adult standard formula with fiber or a soy-based infant formula. (defined for API consistency)
    }
    property baseFormulaType : TFhirCodeableConcept read FBaseFormulaType write SetBaseFormulaType;
    {@member baseFormulaTypeElement
      The type of enteral or infant formula such as an adult standard formula with fiber or a soy-based infant formula.
    }
    property baseFormulaTypeElement : TFhirCodeableConcept read FBaseFormulaType write SetBaseFormulaType;

    {@member baseFormulaProductName
      Typed access to The product or brand name of the enteral or infant formula product such as "ACME Adult Standard Formula".
    }
    property baseFormulaProductName : String read GetBaseFormulaProductNameST write SetBaseFormulaProductNameST;
    {@member baseFormulaProductNameElement
      The product or brand name of the enteral or infant formula product such as "ACME Adult Standard Formula".
    }
    property baseFormulaProductNameElement : TFhirString read FBaseFormulaProductName write SetBaseFormulaProductName;

    {@member additiveType
      Typed access to Indicates the type of modular component such as protein, carbohydrate, fat or fiber to be provided in addition to or mixed with the base formula. (defined for API consistency)
    }
    property additiveType : TFhirCodeableConcept read FAdditiveType write SetAdditiveType;
    {@member additiveTypeElement
      Indicates the type of modular component such as protein, carbohydrate, fat or fiber to be provided in addition to or mixed with the base formula.
    }
    property additiveTypeElement : TFhirCodeableConcept read FAdditiveType write SetAdditiveType;

    {@member additiveProductName
      Typed access to The product or brand name of the type of modular component to be added to the formula.
    }
    property additiveProductName : String read GetAdditiveProductNameST write SetAdditiveProductNameST;
    {@member additiveProductNameElement
      The product or brand name of the type of modular component to be added to the formula.
    }
    property additiveProductNameElement : TFhirString read FAdditiveProductName write SetAdditiveProductName;

    {@member caloricDensity
      Typed access to The amount of energy (Calories) that the formula should provide per specified volume, typically per mL or fluid oz.  For example, an infant may require a formula that provides 24 Calories per fluid ounce or an adult may require an enteral formula that provides 1.5 Calorie/mL. (defined for API consistency)
    }
    property caloricDensity : TFhirQuantity read FCaloricDensity write SetCaloricDensity;
    {@member caloricDensityElement
      The amount of energy (Calories) that the formula should provide per specified volume, typically per mL or fluid oz.  For example, an infant may require a formula that provides 24 Calories per fluid ounce or an adult may require an enteral formula that provides 1.5 Calorie/mL.
    }
    property caloricDensityElement : TFhirQuantity read FCaloricDensity write SetCaloricDensity;

    {@member routeofAdministration
      Typed access to The route or physiological path of administration into the patient's gastrointestinal  tract for purposes of providing the formula feeding, e.g. nasogastric tube. (defined for API consistency)
    }
    property routeofAdministration : TFhirCodeableConcept read FRouteofAdministration write SetRouteofAdministration;
    {@member routeofAdministrationElement
      The route or physiological path of administration into the patient's gastrointestinal  tract for purposes of providing the formula feeding, e.g. nasogastric tube.
    }
    property routeofAdministrationElement : TFhirCodeableConcept read FRouteofAdministration write SetRouteofAdministration;

    {@member administrationList
      Formula administration instructions as structured data.  This repeating structure allows for changing the administration rate or volume over time for both bolus and continuous feeding.  An example of this would be an instruction to increase the rate of continuous feeding every 2 hours.
    }
    property administrationList : TFhirNutritionOrderEnteralFormulaAdministrationList read GetAdministrationList;
    property hasAdministrationList : boolean read GetHasAdministrationList;

    {@member maxVolumeToDeliver
      Typed access to The maximum total quantity of formula that may be administered to a subject over the period of time, e.g. 1440 mL over 24 hours. (defined for API consistency)
    }
    property maxVolumeToDeliver : TFhirQuantity read FMaxVolumeToDeliver write SetMaxVolumeToDeliver;
    {@member maxVolumeToDeliverElement
      The maximum total quantity of formula that may be administered to a subject over the period of time, e.g. 1440 mL over 24 hours.
    }
    property maxVolumeToDeliverElement : TFhirQuantity read FMaxVolumeToDeliver write SetMaxVolumeToDeliver;

    {@member administrationInstruction
      Typed access to Free text formula administration, feeding instructions or additional instructions or information.
    }
    property administrationInstruction : String read GetAdministrationInstructionST write SetAdministrationInstructionST;
    {@member administrationInstructionElement
      Free text formula administration, feeding instructions or additional instructions or information.
    }
    property administrationInstructionElement : TFhirString read FAdministrationInstruction write SetAdministrationInstruction;

  end;


  TFhirNutritionOrderEnteralFormulaListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderEnteralFormulaList;
    function GetCurrent : TFhirNutritionOrderEnteralFormula;
  public
    Constructor Create(list : TFhirNutritionOrderEnteralFormulaList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderEnteralFormula read GetCurrent;
  end;


  {@Class TFhirNutritionOrderEnteralFormulaList
    A list of FhirNutritionOrderEnteralFormula
  }
  {!.Net HL7Connect.Fhir.NutritionOrderEnteralFormulaList}
  TFhirNutritionOrderEnteralFormulaList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderEnteralFormula;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderEnteralFormula);
  public
    {!script hide}
    function Link : TFhirNutritionOrderEnteralFormulaList; Overload;
    function Clone : TFhirNutritionOrderEnteralFormulaList; Overload;
    function GetEnumerator : TFhirNutritionOrderEnteralFormulaListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderEnteralFormula to the end of the list.
    }
    function Append : TFhirNutritionOrderEnteralFormula;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderEnteralFormula to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderEnteralFormula); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderEnteralFormula) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderEnteralFormula before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderEnteralFormula;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderEnteralFormula before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderEnteralFormula);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderEnteralFormula. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderEnteralFormula. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderEnteralFormula);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderEnteralFormula;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderEnteralFormulas[index : Integer] : TFhirNutritionOrderEnteralFormula read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrderEnteralFormulaAdministration : TFhirElement
    Formula administration instructions as structured data.  This repeating structure allows for changing the administration rate or volume over time for both bolus and continuous feeding.  An example of this would be an instruction to increase the rate of continuous feeding every 2 hours.
  }
  {!.Net HL7Connect.Fhir.NutritionOrderEnteralFormulaAdministration}
  TFhirNutritionOrderEnteralFormulaAdministration = class (TFhirBackboneElement)
  private
    FSchedule : TFhirTiming;
    FQuantity : TFhirQuantity;
    FRate : TFhirType;
    Procedure SetSchedule(value : TFhirTiming);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetRate(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrderEnteralFormulaAdministration; overload;
    function Clone : TFhirNutritionOrderEnteralFormulaAdministration; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member schedule
      Typed access to The time period and frequency at which the enteral formula should be delivered to the patient. (defined for API consistency)
    }
    property schedule : TFhirTiming read FSchedule write SetSchedule;
    {@member scheduleElement
      The time period and frequency at which the enteral formula should be delivered to the patient.
    }
    property scheduleElement : TFhirTiming read FSchedule write SetSchedule;

    {@member quantity
      Typed access to The volume of formula to provide to the patient per the specified administration schedule. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The volume of formula to provide to the patient per the specified administration schedule.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member rate
      Typed access to The rate of administration of formula via a feeding pump, e.g. 60 mL per hour, according to the specified schedule. (defined for API consistency)
    }
    property rate : TFhirType read FRate write SetRate;
    {@member rateElement
      The rate of administration of formula via a feeding pump, e.g. 60 mL per hour, according to the specified schedule.
    }
    property rateElement : TFhirType read FRate write SetRate;

  end;


  TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderEnteralFormulaAdministrationList;
    function GetCurrent : TFhirNutritionOrderEnteralFormulaAdministration;
  public
    Constructor Create(list : TFhirNutritionOrderEnteralFormulaAdministrationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrderEnteralFormulaAdministration read GetCurrent;
  end;


  {@Class TFhirNutritionOrderEnteralFormulaAdministrationList
    A list of FhirNutritionOrderEnteralFormulaAdministration
  }
  {!.Net HL7Connect.Fhir.NutritionOrderEnteralFormulaAdministrationList}
  TFhirNutritionOrderEnteralFormulaAdministrationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrderEnteralFormulaAdministration;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrderEnteralFormulaAdministration);
  public
    {!script hide}
    function Link : TFhirNutritionOrderEnteralFormulaAdministrationList; Overload;
    function Clone : TFhirNutritionOrderEnteralFormulaAdministrationList; Overload;
    function GetEnumerator : TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrderEnteralFormulaAdministration to the end of the list.
    }
    function Append : TFhirNutritionOrderEnteralFormulaAdministration;

    
    {@member AddItem
      Add an already existing FhirNutritionOrderEnteralFormulaAdministration to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrderEnteralFormulaAdministration); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrderEnteralFormulaAdministration) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrderEnteralFormulaAdministration before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrderEnteralFormulaAdministration;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrderEnteralFormulaAdministration before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrderEnteralFormulaAdministration);
    
    {@member Item
       Get the iIndexth FhirNutritionOrderEnteralFormulaAdministration. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrderEnteralFormulaAdministration. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrderEnteralFormulaAdministration);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrderEnteralFormulaAdministration;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrderEnteralFormulaAdministrations[index : Integer] : TFhirNutritionOrderEnteralFormulaAdministration read GetItemN write SetItemN; default;
  End;


  {@Class TFhirNutritionOrder : TFhirDomainResource
    A request to supply a diet, formula feeding (enteral) or oral nutritional supplement to a patient/resident.
  }
  {!.Net HL7Connect.Fhir.NutritionOrder}
  TFhirNutritionOrder = class (TFhirDomainResource)
  private
    FPatient : TFhirReference{TFhirPatient};
    FOrderer : TFhirReference{TFhirPractitioner};
    FidentifierList : TFhirIdentifierList;
    FEncounter : TFhirReference{TFhirEncounter};
    FDateTime : TFhirDateTime;
    FStatus : TFhirEnum;
    FallergyIntoleranceList : TFhirReferenceList{TFhirAllergyIntolerance};
    FfoodPreferenceModifierList : TFhirCodeableConceptList;
    FexcludeFoodModifierList : TFhirCodeableConceptList;
    FOralDiet : TFhirNutritionOrderOralDiet;
    FsupplementList : TFhirNutritionOrderSupplementList;
    FEnteralFormula : TFhirNutritionOrderEnteralFormula;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetOrderer(value : TFhirReference{TFhirPractitioner});
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetDateTime(value : TFhirDateTime);
    Function GetDateTimeST : TDateAndTime;
    Procedure SetDateTimeST(value : TDateAndTime);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirNutritionOrderStatus;
    Procedure SetStatusST(value : TFhirNutritionOrderStatus);
    function GetAllergyIntoleranceList : TFhirReferenceList{TFhirAllergyIntolerance};
    function GetHasAllergyIntoleranceList : Boolean;
    function GetFoodPreferenceModifierList : TFhirCodeableConceptList;
    function GetHasFoodPreferenceModifierList : Boolean;
    function GetExcludeFoodModifierList : TFhirCodeableConceptList;
    function GetHasExcludeFoodModifierList : Boolean;
    Procedure SetOralDiet(value : TFhirNutritionOrderOralDiet);
    function GetSupplementList : TFhirNutritionOrderSupplementList;
    function GetHasSupplementList : Boolean;
    Procedure SetEnteralFormula(value : TFhirNutritionOrderEnteralFormula);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirNutritionOrder; overload;
    function Clone : TFhirNutritionOrder; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member patient
      Typed access to The person (patient) who needs the nutrition order for an oral diet, nutritional supplement and/or enteral or formula feeding. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The person (patient) who needs the nutrition order for an oral diet, nutritional supplement and/or enteral or formula feeding.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member orderer
      Typed access to The practitioner that holds legal responsibility for ordering the diet, nutritional supplement, or formula feedings. (defined for API consistency)
    }
    property orderer : TFhirReference{TFhirPractitioner} read FOrderer write SetOrderer;
    {@member ordererElement
      The practitioner that holds legal responsibility for ordering the diet, nutritional supplement, or formula feedings.
    }
    property ordererElement : TFhirReference{TFhirPractitioner} read FOrderer write SetOrderer;

    {@member identifierList
      Identifiers assigned to this order by the order sender or by the order receiver.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member encounter
      Typed access to An encounter that provides additional information about the healthcare context in which this request is made. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      An encounter that provides additional information about the healthcare context in which this request is made.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member dateTime
      Typed access to The date and time that this nutrition order was requested.
    }
    property dateTime : TDateAndTime read GetDateTimeST write SetDateTimeST;
    {@member dateTimeElement
      The date and time that this nutrition order was requested.
    }
    property dateTimeElement : TFhirDateTime read FDateTime write SetDateTime;

    {@member status
      The workflow status of the nutrition order/request.
    }
    property status : TFhirNutritionOrderStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member allergyIntoleranceList
      A link to a record of allergies or intolerances  which should be included in the nutrition order.
    }
    property allergyIntoleranceList : TFhirReferenceList{TFhirAllergyIntolerance} read GetAllergyIntoleranceList;
    property hasAllergyIntoleranceList : boolean read GetHasAllergyIntoleranceList;

    {@member foodPreferenceModifierList
      This modifier is used to convey order-specific modifiers about the type of food that should be given. These can be derived from patient allergies, intolerances, or preferences such as Halal, Vegan or Kosher. This modifier applies to the entire nutrition order inclusive of the oral diet, nutritional supplements and enteral formula feedings.
    }
    property foodPreferenceModifierList : TFhirCodeableConceptList read GetFoodPreferenceModifierList;
    property hasFoodPreferenceModifierList : boolean read GetHasFoodPreferenceModifierList;

    {@member excludeFoodModifierList
      This modifier is used to convey order-specific modifiers about the type of food that should NOT be given. These can be derived from patient allergies, intolerances, or preferences such as No Red Meat, No Soy or No Wheat or  Gluten-Free.  While it should not be necessary to repeat allergy or intolerance information captured in the referenced allergyIntolerance resource in the excludeFoodModifier, this element may be used to convey additional specificity related to foods that should be eliminated from the patient?s diet for any reason.  This modifier applies to the entire nutrition order inclusive of the oral diet, nutritional supplements and enteral formula feedings.
    }
    property excludeFoodModifierList : TFhirCodeableConceptList read GetExcludeFoodModifierList;
    property hasExcludeFoodModifierList : boolean read GetHasExcludeFoodModifierList;

    {@member oralDiet
      Typed access to Diet given orally in contrast to enteral (tube) feeding. (defined for API consistency)
    }
    property oralDiet : TFhirNutritionOrderOralDiet read FOralDiet write SetOralDiet;
    {@member oralDietElement
      Diet given orally in contrast to enteral (tube) feeding.
    }
    property oralDietElement : TFhirNutritionOrderOralDiet read FOralDiet write SetOralDiet;

    {@member supplementList
      Oral nutritional products given in order to add further nutritional value to the patient's diet.
    }
    property supplementList : TFhirNutritionOrderSupplementList read GetSupplementList;
    property hasSupplementList : boolean read GetHasSupplementList;

    {@member enteralFormula
      Typed access to Feeding provided through the gastrointestinal tract via a tube, catheter, or stoma that delivers nutrition distal to the oral cavity. (defined for API consistency)
    }
    property enteralFormula : TFhirNutritionOrderEnteralFormula read FEnteralFormula write SetEnteralFormula;
    {@member enteralFormulaElement
      Feeding provided through the gastrointestinal tract via a tube, catheter, or stoma that delivers nutrition distal to the oral cavity.
    }
    property enteralFormulaElement : TFhirNutritionOrderEnteralFormula read FEnteralFormula write SetEnteralFormula;

  end;


  TFhirNutritionOrderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirNutritionOrderList;
    function GetCurrent : TFhirNutritionOrder;
  public
    Constructor Create(list : TFhirNutritionOrderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirNutritionOrder read GetCurrent;
  end;


  {@Class TFhirNutritionOrderList
    A list of FhirNutritionOrder
  }
  {!.Net HL7Connect.Fhir.NutritionOrderList}
  TFhirNutritionOrderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirNutritionOrder;
    procedure SetItemN(index : Integer; value : TFhirNutritionOrder);
  public
    {!script hide}
    function Link : TFhirNutritionOrderList; Overload;
    function Clone : TFhirNutritionOrderList; Overload;
    function GetEnumerator : TFhirNutritionOrderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirNutritionOrder to the end of the list.
    }
    function Append : TFhirNutritionOrder;

    
    {@member AddItem
      Add an already existing FhirNutritionOrder to the end of the list.
    }
    procedure AddItem(value : TFhirNutritionOrder); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirNutritionOrder) : Integer;
    

    {@member Insert
      Insert FhirNutritionOrder before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirNutritionOrder;
    

    {@member InsertItem
       Insert an existing FhirNutritionOrder before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirNutritionOrder);
    
    {@member Item
       Get the iIndexth FhirNutritionOrder. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirNutritionOrder. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirNutritionOrder);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirNutritionOrder;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirNutritionOrders[index : Integer] : TFhirNutritionOrder read GetItemN write SetItemN; default;
  End;


  {@Class TFhirObservationReferenceRange : TFhirElement
    Guidance on how to interpret the value by comparison to a normal or recommended range.
  }
  {!.Net HL7Connect.Fhir.ObservationReferenceRange}
  TFhirObservationReferenceRange = class (TFhirBackboneElement)
  private
    FLow : TFhirQuantity;
    FHigh : TFhirQuantity;
    FMeaning : TFhirCodeableConcept;
    FAge : TFhirRange;
    FText : TFhirString;
    Procedure SetLow(value : TFhirQuantity);
    Procedure SetHigh(value : TFhirQuantity);
    Procedure SetMeaning(value : TFhirCodeableConcept);
    Procedure SetAge(value : TFhirRange);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirObservationReferenceRange; overload;
    function Clone : TFhirObservationReferenceRange; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member low
      Typed access to The value of the low bound of the reference range.  The low bound of the reference range endpoint is inclusive of the value (e.g.  reference range is >=5 - <=9).   If the low bound is omitted,  it is assumed to be meaningless (e.g. reference range is <=2.3). (defined for API consistency)
    }
    property low : TFhirQuantity read FLow write SetLow;
    {@member lowElement
      The value of the low bound of the reference range.  The low bound of the reference range endpoint is inclusive of the value (e.g.  reference range is >=5 - <=9).   If the low bound is omitted,  it is assumed to be meaningless (e.g. reference range is <=2.3).
    }
    property lowElement : TFhirQuantity read FLow write SetLow;

    {@member high
      Typed access to The value of the high bound of the reference range.  The high bound of the reference range endpoint is inclusive of the value (e.g.  reference range is >=5 - <=9).   If the high bound is omitted,  it is assumed to be meaningless (e.g. reference range is >= 2.3). (defined for API consistency)
    }
    property high : TFhirQuantity read FHigh write SetHigh;
    {@member highElement
      The value of the high bound of the reference range.  The high bound of the reference range endpoint is inclusive of the value (e.g.  reference range is >=5 - <=9).   If the high bound is omitted,  it is assumed to be meaningless (e.g. reference range is >= 2.3).
    }
    property highElement : TFhirQuantity read FHigh write SetHigh;

    {@member meaning
      Typed access to Code for the meaning of the reference range. (defined for API consistency)
    }
    property meaning : TFhirCodeableConcept read FMeaning write SetMeaning;
    {@member meaningElement
      Code for the meaning of the reference range.
    }
    property meaningElement : TFhirCodeableConcept read FMeaning write SetMeaning;

    {@member age
      Typed access to The age at which this reference range is applicable. This is a neonatal age (e.g. number of weeks at term) if the meaning says so. (defined for API consistency)
    }
    property age : TFhirRange read FAge write SetAge;
    {@member ageElement
      The age at which this reference range is applicable. This is a neonatal age (e.g. number of weeks at term) if the meaning says so.
    }
    property ageElement : TFhirRange read FAge write SetAge;

    {@member text
      Typed access to Text based reference range in an observation which may be used when a quantitative range is not appropriate for an observation.  An example would be a reference value of "Negative" or a list or table of 'normals'.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Text based reference range in an observation which may be used when a quantitative range is not appropriate for an observation.  An example would be a reference value of "Negative" or a list or table of 'normals'.
    }
    property textElement : TFhirString read FText write SetText;

  end;


  TFhirObservationReferenceRangeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirObservationReferenceRangeList;
    function GetCurrent : TFhirObservationReferenceRange;
  public
    Constructor Create(list : TFhirObservationReferenceRangeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirObservationReferenceRange read GetCurrent;
  end;


  {@Class TFhirObservationReferenceRangeList
    A list of FhirObservationReferenceRange
  }
  {!.Net HL7Connect.Fhir.ObservationReferenceRangeList}
  TFhirObservationReferenceRangeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirObservationReferenceRange;
    procedure SetItemN(index : Integer; value : TFhirObservationReferenceRange);
  public
    {!script hide}
    function Link : TFhirObservationReferenceRangeList; Overload;
    function Clone : TFhirObservationReferenceRangeList; Overload;
    function GetEnumerator : TFhirObservationReferenceRangeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirObservationReferenceRange to the end of the list.
    }
    function Append : TFhirObservationReferenceRange;

    
    {@member AddItem
      Add an already existing FhirObservationReferenceRange to the end of the list.
    }
    procedure AddItem(value : TFhirObservationReferenceRange); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirObservationReferenceRange) : Integer;
    

    {@member Insert
      Insert FhirObservationReferenceRange before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirObservationReferenceRange;
    

    {@member InsertItem
       Insert an existing FhirObservationReferenceRange before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirObservationReferenceRange);
    
    {@member Item
       Get the iIndexth FhirObservationReferenceRange. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirObservationReferenceRange. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirObservationReferenceRange);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirObservationReferenceRange;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirObservationReferenceRanges[index : Integer] : TFhirObservationReferenceRange read GetItemN write SetItemN; default;
  End;


  {@Class TFhirObservationRelated : TFhirElement
    A  reference to another resource (usually another Observation but could  also be a QuestionnaireAnswer) whose relationship is defined by the relationship type code.
  }
  {!.Net HL7Connect.Fhir.ObservationRelated}
  TFhirObservationRelated = class (TFhirBackboneElement)
  private
    FType_ : TFhirEnum;
    FTarget : TFhirReference{Resource};
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirObservationRelationshiptypes;
    Procedure SetType_ST(value : TFhirObservationRelationshiptypes);
    Procedure SetTarget(value : TFhirReference{Resource});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirObservationRelated; overload;
    function Clone : TFhirObservationRelated; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      A code specifying the kind of relationship that exists with the target resource.
    }
    property type_ : TFhirObservationRelationshiptypes read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member target
      Typed access to A reference to the observation or [[[QuestionnaireResponse]]] resource that is related to this observation. (defined for API consistency)
    }
    property target : TFhirReference{Resource} read FTarget write SetTarget;
    {@member targetElement
      A reference to the observation or [[[QuestionnaireResponse]]] resource that is related to this observation.
    }
    property targetElement : TFhirReference{Resource} read FTarget write SetTarget;

  end;


  TFhirObservationRelatedListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirObservationRelatedList;
    function GetCurrent : TFhirObservationRelated;
  public
    Constructor Create(list : TFhirObservationRelatedList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirObservationRelated read GetCurrent;
  end;


  {@Class TFhirObservationRelatedList
    A list of FhirObservationRelated
  }
  {!.Net HL7Connect.Fhir.ObservationRelatedList}
  TFhirObservationRelatedList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirObservationRelated;
    procedure SetItemN(index : Integer; value : TFhirObservationRelated);
  public
    {!script hide}
    function Link : TFhirObservationRelatedList; Overload;
    function Clone : TFhirObservationRelatedList; Overload;
    function GetEnumerator : TFhirObservationRelatedListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirObservationRelated to the end of the list.
    }
    function Append : TFhirObservationRelated;

    
    {@member AddItem
      Add an already existing FhirObservationRelated to the end of the list.
    }
    procedure AddItem(value : TFhirObservationRelated); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirObservationRelated) : Integer;
    

    {@member Insert
      Insert FhirObservationRelated before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirObservationRelated;
    

    {@member InsertItem
       Insert an existing FhirObservationRelated before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirObservationRelated);
    
    {@member Item
       Get the iIndexth FhirObservationRelated. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirObservationRelated. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirObservationRelated);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirObservationRelated;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirObservationRelateds[index : Integer] : TFhirObservationRelated read GetItemN write SetItemN; default;
  End;


  {@Class TFhirObservationComponent : TFhirElement
    Some observations have multiple component observations.  These component observations are expressed as separate code value pairs that share the same attributes.  Examples include systolic and diastolic component observations for blood pressure measurement and multiple component observations for genetics observations.
  }
  {!.Net HL7Connect.Fhir.ObservationComponent}
  TFhirObservationComponent = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FValue : TFhirType;
    FDataAbsentReason : TFhirCodeableConcept;
    FreferenceRangeList : TFhirObservationReferenceRangeList;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetValue(value : TFhirType);
    Procedure SetDataAbsentReason(value : TFhirCodeableConcept);
    function GetReferenceRangeList : TFhirObservationReferenceRangeList;
    function GetHasReferenceRangeList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirObservationComponent; overload;
    function Clone : TFhirObservationComponent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Describes what was observed. Sometimes this is called the observation "code". (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Describes what was observed. Sometimes this is called the observation "code".
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member value
      Typed access to The information determined as a result of making the observation, if the information has a simple value. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      The information determined as a result of making the observation, if the information has a simple value.
    }
    property valueElement : TFhirType read FValue write SetValue;

    {@member dataAbsentReason
      Typed access to Provides a reason why the expected value in the element Observation.value[x] is missing. (defined for API consistency)
    }
    property dataAbsentReason : TFhirCodeableConcept read FDataAbsentReason write SetDataAbsentReason;
    {@member dataAbsentReasonElement
      Provides a reason why the expected value in the element Observation.value[x] is missing.
    }
    property dataAbsentReasonElement : TFhirCodeableConcept read FDataAbsentReason write SetDataAbsentReason;

    {@member referenceRangeList
      Guidance on how to interpret the value by comparison to a normal or recommended range.
    }
    property referenceRangeList : TFhirObservationReferenceRangeList read GetReferenceRangeList;
    property hasReferenceRangeList : boolean read GetHasReferenceRangeList;

  end;


  TFhirObservationComponentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirObservationComponentList;
    function GetCurrent : TFhirObservationComponent;
  public
    Constructor Create(list : TFhirObservationComponentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirObservationComponent read GetCurrent;
  end;


  {@Class TFhirObservationComponentList
    A list of FhirObservationComponent
  }
  {!.Net HL7Connect.Fhir.ObservationComponentList}
  TFhirObservationComponentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirObservationComponent;
    procedure SetItemN(index : Integer; value : TFhirObservationComponent);
  public
    {!script hide}
    function Link : TFhirObservationComponentList; Overload;
    function Clone : TFhirObservationComponentList; Overload;
    function GetEnumerator : TFhirObservationComponentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirObservationComponent to the end of the list.
    }
    function Append : TFhirObservationComponent;

    
    {@member AddItem
      Add an already existing FhirObservationComponent to the end of the list.
    }
    procedure AddItem(value : TFhirObservationComponent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirObservationComponent) : Integer;
    

    {@member Insert
      Insert FhirObservationComponent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirObservationComponent;
    

    {@member InsertItem
       Insert an existing FhirObservationComponent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirObservationComponent);
    
    {@member Item
       Get the iIndexth FhirObservationComponent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirObservationComponent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirObservationComponent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirObservationComponent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirObservationComponents[index : Integer] : TFhirObservationComponent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirObservation : TFhirDomainResource
    Measurements and simple assertions made about a patient, device or other subject.
  }
  {!.Net HL7Connect.Fhir.Observation}
  TFhirObservation = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FCategory : TFhirCodeableConcept;
    FCode : TFhirCodeableConcept;
    FSubject : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FEffective : TFhirType;
    FIssued : TFhirInstant;
    FperformerList : TFhirReferenceList{Resource};
    FValue : TFhirType;
    FDataAbsentReason : TFhirCodeableConcept;
    FInterpretation : TFhirCodeableConcept;
    FComments : TFhirString;
    FBodySite : TFhirCodeableConcept;
    FMethod : TFhirCodeableConcept;
    FSpecimen : TFhirReference{TFhirSpecimen};
    FDevice : TFhirReference{Resource};
    FreferenceRangeList : TFhirObservationReferenceRangeList;
    FrelatedList : TFhirObservationRelatedList;
    FcomponentList : TFhirObservationComponentList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirObservationStatus;
    Procedure SetStatusST(value : TFhirObservationStatus);
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetEffective(value : TFhirType);
    Procedure SetIssued(value : TFhirInstant);
    Function GetIssuedST : TDateAndTime;
    Procedure SetIssuedST(value : TDateAndTime);
    function GetPerformerList : TFhirReferenceList{Resource};
    function GetHasPerformerList : Boolean;
    Procedure SetValue(value : TFhirType);
    Procedure SetDataAbsentReason(value : TFhirCodeableConcept);
    Procedure SetInterpretation(value : TFhirCodeableConcept);
    Procedure SetComments(value : TFhirString);
    Function GetCommentsST : String;
    Procedure SetCommentsST(value : String);
    Procedure SetBodySite(value : TFhirCodeableConcept);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetSpecimen(value : TFhirReference{TFhirSpecimen});
    Procedure SetDevice(value : TFhirReference{Resource});
    function GetReferenceRangeList : TFhirObservationReferenceRangeList;
    function GetHasReferenceRangeList : Boolean;
    function GetRelatedList : TFhirObservationRelatedList;
    function GetHasRelatedList : Boolean;
    function GetComponentList : TFhirObservationComponentList;
    function GetHasComponentList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirObservation; overload;
    function Clone : TFhirObservation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      A unique identifier for the simple observation instance.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      The status of the result value.
    }
    property status : TFhirObservationStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member category
      Typed access to A code that classifies the general type of observation being made.  This is used  for searching, sorting and display purposes. (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      A code that classifies the general type of observation being made.  This is used  for searching, sorting and display purposes.
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member code
      Typed access to Describes what was observed. Sometimes this is called the observation "name". (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Describes what was observed. Sometimes this is called the observation "name".
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member subject
      Typed access to The patient, or group of patients, location, or device whose characteristics (direct or indirect) are described by the observation and into whose record the observation is placed.  Comments: Indirect characteristics may be those of a specimen, fetus, donor,  other observer (for example a relative or EMT), or any observation made about the subject. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The patient, or group of patients, location, or device whose characteristics (direct or indirect) are described by the observation and into whose record the observation is placed.  Comments: Indirect characteristics may be those of a specimen, fetus, donor,  other observer (for example a relative or EMT), or any observation made about the subject.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member encounter
      Typed access to The healthcare event  (e.g. a patient and healthcare provider interaction) during which this observation is made. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The healthcare event  (e.g. a patient and healthcare provider interaction) during which this observation is made.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member effective
      Typed access to The time or time-period the observed value is asserted as being true. For biological subjects - e.g. human patients - this is usually called the "physiologically relevant time". This is usually either the time of the procedure or of specimen collection, but very often the source of the date/time is not known, only the date/time itself. (defined for API consistency)
    }
    property effective : TFhirType read FEffective write SetEffective;
    {@member effectiveElement
      The time or time-period the observed value is asserted as being true. For biological subjects - e.g. human patients - this is usually called the "physiologically relevant time". This is usually either the time of the procedure or of specimen collection, but very often the source of the date/time is not known, only the date/time itself.
    }
    property effectiveElement : TFhirType read FEffective write SetEffective;

    {@member issued
      Typed access to The date and time this observation was made available to providers, typically after the results have been reviewed and verified.
    }
    property issued : TDateAndTime read GetIssuedST write SetIssuedST;
    {@member issuedElement
      The date and time this observation was made available to providers, typically after the results have been reviewed and verified.
    }
    property issuedElement : TFhirInstant read FIssued write SetIssued;

    {@member performerList
      Who was responsible for asserting the observed value as "true".
    }
    property performerList : TFhirReferenceList{Resource} read GetPerformerList;
    property hasPerformerList : boolean read GetHasPerformerList;

    {@member value
      Typed access to The information determined as a result of making the observation, if the information has a simple value. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      The information determined as a result of making the observation, if the information has a simple value.
    }
    property valueElement : TFhirType read FValue write SetValue;

    {@member dataAbsentReason
      Typed access to Provides a reason why the expected value in the element Observation.value[x] is missing. (defined for API consistency)
    }
    property dataAbsentReason : TFhirCodeableConcept read FDataAbsentReason write SetDataAbsentReason;
    {@member dataAbsentReasonElement
      Provides a reason why the expected value in the element Observation.value[x] is missing.
    }
    property dataAbsentReasonElement : TFhirCodeableConcept read FDataAbsentReason write SetDataAbsentReason;

    {@member interpretation
      Typed access to The assessment made based on the result of the observation.  Intended as a simple compact code often placed adjacent to the result value in reports and flow sheets to signal the meaning/normalcy status of the result. Otherwise known as abnormal flag. (defined for API consistency)
    }
    property interpretation : TFhirCodeableConcept read FInterpretation write SetInterpretation;
    {@member interpretationElement
      The assessment made based on the result of the observation.  Intended as a simple compact code often placed adjacent to the result value in reports and flow sheets to signal the meaning/normalcy status of the result. Otherwise known as abnormal flag.
    }
    property interpretationElement : TFhirCodeableConcept read FInterpretation write SetInterpretation;

    {@member comments
      Typed access to May include statements about significant, unexpected or unreliable values, or information about the source of the value where this may be relevant to the interpretation of the result.
    }
    property comments : String read GetCommentsST write SetCommentsST;
    {@member commentsElement
      May include statements about significant, unexpected or unreliable values, or information about the source of the value where this may be relevant to the interpretation of the result.
    }
    property commentsElement : TFhirString read FComments write SetComments;

    {@member bodySite
      Typed access to Indicates the site on the subject's body where the observation was made (i.e. the target site). (defined for API consistency)
    }
    property bodySite : TFhirCodeableConcept read FBodySite write SetBodySite;
    {@member bodySiteElement
      Indicates the site on the subject's body where the observation was made (i.e. the target site).
    }
    property bodySiteElement : TFhirCodeableConcept read FBodySite write SetBodySite;

    {@member method
      Typed access to Indicates the mechanism used to perform the observation. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      Indicates the mechanism used to perform the observation.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member specimen
      Typed access to The specimen that was used when this observation was made. (defined for API consistency)
    }
    property specimen : TFhirReference{TFhirSpecimen} read FSpecimen write SetSpecimen;
    {@member specimenElement
      The specimen that was used when this observation was made.
    }
    property specimenElement : TFhirReference{TFhirSpecimen} read FSpecimen write SetSpecimen;

    {@member device
      Typed access to The device used to generate the observation data. (defined for API consistency)
    }
    property device : TFhirReference{Resource} read FDevice write SetDevice;
    {@member deviceElement
      The device used to generate the observation data.
    }
    property deviceElement : TFhirReference{Resource} read FDevice write SetDevice;

    {@member referenceRangeList
      Guidance on how to interpret the value by comparison to a normal or recommended range.
    }
    property referenceRangeList : TFhirObservationReferenceRangeList read GetReferenceRangeList;
    property hasReferenceRangeList : boolean read GetHasReferenceRangeList;

    {@member relatedList
      A  reference to another resource (usually another Observation but could  also be a QuestionnaireAnswer) whose relationship is defined by the relationship type code.
    }
    property relatedList : TFhirObservationRelatedList read GetRelatedList;
    property hasRelatedList : boolean read GetHasRelatedList;

    {@member componentList
      Some observations have multiple component observations.  These component observations are expressed as separate code value pairs that share the same attributes.  Examples include systolic and diastolic component observations for blood pressure measurement and multiple component observations for genetics observations.
    }
    property componentList : TFhirObservationComponentList read GetComponentList;
    property hasComponentList : boolean read GetHasComponentList;

  end;


  TFhirObservationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirObservationList;
    function GetCurrent : TFhirObservation;
  public
    Constructor Create(list : TFhirObservationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirObservation read GetCurrent;
  end;


  {@Class TFhirObservationList
    A list of FhirObservation
  }
  {!.Net HL7Connect.Fhir.ObservationList}
  TFhirObservationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirObservation;
    procedure SetItemN(index : Integer; value : TFhirObservation);
  public
    {!script hide}
    function Link : TFhirObservationList; Overload;
    function Clone : TFhirObservationList; Overload;
    function GetEnumerator : TFhirObservationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirObservation to the end of the list.
    }
    function Append : TFhirObservation;

    
    {@member AddItem
      Add an already existing FhirObservation to the end of the list.
    }
    procedure AddItem(value : TFhirObservation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirObservation) : Integer;
    

    {@member Insert
      Insert FhirObservation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirObservation;
    

    {@member InsertItem
       Insert an existing FhirObservation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirObservation);
    
    {@member Item
       Get the iIndexth FhirObservation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirObservation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirObservation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirObservation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirObservations[index : Integer] : TFhirObservation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationDefinitionContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionContact}
  TFhirOperationDefinitionContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationDefinitionContact; overload;
    function Clone : TFhirOperationDefinitionContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the operation definition.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the operation definition.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirOperationDefinitionContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationDefinitionContactList;
    function GetCurrent : TFhirOperationDefinitionContact;
  public
    Constructor Create(list : TFhirOperationDefinitionContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationDefinitionContact read GetCurrent;
  end;


  {@Class TFhirOperationDefinitionContactList
    A list of FhirOperationDefinitionContact
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionContactList}
  TFhirOperationDefinitionContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationDefinitionContact;
    procedure SetItemN(index : Integer; value : TFhirOperationDefinitionContact);
  public
    {!script hide}
    function Link : TFhirOperationDefinitionContactList; Overload;
    function Clone : TFhirOperationDefinitionContactList; Overload;
    function GetEnumerator : TFhirOperationDefinitionContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationDefinitionContact to the end of the list.
    }
    function Append : TFhirOperationDefinitionContact;

    
    {@member AddItem
      Add an already existing FhirOperationDefinitionContact to the end of the list.
    }
    procedure AddItem(value : TFhirOperationDefinitionContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationDefinitionContact) : Integer;
    

    {@member Insert
      Insert FhirOperationDefinitionContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationDefinitionContact;
    

    {@member InsertItem
       Insert an existing FhirOperationDefinitionContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationDefinitionContact);
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationDefinitionContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationDefinitionContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationDefinitionContacts[index : Integer] : TFhirOperationDefinitionContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationDefinitionParameter : TFhirElement
    The parameters for the operation/query.
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionParameter}
  TFhirOperationDefinitionParameter = class (TFhirBackboneElement)
  private
    FName : TFhirCode;
    FUse : TFhirEnum;
    FMin : TFhirInteger;
    FMax : TFhirString;
    FDocumentation : TFhirString;
    FType_ : TFhirEnum;
    FProfile : TFhirReference{TFhirStructureDefinition};
    FBinding : TFhirOperationDefinitionParameterBinding;
    FpartList : TFhirOperationDefinitionParameterList;
    Procedure SetName(value : TFhirCode);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetUse(value : TFhirEnum);
    Function GetUseST : TFhirOperationParameterUse;
    Procedure SetUseST(value : TFhirOperationParameterUse);
    Procedure SetMin(value : TFhirInteger);
    Function GetMinST : String;
    Procedure SetMinST(value : String);
    Procedure SetMax(value : TFhirString);
    Function GetMaxST : String;
    Procedure SetMaxST(value : String);
    Procedure SetDocumentation(value : TFhirString);
    Function GetDocumentationST : String;
    Procedure SetDocumentationST(value : String);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirValuesetOperationParameterType;
    Procedure SetType_ST(value : TFhirValuesetOperationParameterType);
    Procedure SetProfile(value : TFhirReference{TFhirStructureDefinition});
    Procedure SetBinding(value : TFhirOperationDefinitionParameterBinding);
    function GetPartList : TFhirOperationDefinitionParameterList;
    function GetHasPartList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationDefinitionParameter; overload;
    function Clone : TFhirOperationDefinitionParameter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of used to identify the parameter.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of used to identify the parameter.
    }
    property nameElement : TFhirCode read FName write SetName;

    {@member use
      Whether this is an input or an output parameter.
    }
    property use : TFhirOperationParameterUse read GetUseST write SetUseST;
    property useElement : TFhirEnum read FUse write SetUse;

    {@member min
      Typed access to The minimum number of times this parameter SHALL appear in the request or response.
    }
    property min : String read GetMinST write SetMinST;
    {@member minElement
      The minimum number of times this parameter SHALL appear in the request or response.
    }
    property minElement : TFhirInteger read FMin write SetMin;

    {@member max
      Typed access to The maximum number of times this element is permitted to appear in the request or response.
    }
    property max : String read GetMaxST write SetMaxST;
    {@member maxElement
      The maximum number of times this element is permitted to appear in the request or response.
    }
    property maxElement : TFhirString read FMax write SetMax;

    {@member documentation
      Typed access to Describes the meaning or use of this parameter.
    }
    property documentation : String read GetDocumentationST write SetDocumentationST;
    {@member documentationElement
      Describes the meaning or use of this parameter.
    }
    property documentationElement : TFhirString read FDocumentation write SetDocumentation;

    {@member type_
      The type for this parameter.
    }
    property type_ : TFhirValuesetOperationParameterType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member profile
      Typed access to A profile the specifies the rules that this parameter must conform to. (defined for API consistency)
    }
    property profile : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;
    {@member profileElement
      A profile the specifies the rules that this parameter must conform to.
    }
    property profileElement : TFhirReference{TFhirStructureDefinition} read FProfile write SetProfile;

    {@member binding
      Typed access to Binds to a value set if this parameter is coded (code, Coding, CodeableConcept). (defined for API consistency)
    }
    property binding : TFhirOperationDefinitionParameterBinding read FBinding write SetBinding;
    {@member bindingElement
      Binds to a value set if this parameter is coded (code, Coding, CodeableConcept).
    }
    property bindingElement : TFhirOperationDefinitionParameterBinding read FBinding write SetBinding;

    {@member partList
      The parts of a Tuple Parameter.
    }
    property partList : TFhirOperationDefinitionParameterList read GetPartList;
    property hasPartList : boolean read GetHasPartList;

  end;


  TFhirOperationDefinitionParameterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationDefinitionParameterList;
    function GetCurrent : TFhirOperationDefinitionParameter;
  public
    Constructor Create(list : TFhirOperationDefinitionParameterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationDefinitionParameter read GetCurrent;
  end;


  {@Class TFhirOperationDefinitionParameterList
    A list of FhirOperationDefinitionParameter
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionParameterList}
  TFhirOperationDefinitionParameterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationDefinitionParameter;
    procedure SetItemN(index : Integer; value : TFhirOperationDefinitionParameter);
  public
    {!script hide}
    function Link : TFhirOperationDefinitionParameterList; Overload;
    function Clone : TFhirOperationDefinitionParameterList; Overload;
    function GetEnumerator : TFhirOperationDefinitionParameterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationDefinitionParameter to the end of the list.
    }
    function Append : TFhirOperationDefinitionParameter;

    
    {@member AddItem
      Add an already existing FhirOperationDefinitionParameter to the end of the list.
    }
    procedure AddItem(value : TFhirOperationDefinitionParameter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationDefinitionParameter) : Integer;
    

    {@member Insert
      Insert FhirOperationDefinitionParameter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationDefinitionParameter;
    

    {@member InsertItem
       Insert an existing FhirOperationDefinitionParameter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationDefinitionParameter);
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionParameter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionParameter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationDefinitionParameter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationDefinitionParameter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationDefinitionParameters[index : Integer] : TFhirOperationDefinitionParameter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationDefinitionParameterBinding : TFhirElement
    Binds to a value set if this parameter is coded (code, Coding, CodeableConcept).
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionParameterBinding}
  TFhirOperationDefinitionParameterBinding = class (TFhirBackboneElement)
  private
    FStrength : TFhirEnum;
    FValueSet : TFhirType;
    Procedure SetStrength(value : TFhirEnum);
    Function GetStrengthST : TFhirBindingStrength;
    Procedure SetStrengthST(value : TFhirBindingStrength);
    Procedure SetValueSet(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationDefinitionParameterBinding; overload;
    function Clone : TFhirOperationDefinitionParameterBinding; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member strength
      Indicates the degree of conformance expectations associated with this binding - that is, the degree to which the provided value set must be adhered to in the instances.
    }
    property strength : TFhirBindingStrength read GetStrengthST write SetStrengthST;
    property strengthElement : TFhirEnum read FStrength write SetStrength;

    {@member valueSet
      Typed access to Points to the value set or external definition (e.g. implicit value set) that identifies the set of codes to be used. (defined for API consistency)
    }
    property valueSet : TFhirType read FValueSet write SetValueSet;
    {@member valueSetElement
      Points to the value set or external definition (e.g. implicit value set) that identifies the set of codes to be used.
    }
    property valueSetElement : TFhirType read FValueSet write SetValueSet;

  end;


  TFhirOperationDefinitionParameterBindingListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationDefinitionParameterBindingList;
    function GetCurrent : TFhirOperationDefinitionParameterBinding;
  public
    Constructor Create(list : TFhirOperationDefinitionParameterBindingList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationDefinitionParameterBinding read GetCurrent;
  end;


  {@Class TFhirOperationDefinitionParameterBindingList
    A list of FhirOperationDefinitionParameterBinding
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionParameterBindingList}
  TFhirOperationDefinitionParameterBindingList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationDefinitionParameterBinding;
    procedure SetItemN(index : Integer; value : TFhirOperationDefinitionParameterBinding);
  public
    {!script hide}
    function Link : TFhirOperationDefinitionParameterBindingList; Overload;
    function Clone : TFhirOperationDefinitionParameterBindingList; Overload;
    function GetEnumerator : TFhirOperationDefinitionParameterBindingListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationDefinitionParameterBinding to the end of the list.
    }
    function Append : TFhirOperationDefinitionParameterBinding;

    
    {@member AddItem
      Add an already existing FhirOperationDefinitionParameterBinding to the end of the list.
    }
    procedure AddItem(value : TFhirOperationDefinitionParameterBinding); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationDefinitionParameterBinding) : Integer;
    

    {@member Insert
      Insert FhirOperationDefinitionParameterBinding before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationDefinitionParameterBinding;
    

    {@member InsertItem
       Insert an existing FhirOperationDefinitionParameterBinding before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationDefinitionParameterBinding);
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionParameterBinding. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationDefinitionParameterBinding. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationDefinitionParameterBinding);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationDefinitionParameterBinding;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationDefinitionParameterBindings[index : Integer] : TFhirOperationDefinitionParameterBinding read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationDefinition : TFhirDomainResource
    A formal computable definition of an operation (on the RESTful interface) or a named query (using the search interaction).
  }
  {!.Net HL7Connect.Fhir.OperationDefinition}
  TFhirOperationDefinition = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FKind : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirOperationDefinitionContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FRequirements : TFhirString;
    FIdempotent : TFhirBoolean;
    FCode : TFhirCode;
    FNotes : TFhirString;
    FBase : TFhirReference{TFhirOperationDefinition};
    FSystem : TFhirBoolean;
    Ftype_List : TFhirCodeList;
    FInstance : TFhirBoolean;
    FparameterList : TFhirOperationDefinitionParameterList;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetKind(value : TFhirEnum);
    Function GetKindST : TFhirOperationKind;
    Procedure SetKindST(value : TFhirOperationKind);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirOperationDefinitionContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetIdempotent(value : TFhirBoolean);
    Function GetIdempotentST : Boolean;
    Procedure SetIdempotentST(value : Boolean);
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetNotes(value : TFhirString);
    Function GetNotesST : String;
    Procedure SetNotesST(value : String);
    Procedure SetBase(value : TFhirReference{TFhirOperationDefinition});
    Procedure SetSystem(value : TFhirBoolean);
    Function GetSystemST : Boolean;
    Procedure SetSystemST(value : Boolean);
    function GetType_List : TFhirCodeList;
    function GetHasType_List : Boolean;
    Procedure SetInstance(value : TFhirBoolean);
    Function GetInstanceST : Boolean;
    Procedure SetInstanceST(value : Boolean);
    function GetParameterList : TFhirOperationDefinitionParameterList;
    function GetHasParameterList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationDefinition; overload;
    function Clone : TFhirOperationDefinition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this operation definition when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this operation definition is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this operation definition when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this operation definition is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member version
      Typed access to The identifier that is used to identify this version of the profile when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the profile when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name identifying the operation.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the operation.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the profile.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member kind
      Whether this is an operation or a named query.
    }
    property kind : TFhirOperationKind read GetKindST write SetKindST;
    property kindElement : TFhirEnum read FKind write SetKind;

    {@member experimental
      Typed access to This profile was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This profile was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the operation definition.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the operation definition.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirOperationDefinitionContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the operation definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the Operation Definition changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the operation definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the Operation Definition changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the profile and its use.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the profile and its use.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member requirements
      Typed access to Explains why this operation definition is needed and why it's been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this operation definition is needed and why it's been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member idempotent
      Typed access to Operations that are idempotent (see [HTTP specification definition of idempotent](http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html)) may be invoked by performing an HTTP GET operation instead of a POST.
    }
    property idempotent : Boolean read GetIdempotentST write SetIdempotentST;
    {@member idempotentElement
      Operations that are idempotent (see [HTTP specification definition of idempotent](http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html)) may be invoked by performing an HTTP GET operation instead of a POST.
    }
    property idempotentElement : TFhirBoolean read FIdempotent write SetIdempotent;

    {@member code
      Typed access to The name used to invoke the operation.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      The name used to invoke the operation.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member notes
      Typed access to Additional information about how to use this operation or named query.
    }
    property notes : String read GetNotesST write SetNotesST;
    {@member notesElement
      Additional information about how to use this operation or named query.
    }
    property notesElement : TFhirString read FNotes write SetNotes;

    {@member base
      Typed access to Indicates that this operation definition is a constraining profile on the base. (defined for API consistency)
    }
    property base : TFhirReference{TFhirOperationDefinition} read FBase write SetBase;
    {@member baseElement
      Indicates that this operation definition is a constraining profile on the base.
    }
    property baseElement : TFhirReference{TFhirOperationDefinition} read FBase write SetBase;

    {@member system
      Typed access to Indicates whether this operation or named query can be invoked at the system level (e.g. without needing to choose a resource type for the context).
    }
    property system : Boolean read GetSystemST write SetSystemST;
    {@member systemElement
      Indicates whether this operation or named query can be invoked at the system level (e.g. without needing to choose a resource type for the context).
    }
    property systemElement : TFhirBoolean read FSystem write SetSystem;

    {@member type_List
      Indicates whether this operation or named query can be invoked at the resource type level for any given resource type level (e.g. without needing to choose a resource type for the context).
    }
    property type_List : TFhirCodeList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member instance
      Typed access to Indicates whether this operation can be invoked on a particular instance of one of the given types.
    }
    property instance : Boolean read GetInstanceST write SetInstanceST;
    {@member instanceElement
      Indicates whether this operation can be invoked on a particular instance of one of the given types.
    }
    property instanceElement : TFhirBoolean read FInstance write SetInstance;

    {@member parameterList
      The parameters for the operation/query.
    }
    property parameterList : TFhirOperationDefinitionParameterList read GetParameterList;
    property hasParameterList : boolean read GetHasParameterList;

  end;


  TFhirOperationDefinitionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationDefinitionList;
    function GetCurrent : TFhirOperationDefinition;
  public
    Constructor Create(list : TFhirOperationDefinitionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationDefinition read GetCurrent;
  end;


  {@Class TFhirOperationDefinitionList
    A list of FhirOperationDefinition
  }
  {!.Net HL7Connect.Fhir.OperationDefinitionList}
  TFhirOperationDefinitionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationDefinition;
    procedure SetItemN(index : Integer; value : TFhirOperationDefinition);
  public
    {!script hide}
    function Link : TFhirOperationDefinitionList; Overload;
    function Clone : TFhirOperationDefinitionList; Overload;
    function GetEnumerator : TFhirOperationDefinitionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationDefinition to the end of the list.
    }
    function Append : TFhirOperationDefinition;

    
    {@member AddItem
      Add an already existing FhirOperationDefinition to the end of the list.
    }
    procedure AddItem(value : TFhirOperationDefinition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationDefinition) : Integer;
    

    {@member Insert
      Insert FhirOperationDefinition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationDefinition;
    

    {@member InsertItem
       Insert an existing FhirOperationDefinition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationDefinition);
    
    {@member Item
       Get the iIndexth FhirOperationDefinition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationDefinition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationDefinition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationDefinition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationDefinitions[index : Integer] : TFhirOperationDefinition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationOutcomeIssue : TFhirElement
    An error, warning or information message that results from a system action.
  }
  {!.Net HL7Connect.Fhir.OperationOutcomeIssue}
  TFhirOperationOutcomeIssue = class (TFhirBackboneElement)
  private
    FSeverity : TFhirEnum;
    FCode : TFhirEnum;
    FDetails : TFhirCodeableConcept;
    FDiagnostics : TFhirString;
    FlocationList : TFhirStringList;
    Procedure SetSeverity(value : TFhirEnum);
    Function GetSeverityST : TFhirIssueSeverity;
    Procedure SetSeverityST(value : TFhirIssueSeverity);
    Procedure SetCode(value : TFhirEnum);
    Function GetCodeST : TFhirIssueType;
    Procedure SetCodeST(value : TFhirIssueType);
    Procedure SetDetails(value : TFhirCodeableConcept);
    Procedure SetDiagnostics(value : TFhirString);
    Function GetDiagnosticsST : String;
    Procedure SetDiagnosticsST(value : String);
    function GetLocationList : TFhirStringList;
    function GetHasLocationList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationOutcomeIssue; overload;
    function Clone : TFhirOperationOutcomeIssue; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member severity
      Indicates whether the issue indicates a variation from successful processing.
    }
    property severity : TFhirIssueSeverity read GetSeverityST write SetSeverityST;
    property severityElement : TFhirEnum read FSeverity write SetSeverity;

    {@member code
      Describes the type of the issue. The system that creates an OperationOutcome SHALL choose the most applicable code from the IssueType value set, and may additional provide its own code for the error in the details element.
    }
    property code : TFhirIssueType read GetCodeST write SetCodeST;
    property codeElement : TFhirEnum read FCode write SetCode;

    {@member details
      Typed access to Additional details about the error. This may be a text description of the error, or a system code that identifies the error. (defined for API consistency)
    }
    property details : TFhirCodeableConcept read FDetails write SetDetails;
    {@member detailsElement
      Additional details about the error. This may be a text description of the error, or a system code that identifies the error.
    }
    property detailsElement : TFhirCodeableConcept read FDetails write SetDetails;

    {@member diagnostics
      Typed access to Additional diagnostic information about the issue.  Typically, this may be a description of how a value is erroneous, or a stack dump to help trace the issue.
    }
    property diagnostics : String read GetDiagnosticsST write SetDiagnosticsST;
    {@member diagnosticsElement
      Additional diagnostic information about the issue.  Typically, this may be a description of how a value is erroneous, or a stack dump to help trace the issue.
    }
    property diagnosticsElement : TFhirString read FDiagnostics write SetDiagnostics;

    {@member locationList
      A simple XPath limited to element names, repetition indicators and the default child access that identifies one of the elements in the resource that caused this issue to be raised.
    }
    property locationList : TFhirStringList read GetLocationList;
    property hasLocationList : boolean read GetHasLocationList;

  end;


  TFhirOperationOutcomeIssueListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationOutcomeIssueList;
    function GetCurrent : TFhirOperationOutcomeIssue;
  public
    Constructor Create(list : TFhirOperationOutcomeIssueList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationOutcomeIssue read GetCurrent;
  end;


  {@Class TFhirOperationOutcomeIssueList
    A list of FhirOperationOutcomeIssue
  }
  {!.Net HL7Connect.Fhir.OperationOutcomeIssueList}
  TFhirOperationOutcomeIssueList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationOutcomeIssue;
    procedure SetItemN(index : Integer; value : TFhirOperationOutcomeIssue);
  public
    {!script hide}
    function Link : TFhirOperationOutcomeIssueList; Overload;
    function Clone : TFhirOperationOutcomeIssueList; Overload;
    function GetEnumerator : TFhirOperationOutcomeIssueListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationOutcomeIssue to the end of the list.
    }
    function Append : TFhirOperationOutcomeIssue;

    
    {@member AddItem
      Add an already existing FhirOperationOutcomeIssue to the end of the list.
    }
    procedure AddItem(value : TFhirOperationOutcomeIssue); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationOutcomeIssue) : Integer;
    

    {@member Insert
      Insert FhirOperationOutcomeIssue before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationOutcomeIssue;
    

    {@member InsertItem
       Insert an existing FhirOperationOutcomeIssue before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationOutcomeIssue);
    
    {@member Item
       Get the iIndexth FhirOperationOutcomeIssue. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationOutcomeIssue. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationOutcomeIssue);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationOutcomeIssue;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationOutcomeIssues[index : Integer] : TFhirOperationOutcomeIssue read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOperationOutcome : TFhirDomainResource
    A collection of error, warning or information messages that result from a system action.
  }
  {!.Net HL7Connect.Fhir.OperationOutcome}
  TFhirOperationOutcome = class (TFhirDomainResource)
  private
    FissueList : TFhirOperationOutcomeIssueList;
    function GetIssueList : TFhirOperationOutcomeIssueList;
    function GetHasIssueList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOperationOutcome; overload;
    function Clone : TFhirOperationOutcome; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member issueList
      An error, warning or information message that results from a system action.
    }
    property issueList : TFhirOperationOutcomeIssueList read GetIssueList;
    property hasIssueList : boolean read GetHasIssueList;

  end;


  TFhirOperationOutcomeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOperationOutcomeList;
    function GetCurrent : TFhirOperationOutcome;
  public
    Constructor Create(list : TFhirOperationOutcomeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOperationOutcome read GetCurrent;
  end;


  {@Class TFhirOperationOutcomeList
    A list of FhirOperationOutcome
  }
  {!.Net HL7Connect.Fhir.OperationOutcomeList}
  TFhirOperationOutcomeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOperationOutcome;
    procedure SetItemN(index : Integer; value : TFhirOperationOutcome);
  public
    {!script hide}
    function Link : TFhirOperationOutcomeList; Overload;
    function Clone : TFhirOperationOutcomeList; Overload;
    function GetEnumerator : TFhirOperationOutcomeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOperationOutcome to the end of the list.
    }
    function Append : TFhirOperationOutcome;

    
    {@member AddItem
      Add an already existing FhirOperationOutcome to the end of the list.
    }
    procedure AddItem(value : TFhirOperationOutcome); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOperationOutcome) : Integer;
    

    {@member Insert
      Insert FhirOperationOutcome before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOperationOutcome;
    

    {@member InsertItem
       Insert an existing FhirOperationOutcome before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOperationOutcome);
    
    {@member Item
       Get the iIndexth FhirOperationOutcome. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOperationOutcome. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOperationOutcome);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOperationOutcome;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOperationOutcomes[index : Integer] : TFhirOperationOutcome read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOrderWhen : TFhirElement
    When order should be fulfilled.
  }
  {!.Net HL7Connect.Fhir.OrderWhen}
  TFhirOrderWhen = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FSchedule : TFhirTiming;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSchedule(value : TFhirTiming);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOrderWhen; overload;
    function Clone : TFhirOrderWhen; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code specifies when request should be done. The code may simply be a priority code. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Code specifies when request should be done. The code may simply be a priority code.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member schedule
      Typed access to A formal schedule. (defined for API consistency)
    }
    property schedule : TFhirTiming read FSchedule write SetSchedule;
    {@member scheduleElement
      A formal schedule.
    }
    property scheduleElement : TFhirTiming read FSchedule write SetSchedule;

  end;


  TFhirOrderWhenListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOrderWhenList;
    function GetCurrent : TFhirOrderWhen;
  public
    Constructor Create(list : TFhirOrderWhenList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOrderWhen read GetCurrent;
  end;


  {@Class TFhirOrderWhenList
    A list of FhirOrderWhen
  }
  {!.Net HL7Connect.Fhir.OrderWhenList}
  TFhirOrderWhenList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOrderWhen;
    procedure SetItemN(index : Integer; value : TFhirOrderWhen);
  public
    {!script hide}
    function Link : TFhirOrderWhenList; Overload;
    function Clone : TFhirOrderWhenList; Overload;
    function GetEnumerator : TFhirOrderWhenListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOrderWhen to the end of the list.
    }
    function Append : TFhirOrderWhen;

    
    {@member AddItem
      Add an already existing FhirOrderWhen to the end of the list.
    }
    procedure AddItem(value : TFhirOrderWhen); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOrderWhen) : Integer;
    

    {@member Insert
      Insert FhirOrderWhen before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOrderWhen;
    

    {@member InsertItem
       Insert an existing FhirOrderWhen before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOrderWhen);
    
    {@member Item
       Get the iIndexth FhirOrderWhen. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOrderWhen. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOrderWhen);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOrderWhen;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOrderWhens[index : Integer] : TFhirOrderWhen read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOrder : TFhirDomainResource
    A request to perform an action.
  }
  {!.Net HL7Connect.Fhir.Order}
  TFhirOrder = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FDate : TFhirDateTime;
    FSubject : TFhirReference{Resource};
    FSource : TFhirReference{Resource};
    FTarget : TFhirReference{Resource};
    FReason : TFhirType;
    FWhen : TFhirOrderWhen;
    FdetailList : TFhirReferenceList{TFhirReference};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetSource(value : TFhirReference{Resource});
    Procedure SetTarget(value : TFhirReference{Resource});
    Procedure SetReason(value : TFhirType);
    Procedure SetWhen(value : TFhirOrderWhen);
    function GetDetailList : TFhirReferenceList{TFhirReference};
    function GetHasDetailList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOrder; overload;
    function Clone : TFhirOrder; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifiers assigned to this order by the orderer or by the receiver.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member date
      Typed access to When the order was made.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      When the order was made.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member subject
      Typed access to Patient this order is about. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Patient this order is about.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member source
      Typed access to Who initiated the order. (defined for API consistency)
    }
    property source : TFhirReference{Resource} read FSource write SetSource;
    {@member sourceElement
      Who initiated the order.
    }
    property sourceElement : TFhirReference{Resource} read FSource write SetSource;

    {@member target
      Typed access to Who is intended to fulfill the order. (defined for API consistency)
    }
    property target : TFhirReference{Resource} read FTarget write SetTarget;
    {@member targetElement
      Who is intended to fulfill the order.
    }
    property targetElement : TFhirReference{Resource} read FTarget write SetTarget;

    {@member reason
      Typed access to Text - why the order was made. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      Text - why the order was made.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member when
      Typed access to When order should be fulfilled. (defined for API consistency)
    }
    property when : TFhirOrderWhen read FWhen write SetWhen;
    {@member whenElement
      When order should be fulfilled.
    }
    property whenElement : TFhirOrderWhen read FWhen write SetWhen;

    {@member detailList
      What action is being ordered.
    }
    property detailList : TFhirReferenceList{TFhirReference} read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

  end;


  TFhirOrderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOrderList;
    function GetCurrent : TFhirOrder;
  public
    Constructor Create(list : TFhirOrderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOrder read GetCurrent;
  end;


  {@Class TFhirOrderList
    A list of FhirOrder
  }
  {!.Net HL7Connect.Fhir.OrderList}
  TFhirOrderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOrder;
    procedure SetItemN(index : Integer; value : TFhirOrder);
  public
    {!script hide}
    function Link : TFhirOrderList; Overload;
    function Clone : TFhirOrderList; Overload;
    function GetEnumerator : TFhirOrderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOrder to the end of the list.
    }
    function Append : TFhirOrder;

    
    {@member AddItem
      Add an already existing FhirOrder to the end of the list.
    }
    procedure AddItem(value : TFhirOrder); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOrder) : Integer;
    

    {@member Insert
      Insert FhirOrder before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOrder;
    

    {@member InsertItem
       Insert an existing FhirOrder before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOrder);
    
    {@member Item
       Get the iIndexth FhirOrder. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOrder. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOrder);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOrder;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOrders[index : Integer] : TFhirOrder read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOrderResponse : TFhirDomainResource
    A response to an order.
  }
  {!.Net HL7Connect.Fhir.OrderResponse}
  TFhirOrderResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirOrder};
    FDate : TFhirDateTime;
    FWho : TFhirReference{Resource};
    FOrderStatus : TFhirEnum;
    FDescription : TFhirString;
    FfulfillmentList : TFhirReferenceList{TFhirReference};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirOrder});
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetWho(value : TFhirReference{Resource});
    Procedure SetOrderStatus(value : TFhirEnum);
    Function GetOrderStatusST : TFhirOrderStatus;
    Procedure SetOrderStatusST(value : TFhirOrderStatus);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetFulfillmentList : TFhirReferenceList{TFhirReference};
    function GetHasFulfillmentList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOrderResponse; overload;
    function Clone : TFhirOrderResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifiers assigned to this order. The identifiers are usually assigned by the system responding to the order, but they may be provided or added to by other systems.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to A reference to the order that this is in response to. (defined for API consistency)
    }
    property request : TFhirReference{TFhirOrder} read FRequest write SetRequest;
    {@member requestElement
      A reference to the order that this is in response to.
    }
    property requestElement : TFhirReference{TFhirOrder} read FRequest write SetRequest;

    {@member date
      Typed access to The date and time at which this order response was made (created/posted).
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date and time at which this order response was made (created/posted).
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member who
      Typed access to The person, organization, or device credited with making the response. (defined for API consistency)
    }
    property who : TFhirReference{Resource} read FWho write SetWho;
    {@member whoElement
      The person, organization, or device credited with making the response.
    }
    property whoElement : TFhirReference{Resource} read FWho write SetWho;

    {@member orderStatus
      What this response says about the status of the original order.
    }
    property orderStatus : TFhirOrderStatus read GetOrderStatusST write SetOrderStatusST;
    property orderStatusElement : TFhirEnum read FOrderStatus write SetOrderStatus;

    {@member description
      Typed access to Additional description about the response - e.g. a text description provided by a human user when making decisions about the order.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Additional description about the response - e.g. a text description provided by a human user when making decisions about the order.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member fulfillmentList
      Links to resources that provide details of the outcome of performing the order; e.g. Diagnostic Reports in a response that is made to an order that referenced a diagnostic order.
    }
    property fulfillmentList : TFhirReferenceList{TFhirReference} read GetFulfillmentList;
    property hasFulfillmentList : boolean read GetHasFulfillmentList;

  end;


  TFhirOrderResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOrderResponseList;
    function GetCurrent : TFhirOrderResponse;
  public
    Constructor Create(list : TFhirOrderResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOrderResponse read GetCurrent;
  end;


  {@Class TFhirOrderResponseList
    A list of FhirOrderResponse
  }
  {!.Net HL7Connect.Fhir.OrderResponseList}
  TFhirOrderResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOrderResponse;
    procedure SetItemN(index : Integer; value : TFhirOrderResponse);
  public
    {!script hide}
    function Link : TFhirOrderResponseList; Overload;
    function Clone : TFhirOrderResponseList; Overload;
    function GetEnumerator : TFhirOrderResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOrderResponse to the end of the list.
    }
    function Append : TFhirOrderResponse;

    
    {@member AddItem
      Add an already existing FhirOrderResponse to the end of the list.
    }
    procedure AddItem(value : TFhirOrderResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOrderResponse) : Integer;
    

    {@member Insert
      Insert FhirOrderResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOrderResponse;
    

    {@member InsertItem
       Insert an existing FhirOrderResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOrderResponse);
    
    {@member Item
       Get the iIndexth FhirOrderResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOrderResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOrderResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOrderResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOrderResponses[index : Integer] : TFhirOrderResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOrganizationContact : TFhirElement
    Contact for the organization for a certain purpose.
  }
  {!.Net HL7Connect.Fhir.OrganizationContact}
  TFhirOrganizationContact = class (TFhirBackboneElement)
  private
    FPurpose : TFhirCodeableConcept;
    FName : TFhirHumanName;
    FtelecomList : TFhirContactPointList;
    FAddress : TFhirAddress;
    Procedure SetPurpose(value : TFhirCodeableConcept);
    Procedure SetName(value : TFhirHumanName);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetAddress(value : TFhirAddress);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOrganizationContact; overload;
    function Clone : TFhirOrganizationContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member purpose
      Typed access to Indicates a purpose for which the contact can be reached. (defined for API consistency)
    }
    property purpose : TFhirCodeableConcept read FPurpose write SetPurpose;
    {@member purposeElement
      Indicates a purpose for which the contact can be reached.
    }
    property purposeElement : TFhirCodeableConcept read FPurpose write SetPurpose;

    {@member name
      Typed access to A name associated with the contact. (defined for API consistency)
    }
    property name : TFhirHumanName read FName write SetName;
    {@member nameElement
      A name associated with the contact.
    }
    property nameElement : TFhirHumanName read FName write SetName;

    {@member telecomList
      A contact detail (e.g. a telephone number or an email address) by which the party may be contacted.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member address
      Typed access to Visiting or postal addresses for the contact. (defined for API consistency)
    }
    property address : TFhirAddress read FAddress write SetAddress;
    {@member addressElement
      Visiting or postal addresses for the contact.
    }
    property addressElement : TFhirAddress read FAddress write SetAddress;

  end;


  TFhirOrganizationContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOrganizationContactList;
    function GetCurrent : TFhirOrganizationContact;
  public
    Constructor Create(list : TFhirOrganizationContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOrganizationContact read GetCurrent;
  end;


  {@Class TFhirOrganizationContactList
    A list of FhirOrganizationContact
  }
  {!.Net HL7Connect.Fhir.OrganizationContactList}
  TFhirOrganizationContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOrganizationContact;
    procedure SetItemN(index : Integer; value : TFhirOrganizationContact);
  public
    {!script hide}
    function Link : TFhirOrganizationContactList; Overload;
    function Clone : TFhirOrganizationContactList; Overload;
    function GetEnumerator : TFhirOrganizationContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOrganizationContact to the end of the list.
    }
    function Append : TFhirOrganizationContact;

    
    {@member AddItem
      Add an already existing FhirOrganizationContact to the end of the list.
    }
    procedure AddItem(value : TFhirOrganizationContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOrganizationContact) : Integer;
    

    {@member Insert
      Insert FhirOrganizationContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOrganizationContact;
    

    {@member InsertItem
       Insert an existing FhirOrganizationContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOrganizationContact);
    
    {@member Item
       Get the iIndexth FhirOrganizationContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOrganizationContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOrganizationContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOrganizationContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOrganizationContacts[index : Integer] : TFhirOrganizationContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirOrganization : TFhirDomainResource
    A formally or informally recognized grouping of people or organizations formed for the purpose of achieving some form of collective action.  Includes companies, institutions, corporations, departments, community groups, healthcare practice groups, etc.
  }
  {!.Net HL7Connect.Fhir.Organization}
  TFhirOrganization = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FActive : TFhirBoolean;
    FType_ : TFhirCodeableConcept;
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    FaddressList : TFhirAddressList;
    FPartOf : TFhirReference{TFhirOrganization};
    FcontactList : TFhirOrganizationContactList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetActive(value : TFhirBoolean);
    Function GetActiveST : Boolean;
    Procedure SetActiveST(value : Boolean);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    function GetAddressList : TFhirAddressList;
    function GetHasAddressList : Boolean;
    Procedure SetPartOf(value : TFhirReference{TFhirOrganization});
    function GetContactList : TFhirOrganizationContactList;
    function GetHasContactList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirOrganization; overload;
    function Clone : TFhirOrganization; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier for the organization that is used to identify the organization across multiple disparate systems.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member active
      Typed access to Whether the organization's record is still in active use.
    }
    property active : Boolean read GetActiveST write SetActiveST;
    {@member activeElement
      Whether the organization's record is still in active use.
    }
    property activeElement : TFhirBoolean read FActive write SetActive;

    {@member type_
      Typed access to The kind of organization that this is. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The kind of organization that this is.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member name
      Typed access to A name associated with the organization.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A name associated with the organization.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      A contact detail for the organization.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member addressList
      An address for the organization.
    }
    property addressList : TFhirAddressList read GetAddressList;
    property hasAddressList : boolean read GetHasAddressList;

    {@member partOf
      Typed access to The organization of which this organization forms a part. (defined for API consistency)
    }
    property partOf : TFhirReference{TFhirOrganization} read FPartOf write SetPartOf;
    {@member partOfElement
      The organization of which this organization forms a part.
    }
    property partOfElement : TFhirReference{TFhirOrganization} read FPartOf write SetPartOf;

    {@member contactList
      Contact for the organization for a certain purpose.
    }
    property contactList : TFhirOrganizationContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

  end;


  TFhirOrganizationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirOrganizationList;
    function GetCurrent : TFhirOrganization;
  public
    Constructor Create(list : TFhirOrganizationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirOrganization read GetCurrent;
  end;


  {@Class TFhirOrganizationList
    A list of FhirOrganization
  }
  {!.Net HL7Connect.Fhir.OrganizationList}
  TFhirOrganizationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirOrganization;
    procedure SetItemN(index : Integer; value : TFhirOrganization);
  public
    {!script hide}
    function Link : TFhirOrganizationList; Overload;
    function Clone : TFhirOrganizationList; Overload;
    function GetEnumerator : TFhirOrganizationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirOrganization to the end of the list.
    }
    function Append : TFhirOrganization;

    
    {@member AddItem
      Add an already existing FhirOrganization to the end of the list.
    }
    procedure AddItem(value : TFhirOrganization); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirOrganization) : Integer;
    

    {@member Insert
      Insert FhirOrganization before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirOrganization;
    

    {@member InsertItem
       Insert an existing FhirOrganization before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirOrganization);
    
    {@member Item
       Get the iIndexth FhirOrganization. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirOrganization. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirOrganization);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirOrganization;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirOrganizations[index : Integer] : TFhirOrganization read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPatientContact : TFhirElement
    A contact party (e.g. guardian, partner, friend) for the patient.
  }
  {!.Net HL7Connect.Fhir.PatientContact}
  TFhirPatientContact = class (TFhirBackboneElement)
  private
    FrelationshipList : TFhirCodeableConceptList;
    FName : TFhirHumanName;
    FtelecomList : TFhirContactPointList;
    FAddress : TFhirAddress;
    FGender : TFhirEnum;
    FOrganization : TFhirReference{TFhirOrganization};
    FPeriod : TFhirPeriod;
    function GetRelationshipList : TFhirCodeableConceptList;
    function GetHasRelationshipList : Boolean;
    Procedure SetName(value : TFhirHumanName);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetAddress(value : TFhirAddress);
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPatientContact; overload;
    function Clone : TFhirPatientContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member relationshipList
      The nature of the relationship between the patient and the contact person.
    }
    property relationshipList : TFhirCodeableConceptList read GetRelationshipList;
    property hasRelationshipList : boolean read GetHasRelationshipList;

    {@member name
      Typed access to A name associated with the contact person. (defined for API consistency)
    }
    property name : TFhirHumanName read FName write SetName;
    {@member nameElement
      A name associated with the contact person.
    }
    property nameElement : TFhirHumanName read FName write SetName;

    {@member telecomList
      A contact detail for the person, e.g. a telephone number or an email address.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member address
      Typed access to Address for the contact person. (defined for API consistency)
    }
    property address : TFhirAddress read FAddress write SetAddress;
    {@member addressElement
      Address for the contact person.
    }
    property addressElement : TFhirAddress read FAddress write SetAddress;

    {@member gender
      Administrative Gender - the gender that the contact person is considered to have for administration and record keeping purposes.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member organization
      Typed access to Organization on behalf of which the contact is acting or for which the contact is working. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      Organization on behalf of which the contact is acting or for which the contact is working.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member period
      Typed access to The period during which this contact person or organization is valid to be contacted relating to this patient. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period during which this contact person or organization is valid to be contacted relating to this patient.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirPatientContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPatientContactList;
    function GetCurrent : TFhirPatientContact;
  public
    Constructor Create(list : TFhirPatientContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPatientContact read GetCurrent;
  end;


  {@Class TFhirPatientContactList
    A list of FhirPatientContact
  }
  {!.Net HL7Connect.Fhir.PatientContactList}
  TFhirPatientContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPatientContact;
    procedure SetItemN(index : Integer; value : TFhirPatientContact);
  public
    {!script hide}
    function Link : TFhirPatientContactList; Overload;
    function Clone : TFhirPatientContactList; Overload;
    function GetEnumerator : TFhirPatientContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPatientContact to the end of the list.
    }
    function Append : TFhirPatientContact;

    
    {@member AddItem
      Add an already existing FhirPatientContact to the end of the list.
    }
    procedure AddItem(value : TFhirPatientContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPatientContact) : Integer;
    

    {@member Insert
      Insert FhirPatientContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPatientContact;
    

    {@member InsertItem
       Insert an existing FhirPatientContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPatientContact);
    
    {@member Item
       Get the iIndexth FhirPatientContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPatientContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPatientContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPatientContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPatientContacts[index : Integer] : TFhirPatientContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPatientAnimal : TFhirElement
    This patient is known to be an animal.
  }
  {!.Net HL7Connect.Fhir.PatientAnimal}
  TFhirPatientAnimal = class (TFhirBackboneElement)
  private
    FSpecies : TFhirCodeableConcept;
    FBreed : TFhirCodeableConcept;
    FGenderStatus : TFhirCodeableConcept;
    Procedure SetSpecies(value : TFhirCodeableConcept);
    Procedure SetBreed(value : TFhirCodeableConcept);
    Procedure SetGenderStatus(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPatientAnimal; overload;
    function Clone : TFhirPatientAnimal; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member species
      Typed access to Identifies the high level taxonomic categorization of the kind of animal. (defined for API consistency)
    }
    property species : TFhirCodeableConcept read FSpecies write SetSpecies;
    {@member speciesElement
      Identifies the high level taxonomic categorization of the kind of animal.
    }
    property speciesElement : TFhirCodeableConcept read FSpecies write SetSpecies;

    {@member breed
      Typed access to Identifies the detailed categorization of the kind of animal. (defined for API consistency)
    }
    property breed : TFhirCodeableConcept read FBreed write SetBreed;
    {@member breedElement
      Identifies the detailed categorization of the kind of animal.
    }
    property breedElement : TFhirCodeableConcept read FBreed write SetBreed;

    {@member genderStatus
      Typed access to Indicates the current state of the animal's reproductive organs. (defined for API consistency)
    }
    property genderStatus : TFhirCodeableConcept read FGenderStatus write SetGenderStatus;
    {@member genderStatusElement
      Indicates the current state of the animal's reproductive organs.
    }
    property genderStatusElement : TFhirCodeableConcept read FGenderStatus write SetGenderStatus;

  end;


  TFhirPatientAnimalListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPatientAnimalList;
    function GetCurrent : TFhirPatientAnimal;
  public
    Constructor Create(list : TFhirPatientAnimalList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPatientAnimal read GetCurrent;
  end;


  {@Class TFhirPatientAnimalList
    A list of FhirPatientAnimal
  }
  {!.Net HL7Connect.Fhir.PatientAnimalList}
  TFhirPatientAnimalList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPatientAnimal;
    procedure SetItemN(index : Integer; value : TFhirPatientAnimal);
  public
    {!script hide}
    function Link : TFhirPatientAnimalList; Overload;
    function Clone : TFhirPatientAnimalList; Overload;
    function GetEnumerator : TFhirPatientAnimalListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPatientAnimal to the end of the list.
    }
    function Append : TFhirPatientAnimal;

    
    {@member AddItem
      Add an already existing FhirPatientAnimal to the end of the list.
    }
    procedure AddItem(value : TFhirPatientAnimal); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPatientAnimal) : Integer;
    

    {@member Insert
      Insert FhirPatientAnimal before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPatientAnimal;
    

    {@member InsertItem
       Insert an existing FhirPatientAnimal before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPatientAnimal);
    
    {@member Item
       Get the iIndexth FhirPatientAnimal. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPatientAnimal. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPatientAnimal);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPatientAnimal;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPatientAnimals[index : Integer] : TFhirPatientAnimal read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPatientCommunication : TFhirElement
    Languages which may be used to communicate with the patient about his or her health.
  }
  {!.Net HL7Connect.Fhir.PatientCommunication}
  TFhirPatientCommunication = class (TFhirBackboneElement)
  private
    FLanguage : TFhirCodeableConcept;
    FPreferred : TFhirBoolean;
    Procedure SetLanguage(value : TFhirCodeableConcept);
    Procedure SetPreferred(value : TFhirBoolean);
    Function GetPreferredST : Boolean;
    Procedure SetPreferredST(value : Boolean);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPatientCommunication; overload;
    function Clone : TFhirPatientCommunication; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member language
      Typed access to The ISO-639-1 alpha 2 code in lower case for the language, optionally followed by a hyphen and the ISO-3166-1 alpha 2 code for the region in upper case; e.g. "en" for English, or "en-US" for American English versus "en-EN" for England English. (defined for API consistency)
    }
    property language : TFhirCodeableConcept read FLanguage write SetLanguage;
    {@member languageElement
      The ISO-639-1 alpha 2 code in lower case for the language, optionally followed by a hyphen and the ISO-3166-1 alpha 2 code for the region in upper case; e.g. "en" for English, or "en-US" for American English versus "en-EN" for England English.
    }
    property languageElement : TFhirCodeableConcept read FLanguage write SetLanguage;

    {@member preferred
      Typed access to Indicates whether or not the patient prefers this language (over other languages he masters up a certain level).
    }
    property preferred : Boolean read GetPreferredST write SetPreferredST;
    {@member preferredElement
      Indicates whether or not the patient prefers this language (over other languages he masters up a certain level).
    }
    property preferredElement : TFhirBoolean read FPreferred write SetPreferred;

  end;


  TFhirPatientCommunicationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPatientCommunicationList;
    function GetCurrent : TFhirPatientCommunication;
  public
    Constructor Create(list : TFhirPatientCommunicationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPatientCommunication read GetCurrent;
  end;


  {@Class TFhirPatientCommunicationList
    A list of FhirPatientCommunication
  }
  {!.Net HL7Connect.Fhir.PatientCommunicationList}
  TFhirPatientCommunicationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPatientCommunication;
    procedure SetItemN(index : Integer; value : TFhirPatientCommunication);
  public
    {!script hide}
    function Link : TFhirPatientCommunicationList; Overload;
    function Clone : TFhirPatientCommunicationList; Overload;
    function GetEnumerator : TFhirPatientCommunicationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPatientCommunication to the end of the list.
    }
    function Append : TFhirPatientCommunication;

    
    {@member AddItem
      Add an already existing FhirPatientCommunication to the end of the list.
    }
    procedure AddItem(value : TFhirPatientCommunication); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPatientCommunication) : Integer;
    

    {@member Insert
      Insert FhirPatientCommunication before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPatientCommunication;
    

    {@member InsertItem
       Insert an existing FhirPatientCommunication before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPatientCommunication);
    
    {@member Item
       Get the iIndexth FhirPatientCommunication. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPatientCommunication. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPatientCommunication);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPatientCommunication;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPatientCommunications[index : Integer] : TFhirPatientCommunication read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPatientLink : TFhirElement
    Link to another patient resource that concerns the same actual patient.
  }
  {!.Net HL7Connect.Fhir.PatientLink}
  TFhirPatientLink = class (TFhirBackboneElement)
  private
    FOther : TFhirReference{TFhirPatient};
    FType_ : TFhirEnum;
    Procedure SetOther(value : TFhirReference{TFhirPatient});
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirLinkType;
    Procedure SetType_ST(value : TFhirLinkType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPatientLink; overload;
    function Clone : TFhirPatientLink; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member other
      Typed access to The other patient resource that the link refers to. (defined for API consistency)
    }
    property other : TFhirReference{TFhirPatient} read FOther write SetOther;
    {@member otherElement
      The other patient resource that the link refers to.
    }
    property otherElement : TFhirReference{TFhirPatient} read FOther write SetOther;

    {@member type_
      The type of link between this patient resource and another patient resource.
    }
    property type_ : TFhirLinkType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

  end;


  TFhirPatientLinkListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPatientLinkList;
    function GetCurrent : TFhirPatientLink;
  public
    Constructor Create(list : TFhirPatientLinkList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPatientLink read GetCurrent;
  end;


  {@Class TFhirPatientLinkList
    A list of FhirPatientLink
  }
  {!.Net HL7Connect.Fhir.PatientLinkList}
  TFhirPatientLinkList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPatientLink;
    procedure SetItemN(index : Integer; value : TFhirPatientLink);
  public
    {!script hide}
    function Link : TFhirPatientLinkList; Overload;
    function Clone : TFhirPatientLinkList; Overload;
    function GetEnumerator : TFhirPatientLinkListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPatientLink to the end of the list.
    }
    function Append : TFhirPatientLink;

    
    {@member AddItem
      Add an already existing FhirPatientLink to the end of the list.
    }
    procedure AddItem(value : TFhirPatientLink); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPatientLink) : Integer;
    

    {@member Insert
      Insert FhirPatientLink before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPatientLink;
    

    {@member InsertItem
       Insert an existing FhirPatientLink before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPatientLink);
    
    {@member Item
       Get the iIndexth FhirPatientLink. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPatientLink. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPatientLink);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPatientLink;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPatientLinks[index : Integer] : TFhirPatientLink read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPatient : TFhirDomainResource
    Demographics and other administrative information about an individual or animal receiving care or other health-related services.
  }
  {!.Net HL7Connect.Fhir.Patient}
  TFhirPatient = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FActive : TFhirBoolean;
    FnameList : TFhirHumanNameList;
    FtelecomList : TFhirContactPointList;
    FGender : TFhirEnum;
    FBirthDate : TFhirDate;
    FDeceased : TFhirType;
    FaddressList : TFhirAddressList;
    FMaritalStatus : TFhirCodeableConcept;
    FMultipleBirth : TFhirType;
    FphotoList : TFhirAttachmentList;
    FcontactList : TFhirPatientContactList;
    FAnimal : TFhirPatientAnimal;
    FcommunicationList : TFhirPatientCommunicationList;
    FcareProviderList : TFhirReferenceList{Resource};
    FManagingOrganization : TFhirReference{TFhirOrganization};
    Flink_List : TFhirPatientLinkList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetActive(value : TFhirBoolean);
    Function GetActiveST : Boolean;
    Procedure SetActiveST(value : Boolean);
    function GetNameList : TFhirHumanNameList;
    function GetHasNameList : Boolean;
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetBirthDate(value : TFhirDate);
    Function GetBirthDateST : TDateAndTime;
    Procedure SetBirthDateST(value : TDateAndTime);
    Procedure SetDeceased(value : TFhirType);
    function GetAddressList : TFhirAddressList;
    function GetHasAddressList : Boolean;
    Procedure SetMaritalStatus(value : TFhirCodeableConcept);
    Procedure SetMultipleBirth(value : TFhirType);
    function GetPhotoList : TFhirAttachmentList;
    function GetHasPhotoList : Boolean;
    function GetContactList : TFhirPatientContactList;
    function GetHasContactList : Boolean;
    Procedure SetAnimal(value : TFhirPatientAnimal);
    function GetCommunicationList : TFhirPatientCommunicationList;
    function GetHasCommunicationList : Boolean;
    function GetCareProviderList : TFhirReferenceList{Resource};
    function GetHasCareProviderList : Boolean;
    Procedure SetManagingOrganization(value : TFhirReference{TFhirOrganization});
    function GetLink_List : TFhirPatientLinkList;
    function GetHasLink_List : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPatient; overload;
    function Clone : TFhirPatient; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      An identifier for this patient.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member active
      Typed access to Whether this patient record is in active use.
    }
    property active : Boolean read GetActiveST write SetActiveST;
    {@member activeElement
      Whether this patient record is in active use.
    }
    property activeElement : TFhirBoolean read FActive write SetActive;

    {@member nameList
      A name associated with the individual.
    }
    property nameList : TFhirHumanNameList read GetNameList;
    property hasNameList : boolean read GetHasNameList;

    {@member telecomList
      A contact detail (e.g. a telephone number or an email address) by which the individual may be contacted.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member gender
      Administrative Gender - the gender that the patient is considered to have for administration and record keeping purposes.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member birthDate
      Typed access to The date of birth for the individual.
    }
    property birthDate : TDateAndTime read GetBirthDateST write SetBirthDateST;
    {@member birthDateElement
      The date of birth for the individual.
    }
    property birthDateElement : TFhirDate read FBirthDate write SetBirthDate;

    {@member deceased
      Typed access to Indicates if the individual is deceased or not. (defined for API consistency)
    }
    property deceased : TFhirType read FDeceased write SetDeceased;
    {@member deceasedElement
      Indicates if the individual is deceased or not.
    }
    property deceasedElement : TFhirType read FDeceased write SetDeceased;

    {@member addressList
      Addresses for the individual.
    }
    property addressList : TFhirAddressList read GetAddressList;
    property hasAddressList : boolean read GetHasAddressList;

    {@member maritalStatus
      Typed access to This field contains a patient's most recent marital (civil) status. (defined for API consistency)
    }
    property maritalStatus : TFhirCodeableConcept read FMaritalStatus write SetMaritalStatus;
    {@member maritalStatusElement
      This field contains a patient's most recent marital (civil) status.
    }
    property maritalStatusElement : TFhirCodeableConcept read FMaritalStatus write SetMaritalStatus;

    {@member multipleBirth
      Typed access to Indicates whether the patient is part of a multiple or indicates the actual birth order. (defined for API consistency)
    }
    property multipleBirth : TFhirType read FMultipleBirth write SetMultipleBirth;
    {@member multipleBirthElement
      Indicates whether the patient is part of a multiple or indicates the actual birth order.
    }
    property multipleBirthElement : TFhirType read FMultipleBirth write SetMultipleBirth;

    {@member photoList
      Image of the patient.
    }
    property photoList : TFhirAttachmentList read GetPhotoList;
    property hasPhotoList : boolean read GetHasPhotoList;

    {@member contactList
      A contact party (e.g. guardian, partner, friend) for the patient.
    }
    property contactList : TFhirPatientContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member animal
      Typed access to This patient is known to be an animal. (defined for API consistency)
    }
    property animal : TFhirPatientAnimal read FAnimal write SetAnimal;
    {@member animalElement
      This patient is known to be an animal.
    }
    property animalElement : TFhirPatientAnimal read FAnimal write SetAnimal;

    {@member communicationList
      Languages which may be used to communicate with the patient about his or her health.
    }
    property communicationList : TFhirPatientCommunicationList read GetCommunicationList;
    property hasCommunicationList : boolean read GetHasCommunicationList;

    {@member careProviderList
      Patient's nominated care provider.
    }
    property careProviderList : TFhirReferenceList{Resource} read GetCareProviderList;
    property hasCareProviderList : boolean read GetHasCareProviderList;

    {@member managingOrganization
      Typed access to Organization that is the custodian of the patient record. (defined for API consistency)
    }
    property managingOrganization : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;
    {@member managingOrganizationElement
      Organization that is the custodian of the patient record.
    }
    property managingOrganizationElement : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;

    {@member link_List
      Link to another patient resource that concerns the same actual patient.
    }
    property link_List : TFhirPatientLinkList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

  end;


  TFhirPatientListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPatientList;
    function GetCurrent : TFhirPatient;
  public
    Constructor Create(list : TFhirPatientList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPatient read GetCurrent;
  end;


  {@Class TFhirPatientList
    A list of FhirPatient
  }
  {!.Net HL7Connect.Fhir.PatientList}
  TFhirPatientList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPatient;
    procedure SetItemN(index : Integer; value : TFhirPatient);
  public
    {!script hide}
    function Link : TFhirPatientList; Overload;
    function Clone : TFhirPatientList; Overload;
    function GetEnumerator : TFhirPatientListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPatient to the end of the list.
    }
    function Append : TFhirPatient;

    
    {@member AddItem
      Add an already existing FhirPatient to the end of the list.
    }
    procedure AddItem(value : TFhirPatient); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPatient) : Integer;
    

    {@member Insert
      Insert FhirPatient before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPatient;
    

    {@member InsertItem
       Insert an existing FhirPatient before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPatient);
    
    {@member Item
       Get the iIndexth FhirPatient. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPatient. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPatient);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPatient;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPatients[index : Integer] : TFhirPatient read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPaymentNotice : TFhirDomainResource
    This resource provides the status of the payment for goods and services rendered, and the request and response resource references.
  }
  {!.Net HL7Connect.Fhir.PaymentNotice}
  TFhirPaymentNotice = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FTarget : TFhirReference{TFhirOrganization};
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    FRequest : TFhirReference{TFhirReference};
    FResponse : TFhirReference{TFhirReference};
    FPaymentStatus : TFhirCoding;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetTarget(value : TFhirReference{TFhirOrganization});
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequest(value : TFhirReference{TFhirReference});
    Procedure SetResponse(value : TFhirReference{TFhirReference});
    Procedure SetPaymentStatus(value : TFhirCoding);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPaymentNotice; overload;
    function Clone : TFhirPaymentNotice; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when this resource was created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when this resource was created.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member target
      Typed access to The Insurer who is target  of the request. (defined for API consistency)
    }
    property target : TFhirReference{TFhirOrganization} read FTarget write SetTarget;
    {@member targetElement
      The Insurer who is target  of the request.
    }
    property targetElement : TFhirReference{TFhirOrganization} read FTarget write SetTarget;

    {@member provider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member request
      Typed access to Reference of resource to reverse. (defined for API consistency)
    }
    property request : TFhirReference{TFhirReference} read FRequest write SetRequest;
    {@member requestElement
      Reference of resource to reverse.
    }
    property requestElement : TFhirReference{TFhirReference} read FRequest write SetRequest;

    {@member response
      Typed access to Reference of response to resource to reverse. (defined for API consistency)
    }
    property response : TFhirReference{TFhirReference} read FResponse write SetResponse;
    {@member responseElement
      Reference of response to resource to reverse.
    }
    property responseElement : TFhirReference{TFhirReference} read FResponse write SetResponse;

    {@member paymentStatus
      Typed access to The payment status, typically paid: payment sent, cleared: payment received. (defined for API consistency)
    }
    property paymentStatus : TFhirCoding read FPaymentStatus write SetPaymentStatus;
    {@member paymentStatusElement
      The payment status, typically paid: payment sent, cleared: payment received.
    }
    property paymentStatusElement : TFhirCoding read FPaymentStatus write SetPaymentStatus;

  end;


  TFhirPaymentNoticeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPaymentNoticeList;
    function GetCurrent : TFhirPaymentNotice;
  public
    Constructor Create(list : TFhirPaymentNoticeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPaymentNotice read GetCurrent;
  end;


  {@Class TFhirPaymentNoticeList
    A list of FhirPaymentNotice
  }
  {!.Net HL7Connect.Fhir.PaymentNoticeList}
  TFhirPaymentNoticeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPaymentNotice;
    procedure SetItemN(index : Integer; value : TFhirPaymentNotice);
  public
    {!script hide}
    function Link : TFhirPaymentNoticeList; Overload;
    function Clone : TFhirPaymentNoticeList; Overload;
    function GetEnumerator : TFhirPaymentNoticeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPaymentNotice to the end of the list.
    }
    function Append : TFhirPaymentNotice;

    
    {@member AddItem
      Add an already existing FhirPaymentNotice to the end of the list.
    }
    procedure AddItem(value : TFhirPaymentNotice); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPaymentNotice) : Integer;
    

    {@member Insert
      Insert FhirPaymentNotice before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPaymentNotice;
    

    {@member InsertItem
       Insert an existing FhirPaymentNotice before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPaymentNotice);
    
    {@member Item
       Get the iIndexth FhirPaymentNotice. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPaymentNotice. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPaymentNotice);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPaymentNotice;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPaymentNotices[index : Integer] : TFhirPaymentNotice read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPaymentReconciliationDetail : TFhirElement
    List of individual settlement amounts and the corresponding transaction.
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliationDetail}
  TFhirPaymentReconciliationDetail = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FRequest : TFhirReference{TFhirReference};
    FResponce : TFhirReference{TFhirReference};
    FSubmitter : TFhirReference{TFhirOrganization};
    FPayee : TFhirReference{TFhirOrganization};
    FDate : TFhirDate;
    FAmount : TFhirQuantity;
    Procedure SetType_(value : TFhirCoding);
    Procedure SetRequest(value : TFhirReference{TFhirReference});
    Procedure SetResponce(value : TFhirReference{TFhirReference});
    Procedure SetSubmitter(value : TFhirReference{TFhirOrganization});
    Procedure SetPayee(value : TFhirReference{TFhirOrganization});
    Procedure SetDate(value : TFhirDate);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetAmount(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPaymentReconciliationDetail; overload;
    function Clone : TFhirPaymentReconciliationDetail; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Code to indicate the nature of the payment, adjustment, funds advance, etc. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      Code to indicate the nature of the payment, adjustment, funds advance, etc.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member request
      Typed access to The claim or financial resource. (defined for API consistency)
    }
    property request : TFhirReference{TFhirReference} read FRequest write SetRequest;
    {@member requestElement
      The claim or financial resource.
    }
    property requestElement : TFhirReference{TFhirReference} read FRequest write SetRequest;

    {@member responce
      Typed access to The claim response resource. (defined for API consistency)
    }
    property responce : TFhirReference{TFhirReference} read FResponce write SetResponce;
    {@member responceElement
      The claim response resource.
    }
    property responceElement : TFhirReference{TFhirReference} read FResponce write SetResponce;

    {@member submitter
      Typed access to The Organization which submitted the invoice or financial transaction. (defined for API consistency)
    }
    property submitter : TFhirReference{TFhirOrganization} read FSubmitter write SetSubmitter;
    {@member submitterElement
      The Organization which submitted the invoice or financial transaction.
    }
    property submitterElement : TFhirReference{TFhirOrganization} read FSubmitter write SetSubmitter;

    {@member payee
      Typed access to The organization which is receiving the payment. (defined for API consistency)
    }
    property payee : TFhirReference{TFhirOrganization} read FPayee write SetPayee;
    {@member payeeElement
      The organization which is receiving the payment.
    }
    property payeeElement : TFhirReference{TFhirOrganization} read FPayee write SetPayee;

    {@member date
      Typed access to The date of the invoice or financial resource.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date of the invoice or financial resource.
    }
    property dateElement : TFhirDate read FDate write SetDate;

    {@member amount
      Typed access to Amount paid for this detail. (defined for API consistency)
    }
    property amount : TFhirQuantity read FAmount write SetAmount;
    {@member amountElement
      Amount paid for this detail.
    }
    property amountElement : TFhirQuantity read FAmount write SetAmount;

  end;


  TFhirPaymentReconciliationDetailListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPaymentReconciliationDetailList;
    function GetCurrent : TFhirPaymentReconciliationDetail;
  public
    Constructor Create(list : TFhirPaymentReconciliationDetailList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPaymentReconciliationDetail read GetCurrent;
  end;


  {@Class TFhirPaymentReconciliationDetailList
    A list of FhirPaymentReconciliationDetail
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliationDetailList}
  TFhirPaymentReconciliationDetailList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPaymentReconciliationDetail;
    procedure SetItemN(index : Integer; value : TFhirPaymentReconciliationDetail);
  public
    {!script hide}
    function Link : TFhirPaymentReconciliationDetailList; Overload;
    function Clone : TFhirPaymentReconciliationDetailList; Overload;
    function GetEnumerator : TFhirPaymentReconciliationDetailListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPaymentReconciliationDetail to the end of the list.
    }
    function Append : TFhirPaymentReconciliationDetail;

    
    {@member AddItem
      Add an already existing FhirPaymentReconciliationDetail to the end of the list.
    }
    procedure AddItem(value : TFhirPaymentReconciliationDetail); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPaymentReconciliationDetail) : Integer;
    

    {@member Insert
      Insert FhirPaymentReconciliationDetail before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPaymentReconciliationDetail;
    

    {@member InsertItem
       Insert an existing FhirPaymentReconciliationDetail before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPaymentReconciliationDetail);
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliationDetail. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliationDetail. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPaymentReconciliationDetail);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPaymentReconciliationDetail;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPaymentReconciliationDetails[index : Integer] : TFhirPaymentReconciliationDetail read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPaymentReconciliationNote : TFhirElement
    Suite of notes.
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliationNote}
  TFhirPaymentReconciliationNote = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FText : TFhirString;
    Procedure SetType_(value : TFhirCoding);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPaymentReconciliationNote; overload;
    function Clone : TFhirPaymentReconciliationNote; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The note purpose: Print/Display. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The note purpose: Print/Display.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member text
      Typed access to The note text.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      The note text.
    }
    property textElement : TFhirString read FText write SetText;

  end;


  TFhirPaymentReconciliationNoteListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPaymentReconciliationNoteList;
    function GetCurrent : TFhirPaymentReconciliationNote;
  public
    Constructor Create(list : TFhirPaymentReconciliationNoteList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPaymentReconciliationNote read GetCurrent;
  end;


  {@Class TFhirPaymentReconciliationNoteList
    A list of FhirPaymentReconciliationNote
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliationNoteList}
  TFhirPaymentReconciliationNoteList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPaymentReconciliationNote;
    procedure SetItemN(index : Integer; value : TFhirPaymentReconciliationNote);
  public
    {!script hide}
    function Link : TFhirPaymentReconciliationNoteList; Overload;
    function Clone : TFhirPaymentReconciliationNoteList; Overload;
    function GetEnumerator : TFhirPaymentReconciliationNoteListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPaymentReconciliationNote to the end of the list.
    }
    function Append : TFhirPaymentReconciliationNote;

    
    {@member AddItem
      Add an already existing FhirPaymentReconciliationNote to the end of the list.
    }
    procedure AddItem(value : TFhirPaymentReconciliationNote); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPaymentReconciliationNote) : Integer;
    

    {@member Insert
      Insert FhirPaymentReconciliationNote before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPaymentReconciliationNote;
    

    {@member InsertItem
       Insert an existing FhirPaymentReconciliationNote before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPaymentReconciliationNote);
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliationNote. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliationNote. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPaymentReconciliationNote);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPaymentReconciliationNote;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPaymentReconciliationNotes[index : Integer] : TFhirPaymentReconciliationNote read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPaymentReconciliation : TFhirDomainResource
    This resource provides payment details and claim references supporting a bulk payment.
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliation}
  TFhirPaymentReconciliation = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirProcessRequest};
    FOutcome : TFhirEnum;
    FDisposition : TFhirString;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FPeriod : TFhirPeriod;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    FdetailList : TFhirPaymentReconciliationDetailList;
    FForm : TFhirCoding;
    FTotal : TFhirQuantity;
    FnoteList : TFhirPaymentReconciliationNoteList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirProcessRequest});
    Procedure SetOutcome(value : TFhirEnum);
    Function GetOutcomeST : TFhirRemittanceOutcome;
    Procedure SetOutcomeST(value : TFhirRemittanceOutcome);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
    function GetDetailList : TFhirPaymentReconciliationDetailList;
    function GetHasDetailList : Boolean;
    Procedure SetForm(value : TFhirCoding);
    Procedure SetTotal(value : TFhirQuantity);
    function GetNoteList : TFhirPaymentReconciliationNoteList;
    function GetHasNoteList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPaymentReconciliation; overload;
    function Clone : TFhirPaymentReconciliation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource reference. (defined for API consistency)
    }
    property request : TFhirReference{TFhirProcessRequest} read FRequest write SetRequest;
    {@member requestElement
      Original request resource reference.
    }
    property requestElement : TFhirReference{TFhirProcessRequest} read FRequest write SetRequest;

    {@member outcome
      Transaction status: error, complete.
    }
    property outcome : TFhirRemittanceOutcome read GetOutcomeST write SetOutcomeST;
    property outcomeElement : TFhirEnum read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member period
      Typed access to The period of time for which payments have been gathered into this bulk payment for settlement. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time for which payments have been gathered into this bulk payment for settlement.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member organization
      Typed access to The Insurer who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The Insurer who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

    {@member detailList
      List of individual settlement amounts and the corresponding transaction.
    }
    property detailList : TFhirPaymentReconciliationDetailList read GetDetailList;
    property hasDetailList : boolean read GetHasDetailList;

    {@member form
      Typed access to The form to be used for printing the content. (defined for API consistency)
    }
    property form : TFhirCoding read FForm write SetForm;
    {@member formElement
      The form to be used for printing the content.
    }
    property formElement : TFhirCoding read FForm write SetForm;

    {@member total
      Typed access to Total payment amount. (defined for API consistency)
    }
    property total : TFhirQuantity read FTotal write SetTotal;
    {@member totalElement
      Total payment amount.
    }
    property totalElement : TFhirQuantity read FTotal write SetTotal;

    {@member noteList
      Suite of notes.
    }
    property noteList : TFhirPaymentReconciliationNoteList read GetNoteList;
    property hasNoteList : boolean read GetHasNoteList;

  end;


  TFhirPaymentReconciliationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPaymentReconciliationList;
    function GetCurrent : TFhirPaymentReconciliation;
  public
    Constructor Create(list : TFhirPaymentReconciliationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPaymentReconciliation read GetCurrent;
  end;


  {@Class TFhirPaymentReconciliationList
    A list of FhirPaymentReconciliation
  }
  {!.Net HL7Connect.Fhir.PaymentReconciliationList}
  TFhirPaymentReconciliationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPaymentReconciliation;
    procedure SetItemN(index : Integer; value : TFhirPaymentReconciliation);
  public
    {!script hide}
    function Link : TFhirPaymentReconciliationList; Overload;
    function Clone : TFhirPaymentReconciliationList; Overload;
    function GetEnumerator : TFhirPaymentReconciliationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPaymentReconciliation to the end of the list.
    }
    function Append : TFhirPaymentReconciliation;

    
    {@member AddItem
      Add an already existing FhirPaymentReconciliation to the end of the list.
    }
    procedure AddItem(value : TFhirPaymentReconciliation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPaymentReconciliation) : Integer;
    

    {@member Insert
      Insert FhirPaymentReconciliation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPaymentReconciliation;
    

    {@member InsertItem
       Insert an existing FhirPaymentReconciliation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPaymentReconciliation);
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPaymentReconciliation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPaymentReconciliation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPaymentReconciliation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPaymentReconciliations[index : Integer] : TFhirPaymentReconciliation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPersonLink : TFhirElement
    Link to a resource that concerns the same actual person.
  }
  {!.Net HL7Connect.Fhir.PersonLink}
  TFhirPersonLink = class (TFhirBackboneElement)
  private
    FTarget : TFhirReference{Resource};
    FAssurance : TFhirEnum;
    Procedure SetTarget(value : TFhirReference{Resource});
    Procedure SetAssurance(value : TFhirEnum);
    Function GetAssuranceST : TFhirIdentityAssuranceLevel;
    Procedure SetAssuranceST(value : TFhirIdentityAssuranceLevel);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPersonLink; overload;
    function Clone : TFhirPersonLink; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member target
      Typed access to The resource to which this actual person is associated. (defined for API consistency)
    }
    property target : TFhirReference{Resource} read FTarget write SetTarget;
    {@member targetElement
      The resource to which this actual person is associated.
    }
    property targetElement : TFhirReference{Resource} read FTarget write SetTarget;

    {@member assurance
      Level of assurance that this link is actually associated with the target resource.
    }
    property assurance : TFhirIdentityAssuranceLevel read GetAssuranceST write SetAssuranceST;
    property assuranceElement : TFhirEnum read FAssurance write SetAssurance;

  end;


  TFhirPersonLinkListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPersonLinkList;
    function GetCurrent : TFhirPersonLink;
  public
    Constructor Create(list : TFhirPersonLinkList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPersonLink read GetCurrent;
  end;


  {@Class TFhirPersonLinkList
    A list of FhirPersonLink
  }
  {!.Net HL7Connect.Fhir.PersonLinkList}
  TFhirPersonLinkList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPersonLink;
    procedure SetItemN(index : Integer; value : TFhirPersonLink);
  public
    {!script hide}
    function Link : TFhirPersonLinkList; Overload;
    function Clone : TFhirPersonLinkList; Overload;
    function GetEnumerator : TFhirPersonLinkListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPersonLink to the end of the list.
    }
    function Append : TFhirPersonLink;

    
    {@member AddItem
      Add an already existing FhirPersonLink to the end of the list.
    }
    procedure AddItem(value : TFhirPersonLink); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPersonLink) : Integer;
    

    {@member Insert
      Insert FhirPersonLink before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPersonLink;
    

    {@member InsertItem
       Insert an existing FhirPersonLink before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPersonLink);
    
    {@member Item
       Get the iIndexth FhirPersonLink. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPersonLink. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPersonLink);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPersonLink;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPersonLinks[index : Integer] : TFhirPersonLink read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPerson : TFhirDomainResource
    Demographics and administrative information about a person independent of a specific health-related context.
  }
  {!.Net HL7Connect.Fhir.Person}
  TFhirPerson = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FnameList : TFhirHumanNameList;
    FtelecomList : TFhirContactPointList;
    FGender : TFhirEnum;
    FBirthDate : TFhirDate;
    FaddressList : TFhirAddressList;
    FPhoto : TFhirAttachment;
    FManagingOrganization : TFhirReference{TFhirOrganization};
    FActive : TFhirBoolean;
    Flink_List : TFhirPersonLinkList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetNameList : TFhirHumanNameList;
    function GetHasNameList : Boolean;
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetBirthDate(value : TFhirDate);
    Function GetBirthDateST : TDateAndTime;
    Procedure SetBirthDateST(value : TDateAndTime);
    function GetAddressList : TFhirAddressList;
    function GetHasAddressList : Boolean;
    Procedure SetPhoto(value : TFhirAttachment);
    Procedure SetManagingOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetActive(value : TFhirBoolean);
    Function GetActiveST : Boolean;
    Procedure SetActiveST(value : Boolean);
    function GetLink_List : TFhirPersonLinkList;
    function GetHasLink_List : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPerson; overload;
    function Clone : TFhirPerson; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier for a person within a particular scope.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member nameList
      A name associated with the person.
    }
    property nameList : TFhirHumanNameList read GetNameList;
    property hasNameList : boolean read GetHasNameList;

    {@member telecomList
      A contact detail for the person, e.g. a telephone number or an email address.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member gender
      Administrative Gender.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member birthDate
      Typed access to The birth date for the person.
    }
    property birthDate : TDateAndTime read GetBirthDateST write SetBirthDateST;
    {@member birthDateElement
      The birth date for the person.
    }
    property birthDateElement : TFhirDate read FBirthDate write SetBirthDate;

    {@member addressList
      One or more addresses for the person.
    }
    property addressList : TFhirAddressList read GetAddressList;
    property hasAddressList : boolean read GetHasAddressList;

    {@member photo
      Typed access to An image that can be displayed as a thumbnail of the person to enhance the identification of the individual. (defined for API consistency)
    }
    property photo : TFhirAttachment read FPhoto write SetPhoto;
    {@member photoElement
      An image that can be displayed as a thumbnail of the person to enhance the identification of the individual.
    }
    property photoElement : TFhirAttachment read FPhoto write SetPhoto;

    {@member managingOrganization
      Typed access to The organization that is the custodian of the person record. (defined for API consistency)
    }
    property managingOrganization : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;
    {@member managingOrganizationElement
      The organization that is the custodian of the person record.
    }
    property managingOrganizationElement : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;

    {@member active
      Typed access to Whether this person's record is in active use.
    }
    property active : Boolean read GetActiveST write SetActiveST;
    {@member activeElement
      Whether this person's record is in active use.
    }
    property activeElement : TFhirBoolean read FActive write SetActive;

    {@member link_List
      Link to a resource that concerns the same actual person.
    }
    property link_List : TFhirPersonLinkList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

  end;


  TFhirPersonListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPersonList;
    function GetCurrent : TFhirPerson;
  public
    Constructor Create(list : TFhirPersonList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPerson read GetCurrent;
  end;


  {@Class TFhirPersonList
    A list of FhirPerson
  }
  {!.Net HL7Connect.Fhir.PersonList}
  TFhirPersonList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPerson;
    procedure SetItemN(index : Integer; value : TFhirPerson);
  public
    {!script hide}
    function Link : TFhirPersonList; Overload;
    function Clone : TFhirPersonList; Overload;
    function GetEnumerator : TFhirPersonListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPerson to the end of the list.
    }
    function Append : TFhirPerson;

    
    {@member AddItem
      Add an already existing FhirPerson to the end of the list.
    }
    procedure AddItem(value : TFhirPerson); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPerson) : Integer;
    

    {@member Insert
      Insert FhirPerson before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPerson;
    

    {@member InsertItem
       Insert an existing FhirPerson before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPerson);
    
    {@member Item
       Get the iIndexth FhirPerson. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPerson. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPerson);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPerson;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPeople[index : Integer] : TFhirPerson read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPractitionerPractitionerRole : TFhirElement
    The list of roles/organizations that the practitioner is associated with.
  }
  {!.Net HL7Connect.Fhir.PractitionerPractitionerRole}
  TFhirPractitionerPractitionerRole = class (TFhirBackboneElement)
  private
    FManagingOrganization : TFhirReference{TFhirOrganization};
    FRole : TFhirCodeableConcept;
    FspecialtyList : TFhirCodeableConceptList;
    FPeriod : TFhirPeriod;
    FlocationList : TFhirReferenceList{TFhirLocation};
    FhealthcareServiceList : TFhirReferenceList{TFhirHealthcareService};
    Procedure SetManagingOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRole(value : TFhirCodeableConcept);
    function GetSpecialtyList : TFhirCodeableConceptList;
    function GetHasSpecialtyList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    function GetLocationList : TFhirReferenceList{TFhirLocation};
    function GetHasLocationList : Boolean;
    function GetHealthcareServiceList : TFhirReferenceList{TFhirHealthcareService};
    function GetHasHealthcareServiceList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPractitionerPractitionerRole; overload;
    function Clone : TFhirPractitionerPractitionerRole; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member managingOrganization
      Typed access to The organization where the Practitioner performs the roles associated. (defined for API consistency)
    }
    property managingOrganization : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;
    {@member managingOrganizationElement
      The organization where the Practitioner performs the roles associated.
    }
    property managingOrganizationElement : TFhirReference{TFhirOrganization} read FManagingOrganization write SetManagingOrganization;

    {@member role
      Typed access to Roles which this practitioner is authorized to perform for the organization. (defined for API consistency)
    }
    property role : TFhirCodeableConcept read FRole write SetRole;
    {@member roleElement
      Roles which this practitioner is authorized to perform for the organization.
    }
    property roleElement : TFhirCodeableConcept read FRole write SetRole;

    {@member specialtyList
      Specific specialty of the practitioner.
    }
    property specialtyList : TFhirCodeableConceptList read GetSpecialtyList;
    property hasSpecialtyList : boolean read GetHasSpecialtyList;

    {@member period
      Typed access to The period during which the person is authorized to act as a practitioner in these role(s) for the organization. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period during which the person is authorized to act as a practitioner in these role(s) for the organization.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member locationList
      The location(s) at which this practitioner provides care.
    }
    property locationList : TFhirReferenceList{TFhirLocation} read GetLocationList;
    property hasLocationList : boolean read GetHasLocationList;

    {@member healthcareServiceList
      The list of healthcare services that this worker provides for this role's Organization/Location(s).
    }
    property healthcareServiceList : TFhirReferenceList{TFhirHealthcareService} read GetHealthcareServiceList;
    property hasHealthcareServiceList : boolean read GetHasHealthcareServiceList;

  end;


  TFhirPractitionerPractitionerRoleListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPractitionerPractitionerRoleList;
    function GetCurrent : TFhirPractitionerPractitionerRole;
  public
    Constructor Create(list : TFhirPractitionerPractitionerRoleList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPractitionerPractitionerRole read GetCurrent;
  end;


  {@Class TFhirPractitionerPractitionerRoleList
    A list of FhirPractitionerPractitionerRole
  }
  {!.Net HL7Connect.Fhir.PractitionerPractitionerRoleList}
  TFhirPractitionerPractitionerRoleList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPractitionerPractitionerRole;
    procedure SetItemN(index : Integer; value : TFhirPractitionerPractitionerRole);
  public
    {!script hide}
    function Link : TFhirPractitionerPractitionerRoleList; Overload;
    function Clone : TFhirPractitionerPractitionerRoleList; Overload;
    function GetEnumerator : TFhirPractitionerPractitionerRoleListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPractitionerPractitionerRole to the end of the list.
    }
    function Append : TFhirPractitionerPractitionerRole;

    
    {@member AddItem
      Add an already existing FhirPractitionerPractitionerRole to the end of the list.
    }
    procedure AddItem(value : TFhirPractitionerPractitionerRole); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPractitionerPractitionerRole) : Integer;
    

    {@member Insert
      Insert FhirPractitionerPractitionerRole before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPractitionerPractitionerRole;
    

    {@member InsertItem
       Insert an existing FhirPractitionerPractitionerRole before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPractitionerPractitionerRole);
    
    {@member Item
       Get the iIndexth FhirPractitionerPractitionerRole. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPractitionerPractitionerRole. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPractitionerPractitionerRole);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPractitionerPractitionerRole;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPractitionerPractitionerRoles[index : Integer] : TFhirPractitionerPractitionerRole read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPractitionerQualification : TFhirElement
    Qualifications obtained by training and certification.
  }
  {!.Net HL7Connect.Fhir.PractitionerQualification}
  TFhirPractitionerQualification = class (TFhirBackboneElement)
  private
    FidentifierList : TFhirIdentifierList;
    FCode : TFhirCodeableConcept;
    FPeriod : TFhirPeriod;
    FIssuer : TFhirReference{TFhirOrganization};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetIssuer(value : TFhirReference{TFhirOrganization});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPractitionerQualification; overload;
    function Clone : TFhirPractitionerQualification; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifierList
      An identifier that applies to this person's qualification in this role.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member code
      Typed access to Coded representation of the qualification. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Coded representation of the qualification.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member period
      Typed access to Period during which the qualification is valid. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      Period during which the qualification is valid.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member issuer
      Typed access to Organization that regulates and issues the qualification. (defined for API consistency)
    }
    property issuer : TFhirReference{TFhirOrganization} read FIssuer write SetIssuer;
    {@member issuerElement
      Organization that regulates and issues the qualification.
    }
    property issuerElement : TFhirReference{TFhirOrganization} read FIssuer write SetIssuer;

  end;


  TFhirPractitionerQualificationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPractitionerQualificationList;
    function GetCurrent : TFhirPractitionerQualification;
  public
    Constructor Create(list : TFhirPractitionerQualificationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPractitionerQualification read GetCurrent;
  end;


  {@Class TFhirPractitionerQualificationList
    A list of FhirPractitionerQualification
  }
  {!.Net HL7Connect.Fhir.PractitionerQualificationList}
  TFhirPractitionerQualificationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPractitionerQualification;
    procedure SetItemN(index : Integer; value : TFhirPractitionerQualification);
  public
    {!script hide}
    function Link : TFhirPractitionerQualificationList; Overload;
    function Clone : TFhirPractitionerQualificationList; Overload;
    function GetEnumerator : TFhirPractitionerQualificationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPractitionerQualification to the end of the list.
    }
    function Append : TFhirPractitionerQualification;

    
    {@member AddItem
      Add an already existing FhirPractitionerQualification to the end of the list.
    }
    procedure AddItem(value : TFhirPractitionerQualification); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPractitionerQualification) : Integer;
    

    {@member Insert
      Insert FhirPractitionerQualification before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPractitionerQualification;
    

    {@member InsertItem
       Insert an existing FhirPractitionerQualification before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPractitionerQualification);
    
    {@member Item
       Get the iIndexth FhirPractitionerQualification. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPractitionerQualification. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPractitionerQualification);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPractitionerQualification;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPractitionerQualifications[index : Integer] : TFhirPractitionerQualification read GetItemN write SetItemN; default;
  End;


  {@Class TFhirPractitioner : TFhirDomainResource
    A person who is directly or indirectly involved in the provisioning of healthcare.
  }
  {!.Net HL7Connect.Fhir.Practitioner}
  TFhirPractitioner = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FActive : TFhirBoolean;
    FName : TFhirHumanName;
    FtelecomList : TFhirContactPointList;
    FaddressList : TFhirAddressList;
    FGender : TFhirEnum;
    FBirthDate : TFhirDate;
    FphotoList : TFhirAttachmentList;
    FpractitionerRoleList : TFhirPractitionerPractitionerRoleList;
    FqualificationList : TFhirPractitionerQualificationList;
    FcommunicationList : TFhirCodeableConceptList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetActive(value : TFhirBoolean);
    Function GetActiveST : Boolean;
    Procedure SetActiveST(value : Boolean);
    Procedure SetName(value : TFhirHumanName);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    function GetAddressList : TFhirAddressList;
    function GetHasAddressList : Boolean;
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetBirthDate(value : TFhirDate);
    Function GetBirthDateST : TDateAndTime;
    Procedure SetBirthDateST(value : TDateAndTime);
    function GetPhotoList : TFhirAttachmentList;
    function GetHasPhotoList : Boolean;
    function GetPractitionerRoleList : TFhirPractitionerPractitionerRoleList;
    function GetHasPractitionerRoleList : Boolean;
    function GetQualificationList : TFhirPractitionerQualificationList;
    function GetHasQualificationList : Boolean;
    function GetCommunicationList : TFhirCodeableConceptList;
    function GetHasCommunicationList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirPractitioner; overload;
    function Clone : TFhirPractitioner; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      An identifier that applies to this person in this role.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member active
      Typed access to Whether this practitioner's record is in active use.
    }
    property active : Boolean read GetActiveST write SetActiveST;
    {@member activeElement
      Whether this practitioner's record is in active use.
    }
    property activeElement : TFhirBoolean read FActive write SetActive;

    {@member name
      Typed access to A name associated with the person. (defined for API consistency)
    }
    property name : TFhirHumanName read FName write SetName;
    {@member nameElement
      A name associated with the person.
    }
    property nameElement : TFhirHumanName read FName write SetName;

    {@member telecomList
      A contact detail for the practitioner, e.g. a telephone number or an email address.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member addressList
      The postal address where the practitioner can be found or visited or to which mail can be delivered.
    }
    property addressList : TFhirAddressList read GetAddressList;
    property hasAddressList : boolean read GetHasAddressList;

    {@member gender
      Administrative Gender - the gender that the person is considered to have for administration and record keeping purposes.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member birthDate
      Typed access to The date of birth for the practitioner.
    }
    property birthDate : TDateAndTime read GetBirthDateST write SetBirthDateST;
    {@member birthDateElement
      The date of birth for the practitioner.
    }
    property birthDateElement : TFhirDate read FBirthDate write SetBirthDate;

    {@member photoList
      Image of the person.
    }
    property photoList : TFhirAttachmentList read GetPhotoList;
    property hasPhotoList : boolean read GetHasPhotoList;

    {@member practitionerRoleList
      The list of roles/organizations that the practitioner is associated with.
    }
    property practitionerRoleList : TFhirPractitionerPractitionerRoleList read GetPractitionerRoleList;
    property hasPractitionerRoleList : boolean read GetHasPractitionerRoleList;

    {@member qualificationList
      Qualifications obtained by training and certification.
    }
    property qualificationList : TFhirPractitionerQualificationList read GetQualificationList;
    property hasQualificationList : boolean read GetHasQualificationList;

    {@member communicationList
      A language the practitioner is able to use in patient communication.
    }
    property communicationList : TFhirCodeableConceptList read GetCommunicationList;
    property hasCommunicationList : boolean read GetHasCommunicationList;

  end;


  TFhirPractitionerListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirPractitionerList;
    function GetCurrent : TFhirPractitioner;
  public
    Constructor Create(list : TFhirPractitionerList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirPractitioner read GetCurrent;
  end;


  {@Class TFhirPractitionerList
    A list of FhirPractitioner
  }
  {!.Net HL7Connect.Fhir.PractitionerList}
  TFhirPractitionerList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirPractitioner;
    procedure SetItemN(index : Integer; value : TFhirPractitioner);
  public
    {!script hide}
    function Link : TFhirPractitionerList; Overload;
    function Clone : TFhirPractitionerList; Overload;
    function GetEnumerator : TFhirPractitionerListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirPractitioner to the end of the list.
    }
    function Append : TFhirPractitioner;

    
    {@member AddItem
      Add an already existing FhirPractitioner to the end of the list.
    }
    procedure AddItem(value : TFhirPractitioner); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirPractitioner) : Integer;
    

    {@member Insert
      Insert FhirPractitioner before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirPractitioner;
    

    {@member InsertItem
       Insert an existing FhirPractitioner before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirPractitioner);
    
    {@member Item
       Get the iIndexth FhirPractitioner. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirPractitioner. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirPractitioner);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirPractitioner;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirPractitioners[index : Integer] : TFhirPractitioner read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcedurePerformer : TFhirElement
    Limited to 'real' people rather than equipment.
  }
  {!.Net HL7Connect.Fhir.ProcedurePerformer}
  TFhirProcedurePerformer = class (TFhirBackboneElement)
  private
    FActor : TFhirReference{Resource};
    FRole : TFhirCodeableConcept;
    Procedure SetActor(value : TFhirReference{Resource});
    Procedure SetRole(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcedurePerformer; overload;
    function Clone : TFhirProcedurePerformer; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member actor
      Typed access to The practitioner who was involved in the procedure. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      The practitioner who was involved in the procedure.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

    {@member role
      Typed access to For example: surgeon, anaethetist, endoscopist. (defined for API consistency)
    }
    property role : TFhirCodeableConcept read FRole write SetRole;
    {@member roleElement
      For example: surgeon, anaethetist, endoscopist.
    }
    property roleElement : TFhirCodeableConcept read FRole write SetRole;

  end;


  TFhirProcedurePerformerListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcedurePerformerList;
    function GetCurrent : TFhirProcedurePerformer;
  public
    Constructor Create(list : TFhirProcedurePerformerList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcedurePerformer read GetCurrent;
  end;


  {@Class TFhirProcedurePerformerList
    A list of FhirProcedurePerformer
  }
  {!.Net HL7Connect.Fhir.ProcedurePerformerList}
  TFhirProcedurePerformerList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcedurePerformer;
    procedure SetItemN(index : Integer; value : TFhirProcedurePerformer);
  public
    {!script hide}
    function Link : TFhirProcedurePerformerList; Overload;
    function Clone : TFhirProcedurePerformerList; Overload;
    function GetEnumerator : TFhirProcedurePerformerListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcedurePerformer to the end of the list.
    }
    function Append : TFhirProcedurePerformer;

    
    {@member AddItem
      Add an already existing FhirProcedurePerformer to the end of the list.
    }
    procedure AddItem(value : TFhirProcedurePerformer); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcedurePerformer) : Integer;
    

    {@member Insert
      Insert FhirProcedurePerformer before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcedurePerformer;
    

    {@member InsertItem
       Insert an existing FhirProcedurePerformer before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcedurePerformer);
    
    {@member Item
       Get the iIndexth FhirProcedurePerformer. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcedurePerformer. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcedurePerformer);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcedurePerformer;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcedurePerformers[index : Integer] : TFhirProcedurePerformer read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcedureFocalDevice : TFhirElement
    A device that is implanted, removed or otherwise manipulated (calibration, battery replacement, fitting a prosthesis, attaching a wound-vac, etc.) as a focal portion of the Procedure.
  }
  {!.Net HL7Connect.Fhir.ProcedureFocalDevice}
  TFhirProcedureFocalDevice = class (TFhirBackboneElement)
  private
    FAction : TFhirCodeableConcept;
    FManipulated : TFhirReference{TFhirDevice};
    Procedure SetAction(value : TFhirCodeableConcept);
    Procedure SetManipulated(value : TFhirReference{TFhirDevice});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcedureFocalDevice; overload;
    function Clone : TFhirProcedureFocalDevice; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member action
      Typed access to The kind of change that happened to the device during the procedure. (defined for API consistency)
    }
    property action : TFhirCodeableConcept read FAction write SetAction;
    {@member actionElement
      The kind of change that happened to the device during the procedure.
    }
    property actionElement : TFhirCodeableConcept read FAction write SetAction;

    {@member manipulated
      Typed access to The device that was manipulated (changed) during the procedure. (defined for API consistency)
    }
    property manipulated : TFhirReference{TFhirDevice} read FManipulated write SetManipulated;
    {@member manipulatedElement
      The device that was manipulated (changed) during the procedure.
    }
    property manipulatedElement : TFhirReference{TFhirDevice} read FManipulated write SetManipulated;

  end;


  TFhirProcedureFocalDeviceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcedureFocalDeviceList;
    function GetCurrent : TFhirProcedureFocalDevice;
  public
    Constructor Create(list : TFhirProcedureFocalDeviceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcedureFocalDevice read GetCurrent;
  end;


  {@Class TFhirProcedureFocalDeviceList
    A list of FhirProcedureFocalDevice
  }
  {!.Net HL7Connect.Fhir.ProcedureFocalDeviceList}
  TFhirProcedureFocalDeviceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcedureFocalDevice;
    procedure SetItemN(index : Integer; value : TFhirProcedureFocalDevice);
  public
    {!script hide}
    function Link : TFhirProcedureFocalDeviceList; Overload;
    function Clone : TFhirProcedureFocalDeviceList; Overload;
    function GetEnumerator : TFhirProcedureFocalDeviceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcedureFocalDevice to the end of the list.
    }
    function Append : TFhirProcedureFocalDevice;

    
    {@member AddItem
      Add an already existing FhirProcedureFocalDevice to the end of the list.
    }
    procedure AddItem(value : TFhirProcedureFocalDevice); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcedureFocalDevice) : Integer;
    

    {@member Insert
      Insert FhirProcedureFocalDevice before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcedureFocalDevice;
    

    {@member InsertItem
       Insert an existing FhirProcedureFocalDevice before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcedureFocalDevice);
    
    {@member Item
       Get the iIndexth FhirProcedureFocalDevice. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcedureFocalDevice. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcedureFocalDevice);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcedureFocalDevice;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcedureFocalDevices[index : Integer] : TFhirProcedureFocalDevice read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcedure : TFhirDomainResource
    An action that is or was performed on a patient. This can be a physical intervention like an operation, or less invasive like counseling or hypnotherapy.
  }
  {!.Net HL7Connect.Fhir.Procedure}
  TFhirProcedure = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FStatus : TFhirEnum;
    FCategory : TFhirCodeableConcept;
    FCode : TFhirCodeableConcept;
    FNotPerformed : TFhirBoolean;
    FreasonNotPerformedList : TFhirCodeableConceptList;
    FbodySiteList : TFhirCodeableConceptList;
    FReason : TFhirType;
    FperformerList : TFhirProcedurePerformerList;
    FPerformed : TFhirType;
    FEncounter : TFhirReference{TFhirEncounter};
    FLocation : TFhirReference{TFhirLocation};
    FOutcome : TFhirCodeableConcept;
    FreportList : TFhirReferenceList{TFhirDiagnosticReport};
    FcomplicationList : TFhirCodeableConceptList;
    FfollowUpList : TFhirCodeableConceptList;
    FRequest : TFhirReference{Resource};
    FnotesList : TFhirAnnotationList;
    FfocalDeviceList : TFhirProcedureFocalDeviceList;
    FusedList : TFhirReferenceList{Resource};
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirProcedureStatus;
    Procedure SetStatusST(value : TFhirProcedureStatus);
    Procedure SetCategory(value : TFhirCodeableConcept);
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetNotPerformed(value : TFhirBoolean);
    Function GetNotPerformedST : Boolean;
    Procedure SetNotPerformedST(value : Boolean);
    function GetReasonNotPerformedList : TFhirCodeableConceptList;
    function GetHasReasonNotPerformedList : Boolean;
    function GetBodySiteList : TFhirCodeableConceptList;
    function GetHasBodySiteList : Boolean;
    Procedure SetReason(value : TFhirType);
    function GetPerformerList : TFhirProcedurePerformerList;
    function GetHasPerformerList : Boolean;
    Procedure SetPerformed(value : TFhirType);
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    Procedure SetOutcome(value : TFhirCodeableConcept);
    function GetReportList : TFhirReferenceList{TFhirDiagnosticReport};
    function GetHasReportList : Boolean;
    function GetComplicationList : TFhirCodeableConceptList;
    function GetHasComplicationList : Boolean;
    function GetFollowUpList : TFhirCodeableConceptList;
    function GetHasFollowUpList : Boolean;
    Procedure SetRequest(value : TFhirReference{Resource});
    function GetNotesList : TFhirAnnotationList;
    function GetHasNotesList : Boolean;
    function GetFocalDeviceList : TFhirProcedureFocalDeviceList;
    function GetHasFocalDeviceList : Boolean;
    function GetUsedList : TFhirReferenceList{Resource};
    function GetHasUsedList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcedure; overload;
    function Clone : TFhirProcedure; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this procedure that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to The person, animal or group on which the procedure was performed. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The person, animal or group on which the procedure was performed.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member status
      A code specifying the state of the procedure. Generally this will be in-progress or completed state.
    }
    property status : TFhirProcedureStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member category
      Typed access to A code that classifies the procedure for searching, sorting and display purposes (e.g. "Surgical Procedure"). (defined for API consistency)
    }
    property category : TFhirCodeableConcept read FCategory write SetCategory;
    {@member categoryElement
      A code that classifies the procedure for searching, sorting and display purposes (e.g. "Surgical Procedure").
    }
    property categoryElement : TFhirCodeableConcept read FCategory write SetCategory;

    {@member code
      Typed access to The specific procedure that is performed. Use text if the exact nature of the procedure cannot be coded (e.g. "Laparoscopic Appendectomy"). (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      The specific procedure that is performed. Use text if the exact nature of the procedure cannot be coded (e.g. "Laparoscopic Appendectomy").
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member notPerformed
      Typed access to Set this to true if the record is saying that the procedure was NOT performed.
    }
    property notPerformed : Boolean read GetNotPerformedST write SetNotPerformedST;
    {@member notPerformedElement
      Set this to true if the record is saying that the procedure was NOT performed.
    }
    property notPerformedElement : TFhirBoolean read FNotPerformed write SetNotPerformed;

    {@member reasonNotPerformedList
      A code indicating why the procedure was not performed.
    }
    property reasonNotPerformedList : TFhirCodeableConceptList read GetReasonNotPerformedList;
    property hasReasonNotPerformedList : boolean read GetHasReasonNotPerformedList;

    {@member bodySiteList
      Detailed and structured anatomical location information. Multiple locations are allowed - e.g. multiple punch biopsies of a lesion.
    }
    property bodySiteList : TFhirCodeableConceptList read GetBodySiteList;
    property hasBodySiteList : boolean read GetHasBodySiteList;

    {@member reason
      Typed access to The reason why the procedure was performed. This may be due to a Condition, may be coded entity of some type, or may simply be present as text. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      The reason why the procedure was performed. This may be due to a Condition, may be coded entity of some type, or may simply be present as text.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member performerList
      Limited to 'real' people rather than equipment.
    }
    property performerList : TFhirProcedurePerformerList read GetPerformerList;
    property hasPerformerList : boolean read GetHasPerformerList;

    {@member performed
      Typed access to The date(time)/period over which the procedure was performed. Allows a period to support complex procedures that span more than one date, and also allows for the length of the procedure to be captured. (defined for API consistency)
    }
    property performed : TFhirType read FPerformed write SetPerformed;
    {@member performedElement
      The date(time)/period over which the procedure was performed. Allows a period to support complex procedures that span more than one date, and also allows for the length of the procedure to be captured.
    }
    property performedElement : TFhirType read FPerformed write SetPerformed;

    {@member encounter
      Typed access to The encounter during which the procedure was performed. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter during which the procedure was performed.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member location
      Typed access to The location where the procedure actually happened.  E.g. a newborn at home, a tracheostomy at a restaurant. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      The location where the procedure actually happened.  E.g. a newborn at home, a tracheostomy at a restaurant.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member outcome
      Typed access to The outcome of the procedure - did it resolve reasons for the procedure being performed? (defined for API consistency)
    }
    property outcome : TFhirCodeableConcept read FOutcome write SetOutcome;
    {@member outcomeElement
      The outcome of the procedure - did it resolve reasons for the procedure being performed?
    }
    property outcomeElement : TFhirCodeableConcept read FOutcome write SetOutcome;

    {@member reportList
      This could be a histology result, pathology report, surgical report, etc..
    }
    property reportList : TFhirReferenceList{TFhirDiagnosticReport} read GetReportList;
    property hasReportList : boolean read GetHasReportList;

    {@member complicationList
      Any complications that occurred during the procedure, or in the immediate post-performance period. These are generally tracked separately from the notes, which will typically describe the procedure itself rather than any 'post procedure' issues.
    }
    property complicationList : TFhirCodeableConceptList read GetComplicationList;
    property hasComplicationList : boolean read GetHasComplicationList;

    {@member followUpList
      If the procedure required specific follow up - e.g. removal of sutures. The followup may be represented as a simple note, or could potentially be more complex in which case the CarePlan resource can be used.
    }
    property followUpList : TFhirCodeableConceptList read GetFollowUpList;
    property hasFollowUpList : boolean read GetHasFollowUpList;

    {@member request
      Typed access to A reference to a resource that contains details of the request for this procedure. (defined for API consistency)
    }
    property request : TFhirReference{Resource} read FRequest write SetRequest;
    {@member requestElement
      A reference to a resource that contains details of the request for this procedure.
    }
    property requestElement : TFhirReference{Resource} read FRequest write SetRequest;

    {@member notesList
      Any other notes about the procedure.  E.g. the operative notes.
    }
    property notesList : TFhirAnnotationList read GetNotesList;
    property hasNotesList : boolean read GetHasNotesList;

    {@member focalDeviceList
      A device that is implanted, removed or otherwise manipulated (calibration, battery replacement, fitting a prosthesis, attaching a wound-vac, etc.) as a focal portion of the Procedure.
    }
    property focalDeviceList : TFhirProcedureFocalDeviceList read GetFocalDeviceList;
    property hasFocalDeviceList : boolean read GetHasFocalDeviceList;

    {@member usedList
      Identifies medications, devices and any other substance used as part of the procedure.
    }
    property usedList : TFhirReferenceList{Resource} read GetUsedList;
    property hasUsedList : boolean read GetHasUsedList;

  end;


  TFhirProcedureListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcedureList;
    function GetCurrent : TFhirProcedure;
  public
    Constructor Create(list : TFhirProcedureList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcedure read GetCurrent;
  end;


  {@Class TFhirProcedureList
    A list of FhirProcedure
  }
  {!.Net HL7Connect.Fhir.ProcedureList}
  TFhirProcedureList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcedure;
    procedure SetItemN(index : Integer; value : TFhirProcedure);
  public
    {!script hide}
    function Link : TFhirProcedureList; Overload;
    function Clone : TFhirProcedureList; Overload;
    function GetEnumerator : TFhirProcedureListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcedure to the end of the list.
    }
    function Append : TFhirProcedure;

    
    {@member AddItem
      Add an already existing FhirProcedure to the end of the list.
    }
    procedure AddItem(value : TFhirProcedure); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcedure) : Integer;
    

    {@member Insert
      Insert FhirProcedure before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcedure;
    

    {@member InsertItem
       Insert an existing FhirProcedure before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcedure);
    
    {@member Item
       Get the iIndexth FhirProcedure. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcedure. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcedure);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcedure;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcedures[index : Integer] : TFhirProcedure read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcedureRequest : TFhirDomainResource
    A request for a procedure to be performed. May be a proposal or an order.
  }
  {!.Net HL7Connect.Fhir.ProcedureRequest}
  TFhirProcedureRequest = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FSubject : TFhirReference{Resource};
    FCode : TFhirCodeableConcept;
    FbodySiteList : TFhirCodeableConceptList;
    FReason : TFhirType;
    FScheduled : TFhirType;
    FEncounter : TFhirReference{TFhirEncounter};
    FPerformer : TFhirReference{Resource};
    FStatus : TFhirEnum;
    FnotesList : TFhirAnnotationList;
    FAsNeeded : TFhirType;
    FOrderedOn : TFhirDateTime;
    FOrderer : TFhirReference{Resource};
    FPriority : TFhirEnum;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetCode(value : TFhirCodeableConcept);
    function GetBodySiteList : TFhirCodeableConceptList;
    function GetHasBodySiteList : Boolean;
    Procedure SetReason(value : TFhirType);
    Procedure SetScheduled(value : TFhirType);
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetPerformer(value : TFhirReference{Resource});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirProcedureRequestStatus;
    Procedure SetStatusST(value : TFhirProcedureRequestStatus);
    function GetNotesList : TFhirAnnotationList;
    function GetHasNotesList : Boolean;
    Procedure SetAsNeeded(value : TFhirType);
    Procedure SetOrderedOn(value : TFhirDateTime);
    Function GetOrderedOnST : TDateAndTime;
    Procedure SetOrderedOnST(value : TDateAndTime);
    Procedure SetOrderer(value : TFhirReference{Resource});
    Procedure SetPriority(value : TFhirEnum);
    Function GetPriorityST : TFhirProcedureRequestPriority;
    Procedure SetPriorityST(value : TFhirProcedureRequestPriority);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcedureRequest; overload;
    function Clone : TFhirProcedureRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifiers assigned to this order by the order or by the receiver.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member subject
      Typed access to The person, animal or group that should receive the procedure. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The person, animal or group that should receive the procedure.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member code
      Typed access to The specific procedure that is ordered. Use text if the exact nature of the procedure cannot be coded. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      The specific procedure that is ordered. Use text if the exact nature of the procedure cannot be coded.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member bodySiteList
      Indicates the sites on the subject's body where the procedure should be performed (I.e. the target sites).
    }
    property bodySiteList : TFhirCodeableConceptList read GetBodySiteList;
    property hasBodySiteList : boolean read GetHasBodySiteList;

    {@member reason
      Typed access to The reason why the procedure is being proposed or ordered. This procedure request may be motivated by a Condition for instance. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      The reason why the procedure is being proposed or ordered. This procedure request may be motivated by a Condition for instance.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member scheduled
      Typed access to The timing schedule for the proposed or ordered procedure. The Schedule data type allows many different expressions.  E.g. "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013". (defined for API consistency)
    }
    property scheduled : TFhirType read FScheduled write SetScheduled;
    {@member scheduledElement
      The timing schedule for the proposed or ordered procedure. The Schedule data type allows many different expressions.  E.g. "Every 8 hours"; "Three times a day"; "1/2 an hour before breakfast for 10 days from 23-Dec 2011:"; "15 Oct 2013, 17 Oct 2013 and 1 Nov 2013".
    }
    property scheduledElement : TFhirType read FScheduled write SetScheduled;

    {@member encounter
      Typed access to The encounter within which the procedure proposal or request was created. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter within which the procedure proposal or request was created.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member performer
      Typed access to For example, the surgeon, anaethetist, endoscopist, etc. (defined for API consistency)
    }
    property performer : TFhirReference{Resource} read FPerformer write SetPerformer;
    {@member performerElement
      For example, the surgeon, anaethetist, endoscopist, etc.
    }
    property performerElement : TFhirReference{Resource} read FPerformer write SetPerformer;

    {@member status
      The status of the order.
    }
    property status : TFhirProcedureRequestStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member notesList
      Any other notes associated with this proposal or order - e.g. provider instructions.
    }
    property notesList : TFhirAnnotationList read GetNotesList;
    property hasNotesList : boolean read GetHasNotesList;

    {@member asNeeded
      Typed access to If a CodeableConcept is present, it indicates the pre-condition for performing the procedure. (defined for API consistency)
    }
    property asNeeded : TFhirType read FAsNeeded write SetAsNeeded;
    {@member asNeededElement
      If a CodeableConcept is present, it indicates the pre-condition for performing the procedure.
    }
    property asNeededElement : TFhirType read FAsNeeded write SetAsNeeded;

    {@member orderedOn
      Typed access to The time when the request was made.
    }
    property orderedOn : TDateAndTime read GetOrderedOnST write SetOrderedOnST;
    {@member orderedOnElement
      The time when the request was made.
    }
    property orderedOnElement : TFhirDateTime read FOrderedOn write SetOrderedOn;

    {@member orderer
      Typed access to The healthcare professional responsible for proposing or ordering the procedure. (defined for API consistency)
    }
    property orderer : TFhirReference{Resource} read FOrderer write SetOrderer;
    {@member ordererElement
      The healthcare professional responsible for proposing or ordering the procedure.
    }
    property ordererElement : TFhirReference{Resource} read FOrderer write SetOrderer;

    {@member priority
      The clinical priority associated with this order.
    }
    property priority : TFhirProcedureRequestPriority read GetPriorityST write SetPriorityST;
    property priorityElement : TFhirEnum read FPriority write SetPriority;

  end;


  TFhirProcedureRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcedureRequestList;
    function GetCurrent : TFhirProcedureRequest;
  public
    Constructor Create(list : TFhirProcedureRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcedureRequest read GetCurrent;
  end;


  {@Class TFhirProcedureRequestList
    A list of FhirProcedureRequest
  }
  {!.Net HL7Connect.Fhir.ProcedureRequestList}
  TFhirProcedureRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcedureRequest;
    procedure SetItemN(index : Integer; value : TFhirProcedureRequest);
  public
    {!script hide}
    function Link : TFhirProcedureRequestList; Overload;
    function Clone : TFhirProcedureRequestList; Overload;
    function GetEnumerator : TFhirProcedureRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcedureRequest to the end of the list.
    }
    function Append : TFhirProcedureRequest;

    
    {@member AddItem
      Add an already existing FhirProcedureRequest to the end of the list.
    }
    procedure AddItem(value : TFhirProcedureRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcedureRequest) : Integer;
    

    {@member Insert
      Insert FhirProcedureRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcedureRequest;
    

    {@member InsertItem
       Insert an existing FhirProcedureRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcedureRequest);
    
    {@member Item
       Get the iIndexth FhirProcedureRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcedureRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcedureRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcedureRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcedureRequests[index : Integer] : TFhirProcedureRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcessRequestItem : TFhirElement
    List of top level items to be re-adjudicated, if none specified then the entire submission is re-adjudicated.
  }
  {!.Net HL7Connect.Fhir.ProcessRequestItem}
  TFhirProcessRequestItem = class (TFhirBackboneElement)
  private
    FSequenceLinkId : TFhirInteger;
    Procedure SetSequenceLinkId(value : TFhirInteger);
    Function GetSequenceLinkIdST : String;
    Procedure SetSequenceLinkIdST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcessRequestItem; overload;
    function Clone : TFhirProcessRequestItem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member sequenceLinkId
      Typed access to A service line number.
    }
    property sequenceLinkId : String read GetSequenceLinkIdST write SetSequenceLinkIdST;
    {@member sequenceLinkIdElement
      A service line number.
    }
    property sequenceLinkIdElement : TFhirInteger read FSequenceLinkId write SetSequenceLinkId;

  end;


  TFhirProcessRequestItemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcessRequestItemList;
    function GetCurrent : TFhirProcessRequestItem;
  public
    Constructor Create(list : TFhirProcessRequestItemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcessRequestItem read GetCurrent;
  end;


  {@Class TFhirProcessRequestItemList
    A list of FhirProcessRequestItem
  }
  {!.Net HL7Connect.Fhir.ProcessRequestItemList}
  TFhirProcessRequestItemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcessRequestItem;
    procedure SetItemN(index : Integer; value : TFhirProcessRequestItem);
  public
    {!script hide}
    function Link : TFhirProcessRequestItemList; Overload;
    function Clone : TFhirProcessRequestItemList; Overload;
    function GetEnumerator : TFhirProcessRequestItemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcessRequestItem to the end of the list.
    }
    function Append : TFhirProcessRequestItem;

    
    {@member AddItem
      Add an already existing FhirProcessRequestItem to the end of the list.
    }
    procedure AddItem(value : TFhirProcessRequestItem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcessRequestItem) : Integer;
    

    {@member Insert
      Insert FhirProcessRequestItem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcessRequestItem;
    

    {@member InsertItem
       Insert an existing FhirProcessRequestItem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcessRequestItem);
    
    {@member Item
       Get the iIndexth FhirProcessRequestItem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcessRequestItem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcessRequestItem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcessRequestItem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcessRequestItems[index : Integer] : TFhirProcessRequestItem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcessRequest : TFhirDomainResource
    This resource provides the target, request and response, and action details for an action to be performed by the target on or about existing resources.
  }
  {!.Net HL7Connect.Fhir.ProcessRequest}
  TFhirProcessRequest = class (TFhirDomainResource)
  private
    FAction : TFhirEnum;
    FidentifierList : TFhirIdentifierList;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FTarget : TFhirReference{TFhirOrganization};
    FProvider : TFhirReference{TFhirPractitioner};
    FOrganization : TFhirReference{TFhirOrganization};
    FRequest : TFhirReference{TFhirReference};
    FResponse : TFhirReference{TFhirReference};
    FNullify : TFhirBoolean;
    FReference : TFhirString;
    FitemList : TFhirProcessRequestItemList;
    FincludeList : TFhirStringList;
    FexcludeList : TFhirStringList;
    FPeriod : TFhirPeriod;
    Procedure SetAction(value : TFhirEnum);
    Function GetActionST : TFhirActionlist;
    Procedure SetActionST(value : TFhirActionlist);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetTarget(value : TFhirReference{TFhirOrganization});
    Procedure SetProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequest(value : TFhirReference{TFhirReference});
    Procedure SetResponse(value : TFhirReference{TFhirReference});
    Procedure SetNullify(value : TFhirBoolean);
    Function GetNullifyST : Boolean;
    Procedure SetNullifyST(value : Boolean);
    Procedure SetReference(value : TFhirString);
    Function GetReferenceST : String;
    Procedure SetReferenceST(value : String);
    function GetItemList : TFhirProcessRequestItemList;
    function GetHasItemList : Boolean;
    function GetIncludeList : TFhirStringList;
    function GetHasIncludeList : Boolean;
    function GetExcludeList : TFhirStringList;
    function GetHasExcludeList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcessRequest; overload;
    function Clone : TFhirProcessRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member action
      The type of processing action being requested, for example Reversal, Readjudication, StatusRequest,PendedRequest.
    }
    property action : TFhirActionlist read GetActionST write SetActionST;
    property actionElement : TFhirEnum read FAction write SetAction;

    {@member identifierList
      The ProcessRequest business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when this resource was created.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when this resource was created.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member target
      Typed access to The organization which is the target of the request. (defined for API consistency)
    }
    property target : TFhirReference{TFhirOrganization} read FTarget write SetTarget;
    {@member targetElement
      The organization which is the target of the request.
    }
    property targetElement : TFhirReference{TFhirOrganization} read FTarget write SetTarget;

    {@member provider
      Typed access to The practitioner who is responsible for the action specified in thise request. (defined for API consistency)
    }
    property provider : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;
    {@member providerElement
      The practitioner who is responsible for the action specified in thise request.
    }
    property providerElement : TFhirReference{TFhirPractitioner} read FProvider write SetProvider;

    {@member organization
      Typed access to The organization which is responsible for the action speccified in thise request. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization which is responsible for the action speccified in thise request.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member request
      Typed access to Reference of resource which is the target or subject of this action. (defined for API consistency)
    }
    property request : TFhirReference{TFhirReference} read FRequest write SetRequest;
    {@member requestElement
      Reference of resource which is the target or subject of this action.
    }
    property requestElement : TFhirReference{TFhirReference} read FRequest write SetRequest;

    {@member response
      Typed access to Reference of a prior response to resource which is the target or subject of this action. (defined for API consistency)
    }
    property response : TFhirReference{TFhirReference} read FResponse write SetResponse;
    {@member responseElement
      Reference of a prior response to resource which is the target or subject of this action.
    }
    property responseElement : TFhirReference{TFhirReference} read FResponse write SetResponse;

    {@member nullify
      Typed access to If true remove all history excluding audit.
    }
    property nullify : Boolean read GetNullifyST write SetNullifyST;
    {@member nullifyElement
      If true remove all history excluding audit.
    }
    property nullifyElement : TFhirBoolean read FNullify write SetNullify;

    {@member reference
      Typed access to A reference to supply which authenticates the process.
    }
    property reference : String read GetReferenceST write SetReferenceST;
    {@member referenceElement
      A reference to supply which authenticates the process.
    }
    property referenceElement : TFhirString read FReference write SetReference;

    {@member itemList
      List of top level items to be re-adjudicated, if none specified then the entire submission is re-adjudicated.
    }
    property itemList : TFhirProcessRequestItemList read GetItemList;
    property hasItemList : boolean read GetHasItemList;

    {@member includeList
      Names of resource types to include.
    }
    property includeList : TFhirStringList read GetIncludeList;
    property hasIncludeList : boolean read GetHasIncludeList;

    {@member excludeList
      Names of resource types to exclude.
    }
    property excludeList : TFhirStringList read GetExcludeList;
    property hasExcludeList : boolean read GetHasExcludeList;

    {@member period
      Typed access to A period of time during which the fulfilling resources would have been created. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      A period of time during which the fulfilling resources would have been created.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirProcessRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcessRequestList;
    function GetCurrent : TFhirProcessRequest;
  public
    Constructor Create(list : TFhirProcessRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcessRequest read GetCurrent;
  end;


  {@Class TFhirProcessRequestList
    A list of FhirProcessRequest
  }
  {!.Net HL7Connect.Fhir.ProcessRequestList}
  TFhirProcessRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcessRequest;
    procedure SetItemN(index : Integer; value : TFhirProcessRequest);
  public
    {!script hide}
    function Link : TFhirProcessRequestList; Overload;
    function Clone : TFhirProcessRequestList; Overload;
    function GetEnumerator : TFhirProcessRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcessRequest to the end of the list.
    }
    function Append : TFhirProcessRequest;

    
    {@member AddItem
      Add an already existing FhirProcessRequest to the end of the list.
    }
    procedure AddItem(value : TFhirProcessRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcessRequest) : Integer;
    

    {@member Insert
      Insert FhirProcessRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcessRequest;
    

    {@member InsertItem
       Insert an existing FhirProcessRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcessRequest);
    
    {@member Item
       Get the iIndexth FhirProcessRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcessRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcessRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcessRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcessRequests[index : Integer] : TFhirProcessRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcessResponseNotes : TFhirElement
    Suite of processing note or additional requirements is the processing has been held.
  }
  {!.Net HL7Connect.Fhir.ProcessResponseNotes}
  TFhirProcessResponseNotes = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FText : TFhirString;
    Procedure SetType_(value : TFhirCoding);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcessResponseNotes; overload;
    function Clone : TFhirProcessResponseNotes; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The note purpose: Print/Display. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The note purpose: Print/Display.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member text
      Typed access to The note text.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      The note text.
    }
    property textElement : TFhirString read FText write SetText;

  end;


  TFhirProcessResponseNotesListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcessResponseNotesList;
    function GetCurrent : TFhirProcessResponseNotes;
  public
    Constructor Create(list : TFhirProcessResponseNotesList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcessResponseNotes read GetCurrent;
  end;


  {@Class TFhirProcessResponseNotesList
    A list of FhirProcessResponseNotes
  }
  {!.Net HL7Connect.Fhir.ProcessResponseNotesList}
  TFhirProcessResponseNotesList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcessResponseNotes;
    procedure SetItemN(index : Integer; value : TFhirProcessResponseNotes);
  public
    {!script hide}
    function Link : TFhirProcessResponseNotesList; Overload;
    function Clone : TFhirProcessResponseNotesList; Overload;
    function GetEnumerator : TFhirProcessResponseNotesListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcessResponseNotes to the end of the list.
    }
    function Append : TFhirProcessResponseNotes;

    
    {@member AddItem
      Add an already existing FhirProcessResponseNotes to the end of the list.
    }
    procedure AddItem(value : TFhirProcessResponseNotes); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcessResponseNotes) : Integer;
    

    {@member Insert
      Insert FhirProcessResponseNotes before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcessResponseNotes;
    

    {@member InsertItem
       Insert an existing FhirProcessResponseNotes before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcessResponseNotes);
    
    {@member Item
       Get the iIndexth FhirProcessResponseNotes. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcessResponseNotes. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcessResponseNotes);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcessResponseNotes;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcessResponseNotes[index : Integer] : TFhirProcessResponseNotes read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProcessResponse : TFhirDomainResource
    This resource provides processing status, errors and notes from the processing of a resource.
  }
  {!.Net HL7Connect.Fhir.ProcessResponse}
  TFhirProcessResponse = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FRequest : TFhirReference{TFhirReference};
    FOutcome : TFhirCoding;
    FDisposition : TFhirString;
    FRuleset : TFhirCoding;
    FOriginalRuleset : TFhirCoding;
    FCreated : TFhirDateTime;
    FOrganization : TFhirReference{TFhirOrganization};
    FRequestProvider : TFhirReference{TFhirPractitioner};
    FRequestOrganization : TFhirReference{TFhirOrganization};
    FForm : TFhirCoding;
    FnotesList : TFhirProcessResponseNotesList;
    FerrorList : TFhirCodingList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetRequest(value : TFhirReference{TFhirReference});
    Procedure SetOutcome(value : TFhirCoding);
    Procedure SetDisposition(value : TFhirString);
    Function GetDispositionST : String;
    Procedure SetDispositionST(value : String);
    Procedure SetRuleset(value : TFhirCoding);
    Procedure SetOriginalRuleset(value : TFhirCoding);
    Procedure SetCreated(value : TFhirDateTime);
    Function GetCreatedST : TDateAndTime;
    Procedure SetCreatedST(value : TDateAndTime);
    Procedure SetOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetRequestProvider(value : TFhirReference{TFhirPractitioner});
    Procedure SetRequestOrganization(value : TFhirReference{TFhirOrganization});
    Procedure SetForm(value : TFhirCoding);
    function GetNotesList : TFhirProcessResponseNotesList;
    function GetHasNotesList : Boolean;
    function GetErrorList : TFhirCodingList;
    function GetHasErrorList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProcessResponse; overload;
    function Clone : TFhirProcessResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      The Response business identifier.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member request
      Typed access to Original request resource reference. (defined for API consistency)
    }
    property request : TFhirReference{TFhirReference} read FRequest write SetRequest;
    {@member requestElement
      Original request resource reference.
    }
    property requestElement : TFhirReference{TFhirReference} read FRequest write SetRequest;

    {@member outcome
      Typed access to Transaction status: error, complete, held. (defined for API consistency)
    }
    property outcome : TFhirCoding read FOutcome write SetOutcome;
    {@member outcomeElement
      Transaction status: error, complete, held.
    }
    property outcomeElement : TFhirCoding read FOutcome write SetOutcome;

    {@member disposition
      Typed access to A description of the status of the adjudication or processing.
    }
    property disposition : String read GetDispositionST write SetDispositionST;
    {@member dispositionElement
      A description of the status of the adjudication or processing.
    }
    property dispositionElement : TFhirString read FDisposition write SetDisposition;

    {@member ruleset
      Typed access to The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources. (defined for API consistency)
    }
    property ruleset : TFhirCoding read FRuleset write SetRuleset;
    {@member rulesetElement
      The version of the style of resource contents. This should be mapped to the allowable profiles for this and supporting resources.
    }
    property rulesetElement : TFhirCoding read FRuleset write SetRuleset;

    {@member originalRuleset
      Typed access to The style (standard) and version of the original material which was converted into this resource. (defined for API consistency)
    }
    property originalRuleset : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;
    {@member originalRulesetElement
      The style (standard) and version of the original material which was converted into this resource.
    }
    property originalRulesetElement : TFhirCoding read FOriginalRuleset write SetOriginalRuleset;

    {@member created
      Typed access to The date when the enclosed suite of services were performed or completed.
    }
    property created : TDateAndTime read GetCreatedST write SetCreatedST;
    {@member createdElement
      The date when the enclosed suite of services were performed or completed.
    }
    property createdElement : TFhirDateTime read FCreated write SetCreated;

    {@member organization
      Typed access to The organization who produced this adjudicated response. (defined for API consistency)
    }
    property organization : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;
    {@member organizationElement
      The organization who produced this adjudicated response.
    }
    property organizationElement : TFhirReference{TFhirOrganization} read FOrganization write SetOrganization;

    {@member requestProvider
      Typed access to The practitioner who is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestProvider : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;
    {@member requestProviderElement
      The practitioner who is responsible for the services rendered to the patient.
    }
    property requestProviderElement : TFhirReference{TFhirPractitioner} read FRequestProvider write SetRequestProvider;

    {@member requestOrganization
      Typed access to The organization which is responsible for the services rendered to the patient. (defined for API consistency)
    }
    property requestOrganization : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;
    {@member requestOrganizationElement
      The organization which is responsible for the services rendered to the patient.
    }
    property requestOrganizationElement : TFhirReference{TFhirOrganization} read FRequestOrganization write SetRequestOrganization;

    {@member form
      Typed access to The form to be used for printing the content. (defined for API consistency)
    }
    property form : TFhirCoding read FForm write SetForm;
    {@member formElement
      The form to be used for printing the content.
    }
    property formElement : TFhirCoding read FForm write SetForm;

    {@member notesList
      Suite of processing note or additional requirements is the processing has been held.
    }
    property notesList : TFhirProcessResponseNotesList read GetNotesList;
    property hasNotesList : boolean read GetHasNotesList;

    {@member errorList
      Processing errors.
    }
    property errorList : TFhirCodingList read GetErrorList;
    property hasErrorList : boolean read GetHasErrorList;

  end;


  TFhirProcessResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProcessResponseList;
    function GetCurrent : TFhirProcessResponse;
  public
    Constructor Create(list : TFhirProcessResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProcessResponse read GetCurrent;
  end;


  {@Class TFhirProcessResponseList
    A list of FhirProcessResponse
  }
  {!.Net HL7Connect.Fhir.ProcessResponseList}
  TFhirProcessResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProcessResponse;
    procedure SetItemN(index : Integer; value : TFhirProcessResponse);
  public
    {!script hide}
    function Link : TFhirProcessResponseList; Overload;
    function Clone : TFhirProcessResponseList; Overload;
    function GetEnumerator : TFhirProcessResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProcessResponse to the end of the list.
    }
    function Append : TFhirProcessResponse;

    
    {@member AddItem
      Add an already existing FhirProcessResponse to the end of the list.
    }
    procedure AddItem(value : TFhirProcessResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProcessResponse) : Integer;
    

    {@member Insert
      Insert FhirProcessResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProcessResponse;
    

    {@member InsertItem
       Insert an existing FhirProcessResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProcessResponse);
    
    {@member Item
       Get the iIndexth FhirProcessResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProcessResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProcessResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProcessResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProcessResponses[index : Integer] : TFhirProcessResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProvenanceAgent : TFhirElement
    An agent takes a role in an activity such that the agent can be assigned some degree of responsibility for the activity taking place. An agent can be a person, an organization, software, or other entities that may be ascribed responsibility.
  }
  {!.Net HL7Connect.Fhir.ProvenanceAgent}
  TFhirProvenanceAgent = class (TFhirBackboneElement)
  private
    FRole : TFhirCoding;
    FActor : TFhirReference{Resource};
    FUserId : TFhirIdentifier;
    FrelatedAgentList : TFhirProvenanceAgentRelatedAgentList;
    Procedure SetRole(value : TFhirCoding);
    Procedure SetActor(value : TFhirReference{Resource});
    Procedure SetUserId(value : TFhirIdentifier);
    function GetRelatedAgentList : TFhirProvenanceAgentRelatedAgentList;
    function GetHasRelatedAgentList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProvenanceAgent; overload;
    function Clone : TFhirProvenanceAgent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member role
      Typed access to The function of the agent with respect to the activity. (defined for API consistency)
    }
    property role : TFhirCoding read FRole write SetRole;
    {@member roleElement
      The function of the agent with respect to the activity.
    }
    property roleElement : TFhirCoding read FRole write SetRole;

    {@member actor
      Typed access to The individual, device or organization that participated in the event. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      The individual, device or organization that participated in the event.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

    {@member userId
      Typed access to The identity of the agent as known by the authorization system. (defined for API consistency)
    }
    property userId : TFhirIdentifier read FUserId write SetUserId;
    {@member userIdElement
      The identity of the agent as known by the authorization system.
    }
    property userIdElement : TFhirIdentifier read FUserId write SetUserId;

    {@member relatedAgentList
      A relationship between two the agents referenced in this resource. This is defined to allow for explicit description of the delegation between agents.  For example, this human author used this device, or one person acted on another's behest.
    }
    property relatedAgentList : TFhirProvenanceAgentRelatedAgentList read GetRelatedAgentList;
    property hasRelatedAgentList : boolean read GetHasRelatedAgentList;

  end;


  TFhirProvenanceAgentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProvenanceAgentList;
    function GetCurrent : TFhirProvenanceAgent;
  public
    Constructor Create(list : TFhirProvenanceAgentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProvenanceAgent read GetCurrent;
  end;


  {@Class TFhirProvenanceAgentList
    A list of FhirProvenanceAgent
  }
  {!.Net HL7Connect.Fhir.ProvenanceAgentList}
  TFhirProvenanceAgentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProvenanceAgent;
    procedure SetItemN(index : Integer; value : TFhirProvenanceAgent);
  public
    {!script hide}
    function Link : TFhirProvenanceAgentList; Overload;
    function Clone : TFhirProvenanceAgentList; Overload;
    function GetEnumerator : TFhirProvenanceAgentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProvenanceAgent to the end of the list.
    }
    function Append : TFhirProvenanceAgent;

    
    {@member AddItem
      Add an already existing FhirProvenanceAgent to the end of the list.
    }
    procedure AddItem(value : TFhirProvenanceAgent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProvenanceAgent) : Integer;
    

    {@member Insert
      Insert FhirProvenanceAgent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProvenanceAgent;
    

    {@member InsertItem
       Insert an existing FhirProvenanceAgent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProvenanceAgent);
    
    {@member Item
       Get the iIndexth FhirProvenanceAgent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProvenanceAgent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProvenanceAgent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProvenanceAgent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProvenanceAgents[index : Integer] : TFhirProvenanceAgent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProvenanceAgentRelatedAgent : TFhirElement
    A relationship between two the agents referenced in this resource. This is defined to allow for explicit description of the delegation between agents.  For example, this human author used this device, or one person acted on another's behest.
  }
  {!.Net HL7Connect.Fhir.ProvenanceAgentRelatedAgent}
  TFhirProvenanceAgentRelatedAgent = class (TFhirBackboneElement)
  private
    FType_ : TFhirCodeableConcept;
    FTarget : TFhirUri;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetTarget(value : TFhirUri);
    Function GetTargetST : String;
    Procedure SetTargetST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProvenanceAgentRelatedAgent; overload;
    function Clone : TFhirProvenanceAgentRelatedAgent; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to The type of relationship between agents. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The type of relationship between agents.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member target
      Typed access to An internal reference to another agent listed in this provenance by its identifier.
    }
    property target : String read GetTargetST write SetTargetST;
    {@member targetElement
      An internal reference to another agent listed in this provenance by its identifier.
    }
    property targetElement : TFhirUri read FTarget write SetTarget;

  end;


  TFhirProvenanceAgentRelatedAgentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProvenanceAgentRelatedAgentList;
    function GetCurrent : TFhirProvenanceAgentRelatedAgent;
  public
    Constructor Create(list : TFhirProvenanceAgentRelatedAgentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProvenanceAgentRelatedAgent read GetCurrent;
  end;


  {@Class TFhirProvenanceAgentRelatedAgentList
    A list of FhirProvenanceAgentRelatedAgent
  }
  {!.Net HL7Connect.Fhir.ProvenanceAgentRelatedAgentList}
  TFhirProvenanceAgentRelatedAgentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProvenanceAgentRelatedAgent;
    procedure SetItemN(index : Integer; value : TFhirProvenanceAgentRelatedAgent);
  public
    {!script hide}
    function Link : TFhirProvenanceAgentRelatedAgentList; Overload;
    function Clone : TFhirProvenanceAgentRelatedAgentList; Overload;
    function GetEnumerator : TFhirProvenanceAgentRelatedAgentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProvenanceAgentRelatedAgent to the end of the list.
    }
    function Append : TFhirProvenanceAgentRelatedAgent;

    
    {@member AddItem
      Add an already existing FhirProvenanceAgentRelatedAgent to the end of the list.
    }
    procedure AddItem(value : TFhirProvenanceAgentRelatedAgent); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProvenanceAgentRelatedAgent) : Integer;
    

    {@member Insert
      Insert FhirProvenanceAgentRelatedAgent before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProvenanceAgentRelatedAgent;
    

    {@member InsertItem
       Insert an existing FhirProvenanceAgentRelatedAgent before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProvenanceAgentRelatedAgent);
    
    {@member Item
       Get the iIndexth FhirProvenanceAgentRelatedAgent. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProvenanceAgentRelatedAgent. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProvenanceAgentRelatedAgent);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProvenanceAgentRelatedAgent;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProvenanceAgentRelatedAgents[index : Integer] : TFhirProvenanceAgentRelatedAgent read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProvenanceEntity : TFhirElement
    An entity used in this activity.
  }
  {!.Net HL7Connect.Fhir.ProvenanceEntity}
  TFhirProvenanceEntity = class (TFhirBackboneElement)
  private
    FRole : TFhirEnum;
    FType_ : TFhirCoding;
    FReference : TFhirUri;
    FDisplay : TFhirString;
    FAgent : TFhirProvenanceAgent;
    Procedure SetRole(value : TFhirEnum);
    Function GetRoleST : TFhirProvenanceEntityRole;
    Procedure SetRoleST(value : TFhirProvenanceEntityRole);
    Procedure SetType_(value : TFhirCoding);
    Procedure SetReference(value : TFhirUri);
    Function GetReferenceST : String;
    Procedure SetReferenceST(value : String);
    Procedure SetDisplay(value : TFhirString);
    Function GetDisplayST : String;
    Procedure SetDisplayST(value : String);
    Procedure SetAgent(value : TFhirProvenanceAgent);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProvenanceEntity; overload;
    function Clone : TFhirProvenanceEntity; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member role
      How the entity was used during the activity.
    }
    property role : TFhirProvenanceEntityRole read GetRoleST write SetRoleST;
    property roleElement : TFhirEnum read FRole write SetRole;

    {@member type_
      Typed access to The type of the entity. If the entity is a resource, then this is a resource type. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      The type of the entity. If the entity is a resource, then this is a resource type.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member reference
      Typed access to Identity of the  Entity used. May be a logical or physical uri and maybe absolute or relative.
    }
    property reference : String read GetReferenceST write SetReferenceST;
    {@member referenceElement
      Identity of the  Entity used. May be a logical or physical uri and maybe absolute or relative.
    }
    property referenceElement : TFhirUri read FReference write SetReference;

    {@member display
      Typed access to Human-readable description of the entity.
    }
    property display : String read GetDisplayST write SetDisplayST;
    {@member displayElement
      Human-readable description of the entity.
    }
    property displayElement : TFhirString read FDisplay write SetDisplay;

    {@member agent
      Typed access to The entity is attributed to an agent to express the agent's responsibility for that entity, possibly along with other agents. This description can be understood as shorthand for saying that the agent was responsible for the activity which generated the entity. (defined for API consistency)
    }
    property agent : TFhirProvenanceAgent read FAgent write SetAgent;
    {@member agentElement
      The entity is attributed to an agent to express the agent's responsibility for that entity, possibly along with other agents. This description can be understood as shorthand for saying that the agent was responsible for the activity which generated the entity.
    }
    property agentElement : TFhirProvenanceAgent read FAgent write SetAgent;

  end;


  TFhirProvenanceEntityListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProvenanceEntityList;
    function GetCurrent : TFhirProvenanceEntity;
  public
    Constructor Create(list : TFhirProvenanceEntityList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProvenanceEntity read GetCurrent;
  end;


  {@Class TFhirProvenanceEntityList
    A list of FhirProvenanceEntity
  }
  {!.Net HL7Connect.Fhir.ProvenanceEntityList}
  TFhirProvenanceEntityList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProvenanceEntity;
    procedure SetItemN(index : Integer; value : TFhirProvenanceEntity);
  public
    {!script hide}
    function Link : TFhirProvenanceEntityList; Overload;
    function Clone : TFhirProvenanceEntityList; Overload;
    function GetEnumerator : TFhirProvenanceEntityListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProvenanceEntity to the end of the list.
    }
    function Append : TFhirProvenanceEntity;

    
    {@member AddItem
      Add an already existing FhirProvenanceEntity to the end of the list.
    }
    procedure AddItem(value : TFhirProvenanceEntity); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProvenanceEntity) : Integer;
    

    {@member Insert
      Insert FhirProvenanceEntity before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProvenanceEntity;
    

    {@member InsertItem
       Insert an existing FhirProvenanceEntity before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProvenanceEntity);
    
    {@member Item
       Get the iIndexth FhirProvenanceEntity. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProvenanceEntity. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProvenanceEntity);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProvenanceEntity;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProvenanceEntities[index : Integer] : TFhirProvenanceEntity read GetItemN write SetItemN; default;
  End;


  {@Class TFhirProvenance : TFhirDomainResource
    Provenance of a resource is a record that describes entities and processes involved in producing and delivering or otherwise influencing that resource. Provenance provides a critical foundation for assessing authenticity, enabling trust, and allowing reproducibility. Provenance assertions are a form of contextual metadata and can themselves become important records with their own provenance. Provenance statement indicates clinical significance in terms of confidence in authenticity, reliability, and trustworthiness, integrity, and stage in lifecycle (e.g. Document Completion - has the artifact been legally authenticated), all of which may impact security, privacy, and trust policies.
  }
  {!.Net HL7Connect.Fhir.Provenance}
  TFhirProvenance = class (TFhirDomainResource)
  private
    FtargetList : TFhirReferenceList{TFhirReference};
    FPeriod : TFhirPeriod;
    FRecorded : TFhirInstant;
    FreasonList : TFhirCodeableConceptList;
    FActivity : TFhirCodeableConcept;
    FLocation : TFhirReference{TFhirLocation};
    FpolicyList : TFhirUriList;
    FagentList : TFhirProvenanceAgentList;
    FentityList : TFhirProvenanceEntityList;
    FsignatureList : TFhirSignatureList;
    function GetTargetList : TFhirReferenceList{TFhirReference};
    function GetHasTargetList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
    Procedure SetRecorded(value : TFhirInstant);
    Function GetRecordedST : TDateAndTime;
    Procedure SetRecordedST(value : TDateAndTime);
    function GetReasonList : TFhirCodeableConceptList;
    function GetHasReasonList : Boolean;
    Procedure SetActivity(value : TFhirCodeableConcept);
    Procedure SetLocation(value : TFhirReference{TFhirLocation});
    function GetPolicyList : TFhirUriList;
    function GetHasPolicyList : Boolean;
    function GetAgentList : TFhirProvenanceAgentList;
    function GetHasAgentList : Boolean;
    function GetEntityList : TFhirProvenanceEntityList;
    function GetHasEntityList : Boolean;
    function GetSignatureList : TFhirSignatureList;
    function GetHasSignatureList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirProvenance; overload;
    function Clone : TFhirProvenance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member targetList
      The Reference(s) that were generated or updated by  the activity described in this resource. A provenance can point to more than one target if multiple resources were created/updated by the same activity.
    }
    property targetList : TFhirReferenceList{TFhirReference} read GetTargetList;
    property hasTargetList : boolean read GetHasTargetList;

    {@member period
      Typed access to The period during which the activity occurred. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period during which the activity occurred.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

    {@member recorded
      Typed access to The instant of time at which the activity was recorded.
    }
    property recorded : TDateAndTime read GetRecordedST write SetRecordedST;
    {@member recordedElement
      The instant of time at which the activity was recorded.
    }
    property recordedElement : TFhirInstant read FRecorded write SetRecorded;

    {@member reasonList
      The reason that the activity was taking place.
    }
    property reasonList : TFhirCodeableConceptList read GetReasonList;
    property hasReasonList : boolean read GetHasReasonList;

    {@member activity
      Typed access to An activity is something that occurs over a period of time and acts upon or with entities; it may include consuming, processing, transforming, modifying, relocating, using, or generating entities. (defined for API consistency)
    }
    property activity : TFhirCodeableConcept read FActivity write SetActivity;
    {@member activityElement
      An activity is something that occurs over a period of time and acts upon or with entities; it may include consuming, processing, transforming, modifying, relocating, using, or generating entities.
    }
    property activityElement : TFhirCodeableConcept read FActivity write SetActivity;

    {@member location
      Typed access to Where the activity occurred, if relevant. (defined for API consistency)
    }
    property location : TFhirReference{TFhirLocation} read FLocation write SetLocation;
    {@member locationElement
      Where the activity occurred, if relevant.
    }
    property locationElement : TFhirReference{TFhirLocation} read FLocation write SetLocation;

    {@member policyList
      Policy or plan the activity was defined by. Typically, a single activity may have multiple applicable policy documents, such as patient consent, guarantor funding, etc.
    }
    property policyList : TFhirUriList read GetPolicyList;
    property hasPolicyList : boolean read GetHasPolicyList;

    {@member agentList
      An agent takes a role in an activity such that the agent can be assigned some degree of responsibility for the activity taking place. An agent can be a person, an organization, software, or other entities that may be ascribed responsibility.
    }
    property agentList : TFhirProvenanceAgentList read GetAgentList;
    property hasAgentList : boolean read GetHasAgentList;

    {@member entityList
      An entity used in this activity.
    }
    property entityList : TFhirProvenanceEntityList read GetEntityList;
    property hasEntityList : boolean read GetHasEntityList;

    {@member signatureList
      A digital signature on the target Reference(s). The signer should match a Provenance.agent. The purpose of the signature is indicated.
    }
    property signatureList : TFhirSignatureList read GetSignatureList;
    property hasSignatureList : boolean read GetHasSignatureList;

  end;


  TFhirProvenanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirProvenanceList;
    function GetCurrent : TFhirProvenance;
  public
    Constructor Create(list : TFhirProvenanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirProvenance read GetCurrent;
  end;


  {@Class TFhirProvenanceList
    A list of FhirProvenance
  }
  {!.Net HL7Connect.Fhir.ProvenanceList}
  TFhirProvenanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirProvenance;
    procedure SetItemN(index : Integer; value : TFhirProvenance);
  public
    {!script hide}
    function Link : TFhirProvenanceList; Overload;
    function Clone : TFhirProvenanceList; Overload;
    function GetEnumerator : TFhirProvenanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirProvenance to the end of the list.
    }
    function Append : TFhirProvenance;

    
    {@member AddItem
      Add an already existing FhirProvenance to the end of the list.
    }
    procedure AddItem(value : TFhirProvenance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirProvenance) : Integer;
    

    {@member Insert
      Insert FhirProvenance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirProvenance;
    

    {@member InsertItem
       Insert an existing FhirProvenance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirProvenance);
    
    {@member Item
       Get the iIndexth FhirProvenance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirProvenance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirProvenance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirProvenance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirProvenances[index : Integer] : TFhirProvenance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireGroup : TFhirElement
    A collection of related questions (or further groupings of questions).
  }
  {!.Net HL7Connect.Fhir.QuestionnaireGroup}
  TFhirQuestionnaireGroup = class (TFhirBackboneElement)
  private
    FLinkId : TFhirString;
    FTitle : TFhirString;
    FconceptList : TFhirCodingList;
    FText : TFhirString;
    FRequired : TFhirBoolean;
    FRepeats : TFhirBoolean;
    FgroupList : TFhirQuestionnaireGroupList;
    FquestionList : TFhirQuestionnaireGroupQuestionList;
    Procedure SetLinkId(value : TFhirString);
    Function GetLinkIdST : String;
    Procedure SetLinkIdST(value : String);
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    function GetConceptList : TFhirCodingList;
    function GetHasConceptList : Boolean;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetRequired(value : TFhirBoolean);
    Function GetRequiredST : Boolean;
    Procedure SetRequiredST(value : Boolean);
    Procedure SetRepeats(value : TFhirBoolean);
    Function GetRepeatsST : Boolean;
    Procedure SetRepeatsST(value : Boolean);
    function GetGroupList : TFhirQuestionnaireGroupList;
    function GetHasGroupList : Boolean;
    function GetQuestionList : TFhirQuestionnaireGroupQuestionList;
    function GetHasQuestionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireGroup; overload;
    function Clone : TFhirQuestionnaireGroup; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member linkId
      Typed access to An identifier that is unique within the Questionnaire allowing linkage to the equivalent group in a QuestionnaireResponse resource.
    }
    property linkId : String read GetLinkIdST write SetLinkIdST;
    {@member linkIdElement
      An identifier that is unique within the Questionnaire allowing linkage to the equivalent group in a QuestionnaireResponse resource.
    }
    property linkIdElement : TFhirString read FLinkId write SetLinkId;

    {@member title
      Typed access to The human-readable name for this section of the questionnaire.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      The human-readable name for this section of the questionnaire.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member conceptList
      Identifies a how this group of questions is known in a particular terminology such as LOINC.
    }
    property conceptList : TFhirCodingList read GetConceptList;
    property hasConceptList : boolean read GetHasConceptList;

    {@member text
      Typed access to Additional text for the group, used for display purposes.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Additional text for the group, used for display purposes.
    }
    property textElement : TFhirString read FText write SetText;

    {@member required
      Typed access to If true, indicates that the group must be present and have required questions within it answered.  If false, the group may be skipped when answering the questionnaire.
    }
    property required : Boolean read GetRequiredST write SetRequiredST;
    {@member requiredElement
      If true, indicates that the group must be present and have required questions within it answered.  If false, the group may be skipped when answering the questionnaire.
    }
    property requiredElement : TFhirBoolean read FRequired write SetRequired;

    {@member repeats
      Typed access to Whether the group may occur multiple times in the instance, containing multiple sets of answers.
    }
    property repeats : Boolean read GetRepeatsST write SetRepeatsST;
    {@member repeatsElement
      Whether the group may occur multiple times in the instance, containing multiple sets of answers.
    }
    property repeatsElement : TFhirBoolean read FRepeats write SetRepeats;

    {@member groupList
      A sub-group within a group. The ordering of groups within this group is relevant.
    }
    property groupList : TFhirQuestionnaireGroupList read GetGroupList;
    property hasGroupList : boolean read GetHasGroupList;

    {@member questionList
      Set of questions within this group. The order of questions within the group is relevant.
    }
    property questionList : TFhirQuestionnaireGroupQuestionList read GetQuestionList;
    property hasQuestionList : boolean read GetHasQuestionList;

  end;


  TFhirQuestionnaireGroupListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireGroupList;
    function GetCurrent : TFhirQuestionnaireGroup;
  public
    Constructor Create(list : TFhirQuestionnaireGroupList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireGroup read GetCurrent;
  end;


  {@Class TFhirQuestionnaireGroupList
    A list of FhirQuestionnaireGroup
  }
  {!.Net HL7Connect.Fhir.QuestionnaireGroupList}
  TFhirQuestionnaireGroupList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireGroup;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireGroup);
  public
    {!script hide}
    function Link : TFhirQuestionnaireGroupList; Overload;
    function Clone : TFhirQuestionnaireGroupList; Overload;
    function GetEnumerator : TFhirQuestionnaireGroupListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireGroup to the end of the list.
    }
    function Append : TFhirQuestionnaireGroup;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireGroup to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireGroup); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireGroup) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireGroup before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireGroup;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireGroup before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireGroup);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireGroup. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireGroup. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireGroup);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireGroup;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireGroups[index : Integer] : TFhirQuestionnaireGroup read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireGroupQuestion : TFhirElement
    Set of questions within this group. The order of questions within the group is relevant.
  }
  {!.Net HL7Connect.Fhir.QuestionnaireGroupQuestion}
  TFhirQuestionnaireGroupQuestion = class (TFhirBackboneElement)
  private
    FLinkId : TFhirString;
    FconceptList : TFhirCodingList;
    FText : TFhirString;
    FType_ : TFhirEnum;
    FRequired : TFhirBoolean;
    FRepeats : TFhirBoolean;
    FOptions : TFhirReference{TFhirValueSet};
    FoptionList : TFhirCodingList;
    FgroupList : TFhirQuestionnaireGroupList;
    Procedure SetLinkId(value : TFhirString);
    Function GetLinkIdST : String;
    Procedure SetLinkIdST(value : String);
    function GetConceptList : TFhirCodingList;
    function GetHasConceptList : Boolean;
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirAnswerFormat;
    Procedure SetType_ST(value : TFhirAnswerFormat);
    Procedure SetRequired(value : TFhirBoolean);
    Function GetRequiredST : Boolean;
    Procedure SetRequiredST(value : Boolean);
    Procedure SetRepeats(value : TFhirBoolean);
    Function GetRepeatsST : Boolean;
    Procedure SetRepeatsST(value : Boolean);
    Procedure SetOptions(value : TFhirReference{TFhirValueSet});
    function GetOptionList : TFhirCodingList;
    function GetHasOptionList : Boolean;
    function GetGroupList : TFhirQuestionnaireGroupList;
    function GetHasGroupList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireGroupQuestion; overload;
    function Clone : TFhirQuestionnaireGroupQuestion; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member linkId
      Typed access to An identifier that is unique within the questionnaire allowing linkage to the equivalent group in a [[[QuestionnaireResponse]]] resource.
    }
    property linkId : String read GetLinkIdST write SetLinkIdST;
    {@member linkIdElement
      An identifier that is unique within the questionnaire allowing linkage to the equivalent group in a [[[QuestionnaireResponse]]] resource.
    }
    property linkIdElement : TFhirString read FLinkId write SetLinkId;

    {@member conceptList
      Identifies a how this question is known in a particular terminology such as LOINC.
    }
    property conceptList : TFhirCodingList read GetConceptList;
    property hasConceptList : boolean read GetHasConceptList;

    {@member text
      Typed access to The actual question as shown to the user to prompt them for an answer.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      The actual question as shown to the user to prompt them for an answer.
    }
    property textElement : TFhirString read FText write SetText;

    {@member type_
      The expected format of the answer, e.g. the type of input (string, integer) or whether a (multiple) choice is expected.
    }
    property type_ : TFhirAnswerFormat read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member required
      Typed access to If true, indicates that the question must be answered and have required groups within it also present.  If false, the question and any contained groups may be skipped when answering the questionnaire.
    }
    property required : Boolean read GetRequiredST write SetRequiredST;
    {@member requiredElement
      If true, indicates that the question must be answered and have required groups within it also present.  If false, the question and any contained groups may be skipped when answering the questionnaire.
    }
    property requiredElement : TFhirBoolean read FRequired write SetRequired;

    {@member repeats
      Typed access to If true, the question may have more than one answer.
    }
    property repeats : Boolean read GetRepeatsST write SetRepeatsST;
    {@member repeatsElement
      If true, the question may have more than one answer.
    }
    property repeatsElement : TFhirBoolean read FRepeats write SetRepeats;

    {@member options
      Typed access to Reference to a value set containing a list of codes representing permitted answers for the question. (defined for API consistency)
    }
    property options : TFhirReference{TFhirValueSet} read FOptions write SetOptions;
    {@member optionsElement
      Reference to a value set containing a list of codes representing permitted answers for the question.
    }
    property optionsElement : TFhirReference{TFhirValueSet} read FOptions write SetOptions;

    {@member optionList
      For a "choice" question, identifies one of the permitted answers for the question.
    }
    property optionList : TFhirCodingList read GetOptionList;
    property hasOptionList : boolean read GetHasOptionList;

    {@member groupList
      Nested group, containing nested question for this question. The order of groups within the question is relevant.
    }
    property groupList : TFhirQuestionnaireGroupList read GetGroupList;
    property hasGroupList : boolean read GetHasGroupList;

  end;


  TFhirQuestionnaireGroupQuestionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireGroupQuestionList;
    function GetCurrent : TFhirQuestionnaireGroupQuestion;
  public
    Constructor Create(list : TFhirQuestionnaireGroupQuestionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireGroupQuestion read GetCurrent;
  end;


  {@Class TFhirQuestionnaireGroupQuestionList
    A list of FhirQuestionnaireGroupQuestion
  }
  {!.Net HL7Connect.Fhir.QuestionnaireGroupQuestionList}
  TFhirQuestionnaireGroupQuestionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireGroupQuestion;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireGroupQuestion);
  public
    {!script hide}
    function Link : TFhirQuestionnaireGroupQuestionList; Overload;
    function Clone : TFhirQuestionnaireGroupQuestionList; Overload;
    function GetEnumerator : TFhirQuestionnaireGroupQuestionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireGroupQuestion to the end of the list.
    }
    function Append : TFhirQuestionnaireGroupQuestion;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireGroupQuestion to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireGroupQuestion); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireGroupQuestion) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireGroupQuestion before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireGroupQuestion;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireGroupQuestion before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireGroupQuestion);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireGroupQuestion. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireGroupQuestion. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireGroupQuestion);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireGroupQuestion;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireGroupQuestions[index : Integer] : TFhirQuestionnaireGroupQuestion read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaire : TFhirDomainResource
    A structured set of questions intended to guide the collection of answers. The questions are ordered and grouped into coherent subsets, corresponding to the structure of the grouping of the underlying questions.
  }
  {!.Net HL7Connect.Fhir.Questionnaire}
  TFhirQuestionnaire = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FVersion : TFhirString;
    FStatus : TFhirEnum;
    FDate : TFhirDateTime;
    FPublisher : TFhirString;
    FtelecomList : TFhirContactPointList;
    FsubjectTypeList : TFhirCodeList;
    FGroup : TFhirQuestionnaireGroup;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirQuestionnaireStatus;
    Procedure SetStatusST(value : TFhirQuestionnaireStatus);
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    function GetSubjectTypeList : TFhirCodeList;
    function GetHasSubjectTypeList : Boolean;
    Procedure SetGroup(value : TFhirQuestionnaireGroup);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaire; overload;
    function Clone : TFhirQuestionnaire; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      This records identifiers associated with this question set that are defined by business processes and/or used to refer to it when a direct URL reference to the resource itself is not appropriate (e.g. in CDA documents, or in written / printed documentation).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member version
      Typed access to The version number assigned by the publisher for business reasons.  It may remain the same when the resource is updated.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version number assigned by the publisher for business reasons.  It may remain the same when the resource is updated.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member status
      The lifecycle status of the questionnaire as a whole.
    }
    property status : TFhirQuestionnaireStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member date
      Typed access to The date that this questionnaire was last changed.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date that this questionnaire was last changed.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member publisher
      Typed access to Organization or person responsible for developing and maintaining the questionnaire.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      Organization or person responsible for developing and maintaining the questionnaire.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member telecomList
      Contact details to assist a user in finding and communicating with the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member subjectTypeList
      Identifies the types of subjects that can be the subject of the questionnaire.
    }
    property subjectTypeList : TFhirCodeList read GetSubjectTypeList;
    property hasSubjectTypeList : boolean read GetHasSubjectTypeList;

    {@member group
      Typed access to A collection of related questions (or further groupings of questions). (defined for API consistency)
    }
    property group : TFhirQuestionnaireGroup read FGroup write SetGroup;
    {@member groupElement
      A collection of related questions (or further groupings of questions).
    }
    property groupElement : TFhirQuestionnaireGroup read FGroup write SetGroup;

  end;


  TFhirQuestionnaireListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireList;
    function GetCurrent : TFhirQuestionnaire;
  public
    Constructor Create(list : TFhirQuestionnaireList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaire read GetCurrent;
  end;


  {@Class TFhirQuestionnaireList
    A list of FhirQuestionnaire
  }
  {!.Net HL7Connect.Fhir.QuestionnaireList}
  TFhirQuestionnaireList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaire;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaire);
  public
    {!script hide}
    function Link : TFhirQuestionnaireList; Overload;
    function Clone : TFhirQuestionnaireList; Overload;
    function GetEnumerator : TFhirQuestionnaireListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaire to the end of the list.
    }
    function Append : TFhirQuestionnaire;

    
    {@member AddItem
      Add an already existing FhirQuestionnaire to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaire); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaire) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaire before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaire;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaire before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaire);
    
    {@member Item
       Get the iIndexth FhirQuestionnaire. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaire. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaire);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaire;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaires[index : Integer] : TFhirQuestionnaire read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireResponseGroup : TFhirElement
    A group of questions to a possibly similarly grouped set of questions in the questionnaire response.
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroup}
  TFhirQuestionnaireResponseGroup = class (TFhirBackboneElement)
  private
    FLinkId : TFhirString;
    FTitle : TFhirString;
    FText : TFhirString;
    FSubject : TFhirReference{TFhirReference};
    FgroupList : TFhirQuestionnaireResponseGroupList;
    FquestionList : TFhirQuestionnaireResponseGroupQuestionList;
    Procedure SetLinkId(value : TFhirString);
    Function GetLinkIdST : String;
    Procedure SetLinkIdST(value : String);
    Procedure SetTitle(value : TFhirString);
    Function GetTitleST : String;
    Procedure SetTitleST(value : String);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    Procedure SetSubject(value : TFhirReference{TFhirReference});
    function GetGroupList : TFhirQuestionnaireResponseGroupList;
    function GetHasGroupList : Boolean;
    function GetQuestionList : TFhirQuestionnaireResponseGroupQuestionList;
    function GetHasQuestionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireResponseGroup; overload;
    function Clone : TFhirQuestionnaireResponseGroup; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member linkId
      Typed access to Identifies the group from the Questionnaire that corresponds to this group in the QuestionnaireResponse resource.
    }
    property linkId : String read GetLinkIdST write SetLinkIdST;
    {@member linkIdElement
      Identifies the group from the Questionnaire that corresponds to this group in the QuestionnaireResponse resource.
    }
    property linkIdElement : TFhirString read FLinkId write SetLinkId;

    {@member title
      Typed access to Text that is displayed above the contents of the group.
    }
    property title : String read GetTitleST write SetTitleST;
    {@member titleElement
      Text that is displayed above the contents of the group.
    }
    property titleElement : TFhirString read FTitle write SetTitle;

    {@member text
      Typed access to Additional text for the group, used for display purposes.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      Additional text for the group, used for display purposes.
    }
    property textElement : TFhirString read FText write SetText;

    {@member subject
      Typed access to More specific subject this section's answers are about, details the subject given in QuestionnaireResponse. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirReference} read FSubject write SetSubject;
    {@member subjectElement
      More specific subject this section's answers are about, details the subject given in QuestionnaireResponse.
    }
    property subjectElement : TFhirReference{TFhirReference} read FSubject write SetSubject;

    {@member groupList
      A sub-group within a group. The ordering of groups within this group is relevant.
    }
    property groupList : TFhirQuestionnaireResponseGroupList read GetGroupList;
    property hasGroupList : boolean read GetHasGroupList;

    {@member questionList
      Set of questions within this group. The order of questions within the group is relevant.
    }
    property questionList : TFhirQuestionnaireResponseGroupQuestionList read GetQuestionList;
    property hasQuestionList : boolean read GetHasQuestionList;

  end;


  TFhirQuestionnaireResponseGroupListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireResponseGroupList;
    function GetCurrent : TFhirQuestionnaireResponseGroup;
  public
    Constructor Create(list : TFhirQuestionnaireResponseGroupList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireResponseGroup read GetCurrent;
  end;


  {@Class TFhirQuestionnaireResponseGroupList
    A list of FhirQuestionnaireResponseGroup
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroupList}
  TFhirQuestionnaireResponseGroupList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireResponseGroup;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireResponseGroup);
  public
    {!script hide}
    function Link : TFhirQuestionnaireResponseGroupList; Overload;
    function Clone : TFhirQuestionnaireResponseGroupList; Overload;
    function GetEnumerator : TFhirQuestionnaireResponseGroupListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireResponseGroup to the end of the list.
    }
    function Append : TFhirQuestionnaireResponseGroup;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireResponseGroup to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireResponseGroup); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireResponseGroup) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireResponseGroup before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireResponseGroup;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireResponseGroup before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireResponseGroup);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroup. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroup. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireResponseGroup);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireResponseGroup;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireResponseGroups[index : Integer] : TFhirQuestionnaireResponseGroup read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireResponseGroupQuestion : TFhirElement
    Set of questions within this group. The order of questions within the group is relevant.
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroupQuestion}
  TFhirQuestionnaireResponseGroupQuestion = class (TFhirBackboneElement)
  private
    FLinkId : TFhirString;
    FText : TFhirString;
    FanswerList : TFhirQuestionnaireResponseGroupQuestionAnswerList;
    Procedure SetLinkId(value : TFhirString);
    Function GetLinkIdST : String;
    Procedure SetLinkIdST(value : String);
    Procedure SetText(value : TFhirString);
    Function GetTextST : String;
    Procedure SetTextST(value : String);
    function GetAnswerList : TFhirQuestionnaireResponseGroupQuestionAnswerList;
    function GetHasAnswerList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireResponseGroupQuestion; overload;
    function Clone : TFhirQuestionnaireResponseGroupQuestion; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member linkId
      Typed access to Identifies the question from the Questionnaire that corresponds to this question in the QuestionnaireResponse resource.
    }
    property linkId : String read GetLinkIdST write SetLinkIdST;
    {@member linkIdElement
      Identifies the question from the Questionnaire that corresponds to this question in the QuestionnaireResponse resource.
    }
    property linkIdElement : TFhirString read FLinkId write SetLinkId;

    {@member text
      Typed access to The actual question as shown to the user to prompt them for an answer.
    }
    property text : String read GetTextST write SetTextST;
    {@member textElement
      The actual question as shown to the user to prompt them for an answer.
    }
    property textElement : TFhirString read FText write SetText;

    {@member answerList
      The respondent's answer(s) to the question.
    }
    property answerList : TFhirQuestionnaireResponseGroupQuestionAnswerList read GetAnswerList;
    property hasAnswerList : boolean read GetHasAnswerList;

  end;


  TFhirQuestionnaireResponseGroupQuestionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireResponseGroupQuestionList;
    function GetCurrent : TFhirQuestionnaireResponseGroupQuestion;
  public
    Constructor Create(list : TFhirQuestionnaireResponseGroupQuestionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireResponseGroupQuestion read GetCurrent;
  end;


  {@Class TFhirQuestionnaireResponseGroupQuestionList
    A list of FhirQuestionnaireResponseGroupQuestion
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroupQuestionList}
  TFhirQuestionnaireResponseGroupQuestionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireResponseGroupQuestion;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireResponseGroupQuestion);
  public
    {!script hide}
    function Link : TFhirQuestionnaireResponseGroupQuestionList; Overload;
    function Clone : TFhirQuestionnaireResponseGroupQuestionList; Overload;
    function GetEnumerator : TFhirQuestionnaireResponseGroupQuestionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireResponseGroupQuestion to the end of the list.
    }
    function Append : TFhirQuestionnaireResponseGroupQuestion;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireResponseGroupQuestion to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireResponseGroupQuestion); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireResponseGroupQuestion) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireResponseGroupQuestion before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireResponseGroupQuestion;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireResponseGroupQuestion before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireResponseGroupQuestion);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroupQuestion. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroupQuestion. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireResponseGroupQuestion);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireResponseGroupQuestion;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireResponseGroupQuestions[index : Integer] : TFhirQuestionnaireResponseGroupQuestion read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireResponseGroupQuestionAnswer : TFhirElement
    The respondent's answer(s) to the question.
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroupQuestionAnswer}
  TFhirQuestionnaireResponseGroupQuestionAnswer = class (TFhirBackboneElement)
  private
    FValue : TFhirType;
    FgroupList : TFhirQuestionnaireResponseGroupList;
    Procedure SetValue(value : TFhirType);
    function GetGroupList : TFhirQuestionnaireResponseGroupList;
    function GetHasGroupList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireResponseGroupQuestionAnswer; overload;
    function Clone : TFhirQuestionnaireResponseGroupQuestionAnswer; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member value
      Typed access to The answer (or one of the answers) provided by the respondent to the question. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      The answer (or one of the answers) provided by the respondent to the question.
    }
    property valueElement : TFhirType read FValue write SetValue;

    {@member groupList
      Nested group, containing nested question for this question. The order of groups within the question is relevant.
    }
    property groupList : TFhirQuestionnaireResponseGroupList read GetGroupList;
    property hasGroupList : boolean read GetHasGroupList;

  end;


  TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireResponseGroupQuestionAnswerList;
    function GetCurrent : TFhirQuestionnaireResponseGroupQuestionAnswer;
  public
    Constructor Create(list : TFhirQuestionnaireResponseGroupQuestionAnswerList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireResponseGroupQuestionAnswer read GetCurrent;
  end;


  {@Class TFhirQuestionnaireResponseGroupQuestionAnswerList
    A list of FhirQuestionnaireResponseGroupQuestionAnswer
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseGroupQuestionAnswerList}
  TFhirQuestionnaireResponseGroupQuestionAnswerList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireResponseGroupQuestionAnswer;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireResponseGroupQuestionAnswer);
  public
    {!script hide}
    function Link : TFhirQuestionnaireResponseGroupQuestionAnswerList; Overload;
    function Clone : TFhirQuestionnaireResponseGroupQuestionAnswerList; Overload;
    function GetEnumerator : TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireResponseGroupQuestionAnswer to the end of the list.
    }
    function Append : TFhirQuestionnaireResponseGroupQuestionAnswer;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireResponseGroupQuestionAnswer to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireResponseGroupQuestionAnswer); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireResponseGroupQuestionAnswer) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireResponseGroupQuestionAnswer before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireResponseGroupQuestionAnswer;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireResponseGroupQuestionAnswer before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireResponseGroupQuestionAnswer);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroupQuestionAnswer. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponseGroupQuestionAnswer. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireResponseGroupQuestionAnswer);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireResponseGroupQuestionAnswer;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireResponseGroupQuestionAnswers[index : Integer] : TFhirQuestionnaireResponseGroupQuestionAnswer read GetItemN write SetItemN; default;
  End;


  {@Class TFhirQuestionnaireResponse : TFhirDomainResource
    A structured set of questions and their answers. The questions are ordered and grouped into coherent subsets, corresponding to the structure of the grouping of the underlying questions.
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponse}
  TFhirQuestionnaireResponse = class (TFhirDomainResource)
  private
    FIdentifier : TFhirIdentifier;
    FQuestionnaire : TFhirReference{TFhirQuestionnaire};
    FStatus : TFhirEnum;
    FSubject : TFhirReference{TFhirReference};
    FAuthor : TFhirReference{Resource};
    FAuthored : TFhirDateTime;
    FSource : TFhirReference{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FGroup : TFhirQuestionnaireResponseGroup;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetQuestionnaire(value : TFhirReference{TFhirQuestionnaire});
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirQuestionnaireAnswersStatus;
    Procedure SetStatusST(value : TFhirQuestionnaireAnswersStatus);
    Procedure SetSubject(value : TFhirReference{TFhirReference});
    Procedure SetAuthor(value : TFhirReference{Resource});
    Procedure SetAuthored(value : TFhirDateTime);
    Function GetAuthoredST : TDateAndTime;
    Procedure SetAuthoredST(value : TDateAndTime);
    Procedure SetSource(value : TFhirReference{Resource});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetGroup(value : TFhirQuestionnaireResponseGroup);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirQuestionnaireResponse; overload;
    function Clone : TFhirQuestionnaireResponse; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifier
      Typed access to A business identifier assigned to a particular completed (or partially completed) questionnaire. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      A business identifier assigned to a particular completed (or partially completed) questionnaire.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member questionnaire
      Typed access to Indicates the Questionnaire resource that defines the form for which answers are being provided. (defined for API consistency)
    }
    property questionnaire : TFhirReference{TFhirQuestionnaire} read FQuestionnaire write SetQuestionnaire;
    {@member questionnaireElement
      Indicates the Questionnaire resource that defines the form for which answers are being provided.
    }
    property questionnaireElement : TFhirReference{TFhirQuestionnaire} read FQuestionnaire write SetQuestionnaire;

    {@member status
      The lifecycle status of the questionnaire response as a whole.
    }
    property status : TFhirQuestionnaireAnswersStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member subject
      Typed access to The subject of the questionnaire response.  This could be a patient, organization, practitioner, device, etc.  This is who/what the answers apply to, but is not necessarily the source of information. (defined for API consistency)
    }
    property subject : TFhirReference{TFhirReference} read FSubject write SetSubject;
    {@member subjectElement
      The subject of the questionnaire response.  This could be a patient, organization, practitioner, device, etc.  This is who/what the answers apply to, but is not necessarily the source of information.
    }
    property subjectElement : TFhirReference{TFhirReference} read FSubject write SetSubject;

    {@member author
      Typed access to Person who received the answers to the questions in the QuestionnaireResponse and recorded them in the system. (defined for API consistency)
    }
    property author : TFhirReference{Resource} read FAuthor write SetAuthor;
    {@member authorElement
      Person who received the answers to the questions in the QuestionnaireResponse and recorded them in the system.
    }
    property authorElement : TFhirReference{Resource} read FAuthor write SetAuthor;

    {@member authored
      Typed access to The date and/or time that this version of the questionnaire response was authored.
    }
    property authored : TDateAndTime read GetAuthoredST write SetAuthoredST;
    {@member authoredElement
      The date and/or time that this version of the questionnaire response was authored.
    }
    property authoredElement : TFhirDateTime read FAuthored write SetAuthored;

    {@member source
      Typed access to The person who answered the questions about the subject. (defined for API consistency)
    }
    property source : TFhirReference{Resource} read FSource write SetSource;
    {@member sourceElement
      The person who answered the questions about the subject.
    }
    property sourceElement : TFhirReference{Resource} read FSource write SetSource;

    {@member encounter
      Typed access to Encounter during which this set of questionnaire response were collected. When there were multiple encounters, this is the one considered most relevant to the context of the answers. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      Encounter during which this set of questionnaire response were collected. When there were multiple encounters, this is the one considered most relevant to the context of the answers.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member group
      Typed access to A group of questions to a possibly similarly grouped set of questions in the questionnaire response. (defined for API consistency)
    }
    property group : TFhirQuestionnaireResponseGroup read FGroup write SetGroup;
    {@member groupElement
      A group of questions to a possibly similarly grouped set of questions in the questionnaire response.
    }
    property groupElement : TFhirQuestionnaireResponseGroup read FGroup write SetGroup;

  end;


  TFhirQuestionnaireResponseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirQuestionnaireResponseList;
    function GetCurrent : TFhirQuestionnaireResponse;
  public
    Constructor Create(list : TFhirQuestionnaireResponseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirQuestionnaireResponse read GetCurrent;
  end;


  {@Class TFhirQuestionnaireResponseList
    A list of FhirQuestionnaireResponse
  }
  {!.Net HL7Connect.Fhir.QuestionnaireResponseList}
  TFhirQuestionnaireResponseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirQuestionnaireResponse;
    procedure SetItemN(index : Integer; value : TFhirQuestionnaireResponse);
  public
    {!script hide}
    function Link : TFhirQuestionnaireResponseList; Overload;
    function Clone : TFhirQuestionnaireResponseList; Overload;
    function GetEnumerator : TFhirQuestionnaireResponseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirQuestionnaireResponse to the end of the list.
    }
    function Append : TFhirQuestionnaireResponse;

    
    {@member AddItem
      Add an already existing FhirQuestionnaireResponse to the end of the list.
    }
    procedure AddItem(value : TFhirQuestionnaireResponse); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirQuestionnaireResponse) : Integer;
    

    {@member Insert
      Insert FhirQuestionnaireResponse before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirQuestionnaireResponse;
    

    {@member InsertItem
       Insert an existing FhirQuestionnaireResponse before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirQuestionnaireResponse);
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponse. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirQuestionnaireResponse. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirQuestionnaireResponse);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirQuestionnaireResponse;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirQuestionnaireResponses[index : Integer] : TFhirQuestionnaireResponse read GetItemN write SetItemN; default;
  End;


  {@Class TFhirReferralRequest : TFhirDomainResource
    Used to record and send details about a request for referral service or transfer of a patient to the care of another provider or provider organization.
  }
  {!.Net HL7Connect.Fhir.ReferralRequest}
  TFhirReferralRequest = class (TFhirDomainResource)
  private
    FStatus : TFhirEnum;
    FidentifierList : TFhirIdentifierList;
    FDate : TFhirDateTime;
    FType_ : TFhirCodeableConcept;
    FSpecialty : TFhirCodeableConcept;
    FPriority : TFhirCodeableConcept;
    FPatient : TFhirReference{TFhirPatient};
    FRequester : TFhirReference{Resource};
    FrecipientList : TFhirReferenceList{Resource};
    FEncounter : TFhirReference{TFhirEncounter};
    FDateSent : TFhirDateTime;
    FReason : TFhirCodeableConcept;
    FDescription : TFhirString;
    FserviceRequestedList : TFhirCodeableConceptList;
    FsupportingInformationList : TFhirReferenceList{TFhirReference};
    FFulfillmentTime : TFhirPeriod;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirReferralstatus;
    Procedure SetStatusST(value : TFhirReferralstatus);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetSpecialty(value : TFhirCodeableConcept);
    Procedure SetPriority(value : TFhirCodeableConcept);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetRequester(value : TFhirReference{Resource});
    function GetRecipientList : TFhirReferenceList{Resource};
    function GetHasRecipientList : Boolean;
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetDateSent(value : TFhirDateTime);
    Function GetDateSentST : TDateAndTime;
    Procedure SetDateSentST(value : TDateAndTime);
    Procedure SetReason(value : TFhirCodeableConcept);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetServiceRequestedList : TFhirCodeableConceptList;
    function GetHasServiceRequestedList : Boolean;
    function GetSupportingInformationList : TFhirReferenceList{TFhirReference};
    function GetHasSupportingInformationList : Boolean;
    Procedure SetFulfillmentTime(value : TFhirPeriod);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirReferralRequest; overload;
    function Clone : TFhirReferralRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member status
      The workflow status of the referral or transfer of care request.
    }
    property status : TFhirReferralstatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member identifierList
      Business identifier that uniquely identifies the referral/care transfer request instance.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member date
      Typed access to Date/DateTime of creation for draft requests and date of activation for active requests.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      Date/DateTime of creation for draft requests and date of activation for active requests.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member type_
      Typed access to An indication of the type of referral (or where applicable the type of transfer of care) request. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      An indication of the type of referral (or where applicable the type of transfer of care) request.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member specialty
      Typed access to Indication of the clinical domain or discipline to which the referral or transfer of care request is sent.  For example: Cardiology Gastroenterology Diabetology. (defined for API consistency)
    }
    property specialty : TFhirCodeableConcept read FSpecialty write SetSpecialty;
    {@member specialtyElement
      Indication of the clinical domain or discipline to which the referral or transfer of care request is sent.  For example: Cardiology Gastroenterology Diabetology.
    }
    property specialtyElement : TFhirCodeableConcept read FSpecialty write SetSpecialty;

    {@member priority
      Typed access to An indication of the urgency of referral (or where applicable the type of transfer of care) request. (defined for API consistency)
    }
    property priority : TFhirCodeableConcept read FPriority write SetPriority;
    {@member priorityElement
      An indication of the urgency of referral (or where applicable the type of transfer of care) request.
    }
    property priorityElement : TFhirCodeableConcept read FPriority write SetPriority;

    {@member patient
      Typed access to The patient who is the subject of a referral or transfer of care request. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient who is the subject of a referral or transfer of care request.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member requester
      Typed access to The healthcare provider or provider organization who/which initiated the referral/transfer of care request. Can also be  Patient (a self referral). (defined for API consistency)
    }
    property requester : TFhirReference{Resource} read FRequester write SetRequester;
    {@member requesterElement
      The healthcare provider or provider organization who/which initiated the referral/transfer of care request. Can also be  Patient (a self referral).
    }
    property requesterElement : TFhirReference{Resource} read FRequester write SetRequester;

    {@member recipientList
      The healthcare provider(s) or provider organization(s) who/which is to receive the referral/transfer of care request.
    }
    property recipientList : TFhirReferenceList{Resource} read GetRecipientList;
    property hasRecipientList : boolean read GetHasRecipientList;

    {@member encounter
      Typed access to The encounter at which the request for referral or transfer of care is initiated. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter at which the request for referral or transfer of care is initiated.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member dateSent
      Typed access to Date/DateTime the request for referral or transfer of care is sent by the author.
    }
    property dateSent : TDateAndTime read GetDateSentST write SetDateSentST;
    {@member dateSentElement
      Date/DateTime the request for referral or transfer of care is sent by the author.
    }
    property dateSentElement : TFhirDateTime read FDateSent write SetDateSent;

    {@member reason
      Typed access to Description of clinical condition indicating why referral/transfer of care is requested.  For example:  Pathological Anomalies, Disabled (physical or mental),  Behavioral Management. (defined for API consistency)
    }
    property reason : TFhirCodeableConcept read FReason write SetReason;
    {@member reasonElement
      Description of clinical condition indicating why referral/transfer of care is requested.  For example:  Pathological Anomalies, Disabled (physical or mental),  Behavioral Management.
    }
    property reasonElement : TFhirCodeableConcept read FReason write SetReason;

    {@member description
      Typed access to The reason element gives a short description of why the referral is being made, the description expands on this to support a more complete clinical summary.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      The reason element gives a short description of why the referral is being made, the description expands on this to support a more complete clinical summary.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member serviceRequestedList
      The service(s) that is/are requested to be provided to the patient.  For example: cardiac pacemaker insertion.
    }
    property serviceRequestedList : TFhirCodeableConceptList read GetServiceRequestedList;
    property hasServiceRequestedList : boolean read GetHasServiceRequestedList;

    {@member supportingInformationList
      Any additional (administrative, financial or clinical) information required to support request for referral or transfer of care.  For example: Presenting problems/chief complaints Medical History Family History Alerts Allergy/Intolerance and Adverse Reactions Medications Observations/Assessments (may include cognitive and fundtional assessments) Diagnostic Reports Care Plan.
    }
    property supportingInformationList : TFhirReferenceList{TFhirReference} read GetSupportingInformationList;
    property hasSupportingInformationList : boolean read GetHasSupportingInformationList;

    {@member fulfillmentTime
      Typed access to The period of time within which the services identified in the referral/transfer of care is specified or required to occur. (defined for API consistency)
    }
    property fulfillmentTime : TFhirPeriod read FFulfillmentTime write SetFulfillmentTime;
    {@member fulfillmentTimeElement
      The period of time within which the services identified in the referral/transfer of care is specified or required to occur.
    }
    property fulfillmentTimeElement : TFhirPeriod read FFulfillmentTime write SetFulfillmentTime;

  end;


  TFhirReferralRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirReferralRequestList;
    function GetCurrent : TFhirReferralRequest;
  public
    Constructor Create(list : TFhirReferralRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirReferralRequest read GetCurrent;
  end;


  {@Class TFhirReferralRequestList
    A list of FhirReferralRequest
  }
  {!.Net HL7Connect.Fhir.ReferralRequestList}
  TFhirReferralRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirReferralRequest;
    procedure SetItemN(index : Integer; value : TFhirReferralRequest);
  public
    {!script hide}
    function Link : TFhirReferralRequestList; Overload;
    function Clone : TFhirReferralRequestList; Overload;
    function GetEnumerator : TFhirReferralRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirReferralRequest to the end of the list.
    }
    function Append : TFhirReferralRequest;

    
    {@member AddItem
      Add an already existing FhirReferralRequest to the end of the list.
    }
    procedure AddItem(value : TFhirReferralRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirReferralRequest) : Integer;
    

    {@member Insert
      Insert FhirReferralRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirReferralRequest;
    

    {@member InsertItem
       Insert an existing FhirReferralRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirReferralRequest);
    
    {@member Item
       Get the iIndexth FhirReferralRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirReferralRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirReferralRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirReferralRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirReferralRequests[index : Integer] : TFhirReferralRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirRelatedPerson : TFhirDomainResource
    Information about a person that is involved in the care for a patient, but who is not the target of healthcare, nor has a formal responsibility in the care process.
  }
  {!.Net HL7Connect.Fhir.RelatedPerson}
  TFhirRelatedPerson = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FPatient : TFhirReference{TFhirPatient};
    FRelationship : TFhirCodeableConcept;
    FName : TFhirHumanName;
    FtelecomList : TFhirContactPointList;
    FGender : TFhirEnum;
    FBirthDate : TFhirDate;
    FaddressList : TFhirAddressList;
    FphotoList : TFhirAttachmentList;
    FPeriod : TFhirPeriod;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetRelationship(value : TFhirCodeableConcept);
    Procedure SetName(value : TFhirHumanName);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
    Procedure SetGender(value : TFhirEnum);
    Function GetGenderST : TFhirAdministrativeGender;
    Procedure SetGenderST(value : TFhirAdministrativeGender);
    Procedure SetBirthDate(value : TFhirDate);
    Function GetBirthDateST : TDateAndTime;
    Procedure SetBirthDateST(value : TDateAndTime);
    function GetAddressList : TFhirAddressList;
    function GetHasAddressList : Boolean;
    function GetPhotoList : TFhirAttachmentList;
    function GetHasPhotoList : Boolean;
    Procedure SetPeriod(value : TFhirPeriod);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirRelatedPerson; overload;
    function Clone : TFhirRelatedPerson; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Identifier for a person within a particular scope.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member patient
      Typed access to The patient this person is related to. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      The patient this person is related to.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member relationship
      Typed access to The nature of the relationship between a patient and the related person. (defined for API consistency)
    }
    property relationship : TFhirCodeableConcept read FRelationship write SetRelationship;
    {@member relationshipElement
      The nature of the relationship between a patient and the related person.
    }
    property relationshipElement : TFhirCodeableConcept read FRelationship write SetRelationship;

    {@member name
      Typed access to A name associated with the person. (defined for API consistency)
    }
    property name : TFhirHumanName read FName write SetName;
    {@member nameElement
      A name associated with the person.
    }
    property nameElement : TFhirHumanName read FName write SetName;

    {@member telecomList
      A contact detail for the person, e.g. a telephone number or an email address.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

    {@member gender
      Administrative Gender - the gender that the person is considered to have for administration and record keeping purposes.
    }
    property gender : TFhirAdministrativeGender read GetGenderST write SetGenderST;
    property genderElement : TFhirEnum read FGender write SetGender;

    {@member birthDate
      Typed access to The date on which the related person was born.
    }
    property birthDate : TDateAndTime read GetBirthDateST write SetBirthDateST;
    {@member birthDateElement
      The date on which the related person was born.
    }
    property birthDateElement : TFhirDate read FBirthDate write SetBirthDate;

    {@member addressList
      Address where the related person can be contacted or visited.
    }
    property addressList : TFhirAddressList read GetAddressList;
    property hasAddressList : boolean read GetHasAddressList;

    {@member photoList
      Image of the person.
    }
    property photoList : TFhirAttachmentList read GetPhotoList;
    property hasPhotoList : boolean read GetHasPhotoList;

    {@member period
      Typed access to The period of time that this relationship is considered to be valid. If there are no dates defined, then the interval is unknown. (defined for API consistency)
    }
    property period : TFhirPeriod read FPeriod write SetPeriod;
    {@member periodElement
      The period of time that this relationship is considered to be valid. If there are no dates defined, then the interval is unknown.
    }
    property periodElement : TFhirPeriod read FPeriod write SetPeriod;

  end;


  TFhirRelatedPersonListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirRelatedPersonList;
    function GetCurrent : TFhirRelatedPerson;
  public
    Constructor Create(list : TFhirRelatedPersonList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirRelatedPerson read GetCurrent;
  end;


  {@Class TFhirRelatedPersonList
    A list of FhirRelatedPerson
  }
  {!.Net HL7Connect.Fhir.RelatedPersonList}
  TFhirRelatedPersonList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirRelatedPerson;
    procedure SetItemN(index : Integer; value : TFhirRelatedPerson);
  public
    {!script hide}
    function Link : TFhirRelatedPersonList; Overload;
    function Clone : TFhirRelatedPersonList; Overload;
    function GetEnumerator : TFhirRelatedPersonListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirRelatedPerson to the end of the list.
    }
    function Append : TFhirRelatedPerson;

    
    {@member AddItem
      Add an already existing FhirRelatedPerson to the end of the list.
    }
    procedure AddItem(value : TFhirRelatedPerson); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirRelatedPerson) : Integer;
    

    {@member Insert
      Insert FhirRelatedPerson before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirRelatedPerson;
    

    {@member InsertItem
       Insert an existing FhirRelatedPerson before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirRelatedPerson);
    
    {@member Item
       Get the iIndexth FhirRelatedPerson. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirRelatedPerson. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirRelatedPerson);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirRelatedPerson;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirRelatedPeople[index : Integer] : TFhirRelatedPerson read GetItemN write SetItemN; default;
  End;


  {@Class TFhirRiskAssessmentPrediction : TFhirElement
    Describes the expected outcome for the subject.
  }
  {!.Net HL7Connect.Fhir.RiskAssessmentPrediction}
  TFhirRiskAssessmentPrediction = class (TFhirBackboneElement)
  private
    FOutcome : TFhirCodeableConcept;
    FProbability : TFhirType;
    FRelativeRisk : TFhirDecimal;
    FWhen : TFhirType;
    FRationale : TFhirString;
    Procedure SetOutcome(value : TFhirCodeableConcept);
    Procedure SetProbability(value : TFhirType);
    Procedure SetRelativeRisk(value : TFhirDecimal);
    Function GetRelativeRiskST : String;
    Procedure SetRelativeRiskST(value : String);
    Procedure SetWhen(value : TFhirType);
    Procedure SetRationale(value : TFhirString);
    Function GetRationaleST : String;
    Procedure SetRationaleST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirRiskAssessmentPrediction; overload;
    function Clone : TFhirRiskAssessmentPrediction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member outcome
      Typed access to One of the potential outcomes for the patient (e.g. remission, death,  a particular condition). (defined for API consistency)
    }
    property outcome : TFhirCodeableConcept read FOutcome write SetOutcome;
    {@member outcomeElement
      One of the potential outcomes for the patient (e.g. remission, death,  a particular condition).
    }
    property outcomeElement : TFhirCodeableConcept read FOutcome write SetOutcome;

    {@member probability
      Typed access to How likely is the outcome (in the specified timeframe). (defined for API consistency)
    }
    property probability : TFhirType read FProbability write SetProbability;
    {@member probabilityElement
      How likely is the outcome (in the specified timeframe).
    }
    property probabilityElement : TFhirType read FProbability write SetProbability;

    {@member relativeRisk
      Typed access to Indicates the risk for this particular subject (with their specific characteristics) divided by the risk of the population in general.  (Numbers greater than 1 = higher risk than the population, numbers less than 1 = lower risk.).
    }
    property relativeRisk : String read GetRelativeRiskST write SetRelativeRiskST;
    {@member relativeRiskElement
      Indicates the risk for this particular subject (with their specific characteristics) divided by the risk of the population in general.  (Numbers greater than 1 = higher risk than the population, numbers less than 1 = lower risk.).
    }
    property relativeRiskElement : TFhirDecimal read FRelativeRisk write SetRelativeRisk;

    {@member when
      Typed access to Indicates the period of time or age range of the subject to which the specified probability applies. (defined for API consistency)
    }
    property when : TFhirType read FWhen write SetWhen;
    {@member whenElement
      Indicates the period of time or age range of the subject to which the specified probability applies.
    }
    property whenElement : TFhirType read FWhen write SetWhen;

    {@member rationale
      Typed access to Additional information explaining the basis for the prediction.
    }
    property rationale : String read GetRationaleST write SetRationaleST;
    {@member rationaleElement
      Additional information explaining the basis for the prediction.
    }
    property rationaleElement : TFhirString read FRationale write SetRationale;

  end;


  TFhirRiskAssessmentPredictionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirRiskAssessmentPredictionList;
    function GetCurrent : TFhirRiskAssessmentPrediction;
  public
    Constructor Create(list : TFhirRiskAssessmentPredictionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirRiskAssessmentPrediction read GetCurrent;
  end;


  {@Class TFhirRiskAssessmentPredictionList
    A list of FhirRiskAssessmentPrediction
  }
  {!.Net HL7Connect.Fhir.RiskAssessmentPredictionList}
  TFhirRiskAssessmentPredictionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirRiskAssessmentPrediction;
    procedure SetItemN(index : Integer; value : TFhirRiskAssessmentPrediction);
  public
    {!script hide}
    function Link : TFhirRiskAssessmentPredictionList; Overload;
    function Clone : TFhirRiskAssessmentPredictionList; Overload;
    function GetEnumerator : TFhirRiskAssessmentPredictionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirRiskAssessmentPrediction to the end of the list.
    }
    function Append : TFhirRiskAssessmentPrediction;

    
    {@member AddItem
      Add an already existing FhirRiskAssessmentPrediction to the end of the list.
    }
    procedure AddItem(value : TFhirRiskAssessmentPrediction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirRiskAssessmentPrediction) : Integer;
    

    {@member Insert
      Insert FhirRiskAssessmentPrediction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirRiskAssessmentPrediction;
    

    {@member InsertItem
       Insert an existing FhirRiskAssessmentPrediction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirRiskAssessmentPrediction);
    
    {@member Item
       Get the iIndexth FhirRiskAssessmentPrediction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirRiskAssessmentPrediction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirRiskAssessmentPrediction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirRiskAssessmentPrediction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirRiskAssessmentPredictions[index : Integer] : TFhirRiskAssessmentPrediction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirRiskAssessment : TFhirDomainResource
    An assessment of the likely outcome(s) for a patient or other subject as well as the likelihood of each outcome.
  }
  {!.Net HL7Connect.Fhir.RiskAssessment}
  TFhirRiskAssessment = class (TFhirDomainResource)
  private
    FSubject : TFhirReference{Resource};
    FDate : TFhirDateTime;
    FCondition : TFhirReference{TFhirCondition};
    FEncounter : TFhirReference{TFhirEncounter};
    FPerformer : TFhirReference{Resource};
    FIdentifier : TFhirIdentifier;
    FMethod : TFhirCodeableConcept;
    FbasisList : TFhirReferenceList{TFhirReference};
    FpredictionList : TFhirRiskAssessmentPredictionList;
    FMitigation : TFhirString;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetCondition(value : TFhirReference{TFhirCondition});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetPerformer(value : TFhirReference{Resource});
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetMethod(value : TFhirCodeableConcept);
    function GetBasisList : TFhirReferenceList{TFhirReference};
    function GetHasBasisList : Boolean;
    function GetPredictionList : TFhirRiskAssessmentPredictionList;
    function GetHasPredictionList : Boolean;
    Procedure SetMitigation(value : TFhirString);
    Function GetMitigationST : String;
    Procedure SetMitigationST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirRiskAssessment; overload;
    function Clone : TFhirRiskAssessment; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member subject
      Typed access to The patient or group the risk assessment applies to. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      The patient or group the risk assessment applies to.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member date
      Typed access to The date (and possibly time) the risk assessment was performed.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date (and possibly time) the risk assessment was performed.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member condition
      Typed access to For assessments or prognosis specific to a particular condition, indicates the condition being assessed. (defined for API consistency)
    }
    property condition : TFhirReference{TFhirCondition} read FCondition write SetCondition;
    {@member conditionElement
      For assessments or prognosis specific to a particular condition, indicates the condition being assessed.
    }
    property conditionElement : TFhirReference{TFhirCondition} read FCondition write SetCondition;

    {@member encounter
      Typed access to The encounter where the assessment was performed. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      The encounter where the assessment was performed.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member performer
      Typed access to The provider or software application that performed the assessment. (defined for API consistency)
    }
    property performer : TFhirReference{Resource} read FPerformer write SetPerformer;
    {@member performerElement
      The provider or software application that performed the assessment.
    }
    property performerElement : TFhirReference{Resource} read FPerformer write SetPerformer;

    {@member identifier
      Typed access to Business identifier assigned to the risk assessment. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Business identifier assigned to the risk assessment.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member method
      Typed access to The algorithm, process or mechanism used to evaluate the risk. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      The algorithm, process or mechanism used to evaluate the risk.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member basisList
      Indicates the source data considered as part of the assessment (FamilyHistory, Observations, Procedures, Conditions, etc.).
    }
    property basisList : TFhirReferenceList{TFhirReference} read GetBasisList;
    property hasBasisList : boolean read GetHasBasisList;

    {@member predictionList
      Describes the expected outcome for the subject.
    }
    property predictionList : TFhirRiskAssessmentPredictionList read GetPredictionList;
    property hasPredictionList : boolean read GetHasPredictionList;

    {@member mitigation
      Typed access to A description of the steps that might be taken to reduce the identified risk(s).
    }
    property mitigation : String read GetMitigationST write SetMitigationST;
    {@member mitigationElement
      A description of the steps that might be taken to reduce the identified risk(s).
    }
    property mitigationElement : TFhirString read FMitigation write SetMitigation;

  end;


  TFhirRiskAssessmentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirRiskAssessmentList;
    function GetCurrent : TFhirRiskAssessment;
  public
    Constructor Create(list : TFhirRiskAssessmentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirRiskAssessment read GetCurrent;
  end;


  {@Class TFhirRiskAssessmentList
    A list of FhirRiskAssessment
  }
  {!.Net HL7Connect.Fhir.RiskAssessmentList}
  TFhirRiskAssessmentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirRiskAssessment;
    procedure SetItemN(index : Integer; value : TFhirRiskAssessment);
  public
    {!script hide}
    function Link : TFhirRiskAssessmentList; Overload;
    function Clone : TFhirRiskAssessmentList; Overload;
    function GetEnumerator : TFhirRiskAssessmentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirRiskAssessment to the end of the list.
    }
    function Append : TFhirRiskAssessment;

    
    {@member AddItem
      Add an already existing FhirRiskAssessment to the end of the list.
    }
    procedure AddItem(value : TFhirRiskAssessment); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirRiskAssessment) : Integer;
    

    {@member Insert
      Insert FhirRiskAssessment before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirRiskAssessment;
    

    {@member InsertItem
       Insert an existing FhirRiskAssessment before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirRiskAssessment);
    
    {@member Item
       Get the iIndexth FhirRiskAssessment. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirRiskAssessment. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirRiskAssessment);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirRiskAssessment;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirRiskAssessments[index : Integer] : TFhirRiskAssessment read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSchedule : TFhirDomainResource
    A container for slot(s) of time that may be available for booking appointments.
  }
  {!.Net HL7Connect.Fhir.Schedule}
  TFhirSchedule = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    Ftype_List : TFhirCodeableConceptList;
    FActor : TFhirReference{Resource};
    FPlanningHorizon : TFhirPeriod;
    FComment : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetType_List : TFhirCodeableConceptList;
    function GetHasType_List : Boolean;
    Procedure SetActor(value : TFhirReference{Resource});
    Procedure SetPlanningHorizon(value : TFhirPeriod);
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSchedule; overload;
    function Clone : TFhirSchedule; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External Ids for this item.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member type_List
      The schedule type can be used for the categorization of healthcare services or other appointment types.
    }
    property type_List : TFhirCodeableConceptList read GetType_List;
    property hasType_List : boolean read GetHasType_List;

    {@member actor
      Typed access to The resource this Schedule resource is providing availability information for. These are expected to usually be one of HealthcareService, Location, Practitioner, Device, Patient or RelatedPerson. (defined for API consistency)
    }
    property actor : TFhirReference{Resource} read FActor write SetActor;
    {@member actorElement
      The resource this Schedule resource is providing availability information for. These are expected to usually be one of HealthcareService, Location, Practitioner, Device, Patient or RelatedPerson.
    }
    property actorElement : TFhirReference{Resource} read FActor write SetActor;

    {@member planningHorizon
      Typed access to The period of time that the slots that are attached to this Schedule resource cover (even if none exist). These  cover the amount of time that an organization's planning horizon; the interval for which they are currently accepting appointments. This does not define a "template" for planning outside these dates. (defined for API consistency)
    }
    property planningHorizon : TFhirPeriod read FPlanningHorizon write SetPlanningHorizon;
    {@member planningHorizonElement
      The period of time that the slots that are attached to this Schedule resource cover (even if none exist). These  cover the amount of time that an organization's planning horizon; the interval for which they are currently accepting appointments. This does not define a "template" for planning outside these dates.
    }
    property planningHorizonElement : TFhirPeriod read FPlanningHorizon write SetPlanningHorizon;

    {@member comment
      Typed access to Comments on the availability to describe any extended information. Such as custom constraints on the slot(s) that may be associated.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      Comments on the availability to describe any extended information. Such as custom constraints on the slot(s) that may be associated.
    }
    property commentElement : TFhirString read FComment write SetComment;

  end;


  TFhirScheduleListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirScheduleList;
    function GetCurrent : TFhirSchedule;
  public
    Constructor Create(list : TFhirScheduleList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSchedule read GetCurrent;
  end;


  {@Class TFhirScheduleList
    A list of FhirSchedule
  }
  {!.Net HL7Connect.Fhir.ScheduleList}
  TFhirScheduleList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSchedule;
    procedure SetItemN(index : Integer; value : TFhirSchedule);
  public
    {!script hide}
    function Link : TFhirScheduleList; Overload;
    function Clone : TFhirScheduleList; Overload;
    function GetEnumerator : TFhirScheduleListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSchedule to the end of the list.
    }
    function Append : TFhirSchedule;

    
    {@member AddItem
      Add an already existing FhirSchedule to the end of the list.
    }
    procedure AddItem(value : TFhirSchedule); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSchedule) : Integer;
    

    {@member Insert
      Insert FhirSchedule before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSchedule;
    

    {@member InsertItem
       Insert an existing FhirSchedule before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSchedule);
    
    {@member Item
       Get the iIndexth FhirSchedule. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSchedule. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSchedule);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSchedule;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSchedules[index : Integer] : TFhirSchedule read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSearchParameterContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.SearchParameterContact}
  TFhirSearchParameterContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSearchParameterContact; overload;
    function Clone : TFhirSearchParameterContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the search parameter.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the search parameter.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirSearchParameterContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSearchParameterContactList;
    function GetCurrent : TFhirSearchParameterContact;
  public
    Constructor Create(list : TFhirSearchParameterContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSearchParameterContact read GetCurrent;
  end;


  {@Class TFhirSearchParameterContactList
    A list of FhirSearchParameterContact
  }
  {!.Net HL7Connect.Fhir.SearchParameterContactList}
  TFhirSearchParameterContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSearchParameterContact;
    procedure SetItemN(index : Integer; value : TFhirSearchParameterContact);
  public
    {!script hide}
    function Link : TFhirSearchParameterContactList; Overload;
    function Clone : TFhirSearchParameterContactList; Overload;
    function GetEnumerator : TFhirSearchParameterContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSearchParameterContact to the end of the list.
    }
    function Append : TFhirSearchParameterContact;

    
    {@member AddItem
      Add an already existing FhirSearchParameterContact to the end of the list.
    }
    procedure AddItem(value : TFhirSearchParameterContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSearchParameterContact) : Integer;
    

    {@member Insert
      Insert FhirSearchParameterContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSearchParameterContact;
    

    {@member InsertItem
       Insert an existing FhirSearchParameterContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSearchParameterContact);
    
    {@member Item
       Get the iIndexth FhirSearchParameterContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSearchParameterContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSearchParameterContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSearchParameterContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSearchParameterContacts[index : Integer] : TFhirSearchParameterContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSearchParameter : TFhirDomainResource
    A search parameter that defines a named search item that can be used to search/filter on a resource.
  }
  {!.Net HL7Connect.Fhir.SearchParameter}
  TFhirSearchParameter = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirSearchParameterContactList;
    FDate : TFhirDateTime;
    FRequirements : TFhirString;
    FCode : TFhirCode;
    FBase : TFhirCode;
    FType_ : TFhirEnum;
    FDescription : TFhirString;
    FXpath : TFhirString;
    FXpathUsage : TFhirEnum;
    FtargetList : TFhirCodeList;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirSearchParameterContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetBase(value : TFhirCode);
    Function GetBaseST : String;
    Procedure SetBaseST(value : String);
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirSearchParamType;
    Procedure SetType_ST(value : TFhirSearchParamType);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetXpath(value : TFhirString);
    Function GetXpathST : String;
    Procedure SetXpathST(value : String);
    Procedure SetXpathUsage(value : TFhirEnum);
    Function GetXpathUsageST : TFhirSearchXpathUsage;
    Procedure SetXpathUsageST(value : TFhirSearchXpathUsage);
    function GetTargetList : TFhirCodeList;
    function GetHasTargetList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSearchParameter; overload;
    function Clone : TFhirSearchParameter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this search parameter when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this search parameter is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this search parameter when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this search parameter is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member name
      Typed access to A free text natural language name identifying the search parameter.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the search parameter.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of this search parameter definition.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to A flag to indicate that this search parameter definition is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      A flag to indicate that this search parameter definition is authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the search parameter.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the search parameter.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirSearchParameterContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date  (and optionally time) when the search parameter definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the search parameter changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date  (and optionally time) when the search parameter definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the search parameter changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member requirements
      Typed access to The Scope and Usage that this search parameter was created to meet.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      The Scope and Usage that this search parameter was created to meet.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member code
      Typed access to The code used in the URL or the parameter name in a parameters resource for this search parameter.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      The code used in the URL or the parameter name in a parameters resource for this search parameter.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member base
      Typed access to The base resource type that this search parameter refers to.
    }
    property base : String read GetBaseST write SetBaseST;
    {@member baseElement
      The base resource type that this search parameter refers to.
    }
    property baseElement : TFhirCode read FBase write SetBase;

    {@member type_
      The type of value a search parameter refers to, and how the content is interpreted.
    }
    property type_ : TFhirSearchParamType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member description
      Typed access to A description of the search parameters and how it used.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A description of the search parameters and how it used.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member xpath
      Typed access to An XPath expression that returns a set of elements for the search parameter.
    }
    property xpath : String read GetXpathST write SetXpathST;
    {@member xpathElement
      An XPath expression that returns a set of elements for the search parameter.
    }
    property xpathElement : TFhirString read FXpath write SetXpath;

    {@member xpathUsage
      How the search parameter relates to the set of elements returned by evaluating the xpath query.
    }
    property xpathUsage : TFhirSearchXpathUsage read GetXpathUsageST write SetXpathUsageST;
    property xpathUsageElement : TFhirEnum read FXpathUsage write SetXpathUsage;

    {@member targetList
      Types of resource (if a resource is referenced).
    }
    property targetList : TFhirCodeList read GetTargetList;
    property hasTargetList : boolean read GetHasTargetList;

  end;


  TFhirSearchParameterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSearchParameterList;
    function GetCurrent : TFhirSearchParameter;
  public
    Constructor Create(list : TFhirSearchParameterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSearchParameter read GetCurrent;
  end;


  {@Class TFhirSearchParameterList
    A list of FhirSearchParameter
  }
  {!.Net HL7Connect.Fhir.SearchParameterList}
  TFhirSearchParameterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSearchParameter;
    procedure SetItemN(index : Integer; value : TFhirSearchParameter);
  public
    {!script hide}
    function Link : TFhirSearchParameterList; Overload;
    function Clone : TFhirSearchParameterList; Overload;
    function GetEnumerator : TFhirSearchParameterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSearchParameter to the end of the list.
    }
    function Append : TFhirSearchParameter;

    
    {@member AddItem
      Add an already existing FhirSearchParameter to the end of the list.
    }
    procedure AddItem(value : TFhirSearchParameter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSearchParameter) : Integer;
    

    {@member Insert
      Insert FhirSearchParameter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSearchParameter;
    

    {@member InsertItem
       Insert an existing FhirSearchParameter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSearchParameter);
    
    {@member Item
       Get the iIndexth FhirSearchParameter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSearchParameter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSearchParameter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSearchParameter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSearchParameters[index : Integer] : TFhirSearchParameter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSlot : TFhirDomainResource
    A slot of time on a schedule that may be available for booking appointments.
  }
  {!.Net HL7Connect.Fhir.Slot}
  TFhirSlot = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FType_ : TFhirCodeableConcept;
    FSchedule : TFhirReference{TFhirSchedule};
    FFreeBusyType : TFhirEnum;
    FStart : TFhirInstant;
    FEnd_ : TFhirInstant;
    FOverbooked : TFhirBoolean;
    FComment : TFhirString;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetSchedule(value : TFhirReference{TFhirSchedule});
    Procedure SetFreeBusyType(value : TFhirEnum);
    Function GetFreeBusyTypeST : TFhirSlotstatus;
    Procedure SetFreeBusyTypeST(value : TFhirSlotstatus);
    Procedure SetStart(value : TFhirInstant);
    Function GetStartST : TDateAndTime;
    Procedure SetStartST(value : TDateAndTime);
    Procedure SetEnd_(value : TFhirInstant);
    Function GetEnd_ST : TDateAndTime;
    Procedure SetEnd_ST(value : TDateAndTime);
    Procedure SetOverbooked(value : TFhirBoolean);
    Function GetOverbookedST : Boolean;
    Procedure SetOverbookedST(value : Boolean);
    Procedure SetComment(value : TFhirString);
    Function GetCommentST : String;
    Procedure SetCommentST(value : String);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSlot; overload;
    function Clone : TFhirSlot; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      External Ids for this item.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member type_
      Typed access to The type of appointments that can be booked into this slot (ideally this would be an identifiable service - which is at a location, rather than the location itself). If provided then this overrides the value provided on the availability resource. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The type of appointments that can be booked into this slot (ideally this would be an identifiable service - which is at a location, rather than the location itself). If provided then this overrides the value provided on the availability resource.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member schedule
      Typed access to The schedule resource that this slot defines an interval of status information. (defined for API consistency)
    }
    property schedule : TFhirReference{TFhirSchedule} read FSchedule write SetSchedule;
    {@member scheduleElement
      The schedule resource that this slot defines an interval of status information.
    }
    property scheduleElement : TFhirReference{TFhirSchedule} read FSchedule write SetSchedule;

    {@member freeBusyType
      busy | free | busy-unavailable | busy-tentative.
    }
    property freeBusyType : TFhirSlotstatus read GetFreeBusyTypeST write SetFreeBusyTypeST;
    property freeBusyTypeElement : TFhirEnum read FFreeBusyType write SetFreeBusyType;

    {@member start
      Typed access to Date/Time that the slot is to begin.
    }
    property start : TDateAndTime read GetStartST write SetStartST;
    {@member startElement
      Date/Time that the slot is to begin.
    }
    property startElement : TFhirInstant read FStart write SetStart;

    {@member end_
      Typed access to Date/Time that the slot is to conclude.
    }
    property end_ : TDateAndTime read GetEnd_ST write SetEnd_ST;
    {@member end_Element
      Date/Time that the slot is to conclude.
    }
    property end_Element : TFhirInstant read FEnd_ write SetEnd_;

    {@member overbooked
      Typed access to This slot has already been overbooked, appointments are unlikely to be accepted for this time.
    }
    property overbooked : Boolean read GetOverbookedST write SetOverbookedST;
    {@member overbookedElement
      This slot has already been overbooked, appointments are unlikely to be accepted for this time.
    }
    property overbookedElement : TFhirBoolean read FOverbooked write SetOverbooked;

    {@member comment
      Typed access to Comments on the slot to describe any extended information. Such as custom constraints on the slot.
    }
    property comment : String read GetCommentST write SetCommentST;
    {@member commentElement
      Comments on the slot to describe any extended information. Such as custom constraints on the slot.
    }
    property commentElement : TFhirString read FComment write SetComment;

  end;


  TFhirSlotListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSlotList;
    function GetCurrent : TFhirSlot;
  public
    Constructor Create(list : TFhirSlotList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSlot read GetCurrent;
  end;


  {@Class TFhirSlotList
    A list of FhirSlot
  }
  {!.Net HL7Connect.Fhir.SlotList}
  TFhirSlotList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSlot;
    procedure SetItemN(index : Integer; value : TFhirSlot);
  public
    {!script hide}
    function Link : TFhirSlotList; Overload;
    function Clone : TFhirSlotList; Overload;
    function GetEnumerator : TFhirSlotListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSlot to the end of the list.
    }
    function Append : TFhirSlot;

    
    {@member AddItem
      Add an already existing FhirSlot to the end of the list.
    }
    procedure AddItem(value : TFhirSlot); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSlot) : Integer;
    

    {@member Insert
      Insert FhirSlot before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSlot;
    

    {@member InsertItem
       Insert an existing FhirSlot before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSlot);
    
    {@member Item
       Get the iIndexth FhirSlot. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSlot. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSlot);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSlot;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSlots[index : Integer] : TFhirSlot read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSpecimenCollection : TFhirElement
    Details concerning the specimen collection.
  }
  {!.Net HL7Connect.Fhir.SpecimenCollection}
  TFhirSpecimenCollection = class (TFhirBackboneElement)
  private
    FCollector : TFhirReference{TFhirPractitioner};
    FcommentList : TFhirStringList;
    FCollected : TFhirType;
    FQuantity : TFhirQuantity;
    FMethod : TFhirCodeableConcept;
    FBodySite : TFhirCodeableConcept;
    Procedure SetCollector(value : TFhirReference{TFhirPractitioner});
    function GetCommentList : TFhirStringList;
    function GetHasCommentList : Boolean;
    Procedure SetCollected(value : TFhirType);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetMethod(value : TFhirCodeableConcept);
    Procedure SetBodySite(value : TFhirCodeableConcept);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSpecimenCollection; overload;
    function Clone : TFhirSpecimenCollection; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member collector
      Typed access to Person who collected the specimen. (defined for API consistency)
    }
    property collector : TFhirReference{TFhirPractitioner} read FCollector write SetCollector;
    {@member collectorElement
      Person who collected the specimen.
    }
    property collectorElement : TFhirReference{TFhirPractitioner} read FCollector write SetCollector;

    {@member commentList
      To communicate any details or issues encountered during the specimen collection procedure.
    }
    property commentList : TFhirStringList read GetCommentList;
    property hasCommentList : boolean read GetHasCommentList;

    {@member collected
      Typed access to Time when specimen was collected from subject - the physiologically relevant time. (defined for API consistency)
    }
    property collected : TFhirType read FCollected write SetCollected;
    {@member collectedElement
      Time when specimen was collected from subject - the physiologically relevant time.
    }
    property collectedElement : TFhirType read FCollected write SetCollected;

    {@member quantity
      Typed access to The quantity of specimen collected; for instance the volume of a blood sample, or the physical measurement of an anatomic pathology sample. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The quantity of specimen collected; for instance the volume of a blood sample, or the physical measurement of an anatomic pathology sample.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member method
      Typed access to A coded value specifying the technique that is used to perform the procedure. (defined for API consistency)
    }
    property method : TFhirCodeableConcept read FMethod write SetMethod;
    {@member methodElement
      A coded value specifying the technique that is used to perform the procedure.
    }
    property methodElement : TFhirCodeableConcept read FMethod write SetMethod;

    {@member bodySite
      Typed access to Anatomical location from which the specimen was collected (if subject is a patient). This is the target site.  This element is not used for environmental specimens. (defined for API consistency)
    }
    property bodySite : TFhirCodeableConcept read FBodySite write SetBodySite;
    {@member bodySiteElement
      Anatomical location from which the specimen was collected (if subject is a patient). This is the target site.  This element is not used for environmental specimens.
    }
    property bodySiteElement : TFhirCodeableConcept read FBodySite write SetBodySite;

  end;


  TFhirSpecimenCollectionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSpecimenCollectionList;
    function GetCurrent : TFhirSpecimenCollection;
  public
    Constructor Create(list : TFhirSpecimenCollectionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSpecimenCollection read GetCurrent;
  end;


  {@Class TFhirSpecimenCollectionList
    A list of FhirSpecimenCollection
  }
  {!.Net HL7Connect.Fhir.SpecimenCollectionList}
  TFhirSpecimenCollectionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSpecimenCollection;
    procedure SetItemN(index : Integer; value : TFhirSpecimenCollection);
  public
    {!script hide}
    function Link : TFhirSpecimenCollectionList; Overload;
    function Clone : TFhirSpecimenCollectionList; Overload;
    function GetEnumerator : TFhirSpecimenCollectionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSpecimenCollection to the end of the list.
    }
    function Append : TFhirSpecimenCollection;

    
    {@member AddItem
      Add an already existing FhirSpecimenCollection to the end of the list.
    }
    procedure AddItem(value : TFhirSpecimenCollection); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSpecimenCollection) : Integer;
    

    {@member Insert
      Insert FhirSpecimenCollection before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSpecimenCollection;
    

    {@member InsertItem
       Insert an existing FhirSpecimenCollection before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSpecimenCollection);
    
    {@member Item
       Get the iIndexth FhirSpecimenCollection. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSpecimenCollection. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSpecimenCollection);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSpecimenCollection;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSpecimenCollections[index : Integer] : TFhirSpecimenCollection read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSpecimenTreatment : TFhirElement
    Details concerning treatment and processing steps for the specimen.
  }
  {!.Net HL7Connect.Fhir.SpecimenTreatment}
  TFhirSpecimenTreatment = class (TFhirBackboneElement)
  private
    FDescription : TFhirString;
    FProcedure_ : TFhirCodeableConcept;
    FadditiveList : TFhirReferenceList{TFhirSubstance};
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetProcedure_(value : TFhirCodeableConcept);
    function GetAdditiveList : TFhirReferenceList{TFhirSubstance};
    function GetHasAdditiveList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSpecimenTreatment; overload;
    function Clone : TFhirSpecimenTreatment; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member description
      Typed access to Textual description of procedure.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Textual description of procedure.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member procedure_
      Typed access to A coded value specifying the procedure used to process the specimen. (defined for API consistency)
    }
    property procedure_ : TFhirCodeableConcept read FProcedure_ write SetProcedure_;
    {@member procedure_Element
      A coded value specifying the procedure used to process the specimen.
    }
    property procedure_Element : TFhirCodeableConcept read FProcedure_ write SetProcedure_;

    {@member additiveList
      Material used in the processing step.
    }
    property additiveList : TFhirReferenceList{TFhirSubstance} read GetAdditiveList;
    property hasAdditiveList : boolean read GetHasAdditiveList;

  end;


  TFhirSpecimenTreatmentListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSpecimenTreatmentList;
    function GetCurrent : TFhirSpecimenTreatment;
  public
    Constructor Create(list : TFhirSpecimenTreatmentList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSpecimenTreatment read GetCurrent;
  end;


  {@Class TFhirSpecimenTreatmentList
    A list of FhirSpecimenTreatment
  }
  {!.Net HL7Connect.Fhir.SpecimenTreatmentList}
  TFhirSpecimenTreatmentList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSpecimenTreatment;
    procedure SetItemN(index : Integer; value : TFhirSpecimenTreatment);
  public
    {!script hide}
    function Link : TFhirSpecimenTreatmentList; Overload;
    function Clone : TFhirSpecimenTreatmentList; Overload;
    function GetEnumerator : TFhirSpecimenTreatmentListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSpecimenTreatment to the end of the list.
    }
    function Append : TFhirSpecimenTreatment;

    
    {@member AddItem
      Add an already existing FhirSpecimenTreatment to the end of the list.
    }
    procedure AddItem(value : TFhirSpecimenTreatment); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSpecimenTreatment) : Integer;
    

    {@member Insert
      Insert FhirSpecimenTreatment before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSpecimenTreatment;
    

    {@member InsertItem
       Insert an existing FhirSpecimenTreatment before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSpecimenTreatment);
    
    {@member Item
       Get the iIndexth FhirSpecimenTreatment. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSpecimenTreatment. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSpecimenTreatment);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSpecimenTreatment;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSpecimenTreatments[index : Integer] : TFhirSpecimenTreatment read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSpecimenContainer : TFhirElement
    The container holding the specimen.  The recursive nature of containers; i.e. blood in tube in tray in rack is not addressed here.
  }
  {!.Net HL7Connect.Fhir.SpecimenContainer}
  TFhirSpecimenContainer = class (TFhirBackboneElement)
  private
    FidentifierList : TFhirIdentifierList;
    FDescription : TFhirString;
    FType_ : TFhirCodeableConcept;
    FCapacity : TFhirQuantity;
    FSpecimenQuantity : TFhirQuantity;
    FAdditive : TFhirType;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetCapacity(value : TFhirQuantity);
    Procedure SetSpecimenQuantity(value : TFhirQuantity);
    Procedure SetAdditive(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSpecimenContainer; overload;
    function Clone : TFhirSpecimenContainer; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifierList
      Id for container. There may be multiple; a manufacturer's bar code, lab assigned identifier, etc. The container ID may differ from the specimen id in some circumstances.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member description
      Typed access to Textual description of the container.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Textual description of the container.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member type_
      Typed access to The type of container associated with the specimen (e.g. slide, aliquot, etc.). (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The type of container associated with the specimen (e.g. slide, aliquot, etc.).
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member capacity
      Typed access to The capacity (volume or other measure) the container may contain. (defined for API consistency)
    }
    property capacity : TFhirQuantity read FCapacity write SetCapacity;
    {@member capacityElement
      The capacity (volume or other measure) the container may contain.
    }
    property capacityElement : TFhirQuantity read FCapacity write SetCapacity;

    {@member specimenQuantity
      Typed access to The quantity of specimen in the container; may be volume, dimensions, or other appropriate measurements, depending on the specimen type. (defined for API consistency)
    }
    property specimenQuantity : TFhirQuantity read FSpecimenQuantity write SetSpecimenQuantity;
    {@member specimenQuantityElement
      The quantity of specimen in the container; may be volume, dimensions, or other appropriate measurements, depending on the specimen type.
    }
    property specimenQuantityElement : TFhirQuantity read FSpecimenQuantity write SetSpecimenQuantity;

    {@member additive
      Typed access to Introduced substance to preserve, maintain or enhance the specimen. Examples: Formalin, Citrate, EDTA. (defined for API consistency)
    }
    property additive : TFhirType read FAdditive write SetAdditive;
    {@member additiveElement
      Introduced substance to preserve, maintain or enhance the specimen. Examples: Formalin, Citrate, EDTA.
    }
    property additiveElement : TFhirType read FAdditive write SetAdditive;

  end;


  TFhirSpecimenContainerListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSpecimenContainerList;
    function GetCurrent : TFhirSpecimenContainer;
  public
    Constructor Create(list : TFhirSpecimenContainerList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSpecimenContainer read GetCurrent;
  end;


  {@Class TFhirSpecimenContainerList
    A list of FhirSpecimenContainer
  }
  {!.Net HL7Connect.Fhir.SpecimenContainerList}
  TFhirSpecimenContainerList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSpecimenContainer;
    procedure SetItemN(index : Integer; value : TFhirSpecimenContainer);
  public
    {!script hide}
    function Link : TFhirSpecimenContainerList; Overload;
    function Clone : TFhirSpecimenContainerList; Overload;
    function GetEnumerator : TFhirSpecimenContainerListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSpecimenContainer to the end of the list.
    }
    function Append : TFhirSpecimenContainer;

    
    {@member AddItem
      Add an already existing FhirSpecimenContainer to the end of the list.
    }
    procedure AddItem(value : TFhirSpecimenContainer); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSpecimenContainer) : Integer;
    

    {@member Insert
      Insert FhirSpecimenContainer before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSpecimenContainer;
    

    {@member InsertItem
       Insert an existing FhirSpecimenContainer before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSpecimenContainer);
    
    {@member Item
       Get the iIndexth FhirSpecimenContainer. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSpecimenContainer. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSpecimenContainer);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSpecimenContainer;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSpecimenContainers[index : Integer] : TFhirSpecimenContainer read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSpecimen : TFhirDomainResource
    A sample to be used for analysis.
  }
  {!.Net HL7Connect.Fhir.Specimen}
  TFhirSpecimen = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FStatus : TFhirEnum;
    FType_ : TFhirCodeableConcept;
    FparentList : TFhirReferenceList{TFhirSpecimen};
    FSubject : TFhirReference{Resource};
    FAccessionIdentifier : TFhirIdentifier;
    FReceivedTime : TFhirDateTime;
    FCollection : TFhirSpecimenCollection;
    FtreatmentList : TFhirSpecimenTreatmentList;
    FcontainerList : TFhirSpecimenContainerList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirSpecimenStatus;
    Procedure SetStatusST(value : TFhirSpecimenStatus);
    Procedure SetType_(value : TFhirCodeableConcept);
    function GetParentList : TFhirReferenceList{TFhirSpecimen};
    function GetHasParentList : Boolean;
    Procedure SetSubject(value : TFhirReference{Resource});
    Procedure SetAccessionIdentifier(value : TFhirIdentifier);
    Procedure SetReceivedTime(value : TFhirDateTime);
    Function GetReceivedTimeST : TDateAndTime;
    Procedure SetReceivedTimeST(value : TDateAndTime);
    Procedure SetCollection(value : TFhirSpecimenCollection);
    function GetTreatmentList : TFhirSpecimenTreatmentList;
    function GetHasTreatmentList : Boolean;
    function GetContainerList : TFhirSpecimenContainerList;
    function GetHasContainerList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSpecimen; overload;
    function Clone : TFhirSpecimen; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Id for specimen.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member status
      The availability of the specimen.
    }
    property status : TFhirSpecimenStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member type_
      Typed access to The kind of material that forms the specimen. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      The kind of material that forms the specimen.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member parentList
      Reference to the parent (source) specimen which is used when the specimen was either derived from or a component of another specimen.
    }
    property parentList : TFhirReferenceList{TFhirSpecimen} read GetParentList;
    property hasParentList : boolean read GetHasParentList;

    {@member subject
      Typed access to Where the specimen came from. This may be from the patient(s) or from the environment or a device. (defined for API consistency)
    }
    property subject : TFhirReference{Resource} read FSubject write SetSubject;
    {@member subjectElement
      Where the specimen came from. This may be from the patient(s) or from the environment or a device.
    }
    property subjectElement : TFhirReference{Resource} read FSubject write SetSubject;

    {@member accessionIdentifier
      Typed access to The identifier assigned by the lab when accessioning specimen(s). This is not necessarily the same as the specimen identifier, depending on local lab procedures. (defined for API consistency)
    }
    property accessionIdentifier : TFhirIdentifier read FAccessionIdentifier write SetAccessionIdentifier;
    {@member accessionIdentifierElement
      The identifier assigned by the lab when accessioning specimen(s). This is not necessarily the same as the specimen identifier, depending on local lab procedures.
    }
    property accessionIdentifierElement : TFhirIdentifier read FAccessionIdentifier write SetAccessionIdentifier;

    {@member receivedTime
      Typed access to Time when specimen was received for processing or testing.
    }
    property receivedTime : TDateAndTime read GetReceivedTimeST write SetReceivedTimeST;
    {@member receivedTimeElement
      Time when specimen was received for processing or testing.
    }
    property receivedTimeElement : TFhirDateTime read FReceivedTime write SetReceivedTime;

    {@member collection
      Typed access to Details concerning the specimen collection. (defined for API consistency)
    }
    property collection : TFhirSpecimenCollection read FCollection write SetCollection;
    {@member collectionElement
      Details concerning the specimen collection.
    }
    property collectionElement : TFhirSpecimenCollection read FCollection write SetCollection;

    {@member treatmentList
      Details concerning treatment and processing steps for the specimen.
    }
    property treatmentList : TFhirSpecimenTreatmentList read GetTreatmentList;
    property hasTreatmentList : boolean read GetHasTreatmentList;

    {@member containerList
      The container holding the specimen.  The recursive nature of containers; i.e. blood in tube in tray in rack is not addressed here.
    }
    property containerList : TFhirSpecimenContainerList read GetContainerList;
    property hasContainerList : boolean read GetHasContainerList;

  end;


  TFhirSpecimenListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSpecimenList;
    function GetCurrent : TFhirSpecimen;
  public
    Constructor Create(list : TFhirSpecimenList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSpecimen read GetCurrent;
  end;


  {@Class TFhirSpecimenList
    A list of FhirSpecimen
  }
  {!.Net HL7Connect.Fhir.SpecimenList}
  TFhirSpecimenList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSpecimen;
    procedure SetItemN(index : Integer; value : TFhirSpecimen);
  public
    {!script hide}
    function Link : TFhirSpecimenList; Overload;
    function Clone : TFhirSpecimenList; Overload;
    function GetEnumerator : TFhirSpecimenListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSpecimen to the end of the list.
    }
    function Append : TFhirSpecimen;

    
    {@member AddItem
      Add an already existing FhirSpecimen to the end of the list.
    }
    procedure AddItem(value : TFhirSpecimen); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSpecimen) : Integer;
    

    {@member Insert
      Insert FhirSpecimen before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSpecimen;
    

    {@member InsertItem
       Insert an existing FhirSpecimen before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSpecimen);
    
    {@member Item
       Get the iIndexth FhirSpecimen. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSpecimen. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSpecimen);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSpecimen;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSpecimen[index : Integer] : TFhirSpecimen read GetItemN write SetItemN; default;
  End;


  {@Class TFhirStructureDefinitionContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionContact}
  TFhirStructureDefinitionContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirStructureDefinitionContact; overload;
    function Clone : TFhirStructureDefinitionContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the structure definition.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the structure definition.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirStructureDefinitionContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirStructureDefinitionContactList;
    function GetCurrent : TFhirStructureDefinitionContact;
  public
    Constructor Create(list : TFhirStructureDefinitionContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirStructureDefinitionContact read GetCurrent;
  end;


  {@Class TFhirStructureDefinitionContactList
    A list of FhirStructureDefinitionContact
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionContactList}
  TFhirStructureDefinitionContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirStructureDefinitionContact;
    procedure SetItemN(index : Integer; value : TFhirStructureDefinitionContact);
  public
    {!script hide}
    function Link : TFhirStructureDefinitionContactList; Overload;
    function Clone : TFhirStructureDefinitionContactList; Overload;
    function GetEnumerator : TFhirStructureDefinitionContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirStructureDefinitionContact to the end of the list.
    }
    function Append : TFhirStructureDefinitionContact;

    
    {@member AddItem
      Add an already existing FhirStructureDefinitionContact to the end of the list.
    }
    procedure AddItem(value : TFhirStructureDefinitionContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirStructureDefinitionContact) : Integer;
    

    {@member Insert
      Insert FhirStructureDefinitionContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirStructureDefinitionContact;
    

    {@member InsertItem
       Insert an existing FhirStructureDefinitionContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirStructureDefinitionContact);
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirStructureDefinitionContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirStructureDefinitionContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirStructureDefinitionContacts[index : Integer] : TFhirStructureDefinitionContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirStructureDefinitionMapping : TFhirElement
    An external specification that the content is mapped to.
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionMapping}
  TFhirStructureDefinitionMapping = class (TFhirBackboneElement)
  private
    FIdentity : TFhirId;
    FUri : TFhirUri;
    FName : TFhirString;
    FComments : TFhirString;
    Procedure SetIdentity(value : TFhirId);
    Function GetIdentityST : String;
    Procedure SetIdentityST(value : String);
    Procedure SetUri(value : TFhirUri);
    Function GetUriST : String;
    Procedure SetUriST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetComments(value : TFhirString);
    Function GetCommentsST : String;
    Procedure SetCommentsST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirStructureDefinitionMapping; overload;
    function Clone : TFhirStructureDefinitionMapping; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identity
      Typed access to An Internal id that is used to identify this mapping set when specific mappings are made.
    }
    property identity : String read GetIdentityST write SetIdentityST;
    {@member identityElement
      An Internal id that is used to identify this mapping set when specific mappings are made.
    }
    property identityElement : TFhirId read FIdentity write SetIdentity;

    {@member uri
      Typed access to An absolute URI that identifies the specification that this mapping is expressed to.
    }
    property uri : String read GetUriST write SetUriST;
    {@member uriElement
      An absolute URI that identifies the specification that this mapping is expressed to.
    }
    property uriElement : TFhirUri read FUri write SetUri;

    {@member name
      Typed access to A name for the specification that is being mapped to.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A name for the specification that is being mapped to.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member comments
      Typed access to Comments about this mapping, including version notes, issues, scope limitations, and other important notes for usage.
    }
    property comments : String read GetCommentsST write SetCommentsST;
    {@member commentsElement
      Comments about this mapping, including version notes, issues, scope limitations, and other important notes for usage.
    }
    property commentsElement : TFhirString read FComments write SetComments;

  end;


  TFhirStructureDefinitionMappingListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirStructureDefinitionMappingList;
    function GetCurrent : TFhirStructureDefinitionMapping;
  public
    Constructor Create(list : TFhirStructureDefinitionMappingList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirStructureDefinitionMapping read GetCurrent;
  end;


  {@Class TFhirStructureDefinitionMappingList
    A list of FhirStructureDefinitionMapping
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionMappingList}
  TFhirStructureDefinitionMappingList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirStructureDefinitionMapping;
    procedure SetItemN(index : Integer; value : TFhirStructureDefinitionMapping);
  public
    {!script hide}
    function Link : TFhirStructureDefinitionMappingList; Overload;
    function Clone : TFhirStructureDefinitionMappingList; Overload;
    function GetEnumerator : TFhirStructureDefinitionMappingListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirStructureDefinitionMapping to the end of the list.
    }
    function Append : TFhirStructureDefinitionMapping;

    
    {@member AddItem
      Add an already existing FhirStructureDefinitionMapping to the end of the list.
    }
    procedure AddItem(value : TFhirStructureDefinitionMapping); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirStructureDefinitionMapping) : Integer;
    

    {@member Insert
      Insert FhirStructureDefinitionMapping before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirStructureDefinitionMapping;
    

    {@member InsertItem
       Insert an existing FhirStructureDefinitionMapping before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirStructureDefinitionMapping);
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionMapping. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionMapping. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirStructureDefinitionMapping);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirStructureDefinitionMapping;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirStructureDefinitionMappings[index : Integer] : TFhirStructureDefinitionMapping read GetItemN write SetItemN; default;
  End;


  {@Class TFhirStructureDefinitionSnapshot : TFhirElement
    A snapshot view is expressed in a stand alone form that can be used and interpreted without considering the base StructureDefinition.
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionSnapshot}
  TFhirStructureDefinitionSnapshot = class (TFhirBackboneElement)
  private
    FelementList : TFhirElementDefinitionList;
    function GetElementList : TFhirElementDefinitionList;
    function GetHasElementList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirStructureDefinitionSnapshot; overload;
    function Clone : TFhirStructureDefinitionSnapshot; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member elementList
      Captures constraints on each element within the resource.
    }
    property elementList : TFhirElementDefinitionList read GetElementList;
    property hasElementList : boolean read GetHasElementList;

  end;


  TFhirStructureDefinitionSnapshotListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirStructureDefinitionSnapshotList;
    function GetCurrent : TFhirStructureDefinitionSnapshot;
  public
    Constructor Create(list : TFhirStructureDefinitionSnapshotList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirStructureDefinitionSnapshot read GetCurrent;
  end;


  {@Class TFhirStructureDefinitionSnapshotList
    A list of FhirStructureDefinitionSnapshot
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionSnapshotList}
  TFhirStructureDefinitionSnapshotList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirStructureDefinitionSnapshot;
    procedure SetItemN(index : Integer; value : TFhirStructureDefinitionSnapshot);
  public
    {!script hide}
    function Link : TFhirStructureDefinitionSnapshotList; Overload;
    function Clone : TFhirStructureDefinitionSnapshotList; Overload;
    function GetEnumerator : TFhirStructureDefinitionSnapshotListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirStructureDefinitionSnapshot to the end of the list.
    }
    function Append : TFhirStructureDefinitionSnapshot;

    
    {@member AddItem
      Add an already existing FhirStructureDefinitionSnapshot to the end of the list.
    }
    procedure AddItem(value : TFhirStructureDefinitionSnapshot); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirStructureDefinitionSnapshot) : Integer;
    

    {@member Insert
      Insert FhirStructureDefinitionSnapshot before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirStructureDefinitionSnapshot;
    

    {@member InsertItem
       Insert an existing FhirStructureDefinitionSnapshot before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirStructureDefinitionSnapshot);
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionSnapshot. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionSnapshot. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirStructureDefinitionSnapshot);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirStructureDefinitionSnapshot;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirStructureDefinitionSnapshots[index : Integer] : TFhirStructureDefinitionSnapshot read GetItemN write SetItemN; default;
  End;


  {@Class TFhirStructureDefinitionDifferential : TFhirElement
    A differential view is expressed relative to the base StructureDefinition - a statement of differences that it applies.
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionDifferential}
  TFhirStructureDefinitionDifferential = class (TFhirBackboneElement)
  private
    FelementList : TFhirElementDefinitionList;
    function GetElementList : TFhirElementDefinitionList;
    function GetHasElementList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirStructureDefinitionDifferential; overload;
    function Clone : TFhirStructureDefinitionDifferential; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member elementList
      Captures constraints on each element within the resource.
    }
    property elementList : TFhirElementDefinitionList read GetElementList;
    property hasElementList : boolean read GetHasElementList;

  end;


  TFhirStructureDefinitionDifferentialListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirStructureDefinitionDifferentialList;
    function GetCurrent : TFhirStructureDefinitionDifferential;
  public
    Constructor Create(list : TFhirStructureDefinitionDifferentialList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirStructureDefinitionDifferential read GetCurrent;
  end;


  {@Class TFhirStructureDefinitionDifferentialList
    A list of FhirStructureDefinitionDifferential
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionDifferentialList}
  TFhirStructureDefinitionDifferentialList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirStructureDefinitionDifferential;
    procedure SetItemN(index : Integer; value : TFhirStructureDefinitionDifferential);
  public
    {!script hide}
    function Link : TFhirStructureDefinitionDifferentialList; Overload;
    function Clone : TFhirStructureDefinitionDifferentialList; Overload;
    function GetEnumerator : TFhirStructureDefinitionDifferentialListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirStructureDefinitionDifferential to the end of the list.
    }
    function Append : TFhirStructureDefinitionDifferential;

    
    {@member AddItem
      Add an already existing FhirStructureDefinitionDifferential to the end of the list.
    }
    procedure AddItem(value : TFhirStructureDefinitionDifferential); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirStructureDefinitionDifferential) : Integer;
    

    {@member Insert
      Insert FhirStructureDefinitionDifferential before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirStructureDefinitionDifferential;
    

    {@member InsertItem
       Insert an existing FhirStructureDefinitionDifferential before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirStructureDefinitionDifferential);
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionDifferential. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirStructureDefinitionDifferential. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirStructureDefinitionDifferential);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirStructureDefinitionDifferential;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirStructureDefinitionDifferentials[index : Integer] : TFhirStructureDefinitionDifferential read GetItemN write SetItemN; default;
  End;


  {@Class TFhirStructureDefinition : TFhirDomainResource
    A definition of a FHIR structure. This resource is used to describe the underlying resources, data types defined in FHIR, and also for describing extensions, and constraints on resources and data types.
  }
  {!.Net HL7Connect.Fhir.StructureDefinition}
  TFhirStructureDefinition = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FidentifierList : TFhirIdentifierList;
    FVersion : TFhirString;
    FName : TFhirString;
    FDisplay : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirStructureDefinitionContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FRequirements : TFhirString;
    FCopyright : TFhirString;
    FcodeList : TFhirCodingList;
    FFhirVersion : TFhirId;
    FmappingList : TFhirStructureDefinitionMappingList;
    FKind : TFhirEnum;
    FConstrainedType : TFhirCode;
    FAbstract : TFhirBoolean;
    FContextType : TFhirEnum;
    FcontextList : TFhirStringList;
    FBase : TFhirUri;
    FSnapshot : TFhirStructureDefinitionSnapshot;
    FDifferential : TFhirStructureDefinitionDifferential;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDisplay(value : TFhirString);
    Function GetDisplayST : String;
    Procedure SetDisplayST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirStructureDefinitionContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    function GetCodeList : TFhirCodingList;
    function GetHasCodeList : Boolean;
    Procedure SetFhirVersion(value : TFhirId);
    Function GetFhirVersionST : String;
    Procedure SetFhirVersionST(value : String);
    function GetMappingList : TFhirStructureDefinitionMappingList;
    function GetHasMappingList : Boolean;
    Procedure SetKind(value : TFhirEnum);
    Function GetKindST : TFhirStructureDefinitionKind;
    Procedure SetKindST(value : TFhirStructureDefinitionKind);
    Procedure SetConstrainedType(value : TFhirCode);
    Function GetConstrainedTypeST : String;
    Procedure SetConstrainedTypeST(value : String);
    Procedure SetAbstract(value : TFhirBoolean);
    Function GetAbstractST : Boolean;
    Procedure SetAbstractST(value : Boolean);
    Procedure SetContextType(value : TFhirEnum);
    Function GetContextTypeST : TFhirExtensionContext;
    Procedure SetContextTypeST(value : TFhirExtensionContext);
    function GetContextList : TFhirStringList;
    function GetHasContextList : Boolean;
    Procedure SetBase(value : TFhirUri);
    Function GetBaseST : String;
    Procedure SetBaseST(value : String);
    Procedure SetSnapshot(value : TFhirStructureDefinitionSnapshot);
    Procedure SetDifferential(value : TFhirStructureDefinitionDifferential);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirStructureDefinition; overload;
    function Clone : TFhirStructureDefinition; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this structure definition when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this structure definition is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this structure definition when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this structure definition is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member identifierList
      Formal identifier that is used to identify this StructureDefinition when it is represented in other formats, or referenced in a specification, model, design or an instance  (should be globally unique OID, UUID, or URI), (if it's not possible to use the literal URI).
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member version
      Typed access to The identifier that is used to identify this version of the StructureDefinition when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the StructureDefinition author manually.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the StructureDefinition when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the StructureDefinition author manually.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name identifying the StructureDefinition.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the StructureDefinition.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member display
      Typed access to Defined so that applications can use this name when displaying the value of the extension to the user.
    }
    property display : String read GetDisplayST write SetDisplayST;
    {@member displayElement
      Defined so that applications can use this name when displaying the value of the extension to the user.
    }
    property displayElement : TFhirString read FDisplay write SetDisplay;

    {@member status
      The status of the StructureDefinition.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to This StructureDefinition was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This StructureDefinition was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the structure definition.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the structure definition.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirStructureDefinitionContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the structure definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the structure definition changes.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the structure definition was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the structure definition changes.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the StructureDefinition and its use.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the StructureDefinition and its use.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of structure definitions.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member requirements
      Typed access to Explains why this structure definition is needed and why it's been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this structure definition is needed and why it's been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member copyright
      Typed access to A copyright statement relating to the structure definition and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the structure definition and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member codeList
      A set of terms from external terminologies that may be used to assist with indexing and searching of templates.
    }
    property codeList : TFhirCodingList read GetCodeList;
    property hasCodeList : boolean read GetHasCodeList;

    {@member fhirVersion
      Typed access to The version of the FHIR specification on which this StructureDefinition is based - this is the formal version of the specification, without the revision number, e.g. [publication].[major].[minor], which is 1.0.2 for this version.
    }
    property fhirVersion : String read GetFhirVersionST write SetFhirVersionST;
    {@member fhirVersionElement
      The version of the FHIR specification on which this StructureDefinition is based - this is the formal version of the specification, without the revision number, e.g. [publication].[major].[minor], which is 1.0.2 for this version.
    }
    property fhirVersionElement : TFhirId read FFhirVersion write SetFhirVersion;

    {@member mappingList
      An external specification that the content is mapped to.
    }
    property mappingList : TFhirStructureDefinitionMappingList read GetMappingList;
    property hasMappingList : boolean read GetHasMappingList;

    {@member kind
      Defines the kind of structure that this definition is describing.
    }
    property kind : TFhirStructureDefinitionKind read GetKindST write SetKindST;
    property kindElement : TFhirEnum read FKind write SetKind;

    {@member constrainedType
      Typed access to The type of type that is being constrained - a data type, an extension, a resource, including abstract ones. If this field is present, it indicates that the structure definition is a constraint. If it is not present, then the structure definition is the definition of a base structure.
    }
    property constrainedType : String read GetConstrainedTypeST write SetConstrainedTypeST;
    {@member constrainedTypeElement
      The type of type that is being constrained - a data type, an extension, a resource, including abstract ones. If this field is present, it indicates that the structure definition is a constraint. If it is not present, then the structure definition is the definition of a base structure.
    }
    property constrainedTypeElement : TFhirCode read FConstrainedType write SetConstrainedType;

    {@member abstract
      Typed access to Whether structure this definition describes is abstract or not  - that is, whether an actual exchanged item can ever be of this type.
    }
    property abstract : Boolean read GetAbstractST write SetAbstractST;
    {@member abstractElement
      Whether structure this definition describes is abstract or not  - that is, whether an actual exchanged item can ever be of this type.
    }
    property abstractElement : TFhirBoolean read FAbstract write SetAbstract;

    {@member contextType
      If this is an extension, Identifies the context within FHIR resources where the extension can be used.
    }
    property contextType : TFhirExtensionContext read GetContextTypeST write SetContextTypeST;
    property contextTypeElement : TFhirEnum read FContextType write SetContextType;

    {@member contextList
      Identifies the types of resource or data type elements to which the extension can be applied.
    }
    property contextList : TFhirStringList read GetContextList;
    property hasContextList : boolean read GetHasContextList;

    {@member base
      Typed access to An absolute URI that is the base structure from which this set of constraints is derived.
    }
    property base : String read GetBaseST write SetBaseST;
    {@member baseElement
      An absolute URI that is the base structure from which this set of constraints is derived.
    }
    property baseElement : TFhirUri read FBase write SetBase;

    {@member snapshot
      Typed access to A snapshot view is expressed in a stand alone form that can be used and interpreted without considering the base StructureDefinition. (defined for API consistency)
    }
    property snapshot : TFhirStructureDefinitionSnapshot read FSnapshot write SetSnapshot;
    {@member snapshotElement
      A snapshot view is expressed in a stand alone form that can be used and interpreted without considering the base StructureDefinition.
    }
    property snapshotElement : TFhirStructureDefinitionSnapshot read FSnapshot write SetSnapshot;

    {@member differential
      Typed access to A differential view is expressed relative to the base StructureDefinition - a statement of differences that it applies. (defined for API consistency)
    }
    property differential : TFhirStructureDefinitionDifferential read FDifferential write SetDifferential;
    {@member differentialElement
      A differential view is expressed relative to the base StructureDefinition - a statement of differences that it applies.
    }
    property differentialElement : TFhirStructureDefinitionDifferential read FDifferential write SetDifferential;

  end;


  TFhirStructureDefinitionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirStructureDefinitionList;
    function GetCurrent : TFhirStructureDefinition;
  public
    Constructor Create(list : TFhirStructureDefinitionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirStructureDefinition read GetCurrent;
  end;


  {@Class TFhirStructureDefinitionList
    A list of FhirStructureDefinition
  }
  {!.Net HL7Connect.Fhir.StructureDefinitionList}
  TFhirStructureDefinitionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirStructureDefinition;
    procedure SetItemN(index : Integer; value : TFhirStructureDefinition);
  public
    {!script hide}
    function Link : TFhirStructureDefinitionList; Overload;
    function Clone : TFhirStructureDefinitionList; Overload;
    function GetEnumerator : TFhirStructureDefinitionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirStructureDefinition to the end of the list.
    }
    function Append : TFhirStructureDefinition;

    
    {@member AddItem
      Add an already existing FhirStructureDefinition to the end of the list.
    }
    procedure AddItem(value : TFhirStructureDefinition); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirStructureDefinition) : Integer;
    

    {@member Insert
      Insert FhirStructureDefinition before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirStructureDefinition;
    

    {@member InsertItem
       Insert an existing FhirStructureDefinition before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirStructureDefinition);
    
    {@member Item
       Get the iIndexth FhirStructureDefinition. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirStructureDefinition. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirStructureDefinition);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirStructureDefinition;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirStructureDefinitions[index : Integer] : TFhirStructureDefinition read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSubscriptionChannel : TFhirElement
    Details where to send notifications when resources are received that meet the criteria.
  }
  {!.Net HL7Connect.Fhir.SubscriptionChannel}
  TFhirSubscriptionChannel = class (TFhirBackboneElement)
  private
    FType_ : TFhirEnum;
    FEndpoint : TFhirUri;
    FPayload : TFhirString;
    FHeader : TFhirString;
    Procedure SetType_(value : TFhirEnum);
    Function GetType_ST : TFhirSubscriptionChannelType;
    Procedure SetType_ST(value : TFhirSubscriptionChannelType);
    Procedure SetEndpoint(value : TFhirUri);
    Function GetEndpointST : String;
    Procedure SetEndpointST(value : String);
    Procedure SetPayload(value : TFhirString);
    Function GetPayloadST : String;
    Procedure SetPayloadST(value : String);
    Procedure SetHeader(value : TFhirString);
    Function GetHeaderST : String;
    Procedure SetHeaderST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSubscriptionChannel; overload;
    function Clone : TFhirSubscriptionChannel; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      The type of channel to send notifications on.
    }
    property type_ : TFhirSubscriptionChannelType read GetType_ST write SetType_ST;
    property type_Element : TFhirEnum read FType_ write SetType_;

    {@member endpoint
      Typed access to The uri that describes the actual end-point to send messages to.
    }
    property endpoint : String read GetEndpointST write SetEndpointST;
    {@member endpointElement
      The uri that describes the actual end-point to send messages to.
    }
    property endpointElement : TFhirUri read FEndpoint write SetEndpoint;

    {@member payload
      Typed access to The mime type to send the payload in - either application/xml+fhir, or application/json+fhir. If the mime type is blank, then there is no payload in the notification, just a notification.
    }
    property payload : String read GetPayloadST write SetPayloadST;
    {@member payloadElement
      The mime type to send the payload in - either application/xml+fhir, or application/json+fhir. If the mime type is blank, then there is no payload in the notification, just a notification.
    }
    property payloadElement : TFhirString read FPayload write SetPayload;

    {@member header
      Typed access to Additional headers / information to send as part of the notification.
    }
    property header : String read GetHeaderST write SetHeaderST;
    {@member headerElement
      Additional headers / information to send as part of the notification.
    }
    property headerElement : TFhirString read FHeader write SetHeader;

  end;


  TFhirSubscriptionChannelListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSubscriptionChannelList;
    function GetCurrent : TFhirSubscriptionChannel;
  public
    Constructor Create(list : TFhirSubscriptionChannelList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSubscriptionChannel read GetCurrent;
  end;


  {@Class TFhirSubscriptionChannelList
    A list of FhirSubscriptionChannel
  }
  {!.Net HL7Connect.Fhir.SubscriptionChannelList}
  TFhirSubscriptionChannelList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSubscriptionChannel;
    procedure SetItemN(index : Integer; value : TFhirSubscriptionChannel);
  public
    {!script hide}
    function Link : TFhirSubscriptionChannelList; Overload;
    function Clone : TFhirSubscriptionChannelList; Overload;
    function GetEnumerator : TFhirSubscriptionChannelListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSubscriptionChannel to the end of the list.
    }
    function Append : TFhirSubscriptionChannel;

    
    {@member AddItem
      Add an already existing FhirSubscriptionChannel to the end of the list.
    }
    procedure AddItem(value : TFhirSubscriptionChannel); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSubscriptionChannel) : Integer;
    

    {@member Insert
      Insert FhirSubscriptionChannel before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSubscriptionChannel;
    

    {@member InsertItem
       Insert an existing FhirSubscriptionChannel before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSubscriptionChannel);
    
    {@member Item
       Get the iIndexth FhirSubscriptionChannel. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSubscriptionChannel. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSubscriptionChannel);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSubscriptionChannel;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSubscriptionChannels[index : Integer] : TFhirSubscriptionChannel read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSubscription : TFhirDomainResource
    The subscription resource is used to define a push based subscription from a server to another system. Once a subscription is registered with the server, the server checks every resource that is created or updated, and if the resource matches the given criteria, it sends a message on the defined "channel" so that another system is able to take an appropriate action.
  }
  {!.Net HL7Connect.Fhir.Subscription}
  TFhirSubscription = class (TFhirDomainResource)
  private
    FCriteria : TFhirString;
    FcontactList : TFhirContactPointList;
    FReason : TFhirString;
    FStatus : TFhirEnum;
    FError : TFhirString;
    FChannel : TFhirSubscriptionChannel;
    FEnd_ : TFhirInstant;
    FtagList : TFhirCodingList;
    Procedure SetCriteria(value : TFhirString);
    Function GetCriteriaST : String;
    Procedure SetCriteriaST(value : String);
    function GetContactList : TFhirContactPointList;
    function GetHasContactList : Boolean;
    Procedure SetReason(value : TFhirString);
    Function GetReasonST : String;
    Procedure SetReasonST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirSubscriptionStatus;
    Procedure SetStatusST(value : TFhirSubscriptionStatus);
    Procedure SetError(value : TFhirString);
    Function GetErrorST : String;
    Procedure SetErrorST(value : String);
    Procedure SetChannel(value : TFhirSubscriptionChannel);
    Procedure SetEnd_(value : TFhirInstant);
    Function GetEnd_ST : TDateAndTime;
    Procedure SetEnd_ST(value : TDateAndTime);
    function GetTagList : TFhirCodingList;
    function GetHasTagList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSubscription; overload;
    function Clone : TFhirSubscription; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member criteria
      Typed access to The rules that the server should use to determine when to generate notifications for this subscription.
    }
    property criteria : String read GetCriteriaST write SetCriteriaST;
    {@member criteriaElement
      The rules that the server should use to determine when to generate notifications for this subscription.
    }
    property criteriaElement : TFhirString read FCriteria write SetCriteria;

    {@member contactList
      Contact details for a human to contact about the subscription. The primary use of this for system administrator troubleshooting.
    }
    property contactList : TFhirContactPointList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member reason
      Typed access to A description of why this subscription is defined.
    }
    property reason : String read GetReasonST write SetReasonST;
    {@member reasonElement
      A description of why this subscription is defined.
    }
    property reasonElement : TFhirString read FReason write SetReason;

    {@member status
      The status of the subscription, which marks the server state for managing the subscription.
    }
    property status : TFhirSubscriptionStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member error
      Typed access to A record of the last error that occurred when the server processed a notification.
    }
    property error : String read GetErrorST write SetErrorST;
    {@member errorElement
      A record of the last error that occurred when the server processed a notification.
    }
    property errorElement : TFhirString read FError write SetError;

    {@member channel
      Typed access to Details where to send notifications when resources are received that meet the criteria. (defined for API consistency)
    }
    property channel : TFhirSubscriptionChannel read FChannel write SetChannel;
    {@member channelElement
      Details where to send notifications when resources are received that meet the criteria.
    }
    property channelElement : TFhirSubscriptionChannel read FChannel write SetChannel;

    {@member end_
      Typed access to The time for the server to turn the subscription off.
    }
    property end_ : TDateAndTime read GetEnd_ST write SetEnd_ST;
    {@member end_Element
      The time for the server to turn the subscription off.
    }
    property end_Element : TFhirInstant read FEnd_ write SetEnd_;

    {@member tagList
      A tag to add to any resource that matches the criteria, after the subscription is processed.
    }
    property tagList : TFhirCodingList read GetTagList;
    property hasTagList : boolean read GetHasTagList;

  end;


  TFhirSubscriptionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSubscriptionList;
    function GetCurrent : TFhirSubscription;
  public
    Constructor Create(list : TFhirSubscriptionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSubscription read GetCurrent;
  end;


  {@Class TFhirSubscriptionList
    A list of FhirSubscription
  }
  {!.Net HL7Connect.Fhir.SubscriptionList}
  TFhirSubscriptionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSubscription;
    procedure SetItemN(index : Integer; value : TFhirSubscription);
  public
    {!script hide}
    function Link : TFhirSubscriptionList; Overload;
    function Clone : TFhirSubscriptionList; Overload;
    function GetEnumerator : TFhirSubscriptionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSubscription to the end of the list.
    }
    function Append : TFhirSubscription;

    
    {@member AddItem
      Add an already existing FhirSubscription to the end of the list.
    }
    procedure AddItem(value : TFhirSubscription); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSubscription) : Integer;
    

    {@member Insert
      Insert FhirSubscription before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSubscription;
    

    {@member InsertItem
       Insert an existing FhirSubscription before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSubscription);
    
    {@member Item
       Get the iIndexth FhirSubscription. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSubscription. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSubscription);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSubscription;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSubscriptions[index : Integer] : TFhirSubscription read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSubstanceInstance : TFhirElement
    Substance may be used to describe a kind of substance, or a specific package/container of the substance: an instance.
  }
  {!.Net HL7Connect.Fhir.SubstanceInstance}
  TFhirSubstanceInstance = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirIdentifier;
    FExpiry : TFhirDateTime;
    FQuantity : TFhirQuantity;
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetExpiry(value : TFhirDateTime);
    Function GetExpiryST : TDateAndTime;
    Procedure SetExpiryST(value : TDateAndTime);
    Procedure SetQuantity(value : TFhirQuantity);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSubstanceInstance; overload;
    function Clone : TFhirSubstanceInstance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to Identifier associated with the package/container (usually a label affixed directly). (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifier associated with the package/container (usually a label affixed directly).
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member expiry
      Typed access to When the substance is no longer valid to use. For some substances, a single arbitrary date is used for expiry.
    }
    property expiry : TDateAndTime read GetExpiryST write SetExpiryST;
    {@member expiryElement
      When the substance is no longer valid to use. For some substances, a single arbitrary date is used for expiry.
    }
    property expiryElement : TFhirDateTime read FExpiry write SetExpiry;

    {@member quantity
      Typed access to The amount of the substance. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of the substance.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

  end;


  TFhirSubstanceInstanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSubstanceInstanceList;
    function GetCurrent : TFhirSubstanceInstance;
  public
    Constructor Create(list : TFhirSubstanceInstanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSubstanceInstance read GetCurrent;
  end;


  {@Class TFhirSubstanceInstanceList
    A list of FhirSubstanceInstance
  }
  {!.Net HL7Connect.Fhir.SubstanceInstanceList}
  TFhirSubstanceInstanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSubstanceInstance;
    procedure SetItemN(index : Integer; value : TFhirSubstanceInstance);
  public
    {!script hide}
    function Link : TFhirSubstanceInstanceList; Overload;
    function Clone : TFhirSubstanceInstanceList; Overload;
    function GetEnumerator : TFhirSubstanceInstanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSubstanceInstance to the end of the list.
    }
    function Append : TFhirSubstanceInstance;

    
    {@member AddItem
      Add an already existing FhirSubstanceInstance to the end of the list.
    }
    procedure AddItem(value : TFhirSubstanceInstance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSubstanceInstance) : Integer;
    

    {@member Insert
      Insert FhirSubstanceInstance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSubstanceInstance;
    

    {@member InsertItem
       Insert an existing FhirSubstanceInstance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSubstanceInstance);
    
    {@member Item
       Get the iIndexth FhirSubstanceInstance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSubstanceInstance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSubstanceInstance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSubstanceInstance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSubstanceInstances[index : Integer] : TFhirSubstanceInstance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSubstanceIngredient : TFhirElement
    A substance can be composed of other substances.
  }
  {!.Net HL7Connect.Fhir.SubstanceIngredient}
  TFhirSubstanceIngredient = class (TFhirBackboneElement)
  private
    FQuantity : TFhirRatio;
    FSubstance : TFhirReference{TFhirSubstance};
    Procedure SetQuantity(value : TFhirRatio);
    Procedure SetSubstance(value : TFhirReference{TFhirSubstance});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSubstanceIngredient; overload;
    function Clone : TFhirSubstanceIngredient; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member quantity
      Typed access to The amount of the ingredient in the substance - a concentration ratio. (defined for API consistency)
    }
    property quantity : TFhirRatio read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of the ingredient in the substance - a concentration ratio.
    }
    property quantityElement : TFhirRatio read FQuantity write SetQuantity;

    {@member substance
      Typed access to Another substance that is a component of this substance. (defined for API consistency)
    }
    property substance : TFhirReference{TFhirSubstance} read FSubstance write SetSubstance;
    {@member substanceElement
      Another substance that is a component of this substance.
    }
    property substanceElement : TFhirReference{TFhirSubstance} read FSubstance write SetSubstance;

  end;


  TFhirSubstanceIngredientListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSubstanceIngredientList;
    function GetCurrent : TFhirSubstanceIngredient;
  public
    Constructor Create(list : TFhirSubstanceIngredientList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSubstanceIngredient read GetCurrent;
  end;


  {@Class TFhirSubstanceIngredientList
    A list of FhirSubstanceIngredient
  }
  {!.Net HL7Connect.Fhir.SubstanceIngredientList}
  TFhirSubstanceIngredientList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSubstanceIngredient;
    procedure SetItemN(index : Integer; value : TFhirSubstanceIngredient);
  public
    {!script hide}
    function Link : TFhirSubstanceIngredientList; Overload;
    function Clone : TFhirSubstanceIngredientList; Overload;
    function GetEnumerator : TFhirSubstanceIngredientListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSubstanceIngredient to the end of the list.
    }
    function Append : TFhirSubstanceIngredient;

    
    {@member AddItem
      Add an already existing FhirSubstanceIngredient to the end of the list.
    }
    procedure AddItem(value : TFhirSubstanceIngredient); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSubstanceIngredient) : Integer;
    

    {@member Insert
      Insert FhirSubstanceIngredient before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSubstanceIngredient;
    

    {@member InsertItem
       Insert an existing FhirSubstanceIngredient before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSubstanceIngredient);
    
    {@member Item
       Get the iIndexth FhirSubstanceIngredient. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSubstanceIngredient. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSubstanceIngredient);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSubstanceIngredient;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSubstanceIngredients[index : Integer] : TFhirSubstanceIngredient read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSubstance : TFhirDomainResource
    A homogeneous material with a definite composition.
  }
  {!.Net HL7Connect.Fhir.Substance}
  TFhirSubstance = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FcategoryList : TFhirCodeableConceptList;
    FCode : TFhirCodeableConcept;
    FDescription : TFhirString;
    FinstanceList : TFhirSubstanceInstanceList;
    FingredientList : TFhirSubstanceIngredientList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    function GetCategoryList : TFhirCodeableConceptList;
    function GetHasCategoryList : Boolean;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetInstanceList : TFhirSubstanceInstanceList;
    function GetHasInstanceList : Boolean;
    function GetIngredientList : TFhirSubstanceIngredientList;
    function GetHasIngredientList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSubstance; overload;
    function Clone : TFhirSubstance; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Unique identifier for the substance.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member categoryList
      A code that classifies the general type of substance.  This is used  for searching, sorting and display purposes.
    }
    property categoryList : TFhirCodeableConceptList read GetCategoryList;
    property hasCategoryList : boolean read GetHasCategoryList;

    {@member code
      Typed access to A code (or set of codes) that identify this substance. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      A code (or set of codes) that identify this substance.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member description
      Typed access to A description of the substance - its appearance, handling requirements, and other usage notes.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A description of the substance - its appearance, handling requirements, and other usage notes.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member instanceList
      Substance may be used to describe a kind of substance, or a specific package/container of the substance: an instance.
    }
    property instanceList : TFhirSubstanceInstanceList read GetInstanceList;
    property hasInstanceList : boolean read GetHasInstanceList;

    {@member ingredientList
      A substance can be composed of other substances.
    }
    property ingredientList : TFhirSubstanceIngredientList read GetIngredientList;
    property hasIngredientList : boolean read GetHasIngredientList;

  end;


  TFhirSubstanceListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSubstanceList;
    function GetCurrent : TFhirSubstance;
  public
    Constructor Create(list : TFhirSubstanceList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSubstance read GetCurrent;
  end;


  {@Class TFhirSubstanceList
    A list of FhirSubstance
  }
  {!.Net HL7Connect.Fhir.SubstanceList}
  TFhirSubstanceList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSubstance;
    procedure SetItemN(index : Integer; value : TFhirSubstance);
  public
    {!script hide}
    function Link : TFhirSubstanceList; Overload;
    function Clone : TFhirSubstanceList; Overload;
    function GetEnumerator : TFhirSubstanceListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSubstance to the end of the list.
    }
    function Append : TFhirSubstance;

    
    {@member AddItem
      Add an already existing FhirSubstance to the end of the list.
    }
    procedure AddItem(value : TFhirSubstance); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSubstance) : Integer;
    

    {@member Insert
      Insert FhirSubstance before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSubstance;
    

    {@member InsertItem
       Insert an existing FhirSubstance before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSubstance);
    
    {@member Item
       Get the iIndexth FhirSubstance. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSubstance. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSubstance);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSubstance;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSubstances[index : Integer] : TFhirSubstance read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSupplyDelivery : TFhirDomainResource
    Record of delivery of what is supplied.
  }
  {!.Net HL7Connect.Fhir.SupplyDelivery}
  TFhirSupplyDelivery = class (TFhirDomainResource)
  private
    FIdentifier : TFhirIdentifier;
    FStatus : TFhirEnum;
    FPatient : TFhirReference{TFhirPatient};
    FType_ : TFhirCodeableConcept;
    FQuantity : TFhirQuantity;
    FSuppliedItem : TFhirReference{Resource};
    FSupplier : TFhirReference{TFhirPractitioner};
    FWhenPrepared : TFhirPeriod;
    FTime : TFhirDateTime;
    FDestination : TFhirReference{TFhirLocation};
    FreceiverList : TFhirReferenceList{TFhirPractitioner};
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirSupplydeliveryStatus;
    Procedure SetStatusST(value : TFhirSupplydeliveryStatus);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetType_(value : TFhirCodeableConcept);
    Procedure SetQuantity(value : TFhirQuantity);
    Procedure SetSuppliedItem(value : TFhirReference{Resource});
    Procedure SetSupplier(value : TFhirReference{TFhirPractitioner});
    Procedure SetWhenPrepared(value : TFhirPeriod);
    Procedure SetTime(value : TFhirDateTime);
    Function GetTimeST : TDateAndTime;
    Procedure SetTimeST(value : TDateAndTime);
    Procedure SetDestination(value : TFhirReference{TFhirLocation});
    function GetReceiverList : TFhirReferenceList{TFhirPractitioner};
    function GetHasReceiverList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSupplyDelivery; overload;
    function Clone : TFhirSupplyDelivery; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifier
      Typed access to Identifier assigned by the dispensing facility when the item(s) is dispensed. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifier assigned by the dispensing facility when the item(s) is dispensed.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member status
      A code specifying the state of the dispense event.
    }
    property status : TFhirSupplydeliveryStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member patient
      Typed access to A link to a resource representing the person whom the delivered item is for. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A link to a resource representing the person whom the delivered item is for.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member type_
      Typed access to Indicates the type of dispensing event that is performed. Examples include: Trial Fill, Completion of Trial, Partial Fill, Emergency Fill, Samples, etc. (defined for API consistency)
    }
    property type_ : TFhirCodeableConcept read FType_ write SetType_;
    {@member type_Element
      Indicates the type of dispensing event that is performed. Examples include: Trial Fill, Completion of Trial, Partial Fill, Emergency Fill, Samples, etc.
    }
    property type_Element : TFhirCodeableConcept read FType_ write SetType_;

    {@member quantity
      Typed access to The amount of supply that has been dispensed. Includes unit of measure. (defined for API consistency)
    }
    property quantity : TFhirQuantity read FQuantity write SetQuantity;
    {@member quantityElement
      The amount of supply that has been dispensed. Includes unit of measure.
    }
    property quantityElement : TFhirQuantity read FQuantity write SetQuantity;

    {@member suppliedItem
      Typed access to Identifies the medication, substance or device being dispensed. This is either a link to a resource representing the details of the item or a simple attribute carrying a code that identifies the item from a known list. (defined for API consistency)
    }
    property suppliedItem : TFhirReference{Resource} read FSuppliedItem write SetSuppliedItem;
    {@member suppliedItemElement
      Identifies the medication, substance or device being dispensed. This is either a link to a resource representing the details of the item or a simple attribute carrying a code that identifies the item from a known list.
    }
    property suppliedItemElement : TFhirReference{Resource} read FSuppliedItem write SetSuppliedItem;

    {@member supplier
      Typed access to The individual responsible for dispensing the medication, supplier or device. (defined for API consistency)
    }
    property supplier : TFhirReference{TFhirPractitioner} read FSupplier write SetSupplier;
    {@member supplierElement
      The individual responsible for dispensing the medication, supplier or device.
    }
    property supplierElement : TFhirReference{TFhirPractitioner} read FSupplier write SetSupplier;

    {@member whenPrepared
      Typed access to The time the dispense event occurred. (defined for API consistency)
    }
    property whenPrepared : TFhirPeriod read FWhenPrepared write SetWhenPrepared;
    {@member whenPreparedElement
      The time the dispense event occurred.
    }
    property whenPreparedElement : TFhirPeriod read FWhenPrepared write SetWhenPrepared;

    {@member time
      Typed access to The time the dispensed item was sent or handed to the patient (or agent).
    }
    property time : TDateAndTime read GetTimeST write SetTimeST;
    {@member timeElement
      The time the dispensed item was sent or handed to the patient (or agent).
    }
    property timeElement : TFhirDateTime read FTime write SetTime;

    {@member destination
      Typed access to Identification of the facility/location where the Supply was shipped to, as part of the dispense event. (defined for API consistency)
    }
    property destination : TFhirReference{TFhirLocation} read FDestination write SetDestination;
    {@member destinationElement
      Identification of the facility/location where the Supply was shipped to, as part of the dispense event.
    }
    property destinationElement : TFhirReference{TFhirLocation} read FDestination write SetDestination;

    {@member receiverList
      Identifies the person who picked up the Supply.
    }
    property receiverList : TFhirReferenceList{TFhirPractitioner} read GetReceiverList;
    property hasReceiverList : boolean read GetHasReceiverList;

  end;


  TFhirSupplyDeliveryListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSupplyDeliveryList;
    function GetCurrent : TFhirSupplyDelivery;
  public
    Constructor Create(list : TFhirSupplyDeliveryList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSupplyDelivery read GetCurrent;
  end;


  {@Class TFhirSupplyDeliveryList
    A list of FhirSupplyDelivery
  }
  {!.Net HL7Connect.Fhir.SupplyDeliveryList}
  TFhirSupplyDeliveryList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSupplyDelivery;
    procedure SetItemN(index : Integer; value : TFhirSupplyDelivery);
  public
    {!script hide}
    function Link : TFhirSupplyDeliveryList; Overload;
    function Clone : TFhirSupplyDeliveryList; Overload;
    function GetEnumerator : TFhirSupplyDeliveryListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSupplyDelivery to the end of the list.
    }
    function Append : TFhirSupplyDelivery;

    
    {@member AddItem
      Add an already existing FhirSupplyDelivery to the end of the list.
    }
    procedure AddItem(value : TFhirSupplyDelivery); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSupplyDelivery) : Integer;
    

    {@member Insert
      Insert FhirSupplyDelivery before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSupplyDelivery;
    

    {@member InsertItem
       Insert an existing FhirSupplyDelivery before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSupplyDelivery);
    
    {@member Item
       Get the iIndexth FhirSupplyDelivery. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSupplyDelivery. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSupplyDelivery);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSupplyDelivery;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSupplyDeliveries[index : Integer] : TFhirSupplyDelivery read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSupplyRequestWhen : TFhirElement
    When the request should be fulfilled.
  }
  {!.Net HL7Connect.Fhir.SupplyRequestWhen}
  TFhirSupplyRequestWhen = class (TFhirBackboneElement)
  private
    FCode : TFhirCodeableConcept;
    FSchedule : TFhirTiming;
    Procedure SetCode(value : TFhirCodeableConcept);
    Procedure SetSchedule(value : TFhirTiming);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSupplyRequestWhen; overload;
    function Clone : TFhirSupplyRequestWhen; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Code indicating when the request should be fulfilled. (defined for API consistency)
    }
    property code : TFhirCodeableConcept read FCode write SetCode;
    {@member codeElement
      Code indicating when the request should be fulfilled.
    }
    property codeElement : TFhirCodeableConcept read FCode write SetCode;

    {@member schedule
      Typed access to Formal fulfillment schedule. (defined for API consistency)
    }
    property schedule : TFhirTiming read FSchedule write SetSchedule;
    {@member scheduleElement
      Formal fulfillment schedule.
    }
    property scheduleElement : TFhirTiming read FSchedule write SetSchedule;

  end;


  TFhirSupplyRequestWhenListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSupplyRequestWhenList;
    function GetCurrent : TFhirSupplyRequestWhen;
  public
    Constructor Create(list : TFhirSupplyRequestWhenList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSupplyRequestWhen read GetCurrent;
  end;


  {@Class TFhirSupplyRequestWhenList
    A list of FhirSupplyRequestWhen
  }
  {!.Net HL7Connect.Fhir.SupplyRequestWhenList}
  TFhirSupplyRequestWhenList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSupplyRequestWhen;
    procedure SetItemN(index : Integer; value : TFhirSupplyRequestWhen);
  public
    {!script hide}
    function Link : TFhirSupplyRequestWhenList; Overload;
    function Clone : TFhirSupplyRequestWhenList; Overload;
    function GetEnumerator : TFhirSupplyRequestWhenListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSupplyRequestWhen to the end of the list.
    }
    function Append : TFhirSupplyRequestWhen;

    
    {@member AddItem
      Add an already existing FhirSupplyRequestWhen to the end of the list.
    }
    procedure AddItem(value : TFhirSupplyRequestWhen); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSupplyRequestWhen) : Integer;
    

    {@member Insert
      Insert FhirSupplyRequestWhen before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSupplyRequestWhen;
    

    {@member InsertItem
       Insert an existing FhirSupplyRequestWhen before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSupplyRequestWhen);
    
    {@member Item
       Get the iIndexth FhirSupplyRequestWhen. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSupplyRequestWhen. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSupplyRequestWhen);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSupplyRequestWhen;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSupplyRequestWhens[index : Integer] : TFhirSupplyRequestWhen read GetItemN write SetItemN; default;
  End;


  {@Class TFhirSupplyRequest : TFhirDomainResource
    A record of a request for a medication, substance or device used in the healthcare setting.
  }
  {!.Net HL7Connect.Fhir.SupplyRequest}
  TFhirSupplyRequest = class (TFhirDomainResource)
  private
    FPatient : TFhirReference{TFhirPatient};
    FSource : TFhirReference{Resource};
    FDate : TFhirDateTime;
    FIdentifier : TFhirIdentifier;
    FStatus : TFhirEnum;
    FKind : TFhirCodeableConcept;
    FOrderedItem : TFhirReference{Resource};
    FsupplierList : TFhirReferenceList{TFhirOrganization};
    FReason : TFhirType;
    FWhen : TFhirSupplyRequestWhen;
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetSource(value : TFhirReference{Resource});
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirSupplyrequestStatus;
    Procedure SetStatusST(value : TFhirSupplyrequestStatus);
    Procedure SetKind(value : TFhirCodeableConcept);
    Procedure SetOrderedItem(value : TFhirReference{Resource});
    function GetSupplierList : TFhirReferenceList{TFhirOrganization};
    function GetHasSupplierList : Boolean;
    Procedure SetReason(value : TFhirType);
    Procedure SetWhen(value : TFhirSupplyRequestWhen);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirSupplyRequest; overload;
    function Clone : TFhirSupplyRequest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member patient
      Typed access to A link to a resource representing the person whom the ordered item is for. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A link to a resource representing the person whom the ordered item is for.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member source
      Typed access to The Practitioner , Organization or Patient who initiated this order for the supply. (defined for API consistency)
    }
    property source : TFhirReference{Resource} read FSource write SetSource;
    {@member sourceElement
      The Practitioner , Organization or Patient who initiated this order for the supply.
    }
    property sourceElement : TFhirReference{Resource} read FSource write SetSource;

    {@member date
      Typed access to When the request was made.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      When the request was made.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member identifier
      Typed access to Unique identifier for this supply request. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Unique identifier for this supply request.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member status
      Status of the supply request.
    }
    property status : TFhirSupplyrequestStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member kind
      Typed access to Category of supply, e.g.  central, non-stock, etc. This is used to support work flows associated with the supply process. (defined for API consistency)
    }
    property kind : TFhirCodeableConcept read FKind write SetKind;
    {@member kindElement
      Category of supply, e.g.  central, non-stock, etc. This is used to support work flows associated with the supply process.
    }
    property kindElement : TFhirCodeableConcept read FKind write SetKind;

    {@member orderedItem
      Typed access to The item that is requested to be supplied. (defined for API consistency)
    }
    property orderedItem : TFhirReference{Resource} read FOrderedItem write SetOrderedItem;
    {@member orderedItemElement
      The item that is requested to be supplied.
    }
    property orderedItemElement : TFhirReference{Resource} read FOrderedItem write SetOrderedItem;

    {@member supplierList
      Who is intended to fulfill the request.
    }
    property supplierList : TFhirReferenceList{TFhirOrganization} read GetSupplierList;
    property hasSupplierList : boolean read GetHasSupplierList;

    {@member reason
      Typed access to Why the supply item was requested. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      Why the supply item was requested.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member when
      Typed access to When the request should be fulfilled. (defined for API consistency)
    }
    property when : TFhirSupplyRequestWhen read FWhen write SetWhen;
    {@member whenElement
      When the request should be fulfilled.
    }
    property whenElement : TFhirSupplyRequestWhen read FWhen write SetWhen;

  end;


  TFhirSupplyRequestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirSupplyRequestList;
    function GetCurrent : TFhirSupplyRequest;
  public
    Constructor Create(list : TFhirSupplyRequestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirSupplyRequest read GetCurrent;
  end;


  {@Class TFhirSupplyRequestList
    A list of FhirSupplyRequest
  }
  {!.Net HL7Connect.Fhir.SupplyRequestList}
  TFhirSupplyRequestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirSupplyRequest;
    procedure SetItemN(index : Integer; value : TFhirSupplyRequest);
  public
    {!script hide}
    function Link : TFhirSupplyRequestList; Overload;
    function Clone : TFhirSupplyRequestList; Overload;
    function GetEnumerator : TFhirSupplyRequestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirSupplyRequest to the end of the list.
    }
    function Append : TFhirSupplyRequest;

    
    {@member AddItem
      Add an already existing FhirSupplyRequest to the end of the list.
    }
    procedure AddItem(value : TFhirSupplyRequest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirSupplyRequest) : Integer;
    

    {@member Insert
      Insert FhirSupplyRequest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirSupplyRequest;
    

    {@member InsertItem
       Insert an existing FhirSupplyRequest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirSupplyRequest);
    
    {@member Item
       Get the iIndexth FhirSupplyRequest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirSupplyRequest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirSupplyRequest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirSupplyRequest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirSupplyRequests[index : Integer] : TFhirSupplyRequest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.TestScriptContact}
  TFhirTestScriptContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptContact; overload;
    function Clone : TFhirTestScriptContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the Test Script.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the Test Script.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirTestScriptContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptContactList;
    function GetCurrent : TFhirTestScriptContact;
  public
    Constructor Create(list : TFhirTestScriptContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptContact read GetCurrent;
  end;


  {@Class TFhirTestScriptContactList
    A list of FhirTestScriptContact
  }
  {!.Net HL7Connect.Fhir.TestScriptContactList}
  TFhirTestScriptContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptContact;
    procedure SetItemN(index : Integer; value : TFhirTestScriptContact);
  public
    {!script hide}
    function Link : TFhirTestScriptContactList; Overload;
    function Clone : TFhirTestScriptContactList; Overload;
    function GetEnumerator : TFhirTestScriptContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptContact to the end of the list.
    }
    function Append : TFhirTestScriptContact;

    
    {@member AddItem
      Add an already existing FhirTestScriptContact to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptContact) : Integer;
    

    {@member Insert
      Insert FhirTestScriptContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptContact;
    

    {@member InsertItem
       Insert an existing FhirTestScriptContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptContact);
    
    {@member Item
       Get the iIndexth FhirTestScriptContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptContacts[index : Integer] : TFhirTestScriptContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptMetadata : TFhirElement
    The required capability must exist and are assumed to function correctly on the FHIR server being tested.
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadata}
  TFhirTestScriptMetadata = class (TFhirBackboneElement)
  private
    Flink_List : TFhirTestScriptMetadataLinkList;
    FcapabilityList : TFhirTestScriptMetadataCapabilityList;
    function GetLink_List : TFhirTestScriptMetadataLinkList;
    function GetHasLink_List : Boolean;
    function GetCapabilityList : TFhirTestScriptMetadataCapabilityList;
    function GetHasCapabilityList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptMetadata; overload;
    function Clone : TFhirTestScriptMetadata; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member link_List
      A link to the FHIR specification that this test is covering.
    }
    property link_List : TFhirTestScriptMetadataLinkList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

    {@member capabilityList
      Capabilities that must exist and are assumed to function correctly on the FHIR server being tested.
    }
    property capabilityList : TFhirTestScriptMetadataCapabilityList read GetCapabilityList;
    property hasCapabilityList : boolean read GetHasCapabilityList;

  end;


  TFhirTestScriptMetadataListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptMetadataList;
    function GetCurrent : TFhirTestScriptMetadata;
  public
    Constructor Create(list : TFhirTestScriptMetadataList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptMetadata read GetCurrent;
  end;


  {@Class TFhirTestScriptMetadataList
    A list of FhirTestScriptMetadata
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadataList}
  TFhirTestScriptMetadataList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptMetadata;
    procedure SetItemN(index : Integer; value : TFhirTestScriptMetadata);
  public
    {!script hide}
    function Link : TFhirTestScriptMetadataList; Overload;
    function Clone : TFhirTestScriptMetadataList; Overload;
    function GetEnumerator : TFhirTestScriptMetadataListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptMetadata to the end of the list.
    }
    function Append : TFhirTestScriptMetadata;

    
    {@member AddItem
      Add an already existing FhirTestScriptMetadata to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptMetadata); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptMetadata) : Integer;
    

    {@member Insert
      Insert FhirTestScriptMetadata before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptMetadata;
    

    {@member InsertItem
       Insert an existing FhirTestScriptMetadata before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptMetadata);
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadata. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadata. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptMetadata);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptMetadata;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptMetadata[index : Integer] : TFhirTestScriptMetadata read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptMetadataLink : TFhirElement
    A link to the FHIR specification that this test is covering.
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadataLink}
  TFhirTestScriptMetadataLink = class (TFhirBackboneElement)
  private
    FUrl : TFhirUri;
    FDescription : TFhirString;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptMetadataLink; overload;
    function Clone : TFhirTestScriptMetadataLink; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member url
      Typed access to URL to a particular requirement or feature within the FHIR specification.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      URL to a particular requirement or feature within the FHIR specification.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member description
      Typed access to Short description of the link.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Short description of the link.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

  end;


  TFhirTestScriptMetadataLinkListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptMetadataLinkList;
    function GetCurrent : TFhirTestScriptMetadataLink;
  public
    Constructor Create(list : TFhirTestScriptMetadataLinkList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptMetadataLink read GetCurrent;
  end;


  {@Class TFhirTestScriptMetadataLinkList
    A list of FhirTestScriptMetadataLink
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadataLinkList}
  TFhirTestScriptMetadataLinkList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptMetadataLink;
    procedure SetItemN(index : Integer; value : TFhirTestScriptMetadataLink);
  public
    {!script hide}
    function Link : TFhirTestScriptMetadataLinkList; Overload;
    function Clone : TFhirTestScriptMetadataLinkList; Overload;
    function GetEnumerator : TFhirTestScriptMetadataLinkListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptMetadataLink to the end of the list.
    }
    function Append : TFhirTestScriptMetadataLink;

    
    {@member AddItem
      Add an already existing FhirTestScriptMetadataLink to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptMetadataLink); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptMetadataLink) : Integer;
    

    {@member Insert
      Insert FhirTestScriptMetadataLink before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptMetadataLink;
    

    {@member InsertItem
       Insert an existing FhirTestScriptMetadataLink before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptMetadataLink);
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadataLink. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadataLink. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptMetadataLink);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptMetadataLink;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptMetadataLinks[index : Integer] : TFhirTestScriptMetadataLink read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptMetadataCapability : TFhirElement
    Capabilities that must exist and are assumed to function correctly on the FHIR server being tested.
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadataCapability}
  TFhirTestScriptMetadataCapability = class (TFhirBackboneElement)
  private
    FRequired : TFhirBoolean;
    FValidated : TFhirBoolean;
    FDescription : TFhirString;
    FDestination : TFhirInteger;
    Flink_List : TFhirUriList;
    FConformance : TFhirReference{TFhirConformance};
    Procedure SetRequired(value : TFhirBoolean);
    Function GetRequiredST : Boolean;
    Procedure SetRequiredST(value : Boolean);
    Procedure SetValidated(value : TFhirBoolean);
    Function GetValidatedST : Boolean;
    Procedure SetValidatedST(value : Boolean);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetDestination(value : TFhirInteger);
    Function GetDestinationST : String;
    Procedure SetDestinationST(value : String);
    function GetLink_List : TFhirUriList;
    function GetHasLink_List : Boolean;
    Procedure SetConformance(value : TFhirReference{TFhirConformance});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptMetadataCapability; overload;
    function Clone : TFhirTestScriptMetadataCapability; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member required
      Typed access to Whether or not the test execution will require the given capabilities of the server in order for this test script to execute.
    }
    property required : Boolean read GetRequiredST write SetRequiredST;
    {@member requiredElement
      Whether or not the test execution will require the given capabilities of the server in order for this test script to execute.
    }
    property requiredElement : TFhirBoolean read FRequired write SetRequired;

    {@member validated
      Typed access to Whether or not the test execution will validate the given capabilities of the server in order for this test script to execute.
    }
    property validated : Boolean read GetValidatedST write SetValidatedST;
    {@member validatedElement
      Whether or not the test execution will validate the given capabilities of the server in order for this test script to execute.
    }
    property validatedElement : TFhirBoolean read FValidated write SetValidated;

    {@member description
      Typed access to Description of the capabilities that this test script is requiring the server to support.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      Description of the capabilities that this test script is requiring the server to support.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member destination
      Typed access to Which server these requirements apply to.
    }
    property destination : String read GetDestinationST write SetDestinationST;
    {@member destinationElement
      Which server these requirements apply to.
    }
    property destinationElement : TFhirInteger read FDestination write SetDestination;

    {@member link_List
      Links to the FHIR specification that describes this interaction and the resources involved in more detail.
    }
    property link_List : TFhirUriList read GetLink_List;
    property hasLink_List : boolean read GetHasLink_List;

    {@member conformance
      Typed access to Minimum conformance required of server for test script to execute successfully.   If server does not meet at a minimum the reference conformance definition, then all tests in this script are skipped. (defined for API consistency)
    }
    property conformance : TFhirReference{TFhirConformance} read FConformance write SetConformance;
    {@member conformanceElement
      Minimum conformance required of server for test script to execute successfully.   If server does not meet at a minimum the reference conformance definition, then all tests in this script are skipped.
    }
    property conformanceElement : TFhirReference{TFhirConformance} read FConformance write SetConformance;

  end;


  TFhirTestScriptMetadataCapabilityListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptMetadataCapabilityList;
    function GetCurrent : TFhirTestScriptMetadataCapability;
  public
    Constructor Create(list : TFhirTestScriptMetadataCapabilityList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptMetadataCapability read GetCurrent;
  end;


  {@Class TFhirTestScriptMetadataCapabilityList
    A list of FhirTestScriptMetadataCapability
  }
  {!.Net HL7Connect.Fhir.TestScriptMetadataCapabilityList}
  TFhirTestScriptMetadataCapabilityList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptMetadataCapability;
    procedure SetItemN(index : Integer; value : TFhirTestScriptMetadataCapability);
  public
    {!script hide}
    function Link : TFhirTestScriptMetadataCapabilityList; Overload;
    function Clone : TFhirTestScriptMetadataCapabilityList; Overload;
    function GetEnumerator : TFhirTestScriptMetadataCapabilityListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptMetadataCapability to the end of the list.
    }
    function Append : TFhirTestScriptMetadataCapability;

    
    {@member AddItem
      Add an already existing FhirTestScriptMetadataCapability to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptMetadataCapability); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptMetadataCapability) : Integer;
    

    {@member Insert
      Insert FhirTestScriptMetadataCapability before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptMetadataCapability;
    

    {@member InsertItem
       Insert an existing FhirTestScriptMetadataCapability before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptMetadataCapability);
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadataCapability. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptMetadataCapability. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptMetadataCapability);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptMetadataCapability;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptMetadataCapabilities[index : Integer] : TFhirTestScriptMetadataCapability read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptFixture : TFhirElement
    Fixture in the test script - by reference (uri). All fixtures are required for the test script to execute.
  }
  {!.Net HL7Connect.Fhir.TestScriptFixture}
  TFhirTestScriptFixture = class (TFhirBackboneElement)
  private
    FAutocreate : TFhirBoolean;
    FAutodelete : TFhirBoolean;
    FResource : TFhirReference{TFhirReference};
    Procedure SetAutocreate(value : TFhirBoolean);
    Function GetAutocreateST : Boolean;
    Procedure SetAutocreateST(value : Boolean);
    Procedure SetAutodelete(value : TFhirBoolean);
    Function GetAutodeleteST : Boolean;
    Procedure SetAutodeleteST(value : Boolean);
    Procedure SetResource(value : TFhirReference{TFhirReference});
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptFixture; overload;
    function Clone : TFhirTestScriptFixture; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member autocreate
      Typed access to Whether or not to implicitly create the fixture during setup. If true, the fixture is automatically created on each server being tested during setup, therefore no create operation is required for this fixture in the TestScript.setup section.
    }
    property autocreate : Boolean read GetAutocreateST write SetAutocreateST;
    {@member autocreateElement
      Whether or not to implicitly create the fixture during setup. If true, the fixture is automatically created on each server being tested during setup, therefore no create operation is required for this fixture in the TestScript.setup section.
    }
    property autocreateElement : TFhirBoolean read FAutocreate write SetAutocreate;

    {@member autodelete
      Typed access to Whether or not to implicitly delete the fixture during teardown If true, the fixture is automatically deleted on each server being tested during teardown, therefore no delete operation is required for this fixture in the TestScript.teardown section.
    }
    property autodelete : Boolean read GetAutodeleteST write SetAutodeleteST;
    {@member autodeleteElement
      Whether or not to implicitly delete the fixture during teardown If true, the fixture is automatically deleted on each server being tested during teardown, therefore no delete operation is required for this fixture in the TestScript.teardown section.
    }
    property autodeleteElement : TFhirBoolean read FAutodelete write SetAutodelete;

    {@member resource
      Typed access to Reference to the resource (containing the contents of the resource needed for operations). (defined for API consistency)
    }
    property resource : TFhirReference{TFhirReference} read FResource write SetResource;
    {@member resourceElement
      Reference to the resource (containing the contents of the resource needed for operations).
    }
    property resourceElement : TFhirReference{TFhirReference} read FResource write SetResource;

  end;


  TFhirTestScriptFixtureListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptFixtureList;
    function GetCurrent : TFhirTestScriptFixture;
  public
    Constructor Create(list : TFhirTestScriptFixtureList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptFixture read GetCurrent;
  end;


  {@Class TFhirTestScriptFixtureList
    A list of FhirTestScriptFixture
  }
  {!.Net HL7Connect.Fhir.TestScriptFixtureList}
  TFhirTestScriptFixtureList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptFixture;
    procedure SetItemN(index : Integer; value : TFhirTestScriptFixture);
  public
    {!script hide}
    function Link : TFhirTestScriptFixtureList; Overload;
    function Clone : TFhirTestScriptFixtureList; Overload;
    function GetEnumerator : TFhirTestScriptFixtureListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptFixture to the end of the list.
    }
    function Append : TFhirTestScriptFixture;

    
    {@member AddItem
      Add an already existing FhirTestScriptFixture to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptFixture); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptFixture) : Integer;
    

    {@member Insert
      Insert FhirTestScriptFixture before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptFixture;
    

    {@member InsertItem
       Insert an existing FhirTestScriptFixture before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptFixture);
    
    {@member Item
       Get the iIndexth FhirTestScriptFixture. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptFixture. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptFixture);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptFixture;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptFixtures[index : Integer] : TFhirTestScriptFixture read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptVariable : TFhirElement
    Variable is set based either on element value in response body or on header field value in the response headers.
  }
  {!.Net HL7Connect.Fhir.TestScriptVariable}
  TFhirTestScriptVariable = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FHeaderField : TFhirString;
    FPath : TFhirString;
    FSourceId : TFhirId;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetHeaderField(value : TFhirString);
    Function GetHeaderFieldST : String;
    Procedure SetHeaderFieldST(value : String);
    Procedure SetPath(value : TFhirString);
    Function GetPathST : String;
    Procedure SetPathST(value : String);
    Procedure SetSourceId(value : TFhirId);
    Function GetSourceIdST : String;
    Procedure SetSourceIdST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptVariable; overload;
    function Clone : TFhirTestScriptVariable; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to Descriptive name for this variable.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      Descriptive name for this variable.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member headerField
      Typed access to Will be used to grab the HTTP header field value from the headers that sourceId is pointing to.
    }
    property headerField : String read GetHeaderFieldST write SetHeaderFieldST;
    {@member headerFieldElement
      Will be used to grab the HTTP header field value from the headers that sourceId is pointing to.
    }
    property headerFieldElement : TFhirString read FHeaderField write SetHeaderField;

    {@member path
      Typed access to XPath or JSONPath against the fixture body.  When variables are defined, either headerField must be specified or path, but not both.
    }
    property path : String read GetPathST write SetPathST;
    {@member pathElement
      XPath or JSONPath against the fixture body.  When variables are defined, either headerField must be specified or path, but not both.
    }
    property pathElement : TFhirString read FPath write SetPath;

    {@member sourceId
      Typed access to Fixture to evaluate the XPath/JSONPath expression or the headerField  against within this variable.
    }
    property sourceId : String read GetSourceIdST write SetSourceIdST;
    {@member sourceIdElement
      Fixture to evaluate the XPath/JSONPath expression or the headerField  against within this variable.
    }
    property sourceIdElement : TFhirId read FSourceId write SetSourceId;

  end;


  TFhirTestScriptVariableListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptVariableList;
    function GetCurrent : TFhirTestScriptVariable;
  public
    Constructor Create(list : TFhirTestScriptVariableList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptVariable read GetCurrent;
  end;


  {@Class TFhirTestScriptVariableList
    A list of FhirTestScriptVariable
  }
  {!.Net HL7Connect.Fhir.TestScriptVariableList}
  TFhirTestScriptVariableList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptVariable;
    procedure SetItemN(index : Integer; value : TFhirTestScriptVariable);
  public
    {!script hide}
    function Link : TFhirTestScriptVariableList; Overload;
    function Clone : TFhirTestScriptVariableList; Overload;
    function GetEnumerator : TFhirTestScriptVariableListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptVariable to the end of the list.
    }
    function Append : TFhirTestScriptVariable;

    
    {@member AddItem
      Add an already existing FhirTestScriptVariable to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptVariable); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptVariable) : Integer;
    

    {@member Insert
      Insert FhirTestScriptVariable before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptVariable;
    

    {@member InsertItem
       Insert an existing FhirTestScriptVariable before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptVariable);
    
    {@member Item
       Get the iIndexth FhirTestScriptVariable. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptVariable. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptVariable);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptVariable;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptVariables[index : Integer] : TFhirTestScriptVariable read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptSetup : TFhirElement
    A series of required setup operations before tests are executed.
  }
  {!.Net HL7Connect.Fhir.TestScriptSetup}
  TFhirTestScriptSetup = class (TFhirBackboneElement)
  private
    FMetadata : TFhirTestScriptMetadata;
    FactionList : TFhirTestScriptSetupActionList;
    Procedure SetMetadata(value : TFhirTestScriptMetadata);
    function GetActionList : TFhirTestScriptSetupActionList;
    function GetHasActionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptSetup; overload;
    function Clone : TFhirTestScriptSetup; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member metadata
      Typed access to Capabilities that must exist and are assumed to function correctly on the FHIR server being tested. (defined for API consistency)
    }
    property metadata : TFhirTestScriptMetadata read FMetadata write SetMetadata;
    {@member metadataElement
      Capabilities that must exist and are assumed to function correctly on the FHIR server being tested.
    }
    property metadataElement : TFhirTestScriptMetadata read FMetadata write SetMetadata;

    {@member actionList
      Action would contain either an operation or an assertion.
    }
    property actionList : TFhirTestScriptSetupActionList read GetActionList;
    property hasActionList : boolean read GetHasActionList;

  end;


  TFhirTestScriptSetupListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptSetupList;
    function GetCurrent : TFhirTestScriptSetup;
  public
    Constructor Create(list : TFhirTestScriptSetupList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptSetup read GetCurrent;
  end;


  {@Class TFhirTestScriptSetupList
    A list of FhirTestScriptSetup
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupList}
  TFhirTestScriptSetupList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptSetup;
    procedure SetItemN(index : Integer; value : TFhirTestScriptSetup);
  public
    {!script hide}
    function Link : TFhirTestScriptSetupList; Overload;
    function Clone : TFhirTestScriptSetupList; Overload;
    function GetEnumerator : TFhirTestScriptSetupListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptSetup to the end of the list.
    }
    function Append : TFhirTestScriptSetup;

    
    {@member AddItem
      Add an already existing FhirTestScriptSetup to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptSetup); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptSetup) : Integer;
    

    {@member Insert
      Insert FhirTestScriptSetup before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptSetup;
    

    {@member InsertItem
       Insert an existing FhirTestScriptSetup before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptSetup);
    
    {@member Item
       Get the iIndexth FhirTestScriptSetup. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptSetup. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptSetup);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptSetup;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptSetups[index : Integer] : TFhirTestScriptSetup read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptSetupAction : TFhirElement
    Action would contain either an operation or an assertion.
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupAction}
  TFhirTestScriptSetupAction = class (TFhirBackboneElement)
  private
    FOperation : TFhirTestScriptSetupActionOperation;
    FAssert : TFhirTestScriptSetupActionAssert;
    Procedure SetOperation(value : TFhirTestScriptSetupActionOperation);
    Procedure SetAssert(value : TFhirTestScriptSetupActionAssert);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptSetupAction; overload;
    function Clone : TFhirTestScriptSetupAction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member operation
      Typed access to The operation to perform. (defined for API consistency)
    }
    property operation : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;
    {@member operationElement
      The operation to perform.
    }
    property operationElement : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;

    {@member assert
      Typed access to Evaluates the results of previous operations to determine if the server under test behaves appropriately. (defined for API consistency)
    }
    property assert : TFhirTestScriptSetupActionAssert read FAssert write SetAssert;
    {@member assertElement
      Evaluates the results of previous operations to determine if the server under test behaves appropriately.
    }
    property assertElement : TFhirTestScriptSetupActionAssert read FAssert write SetAssert;

  end;


  TFhirTestScriptSetupActionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptSetupActionList;
    function GetCurrent : TFhirTestScriptSetupAction;
  public
    Constructor Create(list : TFhirTestScriptSetupActionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptSetupAction read GetCurrent;
  end;


  {@Class TFhirTestScriptSetupActionList
    A list of FhirTestScriptSetupAction
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionList}
  TFhirTestScriptSetupActionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptSetupAction;
    procedure SetItemN(index : Integer; value : TFhirTestScriptSetupAction);
  public
    {!script hide}
    function Link : TFhirTestScriptSetupActionList; Overload;
    function Clone : TFhirTestScriptSetupActionList; Overload;
    function GetEnumerator : TFhirTestScriptSetupActionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptSetupAction to the end of the list.
    }
    function Append : TFhirTestScriptSetupAction;

    
    {@member AddItem
      Add an already existing FhirTestScriptSetupAction to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptSetupAction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptSetupAction) : Integer;
    

    {@member Insert
      Insert FhirTestScriptSetupAction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptSetupAction;
    

    {@member InsertItem
       Insert an existing FhirTestScriptSetupAction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptSetupAction);
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupAction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupAction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptSetupAction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptSetupAction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptSetupActions[index : Integer] : TFhirTestScriptSetupAction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptSetupActionOperation : TFhirElement
    The operation to perform.
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionOperation}
  TFhirTestScriptSetupActionOperation = class (TFhirBackboneElement)
  private
    FType_ : TFhirCoding;
    FResource : TFhirCode;
    FLabel_ : TFhirString;
    FDescription : TFhirString;
    FAccept : TFhirEnum;
    FContentType : TFhirEnum;
    FDestination : TFhirInteger;
    FEncodeRequestUrl : TFhirBoolean;
    FParams : TFhirString;
    FrequestHeaderList : TFhirTestScriptSetupActionOperationRequestHeaderList;
    FResponseId : TFhirId;
    FSourceId : TFhirId;
    FTargetId : TFhirId;
    FUrl : TFhirString;
    Procedure SetType_(value : TFhirCoding);
    Procedure SetResource(value : TFhirCode);
    Function GetResourceST : String;
    Procedure SetResourceST(value : String);
    Procedure SetLabel_(value : TFhirString);
    Function GetLabel_ST : String;
    Procedure SetLabel_ST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetAccept(value : TFhirEnum);
    Function GetAcceptST : TFhirContentType;
    Procedure SetAcceptST(value : TFhirContentType);
    Procedure SetContentType(value : TFhirEnum);
    Function GetContentTypeST : TFhirContentType;
    Procedure SetContentTypeST(value : TFhirContentType);
    Procedure SetDestination(value : TFhirInteger);
    Function GetDestinationST : String;
    Procedure SetDestinationST(value : String);
    Procedure SetEncodeRequestUrl(value : TFhirBoolean);
    Function GetEncodeRequestUrlST : Boolean;
    Procedure SetEncodeRequestUrlST(value : Boolean);
    Procedure SetParams(value : TFhirString);
    Function GetParamsST : String;
    Procedure SetParamsST(value : String);
    function GetRequestHeaderList : TFhirTestScriptSetupActionOperationRequestHeaderList;
    function GetHasRequestHeaderList : Boolean;
    Procedure SetResponseId(value : TFhirId);
    Function GetResponseIdST : String;
    Procedure SetResponseIdST(value : String);
    Procedure SetSourceId(value : TFhirId);
    Function GetSourceIdST : String;
    Procedure SetSourceIdST(value : String);
    Procedure SetTargetId(value : TFhirId);
    Function GetTargetIdST : String;
    Procedure SetTargetIdST(value : String);
    Procedure SetUrl(value : TFhirString);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptSetupActionOperation; overload;
    function Clone : TFhirTestScriptSetupActionOperation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member type_
      Typed access to Server interaction or operation type. (defined for API consistency)
    }
    property type_ : TFhirCoding read FType_ write SetType_;
    {@member type_Element
      Server interaction or operation type.
    }
    property type_Element : TFhirCoding read FType_ write SetType_;

    {@member resource
      Typed access to The type of the resource.  See http://hl7-fhir.github.io/resourcelist.html.
    }
    property resource : String read GetResourceST write SetResourceST;
    {@member resourceElement
      The type of the resource.  See http://hl7-fhir.github.io/resourcelist.html.
    }
    property resourceElement : TFhirCode read FResource write SetResource;

    {@member label_
      Typed access to The label would be used for tracking/logging purposes by test engines.
    }
    property label_ : String read GetLabel_ST write SetLabel_ST;
    {@member label_Element
      The label would be used for tracking/logging purposes by test engines.
    }
    property label_Element : TFhirString read FLabel_ write SetLabel_;

    {@member description
      Typed access to The description would be used by test engines for tracking and reporting purposes.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      The description would be used by test engines for tracking and reporting purposes.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member accept
      The content-type or mime-type to use for RESTful operation in the 'Accept' header.
    }
    property accept : TFhirContentType read GetAcceptST write SetAcceptST;
    property acceptElement : TFhirEnum read FAccept write SetAccept;

    {@member contentType
      The content-type or mime-type to use for RESTful operation in the 'Content-Type' header.
    }
    property contentType : TFhirContentType read GetContentTypeST write SetContentTypeST;
    property contentTypeElement : TFhirEnum read FContentType write SetContentType;

    {@member destination
      Typed access to Which server to perform the operation on.
    }
    property destination : String read GetDestinationST write SetDestinationST;
    {@member destinationElement
      Which server to perform the operation on.
    }
    property destinationElement : TFhirInteger read FDestination write SetDestination;

    {@member encodeRequestUrl
      Typed access to Whether or not to implicitly send the request url in encoded format. The default is true to match the standard RESTful client behavior. Set to false when communicating with a server that does not support encoded url paths.
    }
    property encodeRequestUrl : Boolean read GetEncodeRequestUrlST write SetEncodeRequestUrlST;
    {@member encodeRequestUrlElement
      Whether or not to implicitly send the request url in encoded format. The default is true to match the standard RESTful client behavior. Set to false when communicating with a server that does not support encoded url paths.
    }
    property encodeRequestUrlElement : TFhirBoolean read FEncodeRequestUrl write SetEncodeRequestUrl;

    {@member params
      Typed access to Path plus parameters after [type].  Used to set parts of the request URL explicitly.
    }
    property params : String read GetParamsST write SetParamsST;
    {@member paramsElement
      Path plus parameters after [type].  Used to set parts of the request URL explicitly.
    }
    property paramsElement : TFhirString read FParams write SetParams;

    {@member requestHeaderList
      Header elements would be used to set HTTP headers.
    }
    property requestHeaderList : TFhirTestScriptSetupActionOperationRequestHeaderList read GetRequestHeaderList;
    property hasRequestHeaderList : boolean read GetHasRequestHeaderList;

    {@member responseId
      Typed access to The fixture id (maybe new) to map to the response.
    }
    property responseId : String read GetResponseIdST write SetResponseIdST;
    {@member responseIdElement
      The fixture id (maybe new) to map to the response.
    }
    property responseIdElement : TFhirId read FResponseId write SetResponseId;

    {@member sourceId
      Typed access to The id of the fixture used as the body of a PUT or POST request.
    }
    property sourceId : String read GetSourceIdST write SetSourceIdST;
    {@member sourceIdElement
      The id of the fixture used as the body of a PUT or POST request.
    }
    property sourceIdElement : TFhirId read FSourceId write SetSourceId;

    {@member targetId
      Typed access to Id of fixture used for extracting the [id],  [type], and [vid] for GET requests.
    }
    property targetId : String read GetTargetIdST write SetTargetIdST;
    {@member targetIdElement
      Id of fixture used for extracting the [id],  [type], and [vid] for GET requests.
    }
    property targetIdElement : TFhirId read FTargetId write SetTargetId;

    {@member url
      Typed access to Complete request URL.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      Complete request URL.
    }
    property urlElement : TFhirString read FUrl write SetUrl;

  end;


  TFhirTestScriptSetupActionOperationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptSetupActionOperationList;
    function GetCurrent : TFhirTestScriptSetupActionOperation;
  public
    Constructor Create(list : TFhirTestScriptSetupActionOperationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptSetupActionOperation read GetCurrent;
  end;


  {@Class TFhirTestScriptSetupActionOperationList
    A list of FhirTestScriptSetupActionOperation
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionOperationList}
  TFhirTestScriptSetupActionOperationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptSetupActionOperation;
    procedure SetItemN(index : Integer; value : TFhirTestScriptSetupActionOperation);
  public
    {!script hide}
    function Link : TFhirTestScriptSetupActionOperationList; Overload;
    function Clone : TFhirTestScriptSetupActionOperationList; Overload;
    function GetEnumerator : TFhirTestScriptSetupActionOperationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptSetupActionOperation to the end of the list.
    }
    function Append : TFhirTestScriptSetupActionOperation;

    
    {@member AddItem
      Add an already existing FhirTestScriptSetupActionOperation to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptSetupActionOperation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptSetupActionOperation) : Integer;
    

    {@member Insert
      Insert FhirTestScriptSetupActionOperation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptSetupActionOperation;
    

    {@member InsertItem
       Insert an existing FhirTestScriptSetupActionOperation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptSetupActionOperation);
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionOperation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionOperation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptSetupActionOperation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptSetupActionOperation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptSetupActionOperations[index : Integer] : TFhirTestScriptSetupActionOperation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptSetupActionOperationRequestHeader : TFhirElement
    Header elements would be used to set HTTP headers.
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionOperationRequestHeader}
  TFhirTestScriptSetupActionOperationRequestHeader = class (TFhirBackboneElement)
  private
    FField : TFhirString;
    FValue : TFhirString;
    Procedure SetField(value : TFhirString);
    Function GetFieldST : String;
    Procedure SetFieldST(value : String);
    Procedure SetValue(value : TFhirString);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptSetupActionOperationRequestHeader; overload;
    function Clone : TFhirTestScriptSetupActionOperationRequestHeader; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member field
      Typed access to The HTTP header field e.g. "Accept".
    }
    property field : String read GetFieldST write SetFieldST;
    {@member fieldElement
      The HTTP header field e.g. "Accept".
    }
    property fieldElement : TFhirString read FField write SetField;

    {@member value
      Typed access to The value of the header e.g. "application/xml".
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      The value of the header e.g. "application/xml".
    }
    property valueElement : TFhirString read FValue write SetValue;

  end;


  TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptSetupActionOperationRequestHeaderList;
    function GetCurrent : TFhirTestScriptSetupActionOperationRequestHeader;
  public
    Constructor Create(list : TFhirTestScriptSetupActionOperationRequestHeaderList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptSetupActionOperationRequestHeader read GetCurrent;
  end;


  {@Class TFhirTestScriptSetupActionOperationRequestHeaderList
    A list of FhirTestScriptSetupActionOperationRequestHeader
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionOperationRequestHeaderList}
  TFhirTestScriptSetupActionOperationRequestHeaderList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptSetupActionOperationRequestHeader;
    procedure SetItemN(index : Integer; value : TFhirTestScriptSetupActionOperationRequestHeader);
  public
    {!script hide}
    function Link : TFhirTestScriptSetupActionOperationRequestHeaderList; Overload;
    function Clone : TFhirTestScriptSetupActionOperationRequestHeaderList; Overload;
    function GetEnumerator : TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptSetupActionOperationRequestHeader to the end of the list.
    }
    function Append : TFhirTestScriptSetupActionOperationRequestHeader;

    
    {@member AddItem
      Add an already existing FhirTestScriptSetupActionOperationRequestHeader to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptSetupActionOperationRequestHeader); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptSetupActionOperationRequestHeader) : Integer;
    

    {@member Insert
      Insert FhirTestScriptSetupActionOperationRequestHeader before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptSetupActionOperationRequestHeader;
    

    {@member InsertItem
       Insert an existing FhirTestScriptSetupActionOperationRequestHeader before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptSetupActionOperationRequestHeader);
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionOperationRequestHeader. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionOperationRequestHeader. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptSetupActionOperationRequestHeader);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptSetupActionOperationRequestHeader;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptSetupActionOperationRequestHeaders[index : Integer] : TFhirTestScriptSetupActionOperationRequestHeader read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptSetupActionAssert : TFhirElement
    Evaluates the results of previous operations to determine if the server under test behaves appropriately.
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionAssert}
  TFhirTestScriptSetupActionAssert = class (TFhirBackboneElement)
  private
    FLabel_ : TFhirString;
    FDescription : TFhirString;
    FDirection : TFhirEnum;
    FCompareToSourceId : TFhirString;
    FCompareToSourcePath : TFhirString;
    FContentType : TFhirEnum;
    FHeaderField : TFhirString;
    FMinimumId : TFhirString;
    FNavigationLinks : TFhirBoolean;
    FOperator : TFhirEnum;
    FPath : TFhirString;
    FResource : TFhirCode;
    FResponse : TFhirEnum;
    FResponseCode : TFhirString;
    FSourceId : TFhirId;
    FValidateProfileId : TFhirId;
    FValue : TFhirString;
    FWarningOnly : TFhirBoolean;
    Procedure SetLabel_(value : TFhirString);
    Function GetLabel_ST : String;
    Procedure SetLabel_ST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetDirection(value : TFhirEnum);
    Function GetDirectionST : TFhirAssertDirectionCodes;
    Procedure SetDirectionST(value : TFhirAssertDirectionCodes);
    Procedure SetCompareToSourceId(value : TFhirString);
    Function GetCompareToSourceIdST : String;
    Procedure SetCompareToSourceIdST(value : String);
    Procedure SetCompareToSourcePath(value : TFhirString);
    Function GetCompareToSourcePathST : String;
    Procedure SetCompareToSourcePathST(value : String);
    Procedure SetContentType(value : TFhirEnum);
    Function GetContentTypeST : TFhirContentType;
    Procedure SetContentTypeST(value : TFhirContentType);
    Procedure SetHeaderField(value : TFhirString);
    Function GetHeaderFieldST : String;
    Procedure SetHeaderFieldST(value : String);
    Procedure SetMinimumId(value : TFhirString);
    Function GetMinimumIdST : String;
    Procedure SetMinimumIdST(value : String);
    Procedure SetNavigationLinks(value : TFhirBoolean);
    Function GetNavigationLinksST : Boolean;
    Procedure SetNavigationLinksST(value : Boolean);
    Procedure SetOperator(value : TFhirEnum);
    Function GetOperatorST : TFhirAssertOperatorCodes;
    Procedure SetOperatorST(value : TFhirAssertOperatorCodes);
    Procedure SetPath(value : TFhirString);
    Function GetPathST : String;
    Procedure SetPathST(value : String);
    Procedure SetResource(value : TFhirCode);
    Function GetResourceST : String;
    Procedure SetResourceST(value : String);
    Procedure SetResponse(value : TFhirEnum);
    Function GetResponseST : TFhirAssertResponseCodeTypes;
    Procedure SetResponseST(value : TFhirAssertResponseCodeTypes);
    Procedure SetResponseCode(value : TFhirString);
    Function GetResponseCodeST : String;
    Procedure SetResponseCodeST(value : String);
    Procedure SetSourceId(value : TFhirId);
    Function GetSourceIdST : String;
    Procedure SetSourceIdST(value : String);
    Procedure SetValidateProfileId(value : TFhirId);
    Function GetValidateProfileIdST : String;
    Procedure SetValidateProfileIdST(value : String);
    Procedure SetValue(value : TFhirString);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
    Procedure SetWarningOnly(value : TFhirBoolean);
    Function GetWarningOnlyST : Boolean;
    Procedure SetWarningOnlyST(value : Boolean);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptSetupActionAssert; overload;
    function Clone : TFhirTestScriptSetupActionAssert; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member label_
      Typed access to The label would be used for tracking/logging purposes by test engines.
    }
    property label_ : String read GetLabel_ST write SetLabel_ST;
    {@member label_Element
      The label would be used for tracking/logging purposes by test engines.
    }
    property label_Element : TFhirString read FLabel_ write SetLabel_;

    {@member description
      Typed access to The description would be used by test engines for tracking and reporting purposes.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      The description would be used by test engines for tracking and reporting purposes.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member direction
      The direction to use for the assertion.
    }
    property direction : TFhirAssertDirectionCodes read GetDirectionST write SetDirectionST;
    property directionElement : TFhirEnum read FDirection write SetDirection;

    {@member compareToSourceId
      Typed access to Id of fixture used to compare the "sourceId/path" evaluations to.
    }
    property compareToSourceId : String read GetCompareToSourceIdST write SetCompareToSourceIdST;
    {@member compareToSourceIdElement
      Id of fixture used to compare the "sourceId/path" evaluations to.
    }
    property compareToSourceIdElement : TFhirString read FCompareToSourceId write SetCompareToSourceId;

    {@member compareToSourcePath
      Typed access to XPath or JSONPath expression against fixture used to compare the "sourceId/path" evaluations to.
    }
    property compareToSourcePath : String read GetCompareToSourcePathST write SetCompareToSourcePathST;
    {@member compareToSourcePathElement
      XPath or JSONPath expression against fixture used to compare the "sourceId/path" evaluations to.
    }
    property compareToSourcePathElement : TFhirString read FCompareToSourcePath write SetCompareToSourcePath;

    {@member contentType
      The content-type or mime-type to use for RESTful operation in the 'Content-Type' header.
    }
    property contentType : TFhirContentType read GetContentTypeST write SetContentTypeST;
    property contentTypeElement : TFhirEnum read FContentType write SetContentType;

    {@member headerField
      Typed access to The HTTP header field name e.g. 'Location'.
    }
    property headerField : String read GetHeaderFieldST write SetHeaderFieldST;
    {@member headerFieldElement
      The HTTP header field name e.g. 'Location'.
    }
    property headerFieldElement : TFhirString read FHeaderField write SetHeaderField;

    {@member minimumId
      Typed access to The ID of a fixture.  Asserts that the response contains at a minimumId the fixture specified by minimumId.
    }
    property minimumId : String read GetMinimumIdST write SetMinimumIdST;
    {@member minimumIdElement
      The ID of a fixture.  Asserts that the response contains at a minimumId the fixture specified by minimumId.
    }
    property minimumIdElement : TFhirString read FMinimumId write SetMinimumId;

    {@member navigationLinks
      Typed access to Whether or not the test execution performs validation on the bundle navigation links.
    }
    property navigationLinks : Boolean read GetNavigationLinksST write SetNavigationLinksST;
    {@member navigationLinksElement
      Whether or not the test execution performs validation on the bundle navigation links.
    }
    property navigationLinksElement : TFhirBoolean read FNavigationLinks write SetNavigationLinks;

    {@member operator
      The operator type.
    }
    property operator : TFhirAssertOperatorCodes read GetOperatorST write SetOperatorST;
    property operatorElement : TFhirEnum read FOperator write SetOperator;

    {@member path
      Typed access to The XPath or JSONPath expression to be evaluated against the fixture representing the response received from server.
    }
    property path : String read GetPathST write SetPathST;
    {@member pathElement
      The XPath or JSONPath expression to be evaluated against the fixture representing the response received from server.
    }
    property pathElement : TFhirString read FPath write SetPath;

    {@member resource
      Typed access to The type of the resource.  See http://hl7-fhir.github.io/resourcelist.html.
    }
    property resource : String read GetResourceST write SetResourceST;
    {@member resourceElement
      The type of the resource.  See http://hl7-fhir.github.io/resourcelist.html.
    }
    property resourceElement : TFhirCode read FResource write SetResource;

    {@member response
      okay | created | noContent | notModified | bad | forbidden | notFound | methodNotAllowed | conflict | gone | preconditionFailed | unprocessable.
    }
    property response : TFhirAssertResponseCodeTypes read GetResponseST write SetResponseST;
    property responseElement : TFhirEnum read FResponse write SetResponse;

    {@member responseCode
      Typed access to The value of the HTTP response code to be tested.
    }
    property responseCode : String read GetResponseCodeST write SetResponseCodeST;
    {@member responseCodeElement
      The value of the HTTP response code to be tested.
    }
    property responseCodeElement : TFhirString read FResponseCode write SetResponseCode;

    {@member sourceId
      Typed access to Fixture to evaluate the XPath/JSONPath expression or the headerField  against.
    }
    property sourceId : String read GetSourceIdST write SetSourceIdST;
    {@member sourceIdElement
      Fixture to evaluate the XPath/JSONPath expression or the headerField  against.
    }
    property sourceIdElement : TFhirId read FSourceId write SetSourceId;

    {@member validateProfileId
      Typed access to The ID of the Profile to validate against.
    }
    property validateProfileId : String read GetValidateProfileIdST write SetValidateProfileIdST;
    {@member validateProfileIdElement
      The ID of the Profile to validate against.
    }
    property validateProfileIdElement : TFhirId read FValidateProfileId write SetValidateProfileId;

    {@member value
      Typed access to The value to compare to.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      The value to compare to.
    }
    property valueElement : TFhirString read FValue write SetValue;

    {@member warningOnly
      Typed access to Whether or not the test execution will produce a warning only on error for this assert.
    }
    property warningOnly : Boolean read GetWarningOnlyST write SetWarningOnlyST;
    {@member warningOnlyElement
      Whether or not the test execution will produce a warning only on error for this assert.
    }
    property warningOnlyElement : TFhirBoolean read FWarningOnly write SetWarningOnly;

  end;


  TFhirTestScriptSetupActionAssertListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptSetupActionAssertList;
    function GetCurrent : TFhirTestScriptSetupActionAssert;
  public
    Constructor Create(list : TFhirTestScriptSetupActionAssertList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptSetupActionAssert read GetCurrent;
  end;


  {@Class TFhirTestScriptSetupActionAssertList
    A list of FhirTestScriptSetupActionAssert
  }
  {!.Net HL7Connect.Fhir.TestScriptSetupActionAssertList}
  TFhirTestScriptSetupActionAssertList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptSetupActionAssert;
    procedure SetItemN(index : Integer; value : TFhirTestScriptSetupActionAssert);
  public
    {!script hide}
    function Link : TFhirTestScriptSetupActionAssertList; Overload;
    function Clone : TFhirTestScriptSetupActionAssertList; Overload;
    function GetEnumerator : TFhirTestScriptSetupActionAssertListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptSetupActionAssert to the end of the list.
    }
    function Append : TFhirTestScriptSetupActionAssert;

    
    {@member AddItem
      Add an already existing FhirTestScriptSetupActionAssert to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptSetupActionAssert); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptSetupActionAssert) : Integer;
    

    {@member Insert
      Insert FhirTestScriptSetupActionAssert before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptSetupActionAssert;
    

    {@member InsertItem
       Insert an existing FhirTestScriptSetupActionAssert before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptSetupActionAssert);
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionAssert. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptSetupActionAssert. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptSetupActionAssert);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptSetupActionAssert;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptSetupActionAsserts[index : Integer] : TFhirTestScriptSetupActionAssert read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptTest : TFhirElement
    A test in this script.
  }
  {!.Net HL7Connect.Fhir.TestScriptTest}
  TFhirTestScriptTest = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FDescription : TFhirString;
    FMetadata : TFhirTestScriptMetadata;
    FactionList : TFhirTestScriptTestActionList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    Procedure SetMetadata(value : TFhirTestScriptMetadata);
    function GetActionList : TFhirTestScriptTestActionList;
    function GetHasActionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptTest; overload;
    function Clone : TFhirTestScriptTest; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of this test used for tracking/logging purposes by test engines.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of this test used for tracking/logging purposes by test engines.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member description
      Typed access to A short description of the test used by test engines for tracking and reporting purposes.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A short description of the test used by test engines for tracking and reporting purposes.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member metadata
      Typed access to Capabilities that must exist and are assumed to function correctly on the FHIR server being tested. (defined for API consistency)
    }
    property metadata : TFhirTestScriptMetadata read FMetadata write SetMetadata;
    {@member metadataElement
      Capabilities that must exist and are assumed to function correctly on the FHIR server being tested.
    }
    property metadataElement : TFhirTestScriptMetadata read FMetadata write SetMetadata;

    {@member actionList
      Action would contain either an operation or an assertion.
    }
    property actionList : TFhirTestScriptTestActionList read GetActionList;
    property hasActionList : boolean read GetHasActionList;

  end;


  TFhirTestScriptTestListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptTestList;
    function GetCurrent : TFhirTestScriptTest;
  public
    Constructor Create(list : TFhirTestScriptTestList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptTest read GetCurrent;
  end;


  {@Class TFhirTestScriptTestList
    A list of FhirTestScriptTest
  }
  {!.Net HL7Connect.Fhir.TestScriptTestList}
  TFhirTestScriptTestList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptTest;
    procedure SetItemN(index : Integer; value : TFhirTestScriptTest);
  public
    {!script hide}
    function Link : TFhirTestScriptTestList; Overload;
    function Clone : TFhirTestScriptTestList; Overload;
    function GetEnumerator : TFhirTestScriptTestListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptTest to the end of the list.
    }
    function Append : TFhirTestScriptTest;

    
    {@member AddItem
      Add an already existing FhirTestScriptTest to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptTest); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptTest) : Integer;
    

    {@member Insert
      Insert FhirTestScriptTest before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptTest;
    

    {@member InsertItem
       Insert an existing FhirTestScriptTest before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptTest);
    
    {@member Item
       Get the iIndexth FhirTestScriptTest. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptTest. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptTest);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptTest;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptTests[index : Integer] : TFhirTestScriptTest read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptTestAction : TFhirElement
    Action would contain either an operation or an assertion.
  }
  {!.Net HL7Connect.Fhir.TestScriptTestAction}
  TFhirTestScriptTestAction = class (TFhirBackboneElement)
  private
    FOperation : TFhirTestScriptSetupActionOperation;
    FAssert : TFhirTestScriptSetupActionAssert;
    Procedure SetOperation(value : TFhirTestScriptSetupActionOperation);
    Procedure SetAssert(value : TFhirTestScriptSetupActionAssert);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptTestAction; overload;
    function Clone : TFhirTestScriptTestAction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member operation
      Typed access to An operation would involve a REST request to a server. (defined for API consistency)
    }
    property operation : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;
    {@member operationElement
      An operation would involve a REST request to a server.
    }
    property operationElement : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;

    {@member assert
      Typed access to Evaluates the results of previous operations to determine if the server under test behaves appropriately. (defined for API consistency)
    }
    property assert : TFhirTestScriptSetupActionAssert read FAssert write SetAssert;
    {@member assertElement
      Evaluates the results of previous operations to determine if the server under test behaves appropriately.
    }
    property assertElement : TFhirTestScriptSetupActionAssert read FAssert write SetAssert;

  end;


  TFhirTestScriptTestActionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptTestActionList;
    function GetCurrent : TFhirTestScriptTestAction;
  public
    Constructor Create(list : TFhirTestScriptTestActionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptTestAction read GetCurrent;
  end;


  {@Class TFhirTestScriptTestActionList
    A list of FhirTestScriptTestAction
  }
  {!.Net HL7Connect.Fhir.TestScriptTestActionList}
  TFhirTestScriptTestActionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptTestAction;
    procedure SetItemN(index : Integer; value : TFhirTestScriptTestAction);
  public
    {!script hide}
    function Link : TFhirTestScriptTestActionList; Overload;
    function Clone : TFhirTestScriptTestActionList; Overload;
    function GetEnumerator : TFhirTestScriptTestActionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptTestAction to the end of the list.
    }
    function Append : TFhirTestScriptTestAction;

    
    {@member AddItem
      Add an already existing FhirTestScriptTestAction to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptTestAction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptTestAction) : Integer;
    

    {@member Insert
      Insert FhirTestScriptTestAction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptTestAction;
    

    {@member InsertItem
       Insert an existing FhirTestScriptTestAction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptTestAction);
    
    {@member Item
       Get the iIndexth FhirTestScriptTestAction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptTestAction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptTestAction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptTestAction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptTestActions[index : Integer] : TFhirTestScriptTestAction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptTeardown : TFhirElement
    A series of operations required to clean up after the all the tests are executed (successfully or otherwise).
  }
  {!.Net HL7Connect.Fhir.TestScriptTeardown}
  TFhirTestScriptTeardown = class (TFhirBackboneElement)
  private
    FactionList : TFhirTestScriptTeardownActionList;
    function GetActionList : TFhirTestScriptTeardownActionList;
    function GetHasActionList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptTeardown; overload;
    function Clone : TFhirTestScriptTeardown; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member actionList
      The teardown action will only contain an operation.
    }
    property actionList : TFhirTestScriptTeardownActionList read GetActionList;
    property hasActionList : boolean read GetHasActionList;

  end;


  TFhirTestScriptTeardownListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptTeardownList;
    function GetCurrent : TFhirTestScriptTeardown;
  public
    Constructor Create(list : TFhirTestScriptTeardownList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptTeardown read GetCurrent;
  end;


  {@Class TFhirTestScriptTeardownList
    A list of FhirTestScriptTeardown
  }
  {!.Net HL7Connect.Fhir.TestScriptTeardownList}
  TFhirTestScriptTeardownList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptTeardown;
    procedure SetItemN(index : Integer; value : TFhirTestScriptTeardown);
  public
    {!script hide}
    function Link : TFhirTestScriptTeardownList; Overload;
    function Clone : TFhirTestScriptTeardownList; Overload;
    function GetEnumerator : TFhirTestScriptTeardownListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptTeardown to the end of the list.
    }
    function Append : TFhirTestScriptTeardown;

    
    {@member AddItem
      Add an already existing FhirTestScriptTeardown to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptTeardown); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptTeardown) : Integer;
    

    {@member Insert
      Insert FhirTestScriptTeardown before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptTeardown;
    

    {@member InsertItem
       Insert an existing FhirTestScriptTeardown before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptTeardown);
    
    {@member Item
       Get the iIndexth FhirTestScriptTeardown. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptTeardown. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptTeardown);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptTeardown;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptTeardowns[index : Integer] : TFhirTestScriptTeardown read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScriptTeardownAction : TFhirElement
    The teardown action will only contain an operation.
  }
  {!.Net HL7Connect.Fhir.TestScriptTeardownAction}
  TFhirTestScriptTeardownAction = class (TFhirBackboneElement)
  private
    FOperation : TFhirTestScriptSetupActionOperation;
    Procedure SetOperation(value : TFhirTestScriptSetupActionOperation);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScriptTeardownAction; overload;
    function Clone : TFhirTestScriptTeardownAction; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member operation
      Typed access to An operation would involve a REST request to a server. (defined for API consistency)
    }
    property operation : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;
    {@member operationElement
      An operation would involve a REST request to a server.
    }
    property operationElement : TFhirTestScriptSetupActionOperation read FOperation write SetOperation;

  end;


  TFhirTestScriptTeardownActionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptTeardownActionList;
    function GetCurrent : TFhirTestScriptTeardownAction;
  public
    Constructor Create(list : TFhirTestScriptTeardownActionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScriptTeardownAction read GetCurrent;
  end;


  {@Class TFhirTestScriptTeardownActionList
    A list of FhirTestScriptTeardownAction
  }
  {!.Net HL7Connect.Fhir.TestScriptTeardownActionList}
  TFhirTestScriptTeardownActionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScriptTeardownAction;
    procedure SetItemN(index : Integer; value : TFhirTestScriptTeardownAction);
  public
    {!script hide}
    function Link : TFhirTestScriptTeardownActionList; Overload;
    function Clone : TFhirTestScriptTeardownActionList; Overload;
    function GetEnumerator : TFhirTestScriptTeardownActionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScriptTeardownAction to the end of the list.
    }
    function Append : TFhirTestScriptTeardownAction;

    
    {@member AddItem
      Add an already existing FhirTestScriptTeardownAction to the end of the list.
    }
    procedure AddItem(value : TFhirTestScriptTeardownAction); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScriptTeardownAction) : Integer;
    

    {@member Insert
      Insert FhirTestScriptTeardownAction before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScriptTeardownAction;
    

    {@member InsertItem
       Insert an existing FhirTestScriptTeardownAction before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScriptTeardownAction);
    
    {@member Item
       Get the iIndexth FhirTestScriptTeardownAction. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScriptTeardownAction. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScriptTeardownAction);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScriptTeardownAction;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScriptTeardownActions[index : Integer] : TFhirTestScriptTeardownAction read GetItemN write SetItemN; default;
  End;


  {@Class TFhirTestScript : TFhirDomainResource
    TestScript is a resource that specifies a suite of tests against a FHIR server implementation to determine compliance against the FHIR specification.
  }
  {!.Net HL7Connect.Fhir.TestScript}
  TFhirTestScript = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FIdentifier : TFhirIdentifier;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirTestScriptContactList;
    FDate : TFhirDateTime;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FRequirements : TFhirString;
    FCopyright : TFhirString;
    FMetadata : TFhirTestScriptMetadata;
    FMultiserver : TFhirBoolean;
    FfixtureList : TFhirTestScriptFixtureList;
    FprofileList : TFhirReferenceList{TFhirReference};
    FvariableList : TFhirTestScriptVariableList;
    FSetup : TFhirTestScriptSetup;
    FtestList : TFhirTestScriptTestList;
    FTeardown : TFhirTestScriptTeardown;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirTestScriptContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetMetadata(value : TFhirTestScriptMetadata);
    Procedure SetMultiserver(value : TFhirBoolean);
    Function GetMultiserverST : Boolean;
    Procedure SetMultiserverST(value : Boolean);
    function GetFixtureList : TFhirTestScriptFixtureList;
    function GetHasFixtureList : Boolean;
    function GetProfileList : TFhirReferenceList{TFhirReference};
    function GetHasProfileList : Boolean;
    function GetVariableList : TFhirTestScriptVariableList;
    function GetHasVariableList : Boolean;
    Procedure SetSetup(value : TFhirTestScriptSetup);
    function GetTestList : TFhirTestScriptTestList;
    function GetHasTestList : Boolean;
    Procedure SetTeardown(value : TFhirTestScriptTeardown);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirTestScript; overload;
    function Clone : TFhirTestScript; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this Test Script. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this Test Script is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this Test Script. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this Test Script is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member version
      Typed access to The identifier that is used to identify this version of the TestScript. This is an arbitrary value managed by the TestScript author manually.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The identifier that is used to identify this version of the TestScript. This is an arbitrary value managed by the TestScript author manually.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name identifying the TestScript.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name identifying the TestScript.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the TestScript.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member identifier
      Typed access to Identifier for the TestScript assigned for external purposes outside the context of FHIR. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Identifier for the TestScript assigned for external purposes outside the context of FHIR.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member experimental
      Typed access to This TestScript was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This TestScript was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the Test Script.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the Test Script.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirTestScriptContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date this version of the test tcript was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the test cases change.
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date this version of the test tcript was published. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the test cases change.
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member description
      Typed access to A free text natural language description of the TestScript and its use.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the TestScript and its use.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of Test Scripts.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member requirements
      Typed access to Explains why this Test Script is needed and why it's been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this Test Script is needed and why it's been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member copyright
      Typed access to A copyright statement relating to the Test Script and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the Test Script and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the details of the constraints and mappings.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member metadata
      Typed access to The required capability must exist and are assumed to function correctly on the FHIR server being tested. (defined for API consistency)
    }
    property metadata : TFhirTestScriptMetadata read FMetadata write SetMetadata;
    {@member metadataElement
      The required capability must exist and are assumed to function correctly on the FHIR server being tested.
    }
    property metadataElement : TFhirTestScriptMetadata read FMetadata write SetMetadata;

    {@member multiserver
      Typed access to If the tests apply to more than one FHIR server (e.g. cross-server interoperability tests) then multiserver=true. Defaults to false if value is unspecified.
    }
    property multiserver : Boolean read GetMultiserverST write SetMultiserverST;
    {@member multiserverElement
      If the tests apply to more than one FHIR server (e.g. cross-server interoperability tests) then multiserver=true. Defaults to false if value is unspecified.
    }
    property multiserverElement : TFhirBoolean read FMultiserver write SetMultiserver;

    {@member fixtureList
      Fixture in the test script - by reference (uri). All fixtures are required for the test script to execute.
    }
    property fixtureList : TFhirTestScriptFixtureList read GetFixtureList;
    property hasFixtureList : boolean read GetHasFixtureList;

    {@member profileList
      Reference to the profile to be used for validation.
    }
    property profileList : TFhirReferenceList{TFhirReference} read GetProfileList;
    property hasProfileList : boolean read GetHasProfileList;

    {@member variableList
      Variable is set based either on element value in response body or on header field value in the response headers.
    }
    property variableList : TFhirTestScriptVariableList read GetVariableList;
    property hasVariableList : boolean read GetHasVariableList;

    {@member setup
      Typed access to A series of required setup operations before tests are executed. (defined for API consistency)
    }
    property setup : TFhirTestScriptSetup read FSetup write SetSetup;
    {@member setupElement
      A series of required setup operations before tests are executed.
    }
    property setupElement : TFhirTestScriptSetup read FSetup write SetSetup;

    {@member testList
      A test in this script.
    }
    property testList : TFhirTestScriptTestList read GetTestList;
    property hasTestList : boolean read GetHasTestList;

    {@member teardown
      Typed access to A series of operations required to clean up after the all the tests are executed (successfully or otherwise). (defined for API consistency)
    }
    property teardown : TFhirTestScriptTeardown read FTeardown write SetTeardown;
    {@member teardownElement
      A series of operations required to clean up after the all the tests are executed (successfully or otherwise).
    }
    property teardownElement : TFhirTestScriptTeardown read FTeardown write SetTeardown;

  end;


  TFhirTestScriptListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirTestScriptList;
    function GetCurrent : TFhirTestScript;
  public
    Constructor Create(list : TFhirTestScriptList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirTestScript read GetCurrent;
  end;


  {@Class TFhirTestScriptList
    A list of FhirTestScript
  }
  {!.Net HL7Connect.Fhir.TestScriptList}
  TFhirTestScriptList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirTestScript;
    procedure SetItemN(index : Integer; value : TFhirTestScript);
  public
    {!script hide}
    function Link : TFhirTestScriptList; Overload;
    function Clone : TFhirTestScriptList; Overload;
    function GetEnumerator : TFhirTestScriptListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirTestScript to the end of the list.
    }
    function Append : TFhirTestScript;

    
    {@member AddItem
      Add an already existing FhirTestScript to the end of the list.
    }
    procedure AddItem(value : TFhirTestScript); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirTestScript) : Integer;
    

    {@member Insert
      Insert FhirTestScript before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirTestScript;
    

    {@member InsertItem
       Insert an existing FhirTestScript before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirTestScript);
    
    {@member Item
       Get the iIndexth FhirTestScript. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirTestScript. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirTestScript);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirTestScript;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirTestScripts[index : Integer] : TFhirTestScript read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetContact : TFhirElement
    Contacts to assist a user in finding and communicating with the publisher.
  }
  {!.Net HL7Connect.Fhir.ValueSetContact}
  TFhirValueSetContact = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FtelecomList : TFhirContactPointList;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    function GetTelecomList : TFhirContactPointList;
    function GetHasTelecomList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetContact; overload;
    function Clone : TFhirValueSetContact; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of an individual to contact regarding the value set.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of an individual to contact regarding the value set.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member telecomList
      Contact details for individual (if a name was provided) or the publisher.
    }
    property telecomList : TFhirContactPointList read GetTelecomList;
    property hasTelecomList : boolean read GetHasTelecomList;

  end;


  TFhirValueSetContactListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetContactList;
    function GetCurrent : TFhirValueSetContact;
  public
    Constructor Create(list : TFhirValueSetContactList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetContact read GetCurrent;
  end;


  {@Class TFhirValueSetContactList
    A list of FhirValueSetContact
  }
  {!.Net HL7Connect.Fhir.ValueSetContactList}
  TFhirValueSetContactList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetContact;
    procedure SetItemN(index : Integer; value : TFhirValueSetContact);
  public
    {!script hide}
    function Link : TFhirValueSetContactList; Overload;
    function Clone : TFhirValueSetContactList; Overload;
    function GetEnumerator : TFhirValueSetContactListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetContact to the end of the list.
    }
    function Append : TFhirValueSetContact;

    
    {@member AddItem
      Add an already existing FhirValueSetContact to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetContact); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetContact) : Integer;
    

    {@member Insert
      Insert FhirValueSetContact before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetContact;
    

    {@member InsertItem
       Insert an existing FhirValueSetContact before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetContact);
    
    {@member Item
       Get the iIndexth FhirValueSetContact. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetContact. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetContact);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetContact;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetContacts[index : Integer] : TFhirValueSetContact read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetCodeSystem : TFhirElement
    A definition of a code system, inlined into the value set (as a packaging convenience). Note that the inline code system may be used from other value sets by referring to its (codeSystem.system) directly.
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystem}
  TFhirValueSetCodeSystem = class (TFhirBackboneElement)
  private
    FSystem : TFhirUri;
    FVersion : TFhirString;
    FCaseSensitive : TFhirBoolean;
    FconceptList : TFhirValueSetCodeSystemConceptList;
    Procedure SetSystem(value : TFhirUri);
    Function GetSystemST : String;
    Procedure SetSystemST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetCaseSensitive(value : TFhirBoolean);
    Function GetCaseSensitiveST : Boolean;
    Procedure SetCaseSensitiveST(value : Boolean);
    function GetConceptList : TFhirValueSetCodeSystemConceptList;
    function GetHasConceptList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetCodeSystem; overload;
    function Clone : TFhirValueSetCodeSystem; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member system
      Typed access to An absolute URI that is used to reference this code system, including in [Coding].system.
    }
    property system : String read GetSystemST write SetSystemST;
    {@member systemElement
      An absolute URI that is used to reference this code system, including in [Coding].system.
    }
    property systemElement : TFhirUri read FSystem write SetSystem;

    {@member version
      Typed access to The version of this code system that defines the codes. Note that the version is optional because a well maintained code system does not suffer from versioning, and therefore the version does not need to be maintained. However many code systems are not well maintained, and the version needs to be defined and tracked.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version of this code system that defines the codes. Note that the version is optional because a well maintained code system does not suffer from versioning, and therefore the version does not need to be maintained. However many code systems are not well maintained, and the version needs to be defined and tracked.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member caseSensitive
      Typed access to If code comparison is case sensitive when codes within this system are compared to each other.
    }
    property caseSensitive : Boolean read GetCaseSensitiveST write SetCaseSensitiveST;
    {@member caseSensitiveElement
      If code comparison is case sensitive when codes within this system are compared to each other.
    }
    property caseSensitiveElement : TFhirBoolean read FCaseSensitive write SetCaseSensitive;

    {@member conceptList
      Concepts that are in the code system. The concept definitions are inherently hierarchical, but the definitions must be consulted to determine what the meaning of the hierarchical relationships are.
    }
    property conceptList : TFhirValueSetCodeSystemConceptList read GetConceptList;
    property hasConceptList : boolean read GetHasConceptList;

  end;


  TFhirValueSetCodeSystemListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetCodeSystemList;
    function GetCurrent : TFhirValueSetCodeSystem;
  public
    Constructor Create(list : TFhirValueSetCodeSystemList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetCodeSystem read GetCurrent;
  end;


  {@Class TFhirValueSetCodeSystemList
    A list of FhirValueSetCodeSystem
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystemList}
  TFhirValueSetCodeSystemList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetCodeSystem;
    procedure SetItemN(index : Integer; value : TFhirValueSetCodeSystem);
  public
    {!script hide}
    function Link : TFhirValueSetCodeSystemList; Overload;
    function Clone : TFhirValueSetCodeSystemList; Overload;
    function GetEnumerator : TFhirValueSetCodeSystemListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetCodeSystem to the end of the list.
    }
    function Append : TFhirValueSetCodeSystem;

    
    {@member AddItem
      Add an already existing FhirValueSetCodeSystem to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetCodeSystem); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetCodeSystem) : Integer;
    

    {@member Insert
      Insert FhirValueSetCodeSystem before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetCodeSystem;
    

    {@member InsertItem
       Insert an existing FhirValueSetCodeSystem before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetCodeSystem);
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystem. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystem. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetCodeSystem);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetCodeSystem;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetCodeSystems[index : Integer] : TFhirValueSetCodeSystem read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetCodeSystemConcept : TFhirElement
    Concepts that are in the code system. The concept definitions are inherently hierarchical, but the definitions must be consulted to determine what the meaning of the hierarchical relationships are.
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystemConcept}
  TFhirValueSetCodeSystemConcept = class (TFhirBackboneElement)
  private
    FCode : TFhirCode;
    FAbstract : TFhirBoolean;
    FDisplay : TFhirString;
    FDefinition : TFhirString;
    FdesignationList : TFhirValueSetCodeSystemConceptDesignationList;
    FconceptList : TFhirValueSetCodeSystemConceptList;
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetAbstract(value : TFhirBoolean);
    Function GetAbstractST : Boolean;
    Procedure SetAbstractST(value : Boolean);
    Procedure SetDisplay(value : TFhirString);
    Function GetDisplayST : String;
    Procedure SetDisplayST(value : String);
    Procedure SetDefinition(value : TFhirString);
    Function GetDefinitionST : String;
    Procedure SetDefinitionST(value : String);
    function GetDesignationList : TFhirValueSetCodeSystemConceptDesignationList;
    function GetHasDesignationList : Boolean;
    function GetConceptList : TFhirValueSetCodeSystemConceptList;
    function GetHasConceptList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetCodeSystemConcept; overload;
    function Clone : TFhirValueSetCodeSystemConcept; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to A code - a text symbol - that uniquely identifies the concept within the code system.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      A code - a text symbol - that uniquely identifies the concept within the code system.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member abstract
      Typed access to If this code is not for use as a real concept.
    }
    property abstract : Boolean read GetAbstractST write SetAbstractST;
    {@member abstractElement
      If this code is not for use as a real concept.
    }
    property abstractElement : TFhirBoolean read FAbstract write SetAbstract;

    {@member display
      Typed access to A human readable string that is the recommended default way to present this concept to a user.
    }
    property display : String read GetDisplayST write SetDisplayST;
    {@member displayElement
      A human readable string that is the recommended default way to present this concept to a user.
    }
    property displayElement : TFhirString read FDisplay write SetDisplay;

    {@member definition
      Typed access to The formal definition of the concept. The value set resource does not make formal definitions required, because of the prevalence of legacy systems. However, they are highly recommended, as without them there is no formal meaning associated with the concept.
    }
    property definition : String read GetDefinitionST write SetDefinitionST;
    {@member definitionElement
      The formal definition of the concept. The value set resource does not make formal definitions required, because of the prevalence of legacy systems. However, they are highly recommended, as without them there is no formal meaning associated with the concept.
    }
    property definitionElement : TFhirString read FDefinition write SetDefinition;

    {@member designationList
      Additional representations for the concept - other languages, aliases, specialized purposes, used for particular purposes, etc.
    }
    property designationList : TFhirValueSetCodeSystemConceptDesignationList read GetDesignationList;
    property hasDesignationList : boolean read GetHasDesignationList;

    {@member conceptList
      Defines children of a concept to produce a hierarchy of concepts. The nature of the relationships is variable (is-a/contains/categorizes) and can only be determined by examining the definitions of the concepts.
    }
    property conceptList : TFhirValueSetCodeSystemConceptList read GetConceptList;
    property hasConceptList : boolean read GetHasConceptList;

  end;


  TFhirValueSetCodeSystemConceptListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetCodeSystemConceptList;
    function GetCurrent : TFhirValueSetCodeSystemConcept;
  public
    Constructor Create(list : TFhirValueSetCodeSystemConceptList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetCodeSystemConcept read GetCurrent;
  end;


  {@Class TFhirValueSetCodeSystemConceptList
    A list of FhirValueSetCodeSystemConcept
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystemConceptList}
  TFhirValueSetCodeSystemConceptList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetCodeSystemConcept;
    procedure SetItemN(index : Integer; value : TFhirValueSetCodeSystemConcept);
  public
    {!script hide}
    function Link : TFhirValueSetCodeSystemConceptList; Overload;
    function Clone : TFhirValueSetCodeSystemConceptList; Overload;
    function GetEnumerator : TFhirValueSetCodeSystemConceptListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetCodeSystemConcept to the end of the list.
    }
    function Append : TFhirValueSetCodeSystemConcept;

    
    {@member AddItem
      Add an already existing FhirValueSetCodeSystemConcept to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetCodeSystemConcept); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetCodeSystemConcept) : Integer;
    

    {@member Insert
      Insert FhirValueSetCodeSystemConcept before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetCodeSystemConcept;
    

    {@member InsertItem
       Insert an existing FhirValueSetCodeSystemConcept before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetCodeSystemConcept);
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystemConcept. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystemConcept. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetCodeSystemConcept);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetCodeSystemConcept;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetCodeSystemConcepts[index : Integer] : TFhirValueSetCodeSystemConcept read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetCodeSystemConceptDesignation : TFhirElement
    Additional representations for the concept - other languages, aliases, specialized purposes, used for particular purposes, etc.
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystemConceptDesignation}
  TFhirValueSetCodeSystemConceptDesignation = class (TFhirBackboneElement)
  private
    FLanguage : TFhirCode;
    FUse : TFhirCoding;
    FValue : TFhirString;
    Procedure SetLanguage(value : TFhirCode);
    Function GetLanguageST : String;
    Procedure SetLanguageST(value : String);
    Procedure SetUse(value : TFhirCoding);
    Procedure SetValue(value : TFhirString);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetCodeSystemConceptDesignation; overload;
    function Clone : TFhirValueSetCodeSystemConceptDesignation; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member language
      Typed access to The language this designation is defined for.
    }
    property language : String read GetLanguageST write SetLanguageST;
    {@member languageElement
      The language this designation is defined for.
    }
    property languageElement : TFhirCode read FLanguage write SetLanguage;

    {@member use
      Typed access to A code that details how this designation would be used. (defined for API consistency)
    }
    property use : TFhirCoding read FUse write SetUse;
    {@member useElement
      A code that details how this designation would be used.
    }
    property useElement : TFhirCoding read FUse write SetUse;

    {@member value
      Typed access to The text value for this designation.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      The text value for this designation.
    }
    property valueElement : TFhirString read FValue write SetValue;

  end;


  TFhirValueSetCodeSystemConceptDesignationListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetCodeSystemConceptDesignationList;
    function GetCurrent : TFhirValueSetCodeSystemConceptDesignation;
  public
    Constructor Create(list : TFhirValueSetCodeSystemConceptDesignationList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetCodeSystemConceptDesignation read GetCurrent;
  end;


  {@Class TFhirValueSetCodeSystemConceptDesignationList
    A list of FhirValueSetCodeSystemConceptDesignation
  }
  {!.Net HL7Connect.Fhir.ValueSetCodeSystemConceptDesignationList}
  TFhirValueSetCodeSystemConceptDesignationList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetCodeSystemConceptDesignation;
    procedure SetItemN(index : Integer; value : TFhirValueSetCodeSystemConceptDesignation);
  public
    {!script hide}
    function Link : TFhirValueSetCodeSystemConceptDesignationList; Overload;
    function Clone : TFhirValueSetCodeSystemConceptDesignationList; Overload;
    function GetEnumerator : TFhirValueSetCodeSystemConceptDesignationListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetCodeSystemConceptDesignation to the end of the list.
    }
    function Append : TFhirValueSetCodeSystemConceptDesignation;

    
    {@member AddItem
      Add an already existing FhirValueSetCodeSystemConceptDesignation to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetCodeSystemConceptDesignation); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetCodeSystemConceptDesignation) : Integer;
    

    {@member Insert
      Insert FhirValueSetCodeSystemConceptDesignation before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetCodeSystemConceptDesignation;
    

    {@member InsertItem
       Insert an existing FhirValueSetCodeSystemConceptDesignation before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetCodeSystemConceptDesignation);
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystemConceptDesignation. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetCodeSystemConceptDesignation. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetCodeSystemConceptDesignation);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetCodeSystemConceptDesignation;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetCodeSystemConceptDesignations[index : Integer] : TFhirValueSetCodeSystemConceptDesignation read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetCompose : TFhirElement
    A set of criteria that provide the content logical definition of the value set by including or excluding codes from outside this value set.
  }
  {!.Net HL7Connect.Fhir.ValueSetCompose}
  TFhirValueSetCompose = class (TFhirBackboneElement)
  private
    FimportList : TFhirUriList;
    FincludeList : TFhirValueSetComposeIncludeList;
    FexcludeList : TFhirValueSetComposeIncludeList;
    function GetImportList : TFhirUriList;
    function GetHasImportList : Boolean;
    function GetIncludeList : TFhirValueSetComposeIncludeList;
    function GetHasIncludeList : Boolean;
    function GetExcludeList : TFhirValueSetComposeIncludeList;
    function GetHasExcludeList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetCompose; overload;
    function Clone : TFhirValueSetCompose; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member importList
      Includes the contents of the referenced value set as a part of the contents of this value set. This is an absolute URI that is a reference to ValueSet.uri.
    }
    property importList : TFhirUriList read GetImportList;
    property hasImportList : boolean read GetHasImportList;

    {@member includeList
      Include one or more codes from a code system.
    }
    property includeList : TFhirValueSetComposeIncludeList read GetIncludeList;
    property hasIncludeList : boolean read GetHasIncludeList;

    {@member excludeList
      Exclude one or more codes from the value set.
    }
    property excludeList : TFhirValueSetComposeIncludeList read GetExcludeList;
    property hasExcludeList : boolean read GetHasExcludeList;

  end;


  TFhirValueSetComposeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetComposeList;
    function GetCurrent : TFhirValueSetCompose;
  public
    Constructor Create(list : TFhirValueSetComposeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetCompose read GetCurrent;
  end;


  {@Class TFhirValueSetComposeList
    A list of FhirValueSetCompose
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeList}
  TFhirValueSetComposeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetCompose;
    procedure SetItemN(index : Integer; value : TFhirValueSetCompose);
  public
    {!script hide}
    function Link : TFhirValueSetComposeList; Overload;
    function Clone : TFhirValueSetComposeList; Overload;
    function GetEnumerator : TFhirValueSetComposeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetCompose to the end of the list.
    }
    function Append : TFhirValueSetCompose;

    
    {@member AddItem
      Add an already existing FhirValueSetCompose to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetCompose); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetCompose) : Integer;
    

    {@member Insert
      Insert FhirValueSetCompose before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetCompose;
    

    {@member InsertItem
       Insert an existing FhirValueSetCompose before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetCompose);
    
    {@member Item
       Get the iIndexth FhirValueSetCompose. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetCompose. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetCompose);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetCompose;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetComposes[index : Integer] : TFhirValueSetCompose read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetComposeInclude : TFhirElement
    Include one or more codes from a code system.
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeInclude}
  TFhirValueSetComposeInclude = class (TFhirBackboneElement)
  private
    FSystem : TFhirUri;
    FVersion : TFhirString;
    FconceptList : TFhirValueSetComposeIncludeConceptList;
    FfilterList : TFhirValueSetComposeIncludeFilterList;
    Procedure SetSystem(value : TFhirUri);
    Function GetSystemST : String;
    Procedure SetSystemST(value : String);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    function GetConceptList : TFhirValueSetComposeIncludeConceptList;
    function GetHasConceptList : Boolean;
    function GetFilterList : TFhirValueSetComposeIncludeFilterList;
    function GetHasFilterList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetComposeInclude; overload;
    function Clone : TFhirValueSetComposeInclude; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member system
      Typed access to An absolute URI which is the code system from which the selected codes come from.
    }
    property system : String read GetSystemST write SetSystemST;
    {@member systemElement
      An absolute URI which is the code system from which the selected codes come from.
    }
    property systemElement : TFhirUri read FSystem write SetSystem;

    {@member version
      Typed access to The version of the code system that the codes are selected from.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version of the code system that the codes are selected from.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member conceptList
      Specifies a concept to be included or excluded.
    }
    property conceptList : TFhirValueSetComposeIncludeConceptList read GetConceptList;
    property hasConceptList : boolean read GetHasConceptList;

    {@member filterList
      Select concepts by specify a matching criteria based on the properties (including relationships) defined by the system. If multiple filters are specified, they SHALL all be true.
    }
    property filterList : TFhirValueSetComposeIncludeFilterList read GetFilterList;
    property hasFilterList : boolean read GetHasFilterList;

  end;


  TFhirValueSetComposeIncludeListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetComposeIncludeList;
    function GetCurrent : TFhirValueSetComposeInclude;
  public
    Constructor Create(list : TFhirValueSetComposeIncludeList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetComposeInclude read GetCurrent;
  end;


  {@Class TFhirValueSetComposeIncludeList
    A list of FhirValueSetComposeInclude
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeIncludeList}
  TFhirValueSetComposeIncludeList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetComposeInclude;
    procedure SetItemN(index : Integer; value : TFhirValueSetComposeInclude);
  public
    {!script hide}
    function Link : TFhirValueSetComposeIncludeList; Overload;
    function Clone : TFhirValueSetComposeIncludeList; Overload;
    function GetEnumerator : TFhirValueSetComposeIncludeListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetComposeInclude to the end of the list.
    }
    function Append : TFhirValueSetComposeInclude;

    
    {@member AddItem
      Add an already existing FhirValueSetComposeInclude to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetComposeInclude); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetComposeInclude) : Integer;
    

    {@member Insert
      Insert FhirValueSetComposeInclude before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetComposeInclude;
    

    {@member InsertItem
       Insert an existing FhirValueSetComposeInclude before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetComposeInclude);
    
    {@member Item
       Get the iIndexth FhirValueSetComposeInclude. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetComposeInclude. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetComposeInclude);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetComposeInclude;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetComposeIncludes[index : Integer] : TFhirValueSetComposeInclude read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetComposeIncludeConcept : TFhirElement
    Specifies a concept to be included or excluded.
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeIncludeConcept}
  TFhirValueSetComposeIncludeConcept = class (TFhirBackboneElement)
  private
    FCode : TFhirCode;
    FDisplay : TFhirString;
    FdesignationList : TFhirValueSetCodeSystemConceptDesignationList;
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetDisplay(value : TFhirString);
    Function GetDisplayST : String;
    Procedure SetDisplayST(value : String);
    function GetDesignationList : TFhirValueSetCodeSystemConceptDesignationList;
    function GetHasDesignationList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetComposeIncludeConcept; overload;
    function Clone : TFhirValueSetComposeIncludeConcept; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member code
      Typed access to Specifies a code for the concept to be included or excluded.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      Specifies a code for the concept to be included or excluded.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member display
      Typed access to The text to display to the user for this concept in the context of this valueset. If no display is provided, then applications using the value set use the display specified for the code by the system.
    }
    property display : String read GetDisplayST write SetDisplayST;
    {@member displayElement
      The text to display to the user for this concept in the context of this valueset. If no display is provided, then applications using the value set use the display specified for the code by the system.
    }
    property displayElement : TFhirString read FDisplay write SetDisplay;

    {@member designationList
      Additional representations for this concept when used in this value set - other languages, aliases, specialized purposes, used for particular purposes, etc.
    }
    property designationList : TFhirValueSetCodeSystemConceptDesignationList read GetDesignationList;
    property hasDesignationList : boolean read GetHasDesignationList;

  end;


  TFhirValueSetComposeIncludeConceptListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetComposeIncludeConceptList;
    function GetCurrent : TFhirValueSetComposeIncludeConcept;
  public
    Constructor Create(list : TFhirValueSetComposeIncludeConceptList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetComposeIncludeConcept read GetCurrent;
  end;


  {@Class TFhirValueSetComposeIncludeConceptList
    A list of FhirValueSetComposeIncludeConcept
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeIncludeConceptList}
  TFhirValueSetComposeIncludeConceptList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetComposeIncludeConcept;
    procedure SetItemN(index : Integer; value : TFhirValueSetComposeIncludeConcept);
  public
    {!script hide}
    function Link : TFhirValueSetComposeIncludeConceptList; Overload;
    function Clone : TFhirValueSetComposeIncludeConceptList; Overload;
    function GetEnumerator : TFhirValueSetComposeIncludeConceptListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetComposeIncludeConcept to the end of the list.
    }
    function Append : TFhirValueSetComposeIncludeConcept;

    
    {@member AddItem
      Add an already existing FhirValueSetComposeIncludeConcept to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetComposeIncludeConcept); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetComposeIncludeConcept) : Integer;
    

    {@member Insert
      Insert FhirValueSetComposeIncludeConcept before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetComposeIncludeConcept;
    

    {@member InsertItem
       Insert an existing FhirValueSetComposeIncludeConcept before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetComposeIncludeConcept);
    
    {@member Item
       Get the iIndexth FhirValueSetComposeIncludeConcept. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetComposeIncludeConcept. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetComposeIncludeConcept);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetComposeIncludeConcept;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetComposeIncludeConcepts[index : Integer] : TFhirValueSetComposeIncludeConcept read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetComposeIncludeFilter : TFhirElement
    Select concepts by specify a matching criteria based on the properties (including relationships) defined by the system. If multiple filters are specified, they SHALL all be true.
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeIncludeFilter}
  TFhirValueSetComposeIncludeFilter = class (TFhirBackboneElement)
  private
    FProperty_ : TFhirCode;
    FOp : TFhirEnum;
    FValue : TFhirCode;
    Procedure SetProperty_(value : TFhirCode);
    Function GetProperty_ST : String;
    Procedure SetProperty_ST(value : String);
    Procedure SetOp(value : TFhirEnum);
    Function GetOpST : TFhirFilterOperator;
    Procedure SetOpST(value : TFhirFilterOperator);
    Procedure SetValue(value : TFhirCode);
    Function GetValueST : String;
    Procedure SetValueST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetComposeIncludeFilter; overload;
    function Clone : TFhirValueSetComposeIncludeFilter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member property_
      Typed access to A code that identifies a property defined in the code system.
    }
    property property_ : String read GetProperty_ST write SetProperty_ST;
    {@member property_Element
      A code that identifies a property defined in the code system.
    }
    property property_Element : TFhirCode read FProperty_ write SetProperty_;

    {@member op
      The kind of operation to perform as a part of the filter criteria.
    }
    property op : TFhirFilterOperator read GetOpST write SetOpST;
    property opElement : TFhirEnum read FOp write SetOp;

    {@member value
      Typed access to The match value may be either a code defined by the system, or a string value, which is a regex match on the literal string of the property value.
    }
    property value : String read GetValueST write SetValueST;
    {@member valueElement
      The match value may be either a code defined by the system, or a string value, which is a regex match on the literal string of the property value.
    }
    property valueElement : TFhirCode read FValue write SetValue;

  end;


  TFhirValueSetComposeIncludeFilterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetComposeIncludeFilterList;
    function GetCurrent : TFhirValueSetComposeIncludeFilter;
  public
    Constructor Create(list : TFhirValueSetComposeIncludeFilterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetComposeIncludeFilter read GetCurrent;
  end;


  {@Class TFhirValueSetComposeIncludeFilterList
    A list of FhirValueSetComposeIncludeFilter
  }
  {!.Net HL7Connect.Fhir.ValueSetComposeIncludeFilterList}
  TFhirValueSetComposeIncludeFilterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetComposeIncludeFilter;
    procedure SetItemN(index : Integer; value : TFhirValueSetComposeIncludeFilter);
  public
    {!script hide}
    function Link : TFhirValueSetComposeIncludeFilterList; Overload;
    function Clone : TFhirValueSetComposeIncludeFilterList; Overload;
    function GetEnumerator : TFhirValueSetComposeIncludeFilterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetComposeIncludeFilter to the end of the list.
    }
    function Append : TFhirValueSetComposeIncludeFilter;

    
    {@member AddItem
      Add an already existing FhirValueSetComposeIncludeFilter to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetComposeIncludeFilter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetComposeIncludeFilter) : Integer;
    

    {@member Insert
      Insert FhirValueSetComposeIncludeFilter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetComposeIncludeFilter;
    

    {@member InsertItem
       Insert an existing FhirValueSetComposeIncludeFilter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetComposeIncludeFilter);
    
    {@member Item
       Get the iIndexth FhirValueSetComposeIncludeFilter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetComposeIncludeFilter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetComposeIncludeFilter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetComposeIncludeFilter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetComposeIncludeFilters[index : Integer] : TFhirValueSetComposeIncludeFilter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetExpansion : TFhirElement
    A value set can also be "expanded", where the value set is turned into a simple collection of enumerated codes. This element holds the expansion, if it has been performed.
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansion}
  TFhirValueSetExpansion = class (TFhirBackboneElement)
  private
    FIdentifier : TFhirUri;
    FTimestamp : TFhirDateTime;
    FTotal : TFhirInteger;
    FOffset : TFhirInteger;
    FparameterList : TFhirValueSetExpansionParameterList;
    FcontainsList : TFhirValueSetExpansionContainsList;
    Procedure SetIdentifier(value : TFhirUri);
    Function GetIdentifierST : String;
    Procedure SetIdentifierST(value : String);
    Procedure SetTimestamp(value : TFhirDateTime);
    Function GetTimestampST : TDateAndTime;
    Procedure SetTimestampST(value : TDateAndTime);
    Procedure SetTotal(value : TFhirInteger);
    Function GetTotalST : String;
    Procedure SetTotalST(value : String);
    Procedure SetOffset(value : TFhirInteger);
    Function GetOffsetST : String;
    Procedure SetOffsetST(value : String);
    function GetParameterList : TFhirValueSetExpansionParameterList;
    function GetHasParameterList : Boolean;
    function GetContainsList : TFhirValueSetExpansionContainsList;
    function GetHasContainsList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetExpansion; overload;
    function Clone : TFhirValueSetExpansion; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member identifier
      Typed access to An identifier that uniquely identifies this expansion of the valueset. Systems may re-use the same identifier as long as the expansion and the definition remain the same, but are not required to do so.
    }
    property identifier : String read GetIdentifierST write SetIdentifierST;
    {@member identifierElement
      An identifier that uniquely identifies this expansion of the valueset. Systems may re-use the same identifier as long as the expansion and the definition remain the same, but are not required to do so.
    }
    property identifierElement : TFhirUri read FIdentifier write SetIdentifier;

    {@member timestamp
      Typed access to The time at which the expansion was produced by the expanding system.
    }
    property timestamp : TDateAndTime read GetTimestampST write SetTimestampST;
    {@member timestampElement
      The time at which the expansion was produced by the expanding system.
    }
    property timestampElement : TFhirDateTime read FTimestamp write SetTimestamp;

    {@member total
      Typed access to The total number of concepts in the expansion. If the number of concept nodes in this resource is less than the stated number, then the server can return more using the offset parameter.
    }
    property total : String read GetTotalST write SetTotalST;
    {@member totalElement
      The total number of concepts in the expansion. If the number of concept nodes in this resource is less than the stated number, then the server can return more using the offset parameter.
    }
    property totalElement : TFhirInteger read FTotal write SetTotal;

    {@member offset
      Typed access to If paging is being used, the offset at which this resource starts.  I.e. this resource is a partial view into the expansion. If paging is not being used, this element SHALL not be present.
    }
    property offset : String read GetOffsetST write SetOffsetST;
    {@member offsetElement
      If paging is being used, the offset at which this resource starts.  I.e. this resource is a partial view into the expansion. If paging is not being used, this element SHALL not be present.
    }
    property offsetElement : TFhirInteger read FOffset write SetOffset;

    {@member parameterList
      A parameter that controlled the expansion process. These parameters may be used by users of expanded value sets to check whether the expansion is suitable for a particular purpose, or to pick the correct expansion.
    }
    property parameterList : TFhirValueSetExpansionParameterList read GetParameterList;
    property hasParameterList : boolean read GetHasParameterList;

    {@member containsList
      The codes that are contained in the value set expansion.
    }
    property containsList : TFhirValueSetExpansionContainsList read GetContainsList;
    property hasContainsList : boolean read GetHasContainsList;

  end;


  TFhirValueSetExpansionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetExpansionList;
    function GetCurrent : TFhirValueSetExpansion;
  public
    Constructor Create(list : TFhirValueSetExpansionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetExpansion read GetCurrent;
  end;


  {@Class TFhirValueSetExpansionList
    A list of FhirValueSetExpansion
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansionList}
  TFhirValueSetExpansionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetExpansion;
    procedure SetItemN(index : Integer; value : TFhirValueSetExpansion);
  public
    {!script hide}
    function Link : TFhirValueSetExpansionList; Overload;
    function Clone : TFhirValueSetExpansionList; Overload;
    function GetEnumerator : TFhirValueSetExpansionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetExpansion to the end of the list.
    }
    function Append : TFhirValueSetExpansion;

    
    {@member AddItem
      Add an already existing FhirValueSetExpansion to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetExpansion); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetExpansion) : Integer;
    

    {@member Insert
      Insert FhirValueSetExpansion before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetExpansion;
    

    {@member InsertItem
       Insert an existing FhirValueSetExpansion before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetExpansion);
    
    {@member Item
       Get the iIndexth FhirValueSetExpansion. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetExpansion. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetExpansion);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetExpansion;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetExpansions[index : Integer] : TFhirValueSetExpansion read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetExpansionParameter : TFhirElement
    A parameter that controlled the expansion process. These parameters may be used by users of expanded value sets to check whether the expansion is suitable for a particular purpose, or to pick the correct expansion.
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansionParameter}
  TFhirValueSetExpansionParameter = class (TFhirBackboneElement)
  private
    FName : TFhirString;
    FValue : TFhirType;
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetValue(value : TFhirType);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetExpansionParameter; overload;
    function Clone : TFhirValueSetExpansionParameter; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member name
      Typed access to The name of the parameter.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      The name of the parameter.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member value
      Typed access to The value of the parameter. (defined for API consistency)
    }
    property value : TFhirType read FValue write SetValue;
    {@member valueElement
      The value of the parameter.
    }
    property valueElement : TFhirType read FValue write SetValue;

  end;


  TFhirValueSetExpansionParameterListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetExpansionParameterList;
    function GetCurrent : TFhirValueSetExpansionParameter;
  public
    Constructor Create(list : TFhirValueSetExpansionParameterList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetExpansionParameter read GetCurrent;
  end;


  {@Class TFhirValueSetExpansionParameterList
    A list of FhirValueSetExpansionParameter
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansionParameterList}
  TFhirValueSetExpansionParameterList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetExpansionParameter;
    procedure SetItemN(index : Integer; value : TFhirValueSetExpansionParameter);
  public
    {!script hide}
    function Link : TFhirValueSetExpansionParameterList; Overload;
    function Clone : TFhirValueSetExpansionParameterList; Overload;
    function GetEnumerator : TFhirValueSetExpansionParameterListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetExpansionParameter to the end of the list.
    }
    function Append : TFhirValueSetExpansionParameter;

    
    {@member AddItem
      Add an already existing FhirValueSetExpansionParameter to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetExpansionParameter); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetExpansionParameter) : Integer;
    

    {@member Insert
      Insert FhirValueSetExpansionParameter before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetExpansionParameter;
    

    {@member InsertItem
       Insert an existing FhirValueSetExpansionParameter before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetExpansionParameter);
    
    {@member Item
       Get the iIndexth FhirValueSetExpansionParameter. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetExpansionParameter. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetExpansionParameter);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetExpansionParameter;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetExpansionParameters[index : Integer] : TFhirValueSetExpansionParameter read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSetExpansionContains : TFhirElement
    The codes that are contained in the value set expansion.
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansionContains}
  TFhirValueSetExpansionContains = class (TFhirBackboneElement)
  private
    FSystem : TFhirUri;
    FAbstract : TFhirBoolean;
    FVersion : TFhirString;
    FCode : TFhirCode;
    FDisplay : TFhirString;
    FcontainsList : TFhirValueSetExpansionContainsList;
    Procedure SetSystem(value : TFhirUri);
    Function GetSystemST : String;
    Procedure SetSystemST(value : String);
    Procedure SetAbstract(value : TFhirBoolean);
    Function GetAbstractST : Boolean;
    Procedure SetAbstractST(value : Boolean);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetCode(value : TFhirCode);
    Function GetCodeST : String;
    Procedure SetCodeST(value : String);
    Procedure SetDisplay(value : TFhirString);
    Function GetDisplayST : String;
    Procedure SetDisplayST(value : String);
    function GetContainsList : TFhirValueSetExpansionContainsList;
    function GetHasContainsList : Boolean;
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSetExpansionContains; overload;
    function Clone : TFhirValueSetExpansionContains; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member system
      Typed access to An absolute URI which is the code system in which the code for this item in the expansion is defined.
    }
    property system : String read GetSystemST write SetSystemST;
    {@member systemElement
      An absolute URI which is the code system in which the code for this item in the expansion is defined.
    }
    property systemElement : TFhirUri read FSystem write SetSystem;

    {@member abstract
      Typed access to If true, this entry is included in the expansion for navigational purposes, and the user cannot select the code directly as a proper value.
    }
    property abstract : Boolean read GetAbstractST write SetAbstractST;
    {@member abstractElement
      If true, this entry is included in the expansion for navigational purposes, and the user cannot select the code directly as a proper value.
    }
    property abstractElement : TFhirBoolean read FAbstract write SetAbstract;

    {@member version
      Typed access to The version of this code system that defined this code and/or display. This should only be used with code systems that do not enforce concept permanence.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      The version of this code system that defined this code and/or display. This should only be used with code systems that do not enforce concept permanence.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member code
      Typed access to The code for this item in the expansion hierarchy. If this code is missing the entry in the hierarchy is a place holder (abstract) and does not represent a valid code in the value set.
    }
    property code : String read GetCodeST write SetCodeST;
    {@member codeElement
      The code for this item in the expansion hierarchy. If this code is missing the entry in the hierarchy is a place holder (abstract) and does not represent a valid code in the value set.
    }
    property codeElement : TFhirCode read FCode write SetCode;

    {@member display
      Typed access to The recommended display for this item in the expansion.
    }
    property display : String read GetDisplayST write SetDisplayST;
    {@member displayElement
      The recommended display for this item in the expansion.
    }
    property displayElement : TFhirString read FDisplay write SetDisplay;

    {@member containsList
      Other codes and entries contained under this entry in the hierarchy.
    }
    property containsList : TFhirValueSetExpansionContainsList read GetContainsList;
    property hasContainsList : boolean read GetHasContainsList;

  end;


  TFhirValueSetExpansionContainsListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetExpansionContainsList;
    function GetCurrent : TFhirValueSetExpansionContains;
  public
    Constructor Create(list : TFhirValueSetExpansionContainsList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSetExpansionContains read GetCurrent;
  end;


  {@Class TFhirValueSetExpansionContainsList
    A list of FhirValueSetExpansionContains
  }
  {!.Net HL7Connect.Fhir.ValueSetExpansionContainsList}
  TFhirValueSetExpansionContainsList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSetExpansionContains;
    procedure SetItemN(index : Integer; value : TFhirValueSetExpansionContains);
  public
    {!script hide}
    function Link : TFhirValueSetExpansionContainsList; Overload;
    function Clone : TFhirValueSetExpansionContainsList; Overload;
    function GetEnumerator : TFhirValueSetExpansionContainsListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSetExpansionContains to the end of the list.
    }
    function Append : TFhirValueSetExpansionContains;

    
    {@member AddItem
      Add an already existing FhirValueSetExpansionContains to the end of the list.
    }
    procedure AddItem(value : TFhirValueSetExpansionContains); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSetExpansionContains) : Integer;
    

    {@member Insert
      Insert FhirValueSetExpansionContains before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSetExpansionContains;
    

    {@member InsertItem
       Insert an existing FhirValueSetExpansionContains before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSetExpansionContains);
    
    {@member Item
       Get the iIndexth FhirValueSetExpansionContains. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSetExpansionContains. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSetExpansionContains);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSetExpansionContains;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSetExpansionContains[index : Integer] : TFhirValueSetExpansionContains read GetItemN write SetItemN; default;
  End;


  {@Class TFhirValueSet : TFhirDomainResource
    A value set specifies a set of codes drawn from one or more code systems.
  }
  {!.Net HL7Connect.Fhir.ValueSet}
  TFhirValueSet = class (TFhirDomainResource)
  private
    FUrl : TFhirUri;
    FIdentifier : TFhirIdentifier;
    FVersion : TFhirString;
    FName : TFhirString;
    FStatus : TFhirEnum;
    FExperimental : TFhirBoolean;
    FPublisher : TFhirString;
    FcontactList : TFhirValueSetContactList;
    FDate : TFhirDateTime;
    FLockedDate : TFhirDate;
    FDescription : TFhirString;
    FuseContextList : TFhirCodeableConceptList;
    FImmutable : TFhirBoolean;
    FRequirements : TFhirString;
    FCopyright : TFhirString;
    FExtensible : TFhirBoolean;
    FCodeSystem : TFhirValueSetCodeSystem;
    FCompose : TFhirValueSetCompose;
    FExpansion : TFhirValueSetExpansion;
    Procedure SetUrl(value : TFhirUri);
    Function GetUrlST : String;
    Procedure SetUrlST(value : String);
    Procedure SetIdentifier(value : TFhirIdentifier);
    Procedure SetVersion(value : TFhirString);
    Function GetVersionST : String;
    Procedure SetVersionST(value : String);
    Procedure SetName(value : TFhirString);
    Function GetNameST : String;
    Procedure SetNameST(value : String);
    Procedure SetStatus(value : TFhirEnum);
    Function GetStatusST : TFhirConformanceResourceStatus;
    Procedure SetStatusST(value : TFhirConformanceResourceStatus);
    Procedure SetExperimental(value : TFhirBoolean);
    Function GetExperimentalST : Boolean;
    Procedure SetExperimentalST(value : Boolean);
    Procedure SetPublisher(value : TFhirString);
    Function GetPublisherST : String;
    Procedure SetPublisherST(value : String);
    function GetContactList : TFhirValueSetContactList;
    function GetHasContactList : Boolean;
    Procedure SetDate(value : TFhirDateTime);
    Function GetDateST : TDateAndTime;
    Procedure SetDateST(value : TDateAndTime);
    Procedure SetLockedDate(value : TFhirDate);
    Function GetLockedDateST : TDateAndTime;
    Procedure SetLockedDateST(value : TDateAndTime);
    Procedure SetDescription(value : TFhirString);
    Function GetDescriptionST : String;
    Procedure SetDescriptionST(value : String);
    function GetUseContextList : TFhirCodeableConceptList;
    function GetHasUseContextList : Boolean;
    Procedure SetImmutable(value : TFhirBoolean);
    Function GetImmutableST : Boolean;
    Procedure SetImmutableST(value : Boolean);
    Procedure SetRequirements(value : TFhirString);
    Function GetRequirementsST : String;
    Procedure SetRequirementsST(value : String);
    Procedure SetCopyright(value : TFhirString);
    Function GetCopyrightST : String;
    Procedure SetCopyrightST(value : String);
    Procedure SetExtensible(value : TFhirBoolean);
    Function GetExtensibleST : Boolean;
    Procedure SetExtensibleST(value : Boolean);
    Procedure SetCodeSystem(value : TFhirValueSetCodeSystem);
    Procedure SetCompose(value : TFhirValueSetCompose);
    Procedure SetExpansion(value : TFhirValueSetExpansion);
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirValueSet; overload;
    function Clone : TFhirValueSet; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member url
      Typed access to An absolute URL that is used to identify this value set when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this value set is (or will be) published.
    }
    property url : String read GetUrlST write SetUrlST;
    {@member urlElement
      An absolute URL that is used to identify this value set when it is referenced in a specification, model, design or an instance. This SHALL be a URL, SHOULD be globally unique, and SHOULD be an address at which this value set is (or will be) published.
    }
    property urlElement : TFhirUri read FUrl write SetUrl;

    {@member identifier
      Typed access to Formal identifier that is used to identify this value set when it is represented in other formats, or referenced in a specification, model, design or an instance. (defined for API consistency)
    }
    property identifier : TFhirIdentifier read FIdentifier write SetIdentifier;
    {@member identifierElement
      Formal identifier that is used to identify this value set when it is represented in other formats, or referenced in a specification, model, design or an instance.
    }
    property identifierElement : TFhirIdentifier read FIdentifier write SetIdentifier;

    {@member version
      Typed access to Used to identify this version of the value set when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property version : String read GetVersionST write SetVersionST;
    {@member versionElement
      Used to identify this version of the value set when it is referenced in a specification, model, design or instance. This is an arbitrary value managed by the profile author manually and the value should be a timestamp.
    }
    property versionElement : TFhirString read FVersion write SetVersion;

    {@member name
      Typed access to A free text natural language name describing the value set.
    }
    property name : String read GetNameST write SetNameST;
    {@member nameElement
      A free text natural language name describing the value set.
    }
    property nameElement : TFhirString read FName write SetName;

    {@member status
      The status of the value set.
    }
    property status : TFhirConformanceResourceStatus read GetStatusST write SetStatusST;
    property statusElement : TFhirEnum read FStatus write SetStatus;

    {@member experimental
      Typed access to This valueset was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimental : Boolean read GetExperimentalST write SetExperimentalST;
    {@member experimentalElement
      This valueset was authored for testing purposes (or education/evaluation/marketing), and is not intended to be used for genuine usage.
    }
    property experimentalElement : TFhirBoolean read FExperimental write SetExperimental;

    {@member publisher
      Typed access to The name of the individual or organization that published the value set.
    }
    property publisher : String read GetPublisherST write SetPublisherST;
    {@member publisherElement
      The name of the individual or organization that published the value set.
    }
    property publisherElement : TFhirString read FPublisher write SetPublisher;

    {@member contactList
      Contacts to assist a user in finding and communicating with the publisher.
    }
    property contactList : TFhirValueSetContactList read GetContactList;
    property hasContactList : boolean read GetHasContactList;

    {@member date
      Typed access to The date that the value set status was last changed. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the implementation guide changes (e.g. the 'content logical definition').
    }
    property date : TDateAndTime read GetDateST write SetDateST;
    {@member dateElement
      The date that the value set status was last changed. The date must change when the business version changes, if it does, and it must change if the status code changes. In addition, it should change when the substantive content of the implementation guide changes (e.g. the 'content logical definition').
    }
    property dateElement : TFhirDateTime read FDate write SetDate;

    {@member lockedDate
      Typed access to If a locked date is defined, then the Content Logical Definition must be evaluated using the current version of all referenced code system(s) and value set instances as of the locked date.
    }
    property lockedDate : TDateAndTime read GetLockedDateST write SetLockedDateST;
    {@member lockedDateElement
      If a locked date is defined, then the Content Logical Definition must be evaluated using the current version of all referenced code system(s) and value set instances as of the locked date.
    }
    property lockedDateElement : TFhirDate read FLockedDate write SetLockedDate;

    {@member description
      Typed access to A free text natural language description of the use of the value set - reason for definition, "the semantic space" to be included in the value set, conditions of use, etc. The description may include a list of expected usages for the value set and can also describe the approach taken to build the value set.
    }
    property description : String read GetDescriptionST write SetDescriptionST;
    {@member descriptionElement
      A free text natural language description of the use of the value set - reason for definition, "the semantic space" to be included in the value set, conditions of use, etc. The description may include a list of expected usages for the value set and can also describe the approach taken to build the value set.
    }
    property descriptionElement : TFhirString read FDescription write SetDescription;

    {@member useContextList
      The content was developed with a focus and intent of supporting the contexts that are listed. These terms may be used to assist with indexing and searching of value set definitions.
    }
    property useContextList : TFhirCodeableConceptList read GetUseContextList;
    property hasUseContextList : boolean read GetHasUseContextList;

    {@member immutable
      Typed access to If this is set to 'true', then no new versions of the content logical definition can be created.  Note: Other metadata might still change.
    }
    property immutable : Boolean read GetImmutableST write SetImmutableST;
    {@member immutableElement
      If this is set to 'true', then no new versions of the content logical definition can be created.  Note: Other metadata might still change.
    }
    property immutableElement : TFhirBoolean read FImmutable write SetImmutable;

    {@member requirements
      Typed access to Explains why this value set is needed and why it has been constrained as it has.
    }
    property requirements : String read GetRequirementsST write SetRequirementsST;
    {@member requirementsElement
      Explains why this value set is needed and why it has been constrained as it has.
    }
    property requirementsElement : TFhirString read FRequirements write SetRequirements;

    {@member copyright
      Typed access to A copyright statement relating to the value set and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the value set.
    }
    property copyright : String read GetCopyrightST write SetCopyrightST;
    {@member copyrightElement
      A copyright statement relating to the value set and/or its contents. Copyright statements are generally legal restrictions on the use and publishing of the value set.
    }
    property copyrightElement : TFhirString read FCopyright write SetCopyright;

    {@member extensible
      Typed access to Whether this is intended to be used with an extensible binding or not.
    }
    property extensible : Boolean read GetExtensibleST write SetExtensibleST;
    {@member extensibleElement
      Whether this is intended to be used with an extensible binding or not.
    }
    property extensibleElement : TFhirBoolean read FExtensible write SetExtensible;

    {@member codeSystem
      Typed access to A definition of a code system, inlined into the value set (as a packaging convenience). Note that the inline code system may be used from other value sets by referring to its (codeSystem.system) directly. (defined for API consistency)
    }
    property codeSystem : TFhirValueSetCodeSystem read FCodeSystem write SetCodeSystem;
    {@member codeSystemElement
      A definition of a code system, inlined into the value set (as a packaging convenience). Note that the inline code system may be used from other value sets by referring to its (codeSystem.system) directly.
    }
    property codeSystemElement : TFhirValueSetCodeSystem read FCodeSystem write SetCodeSystem;

    {@member compose
      Typed access to A set of criteria that provide the content logical definition of the value set by including or excluding codes from outside this value set. (defined for API consistency)
    }
    property compose : TFhirValueSetCompose read FCompose write SetCompose;
    {@member composeElement
      A set of criteria that provide the content logical definition of the value set by including or excluding codes from outside this value set.
    }
    property composeElement : TFhirValueSetCompose read FCompose write SetCompose;

    {@member expansion
      Typed access to A value set can also be "expanded", where the value set is turned into a simple collection of enumerated codes. This element holds the expansion, if it has been performed. (defined for API consistency)
    }
    property expansion : TFhirValueSetExpansion read FExpansion write SetExpansion;
    {@member expansionElement
      A value set can also be "expanded", where the value set is turned into a simple collection of enumerated codes. This element holds the expansion, if it has been performed.
    }
    property expansionElement : TFhirValueSetExpansion read FExpansion write SetExpansion;

  end;


  TFhirValueSetListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirValueSetList;
    function GetCurrent : TFhirValueSet;
  public
    Constructor Create(list : TFhirValueSetList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirValueSet read GetCurrent;
  end;


  {@Class TFhirValueSetList
    A list of FhirValueSet
  }
  {!.Net HL7Connect.Fhir.ValueSetList}
  TFhirValueSetList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirValueSet;
    procedure SetItemN(index : Integer; value : TFhirValueSet);
  public
    {!script hide}
    function Link : TFhirValueSetList; Overload;
    function Clone : TFhirValueSetList; Overload;
    function GetEnumerator : TFhirValueSetListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirValueSet to the end of the list.
    }
    function Append : TFhirValueSet;

    
    {@member AddItem
      Add an already existing FhirValueSet to the end of the list.
    }
    procedure AddItem(value : TFhirValueSet); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirValueSet) : Integer;
    

    {@member Insert
      Insert FhirValueSet before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirValueSet;
    

    {@member InsertItem
       Insert an existing FhirValueSet before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirValueSet);
    
    {@member Item
       Get the iIndexth FhirValueSet. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirValueSet. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirValueSet);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirValueSet;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirValueSets[index : Integer] : TFhirValueSet read GetItemN write SetItemN; default;
  End;


  {@Class TFhirVisionPrescriptionDispense : TFhirElement
    Deals with details of the dispense part of the supply specification.
  }
  {!.Net HL7Connect.Fhir.VisionPrescriptionDispense}
  TFhirVisionPrescriptionDispense = class (TFhirBackboneElement)
  private
    FProduct : TFhirCoding;
    FEye : TFhirEnum;
    FSphere : TFhirDecimal;
    FCylinder : TFhirDecimal;
    FAxis : TFhirInteger;
    FPrism : TFhirDecimal;
    FBase : TFhirEnum;
    FAdd : TFhirDecimal;
    FPower : TFhirDecimal;
    FBackCurve : TFhirDecimal;
    FDiameter : TFhirDecimal;
    FDuration : TFhirQuantity;
    FColor : TFhirString;
    FBrand : TFhirString;
    FNotes : TFhirString;
    Procedure SetProduct(value : TFhirCoding);
    Procedure SetEye(value : TFhirEnum);
    Function GetEyeST : TFhirVisionEyeCodes;
    Procedure SetEyeST(value : TFhirVisionEyeCodes);
    Procedure SetSphere(value : TFhirDecimal);
    Function GetSphereST : String;
    Procedure SetSphereST(value : String);
    Procedure SetCylinder(value : TFhirDecimal);
    Function GetCylinderST : String;
    Procedure SetCylinderST(value : String);
    Procedure SetAxis(value : TFhirInteger);
    Function GetAxisST : String;
    Procedure SetAxisST(value : String);
    Procedure SetPrism(value : TFhirDecimal);
    Function GetPrismST : String;
    Procedure SetPrismST(value : String);
    Procedure SetBase(value : TFhirEnum);
    Function GetBaseST : TFhirVisionBaseCodes;
    Procedure SetBaseST(value : TFhirVisionBaseCodes);
    Procedure SetAdd(value : TFhirDecimal);
    Function GetAddST : String;
    Procedure SetAddST(value : String);
    Procedure SetPower(value : TFhirDecimal);
    Function GetPowerST : String;
    Procedure SetPowerST(value : String);
    Procedure SetBackCurve(value : TFhirDecimal);
    Function GetBackCurveST : String;
    Procedure SetBackCurveST(value : String);
    Procedure SetDiameter(value : TFhirDecimal);
    Function GetDiameterST : String;
    Procedure SetDiameterST(value : String);
    Procedure SetDuration(value : TFhirQuantity);
    Procedure SetColor(value : TFhirString);
    Function GetColorST : String;
    Procedure SetColorST(value : String);
    Procedure SetBrand(value : TFhirString);
    Function GetBrandST : String;
    Procedure SetBrandST(value : String);
    Procedure SetNotes(value : TFhirString);
    Function GetNotesST : String;
    Procedure SetNotesST(value : String);
  protected
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirVisionPrescriptionDispense; overload;
    function Clone : TFhirVisionPrescriptionDispense; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    function equalsDeep(other : TFHIRObject) : boolean; override;
    function equalsShallow(other : TFHIRObject) : boolean; override;
    {!script show}
  published
    {@member product
      Typed access to Identifies the type of vision correction product which is required for the patient. (defined for API consistency)
    }
    property product : TFhirCoding read FProduct write SetProduct;
    {@member productElement
      Identifies the type of vision correction product which is required for the patient.
    }
    property productElement : TFhirCoding read FProduct write SetProduct;

    {@member eye
      The eye for which the lens applies.
    }
    property eye : TFhirVisionEyeCodes read GetEyeST write SetEyeST;
    property eyeElement : TFhirEnum read FEye write SetEye;

    {@member sphere
      Typed access to Lens power measured in diopters (0.25 units).
    }
    property sphere : String read GetSphereST write SetSphereST;
    {@member sphereElement
      Lens power measured in diopters (0.25 units).
    }
    property sphereElement : TFhirDecimal read FSphere write SetSphere;

    {@member cylinder
      Typed access to Power adjustment for astigmatism measured in diopters (0.25 units).
    }
    property cylinder : String read GetCylinderST write SetCylinderST;
    {@member cylinderElement
      Power adjustment for astigmatism measured in diopters (0.25 units).
    }
    property cylinderElement : TFhirDecimal read FCylinder write SetCylinder;

    {@member axis
      Typed access to Adjustment for astigmatism measured in integer degrees.
    }
    property axis : String read GetAxisST write SetAxisST;
    {@member axisElement
      Adjustment for astigmatism measured in integer degrees.
    }
    property axisElement : TFhirInteger read FAxis write SetAxis;

    {@member prism
      Typed access to Amount of prism to compensate for eye alignment in fractional units.
    }
    property prism : String read GetPrismST write SetPrismST;
    {@member prismElement
      Amount of prism to compensate for eye alignment in fractional units.
    }
    property prismElement : TFhirDecimal read FPrism write SetPrism;

    {@member base
      The relative base, or reference lens edge, for the prism.
    }
    property base : TFhirVisionBaseCodes read GetBaseST write SetBaseST;
    property baseElement : TFhirEnum read FBase write SetBase;

    {@member add
      Typed access to Power adjustment for multifocal lenses measured in diopters (0.25 units).
    }
    property add : String read GetAddST write SetAddST;
    {@member addElement
      Power adjustment for multifocal lenses measured in diopters (0.25 units).
    }
    property addElement : TFhirDecimal read FAdd write SetAdd;

    {@member power
      Typed access to Contact lens power measured in diopters (0.25 units).
    }
    property power : String read GetPowerST write SetPowerST;
    {@member powerElement
      Contact lens power measured in diopters (0.25 units).
    }
    property powerElement : TFhirDecimal read FPower write SetPower;

    {@member backCurve
      Typed access to Back curvature measured in millimeters.
    }
    property backCurve : String read GetBackCurveST write SetBackCurveST;
    {@member backCurveElement
      Back curvature measured in millimeters.
    }
    property backCurveElement : TFhirDecimal read FBackCurve write SetBackCurve;

    {@member diameter
      Typed access to Contact lens diameter measured in millimeters.
    }
    property diameter : String read GetDiameterST write SetDiameterST;
    {@member diameterElement
      Contact lens diameter measured in millimeters.
    }
    property diameterElement : TFhirDecimal read FDiameter write SetDiameter;

    {@member duration
      Typed access to The recommended maximum wear period for the lens. (defined for API consistency)
    }
    property duration : TFhirQuantity read FDuration write SetDuration;
    {@member durationElement
      The recommended maximum wear period for the lens.
    }
    property durationElement : TFhirQuantity read FDuration write SetDuration;

    {@member color
      Typed access to Special color or pattern.
    }
    property color : String read GetColorST write SetColorST;
    {@member colorElement
      Special color or pattern.
    }
    property colorElement : TFhirString read FColor write SetColor;

    {@member brand
      Typed access to Brand recommendations or restrictions.
    }
    property brand : String read GetBrandST write SetBrandST;
    {@member brandElement
      Brand recommendations or restrictions.
    }
    property brandElement : TFhirString read FBrand write SetBrand;

    {@member notes
      Typed access to Notes for special requirements such as coatings and lens materials.
    }
    property notes : String read GetNotesST write SetNotesST;
    {@member notesElement
      Notes for special requirements such as coatings and lens materials.
    }
    property notesElement : TFhirString read FNotes write SetNotes;

  end;


  TFhirVisionPrescriptionDispenseListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirVisionPrescriptionDispenseList;
    function GetCurrent : TFhirVisionPrescriptionDispense;
  public
    Constructor Create(list : TFhirVisionPrescriptionDispenseList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirVisionPrescriptionDispense read GetCurrent;
  end;


  {@Class TFhirVisionPrescriptionDispenseList
    A list of FhirVisionPrescriptionDispense
  }
  {!.Net HL7Connect.Fhir.VisionPrescriptionDispenseList}
  TFhirVisionPrescriptionDispenseList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirVisionPrescriptionDispense;
    procedure SetItemN(index : Integer; value : TFhirVisionPrescriptionDispense);
  public
    {!script hide}
    function Link : TFhirVisionPrescriptionDispenseList; Overload;
    function Clone : TFhirVisionPrescriptionDispenseList; Overload;
    function GetEnumerator : TFhirVisionPrescriptionDispenseListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirVisionPrescriptionDispense to the end of the list.
    }
    function Append : TFhirVisionPrescriptionDispense;

    
    {@member AddItem
      Add an already existing FhirVisionPrescriptionDispense to the end of the list.
    }
    procedure AddItem(value : TFhirVisionPrescriptionDispense); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirVisionPrescriptionDispense) : Integer;
    

    {@member Insert
      Insert FhirVisionPrescriptionDispense before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirVisionPrescriptionDispense;
    

    {@member InsertItem
       Insert an existing FhirVisionPrescriptionDispense before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirVisionPrescriptionDispense);
    
    {@member Item
       Get the iIndexth FhirVisionPrescriptionDispense. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirVisionPrescriptionDispense. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirVisionPrescriptionDispense);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirVisionPrescriptionDispense;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirVisionPrescriptionDispenses[index : Integer] : TFhirVisionPrescriptionDispense read GetItemN write SetItemN; default;
  End;


  {@Class TFhirVisionPrescription : TFhirDomainResource
    An authorization for the supply of glasses and/or contact lenses to a patient.
  }
  {!.Net HL7Connect.Fhir.VisionPrescription}
  TFhirVisionPrescription = class (TFhirDomainResource)
  private
    FidentifierList : TFhirIdentifierList;
    FDateWritten : TFhirDateTime;
    FPatient : TFhirReference{TFhirPatient};
    FPrescriber : TFhirReference{TFhirPractitioner};
    FEncounter : TFhirReference{TFhirEncounter};
    FReason : TFhirType;
    FdispenseList : TFhirVisionPrescriptionDispenseList;
    function GetIdentifierList : TFhirIdentifierList;
    function GetHasIdentifierList : Boolean;
    Procedure SetDateWritten(value : TFhirDateTime);
    Function GetDateWrittenST : TDateAndTime;
    Procedure SetDateWrittenST(value : TDateAndTime);
    Procedure SetPatient(value : TFhirReference{TFhirPatient});
    Procedure SetPrescriber(value : TFhirReference{TFhirPractitioner});
    Procedure SetEncounter(value : TFhirReference{TFhirEncounter});
    Procedure SetReason(value : TFhirType);
    function GetDispenseList : TFhirVisionPrescriptionDispenseList;
    function GetHasDispenseList : Boolean;
  protected
    function GetResourceType : TFhirResourceType; override;
    Procedure GetChildrenByName(child_name : string; list : TFHIRObjectList); override;
    Procedure ListProperties(oList : TFHIRPropertyList; bInheritedProperties : Boolean); Override;
  public
    constructor Create; Override;
    destructor Destroy; override;
    {!script hide}
    procedure Assign(oSource : TAdvObject); override;
    function Link : TFhirVisionPrescription; overload;
    function Clone : TFhirVisionPrescription; overload;
    procedure setProperty(propName : string; propValue : TFHIRObject); override;
    function FhirType : string; override;
    {!script show}
  published
    {@member identifierList
      Business identifier which may be used by other parties to reference or identify the prescription.
    }
    property identifierList : TFhirIdentifierList read GetIdentifierList;
    property hasIdentifierList : boolean read GetHasIdentifierList;

    {@member dateWritten
      Typed access to The date (and perhaps time) when the prescription was written.
    }
    property dateWritten : TDateAndTime read GetDateWrittenST write SetDateWrittenST;
    {@member dateWrittenElement
      The date (and perhaps time) when the prescription was written.
    }
    property dateWrittenElement : TFhirDateTime read FDateWritten write SetDateWritten;

    {@member patient
      Typed access to A link to a resource representing the person to whom the vision products will be supplied. (defined for API consistency)
    }
    property patient : TFhirReference{TFhirPatient} read FPatient write SetPatient;
    {@member patientElement
      A link to a resource representing the person to whom the vision products will be supplied.
    }
    property patientElement : TFhirReference{TFhirPatient} read FPatient write SetPatient;

    {@member prescriber
      Typed access to The healthcare professional responsible for authorizing the prescription. (defined for API consistency)
    }
    property prescriber : TFhirReference{TFhirPractitioner} read FPrescriber write SetPrescriber;
    {@member prescriberElement
      The healthcare professional responsible for authorizing the prescription.
    }
    property prescriberElement : TFhirReference{TFhirPractitioner} read FPrescriber write SetPrescriber;

    {@member encounter
      Typed access to A link to a resource that identifies the particular occurrence of contact between patient and health care provider. (defined for API consistency)
    }
    property encounter : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;
    {@member encounterElement
      A link to a resource that identifies the particular occurrence of contact between patient and health care provider.
    }
    property encounterElement : TFhirReference{TFhirEncounter} read FEncounter write SetEncounter;

    {@member reason
      Typed access to Can be the reason or the indication for writing the prescription. (defined for API consistency)
    }
    property reason : TFhirType read FReason write SetReason;
    {@member reasonElement
      Can be the reason or the indication for writing the prescription.
    }
    property reasonElement : TFhirType read FReason write SetReason;

    {@member dispenseList
      Deals with details of the dispense part of the supply specification.
    }
    property dispenseList : TFhirVisionPrescriptionDispenseList read GetDispenseList;
    property hasDispenseList : boolean read GetHasDispenseList;

  end;


  TFhirVisionPrescriptionListEnumerator = class (TAdvObject)
  private
    FIndex : integer;
    FList : TFhirVisionPrescriptionList;
    function GetCurrent : TFhirVisionPrescription;
  public
    Constructor Create(list : TFhirVisionPrescriptionList);
    Destructor Destroy; override;
    function MoveNext : boolean;
    property Current : TFhirVisionPrescription read GetCurrent;
  end;


  {@Class TFhirVisionPrescriptionList
    A list of FhirVisionPrescription
  }
  {!.Net HL7Connect.Fhir.VisionPrescriptionList}
  TFhirVisionPrescriptionList = class (TFHIRObjectList)
  private
    function GetItemN(index : Integer) : TFhirVisionPrescription;
    procedure SetItemN(index : Integer; value : TFhirVisionPrescription);
  public
    {!script hide}
    function Link : TFhirVisionPrescriptionList; Overload;
    function Clone : TFhirVisionPrescriptionList; Overload;
    function GetEnumerator : TFhirVisionPrescriptionListEnumerator;
    {!script show}
    

    {@member Append
      Add a FhirVisionPrescription to the end of the list.
    }
    function Append : TFhirVisionPrescription;

    
    {@member AddItem
      Add an already existing FhirVisionPrescription to the end of the list.
    }
    procedure AddItem(value : TFhirVisionPrescription); overload;

    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    
    {@member IndexOf
      See if an item is already in the list. returns -1 if not in the list
    }
    function IndexOf(value : TFhirVisionPrescription) : Integer;
    

    {@member Insert
      Insert FhirVisionPrescription before the designated index (0 = first item)
    }
    function Insert(index : Integer) : TFhirVisionPrescription;
    

    {@member InsertItem
       Insert an existing FhirVisionPrescription before the designated index (0 = first item)
    }
    procedure InsertItem(index : Integer; value : TFhirVisionPrescription);
    
    {@member Item
       Get the iIndexth FhirVisionPrescription. (0 = first item)
    }
    
    {@member Item
       Get the iIndexth FhirVisionPrescription. (0 = first item)
    }
    procedure SetItemByIndex(index : Integer; value : TFhirVisionPrescription);
    
    {@member Count
      The number of items in the collection
    }
    function Item(index : Integer) : TFhirVisionPrescription;
    
    {@member Count
      The number of items in the collection
    }
    function Count : Integer; Overload;
    
    {@member remove
      Remove the indexth item. The first item is index 0.
    }
    procedure Remove(index : Integer);
    {@member ClearItems
      Remove All Items from the list
    }
    procedure ClearItems;
    
    Property FhirVisionPrescriptions[index : Integer] : TFhirVisionPrescription read GetItemN write SetItemN; default;
  End;


  {@Class TFhirResourceFactory : TFHIRBaseFactory

     FHIR factory: class constructors and general useful builders

  }

 TFhirResourceFactory = class (TFHIRBaseFactory)
  public
    {@member newElement
      create a new Element
    }
    {!script nolink}
    function newElement : TFhirElement;
    {@member newBackboneElement
      create a new BackboneElement
    }
    {!script nolink}
    function newBackboneElement : TFhirBackboneElement;
    {@member newEnum
      create a new enum
    }
    {!script nolink}
    function newEnum : TFhirEnum;
    {@member makeEnum
      create a new enum with the given value
    }
    {!script nolink}
    function makeEnum(value : String) : TFhirEnum;
    {@member newInteger
      create a new integer
    }
    {!script nolink}
    function newInteger : TFhirInteger;
    {@member makeInteger
      create a new integer with the given value
    }
    {!script nolink}
    function makeInteger(value : String) : TFhirInteger;
    {@member newDateTime
      create a new dateTime
    }
    {!script nolink}
    function newDateTime : TFhirDateTime;
    {@member makeDateTime
      create a new dateTime with the given value
    }
    {!script nolink}
    function makeDateTime(value : TDateAndTime) : TFhirDateTime;
    {@member newDate
      create a new date
    }
    {!script nolink}
    function newDate : TFhirDate;
    {@member makeDate
      create a new date with the given value
    }
    {!script nolink}
    function makeDate(value : TDateAndTime) : TFhirDate;
    {@member newDecimal
      create a new decimal
    }
    {!script nolink}
    function newDecimal : TFhirDecimal;
    {@member makeDecimal
      create a new decimal with the given value
    }
    {!script nolink}
    function makeDecimal(value : String) : TFhirDecimal;
    {@member newUri
      create a new uri
    }
    {!script nolink}
    function newUri : TFhirUri;
    {@member makeUri
      create a new uri with the given value
    }
    {!script nolink}
    function makeUri(value : String) : TFhirUri;
    {@member newBase64Binary
      create a new base64Binary
    }
    {!script nolink}
    function newBase64Binary : TFhirBase64Binary;
    {@member makeBase64Binary
      create a new base64Binary with the given value
    }
    {!script nolink}
    function makeBase64Binary(value : TBytes) : TFhirBase64Binary;
    {@member newTime
      create a new time
    }
    {!script nolink}
    function newTime : TFhirTime;
    {@member makeTime
      create a new time with the given value
    }
    {!script nolink}
    function makeTime(value : String) : TFhirTime;
    {@member newString
      create a new string
    }
    {!script nolink}
    function newString : TFhirString;
    {@member makeString
      create a new string with the given value
    }
    {!script nolink}
    function makeString(value : String) : TFhirString;
    {@member newBoolean
      create a new boolean
    }
    {!script nolink}
    function newBoolean : TFhirBoolean;
    {@member makeBoolean
      create a new boolean with the given value
    }
    {!script nolink}
    function makeBoolean(value : Boolean) : TFhirBoolean;
    {@member newInstant
      create a new instant
    }
    {!script nolink}
    function newInstant : TFhirInstant;
    {@member makeInstant
      create a new instant with the given value
    }
    {!script nolink}
    function makeInstant(value : TDateAndTime) : TFhirInstant;
    {@member newMarkdown
      create a new markdown
    }
    {!script nolink}
    function newMarkdown : TFhirMarkdown;
    {@member makeMarkdown
      create a new markdown with the given value
    }
    {!script nolink}
    function makeMarkdown(value : String) : TFhirMarkdown;
    {@member newUnsignedInt
      create a new unsignedInt
    }
    {!script nolink}
    function newUnsignedInt : TFhirUnsignedInt;
    {@member makeUnsignedInt
      create a new unsignedInt with the given value
    }
    {!script nolink}
    function makeUnsignedInt(value : String) : TFhirUnsignedInt;
    {@member newCode
      create a new code
    }
    {!script nolink}
    function newCode : TFhirCode;
    {@member makeCode
      create a new code with the given value
    }
    {!script nolink}
    function makeCode(value : String) : TFhirCode;
    {@member newId
      create a new id
    }
    {!script nolink}
    function newId : TFhirId;
    {@member makeId
      create a new id with the given value
    }
    {!script nolink}
    function makeId(value : String) : TFhirId;
    {@member newOid
      create a new oid
    }
    {!script nolink}
    function newOid : TFhirOid;
    {@member makeOid
      create a new oid with the given value
    }
    {!script nolink}
    function makeOid(value : String) : TFhirOid;
    {@member newPositiveInt
      create a new positiveInt
    }
    {!script nolink}
    function newPositiveInt : TFhirPositiveInt;
    {@member makePositiveInt
      create a new positiveInt with the given value
    }
    {!script nolink}
    function makePositiveInt(value : String) : TFhirPositiveInt;
    {@member newUuid
      create a new uuid
    }
    {!script nolink}
    function newUuid : TFhirUuid;
    {@member makeUuid
      create a new uuid with the given value
    }
    {!script nolink}
    function makeUuid(value : String) : TFhirUuid;
    {@member newParametersParameter
      create a new parameter
    }
    {!script nolink}
    function newParametersParameter : TFhirParametersParameter;
    {@member newParameters
      create a new Parameters
    }
    {!script nolink}
    function newParameters : TFhirParameters;
    {@member newExtension
      create a new Extension
    }
    {!script nolink}
    function newExtension : TFhirExtension;
    {@member newNarrative
      create a new Narrative
    }
    {!script nolink}
    function newNarrative : TFhirNarrative;
    {@member newPeriod
      create a new Period
    }
    {!script nolink}
    function newPeriod : TFhirPeriod;
    {@member newCoding
      create a new Coding
    }
    {!script nolink}
    function newCoding : TFhirCoding;
    {@member newRange
      create a new Range
    }
    {!script nolink}
    function newRange : TFhirRange;
    {@member newQuantity
      create a new Quantity
    }
    {!script nolink}
    function newQuantity : TFhirQuantity;
    {@member newAttachment
      create a new Attachment
    }
    {!script nolink}
    function newAttachment : TFhirAttachment;
    {@member newRatio
      create a new Ratio
    }
    {!script nolink}
    function newRatio : TFhirRatio;
    {@member newAnnotation
      create a new Annotation
    }
    {!script nolink}
    function newAnnotation : TFhirAnnotation;
    {@member newSampledData
      create a new SampledData
    }
    {!script nolink}
    function newSampledData : TFhirSampledData;
    {@member newReference
      create a new Reference
    }
    {!script nolink}
    function newReference : TFhirReference;
    {@member newCodeableConcept
      create a new CodeableConcept
    }
    {!script nolink}
    function newCodeableConcept : TFhirCodeableConcept;
    {@member newIdentifier
      create a new Identifier
    }
    {!script nolink}
    function newIdentifier : TFhirIdentifier;
    {@member newSignature
      create a new Signature
    }
    {!script nolink}
    function newSignature : TFhirSignature;
    {@member newElementDefinitionSlicing
      create a new slicing
    }
    {!script nolink}
    function newElementDefinitionSlicing : TFhirElementDefinitionSlicing;
    {@member newElementDefinitionBase
      create a new base
    }
    {!script nolink}
    function newElementDefinitionBase : TFhirElementDefinitionBase;
    {@member newElementDefinitionType
      create a new type
    }
    {!script nolink}
    function newElementDefinitionType : TFhirElementDefinitionType;
    {@member newElementDefinitionConstraint
      create a new constraint
    }
    {!script nolink}
    function newElementDefinitionConstraint : TFhirElementDefinitionConstraint;
    {@member newElementDefinitionBinding
      create a new binding
    }
    {!script nolink}
    function newElementDefinitionBinding : TFhirElementDefinitionBinding;
    {@member newElementDefinitionMapping
      create a new mapping
    }
    {!script nolink}
    function newElementDefinitionMapping : TFhirElementDefinitionMapping;
    {@member newElementDefinition
      create a new ElementDefinition
    }
    {!script nolink}
    function newElementDefinition : TFhirElementDefinition;
    {@member newTimingRepeat
      create a new repeat
    }
    {!script nolink}
    function newTimingRepeat : TFhirTimingRepeat;
    {@member newTiming
      create a new Timing
    }
    {!script nolink}
    function newTiming : TFhirTiming;
    {@member newAddress
      create a new Address
    }
    {!script nolink}
    function newAddress : TFhirAddress;
    {@member newHumanName
      create a new HumanName
    }
    {!script nolink}
    function newHumanName : TFhirHumanName;
    {@member newMeta
      create a new Meta
    }
    {!script nolink}
    function newMeta : TFhirMeta;
    {@member newContactPoint
      create a new ContactPoint
    }
    {!script nolink}
    function newContactPoint : TFhirContactPoint;
    {@member newAccount
      create a new Account
    }
    {!script nolink}
    function newAccount : TFhirAccount;
    {@member newAllergyIntoleranceReaction
      create a new reaction
    }
    {!script nolink}
    function newAllergyIntoleranceReaction : TFhirAllergyIntoleranceReaction;
    {@member newAllergyIntolerance
      create a new AllergyIntolerance
    }
    {!script nolink}
    function newAllergyIntolerance : TFhirAllergyIntolerance;
    {@member newAppointmentParticipant
      create a new participant
    }
    {!script nolink}
    function newAppointmentParticipant : TFhirAppointmentParticipant;
    {@member newAppointment
      create a new Appointment
    }
    {!script nolink}
    function newAppointment : TFhirAppointment;
    {@member newAppointmentResponse
      create a new AppointmentResponse
    }
    {!script nolink}
    function newAppointmentResponse : TFhirAppointmentResponse;
    {@member newAuditEventEvent
      create a new event
    }
    {!script nolink}
    function newAuditEventEvent : TFhirAuditEventEvent;
    {@member newAuditEventParticipant
      create a new participant
    }
    {!script nolink}
    function newAuditEventParticipant : TFhirAuditEventParticipant;
    {@member newAuditEventParticipantNetwork
      create a new network
    }
    {!script nolink}
    function newAuditEventParticipantNetwork : TFhirAuditEventParticipantNetwork;
    {@member newAuditEventSource
      create a new source
    }
    {!script nolink}
    function newAuditEventSource : TFhirAuditEventSource;
    {@member newAuditEventObject
      create a new object
    }
    {!script nolink}
    function newAuditEventObject : TFhirAuditEventObject;
    {@member newAuditEventObjectDetail
      create a new detail
    }
    {!script nolink}
    function newAuditEventObjectDetail : TFhirAuditEventObjectDetail;
    {@member newAuditEvent
      create a new AuditEvent
    }
    {!script nolink}
    function newAuditEvent : TFhirAuditEvent;
    {@member newBasic
      create a new Basic
    }
    {!script nolink}
    function newBasic : TFhirBasic;
    {@member newBinary
      create a new Binary
    }
    {!script nolink}
    function newBinary : TFhirBinary;
    {@member newBodySite
      create a new BodySite
    }
    {!script nolink}
    function newBodySite : TFhirBodySite;
    {@member newBundleLink
      create a new link
    }
    {!script nolink}
    function newBundleLink : TFhirBundleLink;
    {@member newBundleEntry
      create a new entry
    }
    {!script nolink}
    function newBundleEntry : TFhirBundleEntry;
    {@member newBundleEntrySearch
      create a new search
    }
    {!script nolink}
    function newBundleEntrySearch : TFhirBundleEntrySearch;
    {@member newBundleEntryRequest
      create a new request
    }
    {!script nolink}
    function newBundleEntryRequest : TFhirBundleEntryRequest;
    {@member newBundleEntryResponse
      create a new response
    }
    {!script nolink}
    function newBundleEntryResponse : TFhirBundleEntryResponse;
    {@member newBundle
      create a new Bundle
    }
    {!script nolink}
    function newBundle : TFhirBundle;
    {@member newCarePlanRelatedPlan
      create a new relatedPlan
    }
    {!script nolink}
    function newCarePlanRelatedPlan : TFhirCarePlanRelatedPlan;
    {@member newCarePlanParticipant
      create a new participant
    }
    {!script nolink}
    function newCarePlanParticipant : TFhirCarePlanParticipant;
    {@member newCarePlanActivity
      create a new activity
    }
    {!script nolink}
    function newCarePlanActivity : TFhirCarePlanActivity;
    {@member newCarePlanActivityDetail
      create a new detail
    }
    {!script nolink}
    function newCarePlanActivityDetail : TFhirCarePlanActivityDetail;
    {@member newCarePlan
      create a new CarePlan
    }
    {!script nolink}
    function newCarePlan : TFhirCarePlan;
    {@member newClaimPayee
      create a new payee
    }
    {!script nolink}
    function newClaimPayee : TFhirClaimPayee;
    {@member newClaimDiagnosis
      create a new diagnosis
    }
    {!script nolink}
    function newClaimDiagnosis : TFhirClaimDiagnosis;
    {@member newClaimCoverage
      create a new coverage
    }
    {!script nolink}
    function newClaimCoverage : TFhirClaimCoverage;
    {@member newClaimItem
      create a new item
    }
    {!script nolink}
    function newClaimItem : TFhirClaimItem;
    {@member newClaimItemDetail
      create a new detail
    }
    {!script nolink}
    function newClaimItemDetail : TFhirClaimItemDetail;
    {@member newClaimItemDetailSubDetail
      create a new subDetail
    }
    {!script nolink}
    function newClaimItemDetailSubDetail : TFhirClaimItemDetailSubDetail;
    {@member newClaimItemProsthesis
      create a new prosthesis
    }
    {!script nolink}
    function newClaimItemProsthesis : TFhirClaimItemProsthesis;
    {@member newClaimMissingTeeth
      create a new missingTeeth
    }
    {!script nolink}
    function newClaimMissingTeeth : TFhirClaimMissingTeeth;
    {@member newClaim
      create a new Claim
    }
    {!script nolink}
    function newClaim : TFhirClaim;
    {@member newClaimResponseItem
      create a new item
    }
    {!script nolink}
    function newClaimResponseItem : TFhirClaimResponseItem;
    {@member newClaimResponseItemAdjudication
      create a new adjudication
    }
    {!script nolink}
    function newClaimResponseItemAdjudication : TFhirClaimResponseItemAdjudication;
    {@member newClaimResponseItemDetail
      create a new detail
    }
    {!script nolink}
    function newClaimResponseItemDetail : TFhirClaimResponseItemDetail;
    {@member newClaimResponseItemDetailAdjudication
      create a new adjudication
    }
    {!script nolink}
    function newClaimResponseItemDetailAdjudication : TFhirClaimResponseItemDetailAdjudication;
    {@member newClaimResponseItemDetailSubDetail
      create a new subDetail
    }
    {!script nolink}
    function newClaimResponseItemDetailSubDetail : TFhirClaimResponseItemDetailSubDetail;
    {@member newClaimResponseItemDetailSubDetailAdjudication
      create a new adjudication
    }
    {!script nolink}
    function newClaimResponseItemDetailSubDetailAdjudication : TFhirClaimResponseItemDetailSubDetailAdjudication;
    {@member newClaimResponseAddItem
      create a new addItem
    }
    {!script nolink}
    function newClaimResponseAddItem : TFhirClaimResponseAddItem;
    {@member newClaimResponseAddItemAdjudication
      create a new adjudication
    }
    {!script nolink}
    function newClaimResponseAddItemAdjudication : TFhirClaimResponseAddItemAdjudication;
    {@member newClaimResponseAddItemDetail
      create a new detail
    }
    {!script nolink}
    function newClaimResponseAddItemDetail : TFhirClaimResponseAddItemDetail;
    {@member newClaimResponseAddItemDetailAdjudication
      create a new adjudication
    }
    {!script nolink}
    function newClaimResponseAddItemDetailAdjudication : TFhirClaimResponseAddItemDetailAdjudication;
    {@member newClaimResponseError
      create a new error
    }
    {!script nolink}
    function newClaimResponseError : TFhirClaimResponseError;
    {@member newClaimResponseNote
      create a new note
    }
    {!script nolink}
    function newClaimResponseNote : TFhirClaimResponseNote;
    {@member newClaimResponseCoverage
      create a new coverage
    }
    {!script nolink}
    function newClaimResponseCoverage : TFhirClaimResponseCoverage;
    {@member newClaimResponse
      create a new ClaimResponse
    }
    {!script nolink}
    function newClaimResponse : TFhirClaimResponse;
    {@member newClinicalImpressionInvestigations
      create a new investigations
    }
    {!script nolink}
    function newClinicalImpressionInvestigations : TFhirClinicalImpressionInvestigations;
    {@member newClinicalImpressionFinding
      create a new finding
    }
    {!script nolink}
    function newClinicalImpressionFinding : TFhirClinicalImpressionFinding;
    {@member newClinicalImpressionRuledOut
      create a new ruledOut
    }
    {!script nolink}
    function newClinicalImpressionRuledOut : TFhirClinicalImpressionRuledOut;
    {@member newClinicalImpression
      create a new ClinicalImpression
    }
    {!script nolink}
    function newClinicalImpression : TFhirClinicalImpression;
    {@member newCommunicationPayload
      create a new payload
    }
    {!script nolink}
    function newCommunicationPayload : TFhirCommunicationPayload;
    {@member newCommunication
      create a new Communication
    }
    {!script nolink}
    function newCommunication : TFhirCommunication;
    {@member newCommunicationRequestPayload
      create a new payload
    }
    {!script nolink}
    function newCommunicationRequestPayload : TFhirCommunicationRequestPayload;
    {@member newCommunicationRequest
      create a new CommunicationRequest
    }
    {!script nolink}
    function newCommunicationRequest : TFhirCommunicationRequest;
    {@member newCompositionAttester
      create a new attester
    }
    {!script nolink}
    function newCompositionAttester : TFhirCompositionAttester;
    {@member newCompositionEvent
      create a new event
    }
    {!script nolink}
    function newCompositionEvent : TFhirCompositionEvent;
    {@member newCompositionSection
      create a new section
    }
    {!script nolink}
    function newCompositionSection : TFhirCompositionSection;
    {@member newComposition
      create a new Composition
    }
    {!script nolink}
    function newComposition : TFhirComposition;
    {@member newConceptMapContact
      create a new contact
    }
    {!script nolink}
    function newConceptMapContact : TFhirConceptMapContact;
    {@member newConceptMapElement
      create a new element
    }
    {!script nolink}
    function newConceptMapElement : TFhirConceptMapElement;
    {@member newConceptMapElementTarget
      create a new target
    }
    {!script nolink}
    function newConceptMapElementTarget : TFhirConceptMapElementTarget;
    {@member newConceptMapElementTargetDependsOn
      create a new dependsOn
    }
    {!script nolink}
    function newConceptMapElementTargetDependsOn : TFhirConceptMapElementTargetDependsOn;
    {@member newConceptMap
      create a new ConceptMap
    }
    {!script nolink}
    function newConceptMap : TFhirConceptMap;
    {@member newConditionStage
      create a new stage
    }
    {!script nolink}
    function newConditionStage : TFhirConditionStage;
    {@member newConditionEvidence
      create a new evidence
    }
    {!script nolink}
    function newConditionEvidence : TFhirConditionEvidence;
    {@member newCondition
      create a new Condition
    }
    {!script nolink}
    function newCondition : TFhirCondition;
    {@member newConformanceContact
      create a new contact
    }
    {!script nolink}
    function newConformanceContact : TFhirConformanceContact;
    {@member newConformanceSoftware
      create a new software
    }
    {!script nolink}
    function newConformanceSoftware : TFhirConformanceSoftware;
    {@member newConformanceImplementation
      create a new implementation
    }
    {!script nolink}
    function newConformanceImplementation : TFhirConformanceImplementation;
    {@member newConformanceRest
      create a new rest
    }
    {!script nolink}
    function newConformanceRest : TFhirConformanceRest;
    {@member newConformanceRestSecurity
      create a new security
    }
    {!script nolink}
    function newConformanceRestSecurity : TFhirConformanceRestSecurity;
    {@member newConformanceRestSecurityCertificate
      create a new certificate
    }
    {!script nolink}
    function newConformanceRestSecurityCertificate : TFhirConformanceRestSecurityCertificate;
    {@member newConformanceRestResource
      create a new resource
    }
    {!script nolink}
    function newConformanceRestResource : TFhirConformanceRestResource;
    {@member newConformanceRestResourceInteraction
      create a new interaction
    }
    {!script nolink}
    function newConformanceRestResourceInteraction : TFhirConformanceRestResourceInteraction;
    {@member newConformanceRestResourceSearchParam
      create a new searchParam
    }
    {!script nolink}
    function newConformanceRestResourceSearchParam : TFhirConformanceRestResourceSearchParam;
    {@member newConformanceRestInteraction
      create a new interaction
    }
    {!script nolink}
    function newConformanceRestInteraction : TFhirConformanceRestInteraction;
    {@member newConformanceRestOperation
      create a new operation
    }
    {!script nolink}
    function newConformanceRestOperation : TFhirConformanceRestOperation;
    {@member newConformanceMessaging
      create a new messaging
    }
    {!script nolink}
    function newConformanceMessaging : TFhirConformanceMessaging;
    {@member newConformanceMessagingEndpoint
      create a new endpoint
    }
    {!script nolink}
    function newConformanceMessagingEndpoint : TFhirConformanceMessagingEndpoint;
    {@member newConformanceMessagingEvent
      create a new event
    }
    {!script nolink}
    function newConformanceMessagingEvent : TFhirConformanceMessagingEvent;
    {@member newConformanceDocument
      create a new document
    }
    {!script nolink}
    function newConformanceDocument : TFhirConformanceDocument;
    {@member newConformance
      create a new Conformance
    }
    {!script nolink}
    function newConformance : TFhirConformance;
    {@member newContractActor
      create a new actor
    }
    {!script nolink}
    function newContractActor : TFhirContractActor;
    {@member newContractValuedItem
      create a new valuedItem
    }
    {!script nolink}
    function newContractValuedItem : TFhirContractValuedItem;
    {@member newContractSigner
      create a new signer
    }
    {!script nolink}
    function newContractSigner : TFhirContractSigner;
    {@member newContractTerm
      create a new term
    }
    {!script nolink}
    function newContractTerm : TFhirContractTerm;
    {@member newContractTermActor
      create a new actor
    }
    {!script nolink}
    function newContractTermActor : TFhirContractTermActor;
    {@member newContractTermValuedItem
      create a new valuedItem
    }
    {!script nolink}
    function newContractTermValuedItem : TFhirContractTermValuedItem;
    {@member newContractFriendly
      create a new friendly
    }
    {!script nolink}
    function newContractFriendly : TFhirContractFriendly;
    {@member newContractLegal
      create a new legal
    }
    {!script nolink}
    function newContractLegal : TFhirContractLegal;
    {@member newContractRule
      create a new rule
    }
    {!script nolink}
    function newContractRule : TFhirContractRule;
    {@member newContract
      create a new Contract
    }
    {!script nolink}
    function newContract : TFhirContract;
    {@member newCoverage
      create a new Coverage
    }
    {!script nolink}
    function newCoverage : TFhirCoverage;
    {@member newDataElementContact
      create a new contact
    }
    {!script nolink}
    function newDataElementContact : TFhirDataElementContact;
    {@member newDataElementMapping
      create a new mapping
    }
    {!script nolink}
    function newDataElementMapping : TFhirDataElementMapping;
    {@member newDataElement
      create a new DataElement
    }
    {!script nolink}
    function newDataElement : TFhirDataElement;
    {@member newDetectedIssueMitigation
      create a new mitigation
    }
    {!script nolink}
    function newDetectedIssueMitigation : TFhirDetectedIssueMitigation;
    {@member newDetectedIssue
      create a new DetectedIssue
    }
    {!script nolink}
    function newDetectedIssue : TFhirDetectedIssue;
    {@member newDevice
      create a new Device
    }
    {!script nolink}
    function newDevice : TFhirDevice;
    {@member newDeviceComponentProductionSpecification
      create a new productionSpecification
    }
    {!script nolink}
    function newDeviceComponentProductionSpecification : TFhirDeviceComponentProductionSpecification;
    {@member newDeviceComponent
      create a new DeviceComponent
    }
    {!script nolink}
    function newDeviceComponent : TFhirDeviceComponent;
    {@member newDeviceMetricCalibration
      create a new calibration
    }
    {!script nolink}
    function newDeviceMetricCalibration : TFhirDeviceMetricCalibration;
    {@member newDeviceMetric
      create a new DeviceMetric
    }
    {!script nolink}
    function newDeviceMetric : TFhirDeviceMetric;
    {@member newDeviceUseRequest
      create a new DeviceUseRequest
    }
    {!script nolink}
    function newDeviceUseRequest : TFhirDeviceUseRequest;
    {@member newDeviceUseStatement
      create a new DeviceUseStatement
    }
    {!script nolink}
    function newDeviceUseStatement : TFhirDeviceUseStatement;
    {@member newDiagnosticOrderEvent
      create a new event
    }
    {!script nolink}
    function newDiagnosticOrderEvent : TFhirDiagnosticOrderEvent;
    {@member newDiagnosticOrderItem
      create a new item
    }
    {!script nolink}
    function newDiagnosticOrderItem : TFhirDiagnosticOrderItem;
    {@member newDiagnosticOrder
      create a new DiagnosticOrder
    }
    {!script nolink}
    function newDiagnosticOrder : TFhirDiagnosticOrder;
    {@member newDiagnosticReportImage
      create a new image
    }
    {!script nolink}
    function newDiagnosticReportImage : TFhirDiagnosticReportImage;
    {@member newDiagnosticReport
      create a new DiagnosticReport
    }
    {!script nolink}
    function newDiagnosticReport : TFhirDiagnosticReport;
    {@member newDocumentManifestContent
      create a new content
    }
    {!script nolink}
    function newDocumentManifestContent : TFhirDocumentManifestContent;
    {@member newDocumentManifestRelated
      create a new related
    }
    {!script nolink}
    function newDocumentManifestRelated : TFhirDocumentManifestRelated;
    {@member newDocumentManifest
      create a new DocumentManifest
    }
    {!script nolink}
    function newDocumentManifest : TFhirDocumentManifest;
    {@member newDocumentReferenceRelatesTo
      create a new relatesTo
    }
    {!script nolink}
    function newDocumentReferenceRelatesTo : TFhirDocumentReferenceRelatesTo;
    {@member newDocumentReferenceContent
      create a new content
    }
    {!script nolink}
    function newDocumentReferenceContent : TFhirDocumentReferenceContent;
    {@member newDocumentReferenceContext
      create a new context
    }
    {!script nolink}
    function newDocumentReferenceContext : TFhirDocumentReferenceContext;
    {@member newDocumentReferenceContextRelated
      create a new related
    }
    {!script nolink}
    function newDocumentReferenceContextRelated : TFhirDocumentReferenceContextRelated;
    {@member newDocumentReference
      create a new DocumentReference
    }
    {!script nolink}
    function newDocumentReference : TFhirDocumentReference;
    {@member newEligibilityRequest
      create a new EligibilityRequest
    }
    {!script nolink}
    function newEligibilityRequest : TFhirEligibilityRequest;
    {@member newEligibilityResponse
      create a new EligibilityResponse
    }
    {!script nolink}
    function newEligibilityResponse : TFhirEligibilityResponse;
    {@member newEncounterStatusHistory
      create a new statusHistory
    }
    {!script nolink}
    function newEncounterStatusHistory : TFhirEncounterStatusHistory;
    {@member newEncounterParticipant
      create a new participant
    }
    {!script nolink}
    function newEncounterParticipant : TFhirEncounterParticipant;
    {@member newEncounterHospitalization
      create a new hospitalization
    }
    {!script nolink}
    function newEncounterHospitalization : TFhirEncounterHospitalization;
    {@member newEncounterLocation
      create a new location
    }
    {!script nolink}
    function newEncounterLocation : TFhirEncounterLocation;
    {@member newEncounter
      create a new Encounter
    }
    {!script nolink}
    function newEncounter : TFhirEncounter;
    {@member newEnrollmentRequest
      create a new EnrollmentRequest
    }
    {!script nolink}
    function newEnrollmentRequest : TFhirEnrollmentRequest;
    {@member newEnrollmentResponse
      create a new EnrollmentResponse
    }
    {!script nolink}
    function newEnrollmentResponse : TFhirEnrollmentResponse;
    {@member newEpisodeOfCareStatusHistory
      create a new statusHistory
    }
    {!script nolink}
    function newEpisodeOfCareStatusHistory : TFhirEpisodeOfCareStatusHistory;
    {@member newEpisodeOfCareCareTeam
      create a new careTeam
    }
    {!script nolink}
    function newEpisodeOfCareCareTeam : TFhirEpisodeOfCareCareTeam;
    {@member newEpisodeOfCare
      create a new EpisodeOfCare
    }
    {!script nolink}
    function newEpisodeOfCare : TFhirEpisodeOfCare;
    {@member newExplanationOfBenefit
      create a new ExplanationOfBenefit
    }
    {!script nolink}
    function newExplanationOfBenefit : TFhirExplanationOfBenefit;
    {@member newFamilyMemberHistoryCondition
      create a new condition
    }
    {!script nolink}
    function newFamilyMemberHistoryCondition : TFhirFamilyMemberHistoryCondition;
    {@member newFamilyMemberHistory
      create a new FamilyMemberHistory
    }
    {!script nolink}
    function newFamilyMemberHistory : TFhirFamilyMemberHistory;
    {@member newFlag
      create a new Flag
    }
    {!script nolink}
    function newFlag : TFhirFlag;
    {@member newGoalOutcome
      create a new outcome
    }
    {!script nolink}
    function newGoalOutcome : TFhirGoalOutcome;
    {@member newGoal
      create a new Goal
    }
    {!script nolink}
    function newGoal : TFhirGoal;
    {@member newGroupCharacteristic
      create a new characteristic
    }
    {!script nolink}
    function newGroupCharacteristic : TFhirGroupCharacteristic;
    {@member newGroupMember
      create a new member
    }
    {!script nolink}
    function newGroupMember : TFhirGroupMember;
    {@member newGroup
      create a new Group
    }
    {!script nolink}
    function newGroup : TFhirGroup;
    {@member newHealthcareServiceServiceType
      create a new serviceType
    }
    {!script nolink}
    function newHealthcareServiceServiceType : TFhirHealthcareServiceServiceType;
    {@member newHealthcareServiceAvailableTime
      create a new availableTime
    }
    {!script nolink}
    function newHealthcareServiceAvailableTime : TFhirHealthcareServiceAvailableTime;
    {@member newHealthcareServiceNotAvailable
      create a new notAvailable
    }
    {!script nolink}
    function newHealthcareServiceNotAvailable : TFhirHealthcareServiceNotAvailable;
    {@member newHealthcareService
      create a new HealthcareService
    }
    {!script nolink}
    function newHealthcareService : TFhirHealthcareService;
    {@member newImagingObjectSelectionStudy
      create a new study
    }
    {!script nolink}
    function newImagingObjectSelectionStudy : TFhirImagingObjectSelectionStudy;
    {@member newImagingObjectSelectionStudySeries
      create a new series
    }
    {!script nolink}
    function newImagingObjectSelectionStudySeries : TFhirImagingObjectSelectionStudySeries;
    {@member newImagingObjectSelectionStudySeriesInstance
      create a new instance
    }
    {!script nolink}
    function newImagingObjectSelectionStudySeriesInstance : TFhirImagingObjectSelectionStudySeriesInstance;
    {@member newImagingObjectSelectionStudySeriesInstanceFrames
      create a new frames
    }
    {!script nolink}
    function newImagingObjectSelectionStudySeriesInstanceFrames : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
    {@member newImagingObjectSelection
      create a new ImagingObjectSelection
    }
    {!script nolink}
    function newImagingObjectSelection : TFhirImagingObjectSelection;
    {@member newImagingStudySeries
      create a new series
    }
    {!script nolink}
    function newImagingStudySeries : TFhirImagingStudySeries;
    {@member newImagingStudySeriesInstance
      create a new instance
    }
    {!script nolink}
    function newImagingStudySeriesInstance : TFhirImagingStudySeriesInstance;
    {@member newImagingStudy
      create a new ImagingStudy
    }
    {!script nolink}
    function newImagingStudy : TFhirImagingStudy;
    {@member newImmunizationExplanation
      create a new explanation
    }
    {!script nolink}
    function newImmunizationExplanation : TFhirImmunizationExplanation;
    {@member newImmunizationReaction
      create a new reaction
    }
    {!script nolink}
    function newImmunizationReaction : TFhirImmunizationReaction;
    {@member newImmunizationVaccinationProtocol
      create a new vaccinationProtocol
    }
    {!script nolink}
    function newImmunizationVaccinationProtocol : TFhirImmunizationVaccinationProtocol;
    {@member newImmunization
      create a new Immunization
    }
    {!script nolink}
    function newImmunization : TFhirImmunization;
    {@member newImmunizationRecommendationRecommendation
      create a new recommendation
    }
    {!script nolink}
    function newImmunizationRecommendationRecommendation : TFhirImmunizationRecommendationRecommendation;
    {@member newImmunizationRecommendationRecommendationDateCriterion
      create a new dateCriterion
    }
    {!script nolink}
    function newImmunizationRecommendationRecommendationDateCriterion : TFhirImmunizationRecommendationRecommendationDateCriterion;
    {@member newImmunizationRecommendationRecommendationProtocol
      create a new protocol
    }
    {!script nolink}
    function newImmunizationRecommendationRecommendationProtocol : TFhirImmunizationRecommendationRecommendationProtocol;
    {@member newImmunizationRecommendation
      create a new ImmunizationRecommendation
    }
    {!script nolink}
    function newImmunizationRecommendation : TFhirImmunizationRecommendation;
    {@member newImplementationGuideContact
      create a new contact
    }
    {!script nolink}
    function newImplementationGuideContact : TFhirImplementationGuideContact;
    {@member newImplementationGuideDependency
      create a new dependency
    }
    {!script nolink}
    function newImplementationGuideDependency : TFhirImplementationGuideDependency;
    {@member newImplementationGuidePackage
      create a new package
    }
    {!script nolink}
    function newImplementationGuidePackage : TFhirImplementationGuidePackage;
    {@member newImplementationGuidePackageResource
      create a new resource
    }
    {!script nolink}
    function newImplementationGuidePackageResource : TFhirImplementationGuidePackageResource;
    {@member newImplementationGuideGlobal
      create a new global
    }
    {!script nolink}
    function newImplementationGuideGlobal : TFhirImplementationGuideGlobal;
    {@member newImplementationGuidePage
      create a new page
    }
    {!script nolink}
    function newImplementationGuidePage : TFhirImplementationGuidePage;
    {@member newImplementationGuide
      create a new ImplementationGuide
    }
    {!script nolink}
    function newImplementationGuide : TFhirImplementationGuide;
    {@member newListEntry
      create a new entry
    }
    {!script nolink}
    function newListEntry : TFhirListEntry;
    {@member newList
      create a new List
    }
    {!script nolink}
    function newList : TFhirList;
    {@member newLocationPosition
      create a new position
    }
    {!script nolink}
    function newLocationPosition : TFhirLocationPosition;
    {@member newLocation
      create a new Location
    }
    {!script nolink}
    function newLocation : TFhirLocation;
    {@member newMedia
      create a new Media
    }
    {!script nolink}
    function newMedia : TFhirMedia;
    {@member newMedicationProduct
      create a new product
    }
    {!script nolink}
    function newMedicationProduct : TFhirMedicationProduct;
    {@member newMedicationProductIngredient
      create a new ingredient
    }
    {!script nolink}
    function newMedicationProductIngredient : TFhirMedicationProductIngredient;
    {@member newMedicationProductBatch
      create a new batch
    }
    {!script nolink}
    function newMedicationProductBatch : TFhirMedicationProductBatch;
    {@member newMedicationPackage
      create a new package
    }
    {!script nolink}
    function newMedicationPackage : TFhirMedicationPackage;
    {@member newMedicationPackageContent
      create a new content
    }
    {!script nolink}
    function newMedicationPackageContent : TFhirMedicationPackageContent;
    {@member newMedication
      create a new Medication
    }
    {!script nolink}
    function newMedication : TFhirMedication;
    {@member newMedicationAdministrationDosage
      create a new dosage
    }
    {!script nolink}
    function newMedicationAdministrationDosage : TFhirMedicationAdministrationDosage;
    {@member newMedicationAdministration
      create a new MedicationAdministration
    }
    {!script nolink}
    function newMedicationAdministration : TFhirMedicationAdministration;
    {@member newMedicationDispenseDosageInstruction
      create a new dosageInstruction
    }
    {!script nolink}
    function newMedicationDispenseDosageInstruction : TFhirMedicationDispenseDosageInstruction;
    {@member newMedicationDispenseSubstitution
      create a new substitution
    }
    {!script nolink}
    function newMedicationDispenseSubstitution : TFhirMedicationDispenseSubstitution;
    {@member newMedicationDispense
      create a new MedicationDispense
    }
    {!script nolink}
    function newMedicationDispense : TFhirMedicationDispense;
    {@member newMedicationOrderDosageInstruction
      create a new dosageInstruction
    }
    {!script nolink}
    function newMedicationOrderDosageInstruction : TFhirMedicationOrderDosageInstruction;
    {@member newMedicationOrderDispenseRequest
      create a new dispenseRequest
    }
    {!script nolink}
    function newMedicationOrderDispenseRequest : TFhirMedicationOrderDispenseRequest;
    {@member newMedicationOrderSubstitution
      create a new substitution
    }
    {!script nolink}
    function newMedicationOrderSubstitution : TFhirMedicationOrderSubstitution;
    {@member newMedicationOrder
      create a new MedicationOrder
    }
    {!script nolink}
    function newMedicationOrder : TFhirMedicationOrder;
    {@member newMedicationStatementDosage
      create a new dosage
    }
    {!script nolink}
    function newMedicationStatementDosage : TFhirMedicationStatementDosage;
    {@member newMedicationStatement
      create a new MedicationStatement
    }
    {!script nolink}
    function newMedicationStatement : TFhirMedicationStatement;
    {@member newMessageHeaderResponse
      create a new response
    }
    {!script nolink}
    function newMessageHeaderResponse : TFhirMessageHeaderResponse;
    {@member newMessageHeaderSource
      create a new source
    }
    {!script nolink}
    function newMessageHeaderSource : TFhirMessageHeaderSource;
    {@member newMessageHeaderDestination
      create a new destination
    }
    {!script nolink}
    function newMessageHeaderDestination : TFhirMessageHeaderDestination;
    {@member newMessageHeader
      create a new MessageHeader
    }
    {!script nolink}
    function newMessageHeader : TFhirMessageHeader;
    {@member newNamingSystemContact
      create a new contact
    }
    {!script nolink}
    function newNamingSystemContact : TFhirNamingSystemContact;
    {@member newNamingSystemUniqueId
      create a new uniqueId
    }
    {!script nolink}
    function newNamingSystemUniqueId : TFhirNamingSystemUniqueId;
    {@member newNamingSystem
      create a new NamingSystem
    }
    {!script nolink}
    function newNamingSystem : TFhirNamingSystem;
    {@member newNutritionOrderOralDiet
      create a new oralDiet
    }
    {!script nolink}
    function newNutritionOrderOralDiet : TFhirNutritionOrderOralDiet;
    {@member newNutritionOrderOralDietNutrient
      create a new nutrient
    }
    {!script nolink}
    function newNutritionOrderOralDietNutrient : TFhirNutritionOrderOralDietNutrient;
    {@member newNutritionOrderOralDietTexture
      create a new texture
    }
    {!script nolink}
    function newNutritionOrderOralDietTexture : TFhirNutritionOrderOralDietTexture;
    {@member newNutritionOrderSupplement
      create a new supplement
    }
    {!script nolink}
    function newNutritionOrderSupplement : TFhirNutritionOrderSupplement;
    {@member newNutritionOrderEnteralFormula
      create a new enteralFormula
    }
    {!script nolink}
    function newNutritionOrderEnteralFormula : TFhirNutritionOrderEnteralFormula;
    {@member newNutritionOrderEnteralFormulaAdministration
      create a new administration
    }
    {!script nolink}
    function newNutritionOrderEnteralFormulaAdministration : TFhirNutritionOrderEnteralFormulaAdministration;
    {@member newNutritionOrder
      create a new NutritionOrder
    }
    {!script nolink}
    function newNutritionOrder : TFhirNutritionOrder;
    {@member newObservationReferenceRange
      create a new referenceRange
    }
    {!script nolink}
    function newObservationReferenceRange : TFhirObservationReferenceRange;
    {@member newObservationRelated
      create a new related
    }
    {!script nolink}
    function newObservationRelated : TFhirObservationRelated;
    {@member newObservationComponent
      create a new component
    }
    {!script nolink}
    function newObservationComponent : TFhirObservationComponent;
    {@member newObservation
      create a new Observation
    }
    {!script nolink}
    function newObservation : TFhirObservation;
    {@member newOperationDefinitionContact
      create a new contact
    }
    {!script nolink}
    function newOperationDefinitionContact : TFhirOperationDefinitionContact;
    {@member newOperationDefinitionParameter
      create a new parameter
    }
    {!script nolink}
    function newOperationDefinitionParameter : TFhirOperationDefinitionParameter;
    {@member newOperationDefinitionParameterBinding
      create a new binding
    }
    {!script nolink}
    function newOperationDefinitionParameterBinding : TFhirOperationDefinitionParameterBinding;
    {@member newOperationDefinition
      create a new OperationDefinition
    }
    {!script nolink}
    function newOperationDefinition : TFhirOperationDefinition;
    {@member newOperationOutcomeIssue
      create a new issue
    }
    {!script nolink}
    function newOperationOutcomeIssue : TFhirOperationOutcomeIssue;
    {@member newOperationOutcome
      create a new OperationOutcome
    }
    {!script nolink}
    function newOperationOutcome : TFhirOperationOutcome;
    {@member newOrderWhen
      create a new when
    }
    {!script nolink}
    function newOrderWhen : TFhirOrderWhen;
    {@member newOrder
      create a new Order
    }
    {!script nolink}
    function newOrder : TFhirOrder;
    {@member newOrderResponse
      create a new OrderResponse
    }
    {!script nolink}
    function newOrderResponse : TFhirOrderResponse;
    {@member newOrganizationContact
      create a new contact
    }
    {!script nolink}
    function newOrganizationContact : TFhirOrganizationContact;
    {@member newOrganization
      create a new Organization
    }
    {!script nolink}
    function newOrganization : TFhirOrganization;
    {@member newPatientContact
      create a new contact
    }
    {!script nolink}
    function newPatientContact : TFhirPatientContact;
    {@member newPatientAnimal
      create a new animal
    }
    {!script nolink}
    function newPatientAnimal : TFhirPatientAnimal;
    {@member newPatientCommunication
      create a new communication
    }
    {!script nolink}
    function newPatientCommunication : TFhirPatientCommunication;
    {@member newPatientLink
      create a new link
    }
    {!script nolink}
    function newPatientLink : TFhirPatientLink;
    {@member newPatient
      create a new Patient
    }
    {!script nolink}
    function newPatient : TFhirPatient;
    {@member newPaymentNotice
      create a new PaymentNotice
    }
    {!script nolink}
    function newPaymentNotice : TFhirPaymentNotice;
    {@member newPaymentReconciliationDetail
      create a new detail
    }
    {!script nolink}
    function newPaymentReconciliationDetail : TFhirPaymentReconciliationDetail;
    {@member newPaymentReconciliationNote
      create a new note
    }
    {!script nolink}
    function newPaymentReconciliationNote : TFhirPaymentReconciliationNote;
    {@member newPaymentReconciliation
      create a new PaymentReconciliation
    }
    {!script nolink}
    function newPaymentReconciliation : TFhirPaymentReconciliation;
    {@member newPersonLink
      create a new link
    }
    {!script nolink}
    function newPersonLink : TFhirPersonLink;
    {@member newPerson
      create a new Person
    }
    {!script nolink}
    function newPerson : TFhirPerson;
    {@member newPractitionerPractitionerRole
      create a new practitionerRole
    }
    {!script nolink}
    function newPractitionerPractitionerRole : TFhirPractitionerPractitionerRole;
    {@member newPractitionerQualification
      create a new qualification
    }
    {!script nolink}
    function newPractitionerQualification : TFhirPractitionerQualification;
    {@member newPractitioner
      create a new Practitioner
    }
    {!script nolink}
    function newPractitioner : TFhirPractitioner;
    {@member newProcedurePerformer
      create a new performer
    }
    {!script nolink}
    function newProcedurePerformer : TFhirProcedurePerformer;
    {@member newProcedureFocalDevice
      create a new focalDevice
    }
    {!script nolink}
    function newProcedureFocalDevice : TFhirProcedureFocalDevice;
    {@member newProcedure
      create a new Procedure
    }
    {!script nolink}
    function newProcedure : TFhirProcedure;
    {@member newProcedureRequest
      create a new ProcedureRequest
    }
    {!script nolink}
    function newProcedureRequest : TFhirProcedureRequest;
    {@member newProcessRequestItem
      create a new item
    }
    {!script nolink}
    function newProcessRequestItem : TFhirProcessRequestItem;
    {@member newProcessRequest
      create a new ProcessRequest
    }
    {!script nolink}
    function newProcessRequest : TFhirProcessRequest;
    {@member newProcessResponseNotes
      create a new notes
    }
    {!script nolink}
    function newProcessResponseNotes : TFhirProcessResponseNotes;
    {@member newProcessResponse
      create a new ProcessResponse
    }
    {!script nolink}
    function newProcessResponse : TFhirProcessResponse;
    {@member newProvenanceAgent
      create a new agent
    }
    {!script nolink}
    function newProvenanceAgent : TFhirProvenanceAgent;
    {@member newProvenanceAgentRelatedAgent
      create a new relatedAgent
    }
    {!script nolink}
    function newProvenanceAgentRelatedAgent : TFhirProvenanceAgentRelatedAgent;
    {@member newProvenanceEntity
      create a new entity
    }
    {!script nolink}
    function newProvenanceEntity : TFhirProvenanceEntity;
    {@member newProvenance
      create a new Provenance
    }
    {!script nolink}
    function newProvenance : TFhirProvenance;
    {@member newQuestionnaireGroup
      create a new group
    }
    {!script nolink}
    function newQuestionnaireGroup : TFhirQuestionnaireGroup;
    {@member newQuestionnaireGroupQuestion
      create a new question
    }
    {!script nolink}
    function newQuestionnaireGroupQuestion : TFhirQuestionnaireGroupQuestion;
    {@member newQuestionnaire
      create a new Questionnaire
    }
    {!script nolink}
    function newQuestionnaire : TFhirQuestionnaire;
    {@member newQuestionnaireResponseGroup
      create a new group
    }
    {!script nolink}
    function newQuestionnaireResponseGroup : TFhirQuestionnaireResponseGroup;
    {@member newQuestionnaireResponseGroupQuestion
      create a new question
    }
    {!script nolink}
    function newQuestionnaireResponseGroupQuestion : TFhirQuestionnaireResponseGroupQuestion;
    {@member newQuestionnaireResponseGroupQuestionAnswer
      create a new answer
    }
    {!script nolink}
    function newQuestionnaireResponseGroupQuestionAnswer : TFhirQuestionnaireResponseGroupQuestionAnswer;
    {@member newQuestionnaireResponse
      create a new QuestionnaireResponse
    }
    {!script nolink}
    function newQuestionnaireResponse : TFhirQuestionnaireResponse;
    {@member newReferralRequest
      create a new ReferralRequest
    }
    {!script nolink}
    function newReferralRequest : TFhirReferralRequest;
    {@member newRelatedPerson
      create a new RelatedPerson
    }
    {!script nolink}
    function newRelatedPerson : TFhirRelatedPerson;
    {@member newRiskAssessmentPrediction
      create a new prediction
    }
    {!script nolink}
    function newRiskAssessmentPrediction : TFhirRiskAssessmentPrediction;
    {@member newRiskAssessment
      create a new RiskAssessment
    }
    {!script nolink}
    function newRiskAssessment : TFhirRiskAssessment;
    {@member newSchedule
      create a new Schedule
    }
    {!script nolink}
    function newSchedule : TFhirSchedule;
    {@member newSearchParameterContact
      create a new contact
    }
    {!script nolink}
    function newSearchParameterContact : TFhirSearchParameterContact;
    {@member newSearchParameter
      create a new SearchParameter
    }
    {!script nolink}
    function newSearchParameter : TFhirSearchParameter;
    {@member newSlot
      create a new Slot
    }
    {!script nolink}
    function newSlot : TFhirSlot;
    {@member newSpecimenCollection
      create a new collection
    }
    {!script nolink}
    function newSpecimenCollection : TFhirSpecimenCollection;
    {@member newSpecimenTreatment
      create a new treatment
    }
    {!script nolink}
    function newSpecimenTreatment : TFhirSpecimenTreatment;
    {@member newSpecimenContainer
      create a new container
    }
    {!script nolink}
    function newSpecimenContainer : TFhirSpecimenContainer;
    {@member newSpecimen
      create a new Specimen
    }
    {!script nolink}
    function newSpecimen : TFhirSpecimen;
    {@member newStructureDefinitionContact
      create a new contact
    }
    {!script nolink}
    function newStructureDefinitionContact : TFhirStructureDefinitionContact;
    {@member newStructureDefinitionMapping
      create a new mapping
    }
    {!script nolink}
    function newStructureDefinitionMapping : TFhirStructureDefinitionMapping;
    {@member newStructureDefinitionSnapshot
      create a new snapshot
    }
    {!script nolink}
    function newStructureDefinitionSnapshot : TFhirStructureDefinitionSnapshot;
    {@member newStructureDefinitionDifferential
      create a new differential
    }
    {!script nolink}
    function newStructureDefinitionDifferential : TFhirStructureDefinitionDifferential;
    {@member newStructureDefinition
      create a new StructureDefinition
    }
    {!script nolink}
    function newStructureDefinition : TFhirStructureDefinition;
    {@member newSubscriptionChannel
      create a new channel
    }
    {!script nolink}
    function newSubscriptionChannel : TFhirSubscriptionChannel;
    {@member newSubscription
      create a new Subscription
    }
    {!script nolink}
    function newSubscription : TFhirSubscription;
    {@member newSubstanceInstance
      create a new instance
    }
    {!script nolink}
    function newSubstanceInstance : TFhirSubstanceInstance;
    {@member newSubstanceIngredient
      create a new ingredient
    }
    {!script nolink}
    function newSubstanceIngredient : TFhirSubstanceIngredient;
    {@member newSubstance
      create a new Substance
    }
    {!script nolink}
    function newSubstance : TFhirSubstance;
    {@member newSupplyDelivery
      create a new SupplyDelivery
    }
    {!script nolink}
    function newSupplyDelivery : TFhirSupplyDelivery;
    {@member newSupplyRequestWhen
      create a new when
    }
    {!script nolink}
    function newSupplyRequestWhen : TFhirSupplyRequestWhen;
    {@member newSupplyRequest
      create a new SupplyRequest
    }
    {!script nolink}
    function newSupplyRequest : TFhirSupplyRequest;
    {@member newTestScriptContact
      create a new contact
    }
    {!script nolink}
    function newTestScriptContact : TFhirTestScriptContact;
    {@member newTestScriptMetadata
      create a new metadata
    }
    {!script nolink}
    function newTestScriptMetadata : TFhirTestScriptMetadata;
    {@member newTestScriptMetadataLink
      create a new link
    }
    {!script nolink}
    function newTestScriptMetadataLink : TFhirTestScriptMetadataLink;
    {@member newTestScriptMetadataCapability
      create a new capability
    }
    {!script nolink}
    function newTestScriptMetadataCapability : TFhirTestScriptMetadataCapability;
    {@member newTestScriptFixture
      create a new fixture
    }
    {!script nolink}
    function newTestScriptFixture : TFhirTestScriptFixture;
    {@member newTestScriptVariable
      create a new variable
    }
    {!script nolink}
    function newTestScriptVariable : TFhirTestScriptVariable;
    {@member newTestScriptSetup
      create a new setup
    }
    {!script nolink}
    function newTestScriptSetup : TFhirTestScriptSetup;
    {@member newTestScriptSetupAction
      create a new action
    }
    {!script nolink}
    function newTestScriptSetupAction : TFhirTestScriptSetupAction;
    {@member newTestScriptSetupActionOperation
      create a new operation
    }
    {!script nolink}
    function newTestScriptSetupActionOperation : TFhirTestScriptSetupActionOperation;
    {@member newTestScriptSetupActionOperationRequestHeader
      create a new requestHeader
    }
    {!script nolink}
    function newTestScriptSetupActionOperationRequestHeader : TFhirTestScriptSetupActionOperationRequestHeader;
    {@member newTestScriptSetupActionAssert
      create a new assert
    }
    {!script nolink}
    function newTestScriptSetupActionAssert : TFhirTestScriptSetupActionAssert;
    {@member newTestScriptTest
      create a new test
    }
    {!script nolink}
    function newTestScriptTest : TFhirTestScriptTest;
    {@member newTestScriptTestAction
      create a new action
    }
    {!script nolink}
    function newTestScriptTestAction : TFhirTestScriptTestAction;
    {@member newTestScriptTeardown
      create a new teardown
    }
    {!script nolink}
    function newTestScriptTeardown : TFhirTestScriptTeardown;
    {@member newTestScriptTeardownAction
      create a new action
    }
    {!script nolink}
    function newTestScriptTeardownAction : TFhirTestScriptTeardownAction;
    {@member newTestScript
      create a new TestScript
    }
    {!script nolink}
    function newTestScript : TFhirTestScript;
    {@member newValueSetContact
      create a new contact
    }
    {!script nolink}
    function newValueSetContact : TFhirValueSetContact;
    {@member newValueSetCodeSystem
      create a new codeSystem
    }
    {!script nolink}
    function newValueSetCodeSystem : TFhirValueSetCodeSystem;
    {@member newValueSetCodeSystemConcept
      create a new concept
    }
    {!script nolink}
    function newValueSetCodeSystemConcept : TFhirValueSetCodeSystemConcept;
    {@member newValueSetCodeSystemConceptDesignation
      create a new designation
    }
    {!script nolink}
    function newValueSetCodeSystemConceptDesignation : TFhirValueSetCodeSystemConceptDesignation;
    {@member newValueSetCompose
      create a new compose
    }
    {!script nolink}
    function newValueSetCompose : TFhirValueSetCompose;
    {@member newValueSetComposeInclude
      create a new include
    }
    {!script nolink}
    function newValueSetComposeInclude : TFhirValueSetComposeInclude;
    {@member newValueSetComposeIncludeConcept
      create a new concept
    }
    {!script nolink}
    function newValueSetComposeIncludeConcept : TFhirValueSetComposeIncludeConcept;
    {@member newValueSetComposeIncludeFilter
      create a new filter
    }
    {!script nolink}
    function newValueSetComposeIncludeFilter : TFhirValueSetComposeIncludeFilter;
    {@member newValueSetExpansion
      create a new expansion
    }
    {!script nolink}
    function newValueSetExpansion : TFhirValueSetExpansion;
    {@member newValueSetExpansionParameter
      create a new parameter
    }
    {!script nolink}
    function newValueSetExpansionParameter : TFhirValueSetExpansionParameter;
    {@member newValueSetExpansionContains
      create a new contains
    }
    {!script nolink}
    function newValueSetExpansionContains : TFhirValueSetExpansionContains;
    {@member newValueSet
      create a new ValueSet
    }
    {!script nolink}
    function newValueSet : TFhirValueSet;
    {@member newVisionPrescriptionDispense
      create a new dispense
    }
    {!script nolink}
    function newVisionPrescriptionDispense : TFhirVisionPrescriptionDispense;
    {@member newVisionPrescription
      create a new VisionPrescription
    }
    {!script nolink}
    function newVisionPrescription : TFhirVisionPrescription;
    function makeByName(const name : String) : TFHIRBase;
  end;

implementation

uses
  FHIRUtilities;

{ TFhirResource }

constructor TFhirResource.Create;
begin
  inherited;
end;

destructor TFhirResource.Destroy;
begin
  FId.free;
  FMeta.free;
  FImplicitRules.free;
  FLanguage.free;
  inherited;
end;

procedure TFhirResource.Assign(oSource : TAdvObject);
begin
  inherited;
  idElement := TFhirResource(oSource).idElement.Clone;
  meta := TFhirResource(oSource).meta.Clone;
  implicitRulesElement := TFhirResource(oSource).implicitRulesElement.Clone;
  languageElement := TFhirResource(oSource).languageElement.Clone;
end;

procedure TFhirResource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'id') Then
     list.add(FId.Link);
  if (child_name = 'meta') Then
     list.add(FMeta.Link);
  if (child_name = 'implicitRules') Then
     list.add(FImplicitRules.Link);
  if (child_name = 'language') Then
     list.add(FLanguage.Link);
end;

procedure TFhirResource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'id', 'id', FId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'meta', 'Meta', FMeta.Link));{2}
  oList.add(TFHIRProperty.create(self, 'implicitRules', 'uri', FImplicitRules.Link));{2}
  oList.add(TFHIRProperty.create(self, 'language', 'code', FLanguage.Link));{2}
end;

procedure TFhirResource.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'id') then IdElement := propValue as TFhirId{5a}
  else if (propName = 'meta') then Meta := propValue as TFhirMeta{4b}
  else if (propName = 'implicitRules') then ImplicitRulesElement := propValue as TFhirUri{5a}
  else if (propName = 'language') then
    if propValue is TFHIRCode then
      LanguageElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      LanguageElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else inherited;
end;

function TFhirResource.FhirType : string;
begin
  result := 'Resource';
end;

function TFhirResource.Link : TFhirResource;
begin
  result := TFhirResource(inherited Link);
end;

function TFhirResource.Clone : TFhirResource;
begin
  result := TFhirResource(inherited Clone);
end;

{ TFhirResource }

Procedure TFhirResource.SetId(value : TFhirId);
begin
  FId.free;
  FId := value;
end;

Function TFhirResource.GetIdST : String;
begin
  if FId = nil then
    result := ''
  else
    result := FId.value;
end;

Procedure TFhirResource.SetIdST(value : String);
begin
  if value <> '' then
  begin
    if FId = nil then
      FId := TFhirId.create;
    FId.value := value
  end
  else if FId <> nil then
    FId.value := '';
end;

Procedure TFhirResource.SetMeta(value : TFhirMeta);
begin
  FMeta.free;
  FMeta := value;
end;

Procedure TFhirResource.SetImplicitRules(value : TFhirUri);
begin
  FImplicitRules.free;
  FImplicitRules := value;
end;

Function TFhirResource.GetImplicitRulesST : String;
begin
  if FImplicitRules = nil then
    result := ''
  else
    result := FImplicitRules.value;
end;

Procedure TFhirResource.SetImplicitRulesST(value : String);
begin
  if value <> '' then
  begin
    if FImplicitRules = nil then
      FImplicitRules := TFhirUri.create;
    FImplicitRules.value := value
  end
  else if FImplicitRules <> nil then
    FImplicitRules.value := '';
end;

Procedure TFhirResource.SetLanguage(value : TFhirCode);
begin
  FLanguage.free;
  FLanguage := value;
end;

Function TFhirResource.GetLanguageST : String;
begin
  if FLanguage = nil then
    result := ''
  else
    result := FLanguage.value;
end;

Procedure TFhirResource.SetLanguageST(value : String);
begin
  if value <> '' then
  begin
    if FLanguage = nil then
      FLanguage := TFhirCode.create;
    FLanguage.value := value
  end
  else if FLanguage <> nil then
    FLanguage.value := '';
end;


{ TFhirResourceListEnumerator }

Constructor TFhirResourceListEnumerator.Create(list : TFhirResourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirResourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirResourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirResourceListEnumerator.GetCurrent : TFhirResource;
begin
  Result := FList[FIndex];
end;


{ TFhirResourceList }
procedure TFhirResourceList.AddItem(value: TFhirResource);
begin
  assert(value.ClassName = 'TFhirResource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirResource');
  add(value);
end;


procedure TFhirResourceList.ClearItems;
begin
  Clear;
end;

function TFhirResourceList.GetEnumerator : TFhirResourceListEnumerator;
begin
  result := TFhirResourceListEnumerator.Create(self.link);
end;

function TFhirResourceList.Clone: TFhirResourceList;
begin
  result := TFhirResourceList(inherited Clone);
end;

function TFhirResourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirResourceList.GetItemN(index: Integer): TFhirResource;
begin
  result := TFhirResource(ObjectByIndex[index]);
end;

function TFhirResourceList.IndexOf(value: TFhirResource): Integer;
begin
  result := IndexByReference(value);
end;


procedure TFhirResourceList.InsertItem(index: Integer; value: TFhirResource);
begin
  assert(value is TFhirResource);
  Inherited Insert(index, value);
end;

function TFhirResourceList.Item(index: Integer): TFhirResource;
begin
  result := TFhirResource(ObjectByIndex[index]);
end;

function TFhirResourceList.Link: TFhirResourceList;
begin
  result := TFhirResourceList(inherited Link);
end;

procedure TFhirResourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirResourceList.SetItemByIndex(index: Integer; value: TFhirResource);
begin
  assert(value is TFhirResource);
  FhirResources[index] := value;
end;

procedure TFhirResourceList.SetItemN(index: Integer; value: TFhirResource);
begin
  assert(value is TFhirResource);
  ObjectByIndex[index] := value;
end;

{ TFhirDomainResource }

constructor TFhirDomainResource.Create;
begin
  inherited;
end;

destructor TFhirDomainResource.Destroy;
begin
  FText.free;
  FContainedList.Free;
  FExtensionList.Free;
  FModifierExtensionList.Free;
  inherited;
end;

procedure TFhirDomainResource.Assign(oSource : TAdvObject);
begin
  inherited;
  text := TFhirDomainResource(oSource).text.Clone;
  if (TFhirDomainResource(oSource).FContainedList = nil) then
  begin
    FContainedList.free;
    FContainedList := nil;
  end
  else
  begin
    FContainedList := TFhirResourceList.Create;
    FContainedList.Assign(TFhirDomainResource(oSource).FContainedList);
  end;
  if (TFhirDomainResource(oSource).FExtensionList = nil) then
  begin
    FExtensionList.free;
    FExtensionList := nil;
  end
  else
  begin
    FExtensionList := TFhirExtensionList.Create;
    FExtensionList.Assign(TFhirDomainResource(oSource).FExtensionList);
  end;
  if (TFhirDomainResource(oSource).FModifierExtensionList = nil) then
  begin
    FModifierExtensionList.free;
    FModifierExtensionList := nil;
  end
  else
  begin
    FModifierExtensionList := TFhirExtensionList.Create;
    FModifierExtensionList.Assign(TFhirDomainResource(oSource).FModifierExtensionList);
  end;
end;

procedure TFhirDomainResource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'contained') Then
    list.addAll(FContainedList);
  if (child_name = 'extension') Then
    list.addAll(FExtensionList);
  if (child_name = 'modifierExtension') Then
    list.addAll(FModifierExtensionList);
end;

procedure TFhirDomainResource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'text', 'Narrative', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contained', 'Resource', FContainedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'extension', 'Extension', FExtensionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'modifierExtension', 'Extension', FModifierExtensionList.Link)){3};
end;

procedure TFhirDomainResource.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'text') then Text := propValue as TFhirNarrative{4b}
  else if (propName = 'contained') then ContainedList.add(propValue as TFhirResource){2}
  else if (propName = 'extension') then ExtensionList.add(propValue as TFhirExtension){2}
  else if (propName = 'modifierExtension') then ModifierExtensionList.add(propValue as TFhirExtension){2}
  else inherited;
end;

function TFhirDomainResource.FhirType : string;
begin
  result := 'DomainResource';
end;

function TFhirDomainResource.Link : TFhirDomainResource;
begin
  result := TFhirDomainResource(inherited Link);
end;

function TFhirDomainResource.Clone : TFhirDomainResource;
begin
  result := TFhirDomainResource(inherited Clone);
end;

{ TFhirDomainResource }

Procedure TFhirDomainResource.SetText(value : TFhirNarrative);
begin
  FText.free;
  FText := value;
end;

Function TFhirDomainResource.GetContainedList : TFhirResourceList;
begin
  if FContainedList = nil then
    FContainedList := TFhirResourceList.Create;
  result := FContainedList;
end;

Function TFhirDomainResource.GetHasContainedList : boolean;
begin
  result := (FContainedList <> nil) and (FContainedList.count > 0);
end;

Function TFhirDomainResource.GetExtensionList : TFhirExtensionList;
begin
  if FExtensionList = nil then
    FExtensionList := TFhirExtensionList.Create;
  result := FExtensionList;
end;

Function TFhirDomainResource.GetHasExtensionList : boolean;
begin
  result := (FExtensionList <> nil) and (FExtensionList.count > 0);
end;

Function TFhirDomainResource.GetModifierExtensionList : TFhirExtensionList;
begin
  if FModifierExtensionList = nil then
    FModifierExtensionList := TFhirExtensionList.Create;
  result := FModifierExtensionList;
end;

Function TFhirDomainResource.GetHasModifierExtensionList : boolean;
begin
  result := (FModifierExtensionList <> nil) and (FModifierExtensionList.count > 0);
end;


{ TFhirDomainResourceListEnumerator }

Constructor TFhirDomainResourceListEnumerator.Create(list : TFhirDomainResourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDomainResourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDomainResourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDomainResourceListEnumerator.GetCurrent : TFhirDomainResource;
begin
  Result := FList[FIndex];
end;


{ TFhirDomainResourceList }
procedure TFhirDomainResourceList.AddItem(value: TFhirDomainResource);
begin
  assert(value.ClassName = 'TFhirDomainResource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDomainResource');
  add(value);
end;


procedure TFhirDomainResourceList.ClearItems;
begin
  Clear;
end;

function TFhirDomainResourceList.GetEnumerator : TFhirDomainResourceListEnumerator;
begin
  result := TFhirDomainResourceListEnumerator.Create(self.link);
end;

function TFhirDomainResourceList.Clone: TFhirDomainResourceList;
begin
  result := TFhirDomainResourceList(inherited Clone);
end;

function TFhirDomainResourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDomainResourceList.GetItemN(index: Integer): TFhirDomainResource;
begin
  result := TFhirDomainResource(ObjectByIndex[index]);
end;

function TFhirDomainResourceList.IndexOf(value: TFhirDomainResource): Integer;
begin
  result := IndexByReference(value);
end;


procedure TFhirDomainResourceList.InsertItem(index: Integer; value: TFhirDomainResource);
begin
  assert(value is TFhirDomainResource);
  Inherited Insert(index, value);
end;

function TFhirDomainResourceList.Item(index: Integer): TFhirDomainResource;
begin
  result := TFhirDomainResource(ObjectByIndex[index]);
end;

function TFhirDomainResourceList.Link: TFhirDomainResourceList;
begin
  result := TFhirDomainResourceList(inherited Link);
end;

procedure TFhirDomainResourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDomainResourceList.SetItemByIndex(index: Integer; value: TFhirDomainResource);
begin
  assert(value is TFhirDomainResource);
  FhirDomainResources[index] := value;
end;

procedure TFhirDomainResourceList.SetItemN(index: Integer; value: TFhirDomainResource);
begin
  assert(value is TFhirDomainResource);
  ObjectByIndex[index] := value;
end;

{ TFhirParametersParameter }

constructor TFhirParametersParameter.Create;
begin
  inherited;
end;

destructor TFhirParametersParameter.Destroy;
begin
  FName.free;
  FValue.free;
  FResource.free;
  FPartList.Free;
  inherited;
end;

procedure TFhirParametersParameter.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirParametersParameter(oSource).nameElement.Clone;
  value := TFhirParametersParameter(oSource).value.Clone;
  resource := TFhirParametersParameter(oSource).resource.Clone;
  if (TFhirParametersParameter(oSource).FPartList = nil) then
  begin
    FPartList.free;
    FPartList := nil;
  end
  else
  begin
    FPartList := TFhirParametersParameterList.Create;
    FPartList.Assign(TFhirParametersParameter(oSource).FPartList);
  end;
end;

procedure TFhirParametersParameter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
  if (child_name = 'resource') Then
     list.add(FResource.Link);
  if (child_name = 'part') Then
    list.addAll(FPartList);
end;

procedure TFhirParametersParameter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value[x]', '*', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', 'Resource', FResource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'part', '@Parameters.parameter', FPartList.Link)){3};
end;

procedure TFhirParametersParameter.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else if (propName = 'resource') then Resource := propValue as TFhirResource{4b}
  else if (propName = 'part') then PartList.add(propValue as TFhirParametersParameter){2}
  else inherited;
end;

function TFhirParametersParameter.FhirType : string;
begin
  result := 'parameter';
end;

function TFhirParametersParameter.Link : TFhirParametersParameter;
begin
  result := TFhirParametersParameter(inherited Link);
end;

function TFhirParametersParameter.Clone : TFhirParametersParameter;
begin
  result := TFhirParametersParameter(inherited Clone);
end;

function TFhirParametersParameter.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirParametersParameter;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirParametersParameter)) then
    result := false
  else
  begin
    o := TFhirParametersParameter(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(valueElement, o.valueElement, true) and 
      compareDeep(resourceElement, o.resourceElement, true) and compareDeep(partList, o.partList, true);
  end;
end;

function TFhirParametersParameter.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirParametersParameter;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirParametersParameter)) then
    result := false
  else
  begin
    o := TFhirParametersParameter(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirParametersParameter }

Procedure TFhirParametersParameter.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirParametersParameter.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirParametersParameter.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirParametersParameter.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;

Procedure TFhirParametersParameter.SetResource(value : TFhirResource);
begin
  FResource.free;
  FResource := value;
end;

Function TFhirParametersParameter.GetPartList : TFhirParametersParameterList;
begin
  if FPartList = nil then
    FPartList := TFhirParametersParameterList.Create;
  result := FPartList;
end;

Function TFhirParametersParameter.GetHasPartList : boolean;
begin
  result := (FPartList <> nil) and (FPartList.count > 0);
end;


{ TFhirParametersParameterListEnumerator }

Constructor TFhirParametersParameterListEnumerator.Create(list : TFhirParametersParameterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirParametersParameterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirParametersParameterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirParametersParameterListEnumerator.GetCurrent : TFhirParametersParameter;
begin
  Result := FList[FIndex];
end;


{ TFhirParametersParameterList }
procedure TFhirParametersParameterList.AddItem(value: TFhirParametersParameter);
begin
  assert(value.ClassName = 'TFhirParametersParameter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirParametersParameter');
  add(value);
end;


function TFhirParametersParameterList.Append: TFhirParametersParameter;
begin
  result := TFhirParametersParameter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirParametersParameterList.ClearItems;
begin
  Clear;
end;

function TFhirParametersParameterList.GetEnumerator : TFhirParametersParameterListEnumerator;
begin
  result := TFhirParametersParameterListEnumerator.Create(self.link);
end;

function TFhirParametersParameterList.Clone: TFhirParametersParameterList;
begin
  result := TFhirParametersParameterList(inherited Clone);
end;

function TFhirParametersParameterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirParametersParameterList.GetItemN(index: Integer): TFhirParametersParameter;
begin
  result := TFhirParametersParameter(ObjectByIndex[index]);
end;

function TFhirParametersParameterList.IndexOf(value: TFhirParametersParameter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirParametersParameterList.Insert(index: Integer): TFhirParametersParameter;
begin
  result := TFhirParametersParameter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirParametersParameterList.InsertItem(index: Integer; value: TFhirParametersParameter);
begin
  assert(value is TFhirParametersParameter);
  Inherited Insert(index, value);
end;

function TFhirParametersParameterList.Item(index: Integer): TFhirParametersParameter;
begin
  result := TFhirParametersParameter(ObjectByIndex[index]);
end;

function TFhirParametersParameterList.Link: TFhirParametersParameterList;
begin
  result := TFhirParametersParameterList(inherited Link);
end;

procedure TFhirParametersParameterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirParametersParameterList.SetItemByIndex(index: Integer; value: TFhirParametersParameter);
begin
  assert(value is TFhirParametersParameter);
  FhirParametersParameters[index] := value;
end;

procedure TFhirParametersParameterList.SetItemN(index: Integer; value: TFhirParametersParameter);
begin
  assert(value is TFhirParametersParameter);
  ObjectByIndex[index] := value;
end;

{ TFhirParameters }

constructor TFhirParameters.Create;
begin
  inherited;
end;

destructor TFhirParameters.Destroy;
begin
  FParameterList.Free;
  inherited;
end;

function TFhirParameters.GetResourceType : TFhirResourceType;
begin
  result := frtParameters;
end;

procedure TFhirParameters.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirParameters(oSource).FParameterList = nil) then
  begin
    FParameterList.free;
    FParameterList := nil;
  end
  else
  begin
    FParameterList := TFhirParametersParameterList.Create;
    FParameterList.Assign(TFhirParameters(oSource).FParameterList);
  end;
end;

procedure TFhirParameters.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'parameter') Then
    list.addAll(FParameterList);
end;

procedure TFhirParameters.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'parameter', '', FParameterList.Link)){3};
end;

procedure TFhirParameters.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'parameter') then ParameterList.add(propValue as TFhirParametersParameter){2}
  else inherited;
end;

function TFhirParameters.FhirType : string;
begin
  result := 'Parameters';
end;

function TFhirParameters.Link : TFhirParameters;
begin
  result := TFhirParameters(inherited Link);
end;

function TFhirParameters.Clone : TFhirParameters;
begin
  result := TFhirParameters(inherited Clone);
end;

{ TFhirParameters }

Function TFhirParameters.GetParameterList : TFhirParametersParameterList;
begin
  if FParameterList = nil then
    FParameterList := TFhirParametersParameterList.Create;
  result := FParameterList;
end;

Function TFhirParameters.GetHasParameterList : boolean;
begin
  result := (FParameterList <> nil) and (FParameterList.count > 0);
end;


{ TFhirParametersListEnumerator }

Constructor TFhirParametersListEnumerator.Create(list : TFhirParametersList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirParametersListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirParametersListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirParametersListEnumerator.GetCurrent : TFhirParameters;
begin
  Result := FList[FIndex];
end;


{ TFhirParametersList }
procedure TFhirParametersList.AddItem(value: TFhirParameters);
begin
  assert(value.ClassName = 'TFhirParameters', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirParameters');
  add(value);
end;


function TFhirParametersList.Append: TFhirParameters;
begin
  result := TFhirParameters.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirParametersList.ClearItems;
begin
  Clear;
end;

function TFhirParametersList.GetEnumerator : TFhirParametersListEnumerator;
begin
  result := TFhirParametersListEnumerator.Create(self.link);
end;

function TFhirParametersList.Clone: TFhirParametersList;
begin
  result := TFhirParametersList(inherited Clone);
end;

function TFhirParametersList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirParametersList.GetItemN(index: Integer): TFhirParameters;
begin
  result := TFhirParameters(ObjectByIndex[index]);
end;

function TFhirParametersList.IndexOf(value: TFhirParameters): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirParametersList.Insert(index: Integer): TFhirParameters;
begin
  result := TFhirParameters.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirParametersList.InsertItem(index: Integer; value: TFhirParameters);
begin
  assert(value is TFhirParameters);
  Inherited Insert(index, value);
end;

function TFhirParametersList.Item(index: Integer): TFhirParameters;
begin
  result := TFhirParameters(ObjectByIndex[index]);
end;

function TFhirParametersList.Link: TFhirParametersList;
begin
  result := TFhirParametersList(inherited Link);
end;

procedure TFhirParametersList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirParametersList.SetItemByIndex(index: Integer; value: TFhirParameters);
begin
  assert(value is TFhirParameters);
  FhirParameters[index] := value;
end;

procedure TFhirParametersList.SetItemN(index: Integer; value: TFhirParameters);
begin
  assert(value is TFhirParameters);
  ObjectByIndex[index] := value;
end;

{ TFhirAccount }

constructor TFhirAccount.Create;
begin
  inherited;
end;

destructor TFhirAccount.Destroy;
begin
  FIdentifierList.Free;
  FName.free;
  FType_.free;
  FStatus.free;
  FActivePeriod.free;
  FCurrency.free;
  FBalance.free;
  FCoveragePeriod.free;
  FSubject.free;
  FOwner.free;
  FDescription.free;
  inherited;
end;

function TFhirAccount.GetResourceType : TFhirResourceType;
begin
  result := frtAccount;
end;

procedure TFhirAccount.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAccount(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirAccount(oSource).FIdentifierList);
  end;
  nameElement := TFhirAccount(oSource).nameElement.Clone;
  type_ := TFhirAccount(oSource).type_.Clone;
  FStatus := TFhirAccount(oSource).FStatus.Link;
  activePeriod := TFhirAccount(oSource).activePeriod.Clone;
  currency := TFhirAccount(oSource).currency.Clone;
  balance := TFhirAccount(oSource).balance.Clone;
  coveragePeriod := TFhirAccount(oSource).coveragePeriod.Clone;
  subject := TFhirAccount(oSource).subject.Clone;
  owner := TFhirAccount(oSource).owner.Clone;
  descriptionElement := TFhirAccount(oSource).descriptionElement.Clone;
end;

procedure TFhirAccount.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'activePeriod') Then
     list.add(FActivePeriod.Link);
  if (child_name = 'currency') Then
     list.add(FCurrency.Link);
  if (child_name = 'balance') Then
     list.add(FBalance.Link);
  if (child_name = 'coveragePeriod') Then
     list.add(FCoveragePeriod.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'owner') Then
     list.add(FOwner.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
end;

procedure TFhirAccount.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'activePeriod', 'Period', FActivePeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'currency', 'Coding', FCurrency.Link));{2}
  oList.add(TFHIRProperty.create(self, 'balance', 'Money', FBalance.Link));{2}
  oList.add(TFHIRProperty.create(self, 'coveragePeriod', 'Period', FCoveragePeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Device|Practitioner|Location|HealthcareService|Organization)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'owner', 'Reference(Organization)', FOwner.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
end;

procedure TFhirAccount.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'activePeriod') then ActivePeriod := propValue as TFhirPeriod{4b}
  else if (propName = 'currency') then Currency := propValue as TFhirCoding{4b}
  else if (propName = 'balance') then Balance := propValue as TFhirQuantity{4b}
  else if (propName = 'coveragePeriod') then CoveragePeriod := propValue as TFhirPeriod{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'owner') then Owner := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirAccount.FhirType : string;
begin
  result := 'Account';
end;

function TFhirAccount.Link : TFhirAccount;
begin
  result := TFhirAccount(inherited Link);
end;

function TFhirAccount.Clone : TFhirAccount;
begin
  result := TFhirAccount(inherited Clone);
end;

{ TFhirAccount }

Function TFhirAccount.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirAccount.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirAccount.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirAccount.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirAccount.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirAccount.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirAccount.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirAccount.GetStatusST : TFhirAccountStatus;
begin
  if FStatus = nil then
    result := TFhirAccountStatus(0)
  else
    result := TFhirAccountStatus(StringArrayIndexOfSensitive(CODES_TFhirAccountStatus, FStatus.value));
end;

Procedure TFhirAccount.SetStatusST(value : TFhirAccountStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirAccountStatus[value]);
end;

Procedure TFhirAccount.SetActivePeriod(value : TFhirPeriod);
begin
  FActivePeriod.free;
  FActivePeriod := value;
end;

Procedure TFhirAccount.SetCurrency(value : TFhirCoding);
begin
  FCurrency.free;
  FCurrency := value;
end;

Procedure TFhirAccount.SetBalance(value : TFhirQuantity);
begin
  FBalance.free;
  FBalance := value;
end;

Procedure TFhirAccount.SetCoveragePeriod(value : TFhirPeriod);
begin
  FCoveragePeriod.free;
  FCoveragePeriod := value;
end;

Procedure TFhirAccount.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirAccount.SetOwner(value : TFhirReference{TFhirOrganization});
begin
  FOwner.free;
  FOwner := value;
end;

Procedure TFhirAccount.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirAccount.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirAccount.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;


{ TFhirAccountListEnumerator }

Constructor TFhirAccountListEnumerator.Create(list : TFhirAccountList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAccountListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAccountListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAccountListEnumerator.GetCurrent : TFhirAccount;
begin
  Result := FList[FIndex];
end;


{ TFhirAccountList }
procedure TFhirAccountList.AddItem(value: TFhirAccount);
begin
  assert(value.ClassName = 'TFhirAccount', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAccount');
  add(value);
end;


function TFhirAccountList.Append: TFhirAccount;
begin
  result := TFhirAccount.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAccountList.ClearItems;
begin
  Clear;
end;

function TFhirAccountList.GetEnumerator : TFhirAccountListEnumerator;
begin
  result := TFhirAccountListEnumerator.Create(self.link);
end;

function TFhirAccountList.Clone: TFhirAccountList;
begin
  result := TFhirAccountList(inherited Clone);
end;

function TFhirAccountList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAccountList.GetItemN(index: Integer): TFhirAccount;
begin
  result := TFhirAccount(ObjectByIndex[index]);
end;

function TFhirAccountList.IndexOf(value: TFhirAccount): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAccountList.Insert(index: Integer): TFhirAccount;
begin
  result := TFhirAccount.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAccountList.InsertItem(index: Integer; value: TFhirAccount);
begin
  assert(value is TFhirAccount);
  Inherited Insert(index, value);
end;

function TFhirAccountList.Item(index: Integer): TFhirAccount;
begin
  result := TFhirAccount(ObjectByIndex[index]);
end;

function TFhirAccountList.Link: TFhirAccountList;
begin
  result := TFhirAccountList(inherited Link);
end;

procedure TFhirAccountList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAccountList.SetItemByIndex(index: Integer; value: TFhirAccount);
begin
  assert(value is TFhirAccount);
  FhirAccounts[index] := value;
end;

procedure TFhirAccountList.SetItemN(index: Integer; value: TFhirAccount);
begin
  assert(value is TFhirAccount);
  ObjectByIndex[index] := value;
end;

{ TFhirAllergyIntoleranceReaction }

constructor TFhirAllergyIntoleranceReaction.Create;
begin
  inherited;
end;

destructor TFhirAllergyIntoleranceReaction.Destroy;
begin
  FSubstance.free;
  FCertainty.free;
  FManifestationList.Free;
  FDescription.free;
  FOnset.free;
  FSeverity.free;
  FExposureRoute.free;
  FNote.free;
  inherited;
end;

procedure TFhirAllergyIntoleranceReaction.Assign(oSource : TAdvObject);
begin
  inherited;
  substance := TFhirAllergyIntoleranceReaction(oSource).substance.Clone;
  FCertainty := TFhirAllergyIntoleranceReaction(oSource).FCertainty.Link;
  if (TFhirAllergyIntoleranceReaction(oSource).FManifestationList = nil) then
  begin
    FManifestationList.free;
    FManifestationList := nil;
  end
  else
  begin
    FManifestationList := TFhirCodeableConceptList.Create;
    FManifestationList.Assign(TFhirAllergyIntoleranceReaction(oSource).FManifestationList);
  end;
  descriptionElement := TFhirAllergyIntoleranceReaction(oSource).descriptionElement.Clone;
  onsetElement := TFhirAllergyIntoleranceReaction(oSource).onsetElement.Clone;
  FSeverity := TFhirAllergyIntoleranceReaction(oSource).FSeverity.Link;
  exposureRoute := TFhirAllergyIntoleranceReaction(oSource).exposureRoute.Clone;
  note := TFhirAllergyIntoleranceReaction(oSource).note.Clone;
end;

procedure TFhirAllergyIntoleranceReaction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'substance') Then
     list.add(FSubstance.Link);
  if (child_name = 'certainty') Then
     list.add(FCertainty.Link);
  if (child_name = 'manifestation') Then
    list.addAll(FManifestationList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'onset') Then
     list.add(FOnset.Link);
  if (child_name = 'severity') Then
     list.add(FSeverity.Link);
  if (child_name = 'exposureRoute') Then
     list.add(FExposureRoute.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
end;

procedure TFhirAllergyIntoleranceReaction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'substance', 'CodeableConcept', FSubstance.Link));{2}
  oList.add(TFHIRProperty.create(self, 'certainty', 'code', FCertainty.Link));{1}
  oList.add(TFHIRProperty.create(self, 'manifestation', 'CodeableConcept', FManifestationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'onset', 'dateTime', FOnset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'severity', 'code', FSeverity.Link));{1}
  oList.add(TFHIRProperty.create(self, 'exposureRoute', 'CodeableConcept', FExposureRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNote.Link));{2}
end;

procedure TFhirAllergyIntoleranceReaction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'substance') then Substance := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'certainty') then CertaintyElement := propValue as TFHIREnum
  else if (propName = 'manifestation') then ManifestationList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'onset') then OnsetElement := propValue as TFhirDateTime{5a}
  else if (propName = 'severity') then SeverityElement := propValue as TFHIREnum
  else if (propName = 'exposureRoute') then ExposureRoute := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'note') then Note := propValue as TFhirAnnotation{4b}
  else inherited;
end;

function TFhirAllergyIntoleranceReaction.FhirType : string;
begin
  result := 'reaction';
end;

function TFhirAllergyIntoleranceReaction.Link : TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction(inherited Link);
end;

function TFhirAllergyIntoleranceReaction.Clone : TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction(inherited Clone);
end;

function TFhirAllergyIntoleranceReaction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAllergyIntoleranceReaction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAllergyIntoleranceReaction)) then
    result := false
  else
  begin
    o := TFhirAllergyIntoleranceReaction(other);
    result := compareDeep(substanceElement, o.substanceElement, true) and compareDeep(certaintyElement, o.certaintyElement, true) and 
      compareDeep(manifestationList, o.manifestationList, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(onsetElement, o.onsetElement, true) and compareDeep(severityElement, o.severityElement, true) and 
      compareDeep(exposureRouteElement, o.exposureRouteElement, true) and compareDeep(noteElement, o.noteElement, true);
  end;
end;

function TFhirAllergyIntoleranceReaction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAllergyIntoleranceReaction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAllergyIntoleranceReaction)) then
    result := false
  else
  begin
    o := TFhirAllergyIntoleranceReaction(other);
    result := compareValues(certaintyElement, o.certaintyElement, true) and compareValues(descriptionElement, o.descriptionElement, true) and 
      compareValues(onsetElement, o.onsetElement, true) and compareValues(severityElement, o.severityElement, true);
  end;
end;

{ TFhirAllergyIntoleranceReaction }

Procedure TFhirAllergyIntoleranceReaction.SetSubstance(value : TFhirCodeableConcept);
begin
  FSubstance.free;
  FSubstance := value;
end;

Procedure TFhirAllergyIntoleranceReaction.SetCertainty(value : TFhirEnum);
begin
  FCertainty.free;
  FCertainty := value;
end;

Function TFhirAllergyIntoleranceReaction.GetCertaintyST : TFhirReactionEventCertainty;
begin
  if FCertainty = nil then
    result := TFhirReactionEventCertainty(0)
  else
    result := TFhirReactionEventCertainty(StringArrayIndexOfSensitive(CODES_TFhirReactionEventCertainty, FCertainty.value));
end;

Procedure TFhirAllergyIntoleranceReaction.SetCertaintyST(value : TFhirReactionEventCertainty);
begin
  if ord(value) = 0 then
    CertaintyElement := nil
  else
    CertaintyElement := TFhirEnum.create(CODES_TFhirReactionEventCertainty[value]);
end;

Function TFhirAllergyIntoleranceReaction.GetManifestationList : TFhirCodeableConceptList;
begin
  if FManifestationList = nil then
    FManifestationList := TFhirCodeableConceptList.Create;
  result := FManifestationList;
end;

Function TFhirAllergyIntoleranceReaction.GetHasManifestationList : boolean;
begin
  result := (FManifestationList <> nil) and (FManifestationList.count > 0);
end;

Procedure TFhirAllergyIntoleranceReaction.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirAllergyIntoleranceReaction.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirAllergyIntoleranceReaction.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirAllergyIntoleranceReaction.SetOnset(value : TFhirDateTime);
begin
  FOnset.free;
  FOnset := value;
end;

Function TFhirAllergyIntoleranceReaction.GetOnsetST : TDateAndTime;
begin
  if FOnset = nil then
    result := nil
  else
    result := FOnset.value;
end;

Procedure TFhirAllergyIntoleranceReaction.SetOnsetST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FOnset = nil then
      FOnset := TFhirDateTime.create;
    FOnset.value := value
  end
  else if FOnset <> nil then
    FOnset.value := nil;
end;

Procedure TFhirAllergyIntoleranceReaction.SetSeverity(value : TFhirEnum);
begin
  FSeverity.free;
  FSeverity := value;
end;

Function TFhirAllergyIntoleranceReaction.GetSeverityST : TFhirReactionEventSeverity;
begin
  if FSeverity = nil then
    result := TFhirReactionEventSeverity(0)
  else
    result := TFhirReactionEventSeverity(StringArrayIndexOfSensitive(CODES_TFhirReactionEventSeverity, FSeverity.value));
end;

Procedure TFhirAllergyIntoleranceReaction.SetSeverityST(value : TFhirReactionEventSeverity);
begin
  if ord(value) = 0 then
    SeverityElement := nil
  else
    SeverityElement := TFhirEnum.create(CODES_TFhirReactionEventSeverity[value]);
end;

Procedure TFhirAllergyIntoleranceReaction.SetExposureRoute(value : TFhirCodeableConcept);
begin
  FExposureRoute.free;
  FExposureRoute := value;
end;

Procedure TFhirAllergyIntoleranceReaction.SetNote(value : TFhirAnnotation);
begin
  FNote.free;
  FNote := value;
end;


{ TFhirAllergyIntoleranceReactionListEnumerator }

Constructor TFhirAllergyIntoleranceReactionListEnumerator.Create(list : TFhirAllergyIntoleranceReactionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAllergyIntoleranceReactionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAllergyIntoleranceReactionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAllergyIntoleranceReactionListEnumerator.GetCurrent : TFhirAllergyIntoleranceReaction;
begin
  Result := FList[FIndex];
end;


{ TFhirAllergyIntoleranceReactionList }
procedure TFhirAllergyIntoleranceReactionList.AddItem(value: TFhirAllergyIntoleranceReaction);
begin
  assert(value.ClassName = 'TFhirAllergyIntoleranceReaction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAllergyIntoleranceReaction');
  add(value);
end;


function TFhirAllergyIntoleranceReactionList.Append: TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAllergyIntoleranceReactionList.ClearItems;
begin
  Clear;
end;

function TFhirAllergyIntoleranceReactionList.GetEnumerator : TFhirAllergyIntoleranceReactionListEnumerator;
begin
  result := TFhirAllergyIntoleranceReactionListEnumerator.Create(self.link);
end;

function TFhirAllergyIntoleranceReactionList.Clone: TFhirAllergyIntoleranceReactionList;
begin
  result := TFhirAllergyIntoleranceReactionList(inherited Clone);
end;

function TFhirAllergyIntoleranceReactionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAllergyIntoleranceReactionList.GetItemN(index: Integer): TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction(ObjectByIndex[index]);
end;

function TFhirAllergyIntoleranceReactionList.IndexOf(value: TFhirAllergyIntoleranceReaction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAllergyIntoleranceReactionList.Insert(index: Integer): TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAllergyIntoleranceReactionList.InsertItem(index: Integer; value: TFhirAllergyIntoleranceReaction);
begin
  assert(value is TFhirAllergyIntoleranceReaction);
  Inherited Insert(index, value);
end;

function TFhirAllergyIntoleranceReactionList.Item(index: Integer): TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction(ObjectByIndex[index]);
end;

function TFhirAllergyIntoleranceReactionList.Link: TFhirAllergyIntoleranceReactionList;
begin
  result := TFhirAllergyIntoleranceReactionList(inherited Link);
end;

procedure TFhirAllergyIntoleranceReactionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAllergyIntoleranceReactionList.SetItemByIndex(index: Integer; value: TFhirAllergyIntoleranceReaction);
begin
  assert(value is TFhirAllergyIntoleranceReaction);
  FhirAllergyIntoleranceReactions[index] := value;
end;

procedure TFhirAllergyIntoleranceReactionList.SetItemN(index: Integer; value: TFhirAllergyIntoleranceReaction);
begin
  assert(value is TFhirAllergyIntoleranceReaction);
  ObjectByIndex[index] := value;
end;

{ TFhirAllergyIntolerance }

constructor TFhirAllergyIntolerance.Create;
begin
  inherited;
end;

destructor TFhirAllergyIntolerance.Destroy;
begin
  FIdentifierList.Free;
  FOnset.free;
  FRecordedDate.free;
  FRecorder.free;
  FPatient.free;
  FReporter.free;
  FSubstance.free;
  FStatus.free;
  FCriticality.free;
  FType_.free;
  FCategory.free;
  FLastOccurence.free;
  FNote.free;
  FReactionList.Free;
  inherited;
end;

function TFhirAllergyIntolerance.GetResourceType : TFhirResourceType;
begin
  result := frtAllergyIntolerance;
end;

procedure TFhirAllergyIntolerance.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAllergyIntolerance(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirAllergyIntolerance(oSource).FIdentifierList);
  end;
  onsetElement := TFhirAllergyIntolerance(oSource).onsetElement.Clone;
  recordedDateElement := TFhirAllergyIntolerance(oSource).recordedDateElement.Clone;
  recorder := TFhirAllergyIntolerance(oSource).recorder.Clone;
  patient := TFhirAllergyIntolerance(oSource).patient.Clone;
  reporter := TFhirAllergyIntolerance(oSource).reporter.Clone;
  substance := TFhirAllergyIntolerance(oSource).substance.Clone;
  FStatus := TFhirAllergyIntolerance(oSource).FStatus.Link;
  FCriticality := TFhirAllergyIntolerance(oSource).FCriticality.Link;
  FType_ := TFhirAllergyIntolerance(oSource).FType_.Link;
  FCategory := TFhirAllergyIntolerance(oSource).FCategory.Link;
  lastOccurenceElement := TFhirAllergyIntolerance(oSource).lastOccurenceElement.Clone;
  note := TFhirAllergyIntolerance(oSource).note.Clone;
  if (TFhirAllergyIntolerance(oSource).FReactionList = nil) then
  begin
    FReactionList.free;
    FReactionList := nil;
  end
  else
  begin
    FReactionList := TFhirAllergyIntoleranceReactionList.Create;
    FReactionList.Assign(TFhirAllergyIntolerance(oSource).FReactionList);
  end;
end;

procedure TFhirAllergyIntolerance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'onset') Then
     list.add(FOnset.Link);
  if (child_name = 'recordedDate') Then
     list.add(FRecordedDate.Link);
  if (child_name = 'recorder') Then
     list.add(FRecorder.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'reporter') Then
     list.add(FReporter.Link);
  if (child_name = 'substance') Then
     list.add(FSubstance.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'criticality') Then
     list.add(FCriticality.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'lastOccurence') Then
     list.add(FLastOccurence.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'reaction') Then
    list.addAll(FReactionList);
end;

procedure TFhirAllergyIntolerance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'onset', 'dateTime', FOnset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recordedDate', 'dateTime', FRecordedDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recorder', 'Reference(Practitioner|Patient)', FRecorder.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reporter', 'Reference(Patient|RelatedPerson|Practitioner)', FReporter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'substance', 'CodeableConcept', FSubstance.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'criticality', 'code', FCriticality.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'category', 'code', FCategory.Link));{1}
  oList.add(TFHIRProperty.create(self, 'lastOccurence', 'dateTime', FLastOccurence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reaction', '', FReactionList.Link)){3};
end;

procedure TFhirAllergyIntolerance.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'onset') then OnsetElement := propValue as TFhirDateTime{5a}
  else if (propName = 'recordedDate') then RecordedDateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'recorder') then Recorder := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'reporter') then Reporter := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'substance') then Substance := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'criticality') then CriticalityElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'category') then CategoryElement := propValue as TFHIREnum
  else if (propName = 'lastOccurence') then LastOccurenceElement := propValue as TFhirDateTime{5a}
  else if (propName = 'note') then Note := propValue as TFhirAnnotation{4b}
  else if (propName = 'reaction') then ReactionList.add(propValue as TFhirAllergyIntoleranceReaction){2}
  else inherited;
end;

function TFhirAllergyIntolerance.FhirType : string;
begin
  result := 'AllergyIntolerance';
end;

function TFhirAllergyIntolerance.Link : TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance(inherited Link);
end;

function TFhirAllergyIntolerance.Clone : TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance(inherited Clone);
end;

{ TFhirAllergyIntolerance }

Function TFhirAllergyIntolerance.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirAllergyIntolerance.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirAllergyIntolerance.SetOnset(value : TFhirDateTime);
begin
  FOnset.free;
  FOnset := value;
end;

Function TFhirAllergyIntolerance.GetOnsetST : TDateAndTime;
begin
  if FOnset = nil then
    result := nil
  else
    result := FOnset.value;
end;

Procedure TFhirAllergyIntolerance.SetOnsetST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FOnset = nil then
      FOnset := TFhirDateTime.create;
    FOnset.value := value
  end
  else if FOnset <> nil then
    FOnset.value := nil;
end;

Procedure TFhirAllergyIntolerance.SetRecordedDate(value : TFhirDateTime);
begin
  FRecordedDate.free;
  FRecordedDate := value;
end;

Function TFhirAllergyIntolerance.GetRecordedDateST : TDateAndTime;
begin
  if FRecordedDate = nil then
    result := nil
  else
    result := FRecordedDate.value;
end;

Procedure TFhirAllergyIntolerance.SetRecordedDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FRecordedDate = nil then
      FRecordedDate := TFhirDateTime.create;
    FRecordedDate.value := value
  end
  else if FRecordedDate <> nil then
    FRecordedDate.value := nil;
end;

Procedure TFhirAllergyIntolerance.SetRecorder(value : TFhirReference{Resource});
begin
  FRecorder.free;
  FRecorder := value;
end;

Procedure TFhirAllergyIntolerance.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirAllergyIntolerance.SetReporter(value : TFhirReference{Resource});
begin
  FReporter.free;
  FReporter := value;
end;

Procedure TFhirAllergyIntolerance.SetSubstance(value : TFhirCodeableConcept);
begin
  FSubstance.free;
  FSubstance := value;
end;

Procedure TFhirAllergyIntolerance.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirAllergyIntolerance.GetStatusST : TFhirAllergyIntoleranceStatus;
begin
  if FStatus = nil then
    result := TFhirAllergyIntoleranceStatus(0)
  else
    result := TFhirAllergyIntoleranceStatus(StringArrayIndexOfSensitive(CODES_TFhirAllergyIntoleranceStatus, FStatus.value));
end;

Procedure TFhirAllergyIntolerance.SetStatusST(value : TFhirAllergyIntoleranceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirAllergyIntoleranceStatus[value]);
end;

Procedure TFhirAllergyIntolerance.SetCriticality(value : TFhirEnum);
begin
  FCriticality.free;
  FCriticality := value;
end;

Function TFhirAllergyIntolerance.GetCriticalityST : TFhirAllergyIntoleranceCriticality;
begin
  if FCriticality = nil then
    result := TFhirAllergyIntoleranceCriticality(0)
  else
    result := TFhirAllergyIntoleranceCriticality(StringArrayIndexOfSensitive(CODES_TFhirAllergyIntoleranceCriticality, FCriticality.value));
end;

Procedure TFhirAllergyIntolerance.SetCriticalityST(value : TFhirAllergyIntoleranceCriticality);
begin
  if ord(value) = 0 then
    CriticalityElement := nil
  else
    CriticalityElement := TFhirEnum.create(CODES_TFhirAllergyIntoleranceCriticality[value]);
end;

Procedure TFhirAllergyIntolerance.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirAllergyIntolerance.GetType_ST : TFhirAllergyIntoleranceType;
begin
  if FType_ = nil then
    result := TFhirAllergyIntoleranceType(0)
  else
    result := TFhirAllergyIntoleranceType(StringArrayIndexOfSensitive(CODES_TFhirAllergyIntoleranceType, FType_.value));
end;

Procedure TFhirAllergyIntolerance.SetType_ST(value : TFhirAllergyIntoleranceType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirAllergyIntoleranceType[value]);
end;

Procedure TFhirAllergyIntolerance.SetCategory(value : TFhirEnum);
begin
  FCategory.free;
  FCategory := value;
end;

Function TFhirAllergyIntolerance.GetCategoryST : TFhirAllergyIntoleranceCategory;
begin
  if FCategory = nil then
    result := TFhirAllergyIntoleranceCategory(0)
  else
    result := TFhirAllergyIntoleranceCategory(StringArrayIndexOfSensitive(CODES_TFhirAllergyIntoleranceCategory, FCategory.value));
end;

Procedure TFhirAllergyIntolerance.SetCategoryST(value : TFhirAllergyIntoleranceCategory);
begin
  if ord(value) = 0 then
    CategoryElement := nil
  else
    CategoryElement := TFhirEnum.create(CODES_TFhirAllergyIntoleranceCategory[value]);
end;

Procedure TFhirAllergyIntolerance.SetLastOccurence(value : TFhirDateTime);
begin
  FLastOccurence.free;
  FLastOccurence := value;
end;

Function TFhirAllergyIntolerance.GetLastOccurenceST : TDateAndTime;
begin
  if FLastOccurence = nil then
    result := nil
  else
    result := FLastOccurence.value;
end;

Procedure TFhirAllergyIntolerance.SetLastOccurenceST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FLastOccurence = nil then
      FLastOccurence := TFhirDateTime.create;
    FLastOccurence.value := value
  end
  else if FLastOccurence <> nil then
    FLastOccurence.value := nil;
end;

Procedure TFhirAllergyIntolerance.SetNote(value : TFhirAnnotation);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirAllergyIntolerance.GetReactionList : TFhirAllergyIntoleranceReactionList;
begin
  if FReactionList = nil then
    FReactionList := TFhirAllergyIntoleranceReactionList.Create;
  result := FReactionList;
end;

Function TFhirAllergyIntolerance.GetHasReactionList : boolean;
begin
  result := (FReactionList <> nil) and (FReactionList.count > 0);
end;


{ TFhirAllergyIntoleranceListEnumerator }

Constructor TFhirAllergyIntoleranceListEnumerator.Create(list : TFhirAllergyIntoleranceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAllergyIntoleranceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAllergyIntoleranceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAllergyIntoleranceListEnumerator.GetCurrent : TFhirAllergyIntolerance;
begin
  Result := FList[FIndex];
end;


{ TFhirAllergyIntoleranceList }
procedure TFhirAllergyIntoleranceList.AddItem(value: TFhirAllergyIntolerance);
begin
  assert(value.ClassName = 'TFhirAllergyIntolerance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAllergyIntolerance');
  add(value);
end;


function TFhirAllergyIntoleranceList.Append: TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAllergyIntoleranceList.ClearItems;
begin
  Clear;
end;

function TFhirAllergyIntoleranceList.GetEnumerator : TFhirAllergyIntoleranceListEnumerator;
begin
  result := TFhirAllergyIntoleranceListEnumerator.Create(self.link);
end;

function TFhirAllergyIntoleranceList.Clone: TFhirAllergyIntoleranceList;
begin
  result := TFhirAllergyIntoleranceList(inherited Clone);
end;

function TFhirAllergyIntoleranceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAllergyIntoleranceList.GetItemN(index: Integer): TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance(ObjectByIndex[index]);
end;

function TFhirAllergyIntoleranceList.IndexOf(value: TFhirAllergyIntolerance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAllergyIntoleranceList.Insert(index: Integer): TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAllergyIntoleranceList.InsertItem(index: Integer; value: TFhirAllergyIntolerance);
begin
  assert(value is TFhirAllergyIntolerance);
  Inherited Insert(index, value);
end;

function TFhirAllergyIntoleranceList.Item(index: Integer): TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance(ObjectByIndex[index]);
end;

function TFhirAllergyIntoleranceList.Link: TFhirAllergyIntoleranceList;
begin
  result := TFhirAllergyIntoleranceList(inherited Link);
end;

procedure TFhirAllergyIntoleranceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAllergyIntoleranceList.SetItemByIndex(index: Integer; value: TFhirAllergyIntolerance);
begin
  assert(value is TFhirAllergyIntolerance);
  FhirAllergyIntolerances[index] := value;
end;

procedure TFhirAllergyIntoleranceList.SetItemN(index: Integer; value: TFhirAllergyIntolerance);
begin
  assert(value is TFhirAllergyIntolerance);
  ObjectByIndex[index] := value;
end;

{ TFhirAppointmentParticipant }

constructor TFhirAppointmentParticipant.Create;
begin
  inherited;
end;

destructor TFhirAppointmentParticipant.Destroy;
begin
  FType_List.Free;
  FActor.free;
  FRequired.free;
  FStatus.free;
  inherited;
end;

procedure TFhirAppointmentParticipant.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAppointmentParticipant(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirAppointmentParticipant(oSource).FType_List);
  end;
  actor := TFhirAppointmentParticipant(oSource).actor.Clone;
  FRequired := TFhirAppointmentParticipant(oSource).FRequired.Link;
  FStatus := TFhirAppointmentParticipant(oSource).FStatus.Link;
end;

procedure TFhirAppointmentParticipant.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'actor') Then
     list.add(FActor.Link);
  if (child_name = 'required') Then
     list.add(FRequired.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
end;

procedure TFhirAppointmentParticipant.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Patient|Practitioner|RelatedPerson|Device|HealthcareService|Location)', FActor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'required', 'code', FRequired.Link));{1}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
end;

procedure TFhirAppointmentParticipant.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'required') then RequiredElement := propValue as TFHIREnum
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else inherited;
end;

function TFhirAppointmentParticipant.FhirType : string;
begin
  result := 'participant';
end;

function TFhirAppointmentParticipant.Link : TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant(inherited Link);
end;

function TFhirAppointmentParticipant.Clone : TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant(inherited Clone);
end;

function TFhirAppointmentParticipant.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAppointmentParticipant;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAppointmentParticipant)) then
    result := false
  else
  begin
    o := TFhirAppointmentParticipant(other);
    result := compareDeep(type_List, o.type_List, true) and compareDeep(actorElement, o.actorElement, true) and 
      compareDeep(requiredElement, o.requiredElement, true) and compareDeep(statusElement, o.statusElement, true);
  end;
end;

function TFhirAppointmentParticipant.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAppointmentParticipant;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAppointmentParticipant)) then
    result := false
  else
  begin
    o := TFhirAppointmentParticipant(other);
    result := compareValues(requiredElement, o.requiredElement, true) and compareValues(statusElement, o.statusElement, true);
  end;
end;

{ TFhirAppointmentParticipant }

Function TFhirAppointmentParticipant.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirAppointmentParticipant.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Procedure TFhirAppointmentParticipant.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;

Procedure TFhirAppointmentParticipant.SetRequired(value : TFhirEnum);
begin
  FRequired.free;
  FRequired := value;
end;

Function TFhirAppointmentParticipant.GetRequiredST : TFhirParticipantrequired;
begin
  if FRequired = nil then
    result := TFhirParticipantrequired(0)
  else
    result := TFhirParticipantrequired(StringArrayIndexOfSensitive(CODES_TFhirParticipantrequired, FRequired.value));
end;

Procedure TFhirAppointmentParticipant.SetRequiredST(value : TFhirParticipantrequired);
begin
  if ord(value) = 0 then
    RequiredElement := nil
  else
    RequiredElement := TFhirEnum.create(CODES_TFhirParticipantrequired[value]);
end;

Procedure TFhirAppointmentParticipant.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirAppointmentParticipant.GetStatusST : TFhirParticipationstatus;
begin
  if FStatus = nil then
    result := TFhirParticipationstatus(0)
  else
    result := TFhirParticipationstatus(StringArrayIndexOfSensitive(CODES_TFhirParticipationstatus, FStatus.value));
end;

Procedure TFhirAppointmentParticipant.SetStatusST(value : TFhirParticipationstatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirParticipationstatus[value]);
end;


{ TFhirAppointmentParticipantListEnumerator }

Constructor TFhirAppointmentParticipantListEnumerator.Create(list : TFhirAppointmentParticipantList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAppointmentParticipantListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAppointmentParticipantListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAppointmentParticipantListEnumerator.GetCurrent : TFhirAppointmentParticipant;
begin
  Result := FList[FIndex];
end;


{ TFhirAppointmentParticipantList }
procedure TFhirAppointmentParticipantList.AddItem(value: TFhirAppointmentParticipant);
begin
  assert(value.ClassName = 'TFhirAppointmentParticipant', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAppointmentParticipant');
  add(value);
end;


function TFhirAppointmentParticipantList.Append: TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentParticipantList.ClearItems;
begin
  Clear;
end;

function TFhirAppointmentParticipantList.GetEnumerator : TFhirAppointmentParticipantListEnumerator;
begin
  result := TFhirAppointmentParticipantListEnumerator.Create(self.link);
end;

function TFhirAppointmentParticipantList.Clone: TFhirAppointmentParticipantList;
begin
  result := TFhirAppointmentParticipantList(inherited Clone);
end;

function TFhirAppointmentParticipantList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAppointmentParticipantList.GetItemN(index: Integer): TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant(ObjectByIndex[index]);
end;

function TFhirAppointmentParticipantList.IndexOf(value: TFhirAppointmentParticipant): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAppointmentParticipantList.Insert(index: Integer): TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentParticipantList.InsertItem(index: Integer; value: TFhirAppointmentParticipant);
begin
  assert(value is TFhirAppointmentParticipant);
  Inherited Insert(index, value);
end;

function TFhirAppointmentParticipantList.Item(index: Integer): TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant(ObjectByIndex[index]);
end;

function TFhirAppointmentParticipantList.Link: TFhirAppointmentParticipantList;
begin
  result := TFhirAppointmentParticipantList(inherited Link);
end;

procedure TFhirAppointmentParticipantList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAppointmentParticipantList.SetItemByIndex(index: Integer; value: TFhirAppointmentParticipant);
begin
  assert(value is TFhirAppointmentParticipant);
  FhirAppointmentParticipants[index] := value;
end;

procedure TFhirAppointmentParticipantList.SetItemN(index: Integer; value: TFhirAppointmentParticipant);
begin
  assert(value is TFhirAppointmentParticipant);
  ObjectByIndex[index] := value;
end;

{ TFhirAppointment }

constructor TFhirAppointment.Create;
begin
  inherited;
end;

destructor TFhirAppointment.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FType_.free;
  FReason.free;
  FPriority.free;
  FDescription.free;
  FStart.free;
  FEnd_.free;
  FMinutesDuration.free;
  FSlotList.Free;
  FComment.free;
  FParticipantList.Free;
  inherited;
end;

function TFhirAppointment.GetResourceType : TFhirResourceType;
begin
  result := frtAppointment;
end;

procedure TFhirAppointment.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAppointment(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirAppointment(oSource).FIdentifierList);
  end;
  FStatus := TFhirAppointment(oSource).FStatus.Link;
  type_ := TFhirAppointment(oSource).type_.Clone;
  reason := TFhirAppointment(oSource).reason.Clone;
  priorityElement := TFhirAppointment(oSource).priorityElement.Clone;
  descriptionElement := TFhirAppointment(oSource).descriptionElement.Clone;
  startElement := TFhirAppointment(oSource).startElement.Clone;
  end_Element := TFhirAppointment(oSource).end_Element.Clone;
  minutesDurationElement := TFhirAppointment(oSource).minutesDurationElement.Clone;
  if (TFhirAppointment(oSource).FSlotList = nil) then
  begin
    FSlotList.free;
    FSlotList := nil;
  end
  else
  begin
    FSlotList := TFhirReferenceList{TFhirSlot}.Create;
    FSlotList.Assign(TFhirAppointment(oSource).FSlotList);
  end;
  commentElement := TFhirAppointment(oSource).commentElement.Clone;
  if (TFhirAppointment(oSource).FParticipantList = nil) then
  begin
    FParticipantList.free;
    FParticipantList := nil;
  end
  else
  begin
    FParticipantList := TFhirAppointmentParticipantList.Create;
    FParticipantList.Assign(TFhirAppointment(oSource).FParticipantList);
  end;
end;

procedure TFhirAppointment.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'start') Then
     list.add(FStart.Link);
  if (child_name = 'end') Then
     list.add(FEnd_.Link);
  if (child_name = 'minutesDuration') Then
     list.add(FMinutesDuration.Link);
  if (child_name = 'slot') Then
    list.addAll(FSlotList);
  if (child_name = 'comment') Then
     list.add(FComment.Link);
  if (child_name = 'participant') Then
    list.addAll(FParticipantList);
end;

procedure TFhirAppointment.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'unsignedInt', FPriority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'start', 'instant', FStart.Link));{2}
  oList.add(TFHIRProperty.create(self, 'end', 'instant', FEnd_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'minutesDuration', 'positiveInt', FMinutesDuration.Link));{2}
  oList.add(TFHIRProperty.create(self, 'slot', 'Reference(Slot)', FSlotList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'participant', '', FParticipantList.Link)){3};
end;

procedure TFhirAppointment.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'reason') then Reason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'priority') then PriorityElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'start') then StartElement := propValue as TFhirInstant{5a}
  else if (propName = 'end') then End_Element := propValue as TFhirInstant{5a}
  else if (propName = 'minutesDuration') then MinutesDurationElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'slot') then SlotList.add(propValue as TFhirReference{TFhirSlot}){2}
  else if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else if (propName = 'participant') then ParticipantList.add(propValue as TFhirAppointmentParticipant){2}
  else inherited;
end;

function TFhirAppointment.FhirType : string;
begin
  result := 'Appointment';
end;

function TFhirAppointment.Link : TFhirAppointment;
begin
  result := TFhirAppointment(inherited Link);
end;

function TFhirAppointment.Clone : TFhirAppointment;
begin
  result := TFhirAppointment(inherited Clone);
end;

{ TFhirAppointment }

Function TFhirAppointment.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirAppointment.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirAppointment.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirAppointment.GetStatusST : TFhirAppointmentstatus;
begin
  if FStatus = nil then
    result := TFhirAppointmentstatus(0)
  else
    result := TFhirAppointmentstatus(StringArrayIndexOfSensitive(CODES_TFhirAppointmentstatus, FStatus.value));
end;

Procedure TFhirAppointment.SetStatusST(value : TFhirAppointmentstatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirAppointmentstatus[value]);
end;

Procedure TFhirAppointment.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirAppointment.SetReason(value : TFhirCodeableConcept);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirAppointment.SetPriority(value : TFhirUnsignedInt);
begin
  FPriority.free;
  FPriority := value;
end;

Function TFhirAppointment.GetPriorityST : String;
begin
  if FPriority = nil then
    result := ''
  else
    result := FPriority.value;
end;

Procedure TFhirAppointment.SetPriorityST(value : String);
begin
  if value <> '' then
  begin
    if FPriority = nil then
      FPriority := TFhirUnsignedInt.create;
    FPriority.value := value
  end
  else if FPriority <> nil then
    FPriority.value := '';
end;

Procedure TFhirAppointment.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirAppointment.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirAppointment.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirAppointment.SetStart(value : TFhirInstant);
begin
  FStart.free;
  FStart := value;
end;

Function TFhirAppointment.GetStartST : TDateAndTime;
begin
  if FStart = nil then
    result := nil
  else
    result := FStart.value;
end;

Procedure TFhirAppointment.SetStartST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStart = nil then
      FStart := TFhirInstant.create;
    FStart.value := value
  end
  else if FStart <> nil then
    FStart.value := nil;
end;

Procedure TFhirAppointment.SetEnd_(value : TFhirInstant);
begin
  FEnd_.free;
  FEnd_ := value;
end;

Function TFhirAppointment.GetEnd_ST : TDateAndTime;
begin
  if FEnd_ = nil then
    result := nil
  else
    result := FEnd_.value;
end;

Procedure TFhirAppointment.SetEnd_ST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEnd_ = nil then
      FEnd_ := TFhirInstant.create;
    FEnd_.value := value
  end
  else if FEnd_ <> nil then
    FEnd_.value := nil;
end;

Procedure TFhirAppointment.SetMinutesDuration(value : TFhirPositiveInt);
begin
  FMinutesDuration.free;
  FMinutesDuration := value;
end;

Function TFhirAppointment.GetMinutesDurationST : String;
begin
  if FMinutesDuration = nil then
    result := ''
  else
    result := FMinutesDuration.value;
end;

Procedure TFhirAppointment.SetMinutesDurationST(value : String);
begin
  if value <> '' then
  begin
    if FMinutesDuration = nil then
      FMinutesDuration := TFhirPositiveInt.create;
    FMinutesDuration.value := value
  end
  else if FMinutesDuration <> nil then
    FMinutesDuration.value := '';
end;

Function TFhirAppointment.GetSlotList : TFhirReferenceList{TFhirSlot};
begin
  if FSlotList = nil then
    FSlotList := TFhirReferenceList{TFhirSlot}.Create;
  result := FSlotList;
end;

Function TFhirAppointment.GetHasSlotList : boolean;
begin
  result := (FSlotList <> nil) and (FSlotList.count > 0);
end;

Procedure TFhirAppointment.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirAppointment.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirAppointment.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;

Function TFhirAppointment.GetParticipantList : TFhirAppointmentParticipantList;
begin
  if FParticipantList = nil then
    FParticipantList := TFhirAppointmentParticipantList.Create;
  result := FParticipantList;
end;

Function TFhirAppointment.GetHasParticipantList : boolean;
begin
  result := (FParticipantList <> nil) and (FParticipantList.count > 0);
end;


{ TFhirAppointmentListEnumerator }

Constructor TFhirAppointmentListEnumerator.Create(list : TFhirAppointmentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAppointmentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAppointmentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAppointmentListEnumerator.GetCurrent : TFhirAppointment;
begin
  Result := FList[FIndex];
end;


{ TFhirAppointmentList }
procedure TFhirAppointmentList.AddItem(value: TFhirAppointment);
begin
  assert(value.ClassName = 'TFhirAppointment', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAppointment');
  add(value);
end;


function TFhirAppointmentList.Append: TFhirAppointment;
begin
  result := TFhirAppointment.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentList.ClearItems;
begin
  Clear;
end;

function TFhirAppointmentList.GetEnumerator : TFhirAppointmentListEnumerator;
begin
  result := TFhirAppointmentListEnumerator.Create(self.link);
end;

function TFhirAppointmentList.Clone: TFhirAppointmentList;
begin
  result := TFhirAppointmentList(inherited Clone);
end;

function TFhirAppointmentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAppointmentList.GetItemN(index: Integer): TFhirAppointment;
begin
  result := TFhirAppointment(ObjectByIndex[index]);
end;

function TFhirAppointmentList.IndexOf(value: TFhirAppointment): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAppointmentList.Insert(index: Integer): TFhirAppointment;
begin
  result := TFhirAppointment.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentList.InsertItem(index: Integer; value: TFhirAppointment);
begin
  assert(value is TFhirAppointment);
  Inherited Insert(index, value);
end;

function TFhirAppointmentList.Item(index: Integer): TFhirAppointment;
begin
  result := TFhirAppointment(ObjectByIndex[index]);
end;

function TFhirAppointmentList.Link: TFhirAppointmentList;
begin
  result := TFhirAppointmentList(inherited Link);
end;

procedure TFhirAppointmentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAppointmentList.SetItemByIndex(index: Integer; value: TFhirAppointment);
begin
  assert(value is TFhirAppointment);
  FhirAppointments[index] := value;
end;

procedure TFhirAppointmentList.SetItemN(index: Integer; value: TFhirAppointment);
begin
  assert(value is TFhirAppointment);
  ObjectByIndex[index] := value;
end;

{ TFhirAppointmentResponse }

constructor TFhirAppointmentResponse.Create;
begin
  inherited;
end;

destructor TFhirAppointmentResponse.Destroy;
begin
  FIdentifierList.Free;
  FAppointment.free;
  FStart.free;
  FEnd_.free;
  FParticipantTypeList.Free;
  FActor.free;
  FParticipantStatus.free;
  FComment.free;
  inherited;
end;

function TFhirAppointmentResponse.GetResourceType : TFhirResourceType;
begin
  result := frtAppointmentResponse;
end;

procedure TFhirAppointmentResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAppointmentResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirAppointmentResponse(oSource).FIdentifierList);
  end;
  appointment := TFhirAppointmentResponse(oSource).appointment.Clone;
  startElement := TFhirAppointmentResponse(oSource).startElement.Clone;
  end_Element := TFhirAppointmentResponse(oSource).end_Element.Clone;
  if (TFhirAppointmentResponse(oSource).FParticipantTypeList = nil) then
  begin
    FParticipantTypeList.free;
    FParticipantTypeList := nil;
  end
  else
  begin
    FParticipantTypeList := TFhirCodeableConceptList.Create;
    FParticipantTypeList.Assign(TFhirAppointmentResponse(oSource).FParticipantTypeList);
  end;
  actor := TFhirAppointmentResponse(oSource).actor.Clone;
  FParticipantStatus := TFhirAppointmentResponse(oSource).FParticipantStatus.Link;
  commentElement := TFhirAppointmentResponse(oSource).commentElement.Clone;
end;

procedure TFhirAppointmentResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'appointment') Then
     list.add(FAppointment.Link);
  if (child_name = 'start') Then
     list.add(FStart.Link);
  if (child_name = 'end') Then
     list.add(FEnd_.Link);
  if (child_name = 'participantType') Then
    list.addAll(FParticipantTypeList);
  if (child_name = 'actor') Then
     list.add(FActor.Link);
  if (child_name = 'participantStatus') Then
     list.add(FParticipantStatus.Link);
  if (child_name = 'comment') Then
     list.add(FComment.Link);
end;

procedure TFhirAppointmentResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'appointment', 'Reference(Appointment)', FAppointment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'start', 'instant', FStart.Link));{2}
  oList.add(TFHIRProperty.create(self, 'end', 'instant', FEnd_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'participantType', 'CodeableConcept', FParticipantTypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Patient|Practitioner|RelatedPerson|Device|HealthcareService|Location)', FActor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'participantStatus', 'code', FParticipantStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
end;

procedure TFhirAppointmentResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'appointment') then Appointment := propValue as TFhirReference{TFhirAppointment}{4b}
  else if (propName = 'start') then StartElement := propValue as TFhirInstant{5a}
  else if (propName = 'end') then End_Element := propValue as TFhirInstant{5a}
  else if (propName = 'participantType') then ParticipantTypeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'participantStatus') then ParticipantStatusElement := propValue as TFHIREnum
  else if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirAppointmentResponse.FhirType : string;
begin
  result := 'AppointmentResponse';
end;

function TFhirAppointmentResponse.Link : TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse(inherited Link);
end;

function TFhirAppointmentResponse.Clone : TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse(inherited Clone);
end;

{ TFhirAppointmentResponse }

Function TFhirAppointmentResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirAppointmentResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirAppointmentResponse.SetAppointment(value : TFhirReference{TFhirAppointment});
begin
  FAppointment.free;
  FAppointment := value;
end;

Procedure TFhirAppointmentResponse.SetStart(value : TFhirInstant);
begin
  FStart.free;
  FStart := value;
end;

Function TFhirAppointmentResponse.GetStartST : TDateAndTime;
begin
  if FStart = nil then
    result := nil
  else
    result := FStart.value;
end;

Procedure TFhirAppointmentResponse.SetStartST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStart = nil then
      FStart := TFhirInstant.create;
    FStart.value := value
  end
  else if FStart <> nil then
    FStart.value := nil;
end;

Procedure TFhirAppointmentResponse.SetEnd_(value : TFhirInstant);
begin
  FEnd_.free;
  FEnd_ := value;
end;

Function TFhirAppointmentResponse.GetEnd_ST : TDateAndTime;
begin
  if FEnd_ = nil then
    result := nil
  else
    result := FEnd_.value;
end;

Procedure TFhirAppointmentResponse.SetEnd_ST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEnd_ = nil then
      FEnd_ := TFhirInstant.create;
    FEnd_.value := value
  end
  else if FEnd_ <> nil then
    FEnd_.value := nil;
end;

Function TFhirAppointmentResponse.GetParticipantTypeList : TFhirCodeableConceptList;
begin
  if FParticipantTypeList = nil then
    FParticipantTypeList := TFhirCodeableConceptList.Create;
  result := FParticipantTypeList;
end;

Function TFhirAppointmentResponse.GetHasParticipantTypeList : boolean;
begin
  result := (FParticipantTypeList <> nil) and (FParticipantTypeList.count > 0);
end;

Procedure TFhirAppointmentResponse.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;

Procedure TFhirAppointmentResponse.SetParticipantStatus(value : TFhirEnum);
begin
  FParticipantStatus.free;
  FParticipantStatus := value;
end;

Function TFhirAppointmentResponse.GetParticipantStatusST : TFhirParticipantstatus;
begin
  if FParticipantStatus = nil then
    result := TFhirParticipantstatus(0)
  else
    result := TFhirParticipantstatus(StringArrayIndexOfSensitive(CODES_TFhirParticipantstatus, FParticipantStatus.value));
end;

Procedure TFhirAppointmentResponse.SetParticipantStatusST(value : TFhirParticipantstatus);
begin
  if ord(value) = 0 then
    ParticipantStatusElement := nil
  else
    ParticipantStatusElement := TFhirEnum.create(CODES_TFhirParticipantstatus[value]);
end;

Procedure TFhirAppointmentResponse.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirAppointmentResponse.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirAppointmentResponse.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;


{ TFhirAppointmentResponseListEnumerator }

Constructor TFhirAppointmentResponseListEnumerator.Create(list : TFhirAppointmentResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAppointmentResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAppointmentResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAppointmentResponseListEnumerator.GetCurrent : TFhirAppointmentResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirAppointmentResponseList }
procedure TFhirAppointmentResponseList.AddItem(value: TFhirAppointmentResponse);
begin
  assert(value.ClassName = 'TFhirAppointmentResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAppointmentResponse');
  add(value);
end;


function TFhirAppointmentResponseList.Append: TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentResponseList.ClearItems;
begin
  Clear;
end;

function TFhirAppointmentResponseList.GetEnumerator : TFhirAppointmentResponseListEnumerator;
begin
  result := TFhirAppointmentResponseListEnumerator.Create(self.link);
end;

function TFhirAppointmentResponseList.Clone: TFhirAppointmentResponseList;
begin
  result := TFhirAppointmentResponseList(inherited Clone);
end;

function TFhirAppointmentResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAppointmentResponseList.GetItemN(index: Integer): TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse(ObjectByIndex[index]);
end;

function TFhirAppointmentResponseList.IndexOf(value: TFhirAppointmentResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAppointmentResponseList.Insert(index: Integer): TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAppointmentResponseList.InsertItem(index: Integer; value: TFhirAppointmentResponse);
begin
  assert(value is TFhirAppointmentResponse);
  Inherited Insert(index, value);
end;

function TFhirAppointmentResponseList.Item(index: Integer): TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse(ObjectByIndex[index]);
end;

function TFhirAppointmentResponseList.Link: TFhirAppointmentResponseList;
begin
  result := TFhirAppointmentResponseList(inherited Link);
end;

procedure TFhirAppointmentResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAppointmentResponseList.SetItemByIndex(index: Integer; value: TFhirAppointmentResponse);
begin
  assert(value is TFhirAppointmentResponse);
  FhirAppointmentResponses[index] := value;
end;

procedure TFhirAppointmentResponseList.SetItemN(index: Integer; value: TFhirAppointmentResponse);
begin
  assert(value is TFhirAppointmentResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventEvent }

constructor TFhirAuditEventEvent.Create;
begin
  inherited;
end;

destructor TFhirAuditEventEvent.Destroy;
begin
  FType_.free;
  FSubtypeList.Free;
  FAction.free;
  FDateTime.free;
  FOutcome.free;
  FOutcomeDesc.free;
  FPurposeOfEventList.Free;
  inherited;
end;

procedure TFhirAuditEventEvent.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirAuditEventEvent(oSource).type_.Clone;
  if (TFhirAuditEventEvent(oSource).FSubtypeList = nil) then
  begin
    FSubtypeList.free;
    FSubtypeList := nil;
  end
  else
  begin
    FSubtypeList := TFhirCodingList.Create;
    FSubtypeList.Assign(TFhirAuditEventEvent(oSource).FSubtypeList);
  end;
  FAction := TFhirAuditEventEvent(oSource).FAction.Link;
  dateTimeElement := TFhirAuditEventEvent(oSource).dateTimeElement.Clone;
  FOutcome := TFhirAuditEventEvent(oSource).FOutcome.Link;
  outcomeDescElement := TFhirAuditEventEvent(oSource).outcomeDescElement.Clone;
  if (TFhirAuditEventEvent(oSource).FPurposeOfEventList = nil) then
  begin
    FPurposeOfEventList.free;
    FPurposeOfEventList := nil;
  end
  else
  begin
    FPurposeOfEventList := TFhirCodingList.Create;
    FPurposeOfEventList.Assign(TFhirAuditEventEvent(oSource).FPurposeOfEventList);
  end;
end;

procedure TFhirAuditEventEvent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'subtype') Then
    list.addAll(FSubtypeList);
  if (child_name = 'action') Then
     list.add(FAction.Link);
  if (child_name = 'dateTime') Then
     list.add(FDateTime.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'outcomeDesc') Then
     list.add(FOutcomeDesc.Link);
  if (child_name = 'purposeOfEvent') Then
    list.addAll(FPurposeOfEventList);
end;

procedure TFhirAuditEventEvent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subtype', 'Coding', FSubtypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'action', 'code', FAction.Link));{1}
  oList.add(TFHIRProperty.create(self, 'dateTime', 'instant', FDateTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'outcomeDesc', 'string', FOutcomeDesc.Link));{2}
  oList.add(TFHIRProperty.create(self, 'purposeOfEvent', 'Coding', FPurposeOfEventList.Link)){3};
end;

procedure TFhirAuditEventEvent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'subtype') then SubtypeList.add(propValue as TFhirCoding){2}
  else if (propName = 'action') then ActionElement := propValue as TFHIREnum
  else if (propName = 'dateTime') then DateTimeElement := propValue as TFhirInstant{5a}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'outcomeDesc') then OutcomeDescElement := propValue as TFhirString{5a}
  else if (propName = 'purposeOfEvent') then PurposeOfEventList.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirAuditEventEvent.FhirType : string;
begin
  result := 'event';
end;

function TFhirAuditEventEvent.Link : TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent(inherited Link);
end;

function TFhirAuditEventEvent.Clone : TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent(inherited Clone);
end;

function TFhirAuditEventEvent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventEvent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventEvent)) then
    result := false
  else
  begin
    o := TFhirAuditEventEvent(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(subtypeList, o.subtypeList, true) and 
      compareDeep(actionElement, o.actionElement, true) and compareDeep(dateTimeElement, o.dateTimeElement, true) and 
      compareDeep(outcomeElement, o.outcomeElement, true) and compareDeep(outcomeDescElement, o.outcomeDescElement, true) and 
      compareDeep(purposeOfEventList, o.purposeOfEventList, true);
  end;
end;

function TFhirAuditEventEvent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventEvent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventEvent)) then
    result := false
  else
  begin
    o := TFhirAuditEventEvent(other);
    result := compareValues(actionElement, o.actionElement, true) and compareValues(dateTimeElement, o.dateTimeElement, true) and 
      compareValues(outcomeElement, o.outcomeElement, true) and compareValues(outcomeDescElement, o.outcomeDescElement, true);
  end;
end;

{ TFhirAuditEventEvent }

Procedure TFhirAuditEventEvent.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirAuditEventEvent.GetSubtypeList : TFhirCodingList;
begin
  if FSubtypeList = nil then
    FSubtypeList := TFhirCodingList.Create;
  result := FSubtypeList;
end;

Function TFhirAuditEventEvent.GetHasSubtypeList : boolean;
begin
  result := (FSubtypeList <> nil) and (FSubtypeList.count > 0);
end;

Procedure TFhirAuditEventEvent.SetAction(value : TFhirEnum);
begin
  FAction.free;
  FAction := value;
end;

Function TFhirAuditEventEvent.GetActionST : TFhirAuditEventAction;
begin
  if FAction = nil then
    result := TFhirAuditEventAction(0)
  else
    result := TFhirAuditEventAction(StringArrayIndexOfSensitive(CODES_TFhirAuditEventAction, FAction.value));
end;

Procedure TFhirAuditEventEvent.SetActionST(value : TFhirAuditEventAction);
begin
  if ord(value) = 0 then
    ActionElement := nil
  else
    ActionElement := TFhirEnum.create(CODES_TFhirAuditEventAction[value]);
end;

Procedure TFhirAuditEventEvent.SetDateTime(value : TFhirInstant);
begin
  FDateTime.free;
  FDateTime := value;
end;

Function TFhirAuditEventEvent.GetDateTimeST : TDateAndTime;
begin
  if FDateTime = nil then
    result := nil
  else
    result := FDateTime.value;
end;

Procedure TFhirAuditEventEvent.SetDateTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateTime = nil then
      FDateTime := TFhirInstant.create;
    FDateTime.value := value
  end
  else if FDateTime <> nil then
    FDateTime.value := nil;
end;

Procedure TFhirAuditEventEvent.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirAuditEventEvent.GetOutcomeST : TFhirAuditEventOutcome;
begin
  if FOutcome = nil then
    result := TFhirAuditEventOutcome(0)
  else
    result := TFhirAuditEventOutcome(StringArrayIndexOfSensitive(CODES_TFhirAuditEventOutcome, FOutcome.value));
end;

Procedure TFhirAuditEventEvent.SetOutcomeST(value : TFhirAuditEventOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirAuditEventOutcome[value]);
end;

Procedure TFhirAuditEventEvent.SetOutcomeDesc(value : TFhirString);
begin
  FOutcomeDesc.free;
  FOutcomeDesc := value;
end;

Function TFhirAuditEventEvent.GetOutcomeDescST : String;
begin
  if FOutcomeDesc = nil then
    result := ''
  else
    result := FOutcomeDesc.value;
end;

Procedure TFhirAuditEventEvent.SetOutcomeDescST(value : String);
begin
  if value <> '' then
  begin
    if FOutcomeDesc = nil then
      FOutcomeDesc := TFhirString.create;
    FOutcomeDesc.value := value
  end
  else if FOutcomeDesc <> nil then
    FOutcomeDesc.value := '';
end;

Function TFhirAuditEventEvent.GetPurposeOfEventList : TFhirCodingList;
begin
  if FPurposeOfEventList = nil then
    FPurposeOfEventList := TFhirCodingList.Create;
  result := FPurposeOfEventList;
end;

Function TFhirAuditEventEvent.GetHasPurposeOfEventList : boolean;
begin
  result := (FPurposeOfEventList <> nil) and (FPurposeOfEventList.count > 0);
end;


{ TFhirAuditEventEventListEnumerator }

Constructor TFhirAuditEventEventListEnumerator.Create(list : TFhirAuditEventEventList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventEventListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventEventListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventEventListEnumerator.GetCurrent : TFhirAuditEventEvent;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventEventList }
procedure TFhirAuditEventEventList.AddItem(value: TFhirAuditEventEvent);
begin
  assert(value.ClassName = 'TFhirAuditEventEvent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventEvent');
  add(value);
end;


function TFhirAuditEventEventList.Append: TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventEventList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventEventList.GetEnumerator : TFhirAuditEventEventListEnumerator;
begin
  result := TFhirAuditEventEventListEnumerator.Create(self.link);
end;

function TFhirAuditEventEventList.Clone: TFhirAuditEventEventList;
begin
  result := TFhirAuditEventEventList(inherited Clone);
end;

function TFhirAuditEventEventList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventEventList.GetItemN(index: Integer): TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent(ObjectByIndex[index]);
end;

function TFhirAuditEventEventList.IndexOf(value: TFhirAuditEventEvent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventEventList.Insert(index: Integer): TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventEventList.InsertItem(index: Integer; value: TFhirAuditEventEvent);
begin
  assert(value is TFhirAuditEventEvent);
  Inherited Insert(index, value);
end;

function TFhirAuditEventEventList.Item(index: Integer): TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent(ObjectByIndex[index]);
end;

function TFhirAuditEventEventList.Link: TFhirAuditEventEventList;
begin
  result := TFhirAuditEventEventList(inherited Link);
end;

procedure TFhirAuditEventEventList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventEventList.SetItemByIndex(index: Integer; value: TFhirAuditEventEvent);
begin
  assert(value is TFhirAuditEventEvent);
  FhirAuditEventEvents[index] := value;
end;

procedure TFhirAuditEventEventList.SetItemN(index: Integer; value: TFhirAuditEventEvent);
begin
  assert(value is TFhirAuditEventEvent);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventParticipant }

constructor TFhirAuditEventParticipant.Create;
begin
  inherited;
end;

destructor TFhirAuditEventParticipant.Destroy;
begin
  FRoleList.Free;
  FReference.free;
  FUserId.free;
  FAltId.free;
  FName.free;
  FRequestor.free;
  FLocation.free;
  FPolicyList.Free;
  FMedia.free;
  FNetwork.free;
  FPurposeOfUseList.Free;
  inherited;
end;

procedure TFhirAuditEventParticipant.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirAuditEventParticipant(oSource).FRoleList = nil) then
  begin
    FRoleList.free;
    FRoleList := nil;
  end
  else
  begin
    FRoleList := TFhirCodeableConceptList.Create;
    FRoleList.Assign(TFhirAuditEventParticipant(oSource).FRoleList);
  end;
  reference := TFhirAuditEventParticipant(oSource).reference.Clone;
  userId := TFhirAuditEventParticipant(oSource).userId.Clone;
  altIdElement := TFhirAuditEventParticipant(oSource).altIdElement.Clone;
  nameElement := TFhirAuditEventParticipant(oSource).nameElement.Clone;
  requestorElement := TFhirAuditEventParticipant(oSource).requestorElement.Clone;
  location := TFhirAuditEventParticipant(oSource).location.Clone;
  if (TFhirAuditEventParticipant(oSource).FPolicyList = nil) then
  begin
    FPolicyList.free;
    FPolicyList := nil;
  end
  else
  begin
    FPolicyList := TFhirUriList.Create;
    FPolicyList.Assign(TFhirAuditEventParticipant(oSource).FPolicyList);
  end;
  media := TFhirAuditEventParticipant(oSource).media.Clone;
  network := TFhirAuditEventParticipant(oSource).network.Clone;
  if (TFhirAuditEventParticipant(oSource).FPurposeOfUseList = nil) then
  begin
    FPurposeOfUseList.free;
    FPurposeOfUseList := nil;
  end
  else
  begin
    FPurposeOfUseList := TFhirCodingList.Create;
    FPurposeOfUseList.Assign(TFhirAuditEventParticipant(oSource).FPurposeOfUseList);
  end;
end;

procedure TFhirAuditEventParticipant.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'role') Then
    list.addAll(FRoleList);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'userId') Then
     list.add(FUserId.Link);
  if (child_name = 'altId') Then
     list.add(FAltId.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'requestor') Then
     list.add(FRequestor.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'policy') Then
    list.addAll(FPolicyList);
  if (child_name = 'media') Then
     list.add(FMedia.Link);
  if (child_name = 'network') Then
     list.add(FNetwork.Link);
  if (child_name = 'purposeOfUse') Then
    list.addAll(FPurposeOfUseList);
end;

procedure TFhirAuditEventParticipant.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRoleList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reference', 'Reference(Practitioner|Organization|Device|Patient|RelatedPerson)', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'userId', 'Identifier', FUserId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'altId', 'string', FAltId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestor', 'boolean', FRequestor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'policy', 'uri', FPolicyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'media', 'Coding', FMedia.Link));{2}
  oList.add(TFHIRProperty.create(self, 'network', '', FNetwork.Link));{2}
  oList.add(TFHIRProperty.create(self, 'purposeOfUse', 'Coding', FPurposeOfUseList.Link)){3};
end;

procedure TFhirAuditEventParticipant.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'role') then RoleList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'reference') then Reference := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'userId') then UserId := propValue as TFhirIdentifier{4b}
  else if (propName = 'altId') then AltIdElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'requestor') then RequestorElement := propValue as TFhirBoolean{5a}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'policy') then PolicyList.add(propValue as TFhirUri){2}
  else if (propName = 'media') then Media := propValue as TFhirCoding{4b}
  else if (propName = 'network') then Network := propValue as TFhirAuditEventParticipantNetwork{4b}
  else if (propName = 'purposeOfUse') then PurposeOfUseList.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirAuditEventParticipant.FhirType : string;
begin
  result := 'participant';
end;

function TFhirAuditEventParticipant.Link : TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant(inherited Link);
end;

function TFhirAuditEventParticipant.Clone : TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant(inherited Clone);
end;

function TFhirAuditEventParticipant.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventParticipant;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventParticipant)) then
    result := false
  else
  begin
    o := TFhirAuditEventParticipant(other);
    result := compareDeep(roleList, o.roleList, true) and compareDeep(referenceElement, o.referenceElement, true) and 
      compareDeep(userIdElement, o.userIdElement, true) and compareDeep(altIdElement, o.altIdElement, true) and 
      compareDeep(nameElement, o.nameElement, true) and compareDeep(requestorElement, o.requestorElement, true) and 
      compareDeep(locationElement, o.locationElement, true) and compareDeep(policyList, o.policyList, true) and 
      compareDeep(mediaElement, o.mediaElement, true) and compareDeep(networkElement, o.networkElement, true) and 
      compareDeep(purposeOfUseList, o.purposeOfUseList, true);
  end;
end;

function TFhirAuditEventParticipant.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventParticipant;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventParticipant)) then
    result := false
  else
  begin
    o := TFhirAuditEventParticipant(other);
    result := compareValues(altIdElement, o.altIdElement, true) and compareValues(nameElement, o.nameElement, true) and 
      compareValues(requestorElement, o.requestorElement, true) and compareValues(policyList, o.policyList, true);
  end;
end;

{ TFhirAuditEventParticipant }

Function TFhirAuditEventParticipant.GetRoleList : TFhirCodeableConceptList;
begin
  if FRoleList = nil then
    FRoleList := TFhirCodeableConceptList.Create;
  result := FRoleList;
end;

Function TFhirAuditEventParticipant.GetHasRoleList : boolean;
begin
  result := (FRoleList <> nil) and (FRoleList.count > 0);
end;

Procedure TFhirAuditEventParticipant.SetReference(value : TFhirReference{Resource});
begin
  FReference.free;
  FReference := value;
end;

Procedure TFhirAuditEventParticipant.SetUserId(value : TFhirIdentifier);
begin
  FUserId.free;
  FUserId := value;
end;

Procedure TFhirAuditEventParticipant.SetAltId(value : TFhirString);
begin
  FAltId.free;
  FAltId := value;
end;

Function TFhirAuditEventParticipant.GetAltIdST : String;
begin
  if FAltId = nil then
    result := ''
  else
    result := FAltId.value;
end;

Procedure TFhirAuditEventParticipant.SetAltIdST(value : String);
begin
  if value <> '' then
  begin
    if FAltId = nil then
      FAltId := TFhirString.create;
    FAltId.value := value
  end
  else if FAltId <> nil then
    FAltId.value := '';
end;

Procedure TFhirAuditEventParticipant.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirAuditEventParticipant.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirAuditEventParticipant.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirAuditEventParticipant.SetRequestor(value : TFhirBoolean);
begin
  FRequestor.free;
  FRequestor := value;
end;

Function TFhirAuditEventParticipant.GetRequestorST : Boolean;
begin
  if FRequestor = nil then
    result := false
  else
    result := FRequestor.value;
end;

Procedure TFhirAuditEventParticipant.SetRequestorST(value : Boolean);
begin
  if FRequestor = nil then
    FRequestor := TFhirBoolean.create;
  FRequestor.value := value
end;

Procedure TFhirAuditEventParticipant.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Function TFhirAuditEventParticipant.GetPolicyList : TFhirUriList;
begin
  if FPolicyList = nil then
    FPolicyList := TFhirUriList.Create;
  result := FPolicyList;
end;

Function TFhirAuditEventParticipant.GetHasPolicyList : boolean;
begin
  result := (FPolicyList <> nil) and (FPolicyList.count > 0);
end;

Procedure TFhirAuditEventParticipant.SetMedia(value : TFhirCoding);
begin
  FMedia.free;
  FMedia := value;
end;

Procedure TFhirAuditEventParticipant.SetNetwork(value : TFhirAuditEventParticipantNetwork);
begin
  FNetwork.free;
  FNetwork := value;
end;

Function TFhirAuditEventParticipant.GetPurposeOfUseList : TFhirCodingList;
begin
  if FPurposeOfUseList = nil then
    FPurposeOfUseList := TFhirCodingList.Create;
  result := FPurposeOfUseList;
end;

Function TFhirAuditEventParticipant.GetHasPurposeOfUseList : boolean;
begin
  result := (FPurposeOfUseList <> nil) and (FPurposeOfUseList.count > 0);
end;


{ TFhirAuditEventParticipantListEnumerator }

Constructor TFhirAuditEventParticipantListEnumerator.Create(list : TFhirAuditEventParticipantList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventParticipantListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventParticipantListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventParticipantListEnumerator.GetCurrent : TFhirAuditEventParticipant;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventParticipantList }
procedure TFhirAuditEventParticipantList.AddItem(value: TFhirAuditEventParticipant);
begin
  assert(value.ClassName = 'TFhirAuditEventParticipant', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventParticipant');
  add(value);
end;


function TFhirAuditEventParticipantList.Append: TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventParticipantList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventParticipantList.GetEnumerator : TFhirAuditEventParticipantListEnumerator;
begin
  result := TFhirAuditEventParticipantListEnumerator.Create(self.link);
end;

function TFhirAuditEventParticipantList.Clone: TFhirAuditEventParticipantList;
begin
  result := TFhirAuditEventParticipantList(inherited Clone);
end;

function TFhirAuditEventParticipantList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventParticipantList.GetItemN(index: Integer): TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant(ObjectByIndex[index]);
end;

function TFhirAuditEventParticipantList.IndexOf(value: TFhirAuditEventParticipant): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventParticipantList.Insert(index: Integer): TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventParticipantList.InsertItem(index: Integer; value: TFhirAuditEventParticipant);
begin
  assert(value is TFhirAuditEventParticipant);
  Inherited Insert(index, value);
end;

function TFhirAuditEventParticipantList.Item(index: Integer): TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant(ObjectByIndex[index]);
end;

function TFhirAuditEventParticipantList.Link: TFhirAuditEventParticipantList;
begin
  result := TFhirAuditEventParticipantList(inherited Link);
end;

procedure TFhirAuditEventParticipantList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventParticipantList.SetItemByIndex(index: Integer; value: TFhirAuditEventParticipant);
begin
  assert(value is TFhirAuditEventParticipant);
  FhirAuditEventParticipants[index] := value;
end;

procedure TFhirAuditEventParticipantList.SetItemN(index: Integer; value: TFhirAuditEventParticipant);
begin
  assert(value is TFhirAuditEventParticipant);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventParticipantNetwork }

constructor TFhirAuditEventParticipantNetwork.Create;
begin
  inherited;
end;

destructor TFhirAuditEventParticipantNetwork.Destroy;
begin
  FAddress.free;
  FType_.free;
  inherited;
end;

procedure TFhirAuditEventParticipantNetwork.Assign(oSource : TAdvObject);
begin
  inherited;
  addressElement := TFhirAuditEventParticipantNetwork(oSource).addressElement.Clone;
  FType_ := TFhirAuditEventParticipantNetwork(oSource).FType_.Link;
end;

procedure TFhirAuditEventParticipantNetwork.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'address') Then
     list.add(FAddress.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
end;

procedure TFhirAuditEventParticipantNetwork.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'address', 'string', FAddress.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
end;

procedure TFhirAuditEventParticipantNetwork.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'address') then AddressElement := propValue as TFhirString{5a}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else inherited;
end;

function TFhirAuditEventParticipantNetwork.FhirType : string;
begin
  result := 'network';
end;

function TFhirAuditEventParticipantNetwork.Link : TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork(inherited Link);
end;

function TFhirAuditEventParticipantNetwork.Clone : TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork(inherited Clone);
end;

function TFhirAuditEventParticipantNetwork.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventParticipantNetwork;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventParticipantNetwork)) then
    result := false
  else
  begin
    o := TFhirAuditEventParticipantNetwork(other);
    result := compareDeep(addressElement, o.addressElement, true) and compareDeep(type_Element, o.type_Element, true);
  end;
end;

function TFhirAuditEventParticipantNetwork.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventParticipantNetwork;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventParticipantNetwork)) then
    result := false
  else
  begin
    o := TFhirAuditEventParticipantNetwork(other);
    result := compareValues(addressElement, o.addressElement, true) and compareValues(type_Element, o.type_Element, true);
  end;
end;

{ TFhirAuditEventParticipantNetwork }

Procedure TFhirAuditEventParticipantNetwork.SetAddress(value : TFhirString);
begin
  FAddress.free;
  FAddress := value;
end;

Function TFhirAuditEventParticipantNetwork.GetAddressST : String;
begin
  if FAddress = nil then
    result := ''
  else
    result := FAddress.value;
end;

Procedure TFhirAuditEventParticipantNetwork.SetAddressST(value : String);
begin
  if value <> '' then
  begin
    if FAddress = nil then
      FAddress := TFhirString.create;
    FAddress.value := value
  end
  else if FAddress <> nil then
    FAddress.value := '';
end;

Procedure TFhirAuditEventParticipantNetwork.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirAuditEventParticipantNetwork.GetType_ST : TFhirNetworkType;
begin
  if FType_ = nil then
    result := TFhirNetworkType(0)
  else
    result := TFhirNetworkType(StringArrayIndexOfSensitive(CODES_TFhirNetworkType, FType_.value));
end;

Procedure TFhirAuditEventParticipantNetwork.SetType_ST(value : TFhirNetworkType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirNetworkType[value]);
end;


{ TFhirAuditEventParticipantNetworkListEnumerator }

Constructor TFhirAuditEventParticipantNetworkListEnumerator.Create(list : TFhirAuditEventParticipantNetworkList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventParticipantNetworkListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventParticipantNetworkListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventParticipantNetworkListEnumerator.GetCurrent : TFhirAuditEventParticipantNetwork;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventParticipantNetworkList }
procedure TFhirAuditEventParticipantNetworkList.AddItem(value: TFhirAuditEventParticipantNetwork);
begin
  assert(value.ClassName = 'TFhirAuditEventParticipantNetwork', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventParticipantNetwork');
  add(value);
end;


function TFhirAuditEventParticipantNetworkList.Append: TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventParticipantNetworkList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventParticipantNetworkList.GetEnumerator : TFhirAuditEventParticipantNetworkListEnumerator;
begin
  result := TFhirAuditEventParticipantNetworkListEnumerator.Create(self.link);
end;

function TFhirAuditEventParticipantNetworkList.Clone: TFhirAuditEventParticipantNetworkList;
begin
  result := TFhirAuditEventParticipantNetworkList(inherited Clone);
end;

function TFhirAuditEventParticipantNetworkList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventParticipantNetworkList.GetItemN(index: Integer): TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork(ObjectByIndex[index]);
end;

function TFhirAuditEventParticipantNetworkList.IndexOf(value: TFhirAuditEventParticipantNetwork): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventParticipantNetworkList.Insert(index: Integer): TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventParticipantNetworkList.InsertItem(index: Integer; value: TFhirAuditEventParticipantNetwork);
begin
  assert(value is TFhirAuditEventParticipantNetwork);
  Inherited Insert(index, value);
end;

function TFhirAuditEventParticipantNetworkList.Item(index: Integer): TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork(ObjectByIndex[index]);
end;

function TFhirAuditEventParticipantNetworkList.Link: TFhirAuditEventParticipantNetworkList;
begin
  result := TFhirAuditEventParticipantNetworkList(inherited Link);
end;

procedure TFhirAuditEventParticipantNetworkList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventParticipantNetworkList.SetItemByIndex(index: Integer; value: TFhirAuditEventParticipantNetwork);
begin
  assert(value is TFhirAuditEventParticipantNetwork);
  FhirAuditEventParticipantNetworks[index] := value;
end;

procedure TFhirAuditEventParticipantNetworkList.SetItemN(index: Integer; value: TFhirAuditEventParticipantNetwork);
begin
  assert(value is TFhirAuditEventParticipantNetwork);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventSource }

constructor TFhirAuditEventSource.Create;
begin
  inherited;
end;

destructor TFhirAuditEventSource.Destroy;
begin
  FSite.free;
  FIdentifier.free;
  FType_List.Free;
  inherited;
end;

procedure TFhirAuditEventSource.Assign(oSource : TAdvObject);
begin
  inherited;
  siteElement := TFhirAuditEventSource(oSource).siteElement.Clone;
  identifier := TFhirAuditEventSource(oSource).identifier.Clone;
  if (TFhirAuditEventSource(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodingList.Create;
    FType_List.Assign(TFhirAuditEventSource(oSource).FType_List);
  end;
end;

procedure TFhirAuditEventSource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'site') Then
     list.add(FSite.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'type') Then
    list.addAll(FType_List);
end;

procedure TFhirAuditEventSource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'site', 'string', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_List.Link)){3};
end;

procedure TFhirAuditEventSource.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'site') then SiteElement := propValue as TFhirString{5a}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'type') then Type_List.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirAuditEventSource.FhirType : string;
begin
  result := 'source';
end;

function TFhirAuditEventSource.Link : TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource(inherited Link);
end;

function TFhirAuditEventSource.Clone : TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource(inherited Clone);
end;

function TFhirAuditEventSource.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventSource;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventSource)) then
    result := false
  else
  begin
    o := TFhirAuditEventSource(other);
    result := compareDeep(siteElement, o.siteElement, true) and compareDeep(identifierElement, o.identifierElement, true) and 
      compareDeep(type_List, o.type_List, true);
  end;
end;

function TFhirAuditEventSource.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventSource;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventSource)) then
    result := false
  else
  begin
    o := TFhirAuditEventSource(other);
    result := compareValues(siteElement, o.siteElement, true);
  end;
end;

{ TFhirAuditEventSource }

Procedure TFhirAuditEventSource.SetSite(value : TFhirString);
begin
  FSite.free;
  FSite := value;
end;

Function TFhirAuditEventSource.GetSiteST : String;
begin
  if FSite = nil then
    result := ''
  else
    result := FSite.value;
end;

Procedure TFhirAuditEventSource.SetSiteST(value : String);
begin
  if value <> '' then
  begin
    if FSite = nil then
      FSite := TFhirString.create;
    FSite.value := value
  end
  else if FSite <> nil then
    FSite.value := '';
end;

Procedure TFhirAuditEventSource.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Function TFhirAuditEventSource.GetType_List : TFhirCodingList;
begin
  if FType_List = nil then
    FType_List := TFhirCodingList.Create;
  result := FType_List;
end;

Function TFhirAuditEventSource.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;


{ TFhirAuditEventSourceListEnumerator }

Constructor TFhirAuditEventSourceListEnumerator.Create(list : TFhirAuditEventSourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventSourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventSourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventSourceListEnumerator.GetCurrent : TFhirAuditEventSource;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventSourceList }
procedure TFhirAuditEventSourceList.AddItem(value: TFhirAuditEventSource);
begin
  assert(value.ClassName = 'TFhirAuditEventSource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventSource');
  add(value);
end;


function TFhirAuditEventSourceList.Append: TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventSourceList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventSourceList.GetEnumerator : TFhirAuditEventSourceListEnumerator;
begin
  result := TFhirAuditEventSourceListEnumerator.Create(self.link);
end;

function TFhirAuditEventSourceList.Clone: TFhirAuditEventSourceList;
begin
  result := TFhirAuditEventSourceList(inherited Clone);
end;

function TFhirAuditEventSourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventSourceList.GetItemN(index: Integer): TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource(ObjectByIndex[index]);
end;

function TFhirAuditEventSourceList.IndexOf(value: TFhirAuditEventSource): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventSourceList.Insert(index: Integer): TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventSourceList.InsertItem(index: Integer; value: TFhirAuditEventSource);
begin
  assert(value is TFhirAuditEventSource);
  Inherited Insert(index, value);
end;

function TFhirAuditEventSourceList.Item(index: Integer): TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource(ObjectByIndex[index]);
end;

function TFhirAuditEventSourceList.Link: TFhirAuditEventSourceList;
begin
  result := TFhirAuditEventSourceList(inherited Link);
end;

procedure TFhirAuditEventSourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventSourceList.SetItemByIndex(index: Integer; value: TFhirAuditEventSource);
begin
  assert(value is TFhirAuditEventSource);
  FhirAuditEventSources[index] := value;
end;

procedure TFhirAuditEventSourceList.SetItemN(index: Integer; value: TFhirAuditEventSource);
begin
  assert(value is TFhirAuditEventSource);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventObject }

constructor TFhirAuditEventObject.Create;
begin
  inherited;
end;

destructor TFhirAuditEventObject.Destroy;
begin
  FIdentifier.free;
  FReference.free;
  FType_.free;
  FRole.free;
  FLifecycle.free;
  FSecurityLabelList.Free;
  FName.free;
  FDescription.free;
  FQuery.free;
  FDetailList.Free;
  inherited;
end;

procedure TFhirAuditEventObject.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirAuditEventObject(oSource).identifier.Clone;
  reference := TFhirAuditEventObject(oSource).reference.Clone;
  type_ := TFhirAuditEventObject(oSource).type_.Clone;
  role := TFhirAuditEventObject(oSource).role.Clone;
  lifecycle := TFhirAuditEventObject(oSource).lifecycle.Clone;
  if (TFhirAuditEventObject(oSource).FSecurityLabelList = nil) then
  begin
    FSecurityLabelList.free;
    FSecurityLabelList := nil;
  end
  else
  begin
    FSecurityLabelList := TFhirCodingList.Create;
    FSecurityLabelList.Assign(TFhirAuditEventObject(oSource).FSecurityLabelList);
  end;
  nameElement := TFhirAuditEventObject(oSource).nameElement.Clone;
  descriptionElement := TFhirAuditEventObject(oSource).descriptionElement.Clone;
  queryElement := TFhirAuditEventObject(oSource).queryElement.Clone;
  if (TFhirAuditEventObject(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirAuditEventObjectDetailList.Create;
    FDetailList.Assign(TFhirAuditEventObject(oSource).FDetailList);
  end;
end;

procedure TFhirAuditEventObject.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'role') Then
     list.add(FRole.Link);
  if (child_name = 'lifecycle') Then
     list.add(FLifecycle.Link);
  if (child_name = 'securityLabel') Then
    list.addAll(FSecurityLabelList);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'query') Then
     list.add(FQuery.Link);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirAuditEventObject.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reference', 'Reference(Any)', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'role', 'Coding', FRole.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lifecycle', 'Coding', FLifecycle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'securityLabel', 'Coding', FSecurityLabelList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'query', 'base64Binary', FQuery.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetailList.Link)){3};
end;

procedure TFhirAuditEventObject.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'reference') then Reference := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'role') then Role := propValue as TFhirCoding{4b}
  else if (propName = 'lifecycle') then Lifecycle := propValue as TFhirCoding{4b}
  else if (propName = 'securityLabel') then SecurityLabelList.add(propValue as TFhirCoding){2}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'query') then QueryElement := propValue as TFhirBase64Binary{5a}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirAuditEventObjectDetail){2}
  else inherited;
end;

function TFhirAuditEventObject.FhirType : string;
begin
  result := 'object';
end;

function TFhirAuditEventObject.Link : TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject(inherited Link);
end;

function TFhirAuditEventObject.Clone : TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject(inherited Clone);
end;

function TFhirAuditEventObject.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventObject;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventObject)) then
    result := false
  else
  begin
    o := TFhirAuditEventObject(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(referenceElement, o.referenceElement, true) and 
      compareDeep(type_Element, o.type_Element, true) and compareDeep(roleElement, o.roleElement, true) and 
      compareDeep(lifecycleElement, o.lifecycleElement, true) and compareDeep(securityLabelList, o.securityLabelList, true) and 
      compareDeep(nameElement, o.nameElement, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(queryElement, o.queryElement, true) and compareDeep(detailList, o.detailList, true);
  end;
end;

function TFhirAuditEventObject.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventObject;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventObject)) then
    result := false
  else
  begin
    o := TFhirAuditEventObject(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(descriptionElement, o.descriptionElement, true) and 
      compareValues(queryElement, o.queryElement, true);
  end;
end;

{ TFhirAuditEventObject }

Procedure TFhirAuditEventObject.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirAuditEventObject.SetReference(value : TFhirReference{TFhirReference});
begin
  FReference.free;
  FReference := value;
end;

Procedure TFhirAuditEventObject.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirAuditEventObject.SetRole(value : TFhirCoding);
begin
  FRole.free;
  FRole := value;
end;

Procedure TFhirAuditEventObject.SetLifecycle(value : TFhirCoding);
begin
  FLifecycle.free;
  FLifecycle := value;
end;

Function TFhirAuditEventObject.GetSecurityLabelList : TFhirCodingList;
begin
  if FSecurityLabelList = nil then
    FSecurityLabelList := TFhirCodingList.Create;
  result := FSecurityLabelList;
end;

Function TFhirAuditEventObject.GetHasSecurityLabelList : boolean;
begin
  result := (FSecurityLabelList <> nil) and (FSecurityLabelList.count > 0);
end;

Procedure TFhirAuditEventObject.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirAuditEventObject.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirAuditEventObject.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirAuditEventObject.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirAuditEventObject.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirAuditEventObject.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirAuditEventObject.SetQuery(value : TFhirBase64Binary);
begin
  FQuery.free;
  FQuery := value;
end;

Function TFhirAuditEventObject.GetQueryST : TBytes;
begin
  if FQuery = nil then
    result := nil
  else
    result := FQuery.value;
end;

Procedure TFhirAuditEventObject.SetQueryST(value : TBytes);
begin
  if value <> nil then
  begin
    if FQuery = nil then
      FQuery := TFhirBase64Binary.create;
    FQuery.value := value
  end
  else if FQuery <> nil then
    FQuery.value := nil;
end;

Function TFhirAuditEventObject.GetDetailList : TFhirAuditEventObjectDetailList;
begin
  if FDetailList = nil then
    FDetailList := TFhirAuditEventObjectDetailList.Create;
  result := FDetailList;
end;

Function TFhirAuditEventObject.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirAuditEventObjectListEnumerator }

Constructor TFhirAuditEventObjectListEnumerator.Create(list : TFhirAuditEventObjectList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventObjectListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventObjectListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventObjectListEnumerator.GetCurrent : TFhirAuditEventObject;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventObjectList }
procedure TFhirAuditEventObjectList.AddItem(value: TFhirAuditEventObject);
begin
  assert(value.ClassName = 'TFhirAuditEventObject', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventObject');
  add(value);
end;


function TFhirAuditEventObjectList.Append: TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventObjectList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventObjectList.GetEnumerator : TFhirAuditEventObjectListEnumerator;
begin
  result := TFhirAuditEventObjectListEnumerator.Create(self.link);
end;

function TFhirAuditEventObjectList.Clone: TFhirAuditEventObjectList;
begin
  result := TFhirAuditEventObjectList(inherited Clone);
end;

function TFhirAuditEventObjectList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventObjectList.GetItemN(index: Integer): TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject(ObjectByIndex[index]);
end;

function TFhirAuditEventObjectList.IndexOf(value: TFhirAuditEventObject): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventObjectList.Insert(index: Integer): TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventObjectList.InsertItem(index: Integer; value: TFhirAuditEventObject);
begin
  assert(value is TFhirAuditEventObject);
  Inherited Insert(index, value);
end;

function TFhirAuditEventObjectList.Item(index: Integer): TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject(ObjectByIndex[index]);
end;

function TFhirAuditEventObjectList.Link: TFhirAuditEventObjectList;
begin
  result := TFhirAuditEventObjectList(inherited Link);
end;

procedure TFhirAuditEventObjectList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventObjectList.SetItemByIndex(index: Integer; value: TFhirAuditEventObject);
begin
  assert(value is TFhirAuditEventObject);
  FhirAuditEventObjects[index] := value;
end;

procedure TFhirAuditEventObjectList.SetItemN(index: Integer; value: TFhirAuditEventObject);
begin
  assert(value is TFhirAuditEventObject);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEventObjectDetail }

constructor TFhirAuditEventObjectDetail.Create;
begin
  inherited;
end;

destructor TFhirAuditEventObjectDetail.Destroy;
begin
  FType_.free;
  FValue.free;
  inherited;
end;

procedure TFhirAuditEventObjectDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  type_Element := TFhirAuditEventObjectDetail(oSource).type_Element.Clone;
  valueElement := TFhirAuditEventObjectDetail(oSource).valueElement.Clone;
end;

procedure TFhirAuditEventObjectDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirAuditEventObjectDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'string', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'base64Binary', FValue.Link));{2}
end;

procedure TFhirAuditEventObjectDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFhirString{5a}
  else if (propName = 'value') then ValueElement := propValue as TFhirBase64Binary{5a}
  else inherited;
end;

function TFhirAuditEventObjectDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirAuditEventObjectDetail.Link : TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail(inherited Link);
end;

function TFhirAuditEventObjectDetail.Clone : TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail(inherited Clone);
end;

function TFhirAuditEventObjectDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventObjectDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirAuditEventObjectDetail)) then
    result := false
  else
  begin
    o := TFhirAuditEventObjectDetail(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirAuditEventObjectDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirAuditEventObjectDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirAuditEventObjectDetail)) then
    result := false
  else
  begin
    o := TFhirAuditEventObjectDetail(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirAuditEventObjectDetail }

Procedure TFhirAuditEventObjectDetail.SetType_(value : TFhirString);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirAuditEventObjectDetail.GetType_ST : String;
begin
  if FType_ = nil then
    result := ''
  else
    result := FType_.value;
end;

Procedure TFhirAuditEventObjectDetail.SetType_ST(value : String);
begin
  if value <> '' then
  begin
    if FType_ = nil then
      FType_ := TFhirString.create;
    FType_.value := value
  end
  else if FType_ <> nil then
    FType_.value := '';
end;

Procedure TFhirAuditEventObjectDetail.SetValue(value : TFhirBase64Binary);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirAuditEventObjectDetail.GetValueST : TBytes;
begin
  if FValue = nil then
    result := nil
  else
    result := FValue.value;
end;

Procedure TFhirAuditEventObjectDetail.SetValueST(value : TBytes);
begin
  if value <> nil then
  begin
    if FValue = nil then
      FValue := TFhirBase64Binary.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := nil;
end;


{ TFhirAuditEventObjectDetailListEnumerator }

Constructor TFhirAuditEventObjectDetailListEnumerator.Create(list : TFhirAuditEventObjectDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventObjectDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventObjectDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventObjectDetailListEnumerator.GetCurrent : TFhirAuditEventObjectDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventObjectDetailList }
procedure TFhirAuditEventObjectDetailList.AddItem(value: TFhirAuditEventObjectDetail);
begin
  assert(value.ClassName = 'TFhirAuditEventObjectDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEventObjectDetail');
  add(value);
end;


function TFhirAuditEventObjectDetailList.Append: TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventObjectDetailList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventObjectDetailList.GetEnumerator : TFhirAuditEventObjectDetailListEnumerator;
begin
  result := TFhirAuditEventObjectDetailListEnumerator.Create(self.link);
end;

function TFhirAuditEventObjectDetailList.Clone: TFhirAuditEventObjectDetailList;
begin
  result := TFhirAuditEventObjectDetailList(inherited Clone);
end;

function TFhirAuditEventObjectDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventObjectDetailList.GetItemN(index: Integer): TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail(ObjectByIndex[index]);
end;

function TFhirAuditEventObjectDetailList.IndexOf(value: TFhirAuditEventObjectDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventObjectDetailList.Insert(index: Integer): TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventObjectDetailList.InsertItem(index: Integer; value: TFhirAuditEventObjectDetail);
begin
  assert(value is TFhirAuditEventObjectDetail);
  Inherited Insert(index, value);
end;

function TFhirAuditEventObjectDetailList.Item(index: Integer): TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail(ObjectByIndex[index]);
end;

function TFhirAuditEventObjectDetailList.Link: TFhirAuditEventObjectDetailList;
begin
  result := TFhirAuditEventObjectDetailList(inherited Link);
end;

procedure TFhirAuditEventObjectDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventObjectDetailList.SetItemByIndex(index: Integer; value: TFhirAuditEventObjectDetail);
begin
  assert(value is TFhirAuditEventObjectDetail);
  FhirAuditEventObjectDetails[index] := value;
end;

procedure TFhirAuditEventObjectDetailList.SetItemN(index: Integer; value: TFhirAuditEventObjectDetail);
begin
  assert(value is TFhirAuditEventObjectDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirAuditEvent }

constructor TFhirAuditEvent.Create;
begin
  inherited;
end;

destructor TFhirAuditEvent.Destroy;
begin
  FEvent.free;
  FParticipantList.Free;
  FSource.free;
  FObject_List.Free;
  inherited;
end;

function TFhirAuditEvent.GetResourceType : TFhirResourceType;
begin
  result := frtAuditEvent;
end;

procedure TFhirAuditEvent.Assign(oSource : TAdvObject);
begin
  inherited;
  event := TFhirAuditEvent(oSource).event.Clone;
  if (TFhirAuditEvent(oSource).FParticipantList = nil) then
  begin
    FParticipantList.free;
    FParticipantList := nil;
  end
  else
  begin
    FParticipantList := TFhirAuditEventParticipantList.Create;
    FParticipantList.Assign(TFhirAuditEvent(oSource).FParticipantList);
  end;
  source := TFhirAuditEvent(oSource).source.Clone;
  if (TFhirAuditEvent(oSource).FObject_List = nil) then
  begin
    FObject_List.free;
    FObject_List := nil;
  end
  else
  begin
    FObject_List := TFhirAuditEventObjectList.Create;
    FObject_List.Assign(TFhirAuditEvent(oSource).FObject_List);
  end;
end;

procedure TFhirAuditEvent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'event') Then
     list.add(FEvent.Link);
  if (child_name = 'participant') Then
    list.addAll(FParticipantList);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'object') Then
    list.addAll(FObject_List);
end;

procedure TFhirAuditEvent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'event', '', FEvent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'participant', '', FParticipantList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'source', '', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'object', '', FObject_List.Link)){3};
end;

procedure TFhirAuditEvent.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'event') then Event := propValue as TFhirAuditEventEvent{4b}
  else if (propName = 'participant') then ParticipantList.add(propValue as TFhirAuditEventParticipant){2}
  else if (propName = 'source') then Source := propValue as TFhirAuditEventSource{4b}
  else if (propName = 'object') then Object_List.add(propValue as TFhirAuditEventObject){2}
  else inherited;
end;

function TFhirAuditEvent.FhirType : string;
begin
  result := 'AuditEvent';
end;

function TFhirAuditEvent.Link : TFhirAuditEvent;
begin
  result := TFhirAuditEvent(inherited Link);
end;

function TFhirAuditEvent.Clone : TFhirAuditEvent;
begin
  result := TFhirAuditEvent(inherited Clone);
end;

{ TFhirAuditEvent }

Procedure TFhirAuditEvent.SetEvent(value : TFhirAuditEventEvent);
begin
  FEvent.free;
  FEvent := value;
end;

Function TFhirAuditEvent.GetParticipantList : TFhirAuditEventParticipantList;
begin
  if FParticipantList = nil then
    FParticipantList := TFhirAuditEventParticipantList.Create;
  result := FParticipantList;
end;

Function TFhirAuditEvent.GetHasParticipantList : boolean;
begin
  result := (FParticipantList <> nil) and (FParticipantList.count > 0);
end;

Procedure TFhirAuditEvent.SetSource(value : TFhirAuditEventSource);
begin
  FSource.free;
  FSource := value;
end;

Function TFhirAuditEvent.GetObject_List : TFhirAuditEventObjectList;
begin
  if FObject_List = nil then
    FObject_List := TFhirAuditEventObjectList.Create;
  result := FObject_List;
end;

Function TFhirAuditEvent.GetHasObject_List : boolean;
begin
  result := (FObject_List <> nil) and (FObject_List.count > 0);
end;


{ TFhirAuditEventListEnumerator }

Constructor TFhirAuditEventListEnumerator.Create(list : TFhirAuditEventList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirAuditEventListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirAuditEventListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirAuditEventListEnumerator.GetCurrent : TFhirAuditEvent;
begin
  Result := FList[FIndex];
end;


{ TFhirAuditEventList }
procedure TFhirAuditEventList.AddItem(value: TFhirAuditEvent);
begin
  assert(value.ClassName = 'TFhirAuditEvent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirAuditEvent');
  add(value);
end;


function TFhirAuditEventList.Append: TFhirAuditEvent;
begin
  result := TFhirAuditEvent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventList.ClearItems;
begin
  Clear;
end;

function TFhirAuditEventList.GetEnumerator : TFhirAuditEventListEnumerator;
begin
  result := TFhirAuditEventListEnumerator.Create(self.link);
end;

function TFhirAuditEventList.Clone: TFhirAuditEventList;
begin
  result := TFhirAuditEventList(inherited Clone);
end;

function TFhirAuditEventList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirAuditEventList.GetItemN(index: Integer): TFhirAuditEvent;
begin
  result := TFhirAuditEvent(ObjectByIndex[index]);
end;

function TFhirAuditEventList.IndexOf(value: TFhirAuditEvent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirAuditEventList.Insert(index: Integer): TFhirAuditEvent;
begin
  result := TFhirAuditEvent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirAuditEventList.InsertItem(index: Integer; value: TFhirAuditEvent);
begin
  assert(value is TFhirAuditEvent);
  Inherited Insert(index, value);
end;

function TFhirAuditEventList.Item(index: Integer): TFhirAuditEvent;
begin
  result := TFhirAuditEvent(ObjectByIndex[index]);
end;

function TFhirAuditEventList.Link: TFhirAuditEventList;
begin
  result := TFhirAuditEventList(inherited Link);
end;

procedure TFhirAuditEventList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirAuditEventList.SetItemByIndex(index: Integer; value: TFhirAuditEvent);
begin
  assert(value is TFhirAuditEvent);
  FhirAuditEvents[index] := value;
end;

procedure TFhirAuditEventList.SetItemN(index: Integer; value: TFhirAuditEvent);
begin
  assert(value is TFhirAuditEvent);
  ObjectByIndex[index] := value;
end;

{ TFhirBasic }

constructor TFhirBasic.Create;
begin
  inherited;
end;

destructor TFhirBasic.Destroy;
begin
  FIdentifierList.Free;
  FCode.free;
  FSubject.free;
  FAuthor.free;
  FCreated.free;
  inherited;
end;

function TFhirBasic.GetResourceType : TFhirResourceType;
begin
  result := frtBasic;
end;

procedure TFhirBasic.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirBasic(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirBasic(oSource).FIdentifierList);
  end;
  code := TFhirBasic(oSource).code.Clone;
  subject := TFhirBasic(oSource).subject.Clone;
  author := TFhirBasic(oSource).author.Clone;
  createdElement := TFhirBasic(oSource).createdElement.Clone;
end;

procedure TFhirBasic.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
end;

procedure TFhirBasic.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Patient|RelatedPerson)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'date', FCreated.Link));{2}
end;

procedure TFhirBasic.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDate{5a}
  else inherited;
end;

function TFhirBasic.FhirType : string;
begin
  result := 'Basic';
end;

function TFhirBasic.Link : TFhirBasic;
begin
  result := TFhirBasic(inherited Link);
end;

function TFhirBasic.Clone : TFhirBasic;
begin
  result := TFhirBasic(inherited Clone);
end;

{ TFhirBasic }

Function TFhirBasic.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirBasic.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirBasic.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirBasic.SetSubject(value : TFhirReference{TFhirReference});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirBasic.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirBasic.SetCreated(value : TFhirDate);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirBasic.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirBasic.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDate.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;


{ TFhirBasicListEnumerator }

Constructor TFhirBasicListEnumerator.Create(list : TFhirBasicList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBasicListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBasicListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBasicListEnumerator.GetCurrent : TFhirBasic;
begin
  Result := FList[FIndex];
end;


{ TFhirBasicList }
procedure TFhirBasicList.AddItem(value: TFhirBasic);
begin
  assert(value.ClassName = 'TFhirBasic', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBasic');
  add(value);
end;


function TFhirBasicList.Append: TFhirBasic;
begin
  result := TFhirBasic.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBasicList.ClearItems;
begin
  Clear;
end;

function TFhirBasicList.GetEnumerator : TFhirBasicListEnumerator;
begin
  result := TFhirBasicListEnumerator.Create(self.link);
end;

function TFhirBasicList.Clone: TFhirBasicList;
begin
  result := TFhirBasicList(inherited Clone);
end;

function TFhirBasicList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBasicList.GetItemN(index: Integer): TFhirBasic;
begin
  result := TFhirBasic(ObjectByIndex[index]);
end;

function TFhirBasicList.IndexOf(value: TFhirBasic): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBasicList.Insert(index: Integer): TFhirBasic;
begin
  result := TFhirBasic.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBasicList.InsertItem(index: Integer; value: TFhirBasic);
begin
  assert(value is TFhirBasic);
  Inherited Insert(index, value);
end;

function TFhirBasicList.Item(index: Integer): TFhirBasic;
begin
  result := TFhirBasic(ObjectByIndex[index]);
end;

function TFhirBasicList.Link: TFhirBasicList;
begin
  result := TFhirBasicList(inherited Link);
end;

procedure TFhirBasicList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBasicList.SetItemByIndex(index: Integer; value: TFhirBasic);
begin
  assert(value is TFhirBasic);
  FhirBasics[index] := value;
end;

procedure TFhirBasicList.SetItemN(index: Integer; value: TFhirBasic);
begin
  assert(value is TFhirBasic);
  ObjectByIndex[index] := value;
end;

{ TFhirBinary }

constructor TFhirBinary.Create;
begin
  inherited;
end;

destructor TFhirBinary.Destroy;
begin
  FContentType.free;
  FContent.free;
  inherited;
end;

function TFhirBinary.GetResourceType : TFhirResourceType;
begin
  result := frtBinary;
end;

procedure TFhirBinary.Assign(oSource : TAdvObject);
begin
  inherited;
  contentTypeElement := TFhirBinary(oSource).contentTypeElement.Clone;
  contentElement := TFhirBinary(oSource).contentElement.Clone;
end;

procedure TFhirBinary.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'contentType') Then
     list.add(FContentType.Link);
  if (child_name = 'content') Then
     list.add(FContent.Link);
end;

procedure TFhirBinary.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'contentType', 'code', FContentType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'content', 'base64Binary', FContent.Link));{2}
end;

procedure TFhirBinary.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'contentType') then
    if propValue is TFHIRCode then
      ContentTypeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      ContentTypeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'content') then ContentElement := propValue as TFhirBase64Binary{5a}
  else inherited;
end;

function TFhirBinary.FhirType : string;
begin
  result := 'Binary';
end;

function TFhirBinary.Link : TFhirBinary;
begin
  result := TFhirBinary(inherited Link);
end;

function TFhirBinary.Clone : TFhirBinary;
begin
  result := TFhirBinary(inherited Clone);
end;

{ TFhirBinary }

Procedure TFhirBinary.SetContentType(value : TFhirCode);
begin
  FContentType.free;
  FContentType := value;
end;

Function TFhirBinary.GetContentTypeST : String;
begin
  if FContentType = nil then
    result := ''
  else
    result := FContentType.value;
end;

Procedure TFhirBinary.SetContentTypeST(value : String);
begin
  if value <> '' then
  begin
    if FContentType = nil then
      FContentType := TFhirCode.create;
    FContentType.value := value
  end
  else if FContentType <> nil then
    FContentType.value := '';
end;

Procedure TFhirBinary.SetContent(value : TFhirBase64Binary);
begin
  FContent.free;
  FContent := value;
end;

Function TFhirBinary.GetContentST : TBytes;
begin
  if FContent = nil then
    result := nil
  else
    result := FContent.value;
end;

Procedure TFhirBinary.SetContentST(value : TBytes);
begin
  if value <> nil then
  begin
    if FContent = nil then
      FContent := TFhirBase64Binary.create;
    FContent.value := value
  end
  else if FContent <> nil then
    FContent.value := nil;
end;


{ TFhirBinaryListEnumerator }

Constructor TFhirBinaryListEnumerator.Create(list : TFhirBinaryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBinaryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBinaryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBinaryListEnumerator.GetCurrent : TFhirBinary;
begin
  Result := FList[FIndex];
end;


{ TFhirBinaryList }
procedure TFhirBinaryList.AddItem(value: TFhirBinary);
begin
  assert(value.ClassName = 'TFhirBinary', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBinary');
  add(value);
end;


function TFhirBinaryList.Append: TFhirBinary;
begin
  result := TFhirBinary.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBinaryList.ClearItems;
begin
  Clear;
end;

function TFhirBinaryList.GetEnumerator : TFhirBinaryListEnumerator;
begin
  result := TFhirBinaryListEnumerator.Create(self.link);
end;

function TFhirBinaryList.Clone: TFhirBinaryList;
begin
  result := TFhirBinaryList(inherited Clone);
end;

function TFhirBinaryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBinaryList.GetItemN(index: Integer): TFhirBinary;
begin
  result := TFhirBinary(ObjectByIndex[index]);
end;

function TFhirBinaryList.IndexOf(value: TFhirBinary): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBinaryList.Insert(index: Integer): TFhirBinary;
begin
  result := TFhirBinary.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBinaryList.InsertItem(index: Integer; value: TFhirBinary);
begin
  assert(value is TFhirBinary);
  Inherited Insert(index, value);
end;

function TFhirBinaryList.Item(index: Integer): TFhirBinary;
begin
  result := TFhirBinary(ObjectByIndex[index]);
end;

function TFhirBinaryList.Link: TFhirBinaryList;
begin
  result := TFhirBinaryList(inherited Link);
end;

procedure TFhirBinaryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBinaryList.SetItemByIndex(index: Integer; value: TFhirBinary);
begin
  assert(value is TFhirBinary);
  FhirBinaries[index] := value;
end;

procedure TFhirBinaryList.SetItemN(index: Integer; value: TFhirBinary);
begin
  assert(value is TFhirBinary);
  ObjectByIndex[index] := value;
end;

{ TFhirBodySite }

constructor TFhirBodySite.Create;
begin
  inherited;
end;

destructor TFhirBodySite.Destroy;
begin
  FPatient.free;
  FIdentifierList.Free;
  FCode.free;
  FModifierList.Free;
  FDescription.free;
  FImageList.Free;
  inherited;
end;

function TFhirBodySite.GetResourceType : TFhirResourceType;
begin
  result := frtBodySite;
end;

procedure TFhirBodySite.Assign(oSource : TAdvObject);
begin
  inherited;
  patient := TFhirBodySite(oSource).patient.Clone;
  if (TFhirBodySite(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirBodySite(oSource).FIdentifierList);
  end;
  code := TFhirBodySite(oSource).code.Clone;
  if (TFhirBodySite(oSource).FModifierList = nil) then
  begin
    FModifierList.free;
    FModifierList := nil;
  end
  else
  begin
    FModifierList := TFhirCodeableConceptList.Create;
    FModifierList.Assign(TFhirBodySite(oSource).FModifierList);
  end;
  descriptionElement := TFhirBodySite(oSource).descriptionElement.Clone;
  if (TFhirBodySite(oSource).FImageList = nil) then
  begin
    FImageList.free;
    FImageList := nil;
  end
  else
  begin
    FImageList := TFhirAttachmentList.Create;
    FImageList.Assign(TFhirBodySite(oSource).FImageList);
  end;
end;

procedure TFhirBodySite.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'modifier') Then
    list.addAll(FModifierList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'image') Then
    list.addAll(FImageList);
end;

procedure TFhirBodySite.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'modifier', 'CodeableConcept', FModifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'image', 'Attachment', FImageList.Link)){3};
end;

procedure TFhirBodySite.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'modifier') then ModifierList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'image') then ImageList.add(propValue as TFhirAttachment){2}
  else inherited;
end;

function TFhirBodySite.FhirType : string;
begin
  result := 'BodySite';
end;

function TFhirBodySite.Link : TFhirBodySite;
begin
  result := TFhirBodySite(inherited Link);
end;

function TFhirBodySite.Clone : TFhirBodySite;
begin
  result := TFhirBodySite(inherited Clone);
end;

{ TFhirBodySite }

Procedure TFhirBodySite.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Function TFhirBodySite.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirBodySite.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirBodySite.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirBodySite.GetModifierList : TFhirCodeableConceptList;
begin
  if FModifierList = nil then
    FModifierList := TFhirCodeableConceptList.Create;
  result := FModifierList;
end;

Function TFhirBodySite.GetHasModifierList : boolean;
begin
  result := (FModifierList <> nil) and (FModifierList.count > 0);
end;

Procedure TFhirBodySite.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirBodySite.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirBodySite.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirBodySite.GetImageList : TFhirAttachmentList;
begin
  if FImageList = nil then
    FImageList := TFhirAttachmentList.Create;
  result := FImageList;
end;

Function TFhirBodySite.GetHasImageList : boolean;
begin
  result := (FImageList <> nil) and (FImageList.count > 0);
end;


{ TFhirBodySiteListEnumerator }

Constructor TFhirBodySiteListEnumerator.Create(list : TFhirBodySiteList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBodySiteListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBodySiteListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBodySiteListEnumerator.GetCurrent : TFhirBodySite;
begin
  Result := FList[FIndex];
end;


{ TFhirBodySiteList }
procedure TFhirBodySiteList.AddItem(value: TFhirBodySite);
begin
  assert(value.ClassName = 'TFhirBodySite', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBodySite');
  add(value);
end;


function TFhirBodySiteList.Append: TFhirBodySite;
begin
  result := TFhirBodySite.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBodySiteList.ClearItems;
begin
  Clear;
end;

function TFhirBodySiteList.GetEnumerator : TFhirBodySiteListEnumerator;
begin
  result := TFhirBodySiteListEnumerator.Create(self.link);
end;

function TFhirBodySiteList.Clone: TFhirBodySiteList;
begin
  result := TFhirBodySiteList(inherited Clone);
end;

function TFhirBodySiteList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBodySiteList.GetItemN(index: Integer): TFhirBodySite;
begin
  result := TFhirBodySite(ObjectByIndex[index]);
end;

function TFhirBodySiteList.IndexOf(value: TFhirBodySite): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBodySiteList.Insert(index: Integer): TFhirBodySite;
begin
  result := TFhirBodySite.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBodySiteList.InsertItem(index: Integer; value: TFhirBodySite);
begin
  assert(value is TFhirBodySite);
  Inherited Insert(index, value);
end;

function TFhirBodySiteList.Item(index: Integer): TFhirBodySite;
begin
  result := TFhirBodySite(ObjectByIndex[index]);
end;

function TFhirBodySiteList.Link: TFhirBodySiteList;
begin
  result := TFhirBodySiteList(inherited Link);
end;

procedure TFhirBodySiteList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBodySiteList.SetItemByIndex(index: Integer; value: TFhirBodySite);
begin
  assert(value is TFhirBodySite);
  FhirBodySites[index] := value;
end;

procedure TFhirBodySiteList.SetItemN(index: Integer; value: TFhirBodySite);
begin
  assert(value is TFhirBodySite);
  ObjectByIndex[index] := value;
end;

{ TFhirBundleLink }

constructor TFhirBundleLink.Create;
begin
  inherited;
end;

destructor TFhirBundleLink.Destroy;
begin
  FRelation.free;
  FUrl.free;
  inherited;
end;

procedure TFhirBundleLink.Assign(oSource : TAdvObject);
begin
  inherited;
  relationElement := TFhirBundleLink(oSource).relationElement.Clone;
  urlElement := TFhirBundleLink(oSource).urlElement.Clone;
end;

procedure TFhirBundleLink.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'relation') Then
     list.add(FRelation.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
end;

procedure TFhirBundleLink.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'relation', 'string', FRelation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
end;

procedure TFhirBundleLink.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'relation') then RelationElement := propValue as TFhirString{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirBundleLink.FhirType : string;
begin
  result := 'link';
end;

function TFhirBundleLink.Link : TFhirBundleLink;
begin
  result := TFhirBundleLink(inherited Link);
end;

function TFhirBundleLink.Clone : TFhirBundleLink;
begin
  result := TFhirBundleLink(inherited Clone);
end;

function TFhirBundleLink.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleLink;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirBundleLink)) then
    result := false
  else
  begin
    o := TFhirBundleLink(other);
    result := compareDeep(relationElement, o.relationElement, true) and compareDeep(urlElement, o.urlElement, true);
  end;
end;

function TFhirBundleLink.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleLink;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirBundleLink)) then
    result := false
  else
  begin
    o := TFhirBundleLink(other);
    result := compareValues(relationElement, o.relationElement, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirBundleLink }

Procedure TFhirBundleLink.SetRelation(value : TFhirString);
begin
  FRelation.free;
  FRelation := value;
end;

Function TFhirBundleLink.GetRelationST : String;
begin
  if FRelation = nil then
    result := ''
  else
    result := FRelation.value;
end;

Procedure TFhirBundleLink.SetRelationST(value : String);
begin
  if value <> '' then
  begin
    if FRelation = nil then
      FRelation := TFhirString.create;
    FRelation.value := value
  end
  else if FRelation <> nil then
    FRelation.value := '';
end;

Procedure TFhirBundleLink.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirBundleLink.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirBundleLink.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;


{ TFhirBundleLinkListEnumerator }

Constructor TFhirBundleLinkListEnumerator.Create(list : TFhirBundleLinkList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleLinkListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleLinkListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleLinkListEnumerator.GetCurrent : TFhirBundleLink;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleLinkList }
procedure TFhirBundleLinkList.AddItem(value: TFhirBundleLink);
begin
  assert(value.ClassName = 'TFhirBundleLink', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundleLink');
  add(value);
end;


function TFhirBundleLinkList.Append: TFhirBundleLink;
begin
  result := TFhirBundleLink.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleLinkList.ClearItems;
begin
  Clear;
end;

function TFhirBundleLinkList.GetEnumerator : TFhirBundleLinkListEnumerator;
begin
  result := TFhirBundleLinkListEnumerator.Create(self.link);
end;

function TFhirBundleLinkList.Clone: TFhirBundleLinkList;
begin
  result := TFhirBundleLinkList(inherited Clone);
end;

function TFhirBundleLinkList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleLinkList.GetItemN(index: Integer): TFhirBundleLink;
begin
  result := TFhirBundleLink(ObjectByIndex[index]);
end;

function TFhirBundleLinkList.IndexOf(value: TFhirBundleLink): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleLinkList.Insert(index: Integer): TFhirBundleLink;
begin
  result := TFhirBundleLink.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleLinkList.InsertItem(index: Integer; value: TFhirBundleLink);
begin
  assert(value is TFhirBundleLink);
  Inherited Insert(index, value);
end;

function TFhirBundleLinkList.Item(index: Integer): TFhirBundleLink;
begin
  result := TFhirBundleLink(ObjectByIndex[index]);
end;

function TFhirBundleLinkList.Link: TFhirBundleLinkList;
begin
  result := TFhirBundleLinkList(inherited Link);
end;

procedure TFhirBundleLinkList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleLinkList.SetItemByIndex(index: Integer; value: TFhirBundleLink);
begin
  assert(value is TFhirBundleLink);
  FhirBundleLinks[index] := value;
end;

procedure TFhirBundleLinkList.SetItemN(index: Integer; value: TFhirBundleLink);
begin
  assert(value is TFhirBundleLink);
  ObjectByIndex[index] := value;
end;

{ TFhirBundleEntry }

constructor TFhirBundleEntry.Create;
begin
  inherited;
end;

destructor TFhirBundleEntry.Destroy;
begin
  FLink_List.Free;
  FFullUrl.free;
  FResource.free;
  FSearch.free;
  FRequest.free;
  FResponse.free;
  inherited;
end;

procedure TFhirBundleEntry.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirBundleEntry(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirBundleLinkList.Create;
    FLink_List.Assign(TFhirBundleEntry(oSource).FLink_List);
  end;
  fullUrlElement := TFhirBundleEntry(oSource).fullUrlElement.Clone;
  resource := TFhirBundleEntry(oSource).resource.Clone;
  search := TFhirBundleEntry(oSource).search.Clone;
  request := TFhirBundleEntry(oSource).request.Clone;
  response := TFhirBundleEntry(oSource).response.Clone;
end;

procedure TFhirBundleEntry.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'link') Then
    list.addAll(FLink_List);
  if (child_name = 'fullUrl') Then
     list.add(FFullUrl.Link);
  if (child_name = 'resource') Then
     list.add(FResource.Link);
  if (child_name = 'search') Then
     list.add(FSearch.Link);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
end;

procedure TFhirBundleEntry.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'link', '@Bundle.link', FLink_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'fullUrl', 'uri', FFullUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', 'Resource', FResource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'search', '', FSearch.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', '', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', '', FResponse.Link));{2}
end;

procedure TFhirBundleEntry.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'link') then Link_List.add(propValue as TFhirBundleLink){2}
  else if (propName = 'fullUrl') then FullUrlElement := propValue as TFhirUri{5a}
  else if (propName = 'resource') then Resource := propValue as TFhirResource{4b}
  else if (propName = 'search') then Search := propValue as TFhirBundleEntrySearch{4b}
  else if (propName = 'request') then Request := propValue as TFhirBundleEntryRequest{4b}
  else if (propName = 'response') then Response := propValue as TFhirBundleEntryResponse{4b}
  else inherited;
end;

function TFhirBundleEntry.FhirType : string;
begin
  result := 'entry';
end;

function TFhirBundleEntry.Link : TFhirBundleEntry;
begin
  result := TFhirBundleEntry(inherited Link);
end;

function TFhirBundleEntry.Clone : TFhirBundleEntry;
begin
  result := TFhirBundleEntry(inherited Clone);
end;

function TFhirBundleEntry.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntry;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirBundleEntry)) then
    result := false
  else
  begin
    o := TFhirBundleEntry(other);
    result := compareDeep(link_List, o.link_List, true) and compareDeep(fullUrlElement, o.fullUrlElement, true) and 
      compareDeep(resourceElement, o.resourceElement, true) and compareDeep(searchElement, o.searchElement, true) and 
      compareDeep(requestElement, o.requestElement, true) and compareDeep(responseElement, o.responseElement, true);
  end;
end;

function TFhirBundleEntry.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntry;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirBundleEntry)) then
    result := false
  else
  begin
    o := TFhirBundleEntry(other);
    result := compareValues(fullUrlElement, o.fullUrlElement, true);
  end;
end;

{ TFhirBundleEntry }

Function TFhirBundleEntry.GetLink_List : TFhirBundleLinkList;
begin
  if FLink_List = nil then
    FLink_List := TFhirBundleLinkList.Create;
  result := FLink_List;
end;

Function TFhirBundleEntry.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;

Procedure TFhirBundleEntry.SetFullUrl(value : TFhirUri);
begin
  FFullUrl.free;
  FFullUrl := value;
end;

Function TFhirBundleEntry.GetFullUrlST : String;
begin
  if FFullUrl = nil then
    result := ''
  else
    result := FFullUrl.value;
end;

Procedure TFhirBundleEntry.SetFullUrlST(value : String);
begin
  if value <> '' then
  begin
    if FFullUrl = nil then
      FFullUrl := TFhirUri.create;
    FFullUrl.value := value
  end
  else if FFullUrl <> nil then
    FFullUrl.value := '';
end;

Procedure TFhirBundleEntry.SetResource(value : TFhirResource);
begin
  FResource.free;
  FResource := value;
end;

Procedure TFhirBundleEntry.SetSearch(value : TFhirBundleEntrySearch);
begin
  FSearch.free;
  FSearch := value;
end;

Procedure TFhirBundleEntry.SetRequest(value : TFhirBundleEntryRequest);
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirBundleEntry.SetResponse(value : TFhirBundleEntryResponse);
begin
  FResponse.free;
  FResponse := value;
end;


{ TFhirBundleEntryListEnumerator }

Constructor TFhirBundleEntryListEnumerator.Create(list : TFhirBundleEntryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleEntryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleEntryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleEntryListEnumerator.GetCurrent : TFhirBundleEntry;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleEntryList }
procedure TFhirBundleEntryList.AddItem(value: TFhirBundleEntry);
begin
  assert(value.ClassName = 'TFhirBundleEntry', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundleEntry');
  add(value);
end;


function TFhirBundleEntryList.Append: TFhirBundleEntry;
begin
  result := TFhirBundleEntry.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryList.ClearItems;
begin
  Clear;
end;

function TFhirBundleEntryList.GetEnumerator : TFhirBundleEntryListEnumerator;
begin
  result := TFhirBundleEntryListEnumerator.Create(self.link);
end;

function TFhirBundleEntryList.Clone: TFhirBundleEntryList;
begin
  result := TFhirBundleEntryList(inherited Clone);
end;

function TFhirBundleEntryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleEntryList.GetItemN(index: Integer): TFhirBundleEntry;
begin
  result := TFhirBundleEntry(ObjectByIndex[index]);
end;

function TFhirBundleEntryList.IndexOf(value: TFhirBundleEntry): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleEntryList.Insert(index: Integer): TFhirBundleEntry;
begin
  result := TFhirBundleEntry.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryList.InsertItem(index: Integer; value: TFhirBundleEntry);
begin
  assert(value is TFhirBundleEntry);
  Inherited Insert(index, value);
end;

function TFhirBundleEntryList.Item(index: Integer): TFhirBundleEntry;
begin
  result := TFhirBundleEntry(ObjectByIndex[index]);
end;

function TFhirBundleEntryList.Link: TFhirBundleEntryList;
begin
  result := TFhirBundleEntryList(inherited Link);
end;

procedure TFhirBundleEntryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleEntryList.SetItemByIndex(index: Integer; value: TFhirBundleEntry);
begin
  assert(value is TFhirBundleEntry);
  FhirBundleEntries[index] := value;
end;

procedure TFhirBundleEntryList.SetItemN(index: Integer; value: TFhirBundleEntry);
begin
  assert(value is TFhirBundleEntry);
  ObjectByIndex[index] := value;
end;

{ TFhirBundleEntrySearch }

constructor TFhirBundleEntrySearch.Create;
begin
  inherited;
end;

destructor TFhirBundleEntrySearch.Destroy;
begin
  FMode.free;
  FScore.free;
  inherited;
end;

procedure TFhirBundleEntrySearch.Assign(oSource : TAdvObject);
begin
  inherited;
  FMode := TFhirBundleEntrySearch(oSource).FMode.Link;
  scoreElement := TFhirBundleEntrySearch(oSource).scoreElement.Clone;
end;

procedure TFhirBundleEntrySearch.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'score') Then
     list.add(FScore.Link);
end;

procedure TFhirBundleEntrySearch.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'score', 'decimal', FScore.Link));{2}
end;

procedure TFhirBundleEntrySearch.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'score') then ScoreElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirBundleEntrySearch.FhirType : string;
begin
  result := 'search';
end;

function TFhirBundleEntrySearch.Link : TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch(inherited Link);
end;

function TFhirBundleEntrySearch.Clone : TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch(inherited Clone);
end;

function TFhirBundleEntrySearch.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntrySearch;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirBundleEntrySearch)) then
    result := false
  else
  begin
    o := TFhirBundleEntrySearch(other);
    result := compareDeep(modeElement, o.modeElement, true) and compareDeep(scoreElement, o.scoreElement, true);
  end;
end;

function TFhirBundleEntrySearch.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntrySearch;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirBundleEntrySearch)) then
    result := false
  else
  begin
    o := TFhirBundleEntrySearch(other);
    result := compareValues(modeElement, o.modeElement, true) and compareValues(scoreElement, o.scoreElement, true);
  end;
end;

{ TFhirBundleEntrySearch }

Procedure TFhirBundleEntrySearch.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirBundleEntrySearch.GetModeST : TFhirSearchEntryMode;
begin
  if FMode = nil then
    result := TFhirSearchEntryMode(0)
  else
    result := TFhirSearchEntryMode(StringArrayIndexOfSensitive(CODES_TFhirSearchEntryMode, FMode.value));
end;

Procedure TFhirBundleEntrySearch.SetModeST(value : TFhirSearchEntryMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirSearchEntryMode[value]);
end;

Procedure TFhirBundleEntrySearch.SetScore(value : TFhirDecimal);
begin
  FScore.free;
  FScore := value;
end;

Function TFhirBundleEntrySearch.GetScoreST : String;
begin
  if FScore = nil then
    result := ''
  else
    result := FScore.value;
end;

Procedure TFhirBundleEntrySearch.SetScoreST(value : String);
begin
  if value <> '' then
  begin
    if FScore = nil then
      FScore := TFhirDecimal.create;
    FScore.value := value
  end
  else if FScore <> nil then
    FScore.value := '';
end;


{ TFhirBundleEntrySearchListEnumerator }

Constructor TFhirBundleEntrySearchListEnumerator.Create(list : TFhirBundleEntrySearchList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleEntrySearchListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleEntrySearchListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleEntrySearchListEnumerator.GetCurrent : TFhirBundleEntrySearch;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleEntrySearchList }
procedure TFhirBundleEntrySearchList.AddItem(value: TFhirBundleEntrySearch);
begin
  assert(value.ClassName = 'TFhirBundleEntrySearch', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundleEntrySearch');
  add(value);
end;


function TFhirBundleEntrySearchList.Append: TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntrySearchList.ClearItems;
begin
  Clear;
end;

function TFhirBundleEntrySearchList.GetEnumerator : TFhirBundleEntrySearchListEnumerator;
begin
  result := TFhirBundleEntrySearchListEnumerator.Create(self.link);
end;

function TFhirBundleEntrySearchList.Clone: TFhirBundleEntrySearchList;
begin
  result := TFhirBundleEntrySearchList(inherited Clone);
end;

function TFhirBundleEntrySearchList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleEntrySearchList.GetItemN(index: Integer): TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch(ObjectByIndex[index]);
end;

function TFhirBundleEntrySearchList.IndexOf(value: TFhirBundleEntrySearch): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleEntrySearchList.Insert(index: Integer): TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntrySearchList.InsertItem(index: Integer; value: TFhirBundleEntrySearch);
begin
  assert(value is TFhirBundleEntrySearch);
  Inherited Insert(index, value);
end;

function TFhirBundleEntrySearchList.Item(index: Integer): TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch(ObjectByIndex[index]);
end;

function TFhirBundleEntrySearchList.Link: TFhirBundleEntrySearchList;
begin
  result := TFhirBundleEntrySearchList(inherited Link);
end;

procedure TFhirBundleEntrySearchList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleEntrySearchList.SetItemByIndex(index: Integer; value: TFhirBundleEntrySearch);
begin
  assert(value is TFhirBundleEntrySearch);
  FhirBundleEntrySearches[index] := value;
end;

procedure TFhirBundleEntrySearchList.SetItemN(index: Integer; value: TFhirBundleEntrySearch);
begin
  assert(value is TFhirBundleEntrySearch);
  ObjectByIndex[index] := value;
end;

{ TFhirBundleEntryRequest }

constructor TFhirBundleEntryRequest.Create;
begin
  inherited;
end;

destructor TFhirBundleEntryRequest.Destroy;
begin
  FMethod.free;
  FUrl.free;
  FIfNoneMatch.free;
  FIfModifiedSince.free;
  FIfMatch.free;
  FIfNoneExist.free;
  inherited;
end;

procedure TFhirBundleEntryRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  FMethod := TFhirBundleEntryRequest(oSource).FMethod.Link;
  urlElement := TFhirBundleEntryRequest(oSource).urlElement.Clone;
  ifNoneMatchElement := TFhirBundleEntryRequest(oSource).ifNoneMatchElement.Clone;
  ifModifiedSinceElement := TFhirBundleEntryRequest(oSource).ifModifiedSinceElement.Clone;
  ifMatchElement := TFhirBundleEntryRequest(oSource).ifMatchElement.Clone;
  ifNoneExistElement := TFhirBundleEntryRequest(oSource).ifNoneExistElement.Clone;
end;

procedure TFhirBundleEntryRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'ifNoneMatch') Then
     list.add(FIfNoneMatch.Link);
  if (child_name = 'ifModifiedSince') Then
     list.add(FIfModifiedSince.Link);
  if (child_name = 'ifMatch') Then
     list.add(FIfMatch.Link);
  if (child_name = 'ifNoneExist') Then
     list.add(FIfNoneExist.Link);
end;

procedure TFhirBundleEntryRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'method', 'code', FMethod.Link));{1}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ifNoneMatch', 'string', FIfNoneMatch.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ifModifiedSince', 'instant', FIfModifiedSince.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ifMatch', 'string', FIfMatch.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ifNoneExist', 'string', FIfNoneExist.Link));{2}
end;

procedure TFhirBundleEntryRequest.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'method') then MethodElement := propValue as TFHIREnum
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'ifNoneMatch') then IfNoneMatchElement := propValue as TFhirString{5a}
  else if (propName = 'ifModifiedSince') then IfModifiedSinceElement := propValue as TFhirInstant{5a}
  else if (propName = 'ifMatch') then IfMatchElement := propValue as TFhirString{5a}
  else if (propName = 'ifNoneExist') then IfNoneExistElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirBundleEntryRequest.FhirType : string;
begin
  result := 'request';
end;

function TFhirBundleEntryRequest.Link : TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest(inherited Link);
end;

function TFhirBundleEntryRequest.Clone : TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest(inherited Clone);
end;

function TFhirBundleEntryRequest.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntryRequest;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirBundleEntryRequest)) then
    result := false
  else
  begin
    o := TFhirBundleEntryRequest(other);
    result := compareDeep(methodElement, o.methodElement, true) and compareDeep(urlElement, o.urlElement, true) and 
      compareDeep(ifNoneMatchElement, o.ifNoneMatchElement, true) and compareDeep(ifModifiedSinceElement, o.ifModifiedSinceElement, true) and 
      compareDeep(ifMatchElement, o.ifMatchElement, true) and compareDeep(ifNoneExistElement, o.ifNoneExistElement, true);
  end;
end;

function TFhirBundleEntryRequest.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntryRequest;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirBundleEntryRequest)) then
    result := false
  else
  begin
    o := TFhirBundleEntryRequest(other);
    result := compareValues(methodElement, o.methodElement, true) and compareValues(urlElement, o.urlElement, true) and 
      compareValues(ifNoneMatchElement, o.ifNoneMatchElement, true) and compareValues(ifModifiedSinceElement, o.ifModifiedSinceElement, true) and 
      compareValues(ifMatchElement, o.ifMatchElement, true) and compareValues(ifNoneExistElement, o.ifNoneExistElement, true);
  end;
end;

{ TFhirBundleEntryRequest }

Procedure TFhirBundleEntryRequest.SetMethod(value : TFhirEnum);
begin
  FMethod.free;
  FMethod := value;
end;

Function TFhirBundleEntryRequest.GetMethodST : TFhirHttpVerb;
begin
  if FMethod = nil then
    result := TFhirHttpVerb(0)
  else
    result := TFhirHttpVerb(StringArrayIndexOfSensitive(CODES_TFhirHttpVerb, FMethod.value));
end;

Procedure TFhirBundleEntryRequest.SetMethodST(value : TFhirHttpVerb);
begin
  if ord(value) = 0 then
    MethodElement := nil
  else
    MethodElement := TFhirEnum.create(CODES_TFhirHttpVerb[value]);
end;

Procedure TFhirBundleEntryRequest.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirBundleEntryRequest.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirBundleEntryRequest.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirBundleEntryRequest.SetIfNoneMatch(value : TFhirString);
begin
  FIfNoneMatch.free;
  FIfNoneMatch := value;
end;

Function TFhirBundleEntryRequest.GetIfNoneMatchST : String;
begin
  if FIfNoneMatch = nil then
    result := ''
  else
    result := FIfNoneMatch.value;
end;

Procedure TFhirBundleEntryRequest.SetIfNoneMatchST(value : String);
begin
  if value <> '' then
  begin
    if FIfNoneMatch = nil then
      FIfNoneMatch := TFhirString.create;
    FIfNoneMatch.value := value
  end
  else if FIfNoneMatch <> nil then
    FIfNoneMatch.value := '';
end;

Procedure TFhirBundleEntryRequest.SetIfModifiedSince(value : TFhirInstant);
begin
  FIfModifiedSince.free;
  FIfModifiedSince := value;
end;

Function TFhirBundleEntryRequest.GetIfModifiedSinceST : TDateAndTime;
begin
  if FIfModifiedSince = nil then
    result := nil
  else
    result := FIfModifiedSince.value;
end;

Procedure TFhirBundleEntryRequest.SetIfModifiedSinceST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIfModifiedSince = nil then
      FIfModifiedSince := TFhirInstant.create;
    FIfModifiedSince.value := value
  end
  else if FIfModifiedSince <> nil then
    FIfModifiedSince.value := nil;
end;

Procedure TFhirBundleEntryRequest.SetIfMatch(value : TFhirString);
begin
  FIfMatch.free;
  FIfMatch := value;
end;

Function TFhirBundleEntryRequest.GetIfMatchST : String;
begin
  if FIfMatch = nil then
    result := ''
  else
    result := FIfMatch.value;
end;

Procedure TFhirBundleEntryRequest.SetIfMatchST(value : String);
begin
  if value <> '' then
  begin
    if FIfMatch = nil then
      FIfMatch := TFhirString.create;
    FIfMatch.value := value
  end
  else if FIfMatch <> nil then
    FIfMatch.value := '';
end;

Procedure TFhirBundleEntryRequest.SetIfNoneExist(value : TFhirString);
begin
  FIfNoneExist.free;
  FIfNoneExist := value;
end;

Function TFhirBundleEntryRequest.GetIfNoneExistST : String;
begin
  if FIfNoneExist = nil then
    result := ''
  else
    result := FIfNoneExist.value;
end;

Procedure TFhirBundleEntryRequest.SetIfNoneExistST(value : String);
begin
  if value <> '' then
  begin
    if FIfNoneExist = nil then
      FIfNoneExist := TFhirString.create;
    FIfNoneExist.value := value
  end
  else if FIfNoneExist <> nil then
    FIfNoneExist.value := '';
end;


{ TFhirBundleEntryRequestListEnumerator }

Constructor TFhirBundleEntryRequestListEnumerator.Create(list : TFhirBundleEntryRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleEntryRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleEntryRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleEntryRequestListEnumerator.GetCurrent : TFhirBundleEntryRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleEntryRequestList }
procedure TFhirBundleEntryRequestList.AddItem(value: TFhirBundleEntryRequest);
begin
  assert(value.ClassName = 'TFhirBundleEntryRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundleEntryRequest');
  add(value);
end;


function TFhirBundleEntryRequestList.Append: TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryRequestList.ClearItems;
begin
  Clear;
end;

function TFhirBundleEntryRequestList.GetEnumerator : TFhirBundleEntryRequestListEnumerator;
begin
  result := TFhirBundleEntryRequestListEnumerator.Create(self.link);
end;

function TFhirBundleEntryRequestList.Clone: TFhirBundleEntryRequestList;
begin
  result := TFhirBundleEntryRequestList(inherited Clone);
end;

function TFhirBundleEntryRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleEntryRequestList.GetItemN(index: Integer): TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest(ObjectByIndex[index]);
end;

function TFhirBundleEntryRequestList.IndexOf(value: TFhirBundleEntryRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleEntryRequestList.Insert(index: Integer): TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryRequestList.InsertItem(index: Integer; value: TFhirBundleEntryRequest);
begin
  assert(value is TFhirBundleEntryRequest);
  Inherited Insert(index, value);
end;

function TFhirBundleEntryRequestList.Item(index: Integer): TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest(ObjectByIndex[index]);
end;

function TFhirBundleEntryRequestList.Link: TFhirBundleEntryRequestList;
begin
  result := TFhirBundleEntryRequestList(inherited Link);
end;

procedure TFhirBundleEntryRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleEntryRequestList.SetItemByIndex(index: Integer; value: TFhirBundleEntryRequest);
begin
  assert(value is TFhirBundleEntryRequest);
  FhirBundleEntryRequests[index] := value;
end;

procedure TFhirBundleEntryRequestList.SetItemN(index: Integer; value: TFhirBundleEntryRequest);
begin
  assert(value is TFhirBundleEntryRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirBundleEntryResponse }

constructor TFhirBundleEntryResponse.Create;
begin
  inherited;
end;

destructor TFhirBundleEntryResponse.Destroy;
begin
  FStatus.free;
  FLocation.free;
  FEtag.free;
  FLastModified.free;
  inherited;
end;

procedure TFhirBundleEntryResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  statusElement := TFhirBundleEntryResponse(oSource).statusElement.Clone;
  locationElement := TFhirBundleEntryResponse(oSource).locationElement.Clone;
  etagElement := TFhirBundleEntryResponse(oSource).etagElement.Clone;
  lastModifiedElement := TFhirBundleEntryResponse(oSource).lastModifiedElement.Clone;
end;

procedure TFhirBundleEntryResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'etag') Then
     list.add(FEtag.Link);
  if (child_name = 'lastModified') Then
     list.add(FLastModified.Link);
end;

procedure TFhirBundleEntryResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'status', 'string', FStatus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'uri', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'etag', 'string', FEtag.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lastModified', 'instant', FLastModified.Link));{2}
end;

procedure TFhirBundleEntryResponse.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'status') then StatusElement := propValue as TFhirString{5a}
  else if (propName = 'location') then LocationElement := propValue as TFhirUri{5a}
  else if (propName = 'etag') then EtagElement := propValue as TFhirString{5a}
  else if (propName = 'lastModified') then LastModifiedElement := propValue as TFhirInstant{5a}
  else inherited;
end;

function TFhirBundleEntryResponse.FhirType : string;
begin
  result := 'response';
end;

function TFhirBundleEntryResponse.Link : TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse(inherited Link);
end;

function TFhirBundleEntryResponse.Clone : TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse(inherited Clone);
end;

function TFhirBundleEntryResponse.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntryResponse;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirBundleEntryResponse)) then
    result := false
  else
  begin
    o := TFhirBundleEntryResponse(other);
    result := compareDeep(statusElement, o.statusElement, true) and compareDeep(locationElement, o.locationElement, true) and 
      compareDeep(etagElement, o.etagElement, true) and compareDeep(lastModifiedElement, o.lastModifiedElement, true);
  end;
end;

function TFhirBundleEntryResponse.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirBundleEntryResponse;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirBundleEntryResponse)) then
    result := false
  else
  begin
    o := TFhirBundleEntryResponse(other);
    result := compareValues(statusElement, o.statusElement, true) and compareValues(locationElement, o.locationElement, true) and 
      compareValues(etagElement, o.etagElement, true) and compareValues(lastModifiedElement, o.lastModifiedElement, true);
  end;
end;

{ TFhirBundleEntryResponse }

Procedure TFhirBundleEntryResponse.SetStatus(value : TFhirString);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirBundleEntryResponse.GetStatusST : String;
begin
  if FStatus = nil then
    result := ''
  else
    result := FStatus.value;
end;

Procedure TFhirBundleEntryResponse.SetStatusST(value : String);
begin
  if value <> '' then
  begin
    if FStatus = nil then
      FStatus := TFhirString.create;
    FStatus.value := value
  end
  else if FStatus <> nil then
    FStatus.value := '';
end;

Procedure TFhirBundleEntryResponse.SetLocation(value : TFhirUri);
begin
  FLocation.free;
  FLocation := value;
end;

Function TFhirBundleEntryResponse.GetLocationST : String;
begin
  if FLocation = nil then
    result := ''
  else
    result := FLocation.value;
end;

Procedure TFhirBundleEntryResponse.SetLocationST(value : String);
begin
  if value <> '' then
  begin
    if FLocation = nil then
      FLocation := TFhirUri.create;
    FLocation.value := value
  end
  else if FLocation <> nil then
    FLocation.value := '';
end;

Procedure TFhirBundleEntryResponse.SetEtag(value : TFhirString);
begin
  FEtag.free;
  FEtag := value;
end;

Function TFhirBundleEntryResponse.GetEtagST : String;
begin
  if FEtag = nil then
    result := ''
  else
    result := FEtag.value;
end;

Procedure TFhirBundleEntryResponse.SetEtagST(value : String);
begin
  if value <> '' then
  begin
    if FEtag = nil then
      FEtag := TFhirString.create;
    FEtag.value := value
  end
  else if FEtag <> nil then
    FEtag.value := '';
end;

Procedure TFhirBundleEntryResponse.SetLastModified(value : TFhirInstant);
begin
  FLastModified.free;
  FLastModified := value;
end;

Function TFhirBundleEntryResponse.GetLastModifiedST : TDateAndTime;
begin
  if FLastModified = nil then
    result := nil
  else
    result := FLastModified.value;
end;

Procedure TFhirBundleEntryResponse.SetLastModifiedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FLastModified = nil then
      FLastModified := TFhirInstant.create;
    FLastModified.value := value
  end
  else if FLastModified <> nil then
    FLastModified.value := nil;
end;


{ TFhirBundleEntryResponseListEnumerator }

Constructor TFhirBundleEntryResponseListEnumerator.Create(list : TFhirBundleEntryResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleEntryResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleEntryResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleEntryResponseListEnumerator.GetCurrent : TFhirBundleEntryResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleEntryResponseList }
procedure TFhirBundleEntryResponseList.AddItem(value: TFhirBundleEntryResponse);
begin
  assert(value.ClassName = 'TFhirBundleEntryResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundleEntryResponse');
  add(value);
end;


function TFhirBundleEntryResponseList.Append: TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryResponseList.ClearItems;
begin
  Clear;
end;

function TFhirBundleEntryResponseList.GetEnumerator : TFhirBundleEntryResponseListEnumerator;
begin
  result := TFhirBundleEntryResponseListEnumerator.Create(self.link);
end;

function TFhirBundleEntryResponseList.Clone: TFhirBundleEntryResponseList;
begin
  result := TFhirBundleEntryResponseList(inherited Clone);
end;

function TFhirBundleEntryResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleEntryResponseList.GetItemN(index: Integer): TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse(ObjectByIndex[index]);
end;

function TFhirBundleEntryResponseList.IndexOf(value: TFhirBundleEntryResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleEntryResponseList.Insert(index: Integer): TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleEntryResponseList.InsertItem(index: Integer; value: TFhirBundleEntryResponse);
begin
  assert(value is TFhirBundleEntryResponse);
  Inherited Insert(index, value);
end;

function TFhirBundleEntryResponseList.Item(index: Integer): TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse(ObjectByIndex[index]);
end;

function TFhirBundleEntryResponseList.Link: TFhirBundleEntryResponseList;
begin
  result := TFhirBundleEntryResponseList(inherited Link);
end;

procedure TFhirBundleEntryResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleEntryResponseList.SetItemByIndex(index: Integer; value: TFhirBundleEntryResponse);
begin
  assert(value is TFhirBundleEntryResponse);
  FhirBundleEntryResponses[index] := value;
end;

procedure TFhirBundleEntryResponseList.SetItemN(index: Integer; value: TFhirBundleEntryResponse);
begin
  assert(value is TFhirBundleEntryResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirBundle }

constructor TFhirBundle.Create;
begin
  inherited;
end;

destructor TFhirBundle.Destroy;
begin
  FType_.free;
  FTotal.free;
  FLink_List.Free;
  FEntryList.Free;
  FSignature.free;
  inherited;
end;

function TFhirBundle.GetResourceType : TFhirResourceType;
begin
  result := frtBundle;
end;

procedure TFhirBundle.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirBundle(oSource).FType_.Link;
  totalElement := TFhirBundle(oSource).totalElement.Clone;
  if (TFhirBundle(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirBundleLinkList.Create;
    FLink_List.Assign(TFhirBundle(oSource).FLink_List);
  end;
  if (TFhirBundle(oSource).FEntryList = nil) then
  begin
    FEntryList.free;
    FEntryList := nil;
  end
  else
  begin
    FEntryList := TFhirBundleEntryList.Create;
    FEntryList.Assign(TFhirBundle(oSource).FEntryList);
  end;
  signature := TFhirBundle(oSource).signature.Clone;
end;

procedure TFhirBundle.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'total') Then
     list.add(FTotal.Link);
  if (child_name = 'link') Then
    list.addAll(FLink_List);
  if (child_name = 'entry') Then
    list.addAll(FEntryList);
  if (child_name = 'signature') Then
     list.add(FSignature.Link);
end;

procedure TFhirBundle.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'total', 'unsignedInt', FTotal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'link', '', FLink_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'entry', '', FEntryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'signature', 'Signature', FSignature.Link));{2}
end;

procedure TFhirBundle.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'total') then TotalElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'link') then Link_List.add(propValue as TFhirBundleLink){2}
  else if (propName = 'entry') then EntryList.add(propValue as TFhirBundleEntry){2}
  else if (propName = 'signature') then Signature := propValue as TFhirSignature{4b}
  else inherited;
end;

function TFhirBundle.FhirType : string;
begin
  result := 'Bundle';
end;

function TFhirBundle.Link : TFhirBundle;
begin
  result := TFhirBundle(inherited Link);
end;

function TFhirBundle.Clone : TFhirBundle;
begin
  result := TFhirBundle(inherited Clone);
end;

{ TFhirBundle }

Procedure TFhirBundle.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirBundle.GetType_ST : TFhirBundleType;
begin
  if FType_ = nil then
    result := TFhirBundleType(0)
  else
    result := TFhirBundleType(StringArrayIndexOfSensitive(CODES_TFhirBundleType, FType_.value));
end;

Procedure TFhirBundle.SetType_ST(value : TFhirBundleType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirBundleType[value]);
end;

Procedure TFhirBundle.SetTotal(value : TFhirUnsignedInt);
begin
  FTotal.free;
  FTotal := value;
end;

Function TFhirBundle.GetTotalST : String;
begin
  if FTotal = nil then
    result := ''
  else
    result := FTotal.value;
end;

Procedure TFhirBundle.SetTotalST(value : String);
begin
  if value <> '' then
  begin
    if FTotal = nil then
      FTotal := TFhirUnsignedInt.create;
    FTotal.value := value
  end
  else if FTotal <> nil then
    FTotal.value := '';
end;

Function TFhirBundle.GetLink_List : TFhirBundleLinkList;
begin
  if FLink_List = nil then
    FLink_List := TFhirBundleLinkList.Create;
  result := FLink_List;
end;

Function TFhirBundle.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;

Function TFhirBundle.GetEntryList : TFhirBundleEntryList;
begin
  if FEntryList = nil then
    FEntryList := TFhirBundleEntryList.Create;
  result := FEntryList;
end;

Function TFhirBundle.GetHasEntryList : boolean;
begin
  result := (FEntryList <> nil) and (FEntryList.count > 0);
end;

Procedure TFhirBundle.SetSignature(value : TFhirSignature);
begin
  FSignature.free;
  FSignature := value;
end;


{ TFhirBundleListEnumerator }

Constructor TFhirBundleListEnumerator.Create(list : TFhirBundleList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirBundleListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirBundleListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirBundleListEnumerator.GetCurrent : TFhirBundle;
begin
  Result := FList[FIndex];
end;


{ TFhirBundleList }
procedure TFhirBundleList.AddItem(value: TFhirBundle);
begin
  assert(value.ClassName = 'TFhirBundle', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirBundle');
  add(value);
end;


function TFhirBundleList.Append: TFhirBundle;
begin
  result := TFhirBundle.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleList.ClearItems;
begin
  Clear;
end;

function TFhirBundleList.GetEnumerator : TFhirBundleListEnumerator;
begin
  result := TFhirBundleListEnumerator.Create(self.link);
end;

function TFhirBundleList.Clone: TFhirBundleList;
begin
  result := TFhirBundleList(inherited Clone);
end;

function TFhirBundleList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirBundleList.GetItemN(index: Integer): TFhirBundle;
begin
  result := TFhirBundle(ObjectByIndex[index]);
end;

function TFhirBundleList.IndexOf(value: TFhirBundle): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirBundleList.Insert(index: Integer): TFhirBundle;
begin
  result := TFhirBundle.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirBundleList.InsertItem(index: Integer; value: TFhirBundle);
begin
  assert(value is TFhirBundle);
  Inherited Insert(index, value);
end;

function TFhirBundleList.Item(index: Integer): TFhirBundle;
begin
  result := TFhirBundle(ObjectByIndex[index]);
end;

function TFhirBundleList.Link: TFhirBundleList;
begin
  result := TFhirBundleList(inherited Link);
end;

procedure TFhirBundleList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirBundleList.SetItemByIndex(index: Integer; value: TFhirBundle);
begin
  assert(value is TFhirBundle);
  FhirBundles[index] := value;
end;

procedure TFhirBundleList.SetItemN(index: Integer; value: TFhirBundle);
begin
  assert(value is TFhirBundle);
  ObjectByIndex[index] := value;
end;

{ TFhirCarePlanRelatedPlan }

constructor TFhirCarePlanRelatedPlan.Create;
begin
  inherited;
end;

destructor TFhirCarePlanRelatedPlan.Destroy;
begin
  FCode.free;
  FPlan.free;
  inherited;
end;

procedure TFhirCarePlanRelatedPlan.Assign(oSource : TAdvObject);
begin
  inherited;
  FCode := TFhirCarePlanRelatedPlan(oSource).FCode.Link;
  plan := TFhirCarePlanRelatedPlan(oSource).plan.Clone;
end;

procedure TFhirCarePlanRelatedPlan.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'plan') Then
     list.add(FPlan.Link);
end;

procedure TFhirCarePlanRelatedPlan.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'plan', 'Reference(CarePlan)', FPlan.Link));{2}
end;

procedure TFhirCarePlanRelatedPlan.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'plan') then Plan := propValue as TFhirReference{TFhirCarePlan}{4b}
  else inherited;
end;

function TFhirCarePlanRelatedPlan.FhirType : string;
begin
  result := 'relatedPlan';
end;

function TFhirCarePlanRelatedPlan.Link : TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan(inherited Link);
end;

function TFhirCarePlanRelatedPlan.Clone : TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan(inherited Clone);
end;

function TFhirCarePlanRelatedPlan.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanRelatedPlan;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCarePlanRelatedPlan)) then
    result := false
  else
  begin
    o := TFhirCarePlanRelatedPlan(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(planElement, o.planElement, true);
  end;
end;

function TFhirCarePlanRelatedPlan.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanRelatedPlan;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCarePlanRelatedPlan)) then
    result := false
  else
  begin
    o := TFhirCarePlanRelatedPlan(other);
    result := compareValues(codeElement, o.codeElement, true);
  end;
end;

{ TFhirCarePlanRelatedPlan }

Procedure TFhirCarePlanRelatedPlan.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirCarePlanRelatedPlan.GetCodeST : TFhirCarePlanRelationship;
begin
  if FCode = nil then
    result := TFhirCarePlanRelationship(0)
  else
    result := TFhirCarePlanRelationship(StringArrayIndexOfSensitive(CODES_TFhirCarePlanRelationship, FCode.value));
end;

Procedure TFhirCarePlanRelatedPlan.SetCodeST(value : TFhirCarePlanRelationship);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirCarePlanRelationship[value]);
end;

Procedure TFhirCarePlanRelatedPlan.SetPlan(value : TFhirReference{TFhirCarePlan});
begin
  FPlan.free;
  FPlan := value;
end;


{ TFhirCarePlanRelatedPlanListEnumerator }

Constructor TFhirCarePlanRelatedPlanListEnumerator.Create(list : TFhirCarePlanRelatedPlanList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCarePlanRelatedPlanListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCarePlanRelatedPlanListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCarePlanRelatedPlanListEnumerator.GetCurrent : TFhirCarePlanRelatedPlan;
begin
  Result := FList[FIndex];
end;


{ TFhirCarePlanRelatedPlanList }
procedure TFhirCarePlanRelatedPlanList.AddItem(value: TFhirCarePlanRelatedPlan);
begin
  assert(value.ClassName = 'TFhirCarePlanRelatedPlan', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCarePlanRelatedPlan');
  add(value);
end;


function TFhirCarePlanRelatedPlanList.Append: TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanRelatedPlanList.ClearItems;
begin
  Clear;
end;

function TFhirCarePlanRelatedPlanList.GetEnumerator : TFhirCarePlanRelatedPlanListEnumerator;
begin
  result := TFhirCarePlanRelatedPlanListEnumerator.Create(self.link);
end;

function TFhirCarePlanRelatedPlanList.Clone: TFhirCarePlanRelatedPlanList;
begin
  result := TFhirCarePlanRelatedPlanList(inherited Clone);
end;

function TFhirCarePlanRelatedPlanList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCarePlanRelatedPlanList.GetItemN(index: Integer): TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan(ObjectByIndex[index]);
end;

function TFhirCarePlanRelatedPlanList.IndexOf(value: TFhirCarePlanRelatedPlan): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCarePlanRelatedPlanList.Insert(index: Integer): TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanRelatedPlanList.InsertItem(index: Integer; value: TFhirCarePlanRelatedPlan);
begin
  assert(value is TFhirCarePlanRelatedPlan);
  Inherited Insert(index, value);
end;

function TFhirCarePlanRelatedPlanList.Item(index: Integer): TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan(ObjectByIndex[index]);
end;

function TFhirCarePlanRelatedPlanList.Link: TFhirCarePlanRelatedPlanList;
begin
  result := TFhirCarePlanRelatedPlanList(inherited Link);
end;

procedure TFhirCarePlanRelatedPlanList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCarePlanRelatedPlanList.SetItemByIndex(index: Integer; value: TFhirCarePlanRelatedPlan);
begin
  assert(value is TFhirCarePlanRelatedPlan);
  FhirCarePlanRelatedPlans[index] := value;
end;

procedure TFhirCarePlanRelatedPlanList.SetItemN(index: Integer; value: TFhirCarePlanRelatedPlan);
begin
  assert(value is TFhirCarePlanRelatedPlan);
  ObjectByIndex[index] := value;
end;

{ TFhirCarePlanParticipant }

constructor TFhirCarePlanParticipant.Create;
begin
  inherited;
end;

destructor TFhirCarePlanParticipant.Destroy;
begin
  FRole.free;
  FMember.free;
  inherited;
end;

procedure TFhirCarePlanParticipant.Assign(oSource : TAdvObject);
begin
  inherited;
  role := TFhirCarePlanParticipant(oSource).role.Clone;
  member := TFhirCarePlanParticipant(oSource).member.Clone;
end;

procedure TFhirCarePlanParticipant.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'role') Then
     list.add(FRole.Link);
  if (child_name = 'member') Then
     list.add(FMember.Link);
end;

procedure TFhirCarePlanParticipant.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRole.Link));{2}
  oList.add(TFHIRProperty.create(self, 'member', 'Reference(Practitioner|RelatedPerson|Patient|Organization)', FMember.Link));{2}
end;

procedure TFhirCarePlanParticipant.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'role') then Role := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'member') then Member := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirCarePlanParticipant.FhirType : string;
begin
  result := 'participant';
end;

function TFhirCarePlanParticipant.Link : TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant(inherited Link);
end;

function TFhirCarePlanParticipant.Clone : TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant(inherited Clone);
end;

function TFhirCarePlanParticipant.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanParticipant;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCarePlanParticipant)) then
    result := false
  else
  begin
    o := TFhirCarePlanParticipant(other);
    result := compareDeep(roleElement, o.roleElement, true) and compareDeep(memberElement, o.memberElement, true);
  end;
end;

function TFhirCarePlanParticipant.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanParticipant;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCarePlanParticipant)) then
    result := false
  else
  begin
    o := TFhirCarePlanParticipant(other);
    result := true;
  end;
end;

{ TFhirCarePlanParticipant }

Procedure TFhirCarePlanParticipant.SetRole(value : TFhirCodeableConcept);
begin
  FRole.free;
  FRole := value;
end;

Procedure TFhirCarePlanParticipant.SetMember(value : TFhirReference{Resource});
begin
  FMember.free;
  FMember := value;
end;


{ TFhirCarePlanParticipantListEnumerator }

Constructor TFhirCarePlanParticipantListEnumerator.Create(list : TFhirCarePlanParticipantList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCarePlanParticipantListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCarePlanParticipantListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCarePlanParticipantListEnumerator.GetCurrent : TFhirCarePlanParticipant;
begin
  Result := FList[FIndex];
end;


{ TFhirCarePlanParticipantList }
procedure TFhirCarePlanParticipantList.AddItem(value: TFhirCarePlanParticipant);
begin
  assert(value.ClassName = 'TFhirCarePlanParticipant', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCarePlanParticipant');
  add(value);
end;


function TFhirCarePlanParticipantList.Append: TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanParticipantList.ClearItems;
begin
  Clear;
end;

function TFhirCarePlanParticipantList.GetEnumerator : TFhirCarePlanParticipantListEnumerator;
begin
  result := TFhirCarePlanParticipantListEnumerator.Create(self.link);
end;

function TFhirCarePlanParticipantList.Clone: TFhirCarePlanParticipantList;
begin
  result := TFhirCarePlanParticipantList(inherited Clone);
end;

function TFhirCarePlanParticipantList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCarePlanParticipantList.GetItemN(index: Integer): TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant(ObjectByIndex[index]);
end;

function TFhirCarePlanParticipantList.IndexOf(value: TFhirCarePlanParticipant): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCarePlanParticipantList.Insert(index: Integer): TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanParticipantList.InsertItem(index: Integer; value: TFhirCarePlanParticipant);
begin
  assert(value is TFhirCarePlanParticipant);
  Inherited Insert(index, value);
end;

function TFhirCarePlanParticipantList.Item(index: Integer): TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant(ObjectByIndex[index]);
end;

function TFhirCarePlanParticipantList.Link: TFhirCarePlanParticipantList;
begin
  result := TFhirCarePlanParticipantList(inherited Link);
end;

procedure TFhirCarePlanParticipantList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCarePlanParticipantList.SetItemByIndex(index: Integer; value: TFhirCarePlanParticipant);
begin
  assert(value is TFhirCarePlanParticipant);
  FhirCarePlanParticipants[index] := value;
end;

procedure TFhirCarePlanParticipantList.SetItemN(index: Integer; value: TFhirCarePlanParticipant);
begin
  assert(value is TFhirCarePlanParticipant);
  ObjectByIndex[index] := value;
end;

{ TFhirCarePlanActivity }

constructor TFhirCarePlanActivity.Create;
begin
  inherited;
end;

destructor TFhirCarePlanActivity.Destroy;
begin
  FActionResultingList.Free;
  FProgressList.Free;
  FReference.free;
  FDetail.free;
  inherited;
end;

procedure TFhirCarePlanActivity.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCarePlanActivity(oSource).FActionResultingList = nil) then
  begin
    FActionResultingList.free;
    FActionResultingList := nil;
  end
  else
  begin
    FActionResultingList := TFhirReferenceList{TFhirReference}.Create;
    FActionResultingList.Assign(TFhirCarePlanActivity(oSource).FActionResultingList);
  end;
  if (TFhirCarePlanActivity(oSource).FProgressList = nil) then
  begin
    FProgressList.free;
    FProgressList := nil;
  end
  else
  begin
    FProgressList := TFhirAnnotationList.Create;
    FProgressList.Assign(TFhirCarePlanActivity(oSource).FProgressList);
  end;
  reference := TFhirCarePlanActivity(oSource).reference.Clone;
  detail := TFhirCarePlanActivity(oSource).detail.Clone;
end;

procedure TFhirCarePlanActivity.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'actionResulting') Then
    list.addAll(FActionResultingList);
  if (child_name = 'progress') Then
    list.addAll(FProgressList);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'detail') Then
     list.add(FDetail.Link);
end;

procedure TFhirCarePlanActivity.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'actionResulting', 'Reference(Any)', FActionResultingList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'progress', 'Annotation', FProgressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reference', 'Reference(Appointment|CommunicationRequest|DeviceUseRequest|DiagnosticOrder|MedicationOrder|NutritionOrder|Order|ProcedureRequest|ProcessRequest|ReferralRequest|SupplyRequest|VisionPrescription)', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetail.Link));{2}
end;

procedure TFhirCarePlanActivity.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'actionResulting') then ActionResultingList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'progress') then ProgressList.add(propValue as TFhirAnnotation){2}
  else if (propName = 'reference') then Reference := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'detail') then Detail := propValue as TFhirCarePlanActivityDetail{4b}
  else inherited;
end;

function TFhirCarePlanActivity.FhirType : string;
begin
  result := 'activity';
end;

function TFhirCarePlanActivity.Link : TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity(inherited Link);
end;

function TFhirCarePlanActivity.Clone : TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity(inherited Clone);
end;

function TFhirCarePlanActivity.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanActivity;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCarePlanActivity)) then
    result := false
  else
  begin
    o := TFhirCarePlanActivity(other);
    result := compareDeep(actionResultingList, o.actionResultingList, true) and 
      compareDeep(progressList, o.progressList, true) and compareDeep(referenceElement, o.referenceElement, true) and 
      compareDeep(detailElement, o.detailElement, true);
  end;
end;

function TFhirCarePlanActivity.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanActivity;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCarePlanActivity)) then
    result := false
  else
  begin
    o := TFhirCarePlanActivity(other);
    result := true;
  end;
end;

{ TFhirCarePlanActivity }

Function TFhirCarePlanActivity.GetActionResultingList : TFhirReferenceList{TFhirReference};
begin
  if FActionResultingList = nil then
    FActionResultingList := TFhirReferenceList{TFhirReference}.Create;
  result := FActionResultingList;
end;

Function TFhirCarePlanActivity.GetHasActionResultingList : boolean;
begin
  result := (FActionResultingList <> nil) and (FActionResultingList.count > 0);
end;

Function TFhirCarePlanActivity.GetProgressList : TFhirAnnotationList;
begin
  if FProgressList = nil then
    FProgressList := TFhirAnnotationList.Create;
  result := FProgressList;
end;

Function TFhirCarePlanActivity.GetHasProgressList : boolean;
begin
  result := (FProgressList <> nil) and (FProgressList.count > 0);
end;

Procedure TFhirCarePlanActivity.SetReference(value : TFhirReference{Resource});
begin
  FReference.free;
  FReference := value;
end;

Procedure TFhirCarePlanActivity.SetDetail(value : TFhirCarePlanActivityDetail);
begin
  FDetail.free;
  FDetail := value;
end;


{ TFhirCarePlanActivityListEnumerator }

Constructor TFhirCarePlanActivityListEnumerator.Create(list : TFhirCarePlanActivityList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCarePlanActivityListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCarePlanActivityListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCarePlanActivityListEnumerator.GetCurrent : TFhirCarePlanActivity;
begin
  Result := FList[FIndex];
end;


{ TFhirCarePlanActivityList }
procedure TFhirCarePlanActivityList.AddItem(value: TFhirCarePlanActivity);
begin
  assert(value.ClassName = 'TFhirCarePlanActivity', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCarePlanActivity');
  add(value);
end;


function TFhirCarePlanActivityList.Append: TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanActivityList.ClearItems;
begin
  Clear;
end;

function TFhirCarePlanActivityList.GetEnumerator : TFhirCarePlanActivityListEnumerator;
begin
  result := TFhirCarePlanActivityListEnumerator.Create(self.link);
end;

function TFhirCarePlanActivityList.Clone: TFhirCarePlanActivityList;
begin
  result := TFhirCarePlanActivityList(inherited Clone);
end;

function TFhirCarePlanActivityList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCarePlanActivityList.GetItemN(index: Integer): TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity(ObjectByIndex[index]);
end;

function TFhirCarePlanActivityList.IndexOf(value: TFhirCarePlanActivity): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCarePlanActivityList.Insert(index: Integer): TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanActivityList.InsertItem(index: Integer; value: TFhirCarePlanActivity);
begin
  assert(value is TFhirCarePlanActivity);
  Inherited Insert(index, value);
end;

function TFhirCarePlanActivityList.Item(index: Integer): TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity(ObjectByIndex[index]);
end;

function TFhirCarePlanActivityList.Link: TFhirCarePlanActivityList;
begin
  result := TFhirCarePlanActivityList(inherited Link);
end;

procedure TFhirCarePlanActivityList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCarePlanActivityList.SetItemByIndex(index: Integer; value: TFhirCarePlanActivity);
begin
  assert(value is TFhirCarePlanActivity);
  FhirCarePlanActivities[index] := value;
end;

procedure TFhirCarePlanActivityList.SetItemN(index: Integer; value: TFhirCarePlanActivity);
begin
  assert(value is TFhirCarePlanActivity);
  ObjectByIndex[index] := value;
end;

{ TFhirCarePlanActivityDetail }

constructor TFhirCarePlanActivityDetail.Create;
begin
  inherited;
end;

destructor TFhirCarePlanActivityDetail.Destroy;
begin
  FCategory.free;
  FCode.free;
  FReasonCodeList.Free;
  FReasonReferenceList.Free;
  FGoalList.Free;
  FStatus.free;
  FStatusReason.free;
  FProhibited.free;
  FScheduled.free;
  FLocation.free;
  FPerformerList.Free;
  FProduct.free;
  FDailyAmount.free;
  FQuantity.free;
  FDescription.free;
  inherited;
end;

procedure TFhirCarePlanActivityDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  category := TFhirCarePlanActivityDetail(oSource).category.Clone;
  code := TFhirCarePlanActivityDetail(oSource).code.Clone;
  if (TFhirCarePlanActivityDetail(oSource).FReasonCodeList = nil) then
  begin
    FReasonCodeList.free;
    FReasonCodeList := nil;
  end
  else
  begin
    FReasonCodeList := TFhirCodeableConceptList.Create;
    FReasonCodeList.Assign(TFhirCarePlanActivityDetail(oSource).FReasonCodeList);
  end;
  if (TFhirCarePlanActivityDetail(oSource).FReasonReferenceList = nil) then
  begin
    FReasonReferenceList.free;
    FReasonReferenceList := nil;
  end
  else
  begin
    FReasonReferenceList := TFhirReferenceList{TFhirCondition}.Create;
    FReasonReferenceList.Assign(TFhirCarePlanActivityDetail(oSource).FReasonReferenceList);
  end;
  if (TFhirCarePlanActivityDetail(oSource).FGoalList = nil) then
  begin
    FGoalList.free;
    FGoalList := nil;
  end
  else
  begin
    FGoalList := TFhirReferenceList{TFhirGoal}.Create;
    FGoalList.Assign(TFhirCarePlanActivityDetail(oSource).FGoalList);
  end;
  FStatus := TFhirCarePlanActivityDetail(oSource).FStatus.Link;
  statusReason := TFhirCarePlanActivityDetail(oSource).statusReason.Clone;
  prohibitedElement := TFhirCarePlanActivityDetail(oSource).prohibitedElement.Clone;
  scheduled := TFhirCarePlanActivityDetail(oSource).scheduled.Clone;
  location := TFhirCarePlanActivityDetail(oSource).location.Clone;
  if (TFhirCarePlanActivityDetail(oSource).FPerformerList = nil) then
  begin
    FPerformerList.free;
    FPerformerList := nil;
  end
  else
  begin
    FPerformerList := TFhirReferenceList{Resource}.Create;
    FPerformerList.Assign(TFhirCarePlanActivityDetail(oSource).FPerformerList);
  end;
  product := TFhirCarePlanActivityDetail(oSource).product.Clone;
  dailyAmount := TFhirCarePlanActivityDetail(oSource).dailyAmount.Clone;
  quantity := TFhirCarePlanActivityDetail(oSource).quantity.Clone;
  descriptionElement := TFhirCarePlanActivityDetail(oSource).descriptionElement.Clone;
end;

procedure TFhirCarePlanActivityDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'reasonCode') Then
    list.addAll(FReasonCodeList);
  if (child_name = 'reasonReference') Then
    list.addAll(FReasonReferenceList);
  if (child_name = 'goal') Then
    list.addAll(FGoalList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'statusReason') Then
     list.add(FStatusReason.Link);
  if (child_name = 'prohibited') Then
     list.add(FProhibited.Link);
  if (child_name = 'scheduled[x]') Then
     list.add(FScheduled.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'performer') Then
    list.addAll(FPerformerList);
  if (child_name = 'product[x]') Then
     list.add(FProduct.Link);
  if (child_name = 'dailyAmount') Then
     list.add(FDailyAmount.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
end;

procedure TFhirCarePlanActivityDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reasonCode', 'CodeableConcept', FReasonCodeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reasonReference', 'Reference(Condition)', FReasonReferenceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'goal', 'Reference(Goal)', FGoalList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'statusReason', 'CodeableConcept', FStatusReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prohibited', 'boolean', FProhibited.Link));{2}
  oList.add(TFHIRProperty.create(self, 'scheduled[x]', 'Timing|Period|string', FScheduled.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner|Organization|RelatedPerson|Patient)', FPerformerList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'product[x]', 'CodeableConcept|Reference(Medication|Substance)', FProduct.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dailyAmount', 'SimpleQuantity', FDailyAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
end;

procedure TFhirCarePlanActivityDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'reasonCode') then ReasonCodeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'reasonReference') then ReasonReferenceList.add(propValue as TFhirReference{TFhirCondition}){2}
  else if (propName = 'goal') then GoalList.add(propValue as TFhirReference{TFhirGoal}){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'statusReason') then StatusReason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'prohibited') then ProhibitedElement := propValue as TFhirBoolean{5a}
  else if (propName.startsWith('scheduled')) then Scheduled := propValue as TFhirType{4}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'performer') then PerformerList.add(propValue as TFhirReference{Resource}){2}
  else if (propName.startsWith('product')) then Product := propValue as TFhirType{4}
  else if (propName = 'dailyAmount') then DailyAmount := propValue as TFhirQuantity{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirCarePlanActivityDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirCarePlanActivityDetail.Link : TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail(inherited Link);
end;

function TFhirCarePlanActivityDetail.Clone : TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail(inherited Clone);
end;

function TFhirCarePlanActivityDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanActivityDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCarePlanActivityDetail)) then
    result := false
  else
  begin
    o := TFhirCarePlanActivityDetail(other);
    result := compareDeep(categoryElement, o.categoryElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(reasonCodeList, o.reasonCodeList, true) and compareDeep(reasonReferenceList, o.reasonReferenceList, true) and 
      compareDeep(goalList, o.goalList, true) and compareDeep(statusElement, o.statusElement, true) and 
      compareDeep(statusReasonElement, o.statusReasonElement, true) and compareDeep(prohibitedElement, o.prohibitedElement, true) and 
      compareDeep(scheduledElement, o.scheduledElement, true) and compareDeep(locationElement, o.locationElement, true) and 
      compareDeep(performerList, o.performerList, true) and compareDeep(productElement, o.productElement, true) and 
      compareDeep(dailyAmountElement, o.dailyAmountElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true);
  end;
end;

function TFhirCarePlanActivityDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCarePlanActivityDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCarePlanActivityDetail)) then
    result := false
  else
  begin
    o := TFhirCarePlanActivityDetail(other);
    result := compareValues(statusElement, o.statusElement, true) and compareValues(prohibitedElement, o.prohibitedElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirCarePlanActivityDetail }

Procedure TFhirCarePlanActivityDetail.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirCarePlanActivityDetail.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirCarePlanActivityDetail.GetReasonCodeList : TFhirCodeableConceptList;
begin
  if FReasonCodeList = nil then
    FReasonCodeList := TFhirCodeableConceptList.Create;
  result := FReasonCodeList;
end;

Function TFhirCarePlanActivityDetail.GetHasReasonCodeList : boolean;
begin
  result := (FReasonCodeList <> nil) and (FReasonCodeList.count > 0);
end;

Function TFhirCarePlanActivityDetail.GetReasonReferenceList : TFhirReferenceList{TFhirCondition};
begin
  if FReasonReferenceList = nil then
    FReasonReferenceList := TFhirReferenceList{TFhirCondition}.Create;
  result := FReasonReferenceList;
end;

Function TFhirCarePlanActivityDetail.GetHasReasonReferenceList : boolean;
begin
  result := (FReasonReferenceList <> nil) and (FReasonReferenceList.count > 0);
end;

Function TFhirCarePlanActivityDetail.GetGoalList : TFhirReferenceList{TFhirGoal};
begin
  if FGoalList = nil then
    FGoalList := TFhirReferenceList{TFhirGoal}.Create;
  result := FGoalList;
end;

Function TFhirCarePlanActivityDetail.GetHasGoalList : boolean;
begin
  result := (FGoalList <> nil) and (FGoalList.count > 0);
end;

Procedure TFhirCarePlanActivityDetail.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirCarePlanActivityDetail.GetStatusST : TFhirCarePlanActivityStatus;
begin
  if FStatus = nil then
    result := TFhirCarePlanActivityStatus(0)
  else
    result := TFhirCarePlanActivityStatus(StringArrayIndexOfSensitive(CODES_TFhirCarePlanActivityStatus, FStatus.value));
end;

Procedure TFhirCarePlanActivityDetail.SetStatusST(value : TFhirCarePlanActivityStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirCarePlanActivityStatus[value]);
end;

Procedure TFhirCarePlanActivityDetail.SetStatusReason(value : TFhirCodeableConcept);
begin
  FStatusReason.free;
  FStatusReason := value;
end;

Procedure TFhirCarePlanActivityDetail.SetProhibited(value : TFhirBoolean);
begin
  FProhibited.free;
  FProhibited := value;
end;

Function TFhirCarePlanActivityDetail.GetProhibitedST : Boolean;
begin
  if FProhibited = nil then
    result := false
  else
    result := FProhibited.value;
end;

Procedure TFhirCarePlanActivityDetail.SetProhibitedST(value : Boolean);
begin
  if FProhibited = nil then
    FProhibited := TFhirBoolean.create;
  FProhibited.value := value
end;

Procedure TFhirCarePlanActivityDetail.SetScheduled(value : TFhirType);
begin
  FScheduled.free;
  FScheduled := value;
end;

Procedure TFhirCarePlanActivityDetail.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Function TFhirCarePlanActivityDetail.GetPerformerList : TFhirReferenceList{Resource};
begin
  if FPerformerList = nil then
    FPerformerList := TFhirReferenceList{Resource}.Create;
  result := FPerformerList;
end;

Function TFhirCarePlanActivityDetail.GetHasPerformerList : boolean;
begin
  result := (FPerformerList <> nil) and (FPerformerList.count > 0);
end;

Procedure TFhirCarePlanActivityDetail.SetProduct(value : TFhirType);
begin
  FProduct.free;
  FProduct := value;
end;

Procedure TFhirCarePlanActivityDetail.SetDailyAmount(value : TFhirQuantity);
begin
  FDailyAmount.free;
  FDailyAmount := value;
end;

Procedure TFhirCarePlanActivityDetail.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirCarePlanActivityDetail.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirCarePlanActivityDetail.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirCarePlanActivityDetail.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;


{ TFhirCarePlanActivityDetailListEnumerator }

Constructor TFhirCarePlanActivityDetailListEnumerator.Create(list : TFhirCarePlanActivityDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCarePlanActivityDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCarePlanActivityDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCarePlanActivityDetailListEnumerator.GetCurrent : TFhirCarePlanActivityDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirCarePlanActivityDetailList }
procedure TFhirCarePlanActivityDetailList.AddItem(value: TFhirCarePlanActivityDetail);
begin
  assert(value.ClassName = 'TFhirCarePlanActivityDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCarePlanActivityDetail');
  add(value);
end;


function TFhirCarePlanActivityDetailList.Append: TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanActivityDetailList.ClearItems;
begin
  Clear;
end;

function TFhirCarePlanActivityDetailList.GetEnumerator : TFhirCarePlanActivityDetailListEnumerator;
begin
  result := TFhirCarePlanActivityDetailListEnumerator.Create(self.link);
end;

function TFhirCarePlanActivityDetailList.Clone: TFhirCarePlanActivityDetailList;
begin
  result := TFhirCarePlanActivityDetailList(inherited Clone);
end;

function TFhirCarePlanActivityDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCarePlanActivityDetailList.GetItemN(index: Integer): TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail(ObjectByIndex[index]);
end;

function TFhirCarePlanActivityDetailList.IndexOf(value: TFhirCarePlanActivityDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCarePlanActivityDetailList.Insert(index: Integer): TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanActivityDetailList.InsertItem(index: Integer; value: TFhirCarePlanActivityDetail);
begin
  assert(value is TFhirCarePlanActivityDetail);
  Inherited Insert(index, value);
end;

function TFhirCarePlanActivityDetailList.Item(index: Integer): TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail(ObjectByIndex[index]);
end;

function TFhirCarePlanActivityDetailList.Link: TFhirCarePlanActivityDetailList;
begin
  result := TFhirCarePlanActivityDetailList(inherited Link);
end;

procedure TFhirCarePlanActivityDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCarePlanActivityDetailList.SetItemByIndex(index: Integer; value: TFhirCarePlanActivityDetail);
begin
  assert(value is TFhirCarePlanActivityDetail);
  FhirCarePlanActivityDetails[index] := value;
end;

procedure TFhirCarePlanActivityDetailList.SetItemN(index: Integer; value: TFhirCarePlanActivityDetail);
begin
  assert(value is TFhirCarePlanActivityDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirCarePlan }

constructor TFhirCarePlan.Create;
begin
  inherited;
end;

destructor TFhirCarePlan.Destroy;
begin
  FIdentifierList.Free;
  FSubject.free;
  FStatus.free;
  FContext.free;
  FPeriod.free;
  FAuthorList.Free;
  FModified.free;
  FCategoryList.Free;
  FDescription.free;
  FAddressesList.Free;
  FSupportList.Free;
  FRelatedPlanList.Free;
  FParticipantList.Free;
  FGoalList.Free;
  FActivityList.Free;
  FNote.free;
  inherited;
end;

function TFhirCarePlan.GetResourceType : TFhirResourceType;
begin
  result := frtCarePlan;
end;

procedure TFhirCarePlan.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCarePlan(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirCarePlan(oSource).FIdentifierList);
  end;
  subject := TFhirCarePlan(oSource).subject.Clone;
  FStatus := TFhirCarePlan(oSource).FStatus.Link;
  context := TFhirCarePlan(oSource).context.Clone;
  period := TFhirCarePlan(oSource).period.Clone;
  if (TFhirCarePlan(oSource).FAuthorList = nil) then
  begin
    FAuthorList.free;
    FAuthorList := nil;
  end
  else
  begin
    FAuthorList := TFhirReferenceList{Resource}.Create;
    FAuthorList.Assign(TFhirCarePlan(oSource).FAuthorList);
  end;
  modifiedElement := TFhirCarePlan(oSource).modifiedElement.Clone;
  if (TFhirCarePlan(oSource).FCategoryList = nil) then
  begin
    FCategoryList.free;
    FCategoryList := nil;
  end
  else
  begin
    FCategoryList := TFhirCodeableConceptList.Create;
    FCategoryList.Assign(TFhirCarePlan(oSource).FCategoryList);
  end;
  descriptionElement := TFhirCarePlan(oSource).descriptionElement.Clone;
  if (TFhirCarePlan(oSource).FAddressesList = nil) then
  begin
    FAddressesList.free;
    FAddressesList := nil;
  end
  else
  begin
    FAddressesList := TFhirReferenceList{TFhirCondition}.Create;
    FAddressesList.Assign(TFhirCarePlan(oSource).FAddressesList);
  end;
  if (TFhirCarePlan(oSource).FSupportList = nil) then
  begin
    FSupportList.free;
    FSupportList := nil;
  end
  else
  begin
    FSupportList := TFhirReferenceList{TFhirReference}.Create;
    FSupportList.Assign(TFhirCarePlan(oSource).FSupportList);
  end;
  if (TFhirCarePlan(oSource).FRelatedPlanList = nil) then
  begin
    FRelatedPlanList.free;
    FRelatedPlanList := nil;
  end
  else
  begin
    FRelatedPlanList := TFhirCarePlanRelatedPlanList.Create;
    FRelatedPlanList.Assign(TFhirCarePlan(oSource).FRelatedPlanList);
  end;
  if (TFhirCarePlan(oSource).FParticipantList = nil) then
  begin
    FParticipantList.free;
    FParticipantList := nil;
  end
  else
  begin
    FParticipantList := TFhirCarePlanParticipantList.Create;
    FParticipantList.Assign(TFhirCarePlan(oSource).FParticipantList);
  end;
  if (TFhirCarePlan(oSource).FGoalList = nil) then
  begin
    FGoalList.free;
    FGoalList := nil;
  end
  else
  begin
    FGoalList := TFhirReferenceList{TFhirGoal}.Create;
    FGoalList.Assign(TFhirCarePlan(oSource).FGoalList);
  end;
  if (TFhirCarePlan(oSource).FActivityList = nil) then
  begin
    FActivityList.free;
    FActivityList := nil;
  end
  else
  begin
    FActivityList := TFhirCarePlanActivityList.Create;
    FActivityList.Assign(TFhirCarePlan(oSource).FActivityList);
  end;
  note := TFhirCarePlan(oSource).note.Clone;
end;

procedure TFhirCarePlan.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'context') Then
     list.add(FContext.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'author') Then
    list.addAll(FAuthorList);
  if (child_name = 'modified') Then
     list.add(FModified.Link);
  if (child_name = 'category') Then
    list.addAll(FCategoryList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'addresses') Then
    list.addAll(FAddressesList);
  if (child_name = 'support') Then
    list.addAll(FSupportList);
  if (child_name = 'relatedPlan') Then
    list.addAll(FRelatedPlanList);
  if (child_name = 'participant') Then
    list.addAll(FParticipantList);
  if (child_name = 'goal') Then
    list.addAll(FGoalList);
  if (child_name = 'activity') Then
    list.addAll(FActivityList);
  if (child_name = 'note') Then
     list.add(FNote.Link);
end;

procedure TFhirCarePlan.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'context', 'Reference(Encounter|EpisodeOfCare)', FContext.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Patient|Practitioner|RelatedPerson|Organization)', FAuthorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'modified', 'dateTime', FModified.Link));{2}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategoryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'addresses', 'Reference(Condition)', FAddressesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'support', 'Reference(Any)', FSupportList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'relatedPlan', '', FRelatedPlanList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'participant', '', FParticipantList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'goal', 'Reference(Goal)', FGoalList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'activity', '', FActivityList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNote.Link));{2}
end;

procedure TFhirCarePlan.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'context') then Context := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'author') then AuthorList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'modified') then ModifiedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'category') then CategoryList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'addresses') then AddressesList.add(propValue as TFhirReference{TFhirCondition}){2}
  else if (propName = 'support') then SupportList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'relatedPlan') then RelatedPlanList.add(propValue as TFhirCarePlanRelatedPlan){2}
  else if (propName = 'participant') then ParticipantList.add(propValue as TFhirCarePlanParticipant){2}
  else if (propName = 'goal') then GoalList.add(propValue as TFhirReference{TFhirGoal}){2}
  else if (propName = 'activity') then ActivityList.add(propValue as TFhirCarePlanActivity){2}
  else if (propName = 'note') then Note := propValue as TFhirAnnotation{4b}
  else inherited;
end;

function TFhirCarePlan.FhirType : string;
begin
  result := 'CarePlan';
end;

function TFhirCarePlan.Link : TFhirCarePlan;
begin
  result := TFhirCarePlan(inherited Link);
end;

function TFhirCarePlan.Clone : TFhirCarePlan;
begin
  result := TFhirCarePlan(inherited Clone);
end;

{ TFhirCarePlan }

Function TFhirCarePlan.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirCarePlan.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirCarePlan.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirCarePlan.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirCarePlan.GetStatusST : TFhirCarePlanStatus;
begin
  if FStatus = nil then
    result := TFhirCarePlanStatus(0)
  else
    result := TFhirCarePlanStatus(StringArrayIndexOfSensitive(CODES_TFhirCarePlanStatus, FStatus.value));
end;

Procedure TFhirCarePlan.SetStatusST(value : TFhirCarePlanStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirCarePlanStatus[value]);
end;

Procedure TFhirCarePlan.SetContext(value : TFhirReference{Resource});
begin
  FContext.free;
  FContext := value;
end;

Procedure TFhirCarePlan.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Function TFhirCarePlan.GetAuthorList : TFhirReferenceList{Resource};
begin
  if FAuthorList = nil then
    FAuthorList := TFhirReferenceList{Resource}.Create;
  result := FAuthorList;
end;

Function TFhirCarePlan.GetHasAuthorList : boolean;
begin
  result := (FAuthorList <> nil) and (FAuthorList.count > 0);
end;

Procedure TFhirCarePlan.SetModified(value : TFhirDateTime);
begin
  FModified.free;
  FModified := value;
end;

Function TFhirCarePlan.GetModifiedST : TDateAndTime;
begin
  if FModified = nil then
    result := nil
  else
    result := FModified.value;
end;

Procedure TFhirCarePlan.SetModifiedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FModified = nil then
      FModified := TFhirDateTime.create;
    FModified.value := value
  end
  else if FModified <> nil then
    FModified.value := nil;
end;

Function TFhirCarePlan.GetCategoryList : TFhirCodeableConceptList;
begin
  if FCategoryList = nil then
    FCategoryList := TFhirCodeableConceptList.Create;
  result := FCategoryList;
end;

Function TFhirCarePlan.GetHasCategoryList : boolean;
begin
  result := (FCategoryList <> nil) and (FCategoryList.count > 0);
end;

Procedure TFhirCarePlan.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirCarePlan.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirCarePlan.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirCarePlan.GetAddressesList : TFhirReferenceList{TFhirCondition};
begin
  if FAddressesList = nil then
    FAddressesList := TFhirReferenceList{TFhirCondition}.Create;
  result := FAddressesList;
end;

Function TFhirCarePlan.GetHasAddressesList : boolean;
begin
  result := (FAddressesList <> nil) and (FAddressesList.count > 0);
end;

Function TFhirCarePlan.GetSupportList : TFhirReferenceList{TFhirReference};
begin
  if FSupportList = nil then
    FSupportList := TFhirReferenceList{TFhirReference}.Create;
  result := FSupportList;
end;

Function TFhirCarePlan.GetHasSupportList : boolean;
begin
  result := (FSupportList <> nil) and (FSupportList.count > 0);
end;

Function TFhirCarePlan.GetRelatedPlanList : TFhirCarePlanRelatedPlanList;
begin
  if FRelatedPlanList = nil then
    FRelatedPlanList := TFhirCarePlanRelatedPlanList.Create;
  result := FRelatedPlanList;
end;

Function TFhirCarePlan.GetHasRelatedPlanList : boolean;
begin
  result := (FRelatedPlanList <> nil) and (FRelatedPlanList.count > 0);
end;

Function TFhirCarePlan.GetParticipantList : TFhirCarePlanParticipantList;
begin
  if FParticipantList = nil then
    FParticipantList := TFhirCarePlanParticipantList.Create;
  result := FParticipantList;
end;

Function TFhirCarePlan.GetHasParticipantList : boolean;
begin
  result := (FParticipantList <> nil) and (FParticipantList.count > 0);
end;

Function TFhirCarePlan.GetGoalList : TFhirReferenceList{TFhirGoal};
begin
  if FGoalList = nil then
    FGoalList := TFhirReferenceList{TFhirGoal}.Create;
  result := FGoalList;
end;

Function TFhirCarePlan.GetHasGoalList : boolean;
begin
  result := (FGoalList <> nil) and (FGoalList.count > 0);
end;

Function TFhirCarePlan.GetActivityList : TFhirCarePlanActivityList;
begin
  if FActivityList = nil then
    FActivityList := TFhirCarePlanActivityList.Create;
  result := FActivityList;
end;

Function TFhirCarePlan.GetHasActivityList : boolean;
begin
  result := (FActivityList <> nil) and (FActivityList.count > 0);
end;

Procedure TFhirCarePlan.SetNote(value : TFhirAnnotation);
begin
  FNote.free;
  FNote := value;
end;


{ TFhirCarePlanListEnumerator }

Constructor TFhirCarePlanListEnumerator.Create(list : TFhirCarePlanList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCarePlanListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCarePlanListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCarePlanListEnumerator.GetCurrent : TFhirCarePlan;
begin
  Result := FList[FIndex];
end;


{ TFhirCarePlanList }
procedure TFhirCarePlanList.AddItem(value: TFhirCarePlan);
begin
  assert(value.ClassName = 'TFhirCarePlan', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCarePlan');
  add(value);
end;


function TFhirCarePlanList.Append: TFhirCarePlan;
begin
  result := TFhirCarePlan.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanList.ClearItems;
begin
  Clear;
end;

function TFhirCarePlanList.GetEnumerator : TFhirCarePlanListEnumerator;
begin
  result := TFhirCarePlanListEnumerator.Create(self.link);
end;

function TFhirCarePlanList.Clone: TFhirCarePlanList;
begin
  result := TFhirCarePlanList(inherited Clone);
end;

function TFhirCarePlanList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCarePlanList.GetItemN(index: Integer): TFhirCarePlan;
begin
  result := TFhirCarePlan(ObjectByIndex[index]);
end;

function TFhirCarePlanList.IndexOf(value: TFhirCarePlan): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCarePlanList.Insert(index: Integer): TFhirCarePlan;
begin
  result := TFhirCarePlan.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCarePlanList.InsertItem(index: Integer; value: TFhirCarePlan);
begin
  assert(value is TFhirCarePlan);
  Inherited Insert(index, value);
end;

function TFhirCarePlanList.Item(index: Integer): TFhirCarePlan;
begin
  result := TFhirCarePlan(ObjectByIndex[index]);
end;

function TFhirCarePlanList.Link: TFhirCarePlanList;
begin
  result := TFhirCarePlanList(inherited Link);
end;

procedure TFhirCarePlanList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCarePlanList.SetItemByIndex(index: Integer; value: TFhirCarePlan);
begin
  assert(value is TFhirCarePlan);
  FhirCarePlans[index] := value;
end;

procedure TFhirCarePlanList.SetItemN(index: Integer; value: TFhirCarePlan);
begin
  assert(value is TFhirCarePlan);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimPayee }

constructor TFhirClaimPayee.Create;
begin
  inherited;
end;

destructor TFhirClaimPayee.Destroy;
begin
  FType_.free;
  FProvider.free;
  FOrganization.free;
  FPerson.free;
  inherited;
end;

procedure TFhirClaimPayee.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirClaimPayee(oSource).type_.Clone;
  provider := TFhirClaimPayee(oSource).provider.Clone;
  organization := TFhirClaimPayee(oSource).organization.Clone;
  person := TFhirClaimPayee(oSource).person.Clone;
end;

procedure TFhirClaimPayee.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'person') Then
     list.add(FPerson.Link);
end;

procedure TFhirClaimPayee.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'person', 'Reference(Patient)', FPerson.Link));{2}
end;

procedure TFhirClaimPayee.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'person') then Person := propValue as TFhirReference{TFhirPatient}{4b}
  else inherited;
end;

function TFhirClaimPayee.FhirType : string;
begin
  result := 'payee';
end;

function TFhirClaimPayee.Link : TFhirClaimPayee;
begin
  result := TFhirClaimPayee(inherited Link);
end;

function TFhirClaimPayee.Clone : TFhirClaimPayee;
begin
  result := TFhirClaimPayee(inherited Clone);
end;

function TFhirClaimPayee.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimPayee;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimPayee)) then
    result := false
  else
  begin
    o := TFhirClaimPayee(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(providerElement, o.providerElement, true) and 
      compareDeep(organizationElement, o.organizationElement, true) and compareDeep(personElement, o.personElement, true);
  end;
end;

function TFhirClaimPayee.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimPayee;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimPayee)) then
    result := false
  else
  begin
    o := TFhirClaimPayee(other);
    result := true;
  end;
end;

{ TFhirClaimPayee }

Procedure TFhirClaimPayee.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirClaimPayee.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirClaimPayee.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirClaimPayee.SetPerson(value : TFhirReference{TFhirPatient});
begin
  FPerson.free;
  FPerson := value;
end;


{ TFhirClaimPayeeListEnumerator }

Constructor TFhirClaimPayeeListEnumerator.Create(list : TFhirClaimPayeeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimPayeeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimPayeeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimPayeeListEnumerator.GetCurrent : TFhirClaimPayee;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimPayeeList }
procedure TFhirClaimPayeeList.AddItem(value: TFhirClaimPayee);
begin
  assert(value.ClassName = 'TFhirClaimPayee', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimPayee');
  add(value);
end;


function TFhirClaimPayeeList.Append: TFhirClaimPayee;
begin
  result := TFhirClaimPayee.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimPayeeList.ClearItems;
begin
  Clear;
end;

function TFhirClaimPayeeList.GetEnumerator : TFhirClaimPayeeListEnumerator;
begin
  result := TFhirClaimPayeeListEnumerator.Create(self.link);
end;

function TFhirClaimPayeeList.Clone: TFhirClaimPayeeList;
begin
  result := TFhirClaimPayeeList(inherited Clone);
end;

function TFhirClaimPayeeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimPayeeList.GetItemN(index: Integer): TFhirClaimPayee;
begin
  result := TFhirClaimPayee(ObjectByIndex[index]);
end;

function TFhirClaimPayeeList.IndexOf(value: TFhirClaimPayee): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimPayeeList.Insert(index: Integer): TFhirClaimPayee;
begin
  result := TFhirClaimPayee.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimPayeeList.InsertItem(index: Integer; value: TFhirClaimPayee);
begin
  assert(value is TFhirClaimPayee);
  Inherited Insert(index, value);
end;

function TFhirClaimPayeeList.Item(index: Integer): TFhirClaimPayee;
begin
  result := TFhirClaimPayee(ObjectByIndex[index]);
end;

function TFhirClaimPayeeList.Link: TFhirClaimPayeeList;
begin
  result := TFhirClaimPayeeList(inherited Link);
end;

procedure TFhirClaimPayeeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimPayeeList.SetItemByIndex(index: Integer; value: TFhirClaimPayee);
begin
  assert(value is TFhirClaimPayee);
  FhirClaimPayees[index] := value;
end;

procedure TFhirClaimPayeeList.SetItemN(index: Integer; value: TFhirClaimPayee);
begin
  assert(value is TFhirClaimPayee);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimDiagnosis }

constructor TFhirClaimDiagnosis.Create;
begin
  inherited;
end;

destructor TFhirClaimDiagnosis.Destroy;
begin
  FSequence.free;
  FDiagnosis.free;
  inherited;
end;

procedure TFhirClaimDiagnosis.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimDiagnosis(oSource).sequenceElement.Clone;
  diagnosis := TFhirClaimDiagnosis(oSource).diagnosis.Clone;
end;

procedure TFhirClaimDiagnosis.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'diagnosis') Then
     list.add(FDiagnosis.Link);
end;

procedure TFhirClaimDiagnosis.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'diagnosis', 'Coding', FDiagnosis.Link));{2}
end;

procedure TFhirClaimDiagnosis.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'diagnosis') then Diagnosis := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimDiagnosis.FhirType : string;
begin
  result := 'diagnosis';
end;

function TFhirClaimDiagnosis.Link : TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis(inherited Link);
end;

function TFhirClaimDiagnosis.Clone : TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis(inherited Clone);
end;

function TFhirClaimDiagnosis.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimDiagnosis;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimDiagnosis)) then
    result := false
  else
  begin
    o := TFhirClaimDiagnosis(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(diagnosisElement, o.diagnosisElement, true);
  end;
end;

function TFhirClaimDiagnosis.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimDiagnosis;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimDiagnosis)) then
    result := false
  else
  begin
    o := TFhirClaimDiagnosis(other);
    result := compareValues(sequenceElement, o.sequenceElement, true);
  end;
end;

{ TFhirClaimDiagnosis }

Procedure TFhirClaimDiagnosis.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimDiagnosis.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimDiagnosis.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimDiagnosis.SetDiagnosis(value : TFhirCoding);
begin
  FDiagnosis.free;
  FDiagnosis := value;
end;


{ TFhirClaimDiagnosisListEnumerator }

Constructor TFhirClaimDiagnosisListEnumerator.Create(list : TFhirClaimDiagnosisList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimDiagnosisListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimDiagnosisListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimDiagnosisListEnumerator.GetCurrent : TFhirClaimDiagnosis;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimDiagnosisList }
procedure TFhirClaimDiagnosisList.AddItem(value: TFhirClaimDiagnosis);
begin
  assert(value.ClassName = 'TFhirClaimDiagnosis', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimDiagnosis');
  add(value);
end;


function TFhirClaimDiagnosisList.Append: TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimDiagnosisList.ClearItems;
begin
  Clear;
end;

function TFhirClaimDiagnosisList.GetEnumerator : TFhirClaimDiagnosisListEnumerator;
begin
  result := TFhirClaimDiagnosisListEnumerator.Create(self.link);
end;

function TFhirClaimDiagnosisList.Clone: TFhirClaimDiagnosisList;
begin
  result := TFhirClaimDiagnosisList(inherited Clone);
end;

function TFhirClaimDiagnosisList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimDiagnosisList.GetItemN(index: Integer): TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis(ObjectByIndex[index]);
end;

function TFhirClaimDiagnosisList.IndexOf(value: TFhirClaimDiagnosis): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimDiagnosisList.Insert(index: Integer): TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimDiagnosisList.InsertItem(index: Integer; value: TFhirClaimDiagnosis);
begin
  assert(value is TFhirClaimDiagnosis);
  Inherited Insert(index, value);
end;

function TFhirClaimDiagnosisList.Item(index: Integer): TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis(ObjectByIndex[index]);
end;

function TFhirClaimDiagnosisList.Link: TFhirClaimDiagnosisList;
begin
  result := TFhirClaimDiagnosisList(inherited Link);
end;

procedure TFhirClaimDiagnosisList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimDiagnosisList.SetItemByIndex(index: Integer; value: TFhirClaimDiagnosis);
begin
  assert(value is TFhirClaimDiagnosis);
  FhirClaimDiagnoses[index] := value;
end;

procedure TFhirClaimDiagnosisList.SetItemN(index: Integer; value: TFhirClaimDiagnosis);
begin
  assert(value is TFhirClaimDiagnosis);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimCoverage }

constructor TFhirClaimCoverage.Create;
begin
  inherited;
end;

destructor TFhirClaimCoverage.Destroy;
begin
  FSequence.free;
  FFocal.free;
  FCoverage.free;
  FBusinessArrangement.free;
  FRelationship.free;
  FPreAuthRefList.Free;
  FClaimResponse.free;
  FOriginalRuleset.free;
  inherited;
end;

procedure TFhirClaimCoverage.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimCoverage(oSource).sequenceElement.Clone;
  focalElement := TFhirClaimCoverage(oSource).focalElement.Clone;
  coverage := TFhirClaimCoverage(oSource).coverage.Clone;
  businessArrangementElement := TFhirClaimCoverage(oSource).businessArrangementElement.Clone;
  relationship := TFhirClaimCoverage(oSource).relationship.Clone;
  if (TFhirClaimCoverage(oSource).FPreAuthRefList = nil) then
  begin
    FPreAuthRefList.free;
    FPreAuthRefList := nil;
  end
  else
  begin
    FPreAuthRefList := TFhirStringList.Create;
    FPreAuthRefList.Assign(TFhirClaimCoverage(oSource).FPreAuthRefList);
  end;
  claimResponse := TFhirClaimCoverage(oSource).claimResponse.Clone;
  originalRuleset := TFhirClaimCoverage(oSource).originalRuleset.Clone;
end;

procedure TFhirClaimCoverage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'focal') Then
     list.add(FFocal.Link);
  if (child_name = 'coverage') Then
     list.add(FCoverage.Link);
  if (child_name = 'businessArrangement') Then
     list.add(FBusinessArrangement.Link);
  if (child_name = 'relationship') Then
     list.add(FRelationship.Link);
  if (child_name = 'preAuthRef') Then
    list.addAll(FPreAuthRefList);
  if (child_name = 'claimResponse') Then
     list.add(FClaimResponse.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
end;

procedure TFhirClaimCoverage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'focal', 'boolean', FFocal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'coverage', 'Reference(Coverage)', FCoverage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'businessArrangement', 'string', FBusinessArrangement.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relationship', 'Coding', FRelationship.Link));{2}
  oList.add(TFHIRProperty.create(self, 'preAuthRef', 'string', FPreAuthRefList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'claimResponse', 'Reference(ClaimResponse)', FClaimResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
end;

procedure TFhirClaimCoverage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'focal') then FocalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'coverage') then Coverage := propValue as TFhirReference{TFhirCoverage}{4b}
  else if (propName = 'businessArrangement') then BusinessArrangementElement := propValue as TFhirString{5a}
  else if (propName = 'relationship') then Relationship := propValue as TFhirCoding{4b}
  else if (propName = 'preAuthRef') then PreAuthRefList.add(propValue as TFhirString){2}
  else if (propName = 'claimResponse') then ClaimResponse := propValue as TFhirReference{TFhirClaimResponse}{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimCoverage.FhirType : string;
begin
  result := 'coverage';
end;

function TFhirClaimCoverage.Link : TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage(inherited Link);
end;

function TFhirClaimCoverage.Clone : TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage(inherited Clone);
end;

function TFhirClaimCoverage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimCoverage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimCoverage)) then
    result := false
  else
  begin
    o := TFhirClaimCoverage(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(focalElement, o.focalElement, true) and 
      compareDeep(coverageElement, o.coverageElement, true) and compareDeep(businessArrangementElement, o.businessArrangementElement, true) and 
      compareDeep(relationshipElement, o.relationshipElement, true) and compareDeep(preAuthRefList, o.preAuthRefList, true) and 
      compareDeep(claimResponseElement, o.claimResponseElement, true) and compareDeep(originalRulesetElement, o.originalRulesetElement, true);
  end;
end;

function TFhirClaimCoverage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimCoverage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimCoverage)) then
    result := false
  else
  begin
    o := TFhirClaimCoverage(other);
    result := compareValues(sequenceElement, o.sequenceElement, true) and compareValues(focalElement, o.focalElement, true) and 
      compareValues(businessArrangementElement, o.businessArrangementElement, true) and 
      compareValues(preAuthRefList, o.preAuthRefList, true);
  end;
end;

{ TFhirClaimCoverage }

Procedure TFhirClaimCoverage.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimCoverage.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimCoverage.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimCoverage.SetFocal(value : TFhirBoolean);
begin
  FFocal.free;
  FFocal := value;
end;

Function TFhirClaimCoverage.GetFocalST : Boolean;
begin
  if FFocal = nil then
    result := false
  else
    result := FFocal.value;
end;

Procedure TFhirClaimCoverage.SetFocalST(value : Boolean);
begin
  if FFocal = nil then
    FFocal := TFhirBoolean.create;
  FFocal.value := value
end;

Procedure TFhirClaimCoverage.SetCoverage(value : TFhirReference{TFhirCoverage});
begin
  FCoverage.free;
  FCoverage := value;
end;

Procedure TFhirClaimCoverage.SetBusinessArrangement(value : TFhirString);
begin
  FBusinessArrangement.free;
  FBusinessArrangement := value;
end;

Function TFhirClaimCoverage.GetBusinessArrangementST : String;
begin
  if FBusinessArrangement = nil then
    result := ''
  else
    result := FBusinessArrangement.value;
end;

Procedure TFhirClaimCoverage.SetBusinessArrangementST(value : String);
begin
  if value <> '' then
  begin
    if FBusinessArrangement = nil then
      FBusinessArrangement := TFhirString.create;
    FBusinessArrangement.value := value
  end
  else if FBusinessArrangement <> nil then
    FBusinessArrangement.value := '';
end;

Procedure TFhirClaimCoverage.SetRelationship(value : TFhirCoding);
begin
  FRelationship.free;
  FRelationship := value;
end;

Function TFhirClaimCoverage.GetPreAuthRefList : TFhirStringList;
begin
  if FPreAuthRefList = nil then
    FPreAuthRefList := TFhirStringList.Create;
  result := FPreAuthRefList;
end;

Function TFhirClaimCoverage.GetHasPreAuthRefList : boolean;
begin
  result := (FPreAuthRefList <> nil) and (FPreAuthRefList.count > 0);
end;

Procedure TFhirClaimCoverage.SetClaimResponse(value : TFhirReference{TFhirClaimResponse});
begin
  FClaimResponse.free;
  FClaimResponse := value;
end;

Procedure TFhirClaimCoverage.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;


{ TFhirClaimCoverageListEnumerator }

Constructor TFhirClaimCoverageListEnumerator.Create(list : TFhirClaimCoverageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimCoverageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimCoverageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimCoverageListEnumerator.GetCurrent : TFhirClaimCoverage;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimCoverageList }
procedure TFhirClaimCoverageList.AddItem(value: TFhirClaimCoverage);
begin
  assert(value.ClassName = 'TFhirClaimCoverage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimCoverage');
  add(value);
end;


function TFhirClaimCoverageList.Append: TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimCoverageList.ClearItems;
begin
  Clear;
end;

function TFhirClaimCoverageList.GetEnumerator : TFhirClaimCoverageListEnumerator;
begin
  result := TFhirClaimCoverageListEnumerator.Create(self.link);
end;

function TFhirClaimCoverageList.Clone: TFhirClaimCoverageList;
begin
  result := TFhirClaimCoverageList(inherited Clone);
end;

function TFhirClaimCoverageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimCoverageList.GetItemN(index: Integer): TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage(ObjectByIndex[index]);
end;

function TFhirClaimCoverageList.IndexOf(value: TFhirClaimCoverage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimCoverageList.Insert(index: Integer): TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimCoverageList.InsertItem(index: Integer; value: TFhirClaimCoverage);
begin
  assert(value is TFhirClaimCoverage);
  Inherited Insert(index, value);
end;

function TFhirClaimCoverageList.Item(index: Integer): TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage(ObjectByIndex[index]);
end;

function TFhirClaimCoverageList.Link: TFhirClaimCoverageList;
begin
  result := TFhirClaimCoverageList(inherited Link);
end;

procedure TFhirClaimCoverageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimCoverageList.SetItemByIndex(index: Integer; value: TFhirClaimCoverage);
begin
  assert(value is TFhirClaimCoverage);
  FhirClaimCoverages[index] := value;
end;

procedure TFhirClaimCoverageList.SetItemN(index: Integer; value: TFhirClaimCoverage);
begin
  assert(value is TFhirClaimCoverage);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimItem }

constructor TFhirClaimItem.Create;
begin
  inherited;
end;

destructor TFhirClaimItem.Destroy;
begin
  FSequence.free;
  FType_.free;
  FProvider.free;
  FDiagnosisLinkIdList.Free;
  FService.free;
  FServiceDate.free;
  FQuantity.free;
  FUnitPrice.free;
  FFactor.free;
  FPoints.free;
  FNet.free;
  FUdi.free;
  FBodySite.free;
  FSubSiteList.Free;
  FModifierList.Free;
  FDetailList.Free;
  FProsthesis.free;
  inherited;
end;

procedure TFhirClaimItem.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimItem(oSource).sequenceElement.Clone;
  type_ := TFhirClaimItem(oSource).type_.Clone;
  provider := TFhirClaimItem(oSource).provider.Clone;
  if (TFhirClaimItem(oSource).FDiagnosisLinkIdList = nil) then
  begin
    FDiagnosisLinkIdList.free;
    FDiagnosisLinkIdList := nil;
  end
  else
  begin
    FDiagnosisLinkIdList := TFhirPositiveIntList.Create;
    FDiagnosisLinkIdList.Assign(TFhirClaimItem(oSource).FDiagnosisLinkIdList);
  end;
  service := TFhirClaimItem(oSource).service.Clone;
  serviceDateElement := TFhirClaimItem(oSource).serviceDateElement.Clone;
  quantity := TFhirClaimItem(oSource).quantity.Clone;
  unitPrice := TFhirClaimItem(oSource).unitPrice.Clone;
  factorElement := TFhirClaimItem(oSource).factorElement.Clone;
  pointsElement := TFhirClaimItem(oSource).pointsElement.Clone;
  net := TFhirClaimItem(oSource).net.Clone;
  udi := TFhirClaimItem(oSource).udi.Clone;
  bodySite := TFhirClaimItem(oSource).bodySite.Clone;
  if (TFhirClaimItem(oSource).FSubSiteList = nil) then
  begin
    FSubSiteList.free;
    FSubSiteList := nil;
  end
  else
  begin
    FSubSiteList := TFhirCodingList.Create;
    FSubSiteList.Assign(TFhirClaimItem(oSource).FSubSiteList);
  end;
  if (TFhirClaimItem(oSource).FModifierList = nil) then
  begin
    FModifierList.free;
    FModifierList := nil;
  end
  else
  begin
    FModifierList := TFhirCodingList.Create;
    FModifierList.Assign(TFhirClaimItem(oSource).FModifierList);
  end;
  if (TFhirClaimItem(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirClaimItemDetailList.Create;
    FDetailList.Assign(TFhirClaimItem(oSource).FDetailList);
  end;
  prosthesis := TFhirClaimItem(oSource).prosthesis.Clone;
end;

procedure TFhirClaimItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'diagnosisLinkId') Then
    list.addAll(FDiagnosisLinkIdList);
  if (child_name = 'service') Then
     list.add(FService.Link);
  if (child_name = 'serviceDate') Then
     list.add(FServiceDate.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'unitPrice') Then
     list.add(FUnitPrice.Link);
  if (child_name = 'factor') Then
     list.add(FFactor.Link);
  if (child_name = 'points') Then
     list.add(FPoints.Link);
  if (child_name = 'net') Then
     list.add(FNet.Link);
  if (child_name = 'udi') Then
     list.add(FUdi.Link);
  if (child_name = 'bodySite') Then
     list.add(FBodySite.Link);
  if (child_name = 'subSite') Then
    list.addAll(FSubSiteList);
  if (child_name = 'modifier') Then
    list.addAll(FModifierList);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
  if (child_name = 'prosthesis') Then
     list.add(FProsthesis.Link);
end;

procedure TFhirClaimItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'diagnosisLinkId', 'positiveInt', FDiagnosisLinkIdList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'service', 'Coding', FService.Link));{2}
  oList.add(TFHIRProperty.create(self, 'serviceDate', 'date', FServiceDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unitPrice', 'Money', FUnitPrice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'factor', 'decimal', FFactor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'points', 'decimal', FPoints.Link));{2}
  oList.add(TFHIRProperty.create(self, 'net', 'Money', FNet.Link));{2}
  oList.add(TFHIRProperty.create(self, 'udi', 'Coding', FUdi.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bodySite', 'Coding', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subSite', 'Coding', FSubSiteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'modifier', 'Coding', FModifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetailList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'prosthesis', '', FProsthesis.Link));{2}
end;

procedure TFhirClaimItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'diagnosisLinkId') then DiagnosisLinkIdList.add(propValue as TFhirPositiveInt){2}
  else if (propName = 'service') then Service := propValue as TFhirCoding{4b}
  else if (propName = 'serviceDate') then ServiceDateElement := propValue as TFhirDate{5a}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'unitPrice') then UnitPrice := propValue as TFhirQuantity{4b}
  else if (propName = 'factor') then FactorElement := propValue as TFhirDecimal{5a}
  else if (propName = 'points') then PointsElement := propValue as TFhirDecimal{5a}
  else if (propName = 'net') then Net := propValue as TFhirQuantity{4b}
  else if (propName = 'udi') then Udi := propValue as TFhirCoding{4b}
  else if (propName = 'bodySite') then BodySite := propValue as TFhirCoding{4b}
  else if (propName = 'subSite') then SubSiteList.add(propValue as TFhirCoding){2}
  else if (propName = 'modifier') then ModifierList.add(propValue as TFhirCoding){2}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirClaimItemDetail){2}
  else if (propName = 'prosthesis') then Prosthesis := propValue as TFhirClaimItemProsthesis{4b}
  else inherited;
end;

function TFhirClaimItem.FhirType : string;
begin
  result := 'item';
end;

function TFhirClaimItem.Link : TFhirClaimItem;
begin
  result := TFhirClaimItem(inherited Link);
end;

function TFhirClaimItem.Clone : TFhirClaimItem;
begin
  result := TFhirClaimItem(inherited Clone);
end;

function TFhirClaimItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimItem)) then
    result := false
  else
  begin
    o := TFhirClaimItem(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(providerElement, o.providerElement, true) and compareDeep(diagnosisLinkIdList, o.diagnosisLinkIdList, true) and 
      compareDeep(serviceElement, o.serviceElement, true) and compareDeep(serviceDateElement, o.serviceDateElement, true) and 
      compareDeep(quantityElement, o.quantityElement, true) and compareDeep(unitPriceElement, o.unitPriceElement, true) and 
      compareDeep(factorElement, o.factorElement, true) and compareDeep(pointsElement, o.pointsElement, true) and 
      compareDeep(netElement, o.netElement, true) and compareDeep(udiElement, o.udiElement, true) and 
      compareDeep(bodySiteElement, o.bodySiteElement, true) and compareDeep(subSiteList, o.subSiteList, true) and 
      compareDeep(modifierList, o.modifierList, true) and compareDeep(detailList, o.detailList, true) and 
      compareDeep(prosthesisElement, o.prosthesisElement, true);
  end;
end;

function TFhirClaimItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimItem)) then
    result := false
  else
  begin
    o := TFhirClaimItem(other);
    result := compareValues(sequenceElement, o.sequenceElement, true) and compareValues(diagnosisLinkIdList, o.diagnosisLinkIdList, true) and 
      compareValues(serviceDateElement, o.serviceDateElement, true) and compareValues(factorElement, o.factorElement, true) and 
      compareValues(pointsElement, o.pointsElement, true);
  end;
end;

{ TFhirClaimItem }

Procedure TFhirClaimItem.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimItem.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimItem.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimItem.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirClaimItem.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Function TFhirClaimItem.GetDiagnosisLinkIdList : TFhirPositiveIntList;
begin
  if FDiagnosisLinkIdList = nil then
    FDiagnosisLinkIdList := TFhirPositiveIntList.Create;
  result := FDiagnosisLinkIdList;
end;

Function TFhirClaimItem.GetHasDiagnosisLinkIdList : boolean;
begin
  result := (FDiagnosisLinkIdList <> nil) and (FDiagnosisLinkIdList.count > 0);
end;

Procedure TFhirClaimItem.SetService(value : TFhirCoding);
begin
  FService.free;
  FService := value;
end;

Procedure TFhirClaimItem.SetServiceDate(value : TFhirDate);
begin
  FServiceDate.free;
  FServiceDate := value;
end;

Function TFhirClaimItem.GetServiceDateST : TDateAndTime;
begin
  if FServiceDate = nil then
    result := nil
  else
    result := FServiceDate.value;
end;

Procedure TFhirClaimItem.SetServiceDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FServiceDate = nil then
      FServiceDate := TFhirDate.create;
    FServiceDate.value := value
  end
  else if FServiceDate <> nil then
    FServiceDate.value := nil;
end;

Procedure TFhirClaimItem.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirClaimItem.SetUnitPrice(value : TFhirQuantity);
begin
  FUnitPrice.free;
  FUnitPrice := value;
end;

Procedure TFhirClaimItem.SetFactor(value : TFhirDecimal);
begin
  FFactor.free;
  FFactor := value;
end;

Function TFhirClaimItem.GetFactorST : String;
begin
  if FFactor = nil then
    result := ''
  else
    result := FFactor.value;
end;

Procedure TFhirClaimItem.SetFactorST(value : String);
begin
  if value <> '' then
  begin
    if FFactor = nil then
      FFactor := TFhirDecimal.create;
    FFactor.value := value
  end
  else if FFactor <> nil then
    FFactor.value := '';
end;

Procedure TFhirClaimItem.SetPoints(value : TFhirDecimal);
begin
  FPoints.free;
  FPoints := value;
end;

Function TFhirClaimItem.GetPointsST : String;
begin
  if FPoints = nil then
    result := ''
  else
    result := FPoints.value;
end;

Procedure TFhirClaimItem.SetPointsST(value : String);
begin
  if value <> '' then
  begin
    if FPoints = nil then
      FPoints := TFhirDecimal.create;
    FPoints.value := value
  end
  else if FPoints <> nil then
    FPoints.value := '';
end;

Procedure TFhirClaimItem.SetNet(value : TFhirQuantity);
begin
  FNet.free;
  FNet := value;
end;

Procedure TFhirClaimItem.SetUdi(value : TFhirCoding);
begin
  FUdi.free;
  FUdi := value;
end;

Procedure TFhirClaimItem.SetBodySite(value : TFhirCoding);
begin
  FBodySite.free;
  FBodySite := value;
end;

Function TFhirClaimItem.GetSubSiteList : TFhirCodingList;
begin
  if FSubSiteList = nil then
    FSubSiteList := TFhirCodingList.Create;
  result := FSubSiteList;
end;

Function TFhirClaimItem.GetHasSubSiteList : boolean;
begin
  result := (FSubSiteList <> nil) and (FSubSiteList.count > 0);
end;

Function TFhirClaimItem.GetModifierList : TFhirCodingList;
begin
  if FModifierList = nil then
    FModifierList := TFhirCodingList.Create;
  result := FModifierList;
end;

Function TFhirClaimItem.GetHasModifierList : boolean;
begin
  result := (FModifierList <> nil) and (FModifierList.count > 0);
end;

Function TFhirClaimItem.GetDetailList : TFhirClaimItemDetailList;
begin
  if FDetailList = nil then
    FDetailList := TFhirClaimItemDetailList.Create;
  result := FDetailList;
end;

Function TFhirClaimItem.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;

Procedure TFhirClaimItem.SetProsthesis(value : TFhirClaimItemProsthesis);
begin
  FProsthesis.free;
  FProsthesis := value;
end;


{ TFhirClaimItemListEnumerator }

Constructor TFhirClaimItemListEnumerator.Create(list : TFhirClaimItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimItemListEnumerator.GetCurrent : TFhirClaimItem;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimItemList }
procedure TFhirClaimItemList.AddItem(value: TFhirClaimItem);
begin
  assert(value.ClassName = 'TFhirClaimItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimItem');
  add(value);
end;


function TFhirClaimItemList.Append: TFhirClaimItem;
begin
  result := TFhirClaimItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemList.ClearItems;
begin
  Clear;
end;

function TFhirClaimItemList.GetEnumerator : TFhirClaimItemListEnumerator;
begin
  result := TFhirClaimItemListEnumerator.Create(self.link);
end;

function TFhirClaimItemList.Clone: TFhirClaimItemList;
begin
  result := TFhirClaimItemList(inherited Clone);
end;

function TFhirClaimItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimItemList.GetItemN(index: Integer): TFhirClaimItem;
begin
  result := TFhirClaimItem(ObjectByIndex[index]);
end;

function TFhirClaimItemList.IndexOf(value: TFhirClaimItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimItemList.Insert(index: Integer): TFhirClaimItem;
begin
  result := TFhirClaimItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemList.InsertItem(index: Integer; value: TFhirClaimItem);
begin
  assert(value is TFhirClaimItem);
  Inherited Insert(index, value);
end;

function TFhirClaimItemList.Item(index: Integer): TFhirClaimItem;
begin
  result := TFhirClaimItem(ObjectByIndex[index]);
end;

function TFhirClaimItemList.Link: TFhirClaimItemList;
begin
  result := TFhirClaimItemList(inherited Link);
end;

procedure TFhirClaimItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimItemList.SetItemByIndex(index: Integer; value: TFhirClaimItem);
begin
  assert(value is TFhirClaimItem);
  FhirClaimItems[index] := value;
end;

procedure TFhirClaimItemList.SetItemN(index: Integer; value: TFhirClaimItem);
begin
  assert(value is TFhirClaimItem);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimItemDetail }

constructor TFhirClaimItemDetail.Create;
begin
  inherited;
end;

destructor TFhirClaimItemDetail.Destroy;
begin
  FSequence.free;
  FType_.free;
  FService.free;
  FQuantity.free;
  FUnitPrice.free;
  FFactor.free;
  FPoints.free;
  FNet.free;
  FUdi.free;
  FSubDetailList.Free;
  inherited;
end;

procedure TFhirClaimItemDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimItemDetail(oSource).sequenceElement.Clone;
  type_ := TFhirClaimItemDetail(oSource).type_.Clone;
  service := TFhirClaimItemDetail(oSource).service.Clone;
  quantity := TFhirClaimItemDetail(oSource).quantity.Clone;
  unitPrice := TFhirClaimItemDetail(oSource).unitPrice.Clone;
  factorElement := TFhirClaimItemDetail(oSource).factorElement.Clone;
  pointsElement := TFhirClaimItemDetail(oSource).pointsElement.Clone;
  net := TFhirClaimItemDetail(oSource).net.Clone;
  udi := TFhirClaimItemDetail(oSource).udi.Clone;
  if (TFhirClaimItemDetail(oSource).FSubDetailList = nil) then
  begin
    FSubDetailList.free;
    FSubDetailList := nil;
  end
  else
  begin
    FSubDetailList := TFhirClaimItemDetailSubDetailList.Create;
    FSubDetailList.Assign(TFhirClaimItemDetail(oSource).FSubDetailList);
  end;
end;

procedure TFhirClaimItemDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'service') Then
     list.add(FService.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'unitPrice') Then
     list.add(FUnitPrice.Link);
  if (child_name = 'factor') Then
     list.add(FFactor.Link);
  if (child_name = 'points') Then
     list.add(FPoints.Link);
  if (child_name = 'net') Then
     list.add(FNet.Link);
  if (child_name = 'udi') Then
     list.add(FUdi.Link);
  if (child_name = 'subDetail') Then
    list.addAll(FSubDetailList);
end;

procedure TFhirClaimItemDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'service', 'Coding', FService.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unitPrice', 'Money', FUnitPrice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'factor', 'decimal', FFactor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'points', 'decimal', FPoints.Link));{2}
  oList.add(TFHIRProperty.create(self, 'net', 'Money', FNet.Link));{2}
  oList.add(TFHIRProperty.create(self, 'udi', 'Coding', FUdi.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subDetail', '', FSubDetailList.Link)){3};
end;

procedure TFhirClaimItemDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'service') then Service := propValue as TFhirCoding{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'unitPrice') then UnitPrice := propValue as TFhirQuantity{4b}
  else if (propName = 'factor') then FactorElement := propValue as TFhirDecimal{5a}
  else if (propName = 'points') then PointsElement := propValue as TFhirDecimal{5a}
  else if (propName = 'net') then Net := propValue as TFhirQuantity{4b}
  else if (propName = 'udi') then Udi := propValue as TFhirCoding{4b}
  else if (propName = 'subDetail') then SubDetailList.add(propValue as TFhirClaimItemDetailSubDetail){2}
  else inherited;
end;

function TFhirClaimItemDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirClaimItemDetail.Link : TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail(inherited Link);
end;

function TFhirClaimItemDetail.Clone : TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail(inherited Clone);
end;

function TFhirClaimItemDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimItemDetail(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(serviceElement, o.serviceElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(unitPriceElement, o.unitPriceElement, true) and compareDeep(factorElement, o.factorElement, true) and 
      compareDeep(pointsElement, o.pointsElement, true) and compareDeep(netElement, o.netElement, true) and 
      compareDeep(udiElement, o.udiElement, true) and compareDeep(subDetailList, o.subDetailList, true);
  end;
end;

function TFhirClaimItemDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimItemDetail(other);
    result := compareValues(sequenceElement, o.sequenceElement, true) and compareValues(factorElement, o.factorElement, true) and 
      compareValues(pointsElement, o.pointsElement, true);
  end;
end;

{ TFhirClaimItemDetail }

Procedure TFhirClaimItemDetail.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimItemDetail.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimItemDetail.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimItemDetail.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirClaimItemDetail.SetService(value : TFhirCoding);
begin
  FService.free;
  FService := value;
end;

Procedure TFhirClaimItemDetail.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirClaimItemDetail.SetUnitPrice(value : TFhirQuantity);
begin
  FUnitPrice.free;
  FUnitPrice := value;
end;

Procedure TFhirClaimItemDetail.SetFactor(value : TFhirDecimal);
begin
  FFactor.free;
  FFactor := value;
end;

Function TFhirClaimItemDetail.GetFactorST : String;
begin
  if FFactor = nil then
    result := ''
  else
    result := FFactor.value;
end;

Procedure TFhirClaimItemDetail.SetFactorST(value : String);
begin
  if value <> '' then
  begin
    if FFactor = nil then
      FFactor := TFhirDecimal.create;
    FFactor.value := value
  end
  else if FFactor <> nil then
    FFactor.value := '';
end;

Procedure TFhirClaimItemDetail.SetPoints(value : TFhirDecimal);
begin
  FPoints.free;
  FPoints := value;
end;

Function TFhirClaimItemDetail.GetPointsST : String;
begin
  if FPoints = nil then
    result := ''
  else
    result := FPoints.value;
end;

Procedure TFhirClaimItemDetail.SetPointsST(value : String);
begin
  if value <> '' then
  begin
    if FPoints = nil then
      FPoints := TFhirDecimal.create;
    FPoints.value := value
  end
  else if FPoints <> nil then
    FPoints.value := '';
end;

Procedure TFhirClaimItemDetail.SetNet(value : TFhirQuantity);
begin
  FNet.free;
  FNet := value;
end;

Procedure TFhirClaimItemDetail.SetUdi(value : TFhirCoding);
begin
  FUdi.free;
  FUdi := value;
end;

Function TFhirClaimItemDetail.GetSubDetailList : TFhirClaimItemDetailSubDetailList;
begin
  if FSubDetailList = nil then
    FSubDetailList := TFhirClaimItemDetailSubDetailList.Create;
  result := FSubDetailList;
end;

Function TFhirClaimItemDetail.GetHasSubDetailList : boolean;
begin
  result := (FSubDetailList <> nil) and (FSubDetailList.count > 0);
end;


{ TFhirClaimItemDetailListEnumerator }

Constructor TFhirClaimItemDetailListEnumerator.Create(list : TFhirClaimItemDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimItemDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimItemDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimItemDetailListEnumerator.GetCurrent : TFhirClaimItemDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimItemDetailList }
procedure TFhirClaimItemDetailList.AddItem(value: TFhirClaimItemDetail);
begin
  assert(value.ClassName = 'TFhirClaimItemDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimItemDetail');
  add(value);
end;


function TFhirClaimItemDetailList.Append: TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemDetailList.ClearItems;
begin
  Clear;
end;

function TFhirClaimItemDetailList.GetEnumerator : TFhirClaimItemDetailListEnumerator;
begin
  result := TFhirClaimItemDetailListEnumerator.Create(self.link);
end;

function TFhirClaimItemDetailList.Clone: TFhirClaimItemDetailList;
begin
  result := TFhirClaimItemDetailList(inherited Clone);
end;

function TFhirClaimItemDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimItemDetailList.GetItemN(index: Integer): TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimItemDetailList.IndexOf(value: TFhirClaimItemDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimItemDetailList.Insert(index: Integer): TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemDetailList.InsertItem(index: Integer; value: TFhirClaimItemDetail);
begin
  assert(value is TFhirClaimItemDetail);
  Inherited Insert(index, value);
end;

function TFhirClaimItemDetailList.Item(index: Integer): TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimItemDetailList.Link: TFhirClaimItemDetailList;
begin
  result := TFhirClaimItemDetailList(inherited Link);
end;

procedure TFhirClaimItemDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimItemDetailList.SetItemByIndex(index: Integer; value: TFhirClaimItemDetail);
begin
  assert(value is TFhirClaimItemDetail);
  FhirClaimItemDetails[index] := value;
end;

procedure TFhirClaimItemDetailList.SetItemN(index: Integer; value: TFhirClaimItemDetail);
begin
  assert(value is TFhirClaimItemDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimItemDetailSubDetail }

constructor TFhirClaimItemDetailSubDetail.Create;
begin
  inherited;
end;

destructor TFhirClaimItemDetailSubDetail.Destroy;
begin
  FSequence.free;
  FType_.free;
  FService.free;
  FQuantity.free;
  FUnitPrice.free;
  FFactor.free;
  FPoints.free;
  FNet.free;
  FUdi.free;
  inherited;
end;

procedure TFhirClaimItemDetailSubDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimItemDetailSubDetail(oSource).sequenceElement.Clone;
  type_ := TFhirClaimItemDetailSubDetail(oSource).type_.Clone;
  service := TFhirClaimItemDetailSubDetail(oSource).service.Clone;
  quantity := TFhirClaimItemDetailSubDetail(oSource).quantity.Clone;
  unitPrice := TFhirClaimItemDetailSubDetail(oSource).unitPrice.Clone;
  factorElement := TFhirClaimItemDetailSubDetail(oSource).factorElement.Clone;
  pointsElement := TFhirClaimItemDetailSubDetail(oSource).pointsElement.Clone;
  net := TFhirClaimItemDetailSubDetail(oSource).net.Clone;
  udi := TFhirClaimItemDetailSubDetail(oSource).udi.Clone;
end;

procedure TFhirClaimItemDetailSubDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'service') Then
     list.add(FService.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'unitPrice') Then
     list.add(FUnitPrice.Link);
  if (child_name = 'factor') Then
     list.add(FFactor.Link);
  if (child_name = 'points') Then
     list.add(FPoints.Link);
  if (child_name = 'net') Then
     list.add(FNet.Link);
  if (child_name = 'udi') Then
     list.add(FUdi.Link);
end;

procedure TFhirClaimItemDetailSubDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'service', 'Coding', FService.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unitPrice', 'Money', FUnitPrice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'factor', 'decimal', FFactor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'points', 'decimal', FPoints.Link));{2}
  oList.add(TFHIRProperty.create(self, 'net', 'Money', FNet.Link));{2}
  oList.add(TFHIRProperty.create(self, 'udi', 'Coding', FUdi.Link));{2}
end;

procedure TFhirClaimItemDetailSubDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'service') then Service := propValue as TFhirCoding{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'unitPrice') then UnitPrice := propValue as TFhirQuantity{4b}
  else if (propName = 'factor') then FactorElement := propValue as TFhirDecimal{5a}
  else if (propName = 'points') then PointsElement := propValue as TFhirDecimal{5a}
  else if (propName = 'net') then Net := propValue as TFhirQuantity{4b}
  else if (propName = 'udi') then Udi := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimItemDetailSubDetail.FhirType : string;
begin
  result := 'subDetail';
end;

function TFhirClaimItemDetailSubDetail.Link : TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail(inherited Link);
end;

function TFhirClaimItemDetailSubDetail.Clone : TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail(inherited Clone);
end;

function TFhirClaimItemDetailSubDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemDetailSubDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimItemDetailSubDetail)) then
    result := false
  else
  begin
    o := TFhirClaimItemDetailSubDetail(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(serviceElement, o.serviceElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(unitPriceElement, o.unitPriceElement, true) and compareDeep(factorElement, o.factorElement, true) and 
      compareDeep(pointsElement, o.pointsElement, true) and compareDeep(netElement, o.netElement, true) and 
      compareDeep(udiElement, o.udiElement, true);
  end;
end;

function TFhirClaimItemDetailSubDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemDetailSubDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimItemDetailSubDetail)) then
    result := false
  else
  begin
    o := TFhirClaimItemDetailSubDetail(other);
    result := compareValues(sequenceElement, o.sequenceElement, true) and compareValues(factorElement, o.factorElement, true) and 
      compareValues(pointsElement, o.pointsElement, true);
  end;
end;

{ TFhirClaimItemDetailSubDetail }

Procedure TFhirClaimItemDetailSubDetail.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimItemDetailSubDetail.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimItemDetailSubDetail.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetService(value : TFhirCoding);
begin
  FService.free;
  FService := value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetUnitPrice(value : TFhirQuantity);
begin
  FUnitPrice.free;
  FUnitPrice := value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetFactor(value : TFhirDecimal);
begin
  FFactor.free;
  FFactor := value;
end;

Function TFhirClaimItemDetailSubDetail.GetFactorST : String;
begin
  if FFactor = nil then
    result := ''
  else
    result := FFactor.value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetFactorST(value : String);
begin
  if value <> '' then
  begin
    if FFactor = nil then
      FFactor := TFhirDecimal.create;
    FFactor.value := value
  end
  else if FFactor <> nil then
    FFactor.value := '';
end;

Procedure TFhirClaimItemDetailSubDetail.SetPoints(value : TFhirDecimal);
begin
  FPoints.free;
  FPoints := value;
end;

Function TFhirClaimItemDetailSubDetail.GetPointsST : String;
begin
  if FPoints = nil then
    result := ''
  else
    result := FPoints.value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetPointsST(value : String);
begin
  if value <> '' then
  begin
    if FPoints = nil then
      FPoints := TFhirDecimal.create;
    FPoints.value := value
  end
  else if FPoints <> nil then
    FPoints.value := '';
end;

Procedure TFhirClaimItemDetailSubDetail.SetNet(value : TFhirQuantity);
begin
  FNet.free;
  FNet := value;
end;

Procedure TFhirClaimItemDetailSubDetail.SetUdi(value : TFhirCoding);
begin
  FUdi.free;
  FUdi := value;
end;


{ TFhirClaimItemDetailSubDetailListEnumerator }

Constructor TFhirClaimItemDetailSubDetailListEnumerator.Create(list : TFhirClaimItemDetailSubDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimItemDetailSubDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimItemDetailSubDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimItemDetailSubDetailListEnumerator.GetCurrent : TFhirClaimItemDetailSubDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimItemDetailSubDetailList }
procedure TFhirClaimItemDetailSubDetailList.AddItem(value: TFhirClaimItemDetailSubDetail);
begin
  assert(value.ClassName = 'TFhirClaimItemDetailSubDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimItemDetailSubDetail');
  add(value);
end;


function TFhirClaimItemDetailSubDetailList.Append: TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemDetailSubDetailList.ClearItems;
begin
  Clear;
end;

function TFhirClaimItemDetailSubDetailList.GetEnumerator : TFhirClaimItemDetailSubDetailListEnumerator;
begin
  result := TFhirClaimItemDetailSubDetailListEnumerator.Create(self.link);
end;

function TFhirClaimItemDetailSubDetailList.Clone: TFhirClaimItemDetailSubDetailList;
begin
  result := TFhirClaimItemDetailSubDetailList(inherited Clone);
end;

function TFhirClaimItemDetailSubDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimItemDetailSubDetailList.GetItemN(index: Integer): TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail(ObjectByIndex[index]);
end;

function TFhirClaimItemDetailSubDetailList.IndexOf(value: TFhirClaimItemDetailSubDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimItemDetailSubDetailList.Insert(index: Integer): TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemDetailSubDetailList.InsertItem(index: Integer; value: TFhirClaimItemDetailSubDetail);
begin
  assert(value is TFhirClaimItemDetailSubDetail);
  Inherited Insert(index, value);
end;

function TFhirClaimItemDetailSubDetailList.Item(index: Integer): TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail(ObjectByIndex[index]);
end;

function TFhirClaimItemDetailSubDetailList.Link: TFhirClaimItemDetailSubDetailList;
begin
  result := TFhirClaimItemDetailSubDetailList(inherited Link);
end;

procedure TFhirClaimItemDetailSubDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimItemDetailSubDetailList.SetItemByIndex(index: Integer; value: TFhirClaimItemDetailSubDetail);
begin
  assert(value is TFhirClaimItemDetailSubDetail);
  FhirClaimItemDetailSubDetails[index] := value;
end;

procedure TFhirClaimItemDetailSubDetailList.SetItemN(index: Integer; value: TFhirClaimItemDetailSubDetail);
begin
  assert(value is TFhirClaimItemDetailSubDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimItemProsthesis }

constructor TFhirClaimItemProsthesis.Create;
begin
  inherited;
end;

destructor TFhirClaimItemProsthesis.Destroy;
begin
  FInitial.free;
  FPriorDate.free;
  FPriorMaterial.free;
  inherited;
end;

procedure TFhirClaimItemProsthesis.Assign(oSource : TAdvObject);
begin
  inherited;
  initialElement := TFhirClaimItemProsthesis(oSource).initialElement.Clone;
  priorDateElement := TFhirClaimItemProsthesis(oSource).priorDateElement.Clone;
  priorMaterial := TFhirClaimItemProsthesis(oSource).priorMaterial.Clone;
end;

procedure TFhirClaimItemProsthesis.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'initial') Then
     list.add(FInitial.Link);
  if (child_name = 'priorDate') Then
     list.add(FPriorDate.Link);
  if (child_name = 'priorMaterial') Then
     list.add(FPriorMaterial.Link);
end;

procedure TFhirClaimItemProsthesis.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'initial', 'boolean', FInitial.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priorDate', 'date', FPriorDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priorMaterial', 'Coding', FPriorMaterial.Link));{2}
end;

procedure TFhirClaimItemProsthesis.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'initial') then InitialElement := propValue as TFhirBoolean{5a}
  else if (propName = 'priorDate') then PriorDateElement := propValue as TFhirDate{5a}
  else if (propName = 'priorMaterial') then PriorMaterial := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimItemProsthesis.FhirType : string;
begin
  result := 'prosthesis';
end;

function TFhirClaimItemProsthesis.Link : TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis(inherited Link);
end;

function TFhirClaimItemProsthesis.Clone : TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis(inherited Clone);
end;

function TFhirClaimItemProsthesis.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemProsthesis;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimItemProsthesis)) then
    result := false
  else
  begin
    o := TFhirClaimItemProsthesis(other);
    result := compareDeep(initialElement, o.initialElement, true) and compareDeep(priorDateElement, o.priorDateElement, true) and 
      compareDeep(priorMaterialElement, o.priorMaterialElement, true);
  end;
end;

function TFhirClaimItemProsthesis.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimItemProsthesis;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimItemProsthesis)) then
    result := false
  else
  begin
    o := TFhirClaimItemProsthesis(other);
    result := compareValues(initialElement, o.initialElement, true) and compareValues(priorDateElement, o.priorDateElement, true);
  end;
end;

{ TFhirClaimItemProsthesis }

Procedure TFhirClaimItemProsthesis.SetInitial(value : TFhirBoolean);
begin
  FInitial.free;
  FInitial := value;
end;

Function TFhirClaimItemProsthesis.GetInitialST : Boolean;
begin
  if FInitial = nil then
    result := false
  else
    result := FInitial.value;
end;

Procedure TFhirClaimItemProsthesis.SetInitialST(value : Boolean);
begin
  if FInitial = nil then
    FInitial := TFhirBoolean.create;
  FInitial.value := value
end;

Procedure TFhirClaimItemProsthesis.SetPriorDate(value : TFhirDate);
begin
  FPriorDate.free;
  FPriorDate := value;
end;

Function TFhirClaimItemProsthesis.GetPriorDateST : TDateAndTime;
begin
  if FPriorDate = nil then
    result := nil
  else
    result := FPriorDate.value;
end;

Procedure TFhirClaimItemProsthesis.SetPriorDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FPriorDate = nil then
      FPriorDate := TFhirDate.create;
    FPriorDate.value := value
  end
  else if FPriorDate <> nil then
    FPriorDate.value := nil;
end;

Procedure TFhirClaimItemProsthesis.SetPriorMaterial(value : TFhirCoding);
begin
  FPriorMaterial.free;
  FPriorMaterial := value;
end;


{ TFhirClaimItemProsthesisListEnumerator }

Constructor TFhirClaimItemProsthesisListEnumerator.Create(list : TFhirClaimItemProsthesisList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimItemProsthesisListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimItemProsthesisListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimItemProsthesisListEnumerator.GetCurrent : TFhirClaimItemProsthesis;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimItemProsthesisList }
procedure TFhirClaimItemProsthesisList.AddItem(value: TFhirClaimItemProsthesis);
begin
  assert(value.ClassName = 'TFhirClaimItemProsthesis', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimItemProsthesis');
  add(value);
end;


function TFhirClaimItemProsthesisList.Append: TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemProsthesisList.ClearItems;
begin
  Clear;
end;

function TFhirClaimItemProsthesisList.GetEnumerator : TFhirClaimItemProsthesisListEnumerator;
begin
  result := TFhirClaimItemProsthesisListEnumerator.Create(self.link);
end;

function TFhirClaimItemProsthesisList.Clone: TFhirClaimItemProsthesisList;
begin
  result := TFhirClaimItemProsthesisList(inherited Clone);
end;

function TFhirClaimItemProsthesisList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimItemProsthesisList.GetItemN(index: Integer): TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis(ObjectByIndex[index]);
end;

function TFhirClaimItemProsthesisList.IndexOf(value: TFhirClaimItemProsthesis): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimItemProsthesisList.Insert(index: Integer): TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimItemProsthesisList.InsertItem(index: Integer; value: TFhirClaimItemProsthesis);
begin
  assert(value is TFhirClaimItemProsthesis);
  Inherited Insert(index, value);
end;

function TFhirClaimItemProsthesisList.Item(index: Integer): TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis(ObjectByIndex[index]);
end;

function TFhirClaimItemProsthesisList.Link: TFhirClaimItemProsthesisList;
begin
  result := TFhirClaimItemProsthesisList(inherited Link);
end;

procedure TFhirClaimItemProsthesisList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimItemProsthesisList.SetItemByIndex(index: Integer; value: TFhirClaimItemProsthesis);
begin
  assert(value is TFhirClaimItemProsthesis);
  FhirClaimItemProstheses[index] := value;
end;

procedure TFhirClaimItemProsthesisList.SetItemN(index: Integer; value: TFhirClaimItemProsthesis);
begin
  assert(value is TFhirClaimItemProsthesis);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimMissingTeeth }

constructor TFhirClaimMissingTeeth.Create;
begin
  inherited;
end;

destructor TFhirClaimMissingTeeth.Destroy;
begin
  FTooth.free;
  FReason.free;
  FExtractionDate.free;
  inherited;
end;

procedure TFhirClaimMissingTeeth.Assign(oSource : TAdvObject);
begin
  inherited;
  tooth := TFhirClaimMissingTeeth(oSource).tooth.Clone;
  reason := TFhirClaimMissingTeeth(oSource).reason.Clone;
  extractionDateElement := TFhirClaimMissingTeeth(oSource).extractionDateElement.Clone;
end;

procedure TFhirClaimMissingTeeth.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'tooth') Then
     list.add(FTooth.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
  if (child_name = 'extractionDate') Then
     list.add(FExtractionDate.Link);
end;

procedure TFhirClaimMissingTeeth.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'tooth', 'Coding', FTooth.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'Coding', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'extractionDate', 'date', FExtractionDate.Link));{2}
end;

procedure TFhirClaimMissingTeeth.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'tooth') then Tooth := propValue as TFhirCoding{4b}
  else if (propName = 'reason') then Reason := propValue as TFhirCoding{4b}
  else if (propName = 'extractionDate') then ExtractionDateElement := propValue as TFhirDate{5a}
  else inherited;
end;

function TFhirClaimMissingTeeth.FhirType : string;
begin
  result := 'missingTeeth';
end;

function TFhirClaimMissingTeeth.Link : TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth(inherited Link);
end;

function TFhirClaimMissingTeeth.Clone : TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth(inherited Clone);
end;

function TFhirClaimMissingTeeth.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimMissingTeeth;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimMissingTeeth)) then
    result := false
  else
  begin
    o := TFhirClaimMissingTeeth(other);
    result := compareDeep(toothElement, o.toothElement, true) and compareDeep(reasonElement, o.reasonElement, true) and 
      compareDeep(extractionDateElement, o.extractionDateElement, true);
  end;
end;

function TFhirClaimMissingTeeth.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimMissingTeeth;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimMissingTeeth)) then
    result := false
  else
  begin
    o := TFhirClaimMissingTeeth(other);
    result := compareValues(extractionDateElement, o.extractionDateElement, true);
  end;
end;

{ TFhirClaimMissingTeeth }

Procedure TFhirClaimMissingTeeth.SetTooth(value : TFhirCoding);
begin
  FTooth.free;
  FTooth := value;
end;

Procedure TFhirClaimMissingTeeth.SetReason(value : TFhirCoding);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirClaimMissingTeeth.SetExtractionDate(value : TFhirDate);
begin
  FExtractionDate.free;
  FExtractionDate := value;
end;

Function TFhirClaimMissingTeeth.GetExtractionDateST : TDateAndTime;
begin
  if FExtractionDate = nil then
    result := nil
  else
    result := FExtractionDate.value;
end;

Procedure TFhirClaimMissingTeeth.SetExtractionDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FExtractionDate = nil then
      FExtractionDate := TFhirDate.create;
    FExtractionDate.value := value
  end
  else if FExtractionDate <> nil then
    FExtractionDate.value := nil;
end;


{ TFhirClaimMissingTeethListEnumerator }

Constructor TFhirClaimMissingTeethListEnumerator.Create(list : TFhirClaimMissingTeethList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimMissingTeethListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimMissingTeethListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimMissingTeethListEnumerator.GetCurrent : TFhirClaimMissingTeeth;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimMissingTeethList }
procedure TFhirClaimMissingTeethList.AddItem(value: TFhirClaimMissingTeeth);
begin
  assert(value.ClassName = 'TFhirClaimMissingTeeth', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimMissingTeeth');
  add(value);
end;


function TFhirClaimMissingTeethList.Append: TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimMissingTeethList.ClearItems;
begin
  Clear;
end;

function TFhirClaimMissingTeethList.GetEnumerator : TFhirClaimMissingTeethListEnumerator;
begin
  result := TFhirClaimMissingTeethListEnumerator.Create(self.link);
end;

function TFhirClaimMissingTeethList.Clone: TFhirClaimMissingTeethList;
begin
  result := TFhirClaimMissingTeethList(inherited Clone);
end;

function TFhirClaimMissingTeethList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimMissingTeethList.GetItemN(index: Integer): TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth(ObjectByIndex[index]);
end;

function TFhirClaimMissingTeethList.IndexOf(value: TFhirClaimMissingTeeth): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimMissingTeethList.Insert(index: Integer): TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimMissingTeethList.InsertItem(index: Integer; value: TFhirClaimMissingTeeth);
begin
  assert(value is TFhirClaimMissingTeeth);
  Inherited Insert(index, value);
end;

function TFhirClaimMissingTeethList.Item(index: Integer): TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth(ObjectByIndex[index]);
end;

function TFhirClaimMissingTeethList.Link: TFhirClaimMissingTeethList;
begin
  result := TFhirClaimMissingTeethList(inherited Link);
end;

procedure TFhirClaimMissingTeethList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimMissingTeethList.SetItemByIndex(index: Integer; value: TFhirClaimMissingTeeth);
begin
  assert(value is TFhirClaimMissingTeeth);
  FhirClaimMissingTeeths[index] := value;
end;

procedure TFhirClaimMissingTeethList.SetItemN(index: Integer; value: TFhirClaimMissingTeeth);
begin
  assert(value is TFhirClaimMissingTeeth);
  ObjectByIndex[index] := value;
end;

{ TFhirClaim }

constructor TFhirClaim.Create;
begin
  inherited;
end;

destructor TFhirClaim.Destroy;
begin
  FType_.free;
  FIdentifierList.Free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FTarget.free;
  FProvider.free;
  FOrganization.free;
  FUse.free;
  FPriority.free;
  FFundsReserve.free;
  FEnterer.free;
  FFacility.free;
  FPrescription.free;
  FOriginalPrescription.free;
  FPayee.free;
  FReferral.free;
  FDiagnosisList.Free;
  FConditionList.Free;
  FPatient.free;
  FCoverageList.Free;
  FExceptionList.Free;
  FSchool.free;
  FAccident.free;
  FAccidentType.free;
  FInterventionExceptionList.Free;
  FItemList.Free;
  FAdditionalMaterialsList.Free;
  FMissingTeethList.Free;
  inherited;
end;

function TFhirClaim.GetResourceType : TFhirResourceType;
begin
  result := frtClaim;
end;

procedure TFhirClaim.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirClaim(oSource).FType_.Link;
  if (TFhirClaim(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirClaim(oSource).FIdentifierList);
  end;
  ruleset := TFhirClaim(oSource).ruleset.Clone;
  originalRuleset := TFhirClaim(oSource).originalRuleset.Clone;
  createdElement := TFhirClaim(oSource).createdElement.Clone;
  target := TFhirClaim(oSource).target.Clone;
  provider := TFhirClaim(oSource).provider.Clone;
  organization := TFhirClaim(oSource).organization.Clone;
  FUse := TFhirClaim(oSource).FUse.Link;
  priority := TFhirClaim(oSource).priority.Clone;
  fundsReserve := TFhirClaim(oSource).fundsReserve.Clone;
  enterer := TFhirClaim(oSource).enterer.Clone;
  facility := TFhirClaim(oSource).facility.Clone;
  prescription := TFhirClaim(oSource).prescription.Clone;
  originalPrescription := TFhirClaim(oSource).originalPrescription.Clone;
  payee := TFhirClaim(oSource).payee.Clone;
  referral := TFhirClaim(oSource).referral.Clone;
  if (TFhirClaim(oSource).FDiagnosisList = nil) then
  begin
    FDiagnosisList.free;
    FDiagnosisList := nil;
  end
  else
  begin
    FDiagnosisList := TFhirClaimDiagnosisList.Create;
    FDiagnosisList.Assign(TFhirClaim(oSource).FDiagnosisList);
  end;
  if (TFhirClaim(oSource).FConditionList = nil) then
  begin
    FConditionList.free;
    FConditionList := nil;
  end
  else
  begin
    FConditionList := TFhirCodingList.Create;
    FConditionList.Assign(TFhirClaim(oSource).FConditionList);
  end;
  patient := TFhirClaim(oSource).patient.Clone;
  if (TFhirClaim(oSource).FCoverageList = nil) then
  begin
    FCoverageList.free;
    FCoverageList := nil;
  end
  else
  begin
    FCoverageList := TFhirClaimCoverageList.Create;
    FCoverageList.Assign(TFhirClaim(oSource).FCoverageList);
  end;
  if (TFhirClaim(oSource).FExceptionList = nil) then
  begin
    FExceptionList.free;
    FExceptionList := nil;
  end
  else
  begin
    FExceptionList := TFhirCodingList.Create;
    FExceptionList.Assign(TFhirClaim(oSource).FExceptionList);
  end;
  schoolElement := TFhirClaim(oSource).schoolElement.Clone;
  accidentElement := TFhirClaim(oSource).accidentElement.Clone;
  accidentType := TFhirClaim(oSource).accidentType.Clone;
  if (TFhirClaim(oSource).FInterventionExceptionList = nil) then
  begin
    FInterventionExceptionList.free;
    FInterventionExceptionList := nil;
  end
  else
  begin
    FInterventionExceptionList := TFhirCodingList.Create;
    FInterventionExceptionList.Assign(TFhirClaim(oSource).FInterventionExceptionList);
  end;
  if (TFhirClaim(oSource).FItemList = nil) then
  begin
    FItemList.free;
    FItemList := nil;
  end
  else
  begin
    FItemList := TFhirClaimItemList.Create;
    FItemList.Assign(TFhirClaim(oSource).FItemList);
  end;
  if (TFhirClaim(oSource).FAdditionalMaterialsList = nil) then
  begin
    FAdditionalMaterialsList.free;
    FAdditionalMaterialsList := nil;
  end
  else
  begin
    FAdditionalMaterialsList := TFhirCodingList.Create;
    FAdditionalMaterialsList.Assign(TFhirClaim(oSource).FAdditionalMaterialsList);
  end;
  if (TFhirClaim(oSource).FMissingTeethList = nil) then
  begin
    FMissingTeethList.free;
    FMissingTeethList := nil;
  end
  else
  begin
    FMissingTeethList := TFhirClaimMissingTeethList.Create;
    FMissingTeethList.Assign(TFhirClaim(oSource).FMissingTeethList);
  end;
end;

procedure TFhirClaim.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'use') Then
     list.add(FUse.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'fundsReserve') Then
     list.add(FFundsReserve.Link);
  if (child_name = 'enterer') Then
     list.add(FEnterer.Link);
  if (child_name = 'facility') Then
     list.add(FFacility.Link);
  if (child_name = 'prescription') Then
     list.add(FPrescription.Link);
  if (child_name = 'originalPrescription') Then
     list.add(FOriginalPrescription.Link);
  if (child_name = 'payee') Then
     list.add(FPayee.Link);
  if (child_name = 'referral') Then
     list.add(FReferral.Link);
  if (child_name = 'diagnosis') Then
    list.addAll(FDiagnosisList);
  if (child_name = 'condition') Then
    list.addAll(FConditionList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'coverage') Then
    list.addAll(FCoverageList);
  if (child_name = 'exception') Then
    list.addAll(FExceptionList);
  if (child_name = 'school') Then
     list.add(FSchool.Link);
  if (child_name = 'accident') Then
     list.add(FAccident.Link);
  if (child_name = 'accidentType') Then
     list.add(FAccidentType.Link);
  if (child_name = 'interventionException') Then
    list.addAll(FInterventionExceptionList);
  if (child_name = 'item') Then
    list.addAll(FItemList);
  if (child_name = 'additionalMaterials') Then
    list.addAll(FAdditionalMaterialsList);
  if (child_name = 'missingTeeth') Then
    list.addAll(FMissingTeethList);
end;

procedure TFhirClaim.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'use', 'code', FUse.Link));{1}
  oList.add(TFHIRProperty.create(self, 'priority', 'Coding', FPriority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fundsReserve', 'Coding', FFundsReserve.Link));{2}
  oList.add(TFHIRProperty.create(self, 'enterer', 'Reference(Practitioner)', FEnterer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'facility', 'Reference(Location)', FFacility.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prescription', 'Reference(MedicationOrder|VisionPrescription)', FPrescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalPrescription', 'Reference(MedicationOrder)', FOriginalPrescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'payee', '', FPayee.Link));{2}
  oList.add(TFHIRProperty.create(self, 'referral', 'Reference(ReferralRequest)', FReferral.Link));{2}
  oList.add(TFHIRProperty.create(self, 'diagnosis', '', FDiagnosisList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'condition', 'Coding', FConditionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'coverage', '', FCoverageList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'exception', 'Coding', FExceptionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'school', 'string', FSchool.Link));{2}
  oList.add(TFHIRProperty.create(self, 'accident', 'date', FAccident.Link));{2}
  oList.add(TFHIRProperty.create(self, 'accidentType', 'Coding', FAccidentType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'interventionException', 'Coding', FInterventionExceptionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'item', '', FItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'additionalMaterials', 'Coding', FAdditionalMaterialsList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'missingTeeth', '', FMissingTeethList.Link)){3};
end;

procedure TFhirClaim.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'use') then UseElement := propValue as TFHIREnum
  else if (propName = 'priority') then Priority := propValue as TFhirCoding{4b}
  else if (propName = 'fundsReserve') then FundsReserve := propValue as TFhirCoding{4b}
  else if (propName = 'enterer') then Enterer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'facility') then Facility := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'prescription') then Prescription := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'originalPrescription') then OriginalPrescription := propValue as TFhirReference{TFhirMedicationOrder}{4b}
  else if (propName = 'payee') then Payee := propValue as TFhirClaimPayee{4b}
  else if (propName = 'referral') then Referral := propValue as TFhirReference{TFhirReferralRequest}{4b}
  else if (propName = 'diagnosis') then DiagnosisList.add(propValue as TFhirClaimDiagnosis){2}
  else if (propName = 'condition') then ConditionList.add(propValue as TFhirCoding){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'coverage') then CoverageList.add(propValue as TFhirClaimCoverage){2}
  else if (propName = 'exception') then ExceptionList.add(propValue as TFhirCoding){2}
  else if (propName = 'school') then SchoolElement := propValue as TFhirString{5a}
  else if (propName = 'accident') then AccidentElement := propValue as TFhirDate{5a}
  else if (propName = 'accidentType') then AccidentType := propValue as TFhirCoding{4b}
  else if (propName = 'interventionException') then InterventionExceptionList.add(propValue as TFhirCoding){2}
  else if (propName = 'item') then ItemList.add(propValue as TFhirClaimItem){2}
  else if (propName = 'additionalMaterials') then AdditionalMaterialsList.add(propValue as TFhirCoding){2}
  else if (propName = 'missingTeeth') then MissingTeethList.add(propValue as TFhirClaimMissingTeeth){2}
  else inherited;
end;

function TFhirClaim.FhirType : string;
begin
  result := 'Claim';
end;

function TFhirClaim.Link : TFhirClaim;
begin
  result := TFhirClaim(inherited Link);
end;

function TFhirClaim.Clone : TFhirClaim;
begin
  result := TFhirClaim(inherited Clone);
end;

{ TFhirClaim }

Procedure TFhirClaim.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirClaim.GetType_ST : TFhirClaimTypeLink;
begin
  if FType_ = nil then
    result := TFhirClaimTypeLink(0)
  else
    result := TFhirClaimTypeLink(StringArrayIndexOfSensitive(CODES_TFhirClaimTypeLink, FType_.value));
end;

Procedure TFhirClaim.SetType_ST(value : TFhirClaimTypeLink);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirClaimTypeLink[value]);
end;

Function TFhirClaim.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirClaim.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirClaim.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirClaim.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirClaim.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirClaim.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirClaim.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirClaim.SetTarget(value : TFhirReference{TFhirOrganization});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirClaim.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirClaim.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirClaim.SetUse(value : TFhirEnum);
begin
  FUse.free;
  FUse := value;
end;

Function TFhirClaim.GetUseST : TFhirClaimUseLink;
begin
  if FUse = nil then
    result := TFhirClaimUseLink(0)
  else
    result := TFhirClaimUseLink(StringArrayIndexOfSensitive(CODES_TFhirClaimUseLink, FUse.value));
end;

Procedure TFhirClaim.SetUseST(value : TFhirClaimUseLink);
begin
  if ord(value) = 0 then
    UseElement := nil
  else
    UseElement := TFhirEnum.create(CODES_TFhirClaimUseLink[value]);
end;

Procedure TFhirClaim.SetPriority(value : TFhirCoding);
begin
  FPriority.free;
  FPriority := value;
end;

Procedure TFhirClaim.SetFundsReserve(value : TFhirCoding);
begin
  FFundsReserve.free;
  FFundsReserve := value;
end;

Procedure TFhirClaim.SetEnterer(value : TFhirReference{TFhirPractitioner});
begin
  FEnterer.free;
  FEnterer := value;
end;

Procedure TFhirClaim.SetFacility(value : TFhirReference{TFhirLocation});
begin
  FFacility.free;
  FFacility := value;
end;

Procedure TFhirClaim.SetPrescription(value : TFhirReference{Resource});
begin
  FPrescription.free;
  FPrescription := value;
end;

Procedure TFhirClaim.SetOriginalPrescription(value : TFhirReference{TFhirMedicationOrder});
begin
  FOriginalPrescription.free;
  FOriginalPrescription := value;
end;

Procedure TFhirClaim.SetPayee(value : TFhirClaimPayee);
begin
  FPayee.free;
  FPayee := value;
end;

Procedure TFhirClaim.SetReferral(value : TFhirReference{TFhirReferralRequest});
begin
  FReferral.free;
  FReferral := value;
end;

Function TFhirClaim.GetDiagnosisList : TFhirClaimDiagnosisList;
begin
  if FDiagnosisList = nil then
    FDiagnosisList := TFhirClaimDiagnosisList.Create;
  result := FDiagnosisList;
end;

Function TFhirClaim.GetHasDiagnosisList : boolean;
begin
  result := (FDiagnosisList <> nil) and (FDiagnosisList.count > 0);
end;

Function TFhirClaim.GetConditionList : TFhirCodingList;
begin
  if FConditionList = nil then
    FConditionList := TFhirCodingList.Create;
  result := FConditionList;
end;

Function TFhirClaim.GetHasConditionList : boolean;
begin
  result := (FConditionList <> nil) and (FConditionList.count > 0);
end;

Procedure TFhirClaim.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Function TFhirClaim.GetCoverageList : TFhirClaimCoverageList;
begin
  if FCoverageList = nil then
    FCoverageList := TFhirClaimCoverageList.Create;
  result := FCoverageList;
end;

Function TFhirClaim.GetHasCoverageList : boolean;
begin
  result := (FCoverageList <> nil) and (FCoverageList.count > 0);
end;

Function TFhirClaim.GetExceptionList : TFhirCodingList;
begin
  if FExceptionList = nil then
    FExceptionList := TFhirCodingList.Create;
  result := FExceptionList;
end;

Function TFhirClaim.GetHasExceptionList : boolean;
begin
  result := (FExceptionList <> nil) and (FExceptionList.count > 0);
end;

Procedure TFhirClaim.SetSchool(value : TFhirString);
begin
  FSchool.free;
  FSchool := value;
end;

Function TFhirClaim.GetSchoolST : String;
begin
  if FSchool = nil then
    result := ''
  else
    result := FSchool.value;
end;

Procedure TFhirClaim.SetSchoolST(value : String);
begin
  if value <> '' then
  begin
    if FSchool = nil then
      FSchool := TFhirString.create;
    FSchool.value := value
  end
  else if FSchool <> nil then
    FSchool.value := '';
end;

Procedure TFhirClaim.SetAccident(value : TFhirDate);
begin
  FAccident.free;
  FAccident := value;
end;

Function TFhirClaim.GetAccidentST : TDateAndTime;
begin
  if FAccident = nil then
    result := nil
  else
    result := FAccident.value;
end;

Procedure TFhirClaim.SetAccidentST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FAccident = nil then
      FAccident := TFhirDate.create;
    FAccident.value := value
  end
  else if FAccident <> nil then
    FAccident.value := nil;
end;

Procedure TFhirClaim.SetAccidentType(value : TFhirCoding);
begin
  FAccidentType.free;
  FAccidentType := value;
end;

Function TFhirClaim.GetInterventionExceptionList : TFhirCodingList;
begin
  if FInterventionExceptionList = nil then
    FInterventionExceptionList := TFhirCodingList.Create;
  result := FInterventionExceptionList;
end;

Function TFhirClaim.GetHasInterventionExceptionList : boolean;
begin
  result := (FInterventionExceptionList <> nil) and (FInterventionExceptionList.count > 0);
end;

Function TFhirClaim.GetItemList : TFhirClaimItemList;
begin
  if FItemList = nil then
    FItemList := TFhirClaimItemList.Create;
  result := FItemList;
end;

Function TFhirClaim.GetHasItemList : boolean;
begin
  result := (FItemList <> nil) and (FItemList.count > 0);
end;

Function TFhirClaim.GetAdditionalMaterialsList : TFhirCodingList;
begin
  if FAdditionalMaterialsList = nil then
    FAdditionalMaterialsList := TFhirCodingList.Create;
  result := FAdditionalMaterialsList;
end;

Function TFhirClaim.GetHasAdditionalMaterialsList : boolean;
begin
  result := (FAdditionalMaterialsList <> nil) and (FAdditionalMaterialsList.count > 0);
end;

Function TFhirClaim.GetMissingTeethList : TFhirClaimMissingTeethList;
begin
  if FMissingTeethList = nil then
    FMissingTeethList := TFhirClaimMissingTeethList.Create;
  result := FMissingTeethList;
end;

Function TFhirClaim.GetHasMissingTeethList : boolean;
begin
  result := (FMissingTeethList <> nil) and (FMissingTeethList.count > 0);
end;


{ TFhirClaimListEnumerator }

Constructor TFhirClaimListEnumerator.Create(list : TFhirClaimList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimListEnumerator.GetCurrent : TFhirClaim;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimList }
procedure TFhirClaimList.AddItem(value: TFhirClaim);
begin
  assert(value.ClassName = 'TFhirClaim', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaim');
  add(value);
end;


function TFhirClaimList.Append: TFhirClaim;
begin
  result := TFhirClaim.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimList.ClearItems;
begin
  Clear;
end;

function TFhirClaimList.GetEnumerator : TFhirClaimListEnumerator;
begin
  result := TFhirClaimListEnumerator.Create(self.link);
end;

function TFhirClaimList.Clone: TFhirClaimList;
begin
  result := TFhirClaimList(inherited Clone);
end;

function TFhirClaimList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimList.GetItemN(index: Integer): TFhirClaim;
begin
  result := TFhirClaim(ObjectByIndex[index]);
end;

function TFhirClaimList.IndexOf(value: TFhirClaim): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimList.Insert(index: Integer): TFhirClaim;
begin
  result := TFhirClaim.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimList.InsertItem(index: Integer; value: TFhirClaim);
begin
  assert(value is TFhirClaim);
  Inherited Insert(index, value);
end;

function TFhirClaimList.Item(index: Integer): TFhirClaim;
begin
  result := TFhirClaim(ObjectByIndex[index]);
end;

function TFhirClaimList.Link: TFhirClaimList;
begin
  result := TFhirClaimList(inherited Link);
end;

procedure TFhirClaimList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimList.SetItemByIndex(index: Integer; value: TFhirClaim);
begin
  assert(value is TFhirClaim);
  FhirClaims[index] := value;
end;

procedure TFhirClaimList.SetItemN(index: Integer; value: TFhirClaim);
begin
  assert(value is TFhirClaim);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItem }

constructor TFhirClaimResponseItem.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItem.Destroy;
begin
  FSequenceLinkId.free;
  FNoteNumberList.Free;
  FAdjudicationList.Free;
  FDetailList.Free;
  inherited;
end;

procedure TFhirClaimResponseItem.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceLinkIdElement := TFhirClaimResponseItem(oSource).sequenceLinkIdElement.Clone;
  if (TFhirClaimResponseItem(oSource).FNoteNumberList = nil) then
  begin
    FNoteNumberList.free;
    FNoteNumberList := nil;
  end
  else
  begin
    FNoteNumberList := TFhirPositiveIntList.Create;
    FNoteNumberList.Assign(TFhirClaimResponseItem(oSource).FNoteNumberList);
  end;
  if (TFhirClaimResponseItem(oSource).FAdjudicationList = nil) then
  begin
    FAdjudicationList.free;
    FAdjudicationList := nil;
  end
  else
  begin
    FAdjudicationList := TFhirClaimResponseItemAdjudicationList.Create;
    FAdjudicationList.Assign(TFhirClaimResponseItem(oSource).FAdjudicationList);
  end;
  if (TFhirClaimResponseItem(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirClaimResponseItemDetailList.Create;
    FDetailList.Assign(TFhirClaimResponseItem(oSource).FDetailList);
  end;
end;

procedure TFhirClaimResponseItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
     list.add(FSequenceLinkId.Link);
  if (child_name = 'noteNumber') Then
    list.addAll(FNoteNumberList);
  if (child_name = 'adjudication') Then
    list.addAll(FAdjudicationList);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirClaimResponseItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'positiveInt', FSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'noteNumber', 'positiveInt', FNoteNumberList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'adjudication', '', FAdjudicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetailList.Link)){3};
end;

procedure TFhirClaimResponseItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'noteNumber') then NoteNumberList.add(propValue as TFhirPositiveInt){2}
  else if (propName = 'adjudication') then AdjudicationList.add(propValue as TFhirClaimResponseItemAdjudication){2}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirClaimResponseItemDetail){2}
  else inherited;
end;

function TFhirClaimResponseItem.FhirType : string;
begin
  result := 'item';
end;

function TFhirClaimResponseItem.Link : TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem(inherited Link);
end;

function TFhirClaimResponseItem.Clone : TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem(inherited Clone);
end;

function TFhirClaimResponseItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItem)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItem(other);
    result := compareDeep(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareDeep(noteNumberList, o.noteNumberList, true) and compareDeep(adjudicationList, o.adjudicationList, true) and 
      compareDeep(detailList, o.detailList, true);
  end;
end;

function TFhirClaimResponseItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItem)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItem(other);
    result := compareValues(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareValues(noteNumberList, o.noteNumberList, true);
  end;
end;

{ TFhirClaimResponseItem }

Procedure TFhirClaimResponseItem.SetSequenceLinkId(value : TFhirPositiveInt);
begin
  FSequenceLinkId.free;
  FSequenceLinkId := value;
end;

Function TFhirClaimResponseItem.GetSequenceLinkIdST : String;
begin
  if FSequenceLinkId = nil then
    result := ''
  else
    result := FSequenceLinkId.value;
end;

Procedure TFhirClaimResponseItem.SetSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSequenceLinkId = nil then
      FSequenceLinkId := TFhirPositiveInt.create;
    FSequenceLinkId.value := value
  end
  else if FSequenceLinkId <> nil then
    FSequenceLinkId.value := '';
end;

Function TFhirClaimResponseItem.GetNoteNumberList : TFhirPositiveIntList;
begin
  if FNoteNumberList = nil then
    FNoteNumberList := TFhirPositiveIntList.Create;
  result := FNoteNumberList;
end;

Function TFhirClaimResponseItem.GetHasNoteNumberList : boolean;
begin
  result := (FNoteNumberList <> nil) and (FNoteNumberList.count > 0);
end;

Function TFhirClaimResponseItem.GetAdjudicationList : TFhirClaimResponseItemAdjudicationList;
begin
  if FAdjudicationList = nil then
    FAdjudicationList := TFhirClaimResponseItemAdjudicationList.Create;
  result := FAdjudicationList;
end;

Function TFhirClaimResponseItem.GetHasAdjudicationList : boolean;
begin
  result := (FAdjudicationList <> nil) and (FAdjudicationList.count > 0);
end;

Function TFhirClaimResponseItem.GetDetailList : TFhirClaimResponseItemDetailList;
begin
  if FDetailList = nil then
    FDetailList := TFhirClaimResponseItemDetailList.Create;
  result := FDetailList;
end;

Function TFhirClaimResponseItem.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirClaimResponseItemListEnumerator }

Constructor TFhirClaimResponseItemListEnumerator.Create(list : TFhirClaimResponseItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemListEnumerator.GetCurrent : TFhirClaimResponseItem;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemList }
procedure TFhirClaimResponseItemList.AddItem(value: TFhirClaimResponseItem);
begin
  assert(value.ClassName = 'TFhirClaimResponseItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItem');
  add(value);
end;


function TFhirClaimResponseItemList.Append: TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemList.GetEnumerator : TFhirClaimResponseItemListEnumerator;
begin
  result := TFhirClaimResponseItemListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemList.Clone: TFhirClaimResponseItemList;
begin
  result := TFhirClaimResponseItemList(inherited Clone);
end;

function TFhirClaimResponseItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemList.GetItemN(index: Integer): TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemList.IndexOf(value: TFhirClaimResponseItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemList.Insert(index: Integer): TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemList.InsertItem(index: Integer; value: TFhirClaimResponseItem);
begin
  assert(value is TFhirClaimResponseItem);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemList.Item(index: Integer): TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemList.Link: TFhirClaimResponseItemList;
begin
  result := TFhirClaimResponseItemList(inherited Link);
end;

procedure TFhirClaimResponseItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItem);
begin
  assert(value is TFhirClaimResponseItem);
  FhirClaimResponseItems[index] := value;
end;

procedure TFhirClaimResponseItemList.SetItemN(index: Integer; value: TFhirClaimResponseItem);
begin
  assert(value is TFhirClaimResponseItem);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItemAdjudication }

constructor TFhirClaimResponseItemAdjudication.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItemAdjudication.Destroy;
begin
  FCode.free;
  FAmount.free;
  FValue.free;
  inherited;
end;

procedure TFhirClaimResponseItemAdjudication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClaimResponseItemAdjudication(oSource).code.Clone;
  amount := TFhirClaimResponseItemAdjudication(oSource).amount.Clone;
  valueElement := TFhirClaimResponseItemAdjudication(oSource).valueElement.Clone;
end;

procedure TFhirClaimResponseItemAdjudication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirClaimResponseItemAdjudication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'decimal', FValue.Link));{2}
end;

procedure TFhirClaimResponseItemAdjudication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirClaimResponseItemAdjudication.FhirType : string;
begin
  result := 'adjudication';
end;

function TFhirClaimResponseItemAdjudication.Link : TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication(inherited Link);
end;

function TFhirClaimResponseItemAdjudication.Clone : TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication(inherited Clone);
end;

function TFhirClaimResponseItemAdjudication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemAdjudication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemAdjudication(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(amountElement, o.amountElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirClaimResponseItemAdjudication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemAdjudication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemAdjudication(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirClaimResponseItemAdjudication }

Procedure TFhirClaimResponseItemAdjudication.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirClaimResponseItemAdjudication.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;

Procedure TFhirClaimResponseItemAdjudication.SetValue(value : TFhirDecimal);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirClaimResponseItemAdjudication.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirClaimResponseItemAdjudication.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirDecimal.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirClaimResponseItemAdjudicationListEnumerator }

Constructor TFhirClaimResponseItemAdjudicationListEnumerator.Create(list : TFhirClaimResponseItemAdjudicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemAdjudicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemAdjudicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemAdjudicationListEnumerator.GetCurrent : TFhirClaimResponseItemAdjudication;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemAdjudicationList }
procedure TFhirClaimResponseItemAdjudicationList.AddItem(value: TFhirClaimResponseItemAdjudication);
begin
  assert(value.ClassName = 'TFhirClaimResponseItemAdjudication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItemAdjudication');
  add(value);
end;


function TFhirClaimResponseItemAdjudicationList.Append: TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemAdjudicationList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemAdjudicationList.GetEnumerator : TFhirClaimResponseItemAdjudicationListEnumerator;
begin
  result := TFhirClaimResponseItemAdjudicationListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemAdjudicationList.Clone: TFhirClaimResponseItemAdjudicationList;
begin
  result := TFhirClaimResponseItemAdjudicationList(inherited Clone);
end;

function TFhirClaimResponseItemAdjudicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemAdjudicationList.GetItemN(index: Integer): TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemAdjudicationList.IndexOf(value: TFhirClaimResponseItemAdjudication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemAdjudicationList.Insert(index: Integer): TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemAdjudicationList.InsertItem(index: Integer; value: TFhirClaimResponseItemAdjudication);
begin
  assert(value is TFhirClaimResponseItemAdjudication);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemAdjudicationList.Item(index: Integer): TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemAdjudicationList.Link: TFhirClaimResponseItemAdjudicationList;
begin
  result := TFhirClaimResponseItemAdjudicationList(inherited Link);
end;

procedure TFhirClaimResponseItemAdjudicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemAdjudicationList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItemAdjudication);
begin
  assert(value is TFhirClaimResponseItemAdjudication);
  FhirClaimResponseItemAdjudications[index] := value;
end;

procedure TFhirClaimResponseItemAdjudicationList.SetItemN(index: Integer; value: TFhirClaimResponseItemAdjudication);
begin
  assert(value is TFhirClaimResponseItemAdjudication);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItemDetail }

constructor TFhirClaimResponseItemDetail.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItemDetail.Destroy;
begin
  FSequenceLinkId.free;
  FAdjudicationList.Free;
  FSubDetailList.Free;
  inherited;
end;

procedure TFhirClaimResponseItemDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceLinkIdElement := TFhirClaimResponseItemDetail(oSource).sequenceLinkIdElement.Clone;
  if (TFhirClaimResponseItemDetail(oSource).FAdjudicationList = nil) then
  begin
    FAdjudicationList.free;
    FAdjudicationList := nil;
  end
  else
  begin
    FAdjudicationList := TFhirClaimResponseItemDetailAdjudicationList.Create;
    FAdjudicationList.Assign(TFhirClaimResponseItemDetail(oSource).FAdjudicationList);
  end;
  if (TFhirClaimResponseItemDetail(oSource).FSubDetailList = nil) then
  begin
    FSubDetailList.free;
    FSubDetailList := nil;
  end
  else
  begin
    FSubDetailList := TFhirClaimResponseItemDetailSubDetailList.Create;
    FSubDetailList.Assign(TFhirClaimResponseItemDetail(oSource).FSubDetailList);
  end;
end;

procedure TFhirClaimResponseItemDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
     list.add(FSequenceLinkId.Link);
  if (child_name = 'adjudication') Then
    list.addAll(FAdjudicationList);
  if (child_name = 'subDetail') Then
    list.addAll(FSubDetailList);
end;

procedure TFhirClaimResponseItemDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'positiveInt', FSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'adjudication', '', FAdjudicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subDetail', '', FSubDetailList.Link)){3};
end;

procedure TFhirClaimResponseItemDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'adjudication') then AdjudicationList.add(propValue as TFhirClaimResponseItemDetailAdjudication){2}
  else if (propName = 'subDetail') then SubDetailList.add(propValue as TFhirClaimResponseItemDetailSubDetail){2}
  else inherited;
end;

function TFhirClaimResponseItemDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirClaimResponseItemDetail.Link : TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail(inherited Link);
end;

function TFhirClaimResponseItemDetail.Clone : TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail(inherited Clone);
end;

function TFhirClaimResponseItemDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetail(other);
    result := compareDeep(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareDeep(adjudicationList, o.adjudicationList, true) and compareDeep(subDetailList, o.subDetailList, true);
  end;
end;

function TFhirClaimResponseItemDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetail(other);
    result := compareValues(sequenceLinkIdElement, o.sequenceLinkIdElement, true);
  end;
end;

{ TFhirClaimResponseItemDetail }

Procedure TFhirClaimResponseItemDetail.SetSequenceLinkId(value : TFhirPositiveInt);
begin
  FSequenceLinkId.free;
  FSequenceLinkId := value;
end;

Function TFhirClaimResponseItemDetail.GetSequenceLinkIdST : String;
begin
  if FSequenceLinkId = nil then
    result := ''
  else
    result := FSequenceLinkId.value;
end;

Procedure TFhirClaimResponseItemDetail.SetSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSequenceLinkId = nil then
      FSequenceLinkId := TFhirPositiveInt.create;
    FSequenceLinkId.value := value
  end
  else if FSequenceLinkId <> nil then
    FSequenceLinkId.value := '';
end;

Function TFhirClaimResponseItemDetail.GetAdjudicationList : TFhirClaimResponseItemDetailAdjudicationList;
begin
  if FAdjudicationList = nil then
    FAdjudicationList := TFhirClaimResponseItemDetailAdjudicationList.Create;
  result := FAdjudicationList;
end;

Function TFhirClaimResponseItemDetail.GetHasAdjudicationList : boolean;
begin
  result := (FAdjudicationList <> nil) and (FAdjudicationList.count > 0);
end;

Function TFhirClaimResponseItemDetail.GetSubDetailList : TFhirClaimResponseItemDetailSubDetailList;
begin
  if FSubDetailList = nil then
    FSubDetailList := TFhirClaimResponseItemDetailSubDetailList.Create;
  result := FSubDetailList;
end;

Function TFhirClaimResponseItemDetail.GetHasSubDetailList : boolean;
begin
  result := (FSubDetailList <> nil) and (FSubDetailList.count > 0);
end;


{ TFhirClaimResponseItemDetailListEnumerator }

Constructor TFhirClaimResponseItemDetailListEnumerator.Create(list : TFhirClaimResponseItemDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemDetailListEnumerator.GetCurrent : TFhirClaimResponseItemDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemDetailList }
procedure TFhirClaimResponseItemDetailList.AddItem(value: TFhirClaimResponseItemDetail);
begin
  assert(value.ClassName = 'TFhirClaimResponseItemDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItemDetail');
  add(value);
end;


function TFhirClaimResponseItemDetailList.Append: TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemDetailList.GetEnumerator : TFhirClaimResponseItemDetailListEnumerator;
begin
  result := TFhirClaimResponseItemDetailListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemDetailList.Clone: TFhirClaimResponseItemDetailList;
begin
  result := TFhirClaimResponseItemDetailList(inherited Clone);
end;

function TFhirClaimResponseItemDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemDetailList.GetItemN(index: Integer): TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailList.IndexOf(value: TFhirClaimResponseItemDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemDetailList.Insert(index: Integer): TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailList.InsertItem(index: Integer; value: TFhirClaimResponseItemDetail);
begin
  assert(value is TFhirClaimResponseItemDetail);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemDetailList.Item(index: Integer): TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailList.Link: TFhirClaimResponseItemDetailList;
begin
  result := TFhirClaimResponseItemDetailList(inherited Link);
end;

procedure TFhirClaimResponseItemDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemDetailList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItemDetail);
begin
  assert(value is TFhirClaimResponseItemDetail);
  FhirClaimResponseItemDetails[index] := value;
end;

procedure TFhirClaimResponseItemDetailList.SetItemN(index: Integer; value: TFhirClaimResponseItemDetail);
begin
  assert(value is TFhirClaimResponseItemDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItemDetailAdjudication }

constructor TFhirClaimResponseItemDetailAdjudication.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItemDetailAdjudication.Destroy;
begin
  FCode.free;
  FAmount.free;
  FValue.free;
  inherited;
end;

procedure TFhirClaimResponseItemDetailAdjudication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClaimResponseItemDetailAdjudication(oSource).code.Clone;
  amount := TFhirClaimResponseItemDetailAdjudication(oSource).amount.Clone;
  valueElement := TFhirClaimResponseItemDetailAdjudication(oSource).valueElement.Clone;
end;

procedure TFhirClaimResponseItemDetailAdjudication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirClaimResponseItemDetailAdjudication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'decimal', FValue.Link));{2}
end;

procedure TFhirClaimResponseItemDetailAdjudication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirClaimResponseItemDetailAdjudication.FhirType : string;
begin
  result := 'adjudication';
end;

function TFhirClaimResponseItemDetailAdjudication.Link : TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication(inherited Link);
end;

function TFhirClaimResponseItemDetailAdjudication.Clone : TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication(inherited Clone);
end;

function TFhirClaimResponseItemDetailAdjudication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailAdjudication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailAdjudication(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(amountElement, o.amountElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirClaimResponseItemDetailAdjudication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailAdjudication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailAdjudication(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirClaimResponseItemDetailAdjudication }

Procedure TFhirClaimResponseItemDetailAdjudication.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirClaimResponseItemDetailAdjudication.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;

Procedure TFhirClaimResponseItemDetailAdjudication.SetValue(value : TFhirDecimal);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirClaimResponseItemDetailAdjudication.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirClaimResponseItemDetailAdjudication.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirDecimal.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirClaimResponseItemDetailAdjudicationListEnumerator }

Constructor TFhirClaimResponseItemDetailAdjudicationListEnumerator.Create(list : TFhirClaimResponseItemDetailAdjudicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemDetailAdjudicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemDetailAdjudicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemDetailAdjudicationListEnumerator.GetCurrent : TFhirClaimResponseItemDetailAdjudication;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemDetailAdjudicationList }
procedure TFhirClaimResponseItemDetailAdjudicationList.AddItem(value: TFhirClaimResponseItemDetailAdjudication);
begin
  assert(value.ClassName = 'TFhirClaimResponseItemDetailAdjudication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItemDetailAdjudication');
  add(value);
end;


function TFhirClaimResponseItemDetailAdjudicationList.Append: TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailAdjudicationList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemDetailAdjudicationList.GetEnumerator : TFhirClaimResponseItemDetailAdjudicationListEnumerator;
begin
  result := TFhirClaimResponseItemDetailAdjudicationListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemDetailAdjudicationList.Clone: TFhirClaimResponseItemDetailAdjudicationList;
begin
  result := TFhirClaimResponseItemDetailAdjudicationList(inherited Clone);
end;

function TFhirClaimResponseItemDetailAdjudicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemDetailAdjudicationList.GetItemN(index: Integer): TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailAdjudicationList.IndexOf(value: TFhirClaimResponseItemDetailAdjudication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemDetailAdjudicationList.Insert(index: Integer): TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailAdjudicationList.InsertItem(index: Integer; value: TFhirClaimResponseItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailAdjudication);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemDetailAdjudicationList.Item(index: Integer): TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailAdjudicationList.Link: TFhirClaimResponseItemDetailAdjudicationList;
begin
  result := TFhirClaimResponseItemDetailAdjudicationList(inherited Link);
end;

procedure TFhirClaimResponseItemDetailAdjudicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemDetailAdjudicationList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailAdjudication);
  FhirClaimResponseItemDetailAdjudications[index] := value;
end;

procedure TFhirClaimResponseItemDetailAdjudicationList.SetItemN(index: Integer; value: TFhirClaimResponseItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailAdjudication);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItemDetailSubDetail }

constructor TFhirClaimResponseItemDetailSubDetail.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItemDetailSubDetail.Destroy;
begin
  FSequenceLinkId.free;
  FAdjudicationList.Free;
  inherited;
end;

procedure TFhirClaimResponseItemDetailSubDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceLinkIdElement := TFhirClaimResponseItemDetailSubDetail(oSource).sequenceLinkIdElement.Clone;
  if (TFhirClaimResponseItemDetailSubDetail(oSource).FAdjudicationList = nil) then
  begin
    FAdjudicationList.free;
    FAdjudicationList := nil;
  end
  else
  begin
    FAdjudicationList := TFhirClaimResponseItemDetailSubDetailAdjudicationList.Create;
    FAdjudicationList.Assign(TFhirClaimResponseItemDetailSubDetail(oSource).FAdjudicationList);
  end;
end;

procedure TFhirClaimResponseItemDetailSubDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
     list.add(FSequenceLinkId.Link);
  if (child_name = 'adjudication') Then
    list.addAll(FAdjudicationList);
end;

procedure TFhirClaimResponseItemDetailSubDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'positiveInt', FSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'adjudication', '', FAdjudicationList.Link)){3};
end;

procedure TFhirClaimResponseItemDetailSubDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'adjudication') then AdjudicationList.add(propValue as TFhirClaimResponseItemDetailSubDetailAdjudication){2}
  else inherited;
end;

function TFhirClaimResponseItemDetailSubDetail.FhirType : string;
begin
  result := 'subDetail';
end;

function TFhirClaimResponseItemDetailSubDetail.Link : TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail(inherited Link);
end;

function TFhirClaimResponseItemDetailSubDetail.Clone : TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail(inherited Clone);
end;

function TFhirClaimResponseItemDetailSubDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailSubDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailSubDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailSubDetail(other);
    result := compareDeep(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareDeep(adjudicationList, o.adjudicationList, true);
  end;
end;

function TFhirClaimResponseItemDetailSubDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailSubDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailSubDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailSubDetail(other);
    result := compareValues(sequenceLinkIdElement, o.sequenceLinkIdElement, true);
  end;
end;

{ TFhirClaimResponseItemDetailSubDetail }

Procedure TFhirClaimResponseItemDetailSubDetail.SetSequenceLinkId(value : TFhirPositiveInt);
begin
  FSequenceLinkId.free;
  FSequenceLinkId := value;
end;

Function TFhirClaimResponseItemDetailSubDetail.GetSequenceLinkIdST : String;
begin
  if FSequenceLinkId = nil then
    result := ''
  else
    result := FSequenceLinkId.value;
end;

Procedure TFhirClaimResponseItemDetailSubDetail.SetSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSequenceLinkId = nil then
      FSequenceLinkId := TFhirPositiveInt.create;
    FSequenceLinkId.value := value
  end
  else if FSequenceLinkId <> nil then
    FSequenceLinkId.value := '';
end;

Function TFhirClaimResponseItemDetailSubDetail.GetAdjudicationList : TFhirClaimResponseItemDetailSubDetailAdjudicationList;
begin
  if FAdjudicationList = nil then
    FAdjudicationList := TFhirClaimResponseItemDetailSubDetailAdjudicationList.Create;
  result := FAdjudicationList;
end;

Function TFhirClaimResponseItemDetailSubDetail.GetHasAdjudicationList : boolean;
begin
  result := (FAdjudicationList <> nil) and (FAdjudicationList.count > 0);
end;


{ TFhirClaimResponseItemDetailSubDetailListEnumerator }

Constructor TFhirClaimResponseItemDetailSubDetailListEnumerator.Create(list : TFhirClaimResponseItemDetailSubDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemDetailSubDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemDetailSubDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemDetailSubDetailListEnumerator.GetCurrent : TFhirClaimResponseItemDetailSubDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemDetailSubDetailList }
procedure TFhirClaimResponseItemDetailSubDetailList.AddItem(value: TFhirClaimResponseItemDetailSubDetail);
begin
  assert(value.ClassName = 'TFhirClaimResponseItemDetailSubDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItemDetailSubDetail');
  add(value);
end;


function TFhirClaimResponseItemDetailSubDetailList.Append: TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailSubDetailList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemDetailSubDetailList.GetEnumerator : TFhirClaimResponseItemDetailSubDetailListEnumerator;
begin
  result := TFhirClaimResponseItemDetailSubDetailListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemDetailSubDetailList.Clone: TFhirClaimResponseItemDetailSubDetailList;
begin
  result := TFhirClaimResponseItemDetailSubDetailList(inherited Clone);
end;

function TFhirClaimResponseItemDetailSubDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemDetailSubDetailList.GetItemN(index: Integer): TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailSubDetailList.IndexOf(value: TFhirClaimResponseItemDetailSubDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemDetailSubDetailList.Insert(index: Integer): TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailSubDetailList.InsertItem(index: Integer; value: TFhirClaimResponseItemDetailSubDetail);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetail);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemDetailSubDetailList.Item(index: Integer): TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailSubDetailList.Link: TFhirClaimResponseItemDetailSubDetailList;
begin
  result := TFhirClaimResponseItemDetailSubDetailList(inherited Link);
end;

procedure TFhirClaimResponseItemDetailSubDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemDetailSubDetailList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItemDetailSubDetail);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetail);
  FhirClaimResponseItemDetailSubDetails[index] := value;
end;

procedure TFhirClaimResponseItemDetailSubDetailList.SetItemN(index: Integer; value: TFhirClaimResponseItemDetailSubDetail);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseItemDetailSubDetailAdjudication }

constructor TFhirClaimResponseItemDetailSubDetailAdjudication.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseItemDetailSubDetailAdjudication.Destroy;
begin
  FCode.free;
  FAmount.free;
  FValue.free;
  inherited;
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClaimResponseItemDetailSubDetailAdjudication(oSource).code.Clone;
  amount := TFhirClaimResponseItemDetailSubDetailAdjudication(oSource).amount.Clone;
  valueElement := TFhirClaimResponseItemDetailSubDetailAdjudication(oSource).valueElement.Clone;
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'decimal', FValue.Link));{2}
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudication.FhirType : string;
begin
  result := 'adjudication';
end;

function TFhirClaimResponseItemDetailSubDetailAdjudication.Link : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication(inherited Link);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudication.Clone : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication(inherited Clone);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailSubDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailSubDetailAdjudication(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(amountElement, o.amountElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseItemDetailSubDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseItemDetailSubDetailAdjudication(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirClaimResponseItemDetailSubDetailAdjudication }

Procedure TFhirClaimResponseItemDetailSubDetailAdjudication.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirClaimResponseItemDetailSubDetailAdjudication.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;

Procedure TFhirClaimResponseItemDetailSubDetailAdjudication.SetValue(value : TFhirDecimal);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirClaimResponseItemDetailSubDetailAdjudication.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirClaimResponseItemDetailSubDetailAdjudication.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirDecimal.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator }

Constructor TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator.Create(list : TFhirClaimResponseItemDetailSubDetailAdjudicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator.GetCurrent : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseItemDetailSubDetailAdjudicationList }
procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.AddItem(value: TFhirClaimResponseItemDetailSubDetailAdjudication);
begin
  assert(value.ClassName = 'TFhirClaimResponseItemDetailSubDetailAdjudication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseItemDetailSubDetailAdjudication');
  add(value);
end;


function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Append: TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.GetEnumerator : TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudicationListEnumerator.Create(self.link);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Clone: TFhirClaimResponseItemDetailSubDetailAdjudicationList;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudicationList(inherited Clone);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.GetItemN(index: Integer): TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.IndexOf(value: TFhirClaimResponseItemDetailSubDetailAdjudication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Insert(index: Integer): TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.InsertItem(index: Integer; value: TFhirClaimResponseItemDetailSubDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetailAdjudication);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Item(index: Integer): TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseItemDetailSubDetailAdjudicationList.Link: TFhirClaimResponseItemDetailSubDetailAdjudicationList;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudicationList(inherited Link);
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.SetItemByIndex(index: Integer; value: TFhirClaimResponseItemDetailSubDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetailAdjudication);
  FhirClaimResponseItemDetailSubDetailAdjudications[index] := value;
end;

procedure TFhirClaimResponseItemDetailSubDetailAdjudicationList.SetItemN(index: Integer; value: TFhirClaimResponseItemDetailSubDetailAdjudication);
begin
  assert(value is TFhirClaimResponseItemDetailSubDetailAdjudication);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseAddItem }

constructor TFhirClaimResponseAddItem.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseAddItem.Destroy;
begin
  FSequenceLinkIdList.Free;
  FService.free;
  FFee.free;
  FNoteNumberLinkIdList.Free;
  FAdjudicationList.Free;
  FDetailList.Free;
  inherited;
end;

procedure TFhirClaimResponseAddItem.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirClaimResponseAddItem(oSource).FSequenceLinkIdList = nil) then
  begin
    FSequenceLinkIdList.free;
    FSequenceLinkIdList := nil;
  end
  else
  begin
    FSequenceLinkIdList := TFhirPositiveIntList.Create;
    FSequenceLinkIdList.Assign(TFhirClaimResponseAddItem(oSource).FSequenceLinkIdList);
  end;
  service := TFhirClaimResponseAddItem(oSource).service.Clone;
  fee := TFhirClaimResponseAddItem(oSource).fee.Clone;
  if (TFhirClaimResponseAddItem(oSource).FNoteNumberLinkIdList = nil) then
  begin
    FNoteNumberLinkIdList.free;
    FNoteNumberLinkIdList := nil;
  end
  else
  begin
    FNoteNumberLinkIdList := TFhirPositiveIntList.Create;
    FNoteNumberLinkIdList.Assign(TFhirClaimResponseAddItem(oSource).FNoteNumberLinkIdList);
  end;
  if (TFhirClaimResponseAddItem(oSource).FAdjudicationList = nil) then
  begin
    FAdjudicationList.free;
    FAdjudicationList := nil;
  end
  else
  begin
    FAdjudicationList := TFhirClaimResponseAddItemAdjudicationList.Create;
    FAdjudicationList.Assign(TFhirClaimResponseAddItem(oSource).FAdjudicationList);
  end;
  if (TFhirClaimResponseAddItem(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirClaimResponseAddItemDetailList.Create;
    FDetailList.Assign(TFhirClaimResponseAddItem(oSource).FDetailList);
  end;
end;

procedure TFhirClaimResponseAddItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
    list.addAll(FSequenceLinkIdList);
  if (child_name = 'service') Then
     list.add(FService.Link);
  if (child_name = 'fee') Then
     list.add(FFee.Link);
  if (child_name = 'noteNumberLinkId') Then
    list.addAll(FNoteNumberLinkIdList);
  if (child_name = 'adjudication') Then
    list.addAll(FAdjudicationList);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirClaimResponseAddItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'positiveInt', FSequenceLinkIdList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'service', 'Coding', FService.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fee', 'Money', FFee.Link));{2}
  oList.add(TFHIRProperty.create(self, 'noteNumberLinkId', 'positiveInt', FNoteNumberLinkIdList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'adjudication', '', FAdjudicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetailList.Link)){3};
end;

procedure TFhirClaimResponseAddItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdList.add(propValue as TFhirPositiveInt){2}
  else if (propName = 'service') then Service := propValue as TFhirCoding{4b}
  else if (propName = 'fee') then Fee := propValue as TFhirQuantity{4b}
  else if (propName = 'noteNumberLinkId') then NoteNumberLinkIdList.add(propValue as TFhirPositiveInt){2}
  else if (propName = 'adjudication') then AdjudicationList.add(propValue as TFhirClaimResponseAddItemAdjudication){2}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirClaimResponseAddItemDetail){2}
  else inherited;
end;

function TFhirClaimResponseAddItem.FhirType : string;
begin
  result := 'addItem';
end;

function TFhirClaimResponseAddItem.Link : TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem(inherited Link);
end;

function TFhirClaimResponseAddItem.Clone : TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem(inherited Clone);
end;

function TFhirClaimResponseAddItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItem)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItem(other);
    result := compareDeep(sequenceLinkIdList, o.sequenceLinkIdList, true) and compareDeep(serviceElement, o.serviceElement, true) and 
      compareDeep(feeElement, o.feeElement, true) and compareDeep(noteNumberLinkIdList, o.noteNumberLinkIdList, true) and 
      compareDeep(adjudicationList, o.adjudicationList, true) and compareDeep(detailList, o.detailList, true);
  end;
end;

function TFhirClaimResponseAddItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItem)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItem(other);
    result := compareValues(sequenceLinkIdList, o.sequenceLinkIdList, true) and compareValues(noteNumberLinkIdList, o.noteNumberLinkIdList, true);
  end;
end;

{ TFhirClaimResponseAddItem }

Function TFhirClaimResponseAddItem.GetSequenceLinkIdList : TFhirPositiveIntList;
begin
  if FSequenceLinkIdList = nil then
    FSequenceLinkIdList := TFhirPositiveIntList.Create;
  result := FSequenceLinkIdList;
end;

Function TFhirClaimResponseAddItem.GetHasSequenceLinkIdList : boolean;
begin
  result := (FSequenceLinkIdList <> nil) and (FSequenceLinkIdList.count > 0);
end;

Procedure TFhirClaimResponseAddItem.SetService(value : TFhirCoding);
begin
  FService.free;
  FService := value;
end;

Procedure TFhirClaimResponseAddItem.SetFee(value : TFhirQuantity);
begin
  FFee.free;
  FFee := value;
end;

Function TFhirClaimResponseAddItem.GetNoteNumberLinkIdList : TFhirPositiveIntList;
begin
  if FNoteNumberLinkIdList = nil then
    FNoteNumberLinkIdList := TFhirPositiveIntList.Create;
  result := FNoteNumberLinkIdList;
end;

Function TFhirClaimResponseAddItem.GetHasNoteNumberLinkIdList : boolean;
begin
  result := (FNoteNumberLinkIdList <> nil) and (FNoteNumberLinkIdList.count > 0);
end;

Function TFhirClaimResponseAddItem.GetAdjudicationList : TFhirClaimResponseAddItemAdjudicationList;
begin
  if FAdjudicationList = nil then
    FAdjudicationList := TFhirClaimResponseAddItemAdjudicationList.Create;
  result := FAdjudicationList;
end;

Function TFhirClaimResponseAddItem.GetHasAdjudicationList : boolean;
begin
  result := (FAdjudicationList <> nil) and (FAdjudicationList.count > 0);
end;

Function TFhirClaimResponseAddItem.GetDetailList : TFhirClaimResponseAddItemDetailList;
begin
  if FDetailList = nil then
    FDetailList := TFhirClaimResponseAddItemDetailList.Create;
  result := FDetailList;
end;

Function TFhirClaimResponseAddItem.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirClaimResponseAddItemListEnumerator }

Constructor TFhirClaimResponseAddItemListEnumerator.Create(list : TFhirClaimResponseAddItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseAddItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseAddItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseAddItemListEnumerator.GetCurrent : TFhirClaimResponseAddItem;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseAddItemList }
procedure TFhirClaimResponseAddItemList.AddItem(value: TFhirClaimResponseAddItem);
begin
  assert(value.ClassName = 'TFhirClaimResponseAddItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseAddItem');
  add(value);
end;


function TFhirClaimResponseAddItemList.Append: TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseAddItemList.GetEnumerator : TFhirClaimResponseAddItemListEnumerator;
begin
  result := TFhirClaimResponseAddItemListEnumerator.Create(self.link);
end;

function TFhirClaimResponseAddItemList.Clone: TFhirClaimResponseAddItemList;
begin
  result := TFhirClaimResponseAddItemList(inherited Clone);
end;

function TFhirClaimResponseAddItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseAddItemList.GetItemN(index: Integer): TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemList.IndexOf(value: TFhirClaimResponseAddItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseAddItemList.Insert(index: Integer): TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemList.InsertItem(index: Integer; value: TFhirClaimResponseAddItem);
begin
  assert(value is TFhirClaimResponseAddItem);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseAddItemList.Item(index: Integer): TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemList.Link: TFhirClaimResponseAddItemList;
begin
  result := TFhirClaimResponseAddItemList(inherited Link);
end;

procedure TFhirClaimResponseAddItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseAddItemList.SetItemByIndex(index: Integer; value: TFhirClaimResponseAddItem);
begin
  assert(value is TFhirClaimResponseAddItem);
  FhirClaimResponseAddItems[index] := value;
end;

procedure TFhirClaimResponseAddItemList.SetItemN(index: Integer; value: TFhirClaimResponseAddItem);
begin
  assert(value is TFhirClaimResponseAddItem);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseAddItemAdjudication }

constructor TFhirClaimResponseAddItemAdjudication.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseAddItemAdjudication.Destroy;
begin
  FCode.free;
  FAmount.free;
  FValue.free;
  inherited;
end;

procedure TFhirClaimResponseAddItemAdjudication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClaimResponseAddItemAdjudication(oSource).code.Clone;
  amount := TFhirClaimResponseAddItemAdjudication(oSource).amount.Clone;
  valueElement := TFhirClaimResponseAddItemAdjudication(oSource).valueElement.Clone;
end;

procedure TFhirClaimResponseAddItemAdjudication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirClaimResponseAddItemAdjudication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'decimal', FValue.Link));{2}
end;

procedure TFhirClaimResponseAddItemAdjudication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirClaimResponseAddItemAdjudication.FhirType : string;
begin
  result := 'adjudication';
end;

function TFhirClaimResponseAddItemAdjudication.Link : TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication(inherited Link);
end;

function TFhirClaimResponseAddItemAdjudication.Clone : TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication(inherited Clone);
end;

function TFhirClaimResponseAddItemAdjudication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemAdjudication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemAdjudication(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(amountElement, o.amountElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirClaimResponseAddItemAdjudication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemAdjudication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemAdjudication(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirClaimResponseAddItemAdjudication }

Procedure TFhirClaimResponseAddItemAdjudication.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirClaimResponseAddItemAdjudication.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;

Procedure TFhirClaimResponseAddItemAdjudication.SetValue(value : TFhirDecimal);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirClaimResponseAddItemAdjudication.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirClaimResponseAddItemAdjudication.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirDecimal.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirClaimResponseAddItemAdjudicationListEnumerator }

Constructor TFhirClaimResponseAddItemAdjudicationListEnumerator.Create(list : TFhirClaimResponseAddItemAdjudicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseAddItemAdjudicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseAddItemAdjudicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseAddItemAdjudicationListEnumerator.GetCurrent : TFhirClaimResponseAddItemAdjudication;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseAddItemAdjudicationList }
procedure TFhirClaimResponseAddItemAdjudicationList.AddItem(value: TFhirClaimResponseAddItemAdjudication);
begin
  assert(value.ClassName = 'TFhirClaimResponseAddItemAdjudication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseAddItemAdjudication');
  add(value);
end;


function TFhirClaimResponseAddItemAdjudicationList.Append: TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemAdjudicationList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseAddItemAdjudicationList.GetEnumerator : TFhirClaimResponseAddItemAdjudicationListEnumerator;
begin
  result := TFhirClaimResponseAddItemAdjudicationListEnumerator.Create(self.link);
end;

function TFhirClaimResponseAddItemAdjudicationList.Clone: TFhirClaimResponseAddItemAdjudicationList;
begin
  result := TFhirClaimResponseAddItemAdjudicationList(inherited Clone);
end;

function TFhirClaimResponseAddItemAdjudicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseAddItemAdjudicationList.GetItemN(index: Integer): TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemAdjudicationList.IndexOf(value: TFhirClaimResponseAddItemAdjudication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseAddItemAdjudicationList.Insert(index: Integer): TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemAdjudicationList.InsertItem(index: Integer; value: TFhirClaimResponseAddItemAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemAdjudication);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseAddItemAdjudicationList.Item(index: Integer): TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemAdjudicationList.Link: TFhirClaimResponseAddItemAdjudicationList;
begin
  result := TFhirClaimResponseAddItemAdjudicationList(inherited Link);
end;

procedure TFhirClaimResponseAddItemAdjudicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseAddItemAdjudicationList.SetItemByIndex(index: Integer; value: TFhirClaimResponseAddItemAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemAdjudication);
  FhirClaimResponseAddItemAdjudications[index] := value;
end;

procedure TFhirClaimResponseAddItemAdjudicationList.SetItemN(index: Integer; value: TFhirClaimResponseAddItemAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemAdjudication);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseAddItemDetail }

constructor TFhirClaimResponseAddItemDetail.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseAddItemDetail.Destroy;
begin
  FService.free;
  FFee.free;
  FAdjudicationList.Free;
  inherited;
end;

procedure TFhirClaimResponseAddItemDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  service := TFhirClaimResponseAddItemDetail(oSource).service.Clone;
  fee := TFhirClaimResponseAddItemDetail(oSource).fee.Clone;
  if (TFhirClaimResponseAddItemDetail(oSource).FAdjudicationList = nil) then
  begin
    FAdjudicationList.free;
    FAdjudicationList := nil;
  end
  else
  begin
    FAdjudicationList := TFhirClaimResponseAddItemDetailAdjudicationList.Create;
    FAdjudicationList.Assign(TFhirClaimResponseAddItemDetail(oSource).FAdjudicationList);
  end;
end;

procedure TFhirClaimResponseAddItemDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'service') Then
     list.add(FService.Link);
  if (child_name = 'fee') Then
     list.add(FFee.Link);
  if (child_name = 'adjudication') Then
    list.addAll(FAdjudicationList);
end;

procedure TFhirClaimResponseAddItemDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'service', 'Coding', FService.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fee', 'Money', FFee.Link));{2}
  oList.add(TFHIRProperty.create(self, 'adjudication', '', FAdjudicationList.Link)){3};
end;

procedure TFhirClaimResponseAddItemDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'service') then Service := propValue as TFhirCoding{4b}
  else if (propName = 'fee') then Fee := propValue as TFhirQuantity{4b}
  else if (propName = 'adjudication') then AdjudicationList.add(propValue as TFhirClaimResponseAddItemDetailAdjudication){2}
  else inherited;
end;

function TFhirClaimResponseAddItemDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirClaimResponseAddItemDetail.Link : TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail(inherited Link);
end;

function TFhirClaimResponseAddItemDetail.Clone : TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail(inherited Clone);
end;

function TFhirClaimResponseAddItemDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemDetail(other);
    result := compareDeep(serviceElement, o.serviceElement, true) and compareDeep(feeElement, o.feeElement, true) and 
      compareDeep(adjudicationList, o.adjudicationList, true);
  end;
end;

function TFhirClaimResponseAddItemDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemDetail)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemDetail(other);
    result := true;
  end;
end;

{ TFhirClaimResponseAddItemDetail }

Procedure TFhirClaimResponseAddItemDetail.SetService(value : TFhirCoding);
begin
  FService.free;
  FService := value;
end;

Procedure TFhirClaimResponseAddItemDetail.SetFee(value : TFhirQuantity);
begin
  FFee.free;
  FFee := value;
end;

Function TFhirClaimResponseAddItemDetail.GetAdjudicationList : TFhirClaimResponseAddItemDetailAdjudicationList;
begin
  if FAdjudicationList = nil then
    FAdjudicationList := TFhirClaimResponseAddItemDetailAdjudicationList.Create;
  result := FAdjudicationList;
end;

Function TFhirClaimResponseAddItemDetail.GetHasAdjudicationList : boolean;
begin
  result := (FAdjudicationList <> nil) and (FAdjudicationList.count > 0);
end;


{ TFhirClaimResponseAddItemDetailListEnumerator }

Constructor TFhirClaimResponseAddItemDetailListEnumerator.Create(list : TFhirClaimResponseAddItemDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseAddItemDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseAddItemDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseAddItemDetailListEnumerator.GetCurrent : TFhirClaimResponseAddItemDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseAddItemDetailList }
procedure TFhirClaimResponseAddItemDetailList.AddItem(value: TFhirClaimResponseAddItemDetail);
begin
  assert(value.ClassName = 'TFhirClaimResponseAddItemDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseAddItemDetail');
  add(value);
end;


function TFhirClaimResponseAddItemDetailList.Append: TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemDetailList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseAddItemDetailList.GetEnumerator : TFhirClaimResponseAddItemDetailListEnumerator;
begin
  result := TFhirClaimResponseAddItemDetailListEnumerator.Create(self.link);
end;

function TFhirClaimResponseAddItemDetailList.Clone: TFhirClaimResponseAddItemDetailList;
begin
  result := TFhirClaimResponseAddItemDetailList(inherited Clone);
end;

function TFhirClaimResponseAddItemDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseAddItemDetailList.GetItemN(index: Integer): TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemDetailList.IndexOf(value: TFhirClaimResponseAddItemDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseAddItemDetailList.Insert(index: Integer): TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemDetailList.InsertItem(index: Integer; value: TFhirClaimResponseAddItemDetail);
begin
  assert(value is TFhirClaimResponseAddItemDetail);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseAddItemDetailList.Item(index: Integer): TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemDetailList.Link: TFhirClaimResponseAddItemDetailList;
begin
  result := TFhirClaimResponseAddItemDetailList(inherited Link);
end;

procedure TFhirClaimResponseAddItemDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseAddItemDetailList.SetItemByIndex(index: Integer; value: TFhirClaimResponseAddItemDetail);
begin
  assert(value is TFhirClaimResponseAddItemDetail);
  FhirClaimResponseAddItemDetails[index] := value;
end;

procedure TFhirClaimResponseAddItemDetailList.SetItemN(index: Integer; value: TFhirClaimResponseAddItemDetail);
begin
  assert(value is TFhirClaimResponseAddItemDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseAddItemDetailAdjudication }

constructor TFhirClaimResponseAddItemDetailAdjudication.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseAddItemDetailAdjudication.Destroy;
begin
  FCode.free;
  FAmount.free;
  FValue.free;
  inherited;
end;

procedure TFhirClaimResponseAddItemDetailAdjudication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClaimResponseAddItemDetailAdjudication(oSource).code.Clone;
  amount := TFhirClaimResponseAddItemDetailAdjudication(oSource).amount.Clone;
  valueElement := TFhirClaimResponseAddItemDetailAdjudication(oSource).valueElement.Clone;
end;

procedure TFhirClaimResponseAddItemDetailAdjudication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirClaimResponseAddItemDetailAdjudication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'decimal', FValue.Link));{2}
end;

procedure TFhirClaimResponseAddItemDetailAdjudication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirClaimResponseAddItemDetailAdjudication.FhirType : string;
begin
  result := 'adjudication';
end;

function TFhirClaimResponseAddItemDetailAdjudication.Link : TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication(inherited Link);
end;

function TFhirClaimResponseAddItemDetailAdjudication.Clone : TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication(inherited Clone);
end;

function TFhirClaimResponseAddItemDetailAdjudication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemDetailAdjudication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemDetailAdjudication(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(amountElement, o.amountElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirClaimResponseAddItemDetailAdjudication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseAddItemDetailAdjudication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseAddItemDetailAdjudication)) then
    result := false
  else
  begin
    o := TFhirClaimResponseAddItemDetailAdjudication(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirClaimResponseAddItemDetailAdjudication }

Procedure TFhirClaimResponseAddItemDetailAdjudication.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirClaimResponseAddItemDetailAdjudication.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;

Procedure TFhirClaimResponseAddItemDetailAdjudication.SetValue(value : TFhirDecimal);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirClaimResponseAddItemDetailAdjudication.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirClaimResponseAddItemDetailAdjudication.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirDecimal.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirClaimResponseAddItemDetailAdjudicationListEnumerator }

Constructor TFhirClaimResponseAddItemDetailAdjudicationListEnumerator.Create(list : TFhirClaimResponseAddItemDetailAdjudicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseAddItemDetailAdjudicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseAddItemDetailAdjudicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseAddItemDetailAdjudicationListEnumerator.GetCurrent : TFhirClaimResponseAddItemDetailAdjudication;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseAddItemDetailAdjudicationList }
procedure TFhirClaimResponseAddItemDetailAdjudicationList.AddItem(value: TFhirClaimResponseAddItemDetailAdjudication);
begin
  assert(value.ClassName = 'TFhirClaimResponseAddItemDetailAdjudication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseAddItemDetailAdjudication');
  add(value);
end;


function TFhirClaimResponseAddItemDetailAdjudicationList.Append: TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemDetailAdjudicationList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.GetEnumerator : TFhirClaimResponseAddItemDetailAdjudicationListEnumerator;
begin
  result := TFhirClaimResponseAddItemDetailAdjudicationListEnumerator.Create(self.link);
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.Clone: TFhirClaimResponseAddItemDetailAdjudicationList;
begin
  result := TFhirClaimResponseAddItemDetailAdjudicationList(inherited Clone);
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.GetItemN(index: Integer): TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.IndexOf(value: TFhirClaimResponseAddItemDetailAdjudication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseAddItemDetailAdjudicationList.Insert(index: Integer): TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseAddItemDetailAdjudicationList.InsertItem(index: Integer; value: TFhirClaimResponseAddItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemDetailAdjudication);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.Item(index: Integer): TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication(ObjectByIndex[index]);
end;

function TFhirClaimResponseAddItemDetailAdjudicationList.Link: TFhirClaimResponseAddItemDetailAdjudicationList;
begin
  result := TFhirClaimResponseAddItemDetailAdjudicationList(inherited Link);
end;

procedure TFhirClaimResponseAddItemDetailAdjudicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseAddItemDetailAdjudicationList.SetItemByIndex(index: Integer; value: TFhirClaimResponseAddItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemDetailAdjudication);
  FhirClaimResponseAddItemDetailAdjudications[index] := value;
end;

procedure TFhirClaimResponseAddItemDetailAdjudicationList.SetItemN(index: Integer; value: TFhirClaimResponseAddItemDetailAdjudication);
begin
  assert(value is TFhirClaimResponseAddItemDetailAdjudication);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseError }

constructor TFhirClaimResponseError.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseError.Destroy;
begin
  FSequenceLinkId.free;
  FDetailSequenceLinkId.free;
  FSubdetailSequenceLinkId.free;
  FCode.free;
  inherited;
end;

procedure TFhirClaimResponseError.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceLinkIdElement := TFhirClaimResponseError(oSource).sequenceLinkIdElement.Clone;
  detailSequenceLinkIdElement := TFhirClaimResponseError(oSource).detailSequenceLinkIdElement.Clone;
  subdetailSequenceLinkIdElement := TFhirClaimResponseError(oSource).subdetailSequenceLinkIdElement.Clone;
  code := TFhirClaimResponseError(oSource).code.Clone;
end;

procedure TFhirClaimResponseError.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
     list.add(FSequenceLinkId.Link);
  if (child_name = 'detailSequenceLinkId') Then
     list.add(FDetailSequenceLinkId.Link);
  if (child_name = 'subdetailSequenceLinkId') Then
     list.add(FSubdetailSequenceLinkId.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
end;

procedure TFhirClaimResponseError.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'positiveInt', FSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detailSequenceLinkId', 'positiveInt', FDetailSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subdetailSequenceLinkId', 'positiveInt', FSubdetailSequenceLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
end;

procedure TFhirClaimResponseError.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'detailSequenceLinkId') then DetailSequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'subdetailSequenceLinkId') then SubdetailSequenceLinkIdElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimResponseError.FhirType : string;
begin
  result := 'error';
end;

function TFhirClaimResponseError.Link : TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError(inherited Link);
end;

function TFhirClaimResponseError.Clone : TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError(inherited Clone);
end;

function TFhirClaimResponseError.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseError;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseError)) then
    result := false
  else
  begin
    o := TFhirClaimResponseError(other);
    result := compareDeep(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareDeep(detailSequenceLinkIdElement, o.detailSequenceLinkIdElement, true) and 
      compareDeep(subdetailSequenceLinkIdElement, o.subdetailSequenceLinkIdElement, true) and 
      compareDeep(codeElement, o.codeElement, true);
  end;
end;

function TFhirClaimResponseError.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseError;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseError)) then
    result := false
  else
  begin
    o := TFhirClaimResponseError(other);
    result := compareValues(sequenceLinkIdElement, o.sequenceLinkIdElement, true) and 
      compareValues(detailSequenceLinkIdElement, o.detailSequenceLinkIdElement, true) and 
      compareValues(subdetailSequenceLinkIdElement, o.subdetailSequenceLinkIdElement, true);
  end;
end;

{ TFhirClaimResponseError }

Procedure TFhirClaimResponseError.SetSequenceLinkId(value : TFhirPositiveInt);
begin
  FSequenceLinkId.free;
  FSequenceLinkId := value;
end;

Function TFhirClaimResponseError.GetSequenceLinkIdST : String;
begin
  if FSequenceLinkId = nil then
    result := ''
  else
    result := FSequenceLinkId.value;
end;

Procedure TFhirClaimResponseError.SetSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSequenceLinkId = nil then
      FSequenceLinkId := TFhirPositiveInt.create;
    FSequenceLinkId.value := value
  end
  else if FSequenceLinkId <> nil then
    FSequenceLinkId.value := '';
end;

Procedure TFhirClaimResponseError.SetDetailSequenceLinkId(value : TFhirPositiveInt);
begin
  FDetailSequenceLinkId.free;
  FDetailSequenceLinkId := value;
end;

Function TFhirClaimResponseError.GetDetailSequenceLinkIdST : String;
begin
  if FDetailSequenceLinkId = nil then
    result := ''
  else
    result := FDetailSequenceLinkId.value;
end;

Procedure TFhirClaimResponseError.SetDetailSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FDetailSequenceLinkId = nil then
      FDetailSequenceLinkId := TFhirPositiveInt.create;
    FDetailSequenceLinkId.value := value
  end
  else if FDetailSequenceLinkId <> nil then
    FDetailSequenceLinkId.value := '';
end;

Procedure TFhirClaimResponseError.SetSubdetailSequenceLinkId(value : TFhirPositiveInt);
begin
  FSubdetailSequenceLinkId.free;
  FSubdetailSequenceLinkId := value;
end;

Function TFhirClaimResponseError.GetSubdetailSequenceLinkIdST : String;
begin
  if FSubdetailSequenceLinkId = nil then
    result := ''
  else
    result := FSubdetailSequenceLinkId.value;
end;

Procedure TFhirClaimResponseError.SetSubdetailSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSubdetailSequenceLinkId = nil then
      FSubdetailSequenceLinkId := TFhirPositiveInt.create;
    FSubdetailSequenceLinkId.value := value
  end
  else if FSubdetailSequenceLinkId <> nil then
    FSubdetailSequenceLinkId.value := '';
end;

Procedure TFhirClaimResponseError.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;


{ TFhirClaimResponseErrorListEnumerator }

Constructor TFhirClaimResponseErrorListEnumerator.Create(list : TFhirClaimResponseErrorList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseErrorListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseErrorListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseErrorListEnumerator.GetCurrent : TFhirClaimResponseError;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseErrorList }
procedure TFhirClaimResponseErrorList.AddItem(value: TFhirClaimResponseError);
begin
  assert(value.ClassName = 'TFhirClaimResponseError', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseError');
  add(value);
end;


function TFhirClaimResponseErrorList.Append: TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseErrorList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseErrorList.GetEnumerator : TFhirClaimResponseErrorListEnumerator;
begin
  result := TFhirClaimResponseErrorListEnumerator.Create(self.link);
end;

function TFhirClaimResponseErrorList.Clone: TFhirClaimResponseErrorList;
begin
  result := TFhirClaimResponseErrorList(inherited Clone);
end;

function TFhirClaimResponseErrorList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseErrorList.GetItemN(index: Integer): TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError(ObjectByIndex[index]);
end;

function TFhirClaimResponseErrorList.IndexOf(value: TFhirClaimResponseError): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseErrorList.Insert(index: Integer): TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseErrorList.InsertItem(index: Integer; value: TFhirClaimResponseError);
begin
  assert(value is TFhirClaimResponseError);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseErrorList.Item(index: Integer): TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError(ObjectByIndex[index]);
end;

function TFhirClaimResponseErrorList.Link: TFhirClaimResponseErrorList;
begin
  result := TFhirClaimResponseErrorList(inherited Link);
end;

procedure TFhirClaimResponseErrorList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseErrorList.SetItemByIndex(index: Integer; value: TFhirClaimResponseError);
begin
  assert(value is TFhirClaimResponseError);
  FhirClaimResponseErrors[index] := value;
end;

procedure TFhirClaimResponseErrorList.SetItemN(index: Integer; value: TFhirClaimResponseError);
begin
  assert(value is TFhirClaimResponseError);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseNote }

constructor TFhirClaimResponseNote.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseNote.Destroy;
begin
  FNumber.free;
  FType_.free;
  FText.free;
  inherited;
end;

procedure TFhirClaimResponseNote.Assign(oSource : TAdvObject);
begin
  inherited;
  numberElement := TFhirClaimResponseNote(oSource).numberElement.Clone;
  type_ := TFhirClaimResponseNote(oSource).type_.Clone;
  textElement := TFhirClaimResponseNote(oSource).textElement.Clone;
end;

procedure TFhirClaimResponseNote.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'number') Then
     list.add(FNumber.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
end;

procedure TFhirClaimResponseNote.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'number', 'positiveInt', FNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
end;

procedure TFhirClaimResponseNote.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'number') then NumberElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirClaimResponseNote.FhirType : string;
begin
  result := 'note';
end;

function TFhirClaimResponseNote.Link : TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote(inherited Link);
end;

function TFhirClaimResponseNote.Clone : TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote(inherited Clone);
end;

function TFhirClaimResponseNote.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseNote;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseNote)) then
    result := false
  else
  begin
    o := TFhirClaimResponseNote(other);
    result := compareDeep(numberElement, o.numberElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(textElement, o.textElement, true);
  end;
end;

function TFhirClaimResponseNote.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseNote;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseNote)) then
    result := false
  else
  begin
    o := TFhirClaimResponseNote(other);
    result := compareValues(numberElement, o.numberElement, true) and compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirClaimResponseNote }

Procedure TFhirClaimResponseNote.SetNumber(value : TFhirPositiveInt);
begin
  FNumber.free;
  FNumber := value;
end;

Function TFhirClaimResponseNote.GetNumberST : String;
begin
  if FNumber = nil then
    result := ''
  else
    result := FNumber.value;
end;

Procedure TFhirClaimResponseNote.SetNumberST(value : String);
begin
  if value <> '' then
  begin
    if FNumber = nil then
      FNumber := TFhirPositiveInt.create;
    FNumber.value := value
  end
  else if FNumber <> nil then
    FNumber.value := '';
end;

Procedure TFhirClaimResponseNote.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirClaimResponseNote.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirClaimResponseNote.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirClaimResponseNote.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;


{ TFhirClaimResponseNoteListEnumerator }

Constructor TFhirClaimResponseNoteListEnumerator.Create(list : TFhirClaimResponseNoteList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseNoteListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseNoteListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseNoteListEnumerator.GetCurrent : TFhirClaimResponseNote;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseNoteList }
procedure TFhirClaimResponseNoteList.AddItem(value: TFhirClaimResponseNote);
begin
  assert(value.ClassName = 'TFhirClaimResponseNote', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseNote');
  add(value);
end;


function TFhirClaimResponseNoteList.Append: TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseNoteList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseNoteList.GetEnumerator : TFhirClaimResponseNoteListEnumerator;
begin
  result := TFhirClaimResponseNoteListEnumerator.Create(self.link);
end;

function TFhirClaimResponseNoteList.Clone: TFhirClaimResponseNoteList;
begin
  result := TFhirClaimResponseNoteList(inherited Clone);
end;

function TFhirClaimResponseNoteList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseNoteList.GetItemN(index: Integer): TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote(ObjectByIndex[index]);
end;

function TFhirClaimResponseNoteList.IndexOf(value: TFhirClaimResponseNote): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseNoteList.Insert(index: Integer): TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseNoteList.InsertItem(index: Integer; value: TFhirClaimResponseNote);
begin
  assert(value is TFhirClaimResponseNote);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseNoteList.Item(index: Integer): TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote(ObjectByIndex[index]);
end;

function TFhirClaimResponseNoteList.Link: TFhirClaimResponseNoteList;
begin
  result := TFhirClaimResponseNoteList(inherited Link);
end;

procedure TFhirClaimResponseNoteList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseNoteList.SetItemByIndex(index: Integer; value: TFhirClaimResponseNote);
begin
  assert(value is TFhirClaimResponseNote);
  FhirClaimResponseNotes[index] := value;
end;

procedure TFhirClaimResponseNoteList.SetItemN(index: Integer; value: TFhirClaimResponseNote);
begin
  assert(value is TFhirClaimResponseNote);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponseCoverage }

constructor TFhirClaimResponseCoverage.Create;
begin
  inherited;
end;

destructor TFhirClaimResponseCoverage.Destroy;
begin
  FSequence.free;
  FFocal.free;
  FCoverage.free;
  FBusinessArrangement.free;
  FRelationship.free;
  FPreAuthRefList.Free;
  FClaimResponse.free;
  FOriginalRuleset.free;
  inherited;
end;

procedure TFhirClaimResponseCoverage.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceElement := TFhirClaimResponseCoverage(oSource).sequenceElement.Clone;
  focalElement := TFhirClaimResponseCoverage(oSource).focalElement.Clone;
  coverage := TFhirClaimResponseCoverage(oSource).coverage.Clone;
  businessArrangementElement := TFhirClaimResponseCoverage(oSource).businessArrangementElement.Clone;
  relationship := TFhirClaimResponseCoverage(oSource).relationship.Clone;
  if (TFhirClaimResponseCoverage(oSource).FPreAuthRefList = nil) then
  begin
    FPreAuthRefList.free;
    FPreAuthRefList := nil;
  end
  else
  begin
    FPreAuthRefList := TFhirStringList.Create;
    FPreAuthRefList.Assign(TFhirClaimResponseCoverage(oSource).FPreAuthRefList);
  end;
  claimResponse := TFhirClaimResponseCoverage(oSource).claimResponse.Clone;
  originalRuleset := TFhirClaimResponseCoverage(oSource).originalRuleset.Clone;
end;

procedure TFhirClaimResponseCoverage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'focal') Then
     list.add(FFocal.Link);
  if (child_name = 'coverage') Then
     list.add(FCoverage.Link);
  if (child_name = 'businessArrangement') Then
     list.add(FBusinessArrangement.Link);
  if (child_name = 'relationship') Then
     list.add(FRelationship.Link);
  if (child_name = 'preAuthRef') Then
    list.addAll(FPreAuthRefList);
  if (child_name = 'claimResponse') Then
     list.add(FClaimResponse.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
end;

procedure TFhirClaimResponseCoverage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'focal', 'boolean', FFocal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'coverage', 'Reference(Coverage)', FCoverage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'businessArrangement', 'string', FBusinessArrangement.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relationship', 'Coding', FRelationship.Link));{2}
  oList.add(TFHIRProperty.create(self, 'preAuthRef', 'string', FPreAuthRefList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'claimResponse', 'Reference(ClaimResponse)', FClaimResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
end;

procedure TFhirClaimResponseCoverage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'focal') then FocalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'coverage') then Coverage := propValue as TFhirReference{TFhirCoverage}{4b}
  else if (propName = 'businessArrangement') then BusinessArrangementElement := propValue as TFhirString{5a}
  else if (propName = 'relationship') then Relationship := propValue as TFhirCoding{4b}
  else if (propName = 'preAuthRef') then PreAuthRefList.add(propValue as TFhirString){2}
  else if (propName = 'claimResponse') then ClaimResponse := propValue as TFhirReference{TFhirClaimResponse}{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirClaimResponseCoverage.FhirType : string;
begin
  result := 'coverage';
end;

function TFhirClaimResponseCoverage.Link : TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage(inherited Link);
end;

function TFhirClaimResponseCoverage.Clone : TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage(inherited Clone);
end;

function TFhirClaimResponseCoverage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseCoverage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClaimResponseCoverage)) then
    result := false
  else
  begin
    o := TFhirClaimResponseCoverage(other);
    result := compareDeep(sequenceElement, o.sequenceElement, true) and compareDeep(focalElement, o.focalElement, true) and 
      compareDeep(coverageElement, o.coverageElement, true) and compareDeep(businessArrangementElement, o.businessArrangementElement, true) and 
      compareDeep(relationshipElement, o.relationshipElement, true) and compareDeep(preAuthRefList, o.preAuthRefList, true) and 
      compareDeep(claimResponseElement, o.claimResponseElement, true) and compareDeep(originalRulesetElement, o.originalRulesetElement, true);
  end;
end;

function TFhirClaimResponseCoverage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClaimResponseCoverage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClaimResponseCoverage)) then
    result := false
  else
  begin
    o := TFhirClaimResponseCoverage(other);
    result := compareValues(sequenceElement, o.sequenceElement, true) and compareValues(focalElement, o.focalElement, true) and 
      compareValues(businessArrangementElement, o.businessArrangementElement, true) and 
      compareValues(preAuthRefList, o.preAuthRefList, true);
  end;
end;

{ TFhirClaimResponseCoverage }

Procedure TFhirClaimResponseCoverage.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirClaimResponseCoverage.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirClaimResponseCoverage.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirClaimResponseCoverage.SetFocal(value : TFhirBoolean);
begin
  FFocal.free;
  FFocal := value;
end;

Function TFhirClaimResponseCoverage.GetFocalST : Boolean;
begin
  if FFocal = nil then
    result := false
  else
    result := FFocal.value;
end;

Procedure TFhirClaimResponseCoverage.SetFocalST(value : Boolean);
begin
  if FFocal = nil then
    FFocal := TFhirBoolean.create;
  FFocal.value := value
end;

Procedure TFhirClaimResponseCoverage.SetCoverage(value : TFhirReference{TFhirCoverage});
begin
  FCoverage.free;
  FCoverage := value;
end;

Procedure TFhirClaimResponseCoverage.SetBusinessArrangement(value : TFhirString);
begin
  FBusinessArrangement.free;
  FBusinessArrangement := value;
end;

Function TFhirClaimResponseCoverage.GetBusinessArrangementST : String;
begin
  if FBusinessArrangement = nil then
    result := ''
  else
    result := FBusinessArrangement.value;
end;

Procedure TFhirClaimResponseCoverage.SetBusinessArrangementST(value : String);
begin
  if value <> '' then
  begin
    if FBusinessArrangement = nil then
      FBusinessArrangement := TFhirString.create;
    FBusinessArrangement.value := value
  end
  else if FBusinessArrangement <> nil then
    FBusinessArrangement.value := '';
end;

Procedure TFhirClaimResponseCoverage.SetRelationship(value : TFhirCoding);
begin
  FRelationship.free;
  FRelationship := value;
end;

Function TFhirClaimResponseCoverage.GetPreAuthRefList : TFhirStringList;
begin
  if FPreAuthRefList = nil then
    FPreAuthRefList := TFhirStringList.Create;
  result := FPreAuthRefList;
end;

Function TFhirClaimResponseCoverage.GetHasPreAuthRefList : boolean;
begin
  result := (FPreAuthRefList <> nil) and (FPreAuthRefList.count > 0);
end;

Procedure TFhirClaimResponseCoverage.SetClaimResponse(value : TFhirReference{TFhirClaimResponse});
begin
  FClaimResponse.free;
  FClaimResponse := value;
end;

Procedure TFhirClaimResponseCoverage.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;


{ TFhirClaimResponseCoverageListEnumerator }

Constructor TFhirClaimResponseCoverageListEnumerator.Create(list : TFhirClaimResponseCoverageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseCoverageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseCoverageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseCoverageListEnumerator.GetCurrent : TFhirClaimResponseCoverage;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseCoverageList }
procedure TFhirClaimResponseCoverageList.AddItem(value: TFhirClaimResponseCoverage);
begin
  assert(value.ClassName = 'TFhirClaimResponseCoverage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponseCoverage');
  add(value);
end;


function TFhirClaimResponseCoverageList.Append: TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseCoverageList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseCoverageList.GetEnumerator : TFhirClaimResponseCoverageListEnumerator;
begin
  result := TFhirClaimResponseCoverageListEnumerator.Create(self.link);
end;

function TFhirClaimResponseCoverageList.Clone: TFhirClaimResponseCoverageList;
begin
  result := TFhirClaimResponseCoverageList(inherited Clone);
end;

function TFhirClaimResponseCoverageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseCoverageList.GetItemN(index: Integer): TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage(ObjectByIndex[index]);
end;

function TFhirClaimResponseCoverageList.IndexOf(value: TFhirClaimResponseCoverage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseCoverageList.Insert(index: Integer): TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseCoverageList.InsertItem(index: Integer; value: TFhirClaimResponseCoverage);
begin
  assert(value is TFhirClaimResponseCoverage);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseCoverageList.Item(index: Integer): TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage(ObjectByIndex[index]);
end;

function TFhirClaimResponseCoverageList.Link: TFhirClaimResponseCoverageList;
begin
  result := TFhirClaimResponseCoverageList(inherited Link);
end;

procedure TFhirClaimResponseCoverageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseCoverageList.SetItemByIndex(index: Integer; value: TFhirClaimResponseCoverage);
begin
  assert(value is TFhirClaimResponseCoverage);
  FhirClaimResponseCoverages[index] := value;
end;

procedure TFhirClaimResponseCoverageList.SetItemN(index: Integer; value: TFhirClaimResponseCoverage);
begin
  assert(value is TFhirClaimResponseCoverage);
  ObjectByIndex[index] := value;
end;

{ TFhirClaimResponse }

constructor TFhirClaimResponse.Create;
begin
  inherited;
end;

destructor TFhirClaimResponse.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  FOutcome.free;
  FDisposition.free;
  FPayeeType.free;
  FItemList.Free;
  FAddItemList.Free;
  FErrorList.Free;
  FTotalCost.free;
  FUnallocDeductable.free;
  FTotalBenefit.free;
  FPaymentAdjustment.free;
  FPaymentAdjustmentReason.free;
  FPaymentDate.free;
  FPaymentAmount.free;
  FPaymentRef.free;
  FReserved.free;
  FForm.free;
  FNoteList.Free;
  FCoverageList.Free;
  inherited;
end;

function TFhirClaimResponse.GetResourceType : TFhirResourceType;
begin
  result := frtClaimResponse;
end;

procedure TFhirClaimResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirClaimResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirClaimResponse(oSource).FIdentifierList);
  end;
  request := TFhirClaimResponse(oSource).request.Clone;
  ruleset := TFhirClaimResponse(oSource).ruleset.Clone;
  originalRuleset := TFhirClaimResponse(oSource).originalRuleset.Clone;
  createdElement := TFhirClaimResponse(oSource).createdElement.Clone;
  organization := TFhirClaimResponse(oSource).organization.Clone;
  requestProvider := TFhirClaimResponse(oSource).requestProvider.Clone;
  requestOrganization := TFhirClaimResponse(oSource).requestOrganization.Clone;
  FOutcome := TFhirClaimResponse(oSource).FOutcome.Link;
  dispositionElement := TFhirClaimResponse(oSource).dispositionElement.Clone;
  payeeType := TFhirClaimResponse(oSource).payeeType.Clone;
  if (TFhirClaimResponse(oSource).FItemList = nil) then
  begin
    FItemList.free;
    FItemList := nil;
  end
  else
  begin
    FItemList := TFhirClaimResponseItemList.Create;
    FItemList.Assign(TFhirClaimResponse(oSource).FItemList);
  end;
  if (TFhirClaimResponse(oSource).FAddItemList = nil) then
  begin
    FAddItemList.free;
    FAddItemList := nil;
  end
  else
  begin
    FAddItemList := TFhirClaimResponseAddItemList.Create;
    FAddItemList.Assign(TFhirClaimResponse(oSource).FAddItemList);
  end;
  if (TFhirClaimResponse(oSource).FErrorList = nil) then
  begin
    FErrorList.free;
    FErrorList := nil;
  end
  else
  begin
    FErrorList := TFhirClaimResponseErrorList.Create;
    FErrorList.Assign(TFhirClaimResponse(oSource).FErrorList);
  end;
  totalCost := TFhirClaimResponse(oSource).totalCost.Clone;
  unallocDeductable := TFhirClaimResponse(oSource).unallocDeductable.Clone;
  totalBenefit := TFhirClaimResponse(oSource).totalBenefit.Clone;
  paymentAdjustment := TFhirClaimResponse(oSource).paymentAdjustment.Clone;
  paymentAdjustmentReason := TFhirClaimResponse(oSource).paymentAdjustmentReason.Clone;
  paymentDateElement := TFhirClaimResponse(oSource).paymentDateElement.Clone;
  paymentAmount := TFhirClaimResponse(oSource).paymentAmount.Clone;
  paymentRef := TFhirClaimResponse(oSource).paymentRef.Clone;
  reserved := TFhirClaimResponse(oSource).reserved.Clone;
  form := TFhirClaimResponse(oSource).form.Clone;
  if (TFhirClaimResponse(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirClaimResponseNoteList.Create;
    FNoteList.Assign(TFhirClaimResponse(oSource).FNoteList);
  end;
  if (TFhirClaimResponse(oSource).FCoverageList = nil) then
  begin
    FCoverageList.free;
    FCoverageList := nil;
  end
  else
  begin
    FCoverageList := TFhirClaimResponseCoverageList.Create;
    FCoverageList.Assign(TFhirClaimResponse(oSource).FCoverageList);
  end;
end;

procedure TFhirClaimResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'payeeType') Then
     list.add(FPayeeType.Link);
  if (child_name = 'item') Then
    list.addAll(FItemList);
  if (child_name = 'addItem') Then
    list.addAll(FAddItemList);
  if (child_name = 'error') Then
    list.addAll(FErrorList);
  if (child_name = 'totalCost') Then
     list.add(FTotalCost.Link);
  if (child_name = 'unallocDeductable') Then
     list.add(FUnallocDeductable.Link);
  if (child_name = 'totalBenefit') Then
     list.add(FTotalBenefit.Link);
  if (child_name = 'paymentAdjustment') Then
     list.add(FPaymentAdjustment.Link);
  if (child_name = 'paymentAdjustmentReason') Then
     list.add(FPaymentAdjustmentReason.Link);
  if (child_name = 'paymentDate') Then
     list.add(FPaymentDate.Link);
  if (child_name = 'paymentAmount') Then
     list.add(FPaymentAmount.Link);
  if (child_name = 'paymentRef') Then
     list.add(FPaymentRef.Link);
  if (child_name = 'reserved') Then
     list.add(FReserved.Link);
  if (child_name = 'form') Then
     list.add(FForm.Link);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
  if (child_name = 'coverage') Then
    list.addAll(FCoverageList);
end;

procedure TFhirClaimResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Claim)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'payeeType', 'Coding', FPayeeType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'item', '', FItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'addItem', '', FAddItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'error', '', FErrorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'totalCost', 'Money', FTotalCost.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unallocDeductable', 'Money', FUnallocDeductable.Link));{2}
  oList.add(TFHIRProperty.create(self, 'totalBenefit', 'Money', FTotalBenefit.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentAdjustment', 'Money', FPaymentAdjustment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentAdjustmentReason', 'Coding', FPaymentAdjustmentReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentDate', 'date', FPaymentDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentAmount', 'Money', FPaymentAmount.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentRef', 'Identifier', FPaymentRef.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reserved', 'Coding', FReserved.Link));{2}
  oList.add(TFHIRProperty.create(self, 'form', 'Coding', FForm.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', '', FNoteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'coverage', '', FCoverageList.Link)){3};
end;

procedure TFhirClaimResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirClaim}{4b}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'payeeType') then PayeeType := propValue as TFhirCoding{4b}
  else if (propName = 'item') then ItemList.add(propValue as TFhirClaimResponseItem){2}
  else if (propName = 'addItem') then AddItemList.add(propValue as TFhirClaimResponseAddItem){2}
  else if (propName = 'error') then ErrorList.add(propValue as TFhirClaimResponseError){2}
  else if (propName = 'totalCost') then TotalCost := propValue as TFhirQuantity{4b}
  else if (propName = 'unallocDeductable') then UnallocDeductable := propValue as TFhirQuantity{4b}
  else if (propName = 'totalBenefit') then TotalBenefit := propValue as TFhirQuantity{4b}
  else if (propName = 'paymentAdjustment') then PaymentAdjustment := propValue as TFhirQuantity{4b}
  else if (propName = 'paymentAdjustmentReason') then PaymentAdjustmentReason := propValue as TFhirCoding{4b}
  else if (propName = 'paymentDate') then PaymentDateElement := propValue as TFhirDate{5a}
  else if (propName = 'paymentAmount') then PaymentAmount := propValue as TFhirQuantity{4b}
  else if (propName = 'paymentRef') then PaymentRef := propValue as TFhirIdentifier{4b}
  else if (propName = 'reserved') then Reserved := propValue as TFhirCoding{4b}
  else if (propName = 'form') then Form := propValue as TFhirCoding{4b}
  else if (propName = 'note') then NoteList.add(propValue as TFhirClaimResponseNote){2}
  else if (propName = 'coverage') then CoverageList.add(propValue as TFhirClaimResponseCoverage){2}
  else inherited;
end;

function TFhirClaimResponse.FhirType : string;
begin
  result := 'ClaimResponse';
end;

function TFhirClaimResponse.Link : TFhirClaimResponse;
begin
  result := TFhirClaimResponse(inherited Link);
end;

function TFhirClaimResponse.Clone : TFhirClaimResponse;
begin
  result := TFhirClaimResponse(inherited Clone);
end;

{ TFhirClaimResponse }

Function TFhirClaimResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirClaimResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirClaimResponse.SetRequest(value : TFhirReference{TFhirClaim});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirClaimResponse.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirClaimResponse.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirClaimResponse.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirClaimResponse.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirClaimResponse.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirClaimResponse.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirClaimResponse.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirClaimResponse.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;

Procedure TFhirClaimResponse.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirClaimResponse.GetOutcomeST : TFhirRemittanceOutcome;
begin
  if FOutcome = nil then
    result := TFhirRemittanceOutcome(0)
  else
    result := TFhirRemittanceOutcome(StringArrayIndexOfSensitive(CODES_TFhirRemittanceOutcome, FOutcome.value));
end;

Procedure TFhirClaimResponse.SetOutcomeST(value : TFhirRemittanceOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirRemittanceOutcome[value]);
end;

Procedure TFhirClaimResponse.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirClaimResponse.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirClaimResponse.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirClaimResponse.SetPayeeType(value : TFhirCoding);
begin
  FPayeeType.free;
  FPayeeType := value;
end;

Function TFhirClaimResponse.GetItemList : TFhirClaimResponseItemList;
begin
  if FItemList = nil then
    FItemList := TFhirClaimResponseItemList.Create;
  result := FItemList;
end;

Function TFhirClaimResponse.GetHasItemList : boolean;
begin
  result := (FItemList <> nil) and (FItemList.count > 0);
end;

Function TFhirClaimResponse.GetAddItemList : TFhirClaimResponseAddItemList;
begin
  if FAddItemList = nil then
    FAddItemList := TFhirClaimResponseAddItemList.Create;
  result := FAddItemList;
end;

Function TFhirClaimResponse.GetHasAddItemList : boolean;
begin
  result := (FAddItemList <> nil) and (FAddItemList.count > 0);
end;

Function TFhirClaimResponse.GetErrorList : TFhirClaimResponseErrorList;
begin
  if FErrorList = nil then
    FErrorList := TFhirClaimResponseErrorList.Create;
  result := FErrorList;
end;

Function TFhirClaimResponse.GetHasErrorList : boolean;
begin
  result := (FErrorList <> nil) and (FErrorList.count > 0);
end;

Procedure TFhirClaimResponse.SetTotalCost(value : TFhirQuantity);
begin
  FTotalCost.free;
  FTotalCost := value;
end;

Procedure TFhirClaimResponse.SetUnallocDeductable(value : TFhirQuantity);
begin
  FUnallocDeductable.free;
  FUnallocDeductable := value;
end;

Procedure TFhirClaimResponse.SetTotalBenefit(value : TFhirQuantity);
begin
  FTotalBenefit.free;
  FTotalBenefit := value;
end;

Procedure TFhirClaimResponse.SetPaymentAdjustment(value : TFhirQuantity);
begin
  FPaymentAdjustment.free;
  FPaymentAdjustment := value;
end;

Procedure TFhirClaimResponse.SetPaymentAdjustmentReason(value : TFhirCoding);
begin
  FPaymentAdjustmentReason.free;
  FPaymentAdjustmentReason := value;
end;

Procedure TFhirClaimResponse.SetPaymentDate(value : TFhirDate);
begin
  FPaymentDate.free;
  FPaymentDate := value;
end;

Function TFhirClaimResponse.GetPaymentDateST : TDateAndTime;
begin
  if FPaymentDate = nil then
    result := nil
  else
    result := FPaymentDate.value;
end;

Procedure TFhirClaimResponse.SetPaymentDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FPaymentDate = nil then
      FPaymentDate := TFhirDate.create;
    FPaymentDate.value := value
  end
  else if FPaymentDate <> nil then
    FPaymentDate.value := nil;
end;

Procedure TFhirClaimResponse.SetPaymentAmount(value : TFhirQuantity);
begin
  FPaymentAmount.free;
  FPaymentAmount := value;
end;

Procedure TFhirClaimResponse.SetPaymentRef(value : TFhirIdentifier);
begin
  FPaymentRef.free;
  FPaymentRef := value;
end;

Procedure TFhirClaimResponse.SetReserved(value : TFhirCoding);
begin
  FReserved.free;
  FReserved := value;
end;

Procedure TFhirClaimResponse.SetForm(value : TFhirCoding);
begin
  FForm.free;
  FForm := value;
end;

Function TFhirClaimResponse.GetNoteList : TFhirClaimResponseNoteList;
begin
  if FNoteList = nil then
    FNoteList := TFhirClaimResponseNoteList.Create;
  result := FNoteList;
end;

Function TFhirClaimResponse.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;

Function TFhirClaimResponse.GetCoverageList : TFhirClaimResponseCoverageList;
begin
  if FCoverageList = nil then
    FCoverageList := TFhirClaimResponseCoverageList.Create;
  result := FCoverageList;
end;

Function TFhirClaimResponse.GetHasCoverageList : boolean;
begin
  result := (FCoverageList <> nil) and (FCoverageList.count > 0);
end;


{ TFhirClaimResponseListEnumerator }

Constructor TFhirClaimResponseListEnumerator.Create(list : TFhirClaimResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClaimResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClaimResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClaimResponseListEnumerator.GetCurrent : TFhirClaimResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirClaimResponseList }
procedure TFhirClaimResponseList.AddItem(value: TFhirClaimResponse);
begin
  assert(value.ClassName = 'TFhirClaimResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClaimResponse');
  add(value);
end;


function TFhirClaimResponseList.Append: TFhirClaimResponse;
begin
  result := TFhirClaimResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseList.ClearItems;
begin
  Clear;
end;

function TFhirClaimResponseList.GetEnumerator : TFhirClaimResponseListEnumerator;
begin
  result := TFhirClaimResponseListEnumerator.Create(self.link);
end;

function TFhirClaimResponseList.Clone: TFhirClaimResponseList;
begin
  result := TFhirClaimResponseList(inherited Clone);
end;

function TFhirClaimResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClaimResponseList.GetItemN(index: Integer): TFhirClaimResponse;
begin
  result := TFhirClaimResponse(ObjectByIndex[index]);
end;

function TFhirClaimResponseList.IndexOf(value: TFhirClaimResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClaimResponseList.Insert(index: Integer): TFhirClaimResponse;
begin
  result := TFhirClaimResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClaimResponseList.InsertItem(index: Integer; value: TFhirClaimResponse);
begin
  assert(value is TFhirClaimResponse);
  Inherited Insert(index, value);
end;

function TFhirClaimResponseList.Item(index: Integer): TFhirClaimResponse;
begin
  result := TFhirClaimResponse(ObjectByIndex[index]);
end;

function TFhirClaimResponseList.Link: TFhirClaimResponseList;
begin
  result := TFhirClaimResponseList(inherited Link);
end;

procedure TFhirClaimResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClaimResponseList.SetItemByIndex(index: Integer; value: TFhirClaimResponse);
begin
  assert(value is TFhirClaimResponse);
  FhirClaimResponses[index] := value;
end;

procedure TFhirClaimResponseList.SetItemN(index: Integer; value: TFhirClaimResponse);
begin
  assert(value is TFhirClaimResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirClinicalImpressionInvestigations }

constructor TFhirClinicalImpressionInvestigations.Create;
begin
  inherited;
end;

destructor TFhirClinicalImpressionInvestigations.Destroy;
begin
  FCode.free;
  FItemList.Free;
  inherited;
end;

procedure TFhirClinicalImpressionInvestigations.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirClinicalImpressionInvestigations(oSource).code.Clone;
  if (TFhirClinicalImpressionInvestigations(oSource).FItemList = nil) then
  begin
    FItemList.free;
    FItemList := nil;
  end
  else
  begin
    FItemList := TFhirReferenceList{Resource}.Create;
    FItemList.Assign(TFhirClinicalImpressionInvestigations(oSource).FItemList);
  end;
end;

procedure TFhirClinicalImpressionInvestigations.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'item') Then
    list.addAll(FItemList);
end;

procedure TFhirClinicalImpressionInvestigations.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'item', 'Reference(Observation|QuestionnaireResponse|FamilyMemberHistory|DiagnosticReport)', FItemList.Link)){3};
end;

procedure TFhirClinicalImpressionInvestigations.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'item') then ItemList.add(propValue as TFhirReference{Resource}){2}
  else inherited;
end;

function TFhirClinicalImpressionInvestigations.FhirType : string;
begin
  result := 'investigations';
end;

function TFhirClinicalImpressionInvestigations.Link : TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations(inherited Link);
end;

function TFhirClinicalImpressionInvestigations.Clone : TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations(inherited Clone);
end;

function TFhirClinicalImpressionInvestigations.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionInvestigations;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionInvestigations)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionInvestigations(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(itemList, o.itemList, true);
  end;
end;

function TFhirClinicalImpressionInvestigations.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionInvestigations;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionInvestigations)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionInvestigations(other);
    result := true;
  end;
end;

{ TFhirClinicalImpressionInvestigations }

Procedure TFhirClinicalImpressionInvestigations.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirClinicalImpressionInvestigations.GetItemList : TFhirReferenceList{Resource};
begin
  if FItemList = nil then
    FItemList := TFhirReferenceList{Resource}.Create;
  result := FItemList;
end;

Function TFhirClinicalImpressionInvestigations.GetHasItemList : boolean;
begin
  result := (FItemList <> nil) and (FItemList.count > 0);
end;


{ TFhirClinicalImpressionInvestigationsListEnumerator }

Constructor TFhirClinicalImpressionInvestigationsListEnumerator.Create(list : TFhirClinicalImpressionInvestigationsList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClinicalImpressionInvestigationsListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClinicalImpressionInvestigationsListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClinicalImpressionInvestigationsListEnumerator.GetCurrent : TFhirClinicalImpressionInvestigations;
begin
  Result := FList[FIndex];
end;


{ TFhirClinicalImpressionInvestigationsList }
procedure TFhirClinicalImpressionInvestigationsList.AddItem(value: TFhirClinicalImpressionInvestigations);
begin
  assert(value.ClassName = 'TFhirClinicalImpressionInvestigations', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClinicalImpressionInvestigations');
  add(value);
end;


function TFhirClinicalImpressionInvestigationsList.Append: TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionInvestigationsList.ClearItems;
begin
  Clear;
end;

function TFhirClinicalImpressionInvestigationsList.GetEnumerator : TFhirClinicalImpressionInvestigationsListEnumerator;
begin
  result := TFhirClinicalImpressionInvestigationsListEnumerator.Create(self.link);
end;

function TFhirClinicalImpressionInvestigationsList.Clone: TFhirClinicalImpressionInvestigationsList;
begin
  result := TFhirClinicalImpressionInvestigationsList(inherited Clone);
end;

function TFhirClinicalImpressionInvestigationsList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClinicalImpressionInvestigationsList.GetItemN(index: Integer): TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionInvestigationsList.IndexOf(value: TFhirClinicalImpressionInvestigations): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClinicalImpressionInvestigationsList.Insert(index: Integer): TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionInvestigationsList.InsertItem(index: Integer; value: TFhirClinicalImpressionInvestigations);
begin
  assert(value is TFhirClinicalImpressionInvestigations);
  Inherited Insert(index, value);
end;

function TFhirClinicalImpressionInvestigationsList.Item(index: Integer): TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionInvestigationsList.Link: TFhirClinicalImpressionInvestigationsList;
begin
  result := TFhirClinicalImpressionInvestigationsList(inherited Link);
end;

procedure TFhirClinicalImpressionInvestigationsList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClinicalImpressionInvestigationsList.SetItemByIndex(index: Integer; value: TFhirClinicalImpressionInvestigations);
begin
  assert(value is TFhirClinicalImpressionInvestigations);
  FhirClinicalImpressionInvestigations[index] := value;
end;

procedure TFhirClinicalImpressionInvestigationsList.SetItemN(index: Integer; value: TFhirClinicalImpressionInvestigations);
begin
  assert(value is TFhirClinicalImpressionInvestigations);
  ObjectByIndex[index] := value;
end;

{ TFhirClinicalImpressionFinding }

constructor TFhirClinicalImpressionFinding.Create;
begin
  inherited;
end;

destructor TFhirClinicalImpressionFinding.Destroy;
begin
  FItem.free;
  FCause.free;
  inherited;
end;

procedure TFhirClinicalImpressionFinding.Assign(oSource : TAdvObject);
begin
  inherited;
  item := TFhirClinicalImpressionFinding(oSource).item.Clone;
  causeElement := TFhirClinicalImpressionFinding(oSource).causeElement.Clone;
end;

procedure TFhirClinicalImpressionFinding.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'item') Then
     list.add(FItem.Link);
  if (child_name = 'cause') Then
     list.add(FCause.Link);
end;

procedure TFhirClinicalImpressionFinding.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'item', 'CodeableConcept', FItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'cause', 'string', FCause.Link));{2}
end;

procedure TFhirClinicalImpressionFinding.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'item') then Item := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'cause') then CauseElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirClinicalImpressionFinding.FhirType : string;
begin
  result := 'finding';
end;

function TFhirClinicalImpressionFinding.Link : TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding(inherited Link);
end;

function TFhirClinicalImpressionFinding.Clone : TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding(inherited Clone);
end;

function TFhirClinicalImpressionFinding.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionFinding;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionFinding)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionFinding(other);
    result := compareDeep(itemElement, o.itemElement, true) and compareDeep(causeElement, o.causeElement, true);
  end;
end;

function TFhirClinicalImpressionFinding.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionFinding;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionFinding)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionFinding(other);
    result := compareValues(causeElement, o.causeElement, true);
  end;
end;

{ TFhirClinicalImpressionFinding }

Procedure TFhirClinicalImpressionFinding.SetItem(value : TFhirCodeableConcept);
begin
  FItem.free;
  FItem := value;
end;

Procedure TFhirClinicalImpressionFinding.SetCause(value : TFhirString);
begin
  FCause.free;
  FCause := value;
end;

Function TFhirClinicalImpressionFinding.GetCauseST : String;
begin
  if FCause = nil then
    result := ''
  else
    result := FCause.value;
end;

Procedure TFhirClinicalImpressionFinding.SetCauseST(value : String);
begin
  if value <> '' then
  begin
    if FCause = nil then
      FCause := TFhirString.create;
    FCause.value := value
  end
  else if FCause <> nil then
    FCause.value := '';
end;


{ TFhirClinicalImpressionFindingListEnumerator }

Constructor TFhirClinicalImpressionFindingListEnumerator.Create(list : TFhirClinicalImpressionFindingList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClinicalImpressionFindingListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClinicalImpressionFindingListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClinicalImpressionFindingListEnumerator.GetCurrent : TFhirClinicalImpressionFinding;
begin
  Result := FList[FIndex];
end;


{ TFhirClinicalImpressionFindingList }
procedure TFhirClinicalImpressionFindingList.AddItem(value: TFhirClinicalImpressionFinding);
begin
  assert(value.ClassName = 'TFhirClinicalImpressionFinding', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClinicalImpressionFinding');
  add(value);
end;


function TFhirClinicalImpressionFindingList.Append: TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionFindingList.ClearItems;
begin
  Clear;
end;

function TFhirClinicalImpressionFindingList.GetEnumerator : TFhirClinicalImpressionFindingListEnumerator;
begin
  result := TFhirClinicalImpressionFindingListEnumerator.Create(self.link);
end;

function TFhirClinicalImpressionFindingList.Clone: TFhirClinicalImpressionFindingList;
begin
  result := TFhirClinicalImpressionFindingList(inherited Clone);
end;

function TFhirClinicalImpressionFindingList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClinicalImpressionFindingList.GetItemN(index: Integer): TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionFindingList.IndexOf(value: TFhirClinicalImpressionFinding): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClinicalImpressionFindingList.Insert(index: Integer): TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionFindingList.InsertItem(index: Integer; value: TFhirClinicalImpressionFinding);
begin
  assert(value is TFhirClinicalImpressionFinding);
  Inherited Insert(index, value);
end;

function TFhirClinicalImpressionFindingList.Item(index: Integer): TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionFindingList.Link: TFhirClinicalImpressionFindingList;
begin
  result := TFhirClinicalImpressionFindingList(inherited Link);
end;

procedure TFhirClinicalImpressionFindingList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClinicalImpressionFindingList.SetItemByIndex(index: Integer; value: TFhirClinicalImpressionFinding);
begin
  assert(value is TFhirClinicalImpressionFinding);
  FhirClinicalImpressionFindings[index] := value;
end;

procedure TFhirClinicalImpressionFindingList.SetItemN(index: Integer; value: TFhirClinicalImpressionFinding);
begin
  assert(value is TFhirClinicalImpressionFinding);
  ObjectByIndex[index] := value;
end;

{ TFhirClinicalImpressionRuledOut }

constructor TFhirClinicalImpressionRuledOut.Create;
begin
  inherited;
end;

destructor TFhirClinicalImpressionRuledOut.Destroy;
begin
  FItem.free;
  FReason.free;
  inherited;
end;

procedure TFhirClinicalImpressionRuledOut.Assign(oSource : TAdvObject);
begin
  inherited;
  item := TFhirClinicalImpressionRuledOut(oSource).item.Clone;
  reasonElement := TFhirClinicalImpressionRuledOut(oSource).reasonElement.Clone;
end;

procedure TFhirClinicalImpressionRuledOut.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'item') Then
     list.add(FItem.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
end;

procedure TFhirClinicalImpressionRuledOut.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'item', 'CodeableConcept', FItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'string', FReason.Link));{2}
end;

procedure TFhirClinicalImpressionRuledOut.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'item') then Item := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'reason') then ReasonElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirClinicalImpressionRuledOut.FhirType : string;
begin
  result := 'ruledOut';
end;

function TFhirClinicalImpressionRuledOut.Link : TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut(inherited Link);
end;

function TFhirClinicalImpressionRuledOut.Clone : TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut(inherited Clone);
end;

function TFhirClinicalImpressionRuledOut.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionRuledOut;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionRuledOut)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionRuledOut(other);
    result := compareDeep(itemElement, o.itemElement, true) and compareDeep(reasonElement, o.reasonElement, true);
  end;
end;

function TFhirClinicalImpressionRuledOut.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirClinicalImpressionRuledOut;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirClinicalImpressionRuledOut)) then
    result := false
  else
  begin
    o := TFhirClinicalImpressionRuledOut(other);
    result := compareValues(reasonElement, o.reasonElement, true);
  end;
end;

{ TFhirClinicalImpressionRuledOut }

Procedure TFhirClinicalImpressionRuledOut.SetItem(value : TFhirCodeableConcept);
begin
  FItem.free;
  FItem := value;
end;

Procedure TFhirClinicalImpressionRuledOut.SetReason(value : TFhirString);
begin
  FReason.free;
  FReason := value;
end;

Function TFhirClinicalImpressionRuledOut.GetReasonST : String;
begin
  if FReason = nil then
    result := ''
  else
    result := FReason.value;
end;

Procedure TFhirClinicalImpressionRuledOut.SetReasonST(value : String);
begin
  if value <> '' then
  begin
    if FReason = nil then
      FReason := TFhirString.create;
    FReason.value := value
  end
  else if FReason <> nil then
    FReason.value := '';
end;


{ TFhirClinicalImpressionRuledOutListEnumerator }

Constructor TFhirClinicalImpressionRuledOutListEnumerator.Create(list : TFhirClinicalImpressionRuledOutList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClinicalImpressionRuledOutListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClinicalImpressionRuledOutListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClinicalImpressionRuledOutListEnumerator.GetCurrent : TFhirClinicalImpressionRuledOut;
begin
  Result := FList[FIndex];
end;


{ TFhirClinicalImpressionRuledOutList }
procedure TFhirClinicalImpressionRuledOutList.AddItem(value: TFhirClinicalImpressionRuledOut);
begin
  assert(value.ClassName = 'TFhirClinicalImpressionRuledOut', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClinicalImpressionRuledOut');
  add(value);
end;


function TFhirClinicalImpressionRuledOutList.Append: TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionRuledOutList.ClearItems;
begin
  Clear;
end;

function TFhirClinicalImpressionRuledOutList.GetEnumerator : TFhirClinicalImpressionRuledOutListEnumerator;
begin
  result := TFhirClinicalImpressionRuledOutListEnumerator.Create(self.link);
end;

function TFhirClinicalImpressionRuledOutList.Clone: TFhirClinicalImpressionRuledOutList;
begin
  result := TFhirClinicalImpressionRuledOutList(inherited Clone);
end;

function TFhirClinicalImpressionRuledOutList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClinicalImpressionRuledOutList.GetItemN(index: Integer): TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionRuledOutList.IndexOf(value: TFhirClinicalImpressionRuledOut): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClinicalImpressionRuledOutList.Insert(index: Integer): TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionRuledOutList.InsertItem(index: Integer; value: TFhirClinicalImpressionRuledOut);
begin
  assert(value is TFhirClinicalImpressionRuledOut);
  Inherited Insert(index, value);
end;

function TFhirClinicalImpressionRuledOutList.Item(index: Integer): TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionRuledOutList.Link: TFhirClinicalImpressionRuledOutList;
begin
  result := TFhirClinicalImpressionRuledOutList(inherited Link);
end;

procedure TFhirClinicalImpressionRuledOutList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClinicalImpressionRuledOutList.SetItemByIndex(index: Integer; value: TFhirClinicalImpressionRuledOut);
begin
  assert(value is TFhirClinicalImpressionRuledOut);
  FhirClinicalImpressionRuledOuts[index] := value;
end;

procedure TFhirClinicalImpressionRuledOutList.SetItemN(index: Integer; value: TFhirClinicalImpressionRuledOut);
begin
  assert(value is TFhirClinicalImpressionRuledOut);
  ObjectByIndex[index] := value;
end;

{ TFhirClinicalImpression }

constructor TFhirClinicalImpression.Create;
begin
  inherited;
end;

destructor TFhirClinicalImpression.Destroy;
begin
  FPatient.free;
  FAssessor.free;
  FStatus.free;
  FDate.free;
  FDescription.free;
  FPrevious.free;
  FProblemList.Free;
  FTrigger.free;
  FInvestigationsList.Free;
  FProtocol.free;
  FSummary.free;
  FFindingList.Free;
  FResolvedList.Free;
  FRuledOutList.Free;
  FPrognosis.free;
  FPlanList.Free;
  FActionList.Free;
  inherited;
end;

function TFhirClinicalImpression.GetResourceType : TFhirResourceType;
begin
  result := frtClinicalImpression;
end;

procedure TFhirClinicalImpression.Assign(oSource : TAdvObject);
begin
  inherited;
  patient := TFhirClinicalImpression(oSource).patient.Clone;
  assessor := TFhirClinicalImpression(oSource).assessor.Clone;
  FStatus := TFhirClinicalImpression(oSource).FStatus.Link;
  dateElement := TFhirClinicalImpression(oSource).dateElement.Clone;
  descriptionElement := TFhirClinicalImpression(oSource).descriptionElement.Clone;
  previous := TFhirClinicalImpression(oSource).previous.Clone;
  if (TFhirClinicalImpression(oSource).FProblemList = nil) then
  begin
    FProblemList.free;
    FProblemList := nil;
  end
  else
  begin
    FProblemList := TFhirReferenceList{Resource}.Create;
    FProblemList.Assign(TFhirClinicalImpression(oSource).FProblemList);
  end;
  trigger := TFhirClinicalImpression(oSource).trigger.Clone;
  if (TFhirClinicalImpression(oSource).FInvestigationsList = nil) then
  begin
    FInvestigationsList.free;
    FInvestigationsList := nil;
  end
  else
  begin
    FInvestigationsList := TFhirClinicalImpressionInvestigationsList.Create;
    FInvestigationsList.Assign(TFhirClinicalImpression(oSource).FInvestigationsList);
  end;
  protocolElement := TFhirClinicalImpression(oSource).protocolElement.Clone;
  summaryElement := TFhirClinicalImpression(oSource).summaryElement.Clone;
  if (TFhirClinicalImpression(oSource).FFindingList = nil) then
  begin
    FFindingList.free;
    FFindingList := nil;
  end
  else
  begin
    FFindingList := TFhirClinicalImpressionFindingList.Create;
    FFindingList.Assign(TFhirClinicalImpression(oSource).FFindingList);
  end;
  if (TFhirClinicalImpression(oSource).FResolvedList = nil) then
  begin
    FResolvedList.free;
    FResolvedList := nil;
  end
  else
  begin
    FResolvedList := TFhirCodeableConceptList.Create;
    FResolvedList.Assign(TFhirClinicalImpression(oSource).FResolvedList);
  end;
  if (TFhirClinicalImpression(oSource).FRuledOutList = nil) then
  begin
    FRuledOutList.free;
    FRuledOutList := nil;
  end
  else
  begin
    FRuledOutList := TFhirClinicalImpressionRuledOutList.Create;
    FRuledOutList.Assign(TFhirClinicalImpression(oSource).FRuledOutList);
  end;
  prognosisElement := TFhirClinicalImpression(oSource).prognosisElement.Clone;
  if (TFhirClinicalImpression(oSource).FPlanList = nil) then
  begin
    FPlanList.free;
    FPlanList := nil;
  end
  else
  begin
    FPlanList := TFhirReferenceList{Resource}.Create;
    FPlanList.Assign(TFhirClinicalImpression(oSource).FPlanList);
  end;
  if (TFhirClinicalImpression(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirReferenceList{Resource}.Create;
    FActionList.Assign(TFhirClinicalImpression(oSource).FActionList);
  end;
end;

procedure TFhirClinicalImpression.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'assessor') Then
     list.add(FAssessor.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'previous') Then
     list.add(FPrevious.Link);
  if (child_name = 'problem') Then
    list.addAll(FProblemList);
  if (child_name = 'trigger[x]') Then
     list.add(FTrigger.Link);
  if (child_name = 'investigations') Then
    list.addAll(FInvestigationsList);
  if (child_name = 'protocol') Then
     list.add(FProtocol.Link);
  if (child_name = 'summary') Then
     list.add(FSummary.Link);
  if (child_name = 'finding') Then
    list.addAll(FFindingList);
  if (child_name = 'resolved') Then
    list.addAll(FResolvedList);
  if (child_name = 'ruledOut') Then
    list.addAll(FRuledOutList);
  if (child_name = 'prognosis') Then
     list.add(FPrognosis.Link);
  if (child_name = 'plan') Then
    list.addAll(FPlanList);
  if (child_name = 'action') Then
    list.addAll(FActionList);
end;

procedure TFhirClinicalImpression.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'assessor', 'Reference(Practitioner)', FAssessor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'previous', 'Reference(ClinicalImpression)', FPrevious.Link));{2}
  oList.add(TFHIRProperty.create(self, 'problem', 'Reference(Condition|AllergyIntolerance)', FProblemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'trigger[x]', 'CodeableConcept|Reference(Any)', FTrigger.Link));{2}
  oList.add(TFHIRProperty.create(self, 'investigations', '', FInvestigationsList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'protocol', 'uri', FProtocol.Link));{2}
  oList.add(TFHIRProperty.create(self, 'summary', 'string', FSummary.Link));{2}
  oList.add(TFHIRProperty.create(self, 'finding', '', FFindingList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'resolved', 'CodeableConcept', FResolvedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruledOut', '', FRuledOutList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'prognosis', 'string', FPrognosis.Link));{2}
  oList.add(TFHIRProperty.create(self, 'plan', 'Reference(CarePlan|Appointment|CommunicationRequest|DeviceUseRequest|DiagnosticOrder|MedicationOrder|NutritionOrder|Order|ProcedureRequest|ProcessRequest|ReferralRequest|SupplyRequest|VisionPrescription)', FPlanList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'action', 'Reference(ReferralRequest|ProcedureRequest|Procedure|MedicationOrder|DiagnosticOrder|NutritionOrder|SupplyRequest|Appointment)', FActionList.Link)){3};
end;

procedure TFhirClinicalImpression.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'assessor') then Assessor := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'previous') then Previous := propValue as TFhirReference{TFhirClinicalImpression}{4b}
  else if (propName = 'problem') then ProblemList.add(propValue as TFhirReference{Resource}){2}
  else if (propName.startsWith('trigger')) then Trigger := propValue as TFhirType{4}
  else if (propName = 'investigations') then InvestigationsList.add(propValue as TFhirClinicalImpressionInvestigations){2}
  else if (propName = 'protocol') then ProtocolElement := propValue as TFhirUri{5a}
  else if (propName = 'summary') then SummaryElement := propValue as TFhirString{5a}
  else if (propName = 'finding') then FindingList.add(propValue as TFhirClinicalImpressionFinding){2}
  else if (propName = 'resolved') then ResolvedList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'ruledOut') then RuledOutList.add(propValue as TFhirClinicalImpressionRuledOut){2}
  else if (propName = 'prognosis') then PrognosisElement := propValue as TFhirString{5a}
  else if (propName = 'plan') then PlanList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'action') then ActionList.add(propValue as TFhirReference{Resource}){2}
  else inherited;
end;

function TFhirClinicalImpression.FhirType : string;
begin
  result := 'ClinicalImpression';
end;

function TFhirClinicalImpression.Link : TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression(inherited Link);
end;

function TFhirClinicalImpression.Clone : TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression(inherited Clone);
end;

{ TFhirClinicalImpression }

Procedure TFhirClinicalImpression.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirClinicalImpression.SetAssessor(value : TFhirReference{TFhirPractitioner});
begin
  FAssessor.free;
  FAssessor := value;
end;

Procedure TFhirClinicalImpression.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirClinicalImpression.GetStatusST : TFhirClinicalImpressionStatus;
begin
  if FStatus = nil then
    result := TFhirClinicalImpressionStatus(0)
  else
    result := TFhirClinicalImpressionStatus(StringArrayIndexOfSensitive(CODES_TFhirClinicalImpressionStatus, FStatus.value));
end;

Procedure TFhirClinicalImpression.SetStatusST(value : TFhirClinicalImpressionStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirClinicalImpressionStatus[value]);
end;

Procedure TFhirClinicalImpression.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirClinicalImpression.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirClinicalImpression.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirClinicalImpression.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirClinicalImpression.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirClinicalImpression.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirClinicalImpression.SetPrevious(value : TFhirReference{TFhirClinicalImpression});
begin
  FPrevious.free;
  FPrevious := value;
end;

Function TFhirClinicalImpression.GetProblemList : TFhirReferenceList{Resource};
begin
  if FProblemList = nil then
    FProblemList := TFhirReferenceList{Resource}.Create;
  result := FProblemList;
end;

Function TFhirClinicalImpression.GetHasProblemList : boolean;
begin
  result := (FProblemList <> nil) and (FProblemList.count > 0);
end;

Procedure TFhirClinicalImpression.SetTrigger(value : TFhirType);
begin
  FTrigger.free;
  FTrigger := value;
end;

Function TFhirClinicalImpression.GetInvestigationsList : TFhirClinicalImpressionInvestigationsList;
begin
  if FInvestigationsList = nil then
    FInvestigationsList := TFhirClinicalImpressionInvestigationsList.Create;
  result := FInvestigationsList;
end;

Function TFhirClinicalImpression.GetHasInvestigationsList : boolean;
begin
  result := (FInvestigationsList <> nil) and (FInvestigationsList.count > 0);
end;

Procedure TFhirClinicalImpression.SetProtocol(value : TFhirUri);
begin
  FProtocol.free;
  FProtocol := value;
end;

Function TFhirClinicalImpression.GetProtocolST : String;
begin
  if FProtocol = nil then
    result := ''
  else
    result := FProtocol.value;
end;

Procedure TFhirClinicalImpression.SetProtocolST(value : String);
begin
  if value <> '' then
  begin
    if FProtocol = nil then
      FProtocol := TFhirUri.create;
    FProtocol.value := value
  end
  else if FProtocol <> nil then
    FProtocol.value := '';
end;

Procedure TFhirClinicalImpression.SetSummary(value : TFhirString);
begin
  FSummary.free;
  FSummary := value;
end;

Function TFhirClinicalImpression.GetSummaryST : String;
begin
  if FSummary = nil then
    result := ''
  else
    result := FSummary.value;
end;

Procedure TFhirClinicalImpression.SetSummaryST(value : String);
begin
  if value <> '' then
  begin
    if FSummary = nil then
      FSummary := TFhirString.create;
    FSummary.value := value
  end
  else if FSummary <> nil then
    FSummary.value := '';
end;

Function TFhirClinicalImpression.GetFindingList : TFhirClinicalImpressionFindingList;
begin
  if FFindingList = nil then
    FFindingList := TFhirClinicalImpressionFindingList.Create;
  result := FFindingList;
end;

Function TFhirClinicalImpression.GetHasFindingList : boolean;
begin
  result := (FFindingList <> nil) and (FFindingList.count > 0);
end;

Function TFhirClinicalImpression.GetResolvedList : TFhirCodeableConceptList;
begin
  if FResolvedList = nil then
    FResolvedList := TFhirCodeableConceptList.Create;
  result := FResolvedList;
end;

Function TFhirClinicalImpression.GetHasResolvedList : boolean;
begin
  result := (FResolvedList <> nil) and (FResolvedList.count > 0);
end;

Function TFhirClinicalImpression.GetRuledOutList : TFhirClinicalImpressionRuledOutList;
begin
  if FRuledOutList = nil then
    FRuledOutList := TFhirClinicalImpressionRuledOutList.Create;
  result := FRuledOutList;
end;

Function TFhirClinicalImpression.GetHasRuledOutList : boolean;
begin
  result := (FRuledOutList <> nil) and (FRuledOutList.count > 0);
end;

Procedure TFhirClinicalImpression.SetPrognosis(value : TFhirString);
begin
  FPrognosis.free;
  FPrognosis := value;
end;

Function TFhirClinicalImpression.GetPrognosisST : String;
begin
  if FPrognosis = nil then
    result := ''
  else
    result := FPrognosis.value;
end;

Procedure TFhirClinicalImpression.SetPrognosisST(value : String);
begin
  if value <> '' then
  begin
    if FPrognosis = nil then
      FPrognosis := TFhirString.create;
    FPrognosis.value := value
  end
  else if FPrognosis <> nil then
    FPrognosis.value := '';
end;

Function TFhirClinicalImpression.GetPlanList : TFhirReferenceList{Resource};
begin
  if FPlanList = nil then
    FPlanList := TFhirReferenceList{Resource}.Create;
  result := FPlanList;
end;

Function TFhirClinicalImpression.GetHasPlanList : boolean;
begin
  result := (FPlanList <> nil) and (FPlanList.count > 0);
end;

Function TFhirClinicalImpression.GetActionList : TFhirReferenceList{Resource};
begin
  if FActionList = nil then
    FActionList := TFhirReferenceList{Resource}.Create;
  result := FActionList;
end;

Function TFhirClinicalImpression.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;


{ TFhirClinicalImpressionListEnumerator }

Constructor TFhirClinicalImpressionListEnumerator.Create(list : TFhirClinicalImpressionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirClinicalImpressionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirClinicalImpressionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirClinicalImpressionListEnumerator.GetCurrent : TFhirClinicalImpression;
begin
  Result := FList[FIndex];
end;


{ TFhirClinicalImpressionList }
procedure TFhirClinicalImpressionList.AddItem(value: TFhirClinicalImpression);
begin
  assert(value.ClassName = 'TFhirClinicalImpression', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirClinicalImpression');
  add(value);
end;


function TFhirClinicalImpressionList.Append: TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionList.ClearItems;
begin
  Clear;
end;

function TFhirClinicalImpressionList.GetEnumerator : TFhirClinicalImpressionListEnumerator;
begin
  result := TFhirClinicalImpressionListEnumerator.Create(self.link);
end;

function TFhirClinicalImpressionList.Clone: TFhirClinicalImpressionList;
begin
  result := TFhirClinicalImpressionList(inherited Clone);
end;

function TFhirClinicalImpressionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirClinicalImpressionList.GetItemN(index: Integer): TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionList.IndexOf(value: TFhirClinicalImpression): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirClinicalImpressionList.Insert(index: Integer): TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirClinicalImpressionList.InsertItem(index: Integer; value: TFhirClinicalImpression);
begin
  assert(value is TFhirClinicalImpression);
  Inherited Insert(index, value);
end;

function TFhirClinicalImpressionList.Item(index: Integer): TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression(ObjectByIndex[index]);
end;

function TFhirClinicalImpressionList.Link: TFhirClinicalImpressionList;
begin
  result := TFhirClinicalImpressionList(inherited Link);
end;

procedure TFhirClinicalImpressionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirClinicalImpressionList.SetItemByIndex(index: Integer; value: TFhirClinicalImpression);
begin
  assert(value is TFhirClinicalImpression);
  FhirClinicalImpressions[index] := value;
end;

procedure TFhirClinicalImpressionList.SetItemN(index: Integer; value: TFhirClinicalImpression);
begin
  assert(value is TFhirClinicalImpression);
  ObjectByIndex[index] := value;
end;

{ TFhirCommunicationPayload }

constructor TFhirCommunicationPayload.Create;
begin
  inherited;
end;

destructor TFhirCommunicationPayload.Destroy;
begin
  FContent.free;
  inherited;
end;

procedure TFhirCommunicationPayload.Assign(oSource : TAdvObject);
begin
  inherited;
  content := TFhirCommunicationPayload(oSource).content.Clone;
end;

procedure TFhirCommunicationPayload.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'content[x]') Then
     list.add(FContent.Link);
end;

procedure TFhirCommunicationPayload.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'content[x]', 'string|Attachment|Reference(Any)', FContent.Link));{2}
end;

procedure TFhirCommunicationPayload.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('content')) then Content := propValue as TFhirType{4}
  else inherited;
end;

function TFhirCommunicationPayload.FhirType : string;
begin
  result := 'payload';
end;

function TFhirCommunicationPayload.Link : TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload(inherited Link);
end;

function TFhirCommunicationPayload.Clone : TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload(inherited Clone);
end;

function TFhirCommunicationPayload.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCommunicationPayload;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCommunicationPayload)) then
    result := false
  else
  begin
    o := TFhirCommunicationPayload(other);
    result := compareDeep(contentElement, o.contentElement, true);
  end;
end;

function TFhirCommunicationPayload.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCommunicationPayload;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCommunicationPayload)) then
    result := false
  else
  begin
    o := TFhirCommunicationPayload(other);
    result := true;
  end;
end;

{ TFhirCommunicationPayload }

Procedure TFhirCommunicationPayload.SetContent(value : TFhirType);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirCommunicationPayloadListEnumerator }

Constructor TFhirCommunicationPayloadListEnumerator.Create(list : TFhirCommunicationPayloadList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCommunicationPayloadListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCommunicationPayloadListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCommunicationPayloadListEnumerator.GetCurrent : TFhirCommunicationPayload;
begin
  Result := FList[FIndex];
end;


{ TFhirCommunicationPayloadList }
procedure TFhirCommunicationPayloadList.AddItem(value: TFhirCommunicationPayload);
begin
  assert(value.ClassName = 'TFhirCommunicationPayload', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCommunicationPayload');
  add(value);
end;


function TFhirCommunicationPayloadList.Append: TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationPayloadList.ClearItems;
begin
  Clear;
end;

function TFhirCommunicationPayloadList.GetEnumerator : TFhirCommunicationPayloadListEnumerator;
begin
  result := TFhirCommunicationPayloadListEnumerator.Create(self.link);
end;

function TFhirCommunicationPayloadList.Clone: TFhirCommunicationPayloadList;
begin
  result := TFhirCommunicationPayloadList(inherited Clone);
end;

function TFhirCommunicationPayloadList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCommunicationPayloadList.GetItemN(index: Integer): TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload(ObjectByIndex[index]);
end;

function TFhirCommunicationPayloadList.IndexOf(value: TFhirCommunicationPayload): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCommunicationPayloadList.Insert(index: Integer): TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationPayloadList.InsertItem(index: Integer; value: TFhirCommunicationPayload);
begin
  assert(value is TFhirCommunicationPayload);
  Inherited Insert(index, value);
end;

function TFhirCommunicationPayloadList.Item(index: Integer): TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload(ObjectByIndex[index]);
end;

function TFhirCommunicationPayloadList.Link: TFhirCommunicationPayloadList;
begin
  result := TFhirCommunicationPayloadList(inherited Link);
end;

procedure TFhirCommunicationPayloadList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCommunicationPayloadList.SetItemByIndex(index: Integer; value: TFhirCommunicationPayload);
begin
  assert(value is TFhirCommunicationPayload);
  FhirCommunicationPayloads[index] := value;
end;

procedure TFhirCommunicationPayloadList.SetItemN(index: Integer; value: TFhirCommunicationPayload);
begin
  assert(value is TFhirCommunicationPayload);
  ObjectByIndex[index] := value;
end;

{ TFhirCommunication }

constructor TFhirCommunication.Create;
begin
  inherited;
end;

destructor TFhirCommunication.Destroy;
begin
  FIdentifierList.Free;
  FCategory.free;
  FSender.free;
  FRecipientList.Free;
  FPayloadList.Free;
  FMediumList.Free;
  FStatus.free;
  FEncounter.free;
  FSent.free;
  FReceived.free;
  FReasonList.Free;
  FSubject.free;
  FRequestDetail.free;
  inherited;
end;

function TFhirCommunication.GetResourceType : TFhirResourceType;
begin
  result := frtCommunication;
end;

procedure TFhirCommunication.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCommunication(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirCommunication(oSource).FIdentifierList);
  end;
  category := TFhirCommunication(oSource).category.Clone;
  sender := TFhirCommunication(oSource).sender.Clone;
  if (TFhirCommunication(oSource).FRecipientList = nil) then
  begin
    FRecipientList.free;
    FRecipientList := nil;
  end
  else
  begin
    FRecipientList := TFhirReferenceList{Resource}.Create;
    FRecipientList.Assign(TFhirCommunication(oSource).FRecipientList);
  end;
  if (TFhirCommunication(oSource).FPayloadList = nil) then
  begin
    FPayloadList.free;
    FPayloadList := nil;
  end
  else
  begin
    FPayloadList := TFhirCommunicationPayloadList.Create;
    FPayloadList.Assign(TFhirCommunication(oSource).FPayloadList);
  end;
  if (TFhirCommunication(oSource).FMediumList = nil) then
  begin
    FMediumList.free;
    FMediumList := nil;
  end
  else
  begin
    FMediumList := TFhirCodeableConceptList.Create;
    FMediumList.Assign(TFhirCommunication(oSource).FMediumList);
  end;
  FStatus := TFhirCommunication(oSource).FStatus.Link;
  encounter := TFhirCommunication(oSource).encounter.Clone;
  sentElement := TFhirCommunication(oSource).sentElement.Clone;
  receivedElement := TFhirCommunication(oSource).receivedElement.Clone;
  if (TFhirCommunication(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirCommunication(oSource).FReasonList);
  end;
  subject := TFhirCommunication(oSource).subject.Clone;
  requestDetail := TFhirCommunication(oSource).requestDetail.Clone;
end;

procedure TFhirCommunication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'sender') Then
     list.add(FSender.Link);
  if (child_name = 'recipient') Then
    list.addAll(FRecipientList);
  if (child_name = 'payload') Then
    list.addAll(FPayloadList);
  if (child_name = 'medium') Then
    list.addAll(FMediumList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'sent') Then
     list.add(FSent.Link);
  if (child_name = 'received') Then
     list.add(FReceived.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'requestDetail') Then
     list.add(FRequestDetail.Link);
end;

procedure TFhirCommunication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sender', 'Reference(Device|Organization|Patient|Practitioner|RelatedPerson)', FSender.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recipient', 'Reference(Device|Organization|Patient|Practitioner|RelatedPerson|Group)', FRecipientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'payload', '', FPayloadList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'medium', 'CodeableConcept', FMediumList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sent', 'dateTime', FSent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'received', 'dateTime', FReceived.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestDetail', 'Reference(CommunicationRequest)', FRequestDetail.Link));{2}
end;

procedure TFhirCommunication.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'sender') then Sender := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'recipient') then RecipientList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'payload') then PayloadList.add(propValue as TFhirCommunicationPayload){2}
  else if (propName = 'medium') then MediumList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'sent') then SentElement := propValue as TFhirDateTime{5a}
  else if (propName = 'received') then ReceivedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'requestDetail') then RequestDetail := propValue as TFhirReference{TFhirCommunicationRequest}{4b}
  else inherited;
end;

function TFhirCommunication.FhirType : string;
begin
  result := 'Communication';
end;

function TFhirCommunication.Link : TFhirCommunication;
begin
  result := TFhirCommunication(inherited Link);
end;

function TFhirCommunication.Clone : TFhirCommunication;
begin
  result := TFhirCommunication(inherited Clone);
end;

{ TFhirCommunication }

Function TFhirCommunication.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirCommunication.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirCommunication.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirCommunication.SetSender(value : TFhirReference{Resource});
begin
  FSender.free;
  FSender := value;
end;

Function TFhirCommunication.GetRecipientList : TFhirReferenceList{Resource};
begin
  if FRecipientList = nil then
    FRecipientList := TFhirReferenceList{Resource}.Create;
  result := FRecipientList;
end;

Function TFhirCommunication.GetHasRecipientList : boolean;
begin
  result := (FRecipientList <> nil) and (FRecipientList.count > 0);
end;

Function TFhirCommunication.GetPayloadList : TFhirCommunicationPayloadList;
begin
  if FPayloadList = nil then
    FPayloadList := TFhirCommunicationPayloadList.Create;
  result := FPayloadList;
end;

Function TFhirCommunication.GetHasPayloadList : boolean;
begin
  result := (FPayloadList <> nil) and (FPayloadList.count > 0);
end;

Function TFhirCommunication.GetMediumList : TFhirCodeableConceptList;
begin
  if FMediumList = nil then
    FMediumList := TFhirCodeableConceptList.Create;
  result := FMediumList;
end;

Function TFhirCommunication.GetHasMediumList : boolean;
begin
  result := (FMediumList <> nil) and (FMediumList.count > 0);
end;

Procedure TFhirCommunication.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirCommunication.GetStatusST : TFhirCommunicationStatus;
begin
  if FStatus = nil then
    result := TFhirCommunicationStatus(0)
  else
    result := TFhirCommunicationStatus(StringArrayIndexOfSensitive(CODES_TFhirCommunicationStatus, FStatus.value));
end;

Procedure TFhirCommunication.SetStatusST(value : TFhirCommunicationStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirCommunicationStatus[value]);
end;

Procedure TFhirCommunication.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirCommunication.SetSent(value : TFhirDateTime);
begin
  FSent.free;
  FSent := value;
end;

Function TFhirCommunication.GetSentST : TDateAndTime;
begin
  if FSent = nil then
    result := nil
  else
    result := FSent.value;
end;

Procedure TFhirCommunication.SetSentST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FSent = nil then
      FSent := TFhirDateTime.create;
    FSent.value := value
  end
  else if FSent <> nil then
    FSent.value := nil;
end;

Procedure TFhirCommunication.SetReceived(value : TFhirDateTime);
begin
  FReceived.free;
  FReceived := value;
end;

Function TFhirCommunication.GetReceivedST : TDateAndTime;
begin
  if FReceived = nil then
    result := nil
  else
    result := FReceived.value;
end;

Procedure TFhirCommunication.SetReceivedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FReceived = nil then
      FReceived := TFhirDateTime.create;
    FReceived.value := value
  end
  else if FReceived <> nil then
    FReceived.value := nil;
end;

Function TFhirCommunication.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirCommunication.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Procedure TFhirCommunication.SetSubject(value : TFhirReference{TFhirPatient});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirCommunication.SetRequestDetail(value : TFhirReference{TFhirCommunicationRequest});
begin
  FRequestDetail.free;
  FRequestDetail := value;
end;


{ TFhirCommunicationListEnumerator }

Constructor TFhirCommunicationListEnumerator.Create(list : TFhirCommunicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCommunicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCommunicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCommunicationListEnumerator.GetCurrent : TFhirCommunication;
begin
  Result := FList[FIndex];
end;


{ TFhirCommunicationList }
procedure TFhirCommunicationList.AddItem(value: TFhirCommunication);
begin
  assert(value.ClassName = 'TFhirCommunication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCommunication');
  add(value);
end;


function TFhirCommunicationList.Append: TFhirCommunication;
begin
  result := TFhirCommunication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationList.ClearItems;
begin
  Clear;
end;

function TFhirCommunicationList.GetEnumerator : TFhirCommunicationListEnumerator;
begin
  result := TFhirCommunicationListEnumerator.Create(self.link);
end;

function TFhirCommunicationList.Clone: TFhirCommunicationList;
begin
  result := TFhirCommunicationList(inherited Clone);
end;

function TFhirCommunicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCommunicationList.GetItemN(index: Integer): TFhirCommunication;
begin
  result := TFhirCommunication(ObjectByIndex[index]);
end;

function TFhirCommunicationList.IndexOf(value: TFhirCommunication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCommunicationList.Insert(index: Integer): TFhirCommunication;
begin
  result := TFhirCommunication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationList.InsertItem(index: Integer; value: TFhirCommunication);
begin
  assert(value is TFhirCommunication);
  Inherited Insert(index, value);
end;

function TFhirCommunicationList.Item(index: Integer): TFhirCommunication;
begin
  result := TFhirCommunication(ObjectByIndex[index]);
end;

function TFhirCommunicationList.Link: TFhirCommunicationList;
begin
  result := TFhirCommunicationList(inherited Link);
end;

procedure TFhirCommunicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCommunicationList.SetItemByIndex(index: Integer; value: TFhirCommunication);
begin
  assert(value is TFhirCommunication);
  FhirCommunications[index] := value;
end;

procedure TFhirCommunicationList.SetItemN(index: Integer; value: TFhirCommunication);
begin
  assert(value is TFhirCommunication);
  ObjectByIndex[index] := value;
end;

{ TFhirCommunicationRequestPayload }

constructor TFhirCommunicationRequestPayload.Create;
begin
  inherited;
end;

destructor TFhirCommunicationRequestPayload.Destroy;
begin
  FContent.free;
  inherited;
end;

procedure TFhirCommunicationRequestPayload.Assign(oSource : TAdvObject);
begin
  inherited;
  content := TFhirCommunicationRequestPayload(oSource).content.Clone;
end;

procedure TFhirCommunicationRequestPayload.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'content[x]') Then
     list.add(FContent.Link);
end;

procedure TFhirCommunicationRequestPayload.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'content[x]', 'string|Attachment|Reference(Any)', FContent.Link));{2}
end;

procedure TFhirCommunicationRequestPayload.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('content')) then Content := propValue as TFhirType{4}
  else inherited;
end;

function TFhirCommunicationRequestPayload.FhirType : string;
begin
  result := 'payload';
end;

function TFhirCommunicationRequestPayload.Link : TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload(inherited Link);
end;

function TFhirCommunicationRequestPayload.Clone : TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload(inherited Clone);
end;

function TFhirCommunicationRequestPayload.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCommunicationRequestPayload;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCommunicationRequestPayload)) then
    result := false
  else
  begin
    o := TFhirCommunicationRequestPayload(other);
    result := compareDeep(contentElement, o.contentElement, true);
  end;
end;

function TFhirCommunicationRequestPayload.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCommunicationRequestPayload;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCommunicationRequestPayload)) then
    result := false
  else
  begin
    o := TFhirCommunicationRequestPayload(other);
    result := true;
  end;
end;

{ TFhirCommunicationRequestPayload }

Procedure TFhirCommunicationRequestPayload.SetContent(value : TFhirType);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirCommunicationRequestPayloadListEnumerator }

Constructor TFhirCommunicationRequestPayloadListEnumerator.Create(list : TFhirCommunicationRequestPayloadList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCommunicationRequestPayloadListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCommunicationRequestPayloadListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCommunicationRequestPayloadListEnumerator.GetCurrent : TFhirCommunicationRequestPayload;
begin
  Result := FList[FIndex];
end;


{ TFhirCommunicationRequestPayloadList }
procedure TFhirCommunicationRequestPayloadList.AddItem(value: TFhirCommunicationRequestPayload);
begin
  assert(value.ClassName = 'TFhirCommunicationRequestPayload', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCommunicationRequestPayload');
  add(value);
end;


function TFhirCommunicationRequestPayloadList.Append: TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationRequestPayloadList.ClearItems;
begin
  Clear;
end;

function TFhirCommunicationRequestPayloadList.GetEnumerator : TFhirCommunicationRequestPayloadListEnumerator;
begin
  result := TFhirCommunicationRequestPayloadListEnumerator.Create(self.link);
end;

function TFhirCommunicationRequestPayloadList.Clone: TFhirCommunicationRequestPayloadList;
begin
  result := TFhirCommunicationRequestPayloadList(inherited Clone);
end;

function TFhirCommunicationRequestPayloadList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCommunicationRequestPayloadList.GetItemN(index: Integer): TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload(ObjectByIndex[index]);
end;

function TFhirCommunicationRequestPayloadList.IndexOf(value: TFhirCommunicationRequestPayload): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCommunicationRequestPayloadList.Insert(index: Integer): TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationRequestPayloadList.InsertItem(index: Integer; value: TFhirCommunicationRequestPayload);
begin
  assert(value is TFhirCommunicationRequestPayload);
  Inherited Insert(index, value);
end;

function TFhirCommunicationRequestPayloadList.Item(index: Integer): TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload(ObjectByIndex[index]);
end;

function TFhirCommunicationRequestPayloadList.Link: TFhirCommunicationRequestPayloadList;
begin
  result := TFhirCommunicationRequestPayloadList(inherited Link);
end;

procedure TFhirCommunicationRequestPayloadList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCommunicationRequestPayloadList.SetItemByIndex(index: Integer; value: TFhirCommunicationRequestPayload);
begin
  assert(value is TFhirCommunicationRequestPayload);
  FhirCommunicationRequestPayloads[index] := value;
end;

procedure TFhirCommunicationRequestPayloadList.SetItemN(index: Integer; value: TFhirCommunicationRequestPayload);
begin
  assert(value is TFhirCommunicationRequestPayload);
  ObjectByIndex[index] := value;
end;

{ TFhirCommunicationRequest }

constructor TFhirCommunicationRequest.Create;
begin
  inherited;
end;

destructor TFhirCommunicationRequest.Destroy;
begin
  FIdentifierList.Free;
  FCategory.free;
  FSender.free;
  FRecipientList.Free;
  FPayloadList.Free;
  FMediumList.Free;
  FRequester.free;
  FStatus.free;
  FEncounter.free;
  FScheduled.free;
  FReasonList.Free;
  FRequestedOn.free;
  FSubject.free;
  FPriority.free;
  inherited;
end;

function TFhirCommunicationRequest.GetResourceType : TFhirResourceType;
begin
  result := frtCommunicationRequest;
end;

procedure TFhirCommunicationRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCommunicationRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirCommunicationRequest(oSource).FIdentifierList);
  end;
  category := TFhirCommunicationRequest(oSource).category.Clone;
  sender := TFhirCommunicationRequest(oSource).sender.Clone;
  if (TFhirCommunicationRequest(oSource).FRecipientList = nil) then
  begin
    FRecipientList.free;
    FRecipientList := nil;
  end
  else
  begin
    FRecipientList := TFhirReferenceList{Resource}.Create;
    FRecipientList.Assign(TFhirCommunicationRequest(oSource).FRecipientList);
  end;
  if (TFhirCommunicationRequest(oSource).FPayloadList = nil) then
  begin
    FPayloadList.free;
    FPayloadList := nil;
  end
  else
  begin
    FPayloadList := TFhirCommunicationRequestPayloadList.Create;
    FPayloadList.Assign(TFhirCommunicationRequest(oSource).FPayloadList);
  end;
  if (TFhirCommunicationRequest(oSource).FMediumList = nil) then
  begin
    FMediumList.free;
    FMediumList := nil;
  end
  else
  begin
    FMediumList := TFhirCodeableConceptList.Create;
    FMediumList.Assign(TFhirCommunicationRequest(oSource).FMediumList);
  end;
  requester := TFhirCommunicationRequest(oSource).requester.Clone;
  FStatus := TFhirCommunicationRequest(oSource).FStatus.Link;
  encounter := TFhirCommunicationRequest(oSource).encounter.Clone;
  scheduled := TFhirCommunicationRequest(oSource).scheduled.Clone;
  if (TFhirCommunicationRequest(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirCommunicationRequest(oSource).FReasonList);
  end;
  requestedOnElement := TFhirCommunicationRequest(oSource).requestedOnElement.Clone;
  subject := TFhirCommunicationRequest(oSource).subject.Clone;
  priority := TFhirCommunicationRequest(oSource).priority.Clone;
end;

procedure TFhirCommunicationRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'sender') Then
     list.add(FSender.Link);
  if (child_name = 'recipient') Then
    list.addAll(FRecipientList);
  if (child_name = 'payload') Then
    list.addAll(FPayloadList);
  if (child_name = 'medium') Then
    list.addAll(FMediumList);
  if (child_name = 'requester') Then
     list.add(FRequester.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'scheduled[x]') Then
     list.add(FScheduled.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'requestedOn') Then
     list.add(FRequestedOn.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
end;

procedure TFhirCommunicationRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sender', 'Reference(Device|Organization|Patient|Practitioner|RelatedPerson)', FSender.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recipient', 'Reference(Device|Organization|Patient|Practitioner|RelatedPerson)', FRecipientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'payload', '', FPayloadList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'medium', 'CodeableConcept', FMediumList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'requester', 'Reference(Practitioner|Patient|RelatedPerson)', FRequester.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'scheduled[x]', 'dateTime|Period', FScheduled.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'requestedOn', 'dateTime', FRequestedOn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'CodeableConcept', FPriority.Link));{2}
end;

procedure TFhirCommunicationRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'sender') then Sender := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'recipient') then RecipientList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'payload') then PayloadList.add(propValue as TFhirCommunicationRequestPayload){2}
  else if (propName = 'medium') then MediumList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'requester') then Requester := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName.startsWith('scheduled')) then Scheduled := propValue as TFhirType{4}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'requestedOn') then RequestedOnElement := propValue as TFhirDateTime{5a}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'priority') then Priority := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirCommunicationRequest.FhirType : string;
begin
  result := 'CommunicationRequest';
end;

function TFhirCommunicationRequest.Link : TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest(inherited Link);
end;

function TFhirCommunicationRequest.Clone : TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest(inherited Clone);
end;

{ TFhirCommunicationRequest }

Function TFhirCommunicationRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirCommunicationRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirCommunicationRequest.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirCommunicationRequest.SetSender(value : TFhirReference{Resource});
begin
  FSender.free;
  FSender := value;
end;

Function TFhirCommunicationRequest.GetRecipientList : TFhirReferenceList{Resource};
begin
  if FRecipientList = nil then
    FRecipientList := TFhirReferenceList{Resource}.Create;
  result := FRecipientList;
end;

Function TFhirCommunicationRequest.GetHasRecipientList : boolean;
begin
  result := (FRecipientList <> nil) and (FRecipientList.count > 0);
end;

Function TFhirCommunicationRequest.GetPayloadList : TFhirCommunicationRequestPayloadList;
begin
  if FPayloadList = nil then
    FPayloadList := TFhirCommunicationRequestPayloadList.Create;
  result := FPayloadList;
end;

Function TFhirCommunicationRequest.GetHasPayloadList : boolean;
begin
  result := (FPayloadList <> nil) and (FPayloadList.count > 0);
end;

Function TFhirCommunicationRequest.GetMediumList : TFhirCodeableConceptList;
begin
  if FMediumList = nil then
    FMediumList := TFhirCodeableConceptList.Create;
  result := FMediumList;
end;

Function TFhirCommunicationRequest.GetHasMediumList : boolean;
begin
  result := (FMediumList <> nil) and (FMediumList.count > 0);
end;

Procedure TFhirCommunicationRequest.SetRequester(value : TFhirReference{Resource});
begin
  FRequester.free;
  FRequester := value;
end;

Procedure TFhirCommunicationRequest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirCommunicationRequest.GetStatusST : TFhirCommunicationRequestStatus;
begin
  if FStatus = nil then
    result := TFhirCommunicationRequestStatus(0)
  else
    result := TFhirCommunicationRequestStatus(StringArrayIndexOfSensitive(CODES_TFhirCommunicationRequestStatus, FStatus.value));
end;

Procedure TFhirCommunicationRequest.SetStatusST(value : TFhirCommunicationRequestStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirCommunicationRequestStatus[value]);
end;

Procedure TFhirCommunicationRequest.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirCommunicationRequest.SetScheduled(value : TFhirType);
begin
  FScheduled.free;
  FScheduled := value;
end;

Function TFhirCommunicationRequest.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirCommunicationRequest.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Procedure TFhirCommunicationRequest.SetRequestedOn(value : TFhirDateTime);
begin
  FRequestedOn.free;
  FRequestedOn := value;
end;

Function TFhirCommunicationRequest.GetRequestedOnST : TDateAndTime;
begin
  if FRequestedOn = nil then
    result := nil
  else
    result := FRequestedOn.value;
end;

Procedure TFhirCommunicationRequest.SetRequestedOnST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FRequestedOn = nil then
      FRequestedOn := TFhirDateTime.create;
    FRequestedOn.value := value
  end
  else if FRequestedOn <> nil then
    FRequestedOn.value := nil;
end;

Procedure TFhirCommunicationRequest.SetSubject(value : TFhirReference{TFhirPatient});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirCommunicationRequest.SetPriority(value : TFhirCodeableConcept);
begin
  FPriority.free;
  FPriority := value;
end;


{ TFhirCommunicationRequestListEnumerator }

Constructor TFhirCommunicationRequestListEnumerator.Create(list : TFhirCommunicationRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCommunicationRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCommunicationRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCommunicationRequestListEnumerator.GetCurrent : TFhirCommunicationRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirCommunicationRequestList }
procedure TFhirCommunicationRequestList.AddItem(value: TFhirCommunicationRequest);
begin
  assert(value.ClassName = 'TFhirCommunicationRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCommunicationRequest');
  add(value);
end;


function TFhirCommunicationRequestList.Append: TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationRequestList.ClearItems;
begin
  Clear;
end;

function TFhirCommunicationRequestList.GetEnumerator : TFhirCommunicationRequestListEnumerator;
begin
  result := TFhirCommunicationRequestListEnumerator.Create(self.link);
end;

function TFhirCommunicationRequestList.Clone: TFhirCommunicationRequestList;
begin
  result := TFhirCommunicationRequestList(inherited Clone);
end;

function TFhirCommunicationRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCommunicationRequestList.GetItemN(index: Integer): TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest(ObjectByIndex[index]);
end;

function TFhirCommunicationRequestList.IndexOf(value: TFhirCommunicationRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCommunicationRequestList.Insert(index: Integer): TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCommunicationRequestList.InsertItem(index: Integer; value: TFhirCommunicationRequest);
begin
  assert(value is TFhirCommunicationRequest);
  Inherited Insert(index, value);
end;

function TFhirCommunicationRequestList.Item(index: Integer): TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest(ObjectByIndex[index]);
end;

function TFhirCommunicationRequestList.Link: TFhirCommunicationRequestList;
begin
  result := TFhirCommunicationRequestList(inherited Link);
end;

procedure TFhirCommunicationRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCommunicationRequestList.SetItemByIndex(index: Integer; value: TFhirCommunicationRequest);
begin
  assert(value is TFhirCommunicationRequest);
  FhirCommunicationRequests[index] := value;
end;

procedure TFhirCommunicationRequestList.SetItemN(index: Integer; value: TFhirCommunicationRequest);
begin
  assert(value is TFhirCommunicationRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirCompositionAttester }

constructor TFhirCompositionAttester.Create;
begin
  inherited;
end;

destructor TFhirCompositionAttester.Destroy;
begin
  FMode.Free;
  FTime.free;
  FParty.free;
  inherited;
end;

procedure TFhirCompositionAttester.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCompositionAttester(oSource).FMode = nil) then
  begin
    FMode.free;
    FMode := nil;
  end
  else
  begin
    FMode := TFHIREnumList.Create;
    FMode.Assign(TFhirCompositionAttester(oSource).FMode);
  end;
  timeElement := TFhirCompositionAttester(oSource).timeElement.Clone;
  party := TFhirCompositionAttester(oSource).party.Clone;
end;

procedure TFhirCompositionAttester.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'mode') Then
     list.addAll(FMode);
  if (child_name = 'time') Then
     list.add(FTime.Link);
  if (child_name = 'party') Then
     list.add(FParty.Link);
end;

procedure TFhirCompositionAttester.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link)){3};
  oList.add(TFHIRProperty.create(self, 'time', 'dateTime', FTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'party', 'Reference(Patient|Practitioner|Organization)', FParty.Link));{2}
end;

procedure TFhirCompositionAttester.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'mode') then FMode.add(propValue as TFHIREnum) {1}
  else if (propName = 'time') then TimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'party') then Party := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirCompositionAttester.FhirType : string;
begin
  result := 'attester';
end;

function TFhirCompositionAttester.Link : TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester(inherited Link);
end;

function TFhirCompositionAttester.Clone : TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester(inherited Clone);
end;

function TFhirCompositionAttester.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionAttester;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCompositionAttester)) then
    result := false
  else
  begin
    o := TFhirCompositionAttester(other);
    result := compareDeep(modeList, o.modeList, true) and compareDeep(timeElement, o.timeElement, true) and 
      compareDeep(partyElement, o.partyElement, true);
  end;
end;

function TFhirCompositionAttester.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionAttester;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCompositionAttester)) then
    result := false
  else
  begin
    o := TFhirCompositionAttester(other);
    result := compareValues(modeList, o.modeList, true) and compareValues(timeElement, o.timeElement, true);
  end;
end;

{ TFhirCompositionAttester }

Function TFhirCompositionAttester.GetMode : TFhirEnumList;
begin
  if FMode = nil then
    FMode := TFHIREnumList.Create;
  result := FMode;
end;

Function TFhirCompositionAttester.GetHasMode : boolean;
begin
  result := (FMode <> nil) and (FMode.count > 0);
end;

Function TFhirCompositionAttester.GetModeST : TFhirCompositionAttestationModeList;
  var i : integer;
begin
  result := [];
  if Fmode <> nil then
    for i := 0 to Fmode.count - 1 do
      result := result + [TFhirCompositionAttestationMode(StringArrayIndexOfSensitive(CODES_TFhirCompositionAttestationMode, Fmode[i].value))];
end;

Procedure TFhirCompositionAttester.SetModeST(value : TFhirCompositionAttestationModeList);
var a : TFhirCompositionAttestationMode;
begin
  Fmode.clear;
  for a := low(TFhirCompositionAttestationMode) to high(TFhirCompositionAttestationMode) do
    if a in value then
      begin
         if Fmode = nil then
           Fmode := TFhirEnumList.create;
         Fmode.add(TFhirEnum.create(CODES_TFhirCompositionAttestationMode[a]));
      end;
end;

Procedure TFhirCompositionAttester.SetTime(value : TFhirDateTime);
begin
  FTime.free;
  FTime := value;
end;

Function TFhirCompositionAttester.GetTimeST : TDateAndTime;
begin
  if FTime = nil then
    result := nil
  else
    result := FTime.value;
end;

Procedure TFhirCompositionAttester.SetTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FTime = nil then
      FTime := TFhirDateTime.create;
    FTime.value := value
  end
  else if FTime <> nil then
    FTime.value := nil;
end;

Procedure TFhirCompositionAttester.SetParty(value : TFhirReference{Resource});
begin
  FParty.free;
  FParty := value;
end;


{ TFhirCompositionAttesterListEnumerator }

Constructor TFhirCompositionAttesterListEnumerator.Create(list : TFhirCompositionAttesterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCompositionAttesterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCompositionAttesterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCompositionAttesterListEnumerator.GetCurrent : TFhirCompositionAttester;
begin
  Result := FList[FIndex];
end;


{ TFhirCompositionAttesterList }
procedure TFhirCompositionAttesterList.AddItem(value: TFhirCompositionAttester);
begin
  assert(value.ClassName = 'TFhirCompositionAttester', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCompositionAttester');
  add(value);
end;


function TFhirCompositionAttesterList.Append: TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionAttesterList.ClearItems;
begin
  Clear;
end;

function TFhirCompositionAttesterList.GetEnumerator : TFhirCompositionAttesterListEnumerator;
begin
  result := TFhirCompositionAttesterListEnumerator.Create(self.link);
end;

function TFhirCompositionAttesterList.Clone: TFhirCompositionAttesterList;
begin
  result := TFhirCompositionAttesterList(inherited Clone);
end;

function TFhirCompositionAttesterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCompositionAttesterList.GetItemN(index: Integer): TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester(ObjectByIndex[index]);
end;

function TFhirCompositionAttesterList.IndexOf(value: TFhirCompositionAttester): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCompositionAttesterList.Insert(index: Integer): TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionAttesterList.InsertItem(index: Integer; value: TFhirCompositionAttester);
begin
  assert(value is TFhirCompositionAttester);
  Inherited Insert(index, value);
end;

function TFhirCompositionAttesterList.Item(index: Integer): TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester(ObjectByIndex[index]);
end;

function TFhirCompositionAttesterList.Link: TFhirCompositionAttesterList;
begin
  result := TFhirCompositionAttesterList(inherited Link);
end;

procedure TFhirCompositionAttesterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCompositionAttesterList.SetItemByIndex(index: Integer; value: TFhirCompositionAttester);
begin
  assert(value is TFhirCompositionAttester);
  FhirCompositionAttesters[index] := value;
end;

procedure TFhirCompositionAttesterList.SetItemN(index: Integer; value: TFhirCompositionAttester);
begin
  assert(value is TFhirCompositionAttester);
  ObjectByIndex[index] := value;
end;

{ TFhirCompositionEvent }

constructor TFhirCompositionEvent.Create;
begin
  inherited;
end;

destructor TFhirCompositionEvent.Destroy;
begin
  FCodeList.Free;
  FPeriod.free;
  FDetailList.Free;
  inherited;
end;

procedure TFhirCompositionEvent.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCompositionEvent(oSource).FCodeList = nil) then
  begin
    FCodeList.free;
    FCodeList := nil;
  end
  else
  begin
    FCodeList := TFhirCodeableConceptList.Create;
    FCodeList.Assign(TFhirCompositionEvent(oSource).FCodeList);
  end;
  period := TFhirCompositionEvent(oSource).period.Clone;
  if (TFhirCompositionEvent(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
    FDetailList.Assign(TFhirCompositionEvent(oSource).FDetailList);
  end;
end;

procedure TFhirCompositionEvent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
    list.addAll(FCodeList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirCompositionEvent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCodeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', 'Reference(Any)', FDetailList.Link)){3};
end;

procedure TFhirCompositionEvent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then CodeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirReference{TFhirReference}){2}
  else inherited;
end;

function TFhirCompositionEvent.FhirType : string;
begin
  result := 'event';
end;

function TFhirCompositionEvent.Link : TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent(inherited Link);
end;

function TFhirCompositionEvent.Clone : TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent(inherited Clone);
end;

function TFhirCompositionEvent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionEvent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCompositionEvent)) then
    result := false
  else
  begin
    o := TFhirCompositionEvent(other);
    result := compareDeep(codeList, o.codeList, true) and compareDeep(periodElement, o.periodElement, true) and 
      compareDeep(detailList, o.detailList, true);
  end;
end;

function TFhirCompositionEvent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionEvent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCompositionEvent)) then
    result := false
  else
  begin
    o := TFhirCompositionEvent(other);
    result := true;
  end;
end;

{ TFhirCompositionEvent }

Function TFhirCompositionEvent.GetCodeList : TFhirCodeableConceptList;
begin
  if FCodeList = nil then
    FCodeList := TFhirCodeableConceptList.Create;
  result := FCodeList;
end;

Function TFhirCompositionEvent.GetHasCodeList : boolean;
begin
  result := (FCodeList <> nil) and (FCodeList.count > 0);
end;

Procedure TFhirCompositionEvent.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Function TFhirCompositionEvent.GetDetailList : TFhirReferenceList{TFhirReference};
begin
  if FDetailList = nil then
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
  result := FDetailList;
end;

Function TFhirCompositionEvent.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirCompositionEventListEnumerator }

Constructor TFhirCompositionEventListEnumerator.Create(list : TFhirCompositionEventList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCompositionEventListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCompositionEventListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCompositionEventListEnumerator.GetCurrent : TFhirCompositionEvent;
begin
  Result := FList[FIndex];
end;


{ TFhirCompositionEventList }
procedure TFhirCompositionEventList.AddItem(value: TFhirCompositionEvent);
begin
  assert(value.ClassName = 'TFhirCompositionEvent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCompositionEvent');
  add(value);
end;


function TFhirCompositionEventList.Append: TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionEventList.ClearItems;
begin
  Clear;
end;

function TFhirCompositionEventList.GetEnumerator : TFhirCompositionEventListEnumerator;
begin
  result := TFhirCompositionEventListEnumerator.Create(self.link);
end;

function TFhirCompositionEventList.Clone: TFhirCompositionEventList;
begin
  result := TFhirCompositionEventList(inherited Clone);
end;

function TFhirCompositionEventList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCompositionEventList.GetItemN(index: Integer): TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent(ObjectByIndex[index]);
end;

function TFhirCompositionEventList.IndexOf(value: TFhirCompositionEvent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCompositionEventList.Insert(index: Integer): TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionEventList.InsertItem(index: Integer; value: TFhirCompositionEvent);
begin
  assert(value is TFhirCompositionEvent);
  Inherited Insert(index, value);
end;

function TFhirCompositionEventList.Item(index: Integer): TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent(ObjectByIndex[index]);
end;

function TFhirCompositionEventList.Link: TFhirCompositionEventList;
begin
  result := TFhirCompositionEventList(inherited Link);
end;

procedure TFhirCompositionEventList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCompositionEventList.SetItemByIndex(index: Integer; value: TFhirCompositionEvent);
begin
  assert(value is TFhirCompositionEvent);
  FhirCompositionEvents[index] := value;
end;

procedure TFhirCompositionEventList.SetItemN(index: Integer; value: TFhirCompositionEvent);
begin
  assert(value is TFhirCompositionEvent);
  ObjectByIndex[index] := value;
end;

{ TFhirCompositionSection }

constructor TFhirCompositionSection.Create;
begin
  inherited;
end;

destructor TFhirCompositionSection.Destroy;
begin
  FTitle.free;
  FCode.free;
  FText.free;
  FMode.free;
  FOrderedBy.free;
  FEntryList.Free;
  FEmptyReason.free;
  FSectionList.Free;
  inherited;
end;

procedure TFhirCompositionSection.Assign(oSource : TAdvObject);
begin
  inherited;
  titleElement := TFhirCompositionSection(oSource).titleElement.Clone;
  code := TFhirCompositionSection(oSource).code.Clone;
  text := TFhirCompositionSection(oSource).text.Clone;
  FMode := TFhirCompositionSection(oSource).FMode.Link;
  orderedBy := TFhirCompositionSection(oSource).orderedBy.Clone;
  if (TFhirCompositionSection(oSource).FEntryList = nil) then
  begin
    FEntryList.free;
    FEntryList := nil;
  end
  else
  begin
    FEntryList := TFhirReferenceList{TFhirReference}.Create;
    FEntryList.Assign(TFhirCompositionSection(oSource).FEntryList);
  end;
  emptyReason := TFhirCompositionSection(oSource).emptyReason.Clone;
  if (TFhirCompositionSection(oSource).FSectionList = nil) then
  begin
    FSectionList.free;
    FSectionList := nil;
  end
  else
  begin
    FSectionList := TFhirCompositionSectionList.Create;
    FSectionList.Assign(TFhirCompositionSection(oSource).FSectionList);
  end;
end;

procedure TFhirCompositionSection.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'orderedBy') Then
     list.add(FOrderedBy.Link);
  if (child_name = 'entry') Then
    list.addAll(FEntryList);
  if (child_name = 'emptyReason') Then
     list.add(FEmptyReason.Link);
  if (child_name = 'section') Then
    list.addAll(FSectionList);
end;

procedure TFhirCompositionSection.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'Narrative', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'orderedBy', 'CodeableConcept', FOrderedBy.Link));{2}
  oList.add(TFHIRProperty.create(self, 'entry', 'Reference(Any)', FEntryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'emptyReason', 'CodeableConcept', FEmptyReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'section', '@Composition.section', FSectionList.Link)){3};
end;

procedure TFhirCompositionSection.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'text') then Text := propValue as TFhirNarrative{4b}
  else if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'orderedBy') then OrderedBy := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'entry') then EntryList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'emptyReason') then EmptyReason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'section') then SectionList.add(propValue as TFhirCompositionSection){2}
  else inherited;
end;

function TFhirCompositionSection.FhirType : string;
begin
  result := 'section';
end;

function TFhirCompositionSection.Link : TFhirCompositionSection;
begin
  result := TFhirCompositionSection(inherited Link);
end;

function TFhirCompositionSection.Clone : TFhirCompositionSection;
begin
  result := TFhirCompositionSection(inherited Clone);
end;

function TFhirCompositionSection.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionSection;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirCompositionSection)) then
    result := false
  else
  begin
    o := TFhirCompositionSection(other);
    result := compareDeep(titleElement, o.titleElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(textElement, o.textElement, true) and compareDeep(modeElement, o.modeElement, true) and 
      compareDeep(orderedByElement, o.orderedByElement, true) and compareDeep(entryList, o.entryList, true) and 
      compareDeep(emptyReasonElement, o.emptyReasonElement, true) and compareDeep(sectionList, o.sectionList, true);
  end;
end;

function TFhirCompositionSection.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirCompositionSection;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirCompositionSection)) then
    result := false
  else
  begin
    o := TFhirCompositionSection(other);
    result := compareValues(titleElement, o.titleElement, true) and compareValues(modeElement, o.modeElement, true);
  end;
end;

{ TFhirCompositionSection }

Procedure TFhirCompositionSection.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirCompositionSection.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirCompositionSection.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Procedure TFhirCompositionSection.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirCompositionSection.SetText(value : TFhirNarrative);
begin
  FText.free;
  FText := value;
end;

Procedure TFhirCompositionSection.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirCompositionSection.GetModeST : TFhirListMode;
begin
  if FMode = nil then
    result := TFhirListMode(0)
  else
    result := TFhirListMode(StringArrayIndexOfSensitive(CODES_TFhirListMode, FMode.value));
end;

Procedure TFhirCompositionSection.SetModeST(value : TFhirListMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirListMode[value]);
end;

Procedure TFhirCompositionSection.SetOrderedBy(value : TFhirCodeableConcept);
begin
  FOrderedBy.free;
  FOrderedBy := value;
end;

Function TFhirCompositionSection.GetEntryList : TFhirReferenceList{TFhirReference};
begin
  if FEntryList = nil then
    FEntryList := TFhirReferenceList{TFhirReference}.Create;
  result := FEntryList;
end;

Function TFhirCompositionSection.GetHasEntryList : boolean;
begin
  result := (FEntryList <> nil) and (FEntryList.count > 0);
end;

Procedure TFhirCompositionSection.SetEmptyReason(value : TFhirCodeableConcept);
begin
  FEmptyReason.free;
  FEmptyReason := value;
end;

Function TFhirCompositionSection.GetSectionList : TFhirCompositionSectionList;
begin
  if FSectionList = nil then
    FSectionList := TFhirCompositionSectionList.Create;
  result := FSectionList;
end;

Function TFhirCompositionSection.GetHasSectionList : boolean;
begin
  result := (FSectionList <> nil) and (FSectionList.count > 0);
end;


{ TFhirCompositionSectionListEnumerator }

Constructor TFhirCompositionSectionListEnumerator.Create(list : TFhirCompositionSectionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCompositionSectionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCompositionSectionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCompositionSectionListEnumerator.GetCurrent : TFhirCompositionSection;
begin
  Result := FList[FIndex];
end;


{ TFhirCompositionSectionList }
procedure TFhirCompositionSectionList.AddItem(value: TFhirCompositionSection);
begin
  assert(value.ClassName = 'TFhirCompositionSection', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCompositionSection');
  add(value);
end;


function TFhirCompositionSectionList.Append: TFhirCompositionSection;
begin
  result := TFhirCompositionSection.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionSectionList.ClearItems;
begin
  Clear;
end;

function TFhirCompositionSectionList.GetEnumerator : TFhirCompositionSectionListEnumerator;
begin
  result := TFhirCompositionSectionListEnumerator.Create(self.link);
end;

function TFhirCompositionSectionList.Clone: TFhirCompositionSectionList;
begin
  result := TFhirCompositionSectionList(inherited Clone);
end;

function TFhirCompositionSectionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCompositionSectionList.GetItemN(index: Integer): TFhirCompositionSection;
begin
  result := TFhirCompositionSection(ObjectByIndex[index]);
end;

function TFhirCompositionSectionList.IndexOf(value: TFhirCompositionSection): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCompositionSectionList.Insert(index: Integer): TFhirCompositionSection;
begin
  result := TFhirCompositionSection.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionSectionList.InsertItem(index: Integer; value: TFhirCompositionSection);
begin
  assert(value is TFhirCompositionSection);
  Inherited Insert(index, value);
end;

function TFhirCompositionSectionList.Item(index: Integer): TFhirCompositionSection;
begin
  result := TFhirCompositionSection(ObjectByIndex[index]);
end;

function TFhirCompositionSectionList.Link: TFhirCompositionSectionList;
begin
  result := TFhirCompositionSectionList(inherited Link);
end;

procedure TFhirCompositionSectionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCompositionSectionList.SetItemByIndex(index: Integer; value: TFhirCompositionSection);
begin
  assert(value is TFhirCompositionSection);
  FhirCompositionSections[index] := value;
end;

procedure TFhirCompositionSectionList.SetItemN(index: Integer; value: TFhirCompositionSection);
begin
  assert(value is TFhirCompositionSection);
  ObjectByIndex[index] := value;
end;

{ TFhirComposition }

constructor TFhirComposition.Create;
begin
  inherited;
end;

destructor TFhirComposition.Destroy;
begin
  FIdentifier.free;
  FDate.free;
  FType_.free;
  FClass_.free;
  FTitle.free;
  FStatus.free;
  FConfidentiality.free;
  FSubject.free;
  FAuthorList.Free;
  FAttesterList.Free;
  FCustodian.free;
  FEventList.Free;
  FEncounter.free;
  FSectionList.Free;
  inherited;
end;

function TFhirComposition.GetResourceType : TFhirResourceType;
begin
  result := frtComposition;
end;

procedure TFhirComposition.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirComposition(oSource).identifier.Clone;
  dateElement := TFhirComposition(oSource).dateElement.Clone;
  type_ := TFhirComposition(oSource).type_.Clone;
  class_ := TFhirComposition(oSource).class_.Clone;
  titleElement := TFhirComposition(oSource).titleElement.Clone;
  FStatus := TFhirComposition(oSource).FStatus.Link;
  FConfidentiality := TFhirComposition(oSource).FConfidentiality.Link;
  subject := TFhirComposition(oSource).subject.Clone;
  if (TFhirComposition(oSource).FAuthorList = nil) then
  begin
    FAuthorList.free;
    FAuthorList := nil;
  end
  else
  begin
    FAuthorList := TFhirReferenceList{Resource}.Create;
    FAuthorList.Assign(TFhirComposition(oSource).FAuthorList);
  end;
  if (TFhirComposition(oSource).FAttesterList = nil) then
  begin
    FAttesterList.free;
    FAttesterList := nil;
  end
  else
  begin
    FAttesterList := TFhirCompositionAttesterList.Create;
    FAttesterList.Assign(TFhirComposition(oSource).FAttesterList);
  end;
  custodian := TFhirComposition(oSource).custodian.Clone;
  if (TFhirComposition(oSource).FEventList = nil) then
  begin
    FEventList.free;
    FEventList := nil;
  end
  else
  begin
    FEventList := TFhirCompositionEventList.Create;
    FEventList.Assign(TFhirComposition(oSource).FEventList);
  end;
  encounter := TFhirComposition(oSource).encounter.Clone;
  if (TFhirComposition(oSource).FSectionList = nil) then
  begin
    FSectionList.free;
    FSectionList := nil;
  end
  else
  begin
    FSectionList := TFhirCompositionSectionList.Create;
    FSectionList.Assign(TFhirComposition(oSource).FSectionList);
  end;
end;

procedure TFhirComposition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'class') Then
     list.add(FClass_.Link);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'confidentiality') Then
     list.add(FConfidentiality.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'author') Then
    list.addAll(FAuthorList);
  if (child_name = 'attester') Then
    list.addAll(FAttesterList);
  if (child_name = 'custodian') Then
     list.add(FCustodian.Link);
  if (child_name = 'event') Then
    list.addAll(FEventList);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'section') Then
    list.addAll(FSectionList);
end;

procedure TFhirComposition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'class', 'CodeableConcept', FClass_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'confidentiality', 'code', FConfidentiality.Link));{1}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Device|Patient|RelatedPerson)', FAuthorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'attester', '', FAttesterList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'custodian', 'Reference(Organization)', FCustodian.Link));{2}
  oList.add(TFHIRProperty.create(self, 'event', '', FEventList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'section', '', FSectionList.Link)){3};
end;

procedure TFhirComposition.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'class') then Class_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'confidentiality') then ConfidentialityElement := propValue as TFHIREnum
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'author') then AuthorList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'attester') then AttesterList.add(propValue as TFhirCompositionAttester){2}
  else if (propName = 'custodian') then Custodian := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'event') then EventList.add(propValue as TFhirCompositionEvent){2}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'section') then SectionList.add(propValue as TFhirCompositionSection){2}
  else inherited;
end;

function TFhirComposition.FhirType : string;
begin
  result := 'Composition';
end;

function TFhirComposition.Link : TFhirComposition;
begin
  result := TFhirComposition(inherited Link);
end;

function TFhirComposition.Clone : TFhirComposition;
begin
  result := TFhirComposition(inherited Clone);
end;

{ TFhirComposition }

Procedure TFhirComposition.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirComposition.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirComposition.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirComposition.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirComposition.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirComposition.SetClass_(value : TFhirCodeableConcept);
begin
  FClass_.free;
  FClass_ := value;
end;

Procedure TFhirComposition.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirComposition.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirComposition.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Procedure TFhirComposition.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirComposition.GetStatusST : TFhirCompositionStatus;
begin
  if FStatus = nil then
    result := TFhirCompositionStatus(0)
  else
    result := TFhirCompositionStatus(StringArrayIndexOfSensitive(CODES_TFhirCompositionStatus, FStatus.value));
end;

Procedure TFhirComposition.SetStatusST(value : TFhirCompositionStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirCompositionStatus[value]);
end;

Procedure TFhirComposition.SetConfidentiality(value : TFhirEnum);
begin
  FConfidentiality.free;
  FConfidentiality := value;
end;

Function TFhirComposition.GetConfidentialityST : TFhirV3Confidentiality;
begin
  if FConfidentiality = nil then
    result := TFhirV3Confidentiality(0)
  else
    result := TFhirV3Confidentiality(StringArrayIndexOfSensitive(CODES_TFhirV3Confidentiality, FConfidentiality.value));
end;

Procedure TFhirComposition.SetConfidentialityST(value : TFhirV3Confidentiality);
begin
  if ord(value) = 0 then
    ConfidentialityElement := nil
  else
    ConfidentialityElement := TFhirEnum.create(CODES_TFhirV3Confidentiality[value]);
end;

Procedure TFhirComposition.SetSubject(value : TFhirReference{TFhirReference});
begin
  FSubject.free;
  FSubject := value;
end;

Function TFhirComposition.GetAuthorList : TFhirReferenceList{Resource};
begin
  if FAuthorList = nil then
    FAuthorList := TFhirReferenceList{Resource}.Create;
  result := FAuthorList;
end;

Function TFhirComposition.GetHasAuthorList : boolean;
begin
  result := (FAuthorList <> nil) and (FAuthorList.count > 0);
end;

Function TFhirComposition.GetAttesterList : TFhirCompositionAttesterList;
begin
  if FAttesterList = nil then
    FAttesterList := TFhirCompositionAttesterList.Create;
  result := FAttesterList;
end;

Function TFhirComposition.GetHasAttesterList : boolean;
begin
  result := (FAttesterList <> nil) and (FAttesterList.count > 0);
end;

Procedure TFhirComposition.SetCustodian(value : TFhirReference{TFhirOrganization});
begin
  FCustodian.free;
  FCustodian := value;
end;

Function TFhirComposition.GetEventList : TFhirCompositionEventList;
begin
  if FEventList = nil then
    FEventList := TFhirCompositionEventList.Create;
  result := FEventList;
end;

Function TFhirComposition.GetHasEventList : boolean;
begin
  result := (FEventList <> nil) and (FEventList.count > 0);
end;

Procedure TFhirComposition.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Function TFhirComposition.GetSectionList : TFhirCompositionSectionList;
begin
  if FSectionList = nil then
    FSectionList := TFhirCompositionSectionList.Create;
  result := FSectionList;
end;

Function TFhirComposition.GetHasSectionList : boolean;
begin
  result := (FSectionList <> nil) and (FSectionList.count > 0);
end;


{ TFhirCompositionListEnumerator }

Constructor TFhirCompositionListEnumerator.Create(list : TFhirCompositionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCompositionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCompositionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCompositionListEnumerator.GetCurrent : TFhirComposition;
begin
  Result := FList[FIndex];
end;


{ TFhirCompositionList }
procedure TFhirCompositionList.AddItem(value: TFhirComposition);
begin
  assert(value.ClassName = 'TFhirComposition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirComposition');
  add(value);
end;


function TFhirCompositionList.Append: TFhirComposition;
begin
  result := TFhirComposition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionList.ClearItems;
begin
  Clear;
end;

function TFhirCompositionList.GetEnumerator : TFhirCompositionListEnumerator;
begin
  result := TFhirCompositionListEnumerator.Create(self.link);
end;

function TFhirCompositionList.Clone: TFhirCompositionList;
begin
  result := TFhirCompositionList(inherited Clone);
end;

function TFhirCompositionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCompositionList.GetItemN(index: Integer): TFhirComposition;
begin
  result := TFhirComposition(ObjectByIndex[index]);
end;

function TFhirCompositionList.IndexOf(value: TFhirComposition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCompositionList.Insert(index: Integer): TFhirComposition;
begin
  result := TFhirComposition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCompositionList.InsertItem(index: Integer; value: TFhirComposition);
begin
  assert(value is TFhirComposition);
  Inherited Insert(index, value);
end;

function TFhirCompositionList.Item(index: Integer): TFhirComposition;
begin
  result := TFhirComposition(ObjectByIndex[index]);
end;

function TFhirCompositionList.Link: TFhirCompositionList;
begin
  result := TFhirCompositionList(inherited Link);
end;

procedure TFhirCompositionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCompositionList.SetItemByIndex(index: Integer; value: TFhirComposition);
begin
  assert(value is TFhirComposition);
  FhirCompositions[index] := value;
end;

procedure TFhirCompositionList.SetItemN(index: Integer; value: TFhirComposition);
begin
  assert(value is TFhirComposition);
  ObjectByIndex[index] := value;
end;

{ TFhirConceptMapContact }

constructor TFhirConceptMapContact.Create;
begin
  inherited;
end;

destructor TFhirConceptMapContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirConceptMapContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirConceptMapContact(oSource).nameElement.Clone;
  if (TFhirConceptMapContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirConceptMapContact(oSource).FTelecomList);
  end;
end;

procedure TFhirConceptMapContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirConceptMapContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirConceptMapContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirConceptMapContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirConceptMapContact.Link : TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact(inherited Link);
end;

function TFhirConceptMapContact.Clone : TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact(inherited Clone);
end;

function TFhirConceptMapContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConceptMapContact)) then
    result := false
  else
  begin
    o := TFhirConceptMapContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirConceptMapContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConceptMapContact)) then
    result := false
  else
  begin
    o := TFhirConceptMapContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirConceptMapContact }

Procedure TFhirConceptMapContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConceptMapContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConceptMapContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirConceptMapContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirConceptMapContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirConceptMapContactListEnumerator }

Constructor TFhirConceptMapContactListEnumerator.Create(list : TFhirConceptMapContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConceptMapContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConceptMapContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConceptMapContactListEnumerator.GetCurrent : TFhirConceptMapContact;
begin
  Result := FList[FIndex];
end;


{ TFhirConceptMapContactList }
procedure TFhirConceptMapContactList.AddItem(value: TFhirConceptMapContact);
begin
  assert(value.ClassName = 'TFhirConceptMapContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConceptMapContact');
  add(value);
end;


function TFhirConceptMapContactList.Append: TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapContactList.ClearItems;
begin
  Clear;
end;

function TFhirConceptMapContactList.GetEnumerator : TFhirConceptMapContactListEnumerator;
begin
  result := TFhirConceptMapContactListEnumerator.Create(self.link);
end;

function TFhirConceptMapContactList.Clone: TFhirConceptMapContactList;
begin
  result := TFhirConceptMapContactList(inherited Clone);
end;

function TFhirConceptMapContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConceptMapContactList.GetItemN(index: Integer): TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact(ObjectByIndex[index]);
end;

function TFhirConceptMapContactList.IndexOf(value: TFhirConceptMapContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConceptMapContactList.Insert(index: Integer): TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapContactList.InsertItem(index: Integer; value: TFhirConceptMapContact);
begin
  assert(value is TFhirConceptMapContact);
  Inherited Insert(index, value);
end;

function TFhirConceptMapContactList.Item(index: Integer): TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact(ObjectByIndex[index]);
end;

function TFhirConceptMapContactList.Link: TFhirConceptMapContactList;
begin
  result := TFhirConceptMapContactList(inherited Link);
end;

procedure TFhirConceptMapContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConceptMapContactList.SetItemByIndex(index: Integer; value: TFhirConceptMapContact);
begin
  assert(value is TFhirConceptMapContact);
  FhirConceptMapContacts[index] := value;
end;

procedure TFhirConceptMapContactList.SetItemN(index: Integer; value: TFhirConceptMapContact);
begin
  assert(value is TFhirConceptMapContact);
  ObjectByIndex[index] := value;
end;

{ TFhirConceptMapElement }

constructor TFhirConceptMapElement.Create;
begin
  inherited;
end;

destructor TFhirConceptMapElement.Destroy;
begin
  FCodeSystem.free;
  FCode.free;
  FTargetList.Free;
  inherited;
end;

procedure TFhirConceptMapElement.Assign(oSource : TAdvObject);
begin
  inherited;
  codeSystemElement := TFhirConceptMapElement(oSource).codeSystemElement.Clone;
  codeElement := TFhirConceptMapElement(oSource).codeElement.Clone;
  if (TFhirConceptMapElement(oSource).FTargetList = nil) then
  begin
    FTargetList.free;
    FTargetList := nil;
  end
  else
  begin
    FTargetList := TFhirConceptMapElementTargetList.Create;
    FTargetList.Assign(TFhirConceptMapElement(oSource).FTargetList);
  end;
end;

procedure TFhirConceptMapElement.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'codeSystem') Then
     list.add(FCodeSystem.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'target') Then
    list.addAll(FTargetList);
end;

procedure TFhirConceptMapElement.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'codeSystem', 'uri', FCodeSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', '', FTargetList.Link)){3};
end;

procedure TFhirConceptMapElement.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'codeSystem') then CodeSystemElement := propValue as TFhirUri{5a}
  else if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'target') then TargetList.add(propValue as TFhirConceptMapElementTarget){2}
  else inherited;
end;

function TFhirConceptMapElement.FhirType : string;
begin
  result := 'element';
end;

function TFhirConceptMapElement.Link : TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement(inherited Link);
end;

function TFhirConceptMapElement.Clone : TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement(inherited Clone);
end;

function TFhirConceptMapElement.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElement;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConceptMapElement)) then
    result := false
  else
  begin
    o := TFhirConceptMapElement(other);
    result := compareDeep(codeSystemElement, o.codeSystemElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(targetList, o.targetList, true);
  end;
end;

function TFhirConceptMapElement.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElement;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConceptMapElement)) then
    result := false
  else
  begin
    o := TFhirConceptMapElement(other);
    result := compareValues(codeSystemElement, o.codeSystemElement, true) and compareValues(codeElement, o.codeElement, true);
  end;
end;

{ TFhirConceptMapElement }

Procedure TFhirConceptMapElement.SetCodeSystem(value : TFhirUri);
begin
  FCodeSystem.free;
  FCodeSystem := value;
end;

Function TFhirConceptMapElement.GetCodeSystemST : String;
begin
  if FCodeSystem = nil then
    result := ''
  else
    result := FCodeSystem.value;
end;

Procedure TFhirConceptMapElement.SetCodeSystemST(value : String);
begin
  if value <> '' then
  begin
    if FCodeSystem = nil then
      FCodeSystem := TFhirUri.create;
    FCodeSystem.value := value
  end
  else if FCodeSystem <> nil then
    FCodeSystem.value := '';
end;

Procedure TFhirConceptMapElement.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConceptMapElement.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirConceptMapElement.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Function TFhirConceptMapElement.GetTargetList : TFhirConceptMapElementTargetList;
begin
  if FTargetList = nil then
    FTargetList := TFhirConceptMapElementTargetList.Create;
  result := FTargetList;
end;

Function TFhirConceptMapElement.GetHasTargetList : boolean;
begin
  result := (FTargetList <> nil) and (FTargetList.count > 0);
end;


{ TFhirConceptMapElementListEnumerator }

Constructor TFhirConceptMapElementListEnumerator.Create(list : TFhirConceptMapElementList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConceptMapElementListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConceptMapElementListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConceptMapElementListEnumerator.GetCurrent : TFhirConceptMapElement;
begin
  Result := FList[FIndex];
end;


{ TFhirConceptMapElementList }
procedure TFhirConceptMapElementList.AddItem(value: TFhirConceptMapElement);
begin
  assert(value.ClassName = 'TFhirConceptMapElement', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConceptMapElement');
  add(value);
end;


function TFhirConceptMapElementList.Append: TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementList.ClearItems;
begin
  Clear;
end;

function TFhirConceptMapElementList.GetEnumerator : TFhirConceptMapElementListEnumerator;
begin
  result := TFhirConceptMapElementListEnumerator.Create(self.link);
end;

function TFhirConceptMapElementList.Clone: TFhirConceptMapElementList;
begin
  result := TFhirConceptMapElementList(inherited Clone);
end;

function TFhirConceptMapElementList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConceptMapElementList.GetItemN(index: Integer): TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement(ObjectByIndex[index]);
end;

function TFhirConceptMapElementList.IndexOf(value: TFhirConceptMapElement): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConceptMapElementList.Insert(index: Integer): TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementList.InsertItem(index: Integer; value: TFhirConceptMapElement);
begin
  assert(value is TFhirConceptMapElement);
  Inherited Insert(index, value);
end;

function TFhirConceptMapElementList.Item(index: Integer): TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement(ObjectByIndex[index]);
end;

function TFhirConceptMapElementList.Link: TFhirConceptMapElementList;
begin
  result := TFhirConceptMapElementList(inherited Link);
end;

procedure TFhirConceptMapElementList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConceptMapElementList.SetItemByIndex(index: Integer; value: TFhirConceptMapElement);
begin
  assert(value is TFhirConceptMapElement);
  FhirConceptMapElements[index] := value;
end;

procedure TFhirConceptMapElementList.SetItemN(index: Integer; value: TFhirConceptMapElement);
begin
  assert(value is TFhirConceptMapElement);
  ObjectByIndex[index] := value;
end;

{ TFhirConceptMapElementTarget }

constructor TFhirConceptMapElementTarget.Create;
begin
  inherited;
end;

destructor TFhirConceptMapElementTarget.Destroy;
begin
  FCodeSystem.free;
  FCode.free;
  FEquivalence.free;
  FComments.free;
  FDependsOnList.Free;
  FProductList.Free;
  inherited;
end;

procedure TFhirConceptMapElementTarget.Assign(oSource : TAdvObject);
begin
  inherited;
  codeSystemElement := TFhirConceptMapElementTarget(oSource).codeSystemElement.Clone;
  codeElement := TFhirConceptMapElementTarget(oSource).codeElement.Clone;
  FEquivalence := TFhirConceptMapElementTarget(oSource).FEquivalence.Link;
  commentsElement := TFhirConceptMapElementTarget(oSource).commentsElement.Clone;
  if (TFhirConceptMapElementTarget(oSource).FDependsOnList = nil) then
  begin
    FDependsOnList.free;
    FDependsOnList := nil;
  end
  else
  begin
    FDependsOnList := TFhirConceptMapElementTargetDependsOnList.Create;
    FDependsOnList.Assign(TFhirConceptMapElementTarget(oSource).FDependsOnList);
  end;
  if (TFhirConceptMapElementTarget(oSource).FProductList = nil) then
  begin
    FProductList.free;
    FProductList := nil;
  end
  else
  begin
    FProductList := TFhirConceptMapElementTargetDependsOnList.Create;
    FProductList.Assign(TFhirConceptMapElementTarget(oSource).FProductList);
  end;
end;

procedure TFhirConceptMapElementTarget.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'codeSystem') Then
     list.add(FCodeSystem.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'equivalence') Then
     list.add(FEquivalence.Link);
  if (child_name = 'comments') Then
     list.add(FComments.Link);
  if (child_name = 'dependsOn') Then
    list.addAll(FDependsOnList);
  if (child_name = 'product') Then
    list.addAll(FProductList);
end;

procedure TFhirConceptMapElementTarget.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'codeSystem', 'uri', FCodeSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'equivalence', 'code', FEquivalence.Link));{1}
  oList.add(TFHIRProperty.create(self, 'comments', 'string', FComments.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dependsOn', '', FDependsOnList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'product', '@ConceptMap.element.target.dependsOn', FProductList.Link)){3};
end;

procedure TFhirConceptMapElementTarget.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'codeSystem') then CodeSystemElement := propValue as TFhirUri{5a}
  else if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'equivalence') then EquivalenceElement := propValue as TFHIREnum
  else if (propName = 'comments') then CommentsElement := propValue as TFhirString{5a}
  else if (propName = 'dependsOn') then DependsOnList.add(propValue as TFhirConceptMapElementTargetDependsOn){2}
  else if (propName = 'product') then ProductList.add(propValue as TFhirConceptMapElementTargetDependsOn){2}
  else inherited;
end;

function TFhirConceptMapElementTarget.FhirType : string;
begin
  result := 'target';
end;

function TFhirConceptMapElementTarget.Link : TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget(inherited Link);
end;

function TFhirConceptMapElementTarget.Clone : TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget(inherited Clone);
end;

function TFhirConceptMapElementTarget.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElementTarget;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConceptMapElementTarget)) then
    result := false
  else
  begin
    o := TFhirConceptMapElementTarget(other);
    result := compareDeep(codeSystemElement, o.codeSystemElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(equivalenceElement, o.equivalenceElement, true) and compareDeep(commentsElement, o.commentsElement, true) and 
      compareDeep(dependsOnList, o.dependsOnList, true) and compareDeep(productList, o.productList, true);
  end;
end;

function TFhirConceptMapElementTarget.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElementTarget;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConceptMapElementTarget)) then
    result := false
  else
  begin
    o := TFhirConceptMapElementTarget(other);
    result := compareValues(codeSystemElement, o.codeSystemElement, true) and compareValues(codeElement, o.codeElement, true) and 
      compareValues(equivalenceElement, o.equivalenceElement, true) and compareValues(commentsElement, o.commentsElement, true);
  end;
end;

{ TFhirConceptMapElementTarget }

Procedure TFhirConceptMapElementTarget.SetCodeSystem(value : TFhirUri);
begin
  FCodeSystem.free;
  FCodeSystem := value;
end;

Function TFhirConceptMapElementTarget.GetCodeSystemST : String;
begin
  if FCodeSystem = nil then
    result := ''
  else
    result := FCodeSystem.value;
end;

Procedure TFhirConceptMapElementTarget.SetCodeSystemST(value : String);
begin
  if value <> '' then
  begin
    if FCodeSystem = nil then
      FCodeSystem := TFhirUri.create;
    FCodeSystem.value := value
  end
  else if FCodeSystem <> nil then
    FCodeSystem.value := '';
end;

Procedure TFhirConceptMapElementTarget.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConceptMapElementTarget.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirConceptMapElementTarget.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirConceptMapElementTarget.SetEquivalence(value : TFhirEnum);
begin
  FEquivalence.free;
  FEquivalence := value;
end;

Function TFhirConceptMapElementTarget.GetEquivalenceST : TFhirConceptMapEquivalence;
begin
  if FEquivalence = nil then
    result := TFhirConceptMapEquivalence(0)
  else
    result := TFhirConceptMapEquivalence(StringArrayIndexOfSensitive(CODES_TFhirConceptMapEquivalence, FEquivalence.value));
end;

Procedure TFhirConceptMapElementTarget.SetEquivalenceST(value : TFhirConceptMapEquivalence);
begin
  if ord(value) = 0 then
    EquivalenceElement := nil
  else
    EquivalenceElement := TFhirEnum.create(CODES_TFhirConceptMapEquivalence[value]);
end;

Procedure TFhirConceptMapElementTarget.SetComments(value : TFhirString);
begin
  FComments.free;
  FComments := value;
end;

Function TFhirConceptMapElementTarget.GetCommentsST : String;
begin
  if FComments = nil then
    result := ''
  else
    result := FComments.value;
end;

Procedure TFhirConceptMapElementTarget.SetCommentsST(value : String);
begin
  if value <> '' then
  begin
    if FComments = nil then
      FComments := TFhirString.create;
    FComments.value := value
  end
  else if FComments <> nil then
    FComments.value := '';
end;

Function TFhirConceptMapElementTarget.GetDependsOnList : TFhirConceptMapElementTargetDependsOnList;
begin
  if FDependsOnList = nil then
    FDependsOnList := TFhirConceptMapElementTargetDependsOnList.Create;
  result := FDependsOnList;
end;

Function TFhirConceptMapElementTarget.GetHasDependsOnList : boolean;
begin
  result := (FDependsOnList <> nil) and (FDependsOnList.count > 0);
end;

Function TFhirConceptMapElementTarget.GetProductList : TFhirConceptMapElementTargetDependsOnList;
begin
  if FProductList = nil then
    FProductList := TFhirConceptMapElementTargetDependsOnList.Create;
  result := FProductList;
end;

Function TFhirConceptMapElementTarget.GetHasProductList : boolean;
begin
  result := (FProductList <> nil) and (FProductList.count > 0);
end;


{ TFhirConceptMapElementTargetListEnumerator }

Constructor TFhirConceptMapElementTargetListEnumerator.Create(list : TFhirConceptMapElementTargetList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConceptMapElementTargetListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConceptMapElementTargetListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConceptMapElementTargetListEnumerator.GetCurrent : TFhirConceptMapElementTarget;
begin
  Result := FList[FIndex];
end;


{ TFhirConceptMapElementTargetList }
procedure TFhirConceptMapElementTargetList.AddItem(value: TFhirConceptMapElementTarget);
begin
  assert(value.ClassName = 'TFhirConceptMapElementTarget', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConceptMapElementTarget');
  add(value);
end;


function TFhirConceptMapElementTargetList.Append: TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementTargetList.ClearItems;
begin
  Clear;
end;

function TFhirConceptMapElementTargetList.GetEnumerator : TFhirConceptMapElementTargetListEnumerator;
begin
  result := TFhirConceptMapElementTargetListEnumerator.Create(self.link);
end;

function TFhirConceptMapElementTargetList.Clone: TFhirConceptMapElementTargetList;
begin
  result := TFhirConceptMapElementTargetList(inherited Clone);
end;

function TFhirConceptMapElementTargetList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConceptMapElementTargetList.GetItemN(index: Integer): TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget(ObjectByIndex[index]);
end;

function TFhirConceptMapElementTargetList.IndexOf(value: TFhirConceptMapElementTarget): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConceptMapElementTargetList.Insert(index: Integer): TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementTargetList.InsertItem(index: Integer; value: TFhirConceptMapElementTarget);
begin
  assert(value is TFhirConceptMapElementTarget);
  Inherited Insert(index, value);
end;

function TFhirConceptMapElementTargetList.Item(index: Integer): TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget(ObjectByIndex[index]);
end;

function TFhirConceptMapElementTargetList.Link: TFhirConceptMapElementTargetList;
begin
  result := TFhirConceptMapElementTargetList(inherited Link);
end;

procedure TFhirConceptMapElementTargetList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConceptMapElementTargetList.SetItemByIndex(index: Integer; value: TFhirConceptMapElementTarget);
begin
  assert(value is TFhirConceptMapElementTarget);
  FhirConceptMapElementTargets[index] := value;
end;

procedure TFhirConceptMapElementTargetList.SetItemN(index: Integer; value: TFhirConceptMapElementTarget);
begin
  assert(value is TFhirConceptMapElementTarget);
  ObjectByIndex[index] := value;
end;

{ TFhirConceptMapElementTargetDependsOn }

constructor TFhirConceptMapElementTargetDependsOn.Create;
begin
  inherited;
end;

destructor TFhirConceptMapElementTargetDependsOn.Destroy;
begin
  FElement.free;
  FCodeSystem.free;
  FCode.free;
  inherited;
end;

procedure TFhirConceptMapElementTargetDependsOn.Assign(oSource : TAdvObject);
begin
  inherited;
  elementElement := TFhirConceptMapElementTargetDependsOn(oSource).elementElement.Clone;
  codeSystemElement := TFhirConceptMapElementTargetDependsOn(oSource).codeSystemElement.Clone;
  codeElement := TFhirConceptMapElementTargetDependsOn(oSource).codeElement.Clone;
end;

procedure TFhirConceptMapElementTargetDependsOn.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'element') Then
     list.add(FElement.Link);
  if (child_name = 'codeSystem') Then
     list.add(FCodeSystem.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
end;

procedure TFhirConceptMapElementTargetDependsOn.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'element', 'uri', FElement.Link));{2}
  oList.add(TFHIRProperty.create(self, 'codeSystem', 'uri', FCodeSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'string', FCode.Link));{2}
end;

procedure TFhirConceptMapElementTargetDependsOn.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'element') then ElementElement := propValue as TFhirUri{5a}
  else if (propName = 'codeSystem') then CodeSystemElement := propValue as TFhirUri{5a}
  else if (propName = 'code') then CodeElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirConceptMapElementTargetDependsOn.FhirType : string;
begin
  result := 'dependsOn';
end;

function TFhirConceptMapElementTargetDependsOn.Link : TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn(inherited Link);
end;

function TFhirConceptMapElementTargetDependsOn.Clone : TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn(inherited Clone);
end;

function TFhirConceptMapElementTargetDependsOn.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElementTargetDependsOn;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConceptMapElementTargetDependsOn)) then
    result := false
  else
  begin
    o := TFhirConceptMapElementTargetDependsOn(other);
    result := compareDeep(elementElement, o.elementElement, true) and compareDeep(codeSystemElement, o.codeSystemElement, true) and 
      compareDeep(codeElement, o.codeElement, true);
  end;
end;

function TFhirConceptMapElementTargetDependsOn.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConceptMapElementTargetDependsOn;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConceptMapElementTargetDependsOn)) then
    result := false
  else
  begin
    o := TFhirConceptMapElementTargetDependsOn(other);
    result := compareValues(elementElement, o.elementElement, true) and compareValues(codeSystemElement, o.codeSystemElement, true) and 
      compareValues(codeElement, o.codeElement, true);
  end;
end;

{ TFhirConceptMapElementTargetDependsOn }

Procedure TFhirConceptMapElementTargetDependsOn.SetElement(value : TFhirUri);
begin
  FElement.free;
  FElement := value;
end;

Function TFhirConceptMapElementTargetDependsOn.GetElementST : String;
begin
  if FElement = nil then
    result := ''
  else
    result := FElement.value;
end;

Procedure TFhirConceptMapElementTargetDependsOn.SetElementST(value : String);
begin
  if value <> '' then
  begin
    if FElement = nil then
      FElement := TFhirUri.create;
    FElement.value := value
  end
  else if FElement <> nil then
    FElement.value := '';
end;

Procedure TFhirConceptMapElementTargetDependsOn.SetCodeSystem(value : TFhirUri);
begin
  FCodeSystem.free;
  FCodeSystem := value;
end;

Function TFhirConceptMapElementTargetDependsOn.GetCodeSystemST : String;
begin
  if FCodeSystem = nil then
    result := ''
  else
    result := FCodeSystem.value;
end;

Procedure TFhirConceptMapElementTargetDependsOn.SetCodeSystemST(value : String);
begin
  if value <> '' then
  begin
    if FCodeSystem = nil then
      FCodeSystem := TFhirUri.create;
    FCodeSystem.value := value
  end
  else if FCodeSystem <> nil then
    FCodeSystem.value := '';
end;

Procedure TFhirConceptMapElementTargetDependsOn.SetCode(value : TFhirString);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConceptMapElementTargetDependsOn.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirConceptMapElementTargetDependsOn.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirString.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;


{ TFhirConceptMapElementTargetDependsOnListEnumerator }

Constructor TFhirConceptMapElementTargetDependsOnListEnumerator.Create(list : TFhirConceptMapElementTargetDependsOnList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConceptMapElementTargetDependsOnListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConceptMapElementTargetDependsOnListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConceptMapElementTargetDependsOnListEnumerator.GetCurrent : TFhirConceptMapElementTargetDependsOn;
begin
  Result := FList[FIndex];
end;


{ TFhirConceptMapElementTargetDependsOnList }
procedure TFhirConceptMapElementTargetDependsOnList.AddItem(value: TFhirConceptMapElementTargetDependsOn);
begin
  assert(value.ClassName = 'TFhirConceptMapElementTargetDependsOn', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConceptMapElementTargetDependsOn');
  add(value);
end;


function TFhirConceptMapElementTargetDependsOnList.Append: TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementTargetDependsOnList.ClearItems;
begin
  Clear;
end;

function TFhirConceptMapElementTargetDependsOnList.GetEnumerator : TFhirConceptMapElementTargetDependsOnListEnumerator;
begin
  result := TFhirConceptMapElementTargetDependsOnListEnumerator.Create(self.link);
end;

function TFhirConceptMapElementTargetDependsOnList.Clone: TFhirConceptMapElementTargetDependsOnList;
begin
  result := TFhirConceptMapElementTargetDependsOnList(inherited Clone);
end;

function TFhirConceptMapElementTargetDependsOnList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConceptMapElementTargetDependsOnList.GetItemN(index: Integer): TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn(ObjectByIndex[index]);
end;

function TFhirConceptMapElementTargetDependsOnList.IndexOf(value: TFhirConceptMapElementTargetDependsOn): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConceptMapElementTargetDependsOnList.Insert(index: Integer): TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapElementTargetDependsOnList.InsertItem(index: Integer; value: TFhirConceptMapElementTargetDependsOn);
begin
  assert(value is TFhirConceptMapElementTargetDependsOn);
  Inherited Insert(index, value);
end;

function TFhirConceptMapElementTargetDependsOnList.Item(index: Integer): TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn(ObjectByIndex[index]);
end;

function TFhirConceptMapElementTargetDependsOnList.Link: TFhirConceptMapElementTargetDependsOnList;
begin
  result := TFhirConceptMapElementTargetDependsOnList(inherited Link);
end;

procedure TFhirConceptMapElementTargetDependsOnList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConceptMapElementTargetDependsOnList.SetItemByIndex(index: Integer; value: TFhirConceptMapElementTargetDependsOn);
begin
  assert(value is TFhirConceptMapElementTargetDependsOn);
  FhirConceptMapElementTargetDependsOns[index] := value;
end;

procedure TFhirConceptMapElementTargetDependsOnList.SetItemN(index: Integer; value: TFhirConceptMapElementTargetDependsOn);
begin
  assert(value is TFhirConceptMapElementTargetDependsOn);
  ObjectByIndex[index] := value;
end;

{ TFhirConceptMap }

constructor TFhirConceptMap.Create;
begin
  inherited;
end;

destructor TFhirConceptMap.Destroy;
begin
  FUrl.free;
  FIdentifier.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FUseContextList.Free;
  FRequirements.free;
  FCopyright.free;
  FSource.free;
  FTarget.free;
  FElementList.Free;
  inherited;
end;

function TFhirConceptMap.GetResourceType : TFhirResourceType;
begin
  result := frtConceptMap;
end;

procedure TFhirConceptMap.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirConceptMap(oSource).urlElement.Clone;
  identifier := TFhirConceptMap(oSource).identifier.Clone;
  versionElement := TFhirConceptMap(oSource).versionElement.Clone;
  nameElement := TFhirConceptMap(oSource).nameElement.Clone;
  FStatus := TFhirConceptMap(oSource).FStatus.Link;
  experimentalElement := TFhirConceptMap(oSource).experimentalElement.Clone;
  publisherElement := TFhirConceptMap(oSource).publisherElement.Clone;
  if (TFhirConceptMap(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirConceptMapContactList.Create;
    FContactList.Assign(TFhirConceptMap(oSource).FContactList);
  end;
  dateElement := TFhirConceptMap(oSource).dateElement.Clone;
  descriptionElement := TFhirConceptMap(oSource).descriptionElement.Clone;
  if (TFhirConceptMap(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirConceptMap(oSource).FUseContextList);
  end;
  requirementsElement := TFhirConceptMap(oSource).requirementsElement.Clone;
  copyrightElement := TFhirConceptMap(oSource).copyrightElement.Clone;
  source := TFhirConceptMap(oSource).source.Clone;
  target := TFhirConceptMap(oSource).target.Clone;
  if (TFhirConceptMap(oSource).FElementList = nil) then
  begin
    FElementList.free;
    FElementList := nil;
  end
  else
  begin
    FElementList := TFhirConceptMapElementList.Create;
    FElementList.Assign(TFhirConceptMap(oSource).FElementList);
  end;
end;

procedure TFhirConceptMap.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'source[x]') Then
     list.add(FSource.Link);
  if (child_name = 'target[x]') Then
     list.add(FTarget.Link);
  if (child_name = 'element') Then
    list.addAll(FElementList);
end;

procedure TFhirConceptMap.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source[x]', 'uri|Reference(ValueSet|StructureDefinition)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target[x]', 'uri|Reference(ValueSet|StructureDefinition)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'element', '', FElementList.Link)){3};
end;

procedure TFhirConceptMap.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirConceptMapContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName.startsWith('source')) then Source := propValue as TFhirType{4}
  else if (propName.startsWith('target')) then Target := propValue as TFhirType{4}
  else if (propName = 'element') then ElementList.add(propValue as TFhirConceptMapElement){2}
  else inherited;
end;

function TFhirConceptMap.FhirType : string;
begin
  result := 'ConceptMap';
end;

function TFhirConceptMap.Link : TFhirConceptMap;
begin
  result := TFhirConceptMap(inherited Link);
end;

function TFhirConceptMap.Clone : TFhirConceptMap;
begin
  result := TFhirConceptMap(inherited Clone);
end;

{ TFhirConceptMap }

Procedure TFhirConceptMap.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirConceptMap.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirConceptMap.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirConceptMap.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirConceptMap.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirConceptMap.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirConceptMap.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirConceptMap.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConceptMap.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConceptMap.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirConceptMap.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirConceptMap.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirConceptMap.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirConceptMap.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirConceptMap.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirConceptMap.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirConceptMap.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirConceptMap.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirConceptMap.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirConceptMap.GetContactList : TFhirConceptMapContactList;
begin
  if FContactList = nil then
    FContactList := TFhirConceptMapContactList.Create;
  result := FContactList;
end;

Function TFhirConceptMap.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirConceptMap.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirConceptMap.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirConceptMap.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirConceptMap.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirConceptMap.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirConceptMap.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirConceptMap.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirConceptMap.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirConceptMap.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirConceptMap.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirConceptMap.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirConceptMap.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirConceptMap.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirConceptMap.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirConceptMap.SetSource(value : TFhirType);
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirConceptMap.SetTarget(value : TFhirType);
begin
  FTarget.free;
  FTarget := value;
end;

Function TFhirConceptMap.GetElementList : TFhirConceptMapElementList;
begin
  if FElementList = nil then
    FElementList := TFhirConceptMapElementList.Create;
  result := FElementList;
end;

Function TFhirConceptMap.GetHasElementList : boolean;
begin
  result := (FElementList <> nil) and (FElementList.count > 0);
end;


{ TFhirConceptMapListEnumerator }

Constructor TFhirConceptMapListEnumerator.Create(list : TFhirConceptMapList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConceptMapListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConceptMapListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConceptMapListEnumerator.GetCurrent : TFhirConceptMap;
begin
  Result := FList[FIndex];
end;


{ TFhirConceptMapList }
procedure TFhirConceptMapList.AddItem(value: TFhirConceptMap);
begin
  assert(value.ClassName = 'TFhirConceptMap', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConceptMap');
  add(value);
end;


function TFhirConceptMapList.Append: TFhirConceptMap;
begin
  result := TFhirConceptMap.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapList.ClearItems;
begin
  Clear;
end;

function TFhirConceptMapList.GetEnumerator : TFhirConceptMapListEnumerator;
begin
  result := TFhirConceptMapListEnumerator.Create(self.link);
end;

function TFhirConceptMapList.Clone: TFhirConceptMapList;
begin
  result := TFhirConceptMapList(inherited Clone);
end;

function TFhirConceptMapList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConceptMapList.GetItemN(index: Integer): TFhirConceptMap;
begin
  result := TFhirConceptMap(ObjectByIndex[index]);
end;

function TFhirConceptMapList.IndexOf(value: TFhirConceptMap): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConceptMapList.Insert(index: Integer): TFhirConceptMap;
begin
  result := TFhirConceptMap.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConceptMapList.InsertItem(index: Integer; value: TFhirConceptMap);
begin
  assert(value is TFhirConceptMap);
  Inherited Insert(index, value);
end;

function TFhirConceptMapList.Item(index: Integer): TFhirConceptMap;
begin
  result := TFhirConceptMap(ObjectByIndex[index]);
end;

function TFhirConceptMapList.Link: TFhirConceptMapList;
begin
  result := TFhirConceptMapList(inherited Link);
end;

procedure TFhirConceptMapList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConceptMapList.SetItemByIndex(index: Integer; value: TFhirConceptMap);
begin
  assert(value is TFhirConceptMap);
  FhirConceptMaps[index] := value;
end;

procedure TFhirConceptMapList.SetItemN(index: Integer; value: TFhirConceptMap);
begin
  assert(value is TFhirConceptMap);
  ObjectByIndex[index] := value;
end;

{ TFhirConditionStage }

constructor TFhirConditionStage.Create;
begin
  inherited;
end;

destructor TFhirConditionStage.Destroy;
begin
  FSummary.free;
  FAssessmentList.Free;
  inherited;
end;

procedure TFhirConditionStage.Assign(oSource : TAdvObject);
begin
  inherited;
  summary := TFhirConditionStage(oSource).summary.Clone;
  if (TFhirConditionStage(oSource).FAssessmentList = nil) then
  begin
    FAssessmentList.free;
    FAssessmentList := nil;
  end
  else
  begin
    FAssessmentList := TFhirReferenceList{Resource}.Create;
    FAssessmentList.Assign(TFhirConditionStage(oSource).FAssessmentList);
  end;
end;

procedure TFhirConditionStage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'summary') Then
     list.add(FSummary.Link);
  if (child_name = 'assessment') Then
    list.addAll(FAssessmentList);
end;

procedure TFhirConditionStage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'summary', 'CodeableConcept', FSummary.Link));{2}
  oList.add(TFHIRProperty.create(self, 'assessment', 'Reference(ClinicalImpression|DiagnosticReport|Observation)', FAssessmentList.Link)){3};
end;

procedure TFhirConditionStage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'summary') then Summary := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'assessment') then AssessmentList.add(propValue as TFhirReference{Resource}){2}
  else inherited;
end;

function TFhirConditionStage.FhirType : string;
begin
  result := 'stage';
end;

function TFhirConditionStage.Link : TFhirConditionStage;
begin
  result := TFhirConditionStage(inherited Link);
end;

function TFhirConditionStage.Clone : TFhirConditionStage;
begin
  result := TFhirConditionStage(inherited Clone);
end;

function TFhirConditionStage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConditionStage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConditionStage)) then
    result := false
  else
  begin
    o := TFhirConditionStage(other);
    result := compareDeep(summaryElement, o.summaryElement, true) and compareDeep(assessmentList, o.assessmentList, true);
  end;
end;

function TFhirConditionStage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConditionStage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConditionStage)) then
    result := false
  else
  begin
    o := TFhirConditionStage(other);
    result := true;
  end;
end;

{ TFhirConditionStage }

Procedure TFhirConditionStage.SetSummary(value : TFhirCodeableConcept);
begin
  FSummary.free;
  FSummary := value;
end;

Function TFhirConditionStage.GetAssessmentList : TFhirReferenceList{Resource};
begin
  if FAssessmentList = nil then
    FAssessmentList := TFhirReferenceList{Resource}.Create;
  result := FAssessmentList;
end;

Function TFhirConditionStage.GetHasAssessmentList : boolean;
begin
  result := (FAssessmentList <> nil) and (FAssessmentList.count > 0);
end;


{ TFhirConditionStageListEnumerator }

Constructor TFhirConditionStageListEnumerator.Create(list : TFhirConditionStageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConditionStageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConditionStageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConditionStageListEnumerator.GetCurrent : TFhirConditionStage;
begin
  Result := FList[FIndex];
end;


{ TFhirConditionStageList }
procedure TFhirConditionStageList.AddItem(value: TFhirConditionStage);
begin
  assert(value.ClassName = 'TFhirConditionStage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConditionStage');
  add(value);
end;


function TFhirConditionStageList.Append: TFhirConditionStage;
begin
  result := TFhirConditionStage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionStageList.ClearItems;
begin
  Clear;
end;

function TFhirConditionStageList.GetEnumerator : TFhirConditionStageListEnumerator;
begin
  result := TFhirConditionStageListEnumerator.Create(self.link);
end;

function TFhirConditionStageList.Clone: TFhirConditionStageList;
begin
  result := TFhirConditionStageList(inherited Clone);
end;

function TFhirConditionStageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConditionStageList.GetItemN(index: Integer): TFhirConditionStage;
begin
  result := TFhirConditionStage(ObjectByIndex[index]);
end;

function TFhirConditionStageList.IndexOf(value: TFhirConditionStage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConditionStageList.Insert(index: Integer): TFhirConditionStage;
begin
  result := TFhirConditionStage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionStageList.InsertItem(index: Integer; value: TFhirConditionStage);
begin
  assert(value is TFhirConditionStage);
  Inherited Insert(index, value);
end;

function TFhirConditionStageList.Item(index: Integer): TFhirConditionStage;
begin
  result := TFhirConditionStage(ObjectByIndex[index]);
end;

function TFhirConditionStageList.Link: TFhirConditionStageList;
begin
  result := TFhirConditionStageList(inherited Link);
end;

procedure TFhirConditionStageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConditionStageList.SetItemByIndex(index: Integer; value: TFhirConditionStage);
begin
  assert(value is TFhirConditionStage);
  FhirConditionStages[index] := value;
end;

procedure TFhirConditionStageList.SetItemN(index: Integer; value: TFhirConditionStage);
begin
  assert(value is TFhirConditionStage);
  ObjectByIndex[index] := value;
end;

{ TFhirConditionEvidence }

constructor TFhirConditionEvidence.Create;
begin
  inherited;
end;

destructor TFhirConditionEvidence.Destroy;
begin
  FCode.free;
  FDetailList.Free;
  inherited;
end;

procedure TFhirConditionEvidence.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirConditionEvidence(oSource).code.Clone;
  if (TFhirConditionEvidence(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
    FDetailList.Assign(TFhirConditionEvidence(oSource).FDetailList);
  end;
end;

procedure TFhirConditionEvidence.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirConditionEvidence.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', 'Reference(Any)', FDetailList.Link)){3};
end;

procedure TFhirConditionEvidence.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirReference{TFhirReference}){2}
  else inherited;
end;

function TFhirConditionEvidence.FhirType : string;
begin
  result := 'evidence';
end;

function TFhirConditionEvidence.Link : TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence(inherited Link);
end;

function TFhirConditionEvidence.Clone : TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence(inherited Clone);
end;

function TFhirConditionEvidence.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConditionEvidence;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConditionEvidence)) then
    result := false
  else
  begin
    o := TFhirConditionEvidence(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(detailList, o.detailList, true);
  end;
end;

function TFhirConditionEvidence.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConditionEvidence;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConditionEvidence)) then
    result := false
  else
  begin
    o := TFhirConditionEvidence(other);
    result := true;
  end;
end;

{ TFhirConditionEvidence }

Procedure TFhirConditionEvidence.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConditionEvidence.GetDetailList : TFhirReferenceList{TFhirReference};
begin
  if FDetailList = nil then
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
  result := FDetailList;
end;

Function TFhirConditionEvidence.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirConditionEvidenceListEnumerator }

Constructor TFhirConditionEvidenceListEnumerator.Create(list : TFhirConditionEvidenceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConditionEvidenceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConditionEvidenceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConditionEvidenceListEnumerator.GetCurrent : TFhirConditionEvidence;
begin
  Result := FList[FIndex];
end;


{ TFhirConditionEvidenceList }
procedure TFhirConditionEvidenceList.AddItem(value: TFhirConditionEvidence);
begin
  assert(value.ClassName = 'TFhirConditionEvidence', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConditionEvidence');
  add(value);
end;


function TFhirConditionEvidenceList.Append: TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionEvidenceList.ClearItems;
begin
  Clear;
end;

function TFhirConditionEvidenceList.GetEnumerator : TFhirConditionEvidenceListEnumerator;
begin
  result := TFhirConditionEvidenceListEnumerator.Create(self.link);
end;

function TFhirConditionEvidenceList.Clone: TFhirConditionEvidenceList;
begin
  result := TFhirConditionEvidenceList(inherited Clone);
end;

function TFhirConditionEvidenceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConditionEvidenceList.GetItemN(index: Integer): TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence(ObjectByIndex[index]);
end;

function TFhirConditionEvidenceList.IndexOf(value: TFhirConditionEvidence): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConditionEvidenceList.Insert(index: Integer): TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionEvidenceList.InsertItem(index: Integer; value: TFhirConditionEvidence);
begin
  assert(value is TFhirConditionEvidence);
  Inherited Insert(index, value);
end;

function TFhirConditionEvidenceList.Item(index: Integer): TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence(ObjectByIndex[index]);
end;

function TFhirConditionEvidenceList.Link: TFhirConditionEvidenceList;
begin
  result := TFhirConditionEvidenceList(inherited Link);
end;

procedure TFhirConditionEvidenceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConditionEvidenceList.SetItemByIndex(index: Integer; value: TFhirConditionEvidence);
begin
  assert(value is TFhirConditionEvidence);
  FhirConditionEvidences[index] := value;
end;

procedure TFhirConditionEvidenceList.SetItemN(index: Integer; value: TFhirConditionEvidence);
begin
  assert(value is TFhirConditionEvidence);
  ObjectByIndex[index] := value;
end;

{ TFhirCondition }

constructor TFhirCondition.Create;
begin
  inherited;
end;

destructor TFhirCondition.Destroy;
begin
  FIdentifierList.Free;
  FPatient.free;
  FEncounter.free;
  FAsserter.free;
  FDateRecorded.free;
  FCode.free;
  FCategory.free;
  FClinicalStatus.free;
  FVerificationStatus.free;
  FSeverity.free;
  FOnset.free;
  FAbatement.free;
  FStage.free;
  FEvidenceList.Free;
  FBodySiteList.Free;
  FNotes.free;
  inherited;
end;

function TFhirCondition.GetResourceType : TFhirResourceType;
begin
  result := frtCondition;
end;

procedure TFhirCondition.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirCondition(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirCondition(oSource).FIdentifierList);
  end;
  patient := TFhirCondition(oSource).patient.Clone;
  encounter := TFhirCondition(oSource).encounter.Clone;
  asserter := TFhirCondition(oSource).asserter.Clone;
  dateRecordedElement := TFhirCondition(oSource).dateRecordedElement.Clone;
  code := TFhirCondition(oSource).code.Clone;
  category := TFhirCondition(oSource).category.Clone;
  FClinicalStatus := TFhirCondition(oSource).FClinicalStatus.Link;
  FVerificationStatus := TFhirCondition(oSource).FVerificationStatus.Link;
  severity := TFhirCondition(oSource).severity.Clone;
  onset := TFhirCondition(oSource).onset.Clone;
  abatement := TFhirCondition(oSource).abatement.Clone;
  stage := TFhirCondition(oSource).stage.Clone;
  if (TFhirCondition(oSource).FEvidenceList = nil) then
  begin
    FEvidenceList.free;
    FEvidenceList := nil;
  end
  else
  begin
    FEvidenceList := TFhirConditionEvidenceList.Create;
    FEvidenceList.Assign(TFhirCondition(oSource).FEvidenceList);
  end;
  if (TFhirCondition(oSource).FBodySiteList = nil) then
  begin
    FBodySiteList.free;
    FBodySiteList := nil;
  end
  else
  begin
    FBodySiteList := TFhirCodeableConceptList.Create;
    FBodySiteList.Assign(TFhirCondition(oSource).FBodySiteList);
  end;
  notesElement := TFhirCondition(oSource).notesElement.Clone;
end;

procedure TFhirCondition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'asserter') Then
     list.add(FAsserter.Link);
  if (child_name = 'dateRecorded') Then
     list.add(FDateRecorded.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'clinicalStatus') Then
     list.add(FClinicalStatus.Link);
  if (child_name = 'verificationStatus') Then
     list.add(FVerificationStatus.Link);
  if (child_name = 'severity') Then
     list.add(FSeverity.Link);
  if (child_name = 'onset[x]') Then
     list.add(FOnset.Link);
  if (child_name = 'abatement[x]') Then
     list.add(FAbatement.Link);
  if (child_name = 'stage') Then
     list.add(FStage.Link);
  if (child_name = 'evidence') Then
    list.addAll(FEvidenceList);
  if (child_name = 'bodySite') Then
    list.addAll(FBodySiteList);
  if (child_name = 'notes') Then
     list.add(FNotes.Link);
end;

procedure TFhirCondition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'asserter', 'Reference(Practitioner|Patient)', FAsserter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateRecorded', 'date', FDateRecorded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'clinicalStatus', 'code', FClinicalStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'verificationStatus', 'code', FVerificationStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'severity', 'CodeableConcept', FSeverity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'onset[x]', 'dateTime|Age|Period|Range|string', FOnset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'abatement[x]', 'dateTime|Age|boolean|Period|Range|string', FAbatement.Link));{2}
  oList.add(TFHIRProperty.create(self, 'stage', '', FStage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'evidence', '', FEvidenceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySiteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'notes', 'string', FNotes.Link));{2}
end;

procedure TFhirCondition.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'asserter') then Asserter := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'dateRecorded') then DateRecordedElement := propValue as TFhirDate{5a}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'clinicalStatus') then ClinicalStatusElement := propValue as TFHIREnum
  else if (propName = 'verificationStatus') then VerificationStatusElement := propValue as TFHIREnum
  else if (propName = 'severity') then Severity := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('onset')) then Onset := propValue as TFhirType{4}
  else if (propName.startsWith('abatement')) then Abatement := propValue as TFhirType{4}
  else if (propName = 'stage') then Stage := propValue as TFhirConditionStage{4b}
  else if (propName = 'evidence') then EvidenceList.add(propValue as TFhirConditionEvidence){2}
  else if (propName = 'bodySite') then BodySiteList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'notes') then NotesElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirCondition.FhirType : string;
begin
  result := 'Condition';
end;

function TFhirCondition.Link : TFhirCondition;
begin
  result := TFhirCondition(inherited Link);
end;

function TFhirCondition.Clone : TFhirCondition;
begin
  result := TFhirCondition(inherited Clone);
end;

{ TFhirCondition }

Function TFhirCondition.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirCondition.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirCondition.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirCondition.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirCondition.SetAsserter(value : TFhirReference{Resource});
begin
  FAsserter.free;
  FAsserter := value;
end;

Procedure TFhirCondition.SetDateRecorded(value : TFhirDate);
begin
  FDateRecorded.free;
  FDateRecorded := value;
end;

Function TFhirCondition.GetDateRecordedST : TDateAndTime;
begin
  if FDateRecorded = nil then
    result := nil
  else
    result := FDateRecorded.value;
end;

Procedure TFhirCondition.SetDateRecordedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateRecorded = nil then
      FDateRecorded := TFhirDate.create;
    FDateRecorded.value := value
  end
  else if FDateRecorded <> nil then
    FDateRecorded.value := nil;
end;

Procedure TFhirCondition.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirCondition.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirCondition.SetClinicalStatus(value : TFhirEnum);
begin
  FClinicalStatus.free;
  FClinicalStatus := value;
end;

Function TFhirCondition.GetClinicalStatusST : TFhirValuesetConditionClinical;
begin
  if FClinicalStatus = nil then
    result := TFhirValuesetConditionClinical(0)
  else
    result := TFhirValuesetConditionClinical(StringArrayIndexOfSensitive(CODES_TFhirValuesetConditionClinical, FClinicalStatus.value));
end;

Procedure TFhirCondition.SetClinicalStatusST(value : TFhirValuesetConditionClinical);
begin
  if ord(value) = 0 then
    ClinicalStatusElement := nil
  else
    ClinicalStatusElement := TFhirEnum.create(CODES_TFhirValuesetConditionClinical[value]);
end;

Procedure TFhirCondition.SetVerificationStatus(value : TFhirEnum);
begin
  FVerificationStatus.free;
  FVerificationStatus := value;
end;

Function TFhirCondition.GetVerificationStatusST : TFhirConditionVerStatus;
begin
  if FVerificationStatus = nil then
    result := TFhirConditionVerStatus(0)
  else
    result := TFhirConditionVerStatus(StringArrayIndexOfSensitive(CODES_TFhirConditionVerStatus, FVerificationStatus.value));
end;

Procedure TFhirCondition.SetVerificationStatusST(value : TFhirConditionVerStatus);
begin
  if ord(value) = 0 then
    VerificationStatusElement := nil
  else
    VerificationStatusElement := TFhirEnum.create(CODES_TFhirConditionVerStatus[value]);
end;

Procedure TFhirCondition.SetSeverity(value : TFhirCodeableConcept);
begin
  FSeverity.free;
  FSeverity := value;
end;

Procedure TFhirCondition.SetOnset(value : TFhirType);
begin
  FOnset.free;
  FOnset := value;
end;

Procedure TFhirCondition.SetAbatement(value : TFhirType);
begin
  FAbatement.free;
  FAbatement := value;
end;

Procedure TFhirCondition.SetStage(value : TFhirConditionStage);
begin
  FStage.free;
  FStage := value;
end;

Function TFhirCondition.GetEvidenceList : TFhirConditionEvidenceList;
begin
  if FEvidenceList = nil then
    FEvidenceList := TFhirConditionEvidenceList.Create;
  result := FEvidenceList;
end;

Function TFhirCondition.GetHasEvidenceList : boolean;
begin
  result := (FEvidenceList <> nil) and (FEvidenceList.count > 0);
end;

Function TFhirCondition.GetBodySiteList : TFhirCodeableConceptList;
begin
  if FBodySiteList = nil then
    FBodySiteList := TFhirCodeableConceptList.Create;
  result := FBodySiteList;
end;

Function TFhirCondition.GetHasBodySiteList : boolean;
begin
  result := (FBodySiteList <> nil) and (FBodySiteList.count > 0);
end;

Procedure TFhirCondition.SetNotes(value : TFhirString);
begin
  FNotes.free;
  FNotes := value;
end;

Function TFhirCondition.GetNotesST : String;
begin
  if FNotes = nil then
    result := ''
  else
    result := FNotes.value;
end;

Procedure TFhirCondition.SetNotesST(value : String);
begin
  if value <> '' then
  begin
    if FNotes = nil then
      FNotes := TFhirString.create;
    FNotes.value := value
  end
  else if FNotes <> nil then
    FNotes.value := '';
end;


{ TFhirConditionListEnumerator }

Constructor TFhirConditionListEnumerator.Create(list : TFhirConditionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConditionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConditionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConditionListEnumerator.GetCurrent : TFhirCondition;
begin
  Result := FList[FIndex];
end;


{ TFhirConditionList }
procedure TFhirConditionList.AddItem(value: TFhirCondition);
begin
  assert(value.ClassName = 'TFhirCondition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCondition');
  add(value);
end;


function TFhirConditionList.Append: TFhirCondition;
begin
  result := TFhirCondition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionList.ClearItems;
begin
  Clear;
end;

function TFhirConditionList.GetEnumerator : TFhirConditionListEnumerator;
begin
  result := TFhirConditionListEnumerator.Create(self.link);
end;

function TFhirConditionList.Clone: TFhirConditionList;
begin
  result := TFhirConditionList(inherited Clone);
end;

function TFhirConditionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConditionList.GetItemN(index: Integer): TFhirCondition;
begin
  result := TFhirCondition(ObjectByIndex[index]);
end;

function TFhirConditionList.IndexOf(value: TFhirCondition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConditionList.Insert(index: Integer): TFhirCondition;
begin
  result := TFhirCondition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConditionList.InsertItem(index: Integer; value: TFhirCondition);
begin
  assert(value is TFhirCondition);
  Inherited Insert(index, value);
end;

function TFhirConditionList.Item(index: Integer): TFhirCondition;
begin
  result := TFhirCondition(ObjectByIndex[index]);
end;

function TFhirConditionList.Link: TFhirConditionList;
begin
  result := TFhirConditionList(inherited Link);
end;

procedure TFhirConditionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConditionList.SetItemByIndex(index: Integer; value: TFhirCondition);
begin
  assert(value is TFhirCondition);
  FhirConditions[index] := value;
end;

procedure TFhirConditionList.SetItemN(index: Integer; value: TFhirCondition);
begin
  assert(value is TFhirCondition);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceContact }

constructor TFhirConformanceContact.Create;
begin
  inherited;
end;

destructor TFhirConformanceContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirConformanceContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirConformanceContact(oSource).nameElement.Clone;
  if (TFhirConformanceContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirConformanceContact(oSource).FTelecomList);
  end;
end;

procedure TFhirConformanceContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirConformanceContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirConformanceContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirConformanceContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirConformanceContact.Link : TFhirConformanceContact;
begin
  result := TFhirConformanceContact(inherited Link);
end;

function TFhirConformanceContact.Clone : TFhirConformanceContact;
begin
  result := TFhirConformanceContact(inherited Clone);
end;

function TFhirConformanceContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceContact)) then
    result := false
  else
  begin
    o := TFhirConformanceContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirConformanceContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceContact)) then
    result := false
  else
  begin
    o := TFhirConformanceContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirConformanceContact }

Procedure TFhirConformanceContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConformanceContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConformanceContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirConformanceContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirConformanceContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirConformanceContactListEnumerator }

Constructor TFhirConformanceContactListEnumerator.Create(list : TFhirConformanceContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceContactListEnumerator.GetCurrent : TFhirConformanceContact;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceContactList }
procedure TFhirConformanceContactList.AddItem(value: TFhirConformanceContact);
begin
  assert(value.ClassName = 'TFhirConformanceContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceContact');
  add(value);
end;


function TFhirConformanceContactList.Append: TFhirConformanceContact;
begin
  result := TFhirConformanceContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceContactList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceContactList.GetEnumerator : TFhirConformanceContactListEnumerator;
begin
  result := TFhirConformanceContactListEnumerator.Create(self.link);
end;

function TFhirConformanceContactList.Clone: TFhirConformanceContactList;
begin
  result := TFhirConformanceContactList(inherited Clone);
end;

function TFhirConformanceContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceContactList.GetItemN(index: Integer): TFhirConformanceContact;
begin
  result := TFhirConformanceContact(ObjectByIndex[index]);
end;

function TFhirConformanceContactList.IndexOf(value: TFhirConformanceContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceContactList.Insert(index: Integer): TFhirConformanceContact;
begin
  result := TFhirConformanceContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceContactList.InsertItem(index: Integer; value: TFhirConformanceContact);
begin
  assert(value is TFhirConformanceContact);
  Inherited Insert(index, value);
end;

function TFhirConformanceContactList.Item(index: Integer): TFhirConformanceContact;
begin
  result := TFhirConformanceContact(ObjectByIndex[index]);
end;

function TFhirConformanceContactList.Link: TFhirConformanceContactList;
begin
  result := TFhirConformanceContactList(inherited Link);
end;

procedure TFhirConformanceContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceContactList.SetItemByIndex(index: Integer; value: TFhirConformanceContact);
begin
  assert(value is TFhirConformanceContact);
  FhirConformanceContacts[index] := value;
end;

procedure TFhirConformanceContactList.SetItemN(index: Integer; value: TFhirConformanceContact);
begin
  assert(value is TFhirConformanceContact);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceSoftware }

constructor TFhirConformanceSoftware.Create;
begin
  inherited;
end;

destructor TFhirConformanceSoftware.Destroy;
begin
  FName.free;
  FVersion.free;
  FReleaseDate.free;
  inherited;
end;

procedure TFhirConformanceSoftware.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirConformanceSoftware(oSource).nameElement.Clone;
  versionElement := TFhirConformanceSoftware(oSource).versionElement.Clone;
  releaseDateElement := TFhirConformanceSoftware(oSource).releaseDateElement.Clone;
end;

procedure TFhirConformanceSoftware.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'releaseDate') Then
     list.add(FReleaseDate.Link);
end;

procedure TFhirConformanceSoftware.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'releaseDate', 'dateTime', FReleaseDate.Link));{2}
end;

procedure TFhirConformanceSoftware.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'releaseDate') then ReleaseDateElement := propValue as TFhirDateTime{5a}
  else inherited;
end;

function TFhirConformanceSoftware.FhirType : string;
begin
  result := 'software';
end;

function TFhirConformanceSoftware.Link : TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware(inherited Link);
end;

function TFhirConformanceSoftware.Clone : TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware(inherited Clone);
end;

function TFhirConformanceSoftware.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceSoftware;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceSoftware)) then
    result := false
  else
  begin
    o := TFhirConformanceSoftware(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(versionElement, o.versionElement, true) and 
      compareDeep(releaseDateElement, o.releaseDateElement, true);
  end;
end;

function TFhirConformanceSoftware.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceSoftware;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceSoftware)) then
    result := false
  else
  begin
    o := TFhirConformanceSoftware(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(versionElement, o.versionElement, true) and 
      compareValues(releaseDateElement, o.releaseDateElement, true);
  end;
end;

{ TFhirConformanceSoftware }

Procedure TFhirConformanceSoftware.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConformanceSoftware.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConformanceSoftware.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirConformanceSoftware.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirConformanceSoftware.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirConformanceSoftware.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirConformanceSoftware.SetReleaseDate(value : TFhirDateTime);
begin
  FReleaseDate.free;
  FReleaseDate := value;
end;

Function TFhirConformanceSoftware.GetReleaseDateST : TDateAndTime;
begin
  if FReleaseDate = nil then
    result := nil
  else
    result := FReleaseDate.value;
end;

Procedure TFhirConformanceSoftware.SetReleaseDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FReleaseDate = nil then
      FReleaseDate := TFhirDateTime.create;
    FReleaseDate.value := value
  end
  else if FReleaseDate <> nil then
    FReleaseDate.value := nil;
end;


{ TFhirConformanceSoftwareListEnumerator }

Constructor TFhirConformanceSoftwareListEnumerator.Create(list : TFhirConformanceSoftwareList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceSoftwareListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceSoftwareListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceSoftwareListEnumerator.GetCurrent : TFhirConformanceSoftware;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceSoftwareList }
procedure TFhirConformanceSoftwareList.AddItem(value: TFhirConformanceSoftware);
begin
  assert(value.ClassName = 'TFhirConformanceSoftware', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceSoftware');
  add(value);
end;


function TFhirConformanceSoftwareList.Append: TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceSoftwareList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceSoftwareList.GetEnumerator : TFhirConformanceSoftwareListEnumerator;
begin
  result := TFhirConformanceSoftwareListEnumerator.Create(self.link);
end;

function TFhirConformanceSoftwareList.Clone: TFhirConformanceSoftwareList;
begin
  result := TFhirConformanceSoftwareList(inherited Clone);
end;

function TFhirConformanceSoftwareList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceSoftwareList.GetItemN(index: Integer): TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware(ObjectByIndex[index]);
end;

function TFhirConformanceSoftwareList.IndexOf(value: TFhirConformanceSoftware): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceSoftwareList.Insert(index: Integer): TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceSoftwareList.InsertItem(index: Integer; value: TFhirConformanceSoftware);
begin
  assert(value is TFhirConformanceSoftware);
  Inherited Insert(index, value);
end;

function TFhirConformanceSoftwareList.Item(index: Integer): TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware(ObjectByIndex[index]);
end;

function TFhirConformanceSoftwareList.Link: TFhirConformanceSoftwareList;
begin
  result := TFhirConformanceSoftwareList(inherited Link);
end;

procedure TFhirConformanceSoftwareList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceSoftwareList.SetItemByIndex(index: Integer; value: TFhirConformanceSoftware);
begin
  assert(value is TFhirConformanceSoftware);
  FhirConformanceSoftwares[index] := value;
end;

procedure TFhirConformanceSoftwareList.SetItemN(index: Integer; value: TFhirConformanceSoftware);
begin
  assert(value is TFhirConformanceSoftware);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceImplementation }

constructor TFhirConformanceImplementation.Create;
begin
  inherited;
end;

destructor TFhirConformanceImplementation.Destroy;
begin
  FDescription.free;
  FUrl.free;
  inherited;
end;

procedure TFhirConformanceImplementation.Assign(oSource : TAdvObject);
begin
  inherited;
  descriptionElement := TFhirConformanceImplementation(oSource).descriptionElement.Clone;
  urlElement := TFhirConformanceImplementation(oSource).urlElement.Clone;
end;

procedure TFhirConformanceImplementation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
end;

procedure TFhirConformanceImplementation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
end;

procedure TFhirConformanceImplementation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirConformanceImplementation.FhirType : string;
begin
  result := 'implementation';
end;

function TFhirConformanceImplementation.Link : TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation(inherited Link);
end;

function TFhirConformanceImplementation.Clone : TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation(inherited Clone);
end;

function TFhirConformanceImplementation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceImplementation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceImplementation)) then
    result := false
  else
  begin
    o := TFhirConformanceImplementation(other);
    result := compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(urlElement, o.urlElement, true);
  end;
end;

function TFhirConformanceImplementation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceImplementation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceImplementation)) then
    result := false
  else
  begin
    o := TFhirConformanceImplementation(other);
    result := compareValues(descriptionElement, o.descriptionElement, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirConformanceImplementation }

Procedure TFhirConformanceImplementation.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirConformanceImplementation.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirConformanceImplementation.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirConformanceImplementation.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirConformanceImplementation.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirConformanceImplementation.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;


{ TFhirConformanceImplementationListEnumerator }

Constructor TFhirConformanceImplementationListEnumerator.Create(list : TFhirConformanceImplementationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceImplementationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceImplementationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceImplementationListEnumerator.GetCurrent : TFhirConformanceImplementation;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceImplementationList }
procedure TFhirConformanceImplementationList.AddItem(value: TFhirConformanceImplementation);
begin
  assert(value.ClassName = 'TFhirConformanceImplementation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceImplementation');
  add(value);
end;


function TFhirConformanceImplementationList.Append: TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceImplementationList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceImplementationList.GetEnumerator : TFhirConformanceImplementationListEnumerator;
begin
  result := TFhirConformanceImplementationListEnumerator.Create(self.link);
end;

function TFhirConformanceImplementationList.Clone: TFhirConformanceImplementationList;
begin
  result := TFhirConformanceImplementationList(inherited Clone);
end;

function TFhirConformanceImplementationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceImplementationList.GetItemN(index: Integer): TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation(ObjectByIndex[index]);
end;

function TFhirConformanceImplementationList.IndexOf(value: TFhirConformanceImplementation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceImplementationList.Insert(index: Integer): TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceImplementationList.InsertItem(index: Integer; value: TFhirConformanceImplementation);
begin
  assert(value is TFhirConformanceImplementation);
  Inherited Insert(index, value);
end;

function TFhirConformanceImplementationList.Item(index: Integer): TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation(ObjectByIndex[index]);
end;

function TFhirConformanceImplementationList.Link: TFhirConformanceImplementationList;
begin
  result := TFhirConformanceImplementationList(inherited Link);
end;

procedure TFhirConformanceImplementationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceImplementationList.SetItemByIndex(index: Integer; value: TFhirConformanceImplementation);
begin
  assert(value is TFhirConformanceImplementation);
  FhirConformanceImplementations[index] := value;
end;

procedure TFhirConformanceImplementationList.SetItemN(index: Integer; value: TFhirConformanceImplementation);
begin
  assert(value is TFhirConformanceImplementation);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRest }

constructor TFhirConformanceRest.Create;
begin
  inherited;
end;

destructor TFhirConformanceRest.Destroy;
begin
  FMode.free;
  FDocumentation.free;
  FSecurity.free;
  FResourceList.Free;
  FInteractionList.Free;
  FTransactionMode.free;
  FSearchParamList.Free;
  FOperationList.Free;
  FCompartmentList.Free;
  inherited;
end;

procedure TFhirConformanceRest.Assign(oSource : TAdvObject);
begin
  inherited;
  FMode := TFhirConformanceRest(oSource).FMode.Link;
  documentationElement := TFhirConformanceRest(oSource).documentationElement.Clone;
  security := TFhirConformanceRest(oSource).security.Clone;
  if (TFhirConformanceRest(oSource).FResourceList = nil) then
  begin
    FResourceList.free;
    FResourceList := nil;
  end
  else
  begin
    FResourceList := TFhirConformanceRestResourceList.Create;
    FResourceList.Assign(TFhirConformanceRest(oSource).FResourceList);
  end;
  if (TFhirConformanceRest(oSource).FInteractionList = nil) then
  begin
    FInteractionList.free;
    FInteractionList := nil;
  end
  else
  begin
    FInteractionList := TFhirConformanceRestInteractionList.Create;
    FInteractionList.Assign(TFhirConformanceRest(oSource).FInteractionList);
  end;
  FTransactionMode := TFhirConformanceRest(oSource).FTransactionMode.Link;
  if (TFhirConformanceRest(oSource).FSearchParamList = nil) then
  begin
    FSearchParamList.free;
    FSearchParamList := nil;
  end
  else
  begin
    FSearchParamList := TFhirConformanceRestResourceSearchParamList.Create;
    FSearchParamList.Assign(TFhirConformanceRest(oSource).FSearchParamList);
  end;
  if (TFhirConformanceRest(oSource).FOperationList = nil) then
  begin
    FOperationList.free;
    FOperationList := nil;
  end
  else
  begin
    FOperationList := TFhirConformanceRestOperationList.Create;
    FOperationList.Assign(TFhirConformanceRest(oSource).FOperationList);
  end;
  if (TFhirConformanceRest(oSource).FCompartmentList = nil) then
  begin
    FCompartmentList.free;
    FCompartmentList := nil;
  end
  else
  begin
    FCompartmentList := TFhirUriList.Create;
    FCompartmentList.Assign(TFhirConformanceRest(oSource).FCompartmentList);
  end;
end;

procedure TFhirConformanceRest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
  if (child_name = 'security') Then
     list.add(FSecurity.Link);
  if (child_name = 'resource') Then
    list.addAll(FResourceList);
  if (child_name = 'interaction') Then
    list.addAll(FInteractionList);
  if (child_name = 'transactionMode') Then
     list.add(FTransactionMode.Link);
  if (child_name = 'searchParam') Then
    list.addAll(FSearchParamList);
  if (child_name = 'operation') Then
    list.addAll(FOperationList);
  if (child_name = 'compartment') Then
    list.addAll(FCompartmentList);
end;

procedure TFhirConformanceRest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'security', '', FSecurity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', '', FResourceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'interaction', '', FInteractionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'transactionMode', 'code', FTransactionMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'searchParam', '@Conformance.rest.resource.searchParam', FSearchParamList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'operation', '', FOperationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'compartment', 'uri', FCompartmentList.Link)){3};
end;

procedure TFhirConformanceRest.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else if (propName = 'security') then Security := propValue as TFhirConformanceRestSecurity{4b}
  else if (propName = 'resource') then ResourceList.add(propValue as TFhirConformanceRestResource){2}
  else if (propName = 'interaction') then InteractionList.add(propValue as TFhirConformanceRestInteraction){2}
  else if (propName = 'transactionMode') then TransactionModeElement := propValue as TFHIREnum
  else if (propName = 'searchParam') then SearchParamList.add(propValue as TFhirConformanceRestResourceSearchParam){2}
  else if (propName = 'operation') then OperationList.add(propValue as TFhirConformanceRestOperation){2}
  else if (propName = 'compartment') then CompartmentList.add(propValue as TFhirUri){2}
  else inherited;
end;

function TFhirConformanceRest.FhirType : string;
begin
  result := 'rest';
end;

function TFhirConformanceRest.Link : TFhirConformanceRest;
begin
  result := TFhirConformanceRest(inherited Link);
end;

function TFhirConformanceRest.Clone : TFhirConformanceRest;
begin
  result := TFhirConformanceRest(inherited Clone);
end;

function TFhirConformanceRest.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRest;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRest)) then
    result := false
  else
  begin
    o := TFhirConformanceRest(other);
    result := compareDeep(modeElement, o.modeElement, true) and compareDeep(documentationElement, o.documentationElement, true) and 
      compareDeep(securityElement, o.securityElement, true) and compareDeep(resourceList, o.resourceList, true) and 
      compareDeep(interactionList, o.interactionList, true) and compareDeep(transactionModeElement, o.transactionModeElement, true) and 
      compareDeep(searchParamList, o.searchParamList, true) and compareDeep(operationList, o.operationList, true) and 
      compareDeep(compartmentList, o.compartmentList, true);
  end;
end;

function TFhirConformanceRest.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRest;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRest)) then
    result := false
  else
  begin
    o := TFhirConformanceRest(other);
    result := compareValues(modeElement, o.modeElement, true) and compareValues(documentationElement, o.documentationElement, true) and 
      compareValues(transactionModeElement, o.transactionModeElement, true) and compareValues(compartmentList, o.compartmentList, true);
  end;
end;

{ TFhirConformanceRest }

Procedure TFhirConformanceRest.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirConformanceRest.GetModeST : TFhirRestfulConformanceMode;
begin
  if FMode = nil then
    result := TFhirRestfulConformanceMode(0)
  else
    result := TFhirRestfulConformanceMode(StringArrayIndexOfSensitive(CODES_TFhirRestfulConformanceMode, FMode.value));
end;

Procedure TFhirConformanceRest.SetModeST(value : TFhirRestfulConformanceMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirRestfulConformanceMode[value]);
end;

Procedure TFhirConformanceRest.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceRest.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceRest.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;

Procedure TFhirConformanceRest.SetSecurity(value : TFhirConformanceRestSecurity);
begin
  FSecurity.free;
  FSecurity := value;
end;

Function TFhirConformanceRest.GetResourceList : TFhirConformanceRestResourceList;
begin
  if FResourceList = nil then
    FResourceList := TFhirConformanceRestResourceList.Create;
  result := FResourceList;
end;

Function TFhirConformanceRest.GetHasResourceList : boolean;
begin
  result := (FResourceList <> nil) and (FResourceList.count > 0);
end;

Function TFhirConformanceRest.GetInteractionList : TFhirConformanceRestInteractionList;
begin
  if FInteractionList = nil then
    FInteractionList := TFhirConformanceRestInteractionList.Create;
  result := FInteractionList;
end;

Function TFhirConformanceRest.GetHasInteractionList : boolean;
begin
  result := (FInteractionList <> nil) and (FInteractionList.count > 0);
end;

Procedure TFhirConformanceRest.SetTransactionMode(value : TFhirEnum);
begin
  FTransactionMode.free;
  FTransactionMode := value;
end;

Function TFhirConformanceRest.GetTransactionModeST : TFhirTransactionMode;
begin
  if FTransactionMode = nil then
    result := TFhirTransactionMode(0)
  else
    result := TFhirTransactionMode(StringArrayIndexOfSensitive(CODES_TFhirTransactionMode, FTransactionMode.value));
end;

Procedure TFhirConformanceRest.SetTransactionModeST(value : TFhirTransactionMode);
begin
  if ord(value) = 0 then
    TransactionModeElement := nil
  else
    TransactionModeElement := TFhirEnum.create(CODES_TFhirTransactionMode[value]);
end;

Function TFhirConformanceRest.GetSearchParamList : TFhirConformanceRestResourceSearchParamList;
begin
  if FSearchParamList = nil then
    FSearchParamList := TFhirConformanceRestResourceSearchParamList.Create;
  result := FSearchParamList;
end;

Function TFhirConformanceRest.GetHasSearchParamList : boolean;
begin
  result := (FSearchParamList <> nil) and (FSearchParamList.count > 0);
end;

Function TFhirConformanceRest.GetOperationList : TFhirConformanceRestOperationList;
begin
  if FOperationList = nil then
    FOperationList := TFhirConformanceRestOperationList.Create;
  result := FOperationList;
end;

Function TFhirConformanceRest.GetHasOperationList : boolean;
begin
  result := (FOperationList <> nil) and (FOperationList.count > 0);
end;

Function TFhirConformanceRest.GetCompartmentList : TFhirUriList;
begin
  if FCompartmentList = nil then
    FCompartmentList := TFhirUriList.Create;
  result := FCompartmentList;
end;

Function TFhirConformanceRest.GetHasCompartmentList : boolean;
begin
  result := (FCompartmentList <> nil) and (FCompartmentList.count > 0);
end;


{ TFhirConformanceRestListEnumerator }

Constructor TFhirConformanceRestListEnumerator.Create(list : TFhirConformanceRestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestListEnumerator.GetCurrent : TFhirConformanceRest;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestList }
procedure TFhirConformanceRestList.AddItem(value: TFhirConformanceRest);
begin
  assert(value.ClassName = 'TFhirConformanceRest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRest');
  add(value);
end;


function TFhirConformanceRestList.Append: TFhirConformanceRest;
begin
  result := TFhirConformanceRest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestList.GetEnumerator : TFhirConformanceRestListEnumerator;
begin
  result := TFhirConformanceRestListEnumerator.Create(self.link);
end;

function TFhirConformanceRestList.Clone: TFhirConformanceRestList;
begin
  result := TFhirConformanceRestList(inherited Clone);
end;

function TFhirConformanceRestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestList.GetItemN(index: Integer): TFhirConformanceRest;
begin
  result := TFhirConformanceRest(ObjectByIndex[index]);
end;

function TFhirConformanceRestList.IndexOf(value: TFhirConformanceRest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestList.Insert(index: Integer): TFhirConformanceRest;
begin
  result := TFhirConformanceRest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestList.InsertItem(index: Integer; value: TFhirConformanceRest);
begin
  assert(value is TFhirConformanceRest);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestList.Item(index: Integer): TFhirConformanceRest;
begin
  result := TFhirConformanceRest(ObjectByIndex[index]);
end;

function TFhirConformanceRestList.Link: TFhirConformanceRestList;
begin
  result := TFhirConformanceRestList(inherited Link);
end;

procedure TFhirConformanceRestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestList.SetItemByIndex(index: Integer; value: TFhirConformanceRest);
begin
  assert(value is TFhirConformanceRest);
  FhirConformanceRests[index] := value;
end;

procedure TFhirConformanceRestList.SetItemN(index: Integer; value: TFhirConformanceRest);
begin
  assert(value is TFhirConformanceRest);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestSecurity }

constructor TFhirConformanceRestSecurity.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestSecurity.Destroy;
begin
  FCors.free;
  FServiceList.Free;
  FDescription.free;
  FCertificateList.Free;
  inherited;
end;

procedure TFhirConformanceRestSecurity.Assign(oSource : TAdvObject);
begin
  inherited;
  corsElement := TFhirConformanceRestSecurity(oSource).corsElement.Clone;
  if (TFhirConformanceRestSecurity(oSource).FServiceList = nil) then
  begin
    FServiceList.free;
    FServiceList := nil;
  end
  else
  begin
    FServiceList := TFhirCodeableConceptList.Create;
    FServiceList.Assign(TFhirConformanceRestSecurity(oSource).FServiceList);
  end;
  descriptionElement := TFhirConformanceRestSecurity(oSource).descriptionElement.Clone;
  if (TFhirConformanceRestSecurity(oSource).FCertificateList = nil) then
  begin
    FCertificateList.free;
    FCertificateList := nil;
  end
  else
  begin
    FCertificateList := TFhirConformanceRestSecurityCertificateList.Create;
    FCertificateList.Assign(TFhirConformanceRestSecurity(oSource).FCertificateList);
  end;
end;

procedure TFhirConformanceRestSecurity.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'cors') Then
     list.add(FCors.Link);
  if (child_name = 'service') Then
    list.addAll(FServiceList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'certificate') Then
    list.addAll(FCertificateList);
end;

procedure TFhirConformanceRestSecurity.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'cors', 'boolean', FCors.Link));{2}
  oList.add(TFHIRProperty.create(self, 'service', 'CodeableConcept', FServiceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'certificate', '', FCertificateList.Link)){3};
end;

procedure TFhirConformanceRestSecurity.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'cors') then CorsElement := propValue as TFhirBoolean{5a}
  else if (propName = 'service') then ServiceList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'certificate') then CertificateList.add(propValue as TFhirConformanceRestSecurityCertificate){2}
  else inherited;
end;

function TFhirConformanceRestSecurity.FhirType : string;
begin
  result := 'security';
end;

function TFhirConformanceRestSecurity.Link : TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity(inherited Link);
end;

function TFhirConformanceRestSecurity.Clone : TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity(inherited Clone);
end;

function TFhirConformanceRestSecurity.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestSecurity;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestSecurity)) then
    result := false
  else
  begin
    o := TFhirConformanceRestSecurity(other);
    result := compareDeep(corsElement, o.corsElement, true) and compareDeep(serviceList, o.serviceList, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(certificateList, o.certificateList, true);
  end;
end;

function TFhirConformanceRestSecurity.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestSecurity;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestSecurity)) then
    result := false
  else
  begin
    o := TFhirConformanceRestSecurity(other);
    result := compareValues(corsElement, o.corsElement, true) and compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirConformanceRestSecurity }

Procedure TFhirConformanceRestSecurity.SetCors(value : TFhirBoolean);
begin
  FCors.free;
  FCors := value;
end;

Function TFhirConformanceRestSecurity.GetCorsST : Boolean;
begin
  if FCors = nil then
    result := false
  else
    result := FCors.value;
end;

Procedure TFhirConformanceRestSecurity.SetCorsST(value : Boolean);
begin
  if FCors = nil then
    FCors := TFhirBoolean.create;
  FCors.value := value
end;

Function TFhirConformanceRestSecurity.GetServiceList : TFhirCodeableConceptList;
begin
  if FServiceList = nil then
    FServiceList := TFhirCodeableConceptList.Create;
  result := FServiceList;
end;

Function TFhirConformanceRestSecurity.GetHasServiceList : boolean;
begin
  result := (FServiceList <> nil) and (FServiceList.count > 0);
end;

Procedure TFhirConformanceRestSecurity.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirConformanceRestSecurity.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirConformanceRestSecurity.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirConformanceRestSecurity.GetCertificateList : TFhirConformanceRestSecurityCertificateList;
begin
  if FCertificateList = nil then
    FCertificateList := TFhirConformanceRestSecurityCertificateList.Create;
  result := FCertificateList;
end;

Function TFhirConformanceRestSecurity.GetHasCertificateList : boolean;
begin
  result := (FCertificateList <> nil) and (FCertificateList.count > 0);
end;


{ TFhirConformanceRestSecurityListEnumerator }

Constructor TFhirConformanceRestSecurityListEnumerator.Create(list : TFhirConformanceRestSecurityList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestSecurityListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestSecurityListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestSecurityListEnumerator.GetCurrent : TFhirConformanceRestSecurity;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestSecurityList }
procedure TFhirConformanceRestSecurityList.AddItem(value: TFhirConformanceRestSecurity);
begin
  assert(value.ClassName = 'TFhirConformanceRestSecurity', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestSecurity');
  add(value);
end;


function TFhirConformanceRestSecurityList.Append: TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestSecurityList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestSecurityList.GetEnumerator : TFhirConformanceRestSecurityListEnumerator;
begin
  result := TFhirConformanceRestSecurityListEnumerator.Create(self.link);
end;

function TFhirConformanceRestSecurityList.Clone: TFhirConformanceRestSecurityList;
begin
  result := TFhirConformanceRestSecurityList(inherited Clone);
end;

function TFhirConformanceRestSecurityList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestSecurityList.GetItemN(index: Integer): TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity(ObjectByIndex[index]);
end;

function TFhirConformanceRestSecurityList.IndexOf(value: TFhirConformanceRestSecurity): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestSecurityList.Insert(index: Integer): TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestSecurityList.InsertItem(index: Integer; value: TFhirConformanceRestSecurity);
begin
  assert(value is TFhirConformanceRestSecurity);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestSecurityList.Item(index: Integer): TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity(ObjectByIndex[index]);
end;

function TFhirConformanceRestSecurityList.Link: TFhirConformanceRestSecurityList;
begin
  result := TFhirConformanceRestSecurityList(inherited Link);
end;

procedure TFhirConformanceRestSecurityList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestSecurityList.SetItemByIndex(index: Integer; value: TFhirConformanceRestSecurity);
begin
  assert(value is TFhirConformanceRestSecurity);
  FhirConformanceRestSecurities[index] := value;
end;

procedure TFhirConformanceRestSecurityList.SetItemN(index: Integer; value: TFhirConformanceRestSecurity);
begin
  assert(value is TFhirConformanceRestSecurity);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestSecurityCertificate }

constructor TFhirConformanceRestSecurityCertificate.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestSecurityCertificate.Destroy;
begin
  FType_.free;
  FBlob.free;
  inherited;
end;

procedure TFhirConformanceRestSecurityCertificate.Assign(oSource : TAdvObject);
begin
  inherited;
  type_Element := TFhirConformanceRestSecurityCertificate(oSource).type_Element.Clone;
  blobElement := TFhirConformanceRestSecurityCertificate(oSource).blobElement.Clone;
end;

procedure TFhirConformanceRestSecurityCertificate.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'blob') Then
     list.add(FBlob.Link);
end;

procedure TFhirConformanceRestSecurityCertificate.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'blob', 'base64Binary', FBlob.Link));{2}
end;

procedure TFhirConformanceRestSecurityCertificate.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then
    if propValue is TFHIRCode then
      Type_Element := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      Type_Element := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'blob') then BlobElement := propValue as TFhirBase64Binary{5a}
  else inherited;
end;

function TFhirConformanceRestSecurityCertificate.FhirType : string;
begin
  result := 'certificate';
end;

function TFhirConformanceRestSecurityCertificate.Link : TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate(inherited Link);
end;

function TFhirConformanceRestSecurityCertificate.Clone : TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate(inherited Clone);
end;

function TFhirConformanceRestSecurityCertificate.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestSecurityCertificate;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestSecurityCertificate)) then
    result := false
  else
  begin
    o := TFhirConformanceRestSecurityCertificate(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(blobElement, o.blobElement, true);
  end;
end;

function TFhirConformanceRestSecurityCertificate.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestSecurityCertificate;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestSecurityCertificate)) then
    result := false
  else
  begin
    o := TFhirConformanceRestSecurityCertificate(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(blobElement, o.blobElement, true);
  end;
end;

{ TFhirConformanceRestSecurityCertificate }

Procedure TFhirConformanceRestSecurityCertificate.SetType_(value : TFhirCode);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirConformanceRestSecurityCertificate.GetType_ST : String;
begin
  if FType_ = nil then
    result := ''
  else
    result := FType_.value;
end;

Procedure TFhirConformanceRestSecurityCertificate.SetType_ST(value : String);
begin
  if value <> '' then
  begin
    if FType_ = nil then
      FType_ := TFhirCode.create;
    FType_.value := value
  end
  else if FType_ <> nil then
    FType_.value := '';
end;

Procedure TFhirConformanceRestSecurityCertificate.SetBlob(value : TFhirBase64Binary);
begin
  FBlob.free;
  FBlob := value;
end;

Function TFhirConformanceRestSecurityCertificate.GetBlobST : TBytes;
begin
  if FBlob = nil then
    result := nil
  else
    result := FBlob.value;
end;

Procedure TFhirConformanceRestSecurityCertificate.SetBlobST(value : TBytes);
begin
  if value <> nil then
  begin
    if FBlob = nil then
      FBlob := TFhirBase64Binary.create;
    FBlob.value := value
  end
  else if FBlob <> nil then
    FBlob.value := nil;
end;


{ TFhirConformanceRestSecurityCertificateListEnumerator }

Constructor TFhirConformanceRestSecurityCertificateListEnumerator.Create(list : TFhirConformanceRestSecurityCertificateList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestSecurityCertificateListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestSecurityCertificateListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestSecurityCertificateListEnumerator.GetCurrent : TFhirConformanceRestSecurityCertificate;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestSecurityCertificateList }
procedure TFhirConformanceRestSecurityCertificateList.AddItem(value: TFhirConformanceRestSecurityCertificate);
begin
  assert(value.ClassName = 'TFhirConformanceRestSecurityCertificate', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestSecurityCertificate');
  add(value);
end;


function TFhirConformanceRestSecurityCertificateList.Append: TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestSecurityCertificateList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestSecurityCertificateList.GetEnumerator : TFhirConformanceRestSecurityCertificateListEnumerator;
begin
  result := TFhirConformanceRestSecurityCertificateListEnumerator.Create(self.link);
end;

function TFhirConformanceRestSecurityCertificateList.Clone: TFhirConformanceRestSecurityCertificateList;
begin
  result := TFhirConformanceRestSecurityCertificateList(inherited Clone);
end;

function TFhirConformanceRestSecurityCertificateList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestSecurityCertificateList.GetItemN(index: Integer): TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate(ObjectByIndex[index]);
end;

function TFhirConformanceRestSecurityCertificateList.IndexOf(value: TFhirConformanceRestSecurityCertificate): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestSecurityCertificateList.Insert(index: Integer): TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestSecurityCertificateList.InsertItem(index: Integer; value: TFhirConformanceRestSecurityCertificate);
begin
  assert(value is TFhirConformanceRestSecurityCertificate);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestSecurityCertificateList.Item(index: Integer): TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate(ObjectByIndex[index]);
end;

function TFhirConformanceRestSecurityCertificateList.Link: TFhirConformanceRestSecurityCertificateList;
begin
  result := TFhirConformanceRestSecurityCertificateList(inherited Link);
end;

procedure TFhirConformanceRestSecurityCertificateList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestSecurityCertificateList.SetItemByIndex(index: Integer; value: TFhirConformanceRestSecurityCertificate);
begin
  assert(value is TFhirConformanceRestSecurityCertificate);
  FhirConformanceRestSecurityCertificates[index] := value;
end;

procedure TFhirConformanceRestSecurityCertificateList.SetItemN(index: Integer; value: TFhirConformanceRestSecurityCertificate);
begin
  assert(value is TFhirConformanceRestSecurityCertificate);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestResource }

constructor TFhirConformanceRestResource.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestResource.Destroy;
begin
  FType_.free;
  FProfile.free;
  FInteractionList.Free;
  FVersioning.free;
  FReadHistory.free;
  FUpdateCreate.free;
  FConditionalCreate.free;
  FConditionalUpdate.free;
  FConditionalDelete.free;
  FSearchIncludeList.Free;
  FSearchRevIncludeList.Free;
  FSearchParamList.Free;
  inherited;
end;

procedure TFhirConformanceRestResource.Assign(oSource : TAdvObject);
begin
  inherited;
  type_Element := TFhirConformanceRestResource(oSource).type_Element.Clone;
  profile := TFhirConformanceRestResource(oSource).profile.Clone;
  if (TFhirConformanceRestResource(oSource).FInteractionList = nil) then
  begin
    FInteractionList.free;
    FInteractionList := nil;
  end
  else
  begin
    FInteractionList := TFhirConformanceRestResourceInteractionList.Create;
    FInteractionList.Assign(TFhirConformanceRestResource(oSource).FInteractionList);
  end;
  FVersioning := TFhirConformanceRestResource(oSource).FVersioning.Link;
  readHistoryElement := TFhirConformanceRestResource(oSource).readHistoryElement.Clone;
  updateCreateElement := TFhirConformanceRestResource(oSource).updateCreateElement.Clone;
  conditionalCreateElement := TFhirConformanceRestResource(oSource).conditionalCreateElement.Clone;
  conditionalUpdateElement := TFhirConformanceRestResource(oSource).conditionalUpdateElement.Clone;
  FConditionalDelete := TFhirConformanceRestResource(oSource).FConditionalDelete.Link;
  if (TFhirConformanceRestResource(oSource).FSearchIncludeList = nil) then
  begin
    FSearchIncludeList.free;
    FSearchIncludeList := nil;
  end
  else
  begin
    FSearchIncludeList := TFhirStringList.Create;
    FSearchIncludeList.Assign(TFhirConformanceRestResource(oSource).FSearchIncludeList);
  end;
  if (TFhirConformanceRestResource(oSource).FSearchRevIncludeList = nil) then
  begin
    FSearchRevIncludeList.free;
    FSearchRevIncludeList := nil;
  end
  else
  begin
    FSearchRevIncludeList := TFhirStringList.Create;
    FSearchRevIncludeList.Assign(TFhirConformanceRestResource(oSource).FSearchRevIncludeList);
  end;
  if (TFhirConformanceRestResource(oSource).FSearchParamList = nil) then
  begin
    FSearchParamList.free;
    FSearchParamList := nil;
  end
  else
  begin
    FSearchParamList := TFhirConformanceRestResourceSearchParamList.Create;
    FSearchParamList.Assign(TFhirConformanceRestResource(oSource).FSearchParamList);
  end;
end;

procedure TFhirConformanceRestResource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'profile') Then
     list.add(FProfile.Link);
  if (child_name = 'interaction') Then
    list.addAll(FInteractionList);
  if (child_name = 'versioning') Then
     list.add(FVersioning.Link);
  if (child_name = 'readHistory') Then
     list.add(FReadHistory.Link);
  if (child_name = 'updateCreate') Then
     list.add(FUpdateCreate.Link);
  if (child_name = 'conditionalCreate') Then
     list.add(FConditionalCreate.Link);
  if (child_name = 'conditionalUpdate') Then
     list.add(FConditionalUpdate.Link);
  if (child_name = 'conditionalDelete') Then
     list.add(FConditionalDelete.Link);
  if (child_name = 'searchInclude') Then
    list.addAll(FSearchIncludeList);
  if (child_name = 'searchRevInclude') Then
    list.addAll(FSearchRevIncludeList);
  if (child_name = 'searchParam') Then
    list.addAll(FSearchParamList);
end;

procedure TFhirConformanceRestResource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(StructureDefinition)', FProfile.Link));{2}
  oList.add(TFHIRProperty.create(self, 'interaction', '', FInteractionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'versioning', 'code', FVersioning.Link));{1}
  oList.add(TFHIRProperty.create(self, 'readHistory', 'boolean', FReadHistory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'updateCreate', 'boolean', FUpdateCreate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'conditionalCreate', 'boolean', FConditionalCreate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'conditionalUpdate', 'boolean', FConditionalUpdate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'conditionalDelete', 'code', FConditionalDelete.Link));{1}
  oList.add(TFHIRProperty.create(self, 'searchInclude', 'string', FSearchIncludeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'searchRevInclude', 'string', FSearchRevIncludeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'searchParam', '', FSearchParamList.Link)){3};
end;

procedure TFhirConformanceRestResource.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then
    if propValue is TFHIRCode then
      Type_Element := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      Type_Element := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'profile') then Profile := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else if (propName = 'interaction') then InteractionList.add(propValue as TFhirConformanceRestResourceInteraction){2}
  else if (propName = 'versioning') then VersioningElement := propValue as TFHIREnum
  else if (propName = 'readHistory') then ReadHistoryElement := propValue as TFhirBoolean{5a}
  else if (propName = 'updateCreate') then UpdateCreateElement := propValue as TFhirBoolean{5a}
  else if (propName = 'conditionalCreate') then ConditionalCreateElement := propValue as TFhirBoolean{5a}
  else if (propName = 'conditionalUpdate') then ConditionalUpdateElement := propValue as TFhirBoolean{5a}
  else if (propName = 'conditionalDelete') then ConditionalDeleteElement := propValue as TFHIREnum
  else if (propName = 'searchInclude') then SearchIncludeList.add(propValue as TFhirString){2}
  else if (propName = 'searchRevInclude') then SearchRevIncludeList.add(propValue as TFhirString){2}
  else if (propName = 'searchParam') then SearchParamList.add(propValue as TFhirConformanceRestResourceSearchParam){2}
  else inherited;
end;

function TFhirConformanceRestResource.FhirType : string;
begin
  result := 'resource';
end;

function TFhirConformanceRestResource.Link : TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource(inherited Link);
end;

function TFhirConformanceRestResource.Clone : TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource(inherited Clone);
end;

function TFhirConformanceRestResource.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResource;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResource)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResource(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(profileElement, o.profileElement, true) and 
      compareDeep(interactionList, o.interactionList, true) and compareDeep(versioningElement, o.versioningElement, true) and 
      compareDeep(readHistoryElement, o.readHistoryElement, true) and compareDeep(updateCreateElement, o.updateCreateElement, true) and 
      compareDeep(conditionalCreateElement, o.conditionalCreateElement, true) and compareDeep(conditionalUpdateElement, o.conditionalUpdateElement, true) and 
      compareDeep(conditionalDeleteElement, o.conditionalDeleteElement, true) and compareDeep(searchIncludeList, o.searchIncludeList, true) and 
      compareDeep(searchRevIncludeList, o.searchRevIncludeList, true) and compareDeep(searchParamList, o.searchParamList, true);
  end;
end;

function TFhirConformanceRestResource.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResource;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResource)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResource(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(versioningElement, o.versioningElement, true) and 
      compareValues(readHistoryElement, o.readHistoryElement, true) and compareValues(updateCreateElement, o.updateCreateElement, true) and 
      compareValues(conditionalCreateElement, o.conditionalCreateElement, true) and compareValues(conditionalUpdateElement, o.conditionalUpdateElement, true) and 
      compareValues(conditionalDeleteElement, o.conditionalDeleteElement, true) and compareValues(searchIncludeList, o.searchIncludeList, true) and 
      compareValues(searchRevIncludeList, o.searchRevIncludeList, true);
  end;
end;

{ TFhirConformanceRestResource }

Procedure TFhirConformanceRestResource.SetType_(value : TFhirCode);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirConformanceRestResource.GetType_ST : String;
begin
  if FType_ = nil then
    result := ''
  else
    result := FType_.value;
end;

Procedure TFhirConformanceRestResource.SetType_ST(value : String);
begin
  if value <> '' then
  begin
    if FType_ = nil then
      FType_ := TFhirCode.create;
    FType_.value := value
  end
  else if FType_ <> nil then
    FType_.value := '';
end;

Procedure TFhirConformanceRestResource.SetProfile(value : TFhirReference{TFhirStructureDefinition});
begin
  FProfile.free;
  FProfile := value;
end;

Function TFhirConformanceRestResource.GetInteractionList : TFhirConformanceRestResourceInteractionList;
begin
  if FInteractionList = nil then
    FInteractionList := TFhirConformanceRestResourceInteractionList.Create;
  result := FInteractionList;
end;

Function TFhirConformanceRestResource.GetHasInteractionList : boolean;
begin
  result := (FInteractionList <> nil) and (FInteractionList.count > 0);
end;

Procedure TFhirConformanceRestResource.SetVersioning(value : TFhirEnum);
begin
  FVersioning.free;
  FVersioning := value;
end;

Function TFhirConformanceRestResource.GetVersioningST : TFhirVersioningPolicy;
begin
  if FVersioning = nil then
    result := TFhirVersioningPolicy(0)
  else
    result := TFhirVersioningPolicy(StringArrayIndexOfSensitive(CODES_TFhirVersioningPolicy, FVersioning.value));
end;

Procedure TFhirConformanceRestResource.SetVersioningST(value : TFhirVersioningPolicy);
begin
  if ord(value) = 0 then
    VersioningElement := nil
  else
    VersioningElement := TFhirEnum.create(CODES_TFhirVersioningPolicy[value]);
end;

Procedure TFhirConformanceRestResource.SetReadHistory(value : TFhirBoolean);
begin
  FReadHistory.free;
  FReadHistory := value;
end;

Function TFhirConformanceRestResource.GetReadHistoryST : Boolean;
begin
  if FReadHistory = nil then
    result := false
  else
    result := FReadHistory.value;
end;

Procedure TFhirConformanceRestResource.SetReadHistoryST(value : Boolean);
begin
  if FReadHistory = nil then
    FReadHistory := TFhirBoolean.create;
  FReadHistory.value := value
end;

Procedure TFhirConformanceRestResource.SetUpdateCreate(value : TFhirBoolean);
begin
  FUpdateCreate.free;
  FUpdateCreate := value;
end;

Function TFhirConformanceRestResource.GetUpdateCreateST : Boolean;
begin
  if FUpdateCreate = nil then
    result := false
  else
    result := FUpdateCreate.value;
end;

Procedure TFhirConformanceRestResource.SetUpdateCreateST(value : Boolean);
begin
  if FUpdateCreate = nil then
    FUpdateCreate := TFhirBoolean.create;
  FUpdateCreate.value := value
end;

Procedure TFhirConformanceRestResource.SetConditionalCreate(value : TFhirBoolean);
begin
  FConditionalCreate.free;
  FConditionalCreate := value;
end;

Function TFhirConformanceRestResource.GetConditionalCreateST : Boolean;
begin
  if FConditionalCreate = nil then
    result := false
  else
    result := FConditionalCreate.value;
end;

Procedure TFhirConformanceRestResource.SetConditionalCreateST(value : Boolean);
begin
  if FConditionalCreate = nil then
    FConditionalCreate := TFhirBoolean.create;
  FConditionalCreate.value := value
end;

Procedure TFhirConformanceRestResource.SetConditionalUpdate(value : TFhirBoolean);
begin
  FConditionalUpdate.free;
  FConditionalUpdate := value;
end;

Function TFhirConformanceRestResource.GetConditionalUpdateST : Boolean;
begin
  if FConditionalUpdate = nil then
    result := false
  else
    result := FConditionalUpdate.value;
end;

Procedure TFhirConformanceRestResource.SetConditionalUpdateST(value : Boolean);
begin
  if FConditionalUpdate = nil then
    FConditionalUpdate := TFhirBoolean.create;
  FConditionalUpdate.value := value
end;

Procedure TFhirConformanceRestResource.SetConditionalDelete(value : TFhirEnum);
begin
  FConditionalDelete.free;
  FConditionalDelete := value;
end;

Function TFhirConformanceRestResource.GetConditionalDeleteST : TFhirConditionalDeleteStatus;
begin
  if FConditionalDelete = nil then
    result := TFhirConditionalDeleteStatus(0)
  else
    result := TFhirConditionalDeleteStatus(StringArrayIndexOfSensitive(CODES_TFhirConditionalDeleteStatus, FConditionalDelete.value));
end;

Procedure TFhirConformanceRestResource.SetConditionalDeleteST(value : TFhirConditionalDeleteStatus);
begin
  if ord(value) = 0 then
    ConditionalDeleteElement := nil
  else
    ConditionalDeleteElement := TFhirEnum.create(CODES_TFhirConditionalDeleteStatus[value]);
end;

Function TFhirConformanceRestResource.GetSearchIncludeList : TFhirStringList;
begin
  if FSearchIncludeList = nil then
    FSearchIncludeList := TFhirStringList.Create;
  result := FSearchIncludeList;
end;

Function TFhirConformanceRestResource.GetHasSearchIncludeList : boolean;
begin
  result := (FSearchIncludeList <> nil) and (FSearchIncludeList.count > 0);
end;

Function TFhirConformanceRestResource.GetSearchRevIncludeList : TFhirStringList;
begin
  if FSearchRevIncludeList = nil then
    FSearchRevIncludeList := TFhirStringList.Create;
  result := FSearchRevIncludeList;
end;

Function TFhirConformanceRestResource.GetHasSearchRevIncludeList : boolean;
begin
  result := (FSearchRevIncludeList <> nil) and (FSearchRevIncludeList.count > 0);
end;

Function TFhirConformanceRestResource.GetSearchParamList : TFhirConformanceRestResourceSearchParamList;
begin
  if FSearchParamList = nil then
    FSearchParamList := TFhirConformanceRestResourceSearchParamList.Create;
  result := FSearchParamList;
end;

Function TFhirConformanceRestResource.GetHasSearchParamList : boolean;
begin
  result := (FSearchParamList <> nil) and (FSearchParamList.count > 0);
end;


{ TFhirConformanceRestResourceListEnumerator }

Constructor TFhirConformanceRestResourceListEnumerator.Create(list : TFhirConformanceRestResourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestResourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestResourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestResourceListEnumerator.GetCurrent : TFhirConformanceRestResource;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestResourceList }
procedure TFhirConformanceRestResourceList.AddItem(value: TFhirConformanceRestResource);
begin
  assert(value.ClassName = 'TFhirConformanceRestResource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestResource');
  add(value);
end;


function TFhirConformanceRestResourceList.Append: TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestResourceList.GetEnumerator : TFhirConformanceRestResourceListEnumerator;
begin
  result := TFhirConformanceRestResourceListEnumerator.Create(self.link);
end;

function TFhirConformanceRestResourceList.Clone: TFhirConformanceRestResourceList;
begin
  result := TFhirConformanceRestResourceList(inherited Clone);
end;

function TFhirConformanceRestResourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestResourceList.GetItemN(index: Integer): TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceList.IndexOf(value: TFhirConformanceRestResource): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestResourceList.Insert(index: Integer): TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceList.InsertItem(index: Integer; value: TFhirConformanceRestResource);
begin
  assert(value is TFhirConformanceRestResource);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestResourceList.Item(index: Integer): TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceList.Link: TFhirConformanceRestResourceList;
begin
  result := TFhirConformanceRestResourceList(inherited Link);
end;

procedure TFhirConformanceRestResourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestResourceList.SetItemByIndex(index: Integer; value: TFhirConformanceRestResource);
begin
  assert(value is TFhirConformanceRestResource);
  FhirConformanceRestResources[index] := value;
end;

procedure TFhirConformanceRestResourceList.SetItemN(index: Integer; value: TFhirConformanceRestResource);
begin
  assert(value is TFhirConformanceRestResource);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestResourceInteraction }

constructor TFhirConformanceRestResourceInteraction.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestResourceInteraction.Destroy;
begin
  FCode.free;
  FDocumentation.free;
  inherited;
end;

procedure TFhirConformanceRestResourceInteraction.Assign(oSource : TAdvObject);
begin
  inherited;
  FCode := TFhirConformanceRestResourceInteraction(oSource).FCode.Link;
  documentationElement := TFhirConformanceRestResourceInteraction(oSource).documentationElement.Clone;
end;

procedure TFhirConformanceRestResourceInteraction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
end;

procedure TFhirConformanceRestResourceInteraction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
end;

procedure TFhirConformanceRestResourceInteraction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirConformanceRestResourceInteraction.FhirType : string;
begin
  result := 'interaction';
end;

function TFhirConformanceRestResourceInteraction.Link : TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction(inherited Link);
end;

function TFhirConformanceRestResourceInteraction.Clone : TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction(inherited Clone);
end;

function TFhirConformanceRestResourceInteraction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResourceInteraction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResourceInteraction)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResourceInteraction(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(documentationElement, o.documentationElement, true);
  end;
end;

function TFhirConformanceRestResourceInteraction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResourceInteraction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResourceInteraction)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResourceInteraction(other);
    result := compareValues(codeElement, o.codeElement, true) and compareValues(documentationElement, o.documentationElement, true);
  end;
end;

{ TFhirConformanceRestResourceInteraction }

Procedure TFhirConformanceRestResourceInteraction.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConformanceRestResourceInteraction.GetCodeST : TFhirTypeRestfulInteraction;
begin
  if FCode = nil then
    result := TFhirTypeRestfulInteraction(0)
  else
    result := TFhirTypeRestfulInteraction(StringArrayIndexOfSensitive(CODES_TFhirTypeRestfulInteraction, FCode.value));
end;

Procedure TFhirConformanceRestResourceInteraction.SetCodeST(value : TFhirTypeRestfulInteraction);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirTypeRestfulInteraction[value]);
end;

Procedure TFhirConformanceRestResourceInteraction.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceRestResourceInteraction.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceRestResourceInteraction.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;


{ TFhirConformanceRestResourceInteractionListEnumerator }

Constructor TFhirConformanceRestResourceInteractionListEnumerator.Create(list : TFhirConformanceRestResourceInteractionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestResourceInteractionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestResourceInteractionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestResourceInteractionListEnumerator.GetCurrent : TFhirConformanceRestResourceInteraction;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestResourceInteractionList }
procedure TFhirConformanceRestResourceInteractionList.AddItem(value: TFhirConformanceRestResourceInteraction);
begin
  assert(value.ClassName = 'TFhirConformanceRestResourceInteraction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestResourceInteraction');
  add(value);
end;


function TFhirConformanceRestResourceInteractionList.Append: TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceInteractionList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestResourceInteractionList.GetEnumerator : TFhirConformanceRestResourceInteractionListEnumerator;
begin
  result := TFhirConformanceRestResourceInteractionListEnumerator.Create(self.link);
end;

function TFhirConformanceRestResourceInteractionList.Clone: TFhirConformanceRestResourceInteractionList;
begin
  result := TFhirConformanceRestResourceInteractionList(inherited Clone);
end;

function TFhirConformanceRestResourceInteractionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestResourceInteractionList.GetItemN(index: Integer): TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceInteractionList.IndexOf(value: TFhirConformanceRestResourceInteraction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestResourceInteractionList.Insert(index: Integer): TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceInteractionList.InsertItem(index: Integer; value: TFhirConformanceRestResourceInteraction);
begin
  assert(value is TFhirConformanceRestResourceInteraction);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestResourceInteractionList.Item(index: Integer): TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceInteractionList.Link: TFhirConformanceRestResourceInteractionList;
begin
  result := TFhirConformanceRestResourceInteractionList(inherited Link);
end;

procedure TFhirConformanceRestResourceInteractionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestResourceInteractionList.SetItemByIndex(index: Integer; value: TFhirConformanceRestResourceInteraction);
begin
  assert(value is TFhirConformanceRestResourceInteraction);
  FhirConformanceRestResourceInteractions[index] := value;
end;

procedure TFhirConformanceRestResourceInteractionList.SetItemN(index: Integer; value: TFhirConformanceRestResourceInteraction);
begin
  assert(value is TFhirConformanceRestResourceInteraction);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestResourceSearchParam }

constructor TFhirConformanceRestResourceSearchParam.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestResourceSearchParam.Destroy;
begin
  FName.free;
  FDefinition.free;
  FType_.free;
  FDocumentation.free;
  FTargetList.Free;
  FModifier.Free;
  FChainList.Free;
  inherited;
end;

procedure TFhirConformanceRestResourceSearchParam.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirConformanceRestResourceSearchParam(oSource).nameElement.Clone;
  definitionElement := TFhirConformanceRestResourceSearchParam(oSource).definitionElement.Clone;
  FType_ := TFhirConformanceRestResourceSearchParam(oSource).FType_.Link;
  documentationElement := TFhirConformanceRestResourceSearchParam(oSource).documentationElement.Clone;
  if (TFhirConformanceRestResourceSearchParam(oSource).FTargetList = nil) then
  begin
    FTargetList.free;
    FTargetList := nil;
  end
  else
  begin
    FTargetList := TFhirCodeList.Create;
    FTargetList.Assign(TFhirConformanceRestResourceSearchParam(oSource).FTargetList);
  end;
  if (TFhirConformanceRestResourceSearchParam(oSource).FModifier = nil) then
  begin
    FModifier.free;
    FModifier := nil;
  end
  else
  begin
    FModifier := TFHIREnumList.Create;
    FModifier.Assign(TFhirConformanceRestResourceSearchParam(oSource).FModifier);
  end;
  if (TFhirConformanceRestResourceSearchParam(oSource).FChainList = nil) then
  begin
    FChainList.free;
    FChainList := nil;
  end
  else
  begin
    FChainList := TFhirStringList.Create;
    FChainList.Assign(TFhirConformanceRestResourceSearchParam(oSource).FChainList);
  end;
end;

procedure TFhirConformanceRestResourceSearchParam.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'definition') Then
     list.add(FDefinition.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
  if (child_name = 'target') Then
    list.addAll(FTargetList);
  if (child_name = 'modifier') Then
     list.addAll(FModifier);
  if (child_name = 'chain') Then
    list.addAll(FChainList);
end;

procedure TFhirConformanceRestResourceSearchParam.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'definition', 'uri', FDefinition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'code', FTargetList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'modifier', 'code', FModifier.Link)){3};
  oList.add(TFHIRProperty.create(self, 'chain', 'string', FChainList.Link)){3};
end;

procedure TFhirConformanceRestResourceSearchParam.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'definition') then DefinitionElement := propValue as TFhirUri{5a}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else if (propName = 'target') then TargetList.add(propValue as TFhirCode){2}
  else if (propName = 'modifier') then FModifier.add(propValue as TFHIREnum) {1}
  else if (propName = 'chain') then ChainList.add(propValue as TFhirString){2}
  else inherited;
end;

function TFhirConformanceRestResourceSearchParam.FhirType : string;
begin
  result := 'searchParam';
end;

function TFhirConformanceRestResourceSearchParam.Link : TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam(inherited Link);
end;

function TFhirConformanceRestResourceSearchParam.Clone : TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam(inherited Clone);
end;

function TFhirConformanceRestResourceSearchParam.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResourceSearchParam;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResourceSearchParam)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResourceSearchParam(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(definitionElement, o.definitionElement, true) and 
      compareDeep(type_Element, o.type_Element, true) and compareDeep(documentationElement, o.documentationElement, true) and 
      compareDeep(targetList, o.targetList, true) and compareDeep(modifierList, o.modifierList, true) and 
      compareDeep(chainList, o.chainList, true);
  end;
end;

function TFhirConformanceRestResourceSearchParam.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestResourceSearchParam;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestResourceSearchParam)) then
    result := false
  else
  begin
    o := TFhirConformanceRestResourceSearchParam(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(definitionElement, o.definitionElement, true) and 
      compareValues(type_Element, o.type_Element, true) and compareValues(documentationElement, o.documentationElement, true) and 
      compareValues(targetList, o.targetList, true) and compareValues(modifierList, o.modifierList, true) and 
      compareValues(chainList, o.chainList, true);
  end;
end;

{ TFhirConformanceRestResourceSearchParam }

Procedure TFhirConformanceRestResourceSearchParam.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConformanceRestResourceSearchParam.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConformanceRestResourceSearchParam.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirConformanceRestResourceSearchParam.SetDefinition(value : TFhirUri);
begin
  FDefinition.free;
  FDefinition := value;
end;

Function TFhirConformanceRestResourceSearchParam.GetDefinitionST : String;
begin
  if FDefinition = nil then
    result := ''
  else
    result := FDefinition.value;
end;

Procedure TFhirConformanceRestResourceSearchParam.SetDefinitionST(value : String);
begin
  if value <> '' then
  begin
    if FDefinition = nil then
      FDefinition := TFhirUri.create;
    FDefinition.value := value
  end
  else if FDefinition <> nil then
    FDefinition.value := '';
end;

Procedure TFhirConformanceRestResourceSearchParam.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirConformanceRestResourceSearchParam.GetType_ST : TFhirSearchParamType;
begin
  if FType_ = nil then
    result := TFhirSearchParamType(0)
  else
    result := TFhirSearchParamType(StringArrayIndexOfSensitive(CODES_TFhirSearchParamType, FType_.value));
end;

Procedure TFhirConformanceRestResourceSearchParam.SetType_ST(value : TFhirSearchParamType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirSearchParamType[value]);
end;

Procedure TFhirConformanceRestResourceSearchParam.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceRestResourceSearchParam.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceRestResourceSearchParam.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;

Function TFhirConformanceRestResourceSearchParam.GetTargetList : TFhirCodeList;
begin
  if FTargetList = nil then
    FTargetList := TFhirCodeList.Create;
  result := FTargetList;
end;

Function TFhirConformanceRestResourceSearchParam.GetHasTargetList : boolean;
begin
  result := (FTargetList <> nil) and (FTargetList.count > 0);
end;

Function TFhirConformanceRestResourceSearchParam.GetModifier : TFhirEnumList;
begin
  if FModifier = nil then
    FModifier := TFHIREnumList.Create;
  result := FModifier;
end;

Function TFhirConformanceRestResourceSearchParam.GetHasModifier : boolean;
begin
  result := (FModifier <> nil) and (FModifier.count > 0);
end;

Function TFhirConformanceRestResourceSearchParam.GetModifierST : TFhirSearchModifierCodeList;
  var i : integer;
begin
  result := [];
  if Fmodifier <> nil then
    for i := 0 to Fmodifier.count - 1 do
      result := result + [TFhirSearchModifierCode(StringArrayIndexOfSensitive(CODES_TFhirSearchModifierCode, Fmodifier[i].value))];
end;

Procedure TFhirConformanceRestResourceSearchParam.SetModifierST(value : TFhirSearchModifierCodeList);
var a : TFhirSearchModifierCode;
begin
  Fmodifier.clear;
  for a := low(TFhirSearchModifierCode) to high(TFhirSearchModifierCode) do
    if a in value then
      begin
         if Fmodifier = nil then
           Fmodifier := TFhirEnumList.create;
         Fmodifier.add(TFhirEnum.create(CODES_TFhirSearchModifierCode[a]));
      end;
end;

Function TFhirConformanceRestResourceSearchParam.GetChainList : TFhirStringList;
begin
  if FChainList = nil then
    FChainList := TFhirStringList.Create;
  result := FChainList;
end;

Function TFhirConformanceRestResourceSearchParam.GetHasChainList : boolean;
begin
  result := (FChainList <> nil) and (FChainList.count > 0);
end;


{ TFhirConformanceRestResourceSearchParamListEnumerator }

Constructor TFhirConformanceRestResourceSearchParamListEnumerator.Create(list : TFhirConformanceRestResourceSearchParamList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestResourceSearchParamListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestResourceSearchParamListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestResourceSearchParamListEnumerator.GetCurrent : TFhirConformanceRestResourceSearchParam;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestResourceSearchParamList }
procedure TFhirConformanceRestResourceSearchParamList.AddItem(value: TFhirConformanceRestResourceSearchParam);
begin
  assert(value.ClassName = 'TFhirConformanceRestResourceSearchParam', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestResourceSearchParam');
  add(value);
end;


function TFhirConformanceRestResourceSearchParamList.Append: TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceSearchParamList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestResourceSearchParamList.GetEnumerator : TFhirConformanceRestResourceSearchParamListEnumerator;
begin
  result := TFhirConformanceRestResourceSearchParamListEnumerator.Create(self.link);
end;

function TFhirConformanceRestResourceSearchParamList.Clone: TFhirConformanceRestResourceSearchParamList;
begin
  result := TFhirConformanceRestResourceSearchParamList(inherited Clone);
end;

function TFhirConformanceRestResourceSearchParamList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestResourceSearchParamList.GetItemN(index: Integer): TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceSearchParamList.IndexOf(value: TFhirConformanceRestResourceSearchParam): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestResourceSearchParamList.Insert(index: Integer): TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestResourceSearchParamList.InsertItem(index: Integer; value: TFhirConformanceRestResourceSearchParam);
begin
  assert(value is TFhirConformanceRestResourceSearchParam);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestResourceSearchParamList.Item(index: Integer): TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam(ObjectByIndex[index]);
end;

function TFhirConformanceRestResourceSearchParamList.Link: TFhirConformanceRestResourceSearchParamList;
begin
  result := TFhirConformanceRestResourceSearchParamList(inherited Link);
end;

procedure TFhirConformanceRestResourceSearchParamList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestResourceSearchParamList.SetItemByIndex(index: Integer; value: TFhirConformanceRestResourceSearchParam);
begin
  assert(value is TFhirConformanceRestResourceSearchParam);
  FhirConformanceRestResourceSearchParams[index] := value;
end;

procedure TFhirConformanceRestResourceSearchParamList.SetItemN(index: Integer; value: TFhirConformanceRestResourceSearchParam);
begin
  assert(value is TFhirConformanceRestResourceSearchParam);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestInteraction }

constructor TFhirConformanceRestInteraction.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestInteraction.Destroy;
begin
  FCode.free;
  FDocumentation.free;
  inherited;
end;

procedure TFhirConformanceRestInteraction.Assign(oSource : TAdvObject);
begin
  inherited;
  FCode := TFhirConformanceRestInteraction(oSource).FCode.Link;
  documentationElement := TFhirConformanceRestInteraction(oSource).documentationElement.Clone;
end;

procedure TFhirConformanceRestInteraction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
end;

procedure TFhirConformanceRestInteraction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
end;

procedure TFhirConformanceRestInteraction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirConformanceRestInteraction.FhirType : string;
begin
  result := 'interaction';
end;

function TFhirConformanceRestInteraction.Link : TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction(inherited Link);
end;

function TFhirConformanceRestInteraction.Clone : TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction(inherited Clone);
end;

function TFhirConformanceRestInteraction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestInteraction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestInteraction)) then
    result := false
  else
  begin
    o := TFhirConformanceRestInteraction(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(documentationElement, o.documentationElement, true);
  end;
end;

function TFhirConformanceRestInteraction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestInteraction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestInteraction)) then
    result := false
  else
  begin
    o := TFhirConformanceRestInteraction(other);
    result := compareValues(codeElement, o.codeElement, true) and compareValues(documentationElement, o.documentationElement, true);
  end;
end;

{ TFhirConformanceRestInteraction }

Procedure TFhirConformanceRestInteraction.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirConformanceRestInteraction.GetCodeST : TFhirSystemRestfulInteraction;
begin
  if FCode = nil then
    result := TFhirSystemRestfulInteraction(0)
  else
    result := TFhirSystemRestfulInteraction(StringArrayIndexOfSensitive(CODES_TFhirSystemRestfulInteraction, FCode.value));
end;

Procedure TFhirConformanceRestInteraction.SetCodeST(value : TFhirSystemRestfulInteraction);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirSystemRestfulInteraction[value]);
end;

Procedure TFhirConformanceRestInteraction.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceRestInteraction.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceRestInteraction.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;


{ TFhirConformanceRestInteractionListEnumerator }

Constructor TFhirConformanceRestInteractionListEnumerator.Create(list : TFhirConformanceRestInteractionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestInteractionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestInteractionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestInteractionListEnumerator.GetCurrent : TFhirConformanceRestInteraction;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestInteractionList }
procedure TFhirConformanceRestInteractionList.AddItem(value: TFhirConformanceRestInteraction);
begin
  assert(value.ClassName = 'TFhirConformanceRestInteraction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestInteraction');
  add(value);
end;


function TFhirConformanceRestInteractionList.Append: TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestInteractionList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestInteractionList.GetEnumerator : TFhirConformanceRestInteractionListEnumerator;
begin
  result := TFhirConformanceRestInteractionListEnumerator.Create(self.link);
end;

function TFhirConformanceRestInteractionList.Clone: TFhirConformanceRestInteractionList;
begin
  result := TFhirConformanceRestInteractionList(inherited Clone);
end;

function TFhirConformanceRestInteractionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestInteractionList.GetItemN(index: Integer): TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction(ObjectByIndex[index]);
end;

function TFhirConformanceRestInteractionList.IndexOf(value: TFhirConformanceRestInteraction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestInteractionList.Insert(index: Integer): TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestInteractionList.InsertItem(index: Integer; value: TFhirConformanceRestInteraction);
begin
  assert(value is TFhirConformanceRestInteraction);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestInteractionList.Item(index: Integer): TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction(ObjectByIndex[index]);
end;

function TFhirConformanceRestInteractionList.Link: TFhirConformanceRestInteractionList;
begin
  result := TFhirConformanceRestInteractionList(inherited Link);
end;

procedure TFhirConformanceRestInteractionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestInteractionList.SetItemByIndex(index: Integer; value: TFhirConformanceRestInteraction);
begin
  assert(value is TFhirConformanceRestInteraction);
  FhirConformanceRestInteractions[index] := value;
end;

procedure TFhirConformanceRestInteractionList.SetItemN(index: Integer; value: TFhirConformanceRestInteraction);
begin
  assert(value is TFhirConformanceRestInteraction);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceRestOperation }

constructor TFhirConformanceRestOperation.Create;
begin
  inherited;
end;

destructor TFhirConformanceRestOperation.Destroy;
begin
  FName.free;
  FDefinition.free;
  inherited;
end;

procedure TFhirConformanceRestOperation.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirConformanceRestOperation(oSource).nameElement.Clone;
  definition := TFhirConformanceRestOperation(oSource).definition.Clone;
end;

procedure TFhirConformanceRestOperation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'definition') Then
     list.add(FDefinition.Link);
end;

procedure TFhirConformanceRestOperation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'definition', 'Reference(OperationDefinition)', FDefinition.Link));{2}
end;

procedure TFhirConformanceRestOperation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'definition') then Definition := propValue as TFhirReference{TFhirOperationDefinition}{4b}
  else inherited;
end;

function TFhirConformanceRestOperation.FhirType : string;
begin
  result := 'operation';
end;

function TFhirConformanceRestOperation.Link : TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation(inherited Link);
end;

function TFhirConformanceRestOperation.Clone : TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation(inherited Clone);
end;

function TFhirConformanceRestOperation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestOperation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceRestOperation)) then
    result := false
  else
  begin
    o := TFhirConformanceRestOperation(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(definitionElement, o.definitionElement, true);
  end;
end;

function TFhirConformanceRestOperation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceRestOperation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceRestOperation)) then
    result := false
  else
  begin
    o := TFhirConformanceRestOperation(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirConformanceRestOperation }

Procedure TFhirConformanceRestOperation.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConformanceRestOperation.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConformanceRestOperation.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirConformanceRestOperation.SetDefinition(value : TFhirReference{TFhirOperationDefinition});
begin
  FDefinition.free;
  FDefinition := value;
end;


{ TFhirConformanceRestOperationListEnumerator }

Constructor TFhirConformanceRestOperationListEnumerator.Create(list : TFhirConformanceRestOperationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceRestOperationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceRestOperationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceRestOperationListEnumerator.GetCurrent : TFhirConformanceRestOperation;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceRestOperationList }
procedure TFhirConformanceRestOperationList.AddItem(value: TFhirConformanceRestOperation);
begin
  assert(value.ClassName = 'TFhirConformanceRestOperation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceRestOperation');
  add(value);
end;


function TFhirConformanceRestOperationList.Append: TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestOperationList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceRestOperationList.GetEnumerator : TFhirConformanceRestOperationListEnumerator;
begin
  result := TFhirConformanceRestOperationListEnumerator.Create(self.link);
end;

function TFhirConformanceRestOperationList.Clone: TFhirConformanceRestOperationList;
begin
  result := TFhirConformanceRestOperationList(inherited Clone);
end;

function TFhirConformanceRestOperationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceRestOperationList.GetItemN(index: Integer): TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation(ObjectByIndex[index]);
end;

function TFhirConformanceRestOperationList.IndexOf(value: TFhirConformanceRestOperation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceRestOperationList.Insert(index: Integer): TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceRestOperationList.InsertItem(index: Integer; value: TFhirConformanceRestOperation);
begin
  assert(value is TFhirConformanceRestOperation);
  Inherited Insert(index, value);
end;

function TFhirConformanceRestOperationList.Item(index: Integer): TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation(ObjectByIndex[index]);
end;

function TFhirConformanceRestOperationList.Link: TFhirConformanceRestOperationList;
begin
  result := TFhirConformanceRestOperationList(inherited Link);
end;

procedure TFhirConformanceRestOperationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceRestOperationList.SetItemByIndex(index: Integer; value: TFhirConformanceRestOperation);
begin
  assert(value is TFhirConformanceRestOperation);
  FhirConformanceRestOperations[index] := value;
end;

procedure TFhirConformanceRestOperationList.SetItemN(index: Integer; value: TFhirConformanceRestOperation);
begin
  assert(value is TFhirConformanceRestOperation);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceMessaging }

constructor TFhirConformanceMessaging.Create;
begin
  inherited;
end;

destructor TFhirConformanceMessaging.Destroy;
begin
  FEndpointList.Free;
  FReliableCache.free;
  FDocumentation.free;
  FEventList.Free;
  inherited;
end;

procedure TFhirConformanceMessaging.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirConformanceMessaging(oSource).FEndpointList = nil) then
  begin
    FEndpointList.free;
    FEndpointList := nil;
  end
  else
  begin
    FEndpointList := TFhirConformanceMessagingEndpointList.Create;
    FEndpointList.Assign(TFhirConformanceMessaging(oSource).FEndpointList);
  end;
  reliableCacheElement := TFhirConformanceMessaging(oSource).reliableCacheElement.Clone;
  documentationElement := TFhirConformanceMessaging(oSource).documentationElement.Clone;
  if (TFhirConformanceMessaging(oSource).FEventList = nil) then
  begin
    FEventList.free;
    FEventList := nil;
  end
  else
  begin
    FEventList := TFhirConformanceMessagingEventList.Create;
    FEventList.Assign(TFhirConformanceMessaging(oSource).FEventList);
  end;
end;

procedure TFhirConformanceMessaging.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'endpoint') Then
    list.addAll(FEndpointList);
  if (child_name = 'reliableCache') Then
     list.add(FReliableCache.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
  if (child_name = 'event') Then
    list.addAll(FEventList);
end;

procedure TFhirConformanceMessaging.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'endpoint', '', FEndpointList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reliableCache', 'unsignedInt', FReliableCache.Link));{2}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'event', '', FEventList.Link)){3};
end;

procedure TFhirConformanceMessaging.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'endpoint') then EndpointList.add(propValue as TFhirConformanceMessagingEndpoint){2}
  else if (propName = 'reliableCache') then ReliableCacheElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else if (propName = 'event') then EventList.add(propValue as TFhirConformanceMessagingEvent){2}
  else inherited;
end;

function TFhirConformanceMessaging.FhirType : string;
begin
  result := 'messaging';
end;

function TFhirConformanceMessaging.Link : TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging(inherited Link);
end;

function TFhirConformanceMessaging.Clone : TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging(inherited Clone);
end;

function TFhirConformanceMessaging.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessaging;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceMessaging)) then
    result := false
  else
  begin
    o := TFhirConformanceMessaging(other);
    result := compareDeep(endpointList, o.endpointList, true) and compareDeep(reliableCacheElement, o.reliableCacheElement, true) and 
      compareDeep(documentationElement, o.documentationElement, true) and compareDeep(eventList, o.eventList, true);
  end;
end;

function TFhirConformanceMessaging.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessaging;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceMessaging)) then
    result := false
  else
  begin
    o := TFhirConformanceMessaging(other);
    result := compareValues(reliableCacheElement, o.reliableCacheElement, true) and 
      compareValues(documentationElement, o.documentationElement, true);
  end;
end;

{ TFhirConformanceMessaging }

Function TFhirConformanceMessaging.GetEndpointList : TFhirConformanceMessagingEndpointList;
begin
  if FEndpointList = nil then
    FEndpointList := TFhirConformanceMessagingEndpointList.Create;
  result := FEndpointList;
end;

Function TFhirConformanceMessaging.GetHasEndpointList : boolean;
begin
  result := (FEndpointList <> nil) and (FEndpointList.count > 0);
end;

Procedure TFhirConformanceMessaging.SetReliableCache(value : TFhirUnsignedInt);
begin
  FReliableCache.free;
  FReliableCache := value;
end;

Function TFhirConformanceMessaging.GetReliableCacheST : String;
begin
  if FReliableCache = nil then
    result := ''
  else
    result := FReliableCache.value;
end;

Procedure TFhirConformanceMessaging.SetReliableCacheST(value : String);
begin
  if value <> '' then
  begin
    if FReliableCache = nil then
      FReliableCache := TFhirUnsignedInt.create;
    FReliableCache.value := value
  end
  else if FReliableCache <> nil then
    FReliableCache.value := '';
end;

Procedure TFhirConformanceMessaging.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceMessaging.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceMessaging.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;

Function TFhirConformanceMessaging.GetEventList : TFhirConformanceMessagingEventList;
begin
  if FEventList = nil then
    FEventList := TFhirConformanceMessagingEventList.Create;
  result := FEventList;
end;

Function TFhirConformanceMessaging.GetHasEventList : boolean;
begin
  result := (FEventList <> nil) and (FEventList.count > 0);
end;


{ TFhirConformanceMessagingListEnumerator }

Constructor TFhirConformanceMessagingListEnumerator.Create(list : TFhirConformanceMessagingList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceMessagingListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceMessagingListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceMessagingListEnumerator.GetCurrent : TFhirConformanceMessaging;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceMessagingList }
procedure TFhirConformanceMessagingList.AddItem(value: TFhirConformanceMessaging);
begin
  assert(value.ClassName = 'TFhirConformanceMessaging', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceMessaging');
  add(value);
end;


function TFhirConformanceMessagingList.Append: TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceMessagingList.GetEnumerator : TFhirConformanceMessagingListEnumerator;
begin
  result := TFhirConformanceMessagingListEnumerator.Create(self.link);
end;

function TFhirConformanceMessagingList.Clone: TFhirConformanceMessagingList;
begin
  result := TFhirConformanceMessagingList(inherited Clone);
end;

function TFhirConformanceMessagingList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceMessagingList.GetItemN(index: Integer): TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingList.IndexOf(value: TFhirConformanceMessaging): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceMessagingList.Insert(index: Integer): TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingList.InsertItem(index: Integer; value: TFhirConformanceMessaging);
begin
  assert(value is TFhirConformanceMessaging);
  Inherited Insert(index, value);
end;

function TFhirConformanceMessagingList.Item(index: Integer): TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingList.Link: TFhirConformanceMessagingList;
begin
  result := TFhirConformanceMessagingList(inherited Link);
end;

procedure TFhirConformanceMessagingList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceMessagingList.SetItemByIndex(index: Integer; value: TFhirConformanceMessaging);
begin
  assert(value is TFhirConformanceMessaging);
  FhirConformanceMessagings[index] := value;
end;

procedure TFhirConformanceMessagingList.SetItemN(index: Integer; value: TFhirConformanceMessaging);
begin
  assert(value is TFhirConformanceMessaging);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceMessagingEndpoint }

constructor TFhirConformanceMessagingEndpoint.Create;
begin
  inherited;
end;

destructor TFhirConformanceMessagingEndpoint.Destroy;
begin
  FProtocol.free;
  FAddress.free;
  inherited;
end;

procedure TFhirConformanceMessagingEndpoint.Assign(oSource : TAdvObject);
begin
  inherited;
  protocol := TFhirConformanceMessagingEndpoint(oSource).protocol.Clone;
  addressElement := TFhirConformanceMessagingEndpoint(oSource).addressElement.Clone;
end;

procedure TFhirConformanceMessagingEndpoint.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'protocol') Then
     list.add(FProtocol.Link);
  if (child_name = 'address') Then
     list.add(FAddress.Link);
end;

procedure TFhirConformanceMessagingEndpoint.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'protocol', 'Coding', FProtocol.Link));{2}
  oList.add(TFHIRProperty.create(self, 'address', 'uri', FAddress.Link));{2}
end;

procedure TFhirConformanceMessagingEndpoint.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'protocol') then Protocol := propValue as TFhirCoding{4b}
  else if (propName = 'address') then AddressElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirConformanceMessagingEndpoint.FhirType : string;
begin
  result := 'endpoint';
end;

function TFhirConformanceMessagingEndpoint.Link : TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint(inherited Link);
end;

function TFhirConformanceMessagingEndpoint.Clone : TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint(inherited Clone);
end;

function TFhirConformanceMessagingEndpoint.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessagingEndpoint;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceMessagingEndpoint)) then
    result := false
  else
  begin
    o := TFhirConformanceMessagingEndpoint(other);
    result := compareDeep(protocolElement, o.protocolElement, true) and compareDeep(addressElement, o.addressElement, true);
  end;
end;

function TFhirConformanceMessagingEndpoint.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessagingEndpoint;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceMessagingEndpoint)) then
    result := false
  else
  begin
    o := TFhirConformanceMessagingEndpoint(other);
    result := compareValues(addressElement, o.addressElement, true);
  end;
end;

{ TFhirConformanceMessagingEndpoint }

Procedure TFhirConformanceMessagingEndpoint.SetProtocol(value : TFhirCoding);
begin
  FProtocol.free;
  FProtocol := value;
end;

Procedure TFhirConformanceMessagingEndpoint.SetAddress(value : TFhirUri);
begin
  FAddress.free;
  FAddress := value;
end;

Function TFhirConformanceMessagingEndpoint.GetAddressST : String;
begin
  if FAddress = nil then
    result := ''
  else
    result := FAddress.value;
end;

Procedure TFhirConformanceMessagingEndpoint.SetAddressST(value : String);
begin
  if value <> '' then
  begin
    if FAddress = nil then
      FAddress := TFhirUri.create;
    FAddress.value := value
  end
  else if FAddress <> nil then
    FAddress.value := '';
end;


{ TFhirConformanceMessagingEndpointListEnumerator }

Constructor TFhirConformanceMessagingEndpointListEnumerator.Create(list : TFhirConformanceMessagingEndpointList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceMessagingEndpointListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceMessagingEndpointListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceMessagingEndpointListEnumerator.GetCurrent : TFhirConformanceMessagingEndpoint;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceMessagingEndpointList }
procedure TFhirConformanceMessagingEndpointList.AddItem(value: TFhirConformanceMessagingEndpoint);
begin
  assert(value.ClassName = 'TFhirConformanceMessagingEndpoint', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceMessagingEndpoint');
  add(value);
end;


function TFhirConformanceMessagingEndpointList.Append: TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingEndpointList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceMessagingEndpointList.GetEnumerator : TFhirConformanceMessagingEndpointListEnumerator;
begin
  result := TFhirConformanceMessagingEndpointListEnumerator.Create(self.link);
end;

function TFhirConformanceMessagingEndpointList.Clone: TFhirConformanceMessagingEndpointList;
begin
  result := TFhirConformanceMessagingEndpointList(inherited Clone);
end;

function TFhirConformanceMessagingEndpointList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceMessagingEndpointList.GetItemN(index: Integer): TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingEndpointList.IndexOf(value: TFhirConformanceMessagingEndpoint): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceMessagingEndpointList.Insert(index: Integer): TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingEndpointList.InsertItem(index: Integer; value: TFhirConformanceMessagingEndpoint);
begin
  assert(value is TFhirConformanceMessagingEndpoint);
  Inherited Insert(index, value);
end;

function TFhirConformanceMessagingEndpointList.Item(index: Integer): TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingEndpointList.Link: TFhirConformanceMessagingEndpointList;
begin
  result := TFhirConformanceMessagingEndpointList(inherited Link);
end;

procedure TFhirConformanceMessagingEndpointList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceMessagingEndpointList.SetItemByIndex(index: Integer; value: TFhirConformanceMessagingEndpoint);
begin
  assert(value is TFhirConformanceMessagingEndpoint);
  FhirConformanceMessagingEndpoints[index] := value;
end;

procedure TFhirConformanceMessagingEndpointList.SetItemN(index: Integer; value: TFhirConformanceMessagingEndpoint);
begin
  assert(value is TFhirConformanceMessagingEndpoint);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceMessagingEvent }

constructor TFhirConformanceMessagingEvent.Create;
begin
  inherited;
end;

destructor TFhirConformanceMessagingEvent.Destroy;
begin
  FCode.free;
  FCategory.free;
  FMode.free;
  FFocus.free;
  FRequest.free;
  FResponse.free;
  FDocumentation.free;
  inherited;
end;

procedure TFhirConformanceMessagingEvent.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirConformanceMessagingEvent(oSource).code.Clone;
  FCategory := TFhirConformanceMessagingEvent(oSource).FCategory.Link;
  FMode := TFhirConformanceMessagingEvent(oSource).FMode.Link;
  focusElement := TFhirConformanceMessagingEvent(oSource).focusElement.Clone;
  request := TFhirConformanceMessagingEvent(oSource).request.Clone;
  response := TFhirConformanceMessagingEvent(oSource).response.Clone;
  documentationElement := TFhirConformanceMessagingEvent(oSource).documentationElement.Clone;
end;

procedure TFhirConformanceMessagingEvent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'focus') Then
     list.add(FFocus.Link);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
end;

procedure TFhirConformanceMessagingEvent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'category', 'code', FCategory.Link));{1}
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'focus', 'code', FFocus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(StructureDefinition)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', 'Reference(StructureDefinition)', FResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
end;

procedure TFhirConformanceMessagingEvent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCoding{4b}
  else if (propName = 'category') then CategoryElement := propValue as TFHIREnum
  else if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'focus') then
    if propValue is TFHIRCode then
      FocusElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      FocusElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else if (propName = 'response') then Response := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirConformanceMessagingEvent.FhirType : string;
begin
  result := 'event';
end;

function TFhirConformanceMessagingEvent.Link : TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent(inherited Link);
end;

function TFhirConformanceMessagingEvent.Clone : TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent(inherited Clone);
end;

function TFhirConformanceMessagingEvent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessagingEvent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceMessagingEvent)) then
    result := false
  else
  begin
    o := TFhirConformanceMessagingEvent(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(categoryElement, o.categoryElement, true) and 
      compareDeep(modeElement, o.modeElement, true) and compareDeep(focusElement, o.focusElement, true) and 
      compareDeep(requestElement, o.requestElement, true) and compareDeep(responseElement, o.responseElement, true) and 
      compareDeep(documentationElement, o.documentationElement, true);
  end;
end;

function TFhirConformanceMessagingEvent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceMessagingEvent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceMessagingEvent)) then
    result := false
  else
  begin
    o := TFhirConformanceMessagingEvent(other);
    result := compareValues(categoryElement, o.categoryElement, true) and compareValues(modeElement, o.modeElement, true) and 
      compareValues(focusElement, o.focusElement, true) and compareValues(documentationElement, o.documentationElement, true);
  end;
end;

{ TFhirConformanceMessagingEvent }

Procedure TFhirConformanceMessagingEvent.SetCode(value : TFhirCoding);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirConformanceMessagingEvent.SetCategory(value : TFhirEnum);
begin
  FCategory.free;
  FCategory := value;
end;

Function TFhirConformanceMessagingEvent.GetCategoryST : TFhirMessageSignificanceCategory;
begin
  if FCategory = nil then
    result := TFhirMessageSignificanceCategory(0)
  else
    result := TFhirMessageSignificanceCategory(StringArrayIndexOfSensitive(CODES_TFhirMessageSignificanceCategory, FCategory.value));
end;

Procedure TFhirConformanceMessagingEvent.SetCategoryST(value : TFhirMessageSignificanceCategory);
begin
  if ord(value) = 0 then
    CategoryElement := nil
  else
    CategoryElement := TFhirEnum.create(CODES_TFhirMessageSignificanceCategory[value]);
end;

Procedure TFhirConformanceMessagingEvent.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirConformanceMessagingEvent.GetModeST : TFhirMessageConformanceEventMode;
begin
  if FMode = nil then
    result := TFhirMessageConformanceEventMode(0)
  else
    result := TFhirMessageConformanceEventMode(StringArrayIndexOfSensitive(CODES_TFhirMessageConformanceEventMode, FMode.value));
end;

Procedure TFhirConformanceMessagingEvent.SetModeST(value : TFhirMessageConformanceEventMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirMessageConformanceEventMode[value]);
end;

Procedure TFhirConformanceMessagingEvent.SetFocus(value : TFhirCode);
begin
  FFocus.free;
  FFocus := value;
end;

Function TFhirConformanceMessagingEvent.GetFocusST : String;
begin
  if FFocus = nil then
    result := ''
  else
    result := FFocus.value;
end;

Procedure TFhirConformanceMessagingEvent.SetFocusST(value : String);
begin
  if value <> '' then
  begin
    if FFocus = nil then
      FFocus := TFhirCode.create;
    FFocus.value := value
  end
  else if FFocus <> nil then
    FFocus.value := '';
end;

Procedure TFhirConformanceMessagingEvent.SetRequest(value : TFhirReference{TFhirStructureDefinition});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirConformanceMessagingEvent.SetResponse(value : TFhirReference{TFhirStructureDefinition});
begin
  FResponse.free;
  FResponse := value;
end;

Procedure TFhirConformanceMessagingEvent.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceMessagingEvent.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceMessagingEvent.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;


{ TFhirConformanceMessagingEventListEnumerator }

Constructor TFhirConformanceMessagingEventListEnumerator.Create(list : TFhirConformanceMessagingEventList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceMessagingEventListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceMessagingEventListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceMessagingEventListEnumerator.GetCurrent : TFhirConformanceMessagingEvent;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceMessagingEventList }
procedure TFhirConformanceMessagingEventList.AddItem(value: TFhirConformanceMessagingEvent);
begin
  assert(value.ClassName = 'TFhirConformanceMessagingEvent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceMessagingEvent');
  add(value);
end;


function TFhirConformanceMessagingEventList.Append: TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingEventList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceMessagingEventList.GetEnumerator : TFhirConformanceMessagingEventListEnumerator;
begin
  result := TFhirConformanceMessagingEventListEnumerator.Create(self.link);
end;

function TFhirConformanceMessagingEventList.Clone: TFhirConformanceMessagingEventList;
begin
  result := TFhirConformanceMessagingEventList(inherited Clone);
end;

function TFhirConformanceMessagingEventList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceMessagingEventList.GetItemN(index: Integer): TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingEventList.IndexOf(value: TFhirConformanceMessagingEvent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceMessagingEventList.Insert(index: Integer): TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceMessagingEventList.InsertItem(index: Integer; value: TFhirConformanceMessagingEvent);
begin
  assert(value is TFhirConformanceMessagingEvent);
  Inherited Insert(index, value);
end;

function TFhirConformanceMessagingEventList.Item(index: Integer): TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent(ObjectByIndex[index]);
end;

function TFhirConformanceMessagingEventList.Link: TFhirConformanceMessagingEventList;
begin
  result := TFhirConformanceMessagingEventList(inherited Link);
end;

procedure TFhirConformanceMessagingEventList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceMessagingEventList.SetItemByIndex(index: Integer; value: TFhirConformanceMessagingEvent);
begin
  assert(value is TFhirConformanceMessagingEvent);
  FhirConformanceMessagingEvents[index] := value;
end;

procedure TFhirConformanceMessagingEventList.SetItemN(index: Integer; value: TFhirConformanceMessagingEvent);
begin
  assert(value is TFhirConformanceMessagingEvent);
  ObjectByIndex[index] := value;
end;

{ TFhirConformanceDocument }

constructor TFhirConformanceDocument.Create;
begin
  inherited;
end;

destructor TFhirConformanceDocument.Destroy;
begin
  FMode.free;
  FDocumentation.free;
  FProfile.free;
  inherited;
end;

procedure TFhirConformanceDocument.Assign(oSource : TAdvObject);
begin
  inherited;
  FMode := TFhirConformanceDocument(oSource).FMode.Link;
  documentationElement := TFhirConformanceDocument(oSource).documentationElement.Clone;
  profile := TFhirConformanceDocument(oSource).profile.Clone;
end;

procedure TFhirConformanceDocument.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
  if (child_name = 'profile') Then
     list.add(FProfile.Link);
end;

procedure TFhirConformanceDocument.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(StructureDefinition)', FProfile.Link));{2}
end;

procedure TFhirConformanceDocument.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else if (propName = 'profile') then Profile := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else inherited;
end;

function TFhirConformanceDocument.FhirType : string;
begin
  result := 'document';
end;

function TFhirConformanceDocument.Link : TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument(inherited Link);
end;

function TFhirConformanceDocument.Clone : TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument(inherited Clone);
end;

function TFhirConformanceDocument.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceDocument;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirConformanceDocument)) then
    result := false
  else
  begin
    o := TFhirConformanceDocument(other);
    result := compareDeep(modeElement, o.modeElement, true) and compareDeep(documentationElement, o.documentationElement, true) and 
      compareDeep(profileElement, o.profileElement, true);
  end;
end;

function TFhirConformanceDocument.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirConformanceDocument;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirConformanceDocument)) then
    result := false
  else
  begin
    o := TFhirConformanceDocument(other);
    result := compareValues(modeElement, o.modeElement, true) and compareValues(documentationElement, o.documentationElement, true);
  end;
end;

{ TFhirConformanceDocument }

Procedure TFhirConformanceDocument.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirConformanceDocument.GetModeST : TFhirDocumentMode;
begin
  if FMode = nil then
    result := TFhirDocumentMode(0)
  else
    result := TFhirDocumentMode(StringArrayIndexOfSensitive(CODES_TFhirDocumentMode, FMode.value));
end;

Procedure TFhirConformanceDocument.SetModeST(value : TFhirDocumentMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirDocumentMode[value]);
end;

Procedure TFhirConformanceDocument.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirConformanceDocument.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirConformanceDocument.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;

Procedure TFhirConformanceDocument.SetProfile(value : TFhirReference{TFhirStructureDefinition});
begin
  FProfile.free;
  FProfile := value;
end;


{ TFhirConformanceDocumentListEnumerator }

Constructor TFhirConformanceDocumentListEnumerator.Create(list : TFhirConformanceDocumentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceDocumentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceDocumentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceDocumentListEnumerator.GetCurrent : TFhirConformanceDocument;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceDocumentList }
procedure TFhirConformanceDocumentList.AddItem(value: TFhirConformanceDocument);
begin
  assert(value.ClassName = 'TFhirConformanceDocument', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformanceDocument');
  add(value);
end;


function TFhirConformanceDocumentList.Append: TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceDocumentList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceDocumentList.GetEnumerator : TFhirConformanceDocumentListEnumerator;
begin
  result := TFhirConformanceDocumentListEnumerator.Create(self.link);
end;

function TFhirConformanceDocumentList.Clone: TFhirConformanceDocumentList;
begin
  result := TFhirConformanceDocumentList(inherited Clone);
end;

function TFhirConformanceDocumentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceDocumentList.GetItemN(index: Integer): TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument(ObjectByIndex[index]);
end;

function TFhirConformanceDocumentList.IndexOf(value: TFhirConformanceDocument): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceDocumentList.Insert(index: Integer): TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceDocumentList.InsertItem(index: Integer; value: TFhirConformanceDocument);
begin
  assert(value is TFhirConformanceDocument);
  Inherited Insert(index, value);
end;

function TFhirConformanceDocumentList.Item(index: Integer): TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument(ObjectByIndex[index]);
end;

function TFhirConformanceDocumentList.Link: TFhirConformanceDocumentList;
begin
  result := TFhirConformanceDocumentList(inherited Link);
end;

procedure TFhirConformanceDocumentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceDocumentList.SetItemByIndex(index: Integer; value: TFhirConformanceDocument);
begin
  assert(value is TFhirConformanceDocument);
  FhirConformanceDocuments[index] := value;
end;

procedure TFhirConformanceDocumentList.SetItemN(index: Integer; value: TFhirConformanceDocument);
begin
  assert(value is TFhirConformanceDocument);
  ObjectByIndex[index] := value;
end;

{ TFhirConformance }

constructor TFhirConformance.Create;
begin
  inherited;
end;

destructor TFhirConformance.Destroy;
begin
  FUrl.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FRequirements.free;
  FCopyright.free;
  FKind.free;
  FSoftware.free;
  FImplementation_.free;
  FFhirVersion.free;
  FAcceptUnknown.free;
  FFormatList.Free;
  FProfileList.Free;
  FRestList.Free;
  FMessagingList.Free;
  FDocumentList.Free;
  inherited;
end;

function TFhirConformance.GetResourceType : TFhirResourceType;
begin
  result := frtConformance;
end;

procedure TFhirConformance.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirConformance(oSource).urlElement.Clone;
  versionElement := TFhirConformance(oSource).versionElement.Clone;
  nameElement := TFhirConformance(oSource).nameElement.Clone;
  FStatus := TFhirConformance(oSource).FStatus.Link;
  experimentalElement := TFhirConformance(oSource).experimentalElement.Clone;
  publisherElement := TFhirConformance(oSource).publisherElement.Clone;
  if (TFhirConformance(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirConformanceContactList.Create;
    FContactList.Assign(TFhirConformance(oSource).FContactList);
  end;
  dateElement := TFhirConformance(oSource).dateElement.Clone;
  descriptionElement := TFhirConformance(oSource).descriptionElement.Clone;
  requirementsElement := TFhirConformance(oSource).requirementsElement.Clone;
  copyrightElement := TFhirConformance(oSource).copyrightElement.Clone;
  FKind := TFhirConformance(oSource).FKind.Link;
  software := TFhirConformance(oSource).software.Clone;
  implementation_ := TFhirConformance(oSource).implementation_.Clone;
  fhirVersionElement := TFhirConformance(oSource).fhirVersionElement.Clone;
  FAcceptUnknown := TFhirConformance(oSource).FAcceptUnknown.Link;
  if (TFhirConformance(oSource).FFormatList = nil) then
  begin
    FFormatList.free;
    FFormatList := nil;
  end
  else
  begin
    FFormatList := TFhirCodeList.Create;
    FFormatList.Assign(TFhirConformance(oSource).FFormatList);
  end;
  if (TFhirConformance(oSource).FProfileList = nil) then
  begin
    FProfileList.free;
    FProfileList := nil;
  end
  else
  begin
    FProfileList := TFhirReferenceList{TFhirStructureDefinition}.Create;
    FProfileList.Assign(TFhirConformance(oSource).FProfileList);
  end;
  if (TFhirConformance(oSource).FRestList = nil) then
  begin
    FRestList.free;
    FRestList := nil;
  end
  else
  begin
    FRestList := TFhirConformanceRestList.Create;
    FRestList.Assign(TFhirConformance(oSource).FRestList);
  end;
  if (TFhirConformance(oSource).FMessagingList = nil) then
  begin
    FMessagingList.free;
    FMessagingList := nil;
  end
  else
  begin
    FMessagingList := TFhirConformanceMessagingList.Create;
    FMessagingList.Assign(TFhirConformance(oSource).FMessagingList);
  end;
  if (TFhirConformance(oSource).FDocumentList = nil) then
  begin
    FDocumentList.free;
    FDocumentList := nil;
  end
  else
  begin
    FDocumentList := TFhirConformanceDocumentList.Create;
    FDocumentList.Assign(TFhirConformance(oSource).FDocumentList);
  end;
end;

procedure TFhirConformance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'software') Then
     list.add(FSoftware.Link);
  if (child_name = 'implementation') Then
     list.add(FImplementation_.Link);
  if (child_name = 'fhirVersion') Then
     list.add(FFhirVersion.Link);
  if (child_name = 'acceptUnknown') Then
     list.add(FAcceptUnknown.Link);
  if (child_name = 'format') Then
    list.addAll(FFormatList);
  if (child_name = 'profile') Then
    list.addAll(FProfileList);
  if (child_name = 'rest') Then
    list.addAll(FRestList);
  if (child_name = 'messaging') Then
    list.addAll(FMessagingList);
  if (child_name = 'document') Then
    list.addAll(FDocumentList);
end;

procedure TFhirConformance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'kind', 'code', FKind.Link));{1}
  oList.add(TFHIRProperty.create(self, 'software', '', FSoftware.Link));{2}
  oList.add(TFHIRProperty.create(self, 'implementation', '', FImplementation_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fhirVersion', 'id', FFhirVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'acceptUnknown', 'code', FAcceptUnknown.Link));{1}
  oList.add(TFHIRProperty.create(self, 'format', 'code', FFormatList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(StructureDefinition)', FProfileList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'rest', '', FRestList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'messaging', '', FMessagingList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'document', '', FDocumentList.Link)){3};
end;

procedure TFhirConformance.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirConformanceContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'kind') then KindElement := propValue as TFHIREnum
  else if (propName = 'software') then Software := propValue as TFhirConformanceSoftware{4b}
  else if (propName = 'implementation') then Implementation_ := propValue as TFhirConformanceImplementation{4b}
  else if (propName = 'fhirVersion') then FhirVersionElement := propValue as TFhirId{5a}
  else if (propName = 'acceptUnknown') then AcceptUnknownElement := propValue as TFHIREnum
  else if (propName = 'format') then FormatList.add(propValue as TFhirCode){2}
  else if (propName = 'profile') then ProfileList.add(propValue as TFhirReference{TFhirStructureDefinition}){2}
  else if (propName = 'rest') then RestList.add(propValue as TFhirConformanceRest){2}
  else if (propName = 'messaging') then MessagingList.add(propValue as TFhirConformanceMessaging){2}
  else if (propName = 'document') then DocumentList.add(propValue as TFhirConformanceDocument){2}
  else inherited;
end;

function TFhirConformance.FhirType : string;
begin
  result := 'Conformance';
end;

function TFhirConformance.Link : TFhirConformance;
begin
  result := TFhirConformance(inherited Link);
end;

function TFhirConformance.Clone : TFhirConformance;
begin
  result := TFhirConformance(inherited Clone);
end;

{ TFhirConformance }

Procedure TFhirConformance.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirConformance.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirConformance.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirConformance.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirConformance.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirConformance.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirConformance.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirConformance.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirConformance.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirConformance.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirConformance.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirConformance.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirConformance.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirConformance.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirConformance.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirConformance.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirConformance.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirConformance.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirConformance.GetContactList : TFhirConformanceContactList;
begin
  if FContactList = nil then
    FContactList := TFhirConformanceContactList.Create;
  result := FContactList;
end;

Function TFhirConformance.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirConformance.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirConformance.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirConformance.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirConformance.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirConformance.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirConformance.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirConformance.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirConformance.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirConformance.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirConformance.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirConformance.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirConformance.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirConformance.SetKind(value : TFhirEnum);
begin
  FKind.free;
  FKind := value;
end;

Function TFhirConformance.GetKindST : TFhirConformanceStatementKind;
begin
  if FKind = nil then
    result := TFhirConformanceStatementKind(0)
  else
    result := TFhirConformanceStatementKind(StringArrayIndexOfSensitive(CODES_TFhirConformanceStatementKind, FKind.value));
end;

Procedure TFhirConformance.SetKindST(value : TFhirConformanceStatementKind);
begin
  if ord(value) = 0 then
    KindElement := nil
  else
    KindElement := TFhirEnum.create(CODES_TFhirConformanceStatementKind[value]);
end;

Procedure TFhirConformance.SetSoftware(value : TFhirConformanceSoftware);
begin
  FSoftware.free;
  FSoftware := value;
end;

Procedure TFhirConformance.SetImplementation_(value : TFhirConformanceImplementation);
begin
  FImplementation_.free;
  FImplementation_ := value;
end;

Procedure TFhirConformance.SetFhirVersion(value : TFhirId);
begin
  FFhirVersion.free;
  FFhirVersion := value;
end;

Function TFhirConformance.GetFhirVersionST : String;
begin
  if FFhirVersion = nil then
    result := ''
  else
    result := FFhirVersion.value;
end;

Procedure TFhirConformance.SetFhirVersionST(value : String);
begin
  if value <> '' then
  begin
    if FFhirVersion = nil then
      FFhirVersion := TFhirId.create;
    FFhirVersion.value := value
  end
  else if FFhirVersion <> nil then
    FFhirVersion.value := '';
end;

Procedure TFhirConformance.SetAcceptUnknown(value : TFhirEnum);
begin
  FAcceptUnknown.free;
  FAcceptUnknown := value;
end;

Function TFhirConformance.GetAcceptUnknownST : TFhirUnknownContentCode;
begin
  if FAcceptUnknown = nil then
    result := TFhirUnknownContentCode(0)
  else
    result := TFhirUnknownContentCode(StringArrayIndexOfSensitive(CODES_TFhirUnknownContentCode, FAcceptUnknown.value));
end;

Procedure TFhirConformance.SetAcceptUnknownST(value : TFhirUnknownContentCode);
begin
  if ord(value) = 0 then
    AcceptUnknownElement := nil
  else
    AcceptUnknownElement := TFhirEnum.create(CODES_TFhirUnknownContentCode[value]);
end;

Function TFhirConformance.GetFormatList : TFhirCodeList;
begin
  if FFormatList = nil then
    FFormatList := TFhirCodeList.Create;
  result := FFormatList;
end;

Function TFhirConformance.GetHasFormatList : boolean;
begin
  result := (FFormatList <> nil) and (FFormatList.count > 0);
end;

Function TFhirConformance.GetProfileList : TFhirReferenceList{TFhirStructureDefinition};
begin
  if FProfileList = nil then
    FProfileList := TFhirReferenceList{TFhirStructureDefinition}.Create;
  result := FProfileList;
end;

Function TFhirConformance.GetHasProfileList : boolean;
begin
  result := (FProfileList <> nil) and (FProfileList.count > 0);
end;

Function TFhirConformance.GetRestList : TFhirConformanceRestList;
begin
  if FRestList = nil then
    FRestList := TFhirConformanceRestList.Create;
  result := FRestList;
end;

Function TFhirConformance.GetHasRestList : boolean;
begin
  result := (FRestList <> nil) and (FRestList.count > 0);
end;

Function TFhirConformance.GetMessagingList : TFhirConformanceMessagingList;
begin
  if FMessagingList = nil then
    FMessagingList := TFhirConformanceMessagingList.Create;
  result := FMessagingList;
end;

Function TFhirConformance.GetHasMessagingList : boolean;
begin
  result := (FMessagingList <> nil) and (FMessagingList.count > 0);
end;

Function TFhirConformance.GetDocumentList : TFhirConformanceDocumentList;
begin
  if FDocumentList = nil then
    FDocumentList := TFhirConformanceDocumentList.Create;
  result := FDocumentList;
end;

Function TFhirConformance.GetHasDocumentList : boolean;
begin
  result := (FDocumentList <> nil) and (FDocumentList.count > 0);
end;


{ TFhirConformanceListEnumerator }

Constructor TFhirConformanceListEnumerator.Create(list : TFhirConformanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirConformanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirConformanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirConformanceListEnumerator.GetCurrent : TFhirConformance;
begin
  Result := FList[FIndex];
end;


{ TFhirConformanceList }
procedure TFhirConformanceList.AddItem(value: TFhirConformance);
begin
  assert(value.ClassName = 'TFhirConformance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirConformance');
  add(value);
end;


function TFhirConformanceList.Append: TFhirConformance;
begin
  result := TFhirConformance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceList.ClearItems;
begin
  Clear;
end;

function TFhirConformanceList.GetEnumerator : TFhirConformanceListEnumerator;
begin
  result := TFhirConformanceListEnumerator.Create(self.link);
end;

function TFhirConformanceList.Clone: TFhirConformanceList;
begin
  result := TFhirConformanceList(inherited Clone);
end;

function TFhirConformanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirConformanceList.GetItemN(index: Integer): TFhirConformance;
begin
  result := TFhirConformance(ObjectByIndex[index]);
end;

function TFhirConformanceList.IndexOf(value: TFhirConformance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirConformanceList.Insert(index: Integer): TFhirConformance;
begin
  result := TFhirConformance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirConformanceList.InsertItem(index: Integer; value: TFhirConformance);
begin
  assert(value is TFhirConformance);
  Inherited Insert(index, value);
end;

function TFhirConformanceList.Item(index: Integer): TFhirConformance;
begin
  result := TFhirConformance(ObjectByIndex[index]);
end;

function TFhirConformanceList.Link: TFhirConformanceList;
begin
  result := TFhirConformanceList(inherited Link);
end;

procedure TFhirConformanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirConformanceList.SetItemByIndex(index: Integer; value: TFhirConformance);
begin
  assert(value is TFhirConformance);
  FhirConformances[index] := value;
end;

procedure TFhirConformanceList.SetItemN(index: Integer; value: TFhirConformance);
begin
  assert(value is TFhirConformance);
  ObjectByIndex[index] := value;
end;

{ TFhirContractActor }

constructor TFhirContractActor.Create;
begin
  inherited;
end;

destructor TFhirContractActor.Destroy;
begin
  FEntity.free;
  FRoleList.Free;
  inherited;
end;

procedure TFhirContractActor.Assign(oSource : TAdvObject);
begin
  inherited;
  entity := TFhirContractActor(oSource).entity.Clone;
  if (TFhirContractActor(oSource).FRoleList = nil) then
  begin
    FRoleList.free;
    FRoleList := nil;
  end
  else
  begin
    FRoleList := TFhirCodeableConceptList.Create;
    FRoleList.Assign(TFhirContractActor(oSource).FRoleList);
  end;
end;

procedure TFhirContractActor.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'entity') Then
     list.add(FEntity.Link);
  if (child_name = 'role') Then
    list.addAll(FRoleList);
end;

procedure TFhirContractActor.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'entity', 'Reference(Contract|Device|Group|Location|Organization|Patient|Practitioner|RelatedPerson|Substance)', FEntity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRoleList.Link)){3};
end;

procedure TFhirContractActor.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'entity') then Entity := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'role') then RoleList.add(propValue as TFhirCodeableConcept){2}
  else inherited;
end;

function TFhirContractActor.FhirType : string;
begin
  result := 'actor';
end;

function TFhirContractActor.Link : TFhirContractActor;
begin
  result := TFhirContractActor(inherited Link);
end;

function TFhirContractActor.Clone : TFhirContractActor;
begin
  result := TFhirContractActor(inherited Clone);
end;

function TFhirContractActor.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractActor;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractActor)) then
    result := false
  else
  begin
    o := TFhirContractActor(other);
    result := compareDeep(entityElement, o.entityElement, true) and compareDeep(roleList, o.roleList, true);
  end;
end;

function TFhirContractActor.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractActor;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractActor)) then
    result := false
  else
  begin
    o := TFhirContractActor(other);
    result := true;
  end;
end;

{ TFhirContractActor }

Procedure TFhirContractActor.SetEntity(value : TFhirReference{Resource});
begin
  FEntity.free;
  FEntity := value;
end;

Function TFhirContractActor.GetRoleList : TFhirCodeableConceptList;
begin
  if FRoleList = nil then
    FRoleList := TFhirCodeableConceptList.Create;
  result := FRoleList;
end;

Function TFhirContractActor.GetHasRoleList : boolean;
begin
  result := (FRoleList <> nil) and (FRoleList.count > 0);
end;


{ TFhirContractActorListEnumerator }

Constructor TFhirContractActorListEnumerator.Create(list : TFhirContractActorList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractActorListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractActorListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractActorListEnumerator.GetCurrent : TFhirContractActor;
begin
  Result := FList[FIndex];
end;


{ TFhirContractActorList }
procedure TFhirContractActorList.AddItem(value: TFhirContractActor);
begin
  assert(value.ClassName = 'TFhirContractActor', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractActor');
  add(value);
end;


function TFhirContractActorList.Append: TFhirContractActor;
begin
  result := TFhirContractActor.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractActorList.ClearItems;
begin
  Clear;
end;

function TFhirContractActorList.GetEnumerator : TFhirContractActorListEnumerator;
begin
  result := TFhirContractActorListEnumerator.Create(self.link);
end;

function TFhirContractActorList.Clone: TFhirContractActorList;
begin
  result := TFhirContractActorList(inherited Clone);
end;

function TFhirContractActorList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractActorList.GetItemN(index: Integer): TFhirContractActor;
begin
  result := TFhirContractActor(ObjectByIndex[index]);
end;

function TFhirContractActorList.IndexOf(value: TFhirContractActor): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractActorList.Insert(index: Integer): TFhirContractActor;
begin
  result := TFhirContractActor.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractActorList.InsertItem(index: Integer; value: TFhirContractActor);
begin
  assert(value is TFhirContractActor);
  Inherited Insert(index, value);
end;

function TFhirContractActorList.Item(index: Integer): TFhirContractActor;
begin
  result := TFhirContractActor(ObjectByIndex[index]);
end;

function TFhirContractActorList.Link: TFhirContractActorList;
begin
  result := TFhirContractActorList(inherited Link);
end;

procedure TFhirContractActorList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractActorList.SetItemByIndex(index: Integer; value: TFhirContractActor);
begin
  assert(value is TFhirContractActor);
  FhirContractActors[index] := value;
end;

procedure TFhirContractActorList.SetItemN(index: Integer; value: TFhirContractActor);
begin
  assert(value is TFhirContractActor);
  ObjectByIndex[index] := value;
end;

{ TFhirContractValuedItem }

constructor TFhirContractValuedItem.Create;
begin
  inherited;
end;

destructor TFhirContractValuedItem.Destroy;
begin
  FEntity.free;
  FIdentifier.free;
  FEffectiveTime.free;
  FQuantity.free;
  FUnitPrice.free;
  FFactor.free;
  FPoints.free;
  FNet.free;
  inherited;
end;

procedure TFhirContractValuedItem.Assign(oSource : TAdvObject);
begin
  inherited;
  entity := TFhirContractValuedItem(oSource).entity.Clone;
  identifier := TFhirContractValuedItem(oSource).identifier.Clone;
  effectiveTimeElement := TFhirContractValuedItem(oSource).effectiveTimeElement.Clone;
  quantity := TFhirContractValuedItem(oSource).quantity.Clone;
  unitPrice := TFhirContractValuedItem(oSource).unitPrice.Clone;
  factorElement := TFhirContractValuedItem(oSource).factorElement.Clone;
  pointsElement := TFhirContractValuedItem(oSource).pointsElement.Clone;
  net := TFhirContractValuedItem(oSource).net.Clone;
end;

procedure TFhirContractValuedItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'entity[x]') Then
     list.add(FEntity.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'effectiveTime') Then
     list.add(FEffectiveTime.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'unitPrice') Then
     list.add(FUnitPrice.Link);
  if (child_name = 'factor') Then
     list.add(FFactor.Link);
  if (child_name = 'points') Then
     list.add(FPoints.Link);
  if (child_name = 'net') Then
     list.add(FNet.Link);
end;

procedure TFhirContractValuedItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'entity[x]', 'CodeableConcept|Reference(Any)', FEntity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'effectiveTime', 'dateTime', FEffectiveTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unitPrice', 'Money', FUnitPrice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'factor', 'decimal', FFactor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'points', 'decimal', FPoints.Link));{2}
  oList.add(TFHIRProperty.create(self, 'net', 'Money', FNet.Link));{2}
end;

procedure TFhirContractValuedItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('entity')) then Entity := propValue as TFhirType{4}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'effectiveTime') then EffectiveTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'unitPrice') then UnitPrice := propValue as TFhirQuantity{4b}
  else if (propName = 'factor') then FactorElement := propValue as TFhirDecimal{5a}
  else if (propName = 'points') then PointsElement := propValue as TFhirDecimal{5a}
  else if (propName = 'net') then Net := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirContractValuedItem.FhirType : string;
begin
  result := 'valuedItem';
end;

function TFhirContractValuedItem.Link : TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem(inherited Link);
end;

function TFhirContractValuedItem.Clone : TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem(inherited Clone);
end;

function TFhirContractValuedItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractValuedItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractValuedItem)) then
    result := false
  else
  begin
    o := TFhirContractValuedItem(other);
    result := compareDeep(entityElement, o.entityElement, true) and compareDeep(identifierElement, o.identifierElement, true) and 
      compareDeep(effectiveTimeElement, o.effectiveTimeElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(unitPriceElement, o.unitPriceElement, true) and compareDeep(factorElement, o.factorElement, true) and 
      compareDeep(pointsElement, o.pointsElement, true) and compareDeep(netElement, o.netElement, true);
  end;
end;

function TFhirContractValuedItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractValuedItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractValuedItem)) then
    result := false
  else
  begin
    o := TFhirContractValuedItem(other);
    result := compareValues(effectiveTimeElement, o.effectiveTimeElement, true) and 
      compareValues(factorElement, o.factorElement, true) and compareValues(pointsElement, o.pointsElement, true);
  end;
end;

{ TFhirContractValuedItem }

Procedure TFhirContractValuedItem.SetEntity(value : TFhirType);
begin
  FEntity.free;
  FEntity := value;
end;

Procedure TFhirContractValuedItem.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirContractValuedItem.SetEffectiveTime(value : TFhirDateTime);
begin
  FEffectiveTime.free;
  FEffectiveTime := value;
end;

Function TFhirContractValuedItem.GetEffectiveTimeST : TDateAndTime;
begin
  if FEffectiveTime = nil then
    result := nil
  else
    result := FEffectiveTime.value;
end;

Procedure TFhirContractValuedItem.SetEffectiveTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEffectiveTime = nil then
      FEffectiveTime := TFhirDateTime.create;
    FEffectiveTime.value := value
  end
  else if FEffectiveTime <> nil then
    FEffectiveTime.value := nil;
end;

Procedure TFhirContractValuedItem.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirContractValuedItem.SetUnitPrice(value : TFhirQuantity);
begin
  FUnitPrice.free;
  FUnitPrice := value;
end;

Procedure TFhirContractValuedItem.SetFactor(value : TFhirDecimal);
begin
  FFactor.free;
  FFactor := value;
end;

Function TFhirContractValuedItem.GetFactorST : String;
begin
  if FFactor = nil then
    result := ''
  else
    result := FFactor.value;
end;

Procedure TFhirContractValuedItem.SetFactorST(value : String);
begin
  if value <> '' then
  begin
    if FFactor = nil then
      FFactor := TFhirDecimal.create;
    FFactor.value := value
  end
  else if FFactor <> nil then
    FFactor.value := '';
end;

Procedure TFhirContractValuedItem.SetPoints(value : TFhirDecimal);
begin
  FPoints.free;
  FPoints := value;
end;

Function TFhirContractValuedItem.GetPointsST : String;
begin
  if FPoints = nil then
    result := ''
  else
    result := FPoints.value;
end;

Procedure TFhirContractValuedItem.SetPointsST(value : String);
begin
  if value <> '' then
  begin
    if FPoints = nil then
      FPoints := TFhirDecimal.create;
    FPoints.value := value
  end
  else if FPoints <> nil then
    FPoints.value := '';
end;

Procedure TFhirContractValuedItem.SetNet(value : TFhirQuantity);
begin
  FNet.free;
  FNet := value;
end;


{ TFhirContractValuedItemListEnumerator }

Constructor TFhirContractValuedItemListEnumerator.Create(list : TFhirContractValuedItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractValuedItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractValuedItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractValuedItemListEnumerator.GetCurrent : TFhirContractValuedItem;
begin
  Result := FList[FIndex];
end;


{ TFhirContractValuedItemList }
procedure TFhirContractValuedItemList.AddItem(value: TFhirContractValuedItem);
begin
  assert(value.ClassName = 'TFhirContractValuedItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractValuedItem');
  add(value);
end;


function TFhirContractValuedItemList.Append: TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractValuedItemList.ClearItems;
begin
  Clear;
end;

function TFhirContractValuedItemList.GetEnumerator : TFhirContractValuedItemListEnumerator;
begin
  result := TFhirContractValuedItemListEnumerator.Create(self.link);
end;

function TFhirContractValuedItemList.Clone: TFhirContractValuedItemList;
begin
  result := TFhirContractValuedItemList(inherited Clone);
end;

function TFhirContractValuedItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractValuedItemList.GetItemN(index: Integer): TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem(ObjectByIndex[index]);
end;

function TFhirContractValuedItemList.IndexOf(value: TFhirContractValuedItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractValuedItemList.Insert(index: Integer): TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractValuedItemList.InsertItem(index: Integer; value: TFhirContractValuedItem);
begin
  assert(value is TFhirContractValuedItem);
  Inherited Insert(index, value);
end;

function TFhirContractValuedItemList.Item(index: Integer): TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem(ObjectByIndex[index]);
end;

function TFhirContractValuedItemList.Link: TFhirContractValuedItemList;
begin
  result := TFhirContractValuedItemList(inherited Link);
end;

procedure TFhirContractValuedItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractValuedItemList.SetItemByIndex(index: Integer; value: TFhirContractValuedItem);
begin
  assert(value is TFhirContractValuedItem);
  FhirContractValuedItems[index] := value;
end;

procedure TFhirContractValuedItemList.SetItemN(index: Integer; value: TFhirContractValuedItem);
begin
  assert(value is TFhirContractValuedItem);
  ObjectByIndex[index] := value;
end;

{ TFhirContractSigner }

constructor TFhirContractSigner.Create;
begin
  inherited;
end;

destructor TFhirContractSigner.Destroy;
begin
  FType_.free;
  FParty.free;
  FSignature.free;
  inherited;
end;

procedure TFhirContractSigner.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirContractSigner(oSource).type_.Clone;
  party := TFhirContractSigner(oSource).party.Clone;
  signatureElement := TFhirContractSigner(oSource).signatureElement.Clone;
end;

procedure TFhirContractSigner.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'party') Then
     list.add(FParty.Link);
  if (child_name = 'signature') Then
     list.add(FSignature.Link);
end;

procedure TFhirContractSigner.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'party', 'Reference(Organization|Patient|Practitioner|RelatedPerson)', FParty.Link));{2}
  oList.add(TFHIRProperty.create(self, 'signature', 'string', FSignature.Link));{2}
end;

procedure TFhirContractSigner.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'party') then Party := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'signature') then SignatureElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirContractSigner.FhirType : string;
begin
  result := 'signer';
end;

function TFhirContractSigner.Link : TFhirContractSigner;
begin
  result := TFhirContractSigner(inherited Link);
end;

function TFhirContractSigner.Clone : TFhirContractSigner;
begin
  result := TFhirContractSigner(inherited Clone);
end;

function TFhirContractSigner.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractSigner;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractSigner)) then
    result := false
  else
  begin
    o := TFhirContractSigner(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(partyElement, o.partyElement, true) and 
      compareDeep(signatureElement, o.signatureElement, true);
  end;
end;

function TFhirContractSigner.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractSigner;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractSigner)) then
    result := false
  else
  begin
    o := TFhirContractSigner(other);
    result := compareValues(signatureElement, o.signatureElement, true);
  end;
end;

{ TFhirContractSigner }

Procedure TFhirContractSigner.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirContractSigner.SetParty(value : TFhirReference{Resource});
begin
  FParty.free;
  FParty := value;
end;

Procedure TFhirContractSigner.SetSignature(value : TFhirString);
begin
  FSignature.free;
  FSignature := value;
end;

Function TFhirContractSigner.GetSignatureST : String;
begin
  if FSignature = nil then
    result := ''
  else
    result := FSignature.value;
end;

Procedure TFhirContractSigner.SetSignatureST(value : String);
begin
  if value <> '' then
  begin
    if FSignature = nil then
      FSignature := TFhirString.create;
    FSignature.value := value
  end
  else if FSignature <> nil then
    FSignature.value := '';
end;


{ TFhirContractSignerListEnumerator }

Constructor TFhirContractSignerListEnumerator.Create(list : TFhirContractSignerList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractSignerListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractSignerListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractSignerListEnumerator.GetCurrent : TFhirContractSigner;
begin
  Result := FList[FIndex];
end;


{ TFhirContractSignerList }
procedure TFhirContractSignerList.AddItem(value: TFhirContractSigner);
begin
  assert(value.ClassName = 'TFhirContractSigner', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractSigner');
  add(value);
end;


function TFhirContractSignerList.Append: TFhirContractSigner;
begin
  result := TFhirContractSigner.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractSignerList.ClearItems;
begin
  Clear;
end;

function TFhirContractSignerList.GetEnumerator : TFhirContractSignerListEnumerator;
begin
  result := TFhirContractSignerListEnumerator.Create(self.link);
end;

function TFhirContractSignerList.Clone: TFhirContractSignerList;
begin
  result := TFhirContractSignerList(inherited Clone);
end;

function TFhirContractSignerList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractSignerList.GetItemN(index: Integer): TFhirContractSigner;
begin
  result := TFhirContractSigner(ObjectByIndex[index]);
end;

function TFhirContractSignerList.IndexOf(value: TFhirContractSigner): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractSignerList.Insert(index: Integer): TFhirContractSigner;
begin
  result := TFhirContractSigner.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractSignerList.InsertItem(index: Integer; value: TFhirContractSigner);
begin
  assert(value is TFhirContractSigner);
  Inherited Insert(index, value);
end;

function TFhirContractSignerList.Item(index: Integer): TFhirContractSigner;
begin
  result := TFhirContractSigner(ObjectByIndex[index]);
end;

function TFhirContractSignerList.Link: TFhirContractSignerList;
begin
  result := TFhirContractSignerList(inherited Link);
end;

procedure TFhirContractSignerList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractSignerList.SetItemByIndex(index: Integer; value: TFhirContractSigner);
begin
  assert(value is TFhirContractSigner);
  FhirContractSigners[index] := value;
end;

procedure TFhirContractSignerList.SetItemN(index: Integer; value: TFhirContractSigner);
begin
  assert(value is TFhirContractSigner);
  ObjectByIndex[index] := value;
end;

{ TFhirContractTerm }

constructor TFhirContractTerm.Create;
begin
  inherited;
end;

destructor TFhirContractTerm.Destroy;
begin
  FIdentifier.free;
  FIssued.free;
  FApplies.free;
  FType_.free;
  FSubType.free;
  FSubject.free;
  FActionList.Free;
  FActionReasonList.Free;
  FActorList.Free;
  FText.free;
  FValuedItemList.Free;
  FGroupList.Free;
  inherited;
end;

procedure TFhirContractTerm.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirContractTerm(oSource).identifier.Clone;
  issuedElement := TFhirContractTerm(oSource).issuedElement.Clone;
  applies := TFhirContractTerm(oSource).applies.Clone;
  type_ := TFhirContractTerm(oSource).type_.Clone;
  subType := TFhirContractTerm(oSource).subType.Clone;
  subject := TFhirContractTerm(oSource).subject.Clone;
  if (TFhirContractTerm(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirCodeableConceptList.Create;
    FActionList.Assign(TFhirContractTerm(oSource).FActionList);
  end;
  if (TFhirContractTerm(oSource).FActionReasonList = nil) then
  begin
    FActionReasonList.free;
    FActionReasonList := nil;
  end
  else
  begin
    FActionReasonList := TFhirCodeableConceptList.Create;
    FActionReasonList.Assign(TFhirContractTerm(oSource).FActionReasonList);
  end;
  if (TFhirContractTerm(oSource).FActorList = nil) then
  begin
    FActorList.free;
    FActorList := nil;
  end
  else
  begin
    FActorList := TFhirContractTermActorList.Create;
    FActorList.Assign(TFhirContractTerm(oSource).FActorList);
  end;
  textElement := TFhirContractTerm(oSource).textElement.Clone;
  if (TFhirContractTerm(oSource).FValuedItemList = nil) then
  begin
    FValuedItemList.free;
    FValuedItemList := nil;
  end
  else
  begin
    FValuedItemList := TFhirContractTermValuedItemList.Create;
    FValuedItemList.Assign(TFhirContractTerm(oSource).FValuedItemList);
  end;
  if (TFhirContractTerm(oSource).FGroupList = nil) then
  begin
    FGroupList.free;
    FGroupList := nil;
  end
  else
  begin
    FGroupList := TFhirContractTermList.Create;
    FGroupList.Assign(TFhirContractTerm(oSource).FGroupList);
  end;
end;

procedure TFhirContractTerm.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'issued') Then
     list.add(FIssued.Link);
  if (child_name = 'applies') Then
     list.add(FApplies.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'subType') Then
     list.add(FSubType.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'action') Then
    list.addAll(FActionList);
  if (child_name = 'actionReason') Then
    list.addAll(FActionReasonList);
  if (child_name = 'actor') Then
    list.addAll(FActorList);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'valuedItem') Then
    list.addAll(FValuedItemList);
  if (child_name = 'group') Then
    list.addAll(FGroupList);
end;

procedure TFhirContractTerm.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'issued', 'dateTime', FIssued.Link));{2}
  oList.add(TFHIRProperty.create(self, 'applies', 'Period', FApplies.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subType', 'CodeableConcept', FSubType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'action', 'CodeableConcept', FActionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actionReason', 'CodeableConcept', FActionReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actor', '', FActorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'valuedItem', '', FValuedItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'group', '@Contract.term', FGroupList.Link)){3};
end;

procedure TFhirContractTerm.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'issued') then IssuedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'applies') then Applies := propValue as TFhirPeriod{4b}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subType') then SubType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'action') then ActionList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actionReason') then ActionReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actor') then ActorList.add(propValue as TFhirContractTermActor){2}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'valuedItem') then ValuedItemList.add(propValue as TFhirContractTermValuedItem){2}
  else if (propName = 'group') then GroupList.add(propValue as TFhirContractTerm){2}
  else inherited;
end;

function TFhirContractTerm.FhirType : string;
begin
  result := 'term';
end;

function TFhirContractTerm.Link : TFhirContractTerm;
begin
  result := TFhirContractTerm(inherited Link);
end;

function TFhirContractTerm.Clone : TFhirContractTerm;
begin
  result := TFhirContractTerm(inherited Clone);
end;

function TFhirContractTerm.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTerm;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractTerm)) then
    result := false
  else
  begin
    o := TFhirContractTerm(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(issuedElement, o.issuedElement, true) and 
      compareDeep(appliesElement, o.appliesElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(subTypeElement, o.subTypeElement, true) and compareDeep(subjectElement, o.subjectElement, true) and 
      compareDeep(actionList, o.actionList, true) and compareDeep(actionReasonList, o.actionReasonList, true) and 
      compareDeep(actorList, o.actorList, true) and compareDeep(textElement, o.textElement, true) and 
      compareDeep(valuedItemList, o.valuedItemList, true) and compareDeep(groupList, o.groupList, true);
  end;
end;

function TFhirContractTerm.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTerm;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractTerm)) then
    result := false
  else
  begin
    o := TFhirContractTerm(other);
    result := compareValues(issuedElement, o.issuedElement, true) and compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirContractTerm }

Procedure TFhirContractTerm.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirContractTerm.SetIssued(value : TFhirDateTime);
begin
  FIssued.free;
  FIssued := value;
end;

Function TFhirContractTerm.GetIssuedST : TDateAndTime;
begin
  if FIssued = nil then
    result := nil
  else
    result := FIssued.value;
end;

Procedure TFhirContractTerm.SetIssuedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIssued = nil then
      FIssued := TFhirDateTime.create;
    FIssued.value := value
  end
  else if FIssued <> nil then
    FIssued.value := nil;
end;

Procedure TFhirContractTerm.SetApplies(value : TFhirPeriod);
begin
  FApplies.free;
  FApplies := value;
end;

Procedure TFhirContractTerm.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirContractTerm.SetSubType(value : TFhirCodeableConcept);
begin
  FSubType.free;
  FSubType := value;
end;

Procedure TFhirContractTerm.SetSubject(value : TFhirReference{TFhirReference});
begin
  FSubject.free;
  FSubject := value;
end;

Function TFhirContractTerm.GetActionList : TFhirCodeableConceptList;
begin
  if FActionList = nil then
    FActionList := TFhirCodeableConceptList.Create;
  result := FActionList;
end;

Function TFhirContractTerm.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;

Function TFhirContractTerm.GetActionReasonList : TFhirCodeableConceptList;
begin
  if FActionReasonList = nil then
    FActionReasonList := TFhirCodeableConceptList.Create;
  result := FActionReasonList;
end;

Function TFhirContractTerm.GetHasActionReasonList : boolean;
begin
  result := (FActionReasonList <> nil) and (FActionReasonList.count > 0);
end;

Function TFhirContractTerm.GetActorList : TFhirContractTermActorList;
begin
  if FActorList = nil then
    FActorList := TFhirContractTermActorList.Create;
  result := FActorList;
end;

Function TFhirContractTerm.GetHasActorList : boolean;
begin
  result := (FActorList <> nil) and (FActorList.count > 0);
end;

Procedure TFhirContractTerm.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirContractTerm.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirContractTerm.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Function TFhirContractTerm.GetValuedItemList : TFhirContractTermValuedItemList;
begin
  if FValuedItemList = nil then
    FValuedItemList := TFhirContractTermValuedItemList.Create;
  result := FValuedItemList;
end;

Function TFhirContractTerm.GetHasValuedItemList : boolean;
begin
  result := (FValuedItemList <> nil) and (FValuedItemList.count > 0);
end;

Function TFhirContractTerm.GetGroupList : TFhirContractTermList;
begin
  if FGroupList = nil then
    FGroupList := TFhirContractTermList.Create;
  result := FGroupList;
end;

Function TFhirContractTerm.GetHasGroupList : boolean;
begin
  result := (FGroupList <> nil) and (FGroupList.count > 0);
end;


{ TFhirContractTermListEnumerator }

Constructor TFhirContractTermListEnumerator.Create(list : TFhirContractTermList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractTermListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractTermListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractTermListEnumerator.GetCurrent : TFhirContractTerm;
begin
  Result := FList[FIndex];
end;


{ TFhirContractTermList }
procedure TFhirContractTermList.AddItem(value: TFhirContractTerm);
begin
  assert(value.ClassName = 'TFhirContractTerm', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractTerm');
  add(value);
end;


function TFhirContractTermList.Append: TFhirContractTerm;
begin
  result := TFhirContractTerm.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermList.ClearItems;
begin
  Clear;
end;

function TFhirContractTermList.GetEnumerator : TFhirContractTermListEnumerator;
begin
  result := TFhirContractTermListEnumerator.Create(self.link);
end;

function TFhirContractTermList.Clone: TFhirContractTermList;
begin
  result := TFhirContractTermList(inherited Clone);
end;

function TFhirContractTermList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractTermList.GetItemN(index: Integer): TFhirContractTerm;
begin
  result := TFhirContractTerm(ObjectByIndex[index]);
end;

function TFhirContractTermList.IndexOf(value: TFhirContractTerm): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractTermList.Insert(index: Integer): TFhirContractTerm;
begin
  result := TFhirContractTerm.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermList.InsertItem(index: Integer; value: TFhirContractTerm);
begin
  assert(value is TFhirContractTerm);
  Inherited Insert(index, value);
end;

function TFhirContractTermList.Item(index: Integer): TFhirContractTerm;
begin
  result := TFhirContractTerm(ObjectByIndex[index]);
end;

function TFhirContractTermList.Link: TFhirContractTermList;
begin
  result := TFhirContractTermList(inherited Link);
end;

procedure TFhirContractTermList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractTermList.SetItemByIndex(index: Integer; value: TFhirContractTerm);
begin
  assert(value is TFhirContractTerm);
  FhirContractTerms[index] := value;
end;

procedure TFhirContractTermList.SetItemN(index: Integer; value: TFhirContractTerm);
begin
  assert(value is TFhirContractTerm);
  ObjectByIndex[index] := value;
end;

{ TFhirContractTermActor }

constructor TFhirContractTermActor.Create;
begin
  inherited;
end;

destructor TFhirContractTermActor.Destroy;
begin
  FEntity.free;
  FRoleList.Free;
  inherited;
end;

procedure TFhirContractTermActor.Assign(oSource : TAdvObject);
begin
  inherited;
  entity := TFhirContractTermActor(oSource).entity.Clone;
  if (TFhirContractTermActor(oSource).FRoleList = nil) then
  begin
    FRoleList.free;
    FRoleList := nil;
  end
  else
  begin
    FRoleList := TFhirCodeableConceptList.Create;
    FRoleList.Assign(TFhirContractTermActor(oSource).FRoleList);
  end;
end;

procedure TFhirContractTermActor.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'entity') Then
     list.add(FEntity.Link);
  if (child_name = 'role') Then
    list.addAll(FRoleList);
end;

procedure TFhirContractTermActor.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'entity', 'Reference(Contract|Device|Group|Location|Organization|Patient|Practitioner|RelatedPerson|Substance)', FEntity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRoleList.Link)){3};
end;

procedure TFhirContractTermActor.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'entity') then Entity := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'role') then RoleList.add(propValue as TFhirCodeableConcept){2}
  else inherited;
end;

function TFhirContractTermActor.FhirType : string;
begin
  result := 'actor';
end;

function TFhirContractTermActor.Link : TFhirContractTermActor;
begin
  result := TFhirContractTermActor(inherited Link);
end;

function TFhirContractTermActor.Clone : TFhirContractTermActor;
begin
  result := TFhirContractTermActor(inherited Clone);
end;

function TFhirContractTermActor.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTermActor;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractTermActor)) then
    result := false
  else
  begin
    o := TFhirContractTermActor(other);
    result := compareDeep(entityElement, o.entityElement, true) and compareDeep(roleList, o.roleList, true);
  end;
end;

function TFhirContractTermActor.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTermActor;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractTermActor)) then
    result := false
  else
  begin
    o := TFhirContractTermActor(other);
    result := true;
  end;
end;

{ TFhirContractTermActor }

Procedure TFhirContractTermActor.SetEntity(value : TFhirReference{Resource});
begin
  FEntity.free;
  FEntity := value;
end;

Function TFhirContractTermActor.GetRoleList : TFhirCodeableConceptList;
begin
  if FRoleList = nil then
    FRoleList := TFhirCodeableConceptList.Create;
  result := FRoleList;
end;

Function TFhirContractTermActor.GetHasRoleList : boolean;
begin
  result := (FRoleList <> nil) and (FRoleList.count > 0);
end;


{ TFhirContractTermActorListEnumerator }

Constructor TFhirContractTermActorListEnumerator.Create(list : TFhirContractTermActorList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractTermActorListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractTermActorListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractTermActorListEnumerator.GetCurrent : TFhirContractTermActor;
begin
  Result := FList[FIndex];
end;


{ TFhirContractTermActorList }
procedure TFhirContractTermActorList.AddItem(value: TFhirContractTermActor);
begin
  assert(value.ClassName = 'TFhirContractTermActor', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractTermActor');
  add(value);
end;


function TFhirContractTermActorList.Append: TFhirContractTermActor;
begin
  result := TFhirContractTermActor.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermActorList.ClearItems;
begin
  Clear;
end;

function TFhirContractTermActorList.GetEnumerator : TFhirContractTermActorListEnumerator;
begin
  result := TFhirContractTermActorListEnumerator.Create(self.link);
end;

function TFhirContractTermActorList.Clone: TFhirContractTermActorList;
begin
  result := TFhirContractTermActorList(inherited Clone);
end;

function TFhirContractTermActorList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractTermActorList.GetItemN(index: Integer): TFhirContractTermActor;
begin
  result := TFhirContractTermActor(ObjectByIndex[index]);
end;

function TFhirContractTermActorList.IndexOf(value: TFhirContractTermActor): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractTermActorList.Insert(index: Integer): TFhirContractTermActor;
begin
  result := TFhirContractTermActor.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermActorList.InsertItem(index: Integer; value: TFhirContractTermActor);
begin
  assert(value is TFhirContractTermActor);
  Inherited Insert(index, value);
end;

function TFhirContractTermActorList.Item(index: Integer): TFhirContractTermActor;
begin
  result := TFhirContractTermActor(ObjectByIndex[index]);
end;

function TFhirContractTermActorList.Link: TFhirContractTermActorList;
begin
  result := TFhirContractTermActorList(inherited Link);
end;

procedure TFhirContractTermActorList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractTermActorList.SetItemByIndex(index: Integer; value: TFhirContractTermActor);
begin
  assert(value is TFhirContractTermActor);
  FhirContractTermActors[index] := value;
end;

procedure TFhirContractTermActorList.SetItemN(index: Integer; value: TFhirContractTermActor);
begin
  assert(value is TFhirContractTermActor);
  ObjectByIndex[index] := value;
end;

{ TFhirContractTermValuedItem }

constructor TFhirContractTermValuedItem.Create;
begin
  inherited;
end;

destructor TFhirContractTermValuedItem.Destroy;
begin
  FEntity.free;
  FIdentifier.free;
  FEffectiveTime.free;
  FQuantity.free;
  FUnitPrice.free;
  FFactor.free;
  FPoints.free;
  FNet.free;
  inherited;
end;

procedure TFhirContractTermValuedItem.Assign(oSource : TAdvObject);
begin
  inherited;
  entity := TFhirContractTermValuedItem(oSource).entity.Clone;
  identifier := TFhirContractTermValuedItem(oSource).identifier.Clone;
  effectiveTimeElement := TFhirContractTermValuedItem(oSource).effectiveTimeElement.Clone;
  quantity := TFhirContractTermValuedItem(oSource).quantity.Clone;
  unitPrice := TFhirContractTermValuedItem(oSource).unitPrice.Clone;
  factorElement := TFhirContractTermValuedItem(oSource).factorElement.Clone;
  pointsElement := TFhirContractTermValuedItem(oSource).pointsElement.Clone;
  net := TFhirContractTermValuedItem(oSource).net.Clone;
end;

procedure TFhirContractTermValuedItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'entity[x]') Then
     list.add(FEntity.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'effectiveTime') Then
     list.add(FEffectiveTime.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'unitPrice') Then
     list.add(FUnitPrice.Link);
  if (child_name = 'factor') Then
     list.add(FFactor.Link);
  if (child_name = 'points') Then
     list.add(FPoints.Link);
  if (child_name = 'net') Then
     list.add(FNet.Link);
end;

procedure TFhirContractTermValuedItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'entity[x]', 'CodeableConcept|Reference(Any)', FEntity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'effectiveTime', 'dateTime', FEffectiveTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unitPrice', 'Money', FUnitPrice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'factor', 'decimal', FFactor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'points', 'decimal', FPoints.Link));{2}
  oList.add(TFHIRProperty.create(self, 'net', 'Money', FNet.Link));{2}
end;

procedure TFhirContractTermValuedItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('entity')) then Entity := propValue as TFhirType{4}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'effectiveTime') then EffectiveTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'unitPrice') then UnitPrice := propValue as TFhirQuantity{4b}
  else if (propName = 'factor') then FactorElement := propValue as TFhirDecimal{5a}
  else if (propName = 'points') then PointsElement := propValue as TFhirDecimal{5a}
  else if (propName = 'net') then Net := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirContractTermValuedItem.FhirType : string;
begin
  result := 'valuedItem';
end;

function TFhirContractTermValuedItem.Link : TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem(inherited Link);
end;

function TFhirContractTermValuedItem.Clone : TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem(inherited Clone);
end;

function TFhirContractTermValuedItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTermValuedItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractTermValuedItem)) then
    result := false
  else
  begin
    o := TFhirContractTermValuedItem(other);
    result := compareDeep(entityElement, o.entityElement, true) and compareDeep(identifierElement, o.identifierElement, true) and 
      compareDeep(effectiveTimeElement, o.effectiveTimeElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(unitPriceElement, o.unitPriceElement, true) and compareDeep(factorElement, o.factorElement, true) and 
      compareDeep(pointsElement, o.pointsElement, true) and compareDeep(netElement, o.netElement, true);
  end;
end;

function TFhirContractTermValuedItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractTermValuedItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractTermValuedItem)) then
    result := false
  else
  begin
    o := TFhirContractTermValuedItem(other);
    result := compareValues(effectiveTimeElement, o.effectiveTimeElement, true) and 
      compareValues(factorElement, o.factorElement, true) and compareValues(pointsElement, o.pointsElement, true);
  end;
end;

{ TFhirContractTermValuedItem }

Procedure TFhirContractTermValuedItem.SetEntity(value : TFhirType);
begin
  FEntity.free;
  FEntity := value;
end;

Procedure TFhirContractTermValuedItem.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirContractTermValuedItem.SetEffectiveTime(value : TFhirDateTime);
begin
  FEffectiveTime.free;
  FEffectiveTime := value;
end;

Function TFhirContractTermValuedItem.GetEffectiveTimeST : TDateAndTime;
begin
  if FEffectiveTime = nil then
    result := nil
  else
    result := FEffectiveTime.value;
end;

Procedure TFhirContractTermValuedItem.SetEffectiveTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEffectiveTime = nil then
      FEffectiveTime := TFhirDateTime.create;
    FEffectiveTime.value := value
  end
  else if FEffectiveTime <> nil then
    FEffectiveTime.value := nil;
end;

Procedure TFhirContractTermValuedItem.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirContractTermValuedItem.SetUnitPrice(value : TFhirQuantity);
begin
  FUnitPrice.free;
  FUnitPrice := value;
end;

Procedure TFhirContractTermValuedItem.SetFactor(value : TFhirDecimal);
begin
  FFactor.free;
  FFactor := value;
end;

Function TFhirContractTermValuedItem.GetFactorST : String;
begin
  if FFactor = nil then
    result := ''
  else
    result := FFactor.value;
end;

Procedure TFhirContractTermValuedItem.SetFactorST(value : String);
begin
  if value <> '' then
  begin
    if FFactor = nil then
      FFactor := TFhirDecimal.create;
    FFactor.value := value
  end
  else if FFactor <> nil then
    FFactor.value := '';
end;

Procedure TFhirContractTermValuedItem.SetPoints(value : TFhirDecimal);
begin
  FPoints.free;
  FPoints := value;
end;

Function TFhirContractTermValuedItem.GetPointsST : String;
begin
  if FPoints = nil then
    result := ''
  else
    result := FPoints.value;
end;

Procedure TFhirContractTermValuedItem.SetPointsST(value : String);
begin
  if value <> '' then
  begin
    if FPoints = nil then
      FPoints := TFhirDecimal.create;
    FPoints.value := value
  end
  else if FPoints <> nil then
    FPoints.value := '';
end;

Procedure TFhirContractTermValuedItem.SetNet(value : TFhirQuantity);
begin
  FNet.free;
  FNet := value;
end;


{ TFhirContractTermValuedItemListEnumerator }

Constructor TFhirContractTermValuedItemListEnumerator.Create(list : TFhirContractTermValuedItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractTermValuedItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractTermValuedItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractTermValuedItemListEnumerator.GetCurrent : TFhirContractTermValuedItem;
begin
  Result := FList[FIndex];
end;


{ TFhirContractTermValuedItemList }
procedure TFhirContractTermValuedItemList.AddItem(value: TFhirContractTermValuedItem);
begin
  assert(value.ClassName = 'TFhirContractTermValuedItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractTermValuedItem');
  add(value);
end;


function TFhirContractTermValuedItemList.Append: TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermValuedItemList.ClearItems;
begin
  Clear;
end;

function TFhirContractTermValuedItemList.GetEnumerator : TFhirContractTermValuedItemListEnumerator;
begin
  result := TFhirContractTermValuedItemListEnumerator.Create(self.link);
end;

function TFhirContractTermValuedItemList.Clone: TFhirContractTermValuedItemList;
begin
  result := TFhirContractTermValuedItemList(inherited Clone);
end;

function TFhirContractTermValuedItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractTermValuedItemList.GetItemN(index: Integer): TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem(ObjectByIndex[index]);
end;

function TFhirContractTermValuedItemList.IndexOf(value: TFhirContractTermValuedItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractTermValuedItemList.Insert(index: Integer): TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractTermValuedItemList.InsertItem(index: Integer; value: TFhirContractTermValuedItem);
begin
  assert(value is TFhirContractTermValuedItem);
  Inherited Insert(index, value);
end;

function TFhirContractTermValuedItemList.Item(index: Integer): TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem(ObjectByIndex[index]);
end;

function TFhirContractTermValuedItemList.Link: TFhirContractTermValuedItemList;
begin
  result := TFhirContractTermValuedItemList(inherited Link);
end;

procedure TFhirContractTermValuedItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractTermValuedItemList.SetItemByIndex(index: Integer; value: TFhirContractTermValuedItem);
begin
  assert(value is TFhirContractTermValuedItem);
  FhirContractTermValuedItems[index] := value;
end;

procedure TFhirContractTermValuedItemList.SetItemN(index: Integer; value: TFhirContractTermValuedItem);
begin
  assert(value is TFhirContractTermValuedItem);
  ObjectByIndex[index] := value;
end;

{ TFhirContractFriendly }

constructor TFhirContractFriendly.Create;
begin
  inherited;
end;

destructor TFhirContractFriendly.Destroy;
begin
  FContent.free;
  inherited;
end;

procedure TFhirContractFriendly.Assign(oSource : TAdvObject);
begin
  inherited;
  content := TFhirContractFriendly(oSource).content.Clone;
end;

procedure TFhirContractFriendly.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'content[x]') Then
     list.add(FContent.Link);
end;

procedure TFhirContractFriendly.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'content[x]', 'Attachment|Reference(Composition|DocumentReference|QuestionnaireResponse)', FContent.Link));{2}
end;

procedure TFhirContractFriendly.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('content')) then Content := propValue as TFhirType{4}
  else inherited;
end;

function TFhirContractFriendly.FhirType : string;
begin
  result := 'friendly';
end;

function TFhirContractFriendly.Link : TFhirContractFriendly;
begin
  result := TFhirContractFriendly(inherited Link);
end;

function TFhirContractFriendly.Clone : TFhirContractFriendly;
begin
  result := TFhirContractFriendly(inherited Clone);
end;

function TFhirContractFriendly.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractFriendly;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractFriendly)) then
    result := false
  else
  begin
    o := TFhirContractFriendly(other);
    result := compareDeep(contentElement, o.contentElement, true);
  end;
end;

function TFhirContractFriendly.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractFriendly;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractFriendly)) then
    result := false
  else
  begin
    o := TFhirContractFriendly(other);
    result := true;
  end;
end;

{ TFhirContractFriendly }

Procedure TFhirContractFriendly.SetContent(value : TFhirType);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirContractFriendlyListEnumerator }

Constructor TFhirContractFriendlyListEnumerator.Create(list : TFhirContractFriendlyList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractFriendlyListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractFriendlyListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractFriendlyListEnumerator.GetCurrent : TFhirContractFriendly;
begin
  Result := FList[FIndex];
end;


{ TFhirContractFriendlyList }
procedure TFhirContractFriendlyList.AddItem(value: TFhirContractFriendly);
begin
  assert(value.ClassName = 'TFhirContractFriendly', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractFriendly');
  add(value);
end;


function TFhirContractFriendlyList.Append: TFhirContractFriendly;
begin
  result := TFhirContractFriendly.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractFriendlyList.ClearItems;
begin
  Clear;
end;

function TFhirContractFriendlyList.GetEnumerator : TFhirContractFriendlyListEnumerator;
begin
  result := TFhirContractFriendlyListEnumerator.Create(self.link);
end;

function TFhirContractFriendlyList.Clone: TFhirContractFriendlyList;
begin
  result := TFhirContractFriendlyList(inherited Clone);
end;

function TFhirContractFriendlyList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractFriendlyList.GetItemN(index: Integer): TFhirContractFriendly;
begin
  result := TFhirContractFriendly(ObjectByIndex[index]);
end;

function TFhirContractFriendlyList.IndexOf(value: TFhirContractFriendly): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractFriendlyList.Insert(index: Integer): TFhirContractFriendly;
begin
  result := TFhirContractFriendly.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractFriendlyList.InsertItem(index: Integer; value: TFhirContractFriendly);
begin
  assert(value is TFhirContractFriendly);
  Inherited Insert(index, value);
end;

function TFhirContractFriendlyList.Item(index: Integer): TFhirContractFriendly;
begin
  result := TFhirContractFriendly(ObjectByIndex[index]);
end;

function TFhirContractFriendlyList.Link: TFhirContractFriendlyList;
begin
  result := TFhirContractFriendlyList(inherited Link);
end;

procedure TFhirContractFriendlyList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractFriendlyList.SetItemByIndex(index: Integer; value: TFhirContractFriendly);
begin
  assert(value is TFhirContractFriendly);
  FhirContractFriendlies[index] := value;
end;

procedure TFhirContractFriendlyList.SetItemN(index: Integer; value: TFhirContractFriendly);
begin
  assert(value is TFhirContractFriendly);
  ObjectByIndex[index] := value;
end;

{ TFhirContractLegal }

constructor TFhirContractLegal.Create;
begin
  inherited;
end;

destructor TFhirContractLegal.Destroy;
begin
  FContent.free;
  inherited;
end;

procedure TFhirContractLegal.Assign(oSource : TAdvObject);
begin
  inherited;
  content := TFhirContractLegal(oSource).content.Clone;
end;

procedure TFhirContractLegal.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'content[x]') Then
     list.add(FContent.Link);
end;

procedure TFhirContractLegal.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'content[x]', 'Attachment|Reference(Composition|DocumentReference|QuestionnaireResponse)', FContent.Link));{2}
end;

procedure TFhirContractLegal.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('content')) then Content := propValue as TFhirType{4}
  else inherited;
end;

function TFhirContractLegal.FhirType : string;
begin
  result := 'legal';
end;

function TFhirContractLegal.Link : TFhirContractLegal;
begin
  result := TFhirContractLegal(inherited Link);
end;

function TFhirContractLegal.Clone : TFhirContractLegal;
begin
  result := TFhirContractLegal(inherited Clone);
end;

function TFhirContractLegal.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractLegal;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractLegal)) then
    result := false
  else
  begin
    o := TFhirContractLegal(other);
    result := compareDeep(contentElement, o.contentElement, true);
  end;
end;

function TFhirContractLegal.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractLegal;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractLegal)) then
    result := false
  else
  begin
    o := TFhirContractLegal(other);
    result := true;
  end;
end;

{ TFhirContractLegal }

Procedure TFhirContractLegal.SetContent(value : TFhirType);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirContractLegalListEnumerator }

Constructor TFhirContractLegalListEnumerator.Create(list : TFhirContractLegalList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractLegalListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractLegalListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractLegalListEnumerator.GetCurrent : TFhirContractLegal;
begin
  Result := FList[FIndex];
end;


{ TFhirContractLegalList }
procedure TFhirContractLegalList.AddItem(value: TFhirContractLegal);
begin
  assert(value.ClassName = 'TFhirContractLegal', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractLegal');
  add(value);
end;


function TFhirContractLegalList.Append: TFhirContractLegal;
begin
  result := TFhirContractLegal.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractLegalList.ClearItems;
begin
  Clear;
end;

function TFhirContractLegalList.GetEnumerator : TFhirContractLegalListEnumerator;
begin
  result := TFhirContractLegalListEnumerator.Create(self.link);
end;

function TFhirContractLegalList.Clone: TFhirContractLegalList;
begin
  result := TFhirContractLegalList(inherited Clone);
end;

function TFhirContractLegalList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractLegalList.GetItemN(index: Integer): TFhirContractLegal;
begin
  result := TFhirContractLegal(ObjectByIndex[index]);
end;

function TFhirContractLegalList.IndexOf(value: TFhirContractLegal): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractLegalList.Insert(index: Integer): TFhirContractLegal;
begin
  result := TFhirContractLegal.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractLegalList.InsertItem(index: Integer; value: TFhirContractLegal);
begin
  assert(value is TFhirContractLegal);
  Inherited Insert(index, value);
end;

function TFhirContractLegalList.Item(index: Integer): TFhirContractLegal;
begin
  result := TFhirContractLegal(ObjectByIndex[index]);
end;

function TFhirContractLegalList.Link: TFhirContractLegalList;
begin
  result := TFhirContractLegalList(inherited Link);
end;

procedure TFhirContractLegalList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractLegalList.SetItemByIndex(index: Integer; value: TFhirContractLegal);
begin
  assert(value is TFhirContractLegal);
  FhirContractLegals[index] := value;
end;

procedure TFhirContractLegalList.SetItemN(index: Integer; value: TFhirContractLegal);
begin
  assert(value is TFhirContractLegal);
  ObjectByIndex[index] := value;
end;

{ TFhirContractRule }

constructor TFhirContractRule.Create;
begin
  inherited;
end;

destructor TFhirContractRule.Destroy;
begin
  FContent.free;
  inherited;
end;

procedure TFhirContractRule.Assign(oSource : TAdvObject);
begin
  inherited;
  content := TFhirContractRule(oSource).content.Clone;
end;

procedure TFhirContractRule.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'content[x]') Then
     list.add(FContent.Link);
end;

procedure TFhirContractRule.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'content[x]', 'Attachment|Reference(DocumentReference)', FContent.Link));{2}
end;

procedure TFhirContractRule.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('content')) then Content := propValue as TFhirType{4}
  else inherited;
end;

function TFhirContractRule.FhirType : string;
begin
  result := 'rule';
end;

function TFhirContractRule.Link : TFhirContractRule;
begin
  result := TFhirContractRule(inherited Link);
end;

function TFhirContractRule.Clone : TFhirContractRule;
begin
  result := TFhirContractRule(inherited Clone);
end;

function TFhirContractRule.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirContractRule;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirContractRule)) then
    result := false
  else
  begin
    o := TFhirContractRule(other);
    result := compareDeep(contentElement, o.contentElement, true);
  end;
end;

function TFhirContractRule.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirContractRule;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirContractRule)) then
    result := false
  else
  begin
    o := TFhirContractRule(other);
    result := true;
  end;
end;

{ TFhirContractRule }

Procedure TFhirContractRule.SetContent(value : TFhirType);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirContractRuleListEnumerator }

Constructor TFhirContractRuleListEnumerator.Create(list : TFhirContractRuleList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractRuleListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractRuleListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractRuleListEnumerator.GetCurrent : TFhirContractRule;
begin
  Result := FList[FIndex];
end;


{ TFhirContractRuleList }
procedure TFhirContractRuleList.AddItem(value: TFhirContractRule);
begin
  assert(value.ClassName = 'TFhirContractRule', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContractRule');
  add(value);
end;


function TFhirContractRuleList.Append: TFhirContractRule;
begin
  result := TFhirContractRule.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractRuleList.ClearItems;
begin
  Clear;
end;

function TFhirContractRuleList.GetEnumerator : TFhirContractRuleListEnumerator;
begin
  result := TFhirContractRuleListEnumerator.Create(self.link);
end;

function TFhirContractRuleList.Clone: TFhirContractRuleList;
begin
  result := TFhirContractRuleList(inherited Clone);
end;

function TFhirContractRuleList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractRuleList.GetItemN(index: Integer): TFhirContractRule;
begin
  result := TFhirContractRule(ObjectByIndex[index]);
end;

function TFhirContractRuleList.IndexOf(value: TFhirContractRule): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractRuleList.Insert(index: Integer): TFhirContractRule;
begin
  result := TFhirContractRule.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractRuleList.InsertItem(index: Integer; value: TFhirContractRule);
begin
  assert(value is TFhirContractRule);
  Inherited Insert(index, value);
end;

function TFhirContractRuleList.Item(index: Integer): TFhirContractRule;
begin
  result := TFhirContractRule(ObjectByIndex[index]);
end;

function TFhirContractRuleList.Link: TFhirContractRuleList;
begin
  result := TFhirContractRuleList(inherited Link);
end;

procedure TFhirContractRuleList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractRuleList.SetItemByIndex(index: Integer; value: TFhirContractRule);
begin
  assert(value is TFhirContractRule);
  FhirContractRules[index] := value;
end;

procedure TFhirContractRuleList.SetItemN(index: Integer; value: TFhirContractRule);
begin
  assert(value is TFhirContractRule);
  ObjectByIndex[index] := value;
end;

{ TFhirContract }

constructor TFhirContract.Create;
begin
  inherited;
end;

destructor TFhirContract.Destroy;
begin
  FIdentifier.free;
  FIssued.free;
  FApplies.free;
  FSubjectList.Free;
  FAuthorityList.Free;
  FDomainList.Free;
  FType_.free;
  FSubTypeList.Free;
  FActionList.Free;
  FActionReasonList.Free;
  FActorList.Free;
  FValuedItemList.Free;
  FSignerList.Free;
  FTermList.Free;
  FBinding.free;
  FFriendlyList.Free;
  FLegalList.Free;
  FRuleList.Free;
  inherited;
end;

function TFhirContract.GetResourceType : TFhirResourceType;
begin
  result := frtContract;
end;

procedure TFhirContract.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirContract(oSource).identifier.Clone;
  issuedElement := TFhirContract(oSource).issuedElement.Clone;
  applies := TFhirContract(oSource).applies.Clone;
  if (TFhirContract(oSource).FSubjectList = nil) then
  begin
    FSubjectList.free;
    FSubjectList := nil;
  end
  else
  begin
    FSubjectList := TFhirReferenceList{TFhirReference}.Create;
    FSubjectList.Assign(TFhirContract(oSource).FSubjectList);
  end;
  if (TFhirContract(oSource).FAuthorityList = nil) then
  begin
    FAuthorityList.free;
    FAuthorityList := nil;
  end
  else
  begin
    FAuthorityList := TFhirReferenceList{TFhirOrganization}.Create;
    FAuthorityList.Assign(TFhirContract(oSource).FAuthorityList);
  end;
  if (TFhirContract(oSource).FDomainList = nil) then
  begin
    FDomainList.free;
    FDomainList := nil;
  end
  else
  begin
    FDomainList := TFhirReferenceList{TFhirLocation}.Create;
    FDomainList.Assign(TFhirContract(oSource).FDomainList);
  end;
  type_ := TFhirContract(oSource).type_.Clone;
  if (TFhirContract(oSource).FSubTypeList = nil) then
  begin
    FSubTypeList.free;
    FSubTypeList := nil;
  end
  else
  begin
    FSubTypeList := TFhirCodeableConceptList.Create;
    FSubTypeList.Assign(TFhirContract(oSource).FSubTypeList);
  end;
  if (TFhirContract(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirCodeableConceptList.Create;
    FActionList.Assign(TFhirContract(oSource).FActionList);
  end;
  if (TFhirContract(oSource).FActionReasonList = nil) then
  begin
    FActionReasonList.free;
    FActionReasonList := nil;
  end
  else
  begin
    FActionReasonList := TFhirCodeableConceptList.Create;
    FActionReasonList.Assign(TFhirContract(oSource).FActionReasonList);
  end;
  if (TFhirContract(oSource).FActorList = nil) then
  begin
    FActorList.free;
    FActorList := nil;
  end
  else
  begin
    FActorList := TFhirContractActorList.Create;
    FActorList.Assign(TFhirContract(oSource).FActorList);
  end;
  if (TFhirContract(oSource).FValuedItemList = nil) then
  begin
    FValuedItemList.free;
    FValuedItemList := nil;
  end
  else
  begin
    FValuedItemList := TFhirContractValuedItemList.Create;
    FValuedItemList.Assign(TFhirContract(oSource).FValuedItemList);
  end;
  if (TFhirContract(oSource).FSignerList = nil) then
  begin
    FSignerList.free;
    FSignerList := nil;
  end
  else
  begin
    FSignerList := TFhirContractSignerList.Create;
    FSignerList.Assign(TFhirContract(oSource).FSignerList);
  end;
  if (TFhirContract(oSource).FTermList = nil) then
  begin
    FTermList.free;
    FTermList := nil;
  end
  else
  begin
    FTermList := TFhirContractTermList.Create;
    FTermList.Assign(TFhirContract(oSource).FTermList);
  end;
  binding := TFhirContract(oSource).binding.Clone;
  if (TFhirContract(oSource).FFriendlyList = nil) then
  begin
    FFriendlyList.free;
    FFriendlyList := nil;
  end
  else
  begin
    FFriendlyList := TFhirContractFriendlyList.Create;
    FFriendlyList.Assign(TFhirContract(oSource).FFriendlyList);
  end;
  if (TFhirContract(oSource).FLegalList = nil) then
  begin
    FLegalList.free;
    FLegalList := nil;
  end
  else
  begin
    FLegalList := TFhirContractLegalList.Create;
    FLegalList.Assign(TFhirContract(oSource).FLegalList);
  end;
  if (TFhirContract(oSource).FRuleList = nil) then
  begin
    FRuleList.free;
    FRuleList := nil;
  end
  else
  begin
    FRuleList := TFhirContractRuleList.Create;
    FRuleList.Assign(TFhirContract(oSource).FRuleList);
  end;
end;

procedure TFhirContract.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'issued') Then
     list.add(FIssued.Link);
  if (child_name = 'applies') Then
     list.add(FApplies.Link);
  if (child_name = 'subject') Then
    list.addAll(FSubjectList);
  if (child_name = 'authority') Then
    list.addAll(FAuthorityList);
  if (child_name = 'domain') Then
    list.addAll(FDomainList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'subType') Then
    list.addAll(FSubTypeList);
  if (child_name = 'action') Then
    list.addAll(FActionList);
  if (child_name = 'actionReason') Then
    list.addAll(FActionReasonList);
  if (child_name = 'actor') Then
    list.addAll(FActorList);
  if (child_name = 'valuedItem') Then
    list.addAll(FValuedItemList);
  if (child_name = 'signer') Then
    list.addAll(FSignerList);
  if (child_name = 'term') Then
    list.addAll(FTermList);
  if (child_name = 'binding[x]') Then
     list.add(FBinding.Link);
  if (child_name = 'friendly') Then
    list.addAll(FFriendlyList);
  if (child_name = 'legal') Then
    list.addAll(FLegalList);
  if (child_name = 'rule') Then
    list.addAll(FRuleList);
end;

procedure TFhirContract.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'issued', 'dateTime', FIssued.Link));{2}
  oList.add(TFHIRProperty.create(self, 'applies', 'Period', FApplies.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubjectList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'authority', 'Reference(Organization)', FAuthorityList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'domain', 'Reference(Location)', FDomainList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subType', 'CodeableConcept', FSubTypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'action', 'CodeableConcept', FActionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actionReason', 'CodeableConcept', FActionReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actor', '', FActorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'valuedItem', '', FValuedItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'signer', '', FSignerList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'term', '', FTermList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'binding[x]', 'Attachment|Reference(Composition|DocumentReference|QuestionnaireResponse)', FBinding.Link));{2}
  oList.add(TFHIRProperty.create(self, 'friendly', '', FFriendlyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'legal', '', FLegalList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'rule', '', FRuleList.Link)){3};
end;

procedure TFhirContract.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'issued') then IssuedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'applies') then Applies := propValue as TFhirPeriod{4b}
  else if (propName = 'subject') then SubjectList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'authority') then AuthorityList.add(propValue as TFhirReference{TFhirOrganization}){2}
  else if (propName = 'domain') then DomainList.add(propValue as TFhirReference{TFhirLocation}){2}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subType') then SubTypeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'action') then ActionList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actionReason') then ActionReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actor') then ActorList.add(propValue as TFhirContractActor){2}
  else if (propName = 'valuedItem') then ValuedItemList.add(propValue as TFhirContractValuedItem){2}
  else if (propName = 'signer') then SignerList.add(propValue as TFhirContractSigner){2}
  else if (propName = 'term') then TermList.add(propValue as TFhirContractTerm){2}
  else if (propName.startsWith('binding')) then Binding := propValue as TFhirType{4}
  else if (propName = 'friendly') then FriendlyList.add(propValue as TFhirContractFriendly){2}
  else if (propName = 'legal') then LegalList.add(propValue as TFhirContractLegal){2}
  else if (propName = 'rule') then RuleList.add(propValue as TFhirContractRule){2}
  else inherited;
end;

function TFhirContract.FhirType : string;
begin
  result := 'Contract';
end;

function TFhirContract.Link : TFhirContract;
begin
  result := TFhirContract(inherited Link);
end;

function TFhirContract.Clone : TFhirContract;
begin
  result := TFhirContract(inherited Clone);
end;

{ TFhirContract }

Procedure TFhirContract.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirContract.SetIssued(value : TFhirDateTime);
begin
  FIssued.free;
  FIssued := value;
end;

Function TFhirContract.GetIssuedST : TDateAndTime;
begin
  if FIssued = nil then
    result := nil
  else
    result := FIssued.value;
end;

Procedure TFhirContract.SetIssuedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIssued = nil then
      FIssued := TFhirDateTime.create;
    FIssued.value := value
  end
  else if FIssued <> nil then
    FIssued.value := nil;
end;

Procedure TFhirContract.SetApplies(value : TFhirPeriod);
begin
  FApplies.free;
  FApplies := value;
end;

Function TFhirContract.GetSubjectList : TFhirReferenceList{TFhirReference};
begin
  if FSubjectList = nil then
    FSubjectList := TFhirReferenceList{TFhirReference}.Create;
  result := FSubjectList;
end;

Function TFhirContract.GetHasSubjectList : boolean;
begin
  result := (FSubjectList <> nil) and (FSubjectList.count > 0);
end;

Function TFhirContract.GetAuthorityList : TFhirReferenceList{TFhirOrganization};
begin
  if FAuthorityList = nil then
    FAuthorityList := TFhirReferenceList{TFhirOrganization}.Create;
  result := FAuthorityList;
end;

Function TFhirContract.GetHasAuthorityList : boolean;
begin
  result := (FAuthorityList <> nil) and (FAuthorityList.count > 0);
end;

Function TFhirContract.GetDomainList : TFhirReferenceList{TFhirLocation};
begin
  if FDomainList = nil then
    FDomainList := TFhirReferenceList{TFhirLocation}.Create;
  result := FDomainList;
end;

Function TFhirContract.GetHasDomainList : boolean;
begin
  result := (FDomainList <> nil) and (FDomainList.count > 0);
end;

Procedure TFhirContract.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirContract.GetSubTypeList : TFhirCodeableConceptList;
begin
  if FSubTypeList = nil then
    FSubTypeList := TFhirCodeableConceptList.Create;
  result := FSubTypeList;
end;

Function TFhirContract.GetHasSubTypeList : boolean;
begin
  result := (FSubTypeList <> nil) and (FSubTypeList.count > 0);
end;

Function TFhirContract.GetActionList : TFhirCodeableConceptList;
begin
  if FActionList = nil then
    FActionList := TFhirCodeableConceptList.Create;
  result := FActionList;
end;

Function TFhirContract.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;

Function TFhirContract.GetActionReasonList : TFhirCodeableConceptList;
begin
  if FActionReasonList = nil then
    FActionReasonList := TFhirCodeableConceptList.Create;
  result := FActionReasonList;
end;

Function TFhirContract.GetHasActionReasonList : boolean;
begin
  result := (FActionReasonList <> nil) and (FActionReasonList.count > 0);
end;

Function TFhirContract.GetActorList : TFhirContractActorList;
begin
  if FActorList = nil then
    FActorList := TFhirContractActorList.Create;
  result := FActorList;
end;

Function TFhirContract.GetHasActorList : boolean;
begin
  result := (FActorList <> nil) and (FActorList.count > 0);
end;

Function TFhirContract.GetValuedItemList : TFhirContractValuedItemList;
begin
  if FValuedItemList = nil then
    FValuedItemList := TFhirContractValuedItemList.Create;
  result := FValuedItemList;
end;

Function TFhirContract.GetHasValuedItemList : boolean;
begin
  result := (FValuedItemList <> nil) and (FValuedItemList.count > 0);
end;

Function TFhirContract.GetSignerList : TFhirContractSignerList;
begin
  if FSignerList = nil then
    FSignerList := TFhirContractSignerList.Create;
  result := FSignerList;
end;

Function TFhirContract.GetHasSignerList : boolean;
begin
  result := (FSignerList <> nil) and (FSignerList.count > 0);
end;

Function TFhirContract.GetTermList : TFhirContractTermList;
begin
  if FTermList = nil then
    FTermList := TFhirContractTermList.Create;
  result := FTermList;
end;

Function TFhirContract.GetHasTermList : boolean;
begin
  result := (FTermList <> nil) and (FTermList.count > 0);
end;

Procedure TFhirContract.SetBinding(value : TFhirType);
begin
  FBinding.free;
  FBinding := value;
end;

Function TFhirContract.GetFriendlyList : TFhirContractFriendlyList;
begin
  if FFriendlyList = nil then
    FFriendlyList := TFhirContractFriendlyList.Create;
  result := FFriendlyList;
end;

Function TFhirContract.GetHasFriendlyList : boolean;
begin
  result := (FFriendlyList <> nil) and (FFriendlyList.count > 0);
end;

Function TFhirContract.GetLegalList : TFhirContractLegalList;
begin
  if FLegalList = nil then
    FLegalList := TFhirContractLegalList.Create;
  result := FLegalList;
end;

Function TFhirContract.GetHasLegalList : boolean;
begin
  result := (FLegalList <> nil) and (FLegalList.count > 0);
end;

Function TFhirContract.GetRuleList : TFhirContractRuleList;
begin
  if FRuleList = nil then
    FRuleList := TFhirContractRuleList.Create;
  result := FRuleList;
end;

Function TFhirContract.GetHasRuleList : boolean;
begin
  result := (FRuleList <> nil) and (FRuleList.count > 0);
end;


{ TFhirContractListEnumerator }

Constructor TFhirContractListEnumerator.Create(list : TFhirContractList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirContractListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirContractListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirContractListEnumerator.GetCurrent : TFhirContract;
begin
  Result := FList[FIndex];
end;


{ TFhirContractList }
procedure TFhirContractList.AddItem(value: TFhirContract);
begin
  assert(value.ClassName = 'TFhirContract', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirContract');
  add(value);
end;


function TFhirContractList.Append: TFhirContract;
begin
  result := TFhirContract.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractList.ClearItems;
begin
  Clear;
end;

function TFhirContractList.GetEnumerator : TFhirContractListEnumerator;
begin
  result := TFhirContractListEnumerator.Create(self.link);
end;

function TFhirContractList.Clone: TFhirContractList;
begin
  result := TFhirContractList(inherited Clone);
end;

function TFhirContractList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirContractList.GetItemN(index: Integer): TFhirContract;
begin
  result := TFhirContract(ObjectByIndex[index]);
end;

function TFhirContractList.IndexOf(value: TFhirContract): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirContractList.Insert(index: Integer): TFhirContract;
begin
  result := TFhirContract.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirContractList.InsertItem(index: Integer; value: TFhirContract);
begin
  assert(value is TFhirContract);
  Inherited Insert(index, value);
end;

function TFhirContractList.Item(index: Integer): TFhirContract;
begin
  result := TFhirContract(ObjectByIndex[index]);
end;

function TFhirContractList.Link: TFhirContractList;
begin
  result := TFhirContractList(inherited Link);
end;

procedure TFhirContractList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirContractList.SetItemByIndex(index: Integer; value: TFhirContract);
begin
  assert(value is TFhirContract);
  FhirContracts[index] := value;
end;

procedure TFhirContractList.SetItemN(index: Integer; value: TFhirContract);
begin
  assert(value is TFhirContract);
  ObjectByIndex[index] := value;
end;

{ TFhirCoverage }

constructor TFhirCoverage.Create;
begin
  inherited;
end;

destructor TFhirCoverage.Destroy;
begin
  FIssuer.free;
  FBin.free;
  FPeriod.free;
  FType_.free;
  FSubscriberId.free;
  FIdentifierList.Free;
  FGroup.free;
  FPlan.free;
  FSubPlan.free;
  FDependent.free;
  FSequence.free;
  FSubscriber.free;
  FNetwork.free;
  FContractList.Free;
  inherited;
end;

function TFhirCoverage.GetResourceType : TFhirResourceType;
begin
  result := frtCoverage;
end;

procedure TFhirCoverage.Assign(oSource : TAdvObject);
begin
  inherited;
  issuer := TFhirCoverage(oSource).issuer.Clone;
  bin := TFhirCoverage(oSource).bin.Clone;
  period := TFhirCoverage(oSource).period.Clone;
  type_ := TFhirCoverage(oSource).type_.Clone;
  subscriberId := TFhirCoverage(oSource).subscriberId.Clone;
  if (TFhirCoverage(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirCoverage(oSource).FIdentifierList);
  end;
  groupElement := TFhirCoverage(oSource).groupElement.Clone;
  planElement := TFhirCoverage(oSource).planElement.Clone;
  subPlanElement := TFhirCoverage(oSource).subPlanElement.Clone;
  dependentElement := TFhirCoverage(oSource).dependentElement.Clone;
  sequenceElement := TFhirCoverage(oSource).sequenceElement.Clone;
  subscriber := TFhirCoverage(oSource).subscriber.Clone;
  network := TFhirCoverage(oSource).network.Clone;
  if (TFhirCoverage(oSource).FContractList = nil) then
  begin
    FContractList.free;
    FContractList := nil;
  end
  else
  begin
    FContractList := TFhirReferenceList{TFhirContract}.Create;
    FContractList.Assign(TFhirCoverage(oSource).FContractList);
  end;
end;

procedure TFhirCoverage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'issuer') Then
     list.add(FIssuer.Link);
  if (child_name = 'bin') Then
     list.add(FBin.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'subscriberId') Then
     list.add(FSubscriberId.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'group') Then
     list.add(FGroup.Link);
  if (child_name = 'plan') Then
     list.add(FPlan.Link);
  if (child_name = 'subPlan') Then
     list.add(FSubPlan.Link);
  if (child_name = 'dependent') Then
     list.add(FDependent.Link);
  if (child_name = 'sequence') Then
     list.add(FSequence.Link);
  if (child_name = 'subscriber') Then
     list.add(FSubscriber.Link);
  if (child_name = 'network') Then
     list.add(FNetwork.Link);
  if (child_name = 'contract') Then
    list.addAll(FContractList);
end;

procedure TFhirCoverage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'issuer', 'Reference(Organization)', FIssuer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bin', 'Identifier', FBin.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subscriberId', 'Identifier', FSubscriberId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'group', 'string', FGroup.Link));{2}
  oList.add(TFHIRProperty.create(self, 'plan', 'string', FPlan.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subPlan', 'string', FSubPlan.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dependent', 'positiveInt', FDependent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sequence', 'positiveInt', FSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subscriber', 'Reference(Patient)', FSubscriber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'network', 'Identifier', FNetwork.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contract', 'Reference(Contract)', FContractList.Link)){3};
end;

procedure TFhirCoverage.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'issuer') then Issuer := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'bin') then Bin := propValue as TFhirIdentifier{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'subscriberId') then SubscriberId := propValue as TFhirIdentifier{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'group') then GroupElement := propValue as TFhirString{5a}
  else if (propName = 'plan') then PlanElement := propValue as TFhirString{5a}
  else if (propName = 'subPlan') then SubPlanElement := propValue as TFhirString{5a}
  else if (propName = 'dependent') then DependentElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'sequence') then SequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'subscriber') then Subscriber := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'network') then Network := propValue as TFhirIdentifier{4b}
  else if (propName = 'contract') then ContractList.add(propValue as TFhirReference{TFhirContract}){2}
  else inherited;
end;

function TFhirCoverage.FhirType : string;
begin
  result := 'Coverage';
end;

function TFhirCoverage.Link : TFhirCoverage;
begin
  result := TFhirCoverage(inherited Link);
end;

function TFhirCoverage.Clone : TFhirCoverage;
begin
  result := TFhirCoverage(inherited Clone);
end;

{ TFhirCoverage }

Procedure TFhirCoverage.SetIssuer(value : TFhirReference{TFhirOrganization});
begin
  FIssuer.free;
  FIssuer := value;
end;

Procedure TFhirCoverage.SetBin(value : TFhirIdentifier);
begin
  FBin.free;
  FBin := value;
end;

Procedure TFhirCoverage.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirCoverage.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirCoverage.SetSubscriberId(value : TFhirIdentifier);
begin
  FSubscriberId.free;
  FSubscriberId := value;
end;

Function TFhirCoverage.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirCoverage.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirCoverage.SetGroup(value : TFhirString);
begin
  FGroup.free;
  FGroup := value;
end;

Function TFhirCoverage.GetGroupST : String;
begin
  if FGroup = nil then
    result := ''
  else
    result := FGroup.value;
end;

Procedure TFhirCoverage.SetGroupST(value : String);
begin
  if value <> '' then
  begin
    if FGroup = nil then
      FGroup := TFhirString.create;
    FGroup.value := value
  end
  else if FGroup <> nil then
    FGroup.value := '';
end;

Procedure TFhirCoverage.SetPlan(value : TFhirString);
begin
  FPlan.free;
  FPlan := value;
end;

Function TFhirCoverage.GetPlanST : String;
begin
  if FPlan = nil then
    result := ''
  else
    result := FPlan.value;
end;

Procedure TFhirCoverage.SetPlanST(value : String);
begin
  if value <> '' then
  begin
    if FPlan = nil then
      FPlan := TFhirString.create;
    FPlan.value := value
  end
  else if FPlan <> nil then
    FPlan.value := '';
end;

Procedure TFhirCoverage.SetSubPlan(value : TFhirString);
begin
  FSubPlan.free;
  FSubPlan := value;
end;

Function TFhirCoverage.GetSubPlanST : String;
begin
  if FSubPlan = nil then
    result := ''
  else
    result := FSubPlan.value;
end;

Procedure TFhirCoverage.SetSubPlanST(value : String);
begin
  if value <> '' then
  begin
    if FSubPlan = nil then
      FSubPlan := TFhirString.create;
    FSubPlan.value := value
  end
  else if FSubPlan <> nil then
    FSubPlan.value := '';
end;

Procedure TFhirCoverage.SetDependent(value : TFhirPositiveInt);
begin
  FDependent.free;
  FDependent := value;
end;

Function TFhirCoverage.GetDependentST : String;
begin
  if FDependent = nil then
    result := ''
  else
    result := FDependent.value;
end;

Procedure TFhirCoverage.SetDependentST(value : String);
begin
  if value <> '' then
  begin
    if FDependent = nil then
      FDependent := TFhirPositiveInt.create;
    FDependent.value := value
  end
  else if FDependent <> nil then
    FDependent.value := '';
end;

Procedure TFhirCoverage.SetSequence(value : TFhirPositiveInt);
begin
  FSequence.free;
  FSequence := value;
end;

Function TFhirCoverage.GetSequenceST : String;
begin
  if FSequence = nil then
    result := ''
  else
    result := FSequence.value;
end;

Procedure TFhirCoverage.SetSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FSequence = nil then
      FSequence := TFhirPositiveInt.create;
    FSequence.value := value
  end
  else if FSequence <> nil then
    FSequence.value := '';
end;

Procedure TFhirCoverage.SetSubscriber(value : TFhirReference{TFhirPatient});
begin
  FSubscriber.free;
  FSubscriber := value;
end;

Procedure TFhirCoverage.SetNetwork(value : TFhirIdentifier);
begin
  FNetwork.free;
  FNetwork := value;
end;

Function TFhirCoverage.GetContractList : TFhirReferenceList{TFhirContract};
begin
  if FContractList = nil then
    FContractList := TFhirReferenceList{TFhirContract}.Create;
  result := FContractList;
end;

Function TFhirCoverage.GetHasContractList : boolean;
begin
  result := (FContractList <> nil) and (FContractList.count > 0);
end;


{ TFhirCoverageListEnumerator }

Constructor TFhirCoverageListEnumerator.Create(list : TFhirCoverageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirCoverageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirCoverageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirCoverageListEnumerator.GetCurrent : TFhirCoverage;
begin
  Result := FList[FIndex];
end;


{ TFhirCoverageList }
procedure TFhirCoverageList.AddItem(value: TFhirCoverage);
begin
  assert(value.ClassName = 'TFhirCoverage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirCoverage');
  add(value);
end;


function TFhirCoverageList.Append: TFhirCoverage;
begin
  result := TFhirCoverage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCoverageList.ClearItems;
begin
  Clear;
end;

function TFhirCoverageList.GetEnumerator : TFhirCoverageListEnumerator;
begin
  result := TFhirCoverageListEnumerator.Create(self.link);
end;

function TFhirCoverageList.Clone: TFhirCoverageList;
begin
  result := TFhirCoverageList(inherited Clone);
end;

function TFhirCoverageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirCoverageList.GetItemN(index: Integer): TFhirCoverage;
begin
  result := TFhirCoverage(ObjectByIndex[index]);
end;

function TFhirCoverageList.IndexOf(value: TFhirCoverage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirCoverageList.Insert(index: Integer): TFhirCoverage;
begin
  result := TFhirCoverage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirCoverageList.InsertItem(index: Integer; value: TFhirCoverage);
begin
  assert(value is TFhirCoverage);
  Inherited Insert(index, value);
end;

function TFhirCoverageList.Item(index: Integer): TFhirCoverage;
begin
  result := TFhirCoverage(ObjectByIndex[index]);
end;

function TFhirCoverageList.Link: TFhirCoverageList;
begin
  result := TFhirCoverageList(inherited Link);
end;

procedure TFhirCoverageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirCoverageList.SetItemByIndex(index: Integer; value: TFhirCoverage);
begin
  assert(value is TFhirCoverage);
  FhirCoverages[index] := value;
end;

procedure TFhirCoverageList.SetItemN(index: Integer; value: TFhirCoverage);
begin
  assert(value is TFhirCoverage);
  ObjectByIndex[index] := value;
end;

{ TFhirDataElementContact }

constructor TFhirDataElementContact.Create;
begin
  inherited;
end;

destructor TFhirDataElementContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirDataElementContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirDataElementContact(oSource).nameElement.Clone;
  if (TFhirDataElementContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirDataElementContact(oSource).FTelecomList);
  end;
end;

procedure TFhirDataElementContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirDataElementContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirDataElementContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirDataElementContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirDataElementContact.Link : TFhirDataElementContact;
begin
  result := TFhirDataElementContact(inherited Link);
end;

function TFhirDataElementContact.Clone : TFhirDataElementContact;
begin
  result := TFhirDataElementContact(inherited Clone);
end;

function TFhirDataElementContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDataElementContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDataElementContact)) then
    result := false
  else
  begin
    o := TFhirDataElementContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirDataElementContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDataElementContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDataElementContact)) then
    result := false
  else
  begin
    o := TFhirDataElementContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirDataElementContact }

Procedure TFhirDataElementContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirDataElementContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirDataElementContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirDataElementContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirDataElementContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirDataElementContactListEnumerator }

Constructor TFhirDataElementContactListEnumerator.Create(list : TFhirDataElementContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDataElementContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDataElementContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDataElementContactListEnumerator.GetCurrent : TFhirDataElementContact;
begin
  Result := FList[FIndex];
end;


{ TFhirDataElementContactList }
procedure TFhirDataElementContactList.AddItem(value: TFhirDataElementContact);
begin
  assert(value.ClassName = 'TFhirDataElementContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDataElementContact');
  add(value);
end;


function TFhirDataElementContactList.Append: TFhirDataElementContact;
begin
  result := TFhirDataElementContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementContactList.ClearItems;
begin
  Clear;
end;

function TFhirDataElementContactList.GetEnumerator : TFhirDataElementContactListEnumerator;
begin
  result := TFhirDataElementContactListEnumerator.Create(self.link);
end;

function TFhirDataElementContactList.Clone: TFhirDataElementContactList;
begin
  result := TFhirDataElementContactList(inherited Clone);
end;

function TFhirDataElementContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDataElementContactList.GetItemN(index: Integer): TFhirDataElementContact;
begin
  result := TFhirDataElementContact(ObjectByIndex[index]);
end;

function TFhirDataElementContactList.IndexOf(value: TFhirDataElementContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDataElementContactList.Insert(index: Integer): TFhirDataElementContact;
begin
  result := TFhirDataElementContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementContactList.InsertItem(index: Integer; value: TFhirDataElementContact);
begin
  assert(value is TFhirDataElementContact);
  Inherited Insert(index, value);
end;

function TFhirDataElementContactList.Item(index: Integer): TFhirDataElementContact;
begin
  result := TFhirDataElementContact(ObjectByIndex[index]);
end;

function TFhirDataElementContactList.Link: TFhirDataElementContactList;
begin
  result := TFhirDataElementContactList(inherited Link);
end;

procedure TFhirDataElementContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDataElementContactList.SetItemByIndex(index: Integer; value: TFhirDataElementContact);
begin
  assert(value is TFhirDataElementContact);
  FhirDataElementContacts[index] := value;
end;

procedure TFhirDataElementContactList.SetItemN(index: Integer; value: TFhirDataElementContact);
begin
  assert(value is TFhirDataElementContact);
  ObjectByIndex[index] := value;
end;

{ TFhirDataElementMapping }

constructor TFhirDataElementMapping.Create;
begin
  inherited;
end;

destructor TFhirDataElementMapping.Destroy;
begin
  FIdentity.free;
  FUri.free;
  FName.free;
  FComments.free;
  inherited;
end;

procedure TFhirDataElementMapping.Assign(oSource : TAdvObject);
begin
  inherited;
  identityElement := TFhirDataElementMapping(oSource).identityElement.Clone;
  uriElement := TFhirDataElementMapping(oSource).uriElement.Clone;
  nameElement := TFhirDataElementMapping(oSource).nameElement.Clone;
  commentsElement := TFhirDataElementMapping(oSource).commentsElement.Clone;
end;

procedure TFhirDataElementMapping.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identity') Then
     list.add(FIdentity.Link);
  if (child_name = 'uri') Then
     list.add(FUri.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'comments') Then
     list.add(FComments.Link);
end;

procedure TFhirDataElementMapping.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identity', 'id', FIdentity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uri', 'uri', FUri.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comments', 'string', FComments.Link));{2}
end;

procedure TFhirDataElementMapping.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identity') then IdentityElement := propValue as TFhirId{5a}
  else if (propName = 'uri') then UriElement := propValue as TFhirUri{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'comments') then CommentsElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirDataElementMapping.FhirType : string;
begin
  result := 'mapping';
end;

function TFhirDataElementMapping.Link : TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping(inherited Link);
end;

function TFhirDataElementMapping.Clone : TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping(inherited Clone);
end;

function TFhirDataElementMapping.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDataElementMapping;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDataElementMapping)) then
    result := false
  else
  begin
    o := TFhirDataElementMapping(other);
    result := compareDeep(identityElement, o.identityElement, true) and compareDeep(uriElement, o.uriElement, true) and 
      compareDeep(nameElement, o.nameElement, true) and compareDeep(commentsElement, o.commentsElement, true);
  end;
end;

function TFhirDataElementMapping.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDataElementMapping;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDataElementMapping)) then
    result := false
  else
  begin
    o := TFhirDataElementMapping(other);
    result := compareValues(identityElement, o.identityElement, true) and compareValues(uriElement, o.uriElement, true) and 
      compareValues(nameElement, o.nameElement, true) and compareValues(commentsElement, o.commentsElement, true);
  end;
end;

{ TFhirDataElementMapping }

Procedure TFhirDataElementMapping.SetIdentity(value : TFhirId);
begin
  FIdentity.free;
  FIdentity := value;
end;

Function TFhirDataElementMapping.GetIdentityST : String;
begin
  if FIdentity = nil then
    result := ''
  else
    result := FIdentity.value;
end;

Procedure TFhirDataElementMapping.SetIdentityST(value : String);
begin
  if value <> '' then
  begin
    if FIdentity = nil then
      FIdentity := TFhirId.create;
    FIdentity.value := value
  end
  else if FIdentity <> nil then
    FIdentity.value := '';
end;

Procedure TFhirDataElementMapping.SetUri(value : TFhirUri);
begin
  FUri.free;
  FUri := value;
end;

Function TFhirDataElementMapping.GetUriST : String;
begin
  if FUri = nil then
    result := ''
  else
    result := FUri.value;
end;

Procedure TFhirDataElementMapping.SetUriST(value : String);
begin
  if value <> '' then
  begin
    if FUri = nil then
      FUri := TFhirUri.create;
    FUri.value := value
  end
  else if FUri <> nil then
    FUri.value := '';
end;

Procedure TFhirDataElementMapping.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirDataElementMapping.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirDataElementMapping.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirDataElementMapping.SetComments(value : TFhirString);
begin
  FComments.free;
  FComments := value;
end;

Function TFhirDataElementMapping.GetCommentsST : String;
begin
  if FComments = nil then
    result := ''
  else
    result := FComments.value;
end;

Procedure TFhirDataElementMapping.SetCommentsST(value : String);
begin
  if value <> '' then
  begin
    if FComments = nil then
      FComments := TFhirString.create;
    FComments.value := value
  end
  else if FComments <> nil then
    FComments.value := '';
end;


{ TFhirDataElementMappingListEnumerator }

Constructor TFhirDataElementMappingListEnumerator.Create(list : TFhirDataElementMappingList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDataElementMappingListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDataElementMappingListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDataElementMappingListEnumerator.GetCurrent : TFhirDataElementMapping;
begin
  Result := FList[FIndex];
end;


{ TFhirDataElementMappingList }
procedure TFhirDataElementMappingList.AddItem(value: TFhirDataElementMapping);
begin
  assert(value.ClassName = 'TFhirDataElementMapping', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDataElementMapping');
  add(value);
end;


function TFhirDataElementMappingList.Append: TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementMappingList.ClearItems;
begin
  Clear;
end;

function TFhirDataElementMappingList.GetEnumerator : TFhirDataElementMappingListEnumerator;
begin
  result := TFhirDataElementMappingListEnumerator.Create(self.link);
end;

function TFhirDataElementMappingList.Clone: TFhirDataElementMappingList;
begin
  result := TFhirDataElementMappingList(inherited Clone);
end;

function TFhirDataElementMappingList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDataElementMappingList.GetItemN(index: Integer): TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping(ObjectByIndex[index]);
end;

function TFhirDataElementMappingList.IndexOf(value: TFhirDataElementMapping): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDataElementMappingList.Insert(index: Integer): TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementMappingList.InsertItem(index: Integer; value: TFhirDataElementMapping);
begin
  assert(value is TFhirDataElementMapping);
  Inherited Insert(index, value);
end;

function TFhirDataElementMappingList.Item(index: Integer): TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping(ObjectByIndex[index]);
end;

function TFhirDataElementMappingList.Link: TFhirDataElementMappingList;
begin
  result := TFhirDataElementMappingList(inherited Link);
end;

procedure TFhirDataElementMappingList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDataElementMappingList.SetItemByIndex(index: Integer; value: TFhirDataElementMapping);
begin
  assert(value is TFhirDataElementMapping);
  FhirDataElementMappings[index] := value;
end;

procedure TFhirDataElementMappingList.SetItemN(index: Integer; value: TFhirDataElementMapping);
begin
  assert(value is TFhirDataElementMapping);
  ObjectByIndex[index] := value;
end;

{ TFhirDataElement }

constructor TFhirDataElement.Create;
begin
  inherited;
end;

destructor TFhirDataElement.Destroy;
begin
  FUrl.free;
  FIdentifierList.Free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FUseContextList.Free;
  FCopyright.free;
  FStringency.free;
  FMappingList.Free;
  FElementList.Free;
  inherited;
end;

function TFhirDataElement.GetResourceType : TFhirResourceType;
begin
  result := frtDataElement;
end;

procedure TFhirDataElement.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirDataElement(oSource).urlElement.Clone;
  if (TFhirDataElement(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDataElement(oSource).FIdentifierList);
  end;
  versionElement := TFhirDataElement(oSource).versionElement.Clone;
  nameElement := TFhirDataElement(oSource).nameElement.Clone;
  FStatus := TFhirDataElement(oSource).FStatus.Link;
  experimentalElement := TFhirDataElement(oSource).experimentalElement.Clone;
  publisherElement := TFhirDataElement(oSource).publisherElement.Clone;
  if (TFhirDataElement(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirDataElementContactList.Create;
    FContactList.Assign(TFhirDataElement(oSource).FContactList);
  end;
  dateElement := TFhirDataElement(oSource).dateElement.Clone;
  if (TFhirDataElement(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirDataElement(oSource).FUseContextList);
  end;
  copyrightElement := TFhirDataElement(oSource).copyrightElement.Clone;
  FStringency := TFhirDataElement(oSource).FStringency.Link;
  if (TFhirDataElement(oSource).FMappingList = nil) then
  begin
    FMappingList.free;
    FMappingList := nil;
  end
  else
  begin
    FMappingList := TFhirDataElementMappingList.Create;
    FMappingList.Assign(TFhirDataElement(oSource).FMappingList);
  end;
  if (TFhirDataElement(oSource).FElementList = nil) then
  begin
    FElementList.free;
    FElementList := nil;
  end
  else
  begin
    FElementList := TFhirElementDefinitionList.Create;
    FElementList.Assign(TFhirDataElement(oSource).FElementList);
  end;
end;

procedure TFhirDataElement.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'stringency') Then
     list.add(FStringency.Link);
  if (child_name = 'mapping') Then
    list.addAll(FMappingList);
  if (child_name = 'element') Then
    list.addAll(FElementList);
end;

procedure TFhirDataElement.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'stringency', 'code', FStringency.Link));{1}
  oList.add(TFHIRProperty.create(self, 'mapping', '', FMappingList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'element', 'ElementDefinition', FElementList.Link)){3};
end;

procedure TFhirDataElement.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirDataElementContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'stringency') then StringencyElement := propValue as TFHIREnum
  else if (propName = 'mapping') then MappingList.add(propValue as TFhirDataElementMapping){2}
  else if (propName = 'element') then ElementList.add(propValue as TFhirElementDefinition){2}
  else inherited;
end;

function TFhirDataElement.FhirType : string;
begin
  result := 'DataElement';
end;

function TFhirDataElement.Link : TFhirDataElement;
begin
  result := TFhirDataElement(inherited Link);
end;

function TFhirDataElement.Clone : TFhirDataElement;
begin
  result := TFhirDataElement(inherited Clone);
end;

{ TFhirDataElement }

Procedure TFhirDataElement.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirDataElement.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirDataElement.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Function TFhirDataElement.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDataElement.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDataElement.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirDataElement.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirDataElement.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirDataElement.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirDataElement.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirDataElement.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirDataElement.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDataElement.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirDataElement.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirDataElement.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirDataElement.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirDataElement.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirDataElement.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirDataElement.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirDataElement.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirDataElement.GetContactList : TFhirDataElementContactList;
begin
  if FContactList = nil then
    FContactList := TFhirDataElementContactList.Create;
  result := FContactList;
end;

Function TFhirDataElement.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirDataElement.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirDataElement.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirDataElement.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Function TFhirDataElement.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirDataElement.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirDataElement.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirDataElement.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirDataElement.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirDataElement.SetStringency(value : TFhirEnum);
begin
  FStringency.free;
  FStringency := value;
end;

Function TFhirDataElement.GetStringencyST : TFhirDataelementStringency;
begin
  if FStringency = nil then
    result := TFhirDataelementStringency(0)
  else
    result := TFhirDataelementStringency(StringArrayIndexOfSensitive(CODES_TFhirDataelementStringency, FStringency.value));
end;

Procedure TFhirDataElement.SetStringencyST(value : TFhirDataelementStringency);
begin
  if ord(value) = 0 then
    StringencyElement := nil
  else
    StringencyElement := TFhirEnum.create(CODES_TFhirDataelementStringency[value]);
end;

Function TFhirDataElement.GetMappingList : TFhirDataElementMappingList;
begin
  if FMappingList = nil then
    FMappingList := TFhirDataElementMappingList.Create;
  result := FMappingList;
end;

Function TFhirDataElement.GetHasMappingList : boolean;
begin
  result := (FMappingList <> nil) and (FMappingList.count > 0);
end;

Function TFhirDataElement.GetElementList : TFhirElementDefinitionList;
begin
  if FElementList = nil then
    FElementList := TFhirElementDefinitionList.Create;
  result := FElementList;
end;

Function TFhirDataElement.GetHasElementList : boolean;
begin
  result := (FElementList <> nil) and (FElementList.count > 0);
end;


{ TFhirDataElementListEnumerator }

Constructor TFhirDataElementListEnumerator.Create(list : TFhirDataElementList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDataElementListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDataElementListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDataElementListEnumerator.GetCurrent : TFhirDataElement;
begin
  Result := FList[FIndex];
end;


{ TFhirDataElementList }
procedure TFhirDataElementList.AddItem(value: TFhirDataElement);
begin
  assert(value.ClassName = 'TFhirDataElement', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDataElement');
  add(value);
end;


function TFhirDataElementList.Append: TFhirDataElement;
begin
  result := TFhirDataElement.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementList.ClearItems;
begin
  Clear;
end;

function TFhirDataElementList.GetEnumerator : TFhirDataElementListEnumerator;
begin
  result := TFhirDataElementListEnumerator.Create(self.link);
end;

function TFhirDataElementList.Clone: TFhirDataElementList;
begin
  result := TFhirDataElementList(inherited Clone);
end;

function TFhirDataElementList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDataElementList.GetItemN(index: Integer): TFhirDataElement;
begin
  result := TFhirDataElement(ObjectByIndex[index]);
end;

function TFhirDataElementList.IndexOf(value: TFhirDataElement): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDataElementList.Insert(index: Integer): TFhirDataElement;
begin
  result := TFhirDataElement.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDataElementList.InsertItem(index: Integer; value: TFhirDataElement);
begin
  assert(value is TFhirDataElement);
  Inherited Insert(index, value);
end;

function TFhirDataElementList.Item(index: Integer): TFhirDataElement;
begin
  result := TFhirDataElement(ObjectByIndex[index]);
end;

function TFhirDataElementList.Link: TFhirDataElementList;
begin
  result := TFhirDataElementList(inherited Link);
end;

procedure TFhirDataElementList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDataElementList.SetItemByIndex(index: Integer; value: TFhirDataElement);
begin
  assert(value is TFhirDataElement);
  FhirDataElements[index] := value;
end;

procedure TFhirDataElementList.SetItemN(index: Integer; value: TFhirDataElement);
begin
  assert(value is TFhirDataElement);
  ObjectByIndex[index] := value;
end;

{ TFhirDetectedIssueMitigation }

constructor TFhirDetectedIssueMitigation.Create;
begin
  inherited;
end;

destructor TFhirDetectedIssueMitigation.Destroy;
begin
  FAction.free;
  FDate.free;
  FAuthor.free;
  inherited;
end;

procedure TFhirDetectedIssueMitigation.Assign(oSource : TAdvObject);
begin
  inherited;
  action := TFhirDetectedIssueMitigation(oSource).action.Clone;
  dateElement := TFhirDetectedIssueMitigation(oSource).dateElement.Clone;
  author := TFhirDetectedIssueMitigation(oSource).author.Clone;
end;

procedure TFhirDetectedIssueMitigation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'action') Then
     list.add(FAction.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
end;

procedure TFhirDetectedIssueMitigation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'action', 'CodeableConcept', FAction.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner)', FAuthor.Link));{2}
end;

procedure TFhirDetectedIssueMitigation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'action') then Action := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'author') then Author := propValue as TFhirReference{TFhirPractitioner}{4b}
  else inherited;
end;

function TFhirDetectedIssueMitigation.FhirType : string;
begin
  result := 'mitigation';
end;

function TFhirDetectedIssueMitigation.Link : TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation(inherited Link);
end;

function TFhirDetectedIssueMitigation.Clone : TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation(inherited Clone);
end;

function TFhirDetectedIssueMitigation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDetectedIssueMitigation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDetectedIssueMitigation)) then
    result := false
  else
  begin
    o := TFhirDetectedIssueMitigation(other);
    result := compareDeep(actionElement, o.actionElement, true) and compareDeep(dateElement, o.dateElement, true) and 
      compareDeep(authorElement, o.authorElement, true);
  end;
end;

function TFhirDetectedIssueMitigation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDetectedIssueMitigation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDetectedIssueMitigation)) then
    result := false
  else
  begin
    o := TFhirDetectedIssueMitigation(other);
    result := compareValues(dateElement, o.dateElement, true);
  end;
end;

{ TFhirDetectedIssueMitigation }

Procedure TFhirDetectedIssueMitigation.SetAction(value : TFhirCodeableConcept);
begin
  FAction.free;
  FAction := value;
end;

Procedure TFhirDetectedIssueMitigation.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirDetectedIssueMitigation.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirDetectedIssueMitigation.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirDetectedIssueMitigation.SetAuthor(value : TFhirReference{TFhirPractitioner});
begin
  FAuthor.free;
  FAuthor := value;
end;


{ TFhirDetectedIssueMitigationListEnumerator }

Constructor TFhirDetectedIssueMitigationListEnumerator.Create(list : TFhirDetectedIssueMitigationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDetectedIssueMitigationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDetectedIssueMitigationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDetectedIssueMitigationListEnumerator.GetCurrent : TFhirDetectedIssueMitigation;
begin
  Result := FList[FIndex];
end;


{ TFhirDetectedIssueMitigationList }
procedure TFhirDetectedIssueMitigationList.AddItem(value: TFhirDetectedIssueMitigation);
begin
  assert(value.ClassName = 'TFhirDetectedIssueMitigation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDetectedIssueMitigation');
  add(value);
end;


function TFhirDetectedIssueMitigationList.Append: TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDetectedIssueMitigationList.ClearItems;
begin
  Clear;
end;

function TFhirDetectedIssueMitigationList.GetEnumerator : TFhirDetectedIssueMitigationListEnumerator;
begin
  result := TFhirDetectedIssueMitigationListEnumerator.Create(self.link);
end;

function TFhirDetectedIssueMitigationList.Clone: TFhirDetectedIssueMitigationList;
begin
  result := TFhirDetectedIssueMitigationList(inherited Clone);
end;

function TFhirDetectedIssueMitigationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDetectedIssueMitigationList.GetItemN(index: Integer): TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation(ObjectByIndex[index]);
end;

function TFhirDetectedIssueMitigationList.IndexOf(value: TFhirDetectedIssueMitigation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDetectedIssueMitigationList.Insert(index: Integer): TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDetectedIssueMitigationList.InsertItem(index: Integer; value: TFhirDetectedIssueMitigation);
begin
  assert(value is TFhirDetectedIssueMitigation);
  Inherited Insert(index, value);
end;

function TFhirDetectedIssueMitigationList.Item(index: Integer): TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation(ObjectByIndex[index]);
end;

function TFhirDetectedIssueMitigationList.Link: TFhirDetectedIssueMitigationList;
begin
  result := TFhirDetectedIssueMitigationList(inherited Link);
end;

procedure TFhirDetectedIssueMitigationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDetectedIssueMitigationList.SetItemByIndex(index: Integer; value: TFhirDetectedIssueMitigation);
begin
  assert(value is TFhirDetectedIssueMitigation);
  FhirDetectedIssueMitigations[index] := value;
end;

procedure TFhirDetectedIssueMitigationList.SetItemN(index: Integer; value: TFhirDetectedIssueMitigation);
begin
  assert(value is TFhirDetectedIssueMitigation);
  ObjectByIndex[index] := value;
end;

{ TFhirDetectedIssue }

constructor TFhirDetectedIssue.Create;
begin
  inherited;
end;

destructor TFhirDetectedIssue.Destroy;
begin
  FPatient.free;
  FCategory.free;
  FSeverity.free;
  FImplicatedList.Free;
  FDetail.free;
  FDate.free;
  FAuthor.free;
  FIdentifier.free;
  FReference.free;
  FMitigationList.Free;
  inherited;
end;

function TFhirDetectedIssue.GetResourceType : TFhirResourceType;
begin
  result := frtDetectedIssue;
end;

procedure TFhirDetectedIssue.Assign(oSource : TAdvObject);
begin
  inherited;
  patient := TFhirDetectedIssue(oSource).patient.Clone;
  category := TFhirDetectedIssue(oSource).category.Clone;
  FSeverity := TFhirDetectedIssue(oSource).FSeverity.Link;
  if (TFhirDetectedIssue(oSource).FImplicatedList = nil) then
  begin
    FImplicatedList.free;
    FImplicatedList := nil;
  end
  else
  begin
    FImplicatedList := TFhirReferenceList{TFhirReference}.Create;
    FImplicatedList.Assign(TFhirDetectedIssue(oSource).FImplicatedList);
  end;
  detailElement := TFhirDetectedIssue(oSource).detailElement.Clone;
  dateElement := TFhirDetectedIssue(oSource).dateElement.Clone;
  author := TFhirDetectedIssue(oSource).author.Clone;
  identifier := TFhirDetectedIssue(oSource).identifier.Clone;
  referenceElement := TFhirDetectedIssue(oSource).referenceElement.Clone;
  if (TFhirDetectedIssue(oSource).FMitigationList = nil) then
  begin
    FMitigationList.free;
    FMitigationList := nil;
  end
  else
  begin
    FMitigationList := TFhirDetectedIssueMitigationList.Create;
    FMitigationList.Assign(TFhirDetectedIssue(oSource).FMitigationList);
  end;
end;

procedure TFhirDetectedIssue.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'severity') Then
     list.add(FSeverity.Link);
  if (child_name = 'implicated') Then
    list.addAll(FImplicatedList);
  if (child_name = 'detail') Then
     list.add(FDetail.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'mitigation') Then
    list.addAll(FMitigationList);
end;

procedure TFhirDetectedIssue.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'severity', 'code', FSeverity.Link));{1}
  oList.add(TFHIRProperty.create(self, 'implicated', 'Reference(Any)', FImplicatedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'detail', 'string', FDetail.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Device)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reference', 'uri', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'mitigation', '', FMitigationList.Link)){3};
end;

procedure TFhirDetectedIssue.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'severity') then SeverityElement := propValue as TFHIREnum
  else if (propName = 'implicated') then ImplicatedList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'detail') then DetailElement := propValue as TFhirString{5a}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'reference') then ReferenceElement := propValue as TFhirUri{5a}
  else if (propName = 'mitigation') then MitigationList.add(propValue as TFhirDetectedIssueMitigation){2}
  else inherited;
end;

function TFhirDetectedIssue.FhirType : string;
begin
  result := 'DetectedIssue';
end;

function TFhirDetectedIssue.Link : TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue(inherited Link);
end;

function TFhirDetectedIssue.Clone : TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue(inherited Clone);
end;

{ TFhirDetectedIssue }

Procedure TFhirDetectedIssue.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirDetectedIssue.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirDetectedIssue.SetSeverity(value : TFhirEnum);
begin
  FSeverity.free;
  FSeverity := value;
end;

Function TFhirDetectedIssue.GetSeverityST : TFhirDetectedissueSeverity;
begin
  if FSeverity = nil then
    result := TFhirDetectedissueSeverity(0)
  else
    result := TFhirDetectedissueSeverity(StringArrayIndexOfSensitive(CODES_TFhirDetectedissueSeverity, FSeverity.value));
end;

Procedure TFhirDetectedIssue.SetSeverityST(value : TFhirDetectedissueSeverity);
begin
  if ord(value) = 0 then
    SeverityElement := nil
  else
    SeverityElement := TFhirEnum.create(CODES_TFhirDetectedissueSeverity[value]);
end;

Function TFhirDetectedIssue.GetImplicatedList : TFhirReferenceList{TFhirReference};
begin
  if FImplicatedList = nil then
    FImplicatedList := TFhirReferenceList{TFhirReference}.Create;
  result := FImplicatedList;
end;

Function TFhirDetectedIssue.GetHasImplicatedList : boolean;
begin
  result := (FImplicatedList <> nil) and (FImplicatedList.count > 0);
end;

Procedure TFhirDetectedIssue.SetDetail(value : TFhirString);
begin
  FDetail.free;
  FDetail := value;
end;

Function TFhirDetectedIssue.GetDetailST : String;
begin
  if FDetail = nil then
    result := ''
  else
    result := FDetail.value;
end;

Procedure TFhirDetectedIssue.SetDetailST(value : String);
begin
  if value <> '' then
  begin
    if FDetail = nil then
      FDetail := TFhirString.create;
    FDetail.value := value
  end
  else if FDetail <> nil then
    FDetail.value := '';
end;

Procedure TFhirDetectedIssue.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirDetectedIssue.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirDetectedIssue.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirDetectedIssue.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirDetectedIssue.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirDetectedIssue.SetReference(value : TFhirUri);
begin
  FReference.free;
  FReference := value;
end;

Function TFhirDetectedIssue.GetReferenceST : String;
begin
  if FReference = nil then
    result := ''
  else
    result := FReference.value;
end;

Procedure TFhirDetectedIssue.SetReferenceST(value : String);
begin
  if value <> '' then
  begin
    if FReference = nil then
      FReference := TFhirUri.create;
    FReference.value := value
  end
  else if FReference <> nil then
    FReference.value := '';
end;

Function TFhirDetectedIssue.GetMitigationList : TFhirDetectedIssueMitigationList;
begin
  if FMitigationList = nil then
    FMitigationList := TFhirDetectedIssueMitigationList.Create;
  result := FMitigationList;
end;

Function TFhirDetectedIssue.GetHasMitigationList : boolean;
begin
  result := (FMitigationList <> nil) and (FMitigationList.count > 0);
end;


{ TFhirDetectedIssueListEnumerator }

Constructor TFhirDetectedIssueListEnumerator.Create(list : TFhirDetectedIssueList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDetectedIssueListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDetectedIssueListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDetectedIssueListEnumerator.GetCurrent : TFhirDetectedIssue;
begin
  Result := FList[FIndex];
end;


{ TFhirDetectedIssueList }
procedure TFhirDetectedIssueList.AddItem(value: TFhirDetectedIssue);
begin
  assert(value.ClassName = 'TFhirDetectedIssue', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDetectedIssue');
  add(value);
end;


function TFhirDetectedIssueList.Append: TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDetectedIssueList.ClearItems;
begin
  Clear;
end;

function TFhirDetectedIssueList.GetEnumerator : TFhirDetectedIssueListEnumerator;
begin
  result := TFhirDetectedIssueListEnumerator.Create(self.link);
end;

function TFhirDetectedIssueList.Clone: TFhirDetectedIssueList;
begin
  result := TFhirDetectedIssueList(inherited Clone);
end;

function TFhirDetectedIssueList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDetectedIssueList.GetItemN(index: Integer): TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue(ObjectByIndex[index]);
end;

function TFhirDetectedIssueList.IndexOf(value: TFhirDetectedIssue): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDetectedIssueList.Insert(index: Integer): TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDetectedIssueList.InsertItem(index: Integer; value: TFhirDetectedIssue);
begin
  assert(value is TFhirDetectedIssue);
  Inherited Insert(index, value);
end;

function TFhirDetectedIssueList.Item(index: Integer): TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue(ObjectByIndex[index]);
end;

function TFhirDetectedIssueList.Link: TFhirDetectedIssueList;
begin
  result := TFhirDetectedIssueList(inherited Link);
end;

procedure TFhirDetectedIssueList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDetectedIssueList.SetItemByIndex(index: Integer; value: TFhirDetectedIssue);
begin
  assert(value is TFhirDetectedIssue);
  FhirDetectedIssues[index] := value;
end;

procedure TFhirDetectedIssueList.SetItemN(index: Integer; value: TFhirDetectedIssue);
begin
  assert(value is TFhirDetectedIssue);
  ObjectByIndex[index] := value;
end;

{ TFhirDevice }

constructor TFhirDevice.Create;
begin
  inherited;
end;

destructor TFhirDevice.Destroy;
begin
  FIdentifierList.Free;
  FType_.free;
  FNoteList.Free;
  FStatus.free;
  FManufacturer.free;
  FModel.free;
  FVersion.free;
  FManufactureDate.free;
  FExpiry.free;
  FUdi.free;
  FLotNumber.free;
  FOwner.free;
  FLocation.free;
  FPatient.free;
  FContactList.Free;
  FUrl.free;
  inherited;
end;

function TFhirDevice.GetResourceType : TFhirResourceType;
begin
  result := frtDevice;
end;

procedure TFhirDevice.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirDevice(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDevice(oSource).FIdentifierList);
  end;
  type_ := TFhirDevice(oSource).type_.Clone;
  if (TFhirDevice(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirAnnotationList.Create;
    FNoteList.Assign(TFhirDevice(oSource).FNoteList);
  end;
  FStatus := TFhirDevice(oSource).FStatus.Link;
  manufacturerElement := TFhirDevice(oSource).manufacturerElement.Clone;
  modelElement := TFhirDevice(oSource).modelElement.Clone;
  versionElement := TFhirDevice(oSource).versionElement.Clone;
  manufactureDateElement := TFhirDevice(oSource).manufactureDateElement.Clone;
  expiryElement := TFhirDevice(oSource).expiryElement.Clone;
  udiElement := TFhirDevice(oSource).udiElement.Clone;
  lotNumberElement := TFhirDevice(oSource).lotNumberElement.Clone;
  owner := TFhirDevice(oSource).owner.Clone;
  location := TFhirDevice(oSource).location.Clone;
  patient := TFhirDevice(oSource).patient.Clone;
  if (TFhirDevice(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirContactPointList.Create;
    FContactList.Assign(TFhirDevice(oSource).FContactList);
  end;
  urlElement := TFhirDevice(oSource).urlElement.Clone;
end;

procedure TFhirDevice.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'manufacturer') Then
     list.add(FManufacturer.Link);
  if (child_name = 'model') Then
     list.add(FModel.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'manufactureDate') Then
     list.add(FManufactureDate.Link);
  if (child_name = 'expiry') Then
     list.add(FExpiry.Link);
  if (child_name = 'udi') Then
     list.add(FUdi.Link);
  if (child_name = 'lotNumber') Then
     list.add(FLotNumber.Link);
  if (child_name = 'owner') Then
     list.add(FOwner.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
end;

procedure TFhirDevice.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNoteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'manufacturer', 'string', FManufacturer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'model', 'string', FModel.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'manufactureDate', 'dateTime', FManufactureDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expiry', 'dateTime', FExpiry.Link));{2}
  oList.add(TFHIRProperty.create(self, 'udi', 'string', FUdi.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lotNumber', 'string', FLotNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'owner', 'Reference(Organization)', FOwner.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', 'ContactPoint', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
end;

procedure TFhirDevice.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'note') then NoteList.add(propValue as TFhirAnnotation){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'manufacturer') then ManufacturerElement := propValue as TFhirString{5a}
  else if (propName = 'model') then ModelElement := propValue as TFhirString{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'manufactureDate') then ManufactureDateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'expiry') then ExpiryElement := propValue as TFhirDateTime{5a}
  else if (propName = 'udi') then UdiElement := propValue as TFhirString{5a}
  else if (propName = 'lotNumber') then LotNumberElement := propValue as TFhirString{5a}
  else if (propName = 'owner') then Owner := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirDevice.FhirType : string;
begin
  result := 'Device';
end;

function TFhirDevice.Link : TFhirDevice;
begin
  result := TFhirDevice(inherited Link);
end;

function TFhirDevice.Clone : TFhirDevice;
begin
  result := TFhirDevice(inherited Clone);
end;

{ TFhirDevice }

Function TFhirDevice.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDevice.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDevice.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirDevice.GetNoteList : TFhirAnnotationList;
begin
  if FNoteList = nil then
    FNoteList := TFhirAnnotationList.Create;
  result := FNoteList;
end;

Function TFhirDevice.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;

Procedure TFhirDevice.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDevice.GetStatusST : TFhirDevicestatus;
begin
  if FStatus = nil then
    result := TFhirDevicestatus(0)
  else
    result := TFhirDevicestatus(StringArrayIndexOfSensitive(CODES_TFhirDevicestatus, FStatus.value));
end;

Procedure TFhirDevice.SetStatusST(value : TFhirDevicestatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDevicestatus[value]);
end;

Procedure TFhirDevice.SetManufacturer(value : TFhirString);
begin
  FManufacturer.free;
  FManufacturer := value;
end;

Function TFhirDevice.GetManufacturerST : String;
begin
  if FManufacturer = nil then
    result := ''
  else
    result := FManufacturer.value;
end;

Procedure TFhirDevice.SetManufacturerST(value : String);
begin
  if value <> '' then
  begin
    if FManufacturer = nil then
      FManufacturer := TFhirString.create;
    FManufacturer.value := value
  end
  else if FManufacturer <> nil then
    FManufacturer.value := '';
end;

Procedure TFhirDevice.SetModel(value : TFhirString);
begin
  FModel.free;
  FModel := value;
end;

Function TFhirDevice.GetModelST : String;
begin
  if FModel = nil then
    result := ''
  else
    result := FModel.value;
end;

Procedure TFhirDevice.SetModelST(value : String);
begin
  if value <> '' then
  begin
    if FModel = nil then
      FModel := TFhirString.create;
    FModel.value := value
  end
  else if FModel <> nil then
    FModel.value := '';
end;

Procedure TFhirDevice.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirDevice.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirDevice.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirDevice.SetManufactureDate(value : TFhirDateTime);
begin
  FManufactureDate.free;
  FManufactureDate := value;
end;

Function TFhirDevice.GetManufactureDateST : TDateAndTime;
begin
  if FManufactureDate = nil then
    result := nil
  else
    result := FManufactureDate.value;
end;

Procedure TFhirDevice.SetManufactureDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FManufactureDate = nil then
      FManufactureDate := TFhirDateTime.create;
    FManufactureDate.value := value
  end
  else if FManufactureDate <> nil then
    FManufactureDate.value := nil;
end;

Procedure TFhirDevice.SetExpiry(value : TFhirDateTime);
begin
  FExpiry.free;
  FExpiry := value;
end;

Function TFhirDevice.GetExpiryST : TDateAndTime;
begin
  if FExpiry = nil then
    result := nil
  else
    result := FExpiry.value;
end;

Procedure TFhirDevice.SetExpiryST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FExpiry = nil then
      FExpiry := TFhirDateTime.create;
    FExpiry.value := value
  end
  else if FExpiry <> nil then
    FExpiry.value := nil;
end;

Procedure TFhirDevice.SetUdi(value : TFhirString);
begin
  FUdi.free;
  FUdi := value;
end;

Function TFhirDevice.GetUdiST : String;
begin
  if FUdi = nil then
    result := ''
  else
    result := FUdi.value;
end;

Procedure TFhirDevice.SetUdiST(value : String);
begin
  if value <> '' then
  begin
    if FUdi = nil then
      FUdi := TFhirString.create;
    FUdi.value := value
  end
  else if FUdi <> nil then
    FUdi.value := '';
end;

Procedure TFhirDevice.SetLotNumber(value : TFhirString);
begin
  FLotNumber.free;
  FLotNumber := value;
end;

Function TFhirDevice.GetLotNumberST : String;
begin
  if FLotNumber = nil then
    result := ''
  else
    result := FLotNumber.value;
end;

Procedure TFhirDevice.SetLotNumberST(value : String);
begin
  if value <> '' then
  begin
    if FLotNumber = nil then
      FLotNumber := TFhirString.create;
    FLotNumber.value := value
  end
  else if FLotNumber <> nil then
    FLotNumber.value := '';
end;

Procedure TFhirDevice.SetOwner(value : TFhirReference{TFhirOrganization});
begin
  FOwner.free;
  FOwner := value;
end;

Procedure TFhirDevice.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Procedure TFhirDevice.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Function TFhirDevice.GetContactList : TFhirContactPointList;
begin
  if FContactList = nil then
    FContactList := TFhirContactPointList.Create;
  result := FContactList;
end;

Function TFhirDevice.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirDevice.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirDevice.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirDevice.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;


{ TFhirDeviceListEnumerator }

Constructor TFhirDeviceListEnumerator.Create(list : TFhirDeviceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceListEnumerator.GetCurrent : TFhirDevice;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceList }
procedure TFhirDeviceList.AddItem(value: TFhirDevice);
begin
  assert(value.ClassName = 'TFhirDevice', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDevice');
  add(value);
end;


function TFhirDeviceList.Append: TFhirDevice;
begin
  result := TFhirDevice.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceList.GetEnumerator : TFhirDeviceListEnumerator;
begin
  result := TFhirDeviceListEnumerator.Create(self.link);
end;

function TFhirDeviceList.Clone: TFhirDeviceList;
begin
  result := TFhirDeviceList(inherited Clone);
end;

function TFhirDeviceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceList.GetItemN(index: Integer): TFhirDevice;
begin
  result := TFhirDevice(ObjectByIndex[index]);
end;

function TFhirDeviceList.IndexOf(value: TFhirDevice): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceList.Insert(index: Integer): TFhirDevice;
begin
  result := TFhirDevice.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceList.InsertItem(index: Integer; value: TFhirDevice);
begin
  assert(value is TFhirDevice);
  Inherited Insert(index, value);
end;

function TFhirDeviceList.Item(index: Integer): TFhirDevice;
begin
  result := TFhirDevice(ObjectByIndex[index]);
end;

function TFhirDeviceList.Link: TFhirDeviceList;
begin
  result := TFhirDeviceList(inherited Link);
end;

procedure TFhirDeviceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceList.SetItemByIndex(index: Integer; value: TFhirDevice);
begin
  assert(value is TFhirDevice);
  FhirDevices[index] := value;
end;

procedure TFhirDeviceList.SetItemN(index: Integer; value: TFhirDevice);
begin
  assert(value is TFhirDevice);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceComponentProductionSpecification }

constructor TFhirDeviceComponentProductionSpecification.Create;
begin
  inherited;
end;

destructor TFhirDeviceComponentProductionSpecification.Destroy;
begin
  FSpecType.free;
  FComponentId.free;
  FProductionSpec.free;
  inherited;
end;

procedure TFhirDeviceComponentProductionSpecification.Assign(oSource : TAdvObject);
begin
  inherited;
  specType := TFhirDeviceComponentProductionSpecification(oSource).specType.Clone;
  componentId := TFhirDeviceComponentProductionSpecification(oSource).componentId.Clone;
  productionSpecElement := TFhirDeviceComponentProductionSpecification(oSource).productionSpecElement.Clone;
end;

procedure TFhirDeviceComponentProductionSpecification.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'specType') Then
     list.add(FSpecType.Link);
  if (child_name = 'componentId') Then
     list.add(FComponentId.Link);
  if (child_name = 'productionSpec') Then
     list.add(FProductionSpec.Link);
end;

procedure TFhirDeviceComponentProductionSpecification.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'specType', 'CodeableConcept', FSpecType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'componentId', 'Identifier', FComponentId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'productionSpec', 'string', FProductionSpec.Link));{2}
end;

procedure TFhirDeviceComponentProductionSpecification.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'specType') then SpecType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'componentId') then ComponentId := propValue as TFhirIdentifier{4b}
  else if (propName = 'productionSpec') then ProductionSpecElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirDeviceComponentProductionSpecification.FhirType : string;
begin
  result := 'productionSpecification';
end;

function TFhirDeviceComponentProductionSpecification.Link : TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification(inherited Link);
end;

function TFhirDeviceComponentProductionSpecification.Clone : TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification(inherited Clone);
end;

function TFhirDeviceComponentProductionSpecification.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDeviceComponentProductionSpecification;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDeviceComponentProductionSpecification)) then
    result := false
  else
  begin
    o := TFhirDeviceComponentProductionSpecification(other);
    result := compareDeep(specTypeElement, o.specTypeElement, true) and compareDeep(componentIdElement, o.componentIdElement, true) and 
      compareDeep(productionSpecElement, o.productionSpecElement, true);
  end;
end;

function TFhirDeviceComponentProductionSpecification.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDeviceComponentProductionSpecification;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDeviceComponentProductionSpecification)) then
    result := false
  else
  begin
    o := TFhirDeviceComponentProductionSpecification(other);
    result := compareValues(productionSpecElement, o.productionSpecElement, true);
  end;
end;

{ TFhirDeviceComponentProductionSpecification }

Procedure TFhirDeviceComponentProductionSpecification.SetSpecType(value : TFhirCodeableConcept);
begin
  FSpecType.free;
  FSpecType := value;
end;

Procedure TFhirDeviceComponentProductionSpecification.SetComponentId(value : TFhirIdentifier);
begin
  FComponentId.free;
  FComponentId := value;
end;

Procedure TFhirDeviceComponentProductionSpecification.SetProductionSpec(value : TFhirString);
begin
  FProductionSpec.free;
  FProductionSpec := value;
end;

Function TFhirDeviceComponentProductionSpecification.GetProductionSpecST : String;
begin
  if FProductionSpec = nil then
    result := ''
  else
    result := FProductionSpec.value;
end;

Procedure TFhirDeviceComponentProductionSpecification.SetProductionSpecST(value : String);
begin
  if value <> '' then
  begin
    if FProductionSpec = nil then
      FProductionSpec := TFhirString.create;
    FProductionSpec.value := value
  end
  else if FProductionSpec <> nil then
    FProductionSpec.value := '';
end;


{ TFhirDeviceComponentProductionSpecificationListEnumerator }

Constructor TFhirDeviceComponentProductionSpecificationListEnumerator.Create(list : TFhirDeviceComponentProductionSpecificationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceComponentProductionSpecificationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceComponentProductionSpecificationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceComponentProductionSpecificationListEnumerator.GetCurrent : TFhirDeviceComponentProductionSpecification;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceComponentProductionSpecificationList }
procedure TFhirDeviceComponentProductionSpecificationList.AddItem(value: TFhirDeviceComponentProductionSpecification);
begin
  assert(value.ClassName = 'TFhirDeviceComponentProductionSpecification', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceComponentProductionSpecification');
  add(value);
end;


function TFhirDeviceComponentProductionSpecificationList.Append: TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceComponentProductionSpecificationList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceComponentProductionSpecificationList.GetEnumerator : TFhirDeviceComponentProductionSpecificationListEnumerator;
begin
  result := TFhirDeviceComponentProductionSpecificationListEnumerator.Create(self.link);
end;

function TFhirDeviceComponentProductionSpecificationList.Clone: TFhirDeviceComponentProductionSpecificationList;
begin
  result := TFhirDeviceComponentProductionSpecificationList(inherited Clone);
end;

function TFhirDeviceComponentProductionSpecificationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceComponentProductionSpecificationList.GetItemN(index: Integer): TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification(ObjectByIndex[index]);
end;

function TFhirDeviceComponentProductionSpecificationList.IndexOf(value: TFhirDeviceComponentProductionSpecification): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceComponentProductionSpecificationList.Insert(index: Integer): TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceComponentProductionSpecificationList.InsertItem(index: Integer; value: TFhirDeviceComponentProductionSpecification);
begin
  assert(value is TFhirDeviceComponentProductionSpecification);
  Inherited Insert(index, value);
end;

function TFhirDeviceComponentProductionSpecificationList.Item(index: Integer): TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification(ObjectByIndex[index]);
end;

function TFhirDeviceComponentProductionSpecificationList.Link: TFhirDeviceComponentProductionSpecificationList;
begin
  result := TFhirDeviceComponentProductionSpecificationList(inherited Link);
end;

procedure TFhirDeviceComponentProductionSpecificationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceComponentProductionSpecificationList.SetItemByIndex(index: Integer; value: TFhirDeviceComponentProductionSpecification);
begin
  assert(value is TFhirDeviceComponentProductionSpecification);
  FhirDeviceComponentProductionSpecifications[index] := value;
end;

procedure TFhirDeviceComponentProductionSpecificationList.SetItemN(index: Integer; value: TFhirDeviceComponentProductionSpecification);
begin
  assert(value is TFhirDeviceComponentProductionSpecification);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceComponent }

constructor TFhirDeviceComponent.Create;
begin
  inherited;
end;

destructor TFhirDeviceComponent.Destroy;
begin
  FType_.free;
  FIdentifier.free;
  FLastSystemChange.free;
  FSource.free;
  FParent.free;
  FOperationalStatusList.Free;
  FParameterGroup.free;
  FMeasurementPrinciple.free;
  FProductionSpecificationList.Free;
  FLanguageCode.free;
  inherited;
end;

function TFhirDeviceComponent.GetResourceType : TFhirResourceType;
begin
  result := frtDeviceComponent;
end;

procedure TFhirDeviceComponent.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirDeviceComponent(oSource).type_.Clone;
  identifier := TFhirDeviceComponent(oSource).identifier.Clone;
  lastSystemChangeElement := TFhirDeviceComponent(oSource).lastSystemChangeElement.Clone;
  source := TFhirDeviceComponent(oSource).source.Clone;
  parent := TFhirDeviceComponent(oSource).parent.Clone;
  if (TFhirDeviceComponent(oSource).FOperationalStatusList = nil) then
  begin
    FOperationalStatusList.free;
    FOperationalStatusList := nil;
  end
  else
  begin
    FOperationalStatusList := TFhirCodeableConceptList.Create;
    FOperationalStatusList.Assign(TFhirDeviceComponent(oSource).FOperationalStatusList);
  end;
  parameterGroup := TFhirDeviceComponent(oSource).parameterGroup.Clone;
  FMeasurementPrinciple := TFhirDeviceComponent(oSource).FMeasurementPrinciple.Link;
  if (TFhirDeviceComponent(oSource).FProductionSpecificationList = nil) then
  begin
    FProductionSpecificationList.free;
    FProductionSpecificationList := nil;
  end
  else
  begin
    FProductionSpecificationList := TFhirDeviceComponentProductionSpecificationList.Create;
    FProductionSpecificationList.Assign(TFhirDeviceComponent(oSource).FProductionSpecificationList);
  end;
  languageCode := TFhirDeviceComponent(oSource).languageCode.Clone;
end;

procedure TFhirDeviceComponent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'lastSystemChange') Then
     list.add(FLastSystemChange.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'parent') Then
     list.add(FParent.Link);
  if (child_name = 'operationalStatus') Then
    list.addAll(FOperationalStatusList);
  if (child_name = 'parameterGroup') Then
     list.add(FParameterGroup.Link);
  if (child_name = 'measurementPrinciple') Then
     list.add(FMeasurementPrinciple.Link);
  if (child_name = 'productionSpecification') Then
    list.addAll(FProductionSpecificationList);
  if (child_name = 'languageCode') Then
     list.add(FLanguageCode.Link);
end;

procedure TFhirDeviceComponent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lastSystemChange', 'instant', FLastSystemChange.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Device)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'parent', 'Reference(DeviceComponent)', FParent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'operationalStatus', 'CodeableConcept', FOperationalStatusList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'parameterGroup', 'CodeableConcept', FParameterGroup.Link));{2}
  oList.add(TFHIRProperty.create(self, 'measurementPrinciple', 'code', FMeasurementPrinciple.Link));{1}
  oList.add(TFHIRProperty.create(self, 'productionSpecification', '', FProductionSpecificationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'languageCode', 'CodeableConcept', FLanguageCode.Link));{2}
end;

procedure TFhirDeviceComponent.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'lastSystemChange') then LastSystemChangeElement := propValue as TFhirInstant{5a}
  else if (propName = 'source') then Source := propValue as TFhirReference{TFhirDevice}{4b}
  else if (propName = 'parent') then Parent := propValue as TFhirReference{TFhirDeviceComponent}{4b}
  else if (propName = 'operationalStatus') then OperationalStatusList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'parameterGroup') then ParameterGroup := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'measurementPrinciple') then MeasurementPrincipleElement := propValue as TFHIREnum
  else if (propName = 'productionSpecification') then ProductionSpecificationList.add(propValue as TFhirDeviceComponentProductionSpecification){2}
  else if (propName = 'languageCode') then LanguageCode := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirDeviceComponent.FhirType : string;
begin
  result := 'DeviceComponent';
end;

function TFhirDeviceComponent.Link : TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent(inherited Link);
end;

function TFhirDeviceComponent.Clone : TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent(inherited Clone);
end;

{ TFhirDeviceComponent }

Procedure TFhirDeviceComponent.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirDeviceComponent.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirDeviceComponent.SetLastSystemChange(value : TFhirInstant);
begin
  FLastSystemChange.free;
  FLastSystemChange := value;
end;

Function TFhirDeviceComponent.GetLastSystemChangeST : TDateAndTime;
begin
  if FLastSystemChange = nil then
    result := nil
  else
    result := FLastSystemChange.value;
end;

Procedure TFhirDeviceComponent.SetLastSystemChangeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FLastSystemChange = nil then
      FLastSystemChange := TFhirInstant.create;
    FLastSystemChange.value := value
  end
  else if FLastSystemChange <> nil then
    FLastSystemChange.value := nil;
end;

Procedure TFhirDeviceComponent.SetSource(value : TFhirReference{TFhirDevice});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirDeviceComponent.SetParent(value : TFhirReference{TFhirDeviceComponent});
begin
  FParent.free;
  FParent := value;
end;

Function TFhirDeviceComponent.GetOperationalStatusList : TFhirCodeableConceptList;
begin
  if FOperationalStatusList = nil then
    FOperationalStatusList := TFhirCodeableConceptList.Create;
  result := FOperationalStatusList;
end;

Function TFhirDeviceComponent.GetHasOperationalStatusList : boolean;
begin
  result := (FOperationalStatusList <> nil) and (FOperationalStatusList.count > 0);
end;

Procedure TFhirDeviceComponent.SetParameterGroup(value : TFhirCodeableConcept);
begin
  FParameterGroup.free;
  FParameterGroup := value;
end;

Procedure TFhirDeviceComponent.SetMeasurementPrinciple(value : TFhirEnum);
begin
  FMeasurementPrinciple.free;
  FMeasurementPrinciple := value;
end;

Function TFhirDeviceComponent.GetMeasurementPrincipleST : TFhirMeasurementPrinciple;
begin
  if FMeasurementPrinciple = nil then
    result := TFhirMeasurementPrinciple(0)
  else
    result := TFhirMeasurementPrinciple(StringArrayIndexOfSensitive(CODES_TFhirMeasurementPrinciple, FMeasurementPrinciple.value));
end;

Procedure TFhirDeviceComponent.SetMeasurementPrincipleST(value : TFhirMeasurementPrinciple);
begin
  if ord(value) = 0 then
    MeasurementPrincipleElement := nil
  else
    MeasurementPrincipleElement := TFhirEnum.create(CODES_TFhirMeasurementPrinciple[value]);
end;

Function TFhirDeviceComponent.GetProductionSpecificationList : TFhirDeviceComponentProductionSpecificationList;
begin
  if FProductionSpecificationList = nil then
    FProductionSpecificationList := TFhirDeviceComponentProductionSpecificationList.Create;
  result := FProductionSpecificationList;
end;

Function TFhirDeviceComponent.GetHasProductionSpecificationList : boolean;
begin
  result := (FProductionSpecificationList <> nil) and (FProductionSpecificationList.count > 0);
end;

Procedure TFhirDeviceComponent.SetLanguageCode(value : TFhirCodeableConcept);
begin
  FLanguageCode.free;
  FLanguageCode := value;
end;


{ TFhirDeviceComponentListEnumerator }

Constructor TFhirDeviceComponentListEnumerator.Create(list : TFhirDeviceComponentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceComponentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceComponentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceComponentListEnumerator.GetCurrent : TFhirDeviceComponent;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceComponentList }
procedure TFhirDeviceComponentList.AddItem(value: TFhirDeviceComponent);
begin
  assert(value.ClassName = 'TFhirDeviceComponent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceComponent');
  add(value);
end;


function TFhirDeviceComponentList.Append: TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceComponentList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceComponentList.GetEnumerator : TFhirDeviceComponentListEnumerator;
begin
  result := TFhirDeviceComponentListEnumerator.Create(self.link);
end;

function TFhirDeviceComponentList.Clone: TFhirDeviceComponentList;
begin
  result := TFhirDeviceComponentList(inherited Clone);
end;

function TFhirDeviceComponentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceComponentList.GetItemN(index: Integer): TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent(ObjectByIndex[index]);
end;

function TFhirDeviceComponentList.IndexOf(value: TFhirDeviceComponent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceComponentList.Insert(index: Integer): TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceComponentList.InsertItem(index: Integer; value: TFhirDeviceComponent);
begin
  assert(value is TFhirDeviceComponent);
  Inherited Insert(index, value);
end;

function TFhirDeviceComponentList.Item(index: Integer): TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent(ObjectByIndex[index]);
end;

function TFhirDeviceComponentList.Link: TFhirDeviceComponentList;
begin
  result := TFhirDeviceComponentList(inherited Link);
end;

procedure TFhirDeviceComponentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceComponentList.SetItemByIndex(index: Integer; value: TFhirDeviceComponent);
begin
  assert(value is TFhirDeviceComponent);
  FhirDeviceComponents[index] := value;
end;

procedure TFhirDeviceComponentList.SetItemN(index: Integer; value: TFhirDeviceComponent);
begin
  assert(value is TFhirDeviceComponent);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceMetricCalibration }

constructor TFhirDeviceMetricCalibration.Create;
begin
  inherited;
end;

destructor TFhirDeviceMetricCalibration.Destroy;
begin
  FType_.free;
  FState.free;
  FTime.free;
  inherited;
end;

procedure TFhirDeviceMetricCalibration.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirDeviceMetricCalibration(oSource).FType_.Link;
  FState := TFhirDeviceMetricCalibration(oSource).FState.Link;
  timeElement := TFhirDeviceMetricCalibration(oSource).timeElement.Clone;
end;

procedure TFhirDeviceMetricCalibration.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'state') Then
     list.add(FState.Link);
  if (child_name = 'time') Then
     list.add(FTime.Link);
end;

procedure TFhirDeviceMetricCalibration.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'state', 'code', FState.Link));{1}
  oList.add(TFHIRProperty.create(self, 'time', 'instant', FTime.Link));{2}
end;

procedure TFhirDeviceMetricCalibration.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'state') then StateElement := propValue as TFHIREnum
  else if (propName = 'time') then TimeElement := propValue as TFhirInstant{5a}
  else inherited;
end;

function TFhirDeviceMetricCalibration.FhirType : string;
begin
  result := 'calibration';
end;

function TFhirDeviceMetricCalibration.Link : TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration(inherited Link);
end;

function TFhirDeviceMetricCalibration.Clone : TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration(inherited Clone);
end;

function TFhirDeviceMetricCalibration.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDeviceMetricCalibration;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDeviceMetricCalibration)) then
    result := false
  else
  begin
    o := TFhirDeviceMetricCalibration(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(stateElement, o.stateElement, true) and 
      compareDeep(timeElement, o.timeElement, true);
  end;
end;

function TFhirDeviceMetricCalibration.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDeviceMetricCalibration;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDeviceMetricCalibration)) then
    result := false
  else
  begin
    o := TFhirDeviceMetricCalibration(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(stateElement, o.stateElement, true) and 
      compareValues(timeElement, o.timeElement, true);
  end;
end;

{ TFhirDeviceMetricCalibration }

Procedure TFhirDeviceMetricCalibration.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirDeviceMetricCalibration.GetType_ST : TFhirMetricCalibrationType;
begin
  if FType_ = nil then
    result := TFhirMetricCalibrationType(0)
  else
    result := TFhirMetricCalibrationType(StringArrayIndexOfSensitive(CODES_TFhirMetricCalibrationType, FType_.value));
end;

Procedure TFhirDeviceMetricCalibration.SetType_ST(value : TFhirMetricCalibrationType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirMetricCalibrationType[value]);
end;

Procedure TFhirDeviceMetricCalibration.SetState(value : TFhirEnum);
begin
  FState.free;
  FState := value;
end;

Function TFhirDeviceMetricCalibration.GetStateST : TFhirMetricCalibrationState;
begin
  if FState = nil then
    result := TFhirMetricCalibrationState(0)
  else
    result := TFhirMetricCalibrationState(StringArrayIndexOfSensitive(CODES_TFhirMetricCalibrationState, FState.value));
end;

Procedure TFhirDeviceMetricCalibration.SetStateST(value : TFhirMetricCalibrationState);
begin
  if ord(value) = 0 then
    StateElement := nil
  else
    StateElement := TFhirEnum.create(CODES_TFhirMetricCalibrationState[value]);
end;

Procedure TFhirDeviceMetricCalibration.SetTime(value : TFhirInstant);
begin
  FTime.free;
  FTime := value;
end;

Function TFhirDeviceMetricCalibration.GetTimeST : TDateAndTime;
begin
  if FTime = nil then
    result := nil
  else
    result := FTime.value;
end;

Procedure TFhirDeviceMetricCalibration.SetTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FTime = nil then
      FTime := TFhirInstant.create;
    FTime.value := value
  end
  else if FTime <> nil then
    FTime.value := nil;
end;


{ TFhirDeviceMetricCalibrationListEnumerator }

Constructor TFhirDeviceMetricCalibrationListEnumerator.Create(list : TFhirDeviceMetricCalibrationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceMetricCalibrationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceMetricCalibrationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceMetricCalibrationListEnumerator.GetCurrent : TFhirDeviceMetricCalibration;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceMetricCalibrationList }
procedure TFhirDeviceMetricCalibrationList.AddItem(value: TFhirDeviceMetricCalibration);
begin
  assert(value.ClassName = 'TFhirDeviceMetricCalibration', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceMetricCalibration');
  add(value);
end;


function TFhirDeviceMetricCalibrationList.Append: TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceMetricCalibrationList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceMetricCalibrationList.GetEnumerator : TFhirDeviceMetricCalibrationListEnumerator;
begin
  result := TFhirDeviceMetricCalibrationListEnumerator.Create(self.link);
end;

function TFhirDeviceMetricCalibrationList.Clone: TFhirDeviceMetricCalibrationList;
begin
  result := TFhirDeviceMetricCalibrationList(inherited Clone);
end;

function TFhirDeviceMetricCalibrationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceMetricCalibrationList.GetItemN(index: Integer): TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration(ObjectByIndex[index]);
end;

function TFhirDeviceMetricCalibrationList.IndexOf(value: TFhirDeviceMetricCalibration): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceMetricCalibrationList.Insert(index: Integer): TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceMetricCalibrationList.InsertItem(index: Integer; value: TFhirDeviceMetricCalibration);
begin
  assert(value is TFhirDeviceMetricCalibration);
  Inherited Insert(index, value);
end;

function TFhirDeviceMetricCalibrationList.Item(index: Integer): TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration(ObjectByIndex[index]);
end;

function TFhirDeviceMetricCalibrationList.Link: TFhirDeviceMetricCalibrationList;
begin
  result := TFhirDeviceMetricCalibrationList(inherited Link);
end;

procedure TFhirDeviceMetricCalibrationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceMetricCalibrationList.SetItemByIndex(index: Integer; value: TFhirDeviceMetricCalibration);
begin
  assert(value is TFhirDeviceMetricCalibration);
  FhirDeviceMetricCalibrations[index] := value;
end;

procedure TFhirDeviceMetricCalibrationList.SetItemN(index: Integer; value: TFhirDeviceMetricCalibration);
begin
  assert(value is TFhirDeviceMetricCalibration);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceMetric }

constructor TFhirDeviceMetric.Create;
begin
  inherited;
end;

destructor TFhirDeviceMetric.Destroy;
begin
  FType_.free;
  FIdentifier.free;
  FUnit_.free;
  FSource.free;
  FParent.free;
  FOperationalStatus.free;
  FColor.free;
  FCategory.free;
  FMeasurementPeriod.free;
  FCalibrationList.Free;
  inherited;
end;

function TFhirDeviceMetric.GetResourceType : TFhirResourceType;
begin
  result := frtDeviceMetric;
end;

procedure TFhirDeviceMetric.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirDeviceMetric(oSource).type_.Clone;
  identifier := TFhirDeviceMetric(oSource).identifier.Clone;
  unit_ := TFhirDeviceMetric(oSource).unit_.Clone;
  source := TFhirDeviceMetric(oSource).source.Clone;
  parent := TFhirDeviceMetric(oSource).parent.Clone;
  FOperationalStatus := TFhirDeviceMetric(oSource).FOperationalStatus.Link;
  FColor := TFhirDeviceMetric(oSource).FColor.Link;
  FCategory := TFhirDeviceMetric(oSource).FCategory.Link;
  measurementPeriod := TFhirDeviceMetric(oSource).measurementPeriod.Clone;
  if (TFhirDeviceMetric(oSource).FCalibrationList = nil) then
  begin
    FCalibrationList.free;
    FCalibrationList := nil;
  end
  else
  begin
    FCalibrationList := TFhirDeviceMetricCalibrationList.Create;
    FCalibrationList.Assign(TFhirDeviceMetric(oSource).FCalibrationList);
  end;
end;

procedure TFhirDeviceMetric.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'unit') Then
     list.add(FUnit_.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'parent') Then
     list.add(FParent.Link);
  if (child_name = 'operationalStatus') Then
     list.add(FOperationalStatus.Link);
  if (child_name = 'color') Then
     list.add(FColor.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'measurementPeriod') Then
     list.add(FMeasurementPeriod.Link);
  if (child_name = 'calibration') Then
    list.addAll(FCalibrationList);
end;

procedure TFhirDeviceMetric.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'unit', 'CodeableConcept', FUnit_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Device)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'parent', 'Reference(DeviceComponent)', FParent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'operationalStatus', 'code', FOperationalStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'color', 'code', FColor.Link));{1}
  oList.add(TFHIRProperty.create(self, 'category', 'code', FCategory.Link));{1}
  oList.add(TFHIRProperty.create(self, 'measurementPeriod', 'Timing', FMeasurementPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'calibration', '', FCalibrationList.Link)){3};
end;

procedure TFhirDeviceMetric.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'unit') then Unit_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'source') then Source := propValue as TFhirReference{TFhirDevice}{4b}
  else if (propName = 'parent') then Parent := propValue as TFhirReference{TFhirDeviceComponent}{4b}
  else if (propName = 'operationalStatus') then OperationalStatusElement := propValue as TFHIREnum
  else if (propName = 'color') then ColorElement := propValue as TFHIREnum
  else if (propName = 'category') then CategoryElement := propValue as TFHIREnum
  else if (propName = 'measurementPeriod') then MeasurementPeriod := propValue as TFhirTiming{4b}
  else if (propName = 'calibration') then CalibrationList.add(propValue as TFhirDeviceMetricCalibration){2}
  else inherited;
end;

function TFhirDeviceMetric.FhirType : string;
begin
  result := 'DeviceMetric';
end;

function TFhirDeviceMetric.Link : TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric(inherited Link);
end;

function TFhirDeviceMetric.Clone : TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric(inherited Clone);
end;

{ TFhirDeviceMetric }

Procedure TFhirDeviceMetric.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirDeviceMetric.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirDeviceMetric.SetUnit_(value : TFhirCodeableConcept);
begin
  FUnit_.free;
  FUnit_ := value;
end;

Procedure TFhirDeviceMetric.SetSource(value : TFhirReference{TFhirDevice});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirDeviceMetric.SetParent(value : TFhirReference{TFhirDeviceComponent});
begin
  FParent.free;
  FParent := value;
end;

Procedure TFhirDeviceMetric.SetOperationalStatus(value : TFhirEnum);
begin
  FOperationalStatus.free;
  FOperationalStatus := value;
end;

Function TFhirDeviceMetric.GetOperationalStatusST : TFhirMetricOperationalStatus;
begin
  if FOperationalStatus = nil then
    result := TFhirMetricOperationalStatus(0)
  else
    result := TFhirMetricOperationalStatus(StringArrayIndexOfSensitive(CODES_TFhirMetricOperationalStatus, FOperationalStatus.value));
end;

Procedure TFhirDeviceMetric.SetOperationalStatusST(value : TFhirMetricOperationalStatus);
begin
  if ord(value) = 0 then
    OperationalStatusElement := nil
  else
    OperationalStatusElement := TFhirEnum.create(CODES_TFhirMetricOperationalStatus[value]);
end;

Procedure TFhirDeviceMetric.SetColor(value : TFhirEnum);
begin
  FColor.free;
  FColor := value;
end;

Function TFhirDeviceMetric.GetColorST : TFhirMetricColor;
begin
  if FColor = nil then
    result := TFhirMetricColor(0)
  else
    result := TFhirMetricColor(StringArrayIndexOfSensitive(CODES_TFhirMetricColor, FColor.value));
end;

Procedure TFhirDeviceMetric.SetColorST(value : TFhirMetricColor);
begin
  if ord(value) = 0 then
    ColorElement := nil
  else
    ColorElement := TFhirEnum.create(CODES_TFhirMetricColor[value]);
end;

Procedure TFhirDeviceMetric.SetCategory(value : TFhirEnum);
begin
  FCategory.free;
  FCategory := value;
end;

Function TFhirDeviceMetric.GetCategoryST : TFhirMetricCategory;
begin
  if FCategory = nil then
    result := TFhirMetricCategory(0)
  else
    result := TFhirMetricCategory(StringArrayIndexOfSensitive(CODES_TFhirMetricCategory, FCategory.value));
end;

Procedure TFhirDeviceMetric.SetCategoryST(value : TFhirMetricCategory);
begin
  if ord(value) = 0 then
    CategoryElement := nil
  else
    CategoryElement := TFhirEnum.create(CODES_TFhirMetricCategory[value]);
end;

Procedure TFhirDeviceMetric.SetMeasurementPeriod(value : TFhirTiming);
begin
  FMeasurementPeriod.free;
  FMeasurementPeriod := value;
end;

Function TFhirDeviceMetric.GetCalibrationList : TFhirDeviceMetricCalibrationList;
begin
  if FCalibrationList = nil then
    FCalibrationList := TFhirDeviceMetricCalibrationList.Create;
  result := FCalibrationList;
end;

Function TFhirDeviceMetric.GetHasCalibrationList : boolean;
begin
  result := (FCalibrationList <> nil) and (FCalibrationList.count > 0);
end;


{ TFhirDeviceMetricListEnumerator }

Constructor TFhirDeviceMetricListEnumerator.Create(list : TFhirDeviceMetricList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceMetricListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceMetricListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceMetricListEnumerator.GetCurrent : TFhirDeviceMetric;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceMetricList }
procedure TFhirDeviceMetricList.AddItem(value: TFhirDeviceMetric);
begin
  assert(value.ClassName = 'TFhirDeviceMetric', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceMetric');
  add(value);
end;


function TFhirDeviceMetricList.Append: TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceMetricList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceMetricList.GetEnumerator : TFhirDeviceMetricListEnumerator;
begin
  result := TFhirDeviceMetricListEnumerator.Create(self.link);
end;

function TFhirDeviceMetricList.Clone: TFhirDeviceMetricList;
begin
  result := TFhirDeviceMetricList(inherited Clone);
end;

function TFhirDeviceMetricList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceMetricList.GetItemN(index: Integer): TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric(ObjectByIndex[index]);
end;

function TFhirDeviceMetricList.IndexOf(value: TFhirDeviceMetric): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceMetricList.Insert(index: Integer): TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceMetricList.InsertItem(index: Integer; value: TFhirDeviceMetric);
begin
  assert(value is TFhirDeviceMetric);
  Inherited Insert(index, value);
end;

function TFhirDeviceMetricList.Item(index: Integer): TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric(ObjectByIndex[index]);
end;

function TFhirDeviceMetricList.Link: TFhirDeviceMetricList;
begin
  result := TFhirDeviceMetricList(inherited Link);
end;

procedure TFhirDeviceMetricList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceMetricList.SetItemByIndex(index: Integer; value: TFhirDeviceMetric);
begin
  assert(value is TFhirDeviceMetric);
  FhirDeviceMetrics[index] := value;
end;

procedure TFhirDeviceMetricList.SetItemN(index: Integer; value: TFhirDeviceMetric);
begin
  assert(value is TFhirDeviceMetric);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceUseRequest }

constructor TFhirDeviceUseRequest.Create;
begin
  inherited;
end;

destructor TFhirDeviceUseRequest.Destroy;
begin
  FBodySite.free;
  FStatus.free;
  FDevice.free;
  FEncounter.free;
  FIdentifierList.Free;
  FIndicationList.Free;
  FNotesList.Free;
  FPrnReasonList.Free;
  FOrderedOn.free;
  FRecordedOn.free;
  FSubject.free;
  FTiming.free;
  FPriority.free;
  inherited;
end;

function TFhirDeviceUseRequest.GetResourceType : TFhirResourceType;
begin
  result := frtDeviceUseRequest;
end;

procedure TFhirDeviceUseRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  bodySite := TFhirDeviceUseRequest(oSource).bodySite.Clone;
  FStatus := TFhirDeviceUseRequest(oSource).FStatus.Link;
  device := TFhirDeviceUseRequest(oSource).device.Clone;
  encounter := TFhirDeviceUseRequest(oSource).encounter.Clone;
  if (TFhirDeviceUseRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDeviceUseRequest(oSource).FIdentifierList);
  end;
  if (TFhirDeviceUseRequest(oSource).FIndicationList = nil) then
  begin
    FIndicationList.free;
    FIndicationList := nil;
  end
  else
  begin
    FIndicationList := TFhirCodeableConceptList.Create;
    FIndicationList.Assign(TFhirDeviceUseRequest(oSource).FIndicationList);
  end;
  if (TFhirDeviceUseRequest(oSource).FNotesList = nil) then
  begin
    FNotesList.free;
    FNotesList := nil;
  end
  else
  begin
    FNotesList := TFhirStringList.Create;
    FNotesList.Assign(TFhirDeviceUseRequest(oSource).FNotesList);
  end;
  if (TFhirDeviceUseRequest(oSource).FPrnReasonList = nil) then
  begin
    FPrnReasonList.free;
    FPrnReasonList := nil;
  end
  else
  begin
    FPrnReasonList := TFhirCodeableConceptList.Create;
    FPrnReasonList.Assign(TFhirDeviceUseRequest(oSource).FPrnReasonList);
  end;
  orderedOnElement := TFhirDeviceUseRequest(oSource).orderedOnElement.Clone;
  recordedOnElement := TFhirDeviceUseRequest(oSource).recordedOnElement.Clone;
  subject := TFhirDeviceUseRequest(oSource).subject.Clone;
  timing := TFhirDeviceUseRequest(oSource).timing.Clone;
  FPriority := TFhirDeviceUseRequest(oSource).FPriority.Link;
end;

procedure TFhirDeviceUseRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'bodySite[x]') Then
     list.add(FBodySite.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'device') Then
     list.add(FDevice.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'indication') Then
    list.addAll(FIndicationList);
  if (child_name = 'notes') Then
    list.addAll(FNotesList);
  if (child_name = 'prnReason') Then
    list.addAll(FPrnReasonList);
  if (child_name = 'orderedOn') Then
     list.add(FOrderedOn.Link);
  if (child_name = 'recordedOn') Then
     list.add(FRecordedOn.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'timing[x]') Then
     list.add(FTiming.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
end;

procedure TFhirDeviceUseRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'bodySite[x]', 'CodeableConcept|Reference(BodySite)', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'device', 'Reference(Device)', FDevice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'indication', 'CodeableConcept', FIndicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'notes', 'string', FNotesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'prnReason', 'CodeableConcept', FPrnReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'orderedOn', 'dateTime', FOrderedOn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recordedOn', 'dateTime', FRecordedOn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timing[x]', 'Timing|Period|dateTime', FTiming.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'code', FPriority.Link));{1}
end;

procedure TFhirDeviceUseRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName.startsWith('bodySite')) then BodySite := propValue as TFhirType{4}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'device') then Device := propValue as TFhirReference{TFhirDevice}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'indication') then IndicationList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'notes') then NotesList.add(propValue as TFhirString){2}
  else if (propName = 'prnReason') then PrnReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'orderedOn') then OrderedOnElement := propValue as TFhirDateTime{5a}
  else if (propName = 'recordedOn') then RecordedOnElement := propValue as TFhirDateTime{5a}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName.startsWith('timing')) then Timing := propValue as TFhirType{4}
  else if (propName = 'priority') then PriorityElement := propValue as TFHIREnum
  else inherited;
end;

function TFhirDeviceUseRequest.FhirType : string;
begin
  result := 'DeviceUseRequest';
end;

function TFhirDeviceUseRequest.Link : TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest(inherited Link);
end;

function TFhirDeviceUseRequest.Clone : TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest(inherited Clone);
end;

{ TFhirDeviceUseRequest }

Procedure TFhirDeviceUseRequest.SetBodySite(value : TFhirType);
begin
  FBodySite.free;
  FBodySite := value;
end;

Procedure TFhirDeviceUseRequest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDeviceUseRequest.GetStatusST : TFhirDeviceUseRequestStatus;
begin
  if FStatus = nil then
    result := TFhirDeviceUseRequestStatus(0)
  else
    result := TFhirDeviceUseRequestStatus(StringArrayIndexOfSensitive(CODES_TFhirDeviceUseRequestStatus, FStatus.value));
end;

Procedure TFhirDeviceUseRequest.SetStatusST(value : TFhirDeviceUseRequestStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDeviceUseRequestStatus[value]);
end;

Procedure TFhirDeviceUseRequest.SetDevice(value : TFhirReference{TFhirDevice});
begin
  FDevice.free;
  FDevice := value;
end;

Procedure TFhirDeviceUseRequest.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Function TFhirDeviceUseRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDeviceUseRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirDeviceUseRequest.GetIndicationList : TFhirCodeableConceptList;
begin
  if FIndicationList = nil then
    FIndicationList := TFhirCodeableConceptList.Create;
  result := FIndicationList;
end;

Function TFhirDeviceUseRequest.GetHasIndicationList : boolean;
begin
  result := (FIndicationList <> nil) and (FIndicationList.count > 0);
end;

Function TFhirDeviceUseRequest.GetNotesList : TFhirStringList;
begin
  if FNotesList = nil then
    FNotesList := TFhirStringList.Create;
  result := FNotesList;
end;

Function TFhirDeviceUseRequest.GetHasNotesList : boolean;
begin
  result := (FNotesList <> nil) and (FNotesList.count > 0);
end;

Function TFhirDeviceUseRequest.GetPrnReasonList : TFhirCodeableConceptList;
begin
  if FPrnReasonList = nil then
    FPrnReasonList := TFhirCodeableConceptList.Create;
  result := FPrnReasonList;
end;

Function TFhirDeviceUseRequest.GetHasPrnReasonList : boolean;
begin
  result := (FPrnReasonList <> nil) and (FPrnReasonList.count > 0);
end;

Procedure TFhirDeviceUseRequest.SetOrderedOn(value : TFhirDateTime);
begin
  FOrderedOn.free;
  FOrderedOn := value;
end;

Function TFhirDeviceUseRequest.GetOrderedOnST : TDateAndTime;
begin
  if FOrderedOn = nil then
    result := nil
  else
    result := FOrderedOn.value;
end;

Procedure TFhirDeviceUseRequest.SetOrderedOnST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FOrderedOn = nil then
      FOrderedOn := TFhirDateTime.create;
    FOrderedOn.value := value
  end
  else if FOrderedOn <> nil then
    FOrderedOn.value := nil;
end;

Procedure TFhirDeviceUseRequest.SetRecordedOn(value : TFhirDateTime);
begin
  FRecordedOn.free;
  FRecordedOn := value;
end;

Function TFhirDeviceUseRequest.GetRecordedOnST : TDateAndTime;
begin
  if FRecordedOn = nil then
    result := nil
  else
    result := FRecordedOn.value;
end;

Procedure TFhirDeviceUseRequest.SetRecordedOnST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FRecordedOn = nil then
      FRecordedOn := TFhirDateTime.create;
    FRecordedOn.value := value
  end
  else if FRecordedOn <> nil then
    FRecordedOn.value := nil;
end;

Procedure TFhirDeviceUseRequest.SetSubject(value : TFhirReference{TFhirPatient});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirDeviceUseRequest.SetTiming(value : TFhirType);
begin
  FTiming.free;
  FTiming := value;
end;

Procedure TFhirDeviceUseRequest.SetPriority(value : TFhirEnum);
begin
  FPriority.free;
  FPriority := value;
end;

Function TFhirDeviceUseRequest.GetPriorityST : TFhirDeviceUseRequestPriority;
begin
  if FPriority = nil then
    result := TFhirDeviceUseRequestPriority(0)
  else
    result := TFhirDeviceUseRequestPriority(StringArrayIndexOfSensitive(CODES_TFhirDeviceUseRequestPriority, FPriority.value));
end;

Procedure TFhirDeviceUseRequest.SetPriorityST(value : TFhirDeviceUseRequestPriority);
begin
  if ord(value) = 0 then
    PriorityElement := nil
  else
    PriorityElement := TFhirEnum.create(CODES_TFhirDeviceUseRequestPriority[value]);
end;


{ TFhirDeviceUseRequestListEnumerator }

Constructor TFhirDeviceUseRequestListEnumerator.Create(list : TFhirDeviceUseRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceUseRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceUseRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceUseRequestListEnumerator.GetCurrent : TFhirDeviceUseRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceUseRequestList }
procedure TFhirDeviceUseRequestList.AddItem(value: TFhirDeviceUseRequest);
begin
  assert(value.ClassName = 'TFhirDeviceUseRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceUseRequest');
  add(value);
end;


function TFhirDeviceUseRequestList.Append: TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceUseRequestList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceUseRequestList.GetEnumerator : TFhirDeviceUseRequestListEnumerator;
begin
  result := TFhirDeviceUseRequestListEnumerator.Create(self.link);
end;

function TFhirDeviceUseRequestList.Clone: TFhirDeviceUseRequestList;
begin
  result := TFhirDeviceUseRequestList(inherited Clone);
end;

function TFhirDeviceUseRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceUseRequestList.GetItemN(index: Integer): TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest(ObjectByIndex[index]);
end;

function TFhirDeviceUseRequestList.IndexOf(value: TFhirDeviceUseRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceUseRequestList.Insert(index: Integer): TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceUseRequestList.InsertItem(index: Integer; value: TFhirDeviceUseRequest);
begin
  assert(value is TFhirDeviceUseRequest);
  Inherited Insert(index, value);
end;

function TFhirDeviceUseRequestList.Item(index: Integer): TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest(ObjectByIndex[index]);
end;

function TFhirDeviceUseRequestList.Link: TFhirDeviceUseRequestList;
begin
  result := TFhirDeviceUseRequestList(inherited Link);
end;

procedure TFhirDeviceUseRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceUseRequestList.SetItemByIndex(index: Integer; value: TFhirDeviceUseRequest);
begin
  assert(value is TFhirDeviceUseRequest);
  FhirDeviceUseRequests[index] := value;
end;

procedure TFhirDeviceUseRequestList.SetItemN(index: Integer; value: TFhirDeviceUseRequest);
begin
  assert(value is TFhirDeviceUseRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirDeviceUseStatement }

constructor TFhirDeviceUseStatement.Create;
begin
  inherited;
end;

destructor TFhirDeviceUseStatement.Destroy;
begin
  FBodySite.free;
  FWhenUsed.free;
  FDevice.free;
  FIdentifierList.Free;
  FIndicationList.Free;
  FNotesList.Free;
  FRecordedOn.free;
  FSubject.free;
  FTiming.free;
  inherited;
end;

function TFhirDeviceUseStatement.GetResourceType : TFhirResourceType;
begin
  result := frtDeviceUseStatement;
end;

procedure TFhirDeviceUseStatement.Assign(oSource : TAdvObject);
begin
  inherited;
  bodySite := TFhirDeviceUseStatement(oSource).bodySite.Clone;
  whenUsed := TFhirDeviceUseStatement(oSource).whenUsed.Clone;
  device := TFhirDeviceUseStatement(oSource).device.Clone;
  if (TFhirDeviceUseStatement(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDeviceUseStatement(oSource).FIdentifierList);
  end;
  if (TFhirDeviceUseStatement(oSource).FIndicationList = nil) then
  begin
    FIndicationList.free;
    FIndicationList := nil;
  end
  else
  begin
    FIndicationList := TFhirCodeableConceptList.Create;
    FIndicationList.Assign(TFhirDeviceUseStatement(oSource).FIndicationList);
  end;
  if (TFhirDeviceUseStatement(oSource).FNotesList = nil) then
  begin
    FNotesList.free;
    FNotesList := nil;
  end
  else
  begin
    FNotesList := TFhirStringList.Create;
    FNotesList.Assign(TFhirDeviceUseStatement(oSource).FNotesList);
  end;
  recordedOnElement := TFhirDeviceUseStatement(oSource).recordedOnElement.Clone;
  subject := TFhirDeviceUseStatement(oSource).subject.Clone;
  timing := TFhirDeviceUseStatement(oSource).timing.Clone;
end;

procedure TFhirDeviceUseStatement.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'bodySite[x]') Then
     list.add(FBodySite.Link);
  if (child_name = 'whenUsed') Then
     list.add(FWhenUsed.Link);
  if (child_name = 'device') Then
     list.add(FDevice.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'indication') Then
    list.addAll(FIndicationList);
  if (child_name = 'notes') Then
    list.addAll(FNotesList);
  if (child_name = 'recordedOn') Then
     list.add(FRecordedOn.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'timing[x]') Then
     list.add(FTiming.Link);
end;

procedure TFhirDeviceUseStatement.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'bodySite[x]', 'CodeableConcept|Reference(BodySite)', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'whenUsed', 'Period', FWhenUsed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'device', 'Reference(Device)', FDevice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'indication', 'CodeableConcept', FIndicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'notes', 'string', FNotesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'recordedOn', 'dateTime', FRecordedOn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timing[x]', 'Timing|Period|dateTime', FTiming.Link));{2}
end;

procedure TFhirDeviceUseStatement.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName.startsWith('bodySite')) then BodySite := propValue as TFhirType{4}
  else if (propName = 'whenUsed') then WhenUsed := propValue as TFhirPeriod{4b}
  else if (propName = 'device') then Device := propValue as TFhirReference{TFhirDevice}{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'indication') then IndicationList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'notes') then NotesList.add(propValue as TFhirString){2}
  else if (propName = 'recordedOn') then RecordedOnElement := propValue as TFhirDateTime{5a}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName.startsWith('timing')) then Timing := propValue as TFhirType{4}
  else inherited;
end;

function TFhirDeviceUseStatement.FhirType : string;
begin
  result := 'DeviceUseStatement';
end;

function TFhirDeviceUseStatement.Link : TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement(inherited Link);
end;

function TFhirDeviceUseStatement.Clone : TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement(inherited Clone);
end;

{ TFhirDeviceUseStatement }

Procedure TFhirDeviceUseStatement.SetBodySite(value : TFhirType);
begin
  FBodySite.free;
  FBodySite := value;
end;

Procedure TFhirDeviceUseStatement.SetWhenUsed(value : TFhirPeriod);
begin
  FWhenUsed.free;
  FWhenUsed := value;
end;

Procedure TFhirDeviceUseStatement.SetDevice(value : TFhirReference{TFhirDevice});
begin
  FDevice.free;
  FDevice := value;
end;

Function TFhirDeviceUseStatement.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDeviceUseStatement.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirDeviceUseStatement.GetIndicationList : TFhirCodeableConceptList;
begin
  if FIndicationList = nil then
    FIndicationList := TFhirCodeableConceptList.Create;
  result := FIndicationList;
end;

Function TFhirDeviceUseStatement.GetHasIndicationList : boolean;
begin
  result := (FIndicationList <> nil) and (FIndicationList.count > 0);
end;

Function TFhirDeviceUseStatement.GetNotesList : TFhirStringList;
begin
  if FNotesList = nil then
    FNotesList := TFhirStringList.Create;
  result := FNotesList;
end;

Function TFhirDeviceUseStatement.GetHasNotesList : boolean;
begin
  result := (FNotesList <> nil) and (FNotesList.count > 0);
end;

Procedure TFhirDeviceUseStatement.SetRecordedOn(value : TFhirDateTime);
begin
  FRecordedOn.free;
  FRecordedOn := value;
end;

Function TFhirDeviceUseStatement.GetRecordedOnST : TDateAndTime;
begin
  if FRecordedOn = nil then
    result := nil
  else
    result := FRecordedOn.value;
end;

Procedure TFhirDeviceUseStatement.SetRecordedOnST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FRecordedOn = nil then
      FRecordedOn := TFhirDateTime.create;
    FRecordedOn.value := value
  end
  else if FRecordedOn <> nil then
    FRecordedOn.value := nil;
end;

Procedure TFhirDeviceUseStatement.SetSubject(value : TFhirReference{TFhirPatient});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirDeviceUseStatement.SetTiming(value : TFhirType);
begin
  FTiming.free;
  FTiming := value;
end;


{ TFhirDeviceUseStatementListEnumerator }

Constructor TFhirDeviceUseStatementListEnumerator.Create(list : TFhirDeviceUseStatementList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDeviceUseStatementListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDeviceUseStatementListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDeviceUseStatementListEnumerator.GetCurrent : TFhirDeviceUseStatement;
begin
  Result := FList[FIndex];
end;


{ TFhirDeviceUseStatementList }
procedure TFhirDeviceUseStatementList.AddItem(value: TFhirDeviceUseStatement);
begin
  assert(value.ClassName = 'TFhirDeviceUseStatement', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDeviceUseStatement');
  add(value);
end;


function TFhirDeviceUseStatementList.Append: TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceUseStatementList.ClearItems;
begin
  Clear;
end;

function TFhirDeviceUseStatementList.GetEnumerator : TFhirDeviceUseStatementListEnumerator;
begin
  result := TFhirDeviceUseStatementListEnumerator.Create(self.link);
end;

function TFhirDeviceUseStatementList.Clone: TFhirDeviceUseStatementList;
begin
  result := TFhirDeviceUseStatementList(inherited Clone);
end;

function TFhirDeviceUseStatementList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDeviceUseStatementList.GetItemN(index: Integer): TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement(ObjectByIndex[index]);
end;

function TFhirDeviceUseStatementList.IndexOf(value: TFhirDeviceUseStatement): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDeviceUseStatementList.Insert(index: Integer): TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDeviceUseStatementList.InsertItem(index: Integer; value: TFhirDeviceUseStatement);
begin
  assert(value is TFhirDeviceUseStatement);
  Inherited Insert(index, value);
end;

function TFhirDeviceUseStatementList.Item(index: Integer): TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement(ObjectByIndex[index]);
end;

function TFhirDeviceUseStatementList.Link: TFhirDeviceUseStatementList;
begin
  result := TFhirDeviceUseStatementList(inherited Link);
end;

procedure TFhirDeviceUseStatementList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDeviceUseStatementList.SetItemByIndex(index: Integer; value: TFhirDeviceUseStatement);
begin
  assert(value is TFhirDeviceUseStatement);
  FhirDeviceUseStatements[index] := value;
end;

procedure TFhirDeviceUseStatementList.SetItemN(index: Integer; value: TFhirDeviceUseStatement);
begin
  assert(value is TFhirDeviceUseStatement);
  ObjectByIndex[index] := value;
end;

{ TFhirDiagnosticOrderEvent }

constructor TFhirDiagnosticOrderEvent.Create;
begin
  inherited;
end;

destructor TFhirDiagnosticOrderEvent.Destroy;
begin
  FStatus.free;
  FDescription.free;
  FDateTime.free;
  FActor.free;
  inherited;
end;

procedure TFhirDiagnosticOrderEvent.Assign(oSource : TAdvObject);
begin
  inherited;
  FStatus := TFhirDiagnosticOrderEvent(oSource).FStatus.Link;
  description := TFhirDiagnosticOrderEvent(oSource).description.Clone;
  dateTimeElement := TFhirDiagnosticOrderEvent(oSource).dateTimeElement.Clone;
  actor := TFhirDiagnosticOrderEvent(oSource).actor.Clone;
end;

procedure TFhirDiagnosticOrderEvent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'dateTime') Then
     list.add(FDateTime.Link);
  if (child_name = 'actor') Then
     list.add(FActor.Link);
end;

procedure TFhirDiagnosticOrderEvent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'description', 'CodeableConcept', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateTime', 'dateTime', FDateTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Practitioner|Device)', FActor.Link));{2}
end;

procedure TFhirDiagnosticOrderEvent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'description') then Description := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'dateTime') then DateTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirDiagnosticOrderEvent.FhirType : string;
begin
  result := 'event';
end;

function TFhirDiagnosticOrderEvent.Link : TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent(inherited Link);
end;

function TFhirDiagnosticOrderEvent.Clone : TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent(inherited Clone);
end;

function TFhirDiagnosticOrderEvent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticOrderEvent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDiagnosticOrderEvent)) then
    result := false
  else
  begin
    o := TFhirDiagnosticOrderEvent(other);
    result := compareDeep(statusElement, o.statusElement, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(dateTimeElement, o.dateTimeElement, true) and compareDeep(actorElement, o.actorElement, true);
  end;
end;

function TFhirDiagnosticOrderEvent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticOrderEvent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDiagnosticOrderEvent)) then
    result := false
  else
  begin
    o := TFhirDiagnosticOrderEvent(other);
    result := compareValues(statusElement, o.statusElement, true) and compareValues(dateTimeElement, o.dateTimeElement, true);
  end;
end;

{ TFhirDiagnosticOrderEvent }

Procedure TFhirDiagnosticOrderEvent.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDiagnosticOrderEvent.GetStatusST : TFhirDiagnosticOrderStatus;
begin
  if FStatus = nil then
    result := TFhirDiagnosticOrderStatus(0)
  else
    result := TFhirDiagnosticOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirDiagnosticOrderStatus, FStatus.value));
end;

Procedure TFhirDiagnosticOrderEvent.SetStatusST(value : TFhirDiagnosticOrderStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDiagnosticOrderStatus[value]);
end;

Procedure TFhirDiagnosticOrderEvent.SetDescription(value : TFhirCodeableConcept);
begin
  FDescription.free;
  FDescription := value;
end;

Procedure TFhirDiagnosticOrderEvent.SetDateTime(value : TFhirDateTime);
begin
  FDateTime.free;
  FDateTime := value;
end;

Function TFhirDiagnosticOrderEvent.GetDateTimeST : TDateAndTime;
begin
  if FDateTime = nil then
    result := nil
  else
    result := FDateTime.value;
end;

Procedure TFhirDiagnosticOrderEvent.SetDateTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateTime = nil then
      FDateTime := TFhirDateTime.create;
    FDateTime.value := value
  end
  else if FDateTime <> nil then
    FDateTime.value := nil;
end;

Procedure TFhirDiagnosticOrderEvent.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;


{ TFhirDiagnosticOrderEventListEnumerator }

Constructor TFhirDiagnosticOrderEventListEnumerator.Create(list : TFhirDiagnosticOrderEventList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDiagnosticOrderEventListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDiagnosticOrderEventListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDiagnosticOrderEventListEnumerator.GetCurrent : TFhirDiagnosticOrderEvent;
begin
  Result := FList[FIndex];
end;


{ TFhirDiagnosticOrderEventList }
procedure TFhirDiagnosticOrderEventList.AddItem(value: TFhirDiagnosticOrderEvent);
begin
  assert(value.ClassName = 'TFhirDiagnosticOrderEvent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDiagnosticOrderEvent');
  add(value);
end;


function TFhirDiagnosticOrderEventList.Append: TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderEventList.ClearItems;
begin
  Clear;
end;

function TFhirDiagnosticOrderEventList.GetEnumerator : TFhirDiagnosticOrderEventListEnumerator;
begin
  result := TFhirDiagnosticOrderEventListEnumerator.Create(self.link);
end;

function TFhirDiagnosticOrderEventList.Clone: TFhirDiagnosticOrderEventList;
begin
  result := TFhirDiagnosticOrderEventList(inherited Clone);
end;

function TFhirDiagnosticOrderEventList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDiagnosticOrderEventList.GetItemN(index: Integer): TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderEventList.IndexOf(value: TFhirDiagnosticOrderEvent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDiagnosticOrderEventList.Insert(index: Integer): TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderEventList.InsertItem(index: Integer; value: TFhirDiagnosticOrderEvent);
begin
  assert(value is TFhirDiagnosticOrderEvent);
  Inherited Insert(index, value);
end;

function TFhirDiagnosticOrderEventList.Item(index: Integer): TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderEventList.Link: TFhirDiagnosticOrderEventList;
begin
  result := TFhirDiagnosticOrderEventList(inherited Link);
end;

procedure TFhirDiagnosticOrderEventList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDiagnosticOrderEventList.SetItemByIndex(index: Integer; value: TFhirDiagnosticOrderEvent);
begin
  assert(value is TFhirDiagnosticOrderEvent);
  FhirDiagnosticOrderEvents[index] := value;
end;

procedure TFhirDiagnosticOrderEventList.SetItemN(index: Integer; value: TFhirDiagnosticOrderEvent);
begin
  assert(value is TFhirDiagnosticOrderEvent);
  ObjectByIndex[index] := value;
end;

{ TFhirDiagnosticOrderItem }

constructor TFhirDiagnosticOrderItem.Create;
begin
  inherited;
end;

destructor TFhirDiagnosticOrderItem.Destroy;
begin
  FCode.free;
  FSpecimenList.Free;
  FBodySite.free;
  FStatus.free;
  FEventList.Free;
  inherited;
end;

procedure TFhirDiagnosticOrderItem.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirDiagnosticOrderItem(oSource).code.Clone;
  if (TFhirDiagnosticOrderItem(oSource).FSpecimenList = nil) then
  begin
    FSpecimenList.free;
    FSpecimenList := nil;
  end
  else
  begin
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
    FSpecimenList.Assign(TFhirDiagnosticOrderItem(oSource).FSpecimenList);
  end;
  bodySite := TFhirDiagnosticOrderItem(oSource).bodySite.Clone;
  FStatus := TFhirDiagnosticOrderItem(oSource).FStatus.Link;
  if (TFhirDiagnosticOrderItem(oSource).FEventList = nil) then
  begin
    FEventList.free;
    FEventList := nil;
  end
  else
  begin
    FEventList := TFhirDiagnosticOrderEventList.Create;
    FEventList.Assign(TFhirDiagnosticOrderItem(oSource).FEventList);
  end;
end;

procedure TFhirDiagnosticOrderItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'specimen') Then
    list.addAll(FSpecimenList);
  if (child_name = 'bodySite') Then
     list.add(FBodySite.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'event') Then
    list.addAll(FEventList);
end;

procedure TFhirDiagnosticOrderItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specimen', 'Reference(Specimen)', FSpecimenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'event', '@DiagnosticOrder.event', FEventList.Link)){3};
end;

procedure TFhirDiagnosticOrderItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'specimen') then SpecimenList.add(propValue as TFhirReference{TFhirSpecimen}){2}
  else if (propName = 'bodySite') then BodySite := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'event') then EventList.add(propValue as TFhirDiagnosticOrderEvent){2}
  else inherited;
end;

function TFhirDiagnosticOrderItem.FhirType : string;
begin
  result := 'item';
end;

function TFhirDiagnosticOrderItem.Link : TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem(inherited Link);
end;

function TFhirDiagnosticOrderItem.Clone : TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem(inherited Clone);
end;

function TFhirDiagnosticOrderItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticOrderItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDiagnosticOrderItem)) then
    result := false
  else
  begin
    o := TFhirDiagnosticOrderItem(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(specimenList, o.specimenList, true) and 
      compareDeep(bodySiteElement, o.bodySiteElement, true) and compareDeep(statusElement, o.statusElement, true) and 
      compareDeep(eventList, o.eventList, true);
  end;
end;

function TFhirDiagnosticOrderItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticOrderItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDiagnosticOrderItem)) then
    result := false
  else
  begin
    o := TFhirDiagnosticOrderItem(other);
    result := compareValues(statusElement, o.statusElement, true);
  end;
end;

{ TFhirDiagnosticOrderItem }

Procedure TFhirDiagnosticOrderItem.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirDiagnosticOrderItem.GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
begin
  if FSpecimenList = nil then
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
  result := FSpecimenList;
end;

Function TFhirDiagnosticOrderItem.GetHasSpecimenList : boolean;
begin
  result := (FSpecimenList <> nil) and (FSpecimenList.count > 0);
end;

Procedure TFhirDiagnosticOrderItem.SetBodySite(value : TFhirCodeableConcept);
begin
  FBodySite.free;
  FBodySite := value;
end;

Procedure TFhirDiagnosticOrderItem.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDiagnosticOrderItem.GetStatusST : TFhirDiagnosticOrderStatus;
begin
  if FStatus = nil then
    result := TFhirDiagnosticOrderStatus(0)
  else
    result := TFhirDiagnosticOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirDiagnosticOrderStatus, FStatus.value));
end;

Procedure TFhirDiagnosticOrderItem.SetStatusST(value : TFhirDiagnosticOrderStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDiagnosticOrderStatus[value]);
end;

Function TFhirDiagnosticOrderItem.GetEventList : TFhirDiagnosticOrderEventList;
begin
  if FEventList = nil then
    FEventList := TFhirDiagnosticOrderEventList.Create;
  result := FEventList;
end;

Function TFhirDiagnosticOrderItem.GetHasEventList : boolean;
begin
  result := (FEventList <> nil) and (FEventList.count > 0);
end;


{ TFhirDiagnosticOrderItemListEnumerator }

Constructor TFhirDiagnosticOrderItemListEnumerator.Create(list : TFhirDiagnosticOrderItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDiagnosticOrderItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDiagnosticOrderItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDiagnosticOrderItemListEnumerator.GetCurrent : TFhirDiagnosticOrderItem;
begin
  Result := FList[FIndex];
end;


{ TFhirDiagnosticOrderItemList }
procedure TFhirDiagnosticOrderItemList.AddItem(value: TFhirDiagnosticOrderItem);
begin
  assert(value.ClassName = 'TFhirDiagnosticOrderItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDiagnosticOrderItem');
  add(value);
end;


function TFhirDiagnosticOrderItemList.Append: TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderItemList.ClearItems;
begin
  Clear;
end;

function TFhirDiagnosticOrderItemList.GetEnumerator : TFhirDiagnosticOrderItemListEnumerator;
begin
  result := TFhirDiagnosticOrderItemListEnumerator.Create(self.link);
end;

function TFhirDiagnosticOrderItemList.Clone: TFhirDiagnosticOrderItemList;
begin
  result := TFhirDiagnosticOrderItemList(inherited Clone);
end;

function TFhirDiagnosticOrderItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDiagnosticOrderItemList.GetItemN(index: Integer): TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderItemList.IndexOf(value: TFhirDiagnosticOrderItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDiagnosticOrderItemList.Insert(index: Integer): TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderItemList.InsertItem(index: Integer; value: TFhirDiagnosticOrderItem);
begin
  assert(value is TFhirDiagnosticOrderItem);
  Inherited Insert(index, value);
end;

function TFhirDiagnosticOrderItemList.Item(index: Integer): TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderItemList.Link: TFhirDiagnosticOrderItemList;
begin
  result := TFhirDiagnosticOrderItemList(inherited Link);
end;

procedure TFhirDiagnosticOrderItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDiagnosticOrderItemList.SetItemByIndex(index: Integer; value: TFhirDiagnosticOrderItem);
begin
  assert(value is TFhirDiagnosticOrderItem);
  FhirDiagnosticOrderItems[index] := value;
end;

procedure TFhirDiagnosticOrderItemList.SetItemN(index: Integer; value: TFhirDiagnosticOrderItem);
begin
  assert(value is TFhirDiagnosticOrderItem);
  ObjectByIndex[index] := value;
end;

{ TFhirDiagnosticOrder }

constructor TFhirDiagnosticOrder.Create;
begin
  inherited;
end;

destructor TFhirDiagnosticOrder.Destroy;
begin
  FSubject.free;
  FOrderer.free;
  FIdentifierList.Free;
  FEncounter.free;
  FReasonList.Free;
  FSupportingInformationList.Free;
  FSpecimenList.Free;
  FStatus.free;
  FPriority.free;
  FEventList.Free;
  FItemList.Free;
  FNoteList.Free;
  inherited;
end;

function TFhirDiagnosticOrder.GetResourceType : TFhirResourceType;
begin
  result := frtDiagnosticOrder;
end;

procedure TFhirDiagnosticOrder.Assign(oSource : TAdvObject);
begin
  inherited;
  subject := TFhirDiagnosticOrder(oSource).subject.Clone;
  orderer := TFhirDiagnosticOrder(oSource).orderer.Clone;
  if (TFhirDiagnosticOrder(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDiagnosticOrder(oSource).FIdentifierList);
  end;
  encounter := TFhirDiagnosticOrder(oSource).encounter.Clone;
  if (TFhirDiagnosticOrder(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirDiagnosticOrder(oSource).FReasonList);
  end;
  if (TFhirDiagnosticOrder(oSource).FSupportingInformationList = nil) then
  begin
    FSupportingInformationList.free;
    FSupportingInformationList := nil;
  end
  else
  begin
    FSupportingInformationList := TFhirReferenceList{Resource}.Create;
    FSupportingInformationList.Assign(TFhirDiagnosticOrder(oSource).FSupportingInformationList);
  end;
  if (TFhirDiagnosticOrder(oSource).FSpecimenList = nil) then
  begin
    FSpecimenList.free;
    FSpecimenList := nil;
  end
  else
  begin
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
    FSpecimenList.Assign(TFhirDiagnosticOrder(oSource).FSpecimenList);
  end;
  FStatus := TFhirDiagnosticOrder(oSource).FStatus.Link;
  FPriority := TFhirDiagnosticOrder(oSource).FPriority.Link;
  if (TFhirDiagnosticOrder(oSource).FEventList = nil) then
  begin
    FEventList.free;
    FEventList := nil;
  end
  else
  begin
    FEventList := TFhirDiagnosticOrderEventList.Create;
    FEventList.Assign(TFhirDiagnosticOrder(oSource).FEventList);
  end;
  if (TFhirDiagnosticOrder(oSource).FItemList = nil) then
  begin
    FItemList.free;
    FItemList := nil;
  end
  else
  begin
    FItemList := TFhirDiagnosticOrderItemList.Create;
    FItemList.Assign(TFhirDiagnosticOrder(oSource).FItemList);
  end;
  if (TFhirDiagnosticOrder(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirAnnotationList.Create;
    FNoteList.Assign(TFhirDiagnosticOrder(oSource).FNoteList);
  end;
end;

procedure TFhirDiagnosticOrder.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'orderer') Then
     list.add(FOrderer.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'supportingInformation') Then
    list.addAll(FSupportingInformationList);
  if (child_name = 'specimen') Then
    list.addAll(FSpecimenList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'event') Then
    list.addAll(FEventList);
  if (child_name = 'item') Then
    list.addAll(FItemList);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
end;

procedure TFhirDiagnosticOrder.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Location|Device)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderer', 'Reference(Practitioner)', FOrderer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'supportingInformation', 'Reference(Observation|Condition|DocumentReference)', FSupportingInformationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'specimen', 'Reference(Specimen)', FSpecimenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'priority', 'code', FPriority.Link));{1}
  oList.add(TFHIRProperty.create(self, 'event', '', FEventList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'item', '', FItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNoteList.Link)){3};
end;

procedure TFhirDiagnosticOrder.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'orderer') then Orderer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'supportingInformation') then SupportingInformationList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'specimen') then SpecimenList.add(propValue as TFhirReference{TFhirSpecimen}){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'priority') then PriorityElement := propValue as TFHIREnum
  else if (propName = 'event') then EventList.add(propValue as TFhirDiagnosticOrderEvent){2}
  else if (propName = 'item') then ItemList.add(propValue as TFhirDiagnosticOrderItem){2}
  else if (propName = 'note') then NoteList.add(propValue as TFhirAnnotation){2}
  else inherited;
end;

function TFhirDiagnosticOrder.FhirType : string;
begin
  result := 'DiagnosticOrder';
end;

function TFhirDiagnosticOrder.Link : TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder(inherited Link);
end;

function TFhirDiagnosticOrder.Clone : TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder(inherited Clone);
end;

{ TFhirDiagnosticOrder }

Procedure TFhirDiagnosticOrder.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirDiagnosticOrder.SetOrderer(value : TFhirReference{TFhirPractitioner});
begin
  FOrderer.free;
  FOrderer := value;
end;

Function TFhirDiagnosticOrder.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDiagnosticOrder.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDiagnosticOrder.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Function TFhirDiagnosticOrder.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirDiagnosticOrder.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Function TFhirDiagnosticOrder.GetSupportingInformationList : TFhirReferenceList{Resource};
begin
  if FSupportingInformationList = nil then
    FSupportingInformationList := TFhirReferenceList{Resource}.Create;
  result := FSupportingInformationList;
end;

Function TFhirDiagnosticOrder.GetHasSupportingInformationList : boolean;
begin
  result := (FSupportingInformationList <> nil) and (FSupportingInformationList.count > 0);
end;

Function TFhirDiagnosticOrder.GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
begin
  if FSpecimenList = nil then
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
  result := FSpecimenList;
end;

Function TFhirDiagnosticOrder.GetHasSpecimenList : boolean;
begin
  result := (FSpecimenList <> nil) and (FSpecimenList.count > 0);
end;

Procedure TFhirDiagnosticOrder.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDiagnosticOrder.GetStatusST : TFhirDiagnosticOrderStatus;
begin
  if FStatus = nil then
    result := TFhirDiagnosticOrderStatus(0)
  else
    result := TFhirDiagnosticOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirDiagnosticOrderStatus, FStatus.value));
end;

Procedure TFhirDiagnosticOrder.SetStatusST(value : TFhirDiagnosticOrderStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDiagnosticOrderStatus[value]);
end;

Procedure TFhirDiagnosticOrder.SetPriority(value : TFhirEnum);
begin
  FPriority.free;
  FPriority := value;
end;

Function TFhirDiagnosticOrder.GetPriorityST : TFhirDiagnosticOrderPriority;
begin
  if FPriority = nil then
    result := TFhirDiagnosticOrderPriority(0)
  else
    result := TFhirDiagnosticOrderPriority(StringArrayIndexOfSensitive(CODES_TFhirDiagnosticOrderPriority, FPriority.value));
end;

Procedure TFhirDiagnosticOrder.SetPriorityST(value : TFhirDiagnosticOrderPriority);
begin
  if ord(value) = 0 then
    PriorityElement := nil
  else
    PriorityElement := TFhirEnum.create(CODES_TFhirDiagnosticOrderPriority[value]);
end;

Function TFhirDiagnosticOrder.GetEventList : TFhirDiagnosticOrderEventList;
begin
  if FEventList = nil then
    FEventList := TFhirDiagnosticOrderEventList.Create;
  result := FEventList;
end;

Function TFhirDiagnosticOrder.GetHasEventList : boolean;
begin
  result := (FEventList <> nil) and (FEventList.count > 0);
end;

Function TFhirDiagnosticOrder.GetItemList : TFhirDiagnosticOrderItemList;
begin
  if FItemList = nil then
    FItemList := TFhirDiagnosticOrderItemList.Create;
  result := FItemList;
end;

Function TFhirDiagnosticOrder.GetHasItemList : boolean;
begin
  result := (FItemList <> nil) and (FItemList.count > 0);
end;

Function TFhirDiagnosticOrder.GetNoteList : TFhirAnnotationList;
begin
  if FNoteList = nil then
    FNoteList := TFhirAnnotationList.Create;
  result := FNoteList;
end;

Function TFhirDiagnosticOrder.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;


{ TFhirDiagnosticOrderListEnumerator }

Constructor TFhirDiagnosticOrderListEnumerator.Create(list : TFhirDiagnosticOrderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDiagnosticOrderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDiagnosticOrderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDiagnosticOrderListEnumerator.GetCurrent : TFhirDiagnosticOrder;
begin
  Result := FList[FIndex];
end;


{ TFhirDiagnosticOrderList }
procedure TFhirDiagnosticOrderList.AddItem(value: TFhirDiagnosticOrder);
begin
  assert(value.ClassName = 'TFhirDiagnosticOrder', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDiagnosticOrder');
  add(value);
end;


function TFhirDiagnosticOrderList.Append: TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderList.ClearItems;
begin
  Clear;
end;

function TFhirDiagnosticOrderList.GetEnumerator : TFhirDiagnosticOrderListEnumerator;
begin
  result := TFhirDiagnosticOrderListEnumerator.Create(self.link);
end;

function TFhirDiagnosticOrderList.Clone: TFhirDiagnosticOrderList;
begin
  result := TFhirDiagnosticOrderList(inherited Clone);
end;

function TFhirDiagnosticOrderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDiagnosticOrderList.GetItemN(index: Integer): TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderList.IndexOf(value: TFhirDiagnosticOrder): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDiagnosticOrderList.Insert(index: Integer): TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticOrderList.InsertItem(index: Integer; value: TFhirDiagnosticOrder);
begin
  assert(value is TFhirDiagnosticOrder);
  Inherited Insert(index, value);
end;

function TFhirDiagnosticOrderList.Item(index: Integer): TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder(ObjectByIndex[index]);
end;

function TFhirDiagnosticOrderList.Link: TFhirDiagnosticOrderList;
begin
  result := TFhirDiagnosticOrderList(inherited Link);
end;

procedure TFhirDiagnosticOrderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDiagnosticOrderList.SetItemByIndex(index: Integer; value: TFhirDiagnosticOrder);
begin
  assert(value is TFhirDiagnosticOrder);
  FhirDiagnosticOrders[index] := value;
end;

procedure TFhirDiagnosticOrderList.SetItemN(index: Integer; value: TFhirDiagnosticOrder);
begin
  assert(value is TFhirDiagnosticOrder);
  ObjectByIndex[index] := value;
end;

{ TFhirDiagnosticReportImage }

constructor TFhirDiagnosticReportImage.Create;
begin
  inherited;
end;

destructor TFhirDiagnosticReportImage.Destroy;
begin
  FComment.free;
  FLink_.free;
  inherited;
end;

procedure TFhirDiagnosticReportImage.Assign(oSource : TAdvObject);
begin
  inherited;
  commentElement := TFhirDiagnosticReportImage(oSource).commentElement.Clone;
  link_ := TFhirDiagnosticReportImage(oSource).link_.Clone;
end;

procedure TFhirDiagnosticReportImage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'comment') Then
     list.add(FComment.Link);
  if (child_name = 'link') Then
     list.add(FLink_.Link);
end;

procedure TFhirDiagnosticReportImage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'link', 'Reference(Media)', FLink_.Link));{2}
end;

procedure TFhirDiagnosticReportImage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else if (propName = 'link') then Link_ := propValue as TFhirReference{TFhirMedia}{4b}
  else inherited;
end;

function TFhirDiagnosticReportImage.FhirType : string;
begin
  result := 'image';
end;

function TFhirDiagnosticReportImage.Link : TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage(inherited Link);
end;

function TFhirDiagnosticReportImage.Clone : TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage(inherited Clone);
end;

function TFhirDiagnosticReportImage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticReportImage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDiagnosticReportImage)) then
    result := false
  else
  begin
    o := TFhirDiagnosticReportImage(other);
    result := compareDeep(commentElement, o.commentElement, true) and compareDeep(link_Element, o.link_Element, true);
  end;
end;

function TFhirDiagnosticReportImage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDiagnosticReportImage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDiagnosticReportImage)) then
    result := false
  else
  begin
    o := TFhirDiagnosticReportImage(other);
    result := compareValues(commentElement, o.commentElement, true);
  end;
end;

{ TFhirDiagnosticReportImage }

Procedure TFhirDiagnosticReportImage.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirDiagnosticReportImage.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirDiagnosticReportImage.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;

Procedure TFhirDiagnosticReportImage.SetLink_(value : TFhirReference{TFhirMedia});
begin
  FLink_.free;
  FLink_ := value;
end;


{ TFhirDiagnosticReportImageListEnumerator }

Constructor TFhirDiagnosticReportImageListEnumerator.Create(list : TFhirDiagnosticReportImageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDiagnosticReportImageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDiagnosticReportImageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDiagnosticReportImageListEnumerator.GetCurrent : TFhirDiagnosticReportImage;
begin
  Result := FList[FIndex];
end;


{ TFhirDiagnosticReportImageList }
procedure TFhirDiagnosticReportImageList.AddItem(value: TFhirDiagnosticReportImage);
begin
  assert(value.ClassName = 'TFhirDiagnosticReportImage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDiagnosticReportImage');
  add(value);
end;


function TFhirDiagnosticReportImageList.Append: TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticReportImageList.ClearItems;
begin
  Clear;
end;

function TFhirDiagnosticReportImageList.GetEnumerator : TFhirDiagnosticReportImageListEnumerator;
begin
  result := TFhirDiagnosticReportImageListEnumerator.Create(self.link);
end;

function TFhirDiagnosticReportImageList.Clone: TFhirDiagnosticReportImageList;
begin
  result := TFhirDiagnosticReportImageList(inherited Clone);
end;

function TFhirDiagnosticReportImageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDiagnosticReportImageList.GetItemN(index: Integer): TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage(ObjectByIndex[index]);
end;

function TFhirDiagnosticReportImageList.IndexOf(value: TFhirDiagnosticReportImage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDiagnosticReportImageList.Insert(index: Integer): TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticReportImageList.InsertItem(index: Integer; value: TFhirDiagnosticReportImage);
begin
  assert(value is TFhirDiagnosticReportImage);
  Inherited Insert(index, value);
end;

function TFhirDiagnosticReportImageList.Item(index: Integer): TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage(ObjectByIndex[index]);
end;

function TFhirDiagnosticReportImageList.Link: TFhirDiagnosticReportImageList;
begin
  result := TFhirDiagnosticReportImageList(inherited Link);
end;

procedure TFhirDiagnosticReportImageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDiagnosticReportImageList.SetItemByIndex(index: Integer; value: TFhirDiagnosticReportImage);
begin
  assert(value is TFhirDiagnosticReportImage);
  FhirDiagnosticReportImages[index] := value;
end;

procedure TFhirDiagnosticReportImageList.SetItemN(index: Integer; value: TFhirDiagnosticReportImage);
begin
  assert(value is TFhirDiagnosticReportImage);
  ObjectByIndex[index] := value;
end;

{ TFhirDiagnosticReport }

constructor TFhirDiagnosticReport.Create;
begin
  inherited;
end;

destructor TFhirDiagnosticReport.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FCategory.free;
  FCode.free;
  FSubject.free;
  FEncounter.free;
  FEffective.free;
  FIssued.free;
  FPerformer.free;
  FRequestList.Free;
  FSpecimenList.Free;
  FResultList.Free;
  FImagingStudyList.Free;
  FImageList.Free;
  FConclusion.free;
  FCodedDiagnosisList.Free;
  FPresentedFormList.Free;
  inherited;
end;

function TFhirDiagnosticReport.GetResourceType : TFhirResourceType;
begin
  result := frtDiagnosticReport;
end;

procedure TFhirDiagnosticReport.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirDiagnosticReport(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDiagnosticReport(oSource).FIdentifierList);
  end;
  FStatus := TFhirDiagnosticReport(oSource).FStatus.Link;
  category := TFhirDiagnosticReport(oSource).category.Clone;
  code := TFhirDiagnosticReport(oSource).code.Clone;
  subject := TFhirDiagnosticReport(oSource).subject.Clone;
  encounter := TFhirDiagnosticReport(oSource).encounter.Clone;
  effective := TFhirDiagnosticReport(oSource).effective.Clone;
  issuedElement := TFhirDiagnosticReport(oSource).issuedElement.Clone;
  performer := TFhirDiagnosticReport(oSource).performer.Clone;
  if (TFhirDiagnosticReport(oSource).FRequestList = nil) then
  begin
    FRequestList.free;
    FRequestList := nil;
  end
  else
  begin
    FRequestList := TFhirReferenceList{Resource}.Create;
    FRequestList.Assign(TFhirDiagnosticReport(oSource).FRequestList);
  end;
  if (TFhirDiagnosticReport(oSource).FSpecimenList = nil) then
  begin
    FSpecimenList.free;
    FSpecimenList := nil;
  end
  else
  begin
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
    FSpecimenList.Assign(TFhirDiagnosticReport(oSource).FSpecimenList);
  end;
  if (TFhirDiagnosticReport(oSource).FResultList = nil) then
  begin
    FResultList.free;
    FResultList := nil;
  end
  else
  begin
    FResultList := TFhirReferenceList{TFhirObservation}.Create;
    FResultList.Assign(TFhirDiagnosticReport(oSource).FResultList);
  end;
  if (TFhirDiagnosticReport(oSource).FImagingStudyList = nil) then
  begin
    FImagingStudyList.free;
    FImagingStudyList := nil;
  end
  else
  begin
    FImagingStudyList := TFhirReferenceList{Resource}.Create;
    FImagingStudyList.Assign(TFhirDiagnosticReport(oSource).FImagingStudyList);
  end;
  if (TFhirDiagnosticReport(oSource).FImageList = nil) then
  begin
    FImageList.free;
    FImageList := nil;
  end
  else
  begin
    FImageList := TFhirDiagnosticReportImageList.Create;
    FImageList.Assign(TFhirDiagnosticReport(oSource).FImageList);
  end;
  conclusionElement := TFhirDiagnosticReport(oSource).conclusionElement.Clone;
  if (TFhirDiagnosticReport(oSource).FCodedDiagnosisList = nil) then
  begin
    FCodedDiagnosisList.free;
    FCodedDiagnosisList := nil;
  end
  else
  begin
    FCodedDiagnosisList := TFhirCodeableConceptList.Create;
    FCodedDiagnosisList.Assign(TFhirDiagnosticReport(oSource).FCodedDiagnosisList);
  end;
  if (TFhirDiagnosticReport(oSource).FPresentedFormList = nil) then
  begin
    FPresentedFormList.free;
    FPresentedFormList := nil;
  end
  else
  begin
    FPresentedFormList := TFhirAttachmentList.Create;
    FPresentedFormList.Assign(TFhirDiagnosticReport(oSource).FPresentedFormList);
  end;
end;

procedure TFhirDiagnosticReport.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'effective[x]') Then
     list.add(FEffective.Link);
  if (child_name = 'issued') Then
     list.add(FIssued.Link);
  if (child_name = 'performer') Then
     list.add(FPerformer.Link);
  if (child_name = 'request') Then
    list.addAll(FRequestList);
  if (child_name = 'specimen') Then
    list.addAll(FSpecimenList);
  if (child_name = 'result') Then
    list.addAll(FResultList);
  if (child_name = 'imagingStudy') Then
    list.addAll(FImagingStudyList);
  if (child_name = 'image') Then
    list.addAll(FImageList);
  if (child_name = 'conclusion') Then
     list.add(FConclusion.Link);
  if (child_name = 'codedDiagnosis') Then
    list.addAll(FCodedDiagnosisList);
  if (child_name = 'presentedForm') Then
    list.addAll(FPresentedFormList);
end;

procedure TFhirDiagnosticReport.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Device|Location)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'effective[x]', 'dateTime|Period', FEffective.Link));{2}
  oList.add(TFHIRProperty.create(self, 'issued', 'instant', FIssued.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner|Organization)', FPerformer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(DiagnosticOrder|ProcedureRequest|ReferralRequest)', FRequestList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'specimen', 'Reference(Specimen)', FSpecimenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'result', 'Reference(Observation)', FResultList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'imagingStudy', 'Reference(ImagingStudy|ImagingObjectSelection)', FImagingStudyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'image', '', FImageList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'conclusion', 'string', FConclusion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'codedDiagnosis', 'CodeableConcept', FCodedDiagnosisList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'presentedForm', 'Attachment', FPresentedFormList.Link)){3};
end;

procedure TFhirDiagnosticReport.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName.startsWith('effective')) then Effective := propValue as TFhirType{4}
  else if (propName = 'issued') then IssuedElement := propValue as TFhirInstant{5a}
  else if (propName = 'performer') then Performer := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'request') then RequestList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'specimen') then SpecimenList.add(propValue as TFhirReference{TFhirSpecimen}){2}
  else if (propName = 'result') then ResultList.add(propValue as TFhirReference{TFhirObservation}){2}
  else if (propName = 'imagingStudy') then ImagingStudyList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'image') then ImageList.add(propValue as TFhirDiagnosticReportImage){2}
  else if (propName = 'conclusion') then ConclusionElement := propValue as TFhirString{5a}
  else if (propName = 'codedDiagnosis') then CodedDiagnosisList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'presentedForm') then PresentedFormList.add(propValue as TFhirAttachment){2}
  else inherited;
end;

function TFhirDiagnosticReport.FhirType : string;
begin
  result := 'DiagnosticReport';
end;

function TFhirDiagnosticReport.Link : TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport(inherited Link);
end;

function TFhirDiagnosticReport.Clone : TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport(inherited Clone);
end;

{ TFhirDiagnosticReport }

Function TFhirDiagnosticReport.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDiagnosticReport.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDiagnosticReport.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDiagnosticReport.GetStatusST : TFhirDiagnosticReportStatus;
begin
  if FStatus = nil then
    result := TFhirDiagnosticReportStatus(0)
  else
    result := TFhirDiagnosticReportStatus(StringArrayIndexOfSensitive(CODES_TFhirDiagnosticReportStatus, FStatus.value));
end;

Procedure TFhirDiagnosticReport.SetStatusST(value : TFhirDiagnosticReportStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDiagnosticReportStatus[value]);
end;

Procedure TFhirDiagnosticReport.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirDiagnosticReport.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirDiagnosticReport.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirDiagnosticReport.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirDiagnosticReport.SetEffective(value : TFhirType);
begin
  FEffective.free;
  FEffective := value;
end;

Procedure TFhirDiagnosticReport.SetIssued(value : TFhirInstant);
begin
  FIssued.free;
  FIssued := value;
end;

Function TFhirDiagnosticReport.GetIssuedST : TDateAndTime;
begin
  if FIssued = nil then
    result := nil
  else
    result := FIssued.value;
end;

Procedure TFhirDiagnosticReport.SetIssuedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIssued = nil then
      FIssued := TFhirInstant.create;
    FIssued.value := value
  end
  else if FIssued <> nil then
    FIssued.value := nil;
end;

Procedure TFhirDiagnosticReport.SetPerformer(value : TFhirReference{Resource});
begin
  FPerformer.free;
  FPerformer := value;
end;

Function TFhirDiagnosticReport.GetRequestList : TFhirReferenceList{Resource};
begin
  if FRequestList = nil then
    FRequestList := TFhirReferenceList{Resource}.Create;
  result := FRequestList;
end;

Function TFhirDiagnosticReport.GetHasRequestList : boolean;
begin
  result := (FRequestList <> nil) and (FRequestList.count > 0);
end;

Function TFhirDiagnosticReport.GetSpecimenList : TFhirReferenceList{TFhirSpecimen};
begin
  if FSpecimenList = nil then
    FSpecimenList := TFhirReferenceList{TFhirSpecimen}.Create;
  result := FSpecimenList;
end;

Function TFhirDiagnosticReport.GetHasSpecimenList : boolean;
begin
  result := (FSpecimenList <> nil) and (FSpecimenList.count > 0);
end;

Function TFhirDiagnosticReport.GetResultList : TFhirReferenceList{TFhirObservation};
begin
  if FResultList = nil then
    FResultList := TFhirReferenceList{TFhirObservation}.Create;
  result := FResultList;
end;

Function TFhirDiagnosticReport.GetHasResultList : boolean;
begin
  result := (FResultList <> nil) and (FResultList.count > 0);
end;

Function TFhirDiagnosticReport.GetImagingStudyList : TFhirReferenceList{Resource};
begin
  if FImagingStudyList = nil then
    FImagingStudyList := TFhirReferenceList{Resource}.Create;
  result := FImagingStudyList;
end;

Function TFhirDiagnosticReport.GetHasImagingStudyList : boolean;
begin
  result := (FImagingStudyList <> nil) and (FImagingStudyList.count > 0);
end;

Function TFhirDiagnosticReport.GetImageList : TFhirDiagnosticReportImageList;
begin
  if FImageList = nil then
    FImageList := TFhirDiagnosticReportImageList.Create;
  result := FImageList;
end;

Function TFhirDiagnosticReport.GetHasImageList : boolean;
begin
  result := (FImageList <> nil) and (FImageList.count > 0);
end;

Procedure TFhirDiagnosticReport.SetConclusion(value : TFhirString);
begin
  FConclusion.free;
  FConclusion := value;
end;

Function TFhirDiagnosticReport.GetConclusionST : String;
begin
  if FConclusion = nil then
    result := ''
  else
    result := FConclusion.value;
end;

Procedure TFhirDiagnosticReport.SetConclusionST(value : String);
begin
  if value <> '' then
  begin
    if FConclusion = nil then
      FConclusion := TFhirString.create;
    FConclusion.value := value
  end
  else if FConclusion <> nil then
    FConclusion.value := '';
end;

Function TFhirDiagnosticReport.GetCodedDiagnosisList : TFhirCodeableConceptList;
begin
  if FCodedDiagnosisList = nil then
    FCodedDiagnosisList := TFhirCodeableConceptList.Create;
  result := FCodedDiagnosisList;
end;

Function TFhirDiagnosticReport.GetHasCodedDiagnosisList : boolean;
begin
  result := (FCodedDiagnosisList <> nil) and (FCodedDiagnosisList.count > 0);
end;

Function TFhirDiagnosticReport.GetPresentedFormList : TFhirAttachmentList;
begin
  if FPresentedFormList = nil then
    FPresentedFormList := TFhirAttachmentList.Create;
  result := FPresentedFormList;
end;

Function TFhirDiagnosticReport.GetHasPresentedFormList : boolean;
begin
  result := (FPresentedFormList <> nil) and (FPresentedFormList.count > 0);
end;


{ TFhirDiagnosticReportListEnumerator }

Constructor TFhirDiagnosticReportListEnumerator.Create(list : TFhirDiagnosticReportList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDiagnosticReportListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDiagnosticReportListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDiagnosticReportListEnumerator.GetCurrent : TFhirDiagnosticReport;
begin
  Result := FList[FIndex];
end;


{ TFhirDiagnosticReportList }
procedure TFhirDiagnosticReportList.AddItem(value: TFhirDiagnosticReport);
begin
  assert(value.ClassName = 'TFhirDiagnosticReport', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDiagnosticReport');
  add(value);
end;


function TFhirDiagnosticReportList.Append: TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticReportList.ClearItems;
begin
  Clear;
end;

function TFhirDiagnosticReportList.GetEnumerator : TFhirDiagnosticReportListEnumerator;
begin
  result := TFhirDiagnosticReportListEnumerator.Create(self.link);
end;

function TFhirDiagnosticReportList.Clone: TFhirDiagnosticReportList;
begin
  result := TFhirDiagnosticReportList(inherited Clone);
end;

function TFhirDiagnosticReportList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDiagnosticReportList.GetItemN(index: Integer): TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport(ObjectByIndex[index]);
end;

function TFhirDiagnosticReportList.IndexOf(value: TFhirDiagnosticReport): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDiagnosticReportList.Insert(index: Integer): TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDiagnosticReportList.InsertItem(index: Integer; value: TFhirDiagnosticReport);
begin
  assert(value is TFhirDiagnosticReport);
  Inherited Insert(index, value);
end;

function TFhirDiagnosticReportList.Item(index: Integer): TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport(ObjectByIndex[index]);
end;

function TFhirDiagnosticReportList.Link: TFhirDiagnosticReportList;
begin
  result := TFhirDiagnosticReportList(inherited Link);
end;

procedure TFhirDiagnosticReportList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDiagnosticReportList.SetItemByIndex(index: Integer; value: TFhirDiagnosticReport);
begin
  assert(value is TFhirDiagnosticReport);
  FhirDiagnosticReports[index] := value;
end;

procedure TFhirDiagnosticReportList.SetItemN(index: Integer; value: TFhirDiagnosticReport);
begin
  assert(value is TFhirDiagnosticReport);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentManifestContent }

constructor TFhirDocumentManifestContent.Create;
begin
  inherited;
end;

destructor TFhirDocumentManifestContent.Destroy;
begin
  FP.free;
  inherited;
end;

procedure TFhirDocumentManifestContent.Assign(oSource : TAdvObject);
begin
  inherited;
  p := TFhirDocumentManifestContent(oSource).p.Clone;
end;

procedure TFhirDocumentManifestContent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'p[x]') Then
     list.add(FP.Link);
end;

procedure TFhirDocumentManifestContent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'p[x]', 'Attachment|Reference(Any)', FP.Link));{2}
end;

procedure TFhirDocumentManifestContent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('p')) then P := propValue as TFhirType{4}
  else inherited;
end;

function TFhirDocumentManifestContent.FhirType : string;
begin
  result := 'content';
end;

function TFhirDocumentManifestContent.Link : TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent(inherited Link);
end;

function TFhirDocumentManifestContent.Clone : TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent(inherited Clone);
end;

function TFhirDocumentManifestContent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentManifestContent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentManifestContent)) then
    result := false
  else
  begin
    o := TFhirDocumentManifestContent(other);
    result := compareDeep(pElement, o.pElement, true);
  end;
end;

function TFhirDocumentManifestContent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentManifestContent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentManifestContent)) then
    result := false
  else
  begin
    o := TFhirDocumentManifestContent(other);
    result := true;
  end;
end;

{ TFhirDocumentManifestContent }

Procedure TFhirDocumentManifestContent.SetP(value : TFhirType);
begin
  FP.free;
  FP := value;
end;


{ TFhirDocumentManifestContentListEnumerator }

Constructor TFhirDocumentManifestContentListEnumerator.Create(list : TFhirDocumentManifestContentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentManifestContentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentManifestContentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentManifestContentListEnumerator.GetCurrent : TFhirDocumentManifestContent;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentManifestContentList }
procedure TFhirDocumentManifestContentList.AddItem(value: TFhirDocumentManifestContent);
begin
  assert(value.ClassName = 'TFhirDocumentManifestContent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentManifestContent');
  add(value);
end;


function TFhirDocumentManifestContentList.Append: TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestContentList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentManifestContentList.GetEnumerator : TFhirDocumentManifestContentListEnumerator;
begin
  result := TFhirDocumentManifestContentListEnumerator.Create(self.link);
end;

function TFhirDocumentManifestContentList.Clone: TFhirDocumentManifestContentList;
begin
  result := TFhirDocumentManifestContentList(inherited Clone);
end;

function TFhirDocumentManifestContentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentManifestContentList.GetItemN(index: Integer): TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent(ObjectByIndex[index]);
end;

function TFhirDocumentManifestContentList.IndexOf(value: TFhirDocumentManifestContent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentManifestContentList.Insert(index: Integer): TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestContentList.InsertItem(index: Integer; value: TFhirDocumentManifestContent);
begin
  assert(value is TFhirDocumentManifestContent);
  Inherited Insert(index, value);
end;

function TFhirDocumentManifestContentList.Item(index: Integer): TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent(ObjectByIndex[index]);
end;

function TFhirDocumentManifestContentList.Link: TFhirDocumentManifestContentList;
begin
  result := TFhirDocumentManifestContentList(inherited Link);
end;

procedure TFhirDocumentManifestContentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentManifestContentList.SetItemByIndex(index: Integer; value: TFhirDocumentManifestContent);
begin
  assert(value is TFhirDocumentManifestContent);
  FhirDocumentManifestContents[index] := value;
end;

procedure TFhirDocumentManifestContentList.SetItemN(index: Integer; value: TFhirDocumentManifestContent);
begin
  assert(value is TFhirDocumentManifestContent);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentManifestRelated }

constructor TFhirDocumentManifestRelated.Create;
begin
  inherited;
end;

destructor TFhirDocumentManifestRelated.Destroy;
begin
  FIdentifier.free;
  FRef.free;
  inherited;
end;

procedure TFhirDocumentManifestRelated.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirDocumentManifestRelated(oSource).identifier.Clone;
  ref := TFhirDocumentManifestRelated(oSource).ref.Clone;
end;

procedure TFhirDocumentManifestRelated.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'ref') Then
     list.add(FRef.Link);
end;

procedure TFhirDocumentManifestRelated.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ref', 'Reference(Any)', FRef.Link));{2}
end;

procedure TFhirDocumentManifestRelated.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'ref') then Ref := propValue as TFhirReference{TFhirReference}{4b}
  else inherited;
end;

function TFhirDocumentManifestRelated.FhirType : string;
begin
  result := 'related';
end;

function TFhirDocumentManifestRelated.Link : TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated(inherited Link);
end;

function TFhirDocumentManifestRelated.Clone : TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated(inherited Clone);
end;

function TFhirDocumentManifestRelated.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentManifestRelated;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentManifestRelated)) then
    result := false
  else
  begin
    o := TFhirDocumentManifestRelated(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(refElement, o.refElement, true);
  end;
end;

function TFhirDocumentManifestRelated.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentManifestRelated;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentManifestRelated)) then
    result := false
  else
  begin
    o := TFhirDocumentManifestRelated(other);
    result := true;
  end;
end;

{ TFhirDocumentManifestRelated }

Procedure TFhirDocumentManifestRelated.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirDocumentManifestRelated.SetRef(value : TFhirReference{TFhirReference});
begin
  FRef.free;
  FRef := value;
end;


{ TFhirDocumentManifestRelatedListEnumerator }

Constructor TFhirDocumentManifestRelatedListEnumerator.Create(list : TFhirDocumentManifestRelatedList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentManifestRelatedListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentManifestRelatedListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentManifestRelatedListEnumerator.GetCurrent : TFhirDocumentManifestRelated;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentManifestRelatedList }
procedure TFhirDocumentManifestRelatedList.AddItem(value: TFhirDocumentManifestRelated);
begin
  assert(value.ClassName = 'TFhirDocumentManifestRelated', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentManifestRelated');
  add(value);
end;


function TFhirDocumentManifestRelatedList.Append: TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestRelatedList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentManifestRelatedList.GetEnumerator : TFhirDocumentManifestRelatedListEnumerator;
begin
  result := TFhirDocumentManifestRelatedListEnumerator.Create(self.link);
end;

function TFhirDocumentManifestRelatedList.Clone: TFhirDocumentManifestRelatedList;
begin
  result := TFhirDocumentManifestRelatedList(inherited Clone);
end;

function TFhirDocumentManifestRelatedList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentManifestRelatedList.GetItemN(index: Integer): TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated(ObjectByIndex[index]);
end;

function TFhirDocumentManifestRelatedList.IndexOf(value: TFhirDocumentManifestRelated): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentManifestRelatedList.Insert(index: Integer): TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestRelatedList.InsertItem(index: Integer; value: TFhirDocumentManifestRelated);
begin
  assert(value is TFhirDocumentManifestRelated);
  Inherited Insert(index, value);
end;

function TFhirDocumentManifestRelatedList.Item(index: Integer): TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated(ObjectByIndex[index]);
end;

function TFhirDocumentManifestRelatedList.Link: TFhirDocumentManifestRelatedList;
begin
  result := TFhirDocumentManifestRelatedList(inherited Link);
end;

procedure TFhirDocumentManifestRelatedList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentManifestRelatedList.SetItemByIndex(index: Integer; value: TFhirDocumentManifestRelated);
begin
  assert(value is TFhirDocumentManifestRelated);
  FhirDocumentManifestRelateds[index] := value;
end;

procedure TFhirDocumentManifestRelatedList.SetItemN(index: Integer; value: TFhirDocumentManifestRelated);
begin
  assert(value is TFhirDocumentManifestRelated);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentManifest }

constructor TFhirDocumentManifest.Create;
begin
  inherited;
end;

destructor TFhirDocumentManifest.Destroy;
begin
  FMasterIdentifier.free;
  FIdentifierList.Free;
  FSubject.free;
  FRecipientList.Free;
  FType_.free;
  FAuthorList.Free;
  FCreated.free;
  FSource.free;
  FStatus.free;
  FDescription.free;
  FContentList.Free;
  FRelatedList.Free;
  inherited;
end;

function TFhirDocumentManifest.GetResourceType : TFhirResourceType;
begin
  result := frtDocumentManifest;
end;

procedure TFhirDocumentManifest.Assign(oSource : TAdvObject);
begin
  inherited;
  masterIdentifier := TFhirDocumentManifest(oSource).masterIdentifier.Clone;
  if (TFhirDocumentManifest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDocumentManifest(oSource).FIdentifierList);
  end;
  subject := TFhirDocumentManifest(oSource).subject.Clone;
  if (TFhirDocumentManifest(oSource).FRecipientList = nil) then
  begin
    FRecipientList.free;
    FRecipientList := nil;
  end
  else
  begin
    FRecipientList := TFhirReferenceList{Resource}.Create;
    FRecipientList.Assign(TFhirDocumentManifest(oSource).FRecipientList);
  end;
  type_ := TFhirDocumentManifest(oSource).type_.Clone;
  if (TFhirDocumentManifest(oSource).FAuthorList = nil) then
  begin
    FAuthorList.free;
    FAuthorList := nil;
  end
  else
  begin
    FAuthorList := TFhirReferenceList{Resource}.Create;
    FAuthorList.Assign(TFhirDocumentManifest(oSource).FAuthorList);
  end;
  createdElement := TFhirDocumentManifest(oSource).createdElement.Clone;
  sourceElement := TFhirDocumentManifest(oSource).sourceElement.Clone;
  FStatus := TFhirDocumentManifest(oSource).FStatus.Link;
  descriptionElement := TFhirDocumentManifest(oSource).descriptionElement.Clone;
  if (TFhirDocumentManifest(oSource).FContentList = nil) then
  begin
    FContentList.free;
    FContentList := nil;
  end
  else
  begin
    FContentList := TFhirDocumentManifestContentList.Create;
    FContentList.Assign(TFhirDocumentManifest(oSource).FContentList);
  end;
  if (TFhirDocumentManifest(oSource).FRelatedList = nil) then
  begin
    FRelatedList.free;
    FRelatedList := nil;
  end
  else
  begin
    FRelatedList := TFhirDocumentManifestRelatedList.Create;
    FRelatedList.Assign(TFhirDocumentManifest(oSource).FRelatedList);
  end;
end;

procedure TFhirDocumentManifest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'masterIdentifier') Then
     list.add(FMasterIdentifier.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'recipient') Then
    list.addAll(FRecipientList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'author') Then
    list.addAll(FAuthorList);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'content') Then
    list.addAll(FContentList);
  if (child_name = 'related') Then
    list.addAll(FRelatedList);
end;

procedure TFhirDocumentManifest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'masterIdentifier', 'Identifier', FMasterIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Practitioner|Group|Device)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recipient', 'Reference(Patient|Practitioner|RelatedPerson|Organization)', FRecipientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Organization|Device|Patient|RelatedPerson)', FAuthorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'uri', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'content', '', FContentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'related', '', FRelatedList.Link)){3};
end;

procedure TFhirDocumentManifest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'masterIdentifier') then MasterIdentifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'recipient') then RecipientList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'author') then AuthorList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'source') then SourceElement := propValue as TFhirUri{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'content') then ContentList.add(propValue as TFhirDocumentManifestContent){2}
  else if (propName = 'related') then RelatedList.add(propValue as TFhirDocumentManifestRelated){2}
  else inherited;
end;

function TFhirDocumentManifest.FhirType : string;
begin
  result := 'DocumentManifest';
end;

function TFhirDocumentManifest.Link : TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest(inherited Link);
end;

function TFhirDocumentManifest.Clone : TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest(inherited Clone);
end;

{ TFhirDocumentManifest }

Procedure TFhirDocumentManifest.SetMasterIdentifier(value : TFhirIdentifier);
begin
  FMasterIdentifier.free;
  FMasterIdentifier := value;
end;

Function TFhirDocumentManifest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDocumentManifest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDocumentManifest.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Function TFhirDocumentManifest.GetRecipientList : TFhirReferenceList{Resource};
begin
  if FRecipientList = nil then
    FRecipientList := TFhirReferenceList{Resource}.Create;
  result := FRecipientList;
end;

Function TFhirDocumentManifest.GetHasRecipientList : boolean;
begin
  result := (FRecipientList <> nil) and (FRecipientList.count > 0);
end;

Procedure TFhirDocumentManifest.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirDocumentManifest.GetAuthorList : TFhirReferenceList{Resource};
begin
  if FAuthorList = nil then
    FAuthorList := TFhirReferenceList{Resource}.Create;
  result := FAuthorList;
end;

Function TFhirDocumentManifest.GetHasAuthorList : boolean;
begin
  result := (FAuthorList <> nil) and (FAuthorList.count > 0);
end;

Procedure TFhirDocumentManifest.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirDocumentManifest.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirDocumentManifest.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirDocumentManifest.SetSource(value : TFhirUri);
begin
  FSource.free;
  FSource := value;
end;

Function TFhirDocumentManifest.GetSourceST : String;
begin
  if FSource = nil then
    result := ''
  else
    result := FSource.value;
end;

Procedure TFhirDocumentManifest.SetSourceST(value : String);
begin
  if value <> '' then
  begin
    if FSource = nil then
      FSource := TFhirUri.create;
    FSource.value := value
  end
  else if FSource <> nil then
    FSource.value := '';
end;

Procedure TFhirDocumentManifest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDocumentManifest.GetStatusST : TFhirDocumentReferenceStatus;
begin
  if FStatus = nil then
    result := TFhirDocumentReferenceStatus(0)
  else
    result := TFhirDocumentReferenceStatus(StringArrayIndexOfSensitive(CODES_TFhirDocumentReferenceStatus, FStatus.value));
end;

Procedure TFhirDocumentManifest.SetStatusST(value : TFhirDocumentReferenceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDocumentReferenceStatus[value]);
end;

Procedure TFhirDocumentManifest.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirDocumentManifest.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirDocumentManifest.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirDocumentManifest.GetContentList : TFhirDocumentManifestContentList;
begin
  if FContentList = nil then
    FContentList := TFhirDocumentManifestContentList.Create;
  result := FContentList;
end;

Function TFhirDocumentManifest.GetHasContentList : boolean;
begin
  result := (FContentList <> nil) and (FContentList.count > 0);
end;

Function TFhirDocumentManifest.GetRelatedList : TFhirDocumentManifestRelatedList;
begin
  if FRelatedList = nil then
    FRelatedList := TFhirDocumentManifestRelatedList.Create;
  result := FRelatedList;
end;

Function TFhirDocumentManifest.GetHasRelatedList : boolean;
begin
  result := (FRelatedList <> nil) and (FRelatedList.count > 0);
end;


{ TFhirDocumentManifestListEnumerator }

Constructor TFhirDocumentManifestListEnumerator.Create(list : TFhirDocumentManifestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentManifestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentManifestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentManifestListEnumerator.GetCurrent : TFhirDocumentManifest;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentManifestList }
procedure TFhirDocumentManifestList.AddItem(value: TFhirDocumentManifest);
begin
  assert(value.ClassName = 'TFhirDocumentManifest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentManifest');
  add(value);
end;


function TFhirDocumentManifestList.Append: TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentManifestList.GetEnumerator : TFhirDocumentManifestListEnumerator;
begin
  result := TFhirDocumentManifestListEnumerator.Create(self.link);
end;

function TFhirDocumentManifestList.Clone: TFhirDocumentManifestList;
begin
  result := TFhirDocumentManifestList(inherited Clone);
end;

function TFhirDocumentManifestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentManifestList.GetItemN(index: Integer): TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest(ObjectByIndex[index]);
end;

function TFhirDocumentManifestList.IndexOf(value: TFhirDocumentManifest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentManifestList.Insert(index: Integer): TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentManifestList.InsertItem(index: Integer; value: TFhirDocumentManifest);
begin
  assert(value is TFhirDocumentManifest);
  Inherited Insert(index, value);
end;

function TFhirDocumentManifestList.Item(index: Integer): TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest(ObjectByIndex[index]);
end;

function TFhirDocumentManifestList.Link: TFhirDocumentManifestList;
begin
  result := TFhirDocumentManifestList(inherited Link);
end;

procedure TFhirDocumentManifestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentManifestList.SetItemByIndex(index: Integer; value: TFhirDocumentManifest);
begin
  assert(value is TFhirDocumentManifest);
  FhirDocumentManifests[index] := value;
end;

procedure TFhirDocumentManifestList.SetItemN(index: Integer; value: TFhirDocumentManifest);
begin
  assert(value is TFhirDocumentManifest);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentReferenceRelatesTo }

constructor TFhirDocumentReferenceRelatesTo.Create;
begin
  inherited;
end;

destructor TFhirDocumentReferenceRelatesTo.Destroy;
begin
  FCode.free;
  FTarget.free;
  inherited;
end;

procedure TFhirDocumentReferenceRelatesTo.Assign(oSource : TAdvObject);
begin
  inherited;
  FCode := TFhirDocumentReferenceRelatesTo(oSource).FCode.Link;
  target := TFhirDocumentReferenceRelatesTo(oSource).target.Clone;
end;

procedure TFhirDocumentReferenceRelatesTo.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
end;

procedure TFhirDocumentReferenceRelatesTo.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(DocumentReference)', FTarget.Link));{2}
end;

procedure TFhirDocumentReferenceRelatesTo.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirDocumentReference}{4b}
  else inherited;
end;

function TFhirDocumentReferenceRelatesTo.FhirType : string;
begin
  result := 'relatesTo';
end;

function TFhirDocumentReferenceRelatesTo.Link : TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo(inherited Link);
end;

function TFhirDocumentReferenceRelatesTo.Clone : TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo(inherited Clone);
end;

function TFhirDocumentReferenceRelatesTo.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceRelatesTo;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceRelatesTo)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceRelatesTo(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(targetElement, o.targetElement, true);
  end;
end;

function TFhirDocumentReferenceRelatesTo.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceRelatesTo;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceRelatesTo)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceRelatesTo(other);
    result := compareValues(codeElement, o.codeElement, true);
  end;
end;

{ TFhirDocumentReferenceRelatesTo }

Procedure TFhirDocumentReferenceRelatesTo.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirDocumentReferenceRelatesTo.GetCodeST : TFhirDocumentRelationshipType;
begin
  if FCode = nil then
    result := TFhirDocumentRelationshipType(0)
  else
    result := TFhirDocumentRelationshipType(StringArrayIndexOfSensitive(CODES_TFhirDocumentRelationshipType, FCode.value));
end;

Procedure TFhirDocumentReferenceRelatesTo.SetCodeST(value : TFhirDocumentRelationshipType);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirDocumentRelationshipType[value]);
end;

Procedure TFhirDocumentReferenceRelatesTo.SetTarget(value : TFhirReference{TFhirDocumentReference});
begin
  FTarget.free;
  FTarget := value;
end;


{ TFhirDocumentReferenceRelatesToListEnumerator }

Constructor TFhirDocumentReferenceRelatesToListEnumerator.Create(list : TFhirDocumentReferenceRelatesToList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentReferenceRelatesToListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentReferenceRelatesToListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentReferenceRelatesToListEnumerator.GetCurrent : TFhirDocumentReferenceRelatesTo;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentReferenceRelatesToList }
procedure TFhirDocumentReferenceRelatesToList.AddItem(value: TFhirDocumentReferenceRelatesTo);
begin
  assert(value.ClassName = 'TFhirDocumentReferenceRelatesTo', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentReferenceRelatesTo');
  add(value);
end;


function TFhirDocumentReferenceRelatesToList.Append: TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceRelatesToList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentReferenceRelatesToList.GetEnumerator : TFhirDocumentReferenceRelatesToListEnumerator;
begin
  result := TFhirDocumentReferenceRelatesToListEnumerator.Create(self.link);
end;

function TFhirDocumentReferenceRelatesToList.Clone: TFhirDocumentReferenceRelatesToList;
begin
  result := TFhirDocumentReferenceRelatesToList(inherited Clone);
end;

function TFhirDocumentReferenceRelatesToList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentReferenceRelatesToList.GetItemN(index: Integer): TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceRelatesToList.IndexOf(value: TFhirDocumentReferenceRelatesTo): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentReferenceRelatesToList.Insert(index: Integer): TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceRelatesToList.InsertItem(index: Integer; value: TFhirDocumentReferenceRelatesTo);
begin
  assert(value is TFhirDocumentReferenceRelatesTo);
  Inherited Insert(index, value);
end;

function TFhirDocumentReferenceRelatesToList.Item(index: Integer): TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceRelatesToList.Link: TFhirDocumentReferenceRelatesToList;
begin
  result := TFhirDocumentReferenceRelatesToList(inherited Link);
end;

procedure TFhirDocumentReferenceRelatesToList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentReferenceRelatesToList.SetItemByIndex(index: Integer; value: TFhirDocumentReferenceRelatesTo);
begin
  assert(value is TFhirDocumentReferenceRelatesTo);
  FhirDocumentReferenceRelatesTos[index] := value;
end;

procedure TFhirDocumentReferenceRelatesToList.SetItemN(index: Integer; value: TFhirDocumentReferenceRelatesTo);
begin
  assert(value is TFhirDocumentReferenceRelatesTo);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentReferenceContent }

constructor TFhirDocumentReferenceContent.Create;
begin
  inherited;
end;

destructor TFhirDocumentReferenceContent.Destroy;
begin
  FAttachment.free;
  FFormatList.Free;
  inherited;
end;

procedure TFhirDocumentReferenceContent.Assign(oSource : TAdvObject);
begin
  inherited;
  attachment := TFhirDocumentReferenceContent(oSource).attachment.Clone;
  if (TFhirDocumentReferenceContent(oSource).FFormatList = nil) then
  begin
    FFormatList.free;
    FFormatList := nil;
  end
  else
  begin
    FFormatList := TFhirCodingList.Create;
    FFormatList.Assign(TFhirDocumentReferenceContent(oSource).FFormatList);
  end;
end;

procedure TFhirDocumentReferenceContent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'attachment') Then
     list.add(FAttachment.Link);
  if (child_name = 'format') Then
    list.addAll(FFormatList);
end;

procedure TFhirDocumentReferenceContent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'attachment', 'Attachment', FAttachment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'format', 'Coding', FFormatList.Link)){3};
end;

procedure TFhirDocumentReferenceContent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'attachment') then Attachment := propValue as TFhirAttachment{4b}
  else if (propName = 'format') then FormatList.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirDocumentReferenceContent.FhirType : string;
begin
  result := 'content';
end;

function TFhirDocumentReferenceContent.Link : TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent(inherited Link);
end;

function TFhirDocumentReferenceContent.Clone : TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent(inherited Clone);
end;

function TFhirDocumentReferenceContent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContent)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContent(other);
    result := compareDeep(attachmentElement, o.attachmentElement, true) and compareDeep(formatList, o.formatList, true);
  end;
end;

function TFhirDocumentReferenceContent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContent)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContent(other);
    result := true;
  end;
end;

{ TFhirDocumentReferenceContent }

Procedure TFhirDocumentReferenceContent.SetAttachment(value : TFhirAttachment);
begin
  FAttachment.free;
  FAttachment := value;
end;

Function TFhirDocumentReferenceContent.GetFormatList : TFhirCodingList;
begin
  if FFormatList = nil then
    FFormatList := TFhirCodingList.Create;
  result := FFormatList;
end;

Function TFhirDocumentReferenceContent.GetHasFormatList : boolean;
begin
  result := (FFormatList <> nil) and (FFormatList.count > 0);
end;


{ TFhirDocumentReferenceContentListEnumerator }

Constructor TFhirDocumentReferenceContentListEnumerator.Create(list : TFhirDocumentReferenceContentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentReferenceContentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentReferenceContentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentReferenceContentListEnumerator.GetCurrent : TFhirDocumentReferenceContent;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentReferenceContentList }
procedure TFhirDocumentReferenceContentList.AddItem(value: TFhirDocumentReferenceContent);
begin
  assert(value.ClassName = 'TFhirDocumentReferenceContent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentReferenceContent');
  add(value);
end;


function TFhirDocumentReferenceContentList.Append: TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContentList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentReferenceContentList.GetEnumerator : TFhirDocumentReferenceContentListEnumerator;
begin
  result := TFhirDocumentReferenceContentListEnumerator.Create(self.link);
end;

function TFhirDocumentReferenceContentList.Clone: TFhirDocumentReferenceContentList;
begin
  result := TFhirDocumentReferenceContentList(inherited Clone);
end;

function TFhirDocumentReferenceContentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentReferenceContentList.GetItemN(index: Integer): TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContentList.IndexOf(value: TFhirDocumentReferenceContent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentReferenceContentList.Insert(index: Integer): TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContentList.InsertItem(index: Integer; value: TFhirDocumentReferenceContent);
begin
  assert(value is TFhirDocumentReferenceContent);
  Inherited Insert(index, value);
end;

function TFhirDocumentReferenceContentList.Item(index: Integer): TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContentList.Link: TFhirDocumentReferenceContentList;
begin
  result := TFhirDocumentReferenceContentList(inherited Link);
end;

procedure TFhirDocumentReferenceContentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentReferenceContentList.SetItemByIndex(index: Integer; value: TFhirDocumentReferenceContent);
begin
  assert(value is TFhirDocumentReferenceContent);
  FhirDocumentReferenceContents[index] := value;
end;

procedure TFhirDocumentReferenceContentList.SetItemN(index: Integer; value: TFhirDocumentReferenceContent);
begin
  assert(value is TFhirDocumentReferenceContent);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentReferenceContext }

constructor TFhirDocumentReferenceContext.Create;
begin
  inherited;
end;

destructor TFhirDocumentReferenceContext.Destroy;
begin
  FEncounter.free;
  FEventList.Free;
  FPeriod.free;
  FFacilityType.free;
  FPracticeSetting.free;
  FSourcePatientInfo.free;
  FRelatedList.Free;
  inherited;
end;

procedure TFhirDocumentReferenceContext.Assign(oSource : TAdvObject);
begin
  inherited;
  encounter := TFhirDocumentReferenceContext(oSource).encounter.Clone;
  if (TFhirDocumentReferenceContext(oSource).FEventList = nil) then
  begin
    FEventList.free;
    FEventList := nil;
  end
  else
  begin
    FEventList := TFhirCodeableConceptList.Create;
    FEventList.Assign(TFhirDocumentReferenceContext(oSource).FEventList);
  end;
  period := TFhirDocumentReferenceContext(oSource).period.Clone;
  facilityType := TFhirDocumentReferenceContext(oSource).facilityType.Clone;
  practiceSetting := TFhirDocumentReferenceContext(oSource).practiceSetting.Clone;
  sourcePatientInfo := TFhirDocumentReferenceContext(oSource).sourcePatientInfo.Clone;
  if (TFhirDocumentReferenceContext(oSource).FRelatedList = nil) then
  begin
    FRelatedList.free;
    FRelatedList := nil;
  end
  else
  begin
    FRelatedList := TFhirDocumentReferenceContextRelatedList.Create;
    FRelatedList.Assign(TFhirDocumentReferenceContext(oSource).FRelatedList);
  end;
end;

procedure TFhirDocumentReferenceContext.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'event') Then
    list.addAll(FEventList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'facilityType') Then
     list.add(FFacilityType.Link);
  if (child_name = 'practiceSetting') Then
     list.add(FPracticeSetting.Link);
  if (child_name = 'sourcePatientInfo') Then
     list.add(FSourcePatientInfo.Link);
  if (child_name = 'related') Then
    list.addAll(FRelatedList);
end;

procedure TFhirDocumentReferenceContext.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'event', 'CodeableConcept', FEventList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'facilityType', 'CodeableConcept', FFacilityType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'practiceSetting', 'CodeableConcept', FPracticeSetting.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sourcePatientInfo', 'Reference(Patient)', FSourcePatientInfo.Link));{2}
  oList.add(TFHIRProperty.create(self, 'related', '', FRelatedList.Link)){3};
end;

procedure TFhirDocumentReferenceContext.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'event') then EventList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'facilityType') then FacilityType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'practiceSetting') then PracticeSetting := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'sourcePatientInfo') then SourcePatientInfo := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'related') then RelatedList.add(propValue as TFhirDocumentReferenceContextRelated){2}
  else inherited;
end;

function TFhirDocumentReferenceContext.FhirType : string;
begin
  result := 'context';
end;

function TFhirDocumentReferenceContext.Link : TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext(inherited Link);
end;

function TFhirDocumentReferenceContext.Clone : TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext(inherited Clone);
end;

function TFhirDocumentReferenceContext.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContext;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContext)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContext(other);
    result := compareDeep(encounterElement, o.encounterElement, true) and compareDeep(eventList, o.eventList, true) and 
      compareDeep(periodElement, o.periodElement, true) and compareDeep(facilityTypeElement, o.facilityTypeElement, true) and 
      compareDeep(practiceSettingElement, o.practiceSettingElement, true) and compareDeep(sourcePatientInfoElement, o.sourcePatientInfoElement, true) and 
      compareDeep(relatedList, o.relatedList, true);
  end;
end;

function TFhirDocumentReferenceContext.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContext;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContext)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContext(other);
    result := true;
  end;
end;

{ TFhirDocumentReferenceContext }

Procedure TFhirDocumentReferenceContext.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Function TFhirDocumentReferenceContext.GetEventList : TFhirCodeableConceptList;
begin
  if FEventList = nil then
    FEventList := TFhirCodeableConceptList.Create;
  result := FEventList;
end;

Function TFhirDocumentReferenceContext.GetHasEventList : boolean;
begin
  result := (FEventList <> nil) and (FEventList.count > 0);
end;

Procedure TFhirDocumentReferenceContext.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirDocumentReferenceContext.SetFacilityType(value : TFhirCodeableConcept);
begin
  FFacilityType.free;
  FFacilityType := value;
end;

Procedure TFhirDocumentReferenceContext.SetPracticeSetting(value : TFhirCodeableConcept);
begin
  FPracticeSetting.free;
  FPracticeSetting := value;
end;

Procedure TFhirDocumentReferenceContext.SetSourcePatientInfo(value : TFhirReference{TFhirPatient});
begin
  FSourcePatientInfo.free;
  FSourcePatientInfo := value;
end;

Function TFhirDocumentReferenceContext.GetRelatedList : TFhirDocumentReferenceContextRelatedList;
begin
  if FRelatedList = nil then
    FRelatedList := TFhirDocumentReferenceContextRelatedList.Create;
  result := FRelatedList;
end;

Function TFhirDocumentReferenceContext.GetHasRelatedList : boolean;
begin
  result := (FRelatedList <> nil) and (FRelatedList.count > 0);
end;


{ TFhirDocumentReferenceContextListEnumerator }

Constructor TFhirDocumentReferenceContextListEnumerator.Create(list : TFhirDocumentReferenceContextList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentReferenceContextListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentReferenceContextListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentReferenceContextListEnumerator.GetCurrent : TFhirDocumentReferenceContext;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentReferenceContextList }
procedure TFhirDocumentReferenceContextList.AddItem(value: TFhirDocumentReferenceContext);
begin
  assert(value.ClassName = 'TFhirDocumentReferenceContext', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentReferenceContext');
  add(value);
end;


function TFhirDocumentReferenceContextList.Append: TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContextList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentReferenceContextList.GetEnumerator : TFhirDocumentReferenceContextListEnumerator;
begin
  result := TFhirDocumentReferenceContextListEnumerator.Create(self.link);
end;

function TFhirDocumentReferenceContextList.Clone: TFhirDocumentReferenceContextList;
begin
  result := TFhirDocumentReferenceContextList(inherited Clone);
end;

function TFhirDocumentReferenceContextList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentReferenceContextList.GetItemN(index: Integer): TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContextList.IndexOf(value: TFhirDocumentReferenceContext): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentReferenceContextList.Insert(index: Integer): TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContextList.InsertItem(index: Integer; value: TFhirDocumentReferenceContext);
begin
  assert(value is TFhirDocumentReferenceContext);
  Inherited Insert(index, value);
end;

function TFhirDocumentReferenceContextList.Item(index: Integer): TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContextList.Link: TFhirDocumentReferenceContextList;
begin
  result := TFhirDocumentReferenceContextList(inherited Link);
end;

procedure TFhirDocumentReferenceContextList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentReferenceContextList.SetItemByIndex(index: Integer; value: TFhirDocumentReferenceContext);
begin
  assert(value is TFhirDocumentReferenceContext);
  FhirDocumentReferenceContexts[index] := value;
end;

procedure TFhirDocumentReferenceContextList.SetItemN(index: Integer; value: TFhirDocumentReferenceContext);
begin
  assert(value is TFhirDocumentReferenceContext);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentReferenceContextRelated }

constructor TFhirDocumentReferenceContextRelated.Create;
begin
  inherited;
end;

destructor TFhirDocumentReferenceContextRelated.Destroy;
begin
  FIdentifier.free;
  FRef.free;
  inherited;
end;

procedure TFhirDocumentReferenceContextRelated.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirDocumentReferenceContextRelated(oSource).identifier.Clone;
  ref := TFhirDocumentReferenceContextRelated(oSource).ref.Clone;
end;

procedure TFhirDocumentReferenceContextRelated.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'ref') Then
     list.add(FRef.Link);
end;

procedure TFhirDocumentReferenceContextRelated.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ref', 'Reference(Any)', FRef.Link));{2}
end;

procedure TFhirDocumentReferenceContextRelated.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'ref') then Ref := propValue as TFhirReference{TFhirReference}{4b}
  else inherited;
end;

function TFhirDocumentReferenceContextRelated.FhirType : string;
begin
  result := 'related';
end;

function TFhirDocumentReferenceContextRelated.Link : TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated(inherited Link);
end;

function TFhirDocumentReferenceContextRelated.Clone : TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated(inherited Clone);
end;

function TFhirDocumentReferenceContextRelated.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContextRelated;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContextRelated)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContextRelated(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(refElement, o.refElement, true);
  end;
end;

function TFhirDocumentReferenceContextRelated.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirDocumentReferenceContextRelated;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirDocumentReferenceContextRelated)) then
    result := false
  else
  begin
    o := TFhirDocumentReferenceContextRelated(other);
    result := true;
  end;
end;

{ TFhirDocumentReferenceContextRelated }

Procedure TFhirDocumentReferenceContextRelated.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirDocumentReferenceContextRelated.SetRef(value : TFhirReference{TFhirReference});
begin
  FRef.free;
  FRef := value;
end;


{ TFhirDocumentReferenceContextRelatedListEnumerator }

Constructor TFhirDocumentReferenceContextRelatedListEnumerator.Create(list : TFhirDocumentReferenceContextRelatedList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentReferenceContextRelatedListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentReferenceContextRelatedListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentReferenceContextRelatedListEnumerator.GetCurrent : TFhirDocumentReferenceContextRelated;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentReferenceContextRelatedList }
procedure TFhirDocumentReferenceContextRelatedList.AddItem(value: TFhirDocumentReferenceContextRelated);
begin
  assert(value.ClassName = 'TFhirDocumentReferenceContextRelated', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentReferenceContextRelated');
  add(value);
end;


function TFhirDocumentReferenceContextRelatedList.Append: TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContextRelatedList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentReferenceContextRelatedList.GetEnumerator : TFhirDocumentReferenceContextRelatedListEnumerator;
begin
  result := TFhirDocumentReferenceContextRelatedListEnumerator.Create(self.link);
end;

function TFhirDocumentReferenceContextRelatedList.Clone: TFhirDocumentReferenceContextRelatedList;
begin
  result := TFhirDocumentReferenceContextRelatedList(inherited Clone);
end;

function TFhirDocumentReferenceContextRelatedList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentReferenceContextRelatedList.GetItemN(index: Integer): TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContextRelatedList.IndexOf(value: TFhirDocumentReferenceContextRelated): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentReferenceContextRelatedList.Insert(index: Integer): TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceContextRelatedList.InsertItem(index: Integer; value: TFhirDocumentReferenceContextRelated);
begin
  assert(value is TFhirDocumentReferenceContextRelated);
  Inherited Insert(index, value);
end;

function TFhirDocumentReferenceContextRelatedList.Item(index: Integer): TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceContextRelatedList.Link: TFhirDocumentReferenceContextRelatedList;
begin
  result := TFhirDocumentReferenceContextRelatedList(inherited Link);
end;

procedure TFhirDocumentReferenceContextRelatedList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentReferenceContextRelatedList.SetItemByIndex(index: Integer; value: TFhirDocumentReferenceContextRelated);
begin
  assert(value is TFhirDocumentReferenceContextRelated);
  FhirDocumentReferenceContextRelateds[index] := value;
end;

procedure TFhirDocumentReferenceContextRelatedList.SetItemN(index: Integer; value: TFhirDocumentReferenceContextRelated);
begin
  assert(value is TFhirDocumentReferenceContextRelated);
  ObjectByIndex[index] := value;
end;

{ TFhirDocumentReference }

constructor TFhirDocumentReference.Create;
begin
  inherited;
end;

destructor TFhirDocumentReference.Destroy;
begin
  FMasterIdentifier.free;
  FIdentifierList.Free;
  FSubject.free;
  FType_.free;
  FClass_.free;
  FAuthorList.Free;
  FCustodian.free;
  FAuthenticator.free;
  FCreated.free;
  FIndexed.free;
  FStatus.free;
  FDocStatus.free;
  FRelatesToList.Free;
  FDescription.free;
  FSecurityLabelList.Free;
  FContentList.Free;
  FContext.free;
  inherited;
end;

function TFhirDocumentReference.GetResourceType : TFhirResourceType;
begin
  result := frtDocumentReference;
end;

procedure TFhirDocumentReference.Assign(oSource : TAdvObject);
begin
  inherited;
  masterIdentifier := TFhirDocumentReference(oSource).masterIdentifier.Clone;
  if (TFhirDocumentReference(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirDocumentReference(oSource).FIdentifierList);
  end;
  subject := TFhirDocumentReference(oSource).subject.Clone;
  type_ := TFhirDocumentReference(oSource).type_.Clone;
  class_ := TFhirDocumentReference(oSource).class_.Clone;
  if (TFhirDocumentReference(oSource).FAuthorList = nil) then
  begin
    FAuthorList.free;
    FAuthorList := nil;
  end
  else
  begin
    FAuthorList := TFhirReferenceList{Resource}.Create;
    FAuthorList.Assign(TFhirDocumentReference(oSource).FAuthorList);
  end;
  custodian := TFhirDocumentReference(oSource).custodian.Clone;
  authenticator := TFhirDocumentReference(oSource).authenticator.Clone;
  createdElement := TFhirDocumentReference(oSource).createdElement.Clone;
  indexedElement := TFhirDocumentReference(oSource).indexedElement.Clone;
  FStatus := TFhirDocumentReference(oSource).FStatus.Link;
  docStatus := TFhirDocumentReference(oSource).docStatus.Clone;
  if (TFhirDocumentReference(oSource).FRelatesToList = nil) then
  begin
    FRelatesToList.free;
    FRelatesToList := nil;
  end
  else
  begin
    FRelatesToList := TFhirDocumentReferenceRelatesToList.Create;
    FRelatesToList.Assign(TFhirDocumentReference(oSource).FRelatesToList);
  end;
  descriptionElement := TFhirDocumentReference(oSource).descriptionElement.Clone;
  if (TFhirDocumentReference(oSource).FSecurityLabelList = nil) then
  begin
    FSecurityLabelList.free;
    FSecurityLabelList := nil;
  end
  else
  begin
    FSecurityLabelList := TFhirCodeableConceptList.Create;
    FSecurityLabelList.Assign(TFhirDocumentReference(oSource).FSecurityLabelList);
  end;
  if (TFhirDocumentReference(oSource).FContentList = nil) then
  begin
    FContentList.free;
    FContentList := nil;
  end
  else
  begin
    FContentList := TFhirDocumentReferenceContentList.Create;
    FContentList.Assign(TFhirDocumentReference(oSource).FContentList);
  end;
  context := TFhirDocumentReference(oSource).context.Clone;
end;

procedure TFhirDocumentReference.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'masterIdentifier') Then
     list.add(FMasterIdentifier.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'class') Then
     list.add(FClass_.Link);
  if (child_name = 'author') Then
    list.addAll(FAuthorList);
  if (child_name = 'custodian') Then
     list.add(FCustodian.Link);
  if (child_name = 'authenticator') Then
     list.add(FAuthenticator.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'indexed') Then
     list.add(FIndexed.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'docStatus') Then
     list.add(FDocStatus.Link);
  if (child_name = 'relatesTo') Then
    list.addAll(FRelatesToList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'securityLabel') Then
    list.addAll(FSecurityLabelList);
  if (child_name = 'content') Then
    list.addAll(FContentList);
  if (child_name = 'context') Then
     list.add(FContext.Link);
end;

procedure TFhirDocumentReference.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'masterIdentifier', 'Identifier', FMasterIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Practitioner|Group|Device)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'class', 'CodeableConcept', FClass_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Organization|Device|Patient|RelatedPerson)', FAuthorList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'custodian', 'Reference(Organization)', FCustodian.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authenticator', 'Reference(Practitioner|Organization)', FAuthenticator.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'indexed', 'instant', FIndexed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'docStatus', 'CodeableConcept', FDocStatus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relatesTo', '', FRelatesToList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'securityLabel', 'CodeableConcept', FSecurityLabelList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'content', '', FContentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'context', '', FContext.Link));{2}
end;

procedure TFhirDocumentReference.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'masterIdentifier') then MasterIdentifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'class') then Class_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'author') then AuthorList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'custodian') then Custodian := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'authenticator') then Authenticator := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'indexed') then IndexedElement := propValue as TFhirInstant{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'docStatus') then DocStatus := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'relatesTo') then RelatesToList.add(propValue as TFhirDocumentReferenceRelatesTo){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'securityLabel') then SecurityLabelList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'content') then ContentList.add(propValue as TFhirDocumentReferenceContent){2}
  else if (propName = 'context') then Context := propValue as TFhirDocumentReferenceContext{4b}
  else inherited;
end;

function TFhirDocumentReference.FhirType : string;
begin
  result := 'DocumentReference';
end;

function TFhirDocumentReference.Link : TFhirDocumentReference;
begin
  result := TFhirDocumentReference(inherited Link);
end;

function TFhirDocumentReference.Clone : TFhirDocumentReference;
begin
  result := TFhirDocumentReference(inherited Clone);
end;

{ TFhirDocumentReference }

Procedure TFhirDocumentReference.SetMasterIdentifier(value : TFhirIdentifier);
begin
  FMasterIdentifier.free;
  FMasterIdentifier := value;
end;

Function TFhirDocumentReference.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirDocumentReference.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirDocumentReference.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirDocumentReference.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirDocumentReference.SetClass_(value : TFhirCodeableConcept);
begin
  FClass_.free;
  FClass_ := value;
end;

Function TFhirDocumentReference.GetAuthorList : TFhirReferenceList{Resource};
begin
  if FAuthorList = nil then
    FAuthorList := TFhirReferenceList{Resource}.Create;
  result := FAuthorList;
end;

Function TFhirDocumentReference.GetHasAuthorList : boolean;
begin
  result := (FAuthorList <> nil) and (FAuthorList.count > 0);
end;

Procedure TFhirDocumentReference.SetCustodian(value : TFhirReference{TFhirOrganization});
begin
  FCustodian.free;
  FCustodian := value;
end;

Procedure TFhirDocumentReference.SetAuthenticator(value : TFhirReference{Resource});
begin
  FAuthenticator.free;
  FAuthenticator := value;
end;

Procedure TFhirDocumentReference.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirDocumentReference.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirDocumentReference.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirDocumentReference.SetIndexed(value : TFhirInstant);
begin
  FIndexed.free;
  FIndexed := value;
end;

Function TFhirDocumentReference.GetIndexedST : TDateAndTime;
begin
  if FIndexed = nil then
    result := nil
  else
    result := FIndexed.value;
end;

Procedure TFhirDocumentReference.SetIndexedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIndexed = nil then
      FIndexed := TFhirInstant.create;
    FIndexed.value := value
  end
  else if FIndexed <> nil then
    FIndexed.value := nil;
end;

Procedure TFhirDocumentReference.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirDocumentReference.GetStatusST : TFhirDocumentReferenceStatus;
begin
  if FStatus = nil then
    result := TFhirDocumentReferenceStatus(0)
  else
    result := TFhirDocumentReferenceStatus(StringArrayIndexOfSensitive(CODES_TFhirDocumentReferenceStatus, FStatus.value));
end;

Procedure TFhirDocumentReference.SetStatusST(value : TFhirDocumentReferenceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirDocumentReferenceStatus[value]);
end;

Procedure TFhirDocumentReference.SetDocStatus(value : TFhirCodeableConcept);
begin
  FDocStatus.free;
  FDocStatus := value;
end;

Function TFhirDocumentReference.GetRelatesToList : TFhirDocumentReferenceRelatesToList;
begin
  if FRelatesToList = nil then
    FRelatesToList := TFhirDocumentReferenceRelatesToList.Create;
  result := FRelatesToList;
end;

Function TFhirDocumentReference.GetHasRelatesToList : boolean;
begin
  result := (FRelatesToList <> nil) and (FRelatesToList.count > 0);
end;

Procedure TFhirDocumentReference.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirDocumentReference.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirDocumentReference.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirDocumentReference.GetSecurityLabelList : TFhirCodeableConceptList;
begin
  if FSecurityLabelList = nil then
    FSecurityLabelList := TFhirCodeableConceptList.Create;
  result := FSecurityLabelList;
end;

Function TFhirDocumentReference.GetHasSecurityLabelList : boolean;
begin
  result := (FSecurityLabelList <> nil) and (FSecurityLabelList.count > 0);
end;

Function TFhirDocumentReference.GetContentList : TFhirDocumentReferenceContentList;
begin
  if FContentList = nil then
    FContentList := TFhirDocumentReferenceContentList.Create;
  result := FContentList;
end;

Function TFhirDocumentReference.GetHasContentList : boolean;
begin
  result := (FContentList <> nil) and (FContentList.count > 0);
end;

Procedure TFhirDocumentReference.SetContext(value : TFhirDocumentReferenceContext);
begin
  FContext.free;
  FContext := value;
end;


{ TFhirDocumentReferenceListEnumerator }

Constructor TFhirDocumentReferenceListEnumerator.Create(list : TFhirDocumentReferenceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirDocumentReferenceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirDocumentReferenceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirDocumentReferenceListEnumerator.GetCurrent : TFhirDocumentReference;
begin
  Result := FList[FIndex];
end;


{ TFhirDocumentReferenceList }
procedure TFhirDocumentReferenceList.AddItem(value: TFhirDocumentReference);
begin
  assert(value.ClassName = 'TFhirDocumentReference', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirDocumentReference');
  add(value);
end;


function TFhirDocumentReferenceList.Append: TFhirDocumentReference;
begin
  result := TFhirDocumentReference.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceList.ClearItems;
begin
  Clear;
end;

function TFhirDocumentReferenceList.GetEnumerator : TFhirDocumentReferenceListEnumerator;
begin
  result := TFhirDocumentReferenceListEnumerator.Create(self.link);
end;

function TFhirDocumentReferenceList.Clone: TFhirDocumentReferenceList;
begin
  result := TFhirDocumentReferenceList(inherited Clone);
end;

function TFhirDocumentReferenceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirDocumentReferenceList.GetItemN(index: Integer): TFhirDocumentReference;
begin
  result := TFhirDocumentReference(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceList.IndexOf(value: TFhirDocumentReference): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirDocumentReferenceList.Insert(index: Integer): TFhirDocumentReference;
begin
  result := TFhirDocumentReference.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirDocumentReferenceList.InsertItem(index: Integer; value: TFhirDocumentReference);
begin
  assert(value is TFhirDocumentReference);
  Inherited Insert(index, value);
end;

function TFhirDocumentReferenceList.Item(index: Integer): TFhirDocumentReference;
begin
  result := TFhirDocumentReference(ObjectByIndex[index]);
end;

function TFhirDocumentReferenceList.Link: TFhirDocumentReferenceList;
begin
  result := TFhirDocumentReferenceList(inherited Link);
end;

procedure TFhirDocumentReferenceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirDocumentReferenceList.SetItemByIndex(index: Integer; value: TFhirDocumentReference);
begin
  assert(value is TFhirDocumentReference);
  FhirDocumentReferences[index] := value;
end;

procedure TFhirDocumentReferenceList.SetItemN(index: Integer; value: TFhirDocumentReference);
begin
  assert(value is TFhirDocumentReference);
  ObjectByIndex[index] := value;
end;

{ TFhirEligibilityRequest }

constructor TFhirEligibilityRequest.Create;
begin
  inherited;
end;

destructor TFhirEligibilityRequest.Destroy;
begin
  FIdentifierList.Free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FTarget.free;
  FProvider.free;
  FOrganization.free;
  inherited;
end;

function TFhirEligibilityRequest.GetResourceType : TFhirResourceType;
begin
  result := frtEligibilityRequest;
end;

procedure TFhirEligibilityRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEligibilityRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEligibilityRequest(oSource).FIdentifierList);
  end;
  ruleset := TFhirEligibilityRequest(oSource).ruleset.Clone;
  originalRuleset := TFhirEligibilityRequest(oSource).originalRuleset.Clone;
  createdElement := TFhirEligibilityRequest(oSource).createdElement.Clone;
  target := TFhirEligibilityRequest(oSource).target.Clone;
  provider := TFhirEligibilityRequest(oSource).provider.Clone;
  organization := TFhirEligibilityRequest(oSource).organization.Clone;
end;

procedure TFhirEligibilityRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
end;

procedure TFhirEligibilityRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
end;

procedure TFhirEligibilityRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else inherited;
end;

function TFhirEligibilityRequest.FhirType : string;
begin
  result := 'EligibilityRequest';
end;

function TFhirEligibilityRequest.Link : TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest(inherited Link);
end;

function TFhirEligibilityRequest.Clone : TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest(inherited Clone);
end;

{ TFhirEligibilityRequest }

Function TFhirEligibilityRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEligibilityRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEligibilityRequest.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirEligibilityRequest.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirEligibilityRequest.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirEligibilityRequest.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirEligibilityRequest.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirEligibilityRequest.SetTarget(value : TFhirReference{TFhirOrganization});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirEligibilityRequest.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirEligibilityRequest.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;


{ TFhirEligibilityRequestListEnumerator }

Constructor TFhirEligibilityRequestListEnumerator.Create(list : TFhirEligibilityRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEligibilityRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEligibilityRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEligibilityRequestListEnumerator.GetCurrent : TFhirEligibilityRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirEligibilityRequestList }
procedure TFhirEligibilityRequestList.AddItem(value: TFhirEligibilityRequest);
begin
  assert(value.ClassName = 'TFhirEligibilityRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEligibilityRequest');
  add(value);
end;


function TFhirEligibilityRequestList.Append: TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEligibilityRequestList.ClearItems;
begin
  Clear;
end;

function TFhirEligibilityRequestList.GetEnumerator : TFhirEligibilityRequestListEnumerator;
begin
  result := TFhirEligibilityRequestListEnumerator.Create(self.link);
end;

function TFhirEligibilityRequestList.Clone: TFhirEligibilityRequestList;
begin
  result := TFhirEligibilityRequestList(inherited Clone);
end;

function TFhirEligibilityRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEligibilityRequestList.GetItemN(index: Integer): TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest(ObjectByIndex[index]);
end;

function TFhirEligibilityRequestList.IndexOf(value: TFhirEligibilityRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEligibilityRequestList.Insert(index: Integer): TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEligibilityRequestList.InsertItem(index: Integer; value: TFhirEligibilityRequest);
begin
  assert(value is TFhirEligibilityRequest);
  Inherited Insert(index, value);
end;

function TFhirEligibilityRequestList.Item(index: Integer): TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest(ObjectByIndex[index]);
end;

function TFhirEligibilityRequestList.Link: TFhirEligibilityRequestList;
begin
  result := TFhirEligibilityRequestList(inherited Link);
end;

procedure TFhirEligibilityRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEligibilityRequestList.SetItemByIndex(index: Integer; value: TFhirEligibilityRequest);
begin
  assert(value is TFhirEligibilityRequest);
  FhirEligibilityRequests[index] := value;
end;

procedure TFhirEligibilityRequestList.SetItemN(index: Integer; value: TFhirEligibilityRequest);
begin
  assert(value is TFhirEligibilityRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirEligibilityResponse }

constructor TFhirEligibilityResponse.Create;
begin
  inherited;
end;

destructor TFhirEligibilityResponse.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FOutcome.free;
  FDisposition.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  inherited;
end;

function TFhirEligibilityResponse.GetResourceType : TFhirResourceType;
begin
  result := frtEligibilityResponse;
end;

procedure TFhirEligibilityResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEligibilityResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEligibilityResponse(oSource).FIdentifierList);
  end;
  request := TFhirEligibilityResponse(oSource).request.Clone;
  FOutcome := TFhirEligibilityResponse(oSource).FOutcome.Link;
  dispositionElement := TFhirEligibilityResponse(oSource).dispositionElement.Clone;
  ruleset := TFhirEligibilityResponse(oSource).ruleset.Clone;
  originalRuleset := TFhirEligibilityResponse(oSource).originalRuleset.Clone;
  createdElement := TFhirEligibilityResponse(oSource).createdElement.Clone;
  organization := TFhirEligibilityResponse(oSource).organization.Clone;
  requestProvider := TFhirEligibilityResponse(oSource).requestProvider.Clone;
  requestOrganization := TFhirEligibilityResponse(oSource).requestOrganization.Clone;
end;

procedure TFhirEligibilityResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
end;

procedure TFhirEligibilityResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(EligibilityRequest)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
end;

procedure TFhirEligibilityResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirEligibilityRequest}{4b}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else inherited;
end;

function TFhirEligibilityResponse.FhirType : string;
begin
  result := 'EligibilityResponse';
end;

function TFhirEligibilityResponse.Link : TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse(inherited Link);
end;

function TFhirEligibilityResponse.Clone : TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse(inherited Clone);
end;

{ TFhirEligibilityResponse }

Function TFhirEligibilityResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEligibilityResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEligibilityResponse.SetRequest(value : TFhirReference{TFhirEligibilityRequest});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirEligibilityResponse.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirEligibilityResponse.GetOutcomeST : TFhirRemittanceOutcome;
begin
  if FOutcome = nil then
    result := TFhirRemittanceOutcome(0)
  else
    result := TFhirRemittanceOutcome(StringArrayIndexOfSensitive(CODES_TFhirRemittanceOutcome, FOutcome.value));
end;

Procedure TFhirEligibilityResponse.SetOutcomeST(value : TFhirRemittanceOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirRemittanceOutcome[value]);
end;

Procedure TFhirEligibilityResponse.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirEligibilityResponse.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirEligibilityResponse.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirEligibilityResponse.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirEligibilityResponse.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirEligibilityResponse.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirEligibilityResponse.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirEligibilityResponse.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirEligibilityResponse.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirEligibilityResponse.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirEligibilityResponse.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;


{ TFhirEligibilityResponseListEnumerator }

Constructor TFhirEligibilityResponseListEnumerator.Create(list : TFhirEligibilityResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEligibilityResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEligibilityResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEligibilityResponseListEnumerator.GetCurrent : TFhirEligibilityResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirEligibilityResponseList }
procedure TFhirEligibilityResponseList.AddItem(value: TFhirEligibilityResponse);
begin
  assert(value.ClassName = 'TFhirEligibilityResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEligibilityResponse');
  add(value);
end;


function TFhirEligibilityResponseList.Append: TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEligibilityResponseList.ClearItems;
begin
  Clear;
end;

function TFhirEligibilityResponseList.GetEnumerator : TFhirEligibilityResponseListEnumerator;
begin
  result := TFhirEligibilityResponseListEnumerator.Create(self.link);
end;

function TFhirEligibilityResponseList.Clone: TFhirEligibilityResponseList;
begin
  result := TFhirEligibilityResponseList(inherited Clone);
end;

function TFhirEligibilityResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEligibilityResponseList.GetItemN(index: Integer): TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse(ObjectByIndex[index]);
end;

function TFhirEligibilityResponseList.IndexOf(value: TFhirEligibilityResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEligibilityResponseList.Insert(index: Integer): TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEligibilityResponseList.InsertItem(index: Integer; value: TFhirEligibilityResponse);
begin
  assert(value is TFhirEligibilityResponse);
  Inherited Insert(index, value);
end;

function TFhirEligibilityResponseList.Item(index: Integer): TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse(ObjectByIndex[index]);
end;

function TFhirEligibilityResponseList.Link: TFhirEligibilityResponseList;
begin
  result := TFhirEligibilityResponseList(inherited Link);
end;

procedure TFhirEligibilityResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEligibilityResponseList.SetItemByIndex(index: Integer; value: TFhirEligibilityResponse);
begin
  assert(value is TFhirEligibilityResponse);
  FhirEligibilityResponses[index] := value;
end;

procedure TFhirEligibilityResponseList.SetItemN(index: Integer; value: TFhirEligibilityResponse);
begin
  assert(value is TFhirEligibilityResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirEncounterStatusHistory }

constructor TFhirEncounterStatusHistory.Create;
begin
  inherited;
end;

destructor TFhirEncounterStatusHistory.Destroy;
begin
  FStatus.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirEncounterStatusHistory.Assign(oSource : TAdvObject);
begin
  inherited;
  FStatus := TFhirEncounterStatusHistory(oSource).FStatus.Link;
  period := TFhirEncounterStatusHistory(oSource).period.Clone;
end;

procedure TFhirEncounterStatusHistory.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirEncounterStatusHistory.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirEncounterStatusHistory.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirEncounterStatusHistory.FhirType : string;
begin
  result := 'statusHistory';
end;

function TFhirEncounterStatusHistory.Link : TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory(inherited Link);
end;

function TFhirEncounterStatusHistory.Clone : TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory(inherited Clone);
end;

function TFhirEncounterStatusHistory.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterStatusHistory;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEncounterStatusHistory)) then
    result := false
  else
  begin
    o := TFhirEncounterStatusHistory(other);
    result := compareDeep(statusElement, o.statusElement, true) and compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirEncounterStatusHistory.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterStatusHistory;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEncounterStatusHistory)) then
    result := false
  else
  begin
    o := TFhirEncounterStatusHistory(other);
    result := compareValues(statusElement, o.statusElement, true);
  end;
end;

{ TFhirEncounterStatusHistory }

Procedure TFhirEncounterStatusHistory.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirEncounterStatusHistory.GetStatusST : TFhirEncounterState;
begin
  if FStatus = nil then
    result := TFhirEncounterState(0)
  else
    result := TFhirEncounterState(StringArrayIndexOfSensitive(CODES_TFhirEncounterState, FStatus.value));
end;

Procedure TFhirEncounterStatusHistory.SetStatusST(value : TFhirEncounterState);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirEncounterState[value]);
end;

Procedure TFhirEncounterStatusHistory.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirEncounterStatusHistoryListEnumerator }

Constructor TFhirEncounterStatusHistoryListEnumerator.Create(list : TFhirEncounterStatusHistoryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEncounterStatusHistoryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEncounterStatusHistoryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEncounterStatusHistoryListEnumerator.GetCurrent : TFhirEncounterStatusHistory;
begin
  Result := FList[FIndex];
end;


{ TFhirEncounterStatusHistoryList }
procedure TFhirEncounterStatusHistoryList.AddItem(value: TFhirEncounterStatusHistory);
begin
  assert(value.ClassName = 'TFhirEncounterStatusHistory', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEncounterStatusHistory');
  add(value);
end;


function TFhirEncounterStatusHistoryList.Append: TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterStatusHistoryList.ClearItems;
begin
  Clear;
end;

function TFhirEncounterStatusHistoryList.GetEnumerator : TFhirEncounterStatusHistoryListEnumerator;
begin
  result := TFhirEncounterStatusHistoryListEnumerator.Create(self.link);
end;

function TFhirEncounterStatusHistoryList.Clone: TFhirEncounterStatusHistoryList;
begin
  result := TFhirEncounterStatusHistoryList(inherited Clone);
end;

function TFhirEncounterStatusHistoryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEncounterStatusHistoryList.GetItemN(index: Integer): TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory(ObjectByIndex[index]);
end;

function TFhirEncounterStatusHistoryList.IndexOf(value: TFhirEncounterStatusHistory): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEncounterStatusHistoryList.Insert(index: Integer): TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterStatusHistoryList.InsertItem(index: Integer; value: TFhirEncounterStatusHistory);
begin
  assert(value is TFhirEncounterStatusHistory);
  Inherited Insert(index, value);
end;

function TFhirEncounterStatusHistoryList.Item(index: Integer): TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory(ObjectByIndex[index]);
end;

function TFhirEncounterStatusHistoryList.Link: TFhirEncounterStatusHistoryList;
begin
  result := TFhirEncounterStatusHistoryList(inherited Link);
end;

procedure TFhirEncounterStatusHistoryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEncounterStatusHistoryList.SetItemByIndex(index: Integer; value: TFhirEncounterStatusHistory);
begin
  assert(value is TFhirEncounterStatusHistory);
  FhirEncounterStatusHistories[index] := value;
end;

procedure TFhirEncounterStatusHistoryList.SetItemN(index: Integer; value: TFhirEncounterStatusHistory);
begin
  assert(value is TFhirEncounterStatusHistory);
  ObjectByIndex[index] := value;
end;

{ TFhirEncounterParticipant }

constructor TFhirEncounterParticipant.Create;
begin
  inherited;
end;

destructor TFhirEncounterParticipant.Destroy;
begin
  FType_List.Free;
  FPeriod.free;
  FIndividual.free;
  inherited;
end;

procedure TFhirEncounterParticipant.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEncounterParticipant(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirEncounterParticipant(oSource).FType_List);
  end;
  period := TFhirEncounterParticipant(oSource).period.Clone;
  individual := TFhirEncounterParticipant(oSource).individual.Clone;
end;

procedure TFhirEncounterParticipant.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'individual') Then
     list.add(FIndividual.Link);
end;

procedure TFhirEncounterParticipant.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'individual', 'Reference(Practitioner|RelatedPerson)', FIndividual.Link));{2}
end;

procedure TFhirEncounterParticipant.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'individual') then Individual := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirEncounterParticipant.FhirType : string;
begin
  result := 'participant';
end;

function TFhirEncounterParticipant.Link : TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant(inherited Link);
end;

function TFhirEncounterParticipant.Clone : TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant(inherited Clone);
end;

function TFhirEncounterParticipant.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterParticipant;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEncounterParticipant)) then
    result := false
  else
  begin
    o := TFhirEncounterParticipant(other);
    result := compareDeep(type_List, o.type_List, true) and compareDeep(periodElement, o.periodElement, true) and 
      compareDeep(individualElement, o.individualElement, true);
  end;
end;

function TFhirEncounterParticipant.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterParticipant;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEncounterParticipant)) then
    result := false
  else
  begin
    o := TFhirEncounterParticipant(other);
    result := true;
  end;
end;

{ TFhirEncounterParticipant }

Function TFhirEncounterParticipant.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirEncounterParticipant.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Procedure TFhirEncounterParticipant.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirEncounterParticipant.SetIndividual(value : TFhirReference{Resource});
begin
  FIndividual.free;
  FIndividual := value;
end;


{ TFhirEncounterParticipantListEnumerator }

Constructor TFhirEncounterParticipantListEnumerator.Create(list : TFhirEncounterParticipantList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEncounterParticipantListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEncounterParticipantListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEncounterParticipantListEnumerator.GetCurrent : TFhirEncounterParticipant;
begin
  Result := FList[FIndex];
end;


{ TFhirEncounterParticipantList }
procedure TFhirEncounterParticipantList.AddItem(value: TFhirEncounterParticipant);
begin
  assert(value.ClassName = 'TFhirEncounterParticipant', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEncounterParticipant');
  add(value);
end;


function TFhirEncounterParticipantList.Append: TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterParticipantList.ClearItems;
begin
  Clear;
end;

function TFhirEncounterParticipantList.GetEnumerator : TFhirEncounterParticipantListEnumerator;
begin
  result := TFhirEncounterParticipantListEnumerator.Create(self.link);
end;

function TFhirEncounterParticipantList.Clone: TFhirEncounterParticipantList;
begin
  result := TFhirEncounterParticipantList(inherited Clone);
end;

function TFhirEncounterParticipantList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEncounterParticipantList.GetItemN(index: Integer): TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant(ObjectByIndex[index]);
end;

function TFhirEncounterParticipantList.IndexOf(value: TFhirEncounterParticipant): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEncounterParticipantList.Insert(index: Integer): TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterParticipantList.InsertItem(index: Integer; value: TFhirEncounterParticipant);
begin
  assert(value is TFhirEncounterParticipant);
  Inherited Insert(index, value);
end;

function TFhirEncounterParticipantList.Item(index: Integer): TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant(ObjectByIndex[index]);
end;

function TFhirEncounterParticipantList.Link: TFhirEncounterParticipantList;
begin
  result := TFhirEncounterParticipantList(inherited Link);
end;

procedure TFhirEncounterParticipantList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEncounterParticipantList.SetItemByIndex(index: Integer; value: TFhirEncounterParticipant);
begin
  assert(value is TFhirEncounterParticipant);
  FhirEncounterParticipants[index] := value;
end;

procedure TFhirEncounterParticipantList.SetItemN(index: Integer; value: TFhirEncounterParticipant);
begin
  assert(value is TFhirEncounterParticipant);
  ObjectByIndex[index] := value;
end;

{ TFhirEncounterHospitalization }

constructor TFhirEncounterHospitalization.Create;
begin
  inherited;
end;

destructor TFhirEncounterHospitalization.Destroy;
begin
  FPreAdmissionIdentifier.free;
  FOrigin.free;
  FAdmitSource.free;
  FAdmittingDiagnosisList.Free;
  FReAdmission.free;
  FDietPreferenceList.Free;
  FSpecialCourtesyList.Free;
  FSpecialArrangementList.Free;
  FDestination.free;
  FDischargeDisposition.free;
  FDischargeDiagnosisList.Free;
  inherited;
end;

procedure TFhirEncounterHospitalization.Assign(oSource : TAdvObject);
begin
  inherited;
  preAdmissionIdentifier := TFhirEncounterHospitalization(oSource).preAdmissionIdentifier.Clone;
  origin := TFhirEncounterHospitalization(oSource).origin.Clone;
  admitSource := TFhirEncounterHospitalization(oSource).admitSource.Clone;
  if (TFhirEncounterHospitalization(oSource).FAdmittingDiagnosisList = nil) then
  begin
    FAdmittingDiagnosisList.free;
    FAdmittingDiagnosisList := nil;
  end
  else
  begin
    FAdmittingDiagnosisList := TFhirReferenceList{TFhirCondition}.Create;
    FAdmittingDiagnosisList.Assign(TFhirEncounterHospitalization(oSource).FAdmittingDiagnosisList);
  end;
  reAdmission := TFhirEncounterHospitalization(oSource).reAdmission.Clone;
  if (TFhirEncounterHospitalization(oSource).FDietPreferenceList = nil) then
  begin
    FDietPreferenceList.free;
    FDietPreferenceList := nil;
  end
  else
  begin
    FDietPreferenceList := TFhirCodeableConceptList.Create;
    FDietPreferenceList.Assign(TFhirEncounterHospitalization(oSource).FDietPreferenceList);
  end;
  if (TFhirEncounterHospitalization(oSource).FSpecialCourtesyList = nil) then
  begin
    FSpecialCourtesyList.free;
    FSpecialCourtesyList := nil;
  end
  else
  begin
    FSpecialCourtesyList := TFhirCodeableConceptList.Create;
    FSpecialCourtesyList.Assign(TFhirEncounterHospitalization(oSource).FSpecialCourtesyList);
  end;
  if (TFhirEncounterHospitalization(oSource).FSpecialArrangementList = nil) then
  begin
    FSpecialArrangementList.free;
    FSpecialArrangementList := nil;
  end
  else
  begin
    FSpecialArrangementList := TFhirCodeableConceptList.Create;
    FSpecialArrangementList.Assign(TFhirEncounterHospitalization(oSource).FSpecialArrangementList);
  end;
  destination := TFhirEncounterHospitalization(oSource).destination.Clone;
  dischargeDisposition := TFhirEncounterHospitalization(oSource).dischargeDisposition.Clone;
  if (TFhirEncounterHospitalization(oSource).FDischargeDiagnosisList = nil) then
  begin
    FDischargeDiagnosisList.free;
    FDischargeDiagnosisList := nil;
  end
  else
  begin
    FDischargeDiagnosisList := TFhirReferenceList{TFhirCondition}.Create;
    FDischargeDiagnosisList.Assign(TFhirEncounterHospitalization(oSource).FDischargeDiagnosisList);
  end;
end;

procedure TFhirEncounterHospitalization.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'preAdmissionIdentifier') Then
     list.add(FPreAdmissionIdentifier.Link);
  if (child_name = 'origin') Then
     list.add(FOrigin.Link);
  if (child_name = 'admitSource') Then
     list.add(FAdmitSource.Link);
  if (child_name = 'admittingDiagnosis') Then
    list.addAll(FAdmittingDiagnosisList);
  if (child_name = 'reAdmission') Then
     list.add(FReAdmission.Link);
  if (child_name = 'dietPreference') Then
    list.addAll(FDietPreferenceList);
  if (child_name = 'specialCourtesy') Then
    list.addAll(FSpecialCourtesyList);
  if (child_name = 'specialArrangement') Then
    list.addAll(FSpecialArrangementList);
  if (child_name = 'destination') Then
     list.add(FDestination.Link);
  if (child_name = 'dischargeDisposition') Then
     list.add(FDischargeDisposition.Link);
  if (child_name = 'dischargeDiagnosis') Then
    list.addAll(FDischargeDiagnosisList);
end;

procedure TFhirEncounterHospitalization.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'preAdmissionIdentifier', 'Identifier', FPreAdmissionIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'origin', 'Reference(Location)', FOrigin.Link));{2}
  oList.add(TFHIRProperty.create(self, 'admitSource', 'CodeableConcept', FAdmitSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'admittingDiagnosis', 'Reference(Condition)', FAdmittingDiagnosisList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reAdmission', 'CodeableConcept', FReAdmission.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dietPreference', 'CodeableConcept', FDietPreferenceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'specialCourtesy', 'CodeableConcept', FSpecialCourtesyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'specialArrangement', 'CodeableConcept', FSpecialArrangementList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'destination', 'Reference(Location)', FDestination.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dischargeDisposition', 'CodeableConcept', FDischargeDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dischargeDiagnosis', 'Reference(Condition)', FDischargeDiagnosisList.Link)){3};
end;

procedure TFhirEncounterHospitalization.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'preAdmissionIdentifier') then PreAdmissionIdentifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'origin') then Origin := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'admitSource') then AdmitSource := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'admittingDiagnosis') then AdmittingDiagnosisList.add(propValue as TFhirReference{TFhirCondition}){2}
  else if (propName = 'reAdmission') then ReAdmission := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'dietPreference') then DietPreferenceList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'specialCourtesy') then SpecialCourtesyList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'specialArrangement') then SpecialArrangementList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'destination') then Destination := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'dischargeDisposition') then DischargeDisposition := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'dischargeDiagnosis') then DischargeDiagnosisList.add(propValue as TFhirReference{TFhirCondition}){2}
  else inherited;
end;

function TFhirEncounterHospitalization.FhirType : string;
begin
  result := 'hospitalization';
end;

function TFhirEncounterHospitalization.Link : TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization(inherited Link);
end;

function TFhirEncounterHospitalization.Clone : TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization(inherited Clone);
end;

function TFhirEncounterHospitalization.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterHospitalization;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEncounterHospitalization)) then
    result := false
  else
  begin
    o := TFhirEncounterHospitalization(other);
    result := compareDeep(preAdmissionIdentifierElement, o.preAdmissionIdentifierElement, true) and 
      compareDeep(originElement, o.originElement, true) and compareDeep(admitSourceElement, o.admitSourceElement, true) and 
      compareDeep(admittingDiagnosisList, o.admittingDiagnosisList, true) and compareDeep(reAdmissionElement, o.reAdmissionElement, true) and 
      compareDeep(dietPreferenceList, o.dietPreferenceList, true) and compareDeep(specialCourtesyList, o.specialCourtesyList, true) and 
      compareDeep(specialArrangementList, o.specialArrangementList, true) and compareDeep(destinationElement, o.destinationElement, true) and 
      compareDeep(dischargeDispositionElement, o.dischargeDispositionElement, true) and 
      compareDeep(dischargeDiagnosisList, o.dischargeDiagnosisList, true);
  end;
end;

function TFhirEncounterHospitalization.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterHospitalization;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEncounterHospitalization)) then
    result := false
  else
  begin
    o := TFhirEncounterHospitalization(other);
    result := true;
  end;
end;

{ TFhirEncounterHospitalization }

Procedure TFhirEncounterHospitalization.SetPreAdmissionIdentifier(value : TFhirIdentifier);
begin
  FPreAdmissionIdentifier.free;
  FPreAdmissionIdentifier := value;
end;

Procedure TFhirEncounterHospitalization.SetOrigin(value : TFhirReference{TFhirLocation});
begin
  FOrigin.free;
  FOrigin := value;
end;

Procedure TFhirEncounterHospitalization.SetAdmitSource(value : TFhirCodeableConcept);
begin
  FAdmitSource.free;
  FAdmitSource := value;
end;

Function TFhirEncounterHospitalization.GetAdmittingDiagnosisList : TFhirReferenceList{TFhirCondition};
begin
  if FAdmittingDiagnosisList = nil then
    FAdmittingDiagnosisList := TFhirReferenceList{TFhirCondition}.Create;
  result := FAdmittingDiagnosisList;
end;

Function TFhirEncounterHospitalization.GetHasAdmittingDiagnosisList : boolean;
begin
  result := (FAdmittingDiagnosisList <> nil) and (FAdmittingDiagnosisList.count > 0);
end;

Procedure TFhirEncounterHospitalization.SetReAdmission(value : TFhirCodeableConcept);
begin
  FReAdmission.free;
  FReAdmission := value;
end;

Function TFhirEncounterHospitalization.GetDietPreferenceList : TFhirCodeableConceptList;
begin
  if FDietPreferenceList = nil then
    FDietPreferenceList := TFhirCodeableConceptList.Create;
  result := FDietPreferenceList;
end;

Function TFhirEncounterHospitalization.GetHasDietPreferenceList : boolean;
begin
  result := (FDietPreferenceList <> nil) and (FDietPreferenceList.count > 0);
end;

Function TFhirEncounterHospitalization.GetSpecialCourtesyList : TFhirCodeableConceptList;
begin
  if FSpecialCourtesyList = nil then
    FSpecialCourtesyList := TFhirCodeableConceptList.Create;
  result := FSpecialCourtesyList;
end;

Function TFhirEncounterHospitalization.GetHasSpecialCourtesyList : boolean;
begin
  result := (FSpecialCourtesyList <> nil) and (FSpecialCourtesyList.count > 0);
end;

Function TFhirEncounterHospitalization.GetSpecialArrangementList : TFhirCodeableConceptList;
begin
  if FSpecialArrangementList = nil then
    FSpecialArrangementList := TFhirCodeableConceptList.Create;
  result := FSpecialArrangementList;
end;

Function TFhirEncounterHospitalization.GetHasSpecialArrangementList : boolean;
begin
  result := (FSpecialArrangementList <> nil) and (FSpecialArrangementList.count > 0);
end;

Procedure TFhirEncounterHospitalization.SetDestination(value : TFhirReference{TFhirLocation});
begin
  FDestination.free;
  FDestination := value;
end;

Procedure TFhirEncounterHospitalization.SetDischargeDisposition(value : TFhirCodeableConcept);
begin
  FDischargeDisposition.free;
  FDischargeDisposition := value;
end;

Function TFhirEncounterHospitalization.GetDischargeDiagnosisList : TFhirReferenceList{TFhirCondition};
begin
  if FDischargeDiagnosisList = nil then
    FDischargeDiagnosisList := TFhirReferenceList{TFhirCondition}.Create;
  result := FDischargeDiagnosisList;
end;

Function TFhirEncounterHospitalization.GetHasDischargeDiagnosisList : boolean;
begin
  result := (FDischargeDiagnosisList <> nil) and (FDischargeDiagnosisList.count > 0);
end;


{ TFhirEncounterHospitalizationListEnumerator }

Constructor TFhirEncounterHospitalizationListEnumerator.Create(list : TFhirEncounterHospitalizationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEncounterHospitalizationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEncounterHospitalizationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEncounterHospitalizationListEnumerator.GetCurrent : TFhirEncounterHospitalization;
begin
  Result := FList[FIndex];
end;


{ TFhirEncounterHospitalizationList }
procedure TFhirEncounterHospitalizationList.AddItem(value: TFhirEncounterHospitalization);
begin
  assert(value.ClassName = 'TFhirEncounterHospitalization', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEncounterHospitalization');
  add(value);
end;


function TFhirEncounterHospitalizationList.Append: TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterHospitalizationList.ClearItems;
begin
  Clear;
end;

function TFhirEncounterHospitalizationList.GetEnumerator : TFhirEncounterHospitalizationListEnumerator;
begin
  result := TFhirEncounterHospitalizationListEnumerator.Create(self.link);
end;

function TFhirEncounterHospitalizationList.Clone: TFhirEncounterHospitalizationList;
begin
  result := TFhirEncounterHospitalizationList(inherited Clone);
end;

function TFhirEncounterHospitalizationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEncounterHospitalizationList.GetItemN(index: Integer): TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization(ObjectByIndex[index]);
end;

function TFhirEncounterHospitalizationList.IndexOf(value: TFhirEncounterHospitalization): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEncounterHospitalizationList.Insert(index: Integer): TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterHospitalizationList.InsertItem(index: Integer; value: TFhirEncounterHospitalization);
begin
  assert(value is TFhirEncounterHospitalization);
  Inherited Insert(index, value);
end;

function TFhirEncounterHospitalizationList.Item(index: Integer): TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization(ObjectByIndex[index]);
end;

function TFhirEncounterHospitalizationList.Link: TFhirEncounterHospitalizationList;
begin
  result := TFhirEncounterHospitalizationList(inherited Link);
end;

procedure TFhirEncounterHospitalizationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEncounterHospitalizationList.SetItemByIndex(index: Integer; value: TFhirEncounterHospitalization);
begin
  assert(value is TFhirEncounterHospitalization);
  FhirEncounterHospitalizations[index] := value;
end;

procedure TFhirEncounterHospitalizationList.SetItemN(index: Integer; value: TFhirEncounterHospitalization);
begin
  assert(value is TFhirEncounterHospitalization);
  ObjectByIndex[index] := value;
end;

{ TFhirEncounterLocation }

constructor TFhirEncounterLocation.Create;
begin
  inherited;
end;

destructor TFhirEncounterLocation.Destroy;
begin
  FLocation.free;
  FStatus.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirEncounterLocation.Assign(oSource : TAdvObject);
begin
  inherited;
  location := TFhirEncounterLocation(oSource).location.Clone;
  FStatus := TFhirEncounterLocation(oSource).FStatus.Link;
  period := TFhirEncounterLocation(oSource).period.Clone;
end;

procedure TFhirEncounterLocation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirEncounterLocation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirEncounterLocation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirEncounterLocation.FhirType : string;
begin
  result := 'location';
end;

function TFhirEncounterLocation.Link : TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation(inherited Link);
end;

function TFhirEncounterLocation.Clone : TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation(inherited Clone);
end;

function TFhirEncounterLocation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterLocation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEncounterLocation)) then
    result := false
  else
  begin
    o := TFhirEncounterLocation(other);
    result := compareDeep(locationElement, o.locationElement, true) and compareDeep(statusElement, o.statusElement, true) and 
      compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirEncounterLocation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEncounterLocation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEncounterLocation)) then
    result := false
  else
  begin
    o := TFhirEncounterLocation(other);
    result := compareValues(statusElement, o.statusElement, true);
  end;
end;

{ TFhirEncounterLocation }

Procedure TFhirEncounterLocation.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Procedure TFhirEncounterLocation.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirEncounterLocation.GetStatusST : TFhirEncounterLocationStatus;
begin
  if FStatus = nil then
    result := TFhirEncounterLocationStatus(0)
  else
    result := TFhirEncounterLocationStatus(StringArrayIndexOfSensitive(CODES_TFhirEncounterLocationStatus, FStatus.value));
end;

Procedure TFhirEncounterLocation.SetStatusST(value : TFhirEncounterLocationStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirEncounterLocationStatus[value]);
end;

Procedure TFhirEncounterLocation.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirEncounterLocationListEnumerator }

Constructor TFhirEncounterLocationListEnumerator.Create(list : TFhirEncounterLocationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEncounterLocationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEncounterLocationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEncounterLocationListEnumerator.GetCurrent : TFhirEncounterLocation;
begin
  Result := FList[FIndex];
end;


{ TFhirEncounterLocationList }
procedure TFhirEncounterLocationList.AddItem(value: TFhirEncounterLocation);
begin
  assert(value.ClassName = 'TFhirEncounterLocation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEncounterLocation');
  add(value);
end;


function TFhirEncounterLocationList.Append: TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterLocationList.ClearItems;
begin
  Clear;
end;

function TFhirEncounterLocationList.GetEnumerator : TFhirEncounterLocationListEnumerator;
begin
  result := TFhirEncounterLocationListEnumerator.Create(self.link);
end;

function TFhirEncounterLocationList.Clone: TFhirEncounterLocationList;
begin
  result := TFhirEncounterLocationList(inherited Clone);
end;

function TFhirEncounterLocationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEncounterLocationList.GetItemN(index: Integer): TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation(ObjectByIndex[index]);
end;

function TFhirEncounterLocationList.IndexOf(value: TFhirEncounterLocation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEncounterLocationList.Insert(index: Integer): TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterLocationList.InsertItem(index: Integer; value: TFhirEncounterLocation);
begin
  assert(value is TFhirEncounterLocation);
  Inherited Insert(index, value);
end;

function TFhirEncounterLocationList.Item(index: Integer): TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation(ObjectByIndex[index]);
end;

function TFhirEncounterLocationList.Link: TFhirEncounterLocationList;
begin
  result := TFhirEncounterLocationList(inherited Link);
end;

procedure TFhirEncounterLocationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEncounterLocationList.SetItemByIndex(index: Integer; value: TFhirEncounterLocation);
begin
  assert(value is TFhirEncounterLocation);
  FhirEncounterLocations[index] := value;
end;

procedure TFhirEncounterLocationList.SetItemN(index: Integer; value: TFhirEncounterLocation);
begin
  assert(value is TFhirEncounterLocation);
  ObjectByIndex[index] := value;
end;

{ TFhirEncounter }

constructor TFhirEncounter.Create;
begin
  inherited;
end;

destructor TFhirEncounter.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FStatusHistoryList.Free;
  FClass_.free;
  FType_List.Free;
  FPriority.free;
  FPatient.free;
  FEpisodeOfCareList.Free;
  FIncomingReferralList.Free;
  FParticipantList.Free;
  FAppointment.free;
  FPeriod.free;
  FLength.free;
  FReasonList.Free;
  FIndicationList.Free;
  FHospitalization.free;
  FLocationList.Free;
  FServiceProvider.free;
  FPartOf.free;
  inherited;
end;

function TFhirEncounter.GetResourceType : TFhirResourceType;
begin
  result := frtEncounter;
end;

procedure TFhirEncounter.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEncounter(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEncounter(oSource).FIdentifierList);
  end;
  FStatus := TFhirEncounter(oSource).FStatus.Link;
  if (TFhirEncounter(oSource).FStatusHistoryList = nil) then
  begin
    FStatusHistoryList.free;
    FStatusHistoryList := nil;
  end
  else
  begin
    FStatusHistoryList := TFhirEncounterStatusHistoryList.Create;
    FStatusHistoryList.Assign(TFhirEncounter(oSource).FStatusHistoryList);
  end;
  FClass_ := TFhirEncounter(oSource).FClass_.Link;
  if (TFhirEncounter(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirEncounter(oSource).FType_List);
  end;
  priority := TFhirEncounter(oSource).priority.Clone;
  patient := TFhirEncounter(oSource).patient.Clone;
  if (TFhirEncounter(oSource).FEpisodeOfCareList = nil) then
  begin
    FEpisodeOfCareList.free;
    FEpisodeOfCareList := nil;
  end
  else
  begin
    FEpisodeOfCareList := TFhirReferenceList{TFhirEpisodeOfCare}.Create;
    FEpisodeOfCareList.Assign(TFhirEncounter(oSource).FEpisodeOfCareList);
  end;
  if (TFhirEncounter(oSource).FIncomingReferralList = nil) then
  begin
    FIncomingReferralList.free;
    FIncomingReferralList := nil;
  end
  else
  begin
    FIncomingReferralList := TFhirReferenceList{TFhirReferralRequest}.Create;
    FIncomingReferralList.Assign(TFhirEncounter(oSource).FIncomingReferralList);
  end;
  if (TFhirEncounter(oSource).FParticipantList = nil) then
  begin
    FParticipantList.free;
    FParticipantList := nil;
  end
  else
  begin
    FParticipantList := TFhirEncounterParticipantList.Create;
    FParticipantList.Assign(TFhirEncounter(oSource).FParticipantList);
  end;
  appointment := TFhirEncounter(oSource).appointment.Clone;
  period := TFhirEncounter(oSource).period.Clone;
  length := TFhirEncounter(oSource).length.Clone;
  if (TFhirEncounter(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirEncounter(oSource).FReasonList);
  end;
  if (TFhirEncounter(oSource).FIndicationList = nil) then
  begin
    FIndicationList.free;
    FIndicationList := nil;
  end
  else
  begin
    FIndicationList := TFhirReferenceList{Resource}.Create;
    FIndicationList.Assign(TFhirEncounter(oSource).FIndicationList);
  end;
  hospitalization := TFhirEncounter(oSource).hospitalization.Clone;
  if (TFhirEncounter(oSource).FLocationList = nil) then
  begin
    FLocationList.free;
    FLocationList := nil;
  end
  else
  begin
    FLocationList := TFhirEncounterLocationList.Create;
    FLocationList.Assign(TFhirEncounter(oSource).FLocationList);
  end;
  serviceProvider := TFhirEncounter(oSource).serviceProvider.Clone;
  partOf := TFhirEncounter(oSource).partOf.Clone;
end;

procedure TFhirEncounter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'statusHistory') Then
    list.addAll(FStatusHistoryList);
  if (child_name = 'class') Then
     list.add(FClass_.Link);
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'episodeOfCare') Then
    list.addAll(FEpisodeOfCareList);
  if (child_name = 'incomingReferral') Then
    list.addAll(FIncomingReferralList);
  if (child_name = 'participant') Then
    list.addAll(FParticipantList);
  if (child_name = 'appointment') Then
     list.add(FAppointment.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'length') Then
     list.add(FLength.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'indication') Then
    list.addAll(FIndicationList);
  if (child_name = 'hospitalization') Then
     list.add(FHospitalization.Link);
  if (child_name = 'location') Then
    list.addAll(FLocationList);
  if (child_name = 'serviceProvider') Then
     list.add(FServiceProvider.Link);
  if (child_name = 'partOf') Then
     list.add(FPartOf.Link);
end;

procedure TFhirEncounter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'statusHistory', '', FStatusHistoryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'class', 'code', FClass_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'priority', 'CodeableConcept', FPriority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'episodeOfCare', 'Reference(EpisodeOfCare)', FEpisodeOfCareList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'incomingReferral', 'Reference(ReferralRequest)', FIncomingReferralList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'participant', '', FParticipantList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'appointment', 'Reference(Appointment)', FAppointment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'length', 'Duration', FLength.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'indication', 'Reference(Condition|Procedure)', FIndicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'hospitalization', '', FHospitalization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', '', FLocationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'serviceProvider', 'Reference(Organization)', FServiceProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'partOf', 'Reference(Encounter)', FPartOf.Link));{2}
end;

procedure TFhirEncounter.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'statusHistory') then StatusHistoryList.add(propValue as TFhirEncounterStatusHistory){2}
  else if (propName = 'class') then Class_Element := propValue as TFHIREnum
  else if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'priority') then Priority := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'episodeOfCare') then EpisodeOfCareList.add(propValue as TFhirReference{TFhirEpisodeOfCare}){2}
  else if (propName = 'incomingReferral') then IncomingReferralList.add(propValue as TFhirReference{TFhirReferralRequest}){2}
  else if (propName = 'participant') then ParticipantList.add(propValue as TFhirEncounterParticipant){2}
  else if (propName = 'appointment') then Appointment := propValue as TFhirReference{TFhirAppointment}{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'length') then Length := propValue as TFhirQuantity{4b}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'indication') then IndicationList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'hospitalization') then Hospitalization := propValue as TFhirEncounterHospitalization{4b}
  else if (propName = 'location') then LocationList.add(propValue as TFhirEncounterLocation){2}
  else if (propName = 'serviceProvider') then ServiceProvider := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'partOf') then PartOf := propValue as TFhirReference{TFhirEncounter}{4b}
  else inherited;
end;

function TFhirEncounter.FhirType : string;
begin
  result := 'Encounter';
end;

function TFhirEncounter.Link : TFhirEncounter;
begin
  result := TFhirEncounter(inherited Link);
end;

function TFhirEncounter.Clone : TFhirEncounter;
begin
  result := TFhirEncounter(inherited Clone);
end;

{ TFhirEncounter }

Function TFhirEncounter.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEncounter.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEncounter.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirEncounter.GetStatusST : TFhirEncounterState;
begin
  if FStatus = nil then
    result := TFhirEncounterState(0)
  else
    result := TFhirEncounterState(StringArrayIndexOfSensitive(CODES_TFhirEncounterState, FStatus.value));
end;

Procedure TFhirEncounter.SetStatusST(value : TFhirEncounterState);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirEncounterState[value]);
end;

Function TFhirEncounter.GetStatusHistoryList : TFhirEncounterStatusHistoryList;
begin
  if FStatusHistoryList = nil then
    FStatusHistoryList := TFhirEncounterStatusHistoryList.Create;
  result := FStatusHistoryList;
end;

Function TFhirEncounter.GetHasStatusHistoryList : boolean;
begin
  result := (FStatusHistoryList <> nil) and (FStatusHistoryList.count > 0);
end;

Procedure TFhirEncounter.SetClass_(value : TFhirEnum);
begin
  FClass_.free;
  FClass_ := value;
end;

Function TFhirEncounter.GetClass_ST : TFhirEncounterClass;
begin
  if FClass_ = nil then
    result := TFhirEncounterClass(0)
  else
    result := TFhirEncounterClass(StringArrayIndexOfSensitive(CODES_TFhirEncounterClass, FClass_.value));
end;

Procedure TFhirEncounter.SetClass_ST(value : TFhirEncounterClass);
begin
  if ord(value) = 0 then
    Class_Element := nil
  else
    Class_Element := TFhirEnum.create(CODES_TFhirEncounterClass[value]);
end;

Function TFhirEncounter.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirEncounter.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Procedure TFhirEncounter.SetPriority(value : TFhirCodeableConcept);
begin
  FPriority.free;
  FPriority := value;
end;

Procedure TFhirEncounter.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Function TFhirEncounter.GetEpisodeOfCareList : TFhirReferenceList{TFhirEpisodeOfCare};
begin
  if FEpisodeOfCareList = nil then
    FEpisodeOfCareList := TFhirReferenceList{TFhirEpisodeOfCare}.Create;
  result := FEpisodeOfCareList;
end;

Function TFhirEncounter.GetHasEpisodeOfCareList : boolean;
begin
  result := (FEpisodeOfCareList <> nil) and (FEpisodeOfCareList.count > 0);
end;

Function TFhirEncounter.GetIncomingReferralList : TFhirReferenceList{TFhirReferralRequest};
begin
  if FIncomingReferralList = nil then
    FIncomingReferralList := TFhirReferenceList{TFhirReferralRequest}.Create;
  result := FIncomingReferralList;
end;

Function TFhirEncounter.GetHasIncomingReferralList : boolean;
begin
  result := (FIncomingReferralList <> nil) and (FIncomingReferralList.count > 0);
end;

Function TFhirEncounter.GetParticipantList : TFhirEncounterParticipantList;
begin
  if FParticipantList = nil then
    FParticipantList := TFhirEncounterParticipantList.Create;
  result := FParticipantList;
end;

Function TFhirEncounter.GetHasParticipantList : boolean;
begin
  result := (FParticipantList <> nil) and (FParticipantList.count > 0);
end;

Procedure TFhirEncounter.SetAppointment(value : TFhirReference{TFhirAppointment});
begin
  FAppointment.free;
  FAppointment := value;
end;

Procedure TFhirEncounter.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirEncounter.SetLength(value : TFhirQuantity);
begin
  FLength.free;
  FLength := value;
end;

Function TFhirEncounter.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirEncounter.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Function TFhirEncounter.GetIndicationList : TFhirReferenceList{Resource};
begin
  if FIndicationList = nil then
    FIndicationList := TFhirReferenceList{Resource}.Create;
  result := FIndicationList;
end;

Function TFhirEncounter.GetHasIndicationList : boolean;
begin
  result := (FIndicationList <> nil) and (FIndicationList.count > 0);
end;

Procedure TFhirEncounter.SetHospitalization(value : TFhirEncounterHospitalization);
begin
  FHospitalization.free;
  FHospitalization := value;
end;

Function TFhirEncounter.GetLocationList : TFhirEncounterLocationList;
begin
  if FLocationList = nil then
    FLocationList := TFhirEncounterLocationList.Create;
  result := FLocationList;
end;

Function TFhirEncounter.GetHasLocationList : boolean;
begin
  result := (FLocationList <> nil) and (FLocationList.count > 0);
end;

Procedure TFhirEncounter.SetServiceProvider(value : TFhirReference{TFhirOrganization});
begin
  FServiceProvider.free;
  FServiceProvider := value;
end;

Procedure TFhirEncounter.SetPartOf(value : TFhirReference{TFhirEncounter});
begin
  FPartOf.free;
  FPartOf := value;
end;


{ TFhirEncounterListEnumerator }

Constructor TFhirEncounterListEnumerator.Create(list : TFhirEncounterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEncounterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEncounterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEncounterListEnumerator.GetCurrent : TFhirEncounter;
begin
  Result := FList[FIndex];
end;


{ TFhirEncounterList }
procedure TFhirEncounterList.AddItem(value: TFhirEncounter);
begin
  assert(value.ClassName = 'TFhirEncounter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEncounter');
  add(value);
end;


function TFhirEncounterList.Append: TFhirEncounter;
begin
  result := TFhirEncounter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterList.ClearItems;
begin
  Clear;
end;

function TFhirEncounterList.GetEnumerator : TFhirEncounterListEnumerator;
begin
  result := TFhirEncounterListEnumerator.Create(self.link);
end;

function TFhirEncounterList.Clone: TFhirEncounterList;
begin
  result := TFhirEncounterList(inherited Clone);
end;

function TFhirEncounterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEncounterList.GetItemN(index: Integer): TFhirEncounter;
begin
  result := TFhirEncounter(ObjectByIndex[index]);
end;

function TFhirEncounterList.IndexOf(value: TFhirEncounter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEncounterList.Insert(index: Integer): TFhirEncounter;
begin
  result := TFhirEncounter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEncounterList.InsertItem(index: Integer; value: TFhirEncounter);
begin
  assert(value is TFhirEncounter);
  Inherited Insert(index, value);
end;

function TFhirEncounterList.Item(index: Integer): TFhirEncounter;
begin
  result := TFhirEncounter(ObjectByIndex[index]);
end;

function TFhirEncounterList.Link: TFhirEncounterList;
begin
  result := TFhirEncounterList(inherited Link);
end;

procedure TFhirEncounterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEncounterList.SetItemByIndex(index: Integer; value: TFhirEncounter);
begin
  assert(value is TFhirEncounter);
  FhirEncounters[index] := value;
end;

procedure TFhirEncounterList.SetItemN(index: Integer; value: TFhirEncounter);
begin
  assert(value is TFhirEncounter);
  ObjectByIndex[index] := value;
end;

{ TFhirEnrollmentRequest }

constructor TFhirEnrollmentRequest.Create;
begin
  inherited;
end;

destructor TFhirEnrollmentRequest.Destroy;
begin
  FIdentifierList.Free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FTarget.free;
  FProvider.free;
  FOrganization.free;
  FSubject.free;
  FCoverage.free;
  FRelationship.free;
  inherited;
end;

function TFhirEnrollmentRequest.GetResourceType : TFhirResourceType;
begin
  result := frtEnrollmentRequest;
end;

procedure TFhirEnrollmentRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEnrollmentRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEnrollmentRequest(oSource).FIdentifierList);
  end;
  ruleset := TFhirEnrollmentRequest(oSource).ruleset.Clone;
  originalRuleset := TFhirEnrollmentRequest(oSource).originalRuleset.Clone;
  createdElement := TFhirEnrollmentRequest(oSource).createdElement.Clone;
  target := TFhirEnrollmentRequest(oSource).target.Clone;
  provider := TFhirEnrollmentRequest(oSource).provider.Clone;
  organization := TFhirEnrollmentRequest(oSource).organization.Clone;
  subject := TFhirEnrollmentRequest(oSource).subject.Clone;
  coverage := TFhirEnrollmentRequest(oSource).coverage.Clone;
  relationship := TFhirEnrollmentRequest(oSource).relationship.Clone;
end;

procedure TFhirEnrollmentRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'coverage') Then
     list.add(FCoverage.Link);
  if (child_name = 'relationship') Then
     list.add(FRelationship.Link);
end;

procedure TFhirEnrollmentRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'coverage', 'Reference(Coverage)', FCoverage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relationship', 'Coding', FRelationship.Link));{2}
end;

procedure TFhirEnrollmentRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'coverage') then Coverage := propValue as TFhirReference{TFhirCoverage}{4b}
  else if (propName = 'relationship') then Relationship := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirEnrollmentRequest.FhirType : string;
begin
  result := 'EnrollmentRequest';
end;

function TFhirEnrollmentRequest.Link : TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest(inherited Link);
end;

function TFhirEnrollmentRequest.Clone : TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest(inherited Clone);
end;

{ TFhirEnrollmentRequest }

Function TFhirEnrollmentRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEnrollmentRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEnrollmentRequest.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirEnrollmentRequest.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirEnrollmentRequest.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirEnrollmentRequest.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirEnrollmentRequest.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirEnrollmentRequest.SetTarget(value : TFhirReference{TFhirOrganization});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirEnrollmentRequest.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirEnrollmentRequest.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirEnrollmentRequest.SetSubject(value : TFhirReference{TFhirPatient});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirEnrollmentRequest.SetCoverage(value : TFhirReference{TFhirCoverage});
begin
  FCoverage.free;
  FCoverage := value;
end;

Procedure TFhirEnrollmentRequest.SetRelationship(value : TFhirCoding);
begin
  FRelationship.free;
  FRelationship := value;
end;


{ TFhirEnrollmentRequestListEnumerator }

Constructor TFhirEnrollmentRequestListEnumerator.Create(list : TFhirEnrollmentRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEnrollmentRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEnrollmentRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEnrollmentRequestListEnumerator.GetCurrent : TFhirEnrollmentRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirEnrollmentRequestList }
procedure TFhirEnrollmentRequestList.AddItem(value: TFhirEnrollmentRequest);
begin
  assert(value.ClassName = 'TFhirEnrollmentRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEnrollmentRequest');
  add(value);
end;


function TFhirEnrollmentRequestList.Append: TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEnrollmentRequestList.ClearItems;
begin
  Clear;
end;

function TFhirEnrollmentRequestList.GetEnumerator : TFhirEnrollmentRequestListEnumerator;
begin
  result := TFhirEnrollmentRequestListEnumerator.Create(self.link);
end;

function TFhirEnrollmentRequestList.Clone: TFhirEnrollmentRequestList;
begin
  result := TFhirEnrollmentRequestList(inherited Clone);
end;

function TFhirEnrollmentRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEnrollmentRequestList.GetItemN(index: Integer): TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest(ObjectByIndex[index]);
end;

function TFhirEnrollmentRequestList.IndexOf(value: TFhirEnrollmentRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEnrollmentRequestList.Insert(index: Integer): TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEnrollmentRequestList.InsertItem(index: Integer; value: TFhirEnrollmentRequest);
begin
  assert(value is TFhirEnrollmentRequest);
  Inherited Insert(index, value);
end;

function TFhirEnrollmentRequestList.Item(index: Integer): TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest(ObjectByIndex[index]);
end;

function TFhirEnrollmentRequestList.Link: TFhirEnrollmentRequestList;
begin
  result := TFhirEnrollmentRequestList(inherited Link);
end;

procedure TFhirEnrollmentRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEnrollmentRequestList.SetItemByIndex(index: Integer; value: TFhirEnrollmentRequest);
begin
  assert(value is TFhirEnrollmentRequest);
  FhirEnrollmentRequests[index] := value;
end;

procedure TFhirEnrollmentRequestList.SetItemN(index: Integer; value: TFhirEnrollmentRequest);
begin
  assert(value is TFhirEnrollmentRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirEnrollmentResponse }

constructor TFhirEnrollmentResponse.Create;
begin
  inherited;
end;

destructor TFhirEnrollmentResponse.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FOutcome.free;
  FDisposition.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  inherited;
end;

function TFhirEnrollmentResponse.GetResourceType : TFhirResourceType;
begin
  result := frtEnrollmentResponse;
end;

procedure TFhirEnrollmentResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEnrollmentResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEnrollmentResponse(oSource).FIdentifierList);
  end;
  request := TFhirEnrollmentResponse(oSource).request.Clone;
  FOutcome := TFhirEnrollmentResponse(oSource).FOutcome.Link;
  dispositionElement := TFhirEnrollmentResponse(oSource).dispositionElement.Clone;
  ruleset := TFhirEnrollmentResponse(oSource).ruleset.Clone;
  originalRuleset := TFhirEnrollmentResponse(oSource).originalRuleset.Clone;
  createdElement := TFhirEnrollmentResponse(oSource).createdElement.Clone;
  organization := TFhirEnrollmentResponse(oSource).organization.Clone;
  requestProvider := TFhirEnrollmentResponse(oSource).requestProvider.Clone;
  requestOrganization := TFhirEnrollmentResponse(oSource).requestOrganization.Clone;
end;

procedure TFhirEnrollmentResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
end;

procedure TFhirEnrollmentResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(EnrollmentRequest)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
end;

procedure TFhirEnrollmentResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirEnrollmentRequest}{4b}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else inherited;
end;

function TFhirEnrollmentResponse.FhirType : string;
begin
  result := 'EnrollmentResponse';
end;

function TFhirEnrollmentResponse.Link : TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse(inherited Link);
end;

function TFhirEnrollmentResponse.Clone : TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse(inherited Clone);
end;

{ TFhirEnrollmentResponse }

Function TFhirEnrollmentResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEnrollmentResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEnrollmentResponse.SetRequest(value : TFhirReference{TFhirEnrollmentRequest});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirEnrollmentResponse.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirEnrollmentResponse.GetOutcomeST : TFhirRemittanceOutcome;
begin
  if FOutcome = nil then
    result := TFhirRemittanceOutcome(0)
  else
    result := TFhirRemittanceOutcome(StringArrayIndexOfSensitive(CODES_TFhirRemittanceOutcome, FOutcome.value));
end;

Procedure TFhirEnrollmentResponse.SetOutcomeST(value : TFhirRemittanceOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirRemittanceOutcome[value]);
end;

Procedure TFhirEnrollmentResponse.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirEnrollmentResponse.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirEnrollmentResponse.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirEnrollmentResponse.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirEnrollmentResponse.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirEnrollmentResponse.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirEnrollmentResponse.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirEnrollmentResponse.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirEnrollmentResponse.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirEnrollmentResponse.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirEnrollmentResponse.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;


{ TFhirEnrollmentResponseListEnumerator }

Constructor TFhirEnrollmentResponseListEnumerator.Create(list : TFhirEnrollmentResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEnrollmentResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEnrollmentResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEnrollmentResponseListEnumerator.GetCurrent : TFhirEnrollmentResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirEnrollmentResponseList }
procedure TFhirEnrollmentResponseList.AddItem(value: TFhirEnrollmentResponse);
begin
  assert(value.ClassName = 'TFhirEnrollmentResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEnrollmentResponse');
  add(value);
end;


function TFhirEnrollmentResponseList.Append: TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEnrollmentResponseList.ClearItems;
begin
  Clear;
end;

function TFhirEnrollmentResponseList.GetEnumerator : TFhirEnrollmentResponseListEnumerator;
begin
  result := TFhirEnrollmentResponseListEnumerator.Create(self.link);
end;

function TFhirEnrollmentResponseList.Clone: TFhirEnrollmentResponseList;
begin
  result := TFhirEnrollmentResponseList(inherited Clone);
end;

function TFhirEnrollmentResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEnrollmentResponseList.GetItemN(index: Integer): TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse(ObjectByIndex[index]);
end;

function TFhirEnrollmentResponseList.IndexOf(value: TFhirEnrollmentResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEnrollmentResponseList.Insert(index: Integer): TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEnrollmentResponseList.InsertItem(index: Integer; value: TFhirEnrollmentResponse);
begin
  assert(value is TFhirEnrollmentResponse);
  Inherited Insert(index, value);
end;

function TFhirEnrollmentResponseList.Item(index: Integer): TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse(ObjectByIndex[index]);
end;

function TFhirEnrollmentResponseList.Link: TFhirEnrollmentResponseList;
begin
  result := TFhirEnrollmentResponseList(inherited Link);
end;

procedure TFhirEnrollmentResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEnrollmentResponseList.SetItemByIndex(index: Integer; value: TFhirEnrollmentResponse);
begin
  assert(value is TFhirEnrollmentResponse);
  FhirEnrollmentResponses[index] := value;
end;

procedure TFhirEnrollmentResponseList.SetItemN(index: Integer; value: TFhirEnrollmentResponse);
begin
  assert(value is TFhirEnrollmentResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirEpisodeOfCareStatusHistory }

constructor TFhirEpisodeOfCareStatusHistory.Create;
begin
  inherited;
end;

destructor TFhirEpisodeOfCareStatusHistory.Destroy;
begin
  FStatus.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirEpisodeOfCareStatusHistory.Assign(oSource : TAdvObject);
begin
  inherited;
  FStatus := TFhirEpisodeOfCareStatusHistory(oSource).FStatus.Link;
  period := TFhirEpisodeOfCareStatusHistory(oSource).period.Clone;
end;

procedure TFhirEpisodeOfCareStatusHistory.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirEpisodeOfCareStatusHistory.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirEpisodeOfCareStatusHistory.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirEpisodeOfCareStatusHistory.FhirType : string;
begin
  result := 'statusHistory';
end;

function TFhirEpisodeOfCareStatusHistory.Link : TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory(inherited Link);
end;

function TFhirEpisodeOfCareStatusHistory.Clone : TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory(inherited Clone);
end;

function TFhirEpisodeOfCareStatusHistory.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEpisodeOfCareStatusHistory;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEpisodeOfCareStatusHistory)) then
    result := false
  else
  begin
    o := TFhirEpisodeOfCareStatusHistory(other);
    result := compareDeep(statusElement, o.statusElement, true) and compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirEpisodeOfCareStatusHistory.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEpisodeOfCareStatusHistory;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEpisodeOfCareStatusHistory)) then
    result := false
  else
  begin
    o := TFhirEpisodeOfCareStatusHistory(other);
    result := compareValues(statusElement, o.statusElement, true);
  end;
end;

{ TFhirEpisodeOfCareStatusHistory }

Procedure TFhirEpisodeOfCareStatusHistory.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirEpisodeOfCareStatusHistory.GetStatusST : TFhirEpisodeOfCareStatus;
begin
  if FStatus = nil then
    result := TFhirEpisodeOfCareStatus(0)
  else
    result := TFhirEpisodeOfCareStatus(StringArrayIndexOfSensitive(CODES_TFhirEpisodeOfCareStatus, FStatus.value));
end;

Procedure TFhirEpisodeOfCareStatusHistory.SetStatusST(value : TFhirEpisodeOfCareStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirEpisodeOfCareStatus[value]);
end;

Procedure TFhirEpisodeOfCareStatusHistory.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirEpisodeOfCareStatusHistoryListEnumerator }

Constructor TFhirEpisodeOfCareStatusHistoryListEnumerator.Create(list : TFhirEpisodeOfCareStatusHistoryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEpisodeOfCareStatusHistoryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEpisodeOfCareStatusHistoryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEpisodeOfCareStatusHistoryListEnumerator.GetCurrent : TFhirEpisodeOfCareStatusHistory;
begin
  Result := FList[FIndex];
end;


{ TFhirEpisodeOfCareStatusHistoryList }
procedure TFhirEpisodeOfCareStatusHistoryList.AddItem(value: TFhirEpisodeOfCareStatusHistory);
begin
  assert(value.ClassName = 'TFhirEpisodeOfCareStatusHistory', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEpisodeOfCareStatusHistory');
  add(value);
end;


function TFhirEpisodeOfCareStatusHistoryList.Append: TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareStatusHistoryList.ClearItems;
begin
  Clear;
end;

function TFhirEpisodeOfCareStatusHistoryList.GetEnumerator : TFhirEpisodeOfCareStatusHistoryListEnumerator;
begin
  result := TFhirEpisodeOfCareStatusHistoryListEnumerator.Create(self.link);
end;

function TFhirEpisodeOfCareStatusHistoryList.Clone: TFhirEpisodeOfCareStatusHistoryList;
begin
  result := TFhirEpisodeOfCareStatusHistoryList(inherited Clone);
end;

function TFhirEpisodeOfCareStatusHistoryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEpisodeOfCareStatusHistoryList.GetItemN(index: Integer): TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareStatusHistoryList.IndexOf(value: TFhirEpisodeOfCareStatusHistory): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEpisodeOfCareStatusHistoryList.Insert(index: Integer): TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareStatusHistoryList.InsertItem(index: Integer; value: TFhirEpisodeOfCareStatusHistory);
begin
  assert(value is TFhirEpisodeOfCareStatusHistory);
  Inherited Insert(index, value);
end;

function TFhirEpisodeOfCareStatusHistoryList.Item(index: Integer): TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareStatusHistoryList.Link: TFhirEpisodeOfCareStatusHistoryList;
begin
  result := TFhirEpisodeOfCareStatusHistoryList(inherited Link);
end;

procedure TFhirEpisodeOfCareStatusHistoryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEpisodeOfCareStatusHistoryList.SetItemByIndex(index: Integer; value: TFhirEpisodeOfCareStatusHistory);
begin
  assert(value is TFhirEpisodeOfCareStatusHistory);
  FhirEpisodeOfCareStatusHistories[index] := value;
end;

procedure TFhirEpisodeOfCareStatusHistoryList.SetItemN(index: Integer; value: TFhirEpisodeOfCareStatusHistory);
begin
  assert(value is TFhirEpisodeOfCareStatusHistory);
  ObjectByIndex[index] := value;
end;

{ TFhirEpisodeOfCareCareTeam }

constructor TFhirEpisodeOfCareCareTeam.Create;
begin
  inherited;
end;

destructor TFhirEpisodeOfCareCareTeam.Destroy;
begin
  FRoleList.Free;
  FPeriod.free;
  FMember.free;
  inherited;
end;

procedure TFhirEpisodeOfCareCareTeam.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEpisodeOfCareCareTeam(oSource).FRoleList = nil) then
  begin
    FRoleList.free;
    FRoleList := nil;
  end
  else
  begin
    FRoleList := TFhirCodeableConceptList.Create;
    FRoleList.Assign(TFhirEpisodeOfCareCareTeam(oSource).FRoleList);
  end;
  period := TFhirEpisodeOfCareCareTeam(oSource).period.Clone;
  member := TFhirEpisodeOfCareCareTeam(oSource).member.Clone;
end;

procedure TFhirEpisodeOfCareCareTeam.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'role') Then
    list.addAll(FRoleList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'member') Then
     list.add(FMember.Link);
end;

procedure TFhirEpisodeOfCareCareTeam.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRoleList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'member', 'Reference(Practitioner|Organization)', FMember.Link));{2}
end;

procedure TFhirEpisodeOfCareCareTeam.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'role') then RoleList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'member') then Member := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirEpisodeOfCareCareTeam.FhirType : string;
begin
  result := 'careTeam';
end;

function TFhirEpisodeOfCareCareTeam.Link : TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam(inherited Link);
end;

function TFhirEpisodeOfCareCareTeam.Clone : TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam(inherited Clone);
end;

function TFhirEpisodeOfCareCareTeam.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirEpisodeOfCareCareTeam;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirEpisodeOfCareCareTeam)) then
    result := false
  else
  begin
    o := TFhirEpisodeOfCareCareTeam(other);
    result := compareDeep(roleList, o.roleList, true) and compareDeep(periodElement, o.periodElement, true) and 
      compareDeep(memberElement, o.memberElement, true);
  end;
end;

function TFhirEpisodeOfCareCareTeam.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirEpisodeOfCareCareTeam;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirEpisodeOfCareCareTeam)) then
    result := false
  else
  begin
    o := TFhirEpisodeOfCareCareTeam(other);
    result := true;
  end;
end;

{ TFhirEpisodeOfCareCareTeam }

Function TFhirEpisodeOfCareCareTeam.GetRoleList : TFhirCodeableConceptList;
begin
  if FRoleList = nil then
    FRoleList := TFhirCodeableConceptList.Create;
  result := FRoleList;
end;

Function TFhirEpisodeOfCareCareTeam.GetHasRoleList : boolean;
begin
  result := (FRoleList <> nil) and (FRoleList.count > 0);
end;

Procedure TFhirEpisodeOfCareCareTeam.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirEpisodeOfCareCareTeam.SetMember(value : TFhirReference{Resource});
begin
  FMember.free;
  FMember := value;
end;


{ TFhirEpisodeOfCareCareTeamListEnumerator }

Constructor TFhirEpisodeOfCareCareTeamListEnumerator.Create(list : TFhirEpisodeOfCareCareTeamList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEpisodeOfCareCareTeamListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEpisodeOfCareCareTeamListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEpisodeOfCareCareTeamListEnumerator.GetCurrent : TFhirEpisodeOfCareCareTeam;
begin
  Result := FList[FIndex];
end;


{ TFhirEpisodeOfCareCareTeamList }
procedure TFhirEpisodeOfCareCareTeamList.AddItem(value: TFhirEpisodeOfCareCareTeam);
begin
  assert(value.ClassName = 'TFhirEpisodeOfCareCareTeam', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEpisodeOfCareCareTeam');
  add(value);
end;


function TFhirEpisodeOfCareCareTeamList.Append: TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareCareTeamList.ClearItems;
begin
  Clear;
end;

function TFhirEpisodeOfCareCareTeamList.GetEnumerator : TFhirEpisodeOfCareCareTeamListEnumerator;
begin
  result := TFhirEpisodeOfCareCareTeamListEnumerator.Create(self.link);
end;

function TFhirEpisodeOfCareCareTeamList.Clone: TFhirEpisodeOfCareCareTeamList;
begin
  result := TFhirEpisodeOfCareCareTeamList(inherited Clone);
end;

function TFhirEpisodeOfCareCareTeamList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEpisodeOfCareCareTeamList.GetItemN(index: Integer): TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareCareTeamList.IndexOf(value: TFhirEpisodeOfCareCareTeam): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEpisodeOfCareCareTeamList.Insert(index: Integer): TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareCareTeamList.InsertItem(index: Integer; value: TFhirEpisodeOfCareCareTeam);
begin
  assert(value is TFhirEpisodeOfCareCareTeam);
  Inherited Insert(index, value);
end;

function TFhirEpisodeOfCareCareTeamList.Item(index: Integer): TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareCareTeamList.Link: TFhirEpisodeOfCareCareTeamList;
begin
  result := TFhirEpisodeOfCareCareTeamList(inherited Link);
end;

procedure TFhirEpisodeOfCareCareTeamList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEpisodeOfCareCareTeamList.SetItemByIndex(index: Integer; value: TFhirEpisodeOfCareCareTeam);
begin
  assert(value is TFhirEpisodeOfCareCareTeam);
  FhirEpisodeOfCareCareTeams[index] := value;
end;

procedure TFhirEpisodeOfCareCareTeamList.SetItemN(index: Integer; value: TFhirEpisodeOfCareCareTeam);
begin
  assert(value is TFhirEpisodeOfCareCareTeam);
  ObjectByIndex[index] := value;
end;

{ TFhirEpisodeOfCare }

constructor TFhirEpisodeOfCare.Create;
begin
  inherited;
end;

destructor TFhirEpisodeOfCare.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FStatusHistoryList.Free;
  FType_List.Free;
  FConditionList.Free;
  FPatient.free;
  FManagingOrganization.free;
  FPeriod.free;
  FReferralRequestList.Free;
  FCareManager.free;
  FCareTeamList.Free;
  inherited;
end;

function TFhirEpisodeOfCare.GetResourceType : TFhirResourceType;
begin
  result := frtEpisodeOfCare;
end;

procedure TFhirEpisodeOfCare.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirEpisodeOfCare(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirEpisodeOfCare(oSource).FIdentifierList);
  end;
  FStatus := TFhirEpisodeOfCare(oSource).FStatus.Link;
  if (TFhirEpisodeOfCare(oSource).FStatusHistoryList = nil) then
  begin
    FStatusHistoryList.free;
    FStatusHistoryList := nil;
  end
  else
  begin
    FStatusHistoryList := TFhirEpisodeOfCareStatusHistoryList.Create;
    FStatusHistoryList.Assign(TFhirEpisodeOfCare(oSource).FStatusHistoryList);
  end;
  if (TFhirEpisodeOfCare(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirEpisodeOfCare(oSource).FType_List);
  end;
  if (TFhirEpisodeOfCare(oSource).FConditionList = nil) then
  begin
    FConditionList.free;
    FConditionList := nil;
  end
  else
  begin
    FConditionList := TFhirReferenceList{TFhirCondition}.Create;
    FConditionList.Assign(TFhirEpisodeOfCare(oSource).FConditionList);
  end;
  patient := TFhirEpisodeOfCare(oSource).patient.Clone;
  managingOrganization := TFhirEpisodeOfCare(oSource).managingOrganization.Clone;
  period := TFhirEpisodeOfCare(oSource).period.Clone;
  if (TFhirEpisodeOfCare(oSource).FReferralRequestList = nil) then
  begin
    FReferralRequestList.free;
    FReferralRequestList := nil;
  end
  else
  begin
    FReferralRequestList := TFhirReferenceList{TFhirReferralRequest}.Create;
    FReferralRequestList.Assign(TFhirEpisodeOfCare(oSource).FReferralRequestList);
  end;
  careManager := TFhirEpisodeOfCare(oSource).careManager.Clone;
  if (TFhirEpisodeOfCare(oSource).FCareTeamList = nil) then
  begin
    FCareTeamList.free;
    FCareTeamList := nil;
  end
  else
  begin
    FCareTeamList := TFhirEpisodeOfCareCareTeamList.Create;
    FCareTeamList.Assign(TFhirEpisodeOfCare(oSource).FCareTeamList);
  end;
end;

procedure TFhirEpisodeOfCare.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'statusHistory') Then
    list.addAll(FStatusHistoryList);
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'condition') Then
    list.addAll(FConditionList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'managingOrganization') Then
     list.add(FManagingOrganization.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'referralRequest') Then
    list.addAll(FReferralRequestList);
  if (child_name = 'careManager') Then
     list.add(FCareManager.Link);
  if (child_name = 'careTeam') Then
    list.addAll(FCareTeamList);
end;

procedure TFhirEpisodeOfCare.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'statusHistory', '', FStatusHistoryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'condition', 'Reference(Condition)', FConditionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'managingOrganization', 'Reference(Organization)', FManagingOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'referralRequest', 'Reference(ReferralRequest)', FReferralRequestList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'careManager', 'Reference(Practitioner)', FCareManager.Link));{2}
  oList.add(TFHIRProperty.create(self, 'careTeam', '', FCareTeamList.Link)){3};
end;

procedure TFhirEpisodeOfCare.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'statusHistory') then StatusHistoryList.add(propValue as TFhirEpisodeOfCareStatusHistory){2}
  else if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'condition') then ConditionList.add(propValue as TFhirReference{TFhirCondition}){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'managingOrganization') then ManagingOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'referralRequest') then ReferralRequestList.add(propValue as TFhirReference{TFhirReferralRequest}){2}
  else if (propName = 'careManager') then CareManager := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'careTeam') then CareTeamList.add(propValue as TFhirEpisodeOfCareCareTeam){2}
  else inherited;
end;

function TFhirEpisodeOfCare.FhirType : string;
begin
  result := 'EpisodeOfCare';
end;

function TFhirEpisodeOfCare.Link : TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare(inherited Link);
end;

function TFhirEpisodeOfCare.Clone : TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare(inherited Clone);
end;

{ TFhirEpisodeOfCare }

Function TFhirEpisodeOfCare.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirEpisodeOfCare.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirEpisodeOfCare.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirEpisodeOfCare.GetStatusST : TFhirEpisodeOfCareStatus;
begin
  if FStatus = nil then
    result := TFhirEpisodeOfCareStatus(0)
  else
    result := TFhirEpisodeOfCareStatus(StringArrayIndexOfSensitive(CODES_TFhirEpisodeOfCareStatus, FStatus.value));
end;

Procedure TFhirEpisodeOfCare.SetStatusST(value : TFhirEpisodeOfCareStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirEpisodeOfCareStatus[value]);
end;

Function TFhirEpisodeOfCare.GetStatusHistoryList : TFhirEpisodeOfCareStatusHistoryList;
begin
  if FStatusHistoryList = nil then
    FStatusHistoryList := TFhirEpisodeOfCareStatusHistoryList.Create;
  result := FStatusHistoryList;
end;

Function TFhirEpisodeOfCare.GetHasStatusHistoryList : boolean;
begin
  result := (FStatusHistoryList <> nil) and (FStatusHistoryList.count > 0);
end;

Function TFhirEpisodeOfCare.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirEpisodeOfCare.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Function TFhirEpisodeOfCare.GetConditionList : TFhirReferenceList{TFhirCondition};
begin
  if FConditionList = nil then
    FConditionList := TFhirReferenceList{TFhirCondition}.Create;
  result := FConditionList;
end;

Function TFhirEpisodeOfCare.GetHasConditionList : boolean;
begin
  result := (FConditionList <> nil) and (FConditionList.count > 0);
end;

Procedure TFhirEpisodeOfCare.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirEpisodeOfCare.SetManagingOrganization(value : TFhirReference{TFhirOrganization});
begin
  FManagingOrganization.free;
  FManagingOrganization := value;
end;

Procedure TFhirEpisodeOfCare.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Function TFhirEpisodeOfCare.GetReferralRequestList : TFhirReferenceList{TFhirReferralRequest};
begin
  if FReferralRequestList = nil then
    FReferralRequestList := TFhirReferenceList{TFhirReferralRequest}.Create;
  result := FReferralRequestList;
end;

Function TFhirEpisodeOfCare.GetHasReferralRequestList : boolean;
begin
  result := (FReferralRequestList <> nil) and (FReferralRequestList.count > 0);
end;

Procedure TFhirEpisodeOfCare.SetCareManager(value : TFhirReference{TFhirPractitioner});
begin
  FCareManager.free;
  FCareManager := value;
end;

Function TFhirEpisodeOfCare.GetCareTeamList : TFhirEpisodeOfCareCareTeamList;
begin
  if FCareTeamList = nil then
    FCareTeamList := TFhirEpisodeOfCareCareTeamList.Create;
  result := FCareTeamList;
end;

Function TFhirEpisodeOfCare.GetHasCareTeamList : boolean;
begin
  result := (FCareTeamList <> nil) and (FCareTeamList.count > 0);
end;


{ TFhirEpisodeOfCareListEnumerator }

Constructor TFhirEpisodeOfCareListEnumerator.Create(list : TFhirEpisodeOfCareList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirEpisodeOfCareListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirEpisodeOfCareListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirEpisodeOfCareListEnumerator.GetCurrent : TFhirEpisodeOfCare;
begin
  Result := FList[FIndex];
end;


{ TFhirEpisodeOfCareList }
procedure TFhirEpisodeOfCareList.AddItem(value: TFhirEpisodeOfCare);
begin
  assert(value.ClassName = 'TFhirEpisodeOfCare', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirEpisodeOfCare');
  add(value);
end;


function TFhirEpisodeOfCareList.Append: TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareList.ClearItems;
begin
  Clear;
end;

function TFhirEpisodeOfCareList.GetEnumerator : TFhirEpisodeOfCareListEnumerator;
begin
  result := TFhirEpisodeOfCareListEnumerator.Create(self.link);
end;

function TFhirEpisodeOfCareList.Clone: TFhirEpisodeOfCareList;
begin
  result := TFhirEpisodeOfCareList(inherited Clone);
end;

function TFhirEpisodeOfCareList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirEpisodeOfCareList.GetItemN(index: Integer): TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareList.IndexOf(value: TFhirEpisodeOfCare): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirEpisodeOfCareList.Insert(index: Integer): TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirEpisodeOfCareList.InsertItem(index: Integer; value: TFhirEpisodeOfCare);
begin
  assert(value is TFhirEpisodeOfCare);
  Inherited Insert(index, value);
end;

function TFhirEpisodeOfCareList.Item(index: Integer): TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare(ObjectByIndex[index]);
end;

function TFhirEpisodeOfCareList.Link: TFhirEpisodeOfCareList;
begin
  result := TFhirEpisodeOfCareList(inherited Link);
end;

procedure TFhirEpisodeOfCareList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirEpisodeOfCareList.SetItemByIndex(index: Integer; value: TFhirEpisodeOfCare);
begin
  assert(value is TFhirEpisodeOfCare);
  FhirEpisodeOfCares[index] := value;
end;

procedure TFhirEpisodeOfCareList.SetItemN(index: Integer; value: TFhirEpisodeOfCare);
begin
  assert(value is TFhirEpisodeOfCare);
  ObjectByIndex[index] := value;
end;

{ TFhirExplanationOfBenefit }

constructor TFhirExplanationOfBenefit.Create;
begin
  inherited;
end;

destructor TFhirExplanationOfBenefit.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FOutcome.free;
  FDisposition.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  inherited;
end;

function TFhirExplanationOfBenefit.GetResourceType : TFhirResourceType;
begin
  result := frtExplanationOfBenefit;
end;

procedure TFhirExplanationOfBenefit.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirExplanationOfBenefit(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirExplanationOfBenefit(oSource).FIdentifierList);
  end;
  request := TFhirExplanationOfBenefit(oSource).request.Clone;
  FOutcome := TFhirExplanationOfBenefit(oSource).FOutcome.Link;
  dispositionElement := TFhirExplanationOfBenefit(oSource).dispositionElement.Clone;
  ruleset := TFhirExplanationOfBenefit(oSource).ruleset.Clone;
  originalRuleset := TFhirExplanationOfBenefit(oSource).originalRuleset.Clone;
  createdElement := TFhirExplanationOfBenefit(oSource).createdElement.Clone;
  organization := TFhirExplanationOfBenefit(oSource).organization.Clone;
  requestProvider := TFhirExplanationOfBenefit(oSource).requestProvider.Clone;
  requestOrganization := TFhirExplanationOfBenefit(oSource).requestOrganization.Clone;
end;

procedure TFhirExplanationOfBenefit.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
end;

procedure TFhirExplanationOfBenefit.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Claim)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
end;

procedure TFhirExplanationOfBenefit.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirClaim}{4b}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else inherited;
end;

function TFhirExplanationOfBenefit.FhirType : string;
begin
  result := 'ExplanationOfBenefit';
end;

function TFhirExplanationOfBenefit.Link : TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit(inherited Link);
end;

function TFhirExplanationOfBenefit.Clone : TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit(inherited Clone);
end;

{ TFhirExplanationOfBenefit }

Function TFhirExplanationOfBenefit.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirExplanationOfBenefit.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirExplanationOfBenefit.SetRequest(value : TFhirReference{TFhirClaim});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirExplanationOfBenefit.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirExplanationOfBenefit.GetOutcomeST : TFhirRemittanceOutcome;
begin
  if FOutcome = nil then
    result := TFhirRemittanceOutcome(0)
  else
    result := TFhirRemittanceOutcome(StringArrayIndexOfSensitive(CODES_TFhirRemittanceOutcome, FOutcome.value));
end;

Procedure TFhirExplanationOfBenefit.SetOutcomeST(value : TFhirRemittanceOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirRemittanceOutcome[value]);
end;

Procedure TFhirExplanationOfBenefit.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirExplanationOfBenefit.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirExplanationOfBenefit.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirExplanationOfBenefit.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirExplanationOfBenefit.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirExplanationOfBenefit.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirExplanationOfBenefit.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirExplanationOfBenefit.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirExplanationOfBenefit.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirExplanationOfBenefit.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirExplanationOfBenefit.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;


{ TFhirExplanationOfBenefitListEnumerator }

Constructor TFhirExplanationOfBenefitListEnumerator.Create(list : TFhirExplanationOfBenefitList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirExplanationOfBenefitListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirExplanationOfBenefitListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirExplanationOfBenefitListEnumerator.GetCurrent : TFhirExplanationOfBenefit;
begin
  Result := FList[FIndex];
end;


{ TFhirExplanationOfBenefitList }
procedure TFhirExplanationOfBenefitList.AddItem(value: TFhirExplanationOfBenefit);
begin
  assert(value.ClassName = 'TFhirExplanationOfBenefit', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirExplanationOfBenefit');
  add(value);
end;


function TFhirExplanationOfBenefitList.Append: TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirExplanationOfBenefitList.ClearItems;
begin
  Clear;
end;

function TFhirExplanationOfBenefitList.GetEnumerator : TFhirExplanationOfBenefitListEnumerator;
begin
  result := TFhirExplanationOfBenefitListEnumerator.Create(self.link);
end;

function TFhirExplanationOfBenefitList.Clone: TFhirExplanationOfBenefitList;
begin
  result := TFhirExplanationOfBenefitList(inherited Clone);
end;

function TFhirExplanationOfBenefitList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirExplanationOfBenefitList.GetItemN(index: Integer): TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit(ObjectByIndex[index]);
end;

function TFhirExplanationOfBenefitList.IndexOf(value: TFhirExplanationOfBenefit): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirExplanationOfBenefitList.Insert(index: Integer): TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirExplanationOfBenefitList.InsertItem(index: Integer; value: TFhirExplanationOfBenefit);
begin
  assert(value is TFhirExplanationOfBenefit);
  Inherited Insert(index, value);
end;

function TFhirExplanationOfBenefitList.Item(index: Integer): TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit(ObjectByIndex[index]);
end;

function TFhirExplanationOfBenefitList.Link: TFhirExplanationOfBenefitList;
begin
  result := TFhirExplanationOfBenefitList(inherited Link);
end;

procedure TFhirExplanationOfBenefitList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirExplanationOfBenefitList.SetItemByIndex(index: Integer; value: TFhirExplanationOfBenefit);
begin
  assert(value is TFhirExplanationOfBenefit);
  FhirExplanationOfBenefits[index] := value;
end;

procedure TFhirExplanationOfBenefitList.SetItemN(index: Integer; value: TFhirExplanationOfBenefit);
begin
  assert(value is TFhirExplanationOfBenefit);
  ObjectByIndex[index] := value;
end;

{ TFhirFamilyMemberHistoryCondition }

constructor TFhirFamilyMemberHistoryCondition.Create;
begin
  inherited;
end;

destructor TFhirFamilyMemberHistoryCondition.Destroy;
begin
  FCode.free;
  FOutcome.free;
  FOnset.free;
  FNote.free;
  inherited;
end;

procedure TFhirFamilyMemberHistoryCondition.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirFamilyMemberHistoryCondition(oSource).code.Clone;
  outcome := TFhirFamilyMemberHistoryCondition(oSource).outcome.Clone;
  onset := TFhirFamilyMemberHistoryCondition(oSource).onset.Clone;
  note := TFhirFamilyMemberHistoryCondition(oSource).note.Clone;
end;

procedure TFhirFamilyMemberHistoryCondition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'onset[x]') Then
     list.add(FOnset.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
end;

procedure TFhirFamilyMemberHistoryCondition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'CodeableConcept', FOutcome.Link));{2}
  oList.add(TFHIRProperty.create(self, 'onset[x]', 'Age|Range|Period|string', FOnset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNote.Link));{2}
end;

procedure TFhirFamilyMemberHistoryCondition.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'outcome') then Outcome := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('onset')) then Onset := propValue as TFhirType{4}
  else if (propName = 'note') then Note := propValue as TFhirAnnotation{4b}
  else inherited;
end;

function TFhirFamilyMemberHistoryCondition.FhirType : string;
begin
  result := 'condition';
end;

function TFhirFamilyMemberHistoryCondition.Link : TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition(inherited Link);
end;

function TFhirFamilyMemberHistoryCondition.Clone : TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition(inherited Clone);
end;

function TFhirFamilyMemberHistoryCondition.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirFamilyMemberHistoryCondition;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirFamilyMemberHistoryCondition)) then
    result := false
  else
  begin
    o := TFhirFamilyMemberHistoryCondition(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(outcomeElement, o.outcomeElement, true) and 
      compareDeep(onsetElement, o.onsetElement, true) and compareDeep(noteElement, o.noteElement, true);
  end;
end;

function TFhirFamilyMemberHistoryCondition.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirFamilyMemberHistoryCondition;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirFamilyMemberHistoryCondition)) then
    result := false
  else
  begin
    o := TFhirFamilyMemberHistoryCondition(other);
    result := true;
  end;
end;

{ TFhirFamilyMemberHistoryCondition }

Procedure TFhirFamilyMemberHistoryCondition.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirFamilyMemberHistoryCondition.SetOutcome(value : TFhirCodeableConcept);
begin
  FOutcome.free;
  FOutcome := value;
end;

Procedure TFhirFamilyMemberHistoryCondition.SetOnset(value : TFhirType);
begin
  FOnset.free;
  FOnset := value;
end;

Procedure TFhirFamilyMemberHistoryCondition.SetNote(value : TFhirAnnotation);
begin
  FNote.free;
  FNote := value;
end;


{ TFhirFamilyMemberHistoryConditionListEnumerator }

Constructor TFhirFamilyMemberHistoryConditionListEnumerator.Create(list : TFhirFamilyMemberHistoryConditionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirFamilyMemberHistoryConditionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirFamilyMemberHistoryConditionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirFamilyMemberHistoryConditionListEnumerator.GetCurrent : TFhirFamilyMemberHistoryCondition;
begin
  Result := FList[FIndex];
end;


{ TFhirFamilyMemberHistoryConditionList }
procedure TFhirFamilyMemberHistoryConditionList.AddItem(value: TFhirFamilyMemberHistoryCondition);
begin
  assert(value.ClassName = 'TFhirFamilyMemberHistoryCondition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirFamilyMemberHistoryCondition');
  add(value);
end;


function TFhirFamilyMemberHistoryConditionList.Append: TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFamilyMemberHistoryConditionList.ClearItems;
begin
  Clear;
end;

function TFhirFamilyMemberHistoryConditionList.GetEnumerator : TFhirFamilyMemberHistoryConditionListEnumerator;
begin
  result := TFhirFamilyMemberHistoryConditionListEnumerator.Create(self.link);
end;

function TFhirFamilyMemberHistoryConditionList.Clone: TFhirFamilyMemberHistoryConditionList;
begin
  result := TFhirFamilyMemberHistoryConditionList(inherited Clone);
end;

function TFhirFamilyMemberHistoryConditionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirFamilyMemberHistoryConditionList.GetItemN(index: Integer): TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition(ObjectByIndex[index]);
end;

function TFhirFamilyMemberHistoryConditionList.IndexOf(value: TFhirFamilyMemberHistoryCondition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirFamilyMemberHistoryConditionList.Insert(index: Integer): TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFamilyMemberHistoryConditionList.InsertItem(index: Integer; value: TFhirFamilyMemberHistoryCondition);
begin
  assert(value is TFhirFamilyMemberHistoryCondition);
  Inherited Insert(index, value);
end;

function TFhirFamilyMemberHistoryConditionList.Item(index: Integer): TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition(ObjectByIndex[index]);
end;

function TFhirFamilyMemberHistoryConditionList.Link: TFhirFamilyMemberHistoryConditionList;
begin
  result := TFhirFamilyMemberHistoryConditionList(inherited Link);
end;

procedure TFhirFamilyMemberHistoryConditionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirFamilyMemberHistoryConditionList.SetItemByIndex(index: Integer; value: TFhirFamilyMemberHistoryCondition);
begin
  assert(value is TFhirFamilyMemberHistoryCondition);
  FhirFamilyMemberHistoryConditions[index] := value;
end;

procedure TFhirFamilyMemberHistoryConditionList.SetItemN(index: Integer; value: TFhirFamilyMemberHistoryCondition);
begin
  assert(value is TFhirFamilyMemberHistoryCondition);
  ObjectByIndex[index] := value;
end;

{ TFhirFamilyMemberHistory }

constructor TFhirFamilyMemberHistory.Create;
begin
  inherited;
end;

destructor TFhirFamilyMemberHistory.Destroy;
begin
  FIdentifierList.Free;
  FPatient.free;
  FDate.free;
  FStatus.free;
  FName.free;
  FRelationship.free;
  FGender.free;
  FBorn.free;
  FAge.free;
  FDeceased.free;
  FNote.free;
  FConditionList.Free;
  inherited;
end;

function TFhirFamilyMemberHistory.GetResourceType : TFhirResourceType;
begin
  result := frtFamilyMemberHistory;
end;

procedure TFhirFamilyMemberHistory.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirFamilyMemberHistory(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirFamilyMemberHistory(oSource).FIdentifierList);
  end;
  patient := TFhirFamilyMemberHistory(oSource).patient.Clone;
  dateElement := TFhirFamilyMemberHistory(oSource).dateElement.Clone;
  FStatus := TFhirFamilyMemberHistory(oSource).FStatus.Link;
  nameElement := TFhirFamilyMemberHistory(oSource).nameElement.Clone;
  relationship := TFhirFamilyMemberHistory(oSource).relationship.Clone;
  FGender := TFhirFamilyMemberHistory(oSource).FGender.Link;
  born := TFhirFamilyMemberHistory(oSource).born.Clone;
  age := TFhirFamilyMemberHistory(oSource).age.Clone;
  deceased := TFhirFamilyMemberHistory(oSource).deceased.Clone;
  note := TFhirFamilyMemberHistory(oSource).note.Clone;
  if (TFhirFamilyMemberHistory(oSource).FConditionList = nil) then
  begin
    FConditionList.free;
    FConditionList := nil;
  end
  else
  begin
    FConditionList := TFhirFamilyMemberHistoryConditionList.Create;
    FConditionList.Assign(TFhirFamilyMemberHistory(oSource).FConditionList);
  end;
end;

procedure TFhirFamilyMemberHistory.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'relationship') Then
     list.add(FRelationship.Link);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'born[x]') Then
     list.add(FBorn.Link);
  if (child_name = 'age[x]') Then
     list.add(FAge.Link);
  if (child_name = 'deceased[x]') Then
     list.add(FDeceased.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'condition') Then
    list.addAll(FConditionList);
end;

procedure TFhirFamilyMemberHistory.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relationship', 'CodeableConcept', FRelationship.Link));{2}
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'born[x]', 'Period|date|string', FBorn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'age[x]', 'Age|Range|string', FAge.Link));{2}
  oList.add(TFHIRProperty.create(self, 'deceased[x]', 'boolean|Age|Range|date|string', FDeceased.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'condition', '', FConditionList.Link)){3};
end;

procedure TFhirFamilyMemberHistory.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'relationship') then Relationship := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName.startsWith('born')) then Born := propValue as TFhirType{4}
  else if (propName.startsWith('age')) then Age := propValue as TFhirType{4}
  else if (propName.startsWith('deceased')) then Deceased := propValue as TFhirType{4}
  else if (propName = 'note') then Note := propValue as TFhirAnnotation{4b}
  else if (propName = 'condition') then ConditionList.add(propValue as TFhirFamilyMemberHistoryCondition){2}
  else inherited;
end;

function TFhirFamilyMemberHistory.FhirType : string;
begin
  result := 'FamilyMemberHistory';
end;

function TFhirFamilyMemberHistory.Link : TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory(inherited Link);
end;

function TFhirFamilyMemberHistory.Clone : TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory(inherited Clone);
end;

{ TFhirFamilyMemberHistory }

Function TFhirFamilyMemberHistory.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirFamilyMemberHistory.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirFamilyMemberHistory.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirFamilyMemberHistory.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirFamilyMemberHistory.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirFamilyMemberHistory.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirFamilyMemberHistory.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirFamilyMemberHistory.GetStatusST : TFhirHistoryStatus;
begin
  if FStatus = nil then
    result := TFhirHistoryStatus(0)
  else
    result := TFhirHistoryStatus(StringArrayIndexOfSensitive(CODES_TFhirHistoryStatus, FStatus.value));
end;

Procedure TFhirFamilyMemberHistory.SetStatusST(value : TFhirHistoryStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirHistoryStatus[value]);
end;

Procedure TFhirFamilyMemberHistory.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirFamilyMemberHistory.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirFamilyMemberHistory.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirFamilyMemberHistory.SetRelationship(value : TFhirCodeableConcept);
begin
  FRelationship.free;
  FRelationship := value;
end;

Procedure TFhirFamilyMemberHistory.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirFamilyMemberHistory.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirFamilyMemberHistory.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirFamilyMemberHistory.SetBorn(value : TFhirType);
begin
  FBorn.free;
  FBorn := value;
end;

Procedure TFhirFamilyMemberHistory.SetAge(value : TFhirType);
begin
  FAge.free;
  FAge := value;
end;

Procedure TFhirFamilyMemberHistory.SetDeceased(value : TFhirType);
begin
  FDeceased.free;
  FDeceased := value;
end;

Procedure TFhirFamilyMemberHistory.SetNote(value : TFhirAnnotation);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirFamilyMemberHistory.GetConditionList : TFhirFamilyMemberHistoryConditionList;
begin
  if FConditionList = nil then
    FConditionList := TFhirFamilyMemberHistoryConditionList.Create;
  result := FConditionList;
end;

Function TFhirFamilyMemberHistory.GetHasConditionList : boolean;
begin
  result := (FConditionList <> nil) and (FConditionList.count > 0);
end;


{ TFhirFamilyMemberHistoryListEnumerator }

Constructor TFhirFamilyMemberHistoryListEnumerator.Create(list : TFhirFamilyMemberHistoryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirFamilyMemberHistoryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirFamilyMemberHistoryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirFamilyMemberHistoryListEnumerator.GetCurrent : TFhirFamilyMemberHistory;
begin
  Result := FList[FIndex];
end;


{ TFhirFamilyMemberHistoryList }
procedure TFhirFamilyMemberHistoryList.AddItem(value: TFhirFamilyMemberHistory);
begin
  assert(value.ClassName = 'TFhirFamilyMemberHistory', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirFamilyMemberHistory');
  add(value);
end;


function TFhirFamilyMemberHistoryList.Append: TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFamilyMemberHistoryList.ClearItems;
begin
  Clear;
end;

function TFhirFamilyMemberHistoryList.GetEnumerator : TFhirFamilyMemberHistoryListEnumerator;
begin
  result := TFhirFamilyMemberHistoryListEnumerator.Create(self.link);
end;

function TFhirFamilyMemberHistoryList.Clone: TFhirFamilyMemberHistoryList;
begin
  result := TFhirFamilyMemberHistoryList(inherited Clone);
end;

function TFhirFamilyMemberHistoryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirFamilyMemberHistoryList.GetItemN(index: Integer): TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory(ObjectByIndex[index]);
end;

function TFhirFamilyMemberHistoryList.IndexOf(value: TFhirFamilyMemberHistory): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirFamilyMemberHistoryList.Insert(index: Integer): TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFamilyMemberHistoryList.InsertItem(index: Integer; value: TFhirFamilyMemberHistory);
begin
  assert(value is TFhirFamilyMemberHistory);
  Inherited Insert(index, value);
end;

function TFhirFamilyMemberHistoryList.Item(index: Integer): TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory(ObjectByIndex[index]);
end;

function TFhirFamilyMemberHistoryList.Link: TFhirFamilyMemberHistoryList;
begin
  result := TFhirFamilyMemberHistoryList(inherited Link);
end;

procedure TFhirFamilyMemberHistoryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirFamilyMemberHistoryList.SetItemByIndex(index: Integer; value: TFhirFamilyMemberHistory);
begin
  assert(value is TFhirFamilyMemberHistory);
  FhirFamilyMemberHistories[index] := value;
end;

procedure TFhirFamilyMemberHistoryList.SetItemN(index: Integer; value: TFhirFamilyMemberHistory);
begin
  assert(value is TFhirFamilyMemberHistory);
  ObjectByIndex[index] := value;
end;

{ TFhirFlag }

constructor TFhirFlag.Create;
begin
  inherited;
end;

destructor TFhirFlag.Destroy;
begin
  FIdentifierList.Free;
  FCategory.free;
  FStatus.free;
  FPeriod.free;
  FSubject.free;
  FEncounter.free;
  FAuthor.free;
  FCode.free;
  inherited;
end;

function TFhirFlag.GetResourceType : TFhirResourceType;
begin
  result := frtFlag;
end;

procedure TFhirFlag.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirFlag(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirFlag(oSource).FIdentifierList);
  end;
  category := TFhirFlag(oSource).category.Clone;
  FStatus := TFhirFlag(oSource).FStatus.Link;
  period := TFhirFlag(oSource).period.Clone;
  subject := TFhirFlag(oSource).subject.Clone;
  encounter := TFhirFlag(oSource).encounter.Clone;
  author := TFhirFlag(oSource).author.Clone;
  code := TFhirFlag(oSource).code.Clone;
end;

procedure TFhirFlag.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
end;

procedure TFhirFlag.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Location|Group|Organization|Practitioner)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Device|Organization|Patient|Practitioner)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
end;

procedure TFhirFlag.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirFlag.FhirType : string;
begin
  result := 'Flag';
end;

function TFhirFlag.Link : TFhirFlag;
begin
  result := TFhirFlag(inherited Link);
end;

function TFhirFlag.Clone : TFhirFlag;
begin
  result := TFhirFlag(inherited Clone);
end;

{ TFhirFlag }

Function TFhirFlag.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirFlag.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirFlag.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirFlag.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirFlag.GetStatusST : TFhirFlagStatus;
begin
  if FStatus = nil then
    result := TFhirFlagStatus(0)
  else
    result := TFhirFlagStatus(StringArrayIndexOfSensitive(CODES_TFhirFlagStatus, FStatus.value));
end;

Procedure TFhirFlag.SetStatusST(value : TFhirFlagStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirFlagStatus[value]);
end;

Procedure TFhirFlag.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirFlag.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirFlag.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirFlag.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirFlag.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;


{ TFhirFlagListEnumerator }

Constructor TFhirFlagListEnumerator.Create(list : TFhirFlagList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirFlagListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirFlagListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirFlagListEnumerator.GetCurrent : TFhirFlag;
begin
  Result := FList[FIndex];
end;


{ TFhirFlagList }
procedure TFhirFlagList.AddItem(value: TFhirFlag);
begin
  assert(value.ClassName = 'TFhirFlag', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirFlag');
  add(value);
end;


function TFhirFlagList.Append: TFhirFlag;
begin
  result := TFhirFlag.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFlagList.ClearItems;
begin
  Clear;
end;

function TFhirFlagList.GetEnumerator : TFhirFlagListEnumerator;
begin
  result := TFhirFlagListEnumerator.Create(self.link);
end;

function TFhirFlagList.Clone: TFhirFlagList;
begin
  result := TFhirFlagList(inherited Clone);
end;

function TFhirFlagList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirFlagList.GetItemN(index: Integer): TFhirFlag;
begin
  result := TFhirFlag(ObjectByIndex[index]);
end;

function TFhirFlagList.IndexOf(value: TFhirFlag): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirFlagList.Insert(index: Integer): TFhirFlag;
begin
  result := TFhirFlag.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirFlagList.InsertItem(index: Integer; value: TFhirFlag);
begin
  assert(value is TFhirFlag);
  Inherited Insert(index, value);
end;

function TFhirFlagList.Item(index: Integer): TFhirFlag;
begin
  result := TFhirFlag(ObjectByIndex[index]);
end;

function TFhirFlagList.Link: TFhirFlagList;
begin
  result := TFhirFlagList(inherited Link);
end;

procedure TFhirFlagList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirFlagList.SetItemByIndex(index: Integer; value: TFhirFlag);
begin
  assert(value is TFhirFlag);
  FhirFlags[index] := value;
end;

procedure TFhirFlagList.SetItemN(index: Integer; value: TFhirFlag);
begin
  assert(value is TFhirFlag);
  ObjectByIndex[index] := value;
end;

{ TFhirGoalOutcome }

constructor TFhirGoalOutcome.Create;
begin
  inherited;
end;

destructor TFhirGoalOutcome.Destroy;
begin
  FResult.free;
  inherited;
end;

procedure TFhirGoalOutcome.Assign(oSource : TAdvObject);
begin
  inherited;
  result := TFhirGoalOutcome(oSource).result.Clone;
end;

procedure TFhirGoalOutcome.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'result[x]') Then
     list.add(FResult.Link);
end;

procedure TFhirGoalOutcome.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'result[x]', 'CodeableConcept|Reference(Observation)', FResult.Link));{2}
end;

procedure TFhirGoalOutcome.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('result')) then Result := propValue as TFhirType{4}
  else inherited;
end;

function TFhirGoalOutcome.FhirType : string;
begin
  result := 'outcome';
end;

function TFhirGoalOutcome.Link : TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome(inherited Link);
end;

function TFhirGoalOutcome.Clone : TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome(inherited Clone);
end;

function TFhirGoalOutcome.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirGoalOutcome;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirGoalOutcome)) then
    result := false
  else
  begin
    o := TFhirGoalOutcome(other);
    result := compareDeep(resultElement, o.resultElement, true);
  end;
end;

function TFhirGoalOutcome.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirGoalOutcome;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirGoalOutcome)) then
    result := false
  else
  begin
    o := TFhirGoalOutcome(other);
    result := true;
  end;
end;

{ TFhirGoalOutcome }

Procedure TFhirGoalOutcome.SetResult(value : TFhirType);
begin
  FResult.free;
  FResult := value;
end;


{ TFhirGoalOutcomeListEnumerator }

Constructor TFhirGoalOutcomeListEnumerator.Create(list : TFhirGoalOutcomeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirGoalOutcomeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirGoalOutcomeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirGoalOutcomeListEnumerator.GetCurrent : TFhirGoalOutcome;
begin
  Result := FList[FIndex];
end;


{ TFhirGoalOutcomeList }
procedure TFhirGoalOutcomeList.AddItem(value: TFhirGoalOutcome);
begin
  assert(value.ClassName = 'TFhirGoalOutcome', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirGoalOutcome');
  add(value);
end;


function TFhirGoalOutcomeList.Append: TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGoalOutcomeList.ClearItems;
begin
  Clear;
end;

function TFhirGoalOutcomeList.GetEnumerator : TFhirGoalOutcomeListEnumerator;
begin
  result := TFhirGoalOutcomeListEnumerator.Create(self.link);
end;

function TFhirGoalOutcomeList.Clone: TFhirGoalOutcomeList;
begin
  result := TFhirGoalOutcomeList(inherited Clone);
end;

function TFhirGoalOutcomeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirGoalOutcomeList.GetItemN(index: Integer): TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome(ObjectByIndex[index]);
end;

function TFhirGoalOutcomeList.IndexOf(value: TFhirGoalOutcome): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirGoalOutcomeList.Insert(index: Integer): TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGoalOutcomeList.InsertItem(index: Integer; value: TFhirGoalOutcome);
begin
  assert(value is TFhirGoalOutcome);
  Inherited Insert(index, value);
end;

function TFhirGoalOutcomeList.Item(index: Integer): TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome(ObjectByIndex[index]);
end;

function TFhirGoalOutcomeList.Link: TFhirGoalOutcomeList;
begin
  result := TFhirGoalOutcomeList(inherited Link);
end;

procedure TFhirGoalOutcomeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirGoalOutcomeList.SetItemByIndex(index: Integer; value: TFhirGoalOutcome);
begin
  assert(value is TFhirGoalOutcome);
  FhirGoalOutcomes[index] := value;
end;

procedure TFhirGoalOutcomeList.SetItemN(index: Integer; value: TFhirGoalOutcome);
begin
  assert(value is TFhirGoalOutcome);
  ObjectByIndex[index] := value;
end;

{ TFhirGoal }

constructor TFhirGoal.Create;
begin
  inherited;
end;

destructor TFhirGoal.Destroy;
begin
  FIdentifierList.Free;
  FSubject.free;
  FStart.free;
  FTarget.free;
  FCategoryList.Free;
  FDescription.free;
  FStatus.free;
  FStatusDate.free;
  FStatusReason.free;
  FAuthor.free;
  FPriority.free;
  FAddressesList.Free;
  FNoteList.Free;
  FOutcomeList.Free;
  inherited;
end;

function TFhirGoal.GetResourceType : TFhirResourceType;
begin
  result := frtGoal;
end;

procedure TFhirGoal.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirGoal(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirGoal(oSource).FIdentifierList);
  end;
  subject := TFhirGoal(oSource).subject.Clone;
  start := TFhirGoal(oSource).start.Clone;
  target := TFhirGoal(oSource).target.Clone;
  if (TFhirGoal(oSource).FCategoryList = nil) then
  begin
    FCategoryList.free;
    FCategoryList := nil;
  end
  else
  begin
    FCategoryList := TFhirCodeableConceptList.Create;
    FCategoryList.Assign(TFhirGoal(oSource).FCategoryList);
  end;
  descriptionElement := TFhirGoal(oSource).descriptionElement.Clone;
  FStatus := TFhirGoal(oSource).FStatus.Link;
  statusDateElement := TFhirGoal(oSource).statusDateElement.Clone;
  statusReason := TFhirGoal(oSource).statusReason.Clone;
  author := TFhirGoal(oSource).author.Clone;
  priority := TFhirGoal(oSource).priority.Clone;
  if (TFhirGoal(oSource).FAddressesList = nil) then
  begin
    FAddressesList.free;
    FAddressesList := nil;
  end
  else
  begin
    FAddressesList := TFhirReferenceList{Resource}.Create;
    FAddressesList.Assign(TFhirGoal(oSource).FAddressesList);
  end;
  if (TFhirGoal(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirAnnotationList.Create;
    FNoteList.Assign(TFhirGoal(oSource).FNoteList);
  end;
  if (TFhirGoal(oSource).FOutcomeList = nil) then
  begin
    FOutcomeList.free;
    FOutcomeList := nil;
  end
  else
  begin
    FOutcomeList := TFhirGoalOutcomeList.Create;
    FOutcomeList.Assign(TFhirGoal(oSource).FOutcomeList);
  end;
end;

procedure TFhirGoal.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'start[x]') Then
     list.add(FStart.Link);
  if (child_name = 'target[x]') Then
     list.add(FTarget.Link);
  if (child_name = 'category') Then
    list.addAll(FCategoryList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'statusDate') Then
     list.add(FStatusDate.Link);
  if (child_name = 'statusReason') Then
     list.add(FStatusReason.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'addresses') Then
    list.addAll(FAddressesList);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
  if (child_name = 'outcome') Then
    list.addAll(FOutcomeList);
end;

procedure TFhirGoal.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Organization)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'start[x]', 'date|CodeableConcept', FStart.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target[x]', 'date|Duration', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategoryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'statusDate', 'date', FStatusDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'statusReason', 'CodeableConcept', FStatusReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Patient|Practitioner|RelatedPerson)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'CodeableConcept', FPriority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'addresses', 'Reference(Condition|Observation|MedicationStatement|NutritionOrder|ProcedureRequest|RiskAssessment)', FAddressesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNoteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'outcome', '', FOutcomeList.Link)){3};
end;

procedure TFhirGoal.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName.startsWith('start')) then Start := propValue as TFhirType{4}
  else if (propName.startsWith('target')) then Target := propValue as TFhirType{4}
  else if (propName = 'category') then CategoryList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'statusDate') then StatusDateElement := propValue as TFhirDate{5a}
  else if (propName = 'statusReason') then StatusReason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'priority') then Priority := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'addresses') then AddressesList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'note') then NoteList.add(propValue as TFhirAnnotation){2}
  else if (propName = 'outcome') then OutcomeList.add(propValue as TFhirGoalOutcome){2}
  else inherited;
end;

function TFhirGoal.FhirType : string;
begin
  result := 'Goal';
end;

function TFhirGoal.Link : TFhirGoal;
begin
  result := TFhirGoal(inherited Link);
end;

function TFhirGoal.Clone : TFhirGoal;
begin
  result := TFhirGoal(inherited Clone);
end;

{ TFhirGoal }

Function TFhirGoal.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirGoal.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirGoal.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirGoal.SetStart(value : TFhirType);
begin
  FStart.free;
  FStart := value;
end;

Procedure TFhirGoal.SetTarget(value : TFhirType);
begin
  FTarget.free;
  FTarget := value;
end;

Function TFhirGoal.GetCategoryList : TFhirCodeableConceptList;
begin
  if FCategoryList = nil then
    FCategoryList := TFhirCodeableConceptList.Create;
  result := FCategoryList;
end;

Function TFhirGoal.GetHasCategoryList : boolean;
begin
  result := (FCategoryList <> nil) and (FCategoryList.count > 0);
end;

Procedure TFhirGoal.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirGoal.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirGoal.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirGoal.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirGoal.GetStatusST : TFhirGoalStatus;
begin
  if FStatus = nil then
    result := TFhirGoalStatus(0)
  else
    result := TFhirGoalStatus(StringArrayIndexOfSensitive(CODES_TFhirGoalStatus, FStatus.value));
end;

Procedure TFhirGoal.SetStatusST(value : TFhirGoalStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirGoalStatus[value]);
end;

Procedure TFhirGoal.SetStatusDate(value : TFhirDate);
begin
  FStatusDate.free;
  FStatusDate := value;
end;

Function TFhirGoal.GetStatusDateST : TDateAndTime;
begin
  if FStatusDate = nil then
    result := nil
  else
    result := FStatusDate.value;
end;

Procedure TFhirGoal.SetStatusDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStatusDate = nil then
      FStatusDate := TFhirDate.create;
    FStatusDate.value := value
  end
  else if FStatusDate <> nil then
    FStatusDate.value := nil;
end;

Procedure TFhirGoal.SetStatusReason(value : TFhirCodeableConcept);
begin
  FStatusReason.free;
  FStatusReason := value;
end;

Procedure TFhirGoal.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirGoal.SetPriority(value : TFhirCodeableConcept);
begin
  FPriority.free;
  FPriority := value;
end;

Function TFhirGoal.GetAddressesList : TFhirReferenceList{Resource};
begin
  if FAddressesList = nil then
    FAddressesList := TFhirReferenceList{Resource}.Create;
  result := FAddressesList;
end;

Function TFhirGoal.GetHasAddressesList : boolean;
begin
  result := (FAddressesList <> nil) and (FAddressesList.count > 0);
end;

Function TFhirGoal.GetNoteList : TFhirAnnotationList;
begin
  if FNoteList = nil then
    FNoteList := TFhirAnnotationList.Create;
  result := FNoteList;
end;

Function TFhirGoal.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;

Function TFhirGoal.GetOutcomeList : TFhirGoalOutcomeList;
begin
  if FOutcomeList = nil then
    FOutcomeList := TFhirGoalOutcomeList.Create;
  result := FOutcomeList;
end;

Function TFhirGoal.GetHasOutcomeList : boolean;
begin
  result := (FOutcomeList <> nil) and (FOutcomeList.count > 0);
end;


{ TFhirGoalListEnumerator }

Constructor TFhirGoalListEnumerator.Create(list : TFhirGoalList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirGoalListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirGoalListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirGoalListEnumerator.GetCurrent : TFhirGoal;
begin
  Result := FList[FIndex];
end;


{ TFhirGoalList }
procedure TFhirGoalList.AddItem(value: TFhirGoal);
begin
  assert(value.ClassName = 'TFhirGoal', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirGoal');
  add(value);
end;


function TFhirGoalList.Append: TFhirGoal;
begin
  result := TFhirGoal.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGoalList.ClearItems;
begin
  Clear;
end;

function TFhirGoalList.GetEnumerator : TFhirGoalListEnumerator;
begin
  result := TFhirGoalListEnumerator.Create(self.link);
end;

function TFhirGoalList.Clone: TFhirGoalList;
begin
  result := TFhirGoalList(inherited Clone);
end;

function TFhirGoalList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirGoalList.GetItemN(index: Integer): TFhirGoal;
begin
  result := TFhirGoal(ObjectByIndex[index]);
end;

function TFhirGoalList.IndexOf(value: TFhirGoal): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirGoalList.Insert(index: Integer): TFhirGoal;
begin
  result := TFhirGoal.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGoalList.InsertItem(index: Integer; value: TFhirGoal);
begin
  assert(value is TFhirGoal);
  Inherited Insert(index, value);
end;

function TFhirGoalList.Item(index: Integer): TFhirGoal;
begin
  result := TFhirGoal(ObjectByIndex[index]);
end;

function TFhirGoalList.Link: TFhirGoalList;
begin
  result := TFhirGoalList(inherited Link);
end;

procedure TFhirGoalList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirGoalList.SetItemByIndex(index: Integer; value: TFhirGoal);
begin
  assert(value is TFhirGoal);
  FhirGoals[index] := value;
end;

procedure TFhirGoalList.SetItemN(index: Integer; value: TFhirGoal);
begin
  assert(value is TFhirGoal);
  ObjectByIndex[index] := value;
end;

{ TFhirGroupCharacteristic }

constructor TFhirGroupCharacteristic.Create;
begin
  inherited;
end;

destructor TFhirGroupCharacteristic.Destroy;
begin
  FCode.free;
  FValue.free;
  FExclude.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirGroupCharacteristic.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirGroupCharacteristic(oSource).code.Clone;
  value := TFhirGroupCharacteristic(oSource).value.Clone;
  excludeElement := TFhirGroupCharacteristic(oSource).excludeElement.Clone;
  period := TFhirGroupCharacteristic(oSource).period.Clone;
end;

procedure TFhirGroupCharacteristic.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
  if (child_name = 'exclude') Then
     list.add(FExclude.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirGroupCharacteristic.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value[x]', 'CodeableConcept|boolean|Quantity|Range', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'exclude', 'boolean', FExclude.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirGroupCharacteristic.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else if (propName = 'exclude') then ExcludeElement := propValue as TFhirBoolean{5a}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirGroupCharacteristic.FhirType : string;
begin
  result := 'characteristic';
end;

function TFhirGroupCharacteristic.Link : TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic(inherited Link);
end;

function TFhirGroupCharacteristic.Clone : TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic(inherited Clone);
end;

function TFhirGroupCharacteristic.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirGroupCharacteristic;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirGroupCharacteristic)) then
    result := false
  else
  begin
    o := TFhirGroupCharacteristic(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(valueElement, o.valueElement, true) and 
      compareDeep(excludeElement, o.excludeElement, true) and compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirGroupCharacteristic.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirGroupCharacteristic;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirGroupCharacteristic)) then
    result := false
  else
  begin
    o := TFhirGroupCharacteristic(other);
    result := compareValues(excludeElement, o.excludeElement, true);
  end;
end;

{ TFhirGroupCharacteristic }

Procedure TFhirGroupCharacteristic.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirGroupCharacteristic.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;

Procedure TFhirGroupCharacteristic.SetExclude(value : TFhirBoolean);
begin
  FExclude.free;
  FExclude := value;
end;

Function TFhirGroupCharacteristic.GetExcludeST : Boolean;
begin
  if FExclude = nil then
    result := false
  else
    result := FExclude.value;
end;

Procedure TFhirGroupCharacteristic.SetExcludeST(value : Boolean);
begin
  if FExclude = nil then
    FExclude := TFhirBoolean.create;
  FExclude.value := value
end;

Procedure TFhirGroupCharacteristic.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirGroupCharacteristicListEnumerator }

Constructor TFhirGroupCharacteristicListEnumerator.Create(list : TFhirGroupCharacteristicList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirGroupCharacteristicListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirGroupCharacteristicListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirGroupCharacteristicListEnumerator.GetCurrent : TFhirGroupCharacteristic;
begin
  Result := FList[FIndex];
end;


{ TFhirGroupCharacteristicList }
procedure TFhirGroupCharacteristicList.AddItem(value: TFhirGroupCharacteristic);
begin
  assert(value.ClassName = 'TFhirGroupCharacteristic', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirGroupCharacteristic');
  add(value);
end;


function TFhirGroupCharacteristicList.Append: TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupCharacteristicList.ClearItems;
begin
  Clear;
end;

function TFhirGroupCharacteristicList.GetEnumerator : TFhirGroupCharacteristicListEnumerator;
begin
  result := TFhirGroupCharacteristicListEnumerator.Create(self.link);
end;

function TFhirGroupCharacteristicList.Clone: TFhirGroupCharacteristicList;
begin
  result := TFhirGroupCharacteristicList(inherited Clone);
end;

function TFhirGroupCharacteristicList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirGroupCharacteristicList.GetItemN(index: Integer): TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic(ObjectByIndex[index]);
end;

function TFhirGroupCharacteristicList.IndexOf(value: TFhirGroupCharacteristic): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirGroupCharacteristicList.Insert(index: Integer): TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupCharacteristicList.InsertItem(index: Integer; value: TFhirGroupCharacteristic);
begin
  assert(value is TFhirGroupCharacteristic);
  Inherited Insert(index, value);
end;

function TFhirGroupCharacteristicList.Item(index: Integer): TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic(ObjectByIndex[index]);
end;

function TFhirGroupCharacteristicList.Link: TFhirGroupCharacteristicList;
begin
  result := TFhirGroupCharacteristicList(inherited Link);
end;

procedure TFhirGroupCharacteristicList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirGroupCharacteristicList.SetItemByIndex(index: Integer; value: TFhirGroupCharacteristic);
begin
  assert(value is TFhirGroupCharacteristic);
  FhirGroupCharacteristics[index] := value;
end;

procedure TFhirGroupCharacteristicList.SetItemN(index: Integer; value: TFhirGroupCharacteristic);
begin
  assert(value is TFhirGroupCharacteristic);
  ObjectByIndex[index] := value;
end;

{ TFhirGroupMember }

constructor TFhirGroupMember.Create;
begin
  inherited;
end;

destructor TFhirGroupMember.Destroy;
begin
  FEntity.free;
  FPeriod.free;
  FInactive.free;
  inherited;
end;

procedure TFhirGroupMember.Assign(oSource : TAdvObject);
begin
  inherited;
  entity := TFhirGroupMember(oSource).entity.Clone;
  period := TFhirGroupMember(oSource).period.Clone;
  inactiveElement := TFhirGroupMember(oSource).inactiveElement.Clone;
end;

procedure TFhirGroupMember.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'entity') Then
     list.add(FEntity.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'inactive') Then
     list.add(FInactive.Link);
end;

procedure TFhirGroupMember.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'entity', 'Reference(Patient|Practitioner|Device|Medication|Substance)', FEntity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'inactive', 'boolean', FInactive.Link));{2}
end;

procedure TFhirGroupMember.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'entity') then Entity := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'inactive') then InactiveElement := propValue as TFhirBoolean{5a}
  else inherited;
end;

function TFhirGroupMember.FhirType : string;
begin
  result := 'member';
end;

function TFhirGroupMember.Link : TFhirGroupMember;
begin
  result := TFhirGroupMember(inherited Link);
end;

function TFhirGroupMember.Clone : TFhirGroupMember;
begin
  result := TFhirGroupMember(inherited Clone);
end;

function TFhirGroupMember.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirGroupMember;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirGroupMember)) then
    result := false
  else
  begin
    o := TFhirGroupMember(other);
    result := compareDeep(entityElement, o.entityElement, true) and compareDeep(periodElement, o.periodElement, true) and 
      compareDeep(inactiveElement, o.inactiveElement, true);
  end;
end;

function TFhirGroupMember.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirGroupMember;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirGroupMember)) then
    result := false
  else
  begin
    o := TFhirGroupMember(other);
    result := compareValues(inactiveElement, o.inactiveElement, true);
  end;
end;

{ TFhirGroupMember }

Procedure TFhirGroupMember.SetEntity(value : TFhirReference{Resource});
begin
  FEntity.free;
  FEntity := value;
end;

Procedure TFhirGroupMember.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirGroupMember.SetInactive(value : TFhirBoolean);
begin
  FInactive.free;
  FInactive := value;
end;

Function TFhirGroupMember.GetInactiveST : Boolean;
begin
  if FInactive = nil then
    result := false
  else
    result := FInactive.value;
end;

Procedure TFhirGroupMember.SetInactiveST(value : Boolean);
begin
  if FInactive = nil then
    FInactive := TFhirBoolean.create;
  FInactive.value := value
end;


{ TFhirGroupMemberListEnumerator }

Constructor TFhirGroupMemberListEnumerator.Create(list : TFhirGroupMemberList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirGroupMemberListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirGroupMemberListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirGroupMemberListEnumerator.GetCurrent : TFhirGroupMember;
begin
  Result := FList[FIndex];
end;


{ TFhirGroupMemberList }
procedure TFhirGroupMemberList.AddItem(value: TFhirGroupMember);
begin
  assert(value.ClassName = 'TFhirGroupMember', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirGroupMember');
  add(value);
end;


function TFhirGroupMemberList.Append: TFhirGroupMember;
begin
  result := TFhirGroupMember.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupMemberList.ClearItems;
begin
  Clear;
end;

function TFhirGroupMemberList.GetEnumerator : TFhirGroupMemberListEnumerator;
begin
  result := TFhirGroupMemberListEnumerator.Create(self.link);
end;

function TFhirGroupMemberList.Clone: TFhirGroupMemberList;
begin
  result := TFhirGroupMemberList(inherited Clone);
end;

function TFhirGroupMemberList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirGroupMemberList.GetItemN(index: Integer): TFhirGroupMember;
begin
  result := TFhirGroupMember(ObjectByIndex[index]);
end;

function TFhirGroupMemberList.IndexOf(value: TFhirGroupMember): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirGroupMemberList.Insert(index: Integer): TFhirGroupMember;
begin
  result := TFhirGroupMember.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupMemberList.InsertItem(index: Integer; value: TFhirGroupMember);
begin
  assert(value is TFhirGroupMember);
  Inherited Insert(index, value);
end;

function TFhirGroupMemberList.Item(index: Integer): TFhirGroupMember;
begin
  result := TFhirGroupMember(ObjectByIndex[index]);
end;

function TFhirGroupMemberList.Link: TFhirGroupMemberList;
begin
  result := TFhirGroupMemberList(inherited Link);
end;

procedure TFhirGroupMemberList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirGroupMemberList.SetItemByIndex(index: Integer; value: TFhirGroupMember);
begin
  assert(value is TFhirGroupMember);
  FhirGroupMembers[index] := value;
end;

procedure TFhirGroupMemberList.SetItemN(index: Integer; value: TFhirGroupMember);
begin
  assert(value is TFhirGroupMember);
  ObjectByIndex[index] := value;
end;

{ TFhirGroup }

constructor TFhirGroup.Create;
begin
  inherited;
end;

destructor TFhirGroup.Destroy;
begin
  FIdentifierList.Free;
  FType_.free;
  FActual.free;
  FCode.free;
  FName.free;
  FQuantity.free;
  FCharacteristicList.Free;
  FMemberList.Free;
  inherited;
end;

function TFhirGroup.GetResourceType : TFhirResourceType;
begin
  result := frtGroup;
end;

procedure TFhirGroup.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirGroup(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirGroup(oSource).FIdentifierList);
  end;
  FType_ := TFhirGroup(oSource).FType_.Link;
  actualElement := TFhirGroup(oSource).actualElement.Clone;
  code := TFhirGroup(oSource).code.Clone;
  nameElement := TFhirGroup(oSource).nameElement.Clone;
  quantityElement := TFhirGroup(oSource).quantityElement.Clone;
  if (TFhirGroup(oSource).FCharacteristicList = nil) then
  begin
    FCharacteristicList.free;
    FCharacteristicList := nil;
  end
  else
  begin
    FCharacteristicList := TFhirGroupCharacteristicList.Create;
    FCharacteristicList.Assign(TFhirGroup(oSource).FCharacteristicList);
  end;
  if (TFhirGroup(oSource).FMemberList = nil) then
  begin
    FMemberList.free;
    FMemberList := nil;
  end
  else
  begin
    FMemberList := TFhirGroupMemberList.Create;
    FMemberList.Assign(TFhirGroup(oSource).FMemberList);
  end;
end;

procedure TFhirGroup.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'actual') Then
     list.add(FActual.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'characteristic') Then
    list.addAll(FCharacteristicList);
  if (child_name = 'member') Then
    list.addAll(FMemberList);
end;

procedure TFhirGroup.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'actual', 'boolean', FActual.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'unsignedInt', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'characteristic', '', FCharacteristicList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'member', '', FMemberList.Link)){3};
end;

procedure TFhirGroup.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'actual') then ActualElement := propValue as TFhirBoolean{5a}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'quantity') then QuantityElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'characteristic') then CharacteristicList.add(propValue as TFhirGroupCharacteristic){2}
  else if (propName = 'member') then MemberList.add(propValue as TFhirGroupMember){2}
  else inherited;
end;

function TFhirGroup.FhirType : string;
begin
  result := 'Group';
end;

function TFhirGroup.Link : TFhirGroup;
begin
  result := TFhirGroup(inherited Link);
end;

function TFhirGroup.Clone : TFhirGroup;
begin
  result := TFhirGroup(inherited Clone);
end;

{ TFhirGroup }

Function TFhirGroup.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirGroup.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirGroup.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirGroup.GetType_ST : TFhirGroupType;
begin
  if FType_ = nil then
    result := TFhirGroupType(0)
  else
    result := TFhirGroupType(StringArrayIndexOfSensitive(CODES_TFhirGroupType, FType_.value));
end;

Procedure TFhirGroup.SetType_ST(value : TFhirGroupType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirGroupType[value]);
end;

Procedure TFhirGroup.SetActual(value : TFhirBoolean);
begin
  FActual.free;
  FActual := value;
end;

Function TFhirGroup.GetActualST : Boolean;
begin
  if FActual = nil then
    result := false
  else
    result := FActual.value;
end;

Procedure TFhirGroup.SetActualST(value : Boolean);
begin
  if FActual = nil then
    FActual := TFhirBoolean.create;
  FActual.value := value
end;

Procedure TFhirGroup.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirGroup.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirGroup.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirGroup.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirGroup.SetQuantity(value : TFhirUnsignedInt);
begin
  FQuantity.free;
  FQuantity := value;
end;

Function TFhirGroup.GetQuantityST : String;
begin
  if FQuantity = nil then
    result := ''
  else
    result := FQuantity.value;
end;

Procedure TFhirGroup.SetQuantityST(value : String);
begin
  if value <> '' then
  begin
    if FQuantity = nil then
      FQuantity := TFhirUnsignedInt.create;
    FQuantity.value := value
  end
  else if FQuantity <> nil then
    FQuantity.value := '';
end;

Function TFhirGroup.GetCharacteristicList : TFhirGroupCharacteristicList;
begin
  if FCharacteristicList = nil then
    FCharacteristicList := TFhirGroupCharacteristicList.Create;
  result := FCharacteristicList;
end;

Function TFhirGroup.GetHasCharacteristicList : boolean;
begin
  result := (FCharacteristicList <> nil) and (FCharacteristicList.count > 0);
end;

Function TFhirGroup.GetMemberList : TFhirGroupMemberList;
begin
  if FMemberList = nil then
    FMemberList := TFhirGroupMemberList.Create;
  result := FMemberList;
end;

Function TFhirGroup.GetHasMemberList : boolean;
begin
  result := (FMemberList <> nil) and (FMemberList.count > 0);
end;


{ TFhirGroupListEnumerator }

Constructor TFhirGroupListEnumerator.Create(list : TFhirGroupList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirGroupListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirGroupListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirGroupListEnumerator.GetCurrent : TFhirGroup;
begin
  Result := FList[FIndex];
end;


{ TFhirGroupList }
procedure TFhirGroupList.AddItem(value: TFhirGroup);
begin
  assert(value.ClassName = 'TFhirGroup', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirGroup');
  add(value);
end;


function TFhirGroupList.Append: TFhirGroup;
begin
  result := TFhirGroup.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupList.ClearItems;
begin
  Clear;
end;

function TFhirGroupList.GetEnumerator : TFhirGroupListEnumerator;
begin
  result := TFhirGroupListEnumerator.Create(self.link);
end;

function TFhirGroupList.Clone: TFhirGroupList;
begin
  result := TFhirGroupList(inherited Clone);
end;

function TFhirGroupList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirGroupList.GetItemN(index: Integer): TFhirGroup;
begin
  result := TFhirGroup(ObjectByIndex[index]);
end;

function TFhirGroupList.IndexOf(value: TFhirGroup): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirGroupList.Insert(index: Integer): TFhirGroup;
begin
  result := TFhirGroup.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirGroupList.InsertItem(index: Integer; value: TFhirGroup);
begin
  assert(value is TFhirGroup);
  Inherited Insert(index, value);
end;

function TFhirGroupList.Item(index: Integer): TFhirGroup;
begin
  result := TFhirGroup(ObjectByIndex[index]);
end;

function TFhirGroupList.Link: TFhirGroupList;
begin
  result := TFhirGroupList(inherited Link);
end;

procedure TFhirGroupList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirGroupList.SetItemByIndex(index: Integer; value: TFhirGroup);
begin
  assert(value is TFhirGroup);
  FhirGroups[index] := value;
end;

procedure TFhirGroupList.SetItemN(index: Integer; value: TFhirGroup);
begin
  assert(value is TFhirGroup);
  ObjectByIndex[index] := value;
end;

{ TFhirHealthcareServiceServiceType }

constructor TFhirHealthcareServiceServiceType.Create;
begin
  inherited;
end;

destructor TFhirHealthcareServiceServiceType.Destroy;
begin
  FType_.free;
  FSpecialtyList.Free;
  inherited;
end;

procedure TFhirHealthcareServiceServiceType.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirHealthcareServiceServiceType(oSource).type_.Clone;
  if (TFhirHealthcareServiceServiceType(oSource).FSpecialtyList = nil) then
  begin
    FSpecialtyList.free;
    FSpecialtyList := nil;
  end
  else
  begin
    FSpecialtyList := TFhirCodeableConceptList.Create;
    FSpecialtyList.Assign(TFhirHealthcareServiceServiceType(oSource).FSpecialtyList);
  end;
end;

procedure TFhirHealthcareServiceServiceType.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'specialty') Then
    list.addAll(FSpecialtyList);
end;

procedure TFhirHealthcareServiceServiceType.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specialty', 'CodeableConcept', FSpecialtyList.Link)){3};
end;

procedure TFhirHealthcareServiceServiceType.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'specialty') then SpecialtyList.add(propValue as TFhirCodeableConcept){2}
  else inherited;
end;

function TFhirHealthcareServiceServiceType.FhirType : string;
begin
  result := 'serviceType';
end;

function TFhirHealthcareServiceServiceType.Link : TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType(inherited Link);
end;

function TFhirHealthcareServiceServiceType.Clone : TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType(inherited Clone);
end;

function TFhirHealthcareServiceServiceType.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceServiceType;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceServiceType)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceServiceType(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(specialtyList, o.specialtyList, true);
  end;
end;

function TFhirHealthcareServiceServiceType.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceServiceType;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceServiceType)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceServiceType(other);
    result := true;
  end;
end;

{ TFhirHealthcareServiceServiceType }

Procedure TFhirHealthcareServiceServiceType.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirHealthcareServiceServiceType.GetSpecialtyList : TFhirCodeableConceptList;
begin
  if FSpecialtyList = nil then
    FSpecialtyList := TFhirCodeableConceptList.Create;
  result := FSpecialtyList;
end;

Function TFhirHealthcareServiceServiceType.GetHasSpecialtyList : boolean;
begin
  result := (FSpecialtyList <> nil) and (FSpecialtyList.count > 0);
end;


{ TFhirHealthcareServiceServiceTypeListEnumerator }

Constructor TFhirHealthcareServiceServiceTypeListEnumerator.Create(list : TFhirHealthcareServiceServiceTypeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirHealthcareServiceServiceTypeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirHealthcareServiceServiceTypeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirHealthcareServiceServiceTypeListEnumerator.GetCurrent : TFhirHealthcareServiceServiceType;
begin
  Result := FList[FIndex];
end;


{ TFhirHealthcareServiceServiceTypeList }
procedure TFhirHealthcareServiceServiceTypeList.AddItem(value: TFhirHealthcareServiceServiceType);
begin
  assert(value.ClassName = 'TFhirHealthcareServiceServiceType', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirHealthcareServiceServiceType');
  add(value);
end;


function TFhirHealthcareServiceServiceTypeList.Append: TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceServiceTypeList.ClearItems;
begin
  Clear;
end;

function TFhirHealthcareServiceServiceTypeList.GetEnumerator : TFhirHealthcareServiceServiceTypeListEnumerator;
begin
  result := TFhirHealthcareServiceServiceTypeListEnumerator.Create(self.link);
end;

function TFhirHealthcareServiceServiceTypeList.Clone: TFhirHealthcareServiceServiceTypeList;
begin
  result := TFhirHealthcareServiceServiceTypeList(inherited Clone);
end;

function TFhirHealthcareServiceServiceTypeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirHealthcareServiceServiceTypeList.GetItemN(index: Integer): TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceServiceTypeList.IndexOf(value: TFhirHealthcareServiceServiceType): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirHealthcareServiceServiceTypeList.Insert(index: Integer): TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceServiceTypeList.InsertItem(index: Integer; value: TFhirHealthcareServiceServiceType);
begin
  assert(value is TFhirHealthcareServiceServiceType);
  Inherited Insert(index, value);
end;

function TFhirHealthcareServiceServiceTypeList.Item(index: Integer): TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceServiceTypeList.Link: TFhirHealthcareServiceServiceTypeList;
begin
  result := TFhirHealthcareServiceServiceTypeList(inherited Link);
end;

procedure TFhirHealthcareServiceServiceTypeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirHealthcareServiceServiceTypeList.SetItemByIndex(index: Integer; value: TFhirHealthcareServiceServiceType);
begin
  assert(value is TFhirHealthcareServiceServiceType);
  FhirHealthcareServiceServiceTypes[index] := value;
end;

procedure TFhirHealthcareServiceServiceTypeList.SetItemN(index: Integer; value: TFhirHealthcareServiceServiceType);
begin
  assert(value is TFhirHealthcareServiceServiceType);
  ObjectByIndex[index] := value;
end;

{ TFhirHealthcareServiceAvailableTime }

constructor TFhirHealthcareServiceAvailableTime.Create;
begin
  inherited;
end;

destructor TFhirHealthcareServiceAvailableTime.Destroy;
begin
  FDaysOfWeek.Free;
  FAllDay.free;
  FAvailableStartTime.free;
  FAvailableEndTime.free;
  inherited;
end;

procedure TFhirHealthcareServiceAvailableTime.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirHealthcareServiceAvailableTime(oSource).FDaysOfWeek = nil) then
  begin
    FDaysOfWeek.free;
    FDaysOfWeek := nil;
  end
  else
  begin
    FDaysOfWeek := TFHIREnumList.Create;
    FDaysOfWeek.Assign(TFhirHealthcareServiceAvailableTime(oSource).FDaysOfWeek);
  end;
  allDayElement := TFhirHealthcareServiceAvailableTime(oSource).allDayElement.Clone;
  availableStartTimeElement := TFhirHealthcareServiceAvailableTime(oSource).availableStartTimeElement.Clone;
  availableEndTimeElement := TFhirHealthcareServiceAvailableTime(oSource).availableEndTimeElement.Clone;
end;

procedure TFhirHealthcareServiceAvailableTime.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'daysOfWeek') Then
     list.addAll(FDaysOfWeek);
  if (child_name = 'allDay') Then
     list.add(FAllDay.Link);
  if (child_name = 'availableStartTime') Then
     list.add(FAvailableStartTime.Link);
  if (child_name = 'availableEndTime') Then
     list.add(FAvailableEndTime.Link);
end;

procedure TFhirHealthcareServiceAvailableTime.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'daysOfWeek', 'code', FDaysOfWeek.Link)){3};
  oList.add(TFHIRProperty.create(self, 'allDay', 'boolean', FAllDay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'availableStartTime', 'time', FAvailableStartTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'availableEndTime', 'time', FAvailableEndTime.Link));{2}
end;

procedure TFhirHealthcareServiceAvailableTime.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'daysOfWeek') then FDaysOfWeek.add(propValue as TFHIREnum) {1}
  else if (propName = 'allDay') then AllDayElement := propValue as TFhirBoolean{5a}
  else if (propName = 'availableStartTime') then AvailableStartTimeElement := propValue as TFhirTime{5a}
  else if (propName = 'availableEndTime') then AvailableEndTimeElement := propValue as TFhirTime{5a}
  else inherited;
end;

function TFhirHealthcareServiceAvailableTime.FhirType : string;
begin
  result := 'availableTime';
end;

function TFhirHealthcareServiceAvailableTime.Link : TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime(inherited Link);
end;

function TFhirHealthcareServiceAvailableTime.Clone : TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime(inherited Clone);
end;

function TFhirHealthcareServiceAvailableTime.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceAvailableTime;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceAvailableTime)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceAvailableTime(other);
    result := compareDeep(daysOfWeekList, o.daysOfWeekList, true) and compareDeep(allDayElement, o.allDayElement, true) and 
      compareDeep(availableStartTimeElement, o.availableStartTimeElement, true) and 
      compareDeep(availableEndTimeElement, o.availableEndTimeElement, true);
  end;
end;

function TFhirHealthcareServiceAvailableTime.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceAvailableTime;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceAvailableTime)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceAvailableTime(other);
    result := compareValues(daysOfWeekList, o.daysOfWeekList, true) and compareValues(allDayElement, o.allDayElement, true) and 
      compareValues(availableStartTimeElement, o.availableStartTimeElement, true) and 
      compareValues(availableEndTimeElement, o.availableEndTimeElement, true);
  end;
end;

{ TFhirHealthcareServiceAvailableTime }

Function TFhirHealthcareServiceAvailableTime.GetDaysOfWeek : TFhirEnumList;
begin
  if FDaysOfWeek = nil then
    FDaysOfWeek := TFHIREnumList.Create;
  result := FDaysOfWeek;
end;

Function TFhirHealthcareServiceAvailableTime.GetHasDaysOfWeek : boolean;
begin
  result := (FDaysOfWeek <> nil) and (FDaysOfWeek.count > 0);
end;

Function TFhirHealthcareServiceAvailableTime.GetDaysOfWeekST : TFhirDaysOfWeekList;
  var i : integer;
begin
  result := [];
  if FdaysOfWeek <> nil then
    for i := 0 to FdaysOfWeek.count - 1 do
      result := result + [TFhirDaysOfWeek(StringArrayIndexOfSensitive(CODES_TFhirDaysOfWeek, FdaysOfWeek[i].value))];
end;

Procedure TFhirHealthcareServiceAvailableTime.SetDaysOfWeekST(value : TFhirDaysOfWeekList);
var a : TFhirDaysOfWeek;
begin
  FdaysOfWeek.clear;
  for a := low(TFhirDaysOfWeek) to high(TFhirDaysOfWeek) do
    if a in value then
      begin
         if FdaysOfWeek = nil then
           FdaysOfWeek := TFhirEnumList.create;
         FdaysOfWeek.add(TFhirEnum.create(CODES_TFhirDaysOfWeek[a]));
      end;
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAllDay(value : TFhirBoolean);
begin
  FAllDay.free;
  FAllDay := value;
end;

Function TFhirHealthcareServiceAvailableTime.GetAllDayST : Boolean;
begin
  if FAllDay = nil then
    result := false
  else
    result := FAllDay.value;
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAllDayST(value : Boolean);
begin
  if FAllDay = nil then
    FAllDay := TFhirBoolean.create;
  FAllDay.value := value
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAvailableStartTime(value : TFhirTime);
begin
  FAvailableStartTime.free;
  FAvailableStartTime := value;
end;

Function TFhirHealthcareServiceAvailableTime.GetAvailableStartTimeST : String;
begin
  if FAvailableStartTime = nil then
    result := ''
  else
    result := FAvailableStartTime.value;
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAvailableStartTimeST(value : String);
begin
  if value <> '' then
  begin
    if FAvailableStartTime = nil then
      FAvailableStartTime := TFhirTime.create;
    FAvailableStartTime.value := value
  end
  else if FAvailableStartTime <> nil then
    FAvailableStartTime.value := '';
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAvailableEndTime(value : TFhirTime);
begin
  FAvailableEndTime.free;
  FAvailableEndTime := value;
end;

Function TFhirHealthcareServiceAvailableTime.GetAvailableEndTimeST : String;
begin
  if FAvailableEndTime = nil then
    result := ''
  else
    result := FAvailableEndTime.value;
end;

Procedure TFhirHealthcareServiceAvailableTime.SetAvailableEndTimeST(value : String);
begin
  if value <> '' then
  begin
    if FAvailableEndTime = nil then
      FAvailableEndTime := TFhirTime.create;
    FAvailableEndTime.value := value
  end
  else if FAvailableEndTime <> nil then
    FAvailableEndTime.value := '';
end;


{ TFhirHealthcareServiceAvailableTimeListEnumerator }

Constructor TFhirHealthcareServiceAvailableTimeListEnumerator.Create(list : TFhirHealthcareServiceAvailableTimeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirHealthcareServiceAvailableTimeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirHealthcareServiceAvailableTimeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirHealthcareServiceAvailableTimeListEnumerator.GetCurrent : TFhirHealthcareServiceAvailableTime;
begin
  Result := FList[FIndex];
end;


{ TFhirHealthcareServiceAvailableTimeList }
procedure TFhirHealthcareServiceAvailableTimeList.AddItem(value: TFhirHealthcareServiceAvailableTime);
begin
  assert(value.ClassName = 'TFhirHealthcareServiceAvailableTime', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirHealthcareServiceAvailableTime');
  add(value);
end;


function TFhirHealthcareServiceAvailableTimeList.Append: TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceAvailableTimeList.ClearItems;
begin
  Clear;
end;

function TFhirHealthcareServiceAvailableTimeList.GetEnumerator : TFhirHealthcareServiceAvailableTimeListEnumerator;
begin
  result := TFhirHealthcareServiceAvailableTimeListEnumerator.Create(self.link);
end;

function TFhirHealthcareServiceAvailableTimeList.Clone: TFhirHealthcareServiceAvailableTimeList;
begin
  result := TFhirHealthcareServiceAvailableTimeList(inherited Clone);
end;

function TFhirHealthcareServiceAvailableTimeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirHealthcareServiceAvailableTimeList.GetItemN(index: Integer): TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceAvailableTimeList.IndexOf(value: TFhirHealthcareServiceAvailableTime): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirHealthcareServiceAvailableTimeList.Insert(index: Integer): TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceAvailableTimeList.InsertItem(index: Integer; value: TFhirHealthcareServiceAvailableTime);
begin
  assert(value is TFhirHealthcareServiceAvailableTime);
  Inherited Insert(index, value);
end;

function TFhirHealthcareServiceAvailableTimeList.Item(index: Integer): TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceAvailableTimeList.Link: TFhirHealthcareServiceAvailableTimeList;
begin
  result := TFhirHealthcareServiceAvailableTimeList(inherited Link);
end;

procedure TFhirHealthcareServiceAvailableTimeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirHealthcareServiceAvailableTimeList.SetItemByIndex(index: Integer; value: TFhirHealthcareServiceAvailableTime);
begin
  assert(value is TFhirHealthcareServiceAvailableTime);
  FhirHealthcareServiceAvailableTimes[index] := value;
end;

procedure TFhirHealthcareServiceAvailableTimeList.SetItemN(index: Integer; value: TFhirHealthcareServiceAvailableTime);
begin
  assert(value is TFhirHealthcareServiceAvailableTime);
  ObjectByIndex[index] := value;
end;

{ TFhirHealthcareServiceNotAvailable }

constructor TFhirHealthcareServiceNotAvailable.Create;
begin
  inherited;
end;

destructor TFhirHealthcareServiceNotAvailable.Destroy;
begin
  FDescription.free;
  FDuring.free;
  inherited;
end;

procedure TFhirHealthcareServiceNotAvailable.Assign(oSource : TAdvObject);
begin
  inherited;
  descriptionElement := TFhirHealthcareServiceNotAvailable(oSource).descriptionElement.Clone;
  during := TFhirHealthcareServiceNotAvailable(oSource).during.Clone;
end;

procedure TFhirHealthcareServiceNotAvailable.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'during') Then
     list.add(FDuring.Link);
end;

procedure TFhirHealthcareServiceNotAvailable.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'during', 'Period', FDuring.Link));{2}
end;

procedure TFhirHealthcareServiceNotAvailable.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'during') then During := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirHealthcareServiceNotAvailable.FhirType : string;
begin
  result := 'notAvailable';
end;

function TFhirHealthcareServiceNotAvailable.Link : TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable(inherited Link);
end;

function TFhirHealthcareServiceNotAvailable.Clone : TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable(inherited Clone);
end;

function TFhirHealthcareServiceNotAvailable.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceNotAvailable;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceNotAvailable)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceNotAvailable(other);
    result := compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(duringElement, o.duringElement, true);
  end;
end;

function TFhirHealthcareServiceNotAvailable.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirHealthcareServiceNotAvailable;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirHealthcareServiceNotAvailable)) then
    result := false
  else
  begin
    o := TFhirHealthcareServiceNotAvailable(other);
    result := compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirHealthcareServiceNotAvailable }

Procedure TFhirHealthcareServiceNotAvailable.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirHealthcareServiceNotAvailable.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirHealthcareServiceNotAvailable.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirHealthcareServiceNotAvailable.SetDuring(value : TFhirPeriod);
begin
  FDuring.free;
  FDuring := value;
end;


{ TFhirHealthcareServiceNotAvailableListEnumerator }

Constructor TFhirHealthcareServiceNotAvailableListEnumerator.Create(list : TFhirHealthcareServiceNotAvailableList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirHealthcareServiceNotAvailableListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirHealthcareServiceNotAvailableListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirHealthcareServiceNotAvailableListEnumerator.GetCurrent : TFhirHealthcareServiceNotAvailable;
begin
  Result := FList[FIndex];
end;


{ TFhirHealthcareServiceNotAvailableList }
procedure TFhirHealthcareServiceNotAvailableList.AddItem(value: TFhirHealthcareServiceNotAvailable);
begin
  assert(value.ClassName = 'TFhirHealthcareServiceNotAvailable', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirHealthcareServiceNotAvailable');
  add(value);
end;


function TFhirHealthcareServiceNotAvailableList.Append: TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceNotAvailableList.ClearItems;
begin
  Clear;
end;

function TFhirHealthcareServiceNotAvailableList.GetEnumerator : TFhirHealthcareServiceNotAvailableListEnumerator;
begin
  result := TFhirHealthcareServiceNotAvailableListEnumerator.Create(self.link);
end;

function TFhirHealthcareServiceNotAvailableList.Clone: TFhirHealthcareServiceNotAvailableList;
begin
  result := TFhirHealthcareServiceNotAvailableList(inherited Clone);
end;

function TFhirHealthcareServiceNotAvailableList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirHealthcareServiceNotAvailableList.GetItemN(index: Integer): TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceNotAvailableList.IndexOf(value: TFhirHealthcareServiceNotAvailable): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirHealthcareServiceNotAvailableList.Insert(index: Integer): TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceNotAvailableList.InsertItem(index: Integer; value: TFhirHealthcareServiceNotAvailable);
begin
  assert(value is TFhirHealthcareServiceNotAvailable);
  Inherited Insert(index, value);
end;

function TFhirHealthcareServiceNotAvailableList.Item(index: Integer): TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceNotAvailableList.Link: TFhirHealthcareServiceNotAvailableList;
begin
  result := TFhirHealthcareServiceNotAvailableList(inherited Link);
end;

procedure TFhirHealthcareServiceNotAvailableList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirHealthcareServiceNotAvailableList.SetItemByIndex(index: Integer; value: TFhirHealthcareServiceNotAvailable);
begin
  assert(value is TFhirHealthcareServiceNotAvailable);
  FhirHealthcareServiceNotAvailables[index] := value;
end;

procedure TFhirHealthcareServiceNotAvailableList.SetItemN(index: Integer; value: TFhirHealthcareServiceNotAvailable);
begin
  assert(value is TFhirHealthcareServiceNotAvailable);
  ObjectByIndex[index] := value;
end;

{ TFhirHealthcareService }

constructor TFhirHealthcareService.Create;
begin
  inherited;
end;

destructor TFhirHealthcareService.Destroy;
begin
  FIdentifierList.Free;
  FProvidedBy.free;
  FServiceCategory.free;
  FServiceTypeList.Free;
  FLocation.free;
  FServiceName.free;
  FComment.free;
  FExtraDetails.free;
  FPhoto.free;
  FTelecomList.Free;
  FCoverageAreaList.Free;
  FServiceProvisionCodeList.Free;
  FEligibility.free;
  FEligibilityNote.free;
  FProgramNameList.Free;
  FCharacteristicList.Free;
  FReferralMethodList.Free;
  FPublicKey.free;
  FAppointmentRequired.free;
  FAvailableTimeList.Free;
  FNotAvailableList.Free;
  FAvailabilityExceptions.free;
  inherited;
end;

function TFhirHealthcareService.GetResourceType : TFhirResourceType;
begin
  result := frtHealthcareService;
end;

procedure TFhirHealthcareService.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirHealthcareService(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirHealthcareService(oSource).FIdentifierList);
  end;
  providedBy := TFhirHealthcareService(oSource).providedBy.Clone;
  serviceCategory := TFhirHealthcareService(oSource).serviceCategory.Clone;
  if (TFhirHealthcareService(oSource).FServiceTypeList = nil) then
  begin
    FServiceTypeList.free;
    FServiceTypeList := nil;
  end
  else
  begin
    FServiceTypeList := TFhirHealthcareServiceServiceTypeList.Create;
    FServiceTypeList.Assign(TFhirHealthcareService(oSource).FServiceTypeList);
  end;
  location := TFhirHealthcareService(oSource).location.Clone;
  serviceNameElement := TFhirHealthcareService(oSource).serviceNameElement.Clone;
  commentElement := TFhirHealthcareService(oSource).commentElement.Clone;
  extraDetailsElement := TFhirHealthcareService(oSource).extraDetailsElement.Clone;
  photo := TFhirHealthcareService(oSource).photo.Clone;
  if (TFhirHealthcareService(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirHealthcareService(oSource).FTelecomList);
  end;
  if (TFhirHealthcareService(oSource).FCoverageAreaList = nil) then
  begin
    FCoverageAreaList.free;
    FCoverageAreaList := nil;
  end
  else
  begin
    FCoverageAreaList := TFhirReferenceList{TFhirLocation}.Create;
    FCoverageAreaList.Assign(TFhirHealthcareService(oSource).FCoverageAreaList);
  end;
  if (TFhirHealthcareService(oSource).FServiceProvisionCodeList = nil) then
  begin
    FServiceProvisionCodeList.free;
    FServiceProvisionCodeList := nil;
  end
  else
  begin
    FServiceProvisionCodeList := TFhirCodeableConceptList.Create;
    FServiceProvisionCodeList.Assign(TFhirHealthcareService(oSource).FServiceProvisionCodeList);
  end;
  eligibility := TFhirHealthcareService(oSource).eligibility.Clone;
  eligibilityNoteElement := TFhirHealthcareService(oSource).eligibilityNoteElement.Clone;
  if (TFhirHealthcareService(oSource).FProgramNameList = nil) then
  begin
    FProgramNameList.free;
    FProgramNameList := nil;
  end
  else
  begin
    FProgramNameList := TFhirStringList.Create;
    FProgramNameList.Assign(TFhirHealthcareService(oSource).FProgramNameList);
  end;
  if (TFhirHealthcareService(oSource).FCharacteristicList = nil) then
  begin
    FCharacteristicList.free;
    FCharacteristicList := nil;
  end
  else
  begin
    FCharacteristicList := TFhirCodeableConceptList.Create;
    FCharacteristicList.Assign(TFhirHealthcareService(oSource).FCharacteristicList);
  end;
  if (TFhirHealthcareService(oSource).FReferralMethodList = nil) then
  begin
    FReferralMethodList.free;
    FReferralMethodList := nil;
  end
  else
  begin
    FReferralMethodList := TFhirCodeableConceptList.Create;
    FReferralMethodList.Assign(TFhirHealthcareService(oSource).FReferralMethodList);
  end;
  publicKeyElement := TFhirHealthcareService(oSource).publicKeyElement.Clone;
  appointmentRequiredElement := TFhirHealthcareService(oSource).appointmentRequiredElement.Clone;
  if (TFhirHealthcareService(oSource).FAvailableTimeList = nil) then
  begin
    FAvailableTimeList.free;
    FAvailableTimeList := nil;
  end
  else
  begin
    FAvailableTimeList := TFhirHealthcareServiceAvailableTimeList.Create;
    FAvailableTimeList.Assign(TFhirHealthcareService(oSource).FAvailableTimeList);
  end;
  if (TFhirHealthcareService(oSource).FNotAvailableList = nil) then
  begin
    FNotAvailableList.free;
    FNotAvailableList := nil;
  end
  else
  begin
    FNotAvailableList := TFhirHealthcareServiceNotAvailableList.Create;
    FNotAvailableList.Assign(TFhirHealthcareService(oSource).FNotAvailableList);
  end;
  availabilityExceptionsElement := TFhirHealthcareService(oSource).availabilityExceptionsElement.Clone;
end;

procedure TFhirHealthcareService.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'providedBy') Then
     list.add(FProvidedBy.Link);
  if (child_name = 'serviceCategory') Then
     list.add(FServiceCategory.Link);
  if (child_name = 'serviceType') Then
    list.addAll(FServiceTypeList);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'serviceName') Then
     list.add(FServiceName.Link);
  if (child_name = 'comment') Then
     list.add(FComment.Link);
  if (child_name = 'extraDetails') Then
     list.add(FExtraDetails.Link);
  if (child_name = 'photo') Then
     list.add(FPhoto.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'coverageArea') Then
    list.addAll(FCoverageAreaList);
  if (child_name = 'serviceProvisionCode') Then
    list.addAll(FServiceProvisionCodeList);
  if (child_name = 'eligibility') Then
     list.add(FEligibility.Link);
  if (child_name = 'eligibilityNote') Then
     list.add(FEligibilityNote.Link);
  if (child_name = 'programName') Then
    list.addAll(FProgramNameList);
  if (child_name = 'characteristic') Then
    list.addAll(FCharacteristicList);
  if (child_name = 'referralMethod') Then
    list.addAll(FReferralMethodList);
  if (child_name = 'publicKey') Then
     list.add(FPublicKey.Link);
  if (child_name = 'appointmentRequired') Then
     list.add(FAppointmentRequired.Link);
  if (child_name = 'availableTime') Then
    list.addAll(FAvailableTimeList);
  if (child_name = 'notAvailable') Then
    list.addAll(FNotAvailableList);
  if (child_name = 'availabilityExceptions') Then
     list.add(FAvailabilityExceptions.Link);
end;

procedure TFhirHealthcareService.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'providedBy', 'Reference(Organization)', FProvidedBy.Link));{2}
  oList.add(TFHIRProperty.create(self, 'serviceCategory', 'CodeableConcept', FServiceCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'serviceType', '', FServiceTypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'serviceName', 'string', FServiceName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
  oList.add(TFHIRProperty.create(self, 'extraDetails', 'string', FExtraDetails.Link));{2}
  oList.add(TFHIRProperty.create(self, 'photo', 'Attachment', FPhoto.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'coverageArea', 'Reference(Location)', FCoverageAreaList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'serviceProvisionCode', 'CodeableConcept', FServiceProvisionCodeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'eligibility', 'CodeableConcept', FEligibility.Link));{2}
  oList.add(TFHIRProperty.create(self, 'eligibilityNote', 'string', FEligibilityNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'programName', 'string', FProgramNameList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'characteristic', 'CodeableConcept', FCharacteristicList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'referralMethod', 'CodeableConcept', FReferralMethodList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'publicKey', 'string', FPublicKey.Link));{2}
  oList.add(TFHIRProperty.create(self, 'appointmentRequired', 'boolean', FAppointmentRequired.Link));{2}
  oList.add(TFHIRProperty.create(self, 'availableTime', '', FAvailableTimeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'notAvailable', '', FNotAvailableList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'availabilityExceptions', 'string', FAvailabilityExceptions.Link));{2}
end;

procedure TFhirHealthcareService.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'providedBy') then ProvidedBy := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'serviceCategory') then ServiceCategory := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'serviceType') then ServiceTypeList.add(propValue as TFhirHealthcareServiceServiceType){2}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'serviceName') then ServiceNameElement := propValue as TFhirString{5a}
  else if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else if (propName = 'extraDetails') then ExtraDetailsElement := propValue as TFhirString{5a}
  else if (propName = 'photo') then Photo := propValue as TFhirAttachment{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'coverageArea') then CoverageAreaList.add(propValue as TFhirReference{TFhirLocation}){2}
  else if (propName = 'serviceProvisionCode') then ServiceProvisionCodeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'eligibility') then Eligibility := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'eligibilityNote') then EligibilityNoteElement := propValue as TFhirString{5a}
  else if (propName = 'programName') then ProgramNameList.add(propValue as TFhirString){2}
  else if (propName = 'characteristic') then CharacteristicList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'referralMethod') then ReferralMethodList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'publicKey') then PublicKeyElement := propValue as TFhirString{5a}
  else if (propName = 'appointmentRequired') then AppointmentRequiredElement := propValue as TFhirBoolean{5a}
  else if (propName = 'availableTime') then AvailableTimeList.add(propValue as TFhirHealthcareServiceAvailableTime){2}
  else if (propName = 'notAvailable') then NotAvailableList.add(propValue as TFhirHealthcareServiceNotAvailable){2}
  else if (propName = 'availabilityExceptions') then AvailabilityExceptionsElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirHealthcareService.FhirType : string;
begin
  result := 'HealthcareService';
end;

function TFhirHealthcareService.Link : TFhirHealthcareService;
begin
  result := TFhirHealthcareService(inherited Link);
end;

function TFhirHealthcareService.Clone : TFhirHealthcareService;
begin
  result := TFhirHealthcareService(inherited Clone);
end;

{ TFhirHealthcareService }

Function TFhirHealthcareService.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirHealthcareService.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirHealthcareService.SetProvidedBy(value : TFhirReference{TFhirOrganization});
begin
  FProvidedBy.free;
  FProvidedBy := value;
end;

Procedure TFhirHealthcareService.SetServiceCategory(value : TFhirCodeableConcept);
begin
  FServiceCategory.free;
  FServiceCategory := value;
end;

Function TFhirHealthcareService.GetServiceTypeList : TFhirHealthcareServiceServiceTypeList;
begin
  if FServiceTypeList = nil then
    FServiceTypeList := TFhirHealthcareServiceServiceTypeList.Create;
  result := FServiceTypeList;
end;

Function TFhirHealthcareService.GetHasServiceTypeList : boolean;
begin
  result := (FServiceTypeList <> nil) and (FServiceTypeList.count > 0);
end;

Procedure TFhirHealthcareService.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Procedure TFhirHealthcareService.SetServiceName(value : TFhirString);
begin
  FServiceName.free;
  FServiceName := value;
end;

Function TFhirHealthcareService.GetServiceNameST : String;
begin
  if FServiceName = nil then
    result := ''
  else
    result := FServiceName.value;
end;

Procedure TFhirHealthcareService.SetServiceNameST(value : String);
begin
  if value <> '' then
  begin
    if FServiceName = nil then
      FServiceName := TFhirString.create;
    FServiceName.value := value
  end
  else if FServiceName <> nil then
    FServiceName.value := '';
end;

Procedure TFhirHealthcareService.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirHealthcareService.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirHealthcareService.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;

Procedure TFhirHealthcareService.SetExtraDetails(value : TFhirString);
begin
  FExtraDetails.free;
  FExtraDetails := value;
end;

Function TFhirHealthcareService.GetExtraDetailsST : String;
begin
  if FExtraDetails = nil then
    result := ''
  else
    result := FExtraDetails.value;
end;

Procedure TFhirHealthcareService.SetExtraDetailsST(value : String);
begin
  if value <> '' then
  begin
    if FExtraDetails = nil then
      FExtraDetails := TFhirString.create;
    FExtraDetails.value := value
  end
  else if FExtraDetails <> nil then
    FExtraDetails.value := '';
end;

Procedure TFhirHealthcareService.SetPhoto(value : TFhirAttachment);
begin
  FPhoto.free;
  FPhoto := value;
end;

Function TFhirHealthcareService.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirHealthcareService.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Function TFhirHealthcareService.GetCoverageAreaList : TFhirReferenceList{TFhirLocation};
begin
  if FCoverageAreaList = nil then
    FCoverageAreaList := TFhirReferenceList{TFhirLocation}.Create;
  result := FCoverageAreaList;
end;

Function TFhirHealthcareService.GetHasCoverageAreaList : boolean;
begin
  result := (FCoverageAreaList <> nil) and (FCoverageAreaList.count > 0);
end;

Function TFhirHealthcareService.GetServiceProvisionCodeList : TFhirCodeableConceptList;
begin
  if FServiceProvisionCodeList = nil then
    FServiceProvisionCodeList := TFhirCodeableConceptList.Create;
  result := FServiceProvisionCodeList;
end;

Function TFhirHealthcareService.GetHasServiceProvisionCodeList : boolean;
begin
  result := (FServiceProvisionCodeList <> nil) and (FServiceProvisionCodeList.count > 0);
end;

Procedure TFhirHealthcareService.SetEligibility(value : TFhirCodeableConcept);
begin
  FEligibility.free;
  FEligibility := value;
end;

Procedure TFhirHealthcareService.SetEligibilityNote(value : TFhirString);
begin
  FEligibilityNote.free;
  FEligibilityNote := value;
end;

Function TFhirHealthcareService.GetEligibilityNoteST : String;
begin
  if FEligibilityNote = nil then
    result := ''
  else
    result := FEligibilityNote.value;
end;

Procedure TFhirHealthcareService.SetEligibilityNoteST(value : String);
begin
  if value <> '' then
  begin
    if FEligibilityNote = nil then
      FEligibilityNote := TFhirString.create;
    FEligibilityNote.value := value
  end
  else if FEligibilityNote <> nil then
    FEligibilityNote.value := '';
end;

Function TFhirHealthcareService.GetProgramNameList : TFhirStringList;
begin
  if FProgramNameList = nil then
    FProgramNameList := TFhirStringList.Create;
  result := FProgramNameList;
end;

Function TFhirHealthcareService.GetHasProgramNameList : boolean;
begin
  result := (FProgramNameList <> nil) and (FProgramNameList.count > 0);
end;

Function TFhirHealthcareService.GetCharacteristicList : TFhirCodeableConceptList;
begin
  if FCharacteristicList = nil then
    FCharacteristicList := TFhirCodeableConceptList.Create;
  result := FCharacteristicList;
end;

Function TFhirHealthcareService.GetHasCharacteristicList : boolean;
begin
  result := (FCharacteristicList <> nil) and (FCharacteristicList.count > 0);
end;

Function TFhirHealthcareService.GetReferralMethodList : TFhirCodeableConceptList;
begin
  if FReferralMethodList = nil then
    FReferralMethodList := TFhirCodeableConceptList.Create;
  result := FReferralMethodList;
end;

Function TFhirHealthcareService.GetHasReferralMethodList : boolean;
begin
  result := (FReferralMethodList <> nil) and (FReferralMethodList.count > 0);
end;

Procedure TFhirHealthcareService.SetPublicKey(value : TFhirString);
begin
  FPublicKey.free;
  FPublicKey := value;
end;

Function TFhirHealthcareService.GetPublicKeyST : String;
begin
  if FPublicKey = nil then
    result := ''
  else
    result := FPublicKey.value;
end;

Procedure TFhirHealthcareService.SetPublicKeyST(value : String);
begin
  if value <> '' then
  begin
    if FPublicKey = nil then
      FPublicKey := TFhirString.create;
    FPublicKey.value := value
  end
  else if FPublicKey <> nil then
    FPublicKey.value := '';
end;

Procedure TFhirHealthcareService.SetAppointmentRequired(value : TFhirBoolean);
begin
  FAppointmentRequired.free;
  FAppointmentRequired := value;
end;

Function TFhirHealthcareService.GetAppointmentRequiredST : Boolean;
begin
  if FAppointmentRequired = nil then
    result := false
  else
    result := FAppointmentRequired.value;
end;

Procedure TFhirHealthcareService.SetAppointmentRequiredST(value : Boolean);
begin
  if FAppointmentRequired = nil then
    FAppointmentRequired := TFhirBoolean.create;
  FAppointmentRequired.value := value
end;

Function TFhirHealthcareService.GetAvailableTimeList : TFhirHealthcareServiceAvailableTimeList;
begin
  if FAvailableTimeList = nil then
    FAvailableTimeList := TFhirHealthcareServiceAvailableTimeList.Create;
  result := FAvailableTimeList;
end;

Function TFhirHealthcareService.GetHasAvailableTimeList : boolean;
begin
  result := (FAvailableTimeList <> nil) and (FAvailableTimeList.count > 0);
end;

Function TFhirHealthcareService.GetNotAvailableList : TFhirHealthcareServiceNotAvailableList;
begin
  if FNotAvailableList = nil then
    FNotAvailableList := TFhirHealthcareServiceNotAvailableList.Create;
  result := FNotAvailableList;
end;

Function TFhirHealthcareService.GetHasNotAvailableList : boolean;
begin
  result := (FNotAvailableList <> nil) and (FNotAvailableList.count > 0);
end;

Procedure TFhirHealthcareService.SetAvailabilityExceptions(value : TFhirString);
begin
  FAvailabilityExceptions.free;
  FAvailabilityExceptions := value;
end;

Function TFhirHealthcareService.GetAvailabilityExceptionsST : String;
begin
  if FAvailabilityExceptions = nil then
    result := ''
  else
    result := FAvailabilityExceptions.value;
end;

Procedure TFhirHealthcareService.SetAvailabilityExceptionsST(value : String);
begin
  if value <> '' then
  begin
    if FAvailabilityExceptions = nil then
      FAvailabilityExceptions := TFhirString.create;
    FAvailabilityExceptions.value := value
  end
  else if FAvailabilityExceptions <> nil then
    FAvailabilityExceptions.value := '';
end;


{ TFhirHealthcareServiceListEnumerator }

Constructor TFhirHealthcareServiceListEnumerator.Create(list : TFhirHealthcareServiceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirHealthcareServiceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirHealthcareServiceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirHealthcareServiceListEnumerator.GetCurrent : TFhirHealthcareService;
begin
  Result := FList[FIndex];
end;


{ TFhirHealthcareServiceList }
procedure TFhirHealthcareServiceList.AddItem(value: TFhirHealthcareService);
begin
  assert(value.ClassName = 'TFhirHealthcareService', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirHealthcareService');
  add(value);
end;


function TFhirHealthcareServiceList.Append: TFhirHealthcareService;
begin
  result := TFhirHealthcareService.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceList.ClearItems;
begin
  Clear;
end;

function TFhirHealthcareServiceList.GetEnumerator : TFhirHealthcareServiceListEnumerator;
begin
  result := TFhirHealthcareServiceListEnumerator.Create(self.link);
end;

function TFhirHealthcareServiceList.Clone: TFhirHealthcareServiceList;
begin
  result := TFhirHealthcareServiceList(inherited Clone);
end;

function TFhirHealthcareServiceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirHealthcareServiceList.GetItemN(index: Integer): TFhirHealthcareService;
begin
  result := TFhirHealthcareService(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceList.IndexOf(value: TFhirHealthcareService): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirHealthcareServiceList.Insert(index: Integer): TFhirHealthcareService;
begin
  result := TFhirHealthcareService.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirHealthcareServiceList.InsertItem(index: Integer; value: TFhirHealthcareService);
begin
  assert(value is TFhirHealthcareService);
  Inherited Insert(index, value);
end;

function TFhirHealthcareServiceList.Item(index: Integer): TFhirHealthcareService;
begin
  result := TFhirHealthcareService(ObjectByIndex[index]);
end;

function TFhirHealthcareServiceList.Link: TFhirHealthcareServiceList;
begin
  result := TFhirHealthcareServiceList(inherited Link);
end;

procedure TFhirHealthcareServiceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirHealthcareServiceList.SetItemByIndex(index: Integer; value: TFhirHealthcareService);
begin
  assert(value is TFhirHealthcareService);
  FhirHealthcareServices[index] := value;
end;

procedure TFhirHealthcareServiceList.SetItemN(index: Integer; value: TFhirHealthcareService);
begin
  assert(value is TFhirHealthcareService);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingObjectSelectionStudy }

constructor TFhirImagingObjectSelectionStudy.Create;
begin
  inherited;
end;

destructor TFhirImagingObjectSelectionStudy.Destroy;
begin
  FUid.free;
  FUrl.free;
  FImagingStudy.free;
  FSeriesList.Free;
  inherited;
end;

procedure TFhirImagingObjectSelectionStudy.Assign(oSource : TAdvObject);
begin
  inherited;
  uidElement := TFhirImagingObjectSelectionStudy(oSource).uidElement.Clone;
  urlElement := TFhirImagingObjectSelectionStudy(oSource).urlElement.Clone;
  imagingStudy := TFhirImagingObjectSelectionStudy(oSource).imagingStudy.Clone;
  if (TFhirImagingObjectSelectionStudy(oSource).FSeriesList = nil) then
  begin
    FSeriesList.free;
    FSeriesList := nil;
  end
  else
  begin
    FSeriesList := TFhirImagingObjectSelectionStudySeriesList.Create;
    FSeriesList.Assign(TFhirImagingObjectSelectionStudy(oSource).FSeriesList);
  end;
end;

procedure TFhirImagingObjectSelectionStudy.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'imagingStudy') Then
     list.add(FImagingStudy.Link);
  if (child_name = 'series') Then
    list.addAll(FSeriesList);
end;

procedure TFhirImagingObjectSelectionStudy.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'imagingStudy', 'Reference(ImagingStudy)', FImagingStudy.Link));{2}
  oList.add(TFHIRProperty.create(self, 'series', '', FSeriesList.Link)){3};
end;

procedure TFhirImagingObjectSelectionStudy.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'imagingStudy') then ImagingStudy := propValue as TFhirReference{TFhirImagingStudy}{4b}
  else if (propName = 'series') then SeriesList.add(propValue as TFhirImagingObjectSelectionStudySeries){2}
  else inherited;
end;

function TFhirImagingObjectSelectionStudy.FhirType : string;
begin
  result := 'study';
end;

function TFhirImagingObjectSelectionStudy.Link : TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy(inherited Link);
end;

function TFhirImagingObjectSelectionStudy.Clone : TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy(inherited Clone);
end;

function TFhirImagingObjectSelectionStudy.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudy;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudy)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudy(other);
    result := compareDeep(uidElement, o.uidElement, true) and compareDeep(urlElement, o.urlElement, true) and 
      compareDeep(imagingStudyElement, o.imagingStudyElement, true) and compareDeep(seriesList, o.seriesList, true);
  end;
end;

function TFhirImagingObjectSelectionStudy.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudy;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudy)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudy(other);
    result := compareValues(uidElement, o.uidElement, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirImagingObjectSelectionStudy }

Procedure TFhirImagingObjectSelectionStudy.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingObjectSelectionStudy.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingObjectSelectionStudy.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingObjectSelectionStudy.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingObjectSelectionStudy.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingObjectSelectionStudy.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirImagingObjectSelectionStudy.SetImagingStudy(value : TFhirReference{TFhirImagingStudy});
begin
  FImagingStudy.free;
  FImagingStudy := value;
end;

Function TFhirImagingObjectSelectionStudy.GetSeriesList : TFhirImagingObjectSelectionStudySeriesList;
begin
  if FSeriesList = nil then
    FSeriesList := TFhirImagingObjectSelectionStudySeriesList.Create;
  result := FSeriesList;
end;

Function TFhirImagingObjectSelectionStudy.GetHasSeriesList : boolean;
begin
  result := (FSeriesList <> nil) and (FSeriesList.count > 0);
end;


{ TFhirImagingObjectSelectionStudyListEnumerator }

Constructor TFhirImagingObjectSelectionStudyListEnumerator.Create(list : TFhirImagingObjectSelectionStudyList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingObjectSelectionStudyListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingObjectSelectionStudyListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingObjectSelectionStudyListEnumerator.GetCurrent : TFhirImagingObjectSelectionStudy;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingObjectSelectionStudyList }
procedure TFhirImagingObjectSelectionStudyList.AddItem(value: TFhirImagingObjectSelectionStudy);
begin
  assert(value.ClassName = 'TFhirImagingObjectSelectionStudy', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingObjectSelectionStudy');
  add(value);
end;


function TFhirImagingObjectSelectionStudyList.Append: TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudyList.ClearItems;
begin
  Clear;
end;

function TFhirImagingObjectSelectionStudyList.GetEnumerator : TFhirImagingObjectSelectionStudyListEnumerator;
begin
  result := TFhirImagingObjectSelectionStudyListEnumerator.Create(self.link);
end;

function TFhirImagingObjectSelectionStudyList.Clone: TFhirImagingObjectSelectionStudyList;
begin
  result := TFhirImagingObjectSelectionStudyList(inherited Clone);
end;

function TFhirImagingObjectSelectionStudyList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingObjectSelectionStudyList.GetItemN(index: Integer): TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudyList.IndexOf(value: TFhirImagingObjectSelectionStudy): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingObjectSelectionStudyList.Insert(index: Integer): TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudyList.InsertItem(index: Integer; value: TFhirImagingObjectSelectionStudy);
begin
  assert(value is TFhirImagingObjectSelectionStudy);
  Inherited Insert(index, value);
end;

function TFhirImagingObjectSelectionStudyList.Item(index: Integer): TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudyList.Link: TFhirImagingObjectSelectionStudyList;
begin
  result := TFhirImagingObjectSelectionStudyList(inherited Link);
end;

procedure TFhirImagingObjectSelectionStudyList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingObjectSelectionStudyList.SetItemByIndex(index: Integer; value: TFhirImagingObjectSelectionStudy);
begin
  assert(value is TFhirImagingObjectSelectionStudy);
  FhirImagingObjectSelectionStudies[index] := value;
end;

procedure TFhirImagingObjectSelectionStudyList.SetItemN(index: Integer; value: TFhirImagingObjectSelectionStudy);
begin
  assert(value is TFhirImagingObjectSelectionStudy);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingObjectSelectionStudySeries }

constructor TFhirImagingObjectSelectionStudySeries.Create;
begin
  inherited;
end;

destructor TFhirImagingObjectSelectionStudySeries.Destroy;
begin
  FUid.free;
  FUrl.free;
  FInstanceList.Free;
  inherited;
end;

procedure TFhirImagingObjectSelectionStudySeries.Assign(oSource : TAdvObject);
begin
  inherited;
  uidElement := TFhirImagingObjectSelectionStudySeries(oSource).uidElement.Clone;
  urlElement := TFhirImagingObjectSelectionStudySeries(oSource).urlElement.Clone;
  if (TFhirImagingObjectSelectionStudySeries(oSource).FInstanceList = nil) then
  begin
    FInstanceList.free;
    FInstanceList := nil;
  end
  else
  begin
    FInstanceList := TFhirImagingObjectSelectionStudySeriesInstanceList.Create;
    FInstanceList.Assign(TFhirImagingObjectSelectionStudySeries(oSource).FInstanceList);
  end;
end;

procedure TFhirImagingObjectSelectionStudySeries.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'instance') Then
    list.addAll(FInstanceList);
end;

procedure TFhirImagingObjectSelectionStudySeries.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'instance', '', FInstanceList.Link)){3};
end;

procedure TFhirImagingObjectSelectionStudySeries.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'instance') then InstanceList.add(propValue as TFhirImagingObjectSelectionStudySeriesInstance){2}
  else inherited;
end;

function TFhirImagingObjectSelectionStudySeries.FhirType : string;
begin
  result := 'series';
end;

function TFhirImagingObjectSelectionStudySeries.Link : TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries(inherited Link);
end;

function TFhirImagingObjectSelectionStudySeries.Clone : TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeries.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeries;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeries)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeries(other);
    result := compareDeep(uidElement, o.uidElement, true) and compareDeep(urlElement, o.urlElement, true) and 
      compareDeep(instanceList, o.instanceList, true);
  end;
end;

function TFhirImagingObjectSelectionStudySeries.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeries;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeries)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeries(other);
    result := compareValues(uidElement, o.uidElement, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirImagingObjectSelectionStudySeries }

Procedure TFhirImagingObjectSelectionStudySeries.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingObjectSelectionStudySeries.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingObjectSelectionStudySeries.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingObjectSelectionStudySeries.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingObjectSelectionStudySeries.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingObjectSelectionStudySeries.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Function TFhirImagingObjectSelectionStudySeries.GetInstanceList : TFhirImagingObjectSelectionStudySeriesInstanceList;
begin
  if FInstanceList = nil then
    FInstanceList := TFhirImagingObjectSelectionStudySeriesInstanceList.Create;
  result := FInstanceList;
end;

Function TFhirImagingObjectSelectionStudySeries.GetHasInstanceList : boolean;
begin
  result := (FInstanceList <> nil) and (FInstanceList.count > 0);
end;


{ TFhirImagingObjectSelectionStudySeriesListEnumerator }

Constructor TFhirImagingObjectSelectionStudySeriesListEnumerator.Create(list : TFhirImagingObjectSelectionStudySeriesList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingObjectSelectionStudySeriesListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingObjectSelectionStudySeriesListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingObjectSelectionStudySeriesListEnumerator.GetCurrent : TFhirImagingObjectSelectionStudySeries;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingObjectSelectionStudySeriesList }
procedure TFhirImagingObjectSelectionStudySeriesList.AddItem(value: TFhirImagingObjectSelectionStudySeries);
begin
  assert(value.ClassName = 'TFhirImagingObjectSelectionStudySeries', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingObjectSelectionStudySeries');
  add(value);
end;


function TFhirImagingObjectSelectionStudySeriesList.Append: TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesList.ClearItems;
begin
  Clear;
end;

function TFhirImagingObjectSelectionStudySeriesList.GetEnumerator : TFhirImagingObjectSelectionStudySeriesListEnumerator;
begin
  result := TFhirImagingObjectSelectionStudySeriesListEnumerator.Create(self.link);
end;

function TFhirImagingObjectSelectionStudySeriesList.Clone: TFhirImagingObjectSelectionStudySeriesList;
begin
  result := TFhirImagingObjectSelectionStudySeriesList(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeriesList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingObjectSelectionStudySeriesList.GetItemN(index: Integer): TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesList.IndexOf(value: TFhirImagingObjectSelectionStudySeries): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingObjectSelectionStudySeriesList.Insert(index: Integer): TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesList.InsertItem(index: Integer; value: TFhirImagingObjectSelectionStudySeries);
begin
  assert(value is TFhirImagingObjectSelectionStudySeries);
  Inherited Insert(index, value);
end;

function TFhirImagingObjectSelectionStudySeriesList.Item(index: Integer): TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesList.Link: TFhirImagingObjectSelectionStudySeriesList;
begin
  result := TFhirImagingObjectSelectionStudySeriesList(inherited Link);
end;

procedure TFhirImagingObjectSelectionStudySeriesList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingObjectSelectionStudySeriesList.SetItemByIndex(index: Integer; value: TFhirImagingObjectSelectionStudySeries);
begin
  assert(value is TFhirImagingObjectSelectionStudySeries);
  FhirImagingObjectSelectionStudySeries[index] := value;
end;

procedure TFhirImagingObjectSelectionStudySeriesList.SetItemN(index: Integer; value: TFhirImagingObjectSelectionStudySeries);
begin
  assert(value is TFhirImagingObjectSelectionStudySeries);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingObjectSelectionStudySeriesInstance }

constructor TFhirImagingObjectSelectionStudySeriesInstance.Create;
begin
  inherited;
end;

destructor TFhirImagingObjectSelectionStudySeriesInstance.Destroy;
begin
  FSopClass.free;
  FUid.free;
  FUrl.free;
  FFramesList.Free;
  inherited;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstance.Assign(oSource : TAdvObject);
begin
  inherited;
  sopClassElement := TFhirImagingObjectSelectionStudySeriesInstance(oSource).sopClassElement.Clone;
  uidElement := TFhirImagingObjectSelectionStudySeriesInstance(oSource).uidElement.Clone;
  urlElement := TFhirImagingObjectSelectionStudySeriesInstance(oSource).urlElement.Clone;
  if (TFhirImagingObjectSelectionStudySeriesInstance(oSource).FFramesList = nil) then
  begin
    FFramesList.free;
    FFramesList := nil;
  end
  else
  begin
    FFramesList := TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Create;
    FFramesList.Assign(TFhirImagingObjectSelectionStudySeriesInstance(oSource).FFramesList);
  end;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sopClass') Then
     list.add(FSopClass.Link);
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'frames') Then
    list.addAll(FFramesList);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sopClass', 'oid', FSopClass.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'frames', '', FFramesList.Link)){3};
end;

procedure TFhirImagingObjectSelectionStudySeriesInstance.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sopClass') then SopClassElement := propValue as TFhirOid{5a}
  else if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'frames') then FramesList.add(propValue as TFhirImagingObjectSelectionStudySeriesInstanceFrames){2}
  else inherited;
end;

function TFhirImagingObjectSelectionStudySeriesInstance.FhirType : string;
begin
  result := 'instance';
end;

function TFhirImagingObjectSelectionStudySeriesInstance.Link : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance(inherited Link);
end;

function TFhirImagingObjectSelectionStudySeriesInstance.Clone : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeriesInstance.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeriesInstance)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeriesInstance(other);
    result := compareDeep(sopClassElement, o.sopClassElement, true) and compareDeep(uidElement, o.uidElement, true) and 
      compareDeep(urlElement, o.urlElement, true) and compareDeep(framesList, o.framesList, true);
  end;
end;

function TFhirImagingObjectSelectionStudySeriesInstance.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeriesInstance)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeriesInstance(other);
    result := compareValues(sopClassElement, o.sopClassElement, true) and compareValues(uidElement, o.uidElement, true) and 
      compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirImagingObjectSelectionStudySeriesInstance }

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetSopClass(value : TFhirOid);
begin
  FSopClass.free;
  FSopClass := value;
end;

Function TFhirImagingObjectSelectionStudySeriesInstance.GetSopClassST : String;
begin
  if FSopClass = nil then
    result := ''
  else
    result := FSopClass.value;
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetSopClassST(value : String);
begin
  if value <> '' then
  begin
    if FSopClass = nil then
      FSopClass := TFhirOid.create;
    FSopClass.value := value
  end
  else if FSopClass <> nil then
    FSopClass.value := '';
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingObjectSelectionStudySeriesInstance.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingObjectSelectionStudySeriesInstance.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstance.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Function TFhirImagingObjectSelectionStudySeriesInstance.GetFramesList : TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
begin
  if FFramesList = nil then
    FFramesList := TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Create;
  result := FFramesList;
end;

Function TFhirImagingObjectSelectionStudySeriesInstance.GetHasFramesList : boolean;
begin
  result := (FFramesList <> nil) and (FFramesList.count > 0);
end;


{ TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator }

Constructor TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator.Create(list : TFhirImagingObjectSelectionStudySeriesInstanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator.GetCurrent : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingObjectSelectionStudySeriesInstanceList }
procedure TFhirImagingObjectSelectionStudySeriesInstanceList.AddItem(value: TFhirImagingObjectSelectionStudySeriesInstance);
begin
  assert(value.ClassName = 'TFhirImagingObjectSelectionStudySeriesInstance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingObjectSelectionStudySeriesInstance');
  add(value);
end;


function TFhirImagingObjectSelectionStudySeriesInstanceList.Append: TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesInstanceList.ClearItems;
begin
  Clear;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.GetEnumerator : TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceListEnumerator.Create(self.link);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.Clone: TFhirImagingObjectSelectionStudySeriesInstanceList;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceList(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.GetItemN(index: Integer): TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.IndexOf(value: TFhirImagingObjectSelectionStudySeriesInstance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingObjectSelectionStudySeriesInstanceList.Insert(index: Integer): TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesInstanceList.InsertItem(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstance);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstance);
  Inherited Insert(index, value);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.Item(index: Integer): TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceList.Link: TFhirImagingObjectSelectionStudySeriesInstanceList;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceList(inherited Link);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceList.SetItemByIndex(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstance);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstance);
  FhirImagingObjectSelectionStudySeriesInstances[index] := value;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceList.SetItemN(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstance);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstance);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingObjectSelectionStudySeriesInstanceFrames }

constructor TFhirImagingObjectSelectionStudySeriesInstanceFrames.Create;
begin
  inherited;
end;

destructor TFhirImagingObjectSelectionStudySeriesInstanceFrames.Destroy;
begin
  FFrameNumbersList.Free;
  FUrl.free;
  inherited;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirImagingObjectSelectionStudySeriesInstanceFrames(oSource).FFrameNumbersList = nil) then
  begin
    FFrameNumbersList.free;
    FFrameNumbersList := nil;
  end
  else
  begin
    FFrameNumbersList := TFhirUnsignedIntList.Create;
    FFrameNumbersList.Assign(TFhirImagingObjectSelectionStudySeriesInstanceFrames(oSource).FFrameNumbersList);
  end;
  urlElement := TFhirImagingObjectSelectionStudySeriesInstanceFrames(oSource).urlElement.Clone;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'frameNumbers') Then
    list.addAll(FFrameNumbersList);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'frameNumbers', 'unsignedInt', FFrameNumbersList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'frameNumbers') then FrameNumbersList.add(propValue as TFhirUnsignedInt){2}
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFrames.FhirType : string;
begin
  result := 'frames';
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFrames.Link : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames(inherited Link);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFrames.Clone : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFrames.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeriesInstanceFrames)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeriesInstanceFrames(other);
    result := compareDeep(frameNumbersList, o.frameNumbersList, true) and compareDeep(urlElement, o.urlElement, true);
  end;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFrames.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingObjectSelectionStudySeriesInstanceFrames)) then
    result := false
  else
  begin
    o := TFhirImagingObjectSelectionStudySeriesInstanceFrames(other);
    result := compareValues(frameNumbersList, o.frameNumbersList, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirImagingObjectSelectionStudySeriesInstanceFrames }

Function TFhirImagingObjectSelectionStudySeriesInstanceFrames.GetFrameNumbersList : TFhirUnsignedIntList;
begin
  if FFrameNumbersList = nil then
    FFrameNumbersList := TFhirUnsignedIntList.Create;
  result := FFrameNumbersList;
end;

Function TFhirImagingObjectSelectionStudySeriesInstanceFrames.GetHasFrameNumbersList : boolean;
begin
  result := (FFrameNumbersList <> nil) and (FFrameNumbersList.count > 0);
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingObjectSelectionStudySeriesInstanceFrames.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingObjectSelectionStudySeriesInstanceFrames.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;


{ TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator }

Constructor TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator.Create(list : TFhirImagingObjectSelectionStudySeriesInstanceFramesList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator.GetCurrent : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingObjectSelectionStudySeriesInstanceFramesList }
procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.AddItem(value: TFhirImagingObjectSelectionStudySeriesInstanceFrames);
begin
  assert(value.ClassName = 'TFhirImagingObjectSelectionStudySeriesInstanceFrames', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingObjectSelectionStudySeriesInstanceFrames');
  add(value);
end;


function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Append: TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.ClearItems;
begin
  Clear;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.GetEnumerator : TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFramesListEnumerator.Create(self.link);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Clone: TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFramesList(inherited Clone);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.GetItemN(index: Integer): TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.IndexOf(value: TFhirImagingObjectSelectionStudySeriesInstanceFrames): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Insert(index: Integer): TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.InsertItem(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstanceFrames);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstanceFrames);
  Inherited Insert(index, value);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Item(index: Integer): TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Link: TFhirImagingObjectSelectionStudySeriesInstanceFramesList;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFramesList(inherited Link);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.SetItemByIndex(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstanceFrames);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstanceFrames);
  FhirImagingObjectSelectionStudySeriesInstanceFrames[index] := value;
end;

procedure TFhirImagingObjectSelectionStudySeriesInstanceFramesList.SetItemN(index: Integer; value: TFhirImagingObjectSelectionStudySeriesInstanceFrames);
begin
  assert(value is TFhirImagingObjectSelectionStudySeriesInstanceFrames);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingObjectSelection }

constructor TFhirImagingObjectSelection.Create;
begin
  inherited;
end;

destructor TFhirImagingObjectSelection.Destroy;
begin
  FUid.free;
  FPatient.free;
  FTitle.free;
  FDescription.free;
  FAuthor.free;
  FAuthoringTime.free;
  FStudyList.Free;
  inherited;
end;

function TFhirImagingObjectSelection.GetResourceType : TFhirResourceType;
begin
  result := frtImagingObjectSelection;
end;

procedure TFhirImagingObjectSelection.Assign(oSource : TAdvObject);
begin
  inherited;
  uidElement := TFhirImagingObjectSelection(oSource).uidElement.Clone;
  patient := TFhirImagingObjectSelection(oSource).patient.Clone;
  title := TFhirImagingObjectSelection(oSource).title.Clone;
  descriptionElement := TFhirImagingObjectSelection(oSource).descriptionElement.Clone;
  author := TFhirImagingObjectSelection(oSource).author.Clone;
  authoringTimeElement := TFhirImagingObjectSelection(oSource).authoringTimeElement.Clone;
  if (TFhirImagingObjectSelection(oSource).FStudyList = nil) then
  begin
    FStudyList.free;
    FStudyList := nil;
  end
  else
  begin
    FStudyList := TFhirImagingObjectSelectionStudyList.Create;
    FStudyList.Assign(TFhirImagingObjectSelection(oSource).FStudyList);
  end;
end;

procedure TFhirImagingObjectSelection.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'authoringTime') Then
     list.add(FAuthoringTime.Link);
  if (child_name = 'study') Then
    list.addAll(FStudyList);
end;

procedure TFhirImagingObjectSelection.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'title', 'CodeableConcept', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner|Device|Organization|Patient|RelatedPerson)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authoringTime', 'dateTime', FAuthoringTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'study', '', FStudyList.Link)){3};
end;

procedure TFhirImagingObjectSelection.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'title') then Title := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'authoringTime') then AuthoringTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'study') then StudyList.add(propValue as TFhirImagingObjectSelectionStudy){2}
  else inherited;
end;

function TFhirImagingObjectSelection.FhirType : string;
begin
  result := 'ImagingObjectSelection';
end;

function TFhirImagingObjectSelection.Link : TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection(inherited Link);
end;

function TFhirImagingObjectSelection.Clone : TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection(inherited Clone);
end;

{ TFhirImagingObjectSelection }

Procedure TFhirImagingObjectSelection.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingObjectSelection.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingObjectSelection.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingObjectSelection.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirImagingObjectSelection.SetTitle(value : TFhirCodeableConcept);
begin
  FTitle.free;
  FTitle := value;
end;

Procedure TFhirImagingObjectSelection.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImagingObjectSelection.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImagingObjectSelection.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirImagingObjectSelection.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirImagingObjectSelection.SetAuthoringTime(value : TFhirDateTime);
begin
  FAuthoringTime.free;
  FAuthoringTime := value;
end;

Function TFhirImagingObjectSelection.GetAuthoringTimeST : TDateAndTime;
begin
  if FAuthoringTime = nil then
    result := nil
  else
    result := FAuthoringTime.value;
end;

Procedure TFhirImagingObjectSelection.SetAuthoringTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FAuthoringTime = nil then
      FAuthoringTime := TFhirDateTime.create;
    FAuthoringTime.value := value
  end
  else if FAuthoringTime <> nil then
    FAuthoringTime.value := nil;
end;

Function TFhirImagingObjectSelection.GetStudyList : TFhirImagingObjectSelectionStudyList;
begin
  if FStudyList = nil then
    FStudyList := TFhirImagingObjectSelectionStudyList.Create;
  result := FStudyList;
end;

Function TFhirImagingObjectSelection.GetHasStudyList : boolean;
begin
  result := (FStudyList <> nil) and (FStudyList.count > 0);
end;


{ TFhirImagingObjectSelectionListEnumerator }

Constructor TFhirImagingObjectSelectionListEnumerator.Create(list : TFhirImagingObjectSelectionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingObjectSelectionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingObjectSelectionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingObjectSelectionListEnumerator.GetCurrent : TFhirImagingObjectSelection;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingObjectSelectionList }
procedure TFhirImagingObjectSelectionList.AddItem(value: TFhirImagingObjectSelection);
begin
  assert(value.ClassName = 'TFhirImagingObjectSelection', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingObjectSelection');
  add(value);
end;


function TFhirImagingObjectSelectionList.Append: TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionList.ClearItems;
begin
  Clear;
end;

function TFhirImagingObjectSelectionList.GetEnumerator : TFhirImagingObjectSelectionListEnumerator;
begin
  result := TFhirImagingObjectSelectionListEnumerator.Create(self.link);
end;

function TFhirImagingObjectSelectionList.Clone: TFhirImagingObjectSelectionList;
begin
  result := TFhirImagingObjectSelectionList(inherited Clone);
end;

function TFhirImagingObjectSelectionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingObjectSelectionList.GetItemN(index: Integer): TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionList.IndexOf(value: TFhirImagingObjectSelection): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingObjectSelectionList.Insert(index: Integer): TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingObjectSelectionList.InsertItem(index: Integer; value: TFhirImagingObjectSelection);
begin
  assert(value is TFhirImagingObjectSelection);
  Inherited Insert(index, value);
end;

function TFhirImagingObjectSelectionList.Item(index: Integer): TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection(ObjectByIndex[index]);
end;

function TFhirImagingObjectSelectionList.Link: TFhirImagingObjectSelectionList;
begin
  result := TFhirImagingObjectSelectionList(inherited Link);
end;

procedure TFhirImagingObjectSelectionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingObjectSelectionList.SetItemByIndex(index: Integer; value: TFhirImagingObjectSelection);
begin
  assert(value is TFhirImagingObjectSelection);
  FhirImagingObjectSelections[index] := value;
end;

procedure TFhirImagingObjectSelectionList.SetItemN(index: Integer; value: TFhirImagingObjectSelection);
begin
  assert(value is TFhirImagingObjectSelection);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingStudySeries }

constructor TFhirImagingStudySeries.Create;
begin
  inherited;
end;

destructor TFhirImagingStudySeries.Destroy;
begin
  FNumber.free;
  FModality.free;
  FUid.free;
  FDescription.free;
  FNumberOfInstances.free;
  FAvailability.free;
  FUrl.free;
  FBodySite.free;
  FLaterality.free;
  FStarted.free;
  FInstanceList.Free;
  inherited;
end;

procedure TFhirImagingStudySeries.Assign(oSource : TAdvObject);
begin
  inherited;
  numberElement := TFhirImagingStudySeries(oSource).numberElement.Clone;
  modality := TFhirImagingStudySeries(oSource).modality.Clone;
  uidElement := TFhirImagingStudySeries(oSource).uidElement.Clone;
  descriptionElement := TFhirImagingStudySeries(oSource).descriptionElement.Clone;
  numberOfInstancesElement := TFhirImagingStudySeries(oSource).numberOfInstancesElement.Clone;
  FAvailability := TFhirImagingStudySeries(oSource).FAvailability.Link;
  urlElement := TFhirImagingStudySeries(oSource).urlElement.Clone;
  bodySite := TFhirImagingStudySeries(oSource).bodySite.Clone;
  laterality := TFhirImagingStudySeries(oSource).laterality.Clone;
  startedElement := TFhirImagingStudySeries(oSource).startedElement.Clone;
  if (TFhirImagingStudySeries(oSource).FInstanceList = nil) then
  begin
    FInstanceList.free;
    FInstanceList := nil;
  end
  else
  begin
    FInstanceList := TFhirImagingStudySeriesInstanceList.Create;
    FInstanceList.Assign(TFhirImagingStudySeries(oSource).FInstanceList);
  end;
end;

procedure TFhirImagingStudySeries.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'number') Then
     list.add(FNumber.Link);
  if (child_name = 'modality') Then
     list.add(FModality.Link);
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'numberOfInstances') Then
     list.add(FNumberOfInstances.Link);
  if (child_name = 'availability') Then
     list.add(FAvailability.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'bodySite') Then
     list.add(FBodySite.Link);
  if (child_name = 'laterality') Then
     list.add(FLaterality.Link);
  if (child_name = 'started') Then
     list.add(FStarted.Link);
  if (child_name = 'instance') Then
    list.addAll(FInstanceList);
end;

procedure TFhirImagingStudySeries.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'number', 'unsignedInt', FNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'modality', 'Coding', FModality.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'numberOfInstances', 'unsignedInt', FNumberOfInstances.Link));{2}
  oList.add(TFHIRProperty.create(self, 'availability', 'code', FAvailability.Link));{1}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bodySite', 'Coding', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'laterality', 'Coding', FLaterality.Link));{2}
  oList.add(TFHIRProperty.create(self, 'started', 'dateTime', FStarted.Link));{2}
  oList.add(TFHIRProperty.create(self, 'instance', '', FInstanceList.Link)){3};
end;

procedure TFhirImagingStudySeries.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'number') then NumberElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'modality') then Modality := propValue as TFhirCoding{4b}
  else if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'numberOfInstances') then NumberOfInstancesElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'availability') then AvailabilityElement := propValue as TFHIREnum
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'bodySite') then BodySite := propValue as TFhirCoding{4b}
  else if (propName = 'laterality') then Laterality := propValue as TFhirCoding{4b}
  else if (propName = 'started') then StartedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'instance') then InstanceList.add(propValue as TFhirImagingStudySeriesInstance){2}
  else inherited;
end;

function TFhirImagingStudySeries.FhirType : string;
begin
  result := 'series';
end;

function TFhirImagingStudySeries.Link : TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries(inherited Link);
end;

function TFhirImagingStudySeries.Clone : TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries(inherited Clone);
end;

function TFhirImagingStudySeries.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingStudySeries;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingStudySeries)) then
    result := false
  else
  begin
    o := TFhirImagingStudySeries(other);
    result := compareDeep(numberElement, o.numberElement, true) and compareDeep(modalityElement, o.modalityElement, true) and 
      compareDeep(uidElement, o.uidElement, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(numberOfInstancesElement, o.numberOfInstancesElement, true) and compareDeep(availabilityElement, o.availabilityElement, true) and 
      compareDeep(urlElement, o.urlElement, true) and compareDeep(bodySiteElement, o.bodySiteElement, true) and 
      compareDeep(lateralityElement, o.lateralityElement, true) and compareDeep(startedElement, o.startedElement, true) and 
      compareDeep(instanceList, o.instanceList, true);
  end;
end;

function TFhirImagingStudySeries.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingStudySeries;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingStudySeries)) then
    result := false
  else
  begin
    o := TFhirImagingStudySeries(other);
    result := compareValues(numberElement, o.numberElement, true) and compareValues(uidElement, o.uidElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(numberOfInstancesElement, o.numberOfInstancesElement, true) and 
      compareValues(availabilityElement, o.availabilityElement, true) and compareValues(urlElement, o.urlElement, true) and 
      compareValues(startedElement, o.startedElement, true);
  end;
end;

{ TFhirImagingStudySeries }

Procedure TFhirImagingStudySeries.SetNumber(value : TFhirUnsignedInt);
begin
  FNumber.free;
  FNumber := value;
end;

Function TFhirImagingStudySeries.GetNumberST : String;
begin
  if FNumber = nil then
    result := ''
  else
    result := FNumber.value;
end;

Procedure TFhirImagingStudySeries.SetNumberST(value : String);
begin
  if value <> '' then
  begin
    if FNumber = nil then
      FNumber := TFhirUnsignedInt.create;
    FNumber.value := value
  end
  else if FNumber <> nil then
    FNumber.value := '';
end;

Procedure TFhirImagingStudySeries.SetModality(value : TFhirCoding);
begin
  FModality.free;
  FModality := value;
end;

Procedure TFhirImagingStudySeries.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingStudySeries.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingStudySeries.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingStudySeries.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImagingStudySeries.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImagingStudySeries.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirImagingStudySeries.SetNumberOfInstances(value : TFhirUnsignedInt);
begin
  FNumberOfInstances.free;
  FNumberOfInstances := value;
end;

Function TFhirImagingStudySeries.GetNumberOfInstancesST : String;
begin
  if FNumberOfInstances = nil then
    result := ''
  else
    result := FNumberOfInstances.value;
end;

Procedure TFhirImagingStudySeries.SetNumberOfInstancesST(value : String);
begin
  if value <> '' then
  begin
    if FNumberOfInstances = nil then
      FNumberOfInstances := TFhirUnsignedInt.create;
    FNumberOfInstances.value := value
  end
  else if FNumberOfInstances <> nil then
    FNumberOfInstances.value := '';
end;

Procedure TFhirImagingStudySeries.SetAvailability(value : TFhirEnum);
begin
  FAvailability.free;
  FAvailability := value;
end;

Function TFhirImagingStudySeries.GetAvailabilityST : TFhirInstanceAvailability;
begin
  if FAvailability = nil then
    result := TFhirInstanceAvailability(0)
  else
    result := TFhirInstanceAvailability(StringArrayIndexOfSensitive(CODES_TFhirInstanceAvailability, FAvailability.value));
end;

Procedure TFhirImagingStudySeries.SetAvailabilityST(value : TFhirInstanceAvailability);
begin
  if ord(value) = 0 then
    AvailabilityElement := nil
  else
    AvailabilityElement := TFhirEnum.create(CODES_TFhirInstanceAvailability[value]);
end;

Procedure TFhirImagingStudySeries.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingStudySeries.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingStudySeries.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirImagingStudySeries.SetBodySite(value : TFhirCoding);
begin
  FBodySite.free;
  FBodySite := value;
end;

Procedure TFhirImagingStudySeries.SetLaterality(value : TFhirCoding);
begin
  FLaterality.free;
  FLaterality := value;
end;

Procedure TFhirImagingStudySeries.SetStarted(value : TFhirDateTime);
begin
  FStarted.free;
  FStarted := value;
end;

Function TFhirImagingStudySeries.GetStartedST : TDateAndTime;
begin
  if FStarted = nil then
    result := nil
  else
    result := FStarted.value;
end;

Procedure TFhirImagingStudySeries.SetStartedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStarted = nil then
      FStarted := TFhirDateTime.create;
    FStarted.value := value
  end
  else if FStarted <> nil then
    FStarted.value := nil;
end;

Function TFhirImagingStudySeries.GetInstanceList : TFhirImagingStudySeriesInstanceList;
begin
  if FInstanceList = nil then
    FInstanceList := TFhirImagingStudySeriesInstanceList.Create;
  result := FInstanceList;
end;

Function TFhirImagingStudySeries.GetHasInstanceList : boolean;
begin
  result := (FInstanceList <> nil) and (FInstanceList.count > 0);
end;


{ TFhirImagingStudySeriesListEnumerator }

Constructor TFhirImagingStudySeriesListEnumerator.Create(list : TFhirImagingStudySeriesList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingStudySeriesListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingStudySeriesListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingStudySeriesListEnumerator.GetCurrent : TFhirImagingStudySeries;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingStudySeriesList }
procedure TFhirImagingStudySeriesList.AddItem(value: TFhirImagingStudySeries);
begin
  assert(value.ClassName = 'TFhirImagingStudySeries', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingStudySeries');
  add(value);
end;


function TFhirImagingStudySeriesList.Append: TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudySeriesList.ClearItems;
begin
  Clear;
end;

function TFhirImagingStudySeriesList.GetEnumerator : TFhirImagingStudySeriesListEnumerator;
begin
  result := TFhirImagingStudySeriesListEnumerator.Create(self.link);
end;

function TFhirImagingStudySeriesList.Clone: TFhirImagingStudySeriesList;
begin
  result := TFhirImagingStudySeriesList(inherited Clone);
end;

function TFhirImagingStudySeriesList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingStudySeriesList.GetItemN(index: Integer): TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries(ObjectByIndex[index]);
end;

function TFhirImagingStudySeriesList.IndexOf(value: TFhirImagingStudySeries): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingStudySeriesList.Insert(index: Integer): TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudySeriesList.InsertItem(index: Integer; value: TFhirImagingStudySeries);
begin
  assert(value is TFhirImagingStudySeries);
  Inherited Insert(index, value);
end;

function TFhirImagingStudySeriesList.Item(index: Integer): TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries(ObjectByIndex[index]);
end;

function TFhirImagingStudySeriesList.Link: TFhirImagingStudySeriesList;
begin
  result := TFhirImagingStudySeriesList(inherited Link);
end;

procedure TFhirImagingStudySeriesList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingStudySeriesList.SetItemByIndex(index: Integer; value: TFhirImagingStudySeries);
begin
  assert(value is TFhirImagingStudySeries);
  FhirImagingStudySeries[index] := value;
end;

procedure TFhirImagingStudySeriesList.SetItemN(index: Integer; value: TFhirImagingStudySeries);
begin
  assert(value is TFhirImagingStudySeries);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingStudySeriesInstance }

constructor TFhirImagingStudySeriesInstance.Create;
begin
  inherited;
end;

destructor TFhirImagingStudySeriesInstance.Destroy;
begin
  FNumber.free;
  FUid.free;
  FSopClass.free;
  FType_.free;
  FTitle.free;
  FContentList.Free;
  inherited;
end;

procedure TFhirImagingStudySeriesInstance.Assign(oSource : TAdvObject);
begin
  inherited;
  numberElement := TFhirImagingStudySeriesInstance(oSource).numberElement.Clone;
  uidElement := TFhirImagingStudySeriesInstance(oSource).uidElement.Clone;
  sopClassElement := TFhirImagingStudySeriesInstance(oSource).sopClassElement.Clone;
  type_Element := TFhirImagingStudySeriesInstance(oSource).type_Element.Clone;
  titleElement := TFhirImagingStudySeriesInstance(oSource).titleElement.Clone;
  if (TFhirImagingStudySeriesInstance(oSource).FContentList = nil) then
  begin
    FContentList.free;
    FContentList := nil;
  end
  else
  begin
    FContentList := TFhirAttachmentList.Create;
    FContentList.Assign(TFhirImagingStudySeriesInstance(oSource).FContentList);
  end;
end;

procedure TFhirImagingStudySeriesInstance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'number') Then
     list.add(FNumber.Link);
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'sopClass') Then
     list.add(FSopClass.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'content') Then
    list.addAll(FContentList);
end;

procedure TFhirImagingStudySeriesInstance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'number', 'unsignedInt', FNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sopClass', 'oid', FSopClass.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'string', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'content', 'Attachment', FContentList.Link)){3};
end;

procedure TFhirImagingStudySeriesInstance.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'number') then NumberElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'sopClass') then SopClassElement := propValue as TFhirOid{5a}
  else if (propName = 'type') then Type_Element := propValue as TFhirString{5a}
  else if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'content') then ContentList.add(propValue as TFhirAttachment){2}
  else inherited;
end;

function TFhirImagingStudySeriesInstance.FhirType : string;
begin
  result := 'instance';
end;

function TFhirImagingStudySeriesInstance.Link : TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance(inherited Link);
end;

function TFhirImagingStudySeriesInstance.Clone : TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance(inherited Clone);
end;

function TFhirImagingStudySeriesInstance.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingStudySeriesInstance;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImagingStudySeriesInstance)) then
    result := false
  else
  begin
    o := TFhirImagingStudySeriesInstance(other);
    result := compareDeep(numberElement, o.numberElement, true) and compareDeep(uidElement, o.uidElement, true) and 
      compareDeep(sopClassElement, o.sopClassElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(titleElement, o.titleElement, true) and compareDeep(contentList, o.contentList, true);
  end;
end;

function TFhirImagingStudySeriesInstance.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImagingStudySeriesInstance;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImagingStudySeriesInstance)) then
    result := false
  else
  begin
    o := TFhirImagingStudySeriesInstance(other);
    result := compareValues(numberElement, o.numberElement, true) and compareValues(uidElement, o.uidElement, true) and 
      compareValues(sopClassElement, o.sopClassElement, true) and compareValues(type_Element, o.type_Element, true) and 
      compareValues(titleElement, o.titleElement, true);
  end;
end;

{ TFhirImagingStudySeriesInstance }

Procedure TFhirImagingStudySeriesInstance.SetNumber(value : TFhirUnsignedInt);
begin
  FNumber.free;
  FNumber := value;
end;

Function TFhirImagingStudySeriesInstance.GetNumberST : String;
begin
  if FNumber = nil then
    result := ''
  else
    result := FNumber.value;
end;

Procedure TFhirImagingStudySeriesInstance.SetNumberST(value : String);
begin
  if value <> '' then
  begin
    if FNumber = nil then
      FNumber := TFhirUnsignedInt.create;
    FNumber.value := value
  end
  else if FNumber <> nil then
    FNumber.value := '';
end;

Procedure TFhirImagingStudySeriesInstance.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingStudySeriesInstance.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingStudySeriesInstance.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingStudySeriesInstance.SetSopClass(value : TFhirOid);
begin
  FSopClass.free;
  FSopClass := value;
end;

Function TFhirImagingStudySeriesInstance.GetSopClassST : String;
begin
  if FSopClass = nil then
    result := ''
  else
    result := FSopClass.value;
end;

Procedure TFhirImagingStudySeriesInstance.SetSopClassST(value : String);
begin
  if value <> '' then
  begin
    if FSopClass = nil then
      FSopClass := TFhirOid.create;
    FSopClass.value := value
  end
  else if FSopClass <> nil then
    FSopClass.value := '';
end;

Procedure TFhirImagingStudySeriesInstance.SetType_(value : TFhirString);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirImagingStudySeriesInstance.GetType_ST : String;
begin
  if FType_ = nil then
    result := ''
  else
    result := FType_.value;
end;

Procedure TFhirImagingStudySeriesInstance.SetType_ST(value : String);
begin
  if value <> '' then
  begin
    if FType_ = nil then
      FType_ := TFhirString.create;
    FType_.value := value
  end
  else if FType_ <> nil then
    FType_.value := '';
end;

Procedure TFhirImagingStudySeriesInstance.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirImagingStudySeriesInstance.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirImagingStudySeriesInstance.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Function TFhirImagingStudySeriesInstance.GetContentList : TFhirAttachmentList;
begin
  if FContentList = nil then
    FContentList := TFhirAttachmentList.Create;
  result := FContentList;
end;

Function TFhirImagingStudySeriesInstance.GetHasContentList : boolean;
begin
  result := (FContentList <> nil) and (FContentList.count > 0);
end;


{ TFhirImagingStudySeriesInstanceListEnumerator }

Constructor TFhirImagingStudySeriesInstanceListEnumerator.Create(list : TFhirImagingStudySeriesInstanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingStudySeriesInstanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingStudySeriesInstanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingStudySeriesInstanceListEnumerator.GetCurrent : TFhirImagingStudySeriesInstance;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingStudySeriesInstanceList }
procedure TFhirImagingStudySeriesInstanceList.AddItem(value: TFhirImagingStudySeriesInstance);
begin
  assert(value.ClassName = 'TFhirImagingStudySeriesInstance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingStudySeriesInstance');
  add(value);
end;


function TFhirImagingStudySeriesInstanceList.Append: TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudySeriesInstanceList.ClearItems;
begin
  Clear;
end;

function TFhirImagingStudySeriesInstanceList.GetEnumerator : TFhirImagingStudySeriesInstanceListEnumerator;
begin
  result := TFhirImagingStudySeriesInstanceListEnumerator.Create(self.link);
end;

function TFhirImagingStudySeriesInstanceList.Clone: TFhirImagingStudySeriesInstanceList;
begin
  result := TFhirImagingStudySeriesInstanceList(inherited Clone);
end;

function TFhirImagingStudySeriesInstanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingStudySeriesInstanceList.GetItemN(index: Integer): TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance(ObjectByIndex[index]);
end;

function TFhirImagingStudySeriesInstanceList.IndexOf(value: TFhirImagingStudySeriesInstance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingStudySeriesInstanceList.Insert(index: Integer): TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudySeriesInstanceList.InsertItem(index: Integer; value: TFhirImagingStudySeriesInstance);
begin
  assert(value is TFhirImagingStudySeriesInstance);
  Inherited Insert(index, value);
end;

function TFhirImagingStudySeriesInstanceList.Item(index: Integer): TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance(ObjectByIndex[index]);
end;

function TFhirImagingStudySeriesInstanceList.Link: TFhirImagingStudySeriesInstanceList;
begin
  result := TFhirImagingStudySeriesInstanceList(inherited Link);
end;

procedure TFhirImagingStudySeriesInstanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingStudySeriesInstanceList.SetItemByIndex(index: Integer; value: TFhirImagingStudySeriesInstance);
begin
  assert(value is TFhirImagingStudySeriesInstance);
  FhirImagingStudySeriesInstances[index] := value;
end;

procedure TFhirImagingStudySeriesInstanceList.SetItemN(index: Integer; value: TFhirImagingStudySeriesInstance);
begin
  assert(value is TFhirImagingStudySeriesInstance);
  ObjectByIndex[index] := value;
end;

{ TFhirImagingStudy }

constructor TFhirImagingStudy.Create;
begin
  inherited;
end;

destructor TFhirImagingStudy.Destroy;
begin
  FStarted.free;
  FPatient.free;
  FUid.free;
  FAccession.free;
  FIdentifierList.Free;
  FOrderList.Free;
  FModalityListList.Free;
  FReferrer.free;
  FAvailability.free;
  FUrl.free;
  FNumberOfSeries.free;
  FNumberOfInstances.free;
  FProcedure_List.Free;
  FInterpreter.free;
  FDescription.free;
  FSeriesList.Free;
  inherited;
end;

function TFhirImagingStudy.GetResourceType : TFhirResourceType;
begin
  result := frtImagingStudy;
end;

procedure TFhirImagingStudy.Assign(oSource : TAdvObject);
begin
  inherited;
  startedElement := TFhirImagingStudy(oSource).startedElement.Clone;
  patient := TFhirImagingStudy(oSource).patient.Clone;
  uidElement := TFhirImagingStudy(oSource).uidElement.Clone;
  accession := TFhirImagingStudy(oSource).accession.Clone;
  if (TFhirImagingStudy(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirImagingStudy(oSource).FIdentifierList);
  end;
  if (TFhirImagingStudy(oSource).FOrderList = nil) then
  begin
    FOrderList.free;
    FOrderList := nil;
  end
  else
  begin
    FOrderList := TFhirReferenceList{TFhirDiagnosticOrder}.Create;
    FOrderList.Assign(TFhirImagingStudy(oSource).FOrderList);
  end;
  if (TFhirImagingStudy(oSource).FModalityListList = nil) then
  begin
    FModalityListList.free;
    FModalityListList := nil;
  end
  else
  begin
    FModalityListList := TFhirCodingList.Create;
    FModalityListList.Assign(TFhirImagingStudy(oSource).FModalityListList);
  end;
  referrer := TFhirImagingStudy(oSource).referrer.Clone;
  FAvailability := TFhirImagingStudy(oSource).FAvailability.Link;
  urlElement := TFhirImagingStudy(oSource).urlElement.Clone;
  numberOfSeriesElement := TFhirImagingStudy(oSource).numberOfSeriesElement.Clone;
  numberOfInstancesElement := TFhirImagingStudy(oSource).numberOfInstancesElement.Clone;
  if (TFhirImagingStudy(oSource).FProcedure_List = nil) then
  begin
    FProcedure_List.free;
    FProcedure_List := nil;
  end
  else
  begin
    FProcedure_List := TFhirReferenceList{TFhirProcedure}.Create;
    FProcedure_List.Assign(TFhirImagingStudy(oSource).FProcedure_List);
  end;
  interpreter := TFhirImagingStudy(oSource).interpreter.Clone;
  descriptionElement := TFhirImagingStudy(oSource).descriptionElement.Clone;
  if (TFhirImagingStudy(oSource).FSeriesList = nil) then
  begin
    FSeriesList.free;
    FSeriesList := nil;
  end
  else
  begin
    FSeriesList := TFhirImagingStudySeriesList.Create;
    FSeriesList.Assign(TFhirImagingStudy(oSource).FSeriesList);
  end;
end;

procedure TFhirImagingStudy.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'started') Then
     list.add(FStarted.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'uid') Then
     list.add(FUid.Link);
  if (child_name = 'accession') Then
     list.add(FAccession.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'order') Then
    list.addAll(FOrderList);
  if (child_name = 'modalityList') Then
    list.addAll(FModalityListList);
  if (child_name = 'referrer') Then
     list.add(FReferrer.Link);
  if (child_name = 'availability') Then
     list.add(FAvailability.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'numberOfSeries') Then
     list.add(FNumberOfSeries.Link);
  if (child_name = 'numberOfInstances') Then
     list.add(FNumberOfInstances.Link);
  if (child_name = 'procedure') Then
    list.addAll(FProcedure_List);
  if (child_name = 'interpreter') Then
     list.add(FInterpreter.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'series') Then
    list.addAll(FSeriesList);
end;

procedure TFhirImagingStudy.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'started', 'dateTime', FStarted.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uid', 'oid', FUid.Link));{2}
  oList.add(TFHIRProperty.create(self, 'accession', 'Identifier', FAccession.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'order', 'Reference(DiagnosticOrder)', FOrderList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'modalityList', 'Coding', FModalityListList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'referrer', 'Reference(Practitioner)', FReferrer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'availability', 'code', FAvailability.Link));{1}
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'numberOfSeries', 'unsignedInt', FNumberOfSeries.Link));{2}
  oList.add(TFHIRProperty.create(self, 'numberOfInstances', 'unsignedInt', FNumberOfInstances.Link));{2}
  oList.add(TFHIRProperty.create(self, 'procedure', 'Reference(Procedure)', FProcedure_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'interpreter', 'Reference(Practitioner)', FInterpreter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'series', '', FSeriesList.Link)){3};
end;

procedure TFhirImagingStudy.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'started') then StartedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'uid') then UidElement := propValue as TFhirOid{5a}
  else if (propName = 'accession') then Accession := propValue as TFhirIdentifier{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'order') then OrderList.add(propValue as TFhirReference{TFhirDiagnosticOrder}){2}
  else if (propName = 'modalityList') then ModalityListList.add(propValue as TFhirCoding){2}
  else if (propName = 'referrer') then Referrer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'availability') then AvailabilityElement := propValue as TFHIREnum
  else if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'numberOfSeries') then NumberOfSeriesElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'numberOfInstances') then NumberOfInstancesElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'procedure') then Procedure_List.add(propValue as TFhirReference{TFhirProcedure}){2}
  else if (propName = 'interpreter') then Interpreter := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'series') then SeriesList.add(propValue as TFhirImagingStudySeries){2}
  else inherited;
end;

function TFhirImagingStudy.FhirType : string;
begin
  result := 'ImagingStudy';
end;

function TFhirImagingStudy.Link : TFhirImagingStudy;
begin
  result := TFhirImagingStudy(inherited Link);
end;

function TFhirImagingStudy.Clone : TFhirImagingStudy;
begin
  result := TFhirImagingStudy(inherited Clone);
end;

{ TFhirImagingStudy }

Procedure TFhirImagingStudy.SetStarted(value : TFhirDateTime);
begin
  FStarted.free;
  FStarted := value;
end;

Function TFhirImagingStudy.GetStartedST : TDateAndTime;
begin
  if FStarted = nil then
    result := nil
  else
    result := FStarted.value;
end;

Procedure TFhirImagingStudy.SetStartedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStarted = nil then
      FStarted := TFhirDateTime.create;
    FStarted.value := value
  end
  else if FStarted <> nil then
    FStarted.value := nil;
end;

Procedure TFhirImagingStudy.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirImagingStudy.SetUid(value : TFhirOid);
begin
  FUid.free;
  FUid := value;
end;

Function TFhirImagingStudy.GetUidST : String;
begin
  if FUid = nil then
    result := ''
  else
    result := FUid.value;
end;

Procedure TFhirImagingStudy.SetUidST(value : String);
begin
  if value <> '' then
  begin
    if FUid = nil then
      FUid := TFhirOid.create;
    FUid.value := value
  end
  else if FUid <> nil then
    FUid.value := '';
end;

Procedure TFhirImagingStudy.SetAccession(value : TFhirIdentifier);
begin
  FAccession.free;
  FAccession := value;
end;

Function TFhirImagingStudy.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirImagingStudy.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirImagingStudy.GetOrderList : TFhirReferenceList{TFhirDiagnosticOrder};
begin
  if FOrderList = nil then
    FOrderList := TFhirReferenceList{TFhirDiagnosticOrder}.Create;
  result := FOrderList;
end;

Function TFhirImagingStudy.GetHasOrderList : boolean;
begin
  result := (FOrderList <> nil) and (FOrderList.count > 0);
end;

Function TFhirImagingStudy.GetModalityListList : TFhirCodingList;
begin
  if FModalityListList = nil then
    FModalityListList := TFhirCodingList.Create;
  result := FModalityListList;
end;

Function TFhirImagingStudy.GetHasModalityListList : boolean;
begin
  result := (FModalityListList <> nil) and (FModalityListList.count > 0);
end;

Procedure TFhirImagingStudy.SetReferrer(value : TFhirReference{TFhirPractitioner});
begin
  FReferrer.free;
  FReferrer := value;
end;

Procedure TFhirImagingStudy.SetAvailability(value : TFhirEnum);
begin
  FAvailability.free;
  FAvailability := value;
end;

Function TFhirImagingStudy.GetAvailabilityST : TFhirInstanceAvailability;
begin
  if FAvailability = nil then
    result := TFhirInstanceAvailability(0)
  else
    result := TFhirInstanceAvailability(StringArrayIndexOfSensitive(CODES_TFhirInstanceAvailability, FAvailability.value));
end;

Procedure TFhirImagingStudy.SetAvailabilityST(value : TFhirInstanceAvailability);
begin
  if ord(value) = 0 then
    AvailabilityElement := nil
  else
    AvailabilityElement := TFhirEnum.create(CODES_TFhirInstanceAvailability[value]);
end;

Procedure TFhirImagingStudy.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImagingStudy.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImagingStudy.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirImagingStudy.SetNumberOfSeries(value : TFhirUnsignedInt);
begin
  FNumberOfSeries.free;
  FNumberOfSeries := value;
end;

Function TFhirImagingStudy.GetNumberOfSeriesST : String;
begin
  if FNumberOfSeries = nil then
    result := ''
  else
    result := FNumberOfSeries.value;
end;

Procedure TFhirImagingStudy.SetNumberOfSeriesST(value : String);
begin
  if value <> '' then
  begin
    if FNumberOfSeries = nil then
      FNumberOfSeries := TFhirUnsignedInt.create;
    FNumberOfSeries.value := value
  end
  else if FNumberOfSeries <> nil then
    FNumberOfSeries.value := '';
end;

Procedure TFhirImagingStudy.SetNumberOfInstances(value : TFhirUnsignedInt);
begin
  FNumberOfInstances.free;
  FNumberOfInstances := value;
end;

Function TFhirImagingStudy.GetNumberOfInstancesST : String;
begin
  if FNumberOfInstances = nil then
    result := ''
  else
    result := FNumberOfInstances.value;
end;

Procedure TFhirImagingStudy.SetNumberOfInstancesST(value : String);
begin
  if value <> '' then
  begin
    if FNumberOfInstances = nil then
      FNumberOfInstances := TFhirUnsignedInt.create;
    FNumberOfInstances.value := value
  end
  else if FNumberOfInstances <> nil then
    FNumberOfInstances.value := '';
end;

Function TFhirImagingStudy.GetProcedure_List : TFhirReferenceList{TFhirProcedure};
begin
  if FProcedure_List = nil then
    FProcedure_List := TFhirReferenceList{TFhirProcedure}.Create;
  result := FProcedure_List;
end;

Function TFhirImagingStudy.GetHasProcedure_List : boolean;
begin
  result := (FProcedure_List <> nil) and (FProcedure_List.count > 0);
end;

Procedure TFhirImagingStudy.SetInterpreter(value : TFhirReference{TFhirPractitioner});
begin
  FInterpreter.free;
  FInterpreter := value;
end;

Procedure TFhirImagingStudy.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImagingStudy.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImagingStudy.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirImagingStudy.GetSeriesList : TFhirImagingStudySeriesList;
begin
  if FSeriesList = nil then
    FSeriesList := TFhirImagingStudySeriesList.Create;
  result := FSeriesList;
end;

Function TFhirImagingStudy.GetHasSeriesList : boolean;
begin
  result := (FSeriesList <> nil) and (FSeriesList.count > 0);
end;


{ TFhirImagingStudyListEnumerator }

Constructor TFhirImagingStudyListEnumerator.Create(list : TFhirImagingStudyList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImagingStudyListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImagingStudyListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImagingStudyListEnumerator.GetCurrent : TFhirImagingStudy;
begin
  Result := FList[FIndex];
end;


{ TFhirImagingStudyList }
procedure TFhirImagingStudyList.AddItem(value: TFhirImagingStudy);
begin
  assert(value.ClassName = 'TFhirImagingStudy', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImagingStudy');
  add(value);
end;


function TFhirImagingStudyList.Append: TFhirImagingStudy;
begin
  result := TFhirImagingStudy.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudyList.ClearItems;
begin
  Clear;
end;

function TFhirImagingStudyList.GetEnumerator : TFhirImagingStudyListEnumerator;
begin
  result := TFhirImagingStudyListEnumerator.Create(self.link);
end;

function TFhirImagingStudyList.Clone: TFhirImagingStudyList;
begin
  result := TFhirImagingStudyList(inherited Clone);
end;

function TFhirImagingStudyList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImagingStudyList.GetItemN(index: Integer): TFhirImagingStudy;
begin
  result := TFhirImagingStudy(ObjectByIndex[index]);
end;

function TFhirImagingStudyList.IndexOf(value: TFhirImagingStudy): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImagingStudyList.Insert(index: Integer): TFhirImagingStudy;
begin
  result := TFhirImagingStudy.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImagingStudyList.InsertItem(index: Integer; value: TFhirImagingStudy);
begin
  assert(value is TFhirImagingStudy);
  Inherited Insert(index, value);
end;

function TFhirImagingStudyList.Item(index: Integer): TFhirImagingStudy;
begin
  result := TFhirImagingStudy(ObjectByIndex[index]);
end;

function TFhirImagingStudyList.Link: TFhirImagingStudyList;
begin
  result := TFhirImagingStudyList(inherited Link);
end;

procedure TFhirImagingStudyList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImagingStudyList.SetItemByIndex(index: Integer; value: TFhirImagingStudy);
begin
  assert(value is TFhirImagingStudy);
  FhirImagingStudies[index] := value;
end;

procedure TFhirImagingStudyList.SetItemN(index: Integer; value: TFhirImagingStudy);
begin
  assert(value is TFhirImagingStudy);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationExplanation }

constructor TFhirImmunizationExplanation.Create;
begin
  inherited;
end;

destructor TFhirImmunizationExplanation.Destroy;
begin
  FReasonList.Free;
  FReasonNotGivenList.Free;
  inherited;
end;

procedure TFhirImmunizationExplanation.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirImmunizationExplanation(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirImmunizationExplanation(oSource).FReasonList);
  end;
  if (TFhirImmunizationExplanation(oSource).FReasonNotGivenList = nil) then
  begin
    FReasonNotGivenList.free;
    FReasonNotGivenList := nil;
  end
  else
  begin
    FReasonNotGivenList := TFhirCodeableConceptList.Create;
    FReasonNotGivenList.Assign(TFhirImmunizationExplanation(oSource).FReasonNotGivenList);
  end;
end;

procedure TFhirImmunizationExplanation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'reasonNotGiven') Then
    list.addAll(FReasonNotGivenList);
end;

procedure TFhirImmunizationExplanation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reasonNotGiven', 'CodeableConcept', FReasonNotGivenList.Link)){3};
end;

procedure TFhirImmunizationExplanation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'reasonNotGiven') then ReasonNotGivenList.add(propValue as TFhirCodeableConcept){2}
  else inherited;
end;

function TFhirImmunizationExplanation.FhirType : string;
begin
  result := 'explanation';
end;

function TFhirImmunizationExplanation.Link : TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation(inherited Link);
end;

function TFhirImmunizationExplanation.Clone : TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation(inherited Clone);
end;

function TFhirImmunizationExplanation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationExplanation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationExplanation)) then
    result := false
  else
  begin
    o := TFhirImmunizationExplanation(other);
    result := compareDeep(reasonList, o.reasonList, true) and compareDeep(reasonNotGivenList, o.reasonNotGivenList, true);
  end;
end;

function TFhirImmunizationExplanation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationExplanation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationExplanation)) then
    result := false
  else
  begin
    o := TFhirImmunizationExplanation(other);
    result := true;
  end;
end;

{ TFhirImmunizationExplanation }

Function TFhirImmunizationExplanation.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirImmunizationExplanation.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Function TFhirImmunizationExplanation.GetReasonNotGivenList : TFhirCodeableConceptList;
begin
  if FReasonNotGivenList = nil then
    FReasonNotGivenList := TFhirCodeableConceptList.Create;
  result := FReasonNotGivenList;
end;

Function TFhirImmunizationExplanation.GetHasReasonNotGivenList : boolean;
begin
  result := (FReasonNotGivenList <> nil) and (FReasonNotGivenList.count > 0);
end;


{ TFhirImmunizationExplanationListEnumerator }

Constructor TFhirImmunizationExplanationListEnumerator.Create(list : TFhirImmunizationExplanationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationExplanationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationExplanationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationExplanationListEnumerator.GetCurrent : TFhirImmunizationExplanation;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationExplanationList }
procedure TFhirImmunizationExplanationList.AddItem(value: TFhirImmunizationExplanation);
begin
  assert(value.ClassName = 'TFhirImmunizationExplanation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationExplanation');
  add(value);
end;


function TFhirImmunizationExplanationList.Append: TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationExplanationList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationExplanationList.GetEnumerator : TFhirImmunizationExplanationListEnumerator;
begin
  result := TFhirImmunizationExplanationListEnumerator.Create(self.link);
end;

function TFhirImmunizationExplanationList.Clone: TFhirImmunizationExplanationList;
begin
  result := TFhirImmunizationExplanationList(inherited Clone);
end;

function TFhirImmunizationExplanationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationExplanationList.GetItemN(index: Integer): TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation(ObjectByIndex[index]);
end;

function TFhirImmunizationExplanationList.IndexOf(value: TFhirImmunizationExplanation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationExplanationList.Insert(index: Integer): TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationExplanationList.InsertItem(index: Integer; value: TFhirImmunizationExplanation);
begin
  assert(value is TFhirImmunizationExplanation);
  Inherited Insert(index, value);
end;

function TFhirImmunizationExplanationList.Item(index: Integer): TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation(ObjectByIndex[index]);
end;

function TFhirImmunizationExplanationList.Link: TFhirImmunizationExplanationList;
begin
  result := TFhirImmunizationExplanationList(inherited Link);
end;

procedure TFhirImmunizationExplanationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationExplanationList.SetItemByIndex(index: Integer; value: TFhirImmunizationExplanation);
begin
  assert(value is TFhirImmunizationExplanation);
  FhirImmunizationExplanations[index] := value;
end;

procedure TFhirImmunizationExplanationList.SetItemN(index: Integer; value: TFhirImmunizationExplanation);
begin
  assert(value is TFhirImmunizationExplanation);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationReaction }

constructor TFhirImmunizationReaction.Create;
begin
  inherited;
end;

destructor TFhirImmunizationReaction.Destroy;
begin
  FDate.free;
  FDetail.free;
  FReported.free;
  inherited;
end;

procedure TFhirImmunizationReaction.Assign(oSource : TAdvObject);
begin
  inherited;
  dateElement := TFhirImmunizationReaction(oSource).dateElement.Clone;
  detail := TFhirImmunizationReaction(oSource).detail.Clone;
  reportedElement := TFhirImmunizationReaction(oSource).reportedElement.Clone;
end;

procedure TFhirImmunizationReaction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'detail') Then
     list.add(FDetail.Link);
  if (child_name = 'reported') Then
     list.add(FReported.Link);
end;

procedure TFhirImmunizationReaction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', 'Reference(Observation)', FDetail.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reported', 'boolean', FReported.Link));{2}
end;

procedure TFhirImmunizationReaction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'detail') then Detail := propValue as TFhirReference{TFhirObservation}{4b}
  else if (propName = 'reported') then ReportedElement := propValue as TFhirBoolean{5a}
  else inherited;
end;

function TFhirImmunizationReaction.FhirType : string;
begin
  result := 'reaction';
end;

function TFhirImmunizationReaction.Link : TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction(inherited Link);
end;

function TFhirImmunizationReaction.Clone : TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction(inherited Clone);
end;

function TFhirImmunizationReaction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationReaction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationReaction)) then
    result := false
  else
  begin
    o := TFhirImmunizationReaction(other);
    result := compareDeep(dateElement, o.dateElement, true) and compareDeep(detailElement, o.detailElement, true) and 
      compareDeep(reportedElement, o.reportedElement, true);
  end;
end;

function TFhirImmunizationReaction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationReaction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationReaction)) then
    result := false
  else
  begin
    o := TFhirImmunizationReaction(other);
    result := compareValues(dateElement, o.dateElement, true) and compareValues(reportedElement, o.reportedElement, true);
  end;
end;

{ TFhirImmunizationReaction }

Procedure TFhirImmunizationReaction.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirImmunizationReaction.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirImmunizationReaction.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirImmunizationReaction.SetDetail(value : TFhirReference{TFhirObservation});
begin
  FDetail.free;
  FDetail := value;
end;

Procedure TFhirImmunizationReaction.SetReported(value : TFhirBoolean);
begin
  FReported.free;
  FReported := value;
end;

Function TFhirImmunizationReaction.GetReportedST : Boolean;
begin
  if FReported = nil then
    result := false
  else
    result := FReported.value;
end;

Procedure TFhirImmunizationReaction.SetReportedST(value : Boolean);
begin
  if FReported = nil then
    FReported := TFhirBoolean.create;
  FReported.value := value
end;


{ TFhirImmunizationReactionListEnumerator }

Constructor TFhirImmunizationReactionListEnumerator.Create(list : TFhirImmunizationReactionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationReactionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationReactionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationReactionListEnumerator.GetCurrent : TFhirImmunizationReaction;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationReactionList }
procedure TFhirImmunizationReactionList.AddItem(value: TFhirImmunizationReaction);
begin
  assert(value.ClassName = 'TFhirImmunizationReaction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationReaction');
  add(value);
end;


function TFhirImmunizationReactionList.Append: TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationReactionList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationReactionList.GetEnumerator : TFhirImmunizationReactionListEnumerator;
begin
  result := TFhirImmunizationReactionListEnumerator.Create(self.link);
end;

function TFhirImmunizationReactionList.Clone: TFhirImmunizationReactionList;
begin
  result := TFhirImmunizationReactionList(inherited Clone);
end;

function TFhirImmunizationReactionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationReactionList.GetItemN(index: Integer): TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction(ObjectByIndex[index]);
end;

function TFhirImmunizationReactionList.IndexOf(value: TFhirImmunizationReaction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationReactionList.Insert(index: Integer): TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationReactionList.InsertItem(index: Integer; value: TFhirImmunizationReaction);
begin
  assert(value is TFhirImmunizationReaction);
  Inherited Insert(index, value);
end;

function TFhirImmunizationReactionList.Item(index: Integer): TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction(ObjectByIndex[index]);
end;

function TFhirImmunizationReactionList.Link: TFhirImmunizationReactionList;
begin
  result := TFhirImmunizationReactionList(inherited Link);
end;

procedure TFhirImmunizationReactionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationReactionList.SetItemByIndex(index: Integer; value: TFhirImmunizationReaction);
begin
  assert(value is TFhirImmunizationReaction);
  FhirImmunizationReactions[index] := value;
end;

procedure TFhirImmunizationReactionList.SetItemN(index: Integer; value: TFhirImmunizationReaction);
begin
  assert(value is TFhirImmunizationReaction);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationVaccinationProtocol }

constructor TFhirImmunizationVaccinationProtocol.Create;
begin
  inherited;
end;

destructor TFhirImmunizationVaccinationProtocol.Destroy;
begin
  FDoseSequence.free;
  FDescription.free;
  FAuthority.free;
  FSeries.free;
  FSeriesDoses.free;
  FTargetDiseaseList.Free;
  FDoseStatus.free;
  FDoseStatusReason.free;
  inherited;
end;

procedure TFhirImmunizationVaccinationProtocol.Assign(oSource : TAdvObject);
begin
  inherited;
  doseSequenceElement := TFhirImmunizationVaccinationProtocol(oSource).doseSequenceElement.Clone;
  descriptionElement := TFhirImmunizationVaccinationProtocol(oSource).descriptionElement.Clone;
  authority := TFhirImmunizationVaccinationProtocol(oSource).authority.Clone;
  seriesElement := TFhirImmunizationVaccinationProtocol(oSource).seriesElement.Clone;
  seriesDosesElement := TFhirImmunizationVaccinationProtocol(oSource).seriesDosesElement.Clone;
  if (TFhirImmunizationVaccinationProtocol(oSource).FTargetDiseaseList = nil) then
  begin
    FTargetDiseaseList.free;
    FTargetDiseaseList := nil;
  end
  else
  begin
    FTargetDiseaseList := TFhirCodeableConceptList.Create;
    FTargetDiseaseList.Assign(TFhirImmunizationVaccinationProtocol(oSource).FTargetDiseaseList);
  end;
  doseStatus := TFhirImmunizationVaccinationProtocol(oSource).doseStatus.Clone;
  doseStatusReason := TFhirImmunizationVaccinationProtocol(oSource).doseStatusReason.Clone;
end;

procedure TFhirImmunizationVaccinationProtocol.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'doseSequence') Then
     list.add(FDoseSequence.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'authority') Then
     list.add(FAuthority.Link);
  if (child_name = 'series') Then
     list.add(FSeries.Link);
  if (child_name = 'seriesDoses') Then
     list.add(FSeriesDoses.Link);
  if (child_name = 'targetDisease') Then
    list.addAll(FTargetDiseaseList);
  if (child_name = 'doseStatus') Then
     list.add(FDoseStatus.Link);
  if (child_name = 'doseStatusReason') Then
     list.add(FDoseStatusReason.Link);
end;

procedure TFhirImmunizationVaccinationProtocol.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'doseSequence', 'positiveInt', FDoseSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authority', 'Reference(Organization)', FAuthority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'series', 'string', FSeries.Link));{2}
  oList.add(TFHIRProperty.create(self, 'seriesDoses', 'positiveInt', FSeriesDoses.Link));{2}
  oList.add(TFHIRProperty.create(self, 'targetDisease', 'CodeableConcept', FTargetDiseaseList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'doseStatus', 'CodeableConcept', FDoseStatus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'doseStatusReason', 'CodeableConcept', FDoseStatusReason.Link));{2}
end;

procedure TFhirImmunizationVaccinationProtocol.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'doseSequence') then DoseSequenceElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'authority') then Authority := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'series') then SeriesElement := propValue as TFhirString{5a}
  else if (propName = 'seriesDoses') then SeriesDosesElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'targetDisease') then TargetDiseaseList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'doseStatus') then DoseStatus := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'doseStatusReason') then DoseStatusReason := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirImmunizationVaccinationProtocol.FhirType : string;
begin
  result := 'vaccinationProtocol';
end;

function TFhirImmunizationVaccinationProtocol.Link : TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol(inherited Link);
end;

function TFhirImmunizationVaccinationProtocol.Clone : TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol(inherited Clone);
end;

function TFhirImmunizationVaccinationProtocol.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationVaccinationProtocol;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationVaccinationProtocol)) then
    result := false
  else
  begin
    o := TFhirImmunizationVaccinationProtocol(other);
    result := compareDeep(doseSequenceElement, o.doseSequenceElement, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(authorityElement, o.authorityElement, true) and 
      compareDeep(seriesElement, o.seriesElement, true) and compareDeep(seriesDosesElement, o.seriesDosesElement, true) and 
      compareDeep(targetDiseaseList, o.targetDiseaseList, true) and compareDeep(doseStatusElement, o.doseStatusElement, true) and 
      compareDeep(doseStatusReasonElement, o.doseStatusReasonElement, true);
  end;
end;

function TFhirImmunizationVaccinationProtocol.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationVaccinationProtocol;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationVaccinationProtocol)) then
    result := false
  else
  begin
    o := TFhirImmunizationVaccinationProtocol(other);
    result := compareValues(doseSequenceElement, o.doseSequenceElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(seriesElement, o.seriesElement, true) and 
      compareValues(seriesDosesElement, o.seriesDosesElement, true);
  end;
end;

{ TFhirImmunizationVaccinationProtocol }

Procedure TFhirImmunizationVaccinationProtocol.SetDoseSequence(value : TFhirPositiveInt);
begin
  FDoseSequence.free;
  FDoseSequence := value;
end;

Function TFhirImmunizationVaccinationProtocol.GetDoseSequenceST : String;
begin
  if FDoseSequence = nil then
    result := ''
  else
    result := FDoseSequence.value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetDoseSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FDoseSequence = nil then
      FDoseSequence := TFhirPositiveInt.create;
    FDoseSequence.value := value
  end
  else if FDoseSequence <> nil then
    FDoseSequence.value := '';
end;

Procedure TFhirImmunizationVaccinationProtocol.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImmunizationVaccinationProtocol.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirImmunizationVaccinationProtocol.SetAuthority(value : TFhirReference{TFhirOrganization});
begin
  FAuthority.free;
  FAuthority := value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetSeries(value : TFhirString);
begin
  FSeries.free;
  FSeries := value;
end;

Function TFhirImmunizationVaccinationProtocol.GetSeriesST : String;
begin
  if FSeries = nil then
    result := ''
  else
    result := FSeries.value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetSeriesST(value : String);
begin
  if value <> '' then
  begin
    if FSeries = nil then
      FSeries := TFhirString.create;
    FSeries.value := value
  end
  else if FSeries <> nil then
    FSeries.value := '';
end;

Procedure TFhirImmunizationVaccinationProtocol.SetSeriesDoses(value : TFhirPositiveInt);
begin
  FSeriesDoses.free;
  FSeriesDoses := value;
end;

Function TFhirImmunizationVaccinationProtocol.GetSeriesDosesST : String;
begin
  if FSeriesDoses = nil then
    result := ''
  else
    result := FSeriesDoses.value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetSeriesDosesST(value : String);
begin
  if value <> '' then
  begin
    if FSeriesDoses = nil then
      FSeriesDoses := TFhirPositiveInt.create;
    FSeriesDoses.value := value
  end
  else if FSeriesDoses <> nil then
    FSeriesDoses.value := '';
end;

Function TFhirImmunizationVaccinationProtocol.GetTargetDiseaseList : TFhirCodeableConceptList;
begin
  if FTargetDiseaseList = nil then
    FTargetDiseaseList := TFhirCodeableConceptList.Create;
  result := FTargetDiseaseList;
end;

Function TFhirImmunizationVaccinationProtocol.GetHasTargetDiseaseList : boolean;
begin
  result := (FTargetDiseaseList <> nil) and (FTargetDiseaseList.count > 0);
end;

Procedure TFhirImmunizationVaccinationProtocol.SetDoseStatus(value : TFhirCodeableConcept);
begin
  FDoseStatus.free;
  FDoseStatus := value;
end;

Procedure TFhirImmunizationVaccinationProtocol.SetDoseStatusReason(value : TFhirCodeableConcept);
begin
  FDoseStatusReason.free;
  FDoseStatusReason := value;
end;


{ TFhirImmunizationVaccinationProtocolListEnumerator }

Constructor TFhirImmunizationVaccinationProtocolListEnumerator.Create(list : TFhirImmunizationVaccinationProtocolList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationVaccinationProtocolListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationVaccinationProtocolListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationVaccinationProtocolListEnumerator.GetCurrent : TFhirImmunizationVaccinationProtocol;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationVaccinationProtocolList }
procedure TFhirImmunizationVaccinationProtocolList.AddItem(value: TFhirImmunizationVaccinationProtocol);
begin
  assert(value.ClassName = 'TFhirImmunizationVaccinationProtocol', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationVaccinationProtocol');
  add(value);
end;


function TFhirImmunizationVaccinationProtocolList.Append: TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationVaccinationProtocolList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationVaccinationProtocolList.GetEnumerator : TFhirImmunizationVaccinationProtocolListEnumerator;
begin
  result := TFhirImmunizationVaccinationProtocolListEnumerator.Create(self.link);
end;

function TFhirImmunizationVaccinationProtocolList.Clone: TFhirImmunizationVaccinationProtocolList;
begin
  result := TFhirImmunizationVaccinationProtocolList(inherited Clone);
end;

function TFhirImmunizationVaccinationProtocolList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationVaccinationProtocolList.GetItemN(index: Integer): TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol(ObjectByIndex[index]);
end;

function TFhirImmunizationVaccinationProtocolList.IndexOf(value: TFhirImmunizationVaccinationProtocol): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationVaccinationProtocolList.Insert(index: Integer): TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationVaccinationProtocolList.InsertItem(index: Integer; value: TFhirImmunizationVaccinationProtocol);
begin
  assert(value is TFhirImmunizationVaccinationProtocol);
  Inherited Insert(index, value);
end;

function TFhirImmunizationVaccinationProtocolList.Item(index: Integer): TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol(ObjectByIndex[index]);
end;

function TFhirImmunizationVaccinationProtocolList.Link: TFhirImmunizationVaccinationProtocolList;
begin
  result := TFhirImmunizationVaccinationProtocolList(inherited Link);
end;

procedure TFhirImmunizationVaccinationProtocolList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationVaccinationProtocolList.SetItemByIndex(index: Integer; value: TFhirImmunizationVaccinationProtocol);
begin
  assert(value is TFhirImmunizationVaccinationProtocol);
  FhirImmunizationVaccinationProtocols[index] := value;
end;

procedure TFhirImmunizationVaccinationProtocolList.SetItemN(index: Integer; value: TFhirImmunizationVaccinationProtocol);
begin
  assert(value is TFhirImmunizationVaccinationProtocol);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunization }

constructor TFhirImmunization.Create;
begin
  inherited;
end;

destructor TFhirImmunization.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FDate.free;
  FVaccineCode.free;
  FPatient.free;
  FWasNotGiven.free;
  FReported.free;
  FPerformer.free;
  FRequester.free;
  FEncounter.free;
  FManufacturer.free;
  FLocation.free;
  FLotNumber.free;
  FExpirationDate.free;
  FSite.free;
  FRoute.free;
  FDoseQuantity.free;
  FNoteList.Free;
  FExplanation.free;
  FReactionList.Free;
  FVaccinationProtocolList.Free;
  inherited;
end;

function TFhirImmunization.GetResourceType : TFhirResourceType;
begin
  result := frtImmunization;
end;

procedure TFhirImmunization.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirImmunization(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirImmunization(oSource).FIdentifierList);
  end;
  FStatus := TFhirImmunization(oSource).FStatus.Link;
  dateElement := TFhirImmunization(oSource).dateElement.Clone;
  vaccineCode := TFhirImmunization(oSource).vaccineCode.Clone;
  patient := TFhirImmunization(oSource).patient.Clone;
  wasNotGivenElement := TFhirImmunization(oSource).wasNotGivenElement.Clone;
  reportedElement := TFhirImmunization(oSource).reportedElement.Clone;
  performer := TFhirImmunization(oSource).performer.Clone;
  requester := TFhirImmunization(oSource).requester.Clone;
  encounter := TFhirImmunization(oSource).encounter.Clone;
  manufacturer := TFhirImmunization(oSource).manufacturer.Clone;
  location := TFhirImmunization(oSource).location.Clone;
  lotNumberElement := TFhirImmunization(oSource).lotNumberElement.Clone;
  expirationDateElement := TFhirImmunization(oSource).expirationDateElement.Clone;
  site := TFhirImmunization(oSource).site.Clone;
  route := TFhirImmunization(oSource).route.Clone;
  doseQuantity := TFhirImmunization(oSource).doseQuantity.Clone;
  if (TFhirImmunization(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirAnnotationList.Create;
    FNoteList.Assign(TFhirImmunization(oSource).FNoteList);
  end;
  explanation := TFhirImmunization(oSource).explanation.Clone;
  if (TFhirImmunization(oSource).FReactionList = nil) then
  begin
    FReactionList.free;
    FReactionList := nil;
  end
  else
  begin
    FReactionList := TFhirImmunizationReactionList.Create;
    FReactionList.Assign(TFhirImmunization(oSource).FReactionList);
  end;
  if (TFhirImmunization(oSource).FVaccinationProtocolList = nil) then
  begin
    FVaccinationProtocolList.free;
    FVaccinationProtocolList := nil;
  end
  else
  begin
    FVaccinationProtocolList := TFhirImmunizationVaccinationProtocolList.Create;
    FVaccinationProtocolList.Assign(TFhirImmunization(oSource).FVaccinationProtocolList);
  end;
end;

procedure TFhirImmunization.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'vaccineCode') Then
     list.add(FVaccineCode.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'wasNotGiven') Then
     list.add(FWasNotGiven.Link);
  if (child_name = 'reported') Then
     list.add(FReported.Link);
  if (child_name = 'performer') Then
     list.add(FPerformer.Link);
  if (child_name = 'requester') Then
     list.add(FRequester.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'manufacturer') Then
     list.add(FManufacturer.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'lotNumber') Then
     list.add(FLotNumber.Link);
  if (child_name = 'expirationDate') Then
     list.add(FExpirationDate.Link);
  if (child_name = 'site') Then
     list.add(FSite.Link);
  if (child_name = 'route') Then
     list.add(FRoute.Link);
  if (child_name = 'doseQuantity') Then
     list.add(FDoseQuantity.Link);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
  if (child_name = 'explanation') Then
     list.add(FExplanation.Link);
  if (child_name = 'reaction') Then
    list.addAll(FReactionList);
  if (child_name = 'vaccinationProtocol') Then
    list.addAll(FVaccinationProtocolList);
end;

procedure TFhirImmunization.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'vaccineCode', 'CodeableConcept', FVaccineCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'wasNotGiven', 'boolean', FWasNotGiven.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reported', 'boolean', FReported.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner)', FPerformer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requester', 'Reference(Practitioner)', FRequester.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'manufacturer', 'Reference(Organization)', FManufacturer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lotNumber', 'string', FLotNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expirationDate', 'date', FExpirationDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'site', 'CodeableConcept', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'route', 'CodeableConcept', FRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'doseQuantity', 'SimpleQuantity', FDoseQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'Annotation', FNoteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'explanation', '', FExplanation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reaction', '', FReactionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'vaccinationProtocol', '', FVaccinationProtocolList.Link)){3};
end;

procedure TFhirImmunization.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'vaccineCode') then VaccineCode := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'wasNotGiven') then WasNotGivenElement := propValue as TFhirBoolean{5a}
  else if (propName = 'reported') then ReportedElement := propValue as TFhirBoolean{5a}
  else if (propName = 'performer') then Performer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requester') then Requester := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'manufacturer') then Manufacturer := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'lotNumber') then LotNumberElement := propValue as TFhirString{5a}
  else if (propName = 'expirationDate') then ExpirationDateElement := propValue as TFhirDate{5a}
  else if (propName = 'site') then Site := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'route') then Route := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'doseQuantity') then DoseQuantity := propValue as TFhirQuantity{4b}
  else if (propName = 'note') then NoteList.add(propValue as TFhirAnnotation){2}
  else if (propName = 'explanation') then Explanation := propValue as TFhirImmunizationExplanation{4b}
  else if (propName = 'reaction') then ReactionList.add(propValue as TFhirImmunizationReaction){2}
  else if (propName = 'vaccinationProtocol') then VaccinationProtocolList.add(propValue as TFhirImmunizationVaccinationProtocol){2}
  else inherited;
end;

function TFhirImmunization.FhirType : string;
begin
  result := 'Immunization';
end;

function TFhirImmunization.Link : TFhirImmunization;
begin
  result := TFhirImmunization(inherited Link);
end;

function TFhirImmunization.Clone : TFhirImmunization;
begin
  result := TFhirImmunization(inherited Clone);
end;

{ TFhirImmunization }

Function TFhirImmunization.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirImmunization.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirImmunization.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirImmunization.GetStatusST : TFhirMedicationAdminStatus;
begin
  if FStatus = nil then
    result := TFhirMedicationAdminStatus(0)
  else
    result := TFhirMedicationAdminStatus(StringArrayIndexOfSensitive(CODES_TFhirMedicationAdminStatus, FStatus.value));
end;

Procedure TFhirImmunization.SetStatusST(value : TFhirMedicationAdminStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirMedicationAdminStatus[value]);
end;

Procedure TFhirImmunization.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirImmunization.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirImmunization.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirImmunization.SetVaccineCode(value : TFhirCodeableConcept);
begin
  FVaccineCode.free;
  FVaccineCode := value;
end;

Procedure TFhirImmunization.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirImmunization.SetWasNotGiven(value : TFhirBoolean);
begin
  FWasNotGiven.free;
  FWasNotGiven := value;
end;

Function TFhirImmunization.GetWasNotGivenST : Boolean;
begin
  if FWasNotGiven = nil then
    result := false
  else
    result := FWasNotGiven.value;
end;

Procedure TFhirImmunization.SetWasNotGivenST(value : Boolean);
begin
  if FWasNotGiven = nil then
    FWasNotGiven := TFhirBoolean.create;
  FWasNotGiven.value := value
end;

Procedure TFhirImmunization.SetReported(value : TFhirBoolean);
begin
  FReported.free;
  FReported := value;
end;

Function TFhirImmunization.GetReportedST : Boolean;
begin
  if FReported = nil then
    result := false
  else
    result := FReported.value;
end;

Procedure TFhirImmunization.SetReportedST(value : Boolean);
begin
  if FReported = nil then
    FReported := TFhirBoolean.create;
  FReported.value := value
end;

Procedure TFhirImmunization.SetPerformer(value : TFhirReference{TFhirPractitioner});
begin
  FPerformer.free;
  FPerformer := value;
end;

Procedure TFhirImmunization.SetRequester(value : TFhirReference{TFhirPractitioner});
begin
  FRequester.free;
  FRequester := value;
end;

Procedure TFhirImmunization.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirImmunization.SetManufacturer(value : TFhirReference{TFhirOrganization});
begin
  FManufacturer.free;
  FManufacturer := value;
end;

Procedure TFhirImmunization.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Procedure TFhirImmunization.SetLotNumber(value : TFhirString);
begin
  FLotNumber.free;
  FLotNumber := value;
end;

Function TFhirImmunization.GetLotNumberST : String;
begin
  if FLotNumber = nil then
    result := ''
  else
    result := FLotNumber.value;
end;

Procedure TFhirImmunization.SetLotNumberST(value : String);
begin
  if value <> '' then
  begin
    if FLotNumber = nil then
      FLotNumber := TFhirString.create;
    FLotNumber.value := value
  end
  else if FLotNumber <> nil then
    FLotNumber.value := '';
end;

Procedure TFhirImmunization.SetExpirationDate(value : TFhirDate);
begin
  FExpirationDate.free;
  FExpirationDate := value;
end;

Function TFhirImmunization.GetExpirationDateST : TDateAndTime;
begin
  if FExpirationDate = nil then
    result := nil
  else
    result := FExpirationDate.value;
end;

Procedure TFhirImmunization.SetExpirationDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FExpirationDate = nil then
      FExpirationDate := TFhirDate.create;
    FExpirationDate.value := value
  end
  else if FExpirationDate <> nil then
    FExpirationDate.value := nil;
end;

Procedure TFhirImmunization.SetSite(value : TFhirCodeableConcept);
begin
  FSite.free;
  FSite := value;
end;

Procedure TFhirImmunization.SetRoute(value : TFhirCodeableConcept);
begin
  FRoute.free;
  FRoute := value;
end;

Procedure TFhirImmunization.SetDoseQuantity(value : TFhirQuantity);
begin
  FDoseQuantity.free;
  FDoseQuantity := value;
end;

Function TFhirImmunization.GetNoteList : TFhirAnnotationList;
begin
  if FNoteList = nil then
    FNoteList := TFhirAnnotationList.Create;
  result := FNoteList;
end;

Function TFhirImmunization.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;

Procedure TFhirImmunization.SetExplanation(value : TFhirImmunizationExplanation);
begin
  FExplanation.free;
  FExplanation := value;
end;

Function TFhirImmunization.GetReactionList : TFhirImmunizationReactionList;
begin
  if FReactionList = nil then
    FReactionList := TFhirImmunizationReactionList.Create;
  result := FReactionList;
end;

Function TFhirImmunization.GetHasReactionList : boolean;
begin
  result := (FReactionList <> nil) and (FReactionList.count > 0);
end;

Function TFhirImmunization.GetVaccinationProtocolList : TFhirImmunizationVaccinationProtocolList;
begin
  if FVaccinationProtocolList = nil then
    FVaccinationProtocolList := TFhirImmunizationVaccinationProtocolList.Create;
  result := FVaccinationProtocolList;
end;

Function TFhirImmunization.GetHasVaccinationProtocolList : boolean;
begin
  result := (FVaccinationProtocolList <> nil) and (FVaccinationProtocolList.count > 0);
end;


{ TFhirImmunizationListEnumerator }

Constructor TFhirImmunizationListEnumerator.Create(list : TFhirImmunizationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationListEnumerator.GetCurrent : TFhirImmunization;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationList }
procedure TFhirImmunizationList.AddItem(value: TFhirImmunization);
begin
  assert(value.ClassName = 'TFhirImmunization', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunization');
  add(value);
end;


function TFhirImmunizationList.Append: TFhirImmunization;
begin
  result := TFhirImmunization.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationList.GetEnumerator : TFhirImmunizationListEnumerator;
begin
  result := TFhirImmunizationListEnumerator.Create(self.link);
end;

function TFhirImmunizationList.Clone: TFhirImmunizationList;
begin
  result := TFhirImmunizationList(inherited Clone);
end;

function TFhirImmunizationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationList.GetItemN(index: Integer): TFhirImmunization;
begin
  result := TFhirImmunization(ObjectByIndex[index]);
end;

function TFhirImmunizationList.IndexOf(value: TFhirImmunization): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationList.Insert(index: Integer): TFhirImmunization;
begin
  result := TFhirImmunization.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationList.InsertItem(index: Integer; value: TFhirImmunization);
begin
  assert(value is TFhirImmunization);
  Inherited Insert(index, value);
end;

function TFhirImmunizationList.Item(index: Integer): TFhirImmunization;
begin
  result := TFhirImmunization(ObjectByIndex[index]);
end;

function TFhirImmunizationList.Link: TFhirImmunizationList;
begin
  result := TFhirImmunizationList(inherited Link);
end;

procedure TFhirImmunizationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationList.SetItemByIndex(index: Integer; value: TFhirImmunization);
begin
  assert(value is TFhirImmunization);
  FhirImmunizations[index] := value;
end;

procedure TFhirImmunizationList.SetItemN(index: Integer; value: TFhirImmunization);
begin
  assert(value is TFhirImmunization);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationRecommendationRecommendation }

constructor TFhirImmunizationRecommendationRecommendation.Create;
begin
  inherited;
end;

destructor TFhirImmunizationRecommendationRecommendation.Destroy;
begin
  FDate.free;
  FVaccineCode.free;
  FDoseNumber.free;
  FForecastStatus.free;
  FDateCriterionList.Free;
  FProtocol.free;
  FSupportingImmunizationList.Free;
  FSupportingPatientInformationList.Free;
  inherited;
end;

procedure TFhirImmunizationRecommendationRecommendation.Assign(oSource : TAdvObject);
begin
  inherited;
  dateElement := TFhirImmunizationRecommendationRecommendation(oSource).dateElement.Clone;
  vaccineCode := TFhirImmunizationRecommendationRecommendation(oSource).vaccineCode.Clone;
  doseNumberElement := TFhirImmunizationRecommendationRecommendation(oSource).doseNumberElement.Clone;
  forecastStatus := TFhirImmunizationRecommendationRecommendation(oSource).forecastStatus.Clone;
  if (TFhirImmunizationRecommendationRecommendation(oSource).FDateCriterionList = nil) then
  begin
    FDateCriterionList.free;
    FDateCriterionList := nil;
  end
  else
  begin
    FDateCriterionList := TFhirImmunizationRecommendationRecommendationDateCriterionList.Create;
    FDateCriterionList.Assign(TFhirImmunizationRecommendationRecommendation(oSource).FDateCriterionList);
  end;
  protocol := TFhirImmunizationRecommendationRecommendation(oSource).protocol.Clone;
  if (TFhirImmunizationRecommendationRecommendation(oSource).FSupportingImmunizationList = nil) then
  begin
    FSupportingImmunizationList.free;
    FSupportingImmunizationList := nil;
  end
  else
  begin
    FSupportingImmunizationList := TFhirReferenceList{TFhirImmunization}.Create;
    FSupportingImmunizationList.Assign(TFhirImmunizationRecommendationRecommendation(oSource).FSupportingImmunizationList);
  end;
  if (TFhirImmunizationRecommendationRecommendation(oSource).FSupportingPatientInformationList = nil) then
  begin
    FSupportingPatientInformationList.free;
    FSupportingPatientInformationList := nil;
  end
  else
  begin
    FSupportingPatientInformationList := TFhirReferenceList{Resource}.Create;
    FSupportingPatientInformationList.Assign(TFhirImmunizationRecommendationRecommendation(oSource).FSupportingPatientInformationList);
  end;
end;

procedure TFhirImmunizationRecommendationRecommendation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'vaccineCode') Then
     list.add(FVaccineCode.Link);
  if (child_name = 'doseNumber') Then
     list.add(FDoseNumber.Link);
  if (child_name = 'forecastStatus') Then
     list.add(FForecastStatus.Link);
  if (child_name = 'dateCriterion') Then
    list.addAll(FDateCriterionList);
  if (child_name = 'protocol') Then
     list.add(FProtocol.Link);
  if (child_name = 'supportingImmunization') Then
    list.addAll(FSupportingImmunizationList);
  if (child_name = 'supportingPatientInformation') Then
    list.addAll(FSupportingPatientInformationList);
end;

procedure TFhirImmunizationRecommendationRecommendation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'vaccineCode', 'CodeableConcept', FVaccineCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'doseNumber', 'positiveInt', FDoseNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'forecastStatus', 'CodeableConcept', FForecastStatus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateCriterion', '', FDateCriterionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'protocol', '', FProtocol.Link));{2}
  oList.add(TFHIRProperty.create(self, 'supportingImmunization', 'Reference(Immunization)', FSupportingImmunizationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'supportingPatientInformation', 'Reference(Observation|AllergyIntolerance)', FSupportingPatientInformationList.Link)){3};
end;

procedure TFhirImmunizationRecommendationRecommendation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'vaccineCode') then VaccineCode := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'doseNumber') then DoseNumberElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'forecastStatus') then ForecastStatus := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'dateCriterion') then DateCriterionList.add(propValue as TFhirImmunizationRecommendationRecommendationDateCriterion){2}
  else if (propName = 'protocol') then Protocol := propValue as TFhirImmunizationRecommendationRecommendationProtocol{4b}
  else if (propName = 'supportingImmunization') then SupportingImmunizationList.add(propValue as TFhirReference{TFhirImmunization}){2}
  else if (propName = 'supportingPatientInformation') then SupportingPatientInformationList.add(propValue as TFhirReference{Resource}){2}
  else inherited;
end;

function TFhirImmunizationRecommendationRecommendation.FhirType : string;
begin
  result := 'recommendation';
end;

function TFhirImmunizationRecommendationRecommendation.Link : TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation(inherited Link);
end;

function TFhirImmunizationRecommendationRecommendation.Clone : TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendation)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendation(other);
    result := compareDeep(dateElement, o.dateElement, true) and compareDeep(vaccineCodeElement, o.vaccineCodeElement, true) and 
      compareDeep(doseNumberElement, o.doseNumberElement, true) and compareDeep(forecastStatusElement, o.forecastStatusElement, true) and 
      compareDeep(dateCriterionList, o.dateCriterionList, true) and compareDeep(protocolElement, o.protocolElement, true) and 
      compareDeep(supportingImmunizationList, o.supportingImmunizationList, true) and 
      compareDeep(supportingPatientInformationList, o.supportingPatientInformationList, true);
  end;
end;

function TFhirImmunizationRecommendationRecommendation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendation)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendation(other);
    result := compareValues(dateElement, o.dateElement, true) and compareValues(doseNumberElement, o.doseNumberElement, true);
  end;
end;

{ TFhirImmunizationRecommendationRecommendation }

Procedure TFhirImmunizationRecommendationRecommendation.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirImmunizationRecommendationRecommendation.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetVaccineCode(value : TFhirCodeableConcept);
begin
  FVaccineCode.free;
  FVaccineCode := value;
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetDoseNumber(value : TFhirPositiveInt);
begin
  FDoseNumber.free;
  FDoseNumber := value;
end;

Function TFhirImmunizationRecommendationRecommendation.GetDoseNumberST : String;
begin
  if FDoseNumber = nil then
    result := ''
  else
    result := FDoseNumber.value;
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetDoseNumberST(value : String);
begin
  if value <> '' then
  begin
    if FDoseNumber = nil then
      FDoseNumber := TFhirPositiveInt.create;
    FDoseNumber.value := value
  end
  else if FDoseNumber <> nil then
    FDoseNumber.value := '';
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetForecastStatus(value : TFhirCodeableConcept);
begin
  FForecastStatus.free;
  FForecastStatus := value;
end;

Function TFhirImmunizationRecommendationRecommendation.GetDateCriterionList : TFhirImmunizationRecommendationRecommendationDateCriterionList;
begin
  if FDateCriterionList = nil then
    FDateCriterionList := TFhirImmunizationRecommendationRecommendationDateCriterionList.Create;
  result := FDateCriterionList;
end;

Function TFhirImmunizationRecommendationRecommendation.GetHasDateCriterionList : boolean;
begin
  result := (FDateCriterionList <> nil) and (FDateCriterionList.count > 0);
end;

Procedure TFhirImmunizationRecommendationRecommendation.SetProtocol(value : TFhirImmunizationRecommendationRecommendationProtocol);
begin
  FProtocol.free;
  FProtocol := value;
end;

Function TFhirImmunizationRecommendationRecommendation.GetSupportingImmunizationList : TFhirReferenceList{TFhirImmunization};
begin
  if FSupportingImmunizationList = nil then
    FSupportingImmunizationList := TFhirReferenceList{TFhirImmunization}.Create;
  result := FSupportingImmunizationList;
end;

Function TFhirImmunizationRecommendationRecommendation.GetHasSupportingImmunizationList : boolean;
begin
  result := (FSupportingImmunizationList <> nil) and (FSupportingImmunizationList.count > 0);
end;

Function TFhirImmunizationRecommendationRecommendation.GetSupportingPatientInformationList : TFhirReferenceList{Resource};
begin
  if FSupportingPatientInformationList = nil then
    FSupportingPatientInformationList := TFhirReferenceList{Resource}.Create;
  result := FSupportingPatientInformationList;
end;

Function TFhirImmunizationRecommendationRecommendation.GetHasSupportingPatientInformationList : boolean;
begin
  result := (FSupportingPatientInformationList <> nil) and (FSupportingPatientInformationList.count > 0);
end;


{ TFhirImmunizationRecommendationRecommendationListEnumerator }

Constructor TFhirImmunizationRecommendationRecommendationListEnumerator.Create(list : TFhirImmunizationRecommendationRecommendationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationRecommendationRecommendationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationRecommendationRecommendationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationRecommendationRecommendationListEnumerator.GetCurrent : TFhirImmunizationRecommendationRecommendation;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationRecommendationRecommendationList }
procedure TFhirImmunizationRecommendationRecommendationList.AddItem(value: TFhirImmunizationRecommendationRecommendation);
begin
  assert(value.ClassName = 'TFhirImmunizationRecommendationRecommendation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationRecommendationRecommendation');
  add(value);
end;


function TFhirImmunizationRecommendationRecommendationList.Append: TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationRecommendationRecommendationList.GetEnumerator : TFhirImmunizationRecommendationRecommendationListEnumerator;
begin
  result := TFhirImmunizationRecommendationRecommendationListEnumerator.Create(self.link);
end;

function TFhirImmunizationRecommendationRecommendationList.Clone: TFhirImmunizationRecommendationRecommendationList;
begin
  result := TFhirImmunizationRecommendationRecommendationList(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationRecommendationRecommendationList.GetItemN(index: Integer): TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationList.IndexOf(value: TFhirImmunizationRecommendationRecommendation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationRecommendationRecommendationList.Insert(index: Integer): TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationList.InsertItem(index: Integer; value: TFhirImmunizationRecommendationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendationRecommendation);
  Inherited Insert(index, value);
end;

function TFhirImmunizationRecommendationRecommendationList.Item(index: Integer): TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationList.Link: TFhirImmunizationRecommendationRecommendationList;
begin
  result := TFhirImmunizationRecommendationRecommendationList(inherited Link);
end;

procedure TFhirImmunizationRecommendationRecommendationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationRecommendationRecommendationList.SetItemByIndex(index: Integer; value: TFhirImmunizationRecommendationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendationRecommendation);
  FhirImmunizationRecommendationRecommendations[index] := value;
end;

procedure TFhirImmunizationRecommendationRecommendationList.SetItemN(index: Integer; value: TFhirImmunizationRecommendationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendationRecommendation);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationRecommendationRecommendationDateCriterion }

constructor TFhirImmunizationRecommendationRecommendationDateCriterion.Create;
begin
  inherited;
end;

destructor TFhirImmunizationRecommendationRecommendationDateCriterion.Destroy;
begin
  FCode.free;
  FValue.free;
  inherited;
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterion.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirImmunizationRecommendationRecommendationDateCriterion(oSource).code.Clone;
  valueElement := TFhirImmunizationRecommendationRecommendationDateCriterion(oSource).valueElement.Clone;
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterion.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterion.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'dateTime', FValue.Link));{2}
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterion.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirDateTime{5a}
  else inherited;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterion.FhirType : string;
begin
  result := 'dateCriterion';
end;

function TFhirImmunizationRecommendationRecommendationDateCriterion.Link : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion(inherited Link);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterion.Clone : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterion.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendationDateCriterion)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendationDateCriterion(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterion.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendationDateCriterion)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendationDateCriterion(other);
    result := compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirImmunizationRecommendationRecommendationDateCriterion }

Procedure TFhirImmunizationRecommendationRecommendationDateCriterion.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirImmunizationRecommendationRecommendationDateCriterion.SetValue(value : TFhirDateTime);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirImmunizationRecommendationRecommendationDateCriterion.GetValueST : TDateAndTime;
begin
  if FValue = nil then
    result := nil
  else
    result := FValue.value;
end;

Procedure TFhirImmunizationRecommendationRecommendationDateCriterion.SetValueST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FValue = nil then
      FValue := TFhirDateTime.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := nil;
end;


{ TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator }

Constructor TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator.Create(list : TFhirImmunizationRecommendationRecommendationDateCriterionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator.GetCurrent : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationRecommendationRecommendationDateCriterionList }
procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.AddItem(value: TFhirImmunizationRecommendationRecommendationDateCriterion);
begin
  assert(value.ClassName = 'TFhirImmunizationRecommendationRecommendationDateCriterion', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationRecommendationRecommendationDateCriterion');
  add(value);
end;


function TFhirImmunizationRecommendationRecommendationDateCriterionList.Append: TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.GetEnumerator : TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterionListEnumerator.Create(self.link);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.Clone: TFhirImmunizationRecommendationRecommendationDateCriterionList;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterionList(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.GetItemN(index: Integer): TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.IndexOf(value: TFhirImmunizationRecommendationRecommendationDateCriterion): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationRecommendationRecommendationDateCriterionList.Insert(index: Integer): TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.InsertItem(index: Integer; value: TFhirImmunizationRecommendationRecommendationDateCriterion);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationDateCriterion);
  Inherited Insert(index, value);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.Item(index: Integer): TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationDateCriterionList.Link: TFhirImmunizationRecommendationRecommendationDateCriterionList;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterionList(inherited Link);
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.SetItemByIndex(index: Integer; value: TFhirImmunizationRecommendationRecommendationDateCriterion);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationDateCriterion);
  FhirImmunizationRecommendationRecommendationDateCriterions[index] := value;
end;

procedure TFhirImmunizationRecommendationRecommendationDateCriterionList.SetItemN(index: Integer; value: TFhirImmunizationRecommendationRecommendationDateCriterion);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationDateCriterion);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationRecommendationRecommendationProtocol }

constructor TFhirImmunizationRecommendationRecommendationProtocol.Create;
begin
  inherited;
end;

destructor TFhirImmunizationRecommendationRecommendationProtocol.Destroy;
begin
  FDoseSequence.free;
  FDescription.free;
  FAuthority.free;
  FSeries.free;
  inherited;
end;

procedure TFhirImmunizationRecommendationRecommendationProtocol.Assign(oSource : TAdvObject);
begin
  inherited;
  doseSequenceElement := TFhirImmunizationRecommendationRecommendationProtocol(oSource).doseSequenceElement.Clone;
  descriptionElement := TFhirImmunizationRecommendationRecommendationProtocol(oSource).descriptionElement.Clone;
  authority := TFhirImmunizationRecommendationRecommendationProtocol(oSource).authority.Clone;
  seriesElement := TFhirImmunizationRecommendationRecommendationProtocol(oSource).seriesElement.Clone;
end;

procedure TFhirImmunizationRecommendationRecommendationProtocol.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'doseSequence') Then
     list.add(FDoseSequence.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'authority') Then
     list.add(FAuthority.Link);
  if (child_name = 'series') Then
     list.add(FSeries.Link);
end;

procedure TFhirImmunizationRecommendationRecommendationProtocol.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'doseSequence', 'integer', FDoseSequence.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authority', 'Reference(Organization)', FAuthority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'series', 'string', FSeries.Link));{2}
end;

procedure TFhirImmunizationRecommendationRecommendationProtocol.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'doseSequence') then DoseSequenceElement := propValue as TFhirInteger{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'authority') then Authority := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'series') then SeriesElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirImmunizationRecommendationRecommendationProtocol.FhirType : string;
begin
  result := 'protocol';
end;

function TFhirImmunizationRecommendationRecommendationProtocol.Link : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol(inherited Link);
end;

function TFhirImmunizationRecommendationRecommendationProtocol.Clone : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendationProtocol.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendationProtocol)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendationProtocol(other);
    result := compareDeep(doseSequenceElement, o.doseSequenceElement, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(authorityElement, o.authorityElement, true) and 
      compareDeep(seriesElement, o.seriesElement, true);
  end;
end;

function TFhirImmunizationRecommendationRecommendationProtocol.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImmunizationRecommendationRecommendationProtocol)) then
    result := false
  else
  begin
    o := TFhirImmunizationRecommendationRecommendationProtocol(other);
    result := compareValues(doseSequenceElement, o.doseSequenceElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(seriesElement, o.seriesElement, true);
  end;
end;

{ TFhirImmunizationRecommendationRecommendationProtocol }

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetDoseSequence(value : TFhirInteger);
begin
  FDoseSequence.free;
  FDoseSequence := value;
end;

Function TFhirImmunizationRecommendationRecommendationProtocol.GetDoseSequenceST : String;
begin
  if FDoseSequence = nil then
    result := ''
  else
    result := FDoseSequence.value;
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetDoseSequenceST(value : String);
begin
  if value <> '' then
  begin
    if FDoseSequence = nil then
      FDoseSequence := TFhirInteger.create;
    FDoseSequence.value := value
  end
  else if FDoseSequence <> nil then
    FDoseSequence.value := '';
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImmunizationRecommendationRecommendationProtocol.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetAuthority(value : TFhirReference{TFhirOrganization});
begin
  FAuthority.free;
  FAuthority := value;
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetSeries(value : TFhirString);
begin
  FSeries.free;
  FSeries := value;
end;

Function TFhirImmunizationRecommendationRecommendationProtocol.GetSeriesST : String;
begin
  if FSeries = nil then
    result := ''
  else
    result := FSeries.value;
end;

Procedure TFhirImmunizationRecommendationRecommendationProtocol.SetSeriesST(value : String);
begin
  if value <> '' then
  begin
    if FSeries = nil then
      FSeries := TFhirString.create;
    FSeries.value := value
  end
  else if FSeries <> nil then
    FSeries.value := '';
end;


{ TFhirImmunizationRecommendationRecommendationProtocolListEnumerator }

Constructor TFhirImmunizationRecommendationRecommendationProtocolListEnumerator.Create(list : TFhirImmunizationRecommendationRecommendationProtocolList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationRecommendationRecommendationProtocolListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationRecommendationRecommendationProtocolListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationRecommendationRecommendationProtocolListEnumerator.GetCurrent : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationRecommendationRecommendationProtocolList }
procedure TFhirImmunizationRecommendationRecommendationProtocolList.AddItem(value: TFhirImmunizationRecommendationRecommendationProtocol);
begin
  assert(value.ClassName = 'TFhirImmunizationRecommendationRecommendationProtocol', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationRecommendationRecommendationProtocol');
  add(value);
end;


function TFhirImmunizationRecommendationRecommendationProtocolList.Append: TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationProtocolList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.GetEnumerator : TFhirImmunizationRecommendationRecommendationProtocolListEnumerator;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocolListEnumerator.Create(self.link);
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.Clone: TFhirImmunizationRecommendationRecommendationProtocolList;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocolList(inherited Clone);
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.GetItemN(index: Integer): TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.IndexOf(value: TFhirImmunizationRecommendationRecommendationProtocol): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationRecommendationRecommendationProtocolList.Insert(index: Integer): TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationRecommendationProtocolList.InsertItem(index: Integer; value: TFhirImmunizationRecommendationRecommendationProtocol);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationProtocol);
  Inherited Insert(index, value);
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.Item(index: Integer): TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationRecommendationProtocolList.Link: TFhirImmunizationRecommendationRecommendationProtocolList;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocolList(inherited Link);
end;

procedure TFhirImmunizationRecommendationRecommendationProtocolList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationRecommendationRecommendationProtocolList.SetItemByIndex(index: Integer; value: TFhirImmunizationRecommendationRecommendationProtocol);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationProtocol);
  FhirImmunizationRecommendationRecommendationProtocols[index] := value;
end;

procedure TFhirImmunizationRecommendationRecommendationProtocolList.SetItemN(index: Integer; value: TFhirImmunizationRecommendationRecommendationProtocol);
begin
  assert(value is TFhirImmunizationRecommendationRecommendationProtocol);
  ObjectByIndex[index] := value;
end;

{ TFhirImmunizationRecommendation }

constructor TFhirImmunizationRecommendation.Create;
begin
  inherited;
end;

destructor TFhirImmunizationRecommendation.Destroy;
begin
  FIdentifierList.Free;
  FPatient.free;
  FRecommendationList.Free;
  inherited;
end;

function TFhirImmunizationRecommendation.GetResourceType : TFhirResourceType;
begin
  result := frtImmunizationRecommendation;
end;

procedure TFhirImmunizationRecommendation.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirImmunizationRecommendation(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirImmunizationRecommendation(oSource).FIdentifierList);
  end;
  patient := TFhirImmunizationRecommendation(oSource).patient.Clone;
  if (TFhirImmunizationRecommendation(oSource).FRecommendationList = nil) then
  begin
    FRecommendationList.free;
    FRecommendationList := nil;
  end
  else
  begin
    FRecommendationList := TFhirImmunizationRecommendationRecommendationList.Create;
    FRecommendationList.Assign(TFhirImmunizationRecommendation(oSource).FRecommendationList);
  end;
end;

procedure TFhirImmunizationRecommendation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'recommendation') Then
    list.addAll(FRecommendationList);
end;

procedure TFhirImmunizationRecommendation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recommendation', '', FRecommendationList.Link)){3};
end;

procedure TFhirImmunizationRecommendation.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'recommendation') then RecommendationList.add(propValue as TFhirImmunizationRecommendationRecommendation){2}
  else inherited;
end;

function TFhirImmunizationRecommendation.FhirType : string;
begin
  result := 'ImmunizationRecommendation';
end;

function TFhirImmunizationRecommendation.Link : TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation(inherited Link);
end;

function TFhirImmunizationRecommendation.Clone : TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation(inherited Clone);
end;

{ TFhirImmunizationRecommendation }

Function TFhirImmunizationRecommendation.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirImmunizationRecommendation.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirImmunizationRecommendation.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Function TFhirImmunizationRecommendation.GetRecommendationList : TFhirImmunizationRecommendationRecommendationList;
begin
  if FRecommendationList = nil then
    FRecommendationList := TFhirImmunizationRecommendationRecommendationList.Create;
  result := FRecommendationList;
end;

Function TFhirImmunizationRecommendation.GetHasRecommendationList : boolean;
begin
  result := (FRecommendationList <> nil) and (FRecommendationList.count > 0);
end;


{ TFhirImmunizationRecommendationListEnumerator }

Constructor TFhirImmunizationRecommendationListEnumerator.Create(list : TFhirImmunizationRecommendationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImmunizationRecommendationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImmunizationRecommendationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImmunizationRecommendationListEnumerator.GetCurrent : TFhirImmunizationRecommendation;
begin
  Result := FList[FIndex];
end;


{ TFhirImmunizationRecommendationList }
procedure TFhirImmunizationRecommendationList.AddItem(value: TFhirImmunizationRecommendation);
begin
  assert(value.ClassName = 'TFhirImmunizationRecommendation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImmunizationRecommendation');
  add(value);
end;


function TFhirImmunizationRecommendationList.Append: TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationList.ClearItems;
begin
  Clear;
end;

function TFhirImmunizationRecommendationList.GetEnumerator : TFhirImmunizationRecommendationListEnumerator;
begin
  result := TFhirImmunizationRecommendationListEnumerator.Create(self.link);
end;

function TFhirImmunizationRecommendationList.Clone: TFhirImmunizationRecommendationList;
begin
  result := TFhirImmunizationRecommendationList(inherited Clone);
end;

function TFhirImmunizationRecommendationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImmunizationRecommendationList.GetItemN(index: Integer): TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationList.IndexOf(value: TFhirImmunizationRecommendation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImmunizationRecommendationList.Insert(index: Integer): TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImmunizationRecommendationList.InsertItem(index: Integer; value: TFhirImmunizationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendation);
  Inherited Insert(index, value);
end;

function TFhirImmunizationRecommendationList.Item(index: Integer): TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation(ObjectByIndex[index]);
end;

function TFhirImmunizationRecommendationList.Link: TFhirImmunizationRecommendationList;
begin
  result := TFhirImmunizationRecommendationList(inherited Link);
end;

procedure TFhirImmunizationRecommendationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImmunizationRecommendationList.SetItemByIndex(index: Integer; value: TFhirImmunizationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendation);
  FhirImmunizationRecommendations[index] := value;
end;

procedure TFhirImmunizationRecommendationList.SetItemN(index: Integer; value: TFhirImmunizationRecommendation);
begin
  assert(value is TFhirImmunizationRecommendation);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuideContact }

constructor TFhirImplementationGuideContact.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuideContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirImplementationGuideContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirImplementationGuideContact(oSource).nameElement.Clone;
  if (TFhirImplementationGuideContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirImplementationGuideContact(oSource).FTelecomList);
  end;
end;

procedure TFhirImplementationGuideContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirImplementationGuideContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirImplementationGuideContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirImplementationGuideContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirImplementationGuideContact.Link : TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact(inherited Link);
end;

function TFhirImplementationGuideContact.Clone : TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact(inherited Clone);
end;

function TFhirImplementationGuideContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideContact)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirImplementationGuideContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideContact)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirImplementationGuideContact }

Procedure TFhirImplementationGuideContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirImplementationGuideContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirImplementationGuideContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirImplementationGuideContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirImplementationGuideContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirImplementationGuideContactListEnumerator }

Constructor TFhirImplementationGuideContactListEnumerator.Create(list : TFhirImplementationGuideContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuideContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuideContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuideContactListEnumerator.GetCurrent : TFhirImplementationGuideContact;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuideContactList }
procedure TFhirImplementationGuideContactList.AddItem(value: TFhirImplementationGuideContact);
begin
  assert(value.ClassName = 'TFhirImplementationGuideContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuideContact');
  add(value);
end;


function TFhirImplementationGuideContactList.Append: TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideContactList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuideContactList.GetEnumerator : TFhirImplementationGuideContactListEnumerator;
begin
  result := TFhirImplementationGuideContactListEnumerator.Create(self.link);
end;

function TFhirImplementationGuideContactList.Clone: TFhirImplementationGuideContactList;
begin
  result := TFhirImplementationGuideContactList(inherited Clone);
end;

function TFhirImplementationGuideContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuideContactList.GetItemN(index: Integer): TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact(ObjectByIndex[index]);
end;

function TFhirImplementationGuideContactList.IndexOf(value: TFhirImplementationGuideContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuideContactList.Insert(index: Integer): TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideContactList.InsertItem(index: Integer; value: TFhirImplementationGuideContact);
begin
  assert(value is TFhirImplementationGuideContact);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuideContactList.Item(index: Integer): TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact(ObjectByIndex[index]);
end;

function TFhirImplementationGuideContactList.Link: TFhirImplementationGuideContactList;
begin
  result := TFhirImplementationGuideContactList(inherited Link);
end;

procedure TFhirImplementationGuideContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuideContactList.SetItemByIndex(index: Integer; value: TFhirImplementationGuideContact);
begin
  assert(value is TFhirImplementationGuideContact);
  FhirImplementationGuideContacts[index] := value;
end;

procedure TFhirImplementationGuideContactList.SetItemN(index: Integer; value: TFhirImplementationGuideContact);
begin
  assert(value is TFhirImplementationGuideContact);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuideDependency }

constructor TFhirImplementationGuideDependency.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuideDependency.Destroy;
begin
  FType_.free;
  FUri.free;
  inherited;
end;

procedure TFhirImplementationGuideDependency.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirImplementationGuideDependency(oSource).FType_.Link;
  uriElement := TFhirImplementationGuideDependency(oSource).uriElement.Clone;
end;

procedure TFhirImplementationGuideDependency.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'uri') Then
     list.add(FUri.Link);
end;

procedure TFhirImplementationGuideDependency.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'uri', 'uri', FUri.Link));{2}
end;

procedure TFhirImplementationGuideDependency.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'uri') then UriElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirImplementationGuideDependency.FhirType : string;
begin
  result := 'dependency';
end;

function TFhirImplementationGuideDependency.Link : TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency(inherited Link);
end;

function TFhirImplementationGuideDependency.Clone : TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency(inherited Clone);
end;

function TFhirImplementationGuideDependency.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideDependency;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideDependency)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideDependency(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(uriElement, o.uriElement, true);
  end;
end;

function TFhirImplementationGuideDependency.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideDependency;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideDependency)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideDependency(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(uriElement, o.uriElement, true);
  end;
end;

{ TFhirImplementationGuideDependency }

Procedure TFhirImplementationGuideDependency.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirImplementationGuideDependency.GetType_ST : TFhirGuideDependencyType;
begin
  if FType_ = nil then
    result := TFhirGuideDependencyType(0)
  else
    result := TFhirGuideDependencyType(StringArrayIndexOfSensitive(CODES_TFhirGuideDependencyType, FType_.value));
end;

Procedure TFhirImplementationGuideDependency.SetType_ST(value : TFhirGuideDependencyType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirGuideDependencyType[value]);
end;

Procedure TFhirImplementationGuideDependency.SetUri(value : TFhirUri);
begin
  FUri.free;
  FUri := value;
end;

Function TFhirImplementationGuideDependency.GetUriST : String;
begin
  if FUri = nil then
    result := ''
  else
    result := FUri.value;
end;

Procedure TFhirImplementationGuideDependency.SetUriST(value : String);
begin
  if value <> '' then
  begin
    if FUri = nil then
      FUri := TFhirUri.create;
    FUri.value := value
  end
  else if FUri <> nil then
    FUri.value := '';
end;


{ TFhirImplementationGuideDependencyListEnumerator }

Constructor TFhirImplementationGuideDependencyListEnumerator.Create(list : TFhirImplementationGuideDependencyList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuideDependencyListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuideDependencyListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuideDependencyListEnumerator.GetCurrent : TFhirImplementationGuideDependency;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuideDependencyList }
procedure TFhirImplementationGuideDependencyList.AddItem(value: TFhirImplementationGuideDependency);
begin
  assert(value.ClassName = 'TFhirImplementationGuideDependency', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuideDependency');
  add(value);
end;


function TFhirImplementationGuideDependencyList.Append: TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideDependencyList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuideDependencyList.GetEnumerator : TFhirImplementationGuideDependencyListEnumerator;
begin
  result := TFhirImplementationGuideDependencyListEnumerator.Create(self.link);
end;

function TFhirImplementationGuideDependencyList.Clone: TFhirImplementationGuideDependencyList;
begin
  result := TFhirImplementationGuideDependencyList(inherited Clone);
end;

function TFhirImplementationGuideDependencyList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuideDependencyList.GetItemN(index: Integer): TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency(ObjectByIndex[index]);
end;

function TFhirImplementationGuideDependencyList.IndexOf(value: TFhirImplementationGuideDependency): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuideDependencyList.Insert(index: Integer): TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideDependencyList.InsertItem(index: Integer; value: TFhirImplementationGuideDependency);
begin
  assert(value is TFhirImplementationGuideDependency);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuideDependencyList.Item(index: Integer): TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency(ObjectByIndex[index]);
end;

function TFhirImplementationGuideDependencyList.Link: TFhirImplementationGuideDependencyList;
begin
  result := TFhirImplementationGuideDependencyList(inherited Link);
end;

procedure TFhirImplementationGuideDependencyList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuideDependencyList.SetItemByIndex(index: Integer; value: TFhirImplementationGuideDependency);
begin
  assert(value is TFhirImplementationGuideDependency);
  FhirImplementationGuideDependencies[index] := value;
end;

procedure TFhirImplementationGuideDependencyList.SetItemN(index: Integer; value: TFhirImplementationGuideDependency);
begin
  assert(value is TFhirImplementationGuideDependency);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuidePackage }

constructor TFhirImplementationGuidePackage.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuidePackage.Destroy;
begin
  FName.free;
  FDescription.free;
  FResourceList.Free;
  inherited;
end;

procedure TFhirImplementationGuidePackage.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirImplementationGuidePackage(oSource).nameElement.Clone;
  descriptionElement := TFhirImplementationGuidePackage(oSource).descriptionElement.Clone;
  if (TFhirImplementationGuidePackage(oSource).FResourceList = nil) then
  begin
    FResourceList.free;
    FResourceList := nil;
  end
  else
  begin
    FResourceList := TFhirImplementationGuidePackageResourceList.Create;
    FResourceList.Assign(TFhirImplementationGuidePackage(oSource).FResourceList);
  end;
end;

procedure TFhirImplementationGuidePackage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'resource') Then
    list.addAll(FResourceList);
end;

procedure TFhirImplementationGuidePackage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', '', FResourceList.Link)){3};
end;

procedure TFhirImplementationGuidePackage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'resource') then ResourceList.add(propValue as TFhirImplementationGuidePackageResource){2}
  else inherited;
end;

function TFhirImplementationGuidePackage.FhirType : string;
begin
  result := 'package';
end;

function TFhirImplementationGuidePackage.Link : TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage(inherited Link);
end;

function TFhirImplementationGuidePackage.Clone : TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage(inherited Clone);
end;

function TFhirImplementationGuidePackage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePackage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePackage)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePackage(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(resourceList, o.resourceList, true);
  end;
end;

function TFhirImplementationGuidePackage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePackage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePackage)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePackage(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirImplementationGuidePackage }

Procedure TFhirImplementationGuidePackage.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirImplementationGuidePackage.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirImplementationGuidePackage.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirImplementationGuidePackage.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImplementationGuidePackage.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImplementationGuidePackage.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirImplementationGuidePackage.GetResourceList : TFhirImplementationGuidePackageResourceList;
begin
  if FResourceList = nil then
    FResourceList := TFhirImplementationGuidePackageResourceList.Create;
  result := FResourceList;
end;

Function TFhirImplementationGuidePackage.GetHasResourceList : boolean;
begin
  result := (FResourceList <> nil) and (FResourceList.count > 0);
end;


{ TFhirImplementationGuidePackageListEnumerator }

Constructor TFhirImplementationGuidePackageListEnumerator.Create(list : TFhirImplementationGuidePackageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuidePackageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuidePackageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuidePackageListEnumerator.GetCurrent : TFhirImplementationGuidePackage;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuidePackageList }
procedure TFhirImplementationGuidePackageList.AddItem(value: TFhirImplementationGuidePackage);
begin
  assert(value.ClassName = 'TFhirImplementationGuidePackage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuidePackage');
  add(value);
end;


function TFhirImplementationGuidePackageList.Append: TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePackageList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuidePackageList.GetEnumerator : TFhirImplementationGuidePackageListEnumerator;
begin
  result := TFhirImplementationGuidePackageListEnumerator.Create(self.link);
end;

function TFhirImplementationGuidePackageList.Clone: TFhirImplementationGuidePackageList;
begin
  result := TFhirImplementationGuidePackageList(inherited Clone);
end;

function TFhirImplementationGuidePackageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuidePackageList.GetItemN(index: Integer): TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePackageList.IndexOf(value: TFhirImplementationGuidePackage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuidePackageList.Insert(index: Integer): TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePackageList.InsertItem(index: Integer; value: TFhirImplementationGuidePackage);
begin
  assert(value is TFhirImplementationGuidePackage);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuidePackageList.Item(index: Integer): TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePackageList.Link: TFhirImplementationGuidePackageList;
begin
  result := TFhirImplementationGuidePackageList(inherited Link);
end;

procedure TFhirImplementationGuidePackageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuidePackageList.SetItemByIndex(index: Integer; value: TFhirImplementationGuidePackage);
begin
  assert(value is TFhirImplementationGuidePackage);
  FhirImplementationGuidePackages[index] := value;
end;

procedure TFhirImplementationGuidePackageList.SetItemN(index: Integer; value: TFhirImplementationGuidePackage);
begin
  assert(value is TFhirImplementationGuidePackage);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuidePackageResource }

constructor TFhirImplementationGuidePackageResource.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuidePackageResource.Destroy;
begin
  FPurpose.free;
  FName.free;
  FDescription.free;
  FAcronym.free;
  FSource.free;
  FExampleFor.free;
  inherited;
end;

procedure TFhirImplementationGuidePackageResource.Assign(oSource : TAdvObject);
begin
  inherited;
  FPurpose := TFhirImplementationGuidePackageResource(oSource).FPurpose.Link;
  nameElement := TFhirImplementationGuidePackageResource(oSource).nameElement.Clone;
  descriptionElement := TFhirImplementationGuidePackageResource(oSource).descriptionElement.Clone;
  acronymElement := TFhirImplementationGuidePackageResource(oSource).acronymElement.Clone;
  source := TFhirImplementationGuidePackageResource(oSource).source.Clone;
  exampleFor := TFhirImplementationGuidePackageResource(oSource).exampleFor.Clone;
end;

procedure TFhirImplementationGuidePackageResource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'purpose') Then
     list.add(FPurpose.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'acronym') Then
     list.add(FAcronym.Link);
  if (child_name = 'source[x]') Then
     list.add(FSource.Link);
  if (child_name = 'exampleFor') Then
     list.add(FExampleFor.Link);
end;

procedure TFhirImplementationGuidePackageResource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'purpose', 'code', FPurpose.Link));{1}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'acronym', 'string', FAcronym.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source[x]', 'uri|Reference(Any)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'exampleFor', 'Reference(StructureDefinition)', FExampleFor.Link));{2}
end;

procedure TFhirImplementationGuidePackageResource.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'purpose') then PurposeElement := propValue as TFHIREnum
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'acronym') then AcronymElement := propValue as TFhirString{5a}
  else if (propName.startsWith('source')) then Source := propValue as TFhirType{4}
  else if (propName = 'exampleFor') then ExampleFor := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else inherited;
end;

function TFhirImplementationGuidePackageResource.FhirType : string;
begin
  result := 'resource';
end;

function TFhirImplementationGuidePackageResource.Link : TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource(inherited Link);
end;

function TFhirImplementationGuidePackageResource.Clone : TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource(inherited Clone);
end;

function TFhirImplementationGuidePackageResource.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePackageResource;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePackageResource)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePackageResource(other);
    result := compareDeep(purposeElement, o.purposeElement, true) and compareDeep(nameElement, o.nameElement, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(acronymElement, o.acronymElement, true) and 
      compareDeep(sourceElement, o.sourceElement, true) and compareDeep(exampleForElement, o.exampleForElement, true);
  end;
end;

function TFhirImplementationGuidePackageResource.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePackageResource;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePackageResource)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePackageResource(other);
    result := compareValues(purposeElement, o.purposeElement, true) and compareValues(nameElement, o.nameElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(acronymElement, o.acronymElement, true);
  end;
end;

{ TFhirImplementationGuidePackageResource }

Procedure TFhirImplementationGuidePackageResource.SetPurpose(value : TFhirEnum);
begin
  FPurpose.free;
  FPurpose := value;
end;

Function TFhirImplementationGuidePackageResource.GetPurposeST : TFhirGuideResourcePurpose;
begin
  if FPurpose = nil then
    result := TFhirGuideResourcePurpose(0)
  else
    result := TFhirGuideResourcePurpose(StringArrayIndexOfSensitive(CODES_TFhirGuideResourcePurpose, FPurpose.value));
end;

Procedure TFhirImplementationGuidePackageResource.SetPurposeST(value : TFhirGuideResourcePurpose);
begin
  if ord(value) = 0 then
    PurposeElement := nil
  else
    PurposeElement := TFhirEnum.create(CODES_TFhirGuideResourcePurpose[value]);
end;

Procedure TFhirImplementationGuidePackageResource.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirImplementationGuidePackageResource.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirImplementationGuidePackageResource.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirImplementationGuidePackageResource.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImplementationGuidePackageResource.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImplementationGuidePackageResource.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirImplementationGuidePackageResource.SetAcronym(value : TFhirString);
begin
  FAcronym.free;
  FAcronym := value;
end;

Function TFhirImplementationGuidePackageResource.GetAcronymST : String;
begin
  if FAcronym = nil then
    result := ''
  else
    result := FAcronym.value;
end;

Procedure TFhirImplementationGuidePackageResource.SetAcronymST(value : String);
begin
  if value <> '' then
  begin
    if FAcronym = nil then
      FAcronym := TFhirString.create;
    FAcronym.value := value
  end
  else if FAcronym <> nil then
    FAcronym.value := '';
end;

Procedure TFhirImplementationGuidePackageResource.SetSource(value : TFhirType);
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirImplementationGuidePackageResource.SetExampleFor(value : TFhirReference{TFhirStructureDefinition});
begin
  FExampleFor.free;
  FExampleFor := value;
end;


{ TFhirImplementationGuidePackageResourceListEnumerator }

Constructor TFhirImplementationGuidePackageResourceListEnumerator.Create(list : TFhirImplementationGuidePackageResourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuidePackageResourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuidePackageResourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuidePackageResourceListEnumerator.GetCurrent : TFhirImplementationGuidePackageResource;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuidePackageResourceList }
procedure TFhirImplementationGuidePackageResourceList.AddItem(value: TFhirImplementationGuidePackageResource);
begin
  assert(value.ClassName = 'TFhirImplementationGuidePackageResource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuidePackageResource');
  add(value);
end;


function TFhirImplementationGuidePackageResourceList.Append: TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePackageResourceList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuidePackageResourceList.GetEnumerator : TFhirImplementationGuidePackageResourceListEnumerator;
begin
  result := TFhirImplementationGuidePackageResourceListEnumerator.Create(self.link);
end;

function TFhirImplementationGuidePackageResourceList.Clone: TFhirImplementationGuidePackageResourceList;
begin
  result := TFhirImplementationGuidePackageResourceList(inherited Clone);
end;

function TFhirImplementationGuidePackageResourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuidePackageResourceList.GetItemN(index: Integer): TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePackageResourceList.IndexOf(value: TFhirImplementationGuidePackageResource): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuidePackageResourceList.Insert(index: Integer): TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePackageResourceList.InsertItem(index: Integer; value: TFhirImplementationGuidePackageResource);
begin
  assert(value is TFhirImplementationGuidePackageResource);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuidePackageResourceList.Item(index: Integer): TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePackageResourceList.Link: TFhirImplementationGuidePackageResourceList;
begin
  result := TFhirImplementationGuidePackageResourceList(inherited Link);
end;

procedure TFhirImplementationGuidePackageResourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuidePackageResourceList.SetItemByIndex(index: Integer; value: TFhirImplementationGuidePackageResource);
begin
  assert(value is TFhirImplementationGuidePackageResource);
  FhirImplementationGuidePackageResources[index] := value;
end;

procedure TFhirImplementationGuidePackageResourceList.SetItemN(index: Integer; value: TFhirImplementationGuidePackageResource);
begin
  assert(value is TFhirImplementationGuidePackageResource);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuideGlobal }

constructor TFhirImplementationGuideGlobal.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuideGlobal.Destroy;
begin
  FType_.free;
  FProfile.free;
  inherited;
end;

procedure TFhirImplementationGuideGlobal.Assign(oSource : TAdvObject);
begin
  inherited;
  type_Element := TFhirImplementationGuideGlobal(oSource).type_Element.Clone;
  profile := TFhirImplementationGuideGlobal(oSource).profile.Clone;
end;

procedure TFhirImplementationGuideGlobal.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'profile') Then
     list.add(FProfile.Link);
end;

procedure TFhirImplementationGuideGlobal.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(StructureDefinition)', FProfile.Link));{2}
end;

procedure TFhirImplementationGuideGlobal.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then
    if propValue is TFHIRCode then
      Type_Element := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      Type_Element := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'profile') then Profile := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else inherited;
end;

function TFhirImplementationGuideGlobal.FhirType : string;
begin
  result := 'global';
end;

function TFhirImplementationGuideGlobal.Link : TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal(inherited Link);
end;

function TFhirImplementationGuideGlobal.Clone : TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal(inherited Clone);
end;

function TFhirImplementationGuideGlobal.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideGlobal;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideGlobal)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideGlobal(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(profileElement, o.profileElement, true);
  end;
end;

function TFhirImplementationGuideGlobal.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuideGlobal;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuideGlobal)) then
    result := false
  else
  begin
    o := TFhirImplementationGuideGlobal(other);
    result := compareValues(type_Element, o.type_Element, true);
  end;
end;

{ TFhirImplementationGuideGlobal }

Procedure TFhirImplementationGuideGlobal.SetType_(value : TFhirCode);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirImplementationGuideGlobal.GetType_ST : String;
begin
  if FType_ = nil then
    result := ''
  else
    result := FType_.value;
end;

Procedure TFhirImplementationGuideGlobal.SetType_ST(value : String);
begin
  if value <> '' then
  begin
    if FType_ = nil then
      FType_ := TFhirCode.create;
    FType_.value := value
  end
  else if FType_ <> nil then
    FType_.value := '';
end;

Procedure TFhirImplementationGuideGlobal.SetProfile(value : TFhirReference{TFhirStructureDefinition});
begin
  FProfile.free;
  FProfile := value;
end;


{ TFhirImplementationGuideGlobalListEnumerator }

Constructor TFhirImplementationGuideGlobalListEnumerator.Create(list : TFhirImplementationGuideGlobalList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuideGlobalListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuideGlobalListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuideGlobalListEnumerator.GetCurrent : TFhirImplementationGuideGlobal;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuideGlobalList }
procedure TFhirImplementationGuideGlobalList.AddItem(value: TFhirImplementationGuideGlobal);
begin
  assert(value.ClassName = 'TFhirImplementationGuideGlobal', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuideGlobal');
  add(value);
end;


function TFhirImplementationGuideGlobalList.Append: TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideGlobalList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuideGlobalList.GetEnumerator : TFhirImplementationGuideGlobalListEnumerator;
begin
  result := TFhirImplementationGuideGlobalListEnumerator.Create(self.link);
end;

function TFhirImplementationGuideGlobalList.Clone: TFhirImplementationGuideGlobalList;
begin
  result := TFhirImplementationGuideGlobalList(inherited Clone);
end;

function TFhirImplementationGuideGlobalList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuideGlobalList.GetItemN(index: Integer): TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal(ObjectByIndex[index]);
end;

function TFhirImplementationGuideGlobalList.IndexOf(value: TFhirImplementationGuideGlobal): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuideGlobalList.Insert(index: Integer): TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideGlobalList.InsertItem(index: Integer; value: TFhirImplementationGuideGlobal);
begin
  assert(value is TFhirImplementationGuideGlobal);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuideGlobalList.Item(index: Integer): TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal(ObjectByIndex[index]);
end;

function TFhirImplementationGuideGlobalList.Link: TFhirImplementationGuideGlobalList;
begin
  result := TFhirImplementationGuideGlobalList(inherited Link);
end;

procedure TFhirImplementationGuideGlobalList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuideGlobalList.SetItemByIndex(index: Integer; value: TFhirImplementationGuideGlobal);
begin
  assert(value is TFhirImplementationGuideGlobal);
  FhirImplementationGuideGlobals[index] := value;
end;

procedure TFhirImplementationGuideGlobalList.SetItemN(index: Integer; value: TFhirImplementationGuideGlobal);
begin
  assert(value is TFhirImplementationGuideGlobal);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuidePage }

constructor TFhirImplementationGuidePage.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuidePage.Destroy;
begin
  FSource.free;
  FName.free;
  FKind.free;
  FType_List.Free;
  FPackageList.Free;
  FFormat.free;
  FPageList.Free;
  inherited;
end;

procedure TFhirImplementationGuidePage.Assign(oSource : TAdvObject);
begin
  inherited;
  sourceElement := TFhirImplementationGuidePage(oSource).sourceElement.Clone;
  nameElement := TFhirImplementationGuidePage(oSource).nameElement.Clone;
  FKind := TFhirImplementationGuidePage(oSource).FKind.Link;
  if (TFhirImplementationGuidePage(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeList.Create;
    FType_List.Assign(TFhirImplementationGuidePage(oSource).FType_List);
  end;
  if (TFhirImplementationGuidePage(oSource).FPackageList = nil) then
  begin
    FPackageList.free;
    FPackageList := nil;
  end
  else
  begin
    FPackageList := TFhirStringList.Create;
    FPackageList.Assign(TFhirImplementationGuidePage(oSource).FPackageList);
  end;
  formatElement := TFhirImplementationGuidePage(oSource).formatElement.Clone;
  if (TFhirImplementationGuidePage(oSource).FPageList = nil) then
  begin
    FPageList.free;
    FPageList := nil;
  end
  else
  begin
    FPageList := TFhirImplementationGuidePageList.Create;
    FPageList.Assign(TFhirImplementationGuidePage(oSource).FPageList);
  end;
end;

procedure TFhirImplementationGuidePage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'package') Then
    list.addAll(FPackageList);
  if (child_name = 'format') Then
     list.add(FFormat.Link);
  if (child_name = 'page') Then
    list.addAll(FPageList);
end;

procedure TFhirImplementationGuidePage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'source', 'uri', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'kind', 'code', FKind.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'package', 'string', FPackageList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'format', 'code', FFormat.Link));{2}
  oList.add(TFHIRProperty.create(self, 'page', '@ImplementationGuide.page', FPageList.Link)){3};
end;

procedure TFhirImplementationGuidePage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'source') then SourceElement := propValue as TFhirUri{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'kind') then KindElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_List.add(propValue as TFhirCode){2}
  else if (propName = 'package') then PackageList.add(propValue as TFhirString){2}
  else if (propName = 'format') then
    if propValue is TFHIRCode then
      FormatElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      FormatElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'page') then PageList.add(propValue as TFhirImplementationGuidePage){2}
  else inherited;
end;

function TFhirImplementationGuidePage.FhirType : string;
begin
  result := 'page';
end;

function TFhirImplementationGuidePage.Link : TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage(inherited Link);
end;

function TFhirImplementationGuidePage.Clone : TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage(inherited Clone);
end;

function TFhirImplementationGuidePage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePage)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePage(other);
    result := compareDeep(sourceElement, o.sourceElement, true) and compareDeep(nameElement, o.nameElement, true) and 
      compareDeep(kindElement, o.kindElement, true) and compareDeep(type_List, o.type_List, true) and 
      compareDeep(packageList, o.packageList, true) and compareDeep(formatElement, o.formatElement, true) and 
      compareDeep(pageList, o.pageList, true);
  end;
end;

function TFhirImplementationGuidePage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirImplementationGuidePage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirImplementationGuidePage)) then
    result := false
  else
  begin
    o := TFhirImplementationGuidePage(other);
    result := compareValues(sourceElement, o.sourceElement, true) and compareValues(nameElement, o.nameElement, true) and 
      compareValues(kindElement, o.kindElement, true) and compareValues(type_List, o.type_List, true) and 
      compareValues(packageList, o.packageList, true) and compareValues(formatElement, o.formatElement, true);
  end;
end;

{ TFhirImplementationGuidePage }

Procedure TFhirImplementationGuidePage.SetSource(value : TFhirUri);
begin
  FSource.free;
  FSource := value;
end;

Function TFhirImplementationGuidePage.GetSourceST : String;
begin
  if FSource = nil then
    result := ''
  else
    result := FSource.value;
end;

Procedure TFhirImplementationGuidePage.SetSourceST(value : String);
begin
  if value <> '' then
  begin
    if FSource = nil then
      FSource := TFhirUri.create;
    FSource.value := value
  end
  else if FSource <> nil then
    FSource.value := '';
end;

Procedure TFhirImplementationGuidePage.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirImplementationGuidePage.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirImplementationGuidePage.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirImplementationGuidePage.SetKind(value : TFhirEnum);
begin
  FKind.free;
  FKind := value;
end;

Function TFhirImplementationGuidePage.GetKindST : TFhirGuidePageKind;
begin
  if FKind = nil then
    result := TFhirGuidePageKind(0)
  else
    result := TFhirGuidePageKind(StringArrayIndexOfSensitive(CODES_TFhirGuidePageKind, FKind.value));
end;

Procedure TFhirImplementationGuidePage.SetKindST(value : TFhirGuidePageKind);
begin
  if ord(value) = 0 then
    KindElement := nil
  else
    KindElement := TFhirEnum.create(CODES_TFhirGuidePageKind[value]);
end;

Function TFhirImplementationGuidePage.GetType_List : TFhirCodeList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeList.Create;
  result := FType_List;
end;

Function TFhirImplementationGuidePage.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Function TFhirImplementationGuidePage.GetPackageList : TFhirStringList;
begin
  if FPackageList = nil then
    FPackageList := TFhirStringList.Create;
  result := FPackageList;
end;

Function TFhirImplementationGuidePage.GetHasPackageList : boolean;
begin
  result := (FPackageList <> nil) and (FPackageList.count > 0);
end;

Procedure TFhirImplementationGuidePage.SetFormat(value : TFhirCode);
begin
  FFormat.free;
  FFormat := value;
end;

Function TFhirImplementationGuidePage.GetFormatST : String;
begin
  if FFormat = nil then
    result := ''
  else
    result := FFormat.value;
end;

Procedure TFhirImplementationGuidePage.SetFormatST(value : String);
begin
  if value <> '' then
  begin
    if FFormat = nil then
      FFormat := TFhirCode.create;
    FFormat.value := value
  end
  else if FFormat <> nil then
    FFormat.value := '';
end;

Function TFhirImplementationGuidePage.GetPageList : TFhirImplementationGuidePageList;
begin
  if FPageList = nil then
    FPageList := TFhirImplementationGuidePageList.Create;
  result := FPageList;
end;

Function TFhirImplementationGuidePage.GetHasPageList : boolean;
begin
  result := (FPageList <> nil) and (FPageList.count > 0);
end;


{ TFhirImplementationGuidePageListEnumerator }

Constructor TFhirImplementationGuidePageListEnumerator.Create(list : TFhirImplementationGuidePageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuidePageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuidePageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuidePageListEnumerator.GetCurrent : TFhirImplementationGuidePage;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuidePageList }
procedure TFhirImplementationGuidePageList.AddItem(value: TFhirImplementationGuidePage);
begin
  assert(value.ClassName = 'TFhirImplementationGuidePage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuidePage');
  add(value);
end;


function TFhirImplementationGuidePageList.Append: TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePageList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuidePageList.GetEnumerator : TFhirImplementationGuidePageListEnumerator;
begin
  result := TFhirImplementationGuidePageListEnumerator.Create(self.link);
end;

function TFhirImplementationGuidePageList.Clone: TFhirImplementationGuidePageList;
begin
  result := TFhirImplementationGuidePageList(inherited Clone);
end;

function TFhirImplementationGuidePageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuidePageList.GetItemN(index: Integer): TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePageList.IndexOf(value: TFhirImplementationGuidePage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuidePageList.Insert(index: Integer): TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuidePageList.InsertItem(index: Integer; value: TFhirImplementationGuidePage);
begin
  assert(value is TFhirImplementationGuidePage);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuidePageList.Item(index: Integer): TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage(ObjectByIndex[index]);
end;

function TFhirImplementationGuidePageList.Link: TFhirImplementationGuidePageList;
begin
  result := TFhirImplementationGuidePageList(inherited Link);
end;

procedure TFhirImplementationGuidePageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuidePageList.SetItemByIndex(index: Integer; value: TFhirImplementationGuidePage);
begin
  assert(value is TFhirImplementationGuidePage);
  FhirImplementationGuidePages[index] := value;
end;

procedure TFhirImplementationGuidePageList.SetItemN(index: Integer; value: TFhirImplementationGuidePage);
begin
  assert(value is TFhirImplementationGuidePage);
  ObjectByIndex[index] := value;
end;

{ TFhirImplementationGuide }

constructor TFhirImplementationGuide.Create;
begin
  inherited;
end;

destructor TFhirImplementationGuide.Destroy;
begin
  FUrl.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FUseContextList.Free;
  FCopyright.free;
  FFhirVersion.free;
  FDependencyList.Free;
  FPackageList.Free;
  FGlobalList.Free;
  FBinaryList.Free;
  FPage.free;
  inherited;
end;

function TFhirImplementationGuide.GetResourceType : TFhirResourceType;
begin
  result := frtImplementationGuide;
end;

procedure TFhirImplementationGuide.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirImplementationGuide(oSource).urlElement.Clone;
  versionElement := TFhirImplementationGuide(oSource).versionElement.Clone;
  nameElement := TFhirImplementationGuide(oSource).nameElement.Clone;
  FStatus := TFhirImplementationGuide(oSource).FStatus.Link;
  experimentalElement := TFhirImplementationGuide(oSource).experimentalElement.Clone;
  publisherElement := TFhirImplementationGuide(oSource).publisherElement.Clone;
  if (TFhirImplementationGuide(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirImplementationGuideContactList.Create;
    FContactList.Assign(TFhirImplementationGuide(oSource).FContactList);
  end;
  dateElement := TFhirImplementationGuide(oSource).dateElement.Clone;
  descriptionElement := TFhirImplementationGuide(oSource).descriptionElement.Clone;
  if (TFhirImplementationGuide(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirImplementationGuide(oSource).FUseContextList);
  end;
  copyrightElement := TFhirImplementationGuide(oSource).copyrightElement.Clone;
  fhirVersionElement := TFhirImplementationGuide(oSource).fhirVersionElement.Clone;
  if (TFhirImplementationGuide(oSource).FDependencyList = nil) then
  begin
    FDependencyList.free;
    FDependencyList := nil;
  end
  else
  begin
    FDependencyList := TFhirImplementationGuideDependencyList.Create;
    FDependencyList.Assign(TFhirImplementationGuide(oSource).FDependencyList);
  end;
  if (TFhirImplementationGuide(oSource).FPackageList = nil) then
  begin
    FPackageList.free;
    FPackageList := nil;
  end
  else
  begin
    FPackageList := TFhirImplementationGuidePackageList.Create;
    FPackageList.Assign(TFhirImplementationGuide(oSource).FPackageList);
  end;
  if (TFhirImplementationGuide(oSource).FGlobalList = nil) then
  begin
    FGlobalList.free;
    FGlobalList := nil;
  end
  else
  begin
    FGlobalList := TFhirImplementationGuideGlobalList.Create;
    FGlobalList.Assign(TFhirImplementationGuide(oSource).FGlobalList);
  end;
  if (TFhirImplementationGuide(oSource).FBinaryList = nil) then
  begin
    FBinaryList.free;
    FBinaryList := nil;
  end
  else
  begin
    FBinaryList := TFhirUriList.Create;
    FBinaryList.Assign(TFhirImplementationGuide(oSource).FBinaryList);
  end;
  page := TFhirImplementationGuide(oSource).page.Clone;
end;

procedure TFhirImplementationGuide.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'fhirVersion') Then
     list.add(FFhirVersion.Link);
  if (child_name = 'dependency') Then
    list.addAll(FDependencyList);
  if (child_name = 'package') Then
    list.addAll(FPackageList);
  if (child_name = 'global') Then
    list.addAll(FGlobalList);
  if (child_name = 'binary') Then
    list.addAll(FBinaryList);
  if (child_name = 'page') Then
     list.add(FPage.Link);
end;

procedure TFhirImplementationGuide.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fhirVersion', 'id', FFhirVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dependency', '', FDependencyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'package', '', FPackageList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'global', '', FGlobalList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'binary', 'uri', FBinaryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'page', '', FPage.Link));{2}
end;

procedure TFhirImplementationGuide.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirImplementationGuideContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'fhirVersion') then FhirVersionElement := propValue as TFhirId{5a}
  else if (propName = 'dependency') then DependencyList.add(propValue as TFhirImplementationGuideDependency){2}
  else if (propName = 'package') then PackageList.add(propValue as TFhirImplementationGuidePackage){2}
  else if (propName = 'global') then GlobalList.add(propValue as TFhirImplementationGuideGlobal){2}
  else if (propName = 'binary') then BinaryList.add(propValue as TFhirUri){2}
  else if (propName = 'page') then Page := propValue as TFhirImplementationGuidePage{4b}
  else inherited;
end;

function TFhirImplementationGuide.FhirType : string;
begin
  result := 'ImplementationGuide';
end;

function TFhirImplementationGuide.Link : TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide(inherited Link);
end;

function TFhirImplementationGuide.Clone : TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide(inherited Clone);
end;

{ TFhirImplementationGuide }

Procedure TFhirImplementationGuide.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirImplementationGuide.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirImplementationGuide.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirImplementationGuide.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirImplementationGuide.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirImplementationGuide.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirImplementationGuide.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirImplementationGuide.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirImplementationGuide.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirImplementationGuide.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirImplementationGuide.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirImplementationGuide.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirImplementationGuide.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirImplementationGuide.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirImplementationGuide.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirImplementationGuide.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirImplementationGuide.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirImplementationGuide.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirImplementationGuide.GetContactList : TFhirImplementationGuideContactList;
begin
  if FContactList = nil then
    FContactList := TFhirImplementationGuideContactList.Create;
  result := FContactList;
end;

Function TFhirImplementationGuide.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirImplementationGuide.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirImplementationGuide.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirImplementationGuide.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirImplementationGuide.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirImplementationGuide.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirImplementationGuide.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirImplementationGuide.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirImplementationGuide.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirImplementationGuide.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirImplementationGuide.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirImplementationGuide.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirImplementationGuide.SetFhirVersion(value : TFhirId);
begin
  FFhirVersion.free;
  FFhirVersion := value;
end;

Function TFhirImplementationGuide.GetFhirVersionST : String;
begin
  if FFhirVersion = nil then
    result := ''
  else
    result := FFhirVersion.value;
end;

Procedure TFhirImplementationGuide.SetFhirVersionST(value : String);
begin
  if value <> '' then
  begin
    if FFhirVersion = nil then
      FFhirVersion := TFhirId.create;
    FFhirVersion.value := value
  end
  else if FFhirVersion <> nil then
    FFhirVersion.value := '';
end;

Function TFhirImplementationGuide.GetDependencyList : TFhirImplementationGuideDependencyList;
begin
  if FDependencyList = nil then
    FDependencyList := TFhirImplementationGuideDependencyList.Create;
  result := FDependencyList;
end;

Function TFhirImplementationGuide.GetHasDependencyList : boolean;
begin
  result := (FDependencyList <> nil) and (FDependencyList.count > 0);
end;

Function TFhirImplementationGuide.GetPackageList : TFhirImplementationGuidePackageList;
begin
  if FPackageList = nil then
    FPackageList := TFhirImplementationGuidePackageList.Create;
  result := FPackageList;
end;

Function TFhirImplementationGuide.GetHasPackageList : boolean;
begin
  result := (FPackageList <> nil) and (FPackageList.count > 0);
end;

Function TFhirImplementationGuide.GetGlobalList : TFhirImplementationGuideGlobalList;
begin
  if FGlobalList = nil then
    FGlobalList := TFhirImplementationGuideGlobalList.Create;
  result := FGlobalList;
end;

Function TFhirImplementationGuide.GetHasGlobalList : boolean;
begin
  result := (FGlobalList <> nil) and (FGlobalList.count > 0);
end;

Function TFhirImplementationGuide.GetBinaryList : TFhirUriList;
begin
  if FBinaryList = nil then
    FBinaryList := TFhirUriList.Create;
  result := FBinaryList;
end;

Function TFhirImplementationGuide.GetHasBinaryList : boolean;
begin
  result := (FBinaryList <> nil) and (FBinaryList.count > 0);
end;

Procedure TFhirImplementationGuide.SetPage(value : TFhirImplementationGuidePage);
begin
  FPage.free;
  FPage := value;
end;


{ TFhirImplementationGuideListEnumerator }

Constructor TFhirImplementationGuideListEnumerator.Create(list : TFhirImplementationGuideList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirImplementationGuideListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirImplementationGuideListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirImplementationGuideListEnumerator.GetCurrent : TFhirImplementationGuide;
begin
  Result := FList[FIndex];
end;


{ TFhirImplementationGuideList }
procedure TFhirImplementationGuideList.AddItem(value: TFhirImplementationGuide);
begin
  assert(value.ClassName = 'TFhirImplementationGuide', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirImplementationGuide');
  add(value);
end;


function TFhirImplementationGuideList.Append: TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideList.ClearItems;
begin
  Clear;
end;

function TFhirImplementationGuideList.GetEnumerator : TFhirImplementationGuideListEnumerator;
begin
  result := TFhirImplementationGuideListEnumerator.Create(self.link);
end;

function TFhirImplementationGuideList.Clone: TFhirImplementationGuideList;
begin
  result := TFhirImplementationGuideList(inherited Clone);
end;

function TFhirImplementationGuideList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirImplementationGuideList.GetItemN(index: Integer): TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide(ObjectByIndex[index]);
end;

function TFhirImplementationGuideList.IndexOf(value: TFhirImplementationGuide): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirImplementationGuideList.Insert(index: Integer): TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirImplementationGuideList.InsertItem(index: Integer; value: TFhirImplementationGuide);
begin
  assert(value is TFhirImplementationGuide);
  Inherited Insert(index, value);
end;

function TFhirImplementationGuideList.Item(index: Integer): TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide(ObjectByIndex[index]);
end;

function TFhirImplementationGuideList.Link: TFhirImplementationGuideList;
begin
  result := TFhirImplementationGuideList(inherited Link);
end;

procedure TFhirImplementationGuideList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirImplementationGuideList.SetItemByIndex(index: Integer; value: TFhirImplementationGuide);
begin
  assert(value is TFhirImplementationGuide);
  FhirImplementationGuides[index] := value;
end;

procedure TFhirImplementationGuideList.SetItemN(index: Integer; value: TFhirImplementationGuide);
begin
  assert(value is TFhirImplementationGuide);
  ObjectByIndex[index] := value;
end;

{ TFhirListEntry }

constructor TFhirListEntry.Create;
begin
  inherited;
end;

destructor TFhirListEntry.Destroy;
begin
  FFlag.free;
  FDeleted.free;
  FDate.free;
  FItem.free;
  inherited;
end;

procedure TFhirListEntry.Assign(oSource : TAdvObject);
begin
  inherited;
  flag := TFhirListEntry(oSource).flag.Clone;
  deletedElement := TFhirListEntry(oSource).deletedElement.Clone;
  dateElement := TFhirListEntry(oSource).dateElement.Clone;
  item := TFhirListEntry(oSource).item.Clone;
end;

procedure TFhirListEntry.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'flag') Then
     list.add(FFlag.Link);
  if (child_name = 'deleted') Then
     list.add(FDeleted.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'item') Then
     list.add(FItem.Link);
end;

procedure TFhirListEntry.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'flag', 'CodeableConcept', FFlag.Link));{2}
  oList.add(TFHIRProperty.create(self, 'deleted', 'boolean', FDeleted.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'item', 'Reference(Any)', FItem.Link));{2}
end;

procedure TFhirListEntry.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'flag') then Flag := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'deleted') then DeletedElement := propValue as TFhirBoolean{5a}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'item') then Item := propValue as TFhirReference{TFhirReference}{4b}
  else inherited;
end;

function TFhirListEntry.FhirType : string;
begin
  result := 'entry';
end;

function TFhirListEntry.Link : TFhirListEntry;
begin
  result := TFhirListEntry(inherited Link);
end;

function TFhirListEntry.Clone : TFhirListEntry;
begin
  result := TFhirListEntry(inherited Clone);
end;

function TFhirListEntry.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirListEntry;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirListEntry)) then
    result := false
  else
  begin
    o := TFhirListEntry(other);
    result := compareDeep(flagElement, o.flagElement, true) and compareDeep(deletedElement, o.deletedElement, true) and 
      compareDeep(dateElement, o.dateElement, true) and compareDeep(itemElement, o.itemElement, true);
  end;
end;

function TFhirListEntry.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirListEntry;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirListEntry)) then
    result := false
  else
  begin
    o := TFhirListEntry(other);
    result := compareValues(deletedElement, o.deletedElement, true) and compareValues(dateElement, o.dateElement, true);
  end;
end;

{ TFhirListEntry }

Procedure TFhirListEntry.SetFlag(value : TFhirCodeableConcept);
begin
  FFlag.free;
  FFlag := value;
end;

Procedure TFhirListEntry.SetDeleted(value : TFhirBoolean);
begin
  FDeleted.free;
  FDeleted := value;
end;

Function TFhirListEntry.GetDeletedST : Boolean;
begin
  if FDeleted = nil then
    result := false
  else
    result := FDeleted.value;
end;

Procedure TFhirListEntry.SetDeletedST(value : Boolean);
begin
  if FDeleted = nil then
    FDeleted := TFhirBoolean.create;
  FDeleted.value := value
end;

Procedure TFhirListEntry.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirListEntry.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirListEntry.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirListEntry.SetItem(value : TFhirReference{TFhirReference});
begin
  FItem.free;
  FItem := value;
end;


{ TFhirListEntryListEnumerator }

Constructor TFhirListEntryListEnumerator.Create(list : TFhirListEntryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirListEntryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirListEntryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirListEntryListEnumerator.GetCurrent : TFhirListEntry;
begin
  Result := FList[FIndex];
end;


{ TFhirListEntryList }
procedure TFhirListEntryList.AddItem(value: TFhirListEntry);
begin
  assert(value.ClassName = 'TFhirListEntry', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirListEntry');
  add(value);
end;


function TFhirListEntryList.Append: TFhirListEntry;
begin
  result := TFhirListEntry.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirListEntryList.ClearItems;
begin
  Clear;
end;

function TFhirListEntryList.GetEnumerator : TFhirListEntryListEnumerator;
begin
  result := TFhirListEntryListEnumerator.Create(self.link);
end;

function TFhirListEntryList.Clone: TFhirListEntryList;
begin
  result := TFhirListEntryList(inherited Clone);
end;

function TFhirListEntryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirListEntryList.GetItemN(index: Integer): TFhirListEntry;
begin
  result := TFhirListEntry(ObjectByIndex[index]);
end;

function TFhirListEntryList.IndexOf(value: TFhirListEntry): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirListEntryList.Insert(index: Integer): TFhirListEntry;
begin
  result := TFhirListEntry.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirListEntryList.InsertItem(index: Integer; value: TFhirListEntry);
begin
  assert(value is TFhirListEntry);
  Inherited Insert(index, value);
end;

function TFhirListEntryList.Item(index: Integer): TFhirListEntry;
begin
  result := TFhirListEntry(ObjectByIndex[index]);
end;

function TFhirListEntryList.Link: TFhirListEntryList;
begin
  result := TFhirListEntryList(inherited Link);
end;

procedure TFhirListEntryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirListEntryList.SetItemByIndex(index: Integer; value: TFhirListEntry);
begin
  assert(value is TFhirListEntry);
  FhirListEntries[index] := value;
end;

procedure TFhirListEntryList.SetItemN(index: Integer; value: TFhirListEntry);
begin
  assert(value is TFhirListEntry);
  ObjectByIndex[index] := value;
end;

{ TFhirList }

constructor TFhirList.Create;
begin
  inherited;
end;

destructor TFhirList.Destroy;
begin
  FIdentifierList.Free;
  FTitle.free;
  FCode.free;
  FSubject.free;
  FSource.free;
  FEncounter.free;
  FStatus.free;
  FDate.free;
  FOrderedBy.free;
  FMode.free;
  FNote.free;
  FEntryList.Free;
  FEmptyReason.free;
  inherited;
end;

function TFhirList.GetResourceType : TFhirResourceType;
begin
  result := frtList;
end;

procedure TFhirList.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirList(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirList(oSource).FIdentifierList);
  end;
  titleElement := TFhirList(oSource).titleElement.Clone;
  code := TFhirList(oSource).code.Clone;
  subject := TFhirList(oSource).subject.Clone;
  source := TFhirList(oSource).source.Clone;
  encounter := TFhirList(oSource).encounter.Clone;
  FStatus := TFhirList(oSource).FStatus.Link;
  dateElement := TFhirList(oSource).dateElement.Clone;
  orderedBy := TFhirList(oSource).orderedBy.Clone;
  FMode := TFhirList(oSource).FMode.Link;
  noteElement := TFhirList(oSource).noteElement.Clone;
  if (TFhirList(oSource).FEntryList = nil) then
  begin
    FEntryList.free;
    FEntryList := nil;
  end
  else
  begin
    FEntryList := TFhirListEntryList.Create;
    FEntryList.Assign(TFhirList(oSource).FEntryList);
  end;
  emptyReason := TFhirList(oSource).emptyReason.Clone;
end;

procedure TFhirList.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'orderedBy') Then
     list.add(FOrderedBy.Link);
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'entry') Then
    list.addAll(FEntryList);
  if (child_name = 'emptyReason') Then
     list.add(FEmptyReason.Link);
end;

procedure TFhirList.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Device|Location)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Practitioner|Patient|Device)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderedBy', 'CodeableConcept', FOrderedBy.Link));{2}
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'note', 'string', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'entry', '', FEntryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'emptyReason', 'CodeableConcept', FEmptyReason.Link));{2}
end;

procedure TFhirList.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'source') then Source := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'orderedBy') then OrderedBy := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'note') then NoteElement := propValue as TFhirString{5a}
  else if (propName = 'entry') then EntryList.add(propValue as TFhirListEntry){2}
  else if (propName = 'emptyReason') then EmptyReason := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirList.FhirType : string;
begin
  result := 'List';
end;

function TFhirList.Link : TFhirList;
begin
  result := TFhirList(inherited Link);
end;

function TFhirList.Clone : TFhirList;
begin
  result := TFhirList(inherited Clone);
end;

{ TFhirList }

Function TFhirList.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirList.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirList.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirList.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirList.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Procedure TFhirList.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirList.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirList.SetSource(value : TFhirReference{Resource});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirList.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirList.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirList.GetStatusST : TFhirListStatus;
begin
  if FStatus = nil then
    result := TFhirListStatus(0)
  else
    result := TFhirListStatus(StringArrayIndexOfSensitive(CODES_TFhirListStatus, FStatus.value));
end;

Procedure TFhirList.SetStatusST(value : TFhirListStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirListStatus[value]);
end;

Procedure TFhirList.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirList.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirList.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirList.SetOrderedBy(value : TFhirCodeableConcept);
begin
  FOrderedBy.free;
  FOrderedBy := value;
end;

Procedure TFhirList.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirList.GetModeST : TFhirListMode;
begin
  if FMode = nil then
    result := TFhirListMode(0)
  else
    result := TFhirListMode(StringArrayIndexOfSensitive(CODES_TFhirListMode, FMode.value));
end;

Procedure TFhirList.SetModeST(value : TFhirListMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirListMode[value]);
end;

Procedure TFhirList.SetNote(value : TFhirString);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirList.GetNoteST : String;
begin
  if FNote = nil then
    result := ''
  else
    result := FNote.value;
end;

Procedure TFhirList.SetNoteST(value : String);
begin
  if value <> '' then
  begin
    if FNote = nil then
      FNote := TFhirString.create;
    FNote.value := value
  end
  else if FNote <> nil then
    FNote.value := '';
end;

Function TFhirList.GetEntryList : TFhirListEntryList;
begin
  if FEntryList = nil then
    FEntryList := TFhirListEntryList.Create;
  result := FEntryList;
end;

Function TFhirList.GetHasEntryList : boolean;
begin
  result := (FEntryList <> nil) and (FEntryList.count > 0);
end;

Procedure TFhirList.SetEmptyReason(value : TFhirCodeableConcept);
begin
  FEmptyReason.free;
  FEmptyReason := value;
end;


{ TFhirListListEnumerator }

Constructor TFhirListListEnumerator.Create(list : TFhirListList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirListListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirListListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirListListEnumerator.GetCurrent : TFhirList;
begin
  Result := FList[FIndex];
end;


{ TFhirListList }
procedure TFhirListList.AddItem(value: TFhirList);
begin
  assert(value.ClassName = 'TFhirList', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirList');
  add(value);
end;


function TFhirListList.Append: TFhirList;
begin
  result := TFhirList.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirListList.ClearItems;
begin
  Clear;
end;

function TFhirListList.GetEnumerator : TFhirListListEnumerator;
begin
  result := TFhirListListEnumerator.Create(self.link);
end;

function TFhirListList.Clone: TFhirListList;
begin
  result := TFhirListList(inherited Clone);
end;

function TFhirListList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirListList.GetItemN(index: Integer): TFhirList;
begin
  result := TFhirList(ObjectByIndex[index]);
end;

function TFhirListList.IndexOf(value: TFhirList): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirListList.Insert(index: Integer): TFhirList;
begin
  result := TFhirList.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirListList.InsertItem(index: Integer; value: TFhirList);
begin
  assert(value is TFhirList);
  Inherited Insert(index, value);
end;

function TFhirListList.Item(index: Integer): TFhirList;
begin
  result := TFhirList(ObjectByIndex[index]);
end;

function TFhirListList.Link: TFhirListList;
begin
  result := TFhirListList(inherited Link);
end;

procedure TFhirListList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirListList.SetItemByIndex(index: Integer; value: TFhirList);
begin
  assert(value is TFhirList);
  FhirLists[index] := value;
end;

procedure TFhirListList.SetItemN(index: Integer; value: TFhirList);
begin
  assert(value is TFhirList);
  ObjectByIndex[index] := value;
end;

{ TFhirLocationPosition }

constructor TFhirLocationPosition.Create;
begin
  inherited;
end;

destructor TFhirLocationPosition.Destroy;
begin
  FLongitude.free;
  FLatitude.free;
  FAltitude.free;
  inherited;
end;

procedure TFhirLocationPosition.Assign(oSource : TAdvObject);
begin
  inherited;
  longitudeElement := TFhirLocationPosition(oSource).longitudeElement.Clone;
  latitudeElement := TFhirLocationPosition(oSource).latitudeElement.Clone;
  altitudeElement := TFhirLocationPosition(oSource).altitudeElement.Clone;
end;

procedure TFhirLocationPosition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'longitude') Then
     list.add(FLongitude.Link);
  if (child_name = 'latitude') Then
     list.add(FLatitude.Link);
  if (child_name = 'altitude') Then
     list.add(FAltitude.Link);
end;

procedure TFhirLocationPosition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'longitude', 'decimal', FLongitude.Link));{2}
  oList.add(TFHIRProperty.create(self, 'latitude', 'decimal', FLatitude.Link));{2}
  oList.add(TFHIRProperty.create(self, 'altitude', 'decimal', FAltitude.Link));{2}
end;

procedure TFhirLocationPosition.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'longitude') then LongitudeElement := propValue as TFhirDecimal{5a}
  else if (propName = 'latitude') then LatitudeElement := propValue as TFhirDecimal{5a}
  else if (propName = 'altitude') then AltitudeElement := propValue as TFhirDecimal{5a}
  else inherited;
end;

function TFhirLocationPosition.FhirType : string;
begin
  result := 'position';
end;

function TFhirLocationPosition.Link : TFhirLocationPosition;
begin
  result := TFhirLocationPosition(inherited Link);
end;

function TFhirLocationPosition.Clone : TFhirLocationPosition;
begin
  result := TFhirLocationPosition(inherited Clone);
end;

function TFhirLocationPosition.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirLocationPosition;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirLocationPosition)) then
    result := false
  else
  begin
    o := TFhirLocationPosition(other);
    result := compareDeep(longitudeElement, o.longitudeElement, true) and compareDeep(latitudeElement, o.latitudeElement, true) and 
      compareDeep(altitudeElement, o.altitudeElement, true);
  end;
end;

function TFhirLocationPosition.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirLocationPosition;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirLocationPosition)) then
    result := false
  else
  begin
    o := TFhirLocationPosition(other);
    result := compareValues(longitudeElement, o.longitudeElement, true) and compareValues(latitudeElement, o.latitudeElement, true) and 
      compareValues(altitudeElement, o.altitudeElement, true);
  end;
end;

{ TFhirLocationPosition }

Procedure TFhirLocationPosition.SetLongitude(value : TFhirDecimal);
begin
  FLongitude.free;
  FLongitude := value;
end;

Function TFhirLocationPosition.GetLongitudeST : String;
begin
  if FLongitude = nil then
    result := ''
  else
    result := FLongitude.value;
end;

Procedure TFhirLocationPosition.SetLongitudeST(value : String);
begin
  if value <> '' then
  begin
    if FLongitude = nil then
      FLongitude := TFhirDecimal.create;
    FLongitude.value := value
  end
  else if FLongitude <> nil then
    FLongitude.value := '';
end;

Procedure TFhirLocationPosition.SetLatitude(value : TFhirDecimal);
begin
  FLatitude.free;
  FLatitude := value;
end;

Function TFhirLocationPosition.GetLatitudeST : String;
begin
  if FLatitude = nil then
    result := ''
  else
    result := FLatitude.value;
end;

Procedure TFhirLocationPosition.SetLatitudeST(value : String);
begin
  if value <> '' then
  begin
    if FLatitude = nil then
      FLatitude := TFhirDecimal.create;
    FLatitude.value := value
  end
  else if FLatitude <> nil then
    FLatitude.value := '';
end;

Procedure TFhirLocationPosition.SetAltitude(value : TFhirDecimal);
begin
  FAltitude.free;
  FAltitude := value;
end;

Function TFhirLocationPosition.GetAltitudeST : String;
begin
  if FAltitude = nil then
    result := ''
  else
    result := FAltitude.value;
end;

Procedure TFhirLocationPosition.SetAltitudeST(value : String);
begin
  if value <> '' then
  begin
    if FAltitude = nil then
      FAltitude := TFhirDecimal.create;
    FAltitude.value := value
  end
  else if FAltitude <> nil then
    FAltitude.value := '';
end;


{ TFhirLocationPositionListEnumerator }

Constructor TFhirLocationPositionListEnumerator.Create(list : TFhirLocationPositionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirLocationPositionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirLocationPositionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirLocationPositionListEnumerator.GetCurrent : TFhirLocationPosition;
begin
  Result := FList[FIndex];
end;


{ TFhirLocationPositionList }
procedure TFhirLocationPositionList.AddItem(value: TFhirLocationPosition);
begin
  assert(value.ClassName = 'TFhirLocationPosition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirLocationPosition');
  add(value);
end;


function TFhirLocationPositionList.Append: TFhirLocationPosition;
begin
  result := TFhirLocationPosition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirLocationPositionList.ClearItems;
begin
  Clear;
end;

function TFhirLocationPositionList.GetEnumerator : TFhirLocationPositionListEnumerator;
begin
  result := TFhirLocationPositionListEnumerator.Create(self.link);
end;

function TFhirLocationPositionList.Clone: TFhirLocationPositionList;
begin
  result := TFhirLocationPositionList(inherited Clone);
end;

function TFhirLocationPositionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirLocationPositionList.GetItemN(index: Integer): TFhirLocationPosition;
begin
  result := TFhirLocationPosition(ObjectByIndex[index]);
end;

function TFhirLocationPositionList.IndexOf(value: TFhirLocationPosition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirLocationPositionList.Insert(index: Integer): TFhirLocationPosition;
begin
  result := TFhirLocationPosition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirLocationPositionList.InsertItem(index: Integer; value: TFhirLocationPosition);
begin
  assert(value is TFhirLocationPosition);
  Inherited Insert(index, value);
end;

function TFhirLocationPositionList.Item(index: Integer): TFhirLocationPosition;
begin
  result := TFhirLocationPosition(ObjectByIndex[index]);
end;

function TFhirLocationPositionList.Link: TFhirLocationPositionList;
begin
  result := TFhirLocationPositionList(inherited Link);
end;

procedure TFhirLocationPositionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirLocationPositionList.SetItemByIndex(index: Integer; value: TFhirLocationPosition);
begin
  assert(value is TFhirLocationPosition);
  FhirLocationPositions[index] := value;
end;

procedure TFhirLocationPositionList.SetItemN(index: Integer; value: TFhirLocationPosition);
begin
  assert(value is TFhirLocationPosition);
  ObjectByIndex[index] := value;
end;

{ TFhirLocation }

constructor TFhirLocation.Create;
begin
  inherited;
end;

destructor TFhirLocation.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FName.free;
  FDescription.free;
  FMode.free;
  FType_.free;
  FTelecomList.Free;
  FAddress.free;
  FPhysicalType.free;
  FPosition.free;
  FManagingOrganization.free;
  FPartOf.free;
  inherited;
end;

function TFhirLocation.GetResourceType : TFhirResourceType;
begin
  result := frtLocation;
end;

procedure TFhirLocation.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirLocation(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirLocation(oSource).FIdentifierList);
  end;
  FStatus := TFhirLocation(oSource).FStatus.Link;
  nameElement := TFhirLocation(oSource).nameElement.Clone;
  descriptionElement := TFhirLocation(oSource).descriptionElement.Clone;
  FMode := TFhirLocation(oSource).FMode.Link;
  type_ := TFhirLocation(oSource).type_.Clone;
  if (TFhirLocation(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirLocation(oSource).FTelecomList);
  end;
  address := TFhirLocation(oSource).address.Clone;
  physicalType := TFhirLocation(oSource).physicalType.Clone;
  position := TFhirLocation(oSource).position.Clone;
  managingOrganization := TFhirLocation(oSource).managingOrganization.Clone;
  partOf := TFhirLocation(oSource).partOf.Clone;
end;

procedure TFhirLocation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'mode') Then
     list.add(FMode.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'address') Then
     list.add(FAddress.Link);
  if (child_name = 'physicalType') Then
     list.add(FPhysicalType.Link);
  if (child_name = 'position') Then
     list.add(FPosition.Link);
  if (child_name = 'managingOrganization') Then
     list.add(FManagingOrganization.Link);
  if (child_name = 'partOf') Then
     list.add(FPartOf.Link);
end;

procedure TFhirLocation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'mode', 'code', FMode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddress.Link));{2}
  oList.add(TFHIRProperty.create(self, 'physicalType', 'CodeableConcept', FPhysicalType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'position', '', FPosition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'managingOrganization', 'Reference(Organization)', FManagingOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'partOf', 'Reference(Location)', FPartOf.Link));{2}
end;

procedure TFhirLocation.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'mode') then ModeElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'address') then Address := propValue as TFhirAddress{4b}
  else if (propName = 'physicalType') then PhysicalType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'position') then Position := propValue as TFhirLocationPosition{4b}
  else if (propName = 'managingOrganization') then ManagingOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'partOf') then PartOf := propValue as TFhirReference{TFhirLocation}{4b}
  else inherited;
end;

function TFhirLocation.FhirType : string;
begin
  result := 'Location';
end;

function TFhirLocation.Link : TFhirLocation;
begin
  result := TFhirLocation(inherited Link);
end;

function TFhirLocation.Clone : TFhirLocation;
begin
  result := TFhirLocation(inherited Clone);
end;

{ TFhirLocation }

Function TFhirLocation.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirLocation.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirLocation.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirLocation.GetStatusST : TFhirLocationStatus;
begin
  if FStatus = nil then
    result := TFhirLocationStatus(0)
  else
    result := TFhirLocationStatus(StringArrayIndexOfSensitive(CODES_TFhirLocationStatus, FStatus.value));
end;

Procedure TFhirLocation.SetStatusST(value : TFhirLocationStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirLocationStatus[value]);
end;

Procedure TFhirLocation.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirLocation.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirLocation.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirLocation.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirLocation.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirLocation.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirLocation.SetMode(value : TFhirEnum);
begin
  FMode.free;
  FMode := value;
end;

Function TFhirLocation.GetModeST : TFhirLocationMode;
begin
  if FMode = nil then
    result := TFhirLocationMode(0)
  else
    result := TFhirLocationMode(StringArrayIndexOfSensitive(CODES_TFhirLocationMode, FMode.value));
end;

Procedure TFhirLocation.SetModeST(value : TFhirLocationMode);
begin
  if ord(value) = 0 then
    ModeElement := nil
  else
    ModeElement := TFhirEnum.create(CODES_TFhirLocationMode[value]);
end;

Procedure TFhirLocation.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirLocation.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirLocation.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirLocation.SetAddress(value : TFhirAddress);
begin
  FAddress.free;
  FAddress := value;
end;

Procedure TFhirLocation.SetPhysicalType(value : TFhirCodeableConcept);
begin
  FPhysicalType.free;
  FPhysicalType := value;
end;

Procedure TFhirLocation.SetPosition(value : TFhirLocationPosition);
begin
  FPosition.free;
  FPosition := value;
end;

Procedure TFhirLocation.SetManagingOrganization(value : TFhirReference{TFhirOrganization});
begin
  FManagingOrganization.free;
  FManagingOrganization := value;
end;

Procedure TFhirLocation.SetPartOf(value : TFhirReference{TFhirLocation});
begin
  FPartOf.free;
  FPartOf := value;
end;


{ TFhirLocationListEnumerator }

Constructor TFhirLocationListEnumerator.Create(list : TFhirLocationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirLocationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirLocationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirLocationListEnumerator.GetCurrent : TFhirLocation;
begin
  Result := FList[FIndex];
end;


{ TFhirLocationList }
procedure TFhirLocationList.AddItem(value: TFhirLocation);
begin
  assert(value.ClassName = 'TFhirLocation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirLocation');
  add(value);
end;


function TFhirLocationList.Append: TFhirLocation;
begin
  result := TFhirLocation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirLocationList.ClearItems;
begin
  Clear;
end;

function TFhirLocationList.GetEnumerator : TFhirLocationListEnumerator;
begin
  result := TFhirLocationListEnumerator.Create(self.link);
end;

function TFhirLocationList.Clone: TFhirLocationList;
begin
  result := TFhirLocationList(inherited Clone);
end;

function TFhirLocationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirLocationList.GetItemN(index: Integer): TFhirLocation;
begin
  result := TFhirLocation(ObjectByIndex[index]);
end;

function TFhirLocationList.IndexOf(value: TFhirLocation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirLocationList.Insert(index: Integer): TFhirLocation;
begin
  result := TFhirLocation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirLocationList.InsertItem(index: Integer; value: TFhirLocation);
begin
  assert(value is TFhirLocation);
  Inherited Insert(index, value);
end;

function TFhirLocationList.Item(index: Integer): TFhirLocation;
begin
  result := TFhirLocation(ObjectByIndex[index]);
end;

function TFhirLocationList.Link: TFhirLocationList;
begin
  result := TFhirLocationList(inherited Link);
end;

procedure TFhirLocationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirLocationList.SetItemByIndex(index: Integer; value: TFhirLocation);
begin
  assert(value is TFhirLocation);
  FhirLocations[index] := value;
end;

procedure TFhirLocationList.SetItemN(index: Integer; value: TFhirLocation);
begin
  assert(value is TFhirLocation);
  ObjectByIndex[index] := value;
end;

{ TFhirMedia }

constructor TFhirMedia.Create;
begin
  inherited;
end;

destructor TFhirMedia.Destroy;
begin
  FType_.free;
  FSubtype.free;
  FIdentifierList.Free;
  FSubject.free;
  FOperator.free;
  FView.free;
  FDeviceName.free;
  FHeight.free;
  FWidth.free;
  FFrames.free;
  FDuration.free;
  FContent.free;
  inherited;
end;

function TFhirMedia.GetResourceType : TFhirResourceType;
begin
  result := frtMedia;
end;

procedure TFhirMedia.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirMedia(oSource).FType_.Link;
  subtype := TFhirMedia(oSource).subtype.Clone;
  if (TFhirMedia(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirMedia(oSource).FIdentifierList);
  end;
  subject := TFhirMedia(oSource).subject.Clone;
  operator := TFhirMedia(oSource).operator.Clone;
  view := TFhirMedia(oSource).view.Clone;
  deviceNameElement := TFhirMedia(oSource).deviceNameElement.Clone;
  heightElement := TFhirMedia(oSource).heightElement.Clone;
  widthElement := TFhirMedia(oSource).widthElement.Clone;
  framesElement := TFhirMedia(oSource).framesElement.Clone;
  durationElement := TFhirMedia(oSource).durationElement.Clone;
  content := TFhirMedia(oSource).content.Clone;
end;

procedure TFhirMedia.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'subtype') Then
     list.add(FSubtype.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'operator') Then
     list.add(FOperator.Link);
  if (child_name = 'view') Then
     list.add(FView.Link);
  if (child_name = 'deviceName') Then
     list.add(FDeviceName.Link);
  if (child_name = 'height') Then
     list.add(FHeight.Link);
  if (child_name = 'width') Then
     list.add(FWidth.Link);
  if (child_name = 'frames') Then
     list.add(FFrames.Link);
  if (child_name = 'duration') Then
     list.add(FDuration.Link);
  if (child_name = 'content') Then
     list.add(FContent.Link);
end;

procedure TFhirMedia.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'subtype', 'CodeableConcept', FSubtype.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Practitioner|Group|Device|Specimen)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'operator', 'Reference(Practitioner)', FOperator.Link));{2}
  oList.add(TFHIRProperty.create(self, 'view', 'CodeableConcept', FView.Link));{2}
  oList.add(TFHIRProperty.create(self, 'deviceName', 'string', FDeviceName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'height', 'positiveInt', FHeight.Link));{2}
  oList.add(TFHIRProperty.create(self, 'width', 'positiveInt', FWidth.Link));{2}
  oList.add(TFHIRProperty.create(self, 'frames', 'positiveInt', FFrames.Link));{2}
  oList.add(TFHIRProperty.create(self, 'duration', 'unsignedInt', FDuration.Link));{2}
  oList.add(TFHIRProperty.create(self, 'content', 'Attachment', FContent.Link));{2}
end;

procedure TFhirMedia.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'subtype') then Subtype := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'operator') then Operator := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'view') then View := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'deviceName') then DeviceNameElement := propValue as TFhirString{5a}
  else if (propName = 'height') then HeightElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'width') then WidthElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'frames') then FramesElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'duration') then DurationElement := propValue as TFhirUnsignedInt{5a}
  else if (propName = 'content') then Content := propValue as TFhirAttachment{4b}
  else inherited;
end;

function TFhirMedia.FhirType : string;
begin
  result := 'Media';
end;

function TFhirMedia.Link : TFhirMedia;
begin
  result := TFhirMedia(inherited Link);
end;

function TFhirMedia.Clone : TFhirMedia;
begin
  result := TFhirMedia(inherited Clone);
end;

{ TFhirMedia }

Procedure TFhirMedia.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirMedia.GetType_ST : TFhirDigitalMediaType;
begin
  if FType_ = nil then
    result := TFhirDigitalMediaType(0)
  else
    result := TFhirDigitalMediaType(StringArrayIndexOfSensitive(CODES_TFhirDigitalMediaType, FType_.value));
end;

Procedure TFhirMedia.SetType_ST(value : TFhirDigitalMediaType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirDigitalMediaType[value]);
end;

Procedure TFhirMedia.SetSubtype(value : TFhirCodeableConcept);
begin
  FSubtype.free;
  FSubtype := value;
end;

Function TFhirMedia.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirMedia.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirMedia.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirMedia.SetOperator(value : TFhirReference{TFhirPractitioner});
begin
  FOperator.free;
  FOperator := value;
end;

Procedure TFhirMedia.SetView(value : TFhirCodeableConcept);
begin
  FView.free;
  FView := value;
end;

Procedure TFhirMedia.SetDeviceName(value : TFhirString);
begin
  FDeviceName.free;
  FDeviceName := value;
end;

Function TFhirMedia.GetDeviceNameST : String;
begin
  if FDeviceName = nil then
    result := ''
  else
    result := FDeviceName.value;
end;

Procedure TFhirMedia.SetDeviceNameST(value : String);
begin
  if value <> '' then
  begin
    if FDeviceName = nil then
      FDeviceName := TFhirString.create;
    FDeviceName.value := value
  end
  else if FDeviceName <> nil then
    FDeviceName.value := '';
end;

Procedure TFhirMedia.SetHeight(value : TFhirPositiveInt);
begin
  FHeight.free;
  FHeight := value;
end;

Function TFhirMedia.GetHeightST : String;
begin
  if FHeight = nil then
    result := ''
  else
    result := FHeight.value;
end;

Procedure TFhirMedia.SetHeightST(value : String);
begin
  if value <> '' then
  begin
    if FHeight = nil then
      FHeight := TFhirPositiveInt.create;
    FHeight.value := value
  end
  else if FHeight <> nil then
    FHeight.value := '';
end;

Procedure TFhirMedia.SetWidth(value : TFhirPositiveInt);
begin
  FWidth.free;
  FWidth := value;
end;

Function TFhirMedia.GetWidthST : String;
begin
  if FWidth = nil then
    result := ''
  else
    result := FWidth.value;
end;

Procedure TFhirMedia.SetWidthST(value : String);
begin
  if value <> '' then
  begin
    if FWidth = nil then
      FWidth := TFhirPositiveInt.create;
    FWidth.value := value
  end
  else if FWidth <> nil then
    FWidth.value := '';
end;

Procedure TFhirMedia.SetFrames(value : TFhirPositiveInt);
begin
  FFrames.free;
  FFrames := value;
end;

Function TFhirMedia.GetFramesST : String;
begin
  if FFrames = nil then
    result := ''
  else
    result := FFrames.value;
end;

Procedure TFhirMedia.SetFramesST(value : String);
begin
  if value <> '' then
  begin
    if FFrames = nil then
      FFrames := TFhirPositiveInt.create;
    FFrames.value := value
  end
  else if FFrames <> nil then
    FFrames.value := '';
end;

Procedure TFhirMedia.SetDuration(value : TFhirUnsignedInt);
begin
  FDuration.free;
  FDuration := value;
end;

Function TFhirMedia.GetDurationST : String;
begin
  if FDuration = nil then
    result := ''
  else
    result := FDuration.value;
end;

Procedure TFhirMedia.SetDurationST(value : String);
begin
  if value <> '' then
  begin
    if FDuration = nil then
      FDuration := TFhirUnsignedInt.create;
    FDuration.value := value
  end
  else if FDuration <> nil then
    FDuration.value := '';
end;

Procedure TFhirMedia.SetContent(value : TFhirAttachment);
begin
  FContent.free;
  FContent := value;
end;


{ TFhirMediaListEnumerator }

Constructor TFhirMediaListEnumerator.Create(list : TFhirMediaList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMediaListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMediaListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMediaListEnumerator.GetCurrent : TFhirMedia;
begin
  Result := FList[FIndex];
end;


{ TFhirMediaList }
procedure TFhirMediaList.AddItem(value: TFhirMedia);
begin
  assert(value.ClassName = 'TFhirMedia', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedia');
  add(value);
end;


function TFhirMediaList.Append: TFhirMedia;
begin
  result := TFhirMedia.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMediaList.ClearItems;
begin
  Clear;
end;

function TFhirMediaList.GetEnumerator : TFhirMediaListEnumerator;
begin
  result := TFhirMediaListEnumerator.Create(self.link);
end;

function TFhirMediaList.Clone: TFhirMediaList;
begin
  result := TFhirMediaList(inherited Clone);
end;

function TFhirMediaList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMediaList.GetItemN(index: Integer): TFhirMedia;
begin
  result := TFhirMedia(ObjectByIndex[index]);
end;

function TFhirMediaList.IndexOf(value: TFhirMedia): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMediaList.Insert(index: Integer): TFhirMedia;
begin
  result := TFhirMedia.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMediaList.InsertItem(index: Integer; value: TFhirMedia);
begin
  assert(value is TFhirMedia);
  Inherited Insert(index, value);
end;

function TFhirMediaList.Item(index: Integer): TFhirMedia;
begin
  result := TFhirMedia(ObjectByIndex[index]);
end;

function TFhirMediaList.Link: TFhirMediaList;
begin
  result := TFhirMediaList(inherited Link);
end;

procedure TFhirMediaList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMediaList.SetItemByIndex(index: Integer; value: TFhirMedia);
begin
  assert(value is TFhirMedia);
  FhirMedia[index] := value;
end;

procedure TFhirMediaList.SetItemN(index: Integer; value: TFhirMedia);
begin
  assert(value is TFhirMedia);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationProduct }

constructor TFhirMedicationProduct.Create;
begin
  inherited;
end;

destructor TFhirMedicationProduct.Destroy;
begin
  FForm.free;
  FIngredientList.Free;
  FBatchList.Free;
  inherited;
end;

procedure TFhirMedicationProduct.Assign(oSource : TAdvObject);
begin
  inherited;
  form := TFhirMedicationProduct(oSource).form.Clone;
  if (TFhirMedicationProduct(oSource).FIngredientList = nil) then
  begin
    FIngredientList.free;
    FIngredientList := nil;
  end
  else
  begin
    FIngredientList := TFhirMedicationProductIngredientList.Create;
    FIngredientList.Assign(TFhirMedicationProduct(oSource).FIngredientList);
  end;
  if (TFhirMedicationProduct(oSource).FBatchList = nil) then
  begin
    FBatchList.free;
    FBatchList := nil;
  end
  else
  begin
    FBatchList := TFhirMedicationProductBatchList.Create;
    FBatchList.Assign(TFhirMedicationProduct(oSource).FBatchList);
  end;
end;

procedure TFhirMedicationProduct.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'form') Then
     list.add(FForm.Link);
  if (child_name = 'ingredient') Then
    list.addAll(FIngredientList);
  if (child_name = 'batch') Then
    list.addAll(FBatchList);
end;

procedure TFhirMedicationProduct.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'form', 'CodeableConcept', FForm.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ingredient', '', FIngredientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'batch', '', FBatchList.Link)){3};
end;

procedure TFhirMedicationProduct.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'form') then Form := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'ingredient') then IngredientList.add(propValue as TFhirMedicationProductIngredient){2}
  else if (propName = 'batch') then BatchList.add(propValue as TFhirMedicationProductBatch){2}
  else inherited;
end;

function TFhirMedicationProduct.FhirType : string;
begin
  result := 'product';
end;

function TFhirMedicationProduct.Link : TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct(inherited Link);
end;

function TFhirMedicationProduct.Clone : TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct(inherited Clone);
end;

function TFhirMedicationProduct.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProduct;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationProduct)) then
    result := false
  else
  begin
    o := TFhirMedicationProduct(other);
    result := compareDeep(formElement, o.formElement, true) and compareDeep(ingredientList, o.ingredientList, true) and 
      compareDeep(batchList, o.batchList, true);
  end;
end;

function TFhirMedicationProduct.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProduct;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationProduct)) then
    result := false
  else
  begin
    o := TFhirMedicationProduct(other);
    result := true;
  end;
end;

{ TFhirMedicationProduct }

Procedure TFhirMedicationProduct.SetForm(value : TFhirCodeableConcept);
begin
  FForm.free;
  FForm := value;
end;

Function TFhirMedicationProduct.GetIngredientList : TFhirMedicationProductIngredientList;
begin
  if FIngredientList = nil then
    FIngredientList := TFhirMedicationProductIngredientList.Create;
  result := FIngredientList;
end;

Function TFhirMedicationProduct.GetHasIngredientList : boolean;
begin
  result := (FIngredientList <> nil) and (FIngredientList.count > 0);
end;

Function TFhirMedicationProduct.GetBatchList : TFhirMedicationProductBatchList;
begin
  if FBatchList = nil then
    FBatchList := TFhirMedicationProductBatchList.Create;
  result := FBatchList;
end;

Function TFhirMedicationProduct.GetHasBatchList : boolean;
begin
  result := (FBatchList <> nil) and (FBatchList.count > 0);
end;


{ TFhirMedicationProductListEnumerator }

Constructor TFhirMedicationProductListEnumerator.Create(list : TFhirMedicationProductList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationProductListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationProductListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationProductListEnumerator.GetCurrent : TFhirMedicationProduct;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationProductList }
procedure TFhirMedicationProductList.AddItem(value: TFhirMedicationProduct);
begin
  assert(value.ClassName = 'TFhirMedicationProduct', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationProduct');
  add(value);
end;


function TFhirMedicationProductList.Append: TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationProductList.GetEnumerator : TFhirMedicationProductListEnumerator;
begin
  result := TFhirMedicationProductListEnumerator.Create(self.link);
end;

function TFhirMedicationProductList.Clone: TFhirMedicationProductList;
begin
  result := TFhirMedicationProductList(inherited Clone);
end;

function TFhirMedicationProductList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationProductList.GetItemN(index: Integer): TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct(ObjectByIndex[index]);
end;

function TFhirMedicationProductList.IndexOf(value: TFhirMedicationProduct): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationProductList.Insert(index: Integer): TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductList.InsertItem(index: Integer; value: TFhirMedicationProduct);
begin
  assert(value is TFhirMedicationProduct);
  Inherited Insert(index, value);
end;

function TFhirMedicationProductList.Item(index: Integer): TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct(ObjectByIndex[index]);
end;

function TFhirMedicationProductList.Link: TFhirMedicationProductList;
begin
  result := TFhirMedicationProductList(inherited Link);
end;

procedure TFhirMedicationProductList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationProductList.SetItemByIndex(index: Integer; value: TFhirMedicationProduct);
begin
  assert(value is TFhirMedicationProduct);
  FhirMedicationProducts[index] := value;
end;

procedure TFhirMedicationProductList.SetItemN(index: Integer; value: TFhirMedicationProduct);
begin
  assert(value is TFhirMedicationProduct);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationProductIngredient }

constructor TFhirMedicationProductIngredient.Create;
begin
  inherited;
end;

destructor TFhirMedicationProductIngredient.Destroy;
begin
  FItem.free;
  FAmount.free;
  inherited;
end;

procedure TFhirMedicationProductIngredient.Assign(oSource : TAdvObject);
begin
  inherited;
  item := TFhirMedicationProductIngredient(oSource).item.Clone;
  amount := TFhirMedicationProductIngredient(oSource).amount.Clone;
end;

procedure TFhirMedicationProductIngredient.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'item') Then
     list.add(FItem.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
end;

procedure TFhirMedicationProductIngredient.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'item', 'Reference(Substance|Medication)', FItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Ratio', FAmount.Link));{2}
end;

procedure TFhirMedicationProductIngredient.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'item') then Item := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirRatio{4b}
  else inherited;
end;

function TFhirMedicationProductIngredient.FhirType : string;
begin
  result := 'ingredient';
end;

function TFhirMedicationProductIngredient.Link : TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient(inherited Link);
end;

function TFhirMedicationProductIngredient.Clone : TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient(inherited Clone);
end;

function TFhirMedicationProductIngredient.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProductIngredient;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationProductIngredient)) then
    result := false
  else
  begin
    o := TFhirMedicationProductIngredient(other);
    result := compareDeep(itemElement, o.itemElement, true) and compareDeep(amountElement, o.amountElement, true);
  end;
end;

function TFhirMedicationProductIngredient.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProductIngredient;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationProductIngredient)) then
    result := false
  else
  begin
    o := TFhirMedicationProductIngredient(other);
    result := true;
  end;
end;

{ TFhirMedicationProductIngredient }

Procedure TFhirMedicationProductIngredient.SetItem(value : TFhirReference{Resource});
begin
  FItem.free;
  FItem := value;
end;

Procedure TFhirMedicationProductIngredient.SetAmount(value : TFhirRatio);
begin
  FAmount.free;
  FAmount := value;
end;


{ TFhirMedicationProductIngredientListEnumerator }

Constructor TFhirMedicationProductIngredientListEnumerator.Create(list : TFhirMedicationProductIngredientList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationProductIngredientListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationProductIngredientListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationProductIngredientListEnumerator.GetCurrent : TFhirMedicationProductIngredient;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationProductIngredientList }
procedure TFhirMedicationProductIngredientList.AddItem(value: TFhirMedicationProductIngredient);
begin
  assert(value.ClassName = 'TFhirMedicationProductIngredient', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationProductIngredient');
  add(value);
end;


function TFhirMedicationProductIngredientList.Append: TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductIngredientList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationProductIngredientList.GetEnumerator : TFhirMedicationProductIngredientListEnumerator;
begin
  result := TFhirMedicationProductIngredientListEnumerator.Create(self.link);
end;

function TFhirMedicationProductIngredientList.Clone: TFhirMedicationProductIngredientList;
begin
  result := TFhirMedicationProductIngredientList(inherited Clone);
end;

function TFhirMedicationProductIngredientList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationProductIngredientList.GetItemN(index: Integer): TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient(ObjectByIndex[index]);
end;

function TFhirMedicationProductIngredientList.IndexOf(value: TFhirMedicationProductIngredient): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationProductIngredientList.Insert(index: Integer): TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductIngredientList.InsertItem(index: Integer; value: TFhirMedicationProductIngredient);
begin
  assert(value is TFhirMedicationProductIngredient);
  Inherited Insert(index, value);
end;

function TFhirMedicationProductIngredientList.Item(index: Integer): TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient(ObjectByIndex[index]);
end;

function TFhirMedicationProductIngredientList.Link: TFhirMedicationProductIngredientList;
begin
  result := TFhirMedicationProductIngredientList(inherited Link);
end;

procedure TFhirMedicationProductIngredientList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationProductIngredientList.SetItemByIndex(index: Integer; value: TFhirMedicationProductIngredient);
begin
  assert(value is TFhirMedicationProductIngredient);
  FhirMedicationProductIngredients[index] := value;
end;

procedure TFhirMedicationProductIngredientList.SetItemN(index: Integer; value: TFhirMedicationProductIngredient);
begin
  assert(value is TFhirMedicationProductIngredient);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationProductBatch }

constructor TFhirMedicationProductBatch.Create;
begin
  inherited;
end;

destructor TFhirMedicationProductBatch.Destroy;
begin
  FLotNumber.free;
  FExpirationDate.free;
  inherited;
end;

procedure TFhirMedicationProductBatch.Assign(oSource : TAdvObject);
begin
  inherited;
  lotNumberElement := TFhirMedicationProductBatch(oSource).lotNumberElement.Clone;
  expirationDateElement := TFhirMedicationProductBatch(oSource).expirationDateElement.Clone;
end;

procedure TFhirMedicationProductBatch.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'lotNumber') Then
     list.add(FLotNumber.Link);
  if (child_name = 'expirationDate') Then
     list.add(FExpirationDate.Link);
end;

procedure TFhirMedicationProductBatch.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'lotNumber', 'string', FLotNumber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expirationDate', 'dateTime', FExpirationDate.Link));{2}
end;

procedure TFhirMedicationProductBatch.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'lotNumber') then LotNumberElement := propValue as TFhirString{5a}
  else if (propName = 'expirationDate') then ExpirationDateElement := propValue as TFhirDateTime{5a}
  else inherited;
end;

function TFhirMedicationProductBatch.FhirType : string;
begin
  result := 'batch';
end;

function TFhirMedicationProductBatch.Link : TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch(inherited Link);
end;

function TFhirMedicationProductBatch.Clone : TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch(inherited Clone);
end;

function TFhirMedicationProductBatch.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProductBatch;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationProductBatch)) then
    result := false
  else
  begin
    o := TFhirMedicationProductBatch(other);
    result := compareDeep(lotNumberElement, o.lotNumberElement, true) and compareDeep(expirationDateElement, o.expirationDateElement, true);
  end;
end;

function TFhirMedicationProductBatch.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationProductBatch;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationProductBatch)) then
    result := false
  else
  begin
    o := TFhirMedicationProductBatch(other);
    result := compareValues(lotNumberElement, o.lotNumberElement, true) and compareValues(expirationDateElement, o.expirationDateElement, true);
  end;
end;

{ TFhirMedicationProductBatch }

Procedure TFhirMedicationProductBatch.SetLotNumber(value : TFhirString);
begin
  FLotNumber.free;
  FLotNumber := value;
end;

Function TFhirMedicationProductBatch.GetLotNumberST : String;
begin
  if FLotNumber = nil then
    result := ''
  else
    result := FLotNumber.value;
end;

Procedure TFhirMedicationProductBatch.SetLotNumberST(value : String);
begin
  if value <> '' then
  begin
    if FLotNumber = nil then
      FLotNumber := TFhirString.create;
    FLotNumber.value := value
  end
  else if FLotNumber <> nil then
    FLotNumber.value := '';
end;

Procedure TFhirMedicationProductBatch.SetExpirationDate(value : TFhirDateTime);
begin
  FExpirationDate.free;
  FExpirationDate := value;
end;

Function TFhirMedicationProductBatch.GetExpirationDateST : TDateAndTime;
begin
  if FExpirationDate = nil then
    result := nil
  else
    result := FExpirationDate.value;
end;

Procedure TFhirMedicationProductBatch.SetExpirationDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FExpirationDate = nil then
      FExpirationDate := TFhirDateTime.create;
    FExpirationDate.value := value
  end
  else if FExpirationDate <> nil then
    FExpirationDate.value := nil;
end;


{ TFhirMedicationProductBatchListEnumerator }

Constructor TFhirMedicationProductBatchListEnumerator.Create(list : TFhirMedicationProductBatchList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationProductBatchListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationProductBatchListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationProductBatchListEnumerator.GetCurrent : TFhirMedicationProductBatch;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationProductBatchList }
procedure TFhirMedicationProductBatchList.AddItem(value: TFhirMedicationProductBatch);
begin
  assert(value.ClassName = 'TFhirMedicationProductBatch', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationProductBatch');
  add(value);
end;


function TFhirMedicationProductBatchList.Append: TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductBatchList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationProductBatchList.GetEnumerator : TFhirMedicationProductBatchListEnumerator;
begin
  result := TFhirMedicationProductBatchListEnumerator.Create(self.link);
end;

function TFhirMedicationProductBatchList.Clone: TFhirMedicationProductBatchList;
begin
  result := TFhirMedicationProductBatchList(inherited Clone);
end;

function TFhirMedicationProductBatchList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationProductBatchList.GetItemN(index: Integer): TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch(ObjectByIndex[index]);
end;

function TFhirMedicationProductBatchList.IndexOf(value: TFhirMedicationProductBatch): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationProductBatchList.Insert(index: Integer): TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationProductBatchList.InsertItem(index: Integer; value: TFhirMedicationProductBatch);
begin
  assert(value is TFhirMedicationProductBatch);
  Inherited Insert(index, value);
end;

function TFhirMedicationProductBatchList.Item(index: Integer): TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch(ObjectByIndex[index]);
end;

function TFhirMedicationProductBatchList.Link: TFhirMedicationProductBatchList;
begin
  result := TFhirMedicationProductBatchList(inherited Link);
end;

procedure TFhirMedicationProductBatchList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationProductBatchList.SetItemByIndex(index: Integer; value: TFhirMedicationProductBatch);
begin
  assert(value is TFhirMedicationProductBatch);
  FhirMedicationProductBatches[index] := value;
end;

procedure TFhirMedicationProductBatchList.SetItemN(index: Integer; value: TFhirMedicationProductBatch);
begin
  assert(value is TFhirMedicationProductBatch);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationPackage }

constructor TFhirMedicationPackage.Create;
begin
  inherited;
end;

destructor TFhirMedicationPackage.Destroy;
begin
  FContainer.free;
  FContentList.Free;
  inherited;
end;

procedure TFhirMedicationPackage.Assign(oSource : TAdvObject);
begin
  inherited;
  container := TFhirMedicationPackage(oSource).container.Clone;
  if (TFhirMedicationPackage(oSource).FContentList = nil) then
  begin
    FContentList.free;
    FContentList := nil;
  end
  else
  begin
    FContentList := TFhirMedicationPackageContentList.Create;
    FContentList.Assign(TFhirMedicationPackage(oSource).FContentList);
  end;
end;

procedure TFhirMedicationPackage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'container') Then
     list.add(FContainer.Link);
  if (child_name = 'content') Then
    list.addAll(FContentList);
end;

procedure TFhirMedicationPackage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'container', 'CodeableConcept', FContainer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'content', '', FContentList.Link)){3};
end;

procedure TFhirMedicationPackage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'container') then Container := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'content') then ContentList.add(propValue as TFhirMedicationPackageContent){2}
  else inherited;
end;

function TFhirMedicationPackage.FhirType : string;
begin
  result := 'package';
end;

function TFhirMedicationPackage.Link : TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage(inherited Link);
end;

function TFhirMedicationPackage.Clone : TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage(inherited Clone);
end;

function TFhirMedicationPackage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationPackage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationPackage)) then
    result := false
  else
  begin
    o := TFhirMedicationPackage(other);
    result := compareDeep(containerElement, o.containerElement, true) and compareDeep(contentList, o.contentList, true);
  end;
end;

function TFhirMedicationPackage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationPackage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationPackage)) then
    result := false
  else
  begin
    o := TFhirMedicationPackage(other);
    result := true;
  end;
end;

{ TFhirMedicationPackage }

Procedure TFhirMedicationPackage.SetContainer(value : TFhirCodeableConcept);
begin
  FContainer.free;
  FContainer := value;
end;

Function TFhirMedicationPackage.GetContentList : TFhirMedicationPackageContentList;
begin
  if FContentList = nil then
    FContentList := TFhirMedicationPackageContentList.Create;
  result := FContentList;
end;

Function TFhirMedicationPackage.GetHasContentList : boolean;
begin
  result := (FContentList <> nil) and (FContentList.count > 0);
end;


{ TFhirMedicationPackageListEnumerator }

Constructor TFhirMedicationPackageListEnumerator.Create(list : TFhirMedicationPackageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationPackageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationPackageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationPackageListEnumerator.GetCurrent : TFhirMedicationPackage;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationPackageList }
procedure TFhirMedicationPackageList.AddItem(value: TFhirMedicationPackage);
begin
  assert(value.ClassName = 'TFhirMedicationPackage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationPackage');
  add(value);
end;


function TFhirMedicationPackageList.Append: TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationPackageList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationPackageList.GetEnumerator : TFhirMedicationPackageListEnumerator;
begin
  result := TFhirMedicationPackageListEnumerator.Create(self.link);
end;

function TFhirMedicationPackageList.Clone: TFhirMedicationPackageList;
begin
  result := TFhirMedicationPackageList(inherited Clone);
end;

function TFhirMedicationPackageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationPackageList.GetItemN(index: Integer): TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage(ObjectByIndex[index]);
end;

function TFhirMedicationPackageList.IndexOf(value: TFhirMedicationPackage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationPackageList.Insert(index: Integer): TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationPackageList.InsertItem(index: Integer; value: TFhirMedicationPackage);
begin
  assert(value is TFhirMedicationPackage);
  Inherited Insert(index, value);
end;

function TFhirMedicationPackageList.Item(index: Integer): TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage(ObjectByIndex[index]);
end;

function TFhirMedicationPackageList.Link: TFhirMedicationPackageList;
begin
  result := TFhirMedicationPackageList(inherited Link);
end;

procedure TFhirMedicationPackageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationPackageList.SetItemByIndex(index: Integer; value: TFhirMedicationPackage);
begin
  assert(value is TFhirMedicationPackage);
  FhirMedicationPackages[index] := value;
end;

procedure TFhirMedicationPackageList.SetItemN(index: Integer; value: TFhirMedicationPackage);
begin
  assert(value is TFhirMedicationPackage);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationPackageContent }

constructor TFhirMedicationPackageContent.Create;
begin
  inherited;
end;

destructor TFhirMedicationPackageContent.Destroy;
begin
  FItem.free;
  FAmount.free;
  inherited;
end;

procedure TFhirMedicationPackageContent.Assign(oSource : TAdvObject);
begin
  inherited;
  item := TFhirMedicationPackageContent(oSource).item.Clone;
  amount := TFhirMedicationPackageContent(oSource).amount.Clone;
end;

procedure TFhirMedicationPackageContent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'item') Then
     list.add(FItem.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
end;

procedure TFhirMedicationPackageContent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'item', 'Reference(Medication)', FItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'SimpleQuantity', FAmount.Link));{2}
end;

procedure TFhirMedicationPackageContent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'item') then Item := propValue as TFhirReference{TFhirMedication}{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirMedicationPackageContent.FhirType : string;
begin
  result := 'content';
end;

function TFhirMedicationPackageContent.Link : TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent(inherited Link);
end;

function TFhirMedicationPackageContent.Clone : TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent(inherited Clone);
end;

function TFhirMedicationPackageContent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationPackageContent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationPackageContent)) then
    result := false
  else
  begin
    o := TFhirMedicationPackageContent(other);
    result := compareDeep(itemElement, o.itemElement, true) and compareDeep(amountElement, o.amountElement, true);
  end;
end;

function TFhirMedicationPackageContent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationPackageContent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationPackageContent)) then
    result := false
  else
  begin
    o := TFhirMedicationPackageContent(other);
    result := true;
  end;
end;

{ TFhirMedicationPackageContent }

Procedure TFhirMedicationPackageContent.SetItem(value : TFhirReference{TFhirMedication});
begin
  FItem.free;
  FItem := value;
end;

Procedure TFhirMedicationPackageContent.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;


{ TFhirMedicationPackageContentListEnumerator }

Constructor TFhirMedicationPackageContentListEnumerator.Create(list : TFhirMedicationPackageContentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationPackageContentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationPackageContentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationPackageContentListEnumerator.GetCurrent : TFhirMedicationPackageContent;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationPackageContentList }
procedure TFhirMedicationPackageContentList.AddItem(value: TFhirMedicationPackageContent);
begin
  assert(value.ClassName = 'TFhirMedicationPackageContent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationPackageContent');
  add(value);
end;


function TFhirMedicationPackageContentList.Append: TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationPackageContentList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationPackageContentList.GetEnumerator : TFhirMedicationPackageContentListEnumerator;
begin
  result := TFhirMedicationPackageContentListEnumerator.Create(self.link);
end;

function TFhirMedicationPackageContentList.Clone: TFhirMedicationPackageContentList;
begin
  result := TFhirMedicationPackageContentList(inherited Clone);
end;

function TFhirMedicationPackageContentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationPackageContentList.GetItemN(index: Integer): TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent(ObjectByIndex[index]);
end;

function TFhirMedicationPackageContentList.IndexOf(value: TFhirMedicationPackageContent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationPackageContentList.Insert(index: Integer): TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationPackageContentList.InsertItem(index: Integer; value: TFhirMedicationPackageContent);
begin
  assert(value is TFhirMedicationPackageContent);
  Inherited Insert(index, value);
end;

function TFhirMedicationPackageContentList.Item(index: Integer): TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent(ObjectByIndex[index]);
end;

function TFhirMedicationPackageContentList.Link: TFhirMedicationPackageContentList;
begin
  result := TFhirMedicationPackageContentList(inherited Link);
end;

procedure TFhirMedicationPackageContentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationPackageContentList.SetItemByIndex(index: Integer; value: TFhirMedicationPackageContent);
begin
  assert(value is TFhirMedicationPackageContent);
  FhirMedicationPackageContents[index] := value;
end;

procedure TFhirMedicationPackageContentList.SetItemN(index: Integer; value: TFhirMedicationPackageContent);
begin
  assert(value is TFhirMedicationPackageContent);
  ObjectByIndex[index] := value;
end;

{ TFhirMedication }

constructor TFhirMedication.Create;
begin
  inherited;
end;

destructor TFhirMedication.Destroy;
begin
  FCode.free;
  FIsBrand.free;
  FManufacturer.free;
  FProduct.free;
  FPackage.free;
  inherited;
end;

function TFhirMedication.GetResourceType : TFhirResourceType;
begin
  result := frtMedication;
end;

procedure TFhirMedication.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirMedication(oSource).code.Clone;
  isBrandElement := TFhirMedication(oSource).isBrandElement.Clone;
  manufacturer := TFhirMedication(oSource).manufacturer.Clone;
  product := TFhirMedication(oSource).product.Clone;
  package := TFhirMedication(oSource).package.Clone;
end;

procedure TFhirMedication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'isBrand') Then
     list.add(FIsBrand.Link);
  if (child_name = 'manufacturer') Then
     list.add(FManufacturer.Link);
  if (child_name = 'product') Then
     list.add(FProduct.Link);
  if (child_name = 'package') Then
     list.add(FPackage.Link);
end;

procedure TFhirMedication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'isBrand', 'boolean', FIsBrand.Link));{2}
  oList.add(TFHIRProperty.create(self, 'manufacturer', 'Reference(Organization)', FManufacturer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'product', '', FProduct.Link));{2}
  oList.add(TFHIRProperty.create(self, 'package', '', FPackage.Link));{2}
end;

procedure TFhirMedication.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'isBrand') then IsBrandElement := propValue as TFhirBoolean{5a}
  else if (propName = 'manufacturer') then Manufacturer := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'product') then Product := propValue as TFhirMedicationProduct{4b}
  else if (propName = 'package') then Package := propValue as TFhirMedicationPackage{4b}
  else inherited;
end;

function TFhirMedication.FhirType : string;
begin
  result := 'Medication';
end;

function TFhirMedication.Link : TFhirMedication;
begin
  result := TFhirMedication(inherited Link);
end;

function TFhirMedication.Clone : TFhirMedication;
begin
  result := TFhirMedication(inherited Clone);
end;

{ TFhirMedication }

Procedure TFhirMedication.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirMedication.SetIsBrand(value : TFhirBoolean);
begin
  FIsBrand.free;
  FIsBrand := value;
end;

Function TFhirMedication.GetIsBrandST : Boolean;
begin
  if FIsBrand = nil then
    result := false
  else
    result := FIsBrand.value;
end;

Procedure TFhirMedication.SetIsBrandST(value : Boolean);
begin
  if FIsBrand = nil then
    FIsBrand := TFhirBoolean.create;
  FIsBrand.value := value
end;

Procedure TFhirMedication.SetManufacturer(value : TFhirReference{TFhirOrganization});
begin
  FManufacturer.free;
  FManufacturer := value;
end;

Procedure TFhirMedication.SetProduct(value : TFhirMedicationProduct);
begin
  FProduct.free;
  FProduct := value;
end;

Procedure TFhirMedication.SetPackage(value : TFhirMedicationPackage);
begin
  FPackage.free;
  FPackage := value;
end;


{ TFhirMedicationListEnumerator }

Constructor TFhirMedicationListEnumerator.Create(list : TFhirMedicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationListEnumerator.GetCurrent : TFhirMedication;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationList }
procedure TFhirMedicationList.AddItem(value: TFhirMedication);
begin
  assert(value.ClassName = 'TFhirMedication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedication');
  add(value);
end;


function TFhirMedicationList.Append: TFhirMedication;
begin
  result := TFhirMedication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationList.GetEnumerator : TFhirMedicationListEnumerator;
begin
  result := TFhirMedicationListEnumerator.Create(self.link);
end;

function TFhirMedicationList.Clone: TFhirMedicationList;
begin
  result := TFhirMedicationList(inherited Clone);
end;

function TFhirMedicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationList.GetItemN(index: Integer): TFhirMedication;
begin
  result := TFhirMedication(ObjectByIndex[index]);
end;

function TFhirMedicationList.IndexOf(value: TFhirMedication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationList.Insert(index: Integer): TFhirMedication;
begin
  result := TFhirMedication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationList.InsertItem(index: Integer; value: TFhirMedication);
begin
  assert(value is TFhirMedication);
  Inherited Insert(index, value);
end;

function TFhirMedicationList.Item(index: Integer): TFhirMedication;
begin
  result := TFhirMedication(ObjectByIndex[index]);
end;

function TFhirMedicationList.Link: TFhirMedicationList;
begin
  result := TFhirMedicationList(inherited Link);
end;

procedure TFhirMedicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationList.SetItemByIndex(index: Integer; value: TFhirMedication);
begin
  assert(value is TFhirMedication);
  FhirMedications[index] := value;
end;

procedure TFhirMedicationList.SetItemN(index: Integer; value: TFhirMedication);
begin
  assert(value is TFhirMedication);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationAdministrationDosage }

constructor TFhirMedicationAdministrationDosage.Create;
begin
  inherited;
end;

destructor TFhirMedicationAdministrationDosage.Destroy;
begin
  FText.free;
  FSite.free;
  FRoute.free;
  FMethod.free;
  FQuantity.free;
  FRate.free;
  inherited;
end;

procedure TFhirMedicationAdministrationDosage.Assign(oSource : TAdvObject);
begin
  inherited;
  textElement := TFhirMedicationAdministrationDosage(oSource).textElement.Clone;
  site := TFhirMedicationAdministrationDosage(oSource).site.Clone;
  route := TFhirMedicationAdministrationDosage(oSource).route.Clone;
  method := TFhirMedicationAdministrationDosage(oSource).method.Clone;
  quantity := TFhirMedicationAdministrationDosage(oSource).quantity.Clone;
  rate := TFhirMedicationAdministrationDosage(oSource).rate.Clone;
end;

procedure TFhirMedicationAdministrationDosage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'site[x]') Then
     list.add(FSite.Link);
  if (child_name = 'route') Then
     list.add(FRoute.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'rate[x]') Then
     list.add(FRate.Link);
end;

procedure TFhirMedicationAdministrationDosage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'site[x]', 'CodeableConcept|Reference(BodySite)', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'route', 'CodeableConcept', FRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rate[x]', 'Ratio|Range', FRate.Link));{2}
end;

procedure TFhirMedicationAdministrationDosage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName.startsWith('site')) then Site := propValue as TFhirType{4}
  else if (propName = 'route') then Route := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName.startsWith('rate')) then Rate := propValue as TFhirType{4}
  else inherited;
end;

function TFhirMedicationAdministrationDosage.FhirType : string;
begin
  result := 'dosage';
end;

function TFhirMedicationAdministrationDosage.Link : TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage(inherited Link);
end;

function TFhirMedicationAdministrationDosage.Clone : TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage(inherited Clone);
end;

function TFhirMedicationAdministrationDosage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationAdministrationDosage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationAdministrationDosage)) then
    result := false
  else
  begin
    o := TFhirMedicationAdministrationDosage(other);
    result := compareDeep(textElement, o.textElement, true) and compareDeep(siteElement, o.siteElement, true) and 
      compareDeep(routeElement, o.routeElement, true) and compareDeep(methodElement, o.methodElement, true) and 
      compareDeep(quantityElement, o.quantityElement, true) and compareDeep(rateElement, o.rateElement, true);
  end;
end;

function TFhirMedicationAdministrationDosage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationAdministrationDosage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationAdministrationDosage)) then
    result := false
  else
  begin
    o := TFhirMedicationAdministrationDosage(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirMedicationAdministrationDosage }

Procedure TFhirMedicationAdministrationDosage.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirMedicationAdministrationDosage.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirMedicationAdministrationDosage.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirMedicationAdministrationDosage.SetSite(value : TFhirType);
begin
  FSite.free;
  FSite := value;
end;

Procedure TFhirMedicationAdministrationDosage.SetRoute(value : TFhirCodeableConcept);
begin
  FRoute.free;
  FRoute := value;
end;

Procedure TFhirMedicationAdministrationDosage.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirMedicationAdministrationDosage.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirMedicationAdministrationDosage.SetRate(value : TFhirType);
begin
  FRate.free;
  FRate := value;
end;


{ TFhirMedicationAdministrationDosageListEnumerator }

Constructor TFhirMedicationAdministrationDosageListEnumerator.Create(list : TFhirMedicationAdministrationDosageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationAdministrationDosageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationAdministrationDosageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationAdministrationDosageListEnumerator.GetCurrent : TFhirMedicationAdministrationDosage;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationAdministrationDosageList }
procedure TFhirMedicationAdministrationDosageList.AddItem(value: TFhirMedicationAdministrationDosage);
begin
  assert(value.ClassName = 'TFhirMedicationAdministrationDosage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationAdministrationDosage');
  add(value);
end;


function TFhirMedicationAdministrationDosageList.Append: TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationAdministrationDosageList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationAdministrationDosageList.GetEnumerator : TFhirMedicationAdministrationDosageListEnumerator;
begin
  result := TFhirMedicationAdministrationDosageListEnumerator.Create(self.link);
end;

function TFhirMedicationAdministrationDosageList.Clone: TFhirMedicationAdministrationDosageList;
begin
  result := TFhirMedicationAdministrationDosageList(inherited Clone);
end;

function TFhirMedicationAdministrationDosageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationAdministrationDosageList.GetItemN(index: Integer): TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage(ObjectByIndex[index]);
end;

function TFhirMedicationAdministrationDosageList.IndexOf(value: TFhirMedicationAdministrationDosage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationAdministrationDosageList.Insert(index: Integer): TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationAdministrationDosageList.InsertItem(index: Integer; value: TFhirMedicationAdministrationDosage);
begin
  assert(value is TFhirMedicationAdministrationDosage);
  Inherited Insert(index, value);
end;

function TFhirMedicationAdministrationDosageList.Item(index: Integer): TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage(ObjectByIndex[index]);
end;

function TFhirMedicationAdministrationDosageList.Link: TFhirMedicationAdministrationDosageList;
begin
  result := TFhirMedicationAdministrationDosageList(inherited Link);
end;

procedure TFhirMedicationAdministrationDosageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationAdministrationDosageList.SetItemByIndex(index: Integer; value: TFhirMedicationAdministrationDosage);
begin
  assert(value is TFhirMedicationAdministrationDosage);
  FhirMedicationAdministrationDosages[index] := value;
end;

procedure TFhirMedicationAdministrationDosageList.SetItemN(index: Integer; value: TFhirMedicationAdministrationDosage);
begin
  assert(value is TFhirMedicationAdministrationDosage);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationAdministration }

constructor TFhirMedicationAdministration.Create;
begin
  inherited;
end;

destructor TFhirMedicationAdministration.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FPatient.free;
  FPractitioner.free;
  FEncounter.free;
  FPrescription.free;
  FWasNotGiven.free;
  FReasonNotGivenList.Free;
  FReasonGivenList.Free;
  FEffectiveTime.free;
  FMedication.free;
  FDeviceList.Free;
  FNote.free;
  FDosage.free;
  inherited;
end;

function TFhirMedicationAdministration.GetResourceType : TFhirResourceType;
begin
  result := frtMedicationAdministration;
end;

procedure TFhirMedicationAdministration.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirMedicationAdministration(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirMedicationAdministration(oSource).FIdentifierList);
  end;
  FStatus := TFhirMedicationAdministration(oSource).FStatus.Link;
  patient := TFhirMedicationAdministration(oSource).patient.Clone;
  practitioner := TFhirMedicationAdministration(oSource).practitioner.Clone;
  encounter := TFhirMedicationAdministration(oSource).encounter.Clone;
  prescription := TFhirMedicationAdministration(oSource).prescription.Clone;
  wasNotGivenElement := TFhirMedicationAdministration(oSource).wasNotGivenElement.Clone;
  if (TFhirMedicationAdministration(oSource).FReasonNotGivenList = nil) then
  begin
    FReasonNotGivenList.free;
    FReasonNotGivenList := nil;
  end
  else
  begin
    FReasonNotGivenList := TFhirCodeableConceptList.Create;
    FReasonNotGivenList.Assign(TFhirMedicationAdministration(oSource).FReasonNotGivenList);
  end;
  if (TFhirMedicationAdministration(oSource).FReasonGivenList = nil) then
  begin
    FReasonGivenList.free;
    FReasonGivenList := nil;
  end
  else
  begin
    FReasonGivenList := TFhirCodeableConceptList.Create;
    FReasonGivenList.Assign(TFhirMedicationAdministration(oSource).FReasonGivenList);
  end;
  effectiveTime := TFhirMedicationAdministration(oSource).effectiveTime.Clone;
  medication := TFhirMedicationAdministration(oSource).medication.Clone;
  if (TFhirMedicationAdministration(oSource).FDeviceList = nil) then
  begin
    FDeviceList.free;
    FDeviceList := nil;
  end
  else
  begin
    FDeviceList := TFhirReferenceList{TFhirDevice}.Create;
    FDeviceList.Assign(TFhirMedicationAdministration(oSource).FDeviceList);
  end;
  noteElement := TFhirMedicationAdministration(oSource).noteElement.Clone;
  dosage := TFhirMedicationAdministration(oSource).dosage.Clone;
end;

procedure TFhirMedicationAdministration.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'practitioner') Then
     list.add(FPractitioner.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'prescription') Then
     list.add(FPrescription.Link);
  if (child_name = 'wasNotGiven') Then
     list.add(FWasNotGiven.Link);
  if (child_name = 'reasonNotGiven') Then
    list.addAll(FReasonNotGivenList);
  if (child_name = 'reasonGiven') Then
    list.addAll(FReasonGivenList);
  if (child_name = 'effectiveTime[x]') Then
     list.add(FEffectiveTime.Link);
  if (child_name = 'medication[x]') Then
     list.add(FMedication.Link);
  if (child_name = 'device') Then
    list.addAll(FDeviceList);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'dosage') Then
     list.add(FDosage.Link);
end;

procedure TFhirMedicationAdministration.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'practitioner', 'Reference(Practitioner|Patient|RelatedPerson)', FPractitioner.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prescription', 'Reference(MedicationOrder)', FPrescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'wasNotGiven', 'boolean', FWasNotGiven.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reasonNotGiven', 'CodeableConcept', FReasonNotGivenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reasonGiven', 'CodeableConcept', FReasonGivenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'effectiveTime[x]', 'dateTime|Period', FEffectiveTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'medication[x]', 'CodeableConcept|Reference(Medication)', FMedication.Link));{2}
  oList.add(TFHIRProperty.create(self, 'device', 'Reference(Device)', FDeviceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'note', 'string', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dosage', '', FDosage.Link));{2}
end;

procedure TFhirMedicationAdministration.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'practitioner') then Practitioner := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'prescription') then Prescription := propValue as TFhirReference{TFhirMedicationOrder}{4b}
  else if (propName = 'wasNotGiven') then WasNotGivenElement := propValue as TFhirBoolean{5a}
  else if (propName = 'reasonNotGiven') then ReasonNotGivenList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'reasonGiven') then ReasonGivenList.add(propValue as TFhirCodeableConcept){2}
  else if (propName.startsWith('effectiveTime')) then EffectiveTime := propValue as TFhirType{4}
  else if (propName.startsWith('medication')) then Medication := propValue as TFhirType{4}
  else if (propName = 'device') then DeviceList.add(propValue as TFhirReference{TFhirDevice}){2}
  else if (propName = 'note') then NoteElement := propValue as TFhirString{5a}
  else if (propName = 'dosage') then Dosage := propValue as TFhirMedicationAdministrationDosage{4b}
  else inherited;
end;

function TFhirMedicationAdministration.FhirType : string;
begin
  result := 'MedicationAdministration';
end;

function TFhirMedicationAdministration.Link : TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration(inherited Link);
end;

function TFhirMedicationAdministration.Clone : TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration(inherited Clone);
end;

{ TFhirMedicationAdministration }

Function TFhirMedicationAdministration.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirMedicationAdministration.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirMedicationAdministration.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirMedicationAdministration.GetStatusST : TFhirMedicationAdminStatus;
begin
  if FStatus = nil then
    result := TFhirMedicationAdminStatus(0)
  else
    result := TFhirMedicationAdminStatus(StringArrayIndexOfSensitive(CODES_TFhirMedicationAdminStatus, FStatus.value));
end;

Procedure TFhirMedicationAdministration.SetStatusST(value : TFhirMedicationAdminStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirMedicationAdminStatus[value]);
end;

Procedure TFhirMedicationAdministration.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirMedicationAdministration.SetPractitioner(value : TFhirReference{Resource});
begin
  FPractitioner.free;
  FPractitioner := value;
end;

Procedure TFhirMedicationAdministration.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirMedicationAdministration.SetPrescription(value : TFhirReference{TFhirMedicationOrder});
begin
  FPrescription.free;
  FPrescription := value;
end;

Procedure TFhirMedicationAdministration.SetWasNotGiven(value : TFhirBoolean);
begin
  FWasNotGiven.free;
  FWasNotGiven := value;
end;

Function TFhirMedicationAdministration.GetWasNotGivenST : Boolean;
begin
  if FWasNotGiven = nil then
    result := false
  else
    result := FWasNotGiven.value;
end;

Procedure TFhirMedicationAdministration.SetWasNotGivenST(value : Boolean);
begin
  if FWasNotGiven = nil then
    FWasNotGiven := TFhirBoolean.create;
  FWasNotGiven.value := value
end;

Function TFhirMedicationAdministration.GetReasonNotGivenList : TFhirCodeableConceptList;
begin
  if FReasonNotGivenList = nil then
    FReasonNotGivenList := TFhirCodeableConceptList.Create;
  result := FReasonNotGivenList;
end;

Function TFhirMedicationAdministration.GetHasReasonNotGivenList : boolean;
begin
  result := (FReasonNotGivenList <> nil) and (FReasonNotGivenList.count > 0);
end;

Function TFhirMedicationAdministration.GetReasonGivenList : TFhirCodeableConceptList;
begin
  if FReasonGivenList = nil then
    FReasonGivenList := TFhirCodeableConceptList.Create;
  result := FReasonGivenList;
end;

Function TFhirMedicationAdministration.GetHasReasonGivenList : boolean;
begin
  result := (FReasonGivenList <> nil) and (FReasonGivenList.count > 0);
end;

Procedure TFhirMedicationAdministration.SetEffectiveTime(value : TFhirType);
begin
  FEffectiveTime.free;
  FEffectiveTime := value;
end;

Procedure TFhirMedicationAdministration.SetMedication(value : TFhirType);
begin
  FMedication.free;
  FMedication := value;
end;

Function TFhirMedicationAdministration.GetDeviceList : TFhirReferenceList{TFhirDevice};
begin
  if FDeviceList = nil then
    FDeviceList := TFhirReferenceList{TFhirDevice}.Create;
  result := FDeviceList;
end;

Function TFhirMedicationAdministration.GetHasDeviceList : boolean;
begin
  result := (FDeviceList <> nil) and (FDeviceList.count > 0);
end;

Procedure TFhirMedicationAdministration.SetNote(value : TFhirString);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirMedicationAdministration.GetNoteST : String;
begin
  if FNote = nil then
    result := ''
  else
    result := FNote.value;
end;

Procedure TFhirMedicationAdministration.SetNoteST(value : String);
begin
  if value <> '' then
  begin
    if FNote = nil then
      FNote := TFhirString.create;
    FNote.value := value
  end
  else if FNote <> nil then
    FNote.value := '';
end;

Procedure TFhirMedicationAdministration.SetDosage(value : TFhirMedicationAdministrationDosage);
begin
  FDosage.free;
  FDosage := value;
end;


{ TFhirMedicationAdministrationListEnumerator }

Constructor TFhirMedicationAdministrationListEnumerator.Create(list : TFhirMedicationAdministrationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationAdministrationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationAdministrationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationAdministrationListEnumerator.GetCurrent : TFhirMedicationAdministration;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationAdministrationList }
procedure TFhirMedicationAdministrationList.AddItem(value: TFhirMedicationAdministration);
begin
  assert(value.ClassName = 'TFhirMedicationAdministration', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationAdministration');
  add(value);
end;


function TFhirMedicationAdministrationList.Append: TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationAdministrationList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationAdministrationList.GetEnumerator : TFhirMedicationAdministrationListEnumerator;
begin
  result := TFhirMedicationAdministrationListEnumerator.Create(self.link);
end;

function TFhirMedicationAdministrationList.Clone: TFhirMedicationAdministrationList;
begin
  result := TFhirMedicationAdministrationList(inherited Clone);
end;

function TFhirMedicationAdministrationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationAdministrationList.GetItemN(index: Integer): TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration(ObjectByIndex[index]);
end;

function TFhirMedicationAdministrationList.IndexOf(value: TFhirMedicationAdministration): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationAdministrationList.Insert(index: Integer): TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationAdministrationList.InsertItem(index: Integer; value: TFhirMedicationAdministration);
begin
  assert(value is TFhirMedicationAdministration);
  Inherited Insert(index, value);
end;

function TFhirMedicationAdministrationList.Item(index: Integer): TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration(ObjectByIndex[index]);
end;

function TFhirMedicationAdministrationList.Link: TFhirMedicationAdministrationList;
begin
  result := TFhirMedicationAdministrationList(inherited Link);
end;

procedure TFhirMedicationAdministrationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationAdministrationList.SetItemByIndex(index: Integer; value: TFhirMedicationAdministration);
begin
  assert(value is TFhirMedicationAdministration);
  FhirMedicationAdministrations[index] := value;
end;

procedure TFhirMedicationAdministrationList.SetItemN(index: Integer; value: TFhirMedicationAdministration);
begin
  assert(value is TFhirMedicationAdministration);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationDispenseDosageInstruction }

constructor TFhirMedicationDispenseDosageInstruction.Create;
begin
  inherited;
end;

destructor TFhirMedicationDispenseDosageInstruction.Destroy;
begin
  FText.free;
  FAdditionalInstructions.free;
  FTiming.free;
  FAsNeeded.free;
  FSite.free;
  FRoute.free;
  FMethod.free;
  FDose.free;
  FRate.free;
  FMaxDosePerPeriod.free;
  inherited;
end;

procedure TFhirMedicationDispenseDosageInstruction.Assign(oSource : TAdvObject);
begin
  inherited;
  textElement := TFhirMedicationDispenseDosageInstruction(oSource).textElement.Clone;
  additionalInstructions := TFhirMedicationDispenseDosageInstruction(oSource).additionalInstructions.Clone;
  timing := TFhirMedicationDispenseDosageInstruction(oSource).timing.Clone;
  asNeeded := TFhirMedicationDispenseDosageInstruction(oSource).asNeeded.Clone;
  site := TFhirMedicationDispenseDosageInstruction(oSource).site.Clone;
  route := TFhirMedicationDispenseDosageInstruction(oSource).route.Clone;
  method := TFhirMedicationDispenseDosageInstruction(oSource).method.Clone;
  dose := TFhirMedicationDispenseDosageInstruction(oSource).dose.Clone;
  rate := TFhirMedicationDispenseDosageInstruction(oSource).rate.Clone;
  maxDosePerPeriod := TFhirMedicationDispenseDosageInstruction(oSource).maxDosePerPeriod.Clone;
end;

procedure TFhirMedicationDispenseDosageInstruction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'additionalInstructions') Then
     list.add(FAdditionalInstructions.Link);
  if (child_name = 'timing') Then
     list.add(FTiming.Link);
  if (child_name = 'asNeeded[x]') Then
     list.add(FAsNeeded.Link);
  if (child_name = 'site[x]') Then
     list.add(FSite.Link);
  if (child_name = 'route') Then
     list.add(FRoute.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'dose[x]') Then
     list.add(FDose.Link);
  if (child_name = 'rate[x]') Then
     list.add(FRate.Link);
  if (child_name = 'maxDosePerPeriod') Then
     list.add(FMaxDosePerPeriod.Link);
end;

procedure TFhirMedicationDispenseDosageInstruction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additionalInstructions', 'CodeableConcept', FAdditionalInstructions.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timing', 'Timing', FTiming.Link));{2}
  oList.add(TFHIRProperty.create(self, 'asNeeded[x]', 'boolean|CodeableConcept', FAsNeeded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'site[x]', 'CodeableConcept|Reference(BodySite)', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'route', 'CodeableConcept', FRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dose[x]', 'Range|SimpleQuantity', FDose.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rate[x]', 'Ratio|Range', FRate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'maxDosePerPeriod', 'Ratio', FMaxDosePerPeriod.Link));{2}
end;

procedure TFhirMedicationDispenseDosageInstruction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'additionalInstructions') then AdditionalInstructions := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'timing') then Timing := propValue as TFhirTiming{4b}
  else if (propName.startsWith('asNeeded')) then AsNeeded := propValue as TFhirType{4}
  else if (propName.startsWith('site')) then Site := propValue as TFhirType{4}
  else if (propName = 'route') then Route := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('dose')) then Dose := propValue as TFhirType{4}
  else if (propName.startsWith('rate')) then Rate := propValue as TFhirType{4}
  else if (propName = 'maxDosePerPeriod') then MaxDosePerPeriod := propValue as TFhirRatio{4b}
  else inherited;
end;

function TFhirMedicationDispenseDosageInstruction.FhirType : string;
begin
  result := 'dosageInstruction';
end;

function TFhirMedicationDispenseDosageInstruction.Link : TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction(inherited Link);
end;

function TFhirMedicationDispenseDosageInstruction.Clone : TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction(inherited Clone);
end;

function TFhirMedicationDispenseDosageInstruction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationDispenseDosageInstruction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationDispenseDosageInstruction)) then
    result := false
  else
  begin
    o := TFhirMedicationDispenseDosageInstruction(other);
    result := compareDeep(textElement, o.textElement, true) and compareDeep(additionalInstructionsElement, o.additionalInstructionsElement, true) and 
      compareDeep(timingElement, o.timingElement, true) and compareDeep(asNeededElement, o.asNeededElement, true) and 
      compareDeep(siteElement, o.siteElement, true) and compareDeep(routeElement, o.routeElement, true) and 
      compareDeep(methodElement, o.methodElement, true) and compareDeep(doseElement, o.doseElement, true) and 
      compareDeep(rateElement, o.rateElement, true) and compareDeep(maxDosePerPeriodElement, o.maxDosePerPeriodElement, true);
  end;
end;

function TFhirMedicationDispenseDosageInstruction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationDispenseDosageInstruction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationDispenseDosageInstruction)) then
    result := false
  else
  begin
    o := TFhirMedicationDispenseDosageInstruction(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirMedicationDispenseDosageInstruction }

Procedure TFhirMedicationDispenseDosageInstruction.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirMedicationDispenseDosageInstruction.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetAdditionalInstructions(value : TFhirCodeableConcept);
begin
  FAdditionalInstructions.free;
  FAdditionalInstructions := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetTiming(value : TFhirTiming);
begin
  FTiming.free;
  FTiming := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetAsNeeded(value : TFhirType);
begin
  FAsNeeded.free;
  FAsNeeded := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetSite(value : TFhirType);
begin
  FSite.free;
  FSite := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetRoute(value : TFhirCodeableConcept);
begin
  FRoute.free;
  FRoute := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetDose(value : TFhirType);
begin
  FDose.free;
  FDose := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetRate(value : TFhirType);
begin
  FRate.free;
  FRate := value;
end;

Procedure TFhirMedicationDispenseDosageInstruction.SetMaxDosePerPeriod(value : TFhirRatio);
begin
  FMaxDosePerPeriod.free;
  FMaxDosePerPeriod := value;
end;


{ TFhirMedicationDispenseDosageInstructionListEnumerator }

Constructor TFhirMedicationDispenseDosageInstructionListEnumerator.Create(list : TFhirMedicationDispenseDosageInstructionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationDispenseDosageInstructionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationDispenseDosageInstructionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationDispenseDosageInstructionListEnumerator.GetCurrent : TFhirMedicationDispenseDosageInstruction;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationDispenseDosageInstructionList }
procedure TFhirMedicationDispenseDosageInstructionList.AddItem(value: TFhirMedicationDispenseDosageInstruction);
begin
  assert(value.ClassName = 'TFhirMedicationDispenseDosageInstruction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationDispenseDosageInstruction');
  add(value);
end;


function TFhirMedicationDispenseDosageInstructionList.Append: TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseDosageInstructionList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationDispenseDosageInstructionList.GetEnumerator : TFhirMedicationDispenseDosageInstructionListEnumerator;
begin
  result := TFhirMedicationDispenseDosageInstructionListEnumerator.Create(self.link);
end;

function TFhirMedicationDispenseDosageInstructionList.Clone: TFhirMedicationDispenseDosageInstructionList;
begin
  result := TFhirMedicationDispenseDosageInstructionList(inherited Clone);
end;

function TFhirMedicationDispenseDosageInstructionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationDispenseDosageInstructionList.GetItemN(index: Integer): TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseDosageInstructionList.IndexOf(value: TFhirMedicationDispenseDosageInstruction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationDispenseDosageInstructionList.Insert(index: Integer): TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseDosageInstructionList.InsertItem(index: Integer; value: TFhirMedicationDispenseDosageInstruction);
begin
  assert(value is TFhirMedicationDispenseDosageInstruction);
  Inherited Insert(index, value);
end;

function TFhirMedicationDispenseDosageInstructionList.Item(index: Integer): TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseDosageInstructionList.Link: TFhirMedicationDispenseDosageInstructionList;
begin
  result := TFhirMedicationDispenseDosageInstructionList(inherited Link);
end;

procedure TFhirMedicationDispenseDosageInstructionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationDispenseDosageInstructionList.SetItemByIndex(index: Integer; value: TFhirMedicationDispenseDosageInstruction);
begin
  assert(value is TFhirMedicationDispenseDosageInstruction);
  FhirMedicationDispenseDosageInstructions[index] := value;
end;

procedure TFhirMedicationDispenseDosageInstructionList.SetItemN(index: Integer; value: TFhirMedicationDispenseDosageInstruction);
begin
  assert(value is TFhirMedicationDispenseDosageInstruction);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationDispenseSubstitution }

constructor TFhirMedicationDispenseSubstitution.Create;
begin
  inherited;
end;

destructor TFhirMedicationDispenseSubstitution.Destroy;
begin
  FType_.free;
  FReasonList.Free;
  FResponsiblePartyList.Free;
  inherited;
end;

procedure TFhirMedicationDispenseSubstitution.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirMedicationDispenseSubstitution(oSource).type_.Clone;
  if (TFhirMedicationDispenseSubstitution(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirMedicationDispenseSubstitution(oSource).FReasonList);
  end;
  if (TFhirMedicationDispenseSubstitution(oSource).FResponsiblePartyList = nil) then
  begin
    FResponsiblePartyList.free;
    FResponsiblePartyList := nil;
  end
  else
  begin
    FResponsiblePartyList := TFhirReferenceList{TFhirPractitioner}.Create;
    FResponsiblePartyList.Assign(TFhirMedicationDispenseSubstitution(oSource).FResponsiblePartyList);
  end;
end;

procedure TFhirMedicationDispenseSubstitution.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'responsibleParty') Then
    list.addAll(FResponsiblePartyList);
end;

procedure TFhirMedicationDispenseSubstitution.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'responsibleParty', 'Reference(Practitioner)', FResponsiblePartyList.Link)){3};
end;

procedure TFhirMedicationDispenseSubstitution.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'responsibleParty') then ResponsiblePartyList.add(propValue as TFhirReference{TFhirPractitioner}){2}
  else inherited;
end;

function TFhirMedicationDispenseSubstitution.FhirType : string;
begin
  result := 'substitution';
end;

function TFhirMedicationDispenseSubstitution.Link : TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution(inherited Link);
end;

function TFhirMedicationDispenseSubstitution.Clone : TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution(inherited Clone);
end;

function TFhirMedicationDispenseSubstitution.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationDispenseSubstitution;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationDispenseSubstitution)) then
    result := false
  else
  begin
    o := TFhirMedicationDispenseSubstitution(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(reasonList, o.reasonList, true) and 
      compareDeep(responsiblePartyList, o.responsiblePartyList, true);
  end;
end;

function TFhirMedicationDispenseSubstitution.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationDispenseSubstitution;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationDispenseSubstitution)) then
    result := false
  else
  begin
    o := TFhirMedicationDispenseSubstitution(other);
    result := true;
  end;
end;

{ TFhirMedicationDispenseSubstitution }

Procedure TFhirMedicationDispenseSubstitution.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirMedicationDispenseSubstitution.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirMedicationDispenseSubstitution.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Function TFhirMedicationDispenseSubstitution.GetResponsiblePartyList : TFhirReferenceList{TFhirPractitioner};
begin
  if FResponsiblePartyList = nil then
    FResponsiblePartyList := TFhirReferenceList{TFhirPractitioner}.Create;
  result := FResponsiblePartyList;
end;

Function TFhirMedicationDispenseSubstitution.GetHasResponsiblePartyList : boolean;
begin
  result := (FResponsiblePartyList <> nil) and (FResponsiblePartyList.count > 0);
end;


{ TFhirMedicationDispenseSubstitutionListEnumerator }

Constructor TFhirMedicationDispenseSubstitutionListEnumerator.Create(list : TFhirMedicationDispenseSubstitutionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationDispenseSubstitutionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationDispenseSubstitutionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationDispenseSubstitutionListEnumerator.GetCurrent : TFhirMedicationDispenseSubstitution;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationDispenseSubstitutionList }
procedure TFhirMedicationDispenseSubstitutionList.AddItem(value: TFhirMedicationDispenseSubstitution);
begin
  assert(value.ClassName = 'TFhirMedicationDispenseSubstitution', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationDispenseSubstitution');
  add(value);
end;


function TFhirMedicationDispenseSubstitutionList.Append: TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseSubstitutionList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationDispenseSubstitutionList.GetEnumerator : TFhirMedicationDispenseSubstitutionListEnumerator;
begin
  result := TFhirMedicationDispenseSubstitutionListEnumerator.Create(self.link);
end;

function TFhirMedicationDispenseSubstitutionList.Clone: TFhirMedicationDispenseSubstitutionList;
begin
  result := TFhirMedicationDispenseSubstitutionList(inherited Clone);
end;

function TFhirMedicationDispenseSubstitutionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationDispenseSubstitutionList.GetItemN(index: Integer): TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseSubstitutionList.IndexOf(value: TFhirMedicationDispenseSubstitution): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationDispenseSubstitutionList.Insert(index: Integer): TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseSubstitutionList.InsertItem(index: Integer; value: TFhirMedicationDispenseSubstitution);
begin
  assert(value is TFhirMedicationDispenseSubstitution);
  Inherited Insert(index, value);
end;

function TFhirMedicationDispenseSubstitutionList.Item(index: Integer): TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseSubstitutionList.Link: TFhirMedicationDispenseSubstitutionList;
begin
  result := TFhirMedicationDispenseSubstitutionList(inherited Link);
end;

procedure TFhirMedicationDispenseSubstitutionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationDispenseSubstitutionList.SetItemByIndex(index: Integer; value: TFhirMedicationDispenseSubstitution);
begin
  assert(value is TFhirMedicationDispenseSubstitution);
  FhirMedicationDispenseSubstitutions[index] := value;
end;

procedure TFhirMedicationDispenseSubstitutionList.SetItemN(index: Integer; value: TFhirMedicationDispenseSubstitution);
begin
  assert(value is TFhirMedicationDispenseSubstitution);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationDispense }

constructor TFhirMedicationDispense.Create;
begin
  inherited;
end;

destructor TFhirMedicationDispense.Destroy;
begin
  FIdentifier.free;
  FStatus.free;
  FPatient.free;
  FDispenser.free;
  FAuthorizingPrescriptionList.Free;
  FType_.free;
  FQuantity.free;
  FDaysSupply.free;
  FMedication.free;
  FWhenPrepared.free;
  FWhenHandedOver.free;
  FDestination.free;
  FReceiverList.Free;
  FNote.free;
  FDosageInstructionList.Free;
  FSubstitution.free;
  inherited;
end;

function TFhirMedicationDispense.GetResourceType : TFhirResourceType;
begin
  result := frtMedicationDispense;
end;

procedure TFhirMedicationDispense.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirMedicationDispense(oSource).identifier.Clone;
  FStatus := TFhirMedicationDispense(oSource).FStatus.Link;
  patient := TFhirMedicationDispense(oSource).patient.Clone;
  dispenser := TFhirMedicationDispense(oSource).dispenser.Clone;
  if (TFhirMedicationDispense(oSource).FAuthorizingPrescriptionList = nil) then
  begin
    FAuthorizingPrescriptionList.free;
    FAuthorizingPrescriptionList := nil;
  end
  else
  begin
    FAuthorizingPrescriptionList := TFhirReferenceList{TFhirMedicationOrder}.Create;
    FAuthorizingPrescriptionList.Assign(TFhirMedicationDispense(oSource).FAuthorizingPrescriptionList);
  end;
  type_ := TFhirMedicationDispense(oSource).type_.Clone;
  quantity := TFhirMedicationDispense(oSource).quantity.Clone;
  daysSupply := TFhirMedicationDispense(oSource).daysSupply.Clone;
  medication := TFhirMedicationDispense(oSource).medication.Clone;
  whenPreparedElement := TFhirMedicationDispense(oSource).whenPreparedElement.Clone;
  whenHandedOverElement := TFhirMedicationDispense(oSource).whenHandedOverElement.Clone;
  destination := TFhirMedicationDispense(oSource).destination.Clone;
  if (TFhirMedicationDispense(oSource).FReceiverList = nil) then
  begin
    FReceiverList.free;
    FReceiverList := nil;
  end
  else
  begin
    FReceiverList := TFhirReferenceList{Resource}.Create;
    FReceiverList.Assign(TFhirMedicationDispense(oSource).FReceiverList);
  end;
  noteElement := TFhirMedicationDispense(oSource).noteElement.Clone;
  if (TFhirMedicationDispense(oSource).FDosageInstructionList = nil) then
  begin
    FDosageInstructionList.free;
    FDosageInstructionList := nil;
  end
  else
  begin
    FDosageInstructionList := TFhirMedicationDispenseDosageInstructionList.Create;
    FDosageInstructionList.Assign(TFhirMedicationDispense(oSource).FDosageInstructionList);
  end;
  substitution := TFhirMedicationDispense(oSource).substitution.Clone;
end;

procedure TFhirMedicationDispense.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'dispenser') Then
     list.add(FDispenser.Link);
  if (child_name = 'authorizingPrescription') Then
    list.addAll(FAuthorizingPrescriptionList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'daysSupply') Then
     list.add(FDaysSupply.Link);
  if (child_name = 'medication[x]') Then
     list.add(FMedication.Link);
  if (child_name = 'whenPrepared') Then
     list.add(FWhenPrepared.Link);
  if (child_name = 'whenHandedOver') Then
     list.add(FWhenHandedOver.Link);
  if (child_name = 'destination') Then
     list.add(FDestination.Link);
  if (child_name = 'receiver') Then
    list.addAll(FReceiverList);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'dosageInstruction') Then
    list.addAll(FDosageInstructionList);
  if (child_name = 'substitution') Then
     list.add(FSubstitution.Link);
end;

procedure TFhirMedicationDispense.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dispenser', 'Reference(Practitioner)', FDispenser.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authorizingPrescription', 'Reference(MedicationOrder)', FAuthorizingPrescriptionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'daysSupply', 'SimpleQuantity', FDaysSupply.Link));{2}
  oList.add(TFHIRProperty.create(self, 'medication[x]', 'CodeableConcept|Reference(Medication)', FMedication.Link));{2}
  oList.add(TFHIRProperty.create(self, 'whenPrepared', 'dateTime', FWhenPrepared.Link));{2}
  oList.add(TFHIRProperty.create(self, 'whenHandedOver', 'dateTime', FWhenHandedOver.Link));{2}
  oList.add(TFHIRProperty.create(self, 'destination', 'Reference(Location)', FDestination.Link));{2}
  oList.add(TFHIRProperty.create(self, 'receiver', 'Reference(Patient|Practitioner)', FReceiverList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'note', 'string', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dosageInstruction', '', FDosageInstructionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'substitution', '', FSubstitution.Link));{2}
end;

procedure TFhirMedicationDispense.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'dispenser') then Dispenser := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'authorizingPrescription') then AuthorizingPrescriptionList.add(propValue as TFhirReference{TFhirMedicationOrder}){2}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'daysSupply') then DaysSupply := propValue as TFhirQuantity{4b}
  else if (propName.startsWith('medication')) then Medication := propValue as TFhirType{4}
  else if (propName = 'whenPrepared') then WhenPreparedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'whenHandedOver') then WhenHandedOverElement := propValue as TFhirDateTime{5a}
  else if (propName = 'destination') then Destination := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'receiver') then ReceiverList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'note') then NoteElement := propValue as TFhirString{5a}
  else if (propName = 'dosageInstruction') then DosageInstructionList.add(propValue as TFhirMedicationDispenseDosageInstruction){2}
  else if (propName = 'substitution') then Substitution := propValue as TFhirMedicationDispenseSubstitution{4b}
  else inherited;
end;

function TFhirMedicationDispense.FhirType : string;
begin
  result := 'MedicationDispense';
end;

function TFhirMedicationDispense.Link : TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense(inherited Link);
end;

function TFhirMedicationDispense.Clone : TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense(inherited Clone);
end;

{ TFhirMedicationDispense }

Procedure TFhirMedicationDispense.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirMedicationDispense.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirMedicationDispense.GetStatusST : TFhirMedicationDispenseStatus;
begin
  if FStatus = nil then
    result := TFhirMedicationDispenseStatus(0)
  else
    result := TFhirMedicationDispenseStatus(StringArrayIndexOfSensitive(CODES_TFhirMedicationDispenseStatus, FStatus.value));
end;

Procedure TFhirMedicationDispense.SetStatusST(value : TFhirMedicationDispenseStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirMedicationDispenseStatus[value]);
end;

Procedure TFhirMedicationDispense.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirMedicationDispense.SetDispenser(value : TFhirReference{TFhirPractitioner});
begin
  FDispenser.free;
  FDispenser := value;
end;

Function TFhirMedicationDispense.GetAuthorizingPrescriptionList : TFhirReferenceList{TFhirMedicationOrder};
begin
  if FAuthorizingPrescriptionList = nil then
    FAuthorizingPrescriptionList := TFhirReferenceList{TFhirMedicationOrder}.Create;
  result := FAuthorizingPrescriptionList;
end;

Function TFhirMedicationDispense.GetHasAuthorizingPrescriptionList : boolean;
begin
  result := (FAuthorizingPrescriptionList <> nil) and (FAuthorizingPrescriptionList.count > 0);
end;

Procedure TFhirMedicationDispense.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirMedicationDispense.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirMedicationDispense.SetDaysSupply(value : TFhirQuantity);
begin
  FDaysSupply.free;
  FDaysSupply := value;
end;

Procedure TFhirMedicationDispense.SetMedication(value : TFhirType);
begin
  FMedication.free;
  FMedication := value;
end;

Procedure TFhirMedicationDispense.SetWhenPrepared(value : TFhirDateTime);
begin
  FWhenPrepared.free;
  FWhenPrepared := value;
end;

Function TFhirMedicationDispense.GetWhenPreparedST : TDateAndTime;
begin
  if FWhenPrepared = nil then
    result := nil
  else
    result := FWhenPrepared.value;
end;

Procedure TFhirMedicationDispense.SetWhenPreparedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FWhenPrepared = nil then
      FWhenPrepared := TFhirDateTime.create;
    FWhenPrepared.value := value
  end
  else if FWhenPrepared <> nil then
    FWhenPrepared.value := nil;
end;

Procedure TFhirMedicationDispense.SetWhenHandedOver(value : TFhirDateTime);
begin
  FWhenHandedOver.free;
  FWhenHandedOver := value;
end;

Function TFhirMedicationDispense.GetWhenHandedOverST : TDateAndTime;
begin
  if FWhenHandedOver = nil then
    result := nil
  else
    result := FWhenHandedOver.value;
end;

Procedure TFhirMedicationDispense.SetWhenHandedOverST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FWhenHandedOver = nil then
      FWhenHandedOver := TFhirDateTime.create;
    FWhenHandedOver.value := value
  end
  else if FWhenHandedOver <> nil then
    FWhenHandedOver.value := nil;
end;

Procedure TFhirMedicationDispense.SetDestination(value : TFhirReference{TFhirLocation});
begin
  FDestination.free;
  FDestination := value;
end;

Function TFhirMedicationDispense.GetReceiverList : TFhirReferenceList{Resource};
begin
  if FReceiverList = nil then
    FReceiverList := TFhirReferenceList{Resource}.Create;
  result := FReceiverList;
end;

Function TFhirMedicationDispense.GetHasReceiverList : boolean;
begin
  result := (FReceiverList <> nil) and (FReceiverList.count > 0);
end;

Procedure TFhirMedicationDispense.SetNote(value : TFhirString);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirMedicationDispense.GetNoteST : String;
begin
  if FNote = nil then
    result := ''
  else
    result := FNote.value;
end;

Procedure TFhirMedicationDispense.SetNoteST(value : String);
begin
  if value <> '' then
  begin
    if FNote = nil then
      FNote := TFhirString.create;
    FNote.value := value
  end
  else if FNote <> nil then
    FNote.value := '';
end;

Function TFhirMedicationDispense.GetDosageInstructionList : TFhirMedicationDispenseDosageInstructionList;
begin
  if FDosageInstructionList = nil then
    FDosageInstructionList := TFhirMedicationDispenseDosageInstructionList.Create;
  result := FDosageInstructionList;
end;

Function TFhirMedicationDispense.GetHasDosageInstructionList : boolean;
begin
  result := (FDosageInstructionList <> nil) and (FDosageInstructionList.count > 0);
end;

Procedure TFhirMedicationDispense.SetSubstitution(value : TFhirMedicationDispenseSubstitution);
begin
  FSubstitution.free;
  FSubstitution := value;
end;


{ TFhirMedicationDispenseListEnumerator }

Constructor TFhirMedicationDispenseListEnumerator.Create(list : TFhirMedicationDispenseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationDispenseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationDispenseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationDispenseListEnumerator.GetCurrent : TFhirMedicationDispense;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationDispenseList }
procedure TFhirMedicationDispenseList.AddItem(value: TFhirMedicationDispense);
begin
  assert(value.ClassName = 'TFhirMedicationDispense', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationDispense');
  add(value);
end;


function TFhirMedicationDispenseList.Append: TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationDispenseList.GetEnumerator : TFhirMedicationDispenseListEnumerator;
begin
  result := TFhirMedicationDispenseListEnumerator.Create(self.link);
end;

function TFhirMedicationDispenseList.Clone: TFhirMedicationDispenseList;
begin
  result := TFhirMedicationDispenseList(inherited Clone);
end;

function TFhirMedicationDispenseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationDispenseList.GetItemN(index: Integer): TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseList.IndexOf(value: TFhirMedicationDispense): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationDispenseList.Insert(index: Integer): TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationDispenseList.InsertItem(index: Integer; value: TFhirMedicationDispense);
begin
  assert(value is TFhirMedicationDispense);
  Inherited Insert(index, value);
end;

function TFhirMedicationDispenseList.Item(index: Integer): TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense(ObjectByIndex[index]);
end;

function TFhirMedicationDispenseList.Link: TFhirMedicationDispenseList;
begin
  result := TFhirMedicationDispenseList(inherited Link);
end;

procedure TFhirMedicationDispenseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationDispenseList.SetItemByIndex(index: Integer; value: TFhirMedicationDispense);
begin
  assert(value is TFhirMedicationDispense);
  FhirMedicationDispenses[index] := value;
end;

procedure TFhirMedicationDispenseList.SetItemN(index: Integer; value: TFhirMedicationDispense);
begin
  assert(value is TFhirMedicationDispense);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationOrderDosageInstruction }

constructor TFhirMedicationOrderDosageInstruction.Create;
begin
  inherited;
end;

destructor TFhirMedicationOrderDosageInstruction.Destroy;
begin
  FText.free;
  FAdditionalInstructions.free;
  FTiming.free;
  FAsNeeded.free;
  FSite.free;
  FRoute.free;
  FMethod.free;
  FDose.free;
  FRate.free;
  FMaxDosePerPeriod.free;
  inherited;
end;

procedure TFhirMedicationOrderDosageInstruction.Assign(oSource : TAdvObject);
begin
  inherited;
  textElement := TFhirMedicationOrderDosageInstruction(oSource).textElement.Clone;
  additionalInstructions := TFhirMedicationOrderDosageInstruction(oSource).additionalInstructions.Clone;
  timing := TFhirMedicationOrderDosageInstruction(oSource).timing.Clone;
  asNeeded := TFhirMedicationOrderDosageInstruction(oSource).asNeeded.Clone;
  site := TFhirMedicationOrderDosageInstruction(oSource).site.Clone;
  route := TFhirMedicationOrderDosageInstruction(oSource).route.Clone;
  method := TFhirMedicationOrderDosageInstruction(oSource).method.Clone;
  dose := TFhirMedicationOrderDosageInstruction(oSource).dose.Clone;
  rate := TFhirMedicationOrderDosageInstruction(oSource).rate.Clone;
  maxDosePerPeriod := TFhirMedicationOrderDosageInstruction(oSource).maxDosePerPeriod.Clone;
end;

procedure TFhirMedicationOrderDosageInstruction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'additionalInstructions') Then
     list.add(FAdditionalInstructions.Link);
  if (child_name = 'timing') Then
     list.add(FTiming.Link);
  if (child_name = 'asNeeded[x]') Then
     list.add(FAsNeeded.Link);
  if (child_name = 'site[x]') Then
     list.add(FSite.Link);
  if (child_name = 'route') Then
     list.add(FRoute.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'dose[x]') Then
     list.add(FDose.Link);
  if (child_name = 'rate[x]') Then
     list.add(FRate.Link);
  if (child_name = 'maxDosePerPeriod') Then
     list.add(FMaxDosePerPeriod.Link);
end;

procedure TFhirMedicationOrderDosageInstruction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additionalInstructions', 'CodeableConcept', FAdditionalInstructions.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timing', 'Timing', FTiming.Link));{2}
  oList.add(TFHIRProperty.create(self, 'asNeeded[x]', 'boolean|CodeableConcept', FAsNeeded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'site[x]', 'CodeableConcept|Reference(BodySite)', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'route', 'CodeableConcept', FRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dose[x]', 'Range|SimpleQuantity', FDose.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rate[x]', 'Ratio|Range', FRate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'maxDosePerPeriod', 'Ratio', FMaxDosePerPeriod.Link));{2}
end;

procedure TFhirMedicationOrderDosageInstruction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'additionalInstructions') then AdditionalInstructions := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'timing') then Timing := propValue as TFhirTiming{4b}
  else if (propName.startsWith('asNeeded')) then AsNeeded := propValue as TFhirType{4}
  else if (propName.startsWith('site')) then Site := propValue as TFhirType{4}
  else if (propName = 'route') then Route := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('dose')) then Dose := propValue as TFhirType{4}
  else if (propName.startsWith('rate')) then Rate := propValue as TFhirType{4}
  else if (propName = 'maxDosePerPeriod') then MaxDosePerPeriod := propValue as TFhirRatio{4b}
  else inherited;
end;

function TFhirMedicationOrderDosageInstruction.FhirType : string;
begin
  result := 'dosageInstruction';
end;

function TFhirMedicationOrderDosageInstruction.Link : TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction(inherited Link);
end;

function TFhirMedicationOrderDosageInstruction.Clone : TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction(inherited Clone);
end;

function TFhirMedicationOrderDosageInstruction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderDosageInstruction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderDosageInstruction)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderDosageInstruction(other);
    result := compareDeep(textElement, o.textElement, true) and compareDeep(additionalInstructionsElement, o.additionalInstructionsElement, true) and 
      compareDeep(timingElement, o.timingElement, true) and compareDeep(asNeededElement, o.asNeededElement, true) and 
      compareDeep(siteElement, o.siteElement, true) and compareDeep(routeElement, o.routeElement, true) and 
      compareDeep(methodElement, o.methodElement, true) and compareDeep(doseElement, o.doseElement, true) and 
      compareDeep(rateElement, o.rateElement, true) and compareDeep(maxDosePerPeriodElement, o.maxDosePerPeriodElement, true);
  end;
end;

function TFhirMedicationOrderDosageInstruction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderDosageInstruction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderDosageInstruction)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderDosageInstruction(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirMedicationOrderDosageInstruction }

Procedure TFhirMedicationOrderDosageInstruction.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirMedicationOrderDosageInstruction.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirMedicationOrderDosageInstruction.SetAdditionalInstructions(value : TFhirCodeableConcept);
begin
  FAdditionalInstructions.free;
  FAdditionalInstructions := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetTiming(value : TFhirTiming);
begin
  FTiming.free;
  FTiming := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetAsNeeded(value : TFhirType);
begin
  FAsNeeded.free;
  FAsNeeded := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetSite(value : TFhirType);
begin
  FSite.free;
  FSite := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetRoute(value : TFhirCodeableConcept);
begin
  FRoute.free;
  FRoute := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetDose(value : TFhirType);
begin
  FDose.free;
  FDose := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetRate(value : TFhirType);
begin
  FRate.free;
  FRate := value;
end;

Procedure TFhirMedicationOrderDosageInstruction.SetMaxDosePerPeriod(value : TFhirRatio);
begin
  FMaxDosePerPeriod.free;
  FMaxDosePerPeriod := value;
end;


{ TFhirMedicationOrderDosageInstructionListEnumerator }

Constructor TFhirMedicationOrderDosageInstructionListEnumerator.Create(list : TFhirMedicationOrderDosageInstructionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationOrderDosageInstructionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationOrderDosageInstructionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationOrderDosageInstructionListEnumerator.GetCurrent : TFhirMedicationOrderDosageInstruction;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationOrderDosageInstructionList }
procedure TFhirMedicationOrderDosageInstructionList.AddItem(value: TFhirMedicationOrderDosageInstruction);
begin
  assert(value.ClassName = 'TFhirMedicationOrderDosageInstruction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationOrderDosageInstruction');
  add(value);
end;


function TFhirMedicationOrderDosageInstructionList.Append: TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderDosageInstructionList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationOrderDosageInstructionList.GetEnumerator : TFhirMedicationOrderDosageInstructionListEnumerator;
begin
  result := TFhirMedicationOrderDosageInstructionListEnumerator.Create(self.link);
end;

function TFhirMedicationOrderDosageInstructionList.Clone: TFhirMedicationOrderDosageInstructionList;
begin
  result := TFhirMedicationOrderDosageInstructionList(inherited Clone);
end;

function TFhirMedicationOrderDosageInstructionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationOrderDosageInstructionList.GetItemN(index: Integer): TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction(ObjectByIndex[index]);
end;

function TFhirMedicationOrderDosageInstructionList.IndexOf(value: TFhirMedicationOrderDosageInstruction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationOrderDosageInstructionList.Insert(index: Integer): TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderDosageInstructionList.InsertItem(index: Integer; value: TFhirMedicationOrderDosageInstruction);
begin
  assert(value is TFhirMedicationOrderDosageInstruction);
  Inherited Insert(index, value);
end;

function TFhirMedicationOrderDosageInstructionList.Item(index: Integer): TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction(ObjectByIndex[index]);
end;

function TFhirMedicationOrderDosageInstructionList.Link: TFhirMedicationOrderDosageInstructionList;
begin
  result := TFhirMedicationOrderDosageInstructionList(inherited Link);
end;

procedure TFhirMedicationOrderDosageInstructionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationOrderDosageInstructionList.SetItemByIndex(index: Integer; value: TFhirMedicationOrderDosageInstruction);
begin
  assert(value is TFhirMedicationOrderDosageInstruction);
  FhirMedicationOrderDosageInstructions[index] := value;
end;

procedure TFhirMedicationOrderDosageInstructionList.SetItemN(index: Integer; value: TFhirMedicationOrderDosageInstruction);
begin
  assert(value is TFhirMedicationOrderDosageInstruction);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationOrderDispenseRequest }

constructor TFhirMedicationOrderDispenseRequest.Create;
begin
  inherited;
end;

destructor TFhirMedicationOrderDispenseRequest.Destroy;
begin
  FMedication.free;
  FValidityPeriod.free;
  FNumberOfRepeatsAllowed.free;
  FQuantity.free;
  FExpectedSupplyDuration.free;
  inherited;
end;

procedure TFhirMedicationOrderDispenseRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  medication := TFhirMedicationOrderDispenseRequest(oSource).medication.Clone;
  validityPeriod := TFhirMedicationOrderDispenseRequest(oSource).validityPeriod.Clone;
  numberOfRepeatsAllowedElement := TFhirMedicationOrderDispenseRequest(oSource).numberOfRepeatsAllowedElement.Clone;
  quantity := TFhirMedicationOrderDispenseRequest(oSource).quantity.Clone;
  expectedSupplyDuration := TFhirMedicationOrderDispenseRequest(oSource).expectedSupplyDuration.Clone;
end;

procedure TFhirMedicationOrderDispenseRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'medication[x]') Then
     list.add(FMedication.Link);
  if (child_name = 'validityPeriod') Then
     list.add(FValidityPeriod.Link);
  if (child_name = 'numberOfRepeatsAllowed') Then
     list.add(FNumberOfRepeatsAllowed.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'expectedSupplyDuration') Then
     list.add(FExpectedSupplyDuration.Link);
end;

procedure TFhirMedicationOrderDispenseRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'medication[x]', 'CodeableConcept|Reference(Medication)', FMedication.Link));{2}
  oList.add(TFHIRProperty.create(self, 'validityPeriod', 'Period', FValidityPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'numberOfRepeatsAllowed', 'positiveInt', FNumberOfRepeatsAllowed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expectedSupplyDuration', 'Duration', FExpectedSupplyDuration.Link));{2}
end;

procedure TFhirMedicationOrderDispenseRequest.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('medication')) then Medication := propValue as TFhirType{4}
  else if (propName = 'validityPeriod') then ValidityPeriod := propValue as TFhirPeriod{4b}
  else if (propName = 'numberOfRepeatsAllowed') then NumberOfRepeatsAllowedElement := propValue as TFhirPositiveInt{5a}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'expectedSupplyDuration') then ExpectedSupplyDuration := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirMedicationOrderDispenseRequest.FhirType : string;
begin
  result := 'dispenseRequest';
end;

function TFhirMedicationOrderDispenseRequest.Link : TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest(inherited Link);
end;

function TFhirMedicationOrderDispenseRequest.Clone : TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest(inherited Clone);
end;

function TFhirMedicationOrderDispenseRequest.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderDispenseRequest;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderDispenseRequest)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderDispenseRequest(other);
    result := compareDeep(medicationElement, o.medicationElement, true) and compareDeep(validityPeriodElement, o.validityPeriodElement, true) and 
      compareDeep(numberOfRepeatsAllowedElement, o.numberOfRepeatsAllowedElement, true) and 
      compareDeep(quantityElement, o.quantityElement, true) and compareDeep(expectedSupplyDurationElement, o.expectedSupplyDurationElement, true);
  end;
end;

function TFhirMedicationOrderDispenseRequest.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderDispenseRequest;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderDispenseRequest)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderDispenseRequest(other);
    result := compareValues(numberOfRepeatsAllowedElement, o.numberOfRepeatsAllowedElement, true);
  end;
end;

{ TFhirMedicationOrderDispenseRequest }

Procedure TFhirMedicationOrderDispenseRequest.SetMedication(value : TFhirType);
begin
  FMedication.free;
  FMedication := value;
end;

Procedure TFhirMedicationOrderDispenseRequest.SetValidityPeriod(value : TFhirPeriod);
begin
  FValidityPeriod.free;
  FValidityPeriod := value;
end;

Procedure TFhirMedicationOrderDispenseRequest.SetNumberOfRepeatsAllowed(value : TFhirPositiveInt);
begin
  FNumberOfRepeatsAllowed.free;
  FNumberOfRepeatsAllowed := value;
end;

Function TFhirMedicationOrderDispenseRequest.GetNumberOfRepeatsAllowedST : String;
begin
  if FNumberOfRepeatsAllowed = nil then
    result := ''
  else
    result := FNumberOfRepeatsAllowed.value;
end;

Procedure TFhirMedicationOrderDispenseRequest.SetNumberOfRepeatsAllowedST(value : String);
begin
  if value <> '' then
  begin
    if FNumberOfRepeatsAllowed = nil then
      FNumberOfRepeatsAllowed := TFhirPositiveInt.create;
    FNumberOfRepeatsAllowed.value := value
  end
  else if FNumberOfRepeatsAllowed <> nil then
    FNumberOfRepeatsAllowed.value := '';
end;

Procedure TFhirMedicationOrderDispenseRequest.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirMedicationOrderDispenseRequest.SetExpectedSupplyDuration(value : TFhirQuantity);
begin
  FExpectedSupplyDuration.free;
  FExpectedSupplyDuration := value;
end;


{ TFhirMedicationOrderDispenseRequestListEnumerator }

Constructor TFhirMedicationOrderDispenseRequestListEnumerator.Create(list : TFhirMedicationOrderDispenseRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationOrderDispenseRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationOrderDispenseRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationOrderDispenseRequestListEnumerator.GetCurrent : TFhirMedicationOrderDispenseRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationOrderDispenseRequestList }
procedure TFhirMedicationOrderDispenseRequestList.AddItem(value: TFhirMedicationOrderDispenseRequest);
begin
  assert(value.ClassName = 'TFhirMedicationOrderDispenseRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationOrderDispenseRequest');
  add(value);
end;


function TFhirMedicationOrderDispenseRequestList.Append: TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderDispenseRequestList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationOrderDispenseRequestList.GetEnumerator : TFhirMedicationOrderDispenseRequestListEnumerator;
begin
  result := TFhirMedicationOrderDispenseRequestListEnumerator.Create(self.link);
end;

function TFhirMedicationOrderDispenseRequestList.Clone: TFhirMedicationOrderDispenseRequestList;
begin
  result := TFhirMedicationOrderDispenseRequestList(inherited Clone);
end;

function TFhirMedicationOrderDispenseRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationOrderDispenseRequestList.GetItemN(index: Integer): TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest(ObjectByIndex[index]);
end;

function TFhirMedicationOrderDispenseRequestList.IndexOf(value: TFhirMedicationOrderDispenseRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationOrderDispenseRequestList.Insert(index: Integer): TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderDispenseRequestList.InsertItem(index: Integer; value: TFhirMedicationOrderDispenseRequest);
begin
  assert(value is TFhirMedicationOrderDispenseRequest);
  Inherited Insert(index, value);
end;

function TFhirMedicationOrderDispenseRequestList.Item(index: Integer): TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest(ObjectByIndex[index]);
end;

function TFhirMedicationOrderDispenseRequestList.Link: TFhirMedicationOrderDispenseRequestList;
begin
  result := TFhirMedicationOrderDispenseRequestList(inherited Link);
end;

procedure TFhirMedicationOrderDispenseRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationOrderDispenseRequestList.SetItemByIndex(index: Integer; value: TFhirMedicationOrderDispenseRequest);
begin
  assert(value is TFhirMedicationOrderDispenseRequest);
  FhirMedicationOrderDispenseRequests[index] := value;
end;

procedure TFhirMedicationOrderDispenseRequestList.SetItemN(index: Integer; value: TFhirMedicationOrderDispenseRequest);
begin
  assert(value is TFhirMedicationOrderDispenseRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationOrderSubstitution }

constructor TFhirMedicationOrderSubstitution.Create;
begin
  inherited;
end;

destructor TFhirMedicationOrderSubstitution.Destroy;
begin
  FType_.free;
  FReason.free;
  inherited;
end;

procedure TFhirMedicationOrderSubstitution.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirMedicationOrderSubstitution(oSource).type_.Clone;
  reason := TFhirMedicationOrderSubstitution(oSource).reason.Clone;
end;

procedure TFhirMedicationOrderSubstitution.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
end;

procedure TFhirMedicationOrderSubstitution.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReason.Link));{2}
end;

procedure TFhirMedicationOrderSubstitution.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'reason') then Reason := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirMedicationOrderSubstitution.FhirType : string;
begin
  result := 'substitution';
end;

function TFhirMedicationOrderSubstitution.Link : TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution(inherited Link);
end;

function TFhirMedicationOrderSubstitution.Clone : TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution(inherited Clone);
end;

function TFhirMedicationOrderSubstitution.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderSubstitution;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderSubstitution)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderSubstitution(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(reasonElement, o.reasonElement, true);
  end;
end;

function TFhirMedicationOrderSubstitution.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationOrderSubstitution;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationOrderSubstitution)) then
    result := false
  else
  begin
    o := TFhirMedicationOrderSubstitution(other);
    result := true;
  end;
end;

{ TFhirMedicationOrderSubstitution }

Procedure TFhirMedicationOrderSubstitution.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirMedicationOrderSubstitution.SetReason(value : TFhirCodeableConcept);
begin
  FReason.free;
  FReason := value;
end;


{ TFhirMedicationOrderSubstitutionListEnumerator }

Constructor TFhirMedicationOrderSubstitutionListEnumerator.Create(list : TFhirMedicationOrderSubstitutionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationOrderSubstitutionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationOrderSubstitutionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationOrderSubstitutionListEnumerator.GetCurrent : TFhirMedicationOrderSubstitution;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationOrderSubstitutionList }
procedure TFhirMedicationOrderSubstitutionList.AddItem(value: TFhirMedicationOrderSubstitution);
begin
  assert(value.ClassName = 'TFhirMedicationOrderSubstitution', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationOrderSubstitution');
  add(value);
end;


function TFhirMedicationOrderSubstitutionList.Append: TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderSubstitutionList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationOrderSubstitutionList.GetEnumerator : TFhirMedicationOrderSubstitutionListEnumerator;
begin
  result := TFhirMedicationOrderSubstitutionListEnumerator.Create(self.link);
end;

function TFhirMedicationOrderSubstitutionList.Clone: TFhirMedicationOrderSubstitutionList;
begin
  result := TFhirMedicationOrderSubstitutionList(inherited Clone);
end;

function TFhirMedicationOrderSubstitutionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationOrderSubstitutionList.GetItemN(index: Integer): TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution(ObjectByIndex[index]);
end;

function TFhirMedicationOrderSubstitutionList.IndexOf(value: TFhirMedicationOrderSubstitution): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationOrderSubstitutionList.Insert(index: Integer): TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderSubstitutionList.InsertItem(index: Integer; value: TFhirMedicationOrderSubstitution);
begin
  assert(value is TFhirMedicationOrderSubstitution);
  Inherited Insert(index, value);
end;

function TFhirMedicationOrderSubstitutionList.Item(index: Integer): TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution(ObjectByIndex[index]);
end;

function TFhirMedicationOrderSubstitutionList.Link: TFhirMedicationOrderSubstitutionList;
begin
  result := TFhirMedicationOrderSubstitutionList(inherited Link);
end;

procedure TFhirMedicationOrderSubstitutionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationOrderSubstitutionList.SetItemByIndex(index: Integer; value: TFhirMedicationOrderSubstitution);
begin
  assert(value is TFhirMedicationOrderSubstitution);
  FhirMedicationOrderSubstitutions[index] := value;
end;

procedure TFhirMedicationOrderSubstitutionList.SetItemN(index: Integer; value: TFhirMedicationOrderSubstitution);
begin
  assert(value is TFhirMedicationOrderSubstitution);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationOrder }

constructor TFhirMedicationOrder.Create;
begin
  inherited;
end;

destructor TFhirMedicationOrder.Destroy;
begin
  FIdentifierList.Free;
  FDateWritten.free;
  FStatus.free;
  FDateEnded.free;
  FReasonEnded.free;
  FPatient.free;
  FPrescriber.free;
  FEncounter.free;
  FReason.free;
  FNote.free;
  FMedication.free;
  FDosageInstructionList.Free;
  FDispenseRequest.free;
  FSubstitution.free;
  FPriorPrescription.free;
  inherited;
end;

function TFhirMedicationOrder.GetResourceType : TFhirResourceType;
begin
  result := frtMedicationOrder;
end;

procedure TFhirMedicationOrder.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirMedicationOrder(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirMedicationOrder(oSource).FIdentifierList);
  end;
  dateWrittenElement := TFhirMedicationOrder(oSource).dateWrittenElement.Clone;
  FStatus := TFhirMedicationOrder(oSource).FStatus.Link;
  dateEndedElement := TFhirMedicationOrder(oSource).dateEndedElement.Clone;
  reasonEnded := TFhirMedicationOrder(oSource).reasonEnded.Clone;
  patient := TFhirMedicationOrder(oSource).patient.Clone;
  prescriber := TFhirMedicationOrder(oSource).prescriber.Clone;
  encounter := TFhirMedicationOrder(oSource).encounter.Clone;
  reason := TFhirMedicationOrder(oSource).reason.Clone;
  noteElement := TFhirMedicationOrder(oSource).noteElement.Clone;
  medication := TFhirMedicationOrder(oSource).medication.Clone;
  if (TFhirMedicationOrder(oSource).FDosageInstructionList = nil) then
  begin
    FDosageInstructionList.free;
    FDosageInstructionList := nil;
  end
  else
  begin
    FDosageInstructionList := TFhirMedicationOrderDosageInstructionList.Create;
    FDosageInstructionList.Assign(TFhirMedicationOrder(oSource).FDosageInstructionList);
  end;
  dispenseRequest := TFhirMedicationOrder(oSource).dispenseRequest.Clone;
  substitution := TFhirMedicationOrder(oSource).substitution.Clone;
  priorPrescription := TFhirMedicationOrder(oSource).priorPrescription.Clone;
end;

procedure TFhirMedicationOrder.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'dateWritten') Then
     list.add(FDateWritten.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'dateEnded') Then
     list.add(FDateEnded.Link);
  if (child_name = 'reasonEnded') Then
     list.add(FReasonEnded.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'prescriber') Then
     list.add(FPrescriber.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'medication[x]') Then
     list.add(FMedication.Link);
  if (child_name = 'dosageInstruction') Then
    list.addAll(FDosageInstructionList);
  if (child_name = 'dispenseRequest') Then
     list.add(FDispenseRequest.Link);
  if (child_name = 'substitution') Then
     list.add(FSubstitution.Link);
  if (child_name = 'priorPrescription') Then
     list.add(FPriorPrescription.Link);
end;

procedure TFhirMedicationOrder.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'dateWritten', 'dateTime', FDateWritten.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'dateEnded', 'dateTime', FDateEnded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reasonEnded', 'CodeableConcept', FReasonEnded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prescriber', 'Reference(Practitioner)', FPrescriber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Condition)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'string', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'medication[x]', 'CodeableConcept|Reference(Medication)', FMedication.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dosageInstruction', '', FDosageInstructionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'dispenseRequest', '', FDispenseRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'substitution', '', FSubstitution.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priorPrescription', 'Reference(MedicationOrder)', FPriorPrescription.Link));{2}
end;

procedure TFhirMedicationOrder.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'dateWritten') then DateWrittenElement := propValue as TFhirDateTime{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'dateEnded') then DateEndedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'reasonEnded') then ReasonEnded := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'prescriber') then Prescriber := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName = 'note') then NoteElement := propValue as TFhirString{5a}
  else if (propName.startsWith('medication')) then Medication := propValue as TFhirType{4}
  else if (propName = 'dosageInstruction') then DosageInstructionList.add(propValue as TFhirMedicationOrderDosageInstruction){2}
  else if (propName = 'dispenseRequest') then DispenseRequest := propValue as TFhirMedicationOrderDispenseRequest{4b}
  else if (propName = 'substitution') then Substitution := propValue as TFhirMedicationOrderSubstitution{4b}
  else if (propName = 'priorPrescription') then PriorPrescription := propValue as TFhirReference{TFhirMedicationOrder}{4b}
  else inherited;
end;

function TFhirMedicationOrder.FhirType : string;
begin
  result := 'MedicationOrder';
end;

function TFhirMedicationOrder.Link : TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder(inherited Link);
end;

function TFhirMedicationOrder.Clone : TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder(inherited Clone);
end;

{ TFhirMedicationOrder }

Function TFhirMedicationOrder.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirMedicationOrder.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirMedicationOrder.SetDateWritten(value : TFhirDateTime);
begin
  FDateWritten.free;
  FDateWritten := value;
end;

Function TFhirMedicationOrder.GetDateWrittenST : TDateAndTime;
begin
  if FDateWritten = nil then
    result := nil
  else
    result := FDateWritten.value;
end;

Procedure TFhirMedicationOrder.SetDateWrittenST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateWritten = nil then
      FDateWritten := TFhirDateTime.create;
    FDateWritten.value := value
  end
  else if FDateWritten <> nil then
    FDateWritten.value := nil;
end;

Procedure TFhirMedicationOrder.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirMedicationOrder.GetStatusST : TFhirMedicationOrderStatus;
begin
  if FStatus = nil then
    result := TFhirMedicationOrderStatus(0)
  else
    result := TFhirMedicationOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirMedicationOrderStatus, FStatus.value));
end;

Procedure TFhirMedicationOrder.SetStatusST(value : TFhirMedicationOrderStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirMedicationOrderStatus[value]);
end;

Procedure TFhirMedicationOrder.SetDateEnded(value : TFhirDateTime);
begin
  FDateEnded.free;
  FDateEnded := value;
end;

Function TFhirMedicationOrder.GetDateEndedST : TDateAndTime;
begin
  if FDateEnded = nil then
    result := nil
  else
    result := FDateEnded.value;
end;

Procedure TFhirMedicationOrder.SetDateEndedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateEnded = nil then
      FDateEnded := TFhirDateTime.create;
    FDateEnded.value := value
  end
  else if FDateEnded <> nil then
    FDateEnded.value := nil;
end;

Procedure TFhirMedicationOrder.SetReasonEnded(value : TFhirCodeableConcept);
begin
  FReasonEnded.free;
  FReasonEnded := value;
end;

Procedure TFhirMedicationOrder.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirMedicationOrder.SetPrescriber(value : TFhirReference{TFhirPractitioner});
begin
  FPrescriber.free;
  FPrescriber := value;
end;

Procedure TFhirMedicationOrder.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirMedicationOrder.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirMedicationOrder.SetNote(value : TFhirString);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirMedicationOrder.GetNoteST : String;
begin
  if FNote = nil then
    result := ''
  else
    result := FNote.value;
end;

Procedure TFhirMedicationOrder.SetNoteST(value : String);
begin
  if value <> '' then
  begin
    if FNote = nil then
      FNote := TFhirString.create;
    FNote.value := value
  end
  else if FNote <> nil then
    FNote.value := '';
end;

Procedure TFhirMedicationOrder.SetMedication(value : TFhirType);
begin
  FMedication.free;
  FMedication := value;
end;

Function TFhirMedicationOrder.GetDosageInstructionList : TFhirMedicationOrderDosageInstructionList;
begin
  if FDosageInstructionList = nil then
    FDosageInstructionList := TFhirMedicationOrderDosageInstructionList.Create;
  result := FDosageInstructionList;
end;

Function TFhirMedicationOrder.GetHasDosageInstructionList : boolean;
begin
  result := (FDosageInstructionList <> nil) and (FDosageInstructionList.count > 0);
end;

Procedure TFhirMedicationOrder.SetDispenseRequest(value : TFhirMedicationOrderDispenseRequest);
begin
  FDispenseRequest.free;
  FDispenseRequest := value;
end;

Procedure TFhirMedicationOrder.SetSubstitution(value : TFhirMedicationOrderSubstitution);
begin
  FSubstitution.free;
  FSubstitution := value;
end;

Procedure TFhirMedicationOrder.SetPriorPrescription(value : TFhirReference{TFhirMedicationOrder});
begin
  FPriorPrescription.free;
  FPriorPrescription := value;
end;


{ TFhirMedicationOrderListEnumerator }

Constructor TFhirMedicationOrderListEnumerator.Create(list : TFhirMedicationOrderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationOrderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationOrderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationOrderListEnumerator.GetCurrent : TFhirMedicationOrder;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationOrderList }
procedure TFhirMedicationOrderList.AddItem(value: TFhirMedicationOrder);
begin
  assert(value.ClassName = 'TFhirMedicationOrder', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationOrder');
  add(value);
end;


function TFhirMedicationOrderList.Append: TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationOrderList.GetEnumerator : TFhirMedicationOrderListEnumerator;
begin
  result := TFhirMedicationOrderListEnumerator.Create(self.link);
end;

function TFhirMedicationOrderList.Clone: TFhirMedicationOrderList;
begin
  result := TFhirMedicationOrderList(inherited Clone);
end;

function TFhirMedicationOrderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationOrderList.GetItemN(index: Integer): TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder(ObjectByIndex[index]);
end;

function TFhirMedicationOrderList.IndexOf(value: TFhirMedicationOrder): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationOrderList.Insert(index: Integer): TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationOrderList.InsertItem(index: Integer; value: TFhirMedicationOrder);
begin
  assert(value is TFhirMedicationOrder);
  Inherited Insert(index, value);
end;

function TFhirMedicationOrderList.Item(index: Integer): TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder(ObjectByIndex[index]);
end;

function TFhirMedicationOrderList.Link: TFhirMedicationOrderList;
begin
  result := TFhirMedicationOrderList(inherited Link);
end;

procedure TFhirMedicationOrderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationOrderList.SetItemByIndex(index: Integer; value: TFhirMedicationOrder);
begin
  assert(value is TFhirMedicationOrder);
  FhirMedicationOrders[index] := value;
end;

procedure TFhirMedicationOrderList.SetItemN(index: Integer; value: TFhirMedicationOrder);
begin
  assert(value is TFhirMedicationOrder);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationStatementDosage }

constructor TFhirMedicationStatementDosage.Create;
begin
  inherited;
end;

destructor TFhirMedicationStatementDosage.Destroy;
begin
  FText.free;
  FTiming.free;
  FAsNeeded.free;
  FSite.free;
  FRoute.free;
  FMethod.free;
  FQuantity.free;
  FRate.free;
  FMaxDosePerPeriod.free;
  inherited;
end;

procedure TFhirMedicationStatementDosage.Assign(oSource : TAdvObject);
begin
  inherited;
  textElement := TFhirMedicationStatementDosage(oSource).textElement.Clone;
  timing := TFhirMedicationStatementDosage(oSource).timing.Clone;
  asNeeded := TFhirMedicationStatementDosage(oSource).asNeeded.Clone;
  site := TFhirMedicationStatementDosage(oSource).site.Clone;
  route := TFhirMedicationStatementDosage(oSource).route.Clone;
  method := TFhirMedicationStatementDosage(oSource).method.Clone;
  quantity := TFhirMedicationStatementDosage(oSource).quantity.Clone;
  rate := TFhirMedicationStatementDosage(oSource).rate.Clone;
  maxDosePerPeriod := TFhirMedicationStatementDosage(oSource).maxDosePerPeriod.Clone;
end;

procedure TFhirMedicationStatementDosage.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'timing') Then
     list.add(FTiming.Link);
  if (child_name = 'asNeeded[x]') Then
     list.add(FAsNeeded.Link);
  if (child_name = 'site[x]') Then
     list.add(FSite.Link);
  if (child_name = 'route') Then
     list.add(FRoute.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'quantity[x]') Then
     list.add(FQuantity.Link);
  if (child_name = 'rate[x]') Then
     list.add(FRate.Link);
  if (child_name = 'maxDosePerPeriod') Then
     list.add(FMaxDosePerPeriod.Link);
end;

procedure TFhirMedicationStatementDosage.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timing', 'Timing', FTiming.Link));{2}
  oList.add(TFHIRProperty.create(self, 'asNeeded[x]', 'boolean|CodeableConcept', FAsNeeded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'site[x]', 'CodeableConcept|Reference(BodySite)', FSite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'route', 'CodeableConcept', FRoute.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity[x]', 'SimpleQuantity|Range', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rate[x]', 'Ratio|Range', FRate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'maxDosePerPeriod', 'Ratio', FMaxDosePerPeriod.Link));{2}
end;

procedure TFhirMedicationStatementDosage.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'timing') then Timing := propValue as TFhirTiming{4b}
  else if (propName.startsWith('asNeeded')) then AsNeeded := propValue as TFhirType{4}
  else if (propName.startsWith('site')) then Site := propValue as TFhirType{4}
  else if (propName = 'route') then Route := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('quantity')) then Quantity := propValue as TFhirType{4}
  else if (propName.startsWith('rate')) then Rate := propValue as TFhirType{4}
  else if (propName = 'maxDosePerPeriod') then MaxDosePerPeriod := propValue as TFhirRatio{4b}
  else inherited;
end;

function TFhirMedicationStatementDosage.FhirType : string;
begin
  result := 'dosage';
end;

function TFhirMedicationStatementDosage.Link : TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage(inherited Link);
end;

function TFhirMedicationStatementDosage.Clone : TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage(inherited Clone);
end;

function TFhirMedicationStatementDosage.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationStatementDosage;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMedicationStatementDosage)) then
    result := false
  else
  begin
    o := TFhirMedicationStatementDosage(other);
    result := compareDeep(textElement, o.textElement, true) and compareDeep(timingElement, o.timingElement, true) and 
      compareDeep(asNeededElement, o.asNeededElement, true) and compareDeep(siteElement, o.siteElement, true) and 
      compareDeep(routeElement, o.routeElement, true) and compareDeep(methodElement, o.methodElement, true) and 
      compareDeep(quantityElement, o.quantityElement, true) and compareDeep(rateElement, o.rateElement, true) and 
      compareDeep(maxDosePerPeriodElement, o.maxDosePerPeriodElement, true);
  end;
end;

function TFhirMedicationStatementDosage.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMedicationStatementDosage;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMedicationStatementDosage)) then
    result := false
  else
  begin
    o := TFhirMedicationStatementDosage(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirMedicationStatementDosage }

Procedure TFhirMedicationStatementDosage.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirMedicationStatementDosage.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirMedicationStatementDosage.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirMedicationStatementDosage.SetTiming(value : TFhirTiming);
begin
  FTiming.free;
  FTiming := value;
end;

Procedure TFhirMedicationStatementDosage.SetAsNeeded(value : TFhirType);
begin
  FAsNeeded.free;
  FAsNeeded := value;
end;

Procedure TFhirMedicationStatementDosage.SetSite(value : TFhirType);
begin
  FSite.free;
  FSite := value;
end;

Procedure TFhirMedicationStatementDosage.SetRoute(value : TFhirCodeableConcept);
begin
  FRoute.free;
  FRoute := value;
end;

Procedure TFhirMedicationStatementDosage.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirMedicationStatementDosage.SetQuantity(value : TFhirType);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirMedicationStatementDosage.SetRate(value : TFhirType);
begin
  FRate.free;
  FRate := value;
end;

Procedure TFhirMedicationStatementDosage.SetMaxDosePerPeriod(value : TFhirRatio);
begin
  FMaxDosePerPeriod.free;
  FMaxDosePerPeriod := value;
end;


{ TFhirMedicationStatementDosageListEnumerator }

Constructor TFhirMedicationStatementDosageListEnumerator.Create(list : TFhirMedicationStatementDosageList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationStatementDosageListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationStatementDosageListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationStatementDosageListEnumerator.GetCurrent : TFhirMedicationStatementDosage;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationStatementDosageList }
procedure TFhirMedicationStatementDosageList.AddItem(value: TFhirMedicationStatementDosage);
begin
  assert(value.ClassName = 'TFhirMedicationStatementDosage', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationStatementDosage');
  add(value);
end;


function TFhirMedicationStatementDosageList.Append: TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationStatementDosageList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationStatementDosageList.GetEnumerator : TFhirMedicationStatementDosageListEnumerator;
begin
  result := TFhirMedicationStatementDosageListEnumerator.Create(self.link);
end;

function TFhirMedicationStatementDosageList.Clone: TFhirMedicationStatementDosageList;
begin
  result := TFhirMedicationStatementDosageList(inherited Clone);
end;

function TFhirMedicationStatementDosageList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationStatementDosageList.GetItemN(index: Integer): TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage(ObjectByIndex[index]);
end;

function TFhirMedicationStatementDosageList.IndexOf(value: TFhirMedicationStatementDosage): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationStatementDosageList.Insert(index: Integer): TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationStatementDosageList.InsertItem(index: Integer; value: TFhirMedicationStatementDosage);
begin
  assert(value is TFhirMedicationStatementDosage);
  Inherited Insert(index, value);
end;

function TFhirMedicationStatementDosageList.Item(index: Integer): TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage(ObjectByIndex[index]);
end;

function TFhirMedicationStatementDosageList.Link: TFhirMedicationStatementDosageList;
begin
  result := TFhirMedicationStatementDosageList(inherited Link);
end;

procedure TFhirMedicationStatementDosageList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationStatementDosageList.SetItemByIndex(index: Integer; value: TFhirMedicationStatementDosage);
begin
  assert(value is TFhirMedicationStatementDosage);
  FhirMedicationStatementDosages[index] := value;
end;

procedure TFhirMedicationStatementDosageList.SetItemN(index: Integer; value: TFhirMedicationStatementDosage);
begin
  assert(value is TFhirMedicationStatementDosage);
  ObjectByIndex[index] := value;
end;

{ TFhirMedicationStatement }

constructor TFhirMedicationStatement.Create;
begin
  inherited;
end;

destructor TFhirMedicationStatement.Destroy;
begin
  FIdentifierList.Free;
  FPatient.free;
  FInformationSource.free;
  FDateAsserted.free;
  FStatus.free;
  FWasNotTaken.free;
  FReasonNotTakenList.Free;
  FReasonForUse.free;
  FEffective.free;
  FNote.free;
  FSupportingInformationList.Free;
  FMedication.free;
  FDosageList.Free;
  inherited;
end;

function TFhirMedicationStatement.GetResourceType : TFhirResourceType;
begin
  result := frtMedicationStatement;
end;

procedure TFhirMedicationStatement.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirMedicationStatement(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirMedicationStatement(oSource).FIdentifierList);
  end;
  patient := TFhirMedicationStatement(oSource).patient.Clone;
  informationSource := TFhirMedicationStatement(oSource).informationSource.Clone;
  dateAssertedElement := TFhirMedicationStatement(oSource).dateAssertedElement.Clone;
  FStatus := TFhirMedicationStatement(oSource).FStatus.Link;
  wasNotTakenElement := TFhirMedicationStatement(oSource).wasNotTakenElement.Clone;
  if (TFhirMedicationStatement(oSource).FReasonNotTakenList = nil) then
  begin
    FReasonNotTakenList.free;
    FReasonNotTakenList := nil;
  end
  else
  begin
    FReasonNotTakenList := TFhirCodeableConceptList.Create;
    FReasonNotTakenList.Assign(TFhirMedicationStatement(oSource).FReasonNotTakenList);
  end;
  reasonForUse := TFhirMedicationStatement(oSource).reasonForUse.Clone;
  effective := TFhirMedicationStatement(oSource).effective.Clone;
  noteElement := TFhirMedicationStatement(oSource).noteElement.Clone;
  if (TFhirMedicationStatement(oSource).FSupportingInformationList = nil) then
  begin
    FSupportingInformationList.free;
    FSupportingInformationList := nil;
  end
  else
  begin
    FSupportingInformationList := TFhirReferenceList{TFhirReference}.Create;
    FSupportingInformationList.Assign(TFhirMedicationStatement(oSource).FSupportingInformationList);
  end;
  medication := TFhirMedicationStatement(oSource).medication.Clone;
  if (TFhirMedicationStatement(oSource).FDosageList = nil) then
  begin
    FDosageList.free;
    FDosageList := nil;
  end
  else
  begin
    FDosageList := TFhirMedicationStatementDosageList.Create;
    FDosageList.Assign(TFhirMedicationStatement(oSource).FDosageList);
  end;
end;

procedure TFhirMedicationStatement.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'informationSource') Then
     list.add(FInformationSource.Link);
  if (child_name = 'dateAsserted') Then
     list.add(FDateAsserted.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'wasNotTaken') Then
     list.add(FWasNotTaken.Link);
  if (child_name = 'reasonNotTaken') Then
    list.addAll(FReasonNotTakenList);
  if (child_name = 'reasonForUse[x]') Then
     list.add(FReasonForUse.Link);
  if (child_name = 'effective[x]') Then
     list.add(FEffective.Link);
  if (child_name = 'note') Then
     list.add(FNote.Link);
  if (child_name = 'supportingInformation') Then
    list.addAll(FSupportingInformationList);
  if (child_name = 'medication[x]') Then
     list.add(FMedication.Link);
  if (child_name = 'dosage') Then
    list.addAll(FDosageList);
end;

procedure TFhirMedicationStatement.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'informationSource', 'Reference(Patient|Practitioner|RelatedPerson)', FInformationSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateAsserted', 'dateTime', FDateAsserted.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'wasNotTaken', 'boolean', FWasNotTaken.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reasonNotTaken', 'CodeableConcept', FReasonNotTakenList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reasonForUse[x]', 'CodeableConcept|Reference(Condition)', FReasonForUse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'effective[x]', 'dateTime|Period', FEffective.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', 'string', FNote.Link));{2}
  oList.add(TFHIRProperty.create(self, 'supportingInformation', 'Reference(Any)', FSupportingInformationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'medication[x]', 'CodeableConcept|Reference(Medication)', FMedication.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dosage', '', FDosageList.Link)){3};
end;

procedure TFhirMedicationStatement.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'informationSource') then InformationSource := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'dateAsserted') then DateAssertedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'wasNotTaken') then WasNotTakenElement := propValue as TFhirBoolean{5a}
  else if (propName = 'reasonNotTaken') then ReasonNotTakenList.add(propValue as TFhirCodeableConcept){2}
  else if (propName.startsWith('reasonForUse')) then ReasonForUse := propValue as TFhirType{4}
  else if (propName.startsWith('effective')) then Effective := propValue as TFhirType{4}
  else if (propName = 'note') then NoteElement := propValue as TFhirString{5a}
  else if (propName = 'supportingInformation') then SupportingInformationList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName.startsWith('medication')) then Medication := propValue as TFhirType{4}
  else if (propName = 'dosage') then DosageList.add(propValue as TFhirMedicationStatementDosage){2}
  else inherited;
end;

function TFhirMedicationStatement.FhirType : string;
begin
  result := 'MedicationStatement';
end;

function TFhirMedicationStatement.Link : TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement(inherited Link);
end;

function TFhirMedicationStatement.Clone : TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement(inherited Clone);
end;

{ TFhirMedicationStatement }

Function TFhirMedicationStatement.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirMedicationStatement.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirMedicationStatement.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirMedicationStatement.SetInformationSource(value : TFhirReference{Resource});
begin
  FInformationSource.free;
  FInformationSource := value;
end;

Procedure TFhirMedicationStatement.SetDateAsserted(value : TFhirDateTime);
begin
  FDateAsserted.free;
  FDateAsserted := value;
end;

Function TFhirMedicationStatement.GetDateAssertedST : TDateAndTime;
begin
  if FDateAsserted = nil then
    result := nil
  else
    result := FDateAsserted.value;
end;

Procedure TFhirMedicationStatement.SetDateAssertedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateAsserted = nil then
      FDateAsserted := TFhirDateTime.create;
    FDateAsserted.value := value
  end
  else if FDateAsserted <> nil then
    FDateAsserted.value := nil;
end;

Procedure TFhirMedicationStatement.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirMedicationStatement.GetStatusST : TFhirMedicationStatementStatus;
begin
  if FStatus = nil then
    result := TFhirMedicationStatementStatus(0)
  else
    result := TFhirMedicationStatementStatus(StringArrayIndexOfSensitive(CODES_TFhirMedicationStatementStatus, FStatus.value));
end;

Procedure TFhirMedicationStatement.SetStatusST(value : TFhirMedicationStatementStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirMedicationStatementStatus[value]);
end;

Procedure TFhirMedicationStatement.SetWasNotTaken(value : TFhirBoolean);
begin
  FWasNotTaken.free;
  FWasNotTaken := value;
end;

Function TFhirMedicationStatement.GetWasNotTakenST : Boolean;
begin
  if FWasNotTaken = nil then
    result := false
  else
    result := FWasNotTaken.value;
end;

Procedure TFhirMedicationStatement.SetWasNotTakenST(value : Boolean);
begin
  if FWasNotTaken = nil then
    FWasNotTaken := TFhirBoolean.create;
  FWasNotTaken.value := value
end;

Function TFhirMedicationStatement.GetReasonNotTakenList : TFhirCodeableConceptList;
begin
  if FReasonNotTakenList = nil then
    FReasonNotTakenList := TFhirCodeableConceptList.Create;
  result := FReasonNotTakenList;
end;

Function TFhirMedicationStatement.GetHasReasonNotTakenList : boolean;
begin
  result := (FReasonNotTakenList <> nil) and (FReasonNotTakenList.count > 0);
end;

Procedure TFhirMedicationStatement.SetReasonForUse(value : TFhirType);
begin
  FReasonForUse.free;
  FReasonForUse := value;
end;

Procedure TFhirMedicationStatement.SetEffective(value : TFhirType);
begin
  FEffective.free;
  FEffective := value;
end;

Procedure TFhirMedicationStatement.SetNote(value : TFhirString);
begin
  FNote.free;
  FNote := value;
end;

Function TFhirMedicationStatement.GetNoteST : String;
begin
  if FNote = nil then
    result := ''
  else
    result := FNote.value;
end;

Procedure TFhirMedicationStatement.SetNoteST(value : String);
begin
  if value <> '' then
  begin
    if FNote = nil then
      FNote := TFhirString.create;
    FNote.value := value
  end
  else if FNote <> nil then
    FNote.value := '';
end;

Function TFhirMedicationStatement.GetSupportingInformationList : TFhirReferenceList{TFhirReference};
begin
  if FSupportingInformationList = nil then
    FSupportingInformationList := TFhirReferenceList{TFhirReference}.Create;
  result := FSupportingInformationList;
end;

Function TFhirMedicationStatement.GetHasSupportingInformationList : boolean;
begin
  result := (FSupportingInformationList <> nil) and (FSupportingInformationList.count > 0);
end;

Procedure TFhirMedicationStatement.SetMedication(value : TFhirType);
begin
  FMedication.free;
  FMedication := value;
end;

Function TFhirMedicationStatement.GetDosageList : TFhirMedicationStatementDosageList;
begin
  if FDosageList = nil then
    FDosageList := TFhirMedicationStatementDosageList.Create;
  result := FDosageList;
end;

Function TFhirMedicationStatement.GetHasDosageList : boolean;
begin
  result := (FDosageList <> nil) and (FDosageList.count > 0);
end;


{ TFhirMedicationStatementListEnumerator }

Constructor TFhirMedicationStatementListEnumerator.Create(list : TFhirMedicationStatementList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMedicationStatementListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMedicationStatementListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMedicationStatementListEnumerator.GetCurrent : TFhirMedicationStatement;
begin
  Result := FList[FIndex];
end;


{ TFhirMedicationStatementList }
procedure TFhirMedicationStatementList.AddItem(value: TFhirMedicationStatement);
begin
  assert(value.ClassName = 'TFhirMedicationStatement', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMedicationStatement');
  add(value);
end;


function TFhirMedicationStatementList.Append: TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationStatementList.ClearItems;
begin
  Clear;
end;

function TFhirMedicationStatementList.GetEnumerator : TFhirMedicationStatementListEnumerator;
begin
  result := TFhirMedicationStatementListEnumerator.Create(self.link);
end;

function TFhirMedicationStatementList.Clone: TFhirMedicationStatementList;
begin
  result := TFhirMedicationStatementList(inherited Clone);
end;

function TFhirMedicationStatementList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMedicationStatementList.GetItemN(index: Integer): TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement(ObjectByIndex[index]);
end;

function TFhirMedicationStatementList.IndexOf(value: TFhirMedicationStatement): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMedicationStatementList.Insert(index: Integer): TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMedicationStatementList.InsertItem(index: Integer; value: TFhirMedicationStatement);
begin
  assert(value is TFhirMedicationStatement);
  Inherited Insert(index, value);
end;

function TFhirMedicationStatementList.Item(index: Integer): TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement(ObjectByIndex[index]);
end;

function TFhirMedicationStatementList.Link: TFhirMedicationStatementList;
begin
  result := TFhirMedicationStatementList(inherited Link);
end;

procedure TFhirMedicationStatementList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMedicationStatementList.SetItemByIndex(index: Integer; value: TFhirMedicationStatement);
begin
  assert(value is TFhirMedicationStatement);
  FhirMedicationStatements[index] := value;
end;

procedure TFhirMedicationStatementList.SetItemN(index: Integer; value: TFhirMedicationStatement);
begin
  assert(value is TFhirMedicationStatement);
  ObjectByIndex[index] := value;
end;

{ TFhirMessageHeaderResponse }

constructor TFhirMessageHeaderResponse.Create;
begin
  inherited;
end;

destructor TFhirMessageHeaderResponse.Destroy;
begin
  FIdentifier.free;
  FCode.free;
  FDetails.free;
  inherited;
end;

procedure TFhirMessageHeaderResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  identifierElement := TFhirMessageHeaderResponse(oSource).identifierElement.Clone;
  FCode := TFhirMessageHeaderResponse(oSource).FCode.Link;
  details := TFhirMessageHeaderResponse(oSource).details.Clone;
end;

procedure TFhirMessageHeaderResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'details') Then
     list.add(FDetails.Link);
end;

procedure TFhirMessageHeaderResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'id', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'details', 'Reference(OperationOutcome)', FDetails.Link));{2}
end;

procedure TFhirMessageHeaderResponse.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierElement := propValue as TFhirId{5a}
  else if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'details') then Details := propValue as TFhirReference{TFhirOperationOutcome}{4b}
  else inherited;
end;

function TFhirMessageHeaderResponse.FhirType : string;
begin
  result := 'response';
end;

function TFhirMessageHeaderResponse.Link : TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse(inherited Link);
end;

function TFhirMessageHeaderResponse.Clone : TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse(inherited Clone);
end;

function TFhirMessageHeaderResponse.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderResponse;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderResponse)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderResponse(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(detailsElement, o.detailsElement, true);
  end;
end;

function TFhirMessageHeaderResponse.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderResponse;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderResponse)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderResponse(other);
    result := compareValues(identifierElement, o.identifierElement, true) and compareValues(codeElement, o.codeElement, true);
  end;
end;

{ TFhirMessageHeaderResponse }

Procedure TFhirMessageHeaderResponse.SetIdentifier(value : TFhirId);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Function TFhirMessageHeaderResponse.GetIdentifierST : String;
begin
  if FIdentifier = nil then
    result := ''
  else
    result := FIdentifier.value;
end;

Procedure TFhirMessageHeaderResponse.SetIdentifierST(value : String);
begin
  if value <> '' then
  begin
    if FIdentifier = nil then
      FIdentifier := TFhirId.create;
    FIdentifier.value := value
  end
  else if FIdentifier <> nil then
    FIdentifier.value := '';
end;

Procedure TFhirMessageHeaderResponse.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirMessageHeaderResponse.GetCodeST : TFhirResponseCode;
begin
  if FCode = nil then
    result := TFhirResponseCode(0)
  else
    result := TFhirResponseCode(StringArrayIndexOfSensitive(CODES_TFhirResponseCode, FCode.value));
end;

Procedure TFhirMessageHeaderResponse.SetCodeST(value : TFhirResponseCode);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirResponseCode[value]);
end;

Procedure TFhirMessageHeaderResponse.SetDetails(value : TFhirReference{TFhirOperationOutcome});
begin
  FDetails.free;
  FDetails := value;
end;


{ TFhirMessageHeaderResponseListEnumerator }

Constructor TFhirMessageHeaderResponseListEnumerator.Create(list : TFhirMessageHeaderResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMessageHeaderResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMessageHeaderResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMessageHeaderResponseListEnumerator.GetCurrent : TFhirMessageHeaderResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirMessageHeaderResponseList }
procedure TFhirMessageHeaderResponseList.AddItem(value: TFhirMessageHeaderResponse);
begin
  assert(value.ClassName = 'TFhirMessageHeaderResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMessageHeaderResponse');
  add(value);
end;


function TFhirMessageHeaderResponseList.Append: TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderResponseList.ClearItems;
begin
  Clear;
end;

function TFhirMessageHeaderResponseList.GetEnumerator : TFhirMessageHeaderResponseListEnumerator;
begin
  result := TFhirMessageHeaderResponseListEnumerator.Create(self.link);
end;

function TFhirMessageHeaderResponseList.Clone: TFhirMessageHeaderResponseList;
begin
  result := TFhirMessageHeaderResponseList(inherited Clone);
end;

function TFhirMessageHeaderResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMessageHeaderResponseList.GetItemN(index: Integer): TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse(ObjectByIndex[index]);
end;

function TFhirMessageHeaderResponseList.IndexOf(value: TFhirMessageHeaderResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMessageHeaderResponseList.Insert(index: Integer): TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderResponseList.InsertItem(index: Integer; value: TFhirMessageHeaderResponse);
begin
  assert(value is TFhirMessageHeaderResponse);
  Inherited Insert(index, value);
end;

function TFhirMessageHeaderResponseList.Item(index: Integer): TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse(ObjectByIndex[index]);
end;

function TFhirMessageHeaderResponseList.Link: TFhirMessageHeaderResponseList;
begin
  result := TFhirMessageHeaderResponseList(inherited Link);
end;

procedure TFhirMessageHeaderResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMessageHeaderResponseList.SetItemByIndex(index: Integer; value: TFhirMessageHeaderResponse);
begin
  assert(value is TFhirMessageHeaderResponse);
  FhirMessageHeaderResponses[index] := value;
end;

procedure TFhirMessageHeaderResponseList.SetItemN(index: Integer; value: TFhirMessageHeaderResponse);
begin
  assert(value is TFhirMessageHeaderResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirMessageHeaderSource }

constructor TFhirMessageHeaderSource.Create;
begin
  inherited;
end;

destructor TFhirMessageHeaderSource.Destroy;
begin
  FName.free;
  FSoftware.free;
  FVersion.free;
  FContact.free;
  FEndpoint.free;
  inherited;
end;

procedure TFhirMessageHeaderSource.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirMessageHeaderSource(oSource).nameElement.Clone;
  softwareElement := TFhirMessageHeaderSource(oSource).softwareElement.Clone;
  versionElement := TFhirMessageHeaderSource(oSource).versionElement.Clone;
  contact := TFhirMessageHeaderSource(oSource).contact.Clone;
  endpointElement := TFhirMessageHeaderSource(oSource).endpointElement.Clone;
end;

procedure TFhirMessageHeaderSource.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'software') Then
     list.add(FSoftware.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'contact') Then
     list.add(FContact.Link);
  if (child_name = 'endpoint') Then
     list.add(FEndpoint.Link);
end;

procedure TFhirMessageHeaderSource.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'software', 'string', FSoftware.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', 'ContactPoint', FContact.Link));{2}
  oList.add(TFHIRProperty.create(self, 'endpoint', 'uri', FEndpoint.Link));{2}
end;

procedure TFhirMessageHeaderSource.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'software') then SoftwareElement := propValue as TFhirString{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then Contact := propValue as TFhirContactPoint{4b}
  else if (propName = 'endpoint') then EndpointElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirMessageHeaderSource.FhirType : string;
begin
  result := 'source';
end;

function TFhirMessageHeaderSource.Link : TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource(inherited Link);
end;

function TFhirMessageHeaderSource.Clone : TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource(inherited Clone);
end;

function TFhirMessageHeaderSource.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderSource;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderSource)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderSource(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(softwareElement, o.softwareElement, true) and 
      compareDeep(versionElement, o.versionElement, true) and compareDeep(contactElement, o.contactElement, true) and 
      compareDeep(endpointElement, o.endpointElement, true);
  end;
end;

function TFhirMessageHeaderSource.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderSource;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderSource)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderSource(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(softwareElement, o.softwareElement, true) and 
      compareValues(versionElement, o.versionElement, true) and compareValues(endpointElement, o.endpointElement, true);
  end;
end;

{ TFhirMessageHeaderSource }

Procedure TFhirMessageHeaderSource.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirMessageHeaderSource.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirMessageHeaderSource.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirMessageHeaderSource.SetSoftware(value : TFhirString);
begin
  FSoftware.free;
  FSoftware := value;
end;

Function TFhirMessageHeaderSource.GetSoftwareST : String;
begin
  if FSoftware = nil then
    result := ''
  else
    result := FSoftware.value;
end;

Procedure TFhirMessageHeaderSource.SetSoftwareST(value : String);
begin
  if value <> '' then
  begin
    if FSoftware = nil then
      FSoftware := TFhirString.create;
    FSoftware.value := value
  end
  else if FSoftware <> nil then
    FSoftware.value := '';
end;

Procedure TFhirMessageHeaderSource.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirMessageHeaderSource.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirMessageHeaderSource.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirMessageHeaderSource.SetContact(value : TFhirContactPoint);
begin
  FContact.free;
  FContact := value;
end;

Procedure TFhirMessageHeaderSource.SetEndpoint(value : TFhirUri);
begin
  FEndpoint.free;
  FEndpoint := value;
end;

Function TFhirMessageHeaderSource.GetEndpointST : String;
begin
  if FEndpoint = nil then
    result := ''
  else
    result := FEndpoint.value;
end;

Procedure TFhirMessageHeaderSource.SetEndpointST(value : String);
begin
  if value <> '' then
  begin
    if FEndpoint = nil then
      FEndpoint := TFhirUri.create;
    FEndpoint.value := value
  end
  else if FEndpoint <> nil then
    FEndpoint.value := '';
end;


{ TFhirMessageHeaderSourceListEnumerator }

Constructor TFhirMessageHeaderSourceListEnumerator.Create(list : TFhirMessageHeaderSourceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMessageHeaderSourceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMessageHeaderSourceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMessageHeaderSourceListEnumerator.GetCurrent : TFhirMessageHeaderSource;
begin
  Result := FList[FIndex];
end;


{ TFhirMessageHeaderSourceList }
procedure TFhirMessageHeaderSourceList.AddItem(value: TFhirMessageHeaderSource);
begin
  assert(value.ClassName = 'TFhirMessageHeaderSource', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMessageHeaderSource');
  add(value);
end;


function TFhirMessageHeaderSourceList.Append: TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderSourceList.ClearItems;
begin
  Clear;
end;

function TFhirMessageHeaderSourceList.GetEnumerator : TFhirMessageHeaderSourceListEnumerator;
begin
  result := TFhirMessageHeaderSourceListEnumerator.Create(self.link);
end;

function TFhirMessageHeaderSourceList.Clone: TFhirMessageHeaderSourceList;
begin
  result := TFhirMessageHeaderSourceList(inherited Clone);
end;

function TFhirMessageHeaderSourceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMessageHeaderSourceList.GetItemN(index: Integer): TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource(ObjectByIndex[index]);
end;

function TFhirMessageHeaderSourceList.IndexOf(value: TFhirMessageHeaderSource): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMessageHeaderSourceList.Insert(index: Integer): TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderSourceList.InsertItem(index: Integer; value: TFhirMessageHeaderSource);
begin
  assert(value is TFhirMessageHeaderSource);
  Inherited Insert(index, value);
end;

function TFhirMessageHeaderSourceList.Item(index: Integer): TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource(ObjectByIndex[index]);
end;

function TFhirMessageHeaderSourceList.Link: TFhirMessageHeaderSourceList;
begin
  result := TFhirMessageHeaderSourceList(inherited Link);
end;

procedure TFhirMessageHeaderSourceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMessageHeaderSourceList.SetItemByIndex(index: Integer; value: TFhirMessageHeaderSource);
begin
  assert(value is TFhirMessageHeaderSource);
  FhirMessageHeaderSources[index] := value;
end;

procedure TFhirMessageHeaderSourceList.SetItemN(index: Integer; value: TFhirMessageHeaderSource);
begin
  assert(value is TFhirMessageHeaderSource);
  ObjectByIndex[index] := value;
end;

{ TFhirMessageHeaderDestination }

constructor TFhirMessageHeaderDestination.Create;
begin
  inherited;
end;

destructor TFhirMessageHeaderDestination.Destroy;
begin
  FName.free;
  FTarget.free;
  FEndpoint.free;
  inherited;
end;

procedure TFhirMessageHeaderDestination.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirMessageHeaderDestination(oSource).nameElement.Clone;
  target := TFhirMessageHeaderDestination(oSource).target.Clone;
  endpointElement := TFhirMessageHeaderDestination(oSource).endpointElement.Clone;
end;

procedure TFhirMessageHeaderDestination.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'endpoint') Then
     list.add(FEndpoint.Link);
end;

procedure TFhirMessageHeaderDestination.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Device)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'endpoint', 'uri', FEndpoint.Link));{2}
end;

procedure TFhirMessageHeaderDestination.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirDevice}{4b}
  else if (propName = 'endpoint') then EndpointElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirMessageHeaderDestination.FhirType : string;
begin
  result := 'destination';
end;

function TFhirMessageHeaderDestination.Link : TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination(inherited Link);
end;

function TFhirMessageHeaderDestination.Clone : TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination(inherited Clone);
end;

function TFhirMessageHeaderDestination.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderDestination;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderDestination)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderDestination(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(targetElement, o.targetElement, true) and 
      compareDeep(endpointElement, o.endpointElement, true);
  end;
end;

function TFhirMessageHeaderDestination.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirMessageHeaderDestination;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirMessageHeaderDestination)) then
    result := false
  else
  begin
    o := TFhirMessageHeaderDestination(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(endpointElement, o.endpointElement, true);
  end;
end;

{ TFhirMessageHeaderDestination }

Procedure TFhirMessageHeaderDestination.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirMessageHeaderDestination.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirMessageHeaderDestination.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirMessageHeaderDestination.SetTarget(value : TFhirReference{TFhirDevice});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirMessageHeaderDestination.SetEndpoint(value : TFhirUri);
begin
  FEndpoint.free;
  FEndpoint := value;
end;

Function TFhirMessageHeaderDestination.GetEndpointST : String;
begin
  if FEndpoint = nil then
    result := ''
  else
    result := FEndpoint.value;
end;

Procedure TFhirMessageHeaderDestination.SetEndpointST(value : String);
begin
  if value <> '' then
  begin
    if FEndpoint = nil then
      FEndpoint := TFhirUri.create;
    FEndpoint.value := value
  end
  else if FEndpoint <> nil then
    FEndpoint.value := '';
end;


{ TFhirMessageHeaderDestinationListEnumerator }

Constructor TFhirMessageHeaderDestinationListEnumerator.Create(list : TFhirMessageHeaderDestinationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMessageHeaderDestinationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMessageHeaderDestinationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMessageHeaderDestinationListEnumerator.GetCurrent : TFhirMessageHeaderDestination;
begin
  Result := FList[FIndex];
end;


{ TFhirMessageHeaderDestinationList }
procedure TFhirMessageHeaderDestinationList.AddItem(value: TFhirMessageHeaderDestination);
begin
  assert(value.ClassName = 'TFhirMessageHeaderDestination', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMessageHeaderDestination');
  add(value);
end;


function TFhirMessageHeaderDestinationList.Append: TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderDestinationList.ClearItems;
begin
  Clear;
end;

function TFhirMessageHeaderDestinationList.GetEnumerator : TFhirMessageHeaderDestinationListEnumerator;
begin
  result := TFhirMessageHeaderDestinationListEnumerator.Create(self.link);
end;

function TFhirMessageHeaderDestinationList.Clone: TFhirMessageHeaderDestinationList;
begin
  result := TFhirMessageHeaderDestinationList(inherited Clone);
end;

function TFhirMessageHeaderDestinationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMessageHeaderDestinationList.GetItemN(index: Integer): TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination(ObjectByIndex[index]);
end;

function TFhirMessageHeaderDestinationList.IndexOf(value: TFhirMessageHeaderDestination): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMessageHeaderDestinationList.Insert(index: Integer): TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderDestinationList.InsertItem(index: Integer; value: TFhirMessageHeaderDestination);
begin
  assert(value is TFhirMessageHeaderDestination);
  Inherited Insert(index, value);
end;

function TFhirMessageHeaderDestinationList.Item(index: Integer): TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination(ObjectByIndex[index]);
end;

function TFhirMessageHeaderDestinationList.Link: TFhirMessageHeaderDestinationList;
begin
  result := TFhirMessageHeaderDestinationList(inherited Link);
end;

procedure TFhirMessageHeaderDestinationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMessageHeaderDestinationList.SetItemByIndex(index: Integer; value: TFhirMessageHeaderDestination);
begin
  assert(value is TFhirMessageHeaderDestination);
  FhirMessageHeaderDestinations[index] := value;
end;

procedure TFhirMessageHeaderDestinationList.SetItemN(index: Integer; value: TFhirMessageHeaderDestination);
begin
  assert(value is TFhirMessageHeaderDestination);
  ObjectByIndex[index] := value;
end;

{ TFhirMessageHeader }

constructor TFhirMessageHeader.Create;
begin
  inherited;
end;

destructor TFhirMessageHeader.Destroy;
begin
  FTimestamp.free;
  FEvent.free;
  FResponse.free;
  FSource.free;
  FDestinationList.Free;
  FEnterer.free;
  FAuthor.free;
  FReceiver.free;
  FResponsible.free;
  FReason.free;
  FDataList.Free;
  inherited;
end;

function TFhirMessageHeader.GetResourceType : TFhirResourceType;
begin
  result := frtMessageHeader;
end;

procedure TFhirMessageHeader.Assign(oSource : TAdvObject);
begin
  inherited;
  timestampElement := TFhirMessageHeader(oSource).timestampElement.Clone;
  event := TFhirMessageHeader(oSource).event.Clone;
  response := TFhirMessageHeader(oSource).response.Clone;
  source := TFhirMessageHeader(oSource).source.Clone;
  if (TFhirMessageHeader(oSource).FDestinationList = nil) then
  begin
    FDestinationList.free;
    FDestinationList := nil;
  end
  else
  begin
    FDestinationList := TFhirMessageHeaderDestinationList.Create;
    FDestinationList.Assign(TFhirMessageHeader(oSource).FDestinationList);
  end;
  enterer := TFhirMessageHeader(oSource).enterer.Clone;
  author := TFhirMessageHeader(oSource).author.Clone;
  receiver := TFhirMessageHeader(oSource).receiver.Clone;
  responsible := TFhirMessageHeader(oSource).responsible.Clone;
  reason := TFhirMessageHeader(oSource).reason.Clone;
  if (TFhirMessageHeader(oSource).FDataList = nil) then
  begin
    FDataList.free;
    FDataList := nil;
  end
  else
  begin
    FDataList := TFhirReferenceList{TFhirReference}.Create;
    FDataList.Assign(TFhirMessageHeader(oSource).FDataList);
  end;
end;

procedure TFhirMessageHeader.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'timestamp') Then
     list.add(FTimestamp.Link);
  if (child_name = 'event') Then
     list.add(FEvent.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'destination') Then
    list.addAll(FDestinationList);
  if (child_name = 'enterer') Then
     list.add(FEnterer.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'receiver') Then
     list.add(FReceiver.Link);
  if (child_name = 'responsible') Then
     list.add(FResponsible.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
  if (child_name = 'data') Then
    list.addAll(FDataList);
end;

procedure TFhirMessageHeader.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'timestamp', 'instant', FTimestamp.Link));{2}
  oList.add(TFHIRProperty.create(self, 'event', 'Coding', FEvent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', '', FResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', '', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'destination', '', FDestinationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'enterer', 'Reference(Practitioner)', FEnterer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Practitioner)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'receiver', 'Reference(Practitioner|Organization)', FReceiver.Link));{2}
  oList.add(TFHIRProperty.create(self, 'responsible', 'Reference(Practitioner|Organization)', FResponsible.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'data', 'Reference(Any)', FDataList.Link)){3};
end;

procedure TFhirMessageHeader.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'timestamp') then TimestampElement := propValue as TFhirInstant{5a}
  else if (propName = 'event') then Event := propValue as TFhirCoding{4b}
  else if (propName = 'response') then Response := propValue as TFhirMessageHeaderResponse{4b}
  else if (propName = 'source') then Source := propValue as TFhirMessageHeaderSource{4b}
  else if (propName = 'destination') then DestinationList.add(propValue as TFhirMessageHeaderDestination){2}
  else if (propName = 'enterer') then Enterer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'author') then Author := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'receiver') then Receiver := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'responsible') then Responsible := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'reason') then Reason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'data') then DataList.add(propValue as TFhirReference{TFhirReference}){2}
  else inherited;
end;

function TFhirMessageHeader.FhirType : string;
begin
  result := 'MessageHeader';
end;

function TFhirMessageHeader.Link : TFhirMessageHeader;
begin
  result := TFhirMessageHeader(inherited Link);
end;

function TFhirMessageHeader.Clone : TFhirMessageHeader;
begin
  result := TFhirMessageHeader(inherited Clone);
end;

{ TFhirMessageHeader }

Procedure TFhirMessageHeader.SetTimestamp(value : TFhirInstant);
begin
  FTimestamp.free;
  FTimestamp := value;
end;

Function TFhirMessageHeader.GetTimestampST : TDateAndTime;
begin
  if FTimestamp = nil then
    result := nil
  else
    result := FTimestamp.value;
end;

Procedure TFhirMessageHeader.SetTimestampST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FTimestamp = nil then
      FTimestamp := TFhirInstant.create;
    FTimestamp.value := value
  end
  else if FTimestamp <> nil then
    FTimestamp.value := nil;
end;

Procedure TFhirMessageHeader.SetEvent(value : TFhirCoding);
begin
  FEvent.free;
  FEvent := value;
end;

Procedure TFhirMessageHeader.SetResponse(value : TFhirMessageHeaderResponse);
begin
  FResponse.free;
  FResponse := value;
end;

Procedure TFhirMessageHeader.SetSource(value : TFhirMessageHeaderSource);
begin
  FSource.free;
  FSource := value;
end;

Function TFhirMessageHeader.GetDestinationList : TFhirMessageHeaderDestinationList;
begin
  if FDestinationList = nil then
    FDestinationList := TFhirMessageHeaderDestinationList.Create;
  result := FDestinationList;
end;

Function TFhirMessageHeader.GetHasDestinationList : boolean;
begin
  result := (FDestinationList <> nil) and (FDestinationList.count > 0);
end;

Procedure TFhirMessageHeader.SetEnterer(value : TFhirReference{TFhirPractitioner});
begin
  FEnterer.free;
  FEnterer := value;
end;

Procedure TFhirMessageHeader.SetAuthor(value : TFhirReference{TFhirPractitioner});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirMessageHeader.SetReceiver(value : TFhirReference{Resource});
begin
  FReceiver.free;
  FReceiver := value;
end;

Procedure TFhirMessageHeader.SetResponsible(value : TFhirReference{Resource});
begin
  FResponsible.free;
  FResponsible := value;
end;

Procedure TFhirMessageHeader.SetReason(value : TFhirCodeableConcept);
begin
  FReason.free;
  FReason := value;
end;

Function TFhirMessageHeader.GetDataList : TFhirReferenceList{TFhirReference};
begin
  if FDataList = nil then
    FDataList := TFhirReferenceList{TFhirReference}.Create;
  result := FDataList;
end;

Function TFhirMessageHeader.GetHasDataList : boolean;
begin
  result := (FDataList <> nil) and (FDataList.count > 0);
end;


{ TFhirMessageHeaderListEnumerator }

Constructor TFhirMessageHeaderListEnumerator.Create(list : TFhirMessageHeaderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirMessageHeaderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirMessageHeaderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirMessageHeaderListEnumerator.GetCurrent : TFhirMessageHeader;
begin
  Result := FList[FIndex];
end;


{ TFhirMessageHeaderList }
procedure TFhirMessageHeaderList.AddItem(value: TFhirMessageHeader);
begin
  assert(value.ClassName = 'TFhirMessageHeader', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirMessageHeader');
  add(value);
end;


function TFhirMessageHeaderList.Append: TFhirMessageHeader;
begin
  result := TFhirMessageHeader.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderList.ClearItems;
begin
  Clear;
end;

function TFhirMessageHeaderList.GetEnumerator : TFhirMessageHeaderListEnumerator;
begin
  result := TFhirMessageHeaderListEnumerator.Create(self.link);
end;

function TFhirMessageHeaderList.Clone: TFhirMessageHeaderList;
begin
  result := TFhirMessageHeaderList(inherited Clone);
end;

function TFhirMessageHeaderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirMessageHeaderList.GetItemN(index: Integer): TFhirMessageHeader;
begin
  result := TFhirMessageHeader(ObjectByIndex[index]);
end;

function TFhirMessageHeaderList.IndexOf(value: TFhirMessageHeader): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirMessageHeaderList.Insert(index: Integer): TFhirMessageHeader;
begin
  result := TFhirMessageHeader.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirMessageHeaderList.InsertItem(index: Integer; value: TFhirMessageHeader);
begin
  assert(value is TFhirMessageHeader);
  Inherited Insert(index, value);
end;

function TFhirMessageHeaderList.Item(index: Integer): TFhirMessageHeader;
begin
  result := TFhirMessageHeader(ObjectByIndex[index]);
end;

function TFhirMessageHeaderList.Link: TFhirMessageHeaderList;
begin
  result := TFhirMessageHeaderList(inherited Link);
end;

procedure TFhirMessageHeaderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirMessageHeaderList.SetItemByIndex(index: Integer; value: TFhirMessageHeader);
begin
  assert(value is TFhirMessageHeader);
  FhirMessageHeaders[index] := value;
end;

procedure TFhirMessageHeaderList.SetItemN(index: Integer; value: TFhirMessageHeader);
begin
  assert(value is TFhirMessageHeader);
  ObjectByIndex[index] := value;
end;

{ TFhirNamingSystemContact }

constructor TFhirNamingSystemContact.Create;
begin
  inherited;
end;

destructor TFhirNamingSystemContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirNamingSystemContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirNamingSystemContact(oSource).nameElement.Clone;
  if (TFhirNamingSystemContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirNamingSystemContact(oSource).FTelecomList);
  end;
end;

procedure TFhirNamingSystemContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirNamingSystemContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirNamingSystemContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirNamingSystemContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirNamingSystemContact.Link : TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact(inherited Link);
end;

function TFhirNamingSystemContact.Clone : TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact(inherited Clone);
end;

function TFhirNamingSystemContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNamingSystemContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNamingSystemContact)) then
    result := false
  else
  begin
    o := TFhirNamingSystemContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirNamingSystemContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNamingSystemContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNamingSystemContact)) then
    result := false
  else
  begin
    o := TFhirNamingSystemContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirNamingSystemContact }

Procedure TFhirNamingSystemContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirNamingSystemContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirNamingSystemContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirNamingSystemContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirNamingSystemContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirNamingSystemContactListEnumerator }

Constructor TFhirNamingSystemContactListEnumerator.Create(list : TFhirNamingSystemContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNamingSystemContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNamingSystemContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNamingSystemContactListEnumerator.GetCurrent : TFhirNamingSystemContact;
begin
  Result := FList[FIndex];
end;


{ TFhirNamingSystemContactList }
procedure TFhirNamingSystemContactList.AddItem(value: TFhirNamingSystemContact);
begin
  assert(value.ClassName = 'TFhirNamingSystemContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNamingSystemContact');
  add(value);
end;


function TFhirNamingSystemContactList.Append: TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemContactList.ClearItems;
begin
  Clear;
end;

function TFhirNamingSystemContactList.GetEnumerator : TFhirNamingSystemContactListEnumerator;
begin
  result := TFhirNamingSystemContactListEnumerator.Create(self.link);
end;

function TFhirNamingSystemContactList.Clone: TFhirNamingSystemContactList;
begin
  result := TFhirNamingSystemContactList(inherited Clone);
end;

function TFhirNamingSystemContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNamingSystemContactList.GetItemN(index: Integer): TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact(ObjectByIndex[index]);
end;

function TFhirNamingSystemContactList.IndexOf(value: TFhirNamingSystemContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNamingSystemContactList.Insert(index: Integer): TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemContactList.InsertItem(index: Integer; value: TFhirNamingSystemContact);
begin
  assert(value is TFhirNamingSystemContact);
  Inherited Insert(index, value);
end;

function TFhirNamingSystemContactList.Item(index: Integer): TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact(ObjectByIndex[index]);
end;

function TFhirNamingSystemContactList.Link: TFhirNamingSystemContactList;
begin
  result := TFhirNamingSystemContactList(inherited Link);
end;

procedure TFhirNamingSystemContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNamingSystemContactList.SetItemByIndex(index: Integer; value: TFhirNamingSystemContact);
begin
  assert(value is TFhirNamingSystemContact);
  FhirNamingSystemContacts[index] := value;
end;

procedure TFhirNamingSystemContactList.SetItemN(index: Integer; value: TFhirNamingSystemContact);
begin
  assert(value is TFhirNamingSystemContact);
  ObjectByIndex[index] := value;
end;

{ TFhirNamingSystemUniqueId }

constructor TFhirNamingSystemUniqueId.Create;
begin
  inherited;
end;

destructor TFhirNamingSystemUniqueId.Destroy;
begin
  FType_.free;
  FValue.free;
  FPreferred.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirNamingSystemUniqueId.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirNamingSystemUniqueId(oSource).FType_.Link;
  valueElement := TFhirNamingSystemUniqueId(oSource).valueElement.Clone;
  preferredElement := TFhirNamingSystemUniqueId(oSource).preferredElement.Clone;
  period := TFhirNamingSystemUniqueId(oSource).period.Clone;
end;

procedure TFhirNamingSystemUniqueId.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
  if (child_name = 'preferred') Then
     list.add(FPreferred.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirNamingSystemUniqueId.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'value', 'string', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'preferred', 'boolean', FPreferred.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirNamingSystemUniqueId.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'value') then ValueElement := propValue as TFhirString{5a}
  else if (propName = 'preferred') then PreferredElement := propValue as TFhirBoolean{5a}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirNamingSystemUniqueId.FhirType : string;
begin
  result := 'uniqueId';
end;

function TFhirNamingSystemUniqueId.Link : TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId(inherited Link);
end;

function TFhirNamingSystemUniqueId.Clone : TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId(inherited Clone);
end;

function TFhirNamingSystemUniqueId.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNamingSystemUniqueId;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNamingSystemUniqueId)) then
    result := false
  else
  begin
    o := TFhirNamingSystemUniqueId(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(valueElement, o.valueElement, true) and 
      compareDeep(preferredElement, o.preferredElement, true) and compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirNamingSystemUniqueId.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNamingSystemUniqueId;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNamingSystemUniqueId)) then
    result := false
  else
  begin
    o := TFhirNamingSystemUniqueId(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(valueElement, o.valueElement, true) and 
      compareValues(preferredElement, o.preferredElement, true);
  end;
end;

{ TFhirNamingSystemUniqueId }

Procedure TFhirNamingSystemUniqueId.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirNamingSystemUniqueId.GetType_ST : TFhirNamingsystemIdentifierType;
begin
  if FType_ = nil then
    result := TFhirNamingsystemIdentifierType(0)
  else
    result := TFhirNamingsystemIdentifierType(StringArrayIndexOfSensitive(CODES_TFhirNamingsystemIdentifierType, FType_.value));
end;

Procedure TFhirNamingSystemUniqueId.SetType_ST(value : TFhirNamingsystemIdentifierType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirNamingsystemIdentifierType[value]);
end;

Procedure TFhirNamingSystemUniqueId.SetValue(value : TFhirString);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirNamingSystemUniqueId.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirNamingSystemUniqueId.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirString.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;

Procedure TFhirNamingSystemUniqueId.SetPreferred(value : TFhirBoolean);
begin
  FPreferred.free;
  FPreferred := value;
end;

Function TFhirNamingSystemUniqueId.GetPreferredST : Boolean;
begin
  if FPreferred = nil then
    result := false
  else
    result := FPreferred.value;
end;

Procedure TFhirNamingSystemUniqueId.SetPreferredST(value : Boolean);
begin
  if FPreferred = nil then
    FPreferred := TFhirBoolean.create;
  FPreferred.value := value
end;

Procedure TFhirNamingSystemUniqueId.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirNamingSystemUniqueIdListEnumerator }

Constructor TFhirNamingSystemUniqueIdListEnumerator.Create(list : TFhirNamingSystemUniqueIdList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNamingSystemUniqueIdListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNamingSystemUniqueIdListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNamingSystemUniqueIdListEnumerator.GetCurrent : TFhirNamingSystemUniqueId;
begin
  Result := FList[FIndex];
end;


{ TFhirNamingSystemUniqueIdList }
procedure TFhirNamingSystemUniqueIdList.AddItem(value: TFhirNamingSystemUniqueId);
begin
  assert(value.ClassName = 'TFhirNamingSystemUniqueId', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNamingSystemUniqueId');
  add(value);
end;


function TFhirNamingSystemUniqueIdList.Append: TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemUniqueIdList.ClearItems;
begin
  Clear;
end;

function TFhirNamingSystemUniqueIdList.GetEnumerator : TFhirNamingSystemUniqueIdListEnumerator;
begin
  result := TFhirNamingSystemUniqueIdListEnumerator.Create(self.link);
end;

function TFhirNamingSystemUniqueIdList.Clone: TFhirNamingSystemUniqueIdList;
begin
  result := TFhirNamingSystemUniqueIdList(inherited Clone);
end;

function TFhirNamingSystemUniqueIdList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNamingSystemUniqueIdList.GetItemN(index: Integer): TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId(ObjectByIndex[index]);
end;

function TFhirNamingSystemUniqueIdList.IndexOf(value: TFhirNamingSystemUniqueId): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNamingSystemUniqueIdList.Insert(index: Integer): TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemUniqueIdList.InsertItem(index: Integer; value: TFhirNamingSystemUniqueId);
begin
  assert(value is TFhirNamingSystemUniqueId);
  Inherited Insert(index, value);
end;

function TFhirNamingSystemUniqueIdList.Item(index: Integer): TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId(ObjectByIndex[index]);
end;

function TFhirNamingSystemUniqueIdList.Link: TFhirNamingSystemUniqueIdList;
begin
  result := TFhirNamingSystemUniqueIdList(inherited Link);
end;

procedure TFhirNamingSystemUniqueIdList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNamingSystemUniqueIdList.SetItemByIndex(index: Integer; value: TFhirNamingSystemUniqueId);
begin
  assert(value is TFhirNamingSystemUniqueId);
  FhirNamingSystemUniqueIds[index] := value;
end;

procedure TFhirNamingSystemUniqueIdList.SetItemN(index: Integer; value: TFhirNamingSystemUniqueId);
begin
  assert(value is TFhirNamingSystemUniqueId);
  ObjectByIndex[index] := value;
end;

{ TFhirNamingSystem }

constructor TFhirNamingSystem.Create;
begin
  inherited;
end;

destructor TFhirNamingSystem.Destroy;
begin
  FName.free;
  FStatus.free;
  FKind.free;
  FPublisher.free;
  FContactList.Free;
  FResponsible.free;
  FDate.free;
  FType_.free;
  FDescription.free;
  FUseContextList.Free;
  FUsage.free;
  FUniqueIdList.Free;
  FReplacedBy.free;
  inherited;
end;

function TFhirNamingSystem.GetResourceType : TFhirResourceType;
begin
  result := frtNamingSystem;
end;

procedure TFhirNamingSystem.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirNamingSystem(oSource).nameElement.Clone;
  FStatus := TFhirNamingSystem(oSource).FStatus.Link;
  FKind := TFhirNamingSystem(oSource).FKind.Link;
  publisherElement := TFhirNamingSystem(oSource).publisherElement.Clone;
  if (TFhirNamingSystem(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirNamingSystemContactList.Create;
    FContactList.Assign(TFhirNamingSystem(oSource).FContactList);
  end;
  responsibleElement := TFhirNamingSystem(oSource).responsibleElement.Clone;
  dateElement := TFhirNamingSystem(oSource).dateElement.Clone;
  type_ := TFhirNamingSystem(oSource).type_.Clone;
  descriptionElement := TFhirNamingSystem(oSource).descriptionElement.Clone;
  if (TFhirNamingSystem(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirNamingSystem(oSource).FUseContextList);
  end;
  usageElement := TFhirNamingSystem(oSource).usageElement.Clone;
  if (TFhirNamingSystem(oSource).FUniqueIdList = nil) then
  begin
    FUniqueIdList.free;
    FUniqueIdList := nil;
  end
  else
  begin
    FUniqueIdList := TFhirNamingSystemUniqueIdList.Create;
    FUniqueIdList.Assign(TFhirNamingSystem(oSource).FUniqueIdList);
  end;
  replacedBy := TFhirNamingSystem(oSource).replacedBy.Clone;
end;

procedure TFhirNamingSystem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'responsible') Then
     list.add(FResponsible.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'usage') Then
     list.add(FUsage.Link);
  if (child_name = 'uniqueId') Then
    list.addAll(FUniqueIdList);
  if (child_name = 'replacedBy') Then
     list.add(FReplacedBy.Link);
end;

procedure TFhirNamingSystem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'kind', 'code', FKind.Link));{1}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'responsible', 'string', FResponsible.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'usage', 'string', FUsage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uniqueId', '', FUniqueIdList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'replacedBy', 'Reference(NamingSystem)', FReplacedBy.Link));{2}
end;

procedure TFhirNamingSystem.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'kind') then KindElement := propValue as TFHIREnum
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirNamingSystemContact){2}
  else if (propName = 'responsible') then ResponsibleElement := propValue as TFhirString{5a}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'usage') then UsageElement := propValue as TFhirString{5a}
  else if (propName = 'uniqueId') then UniqueIdList.add(propValue as TFhirNamingSystemUniqueId){2}
  else if (propName = 'replacedBy') then ReplacedBy := propValue as TFhirReference{TFhirNamingSystem}{4b}
  else inherited;
end;

function TFhirNamingSystem.FhirType : string;
begin
  result := 'NamingSystem';
end;

function TFhirNamingSystem.Link : TFhirNamingSystem;
begin
  result := TFhirNamingSystem(inherited Link);
end;

function TFhirNamingSystem.Clone : TFhirNamingSystem;
begin
  result := TFhirNamingSystem(inherited Clone);
end;

{ TFhirNamingSystem }

Procedure TFhirNamingSystem.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirNamingSystem.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirNamingSystem.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirNamingSystem.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirNamingSystem.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirNamingSystem.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirNamingSystem.SetKind(value : TFhirEnum);
begin
  FKind.free;
  FKind := value;
end;

Function TFhirNamingSystem.GetKindST : TFhirNamingsystemType;
begin
  if FKind = nil then
    result := TFhirNamingsystemType(0)
  else
    result := TFhirNamingsystemType(StringArrayIndexOfSensitive(CODES_TFhirNamingsystemType, FKind.value));
end;

Procedure TFhirNamingSystem.SetKindST(value : TFhirNamingsystemType);
begin
  if ord(value) = 0 then
    KindElement := nil
  else
    KindElement := TFhirEnum.create(CODES_TFhirNamingsystemType[value]);
end;

Procedure TFhirNamingSystem.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirNamingSystem.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirNamingSystem.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirNamingSystem.GetContactList : TFhirNamingSystemContactList;
begin
  if FContactList = nil then
    FContactList := TFhirNamingSystemContactList.Create;
  result := FContactList;
end;

Function TFhirNamingSystem.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirNamingSystem.SetResponsible(value : TFhirString);
begin
  FResponsible.free;
  FResponsible := value;
end;

Function TFhirNamingSystem.GetResponsibleST : String;
begin
  if FResponsible = nil then
    result := ''
  else
    result := FResponsible.value;
end;

Procedure TFhirNamingSystem.SetResponsibleST(value : String);
begin
  if value <> '' then
  begin
    if FResponsible = nil then
      FResponsible := TFhirString.create;
    FResponsible.value := value
  end
  else if FResponsible <> nil then
    FResponsible.value := '';
end;

Procedure TFhirNamingSystem.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirNamingSystem.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirNamingSystem.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirNamingSystem.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirNamingSystem.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirNamingSystem.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirNamingSystem.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirNamingSystem.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirNamingSystem.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirNamingSystem.SetUsage(value : TFhirString);
begin
  FUsage.free;
  FUsage := value;
end;

Function TFhirNamingSystem.GetUsageST : String;
begin
  if FUsage = nil then
    result := ''
  else
    result := FUsage.value;
end;

Procedure TFhirNamingSystem.SetUsageST(value : String);
begin
  if value <> '' then
  begin
    if FUsage = nil then
      FUsage := TFhirString.create;
    FUsage.value := value
  end
  else if FUsage <> nil then
    FUsage.value := '';
end;

Function TFhirNamingSystem.GetUniqueIdList : TFhirNamingSystemUniqueIdList;
begin
  if FUniqueIdList = nil then
    FUniqueIdList := TFhirNamingSystemUniqueIdList.Create;
  result := FUniqueIdList;
end;

Function TFhirNamingSystem.GetHasUniqueIdList : boolean;
begin
  result := (FUniqueIdList <> nil) and (FUniqueIdList.count > 0);
end;

Procedure TFhirNamingSystem.SetReplacedBy(value : TFhirReference{TFhirNamingSystem});
begin
  FReplacedBy.free;
  FReplacedBy := value;
end;


{ TFhirNamingSystemListEnumerator }

Constructor TFhirNamingSystemListEnumerator.Create(list : TFhirNamingSystemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNamingSystemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNamingSystemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNamingSystemListEnumerator.GetCurrent : TFhirNamingSystem;
begin
  Result := FList[FIndex];
end;


{ TFhirNamingSystemList }
procedure TFhirNamingSystemList.AddItem(value: TFhirNamingSystem);
begin
  assert(value.ClassName = 'TFhirNamingSystem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNamingSystem');
  add(value);
end;


function TFhirNamingSystemList.Append: TFhirNamingSystem;
begin
  result := TFhirNamingSystem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemList.ClearItems;
begin
  Clear;
end;

function TFhirNamingSystemList.GetEnumerator : TFhirNamingSystemListEnumerator;
begin
  result := TFhirNamingSystemListEnumerator.Create(self.link);
end;

function TFhirNamingSystemList.Clone: TFhirNamingSystemList;
begin
  result := TFhirNamingSystemList(inherited Clone);
end;

function TFhirNamingSystemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNamingSystemList.GetItemN(index: Integer): TFhirNamingSystem;
begin
  result := TFhirNamingSystem(ObjectByIndex[index]);
end;

function TFhirNamingSystemList.IndexOf(value: TFhirNamingSystem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNamingSystemList.Insert(index: Integer): TFhirNamingSystem;
begin
  result := TFhirNamingSystem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNamingSystemList.InsertItem(index: Integer; value: TFhirNamingSystem);
begin
  assert(value is TFhirNamingSystem);
  Inherited Insert(index, value);
end;

function TFhirNamingSystemList.Item(index: Integer): TFhirNamingSystem;
begin
  result := TFhirNamingSystem(ObjectByIndex[index]);
end;

function TFhirNamingSystemList.Link: TFhirNamingSystemList;
begin
  result := TFhirNamingSystemList(inherited Link);
end;

procedure TFhirNamingSystemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNamingSystemList.SetItemByIndex(index: Integer; value: TFhirNamingSystem);
begin
  assert(value is TFhirNamingSystem);
  FhirNamingSystems[index] := value;
end;

procedure TFhirNamingSystemList.SetItemN(index: Integer; value: TFhirNamingSystem);
begin
  assert(value is TFhirNamingSystem);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderOralDiet }

constructor TFhirNutritionOrderOralDiet.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderOralDiet.Destroy;
begin
  FType_List.Free;
  FScheduleList.Free;
  FNutrientList.Free;
  FTextureList.Free;
  FFluidConsistencyTypeList.Free;
  FInstruction.free;
  inherited;
end;

procedure TFhirNutritionOrderOralDiet.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirNutritionOrderOralDiet(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirNutritionOrderOralDiet(oSource).FType_List);
  end;
  if (TFhirNutritionOrderOralDiet(oSource).FScheduleList = nil) then
  begin
    FScheduleList.free;
    FScheduleList := nil;
  end
  else
  begin
    FScheduleList := TFhirTimingList.Create;
    FScheduleList.Assign(TFhirNutritionOrderOralDiet(oSource).FScheduleList);
  end;
  if (TFhirNutritionOrderOralDiet(oSource).FNutrientList = nil) then
  begin
    FNutrientList.free;
    FNutrientList := nil;
  end
  else
  begin
    FNutrientList := TFhirNutritionOrderOralDietNutrientList.Create;
    FNutrientList.Assign(TFhirNutritionOrderOralDiet(oSource).FNutrientList);
  end;
  if (TFhirNutritionOrderOralDiet(oSource).FTextureList = nil) then
  begin
    FTextureList.free;
    FTextureList := nil;
  end
  else
  begin
    FTextureList := TFhirNutritionOrderOralDietTextureList.Create;
    FTextureList.Assign(TFhirNutritionOrderOralDiet(oSource).FTextureList);
  end;
  if (TFhirNutritionOrderOralDiet(oSource).FFluidConsistencyTypeList = nil) then
  begin
    FFluidConsistencyTypeList.free;
    FFluidConsistencyTypeList := nil;
  end
  else
  begin
    FFluidConsistencyTypeList := TFhirCodeableConceptList.Create;
    FFluidConsistencyTypeList.Assign(TFhirNutritionOrderOralDiet(oSource).FFluidConsistencyTypeList);
  end;
  instructionElement := TFhirNutritionOrderOralDiet(oSource).instructionElement.Clone;
end;

procedure TFhirNutritionOrderOralDiet.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'schedule') Then
    list.addAll(FScheduleList);
  if (child_name = 'nutrient') Then
    list.addAll(FNutrientList);
  if (child_name = 'texture') Then
    list.addAll(FTextureList);
  if (child_name = 'fluidConsistencyType') Then
    list.addAll(FFluidConsistencyTypeList);
  if (child_name = 'instruction') Then
     list.add(FInstruction.Link);
end;

procedure TFhirNutritionOrderOralDiet.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'schedule', 'Timing', FScheduleList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'nutrient', '', FNutrientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'texture', '', FTextureList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'fluidConsistencyType', 'CodeableConcept', FFluidConsistencyTypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'instruction', 'string', FInstruction.Link));{2}
end;

procedure TFhirNutritionOrderOralDiet.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'schedule') then ScheduleList.add(propValue as TFhirTiming){2}
  else if (propName = 'nutrient') then NutrientList.add(propValue as TFhirNutritionOrderOralDietNutrient){2}
  else if (propName = 'texture') then TextureList.add(propValue as TFhirNutritionOrderOralDietTexture){2}
  else if (propName = 'fluidConsistencyType') then FluidConsistencyTypeList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'instruction') then InstructionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirNutritionOrderOralDiet.FhirType : string;
begin
  result := 'oralDiet';
end;

function TFhirNutritionOrderOralDiet.Link : TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet(inherited Link);
end;

function TFhirNutritionOrderOralDiet.Clone : TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet(inherited Clone);
end;

function TFhirNutritionOrderOralDiet.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDiet;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDiet)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDiet(other);
    result := compareDeep(type_List, o.type_List, true) and compareDeep(scheduleList, o.scheduleList, true) and 
      compareDeep(nutrientList, o.nutrientList, true) and compareDeep(textureList, o.textureList, true) and 
      compareDeep(fluidConsistencyTypeList, o.fluidConsistencyTypeList, true) and compareDeep(instructionElement, o.instructionElement, true);
  end;
end;

function TFhirNutritionOrderOralDiet.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDiet;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDiet)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDiet(other);
    result := compareValues(instructionElement, o.instructionElement, true);
  end;
end;

{ TFhirNutritionOrderOralDiet }

Function TFhirNutritionOrderOralDiet.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirNutritionOrderOralDiet.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Function TFhirNutritionOrderOralDiet.GetScheduleList : TFhirTimingList;
begin
  if FScheduleList = nil then
    FScheduleList := TFhirTimingList.Create;
  result := FScheduleList;
end;

Function TFhirNutritionOrderOralDiet.GetHasScheduleList : boolean;
begin
  result := (FScheduleList <> nil) and (FScheduleList.count > 0);
end;

Function TFhirNutritionOrderOralDiet.GetNutrientList : TFhirNutritionOrderOralDietNutrientList;
begin
  if FNutrientList = nil then
    FNutrientList := TFhirNutritionOrderOralDietNutrientList.Create;
  result := FNutrientList;
end;

Function TFhirNutritionOrderOralDiet.GetHasNutrientList : boolean;
begin
  result := (FNutrientList <> nil) and (FNutrientList.count > 0);
end;

Function TFhirNutritionOrderOralDiet.GetTextureList : TFhirNutritionOrderOralDietTextureList;
begin
  if FTextureList = nil then
    FTextureList := TFhirNutritionOrderOralDietTextureList.Create;
  result := FTextureList;
end;

Function TFhirNutritionOrderOralDiet.GetHasTextureList : boolean;
begin
  result := (FTextureList <> nil) and (FTextureList.count > 0);
end;

Function TFhirNutritionOrderOralDiet.GetFluidConsistencyTypeList : TFhirCodeableConceptList;
begin
  if FFluidConsistencyTypeList = nil then
    FFluidConsistencyTypeList := TFhirCodeableConceptList.Create;
  result := FFluidConsistencyTypeList;
end;

Function TFhirNutritionOrderOralDiet.GetHasFluidConsistencyTypeList : boolean;
begin
  result := (FFluidConsistencyTypeList <> nil) and (FFluidConsistencyTypeList.count > 0);
end;

Procedure TFhirNutritionOrderOralDiet.SetInstruction(value : TFhirString);
begin
  FInstruction.free;
  FInstruction := value;
end;

Function TFhirNutritionOrderOralDiet.GetInstructionST : String;
begin
  if FInstruction = nil then
    result := ''
  else
    result := FInstruction.value;
end;

Procedure TFhirNutritionOrderOralDiet.SetInstructionST(value : String);
begin
  if value <> '' then
  begin
    if FInstruction = nil then
      FInstruction := TFhirString.create;
    FInstruction.value := value
  end
  else if FInstruction <> nil then
    FInstruction.value := '';
end;


{ TFhirNutritionOrderOralDietListEnumerator }

Constructor TFhirNutritionOrderOralDietListEnumerator.Create(list : TFhirNutritionOrderOralDietList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderOralDietListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderOralDietListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderOralDietListEnumerator.GetCurrent : TFhirNutritionOrderOralDiet;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderOralDietList }
procedure TFhirNutritionOrderOralDietList.AddItem(value: TFhirNutritionOrderOralDiet);
begin
  assert(value.ClassName = 'TFhirNutritionOrderOralDiet', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderOralDiet');
  add(value);
end;


function TFhirNutritionOrderOralDietList.Append: TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderOralDietList.GetEnumerator : TFhirNutritionOrderOralDietListEnumerator;
begin
  result := TFhirNutritionOrderOralDietListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderOralDietList.Clone: TFhirNutritionOrderOralDietList;
begin
  result := TFhirNutritionOrderOralDietList(inherited Clone);
end;

function TFhirNutritionOrderOralDietList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderOralDietList.GetItemN(index: Integer): TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietList.IndexOf(value: TFhirNutritionOrderOralDiet): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderOralDietList.Insert(index: Integer): TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietList.InsertItem(index: Integer; value: TFhirNutritionOrderOralDiet);
begin
  assert(value is TFhirNutritionOrderOralDiet);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderOralDietList.Item(index: Integer): TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietList.Link: TFhirNutritionOrderOralDietList;
begin
  result := TFhirNutritionOrderOralDietList(inherited Link);
end;

procedure TFhirNutritionOrderOralDietList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderOralDietList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderOralDiet);
begin
  assert(value is TFhirNutritionOrderOralDiet);
  FhirNutritionOrderOralDiets[index] := value;
end;

procedure TFhirNutritionOrderOralDietList.SetItemN(index: Integer; value: TFhirNutritionOrderOralDiet);
begin
  assert(value is TFhirNutritionOrderOralDiet);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderOralDietNutrient }

constructor TFhirNutritionOrderOralDietNutrient.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderOralDietNutrient.Destroy;
begin
  FModifier.free;
  FAmount.free;
  inherited;
end;

procedure TFhirNutritionOrderOralDietNutrient.Assign(oSource : TAdvObject);
begin
  inherited;
  modifier := TFhirNutritionOrderOralDietNutrient(oSource).modifier.Clone;
  amount := TFhirNutritionOrderOralDietNutrient(oSource).amount.Clone;
end;

procedure TFhirNutritionOrderOralDietNutrient.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'modifier') Then
     list.add(FModifier.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
end;

procedure TFhirNutritionOrderOralDietNutrient.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'modifier', 'CodeableConcept', FModifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'SimpleQuantity', FAmount.Link));{2}
end;

procedure TFhirNutritionOrderOralDietNutrient.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'modifier') then Modifier := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirNutritionOrderOralDietNutrient.FhirType : string;
begin
  result := 'nutrient';
end;

function TFhirNutritionOrderOralDietNutrient.Link : TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient(inherited Link);
end;

function TFhirNutritionOrderOralDietNutrient.Clone : TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient(inherited Clone);
end;

function TFhirNutritionOrderOralDietNutrient.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDietNutrient;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDietNutrient)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDietNutrient(other);
    result := compareDeep(modifierElement, o.modifierElement, true) and compareDeep(amountElement, o.amountElement, true);
  end;
end;

function TFhirNutritionOrderOralDietNutrient.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDietNutrient;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDietNutrient)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDietNutrient(other);
    result := true;
  end;
end;

{ TFhirNutritionOrderOralDietNutrient }

Procedure TFhirNutritionOrderOralDietNutrient.SetModifier(value : TFhirCodeableConcept);
begin
  FModifier.free;
  FModifier := value;
end;

Procedure TFhirNutritionOrderOralDietNutrient.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;


{ TFhirNutritionOrderOralDietNutrientListEnumerator }

Constructor TFhirNutritionOrderOralDietNutrientListEnumerator.Create(list : TFhirNutritionOrderOralDietNutrientList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderOralDietNutrientListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderOralDietNutrientListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderOralDietNutrientListEnumerator.GetCurrent : TFhirNutritionOrderOralDietNutrient;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderOralDietNutrientList }
procedure TFhirNutritionOrderOralDietNutrientList.AddItem(value: TFhirNutritionOrderOralDietNutrient);
begin
  assert(value.ClassName = 'TFhirNutritionOrderOralDietNutrient', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderOralDietNutrient');
  add(value);
end;


function TFhirNutritionOrderOralDietNutrientList.Append: TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietNutrientList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderOralDietNutrientList.GetEnumerator : TFhirNutritionOrderOralDietNutrientListEnumerator;
begin
  result := TFhirNutritionOrderOralDietNutrientListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderOralDietNutrientList.Clone: TFhirNutritionOrderOralDietNutrientList;
begin
  result := TFhirNutritionOrderOralDietNutrientList(inherited Clone);
end;

function TFhirNutritionOrderOralDietNutrientList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderOralDietNutrientList.GetItemN(index: Integer): TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietNutrientList.IndexOf(value: TFhirNutritionOrderOralDietNutrient): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderOralDietNutrientList.Insert(index: Integer): TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietNutrientList.InsertItem(index: Integer; value: TFhirNutritionOrderOralDietNutrient);
begin
  assert(value is TFhirNutritionOrderOralDietNutrient);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderOralDietNutrientList.Item(index: Integer): TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietNutrientList.Link: TFhirNutritionOrderOralDietNutrientList;
begin
  result := TFhirNutritionOrderOralDietNutrientList(inherited Link);
end;

procedure TFhirNutritionOrderOralDietNutrientList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderOralDietNutrientList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderOralDietNutrient);
begin
  assert(value is TFhirNutritionOrderOralDietNutrient);
  FhirNutritionOrderOralDietNutrients[index] := value;
end;

procedure TFhirNutritionOrderOralDietNutrientList.SetItemN(index: Integer; value: TFhirNutritionOrderOralDietNutrient);
begin
  assert(value is TFhirNutritionOrderOralDietNutrient);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderOralDietTexture }

constructor TFhirNutritionOrderOralDietTexture.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderOralDietTexture.Destroy;
begin
  FModifier.free;
  FFoodType.free;
  inherited;
end;

procedure TFhirNutritionOrderOralDietTexture.Assign(oSource : TAdvObject);
begin
  inherited;
  modifier := TFhirNutritionOrderOralDietTexture(oSource).modifier.Clone;
  foodType := TFhirNutritionOrderOralDietTexture(oSource).foodType.Clone;
end;

procedure TFhirNutritionOrderOralDietTexture.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'modifier') Then
     list.add(FModifier.Link);
  if (child_name = 'foodType') Then
     list.add(FFoodType.Link);
end;

procedure TFhirNutritionOrderOralDietTexture.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'modifier', 'CodeableConcept', FModifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'foodType', 'CodeableConcept', FFoodType.Link));{2}
end;

procedure TFhirNutritionOrderOralDietTexture.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'modifier') then Modifier := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'foodType') then FoodType := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirNutritionOrderOralDietTexture.FhirType : string;
begin
  result := 'texture';
end;

function TFhirNutritionOrderOralDietTexture.Link : TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture(inherited Link);
end;

function TFhirNutritionOrderOralDietTexture.Clone : TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture(inherited Clone);
end;

function TFhirNutritionOrderOralDietTexture.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDietTexture;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDietTexture)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDietTexture(other);
    result := compareDeep(modifierElement, o.modifierElement, true) and compareDeep(foodTypeElement, o.foodTypeElement, true);
  end;
end;

function TFhirNutritionOrderOralDietTexture.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderOralDietTexture;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderOralDietTexture)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderOralDietTexture(other);
    result := true;
  end;
end;

{ TFhirNutritionOrderOralDietTexture }

Procedure TFhirNutritionOrderOralDietTexture.SetModifier(value : TFhirCodeableConcept);
begin
  FModifier.free;
  FModifier := value;
end;

Procedure TFhirNutritionOrderOralDietTexture.SetFoodType(value : TFhirCodeableConcept);
begin
  FFoodType.free;
  FFoodType := value;
end;


{ TFhirNutritionOrderOralDietTextureListEnumerator }

Constructor TFhirNutritionOrderOralDietTextureListEnumerator.Create(list : TFhirNutritionOrderOralDietTextureList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderOralDietTextureListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderOralDietTextureListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderOralDietTextureListEnumerator.GetCurrent : TFhirNutritionOrderOralDietTexture;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderOralDietTextureList }
procedure TFhirNutritionOrderOralDietTextureList.AddItem(value: TFhirNutritionOrderOralDietTexture);
begin
  assert(value.ClassName = 'TFhirNutritionOrderOralDietTexture', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderOralDietTexture');
  add(value);
end;


function TFhirNutritionOrderOralDietTextureList.Append: TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietTextureList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderOralDietTextureList.GetEnumerator : TFhirNutritionOrderOralDietTextureListEnumerator;
begin
  result := TFhirNutritionOrderOralDietTextureListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderOralDietTextureList.Clone: TFhirNutritionOrderOralDietTextureList;
begin
  result := TFhirNutritionOrderOralDietTextureList(inherited Clone);
end;

function TFhirNutritionOrderOralDietTextureList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderOralDietTextureList.GetItemN(index: Integer): TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietTextureList.IndexOf(value: TFhirNutritionOrderOralDietTexture): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderOralDietTextureList.Insert(index: Integer): TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderOralDietTextureList.InsertItem(index: Integer; value: TFhirNutritionOrderOralDietTexture);
begin
  assert(value is TFhirNutritionOrderOralDietTexture);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderOralDietTextureList.Item(index: Integer): TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture(ObjectByIndex[index]);
end;

function TFhirNutritionOrderOralDietTextureList.Link: TFhirNutritionOrderOralDietTextureList;
begin
  result := TFhirNutritionOrderOralDietTextureList(inherited Link);
end;

procedure TFhirNutritionOrderOralDietTextureList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderOralDietTextureList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderOralDietTexture);
begin
  assert(value is TFhirNutritionOrderOralDietTexture);
  FhirNutritionOrderOralDietTextures[index] := value;
end;

procedure TFhirNutritionOrderOralDietTextureList.SetItemN(index: Integer; value: TFhirNutritionOrderOralDietTexture);
begin
  assert(value is TFhirNutritionOrderOralDietTexture);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderSupplement }

constructor TFhirNutritionOrderSupplement.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderSupplement.Destroy;
begin
  FType_.free;
  FProductName.free;
  FScheduleList.Free;
  FQuantity.free;
  FInstruction.free;
  inherited;
end;

procedure TFhirNutritionOrderSupplement.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirNutritionOrderSupplement(oSource).type_.Clone;
  productNameElement := TFhirNutritionOrderSupplement(oSource).productNameElement.Clone;
  if (TFhirNutritionOrderSupplement(oSource).FScheduleList = nil) then
  begin
    FScheduleList.free;
    FScheduleList := nil;
  end
  else
  begin
    FScheduleList := TFhirTimingList.Create;
    FScheduleList.Assign(TFhirNutritionOrderSupplement(oSource).FScheduleList);
  end;
  quantity := TFhirNutritionOrderSupplement(oSource).quantity.Clone;
  instructionElement := TFhirNutritionOrderSupplement(oSource).instructionElement.Clone;
end;

procedure TFhirNutritionOrderSupplement.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'productName') Then
     list.add(FProductName.Link);
  if (child_name = 'schedule') Then
    list.addAll(FScheduleList);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'instruction') Then
     list.add(FInstruction.Link);
end;

procedure TFhirNutritionOrderSupplement.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'productName', 'string', FProductName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'schedule', 'Timing', FScheduleList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'instruction', 'string', FInstruction.Link));{2}
end;

procedure TFhirNutritionOrderSupplement.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'productName') then ProductNameElement := propValue as TFhirString{5a}
  else if (propName = 'schedule') then ScheduleList.add(propValue as TFhirTiming){2}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'instruction') then InstructionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirNutritionOrderSupplement.FhirType : string;
begin
  result := 'supplement';
end;

function TFhirNutritionOrderSupplement.Link : TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement(inherited Link);
end;

function TFhirNutritionOrderSupplement.Clone : TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement(inherited Clone);
end;

function TFhirNutritionOrderSupplement.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderSupplement;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderSupplement)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderSupplement(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(productNameElement, o.productNameElement, true) and 
      compareDeep(scheduleList, o.scheduleList, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(instructionElement, o.instructionElement, true);
  end;
end;

function TFhirNutritionOrderSupplement.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderSupplement;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderSupplement)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderSupplement(other);
    result := compareValues(productNameElement, o.productNameElement, true) and compareValues(instructionElement, o.instructionElement, true);
  end;
end;

{ TFhirNutritionOrderSupplement }

Procedure TFhirNutritionOrderSupplement.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirNutritionOrderSupplement.SetProductName(value : TFhirString);
begin
  FProductName.free;
  FProductName := value;
end;

Function TFhirNutritionOrderSupplement.GetProductNameST : String;
begin
  if FProductName = nil then
    result := ''
  else
    result := FProductName.value;
end;

Procedure TFhirNutritionOrderSupplement.SetProductNameST(value : String);
begin
  if value <> '' then
  begin
    if FProductName = nil then
      FProductName := TFhirString.create;
    FProductName.value := value
  end
  else if FProductName <> nil then
    FProductName.value := '';
end;

Function TFhirNutritionOrderSupplement.GetScheduleList : TFhirTimingList;
begin
  if FScheduleList = nil then
    FScheduleList := TFhirTimingList.Create;
  result := FScheduleList;
end;

Function TFhirNutritionOrderSupplement.GetHasScheduleList : boolean;
begin
  result := (FScheduleList <> nil) and (FScheduleList.count > 0);
end;

Procedure TFhirNutritionOrderSupplement.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirNutritionOrderSupplement.SetInstruction(value : TFhirString);
begin
  FInstruction.free;
  FInstruction := value;
end;

Function TFhirNutritionOrderSupplement.GetInstructionST : String;
begin
  if FInstruction = nil then
    result := ''
  else
    result := FInstruction.value;
end;

Procedure TFhirNutritionOrderSupplement.SetInstructionST(value : String);
begin
  if value <> '' then
  begin
    if FInstruction = nil then
      FInstruction := TFhirString.create;
    FInstruction.value := value
  end
  else if FInstruction <> nil then
    FInstruction.value := '';
end;


{ TFhirNutritionOrderSupplementListEnumerator }

Constructor TFhirNutritionOrderSupplementListEnumerator.Create(list : TFhirNutritionOrderSupplementList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderSupplementListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderSupplementListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderSupplementListEnumerator.GetCurrent : TFhirNutritionOrderSupplement;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderSupplementList }
procedure TFhirNutritionOrderSupplementList.AddItem(value: TFhirNutritionOrderSupplement);
begin
  assert(value.ClassName = 'TFhirNutritionOrderSupplement', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderSupplement');
  add(value);
end;


function TFhirNutritionOrderSupplementList.Append: TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderSupplementList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderSupplementList.GetEnumerator : TFhirNutritionOrderSupplementListEnumerator;
begin
  result := TFhirNutritionOrderSupplementListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderSupplementList.Clone: TFhirNutritionOrderSupplementList;
begin
  result := TFhirNutritionOrderSupplementList(inherited Clone);
end;

function TFhirNutritionOrderSupplementList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderSupplementList.GetItemN(index: Integer): TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement(ObjectByIndex[index]);
end;

function TFhirNutritionOrderSupplementList.IndexOf(value: TFhirNutritionOrderSupplement): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderSupplementList.Insert(index: Integer): TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderSupplementList.InsertItem(index: Integer; value: TFhirNutritionOrderSupplement);
begin
  assert(value is TFhirNutritionOrderSupplement);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderSupplementList.Item(index: Integer): TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement(ObjectByIndex[index]);
end;

function TFhirNutritionOrderSupplementList.Link: TFhirNutritionOrderSupplementList;
begin
  result := TFhirNutritionOrderSupplementList(inherited Link);
end;

procedure TFhirNutritionOrderSupplementList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderSupplementList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderSupplement);
begin
  assert(value is TFhirNutritionOrderSupplement);
  FhirNutritionOrderSupplements[index] := value;
end;

procedure TFhirNutritionOrderSupplementList.SetItemN(index: Integer; value: TFhirNutritionOrderSupplement);
begin
  assert(value is TFhirNutritionOrderSupplement);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderEnteralFormula }

constructor TFhirNutritionOrderEnteralFormula.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderEnteralFormula.Destroy;
begin
  FBaseFormulaType.free;
  FBaseFormulaProductName.free;
  FAdditiveType.free;
  FAdditiveProductName.free;
  FCaloricDensity.free;
  FRouteofAdministration.free;
  FAdministrationList.Free;
  FMaxVolumeToDeliver.free;
  FAdministrationInstruction.free;
  inherited;
end;

procedure TFhirNutritionOrderEnteralFormula.Assign(oSource : TAdvObject);
begin
  inherited;
  baseFormulaType := TFhirNutritionOrderEnteralFormula(oSource).baseFormulaType.Clone;
  baseFormulaProductNameElement := TFhirNutritionOrderEnteralFormula(oSource).baseFormulaProductNameElement.Clone;
  additiveType := TFhirNutritionOrderEnteralFormula(oSource).additiveType.Clone;
  additiveProductNameElement := TFhirNutritionOrderEnteralFormula(oSource).additiveProductNameElement.Clone;
  caloricDensity := TFhirNutritionOrderEnteralFormula(oSource).caloricDensity.Clone;
  routeofAdministration := TFhirNutritionOrderEnteralFormula(oSource).routeofAdministration.Clone;
  if (TFhirNutritionOrderEnteralFormula(oSource).FAdministrationList = nil) then
  begin
    FAdministrationList.free;
    FAdministrationList := nil;
  end
  else
  begin
    FAdministrationList := TFhirNutritionOrderEnteralFormulaAdministrationList.Create;
    FAdministrationList.Assign(TFhirNutritionOrderEnteralFormula(oSource).FAdministrationList);
  end;
  maxVolumeToDeliver := TFhirNutritionOrderEnteralFormula(oSource).maxVolumeToDeliver.Clone;
  administrationInstructionElement := TFhirNutritionOrderEnteralFormula(oSource).administrationInstructionElement.Clone;
end;

procedure TFhirNutritionOrderEnteralFormula.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'baseFormulaType') Then
     list.add(FBaseFormulaType.Link);
  if (child_name = 'baseFormulaProductName') Then
     list.add(FBaseFormulaProductName.Link);
  if (child_name = 'additiveType') Then
     list.add(FAdditiveType.Link);
  if (child_name = 'additiveProductName') Then
     list.add(FAdditiveProductName.Link);
  if (child_name = 'caloricDensity') Then
     list.add(FCaloricDensity.Link);
  if (child_name = 'routeofAdministration') Then
     list.add(FRouteofAdministration.Link);
  if (child_name = 'administration') Then
    list.addAll(FAdministrationList);
  if (child_name = 'maxVolumeToDeliver') Then
     list.add(FMaxVolumeToDeliver.Link);
  if (child_name = 'administrationInstruction') Then
     list.add(FAdministrationInstruction.Link);
end;

procedure TFhirNutritionOrderEnteralFormula.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'baseFormulaType', 'CodeableConcept', FBaseFormulaType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'baseFormulaProductName', 'string', FBaseFormulaProductName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additiveType', 'CodeableConcept', FAdditiveType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additiveProductName', 'string', FAdditiveProductName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'caloricDensity', 'SimpleQuantity', FCaloricDensity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'routeofAdministration', 'CodeableConcept', FRouteofAdministration.Link));{2}
  oList.add(TFHIRProperty.create(self, 'administration', '', FAdministrationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'maxVolumeToDeliver', 'SimpleQuantity', FMaxVolumeToDeliver.Link));{2}
  oList.add(TFHIRProperty.create(self, 'administrationInstruction', 'string', FAdministrationInstruction.Link));{2}
end;

procedure TFhirNutritionOrderEnteralFormula.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'baseFormulaType') then BaseFormulaType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'baseFormulaProductName') then BaseFormulaProductNameElement := propValue as TFhirString{5a}
  else if (propName = 'additiveType') then AdditiveType := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'additiveProductName') then AdditiveProductNameElement := propValue as TFhirString{5a}
  else if (propName = 'caloricDensity') then CaloricDensity := propValue as TFhirQuantity{4b}
  else if (propName = 'routeofAdministration') then RouteofAdministration := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'administration') then AdministrationList.add(propValue as TFhirNutritionOrderEnteralFormulaAdministration){2}
  else if (propName = 'maxVolumeToDeliver') then MaxVolumeToDeliver := propValue as TFhirQuantity{4b}
  else if (propName = 'administrationInstruction') then AdministrationInstructionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirNutritionOrderEnteralFormula.FhirType : string;
begin
  result := 'enteralFormula';
end;

function TFhirNutritionOrderEnteralFormula.Link : TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula(inherited Link);
end;

function TFhirNutritionOrderEnteralFormula.Clone : TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula(inherited Clone);
end;

function TFhirNutritionOrderEnteralFormula.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderEnteralFormula;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderEnteralFormula)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderEnteralFormula(other);
    result := compareDeep(baseFormulaTypeElement, o.baseFormulaTypeElement, true) and 
      compareDeep(baseFormulaProductNameElement, o.baseFormulaProductNameElement, true) and 
      compareDeep(additiveTypeElement, o.additiveTypeElement, true) and compareDeep(additiveProductNameElement, o.additiveProductNameElement, true) and 
      compareDeep(caloricDensityElement, o.caloricDensityElement, true) and compareDeep(routeofAdministrationElement, o.routeofAdministrationElement, true) and 
      compareDeep(administrationList, o.administrationList, true) and compareDeep(maxVolumeToDeliverElement, o.maxVolumeToDeliverElement, true) and 
      compareDeep(administrationInstructionElement, o.administrationInstructionElement, true);
  end;
end;

function TFhirNutritionOrderEnteralFormula.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderEnteralFormula;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderEnteralFormula)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderEnteralFormula(other);
    result := compareValues(baseFormulaProductNameElement, o.baseFormulaProductNameElement, true) and 
      compareValues(additiveProductNameElement, o.additiveProductNameElement, true) and 
      compareValues(administrationInstructionElement, o.administrationInstructionElement, true);
  end;
end;

{ TFhirNutritionOrderEnteralFormula }

Procedure TFhirNutritionOrderEnteralFormula.SetBaseFormulaType(value : TFhirCodeableConcept);
begin
  FBaseFormulaType.free;
  FBaseFormulaType := value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetBaseFormulaProductName(value : TFhirString);
begin
  FBaseFormulaProductName.free;
  FBaseFormulaProductName := value;
end;

Function TFhirNutritionOrderEnteralFormula.GetBaseFormulaProductNameST : String;
begin
  if FBaseFormulaProductName = nil then
    result := ''
  else
    result := FBaseFormulaProductName.value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetBaseFormulaProductNameST(value : String);
begin
  if value <> '' then
  begin
    if FBaseFormulaProductName = nil then
      FBaseFormulaProductName := TFhirString.create;
    FBaseFormulaProductName.value := value
  end
  else if FBaseFormulaProductName <> nil then
    FBaseFormulaProductName.value := '';
end;

Procedure TFhirNutritionOrderEnteralFormula.SetAdditiveType(value : TFhirCodeableConcept);
begin
  FAdditiveType.free;
  FAdditiveType := value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetAdditiveProductName(value : TFhirString);
begin
  FAdditiveProductName.free;
  FAdditiveProductName := value;
end;

Function TFhirNutritionOrderEnteralFormula.GetAdditiveProductNameST : String;
begin
  if FAdditiveProductName = nil then
    result := ''
  else
    result := FAdditiveProductName.value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetAdditiveProductNameST(value : String);
begin
  if value <> '' then
  begin
    if FAdditiveProductName = nil then
      FAdditiveProductName := TFhirString.create;
    FAdditiveProductName.value := value
  end
  else if FAdditiveProductName <> nil then
    FAdditiveProductName.value := '';
end;

Procedure TFhirNutritionOrderEnteralFormula.SetCaloricDensity(value : TFhirQuantity);
begin
  FCaloricDensity.free;
  FCaloricDensity := value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetRouteofAdministration(value : TFhirCodeableConcept);
begin
  FRouteofAdministration.free;
  FRouteofAdministration := value;
end;

Function TFhirNutritionOrderEnteralFormula.GetAdministrationList : TFhirNutritionOrderEnteralFormulaAdministrationList;
begin
  if FAdministrationList = nil then
    FAdministrationList := TFhirNutritionOrderEnteralFormulaAdministrationList.Create;
  result := FAdministrationList;
end;

Function TFhirNutritionOrderEnteralFormula.GetHasAdministrationList : boolean;
begin
  result := (FAdministrationList <> nil) and (FAdministrationList.count > 0);
end;

Procedure TFhirNutritionOrderEnteralFormula.SetMaxVolumeToDeliver(value : TFhirQuantity);
begin
  FMaxVolumeToDeliver.free;
  FMaxVolumeToDeliver := value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetAdministrationInstruction(value : TFhirString);
begin
  FAdministrationInstruction.free;
  FAdministrationInstruction := value;
end;

Function TFhirNutritionOrderEnteralFormula.GetAdministrationInstructionST : String;
begin
  if FAdministrationInstruction = nil then
    result := ''
  else
    result := FAdministrationInstruction.value;
end;

Procedure TFhirNutritionOrderEnteralFormula.SetAdministrationInstructionST(value : String);
begin
  if value <> '' then
  begin
    if FAdministrationInstruction = nil then
      FAdministrationInstruction := TFhirString.create;
    FAdministrationInstruction.value := value
  end
  else if FAdministrationInstruction <> nil then
    FAdministrationInstruction.value := '';
end;


{ TFhirNutritionOrderEnteralFormulaListEnumerator }

Constructor TFhirNutritionOrderEnteralFormulaListEnumerator.Create(list : TFhirNutritionOrderEnteralFormulaList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderEnteralFormulaListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderEnteralFormulaListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderEnteralFormulaListEnumerator.GetCurrent : TFhirNutritionOrderEnteralFormula;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderEnteralFormulaList }
procedure TFhirNutritionOrderEnteralFormulaList.AddItem(value: TFhirNutritionOrderEnteralFormula);
begin
  assert(value.ClassName = 'TFhirNutritionOrderEnteralFormula', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderEnteralFormula');
  add(value);
end;


function TFhirNutritionOrderEnteralFormulaList.Append: TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderEnteralFormulaList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderEnteralFormulaList.GetEnumerator : TFhirNutritionOrderEnteralFormulaListEnumerator;
begin
  result := TFhirNutritionOrderEnteralFormulaListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderEnteralFormulaList.Clone: TFhirNutritionOrderEnteralFormulaList;
begin
  result := TFhirNutritionOrderEnteralFormulaList(inherited Clone);
end;

function TFhirNutritionOrderEnteralFormulaList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderEnteralFormulaList.GetItemN(index: Integer): TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula(ObjectByIndex[index]);
end;

function TFhirNutritionOrderEnteralFormulaList.IndexOf(value: TFhirNutritionOrderEnteralFormula): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderEnteralFormulaList.Insert(index: Integer): TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderEnteralFormulaList.InsertItem(index: Integer; value: TFhirNutritionOrderEnteralFormula);
begin
  assert(value is TFhirNutritionOrderEnteralFormula);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderEnteralFormulaList.Item(index: Integer): TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula(ObjectByIndex[index]);
end;

function TFhirNutritionOrderEnteralFormulaList.Link: TFhirNutritionOrderEnteralFormulaList;
begin
  result := TFhirNutritionOrderEnteralFormulaList(inherited Link);
end;

procedure TFhirNutritionOrderEnteralFormulaList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderEnteralFormulaList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderEnteralFormula);
begin
  assert(value is TFhirNutritionOrderEnteralFormula);
  FhirNutritionOrderEnteralFormulas[index] := value;
end;

procedure TFhirNutritionOrderEnteralFormulaList.SetItemN(index: Integer; value: TFhirNutritionOrderEnteralFormula);
begin
  assert(value is TFhirNutritionOrderEnteralFormula);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrderEnteralFormulaAdministration }

constructor TFhirNutritionOrderEnteralFormulaAdministration.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrderEnteralFormulaAdministration.Destroy;
begin
  FSchedule.free;
  FQuantity.free;
  FRate.free;
  inherited;
end;

procedure TFhirNutritionOrderEnteralFormulaAdministration.Assign(oSource : TAdvObject);
begin
  inherited;
  schedule := TFhirNutritionOrderEnteralFormulaAdministration(oSource).schedule.Clone;
  quantity := TFhirNutritionOrderEnteralFormulaAdministration(oSource).quantity.Clone;
  rate := TFhirNutritionOrderEnteralFormulaAdministration(oSource).rate.Clone;
end;

procedure TFhirNutritionOrderEnteralFormulaAdministration.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'schedule') Then
     list.add(FSchedule.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'rate[x]') Then
     list.add(FRate.Link);
end;

procedure TFhirNutritionOrderEnteralFormulaAdministration.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'schedule', 'Timing', FSchedule.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rate[x]', 'SimpleQuantity|Ratio', FRate.Link));{2}
end;

procedure TFhirNutritionOrderEnteralFormulaAdministration.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'schedule') then Schedule := propValue as TFhirTiming{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName.startsWith('rate')) then Rate := propValue as TFhirType{4}
  else inherited;
end;

function TFhirNutritionOrderEnteralFormulaAdministration.FhirType : string;
begin
  result := 'administration';
end;

function TFhirNutritionOrderEnteralFormulaAdministration.Link : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration(inherited Link);
end;

function TFhirNutritionOrderEnteralFormulaAdministration.Clone : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration(inherited Clone);
end;

function TFhirNutritionOrderEnteralFormulaAdministration.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderEnteralFormulaAdministration)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderEnteralFormulaAdministration(other);
    result := compareDeep(scheduleElement, o.scheduleElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(rateElement, o.rateElement, true);
  end;
end;

function TFhirNutritionOrderEnteralFormulaAdministration.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirNutritionOrderEnteralFormulaAdministration)) then
    result := false
  else
  begin
    o := TFhirNutritionOrderEnteralFormulaAdministration(other);
    result := true;
  end;
end;

{ TFhirNutritionOrderEnteralFormulaAdministration }

Procedure TFhirNutritionOrderEnteralFormulaAdministration.SetSchedule(value : TFhirTiming);
begin
  FSchedule.free;
  FSchedule := value;
end;

Procedure TFhirNutritionOrderEnteralFormulaAdministration.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirNutritionOrderEnteralFormulaAdministration.SetRate(value : TFhirType);
begin
  FRate.free;
  FRate := value;
end;


{ TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator }

Constructor TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator.Create(list : TFhirNutritionOrderEnteralFormulaAdministrationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator.GetCurrent : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderEnteralFormulaAdministrationList }
procedure TFhirNutritionOrderEnteralFormulaAdministrationList.AddItem(value: TFhirNutritionOrderEnteralFormulaAdministration);
begin
  assert(value.ClassName = 'TFhirNutritionOrderEnteralFormulaAdministration', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrderEnteralFormulaAdministration');
  add(value);
end;


function TFhirNutritionOrderEnteralFormulaAdministrationList.Append: TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderEnteralFormulaAdministrationList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.GetEnumerator : TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministrationListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.Clone: TFhirNutritionOrderEnteralFormulaAdministrationList;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministrationList(inherited Clone);
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.GetItemN(index: Integer): TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration(ObjectByIndex[index]);
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.IndexOf(value: TFhirNutritionOrderEnteralFormulaAdministration): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderEnteralFormulaAdministrationList.Insert(index: Integer): TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderEnteralFormulaAdministrationList.InsertItem(index: Integer; value: TFhirNutritionOrderEnteralFormulaAdministration);
begin
  assert(value is TFhirNutritionOrderEnteralFormulaAdministration);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.Item(index: Integer): TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration(ObjectByIndex[index]);
end;

function TFhirNutritionOrderEnteralFormulaAdministrationList.Link: TFhirNutritionOrderEnteralFormulaAdministrationList;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministrationList(inherited Link);
end;

procedure TFhirNutritionOrderEnteralFormulaAdministrationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderEnteralFormulaAdministrationList.SetItemByIndex(index: Integer; value: TFhirNutritionOrderEnteralFormulaAdministration);
begin
  assert(value is TFhirNutritionOrderEnteralFormulaAdministration);
  FhirNutritionOrderEnteralFormulaAdministrations[index] := value;
end;

procedure TFhirNutritionOrderEnteralFormulaAdministrationList.SetItemN(index: Integer; value: TFhirNutritionOrderEnteralFormulaAdministration);
begin
  assert(value is TFhirNutritionOrderEnteralFormulaAdministration);
  ObjectByIndex[index] := value;
end;

{ TFhirNutritionOrder }

constructor TFhirNutritionOrder.Create;
begin
  inherited;
end;

destructor TFhirNutritionOrder.Destroy;
begin
  FPatient.free;
  FOrderer.free;
  FIdentifierList.Free;
  FEncounter.free;
  FDateTime.free;
  FStatus.free;
  FAllergyIntoleranceList.Free;
  FFoodPreferenceModifierList.Free;
  FExcludeFoodModifierList.Free;
  FOralDiet.free;
  FSupplementList.Free;
  FEnteralFormula.free;
  inherited;
end;

function TFhirNutritionOrder.GetResourceType : TFhirResourceType;
begin
  result := frtNutritionOrder;
end;

procedure TFhirNutritionOrder.Assign(oSource : TAdvObject);
begin
  inherited;
  patient := TFhirNutritionOrder(oSource).patient.Clone;
  orderer := TFhirNutritionOrder(oSource).orderer.Clone;
  if (TFhirNutritionOrder(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirNutritionOrder(oSource).FIdentifierList);
  end;
  encounter := TFhirNutritionOrder(oSource).encounter.Clone;
  dateTimeElement := TFhirNutritionOrder(oSource).dateTimeElement.Clone;
  FStatus := TFhirNutritionOrder(oSource).FStatus.Link;
  if (TFhirNutritionOrder(oSource).FAllergyIntoleranceList = nil) then
  begin
    FAllergyIntoleranceList.free;
    FAllergyIntoleranceList := nil;
  end
  else
  begin
    FAllergyIntoleranceList := TFhirReferenceList{TFhirAllergyIntolerance}.Create;
    FAllergyIntoleranceList.Assign(TFhirNutritionOrder(oSource).FAllergyIntoleranceList);
  end;
  if (TFhirNutritionOrder(oSource).FFoodPreferenceModifierList = nil) then
  begin
    FFoodPreferenceModifierList.free;
    FFoodPreferenceModifierList := nil;
  end
  else
  begin
    FFoodPreferenceModifierList := TFhirCodeableConceptList.Create;
    FFoodPreferenceModifierList.Assign(TFhirNutritionOrder(oSource).FFoodPreferenceModifierList);
  end;
  if (TFhirNutritionOrder(oSource).FExcludeFoodModifierList = nil) then
  begin
    FExcludeFoodModifierList.free;
    FExcludeFoodModifierList := nil;
  end
  else
  begin
    FExcludeFoodModifierList := TFhirCodeableConceptList.Create;
    FExcludeFoodModifierList.Assign(TFhirNutritionOrder(oSource).FExcludeFoodModifierList);
  end;
  oralDiet := TFhirNutritionOrder(oSource).oralDiet.Clone;
  if (TFhirNutritionOrder(oSource).FSupplementList = nil) then
  begin
    FSupplementList.free;
    FSupplementList := nil;
  end
  else
  begin
    FSupplementList := TFhirNutritionOrderSupplementList.Create;
    FSupplementList.Assign(TFhirNutritionOrder(oSource).FSupplementList);
  end;
  enteralFormula := TFhirNutritionOrder(oSource).enteralFormula.Clone;
end;

procedure TFhirNutritionOrder.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'orderer') Then
     list.add(FOrderer.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'dateTime') Then
     list.add(FDateTime.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'allergyIntolerance') Then
    list.addAll(FAllergyIntoleranceList);
  if (child_name = 'foodPreferenceModifier') Then
    list.addAll(FFoodPreferenceModifierList);
  if (child_name = 'excludeFoodModifier') Then
    list.addAll(FExcludeFoodModifierList);
  if (child_name = 'oralDiet') Then
     list.add(FOralDiet.Link);
  if (child_name = 'supplement') Then
    list.addAll(FSupplementList);
  if (child_name = 'enteralFormula') Then
     list.add(FEnteralFormula.Link);
end;

procedure TFhirNutritionOrder.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderer', 'Reference(Practitioner)', FOrderer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateTime', 'dateTime', FDateTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'allergyIntolerance', 'Reference(AllergyIntolerance)', FAllergyIntoleranceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'foodPreferenceModifier', 'CodeableConcept', FFoodPreferenceModifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'excludeFoodModifier', 'CodeableConcept', FExcludeFoodModifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'oralDiet', '', FOralDiet.Link));{2}
  oList.add(TFHIRProperty.create(self, 'supplement', '', FSupplementList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'enteralFormula', '', FEnteralFormula.Link));{2}
end;

procedure TFhirNutritionOrder.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'orderer') then Orderer := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'dateTime') then DateTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'allergyIntolerance') then AllergyIntoleranceList.add(propValue as TFhirReference{TFhirAllergyIntolerance}){2}
  else if (propName = 'foodPreferenceModifier') then FoodPreferenceModifierList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'excludeFoodModifier') then ExcludeFoodModifierList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'oralDiet') then OralDiet := propValue as TFhirNutritionOrderOralDiet{4b}
  else if (propName = 'supplement') then SupplementList.add(propValue as TFhirNutritionOrderSupplement){2}
  else if (propName = 'enteralFormula') then EnteralFormula := propValue as TFhirNutritionOrderEnteralFormula{4b}
  else inherited;
end;

function TFhirNutritionOrder.FhirType : string;
begin
  result := 'NutritionOrder';
end;

function TFhirNutritionOrder.Link : TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder(inherited Link);
end;

function TFhirNutritionOrder.Clone : TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder(inherited Clone);
end;

{ TFhirNutritionOrder }

Procedure TFhirNutritionOrder.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirNutritionOrder.SetOrderer(value : TFhirReference{TFhirPractitioner});
begin
  FOrderer.free;
  FOrderer := value;
end;

Function TFhirNutritionOrder.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirNutritionOrder.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirNutritionOrder.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirNutritionOrder.SetDateTime(value : TFhirDateTime);
begin
  FDateTime.free;
  FDateTime := value;
end;

Function TFhirNutritionOrder.GetDateTimeST : TDateAndTime;
begin
  if FDateTime = nil then
    result := nil
  else
    result := FDateTime.value;
end;

Procedure TFhirNutritionOrder.SetDateTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateTime = nil then
      FDateTime := TFhirDateTime.create;
    FDateTime.value := value
  end
  else if FDateTime <> nil then
    FDateTime.value := nil;
end;

Procedure TFhirNutritionOrder.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirNutritionOrder.GetStatusST : TFhirNutritionOrderStatus;
begin
  if FStatus = nil then
    result := TFhirNutritionOrderStatus(0)
  else
    result := TFhirNutritionOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirNutritionOrderStatus, FStatus.value));
end;

Procedure TFhirNutritionOrder.SetStatusST(value : TFhirNutritionOrderStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirNutritionOrderStatus[value]);
end;

Function TFhirNutritionOrder.GetAllergyIntoleranceList : TFhirReferenceList{TFhirAllergyIntolerance};
begin
  if FAllergyIntoleranceList = nil then
    FAllergyIntoleranceList := TFhirReferenceList{TFhirAllergyIntolerance}.Create;
  result := FAllergyIntoleranceList;
end;

Function TFhirNutritionOrder.GetHasAllergyIntoleranceList : boolean;
begin
  result := (FAllergyIntoleranceList <> nil) and (FAllergyIntoleranceList.count > 0);
end;

Function TFhirNutritionOrder.GetFoodPreferenceModifierList : TFhirCodeableConceptList;
begin
  if FFoodPreferenceModifierList = nil then
    FFoodPreferenceModifierList := TFhirCodeableConceptList.Create;
  result := FFoodPreferenceModifierList;
end;

Function TFhirNutritionOrder.GetHasFoodPreferenceModifierList : boolean;
begin
  result := (FFoodPreferenceModifierList <> nil) and (FFoodPreferenceModifierList.count > 0);
end;

Function TFhirNutritionOrder.GetExcludeFoodModifierList : TFhirCodeableConceptList;
begin
  if FExcludeFoodModifierList = nil then
    FExcludeFoodModifierList := TFhirCodeableConceptList.Create;
  result := FExcludeFoodModifierList;
end;

Function TFhirNutritionOrder.GetHasExcludeFoodModifierList : boolean;
begin
  result := (FExcludeFoodModifierList <> nil) and (FExcludeFoodModifierList.count > 0);
end;

Procedure TFhirNutritionOrder.SetOralDiet(value : TFhirNutritionOrderOralDiet);
begin
  FOralDiet.free;
  FOralDiet := value;
end;

Function TFhirNutritionOrder.GetSupplementList : TFhirNutritionOrderSupplementList;
begin
  if FSupplementList = nil then
    FSupplementList := TFhirNutritionOrderSupplementList.Create;
  result := FSupplementList;
end;

Function TFhirNutritionOrder.GetHasSupplementList : boolean;
begin
  result := (FSupplementList <> nil) and (FSupplementList.count > 0);
end;

Procedure TFhirNutritionOrder.SetEnteralFormula(value : TFhirNutritionOrderEnteralFormula);
begin
  FEnteralFormula.free;
  FEnteralFormula := value;
end;


{ TFhirNutritionOrderListEnumerator }

Constructor TFhirNutritionOrderListEnumerator.Create(list : TFhirNutritionOrderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirNutritionOrderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirNutritionOrderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirNutritionOrderListEnumerator.GetCurrent : TFhirNutritionOrder;
begin
  Result := FList[FIndex];
end;


{ TFhirNutritionOrderList }
procedure TFhirNutritionOrderList.AddItem(value: TFhirNutritionOrder);
begin
  assert(value.ClassName = 'TFhirNutritionOrder', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirNutritionOrder');
  add(value);
end;


function TFhirNutritionOrderList.Append: TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderList.ClearItems;
begin
  Clear;
end;

function TFhirNutritionOrderList.GetEnumerator : TFhirNutritionOrderListEnumerator;
begin
  result := TFhirNutritionOrderListEnumerator.Create(self.link);
end;

function TFhirNutritionOrderList.Clone: TFhirNutritionOrderList;
begin
  result := TFhirNutritionOrderList(inherited Clone);
end;

function TFhirNutritionOrderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirNutritionOrderList.GetItemN(index: Integer): TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder(ObjectByIndex[index]);
end;

function TFhirNutritionOrderList.IndexOf(value: TFhirNutritionOrder): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirNutritionOrderList.Insert(index: Integer): TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirNutritionOrderList.InsertItem(index: Integer; value: TFhirNutritionOrder);
begin
  assert(value is TFhirNutritionOrder);
  Inherited Insert(index, value);
end;

function TFhirNutritionOrderList.Item(index: Integer): TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder(ObjectByIndex[index]);
end;

function TFhirNutritionOrderList.Link: TFhirNutritionOrderList;
begin
  result := TFhirNutritionOrderList(inherited Link);
end;

procedure TFhirNutritionOrderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirNutritionOrderList.SetItemByIndex(index: Integer; value: TFhirNutritionOrder);
begin
  assert(value is TFhirNutritionOrder);
  FhirNutritionOrders[index] := value;
end;

procedure TFhirNutritionOrderList.SetItemN(index: Integer; value: TFhirNutritionOrder);
begin
  assert(value is TFhirNutritionOrder);
  ObjectByIndex[index] := value;
end;

{ TFhirObservationReferenceRange }

constructor TFhirObservationReferenceRange.Create;
begin
  inherited;
end;

destructor TFhirObservationReferenceRange.Destroy;
begin
  FLow.free;
  FHigh.free;
  FMeaning.free;
  FAge.free;
  FText.free;
  inherited;
end;

procedure TFhirObservationReferenceRange.Assign(oSource : TAdvObject);
begin
  inherited;
  low := TFhirObservationReferenceRange(oSource).low.Clone;
  high := TFhirObservationReferenceRange(oSource).high.Clone;
  meaning := TFhirObservationReferenceRange(oSource).meaning.Clone;
  age := TFhirObservationReferenceRange(oSource).age.Clone;
  textElement := TFhirObservationReferenceRange(oSource).textElement.Clone;
end;

procedure TFhirObservationReferenceRange.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'low') Then
     list.add(FLow.Link);
  if (child_name = 'high') Then
     list.add(FHigh.Link);
  if (child_name = 'meaning') Then
     list.add(FMeaning.Link);
  if (child_name = 'age') Then
     list.add(FAge.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
end;

procedure TFhirObservationReferenceRange.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'low', 'SimpleQuantity', FLow.Link));{2}
  oList.add(TFHIRProperty.create(self, 'high', 'SimpleQuantity', FHigh.Link));{2}
  oList.add(TFHIRProperty.create(self, 'meaning', 'CodeableConcept', FMeaning.Link));{2}
  oList.add(TFHIRProperty.create(self, 'age', 'Range', FAge.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
end;

procedure TFhirObservationReferenceRange.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'low') then Low := propValue as TFhirQuantity{4b}
  else if (propName = 'high') then High := propValue as TFhirQuantity{4b}
  else if (propName = 'meaning') then Meaning := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'age') then Age := propValue as TFhirRange{4b}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirObservationReferenceRange.FhirType : string;
begin
  result := 'referenceRange';
end;

function TFhirObservationReferenceRange.Link : TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange(inherited Link);
end;

function TFhirObservationReferenceRange.Clone : TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange(inherited Clone);
end;

function TFhirObservationReferenceRange.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationReferenceRange;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirObservationReferenceRange)) then
    result := false
  else
  begin
    o := TFhirObservationReferenceRange(other);
    result := compareDeep(lowElement, o.lowElement, true) and compareDeep(highElement, o.highElement, true) and 
      compareDeep(meaningElement, o.meaningElement, true) and compareDeep(ageElement, o.ageElement, true) and 
      compareDeep(textElement, o.textElement, true);
  end;
end;

function TFhirObservationReferenceRange.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationReferenceRange;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirObservationReferenceRange)) then
    result := false
  else
  begin
    o := TFhirObservationReferenceRange(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirObservationReferenceRange }

Procedure TFhirObservationReferenceRange.SetLow(value : TFhirQuantity);
begin
  FLow.free;
  FLow := value;
end;

Procedure TFhirObservationReferenceRange.SetHigh(value : TFhirQuantity);
begin
  FHigh.free;
  FHigh := value;
end;

Procedure TFhirObservationReferenceRange.SetMeaning(value : TFhirCodeableConcept);
begin
  FMeaning.free;
  FMeaning := value;
end;

Procedure TFhirObservationReferenceRange.SetAge(value : TFhirRange);
begin
  FAge.free;
  FAge := value;
end;

Procedure TFhirObservationReferenceRange.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirObservationReferenceRange.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirObservationReferenceRange.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;


{ TFhirObservationReferenceRangeListEnumerator }

Constructor TFhirObservationReferenceRangeListEnumerator.Create(list : TFhirObservationReferenceRangeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirObservationReferenceRangeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirObservationReferenceRangeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirObservationReferenceRangeListEnumerator.GetCurrent : TFhirObservationReferenceRange;
begin
  Result := FList[FIndex];
end;


{ TFhirObservationReferenceRangeList }
procedure TFhirObservationReferenceRangeList.AddItem(value: TFhirObservationReferenceRange);
begin
  assert(value.ClassName = 'TFhirObservationReferenceRange', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirObservationReferenceRange');
  add(value);
end;


function TFhirObservationReferenceRangeList.Append: TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationReferenceRangeList.ClearItems;
begin
  Clear;
end;

function TFhirObservationReferenceRangeList.GetEnumerator : TFhirObservationReferenceRangeListEnumerator;
begin
  result := TFhirObservationReferenceRangeListEnumerator.Create(self.link);
end;

function TFhirObservationReferenceRangeList.Clone: TFhirObservationReferenceRangeList;
begin
  result := TFhirObservationReferenceRangeList(inherited Clone);
end;

function TFhirObservationReferenceRangeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirObservationReferenceRangeList.GetItemN(index: Integer): TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange(ObjectByIndex[index]);
end;

function TFhirObservationReferenceRangeList.IndexOf(value: TFhirObservationReferenceRange): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirObservationReferenceRangeList.Insert(index: Integer): TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationReferenceRangeList.InsertItem(index: Integer; value: TFhirObservationReferenceRange);
begin
  assert(value is TFhirObservationReferenceRange);
  Inherited Insert(index, value);
end;

function TFhirObservationReferenceRangeList.Item(index: Integer): TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange(ObjectByIndex[index]);
end;

function TFhirObservationReferenceRangeList.Link: TFhirObservationReferenceRangeList;
begin
  result := TFhirObservationReferenceRangeList(inherited Link);
end;

procedure TFhirObservationReferenceRangeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirObservationReferenceRangeList.SetItemByIndex(index: Integer; value: TFhirObservationReferenceRange);
begin
  assert(value is TFhirObservationReferenceRange);
  FhirObservationReferenceRanges[index] := value;
end;

procedure TFhirObservationReferenceRangeList.SetItemN(index: Integer; value: TFhirObservationReferenceRange);
begin
  assert(value is TFhirObservationReferenceRange);
  ObjectByIndex[index] := value;
end;

{ TFhirObservationRelated }

constructor TFhirObservationRelated.Create;
begin
  inherited;
end;

destructor TFhirObservationRelated.Destroy;
begin
  FType_.free;
  FTarget.free;
  inherited;
end;

procedure TFhirObservationRelated.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirObservationRelated(oSource).FType_.Link;
  target := TFhirObservationRelated(oSource).target.Clone;
end;

procedure TFhirObservationRelated.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
end;

procedure TFhirObservationRelated.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Observation|QuestionnaireResponse)', FTarget.Link));{2}
end;

procedure TFhirObservationRelated.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'target') then Target := propValue as TFhirReference{Resource}{4b}
  else inherited;
end;

function TFhirObservationRelated.FhirType : string;
begin
  result := 'related';
end;

function TFhirObservationRelated.Link : TFhirObservationRelated;
begin
  result := TFhirObservationRelated(inherited Link);
end;

function TFhirObservationRelated.Clone : TFhirObservationRelated;
begin
  result := TFhirObservationRelated(inherited Clone);
end;

function TFhirObservationRelated.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationRelated;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirObservationRelated)) then
    result := false
  else
  begin
    o := TFhirObservationRelated(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(targetElement, o.targetElement, true);
  end;
end;

function TFhirObservationRelated.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationRelated;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirObservationRelated)) then
    result := false
  else
  begin
    o := TFhirObservationRelated(other);
    result := compareValues(type_Element, o.type_Element, true);
  end;
end;

{ TFhirObservationRelated }

Procedure TFhirObservationRelated.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirObservationRelated.GetType_ST : TFhirObservationRelationshiptypes;
begin
  if FType_ = nil then
    result := TFhirObservationRelationshiptypes(0)
  else
    result := TFhirObservationRelationshiptypes(StringArrayIndexOfSensitive(CODES_TFhirObservationRelationshiptypes, FType_.value));
end;

Procedure TFhirObservationRelated.SetType_ST(value : TFhirObservationRelationshiptypes);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirObservationRelationshiptypes[value]);
end;

Procedure TFhirObservationRelated.SetTarget(value : TFhirReference{Resource});
begin
  FTarget.free;
  FTarget := value;
end;


{ TFhirObservationRelatedListEnumerator }

Constructor TFhirObservationRelatedListEnumerator.Create(list : TFhirObservationRelatedList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirObservationRelatedListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirObservationRelatedListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirObservationRelatedListEnumerator.GetCurrent : TFhirObservationRelated;
begin
  Result := FList[FIndex];
end;


{ TFhirObservationRelatedList }
procedure TFhirObservationRelatedList.AddItem(value: TFhirObservationRelated);
begin
  assert(value.ClassName = 'TFhirObservationRelated', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirObservationRelated');
  add(value);
end;


function TFhirObservationRelatedList.Append: TFhirObservationRelated;
begin
  result := TFhirObservationRelated.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationRelatedList.ClearItems;
begin
  Clear;
end;

function TFhirObservationRelatedList.GetEnumerator : TFhirObservationRelatedListEnumerator;
begin
  result := TFhirObservationRelatedListEnumerator.Create(self.link);
end;

function TFhirObservationRelatedList.Clone: TFhirObservationRelatedList;
begin
  result := TFhirObservationRelatedList(inherited Clone);
end;

function TFhirObservationRelatedList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirObservationRelatedList.GetItemN(index: Integer): TFhirObservationRelated;
begin
  result := TFhirObservationRelated(ObjectByIndex[index]);
end;

function TFhirObservationRelatedList.IndexOf(value: TFhirObservationRelated): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirObservationRelatedList.Insert(index: Integer): TFhirObservationRelated;
begin
  result := TFhirObservationRelated.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationRelatedList.InsertItem(index: Integer; value: TFhirObservationRelated);
begin
  assert(value is TFhirObservationRelated);
  Inherited Insert(index, value);
end;

function TFhirObservationRelatedList.Item(index: Integer): TFhirObservationRelated;
begin
  result := TFhirObservationRelated(ObjectByIndex[index]);
end;

function TFhirObservationRelatedList.Link: TFhirObservationRelatedList;
begin
  result := TFhirObservationRelatedList(inherited Link);
end;

procedure TFhirObservationRelatedList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirObservationRelatedList.SetItemByIndex(index: Integer; value: TFhirObservationRelated);
begin
  assert(value is TFhirObservationRelated);
  FhirObservationRelateds[index] := value;
end;

procedure TFhirObservationRelatedList.SetItemN(index: Integer; value: TFhirObservationRelated);
begin
  assert(value is TFhirObservationRelated);
  ObjectByIndex[index] := value;
end;

{ TFhirObservationComponent }

constructor TFhirObservationComponent.Create;
begin
  inherited;
end;

destructor TFhirObservationComponent.Destroy;
begin
  FCode.free;
  FValue.free;
  FDataAbsentReason.free;
  FReferenceRangeList.Free;
  inherited;
end;

procedure TFhirObservationComponent.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirObservationComponent(oSource).code.Clone;
  value := TFhirObservationComponent(oSource).value.Clone;
  dataAbsentReason := TFhirObservationComponent(oSource).dataAbsentReason.Clone;
  if (TFhirObservationComponent(oSource).FReferenceRangeList = nil) then
  begin
    FReferenceRangeList.free;
    FReferenceRangeList := nil;
  end
  else
  begin
    FReferenceRangeList := TFhirObservationReferenceRangeList.Create;
    FReferenceRangeList.Assign(TFhirObservationComponent(oSource).FReferenceRangeList);
  end;
end;

procedure TFhirObservationComponent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
  if (child_name = 'dataAbsentReason') Then
     list.add(FDataAbsentReason.Link);
  if (child_name = 'referenceRange') Then
    list.addAll(FReferenceRangeList);
end;

procedure TFhirObservationComponent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value[x]', 'Quantity|CodeableConcept|string|Range|Ratio|SampledData|Attachment|time|dateTime|Period', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dataAbsentReason', 'CodeableConcept', FDataAbsentReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'referenceRange', '@Observation.referenceRange', FReferenceRangeList.Link)){3};
end;

procedure TFhirObservationComponent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else if (propName = 'dataAbsentReason') then DataAbsentReason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'referenceRange') then ReferenceRangeList.add(propValue as TFhirObservationReferenceRange){2}
  else inherited;
end;

function TFhirObservationComponent.FhirType : string;
begin
  result := 'component';
end;

function TFhirObservationComponent.Link : TFhirObservationComponent;
begin
  result := TFhirObservationComponent(inherited Link);
end;

function TFhirObservationComponent.Clone : TFhirObservationComponent;
begin
  result := TFhirObservationComponent(inherited Clone);
end;

function TFhirObservationComponent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationComponent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirObservationComponent)) then
    result := false
  else
  begin
    o := TFhirObservationComponent(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(valueElement, o.valueElement, true) and 
      compareDeep(dataAbsentReasonElement, o.dataAbsentReasonElement, true) and compareDeep(referenceRangeList, o.referenceRangeList, true);
  end;
end;

function TFhirObservationComponent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirObservationComponent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirObservationComponent)) then
    result := false
  else
  begin
    o := TFhirObservationComponent(other);
    result := true;
  end;
end;

{ TFhirObservationComponent }

Procedure TFhirObservationComponent.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirObservationComponent.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;

Procedure TFhirObservationComponent.SetDataAbsentReason(value : TFhirCodeableConcept);
begin
  FDataAbsentReason.free;
  FDataAbsentReason := value;
end;

Function TFhirObservationComponent.GetReferenceRangeList : TFhirObservationReferenceRangeList;
begin
  if FReferenceRangeList = nil then
    FReferenceRangeList := TFhirObservationReferenceRangeList.Create;
  result := FReferenceRangeList;
end;

Function TFhirObservationComponent.GetHasReferenceRangeList : boolean;
begin
  result := (FReferenceRangeList <> nil) and (FReferenceRangeList.count > 0);
end;


{ TFhirObservationComponentListEnumerator }

Constructor TFhirObservationComponentListEnumerator.Create(list : TFhirObservationComponentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirObservationComponentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirObservationComponentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirObservationComponentListEnumerator.GetCurrent : TFhirObservationComponent;
begin
  Result := FList[FIndex];
end;


{ TFhirObservationComponentList }
procedure TFhirObservationComponentList.AddItem(value: TFhirObservationComponent);
begin
  assert(value.ClassName = 'TFhirObservationComponent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirObservationComponent');
  add(value);
end;


function TFhirObservationComponentList.Append: TFhirObservationComponent;
begin
  result := TFhirObservationComponent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationComponentList.ClearItems;
begin
  Clear;
end;

function TFhirObservationComponentList.GetEnumerator : TFhirObservationComponentListEnumerator;
begin
  result := TFhirObservationComponentListEnumerator.Create(self.link);
end;

function TFhirObservationComponentList.Clone: TFhirObservationComponentList;
begin
  result := TFhirObservationComponentList(inherited Clone);
end;

function TFhirObservationComponentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirObservationComponentList.GetItemN(index: Integer): TFhirObservationComponent;
begin
  result := TFhirObservationComponent(ObjectByIndex[index]);
end;

function TFhirObservationComponentList.IndexOf(value: TFhirObservationComponent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirObservationComponentList.Insert(index: Integer): TFhirObservationComponent;
begin
  result := TFhirObservationComponent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationComponentList.InsertItem(index: Integer; value: TFhirObservationComponent);
begin
  assert(value is TFhirObservationComponent);
  Inherited Insert(index, value);
end;

function TFhirObservationComponentList.Item(index: Integer): TFhirObservationComponent;
begin
  result := TFhirObservationComponent(ObjectByIndex[index]);
end;

function TFhirObservationComponentList.Link: TFhirObservationComponentList;
begin
  result := TFhirObservationComponentList(inherited Link);
end;

procedure TFhirObservationComponentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirObservationComponentList.SetItemByIndex(index: Integer; value: TFhirObservationComponent);
begin
  assert(value is TFhirObservationComponent);
  FhirObservationComponents[index] := value;
end;

procedure TFhirObservationComponentList.SetItemN(index: Integer; value: TFhirObservationComponent);
begin
  assert(value is TFhirObservationComponent);
  ObjectByIndex[index] := value;
end;

{ TFhirObservation }

constructor TFhirObservation.Create;
begin
  inherited;
end;

destructor TFhirObservation.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FCategory.free;
  FCode.free;
  FSubject.free;
  FEncounter.free;
  FEffective.free;
  FIssued.free;
  FPerformerList.Free;
  FValue.free;
  FDataAbsentReason.free;
  FInterpretation.free;
  FComments.free;
  FBodySite.free;
  FMethod.free;
  FSpecimen.free;
  FDevice.free;
  FReferenceRangeList.Free;
  FRelatedList.Free;
  FComponentList.Free;
  inherited;
end;

function TFhirObservation.GetResourceType : TFhirResourceType;
begin
  result := frtObservation;
end;

procedure TFhirObservation.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirObservation(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirObservation(oSource).FIdentifierList);
  end;
  FStatus := TFhirObservation(oSource).FStatus.Link;
  category := TFhirObservation(oSource).category.Clone;
  code := TFhirObservation(oSource).code.Clone;
  subject := TFhirObservation(oSource).subject.Clone;
  encounter := TFhirObservation(oSource).encounter.Clone;
  effective := TFhirObservation(oSource).effective.Clone;
  issuedElement := TFhirObservation(oSource).issuedElement.Clone;
  if (TFhirObservation(oSource).FPerformerList = nil) then
  begin
    FPerformerList.free;
    FPerformerList := nil;
  end
  else
  begin
    FPerformerList := TFhirReferenceList{Resource}.Create;
    FPerformerList.Assign(TFhirObservation(oSource).FPerformerList);
  end;
  value := TFhirObservation(oSource).value.Clone;
  dataAbsentReason := TFhirObservation(oSource).dataAbsentReason.Clone;
  interpretation := TFhirObservation(oSource).interpretation.Clone;
  commentsElement := TFhirObservation(oSource).commentsElement.Clone;
  bodySite := TFhirObservation(oSource).bodySite.Clone;
  method := TFhirObservation(oSource).method.Clone;
  specimen := TFhirObservation(oSource).specimen.Clone;
  device := TFhirObservation(oSource).device.Clone;
  if (TFhirObservation(oSource).FReferenceRangeList = nil) then
  begin
    FReferenceRangeList.free;
    FReferenceRangeList := nil;
  end
  else
  begin
    FReferenceRangeList := TFhirObservationReferenceRangeList.Create;
    FReferenceRangeList.Assign(TFhirObservation(oSource).FReferenceRangeList);
  end;
  if (TFhirObservation(oSource).FRelatedList = nil) then
  begin
    FRelatedList.free;
    FRelatedList := nil;
  end
  else
  begin
    FRelatedList := TFhirObservationRelatedList.Create;
    FRelatedList.Assign(TFhirObservation(oSource).FRelatedList);
  end;
  if (TFhirObservation(oSource).FComponentList = nil) then
  begin
    FComponentList.free;
    FComponentList := nil;
  end
  else
  begin
    FComponentList := TFhirObservationComponentList.Create;
    FComponentList.Assign(TFhirObservation(oSource).FComponentList);
  end;
end;

procedure TFhirObservation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'effective[x]') Then
     list.add(FEffective.Link);
  if (child_name = 'issued') Then
     list.add(FIssued.Link);
  if (child_name = 'performer') Then
    list.addAll(FPerformerList);
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
  if (child_name = 'dataAbsentReason') Then
     list.add(FDataAbsentReason.Link);
  if (child_name = 'interpretation') Then
     list.add(FInterpretation.Link);
  if (child_name = 'comments') Then
     list.add(FComments.Link);
  if (child_name = 'bodySite') Then
     list.add(FBodySite.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'specimen') Then
     list.add(FSpecimen.Link);
  if (child_name = 'device') Then
     list.add(FDevice.Link);
  if (child_name = 'referenceRange') Then
    list.addAll(FReferenceRangeList);
  if (child_name = 'related') Then
    list.addAll(FRelatedList);
  if (child_name = 'component') Then
    list.addAll(FComponentList);
end;

procedure TFhirObservation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Device|Location)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'effective[x]', 'dateTime|Period', FEffective.Link));{2}
  oList.add(TFHIRProperty.create(self, 'issued', 'instant', FIssued.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner|Organization|Patient|RelatedPerson)', FPerformerList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'value[x]', 'Quantity|CodeableConcept|string|Range|Ratio|SampledData|Attachment|time|dateTime|Period', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dataAbsentReason', 'CodeableConcept', FDataAbsentReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'interpretation', 'CodeableConcept', FInterpretation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comments', 'string', FComments.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySite.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specimen', 'Reference(Specimen)', FSpecimen.Link));{2}
  oList.add(TFHIRProperty.create(self, 'device', 'Reference(Device|DeviceMetric)', FDevice.Link));{2}
  oList.add(TFHIRProperty.create(self, 'referenceRange', '', FReferenceRangeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'related', '', FRelatedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'component', '', FComponentList.Link)){3};
end;

procedure TFhirObservation.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName.startsWith('effective')) then Effective := propValue as TFhirType{4}
  else if (propName = 'issued') then IssuedElement := propValue as TFhirInstant{5a}
  else if (propName = 'performer') then PerformerList.add(propValue as TFhirReference{Resource}){2}
  else if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else if (propName = 'dataAbsentReason') then DataAbsentReason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'interpretation') then Interpretation := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'comments') then CommentsElement := propValue as TFhirString{5a}
  else if (propName = 'bodySite') then BodySite := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'specimen') then Specimen := propValue as TFhirReference{TFhirSpecimen}{4b}
  else if (propName = 'device') then Device := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'referenceRange') then ReferenceRangeList.add(propValue as TFhirObservationReferenceRange){2}
  else if (propName = 'related') then RelatedList.add(propValue as TFhirObservationRelated){2}
  else if (propName = 'component') then ComponentList.add(propValue as TFhirObservationComponent){2}
  else inherited;
end;

function TFhirObservation.FhirType : string;
begin
  result := 'Observation';
end;

function TFhirObservation.Link : TFhirObservation;
begin
  result := TFhirObservation(inherited Link);
end;

function TFhirObservation.Clone : TFhirObservation;
begin
  result := TFhirObservation(inherited Clone);
end;

{ TFhirObservation }

Function TFhirObservation.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirObservation.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirObservation.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirObservation.GetStatusST : TFhirObservationStatus;
begin
  if FStatus = nil then
    result := TFhirObservationStatus(0)
  else
    result := TFhirObservationStatus(StringArrayIndexOfSensitive(CODES_TFhirObservationStatus, FStatus.value));
end;

Procedure TFhirObservation.SetStatusST(value : TFhirObservationStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirObservationStatus[value]);
end;

Procedure TFhirObservation.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirObservation.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirObservation.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirObservation.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirObservation.SetEffective(value : TFhirType);
begin
  FEffective.free;
  FEffective := value;
end;

Procedure TFhirObservation.SetIssued(value : TFhirInstant);
begin
  FIssued.free;
  FIssued := value;
end;

Function TFhirObservation.GetIssuedST : TDateAndTime;
begin
  if FIssued = nil then
    result := nil
  else
    result := FIssued.value;
end;

Procedure TFhirObservation.SetIssuedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FIssued = nil then
      FIssued := TFhirInstant.create;
    FIssued.value := value
  end
  else if FIssued <> nil then
    FIssued.value := nil;
end;

Function TFhirObservation.GetPerformerList : TFhirReferenceList{Resource};
begin
  if FPerformerList = nil then
    FPerformerList := TFhirReferenceList{Resource}.Create;
  result := FPerformerList;
end;

Function TFhirObservation.GetHasPerformerList : boolean;
begin
  result := (FPerformerList <> nil) and (FPerformerList.count > 0);
end;

Procedure TFhirObservation.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;

Procedure TFhirObservation.SetDataAbsentReason(value : TFhirCodeableConcept);
begin
  FDataAbsentReason.free;
  FDataAbsentReason := value;
end;

Procedure TFhirObservation.SetInterpretation(value : TFhirCodeableConcept);
begin
  FInterpretation.free;
  FInterpretation := value;
end;

Procedure TFhirObservation.SetComments(value : TFhirString);
begin
  FComments.free;
  FComments := value;
end;

Function TFhirObservation.GetCommentsST : String;
begin
  if FComments = nil then
    result := ''
  else
    result := FComments.value;
end;

Procedure TFhirObservation.SetCommentsST(value : String);
begin
  if value <> '' then
  begin
    if FComments = nil then
      FComments := TFhirString.create;
    FComments.value := value
  end
  else if FComments <> nil then
    FComments.value := '';
end;

Procedure TFhirObservation.SetBodySite(value : TFhirCodeableConcept);
begin
  FBodySite.free;
  FBodySite := value;
end;

Procedure TFhirObservation.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirObservation.SetSpecimen(value : TFhirReference{TFhirSpecimen});
begin
  FSpecimen.free;
  FSpecimen := value;
end;

Procedure TFhirObservation.SetDevice(value : TFhirReference{Resource});
begin
  FDevice.free;
  FDevice := value;
end;

Function TFhirObservation.GetReferenceRangeList : TFhirObservationReferenceRangeList;
begin
  if FReferenceRangeList = nil then
    FReferenceRangeList := TFhirObservationReferenceRangeList.Create;
  result := FReferenceRangeList;
end;

Function TFhirObservation.GetHasReferenceRangeList : boolean;
begin
  result := (FReferenceRangeList <> nil) and (FReferenceRangeList.count > 0);
end;

Function TFhirObservation.GetRelatedList : TFhirObservationRelatedList;
begin
  if FRelatedList = nil then
    FRelatedList := TFhirObservationRelatedList.Create;
  result := FRelatedList;
end;

Function TFhirObservation.GetHasRelatedList : boolean;
begin
  result := (FRelatedList <> nil) and (FRelatedList.count > 0);
end;

Function TFhirObservation.GetComponentList : TFhirObservationComponentList;
begin
  if FComponentList = nil then
    FComponentList := TFhirObservationComponentList.Create;
  result := FComponentList;
end;

Function TFhirObservation.GetHasComponentList : boolean;
begin
  result := (FComponentList <> nil) and (FComponentList.count > 0);
end;


{ TFhirObservationListEnumerator }

Constructor TFhirObservationListEnumerator.Create(list : TFhirObservationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirObservationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirObservationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirObservationListEnumerator.GetCurrent : TFhirObservation;
begin
  Result := FList[FIndex];
end;


{ TFhirObservationList }
procedure TFhirObservationList.AddItem(value: TFhirObservation);
begin
  assert(value.ClassName = 'TFhirObservation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirObservation');
  add(value);
end;


function TFhirObservationList.Append: TFhirObservation;
begin
  result := TFhirObservation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationList.ClearItems;
begin
  Clear;
end;

function TFhirObservationList.GetEnumerator : TFhirObservationListEnumerator;
begin
  result := TFhirObservationListEnumerator.Create(self.link);
end;

function TFhirObservationList.Clone: TFhirObservationList;
begin
  result := TFhirObservationList(inherited Clone);
end;

function TFhirObservationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirObservationList.GetItemN(index: Integer): TFhirObservation;
begin
  result := TFhirObservation(ObjectByIndex[index]);
end;

function TFhirObservationList.IndexOf(value: TFhirObservation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirObservationList.Insert(index: Integer): TFhirObservation;
begin
  result := TFhirObservation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirObservationList.InsertItem(index: Integer; value: TFhirObservation);
begin
  assert(value is TFhirObservation);
  Inherited Insert(index, value);
end;

function TFhirObservationList.Item(index: Integer): TFhirObservation;
begin
  result := TFhirObservation(ObjectByIndex[index]);
end;

function TFhirObservationList.Link: TFhirObservationList;
begin
  result := TFhirObservationList(inherited Link);
end;

procedure TFhirObservationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirObservationList.SetItemByIndex(index: Integer; value: TFhirObservation);
begin
  assert(value is TFhirObservation);
  FhirObservations[index] := value;
end;

procedure TFhirObservationList.SetItemN(index: Integer; value: TFhirObservation);
begin
  assert(value is TFhirObservation);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationDefinitionContact }

constructor TFhirOperationDefinitionContact.Create;
begin
  inherited;
end;

destructor TFhirOperationDefinitionContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirOperationDefinitionContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirOperationDefinitionContact(oSource).nameElement.Clone;
  if (TFhirOperationDefinitionContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirOperationDefinitionContact(oSource).FTelecomList);
  end;
end;

procedure TFhirOperationDefinitionContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirOperationDefinitionContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirOperationDefinitionContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirOperationDefinitionContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirOperationDefinitionContact.Link : TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact(inherited Link);
end;

function TFhirOperationDefinitionContact.Clone : TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact(inherited Clone);
end;

function TFhirOperationDefinitionContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionContact)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirOperationDefinitionContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionContact)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirOperationDefinitionContact }

Procedure TFhirOperationDefinitionContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirOperationDefinitionContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirOperationDefinitionContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirOperationDefinitionContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirOperationDefinitionContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirOperationDefinitionContactListEnumerator }

Constructor TFhirOperationDefinitionContactListEnumerator.Create(list : TFhirOperationDefinitionContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationDefinitionContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationDefinitionContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationDefinitionContactListEnumerator.GetCurrent : TFhirOperationDefinitionContact;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationDefinitionContactList }
procedure TFhirOperationDefinitionContactList.AddItem(value: TFhirOperationDefinitionContact);
begin
  assert(value.ClassName = 'TFhirOperationDefinitionContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationDefinitionContact');
  add(value);
end;


function TFhirOperationDefinitionContactList.Append: TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionContactList.ClearItems;
begin
  Clear;
end;

function TFhirOperationDefinitionContactList.GetEnumerator : TFhirOperationDefinitionContactListEnumerator;
begin
  result := TFhirOperationDefinitionContactListEnumerator.Create(self.link);
end;

function TFhirOperationDefinitionContactList.Clone: TFhirOperationDefinitionContactList;
begin
  result := TFhirOperationDefinitionContactList(inherited Clone);
end;

function TFhirOperationDefinitionContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationDefinitionContactList.GetItemN(index: Integer): TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionContactList.IndexOf(value: TFhirOperationDefinitionContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationDefinitionContactList.Insert(index: Integer): TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionContactList.InsertItem(index: Integer; value: TFhirOperationDefinitionContact);
begin
  assert(value is TFhirOperationDefinitionContact);
  Inherited Insert(index, value);
end;

function TFhirOperationDefinitionContactList.Item(index: Integer): TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionContactList.Link: TFhirOperationDefinitionContactList;
begin
  result := TFhirOperationDefinitionContactList(inherited Link);
end;

procedure TFhirOperationDefinitionContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationDefinitionContactList.SetItemByIndex(index: Integer; value: TFhirOperationDefinitionContact);
begin
  assert(value is TFhirOperationDefinitionContact);
  FhirOperationDefinitionContacts[index] := value;
end;

procedure TFhirOperationDefinitionContactList.SetItemN(index: Integer; value: TFhirOperationDefinitionContact);
begin
  assert(value is TFhirOperationDefinitionContact);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationDefinitionParameter }

constructor TFhirOperationDefinitionParameter.Create;
begin
  inherited;
end;

destructor TFhirOperationDefinitionParameter.Destroy;
begin
  FName.free;
  FUse.free;
  FMin.free;
  FMax.free;
  FDocumentation.free;
  FType_.free;
  FProfile.free;
  FBinding.free;
  FPartList.Free;
  inherited;
end;

procedure TFhirOperationDefinitionParameter.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirOperationDefinitionParameter(oSource).nameElement.Clone;
  FUse := TFhirOperationDefinitionParameter(oSource).FUse.Link;
  minElement := TFhirOperationDefinitionParameter(oSource).minElement.Clone;
  maxElement := TFhirOperationDefinitionParameter(oSource).maxElement.Clone;
  documentationElement := TFhirOperationDefinitionParameter(oSource).documentationElement.Clone;
  FType_ := TFhirOperationDefinitionParameter(oSource).FType_.Link;
  profile := TFhirOperationDefinitionParameter(oSource).profile.Clone;
  binding := TFhirOperationDefinitionParameter(oSource).binding.Clone;
  if (TFhirOperationDefinitionParameter(oSource).FPartList = nil) then
  begin
    FPartList.free;
    FPartList := nil;
  end
  else
  begin
    FPartList := TFhirOperationDefinitionParameterList.Create;
    FPartList.Assign(TFhirOperationDefinitionParameter(oSource).FPartList);
  end;
end;

procedure TFhirOperationDefinitionParameter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'use') Then
     list.add(FUse.Link);
  if (child_name = 'min') Then
     list.add(FMin.Link);
  if (child_name = 'max') Then
     list.add(FMax.Link);
  if (child_name = 'documentation') Then
     list.add(FDocumentation.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'profile') Then
     list.add(FProfile.Link);
  if (child_name = 'binding') Then
     list.add(FBinding.Link);
  if (child_name = 'part') Then
    list.addAll(FPartList);
end;

procedure TFhirOperationDefinitionParameter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'code', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'use', 'code', FUse.Link));{1}
  oList.add(TFHIRProperty.create(self, 'min', 'integer', FMin.Link));{2}
  oList.add(TFHIRProperty.create(self, 'max', 'string', FMax.Link));{2}
  oList.add(TFHIRProperty.create(self, 'documentation', 'string', FDocumentation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(StructureDefinition)', FProfile.Link));{2}
  oList.add(TFHIRProperty.create(self, 'binding', '', FBinding.Link));{2}
  oList.add(TFHIRProperty.create(self, 'part', '@OperationDefinition.parameter', FPartList.Link)){3};
end;

procedure TFhirOperationDefinitionParameter.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then
    if propValue is TFHIRCode then
      NameElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      NameElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'use') then UseElement := propValue as TFHIREnum
  else if (propName = 'min') then MinElement := propValue as TFhirInteger{5a}
  else if (propName = 'max') then MaxElement := propValue as TFhirString{5a}
  else if (propName = 'documentation') then DocumentationElement := propValue as TFhirString{5a}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'profile') then Profile := propValue as TFhirReference{TFhirStructureDefinition}{4b}
  else if (propName = 'binding') then Binding := propValue as TFhirOperationDefinitionParameterBinding{4b}
  else if (propName = 'part') then PartList.add(propValue as TFhirOperationDefinitionParameter){2}
  else inherited;
end;

function TFhirOperationDefinitionParameter.FhirType : string;
begin
  result := 'parameter';
end;

function TFhirOperationDefinitionParameter.Link : TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter(inherited Link);
end;

function TFhirOperationDefinitionParameter.Clone : TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter(inherited Clone);
end;

function TFhirOperationDefinitionParameter.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionParameter;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionParameter)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionParameter(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(useElement, o.useElement, true) and 
      compareDeep(minElement, o.minElement, true) and compareDeep(maxElement, o.maxElement, true) and 
      compareDeep(documentationElement, o.documentationElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(profileElement, o.profileElement, true) and compareDeep(bindingElement, o.bindingElement, true) and 
      compareDeep(partList, o.partList, true);
  end;
end;

function TFhirOperationDefinitionParameter.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionParameter;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionParameter)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionParameter(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(useElement, o.useElement, true) and 
      compareValues(minElement, o.minElement, true) and compareValues(maxElement, o.maxElement, true) and 
      compareValues(documentationElement, o.documentationElement, true) and compareValues(type_Element, o.type_Element, true);
  end;
end;

{ TFhirOperationDefinitionParameter }

Procedure TFhirOperationDefinitionParameter.SetName(value : TFhirCode);
begin
  FName.free;
  FName := value;
end;

Function TFhirOperationDefinitionParameter.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirOperationDefinitionParameter.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirCode.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirOperationDefinitionParameter.SetUse(value : TFhirEnum);
begin
  FUse.free;
  FUse := value;
end;

Function TFhirOperationDefinitionParameter.GetUseST : TFhirOperationParameterUse;
begin
  if FUse = nil then
    result := TFhirOperationParameterUse(0)
  else
    result := TFhirOperationParameterUse(StringArrayIndexOfSensitive(CODES_TFhirOperationParameterUse, FUse.value));
end;

Procedure TFhirOperationDefinitionParameter.SetUseST(value : TFhirOperationParameterUse);
begin
  if ord(value) = 0 then
    UseElement := nil
  else
    UseElement := TFhirEnum.create(CODES_TFhirOperationParameterUse[value]);
end;

Procedure TFhirOperationDefinitionParameter.SetMin(value : TFhirInteger);
begin
  FMin.free;
  FMin := value;
end;

Function TFhirOperationDefinitionParameter.GetMinST : String;
begin
  if FMin = nil then
    result := ''
  else
    result := FMin.value;
end;

Procedure TFhirOperationDefinitionParameter.SetMinST(value : String);
begin
  if value <> '' then
  begin
    if FMin = nil then
      FMin := TFhirInteger.create;
    FMin.value := value
  end
  else if FMin <> nil then
    FMin.value := '';
end;

Procedure TFhirOperationDefinitionParameter.SetMax(value : TFhirString);
begin
  FMax.free;
  FMax := value;
end;

Function TFhirOperationDefinitionParameter.GetMaxST : String;
begin
  if FMax = nil then
    result := ''
  else
    result := FMax.value;
end;

Procedure TFhirOperationDefinitionParameter.SetMaxST(value : String);
begin
  if value <> '' then
  begin
    if FMax = nil then
      FMax := TFhirString.create;
    FMax.value := value
  end
  else if FMax <> nil then
    FMax.value := '';
end;

Procedure TFhirOperationDefinitionParameter.SetDocumentation(value : TFhirString);
begin
  FDocumentation.free;
  FDocumentation := value;
end;

Function TFhirOperationDefinitionParameter.GetDocumentationST : String;
begin
  if FDocumentation = nil then
    result := ''
  else
    result := FDocumentation.value;
end;

Procedure TFhirOperationDefinitionParameter.SetDocumentationST(value : String);
begin
  if value <> '' then
  begin
    if FDocumentation = nil then
      FDocumentation := TFhirString.create;
    FDocumentation.value := value
  end
  else if FDocumentation <> nil then
    FDocumentation.value := '';
end;

Procedure TFhirOperationDefinitionParameter.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirOperationDefinitionParameter.GetType_ST : TFhirValuesetOperationParameterType;
begin
  if FType_ = nil then
    result := TFhirValuesetOperationParameterType(0)
  else
    result := TFhirValuesetOperationParameterType(StringArrayIndexOfSensitive(CODES_TFhirValuesetOperationParameterType, FType_.value));
end;

Procedure TFhirOperationDefinitionParameter.SetType_ST(value : TFhirValuesetOperationParameterType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirValuesetOperationParameterType[value]);
end;

Procedure TFhirOperationDefinitionParameter.SetProfile(value : TFhirReference{TFhirStructureDefinition});
begin
  FProfile.free;
  FProfile := value;
end;

Procedure TFhirOperationDefinitionParameter.SetBinding(value : TFhirOperationDefinitionParameterBinding);
begin
  FBinding.free;
  FBinding := value;
end;

Function TFhirOperationDefinitionParameter.GetPartList : TFhirOperationDefinitionParameterList;
begin
  if FPartList = nil then
    FPartList := TFhirOperationDefinitionParameterList.Create;
  result := FPartList;
end;

Function TFhirOperationDefinitionParameter.GetHasPartList : boolean;
begin
  result := (FPartList <> nil) and (FPartList.count > 0);
end;


{ TFhirOperationDefinitionParameterListEnumerator }

Constructor TFhirOperationDefinitionParameterListEnumerator.Create(list : TFhirOperationDefinitionParameterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationDefinitionParameterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationDefinitionParameterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationDefinitionParameterListEnumerator.GetCurrent : TFhirOperationDefinitionParameter;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationDefinitionParameterList }
procedure TFhirOperationDefinitionParameterList.AddItem(value: TFhirOperationDefinitionParameter);
begin
  assert(value.ClassName = 'TFhirOperationDefinitionParameter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationDefinitionParameter');
  add(value);
end;


function TFhirOperationDefinitionParameterList.Append: TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionParameterList.ClearItems;
begin
  Clear;
end;

function TFhirOperationDefinitionParameterList.GetEnumerator : TFhirOperationDefinitionParameterListEnumerator;
begin
  result := TFhirOperationDefinitionParameterListEnumerator.Create(self.link);
end;

function TFhirOperationDefinitionParameterList.Clone: TFhirOperationDefinitionParameterList;
begin
  result := TFhirOperationDefinitionParameterList(inherited Clone);
end;

function TFhirOperationDefinitionParameterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationDefinitionParameterList.GetItemN(index: Integer): TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionParameterList.IndexOf(value: TFhirOperationDefinitionParameter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationDefinitionParameterList.Insert(index: Integer): TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionParameterList.InsertItem(index: Integer; value: TFhirOperationDefinitionParameter);
begin
  assert(value is TFhirOperationDefinitionParameter);
  Inherited Insert(index, value);
end;

function TFhirOperationDefinitionParameterList.Item(index: Integer): TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionParameterList.Link: TFhirOperationDefinitionParameterList;
begin
  result := TFhirOperationDefinitionParameterList(inherited Link);
end;

procedure TFhirOperationDefinitionParameterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationDefinitionParameterList.SetItemByIndex(index: Integer; value: TFhirOperationDefinitionParameter);
begin
  assert(value is TFhirOperationDefinitionParameter);
  FhirOperationDefinitionParameters[index] := value;
end;

procedure TFhirOperationDefinitionParameterList.SetItemN(index: Integer; value: TFhirOperationDefinitionParameter);
begin
  assert(value is TFhirOperationDefinitionParameter);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationDefinitionParameterBinding }

constructor TFhirOperationDefinitionParameterBinding.Create;
begin
  inherited;
end;

destructor TFhirOperationDefinitionParameterBinding.Destroy;
begin
  FStrength.free;
  FValueSet.free;
  inherited;
end;

procedure TFhirOperationDefinitionParameterBinding.Assign(oSource : TAdvObject);
begin
  inherited;
  FStrength := TFhirOperationDefinitionParameterBinding(oSource).FStrength.Link;
  valueSet := TFhirOperationDefinitionParameterBinding(oSource).valueSet.Clone;
end;

procedure TFhirOperationDefinitionParameterBinding.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'strength') Then
     list.add(FStrength.Link);
  if (child_name = 'valueSet[x]') Then
     list.add(FValueSet.Link);
end;

procedure TFhirOperationDefinitionParameterBinding.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'strength', 'code', FStrength.Link));{1}
  oList.add(TFHIRProperty.create(self, 'valueSet[x]', 'uri|Reference(ValueSet)', FValueSet.Link));{2}
end;

procedure TFhirOperationDefinitionParameterBinding.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'strength') then StrengthElement := propValue as TFHIREnum
  else if (propName.startsWith('valueSet')) then ValueSet := propValue as TFhirType{4}
  else inherited;
end;

function TFhirOperationDefinitionParameterBinding.FhirType : string;
begin
  result := 'binding';
end;

function TFhirOperationDefinitionParameterBinding.Link : TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding(inherited Link);
end;

function TFhirOperationDefinitionParameterBinding.Clone : TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding(inherited Clone);
end;

function TFhirOperationDefinitionParameterBinding.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionParameterBinding;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionParameterBinding)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionParameterBinding(other);
    result := compareDeep(strengthElement, o.strengthElement, true) and compareDeep(valueSetElement, o.valueSetElement, true);
  end;
end;

function TFhirOperationDefinitionParameterBinding.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationDefinitionParameterBinding;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOperationDefinitionParameterBinding)) then
    result := false
  else
  begin
    o := TFhirOperationDefinitionParameterBinding(other);
    result := compareValues(strengthElement, o.strengthElement, true);
  end;
end;

{ TFhirOperationDefinitionParameterBinding }

Procedure TFhirOperationDefinitionParameterBinding.SetStrength(value : TFhirEnum);
begin
  FStrength.free;
  FStrength := value;
end;

Function TFhirOperationDefinitionParameterBinding.GetStrengthST : TFhirBindingStrength;
begin
  if FStrength = nil then
    result := TFhirBindingStrength(0)
  else
    result := TFhirBindingStrength(StringArrayIndexOfSensitive(CODES_TFhirBindingStrength, FStrength.value));
end;

Procedure TFhirOperationDefinitionParameterBinding.SetStrengthST(value : TFhirBindingStrength);
begin
  if ord(value) = 0 then
    StrengthElement := nil
  else
    StrengthElement := TFhirEnum.create(CODES_TFhirBindingStrength[value]);
end;

Procedure TFhirOperationDefinitionParameterBinding.SetValueSet(value : TFhirType);
begin
  FValueSet.free;
  FValueSet := value;
end;


{ TFhirOperationDefinitionParameterBindingListEnumerator }

Constructor TFhirOperationDefinitionParameterBindingListEnumerator.Create(list : TFhirOperationDefinitionParameterBindingList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationDefinitionParameterBindingListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationDefinitionParameterBindingListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationDefinitionParameterBindingListEnumerator.GetCurrent : TFhirOperationDefinitionParameterBinding;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationDefinitionParameterBindingList }
procedure TFhirOperationDefinitionParameterBindingList.AddItem(value: TFhirOperationDefinitionParameterBinding);
begin
  assert(value.ClassName = 'TFhirOperationDefinitionParameterBinding', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationDefinitionParameterBinding');
  add(value);
end;


function TFhirOperationDefinitionParameterBindingList.Append: TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionParameterBindingList.ClearItems;
begin
  Clear;
end;

function TFhirOperationDefinitionParameterBindingList.GetEnumerator : TFhirOperationDefinitionParameterBindingListEnumerator;
begin
  result := TFhirOperationDefinitionParameterBindingListEnumerator.Create(self.link);
end;

function TFhirOperationDefinitionParameterBindingList.Clone: TFhirOperationDefinitionParameterBindingList;
begin
  result := TFhirOperationDefinitionParameterBindingList(inherited Clone);
end;

function TFhirOperationDefinitionParameterBindingList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationDefinitionParameterBindingList.GetItemN(index: Integer): TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionParameterBindingList.IndexOf(value: TFhirOperationDefinitionParameterBinding): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationDefinitionParameterBindingList.Insert(index: Integer): TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionParameterBindingList.InsertItem(index: Integer; value: TFhirOperationDefinitionParameterBinding);
begin
  assert(value is TFhirOperationDefinitionParameterBinding);
  Inherited Insert(index, value);
end;

function TFhirOperationDefinitionParameterBindingList.Item(index: Integer): TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionParameterBindingList.Link: TFhirOperationDefinitionParameterBindingList;
begin
  result := TFhirOperationDefinitionParameterBindingList(inherited Link);
end;

procedure TFhirOperationDefinitionParameterBindingList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationDefinitionParameterBindingList.SetItemByIndex(index: Integer; value: TFhirOperationDefinitionParameterBinding);
begin
  assert(value is TFhirOperationDefinitionParameterBinding);
  FhirOperationDefinitionParameterBindings[index] := value;
end;

procedure TFhirOperationDefinitionParameterBindingList.SetItemN(index: Integer; value: TFhirOperationDefinitionParameterBinding);
begin
  assert(value is TFhirOperationDefinitionParameterBinding);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationDefinition }

constructor TFhirOperationDefinition.Create;
begin
  inherited;
end;

destructor TFhirOperationDefinition.Destroy;
begin
  FUrl.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FKind.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FRequirements.free;
  FIdempotent.free;
  FCode.free;
  FNotes.free;
  FBase.free;
  FSystem.free;
  FType_List.Free;
  FInstance.free;
  FParameterList.Free;
  inherited;
end;

function TFhirOperationDefinition.GetResourceType : TFhirResourceType;
begin
  result := frtOperationDefinition;
end;

procedure TFhirOperationDefinition.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirOperationDefinition(oSource).urlElement.Clone;
  versionElement := TFhirOperationDefinition(oSource).versionElement.Clone;
  nameElement := TFhirOperationDefinition(oSource).nameElement.Clone;
  FStatus := TFhirOperationDefinition(oSource).FStatus.Link;
  FKind := TFhirOperationDefinition(oSource).FKind.Link;
  experimentalElement := TFhirOperationDefinition(oSource).experimentalElement.Clone;
  publisherElement := TFhirOperationDefinition(oSource).publisherElement.Clone;
  if (TFhirOperationDefinition(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirOperationDefinitionContactList.Create;
    FContactList.Assign(TFhirOperationDefinition(oSource).FContactList);
  end;
  dateElement := TFhirOperationDefinition(oSource).dateElement.Clone;
  descriptionElement := TFhirOperationDefinition(oSource).descriptionElement.Clone;
  requirementsElement := TFhirOperationDefinition(oSource).requirementsElement.Clone;
  idempotentElement := TFhirOperationDefinition(oSource).idempotentElement.Clone;
  codeElement := TFhirOperationDefinition(oSource).codeElement.Clone;
  notesElement := TFhirOperationDefinition(oSource).notesElement.Clone;
  base := TFhirOperationDefinition(oSource).base.Clone;
  systemElement := TFhirOperationDefinition(oSource).systemElement.Clone;
  if (TFhirOperationDefinition(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeList.Create;
    FType_List.Assign(TFhirOperationDefinition(oSource).FType_List);
  end;
  instanceElement := TFhirOperationDefinition(oSource).instanceElement.Clone;
  if (TFhirOperationDefinition(oSource).FParameterList = nil) then
  begin
    FParameterList.free;
    FParameterList := nil;
  end
  else
  begin
    FParameterList := TFhirOperationDefinitionParameterList.Create;
    FParameterList.Assign(TFhirOperationDefinition(oSource).FParameterList);
  end;
end;

procedure TFhirOperationDefinition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'idempotent') Then
     list.add(FIdempotent.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'notes') Then
     list.add(FNotes.Link);
  if (child_name = 'base') Then
     list.add(FBase.Link);
  if (child_name = 'system') Then
     list.add(FSystem.Link);
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'instance') Then
     list.add(FInstance.Link);
  if (child_name = 'parameter') Then
    list.addAll(FParameterList);
end;

procedure TFhirOperationDefinition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'kind', 'code', FKind.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'idempotent', 'boolean', FIdempotent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'notes', 'string', FNotes.Link));{2}
  oList.add(TFHIRProperty.create(self, 'base', 'Reference(OperationDefinition)', FBase.Link));{2}
  oList.add(TFHIRProperty.create(self, 'system', 'boolean', FSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'instance', 'boolean', FInstance.Link));{2}
  oList.add(TFHIRProperty.create(self, 'parameter', '', FParameterList.Link)){3};
end;

procedure TFhirOperationDefinition.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'kind') then KindElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirOperationDefinitionContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'idempotent') then IdempotentElement := propValue as TFhirBoolean{5a}
  else if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'notes') then NotesElement := propValue as TFhirString{5a}
  else if (propName = 'base') then Base := propValue as TFhirReference{TFhirOperationDefinition}{4b}
  else if (propName = 'system') then SystemElement := propValue as TFhirBoolean{5a}
  else if (propName = 'type') then Type_List.add(propValue as TFhirCode){2}
  else if (propName = 'instance') then InstanceElement := propValue as TFhirBoolean{5a}
  else if (propName = 'parameter') then ParameterList.add(propValue as TFhirOperationDefinitionParameter){2}
  else inherited;
end;

function TFhirOperationDefinition.FhirType : string;
begin
  result := 'OperationDefinition';
end;

function TFhirOperationDefinition.Link : TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition(inherited Link);
end;

function TFhirOperationDefinition.Clone : TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition(inherited Clone);
end;

{ TFhirOperationDefinition }

Procedure TFhirOperationDefinition.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirOperationDefinition.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirOperationDefinition.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirOperationDefinition.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirOperationDefinition.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirOperationDefinition.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirOperationDefinition.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirOperationDefinition.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirOperationDefinition.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirOperationDefinition.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirOperationDefinition.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirOperationDefinition.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirOperationDefinition.SetKind(value : TFhirEnum);
begin
  FKind.free;
  FKind := value;
end;

Function TFhirOperationDefinition.GetKindST : TFhirOperationKind;
begin
  if FKind = nil then
    result := TFhirOperationKind(0)
  else
    result := TFhirOperationKind(StringArrayIndexOfSensitive(CODES_TFhirOperationKind, FKind.value));
end;

Procedure TFhirOperationDefinition.SetKindST(value : TFhirOperationKind);
begin
  if ord(value) = 0 then
    KindElement := nil
  else
    KindElement := TFhirEnum.create(CODES_TFhirOperationKind[value]);
end;

Procedure TFhirOperationDefinition.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirOperationDefinition.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirOperationDefinition.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirOperationDefinition.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirOperationDefinition.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirOperationDefinition.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirOperationDefinition.GetContactList : TFhirOperationDefinitionContactList;
begin
  if FContactList = nil then
    FContactList := TFhirOperationDefinitionContactList.Create;
  result := FContactList;
end;

Function TFhirOperationDefinition.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirOperationDefinition.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirOperationDefinition.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirOperationDefinition.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirOperationDefinition.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirOperationDefinition.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirOperationDefinition.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirOperationDefinition.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirOperationDefinition.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirOperationDefinition.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirOperationDefinition.SetIdempotent(value : TFhirBoolean);
begin
  FIdempotent.free;
  FIdempotent := value;
end;

Function TFhirOperationDefinition.GetIdempotentST : Boolean;
begin
  if FIdempotent = nil then
    result := false
  else
    result := FIdempotent.value;
end;

Procedure TFhirOperationDefinition.SetIdempotentST(value : Boolean);
begin
  if FIdempotent = nil then
    FIdempotent := TFhirBoolean.create;
  FIdempotent.value := value
end;

Procedure TFhirOperationDefinition.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirOperationDefinition.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirOperationDefinition.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirOperationDefinition.SetNotes(value : TFhirString);
begin
  FNotes.free;
  FNotes := value;
end;

Function TFhirOperationDefinition.GetNotesST : String;
begin
  if FNotes = nil then
    result := ''
  else
    result := FNotes.value;
end;

Procedure TFhirOperationDefinition.SetNotesST(value : String);
begin
  if value <> '' then
  begin
    if FNotes = nil then
      FNotes := TFhirString.create;
    FNotes.value := value
  end
  else if FNotes <> nil then
    FNotes.value := '';
end;

Procedure TFhirOperationDefinition.SetBase(value : TFhirReference{TFhirOperationDefinition});
begin
  FBase.free;
  FBase := value;
end;

Procedure TFhirOperationDefinition.SetSystem(value : TFhirBoolean);
begin
  FSystem.free;
  FSystem := value;
end;

Function TFhirOperationDefinition.GetSystemST : Boolean;
begin
  if FSystem = nil then
    result := false
  else
    result := FSystem.value;
end;

Procedure TFhirOperationDefinition.SetSystemST(value : Boolean);
begin
  if FSystem = nil then
    FSystem := TFhirBoolean.create;
  FSystem.value := value
end;

Function TFhirOperationDefinition.GetType_List : TFhirCodeList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeList.Create;
  result := FType_List;
end;

Function TFhirOperationDefinition.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Procedure TFhirOperationDefinition.SetInstance(value : TFhirBoolean);
begin
  FInstance.free;
  FInstance := value;
end;

Function TFhirOperationDefinition.GetInstanceST : Boolean;
begin
  if FInstance = nil then
    result := false
  else
    result := FInstance.value;
end;

Procedure TFhirOperationDefinition.SetInstanceST(value : Boolean);
begin
  if FInstance = nil then
    FInstance := TFhirBoolean.create;
  FInstance.value := value
end;

Function TFhirOperationDefinition.GetParameterList : TFhirOperationDefinitionParameterList;
begin
  if FParameterList = nil then
    FParameterList := TFhirOperationDefinitionParameterList.Create;
  result := FParameterList;
end;

Function TFhirOperationDefinition.GetHasParameterList : boolean;
begin
  result := (FParameterList <> nil) and (FParameterList.count > 0);
end;


{ TFhirOperationDefinitionListEnumerator }

Constructor TFhirOperationDefinitionListEnumerator.Create(list : TFhirOperationDefinitionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationDefinitionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationDefinitionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationDefinitionListEnumerator.GetCurrent : TFhirOperationDefinition;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationDefinitionList }
procedure TFhirOperationDefinitionList.AddItem(value: TFhirOperationDefinition);
begin
  assert(value.ClassName = 'TFhirOperationDefinition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationDefinition');
  add(value);
end;


function TFhirOperationDefinitionList.Append: TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionList.ClearItems;
begin
  Clear;
end;

function TFhirOperationDefinitionList.GetEnumerator : TFhirOperationDefinitionListEnumerator;
begin
  result := TFhirOperationDefinitionListEnumerator.Create(self.link);
end;

function TFhirOperationDefinitionList.Clone: TFhirOperationDefinitionList;
begin
  result := TFhirOperationDefinitionList(inherited Clone);
end;

function TFhirOperationDefinitionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationDefinitionList.GetItemN(index: Integer): TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionList.IndexOf(value: TFhirOperationDefinition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationDefinitionList.Insert(index: Integer): TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationDefinitionList.InsertItem(index: Integer; value: TFhirOperationDefinition);
begin
  assert(value is TFhirOperationDefinition);
  Inherited Insert(index, value);
end;

function TFhirOperationDefinitionList.Item(index: Integer): TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition(ObjectByIndex[index]);
end;

function TFhirOperationDefinitionList.Link: TFhirOperationDefinitionList;
begin
  result := TFhirOperationDefinitionList(inherited Link);
end;

procedure TFhirOperationDefinitionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationDefinitionList.SetItemByIndex(index: Integer; value: TFhirOperationDefinition);
begin
  assert(value is TFhirOperationDefinition);
  FhirOperationDefinitions[index] := value;
end;

procedure TFhirOperationDefinitionList.SetItemN(index: Integer; value: TFhirOperationDefinition);
begin
  assert(value is TFhirOperationDefinition);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationOutcomeIssue }

constructor TFhirOperationOutcomeIssue.Create;
begin
  inherited;
end;

destructor TFhirOperationOutcomeIssue.Destroy;
begin
  FSeverity.free;
  FCode.free;
  FDetails.free;
  FDiagnostics.free;
  FLocationList.Free;
  inherited;
end;

procedure TFhirOperationOutcomeIssue.Assign(oSource : TAdvObject);
begin
  inherited;
  FSeverity := TFhirOperationOutcomeIssue(oSource).FSeverity.Link;
  FCode := TFhirOperationOutcomeIssue(oSource).FCode.Link;
  details := TFhirOperationOutcomeIssue(oSource).details.Clone;
  diagnosticsElement := TFhirOperationOutcomeIssue(oSource).diagnosticsElement.Clone;
  if (TFhirOperationOutcomeIssue(oSource).FLocationList = nil) then
  begin
    FLocationList.free;
    FLocationList := nil;
  end
  else
  begin
    FLocationList := TFhirStringList.Create;
    FLocationList.Assign(TFhirOperationOutcomeIssue(oSource).FLocationList);
  end;
end;

procedure TFhirOperationOutcomeIssue.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'severity') Then
     list.add(FSeverity.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'details') Then
     list.add(FDetails.Link);
  if (child_name = 'diagnostics') Then
     list.add(FDiagnostics.Link);
  if (child_name = 'location') Then
    list.addAll(FLocationList);
end;

procedure TFhirOperationOutcomeIssue.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'severity', 'code', FSeverity.Link));{1}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{1}
  oList.add(TFHIRProperty.create(self, 'details', 'CodeableConcept', FDetails.Link));{2}
  oList.add(TFHIRProperty.create(self, 'diagnostics', 'string', FDiagnostics.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'string', FLocationList.Link)){3};
end;

procedure TFhirOperationOutcomeIssue.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'severity') then SeverityElement := propValue as TFHIREnum
  else if (propName = 'code') then CodeElement := propValue as TFHIREnum
  else if (propName = 'details') then Details := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'diagnostics') then DiagnosticsElement := propValue as TFhirString{5a}
  else if (propName = 'location') then LocationList.add(propValue as TFhirString){2}
  else inherited;
end;

function TFhirOperationOutcomeIssue.FhirType : string;
begin
  result := 'issue';
end;

function TFhirOperationOutcomeIssue.Link : TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue(inherited Link);
end;

function TFhirOperationOutcomeIssue.Clone : TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue(inherited Clone);
end;

function TFhirOperationOutcomeIssue.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationOutcomeIssue;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOperationOutcomeIssue)) then
    result := false
  else
  begin
    o := TFhirOperationOutcomeIssue(other);
    result := compareDeep(severityElement, o.severityElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(detailsElement, o.detailsElement, true) and compareDeep(diagnosticsElement, o.diagnosticsElement, true) and 
      compareDeep(locationList, o.locationList, true);
  end;
end;

function TFhirOperationOutcomeIssue.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOperationOutcomeIssue;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOperationOutcomeIssue)) then
    result := false
  else
  begin
    o := TFhirOperationOutcomeIssue(other);
    result := compareValues(severityElement, o.severityElement, true) and compareValues(codeElement, o.codeElement, true) and 
      compareValues(diagnosticsElement, o.diagnosticsElement, true) and compareValues(locationList, o.locationList, true);
  end;
end;

{ TFhirOperationOutcomeIssue }

Procedure TFhirOperationOutcomeIssue.SetSeverity(value : TFhirEnum);
begin
  FSeverity.free;
  FSeverity := value;
end;

Function TFhirOperationOutcomeIssue.GetSeverityST : TFhirIssueSeverity;
begin
  if FSeverity = nil then
    result := TFhirIssueSeverity(0)
  else
    result := TFhirIssueSeverity(StringArrayIndexOfSensitive(CODES_TFhirIssueSeverity, FSeverity.value));
end;

Procedure TFhirOperationOutcomeIssue.SetSeverityST(value : TFhirIssueSeverity);
begin
  if ord(value) = 0 then
    SeverityElement := nil
  else
    SeverityElement := TFhirEnum.create(CODES_TFhirIssueSeverity[value]);
end;

Procedure TFhirOperationOutcomeIssue.SetCode(value : TFhirEnum);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirOperationOutcomeIssue.GetCodeST : TFhirIssueType;
begin
  if FCode = nil then
    result := TFhirIssueType(0)
  else
    result := TFhirIssueType(StringArrayIndexOfSensitive(CODES_TFhirIssueType, FCode.value));
end;

Procedure TFhirOperationOutcomeIssue.SetCodeST(value : TFhirIssueType);
begin
  if ord(value) = 0 then
    CodeElement := nil
  else
    CodeElement := TFhirEnum.create(CODES_TFhirIssueType[value]);
end;

Procedure TFhirOperationOutcomeIssue.SetDetails(value : TFhirCodeableConcept);
begin
  FDetails.free;
  FDetails := value;
end;

Procedure TFhirOperationOutcomeIssue.SetDiagnostics(value : TFhirString);
begin
  FDiagnostics.free;
  FDiagnostics := value;
end;

Function TFhirOperationOutcomeIssue.GetDiagnosticsST : String;
begin
  if FDiagnostics = nil then
    result := ''
  else
    result := FDiagnostics.value;
end;

Procedure TFhirOperationOutcomeIssue.SetDiagnosticsST(value : String);
begin
  if value <> '' then
  begin
    if FDiagnostics = nil then
      FDiagnostics := TFhirString.create;
    FDiagnostics.value := value
  end
  else if FDiagnostics <> nil then
    FDiagnostics.value := '';
end;

Function TFhirOperationOutcomeIssue.GetLocationList : TFhirStringList;
begin
  if FLocationList = nil then
    FLocationList := TFhirStringList.Create;
  result := FLocationList;
end;

Function TFhirOperationOutcomeIssue.GetHasLocationList : boolean;
begin
  result := (FLocationList <> nil) and (FLocationList.count > 0);
end;


{ TFhirOperationOutcomeIssueListEnumerator }

Constructor TFhirOperationOutcomeIssueListEnumerator.Create(list : TFhirOperationOutcomeIssueList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationOutcomeIssueListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationOutcomeIssueListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationOutcomeIssueListEnumerator.GetCurrent : TFhirOperationOutcomeIssue;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationOutcomeIssueList }
procedure TFhirOperationOutcomeIssueList.AddItem(value: TFhirOperationOutcomeIssue);
begin
  assert(value.ClassName = 'TFhirOperationOutcomeIssue', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationOutcomeIssue');
  add(value);
end;


function TFhirOperationOutcomeIssueList.Append: TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationOutcomeIssueList.ClearItems;
begin
  Clear;
end;

function TFhirOperationOutcomeIssueList.GetEnumerator : TFhirOperationOutcomeIssueListEnumerator;
begin
  result := TFhirOperationOutcomeIssueListEnumerator.Create(self.link);
end;

function TFhirOperationOutcomeIssueList.Clone: TFhirOperationOutcomeIssueList;
begin
  result := TFhirOperationOutcomeIssueList(inherited Clone);
end;

function TFhirOperationOutcomeIssueList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationOutcomeIssueList.GetItemN(index: Integer): TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue(ObjectByIndex[index]);
end;

function TFhirOperationOutcomeIssueList.IndexOf(value: TFhirOperationOutcomeIssue): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationOutcomeIssueList.Insert(index: Integer): TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationOutcomeIssueList.InsertItem(index: Integer; value: TFhirOperationOutcomeIssue);
begin
  assert(value is TFhirOperationOutcomeIssue);
  Inherited Insert(index, value);
end;

function TFhirOperationOutcomeIssueList.Item(index: Integer): TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue(ObjectByIndex[index]);
end;

function TFhirOperationOutcomeIssueList.Link: TFhirOperationOutcomeIssueList;
begin
  result := TFhirOperationOutcomeIssueList(inherited Link);
end;

procedure TFhirOperationOutcomeIssueList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationOutcomeIssueList.SetItemByIndex(index: Integer; value: TFhirOperationOutcomeIssue);
begin
  assert(value is TFhirOperationOutcomeIssue);
  FhirOperationOutcomeIssues[index] := value;
end;

procedure TFhirOperationOutcomeIssueList.SetItemN(index: Integer; value: TFhirOperationOutcomeIssue);
begin
  assert(value is TFhirOperationOutcomeIssue);
  ObjectByIndex[index] := value;
end;

{ TFhirOperationOutcome }

constructor TFhirOperationOutcome.Create;
begin
  inherited;
end;

destructor TFhirOperationOutcome.Destroy;
begin
  FIssueList.Free;
  inherited;
end;

function TFhirOperationOutcome.GetResourceType : TFhirResourceType;
begin
  result := frtOperationOutcome;
end;

procedure TFhirOperationOutcome.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirOperationOutcome(oSource).FIssueList = nil) then
  begin
    FIssueList.free;
    FIssueList := nil;
  end
  else
  begin
    FIssueList := TFhirOperationOutcomeIssueList.Create;
    FIssueList.Assign(TFhirOperationOutcome(oSource).FIssueList);
  end;
end;

procedure TFhirOperationOutcome.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'issue') Then
    list.addAll(FIssueList);
end;

procedure TFhirOperationOutcome.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'issue', '', FIssueList.Link)){3};
end;

procedure TFhirOperationOutcome.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'issue') then IssueList.add(propValue as TFhirOperationOutcomeIssue){2}
  else inherited;
end;

function TFhirOperationOutcome.FhirType : string;
begin
  result := 'OperationOutcome';
end;

function TFhirOperationOutcome.Link : TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome(inherited Link);
end;

function TFhirOperationOutcome.Clone : TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome(inherited Clone);
end;

{ TFhirOperationOutcome }

Function TFhirOperationOutcome.GetIssueList : TFhirOperationOutcomeIssueList;
begin
  if FIssueList = nil then
    FIssueList := TFhirOperationOutcomeIssueList.Create;
  result := FIssueList;
end;

Function TFhirOperationOutcome.GetHasIssueList : boolean;
begin
  result := (FIssueList <> nil) and (FIssueList.count > 0);
end;


{ TFhirOperationOutcomeListEnumerator }

Constructor TFhirOperationOutcomeListEnumerator.Create(list : TFhirOperationOutcomeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOperationOutcomeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOperationOutcomeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOperationOutcomeListEnumerator.GetCurrent : TFhirOperationOutcome;
begin
  Result := FList[FIndex];
end;


{ TFhirOperationOutcomeList }
procedure TFhirOperationOutcomeList.AddItem(value: TFhirOperationOutcome);
begin
  assert(value.ClassName = 'TFhirOperationOutcome', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOperationOutcome');
  add(value);
end;


function TFhirOperationOutcomeList.Append: TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationOutcomeList.ClearItems;
begin
  Clear;
end;

function TFhirOperationOutcomeList.GetEnumerator : TFhirOperationOutcomeListEnumerator;
begin
  result := TFhirOperationOutcomeListEnumerator.Create(self.link);
end;

function TFhirOperationOutcomeList.Clone: TFhirOperationOutcomeList;
begin
  result := TFhirOperationOutcomeList(inherited Clone);
end;

function TFhirOperationOutcomeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOperationOutcomeList.GetItemN(index: Integer): TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome(ObjectByIndex[index]);
end;

function TFhirOperationOutcomeList.IndexOf(value: TFhirOperationOutcome): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOperationOutcomeList.Insert(index: Integer): TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOperationOutcomeList.InsertItem(index: Integer; value: TFhirOperationOutcome);
begin
  assert(value is TFhirOperationOutcome);
  Inherited Insert(index, value);
end;

function TFhirOperationOutcomeList.Item(index: Integer): TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome(ObjectByIndex[index]);
end;

function TFhirOperationOutcomeList.Link: TFhirOperationOutcomeList;
begin
  result := TFhirOperationOutcomeList(inherited Link);
end;

procedure TFhirOperationOutcomeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOperationOutcomeList.SetItemByIndex(index: Integer; value: TFhirOperationOutcome);
begin
  assert(value is TFhirOperationOutcome);
  FhirOperationOutcomes[index] := value;
end;

procedure TFhirOperationOutcomeList.SetItemN(index: Integer; value: TFhirOperationOutcome);
begin
  assert(value is TFhirOperationOutcome);
  ObjectByIndex[index] := value;
end;

{ TFhirOrderWhen }

constructor TFhirOrderWhen.Create;
begin
  inherited;
end;

destructor TFhirOrderWhen.Destroy;
begin
  FCode.free;
  FSchedule.free;
  inherited;
end;

procedure TFhirOrderWhen.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirOrderWhen(oSource).code.Clone;
  schedule := TFhirOrderWhen(oSource).schedule.Clone;
end;

procedure TFhirOrderWhen.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'schedule') Then
     list.add(FSchedule.Link);
end;

procedure TFhirOrderWhen.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'schedule', 'Timing', FSchedule.Link));{2}
end;

procedure TFhirOrderWhen.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'schedule') then Schedule := propValue as TFhirTiming{4b}
  else inherited;
end;

function TFhirOrderWhen.FhirType : string;
begin
  result := 'when';
end;

function TFhirOrderWhen.Link : TFhirOrderWhen;
begin
  result := TFhirOrderWhen(inherited Link);
end;

function TFhirOrderWhen.Clone : TFhirOrderWhen;
begin
  result := TFhirOrderWhen(inherited Clone);
end;

function TFhirOrderWhen.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOrderWhen;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOrderWhen)) then
    result := false
  else
  begin
    o := TFhirOrderWhen(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(scheduleElement, o.scheduleElement, true);
  end;
end;

function TFhirOrderWhen.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOrderWhen;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOrderWhen)) then
    result := false
  else
  begin
    o := TFhirOrderWhen(other);
    result := true;
  end;
end;

{ TFhirOrderWhen }

Procedure TFhirOrderWhen.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirOrderWhen.SetSchedule(value : TFhirTiming);
begin
  FSchedule.free;
  FSchedule := value;
end;


{ TFhirOrderWhenListEnumerator }

Constructor TFhirOrderWhenListEnumerator.Create(list : TFhirOrderWhenList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOrderWhenListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOrderWhenListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOrderWhenListEnumerator.GetCurrent : TFhirOrderWhen;
begin
  Result := FList[FIndex];
end;


{ TFhirOrderWhenList }
procedure TFhirOrderWhenList.AddItem(value: TFhirOrderWhen);
begin
  assert(value.ClassName = 'TFhirOrderWhen', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOrderWhen');
  add(value);
end;


function TFhirOrderWhenList.Append: TFhirOrderWhen;
begin
  result := TFhirOrderWhen.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderWhenList.ClearItems;
begin
  Clear;
end;

function TFhirOrderWhenList.GetEnumerator : TFhirOrderWhenListEnumerator;
begin
  result := TFhirOrderWhenListEnumerator.Create(self.link);
end;

function TFhirOrderWhenList.Clone: TFhirOrderWhenList;
begin
  result := TFhirOrderWhenList(inherited Clone);
end;

function TFhirOrderWhenList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOrderWhenList.GetItemN(index: Integer): TFhirOrderWhen;
begin
  result := TFhirOrderWhen(ObjectByIndex[index]);
end;

function TFhirOrderWhenList.IndexOf(value: TFhirOrderWhen): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOrderWhenList.Insert(index: Integer): TFhirOrderWhen;
begin
  result := TFhirOrderWhen.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderWhenList.InsertItem(index: Integer; value: TFhirOrderWhen);
begin
  assert(value is TFhirOrderWhen);
  Inherited Insert(index, value);
end;

function TFhirOrderWhenList.Item(index: Integer): TFhirOrderWhen;
begin
  result := TFhirOrderWhen(ObjectByIndex[index]);
end;

function TFhirOrderWhenList.Link: TFhirOrderWhenList;
begin
  result := TFhirOrderWhenList(inherited Link);
end;

procedure TFhirOrderWhenList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOrderWhenList.SetItemByIndex(index: Integer; value: TFhirOrderWhen);
begin
  assert(value is TFhirOrderWhen);
  FhirOrderWhens[index] := value;
end;

procedure TFhirOrderWhenList.SetItemN(index: Integer; value: TFhirOrderWhen);
begin
  assert(value is TFhirOrderWhen);
  ObjectByIndex[index] := value;
end;

{ TFhirOrder }

constructor TFhirOrder.Create;
begin
  inherited;
end;

destructor TFhirOrder.Destroy;
begin
  FIdentifierList.Free;
  FDate.free;
  FSubject.free;
  FSource.free;
  FTarget.free;
  FReason.free;
  FWhen.free;
  FDetailList.Free;
  inherited;
end;

function TFhirOrder.GetResourceType : TFhirResourceType;
begin
  result := frtOrder;
end;

procedure TFhirOrder.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirOrder(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirOrder(oSource).FIdentifierList);
  end;
  dateElement := TFhirOrder(oSource).dateElement.Clone;
  subject := TFhirOrder(oSource).subject.Clone;
  source := TFhirOrder(oSource).source.Clone;
  target := TFhirOrder(oSource).target.Clone;
  reason := TFhirOrder(oSource).reason.Clone;
  when := TFhirOrder(oSource).when.Clone;
  if (TFhirOrder(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
    FDetailList.Assign(TFhirOrder(oSource).FDetailList);
  end;
end;

procedure TFhirOrder.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'when') Then
     list.add(FWhen.Link);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
end;

procedure TFhirOrder.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Device|Substance)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Practitioner|Organization)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization|Device|Practitioner)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Any)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'when', '', FWhen.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', 'Reference(Any)', FDetailList.Link)){3};
end;

procedure TFhirOrder.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'source') then Source := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'target') then Target := propValue as TFhirReference{Resource}{4b}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName = 'when') then When := propValue as TFhirOrderWhen{4b}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirReference{TFhirReference}){2}
  else inherited;
end;

function TFhirOrder.FhirType : string;
begin
  result := 'Order';
end;

function TFhirOrder.Link : TFhirOrder;
begin
  result := TFhirOrder(inherited Link);
end;

function TFhirOrder.Clone : TFhirOrder;
begin
  result := TFhirOrder(inherited Clone);
end;

{ TFhirOrder }

Function TFhirOrder.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirOrder.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirOrder.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirOrder.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirOrder.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirOrder.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirOrder.SetSource(value : TFhirReference{Resource});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirOrder.SetTarget(value : TFhirReference{Resource});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirOrder.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirOrder.SetWhen(value : TFhirOrderWhen);
begin
  FWhen.free;
  FWhen := value;
end;

Function TFhirOrder.GetDetailList : TFhirReferenceList{TFhirReference};
begin
  if FDetailList = nil then
    FDetailList := TFhirReferenceList{TFhirReference}.Create;
  result := FDetailList;
end;

Function TFhirOrder.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;


{ TFhirOrderListEnumerator }

Constructor TFhirOrderListEnumerator.Create(list : TFhirOrderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOrderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOrderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOrderListEnumerator.GetCurrent : TFhirOrder;
begin
  Result := FList[FIndex];
end;


{ TFhirOrderList }
procedure TFhirOrderList.AddItem(value: TFhirOrder);
begin
  assert(value.ClassName = 'TFhirOrder', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOrder');
  add(value);
end;


function TFhirOrderList.Append: TFhirOrder;
begin
  result := TFhirOrder.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderList.ClearItems;
begin
  Clear;
end;

function TFhirOrderList.GetEnumerator : TFhirOrderListEnumerator;
begin
  result := TFhirOrderListEnumerator.Create(self.link);
end;

function TFhirOrderList.Clone: TFhirOrderList;
begin
  result := TFhirOrderList(inherited Clone);
end;

function TFhirOrderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOrderList.GetItemN(index: Integer): TFhirOrder;
begin
  result := TFhirOrder(ObjectByIndex[index]);
end;

function TFhirOrderList.IndexOf(value: TFhirOrder): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOrderList.Insert(index: Integer): TFhirOrder;
begin
  result := TFhirOrder.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderList.InsertItem(index: Integer; value: TFhirOrder);
begin
  assert(value is TFhirOrder);
  Inherited Insert(index, value);
end;

function TFhirOrderList.Item(index: Integer): TFhirOrder;
begin
  result := TFhirOrder(ObjectByIndex[index]);
end;

function TFhirOrderList.Link: TFhirOrderList;
begin
  result := TFhirOrderList(inherited Link);
end;

procedure TFhirOrderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOrderList.SetItemByIndex(index: Integer; value: TFhirOrder);
begin
  assert(value is TFhirOrder);
  FhirOrders[index] := value;
end;

procedure TFhirOrderList.SetItemN(index: Integer; value: TFhirOrder);
begin
  assert(value is TFhirOrder);
  ObjectByIndex[index] := value;
end;

{ TFhirOrderResponse }

constructor TFhirOrderResponse.Create;
begin
  inherited;
end;

destructor TFhirOrderResponse.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FDate.free;
  FWho.free;
  FOrderStatus.free;
  FDescription.free;
  FFulfillmentList.Free;
  inherited;
end;

function TFhirOrderResponse.GetResourceType : TFhirResourceType;
begin
  result := frtOrderResponse;
end;

procedure TFhirOrderResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirOrderResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirOrderResponse(oSource).FIdentifierList);
  end;
  request := TFhirOrderResponse(oSource).request.Clone;
  dateElement := TFhirOrderResponse(oSource).dateElement.Clone;
  who := TFhirOrderResponse(oSource).who.Clone;
  FOrderStatus := TFhirOrderResponse(oSource).FOrderStatus.Link;
  descriptionElement := TFhirOrderResponse(oSource).descriptionElement.Clone;
  if (TFhirOrderResponse(oSource).FFulfillmentList = nil) then
  begin
    FFulfillmentList.free;
    FFulfillmentList := nil;
  end
  else
  begin
    FFulfillmentList := TFhirReferenceList{TFhirReference}.Create;
    FFulfillmentList.Assign(TFhirOrderResponse(oSource).FFulfillmentList);
  end;
end;

procedure TFhirOrderResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'who') Then
     list.add(FWho.Link);
  if (child_name = 'orderStatus') Then
     list.add(FOrderStatus.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'fulfillment') Then
    list.addAll(FFulfillmentList);
end;

procedure TFhirOrderResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Order)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'who', 'Reference(Practitioner|Organization|Device)', FWho.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderStatus', 'code', FOrderStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fulfillment', 'Reference(Any)', FFulfillmentList.Link)){3};
end;

procedure TFhirOrderResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirOrder}{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'who') then Who := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'orderStatus') then OrderStatusElement := propValue as TFHIREnum
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'fulfillment') then FulfillmentList.add(propValue as TFhirReference{TFhirReference}){2}
  else inherited;
end;

function TFhirOrderResponse.FhirType : string;
begin
  result := 'OrderResponse';
end;

function TFhirOrderResponse.Link : TFhirOrderResponse;
begin
  result := TFhirOrderResponse(inherited Link);
end;

function TFhirOrderResponse.Clone : TFhirOrderResponse;
begin
  result := TFhirOrderResponse(inherited Clone);
end;

{ TFhirOrderResponse }

Function TFhirOrderResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirOrderResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirOrderResponse.SetRequest(value : TFhirReference{TFhirOrder});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirOrderResponse.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirOrderResponse.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirOrderResponse.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirOrderResponse.SetWho(value : TFhirReference{Resource});
begin
  FWho.free;
  FWho := value;
end;

Procedure TFhirOrderResponse.SetOrderStatus(value : TFhirEnum);
begin
  FOrderStatus.free;
  FOrderStatus := value;
end;

Function TFhirOrderResponse.GetOrderStatusST : TFhirOrderStatus;
begin
  if FOrderStatus = nil then
    result := TFhirOrderStatus(0)
  else
    result := TFhirOrderStatus(StringArrayIndexOfSensitive(CODES_TFhirOrderStatus, FOrderStatus.value));
end;

Procedure TFhirOrderResponse.SetOrderStatusST(value : TFhirOrderStatus);
begin
  if ord(value) = 0 then
    OrderStatusElement := nil
  else
    OrderStatusElement := TFhirEnum.create(CODES_TFhirOrderStatus[value]);
end;

Procedure TFhirOrderResponse.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirOrderResponse.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirOrderResponse.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirOrderResponse.GetFulfillmentList : TFhirReferenceList{TFhirReference};
begin
  if FFulfillmentList = nil then
    FFulfillmentList := TFhirReferenceList{TFhirReference}.Create;
  result := FFulfillmentList;
end;

Function TFhirOrderResponse.GetHasFulfillmentList : boolean;
begin
  result := (FFulfillmentList <> nil) and (FFulfillmentList.count > 0);
end;


{ TFhirOrderResponseListEnumerator }

Constructor TFhirOrderResponseListEnumerator.Create(list : TFhirOrderResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOrderResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOrderResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOrderResponseListEnumerator.GetCurrent : TFhirOrderResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirOrderResponseList }
procedure TFhirOrderResponseList.AddItem(value: TFhirOrderResponse);
begin
  assert(value.ClassName = 'TFhirOrderResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOrderResponse');
  add(value);
end;


function TFhirOrderResponseList.Append: TFhirOrderResponse;
begin
  result := TFhirOrderResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderResponseList.ClearItems;
begin
  Clear;
end;

function TFhirOrderResponseList.GetEnumerator : TFhirOrderResponseListEnumerator;
begin
  result := TFhirOrderResponseListEnumerator.Create(self.link);
end;

function TFhirOrderResponseList.Clone: TFhirOrderResponseList;
begin
  result := TFhirOrderResponseList(inherited Clone);
end;

function TFhirOrderResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOrderResponseList.GetItemN(index: Integer): TFhirOrderResponse;
begin
  result := TFhirOrderResponse(ObjectByIndex[index]);
end;

function TFhirOrderResponseList.IndexOf(value: TFhirOrderResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOrderResponseList.Insert(index: Integer): TFhirOrderResponse;
begin
  result := TFhirOrderResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrderResponseList.InsertItem(index: Integer; value: TFhirOrderResponse);
begin
  assert(value is TFhirOrderResponse);
  Inherited Insert(index, value);
end;

function TFhirOrderResponseList.Item(index: Integer): TFhirOrderResponse;
begin
  result := TFhirOrderResponse(ObjectByIndex[index]);
end;

function TFhirOrderResponseList.Link: TFhirOrderResponseList;
begin
  result := TFhirOrderResponseList(inherited Link);
end;

procedure TFhirOrderResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOrderResponseList.SetItemByIndex(index: Integer; value: TFhirOrderResponse);
begin
  assert(value is TFhirOrderResponse);
  FhirOrderResponses[index] := value;
end;

procedure TFhirOrderResponseList.SetItemN(index: Integer; value: TFhirOrderResponse);
begin
  assert(value is TFhirOrderResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirOrganizationContact }

constructor TFhirOrganizationContact.Create;
begin
  inherited;
end;

destructor TFhirOrganizationContact.Destroy;
begin
  FPurpose.free;
  FName.free;
  FTelecomList.Free;
  FAddress.free;
  inherited;
end;

procedure TFhirOrganizationContact.Assign(oSource : TAdvObject);
begin
  inherited;
  purpose := TFhirOrganizationContact(oSource).purpose.Clone;
  name := TFhirOrganizationContact(oSource).name.Clone;
  if (TFhirOrganizationContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirOrganizationContact(oSource).FTelecomList);
  end;
  address := TFhirOrganizationContact(oSource).address.Clone;
end;

procedure TFhirOrganizationContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'purpose') Then
     list.add(FPurpose.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'address') Then
     list.add(FAddress.Link);
end;

procedure TFhirOrganizationContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'purpose', 'CodeableConcept', FPurpose.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddress.Link));{2}
end;

procedure TFhirOrganizationContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'purpose') then Purpose := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'name') then Name := propValue as TFhirHumanName{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'address') then Address := propValue as TFhirAddress{4b}
  else inherited;
end;

function TFhirOrganizationContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirOrganizationContact.Link : TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact(inherited Link);
end;

function TFhirOrganizationContact.Clone : TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact(inherited Clone);
end;

function TFhirOrganizationContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirOrganizationContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirOrganizationContact)) then
    result := false
  else
  begin
    o := TFhirOrganizationContact(other);
    result := compareDeep(purposeElement, o.purposeElement, true) and compareDeep(nameElement, o.nameElement, true) and 
      compareDeep(telecomList, o.telecomList, true) and compareDeep(addressElement, o.addressElement, true);
  end;
end;

function TFhirOrganizationContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirOrganizationContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirOrganizationContact)) then
    result := false
  else
  begin
    o := TFhirOrganizationContact(other);
    result := true;
  end;
end;

{ TFhirOrganizationContact }

Procedure TFhirOrganizationContact.SetPurpose(value : TFhirCodeableConcept);
begin
  FPurpose.free;
  FPurpose := value;
end;

Procedure TFhirOrganizationContact.SetName(value : TFhirHumanName);
begin
  FName.free;
  FName := value;
end;

Function TFhirOrganizationContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirOrganizationContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirOrganizationContact.SetAddress(value : TFhirAddress);
begin
  FAddress.free;
  FAddress := value;
end;


{ TFhirOrganizationContactListEnumerator }

Constructor TFhirOrganizationContactListEnumerator.Create(list : TFhirOrganizationContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOrganizationContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOrganizationContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOrganizationContactListEnumerator.GetCurrent : TFhirOrganizationContact;
begin
  Result := FList[FIndex];
end;


{ TFhirOrganizationContactList }
procedure TFhirOrganizationContactList.AddItem(value: TFhirOrganizationContact);
begin
  assert(value.ClassName = 'TFhirOrganizationContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOrganizationContact');
  add(value);
end;


function TFhirOrganizationContactList.Append: TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrganizationContactList.ClearItems;
begin
  Clear;
end;

function TFhirOrganizationContactList.GetEnumerator : TFhirOrganizationContactListEnumerator;
begin
  result := TFhirOrganizationContactListEnumerator.Create(self.link);
end;

function TFhirOrganizationContactList.Clone: TFhirOrganizationContactList;
begin
  result := TFhirOrganizationContactList(inherited Clone);
end;

function TFhirOrganizationContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOrganizationContactList.GetItemN(index: Integer): TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact(ObjectByIndex[index]);
end;

function TFhirOrganizationContactList.IndexOf(value: TFhirOrganizationContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOrganizationContactList.Insert(index: Integer): TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrganizationContactList.InsertItem(index: Integer; value: TFhirOrganizationContact);
begin
  assert(value is TFhirOrganizationContact);
  Inherited Insert(index, value);
end;

function TFhirOrganizationContactList.Item(index: Integer): TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact(ObjectByIndex[index]);
end;

function TFhirOrganizationContactList.Link: TFhirOrganizationContactList;
begin
  result := TFhirOrganizationContactList(inherited Link);
end;

procedure TFhirOrganizationContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOrganizationContactList.SetItemByIndex(index: Integer; value: TFhirOrganizationContact);
begin
  assert(value is TFhirOrganizationContact);
  FhirOrganizationContacts[index] := value;
end;

procedure TFhirOrganizationContactList.SetItemN(index: Integer; value: TFhirOrganizationContact);
begin
  assert(value is TFhirOrganizationContact);
  ObjectByIndex[index] := value;
end;

{ TFhirOrganization }

constructor TFhirOrganization.Create;
begin
  inherited;
end;

destructor TFhirOrganization.Destroy;
begin
  FIdentifierList.Free;
  FActive.free;
  FType_.free;
  FName.free;
  FTelecomList.Free;
  FAddressList.Free;
  FPartOf.free;
  FContactList.Free;
  inherited;
end;

function TFhirOrganization.GetResourceType : TFhirResourceType;
begin
  result := frtOrganization;
end;

procedure TFhirOrganization.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirOrganization(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirOrganization(oSource).FIdentifierList);
  end;
  activeElement := TFhirOrganization(oSource).activeElement.Clone;
  type_ := TFhirOrganization(oSource).type_.Clone;
  nameElement := TFhirOrganization(oSource).nameElement.Clone;
  if (TFhirOrganization(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirOrganization(oSource).FTelecomList);
  end;
  if (TFhirOrganization(oSource).FAddressList = nil) then
  begin
    FAddressList.free;
    FAddressList := nil;
  end
  else
  begin
    FAddressList := TFhirAddressList.Create;
    FAddressList.Assign(TFhirOrganization(oSource).FAddressList);
  end;
  partOf := TFhirOrganization(oSource).partOf.Clone;
  if (TFhirOrganization(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirOrganizationContactList.Create;
    FContactList.Assign(TFhirOrganization(oSource).FContactList);
  end;
end;

procedure TFhirOrganization.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'active') Then
     list.add(FActive.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'address') Then
    list.addAll(FAddressList);
  if (child_name = 'partOf') Then
     list.add(FPartOf.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
end;

procedure TFhirOrganization.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'active', 'boolean', FActive.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'partOf', 'Reference(Organization)', FPartOf.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
end;

procedure TFhirOrganization.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'active') then ActiveElement := propValue as TFhirBoolean{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'address') then AddressList.add(propValue as TFhirAddress){2}
  else if (propName = 'partOf') then PartOf := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirOrganizationContact){2}
  else inherited;
end;

function TFhirOrganization.FhirType : string;
begin
  result := 'Organization';
end;

function TFhirOrganization.Link : TFhirOrganization;
begin
  result := TFhirOrganization(inherited Link);
end;

function TFhirOrganization.Clone : TFhirOrganization;
begin
  result := TFhirOrganization(inherited Clone);
end;

{ TFhirOrganization }

Function TFhirOrganization.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirOrganization.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirOrganization.SetActive(value : TFhirBoolean);
begin
  FActive.free;
  FActive := value;
end;

Function TFhirOrganization.GetActiveST : Boolean;
begin
  if FActive = nil then
    result := false
  else
    result := FActive.value;
end;

Procedure TFhirOrganization.SetActiveST(value : Boolean);
begin
  if FActive = nil then
    FActive := TFhirBoolean.create;
  FActive.value := value
end;

Procedure TFhirOrganization.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirOrganization.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirOrganization.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirOrganization.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirOrganization.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirOrganization.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Function TFhirOrganization.GetAddressList : TFhirAddressList;
begin
  if FAddressList = nil then
    FAddressList := TFhirAddressList.Create;
  result := FAddressList;
end;

Function TFhirOrganization.GetHasAddressList : boolean;
begin
  result := (FAddressList <> nil) and (FAddressList.count > 0);
end;

Procedure TFhirOrganization.SetPartOf(value : TFhirReference{TFhirOrganization});
begin
  FPartOf.free;
  FPartOf := value;
end;

Function TFhirOrganization.GetContactList : TFhirOrganizationContactList;
begin
  if FContactList = nil then
    FContactList := TFhirOrganizationContactList.Create;
  result := FContactList;
end;

Function TFhirOrganization.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;


{ TFhirOrganizationListEnumerator }

Constructor TFhirOrganizationListEnumerator.Create(list : TFhirOrganizationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirOrganizationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirOrganizationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirOrganizationListEnumerator.GetCurrent : TFhirOrganization;
begin
  Result := FList[FIndex];
end;


{ TFhirOrganizationList }
procedure TFhirOrganizationList.AddItem(value: TFhirOrganization);
begin
  assert(value.ClassName = 'TFhirOrganization', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirOrganization');
  add(value);
end;


function TFhirOrganizationList.Append: TFhirOrganization;
begin
  result := TFhirOrganization.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrganizationList.ClearItems;
begin
  Clear;
end;

function TFhirOrganizationList.GetEnumerator : TFhirOrganizationListEnumerator;
begin
  result := TFhirOrganizationListEnumerator.Create(self.link);
end;

function TFhirOrganizationList.Clone: TFhirOrganizationList;
begin
  result := TFhirOrganizationList(inherited Clone);
end;

function TFhirOrganizationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirOrganizationList.GetItemN(index: Integer): TFhirOrganization;
begin
  result := TFhirOrganization(ObjectByIndex[index]);
end;

function TFhirOrganizationList.IndexOf(value: TFhirOrganization): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirOrganizationList.Insert(index: Integer): TFhirOrganization;
begin
  result := TFhirOrganization.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirOrganizationList.InsertItem(index: Integer; value: TFhirOrganization);
begin
  assert(value is TFhirOrganization);
  Inherited Insert(index, value);
end;

function TFhirOrganizationList.Item(index: Integer): TFhirOrganization;
begin
  result := TFhirOrganization(ObjectByIndex[index]);
end;

function TFhirOrganizationList.Link: TFhirOrganizationList;
begin
  result := TFhirOrganizationList(inherited Link);
end;

procedure TFhirOrganizationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirOrganizationList.SetItemByIndex(index: Integer; value: TFhirOrganization);
begin
  assert(value is TFhirOrganization);
  FhirOrganizations[index] := value;
end;

procedure TFhirOrganizationList.SetItemN(index: Integer; value: TFhirOrganization);
begin
  assert(value is TFhirOrganization);
  ObjectByIndex[index] := value;
end;

{ TFhirPatientContact }

constructor TFhirPatientContact.Create;
begin
  inherited;
end;

destructor TFhirPatientContact.Destroy;
begin
  FRelationshipList.Free;
  FName.free;
  FTelecomList.Free;
  FAddress.free;
  FGender.free;
  FOrganization.free;
  FPeriod.free;
  inherited;
end;

procedure TFhirPatientContact.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPatientContact(oSource).FRelationshipList = nil) then
  begin
    FRelationshipList.free;
    FRelationshipList := nil;
  end
  else
  begin
    FRelationshipList := TFhirCodeableConceptList.Create;
    FRelationshipList.Assign(TFhirPatientContact(oSource).FRelationshipList);
  end;
  name := TFhirPatientContact(oSource).name.Clone;
  if (TFhirPatientContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirPatientContact(oSource).FTelecomList);
  end;
  address := TFhirPatientContact(oSource).address.Clone;
  FGender := TFhirPatientContact(oSource).FGender.Link;
  organization := TFhirPatientContact(oSource).organization.Clone;
  period := TFhirPatientContact(oSource).period.Clone;
end;

procedure TFhirPatientContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'relationship') Then
    list.addAll(FRelationshipList);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'address') Then
     list.add(FAddress.Link);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirPatientContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'relationship', 'CodeableConcept', FRelationshipList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddress.Link));{2}
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirPatientContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'relationship') then RelationshipList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'name') then Name := propValue as TFhirHumanName{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'address') then Address := propValue as TFhirAddress{4b}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirPatientContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirPatientContact.Link : TFhirPatientContact;
begin
  result := TFhirPatientContact(inherited Link);
end;

function TFhirPatientContact.Clone : TFhirPatientContact;
begin
  result := TFhirPatientContact(inherited Clone);
end;

function TFhirPatientContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPatientContact)) then
    result := false
  else
  begin
    o := TFhirPatientContact(other);
    result := compareDeep(relationshipList, o.relationshipList, true) and compareDeep(nameElement, o.nameElement, true) and 
      compareDeep(telecomList, o.telecomList, true) and compareDeep(addressElement, o.addressElement, true) and 
      compareDeep(genderElement, o.genderElement, true) and compareDeep(organizationElement, o.organizationElement, true) and 
      compareDeep(periodElement, o.periodElement, true);
  end;
end;

function TFhirPatientContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPatientContact)) then
    result := false
  else
  begin
    o := TFhirPatientContact(other);
    result := compareValues(genderElement, o.genderElement, true);
  end;
end;

{ TFhirPatientContact }

Function TFhirPatientContact.GetRelationshipList : TFhirCodeableConceptList;
begin
  if FRelationshipList = nil then
    FRelationshipList := TFhirCodeableConceptList.Create;
  result := FRelationshipList;
end;

Function TFhirPatientContact.GetHasRelationshipList : boolean;
begin
  result := (FRelationshipList <> nil) and (FRelationshipList.count > 0);
end;

Procedure TFhirPatientContact.SetName(value : TFhirHumanName);
begin
  FName.free;
  FName := value;
end;

Function TFhirPatientContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirPatientContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirPatientContact.SetAddress(value : TFhirAddress);
begin
  FAddress.free;
  FAddress := value;
end;

Procedure TFhirPatientContact.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirPatientContact.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirPatientContact.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirPatientContact.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirPatientContact.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirPatientContactListEnumerator }

Constructor TFhirPatientContactListEnumerator.Create(list : TFhirPatientContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPatientContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPatientContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPatientContactListEnumerator.GetCurrent : TFhirPatientContact;
begin
  Result := FList[FIndex];
end;


{ TFhirPatientContactList }
procedure TFhirPatientContactList.AddItem(value: TFhirPatientContact);
begin
  assert(value.ClassName = 'TFhirPatientContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPatientContact');
  add(value);
end;


function TFhirPatientContactList.Append: TFhirPatientContact;
begin
  result := TFhirPatientContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientContactList.ClearItems;
begin
  Clear;
end;

function TFhirPatientContactList.GetEnumerator : TFhirPatientContactListEnumerator;
begin
  result := TFhirPatientContactListEnumerator.Create(self.link);
end;

function TFhirPatientContactList.Clone: TFhirPatientContactList;
begin
  result := TFhirPatientContactList(inherited Clone);
end;

function TFhirPatientContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPatientContactList.GetItemN(index: Integer): TFhirPatientContact;
begin
  result := TFhirPatientContact(ObjectByIndex[index]);
end;

function TFhirPatientContactList.IndexOf(value: TFhirPatientContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPatientContactList.Insert(index: Integer): TFhirPatientContact;
begin
  result := TFhirPatientContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientContactList.InsertItem(index: Integer; value: TFhirPatientContact);
begin
  assert(value is TFhirPatientContact);
  Inherited Insert(index, value);
end;

function TFhirPatientContactList.Item(index: Integer): TFhirPatientContact;
begin
  result := TFhirPatientContact(ObjectByIndex[index]);
end;

function TFhirPatientContactList.Link: TFhirPatientContactList;
begin
  result := TFhirPatientContactList(inherited Link);
end;

procedure TFhirPatientContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPatientContactList.SetItemByIndex(index: Integer; value: TFhirPatientContact);
begin
  assert(value is TFhirPatientContact);
  FhirPatientContacts[index] := value;
end;

procedure TFhirPatientContactList.SetItemN(index: Integer; value: TFhirPatientContact);
begin
  assert(value is TFhirPatientContact);
  ObjectByIndex[index] := value;
end;

{ TFhirPatientAnimal }

constructor TFhirPatientAnimal.Create;
begin
  inherited;
end;

destructor TFhirPatientAnimal.Destroy;
begin
  FSpecies.free;
  FBreed.free;
  FGenderStatus.free;
  inherited;
end;

procedure TFhirPatientAnimal.Assign(oSource : TAdvObject);
begin
  inherited;
  species := TFhirPatientAnimal(oSource).species.Clone;
  breed := TFhirPatientAnimal(oSource).breed.Clone;
  genderStatus := TFhirPatientAnimal(oSource).genderStatus.Clone;
end;

procedure TFhirPatientAnimal.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'species') Then
     list.add(FSpecies.Link);
  if (child_name = 'breed') Then
     list.add(FBreed.Link);
  if (child_name = 'genderStatus') Then
     list.add(FGenderStatus.Link);
end;

procedure TFhirPatientAnimal.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'species', 'CodeableConcept', FSpecies.Link));{2}
  oList.add(TFHIRProperty.create(self, 'breed', 'CodeableConcept', FBreed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'genderStatus', 'CodeableConcept', FGenderStatus.Link));{2}
end;

procedure TFhirPatientAnimal.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'species') then Species := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'breed') then Breed := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'genderStatus') then GenderStatus := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirPatientAnimal.FhirType : string;
begin
  result := 'animal';
end;

function TFhirPatientAnimal.Link : TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal(inherited Link);
end;

function TFhirPatientAnimal.Clone : TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal(inherited Clone);
end;

function TFhirPatientAnimal.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientAnimal;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPatientAnimal)) then
    result := false
  else
  begin
    o := TFhirPatientAnimal(other);
    result := compareDeep(speciesElement, o.speciesElement, true) and compareDeep(breedElement, o.breedElement, true) and 
      compareDeep(genderStatusElement, o.genderStatusElement, true);
  end;
end;

function TFhirPatientAnimal.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientAnimal;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPatientAnimal)) then
    result := false
  else
  begin
    o := TFhirPatientAnimal(other);
    result := true;
  end;
end;

{ TFhirPatientAnimal }

Procedure TFhirPatientAnimal.SetSpecies(value : TFhirCodeableConcept);
begin
  FSpecies.free;
  FSpecies := value;
end;

Procedure TFhirPatientAnimal.SetBreed(value : TFhirCodeableConcept);
begin
  FBreed.free;
  FBreed := value;
end;

Procedure TFhirPatientAnimal.SetGenderStatus(value : TFhirCodeableConcept);
begin
  FGenderStatus.free;
  FGenderStatus := value;
end;


{ TFhirPatientAnimalListEnumerator }

Constructor TFhirPatientAnimalListEnumerator.Create(list : TFhirPatientAnimalList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPatientAnimalListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPatientAnimalListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPatientAnimalListEnumerator.GetCurrent : TFhirPatientAnimal;
begin
  Result := FList[FIndex];
end;


{ TFhirPatientAnimalList }
procedure TFhirPatientAnimalList.AddItem(value: TFhirPatientAnimal);
begin
  assert(value.ClassName = 'TFhirPatientAnimal', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPatientAnimal');
  add(value);
end;


function TFhirPatientAnimalList.Append: TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientAnimalList.ClearItems;
begin
  Clear;
end;

function TFhirPatientAnimalList.GetEnumerator : TFhirPatientAnimalListEnumerator;
begin
  result := TFhirPatientAnimalListEnumerator.Create(self.link);
end;

function TFhirPatientAnimalList.Clone: TFhirPatientAnimalList;
begin
  result := TFhirPatientAnimalList(inherited Clone);
end;

function TFhirPatientAnimalList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPatientAnimalList.GetItemN(index: Integer): TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal(ObjectByIndex[index]);
end;

function TFhirPatientAnimalList.IndexOf(value: TFhirPatientAnimal): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPatientAnimalList.Insert(index: Integer): TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientAnimalList.InsertItem(index: Integer; value: TFhirPatientAnimal);
begin
  assert(value is TFhirPatientAnimal);
  Inherited Insert(index, value);
end;

function TFhirPatientAnimalList.Item(index: Integer): TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal(ObjectByIndex[index]);
end;

function TFhirPatientAnimalList.Link: TFhirPatientAnimalList;
begin
  result := TFhirPatientAnimalList(inherited Link);
end;

procedure TFhirPatientAnimalList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPatientAnimalList.SetItemByIndex(index: Integer; value: TFhirPatientAnimal);
begin
  assert(value is TFhirPatientAnimal);
  FhirPatientAnimals[index] := value;
end;

procedure TFhirPatientAnimalList.SetItemN(index: Integer; value: TFhirPatientAnimal);
begin
  assert(value is TFhirPatientAnimal);
  ObjectByIndex[index] := value;
end;

{ TFhirPatientCommunication }

constructor TFhirPatientCommunication.Create;
begin
  inherited;
end;

destructor TFhirPatientCommunication.Destroy;
begin
  FLanguage.free;
  FPreferred.free;
  inherited;
end;

procedure TFhirPatientCommunication.Assign(oSource : TAdvObject);
begin
  inherited;
  language := TFhirPatientCommunication(oSource).language.Clone;
  preferredElement := TFhirPatientCommunication(oSource).preferredElement.Clone;
end;

procedure TFhirPatientCommunication.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'language') Then
     list.add(FLanguage.Link);
  if (child_name = 'preferred') Then
     list.add(FPreferred.Link);
end;

procedure TFhirPatientCommunication.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'language', 'CodeableConcept', FLanguage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'preferred', 'boolean', FPreferred.Link));{2}
end;

procedure TFhirPatientCommunication.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'language') then Language := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'preferred') then PreferredElement := propValue as TFhirBoolean{5a}
  else inherited;
end;

function TFhirPatientCommunication.FhirType : string;
begin
  result := 'communication';
end;

function TFhirPatientCommunication.Link : TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication(inherited Link);
end;

function TFhirPatientCommunication.Clone : TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication(inherited Clone);
end;

function TFhirPatientCommunication.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientCommunication;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPatientCommunication)) then
    result := false
  else
  begin
    o := TFhirPatientCommunication(other);
    result := compareDeep(languageElement, o.languageElement, true) and compareDeep(preferredElement, o.preferredElement, true);
  end;
end;

function TFhirPatientCommunication.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientCommunication;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPatientCommunication)) then
    result := false
  else
  begin
    o := TFhirPatientCommunication(other);
    result := compareValues(preferredElement, o.preferredElement, true);
  end;
end;

{ TFhirPatientCommunication }

Procedure TFhirPatientCommunication.SetLanguage(value : TFhirCodeableConcept);
begin
  FLanguage.free;
  FLanguage := value;
end;

Procedure TFhirPatientCommunication.SetPreferred(value : TFhirBoolean);
begin
  FPreferred.free;
  FPreferred := value;
end;

Function TFhirPatientCommunication.GetPreferredST : Boolean;
begin
  if FPreferred = nil then
    result := false
  else
    result := FPreferred.value;
end;

Procedure TFhirPatientCommunication.SetPreferredST(value : Boolean);
begin
  if FPreferred = nil then
    FPreferred := TFhirBoolean.create;
  FPreferred.value := value
end;


{ TFhirPatientCommunicationListEnumerator }

Constructor TFhirPatientCommunicationListEnumerator.Create(list : TFhirPatientCommunicationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPatientCommunicationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPatientCommunicationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPatientCommunicationListEnumerator.GetCurrent : TFhirPatientCommunication;
begin
  Result := FList[FIndex];
end;


{ TFhirPatientCommunicationList }
procedure TFhirPatientCommunicationList.AddItem(value: TFhirPatientCommunication);
begin
  assert(value.ClassName = 'TFhirPatientCommunication', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPatientCommunication');
  add(value);
end;


function TFhirPatientCommunicationList.Append: TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientCommunicationList.ClearItems;
begin
  Clear;
end;

function TFhirPatientCommunicationList.GetEnumerator : TFhirPatientCommunicationListEnumerator;
begin
  result := TFhirPatientCommunicationListEnumerator.Create(self.link);
end;

function TFhirPatientCommunicationList.Clone: TFhirPatientCommunicationList;
begin
  result := TFhirPatientCommunicationList(inherited Clone);
end;

function TFhirPatientCommunicationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPatientCommunicationList.GetItemN(index: Integer): TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication(ObjectByIndex[index]);
end;

function TFhirPatientCommunicationList.IndexOf(value: TFhirPatientCommunication): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPatientCommunicationList.Insert(index: Integer): TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientCommunicationList.InsertItem(index: Integer; value: TFhirPatientCommunication);
begin
  assert(value is TFhirPatientCommunication);
  Inherited Insert(index, value);
end;

function TFhirPatientCommunicationList.Item(index: Integer): TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication(ObjectByIndex[index]);
end;

function TFhirPatientCommunicationList.Link: TFhirPatientCommunicationList;
begin
  result := TFhirPatientCommunicationList(inherited Link);
end;

procedure TFhirPatientCommunicationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPatientCommunicationList.SetItemByIndex(index: Integer; value: TFhirPatientCommunication);
begin
  assert(value is TFhirPatientCommunication);
  FhirPatientCommunications[index] := value;
end;

procedure TFhirPatientCommunicationList.SetItemN(index: Integer; value: TFhirPatientCommunication);
begin
  assert(value is TFhirPatientCommunication);
  ObjectByIndex[index] := value;
end;

{ TFhirPatientLink }

constructor TFhirPatientLink.Create;
begin
  inherited;
end;

destructor TFhirPatientLink.Destroy;
begin
  FOther.free;
  FType_.free;
  inherited;
end;

procedure TFhirPatientLink.Assign(oSource : TAdvObject);
begin
  inherited;
  other := TFhirPatientLink(oSource).other.Clone;
  FType_ := TFhirPatientLink(oSource).FType_.Link;
end;

procedure TFhirPatientLink.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'other') Then
     list.add(FOther.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
end;

procedure TFhirPatientLink.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'other', 'Reference(Patient)', FOther.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
end;

procedure TFhirPatientLink.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'other') then Other := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else inherited;
end;

function TFhirPatientLink.FhirType : string;
begin
  result := 'link';
end;

function TFhirPatientLink.Link : TFhirPatientLink;
begin
  result := TFhirPatientLink(inherited Link);
end;

function TFhirPatientLink.Clone : TFhirPatientLink;
begin
  result := TFhirPatientLink(inherited Clone);
end;

function TFhirPatientLink.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientLink;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPatientLink)) then
    result := false
  else
  begin
    o := TFhirPatientLink(other);
    result := compareDeep(otherElement, o.otherElement, true) and compareDeep(type_Element, o.type_Element, true);
  end;
end;

function TFhirPatientLink.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPatientLink;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPatientLink)) then
    result := false
  else
  begin
    o := TFhirPatientLink(other);
    result := compareValues(type_Element, o.type_Element, true);
  end;
end;

{ TFhirPatientLink }

Procedure TFhirPatientLink.SetOther(value : TFhirReference{TFhirPatient});
begin
  FOther.free;
  FOther := value;
end;

Procedure TFhirPatientLink.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirPatientLink.GetType_ST : TFhirLinkType;
begin
  if FType_ = nil then
    result := TFhirLinkType(0)
  else
    result := TFhirLinkType(StringArrayIndexOfSensitive(CODES_TFhirLinkType, FType_.value));
end;

Procedure TFhirPatientLink.SetType_ST(value : TFhirLinkType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirLinkType[value]);
end;


{ TFhirPatientLinkListEnumerator }

Constructor TFhirPatientLinkListEnumerator.Create(list : TFhirPatientLinkList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPatientLinkListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPatientLinkListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPatientLinkListEnumerator.GetCurrent : TFhirPatientLink;
begin
  Result := FList[FIndex];
end;


{ TFhirPatientLinkList }
procedure TFhirPatientLinkList.AddItem(value: TFhirPatientLink);
begin
  assert(value.ClassName = 'TFhirPatientLink', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPatientLink');
  add(value);
end;


function TFhirPatientLinkList.Append: TFhirPatientLink;
begin
  result := TFhirPatientLink.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientLinkList.ClearItems;
begin
  Clear;
end;

function TFhirPatientLinkList.GetEnumerator : TFhirPatientLinkListEnumerator;
begin
  result := TFhirPatientLinkListEnumerator.Create(self.link);
end;

function TFhirPatientLinkList.Clone: TFhirPatientLinkList;
begin
  result := TFhirPatientLinkList(inherited Clone);
end;

function TFhirPatientLinkList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPatientLinkList.GetItemN(index: Integer): TFhirPatientLink;
begin
  result := TFhirPatientLink(ObjectByIndex[index]);
end;

function TFhirPatientLinkList.IndexOf(value: TFhirPatientLink): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPatientLinkList.Insert(index: Integer): TFhirPatientLink;
begin
  result := TFhirPatientLink.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientLinkList.InsertItem(index: Integer; value: TFhirPatientLink);
begin
  assert(value is TFhirPatientLink);
  Inherited Insert(index, value);
end;

function TFhirPatientLinkList.Item(index: Integer): TFhirPatientLink;
begin
  result := TFhirPatientLink(ObjectByIndex[index]);
end;

function TFhirPatientLinkList.Link: TFhirPatientLinkList;
begin
  result := TFhirPatientLinkList(inherited Link);
end;

procedure TFhirPatientLinkList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPatientLinkList.SetItemByIndex(index: Integer; value: TFhirPatientLink);
begin
  assert(value is TFhirPatientLink);
  FhirPatientLinks[index] := value;
end;

procedure TFhirPatientLinkList.SetItemN(index: Integer; value: TFhirPatientLink);
begin
  assert(value is TFhirPatientLink);
  ObjectByIndex[index] := value;
end;

{ TFhirPatient }

constructor TFhirPatient.Create;
begin
  inherited;
end;

destructor TFhirPatient.Destroy;
begin
  FIdentifierList.Free;
  FActive.free;
  FNameList.Free;
  FTelecomList.Free;
  FGender.free;
  FBirthDate.free;
  FDeceased.free;
  FAddressList.Free;
  FMaritalStatus.free;
  FMultipleBirth.free;
  FPhotoList.Free;
  FContactList.Free;
  FAnimal.free;
  FCommunicationList.Free;
  FCareProviderList.Free;
  FManagingOrganization.free;
  FLink_List.Free;
  inherited;
end;

function TFhirPatient.GetResourceType : TFhirResourceType;
begin
  result := frtPatient;
end;

procedure TFhirPatient.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPatient(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPatient(oSource).FIdentifierList);
  end;
  activeElement := TFhirPatient(oSource).activeElement.Clone;
  if (TFhirPatient(oSource).FNameList = nil) then
  begin
    FNameList.free;
    FNameList := nil;
  end
  else
  begin
    FNameList := TFhirHumanNameList.Create;
    FNameList.Assign(TFhirPatient(oSource).FNameList);
  end;
  if (TFhirPatient(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirPatient(oSource).FTelecomList);
  end;
  FGender := TFhirPatient(oSource).FGender.Link;
  birthDateElement := TFhirPatient(oSource).birthDateElement.Clone;
  deceased := TFhirPatient(oSource).deceased.Clone;
  if (TFhirPatient(oSource).FAddressList = nil) then
  begin
    FAddressList.free;
    FAddressList := nil;
  end
  else
  begin
    FAddressList := TFhirAddressList.Create;
    FAddressList.Assign(TFhirPatient(oSource).FAddressList);
  end;
  maritalStatus := TFhirPatient(oSource).maritalStatus.Clone;
  multipleBirth := TFhirPatient(oSource).multipleBirth.Clone;
  if (TFhirPatient(oSource).FPhotoList = nil) then
  begin
    FPhotoList.free;
    FPhotoList := nil;
  end
  else
  begin
    FPhotoList := TFhirAttachmentList.Create;
    FPhotoList.Assign(TFhirPatient(oSource).FPhotoList);
  end;
  if (TFhirPatient(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirPatientContactList.Create;
    FContactList.Assign(TFhirPatient(oSource).FContactList);
  end;
  animal := TFhirPatient(oSource).animal.Clone;
  if (TFhirPatient(oSource).FCommunicationList = nil) then
  begin
    FCommunicationList.free;
    FCommunicationList := nil;
  end
  else
  begin
    FCommunicationList := TFhirPatientCommunicationList.Create;
    FCommunicationList.Assign(TFhirPatient(oSource).FCommunicationList);
  end;
  if (TFhirPatient(oSource).FCareProviderList = nil) then
  begin
    FCareProviderList.free;
    FCareProviderList := nil;
  end
  else
  begin
    FCareProviderList := TFhirReferenceList{Resource}.Create;
    FCareProviderList.Assign(TFhirPatient(oSource).FCareProviderList);
  end;
  managingOrganization := TFhirPatient(oSource).managingOrganization.Clone;
  if (TFhirPatient(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirPatientLinkList.Create;
    FLink_List.Assign(TFhirPatient(oSource).FLink_List);
  end;
end;

procedure TFhirPatient.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'active') Then
     list.add(FActive.Link);
  if (child_name = 'name') Then
    list.addAll(FNameList);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'birthDate') Then
     list.add(FBirthDate.Link);
  if (child_name = 'deceased[x]') Then
     list.add(FDeceased.Link);
  if (child_name = 'address') Then
    list.addAll(FAddressList);
  if (child_name = 'maritalStatus') Then
     list.add(FMaritalStatus.Link);
  if (child_name = 'multipleBirth[x]') Then
     list.add(FMultipleBirth.Link);
  if (child_name = 'photo') Then
    list.addAll(FPhotoList);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'animal') Then
     list.add(FAnimal.Link);
  if (child_name = 'communication') Then
    list.addAll(FCommunicationList);
  if (child_name = 'careProvider') Then
    list.addAll(FCareProviderList);
  if (child_name = 'managingOrganization') Then
     list.add(FManagingOrganization.Link);
  if (child_name = 'link') Then
    list.addAll(FLink_List);
end;

procedure TFhirPatient.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'active', 'boolean', FActive.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FNameList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'birthDate', 'date', FBirthDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'deceased[x]', 'boolean|dateTime', FDeceased.Link));{2}
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'maritalStatus', 'CodeableConcept', FMaritalStatus.Link));{2}
  oList.add(TFHIRProperty.create(self, 'multipleBirth[x]', 'boolean|integer', FMultipleBirth.Link));{2}
  oList.add(TFHIRProperty.create(self, 'photo', 'Attachment', FPhotoList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'animal', '', FAnimal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'communication', '', FCommunicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'careProvider', 'Reference(Organization|Practitioner)', FCareProviderList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'managingOrganization', 'Reference(Organization)', FManagingOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'link', '', FLink_List.Link)){3};
end;

procedure TFhirPatient.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'active') then ActiveElement := propValue as TFhirBoolean{5a}
  else if (propName = 'name') then NameList.add(propValue as TFhirHumanName){2}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName = 'birthDate') then BirthDateElement := propValue as TFhirDate{5a}
  else if (propName.startsWith('deceased')) then Deceased := propValue as TFhirType{4}
  else if (propName = 'address') then AddressList.add(propValue as TFhirAddress){2}
  else if (propName = 'maritalStatus') then MaritalStatus := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('multipleBirth')) then MultipleBirth := propValue as TFhirType{4}
  else if (propName = 'photo') then PhotoList.add(propValue as TFhirAttachment){2}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirPatientContact){2}
  else if (propName = 'animal') then Animal := propValue as TFhirPatientAnimal{4b}
  else if (propName = 'communication') then CommunicationList.add(propValue as TFhirPatientCommunication){2}
  else if (propName = 'careProvider') then CareProviderList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'managingOrganization') then ManagingOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'link') then Link_List.add(propValue as TFhirPatientLink){2}
  else inherited;
end;

function TFhirPatient.FhirType : string;
begin
  result := 'Patient';
end;

function TFhirPatient.Link : TFhirPatient;
begin
  result := TFhirPatient(inherited Link);
end;

function TFhirPatient.Clone : TFhirPatient;
begin
  result := TFhirPatient(inherited Clone);
end;

{ TFhirPatient }

Function TFhirPatient.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPatient.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirPatient.SetActive(value : TFhirBoolean);
begin
  FActive.free;
  FActive := value;
end;

Function TFhirPatient.GetActiveST : Boolean;
begin
  if FActive = nil then
    result := false
  else
    result := FActive.value;
end;

Procedure TFhirPatient.SetActiveST(value : Boolean);
begin
  if FActive = nil then
    FActive := TFhirBoolean.create;
  FActive.value := value
end;

Function TFhirPatient.GetNameList : TFhirHumanNameList;
begin
  if FNameList = nil then
    FNameList := TFhirHumanNameList.Create;
  result := FNameList;
end;

Function TFhirPatient.GetHasNameList : boolean;
begin
  result := (FNameList <> nil) and (FNameList.count > 0);
end;

Function TFhirPatient.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirPatient.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirPatient.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirPatient.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirPatient.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirPatient.SetBirthDate(value : TFhirDate);
begin
  FBirthDate.free;
  FBirthDate := value;
end;

Function TFhirPatient.GetBirthDateST : TDateAndTime;
begin
  if FBirthDate = nil then
    result := nil
  else
    result := FBirthDate.value;
end;

Procedure TFhirPatient.SetBirthDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FBirthDate = nil then
      FBirthDate := TFhirDate.create;
    FBirthDate.value := value
  end
  else if FBirthDate <> nil then
    FBirthDate.value := nil;
end;

Procedure TFhirPatient.SetDeceased(value : TFhirType);
begin
  FDeceased.free;
  FDeceased := value;
end;

Function TFhirPatient.GetAddressList : TFhirAddressList;
begin
  if FAddressList = nil then
    FAddressList := TFhirAddressList.Create;
  result := FAddressList;
end;

Function TFhirPatient.GetHasAddressList : boolean;
begin
  result := (FAddressList <> nil) and (FAddressList.count > 0);
end;

Procedure TFhirPatient.SetMaritalStatus(value : TFhirCodeableConcept);
begin
  FMaritalStatus.free;
  FMaritalStatus := value;
end;

Procedure TFhirPatient.SetMultipleBirth(value : TFhirType);
begin
  FMultipleBirth.free;
  FMultipleBirth := value;
end;

Function TFhirPatient.GetPhotoList : TFhirAttachmentList;
begin
  if FPhotoList = nil then
    FPhotoList := TFhirAttachmentList.Create;
  result := FPhotoList;
end;

Function TFhirPatient.GetHasPhotoList : boolean;
begin
  result := (FPhotoList <> nil) and (FPhotoList.count > 0);
end;

Function TFhirPatient.GetContactList : TFhirPatientContactList;
begin
  if FContactList = nil then
    FContactList := TFhirPatientContactList.Create;
  result := FContactList;
end;

Function TFhirPatient.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirPatient.SetAnimal(value : TFhirPatientAnimal);
begin
  FAnimal.free;
  FAnimal := value;
end;

Function TFhirPatient.GetCommunicationList : TFhirPatientCommunicationList;
begin
  if FCommunicationList = nil then
    FCommunicationList := TFhirPatientCommunicationList.Create;
  result := FCommunicationList;
end;

Function TFhirPatient.GetHasCommunicationList : boolean;
begin
  result := (FCommunicationList <> nil) and (FCommunicationList.count > 0);
end;

Function TFhirPatient.GetCareProviderList : TFhirReferenceList{Resource};
begin
  if FCareProviderList = nil then
    FCareProviderList := TFhirReferenceList{Resource}.Create;
  result := FCareProviderList;
end;

Function TFhirPatient.GetHasCareProviderList : boolean;
begin
  result := (FCareProviderList <> nil) and (FCareProviderList.count > 0);
end;

Procedure TFhirPatient.SetManagingOrganization(value : TFhirReference{TFhirOrganization});
begin
  FManagingOrganization.free;
  FManagingOrganization := value;
end;

Function TFhirPatient.GetLink_List : TFhirPatientLinkList;
begin
  if FLink_List = nil then
    FLink_List := TFhirPatientLinkList.Create;
  result := FLink_List;
end;

Function TFhirPatient.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;


{ TFhirPatientListEnumerator }

Constructor TFhirPatientListEnumerator.Create(list : TFhirPatientList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPatientListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPatientListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPatientListEnumerator.GetCurrent : TFhirPatient;
begin
  Result := FList[FIndex];
end;


{ TFhirPatientList }
procedure TFhirPatientList.AddItem(value: TFhirPatient);
begin
  assert(value.ClassName = 'TFhirPatient', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPatient');
  add(value);
end;


function TFhirPatientList.Append: TFhirPatient;
begin
  result := TFhirPatient.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientList.ClearItems;
begin
  Clear;
end;

function TFhirPatientList.GetEnumerator : TFhirPatientListEnumerator;
begin
  result := TFhirPatientListEnumerator.Create(self.link);
end;

function TFhirPatientList.Clone: TFhirPatientList;
begin
  result := TFhirPatientList(inherited Clone);
end;

function TFhirPatientList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPatientList.GetItemN(index: Integer): TFhirPatient;
begin
  result := TFhirPatient(ObjectByIndex[index]);
end;

function TFhirPatientList.IndexOf(value: TFhirPatient): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPatientList.Insert(index: Integer): TFhirPatient;
begin
  result := TFhirPatient.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPatientList.InsertItem(index: Integer; value: TFhirPatient);
begin
  assert(value is TFhirPatient);
  Inherited Insert(index, value);
end;

function TFhirPatientList.Item(index: Integer): TFhirPatient;
begin
  result := TFhirPatient(ObjectByIndex[index]);
end;

function TFhirPatientList.Link: TFhirPatientList;
begin
  result := TFhirPatientList(inherited Link);
end;

procedure TFhirPatientList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPatientList.SetItemByIndex(index: Integer; value: TFhirPatient);
begin
  assert(value is TFhirPatient);
  FhirPatients[index] := value;
end;

procedure TFhirPatientList.SetItemN(index: Integer; value: TFhirPatient);
begin
  assert(value is TFhirPatient);
  ObjectByIndex[index] := value;
end;

{ TFhirPaymentNotice }

constructor TFhirPaymentNotice.Create;
begin
  inherited;
end;

destructor TFhirPaymentNotice.Destroy;
begin
  FIdentifierList.Free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FTarget.free;
  FProvider.free;
  FOrganization.free;
  FRequest.free;
  FResponse.free;
  FPaymentStatus.free;
  inherited;
end;

function TFhirPaymentNotice.GetResourceType : TFhirResourceType;
begin
  result := frtPaymentNotice;
end;

procedure TFhirPaymentNotice.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPaymentNotice(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPaymentNotice(oSource).FIdentifierList);
  end;
  ruleset := TFhirPaymentNotice(oSource).ruleset.Clone;
  originalRuleset := TFhirPaymentNotice(oSource).originalRuleset.Clone;
  createdElement := TFhirPaymentNotice(oSource).createdElement.Clone;
  target := TFhirPaymentNotice(oSource).target.Clone;
  provider := TFhirPaymentNotice(oSource).provider.Clone;
  organization := TFhirPaymentNotice(oSource).organization.Clone;
  request := TFhirPaymentNotice(oSource).request.Clone;
  response := TFhirPaymentNotice(oSource).response.Clone;
  paymentStatus := TFhirPaymentNotice(oSource).paymentStatus.Clone;
end;

procedure TFhirPaymentNotice.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
  if (child_name = 'paymentStatus') Then
     list.add(FPaymentStatus.Link);
end;

procedure TFhirPaymentNotice.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Any)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', 'Reference(Any)', FResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'paymentStatus', 'Coding', FPaymentStatus.Link));{2}
end;

procedure TFhirPaymentNotice.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'response') then Response := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'paymentStatus') then PaymentStatus := propValue as TFhirCoding{4b}
  else inherited;
end;

function TFhirPaymentNotice.FhirType : string;
begin
  result := 'PaymentNotice';
end;

function TFhirPaymentNotice.Link : TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice(inherited Link);
end;

function TFhirPaymentNotice.Clone : TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice(inherited Clone);
end;

{ TFhirPaymentNotice }

Function TFhirPaymentNotice.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPaymentNotice.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirPaymentNotice.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirPaymentNotice.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirPaymentNotice.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirPaymentNotice.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirPaymentNotice.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirPaymentNotice.SetTarget(value : TFhirReference{TFhirOrganization});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirPaymentNotice.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirPaymentNotice.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirPaymentNotice.SetRequest(value : TFhirReference{TFhirReference});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirPaymentNotice.SetResponse(value : TFhirReference{TFhirReference});
begin
  FResponse.free;
  FResponse := value;
end;

Procedure TFhirPaymentNotice.SetPaymentStatus(value : TFhirCoding);
begin
  FPaymentStatus.free;
  FPaymentStatus := value;
end;


{ TFhirPaymentNoticeListEnumerator }

Constructor TFhirPaymentNoticeListEnumerator.Create(list : TFhirPaymentNoticeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPaymentNoticeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPaymentNoticeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPaymentNoticeListEnumerator.GetCurrent : TFhirPaymentNotice;
begin
  Result := FList[FIndex];
end;


{ TFhirPaymentNoticeList }
procedure TFhirPaymentNoticeList.AddItem(value: TFhirPaymentNotice);
begin
  assert(value.ClassName = 'TFhirPaymentNotice', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPaymentNotice');
  add(value);
end;


function TFhirPaymentNoticeList.Append: TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentNoticeList.ClearItems;
begin
  Clear;
end;

function TFhirPaymentNoticeList.GetEnumerator : TFhirPaymentNoticeListEnumerator;
begin
  result := TFhirPaymentNoticeListEnumerator.Create(self.link);
end;

function TFhirPaymentNoticeList.Clone: TFhirPaymentNoticeList;
begin
  result := TFhirPaymentNoticeList(inherited Clone);
end;

function TFhirPaymentNoticeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPaymentNoticeList.GetItemN(index: Integer): TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice(ObjectByIndex[index]);
end;

function TFhirPaymentNoticeList.IndexOf(value: TFhirPaymentNotice): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPaymentNoticeList.Insert(index: Integer): TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentNoticeList.InsertItem(index: Integer; value: TFhirPaymentNotice);
begin
  assert(value is TFhirPaymentNotice);
  Inherited Insert(index, value);
end;

function TFhirPaymentNoticeList.Item(index: Integer): TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice(ObjectByIndex[index]);
end;

function TFhirPaymentNoticeList.Link: TFhirPaymentNoticeList;
begin
  result := TFhirPaymentNoticeList(inherited Link);
end;

procedure TFhirPaymentNoticeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPaymentNoticeList.SetItemByIndex(index: Integer; value: TFhirPaymentNotice);
begin
  assert(value is TFhirPaymentNotice);
  FhirPaymentNotices[index] := value;
end;

procedure TFhirPaymentNoticeList.SetItemN(index: Integer; value: TFhirPaymentNotice);
begin
  assert(value is TFhirPaymentNotice);
  ObjectByIndex[index] := value;
end;

{ TFhirPaymentReconciliationDetail }

constructor TFhirPaymentReconciliationDetail.Create;
begin
  inherited;
end;

destructor TFhirPaymentReconciliationDetail.Destroy;
begin
  FType_.free;
  FRequest.free;
  FResponce.free;
  FSubmitter.free;
  FPayee.free;
  FDate.free;
  FAmount.free;
  inherited;
end;

procedure TFhirPaymentReconciliationDetail.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirPaymentReconciliationDetail(oSource).type_.Clone;
  request := TFhirPaymentReconciliationDetail(oSource).request.Clone;
  responce := TFhirPaymentReconciliationDetail(oSource).responce.Clone;
  submitter := TFhirPaymentReconciliationDetail(oSource).submitter.Clone;
  payee := TFhirPaymentReconciliationDetail(oSource).payee.Clone;
  dateElement := TFhirPaymentReconciliationDetail(oSource).dateElement.Clone;
  amount := TFhirPaymentReconciliationDetail(oSource).amount.Clone;
end;

procedure TFhirPaymentReconciliationDetail.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'responce') Then
     list.add(FResponce.Link);
  if (child_name = 'submitter') Then
     list.add(FSubmitter.Link);
  if (child_name = 'payee') Then
     list.add(FPayee.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'amount') Then
     list.add(FAmount.Link);
end;

procedure TFhirPaymentReconciliationDetail.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Any)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'responce', 'Reference(Any)', FResponce.Link));{2}
  oList.add(TFHIRProperty.create(self, 'submitter', 'Reference(Organization)', FSubmitter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'payee', 'Reference(Organization)', FPayee.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'date', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'amount', 'Money', FAmount.Link));{2}
end;

procedure TFhirPaymentReconciliationDetail.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'responce') then Responce := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'submitter') then Submitter := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'payee') then Payee := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDate{5a}
  else if (propName = 'amount') then Amount := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirPaymentReconciliationDetail.FhirType : string;
begin
  result := 'detail';
end;

function TFhirPaymentReconciliationDetail.Link : TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail(inherited Link);
end;

function TFhirPaymentReconciliationDetail.Clone : TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail(inherited Clone);
end;

function TFhirPaymentReconciliationDetail.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPaymentReconciliationDetail;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPaymentReconciliationDetail)) then
    result := false
  else
  begin
    o := TFhirPaymentReconciliationDetail(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(requestElement, o.requestElement, true) and 
      compareDeep(responceElement, o.responceElement, true) and compareDeep(submitterElement, o.submitterElement, true) and 
      compareDeep(payeeElement, o.payeeElement, true) and compareDeep(dateElement, o.dateElement, true) and 
      compareDeep(amountElement, o.amountElement, true);
  end;
end;

function TFhirPaymentReconciliationDetail.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPaymentReconciliationDetail;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPaymentReconciliationDetail)) then
    result := false
  else
  begin
    o := TFhirPaymentReconciliationDetail(other);
    result := compareValues(dateElement, o.dateElement, true);
  end;
end;

{ TFhirPaymentReconciliationDetail }

Procedure TFhirPaymentReconciliationDetail.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirPaymentReconciliationDetail.SetRequest(value : TFhirReference{TFhirReference});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirPaymentReconciliationDetail.SetResponce(value : TFhirReference{TFhirReference});
begin
  FResponce.free;
  FResponce := value;
end;

Procedure TFhirPaymentReconciliationDetail.SetSubmitter(value : TFhirReference{TFhirOrganization});
begin
  FSubmitter.free;
  FSubmitter := value;
end;

Procedure TFhirPaymentReconciliationDetail.SetPayee(value : TFhirReference{TFhirOrganization});
begin
  FPayee.free;
  FPayee := value;
end;

Procedure TFhirPaymentReconciliationDetail.SetDate(value : TFhirDate);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirPaymentReconciliationDetail.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirPaymentReconciliationDetail.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDate.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirPaymentReconciliationDetail.SetAmount(value : TFhirQuantity);
begin
  FAmount.free;
  FAmount := value;
end;


{ TFhirPaymentReconciliationDetailListEnumerator }

Constructor TFhirPaymentReconciliationDetailListEnumerator.Create(list : TFhirPaymentReconciliationDetailList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPaymentReconciliationDetailListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPaymentReconciliationDetailListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPaymentReconciliationDetailListEnumerator.GetCurrent : TFhirPaymentReconciliationDetail;
begin
  Result := FList[FIndex];
end;


{ TFhirPaymentReconciliationDetailList }
procedure TFhirPaymentReconciliationDetailList.AddItem(value: TFhirPaymentReconciliationDetail);
begin
  assert(value.ClassName = 'TFhirPaymentReconciliationDetail', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPaymentReconciliationDetail');
  add(value);
end;


function TFhirPaymentReconciliationDetailList.Append: TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationDetailList.ClearItems;
begin
  Clear;
end;

function TFhirPaymentReconciliationDetailList.GetEnumerator : TFhirPaymentReconciliationDetailListEnumerator;
begin
  result := TFhirPaymentReconciliationDetailListEnumerator.Create(self.link);
end;

function TFhirPaymentReconciliationDetailList.Clone: TFhirPaymentReconciliationDetailList;
begin
  result := TFhirPaymentReconciliationDetailList(inherited Clone);
end;

function TFhirPaymentReconciliationDetailList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPaymentReconciliationDetailList.GetItemN(index: Integer): TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationDetailList.IndexOf(value: TFhirPaymentReconciliationDetail): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPaymentReconciliationDetailList.Insert(index: Integer): TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationDetailList.InsertItem(index: Integer; value: TFhirPaymentReconciliationDetail);
begin
  assert(value is TFhirPaymentReconciliationDetail);
  Inherited Insert(index, value);
end;

function TFhirPaymentReconciliationDetailList.Item(index: Integer): TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationDetailList.Link: TFhirPaymentReconciliationDetailList;
begin
  result := TFhirPaymentReconciliationDetailList(inherited Link);
end;

procedure TFhirPaymentReconciliationDetailList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPaymentReconciliationDetailList.SetItemByIndex(index: Integer; value: TFhirPaymentReconciliationDetail);
begin
  assert(value is TFhirPaymentReconciliationDetail);
  FhirPaymentReconciliationDetails[index] := value;
end;

procedure TFhirPaymentReconciliationDetailList.SetItemN(index: Integer; value: TFhirPaymentReconciliationDetail);
begin
  assert(value is TFhirPaymentReconciliationDetail);
  ObjectByIndex[index] := value;
end;

{ TFhirPaymentReconciliationNote }

constructor TFhirPaymentReconciliationNote.Create;
begin
  inherited;
end;

destructor TFhirPaymentReconciliationNote.Destroy;
begin
  FType_.free;
  FText.free;
  inherited;
end;

procedure TFhirPaymentReconciliationNote.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirPaymentReconciliationNote(oSource).type_.Clone;
  textElement := TFhirPaymentReconciliationNote(oSource).textElement.Clone;
end;

procedure TFhirPaymentReconciliationNote.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
end;

procedure TFhirPaymentReconciliationNote.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
end;

procedure TFhirPaymentReconciliationNote.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirPaymentReconciliationNote.FhirType : string;
begin
  result := 'note';
end;

function TFhirPaymentReconciliationNote.Link : TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote(inherited Link);
end;

function TFhirPaymentReconciliationNote.Clone : TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote(inherited Clone);
end;

function TFhirPaymentReconciliationNote.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPaymentReconciliationNote;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPaymentReconciliationNote)) then
    result := false
  else
  begin
    o := TFhirPaymentReconciliationNote(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(textElement, o.textElement, true);
  end;
end;

function TFhirPaymentReconciliationNote.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPaymentReconciliationNote;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPaymentReconciliationNote)) then
    result := false
  else
  begin
    o := TFhirPaymentReconciliationNote(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirPaymentReconciliationNote }

Procedure TFhirPaymentReconciliationNote.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirPaymentReconciliationNote.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirPaymentReconciliationNote.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirPaymentReconciliationNote.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;


{ TFhirPaymentReconciliationNoteListEnumerator }

Constructor TFhirPaymentReconciliationNoteListEnumerator.Create(list : TFhirPaymentReconciliationNoteList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPaymentReconciliationNoteListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPaymentReconciliationNoteListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPaymentReconciliationNoteListEnumerator.GetCurrent : TFhirPaymentReconciliationNote;
begin
  Result := FList[FIndex];
end;


{ TFhirPaymentReconciliationNoteList }
procedure TFhirPaymentReconciliationNoteList.AddItem(value: TFhirPaymentReconciliationNote);
begin
  assert(value.ClassName = 'TFhirPaymentReconciliationNote', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPaymentReconciliationNote');
  add(value);
end;


function TFhirPaymentReconciliationNoteList.Append: TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationNoteList.ClearItems;
begin
  Clear;
end;

function TFhirPaymentReconciliationNoteList.GetEnumerator : TFhirPaymentReconciliationNoteListEnumerator;
begin
  result := TFhirPaymentReconciliationNoteListEnumerator.Create(self.link);
end;

function TFhirPaymentReconciliationNoteList.Clone: TFhirPaymentReconciliationNoteList;
begin
  result := TFhirPaymentReconciliationNoteList(inherited Clone);
end;

function TFhirPaymentReconciliationNoteList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPaymentReconciliationNoteList.GetItemN(index: Integer): TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationNoteList.IndexOf(value: TFhirPaymentReconciliationNote): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPaymentReconciliationNoteList.Insert(index: Integer): TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationNoteList.InsertItem(index: Integer; value: TFhirPaymentReconciliationNote);
begin
  assert(value is TFhirPaymentReconciliationNote);
  Inherited Insert(index, value);
end;

function TFhirPaymentReconciliationNoteList.Item(index: Integer): TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationNoteList.Link: TFhirPaymentReconciliationNoteList;
begin
  result := TFhirPaymentReconciliationNoteList(inherited Link);
end;

procedure TFhirPaymentReconciliationNoteList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPaymentReconciliationNoteList.SetItemByIndex(index: Integer; value: TFhirPaymentReconciliationNote);
begin
  assert(value is TFhirPaymentReconciliationNote);
  FhirPaymentReconciliationNotes[index] := value;
end;

procedure TFhirPaymentReconciliationNoteList.SetItemN(index: Integer; value: TFhirPaymentReconciliationNote);
begin
  assert(value is TFhirPaymentReconciliationNote);
  ObjectByIndex[index] := value;
end;

{ TFhirPaymentReconciliation }

constructor TFhirPaymentReconciliation.Create;
begin
  inherited;
end;

destructor TFhirPaymentReconciliation.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FOutcome.free;
  FDisposition.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FPeriod.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  FDetailList.Free;
  FForm.free;
  FTotal.free;
  FNoteList.Free;
  inherited;
end;

function TFhirPaymentReconciliation.GetResourceType : TFhirResourceType;
begin
  result := frtPaymentReconciliation;
end;

procedure TFhirPaymentReconciliation.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPaymentReconciliation(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPaymentReconciliation(oSource).FIdentifierList);
  end;
  request := TFhirPaymentReconciliation(oSource).request.Clone;
  FOutcome := TFhirPaymentReconciliation(oSource).FOutcome.Link;
  dispositionElement := TFhirPaymentReconciliation(oSource).dispositionElement.Clone;
  ruleset := TFhirPaymentReconciliation(oSource).ruleset.Clone;
  originalRuleset := TFhirPaymentReconciliation(oSource).originalRuleset.Clone;
  createdElement := TFhirPaymentReconciliation(oSource).createdElement.Clone;
  period := TFhirPaymentReconciliation(oSource).period.Clone;
  organization := TFhirPaymentReconciliation(oSource).organization.Clone;
  requestProvider := TFhirPaymentReconciliation(oSource).requestProvider.Clone;
  requestOrganization := TFhirPaymentReconciliation(oSource).requestOrganization.Clone;
  if (TFhirPaymentReconciliation(oSource).FDetailList = nil) then
  begin
    FDetailList.free;
    FDetailList := nil;
  end
  else
  begin
    FDetailList := TFhirPaymentReconciliationDetailList.Create;
    FDetailList.Assign(TFhirPaymentReconciliation(oSource).FDetailList);
  end;
  form := TFhirPaymentReconciliation(oSource).form.Clone;
  total := TFhirPaymentReconciliation(oSource).total.Clone;
  if (TFhirPaymentReconciliation(oSource).FNoteList = nil) then
  begin
    FNoteList.free;
    FNoteList := nil;
  end
  else
  begin
    FNoteList := TFhirPaymentReconciliationNoteList.Create;
    FNoteList.Assign(TFhirPaymentReconciliation(oSource).FNoteList);
  end;
end;

procedure TFhirPaymentReconciliation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
  if (child_name = 'detail') Then
    list.addAll(FDetailList);
  if (child_name = 'form') Then
     list.add(FForm.Link);
  if (child_name = 'total') Then
     list.add(FTotal.Link);
  if (child_name = 'note') Then
    list.addAll(FNoteList);
end;

procedure TFhirPaymentReconciliation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(ProcessRequest)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'code', FOutcome.Link));{1}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'detail', '', FDetailList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'form', 'Coding', FForm.Link));{2}
  oList.add(TFHIRProperty.create(self, 'total', 'Money', FTotal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'note', '', FNoteList.Link)){3};
end;

procedure TFhirPaymentReconciliation.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirProcessRequest}{4b}
  else if (propName = 'outcome') then OutcomeElement := propValue as TFHIREnum
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'detail') then DetailList.add(propValue as TFhirPaymentReconciliationDetail){2}
  else if (propName = 'form') then Form := propValue as TFhirCoding{4b}
  else if (propName = 'total') then Total := propValue as TFhirQuantity{4b}
  else if (propName = 'note') then NoteList.add(propValue as TFhirPaymentReconciliationNote){2}
  else inherited;
end;

function TFhirPaymentReconciliation.FhirType : string;
begin
  result := 'PaymentReconciliation';
end;

function TFhirPaymentReconciliation.Link : TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation(inherited Link);
end;

function TFhirPaymentReconciliation.Clone : TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation(inherited Clone);
end;

{ TFhirPaymentReconciliation }

Function TFhirPaymentReconciliation.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPaymentReconciliation.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirPaymentReconciliation.SetRequest(value : TFhirReference{TFhirProcessRequest});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirPaymentReconciliation.SetOutcome(value : TFhirEnum);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirPaymentReconciliation.GetOutcomeST : TFhirRemittanceOutcome;
begin
  if FOutcome = nil then
    result := TFhirRemittanceOutcome(0)
  else
    result := TFhirRemittanceOutcome(StringArrayIndexOfSensitive(CODES_TFhirRemittanceOutcome, FOutcome.value));
end;

Procedure TFhirPaymentReconciliation.SetOutcomeST(value : TFhirRemittanceOutcome);
begin
  if ord(value) = 0 then
    OutcomeElement := nil
  else
    OutcomeElement := TFhirEnum.create(CODES_TFhirRemittanceOutcome[value]);
end;

Procedure TFhirPaymentReconciliation.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirPaymentReconciliation.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirPaymentReconciliation.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirPaymentReconciliation.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirPaymentReconciliation.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirPaymentReconciliation.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirPaymentReconciliation.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirPaymentReconciliation.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirPaymentReconciliation.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirPaymentReconciliation.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirPaymentReconciliation.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirPaymentReconciliation.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;

Function TFhirPaymentReconciliation.GetDetailList : TFhirPaymentReconciliationDetailList;
begin
  if FDetailList = nil then
    FDetailList := TFhirPaymentReconciliationDetailList.Create;
  result := FDetailList;
end;

Function TFhirPaymentReconciliation.GetHasDetailList : boolean;
begin
  result := (FDetailList <> nil) and (FDetailList.count > 0);
end;

Procedure TFhirPaymentReconciliation.SetForm(value : TFhirCoding);
begin
  FForm.free;
  FForm := value;
end;

Procedure TFhirPaymentReconciliation.SetTotal(value : TFhirQuantity);
begin
  FTotal.free;
  FTotal := value;
end;

Function TFhirPaymentReconciliation.GetNoteList : TFhirPaymentReconciliationNoteList;
begin
  if FNoteList = nil then
    FNoteList := TFhirPaymentReconciliationNoteList.Create;
  result := FNoteList;
end;

Function TFhirPaymentReconciliation.GetHasNoteList : boolean;
begin
  result := (FNoteList <> nil) and (FNoteList.count > 0);
end;


{ TFhirPaymentReconciliationListEnumerator }

Constructor TFhirPaymentReconciliationListEnumerator.Create(list : TFhirPaymentReconciliationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPaymentReconciliationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPaymentReconciliationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPaymentReconciliationListEnumerator.GetCurrent : TFhirPaymentReconciliation;
begin
  Result := FList[FIndex];
end;


{ TFhirPaymentReconciliationList }
procedure TFhirPaymentReconciliationList.AddItem(value: TFhirPaymentReconciliation);
begin
  assert(value.ClassName = 'TFhirPaymentReconciliation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPaymentReconciliation');
  add(value);
end;


function TFhirPaymentReconciliationList.Append: TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationList.ClearItems;
begin
  Clear;
end;

function TFhirPaymentReconciliationList.GetEnumerator : TFhirPaymentReconciliationListEnumerator;
begin
  result := TFhirPaymentReconciliationListEnumerator.Create(self.link);
end;

function TFhirPaymentReconciliationList.Clone: TFhirPaymentReconciliationList;
begin
  result := TFhirPaymentReconciliationList(inherited Clone);
end;

function TFhirPaymentReconciliationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPaymentReconciliationList.GetItemN(index: Integer): TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationList.IndexOf(value: TFhirPaymentReconciliation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPaymentReconciliationList.Insert(index: Integer): TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPaymentReconciliationList.InsertItem(index: Integer; value: TFhirPaymentReconciliation);
begin
  assert(value is TFhirPaymentReconciliation);
  Inherited Insert(index, value);
end;

function TFhirPaymentReconciliationList.Item(index: Integer): TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation(ObjectByIndex[index]);
end;

function TFhirPaymentReconciliationList.Link: TFhirPaymentReconciliationList;
begin
  result := TFhirPaymentReconciliationList(inherited Link);
end;

procedure TFhirPaymentReconciliationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPaymentReconciliationList.SetItemByIndex(index: Integer; value: TFhirPaymentReconciliation);
begin
  assert(value is TFhirPaymentReconciliation);
  FhirPaymentReconciliations[index] := value;
end;

procedure TFhirPaymentReconciliationList.SetItemN(index: Integer; value: TFhirPaymentReconciliation);
begin
  assert(value is TFhirPaymentReconciliation);
  ObjectByIndex[index] := value;
end;

{ TFhirPersonLink }

constructor TFhirPersonLink.Create;
begin
  inherited;
end;

destructor TFhirPersonLink.Destroy;
begin
  FTarget.free;
  FAssurance.free;
  inherited;
end;

procedure TFhirPersonLink.Assign(oSource : TAdvObject);
begin
  inherited;
  target := TFhirPersonLink(oSource).target.Clone;
  FAssurance := TFhirPersonLink(oSource).FAssurance.Link;
end;

procedure TFhirPersonLink.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'assurance') Then
     list.add(FAssurance.Link);
end;

procedure TFhirPersonLink.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Patient|Practitioner|RelatedPerson|Person)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'assurance', 'code', FAssurance.Link));{1}
end;

procedure TFhirPersonLink.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'target') then Target := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'assurance') then AssuranceElement := propValue as TFHIREnum
  else inherited;
end;

function TFhirPersonLink.FhirType : string;
begin
  result := 'link';
end;

function TFhirPersonLink.Link : TFhirPersonLink;
begin
  result := TFhirPersonLink(inherited Link);
end;

function TFhirPersonLink.Clone : TFhirPersonLink;
begin
  result := TFhirPersonLink(inherited Clone);
end;

function TFhirPersonLink.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPersonLink;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPersonLink)) then
    result := false
  else
  begin
    o := TFhirPersonLink(other);
    result := compareDeep(targetElement, o.targetElement, true) and compareDeep(assuranceElement, o.assuranceElement, true);
  end;
end;

function TFhirPersonLink.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPersonLink;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPersonLink)) then
    result := false
  else
  begin
    o := TFhirPersonLink(other);
    result := compareValues(assuranceElement, o.assuranceElement, true);
  end;
end;

{ TFhirPersonLink }

Procedure TFhirPersonLink.SetTarget(value : TFhirReference{Resource});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirPersonLink.SetAssurance(value : TFhirEnum);
begin
  FAssurance.free;
  FAssurance := value;
end;

Function TFhirPersonLink.GetAssuranceST : TFhirIdentityAssuranceLevel;
begin
  if FAssurance = nil then
    result := TFhirIdentityAssuranceLevel(0)
  else
    result := TFhirIdentityAssuranceLevel(StringArrayIndexOfSensitive(CODES_TFhirIdentityAssuranceLevel, FAssurance.value));
end;

Procedure TFhirPersonLink.SetAssuranceST(value : TFhirIdentityAssuranceLevel);
begin
  if ord(value) = 0 then
    AssuranceElement := nil
  else
    AssuranceElement := TFhirEnum.create(CODES_TFhirIdentityAssuranceLevel[value]);
end;


{ TFhirPersonLinkListEnumerator }

Constructor TFhirPersonLinkListEnumerator.Create(list : TFhirPersonLinkList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPersonLinkListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPersonLinkListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPersonLinkListEnumerator.GetCurrent : TFhirPersonLink;
begin
  Result := FList[FIndex];
end;


{ TFhirPersonLinkList }
procedure TFhirPersonLinkList.AddItem(value: TFhirPersonLink);
begin
  assert(value.ClassName = 'TFhirPersonLink', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPersonLink');
  add(value);
end;


function TFhirPersonLinkList.Append: TFhirPersonLink;
begin
  result := TFhirPersonLink.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPersonLinkList.ClearItems;
begin
  Clear;
end;

function TFhirPersonLinkList.GetEnumerator : TFhirPersonLinkListEnumerator;
begin
  result := TFhirPersonLinkListEnumerator.Create(self.link);
end;

function TFhirPersonLinkList.Clone: TFhirPersonLinkList;
begin
  result := TFhirPersonLinkList(inherited Clone);
end;

function TFhirPersonLinkList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPersonLinkList.GetItemN(index: Integer): TFhirPersonLink;
begin
  result := TFhirPersonLink(ObjectByIndex[index]);
end;

function TFhirPersonLinkList.IndexOf(value: TFhirPersonLink): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPersonLinkList.Insert(index: Integer): TFhirPersonLink;
begin
  result := TFhirPersonLink.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPersonLinkList.InsertItem(index: Integer; value: TFhirPersonLink);
begin
  assert(value is TFhirPersonLink);
  Inherited Insert(index, value);
end;

function TFhirPersonLinkList.Item(index: Integer): TFhirPersonLink;
begin
  result := TFhirPersonLink(ObjectByIndex[index]);
end;

function TFhirPersonLinkList.Link: TFhirPersonLinkList;
begin
  result := TFhirPersonLinkList(inherited Link);
end;

procedure TFhirPersonLinkList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPersonLinkList.SetItemByIndex(index: Integer; value: TFhirPersonLink);
begin
  assert(value is TFhirPersonLink);
  FhirPersonLinks[index] := value;
end;

procedure TFhirPersonLinkList.SetItemN(index: Integer; value: TFhirPersonLink);
begin
  assert(value is TFhirPersonLink);
  ObjectByIndex[index] := value;
end;

{ TFhirPerson }

constructor TFhirPerson.Create;
begin
  inherited;
end;

destructor TFhirPerson.Destroy;
begin
  FIdentifierList.Free;
  FNameList.Free;
  FTelecomList.Free;
  FGender.free;
  FBirthDate.free;
  FAddressList.Free;
  FPhoto.free;
  FManagingOrganization.free;
  FActive.free;
  FLink_List.Free;
  inherited;
end;

function TFhirPerson.GetResourceType : TFhirResourceType;
begin
  result := frtPerson;
end;

procedure TFhirPerson.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPerson(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPerson(oSource).FIdentifierList);
  end;
  if (TFhirPerson(oSource).FNameList = nil) then
  begin
    FNameList.free;
    FNameList := nil;
  end
  else
  begin
    FNameList := TFhirHumanNameList.Create;
    FNameList.Assign(TFhirPerson(oSource).FNameList);
  end;
  if (TFhirPerson(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirPerson(oSource).FTelecomList);
  end;
  FGender := TFhirPerson(oSource).FGender.Link;
  birthDateElement := TFhirPerson(oSource).birthDateElement.Clone;
  if (TFhirPerson(oSource).FAddressList = nil) then
  begin
    FAddressList.free;
    FAddressList := nil;
  end
  else
  begin
    FAddressList := TFhirAddressList.Create;
    FAddressList.Assign(TFhirPerson(oSource).FAddressList);
  end;
  photo := TFhirPerson(oSource).photo.Clone;
  managingOrganization := TFhirPerson(oSource).managingOrganization.Clone;
  activeElement := TFhirPerson(oSource).activeElement.Clone;
  if (TFhirPerson(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirPersonLinkList.Create;
    FLink_List.Assign(TFhirPerson(oSource).FLink_List);
  end;
end;

procedure TFhirPerson.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'name') Then
    list.addAll(FNameList);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'birthDate') Then
     list.add(FBirthDate.Link);
  if (child_name = 'address') Then
    list.addAll(FAddressList);
  if (child_name = 'photo') Then
     list.add(FPhoto.Link);
  if (child_name = 'managingOrganization') Then
     list.add(FManagingOrganization.Link);
  if (child_name = 'active') Then
     list.add(FActive.Link);
  if (child_name = 'link') Then
    list.addAll(FLink_List);
end;

procedure TFhirPerson.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FNameList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'birthDate', 'date', FBirthDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'photo', 'Attachment', FPhoto.Link));{2}
  oList.add(TFHIRProperty.create(self, 'managingOrganization', 'Reference(Organization)', FManagingOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'active', 'boolean', FActive.Link));{2}
  oList.add(TFHIRProperty.create(self, 'link', '', FLink_List.Link)){3};
end;

procedure TFhirPerson.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'name') then NameList.add(propValue as TFhirHumanName){2}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName = 'birthDate') then BirthDateElement := propValue as TFhirDate{5a}
  else if (propName = 'address') then AddressList.add(propValue as TFhirAddress){2}
  else if (propName = 'photo') then Photo := propValue as TFhirAttachment{4b}
  else if (propName = 'managingOrganization') then ManagingOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'active') then ActiveElement := propValue as TFhirBoolean{5a}
  else if (propName = 'link') then Link_List.add(propValue as TFhirPersonLink){2}
  else inherited;
end;

function TFhirPerson.FhirType : string;
begin
  result := 'Person';
end;

function TFhirPerson.Link : TFhirPerson;
begin
  result := TFhirPerson(inherited Link);
end;

function TFhirPerson.Clone : TFhirPerson;
begin
  result := TFhirPerson(inherited Clone);
end;

{ TFhirPerson }

Function TFhirPerson.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPerson.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirPerson.GetNameList : TFhirHumanNameList;
begin
  if FNameList = nil then
    FNameList := TFhirHumanNameList.Create;
  result := FNameList;
end;

Function TFhirPerson.GetHasNameList : boolean;
begin
  result := (FNameList <> nil) and (FNameList.count > 0);
end;

Function TFhirPerson.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirPerson.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirPerson.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirPerson.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirPerson.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirPerson.SetBirthDate(value : TFhirDate);
begin
  FBirthDate.free;
  FBirthDate := value;
end;

Function TFhirPerson.GetBirthDateST : TDateAndTime;
begin
  if FBirthDate = nil then
    result := nil
  else
    result := FBirthDate.value;
end;

Procedure TFhirPerson.SetBirthDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FBirthDate = nil then
      FBirthDate := TFhirDate.create;
    FBirthDate.value := value
  end
  else if FBirthDate <> nil then
    FBirthDate.value := nil;
end;

Function TFhirPerson.GetAddressList : TFhirAddressList;
begin
  if FAddressList = nil then
    FAddressList := TFhirAddressList.Create;
  result := FAddressList;
end;

Function TFhirPerson.GetHasAddressList : boolean;
begin
  result := (FAddressList <> nil) and (FAddressList.count > 0);
end;

Procedure TFhirPerson.SetPhoto(value : TFhirAttachment);
begin
  FPhoto.free;
  FPhoto := value;
end;

Procedure TFhirPerson.SetManagingOrganization(value : TFhirReference{TFhirOrganization});
begin
  FManagingOrganization.free;
  FManagingOrganization := value;
end;

Procedure TFhirPerson.SetActive(value : TFhirBoolean);
begin
  FActive.free;
  FActive := value;
end;

Function TFhirPerson.GetActiveST : Boolean;
begin
  if FActive = nil then
    result := false
  else
    result := FActive.value;
end;

Procedure TFhirPerson.SetActiveST(value : Boolean);
begin
  if FActive = nil then
    FActive := TFhirBoolean.create;
  FActive.value := value
end;

Function TFhirPerson.GetLink_List : TFhirPersonLinkList;
begin
  if FLink_List = nil then
    FLink_List := TFhirPersonLinkList.Create;
  result := FLink_List;
end;

Function TFhirPerson.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;


{ TFhirPersonListEnumerator }

Constructor TFhirPersonListEnumerator.Create(list : TFhirPersonList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPersonListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPersonListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPersonListEnumerator.GetCurrent : TFhirPerson;
begin
  Result := FList[FIndex];
end;


{ TFhirPersonList }
procedure TFhirPersonList.AddItem(value: TFhirPerson);
begin
  assert(value.ClassName = 'TFhirPerson', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPerson');
  add(value);
end;


function TFhirPersonList.Append: TFhirPerson;
begin
  result := TFhirPerson.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPersonList.ClearItems;
begin
  Clear;
end;

function TFhirPersonList.GetEnumerator : TFhirPersonListEnumerator;
begin
  result := TFhirPersonListEnumerator.Create(self.link);
end;

function TFhirPersonList.Clone: TFhirPersonList;
begin
  result := TFhirPersonList(inherited Clone);
end;

function TFhirPersonList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPersonList.GetItemN(index: Integer): TFhirPerson;
begin
  result := TFhirPerson(ObjectByIndex[index]);
end;

function TFhirPersonList.IndexOf(value: TFhirPerson): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPersonList.Insert(index: Integer): TFhirPerson;
begin
  result := TFhirPerson.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPersonList.InsertItem(index: Integer; value: TFhirPerson);
begin
  assert(value is TFhirPerson);
  Inherited Insert(index, value);
end;

function TFhirPersonList.Item(index: Integer): TFhirPerson;
begin
  result := TFhirPerson(ObjectByIndex[index]);
end;

function TFhirPersonList.Link: TFhirPersonList;
begin
  result := TFhirPersonList(inherited Link);
end;

procedure TFhirPersonList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPersonList.SetItemByIndex(index: Integer; value: TFhirPerson);
begin
  assert(value is TFhirPerson);
  FhirPeople[index] := value;
end;

procedure TFhirPersonList.SetItemN(index: Integer; value: TFhirPerson);
begin
  assert(value is TFhirPerson);
  ObjectByIndex[index] := value;
end;

{ TFhirPractitionerPractitionerRole }

constructor TFhirPractitionerPractitionerRole.Create;
begin
  inherited;
end;

destructor TFhirPractitionerPractitionerRole.Destroy;
begin
  FManagingOrganization.free;
  FRole.free;
  FSpecialtyList.Free;
  FPeriod.free;
  FLocationList.Free;
  FHealthcareServiceList.Free;
  inherited;
end;

procedure TFhirPractitionerPractitionerRole.Assign(oSource : TAdvObject);
begin
  inherited;
  managingOrganization := TFhirPractitionerPractitionerRole(oSource).managingOrganization.Clone;
  role := TFhirPractitionerPractitionerRole(oSource).role.Clone;
  if (TFhirPractitionerPractitionerRole(oSource).FSpecialtyList = nil) then
  begin
    FSpecialtyList.free;
    FSpecialtyList := nil;
  end
  else
  begin
    FSpecialtyList := TFhirCodeableConceptList.Create;
    FSpecialtyList.Assign(TFhirPractitionerPractitionerRole(oSource).FSpecialtyList);
  end;
  period := TFhirPractitionerPractitionerRole(oSource).period.Clone;
  if (TFhirPractitionerPractitionerRole(oSource).FLocationList = nil) then
  begin
    FLocationList.free;
    FLocationList := nil;
  end
  else
  begin
    FLocationList := TFhirReferenceList{TFhirLocation}.Create;
    FLocationList.Assign(TFhirPractitionerPractitionerRole(oSource).FLocationList);
  end;
  if (TFhirPractitionerPractitionerRole(oSource).FHealthcareServiceList = nil) then
  begin
    FHealthcareServiceList.free;
    FHealthcareServiceList := nil;
  end
  else
  begin
    FHealthcareServiceList := TFhirReferenceList{TFhirHealthcareService}.Create;
    FHealthcareServiceList.Assign(TFhirPractitionerPractitionerRole(oSource).FHealthcareServiceList);
  end;
end;

procedure TFhirPractitionerPractitionerRole.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'managingOrganization') Then
     list.add(FManagingOrganization.Link);
  if (child_name = 'role') Then
     list.add(FRole.Link);
  if (child_name = 'specialty') Then
    list.addAll(FSpecialtyList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'location') Then
    list.addAll(FLocationList);
  if (child_name = 'healthcareService') Then
    list.addAll(FHealthcareServiceList);
end;

procedure TFhirPractitionerPractitionerRole.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'managingOrganization', 'Reference(Organization)', FManagingOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRole.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specialty', 'CodeableConcept', FSpecialtyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'healthcareService', 'Reference(HealthcareService)', FHealthcareServiceList.Link)){3};
end;

procedure TFhirPractitionerPractitionerRole.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'managingOrganization') then ManagingOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'role') then Role := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'specialty') then SpecialtyList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'location') then LocationList.add(propValue as TFhirReference{TFhirLocation}){2}
  else if (propName = 'healthcareService') then HealthcareServiceList.add(propValue as TFhirReference{TFhirHealthcareService}){2}
  else inherited;
end;

function TFhirPractitionerPractitionerRole.FhirType : string;
begin
  result := 'practitionerRole';
end;

function TFhirPractitionerPractitionerRole.Link : TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole(inherited Link);
end;

function TFhirPractitionerPractitionerRole.Clone : TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole(inherited Clone);
end;

function TFhirPractitionerPractitionerRole.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPractitionerPractitionerRole;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPractitionerPractitionerRole)) then
    result := false
  else
  begin
    o := TFhirPractitionerPractitionerRole(other);
    result := compareDeep(managingOrganizationElement, o.managingOrganizationElement, true) and 
      compareDeep(roleElement, o.roleElement, true) and compareDeep(specialtyList, o.specialtyList, true) and 
      compareDeep(periodElement, o.periodElement, true) and compareDeep(locationList, o.locationList, true) and 
      compareDeep(healthcareServiceList, o.healthcareServiceList, true);
  end;
end;

function TFhirPractitionerPractitionerRole.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPractitionerPractitionerRole;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPractitionerPractitionerRole)) then
    result := false
  else
  begin
    o := TFhirPractitionerPractitionerRole(other);
    result := true;
  end;
end;

{ TFhirPractitionerPractitionerRole }

Procedure TFhirPractitionerPractitionerRole.SetManagingOrganization(value : TFhirReference{TFhirOrganization});
begin
  FManagingOrganization.free;
  FManagingOrganization := value;
end;

Procedure TFhirPractitionerPractitionerRole.SetRole(value : TFhirCodeableConcept);
begin
  FRole.free;
  FRole := value;
end;

Function TFhirPractitionerPractitionerRole.GetSpecialtyList : TFhirCodeableConceptList;
begin
  if FSpecialtyList = nil then
    FSpecialtyList := TFhirCodeableConceptList.Create;
  result := FSpecialtyList;
end;

Function TFhirPractitionerPractitionerRole.GetHasSpecialtyList : boolean;
begin
  result := (FSpecialtyList <> nil) and (FSpecialtyList.count > 0);
end;

Procedure TFhirPractitionerPractitionerRole.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Function TFhirPractitionerPractitionerRole.GetLocationList : TFhirReferenceList{TFhirLocation};
begin
  if FLocationList = nil then
    FLocationList := TFhirReferenceList{TFhirLocation}.Create;
  result := FLocationList;
end;

Function TFhirPractitionerPractitionerRole.GetHasLocationList : boolean;
begin
  result := (FLocationList <> nil) and (FLocationList.count > 0);
end;

Function TFhirPractitionerPractitionerRole.GetHealthcareServiceList : TFhirReferenceList{TFhirHealthcareService};
begin
  if FHealthcareServiceList = nil then
    FHealthcareServiceList := TFhirReferenceList{TFhirHealthcareService}.Create;
  result := FHealthcareServiceList;
end;

Function TFhirPractitionerPractitionerRole.GetHasHealthcareServiceList : boolean;
begin
  result := (FHealthcareServiceList <> nil) and (FHealthcareServiceList.count > 0);
end;


{ TFhirPractitionerPractitionerRoleListEnumerator }

Constructor TFhirPractitionerPractitionerRoleListEnumerator.Create(list : TFhirPractitionerPractitionerRoleList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPractitionerPractitionerRoleListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPractitionerPractitionerRoleListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPractitionerPractitionerRoleListEnumerator.GetCurrent : TFhirPractitionerPractitionerRole;
begin
  Result := FList[FIndex];
end;


{ TFhirPractitionerPractitionerRoleList }
procedure TFhirPractitionerPractitionerRoleList.AddItem(value: TFhirPractitionerPractitionerRole);
begin
  assert(value.ClassName = 'TFhirPractitionerPractitionerRole', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPractitionerPractitionerRole');
  add(value);
end;


function TFhirPractitionerPractitionerRoleList.Append: TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerPractitionerRoleList.ClearItems;
begin
  Clear;
end;

function TFhirPractitionerPractitionerRoleList.GetEnumerator : TFhirPractitionerPractitionerRoleListEnumerator;
begin
  result := TFhirPractitionerPractitionerRoleListEnumerator.Create(self.link);
end;

function TFhirPractitionerPractitionerRoleList.Clone: TFhirPractitionerPractitionerRoleList;
begin
  result := TFhirPractitionerPractitionerRoleList(inherited Clone);
end;

function TFhirPractitionerPractitionerRoleList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPractitionerPractitionerRoleList.GetItemN(index: Integer): TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole(ObjectByIndex[index]);
end;

function TFhirPractitionerPractitionerRoleList.IndexOf(value: TFhirPractitionerPractitionerRole): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPractitionerPractitionerRoleList.Insert(index: Integer): TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerPractitionerRoleList.InsertItem(index: Integer; value: TFhirPractitionerPractitionerRole);
begin
  assert(value is TFhirPractitionerPractitionerRole);
  Inherited Insert(index, value);
end;

function TFhirPractitionerPractitionerRoleList.Item(index: Integer): TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole(ObjectByIndex[index]);
end;

function TFhirPractitionerPractitionerRoleList.Link: TFhirPractitionerPractitionerRoleList;
begin
  result := TFhirPractitionerPractitionerRoleList(inherited Link);
end;

procedure TFhirPractitionerPractitionerRoleList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPractitionerPractitionerRoleList.SetItemByIndex(index: Integer; value: TFhirPractitionerPractitionerRole);
begin
  assert(value is TFhirPractitionerPractitionerRole);
  FhirPractitionerPractitionerRoles[index] := value;
end;

procedure TFhirPractitionerPractitionerRoleList.SetItemN(index: Integer; value: TFhirPractitionerPractitionerRole);
begin
  assert(value is TFhirPractitionerPractitionerRole);
  ObjectByIndex[index] := value;
end;

{ TFhirPractitionerQualification }

constructor TFhirPractitionerQualification.Create;
begin
  inherited;
end;

destructor TFhirPractitionerQualification.Destroy;
begin
  FIdentifierList.Free;
  FCode.free;
  FPeriod.free;
  FIssuer.free;
  inherited;
end;

procedure TFhirPractitionerQualification.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPractitionerQualification(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPractitionerQualification(oSource).FIdentifierList);
  end;
  code := TFhirPractitionerQualification(oSource).code.Clone;
  period := TFhirPractitionerQualification(oSource).period.Clone;
  issuer := TFhirPractitionerQualification(oSource).issuer.Clone;
end;

procedure TFhirPractitionerQualification.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'issuer') Then
     list.add(FIssuer.Link);
end;

procedure TFhirPractitionerQualification.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'issuer', 'Reference(Organization)', FIssuer.Link));{2}
end;

procedure TFhirPractitionerQualification.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'issuer') then Issuer := propValue as TFhirReference{TFhirOrganization}{4b}
  else inherited;
end;

function TFhirPractitionerQualification.FhirType : string;
begin
  result := 'qualification';
end;

function TFhirPractitionerQualification.Link : TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification(inherited Link);
end;

function TFhirPractitionerQualification.Clone : TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification(inherited Clone);
end;

function TFhirPractitionerQualification.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirPractitionerQualification;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirPractitionerQualification)) then
    result := false
  else
  begin
    o := TFhirPractitionerQualification(other);
    result := compareDeep(identifierList, o.identifierList, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(periodElement, o.periodElement, true) and compareDeep(issuerElement, o.issuerElement, true);
  end;
end;

function TFhirPractitionerQualification.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirPractitionerQualification;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirPractitionerQualification)) then
    result := false
  else
  begin
    o := TFhirPractitionerQualification(other);
    result := true;
  end;
end;

{ TFhirPractitionerQualification }

Function TFhirPractitionerQualification.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPractitionerQualification.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirPractitionerQualification.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirPractitionerQualification.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirPractitionerQualification.SetIssuer(value : TFhirReference{TFhirOrganization});
begin
  FIssuer.free;
  FIssuer := value;
end;


{ TFhirPractitionerQualificationListEnumerator }

Constructor TFhirPractitionerQualificationListEnumerator.Create(list : TFhirPractitionerQualificationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPractitionerQualificationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPractitionerQualificationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPractitionerQualificationListEnumerator.GetCurrent : TFhirPractitionerQualification;
begin
  Result := FList[FIndex];
end;


{ TFhirPractitionerQualificationList }
procedure TFhirPractitionerQualificationList.AddItem(value: TFhirPractitionerQualification);
begin
  assert(value.ClassName = 'TFhirPractitionerQualification', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPractitionerQualification');
  add(value);
end;


function TFhirPractitionerQualificationList.Append: TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerQualificationList.ClearItems;
begin
  Clear;
end;

function TFhirPractitionerQualificationList.GetEnumerator : TFhirPractitionerQualificationListEnumerator;
begin
  result := TFhirPractitionerQualificationListEnumerator.Create(self.link);
end;

function TFhirPractitionerQualificationList.Clone: TFhirPractitionerQualificationList;
begin
  result := TFhirPractitionerQualificationList(inherited Clone);
end;

function TFhirPractitionerQualificationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPractitionerQualificationList.GetItemN(index: Integer): TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification(ObjectByIndex[index]);
end;

function TFhirPractitionerQualificationList.IndexOf(value: TFhirPractitionerQualification): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPractitionerQualificationList.Insert(index: Integer): TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerQualificationList.InsertItem(index: Integer; value: TFhirPractitionerQualification);
begin
  assert(value is TFhirPractitionerQualification);
  Inherited Insert(index, value);
end;

function TFhirPractitionerQualificationList.Item(index: Integer): TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification(ObjectByIndex[index]);
end;

function TFhirPractitionerQualificationList.Link: TFhirPractitionerQualificationList;
begin
  result := TFhirPractitionerQualificationList(inherited Link);
end;

procedure TFhirPractitionerQualificationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPractitionerQualificationList.SetItemByIndex(index: Integer; value: TFhirPractitionerQualification);
begin
  assert(value is TFhirPractitionerQualification);
  FhirPractitionerQualifications[index] := value;
end;

procedure TFhirPractitionerQualificationList.SetItemN(index: Integer; value: TFhirPractitionerQualification);
begin
  assert(value is TFhirPractitionerQualification);
  ObjectByIndex[index] := value;
end;

{ TFhirPractitioner }

constructor TFhirPractitioner.Create;
begin
  inherited;
end;

destructor TFhirPractitioner.Destroy;
begin
  FIdentifierList.Free;
  FActive.free;
  FName.free;
  FTelecomList.Free;
  FAddressList.Free;
  FGender.free;
  FBirthDate.free;
  FPhotoList.Free;
  FPractitionerRoleList.Free;
  FQualificationList.Free;
  FCommunicationList.Free;
  inherited;
end;

function TFhirPractitioner.GetResourceType : TFhirResourceType;
begin
  result := frtPractitioner;
end;

procedure TFhirPractitioner.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirPractitioner(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirPractitioner(oSource).FIdentifierList);
  end;
  activeElement := TFhirPractitioner(oSource).activeElement.Clone;
  name := TFhirPractitioner(oSource).name.Clone;
  if (TFhirPractitioner(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirPractitioner(oSource).FTelecomList);
  end;
  if (TFhirPractitioner(oSource).FAddressList = nil) then
  begin
    FAddressList.free;
    FAddressList := nil;
  end
  else
  begin
    FAddressList := TFhirAddressList.Create;
    FAddressList.Assign(TFhirPractitioner(oSource).FAddressList);
  end;
  FGender := TFhirPractitioner(oSource).FGender.Link;
  birthDateElement := TFhirPractitioner(oSource).birthDateElement.Clone;
  if (TFhirPractitioner(oSource).FPhotoList = nil) then
  begin
    FPhotoList.free;
    FPhotoList := nil;
  end
  else
  begin
    FPhotoList := TFhirAttachmentList.Create;
    FPhotoList.Assign(TFhirPractitioner(oSource).FPhotoList);
  end;
  if (TFhirPractitioner(oSource).FPractitionerRoleList = nil) then
  begin
    FPractitionerRoleList.free;
    FPractitionerRoleList := nil;
  end
  else
  begin
    FPractitionerRoleList := TFhirPractitionerPractitionerRoleList.Create;
    FPractitionerRoleList.Assign(TFhirPractitioner(oSource).FPractitionerRoleList);
  end;
  if (TFhirPractitioner(oSource).FQualificationList = nil) then
  begin
    FQualificationList.free;
    FQualificationList := nil;
  end
  else
  begin
    FQualificationList := TFhirPractitionerQualificationList.Create;
    FQualificationList.Assign(TFhirPractitioner(oSource).FQualificationList);
  end;
  if (TFhirPractitioner(oSource).FCommunicationList = nil) then
  begin
    FCommunicationList.free;
    FCommunicationList := nil;
  end
  else
  begin
    FCommunicationList := TFhirCodeableConceptList.Create;
    FCommunicationList.Assign(TFhirPractitioner(oSource).FCommunicationList);
  end;
end;

procedure TFhirPractitioner.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'active') Then
     list.add(FActive.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'address') Then
    list.addAll(FAddressList);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'birthDate') Then
     list.add(FBirthDate.Link);
  if (child_name = 'photo') Then
    list.addAll(FPhotoList);
  if (child_name = 'practitionerRole') Then
    list.addAll(FPractitionerRoleList);
  if (child_name = 'qualification') Then
    list.addAll(FQualificationList);
  if (child_name = 'communication') Then
    list.addAll(FCommunicationList);
end;

procedure TFhirPractitioner.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'active', 'boolean', FActive.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'birthDate', 'date', FBirthDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'photo', 'Attachment', FPhotoList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'practitionerRole', '', FPractitionerRoleList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'qualification', '', FQualificationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'communication', 'CodeableConcept', FCommunicationList.Link)){3};
end;

procedure TFhirPractitioner.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'active') then ActiveElement := propValue as TFhirBoolean{5a}
  else if (propName = 'name') then Name := propValue as TFhirHumanName{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'address') then AddressList.add(propValue as TFhirAddress){2}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName = 'birthDate') then BirthDateElement := propValue as TFhirDate{5a}
  else if (propName = 'photo') then PhotoList.add(propValue as TFhirAttachment){2}
  else if (propName = 'practitionerRole') then PractitionerRoleList.add(propValue as TFhirPractitionerPractitionerRole){2}
  else if (propName = 'qualification') then QualificationList.add(propValue as TFhirPractitionerQualification){2}
  else if (propName = 'communication') then CommunicationList.add(propValue as TFhirCodeableConcept){2}
  else inherited;
end;

function TFhirPractitioner.FhirType : string;
begin
  result := 'Practitioner';
end;

function TFhirPractitioner.Link : TFhirPractitioner;
begin
  result := TFhirPractitioner(inherited Link);
end;

function TFhirPractitioner.Clone : TFhirPractitioner;
begin
  result := TFhirPractitioner(inherited Clone);
end;

{ TFhirPractitioner }

Function TFhirPractitioner.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirPractitioner.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirPractitioner.SetActive(value : TFhirBoolean);
begin
  FActive.free;
  FActive := value;
end;

Function TFhirPractitioner.GetActiveST : Boolean;
begin
  if FActive = nil then
    result := false
  else
    result := FActive.value;
end;

Procedure TFhirPractitioner.SetActiveST(value : Boolean);
begin
  if FActive = nil then
    FActive := TFhirBoolean.create;
  FActive.value := value
end;

Procedure TFhirPractitioner.SetName(value : TFhirHumanName);
begin
  FName.free;
  FName := value;
end;

Function TFhirPractitioner.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirPractitioner.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Function TFhirPractitioner.GetAddressList : TFhirAddressList;
begin
  if FAddressList = nil then
    FAddressList := TFhirAddressList.Create;
  result := FAddressList;
end;

Function TFhirPractitioner.GetHasAddressList : boolean;
begin
  result := (FAddressList <> nil) and (FAddressList.count > 0);
end;

Procedure TFhirPractitioner.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirPractitioner.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirPractitioner.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirPractitioner.SetBirthDate(value : TFhirDate);
begin
  FBirthDate.free;
  FBirthDate := value;
end;

Function TFhirPractitioner.GetBirthDateST : TDateAndTime;
begin
  if FBirthDate = nil then
    result := nil
  else
    result := FBirthDate.value;
end;

Procedure TFhirPractitioner.SetBirthDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FBirthDate = nil then
      FBirthDate := TFhirDate.create;
    FBirthDate.value := value
  end
  else if FBirthDate <> nil then
    FBirthDate.value := nil;
end;

Function TFhirPractitioner.GetPhotoList : TFhirAttachmentList;
begin
  if FPhotoList = nil then
    FPhotoList := TFhirAttachmentList.Create;
  result := FPhotoList;
end;

Function TFhirPractitioner.GetHasPhotoList : boolean;
begin
  result := (FPhotoList <> nil) and (FPhotoList.count > 0);
end;

Function TFhirPractitioner.GetPractitionerRoleList : TFhirPractitionerPractitionerRoleList;
begin
  if FPractitionerRoleList = nil then
    FPractitionerRoleList := TFhirPractitionerPractitionerRoleList.Create;
  result := FPractitionerRoleList;
end;

Function TFhirPractitioner.GetHasPractitionerRoleList : boolean;
begin
  result := (FPractitionerRoleList <> nil) and (FPractitionerRoleList.count > 0);
end;

Function TFhirPractitioner.GetQualificationList : TFhirPractitionerQualificationList;
begin
  if FQualificationList = nil then
    FQualificationList := TFhirPractitionerQualificationList.Create;
  result := FQualificationList;
end;

Function TFhirPractitioner.GetHasQualificationList : boolean;
begin
  result := (FQualificationList <> nil) and (FQualificationList.count > 0);
end;

Function TFhirPractitioner.GetCommunicationList : TFhirCodeableConceptList;
begin
  if FCommunicationList = nil then
    FCommunicationList := TFhirCodeableConceptList.Create;
  result := FCommunicationList;
end;

Function TFhirPractitioner.GetHasCommunicationList : boolean;
begin
  result := (FCommunicationList <> nil) and (FCommunicationList.count > 0);
end;


{ TFhirPractitionerListEnumerator }

Constructor TFhirPractitionerListEnumerator.Create(list : TFhirPractitionerList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirPractitionerListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirPractitionerListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirPractitionerListEnumerator.GetCurrent : TFhirPractitioner;
begin
  Result := FList[FIndex];
end;


{ TFhirPractitionerList }
procedure TFhirPractitionerList.AddItem(value: TFhirPractitioner);
begin
  assert(value.ClassName = 'TFhirPractitioner', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirPractitioner');
  add(value);
end;


function TFhirPractitionerList.Append: TFhirPractitioner;
begin
  result := TFhirPractitioner.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerList.ClearItems;
begin
  Clear;
end;

function TFhirPractitionerList.GetEnumerator : TFhirPractitionerListEnumerator;
begin
  result := TFhirPractitionerListEnumerator.Create(self.link);
end;

function TFhirPractitionerList.Clone: TFhirPractitionerList;
begin
  result := TFhirPractitionerList(inherited Clone);
end;

function TFhirPractitionerList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirPractitionerList.GetItemN(index: Integer): TFhirPractitioner;
begin
  result := TFhirPractitioner(ObjectByIndex[index]);
end;

function TFhirPractitionerList.IndexOf(value: TFhirPractitioner): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirPractitionerList.Insert(index: Integer): TFhirPractitioner;
begin
  result := TFhirPractitioner.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirPractitionerList.InsertItem(index: Integer; value: TFhirPractitioner);
begin
  assert(value is TFhirPractitioner);
  Inherited Insert(index, value);
end;

function TFhirPractitionerList.Item(index: Integer): TFhirPractitioner;
begin
  result := TFhirPractitioner(ObjectByIndex[index]);
end;

function TFhirPractitionerList.Link: TFhirPractitionerList;
begin
  result := TFhirPractitionerList(inherited Link);
end;

procedure TFhirPractitionerList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirPractitionerList.SetItemByIndex(index: Integer; value: TFhirPractitioner);
begin
  assert(value is TFhirPractitioner);
  FhirPractitioners[index] := value;
end;

procedure TFhirPractitionerList.SetItemN(index: Integer; value: TFhirPractitioner);
begin
  assert(value is TFhirPractitioner);
  ObjectByIndex[index] := value;
end;

{ TFhirProcedurePerformer }

constructor TFhirProcedurePerformer.Create;
begin
  inherited;
end;

destructor TFhirProcedurePerformer.Destroy;
begin
  FActor.free;
  FRole.free;
  inherited;
end;

procedure TFhirProcedurePerformer.Assign(oSource : TAdvObject);
begin
  inherited;
  actor := TFhirProcedurePerformer(oSource).actor.Clone;
  role := TFhirProcedurePerformer(oSource).role.Clone;
end;

procedure TFhirProcedurePerformer.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'actor') Then
     list.add(FActor.Link);
  if (child_name = 'role') Then
     list.add(FRole.Link);
end;

procedure TFhirProcedurePerformer.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Practitioner|Organization|Patient|RelatedPerson)', FActor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'role', 'CodeableConcept', FRole.Link));{2}
end;

procedure TFhirProcedurePerformer.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'role') then Role := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirProcedurePerformer.FhirType : string;
begin
  result := 'performer';
end;

function TFhirProcedurePerformer.Link : TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer(inherited Link);
end;

function TFhirProcedurePerformer.Clone : TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer(inherited Clone);
end;

function TFhirProcedurePerformer.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProcedurePerformer;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProcedurePerformer)) then
    result := false
  else
  begin
    o := TFhirProcedurePerformer(other);
    result := compareDeep(actorElement, o.actorElement, true) and compareDeep(roleElement, o.roleElement, true);
  end;
end;

function TFhirProcedurePerformer.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProcedurePerformer;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProcedurePerformer)) then
    result := false
  else
  begin
    o := TFhirProcedurePerformer(other);
    result := true;
  end;
end;

{ TFhirProcedurePerformer }

Procedure TFhirProcedurePerformer.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;

Procedure TFhirProcedurePerformer.SetRole(value : TFhirCodeableConcept);
begin
  FRole.free;
  FRole := value;
end;


{ TFhirProcedurePerformerListEnumerator }

Constructor TFhirProcedurePerformerListEnumerator.Create(list : TFhirProcedurePerformerList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcedurePerformerListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcedurePerformerListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcedurePerformerListEnumerator.GetCurrent : TFhirProcedurePerformer;
begin
  Result := FList[FIndex];
end;


{ TFhirProcedurePerformerList }
procedure TFhirProcedurePerformerList.AddItem(value: TFhirProcedurePerformer);
begin
  assert(value.ClassName = 'TFhirProcedurePerformer', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcedurePerformer');
  add(value);
end;


function TFhirProcedurePerformerList.Append: TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedurePerformerList.ClearItems;
begin
  Clear;
end;

function TFhirProcedurePerformerList.GetEnumerator : TFhirProcedurePerformerListEnumerator;
begin
  result := TFhirProcedurePerformerListEnumerator.Create(self.link);
end;

function TFhirProcedurePerformerList.Clone: TFhirProcedurePerformerList;
begin
  result := TFhirProcedurePerformerList(inherited Clone);
end;

function TFhirProcedurePerformerList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcedurePerformerList.GetItemN(index: Integer): TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer(ObjectByIndex[index]);
end;

function TFhirProcedurePerformerList.IndexOf(value: TFhirProcedurePerformer): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcedurePerformerList.Insert(index: Integer): TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedurePerformerList.InsertItem(index: Integer; value: TFhirProcedurePerformer);
begin
  assert(value is TFhirProcedurePerformer);
  Inherited Insert(index, value);
end;

function TFhirProcedurePerformerList.Item(index: Integer): TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer(ObjectByIndex[index]);
end;

function TFhirProcedurePerformerList.Link: TFhirProcedurePerformerList;
begin
  result := TFhirProcedurePerformerList(inherited Link);
end;

procedure TFhirProcedurePerformerList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcedurePerformerList.SetItemByIndex(index: Integer; value: TFhirProcedurePerformer);
begin
  assert(value is TFhirProcedurePerformer);
  FhirProcedurePerformers[index] := value;
end;

procedure TFhirProcedurePerformerList.SetItemN(index: Integer; value: TFhirProcedurePerformer);
begin
  assert(value is TFhirProcedurePerformer);
  ObjectByIndex[index] := value;
end;

{ TFhirProcedureFocalDevice }

constructor TFhirProcedureFocalDevice.Create;
begin
  inherited;
end;

destructor TFhirProcedureFocalDevice.Destroy;
begin
  FAction.free;
  FManipulated.free;
  inherited;
end;

procedure TFhirProcedureFocalDevice.Assign(oSource : TAdvObject);
begin
  inherited;
  action := TFhirProcedureFocalDevice(oSource).action.Clone;
  manipulated := TFhirProcedureFocalDevice(oSource).manipulated.Clone;
end;

procedure TFhirProcedureFocalDevice.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'action') Then
     list.add(FAction.Link);
  if (child_name = 'manipulated') Then
     list.add(FManipulated.Link);
end;

procedure TFhirProcedureFocalDevice.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'action', 'CodeableConcept', FAction.Link));{2}
  oList.add(TFHIRProperty.create(self, 'manipulated', 'Reference(Device)', FManipulated.Link));{2}
end;

procedure TFhirProcedureFocalDevice.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'action') then Action := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'manipulated') then Manipulated := propValue as TFhirReference{TFhirDevice}{4b}
  else inherited;
end;

function TFhirProcedureFocalDevice.FhirType : string;
begin
  result := 'focalDevice';
end;

function TFhirProcedureFocalDevice.Link : TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice(inherited Link);
end;

function TFhirProcedureFocalDevice.Clone : TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice(inherited Clone);
end;

function TFhirProcedureFocalDevice.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProcedureFocalDevice;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProcedureFocalDevice)) then
    result := false
  else
  begin
    o := TFhirProcedureFocalDevice(other);
    result := compareDeep(actionElement, o.actionElement, true) and compareDeep(manipulatedElement, o.manipulatedElement, true);
  end;
end;

function TFhirProcedureFocalDevice.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProcedureFocalDevice;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProcedureFocalDevice)) then
    result := false
  else
  begin
    o := TFhirProcedureFocalDevice(other);
    result := true;
  end;
end;

{ TFhirProcedureFocalDevice }

Procedure TFhirProcedureFocalDevice.SetAction(value : TFhirCodeableConcept);
begin
  FAction.free;
  FAction := value;
end;

Procedure TFhirProcedureFocalDevice.SetManipulated(value : TFhirReference{TFhirDevice});
begin
  FManipulated.free;
  FManipulated := value;
end;


{ TFhirProcedureFocalDeviceListEnumerator }

Constructor TFhirProcedureFocalDeviceListEnumerator.Create(list : TFhirProcedureFocalDeviceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcedureFocalDeviceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcedureFocalDeviceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcedureFocalDeviceListEnumerator.GetCurrent : TFhirProcedureFocalDevice;
begin
  Result := FList[FIndex];
end;


{ TFhirProcedureFocalDeviceList }
procedure TFhirProcedureFocalDeviceList.AddItem(value: TFhirProcedureFocalDevice);
begin
  assert(value.ClassName = 'TFhirProcedureFocalDevice', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcedureFocalDevice');
  add(value);
end;


function TFhirProcedureFocalDeviceList.Append: TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureFocalDeviceList.ClearItems;
begin
  Clear;
end;

function TFhirProcedureFocalDeviceList.GetEnumerator : TFhirProcedureFocalDeviceListEnumerator;
begin
  result := TFhirProcedureFocalDeviceListEnumerator.Create(self.link);
end;

function TFhirProcedureFocalDeviceList.Clone: TFhirProcedureFocalDeviceList;
begin
  result := TFhirProcedureFocalDeviceList(inherited Clone);
end;

function TFhirProcedureFocalDeviceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcedureFocalDeviceList.GetItemN(index: Integer): TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice(ObjectByIndex[index]);
end;

function TFhirProcedureFocalDeviceList.IndexOf(value: TFhirProcedureFocalDevice): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcedureFocalDeviceList.Insert(index: Integer): TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureFocalDeviceList.InsertItem(index: Integer; value: TFhirProcedureFocalDevice);
begin
  assert(value is TFhirProcedureFocalDevice);
  Inherited Insert(index, value);
end;

function TFhirProcedureFocalDeviceList.Item(index: Integer): TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice(ObjectByIndex[index]);
end;

function TFhirProcedureFocalDeviceList.Link: TFhirProcedureFocalDeviceList;
begin
  result := TFhirProcedureFocalDeviceList(inherited Link);
end;

procedure TFhirProcedureFocalDeviceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcedureFocalDeviceList.SetItemByIndex(index: Integer; value: TFhirProcedureFocalDevice);
begin
  assert(value is TFhirProcedureFocalDevice);
  FhirProcedureFocalDevices[index] := value;
end;

procedure TFhirProcedureFocalDeviceList.SetItemN(index: Integer; value: TFhirProcedureFocalDevice);
begin
  assert(value is TFhirProcedureFocalDevice);
  ObjectByIndex[index] := value;
end;

{ TFhirProcedure }

constructor TFhirProcedure.Create;
begin
  inherited;
end;

destructor TFhirProcedure.Destroy;
begin
  FIdentifierList.Free;
  FSubject.free;
  FStatus.free;
  FCategory.free;
  FCode.free;
  FNotPerformed.free;
  FReasonNotPerformedList.Free;
  FBodySiteList.Free;
  FReason.free;
  FPerformerList.Free;
  FPerformed.free;
  FEncounter.free;
  FLocation.free;
  FOutcome.free;
  FReportList.Free;
  FComplicationList.Free;
  FFollowUpList.Free;
  FRequest.free;
  FNotesList.Free;
  FFocalDeviceList.Free;
  FUsedList.Free;
  inherited;
end;

function TFhirProcedure.GetResourceType : TFhirResourceType;
begin
  result := frtProcedure;
end;

procedure TFhirProcedure.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirProcedure(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirProcedure(oSource).FIdentifierList);
  end;
  subject := TFhirProcedure(oSource).subject.Clone;
  FStatus := TFhirProcedure(oSource).FStatus.Link;
  category := TFhirProcedure(oSource).category.Clone;
  code := TFhirProcedure(oSource).code.Clone;
  notPerformedElement := TFhirProcedure(oSource).notPerformedElement.Clone;
  if (TFhirProcedure(oSource).FReasonNotPerformedList = nil) then
  begin
    FReasonNotPerformedList.free;
    FReasonNotPerformedList := nil;
  end
  else
  begin
    FReasonNotPerformedList := TFhirCodeableConceptList.Create;
    FReasonNotPerformedList.Assign(TFhirProcedure(oSource).FReasonNotPerformedList);
  end;
  if (TFhirProcedure(oSource).FBodySiteList = nil) then
  begin
    FBodySiteList.free;
    FBodySiteList := nil;
  end
  else
  begin
    FBodySiteList := TFhirCodeableConceptList.Create;
    FBodySiteList.Assign(TFhirProcedure(oSource).FBodySiteList);
  end;
  reason := TFhirProcedure(oSource).reason.Clone;
  if (TFhirProcedure(oSource).FPerformerList = nil) then
  begin
    FPerformerList.free;
    FPerformerList := nil;
  end
  else
  begin
    FPerformerList := TFhirProcedurePerformerList.Create;
    FPerformerList.Assign(TFhirProcedure(oSource).FPerformerList);
  end;
  performed := TFhirProcedure(oSource).performed.Clone;
  encounter := TFhirProcedure(oSource).encounter.Clone;
  location := TFhirProcedure(oSource).location.Clone;
  outcome := TFhirProcedure(oSource).outcome.Clone;
  if (TFhirProcedure(oSource).FReportList = nil) then
  begin
    FReportList.free;
    FReportList := nil;
  end
  else
  begin
    FReportList := TFhirReferenceList{TFhirDiagnosticReport}.Create;
    FReportList.Assign(TFhirProcedure(oSource).FReportList);
  end;
  if (TFhirProcedure(oSource).FComplicationList = nil) then
  begin
    FComplicationList.free;
    FComplicationList := nil;
  end
  else
  begin
    FComplicationList := TFhirCodeableConceptList.Create;
    FComplicationList.Assign(TFhirProcedure(oSource).FComplicationList);
  end;
  if (TFhirProcedure(oSource).FFollowUpList = nil) then
  begin
    FFollowUpList.free;
    FFollowUpList := nil;
  end
  else
  begin
    FFollowUpList := TFhirCodeableConceptList.Create;
    FFollowUpList.Assign(TFhirProcedure(oSource).FFollowUpList);
  end;
  request := TFhirProcedure(oSource).request.Clone;
  if (TFhirProcedure(oSource).FNotesList = nil) then
  begin
    FNotesList.free;
    FNotesList := nil;
  end
  else
  begin
    FNotesList := TFhirAnnotationList.Create;
    FNotesList.Assign(TFhirProcedure(oSource).FNotesList);
  end;
  if (TFhirProcedure(oSource).FFocalDeviceList = nil) then
  begin
    FFocalDeviceList.free;
    FFocalDeviceList := nil;
  end
  else
  begin
    FFocalDeviceList := TFhirProcedureFocalDeviceList.Create;
    FFocalDeviceList.Assign(TFhirProcedure(oSource).FFocalDeviceList);
  end;
  if (TFhirProcedure(oSource).FUsedList = nil) then
  begin
    FUsedList.free;
    FUsedList := nil;
  end
  else
  begin
    FUsedList := TFhirReferenceList{Resource}.Create;
    FUsedList.Assign(TFhirProcedure(oSource).FUsedList);
  end;
end;

procedure TFhirProcedure.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'category') Then
     list.add(FCategory.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'notPerformed') Then
     list.add(FNotPerformed.Link);
  if (child_name = 'reasonNotPerformed') Then
    list.addAll(FReasonNotPerformedList);
  if (child_name = 'bodySite') Then
    list.addAll(FBodySiteList);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'performer') Then
    list.addAll(FPerformerList);
  if (child_name = 'performed[x]') Then
     list.add(FPerformed.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'report') Then
    list.addAll(FReportList);
  if (child_name = 'complication') Then
    list.addAll(FComplicationList);
  if (child_name = 'followUp') Then
    list.addAll(FFollowUpList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'notes') Then
    list.addAll(FNotesList);
  if (child_name = 'focalDevice') Then
    list.addAll(FFocalDeviceList);
  if (child_name = 'used') Then
    list.addAll(FUsedList);
end;

procedure TFhirProcedure.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategory.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'notPerformed', 'boolean', FNotPerformed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reasonNotPerformed', 'CodeableConcept', FReasonNotPerformedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySiteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Condition)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', '', FPerformerList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'performed[x]', 'dateTime|Period', FPerformed.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'CodeableConcept', FOutcome.Link));{2}
  oList.add(TFHIRProperty.create(self, 'report', 'Reference(DiagnosticReport)', FReportList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'complication', 'CodeableConcept', FComplicationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'followUp', 'CodeableConcept', FFollowUpList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(CarePlan|DiagnosticOrder|ProcedureRequest|ReferralRequest)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'notes', 'Annotation', FNotesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'focalDevice', '', FFocalDeviceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'used', 'Reference(Device|Medication|Substance)', FUsedList.Link)){3};
end;

procedure TFhirProcedure.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'category') then Category := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'notPerformed') then NotPerformedElement := propValue as TFhirBoolean{5a}
  else if (propName = 'reasonNotPerformed') then ReasonNotPerformedList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'bodySite') then BodySiteList.add(propValue as TFhirCodeableConcept){2}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName = 'performer') then PerformerList.add(propValue as TFhirProcedurePerformer){2}
  else if (propName.startsWith('performed')) then Performed := propValue as TFhirType{4}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'outcome') then Outcome := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'report') then ReportList.add(propValue as TFhirReference{TFhirDiagnosticReport}){2}
  else if (propName = 'complication') then ComplicationList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'followUp') then FollowUpList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'notes') then NotesList.add(propValue as TFhirAnnotation){2}
  else if (propName = 'focalDevice') then FocalDeviceList.add(propValue as TFhirProcedureFocalDevice){2}
  else if (propName = 'used') then UsedList.add(propValue as TFhirReference{Resource}){2}
  else inherited;
end;

function TFhirProcedure.FhirType : string;
begin
  result := 'Procedure';
end;

function TFhirProcedure.Link : TFhirProcedure;
begin
  result := TFhirProcedure(inherited Link);
end;

function TFhirProcedure.Clone : TFhirProcedure;
begin
  result := TFhirProcedure(inherited Clone);
end;

{ TFhirProcedure }

Function TFhirProcedure.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirProcedure.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirProcedure.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirProcedure.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirProcedure.GetStatusST : TFhirProcedureStatus;
begin
  if FStatus = nil then
    result := TFhirProcedureStatus(0)
  else
    result := TFhirProcedureStatus(StringArrayIndexOfSensitive(CODES_TFhirProcedureStatus, FStatus.value));
end;

Procedure TFhirProcedure.SetStatusST(value : TFhirProcedureStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirProcedureStatus[value]);
end;

Procedure TFhirProcedure.SetCategory(value : TFhirCodeableConcept);
begin
  FCategory.free;
  FCategory := value;
end;

Procedure TFhirProcedure.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirProcedure.SetNotPerformed(value : TFhirBoolean);
begin
  FNotPerformed.free;
  FNotPerformed := value;
end;

Function TFhirProcedure.GetNotPerformedST : Boolean;
begin
  if FNotPerformed = nil then
    result := false
  else
    result := FNotPerformed.value;
end;

Procedure TFhirProcedure.SetNotPerformedST(value : Boolean);
begin
  if FNotPerformed = nil then
    FNotPerformed := TFhirBoolean.create;
  FNotPerformed.value := value
end;

Function TFhirProcedure.GetReasonNotPerformedList : TFhirCodeableConceptList;
begin
  if FReasonNotPerformedList = nil then
    FReasonNotPerformedList := TFhirCodeableConceptList.Create;
  result := FReasonNotPerformedList;
end;

Function TFhirProcedure.GetHasReasonNotPerformedList : boolean;
begin
  result := (FReasonNotPerformedList <> nil) and (FReasonNotPerformedList.count > 0);
end;

Function TFhirProcedure.GetBodySiteList : TFhirCodeableConceptList;
begin
  if FBodySiteList = nil then
    FBodySiteList := TFhirCodeableConceptList.Create;
  result := FBodySiteList;
end;

Function TFhirProcedure.GetHasBodySiteList : boolean;
begin
  result := (FBodySiteList <> nil) and (FBodySiteList.count > 0);
end;

Procedure TFhirProcedure.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Function TFhirProcedure.GetPerformerList : TFhirProcedurePerformerList;
begin
  if FPerformerList = nil then
    FPerformerList := TFhirProcedurePerformerList.Create;
  result := FPerformerList;
end;

Function TFhirProcedure.GetHasPerformerList : boolean;
begin
  result := (FPerformerList <> nil) and (FPerformerList.count > 0);
end;

Procedure TFhirProcedure.SetPerformed(value : TFhirType);
begin
  FPerformed.free;
  FPerformed := value;
end;

Procedure TFhirProcedure.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirProcedure.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Procedure TFhirProcedure.SetOutcome(value : TFhirCodeableConcept);
begin
  FOutcome.free;
  FOutcome := value;
end;

Function TFhirProcedure.GetReportList : TFhirReferenceList{TFhirDiagnosticReport};
begin
  if FReportList = nil then
    FReportList := TFhirReferenceList{TFhirDiagnosticReport}.Create;
  result := FReportList;
end;

Function TFhirProcedure.GetHasReportList : boolean;
begin
  result := (FReportList <> nil) and (FReportList.count > 0);
end;

Function TFhirProcedure.GetComplicationList : TFhirCodeableConceptList;
begin
  if FComplicationList = nil then
    FComplicationList := TFhirCodeableConceptList.Create;
  result := FComplicationList;
end;

Function TFhirProcedure.GetHasComplicationList : boolean;
begin
  result := (FComplicationList <> nil) and (FComplicationList.count > 0);
end;

Function TFhirProcedure.GetFollowUpList : TFhirCodeableConceptList;
begin
  if FFollowUpList = nil then
    FFollowUpList := TFhirCodeableConceptList.Create;
  result := FFollowUpList;
end;

Function TFhirProcedure.GetHasFollowUpList : boolean;
begin
  result := (FFollowUpList <> nil) and (FFollowUpList.count > 0);
end;

Procedure TFhirProcedure.SetRequest(value : TFhirReference{Resource});
begin
  FRequest.free;
  FRequest := value;
end;

Function TFhirProcedure.GetNotesList : TFhirAnnotationList;
begin
  if FNotesList = nil then
    FNotesList := TFhirAnnotationList.Create;
  result := FNotesList;
end;

Function TFhirProcedure.GetHasNotesList : boolean;
begin
  result := (FNotesList <> nil) and (FNotesList.count > 0);
end;

Function TFhirProcedure.GetFocalDeviceList : TFhirProcedureFocalDeviceList;
begin
  if FFocalDeviceList = nil then
    FFocalDeviceList := TFhirProcedureFocalDeviceList.Create;
  result := FFocalDeviceList;
end;

Function TFhirProcedure.GetHasFocalDeviceList : boolean;
begin
  result := (FFocalDeviceList <> nil) and (FFocalDeviceList.count > 0);
end;

Function TFhirProcedure.GetUsedList : TFhirReferenceList{Resource};
begin
  if FUsedList = nil then
    FUsedList := TFhirReferenceList{Resource}.Create;
  result := FUsedList;
end;

Function TFhirProcedure.GetHasUsedList : boolean;
begin
  result := (FUsedList <> nil) and (FUsedList.count > 0);
end;


{ TFhirProcedureListEnumerator }

Constructor TFhirProcedureListEnumerator.Create(list : TFhirProcedureList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcedureListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcedureListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcedureListEnumerator.GetCurrent : TFhirProcedure;
begin
  Result := FList[FIndex];
end;


{ TFhirProcedureList }
procedure TFhirProcedureList.AddItem(value: TFhirProcedure);
begin
  assert(value.ClassName = 'TFhirProcedure', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcedure');
  add(value);
end;


function TFhirProcedureList.Append: TFhirProcedure;
begin
  result := TFhirProcedure.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureList.ClearItems;
begin
  Clear;
end;

function TFhirProcedureList.GetEnumerator : TFhirProcedureListEnumerator;
begin
  result := TFhirProcedureListEnumerator.Create(self.link);
end;

function TFhirProcedureList.Clone: TFhirProcedureList;
begin
  result := TFhirProcedureList(inherited Clone);
end;

function TFhirProcedureList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcedureList.GetItemN(index: Integer): TFhirProcedure;
begin
  result := TFhirProcedure(ObjectByIndex[index]);
end;

function TFhirProcedureList.IndexOf(value: TFhirProcedure): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcedureList.Insert(index: Integer): TFhirProcedure;
begin
  result := TFhirProcedure.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureList.InsertItem(index: Integer; value: TFhirProcedure);
begin
  assert(value is TFhirProcedure);
  Inherited Insert(index, value);
end;

function TFhirProcedureList.Item(index: Integer): TFhirProcedure;
begin
  result := TFhirProcedure(ObjectByIndex[index]);
end;

function TFhirProcedureList.Link: TFhirProcedureList;
begin
  result := TFhirProcedureList(inherited Link);
end;

procedure TFhirProcedureList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcedureList.SetItemByIndex(index: Integer; value: TFhirProcedure);
begin
  assert(value is TFhirProcedure);
  FhirProcedures[index] := value;
end;

procedure TFhirProcedureList.SetItemN(index: Integer; value: TFhirProcedure);
begin
  assert(value is TFhirProcedure);
  ObjectByIndex[index] := value;
end;

{ TFhirProcedureRequest }

constructor TFhirProcedureRequest.Create;
begin
  inherited;
end;

destructor TFhirProcedureRequest.Destroy;
begin
  FIdentifierList.Free;
  FSubject.free;
  FCode.free;
  FBodySiteList.Free;
  FReason.free;
  FScheduled.free;
  FEncounter.free;
  FPerformer.free;
  FStatus.free;
  FNotesList.Free;
  FAsNeeded.free;
  FOrderedOn.free;
  FOrderer.free;
  FPriority.free;
  inherited;
end;

function TFhirProcedureRequest.GetResourceType : TFhirResourceType;
begin
  result := frtProcedureRequest;
end;

procedure TFhirProcedureRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirProcedureRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirProcedureRequest(oSource).FIdentifierList);
  end;
  subject := TFhirProcedureRequest(oSource).subject.Clone;
  code := TFhirProcedureRequest(oSource).code.Clone;
  if (TFhirProcedureRequest(oSource).FBodySiteList = nil) then
  begin
    FBodySiteList.free;
    FBodySiteList := nil;
  end
  else
  begin
    FBodySiteList := TFhirCodeableConceptList.Create;
    FBodySiteList.Assign(TFhirProcedureRequest(oSource).FBodySiteList);
  end;
  reason := TFhirProcedureRequest(oSource).reason.Clone;
  scheduled := TFhirProcedureRequest(oSource).scheduled.Clone;
  encounter := TFhirProcedureRequest(oSource).encounter.Clone;
  performer := TFhirProcedureRequest(oSource).performer.Clone;
  FStatus := TFhirProcedureRequest(oSource).FStatus.Link;
  if (TFhirProcedureRequest(oSource).FNotesList = nil) then
  begin
    FNotesList.free;
    FNotesList := nil;
  end
  else
  begin
    FNotesList := TFhirAnnotationList.Create;
    FNotesList.Assign(TFhirProcedureRequest(oSource).FNotesList);
  end;
  asNeeded := TFhirProcedureRequest(oSource).asNeeded.Clone;
  orderedOnElement := TFhirProcedureRequest(oSource).orderedOnElement.Clone;
  orderer := TFhirProcedureRequest(oSource).orderer.Clone;
  FPriority := TFhirProcedureRequest(oSource).FPriority.Link;
end;

procedure TFhirProcedureRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'bodySite') Then
    list.addAll(FBodySiteList);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'scheduled[x]') Then
     list.add(FScheduled.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'performer') Then
     list.add(FPerformer.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'notes') Then
    list.addAll(FNotesList);
  if (child_name = 'asNeeded[x]') Then
     list.add(FAsNeeded.Link);
  if (child_name = 'orderedOn') Then
     list.add(FOrderedOn.Link);
  if (child_name = 'orderer') Then
     list.add(FOrderer.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
end;

procedure TFhirProcedureRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySiteList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Condition)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'scheduled[x]', 'dateTime|Period|Timing', FScheduled.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner|Organization|Patient|RelatedPerson)', FPerformer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'notes', 'Annotation', FNotesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'asNeeded[x]', 'boolean|CodeableConcept', FAsNeeded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderedOn', 'dateTime', FOrderedOn.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderer', 'Reference(Practitioner|Patient|RelatedPerson|Device)', FOrderer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'code', FPriority.Link));{1}
end;

procedure TFhirProcedureRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'bodySite') then BodySiteList.add(propValue as TFhirCodeableConcept){2}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName.startsWith('scheduled')) then Scheduled := propValue as TFhirType{4}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'performer') then Performer := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'notes') then NotesList.add(propValue as TFhirAnnotation){2}
  else if (propName.startsWith('asNeeded')) then AsNeeded := propValue as TFhirType{4}
  else if (propName = 'orderedOn') then OrderedOnElement := propValue as TFhirDateTime{5a}
  else if (propName = 'orderer') then Orderer := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'priority') then PriorityElement := propValue as TFHIREnum
  else inherited;
end;

function TFhirProcedureRequest.FhirType : string;
begin
  result := 'ProcedureRequest';
end;

function TFhirProcedureRequest.Link : TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest(inherited Link);
end;

function TFhirProcedureRequest.Clone : TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest(inherited Clone);
end;

{ TFhirProcedureRequest }

Function TFhirProcedureRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirProcedureRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirProcedureRequest.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirProcedureRequest.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirProcedureRequest.GetBodySiteList : TFhirCodeableConceptList;
begin
  if FBodySiteList = nil then
    FBodySiteList := TFhirCodeableConceptList.Create;
  result := FBodySiteList;
end;

Function TFhirProcedureRequest.GetHasBodySiteList : boolean;
begin
  result := (FBodySiteList <> nil) and (FBodySiteList.count > 0);
end;

Procedure TFhirProcedureRequest.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirProcedureRequest.SetScheduled(value : TFhirType);
begin
  FScheduled.free;
  FScheduled := value;
end;

Procedure TFhirProcedureRequest.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirProcedureRequest.SetPerformer(value : TFhirReference{Resource});
begin
  FPerformer.free;
  FPerformer := value;
end;

Procedure TFhirProcedureRequest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirProcedureRequest.GetStatusST : TFhirProcedureRequestStatus;
begin
  if FStatus = nil then
    result := TFhirProcedureRequestStatus(0)
  else
    result := TFhirProcedureRequestStatus(StringArrayIndexOfSensitive(CODES_TFhirProcedureRequestStatus, FStatus.value));
end;

Procedure TFhirProcedureRequest.SetStatusST(value : TFhirProcedureRequestStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirProcedureRequestStatus[value]);
end;

Function TFhirProcedureRequest.GetNotesList : TFhirAnnotationList;
begin
  if FNotesList = nil then
    FNotesList := TFhirAnnotationList.Create;
  result := FNotesList;
end;

Function TFhirProcedureRequest.GetHasNotesList : boolean;
begin
  result := (FNotesList <> nil) and (FNotesList.count > 0);
end;

Procedure TFhirProcedureRequest.SetAsNeeded(value : TFhirType);
begin
  FAsNeeded.free;
  FAsNeeded := value;
end;

Procedure TFhirProcedureRequest.SetOrderedOn(value : TFhirDateTime);
begin
  FOrderedOn.free;
  FOrderedOn := value;
end;

Function TFhirProcedureRequest.GetOrderedOnST : TDateAndTime;
begin
  if FOrderedOn = nil then
    result := nil
  else
    result := FOrderedOn.value;
end;

Procedure TFhirProcedureRequest.SetOrderedOnST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FOrderedOn = nil then
      FOrderedOn := TFhirDateTime.create;
    FOrderedOn.value := value
  end
  else if FOrderedOn <> nil then
    FOrderedOn.value := nil;
end;

Procedure TFhirProcedureRequest.SetOrderer(value : TFhirReference{Resource});
begin
  FOrderer.free;
  FOrderer := value;
end;

Procedure TFhirProcedureRequest.SetPriority(value : TFhirEnum);
begin
  FPriority.free;
  FPriority := value;
end;

Function TFhirProcedureRequest.GetPriorityST : TFhirProcedureRequestPriority;
begin
  if FPriority = nil then
    result := TFhirProcedureRequestPriority(0)
  else
    result := TFhirProcedureRequestPriority(StringArrayIndexOfSensitive(CODES_TFhirProcedureRequestPriority, FPriority.value));
end;

Procedure TFhirProcedureRequest.SetPriorityST(value : TFhirProcedureRequestPriority);
begin
  if ord(value) = 0 then
    PriorityElement := nil
  else
    PriorityElement := TFhirEnum.create(CODES_TFhirProcedureRequestPriority[value]);
end;


{ TFhirProcedureRequestListEnumerator }

Constructor TFhirProcedureRequestListEnumerator.Create(list : TFhirProcedureRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcedureRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcedureRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcedureRequestListEnumerator.GetCurrent : TFhirProcedureRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirProcedureRequestList }
procedure TFhirProcedureRequestList.AddItem(value: TFhirProcedureRequest);
begin
  assert(value.ClassName = 'TFhirProcedureRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcedureRequest');
  add(value);
end;


function TFhirProcedureRequestList.Append: TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureRequestList.ClearItems;
begin
  Clear;
end;

function TFhirProcedureRequestList.GetEnumerator : TFhirProcedureRequestListEnumerator;
begin
  result := TFhirProcedureRequestListEnumerator.Create(self.link);
end;

function TFhirProcedureRequestList.Clone: TFhirProcedureRequestList;
begin
  result := TFhirProcedureRequestList(inherited Clone);
end;

function TFhirProcedureRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcedureRequestList.GetItemN(index: Integer): TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest(ObjectByIndex[index]);
end;

function TFhirProcedureRequestList.IndexOf(value: TFhirProcedureRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcedureRequestList.Insert(index: Integer): TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcedureRequestList.InsertItem(index: Integer; value: TFhirProcedureRequest);
begin
  assert(value is TFhirProcedureRequest);
  Inherited Insert(index, value);
end;

function TFhirProcedureRequestList.Item(index: Integer): TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest(ObjectByIndex[index]);
end;

function TFhirProcedureRequestList.Link: TFhirProcedureRequestList;
begin
  result := TFhirProcedureRequestList(inherited Link);
end;

procedure TFhirProcedureRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcedureRequestList.SetItemByIndex(index: Integer; value: TFhirProcedureRequest);
begin
  assert(value is TFhirProcedureRequest);
  FhirProcedureRequests[index] := value;
end;

procedure TFhirProcedureRequestList.SetItemN(index: Integer; value: TFhirProcedureRequest);
begin
  assert(value is TFhirProcedureRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirProcessRequestItem }

constructor TFhirProcessRequestItem.Create;
begin
  inherited;
end;

destructor TFhirProcessRequestItem.Destroy;
begin
  FSequenceLinkId.free;
  inherited;
end;

procedure TFhirProcessRequestItem.Assign(oSource : TAdvObject);
begin
  inherited;
  sequenceLinkIdElement := TFhirProcessRequestItem(oSource).sequenceLinkIdElement.Clone;
end;

procedure TFhirProcessRequestItem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'sequenceLinkId') Then
     list.add(FSequenceLinkId.Link);
end;

procedure TFhirProcessRequestItem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'sequenceLinkId', 'integer', FSequenceLinkId.Link));{2}
end;

procedure TFhirProcessRequestItem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'sequenceLinkId') then SequenceLinkIdElement := propValue as TFhirInteger{5a}
  else inherited;
end;

function TFhirProcessRequestItem.FhirType : string;
begin
  result := 'item';
end;

function TFhirProcessRequestItem.Link : TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem(inherited Link);
end;

function TFhirProcessRequestItem.Clone : TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem(inherited Clone);
end;

function TFhirProcessRequestItem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProcessRequestItem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProcessRequestItem)) then
    result := false
  else
  begin
    o := TFhirProcessRequestItem(other);
    result := compareDeep(sequenceLinkIdElement, o.sequenceLinkIdElement, true);
  end;
end;

function TFhirProcessRequestItem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProcessRequestItem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProcessRequestItem)) then
    result := false
  else
  begin
    o := TFhirProcessRequestItem(other);
    result := compareValues(sequenceLinkIdElement, o.sequenceLinkIdElement, true);
  end;
end;

{ TFhirProcessRequestItem }

Procedure TFhirProcessRequestItem.SetSequenceLinkId(value : TFhirInteger);
begin
  FSequenceLinkId.free;
  FSequenceLinkId := value;
end;

Function TFhirProcessRequestItem.GetSequenceLinkIdST : String;
begin
  if FSequenceLinkId = nil then
    result := ''
  else
    result := FSequenceLinkId.value;
end;

Procedure TFhirProcessRequestItem.SetSequenceLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FSequenceLinkId = nil then
      FSequenceLinkId := TFhirInteger.create;
    FSequenceLinkId.value := value
  end
  else if FSequenceLinkId <> nil then
    FSequenceLinkId.value := '';
end;


{ TFhirProcessRequestItemListEnumerator }

Constructor TFhirProcessRequestItemListEnumerator.Create(list : TFhirProcessRequestItemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcessRequestItemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcessRequestItemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcessRequestItemListEnumerator.GetCurrent : TFhirProcessRequestItem;
begin
  Result := FList[FIndex];
end;


{ TFhirProcessRequestItemList }
procedure TFhirProcessRequestItemList.AddItem(value: TFhirProcessRequestItem);
begin
  assert(value.ClassName = 'TFhirProcessRequestItem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcessRequestItem');
  add(value);
end;


function TFhirProcessRequestItemList.Append: TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessRequestItemList.ClearItems;
begin
  Clear;
end;

function TFhirProcessRequestItemList.GetEnumerator : TFhirProcessRequestItemListEnumerator;
begin
  result := TFhirProcessRequestItemListEnumerator.Create(self.link);
end;

function TFhirProcessRequestItemList.Clone: TFhirProcessRequestItemList;
begin
  result := TFhirProcessRequestItemList(inherited Clone);
end;

function TFhirProcessRequestItemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcessRequestItemList.GetItemN(index: Integer): TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem(ObjectByIndex[index]);
end;

function TFhirProcessRequestItemList.IndexOf(value: TFhirProcessRequestItem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcessRequestItemList.Insert(index: Integer): TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessRequestItemList.InsertItem(index: Integer; value: TFhirProcessRequestItem);
begin
  assert(value is TFhirProcessRequestItem);
  Inherited Insert(index, value);
end;

function TFhirProcessRequestItemList.Item(index: Integer): TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem(ObjectByIndex[index]);
end;

function TFhirProcessRequestItemList.Link: TFhirProcessRequestItemList;
begin
  result := TFhirProcessRequestItemList(inherited Link);
end;

procedure TFhirProcessRequestItemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcessRequestItemList.SetItemByIndex(index: Integer; value: TFhirProcessRequestItem);
begin
  assert(value is TFhirProcessRequestItem);
  FhirProcessRequestItems[index] := value;
end;

procedure TFhirProcessRequestItemList.SetItemN(index: Integer; value: TFhirProcessRequestItem);
begin
  assert(value is TFhirProcessRequestItem);
  ObjectByIndex[index] := value;
end;

{ TFhirProcessRequest }

constructor TFhirProcessRequest.Create;
begin
  inherited;
end;

destructor TFhirProcessRequest.Destroy;
begin
  FAction.free;
  FIdentifierList.Free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FTarget.free;
  FProvider.free;
  FOrganization.free;
  FRequest.free;
  FResponse.free;
  FNullify.free;
  FReference.free;
  FItemList.Free;
  FIncludeList.Free;
  FExcludeList.Free;
  FPeriod.free;
  inherited;
end;

function TFhirProcessRequest.GetResourceType : TFhirResourceType;
begin
  result := frtProcessRequest;
end;

procedure TFhirProcessRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  FAction := TFhirProcessRequest(oSource).FAction.Link;
  if (TFhirProcessRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirProcessRequest(oSource).FIdentifierList);
  end;
  ruleset := TFhirProcessRequest(oSource).ruleset.Clone;
  originalRuleset := TFhirProcessRequest(oSource).originalRuleset.Clone;
  createdElement := TFhirProcessRequest(oSource).createdElement.Clone;
  target := TFhirProcessRequest(oSource).target.Clone;
  provider := TFhirProcessRequest(oSource).provider.Clone;
  organization := TFhirProcessRequest(oSource).organization.Clone;
  request := TFhirProcessRequest(oSource).request.Clone;
  response := TFhirProcessRequest(oSource).response.Clone;
  nullifyElement := TFhirProcessRequest(oSource).nullifyElement.Clone;
  referenceElement := TFhirProcessRequest(oSource).referenceElement.Clone;
  if (TFhirProcessRequest(oSource).FItemList = nil) then
  begin
    FItemList.free;
    FItemList := nil;
  end
  else
  begin
    FItemList := TFhirProcessRequestItemList.Create;
    FItemList.Assign(TFhirProcessRequest(oSource).FItemList);
  end;
  if (TFhirProcessRequest(oSource).FIncludeList = nil) then
  begin
    FIncludeList.free;
    FIncludeList := nil;
  end
  else
  begin
    FIncludeList := TFhirStringList.Create;
    FIncludeList.Assign(TFhirProcessRequest(oSource).FIncludeList);
  end;
  if (TFhirProcessRequest(oSource).FExcludeList = nil) then
  begin
    FExcludeList.free;
    FExcludeList := nil;
  end
  else
  begin
    FExcludeList := TFhirStringList.Create;
    FExcludeList.Assign(TFhirProcessRequest(oSource).FExcludeList);
  end;
  period := TFhirProcessRequest(oSource).period.Clone;
end;

procedure TFhirProcessRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'action') Then
     list.add(FAction.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
  if (child_name = 'provider') Then
     list.add(FProvider.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
  if (child_name = 'nullify') Then
     list.add(FNullify.Link);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'item') Then
    list.addAll(FItemList);
  if (child_name = 'include') Then
    list.addAll(FIncludeList);
  if (child_name = 'exclude') Then
    list.addAll(FExcludeList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirProcessRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'action', 'code', FAction.Link));{1}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Organization)', FTarget.Link));{2}
  oList.add(TFHIRProperty.create(self, 'provider', 'Reference(Practitioner)', FProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Any)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', 'Reference(Any)', FResponse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'nullify', 'boolean', FNullify.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reference', 'string', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'item', '', FItemList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'include', 'string', FIncludeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'exclude', 'string', FExcludeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirProcessRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'action') then ActionElement := propValue as TFHIREnum
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'target') then Target := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'provider') then Provider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'response') then Response := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'nullify') then NullifyElement := propValue as TFhirBoolean{5a}
  else if (propName = 'reference') then ReferenceElement := propValue as TFhirString{5a}
  else if (propName = 'item') then ItemList.add(propValue as TFhirProcessRequestItem){2}
  else if (propName = 'include') then IncludeList.add(propValue as TFhirString){2}
  else if (propName = 'exclude') then ExcludeList.add(propValue as TFhirString){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirProcessRequest.FhirType : string;
begin
  result := 'ProcessRequest';
end;

function TFhirProcessRequest.Link : TFhirProcessRequest;
begin
  result := TFhirProcessRequest(inherited Link);
end;

function TFhirProcessRequest.Clone : TFhirProcessRequest;
begin
  result := TFhirProcessRequest(inherited Clone);
end;

{ TFhirProcessRequest }

Procedure TFhirProcessRequest.SetAction(value : TFhirEnum);
begin
  FAction.free;
  FAction := value;
end;

Function TFhirProcessRequest.GetActionST : TFhirActionlist;
begin
  if FAction = nil then
    result := TFhirActionlist(0)
  else
    result := TFhirActionlist(StringArrayIndexOfSensitive(CODES_TFhirActionlist, FAction.value));
end;

Procedure TFhirProcessRequest.SetActionST(value : TFhirActionlist);
begin
  if ord(value) = 0 then
    ActionElement := nil
  else
    ActionElement := TFhirEnum.create(CODES_TFhirActionlist[value]);
end;

Function TFhirProcessRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirProcessRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirProcessRequest.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirProcessRequest.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirProcessRequest.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirProcessRequest.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirProcessRequest.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirProcessRequest.SetTarget(value : TFhirReference{TFhirOrganization});
begin
  FTarget.free;
  FTarget := value;
end;

Procedure TFhirProcessRequest.SetProvider(value : TFhirReference{TFhirPractitioner});
begin
  FProvider.free;
  FProvider := value;
end;

Procedure TFhirProcessRequest.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirProcessRequest.SetRequest(value : TFhirReference{TFhirReference});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirProcessRequest.SetResponse(value : TFhirReference{TFhirReference});
begin
  FResponse.free;
  FResponse := value;
end;

Procedure TFhirProcessRequest.SetNullify(value : TFhirBoolean);
begin
  FNullify.free;
  FNullify := value;
end;

Function TFhirProcessRequest.GetNullifyST : Boolean;
begin
  if FNullify = nil then
    result := false
  else
    result := FNullify.value;
end;

Procedure TFhirProcessRequest.SetNullifyST(value : Boolean);
begin
  if FNullify = nil then
    FNullify := TFhirBoolean.create;
  FNullify.value := value
end;

Procedure TFhirProcessRequest.SetReference(value : TFhirString);
begin
  FReference.free;
  FReference := value;
end;

Function TFhirProcessRequest.GetReferenceST : String;
begin
  if FReference = nil then
    result := ''
  else
    result := FReference.value;
end;

Procedure TFhirProcessRequest.SetReferenceST(value : String);
begin
  if value <> '' then
  begin
    if FReference = nil then
      FReference := TFhirString.create;
    FReference.value := value
  end
  else if FReference <> nil then
    FReference.value := '';
end;

Function TFhirProcessRequest.GetItemList : TFhirProcessRequestItemList;
begin
  if FItemList = nil then
    FItemList := TFhirProcessRequestItemList.Create;
  result := FItemList;
end;

Function TFhirProcessRequest.GetHasItemList : boolean;
begin
  result := (FItemList <> nil) and (FItemList.count > 0);
end;

Function TFhirProcessRequest.GetIncludeList : TFhirStringList;
begin
  if FIncludeList = nil then
    FIncludeList := TFhirStringList.Create;
  result := FIncludeList;
end;

Function TFhirProcessRequest.GetHasIncludeList : boolean;
begin
  result := (FIncludeList <> nil) and (FIncludeList.count > 0);
end;

Function TFhirProcessRequest.GetExcludeList : TFhirStringList;
begin
  if FExcludeList = nil then
    FExcludeList := TFhirStringList.Create;
  result := FExcludeList;
end;

Function TFhirProcessRequest.GetHasExcludeList : boolean;
begin
  result := (FExcludeList <> nil) and (FExcludeList.count > 0);
end;

Procedure TFhirProcessRequest.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirProcessRequestListEnumerator }

Constructor TFhirProcessRequestListEnumerator.Create(list : TFhirProcessRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcessRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcessRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcessRequestListEnumerator.GetCurrent : TFhirProcessRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirProcessRequestList }
procedure TFhirProcessRequestList.AddItem(value: TFhirProcessRequest);
begin
  assert(value.ClassName = 'TFhirProcessRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcessRequest');
  add(value);
end;


function TFhirProcessRequestList.Append: TFhirProcessRequest;
begin
  result := TFhirProcessRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessRequestList.ClearItems;
begin
  Clear;
end;

function TFhirProcessRequestList.GetEnumerator : TFhirProcessRequestListEnumerator;
begin
  result := TFhirProcessRequestListEnumerator.Create(self.link);
end;

function TFhirProcessRequestList.Clone: TFhirProcessRequestList;
begin
  result := TFhirProcessRequestList(inherited Clone);
end;

function TFhirProcessRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcessRequestList.GetItemN(index: Integer): TFhirProcessRequest;
begin
  result := TFhirProcessRequest(ObjectByIndex[index]);
end;

function TFhirProcessRequestList.IndexOf(value: TFhirProcessRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcessRequestList.Insert(index: Integer): TFhirProcessRequest;
begin
  result := TFhirProcessRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessRequestList.InsertItem(index: Integer; value: TFhirProcessRequest);
begin
  assert(value is TFhirProcessRequest);
  Inherited Insert(index, value);
end;

function TFhirProcessRequestList.Item(index: Integer): TFhirProcessRequest;
begin
  result := TFhirProcessRequest(ObjectByIndex[index]);
end;

function TFhirProcessRequestList.Link: TFhirProcessRequestList;
begin
  result := TFhirProcessRequestList(inherited Link);
end;

procedure TFhirProcessRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcessRequestList.SetItemByIndex(index: Integer; value: TFhirProcessRequest);
begin
  assert(value is TFhirProcessRequest);
  FhirProcessRequests[index] := value;
end;

procedure TFhirProcessRequestList.SetItemN(index: Integer; value: TFhirProcessRequest);
begin
  assert(value is TFhirProcessRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirProcessResponseNotes }

constructor TFhirProcessResponseNotes.Create;
begin
  inherited;
end;

destructor TFhirProcessResponseNotes.Destroy;
begin
  FType_.free;
  FText.free;
  inherited;
end;

procedure TFhirProcessResponseNotes.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirProcessResponseNotes(oSource).type_.Clone;
  textElement := TFhirProcessResponseNotes(oSource).textElement.Clone;
end;

procedure TFhirProcessResponseNotes.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
end;

procedure TFhirProcessResponseNotes.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
end;

procedure TFhirProcessResponseNotes.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirProcessResponseNotes.FhirType : string;
begin
  result := 'notes';
end;

function TFhirProcessResponseNotes.Link : TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes(inherited Link);
end;

function TFhirProcessResponseNotes.Clone : TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes(inherited Clone);
end;

function TFhirProcessResponseNotes.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProcessResponseNotes;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProcessResponseNotes)) then
    result := false
  else
  begin
    o := TFhirProcessResponseNotes(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(textElement, o.textElement, true);
  end;
end;

function TFhirProcessResponseNotes.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProcessResponseNotes;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProcessResponseNotes)) then
    result := false
  else
  begin
    o := TFhirProcessResponseNotes(other);
    result := compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirProcessResponseNotes }

Procedure TFhirProcessResponseNotes.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirProcessResponseNotes.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirProcessResponseNotes.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirProcessResponseNotes.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;


{ TFhirProcessResponseNotesListEnumerator }

Constructor TFhirProcessResponseNotesListEnumerator.Create(list : TFhirProcessResponseNotesList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcessResponseNotesListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcessResponseNotesListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcessResponseNotesListEnumerator.GetCurrent : TFhirProcessResponseNotes;
begin
  Result := FList[FIndex];
end;


{ TFhirProcessResponseNotesList }
procedure TFhirProcessResponseNotesList.AddItem(value: TFhirProcessResponseNotes);
begin
  assert(value.ClassName = 'TFhirProcessResponseNotes', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcessResponseNotes');
  add(value);
end;


function TFhirProcessResponseNotesList.Append: TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessResponseNotesList.ClearItems;
begin
  Clear;
end;

function TFhirProcessResponseNotesList.GetEnumerator : TFhirProcessResponseNotesListEnumerator;
begin
  result := TFhirProcessResponseNotesListEnumerator.Create(self.link);
end;

function TFhirProcessResponseNotesList.Clone: TFhirProcessResponseNotesList;
begin
  result := TFhirProcessResponseNotesList(inherited Clone);
end;

function TFhirProcessResponseNotesList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcessResponseNotesList.GetItemN(index: Integer): TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes(ObjectByIndex[index]);
end;

function TFhirProcessResponseNotesList.IndexOf(value: TFhirProcessResponseNotes): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcessResponseNotesList.Insert(index: Integer): TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessResponseNotesList.InsertItem(index: Integer; value: TFhirProcessResponseNotes);
begin
  assert(value is TFhirProcessResponseNotes);
  Inherited Insert(index, value);
end;

function TFhirProcessResponseNotesList.Item(index: Integer): TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes(ObjectByIndex[index]);
end;

function TFhirProcessResponseNotesList.Link: TFhirProcessResponseNotesList;
begin
  result := TFhirProcessResponseNotesList(inherited Link);
end;

procedure TFhirProcessResponseNotesList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcessResponseNotesList.SetItemByIndex(index: Integer; value: TFhirProcessResponseNotes);
begin
  assert(value is TFhirProcessResponseNotes);
  FhirProcessResponseNotes[index] := value;
end;

procedure TFhirProcessResponseNotesList.SetItemN(index: Integer; value: TFhirProcessResponseNotes);
begin
  assert(value is TFhirProcessResponseNotes);
  ObjectByIndex[index] := value;
end;

{ TFhirProcessResponse }

constructor TFhirProcessResponse.Create;
begin
  inherited;
end;

destructor TFhirProcessResponse.Destroy;
begin
  FIdentifierList.Free;
  FRequest.free;
  FOutcome.free;
  FDisposition.free;
  FRuleset.free;
  FOriginalRuleset.free;
  FCreated.free;
  FOrganization.free;
  FRequestProvider.free;
  FRequestOrganization.free;
  FForm.free;
  FNotesList.Free;
  FErrorList.Free;
  inherited;
end;

function TFhirProcessResponse.GetResourceType : TFhirResourceType;
begin
  result := frtProcessResponse;
end;

procedure TFhirProcessResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirProcessResponse(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirProcessResponse(oSource).FIdentifierList);
  end;
  request := TFhirProcessResponse(oSource).request.Clone;
  outcome := TFhirProcessResponse(oSource).outcome.Clone;
  dispositionElement := TFhirProcessResponse(oSource).dispositionElement.Clone;
  ruleset := TFhirProcessResponse(oSource).ruleset.Clone;
  originalRuleset := TFhirProcessResponse(oSource).originalRuleset.Clone;
  createdElement := TFhirProcessResponse(oSource).createdElement.Clone;
  organization := TFhirProcessResponse(oSource).organization.Clone;
  requestProvider := TFhirProcessResponse(oSource).requestProvider.Clone;
  requestOrganization := TFhirProcessResponse(oSource).requestOrganization.Clone;
  form := TFhirProcessResponse(oSource).form.Clone;
  if (TFhirProcessResponse(oSource).FNotesList = nil) then
  begin
    FNotesList.free;
    FNotesList := nil;
  end
  else
  begin
    FNotesList := TFhirProcessResponseNotesList.Create;
    FNotesList.Assign(TFhirProcessResponse(oSource).FNotesList);
  end;
  if (TFhirProcessResponse(oSource).FErrorList = nil) then
  begin
    FErrorList.free;
    FErrorList := nil;
  end
  else
  begin
    FErrorList := TFhirCodingList.Create;
    FErrorList.Assign(TFhirProcessResponse(oSource).FErrorList);
  end;
end;

procedure TFhirProcessResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'request') Then
     list.add(FRequest.Link);
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'disposition') Then
     list.add(FDisposition.Link);
  if (child_name = 'ruleset') Then
     list.add(FRuleset.Link);
  if (child_name = 'originalRuleset') Then
     list.add(FOriginalRuleset.Link);
  if (child_name = 'created') Then
     list.add(FCreated.Link);
  if (child_name = 'organization') Then
     list.add(FOrganization.Link);
  if (child_name = 'requestProvider') Then
     list.add(FRequestProvider.Link);
  if (child_name = 'requestOrganization') Then
     list.add(FRequestOrganization.Link);
  if (child_name = 'form') Then
     list.add(FForm.Link);
  if (child_name = 'notes') Then
    list.addAll(FNotesList);
  if (child_name = 'error') Then
    list.addAll(FErrorList);
end;

procedure TFhirProcessResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'request', 'Reference(Any)', FRequest.Link));{2}
  oList.add(TFHIRProperty.create(self, 'outcome', 'Coding', FOutcome.Link));{2}
  oList.add(TFHIRProperty.create(self, 'disposition', 'string', FDisposition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'ruleset', 'Coding', FRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'originalRuleset', 'Coding', FOriginalRuleset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'created', 'dateTime', FCreated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'organization', 'Reference(Organization)', FOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestProvider', 'Reference(Practitioner)', FRequestProvider.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestOrganization', 'Reference(Organization)', FRequestOrganization.Link));{2}
  oList.add(TFHIRProperty.create(self, 'form', 'Coding', FForm.Link));{2}
  oList.add(TFHIRProperty.create(self, 'notes', '', FNotesList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'error', 'Coding', FErrorList.Link)){3};
end;

procedure TFhirProcessResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'request') then Request := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'outcome') then Outcome := propValue as TFhirCoding{4b}
  else if (propName = 'disposition') then DispositionElement := propValue as TFhirString{5a}
  else if (propName = 'ruleset') then Ruleset := propValue as TFhirCoding{4b}
  else if (propName = 'originalRuleset') then OriginalRuleset := propValue as TFhirCoding{4b}
  else if (propName = 'created') then CreatedElement := propValue as TFhirDateTime{5a}
  else if (propName = 'organization') then Organization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'requestProvider') then RequestProvider := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'requestOrganization') then RequestOrganization := propValue as TFhirReference{TFhirOrganization}{4b}
  else if (propName = 'form') then Form := propValue as TFhirCoding{4b}
  else if (propName = 'notes') then NotesList.add(propValue as TFhirProcessResponseNotes){2}
  else if (propName = 'error') then ErrorList.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirProcessResponse.FhirType : string;
begin
  result := 'ProcessResponse';
end;

function TFhirProcessResponse.Link : TFhirProcessResponse;
begin
  result := TFhirProcessResponse(inherited Link);
end;

function TFhirProcessResponse.Clone : TFhirProcessResponse;
begin
  result := TFhirProcessResponse(inherited Clone);
end;

{ TFhirProcessResponse }

Function TFhirProcessResponse.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirProcessResponse.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirProcessResponse.SetRequest(value : TFhirReference{TFhirReference});
begin
  FRequest.free;
  FRequest := value;
end;

Procedure TFhirProcessResponse.SetOutcome(value : TFhirCoding);
begin
  FOutcome.free;
  FOutcome := value;
end;

Procedure TFhirProcessResponse.SetDisposition(value : TFhirString);
begin
  FDisposition.free;
  FDisposition := value;
end;

Function TFhirProcessResponse.GetDispositionST : String;
begin
  if FDisposition = nil then
    result := ''
  else
    result := FDisposition.value;
end;

Procedure TFhirProcessResponse.SetDispositionST(value : String);
begin
  if value <> '' then
  begin
    if FDisposition = nil then
      FDisposition := TFhirString.create;
    FDisposition.value := value
  end
  else if FDisposition <> nil then
    FDisposition.value := '';
end;

Procedure TFhirProcessResponse.SetRuleset(value : TFhirCoding);
begin
  FRuleset.free;
  FRuleset := value;
end;

Procedure TFhirProcessResponse.SetOriginalRuleset(value : TFhirCoding);
begin
  FOriginalRuleset.free;
  FOriginalRuleset := value;
end;

Procedure TFhirProcessResponse.SetCreated(value : TFhirDateTime);
begin
  FCreated.free;
  FCreated := value;
end;

Function TFhirProcessResponse.GetCreatedST : TDateAndTime;
begin
  if FCreated = nil then
    result := nil
  else
    result := FCreated.value;
end;

Procedure TFhirProcessResponse.SetCreatedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FCreated = nil then
      FCreated := TFhirDateTime.create;
    FCreated.value := value
  end
  else if FCreated <> nil then
    FCreated.value := nil;
end;

Procedure TFhirProcessResponse.SetOrganization(value : TFhirReference{TFhirOrganization});
begin
  FOrganization.free;
  FOrganization := value;
end;

Procedure TFhirProcessResponse.SetRequestProvider(value : TFhirReference{TFhirPractitioner});
begin
  FRequestProvider.free;
  FRequestProvider := value;
end;

Procedure TFhirProcessResponse.SetRequestOrganization(value : TFhirReference{TFhirOrganization});
begin
  FRequestOrganization.free;
  FRequestOrganization := value;
end;

Procedure TFhirProcessResponse.SetForm(value : TFhirCoding);
begin
  FForm.free;
  FForm := value;
end;

Function TFhirProcessResponse.GetNotesList : TFhirProcessResponseNotesList;
begin
  if FNotesList = nil then
    FNotesList := TFhirProcessResponseNotesList.Create;
  result := FNotesList;
end;

Function TFhirProcessResponse.GetHasNotesList : boolean;
begin
  result := (FNotesList <> nil) and (FNotesList.count > 0);
end;

Function TFhirProcessResponse.GetErrorList : TFhirCodingList;
begin
  if FErrorList = nil then
    FErrorList := TFhirCodingList.Create;
  result := FErrorList;
end;

Function TFhirProcessResponse.GetHasErrorList : boolean;
begin
  result := (FErrorList <> nil) and (FErrorList.count > 0);
end;


{ TFhirProcessResponseListEnumerator }

Constructor TFhirProcessResponseListEnumerator.Create(list : TFhirProcessResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProcessResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProcessResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProcessResponseListEnumerator.GetCurrent : TFhirProcessResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirProcessResponseList }
procedure TFhirProcessResponseList.AddItem(value: TFhirProcessResponse);
begin
  assert(value.ClassName = 'TFhirProcessResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProcessResponse');
  add(value);
end;


function TFhirProcessResponseList.Append: TFhirProcessResponse;
begin
  result := TFhirProcessResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessResponseList.ClearItems;
begin
  Clear;
end;

function TFhirProcessResponseList.GetEnumerator : TFhirProcessResponseListEnumerator;
begin
  result := TFhirProcessResponseListEnumerator.Create(self.link);
end;

function TFhirProcessResponseList.Clone: TFhirProcessResponseList;
begin
  result := TFhirProcessResponseList(inherited Clone);
end;

function TFhirProcessResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProcessResponseList.GetItemN(index: Integer): TFhirProcessResponse;
begin
  result := TFhirProcessResponse(ObjectByIndex[index]);
end;

function TFhirProcessResponseList.IndexOf(value: TFhirProcessResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProcessResponseList.Insert(index: Integer): TFhirProcessResponse;
begin
  result := TFhirProcessResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProcessResponseList.InsertItem(index: Integer; value: TFhirProcessResponse);
begin
  assert(value is TFhirProcessResponse);
  Inherited Insert(index, value);
end;

function TFhirProcessResponseList.Item(index: Integer): TFhirProcessResponse;
begin
  result := TFhirProcessResponse(ObjectByIndex[index]);
end;

function TFhirProcessResponseList.Link: TFhirProcessResponseList;
begin
  result := TFhirProcessResponseList(inherited Link);
end;

procedure TFhirProcessResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProcessResponseList.SetItemByIndex(index: Integer; value: TFhirProcessResponse);
begin
  assert(value is TFhirProcessResponse);
  FhirProcessResponses[index] := value;
end;

procedure TFhirProcessResponseList.SetItemN(index: Integer; value: TFhirProcessResponse);
begin
  assert(value is TFhirProcessResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirProvenanceAgent }

constructor TFhirProvenanceAgent.Create;
begin
  inherited;
end;

destructor TFhirProvenanceAgent.Destroy;
begin
  FRole.free;
  FActor.free;
  FUserId.free;
  FRelatedAgentList.Free;
  inherited;
end;

procedure TFhirProvenanceAgent.Assign(oSource : TAdvObject);
begin
  inherited;
  role := TFhirProvenanceAgent(oSource).role.Clone;
  actor := TFhirProvenanceAgent(oSource).actor.Clone;
  userId := TFhirProvenanceAgent(oSource).userId.Clone;
  if (TFhirProvenanceAgent(oSource).FRelatedAgentList = nil) then
  begin
    FRelatedAgentList.free;
    FRelatedAgentList := nil;
  end
  else
  begin
    FRelatedAgentList := TFhirProvenanceAgentRelatedAgentList.Create;
    FRelatedAgentList.Assign(TFhirProvenanceAgent(oSource).FRelatedAgentList);
  end;
end;

procedure TFhirProvenanceAgent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'role') Then
     list.add(FRole.Link);
  if (child_name = 'actor') Then
     list.add(FActor.Link);
  if (child_name = 'userId') Then
     list.add(FUserId.Link);
  if (child_name = 'relatedAgent') Then
    list.addAll(FRelatedAgentList);
end;

procedure TFhirProvenanceAgent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'role', 'Coding', FRole.Link));{2}
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Practitioner|RelatedPerson|Patient|Device|Organization)', FActor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'userId', 'Identifier', FUserId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relatedAgent', '', FRelatedAgentList.Link)){3};
end;

procedure TFhirProvenanceAgent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'role') then Role := propValue as TFhirCoding{4b}
  else if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'userId') then UserId := propValue as TFhirIdentifier{4b}
  else if (propName = 'relatedAgent') then RelatedAgentList.add(propValue as TFhirProvenanceAgentRelatedAgent){2}
  else inherited;
end;

function TFhirProvenanceAgent.FhirType : string;
begin
  result := 'agent';
end;

function TFhirProvenanceAgent.Link : TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent(inherited Link);
end;

function TFhirProvenanceAgent.Clone : TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent(inherited Clone);
end;

function TFhirProvenanceAgent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceAgent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProvenanceAgent)) then
    result := false
  else
  begin
    o := TFhirProvenanceAgent(other);
    result := compareDeep(roleElement, o.roleElement, true) and compareDeep(actorElement, o.actorElement, true) and 
      compareDeep(userIdElement, o.userIdElement, true) and compareDeep(relatedAgentList, o.relatedAgentList, true);
  end;
end;

function TFhirProvenanceAgent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceAgent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProvenanceAgent)) then
    result := false
  else
  begin
    o := TFhirProvenanceAgent(other);
    result := true;
  end;
end;

{ TFhirProvenanceAgent }

Procedure TFhirProvenanceAgent.SetRole(value : TFhirCoding);
begin
  FRole.free;
  FRole := value;
end;

Procedure TFhirProvenanceAgent.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;

Procedure TFhirProvenanceAgent.SetUserId(value : TFhirIdentifier);
begin
  FUserId.free;
  FUserId := value;
end;

Function TFhirProvenanceAgent.GetRelatedAgentList : TFhirProvenanceAgentRelatedAgentList;
begin
  if FRelatedAgentList = nil then
    FRelatedAgentList := TFhirProvenanceAgentRelatedAgentList.Create;
  result := FRelatedAgentList;
end;

Function TFhirProvenanceAgent.GetHasRelatedAgentList : boolean;
begin
  result := (FRelatedAgentList <> nil) and (FRelatedAgentList.count > 0);
end;


{ TFhirProvenanceAgentListEnumerator }

Constructor TFhirProvenanceAgentListEnumerator.Create(list : TFhirProvenanceAgentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProvenanceAgentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProvenanceAgentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProvenanceAgentListEnumerator.GetCurrent : TFhirProvenanceAgent;
begin
  Result := FList[FIndex];
end;


{ TFhirProvenanceAgentList }
procedure TFhirProvenanceAgentList.AddItem(value: TFhirProvenanceAgent);
begin
  assert(value.ClassName = 'TFhirProvenanceAgent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProvenanceAgent');
  add(value);
end;


function TFhirProvenanceAgentList.Append: TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceAgentList.ClearItems;
begin
  Clear;
end;

function TFhirProvenanceAgentList.GetEnumerator : TFhirProvenanceAgentListEnumerator;
begin
  result := TFhirProvenanceAgentListEnumerator.Create(self.link);
end;

function TFhirProvenanceAgentList.Clone: TFhirProvenanceAgentList;
begin
  result := TFhirProvenanceAgentList(inherited Clone);
end;

function TFhirProvenanceAgentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProvenanceAgentList.GetItemN(index: Integer): TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent(ObjectByIndex[index]);
end;

function TFhirProvenanceAgentList.IndexOf(value: TFhirProvenanceAgent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProvenanceAgentList.Insert(index: Integer): TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceAgentList.InsertItem(index: Integer; value: TFhirProvenanceAgent);
begin
  assert(value is TFhirProvenanceAgent);
  Inherited Insert(index, value);
end;

function TFhirProvenanceAgentList.Item(index: Integer): TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent(ObjectByIndex[index]);
end;

function TFhirProvenanceAgentList.Link: TFhirProvenanceAgentList;
begin
  result := TFhirProvenanceAgentList(inherited Link);
end;

procedure TFhirProvenanceAgentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProvenanceAgentList.SetItemByIndex(index: Integer; value: TFhirProvenanceAgent);
begin
  assert(value is TFhirProvenanceAgent);
  FhirProvenanceAgents[index] := value;
end;

procedure TFhirProvenanceAgentList.SetItemN(index: Integer; value: TFhirProvenanceAgent);
begin
  assert(value is TFhirProvenanceAgent);
  ObjectByIndex[index] := value;
end;

{ TFhirProvenanceAgentRelatedAgent }

constructor TFhirProvenanceAgentRelatedAgent.Create;
begin
  inherited;
end;

destructor TFhirProvenanceAgentRelatedAgent.Destroy;
begin
  FType_.free;
  FTarget.free;
  inherited;
end;

procedure TFhirProvenanceAgentRelatedAgent.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirProvenanceAgentRelatedAgent(oSource).type_.Clone;
  targetElement := TFhirProvenanceAgentRelatedAgent(oSource).targetElement.Clone;
end;

procedure TFhirProvenanceAgentRelatedAgent.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'target') Then
     list.add(FTarget.Link);
end;

procedure TFhirProvenanceAgentRelatedAgent.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'target', 'uri', FTarget.Link));{2}
end;

procedure TFhirProvenanceAgentRelatedAgent.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'target') then TargetElement := propValue as TFhirUri{5a}
  else inherited;
end;

function TFhirProvenanceAgentRelatedAgent.FhirType : string;
begin
  result := 'relatedAgent';
end;

function TFhirProvenanceAgentRelatedAgent.Link : TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent(inherited Link);
end;

function TFhirProvenanceAgentRelatedAgent.Clone : TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent(inherited Clone);
end;

function TFhirProvenanceAgentRelatedAgent.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceAgentRelatedAgent;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProvenanceAgentRelatedAgent)) then
    result := false
  else
  begin
    o := TFhirProvenanceAgentRelatedAgent(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(targetElement, o.targetElement, true);
  end;
end;

function TFhirProvenanceAgentRelatedAgent.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceAgentRelatedAgent;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProvenanceAgentRelatedAgent)) then
    result := false
  else
  begin
    o := TFhirProvenanceAgentRelatedAgent(other);
    result := compareValues(targetElement, o.targetElement, true);
  end;
end;

{ TFhirProvenanceAgentRelatedAgent }

Procedure TFhirProvenanceAgentRelatedAgent.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirProvenanceAgentRelatedAgent.SetTarget(value : TFhirUri);
begin
  FTarget.free;
  FTarget := value;
end;

Function TFhirProvenanceAgentRelatedAgent.GetTargetST : String;
begin
  if FTarget = nil then
    result := ''
  else
    result := FTarget.value;
end;

Procedure TFhirProvenanceAgentRelatedAgent.SetTargetST(value : String);
begin
  if value <> '' then
  begin
    if FTarget = nil then
      FTarget := TFhirUri.create;
    FTarget.value := value
  end
  else if FTarget <> nil then
    FTarget.value := '';
end;


{ TFhirProvenanceAgentRelatedAgentListEnumerator }

Constructor TFhirProvenanceAgentRelatedAgentListEnumerator.Create(list : TFhirProvenanceAgentRelatedAgentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProvenanceAgentRelatedAgentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProvenanceAgentRelatedAgentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProvenanceAgentRelatedAgentListEnumerator.GetCurrent : TFhirProvenanceAgentRelatedAgent;
begin
  Result := FList[FIndex];
end;


{ TFhirProvenanceAgentRelatedAgentList }
procedure TFhirProvenanceAgentRelatedAgentList.AddItem(value: TFhirProvenanceAgentRelatedAgent);
begin
  assert(value.ClassName = 'TFhirProvenanceAgentRelatedAgent', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProvenanceAgentRelatedAgent');
  add(value);
end;


function TFhirProvenanceAgentRelatedAgentList.Append: TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceAgentRelatedAgentList.ClearItems;
begin
  Clear;
end;

function TFhirProvenanceAgentRelatedAgentList.GetEnumerator : TFhirProvenanceAgentRelatedAgentListEnumerator;
begin
  result := TFhirProvenanceAgentRelatedAgentListEnumerator.Create(self.link);
end;

function TFhirProvenanceAgentRelatedAgentList.Clone: TFhirProvenanceAgentRelatedAgentList;
begin
  result := TFhirProvenanceAgentRelatedAgentList(inherited Clone);
end;

function TFhirProvenanceAgentRelatedAgentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProvenanceAgentRelatedAgentList.GetItemN(index: Integer): TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent(ObjectByIndex[index]);
end;

function TFhirProvenanceAgentRelatedAgentList.IndexOf(value: TFhirProvenanceAgentRelatedAgent): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProvenanceAgentRelatedAgentList.Insert(index: Integer): TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceAgentRelatedAgentList.InsertItem(index: Integer; value: TFhirProvenanceAgentRelatedAgent);
begin
  assert(value is TFhirProvenanceAgentRelatedAgent);
  Inherited Insert(index, value);
end;

function TFhirProvenanceAgentRelatedAgentList.Item(index: Integer): TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent(ObjectByIndex[index]);
end;

function TFhirProvenanceAgentRelatedAgentList.Link: TFhirProvenanceAgentRelatedAgentList;
begin
  result := TFhirProvenanceAgentRelatedAgentList(inherited Link);
end;

procedure TFhirProvenanceAgentRelatedAgentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProvenanceAgentRelatedAgentList.SetItemByIndex(index: Integer; value: TFhirProvenanceAgentRelatedAgent);
begin
  assert(value is TFhirProvenanceAgentRelatedAgent);
  FhirProvenanceAgentRelatedAgents[index] := value;
end;

procedure TFhirProvenanceAgentRelatedAgentList.SetItemN(index: Integer; value: TFhirProvenanceAgentRelatedAgent);
begin
  assert(value is TFhirProvenanceAgentRelatedAgent);
  ObjectByIndex[index] := value;
end;

{ TFhirProvenanceEntity }

constructor TFhirProvenanceEntity.Create;
begin
  inherited;
end;

destructor TFhirProvenanceEntity.Destroy;
begin
  FRole.free;
  FType_.free;
  FReference.free;
  FDisplay.free;
  FAgent.free;
  inherited;
end;

procedure TFhirProvenanceEntity.Assign(oSource : TAdvObject);
begin
  inherited;
  FRole := TFhirProvenanceEntity(oSource).FRole.Link;
  type_ := TFhirProvenanceEntity(oSource).type_.Clone;
  referenceElement := TFhirProvenanceEntity(oSource).referenceElement.Clone;
  displayElement := TFhirProvenanceEntity(oSource).displayElement.Clone;
  agent := TFhirProvenanceEntity(oSource).agent.Clone;
end;

procedure TFhirProvenanceEntity.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'role') Then
     list.add(FRole.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'reference') Then
     list.add(FReference.Link);
  if (child_name = 'display') Then
     list.add(FDisplay.Link);
  if (child_name = 'agent') Then
     list.add(FAgent.Link);
end;

procedure TFhirProvenanceEntity.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'role', 'code', FRole.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reference', 'uri', FReference.Link));{2}
  oList.add(TFHIRProperty.create(self, 'display', 'string', FDisplay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'agent', '@Provenance.agent', FAgent.Link));{2}
end;

procedure TFhirProvenanceEntity.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'role') then RoleElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'reference') then ReferenceElement := propValue as TFhirUri{5a}
  else if (propName = 'display') then DisplayElement := propValue as TFhirString{5a}
  else if (propName = 'agent') then Agent := propValue as TFhirProvenanceAgent{4b}
  else inherited;
end;

function TFhirProvenanceEntity.FhirType : string;
begin
  result := 'entity';
end;

function TFhirProvenanceEntity.Link : TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity(inherited Link);
end;

function TFhirProvenanceEntity.Clone : TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity(inherited Clone);
end;

function TFhirProvenanceEntity.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceEntity;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirProvenanceEntity)) then
    result := false
  else
  begin
    o := TFhirProvenanceEntity(other);
    result := compareDeep(roleElement, o.roleElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(referenceElement, o.referenceElement, true) and compareDeep(displayElement, o.displayElement, true) and 
      compareDeep(agentElement, o.agentElement, true);
  end;
end;

function TFhirProvenanceEntity.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirProvenanceEntity;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirProvenanceEntity)) then
    result := false
  else
  begin
    o := TFhirProvenanceEntity(other);
    result := compareValues(roleElement, o.roleElement, true) and compareValues(referenceElement, o.referenceElement, true) and 
      compareValues(displayElement, o.displayElement, true);
  end;
end;

{ TFhirProvenanceEntity }

Procedure TFhirProvenanceEntity.SetRole(value : TFhirEnum);
begin
  FRole.free;
  FRole := value;
end;

Function TFhirProvenanceEntity.GetRoleST : TFhirProvenanceEntityRole;
begin
  if FRole = nil then
    result := TFhirProvenanceEntityRole(0)
  else
    result := TFhirProvenanceEntityRole(StringArrayIndexOfSensitive(CODES_TFhirProvenanceEntityRole, FRole.value));
end;

Procedure TFhirProvenanceEntity.SetRoleST(value : TFhirProvenanceEntityRole);
begin
  if ord(value) = 0 then
    RoleElement := nil
  else
    RoleElement := TFhirEnum.create(CODES_TFhirProvenanceEntityRole[value]);
end;

Procedure TFhirProvenanceEntity.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirProvenanceEntity.SetReference(value : TFhirUri);
begin
  FReference.free;
  FReference := value;
end;

Function TFhirProvenanceEntity.GetReferenceST : String;
begin
  if FReference = nil then
    result := ''
  else
    result := FReference.value;
end;

Procedure TFhirProvenanceEntity.SetReferenceST(value : String);
begin
  if value <> '' then
  begin
    if FReference = nil then
      FReference := TFhirUri.create;
    FReference.value := value
  end
  else if FReference <> nil then
    FReference.value := '';
end;

Procedure TFhirProvenanceEntity.SetDisplay(value : TFhirString);
begin
  FDisplay.free;
  FDisplay := value;
end;

Function TFhirProvenanceEntity.GetDisplayST : String;
begin
  if FDisplay = nil then
    result := ''
  else
    result := FDisplay.value;
end;

Procedure TFhirProvenanceEntity.SetDisplayST(value : String);
begin
  if value <> '' then
  begin
    if FDisplay = nil then
      FDisplay := TFhirString.create;
    FDisplay.value := value
  end
  else if FDisplay <> nil then
    FDisplay.value := '';
end;

Procedure TFhirProvenanceEntity.SetAgent(value : TFhirProvenanceAgent);
begin
  FAgent.free;
  FAgent := value;
end;


{ TFhirProvenanceEntityListEnumerator }

Constructor TFhirProvenanceEntityListEnumerator.Create(list : TFhirProvenanceEntityList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProvenanceEntityListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProvenanceEntityListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProvenanceEntityListEnumerator.GetCurrent : TFhirProvenanceEntity;
begin
  Result := FList[FIndex];
end;


{ TFhirProvenanceEntityList }
procedure TFhirProvenanceEntityList.AddItem(value: TFhirProvenanceEntity);
begin
  assert(value.ClassName = 'TFhirProvenanceEntity', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProvenanceEntity');
  add(value);
end;


function TFhirProvenanceEntityList.Append: TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceEntityList.ClearItems;
begin
  Clear;
end;

function TFhirProvenanceEntityList.GetEnumerator : TFhirProvenanceEntityListEnumerator;
begin
  result := TFhirProvenanceEntityListEnumerator.Create(self.link);
end;

function TFhirProvenanceEntityList.Clone: TFhirProvenanceEntityList;
begin
  result := TFhirProvenanceEntityList(inherited Clone);
end;

function TFhirProvenanceEntityList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProvenanceEntityList.GetItemN(index: Integer): TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity(ObjectByIndex[index]);
end;

function TFhirProvenanceEntityList.IndexOf(value: TFhirProvenanceEntity): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProvenanceEntityList.Insert(index: Integer): TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceEntityList.InsertItem(index: Integer; value: TFhirProvenanceEntity);
begin
  assert(value is TFhirProvenanceEntity);
  Inherited Insert(index, value);
end;

function TFhirProvenanceEntityList.Item(index: Integer): TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity(ObjectByIndex[index]);
end;

function TFhirProvenanceEntityList.Link: TFhirProvenanceEntityList;
begin
  result := TFhirProvenanceEntityList(inherited Link);
end;

procedure TFhirProvenanceEntityList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProvenanceEntityList.SetItemByIndex(index: Integer; value: TFhirProvenanceEntity);
begin
  assert(value is TFhirProvenanceEntity);
  FhirProvenanceEntities[index] := value;
end;

procedure TFhirProvenanceEntityList.SetItemN(index: Integer; value: TFhirProvenanceEntity);
begin
  assert(value is TFhirProvenanceEntity);
  ObjectByIndex[index] := value;
end;

{ TFhirProvenance }

constructor TFhirProvenance.Create;
begin
  inherited;
end;

destructor TFhirProvenance.Destroy;
begin
  FTargetList.Free;
  FPeriod.free;
  FRecorded.free;
  FReasonList.Free;
  FActivity.free;
  FLocation.free;
  FPolicyList.Free;
  FAgentList.Free;
  FEntityList.Free;
  FSignatureList.Free;
  inherited;
end;

function TFhirProvenance.GetResourceType : TFhirResourceType;
begin
  result := frtProvenance;
end;

procedure TFhirProvenance.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirProvenance(oSource).FTargetList = nil) then
  begin
    FTargetList.free;
    FTargetList := nil;
  end
  else
  begin
    FTargetList := TFhirReferenceList{TFhirReference}.Create;
    FTargetList.Assign(TFhirProvenance(oSource).FTargetList);
  end;
  period := TFhirProvenance(oSource).period.Clone;
  recordedElement := TFhirProvenance(oSource).recordedElement.Clone;
  if (TFhirProvenance(oSource).FReasonList = nil) then
  begin
    FReasonList.free;
    FReasonList := nil;
  end
  else
  begin
    FReasonList := TFhirCodeableConceptList.Create;
    FReasonList.Assign(TFhirProvenance(oSource).FReasonList);
  end;
  activity := TFhirProvenance(oSource).activity.Clone;
  location := TFhirProvenance(oSource).location.Clone;
  if (TFhirProvenance(oSource).FPolicyList = nil) then
  begin
    FPolicyList.free;
    FPolicyList := nil;
  end
  else
  begin
    FPolicyList := TFhirUriList.Create;
    FPolicyList.Assign(TFhirProvenance(oSource).FPolicyList);
  end;
  if (TFhirProvenance(oSource).FAgentList = nil) then
  begin
    FAgentList.free;
    FAgentList := nil;
  end
  else
  begin
    FAgentList := TFhirProvenanceAgentList.Create;
    FAgentList.Assign(TFhirProvenance(oSource).FAgentList);
  end;
  if (TFhirProvenance(oSource).FEntityList = nil) then
  begin
    FEntityList.free;
    FEntityList := nil;
  end
  else
  begin
    FEntityList := TFhirProvenanceEntityList.Create;
    FEntityList.Assign(TFhirProvenance(oSource).FEntityList);
  end;
  if (TFhirProvenance(oSource).FSignatureList = nil) then
  begin
    FSignatureList.free;
    FSignatureList := nil;
  end
  else
  begin
    FSignatureList := TFhirSignatureList.Create;
    FSignatureList.Assign(TFhirProvenance(oSource).FSignatureList);
  end;
end;

procedure TFhirProvenance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'target') Then
    list.addAll(FTargetList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
  if (child_name = 'recorded') Then
     list.add(FRecorded.Link);
  if (child_name = 'reason') Then
    list.addAll(FReasonList);
  if (child_name = 'activity') Then
     list.add(FActivity.Link);
  if (child_name = 'location') Then
     list.add(FLocation.Link);
  if (child_name = 'policy') Then
    list.addAll(FPolicyList);
  if (child_name = 'agent') Then
    list.addAll(FAgentList);
  if (child_name = 'entity') Then
    list.addAll(FEntityList);
  if (child_name = 'signature') Then
    list.addAll(FSignatureList);
end;

procedure TFhirProvenance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'target', 'Reference(Any)', FTargetList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recorded', 'instant', FRecorded.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReasonList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'activity', 'CodeableConcept', FActivity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'location', 'Reference(Location)', FLocation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'policy', 'uri', FPolicyList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'agent', '', FAgentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'entity', '', FEntityList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'signature', 'Signature', FSignatureList.Link)){3};
end;

procedure TFhirProvenance.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'target') then TargetList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else if (propName = 'recorded') then RecordedElement := propValue as TFhirInstant{5a}
  else if (propName = 'reason') then ReasonList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'activity') then Activity := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'location') then Location := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'policy') then PolicyList.add(propValue as TFhirUri){2}
  else if (propName = 'agent') then AgentList.add(propValue as TFhirProvenanceAgent){2}
  else if (propName = 'entity') then EntityList.add(propValue as TFhirProvenanceEntity){2}
  else if (propName = 'signature') then SignatureList.add(propValue as TFhirSignature){2}
  else inherited;
end;

function TFhirProvenance.FhirType : string;
begin
  result := 'Provenance';
end;

function TFhirProvenance.Link : TFhirProvenance;
begin
  result := TFhirProvenance(inherited Link);
end;

function TFhirProvenance.Clone : TFhirProvenance;
begin
  result := TFhirProvenance(inherited Clone);
end;

{ TFhirProvenance }

Function TFhirProvenance.GetTargetList : TFhirReferenceList{TFhirReference};
begin
  if FTargetList = nil then
    FTargetList := TFhirReferenceList{TFhirReference}.Create;
  result := FTargetList;
end;

Function TFhirProvenance.GetHasTargetList : boolean;
begin
  result := (FTargetList <> nil) and (FTargetList.count > 0);
end;

Procedure TFhirProvenance.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;

Procedure TFhirProvenance.SetRecorded(value : TFhirInstant);
begin
  FRecorded.free;
  FRecorded := value;
end;

Function TFhirProvenance.GetRecordedST : TDateAndTime;
begin
  if FRecorded = nil then
    result := nil
  else
    result := FRecorded.value;
end;

Procedure TFhirProvenance.SetRecordedST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FRecorded = nil then
      FRecorded := TFhirInstant.create;
    FRecorded.value := value
  end
  else if FRecorded <> nil then
    FRecorded.value := nil;
end;

Function TFhirProvenance.GetReasonList : TFhirCodeableConceptList;
begin
  if FReasonList = nil then
    FReasonList := TFhirCodeableConceptList.Create;
  result := FReasonList;
end;

Function TFhirProvenance.GetHasReasonList : boolean;
begin
  result := (FReasonList <> nil) and (FReasonList.count > 0);
end;

Procedure TFhirProvenance.SetActivity(value : TFhirCodeableConcept);
begin
  FActivity.free;
  FActivity := value;
end;

Procedure TFhirProvenance.SetLocation(value : TFhirReference{TFhirLocation});
begin
  FLocation.free;
  FLocation := value;
end;

Function TFhirProvenance.GetPolicyList : TFhirUriList;
begin
  if FPolicyList = nil then
    FPolicyList := TFhirUriList.Create;
  result := FPolicyList;
end;

Function TFhirProvenance.GetHasPolicyList : boolean;
begin
  result := (FPolicyList <> nil) and (FPolicyList.count > 0);
end;

Function TFhirProvenance.GetAgentList : TFhirProvenanceAgentList;
begin
  if FAgentList = nil then
    FAgentList := TFhirProvenanceAgentList.Create;
  result := FAgentList;
end;

Function TFhirProvenance.GetHasAgentList : boolean;
begin
  result := (FAgentList <> nil) and (FAgentList.count > 0);
end;

Function TFhirProvenance.GetEntityList : TFhirProvenanceEntityList;
begin
  if FEntityList = nil then
    FEntityList := TFhirProvenanceEntityList.Create;
  result := FEntityList;
end;

Function TFhirProvenance.GetHasEntityList : boolean;
begin
  result := (FEntityList <> nil) and (FEntityList.count > 0);
end;

Function TFhirProvenance.GetSignatureList : TFhirSignatureList;
begin
  if FSignatureList = nil then
    FSignatureList := TFhirSignatureList.Create;
  result := FSignatureList;
end;

Function TFhirProvenance.GetHasSignatureList : boolean;
begin
  result := (FSignatureList <> nil) and (FSignatureList.count > 0);
end;


{ TFhirProvenanceListEnumerator }

Constructor TFhirProvenanceListEnumerator.Create(list : TFhirProvenanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirProvenanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirProvenanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirProvenanceListEnumerator.GetCurrent : TFhirProvenance;
begin
  Result := FList[FIndex];
end;


{ TFhirProvenanceList }
procedure TFhirProvenanceList.AddItem(value: TFhirProvenance);
begin
  assert(value.ClassName = 'TFhirProvenance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirProvenance');
  add(value);
end;


function TFhirProvenanceList.Append: TFhirProvenance;
begin
  result := TFhirProvenance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceList.ClearItems;
begin
  Clear;
end;

function TFhirProvenanceList.GetEnumerator : TFhirProvenanceListEnumerator;
begin
  result := TFhirProvenanceListEnumerator.Create(self.link);
end;

function TFhirProvenanceList.Clone: TFhirProvenanceList;
begin
  result := TFhirProvenanceList(inherited Clone);
end;

function TFhirProvenanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirProvenanceList.GetItemN(index: Integer): TFhirProvenance;
begin
  result := TFhirProvenance(ObjectByIndex[index]);
end;

function TFhirProvenanceList.IndexOf(value: TFhirProvenance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirProvenanceList.Insert(index: Integer): TFhirProvenance;
begin
  result := TFhirProvenance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirProvenanceList.InsertItem(index: Integer; value: TFhirProvenance);
begin
  assert(value is TFhirProvenance);
  Inherited Insert(index, value);
end;

function TFhirProvenanceList.Item(index: Integer): TFhirProvenance;
begin
  result := TFhirProvenance(ObjectByIndex[index]);
end;

function TFhirProvenanceList.Link: TFhirProvenanceList;
begin
  result := TFhirProvenanceList(inherited Link);
end;

procedure TFhirProvenanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirProvenanceList.SetItemByIndex(index: Integer; value: TFhirProvenance);
begin
  assert(value is TFhirProvenance);
  FhirProvenances[index] := value;
end;

procedure TFhirProvenanceList.SetItemN(index: Integer; value: TFhirProvenance);
begin
  assert(value is TFhirProvenance);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireGroup }

constructor TFhirQuestionnaireGroup.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireGroup.Destroy;
begin
  FLinkId.free;
  FTitle.free;
  FConceptList.Free;
  FText.free;
  FRequired.free;
  FRepeats.free;
  FGroupList.Free;
  FQuestionList.Free;
  inherited;
end;

procedure TFhirQuestionnaireGroup.Assign(oSource : TAdvObject);
begin
  inherited;
  linkIdElement := TFhirQuestionnaireGroup(oSource).linkIdElement.Clone;
  titleElement := TFhirQuestionnaireGroup(oSource).titleElement.Clone;
  if (TFhirQuestionnaireGroup(oSource).FConceptList = nil) then
  begin
    FConceptList.free;
    FConceptList := nil;
  end
  else
  begin
    FConceptList := TFhirCodingList.Create;
    FConceptList.Assign(TFhirQuestionnaireGroup(oSource).FConceptList);
  end;
  textElement := TFhirQuestionnaireGroup(oSource).textElement.Clone;
  requiredElement := TFhirQuestionnaireGroup(oSource).requiredElement.Clone;
  repeatsElement := TFhirQuestionnaireGroup(oSource).repeatsElement.Clone;
  if (TFhirQuestionnaireGroup(oSource).FGroupList = nil) then
  begin
    FGroupList.free;
    FGroupList := nil;
  end
  else
  begin
    FGroupList := TFhirQuestionnaireGroupList.Create;
    FGroupList.Assign(TFhirQuestionnaireGroup(oSource).FGroupList);
  end;
  if (TFhirQuestionnaireGroup(oSource).FQuestionList = nil) then
  begin
    FQuestionList.free;
    FQuestionList := nil;
  end
  else
  begin
    FQuestionList := TFhirQuestionnaireGroupQuestionList.Create;
    FQuestionList.Assign(TFhirQuestionnaireGroup(oSource).FQuestionList);
  end;
end;

procedure TFhirQuestionnaireGroup.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'linkId') Then
     list.add(FLinkId.Link);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'concept') Then
    list.addAll(FConceptList);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'required') Then
     list.add(FRequired.Link);
  if (child_name = 'repeats') Then
     list.add(FRepeats.Link);
  if (child_name = 'group') Then
    list.addAll(FGroupList);
  if (child_name = 'question') Then
    list.addAll(FQuestionList);
end;

procedure TFhirQuestionnaireGroup.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'linkId', 'string', FLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'concept', 'Coding', FConceptList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'required', 'boolean', FRequired.Link));{2}
  oList.add(TFHIRProperty.create(self, 'repeats', 'boolean', FRepeats.Link));{2}
  oList.add(TFHIRProperty.create(self, 'group', '@Questionnaire.group', FGroupList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'question', '', FQuestionList.Link)){3};
end;

procedure TFhirQuestionnaireGroup.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'linkId') then LinkIdElement := propValue as TFhirString{5a}
  else if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'concept') then ConceptList.add(propValue as TFhirCoding){2}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'required') then RequiredElement := propValue as TFhirBoolean{5a}
  else if (propName = 'repeats') then RepeatsElement := propValue as TFhirBoolean{5a}
  else if (propName = 'group') then GroupList.add(propValue as TFhirQuestionnaireGroup){2}
  else if (propName = 'question') then QuestionList.add(propValue as TFhirQuestionnaireGroupQuestion){2}
  else inherited;
end;

function TFhirQuestionnaireGroup.FhirType : string;
begin
  result := 'group';
end;

function TFhirQuestionnaireGroup.Link : TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup(inherited Link);
end;

function TFhirQuestionnaireGroup.Clone : TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup(inherited Clone);
end;

function TFhirQuestionnaireGroup.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireGroup;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireGroup)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireGroup(other);
    result := compareDeep(linkIdElement, o.linkIdElement, true) and compareDeep(titleElement, o.titleElement, true) and 
      compareDeep(conceptList, o.conceptList, true) and compareDeep(textElement, o.textElement, true) and 
      compareDeep(requiredElement, o.requiredElement, true) and compareDeep(repeatsElement, o.repeatsElement, true) and 
      compareDeep(groupList, o.groupList, true) and compareDeep(questionList, o.questionList, true);
  end;
end;

function TFhirQuestionnaireGroup.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireGroup;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireGroup)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireGroup(other);
    result := compareValues(linkIdElement, o.linkIdElement, true) and compareValues(titleElement, o.titleElement, true) and 
      compareValues(textElement, o.textElement, true) and compareValues(requiredElement, o.requiredElement, true) and 
      compareValues(repeatsElement, o.repeatsElement, true);
  end;
end;

{ TFhirQuestionnaireGroup }

Procedure TFhirQuestionnaireGroup.SetLinkId(value : TFhirString);
begin
  FLinkId.free;
  FLinkId := value;
end;

Function TFhirQuestionnaireGroup.GetLinkIdST : String;
begin
  if FLinkId = nil then
    result := ''
  else
    result := FLinkId.value;
end;

Procedure TFhirQuestionnaireGroup.SetLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FLinkId = nil then
      FLinkId := TFhirString.create;
    FLinkId.value := value
  end
  else if FLinkId <> nil then
    FLinkId.value := '';
end;

Procedure TFhirQuestionnaireGroup.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirQuestionnaireGroup.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirQuestionnaireGroup.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Function TFhirQuestionnaireGroup.GetConceptList : TFhirCodingList;
begin
  if FConceptList = nil then
    FConceptList := TFhirCodingList.Create;
  result := FConceptList;
end;

Function TFhirQuestionnaireGroup.GetHasConceptList : boolean;
begin
  result := (FConceptList <> nil) and (FConceptList.count > 0);
end;

Procedure TFhirQuestionnaireGroup.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirQuestionnaireGroup.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirQuestionnaireGroup.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirQuestionnaireGroup.SetRequired(value : TFhirBoolean);
begin
  FRequired.free;
  FRequired := value;
end;

Function TFhirQuestionnaireGroup.GetRequiredST : Boolean;
begin
  if FRequired = nil then
    result := false
  else
    result := FRequired.value;
end;

Procedure TFhirQuestionnaireGroup.SetRequiredST(value : Boolean);
begin
  if FRequired = nil then
    FRequired := TFhirBoolean.create;
  FRequired.value := value
end;

Procedure TFhirQuestionnaireGroup.SetRepeats(value : TFhirBoolean);
begin
  FRepeats.free;
  FRepeats := value;
end;

Function TFhirQuestionnaireGroup.GetRepeatsST : Boolean;
begin
  if FRepeats = nil then
    result := false
  else
    result := FRepeats.value;
end;

Procedure TFhirQuestionnaireGroup.SetRepeatsST(value : Boolean);
begin
  if FRepeats = nil then
    FRepeats := TFhirBoolean.create;
  FRepeats.value := value
end;

Function TFhirQuestionnaireGroup.GetGroupList : TFhirQuestionnaireGroupList;
begin
  if FGroupList = nil then
    FGroupList := TFhirQuestionnaireGroupList.Create;
  result := FGroupList;
end;

Function TFhirQuestionnaireGroup.GetHasGroupList : boolean;
begin
  result := (FGroupList <> nil) and (FGroupList.count > 0);
end;

Function TFhirQuestionnaireGroup.GetQuestionList : TFhirQuestionnaireGroupQuestionList;
begin
  if FQuestionList = nil then
    FQuestionList := TFhirQuestionnaireGroupQuestionList.Create;
  result := FQuestionList;
end;

Function TFhirQuestionnaireGroup.GetHasQuestionList : boolean;
begin
  result := (FQuestionList <> nil) and (FQuestionList.count > 0);
end;


{ TFhirQuestionnaireGroupListEnumerator }

Constructor TFhirQuestionnaireGroupListEnumerator.Create(list : TFhirQuestionnaireGroupList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireGroupListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireGroupListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireGroupListEnumerator.GetCurrent : TFhirQuestionnaireGroup;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireGroupList }
procedure TFhirQuestionnaireGroupList.AddItem(value: TFhirQuestionnaireGroup);
begin
  assert(value.ClassName = 'TFhirQuestionnaireGroup', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireGroup');
  add(value);
end;


function TFhirQuestionnaireGroupList.Append: TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireGroupList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireGroupList.GetEnumerator : TFhirQuestionnaireGroupListEnumerator;
begin
  result := TFhirQuestionnaireGroupListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireGroupList.Clone: TFhirQuestionnaireGroupList;
begin
  result := TFhirQuestionnaireGroupList(inherited Clone);
end;

function TFhirQuestionnaireGroupList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireGroupList.GetItemN(index: Integer): TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup(ObjectByIndex[index]);
end;

function TFhirQuestionnaireGroupList.IndexOf(value: TFhirQuestionnaireGroup): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireGroupList.Insert(index: Integer): TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireGroupList.InsertItem(index: Integer; value: TFhirQuestionnaireGroup);
begin
  assert(value is TFhirQuestionnaireGroup);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireGroupList.Item(index: Integer): TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup(ObjectByIndex[index]);
end;

function TFhirQuestionnaireGroupList.Link: TFhirQuestionnaireGroupList;
begin
  result := TFhirQuestionnaireGroupList(inherited Link);
end;

procedure TFhirQuestionnaireGroupList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireGroupList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireGroup);
begin
  assert(value is TFhirQuestionnaireGroup);
  FhirQuestionnaireGroups[index] := value;
end;

procedure TFhirQuestionnaireGroupList.SetItemN(index: Integer; value: TFhirQuestionnaireGroup);
begin
  assert(value is TFhirQuestionnaireGroup);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireGroupQuestion }

constructor TFhirQuestionnaireGroupQuestion.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireGroupQuestion.Destroy;
begin
  FLinkId.free;
  FConceptList.Free;
  FText.free;
  FType_.free;
  FRequired.free;
  FRepeats.free;
  FOptions.free;
  FOptionList.Free;
  FGroupList.Free;
  inherited;
end;

procedure TFhirQuestionnaireGroupQuestion.Assign(oSource : TAdvObject);
begin
  inherited;
  linkIdElement := TFhirQuestionnaireGroupQuestion(oSource).linkIdElement.Clone;
  if (TFhirQuestionnaireGroupQuestion(oSource).FConceptList = nil) then
  begin
    FConceptList.free;
    FConceptList := nil;
  end
  else
  begin
    FConceptList := TFhirCodingList.Create;
    FConceptList.Assign(TFhirQuestionnaireGroupQuestion(oSource).FConceptList);
  end;
  textElement := TFhirQuestionnaireGroupQuestion(oSource).textElement.Clone;
  FType_ := TFhirQuestionnaireGroupQuestion(oSource).FType_.Link;
  requiredElement := TFhirQuestionnaireGroupQuestion(oSource).requiredElement.Clone;
  repeatsElement := TFhirQuestionnaireGroupQuestion(oSource).repeatsElement.Clone;
  options := TFhirQuestionnaireGroupQuestion(oSource).options.Clone;
  if (TFhirQuestionnaireGroupQuestion(oSource).FOptionList = nil) then
  begin
    FOptionList.free;
    FOptionList := nil;
  end
  else
  begin
    FOptionList := TFhirCodingList.Create;
    FOptionList.Assign(TFhirQuestionnaireGroupQuestion(oSource).FOptionList);
  end;
  if (TFhirQuestionnaireGroupQuestion(oSource).FGroupList = nil) then
  begin
    FGroupList.free;
    FGroupList := nil;
  end
  else
  begin
    FGroupList := TFhirQuestionnaireGroupList.Create;
    FGroupList.Assign(TFhirQuestionnaireGroupQuestion(oSource).FGroupList);
  end;
end;

procedure TFhirQuestionnaireGroupQuestion.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'linkId') Then
     list.add(FLinkId.Link);
  if (child_name = 'concept') Then
    list.addAll(FConceptList);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'required') Then
     list.add(FRequired.Link);
  if (child_name = 'repeats') Then
     list.add(FRepeats.Link);
  if (child_name = 'options') Then
     list.add(FOptions.Link);
  if (child_name = 'option') Then
    list.addAll(FOptionList);
  if (child_name = 'group') Then
    list.addAll(FGroupList);
end;

procedure TFhirQuestionnaireGroupQuestion.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'linkId', 'string', FLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'concept', 'Coding', FConceptList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'required', 'boolean', FRequired.Link));{2}
  oList.add(TFHIRProperty.create(self, 'repeats', 'boolean', FRepeats.Link));{2}
  oList.add(TFHIRProperty.create(self, 'options', 'Reference(ValueSet)', FOptions.Link));{2}
  oList.add(TFHIRProperty.create(self, 'option', 'Coding', FOptionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'group', '@Questionnaire.group', FGroupList.Link)){3};
end;

procedure TFhirQuestionnaireGroupQuestion.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'linkId') then LinkIdElement := propValue as TFhirString{5a}
  else if (propName = 'concept') then ConceptList.add(propValue as TFhirCoding){2}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'required') then RequiredElement := propValue as TFhirBoolean{5a}
  else if (propName = 'repeats') then RepeatsElement := propValue as TFhirBoolean{5a}
  else if (propName = 'options') then Options := propValue as TFhirReference{TFhirValueSet}{4b}
  else if (propName = 'option') then OptionList.add(propValue as TFhirCoding){2}
  else if (propName = 'group') then GroupList.add(propValue as TFhirQuestionnaireGroup){2}
  else inherited;
end;

function TFhirQuestionnaireGroupQuestion.FhirType : string;
begin
  result := 'question';
end;

function TFhirQuestionnaireGroupQuestion.Link : TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion(inherited Link);
end;

function TFhirQuestionnaireGroupQuestion.Clone : TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion(inherited Clone);
end;

function TFhirQuestionnaireGroupQuestion.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireGroupQuestion;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireGroupQuestion)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireGroupQuestion(other);
    result := compareDeep(linkIdElement, o.linkIdElement, true) and compareDeep(conceptList, o.conceptList, true) and 
      compareDeep(textElement, o.textElement, true) and compareDeep(type_Element, o.type_Element, true) and 
      compareDeep(requiredElement, o.requiredElement, true) and compareDeep(repeatsElement, o.repeatsElement, true) and 
      compareDeep(optionsElement, o.optionsElement, true) and compareDeep(optionList, o.optionList, true) and 
      compareDeep(groupList, o.groupList, true);
  end;
end;

function TFhirQuestionnaireGroupQuestion.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireGroupQuestion;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireGroupQuestion)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireGroupQuestion(other);
    result := compareValues(linkIdElement, o.linkIdElement, true) and compareValues(textElement, o.textElement, true) and 
      compareValues(type_Element, o.type_Element, true) and compareValues(requiredElement, o.requiredElement, true) and 
      compareValues(repeatsElement, o.repeatsElement, true);
  end;
end;

{ TFhirQuestionnaireGroupQuestion }

Procedure TFhirQuestionnaireGroupQuestion.SetLinkId(value : TFhirString);
begin
  FLinkId.free;
  FLinkId := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetLinkIdST : String;
begin
  if FLinkId = nil then
    result := ''
  else
    result := FLinkId.value;
end;

Procedure TFhirQuestionnaireGroupQuestion.SetLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FLinkId = nil then
      FLinkId := TFhirString.create;
    FLinkId.value := value
  end
  else if FLinkId <> nil then
    FLinkId.value := '';
end;

Function TFhirQuestionnaireGroupQuestion.GetConceptList : TFhirCodingList;
begin
  if FConceptList = nil then
    FConceptList := TFhirCodingList.Create;
  result := FConceptList;
end;

Function TFhirQuestionnaireGroupQuestion.GetHasConceptList : boolean;
begin
  result := (FConceptList <> nil) and (FConceptList.count > 0);
end;

Procedure TFhirQuestionnaireGroupQuestion.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirQuestionnaireGroupQuestion.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirQuestionnaireGroupQuestion.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetType_ST : TFhirAnswerFormat;
begin
  if FType_ = nil then
    result := TFhirAnswerFormat(0)
  else
    result := TFhirAnswerFormat(StringArrayIndexOfSensitive(CODES_TFhirAnswerFormat, FType_.value));
end;

Procedure TFhirQuestionnaireGroupQuestion.SetType_ST(value : TFhirAnswerFormat);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirAnswerFormat[value]);
end;

Procedure TFhirQuestionnaireGroupQuestion.SetRequired(value : TFhirBoolean);
begin
  FRequired.free;
  FRequired := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetRequiredST : Boolean;
begin
  if FRequired = nil then
    result := false
  else
    result := FRequired.value;
end;

Procedure TFhirQuestionnaireGroupQuestion.SetRequiredST(value : Boolean);
begin
  if FRequired = nil then
    FRequired := TFhirBoolean.create;
  FRequired.value := value
end;

Procedure TFhirQuestionnaireGroupQuestion.SetRepeats(value : TFhirBoolean);
begin
  FRepeats.free;
  FRepeats := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetRepeatsST : Boolean;
begin
  if FRepeats = nil then
    result := false
  else
    result := FRepeats.value;
end;

Procedure TFhirQuestionnaireGroupQuestion.SetRepeatsST(value : Boolean);
begin
  if FRepeats = nil then
    FRepeats := TFhirBoolean.create;
  FRepeats.value := value
end;

Procedure TFhirQuestionnaireGroupQuestion.SetOptions(value : TFhirReference{TFhirValueSet});
begin
  FOptions.free;
  FOptions := value;
end;

Function TFhirQuestionnaireGroupQuestion.GetOptionList : TFhirCodingList;
begin
  if FOptionList = nil then
    FOptionList := TFhirCodingList.Create;
  result := FOptionList;
end;

Function TFhirQuestionnaireGroupQuestion.GetHasOptionList : boolean;
begin
  result := (FOptionList <> nil) and (FOptionList.count > 0);
end;

Function TFhirQuestionnaireGroupQuestion.GetGroupList : TFhirQuestionnaireGroupList;
begin
  if FGroupList = nil then
    FGroupList := TFhirQuestionnaireGroupList.Create;
  result := FGroupList;
end;

Function TFhirQuestionnaireGroupQuestion.GetHasGroupList : boolean;
begin
  result := (FGroupList <> nil) and (FGroupList.count > 0);
end;


{ TFhirQuestionnaireGroupQuestionListEnumerator }

Constructor TFhirQuestionnaireGroupQuestionListEnumerator.Create(list : TFhirQuestionnaireGroupQuestionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireGroupQuestionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireGroupQuestionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireGroupQuestionListEnumerator.GetCurrent : TFhirQuestionnaireGroupQuestion;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireGroupQuestionList }
procedure TFhirQuestionnaireGroupQuestionList.AddItem(value: TFhirQuestionnaireGroupQuestion);
begin
  assert(value.ClassName = 'TFhirQuestionnaireGroupQuestion', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireGroupQuestion');
  add(value);
end;


function TFhirQuestionnaireGroupQuestionList.Append: TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireGroupQuestionList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireGroupQuestionList.GetEnumerator : TFhirQuestionnaireGroupQuestionListEnumerator;
begin
  result := TFhirQuestionnaireGroupQuestionListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireGroupQuestionList.Clone: TFhirQuestionnaireGroupQuestionList;
begin
  result := TFhirQuestionnaireGroupQuestionList(inherited Clone);
end;

function TFhirQuestionnaireGroupQuestionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireGroupQuestionList.GetItemN(index: Integer): TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion(ObjectByIndex[index]);
end;

function TFhirQuestionnaireGroupQuestionList.IndexOf(value: TFhirQuestionnaireGroupQuestion): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireGroupQuestionList.Insert(index: Integer): TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireGroupQuestionList.InsertItem(index: Integer; value: TFhirQuestionnaireGroupQuestion);
begin
  assert(value is TFhirQuestionnaireGroupQuestion);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireGroupQuestionList.Item(index: Integer): TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion(ObjectByIndex[index]);
end;

function TFhirQuestionnaireGroupQuestionList.Link: TFhirQuestionnaireGroupQuestionList;
begin
  result := TFhirQuestionnaireGroupQuestionList(inherited Link);
end;

procedure TFhirQuestionnaireGroupQuestionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireGroupQuestionList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireGroupQuestion);
begin
  assert(value is TFhirQuestionnaireGroupQuestion);
  FhirQuestionnaireGroupQuestions[index] := value;
end;

procedure TFhirQuestionnaireGroupQuestionList.SetItemN(index: Integer; value: TFhirQuestionnaireGroupQuestion);
begin
  assert(value is TFhirQuestionnaireGroupQuestion);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaire }

constructor TFhirQuestionnaire.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaire.Destroy;
begin
  FIdentifierList.Free;
  FVersion.free;
  FStatus.free;
  FDate.free;
  FPublisher.free;
  FTelecomList.Free;
  FSubjectTypeList.Free;
  FGroup.free;
  inherited;
end;

function TFhirQuestionnaire.GetResourceType : TFhirResourceType;
begin
  result := frtQuestionnaire;
end;

procedure TFhirQuestionnaire.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirQuestionnaire(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirQuestionnaire(oSource).FIdentifierList);
  end;
  versionElement := TFhirQuestionnaire(oSource).versionElement.Clone;
  FStatus := TFhirQuestionnaire(oSource).FStatus.Link;
  dateElement := TFhirQuestionnaire(oSource).dateElement.Clone;
  publisherElement := TFhirQuestionnaire(oSource).publisherElement.Clone;
  if (TFhirQuestionnaire(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirQuestionnaire(oSource).FTelecomList);
  end;
  if (TFhirQuestionnaire(oSource).FSubjectTypeList = nil) then
  begin
    FSubjectTypeList.free;
    FSubjectTypeList := nil;
  end
  else
  begin
    FSubjectTypeList := TFhirCodeList.Create;
    FSubjectTypeList.Assign(TFhirQuestionnaire(oSource).FSubjectTypeList);
  end;
  group := TFhirQuestionnaire(oSource).group.Clone;
end;

procedure TFhirQuestionnaire.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'subjectType') Then
    list.addAll(FSubjectTypeList);
  if (child_name = 'group') Then
     list.add(FGroup.Link);
end;

procedure TFhirQuestionnaire.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subjectType', 'code', FSubjectTypeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'group', '', FGroup.Link));{2}
end;

procedure TFhirQuestionnaire.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'subjectType') then SubjectTypeList.add(propValue as TFhirCode){2}
  else if (propName = 'group') then Group := propValue as TFhirQuestionnaireGroup{4b}
  else inherited;
end;

function TFhirQuestionnaire.FhirType : string;
begin
  result := 'Questionnaire';
end;

function TFhirQuestionnaire.Link : TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire(inherited Link);
end;

function TFhirQuestionnaire.Clone : TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire(inherited Clone);
end;

{ TFhirQuestionnaire }

Function TFhirQuestionnaire.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirQuestionnaire.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirQuestionnaire.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirQuestionnaire.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirQuestionnaire.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirQuestionnaire.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirQuestionnaire.GetStatusST : TFhirQuestionnaireStatus;
begin
  if FStatus = nil then
    result := TFhirQuestionnaireStatus(0)
  else
    result := TFhirQuestionnaireStatus(StringArrayIndexOfSensitive(CODES_TFhirQuestionnaireStatus, FStatus.value));
end;

Procedure TFhirQuestionnaire.SetStatusST(value : TFhirQuestionnaireStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirQuestionnaireStatus[value]);
end;

Procedure TFhirQuestionnaire.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirQuestionnaire.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirQuestionnaire.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirQuestionnaire.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirQuestionnaire.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirQuestionnaire.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirQuestionnaire.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirQuestionnaire.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Function TFhirQuestionnaire.GetSubjectTypeList : TFhirCodeList;
begin
  if FSubjectTypeList = nil then
    FSubjectTypeList := TFhirCodeList.Create;
  result := FSubjectTypeList;
end;

Function TFhirQuestionnaire.GetHasSubjectTypeList : boolean;
begin
  result := (FSubjectTypeList <> nil) and (FSubjectTypeList.count > 0);
end;

Procedure TFhirQuestionnaire.SetGroup(value : TFhirQuestionnaireGroup);
begin
  FGroup.free;
  FGroup := value;
end;


{ TFhirQuestionnaireListEnumerator }

Constructor TFhirQuestionnaireListEnumerator.Create(list : TFhirQuestionnaireList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireListEnumerator.GetCurrent : TFhirQuestionnaire;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireList }
procedure TFhirQuestionnaireList.AddItem(value: TFhirQuestionnaire);
begin
  assert(value.ClassName = 'TFhirQuestionnaire', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaire');
  add(value);
end;


function TFhirQuestionnaireList.Append: TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireList.GetEnumerator : TFhirQuestionnaireListEnumerator;
begin
  result := TFhirQuestionnaireListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireList.Clone: TFhirQuestionnaireList;
begin
  result := TFhirQuestionnaireList(inherited Clone);
end;

function TFhirQuestionnaireList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireList.GetItemN(index: Integer): TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire(ObjectByIndex[index]);
end;

function TFhirQuestionnaireList.IndexOf(value: TFhirQuestionnaire): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireList.Insert(index: Integer): TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireList.InsertItem(index: Integer; value: TFhirQuestionnaire);
begin
  assert(value is TFhirQuestionnaire);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireList.Item(index: Integer): TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire(ObjectByIndex[index]);
end;

function TFhirQuestionnaireList.Link: TFhirQuestionnaireList;
begin
  result := TFhirQuestionnaireList(inherited Link);
end;

procedure TFhirQuestionnaireList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireList.SetItemByIndex(index: Integer; value: TFhirQuestionnaire);
begin
  assert(value is TFhirQuestionnaire);
  FhirQuestionnaires[index] := value;
end;

procedure TFhirQuestionnaireList.SetItemN(index: Integer; value: TFhirQuestionnaire);
begin
  assert(value is TFhirQuestionnaire);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireResponseGroup }

constructor TFhirQuestionnaireResponseGroup.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireResponseGroup.Destroy;
begin
  FLinkId.free;
  FTitle.free;
  FText.free;
  FSubject.free;
  FGroupList.Free;
  FQuestionList.Free;
  inherited;
end;

procedure TFhirQuestionnaireResponseGroup.Assign(oSource : TAdvObject);
begin
  inherited;
  linkIdElement := TFhirQuestionnaireResponseGroup(oSource).linkIdElement.Clone;
  titleElement := TFhirQuestionnaireResponseGroup(oSource).titleElement.Clone;
  textElement := TFhirQuestionnaireResponseGroup(oSource).textElement.Clone;
  subject := TFhirQuestionnaireResponseGroup(oSource).subject.Clone;
  if (TFhirQuestionnaireResponseGroup(oSource).FGroupList = nil) then
  begin
    FGroupList.free;
    FGroupList := nil;
  end
  else
  begin
    FGroupList := TFhirQuestionnaireResponseGroupList.Create;
    FGroupList.Assign(TFhirQuestionnaireResponseGroup(oSource).FGroupList);
  end;
  if (TFhirQuestionnaireResponseGroup(oSource).FQuestionList = nil) then
  begin
    FQuestionList.free;
    FQuestionList := nil;
  end
  else
  begin
    FQuestionList := TFhirQuestionnaireResponseGroupQuestionList.Create;
    FQuestionList.Assign(TFhirQuestionnaireResponseGroup(oSource).FQuestionList);
  end;
end;

procedure TFhirQuestionnaireResponseGroup.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'linkId') Then
     list.add(FLinkId.Link);
  if (child_name = 'title') Then
     list.add(FTitle.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'group') Then
    list.addAll(FGroupList);
  if (child_name = 'question') Then
    list.addAll(FQuestionList);
end;

procedure TFhirQuestionnaireResponseGroup.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'linkId', 'string', FLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'title', 'string', FTitle.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'group', '@QuestionnaireResponse.group', FGroupList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'question', '', FQuestionList.Link)){3};
end;

procedure TFhirQuestionnaireResponseGroup.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'linkId') then LinkIdElement := propValue as TFhirString{5a}
  else if (propName = 'title') then TitleElement := propValue as TFhirString{5a}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'group') then GroupList.add(propValue as TFhirQuestionnaireResponseGroup){2}
  else if (propName = 'question') then QuestionList.add(propValue as TFhirQuestionnaireResponseGroupQuestion){2}
  else inherited;
end;

function TFhirQuestionnaireResponseGroup.FhirType : string;
begin
  result := 'group';
end;

function TFhirQuestionnaireResponseGroup.Link : TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup(inherited Link);
end;

function TFhirQuestionnaireResponseGroup.Clone : TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup(inherited Clone);
end;

function TFhirQuestionnaireResponseGroup.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroup;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroup)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroup(other);
    result := compareDeep(linkIdElement, o.linkIdElement, true) and compareDeep(titleElement, o.titleElement, true) and 
      compareDeep(textElement, o.textElement, true) and compareDeep(subjectElement, o.subjectElement, true) and 
      compareDeep(groupList, o.groupList, true) and compareDeep(questionList, o.questionList, true);
  end;
end;

function TFhirQuestionnaireResponseGroup.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroup;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroup)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroup(other);
    result := compareValues(linkIdElement, o.linkIdElement, true) and compareValues(titleElement, o.titleElement, true) and 
      compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirQuestionnaireResponseGroup }

Procedure TFhirQuestionnaireResponseGroup.SetLinkId(value : TFhirString);
begin
  FLinkId.free;
  FLinkId := value;
end;

Function TFhirQuestionnaireResponseGroup.GetLinkIdST : String;
begin
  if FLinkId = nil then
    result := ''
  else
    result := FLinkId.value;
end;

Procedure TFhirQuestionnaireResponseGroup.SetLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FLinkId = nil then
      FLinkId := TFhirString.create;
    FLinkId.value := value
  end
  else if FLinkId <> nil then
    FLinkId.value := '';
end;

Procedure TFhirQuestionnaireResponseGroup.SetTitle(value : TFhirString);
begin
  FTitle.free;
  FTitle := value;
end;

Function TFhirQuestionnaireResponseGroup.GetTitleST : String;
begin
  if FTitle = nil then
    result := ''
  else
    result := FTitle.value;
end;

Procedure TFhirQuestionnaireResponseGroup.SetTitleST(value : String);
begin
  if value <> '' then
  begin
    if FTitle = nil then
      FTitle := TFhirString.create;
    FTitle.value := value
  end
  else if FTitle <> nil then
    FTitle.value := '';
end;

Procedure TFhirQuestionnaireResponseGroup.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirQuestionnaireResponseGroup.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirQuestionnaireResponseGroup.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Procedure TFhirQuestionnaireResponseGroup.SetSubject(value : TFhirReference{TFhirReference});
begin
  FSubject.free;
  FSubject := value;
end;

Function TFhirQuestionnaireResponseGroup.GetGroupList : TFhirQuestionnaireResponseGroupList;
begin
  if FGroupList = nil then
    FGroupList := TFhirQuestionnaireResponseGroupList.Create;
  result := FGroupList;
end;

Function TFhirQuestionnaireResponseGroup.GetHasGroupList : boolean;
begin
  result := (FGroupList <> nil) and (FGroupList.count > 0);
end;

Function TFhirQuestionnaireResponseGroup.GetQuestionList : TFhirQuestionnaireResponseGroupQuestionList;
begin
  if FQuestionList = nil then
    FQuestionList := TFhirQuestionnaireResponseGroupQuestionList.Create;
  result := FQuestionList;
end;

Function TFhirQuestionnaireResponseGroup.GetHasQuestionList : boolean;
begin
  result := (FQuestionList <> nil) and (FQuestionList.count > 0);
end;


{ TFhirQuestionnaireResponseGroupListEnumerator }

Constructor TFhirQuestionnaireResponseGroupListEnumerator.Create(list : TFhirQuestionnaireResponseGroupList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireResponseGroupListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireResponseGroupListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireResponseGroupListEnumerator.GetCurrent : TFhirQuestionnaireResponseGroup;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireResponseGroupList }
procedure TFhirQuestionnaireResponseGroupList.AddItem(value: TFhirQuestionnaireResponseGroup);
begin
  assert(value.ClassName = 'TFhirQuestionnaireResponseGroup', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireResponseGroup');
  add(value);
end;


function TFhirQuestionnaireResponseGroupList.Append: TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireResponseGroupList.GetEnumerator : TFhirQuestionnaireResponseGroupListEnumerator;
begin
  result := TFhirQuestionnaireResponseGroupListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireResponseGroupList.Clone: TFhirQuestionnaireResponseGroupList;
begin
  result := TFhirQuestionnaireResponseGroupList(inherited Clone);
end;

function TFhirQuestionnaireResponseGroupList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireResponseGroupList.GetItemN(index: Integer): TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupList.IndexOf(value: TFhirQuestionnaireResponseGroup): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireResponseGroupList.Insert(index: Integer): TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupList.InsertItem(index: Integer; value: TFhirQuestionnaireResponseGroup);
begin
  assert(value is TFhirQuestionnaireResponseGroup);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireResponseGroupList.Item(index: Integer): TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupList.Link: TFhirQuestionnaireResponseGroupList;
begin
  result := TFhirQuestionnaireResponseGroupList(inherited Link);
end;

procedure TFhirQuestionnaireResponseGroupList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireResponseGroupList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireResponseGroup);
begin
  assert(value is TFhirQuestionnaireResponseGroup);
  FhirQuestionnaireResponseGroups[index] := value;
end;

procedure TFhirQuestionnaireResponseGroupList.SetItemN(index: Integer; value: TFhirQuestionnaireResponseGroup);
begin
  assert(value is TFhirQuestionnaireResponseGroup);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireResponseGroupQuestion }

constructor TFhirQuestionnaireResponseGroupQuestion.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireResponseGroupQuestion.Destroy;
begin
  FLinkId.free;
  FText.free;
  FAnswerList.Free;
  inherited;
end;

procedure TFhirQuestionnaireResponseGroupQuestion.Assign(oSource : TAdvObject);
begin
  inherited;
  linkIdElement := TFhirQuestionnaireResponseGroupQuestion(oSource).linkIdElement.Clone;
  textElement := TFhirQuestionnaireResponseGroupQuestion(oSource).textElement.Clone;
  if (TFhirQuestionnaireResponseGroupQuestion(oSource).FAnswerList = nil) then
  begin
    FAnswerList.free;
    FAnswerList := nil;
  end
  else
  begin
    FAnswerList := TFhirQuestionnaireResponseGroupQuestionAnswerList.Create;
    FAnswerList.Assign(TFhirQuestionnaireResponseGroupQuestion(oSource).FAnswerList);
  end;
end;

procedure TFhirQuestionnaireResponseGroupQuestion.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'linkId') Then
     list.add(FLinkId.Link);
  if (child_name = 'text') Then
     list.add(FText.Link);
  if (child_name = 'answer') Then
    list.addAll(FAnswerList);
end;

procedure TFhirQuestionnaireResponseGroupQuestion.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'linkId', 'string', FLinkId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'text', 'string', FText.Link));{2}
  oList.add(TFHIRProperty.create(self, 'answer', '', FAnswerList.Link)){3};
end;

procedure TFhirQuestionnaireResponseGroupQuestion.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'linkId') then LinkIdElement := propValue as TFhirString{5a}
  else if (propName = 'text') then TextElement := propValue as TFhirString{5a}
  else if (propName = 'answer') then AnswerList.add(propValue as TFhirQuestionnaireResponseGroupQuestionAnswer){2}
  else inherited;
end;

function TFhirQuestionnaireResponseGroupQuestion.FhirType : string;
begin
  result := 'question';
end;

function TFhirQuestionnaireResponseGroupQuestion.Link : TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion(inherited Link);
end;

function TFhirQuestionnaireResponseGroupQuestion.Clone : TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion(inherited Clone);
end;

function TFhirQuestionnaireResponseGroupQuestion.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroupQuestion;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroupQuestion)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroupQuestion(other);
    result := compareDeep(linkIdElement, o.linkIdElement, true) and compareDeep(textElement, o.textElement, true) and 
      compareDeep(answerList, o.answerList, true);
  end;
end;

function TFhirQuestionnaireResponseGroupQuestion.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroupQuestion;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroupQuestion)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroupQuestion(other);
    result := compareValues(linkIdElement, o.linkIdElement, true) and compareValues(textElement, o.textElement, true);
  end;
end;

{ TFhirQuestionnaireResponseGroupQuestion }

Procedure TFhirQuestionnaireResponseGroupQuestion.SetLinkId(value : TFhirString);
begin
  FLinkId.free;
  FLinkId := value;
end;

Function TFhirQuestionnaireResponseGroupQuestion.GetLinkIdST : String;
begin
  if FLinkId = nil then
    result := ''
  else
    result := FLinkId.value;
end;

Procedure TFhirQuestionnaireResponseGroupQuestion.SetLinkIdST(value : String);
begin
  if value <> '' then
  begin
    if FLinkId = nil then
      FLinkId := TFhirString.create;
    FLinkId.value := value
  end
  else if FLinkId <> nil then
    FLinkId.value := '';
end;

Procedure TFhirQuestionnaireResponseGroupQuestion.SetText(value : TFhirString);
begin
  FText.free;
  FText := value;
end;

Function TFhirQuestionnaireResponseGroupQuestion.GetTextST : String;
begin
  if FText = nil then
    result := ''
  else
    result := FText.value;
end;

Procedure TFhirQuestionnaireResponseGroupQuestion.SetTextST(value : String);
begin
  if value <> '' then
  begin
    if FText = nil then
      FText := TFhirString.create;
    FText.value := value
  end
  else if FText <> nil then
    FText.value := '';
end;

Function TFhirQuestionnaireResponseGroupQuestion.GetAnswerList : TFhirQuestionnaireResponseGroupQuestionAnswerList;
begin
  if FAnswerList = nil then
    FAnswerList := TFhirQuestionnaireResponseGroupQuestionAnswerList.Create;
  result := FAnswerList;
end;

Function TFhirQuestionnaireResponseGroupQuestion.GetHasAnswerList : boolean;
begin
  result := (FAnswerList <> nil) and (FAnswerList.count > 0);
end;


{ TFhirQuestionnaireResponseGroupQuestionListEnumerator }

Constructor TFhirQuestionnaireResponseGroupQuestionListEnumerator.Create(list : TFhirQuestionnaireResponseGroupQuestionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireResponseGroupQuestionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireResponseGroupQuestionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireResponseGroupQuestionListEnumerator.GetCurrent : TFhirQuestionnaireResponseGroupQuestion;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireResponseGroupQuestionList }
procedure TFhirQuestionnaireResponseGroupQuestionList.AddItem(value: TFhirQuestionnaireResponseGroupQuestion);
begin
  assert(value.ClassName = 'TFhirQuestionnaireResponseGroupQuestion', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireResponseGroupQuestion');
  add(value);
end;


function TFhirQuestionnaireResponseGroupQuestionList.Append: TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupQuestionList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireResponseGroupQuestionList.GetEnumerator : TFhirQuestionnaireResponseGroupQuestionListEnumerator;
begin
  result := TFhirQuestionnaireResponseGroupQuestionListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireResponseGroupQuestionList.Clone: TFhirQuestionnaireResponseGroupQuestionList;
begin
  result := TFhirQuestionnaireResponseGroupQuestionList(inherited Clone);
end;

function TFhirQuestionnaireResponseGroupQuestionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireResponseGroupQuestionList.GetItemN(index: Integer): TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupQuestionList.IndexOf(value: TFhirQuestionnaireResponseGroupQuestion): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireResponseGroupQuestionList.Insert(index: Integer): TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupQuestionList.InsertItem(index: Integer; value: TFhirQuestionnaireResponseGroupQuestion);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestion);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireResponseGroupQuestionList.Item(index: Integer): TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupQuestionList.Link: TFhirQuestionnaireResponseGroupQuestionList;
begin
  result := TFhirQuestionnaireResponseGroupQuestionList(inherited Link);
end;

procedure TFhirQuestionnaireResponseGroupQuestionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireResponseGroupQuestionList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireResponseGroupQuestion);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestion);
  FhirQuestionnaireResponseGroupQuestions[index] := value;
end;

procedure TFhirQuestionnaireResponseGroupQuestionList.SetItemN(index: Integer; value: TFhirQuestionnaireResponseGroupQuestion);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestion);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireResponseGroupQuestionAnswer }

constructor TFhirQuestionnaireResponseGroupQuestionAnswer.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireResponseGroupQuestionAnswer.Destroy;
begin
  FValue.free;
  FGroupList.Free;
  inherited;
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswer.Assign(oSource : TAdvObject);
begin
  inherited;
  value := TFhirQuestionnaireResponseGroupQuestionAnswer(oSource).value.Clone;
  if (TFhirQuestionnaireResponseGroupQuestionAnswer(oSource).FGroupList = nil) then
  begin
    FGroupList.free;
    FGroupList := nil;
  end
  else
  begin
    FGroupList := TFhirQuestionnaireResponseGroupList.Create;
    FGroupList.Assign(TFhirQuestionnaireResponseGroupQuestionAnswer(oSource).FGroupList);
  end;
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswer.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
  if (child_name = 'group') Then
    list.addAll(FGroupList);
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswer.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'value[x]', 'boolean|decimal|integer|date|dateTime|instant|time|string|uri|Attachment|Coding|Quantity|Reference(Any)', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'group', '@QuestionnaireResponse.group', FGroupList.Link)){3};
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswer.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else if (propName = 'group') then GroupList.add(propValue as TFhirQuestionnaireResponseGroup){2}
  else inherited;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswer.FhirType : string;
begin
  result := 'answer';
end;

function TFhirQuestionnaireResponseGroupQuestionAnswer.Link : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer(inherited Link);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswer.Clone : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer(inherited Clone);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswer.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroupQuestionAnswer)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroupQuestionAnswer(other);
    result := compareDeep(valueElement, o.valueElement, true) and compareDeep(groupList, o.groupList, true);
  end;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswer.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirQuestionnaireResponseGroupQuestionAnswer)) then
    result := false
  else
  begin
    o := TFhirQuestionnaireResponseGroupQuestionAnswer(other);
    result := true;
  end;
end;

{ TFhirQuestionnaireResponseGroupQuestionAnswer }

Procedure TFhirQuestionnaireResponseGroupQuestionAnswer.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirQuestionnaireResponseGroupQuestionAnswer.GetGroupList : TFhirQuestionnaireResponseGroupList;
begin
  if FGroupList = nil then
    FGroupList := TFhirQuestionnaireResponseGroupList.Create;
  result := FGroupList;
end;

Function TFhirQuestionnaireResponseGroupQuestionAnswer.GetHasGroupList : boolean;
begin
  result := (FGroupList <> nil) and (FGroupList.count > 0);
end;


{ TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator }

Constructor TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator.Create(list : TFhirQuestionnaireResponseGroupQuestionAnswerList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator.GetCurrent : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireResponseGroupQuestionAnswerList }
procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.AddItem(value: TFhirQuestionnaireResponseGroupQuestionAnswer);
begin
  assert(value.ClassName = 'TFhirQuestionnaireResponseGroupQuestionAnswer', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireResponseGroupQuestionAnswer');
  add(value);
end;


function TFhirQuestionnaireResponseGroupQuestionAnswerList.Append: TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.GetEnumerator : TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswerListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.Clone: TFhirQuestionnaireResponseGroupQuestionAnswerList;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswerList(inherited Clone);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.GetItemN(index: Integer): TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.IndexOf(value: TFhirQuestionnaireResponseGroupQuestionAnswer): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireResponseGroupQuestionAnswerList.Insert(index: Integer): TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.InsertItem(index: Integer; value: TFhirQuestionnaireResponseGroupQuestionAnswer);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestionAnswer);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.Item(index: Integer): TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseGroupQuestionAnswerList.Link: TFhirQuestionnaireResponseGroupQuestionAnswerList;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswerList(inherited Link);
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireResponseGroupQuestionAnswer);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestionAnswer);
  FhirQuestionnaireResponseGroupQuestionAnswers[index] := value;
end;

procedure TFhirQuestionnaireResponseGroupQuestionAnswerList.SetItemN(index: Integer; value: TFhirQuestionnaireResponseGroupQuestionAnswer);
begin
  assert(value is TFhirQuestionnaireResponseGroupQuestionAnswer);
  ObjectByIndex[index] := value;
end;

{ TFhirQuestionnaireResponse }

constructor TFhirQuestionnaireResponse.Create;
begin
  inherited;
end;

destructor TFhirQuestionnaireResponse.Destroy;
begin
  FIdentifier.free;
  FQuestionnaire.free;
  FStatus.free;
  FSubject.free;
  FAuthor.free;
  FAuthored.free;
  FSource.free;
  FEncounter.free;
  FGroup.free;
  inherited;
end;

function TFhirQuestionnaireResponse.GetResourceType : TFhirResourceType;
begin
  result := frtQuestionnaireResponse;
end;

procedure TFhirQuestionnaireResponse.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirQuestionnaireResponse(oSource).identifier.Clone;
  questionnaire := TFhirQuestionnaireResponse(oSource).questionnaire.Clone;
  FStatus := TFhirQuestionnaireResponse(oSource).FStatus.Link;
  subject := TFhirQuestionnaireResponse(oSource).subject.Clone;
  author := TFhirQuestionnaireResponse(oSource).author.Clone;
  authoredElement := TFhirQuestionnaireResponse(oSource).authoredElement.Clone;
  source := TFhirQuestionnaireResponse(oSource).source.Clone;
  encounter := TFhirQuestionnaireResponse(oSource).encounter.Clone;
  group := TFhirQuestionnaireResponse(oSource).group.Clone;
end;

procedure TFhirQuestionnaireResponse.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'questionnaire') Then
     list.add(FQuestionnaire.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'author') Then
     list.add(FAuthor.Link);
  if (child_name = 'authored') Then
     list.add(FAuthored.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'group') Then
     list.add(FGroup.Link);
end;

procedure TFhirQuestionnaireResponse.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'questionnaire', 'Reference(Questionnaire)', FQuestionnaire.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Any)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'author', 'Reference(Device|Practitioner|Patient|RelatedPerson)', FAuthor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'authored', 'dateTime', FAuthored.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Patient|Practitioner|RelatedPerson)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'group', '', FGroup.Link));{2}
end;

procedure TFhirQuestionnaireResponse.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'questionnaire') then Questionnaire := propValue as TFhirReference{TFhirQuestionnaire}{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'subject') then Subject := propValue as TFhirReference{TFhirReference}{4b}
  else if (propName = 'author') then Author := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'authored') then AuthoredElement := propValue as TFhirDateTime{5a}
  else if (propName = 'source') then Source := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'group') then Group := propValue as TFhirQuestionnaireResponseGroup{4b}
  else inherited;
end;

function TFhirQuestionnaireResponse.FhirType : string;
begin
  result := 'QuestionnaireResponse';
end;

function TFhirQuestionnaireResponse.Link : TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse(inherited Link);
end;

function TFhirQuestionnaireResponse.Clone : TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse(inherited Clone);
end;

{ TFhirQuestionnaireResponse }

Procedure TFhirQuestionnaireResponse.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirQuestionnaireResponse.SetQuestionnaire(value : TFhirReference{TFhirQuestionnaire});
begin
  FQuestionnaire.free;
  FQuestionnaire := value;
end;

Procedure TFhirQuestionnaireResponse.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirQuestionnaireResponse.GetStatusST : TFhirQuestionnaireAnswersStatus;
begin
  if FStatus = nil then
    result := TFhirQuestionnaireAnswersStatus(0)
  else
    result := TFhirQuestionnaireAnswersStatus(StringArrayIndexOfSensitive(CODES_TFhirQuestionnaireAnswersStatus, FStatus.value));
end;

Procedure TFhirQuestionnaireResponse.SetStatusST(value : TFhirQuestionnaireAnswersStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirQuestionnaireAnswersStatus[value]);
end;

Procedure TFhirQuestionnaireResponse.SetSubject(value : TFhirReference{TFhirReference});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirQuestionnaireResponse.SetAuthor(value : TFhirReference{Resource});
begin
  FAuthor.free;
  FAuthor := value;
end;

Procedure TFhirQuestionnaireResponse.SetAuthored(value : TFhirDateTime);
begin
  FAuthored.free;
  FAuthored := value;
end;

Function TFhirQuestionnaireResponse.GetAuthoredST : TDateAndTime;
begin
  if FAuthored = nil then
    result := nil
  else
    result := FAuthored.value;
end;

Procedure TFhirQuestionnaireResponse.SetAuthoredST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FAuthored = nil then
      FAuthored := TFhirDateTime.create;
    FAuthored.value := value
  end
  else if FAuthored <> nil then
    FAuthored.value := nil;
end;

Procedure TFhirQuestionnaireResponse.SetSource(value : TFhirReference{Resource});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirQuestionnaireResponse.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirQuestionnaireResponse.SetGroup(value : TFhirQuestionnaireResponseGroup);
begin
  FGroup.free;
  FGroup := value;
end;


{ TFhirQuestionnaireResponseListEnumerator }

Constructor TFhirQuestionnaireResponseListEnumerator.Create(list : TFhirQuestionnaireResponseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirQuestionnaireResponseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirQuestionnaireResponseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirQuestionnaireResponseListEnumerator.GetCurrent : TFhirQuestionnaireResponse;
begin
  Result := FList[FIndex];
end;


{ TFhirQuestionnaireResponseList }
procedure TFhirQuestionnaireResponseList.AddItem(value: TFhirQuestionnaireResponse);
begin
  assert(value.ClassName = 'TFhirQuestionnaireResponse', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirQuestionnaireResponse');
  add(value);
end;


function TFhirQuestionnaireResponseList.Append: TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseList.ClearItems;
begin
  Clear;
end;

function TFhirQuestionnaireResponseList.GetEnumerator : TFhirQuestionnaireResponseListEnumerator;
begin
  result := TFhirQuestionnaireResponseListEnumerator.Create(self.link);
end;

function TFhirQuestionnaireResponseList.Clone: TFhirQuestionnaireResponseList;
begin
  result := TFhirQuestionnaireResponseList(inherited Clone);
end;

function TFhirQuestionnaireResponseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirQuestionnaireResponseList.GetItemN(index: Integer): TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseList.IndexOf(value: TFhirQuestionnaireResponse): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirQuestionnaireResponseList.Insert(index: Integer): TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirQuestionnaireResponseList.InsertItem(index: Integer; value: TFhirQuestionnaireResponse);
begin
  assert(value is TFhirQuestionnaireResponse);
  Inherited Insert(index, value);
end;

function TFhirQuestionnaireResponseList.Item(index: Integer): TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse(ObjectByIndex[index]);
end;

function TFhirQuestionnaireResponseList.Link: TFhirQuestionnaireResponseList;
begin
  result := TFhirQuestionnaireResponseList(inherited Link);
end;

procedure TFhirQuestionnaireResponseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirQuestionnaireResponseList.SetItemByIndex(index: Integer; value: TFhirQuestionnaireResponse);
begin
  assert(value is TFhirQuestionnaireResponse);
  FhirQuestionnaireResponses[index] := value;
end;

procedure TFhirQuestionnaireResponseList.SetItemN(index: Integer; value: TFhirQuestionnaireResponse);
begin
  assert(value is TFhirQuestionnaireResponse);
  ObjectByIndex[index] := value;
end;

{ TFhirReferralRequest }

constructor TFhirReferralRequest.Create;
begin
  inherited;
end;

destructor TFhirReferralRequest.Destroy;
begin
  FStatus.free;
  FIdentifierList.Free;
  FDate.free;
  FType_.free;
  FSpecialty.free;
  FPriority.free;
  FPatient.free;
  FRequester.free;
  FRecipientList.Free;
  FEncounter.free;
  FDateSent.free;
  FReason.free;
  FDescription.free;
  FServiceRequestedList.Free;
  FSupportingInformationList.Free;
  FFulfillmentTime.free;
  inherited;
end;

function TFhirReferralRequest.GetResourceType : TFhirResourceType;
begin
  result := frtReferralRequest;
end;

procedure TFhirReferralRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  FStatus := TFhirReferralRequest(oSource).FStatus.Link;
  if (TFhirReferralRequest(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirReferralRequest(oSource).FIdentifierList);
  end;
  dateElement := TFhirReferralRequest(oSource).dateElement.Clone;
  type_ := TFhirReferralRequest(oSource).type_.Clone;
  specialty := TFhirReferralRequest(oSource).specialty.Clone;
  priority := TFhirReferralRequest(oSource).priority.Clone;
  patient := TFhirReferralRequest(oSource).patient.Clone;
  requester := TFhirReferralRequest(oSource).requester.Clone;
  if (TFhirReferralRequest(oSource).FRecipientList = nil) then
  begin
    FRecipientList.free;
    FRecipientList := nil;
  end
  else
  begin
    FRecipientList := TFhirReferenceList{Resource}.Create;
    FRecipientList.Assign(TFhirReferralRequest(oSource).FRecipientList);
  end;
  encounter := TFhirReferralRequest(oSource).encounter.Clone;
  dateSentElement := TFhirReferralRequest(oSource).dateSentElement.Clone;
  reason := TFhirReferralRequest(oSource).reason.Clone;
  descriptionElement := TFhirReferralRequest(oSource).descriptionElement.Clone;
  if (TFhirReferralRequest(oSource).FServiceRequestedList = nil) then
  begin
    FServiceRequestedList.free;
    FServiceRequestedList := nil;
  end
  else
  begin
    FServiceRequestedList := TFhirCodeableConceptList.Create;
    FServiceRequestedList.Assign(TFhirReferralRequest(oSource).FServiceRequestedList);
  end;
  if (TFhirReferralRequest(oSource).FSupportingInformationList = nil) then
  begin
    FSupportingInformationList.free;
    FSupportingInformationList := nil;
  end
  else
  begin
    FSupportingInformationList := TFhirReferenceList{TFhirReference}.Create;
    FSupportingInformationList.Assign(TFhirReferralRequest(oSource).FSupportingInformationList);
  end;
  fulfillmentTime := TFhirReferralRequest(oSource).fulfillmentTime.Clone;
end;

procedure TFhirReferralRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'specialty') Then
     list.add(FSpecialty.Link);
  if (child_name = 'priority') Then
     list.add(FPriority.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'requester') Then
     list.add(FRequester.Link);
  if (child_name = 'recipient') Then
    list.addAll(FRecipientList);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'dateSent') Then
     list.add(FDateSent.Link);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'serviceRequested') Then
    list.addAll(FServiceRequestedList);
  if (child_name = 'supportingInformation') Then
    list.addAll(FSupportingInformationList);
  if (child_name = 'fulfillmentTime') Then
     list.add(FFulfillmentTime.Link);
end;

procedure TFhirReferralRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specialty', 'CodeableConcept', FSpecialty.Link));{2}
  oList.add(TFHIRProperty.create(self, 'priority', 'CodeableConcept', FPriority.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requester', 'Reference(Practitioner|Organization|Patient)', FRequester.Link));{2}
  oList.add(TFHIRProperty.create(self, 'recipient', 'Reference(Practitioner|Organization)', FRecipientList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dateSent', 'dateTime', FDateSent.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason', 'CodeableConcept', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'serviceRequested', 'CodeableConcept', FServiceRequestedList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'supportingInformation', 'Reference(Any)', FSupportingInformationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'fulfillmentTime', 'Period', FFulfillmentTime.Link));{2}
end;

procedure TFhirReferralRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'specialty') then Specialty := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'priority') then Priority := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'requester') then Requester := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'recipient') then RecipientList.add(propValue as TFhirReference{Resource}){2}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'dateSent') then DateSentElement := propValue as TFhirDateTime{5a}
  else if (propName = 'reason') then Reason := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'serviceRequested') then ServiceRequestedList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'supportingInformation') then SupportingInformationList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'fulfillmentTime') then FulfillmentTime := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirReferralRequest.FhirType : string;
begin
  result := 'ReferralRequest';
end;

function TFhirReferralRequest.Link : TFhirReferralRequest;
begin
  result := TFhirReferralRequest(inherited Link);
end;

function TFhirReferralRequest.Clone : TFhirReferralRequest;
begin
  result := TFhirReferralRequest(inherited Clone);
end;

{ TFhirReferralRequest }

Procedure TFhirReferralRequest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirReferralRequest.GetStatusST : TFhirReferralstatus;
begin
  if FStatus = nil then
    result := TFhirReferralstatus(0)
  else
    result := TFhirReferralstatus(StringArrayIndexOfSensitive(CODES_TFhirReferralstatus, FStatus.value));
end;

Procedure TFhirReferralRequest.SetStatusST(value : TFhirReferralstatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirReferralstatus[value]);
end;

Function TFhirReferralRequest.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirReferralRequest.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirReferralRequest.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirReferralRequest.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirReferralRequest.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirReferralRequest.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirReferralRequest.SetSpecialty(value : TFhirCodeableConcept);
begin
  FSpecialty.free;
  FSpecialty := value;
end;

Procedure TFhirReferralRequest.SetPriority(value : TFhirCodeableConcept);
begin
  FPriority.free;
  FPriority := value;
end;

Procedure TFhirReferralRequest.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirReferralRequest.SetRequester(value : TFhirReference{Resource});
begin
  FRequester.free;
  FRequester := value;
end;

Function TFhirReferralRequest.GetRecipientList : TFhirReferenceList{Resource};
begin
  if FRecipientList = nil then
    FRecipientList := TFhirReferenceList{Resource}.Create;
  result := FRecipientList;
end;

Function TFhirReferralRequest.GetHasRecipientList : boolean;
begin
  result := (FRecipientList <> nil) and (FRecipientList.count > 0);
end;

Procedure TFhirReferralRequest.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirReferralRequest.SetDateSent(value : TFhirDateTime);
begin
  FDateSent.free;
  FDateSent := value;
end;

Function TFhirReferralRequest.GetDateSentST : TDateAndTime;
begin
  if FDateSent = nil then
    result := nil
  else
    result := FDateSent.value;
end;

Procedure TFhirReferralRequest.SetDateSentST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateSent = nil then
      FDateSent := TFhirDateTime.create;
    FDateSent.value := value
  end
  else if FDateSent <> nil then
    FDateSent.value := nil;
end;

Procedure TFhirReferralRequest.SetReason(value : TFhirCodeableConcept);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirReferralRequest.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirReferralRequest.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirReferralRequest.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirReferralRequest.GetServiceRequestedList : TFhirCodeableConceptList;
begin
  if FServiceRequestedList = nil then
    FServiceRequestedList := TFhirCodeableConceptList.Create;
  result := FServiceRequestedList;
end;

Function TFhirReferralRequest.GetHasServiceRequestedList : boolean;
begin
  result := (FServiceRequestedList <> nil) and (FServiceRequestedList.count > 0);
end;

Function TFhirReferralRequest.GetSupportingInformationList : TFhirReferenceList{TFhirReference};
begin
  if FSupportingInformationList = nil then
    FSupportingInformationList := TFhirReferenceList{TFhirReference}.Create;
  result := FSupportingInformationList;
end;

Function TFhirReferralRequest.GetHasSupportingInformationList : boolean;
begin
  result := (FSupportingInformationList <> nil) and (FSupportingInformationList.count > 0);
end;

Procedure TFhirReferralRequest.SetFulfillmentTime(value : TFhirPeriod);
begin
  FFulfillmentTime.free;
  FFulfillmentTime := value;
end;


{ TFhirReferralRequestListEnumerator }

Constructor TFhirReferralRequestListEnumerator.Create(list : TFhirReferralRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirReferralRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirReferralRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirReferralRequestListEnumerator.GetCurrent : TFhirReferralRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirReferralRequestList }
procedure TFhirReferralRequestList.AddItem(value: TFhirReferralRequest);
begin
  assert(value.ClassName = 'TFhirReferralRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirReferralRequest');
  add(value);
end;


function TFhirReferralRequestList.Append: TFhirReferralRequest;
begin
  result := TFhirReferralRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirReferralRequestList.ClearItems;
begin
  Clear;
end;

function TFhirReferralRequestList.GetEnumerator : TFhirReferralRequestListEnumerator;
begin
  result := TFhirReferralRequestListEnumerator.Create(self.link);
end;

function TFhirReferralRequestList.Clone: TFhirReferralRequestList;
begin
  result := TFhirReferralRequestList(inherited Clone);
end;

function TFhirReferralRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirReferralRequestList.GetItemN(index: Integer): TFhirReferralRequest;
begin
  result := TFhirReferralRequest(ObjectByIndex[index]);
end;

function TFhirReferralRequestList.IndexOf(value: TFhirReferralRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirReferralRequestList.Insert(index: Integer): TFhirReferralRequest;
begin
  result := TFhirReferralRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirReferralRequestList.InsertItem(index: Integer; value: TFhirReferralRequest);
begin
  assert(value is TFhirReferralRequest);
  Inherited Insert(index, value);
end;

function TFhirReferralRequestList.Item(index: Integer): TFhirReferralRequest;
begin
  result := TFhirReferralRequest(ObjectByIndex[index]);
end;

function TFhirReferralRequestList.Link: TFhirReferralRequestList;
begin
  result := TFhirReferralRequestList(inherited Link);
end;

procedure TFhirReferralRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirReferralRequestList.SetItemByIndex(index: Integer; value: TFhirReferralRequest);
begin
  assert(value is TFhirReferralRequest);
  FhirReferralRequests[index] := value;
end;

procedure TFhirReferralRequestList.SetItemN(index: Integer; value: TFhirReferralRequest);
begin
  assert(value is TFhirReferralRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirRelatedPerson }

constructor TFhirRelatedPerson.Create;
begin
  inherited;
end;

destructor TFhirRelatedPerson.Destroy;
begin
  FIdentifierList.Free;
  FPatient.free;
  FRelationship.free;
  FName.free;
  FTelecomList.Free;
  FGender.free;
  FBirthDate.free;
  FAddressList.Free;
  FPhotoList.Free;
  FPeriod.free;
  inherited;
end;

function TFhirRelatedPerson.GetResourceType : TFhirResourceType;
begin
  result := frtRelatedPerson;
end;

procedure TFhirRelatedPerson.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirRelatedPerson(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirRelatedPerson(oSource).FIdentifierList);
  end;
  patient := TFhirRelatedPerson(oSource).patient.Clone;
  relationship := TFhirRelatedPerson(oSource).relationship.Clone;
  name := TFhirRelatedPerson(oSource).name.Clone;
  if (TFhirRelatedPerson(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirRelatedPerson(oSource).FTelecomList);
  end;
  FGender := TFhirRelatedPerson(oSource).FGender.Link;
  birthDateElement := TFhirRelatedPerson(oSource).birthDateElement.Clone;
  if (TFhirRelatedPerson(oSource).FAddressList = nil) then
  begin
    FAddressList.free;
    FAddressList := nil;
  end
  else
  begin
    FAddressList := TFhirAddressList.Create;
    FAddressList.Assign(TFhirRelatedPerson(oSource).FAddressList);
  end;
  if (TFhirRelatedPerson(oSource).FPhotoList = nil) then
  begin
    FPhotoList.free;
    FPhotoList := nil;
  end
  else
  begin
    FPhotoList := TFhirAttachmentList.Create;
    FPhotoList.Assign(TFhirRelatedPerson(oSource).FPhotoList);
  end;
  period := TFhirRelatedPerson(oSource).period.Clone;
end;

procedure TFhirRelatedPerson.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'relationship') Then
     list.add(FRelationship.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
  if (child_name = 'gender') Then
     list.add(FGender.Link);
  if (child_name = 'birthDate') Then
     list.add(FBirthDate.Link);
  if (child_name = 'address') Then
    list.addAll(FAddressList);
  if (child_name = 'photo') Then
    list.addAll(FPhotoList);
  if (child_name = 'period') Then
     list.add(FPeriod.Link);
end;

procedure TFhirRelatedPerson.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relationship', 'CodeableConcept', FRelationship.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'HumanName', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'gender', 'code', FGender.Link));{1}
  oList.add(TFHIRProperty.create(self, 'birthDate', 'date', FBirthDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'address', 'Address', FAddressList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'photo', 'Attachment', FPhotoList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'period', 'Period', FPeriod.Link));{2}
end;

procedure TFhirRelatedPerson.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'relationship') then Relationship := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'name') then Name := propValue as TFhirHumanName{4b}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'gender') then GenderElement := propValue as TFHIREnum
  else if (propName = 'birthDate') then BirthDateElement := propValue as TFhirDate{5a}
  else if (propName = 'address') then AddressList.add(propValue as TFhirAddress){2}
  else if (propName = 'photo') then PhotoList.add(propValue as TFhirAttachment){2}
  else if (propName = 'period') then Period := propValue as TFhirPeriod{4b}
  else inherited;
end;

function TFhirRelatedPerson.FhirType : string;
begin
  result := 'RelatedPerson';
end;

function TFhirRelatedPerson.Link : TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson(inherited Link);
end;

function TFhirRelatedPerson.Clone : TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson(inherited Clone);
end;

{ TFhirRelatedPerson }

Function TFhirRelatedPerson.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirRelatedPerson.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirRelatedPerson.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirRelatedPerson.SetRelationship(value : TFhirCodeableConcept);
begin
  FRelationship.free;
  FRelationship := value;
end;

Procedure TFhirRelatedPerson.SetName(value : TFhirHumanName);
begin
  FName.free;
  FName := value;
end;

Function TFhirRelatedPerson.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirRelatedPerson.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;

Procedure TFhirRelatedPerson.SetGender(value : TFhirEnum);
begin
  FGender.free;
  FGender := value;
end;

Function TFhirRelatedPerson.GetGenderST : TFhirAdministrativeGender;
begin
  if FGender = nil then
    result := TFhirAdministrativeGender(0)
  else
    result := TFhirAdministrativeGender(StringArrayIndexOfSensitive(CODES_TFhirAdministrativeGender, FGender.value));
end;

Procedure TFhirRelatedPerson.SetGenderST(value : TFhirAdministrativeGender);
begin
  if ord(value) = 0 then
    GenderElement := nil
  else
    GenderElement := TFhirEnum.create(CODES_TFhirAdministrativeGender[value]);
end;

Procedure TFhirRelatedPerson.SetBirthDate(value : TFhirDate);
begin
  FBirthDate.free;
  FBirthDate := value;
end;

Function TFhirRelatedPerson.GetBirthDateST : TDateAndTime;
begin
  if FBirthDate = nil then
    result := nil
  else
    result := FBirthDate.value;
end;

Procedure TFhirRelatedPerson.SetBirthDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FBirthDate = nil then
      FBirthDate := TFhirDate.create;
    FBirthDate.value := value
  end
  else if FBirthDate <> nil then
    FBirthDate.value := nil;
end;

Function TFhirRelatedPerson.GetAddressList : TFhirAddressList;
begin
  if FAddressList = nil then
    FAddressList := TFhirAddressList.Create;
  result := FAddressList;
end;

Function TFhirRelatedPerson.GetHasAddressList : boolean;
begin
  result := (FAddressList <> nil) and (FAddressList.count > 0);
end;

Function TFhirRelatedPerson.GetPhotoList : TFhirAttachmentList;
begin
  if FPhotoList = nil then
    FPhotoList := TFhirAttachmentList.Create;
  result := FPhotoList;
end;

Function TFhirRelatedPerson.GetHasPhotoList : boolean;
begin
  result := (FPhotoList <> nil) and (FPhotoList.count > 0);
end;

Procedure TFhirRelatedPerson.SetPeriod(value : TFhirPeriod);
begin
  FPeriod.free;
  FPeriod := value;
end;


{ TFhirRelatedPersonListEnumerator }

Constructor TFhirRelatedPersonListEnumerator.Create(list : TFhirRelatedPersonList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirRelatedPersonListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirRelatedPersonListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirRelatedPersonListEnumerator.GetCurrent : TFhirRelatedPerson;
begin
  Result := FList[FIndex];
end;


{ TFhirRelatedPersonList }
procedure TFhirRelatedPersonList.AddItem(value: TFhirRelatedPerson);
begin
  assert(value.ClassName = 'TFhirRelatedPerson', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirRelatedPerson');
  add(value);
end;


function TFhirRelatedPersonList.Append: TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRelatedPersonList.ClearItems;
begin
  Clear;
end;

function TFhirRelatedPersonList.GetEnumerator : TFhirRelatedPersonListEnumerator;
begin
  result := TFhirRelatedPersonListEnumerator.Create(self.link);
end;

function TFhirRelatedPersonList.Clone: TFhirRelatedPersonList;
begin
  result := TFhirRelatedPersonList(inherited Clone);
end;

function TFhirRelatedPersonList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirRelatedPersonList.GetItemN(index: Integer): TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson(ObjectByIndex[index]);
end;

function TFhirRelatedPersonList.IndexOf(value: TFhirRelatedPerson): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirRelatedPersonList.Insert(index: Integer): TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRelatedPersonList.InsertItem(index: Integer; value: TFhirRelatedPerson);
begin
  assert(value is TFhirRelatedPerson);
  Inherited Insert(index, value);
end;

function TFhirRelatedPersonList.Item(index: Integer): TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson(ObjectByIndex[index]);
end;

function TFhirRelatedPersonList.Link: TFhirRelatedPersonList;
begin
  result := TFhirRelatedPersonList(inherited Link);
end;

procedure TFhirRelatedPersonList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirRelatedPersonList.SetItemByIndex(index: Integer; value: TFhirRelatedPerson);
begin
  assert(value is TFhirRelatedPerson);
  FhirRelatedPeople[index] := value;
end;

procedure TFhirRelatedPersonList.SetItemN(index: Integer; value: TFhirRelatedPerson);
begin
  assert(value is TFhirRelatedPerson);
  ObjectByIndex[index] := value;
end;

{ TFhirRiskAssessmentPrediction }

constructor TFhirRiskAssessmentPrediction.Create;
begin
  inherited;
end;

destructor TFhirRiskAssessmentPrediction.Destroy;
begin
  FOutcome.free;
  FProbability.free;
  FRelativeRisk.free;
  FWhen.free;
  FRationale.free;
  inherited;
end;

procedure TFhirRiskAssessmentPrediction.Assign(oSource : TAdvObject);
begin
  inherited;
  outcome := TFhirRiskAssessmentPrediction(oSource).outcome.Clone;
  probability := TFhirRiskAssessmentPrediction(oSource).probability.Clone;
  relativeRiskElement := TFhirRiskAssessmentPrediction(oSource).relativeRiskElement.Clone;
  when := TFhirRiskAssessmentPrediction(oSource).when.Clone;
  rationaleElement := TFhirRiskAssessmentPrediction(oSource).rationaleElement.Clone;
end;

procedure TFhirRiskAssessmentPrediction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'outcome') Then
     list.add(FOutcome.Link);
  if (child_name = 'probability[x]') Then
     list.add(FProbability.Link);
  if (child_name = 'relativeRisk') Then
     list.add(FRelativeRisk.Link);
  if (child_name = 'when[x]') Then
     list.add(FWhen.Link);
  if (child_name = 'rationale') Then
     list.add(FRationale.Link);
end;

procedure TFhirRiskAssessmentPrediction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'outcome', 'CodeableConcept', FOutcome.Link));{2}
  oList.add(TFHIRProperty.create(self, 'probability[x]', 'decimal|Range|CodeableConcept', FProbability.Link));{2}
  oList.add(TFHIRProperty.create(self, 'relativeRisk', 'decimal', FRelativeRisk.Link));{2}
  oList.add(TFHIRProperty.create(self, 'when[x]', 'Period|Range', FWhen.Link));{2}
  oList.add(TFHIRProperty.create(self, 'rationale', 'string', FRationale.Link));{2}
end;

procedure TFhirRiskAssessmentPrediction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'outcome') then Outcome := propValue as TFhirCodeableConcept{4b}
  else if (propName.startsWith('probability')) then Probability := propValue as TFhirType{4}
  else if (propName = 'relativeRisk') then RelativeRiskElement := propValue as TFhirDecimal{5a}
  else if (propName.startsWith('when')) then When := propValue as TFhirType{4}
  else if (propName = 'rationale') then RationaleElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirRiskAssessmentPrediction.FhirType : string;
begin
  result := 'prediction';
end;

function TFhirRiskAssessmentPrediction.Link : TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction(inherited Link);
end;

function TFhirRiskAssessmentPrediction.Clone : TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction(inherited Clone);
end;

function TFhirRiskAssessmentPrediction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirRiskAssessmentPrediction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirRiskAssessmentPrediction)) then
    result := false
  else
  begin
    o := TFhirRiskAssessmentPrediction(other);
    result := compareDeep(outcomeElement, o.outcomeElement, true) and compareDeep(probabilityElement, o.probabilityElement, true) and 
      compareDeep(relativeRiskElement, o.relativeRiskElement, true) and compareDeep(whenElement, o.whenElement, true) and 
      compareDeep(rationaleElement, o.rationaleElement, true);
  end;
end;

function TFhirRiskAssessmentPrediction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirRiskAssessmentPrediction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirRiskAssessmentPrediction)) then
    result := false
  else
  begin
    o := TFhirRiskAssessmentPrediction(other);
    result := compareValues(relativeRiskElement, o.relativeRiskElement, true) and 
      compareValues(rationaleElement, o.rationaleElement, true);
  end;
end;

{ TFhirRiskAssessmentPrediction }

Procedure TFhirRiskAssessmentPrediction.SetOutcome(value : TFhirCodeableConcept);
begin
  FOutcome.free;
  FOutcome := value;
end;

Procedure TFhirRiskAssessmentPrediction.SetProbability(value : TFhirType);
begin
  FProbability.free;
  FProbability := value;
end;

Procedure TFhirRiskAssessmentPrediction.SetRelativeRisk(value : TFhirDecimal);
begin
  FRelativeRisk.free;
  FRelativeRisk := value;
end;

Function TFhirRiskAssessmentPrediction.GetRelativeRiskST : String;
begin
  if FRelativeRisk = nil then
    result := ''
  else
    result := FRelativeRisk.value;
end;

Procedure TFhirRiskAssessmentPrediction.SetRelativeRiskST(value : String);
begin
  if value <> '' then
  begin
    if FRelativeRisk = nil then
      FRelativeRisk := TFhirDecimal.create;
    FRelativeRisk.value := value
  end
  else if FRelativeRisk <> nil then
    FRelativeRisk.value := '';
end;

Procedure TFhirRiskAssessmentPrediction.SetWhen(value : TFhirType);
begin
  FWhen.free;
  FWhen := value;
end;

Procedure TFhirRiskAssessmentPrediction.SetRationale(value : TFhirString);
begin
  FRationale.free;
  FRationale := value;
end;

Function TFhirRiskAssessmentPrediction.GetRationaleST : String;
begin
  if FRationale = nil then
    result := ''
  else
    result := FRationale.value;
end;

Procedure TFhirRiskAssessmentPrediction.SetRationaleST(value : String);
begin
  if value <> '' then
  begin
    if FRationale = nil then
      FRationale := TFhirString.create;
    FRationale.value := value
  end
  else if FRationale <> nil then
    FRationale.value := '';
end;


{ TFhirRiskAssessmentPredictionListEnumerator }

Constructor TFhirRiskAssessmentPredictionListEnumerator.Create(list : TFhirRiskAssessmentPredictionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirRiskAssessmentPredictionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirRiskAssessmentPredictionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirRiskAssessmentPredictionListEnumerator.GetCurrent : TFhirRiskAssessmentPrediction;
begin
  Result := FList[FIndex];
end;


{ TFhirRiskAssessmentPredictionList }
procedure TFhirRiskAssessmentPredictionList.AddItem(value: TFhirRiskAssessmentPrediction);
begin
  assert(value.ClassName = 'TFhirRiskAssessmentPrediction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirRiskAssessmentPrediction');
  add(value);
end;


function TFhirRiskAssessmentPredictionList.Append: TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRiskAssessmentPredictionList.ClearItems;
begin
  Clear;
end;

function TFhirRiskAssessmentPredictionList.GetEnumerator : TFhirRiskAssessmentPredictionListEnumerator;
begin
  result := TFhirRiskAssessmentPredictionListEnumerator.Create(self.link);
end;

function TFhirRiskAssessmentPredictionList.Clone: TFhirRiskAssessmentPredictionList;
begin
  result := TFhirRiskAssessmentPredictionList(inherited Clone);
end;

function TFhirRiskAssessmentPredictionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirRiskAssessmentPredictionList.GetItemN(index: Integer): TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction(ObjectByIndex[index]);
end;

function TFhirRiskAssessmentPredictionList.IndexOf(value: TFhirRiskAssessmentPrediction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirRiskAssessmentPredictionList.Insert(index: Integer): TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRiskAssessmentPredictionList.InsertItem(index: Integer; value: TFhirRiskAssessmentPrediction);
begin
  assert(value is TFhirRiskAssessmentPrediction);
  Inherited Insert(index, value);
end;

function TFhirRiskAssessmentPredictionList.Item(index: Integer): TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction(ObjectByIndex[index]);
end;

function TFhirRiskAssessmentPredictionList.Link: TFhirRiskAssessmentPredictionList;
begin
  result := TFhirRiskAssessmentPredictionList(inherited Link);
end;

procedure TFhirRiskAssessmentPredictionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirRiskAssessmentPredictionList.SetItemByIndex(index: Integer; value: TFhirRiskAssessmentPrediction);
begin
  assert(value is TFhirRiskAssessmentPrediction);
  FhirRiskAssessmentPredictions[index] := value;
end;

procedure TFhirRiskAssessmentPredictionList.SetItemN(index: Integer; value: TFhirRiskAssessmentPrediction);
begin
  assert(value is TFhirRiskAssessmentPrediction);
  ObjectByIndex[index] := value;
end;

{ TFhirRiskAssessment }

constructor TFhirRiskAssessment.Create;
begin
  inherited;
end;

destructor TFhirRiskAssessment.Destroy;
begin
  FSubject.free;
  FDate.free;
  FCondition.free;
  FEncounter.free;
  FPerformer.free;
  FIdentifier.free;
  FMethod.free;
  FBasisList.Free;
  FPredictionList.Free;
  FMitigation.free;
  inherited;
end;

function TFhirRiskAssessment.GetResourceType : TFhirResourceType;
begin
  result := frtRiskAssessment;
end;

procedure TFhirRiskAssessment.Assign(oSource : TAdvObject);
begin
  inherited;
  subject := TFhirRiskAssessment(oSource).subject.Clone;
  dateElement := TFhirRiskAssessment(oSource).dateElement.Clone;
  condition := TFhirRiskAssessment(oSource).condition.Clone;
  encounter := TFhirRiskAssessment(oSource).encounter.Clone;
  performer := TFhirRiskAssessment(oSource).performer.Clone;
  identifier := TFhirRiskAssessment(oSource).identifier.Clone;
  method := TFhirRiskAssessment(oSource).method.Clone;
  if (TFhirRiskAssessment(oSource).FBasisList = nil) then
  begin
    FBasisList.free;
    FBasisList := nil;
  end
  else
  begin
    FBasisList := TFhirReferenceList{TFhirReference}.Create;
    FBasisList.Assign(TFhirRiskAssessment(oSource).FBasisList);
  end;
  if (TFhirRiskAssessment(oSource).FPredictionList = nil) then
  begin
    FPredictionList.free;
    FPredictionList := nil;
  end
  else
  begin
    FPredictionList := TFhirRiskAssessmentPredictionList.Create;
    FPredictionList.Assign(TFhirRiskAssessment(oSource).FPredictionList);
  end;
  mitigationElement := TFhirRiskAssessment(oSource).mitigationElement.Clone;
end;

procedure TFhirRiskAssessment.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'condition') Then
     list.add(FCondition.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'performer') Then
     list.add(FPerformer.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'basis') Then
    list.addAll(FBasisList);
  if (child_name = 'prediction') Then
    list.addAll(FPredictionList);
  if (child_name = 'mitigation') Then
     list.add(FMitigation.Link);
end;

procedure TFhirRiskAssessment.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'condition', 'Reference(Condition)', FCondition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'performer', 'Reference(Practitioner|Device)', FPerformer.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'basis', 'Reference(Any)', FBasisList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'prediction', '', FPredictionList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'mitigation', 'string', FMitigation.Link));{2}
end;

procedure TFhirRiskAssessment.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'condition') then Condition := propValue as TFhirReference{TFhirCondition}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName = 'performer') then Performer := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'basis') then BasisList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'prediction') then PredictionList.add(propValue as TFhirRiskAssessmentPrediction){2}
  else if (propName = 'mitigation') then MitigationElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirRiskAssessment.FhirType : string;
begin
  result := 'RiskAssessment';
end;

function TFhirRiskAssessment.Link : TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment(inherited Link);
end;

function TFhirRiskAssessment.Clone : TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment(inherited Clone);
end;

{ TFhirRiskAssessment }

Procedure TFhirRiskAssessment.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirRiskAssessment.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirRiskAssessment.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirRiskAssessment.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirRiskAssessment.SetCondition(value : TFhirReference{TFhirCondition});
begin
  FCondition.free;
  FCondition := value;
end;

Procedure TFhirRiskAssessment.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirRiskAssessment.SetPerformer(value : TFhirReference{Resource});
begin
  FPerformer.free;
  FPerformer := value;
end;

Procedure TFhirRiskAssessment.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirRiskAssessment.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Function TFhirRiskAssessment.GetBasisList : TFhirReferenceList{TFhirReference};
begin
  if FBasisList = nil then
    FBasisList := TFhirReferenceList{TFhirReference}.Create;
  result := FBasisList;
end;

Function TFhirRiskAssessment.GetHasBasisList : boolean;
begin
  result := (FBasisList <> nil) and (FBasisList.count > 0);
end;

Function TFhirRiskAssessment.GetPredictionList : TFhirRiskAssessmentPredictionList;
begin
  if FPredictionList = nil then
    FPredictionList := TFhirRiskAssessmentPredictionList.Create;
  result := FPredictionList;
end;

Function TFhirRiskAssessment.GetHasPredictionList : boolean;
begin
  result := (FPredictionList <> nil) and (FPredictionList.count > 0);
end;

Procedure TFhirRiskAssessment.SetMitigation(value : TFhirString);
begin
  FMitigation.free;
  FMitigation := value;
end;

Function TFhirRiskAssessment.GetMitigationST : String;
begin
  if FMitigation = nil then
    result := ''
  else
    result := FMitigation.value;
end;

Procedure TFhirRiskAssessment.SetMitigationST(value : String);
begin
  if value <> '' then
  begin
    if FMitigation = nil then
      FMitigation := TFhirString.create;
    FMitigation.value := value
  end
  else if FMitigation <> nil then
    FMitigation.value := '';
end;


{ TFhirRiskAssessmentListEnumerator }

Constructor TFhirRiskAssessmentListEnumerator.Create(list : TFhirRiskAssessmentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirRiskAssessmentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirRiskAssessmentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirRiskAssessmentListEnumerator.GetCurrent : TFhirRiskAssessment;
begin
  Result := FList[FIndex];
end;


{ TFhirRiskAssessmentList }
procedure TFhirRiskAssessmentList.AddItem(value: TFhirRiskAssessment);
begin
  assert(value.ClassName = 'TFhirRiskAssessment', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirRiskAssessment');
  add(value);
end;


function TFhirRiskAssessmentList.Append: TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRiskAssessmentList.ClearItems;
begin
  Clear;
end;

function TFhirRiskAssessmentList.GetEnumerator : TFhirRiskAssessmentListEnumerator;
begin
  result := TFhirRiskAssessmentListEnumerator.Create(self.link);
end;

function TFhirRiskAssessmentList.Clone: TFhirRiskAssessmentList;
begin
  result := TFhirRiskAssessmentList(inherited Clone);
end;

function TFhirRiskAssessmentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirRiskAssessmentList.GetItemN(index: Integer): TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment(ObjectByIndex[index]);
end;

function TFhirRiskAssessmentList.IndexOf(value: TFhirRiskAssessment): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirRiskAssessmentList.Insert(index: Integer): TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirRiskAssessmentList.InsertItem(index: Integer; value: TFhirRiskAssessment);
begin
  assert(value is TFhirRiskAssessment);
  Inherited Insert(index, value);
end;

function TFhirRiskAssessmentList.Item(index: Integer): TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment(ObjectByIndex[index]);
end;

function TFhirRiskAssessmentList.Link: TFhirRiskAssessmentList;
begin
  result := TFhirRiskAssessmentList(inherited Link);
end;

procedure TFhirRiskAssessmentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirRiskAssessmentList.SetItemByIndex(index: Integer; value: TFhirRiskAssessment);
begin
  assert(value is TFhirRiskAssessment);
  FhirRiskAssessments[index] := value;
end;

procedure TFhirRiskAssessmentList.SetItemN(index: Integer; value: TFhirRiskAssessment);
begin
  assert(value is TFhirRiskAssessment);
  ObjectByIndex[index] := value;
end;

{ TFhirSchedule }

constructor TFhirSchedule.Create;
begin
  inherited;
end;

destructor TFhirSchedule.Destroy;
begin
  FIdentifierList.Free;
  FType_List.Free;
  FActor.free;
  FPlanningHorizon.free;
  FComment.free;
  inherited;
end;

function TFhirSchedule.GetResourceType : TFhirResourceType;
begin
  result := frtSchedule;
end;

procedure TFhirSchedule.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirSchedule(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirSchedule(oSource).FIdentifierList);
  end;
  if (TFhirSchedule(oSource).FType_List = nil) then
  begin
    FType_List.free;
    FType_List := nil;
  end
  else
  begin
    FType_List := TFhirCodeableConceptList.Create;
    FType_List.Assign(TFhirSchedule(oSource).FType_List);
  end;
  actor := TFhirSchedule(oSource).actor.Clone;
  planningHorizon := TFhirSchedule(oSource).planningHorizon.Clone;
  commentElement := TFhirSchedule(oSource).commentElement.Clone;
end;

procedure TFhirSchedule.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'type') Then
    list.addAll(FType_List);
  if (child_name = 'actor') Then
     list.add(FActor.Link);
  if (child_name = 'planningHorizon') Then
     list.add(FPlanningHorizon.Link);
  if (child_name = 'comment') Then
     list.add(FComment.Link);
end;

procedure TFhirSchedule.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'actor', 'Reference(Patient|Practitioner|RelatedPerson|Device|HealthcareService|Location)', FActor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'planningHorizon', 'Period', FPlanningHorizon.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
end;

procedure TFhirSchedule.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'type') then Type_List.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'actor') then Actor := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'planningHorizon') then PlanningHorizon := propValue as TFhirPeriod{4b}
  else if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirSchedule.FhirType : string;
begin
  result := 'Schedule';
end;

function TFhirSchedule.Link : TFhirSchedule;
begin
  result := TFhirSchedule(inherited Link);
end;

function TFhirSchedule.Clone : TFhirSchedule;
begin
  result := TFhirSchedule(inherited Clone);
end;

{ TFhirSchedule }

Function TFhirSchedule.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirSchedule.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirSchedule.GetType_List : TFhirCodeableConceptList;
begin
  if FType_List = nil then
    FType_List := TFhirCodeableConceptList.Create;
  result := FType_List;
end;

Function TFhirSchedule.GetHasType_List : boolean;
begin
  result := (FType_List <> nil) and (FType_List.count > 0);
end;

Procedure TFhirSchedule.SetActor(value : TFhirReference{Resource});
begin
  FActor.free;
  FActor := value;
end;

Procedure TFhirSchedule.SetPlanningHorizon(value : TFhirPeriod);
begin
  FPlanningHorizon.free;
  FPlanningHorizon := value;
end;

Procedure TFhirSchedule.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirSchedule.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirSchedule.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;


{ TFhirScheduleListEnumerator }

Constructor TFhirScheduleListEnumerator.Create(list : TFhirScheduleList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirScheduleListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirScheduleListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirScheduleListEnumerator.GetCurrent : TFhirSchedule;
begin
  Result := FList[FIndex];
end;


{ TFhirScheduleList }
procedure TFhirScheduleList.AddItem(value: TFhirSchedule);
begin
  assert(value.ClassName = 'TFhirSchedule', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSchedule');
  add(value);
end;


function TFhirScheduleList.Append: TFhirSchedule;
begin
  result := TFhirSchedule.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirScheduleList.ClearItems;
begin
  Clear;
end;

function TFhirScheduleList.GetEnumerator : TFhirScheduleListEnumerator;
begin
  result := TFhirScheduleListEnumerator.Create(self.link);
end;

function TFhirScheduleList.Clone: TFhirScheduleList;
begin
  result := TFhirScheduleList(inherited Clone);
end;

function TFhirScheduleList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirScheduleList.GetItemN(index: Integer): TFhirSchedule;
begin
  result := TFhirSchedule(ObjectByIndex[index]);
end;

function TFhirScheduleList.IndexOf(value: TFhirSchedule): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirScheduleList.Insert(index: Integer): TFhirSchedule;
begin
  result := TFhirSchedule.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirScheduleList.InsertItem(index: Integer; value: TFhirSchedule);
begin
  assert(value is TFhirSchedule);
  Inherited Insert(index, value);
end;

function TFhirScheduleList.Item(index: Integer): TFhirSchedule;
begin
  result := TFhirSchedule(ObjectByIndex[index]);
end;

function TFhirScheduleList.Link: TFhirScheduleList;
begin
  result := TFhirScheduleList(inherited Link);
end;

procedure TFhirScheduleList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirScheduleList.SetItemByIndex(index: Integer; value: TFhirSchedule);
begin
  assert(value is TFhirSchedule);
  FhirSchedules[index] := value;
end;

procedure TFhirScheduleList.SetItemN(index: Integer; value: TFhirSchedule);
begin
  assert(value is TFhirSchedule);
  ObjectByIndex[index] := value;
end;

{ TFhirSearchParameterContact }

constructor TFhirSearchParameterContact.Create;
begin
  inherited;
end;

destructor TFhirSearchParameterContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirSearchParameterContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirSearchParameterContact(oSource).nameElement.Clone;
  if (TFhirSearchParameterContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirSearchParameterContact(oSource).FTelecomList);
  end;
end;

procedure TFhirSearchParameterContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirSearchParameterContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirSearchParameterContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirSearchParameterContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirSearchParameterContact.Link : TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact(inherited Link);
end;

function TFhirSearchParameterContact.Clone : TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact(inherited Clone);
end;

function TFhirSearchParameterContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSearchParameterContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSearchParameterContact)) then
    result := false
  else
  begin
    o := TFhirSearchParameterContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirSearchParameterContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSearchParameterContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSearchParameterContact)) then
    result := false
  else
  begin
    o := TFhirSearchParameterContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirSearchParameterContact }

Procedure TFhirSearchParameterContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirSearchParameterContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirSearchParameterContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirSearchParameterContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirSearchParameterContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirSearchParameterContactListEnumerator }

Constructor TFhirSearchParameterContactListEnumerator.Create(list : TFhirSearchParameterContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSearchParameterContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSearchParameterContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSearchParameterContactListEnumerator.GetCurrent : TFhirSearchParameterContact;
begin
  Result := FList[FIndex];
end;


{ TFhirSearchParameterContactList }
procedure TFhirSearchParameterContactList.AddItem(value: TFhirSearchParameterContact);
begin
  assert(value.ClassName = 'TFhirSearchParameterContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSearchParameterContact');
  add(value);
end;


function TFhirSearchParameterContactList.Append: TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSearchParameterContactList.ClearItems;
begin
  Clear;
end;

function TFhirSearchParameterContactList.GetEnumerator : TFhirSearchParameterContactListEnumerator;
begin
  result := TFhirSearchParameterContactListEnumerator.Create(self.link);
end;

function TFhirSearchParameterContactList.Clone: TFhirSearchParameterContactList;
begin
  result := TFhirSearchParameterContactList(inherited Clone);
end;

function TFhirSearchParameterContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSearchParameterContactList.GetItemN(index: Integer): TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact(ObjectByIndex[index]);
end;

function TFhirSearchParameterContactList.IndexOf(value: TFhirSearchParameterContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSearchParameterContactList.Insert(index: Integer): TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSearchParameterContactList.InsertItem(index: Integer; value: TFhirSearchParameterContact);
begin
  assert(value is TFhirSearchParameterContact);
  Inherited Insert(index, value);
end;

function TFhirSearchParameterContactList.Item(index: Integer): TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact(ObjectByIndex[index]);
end;

function TFhirSearchParameterContactList.Link: TFhirSearchParameterContactList;
begin
  result := TFhirSearchParameterContactList(inherited Link);
end;

procedure TFhirSearchParameterContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSearchParameterContactList.SetItemByIndex(index: Integer; value: TFhirSearchParameterContact);
begin
  assert(value is TFhirSearchParameterContact);
  FhirSearchParameterContacts[index] := value;
end;

procedure TFhirSearchParameterContactList.SetItemN(index: Integer; value: TFhirSearchParameterContact);
begin
  assert(value is TFhirSearchParameterContact);
  ObjectByIndex[index] := value;
end;

{ TFhirSearchParameter }

constructor TFhirSearchParameter.Create;
begin
  inherited;
end;

destructor TFhirSearchParameter.Destroy;
begin
  FUrl.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FRequirements.free;
  FCode.free;
  FBase.free;
  FType_.free;
  FDescription.free;
  FXpath.free;
  FXpathUsage.free;
  FTargetList.Free;
  inherited;
end;

function TFhirSearchParameter.GetResourceType : TFhirResourceType;
begin
  result := frtSearchParameter;
end;

procedure TFhirSearchParameter.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirSearchParameter(oSource).urlElement.Clone;
  nameElement := TFhirSearchParameter(oSource).nameElement.Clone;
  FStatus := TFhirSearchParameter(oSource).FStatus.Link;
  experimentalElement := TFhirSearchParameter(oSource).experimentalElement.Clone;
  publisherElement := TFhirSearchParameter(oSource).publisherElement.Clone;
  if (TFhirSearchParameter(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirSearchParameterContactList.Create;
    FContactList.Assign(TFhirSearchParameter(oSource).FContactList);
  end;
  dateElement := TFhirSearchParameter(oSource).dateElement.Clone;
  requirementsElement := TFhirSearchParameter(oSource).requirementsElement.Clone;
  codeElement := TFhirSearchParameter(oSource).codeElement.Clone;
  baseElement := TFhirSearchParameter(oSource).baseElement.Clone;
  FType_ := TFhirSearchParameter(oSource).FType_.Link;
  descriptionElement := TFhirSearchParameter(oSource).descriptionElement.Clone;
  xpathElement := TFhirSearchParameter(oSource).xpathElement.Clone;
  FXpathUsage := TFhirSearchParameter(oSource).FXpathUsage.Link;
  if (TFhirSearchParameter(oSource).FTargetList = nil) then
  begin
    FTargetList.free;
    FTargetList := nil;
  end
  else
  begin
    FTargetList := TFhirCodeList.Create;
    FTargetList.Assign(TFhirSearchParameter(oSource).FTargetList);
  end;
end;

procedure TFhirSearchParameter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'base') Then
     list.add(FBase.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'xpath') Then
     list.add(FXpath.Link);
  if (child_name = 'xpathUsage') Then
     list.add(FXpathUsage.Link);
  if (child_name = 'target') Then
    list.addAll(FTargetList);
end;

procedure TFhirSearchParameter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'base', 'code', FBase.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'xpath', 'string', FXpath.Link));{2}
  oList.add(TFHIRProperty.create(self, 'xpathUsage', 'code', FXpathUsage.Link));{1}
  oList.add(TFHIRProperty.create(self, 'target', 'code', FTargetList.Link)){3};
end;

procedure TFhirSearchParameter.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirSearchParameterContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'base') then
    if propValue is TFHIRCode then
      BaseElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      BaseElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'xpath') then XpathElement := propValue as TFhirString{5a}
  else if (propName = 'xpathUsage') then XpathUsageElement := propValue as TFHIREnum
  else if (propName = 'target') then TargetList.add(propValue as TFhirCode){2}
  else inherited;
end;

function TFhirSearchParameter.FhirType : string;
begin
  result := 'SearchParameter';
end;

function TFhirSearchParameter.Link : TFhirSearchParameter;
begin
  result := TFhirSearchParameter(inherited Link);
end;

function TFhirSearchParameter.Clone : TFhirSearchParameter;
begin
  result := TFhirSearchParameter(inherited Clone);
end;

{ TFhirSearchParameter }

Procedure TFhirSearchParameter.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirSearchParameter.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirSearchParameter.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirSearchParameter.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirSearchParameter.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirSearchParameter.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirSearchParameter.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirSearchParameter.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirSearchParameter.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirSearchParameter.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirSearchParameter.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirSearchParameter.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirSearchParameter.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirSearchParameter.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirSearchParameter.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirSearchParameter.GetContactList : TFhirSearchParameterContactList;
begin
  if FContactList = nil then
    FContactList := TFhirSearchParameterContactList.Create;
  result := FContactList;
end;

Function TFhirSearchParameter.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirSearchParameter.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirSearchParameter.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirSearchParameter.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirSearchParameter.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirSearchParameter.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirSearchParameter.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirSearchParameter.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirSearchParameter.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirSearchParameter.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirSearchParameter.SetBase(value : TFhirCode);
begin
  FBase.free;
  FBase := value;
end;

Function TFhirSearchParameter.GetBaseST : String;
begin
  if FBase = nil then
    result := ''
  else
    result := FBase.value;
end;

Procedure TFhirSearchParameter.SetBaseST(value : String);
begin
  if value <> '' then
  begin
    if FBase = nil then
      FBase := TFhirCode.create;
    FBase.value := value
  end
  else if FBase <> nil then
    FBase.value := '';
end;

Procedure TFhirSearchParameter.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirSearchParameter.GetType_ST : TFhirSearchParamType;
begin
  if FType_ = nil then
    result := TFhirSearchParamType(0)
  else
    result := TFhirSearchParamType(StringArrayIndexOfSensitive(CODES_TFhirSearchParamType, FType_.value));
end;

Procedure TFhirSearchParameter.SetType_ST(value : TFhirSearchParamType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirSearchParamType[value]);
end;

Procedure TFhirSearchParameter.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirSearchParameter.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirSearchParameter.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirSearchParameter.SetXpath(value : TFhirString);
begin
  FXpath.free;
  FXpath := value;
end;

Function TFhirSearchParameter.GetXpathST : String;
begin
  if FXpath = nil then
    result := ''
  else
    result := FXpath.value;
end;

Procedure TFhirSearchParameter.SetXpathST(value : String);
begin
  if value <> '' then
  begin
    if FXpath = nil then
      FXpath := TFhirString.create;
    FXpath.value := value
  end
  else if FXpath <> nil then
    FXpath.value := '';
end;

Procedure TFhirSearchParameter.SetXpathUsage(value : TFhirEnum);
begin
  FXpathUsage.free;
  FXpathUsage := value;
end;

Function TFhirSearchParameter.GetXpathUsageST : TFhirSearchXpathUsage;
begin
  if FXpathUsage = nil then
    result := TFhirSearchXpathUsage(0)
  else
    result := TFhirSearchXpathUsage(StringArrayIndexOfSensitive(CODES_TFhirSearchXpathUsage, FXpathUsage.value));
end;

Procedure TFhirSearchParameter.SetXpathUsageST(value : TFhirSearchXpathUsage);
begin
  if ord(value) = 0 then
    XpathUsageElement := nil
  else
    XpathUsageElement := TFhirEnum.create(CODES_TFhirSearchXpathUsage[value]);
end;

Function TFhirSearchParameter.GetTargetList : TFhirCodeList;
begin
  if FTargetList = nil then
    FTargetList := TFhirCodeList.Create;
  result := FTargetList;
end;

Function TFhirSearchParameter.GetHasTargetList : boolean;
begin
  result := (FTargetList <> nil) and (FTargetList.count > 0);
end;


{ TFhirSearchParameterListEnumerator }

Constructor TFhirSearchParameterListEnumerator.Create(list : TFhirSearchParameterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSearchParameterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSearchParameterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSearchParameterListEnumerator.GetCurrent : TFhirSearchParameter;
begin
  Result := FList[FIndex];
end;


{ TFhirSearchParameterList }
procedure TFhirSearchParameterList.AddItem(value: TFhirSearchParameter);
begin
  assert(value.ClassName = 'TFhirSearchParameter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSearchParameter');
  add(value);
end;


function TFhirSearchParameterList.Append: TFhirSearchParameter;
begin
  result := TFhirSearchParameter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSearchParameterList.ClearItems;
begin
  Clear;
end;

function TFhirSearchParameterList.GetEnumerator : TFhirSearchParameterListEnumerator;
begin
  result := TFhirSearchParameterListEnumerator.Create(self.link);
end;

function TFhirSearchParameterList.Clone: TFhirSearchParameterList;
begin
  result := TFhirSearchParameterList(inherited Clone);
end;

function TFhirSearchParameterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSearchParameterList.GetItemN(index: Integer): TFhirSearchParameter;
begin
  result := TFhirSearchParameter(ObjectByIndex[index]);
end;

function TFhirSearchParameterList.IndexOf(value: TFhirSearchParameter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSearchParameterList.Insert(index: Integer): TFhirSearchParameter;
begin
  result := TFhirSearchParameter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSearchParameterList.InsertItem(index: Integer; value: TFhirSearchParameter);
begin
  assert(value is TFhirSearchParameter);
  Inherited Insert(index, value);
end;

function TFhirSearchParameterList.Item(index: Integer): TFhirSearchParameter;
begin
  result := TFhirSearchParameter(ObjectByIndex[index]);
end;

function TFhirSearchParameterList.Link: TFhirSearchParameterList;
begin
  result := TFhirSearchParameterList(inherited Link);
end;

procedure TFhirSearchParameterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSearchParameterList.SetItemByIndex(index: Integer; value: TFhirSearchParameter);
begin
  assert(value is TFhirSearchParameter);
  FhirSearchParameters[index] := value;
end;

procedure TFhirSearchParameterList.SetItemN(index: Integer; value: TFhirSearchParameter);
begin
  assert(value is TFhirSearchParameter);
  ObjectByIndex[index] := value;
end;

{ TFhirSlot }

constructor TFhirSlot.Create;
begin
  inherited;
end;

destructor TFhirSlot.Destroy;
begin
  FIdentifierList.Free;
  FType_.free;
  FSchedule.free;
  FFreeBusyType.free;
  FStart.free;
  FEnd_.free;
  FOverbooked.free;
  FComment.free;
  inherited;
end;

function TFhirSlot.GetResourceType : TFhirResourceType;
begin
  result := frtSlot;
end;

procedure TFhirSlot.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirSlot(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirSlot(oSource).FIdentifierList);
  end;
  type_ := TFhirSlot(oSource).type_.Clone;
  schedule := TFhirSlot(oSource).schedule.Clone;
  FFreeBusyType := TFhirSlot(oSource).FFreeBusyType.Link;
  startElement := TFhirSlot(oSource).startElement.Clone;
  end_Element := TFhirSlot(oSource).end_Element.Clone;
  overbookedElement := TFhirSlot(oSource).overbookedElement.Clone;
  commentElement := TFhirSlot(oSource).commentElement.Clone;
end;

procedure TFhirSlot.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'schedule') Then
     list.add(FSchedule.Link);
  if (child_name = 'freeBusyType') Then
     list.add(FFreeBusyType.Link);
  if (child_name = 'start') Then
     list.add(FStart.Link);
  if (child_name = 'end') Then
     list.add(FEnd_.Link);
  if (child_name = 'overbooked') Then
     list.add(FOverbooked.Link);
  if (child_name = 'comment') Then
     list.add(FComment.Link);
end;

procedure TFhirSlot.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'schedule', 'Reference(Schedule)', FSchedule.Link));{2}
  oList.add(TFHIRProperty.create(self, 'freeBusyType', 'code', FFreeBusyType.Link));{1}
  oList.add(TFHIRProperty.create(self, 'start', 'instant', FStart.Link));{2}
  oList.add(TFHIRProperty.create(self, 'end', 'instant', FEnd_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'overbooked', 'boolean', FOverbooked.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FComment.Link));{2}
end;

procedure TFhirSlot.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'schedule') then Schedule := propValue as TFhirReference{TFhirSchedule}{4b}
  else if (propName = 'freeBusyType') then FreeBusyTypeElement := propValue as TFHIREnum
  else if (propName = 'start') then StartElement := propValue as TFhirInstant{5a}
  else if (propName = 'end') then End_Element := propValue as TFhirInstant{5a}
  else if (propName = 'overbooked') then OverbookedElement := propValue as TFhirBoolean{5a}
  else if (propName = 'comment') then CommentElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirSlot.FhirType : string;
begin
  result := 'Slot';
end;

function TFhirSlot.Link : TFhirSlot;
begin
  result := TFhirSlot(inherited Link);
end;

function TFhirSlot.Clone : TFhirSlot;
begin
  result := TFhirSlot(inherited Clone);
end;

{ TFhirSlot }

Function TFhirSlot.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirSlot.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirSlot.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirSlot.SetSchedule(value : TFhirReference{TFhirSchedule});
begin
  FSchedule.free;
  FSchedule := value;
end;

Procedure TFhirSlot.SetFreeBusyType(value : TFhirEnum);
begin
  FFreeBusyType.free;
  FFreeBusyType := value;
end;

Function TFhirSlot.GetFreeBusyTypeST : TFhirSlotstatus;
begin
  if FFreeBusyType = nil then
    result := TFhirSlotstatus(0)
  else
    result := TFhirSlotstatus(StringArrayIndexOfSensitive(CODES_TFhirSlotstatus, FFreeBusyType.value));
end;

Procedure TFhirSlot.SetFreeBusyTypeST(value : TFhirSlotstatus);
begin
  if ord(value) = 0 then
    FreeBusyTypeElement := nil
  else
    FreeBusyTypeElement := TFhirEnum.create(CODES_TFhirSlotstatus[value]);
end;

Procedure TFhirSlot.SetStart(value : TFhirInstant);
begin
  FStart.free;
  FStart := value;
end;

Function TFhirSlot.GetStartST : TDateAndTime;
begin
  if FStart = nil then
    result := nil
  else
    result := FStart.value;
end;

Procedure TFhirSlot.SetStartST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FStart = nil then
      FStart := TFhirInstant.create;
    FStart.value := value
  end
  else if FStart <> nil then
    FStart.value := nil;
end;

Procedure TFhirSlot.SetEnd_(value : TFhirInstant);
begin
  FEnd_.free;
  FEnd_ := value;
end;

Function TFhirSlot.GetEnd_ST : TDateAndTime;
begin
  if FEnd_ = nil then
    result := nil
  else
    result := FEnd_.value;
end;

Procedure TFhirSlot.SetEnd_ST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEnd_ = nil then
      FEnd_ := TFhirInstant.create;
    FEnd_.value := value
  end
  else if FEnd_ <> nil then
    FEnd_.value := nil;
end;

Procedure TFhirSlot.SetOverbooked(value : TFhirBoolean);
begin
  FOverbooked.free;
  FOverbooked := value;
end;

Function TFhirSlot.GetOverbookedST : Boolean;
begin
  if FOverbooked = nil then
    result := false
  else
    result := FOverbooked.value;
end;

Procedure TFhirSlot.SetOverbookedST(value : Boolean);
begin
  if FOverbooked = nil then
    FOverbooked := TFhirBoolean.create;
  FOverbooked.value := value
end;

Procedure TFhirSlot.SetComment(value : TFhirString);
begin
  FComment.free;
  FComment := value;
end;

Function TFhirSlot.GetCommentST : String;
begin
  if FComment = nil then
    result := ''
  else
    result := FComment.value;
end;

Procedure TFhirSlot.SetCommentST(value : String);
begin
  if value <> '' then
  begin
    if FComment = nil then
      FComment := TFhirString.create;
    FComment.value := value
  end
  else if FComment <> nil then
    FComment.value := '';
end;


{ TFhirSlotListEnumerator }

Constructor TFhirSlotListEnumerator.Create(list : TFhirSlotList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSlotListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSlotListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSlotListEnumerator.GetCurrent : TFhirSlot;
begin
  Result := FList[FIndex];
end;


{ TFhirSlotList }
procedure TFhirSlotList.AddItem(value: TFhirSlot);
begin
  assert(value.ClassName = 'TFhirSlot', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSlot');
  add(value);
end;


function TFhirSlotList.Append: TFhirSlot;
begin
  result := TFhirSlot.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSlotList.ClearItems;
begin
  Clear;
end;

function TFhirSlotList.GetEnumerator : TFhirSlotListEnumerator;
begin
  result := TFhirSlotListEnumerator.Create(self.link);
end;

function TFhirSlotList.Clone: TFhirSlotList;
begin
  result := TFhirSlotList(inherited Clone);
end;

function TFhirSlotList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSlotList.GetItemN(index: Integer): TFhirSlot;
begin
  result := TFhirSlot(ObjectByIndex[index]);
end;

function TFhirSlotList.IndexOf(value: TFhirSlot): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSlotList.Insert(index: Integer): TFhirSlot;
begin
  result := TFhirSlot.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSlotList.InsertItem(index: Integer; value: TFhirSlot);
begin
  assert(value is TFhirSlot);
  Inherited Insert(index, value);
end;

function TFhirSlotList.Item(index: Integer): TFhirSlot;
begin
  result := TFhirSlot(ObjectByIndex[index]);
end;

function TFhirSlotList.Link: TFhirSlotList;
begin
  result := TFhirSlotList(inherited Link);
end;

procedure TFhirSlotList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSlotList.SetItemByIndex(index: Integer; value: TFhirSlot);
begin
  assert(value is TFhirSlot);
  FhirSlots[index] := value;
end;

procedure TFhirSlotList.SetItemN(index: Integer; value: TFhirSlot);
begin
  assert(value is TFhirSlot);
  ObjectByIndex[index] := value;
end;

{ TFhirSpecimenCollection }

constructor TFhirSpecimenCollection.Create;
begin
  inherited;
end;

destructor TFhirSpecimenCollection.Destroy;
begin
  FCollector.free;
  FCommentList.Free;
  FCollected.free;
  FQuantity.free;
  FMethod.free;
  FBodySite.free;
  inherited;
end;

procedure TFhirSpecimenCollection.Assign(oSource : TAdvObject);
begin
  inherited;
  collector := TFhirSpecimenCollection(oSource).collector.Clone;
  if (TFhirSpecimenCollection(oSource).FCommentList = nil) then
  begin
    FCommentList.free;
    FCommentList := nil;
  end
  else
  begin
    FCommentList := TFhirStringList.Create;
    FCommentList.Assign(TFhirSpecimenCollection(oSource).FCommentList);
  end;
  collected := TFhirSpecimenCollection(oSource).collected.Clone;
  quantity := TFhirSpecimenCollection(oSource).quantity.Clone;
  method := TFhirSpecimenCollection(oSource).method.Clone;
  bodySite := TFhirSpecimenCollection(oSource).bodySite.Clone;
end;

procedure TFhirSpecimenCollection.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'collector') Then
     list.add(FCollector.Link);
  if (child_name = 'comment') Then
    list.addAll(FCommentList);
  if (child_name = 'collected[x]') Then
     list.add(FCollected.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'method') Then
     list.add(FMethod.Link);
  if (child_name = 'bodySite') Then
     list.add(FBodySite.Link);
end;

procedure TFhirSpecimenCollection.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'collector', 'Reference(Practitioner)', FCollector.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comment', 'string', FCommentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'collected[x]', 'dateTime|Period', FCollected.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'method', 'CodeableConcept', FMethod.Link));{2}
  oList.add(TFHIRProperty.create(self, 'bodySite', 'CodeableConcept', FBodySite.Link));{2}
end;

procedure TFhirSpecimenCollection.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'collector') then Collector := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'comment') then CommentList.add(propValue as TFhirString){2}
  else if (propName.startsWith('collected')) then Collected := propValue as TFhirType{4}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'method') then Method := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'bodySite') then BodySite := propValue as TFhirCodeableConcept{4b}
  else inherited;
end;

function TFhirSpecimenCollection.FhirType : string;
begin
  result := 'collection';
end;

function TFhirSpecimenCollection.Link : TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection(inherited Link);
end;

function TFhirSpecimenCollection.Clone : TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection(inherited Clone);
end;

function TFhirSpecimenCollection.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenCollection;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSpecimenCollection)) then
    result := false
  else
  begin
    o := TFhirSpecimenCollection(other);
    result := compareDeep(collectorElement, o.collectorElement, true) and compareDeep(commentList, o.commentList, true) and 
      compareDeep(collectedElement, o.collectedElement, true) and compareDeep(quantityElement, o.quantityElement, true) and 
      compareDeep(methodElement, o.methodElement, true) and compareDeep(bodySiteElement, o.bodySiteElement, true);
  end;
end;

function TFhirSpecimenCollection.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenCollection;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSpecimenCollection)) then
    result := false
  else
  begin
    o := TFhirSpecimenCollection(other);
    result := compareValues(commentList, o.commentList, true);
  end;
end;

{ TFhirSpecimenCollection }

Procedure TFhirSpecimenCollection.SetCollector(value : TFhirReference{TFhirPractitioner});
begin
  FCollector.free;
  FCollector := value;
end;

Function TFhirSpecimenCollection.GetCommentList : TFhirStringList;
begin
  if FCommentList = nil then
    FCommentList := TFhirStringList.Create;
  result := FCommentList;
end;

Function TFhirSpecimenCollection.GetHasCommentList : boolean;
begin
  result := (FCommentList <> nil) and (FCommentList.count > 0);
end;

Procedure TFhirSpecimenCollection.SetCollected(value : TFhirType);
begin
  FCollected.free;
  FCollected := value;
end;

Procedure TFhirSpecimenCollection.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirSpecimenCollection.SetMethod(value : TFhirCodeableConcept);
begin
  FMethod.free;
  FMethod := value;
end;

Procedure TFhirSpecimenCollection.SetBodySite(value : TFhirCodeableConcept);
begin
  FBodySite.free;
  FBodySite := value;
end;


{ TFhirSpecimenCollectionListEnumerator }

Constructor TFhirSpecimenCollectionListEnumerator.Create(list : TFhirSpecimenCollectionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSpecimenCollectionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSpecimenCollectionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSpecimenCollectionListEnumerator.GetCurrent : TFhirSpecimenCollection;
begin
  Result := FList[FIndex];
end;


{ TFhirSpecimenCollectionList }
procedure TFhirSpecimenCollectionList.AddItem(value: TFhirSpecimenCollection);
begin
  assert(value.ClassName = 'TFhirSpecimenCollection', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSpecimenCollection');
  add(value);
end;


function TFhirSpecimenCollectionList.Append: TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenCollectionList.ClearItems;
begin
  Clear;
end;

function TFhirSpecimenCollectionList.GetEnumerator : TFhirSpecimenCollectionListEnumerator;
begin
  result := TFhirSpecimenCollectionListEnumerator.Create(self.link);
end;

function TFhirSpecimenCollectionList.Clone: TFhirSpecimenCollectionList;
begin
  result := TFhirSpecimenCollectionList(inherited Clone);
end;

function TFhirSpecimenCollectionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSpecimenCollectionList.GetItemN(index: Integer): TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection(ObjectByIndex[index]);
end;

function TFhirSpecimenCollectionList.IndexOf(value: TFhirSpecimenCollection): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSpecimenCollectionList.Insert(index: Integer): TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenCollectionList.InsertItem(index: Integer; value: TFhirSpecimenCollection);
begin
  assert(value is TFhirSpecimenCollection);
  Inherited Insert(index, value);
end;

function TFhirSpecimenCollectionList.Item(index: Integer): TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection(ObjectByIndex[index]);
end;

function TFhirSpecimenCollectionList.Link: TFhirSpecimenCollectionList;
begin
  result := TFhirSpecimenCollectionList(inherited Link);
end;

procedure TFhirSpecimenCollectionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSpecimenCollectionList.SetItemByIndex(index: Integer; value: TFhirSpecimenCollection);
begin
  assert(value is TFhirSpecimenCollection);
  FhirSpecimenCollections[index] := value;
end;

procedure TFhirSpecimenCollectionList.SetItemN(index: Integer; value: TFhirSpecimenCollection);
begin
  assert(value is TFhirSpecimenCollection);
  ObjectByIndex[index] := value;
end;

{ TFhirSpecimenTreatment }

constructor TFhirSpecimenTreatment.Create;
begin
  inherited;
end;

destructor TFhirSpecimenTreatment.Destroy;
begin
  FDescription.free;
  FProcedure_.free;
  FAdditiveList.Free;
  inherited;
end;

procedure TFhirSpecimenTreatment.Assign(oSource : TAdvObject);
begin
  inherited;
  descriptionElement := TFhirSpecimenTreatment(oSource).descriptionElement.Clone;
  procedure_ := TFhirSpecimenTreatment(oSource).procedure_.Clone;
  if (TFhirSpecimenTreatment(oSource).FAdditiveList = nil) then
  begin
    FAdditiveList.free;
    FAdditiveList := nil;
  end
  else
  begin
    FAdditiveList := TFhirReferenceList{TFhirSubstance}.Create;
    FAdditiveList.Assign(TFhirSpecimenTreatment(oSource).FAdditiveList);
  end;
end;

procedure TFhirSpecimenTreatment.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'procedure') Then
     list.add(FProcedure_.Link);
  if (child_name = 'additive') Then
    list.addAll(FAdditiveList);
end;

procedure TFhirSpecimenTreatment.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'procedure', 'CodeableConcept', FProcedure_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additive', 'Reference(Substance)', FAdditiveList.Link)){3};
end;

procedure TFhirSpecimenTreatment.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'procedure') then Procedure_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'additive') then AdditiveList.add(propValue as TFhirReference{TFhirSubstance}){2}
  else inherited;
end;

function TFhirSpecimenTreatment.FhirType : string;
begin
  result := 'treatment';
end;

function TFhirSpecimenTreatment.Link : TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment(inherited Link);
end;

function TFhirSpecimenTreatment.Clone : TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment(inherited Clone);
end;

function TFhirSpecimenTreatment.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenTreatment;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSpecimenTreatment)) then
    result := false
  else
  begin
    o := TFhirSpecimenTreatment(other);
    result := compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(procedure_Element, o.procedure_Element, true) and 
      compareDeep(additiveList, o.additiveList, true);
  end;
end;

function TFhirSpecimenTreatment.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenTreatment;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSpecimenTreatment)) then
    result := false
  else
  begin
    o := TFhirSpecimenTreatment(other);
    result := compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirSpecimenTreatment }

Procedure TFhirSpecimenTreatment.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirSpecimenTreatment.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirSpecimenTreatment.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirSpecimenTreatment.SetProcedure_(value : TFhirCodeableConcept);
begin
  FProcedure_.free;
  FProcedure_ := value;
end;

Function TFhirSpecimenTreatment.GetAdditiveList : TFhirReferenceList{TFhirSubstance};
begin
  if FAdditiveList = nil then
    FAdditiveList := TFhirReferenceList{TFhirSubstance}.Create;
  result := FAdditiveList;
end;

Function TFhirSpecimenTreatment.GetHasAdditiveList : boolean;
begin
  result := (FAdditiveList <> nil) and (FAdditiveList.count > 0);
end;


{ TFhirSpecimenTreatmentListEnumerator }

Constructor TFhirSpecimenTreatmentListEnumerator.Create(list : TFhirSpecimenTreatmentList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSpecimenTreatmentListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSpecimenTreatmentListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSpecimenTreatmentListEnumerator.GetCurrent : TFhirSpecimenTreatment;
begin
  Result := FList[FIndex];
end;


{ TFhirSpecimenTreatmentList }
procedure TFhirSpecimenTreatmentList.AddItem(value: TFhirSpecimenTreatment);
begin
  assert(value.ClassName = 'TFhirSpecimenTreatment', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSpecimenTreatment');
  add(value);
end;


function TFhirSpecimenTreatmentList.Append: TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenTreatmentList.ClearItems;
begin
  Clear;
end;

function TFhirSpecimenTreatmentList.GetEnumerator : TFhirSpecimenTreatmentListEnumerator;
begin
  result := TFhirSpecimenTreatmentListEnumerator.Create(self.link);
end;

function TFhirSpecimenTreatmentList.Clone: TFhirSpecimenTreatmentList;
begin
  result := TFhirSpecimenTreatmentList(inherited Clone);
end;

function TFhirSpecimenTreatmentList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSpecimenTreatmentList.GetItemN(index: Integer): TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment(ObjectByIndex[index]);
end;

function TFhirSpecimenTreatmentList.IndexOf(value: TFhirSpecimenTreatment): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSpecimenTreatmentList.Insert(index: Integer): TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenTreatmentList.InsertItem(index: Integer; value: TFhirSpecimenTreatment);
begin
  assert(value is TFhirSpecimenTreatment);
  Inherited Insert(index, value);
end;

function TFhirSpecimenTreatmentList.Item(index: Integer): TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment(ObjectByIndex[index]);
end;

function TFhirSpecimenTreatmentList.Link: TFhirSpecimenTreatmentList;
begin
  result := TFhirSpecimenTreatmentList(inherited Link);
end;

procedure TFhirSpecimenTreatmentList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSpecimenTreatmentList.SetItemByIndex(index: Integer; value: TFhirSpecimenTreatment);
begin
  assert(value is TFhirSpecimenTreatment);
  FhirSpecimenTreatments[index] := value;
end;

procedure TFhirSpecimenTreatmentList.SetItemN(index: Integer; value: TFhirSpecimenTreatment);
begin
  assert(value is TFhirSpecimenTreatment);
  ObjectByIndex[index] := value;
end;

{ TFhirSpecimenContainer }

constructor TFhirSpecimenContainer.Create;
begin
  inherited;
end;

destructor TFhirSpecimenContainer.Destroy;
begin
  FIdentifierList.Free;
  FDescription.free;
  FType_.free;
  FCapacity.free;
  FSpecimenQuantity.free;
  FAdditive.free;
  inherited;
end;

procedure TFhirSpecimenContainer.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirSpecimenContainer(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirSpecimenContainer(oSource).FIdentifierList);
  end;
  descriptionElement := TFhirSpecimenContainer(oSource).descriptionElement.Clone;
  type_ := TFhirSpecimenContainer(oSource).type_.Clone;
  capacity := TFhirSpecimenContainer(oSource).capacity.Clone;
  specimenQuantity := TFhirSpecimenContainer(oSource).specimenQuantity.Clone;
  additive := TFhirSpecimenContainer(oSource).additive.Clone;
end;

procedure TFhirSpecimenContainer.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'capacity') Then
     list.add(FCapacity.Link);
  if (child_name = 'specimenQuantity') Then
     list.add(FSpecimenQuantity.Link);
  if (child_name = 'additive[x]') Then
     list.add(FAdditive.Link);
end;

procedure TFhirSpecimenContainer.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'capacity', 'SimpleQuantity', FCapacity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'specimenQuantity', 'SimpleQuantity', FSpecimenQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'additive[x]', 'CodeableConcept|Reference(Substance)', FAdditive.Link));{2}
end;

procedure TFhirSpecimenContainer.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'capacity') then Capacity := propValue as TFhirQuantity{4b}
  else if (propName = 'specimenQuantity') then SpecimenQuantity := propValue as TFhirQuantity{4b}
  else if (propName.startsWith('additive')) then Additive := propValue as TFhirType{4}
  else inherited;
end;

function TFhirSpecimenContainer.FhirType : string;
begin
  result := 'container';
end;

function TFhirSpecimenContainer.Link : TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer(inherited Link);
end;

function TFhirSpecimenContainer.Clone : TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer(inherited Clone);
end;

function TFhirSpecimenContainer.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenContainer;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSpecimenContainer)) then
    result := false
  else
  begin
    o := TFhirSpecimenContainer(other);
    result := compareDeep(identifierList, o.identifierList, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(type_Element, o.type_Element, true) and compareDeep(capacityElement, o.capacityElement, true) and 
      compareDeep(specimenQuantityElement, o.specimenQuantityElement, true) and compareDeep(additiveElement, o.additiveElement, true);
  end;
end;

function TFhirSpecimenContainer.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSpecimenContainer;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSpecimenContainer)) then
    result := false
  else
  begin
    o := TFhirSpecimenContainer(other);
    result := compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirSpecimenContainer }

Function TFhirSpecimenContainer.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirSpecimenContainer.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirSpecimenContainer.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirSpecimenContainer.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirSpecimenContainer.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirSpecimenContainer.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirSpecimenContainer.SetCapacity(value : TFhirQuantity);
begin
  FCapacity.free;
  FCapacity := value;
end;

Procedure TFhirSpecimenContainer.SetSpecimenQuantity(value : TFhirQuantity);
begin
  FSpecimenQuantity.free;
  FSpecimenQuantity := value;
end;

Procedure TFhirSpecimenContainer.SetAdditive(value : TFhirType);
begin
  FAdditive.free;
  FAdditive := value;
end;


{ TFhirSpecimenContainerListEnumerator }

Constructor TFhirSpecimenContainerListEnumerator.Create(list : TFhirSpecimenContainerList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSpecimenContainerListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSpecimenContainerListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSpecimenContainerListEnumerator.GetCurrent : TFhirSpecimenContainer;
begin
  Result := FList[FIndex];
end;


{ TFhirSpecimenContainerList }
procedure TFhirSpecimenContainerList.AddItem(value: TFhirSpecimenContainer);
begin
  assert(value.ClassName = 'TFhirSpecimenContainer', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSpecimenContainer');
  add(value);
end;


function TFhirSpecimenContainerList.Append: TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenContainerList.ClearItems;
begin
  Clear;
end;

function TFhirSpecimenContainerList.GetEnumerator : TFhirSpecimenContainerListEnumerator;
begin
  result := TFhirSpecimenContainerListEnumerator.Create(self.link);
end;

function TFhirSpecimenContainerList.Clone: TFhirSpecimenContainerList;
begin
  result := TFhirSpecimenContainerList(inherited Clone);
end;

function TFhirSpecimenContainerList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSpecimenContainerList.GetItemN(index: Integer): TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer(ObjectByIndex[index]);
end;

function TFhirSpecimenContainerList.IndexOf(value: TFhirSpecimenContainer): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSpecimenContainerList.Insert(index: Integer): TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenContainerList.InsertItem(index: Integer; value: TFhirSpecimenContainer);
begin
  assert(value is TFhirSpecimenContainer);
  Inherited Insert(index, value);
end;

function TFhirSpecimenContainerList.Item(index: Integer): TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer(ObjectByIndex[index]);
end;

function TFhirSpecimenContainerList.Link: TFhirSpecimenContainerList;
begin
  result := TFhirSpecimenContainerList(inherited Link);
end;

procedure TFhirSpecimenContainerList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSpecimenContainerList.SetItemByIndex(index: Integer; value: TFhirSpecimenContainer);
begin
  assert(value is TFhirSpecimenContainer);
  FhirSpecimenContainers[index] := value;
end;

procedure TFhirSpecimenContainerList.SetItemN(index: Integer; value: TFhirSpecimenContainer);
begin
  assert(value is TFhirSpecimenContainer);
  ObjectByIndex[index] := value;
end;

{ TFhirSpecimen }

constructor TFhirSpecimen.Create;
begin
  inherited;
end;

destructor TFhirSpecimen.Destroy;
begin
  FIdentifierList.Free;
  FStatus.free;
  FType_.free;
  FParentList.Free;
  FSubject.free;
  FAccessionIdentifier.free;
  FReceivedTime.free;
  FCollection.free;
  FTreatmentList.Free;
  FContainerList.Free;
  inherited;
end;

function TFhirSpecimen.GetResourceType : TFhirResourceType;
begin
  result := frtSpecimen;
end;

procedure TFhirSpecimen.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirSpecimen(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirSpecimen(oSource).FIdentifierList);
  end;
  FStatus := TFhirSpecimen(oSource).FStatus.Link;
  type_ := TFhirSpecimen(oSource).type_.Clone;
  if (TFhirSpecimen(oSource).FParentList = nil) then
  begin
    FParentList.free;
    FParentList := nil;
  end
  else
  begin
    FParentList := TFhirReferenceList{TFhirSpecimen}.Create;
    FParentList.Assign(TFhirSpecimen(oSource).FParentList);
  end;
  subject := TFhirSpecimen(oSource).subject.Clone;
  accessionIdentifier := TFhirSpecimen(oSource).accessionIdentifier.Clone;
  receivedTimeElement := TFhirSpecimen(oSource).receivedTimeElement.Clone;
  collection := TFhirSpecimen(oSource).collection.Clone;
  if (TFhirSpecimen(oSource).FTreatmentList = nil) then
  begin
    FTreatmentList.free;
    FTreatmentList := nil;
  end
  else
  begin
    FTreatmentList := TFhirSpecimenTreatmentList.Create;
    FTreatmentList.Assign(TFhirSpecimen(oSource).FTreatmentList);
  end;
  if (TFhirSpecimen(oSource).FContainerList = nil) then
  begin
    FContainerList.free;
    FContainerList := nil;
  end
  else
  begin
    FContainerList := TFhirSpecimenContainerList.Create;
    FContainerList.Assign(TFhirSpecimen(oSource).FContainerList);
  end;
end;

procedure TFhirSpecimen.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'parent') Then
    list.addAll(FParentList);
  if (child_name = 'subject') Then
     list.add(FSubject.Link);
  if (child_name = 'accessionIdentifier') Then
     list.add(FAccessionIdentifier.Link);
  if (child_name = 'receivedTime') Then
     list.add(FReceivedTime.Link);
  if (child_name = 'collection') Then
     list.add(FCollection.Link);
  if (child_name = 'treatment') Then
    list.addAll(FTreatmentList);
  if (child_name = 'container') Then
    list.addAll(FContainerList);
end;

procedure TFhirSpecimen.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'parent', 'Reference(Specimen)', FParentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'subject', 'Reference(Patient|Group|Device|Substance)', FSubject.Link));{2}
  oList.add(TFHIRProperty.create(self, 'accessionIdentifier', 'Identifier', FAccessionIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'receivedTime', 'dateTime', FReceivedTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'collection', '', FCollection.Link));{2}
  oList.add(TFHIRProperty.create(self, 'treatment', '', FTreatmentList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'container', '', FContainerList.Link)){3};
end;

procedure TFhirSpecimen.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'parent') then ParentList.add(propValue as TFhirReference{TFhirSpecimen}){2}
  else if (propName = 'subject') then Subject := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'accessionIdentifier') then AccessionIdentifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'receivedTime') then ReceivedTimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'collection') then Collection := propValue as TFhirSpecimenCollection{4b}
  else if (propName = 'treatment') then TreatmentList.add(propValue as TFhirSpecimenTreatment){2}
  else if (propName = 'container') then ContainerList.add(propValue as TFhirSpecimenContainer){2}
  else inherited;
end;

function TFhirSpecimen.FhirType : string;
begin
  result := 'Specimen';
end;

function TFhirSpecimen.Link : TFhirSpecimen;
begin
  result := TFhirSpecimen(inherited Link);
end;

function TFhirSpecimen.Clone : TFhirSpecimen;
begin
  result := TFhirSpecimen(inherited Clone);
end;

{ TFhirSpecimen }

Function TFhirSpecimen.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirSpecimen.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirSpecimen.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirSpecimen.GetStatusST : TFhirSpecimenStatus;
begin
  if FStatus = nil then
    result := TFhirSpecimenStatus(0)
  else
    result := TFhirSpecimenStatus(StringArrayIndexOfSensitive(CODES_TFhirSpecimenStatus, FStatus.value));
end;

Procedure TFhirSpecimen.SetStatusST(value : TFhirSpecimenStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirSpecimenStatus[value]);
end;

Procedure TFhirSpecimen.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirSpecimen.GetParentList : TFhirReferenceList{TFhirSpecimen};
begin
  if FParentList = nil then
    FParentList := TFhirReferenceList{TFhirSpecimen}.Create;
  result := FParentList;
end;

Function TFhirSpecimen.GetHasParentList : boolean;
begin
  result := (FParentList <> nil) and (FParentList.count > 0);
end;

Procedure TFhirSpecimen.SetSubject(value : TFhirReference{Resource});
begin
  FSubject.free;
  FSubject := value;
end;

Procedure TFhirSpecimen.SetAccessionIdentifier(value : TFhirIdentifier);
begin
  FAccessionIdentifier.free;
  FAccessionIdentifier := value;
end;

Procedure TFhirSpecimen.SetReceivedTime(value : TFhirDateTime);
begin
  FReceivedTime.free;
  FReceivedTime := value;
end;

Function TFhirSpecimen.GetReceivedTimeST : TDateAndTime;
begin
  if FReceivedTime = nil then
    result := nil
  else
    result := FReceivedTime.value;
end;

Procedure TFhirSpecimen.SetReceivedTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FReceivedTime = nil then
      FReceivedTime := TFhirDateTime.create;
    FReceivedTime.value := value
  end
  else if FReceivedTime <> nil then
    FReceivedTime.value := nil;
end;

Procedure TFhirSpecimen.SetCollection(value : TFhirSpecimenCollection);
begin
  FCollection.free;
  FCollection := value;
end;

Function TFhirSpecimen.GetTreatmentList : TFhirSpecimenTreatmentList;
begin
  if FTreatmentList = nil then
    FTreatmentList := TFhirSpecimenTreatmentList.Create;
  result := FTreatmentList;
end;

Function TFhirSpecimen.GetHasTreatmentList : boolean;
begin
  result := (FTreatmentList <> nil) and (FTreatmentList.count > 0);
end;

Function TFhirSpecimen.GetContainerList : TFhirSpecimenContainerList;
begin
  if FContainerList = nil then
    FContainerList := TFhirSpecimenContainerList.Create;
  result := FContainerList;
end;

Function TFhirSpecimen.GetHasContainerList : boolean;
begin
  result := (FContainerList <> nil) and (FContainerList.count > 0);
end;


{ TFhirSpecimenListEnumerator }

Constructor TFhirSpecimenListEnumerator.Create(list : TFhirSpecimenList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSpecimenListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSpecimenListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSpecimenListEnumerator.GetCurrent : TFhirSpecimen;
begin
  Result := FList[FIndex];
end;


{ TFhirSpecimenList }
procedure TFhirSpecimenList.AddItem(value: TFhirSpecimen);
begin
  assert(value.ClassName = 'TFhirSpecimen', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSpecimen');
  add(value);
end;


function TFhirSpecimenList.Append: TFhirSpecimen;
begin
  result := TFhirSpecimen.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenList.ClearItems;
begin
  Clear;
end;

function TFhirSpecimenList.GetEnumerator : TFhirSpecimenListEnumerator;
begin
  result := TFhirSpecimenListEnumerator.Create(self.link);
end;

function TFhirSpecimenList.Clone: TFhirSpecimenList;
begin
  result := TFhirSpecimenList(inherited Clone);
end;

function TFhirSpecimenList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSpecimenList.GetItemN(index: Integer): TFhirSpecimen;
begin
  result := TFhirSpecimen(ObjectByIndex[index]);
end;

function TFhirSpecimenList.IndexOf(value: TFhirSpecimen): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSpecimenList.Insert(index: Integer): TFhirSpecimen;
begin
  result := TFhirSpecimen.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSpecimenList.InsertItem(index: Integer; value: TFhirSpecimen);
begin
  assert(value is TFhirSpecimen);
  Inherited Insert(index, value);
end;

function TFhirSpecimenList.Item(index: Integer): TFhirSpecimen;
begin
  result := TFhirSpecimen(ObjectByIndex[index]);
end;

function TFhirSpecimenList.Link: TFhirSpecimenList;
begin
  result := TFhirSpecimenList(inherited Link);
end;

procedure TFhirSpecimenList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSpecimenList.SetItemByIndex(index: Integer; value: TFhirSpecimen);
begin
  assert(value is TFhirSpecimen);
  FhirSpecimen[index] := value;
end;

procedure TFhirSpecimenList.SetItemN(index: Integer; value: TFhirSpecimen);
begin
  assert(value is TFhirSpecimen);
  ObjectByIndex[index] := value;
end;

{ TFhirStructureDefinitionContact }

constructor TFhirStructureDefinitionContact.Create;
begin
  inherited;
end;

destructor TFhirStructureDefinitionContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirStructureDefinitionContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirStructureDefinitionContact(oSource).nameElement.Clone;
  if (TFhirStructureDefinitionContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirStructureDefinitionContact(oSource).FTelecomList);
  end;
end;

procedure TFhirStructureDefinitionContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirStructureDefinitionContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirStructureDefinitionContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirStructureDefinitionContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirStructureDefinitionContact.Link : TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact(inherited Link);
end;

function TFhirStructureDefinitionContact.Clone : TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact(inherited Clone);
end;

function TFhirStructureDefinitionContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionContact)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirStructureDefinitionContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionContact)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirStructureDefinitionContact }

Procedure TFhirStructureDefinitionContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirStructureDefinitionContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirStructureDefinitionContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirStructureDefinitionContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirStructureDefinitionContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirStructureDefinitionContactListEnumerator }

Constructor TFhirStructureDefinitionContactListEnumerator.Create(list : TFhirStructureDefinitionContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirStructureDefinitionContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirStructureDefinitionContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirStructureDefinitionContactListEnumerator.GetCurrent : TFhirStructureDefinitionContact;
begin
  Result := FList[FIndex];
end;


{ TFhirStructureDefinitionContactList }
procedure TFhirStructureDefinitionContactList.AddItem(value: TFhirStructureDefinitionContact);
begin
  assert(value.ClassName = 'TFhirStructureDefinitionContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirStructureDefinitionContact');
  add(value);
end;


function TFhirStructureDefinitionContactList.Append: TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionContactList.ClearItems;
begin
  Clear;
end;

function TFhirStructureDefinitionContactList.GetEnumerator : TFhirStructureDefinitionContactListEnumerator;
begin
  result := TFhirStructureDefinitionContactListEnumerator.Create(self.link);
end;

function TFhirStructureDefinitionContactList.Clone: TFhirStructureDefinitionContactList;
begin
  result := TFhirStructureDefinitionContactList(inherited Clone);
end;

function TFhirStructureDefinitionContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirStructureDefinitionContactList.GetItemN(index: Integer): TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionContactList.IndexOf(value: TFhirStructureDefinitionContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirStructureDefinitionContactList.Insert(index: Integer): TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionContactList.InsertItem(index: Integer; value: TFhirStructureDefinitionContact);
begin
  assert(value is TFhirStructureDefinitionContact);
  Inherited Insert(index, value);
end;

function TFhirStructureDefinitionContactList.Item(index: Integer): TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionContactList.Link: TFhirStructureDefinitionContactList;
begin
  result := TFhirStructureDefinitionContactList(inherited Link);
end;

procedure TFhirStructureDefinitionContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirStructureDefinitionContactList.SetItemByIndex(index: Integer; value: TFhirStructureDefinitionContact);
begin
  assert(value is TFhirStructureDefinitionContact);
  FhirStructureDefinitionContacts[index] := value;
end;

procedure TFhirStructureDefinitionContactList.SetItemN(index: Integer; value: TFhirStructureDefinitionContact);
begin
  assert(value is TFhirStructureDefinitionContact);
  ObjectByIndex[index] := value;
end;

{ TFhirStructureDefinitionMapping }

constructor TFhirStructureDefinitionMapping.Create;
begin
  inherited;
end;

destructor TFhirStructureDefinitionMapping.Destroy;
begin
  FIdentity.free;
  FUri.free;
  FName.free;
  FComments.free;
  inherited;
end;

procedure TFhirStructureDefinitionMapping.Assign(oSource : TAdvObject);
begin
  inherited;
  identityElement := TFhirStructureDefinitionMapping(oSource).identityElement.Clone;
  uriElement := TFhirStructureDefinitionMapping(oSource).uriElement.Clone;
  nameElement := TFhirStructureDefinitionMapping(oSource).nameElement.Clone;
  commentsElement := TFhirStructureDefinitionMapping(oSource).commentsElement.Clone;
end;

procedure TFhirStructureDefinitionMapping.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identity') Then
     list.add(FIdentity.Link);
  if (child_name = 'uri') Then
     list.add(FUri.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'comments') Then
     list.add(FComments.Link);
end;

procedure TFhirStructureDefinitionMapping.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identity', 'id', FIdentity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'uri', 'uri', FUri.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'comments', 'string', FComments.Link));{2}
end;

procedure TFhirStructureDefinitionMapping.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identity') then IdentityElement := propValue as TFhirId{5a}
  else if (propName = 'uri') then UriElement := propValue as TFhirUri{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'comments') then CommentsElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirStructureDefinitionMapping.FhirType : string;
begin
  result := 'mapping';
end;

function TFhirStructureDefinitionMapping.Link : TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping(inherited Link);
end;

function TFhirStructureDefinitionMapping.Clone : TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping(inherited Clone);
end;

function TFhirStructureDefinitionMapping.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionMapping;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionMapping)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionMapping(other);
    result := compareDeep(identityElement, o.identityElement, true) and compareDeep(uriElement, o.uriElement, true) and 
      compareDeep(nameElement, o.nameElement, true) and compareDeep(commentsElement, o.commentsElement, true);
  end;
end;

function TFhirStructureDefinitionMapping.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionMapping;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionMapping)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionMapping(other);
    result := compareValues(identityElement, o.identityElement, true) and compareValues(uriElement, o.uriElement, true) and 
      compareValues(nameElement, o.nameElement, true) and compareValues(commentsElement, o.commentsElement, true);
  end;
end;

{ TFhirStructureDefinitionMapping }

Procedure TFhirStructureDefinitionMapping.SetIdentity(value : TFhirId);
begin
  FIdentity.free;
  FIdentity := value;
end;

Function TFhirStructureDefinitionMapping.GetIdentityST : String;
begin
  if FIdentity = nil then
    result := ''
  else
    result := FIdentity.value;
end;

Procedure TFhirStructureDefinitionMapping.SetIdentityST(value : String);
begin
  if value <> '' then
  begin
    if FIdentity = nil then
      FIdentity := TFhirId.create;
    FIdentity.value := value
  end
  else if FIdentity <> nil then
    FIdentity.value := '';
end;

Procedure TFhirStructureDefinitionMapping.SetUri(value : TFhirUri);
begin
  FUri.free;
  FUri := value;
end;

Function TFhirStructureDefinitionMapping.GetUriST : String;
begin
  if FUri = nil then
    result := ''
  else
    result := FUri.value;
end;

Procedure TFhirStructureDefinitionMapping.SetUriST(value : String);
begin
  if value <> '' then
  begin
    if FUri = nil then
      FUri := TFhirUri.create;
    FUri.value := value
  end
  else if FUri <> nil then
    FUri.value := '';
end;

Procedure TFhirStructureDefinitionMapping.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirStructureDefinitionMapping.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirStructureDefinitionMapping.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirStructureDefinitionMapping.SetComments(value : TFhirString);
begin
  FComments.free;
  FComments := value;
end;

Function TFhirStructureDefinitionMapping.GetCommentsST : String;
begin
  if FComments = nil then
    result := ''
  else
    result := FComments.value;
end;

Procedure TFhirStructureDefinitionMapping.SetCommentsST(value : String);
begin
  if value <> '' then
  begin
    if FComments = nil then
      FComments := TFhirString.create;
    FComments.value := value
  end
  else if FComments <> nil then
    FComments.value := '';
end;


{ TFhirStructureDefinitionMappingListEnumerator }

Constructor TFhirStructureDefinitionMappingListEnumerator.Create(list : TFhirStructureDefinitionMappingList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirStructureDefinitionMappingListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirStructureDefinitionMappingListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirStructureDefinitionMappingListEnumerator.GetCurrent : TFhirStructureDefinitionMapping;
begin
  Result := FList[FIndex];
end;


{ TFhirStructureDefinitionMappingList }
procedure TFhirStructureDefinitionMappingList.AddItem(value: TFhirStructureDefinitionMapping);
begin
  assert(value.ClassName = 'TFhirStructureDefinitionMapping', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirStructureDefinitionMapping');
  add(value);
end;


function TFhirStructureDefinitionMappingList.Append: TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionMappingList.ClearItems;
begin
  Clear;
end;

function TFhirStructureDefinitionMappingList.GetEnumerator : TFhirStructureDefinitionMappingListEnumerator;
begin
  result := TFhirStructureDefinitionMappingListEnumerator.Create(self.link);
end;

function TFhirStructureDefinitionMappingList.Clone: TFhirStructureDefinitionMappingList;
begin
  result := TFhirStructureDefinitionMappingList(inherited Clone);
end;

function TFhirStructureDefinitionMappingList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirStructureDefinitionMappingList.GetItemN(index: Integer): TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionMappingList.IndexOf(value: TFhirStructureDefinitionMapping): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirStructureDefinitionMappingList.Insert(index: Integer): TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionMappingList.InsertItem(index: Integer; value: TFhirStructureDefinitionMapping);
begin
  assert(value is TFhirStructureDefinitionMapping);
  Inherited Insert(index, value);
end;

function TFhirStructureDefinitionMappingList.Item(index: Integer): TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionMappingList.Link: TFhirStructureDefinitionMappingList;
begin
  result := TFhirStructureDefinitionMappingList(inherited Link);
end;

procedure TFhirStructureDefinitionMappingList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirStructureDefinitionMappingList.SetItemByIndex(index: Integer; value: TFhirStructureDefinitionMapping);
begin
  assert(value is TFhirStructureDefinitionMapping);
  FhirStructureDefinitionMappings[index] := value;
end;

procedure TFhirStructureDefinitionMappingList.SetItemN(index: Integer; value: TFhirStructureDefinitionMapping);
begin
  assert(value is TFhirStructureDefinitionMapping);
  ObjectByIndex[index] := value;
end;

{ TFhirStructureDefinitionSnapshot }

constructor TFhirStructureDefinitionSnapshot.Create;
begin
  inherited;
end;

destructor TFhirStructureDefinitionSnapshot.Destroy;
begin
  FElementList.Free;
  inherited;
end;

procedure TFhirStructureDefinitionSnapshot.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirStructureDefinitionSnapshot(oSource).FElementList = nil) then
  begin
    FElementList.free;
    FElementList := nil;
  end
  else
  begin
    FElementList := TFhirElementDefinitionList.Create;
    FElementList.Assign(TFhirStructureDefinitionSnapshot(oSource).FElementList);
  end;
end;

procedure TFhirStructureDefinitionSnapshot.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'element') Then
    list.addAll(FElementList);
end;

procedure TFhirStructureDefinitionSnapshot.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'element', 'ElementDefinition', FElementList.Link)){3};
end;

procedure TFhirStructureDefinitionSnapshot.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'element') then ElementList.add(propValue as TFhirElementDefinition){2}
  else inherited;
end;

function TFhirStructureDefinitionSnapshot.FhirType : string;
begin
  result := 'snapshot';
end;

function TFhirStructureDefinitionSnapshot.Link : TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot(inherited Link);
end;

function TFhirStructureDefinitionSnapshot.Clone : TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot(inherited Clone);
end;

function TFhirStructureDefinitionSnapshot.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionSnapshot;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionSnapshot)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionSnapshot(other);
    result := compareDeep(elementList, o.elementList, true);
  end;
end;

function TFhirStructureDefinitionSnapshot.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionSnapshot;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionSnapshot)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionSnapshot(other);
    result := true;
  end;
end;

{ TFhirStructureDefinitionSnapshot }

Function TFhirStructureDefinitionSnapshot.GetElementList : TFhirElementDefinitionList;
begin
  if FElementList = nil then
    FElementList := TFhirElementDefinitionList.Create;
  result := FElementList;
end;

Function TFhirStructureDefinitionSnapshot.GetHasElementList : boolean;
begin
  result := (FElementList <> nil) and (FElementList.count > 0);
end;


{ TFhirStructureDefinitionSnapshotListEnumerator }

Constructor TFhirStructureDefinitionSnapshotListEnumerator.Create(list : TFhirStructureDefinitionSnapshotList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirStructureDefinitionSnapshotListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirStructureDefinitionSnapshotListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirStructureDefinitionSnapshotListEnumerator.GetCurrent : TFhirStructureDefinitionSnapshot;
begin
  Result := FList[FIndex];
end;


{ TFhirStructureDefinitionSnapshotList }
procedure TFhirStructureDefinitionSnapshotList.AddItem(value: TFhirStructureDefinitionSnapshot);
begin
  assert(value.ClassName = 'TFhirStructureDefinitionSnapshot', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirStructureDefinitionSnapshot');
  add(value);
end;


function TFhirStructureDefinitionSnapshotList.Append: TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionSnapshotList.ClearItems;
begin
  Clear;
end;

function TFhirStructureDefinitionSnapshotList.GetEnumerator : TFhirStructureDefinitionSnapshotListEnumerator;
begin
  result := TFhirStructureDefinitionSnapshotListEnumerator.Create(self.link);
end;

function TFhirStructureDefinitionSnapshotList.Clone: TFhirStructureDefinitionSnapshotList;
begin
  result := TFhirStructureDefinitionSnapshotList(inherited Clone);
end;

function TFhirStructureDefinitionSnapshotList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirStructureDefinitionSnapshotList.GetItemN(index: Integer): TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionSnapshotList.IndexOf(value: TFhirStructureDefinitionSnapshot): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirStructureDefinitionSnapshotList.Insert(index: Integer): TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionSnapshotList.InsertItem(index: Integer; value: TFhirStructureDefinitionSnapshot);
begin
  assert(value is TFhirStructureDefinitionSnapshot);
  Inherited Insert(index, value);
end;

function TFhirStructureDefinitionSnapshotList.Item(index: Integer): TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionSnapshotList.Link: TFhirStructureDefinitionSnapshotList;
begin
  result := TFhirStructureDefinitionSnapshotList(inherited Link);
end;

procedure TFhirStructureDefinitionSnapshotList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirStructureDefinitionSnapshotList.SetItemByIndex(index: Integer; value: TFhirStructureDefinitionSnapshot);
begin
  assert(value is TFhirStructureDefinitionSnapshot);
  FhirStructureDefinitionSnapshots[index] := value;
end;

procedure TFhirStructureDefinitionSnapshotList.SetItemN(index: Integer; value: TFhirStructureDefinitionSnapshot);
begin
  assert(value is TFhirStructureDefinitionSnapshot);
  ObjectByIndex[index] := value;
end;

{ TFhirStructureDefinitionDifferential }

constructor TFhirStructureDefinitionDifferential.Create;
begin
  inherited;
end;

destructor TFhirStructureDefinitionDifferential.Destroy;
begin
  FElementList.Free;
  inherited;
end;

procedure TFhirStructureDefinitionDifferential.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirStructureDefinitionDifferential(oSource).FElementList = nil) then
  begin
    FElementList.free;
    FElementList := nil;
  end
  else
  begin
    FElementList := TFhirElementDefinitionList.Create;
    FElementList.Assign(TFhirStructureDefinitionDifferential(oSource).FElementList);
  end;
end;

procedure TFhirStructureDefinitionDifferential.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'element') Then
    list.addAll(FElementList);
end;

procedure TFhirStructureDefinitionDifferential.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'element', 'ElementDefinition', FElementList.Link)){3};
end;

procedure TFhirStructureDefinitionDifferential.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'element') then ElementList.add(propValue as TFhirElementDefinition){2}
  else inherited;
end;

function TFhirStructureDefinitionDifferential.FhirType : string;
begin
  result := 'differential';
end;

function TFhirStructureDefinitionDifferential.Link : TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential(inherited Link);
end;

function TFhirStructureDefinitionDifferential.Clone : TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential(inherited Clone);
end;

function TFhirStructureDefinitionDifferential.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionDifferential;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionDifferential)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionDifferential(other);
    result := compareDeep(elementList, o.elementList, true);
  end;
end;

function TFhirStructureDefinitionDifferential.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirStructureDefinitionDifferential;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirStructureDefinitionDifferential)) then
    result := false
  else
  begin
    o := TFhirStructureDefinitionDifferential(other);
    result := true;
  end;
end;

{ TFhirStructureDefinitionDifferential }

Function TFhirStructureDefinitionDifferential.GetElementList : TFhirElementDefinitionList;
begin
  if FElementList = nil then
    FElementList := TFhirElementDefinitionList.Create;
  result := FElementList;
end;

Function TFhirStructureDefinitionDifferential.GetHasElementList : boolean;
begin
  result := (FElementList <> nil) and (FElementList.count > 0);
end;


{ TFhirStructureDefinitionDifferentialListEnumerator }

Constructor TFhirStructureDefinitionDifferentialListEnumerator.Create(list : TFhirStructureDefinitionDifferentialList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirStructureDefinitionDifferentialListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirStructureDefinitionDifferentialListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirStructureDefinitionDifferentialListEnumerator.GetCurrent : TFhirStructureDefinitionDifferential;
begin
  Result := FList[FIndex];
end;


{ TFhirStructureDefinitionDifferentialList }
procedure TFhirStructureDefinitionDifferentialList.AddItem(value: TFhirStructureDefinitionDifferential);
begin
  assert(value.ClassName = 'TFhirStructureDefinitionDifferential', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirStructureDefinitionDifferential');
  add(value);
end;


function TFhirStructureDefinitionDifferentialList.Append: TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionDifferentialList.ClearItems;
begin
  Clear;
end;

function TFhirStructureDefinitionDifferentialList.GetEnumerator : TFhirStructureDefinitionDifferentialListEnumerator;
begin
  result := TFhirStructureDefinitionDifferentialListEnumerator.Create(self.link);
end;

function TFhirStructureDefinitionDifferentialList.Clone: TFhirStructureDefinitionDifferentialList;
begin
  result := TFhirStructureDefinitionDifferentialList(inherited Clone);
end;

function TFhirStructureDefinitionDifferentialList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirStructureDefinitionDifferentialList.GetItemN(index: Integer): TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionDifferentialList.IndexOf(value: TFhirStructureDefinitionDifferential): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirStructureDefinitionDifferentialList.Insert(index: Integer): TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionDifferentialList.InsertItem(index: Integer; value: TFhirStructureDefinitionDifferential);
begin
  assert(value is TFhirStructureDefinitionDifferential);
  Inherited Insert(index, value);
end;

function TFhirStructureDefinitionDifferentialList.Item(index: Integer): TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionDifferentialList.Link: TFhirStructureDefinitionDifferentialList;
begin
  result := TFhirStructureDefinitionDifferentialList(inherited Link);
end;

procedure TFhirStructureDefinitionDifferentialList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirStructureDefinitionDifferentialList.SetItemByIndex(index: Integer; value: TFhirStructureDefinitionDifferential);
begin
  assert(value is TFhirStructureDefinitionDifferential);
  FhirStructureDefinitionDifferentials[index] := value;
end;

procedure TFhirStructureDefinitionDifferentialList.SetItemN(index: Integer; value: TFhirStructureDefinitionDifferential);
begin
  assert(value is TFhirStructureDefinitionDifferential);
  ObjectByIndex[index] := value;
end;

{ TFhirStructureDefinition }

constructor TFhirStructureDefinition.Create;
begin
  inherited;
end;

destructor TFhirStructureDefinition.Destroy;
begin
  FUrl.free;
  FIdentifierList.Free;
  FVersion.free;
  FName.free;
  FDisplay.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FUseContextList.Free;
  FRequirements.free;
  FCopyright.free;
  FCodeList.Free;
  FFhirVersion.free;
  FMappingList.Free;
  FKind.free;
  FConstrainedType.free;
  FAbstract.free;
  FContextType.free;
  FContextList.Free;
  FBase.free;
  FSnapshot.free;
  FDifferential.free;
  inherited;
end;

function TFhirStructureDefinition.GetResourceType : TFhirResourceType;
begin
  result := frtStructureDefinition;
end;

procedure TFhirStructureDefinition.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirStructureDefinition(oSource).urlElement.Clone;
  if (TFhirStructureDefinition(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirStructureDefinition(oSource).FIdentifierList);
  end;
  versionElement := TFhirStructureDefinition(oSource).versionElement.Clone;
  nameElement := TFhirStructureDefinition(oSource).nameElement.Clone;
  displayElement := TFhirStructureDefinition(oSource).displayElement.Clone;
  FStatus := TFhirStructureDefinition(oSource).FStatus.Link;
  experimentalElement := TFhirStructureDefinition(oSource).experimentalElement.Clone;
  publisherElement := TFhirStructureDefinition(oSource).publisherElement.Clone;
  if (TFhirStructureDefinition(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirStructureDefinitionContactList.Create;
    FContactList.Assign(TFhirStructureDefinition(oSource).FContactList);
  end;
  dateElement := TFhirStructureDefinition(oSource).dateElement.Clone;
  descriptionElement := TFhirStructureDefinition(oSource).descriptionElement.Clone;
  if (TFhirStructureDefinition(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirStructureDefinition(oSource).FUseContextList);
  end;
  requirementsElement := TFhirStructureDefinition(oSource).requirementsElement.Clone;
  copyrightElement := TFhirStructureDefinition(oSource).copyrightElement.Clone;
  if (TFhirStructureDefinition(oSource).FCodeList = nil) then
  begin
    FCodeList.free;
    FCodeList := nil;
  end
  else
  begin
    FCodeList := TFhirCodingList.Create;
    FCodeList.Assign(TFhirStructureDefinition(oSource).FCodeList);
  end;
  fhirVersionElement := TFhirStructureDefinition(oSource).fhirVersionElement.Clone;
  if (TFhirStructureDefinition(oSource).FMappingList = nil) then
  begin
    FMappingList.free;
    FMappingList := nil;
  end
  else
  begin
    FMappingList := TFhirStructureDefinitionMappingList.Create;
    FMappingList.Assign(TFhirStructureDefinition(oSource).FMappingList);
  end;
  FKind := TFhirStructureDefinition(oSource).FKind.Link;
  constrainedTypeElement := TFhirStructureDefinition(oSource).constrainedTypeElement.Clone;
  abstractElement := TFhirStructureDefinition(oSource).abstractElement.Clone;
  FContextType := TFhirStructureDefinition(oSource).FContextType.Link;
  if (TFhirStructureDefinition(oSource).FContextList = nil) then
  begin
    FContextList.free;
    FContextList := nil;
  end
  else
  begin
    FContextList := TFhirStringList.Create;
    FContextList.Assign(TFhirStructureDefinition(oSource).FContextList);
  end;
  baseElement := TFhirStructureDefinition(oSource).baseElement.Clone;
  snapshot := TFhirStructureDefinition(oSource).snapshot.Clone;
  differential := TFhirStructureDefinition(oSource).differential.Clone;
end;

procedure TFhirStructureDefinition.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'display') Then
     list.add(FDisplay.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'code') Then
    list.addAll(FCodeList);
  if (child_name = 'fhirVersion') Then
     list.add(FFhirVersion.Link);
  if (child_name = 'mapping') Then
    list.addAll(FMappingList);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'constrainedType') Then
     list.add(FConstrainedType.Link);
  if (child_name = 'abstract') Then
     list.add(FAbstract.Link);
  if (child_name = 'contextType') Then
     list.add(FContextType.Link);
  if (child_name = 'context') Then
    list.addAll(FContextList);
  if (child_name = 'base') Then
     list.add(FBase.Link);
  if (child_name = 'snapshot') Then
     list.add(FSnapshot.Link);
  if (child_name = 'differential') Then
     list.add(FDifferential.Link);
end;

procedure TFhirStructureDefinition.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'display', 'string', FDisplay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'Coding', FCodeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'fhirVersion', 'id', FFhirVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'mapping', '', FMappingList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'kind', 'code', FKind.Link));{1}
  oList.add(TFHIRProperty.create(self, 'constrainedType', 'code', FConstrainedType.Link));{2}
  oList.add(TFHIRProperty.create(self, 'abstract', 'boolean', FAbstract.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contextType', 'code', FContextType.Link));{1}
  oList.add(TFHIRProperty.create(self, 'context', 'string', FContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'base', 'uri', FBase.Link));{2}
  oList.add(TFHIRProperty.create(self, 'snapshot', '', FSnapshot.Link));{2}
  oList.add(TFHIRProperty.create(self, 'differential', '', FDifferential.Link));{2}
end;

procedure TFhirStructureDefinition.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'display') then DisplayElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirStructureDefinitionContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'code') then CodeList.add(propValue as TFhirCoding){2}
  else if (propName = 'fhirVersion') then FhirVersionElement := propValue as TFhirId{5a}
  else if (propName = 'mapping') then MappingList.add(propValue as TFhirStructureDefinitionMapping){2}
  else if (propName = 'kind') then KindElement := propValue as TFHIREnum
  else if (propName = 'constrainedType') then
    if propValue is TFHIRCode then
      ConstrainedTypeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      ConstrainedTypeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'abstract') then AbstractElement := propValue as TFhirBoolean{5a}
  else if (propName = 'contextType') then ContextTypeElement := propValue as TFHIREnum
  else if (propName = 'context') then ContextList.add(propValue as TFhirString){2}
  else if (propName = 'base') then BaseElement := propValue as TFhirUri{5a}
  else if (propName = 'snapshot') then Snapshot := propValue as TFhirStructureDefinitionSnapshot{4b}
  else if (propName = 'differential') then Differential := propValue as TFhirStructureDefinitionDifferential{4b}
  else inherited;
end;

function TFhirStructureDefinition.FhirType : string;
begin
  result := 'StructureDefinition';
end;

function TFhirStructureDefinition.Link : TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition(inherited Link);
end;

function TFhirStructureDefinition.Clone : TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition(inherited Clone);
end;

{ TFhirStructureDefinition }

Procedure TFhirStructureDefinition.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirStructureDefinition.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirStructureDefinition.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Function TFhirStructureDefinition.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirStructureDefinition.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirStructureDefinition.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirStructureDefinition.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirStructureDefinition.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirStructureDefinition.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirStructureDefinition.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirStructureDefinition.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirStructureDefinition.SetDisplay(value : TFhirString);
begin
  FDisplay.free;
  FDisplay := value;
end;

Function TFhirStructureDefinition.GetDisplayST : String;
begin
  if FDisplay = nil then
    result := ''
  else
    result := FDisplay.value;
end;

Procedure TFhirStructureDefinition.SetDisplayST(value : String);
begin
  if value <> '' then
  begin
    if FDisplay = nil then
      FDisplay := TFhirString.create;
    FDisplay.value := value
  end
  else if FDisplay <> nil then
    FDisplay.value := '';
end;

Procedure TFhirStructureDefinition.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirStructureDefinition.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirStructureDefinition.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirStructureDefinition.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirStructureDefinition.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirStructureDefinition.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirStructureDefinition.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirStructureDefinition.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirStructureDefinition.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirStructureDefinition.GetContactList : TFhirStructureDefinitionContactList;
begin
  if FContactList = nil then
    FContactList := TFhirStructureDefinitionContactList.Create;
  result := FContactList;
end;

Function TFhirStructureDefinition.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirStructureDefinition.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirStructureDefinition.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirStructureDefinition.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirStructureDefinition.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirStructureDefinition.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirStructureDefinition.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirStructureDefinition.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirStructureDefinition.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirStructureDefinition.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirStructureDefinition.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirStructureDefinition.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirStructureDefinition.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirStructureDefinition.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirStructureDefinition.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Function TFhirStructureDefinition.GetCodeList : TFhirCodingList;
begin
  if FCodeList = nil then
    FCodeList := TFhirCodingList.Create;
  result := FCodeList;
end;

Function TFhirStructureDefinition.GetHasCodeList : boolean;
begin
  result := (FCodeList <> nil) and (FCodeList.count > 0);
end;

Procedure TFhirStructureDefinition.SetFhirVersion(value : TFhirId);
begin
  FFhirVersion.free;
  FFhirVersion := value;
end;

Function TFhirStructureDefinition.GetFhirVersionST : String;
begin
  if FFhirVersion = nil then
    result := ''
  else
    result := FFhirVersion.value;
end;

Procedure TFhirStructureDefinition.SetFhirVersionST(value : String);
begin
  if value <> '' then
  begin
    if FFhirVersion = nil then
      FFhirVersion := TFhirId.create;
    FFhirVersion.value := value
  end
  else if FFhirVersion <> nil then
    FFhirVersion.value := '';
end;

Function TFhirStructureDefinition.GetMappingList : TFhirStructureDefinitionMappingList;
begin
  if FMappingList = nil then
    FMappingList := TFhirStructureDefinitionMappingList.Create;
  result := FMappingList;
end;

Function TFhirStructureDefinition.GetHasMappingList : boolean;
begin
  result := (FMappingList <> nil) and (FMappingList.count > 0);
end;

Procedure TFhirStructureDefinition.SetKind(value : TFhirEnum);
begin
  FKind.free;
  FKind := value;
end;

Function TFhirStructureDefinition.GetKindST : TFhirStructureDefinitionKind;
begin
  if FKind = nil then
    result := TFhirStructureDefinitionKind(0)
  else
    result := TFhirStructureDefinitionKind(StringArrayIndexOfSensitive(CODES_TFhirStructureDefinitionKind, FKind.value));
end;

Procedure TFhirStructureDefinition.SetKindST(value : TFhirStructureDefinitionKind);
begin
  if ord(value) = 0 then
    KindElement := nil
  else
    KindElement := TFhirEnum.create(CODES_TFhirStructureDefinitionKind[value]);
end;

Procedure TFhirStructureDefinition.SetConstrainedType(value : TFhirCode);
begin
  FConstrainedType.free;
  FConstrainedType := value;
end;

Function TFhirStructureDefinition.GetConstrainedTypeST : String;
begin
  if FConstrainedType = nil then
    result := ''
  else
    result := FConstrainedType.value;
end;

Procedure TFhirStructureDefinition.SetConstrainedTypeST(value : String);
begin
  if value <> '' then
  begin
    if FConstrainedType = nil then
      FConstrainedType := TFhirCode.create;
    FConstrainedType.value := value
  end
  else if FConstrainedType <> nil then
    FConstrainedType.value := '';
end;

Procedure TFhirStructureDefinition.SetAbstract(value : TFhirBoolean);
begin
  FAbstract.free;
  FAbstract := value;
end;

Function TFhirStructureDefinition.GetAbstractST : Boolean;
begin
  if FAbstract = nil then
    result := false
  else
    result := FAbstract.value;
end;

Procedure TFhirStructureDefinition.SetAbstractST(value : Boolean);
begin
  if FAbstract = nil then
    FAbstract := TFhirBoolean.create;
  FAbstract.value := value
end;

Procedure TFhirStructureDefinition.SetContextType(value : TFhirEnum);
begin
  FContextType.free;
  FContextType := value;
end;

Function TFhirStructureDefinition.GetContextTypeST : TFhirExtensionContext;
begin
  if FContextType = nil then
    result := TFhirExtensionContext(0)
  else
    result := TFhirExtensionContext(StringArrayIndexOfSensitive(CODES_TFhirExtensionContext, FContextType.value));
end;

Procedure TFhirStructureDefinition.SetContextTypeST(value : TFhirExtensionContext);
begin
  if ord(value) = 0 then
    ContextTypeElement := nil
  else
    ContextTypeElement := TFhirEnum.create(CODES_TFhirExtensionContext[value]);
end;

Function TFhirStructureDefinition.GetContextList : TFhirStringList;
begin
  if FContextList = nil then
    FContextList := TFhirStringList.Create;
  result := FContextList;
end;

Function TFhirStructureDefinition.GetHasContextList : boolean;
begin
  result := (FContextList <> nil) and (FContextList.count > 0);
end;

Procedure TFhirStructureDefinition.SetBase(value : TFhirUri);
begin
  FBase.free;
  FBase := value;
end;

Function TFhirStructureDefinition.GetBaseST : String;
begin
  if FBase = nil then
    result := ''
  else
    result := FBase.value;
end;

Procedure TFhirStructureDefinition.SetBaseST(value : String);
begin
  if value <> '' then
  begin
    if FBase = nil then
      FBase := TFhirUri.create;
    FBase.value := value
  end
  else if FBase <> nil then
    FBase.value := '';
end;

Procedure TFhirStructureDefinition.SetSnapshot(value : TFhirStructureDefinitionSnapshot);
begin
  FSnapshot.free;
  FSnapshot := value;
end;

Procedure TFhirStructureDefinition.SetDifferential(value : TFhirStructureDefinitionDifferential);
begin
  FDifferential.free;
  FDifferential := value;
end;


{ TFhirStructureDefinitionListEnumerator }

Constructor TFhirStructureDefinitionListEnumerator.Create(list : TFhirStructureDefinitionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirStructureDefinitionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirStructureDefinitionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirStructureDefinitionListEnumerator.GetCurrent : TFhirStructureDefinition;
begin
  Result := FList[FIndex];
end;


{ TFhirStructureDefinitionList }
procedure TFhirStructureDefinitionList.AddItem(value: TFhirStructureDefinition);
begin
  assert(value.ClassName = 'TFhirStructureDefinition', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirStructureDefinition');
  add(value);
end;


function TFhirStructureDefinitionList.Append: TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionList.ClearItems;
begin
  Clear;
end;

function TFhirStructureDefinitionList.GetEnumerator : TFhirStructureDefinitionListEnumerator;
begin
  result := TFhirStructureDefinitionListEnumerator.Create(self.link);
end;

function TFhirStructureDefinitionList.Clone: TFhirStructureDefinitionList;
begin
  result := TFhirStructureDefinitionList(inherited Clone);
end;

function TFhirStructureDefinitionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirStructureDefinitionList.GetItemN(index: Integer): TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionList.IndexOf(value: TFhirStructureDefinition): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirStructureDefinitionList.Insert(index: Integer): TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirStructureDefinitionList.InsertItem(index: Integer; value: TFhirStructureDefinition);
begin
  assert(value is TFhirStructureDefinition);
  Inherited Insert(index, value);
end;

function TFhirStructureDefinitionList.Item(index: Integer): TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition(ObjectByIndex[index]);
end;

function TFhirStructureDefinitionList.Link: TFhirStructureDefinitionList;
begin
  result := TFhirStructureDefinitionList(inherited Link);
end;

procedure TFhirStructureDefinitionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirStructureDefinitionList.SetItemByIndex(index: Integer; value: TFhirStructureDefinition);
begin
  assert(value is TFhirStructureDefinition);
  FhirStructureDefinitions[index] := value;
end;

procedure TFhirStructureDefinitionList.SetItemN(index: Integer; value: TFhirStructureDefinition);
begin
  assert(value is TFhirStructureDefinition);
  ObjectByIndex[index] := value;
end;

{ TFhirSubscriptionChannel }

constructor TFhirSubscriptionChannel.Create;
begin
  inherited;
end;

destructor TFhirSubscriptionChannel.Destroy;
begin
  FType_.free;
  FEndpoint.free;
  FPayload.free;
  FHeader.free;
  inherited;
end;

procedure TFhirSubscriptionChannel.Assign(oSource : TAdvObject);
begin
  inherited;
  FType_ := TFhirSubscriptionChannel(oSource).FType_.Link;
  endpointElement := TFhirSubscriptionChannel(oSource).endpointElement.Clone;
  payloadElement := TFhirSubscriptionChannel(oSource).payloadElement.Clone;
  headerElement := TFhirSubscriptionChannel(oSource).headerElement.Clone;
end;

procedure TFhirSubscriptionChannel.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'endpoint') Then
     list.add(FEndpoint.Link);
  if (child_name = 'payload') Then
     list.add(FPayload.Link);
  if (child_name = 'header') Then
     list.add(FHeader.Link);
end;

procedure TFhirSubscriptionChannel.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'code', FType_.Link));{1}
  oList.add(TFHIRProperty.create(self, 'endpoint', 'uri', FEndpoint.Link));{2}
  oList.add(TFHIRProperty.create(self, 'payload', 'string', FPayload.Link));{2}
  oList.add(TFHIRProperty.create(self, 'header', 'string', FHeader.Link));{2}
end;

procedure TFhirSubscriptionChannel.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_Element := propValue as TFHIREnum
  else if (propName = 'endpoint') then EndpointElement := propValue as TFhirUri{5a}
  else if (propName = 'payload') then PayloadElement := propValue as TFhirString{5a}
  else if (propName = 'header') then HeaderElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirSubscriptionChannel.FhirType : string;
begin
  result := 'channel';
end;

function TFhirSubscriptionChannel.Link : TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel(inherited Link);
end;

function TFhirSubscriptionChannel.Clone : TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel(inherited Clone);
end;

function TFhirSubscriptionChannel.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSubscriptionChannel;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSubscriptionChannel)) then
    result := false
  else
  begin
    o := TFhirSubscriptionChannel(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(endpointElement, o.endpointElement, true) and 
      compareDeep(payloadElement, o.payloadElement, true) and compareDeep(headerElement, o.headerElement, true);
  end;
end;

function TFhirSubscriptionChannel.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSubscriptionChannel;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSubscriptionChannel)) then
    result := false
  else
  begin
    o := TFhirSubscriptionChannel(other);
    result := compareValues(type_Element, o.type_Element, true) and compareValues(endpointElement, o.endpointElement, true) and 
      compareValues(payloadElement, o.payloadElement, true) and compareValues(headerElement, o.headerElement, true);
  end;
end;

{ TFhirSubscriptionChannel }

Procedure TFhirSubscriptionChannel.SetType_(value : TFhirEnum);
begin
  FType_.free;
  FType_ := value;
end;

Function TFhirSubscriptionChannel.GetType_ST : TFhirSubscriptionChannelType;
begin
  if FType_ = nil then
    result := TFhirSubscriptionChannelType(0)
  else
    result := TFhirSubscriptionChannelType(StringArrayIndexOfSensitive(CODES_TFhirSubscriptionChannelType, FType_.value));
end;

Procedure TFhirSubscriptionChannel.SetType_ST(value : TFhirSubscriptionChannelType);
begin
  if ord(value) = 0 then
    Type_Element := nil
  else
    Type_Element := TFhirEnum.create(CODES_TFhirSubscriptionChannelType[value]);
end;

Procedure TFhirSubscriptionChannel.SetEndpoint(value : TFhirUri);
begin
  FEndpoint.free;
  FEndpoint := value;
end;

Function TFhirSubscriptionChannel.GetEndpointST : String;
begin
  if FEndpoint = nil then
    result := ''
  else
    result := FEndpoint.value;
end;

Procedure TFhirSubscriptionChannel.SetEndpointST(value : String);
begin
  if value <> '' then
  begin
    if FEndpoint = nil then
      FEndpoint := TFhirUri.create;
    FEndpoint.value := value
  end
  else if FEndpoint <> nil then
    FEndpoint.value := '';
end;

Procedure TFhirSubscriptionChannel.SetPayload(value : TFhirString);
begin
  FPayload.free;
  FPayload := value;
end;

Function TFhirSubscriptionChannel.GetPayloadST : String;
begin
  if FPayload = nil then
    result := ''
  else
    result := FPayload.value;
end;

Procedure TFhirSubscriptionChannel.SetPayloadST(value : String);
begin
  if value <> '' then
  begin
    if FPayload = nil then
      FPayload := TFhirString.create;
    FPayload.value := value
  end
  else if FPayload <> nil then
    FPayload.value := '';
end;

Procedure TFhirSubscriptionChannel.SetHeader(value : TFhirString);
begin
  FHeader.free;
  FHeader := value;
end;

Function TFhirSubscriptionChannel.GetHeaderST : String;
begin
  if FHeader = nil then
    result := ''
  else
    result := FHeader.value;
end;

Procedure TFhirSubscriptionChannel.SetHeaderST(value : String);
begin
  if value <> '' then
  begin
    if FHeader = nil then
      FHeader := TFhirString.create;
    FHeader.value := value
  end
  else if FHeader <> nil then
    FHeader.value := '';
end;


{ TFhirSubscriptionChannelListEnumerator }

Constructor TFhirSubscriptionChannelListEnumerator.Create(list : TFhirSubscriptionChannelList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSubscriptionChannelListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSubscriptionChannelListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSubscriptionChannelListEnumerator.GetCurrent : TFhirSubscriptionChannel;
begin
  Result := FList[FIndex];
end;


{ TFhirSubscriptionChannelList }
procedure TFhirSubscriptionChannelList.AddItem(value: TFhirSubscriptionChannel);
begin
  assert(value.ClassName = 'TFhirSubscriptionChannel', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSubscriptionChannel');
  add(value);
end;


function TFhirSubscriptionChannelList.Append: TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubscriptionChannelList.ClearItems;
begin
  Clear;
end;

function TFhirSubscriptionChannelList.GetEnumerator : TFhirSubscriptionChannelListEnumerator;
begin
  result := TFhirSubscriptionChannelListEnumerator.Create(self.link);
end;

function TFhirSubscriptionChannelList.Clone: TFhirSubscriptionChannelList;
begin
  result := TFhirSubscriptionChannelList(inherited Clone);
end;

function TFhirSubscriptionChannelList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSubscriptionChannelList.GetItemN(index: Integer): TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel(ObjectByIndex[index]);
end;

function TFhirSubscriptionChannelList.IndexOf(value: TFhirSubscriptionChannel): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSubscriptionChannelList.Insert(index: Integer): TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubscriptionChannelList.InsertItem(index: Integer; value: TFhirSubscriptionChannel);
begin
  assert(value is TFhirSubscriptionChannel);
  Inherited Insert(index, value);
end;

function TFhirSubscriptionChannelList.Item(index: Integer): TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel(ObjectByIndex[index]);
end;

function TFhirSubscriptionChannelList.Link: TFhirSubscriptionChannelList;
begin
  result := TFhirSubscriptionChannelList(inherited Link);
end;

procedure TFhirSubscriptionChannelList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSubscriptionChannelList.SetItemByIndex(index: Integer; value: TFhirSubscriptionChannel);
begin
  assert(value is TFhirSubscriptionChannel);
  FhirSubscriptionChannels[index] := value;
end;

procedure TFhirSubscriptionChannelList.SetItemN(index: Integer; value: TFhirSubscriptionChannel);
begin
  assert(value is TFhirSubscriptionChannel);
  ObjectByIndex[index] := value;
end;

{ TFhirSubscription }

constructor TFhirSubscription.Create;
begin
  inherited;
end;

destructor TFhirSubscription.Destroy;
begin
  FCriteria.free;
  FContactList.Free;
  FReason.free;
  FStatus.free;
  FError.free;
  FChannel.free;
  FEnd_.free;
  FTagList.Free;
  inherited;
end;

function TFhirSubscription.GetResourceType : TFhirResourceType;
begin
  result := frtSubscription;
end;

procedure TFhirSubscription.Assign(oSource : TAdvObject);
begin
  inherited;
  criteriaElement := TFhirSubscription(oSource).criteriaElement.Clone;
  if (TFhirSubscription(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirContactPointList.Create;
    FContactList.Assign(TFhirSubscription(oSource).FContactList);
  end;
  reasonElement := TFhirSubscription(oSource).reasonElement.Clone;
  FStatus := TFhirSubscription(oSource).FStatus.Link;
  errorElement := TFhirSubscription(oSource).errorElement.Clone;
  channel := TFhirSubscription(oSource).channel.Clone;
  end_Element := TFhirSubscription(oSource).end_Element.Clone;
  if (TFhirSubscription(oSource).FTagList = nil) then
  begin
    FTagList.free;
    FTagList := nil;
  end
  else
  begin
    FTagList := TFhirCodingList.Create;
    FTagList.Assign(TFhirSubscription(oSource).FTagList);
  end;
end;

procedure TFhirSubscription.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'criteria') Then
     list.add(FCriteria.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'reason') Then
     list.add(FReason.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'error') Then
     list.add(FError.Link);
  if (child_name = 'channel') Then
     list.add(FChannel.Link);
  if (child_name = 'end') Then
     list.add(FEnd_.Link);
  if (child_name = 'tag') Then
    list.addAll(FTagList);
end;

procedure TFhirSubscription.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'criteria', 'string', FCriteria.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', 'ContactPoint', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reason', 'string', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'error', 'string', FError.Link));{2}
  oList.add(TFHIRProperty.create(self, 'channel', '', FChannel.Link));{2}
  oList.add(TFHIRProperty.create(self, 'end', 'instant', FEnd_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'tag', 'Coding', FTagList.Link)){3};
end;

procedure TFhirSubscription.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'criteria') then CriteriaElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirContactPoint){2}
  else if (propName = 'reason') then ReasonElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'error') then ErrorElement := propValue as TFhirString{5a}
  else if (propName = 'channel') then Channel := propValue as TFhirSubscriptionChannel{4b}
  else if (propName = 'end') then End_Element := propValue as TFhirInstant{5a}
  else if (propName = 'tag') then TagList.add(propValue as TFhirCoding){2}
  else inherited;
end;

function TFhirSubscription.FhirType : string;
begin
  result := 'Subscription';
end;

function TFhirSubscription.Link : TFhirSubscription;
begin
  result := TFhirSubscription(inherited Link);
end;

function TFhirSubscription.Clone : TFhirSubscription;
begin
  result := TFhirSubscription(inherited Clone);
end;

{ TFhirSubscription }

Procedure TFhirSubscription.SetCriteria(value : TFhirString);
begin
  FCriteria.free;
  FCriteria := value;
end;

Function TFhirSubscription.GetCriteriaST : String;
begin
  if FCriteria = nil then
    result := ''
  else
    result := FCriteria.value;
end;

Procedure TFhirSubscription.SetCriteriaST(value : String);
begin
  if value <> '' then
  begin
    if FCriteria = nil then
      FCriteria := TFhirString.create;
    FCriteria.value := value
  end
  else if FCriteria <> nil then
    FCriteria.value := '';
end;

Function TFhirSubscription.GetContactList : TFhirContactPointList;
begin
  if FContactList = nil then
    FContactList := TFhirContactPointList.Create;
  result := FContactList;
end;

Function TFhirSubscription.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirSubscription.SetReason(value : TFhirString);
begin
  FReason.free;
  FReason := value;
end;

Function TFhirSubscription.GetReasonST : String;
begin
  if FReason = nil then
    result := ''
  else
    result := FReason.value;
end;

Procedure TFhirSubscription.SetReasonST(value : String);
begin
  if value <> '' then
  begin
    if FReason = nil then
      FReason := TFhirString.create;
    FReason.value := value
  end
  else if FReason <> nil then
    FReason.value := '';
end;

Procedure TFhirSubscription.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirSubscription.GetStatusST : TFhirSubscriptionStatus;
begin
  if FStatus = nil then
    result := TFhirSubscriptionStatus(0)
  else
    result := TFhirSubscriptionStatus(StringArrayIndexOfSensitive(CODES_TFhirSubscriptionStatus, FStatus.value));
end;

Procedure TFhirSubscription.SetStatusST(value : TFhirSubscriptionStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirSubscriptionStatus[value]);
end;

Procedure TFhirSubscription.SetError(value : TFhirString);
begin
  FError.free;
  FError := value;
end;

Function TFhirSubscription.GetErrorST : String;
begin
  if FError = nil then
    result := ''
  else
    result := FError.value;
end;

Procedure TFhirSubscription.SetErrorST(value : String);
begin
  if value <> '' then
  begin
    if FError = nil then
      FError := TFhirString.create;
    FError.value := value
  end
  else if FError <> nil then
    FError.value := '';
end;

Procedure TFhirSubscription.SetChannel(value : TFhirSubscriptionChannel);
begin
  FChannel.free;
  FChannel := value;
end;

Procedure TFhirSubscription.SetEnd_(value : TFhirInstant);
begin
  FEnd_.free;
  FEnd_ := value;
end;

Function TFhirSubscription.GetEnd_ST : TDateAndTime;
begin
  if FEnd_ = nil then
    result := nil
  else
    result := FEnd_.value;
end;

Procedure TFhirSubscription.SetEnd_ST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FEnd_ = nil then
      FEnd_ := TFhirInstant.create;
    FEnd_.value := value
  end
  else if FEnd_ <> nil then
    FEnd_.value := nil;
end;

Function TFhirSubscription.GetTagList : TFhirCodingList;
begin
  if FTagList = nil then
    FTagList := TFhirCodingList.Create;
  result := FTagList;
end;

Function TFhirSubscription.GetHasTagList : boolean;
begin
  result := (FTagList <> nil) and (FTagList.count > 0);
end;


{ TFhirSubscriptionListEnumerator }

Constructor TFhirSubscriptionListEnumerator.Create(list : TFhirSubscriptionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSubscriptionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSubscriptionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSubscriptionListEnumerator.GetCurrent : TFhirSubscription;
begin
  Result := FList[FIndex];
end;


{ TFhirSubscriptionList }
procedure TFhirSubscriptionList.AddItem(value: TFhirSubscription);
begin
  assert(value.ClassName = 'TFhirSubscription', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSubscription');
  add(value);
end;


function TFhirSubscriptionList.Append: TFhirSubscription;
begin
  result := TFhirSubscription.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubscriptionList.ClearItems;
begin
  Clear;
end;

function TFhirSubscriptionList.GetEnumerator : TFhirSubscriptionListEnumerator;
begin
  result := TFhirSubscriptionListEnumerator.Create(self.link);
end;

function TFhirSubscriptionList.Clone: TFhirSubscriptionList;
begin
  result := TFhirSubscriptionList(inherited Clone);
end;

function TFhirSubscriptionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSubscriptionList.GetItemN(index: Integer): TFhirSubscription;
begin
  result := TFhirSubscription(ObjectByIndex[index]);
end;

function TFhirSubscriptionList.IndexOf(value: TFhirSubscription): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSubscriptionList.Insert(index: Integer): TFhirSubscription;
begin
  result := TFhirSubscription.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubscriptionList.InsertItem(index: Integer; value: TFhirSubscription);
begin
  assert(value is TFhirSubscription);
  Inherited Insert(index, value);
end;

function TFhirSubscriptionList.Item(index: Integer): TFhirSubscription;
begin
  result := TFhirSubscription(ObjectByIndex[index]);
end;

function TFhirSubscriptionList.Link: TFhirSubscriptionList;
begin
  result := TFhirSubscriptionList(inherited Link);
end;

procedure TFhirSubscriptionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSubscriptionList.SetItemByIndex(index: Integer; value: TFhirSubscription);
begin
  assert(value is TFhirSubscription);
  FhirSubscriptions[index] := value;
end;

procedure TFhirSubscriptionList.SetItemN(index: Integer; value: TFhirSubscription);
begin
  assert(value is TFhirSubscription);
  ObjectByIndex[index] := value;
end;

{ TFhirSubstanceInstance }

constructor TFhirSubstanceInstance.Create;
begin
  inherited;
end;

destructor TFhirSubstanceInstance.Destroy;
begin
  FIdentifier.free;
  FExpiry.free;
  FQuantity.free;
  inherited;
end;

procedure TFhirSubstanceInstance.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirSubstanceInstance(oSource).identifier.Clone;
  expiryElement := TFhirSubstanceInstance(oSource).expiryElement.Clone;
  quantity := TFhirSubstanceInstance(oSource).quantity.Clone;
end;

procedure TFhirSubstanceInstance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'expiry') Then
     list.add(FExpiry.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
end;

procedure TFhirSubstanceInstance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expiry', 'dateTime', FExpiry.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
end;

procedure TFhirSubstanceInstance.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'expiry') then ExpiryElement := propValue as TFhirDateTime{5a}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else inherited;
end;

function TFhirSubstanceInstance.FhirType : string;
begin
  result := 'instance';
end;

function TFhirSubstanceInstance.Link : TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance(inherited Link);
end;

function TFhirSubstanceInstance.Clone : TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance(inherited Clone);
end;

function TFhirSubstanceInstance.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSubstanceInstance;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSubstanceInstance)) then
    result := false
  else
  begin
    o := TFhirSubstanceInstance(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(expiryElement, o.expiryElement, true) and 
      compareDeep(quantityElement, o.quantityElement, true);
  end;
end;

function TFhirSubstanceInstance.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSubstanceInstance;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSubstanceInstance)) then
    result := false
  else
  begin
    o := TFhirSubstanceInstance(other);
    result := compareValues(expiryElement, o.expiryElement, true);
  end;
end;

{ TFhirSubstanceInstance }

Procedure TFhirSubstanceInstance.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirSubstanceInstance.SetExpiry(value : TFhirDateTime);
begin
  FExpiry.free;
  FExpiry := value;
end;

Function TFhirSubstanceInstance.GetExpiryST : TDateAndTime;
begin
  if FExpiry = nil then
    result := nil
  else
    result := FExpiry.value;
end;

Procedure TFhirSubstanceInstance.SetExpiryST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FExpiry = nil then
      FExpiry := TFhirDateTime.create;
    FExpiry.value := value
  end
  else if FExpiry <> nil then
    FExpiry.value := nil;
end;

Procedure TFhirSubstanceInstance.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;


{ TFhirSubstanceInstanceListEnumerator }

Constructor TFhirSubstanceInstanceListEnumerator.Create(list : TFhirSubstanceInstanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSubstanceInstanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSubstanceInstanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSubstanceInstanceListEnumerator.GetCurrent : TFhirSubstanceInstance;
begin
  Result := FList[FIndex];
end;


{ TFhirSubstanceInstanceList }
procedure TFhirSubstanceInstanceList.AddItem(value: TFhirSubstanceInstance);
begin
  assert(value.ClassName = 'TFhirSubstanceInstance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSubstanceInstance');
  add(value);
end;


function TFhirSubstanceInstanceList.Append: TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceInstanceList.ClearItems;
begin
  Clear;
end;

function TFhirSubstanceInstanceList.GetEnumerator : TFhirSubstanceInstanceListEnumerator;
begin
  result := TFhirSubstanceInstanceListEnumerator.Create(self.link);
end;

function TFhirSubstanceInstanceList.Clone: TFhirSubstanceInstanceList;
begin
  result := TFhirSubstanceInstanceList(inherited Clone);
end;

function TFhirSubstanceInstanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSubstanceInstanceList.GetItemN(index: Integer): TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance(ObjectByIndex[index]);
end;

function TFhirSubstanceInstanceList.IndexOf(value: TFhirSubstanceInstance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSubstanceInstanceList.Insert(index: Integer): TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceInstanceList.InsertItem(index: Integer; value: TFhirSubstanceInstance);
begin
  assert(value is TFhirSubstanceInstance);
  Inherited Insert(index, value);
end;

function TFhirSubstanceInstanceList.Item(index: Integer): TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance(ObjectByIndex[index]);
end;

function TFhirSubstanceInstanceList.Link: TFhirSubstanceInstanceList;
begin
  result := TFhirSubstanceInstanceList(inherited Link);
end;

procedure TFhirSubstanceInstanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSubstanceInstanceList.SetItemByIndex(index: Integer; value: TFhirSubstanceInstance);
begin
  assert(value is TFhirSubstanceInstance);
  FhirSubstanceInstances[index] := value;
end;

procedure TFhirSubstanceInstanceList.SetItemN(index: Integer; value: TFhirSubstanceInstance);
begin
  assert(value is TFhirSubstanceInstance);
  ObjectByIndex[index] := value;
end;

{ TFhirSubstanceIngredient }

constructor TFhirSubstanceIngredient.Create;
begin
  inherited;
end;

destructor TFhirSubstanceIngredient.Destroy;
begin
  FQuantity.free;
  FSubstance.free;
  inherited;
end;

procedure TFhirSubstanceIngredient.Assign(oSource : TAdvObject);
begin
  inherited;
  quantity := TFhirSubstanceIngredient(oSource).quantity.Clone;
  substance := TFhirSubstanceIngredient(oSource).substance.Clone;
end;

procedure TFhirSubstanceIngredient.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'substance') Then
     list.add(FSubstance.Link);
end;

procedure TFhirSubstanceIngredient.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'quantity', 'Ratio', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'substance', 'Reference(Substance)', FSubstance.Link));{2}
end;

procedure TFhirSubstanceIngredient.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'quantity') then Quantity := propValue as TFhirRatio{4b}
  else if (propName = 'substance') then Substance := propValue as TFhirReference{TFhirSubstance}{4b}
  else inherited;
end;

function TFhirSubstanceIngredient.FhirType : string;
begin
  result := 'ingredient';
end;

function TFhirSubstanceIngredient.Link : TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient(inherited Link);
end;

function TFhirSubstanceIngredient.Clone : TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient(inherited Clone);
end;

function TFhirSubstanceIngredient.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSubstanceIngredient;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSubstanceIngredient)) then
    result := false
  else
  begin
    o := TFhirSubstanceIngredient(other);
    result := compareDeep(quantityElement, o.quantityElement, true) and compareDeep(substanceElement, o.substanceElement, true);
  end;
end;

function TFhirSubstanceIngredient.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSubstanceIngredient;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSubstanceIngredient)) then
    result := false
  else
  begin
    o := TFhirSubstanceIngredient(other);
    result := true;
  end;
end;

{ TFhirSubstanceIngredient }

Procedure TFhirSubstanceIngredient.SetQuantity(value : TFhirRatio);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirSubstanceIngredient.SetSubstance(value : TFhirReference{TFhirSubstance});
begin
  FSubstance.free;
  FSubstance := value;
end;


{ TFhirSubstanceIngredientListEnumerator }

Constructor TFhirSubstanceIngredientListEnumerator.Create(list : TFhirSubstanceIngredientList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSubstanceIngredientListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSubstanceIngredientListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSubstanceIngredientListEnumerator.GetCurrent : TFhirSubstanceIngredient;
begin
  Result := FList[FIndex];
end;


{ TFhirSubstanceIngredientList }
procedure TFhirSubstanceIngredientList.AddItem(value: TFhirSubstanceIngredient);
begin
  assert(value.ClassName = 'TFhirSubstanceIngredient', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSubstanceIngredient');
  add(value);
end;


function TFhirSubstanceIngredientList.Append: TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceIngredientList.ClearItems;
begin
  Clear;
end;

function TFhirSubstanceIngredientList.GetEnumerator : TFhirSubstanceIngredientListEnumerator;
begin
  result := TFhirSubstanceIngredientListEnumerator.Create(self.link);
end;

function TFhirSubstanceIngredientList.Clone: TFhirSubstanceIngredientList;
begin
  result := TFhirSubstanceIngredientList(inherited Clone);
end;

function TFhirSubstanceIngredientList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSubstanceIngredientList.GetItemN(index: Integer): TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient(ObjectByIndex[index]);
end;

function TFhirSubstanceIngredientList.IndexOf(value: TFhirSubstanceIngredient): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSubstanceIngredientList.Insert(index: Integer): TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceIngredientList.InsertItem(index: Integer; value: TFhirSubstanceIngredient);
begin
  assert(value is TFhirSubstanceIngredient);
  Inherited Insert(index, value);
end;

function TFhirSubstanceIngredientList.Item(index: Integer): TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient(ObjectByIndex[index]);
end;

function TFhirSubstanceIngredientList.Link: TFhirSubstanceIngredientList;
begin
  result := TFhirSubstanceIngredientList(inherited Link);
end;

procedure TFhirSubstanceIngredientList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSubstanceIngredientList.SetItemByIndex(index: Integer; value: TFhirSubstanceIngredient);
begin
  assert(value is TFhirSubstanceIngredient);
  FhirSubstanceIngredients[index] := value;
end;

procedure TFhirSubstanceIngredientList.SetItemN(index: Integer; value: TFhirSubstanceIngredient);
begin
  assert(value is TFhirSubstanceIngredient);
  ObjectByIndex[index] := value;
end;

{ TFhirSubstance }

constructor TFhirSubstance.Create;
begin
  inherited;
end;

destructor TFhirSubstance.Destroy;
begin
  FIdentifierList.Free;
  FCategoryList.Free;
  FCode.free;
  FDescription.free;
  FInstanceList.Free;
  FIngredientList.Free;
  inherited;
end;

function TFhirSubstance.GetResourceType : TFhirResourceType;
begin
  result := frtSubstance;
end;

procedure TFhirSubstance.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirSubstance(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirSubstance(oSource).FIdentifierList);
  end;
  if (TFhirSubstance(oSource).FCategoryList = nil) then
  begin
    FCategoryList.free;
    FCategoryList := nil;
  end
  else
  begin
    FCategoryList := TFhirCodeableConceptList.Create;
    FCategoryList.Assign(TFhirSubstance(oSource).FCategoryList);
  end;
  code := TFhirSubstance(oSource).code.Clone;
  descriptionElement := TFhirSubstance(oSource).descriptionElement.Clone;
  if (TFhirSubstance(oSource).FInstanceList = nil) then
  begin
    FInstanceList.free;
    FInstanceList := nil;
  end
  else
  begin
    FInstanceList := TFhirSubstanceInstanceList.Create;
    FInstanceList.Assign(TFhirSubstance(oSource).FInstanceList);
  end;
  if (TFhirSubstance(oSource).FIngredientList = nil) then
  begin
    FIngredientList.free;
    FIngredientList := nil;
  end
  else
  begin
    FIngredientList := TFhirSubstanceIngredientList.Create;
    FIngredientList.Assign(TFhirSubstance(oSource).FIngredientList);
  end;
end;

procedure TFhirSubstance.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'category') Then
    list.addAll(FCategoryList);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'instance') Then
    list.addAll(FInstanceList);
  if (child_name = 'ingredient') Then
    list.addAll(FIngredientList);
end;

procedure TFhirSubstance.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'category', 'CodeableConcept', FCategoryList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'instance', '', FInstanceList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'ingredient', '', FIngredientList.Link)){3};
end;

procedure TFhirSubstance.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'category') then CategoryList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'instance') then InstanceList.add(propValue as TFhirSubstanceInstance){2}
  else if (propName = 'ingredient') then IngredientList.add(propValue as TFhirSubstanceIngredient){2}
  else inherited;
end;

function TFhirSubstance.FhirType : string;
begin
  result := 'Substance';
end;

function TFhirSubstance.Link : TFhirSubstance;
begin
  result := TFhirSubstance(inherited Link);
end;

function TFhirSubstance.Clone : TFhirSubstance;
begin
  result := TFhirSubstance(inherited Clone);
end;

{ TFhirSubstance }

Function TFhirSubstance.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirSubstance.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Function TFhirSubstance.GetCategoryList : TFhirCodeableConceptList;
begin
  if FCategoryList = nil then
    FCategoryList := TFhirCodeableConceptList.Create;
  result := FCategoryList;
end;

Function TFhirSubstance.GetHasCategoryList : boolean;
begin
  result := (FCategoryList <> nil) and (FCategoryList.count > 0);
end;

Procedure TFhirSubstance.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirSubstance.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirSubstance.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirSubstance.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirSubstance.GetInstanceList : TFhirSubstanceInstanceList;
begin
  if FInstanceList = nil then
    FInstanceList := TFhirSubstanceInstanceList.Create;
  result := FInstanceList;
end;

Function TFhirSubstance.GetHasInstanceList : boolean;
begin
  result := (FInstanceList <> nil) and (FInstanceList.count > 0);
end;

Function TFhirSubstance.GetIngredientList : TFhirSubstanceIngredientList;
begin
  if FIngredientList = nil then
    FIngredientList := TFhirSubstanceIngredientList.Create;
  result := FIngredientList;
end;

Function TFhirSubstance.GetHasIngredientList : boolean;
begin
  result := (FIngredientList <> nil) and (FIngredientList.count > 0);
end;


{ TFhirSubstanceListEnumerator }

Constructor TFhirSubstanceListEnumerator.Create(list : TFhirSubstanceList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSubstanceListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSubstanceListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSubstanceListEnumerator.GetCurrent : TFhirSubstance;
begin
  Result := FList[FIndex];
end;


{ TFhirSubstanceList }
procedure TFhirSubstanceList.AddItem(value: TFhirSubstance);
begin
  assert(value.ClassName = 'TFhirSubstance', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSubstance');
  add(value);
end;


function TFhirSubstanceList.Append: TFhirSubstance;
begin
  result := TFhirSubstance.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceList.ClearItems;
begin
  Clear;
end;

function TFhirSubstanceList.GetEnumerator : TFhirSubstanceListEnumerator;
begin
  result := TFhirSubstanceListEnumerator.Create(self.link);
end;

function TFhirSubstanceList.Clone: TFhirSubstanceList;
begin
  result := TFhirSubstanceList(inherited Clone);
end;

function TFhirSubstanceList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSubstanceList.GetItemN(index: Integer): TFhirSubstance;
begin
  result := TFhirSubstance(ObjectByIndex[index]);
end;

function TFhirSubstanceList.IndexOf(value: TFhirSubstance): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSubstanceList.Insert(index: Integer): TFhirSubstance;
begin
  result := TFhirSubstance.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSubstanceList.InsertItem(index: Integer; value: TFhirSubstance);
begin
  assert(value is TFhirSubstance);
  Inherited Insert(index, value);
end;

function TFhirSubstanceList.Item(index: Integer): TFhirSubstance;
begin
  result := TFhirSubstance(ObjectByIndex[index]);
end;

function TFhirSubstanceList.Link: TFhirSubstanceList;
begin
  result := TFhirSubstanceList(inherited Link);
end;

procedure TFhirSubstanceList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSubstanceList.SetItemByIndex(index: Integer; value: TFhirSubstance);
begin
  assert(value is TFhirSubstance);
  FhirSubstances[index] := value;
end;

procedure TFhirSubstanceList.SetItemN(index: Integer; value: TFhirSubstance);
begin
  assert(value is TFhirSubstance);
  ObjectByIndex[index] := value;
end;

{ TFhirSupplyDelivery }

constructor TFhirSupplyDelivery.Create;
begin
  inherited;
end;

destructor TFhirSupplyDelivery.Destroy;
begin
  FIdentifier.free;
  FStatus.free;
  FPatient.free;
  FType_.free;
  FQuantity.free;
  FSuppliedItem.free;
  FSupplier.free;
  FWhenPrepared.free;
  FTime.free;
  FDestination.free;
  FReceiverList.Free;
  inherited;
end;

function TFhirSupplyDelivery.GetResourceType : TFhirResourceType;
begin
  result := frtSupplyDelivery;
end;

procedure TFhirSupplyDelivery.Assign(oSource : TAdvObject);
begin
  inherited;
  identifier := TFhirSupplyDelivery(oSource).identifier.Clone;
  FStatus := TFhirSupplyDelivery(oSource).FStatus.Link;
  patient := TFhirSupplyDelivery(oSource).patient.Clone;
  type_ := TFhirSupplyDelivery(oSource).type_.Clone;
  quantity := TFhirSupplyDelivery(oSource).quantity.Clone;
  suppliedItem := TFhirSupplyDelivery(oSource).suppliedItem.Clone;
  supplier := TFhirSupplyDelivery(oSource).supplier.Clone;
  whenPrepared := TFhirSupplyDelivery(oSource).whenPrepared.Clone;
  timeElement := TFhirSupplyDelivery(oSource).timeElement.Clone;
  destination := TFhirSupplyDelivery(oSource).destination.Clone;
  if (TFhirSupplyDelivery(oSource).FReceiverList = nil) then
  begin
    FReceiverList.free;
    FReceiverList := nil;
  end
  else
  begin
    FReceiverList := TFhirReferenceList{TFhirPractitioner}.Create;
    FReceiverList.Assign(TFhirSupplyDelivery(oSource).FReceiverList);
  end;
end;

procedure TFhirSupplyDelivery.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'quantity') Then
     list.add(FQuantity.Link);
  if (child_name = 'suppliedItem') Then
     list.add(FSuppliedItem.Link);
  if (child_name = 'supplier') Then
     list.add(FSupplier.Link);
  if (child_name = 'whenPrepared') Then
     list.add(FWhenPrepared.Link);
  if (child_name = 'time') Then
     list.add(FTime.Link);
  if (child_name = 'destination') Then
     list.add(FDestination.Link);
  if (child_name = 'receiver') Then
    list.addAll(FReceiverList);
end;

procedure TFhirSupplyDelivery.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'type', 'CodeableConcept', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'quantity', 'SimpleQuantity', FQuantity.Link));{2}
  oList.add(TFHIRProperty.create(self, 'suppliedItem', 'Reference(Medication|Substance|Device)', FSuppliedItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'supplier', 'Reference(Practitioner)', FSupplier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'whenPrepared', 'Period', FWhenPrepared.Link));{2}
  oList.add(TFHIRProperty.create(self, 'time', 'dateTime', FTime.Link));{2}
  oList.add(TFHIRProperty.create(self, 'destination', 'Reference(Location)', FDestination.Link));{2}
  oList.add(TFHIRProperty.create(self, 'receiver', 'Reference(Practitioner)', FReceiverList.Link)){3};
end;

procedure TFhirSupplyDelivery.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'type') then Type_ := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'quantity') then Quantity := propValue as TFhirQuantity{4b}
  else if (propName = 'suppliedItem') then SuppliedItem := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'supplier') then Supplier := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'whenPrepared') then WhenPrepared := propValue as TFhirPeriod{4b}
  else if (propName = 'time') then TimeElement := propValue as TFhirDateTime{5a}
  else if (propName = 'destination') then Destination := propValue as TFhirReference{TFhirLocation}{4b}
  else if (propName = 'receiver') then ReceiverList.add(propValue as TFhirReference{TFhirPractitioner}){2}
  else inherited;
end;

function TFhirSupplyDelivery.FhirType : string;
begin
  result := 'SupplyDelivery';
end;

function TFhirSupplyDelivery.Link : TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery(inherited Link);
end;

function TFhirSupplyDelivery.Clone : TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery(inherited Clone);
end;

{ TFhirSupplyDelivery }

Procedure TFhirSupplyDelivery.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirSupplyDelivery.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirSupplyDelivery.GetStatusST : TFhirSupplydeliveryStatus;
begin
  if FStatus = nil then
    result := TFhirSupplydeliveryStatus(0)
  else
    result := TFhirSupplydeliveryStatus(StringArrayIndexOfSensitive(CODES_TFhirSupplydeliveryStatus, FStatus.value));
end;

Procedure TFhirSupplyDelivery.SetStatusST(value : TFhirSupplydeliveryStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirSupplydeliveryStatus[value]);
end;

Procedure TFhirSupplyDelivery.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirSupplyDelivery.SetType_(value : TFhirCodeableConcept);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirSupplyDelivery.SetQuantity(value : TFhirQuantity);
begin
  FQuantity.free;
  FQuantity := value;
end;

Procedure TFhirSupplyDelivery.SetSuppliedItem(value : TFhirReference{Resource});
begin
  FSuppliedItem.free;
  FSuppliedItem := value;
end;

Procedure TFhirSupplyDelivery.SetSupplier(value : TFhirReference{TFhirPractitioner});
begin
  FSupplier.free;
  FSupplier := value;
end;

Procedure TFhirSupplyDelivery.SetWhenPrepared(value : TFhirPeriod);
begin
  FWhenPrepared.free;
  FWhenPrepared := value;
end;

Procedure TFhirSupplyDelivery.SetTime(value : TFhirDateTime);
begin
  FTime.free;
  FTime := value;
end;

Function TFhirSupplyDelivery.GetTimeST : TDateAndTime;
begin
  if FTime = nil then
    result := nil
  else
    result := FTime.value;
end;

Procedure TFhirSupplyDelivery.SetTimeST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FTime = nil then
      FTime := TFhirDateTime.create;
    FTime.value := value
  end
  else if FTime <> nil then
    FTime.value := nil;
end;

Procedure TFhirSupplyDelivery.SetDestination(value : TFhirReference{TFhirLocation});
begin
  FDestination.free;
  FDestination := value;
end;

Function TFhirSupplyDelivery.GetReceiverList : TFhirReferenceList{TFhirPractitioner};
begin
  if FReceiverList = nil then
    FReceiverList := TFhirReferenceList{TFhirPractitioner}.Create;
  result := FReceiverList;
end;

Function TFhirSupplyDelivery.GetHasReceiverList : boolean;
begin
  result := (FReceiverList <> nil) and (FReceiverList.count > 0);
end;


{ TFhirSupplyDeliveryListEnumerator }

Constructor TFhirSupplyDeliveryListEnumerator.Create(list : TFhirSupplyDeliveryList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSupplyDeliveryListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSupplyDeliveryListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSupplyDeliveryListEnumerator.GetCurrent : TFhirSupplyDelivery;
begin
  Result := FList[FIndex];
end;


{ TFhirSupplyDeliveryList }
procedure TFhirSupplyDeliveryList.AddItem(value: TFhirSupplyDelivery);
begin
  assert(value.ClassName = 'TFhirSupplyDelivery', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSupplyDelivery');
  add(value);
end;


function TFhirSupplyDeliveryList.Append: TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyDeliveryList.ClearItems;
begin
  Clear;
end;

function TFhirSupplyDeliveryList.GetEnumerator : TFhirSupplyDeliveryListEnumerator;
begin
  result := TFhirSupplyDeliveryListEnumerator.Create(self.link);
end;

function TFhirSupplyDeliveryList.Clone: TFhirSupplyDeliveryList;
begin
  result := TFhirSupplyDeliveryList(inherited Clone);
end;

function TFhirSupplyDeliveryList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSupplyDeliveryList.GetItemN(index: Integer): TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery(ObjectByIndex[index]);
end;

function TFhirSupplyDeliveryList.IndexOf(value: TFhirSupplyDelivery): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSupplyDeliveryList.Insert(index: Integer): TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyDeliveryList.InsertItem(index: Integer; value: TFhirSupplyDelivery);
begin
  assert(value is TFhirSupplyDelivery);
  Inherited Insert(index, value);
end;

function TFhirSupplyDeliveryList.Item(index: Integer): TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery(ObjectByIndex[index]);
end;

function TFhirSupplyDeliveryList.Link: TFhirSupplyDeliveryList;
begin
  result := TFhirSupplyDeliveryList(inherited Link);
end;

procedure TFhirSupplyDeliveryList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSupplyDeliveryList.SetItemByIndex(index: Integer; value: TFhirSupplyDelivery);
begin
  assert(value is TFhirSupplyDelivery);
  FhirSupplyDeliveries[index] := value;
end;

procedure TFhirSupplyDeliveryList.SetItemN(index: Integer; value: TFhirSupplyDelivery);
begin
  assert(value is TFhirSupplyDelivery);
  ObjectByIndex[index] := value;
end;

{ TFhirSupplyRequestWhen }

constructor TFhirSupplyRequestWhen.Create;
begin
  inherited;
end;

destructor TFhirSupplyRequestWhen.Destroy;
begin
  FCode.free;
  FSchedule.free;
  inherited;
end;

procedure TFhirSupplyRequestWhen.Assign(oSource : TAdvObject);
begin
  inherited;
  code := TFhirSupplyRequestWhen(oSource).code.Clone;
  schedule := TFhirSupplyRequestWhen(oSource).schedule.Clone;
end;

procedure TFhirSupplyRequestWhen.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'schedule') Then
     list.add(FSchedule.Link);
end;

procedure TFhirSupplyRequestWhen.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'CodeableConcept', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'schedule', 'Timing', FSchedule.Link));{2}
end;

procedure TFhirSupplyRequestWhen.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then Code := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'schedule') then Schedule := propValue as TFhirTiming{4b}
  else inherited;
end;

function TFhirSupplyRequestWhen.FhirType : string;
begin
  result := 'when';
end;

function TFhirSupplyRequestWhen.Link : TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen(inherited Link);
end;

function TFhirSupplyRequestWhen.Clone : TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen(inherited Clone);
end;

function TFhirSupplyRequestWhen.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirSupplyRequestWhen;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirSupplyRequestWhen)) then
    result := false
  else
  begin
    o := TFhirSupplyRequestWhen(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(scheduleElement, o.scheduleElement, true);
  end;
end;

function TFhirSupplyRequestWhen.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirSupplyRequestWhen;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirSupplyRequestWhen)) then
    result := false
  else
  begin
    o := TFhirSupplyRequestWhen(other);
    result := true;
  end;
end;

{ TFhirSupplyRequestWhen }

Procedure TFhirSupplyRequestWhen.SetCode(value : TFhirCodeableConcept);
begin
  FCode.free;
  FCode := value;
end;

Procedure TFhirSupplyRequestWhen.SetSchedule(value : TFhirTiming);
begin
  FSchedule.free;
  FSchedule := value;
end;


{ TFhirSupplyRequestWhenListEnumerator }

Constructor TFhirSupplyRequestWhenListEnumerator.Create(list : TFhirSupplyRequestWhenList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSupplyRequestWhenListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSupplyRequestWhenListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSupplyRequestWhenListEnumerator.GetCurrent : TFhirSupplyRequestWhen;
begin
  Result := FList[FIndex];
end;


{ TFhirSupplyRequestWhenList }
procedure TFhirSupplyRequestWhenList.AddItem(value: TFhirSupplyRequestWhen);
begin
  assert(value.ClassName = 'TFhirSupplyRequestWhen', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSupplyRequestWhen');
  add(value);
end;


function TFhirSupplyRequestWhenList.Append: TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyRequestWhenList.ClearItems;
begin
  Clear;
end;

function TFhirSupplyRequestWhenList.GetEnumerator : TFhirSupplyRequestWhenListEnumerator;
begin
  result := TFhirSupplyRequestWhenListEnumerator.Create(self.link);
end;

function TFhirSupplyRequestWhenList.Clone: TFhirSupplyRequestWhenList;
begin
  result := TFhirSupplyRequestWhenList(inherited Clone);
end;

function TFhirSupplyRequestWhenList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSupplyRequestWhenList.GetItemN(index: Integer): TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen(ObjectByIndex[index]);
end;

function TFhirSupplyRequestWhenList.IndexOf(value: TFhirSupplyRequestWhen): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSupplyRequestWhenList.Insert(index: Integer): TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyRequestWhenList.InsertItem(index: Integer; value: TFhirSupplyRequestWhen);
begin
  assert(value is TFhirSupplyRequestWhen);
  Inherited Insert(index, value);
end;

function TFhirSupplyRequestWhenList.Item(index: Integer): TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen(ObjectByIndex[index]);
end;

function TFhirSupplyRequestWhenList.Link: TFhirSupplyRequestWhenList;
begin
  result := TFhirSupplyRequestWhenList(inherited Link);
end;

procedure TFhirSupplyRequestWhenList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSupplyRequestWhenList.SetItemByIndex(index: Integer; value: TFhirSupplyRequestWhen);
begin
  assert(value is TFhirSupplyRequestWhen);
  FhirSupplyRequestWhens[index] := value;
end;

procedure TFhirSupplyRequestWhenList.SetItemN(index: Integer; value: TFhirSupplyRequestWhen);
begin
  assert(value is TFhirSupplyRequestWhen);
  ObjectByIndex[index] := value;
end;

{ TFhirSupplyRequest }

constructor TFhirSupplyRequest.Create;
begin
  inherited;
end;

destructor TFhirSupplyRequest.Destroy;
begin
  FPatient.free;
  FSource.free;
  FDate.free;
  FIdentifier.free;
  FStatus.free;
  FKind.free;
  FOrderedItem.free;
  FSupplierList.Free;
  FReason.free;
  FWhen.free;
  inherited;
end;

function TFhirSupplyRequest.GetResourceType : TFhirResourceType;
begin
  result := frtSupplyRequest;
end;

procedure TFhirSupplyRequest.Assign(oSource : TAdvObject);
begin
  inherited;
  patient := TFhirSupplyRequest(oSource).patient.Clone;
  source := TFhirSupplyRequest(oSource).source.Clone;
  dateElement := TFhirSupplyRequest(oSource).dateElement.Clone;
  identifier := TFhirSupplyRequest(oSource).identifier.Clone;
  FStatus := TFhirSupplyRequest(oSource).FStatus.Link;
  kind := TFhirSupplyRequest(oSource).kind.Clone;
  orderedItem := TFhirSupplyRequest(oSource).orderedItem.Clone;
  if (TFhirSupplyRequest(oSource).FSupplierList = nil) then
  begin
    FSupplierList.free;
    FSupplierList := nil;
  end
  else
  begin
    FSupplierList := TFhirReferenceList{TFhirOrganization}.Create;
    FSupplierList.Assign(TFhirSupplyRequest(oSource).FSupplierList);
  end;
  reason := TFhirSupplyRequest(oSource).reason.Clone;
  when := TFhirSupplyRequest(oSource).when.Clone;
end;

procedure TFhirSupplyRequest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'source') Then
     list.add(FSource.Link);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'kind') Then
     list.add(FKind.Link);
  if (child_name = 'orderedItem') Then
     list.add(FOrderedItem.Link);
  if (child_name = 'supplier') Then
    list.addAll(FSupplierList);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'when') Then
     list.add(FWhen.Link);
end;

procedure TFhirSupplyRequest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'source', 'Reference(Practitioner|Organization|Patient)', FSource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'kind', 'CodeableConcept', FKind.Link));{2}
  oList.add(TFHIRProperty.create(self, 'orderedItem', 'Reference(Medication|Substance|Device)', FOrderedItem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'supplier', 'Reference(Organization)', FSupplierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Any)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'when', '', FWhen.Link));{2}
end;

procedure TFhirSupplyRequest.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'source') then Source := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'kind') then Kind := propValue as TFhirCodeableConcept{4b}
  else if (propName = 'orderedItem') then OrderedItem := propValue as TFhirReference{Resource}{4b}
  else if (propName = 'supplier') then SupplierList.add(propValue as TFhirReference{TFhirOrganization}){2}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName = 'when') then When := propValue as TFhirSupplyRequestWhen{4b}
  else inherited;
end;

function TFhirSupplyRequest.FhirType : string;
begin
  result := 'SupplyRequest';
end;

function TFhirSupplyRequest.Link : TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest(inherited Link);
end;

function TFhirSupplyRequest.Clone : TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest(inherited Clone);
end;

{ TFhirSupplyRequest }

Procedure TFhirSupplyRequest.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirSupplyRequest.SetSource(value : TFhirReference{Resource});
begin
  FSource.free;
  FSource := value;
end;

Procedure TFhirSupplyRequest.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirSupplyRequest.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirSupplyRequest.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirSupplyRequest.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirSupplyRequest.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirSupplyRequest.GetStatusST : TFhirSupplyrequestStatus;
begin
  if FStatus = nil then
    result := TFhirSupplyrequestStatus(0)
  else
    result := TFhirSupplyrequestStatus(StringArrayIndexOfSensitive(CODES_TFhirSupplyrequestStatus, FStatus.value));
end;

Procedure TFhirSupplyRequest.SetStatusST(value : TFhirSupplyrequestStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirSupplyrequestStatus[value]);
end;

Procedure TFhirSupplyRequest.SetKind(value : TFhirCodeableConcept);
begin
  FKind.free;
  FKind := value;
end;

Procedure TFhirSupplyRequest.SetOrderedItem(value : TFhirReference{Resource});
begin
  FOrderedItem.free;
  FOrderedItem := value;
end;

Function TFhirSupplyRequest.GetSupplierList : TFhirReferenceList{TFhirOrganization};
begin
  if FSupplierList = nil then
    FSupplierList := TFhirReferenceList{TFhirOrganization}.Create;
  result := FSupplierList;
end;

Function TFhirSupplyRequest.GetHasSupplierList : boolean;
begin
  result := (FSupplierList <> nil) and (FSupplierList.count > 0);
end;

Procedure TFhirSupplyRequest.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Procedure TFhirSupplyRequest.SetWhen(value : TFhirSupplyRequestWhen);
begin
  FWhen.free;
  FWhen := value;
end;


{ TFhirSupplyRequestListEnumerator }

Constructor TFhirSupplyRequestListEnumerator.Create(list : TFhirSupplyRequestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirSupplyRequestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirSupplyRequestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirSupplyRequestListEnumerator.GetCurrent : TFhirSupplyRequest;
begin
  Result := FList[FIndex];
end;


{ TFhirSupplyRequestList }
procedure TFhirSupplyRequestList.AddItem(value: TFhirSupplyRequest);
begin
  assert(value.ClassName = 'TFhirSupplyRequest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirSupplyRequest');
  add(value);
end;


function TFhirSupplyRequestList.Append: TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyRequestList.ClearItems;
begin
  Clear;
end;

function TFhirSupplyRequestList.GetEnumerator : TFhirSupplyRequestListEnumerator;
begin
  result := TFhirSupplyRequestListEnumerator.Create(self.link);
end;

function TFhirSupplyRequestList.Clone: TFhirSupplyRequestList;
begin
  result := TFhirSupplyRequestList(inherited Clone);
end;

function TFhirSupplyRequestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirSupplyRequestList.GetItemN(index: Integer): TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest(ObjectByIndex[index]);
end;

function TFhirSupplyRequestList.IndexOf(value: TFhirSupplyRequest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirSupplyRequestList.Insert(index: Integer): TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirSupplyRequestList.InsertItem(index: Integer; value: TFhirSupplyRequest);
begin
  assert(value is TFhirSupplyRequest);
  Inherited Insert(index, value);
end;

function TFhirSupplyRequestList.Item(index: Integer): TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest(ObjectByIndex[index]);
end;

function TFhirSupplyRequestList.Link: TFhirSupplyRequestList;
begin
  result := TFhirSupplyRequestList(inherited Link);
end;

procedure TFhirSupplyRequestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirSupplyRequestList.SetItemByIndex(index: Integer; value: TFhirSupplyRequest);
begin
  assert(value is TFhirSupplyRequest);
  FhirSupplyRequests[index] := value;
end;

procedure TFhirSupplyRequestList.SetItemN(index: Integer; value: TFhirSupplyRequest);
begin
  assert(value is TFhirSupplyRequest);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptContact }

constructor TFhirTestScriptContact.Create;
begin
  inherited;
end;

destructor TFhirTestScriptContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirTestScriptContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirTestScriptContact(oSource).nameElement.Clone;
  if (TFhirTestScriptContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirTestScriptContact(oSource).FTelecomList);
  end;
end;

procedure TFhirTestScriptContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirTestScriptContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirTestScriptContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirTestScriptContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirTestScriptContact.Link : TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact(inherited Link);
end;

function TFhirTestScriptContact.Clone : TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact(inherited Clone);
end;

function TFhirTestScriptContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptContact)) then
    result := false
  else
  begin
    o := TFhirTestScriptContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirTestScriptContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptContact)) then
    result := false
  else
  begin
    o := TFhirTestScriptContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirTestScriptContact }

Procedure TFhirTestScriptContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirTestScriptContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirTestScriptContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirTestScriptContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirTestScriptContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirTestScriptContactListEnumerator }

Constructor TFhirTestScriptContactListEnumerator.Create(list : TFhirTestScriptContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptContactListEnumerator.GetCurrent : TFhirTestScriptContact;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptContactList }
procedure TFhirTestScriptContactList.AddItem(value: TFhirTestScriptContact);
begin
  assert(value.ClassName = 'TFhirTestScriptContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptContact');
  add(value);
end;


function TFhirTestScriptContactList.Append: TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptContactList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptContactList.GetEnumerator : TFhirTestScriptContactListEnumerator;
begin
  result := TFhirTestScriptContactListEnumerator.Create(self.link);
end;

function TFhirTestScriptContactList.Clone: TFhirTestScriptContactList;
begin
  result := TFhirTestScriptContactList(inherited Clone);
end;

function TFhirTestScriptContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptContactList.GetItemN(index: Integer): TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact(ObjectByIndex[index]);
end;

function TFhirTestScriptContactList.IndexOf(value: TFhirTestScriptContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptContactList.Insert(index: Integer): TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptContactList.InsertItem(index: Integer; value: TFhirTestScriptContact);
begin
  assert(value is TFhirTestScriptContact);
  Inherited Insert(index, value);
end;

function TFhirTestScriptContactList.Item(index: Integer): TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact(ObjectByIndex[index]);
end;

function TFhirTestScriptContactList.Link: TFhirTestScriptContactList;
begin
  result := TFhirTestScriptContactList(inherited Link);
end;

procedure TFhirTestScriptContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptContactList.SetItemByIndex(index: Integer; value: TFhirTestScriptContact);
begin
  assert(value is TFhirTestScriptContact);
  FhirTestScriptContacts[index] := value;
end;

procedure TFhirTestScriptContactList.SetItemN(index: Integer; value: TFhirTestScriptContact);
begin
  assert(value is TFhirTestScriptContact);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptMetadata }

constructor TFhirTestScriptMetadata.Create;
begin
  inherited;
end;

destructor TFhirTestScriptMetadata.Destroy;
begin
  FLink_List.Free;
  FCapabilityList.Free;
  inherited;
end;

procedure TFhirTestScriptMetadata.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirTestScriptMetadata(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirTestScriptMetadataLinkList.Create;
    FLink_List.Assign(TFhirTestScriptMetadata(oSource).FLink_List);
  end;
  if (TFhirTestScriptMetadata(oSource).FCapabilityList = nil) then
  begin
    FCapabilityList.free;
    FCapabilityList := nil;
  end
  else
  begin
    FCapabilityList := TFhirTestScriptMetadataCapabilityList.Create;
    FCapabilityList.Assign(TFhirTestScriptMetadata(oSource).FCapabilityList);
  end;
end;

procedure TFhirTestScriptMetadata.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'link') Then
    list.addAll(FLink_List);
  if (child_name = 'capability') Then
    list.addAll(FCapabilityList);
end;

procedure TFhirTestScriptMetadata.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'link', '', FLink_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'capability', '', FCapabilityList.Link)){3};
end;

procedure TFhirTestScriptMetadata.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'link') then Link_List.add(propValue as TFhirTestScriptMetadataLink){2}
  else if (propName = 'capability') then CapabilityList.add(propValue as TFhirTestScriptMetadataCapability){2}
  else inherited;
end;

function TFhirTestScriptMetadata.FhirType : string;
begin
  result := 'metadata';
end;

function TFhirTestScriptMetadata.Link : TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata(inherited Link);
end;

function TFhirTestScriptMetadata.Clone : TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata(inherited Clone);
end;

function TFhirTestScriptMetadata.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadata;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadata)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadata(other);
    result := compareDeep(link_List, o.link_List, true) and compareDeep(capabilityList, o.capabilityList, true);
  end;
end;

function TFhirTestScriptMetadata.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadata;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadata)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadata(other);
    result := true;
  end;
end;

{ TFhirTestScriptMetadata }

Function TFhirTestScriptMetadata.GetLink_List : TFhirTestScriptMetadataLinkList;
begin
  if FLink_List = nil then
    FLink_List := TFhirTestScriptMetadataLinkList.Create;
  result := FLink_List;
end;

Function TFhirTestScriptMetadata.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;

Function TFhirTestScriptMetadata.GetCapabilityList : TFhirTestScriptMetadataCapabilityList;
begin
  if FCapabilityList = nil then
    FCapabilityList := TFhirTestScriptMetadataCapabilityList.Create;
  result := FCapabilityList;
end;

Function TFhirTestScriptMetadata.GetHasCapabilityList : boolean;
begin
  result := (FCapabilityList <> nil) and (FCapabilityList.count > 0);
end;


{ TFhirTestScriptMetadataListEnumerator }

Constructor TFhirTestScriptMetadataListEnumerator.Create(list : TFhirTestScriptMetadataList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptMetadataListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptMetadataListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptMetadataListEnumerator.GetCurrent : TFhirTestScriptMetadata;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptMetadataList }
procedure TFhirTestScriptMetadataList.AddItem(value: TFhirTestScriptMetadata);
begin
  assert(value.ClassName = 'TFhirTestScriptMetadata', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptMetadata');
  add(value);
end;


function TFhirTestScriptMetadataList.Append: TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptMetadataList.GetEnumerator : TFhirTestScriptMetadataListEnumerator;
begin
  result := TFhirTestScriptMetadataListEnumerator.Create(self.link);
end;

function TFhirTestScriptMetadataList.Clone: TFhirTestScriptMetadataList;
begin
  result := TFhirTestScriptMetadataList(inherited Clone);
end;

function TFhirTestScriptMetadataList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptMetadataList.GetItemN(index: Integer): TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataList.IndexOf(value: TFhirTestScriptMetadata): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptMetadataList.Insert(index: Integer): TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataList.InsertItem(index: Integer; value: TFhirTestScriptMetadata);
begin
  assert(value is TFhirTestScriptMetadata);
  Inherited Insert(index, value);
end;

function TFhirTestScriptMetadataList.Item(index: Integer): TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataList.Link: TFhirTestScriptMetadataList;
begin
  result := TFhirTestScriptMetadataList(inherited Link);
end;

procedure TFhirTestScriptMetadataList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptMetadataList.SetItemByIndex(index: Integer; value: TFhirTestScriptMetadata);
begin
  assert(value is TFhirTestScriptMetadata);
  FhirTestScriptMetadata[index] := value;
end;

procedure TFhirTestScriptMetadataList.SetItemN(index: Integer; value: TFhirTestScriptMetadata);
begin
  assert(value is TFhirTestScriptMetadata);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptMetadataLink }

constructor TFhirTestScriptMetadataLink.Create;
begin
  inherited;
end;

destructor TFhirTestScriptMetadataLink.Destroy;
begin
  FUrl.free;
  FDescription.free;
  inherited;
end;

procedure TFhirTestScriptMetadataLink.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirTestScriptMetadataLink(oSource).urlElement.Clone;
  descriptionElement := TFhirTestScriptMetadataLink(oSource).descriptionElement.Clone;
end;

procedure TFhirTestScriptMetadataLink.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
end;

procedure TFhirTestScriptMetadataLink.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
end;

procedure TFhirTestScriptMetadataLink.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirTestScriptMetadataLink.FhirType : string;
begin
  result := 'link';
end;

function TFhirTestScriptMetadataLink.Link : TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink(inherited Link);
end;

function TFhirTestScriptMetadataLink.Clone : TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink(inherited Clone);
end;

function TFhirTestScriptMetadataLink.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadataLink;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadataLink)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadataLink(other);
    result := compareDeep(urlElement, o.urlElement, true) and compareDeep(descriptionElement, o.descriptionElement, true);
  end;
end;

function TFhirTestScriptMetadataLink.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadataLink;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadataLink)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadataLink(other);
    result := compareValues(urlElement, o.urlElement, true) and compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirTestScriptMetadataLink }

Procedure TFhirTestScriptMetadataLink.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirTestScriptMetadataLink.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirTestScriptMetadataLink.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirTestScriptMetadataLink.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScriptMetadataLink.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScriptMetadataLink.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;


{ TFhirTestScriptMetadataLinkListEnumerator }

Constructor TFhirTestScriptMetadataLinkListEnumerator.Create(list : TFhirTestScriptMetadataLinkList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptMetadataLinkListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptMetadataLinkListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptMetadataLinkListEnumerator.GetCurrent : TFhirTestScriptMetadataLink;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptMetadataLinkList }
procedure TFhirTestScriptMetadataLinkList.AddItem(value: TFhirTestScriptMetadataLink);
begin
  assert(value.ClassName = 'TFhirTestScriptMetadataLink', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptMetadataLink');
  add(value);
end;


function TFhirTestScriptMetadataLinkList.Append: TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataLinkList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptMetadataLinkList.GetEnumerator : TFhirTestScriptMetadataLinkListEnumerator;
begin
  result := TFhirTestScriptMetadataLinkListEnumerator.Create(self.link);
end;

function TFhirTestScriptMetadataLinkList.Clone: TFhirTestScriptMetadataLinkList;
begin
  result := TFhirTestScriptMetadataLinkList(inherited Clone);
end;

function TFhirTestScriptMetadataLinkList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptMetadataLinkList.GetItemN(index: Integer): TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataLinkList.IndexOf(value: TFhirTestScriptMetadataLink): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptMetadataLinkList.Insert(index: Integer): TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataLinkList.InsertItem(index: Integer; value: TFhirTestScriptMetadataLink);
begin
  assert(value is TFhirTestScriptMetadataLink);
  Inherited Insert(index, value);
end;

function TFhirTestScriptMetadataLinkList.Item(index: Integer): TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataLinkList.Link: TFhirTestScriptMetadataLinkList;
begin
  result := TFhirTestScriptMetadataLinkList(inherited Link);
end;

procedure TFhirTestScriptMetadataLinkList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptMetadataLinkList.SetItemByIndex(index: Integer; value: TFhirTestScriptMetadataLink);
begin
  assert(value is TFhirTestScriptMetadataLink);
  FhirTestScriptMetadataLinks[index] := value;
end;

procedure TFhirTestScriptMetadataLinkList.SetItemN(index: Integer; value: TFhirTestScriptMetadataLink);
begin
  assert(value is TFhirTestScriptMetadataLink);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptMetadataCapability }

constructor TFhirTestScriptMetadataCapability.Create;
begin
  inherited;
end;

destructor TFhirTestScriptMetadataCapability.Destroy;
begin
  FRequired.free;
  FValidated.free;
  FDescription.free;
  FDestination.free;
  FLink_List.Free;
  FConformance.free;
  inherited;
end;

procedure TFhirTestScriptMetadataCapability.Assign(oSource : TAdvObject);
begin
  inherited;
  requiredElement := TFhirTestScriptMetadataCapability(oSource).requiredElement.Clone;
  validatedElement := TFhirTestScriptMetadataCapability(oSource).validatedElement.Clone;
  descriptionElement := TFhirTestScriptMetadataCapability(oSource).descriptionElement.Clone;
  destinationElement := TFhirTestScriptMetadataCapability(oSource).destinationElement.Clone;
  if (TFhirTestScriptMetadataCapability(oSource).FLink_List = nil) then
  begin
    FLink_List.free;
    FLink_List := nil;
  end
  else
  begin
    FLink_List := TFhirUriList.Create;
    FLink_List.Assign(TFhirTestScriptMetadataCapability(oSource).FLink_List);
  end;
  conformance := TFhirTestScriptMetadataCapability(oSource).conformance.Clone;
end;

procedure TFhirTestScriptMetadataCapability.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'required') Then
     list.add(FRequired.Link);
  if (child_name = 'validated') Then
     list.add(FValidated.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'destination') Then
     list.add(FDestination.Link);
  if (child_name = 'link') Then
    list.addAll(FLink_List);
  if (child_name = 'conformance') Then
     list.add(FConformance.Link);
end;

procedure TFhirTestScriptMetadataCapability.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'required', 'boolean', FRequired.Link));{2}
  oList.add(TFHIRProperty.create(self, 'validated', 'boolean', FValidated.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'destination', 'integer', FDestination.Link));{2}
  oList.add(TFHIRProperty.create(self, 'link', 'uri', FLink_List.Link)){3};
  oList.add(TFHIRProperty.create(self, 'conformance', 'Reference(Conformance)', FConformance.Link));{2}
end;

procedure TFhirTestScriptMetadataCapability.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'required') then RequiredElement := propValue as TFhirBoolean{5a}
  else if (propName = 'validated') then ValidatedElement := propValue as TFhirBoolean{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'destination') then DestinationElement := propValue as TFhirInteger{5a}
  else if (propName = 'link') then Link_List.add(propValue as TFhirUri){2}
  else if (propName = 'conformance') then Conformance := propValue as TFhirReference{TFhirConformance}{4b}
  else inherited;
end;

function TFhirTestScriptMetadataCapability.FhirType : string;
begin
  result := 'capability';
end;

function TFhirTestScriptMetadataCapability.Link : TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability(inherited Link);
end;

function TFhirTestScriptMetadataCapability.Clone : TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability(inherited Clone);
end;

function TFhirTestScriptMetadataCapability.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadataCapability;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadataCapability)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadataCapability(other);
    result := compareDeep(requiredElement, o.requiredElement, true) and compareDeep(validatedElement, o.validatedElement, true) and 
      compareDeep(descriptionElement, o.descriptionElement, true) and compareDeep(destinationElement, o.destinationElement, true) and 
      compareDeep(link_List, o.link_List, true) and compareDeep(conformanceElement, o.conformanceElement, true);
  end;
end;

function TFhirTestScriptMetadataCapability.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptMetadataCapability;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptMetadataCapability)) then
    result := false
  else
  begin
    o := TFhirTestScriptMetadataCapability(other);
    result := compareValues(requiredElement, o.requiredElement, true) and compareValues(validatedElement, o.validatedElement, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(destinationElement, o.destinationElement, true) and 
      compareValues(link_List, o.link_List, true);
  end;
end;

{ TFhirTestScriptMetadataCapability }

Procedure TFhirTestScriptMetadataCapability.SetRequired(value : TFhirBoolean);
begin
  FRequired.free;
  FRequired := value;
end;

Function TFhirTestScriptMetadataCapability.GetRequiredST : Boolean;
begin
  if FRequired = nil then
    result := false
  else
    result := FRequired.value;
end;

Procedure TFhirTestScriptMetadataCapability.SetRequiredST(value : Boolean);
begin
  if FRequired = nil then
    FRequired := TFhirBoolean.create;
  FRequired.value := value
end;

Procedure TFhirTestScriptMetadataCapability.SetValidated(value : TFhirBoolean);
begin
  FValidated.free;
  FValidated := value;
end;

Function TFhirTestScriptMetadataCapability.GetValidatedST : Boolean;
begin
  if FValidated = nil then
    result := false
  else
    result := FValidated.value;
end;

Procedure TFhirTestScriptMetadataCapability.SetValidatedST(value : Boolean);
begin
  if FValidated = nil then
    FValidated := TFhirBoolean.create;
  FValidated.value := value
end;

Procedure TFhirTestScriptMetadataCapability.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScriptMetadataCapability.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScriptMetadataCapability.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirTestScriptMetadataCapability.SetDestination(value : TFhirInteger);
begin
  FDestination.free;
  FDestination := value;
end;

Function TFhirTestScriptMetadataCapability.GetDestinationST : String;
begin
  if FDestination = nil then
    result := ''
  else
    result := FDestination.value;
end;

Procedure TFhirTestScriptMetadataCapability.SetDestinationST(value : String);
begin
  if value <> '' then
  begin
    if FDestination = nil then
      FDestination := TFhirInteger.create;
    FDestination.value := value
  end
  else if FDestination <> nil then
    FDestination.value := '';
end;

Function TFhirTestScriptMetadataCapability.GetLink_List : TFhirUriList;
begin
  if FLink_List = nil then
    FLink_List := TFhirUriList.Create;
  result := FLink_List;
end;

Function TFhirTestScriptMetadataCapability.GetHasLink_List : boolean;
begin
  result := (FLink_List <> nil) and (FLink_List.count > 0);
end;

Procedure TFhirTestScriptMetadataCapability.SetConformance(value : TFhirReference{TFhirConformance});
begin
  FConformance.free;
  FConformance := value;
end;


{ TFhirTestScriptMetadataCapabilityListEnumerator }

Constructor TFhirTestScriptMetadataCapabilityListEnumerator.Create(list : TFhirTestScriptMetadataCapabilityList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptMetadataCapabilityListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptMetadataCapabilityListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptMetadataCapabilityListEnumerator.GetCurrent : TFhirTestScriptMetadataCapability;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptMetadataCapabilityList }
procedure TFhirTestScriptMetadataCapabilityList.AddItem(value: TFhirTestScriptMetadataCapability);
begin
  assert(value.ClassName = 'TFhirTestScriptMetadataCapability', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptMetadataCapability');
  add(value);
end;


function TFhirTestScriptMetadataCapabilityList.Append: TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataCapabilityList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptMetadataCapabilityList.GetEnumerator : TFhirTestScriptMetadataCapabilityListEnumerator;
begin
  result := TFhirTestScriptMetadataCapabilityListEnumerator.Create(self.link);
end;

function TFhirTestScriptMetadataCapabilityList.Clone: TFhirTestScriptMetadataCapabilityList;
begin
  result := TFhirTestScriptMetadataCapabilityList(inherited Clone);
end;

function TFhirTestScriptMetadataCapabilityList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptMetadataCapabilityList.GetItemN(index: Integer): TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataCapabilityList.IndexOf(value: TFhirTestScriptMetadataCapability): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptMetadataCapabilityList.Insert(index: Integer): TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptMetadataCapabilityList.InsertItem(index: Integer; value: TFhirTestScriptMetadataCapability);
begin
  assert(value is TFhirTestScriptMetadataCapability);
  Inherited Insert(index, value);
end;

function TFhirTestScriptMetadataCapabilityList.Item(index: Integer): TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability(ObjectByIndex[index]);
end;

function TFhirTestScriptMetadataCapabilityList.Link: TFhirTestScriptMetadataCapabilityList;
begin
  result := TFhirTestScriptMetadataCapabilityList(inherited Link);
end;

procedure TFhirTestScriptMetadataCapabilityList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptMetadataCapabilityList.SetItemByIndex(index: Integer; value: TFhirTestScriptMetadataCapability);
begin
  assert(value is TFhirTestScriptMetadataCapability);
  FhirTestScriptMetadataCapabilities[index] := value;
end;

procedure TFhirTestScriptMetadataCapabilityList.SetItemN(index: Integer; value: TFhirTestScriptMetadataCapability);
begin
  assert(value is TFhirTestScriptMetadataCapability);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptFixture }

constructor TFhirTestScriptFixture.Create;
begin
  inherited;
end;

destructor TFhirTestScriptFixture.Destroy;
begin
  FAutocreate.free;
  FAutodelete.free;
  FResource.free;
  inherited;
end;

procedure TFhirTestScriptFixture.Assign(oSource : TAdvObject);
begin
  inherited;
  autocreateElement := TFhirTestScriptFixture(oSource).autocreateElement.Clone;
  autodeleteElement := TFhirTestScriptFixture(oSource).autodeleteElement.Clone;
  resource := TFhirTestScriptFixture(oSource).resource.Clone;
end;

procedure TFhirTestScriptFixture.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'autocreate') Then
     list.add(FAutocreate.Link);
  if (child_name = 'autodelete') Then
     list.add(FAutodelete.Link);
  if (child_name = 'resource') Then
     list.add(FResource.Link);
end;

procedure TFhirTestScriptFixture.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'autocreate', 'boolean', FAutocreate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'autodelete', 'boolean', FAutodelete.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', 'Reference(Any)', FResource.Link));{2}
end;

procedure TFhirTestScriptFixture.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'autocreate') then AutocreateElement := propValue as TFhirBoolean{5a}
  else if (propName = 'autodelete') then AutodeleteElement := propValue as TFhirBoolean{5a}
  else if (propName = 'resource') then Resource := propValue as TFhirReference{TFhirReference}{4b}
  else inherited;
end;

function TFhirTestScriptFixture.FhirType : string;
begin
  result := 'fixture';
end;

function TFhirTestScriptFixture.Link : TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture(inherited Link);
end;

function TFhirTestScriptFixture.Clone : TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture(inherited Clone);
end;

function TFhirTestScriptFixture.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptFixture;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptFixture)) then
    result := false
  else
  begin
    o := TFhirTestScriptFixture(other);
    result := compareDeep(autocreateElement, o.autocreateElement, true) and compareDeep(autodeleteElement, o.autodeleteElement, true) and 
      compareDeep(resourceElement, o.resourceElement, true);
  end;
end;

function TFhirTestScriptFixture.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptFixture;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptFixture)) then
    result := false
  else
  begin
    o := TFhirTestScriptFixture(other);
    result := compareValues(autocreateElement, o.autocreateElement, true) and compareValues(autodeleteElement, o.autodeleteElement, true);
  end;
end;

{ TFhirTestScriptFixture }

Procedure TFhirTestScriptFixture.SetAutocreate(value : TFhirBoolean);
begin
  FAutocreate.free;
  FAutocreate := value;
end;

Function TFhirTestScriptFixture.GetAutocreateST : Boolean;
begin
  if FAutocreate = nil then
    result := false
  else
    result := FAutocreate.value;
end;

Procedure TFhirTestScriptFixture.SetAutocreateST(value : Boolean);
begin
  if FAutocreate = nil then
    FAutocreate := TFhirBoolean.create;
  FAutocreate.value := value
end;

Procedure TFhirTestScriptFixture.SetAutodelete(value : TFhirBoolean);
begin
  FAutodelete.free;
  FAutodelete := value;
end;

Function TFhirTestScriptFixture.GetAutodeleteST : Boolean;
begin
  if FAutodelete = nil then
    result := false
  else
    result := FAutodelete.value;
end;

Procedure TFhirTestScriptFixture.SetAutodeleteST(value : Boolean);
begin
  if FAutodelete = nil then
    FAutodelete := TFhirBoolean.create;
  FAutodelete.value := value
end;

Procedure TFhirTestScriptFixture.SetResource(value : TFhirReference{TFhirReference});
begin
  FResource.free;
  FResource := value;
end;


{ TFhirTestScriptFixtureListEnumerator }

Constructor TFhirTestScriptFixtureListEnumerator.Create(list : TFhirTestScriptFixtureList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptFixtureListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptFixtureListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptFixtureListEnumerator.GetCurrent : TFhirTestScriptFixture;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptFixtureList }
procedure TFhirTestScriptFixtureList.AddItem(value: TFhirTestScriptFixture);
begin
  assert(value.ClassName = 'TFhirTestScriptFixture', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptFixture');
  add(value);
end;


function TFhirTestScriptFixtureList.Append: TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptFixtureList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptFixtureList.GetEnumerator : TFhirTestScriptFixtureListEnumerator;
begin
  result := TFhirTestScriptFixtureListEnumerator.Create(self.link);
end;

function TFhirTestScriptFixtureList.Clone: TFhirTestScriptFixtureList;
begin
  result := TFhirTestScriptFixtureList(inherited Clone);
end;

function TFhirTestScriptFixtureList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptFixtureList.GetItemN(index: Integer): TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture(ObjectByIndex[index]);
end;

function TFhirTestScriptFixtureList.IndexOf(value: TFhirTestScriptFixture): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptFixtureList.Insert(index: Integer): TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptFixtureList.InsertItem(index: Integer; value: TFhirTestScriptFixture);
begin
  assert(value is TFhirTestScriptFixture);
  Inherited Insert(index, value);
end;

function TFhirTestScriptFixtureList.Item(index: Integer): TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture(ObjectByIndex[index]);
end;

function TFhirTestScriptFixtureList.Link: TFhirTestScriptFixtureList;
begin
  result := TFhirTestScriptFixtureList(inherited Link);
end;

procedure TFhirTestScriptFixtureList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptFixtureList.SetItemByIndex(index: Integer; value: TFhirTestScriptFixture);
begin
  assert(value is TFhirTestScriptFixture);
  FhirTestScriptFixtures[index] := value;
end;

procedure TFhirTestScriptFixtureList.SetItemN(index: Integer; value: TFhirTestScriptFixture);
begin
  assert(value is TFhirTestScriptFixture);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptVariable }

constructor TFhirTestScriptVariable.Create;
begin
  inherited;
end;

destructor TFhirTestScriptVariable.Destroy;
begin
  FName.free;
  FHeaderField.free;
  FPath.free;
  FSourceId.free;
  inherited;
end;

procedure TFhirTestScriptVariable.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirTestScriptVariable(oSource).nameElement.Clone;
  headerFieldElement := TFhirTestScriptVariable(oSource).headerFieldElement.Clone;
  pathElement := TFhirTestScriptVariable(oSource).pathElement.Clone;
  sourceIdElement := TFhirTestScriptVariable(oSource).sourceIdElement.Clone;
end;

procedure TFhirTestScriptVariable.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'headerField') Then
     list.add(FHeaderField.Link);
  if (child_name = 'path') Then
     list.add(FPath.Link);
  if (child_name = 'sourceId') Then
     list.add(FSourceId.Link);
end;

procedure TFhirTestScriptVariable.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'headerField', 'string', FHeaderField.Link));{2}
  oList.add(TFHIRProperty.create(self, 'path', 'string', FPath.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sourceId', 'id', FSourceId.Link));{2}
end;

procedure TFhirTestScriptVariable.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'headerField') then HeaderFieldElement := propValue as TFhirString{5a}
  else if (propName = 'path') then PathElement := propValue as TFhirString{5a}
  else if (propName = 'sourceId') then SourceIdElement := propValue as TFhirId{5a}
  else inherited;
end;

function TFhirTestScriptVariable.FhirType : string;
begin
  result := 'variable';
end;

function TFhirTestScriptVariable.Link : TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable(inherited Link);
end;

function TFhirTestScriptVariable.Clone : TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable(inherited Clone);
end;

function TFhirTestScriptVariable.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptVariable;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptVariable)) then
    result := false
  else
  begin
    o := TFhirTestScriptVariable(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(headerFieldElement, o.headerFieldElement, true) and 
      compareDeep(pathElement, o.pathElement, true) and compareDeep(sourceIdElement, o.sourceIdElement, true);
  end;
end;

function TFhirTestScriptVariable.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptVariable;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptVariable)) then
    result := false
  else
  begin
    o := TFhirTestScriptVariable(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(headerFieldElement, o.headerFieldElement, true) and 
      compareValues(pathElement, o.pathElement, true) and compareValues(sourceIdElement, o.sourceIdElement, true);
  end;
end;

{ TFhirTestScriptVariable }

Procedure TFhirTestScriptVariable.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirTestScriptVariable.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirTestScriptVariable.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirTestScriptVariable.SetHeaderField(value : TFhirString);
begin
  FHeaderField.free;
  FHeaderField := value;
end;

Function TFhirTestScriptVariable.GetHeaderFieldST : String;
begin
  if FHeaderField = nil then
    result := ''
  else
    result := FHeaderField.value;
end;

Procedure TFhirTestScriptVariable.SetHeaderFieldST(value : String);
begin
  if value <> '' then
  begin
    if FHeaderField = nil then
      FHeaderField := TFhirString.create;
    FHeaderField.value := value
  end
  else if FHeaderField <> nil then
    FHeaderField.value := '';
end;

Procedure TFhirTestScriptVariable.SetPath(value : TFhirString);
begin
  FPath.free;
  FPath := value;
end;

Function TFhirTestScriptVariable.GetPathST : String;
begin
  if FPath = nil then
    result := ''
  else
    result := FPath.value;
end;

Procedure TFhirTestScriptVariable.SetPathST(value : String);
begin
  if value <> '' then
  begin
    if FPath = nil then
      FPath := TFhirString.create;
    FPath.value := value
  end
  else if FPath <> nil then
    FPath.value := '';
end;

Procedure TFhirTestScriptVariable.SetSourceId(value : TFhirId);
begin
  FSourceId.free;
  FSourceId := value;
end;

Function TFhirTestScriptVariable.GetSourceIdST : String;
begin
  if FSourceId = nil then
    result := ''
  else
    result := FSourceId.value;
end;

Procedure TFhirTestScriptVariable.SetSourceIdST(value : String);
begin
  if value <> '' then
  begin
    if FSourceId = nil then
      FSourceId := TFhirId.create;
    FSourceId.value := value
  end
  else if FSourceId <> nil then
    FSourceId.value := '';
end;


{ TFhirTestScriptVariableListEnumerator }

Constructor TFhirTestScriptVariableListEnumerator.Create(list : TFhirTestScriptVariableList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptVariableListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptVariableListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptVariableListEnumerator.GetCurrent : TFhirTestScriptVariable;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptVariableList }
procedure TFhirTestScriptVariableList.AddItem(value: TFhirTestScriptVariable);
begin
  assert(value.ClassName = 'TFhirTestScriptVariable', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptVariable');
  add(value);
end;


function TFhirTestScriptVariableList.Append: TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptVariableList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptVariableList.GetEnumerator : TFhirTestScriptVariableListEnumerator;
begin
  result := TFhirTestScriptVariableListEnumerator.Create(self.link);
end;

function TFhirTestScriptVariableList.Clone: TFhirTestScriptVariableList;
begin
  result := TFhirTestScriptVariableList(inherited Clone);
end;

function TFhirTestScriptVariableList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptVariableList.GetItemN(index: Integer): TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable(ObjectByIndex[index]);
end;

function TFhirTestScriptVariableList.IndexOf(value: TFhirTestScriptVariable): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptVariableList.Insert(index: Integer): TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptVariableList.InsertItem(index: Integer; value: TFhirTestScriptVariable);
begin
  assert(value is TFhirTestScriptVariable);
  Inherited Insert(index, value);
end;

function TFhirTestScriptVariableList.Item(index: Integer): TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable(ObjectByIndex[index]);
end;

function TFhirTestScriptVariableList.Link: TFhirTestScriptVariableList;
begin
  result := TFhirTestScriptVariableList(inherited Link);
end;

procedure TFhirTestScriptVariableList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptVariableList.SetItemByIndex(index: Integer; value: TFhirTestScriptVariable);
begin
  assert(value is TFhirTestScriptVariable);
  FhirTestScriptVariables[index] := value;
end;

procedure TFhirTestScriptVariableList.SetItemN(index: Integer; value: TFhirTestScriptVariable);
begin
  assert(value is TFhirTestScriptVariable);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptSetup }

constructor TFhirTestScriptSetup.Create;
begin
  inherited;
end;

destructor TFhirTestScriptSetup.Destroy;
begin
  FMetadata.free;
  FActionList.Free;
  inherited;
end;

procedure TFhirTestScriptSetup.Assign(oSource : TAdvObject);
begin
  inherited;
  metadata := TFhirTestScriptSetup(oSource).metadata.Clone;
  if (TFhirTestScriptSetup(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirTestScriptSetupActionList.Create;
    FActionList.Assign(TFhirTestScriptSetup(oSource).FActionList);
  end;
end;

procedure TFhirTestScriptSetup.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'metadata') Then
     list.add(FMetadata.Link);
  if (child_name = 'action') Then
    list.addAll(FActionList);
end;

procedure TFhirTestScriptSetup.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'metadata', '@TestScript.metadata', FMetadata.Link));{2}
  oList.add(TFHIRProperty.create(self, 'action', '', FActionList.Link)){3};
end;

procedure TFhirTestScriptSetup.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'metadata') then Metadata := propValue as TFhirTestScriptMetadata{4b}
  else if (propName = 'action') then ActionList.add(propValue as TFhirTestScriptSetupAction){2}
  else inherited;
end;

function TFhirTestScriptSetup.FhirType : string;
begin
  result := 'setup';
end;

function TFhirTestScriptSetup.Link : TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup(inherited Link);
end;

function TFhirTestScriptSetup.Clone : TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup(inherited Clone);
end;

function TFhirTestScriptSetup.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetup;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetup)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetup(other);
    result := compareDeep(metadataElement, o.metadataElement, true) and compareDeep(actionList, o.actionList, true);
  end;
end;

function TFhirTestScriptSetup.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetup;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetup)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetup(other);
    result := true;
  end;
end;

{ TFhirTestScriptSetup }

Procedure TFhirTestScriptSetup.SetMetadata(value : TFhirTestScriptMetadata);
begin
  FMetadata.free;
  FMetadata := value;
end;

Function TFhirTestScriptSetup.GetActionList : TFhirTestScriptSetupActionList;
begin
  if FActionList = nil then
    FActionList := TFhirTestScriptSetupActionList.Create;
  result := FActionList;
end;

Function TFhirTestScriptSetup.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;


{ TFhirTestScriptSetupListEnumerator }

Constructor TFhirTestScriptSetupListEnumerator.Create(list : TFhirTestScriptSetupList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptSetupListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptSetupListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptSetupListEnumerator.GetCurrent : TFhirTestScriptSetup;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptSetupList }
procedure TFhirTestScriptSetupList.AddItem(value: TFhirTestScriptSetup);
begin
  assert(value.ClassName = 'TFhirTestScriptSetup', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptSetup');
  add(value);
end;


function TFhirTestScriptSetupList.Append: TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptSetupList.GetEnumerator : TFhirTestScriptSetupListEnumerator;
begin
  result := TFhirTestScriptSetupListEnumerator.Create(self.link);
end;

function TFhirTestScriptSetupList.Clone: TFhirTestScriptSetupList;
begin
  result := TFhirTestScriptSetupList(inherited Clone);
end;

function TFhirTestScriptSetupList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptSetupList.GetItemN(index: Integer): TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupList.IndexOf(value: TFhirTestScriptSetup): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptSetupList.Insert(index: Integer): TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupList.InsertItem(index: Integer; value: TFhirTestScriptSetup);
begin
  assert(value is TFhirTestScriptSetup);
  Inherited Insert(index, value);
end;

function TFhirTestScriptSetupList.Item(index: Integer): TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupList.Link: TFhirTestScriptSetupList;
begin
  result := TFhirTestScriptSetupList(inherited Link);
end;

procedure TFhirTestScriptSetupList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptSetupList.SetItemByIndex(index: Integer; value: TFhirTestScriptSetup);
begin
  assert(value is TFhirTestScriptSetup);
  FhirTestScriptSetups[index] := value;
end;

procedure TFhirTestScriptSetupList.SetItemN(index: Integer; value: TFhirTestScriptSetup);
begin
  assert(value is TFhirTestScriptSetup);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptSetupAction }

constructor TFhirTestScriptSetupAction.Create;
begin
  inherited;
end;

destructor TFhirTestScriptSetupAction.Destroy;
begin
  FOperation.free;
  FAssert.free;
  inherited;
end;

procedure TFhirTestScriptSetupAction.Assign(oSource : TAdvObject);
begin
  inherited;
  operation := TFhirTestScriptSetupAction(oSource).operation.Clone;
  assert := TFhirTestScriptSetupAction(oSource).assert.Clone;
end;

procedure TFhirTestScriptSetupAction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'operation') Then
     list.add(FOperation.Link);
  if (child_name = 'assert') Then
     list.add(FAssert.Link);
end;

procedure TFhirTestScriptSetupAction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'operation', '', FOperation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'assert', '', FAssert.Link));{2}
end;

procedure TFhirTestScriptSetupAction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'operation') then Operation := propValue as TFhirTestScriptSetupActionOperation{4b}
  else if (propName = 'assert') then Assert := propValue as TFhirTestScriptSetupActionAssert{4b}
  else inherited;
end;

function TFhirTestScriptSetupAction.FhirType : string;
begin
  result := 'action';
end;

function TFhirTestScriptSetupAction.Link : TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction(inherited Link);
end;

function TFhirTestScriptSetupAction.Clone : TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction(inherited Clone);
end;

function TFhirTestScriptSetupAction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupAction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupAction(other);
    result := compareDeep(operationElement, o.operationElement, true) and compareDeep(assertElement, o.assertElement, true);
  end;
end;

function TFhirTestScriptSetupAction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupAction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupAction(other);
    result := true;
  end;
end;

{ TFhirTestScriptSetupAction }

Procedure TFhirTestScriptSetupAction.SetOperation(value : TFhirTestScriptSetupActionOperation);
begin
  FOperation.free;
  FOperation := value;
end;

Procedure TFhirTestScriptSetupAction.SetAssert(value : TFhirTestScriptSetupActionAssert);
begin
  FAssert.free;
  FAssert := value;
end;


{ TFhirTestScriptSetupActionListEnumerator }

Constructor TFhirTestScriptSetupActionListEnumerator.Create(list : TFhirTestScriptSetupActionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptSetupActionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptSetupActionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptSetupActionListEnumerator.GetCurrent : TFhirTestScriptSetupAction;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptSetupActionList }
procedure TFhirTestScriptSetupActionList.AddItem(value: TFhirTestScriptSetupAction);
begin
  assert(value.ClassName = 'TFhirTestScriptSetupAction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptSetupAction');
  add(value);
end;


function TFhirTestScriptSetupActionList.Append: TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptSetupActionList.GetEnumerator : TFhirTestScriptSetupActionListEnumerator;
begin
  result := TFhirTestScriptSetupActionListEnumerator.Create(self.link);
end;

function TFhirTestScriptSetupActionList.Clone: TFhirTestScriptSetupActionList;
begin
  result := TFhirTestScriptSetupActionList(inherited Clone);
end;

function TFhirTestScriptSetupActionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptSetupActionList.GetItemN(index: Integer): TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionList.IndexOf(value: TFhirTestScriptSetupAction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptSetupActionList.Insert(index: Integer): TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionList.InsertItem(index: Integer; value: TFhirTestScriptSetupAction);
begin
  assert(value is TFhirTestScriptSetupAction);
  Inherited Insert(index, value);
end;

function TFhirTestScriptSetupActionList.Item(index: Integer): TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionList.Link: TFhirTestScriptSetupActionList;
begin
  result := TFhirTestScriptSetupActionList(inherited Link);
end;

procedure TFhirTestScriptSetupActionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptSetupActionList.SetItemByIndex(index: Integer; value: TFhirTestScriptSetupAction);
begin
  assert(value is TFhirTestScriptSetupAction);
  FhirTestScriptSetupActions[index] := value;
end;

procedure TFhirTestScriptSetupActionList.SetItemN(index: Integer; value: TFhirTestScriptSetupAction);
begin
  assert(value is TFhirTestScriptSetupAction);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptSetupActionOperation }

constructor TFhirTestScriptSetupActionOperation.Create;
begin
  inherited;
end;

destructor TFhirTestScriptSetupActionOperation.Destroy;
begin
  FType_.free;
  FResource.free;
  FLabel_.free;
  FDescription.free;
  FAccept.free;
  FContentType.free;
  FDestination.free;
  FEncodeRequestUrl.free;
  FParams.free;
  FRequestHeaderList.Free;
  FResponseId.free;
  FSourceId.free;
  FTargetId.free;
  FUrl.free;
  inherited;
end;

procedure TFhirTestScriptSetupActionOperation.Assign(oSource : TAdvObject);
begin
  inherited;
  type_ := TFhirTestScriptSetupActionOperation(oSource).type_.Clone;
  resourceElement := TFhirTestScriptSetupActionOperation(oSource).resourceElement.Clone;
  label_Element := TFhirTestScriptSetupActionOperation(oSource).label_Element.Clone;
  descriptionElement := TFhirTestScriptSetupActionOperation(oSource).descriptionElement.Clone;
  FAccept := TFhirTestScriptSetupActionOperation(oSource).FAccept.Link;
  FContentType := TFhirTestScriptSetupActionOperation(oSource).FContentType.Link;
  destinationElement := TFhirTestScriptSetupActionOperation(oSource).destinationElement.Clone;
  encodeRequestUrlElement := TFhirTestScriptSetupActionOperation(oSource).encodeRequestUrlElement.Clone;
  paramsElement := TFhirTestScriptSetupActionOperation(oSource).paramsElement.Clone;
  if (TFhirTestScriptSetupActionOperation(oSource).FRequestHeaderList = nil) then
  begin
    FRequestHeaderList.free;
    FRequestHeaderList := nil;
  end
  else
  begin
    FRequestHeaderList := TFhirTestScriptSetupActionOperationRequestHeaderList.Create;
    FRequestHeaderList.Assign(TFhirTestScriptSetupActionOperation(oSource).FRequestHeaderList);
  end;
  responseIdElement := TFhirTestScriptSetupActionOperation(oSource).responseIdElement.Clone;
  sourceIdElement := TFhirTestScriptSetupActionOperation(oSource).sourceIdElement.Clone;
  targetIdElement := TFhirTestScriptSetupActionOperation(oSource).targetIdElement.Clone;
  urlElement := TFhirTestScriptSetupActionOperation(oSource).urlElement.Clone;
end;

procedure TFhirTestScriptSetupActionOperation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'type') Then
     list.add(FType_.Link);
  if (child_name = 'resource') Then
     list.add(FResource.Link);
  if (child_name = 'label') Then
     list.add(FLabel_.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'accept') Then
     list.add(FAccept.Link);
  if (child_name = 'contentType') Then
     list.add(FContentType.Link);
  if (child_name = 'destination') Then
     list.add(FDestination.Link);
  if (child_name = 'encodeRequestUrl') Then
     list.add(FEncodeRequestUrl.Link);
  if (child_name = 'params') Then
     list.add(FParams.Link);
  if (child_name = 'requestHeader') Then
    list.addAll(FRequestHeaderList);
  if (child_name = 'responseId') Then
     list.add(FResponseId.Link);
  if (child_name = 'sourceId') Then
     list.add(FSourceId.Link);
  if (child_name = 'targetId') Then
     list.add(FTargetId.Link);
  if (child_name = 'url') Then
     list.add(FUrl.Link);
end;

procedure TFhirTestScriptSetupActionOperation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'type', 'Coding', FType_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', 'code', FResource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'label', 'string', FLabel_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'accept', 'code', FAccept.Link));{1}
  oList.add(TFHIRProperty.create(self, 'contentType', 'code', FContentType.Link));{1}
  oList.add(TFHIRProperty.create(self, 'destination', 'integer', FDestination.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encodeRequestUrl', 'boolean', FEncodeRequestUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'params', 'string', FParams.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requestHeader', '', FRequestHeaderList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'responseId', 'id', FResponseId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sourceId', 'id', FSourceId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'targetId', 'id', FTargetId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'url', 'string', FUrl.Link));{2}
end;

procedure TFhirTestScriptSetupActionOperation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'type') then Type_ := propValue as TFhirCoding{4b}
  else if (propName = 'resource') then
    if propValue is TFHIRCode then
      ResourceElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      ResourceElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'label') then Label_Element := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'accept') then AcceptElement := propValue as TFHIREnum
  else if (propName = 'contentType') then ContentTypeElement := propValue as TFHIREnum
  else if (propName = 'destination') then DestinationElement := propValue as TFhirInteger{5a}
  else if (propName = 'encodeRequestUrl') then EncodeRequestUrlElement := propValue as TFhirBoolean{5a}
  else if (propName = 'params') then ParamsElement := propValue as TFhirString{5a}
  else if (propName = 'requestHeader') then RequestHeaderList.add(propValue as TFhirTestScriptSetupActionOperationRequestHeader){2}
  else if (propName = 'responseId') then ResponseIdElement := propValue as TFhirId{5a}
  else if (propName = 'sourceId') then SourceIdElement := propValue as TFhirId{5a}
  else if (propName = 'targetId') then TargetIdElement := propValue as TFhirId{5a}
  else if (propName = 'url') then UrlElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirTestScriptSetupActionOperation.FhirType : string;
begin
  result := 'operation';
end;

function TFhirTestScriptSetupActionOperation.Link : TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation(inherited Link);
end;

function TFhirTestScriptSetupActionOperation.Clone : TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation(inherited Clone);
end;

function TFhirTestScriptSetupActionOperation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionOperation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionOperation)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionOperation(other);
    result := compareDeep(type_Element, o.type_Element, true) and compareDeep(resourceElement, o.resourceElement, true) and 
      compareDeep(label_Element, o.label_Element, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(acceptElement, o.acceptElement, true) and compareDeep(contentTypeElement, o.contentTypeElement, true) and 
      compareDeep(destinationElement, o.destinationElement, true) and compareDeep(encodeRequestUrlElement, o.encodeRequestUrlElement, true) and 
      compareDeep(paramsElement, o.paramsElement, true) and compareDeep(requestHeaderList, o.requestHeaderList, true) and 
      compareDeep(responseIdElement, o.responseIdElement, true) and compareDeep(sourceIdElement, o.sourceIdElement, true) and 
      compareDeep(targetIdElement, o.targetIdElement, true) and compareDeep(urlElement, o.urlElement, true);
  end;
end;

function TFhirTestScriptSetupActionOperation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionOperation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionOperation)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionOperation(other);
    result := compareValues(resourceElement, o.resourceElement, true) and compareValues(label_Element, o.label_Element, true) and 
      compareValues(descriptionElement, o.descriptionElement, true) and compareValues(acceptElement, o.acceptElement, true) and 
      compareValues(contentTypeElement, o.contentTypeElement, true) and compareValues(destinationElement, o.destinationElement, true) and 
      compareValues(encodeRequestUrlElement, o.encodeRequestUrlElement, true) and compareValues(paramsElement, o.paramsElement, true) and 
      compareValues(responseIdElement, o.responseIdElement, true) and compareValues(sourceIdElement, o.sourceIdElement, true) and 
      compareValues(targetIdElement, o.targetIdElement, true) and compareValues(urlElement, o.urlElement, true);
  end;
end;

{ TFhirTestScriptSetupActionOperation }

Procedure TFhirTestScriptSetupActionOperation.SetType_(value : TFhirCoding);
begin
  FType_.free;
  FType_ := value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetResource(value : TFhirCode);
begin
  FResource.free;
  FResource := value;
end;

Function TFhirTestScriptSetupActionOperation.GetResourceST : String;
begin
  if FResource = nil then
    result := ''
  else
    result := FResource.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetResourceST(value : String);
begin
  if value <> '' then
  begin
    if FResource = nil then
      FResource := TFhirCode.create;
    FResource.value := value
  end
  else if FResource <> nil then
    FResource.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetLabel_(value : TFhirString);
begin
  FLabel_.free;
  FLabel_ := value;
end;

Function TFhirTestScriptSetupActionOperation.GetLabel_ST : String;
begin
  if FLabel_ = nil then
    result := ''
  else
    result := FLabel_.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetLabel_ST(value : String);
begin
  if value <> '' then
  begin
    if FLabel_ = nil then
      FLabel_ := TFhirString.create;
    FLabel_.value := value
  end
  else if FLabel_ <> nil then
    FLabel_.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScriptSetupActionOperation.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetAccept(value : TFhirEnum);
begin
  FAccept.free;
  FAccept := value;
end;

Function TFhirTestScriptSetupActionOperation.GetAcceptST : TFhirContentType;
begin
  if FAccept = nil then
    result := TFhirContentType(0)
  else
    result := TFhirContentType(StringArrayIndexOfSensitive(CODES_TFhirContentType, FAccept.value));
end;

Procedure TFhirTestScriptSetupActionOperation.SetAcceptST(value : TFhirContentType);
begin
  if ord(value) = 0 then
    AcceptElement := nil
  else
    AcceptElement := TFhirEnum.create(CODES_TFhirContentType[value]);
end;

Procedure TFhirTestScriptSetupActionOperation.SetContentType(value : TFhirEnum);
begin
  FContentType.free;
  FContentType := value;
end;

Function TFhirTestScriptSetupActionOperation.GetContentTypeST : TFhirContentType;
begin
  if FContentType = nil then
    result := TFhirContentType(0)
  else
    result := TFhirContentType(StringArrayIndexOfSensitive(CODES_TFhirContentType, FContentType.value));
end;

Procedure TFhirTestScriptSetupActionOperation.SetContentTypeST(value : TFhirContentType);
begin
  if ord(value) = 0 then
    ContentTypeElement := nil
  else
    ContentTypeElement := TFhirEnum.create(CODES_TFhirContentType[value]);
end;

Procedure TFhirTestScriptSetupActionOperation.SetDestination(value : TFhirInteger);
begin
  FDestination.free;
  FDestination := value;
end;

Function TFhirTestScriptSetupActionOperation.GetDestinationST : String;
begin
  if FDestination = nil then
    result := ''
  else
    result := FDestination.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetDestinationST(value : String);
begin
  if value <> '' then
  begin
    if FDestination = nil then
      FDestination := TFhirInteger.create;
    FDestination.value := value
  end
  else if FDestination <> nil then
    FDestination.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetEncodeRequestUrl(value : TFhirBoolean);
begin
  FEncodeRequestUrl.free;
  FEncodeRequestUrl := value;
end;

Function TFhirTestScriptSetupActionOperation.GetEncodeRequestUrlST : Boolean;
begin
  if FEncodeRequestUrl = nil then
    result := false
  else
    result := FEncodeRequestUrl.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetEncodeRequestUrlST(value : Boolean);
begin
  if FEncodeRequestUrl = nil then
    FEncodeRequestUrl := TFhirBoolean.create;
  FEncodeRequestUrl.value := value
end;

Procedure TFhirTestScriptSetupActionOperation.SetParams(value : TFhirString);
begin
  FParams.free;
  FParams := value;
end;

Function TFhirTestScriptSetupActionOperation.GetParamsST : String;
begin
  if FParams = nil then
    result := ''
  else
    result := FParams.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetParamsST(value : String);
begin
  if value <> '' then
  begin
    if FParams = nil then
      FParams := TFhirString.create;
    FParams.value := value
  end
  else if FParams <> nil then
    FParams.value := '';
end;

Function TFhirTestScriptSetupActionOperation.GetRequestHeaderList : TFhirTestScriptSetupActionOperationRequestHeaderList;
begin
  if FRequestHeaderList = nil then
    FRequestHeaderList := TFhirTestScriptSetupActionOperationRequestHeaderList.Create;
  result := FRequestHeaderList;
end;

Function TFhirTestScriptSetupActionOperation.GetHasRequestHeaderList : boolean;
begin
  result := (FRequestHeaderList <> nil) and (FRequestHeaderList.count > 0);
end;

Procedure TFhirTestScriptSetupActionOperation.SetResponseId(value : TFhirId);
begin
  FResponseId.free;
  FResponseId := value;
end;

Function TFhirTestScriptSetupActionOperation.GetResponseIdST : String;
begin
  if FResponseId = nil then
    result := ''
  else
    result := FResponseId.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetResponseIdST(value : String);
begin
  if value <> '' then
  begin
    if FResponseId = nil then
      FResponseId := TFhirId.create;
    FResponseId.value := value
  end
  else if FResponseId <> nil then
    FResponseId.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetSourceId(value : TFhirId);
begin
  FSourceId.free;
  FSourceId := value;
end;

Function TFhirTestScriptSetupActionOperation.GetSourceIdST : String;
begin
  if FSourceId = nil then
    result := ''
  else
    result := FSourceId.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetSourceIdST(value : String);
begin
  if value <> '' then
  begin
    if FSourceId = nil then
      FSourceId := TFhirId.create;
    FSourceId.value := value
  end
  else if FSourceId <> nil then
    FSourceId.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetTargetId(value : TFhirId);
begin
  FTargetId.free;
  FTargetId := value;
end;

Function TFhirTestScriptSetupActionOperation.GetTargetIdST : String;
begin
  if FTargetId = nil then
    result := ''
  else
    result := FTargetId.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetTargetIdST(value : String);
begin
  if value <> '' then
  begin
    if FTargetId = nil then
      FTargetId := TFhirId.create;
    FTargetId.value := value
  end
  else if FTargetId <> nil then
    FTargetId.value := '';
end;

Procedure TFhirTestScriptSetupActionOperation.SetUrl(value : TFhirString);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirTestScriptSetupActionOperation.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirTestScriptSetupActionOperation.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirString.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;


{ TFhirTestScriptSetupActionOperationListEnumerator }

Constructor TFhirTestScriptSetupActionOperationListEnumerator.Create(list : TFhirTestScriptSetupActionOperationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptSetupActionOperationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptSetupActionOperationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptSetupActionOperationListEnumerator.GetCurrent : TFhirTestScriptSetupActionOperation;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptSetupActionOperationList }
procedure TFhirTestScriptSetupActionOperationList.AddItem(value: TFhirTestScriptSetupActionOperation);
begin
  assert(value.ClassName = 'TFhirTestScriptSetupActionOperation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptSetupActionOperation');
  add(value);
end;


function TFhirTestScriptSetupActionOperationList.Append: TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionOperationList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptSetupActionOperationList.GetEnumerator : TFhirTestScriptSetupActionOperationListEnumerator;
begin
  result := TFhirTestScriptSetupActionOperationListEnumerator.Create(self.link);
end;

function TFhirTestScriptSetupActionOperationList.Clone: TFhirTestScriptSetupActionOperationList;
begin
  result := TFhirTestScriptSetupActionOperationList(inherited Clone);
end;

function TFhirTestScriptSetupActionOperationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptSetupActionOperationList.GetItemN(index: Integer): TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionOperationList.IndexOf(value: TFhirTestScriptSetupActionOperation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptSetupActionOperationList.Insert(index: Integer): TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionOperationList.InsertItem(index: Integer; value: TFhirTestScriptSetupActionOperation);
begin
  assert(value is TFhirTestScriptSetupActionOperation);
  Inherited Insert(index, value);
end;

function TFhirTestScriptSetupActionOperationList.Item(index: Integer): TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionOperationList.Link: TFhirTestScriptSetupActionOperationList;
begin
  result := TFhirTestScriptSetupActionOperationList(inherited Link);
end;

procedure TFhirTestScriptSetupActionOperationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptSetupActionOperationList.SetItemByIndex(index: Integer; value: TFhirTestScriptSetupActionOperation);
begin
  assert(value is TFhirTestScriptSetupActionOperation);
  FhirTestScriptSetupActionOperations[index] := value;
end;

procedure TFhirTestScriptSetupActionOperationList.SetItemN(index: Integer; value: TFhirTestScriptSetupActionOperation);
begin
  assert(value is TFhirTestScriptSetupActionOperation);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptSetupActionOperationRequestHeader }

constructor TFhirTestScriptSetupActionOperationRequestHeader.Create;
begin
  inherited;
end;

destructor TFhirTestScriptSetupActionOperationRequestHeader.Destroy;
begin
  FField.free;
  FValue.free;
  inherited;
end;

procedure TFhirTestScriptSetupActionOperationRequestHeader.Assign(oSource : TAdvObject);
begin
  inherited;
  fieldElement := TFhirTestScriptSetupActionOperationRequestHeader(oSource).fieldElement.Clone;
  valueElement := TFhirTestScriptSetupActionOperationRequestHeader(oSource).valueElement.Clone;
end;

procedure TFhirTestScriptSetupActionOperationRequestHeader.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'field') Then
     list.add(FField.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirTestScriptSetupActionOperationRequestHeader.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'field', 'string', FField.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'string', FValue.Link));{2}
end;

procedure TFhirTestScriptSetupActionOperationRequestHeader.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'field') then FieldElement := propValue as TFhirString{5a}
  else if (propName = 'value') then ValueElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirTestScriptSetupActionOperationRequestHeader.FhirType : string;
begin
  result := 'requestHeader';
end;

function TFhirTestScriptSetupActionOperationRequestHeader.Link : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader(inherited Link);
end;

function TFhirTestScriptSetupActionOperationRequestHeader.Clone : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader(inherited Clone);
end;

function TFhirTestScriptSetupActionOperationRequestHeader.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionOperationRequestHeader)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionOperationRequestHeader(other);
    result := compareDeep(fieldElement, o.fieldElement, true) and compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirTestScriptSetupActionOperationRequestHeader.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionOperationRequestHeader)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionOperationRequestHeader(other);
    result := compareValues(fieldElement, o.fieldElement, true) and compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirTestScriptSetupActionOperationRequestHeader }

Procedure TFhirTestScriptSetupActionOperationRequestHeader.SetField(value : TFhirString);
begin
  FField.free;
  FField := value;
end;

Function TFhirTestScriptSetupActionOperationRequestHeader.GetFieldST : String;
begin
  if FField = nil then
    result := ''
  else
    result := FField.value;
end;

Procedure TFhirTestScriptSetupActionOperationRequestHeader.SetFieldST(value : String);
begin
  if value <> '' then
  begin
    if FField = nil then
      FField := TFhirString.create;
    FField.value := value
  end
  else if FField <> nil then
    FField.value := '';
end;

Procedure TFhirTestScriptSetupActionOperationRequestHeader.SetValue(value : TFhirString);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirTestScriptSetupActionOperationRequestHeader.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirTestScriptSetupActionOperationRequestHeader.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirString.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator }

Constructor TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator.Create(list : TFhirTestScriptSetupActionOperationRequestHeaderList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator.GetCurrent : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptSetupActionOperationRequestHeaderList }
procedure TFhirTestScriptSetupActionOperationRequestHeaderList.AddItem(value: TFhirTestScriptSetupActionOperationRequestHeader);
begin
  assert(value.ClassName = 'TFhirTestScriptSetupActionOperationRequestHeader', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptSetupActionOperationRequestHeader');
  add(value);
end;


function TFhirTestScriptSetupActionOperationRequestHeaderList.Append: TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionOperationRequestHeaderList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.GetEnumerator : TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeaderListEnumerator.Create(self.link);
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.Clone: TFhirTestScriptSetupActionOperationRequestHeaderList;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeaderList(inherited Clone);
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.GetItemN(index: Integer): TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.IndexOf(value: TFhirTestScriptSetupActionOperationRequestHeader): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptSetupActionOperationRequestHeaderList.Insert(index: Integer): TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionOperationRequestHeaderList.InsertItem(index: Integer; value: TFhirTestScriptSetupActionOperationRequestHeader);
begin
  assert(value is TFhirTestScriptSetupActionOperationRequestHeader);
  Inherited Insert(index, value);
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.Item(index: Integer): TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionOperationRequestHeaderList.Link: TFhirTestScriptSetupActionOperationRequestHeaderList;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeaderList(inherited Link);
end;

procedure TFhirTestScriptSetupActionOperationRequestHeaderList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptSetupActionOperationRequestHeaderList.SetItemByIndex(index: Integer; value: TFhirTestScriptSetupActionOperationRequestHeader);
begin
  assert(value is TFhirTestScriptSetupActionOperationRequestHeader);
  FhirTestScriptSetupActionOperationRequestHeaders[index] := value;
end;

procedure TFhirTestScriptSetupActionOperationRequestHeaderList.SetItemN(index: Integer; value: TFhirTestScriptSetupActionOperationRequestHeader);
begin
  assert(value is TFhirTestScriptSetupActionOperationRequestHeader);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptSetupActionAssert }

constructor TFhirTestScriptSetupActionAssert.Create;
begin
  inherited;
end;

destructor TFhirTestScriptSetupActionAssert.Destroy;
begin
  FLabel_.free;
  FDescription.free;
  FDirection.free;
  FCompareToSourceId.free;
  FCompareToSourcePath.free;
  FContentType.free;
  FHeaderField.free;
  FMinimumId.free;
  FNavigationLinks.free;
  FOperator.free;
  FPath.free;
  FResource.free;
  FResponse.free;
  FResponseCode.free;
  FSourceId.free;
  FValidateProfileId.free;
  FValue.free;
  FWarningOnly.free;
  inherited;
end;

procedure TFhirTestScriptSetupActionAssert.Assign(oSource : TAdvObject);
begin
  inherited;
  label_Element := TFhirTestScriptSetupActionAssert(oSource).label_Element.Clone;
  descriptionElement := TFhirTestScriptSetupActionAssert(oSource).descriptionElement.Clone;
  FDirection := TFhirTestScriptSetupActionAssert(oSource).FDirection.Link;
  compareToSourceIdElement := TFhirTestScriptSetupActionAssert(oSource).compareToSourceIdElement.Clone;
  compareToSourcePathElement := TFhirTestScriptSetupActionAssert(oSource).compareToSourcePathElement.Clone;
  FContentType := TFhirTestScriptSetupActionAssert(oSource).FContentType.Link;
  headerFieldElement := TFhirTestScriptSetupActionAssert(oSource).headerFieldElement.Clone;
  minimumIdElement := TFhirTestScriptSetupActionAssert(oSource).minimumIdElement.Clone;
  navigationLinksElement := TFhirTestScriptSetupActionAssert(oSource).navigationLinksElement.Clone;
  FOperator := TFhirTestScriptSetupActionAssert(oSource).FOperator.Link;
  pathElement := TFhirTestScriptSetupActionAssert(oSource).pathElement.Clone;
  resourceElement := TFhirTestScriptSetupActionAssert(oSource).resourceElement.Clone;
  FResponse := TFhirTestScriptSetupActionAssert(oSource).FResponse.Link;
  responseCodeElement := TFhirTestScriptSetupActionAssert(oSource).responseCodeElement.Clone;
  sourceIdElement := TFhirTestScriptSetupActionAssert(oSource).sourceIdElement.Clone;
  validateProfileIdElement := TFhirTestScriptSetupActionAssert(oSource).validateProfileIdElement.Clone;
  valueElement := TFhirTestScriptSetupActionAssert(oSource).valueElement.Clone;
  warningOnlyElement := TFhirTestScriptSetupActionAssert(oSource).warningOnlyElement.Clone;
end;

procedure TFhirTestScriptSetupActionAssert.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'label') Then
     list.add(FLabel_.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'direction') Then
     list.add(FDirection.Link);
  if (child_name = 'compareToSourceId') Then
     list.add(FCompareToSourceId.Link);
  if (child_name = 'compareToSourcePath') Then
     list.add(FCompareToSourcePath.Link);
  if (child_name = 'contentType') Then
     list.add(FContentType.Link);
  if (child_name = 'headerField') Then
     list.add(FHeaderField.Link);
  if (child_name = 'minimumId') Then
     list.add(FMinimumId.Link);
  if (child_name = 'navigationLinks') Then
     list.add(FNavigationLinks.Link);
  if (child_name = 'operator') Then
     list.add(FOperator.Link);
  if (child_name = 'path') Then
     list.add(FPath.Link);
  if (child_name = 'resource') Then
     list.add(FResource.Link);
  if (child_name = 'response') Then
     list.add(FResponse.Link);
  if (child_name = 'responseCode') Then
     list.add(FResponseCode.Link);
  if (child_name = 'sourceId') Then
     list.add(FSourceId.Link);
  if (child_name = 'validateProfileId') Then
     list.add(FValidateProfileId.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
  if (child_name = 'warningOnly') Then
     list.add(FWarningOnly.Link);
end;

procedure TFhirTestScriptSetupActionAssert.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'label', 'string', FLabel_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'direction', 'code', FDirection.Link));{1}
  oList.add(TFHIRProperty.create(self, 'compareToSourceId', 'string', FCompareToSourceId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'compareToSourcePath', 'string', FCompareToSourcePath.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contentType', 'code', FContentType.Link));{1}
  oList.add(TFHIRProperty.create(self, 'headerField', 'string', FHeaderField.Link));{2}
  oList.add(TFHIRProperty.create(self, 'minimumId', 'string', FMinimumId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'navigationLinks', 'boolean', FNavigationLinks.Link));{2}
  oList.add(TFHIRProperty.create(self, 'operator', 'code', FOperator.Link));{1}
  oList.add(TFHIRProperty.create(self, 'path', 'string', FPath.Link));{2}
  oList.add(TFHIRProperty.create(self, 'resource', 'code', FResource.Link));{2}
  oList.add(TFHIRProperty.create(self, 'response', 'code', FResponse.Link));{1}
  oList.add(TFHIRProperty.create(self, 'responseCode', 'string', FResponseCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'sourceId', 'id', FSourceId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'validateProfileId', 'id', FValidateProfileId.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'string', FValue.Link));{2}
  oList.add(TFHIRProperty.create(self, 'warningOnly', 'boolean', FWarningOnly.Link));{2}
end;

procedure TFhirTestScriptSetupActionAssert.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'label') then Label_Element := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'direction') then DirectionElement := propValue as TFHIREnum
  else if (propName = 'compareToSourceId') then CompareToSourceIdElement := propValue as TFhirString{5a}
  else if (propName = 'compareToSourcePath') then CompareToSourcePathElement := propValue as TFhirString{5a}
  else if (propName = 'contentType') then ContentTypeElement := propValue as TFHIREnum
  else if (propName = 'headerField') then HeaderFieldElement := propValue as TFhirString{5a}
  else if (propName = 'minimumId') then MinimumIdElement := propValue as TFhirString{5a}
  else if (propName = 'navigationLinks') then NavigationLinksElement := propValue as TFhirBoolean{5a}
  else if (propName = 'operator') then OperatorElement := propValue as TFHIREnum
  else if (propName = 'path') then PathElement := propValue as TFhirString{5a}
  else if (propName = 'resource') then
    if propValue is TFHIRCode then
      ResourceElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      ResourceElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'response') then ResponseElement := propValue as TFHIREnum
  else if (propName = 'responseCode') then ResponseCodeElement := propValue as TFhirString{5a}
  else if (propName = 'sourceId') then SourceIdElement := propValue as TFhirId{5a}
  else if (propName = 'validateProfileId') then ValidateProfileIdElement := propValue as TFhirId{5a}
  else if (propName = 'value') then ValueElement := propValue as TFhirString{5a}
  else if (propName = 'warningOnly') then WarningOnlyElement := propValue as TFhirBoolean{5a}
  else inherited;
end;

function TFhirTestScriptSetupActionAssert.FhirType : string;
begin
  result := 'assert';
end;

function TFhirTestScriptSetupActionAssert.Link : TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert(inherited Link);
end;

function TFhirTestScriptSetupActionAssert.Clone : TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert(inherited Clone);
end;

function TFhirTestScriptSetupActionAssert.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionAssert;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionAssert)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionAssert(other);
    result := compareDeep(label_Element, o.label_Element, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(directionElement, o.directionElement, true) and compareDeep(compareToSourceIdElement, o.compareToSourceIdElement, true) and 
      compareDeep(compareToSourcePathElement, o.compareToSourcePathElement, true) and 
      compareDeep(contentTypeElement, o.contentTypeElement, true) and compareDeep(headerFieldElement, o.headerFieldElement, true) and 
      compareDeep(minimumIdElement, o.minimumIdElement, true) and compareDeep(navigationLinksElement, o.navigationLinksElement, true) and 
      compareDeep(operatorElement, o.operatorElement, true) and compareDeep(pathElement, o.pathElement, true) and 
      compareDeep(resourceElement, o.resourceElement, true) and compareDeep(responseElement, o.responseElement, true) and 
      compareDeep(responseCodeElement, o.responseCodeElement, true) and compareDeep(sourceIdElement, o.sourceIdElement, true) and 
      compareDeep(validateProfileIdElement, o.validateProfileIdElement, true) and compareDeep(valueElement, o.valueElement, true) and 
      compareDeep(warningOnlyElement, o.warningOnlyElement, true);
  end;
end;

function TFhirTestScriptSetupActionAssert.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptSetupActionAssert;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptSetupActionAssert)) then
    result := false
  else
  begin
    o := TFhirTestScriptSetupActionAssert(other);
    result := compareValues(label_Element, o.label_Element, true) and compareValues(descriptionElement, o.descriptionElement, true) and 
      compareValues(directionElement, o.directionElement, true) and compareValues(compareToSourceIdElement, o.compareToSourceIdElement, true) and 
      compareValues(compareToSourcePathElement, o.compareToSourcePathElement, true) and 
      compareValues(contentTypeElement, o.contentTypeElement, true) and compareValues(headerFieldElement, o.headerFieldElement, true) and 
      compareValues(minimumIdElement, o.minimumIdElement, true) and compareValues(navigationLinksElement, o.navigationLinksElement, true) and 
      compareValues(operatorElement, o.operatorElement, true) and compareValues(pathElement, o.pathElement, true) and 
      compareValues(resourceElement, o.resourceElement, true) and compareValues(responseElement, o.responseElement, true) and 
      compareValues(responseCodeElement, o.responseCodeElement, true) and compareValues(sourceIdElement, o.sourceIdElement, true) and 
      compareValues(validateProfileIdElement, o.validateProfileIdElement, true) and compareValues(valueElement, o.valueElement, true) and 
      compareValues(warningOnlyElement, o.warningOnlyElement, true);
  end;
end;

{ TFhirTestScriptSetupActionAssert }

Procedure TFhirTestScriptSetupActionAssert.SetLabel_(value : TFhirString);
begin
  FLabel_.free;
  FLabel_ := value;
end;

Function TFhirTestScriptSetupActionAssert.GetLabel_ST : String;
begin
  if FLabel_ = nil then
    result := ''
  else
    result := FLabel_.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetLabel_ST(value : String);
begin
  if value <> '' then
  begin
    if FLabel_ = nil then
      FLabel_ := TFhirString.create;
    FLabel_.value := value
  end
  else if FLabel_ <> nil then
    FLabel_.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScriptSetupActionAssert.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetDirection(value : TFhirEnum);
begin
  FDirection.free;
  FDirection := value;
end;

Function TFhirTestScriptSetupActionAssert.GetDirectionST : TFhirAssertDirectionCodes;
begin
  if FDirection = nil then
    result := TFhirAssertDirectionCodes(0)
  else
    result := TFhirAssertDirectionCodes(StringArrayIndexOfSensitive(CODES_TFhirAssertDirectionCodes, FDirection.value));
end;

Procedure TFhirTestScriptSetupActionAssert.SetDirectionST(value : TFhirAssertDirectionCodes);
begin
  if ord(value) = 0 then
    DirectionElement := nil
  else
    DirectionElement := TFhirEnum.create(CODES_TFhirAssertDirectionCodes[value]);
end;

Procedure TFhirTestScriptSetupActionAssert.SetCompareToSourceId(value : TFhirString);
begin
  FCompareToSourceId.free;
  FCompareToSourceId := value;
end;

Function TFhirTestScriptSetupActionAssert.GetCompareToSourceIdST : String;
begin
  if FCompareToSourceId = nil then
    result := ''
  else
    result := FCompareToSourceId.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetCompareToSourceIdST(value : String);
begin
  if value <> '' then
  begin
    if FCompareToSourceId = nil then
      FCompareToSourceId := TFhirString.create;
    FCompareToSourceId.value := value
  end
  else if FCompareToSourceId <> nil then
    FCompareToSourceId.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetCompareToSourcePath(value : TFhirString);
begin
  FCompareToSourcePath.free;
  FCompareToSourcePath := value;
end;

Function TFhirTestScriptSetupActionAssert.GetCompareToSourcePathST : String;
begin
  if FCompareToSourcePath = nil then
    result := ''
  else
    result := FCompareToSourcePath.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetCompareToSourcePathST(value : String);
begin
  if value <> '' then
  begin
    if FCompareToSourcePath = nil then
      FCompareToSourcePath := TFhirString.create;
    FCompareToSourcePath.value := value
  end
  else if FCompareToSourcePath <> nil then
    FCompareToSourcePath.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetContentType(value : TFhirEnum);
begin
  FContentType.free;
  FContentType := value;
end;

Function TFhirTestScriptSetupActionAssert.GetContentTypeST : TFhirContentType;
begin
  if FContentType = nil then
    result := TFhirContentType(0)
  else
    result := TFhirContentType(StringArrayIndexOfSensitive(CODES_TFhirContentType, FContentType.value));
end;

Procedure TFhirTestScriptSetupActionAssert.SetContentTypeST(value : TFhirContentType);
begin
  if ord(value) = 0 then
    ContentTypeElement := nil
  else
    ContentTypeElement := TFhirEnum.create(CODES_TFhirContentType[value]);
end;

Procedure TFhirTestScriptSetupActionAssert.SetHeaderField(value : TFhirString);
begin
  FHeaderField.free;
  FHeaderField := value;
end;

Function TFhirTestScriptSetupActionAssert.GetHeaderFieldST : String;
begin
  if FHeaderField = nil then
    result := ''
  else
    result := FHeaderField.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetHeaderFieldST(value : String);
begin
  if value <> '' then
  begin
    if FHeaderField = nil then
      FHeaderField := TFhirString.create;
    FHeaderField.value := value
  end
  else if FHeaderField <> nil then
    FHeaderField.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetMinimumId(value : TFhirString);
begin
  FMinimumId.free;
  FMinimumId := value;
end;

Function TFhirTestScriptSetupActionAssert.GetMinimumIdST : String;
begin
  if FMinimumId = nil then
    result := ''
  else
    result := FMinimumId.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetMinimumIdST(value : String);
begin
  if value <> '' then
  begin
    if FMinimumId = nil then
      FMinimumId := TFhirString.create;
    FMinimumId.value := value
  end
  else if FMinimumId <> nil then
    FMinimumId.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetNavigationLinks(value : TFhirBoolean);
begin
  FNavigationLinks.free;
  FNavigationLinks := value;
end;

Function TFhirTestScriptSetupActionAssert.GetNavigationLinksST : Boolean;
begin
  if FNavigationLinks = nil then
    result := false
  else
    result := FNavigationLinks.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetNavigationLinksST(value : Boolean);
begin
  if FNavigationLinks = nil then
    FNavigationLinks := TFhirBoolean.create;
  FNavigationLinks.value := value
end;

Procedure TFhirTestScriptSetupActionAssert.SetOperator(value : TFhirEnum);
begin
  FOperator.free;
  FOperator := value;
end;

Function TFhirTestScriptSetupActionAssert.GetOperatorST : TFhirAssertOperatorCodes;
begin
  if FOperator = nil then
    result := TFhirAssertOperatorCodes(0)
  else
    result := TFhirAssertOperatorCodes(StringArrayIndexOfSensitive(CODES_TFhirAssertOperatorCodes, FOperator.value));
end;

Procedure TFhirTestScriptSetupActionAssert.SetOperatorST(value : TFhirAssertOperatorCodes);
begin
  if ord(value) = 0 then
    OperatorElement := nil
  else
    OperatorElement := TFhirEnum.create(CODES_TFhirAssertOperatorCodes[value]);
end;

Procedure TFhirTestScriptSetupActionAssert.SetPath(value : TFhirString);
begin
  FPath.free;
  FPath := value;
end;

Function TFhirTestScriptSetupActionAssert.GetPathST : String;
begin
  if FPath = nil then
    result := ''
  else
    result := FPath.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetPathST(value : String);
begin
  if value <> '' then
  begin
    if FPath = nil then
      FPath := TFhirString.create;
    FPath.value := value
  end
  else if FPath <> nil then
    FPath.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetResource(value : TFhirCode);
begin
  FResource.free;
  FResource := value;
end;

Function TFhirTestScriptSetupActionAssert.GetResourceST : String;
begin
  if FResource = nil then
    result := ''
  else
    result := FResource.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetResourceST(value : String);
begin
  if value <> '' then
  begin
    if FResource = nil then
      FResource := TFhirCode.create;
    FResource.value := value
  end
  else if FResource <> nil then
    FResource.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetResponse(value : TFhirEnum);
begin
  FResponse.free;
  FResponse := value;
end;

Function TFhirTestScriptSetupActionAssert.GetResponseST : TFhirAssertResponseCodeTypes;
begin
  if FResponse = nil then
    result := TFhirAssertResponseCodeTypes(0)
  else
    result := TFhirAssertResponseCodeTypes(StringArrayIndexOfSensitive(CODES_TFhirAssertResponseCodeTypes, FResponse.value));
end;

Procedure TFhirTestScriptSetupActionAssert.SetResponseST(value : TFhirAssertResponseCodeTypes);
begin
  if ord(value) = 0 then
    ResponseElement := nil
  else
    ResponseElement := TFhirEnum.create(CODES_TFhirAssertResponseCodeTypes[value]);
end;

Procedure TFhirTestScriptSetupActionAssert.SetResponseCode(value : TFhirString);
begin
  FResponseCode.free;
  FResponseCode := value;
end;

Function TFhirTestScriptSetupActionAssert.GetResponseCodeST : String;
begin
  if FResponseCode = nil then
    result := ''
  else
    result := FResponseCode.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetResponseCodeST(value : String);
begin
  if value <> '' then
  begin
    if FResponseCode = nil then
      FResponseCode := TFhirString.create;
    FResponseCode.value := value
  end
  else if FResponseCode <> nil then
    FResponseCode.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetSourceId(value : TFhirId);
begin
  FSourceId.free;
  FSourceId := value;
end;

Function TFhirTestScriptSetupActionAssert.GetSourceIdST : String;
begin
  if FSourceId = nil then
    result := ''
  else
    result := FSourceId.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetSourceIdST(value : String);
begin
  if value <> '' then
  begin
    if FSourceId = nil then
      FSourceId := TFhirId.create;
    FSourceId.value := value
  end
  else if FSourceId <> nil then
    FSourceId.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetValidateProfileId(value : TFhirId);
begin
  FValidateProfileId.free;
  FValidateProfileId := value;
end;

Function TFhirTestScriptSetupActionAssert.GetValidateProfileIdST : String;
begin
  if FValidateProfileId = nil then
    result := ''
  else
    result := FValidateProfileId.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetValidateProfileIdST(value : String);
begin
  if value <> '' then
  begin
    if FValidateProfileId = nil then
      FValidateProfileId := TFhirId.create;
    FValidateProfileId.value := value
  end
  else if FValidateProfileId <> nil then
    FValidateProfileId.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetValue(value : TFhirString);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirTestScriptSetupActionAssert.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirString.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;

Procedure TFhirTestScriptSetupActionAssert.SetWarningOnly(value : TFhirBoolean);
begin
  FWarningOnly.free;
  FWarningOnly := value;
end;

Function TFhirTestScriptSetupActionAssert.GetWarningOnlyST : Boolean;
begin
  if FWarningOnly = nil then
    result := false
  else
    result := FWarningOnly.value;
end;

Procedure TFhirTestScriptSetupActionAssert.SetWarningOnlyST(value : Boolean);
begin
  if FWarningOnly = nil then
    FWarningOnly := TFhirBoolean.create;
  FWarningOnly.value := value
end;


{ TFhirTestScriptSetupActionAssertListEnumerator }

Constructor TFhirTestScriptSetupActionAssertListEnumerator.Create(list : TFhirTestScriptSetupActionAssertList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptSetupActionAssertListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptSetupActionAssertListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptSetupActionAssertListEnumerator.GetCurrent : TFhirTestScriptSetupActionAssert;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptSetupActionAssertList }
procedure TFhirTestScriptSetupActionAssertList.AddItem(value: TFhirTestScriptSetupActionAssert);
begin
  assert(value.ClassName = 'TFhirTestScriptSetupActionAssert', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptSetupActionAssert');
  add(value);
end;


function TFhirTestScriptSetupActionAssertList.Append: TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionAssertList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptSetupActionAssertList.GetEnumerator : TFhirTestScriptSetupActionAssertListEnumerator;
begin
  result := TFhirTestScriptSetupActionAssertListEnumerator.Create(self.link);
end;

function TFhirTestScriptSetupActionAssertList.Clone: TFhirTestScriptSetupActionAssertList;
begin
  result := TFhirTestScriptSetupActionAssertList(inherited Clone);
end;

function TFhirTestScriptSetupActionAssertList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptSetupActionAssertList.GetItemN(index: Integer): TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionAssertList.IndexOf(value: TFhirTestScriptSetupActionAssert): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptSetupActionAssertList.Insert(index: Integer): TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptSetupActionAssertList.InsertItem(index: Integer; value: TFhirTestScriptSetupActionAssert);
begin
  assert(value is TFhirTestScriptSetupActionAssert);
  Inherited Insert(index, value);
end;

function TFhirTestScriptSetupActionAssertList.Item(index: Integer): TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert(ObjectByIndex[index]);
end;

function TFhirTestScriptSetupActionAssertList.Link: TFhirTestScriptSetupActionAssertList;
begin
  result := TFhirTestScriptSetupActionAssertList(inherited Link);
end;

procedure TFhirTestScriptSetupActionAssertList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptSetupActionAssertList.SetItemByIndex(index: Integer; value: TFhirTestScriptSetupActionAssert);
begin
  assert(value is TFhirTestScriptSetupActionAssert);
  FhirTestScriptSetupActionAsserts[index] := value;
end;

procedure TFhirTestScriptSetupActionAssertList.SetItemN(index: Integer; value: TFhirTestScriptSetupActionAssert);
begin
  assert(value is TFhirTestScriptSetupActionAssert);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptTest }

constructor TFhirTestScriptTest.Create;
begin
  inherited;
end;

destructor TFhirTestScriptTest.Destroy;
begin
  FName.free;
  FDescription.free;
  FMetadata.free;
  FActionList.Free;
  inherited;
end;

procedure TFhirTestScriptTest.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirTestScriptTest(oSource).nameElement.Clone;
  descriptionElement := TFhirTestScriptTest(oSource).descriptionElement.Clone;
  metadata := TFhirTestScriptTest(oSource).metadata.Clone;
  if (TFhirTestScriptTest(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirTestScriptTestActionList.Create;
    FActionList.Assign(TFhirTestScriptTest(oSource).FActionList);
  end;
end;

procedure TFhirTestScriptTest.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'metadata') Then
     list.add(FMetadata.Link);
  if (child_name = 'action') Then
    list.addAll(FActionList);
end;

procedure TFhirTestScriptTest.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'metadata', '@TestScript.metadata', FMetadata.Link));{2}
  oList.add(TFHIRProperty.create(self, 'action', '', FActionList.Link)){3};
end;

procedure TFhirTestScriptTest.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'metadata') then Metadata := propValue as TFhirTestScriptMetadata{4b}
  else if (propName = 'action') then ActionList.add(propValue as TFhirTestScriptTestAction){2}
  else inherited;
end;

function TFhirTestScriptTest.FhirType : string;
begin
  result := 'test';
end;

function TFhirTestScriptTest.Link : TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest(inherited Link);
end;

function TFhirTestScriptTest.Clone : TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest(inherited Clone);
end;

function TFhirTestScriptTest.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTest;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptTest)) then
    result := false
  else
  begin
    o := TFhirTestScriptTest(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(descriptionElement, o.descriptionElement, true) and 
      compareDeep(metadataElement, o.metadataElement, true) and compareDeep(actionList, o.actionList, true);
  end;
end;

function TFhirTestScriptTest.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTest;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptTest)) then
    result := false
  else
  begin
    o := TFhirTestScriptTest(other);
    result := compareValues(nameElement, o.nameElement, true) and compareValues(descriptionElement, o.descriptionElement, true);
  end;
end;

{ TFhirTestScriptTest }

Procedure TFhirTestScriptTest.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirTestScriptTest.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirTestScriptTest.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirTestScriptTest.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScriptTest.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScriptTest.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Procedure TFhirTestScriptTest.SetMetadata(value : TFhirTestScriptMetadata);
begin
  FMetadata.free;
  FMetadata := value;
end;

Function TFhirTestScriptTest.GetActionList : TFhirTestScriptTestActionList;
begin
  if FActionList = nil then
    FActionList := TFhirTestScriptTestActionList.Create;
  result := FActionList;
end;

Function TFhirTestScriptTest.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;


{ TFhirTestScriptTestListEnumerator }

Constructor TFhirTestScriptTestListEnumerator.Create(list : TFhirTestScriptTestList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptTestListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptTestListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptTestListEnumerator.GetCurrent : TFhirTestScriptTest;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptTestList }
procedure TFhirTestScriptTestList.AddItem(value: TFhirTestScriptTest);
begin
  assert(value.ClassName = 'TFhirTestScriptTest', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptTest');
  add(value);
end;


function TFhirTestScriptTestList.Append: TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTestList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptTestList.GetEnumerator : TFhirTestScriptTestListEnumerator;
begin
  result := TFhirTestScriptTestListEnumerator.Create(self.link);
end;

function TFhirTestScriptTestList.Clone: TFhirTestScriptTestList;
begin
  result := TFhirTestScriptTestList(inherited Clone);
end;

function TFhirTestScriptTestList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptTestList.GetItemN(index: Integer): TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest(ObjectByIndex[index]);
end;

function TFhirTestScriptTestList.IndexOf(value: TFhirTestScriptTest): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptTestList.Insert(index: Integer): TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTestList.InsertItem(index: Integer; value: TFhirTestScriptTest);
begin
  assert(value is TFhirTestScriptTest);
  Inherited Insert(index, value);
end;

function TFhirTestScriptTestList.Item(index: Integer): TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest(ObjectByIndex[index]);
end;

function TFhirTestScriptTestList.Link: TFhirTestScriptTestList;
begin
  result := TFhirTestScriptTestList(inherited Link);
end;

procedure TFhirTestScriptTestList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptTestList.SetItemByIndex(index: Integer; value: TFhirTestScriptTest);
begin
  assert(value is TFhirTestScriptTest);
  FhirTestScriptTests[index] := value;
end;

procedure TFhirTestScriptTestList.SetItemN(index: Integer; value: TFhirTestScriptTest);
begin
  assert(value is TFhirTestScriptTest);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptTestAction }

constructor TFhirTestScriptTestAction.Create;
begin
  inherited;
end;

destructor TFhirTestScriptTestAction.Destroy;
begin
  FOperation.free;
  FAssert.free;
  inherited;
end;

procedure TFhirTestScriptTestAction.Assign(oSource : TAdvObject);
begin
  inherited;
  operation := TFhirTestScriptTestAction(oSource).operation.Clone;
  assert := TFhirTestScriptTestAction(oSource).assert.Clone;
end;

procedure TFhirTestScriptTestAction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'operation') Then
     list.add(FOperation.Link);
  if (child_name = 'assert') Then
     list.add(FAssert.Link);
end;

procedure TFhirTestScriptTestAction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'operation', '@TestScript.setup.action.operation', FOperation.Link));{2}
  oList.add(TFHIRProperty.create(self, 'assert', '@TestScript.setup.action.assert', FAssert.Link));{2}
end;

procedure TFhirTestScriptTestAction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'operation') then Operation := propValue as TFhirTestScriptSetupActionOperation{4b}
  else if (propName = 'assert') then Assert := propValue as TFhirTestScriptSetupActionAssert{4b}
  else inherited;
end;

function TFhirTestScriptTestAction.FhirType : string;
begin
  result := 'action';
end;

function TFhirTestScriptTestAction.Link : TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction(inherited Link);
end;

function TFhirTestScriptTestAction.Clone : TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction(inherited Clone);
end;

function TFhirTestScriptTestAction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTestAction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptTestAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptTestAction(other);
    result := compareDeep(operationElement, o.operationElement, true) and compareDeep(assertElement, o.assertElement, true);
  end;
end;

function TFhirTestScriptTestAction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTestAction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptTestAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptTestAction(other);
    result := true;
  end;
end;

{ TFhirTestScriptTestAction }

Procedure TFhirTestScriptTestAction.SetOperation(value : TFhirTestScriptSetupActionOperation);
begin
  FOperation.free;
  FOperation := value;
end;

Procedure TFhirTestScriptTestAction.SetAssert(value : TFhirTestScriptSetupActionAssert);
begin
  FAssert.free;
  FAssert := value;
end;


{ TFhirTestScriptTestActionListEnumerator }

Constructor TFhirTestScriptTestActionListEnumerator.Create(list : TFhirTestScriptTestActionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptTestActionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptTestActionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptTestActionListEnumerator.GetCurrent : TFhirTestScriptTestAction;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptTestActionList }
procedure TFhirTestScriptTestActionList.AddItem(value: TFhirTestScriptTestAction);
begin
  assert(value.ClassName = 'TFhirTestScriptTestAction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptTestAction');
  add(value);
end;


function TFhirTestScriptTestActionList.Append: TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTestActionList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptTestActionList.GetEnumerator : TFhirTestScriptTestActionListEnumerator;
begin
  result := TFhirTestScriptTestActionListEnumerator.Create(self.link);
end;

function TFhirTestScriptTestActionList.Clone: TFhirTestScriptTestActionList;
begin
  result := TFhirTestScriptTestActionList(inherited Clone);
end;

function TFhirTestScriptTestActionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptTestActionList.GetItemN(index: Integer): TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction(ObjectByIndex[index]);
end;

function TFhirTestScriptTestActionList.IndexOf(value: TFhirTestScriptTestAction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptTestActionList.Insert(index: Integer): TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTestActionList.InsertItem(index: Integer; value: TFhirTestScriptTestAction);
begin
  assert(value is TFhirTestScriptTestAction);
  Inherited Insert(index, value);
end;

function TFhirTestScriptTestActionList.Item(index: Integer): TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction(ObjectByIndex[index]);
end;

function TFhirTestScriptTestActionList.Link: TFhirTestScriptTestActionList;
begin
  result := TFhirTestScriptTestActionList(inherited Link);
end;

procedure TFhirTestScriptTestActionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptTestActionList.SetItemByIndex(index: Integer; value: TFhirTestScriptTestAction);
begin
  assert(value is TFhirTestScriptTestAction);
  FhirTestScriptTestActions[index] := value;
end;

procedure TFhirTestScriptTestActionList.SetItemN(index: Integer; value: TFhirTestScriptTestAction);
begin
  assert(value is TFhirTestScriptTestAction);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptTeardown }

constructor TFhirTestScriptTeardown.Create;
begin
  inherited;
end;

destructor TFhirTestScriptTeardown.Destroy;
begin
  FActionList.Free;
  inherited;
end;

procedure TFhirTestScriptTeardown.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirTestScriptTeardown(oSource).FActionList = nil) then
  begin
    FActionList.free;
    FActionList := nil;
  end
  else
  begin
    FActionList := TFhirTestScriptTeardownActionList.Create;
    FActionList.Assign(TFhirTestScriptTeardown(oSource).FActionList);
  end;
end;

procedure TFhirTestScriptTeardown.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'action') Then
    list.addAll(FActionList);
end;

procedure TFhirTestScriptTeardown.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'action', '', FActionList.Link)){3};
end;

procedure TFhirTestScriptTeardown.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'action') then ActionList.add(propValue as TFhirTestScriptTeardownAction){2}
  else inherited;
end;

function TFhirTestScriptTeardown.FhirType : string;
begin
  result := 'teardown';
end;

function TFhirTestScriptTeardown.Link : TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown(inherited Link);
end;

function TFhirTestScriptTeardown.Clone : TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown(inherited Clone);
end;

function TFhirTestScriptTeardown.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTeardown;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptTeardown)) then
    result := false
  else
  begin
    o := TFhirTestScriptTeardown(other);
    result := compareDeep(actionList, o.actionList, true);
  end;
end;

function TFhirTestScriptTeardown.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTeardown;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptTeardown)) then
    result := false
  else
  begin
    o := TFhirTestScriptTeardown(other);
    result := true;
  end;
end;

{ TFhirTestScriptTeardown }

Function TFhirTestScriptTeardown.GetActionList : TFhirTestScriptTeardownActionList;
begin
  if FActionList = nil then
    FActionList := TFhirTestScriptTeardownActionList.Create;
  result := FActionList;
end;

Function TFhirTestScriptTeardown.GetHasActionList : boolean;
begin
  result := (FActionList <> nil) and (FActionList.count > 0);
end;


{ TFhirTestScriptTeardownListEnumerator }

Constructor TFhirTestScriptTeardownListEnumerator.Create(list : TFhirTestScriptTeardownList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptTeardownListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptTeardownListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptTeardownListEnumerator.GetCurrent : TFhirTestScriptTeardown;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptTeardownList }
procedure TFhirTestScriptTeardownList.AddItem(value: TFhirTestScriptTeardown);
begin
  assert(value.ClassName = 'TFhirTestScriptTeardown', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptTeardown');
  add(value);
end;


function TFhirTestScriptTeardownList.Append: TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTeardownList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptTeardownList.GetEnumerator : TFhirTestScriptTeardownListEnumerator;
begin
  result := TFhirTestScriptTeardownListEnumerator.Create(self.link);
end;

function TFhirTestScriptTeardownList.Clone: TFhirTestScriptTeardownList;
begin
  result := TFhirTestScriptTeardownList(inherited Clone);
end;

function TFhirTestScriptTeardownList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptTeardownList.GetItemN(index: Integer): TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown(ObjectByIndex[index]);
end;

function TFhirTestScriptTeardownList.IndexOf(value: TFhirTestScriptTeardown): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptTeardownList.Insert(index: Integer): TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTeardownList.InsertItem(index: Integer; value: TFhirTestScriptTeardown);
begin
  assert(value is TFhirTestScriptTeardown);
  Inherited Insert(index, value);
end;

function TFhirTestScriptTeardownList.Item(index: Integer): TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown(ObjectByIndex[index]);
end;

function TFhirTestScriptTeardownList.Link: TFhirTestScriptTeardownList;
begin
  result := TFhirTestScriptTeardownList(inherited Link);
end;

procedure TFhirTestScriptTeardownList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptTeardownList.SetItemByIndex(index: Integer; value: TFhirTestScriptTeardown);
begin
  assert(value is TFhirTestScriptTeardown);
  FhirTestScriptTeardowns[index] := value;
end;

procedure TFhirTestScriptTeardownList.SetItemN(index: Integer; value: TFhirTestScriptTeardown);
begin
  assert(value is TFhirTestScriptTeardown);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScriptTeardownAction }

constructor TFhirTestScriptTeardownAction.Create;
begin
  inherited;
end;

destructor TFhirTestScriptTeardownAction.Destroy;
begin
  FOperation.free;
  inherited;
end;

procedure TFhirTestScriptTeardownAction.Assign(oSource : TAdvObject);
begin
  inherited;
  operation := TFhirTestScriptTeardownAction(oSource).operation.Clone;
end;

procedure TFhirTestScriptTeardownAction.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'operation') Then
     list.add(FOperation.Link);
end;

procedure TFhirTestScriptTeardownAction.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'operation', '@TestScript.setup.action.operation', FOperation.Link));{2}
end;

procedure TFhirTestScriptTeardownAction.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'operation') then Operation := propValue as TFhirTestScriptSetupActionOperation{4b}
  else inherited;
end;

function TFhirTestScriptTeardownAction.FhirType : string;
begin
  result := 'action';
end;

function TFhirTestScriptTeardownAction.Link : TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction(inherited Link);
end;

function TFhirTestScriptTeardownAction.Clone : TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction(inherited Clone);
end;

function TFhirTestScriptTeardownAction.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTeardownAction;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirTestScriptTeardownAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptTeardownAction(other);
    result := compareDeep(operationElement, o.operationElement, true);
  end;
end;

function TFhirTestScriptTeardownAction.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirTestScriptTeardownAction;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirTestScriptTeardownAction)) then
    result := false
  else
  begin
    o := TFhirTestScriptTeardownAction(other);
    result := true;
  end;
end;

{ TFhirTestScriptTeardownAction }

Procedure TFhirTestScriptTeardownAction.SetOperation(value : TFhirTestScriptSetupActionOperation);
begin
  FOperation.free;
  FOperation := value;
end;


{ TFhirTestScriptTeardownActionListEnumerator }

Constructor TFhirTestScriptTeardownActionListEnumerator.Create(list : TFhirTestScriptTeardownActionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptTeardownActionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptTeardownActionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptTeardownActionListEnumerator.GetCurrent : TFhirTestScriptTeardownAction;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptTeardownActionList }
procedure TFhirTestScriptTeardownActionList.AddItem(value: TFhirTestScriptTeardownAction);
begin
  assert(value.ClassName = 'TFhirTestScriptTeardownAction', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScriptTeardownAction');
  add(value);
end;


function TFhirTestScriptTeardownActionList.Append: TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTeardownActionList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptTeardownActionList.GetEnumerator : TFhirTestScriptTeardownActionListEnumerator;
begin
  result := TFhirTestScriptTeardownActionListEnumerator.Create(self.link);
end;

function TFhirTestScriptTeardownActionList.Clone: TFhirTestScriptTeardownActionList;
begin
  result := TFhirTestScriptTeardownActionList(inherited Clone);
end;

function TFhirTestScriptTeardownActionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptTeardownActionList.GetItemN(index: Integer): TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction(ObjectByIndex[index]);
end;

function TFhirTestScriptTeardownActionList.IndexOf(value: TFhirTestScriptTeardownAction): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptTeardownActionList.Insert(index: Integer): TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptTeardownActionList.InsertItem(index: Integer; value: TFhirTestScriptTeardownAction);
begin
  assert(value is TFhirTestScriptTeardownAction);
  Inherited Insert(index, value);
end;

function TFhirTestScriptTeardownActionList.Item(index: Integer): TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction(ObjectByIndex[index]);
end;

function TFhirTestScriptTeardownActionList.Link: TFhirTestScriptTeardownActionList;
begin
  result := TFhirTestScriptTeardownActionList(inherited Link);
end;

procedure TFhirTestScriptTeardownActionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptTeardownActionList.SetItemByIndex(index: Integer; value: TFhirTestScriptTeardownAction);
begin
  assert(value is TFhirTestScriptTeardownAction);
  FhirTestScriptTeardownActions[index] := value;
end;

procedure TFhirTestScriptTeardownActionList.SetItemN(index: Integer; value: TFhirTestScriptTeardownAction);
begin
  assert(value is TFhirTestScriptTeardownAction);
  ObjectByIndex[index] := value;
end;

{ TFhirTestScript }

constructor TFhirTestScript.Create;
begin
  inherited;
end;

destructor TFhirTestScript.Destroy;
begin
  FUrl.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FIdentifier.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FDescription.free;
  FUseContextList.Free;
  FRequirements.free;
  FCopyright.free;
  FMetadata.free;
  FMultiserver.free;
  FFixtureList.Free;
  FProfileList.Free;
  FVariableList.Free;
  FSetup.free;
  FTestList.Free;
  FTeardown.free;
  inherited;
end;

function TFhirTestScript.GetResourceType : TFhirResourceType;
begin
  result := frtTestScript;
end;

procedure TFhirTestScript.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirTestScript(oSource).urlElement.Clone;
  versionElement := TFhirTestScript(oSource).versionElement.Clone;
  nameElement := TFhirTestScript(oSource).nameElement.Clone;
  FStatus := TFhirTestScript(oSource).FStatus.Link;
  identifier := TFhirTestScript(oSource).identifier.Clone;
  experimentalElement := TFhirTestScript(oSource).experimentalElement.Clone;
  publisherElement := TFhirTestScript(oSource).publisherElement.Clone;
  if (TFhirTestScript(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirTestScriptContactList.Create;
    FContactList.Assign(TFhirTestScript(oSource).FContactList);
  end;
  dateElement := TFhirTestScript(oSource).dateElement.Clone;
  descriptionElement := TFhirTestScript(oSource).descriptionElement.Clone;
  if (TFhirTestScript(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirTestScript(oSource).FUseContextList);
  end;
  requirementsElement := TFhirTestScript(oSource).requirementsElement.Clone;
  copyrightElement := TFhirTestScript(oSource).copyrightElement.Clone;
  metadata := TFhirTestScript(oSource).metadata.Clone;
  multiserverElement := TFhirTestScript(oSource).multiserverElement.Clone;
  if (TFhirTestScript(oSource).FFixtureList = nil) then
  begin
    FFixtureList.free;
    FFixtureList := nil;
  end
  else
  begin
    FFixtureList := TFhirTestScriptFixtureList.Create;
    FFixtureList.Assign(TFhirTestScript(oSource).FFixtureList);
  end;
  if (TFhirTestScript(oSource).FProfileList = nil) then
  begin
    FProfileList.free;
    FProfileList := nil;
  end
  else
  begin
    FProfileList := TFhirReferenceList{TFhirReference}.Create;
    FProfileList.Assign(TFhirTestScript(oSource).FProfileList);
  end;
  if (TFhirTestScript(oSource).FVariableList = nil) then
  begin
    FVariableList.free;
    FVariableList := nil;
  end
  else
  begin
    FVariableList := TFhirTestScriptVariableList.Create;
    FVariableList.Assign(TFhirTestScript(oSource).FVariableList);
  end;
  setup := TFhirTestScript(oSource).setup.Clone;
  if (TFhirTestScript(oSource).FTestList = nil) then
  begin
    FTestList.free;
    FTestList := nil;
  end
  else
  begin
    FTestList := TFhirTestScriptTestList.Create;
    FTestList.Assign(TFhirTestScript(oSource).FTestList);
  end;
  teardown := TFhirTestScript(oSource).teardown.Clone;
end;

procedure TFhirTestScript.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'metadata') Then
     list.add(FMetadata.Link);
  if (child_name = 'multiserver') Then
     list.add(FMultiserver.Link);
  if (child_name = 'fixture') Then
    list.addAll(FFixtureList);
  if (child_name = 'profile') Then
    list.addAll(FProfileList);
  if (child_name = 'variable') Then
    list.addAll(FVariableList);
  if (child_name = 'setup') Then
     list.add(FSetup.Link);
  if (child_name = 'test') Then
    list.addAll(FTestList);
  if (child_name = 'teardown') Then
     list.add(FTeardown.Link);
end;

procedure TFhirTestScript.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'metadata', '', FMetadata.Link));{2}
  oList.add(TFHIRProperty.create(self, 'multiserver', 'boolean', FMultiserver.Link));{2}
  oList.add(TFHIRProperty.create(self, 'fixture', '', FFixtureList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'profile', 'Reference(Any)', FProfileList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'variable', '', FVariableList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'setup', '', FSetup.Link));{2}
  oList.add(TFHIRProperty.create(self, 'test', '', FTestList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'teardown', '', FTeardown.Link));{2}
end;

procedure TFhirTestScript.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirTestScriptContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'metadata') then Metadata := propValue as TFhirTestScriptMetadata{4b}
  else if (propName = 'multiserver') then MultiserverElement := propValue as TFhirBoolean{5a}
  else if (propName = 'fixture') then FixtureList.add(propValue as TFhirTestScriptFixture){2}
  else if (propName = 'profile') then ProfileList.add(propValue as TFhirReference{TFhirReference}){2}
  else if (propName = 'variable') then VariableList.add(propValue as TFhirTestScriptVariable){2}
  else if (propName = 'setup') then Setup := propValue as TFhirTestScriptSetup{4b}
  else if (propName = 'test') then TestList.add(propValue as TFhirTestScriptTest){2}
  else if (propName = 'teardown') then Teardown := propValue as TFhirTestScriptTeardown{4b}
  else inherited;
end;

function TFhirTestScript.FhirType : string;
begin
  result := 'TestScript';
end;

function TFhirTestScript.Link : TFhirTestScript;
begin
  result := TFhirTestScript(inherited Link);
end;

function TFhirTestScript.Clone : TFhirTestScript;
begin
  result := TFhirTestScript(inherited Clone);
end;

{ TFhirTestScript }

Procedure TFhirTestScript.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirTestScript.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirTestScript.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirTestScript.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirTestScript.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirTestScript.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirTestScript.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirTestScript.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirTestScript.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirTestScript.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirTestScript.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirTestScript.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirTestScript.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirTestScript.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirTestScript.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirTestScript.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirTestScript.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirTestScript.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirTestScript.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirTestScript.GetContactList : TFhirTestScriptContactList;
begin
  if FContactList = nil then
    FContactList := TFhirTestScriptContactList.Create;
  result := FContactList;
end;

Function TFhirTestScript.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirTestScript.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirTestScript.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirTestScript.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirTestScript.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirTestScript.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirTestScript.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirTestScript.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirTestScript.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirTestScript.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirTestScript.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirTestScript.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirTestScript.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirTestScript.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirTestScript.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirTestScript.SetMetadata(value : TFhirTestScriptMetadata);
begin
  FMetadata.free;
  FMetadata := value;
end;

Procedure TFhirTestScript.SetMultiserver(value : TFhirBoolean);
begin
  FMultiserver.free;
  FMultiserver := value;
end;

Function TFhirTestScript.GetMultiserverST : Boolean;
begin
  if FMultiserver = nil then
    result := false
  else
    result := FMultiserver.value;
end;

Procedure TFhirTestScript.SetMultiserverST(value : Boolean);
begin
  if FMultiserver = nil then
    FMultiserver := TFhirBoolean.create;
  FMultiserver.value := value
end;

Function TFhirTestScript.GetFixtureList : TFhirTestScriptFixtureList;
begin
  if FFixtureList = nil then
    FFixtureList := TFhirTestScriptFixtureList.Create;
  result := FFixtureList;
end;

Function TFhirTestScript.GetHasFixtureList : boolean;
begin
  result := (FFixtureList <> nil) and (FFixtureList.count > 0);
end;

Function TFhirTestScript.GetProfileList : TFhirReferenceList{TFhirReference};
begin
  if FProfileList = nil then
    FProfileList := TFhirReferenceList{TFhirReference}.Create;
  result := FProfileList;
end;

Function TFhirTestScript.GetHasProfileList : boolean;
begin
  result := (FProfileList <> nil) and (FProfileList.count > 0);
end;

Function TFhirTestScript.GetVariableList : TFhirTestScriptVariableList;
begin
  if FVariableList = nil then
    FVariableList := TFhirTestScriptVariableList.Create;
  result := FVariableList;
end;

Function TFhirTestScript.GetHasVariableList : boolean;
begin
  result := (FVariableList <> nil) and (FVariableList.count > 0);
end;

Procedure TFhirTestScript.SetSetup(value : TFhirTestScriptSetup);
begin
  FSetup.free;
  FSetup := value;
end;

Function TFhirTestScript.GetTestList : TFhirTestScriptTestList;
begin
  if FTestList = nil then
    FTestList := TFhirTestScriptTestList.Create;
  result := FTestList;
end;

Function TFhirTestScript.GetHasTestList : boolean;
begin
  result := (FTestList <> nil) and (FTestList.count > 0);
end;

Procedure TFhirTestScript.SetTeardown(value : TFhirTestScriptTeardown);
begin
  FTeardown.free;
  FTeardown := value;
end;


{ TFhirTestScriptListEnumerator }

Constructor TFhirTestScriptListEnumerator.Create(list : TFhirTestScriptList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirTestScriptListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirTestScriptListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirTestScriptListEnumerator.GetCurrent : TFhirTestScript;
begin
  Result := FList[FIndex];
end;


{ TFhirTestScriptList }
procedure TFhirTestScriptList.AddItem(value: TFhirTestScript);
begin
  assert(value.ClassName = 'TFhirTestScript', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirTestScript');
  add(value);
end;


function TFhirTestScriptList.Append: TFhirTestScript;
begin
  result := TFhirTestScript.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptList.ClearItems;
begin
  Clear;
end;

function TFhirTestScriptList.GetEnumerator : TFhirTestScriptListEnumerator;
begin
  result := TFhirTestScriptListEnumerator.Create(self.link);
end;

function TFhirTestScriptList.Clone: TFhirTestScriptList;
begin
  result := TFhirTestScriptList(inherited Clone);
end;

function TFhirTestScriptList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirTestScriptList.GetItemN(index: Integer): TFhirTestScript;
begin
  result := TFhirTestScript(ObjectByIndex[index]);
end;

function TFhirTestScriptList.IndexOf(value: TFhirTestScript): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirTestScriptList.Insert(index: Integer): TFhirTestScript;
begin
  result := TFhirTestScript.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirTestScriptList.InsertItem(index: Integer; value: TFhirTestScript);
begin
  assert(value is TFhirTestScript);
  Inherited Insert(index, value);
end;

function TFhirTestScriptList.Item(index: Integer): TFhirTestScript;
begin
  result := TFhirTestScript(ObjectByIndex[index]);
end;

function TFhirTestScriptList.Link: TFhirTestScriptList;
begin
  result := TFhirTestScriptList(inherited Link);
end;

procedure TFhirTestScriptList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirTestScriptList.SetItemByIndex(index: Integer; value: TFhirTestScript);
begin
  assert(value is TFhirTestScript);
  FhirTestScripts[index] := value;
end;

procedure TFhirTestScriptList.SetItemN(index: Integer; value: TFhirTestScript);
begin
  assert(value is TFhirTestScript);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetContact }

constructor TFhirValueSetContact.Create;
begin
  inherited;
end;

destructor TFhirValueSetContact.Destroy;
begin
  FName.free;
  FTelecomList.Free;
  inherited;
end;

procedure TFhirValueSetContact.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirValueSetContact(oSource).nameElement.Clone;
  if (TFhirValueSetContact(oSource).FTelecomList = nil) then
  begin
    FTelecomList.free;
    FTelecomList := nil;
  end
  else
  begin
    FTelecomList := TFhirContactPointList.Create;
    FTelecomList.Assign(TFhirValueSetContact(oSource).FTelecomList);
  end;
end;

procedure TFhirValueSetContact.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'telecom') Then
    list.addAll(FTelecomList);
end;

procedure TFhirValueSetContact.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'telecom', 'ContactPoint', FTelecomList.Link)){3};
end;

procedure TFhirValueSetContact.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'telecom') then TelecomList.add(propValue as TFhirContactPoint){2}
  else inherited;
end;

function TFhirValueSetContact.FhirType : string;
begin
  result := 'contact';
end;

function TFhirValueSetContact.Link : TFhirValueSetContact;
begin
  result := TFhirValueSetContact(inherited Link);
end;

function TFhirValueSetContact.Clone : TFhirValueSetContact;
begin
  result := TFhirValueSetContact(inherited Clone);
end;

function TFhirValueSetContact.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetContact;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetContact)) then
    result := false
  else
  begin
    o := TFhirValueSetContact(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(telecomList, o.telecomList, true);
  end;
end;

function TFhirValueSetContact.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetContact;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetContact)) then
    result := false
  else
  begin
    o := TFhirValueSetContact(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirValueSetContact }

Procedure TFhirValueSetContact.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirValueSetContact.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirValueSetContact.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Function TFhirValueSetContact.GetTelecomList : TFhirContactPointList;
begin
  if FTelecomList = nil then
    FTelecomList := TFhirContactPointList.Create;
  result := FTelecomList;
end;

Function TFhirValueSetContact.GetHasTelecomList : boolean;
begin
  result := (FTelecomList <> nil) and (FTelecomList.count > 0);
end;


{ TFhirValueSetContactListEnumerator }

Constructor TFhirValueSetContactListEnumerator.Create(list : TFhirValueSetContactList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetContactListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetContactListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetContactListEnumerator.GetCurrent : TFhirValueSetContact;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetContactList }
procedure TFhirValueSetContactList.AddItem(value: TFhirValueSetContact);
begin
  assert(value.ClassName = 'TFhirValueSetContact', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetContact');
  add(value);
end;


function TFhirValueSetContactList.Append: TFhirValueSetContact;
begin
  result := TFhirValueSetContact.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetContactList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetContactList.GetEnumerator : TFhirValueSetContactListEnumerator;
begin
  result := TFhirValueSetContactListEnumerator.Create(self.link);
end;

function TFhirValueSetContactList.Clone: TFhirValueSetContactList;
begin
  result := TFhirValueSetContactList(inherited Clone);
end;

function TFhirValueSetContactList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetContactList.GetItemN(index: Integer): TFhirValueSetContact;
begin
  result := TFhirValueSetContact(ObjectByIndex[index]);
end;

function TFhirValueSetContactList.IndexOf(value: TFhirValueSetContact): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetContactList.Insert(index: Integer): TFhirValueSetContact;
begin
  result := TFhirValueSetContact.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetContactList.InsertItem(index: Integer; value: TFhirValueSetContact);
begin
  assert(value is TFhirValueSetContact);
  Inherited Insert(index, value);
end;

function TFhirValueSetContactList.Item(index: Integer): TFhirValueSetContact;
begin
  result := TFhirValueSetContact(ObjectByIndex[index]);
end;

function TFhirValueSetContactList.Link: TFhirValueSetContactList;
begin
  result := TFhirValueSetContactList(inherited Link);
end;

procedure TFhirValueSetContactList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetContactList.SetItemByIndex(index: Integer; value: TFhirValueSetContact);
begin
  assert(value is TFhirValueSetContact);
  FhirValueSetContacts[index] := value;
end;

procedure TFhirValueSetContactList.SetItemN(index: Integer; value: TFhirValueSetContact);
begin
  assert(value is TFhirValueSetContact);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetCodeSystem }

constructor TFhirValueSetCodeSystem.Create;
begin
  inherited;
end;

destructor TFhirValueSetCodeSystem.Destroy;
begin
  FSystem.free;
  FVersion.free;
  FCaseSensitive.free;
  FConceptList.Free;
  inherited;
end;

procedure TFhirValueSetCodeSystem.Assign(oSource : TAdvObject);
begin
  inherited;
  systemElement := TFhirValueSetCodeSystem(oSource).systemElement.Clone;
  versionElement := TFhirValueSetCodeSystem(oSource).versionElement.Clone;
  caseSensitiveElement := TFhirValueSetCodeSystem(oSource).caseSensitiveElement.Clone;
  if (TFhirValueSetCodeSystem(oSource).FConceptList = nil) then
  begin
    FConceptList.free;
    FConceptList := nil;
  end
  else
  begin
    FConceptList := TFhirValueSetCodeSystemConceptList.Create;
    FConceptList.Assign(TFhirValueSetCodeSystem(oSource).FConceptList);
  end;
end;

procedure TFhirValueSetCodeSystem.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'system') Then
     list.add(FSystem.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'caseSensitive') Then
     list.add(FCaseSensitive.Link);
  if (child_name = 'concept') Then
    list.addAll(FConceptList);
end;

procedure TFhirValueSetCodeSystem.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'system', 'uri', FSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'caseSensitive', 'boolean', FCaseSensitive.Link));{2}
  oList.add(TFHIRProperty.create(self, 'concept', '', FConceptList.Link)){3};
end;

procedure TFhirValueSetCodeSystem.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'system') then SystemElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'caseSensitive') then CaseSensitiveElement := propValue as TFhirBoolean{5a}
  else if (propName = 'concept') then ConceptList.add(propValue as TFhirValueSetCodeSystemConcept){2}
  else inherited;
end;

function TFhirValueSetCodeSystem.FhirType : string;
begin
  result := 'codeSystem';
end;

function TFhirValueSetCodeSystem.Link : TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem(inherited Link);
end;

function TFhirValueSetCodeSystem.Clone : TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem(inherited Clone);
end;

function TFhirValueSetCodeSystem.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystem;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystem)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystem(other);
    result := compareDeep(systemElement, o.systemElement, true) and compareDeep(versionElement, o.versionElement, true) and 
      compareDeep(caseSensitiveElement, o.caseSensitiveElement, true) and compareDeep(conceptList, o.conceptList, true);
  end;
end;

function TFhirValueSetCodeSystem.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystem;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystem)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystem(other);
    result := compareValues(systemElement, o.systemElement, true) and compareValues(versionElement, o.versionElement, true) and 
      compareValues(caseSensitiveElement, o.caseSensitiveElement, true);
  end;
end;

{ TFhirValueSetCodeSystem }

Procedure TFhirValueSetCodeSystem.SetSystem(value : TFhirUri);
begin
  FSystem.free;
  FSystem := value;
end;

Function TFhirValueSetCodeSystem.GetSystemST : String;
begin
  if FSystem = nil then
    result := ''
  else
    result := FSystem.value;
end;

Procedure TFhirValueSetCodeSystem.SetSystemST(value : String);
begin
  if value <> '' then
  begin
    if FSystem = nil then
      FSystem := TFhirUri.create;
    FSystem.value := value
  end
  else if FSystem <> nil then
    FSystem.value := '';
end;

Procedure TFhirValueSetCodeSystem.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirValueSetCodeSystem.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirValueSetCodeSystem.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirValueSetCodeSystem.SetCaseSensitive(value : TFhirBoolean);
begin
  FCaseSensitive.free;
  FCaseSensitive := value;
end;

Function TFhirValueSetCodeSystem.GetCaseSensitiveST : Boolean;
begin
  if FCaseSensitive = nil then
    result := false
  else
    result := FCaseSensitive.value;
end;

Procedure TFhirValueSetCodeSystem.SetCaseSensitiveST(value : Boolean);
begin
  if FCaseSensitive = nil then
    FCaseSensitive := TFhirBoolean.create;
  FCaseSensitive.value := value
end;

Function TFhirValueSetCodeSystem.GetConceptList : TFhirValueSetCodeSystemConceptList;
begin
  if FConceptList = nil then
    FConceptList := TFhirValueSetCodeSystemConceptList.Create;
  result := FConceptList;
end;

Function TFhirValueSetCodeSystem.GetHasConceptList : boolean;
begin
  result := (FConceptList <> nil) and (FConceptList.count > 0);
end;


{ TFhirValueSetCodeSystemListEnumerator }

Constructor TFhirValueSetCodeSystemListEnumerator.Create(list : TFhirValueSetCodeSystemList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetCodeSystemListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetCodeSystemListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetCodeSystemListEnumerator.GetCurrent : TFhirValueSetCodeSystem;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetCodeSystemList }
procedure TFhirValueSetCodeSystemList.AddItem(value: TFhirValueSetCodeSystem);
begin
  assert(value.ClassName = 'TFhirValueSetCodeSystem', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetCodeSystem');
  add(value);
end;


function TFhirValueSetCodeSystemList.Append: TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetCodeSystemList.GetEnumerator : TFhirValueSetCodeSystemListEnumerator;
begin
  result := TFhirValueSetCodeSystemListEnumerator.Create(self.link);
end;

function TFhirValueSetCodeSystemList.Clone: TFhirValueSetCodeSystemList;
begin
  result := TFhirValueSetCodeSystemList(inherited Clone);
end;

function TFhirValueSetCodeSystemList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetCodeSystemList.GetItemN(index: Integer): TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemList.IndexOf(value: TFhirValueSetCodeSystem): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetCodeSystemList.Insert(index: Integer): TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemList.InsertItem(index: Integer; value: TFhirValueSetCodeSystem);
begin
  assert(value is TFhirValueSetCodeSystem);
  Inherited Insert(index, value);
end;

function TFhirValueSetCodeSystemList.Item(index: Integer): TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemList.Link: TFhirValueSetCodeSystemList;
begin
  result := TFhirValueSetCodeSystemList(inherited Link);
end;

procedure TFhirValueSetCodeSystemList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetCodeSystemList.SetItemByIndex(index: Integer; value: TFhirValueSetCodeSystem);
begin
  assert(value is TFhirValueSetCodeSystem);
  FhirValueSetCodeSystems[index] := value;
end;

procedure TFhirValueSetCodeSystemList.SetItemN(index: Integer; value: TFhirValueSetCodeSystem);
begin
  assert(value is TFhirValueSetCodeSystem);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetCodeSystemConcept }

constructor TFhirValueSetCodeSystemConcept.Create;
begin
  inherited;
end;

destructor TFhirValueSetCodeSystemConcept.Destroy;
begin
  FCode.free;
  FAbstract.free;
  FDisplay.free;
  FDefinition.free;
  FDesignationList.Free;
  FConceptList.Free;
  inherited;
end;

procedure TFhirValueSetCodeSystemConcept.Assign(oSource : TAdvObject);
begin
  inherited;
  codeElement := TFhirValueSetCodeSystemConcept(oSource).codeElement.Clone;
  abstractElement := TFhirValueSetCodeSystemConcept(oSource).abstractElement.Clone;
  displayElement := TFhirValueSetCodeSystemConcept(oSource).displayElement.Clone;
  definitionElement := TFhirValueSetCodeSystemConcept(oSource).definitionElement.Clone;
  if (TFhirValueSetCodeSystemConcept(oSource).FDesignationList = nil) then
  begin
    FDesignationList.free;
    FDesignationList := nil;
  end
  else
  begin
    FDesignationList := TFhirValueSetCodeSystemConceptDesignationList.Create;
    FDesignationList.Assign(TFhirValueSetCodeSystemConcept(oSource).FDesignationList);
  end;
  if (TFhirValueSetCodeSystemConcept(oSource).FConceptList = nil) then
  begin
    FConceptList.free;
    FConceptList := nil;
  end
  else
  begin
    FConceptList := TFhirValueSetCodeSystemConceptList.Create;
    FConceptList.Assign(TFhirValueSetCodeSystemConcept(oSource).FConceptList);
  end;
end;

procedure TFhirValueSetCodeSystemConcept.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'abstract') Then
     list.add(FAbstract.Link);
  if (child_name = 'display') Then
     list.add(FDisplay.Link);
  if (child_name = 'definition') Then
     list.add(FDefinition.Link);
  if (child_name = 'designation') Then
    list.addAll(FDesignationList);
  if (child_name = 'concept') Then
    list.addAll(FConceptList);
end;

procedure TFhirValueSetCodeSystemConcept.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'abstract', 'boolean', FAbstract.Link));{2}
  oList.add(TFHIRProperty.create(self, 'display', 'string', FDisplay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'definition', 'string', FDefinition.Link));{2}
  oList.add(TFHIRProperty.create(self, 'designation', '', FDesignationList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'concept', '@ValueSet.codeSystem.concept', FConceptList.Link)){3};
end;

procedure TFhirValueSetCodeSystemConcept.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'abstract') then AbstractElement := propValue as TFhirBoolean{5a}
  else if (propName = 'display') then DisplayElement := propValue as TFhirString{5a}
  else if (propName = 'definition') then DefinitionElement := propValue as TFhirString{5a}
  else if (propName = 'designation') then DesignationList.add(propValue as TFhirValueSetCodeSystemConceptDesignation){2}
  else if (propName = 'concept') then ConceptList.add(propValue as TFhirValueSetCodeSystemConcept){2}
  else inherited;
end;

function TFhirValueSetCodeSystemConcept.FhirType : string;
begin
  result := 'concept';
end;

function TFhirValueSetCodeSystemConcept.Link : TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept(inherited Link);
end;

function TFhirValueSetCodeSystemConcept.Clone : TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept(inherited Clone);
end;

function TFhirValueSetCodeSystemConcept.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystemConcept;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystemConcept)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystemConcept(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(abstractElement, o.abstractElement, true) and 
      compareDeep(displayElement, o.displayElement, true) and compareDeep(definitionElement, o.definitionElement, true) and 
      compareDeep(designationList, o.designationList, true) and compareDeep(conceptList, o.conceptList, true);
  end;
end;

function TFhirValueSetCodeSystemConcept.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystemConcept;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystemConcept)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystemConcept(other);
    result := compareValues(codeElement, o.codeElement, true) and compareValues(abstractElement, o.abstractElement, true) and 
      compareValues(displayElement, o.displayElement, true) and compareValues(definitionElement, o.definitionElement, true);
  end;
end;

{ TFhirValueSetCodeSystemConcept }

Procedure TFhirValueSetCodeSystemConcept.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirValueSetCodeSystemConcept.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirValueSetCodeSystemConcept.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirValueSetCodeSystemConcept.SetAbstract(value : TFhirBoolean);
begin
  FAbstract.free;
  FAbstract := value;
end;

Function TFhirValueSetCodeSystemConcept.GetAbstractST : Boolean;
begin
  if FAbstract = nil then
    result := false
  else
    result := FAbstract.value;
end;

Procedure TFhirValueSetCodeSystemConcept.SetAbstractST(value : Boolean);
begin
  if FAbstract = nil then
    FAbstract := TFhirBoolean.create;
  FAbstract.value := value
end;

Procedure TFhirValueSetCodeSystemConcept.SetDisplay(value : TFhirString);
begin
  FDisplay.free;
  FDisplay := value;
end;

Function TFhirValueSetCodeSystemConcept.GetDisplayST : String;
begin
  if FDisplay = nil then
    result := ''
  else
    result := FDisplay.value;
end;

Procedure TFhirValueSetCodeSystemConcept.SetDisplayST(value : String);
begin
  if value <> '' then
  begin
    if FDisplay = nil then
      FDisplay := TFhirString.create;
    FDisplay.value := value
  end
  else if FDisplay <> nil then
    FDisplay.value := '';
end;

Procedure TFhirValueSetCodeSystemConcept.SetDefinition(value : TFhirString);
begin
  FDefinition.free;
  FDefinition := value;
end;

Function TFhirValueSetCodeSystemConcept.GetDefinitionST : String;
begin
  if FDefinition = nil then
    result := ''
  else
    result := FDefinition.value;
end;

Procedure TFhirValueSetCodeSystemConcept.SetDefinitionST(value : String);
begin
  if value <> '' then
  begin
    if FDefinition = nil then
      FDefinition := TFhirString.create;
    FDefinition.value := value
  end
  else if FDefinition <> nil then
    FDefinition.value := '';
end;

Function TFhirValueSetCodeSystemConcept.GetDesignationList : TFhirValueSetCodeSystemConceptDesignationList;
begin
  if FDesignationList = nil then
    FDesignationList := TFhirValueSetCodeSystemConceptDesignationList.Create;
  result := FDesignationList;
end;

Function TFhirValueSetCodeSystemConcept.GetHasDesignationList : boolean;
begin
  result := (FDesignationList <> nil) and (FDesignationList.count > 0);
end;

Function TFhirValueSetCodeSystemConcept.GetConceptList : TFhirValueSetCodeSystemConceptList;
begin
  if FConceptList = nil then
    FConceptList := TFhirValueSetCodeSystemConceptList.Create;
  result := FConceptList;
end;

Function TFhirValueSetCodeSystemConcept.GetHasConceptList : boolean;
begin
  result := (FConceptList <> nil) and (FConceptList.count > 0);
end;


{ TFhirValueSetCodeSystemConceptListEnumerator }

Constructor TFhirValueSetCodeSystemConceptListEnumerator.Create(list : TFhirValueSetCodeSystemConceptList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetCodeSystemConceptListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetCodeSystemConceptListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetCodeSystemConceptListEnumerator.GetCurrent : TFhirValueSetCodeSystemConcept;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetCodeSystemConceptList }
procedure TFhirValueSetCodeSystemConceptList.AddItem(value: TFhirValueSetCodeSystemConcept);
begin
  assert(value.ClassName = 'TFhirValueSetCodeSystemConcept', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetCodeSystemConcept');
  add(value);
end;


function TFhirValueSetCodeSystemConceptList.Append: TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemConceptList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetCodeSystemConceptList.GetEnumerator : TFhirValueSetCodeSystemConceptListEnumerator;
begin
  result := TFhirValueSetCodeSystemConceptListEnumerator.Create(self.link);
end;

function TFhirValueSetCodeSystemConceptList.Clone: TFhirValueSetCodeSystemConceptList;
begin
  result := TFhirValueSetCodeSystemConceptList(inherited Clone);
end;

function TFhirValueSetCodeSystemConceptList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetCodeSystemConceptList.GetItemN(index: Integer): TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemConceptList.IndexOf(value: TFhirValueSetCodeSystemConcept): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetCodeSystemConceptList.Insert(index: Integer): TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemConceptList.InsertItem(index: Integer; value: TFhirValueSetCodeSystemConcept);
begin
  assert(value is TFhirValueSetCodeSystemConcept);
  Inherited Insert(index, value);
end;

function TFhirValueSetCodeSystemConceptList.Item(index: Integer): TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemConceptList.Link: TFhirValueSetCodeSystemConceptList;
begin
  result := TFhirValueSetCodeSystemConceptList(inherited Link);
end;

procedure TFhirValueSetCodeSystemConceptList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetCodeSystemConceptList.SetItemByIndex(index: Integer; value: TFhirValueSetCodeSystemConcept);
begin
  assert(value is TFhirValueSetCodeSystemConcept);
  FhirValueSetCodeSystemConcepts[index] := value;
end;

procedure TFhirValueSetCodeSystemConceptList.SetItemN(index: Integer; value: TFhirValueSetCodeSystemConcept);
begin
  assert(value is TFhirValueSetCodeSystemConcept);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetCodeSystemConceptDesignation }

constructor TFhirValueSetCodeSystemConceptDesignation.Create;
begin
  inherited;
end;

destructor TFhirValueSetCodeSystemConceptDesignation.Destroy;
begin
  FLanguage.free;
  FUse.free;
  FValue.free;
  inherited;
end;

procedure TFhirValueSetCodeSystemConceptDesignation.Assign(oSource : TAdvObject);
begin
  inherited;
  languageElement := TFhirValueSetCodeSystemConceptDesignation(oSource).languageElement.Clone;
  use := TFhirValueSetCodeSystemConceptDesignation(oSource).use.Clone;
  valueElement := TFhirValueSetCodeSystemConceptDesignation(oSource).valueElement.Clone;
end;

procedure TFhirValueSetCodeSystemConceptDesignation.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'language') Then
     list.add(FLanguage.Link);
  if (child_name = 'use') Then
     list.add(FUse.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirValueSetCodeSystemConceptDesignation.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'language', 'code', FLanguage.Link));{2}
  oList.add(TFHIRProperty.create(self, 'use', 'Coding', FUse.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value', 'string', FValue.Link));{2}
end;

procedure TFhirValueSetCodeSystemConceptDesignation.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'language') then
    if propValue is TFHIRCode then
      LanguageElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      LanguageElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'use') then Use := propValue as TFhirCoding{4b}
  else if (propName = 'value') then ValueElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirValueSetCodeSystemConceptDesignation.FhirType : string;
begin
  result := 'designation';
end;

function TFhirValueSetCodeSystemConceptDesignation.Link : TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation(inherited Link);
end;

function TFhirValueSetCodeSystemConceptDesignation.Clone : TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation(inherited Clone);
end;

function TFhirValueSetCodeSystemConceptDesignation.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystemConceptDesignation;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystemConceptDesignation)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystemConceptDesignation(other);
    result := compareDeep(languageElement, o.languageElement, true) and compareDeep(useElement, o.useElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirValueSetCodeSystemConceptDesignation.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCodeSystemConceptDesignation;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetCodeSystemConceptDesignation)) then
    result := false
  else
  begin
    o := TFhirValueSetCodeSystemConceptDesignation(other);
    result := compareValues(languageElement, o.languageElement, true) and compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirValueSetCodeSystemConceptDesignation }

Procedure TFhirValueSetCodeSystemConceptDesignation.SetLanguage(value : TFhirCode);
begin
  FLanguage.free;
  FLanguage := value;
end;

Function TFhirValueSetCodeSystemConceptDesignation.GetLanguageST : String;
begin
  if FLanguage = nil then
    result := ''
  else
    result := FLanguage.value;
end;

Procedure TFhirValueSetCodeSystemConceptDesignation.SetLanguageST(value : String);
begin
  if value <> '' then
  begin
    if FLanguage = nil then
      FLanguage := TFhirCode.create;
    FLanguage.value := value
  end
  else if FLanguage <> nil then
    FLanguage.value := '';
end;

Procedure TFhirValueSetCodeSystemConceptDesignation.SetUse(value : TFhirCoding);
begin
  FUse.free;
  FUse := value;
end;

Procedure TFhirValueSetCodeSystemConceptDesignation.SetValue(value : TFhirString);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirValueSetCodeSystemConceptDesignation.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirValueSetCodeSystemConceptDesignation.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirString.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirValueSetCodeSystemConceptDesignationListEnumerator }

Constructor TFhirValueSetCodeSystemConceptDesignationListEnumerator.Create(list : TFhirValueSetCodeSystemConceptDesignationList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetCodeSystemConceptDesignationListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetCodeSystemConceptDesignationListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetCodeSystemConceptDesignationListEnumerator.GetCurrent : TFhirValueSetCodeSystemConceptDesignation;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetCodeSystemConceptDesignationList }
procedure TFhirValueSetCodeSystemConceptDesignationList.AddItem(value: TFhirValueSetCodeSystemConceptDesignation);
begin
  assert(value.ClassName = 'TFhirValueSetCodeSystemConceptDesignation', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetCodeSystemConceptDesignation');
  add(value);
end;


function TFhirValueSetCodeSystemConceptDesignationList.Append: TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemConceptDesignationList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetCodeSystemConceptDesignationList.GetEnumerator : TFhirValueSetCodeSystemConceptDesignationListEnumerator;
begin
  result := TFhirValueSetCodeSystemConceptDesignationListEnumerator.Create(self.link);
end;

function TFhirValueSetCodeSystemConceptDesignationList.Clone: TFhirValueSetCodeSystemConceptDesignationList;
begin
  result := TFhirValueSetCodeSystemConceptDesignationList(inherited Clone);
end;

function TFhirValueSetCodeSystemConceptDesignationList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetCodeSystemConceptDesignationList.GetItemN(index: Integer): TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemConceptDesignationList.IndexOf(value: TFhirValueSetCodeSystemConceptDesignation): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetCodeSystemConceptDesignationList.Insert(index: Integer): TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetCodeSystemConceptDesignationList.InsertItem(index: Integer; value: TFhirValueSetCodeSystemConceptDesignation);
begin
  assert(value is TFhirValueSetCodeSystemConceptDesignation);
  Inherited Insert(index, value);
end;

function TFhirValueSetCodeSystemConceptDesignationList.Item(index: Integer): TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation(ObjectByIndex[index]);
end;

function TFhirValueSetCodeSystemConceptDesignationList.Link: TFhirValueSetCodeSystemConceptDesignationList;
begin
  result := TFhirValueSetCodeSystemConceptDesignationList(inherited Link);
end;

procedure TFhirValueSetCodeSystemConceptDesignationList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetCodeSystemConceptDesignationList.SetItemByIndex(index: Integer; value: TFhirValueSetCodeSystemConceptDesignation);
begin
  assert(value is TFhirValueSetCodeSystemConceptDesignation);
  FhirValueSetCodeSystemConceptDesignations[index] := value;
end;

procedure TFhirValueSetCodeSystemConceptDesignationList.SetItemN(index: Integer; value: TFhirValueSetCodeSystemConceptDesignation);
begin
  assert(value is TFhirValueSetCodeSystemConceptDesignation);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetCompose }

constructor TFhirValueSetCompose.Create;
begin
  inherited;
end;

destructor TFhirValueSetCompose.Destroy;
begin
  FImportList.Free;
  FIncludeList.Free;
  FExcludeList.Free;
  inherited;
end;

procedure TFhirValueSetCompose.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirValueSetCompose(oSource).FImportList = nil) then
  begin
    FImportList.free;
    FImportList := nil;
  end
  else
  begin
    FImportList := TFhirUriList.Create;
    FImportList.Assign(TFhirValueSetCompose(oSource).FImportList);
  end;
  if (TFhirValueSetCompose(oSource).FIncludeList = nil) then
  begin
    FIncludeList.free;
    FIncludeList := nil;
  end
  else
  begin
    FIncludeList := TFhirValueSetComposeIncludeList.Create;
    FIncludeList.Assign(TFhirValueSetCompose(oSource).FIncludeList);
  end;
  if (TFhirValueSetCompose(oSource).FExcludeList = nil) then
  begin
    FExcludeList.free;
    FExcludeList := nil;
  end
  else
  begin
    FExcludeList := TFhirValueSetComposeIncludeList.Create;
    FExcludeList.Assign(TFhirValueSetCompose(oSource).FExcludeList);
  end;
end;

procedure TFhirValueSetCompose.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'import') Then
    list.addAll(FImportList);
  if (child_name = 'include') Then
    list.addAll(FIncludeList);
  if (child_name = 'exclude') Then
    list.addAll(FExcludeList);
end;

procedure TFhirValueSetCompose.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'import', 'uri', FImportList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'include', '', FIncludeList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'exclude', '@ValueSet.compose.include', FExcludeList.Link)){3};
end;

procedure TFhirValueSetCompose.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'import') then ImportList.add(propValue as TFhirUri){2}
  else if (propName = 'include') then IncludeList.add(propValue as TFhirValueSetComposeInclude){2}
  else if (propName = 'exclude') then ExcludeList.add(propValue as TFhirValueSetComposeInclude){2}
  else inherited;
end;

function TFhirValueSetCompose.FhirType : string;
begin
  result := 'compose';
end;

function TFhirValueSetCompose.Link : TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose(inherited Link);
end;

function TFhirValueSetCompose.Clone : TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose(inherited Clone);
end;

function TFhirValueSetCompose.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCompose;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetCompose)) then
    result := false
  else
  begin
    o := TFhirValueSetCompose(other);
    result := compareDeep(importList, o.importList, true) and compareDeep(includeList, o.includeList, true) and 
      compareDeep(excludeList, o.excludeList, true);
  end;
end;

function TFhirValueSetCompose.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetCompose;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetCompose)) then
    result := false
  else
  begin
    o := TFhirValueSetCompose(other);
    result := compareValues(importList, o.importList, true);
  end;
end;

{ TFhirValueSetCompose }

Function TFhirValueSetCompose.GetImportList : TFhirUriList;
begin
  if FImportList = nil then
    FImportList := TFhirUriList.Create;
  result := FImportList;
end;

Function TFhirValueSetCompose.GetHasImportList : boolean;
begin
  result := (FImportList <> nil) and (FImportList.count > 0);
end;

Function TFhirValueSetCompose.GetIncludeList : TFhirValueSetComposeIncludeList;
begin
  if FIncludeList = nil then
    FIncludeList := TFhirValueSetComposeIncludeList.Create;
  result := FIncludeList;
end;

Function TFhirValueSetCompose.GetHasIncludeList : boolean;
begin
  result := (FIncludeList <> nil) and (FIncludeList.count > 0);
end;

Function TFhirValueSetCompose.GetExcludeList : TFhirValueSetComposeIncludeList;
begin
  if FExcludeList = nil then
    FExcludeList := TFhirValueSetComposeIncludeList.Create;
  result := FExcludeList;
end;

Function TFhirValueSetCompose.GetHasExcludeList : boolean;
begin
  result := (FExcludeList <> nil) and (FExcludeList.count > 0);
end;


{ TFhirValueSetComposeListEnumerator }

Constructor TFhirValueSetComposeListEnumerator.Create(list : TFhirValueSetComposeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetComposeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetComposeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetComposeListEnumerator.GetCurrent : TFhirValueSetCompose;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetComposeList }
procedure TFhirValueSetComposeList.AddItem(value: TFhirValueSetCompose);
begin
  assert(value.ClassName = 'TFhirValueSetCompose', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetCompose');
  add(value);
end;


function TFhirValueSetComposeList.Append: TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetComposeList.GetEnumerator : TFhirValueSetComposeListEnumerator;
begin
  result := TFhirValueSetComposeListEnumerator.Create(self.link);
end;

function TFhirValueSetComposeList.Clone: TFhirValueSetComposeList;
begin
  result := TFhirValueSetComposeList(inherited Clone);
end;

function TFhirValueSetComposeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetComposeList.GetItemN(index: Integer): TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose(ObjectByIndex[index]);
end;

function TFhirValueSetComposeList.IndexOf(value: TFhirValueSetCompose): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetComposeList.Insert(index: Integer): TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeList.InsertItem(index: Integer; value: TFhirValueSetCompose);
begin
  assert(value is TFhirValueSetCompose);
  Inherited Insert(index, value);
end;

function TFhirValueSetComposeList.Item(index: Integer): TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose(ObjectByIndex[index]);
end;

function TFhirValueSetComposeList.Link: TFhirValueSetComposeList;
begin
  result := TFhirValueSetComposeList(inherited Link);
end;

procedure TFhirValueSetComposeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetComposeList.SetItemByIndex(index: Integer; value: TFhirValueSetCompose);
begin
  assert(value is TFhirValueSetCompose);
  FhirValueSetComposes[index] := value;
end;

procedure TFhirValueSetComposeList.SetItemN(index: Integer; value: TFhirValueSetCompose);
begin
  assert(value is TFhirValueSetCompose);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetComposeInclude }

constructor TFhirValueSetComposeInclude.Create;
begin
  inherited;
end;

destructor TFhirValueSetComposeInclude.Destroy;
begin
  FSystem.free;
  FVersion.free;
  FConceptList.Free;
  FFilterList.Free;
  inherited;
end;

procedure TFhirValueSetComposeInclude.Assign(oSource : TAdvObject);
begin
  inherited;
  systemElement := TFhirValueSetComposeInclude(oSource).systemElement.Clone;
  versionElement := TFhirValueSetComposeInclude(oSource).versionElement.Clone;
  if (TFhirValueSetComposeInclude(oSource).FConceptList = nil) then
  begin
    FConceptList.free;
    FConceptList := nil;
  end
  else
  begin
    FConceptList := TFhirValueSetComposeIncludeConceptList.Create;
    FConceptList.Assign(TFhirValueSetComposeInclude(oSource).FConceptList);
  end;
  if (TFhirValueSetComposeInclude(oSource).FFilterList = nil) then
  begin
    FFilterList.free;
    FFilterList := nil;
  end
  else
  begin
    FFilterList := TFhirValueSetComposeIncludeFilterList.Create;
    FFilterList.Assign(TFhirValueSetComposeInclude(oSource).FFilterList);
  end;
end;

procedure TFhirValueSetComposeInclude.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'system') Then
     list.add(FSystem.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'concept') Then
    list.addAll(FConceptList);
  if (child_name = 'filter') Then
    list.addAll(FFilterList);
end;

procedure TFhirValueSetComposeInclude.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'system', 'uri', FSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'concept', '', FConceptList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'filter', '', FFilterList.Link)){3};
end;

procedure TFhirValueSetComposeInclude.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'system') then SystemElement := propValue as TFhirUri{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'concept') then ConceptList.add(propValue as TFhirValueSetComposeIncludeConcept){2}
  else if (propName = 'filter') then FilterList.add(propValue as TFhirValueSetComposeIncludeFilter){2}
  else inherited;
end;

function TFhirValueSetComposeInclude.FhirType : string;
begin
  result := 'include';
end;

function TFhirValueSetComposeInclude.Link : TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude(inherited Link);
end;

function TFhirValueSetComposeInclude.Clone : TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude(inherited Clone);
end;

function TFhirValueSetComposeInclude.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeInclude;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeInclude)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeInclude(other);
    result := compareDeep(systemElement, o.systemElement, true) and compareDeep(versionElement, o.versionElement, true) and 
      compareDeep(conceptList, o.conceptList, true) and compareDeep(filterList, o.filterList, true);
  end;
end;

function TFhirValueSetComposeInclude.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeInclude;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeInclude)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeInclude(other);
    result := compareValues(systemElement, o.systemElement, true) and compareValues(versionElement, o.versionElement, true);
  end;
end;

{ TFhirValueSetComposeInclude }

Procedure TFhirValueSetComposeInclude.SetSystem(value : TFhirUri);
begin
  FSystem.free;
  FSystem := value;
end;

Function TFhirValueSetComposeInclude.GetSystemST : String;
begin
  if FSystem = nil then
    result := ''
  else
    result := FSystem.value;
end;

Procedure TFhirValueSetComposeInclude.SetSystemST(value : String);
begin
  if value <> '' then
  begin
    if FSystem = nil then
      FSystem := TFhirUri.create;
    FSystem.value := value
  end
  else if FSystem <> nil then
    FSystem.value := '';
end;

Procedure TFhirValueSetComposeInclude.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirValueSetComposeInclude.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirValueSetComposeInclude.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Function TFhirValueSetComposeInclude.GetConceptList : TFhirValueSetComposeIncludeConceptList;
begin
  if FConceptList = nil then
    FConceptList := TFhirValueSetComposeIncludeConceptList.Create;
  result := FConceptList;
end;

Function TFhirValueSetComposeInclude.GetHasConceptList : boolean;
begin
  result := (FConceptList <> nil) and (FConceptList.count > 0);
end;

Function TFhirValueSetComposeInclude.GetFilterList : TFhirValueSetComposeIncludeFilterList;
begin
  if FFilterList = nil then
    FFilterList := TFhirValueSetComposeIncludeFilterList.Create;
  result := FFilterList;
end;

Function TFhirValueSetComposeInclude.GetHasFilterList : boolean;
begin
  result := (FFilterList <> nil) and (FFilterList.count > 0);
end;


{ TFhirValueSetComposeIncludeListEnumerator }

Constructor TFhirValueSetComposeIncludeListEnumerator.Create(list : TFhirValueSetComposeIncludeList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetComposeIncludeListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetComposeIncludeListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetComposeIncludeListEnumerator.GetCurrent : TFhirValueSetComposeInclude;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetComposeIncludeList }
procedure TFhirValueSetComposeIncludeList.AddItem(value: TFhirValueSetComposeInclude);
begin
  assert(value.ClassName = 'TFhirValueSetComposeInclude', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetComposeInclude');
  add(value);
end;


function TFhirValueSetComposeIncludeList.Append: TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetComposeIncludeList.GetEnumerator : TFhirValueSetComposeIncludeListEnumerator;
begin
  result := TFhirValueSetComposeIncludeListEnumerator.Create(self.link);
end;

function TFhirValueSetComposeIncludeList.Clone: TFhirValueSetComposeIncludeList;
begin
  result := TFhirValueSetComposeIncludeList(inherited Clone);
end;

function TFhirValueSetComposeIncludeList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetComposeIncludeList.GetItemN(index: Integer): TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeList.IndexOf(value: TFhirValueSetComposeInclude): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetComposeIncludeList.Insert(index: Integer): TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeList.InsertItem(index: Integer; value: TFhirValueSetComposeInclude);
begin
  assert(value is TFhirValueSetComposeInclude);
  Inherited Insert(index, value);
end;

function TFhirValueSetComposeIncludeList.Item(index: Integer): TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeList.Link: TFhirValueSetComposeIncludeList;
begin
  result := TFhirValueSetComposeIncludeList(inherited Link);
end;

procedure TFhirValueSetComposeIncludeList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetComposeIncludeList.SetItemByIndex(index: Integer; value: TFhirValueSetComposeInclude);
begin
  assert(value is TFhirValueSetComposeInclude);
  FhirValueSetComposeIncludes[index] := value;
end;

procedure TFhirValueSetComposeIncludeList.SetItemN(index: Integer; value: TFhirValueSetComposeInclude);
begin
  assert(value is TFhirValueSetComposeInclude);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetComposeIncludeConcept }

constructor TFhirValueSetComposeIncludeConcept.Create;
begin
  inherited;
end;

destructor TFhirValueSetComposeIncludeConcept.Destroy;
begin
  FCode.free;
  FDisplay.free;
  FDesignationList.Free;
  inherited;
end;

procedure TFhirValueSetComposeIncludeConcept.Assign(oSource : TAdvObject);
begin
  inherited;
  codeElement := TFhirValueSetComposeIncludeConcept(oSource).codeElement.Clone;
  displayElement := TFhirValueSetComposeIncludeConcept(oSource).displayElement.Clone;
  if (TFhirValueSetComposeIncludeConcept(oSource).FDesignationList = nil) then
  begin
    FDesignationList.free;
    FDesignationList := nil;
  end
  else
  begin
    FDesignationList := TFhirValueSetCodeSystemConceptDesignationList.Create;
    FDesignationList.Assign(TFhirValueSetComposeIncludeConcept(oSource).FDesignationList);
  end;
end;

procedure TFhirValueSetComposeIncludeConcept.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'display') Then
     list.add(FDisplay.Link);
  if (child_name = 'designation') Then
    list.addAll(FDesignationList);
end;

procedure TFhirValueSetComposeIncludeConcept.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'display', 'string', FDisplay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'designation', '@ValueSet.codeSystem.concept.designation', FDesignationList.Link)){3};
end;

procedure TFhirValueSetComposeIncludeConcept.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'display') then DisplayElement := propValue as TFhirString{5a}
  else if (propName = 'designation') then DesignationList.add(propValue as TFhirValueSetCodeSystemConceptDesignation){2}
  else inherited;
end;

function TFhirValueSetComposeIncludeConcept.FhirType : string;
begin
  result := 'concept';
end;

function TFhirValueSetComposeIncludeConcept.Link : TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept(inherited Link);
end;

function TFhirValueSetComposeIncludeConcept.Clone : TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept(inherited Clone);
end;

function TFhirValueSetComposeIncludeConcept.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeIncludeConcept;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeIncludeConcept)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeIncludeConcept(other);
    result := compareDeep(codeElement, o.codeElement, true) and compareDeep(displayElement, o.displayElement, true) and 
      compareDeep(designationList, o.designationList, true);
  end;
end;

function TFhirValueSetComposeIncludeConcept.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeIncludeConcept;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeIncludeConcept)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeIncludeConcept(other);
    result := compareValues(codeElement, o.codeElement, true) and compareValues(displayElement, o.displayElement, true);
  end;
end;

{ TFhirValueSetComposeIncludeConcept }

Procedure TFhirValueSetComposeIncludeConcept.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirValueSetComposeIncludeConcept.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirValueSetComposeIncludeConcept.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirValueSetComposeIncludeConcept.SetDisplay(value : TFhirString);
begin
  FDisplay.free;
  FDisplay := value;
end;

Function TFhirValueSetComposeIncludeConcept.GetDisplayST : String;
begin
  if FDisplay = nil then
    result := ''
  else
    result := FDisplay.value;
end;

Procedure TFhirValueSetComposeIncludeConcept.SetDisplayST(value : String);
begin
  if value <> '' then
  begin
    if FDisplay = nil then
      FDisplay := TFhirString.create;
    FDisplay.value := value
  end
  else if FDisplay <> nil then
    FDisplay.value := '';
end;

Function TFhirValueSetComposeIncludeConcept.GetDesignationList : TFhirValueSetCodeSystemConceptDesignationList;
begin
  if FDesignationList = nil then
    FDesignationList := TFhirValueSetCodeSystemConceptDesignationList.Create;
  result := FDesignationList;
end;

Function TFhirValueSetComposeIncludeConcept.GetHasDesignationList : boolean;
begin
  result := (FDesignationList <> nil) and (FDesignationList.count > 0);
end;


{ TFhirValueSetComposeIncludeConceptListEnumerator }

Constructor TFhirValueSetComposeIncludeConceptListEnumerator.Create(list : TFhirValueSetComposeIncludeConceptList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetComposeIncludeConceptListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetComposeIncludeConceptListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetComposeIncludeConceptListEnumerator.GetCurrent : TFhirValueSetComposeIncludeConcept;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetComposeIncludeConceptList }
procedure TFhirValueSetComposeIncludeConceptList.AddItem(value: TFhirValueSetComposeIncludeConcept);
begin
  assert(value.ClassName = 'TFhirValueSetComposeIncludeConcept', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetComposeIncludeConcept');
  add(value);
end;


function TFhirValueSetComposeIncludeConceptList.Append: TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeConceptList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetComposeIncludeConceptList.GetEnumerator : TFhirValueSetComposeIncludeConceptListEnumerator;
begin
  result := TFhirValueSetComposeIncludeConceptListEnumerator.Create(self.link);
end;

function TFhirValueSetComposeIncludeConceptList.Clone: TFhirValueSetComposeIncludeConceptList;
begin
  result := TFhirValueSetComposeIncludeConceptList(inherited Clone);
end;

function TFhirValueSetComposeIncludeConceptList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetComposeIncludeConceptList.GetItemN(index: Integer): TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeConceptList.IndexOf(value: TFhirValueSetComposeIncludeConcept): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetComposeIncludeConceptList.Insert(index: Integer): TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeConceptList.InsertItem(index: Integer; value: TFhirValueSetComposeIncludeConcept);
begin
  assert(value is TFhirValueSetComposeIncludeConcept);
  Inherited Insert(index, value);
end;

function TFhirValueSetComposeIncludeConceptList.Item(index: Integer): TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeConceptList.Link: TFhirValueSetComposeIncludeConceptList;
begin
  result := TFhirValueSetComposeIncludeConceptList(inherited Link);
end;

procedure TFhirValueSetComposeIncludeConceptList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetComposeIncludeConceptList.SetItemByIndex(index: Integer; value: TFhirValueSetComposeIncludeConcept);
begin
  assert(value is TFhirValueSetComposeIncludeConcept);
  FhirValueSetComposeIncludeConcepts[index] := value;
end;

procedure TFhirValueSetComposeIncludeConceptList.SetItemN(index: Integer; value: TFhirValueSetComposeIncludeConcept);
begin
  assert(value is TFhirValueSetComposeIncludeConcept);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetComposeIncludeFilter }

constructor TFhirValueSetComposeIncludeFilter.Create;
begin
  inherited;
end;

destructor TFhirValueSetComposeIncludeFilter.Destroy;
begin
  FProperty_.free;
  FOp.free;
  FValue.free;
  inherited;
end;

procedure TFhirValueSetComposeIncludeFilter.Assign(oSource : TAdvObject);
begin
  inherited;
  property_Element := TFhirValueSetComposeIncludeFilter(oSource).property_Element.Clone;
  FOp := TFhirValueSetComposeIncludeFilter(oSource).FOp.Link;
  valueElement := TFhirValueSetComposeIncludeFilter(oSource).valueElement.Clone;
end;

procedure TFhirValueSetComposeIncludeFilter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'property') Then
     list.add(FProperty_.Link);
  if (child_name = 'op') Then
     list.add(FOp.Link);
  if (child_name = 'value') Then
     list.add(FValue.Link);
end;

procedure TFhirValueSetComposeIncludeFilter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'property', 'code', FProperty_.Link));{2}
  oList.add(TFHIRProperty.create(self, 'op', 'code', FOp.Link));{1}
  oList.add(TFHIRProperty.create(self, 'value', 'code', FValue.Link));{2}
end;

procedure TFhirValueSetComposeIncludeFilter.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'property') then
    if propValue is TFHIRCode then
      Property_Element := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      Property_Element := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'op') then OpElement := propValue as TFHIREnum
  else if (propName = 'value') then
    if propValue is TFHIRCode then
      ValueElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      ValueElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else inherited;
end;

function TFhirValueSetComposeIncludeFilter.FhirType : string;
begin
  result := 'filter';
end;

function TFhirValueSetComposeIncludeFilter.Link : TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter(inherited Link);
end;

function TFhirValueSetComposeIncludeFilter.Clone : TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter(inherited Clone);
end;

function TFhirValueSetComposeIncludeFilter.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeIncludeFilter;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeIncludeFilter)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeIncludeFilter(other);
    result := compareDeep(property_Element, o.property_Element, true) and compareDeep(opElement, o.opElement, true) and 
      compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirValueSetComposeIncludeFilter.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetComposeIncludeFilter;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetComposeIncludeFilter)) then
    result := false
  else
  begin
    o := TFhirValueSetComposeIncludeFilter(other);
    result := compareValues(property_Element, o.property_Element, true) and compareValues(opElement, o.opElement, true) and 
      compareValues(valueElement, o.valueElement, true);
  end;
end;

{ TFhirValueSetComposeIncludeFilter }

Procedure TFhirValueSetComposeIncludeFilter.SetProperty_(value : TFhirCode);
begin
  FProperty_.free;
  FProperty_ := value;
end;

Function TFhirValueSetComposeIncludeFilter.GetProperty_ST : String;
begin
  if FProperty_ = nil then
    result := ''
  else
    result := FProperty_.value;
end;

Procedure TFhirValueSetComposeIncludeFilter.SetProperty_ST(value : String);
begin
  if value <> '' then
  begin
    if FProperty_ = nil then
      FProperty_ := TFhirCode.create;
    FProperty_.value := value
  end
  else if FProperty_ <> nil then
    FProperty_.value := '';
end;

Procedure TFhirValueSetComposeIncludeFilter.SetOp(value : TFhirEnum);
begin
  FOp.free;
  FOp := value;
end;

Function TFhirValueSetComposeIncludeFilter.GetOpST : TFhirFilterOperator;
begin
  if FOp = nil then
    result := TFhirFilterOperator(0)
  else
    result := TFhirFilterOperator(StringArrayIndexOfSensitive(CODES_TFhirFilterOperator, FOp.value));
end;

Procedure TFhirValueSetComposeIncludeFilter.SetOpST(value : TFhirFilterOperator);
begin
  if ord(value) = 0 then
    OpElement := nil
  else
    OpElement := TFhirEnum.create(CODES_TFhirFilterOperator[value]);
end;

Procedure TFhirValueSetComposeIncludeFilter.SetValue(value : TFhirCode);
begin
  FValue.free;
  FValue := value;
end;

Function TFhirValueSetComposeIncludeFilter.GetValueST : String;
begin
  if FValue = nil then
    result := ''
  else
    result := FValue.value;
end;

Procedure TFhirValueSetComposeIncludeFilter.SetValueST(value : String);
begin
  if value <> '' then
  begin
    if FValue = nil then
      FValue := TFhirCode.create;
    FValue.value := value
  end
  else if FValue <> nil then
    FValue.value := '';
end;


{ TFhirValueSetComposeIncludeFilterListEnumerator }

Constructor TFhirValueSetComposeIncludeFilterListEnumerator.Create(list : TFhirValueSetComposeIncludeFilterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetComposeIncludeFilterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetComposeIncludeFilterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetComposeIncludeFilterListEnumerator.GetCurrent : TFhirValueSetComposeIncludeFilter;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetComposeIncludeFilterList }
procedure TFhirValueSetComposeIncludeFilterList.AddItem(value: TFhirValueSetComposeIncludeFilter);
begin
  assert(value.ClassName = 'TFhirValueSetComposeIncludeFilter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetComposeIncludeFilter');
  add(value);
end;


function TFhirValueSetComposeIncludeFilterList.Append: TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeFilterList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetComposeIncludeFilterList.GetEnumerator : TFhirValueSetComposeIncludeFilterListEnumerator;
begin
  result := TFhirValueSetComposeIncludeFilterListEnumerator.Create(self.link);
end;

function TFhirValueSetComposeIncludeFilterList.Clone: TFhirValueSetComposeIncludeFilterList;
begin
  result := TFhirValueSetComposeIncludeFilterList(inherited Clone);
end;

function TFhirValueSetComposeIncludeFilterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetComposeIncludeFilterList.GetItemN(index: Integer): TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeFilterList.IndexOf(value: TFhirValueSetComposeIncludeFilter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetComposeIncludeFilterList.Insert(index: Integer): TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetComposeIncludeFilterList.InsertItem(index: Integer; value: TFhirValueSetComposeIncludeFilter);
begin
  assert(value is TFhirValueSetComposeIncludeFilter);
  Inherited Insert(index, value);
end;

function TFhirValueSetComposeIncludeFilterList.Item(index: Integer): TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter(ObjectByIndex[index]);
end;

function TFhirValueSetComposeIncludeFilterList.Link: TFhirValueSetComposeIncludeFilterList;
begin
  result := TFhirValueSetComposeIncludeFilterList(inherited Link);
end;

procedure TFhirValueSetComposeIncludeFilterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetComposeIncludeFilterList.SetItemByIndex(index: Integer; value: TFhirValueSetComposeIncludeFilter);
begin
  assert(value is TFhirValueSetComposeIncludeFilter);
  FhirValueSetComposeIncludeFilters[index] := value;
end;

procedure TFhirValueSetComposeIncludeFilterList.SetItemN(index: Integer; value: TFhirValueSetComposeIncludeFilter);
begin
  assert(value is TFhirValueSetComposeIncludeFilter);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetExpansion }

constructor TFhirValueSetExpansion.Create;
begin
  inherited;
end;

destructor TFhirValueSetExpansion.Destroy;
begin
  FIdentifier.free;
  FTimestamp.free;
  FTotal.free;
  FOffset.free;
  FParameterList.Free;
  FContainsList.Free;
  inherited;
end;

procedure TFhirValueSetExpansion.Assign(oSource : TAdvObject);
begin
  inherited;
  identifierElement := TFhirValueSetExpansion(oSource).identifierElement.Clone;
  timestampElement := TFhirValueSetExpansion(oSource).timestampElement.Clone;
  totalElement := TFhirValueSetExpansion(oSource).totalElement.Clone;
  offsetElement := TFhirValueSetExpansion(oSource).offsetElement.Clone;
  if (TFhirValueSetExpansion(oSource).FParameterList = nil) then
  begin
    FParameterList.free;
    FParameterList := nil;
  end
  else
  begin
    FParameterList := TFhirValueSetExpansionParameterList.Create;
    FParameterList.Assign(TFhirValueSetExpansion(oSource).FParameterList);
  end;
  if (TFhirValueSetExpansion(oSource).FContainsList = nil) then
  begin
    FContainsList.free;
    FContainsList := nil;
  end
  else
  begin
    FContainsList := TFhirValueSetExpansionContainsList.Create;
    FContainsList.Assign(TFhirValueSetExpansion(oSource).FContainsList);
  end;
end;

procedure TFhirValueSetExpansion.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'timestamp') Then
     list.add(FTimestamp.Link);
  if (child_name = 'total') Then
     list.add(FTotal.Link);
  if (child_name = 'offset') Then
     list.add(FOffset.Link);
  if (child_name = 'parameter') Then
    list.addAll(FParameterList);
  if (child_name = 'contains') Then
    list.addAll(FContainsList);
end;

procedure TFhirValueSetExpansion.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'uri', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'timestamp', 'dateTime', FTimestamp.Link));{2}
  oList.add(TFHIRProperty.create(self, 'total', 'integer', FTotal.Link));{2}
  oList.add(TFHIRProperty.create(self, 'offset', 'integer', FOffset.Link));{2}
  oList.add(TFHIRProperty.create(self, 'parameter', '', FParameterList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'contains', '', FContainsList.Link)){3};
end;

procedure TFhirValueSetExpansion.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierElement := propValue as TFhirUri{5a}
  else if (propName = 'timestamp') then TimestampElement := propValue as TFhirDateTime{5a}
  else if (propName = 'total') then TotalElement := propValue as TFhirInteger{5a}
  else if (propName = 'offset') then OffsetElement := propValue as TFhirInteger{5a}
  else if (propName = 'parameter') then ParameterList.add(propValue as TFhirValueSetExpansionParameter){2}
  else if (propName = 'contains') then ContainsList.add(propValue as TFhirValueSetExpansionContains){2}
  else inherited;
end;

function TFhirValueSetExpansion.FhirType : string;
begin
  result := 'expansion';
end;

function TFhirValueSetExpansion.Link : TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion(inherited Link);
end;

function TFhirValueSetExpansion.Clone : TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion(inherited Clone);
end;

function TFhirValueSetExpansion.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansion;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansion)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansion(other);
    result := compareDeep(identifierElement, o.identifierElement, true) and compareDeep(timestampElement, o.timestampElement, true) and 
      compareDeep(totalElement, o.totalElement, true) and compareDeep(offsetElement, o.offsetElement, true) and 
      compareDeep(parameterList, o.parameterList, true) and compareDeep(containsList, o.containsList, true);
  end;
end;

function TFhirValueSetExpansion.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansion;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansion)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansion(other);
    result := compareValues(identifierElement, o.identifierElement, true) and compareValues(timestampElement, o.timestampElement, true) and 
      compareValues(totalElement, o.totalElement, true) and compareValues(offsetElement, o.offsetElement, true);
  end;
end;

{ TFhirValueSetExpansion }

Procedure TFhirValueSetExpansion.SetIdentifier(value : TFhirUri);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Function TFhirValueSetExpansion.GetIdentifierST : String;
begin
  if FIdentifier = nil then
    result := ''
  else
    result := FIdentifier.value;
end;

Procedure TFhirValueSetExpansion.SetIdentifierST(value : String);
begin
  if value <> '' then
  begin
    if FIdentifier = nil then
      FIdentifier := TFhirUri.create;
    FIdentifier.value := value
  end
  else if FIdentifier <> nil then
    FIdentifier.value := '';
end;

Procedure TFhirValueSetExpansion.SetTimestamp(value : TFhirDateTime);
begin
  FTimestamp.free;
  FTimestamp := value;
end;

Function TFhirValueSetExpansion.GetTimestampST : TDateAndTime;
begin
  if FTimestamp = nil then
    result := nil
  else
    result := FTimestamp.value;
end;

Procedure TFhirValueSetExpansion.SetTimestampST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FTimestamp = nil then
      FTimestamp := TFhirDateTime.create;
    FTimestamp.value := value
  end
  else if FTimestamp <> nil then
    FTimestamp.value := nil;
end;

Procedure TFhirValueSetExpansion.SetTotal(value : TFhirInteger);
begin
  FTotal.free;
  FTotal := value;
end;

Function TFhirValueSetExpansion.GetTotalST : String;
begin
  if FTotal = nil then
    result := ''
  else
    result := FTotal.value;
end;

Procedure TFhirValueSetExpansion.SetTotalST(value : String);
begin
  if value <> '' then
  begin
    if FTotal = nil then
      FTotal := TFhirInteger.create;
    FTotal.value := value
  end
  else if FTotal <> nil then
    FTotal.value := '';
end;

Procedure TFhirValueSetExpansion.SetOffset(value : TFhirInteger);
begin
  FOffset.free;
  FOffset := value;
end;

Function TFhirValueSetExpansion.GetOffsetST : String;
begin
  if FOffset = nil then
    result := ''
  else
    result := FOffset.value;
end;

Procedure TFhirValueSetExpansion.SetOffsetST(value : String);
begin
  if value <> '' then
  begin
    if FOffset = nil then
      FOffset := TFhirInteger.create;
    FOffset.value := value
  end
  else if FOffset <> nil then
    FOffset.value := '';
end;

Function TFhirValueSetExpansion.GetParameterList : TFhirValueSetExpansionParameterList;
begin
  if FParameterList = nil then
    FParameterList := TFhirValueSetExpansionParameterList.Create;
  result := FParameterList;
end;

Function TFhirValueSetExpansion.GetHasParameterList : boolean;
begin
  result := (FParameterList <> nil) and (FParameterList.count > 0);
end;

Function TFhirValueSetExpansion.GetContainsList : TFhirValueSetExpansionContainsList;
begin
  if FContainsList = nil then
    FContainsList := TFhirValueSetExpansionContainsList.Create;
  result := FContainsList;
end;

Function TFhirValueSetExpansion.GetHasContainsList : boolean;
begin
  result := (FContainsList <> nil) and (FContainsList.count > 0);
end;


{ TFhirValueSetExpansionListEnumerator }

Constructor TFhirValueSetExpansionListEnumerator.Create(list : TFhirValueSetExpansionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetExpansionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetExpansionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetExpansionListEnumerator.GetCurrent : TFhirValueSetExpansion;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetExpansionList }
procedure TFhirValueSetExpansionList.AddItem(value: TFhirValueSetExpansion);
begin
  assert(value.ClassName = 'TFhirValueSetExpansion', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetExpansion');
  add(value);
end;


function TFhirValueSetExpansionList.Append: TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetExpansionList.GetEnumerator : TFhirValueSetExpansionListEnumerator;
begin
  result := TFhirValueSetExpansionListEnumerator.Create(self.link);
end;

function TFhirValueSetExpansionList.Clone: TFhirValueSetExpansionList;
begin
  result := TFhirValueSetExpansionList(inherited Clone);
end;

function TFhirValueSetExpansionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetExpansionList.GetItemN(index: Integer): TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionList.IndexOf(value: TFhirValueSetExpansion): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetExpansionList.Insert(index: Integer): TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionList.InsertItem(index: Integer; value: TFhirValueSetExpansion);
begin
  assert(value is TFhirValueSetExpansion);
  Inherited Insert(index, value);
end;

function TFhirValueSetExpansionList.Item(index: Integer): TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionList.Link: TFhirValueSetExpansionList;
begin
  result := TFhirValueSetExpansionList(inherited Link);
end;

procedure TFhirValueSetExpansionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetExpansionList.SetItemByIndex(index: Integer; value: TFhirValueSetExpansion);
begin
  assert(value is TFhirValueSetExpansion);
  FhirValueSetExpansions[index] := value;
end;

procedure TFhirValueSetExpansionList.SetItemN(index: Integer; value: TFhirValueSetExpansion);
begin
  assert(value is TFhirValueSetExpansion);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetExpansionParameter }

constructor TFhirValueSetExpansionParameter.Create;
begin
  inherited;
end;

destructor TFhirValueSetExpansionParameter.Destroy;
begin
  FName.free;
  FValue.free;
  inherited;
end;

procedure TFhirValueSetExpansionParameter.Assign(oSource : TAdvObject);
begin
  inherited;
  nameElement := TFhirValueSetExpansionParameter(oSource).nameElement.Clone;
  value := TFhirValueSetExpansionParameter(oSource).value.Clone;
end;

procedure TFhirValueSetExpansionParameter.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'value[x]') Then
     list.add(FValue.Link);
end;

procedure TFhirValueSetExpansionParameter.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'value[x]', 'string|boolean|integer|decimal|uri|code', FValue.Link));{2}
end;

procedure TFhirValueSetExpansionParameter.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName.startsWith('value')) then Value := propValue as TFhirType{4}
  else inherited;
end;

function TFhirValueSetExpansionParameter.FhirType : string;
begin
  result := 'parameter';
end;

function TFhirValueSetExpansionParameter.Link : TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter(inherited Link);
end;

function TFhirValueSetExpansionParameter.Clone : TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter(inherited Clone);
end;

function TFhirValueSetExpansionParameter.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansionParameter;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansionParameter)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansionParameter(other);
    result := compareDeep(nameElement, o.nameElement, true) and compareDeep(valueElement, o.valueElement, true);
  end;
end;

function TFhirValueSetExpansionParameter.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansionParameter;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansionParameter)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansionParameter(other);
    result := compareValues(nameElement, o.nameElement, true);
  end;
end;

{ TFhirValueSetExpansionParameter }

Procedure TFhirValueSetExpansionParameter.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirValueSetExpansionParameter.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirValueSetExpansionParameter.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirValueSetExpansionParameter.SetValue(value : TFhirType);
begin
  FValue.free;
  FValue := value;
end;


{ TFhirValueSetExpansionParameterListEnumerator }

Constructor TFhirValueSetExpansionParameterListEnumerator.Create(list : TFhirValueSetExpansionParameterList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetExpansionParameterListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetExpansionParameterListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetExpansionParameterListEnumerator.GetCurrent : TFhirValueSetExpansionParameter;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetExpansionParameterList }
procedure TFhirValueSetExpansionParameterList.AddItem(value: TFhirValueSetExpansionParameter);
begin
  assert(value.ClassName = 'TFhirValueSetExpansionParameter', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetExpansionParameter');
  add(value);
end;


function TFhirValueSetExpansionParameterList.Append: TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionParameterList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetExpansionParameterList.GetEnumerator : TFhirValueSetExpansionParameterListEnumerator;
begin
  result := TFhirValueSetExpansionParameterListEnumerator.Create(self.link);
end;

function TFhirValueSetExpansionParameterList.Clone: TFhirValueSetExpansionParameterList;
begin
  result := TFhirValueSetExpansionParameterList(inherited Clone);
end;

function TFhirValueSetExpansionParameterList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetExpansionParameterList.GetItemN(index: Integer): TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionParameterList.IndexOf(value: TFhirValueSetExpansionParameter): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetExpansionParameterList.Insert(index: Integer): TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionParameterList.InsertItem(index: Integer; value: TFhirValueSetExpansionParameter);
begin
  assert(value is TFhirValueSetExpansionParameter);
  Inherited Insert(index, value);
end;

function TFhirValueSetExpansionParameterList.Item(index: Integer): TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionParameterList.Link: TFhirValueSetExpansionParameterList;
begin
  result := TFhirValueSetExpansionParameterList(inherited Link);
end;

procedure TFhirValueSetExpansionParameterList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetExpansionParameterList.SetItemByIndex(index: Integer; value: TFhirValueSetExpansionParameter);
begin
  assert(value is TFhirValueSetExpansionParameter);
  FhirValueSetExpansionParameters[index] := value;
end;

procedure TFhirValueSetExpansionParameterList.SetItemN(index: Integer; value: TFhirValueSetExpansionParameter);
begin
  assert(value is TFhirValueSetExpansionParameter);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSetExpansionContains }

constructor TFhirValueSetExpansionContains.Create;
begin
  inherited;
end;

destructor TFhirValueSetExpansionContains.Destroy;
begin
  FSystem.free;
  FAbstract.free;
  FVersion.free;
  FCode.free;
  FDisplay.free;
  FContainsList.Free;
  inherited;
end;

procedure TFhirValueSetExpansionContains.Assign(oSource : TAdvObject);
begin
  inherited;
  systemElement := TFhirValueSetExpansionContains(oSource).systemElement.Clone;
  abstractElement := TFhirValueSetExpansionContains(oSource).abstractElement.Clone;
  versionElement := TFhirValueSetExpansionContains(oSource).versionElement.Clone;
  codeElement := TFhirValueSetExpansionContains(oSource).codeElement.Clone;
  displayElement := TFhirValueSetExpansionContains(oSource).displayElement.Clone;
  if (TFhirValueSetExpansionContains(oSource).FContainsList = nil) then
  begin
    FContainsList.free;
    FContainsList := nil;
  end
  else
  begin
    FContainsList := TFhirValueSetExpansionContainsList.Create;
    FContainsList.Assign(TFhirValueSetExpansionContains(oSource).FContainsList);
  end;
end;

procedure TFhirValueSetExpansionContains.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'system') Then
     list.add(FSystem.Link);
  if (child_name = 'abstract') Then
     list.add(FAbstract.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'code') Then
     list.add(FCode.Link);
  if (child_name = 'display') Then
     list.add(FDisplay.Link);
  if (child_name = 'contains') Then
    list.addAll(FContainsList);
end;

procedure TFhirValueSetExpansionContains.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'system', 'uri', FSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'abstract', 'boolean', FAbstract.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'code', 'code', FCode.Link));{2}
  oList.add(TFHIRProperty.create(self, 'display', 'string', FDisplay.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contains', '@ValueSet.expansion.contains', FContainsList.Link)){3};
end;

procedure TFhirValueSetExpansionContains.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'system') then SystemElement := propValue as TFhirUri{5a}
  else if (propName = 'abstract') then AbstractElement := propValue as TFhirBoolean{5a}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'code') then
    if propValue is TFHIRCode then
      CodeElement := propValue as TFhirCode{5}
    else if propValue is TFHIREnum then
      CodeElement := TFHIRCode.create(TFHIREnum(propValue).value)
    else
      raise Exception.Create('Type mismatch: cannot convert from "'+propValue.className+'" to "TFHIRCode"'){5a}
  else if (propName = 'display') then DisplayElement := propValue as TFhirString{5a}
  else if (propName = 'contains') then ContainsList.add(propValue as TFhirValueSetExpansionContains){2}
  else inherited;
end;

function TFhirValueSetExpansionContains.FhirType : string;
begin
  result := 'contains';
end;

function TFhirValueSetExpansionContains.Link : TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains(inherited Link);
end;

function TFhirValueSetExpansionContains.Clone : TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains(inherited Clone);
end;

function TFhirValueSetExpansionContains.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansionContains;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansionContains)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansionContains(other);
    result := compareDeep(systemElement, o.systemElement, true) and compareDeep(abstractElement, o.abstractElement, true) and 
      compareDeep(versionElement, o.versionElement, true) and compareDeep(codeElement, o.codeElement, true) and 
      compareDeep(displayElement, o.displayElement, true) and compareDeep(containsList, o.containsList, true);
  end;
end;

function TFhirValueSetExpansionContains.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirValueSetExpansionContains;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirValueSetExpansionContains)) then
    result := false
  else
  begin
    o := TFhirValueSetExpansionContains(other);
    result := compareValues(systemElement, o.systemElement, true) and compareValues(abstractElement, o.abstractElement, true) and 
      compareValues(versionElement, o.versionElement, true) and compareValues(codeElement, o.codeElement, true) and 
      compareValues(displayElement, o.displayElement, true);
  end;
end;

{ TFhirValueSetExpansionContains }

Procedure TFhirValueSetExpansionContains.SetSystem(value : TFhirUri);
begin
  FSystem.free;
  FSystem := value;
end;

Function TFhirValueSetExpansionContains.GetSystemST : String;
begin
  if FSystem = nil then
    result := ''
  else
    result := FSystem.value;
end;

Procedure TFhirValueSetExpansionContains.SetSystemST(value : String);
begin
  if value <> '' then
  begin
    if FSystem = nil then
      FSystem := TFhirUri.create;
    FSystem.value := value
  end
  else if FSystem <> nil then
    FSystem.value := '';
end;

Procedure TFhirValueSetExpansionContains.SetAbstract(value : TFhirBoolean);
begin
  FAbstract.free;
  FAbstract := value;
end;

Function TFhirValueSetExpansionContains.GetAbstractST : Boolean;
begin
  if FAbstract = nil then
    result := false
  else
    result := FAbstract.value;
end;

Procedure TFhirValueSetExpansionContains.SetAbstractST(value : Boolean);
begin
  if FAbstract = nil then
    FAbstract := TFhirBoolean.create;
  FAbstract.value := value
end;

Procedure TFhirValueSetExpansionContains.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirValueSetExpansionContains.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirValueSetExpansionContains.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirValueSetExpansionContains.SetCode(value : TFhirCode);
begin
  FCode.free;
  FCode := value;
end;

Function TFhirValueSetExpansionContains.GetCodeST : String;
begin
  if FCode = nil then
    result := ''
  else
    result := FCode.value;
end;

Procedure TFhirValueSetExpansionContains.SetCodeST(value : String);
begin
  if value <> '' then
  begin
    if FCode = nil then
      FCode := TFhirCode.create;
    FCode.value := value
  end
  else if FCode <> nil then
    FCode.value := '';
end;

Procedure TFhirValueSetExpansionContains.SetDisplay(value : TFhirString);
begin
  FDisplay.free;
  FDisplay := value;
end;

Function TFhirValueSetExpansionContains.GetDisplayST : String;
begin
  if FDisplay = nil then
    result := ''
  else
    result := FDisplay.value;
end;

Procedure TFhirValueSetExpansionContains.SetDisplayST(value : String);
begin
  if value <> '' then
  begin
    if FDisplay = nil then
      FDisplay := TFhirString.create;
    FDisplay.value := value
  end
  else if FDisplay <> nil then
    FDisplay.value := '';
end;

Function TFhirValueSetExpansionContains.GetContainsList : TFhirValueSetExpansionContainsList;
begin
  if FContainsList = nil then
    FContainsList := TFhirValueSetExpansionContainsList.Create;
  result := FContainsList;
end;

Function TFhirValueSetExpansionContains.GetHasContainsList : boolean;
begin
  result := (FContainsList <> nil) and (FContainsList.count > 0);
end;


{ TFhirValueSetExpansionContainsListEnumerator }

Constructor TFhirValueSetExpansionContainsListEnumerator.Create(list : TFhirValueSetExpansionContainsList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetExpansionContainsListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetExpansionContainsListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetExpansionContainsListEnumerator.GetCurrent : TFhirValueSetExpansionContains;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetExpansionContainsList }
procedure TFhirValueSetExpansionContainsList.AddItem(value: TFhirValueSetExpansionContains);
begin
  assert(value.ClassName = 'TFhirValueSetExpansionContains', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSetExpansionContains');
  add(value);
end;


function TFhirValueSetExpansionContainsList.Append: TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionContainsList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetExpansionContainsList.GetEnumerator : TFhirValueSetExpansionContainsListEnumerator;
begin
  result := TFhirValueSetExpansionContainsListEnumerator.Create(self.link);
end;

function TFhirValueSetExpansionContainsList.Clone: TFhirValueSetExpansionContainsList;
begin
  result := TFhirValueSetExpansionContainsList(inherited Clone);
end;

function TFhirValueSetExpansionContainsList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetExpansionContainsList.GetItemN(index: Integer): TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionContainsList.IndexOf(value: TFhirValueSetExpansionContains): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetExpansionContainsList.Insert(index: Integer): TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetExpansionContainsList.InsertItem(index: Integer; value: TFhirValueSetExpansionContains);
begin
  assert(value is TFhirValueSetExpansionContains);
  Inherited Insert(index, value);
end;

function TFhirValueSetExpansionContainsList.Item(index: Integer): TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains(ObjectByIndex[index]);
end;

function TFhirValueSetExpansionContainsList.Link: TFhirValueSetExpansionContainsList;
begin
  result := TFhirValueSetExpansionContainsList(inherited Link);
end;

procedure TFhirValueSetExpansionContainsList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetExpansionContainsList.SetItemByIndex(index: Integer; value: TFhirValueSetExpansionContains);
begin
  assert(value is TFhirValueSetExpansionContains);
  FhirValueSetExpansionContains[index] := value;
end;

procedure TFhirValueSetExpansionContainsList.SetItemN(index: Integer; value: TFhirValueSetExpansionContains);
begin
  assert(value is TFhirValueSetExpansionContains);
  ObjectByIndex[index] := value;
end;

{ TFhirValueSet }

constructor TFhirValueSet.Create;
begin
  inherited;
end;

destructor TFhirValueSet.Destroy;
begin
  FUrl.free;
  FIdentifier.free;
  FVersion.free;
  FName.free;
  FStatus.free;
  FExperimental.free;
  FPublisher.free;
  FContactList.Free;
  FDate.free;
  FLockedDate.free;
  FDescription.free;
  FUseContextList.Free;
  FImmutable.free;
  FRequirements.free;
  FCopyright.free;
  FExtensible.free;
  FCodeSystem.free;
  FCompose.free;
  FExpansion.free;
  inherited;
end;

function TFhirValueSet.GetResourceType : TFhirResourceType;
begin
  result := frtValueSet;
end;

procedure TFhirValueSet.Assign(oSource : TAdvObject);
begin
  inherited;
  urlElement := TFhirValueSet(oSource).urlElement.Clone;
  identifier := TFhirValueSet(oSource).identifier.Clone;
  versionElement := TFhirValueSet(oSource).versionElement.Clone;
  nameElement := TFhirValueSet(oSource).nameElement.Clone;
  FStatus := TFhirValueSet(oSource).FStatus.Link;
  experimentalElement := TFhirValueSet(oSource).experimentalElement.Clone;
  publisherElement := TFhirValueSet(oSource).publisherElement.Clone;
  if (TFhirValueSet(oSource).FContactList = nil) then
  begin
    FContactList.free;
    FContactList := nil;
  end
  else
  begin
    FContactList := TFhirValueSetContactList.Create;
    FContactList.Assign(TFhirValueSet(oSource).FContactList);
  end;
  dateElement := TFhirValueSet(oSource).dateElement.Clone;
  lockedDateElement := TFhirValueSet(oSource).lockedDateElement.Clone;
  descriptionElement := TFhirValueSet(oSource).descriptionElement.Clone;
  if (TFhirValueSet(oSource).FUseContextList = nil) then
  begin
    FUseContextList.free;
    FUseContextList := nil;
  end
  else
  begin
    FUseContextList := TFhirCodeableConceptList.Create;
    FUseContextList.Assign(TFhirValueSet(oSource).FUseContextList);
  end;
  immutableElement := TFhirValueSet(oSource).immutableElement.Clone;
  requirementsElement := TFhirValueSet(oSource).requirementsElement.Clone;
  copyrightElement := TFhirValueSet(oSource).copyrightElement.Clone;
  extensibleElement := TFhirValueSet(oSource).extensibleElement.Clone;
  codeSystem := TFhirValueSet(oSource).codeSystem.Clone;
  compose := TFhirValueSet(oSource).compose.Clone;
  expansion := TFhirValueSet(oSource).expansion.Clone;
end;

procedure TFhirValueSet.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'url') Then
     list.add(FUrl.Link);
  if (child_name = 'identifier') Then
     list.add(FIdentifier.Link);
  if (child_name = 'version') Then
     list.add(FVersion.Link);
  if (child_name = 'name') Then
     list.add(FName.Link);
  if (child_name = 'status') Then
     list.add(FStatus.Link);
  if (child_name = 'experimental') Then
     list.add(FExperimental.Link);
  if (child_name = 'publisher') Then
     list.add(FPublisher.Link);
  if (child_name = 'contact') Then
    list.addAll(FContactList);
  if (child_name = 'date') Then
     list.add(FDate.Link);
  if (child_name = 'lockedDate') Then
     list.add(FLockedDate.Link);
  if (child_name = 'description') Then
     list.add(FDescription.Link);
  if (child_name = 'useContext') Then
    list.addAll(FUseContextList);
  if (child_name = 'immutable') Then
     list.add(FImmutable.Link);
  if (child_name = 'requirements') Then
     list.add(FRequirements.Link);
  if (child_name = 'copyright') Then
     list.add(FCopyright.Link);
  if (child_name = 'extensible') Then
     list.add(FExtensible.Link);
  if (child_name = 'codeSystem') Then
     list.add(FCodeSystem.Link);
  if (child_name = 'compose') Then
     list.add(FCompose.Link);
  if (child_name = 'expansion') Then
     list.add(FExpansion.Link);
end;

procedure TFhirValueSet.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'url', 'uri', FUrl.Link));{2}
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifier.Link));{2}
  oList.add(TFHIRProperty.create(self, 'version', 'string', FVersion.Link));{2}
  oList.add(TFHIRProperty.create(self, 'name', 'string', FName.Link));{2}
  oList.add(TFHIRProperty.create(self, 'status', 'code', FStatus.Link));{1}
  oList.add(TFHIRProperty.create(self, 'experimental', 'boolean', FExperimental.Link));{2}
  oList.add(TFHIRProperty.create(self, 'publisher', 'string', FPublisher.Link));{2}
  oList.add(TFHIRProperty.create(self, 'contact', '', FContactList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'date', 'dateTime', FDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'lockedDate', 'date', FLockedDate.Link));{2}
  oList.add(TFHIRProperty.create(self, 'description', 'string', FDescription.Link));{2}
  oList.add(TFHIRProperty.create(self, 'useContext', 'CodeableConcept', FUseContextList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'immutable', 'boolean', FImmutable.Link));{2}
  oList.add(TFHIRProperty.create(self, 'requirements', 'string', FRequirements.Link));{2}
  oList.add(TFHIRProperty.create(self, 'copyright', 'string', FCopyright.Link));{2}
  oList.add(TFHIRProperty.create(self, 'extensible', 'boolean', FExtensible.Link));{2}
  oList.add(TFHIRProperty.create(self, 'codeSystem', '', FCodeSystem.Link));{2}
  oList.add(TFHIRProperty.create(self, 'compose', '', FCompose.Link));{2}
  oList.add(TFHIRProperty.create(self, 'expansion', '', FExpansion.Link));{2}
end;

procedure TFhirValueSet.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'url') then UrlElement := propValue as TFhirUri{5a}
  else if (propName = 'identifier') then Identifier := propValue as TFhirIdentifier{4b}
  else if (propName = 'version') then VersionElement := propValue as TFhirString{5a}
  else if (propName = 'name') then NameElement := propValue as TFhirString{5a}
  else if (propName = 'status') then StatusElement := propValue as TFHIREnum
  else if (propName = 'experimental') then ExperimentalElement := propValue as TFhirBoolean{5a}
  else if (propName = 'publisher') then PublisherElement := propValue as TFhirString{5a}
  else if (propName = 'contact') then ContactList.add(propValue as TFhirValueSetContact){2}
  else if (propName = 'date') then DateElement := propValue as TFhirDateTime{5a}
  else if (propName = 'lockedDate') then LockedDateElement := propValue as TFhirDate{5a}
  else if (propName = 'description') then DescriptionElement := propValue as TFhirString{5a}
  else if (propName = 'useContext') then UseContextList.add(propValue as TFhirCodeableConcept){2}
  else if (propName = 'immutable') then ImmutableElement := propValue as TFhirBoolean{5a}
  else if (propName = 'requirements') then RequirementsElement := propValue as TFhirString{5a}
  else if (propName = 'copyright') then CopyrightElement := propValue as TFhirString{5a}
  else if (propName = 'extensible') then ExtensibleElement := propValue as TFhirBoolean{5a}
  else if (propName = 'codeSystem') then CodeSystem := propValue as TFhirValueSetCodeSystem{4b}
  else if (propName = 'compose') then Compose := propValue as TFhirValueSetCompose{4b}
  else if (propName = 'expansion') then Expansion := propValue as TFhirValueSetExpansion{4b}
  else inherited;
end;

function TFhirValueSet.FhirType : string;
begin
  result := 'ValueSet';
end;

function TFhirValueSet.Link : TFhirValueSet;
begin
  result := TFhirValueSet(inherited Link);
end;

function TFhirValueSet.Clone : TFhirValueSet;
begin
  result := TFhirValueSet(inherited Clone);
end;

{ TFhirValueSet }

Procedure TFhirValueSet.SetUrl(value : TFhirUri);
begin
  FUrl.free;
  FUrl := value;
end;

Function TFhirValueSet.GetUrlST : String;
begin
  if FUrl = nil then
    result := ''
  else
    result := FUrl.value;
end;

Procedure TFhirValueSet.SetUrlST(value : String);
begin
  if value <> '' then
  begin
    if FUrl = nil then
      FUrl := TFhirUri.create;
    FUrl.value := value
  end
  else if FUrl <> nil then
    FUrl.value := '';
end;

Procedure TFhirValueSet.SetIdentifier(value : TFhirIdentifier);
begin
  FIdentifier.free;
  FIdentifier := value;
end;

Procedure TFhirValueSet.SetVersion(value : TFhirString);
begin
  FVersion.free;
  FVersion := value;
end;

Function TFhirValueSet.GetVersionST : String;
begin
  if FVersion = nil then
    result := ''
  else
    result := FVersion.value;
end;

Procedure TFhirValueSet.SetVersionST(value : String);
begin
  if value <> '' then
  begin
    if FVersion = nil then
      FVersion := TFhirString.create;
    FVersion.value := value
  end
  else if FVersion <> nil then
    FVersion.value := '';
end;

Procedure TFhirValueSet.SetName(value : TFhirString);
begin
  FName.free;
  FName := value;
end;

Function TFhirValueSet.GetNameST : String;
begin
  if FName = nil then
    result := ''
  else
    result := FName.value;
end;

Procedure TFhirValueSet.SetNameST(value : String);
begin
  if value <> '' then
  begin
    if FName = nil then
      FName := TFhirString.create;
    FName.value := value
  end
  else if FName <> nil then
    FName.value := '';
end;

Procedure TFhirValueSet.SetStatus(value : TFhirEnum);
begin
  FStatus.free;
  FStatus := value;
end;

Function TFhirValueSet.GetStatusST : TFhirConformanceResourceStatus;
begin
  if FStatus = nil then
    result := TFhirConformanceResourceStatus(0)
  else
    result := TFhirConformanceResourceStatus(StringArrayIndexOfSensitive(CODES_TFhirConformanceResourceStatus, FStatus.value));
end;

Procedure TFhirValueSet.SetStatusST(value : TFhirConformanceResourceStatus);
begin
  if ord(value) = 0 then
    StatusElement := nil
  else
    StatusElement := TFhirEnum.create(CODES_TFhirConformanceResourceStatus[value]);
end;

Procedure TFhirValueSet.SetExperimental(value : TFhirBoolean);
begin
  FExperimental.free;
  FExperimental := value;
end;

Function TFhirValueSet.GetExperimentalST : Boolean;
begin
  if FExperimental = nil then
    result := false
  else
    result := FExperimental.value;
end;

Procedure TFhirValueSet.SetExperimentalST(value : Boolean);
begin
  if FExperimental = nil then
    FExperimental := TFhirBoolean.create;
  FExperimental.value := value
end;

Procedure TFhirValueSet.SetPublisher(value : TFhirString);
begin
  FPublisher.free;
  FPublisher := value;
end;

Function TFhirValueSet.GetPublisherST : String;
begin
  if FPublisher = nil then
    result := ''
  else
    result := FPublisher.value;
end;

Procedure TFhirValueSet.SetPublisherST(value : String);
begin
  if value <> '' then
  begin
    if FPublisher = nil then
      FPublisher := TFhirString.create;
    FPublisher.value := value
  end
  else if FPublisher <> nil then
    FPublisher.value := '';
end;

Function TFhirValueSet.GetContactList : TFhirValueSetContactList;
begin
  if FContactList = nil then
    FContactList := TFhirValueSetContactList.Create;
  result := FContactList;
end;

Function TFhirValueSet.GetHasContactList : boolean;
begin
  result := (FContactList <> nil) and (FContactList.count > 0);
end;

Procedure TFhirValueSet.SetDate(value : TFhirDateTime);
begin
  FDate.free;
  FDate := value;
end;

Function TFhirValueSet.GetDateST : TDateAndTime;
begin
  if FDate = nil then
    result := nil
  else
    result := FDate.value;
end;

Procedure TFhirValueSet.SetDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDate = nil then
      FDate := TFhirDateTime.create;
    FDate.value := value
  end
  else if FDate <> nil then
    FDate.value := nil;
end;

Procedure TFhirValueSet.SetLockedDate(value : TFhirDate);
begin
  FLockedDate.free;
  FLockedDate := value;
end;

Function TFhirValueSet.GetLockedDateST : TDateAndTime;
begin
  if FLockedDate = nil then
    result := nil
  else
    result := FLockedDate.value;
end;

Procedure TFhirValueSet.SetLockedDateST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FLockedDate = nil then
      FLockedDate := TFhirDate.create;
    FLockedDate.value := value
  end
  else if FLockedDate <> nil then
    FLockedDate.value := nil;
end;

Procedure TFhirValueSet.SetDescription(value : TFhirString);
begin
  FDescription.free;
  FDescription := value;
end;

Function TFhirValueSet.GetDescriptionST : String;
begin
  if FDescription = nil then
    result := ''
  else
    result := FDescription.value;
end;

Procedure TFhirValueSet.SetDescriptionST(value : String);
begin
  if value <> '' then
  begin
    if FDescription = nil then
      FDescription := TFhirString.create;
    FDescription.value := value
  end
  else if FDescription <> nil then
    FDescription.value := '';
end;

Function TFhirValueSet.GetUseContextList : TFhirCodeableConceptList;
begin
  if FUseContextList = nil then
    FUseContextList := TFhirCodeableConceptList.Create;
  result := FUseContextList;
end;

Function TFhirValueSet.GetHasUseContextList : boolean;
begin
  result := (FUseContextList <> nil) and (FUseContextList.count > 0);
end;

Procedure TFhirValueSet.SetImmutable(value : TFhirBoolean);
begin
  FImmutable.free;
  FImmutable := value;
end;

Function TFhirValueSet.GetImmutableST : Boolean;
begin
  if FImmutable = nil then
    result := false
  else
    result := FImmutable.value;
end;

Procedure TFhirValueSet.SetImmutableST(value : Boolean);
begin
  if FImmutable = nil then
    FImmutable := TFhirBoolean.create;
  FImmutable.value := value
end;

Procedure TFhirValueSet.SetRequirements(value : TFhirString);
begin
  FRequirements.free;
  FRequirements := value;
end;

Function TFhirValueSet.GetRequirementsST : String;
begin
  if FRequirements = nil then
    result := ''
  else
    result := FRequirements.value;
end;

Procedure TFhirValueSet.SetRequirementsST(value : String);
begin
  if value <> '' then
  begin
    if FRequirements = nil then
      FRequirements := TFhirString.create;
    FRequirements.value := value
  end
  else if FRequirements <> nil then
    FRequirements.value := '';
end;

Procedure TFhirValueSet.SetCopyright(value : TFhirString);
begin
  FCopyright.free;
  FCopyright := value;
end;

Function TFhirValueSet.GetCopyrightST : String;
begin
  if FCopyright = nil then
    result := ''
  else
    result := FCopyright.value;
end;

Procedure TFhirValueSet.SetCopyrightST(value : String);
begin
  if value <> '' then
  begin
    if FCopyright = nil then
      FCopyright := TFhirString.create;
    FCopyright.value := value
  end
  else if FCopyright <> nil then
    FCopyright.value := '';
end;

Procedure TFhirValueSet.SetExtensible(value : TFhirBoolean);
begin
  FExtensible.free;
  FExtensible := value;
end;

Function TFhirValueSet.GetExtensibleST : Boolean;
begin
  if FExtensible = nil then
    result := false
  else
    result := FExtensible.value;
end;

Procedure TFhirValueSet.SetExtensibleST(value : Boolean);
begin
  if FExtensible = nil then
    FExtensible := TFhirBoolean.create;
  FExtensible.value := value
end;

Procedure TFhirValueSet.SetCodeSystem(value : TFhirValueSetCodeSystem);
begin
  FCodeSystem.free;
  FCodeSystem := value;
end;

Procedure TFhirValueSet.SetCompose(value : TFhirValueSetCompose);
begin
  FCompose.free;
  FCompose := value;
end;

Procedure TFhirValueSet.SetExpansion(value : TFhirValueSetExpansion);
begin
  FExpansion.free;
  FExpansion := value;
end;


{ TFhirValueSetListEnumerator }

Constructor TFhirValueSetListEnumerator.Create(list : TFhirValueSetList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirValueSetListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirValueSetListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirValueSetListEnumerator.GetCurrent : TFhirValueSet;
begin
  Result := FList[FIndex];
end;


{ TFhirValueSetList }
procedure TFhirValueSetList.AddItem(value: TFhirValueSet);
begin
  assert(value.ClassName = 'TFhirValueSet', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirValueSet');
  add(value);
end;


function TFhirValueSetList.Append: TFhirValueSet;
begin
  result := TFhirValueSet.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetList.ClearItems;
begin
  Clear;
end;

function TFhirValueSetList.GetEnumerator : TFhirValueSetListEnumerator;
begin
  result := TFhirValueSetListEnumerator.Create(self.link);
end;

function TFhirValueSetList.Clone: TFhirValueSetList;
begin
  result := TFhirValueSetList(inherited Clone);
end;

function TFhirValueSetList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirValueSetList.GetItemN(index: Integer): TFhirValueSet;
begin
  result := TFhirValueSet(ObjectByIndex[index]);
end;

function TFhirValueSetList.IndexOf(value: TFhirValueSet): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirValueSetList.Insert(index: Integer): TFhirValueSet;
begin
  result := TFhirValueSet.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirValueSetList.InsertItem(index: Integer; value: TFhirValueSet);
begin
  assert(value is TFhirValueSet);
  Inherited Insert(index, value);
end;

function TFhirValueSetList.Item(index: Integer): TFhirValueSet;
begin
  result := TFhirValueSet(ObjectByIndex[index]);
end;

function TFhirValueSetList.Link: TFhirValueSetList;
begin
  result := TFhirValueSetList(inherited Link);
end;

procedure TFhirValueSetList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirValueSetList.SetItemByIndex(index: Integer; value: TFhirValueSet);
begin
  assert(value is TFhirValueSet);
  FhirValueSets[index] := value;
end;

procedure TFhirValueSetList.SetItemN(index: Integer; value: TFhirValueSet);
begin
  assert(value is TFhirValueSet);
  ObjectByIndex[index] := value;
end;

{ TFhirVisionPrescriptionDispense }

constructor TFhirVisionPrescriptionDispense.Create;
begin
  inherited;
end;

destructor TFhirVisionPrescriptionDispense.Destroy;
begin
  FProduct.free;
  FEye.free;
  FSphere.free;
  FCylinder.free;
  FAxis.free;
  FPrism.free;
  FBase.free;
  FAdd.free;
  FPower.free;
  FBackCurve.free;
  FDiameter.free;
  FDuration.free;
  FColor.free;
  FBrand.free;
  FNotes.free;
  inherited;
end;

procedure TFhirVisionPrescriptionDispense.Assign(oSource : TAdvObject);
begin
  inherited;
  product := TFhirVisionPrescriptionDispense(oSource).product.Clone;
  FEye := TFhirVisionPrescriptionDispense(oSource).FEye.Link;
  sphereElement := TFhirVisionPrescriptionDispense(oSource).sphereElement.Clone;
  cylinderElement := TFhirVisionPrescriptionDispense(oSource).cylinderElement.Clone;
  axisElement := TFhirVisionPrescriptionDispense(oSource).axisElement.Clone;
  prismElement := TFhirVisionPrescriptionDispense(oSource).prismElement.Clone;
  FBase := TFhirVisionPrescriptionDispense(oSource).FBase.Link;
  addElement := TFhirVisionPrescriptionDispense(oSource).addElement.Clone;
  powerElement := TFhirVisionPrescriptionDispense(oSource).powerElement.Clone;
  backCurveElement := TFhirVisionPrescriptionDispense(oSource).backCurveElement.Clone;
  diameterElement := TFhirVisionPrescriptionDispense(oSource).diameterElement.Clone;
  duration := TFhirVisionPrescriptionDispense(oSource).duration.Clone;
  colorElement := TFhirVisionPrescriptionDispense(oSource).colorElement.Clone;
  brandElement := TFhirVisionPrescriptionDispense(oSource).brandElement.Clone;
  notesElement := TFhirVisionPrescriptionDispense(oSource).notesElement.Clone;
end;

procedure TFhirVisionPrescriptionDispense.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'product') Then
     list.add(FProduct.Link);
  if (child_name = 'eye') Then
     list.add(FEye.Link);
  if (child_name = 'sphere') Then
     list.add(FSphere.Link);
  if (child_name = 'cylinder') Then
     list.add(FCylinder.Link);
  if (child_name = 'axis') Then
     list.add(FAxis.Link);
  if (child_name = 'prism') Then
     list.add(FPrism.Link);
  if (child_name = 'base') Then
     list.add(FBase.Link);
  if (child_name = 'add') Then
     list.add(FAdd.Link);
  if (child_name = 'power') Then
     list.add(FPower.Link);
  if (child_name = 'backCurve') Then
     list.add(FBackCurve.Link);
  if (child_name = 'diameter') Then
     list.add(FDiameter.Link);
  if (child_name = 'duration') Then
     list.add(FDuration.Link);
  if (child_name = 'color') Then
     list.add(FColor.Link);
  if (child_name = 'brand') Then
     list.add(FBrand.Link);
  if (child_name = 'notes') Then
     list.add(FNotes.Link);
end;

procedure TFhirVisionPrescriptionDispense.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'product', 'Coding', FProduct.Link));{2}
  oList.add(TFHIRProperty.create(self, 'eye', 'code', FEye.Link));{1}
  oList.add(TFHIRProperty.create(self, 'sphere', 'decimal', FSphere.Link));{2}
  oList.add(TFHIRProperty.create(self, 'cylinder', 'decimal', FCylinder.Link));{2}
  oList.add(TFHIRProperty.create(self, 'axis', 'integer', FAxis.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prism', 'decimal', FPrism.Link));{2}
  oList.add(TFHIRProperty.create(self, 'base', 'code', FBase.Link));{1}
  oList.add(TFHIRProperty.create(self, 'add', 'decimal', FAdd.Link));{2}
  oList.add(TFHIRProperty.create(self, 'power', 'decimal', FPower.Link));{2}
  oList.add(TFHIRProperty.create(self, 'backCurve', 'decimal', FBackCurve.Link));{2}
  oList.add(TFHIRProperty.create(self, 'diameter', 'decimal', FDiameter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'duration', 'SimpleQuantity', FDuration.Link));{2}
  oList.add(TFHIRProperty.create(self, 'color', 'string', FColor.Link));{2}
  oList.add(TFHIRProperty.create(self, 'brand', 'string', FBrand.Link));{2}
  oList.add(TFHIRProperty.create(self, 'notes', 'string', FNotes.Link));{2}
end;

procedure TFhirVisionPrescriptionDispense.setProperty(propName : string; propValue: TFHIRObject);
begin
  if (propName = 'product') then Product := propValue as TFhirCoding{4b}
  else if (propName = 'eye') then EyeElement := propValue as TFHIREnum
  else if (propName = 'sphere') then SphereElement := propValue as TFhirDecimal{5a}
  else if (propName = 'cylinder') then CylinderElement := propValue as TFhirDecimal{5a}
  else if (propName = 'axis') then AxisElement := propValue as TFhirInteger{5a}
  else if (propName = 'prism') then PrismElement := propValue as TFhirDecimal{5a}
  else if (propName = 'base') then BaseElement := propValue as TFHIREnum
  else if (propName = 'add') then AddElement := propValue as TFhirDecimal{5a}
  else if (propName = 'power') then PowerElement := propValue as TFhirDecimal{5a}
  else if (propName = 'backCurve') then BackCurveElement := propValue as TFhirDecimal{5a}
  else if (propName = 'diameter') then DiameterElement := propValue as TFhirDecimal{5a}
  else if (propName = 'duration') then Duration := propValue as TFhirQuantity{4b}
  else if (propName = 'color') then ColorElement := propValue as TFhirString{5a}
  else if (propName = 'brand') then BrandElement := propValue as TFhirString{5a}
  else if (propName = 'notes') then NotesElement := propValue as TFhirString{5a}
  else inherited;
end;

function TFhirVisionPrescriptionDispense.FhirType : string;
begin
  result := 'dispense';
end;

function TFhirVisionPrescriptionDispense.Link : TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense(inherited Link);
end;

function TFhirVisionPrescriptionDispense.Clone : TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense(inherited Clone);
end;

function TFhirVisionPrescriptionDispense.equalsDeep(other : TFHIRObject) : boolean; 
var
  o : TFhirVisionPrescriptionDispense;
begin
  if (not inherited equalsDeep(other)) then
    result := false
  else if (not (other is TFhirVisionPrescriptionDispense)) then
    result := false
  else
  begin
    o := TFhirVisionPrescriptionDispense(other);
    result := compareDeep(productElement, o.productElement, true) and compareDeep(eyeElement, o.eyeElement, true) and 
      compareDeep(sphereElement, o.sphereElement, true) and compareDeep(cylinderElement, o.cylinderElement, true) and 
      compareDeep(axisElement, o.axisElement, true) and compareDeep(prismElement, o.prismElement, true) and 
      compareDeep(baseElement, o.baseElement, true) and compareDeep(addElement, o.addElement, true) and 
      compareDeep(powerElement, o.powerElement, true) and compareDeep(backCurveElement, o.backCurveElement, true) and 
      compareDeep(diameterElement, o.diameterElement, true) and compareDeep(durationElement, o.durationElement, true) and 
      compareDeep(colorElement, o.colorElement, true) and compareDeep(brandElement, o.brandElement, true) and 
      compareDeep(notesElement, o.notesElement, true);
  end;
end;

function TFhirVisionPrescriptionDispense.equalsShallow(other : TFHIRObject) : boolean; 
var
  o : TFhirVisionPrescriptionDispense;
begin
  if (not inherited equalsShallow(other)) then
    result := false
  else if (not (other is TFhirVisionPrescriptionDispense)) then
    result := false
  else
  begin
    o := TFhirVisionPrescriptionDispense(other);
    result := compareValues(eyeElement, o.eyeElement, true) and compareValues(sphereElement, o.sphereElement, true) and 
      compareValues(cylinderElement, o.cylinderElement, true) and compareValues(axisElement, o.axisElement, true) and 
      compareValues(prismElement, o.prismElement, true) and compareValues(baseElement, o.baseElement, true) and 
      compareValues(addElement, o.addElement, true) and compareValues(powerElement, o.powerElement, true) and 
      compareValues(backCurveElement, o.backCurveElement, true) and compareValues(diameterElement, o.diameterElement, true) and 
      compareValues(colorElement, o.colorElement, true) and compareValues(brandElement, o.brandElement, true) and 
      compareValues(notesElement, o.notesElement, true);
  end;
end;

{ TFhirVisionPrescriptionDispense }

Procedure TFhirVisionPrescriptionDispense.SetProduct(value : TFhirCoding);
begin
  FProduct.free;
  FProduct := value;
end;

Procedure TFhirVisionPrescriptionDispense.SetEye(value : TFhirEnum);
begin
  FEye.free;
  FEye := value;
end;

Function TFhirVisionPrescriptionDispense.GetEyeST : TFhirVisionEyeCodes;
begin
  if FEye = nil then
    result := TFhirVisionEyeCodes(0)
  else
    result := TFhirVisionEyeCodes(StringArrayIndexOfSensitive(CODES_TFhirVisionEyeCodes, FEye.value));
end;

Procedure TFhirVisionPrescriptionDispense.SetEyeST(value : TFhirVisionEyeCodes);
begin
  if ord(value) = 0 then
    EyeElement := nil
  else
    EyeElement := TFhirEnum.create(CODES_TFhirVisionEyeCodes[value]);
end;

Procedure TFhirVisionPrescriptionDispense.SetSphere(value : TFhirDecimal);
begin
  FSphere.free;
  FSphere := value;
end;

Function TFhirVisionPrescriptionDispense.GetSphereST : String;
begin
  if FSphere = nil then
    result := ''
  else
    result := FSphere.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetSphereST(value : String);
begin
  if value <> '' then
  begin
    if FSphere = nil then
      FSphere := TFhirDecimal.create;
    FSphere.value := value
  end
  else if FSphere <> nil then
    FSphere.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetCylinder(value : TFhirDecimal);
begin
  FCylinder.free;
  FCylinder := value;
end;

Function TFhirVisionPrescriptionDispense.GetCylinderST : String;
begin
  if FCylinder = nil then
    result := ''
  else
    result := FCylinder.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetCylinderST(value : String);
begin
  if value <> '' then
  begin
    if FCylinder = nil then
      FCylinder := TFhirDecimal.create;
    FCylinder.value := value
  end
  else if FCylinder <> nil then
    FCylinder.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetAxis(value : TFhirInteger);
begin
  FAxis.free;
  FAxis := value;
end;

Function TFhirVisionPrescriptionDispense.GetAxisST : String;
begin
  if FAxis = nil then
    result := ''
  else
    result := FAxis.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetAxisST(value : String);
begin
  if value <> '' then
  begin
    if FAxis = nil then
      FAxis := TFhirInteger.create;
    FAxis.value := value
  end
  else if FAxis <> nil then
    FAxis.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetPrism(value : TFhirDecimal);
begin
  FPrism.free;
  FPrism := value;
end;

Function TFhirVisionPrescriptionDispense.GetPrismST : String;
begin
  if FPrism = nil then
    result := ''
  else
    result := FPrism.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetPrismST(value : String);
begin
  if value <> '' then
  begin
    if FPrism = nil then
      FPrism := TFhirDecimal.create;
    FPrism.value := value
  end
  else if FPrism <> nil then
    FPrism.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetBase(value : TFhirEnum);
begin
  FBase.free;
  FBase := value;
end;

Function TFhirVisionPrescriptionDispense.GetBaseST : TFhirVisionBaseCodes;
begin
  if FBase = nil then
    result := TFhirVisionBaseCodes(0)
  else
    result := TFhirVisionBaseCodes(StringArrayIndexOfSensitive(CODES_TFhirVisionBaseCodes, FBase.value));
end;

Procedure TFhirVisionPrescriptionDispense.SetBaseST(value : TFhirVisionBaseCodes);
begin
  if ord(value) = 0 then
    BaseElement := nil
  else
    BaseElement := TFhirEnum.create(CODES_TFhirVisionBaseCodes[value]);
end;

Procedure TFhirVisionPrescriptionDispense.SetAdd(value : TFhirDecimal);
begin
  FAdd.free;
  FAdd := value;
end;

Function TFhirVisionPrescriptionDispense.GetAddST : String;
begin
  if FAdd = nil then
    result := ''
  else
    result := FAdd.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetAddST(value : String);
begin
  if value <> '' then
  begin
    if FAdd = nil then
      FAdd := TFhirDecimal.create;
    FAdd.value := value
  end
  else if FAdd <> nil then
    FAdd.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetPower(value : TFhirDecimal);
begin
  FPower.free;
  FPower := value;
end;

Function TFhirVisionPrescriptionDispense.GetPowerST : String;
begin
  if FPower = nil then
    result := ''
  else
    result := FPower.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetPowerST(value : String);
begin
  if value <> '' then
  begin
    if FPower = nil then
      FPower := TFhirDecimal.create;
    FPower.value := value
  end
  else if FPower <> nil then
    FPower.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetBackCurve(value : TFhirDecimal);
begin
  FBackCurve.free;
  FBackCurve := value;
end;

Function TFhirVisionPrescriptionDispense.GetBackCurveST : String;
begin
  if FBackCurve = nil then
    result := ''
  else
    result := FBackCurve.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetBackCurveST(value : String);
begin
  if value <> '' then
  begin
    if FBackCurve = nil then
      FBackCurve := TFhirDecimal.create;
    FBackCurve.value := value
  end
  else if FBackCurve <> nil then
    FBackCurve.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetDiameter(value : TFhirDecimal);
begin
  FDiameter.free;
  FDiameter := value;
end;

Function TFhirVisionPrescriptionDispense.GetDiameterST : String;
begin
  if FDiameter = nil then
    result := ''
  else
    result := FDiameter.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetDiameterST(value : String);
begin
  if value <> '' then
  begin
    if FDiameter = nil then
      FDiameter := TFhirDecimal.create;
    FDiameter.value := value
  end
  else if FDiameter <> nil then
    FDiameter.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetDuration(value : TFhirQuantity);
begin
  FDuration.free;
  FDuration := value;
end;

Procedure TFhirVisionPrescriptionDispense.SetColor(value : TFhirString);
begin
  FColor.free;
  FColor := value;
end;

Function TFhirVisionPrescriptionDispense.GetColorST : String;
begin
  if FColor = nil then
    result := ''
  else
    result := FColor.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetColorST(value : String);
begin
  if value <> '' then
  begin
    if FColor = nil then
      FColor := TFhirString.create;
    FColor.value := value
  end
  else if FColor <> nil then
    FColor.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetBrand(value : TFhirString);
begin
  FBrand.free;
  FBrand := value;
end;

Function TFhirVisionPrescriptionDispense.GetBrandST : String;
begin
  if FBrand = nil then
    result := ''
  else
    result := FBrand.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetBrandST(value : String);
begin
  if value <> '' then
  begin
    if FBrand = nil then
      FBrand := TFhirString.create;
    FBrand.value := value
  end
  else if FBrand <> nil then
    FBrand.value := '';
end;

Procedure TFhirVisionPrescriptionDispense.SetNotes(value : TFhirString);
begin
  FNotes.free;
  FNotes := value;
end;

Function TFhirVisionPrescriptionDispense.GetNotesST : String;
begin
  if FNotes = nil then
    result := ''
  else
    result := FNotes.value;
end;

Procedure TFhirVisionPrescriptionDispense.SetNotesST(value : String);
begin
  if value <> '' then
  begin
    if FNotes = nil then
      FNotes := TFhirString.create;
    FNotes.value := value
  end
  else if FNotes <> nil then
    FNotes.value := '';
end;


{ TFhirVisionPrescriptionDispenseListEnumerator }

Constructor TFhirVisionPrescriptionDispenseListEnumerator.Create(list : TFhirVisionPrescriptionDispenseList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirVisionPrescriptionDispenseListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirVisionPrescriptionDispenseListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirVisionPrescriptionDispenseListEnumerator.GetCurrent : TFhirVisionPrescriptionDispense;
begin
  Result := FList[FIndex];
end;


{ TFhirVisionPrescriptionDispenseList }
procedure TFhirVisionPrescriptionDispenseList.AddItem(value: TFhirVisionPrescriptionDispense);
begin
  assert(value.ClassName = 'TFhirVisionPrescriptionDispense', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirVisionPrescriptionDispense');
  add(value);
end;


function TFhirVisionPrescriptionDispenseList.Append: TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirVisionPrescriptionDispenseList.ClearItems;
begin
  Clear;
end;

function TFhirVisionPrescriptionDispenseList.GetEnumerator : TFhirVisionPrescriptionDispenseListEnumerator;
begin
  result := TFhirVisionPrescriptionDispenseListEnumerator.Create(self.link);
end;

function TFhirVisionPrescriptionDispenseList.Clone: TFhirVisionPrescriptionDispenseList;
begin
  result := TFhirVisionPrescriptionDispenseList(inherited Clone);
end;

function TFhirVisionPrescriptionDispenseList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirVisionPrescriptionDispenseList.GetItemN(index: Integer): TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense(ObjectByIndex[index]);
end;

function TFhirVisionPrescriptionDispenseList.IndexOf(value: TFhirVisionPrescriptionDispense): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirVisionPrescriptionDispenseList.Insert(index: Integer): TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirVisionPrescriptionDispenseList.InsertItem(index: Integer; value: TFhirVisionPrescriptionDispense);
begin
  assert(value is TFhirVisionPrescriptionDispense);
  Inherited Insert(index, value);
end;

function TFhirVisionPrescriptionDispenseList.Item(index: Integer): TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense(ObjectByIndex[index]);
end;

function TFhirVisionPrescriptionDispenseList.Link: TFhirVisionPrescriptionDispenseList;
begin
  result := TFhirVisionPrescriptionDispenseList(inherited Link);
end;

procedure TFhirVisionPrescriptionDispenseList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirVisionPrescriptionDispenseList.SetItemByIndex(index: Integer; value: TFhirVisionPrescriptionDispense);
begin
  assert(value is TFhirVisionPrescriptionDispense);
  FhirVisionPrescriptionDispenses[index] := value;
end;

procedure TFhirVisionPrescriptionDispenseList.SetItemN(index: Integer; value: TFhirVisionPrescriptionDispense);
begin
  assert(value is TFhirVisionPrescriptionDispense);
  ObjectByIndex[index] := value;
end;

{ TFhirVisionPrescription }

constructor TFhirVisionPrescription.Create;
begin
  inherited;
end;

destructor TFhirVisionPrescription.Destroy;
begin
  FIdentifierList.Free;
  FDateWritten.free;
  FPatient.free;
  FPrescriber.free;
  FEncounter.free;
  FReason.free;
  FDispenseList.Free;
  inherited;
end;

function TFhirVisionPrescription.GetResourceType : TFhirResourceType;
begin
  result := frtVisionPrescription;
end;

procedure TFhirVisionPrescription.Assign(oSource : TAdvObject);
begin
  inherited;
  if (TFhirVisionPrescription(oSource).FIdentifierList = nil) then
  begin
    FIdentifierList.free;
    FIdentifierList := nil;
  end
  else
  begin
    FIdentifierList := TFhirIdentifierList.Create;
    FIdentifierList.Assign(TFhirVisionPrescription(oSource).FIdentifierList);
  end;
  dateWrittenElement := TFhirVisionPrescription(oSource).dateWrittenElement.Clone;
  patient := TFhirVisionPrescription(oSource).patient.Clone;
  prescriber := TFhirVisionPrescription(oSource).prescriber.Clone;
  encounter := TFhirVisionPrescription(oSource).encounter.Clone;
  reason := TFhirVisionPrescription(oSource).reason.Clone;
  if (TFhirVisionPrescription(oSource).FDispenseList = nil) then
  begin
    FDispenseList.free;
    FDispenseList := nil;
  end
  else
  begin
    FDispenseList := TFhirVisionPrescriptionDispenseList.Create;
    FDispenseList.Assign(TFhirVisionPrescription(oSource).FDispenseList);
  end;
end;

procedure TFhirVisionPrescription.GetChildrenByName(child_name : string; list : TFHIRObjectList);
begin
  inherited;
  if (child_name = 'identifier') Then
    list.addAll(FIdentifierList);
  if (child_name = 'dateWritten') Then
     list.add(FDateWritten.Link);
  if (child_name = 'patient') Then
     list.add(FPatient.Link);
  if (child_name = 'prescriber') Then
     list.add(FPrescriber.Link);
  if (child_name = 'encounter') Then
     list.add(FEncounter.Link);
  if (child_name = 'reason[x]') Then
     list.add(FReason.Link);
  if (child_name = 'dispense') Then
    list.addAll(FDispenseList);
end;

procedure TFhirVisionPrescription.ListProperties(oList: TFHIRPropertyList; bInheritedProperties: Boolean);
begin
  inherited;
  oList.add(TFHIRProperty.create(self, 'identifier', 'Identifier', FIdentifierList.Link)){3};
  oList.add(TFHIRProperty.create(self, 'dateWritten', 'dateTime', FDateWritten.Link));{2}
  oList.add(TFHIRProperty.create(self, 'patient', 'Reference(Patient)', FPatient.Link));{2}
  oList.add(TFHIRProperty.create(self, 'prescriber', 'Reference(Practitioner)', FPrescriber.Link));{2}
  oList.add(TFHIRProperty.create(self, 'encounter', 'Reference(Encounter)', FEncounter.Link));{2}
  oList.add(TFHIRProperty.create(self, 'reason[x]', 'CodeableConcept|Reference(Condition)', FReason.Link));{2}
  oList.add(TFHIRProperty.create(self, 'dispense', '', FDispenseList.Link)){3};
end;

procedure TFhirVisionPrescription.setProperty(propName: string; propValue: TFHIRObject);
begin
  if (propName = 'identifier') then IdentifierList.add(propValue as TFhirIdentifier){2}
  else if (propName = 'dateWritten') then DateWrittenElement := propValue as TFhirDateTime{5a}
  else if (propName = 'patient') then Patient := propValue as TFhirReference{TFhirPatient}{4b}
  else if (propName = 'prescriber') then Prescriber := propValue as TFhirReference{TFhirPractitioner}{4b}
  else if (propName = 'encounter') then Encounter := propValue as TFhirReference{TFhirEncounter}{4b}
  else if (propName.startsWith('reason')) then Reason := propValue as TFhirType{4}
  else if (propName = 'dispense') then DispenseList.add(propValue as TFhirVisionPrescriptionDispense){2}
  else inherited;
end;

function TFhirVisionPrescription.FhirType : string;
begin
  result := 'VisionPrescription';
end;

function TFhirVisionPrescription.Link : TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription(inherited Link);
end;

function TFhirVisionPrescription.Clone : TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription(inherited Clone);
end;

{ TFhirVisionPrescription }

Function TFhirVisionPrescription.GetIdentifierList : TFhirIdentifierList;
begin
  if FIdentifierList = nil then
    FIdentifierList := TFhirIdentifierList.Create;
  result := FIdentifierList;
end;

Function TFhirVisionPrescription.GetHasIdentifierList : boolean;
begin
  result := (FIdentifierList <> nil) and (FIdentifierList.count > 0);
end;

Procedure TFhirVisionPrescription.SetDateWritten(value : TFhirDateTime);
begin
  FDateWritten.free;
  FDateWritten := value;
end;

Function TFhirVisionPrescription.GetDateWrittenST : TDateAndTime;
begin
  if FDateWritten = nil then
    result := nil
  else
    result := FDateWritten.value;
end;

Procedure TFhirVisionPrescription.SetDateWrittenST(value : TDateAndTime);
begin
  if value <> nil then
  begin
    if FDateWritten = nil then
      FDateWritten := TFhirDateTime.create;
    FDateWritten.value := value
  end
  else if FDateWritten <> nil then
    FDateWritten.value := nil;
end;

Procedure TFhirVisionPrescription.SetPatient(value : TFhirReference{TFhirPatient});
begin
  FPatient.free;
  FPatient := value;
end;

Procedure TFhirVisionPrescription.SetPrescriber(value : TFhirReference{TFhirPractitioner});
begin
  FPrescriber.free;
  FPrescriber := value;
end;

Procedure TFhirVisionPrescription.SetEncounter(value : TFhirReference{TFhirEncounter});
begin
  FEncounter.free;
  FEncounter := value;
end;

Procedure TFhirVisionPrescription.SetReason(value : TFhirType);
begin
  FReason.free;
  FReason := value;
end;

Function TFhirVisionPrescription.GetDispenseList : TFhirVisionPrescriptionDispenseList;
begin
  if FDispenseList = nil then
    FDispenseList := TFhirVisionPrescriptionDispenseList.Create;
  result := FDispenseList;
end;

Function TFhirVisionPrescription.GetHasDispenseList : boolean;
begin
  result := (FDispenseList <> nil) and (FDispenseList.count > 0);
end;


{ TFhirVisionPrescriptionListEnumerator }

Constructor TFhirVisionPrescriptionListEnumerator.Create(list : TFhirVisionPrescriptionList);
begin
  inherited Create;
  FIndex := -1;
  FList := list;
end;

Destructor TFhirVisionPrescriptionListEnumerator.Destroy;
begin
  FList.Free;
  inherited;
end;

function TFhirVisionPrescriptionListEnumerator.MoveNext : boolean;
begin
  inc(FIndex);
  Result := FIndex < FList.count;
end;

function TFhirVisionPrescriptionListEnumerator.GetCurrent : TFhirVisionPrescription;
begin
  Result := FList[FIndex];
end;


{ TFhirVisionPrescriptionList }
procedure TFhirVisionPrescriptionList.AddItem(value: TFhirVisionPrescription);
begin
  assert(value.ClassName = 'TFhirVisionPrescription', 'Attempt to add an item of type '+value.ClassName+' to a List of TFhirVisionPrescription');
  add(value);
end;


function TFhirVisionPrescriptionList.Append: TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription.create;
  try
    add(result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirVisionPrescriptionList.ClearItems;
begin
  Clear;
end;

function TFhirVisionPrescriptionList.GetEnumerator : TFhirVisionPrescriptionListEnumerator;
begin
  result := TFhirVisionPrescriptionListEnumerator.Create(self.link);
end;

function TFhirVisionPrescriptionList.Clone: TFhirVisionPrescriptionList;
begin
  result := TFhirVisionPrescriptionList(inherited Clone);
end;

function TFhirVisionPrescriptionList.Count: Integer;
begin
  result := Inherited Count;
end;

function TFhirVisionPrescriptionList.GetItemN(index: Integer): TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription(ObjectByIndex[index]);
end;

function TFhirVisionPrescriptionList.IndexOf(value: TFhirVisionPrescription): Integer;
begin
  result := IndexByReference(value);
end;


function TFhirVisionPrescriptionList.Insert(index: Integer): TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription.create;
  try
    inherited insert(index, result.Link);
  finally
    result.free;
  end;
end;


procedure TFhirVisionPrescriptionList.InsertItem(index: Integer; value: TFhirVisionPrescription);
begin
  assert(value is TFhirVisionPrescription);
  Inherited Insert(index, value);
end;

function TFhirVisionPrescriptionList.Item(index: Integer): TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription(ObjectByIndex[index]);
end;

function TFhirVisionPrescriptionList.Link: TFhirVisionPrescriptionList;
begin
  result := TFhirVisionPrescriptionList(inherited Link);
end;

procedure TFhirVisionPrescriptionList.Remove(index: Integer);
begin
  DeleteByIndex(index);
end;

procedure TFhirVisionPrescriptionList.SetItemByIndex(index: Integer; value: TFhirVisionPrescription);
begin
  assert(value is TFhirVisionPrescription);
  FhirVisionPrescriptions[index] := value;
end;

procedure TFhirVisionPrescriptionList.SetItemN(index: Integer; value: TFhirVisionPrescription);
begin
  assert(value is TFhirVisionPrescription);
  ObjectByIndex[index] := value;
end;

function TFhirResourceFactory.newElement : TFhirElement;
begin
  result := TFhirElement.create;
end;

function TFhirResourceFactory.newBackboneElement : TFhirBackboneElement;
begin
  result := TFhirBackboneElement.create;
end;

function TFhirResourceFactory.newEnum : TFhirEnum;
begin
  result := TFhirEnum.create;
end;

function TFhirResourceFactory.makeEnum(value : String) : TFhirEnum;
begin
  result := TFhirEnum.create;
  result.value := value;
end;

function TFhirResourceFactory.newInteger : TFhirInteger;
begin
  result := TFhirInteger.create;
end;

function TFhirResourceFactory.makeInteger(value : String) : TFhirInteger;
begin
  result := TFhirInteger.create;
  result.value := value;
end;

function TFhirResourceFactory.newDateTime : TFhirDateTime;
begin
  result := TFhirDateTime.create;
end;

function TFhirResourceFactory.makeDateTime(value : TDateAndTime) : TFhirDateTime;
begin
  result := TFhirDateTime.create;
  result.value := value;
end;

function TFhirResourceFactory.newDate : TFhirDate;
begin
  result := TFhirDate.create;
end;

function TFhirResourceFactory.makeDate(value : TDateAndTime) : TFhirDate;
begin
  result := TFhirDate.create;
  result.value := value;
end;

function TFhirResourceFactory.newDecimal : TFhirDecimal;
begin
  result := TFhirDecimal.create;
end;

function TFhirResourceFactory.makeDecimal(value : String) : TFhirDecimal;
begin
  result := TFhirDecimal.create;
  result.value := value;
end;

function TFhirResourceFactory.newUri : TFhirUri;
begin
  result := TFhirUri.create;
end;

function TFhirResourceFactory.makeUri(value : String) : TFhirUri;
begin
  result := TFhirUri.create;
  result.value := value;
end;

function TFhirResourceFactory.newBase64Binary : TFhirBase64Binary;
begin
  result := TFhirBase64Binary.create;
end;

function TFhirResourceFactory.makeBase64Binary(value : TBytes) : TFhirBase64Binary;
begin
  result := TFhirBase64Binary.create;
  result.value := value;
end;

function TFhirResourceFactory.newTime : TFhirTime;
begin
  result := TFhirTime.create;
end;

function TFhirResourceFactory.makeTime(value : String) : TFhirTime;
begin
  result := TFhirTime.create;
  result.value := value;
end;

function TFhirResourceFactory.newString : TFhirString;
begin
  result := TFhirString.create;
end;

function TFhirResourceFactory.makeString(value : String) : TFhirString;
begin
  result := TFhirString.create;
  result.value := value;
end;

function TFhirResourceFactory.newBoolean : TFhirBoolean;
begin
  result := TFhirBoolean.create;
end;

function TFhirResourceFactory.makeBoolean(value : Boolean) : TFhirBoolean;
begin
  result := TFhirBoolean.create;
  result.value := value;
end;

function TFhirResourceFactory.newInstant : TFhirInstant;
begin
  result := TFhirInstant.create;
end;

function TFhirResourceFactory.makeInstant(value : TDateAndTime) : TFhirInstant;
begin
  result := TFhirInstant.create;
  result.value := value;
end;

function TFhirResourceFactory.newMarkdown : TFhirMarkdown;
begin
  result := TFhirMarkdown.create;
end;

function TFhirResourceFactory.makeMarkdown(value : String) : TFhirMarkdown;
begin
  result := TFhirMarkdown.create;
  result.value := value;
end;

function TFhirResourceFactory.newUnsignedInt : TFhirUnsignedInt;
begin
  result := TFhirUnsignedInt.create;
end;

function TFhirResourceFactory.makeUnsignedInt(value : String) : TFhirUnsignedInt;
begin
  result := TFhirUnsignedInt.create;
  result.value := value;
end;

function TFhirResourceFactory.newCode : TFhirCode;
begin
  result := TFhirCode.create;
end;

function TFhirResourceFactory.makeCode(value : String) : TFhirCode;
begin
  result := TFhirCode.create;
  result.value := value;
end;

function TFhirResourceFactory.newId : TFhirId;
begin
  result := TFhirId.create;
end;

function TFhirResourceFactory.makeId(value : String) : TFhirId;
begin
  result := TFhirId.create;
  result.value := value;
end;

function TFhirResourceFactory.newOid : TFhirOid;
begin
  result := TFhirOid.create;
end;

function TFhirResourceFactory.makeOid(value : String) : TFhirOid;
begin
  result := TFhirOid.create;
  result.value := value;
end;

function TFhirResourceFactory.newPositiveInt : TFhirPositiveInt;
begin
  result := TFhirPositiveInt.create;
end;

function TFhirResourceFactory.makePositiveInt(value : String) : TFhirPositiveInt;
begin
  result := TFhirPositiveInt.create;
  result.value := value;
end;

function TFhirResourceFactory.newUuid : TFhirUuid;
begin
  result := TFhirUuid.create;
end;

function TFhirResourceFactory.makeUuid(value : String) : TFhirUuid;
begin
  result := TFhirUuid.create;
  result.value := value;
end;

function TFhirResourceFactory.newParametersParameter : TFhirParametersParameter;
begin
  result := TFhirParametersParameter.create;
end;

function TFhirResourceFactory.newParameters : TFhirParameters;
begin
  result := TFhirParameters.create;
end;

function TFhirResourceFactory.newExtension : TFhirExtension;
begin
  result := TFhirExtension.create;
end;

function TFhirResourceFactory.newNarrative : TFhirNarrative;
begin
  result := TFhirNarrative.create;
end;

function TFhirResourceFactory.newPeriod : TFhirPeriod;
begin
  result := TFhirPeriod.create;
end;

function TFhirResourceFactory.newCoding : TFhirCoding;
begin
  result := TFhirCoding.create;
end;

function TFhirResourceFactory.newRange : TFhirRange;
begin
  result := TFhirRange.create;
end;

function TFhirResourceFactory.newQuantity : TFhirQuantity;
begin
  result := TFhirQuantity.create;
end;

function TFhirResourceFactory.newAttachment : TFhirAttachment;
begin
  result := TFhirAttachment.create;
end;

function TFhirResourceFactory.newRatio : TFhirRatio;
begin
  result := TFhirRatio.create;
end;

function TFhirResourceFactory.newAnnotation : TFhirAnnotation;
begin
  result := TFhirAnnotation.create;
end;

function TFhirResourceFactory.newSampledData : TFhirSampledData;
begin
  result := TFhirSampledData.create;
end;

function TFhirResourceFactory.newReference : TFhirReference;
begin
  result := TFhirReference.create;
end;

function TFhirResourceFactory.newCodeableConcept : TFhirCodeableConcept;
begin
  result := TFhirCodeableConcept.create;
end;

function TFhirResourceFactory.newIdentifier : TFhirIdentifier;
begin
  result := TFhirIdentifier.create;
end;

function TFhirResourceFactory.newSignature : TFhirSignature;
begin
  result := TFhirSignature.create;
end;

function TFhirResourceFactory.newElementDefinitionSlicing : TFhirElementDefinitionSlicing;
begin
  result := TFhirElementDefinitionSlicing.create;
end;

function TFhirResourceFactory.newElementDefinitionBase : TFhirElementDefinitionBase;
begin
  result := TFhirElementDefinitionBase.create;
end;

function TFhirResourceFactory.newElementDefinitionType : TFhirElementDefinitionType;
begin
  result := TFhirElementDefinitionType.create;
end;

function TFhirResourceFactory.newElementDefinitionConstraint : TFhirElementDefinitionConstraint;
begin
  result := TFhirElementDefinitionConstraint.create;
end;

function TFhirResourceFactory.newElementDefinitionBinding : TFhirElementDefinitionBinding;
begin
  result := TFhirElementDefinitionBinding.create;
end;

function TFhirResourceFactory.newElementDefinitionMapping : TFhirElementDefinitionMapping;
begin
  result := TFhirElementDefinitionMapping.create;
end;

function TFhirResourceFactory.newElementDefinition : TFhirElementDefinition;
begin
  result := TFhirElementDefinition.create;
end;

function TFhirResourceFactory.newTimingRepeat : TFhirTimingRepeat;
begin
  result := TFhirTimingRepeat.create;
end;

function TFhirResourceFactory.newTiming : TFhirTiming;
begin
  result := TFhirTiming.create;
end;

function TFhirResourceFactory.newAddress : TFhirAddress;
begin
  result := TFhirAddress.create;
end;

function TFhirResourceFactory.newHumanName : TFhirHumanName;
begin
  result := TFhirHumanName.create;
end;

function TFhirResourceFactory.newMeta : TFhirMeta;
begin
  result := TFhirMeta.create;
end;

function TFhirResourceFactory.newContactPoint : TFhirContactPoint;
begin
  result := TFhirContactPoint.create;
end;

function TFhirResourceFactory.newAccount : TFhirAccount;
begin
  result := TFhirAccount.create;
end;

function TFhirResourceFactory.newAllergyIntoleranceReaction : TFhirAllergyIntoleranceReaction;
begin
  result := TFhirAllergyIntoleranceReaction.create;
end;

function TFhirResourceFactory.newAllergyIntolerance : TFhirAllergyIntolerance;
begin
  result := TFhirAllergyIntolerance.create;
end;

function TFhirResourceFactory.newAppointmentParticipant : TFhirAppointmentParticipant;
begin
  result := TFhirAppointmentParticipant.create;
end;

function TFhirResourceFactory.newAppointment : TFhirAppointment;
begin
  result := TFhirAppointment.create;
end;

function TFhirResourceFactory.newAppointmentResponse : TFhirAppointmentResponse;
begin
  result := TFhirAppointmentResponse.create;
end;

function TFhirResourceFactory.newAuditEventEvent : TFhirAuditEventEvent;
begin
  result := TFhirAuditEventEvent.create;
end;

function TFhirResourceFactory.newAuditEventParticipant : TFhirAuditEventParticipant;
begin
  result := TFhirAuditEventParticipant.create;
end;

function TFhirResourceFactory.newAuditEventParticipantNetwork : TFhirAuditEventParticipantNetwork;
begin
  result := TFhirAuditEventParticipantNetwork.create;
end;

function TFhirResourceFactory.newAuditEventSource : TFhirAuditEventSource;
begin
  result := TFhirAuditEventSource.create;
end;

function TFhirResourceFactory.newAuditEventObject : TFhirAuditEventObject;
begin
  result := TFhirAuditEventObject.create;
end;

function TFhirResourceFactory.newAuditEventObjectDetail : TFhirAuditEventObjectDetail;
begin
  result := TFhirAuditEventObjectDetail.create;
end;

function TFhirResourceFactory.newAuditEvent : TFhirAuditEvent;
begin
  result := TFhirAuditEvent.create;
end;

function TFhirResourceFactory.newBasic : TFhirBasic;
begin
  result := TFhirBasic.create;
end;

function TFhirResourceFactory.newBinary : TFhirBinary;
begin
  result := TFhirBinary.create;
end;

function TFhirResourceFactory.newBodySite : TFhirBodySite;
begin
  result := TFhirBodySite.create;
end;

function TFhirResourceFactory.newBundleLink : TFhirBundleLink;
begin
  result := TFhirBundleLink.create;
end;

function TFhirResourceFactory.newBundleEntry : TFhirBundleEntry;
begin
  result := TFhirBundleEntry.create;
end;

function TFhirResourceFactory.newBundleEntrySearch : TFhirBundleEntrySearch;
begin
  result := TFhirBundleEntrySearch.create;
end;

function TFhirResourceFactory.newBundleEntryRequest : TFhirBundleEntryRequest;
begin
  result := TFhirBundleEntryRequest.create;
end;

function TFhirResourceFactory.newBundleEntryResponse : TFhirBundleEntryResponse;
begin
  result := TFhirBundleEntryResponse.create;
end;

function TFhirResourceFactory.newBundle : TFhirBundle;
begin
  result := TFhirBundle.create;
end;

function TFhirResourceFactory.newCarePlanRelatedPlan : TFhirCarePlanRelatedPlan;
begin
  result := TFhirCarePlanRelatedPlan.create;
end;

function TFhirResourceFactory.newCarePlanParticipant : TFhirCarePlanParticipant;
begin
  result := TFhirCarePlanParticipant.create;
end;

function TFhirResourceFactory.newCarePlanActivity : TFhirCarePlanActivity;
begin
  result := TFhirCarePlanActivity.create;
end;

function TFhirResourceFactory.newCarePlanActivityDetail : TFhirCarePlanActivityDetail;
begin
  result := TFhirCarePlanActivityDetail.create;
end;

function TFhirResourceFactory.newCarePlan : TFhirCarePlan;
begin
  result := TFhirCarePlan.create;
end;

function TFhirResourceFactory.newClaimPayee : TFhirClaimPayee;
begin
  result := TFhirClaimPayee.create;
end;

function TFhirResourceFactory.newClaimDiagnosis : TFhirClaimDiagnosis;
begin
  result := TFhirClaimDiagnosis.create;
end;

function TFhirResourceFactory.newClaimCoverage : TFhirClaimCoverage;
begin
  result := TFhirClaimCoverage.create;
end;

function TFhirResourceFactory.newClaimItem : TFhirClaimItem;
begin
  result := TFhirClaimItem.create;
end;

function TFhirResourceFactory.newClaimItemDetail : TFhirClaimItemDetail;
begin
  result := TFhirClaimItemDetail.create;
end;

function TFhirResourceFactory.newClaimItemDetailSubDetail : TFhirClaimItemDetailSubDetail;
begin
  result := TFhirClaimItemDetailSubDetail.create;
end;

function TFhirResourceFactory.newClaimItemProsthesis : TFhirClaimItemProsthesis;
begin
  result := TFhirClaimItemProsthesis.create;
end;

function TFhirResourceFactory.newClaimMissingTeeth : TFhirClaimMissingTeeth;
begin
  result := TFhirClaimMissingTeeth.create;
end;

function TFhirResourceFactory.newClaim : TFhirClaim;
begin
  result := TFhirClaim.create;
end;

function TFhirResourceFactory.newClaimResponseItem : TFhirClaimResponseItem;
begin
  result := TFhirClaimResponseItem.create;
end;

function TFhirResourceFactory.newClaimResponseItemAdjudication : TFhirClaimResponseItemAdjudication;
begin
  result := TFhirClaimResponseItemAdjudication.create;
end;

function TFhirResourceFactory.newClaimResponseItemDetail : TFhirClaimResponseItemDetail;
begin
  result := TFhirClaimResponseItemDetail.create;
end;

function TFhirResourceFactory.newClaimResponseItemDetailAdjudication : TFhirClaimResponseItemDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailAdjudication.create;
end;

function TFhirResourceFactory.newClaimResponseItemDetailSubDetail : TFhirClaimResponseItemDetailSubDetail;
begin
  result := TFhirClaimResponseItemDetailSubDetail.create;
end;

function TFhirResourceFactory.newClaimResponseItemDetailSubDetailAdjudication : TFhirClaimResponseItemDetailSubDetailAdjudication;
begin
  result := TFhirClaimResponseItemDetailSubDetailAdjudication.create;
end;

function TFhirResourceFactory.newClaimResponseAddItem : TFhirClaimResponseAddItem;
begin
  result := TFhirClaimResponseAddItem.create;
end;

function TFhirResourceFactory.newClaimResponseAddItemAdjudication : TFhirClaimResponseAddItemAdjudication;
begin
  result := TFhirClaimResponseAddItemAdjudication.create;
end;

function TFhirResourceFactory.newClaimResponseAddItemDetail : TFhirClaimResponseAddItemDetail;
begin
  result := TFhirClaimResponseAddItemDetail.create;
end;

function TFhirResourceFactory.newClaimResponseAddItemDetailAdjudication : TFhirClaimResponseAddItemDetailAdjudication;
begin
  result := TFhirClaimResponseAddItemDetailAdjudication.create;
end;

function TFhirResourceFactory.newClaimResponseError : TFhirClaimResponseError;
begin
  result := TFhirClaimResponseError.create;
end;

function TFhirResourceFactory.newClaimResponseNote : TFhirClaimResponseNote;
begin
  result := TFhirClaimResponseNote.create;
end;

function TFhirResourceFactory.newClaimResponseCoverage : TFhirClaimResponseCoverage;
begin
  result := TFhirClaimResponseCoverage.create;
end;

function TFhirResourceFactory.newClaimResponse : TFhirClaimResponse;
begin
  result := TFhirClaimResponse.create;
end;

function TFhirResourceFactory.newClinicalImpressionInvestigations : TFhirClinicalImpressionInvestigations;
begin
  result := TFhirClinicalImpressionInvestigations.create;
end;

function TFhirResourceFactory.newClinicalImpressionFinding : TFhirClinicalImpressionFinding;
begin
  result := TFhirClinicalImpressionFinding.create;
end;

function TFhirResourceFactory.newClinicalImpressionRuledOut : TFhirClinicalImpressionRuledOut;
begin
  result := TFhirClinicalImpressionRuledOut.create;
end;

function TFhirResourceFactory.newClinicalImpression : TFhirClinicalImpression;
begin
  result := TFhirClinicalImpression.create;
end;

function TFhirResourceFactory.newCommunicationPayload : TFhirCommunicationPayload;
begin
  result := TFhirCommunicationPayload.create;
end;

function TFhirResourceFactory.newCommunication : TFhirCommunication;
begin
  result := TFhirCommunication.create;
end;

function TFhirResourceFactory.newCommunicationRequestPayload : TFhirCommunicationRequestPayload;
begin
  result := TFhirCommunicationRequestPayload.create;
end;

function TFhirResourceFactory.newCommunicationRequest : TFhirCommunicationRequest;
begin
  result := TFhirCommunicationRequest.create;
end;

function TFhirResourceFactory.newCompositionAttester : TFhirCompositionAttester;
begin
  result := TFhirCompositionAttester.create;
end;

function TFhirResourceFactory.newCompositionEvent : TFhirCompositionEvent;
begin
  result := TFhirCompositionEvent.create;
end;

function TFhirResourceFactory.newCompositionSection : TFhirCompositionSection;
begin
  result := TFhirCompositionSection.create;
end;

function TFhirResourceFactory.newComposition : TFhirComposition;
begin
  result := TFhirComposition.create;
end;

function TFhirResourceFactory.newConceptMapContact : TFhirConceptMapContact;
begin
  result := TFhirConceptMapContact.create;
end;

function TFhirResourceFactory.newConceptMapElement : TFhirConceptMapElement;
begin
  result := TFhirConceptMapElement.create;
end;

function TFhirResourceFactory.newConceptMapElementTarget : TFhirConceptMapElementTarget;
begin
  result := TFhirConceptMapElementTarget.create;
end;

function TFhirResourceFactory.newConceptMapElementTargetDependsOn : TFhirConceptMapElementTargetDependsOn;
begin
  result := TFhirConceptMapElementTargetDependsOn.create;
end;

function TFhirResourceFactory.newConceptMap : TFhirConceptMap;
begin
  result := TFhirConceptMap.create;
end;

function TFhirResourceFactory.newConditionStage : TFhirConditionStage;
begin
  result := TFhirConditionStage.create;
end;

function TFhirResourceFactory.newConditionEvidence : TFhirConditionEvidence;
begin
  result := TFhirConditionEvidence.create;
end;

function TFhirResourceFactory.newCondition : TFhirCondition;
begin
  result := TFhirCondition.create;
end;

function TFhirResourceFactory.newConformanceContact : TFhirConformanceContact;
begin
  result := TFhirConformanceContact.create;
end;

function TFhirResourceFactory.newConformanceSoftware : TFhirConformanceSoftware;
begin
  result := TFhirConformanceSoftware.create;
end;

function TFhirResourceFactory.newConformanceImplementation : TFhirConformanceImplementation;
begin
  result := TFhirConformanceImplementation.create;
end;

function TFhirResourceFactory.newConformanceRest : TFhirConformanceRest;
begin
  result := TFhirConformanceRest.create;
end;

function TFhirResourceFactory.newConformanceRestSecurity : TFhirConformanceRestSecurity;
begin
  result := TFhirConformanceRestSecurity.create;
end;

function TFhirResourceFactory.newConformanceRestSecurityCertificate : TFhirConformanceRestSecurityCertificate;
begin
  result := TFhirConformanceRestSecurityCertificate.create;
end;

function TFhirResourceFactory.newConformanceRestResource : TFhirConformanceRestResource;
begin
  result := TFhirConformanceRestResource.create;
end;

function TFhirResourceFactory.newConformanceRestResourceInteraction : TFhirConformanceRestResourceInteraction;
begin
  result := TFhirConformanceRestResourceInteraction.create;
end;

function TFhirResourceFactory.newConformanceRestResourceSearchParam : TFhirConformanceRestResourceSearchParam;
begin
  result := TFhirConformanceRestResourceSearchParam.create;
end;

function TFhirResourceFactory.newConformanceRestInteraction : TFhirConformanceRestInteraction;
begin
  result := TFhirConformanceRestInteraction.create;
end;

function TFhirResourceFactory.newConformanceRestOperation : TFhirConformanceRestOperation;
begin
  result := TFhirConformanceRestOperation.create;
end;

function TFhirResourceFactory.newConformanceMessaging : TFhirConformanceMessaging;
begin
  result := TFhirConformanceMessaging.create;
end;

function TFhirResourceFactory.newConformanceMessagingEndpoint : TFhirConformanceMessagingEndpoint;
begin
  result := TFhirConformanceMessagingEndpoint.create;
end;

function TFhirResourceFactory.newConformanceMessagingEvent : TFhirConformanceMessagingEvent;
begin
  result := TFhirConformanceMessagingEvent.create;
end;

function TFhirResourceFactory.newConformanceDocument : TFhirConformanceDocument;
begin
  result := TFhirConformanceDocument.create;
end;

function TFhirResourceFactory.newConformance : TFhirConformance;
begin
  result := TFhirConformance.create;
end;

function TFhirResourceFactory.newContractActor : TFhirContractActor;
begin
  result := TFhirContractActor.create;
end;

function TFhirResourceFactory.newContractValuedItem : TFhirContractValuedItem;
begin
  result := TFhirContractValuedItem.create;
end;

function TFhirResourceFactory.newContractSigner : TFhirContractSigner;
begin
  result := TFhirContractSigner.create;
end;

function TFhirResourceFactory.newContractTerm : TFhirContractTerm;
begin
  result := TFhirContractTerm.create;
end;

function TFhirResourceFactory.newContractTermActor : TFhirContractTermActor;
begin
  result := TFhirContractTermActor.create;
end;

function TFhirResourceFactory.newContractTermValuedItem : TFhirContractTermValuedItem;
begin
  result := TFhirContractTermValuedItem.create;
end;

function TFhirResourceFactory.newContractFriendly : TFhirContractFriendly;
begin
  result := TFhirContractFriendly.create;
end;

function TFhirResourceFactory.newContractLegal : TFhirContractLegal;
begin
  result := TFhirContractLegal.create;
end;

function TFhirResourceFactory.newContractRule : TFhirContractRule;
begin
  result := TFhirContractRule.create;
end;

function TFhirResourceFactory.newContract : TFhirContract;
begin
  result := TFhirContract.create;
end;

function TFhirResourceFactory.newCoverage : TFhirCoverage;
begin
  result := TFhirCoverage.create;
end;

function TFhirResourceFactory.newDataElementContact : TFhirDataElementContact;
begin
  result := TFhirDataElementContact.create;
end;

function TFhirResourceFactory.newDataElementMapping : TFhirDataElementMapping;
begin
  result := TFhirDataElementMapping.create;
end;

function TFhirResourceFactory.newDataElement : TFhirDataElement;
begin
  result := TFhirDataElement.create;
end;

function TFhirResourceFactory.newDetectedIssueMitigation : TFhirDetectedIssueMitigation;
begin
  result := TFhirDetectedIssueMitigation.create;
end;

function TFhirResourceFactory.newDetectedIssue : TFhirDetectedIssue;
begin
  result := TFhirDetectedIssue.create;
end;

function TFhirResourceFactory.newDevice : TFhirDevice;
begin
  result := TFhirDevice.create;
end;

function TFhirResourceFactory.newDeviceComponentProductionSpecification : TFhirDeviceComponentProductionSpecification;
begin
  result := TFhirDeviceComponentProductionSpecification.create;
end;

function TFhirResourceFactory.newDeviceComponent : TFhirDeviceComponent;
begin
  result := TFhirDeviceComponent.create;
end;

function TFhirResourceFactory.newDeviceMetricCalibration : TFhirDeviceMetricCalibration;
begin
  result := TFhirDeviceMetricCalibration.create;
end;

function TFhirResourceFactory.newDeviceMetric : TFhirDeviceMetric;
begin
  result := TFhirDeviceMetric.create;
end;

function TFhirResourceFactory.newDeviceUseRequest : TFhirDeviceUseRequest;
begin
  result := TFhirDeviceUseRequest.create;
end;

function TFhirResourceFactory.newDeviceUseStatement : TFhirDeviceUseStatement;
begin
  result := TFhirDeviceUseStatement.create;
end;

function TFhirResourceFactory.newDiagnosticOrderEvent : TFhirDiagnosticOrderEvent;
begin
  result := TFhirDiagnosticOrderEvent.create;
end;

function TFhirResourceFactory.newDiagnosticOrderItem : TFhirDiagnosticOrderItem;
begin
  result := TFhirDiagnosticOrderItem.create;
end;

function TFhirResourceFactory.newDiagnosticOrder : TFhirDiagnosticOrder;
begin
  result := TFhirDiagnosticOrder.create;
end;

function TFhirResourceFactory.newDiagnosticReportImage : TFhirDiagnosticReportImage;
begin
  result := TFhirDiagnosticReportImage.create;
end;

function TFhirResourceFactory.newDiagnosticReport : TFhirDiagnosticReport;
begin
  result := TFhirDiagnosticReport.create;
end;

function TFhirResourceFactory.newDocumentManifestContent : TFhirDocumentManifestContent;
begin
  result := TFhirDocumentManifestContent.create;
end;

function TFhirResourceFactory.newDocumentManifestRelated : TFhirDocumentManifestRelated;
begin
  result := TFhirDocumentManifestRelated.create;
end;

function TFhirResourceFactory.newDocumentManifest : TFhirDocumentManifest;
begin
  result := TFhirDocumentManifest.create;
end;

function TFhirResourceFactory.newDocumentReferenceRelatesTo : TFhirDocumentReferenceRelatesTo;
begin
  result := TFhirDocumentReferenceRelatesTo.create;
end;

function TFhirResourceFactory.newDocumentReferenceContent : TFhirDocumentReferenceContent;
begin
  result := TFhirDocumentReferenceContent.create;
end;

function TFhirResourceFactory.newDocumentReferenceContext : TFhirDocumentReferenceContext;
begin
  result := TFhirDocumentReferenceContext.create;
end;

function TFhirResourceFactory.newDocumentReferenceContextRelated : TFhirDocumentReferenceContextRelated;
begin
  result := TFhirDocumentReferenceContextRelated.create;
end;

function TFhirResourceFactory.newDocumentReference : TFhirDocumentReference;
begin
  result := TFhirDocumentReference.create;
end;

function TFhirResourceFactory.newEligibilityRequest : TFhirEligibilityRequest;
begin
  result := TFhirEligibilityRequest.create;
end;

function TFhirResourceFactory.newEligibilityResponse : TFhirEligibilityResponse;
begin
  result := TFhirEligibilityResponse.create;
end;

function TFhirResourceFactory.newEncounterStatusHistory : TFhirEncounterStatusHistory;
begin
  result := TFhirEncounterStatusHistory.create;
end;

function TFhirResourceFactory.newEncounterParticipant : TFhirEncounterParticipant;
begin
  result := TFhirEncounterParticipant.create;
end;

function TFhirResourceFactory.newEncounterHospitalization : TFhirEncounterHospitalization;
begin
  result := TFhirEncounterHospitalization.create;
end;

function TFhirResourceFactory.newEncounterLocation : TFhirEncounterLocation;
begin
  result := TFhirEncounterLocation.create;
end;

function TFhirResourceFactory.newEncounter : TFhirEncounter;
begin
  result := TFhirEncounter.create;
end;

function TFhirResourceFactory.newEnrollmentRequest : TFhirEnrollmentRequest;
begin
  result := TFhirEnrollmentRequest.create;
end;

function TFhirResourceFactory.newEnrollmentResponse : TFhirEnrollmentResponse;
begin
  result := TFhirEnrollmentResponse.create;
end;

function TFhirResourceFactory.newEpisodeOfCareStatusHistory : TFhirEpisodeOfCareStatusHistory;
begin
  result := TFhirEpisodeOfCareStatusHistory.create;
end;

function TFhirResourceFactory.newEpisodeOfCareCareTeam : TFhirEpisodeOfCareCareTeam;
begin
  result := TFhirEpisodeOfCareCareTeam.create;
end;

function TFhirResourceFactory.newEpisodeOfCare : TFhirEpisodeOfCare;
begin
  result := TFhirEpisodeOfCare.create;
end;

function TFhirResourceFactory.newExplanationOfBenefit : TFhirExplanationOfBenefit;
begin
  result := TFhirExplanationOfBenefit.create;
end;

function TFhirResourceFactory.newFamilyMemberHistoryCondition : TFhirFamilyMemberHistoryCondition;
begin
  result := TFhirFamilyMemberHistoryCondition.create;
end;

function TFhirResourceFactory.newFamilyMemberHistory : TFhirFamilyMemberHistory;
begin
  result := TFhirFamilyMemberHistory.create;
end;

function TFhirResourceFactory.newFlag : TFhirFlag;
begin
  result := TFhirFlag.create;
end;

function TFhirResourceFactory.newGoalOutcome : TFhirGoalOutcome;
begin
  result := TFhirGoalOutcome.create;
end;

function TFhirResourceFactory.newGoal : TFhirGoal;
begin
  result := TFhirGoal.create;
end;

function TFhirResourceFactory.newGroupCharacteristic : TFhirGroupCharacteristic;
begin
  result := TFhirGroupCharacteristic.create;
end;

function TFhirResourceFactory.newGroupMember : TFhirGroupMember;
begin
  result := TFhirGroupMember.create;
end;

function TFhirResourceFactory.newGroup : TFhirGroup;
begin
  result := TFhirGroup.create;
end;

function TFhirResourceFactory.newHealthcareServiceServiceType : TFhirHealthcareServiceServiceType;
begin
  result := TFhirHealthcareServiceServiceType.create;
end;

function TFhirResourceFactory.newHealthcareServiceAvailableTime : TFhirHealthcareServiceAvailableTime;
begin
  result := TFhirHealthcareServiceAvailableTime.create;
end;

function TFhirResourceFactory.newHealthcareServiceNotAvailable : TFhirHealthcareServiceNotAvailable;
begin
  result := TFhirHealthcareServiceNotAvailable.create;
end;

function TFhirResourceFactory.newHealthcareService : TFhirHealthcareService;
begin
  result := TFhirHealthcareService.create;
end;

function TFhirResourceFactory.newImagingObjectSelectionStudy : TFhirImagingObjectSelectionStudy;
begin
  result := TFhirImagingObjectSelectionStudy.create;
end;

function TFhirResourceFactory.newImagingObjectSelectionStudySeries : TFhirImagingObjectSelectionStudySeries;
begin
  result := TFhirImagingObjectSelectionStudySeries.create;
end;

function TFhirResourceFactory.newImagingObjectSelectionStudySeriesInstance : TFhirImagingObjectSelectionStudySeriesInstance;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstance.create;
end;

function TFhirResourceFactory.newImagingObjectSelectionStudySeriesInstanceFrames : TFhirImagingObjectSelectionStudySeriesInstanceFrames;
begin
  result := TFhirImagingObjectSelectionStudySeriesInstanceFrames.create;
end;

function TFhirResourceFactory.newImagingObjectSelection : TFhirImagingObjectSelection;
begin
  result := TFhirImagingObjectSelection.create;
end;

function TFhirResourceFactory.newImagingStudySeries : TFhirImagingStudySeries;
begin
  result := TFhirImagingStudySeries.create;
end;

function TFhirResourceFactory.newImagingStudySeriesInstance : TFhirImagingStudySeriesInstance;
begin
  result := TFhirImagingStudySeriesInstance.create;
end;

function TFhirResourceFactory.newImagingStudy : TFhirImagingStudy;
begin
  result := TFhirImagingStudy.create;
end;

function TFhirResourceFactory.newImmunizationExplanation : TFhirImmunizationExplanation;
begin
  result := TFhirImmunizationExplanation.create;
end;

function TFhirResourceFactory.newImmunizationReaction : TFhirImmunizationReaction;
begin
  result := TFhirImmunizationReaction.create;
end;

function TFhirResourceFactory.newImmunizationVaccinationProtocol : TFhirImmunizationVaccinationProtocol;
begin
  result := TFhirImmunizationVaccinationProtocol.create;
end;

function TFhirResourceFactory.newImmunization : TFhirImmunization;
begin
  result := TFhirImmunization.create;
end;

function TFhirResourceFactory.newImmunizationRecommendationRecommendation : TFhirImmunizationRecommendationRecommendation;
begin
  result := TFhirImmunizationRecommendationRecommendation.create;
end;

function TFhirResourceFactory.newImmunizationRecommendationRecommendationDateCriterion : TFhirImmunizationRecommendationRecommendationDateCriterion;
begin
  result := TFhirImmunizationRecommendationRecommendationDateCriterion.create;
end;

function TFhirResourceFactory.newImmunizationRecommendationRecommendationProtocol : TFhirImmunizationRecommendationRecommendationProtocol;
begin
  result := TFhirImmunizationRecommendationRecommendationProtocol.create;
end;

function TFhirResourceFactory.newImmunizationRecommendation : TFhirImmunizationRecommendation;
begin
  result := TFhirImmunizationRecommendation.create;
end;

function TFhirResourceFactory.newImplementationGuideContact : TFhirImplementationGuideContact;
begin
  result := TFhirImplementationGuideContact.create;
end;

function TFhirResourceFactory.newImplementationGuideDependency : TFhirImplementationGuideDependency;
begin
  result := TFhirImplementationGuideDependency.create;
end;

function TFhirResourceFactory.newImplementationGuidePackage : TFhirImplementationGuidePackage;
begin
  result := TFhirImplementationGuidePackage.create;
end;

function TFhirResourceFactory.newImplementationGuidePackageResource : TFhirImplementationGuidePackageResource;
begin
  result := TFhirImplementationGuidePackageResource.create;
end;

function TFhirResourceFactory.newImplementationGuideGlobal : TFhirImplementationGuideGlobal;
begin
  result := TFhirImplementationGuideGlobal.create;
end;

function TFhirResourceFactory.newImplementationGuidePage : TFhirImplementationGuidePage;
begin
  result := TFhirImplementationGuidePage.create;
end;

function TFhirResourceFactory.newImplementationGuide : TFhirImplementationGuide;
begin
  result := TFhirImplementationGuide.create;
end;

function TFhirResourceFactory.newListEntry : TFhirListEntry;
begin
  result := TFhirListEntry.create;
end;

function TFhirResourceFactory.newList : TFhirList;
begin
  result := TFhirList.create;
end;

function TFhirResourceFactory.newLocationPosition : TFhirLocationPosition;
begin
  result := TFhirLocationPosition.create;
end;

function TFhirResourceFactory.newLocation : TFhirLocation;
begin
  result := TFhirLocation.create;
end;

function TFhirResourceFactory.newMedia : TFhirMedia;
begin
  result := TFhirMedia.create;
end;

function TFhirResourceFactory.newMedicationProduct : TFhirMedicationProduct;
begin
  result := TFhirMedicationProduct.create;
end;

function TFhirResourceFactory.newMedicationProductIngredient : TFhirMedicationProductIngredient;
begin
  result := TFhirMedicationProductIngredient.create;
end;

function TFhirResourceFactory.newMedicationProductBatch : TFhirMedicationProductBatch;
begin
  result := TFhirMedicationProductBatch.create;
end;

function TFhirResourceFactory.newMedicationPackage : TFhirMedicationPackage;
begin
  result := TFhirMedicationPackage.create;
end;

function TFhirResourceFactory.newMedicationPackageContent : TFhirMedicationPackageContent;
begin
  result := TFhirMedicationPackageContent.create;
end;

function TFhirResourceFactory.newMedication : TFhirMedication;
begin
  result := TFhirMedication.create;
end;

function TFhirResourceFactory.newMedicationAdministrationDosage : TFhirMedicationAdministrationDosage;
begin
  result := TFhirMedicationAdministrationDosage.create;
end;

function TFhirResourceFactory.newMedicationAdministration : TFhirMedicationAdministration;
begin
  result := TFhirMedicationAdministration.create;
end;

function TFhirResourceFactory.newMedicationDispenseDosageInstruction : TFhirMedicationDispenseDosageInstruction;
begin
  result := TFhirMedicationDispenseDosageInstruction.create;
end;

function TFhirResourceFactory.newMedicationDispenseSubstitution : TFhirMedicationDispenseSubstitution;
begin
  result := TFhirMedicationDispenseSubstitution.create;
end;

function TFhirResourceFactory.newMedicationDispense : TFhirMedicationDispense;
begin
  result := TFhirMedicationDispense.create;
end;

function TFhirResourceFactory.newMedicationOrderDosageInstruction : TFhirMedicationOrderDosageInstruction;
begin
  result := TFhirMedicationOrderDosageInstruction.create;
end;

function TFhirResourceFactory.newMedicationOrderDispenseRequest : TFhirMedicationOrderDispenseRequest;
begin
  result := TFhirMedicationOrderDispenseRequest.create;
end;

function TFhirResourceFactory.newMedicationOrderSubstitution : TFhirMedicationOrderSubstitution;
begin
  result := TFhirMedicationOrderSubstitution.create;
end;

function TFhirResourceFactory.newMedicationOrder : TFhirMedicationOrder;
begin
  result := TFhirMedicationOrder.create;
end;

function TFhirResourceFactory.newMedicationStatementDosage : TFhirMedicationStatementDosage;
begin
  result := TFhirMedicationStatementDosage.create;
end;

function TFhirResourceFactory.newMedicationStatement : TFhirMedicationStatement;
begin
  result := TFhirMedicationStatement.create;
end;

function TFhirResourceFactory.newMessageHeaderResponse : TFhirMessageHeaderResponse;
begin
  result := TFhirMessageHeaderResponse.create;
end;

function TFhirResourceFactory.newMessageHeaderSource : TFhirMessageHeaderSource;
begin
  result := TFhirMessageHeaderSource.create;
end;

function TFhirResourceFactory.newMessageHeaderDestination : TFhirMessageHeaderDestination;
begin
  result := TFhirMessageHeaderDestination.create;
end;

function TFhirResourceFactory.newMessageHeader : TFhirMessageHeader;
begin
  result := TFhirMessageHeader.create;
end;

function TFhirResourceFactory.newNamingSystemContact : TFhirNamingSystemContact;
begin
  result := TFhirNamingSystemContact.create;
end;

function TFhirResourceFactory.newNamingSystemUniqueId : TFhirNamingSystemUniqueId;
begin
  result := TFhirNamingSystemUniqueId.create;
end;

function TFhirResourceFactory.newNamingSystem : TFhirNamingSystem;
begin
  result := TFhirNamingSystem.create;
end;

function TFhirResourceFactory.newNutritionOrderOralDiet : TFhirNutritionOrderOralDiet;
begin
  result := TFhirNutritionOrderOralDiet.create;
end;

function TFhirResourceFactory.newNutritionOrderOralDietNutrient : TFhirNutritionOrderOralDietNutrient;
begin
  result := TFhirNutritionOrderOralDietNutrient.create;
end;

function TFhirResourceFactory.newNutritionOrderOralDietTexture : TFhirNutritionOrderOralDietTexture;
begin
  result := TFhirNutritionOrderOralDietTexture.create;
end;

function TFhirResourceFactory.newNutritionOrderSupplement : TFhirNutritionOrderSupplement;
begin
  result := TFhirNutritionOrderSupplement.create;
end;

function TFhirResourceFactory.newNutritionOrderEnteralFormula : TFhirNutritionOrderEnteralFormula;
begin
  result := TFhirNutritionOrderEnteralFormula.create;
end;

function TFhirResourceFactory.newNutritionOrderEnteralFormulaAdministration : TFhirNutritionOrderEnteralFormulaAdministration;
begin
  result := TFhirNutritionOrderEnteralFormulaAdministration.create;
end;

function TFhirResourceFactory.newNutritionOrder : TFhirNutritionOrder;
begin
  result := TFhirNutritionOrder.create;
end;

function TFhirResourceFactory.newObservationReferenceRange : TFhirObservationReferenceRange;
begin
  result := TFhirObservationReferenceRange.create;
end;

function TFhirResourceFactory.newObservationRelated : TFhirObservationRelated;
begin
  result := TFhirObservationRelated.create;
end;

function TFhirResourceFactory.newObservationComponent : TFhirObservationComponent;
begin
  result := TFhirObservationComponent.create;
end;

function TFhirResourceFactory.newObservation : TFhirObservation;
begin
  result := TFhirObservation.create;
end;

function TFhirResourceFactory.newOperationDefinitionContact : TFhirOperationDefinitionContact;
begin
  result := TFhirOperationDefinitionContact.create;
end;

function TFhirResourceFactory.newOperationDefinitionParameter : TFhirOperationDefinitionParameter;
begin
  result := TFhirOperationDefinitionParameter.create;
end;

function TFhirResourceFactory.newOperationDefinitionParameterBinding : TFhirOperationDefinitionParameterBinding;
begin
  result := TFhirOperationDefinitionParameterBinding.create;
end;

function TFhirResourceFactory.newOperationDefinition : TFhirOperationDefinition;
begin
  result := TFhirOperationDefinition.create;
end;

function TFhirResourceFactory.newOperationOutcomeIssue : TFhirOperationOutcomeIssue;
begin
  result := TFhirOperationOutcomeIssue.create;
end;

function TFhirResourceFactory.newOperationOutcome : TFhirOperationOutcome;
begin
  result := TFhirOperationOutcome.create;
end;

function TFhirResourceFactory.newOrderWhen : TFhirOrderWhen;
begin
  result := TFhirOrderWhen.create;
end;

function TFhirResourceFactory.newOrder : TFhirOrder;
begin
  result := TFhirOrder.create;
end;

function TFhirResourceFactory.newOrderResponse : TFhirOrderResponse;
begin
  result := TFhirOrderResponse.create;
end;

function TFhirResourceFactory.newOrganizationContact : TFhirOrganizationContact;
begin
  result := TFhirOrganizationContact.create;
end;

function TFhirResourceFactory.newOrganization : TFhirOrganization;
begin
  result := TFhirOrganization.create;
end;

function TFhirResourceFactory.newPatientContact : TFhirPatientContact;
begin
  result := TFhirPatientContact.create;
end;

function TFhirResourceFactory.newPatientAnimal : TFhirPatientAnimal;
begin
  result := TFhirPatientAnimal.create;
end;

function TFhirResourceFactory.newPatientCommunication : TFhirPatientCommunication;
begin
  result := TFhirPatientCommunication.create;
end;

function TFhirResourceFactory.newPatientLink : TFhirPatientLink;
begin
  result := TFhirPatientLink.create;
end;

function TFhirResourceFactory.newPatient : TFhirPatient;
begin
  result := TFhirPatient.create;
end;

function TFhirResourceFactory.newPaymentNotice : TFhirPaymentNotice;
begin
  result := TFhirPaymentNotice.create;
end;

function TFhirResourceFactory.newPaymentReconciliationDetail : TFhirPaymentReconciliationDetail;
begin
  result := TFhirPaymentReconciliationDetail.create;
end;

function TFhirResourceFactory.newPaymentReconciliationNote : TFhirPaymentReconciliationNote;
begin
  result := TFhirPaymentReconciliationNote.create;
end;

function TFhirResourceFactory.newPaymentReconciliation : TFhirPaymentReconciliation;
begin
  result := TFhirPaymentReconciliation.create;
end;

function TFhirResourceFactory.newPersonLink : TFhirPersonLink;
begin
  result := TFhirPersonLink.create;
end;

function TFhirResourceFactory.newPerson : TFhirPerson;
begin
  result := TFhirPerson.create;
end;

function TFhirResourceFactory.newPractitionerPractitionerRole : TFhirPractitionerPractitionerRole;
begin
  result := TFhirPractitionerPractitionerRole.create;
end;

function TFhirResourceFactory.newPractitionerQualification : TFhirPractitionerQualification;
begin
  result := TFhirPractitionerQualification.create;
end;

function TFhirResourceFactory.newPractitioner : TFhirPractitioner;
begin
  result := TFhirPractitioner.create;
end;

function TFhirResourceFactory.newProcedurePerformer : TFhirProcedurePerformer;
begin
  result := TFhirProcedurePerformer.create;
end;

function TFhirResourceFactory.newProcedureFocalDevice : TFhirProcedureFocalDevice;
begin
  result := TFhirProcedureFocalDevice.create;
end;

function TFhirResourceFactory.newProcedure : TFhirProcedure;
begin
  result := TFhirProcedure.create;
end;

function TFhirResourceFactory.newProcedureRequest : TFhirProcedureRequest;
begin
  result := TFhirProcedureRequest.create;
end;

function TFhirResourceFactory.newProcessRequestItem : TFhirProcessRequestItem;
begin
  result := TFhirProcessRequestItem.create;
end;

function TFhirResourceFactory.newProcessRequest : TFhirProcessRequest;
begin
  result := TFhirProcessRequest.create;
end;

function TFhirResourceFactory.newProcessResponseNotes : TFhirProcessResponseNotes;
begin
  result := TFhirProcessResponseNotes.create;
end;

function TFhirResourceFactory.newProcessResponse : TFhirProcessResponse;
begin
  result := TFhirProcessResponse.create;
end;

function TFhirResourceFactory.newProvenanceAgent : TFhirProvenanceAgent;
begin
  result := TFhirProvenanceAgent.create;
end;

function TFhirResourceFactory.newProvenanceAgentRelatedAgent : TFhirProvenanceAgentRelatedAgent;
begin
  result := TFhirProvenanceAgentRelatedAgent.create;
end;

function TFhirResourceFactory.newProvenanceEntity : TFhirProvenanceEntity;
begin
  result := TFhirProvenanceEntity.create;
end;

function TFhirResourceFactory.newProvenance : TFhirProvenance;
begin
  result := TFhirProvenance.create;
end;

function TFhirResourceFactory.newQuestionnaireGroup : TFhirQuestionnaireGroup;
begin
  result := TFhirQuestionnaireGroup.create;
end;

function TFhirResourceFactory.newQuestionnaireGroupQuestion : TFhirQuestionnaireGroupQuestion;
begin
  result := TFhirQuestionnaireGroupQuestion.create;
end;

function TFhirResourceFactory.newQuestionnaire : TFhirQuestionnaire;
begin
  result := TFhirQuestionnaire.create;
end;

function TFhirResourceFactory.newQuestionnaireResponseGroup : TFhirQuestionnaireResponseGroup;
begin
  result := TFhirQuestionnaireResponseGroup.create;
end;

function TFhirResourceFactory.newQuestionnaireResponseGroupQuestion : TFhirQuestionnaireResponseGroupQuestion;
begin
  result := TFhirQuestionnaireResponseGroupQuestion.create;
end;

function TFhirResourceFactory.newQuestionnaireResponseGroupQuestionAnswer : TFhirQuestionnaireResponseGroupQuestionAnswer;
begin
  result := TFhirQuestionnaireResponseGroupQuestionAnswer.create;
end;

function TFhirResourceFactory.newQuestionnaireResponse : TFhirQuestionnaireResponse;
begin
  result := TFhirQuestionnaireResponse.create;
end;

function TFhirResourceFactory.newReferralRequest : TFhirReferralRequest;
begin
  result := TFhirReferralRequest.create;
end;

function TFhirResourceFactory.newRelatedPerson : TFhirRelatedPerson;
begin
  result := TFhirRelatedPerson.create;
end;

function TFhirResourceFactory.newRiskAssessmentPrediction : TFhirRiskAssessmentPrediction;
begin
  result := TFhirRiskAssessmentPrediction.create;
end;

function TFhirResourceFactory.newRiskAssessment : TFhirRiskAssessment;
begin
  result := TFhirRiskAssessment.create;
end;

function TFhirResourceFactory.newSchedule : TFhirSchedule;
begin
  result := TFhirSchedule.create;
end;

function TFhirResourceFactory.newSearchParameterContact : TFhirSearchParameterContact;
begin
  result := TFhirSearchParameterContact.create;
end;

function TFhirResourceFactory.newSearchParameter : TFhirSearchParameter;
begin
  result := TFhirSearchParameter.create;
end;

function TFhirResourceFactory.newSlot : TFhirSlot;
begin
  result := TFhirSlot.create;
end;

function TFhirResourceFactory.newSpecimenCollection : TFhirSpecimenCollection;
begin
  result := TFhirSpecimenCollection.create;
end;

function TFhirResourceFactory.newSpecimenTreatment : TFhirSpecimenTreatment;
begin
  result := TFhirSpecimenTreatment.create;
end;

function TFhirResourceFactory.newSpecimenContainer : TFhirSpecimenContainer;
begin
  result := TFhirSpecimenContainer.create;
end;

function TFhirResourceFactory.newSpecimen : TFhirSpecimen;
begin
  result := TFhirSpecimen.create;
end;

function TFhirResourceFactory.newStructureDefinitionContact : TFhirStructureDefinitionContact;
begin
  result := TFhirStructureDefinitionContact.create;
end;

function TFhirResourceFactory.newStructureDefinitionMapping : TFhirStructureDefinitionMapping;
begin
  result := TFhirStructureDefinitionMapping.create;
end;

function TFhirResourceFactory.newStructureDefinitionSnapshot : TFhirStructureDefinitionSnapshot;
begin
  result := TFhirStructureDefinitionSnapshot.create;
end;

function TFhirResourceFactory.newStructureDefinitionDifferential : TFhirStructureDefinitionDifferential;
begin
  result := TFhirStructureDefinitionDifferential.create;
end;

function TFhirResourceFactory.newStructureDefinition : TFhirStructureDefinition;
begin
  result := TFhirStructureDefinition.create;
end;

function TFhirResourceFactory.newSubscriptionChannel : TFhirSubscriptionChannel;
begin
  result := TFhirSubscriptionChannel.create;
end;

function TFhirResourceFactory.newSubscription : TFhirSubscription;
begin
  result := TFhirSubscription.create;
end;

function TFhirResourceFactory.newSubstanceInstance : TFhirSubstanceInstance;
begin
  result := TFhirSubstanceInstance.create;
end;

function TFhirResourceFactory.newSubstanceIngredient : TFhirSubstanceIngredient;
begin
  result := TFhirSubstanceIngredient.create;
end;

function TFhirResourceFactory.newSubstance : TFhirSubstance;
begin
  result := TFhirSubstance.create;
end;

function TFhirResourceFactory.newSupplyDelivery : TFhirSupplyDelivery;
begin
  result := TFhirSupplyDelivery.create;
end;

function TFhirResourceFactory.newSupplyRequestWhen : TFhirSupplyRequestWhen;
begin
  result := TFhirSupplyRequestWhen.create;
end;

function TFhirResourceFactory.newSupplyRequest : TFhirSupplyRequest;
begin
  result := TFhirSupplyRequest.create;
end;

function TFhirResourceFactory.newTestScriptContact : TFhirTestScriptContact;
begin
  result := TFhirTestScriptContact.create;
end;

function TFhirResourceFactory.newTestScriptMetadata : TFhirTestScriptMetadata;
begin
  result := TFhirTestScriptMetadata.create;
end;

function TFhirResourceFactory.newTestScriptMetadataLink : TFhirTestScriptMetadataLink;
begin
  result := TFhirTestScriptMetadataLink.create;
end;

function TFhirResourceFactory.newTestScriptMetadataCapability : TFhirTestScriptMetadataCapability;
begin
  result := TFhirTestScriptMetadataCapability.create;
end;

function TFhirResourceFactory.newTestScriptFixture : TFhirTestScriptFixture;
begin
  result := TFhirTestScriptFixture.create;
end;

function TFhirResourceFactory.newTestScriptVariable : TFhirTestScriptVariable;
begin
  result := TFhirTestScriptVariable.create;
end;

function TFhirResourceFactory.newTestScriptSetup : TFhirTestScriptSetup;
begin
  result := TFhirTestScriptSetup.create;
end;

function TFhirResourceFactory.newTestScriptSetupAction : TFhirTestScriptSetupAction;
begin
  result := TFhirTestScriptSetupAction.create;
end;

function TFhirResourceFactory.newTestScriptSetupActionOperation : TFhirTestScriptSetupActionOperation;
begin
  result := TFhirTestScriptSetupActionOperation.create;
end;

function TFhirResourceFactory.newTestScriptSetupActionOperationRequestHeader : TFhirTestScriptSetupActionOperationRequestHeader;
begin
  result := TFhirTestScriptSetupActionOperationRequestHeader.create;
end;

function TFhirResourceFactory.newTestScriptSetupActionAssert : TFhirTestScriptSetupActionAssert;
begin
  result := TFhirTestScriptSetupActionAssert.create;
end;

function TFhirResourceFactory.newTestScriptTest : TFhirTestScriptTest;
begin
  result := TFhirTestScriptTest.create;
end;

function TFhirResourceFactory.newTestScriptTestAction : TFhirTestScriptTestAction;
begin
  result := TFhirTestScriptTestAction.create;
end;

function TFhirResourceFactory.newTestScriptTeardown : TFhirTestScriptTeardown;
begin
  result := TFhirTestScriptTeardown.create;
end;

function TFhirResourceFactory.newTestScriptTeardownAction : TFhirTestScriptTeardownAction;
begin
  result := TFhirTestScriptTeardownAction.create;
end;

function TFhirResourceFactory.newTestScript : TFhirTestScript;
begin
  result := TFhirTestScript.create;
end;

function TFhirResourceFactory.newValueSetContact : TFhirValueSetContact;
begin
  result := TFhirValueSetContact.create;
end;

function TFhirResourceFactory.newValueSetCodeSystem : TFhirValueSetCodeSystem;
begin
  result := TFhirValueSetCodeSystem.create;
end;

function TFhirResourceFactory.newValueSetCodeSystemConcept : TFhirValueSetCodeSystemConcept;
begin
  result := TFhirValueSetCodeSystemConcept.create;
end;

function TFhirResourceFactory.newValueSetCodeSystemConceptDesignation : TFhirValueSetCodeSystemConceptDesignation;
begin
  result := TFhirValueSetCodeSystemConceptDesignation.create;
end;

function TFhirResourceFactory.newValueSetCompose : TFhirValueSetCompose;
begin
  result := TFhirValueSetCompose.create;
end;

function TFhirResourceFactory.newValueSetComposeInclude : TFhirValueSetComposeInclude;
begin
  result := TFhirValueSetComposeInclude.create;
end;

function TFhirResourceFactory.newValueSetComposeIncludeConcept : TFhirValueSetComposeIncludeConcept;
begin
  result := TFhirValueSetComposeIncludeConcept.create;
end;

function TFhirResourceFactory.newValueSetComposeIncludeFilter : TFhirValueSetComposeIncludeFilter;
begin
  result := TFhirValueSetComposeIncludeFilter.create;
end;

function TFhirResourceFactory.newValueSetExpansion : TFhirValueSetExpansion;
begin
  result := TFhirValueSetExpansion.create;
end;

function TFhirResourceFactory.newValueSetExpansionParameter : TFhirValueSetExpansionParameter;
begin
  result := TFhirValueSetExpansionParameter.create;
end;

function TFhirResourceFactory.newValueSetExpansionContains : TFhirValueSetExpansionContains;
begin
  result := TFhirValueSetExpansionContains.create;
end;

function TFhirResourceFactory.newValueSet : TFhirValueSet;
begin
  result := TFhirValueSet.create;
end;

function TFhirResourceFactory.newVisionPrescriptionDispense : TFhirVisionPrescriptionDispense;
begin
  result := TFhirVisionPrescriptionDispense.create;
end;

function TFhirResourceFactory.newVisionPrescription : TFhirVisionPrescription;
begin
  result := TFhirVisionPrescription.create;
end;


function TFHIRResourceFactory.makeByName(const name : String) : TFHIRBase;
begin
  if name = 'Element' then
    result := newElement()
  else if name = 'BackboneElement' then
    result := newBackboneElement()
  else if name = 'Enum' then
    result := newEnum()
  else if name = 'Integer' then
    result := newInteger()
  else if name = 'DateTime' then
    result := newDateTime()
  else if name = 'Date' then
    result := newDate()
  else if name = 'Decimal' then
    result := newDecimal()
  else if name = 'Uri' then
    result := newUri()
  else if name = 'Base64Binary' then
    result := newBase64Binary()
  else if name = 'Time' then
    result := newTime()
  else if name = 'String' then
    result := newString()
  else if name = 'Boolean' then
    result := newBoolean()
  else if name = 'Instant' then
    result := newInstant()
  else if name = 'Markdown' then
    result := newMarkdown()
  else if name = 'UnsignedInt' then
    result := newUnsignedInt()
  else if name = 'Code' then
    result := newCode()
  else if name = 'Id' then
    result := newId()
  else if name = 'Oid' then
    result := newOid()
  else if name = 'PositiveInt' then
    result := newPositiveInt()
  else if name = 'Uuid' then
    result := newUuid()
  else if name = 'Parameters.parameter' then
    result := newParametersParameter()
  else if name = 'Parameters' then
    result := newParameters()
  else if name = 'Extension' then
    result := newExtension()
  else if name = 'Narrative' then
    result := newNarrative()
  else if name = 'Period' then
    result := newPeriod()
  else if name = 'Coding' then
    result := newCoding()
  else if name = 'Range' then
    result := newRange()
  else if name = 'Quantity' then
    result := newQuantity()
  else if name = 'Attachment' then
    result := newAttachment()
  else if name = 'Ratio' then
    result := newRatio()
  else if name = 'Annotation' then
    result := newAnnotation()
  else if name = 'SampledData' then
    result := newSampledData()
  else if name = 'Reference' then
    result := newReference()
  else if name = 'CodeableConcept' then
    result := newCodeableConcept()
  else if name = 'Identifier' then
    result := newIdentifier()
  else if name = 'Signature' then
    result := newSignature()
  else if name = 'ElementDefinition.slicing' then
    result := newElementDefinitionSlicing()
  else if name = 'ElementDefinition.base' then
    result := newElementDefinitionBase()
  else if name = 'ElementDefinition.type' then
    result := newElementDefinitionType()
  else if name = 'ElementDefinition.constraint' then
    result := newElementDefinitionConstraint()
  else if name = 'ElementDefinition.binding' then
    result := newElementDefinitionBinding()
  else if name = 'ElementDefinition.mapping' then
    result := newElementDefinitionMapping()
  else if name = 'ElementDefinition' then
    result := newElementDefinition()
  else if name = 'Timing.repeat' then
    result := newTimingRepeat()
  else if name = 'Timing' then
    result := newTiming()
  else if name = 'Address' then
    result := newAddress()
  else if name = 'HumanName' then
    result := newHumanName()
  else if name = 'Meta' then
    result := newMeta()
  else if name = 'ContactPoint' then
    result := newContactPoint()
  else if name = 'Account' then
    result := newAccount()
  else if name = 'AllergyIntolerance.reaction' then
    result := newAllergyIntoleranceReaction()
  else if name = 'AllergyIntolerance' then
    result := newAllergyIntolerance()
  else if name = 'Appointment.participant' then
    result := newAppointmentParticipant()
  else if name = 'Appointment' then
    result := newAppointment()
  else if name = 'AppointmentResponse' then
    result := newAppointmentResponse()
  else if name = 'AuditEvent.event' then
    result := newAuditEventEvent()
  else if name = 'AuditEvent.participant' then
    result := newAuditEventParticipant()
  else if name = 'AuditEvent.participant.network' then
    result := newAuditEventParticipantNetwork()
  else if name = 'AuditEvent.source' then
    result := newAuditEventSource()
  else if name = 'AuditEvent.object' then
    result := newAuditEventObject()
  else if name = 'AuditEvent.object.detail' then
    result := newAuditEventObjectDetail()
  else if name = 'AuditEvent' then
    result := newAuditEvent()
  else if name = 'Basic' then
    result := newBasic()
  else if name = 'Binary' then
    result := newBinary()
  else if name = 'BodySite' then
    result := newBodySite()
  else if name = 'Bundle.link' then
    result := newBundleLink()
  else if name = 'Bundle.entry' then
    result := newBundleEntry()
  else if name = 'Bundle.entry.search' then
    result := newBundleEntrySearch()
  else if name = 'Bundle.entry.request' then
    result := newBundleEntryRequest()
  else if name = 'Bundle.entry.response' then
    result := newBundleEntryResponse()
  else if name = 'Bundle' then
    result := newBundle()
  else if name = 'CarePlan.relatedPlan' then
    result := newCarePlanRelatedPlan()
  else if name = 'CarePlan.participant' then
    result := newCarePlanParticipant()
  else if name = 'CarePlan.activity' then
    result := newCarePlanActivity()
  else if name = 'CarePlan.activity.detail' then
    result := newCarePlanActivityDetail()
  else if name = 'CarePlan' then
    result := newCarePlan()
  else if name = 'Claim.payee' then
    result := newClaimPayee()
  else if name = 'Claim.diagnosis' then
    result := newClaimDiagnosis()
  else if name = 'Claim.coverage' then
    result := newClaimCoverage()
  else if name = 'Claim.item' then
    result := newClaimItem()
  else if name = 'Claim.item.detail' then
    result := newClaimItemDetail()
  else if name = 'Claim.item.detail.subDetail' then
    result := newClaimItemDetailSubDetail()
  else if name = 'Claim.item.prosthesis' then
    result := newClaimItemProsthesis()
  else if name = 'Claim.missingTeeth' then
    result := newClaimMissingTeeth()
  else if name = 'Claim' then
    result := newClaim()
  else if name = 'ClaimResponse.item' then
    result := newClaimResponseItem()
  else if name = 'ClaimResponse.item.adjudication' then
    result := newClaimResponseItemAdjudication()
  else if name = 'ClaimResponse.item.detail' then
    result := newClaimResponseItemDetail()
  else if name = 'ClaimResponse.item.detail.adjudication' then
    result := newClaimResponseItemDetailAdjudication()
  else if name = 'ClaimResponse.item.detail.subDetail' then
    result := newClaimResponseItemDetailSubDetail()
  else if name = 'ClaimResponse.item.detail.subDetail.adjudication' then
    result := newClaimResponseItemDetailSubDetailAdjudication()
  else if name = 'ClaimResponse.addItem' then
    result := newClaimResponseAddItem()
  else if name = 'ClaimResponse.addItem.adjudication' then
    result := newClaimResponseAddItemAdjudication()
  else if name = 'ClaimResponse.addItem.detail' then
    result := newClaimResponseAddItemDetail()
  else if name = 'ClaimResponse.addItem.detail.adjudication' then
    result := newClaimResponseAddItemDetailAdjudication()
  else if name = 'ClaimResponse.error' then
    result := newClaimResponseError()
  else if name = 'ClaimResponse.note' then
    result := newClaimResponseNote()
  else if name = 'ClaimResponse.coverage' then
    result := newClaimResponseCoverage()
  else if name = 'ClaimResponse' then
    result := newClaimResponse()
  else if name = 'ClinicalImpression.investigations' then
    result := newClinicalImpressionInvestigations()
  else if name = 'ClinicalImpression.finding' then
    result := newClinicalImpressionFinding()
  else if name = 'ClinicalImpression.ruledOut' then
    result := newClinicalImpressionRuledOut()
  else if name = 'ClinicalImpression' then
    result := newClinicalImpression()
  else if name = 'Communication.payload' then
    result := newCommunicationPayload()
  else if name = 'Communication' then
    result := newCommunication()
  else if name = 'CommunicationRequest.payload' then
    result := newCommunicationRequestPayload()
  else if name = 'CommunicationRequest' then
    result := newCommunicationRequest()
  else if name = 'Composition.attester' then
    result := newCompositionAttester()
  else if name = 'Composition.event' then
    result := newCompositionEvent()
  else if name = 'Composition.section' then
    result := newCompositionSection()
  else if name = 'Composition' then
    result := newComposition()
  else if name = 'ConceptMap.contact' then
    result := newConceptMapContact()
  else if name = 'ConceptMap.element' then
    result := newConceptMapElement()
  else if name = 'ConceptMap.element.target' then
    result := newConceptMapElementTarget()
  else if name = 'ConceptMap.element.target.dependsOn' then
    result := newConceptMapElementTargetDependsOn()
  else if name = 'ConceptMap' then
    result := newConceptMap()
  else if name = 'Condition.stage' then
    result := newConditionStage()
  else if name = 'Condition.evidence' then
    result := newConditionEvidence()
  else if name = 'Condition' then
    result := newCondition()
  else if name = 'Conformance.contact' then
    result := newConformanceContact()
  else if name = 'Conformance.software' then
    result := newConformanceSoftware()
  else if name = 'Conformance.implementation' then
    result := newConformanceImplementation()
  else if name = 'Conformance.rest' then
    result := newConformanceRest()
  else if name = 'Conformance.rest.security' then
    result := newConformanceRestSecurity()
  else if name = 'Conformance.rest.security.certificate' then
    result := newConformanceRestSecurityCertificate()
  else if name = 'Conformance.rest.resource' then
    result := newConformanceRestResource()
  else if name = 'Conformance.rest.resource.interaction' then
    result := newConformanceRestResourceInteraction()
  else if name = 'Conformance.rest.resource.searchParam' then
    result := newConformanceRestResourceSearchParam()
  else if name = 'Conformance.rest.interaction' then
    result := newConformanceRestInteraction()
  else if name = 'Conformance.rest.operation' then
    result := newConformanceRestOperation()
  else if name = 'Conformance.messaging' then
    result := newConformanceMessaging()
  else if name = 'Conformance.messaging.endpoint' then
    result := newConformanceMessagingEndpoint()
  else if name = 'Conformance.messaging.event' then
    result := newConformanceMessagingEvent()
  else if name = 'Conformance.document' then
    result := newConformanceDocument()
  else if name = 'Conformance' then
    result := newConformance()
  else if name = 'Contract.actor' then
    result := newContractActor()
  else if name = 'Contract.valuedItem' then
    result := newContractValuedItem()
  else if name = 'Contract.signer' then
    result := newContractSigner()
  else if name = 'Contract.term' then
    result := newContractTerm()
  else if name = 'Contract.term.actor' then
    result := newContractTermActor()
  else if name = 'Contract.term.valuedItem' then
    result := newContractTermValuedItem()
  else if name = 'Contract.friendly' then
    result := newContractFriendly()
  else if name = 'Contract.legal' then
    result := newContractLegal()
  else if name = 'Contract.rule' then
    result := newContractRule()
  else if name = 'Contract' then
    result := newContract()
  else if name = 'Coverage' then
    result := newCoverage()
  else if name = 'DataElement.contact' then
    result := newDataElementContact()
  else if name = 'DataElement.mapping' then
    result := newDataElementMapping()
  else if name = 'DataElement' then
    result := newDataElement()
  else if name = 'DetectedIssue.mitigation' then
    result := newDetectedIssueMitigation()
  else if name = 'DetectedIssue' then
    result := newDetectedIssue()
  else if name = 'Device' then
    result := newDevice()
  else if name = 'DeviceComponent.productionSpecification' then
    result := newDeviceComponentProductionSpecification()
  else if name = 'DeviceComponent' then
    result := newDeviceComponent()
  else if name = 'DeviceMetric.calibration' then
    result := newDeviceMetricCalibration()
  else if name = 'DeviceMetric' then
    result := newDeviceMetric()
  else if name = 'DeviceUseRequest' then
    result := newDeviceUseRequest()
  else if name = 'DeviceUseStatement' then
    result := newDeviceUseStatement()
  else if name = 'DiagnosticOrder.event' then
    result := newDiagnosticOrderEvent()
  else if name = 'DiagnosticOrder.item' then
    result := newDiagnosticOrderItem()
  else if name = 'DiagnosticOrder' then
    result := newDiagnosticOrder()
  else if name = 'DiagnosticReport.image' then
    result := newDiagnosticReportImage()
  else if name = 'DiagnosticReport' then
    result := newDiagnosticReport()
  else if name = 'DocumentManifest.content' then
    result := newDocumentManifestContent()
  else if name = 'DocumentManifest.related' then
    result := newDocumentManifestRelated()
  else if name = 'DocumentManifest' then
    result := newDocumentManifest()
  else if name = 'DocumentReference.relatesTo' then
    result := newDocumentReferenceRelatesTo()
  else if name = 'DocumentReference.content' then
    result := newDocumentReferenceContent()
  else if name = 'DocumentReference.context' then
    result := newDocumentReferenceContext()
  else if name = 'DocumentReference.context.related' then
    result := newDocumentReferenceContextRelated()
  else if name = 'DocumentReference' then
    result := newDocumentReference()
  else if name = 'EligibilityRequest' then
    result := newEligibilityRequest()
  else if name = 'EligibilityResponse' then
    result := newEligibilityResponse()
  else if name = 'Encounter.statusHistory' then
    result := newEncounterStatusHistory()
  else if name = 'Encounter.participant' then
    result := newEncounterParticipant()
  else if name = 'Encounter.hospitalization' then
    result := newEncounterHospitalization()
  else if name = 'Encounter.location' then
    result := newEncounterLocation()
  else if name = 'Encounter' then
    result := newEncounter()
  else if name = 'EnrollmentRequest' then
    result := newEnrollmentRequest()
  else if name = 'EnrollmentResponse' then
    result := newEnrollmentResponse()
  else if name = 'EpisodeOfCare.statusHistory' then
    result := newEpisodeOfCareStatusHistory()
  else if name = 'EpisodeOfCare.careTeam' then
    result := newEpisodeOfCareCareTeam()
  else if name = 'EpisodeOfCare' then
    result := newEpisodeOfCare()
  else if name = 'ExplanationOfBenefit' then
    result := newExplanationOfBenefit()
  else if name = 'FamilyMemberHistory.condition' then
    result := newFamilyMemberHistoryCondition()
  else if name = 'FamilyMemberHistory' then
    result := newFamilyMemberHistory()
  else if name = 'Flag' then
    result := newFlag()
  else if name = 'Goal.outcome' then
    result := newGoalOutcome()
  else if name = 'Goal' then
    result := newGoal()
  else if name = 'Group.characteristic' then
    result := newGroupCharacteristic()
  else if name = 'Group.member' then
    result := newGroupMember()
  else if name = 'Group' then
    result := newGroup()
  else if name = 'HealthcareService.serviceType' then
    result := newHealthcareServiceServiceType()
  else if name = 'HealthcareService.availableTime' then
    result := newHealthcareServiceAvailableTime()
  else if name = 'HealthcareService.notAvailable' then
    result := newHealthcareServiceNotAvailable()
  else if name = 'HealthcareService' then
    result := newHealthcareService()
  else if name = 'ImagingObjectSelection.study' then
    result := newImagingObjectSelectionStudy()
  else if name = 'ImagingObjectSelection.study.series' then
    result := newImagingObjectSelectionStudySeries()
  else if name = 'ImagingObjectSelection.study.series.instance' then
    result := newImagingObjectSelectionStudySeriesInstance()
  else if name = 'ImagingObjectSelection.study.series.instance.frames' then
    result := newImagingObjectSelectionStudySeriesInstanceFrames()
  else if name = 'ImagingObjectSelection' then
    result := newImagingObjectSelection()
  else if name = 'ImagingStudy.series' then
    result := newImagingStudySeries()
  else if name = 'ImagingStudy.series.instance' then
    result := newImagingStudySeriesInstance()
  else if name = 'ImagingStudy' then
    result := newImagingStudy()
  else if name = 'Immunization.explanation' then
    result := newImmunizationExplanation()
  else if name = 'Immunization.reaction' then
    result := newImmunizationReaction()
  else if name = 'Immunization.vaccinationProtocol' then
    result := newImmunizationVaccinationProtocol()
  else if name = 'Immunization' then
    result := newImmunization()
  else if name = 'ImmunizationRecommendation.recommendation' then
    result := newImmunizationRecommendationRecommendation()
  else if name = 'ImmunizationRecommendation.recommendation.dateCriterion' then
    result := newImmunizationRecommendationRecommendationDateCriterion()
  else if name = 'ImmunizationRecommendation.recommendation.protocol' then
    result := newImmunizationRecommendationRecommendationProtocol()
  else if name = 'ImmunizationRecommendation' then
    result := newImmunizationRecommendation()
  else if name = 'ImplementationGuide.contact' then
    result := newImplementationGuideContact()
  else if name = 'ImplementationGuide.dependency' then
    result := newImplementationGuideDependency()
  else if name = 'ImplementationGuide.package' then
    result := newImplementationGuidePackage()
  else if name = 'ImplementationGuide.package.resource' then
    result := newImplementationGuidePackageResource()
  else if name = 'ImplementationGuide.global' then
    result := newImplementationGuideGlobal()
  else if name = 'ImplementationGuide.page' then
    result := newImplementationGuidePage()
  else if name = 'ImplementationGuide' then
    result := newImplementationGuide()
  else if name = 'List.entry' then
    result := newListEntry()
  else if name = 'List' then
    result := newList()
  else if name = 'Location.position' then
    result := newLocationPosition()
  else if name = 'Location' then
    result := newLocation()
  else if name = 'Media' then
    result := newMedia()
  else if name = 'Medication.product' then
    result := newMedicationProduct()
  else if name = 'Medication.product.ingredient' then
    result := newMedicationProductIngredient()
  else if name = 'Medication.product.batch' then
    result := newMedicationProductBatch()
  else if name = 'Medication.package' then
    result := newMedicationPackage()
  else if name = 'Medication.package.content' then
    result := newMedicationPackageContent()
  else if name = 'Medication' then
    result := newMedication()
  else if name = 'MedicationAdministration.dosage' then
    result := newMedicationAdministrationDosage()
  else if name = 'MedicationAdministration' then
    result := newMedicationAdministration()
  else if name = 'MedicationDispense.dosageInstruction' then
    result := newMedicationDispenseDosageInstruction()
  else if name = 'MedicationDispense.substitution' then
    result := newMedicationDispenseSubstitution()
  else if name = 'MedicationDispense' then
    result := newMedicationDispense()
  else if name = 'MedicationOrder.dosageInstruction' then
    result := newMedicationOrderDosageInstruction()
  else if name = 'MedicationOrder.dispenseRequest' then
    result := newMedicationOrderDispenseRequest()
  else if name = 'MedicationOrder.substitution' then
    result := newMedicationOrderSubstitution()
  else if name = 'MedicationOrder' then
    result := newMedicationOrder()
  else if name = 'MedicationStatement.dosage' then
    result := newMedicationStatementDosage()
  else if name = 'MedicationStatement' then
    result := newMedicationStatement()
  else if name = 'MessageHeader.response' then
    result := newMessageHeaderResponse()
  else if name = 'MessageHeader.source' then
    result := newMessageHeaderSource()
  else if name = 'MessageHeader.destination' then
    result := newMessageHeaderDestination()
  else if name = 'MessageHeader' then
    result := newMessageHeader()
  else if name = 'NamingSystem.contact' then
    result := newNamingSystemContact()
  else if name = 'NamingSystem.uniqueId' then
    result := newNamingSystemUniqueId()
  else if name = 'NamingSystem' then
    result := newNamingSystem()
  else if name = 'NutritionOrder.oralDiet' then
    result := newNutritionOrderOralDiet()
  else if name = 'NutritionOrder.oralDiet.nutrient' then
    result := newNutritionOrderOralDietNutrient()
  else if name = 'NutritionOrder.oralDiet.texture' then
    result := newNutritionOrderOralDietTexture()
  else if name = 'NutritionOrder.supplement' then
    result := newNutritionOrderSupplement()
  else if name = 'NutritionOrder.enteralFormula' then
    result := newNutritionOrderEnteralFormula()
  else if name = 'NutritionOrder.enteralFormula.administration' then
    result := newNutritionOrderEnteralFormulaAdministration()
  else if name = 'NutritionOrder' then
    result := newNutritionOrder()
  else if name = 'Observation.referenceRange' then
    result := newObservationReferenceRange()
  else if name = 'Observation.related' then
    result := newObservationRelated()
  else if name = 'Observation.component' then
    result := newObservationComponent()
  else if name = 'Observation' then
    result := newObservation()
  else if name = 'OperationDefinition.contact' then
    result := newOperationDefinitionContact()
  else if name = 'OperationDefinition.parameter' then
    result := newOperationDefinitionParameter()
  else if name = 'OperationDefinition.parameter.binding' then
    result := newOperationDefinitionParameterBinding()
  else if name = 'OperationDefinition' then
    result := newOperationDefinition()
  else if name = 'OperationOutcome.issue' then
    result := newOperationOutcomeIssue()
  else if name = 'OperationOutcome' then
    result := newOperationOutcome()
  else if name = 'Order.when' then
    result := newOrderWhen()
  else if name = 'Order' then
    result := newOrder()
  else if name = 'OrderResponse' then
    result := newOrderResponse()
  else if name = 'Organization.contact' then
    result := newOrganizationContact()
  else if name = 'Organization' then
    result := newOrganization()
  else if name = 'Patient.contact' then
    result := newPatientContact()
  else if name = 'Patient.animal' then
    result := newPatientAnimal()
  else if name = 'Patient.communication' then
    result := newPatientCommunication()
  else if name = 'Patient.link' then
    result := newPatientLink()
  else if name = 'Patient' then
    result := newPatient()
  else if name = 'PaymentNotice' then
    result := newPaymentNotice()
  else if name = 'PaymentReconciliation.detail' then
    result := newPaymentReconciliationDetail()
  else if name = 'PaymentReconciliation.note' then
    result := newPaymentReconciliationNote()
  else if name = 'PaymentReconciliation' then
    result := newPaymentReconciliation()
  else if name = 'Person.link' then
    result := newPersonLink()
  else if name = 'Person' then
    result := newPerson()
  else if name = 'Practitioner.practitionerRole' then
    result := newPractitionerPractitionerRole()
  else if name = 'Practitioner.qualification' then
    result := newPractitionerQualification()
  else if name = 'Practitioner' then
    result := newPractitioner()
  else if name = 'Procedure.performer' then
    result := newProcedurePerformer()
  else if name = 'Procedure.focalDevice' then
    result := newProcedureFocalDevice()
  else if name = 'Procedure' then
    result := newProcedure()
  else if name = 'ProcedureRequest' then
    result := newProcedureRequest()
  else if name = 'ProcessRequest.item' then
    result := newProcessRequestItem()
  else if name = 'ProcessRequest' then
    result := newProcessRequest()
  else if name = 'ProcessResponse.notes' then
    result := newProcessResponseNotes()
  else if name = 'ProcessResponse' then
    result := newProcessResponse()
  else if name = 'Provenance.agent' then
    result := newProvenanceAgent()
  else if name = 'Provenance.agent.relatedAgent' then
    result := newProvenanceAgentRelatedAgent()
  else if name = 'Provenance.entity' then
    result := newProvenanceEntity()
  else if name = 'Provenance' then
    result := newProvenance()
  else if name = 'Questionnaire.group' then
    result := newQuestionnaireGroup()
  else if name = 'Questionnaire.group.question' then
    result := newQuestionnaireGroupQuestion()
  else if name = 'Questionnaire' then
    result := newQuestionnaire()
  else if name = 'QuestionnaireResponse.group' then
    result := newQuestionnaireResponseGroup()
  else if name = 'QuestionnaireResponse.group.question' then
    result := newQuestionnaireResponseGroupQuestion()
  else if name = 'QuestionnaireResponse.group.question.answer' then
    result := newQuestionnaireResponseGroupQuestionAnswer()
  else if name = 'QuestionnaireResponse' then
    result := newQuestionnaireResponse()
  else if name = 'ReferralRequest' then
    result := newReferralRequest()
  else if name = 'RelatedPerson' then
    result := newRelatedPerson()
  else if name = 'RiskAssessment.prediction' then
    result := newRiskAssessmentPrediction()
  else if name = 'RiskAssessment' then
    result := newRiskAssessment()
  else if name = 'Schedule' then
    result := newSchedule()
  else if name = 'SearchParameter.contact' then
    result := newSearchParameterContact()
  else if name = 'SearchParameter' then
    result := newSearchParameter()
  else if name = 'Slot' then
    result := newSlot()
  else if name = 'Specimen.collection' then
    result := newSpecimenCollection()
  else if name = 'Specimen.treatment' then
    result := newSpecimenTreatment()
  else if name = 'Specimen.container' then
    result := newSpecimenContainer()
  else if name = 'Specimen' then
    result := newSpecimen()
  else if name = 'StructureDefinition.contact' then
    result := newStructureDefinitionContact()
  else if name = 'StructureDefinition.mapping' then
    result := newStructureDefinitionMapping()
  else if name = 'StructureDefinition.snapshot' then
    result := newStructureDefinitionSnapshot()
  else if name = 'StructureDefinition.differential' then
    result := newStructureDefinitionDifferential()
  else if name = 'StructureDefinition' then
    result := newStructureDefinition()
  else if name = 'Subscription.channel' then
    result := newSubscriptionChannel()
  else if name = 'Subscription' then
    result := newSubscription()
  else if name = 'Substance.instance' then
    result := newSubstanceInstance()
  else if name = 'Substance.ingredient' then
    result := newSubstanceIngredient()
  else if name = 'Substance' then
    result := newSubstance()
  else if name = 'SupplyDelivery' then
    result := newSupplyDelivery()
  else if name = 'SupplyRequest.when' then
    result := newSupplyRequestWhen()
  else if name = 'SupplyRequest' then
    result := newSupplyRequest()
  else if name = 'TestScript.contact' then
    result := newTestScriptContact()
  else if name = 'TestScript.metadata' then
    result := newTestScriptMetadata()
  else if name = 'TestScript.metadata.link' then
    result := newTestScriptMetadataLink()
  else if name = 'TestScript.metadata.capability' then
    result := newTestScriptMetadataCapability()
  else if name = 'TestScript.fixture' then
    result := newTestScriptFixture()
  else if name = 'TestScript.variable' then
    result := newTestScriptVariable()
  else if name = 'TestScript.setup' then
    result := newTestScriptSetup()
  else if name = 'TestScript.setup.action' then
    result := newTestScriptSetupAction()
  else if name = 'TestScript.setup.action.operation' then
    result := newTestScriptSetupActionOperation()
  else if name = 'TestScript.setup.action.operation.requestHeader' then
    result := newTestScriptSetupActionOperationRequestHeader()
  else if name = 'TestScript.setup.action.assert' then
    result := newTestScriptSetupActionAssert()
  else if name = 'TestScript.test' then
    result := newTestScriptTest()
  else if name = 'TestScript.test.action' then
    result := newTestScriptTestAction()
  else if name = 'TestScript.teardown' then
    result := newTestScriptTeardown()
  else if name = 'TestScript.teardown.action' then
    result := newTestScriptTeardownAction()
  else if name = 'TestScript' then
    result := newTestScript()
  else if name = 'ValueSet.contact' then
    result := newValueSetContact()
  else if name = 'ValueSet.codeSystem' then
    result := newValueSetCodeSystem()
  else if name = 'ValueSet.codeSystem.concept' then
    result := newValueSetCodeSystemConcept()
  else if name = 'ValueSet.codeSystem.concept.designation' then
    result := newValueSetCodeSystemConceptDesignation()
  else if name = 'ValueSet.compose' then
    result := newValueSetCompose()
  else if name = 'ValueSet.compose.include' then
    result := newValueSetComposeInclude()
  else if name = 'ValueSet.compose.include.concept' then
    result := newValueSetComposeIncludeConcept()
  else if name = 'ValueSet.compose.include.filter' then
    result := newValueSetComposeIncludeFilter()
  else if name = 'ValueSet.expansion' then
    result := newValueSetExpansion()
  else if name = 'ValueSet.expansion.parameter' then
    result := newValueSetExpansionParameter()
  else if name = 'ValueSet.expansion.contains' then
    result := newValueSetExpansionContains()
  else if name = 'ValueSet' then
    result := newValueSet()
  else if name = 'VisionPrescription.dispense' then
    result := newVisionPrescriptionDispense()
  else if name = 'VisionPrescription' then
    result := newVisionPrescription()
  else
    result := nil;
end;


end.

